/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * <p/>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p/>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

(function(/*! Brunch !*/) {
  'use strict';

  var globals = typeof window !== 'undefined' ? window : global;
  if (typeof globals.require === 'function') return;

  var modules = {};
  var cache = {};

  var has = function(object, name) {
    return ({}).hasOwnProperty.call(object, name);
  };

  var expand = function(root, name) {
    var results = [], parts, part;
    if (/^\.\.?(\/|$)/.test(name)) {
      parts = [root, name].join('/').split('/');
    } else {
      parts = name.split('/');
    }
    for (var i = 0, length = parts.length; i < length; i++) {
      part = parts[i];
      if (part === '..') {
        results.pop();
      } else if (part !== '.' && part !== '') {
        results.push(part);
      }
    }
    return results.join('/');
  };

  var dirname = function(path) {
    return path.split('/').slice(0, -1).join('/');
  };

  var localRequire = function(path) {
    return function(name) {
      var dir = dirname(path);
      var absolute = expand(dir, name);
      return globals.require(absolute, path);
    };
  };

  var initModule = function(name, definition) {
    var module = {id: name, exports: {}};
    cache[name] = module;
    definition(module.exports, localRequire(name), module);
    return module.exports;
  };

  var require = function(name, loaderPath) {
    var path = expand(name, '.');
    if (loaderPath == null) loaderPath = '/';

    if (has(cache, path)) return cache[path].exports;
    if (has(modules, path)) return initModule(path, modules[path]);

    var dirIndex = expand(path, './index');
    if (has(cache, dirIndex)) return cache[dirIndex].exports;
    if (has(modules, dirIndex)) return initModule(dirIndex, modules[dirIndex]);

    throw new Error('Cannot find module "' + name + '" from '+ '"' + loaderPath + '"');
  };

  var define = function(bundle, fn) {
    if (typeof bundle === 'object') {
      for (var key in bundle) {
        if (has(bundle, key)) {
          modules[key] = bundle[key];
        }
      }
    } else {
      modules[bundle] = fn;
    }
  };

  var list = function() {
    var result = [];
    for (var item in modules) {
      if (has(modules, item)) {
        result.push(item);
      }
    }
    return result;
  };

  globals.require = require;
  globals.require.define = define;
  globals.require.register = define;
  globals.require.list = list;
  globals.require.brunch = true;
})();
require.register("app", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// Application bootstrapper
require('utils/ember_reopen');
var stringUtils = require('utils/string_utils');

module.exports = Em.Application.create({
  name: 'Ambari Web',
  rootElement: '#wrapper',

  store: DS.Store.create({
    revision: 4,
    adapter: DS.FixtureAdapter.create({
      simulateRemoteResponse: false
    }),
    typeMaps: {},
    recordCache: []
  }),
  isAdmin: false,
  isOperator: false,

  /**
   * state of stack upgrade process
   * states:
   *  - INIT
   *  - PENDING
   *  - IN_PROGRESS
   *  - HOLDING
   *  - COMPLETED
   * @type {String}
   */
  upgradeState: 'INIT',

  /**
   * flag is true when upgrade process is running
   * @returns {boolean}
   */
  upgradeInProgress: function() {
    return ["IN_PROGRESS"].contains(this.get('upgradeState'));
  }.property('upgradeState'),

  /**
   * flag is true when upgrade process is waiting for user action
   * to procced, retry, perform manual steps etc.
   * @returns {boolean}
   */
  upgradeHolding: function() {
    return this.get('upgradeState').contains("HOLDING");
  }.property('upgradeState'),

  /**
   * RU is running
   * @type {boolean}
   */
  upgradeIsRunning: function() {
    return this.get('upgradeInProgress') || this.get('upgradeHolding');
  }.property('upgradeInProgress', 'upgradeHolding'),

  /**
   * compute user access rights by permission type
   * types:
   *  - ADMIN
   *  - MANAGER
   *  - OPERATOR
   *  - ONLY_ADMIN
   * prefix "upgrade_" mean that element will not be unconditionally blocked while stack upgrade running
   * @param type {string}
   * @return {boolean}
   */
  isAccessible: function (type) {
    if (!App.get('supports.opsDuringRollingUpgrade') && !['INIT', 'COMPLETED'].contains(this.get('upgradeState')) && !type.contains('upgrade_')) {
      return false;
    }

    if (type.contains('upgrade_')) {
      //slice off "upgrade_" prefix to have actual permission type
      type = type.slice(8);
    }

    switch (type) {
      case 'ADMIN':
        return this.get('isAdmin');
      case 'NON_ADMIN':
        return !this.get('isAdmin');
      case 'MANAGER':
        return this.get('isAdmin') || this.get('isOperator');
      case 'OPERATOR':
        return this.get('isOperator');
      case 'ONLY_ADMIN':
        return this.get('isAdmin') && !this.get('isOperator');
      default:
        return false;
    }
  },

  isStackServicesLoaded: false,
  /**
   * return url prefix with number value of version of HDP stack
   */
  stackVersionURL: function () {
    return '/stacks/{0}/versions/{1}'.format(this.get('currentStackName') || 'HDP', this.get('currentStackVersionNumber'));
  }.property('currentStackName','currentStackVersionNumber'),

  falconServerURL: function () {
    var falconService = this.Service.find().findProperty('serviceName', 'FALCON');
    if (falconService) {
      return falconService.get('hostComponents').findProperty('componentName', 'FALCON_SERVER').get('hostName');
    }
    return '';
  }.property().volatile(),

  /* Determine if Application Timeline Service supports Kerberization.
   * Because this value is retrieved from the cardinality of the component, it is safe to keep in app.js
   * since its value will not change during the lifetime of the application.
   */
  doesATSSupportKerberos: function() {
    var YARNService = App.StackServiceComponent.find().filterProperty('serviceName', 'YARN');
    if (YARNService.length) {
      var ATS = App.StackServiceComponent.find().findProperty('componentName', 'APP_TIMELINE_SERVER');
      return (!!ATS && !!ATS.get('minToInstall'));
    }
    return false;
  }.property('router.clusterController.isLoaded'),

  clusterName: null,
  clockDistance: null, // server clock - client clock
  currentStackVersion: '',
  currentStackName: function() {
    return Em.get((this.get('currentStackVersion') || this.get('defaultStackVersion')).match(/(.+)-\d.+/), '1');
  }.property('currentStackVersion', 'defaultStackVersion'),

  /**
   * true if cluster has only 1 host
   * for now is used to disable move/HA actions
   * @type {boolean}
   */
  isSingleNode: function() {
    return this.get('allHostNames.length') === 1;
  }.property('allHostNames.length'),

  allHostNames: [],

  currentStackVersionNumber: function () {
    var regExp = new RegExp(this.get('currentStackName') + '-');
    return (this.get('currentStackVersion') || this.get('defaultStackVersion')).replace(regExp, '');
  }.property('currentStackVersion', 'defaultStackVersion', 'currentStackName'),

  isHadoop22Stack: function () {
    return (stringUtils.compareVersions(this.get('currentStackVersionNumber'), "2.2") > -1);
  }.property('currentStackVersionNumber'),

  /**
   * Determines if current stack is 2.0.*
   * @type {boolean}
   */
  isHadoop20Stack: function () {
    return (stringUtils.compareVersions(this.get('currentStackVersionNumber'), "2.1") == -1 && stringUtils.compareVersions(this.get('currentStackVersionNumber'), "2.0") > -1);
  }.property('currentStackVersionNumber'),

  isHadoopWindowsStack: function() {
    return this.get('currentStackName') == "HDPWIN";
  }.property('currentStackName'),

  /**
   * If NameNode High Availability is enabled
   * Based on <code>clusterStatus.isInstalled</code>, stack version, <code>SNameNode</code> availability
   *
   * @type {bool}
   */
  isHaEnabled: function () {
    var isHDFSInstalled = App.Service.find().findProperty('serviceName','HDFS');
    return !!isHDFSInstalled && !this.HostComponent.find().someProperty('componentName', 'SECONDARY_NAMENODE');
  }.property('router.clusterController.isLoaded', 'router.clusterController.dataLoadList.serviceMetrics'),

  /**
   * If ResourceManager High Availability is enabled
   * Based on number of ResourceManager components host components installed
   *
   * @type {bool}
   */
  isRMHaEnabled: function () {
    var result = false;
    var rmStackComponent = App.StackServiceComponent.find().findProperty('componentName','RESOURCEMANAGER');
    if (rmStackComponent && rmStackComponent.get('isMultipleAllowed')) {
      result = this.HostComponent.find().filterProperty('componentName', 'RESOURCEMANAGER').length > 1;
    }
    return result;
  }.property('router.clusterController.isLoaded', 'isStackServicesLoaded'),

  /**
   * Object with utility functions for list of service names with similar behavior
   */
  services: Em.Object.create({
    all: function () {
      return App.StackService.find().mapProperty('serviceName');
    }.property('App.router.clusterController.isLoaded'),

    clientOnly: function () {
      return App.StackService.find().filterProperty('isClientOnlyService').mapProperty('serviceName');
    }.property('App.router.clusterController.isLoaded'),

    hasClient: function () {
      return App.StackService.find().filterProperty('hasClient').mapProperty('serviceName');
    }.property('App.router.clusterController.isLoaded'),

    hasMaster: function () {
      return App.StackService.find().filterProperty('hasMaster').mapProperty('serviceName');
    }.property('App.router.clusterController.isLoaded'),

    hasSlave: function () {
      return App.StackService.find().filterProperty('hasSlave').mapProperty('serviceName');
    }.property('App.router.clusterController.isLoaded'),

    noConfigTypes: function () {
      return App.StackService.find().filterProperty('isNoConfigTypes').mapProperty('serviceName');
    }.property('App.router.clusterController.isLoaded'),

    monitoring: function () {
      return App.StackService.find().filterProperty('isMonitoringService').mapProperty('serviceName');
    }.property('App.router.clusterController.isLoaded'),

    hostMetrics: function () {
      return App.StackService.find().filterProperty('isHostMetricsService').mapProperty('serviceName');
    }.property('App.router.clusterController.isLoaded'),

    serviceMetrics: function () {
      return App.StackService.find().filterProperty('isServiceMetricsService').mapProperty('serviceName');
    }.property('App.router.clusterController.isLoaded'),

    supportsServiceCheck: function() {
      return App.StackService.find().filterProperty('serviceCheckSupported').mapProperty('serviceName');
    }.property('App.router.clusterController.isLoaded')
  }),

  /**
   * List of components with allowed action for them
   * @type {Em.Object}
   */
  components: Em.Object.create({
    allComponents: function () {
      return App.StackServiceComponent.find().mapProperty('componentName')
    }.property('App.router.clusterController.isLoaded'),

    reassignable: function () {
      return App.StackServiceComponent.find().filterProperty('isReassignable').mapProperty('componentName')
    }.property('App.router.clusterController.isLoaded'),

    restartable: function () {
      return App.StackServiceComponent.find().filterProperty('isRestartable').mapProperty('componentName')
    }.property('App.router.clusterController.isLoaded'),

    deletable: function () {
      return App.StackServiceComponent.find().filterProperty('isDeletable').mapProperty('componentName')
    }.property('App.router.clusterController.isLoaded'),

    rollinRestartAllowed: function () {
      return App.StackServiceComponent.find().filterProperty('isRollinRestartAllowed').mapProperty('componentName')
    }.property('App.router.clusterController.isLoaded'),

    decommissionAllowed: function () {
      return App.StackServiceComponent.find().filterProperty('isDecommissionAllowed').mapProperty('componentName')
    }.property('App.router.clusterController.isLoaded'),

    refreshConfigsAllowed: function () {
      return App.StackServiceComponent.find().filterProperty('isRefreshConfigsAllowed').mapProperty('componentName')
    }.property('App.router.clusterController.isLoaded'),

    addableToHost: function () {
      return App.StackServiceComponent.find().filterProperty('isAddableToHost').mapProperty('componentName')
    }.property('App.router.clusterController.isLoaded'),

    addableMasterInstallerWizard: function () {
      return App.StackServiceComponent.find().filterProperty('isMasterAddableInstallerWizard').filterProperty('showAddBtnInInstall').mapProperty('componentName')
    }.property('App.router.clusterController.isLoaded'),

    multipleMasters: function () {
      return App.StackServiceComponent.find().filterProperty('isMasterWithMultipleInstances').mapProperty('componentName')
    }.property('App.router.clusterController.isLoaded'),

    slaves: function () {
      return App.StackServiceComponent.find().filterProperty('isSlave').mapProperty('componentName')
    }.property('App.router.clusterController.isLoaded'),

    masters: function () {
      return App.StackServiceComponent.find().filterProperty('isMaster').mapProperty('componentName')
    }.property('App.router.clusterController.isLoaded'),

    clients: function () {
      return App.StackServiceComponent.find().filterProperty('isClient').mapProperty('componentName')
    }.property('App.router.clusterController.isLoaded')
  })
});

});

require.register("config", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.version = '2.0.0'; // filled out by set-ambari-version.sh script
App.testMode = (location.port == '3333'); // test mode is automatically enabled if running on brunch server
App.testModeDelayForActions = 10000;
App.skipBootstrap = false;
App.alwaysGoToInstaller = false;
App.testEnableSecurity = true; // By default enable security is tested; turning it false tests disable security
App.testNameNodeHA = true;
App.apiPrefix = '/api/v1';
App.defaultStackVersion = 'HDP-2.2';
App.defaultWindowsStackVersion = 'HDPWIN-2.1';

App.defaultJavaHome = '/usr/jdk/jdk1.6.0_31';
App.timeout = 180000; // default AJAX timeout
App.maxRetries = 3; // max number of retries for certain AJAX calls
App.sessionKeepAliveInterval  = 60000;
App.bgOperationsUpdateInterval = 6000;
App.componentsUpdateInterval = 6000;
App.contentUpdateInterval = 15000;
App.hostStatusCountersUpdateInterval = 10000;
App.alertDefinitionsUpdateInterval = 10000;
App.alertInstancesUpdateInterval = 10000;
App.alertGroupsUpdateInterval = 10000;
App.pageReloadTime=3600000;
App.singleNodeInstall = true;
App.singleNodeAlias = document.location.hostname;
App.minDiskSpace = 2.0; // minimum disk space required for '/' for each host before install, unit GB
App.minDiskSpaceUsrLib = 1.0; // minimum disk space for '/usr/lib' for each host before install, unit GB
App.healthIconClassGreen = 'icon-ok-sign'; // bootstrap icon class for healthy/started service/host/host-component
App.healthIconClassRed = 'icon-warning-sign'; // bootstrap icon class for master down/stopped service/host/host-component
App.healthIconClassOrange = 'icon-minus-sign'; // bootstrap icon class for slave down/decommissioned host/host-component
App.healthIconClassYellow = 'icon-question-sign'; // bootstrap icon class for heartbeat lost service/host/host-component
App.isManagedMySQLForHiveEnabled = false;
App.isStormMetricsSupported = true;
App.healthStatusRed = '#ff0000';
App.healthStatusGreen = '#5AB400';
App.healthStatusOrange = '#FF8E00';

App.stackVersionsAvailable = true;

// experimental features are automatically enabled if running on brunch server
App.enableExperimental = false;

App.supports = {
  preUpgradeCheck: true,
  stackUpgrade: true,
  displayOlderVersions: false,
  autoRollbackHA: false,
  alwaysEnableManagedMySQLForHive: false,
  preKerberizeCheck: false,
  automatedKerberos: true,
  customizeAgentUserAccount: false,
  installGanglia: false,
  opsDuringRollingUpgrade: false
};

if (App.enableExperimental) {
  for (var support in App.supports) {
    App.supports[support] = true;
  }
}

// this is to make sure that IE does not cache data when making AJAX calls to the server
if (!$.mocho) {
  $.ajaxSetup({
    cache: false,
    headers: {"X-Requested-By": "X-Requested-By"}
  });
}

/**
 * Test Mode values
 */
App.test_hostname = 'hostname';

});

require.register("controllers", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


// load all controllers here

require('controllers/application');
require('controllers/login_controller');
require('controllers/wizard');
require('controllers/installer');
require('controllers/global/background_operations_controller');
require('controllers/main');
require('controllers/main/dashboard');
require('controllers/main/dashboard/config_history_controller');
require('controllers/main/admin');
require('controllers/main/admin/highAvailability_controller');
require('controllers/main/admin/highAvailability/nameNode/wizard_controller');
require('controllers/main/admin/highAvailability/progress_controller');
require('controllers/main/admin/highAvailability/progress_popup_controller');
require('controllers/main/admin/highAvailability/nameNode/rollback_controller');
require('controllers/main/admin/highAvailability/nameNode/step1_controller');
require('controllers/main/admin/highAvailability/nameNode/step2_controller');
require('controllers/main/admin/highAvailability/nameNode/step3_controller');
require('controllers/main/admin/highAvailability/nameNode/step4_controller');
require('controllers/main/admin/highAvailability/nameNode/step5_controller');
require('controllers/main/admin/highAvailability/nameNode/step6_controller');
require('controllers/main/admin/highAvailability/nameNode/step7_controller');
require('controllers/main/admin/highAvailability/nameNode/step8_controller');
require('controllers/main/admin/highAvailability/nameNode/step9_controller');
require('controllers/main/admin/highAvailability/nameNode/rollbackHA/step1_controller');
require('controllers/main/admin/highAvailability/nameNode/rollbackHA/step2_controller');
require('controllers/main/admin/highAvailability/nameNode/rollbackHA/step3_controller');
require('controllers/main/admin/highAvailability/nameNode/rollbackHA/rollback_wizard_controller');
require('controllers/main/admin/highAvailability/resourceManager/wizard_controller');
require('controllers/main/admin/highAvailability/resourceManager/step1_controller');
require('controllers/main/admin/highAvailability/resourceManager/step2_controller');
require('controllers/main/admin/highAvailability/resourceManager/step3_controller');
require('controllers/main/admin/highAvailability/resourceManager/step4_controller');
require('controllers/main/admin/stack_and_upgrade_controller');
require('controllers/main/admin/serviceAccounts_controller');
require('controllers/main/admin/advanced');
require('utils/polling');
require('controllers/main/admin/kerberos');
require('controllers/main/admin/kerberos/wizard_controller');
require('controllers/main/admin/kerberos/disable_controller');
require('controllers/main/admin/kerberos/progress_controller');
require('controllers/main/admin/kerberos/step1_controller');
require('controllers/main/admin/kerberos/step2_controller');
require('controllers/main/admin/kerberos/step3_controller');
require('controllers/main/admin/kerberos/step4_controller');
require('controllers/main/admin/kerberos/step5_controller');
require('controllers/main/admin/kerberos/step6_controller');
require('controllers/main/admin/kerberos/step7_controller');
require('controllers/main/admin/security');
require('controllers/main/admin/security/security_progress_controller');
require('controllers/main/admin/security/disable');
require('controllers/main/admin/security/add/addSecurity_controller');
require('controllers/main/admin/security/add/step1');
require('controllers/main/admin/security/add/step2');
require('controllers/main/admin/security/add/step3');
require('controllers/main/admin/security/add/step4');
require('controllers/main/admin/authentication');
require('controllers/main/alert_definitions_controller');
require('controllers/main/alerts/alert_definitions_actions_controller');
require('controllers/main/alerts/add_alert_definition/add_alert_definition_controller');
require('controllers/main/alerts/add_alert_definition/step1_controller');
require('controllers/main/alerts/add_alert_definition/step2_controller');
require('controllers/main/alerts/add_alert_definition/step3_controller');
require('controllers/main/alerts/definition_details_controller');
require('controllers/main/alerts/definition_configs_controller');
require('controllers/main/alerts/alert_instances_controller');
require('controllers/main/alerts/manage_alert_groups_controller');
require('controllers/main/alerts/manage_alert_notifications_controller');
require('controllers/main/service');
require('controllers/main/service/item');
require('controllers/main/service/info/summary');
require('controllers/main/service/info/configs');
require('controllers/main/service/info/audit');
require('controllers/main/service/add_controller');
require('controllers/main/service/reassign_controller');
require('controllers/main/service/reassign/step1_controller');
require('controllers/main/service/reassign/step2_controller');
require('controllers/main/service/reassign/step3_controller');
require('controllers/main/service/reassign/step4_controller');
require('controllers/main/service/reassign/step5_controller');
require('controllers/main/service/reassign/step6_controller');
require('controllers/main/service/manage_config_groups_controller');
require('controllers/main/host');
require('controllers/main/host/details');
require('controllers/main/host/configs_service');
require('controllers/main/host/add_controller');
require('controllers/main/host/addHost/step4_controller');
require('controllers/main/host/host_alerts_controller');
require('controllers/main/charts');
require('controllers/main/charts/heatmap_metrics/heatmap_metric');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_processrun');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_diskspaceused');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_cpuWaitIO');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_memoryused');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_dfs');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_bytesread');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_byteswritten');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_gctime');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_memHeapUsed');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_yarn');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_yarn_gctime');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_yarn_memHeapUsed');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_yarn_ResourceUsed');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_hbase');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_readrequest');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_writerequest');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_compactionqueue');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_regions');
require('controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_memstoresize');
require('controllers/main/charts/heatmap');
require('controllers/main/views_controller');
require('controllers/main/views/details_controller');
require('controllers/wizard/slave_component_groups_controller');
require('controllers/wizard/step0_controller');
require('controllers/wizard/step1_controller');
require('controllers/wizard/step2_controller');
require('controllers/wizard/step3_controller');
require('controllers/wizard/step4_controller');
require('controllers/wizard/step5_controller');
require('controllers/wizard/step6_controller');
require('controllers/wizard/step7_controller');
require('controllers/wizard/step8_controller');
require('controllers/wizard/step9_controller');
require('controllers/wizard/step10_controller');
require('controllers/global/cluster_controller');
require('controllers/global/update_controller');
require('controllers/global/configuration_controller');
require('controllers/main/service/reassign/step7_controller');

});

require.register("controllers/application", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.ApplicationController = Em.Controller.extend(App.UserPref, {

  name: 'applicationController',

  isPollerRunning: false,

  clusterName: function () {
    return (App.router.get('clusterController.clusterName') || 'My Cluster');
  }.property('App.router.clusterController.clusterName'),

  /**
   * set ambari server version from installerController or mainController, making sure version shown up all the time
   */
  ambariVersion: function () {
    return (App.router.get('installerController.ambariServerVersion') || App.router.get('mainController.ambariServerVersion') || Em.I18n.t('common.notAvailable'));
  }.property('App.router.installerController.ambariServerVersion', 'App.router.mainController.ambariServerVersion'),

  clusterDisplayName: function () {
    var name = this.get('clusterName');
    return name.length > 13 ? name.substr(0, 10) + "..." : name;
  }.property('clusterName'),

  isClusterDataLoaded: function() {
    return App.router.get('clusterController.isLoaded') && App.router.get('loggedIn');
  }.property('App.router.clusterController.isLoaded','App.router.loggedIn'),

  isExistingClusterDataLoaded: function () {
    return App.router.get('clusterInstallCompleted') && this.get('isClusterDataLoaded');
  }.property('App.router.clusterInstallCompleted', 'isClusterDataLoaded'),

  init: function(){
    this._super();
  },

  startKeepAlivePoller: function() {
    if (!this.get('isPollerRunning')) {
     this.set('isPollerRunning',true);
      App.updater.run(this, 'getStack', 'isPollerRunning', App.sessionKeepAliveInterval);
    }
  },

  getStack: function(callback) {
    App.ajax.send({
      name: 'router.login.clusters',
      sender: this,
      callback: callback
    });
  },

  dataLoading: function () {
    var dfd = $.Deferred();
    var self = this;
    this.getUserPref(this.persistKey()).complete(function () {
      var curPref = self.get('currentPrefObject');
      self.set('currentPrefObject', null);
      dfd.resolve(curPref);
    });
    return dfd.promise();
  },
  persistKey: function (loginName) {
    if (App.get('testMode')) {
      return 'admin_settings_show_bg';
    }
    if (!loginName)
      loginName = App.router.get('loginName');
    return 'admin-settings-show-bg-' + loginName;
  },
  currentPrefObject: null,

  getUserPrefSuccessCallback: function (response, request, data) {
    if (response != null) {
      console.log('Got persist value from server with key ' + data.key + '. Value is: ' + response);
      this.set('currentPrefObject', response);
      return response;
    }
  },
  getUserPrefErrorCallback: function (request, ajaxOptions, error) {
    // this user is first time login
    if (request.status == 404) {
      console.log('Persist did NOT find the key');
      this.set('currentPrefObject', true);
      this.postUserPref(this.persistKey(), true);
      return true;
    }
  },

  goToAdminView: function () {
    App.router.route("adminView");
  },

  showSettingsPopup: function() {
    // Settings only for admins
    if (!App.isAccessible('upgrade_ADMIN')) return;

    var self = this;
    var curValue = null;
    this.dataLoading().done(function (initValue) {
      App.ModalPopup.show({
        header: Em.I18n.t('common.userSettings'),
        bodyClass: Em.View.extend({
          templateName: require('templates/common/settings'),
          isNotShowBgChecked: !initValue,
          updateValue: function () {
            curValue = !this.get('isNotShowBgChecked');
          }.observes('isNotShowBgChecked')
        }),
        primary: Em.I18n.t('common.save'),
        onPrimary: function() {
          if (curValue == null) {
            curValue = initValue;
          }
          var key = self.persistKey();
          if (!App.get('testMode')) {
            self.postUserPref(key, curValue);
          }
          this.hide();
        }
      })
    });
  },

  showAboutPopup: function() {

    var self = this;
    App.ModalPopup.show({
      header: Em.I18n.t('common.aboutAmbari'),
      secondary: false,
      bodyClass: Em.View.extend({
        templateName: require('templates/common/about'),
        ambariVersion: this.get('ambariVersion')
      })
    });
  }

});
});

require.register("controllers/global/background_operations_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.BackgroundOperationsController = Em.Controller.extend({
  name: 'backgroundOperationsController',

  /**
   * Whether we need to refresh background operations or not
   */
  isWorking : false,

  allOperationsCount : 0,

  /**
   * For host component popup
   */
  services:[],
  serviceTimestamp: null,

  /**
   * Number of operation to load
   */
  operationsCount: 10,
  /**
   * Possible levels:
   * REQUESTS_LIST
   * HOSTS_LIST
   * TASKS_LIST
   * TASK_DETAILS
   */
  levelInfo: Em.Object.create({
    name: 'REQUESTS_LIST',
    requestId: null,
    taskId: null
  }),

  /**
   * Start polling, when <code>isWorking</code> become true
   */
  startPolling: function(){
    if(this.get('isWorking')){
      this.requestMostRecent();
      App.updater.run(this, 'requestMostRecent', 'isWorking', App.bgOperationsUpdateInterval);
    }
  }.observes('isWorking'),

  /**
   * Get requests data from server
   * @param callback
   */
  requestMostRecent: function (callback) {
    var queryParams = this.getQueryParams();
    App.ajax.send({
      'name': queryParams.name,
      'sender': this,
      'success': queryParams.successCallback,
      'callback': callback,
      'data': queryParams.data
    });
    return !this.isInitLoading();
  },

  /**
   * indicate whether data for current level has already been loaded or not
   * @return {Boolean}
   */
  isInitLoading: function () {
    var levelInfo = this.get('levelInfo');
    var request = this.get('services').findProperty('id', levelInfo.get('requestId'));

    if (levelInfo.get('name') === 'HOSTS_LIST') {
      return !!(request && App.isEmptyObject(request.get('hostsMap')));
    }
    return false;
  },
  /**
   * construct params of ajax query regarding displayed level
   */
  getQueryParams: function () {
    var levelInfo = this.get('levelInfo');
    var count = this.get('operationsCount');
    var result = {
      name: 'background_operations.get_most_recent',
      successCallback: 'callBackForMostRecent',
      data: {
        'operationsCount': count
      }
    };
    if (levelInfo.get('name') === 'TASK_DETAILS' && !App.get('testMode')) {
      result.name = 'background_operations.get_by_task';
      result.successCallback = 'callBackFilteredByTask';
      result.data = {
        'taskId': levelInfo.get('taskId'),
        'requestId': levelInfo.get('requestId')
      };
    } else if (levelInfo.get('name') === 'TASKS_LIST' || levelInfo.get('name') === 'HOSTS_LIST') {
      result.name = 'background_operations.get_by_request';
      result.successCallback = 'callBackFilteredByRequest';
      result.data = {
        'requestId': levelInfo.get('requestId')
      };
    }
    return result;
  },

  /**
   * Push hosts and their tasks to request
   * @param data
   * @param ajaxQuery
   * @param params
   */
  callBackFilteredByRequest: function (data, ajaxQuery, params) {
    var requestId = data.Requests.id;
    var requestInputs = data.Requests.inputs;
    var request = this.get('services').findProperty('id', requestId);
    var hostsMap = {};
    var previousTaskStatusMap = request.get('previousTaskStatusMap');
    var currentTaskStatusMap = {};
    data.tasks.forEach(function (task) {
      var host = hostsMap[task.Tasks.host_name];
      task.Tasks.request_id = requestId;
      task.Tasks.request_inputs = requestInputs;
      if (host) {
        host.logTasks.push(task);
        host.isModified = (host.isModified) ? true : previousTaskStatusMap[task.Tasks.id] !== task.Tasks.status;
      } else {
        hostsMap[task.Tasks.host_name] = {
          name: task.Tasks.host_name,
          publicName: task.Tasks.host_name,
          logTasks: [task],
          isModified: previousTaskStatusMap[task.Tasks.id] !== task.Tasks.status
        };
      }
      currentTaskStatusMap[task.Tasks.id] = task.Tasks.status;
    }, this);
    /**
     * sync up request progress with up to date progress of hosts on Host's list,
     * to avoid discrepancies while waiting for response with latest progress of request
     * after switching to operation's list
     */
    if (request.get('isRunning')) {
      request.set('progress', App.HostPopup.getProgress(data.tasks));
      request.set('status', App.HostPopup.getStatus(data.tasks)[0]);
      request.set('isRunning', (request.get('progress') !== 100));
    }
    request.set('previousTaskStatusMap', currentTaskStatusMap);
    request.set('hostsMap', hostsMap);
    this.set('serviceTimestamp', App.dateTime());
  },
  /**
   * Update task, with uploading two additional properties: stdout and stderr
   * @param data
   * @param ajaxQuery
   * @param params
   */
  callBackFilteredByTask: function (data, ajaxQuery, params) {
    var request = this.get('services').findProperty('id', data.Tasks.request_id);
    var host = request.get('hostsMap')[data.Tasks.host_name];
    var task = host.logTasks.findProperty('Tasks.id', data.Tasks.id);
    task.Tasks.status = data.Tasks.status;
    task.Tasks.stdout = data.Tasks.stdout;
    task.Tasks.stderr = data.Tasks.stderr;

    // Put some command information to task object
    task.Tasks.command = data.Tasks.command;
    task.Tasks.custom_command_name = data.Tasks.custom_command_name;
    task.Tasks.structured_out = data.Tasks.structured_out;

    task.Tasks.output_log = data.Tasks.output_log;
    task.Tasks.error_log = data.Tasks.error_log;
    this.set('serviceTimestamp', App.dateTime());
  },

  /**
   * returns true if it's upgrade equest
   * use this flag to exclude upgrade requests from bgo
   * @param {object} request
   * @returns {boolean}
   */
  isUpgradeRequest: function(request) {
    var context = Em.get(request, 'Requests.request_context');
    return context ? /(upgrading|downgrading)/.test(context.toLowerCase()) : false;
  },
  /**
   * Prepare, received from server, requests for host component popup
   * @param data
   */
  callBackForMostRecent: function (data) {
    var runningServices = 0;
    var currentRequestIds = [];
    var countIssued = this.get('operationsCount');
    var countGot = data.itemTotal;

    data.items.forEach(function (request) {
      if (this.isUpgradeRequest(request)) {
        return;
      }
      var rq = this.get("services").findProperty('id', request.Requests.id);
      var isRunning = this.isRequestRunning(request);
      var requestParams = this.parseRequestContext(request.Requests.request_context);
      this.assignScheduleId(request, requestParams);
      currentRequestIds.push(request.Requests.id);

      if (rq) {
        rq.set('progress', Math.floor(request.Requests.progress_percent));
        rq.set('status', request.Requests.request_status);
        rq.set('isRunning', isRunning);
        rq.set('startTime', request.Requests.start_time);
        rq.set('endTime', request.Requests.end_time);
      } else {
        rq = Em.Object.create({
          id: request.Requests.id,
          name: requestParams.requestContext,
          displayName: requestParams.requestContext,
          progress: Math.floor(request.Requests.progress_percent),
          status: request.Requests.request_status,
          isRunning: isRunning,
          hostsMap: {},
          tasks: [],
          startTime: request.Requests.start_time,
          endTime: request.Requests.end_time,
          dependentService: requestParams.dependentService,
          sourceRequestScheduleId: request.Requests.request_schedule && request.Requests.request_schedule.schedule_id,
          previousTaskStatusMap: {},
          contextCommand: requestParams.contextCommand
        });
        this.get("services").unshift(rq);
        //To sort DESC by request id
        this.set("services", this.get("services").sort( function(a,b) { return b.get('id') - a.get('id'); }));
      }
      runningServices += ~~isRunning;
    }, this);
    this.removeOldRequests(currentRequestIds);
    this.set("allOperationsCount", runningServices);
    this.set('isShowMoreAvailable', countGot >= countIssued);
    this.set('serviceTimestamp', App.dateTime());
  },
  isShowMoreAvailable: null,
  /**
   * remove old requests
   * as API returns 10, or  20 , or 30 ...etc latest request, the requests that absent in response should be removed
   * @param currentRequestIds
   */
  removeOldRequests: function (currentRequestIds) {
    this.get('services').forEach(function (service, index, services) {
      if (!currentRequestIds.contains(service.id)) {
        services.splice(index, 1);
      }
    });
  },
  /**
   * identify whether request is running by task counters
   * @param request
   * @return {Boolean}
   */
  isRequestRunning: function (request) {
    return (request.Requests.task_count -
      (request.Requests.aborted_task_count + request.Requests.completed_task_count + request.Requests.failed_task_count
        + request.Requests.timed_out_task_count - request.Requests.queued_task_count)) > 0;
  },
  /**
   * identify whether there is only one host in request
   * @param inputs
   * @return {Boolean}
   */
  isOneHost: function (inputs) {
    if (!inputs) {
      return false;
    }
    inputs = JSON.parse(inputs);
    if (inputs && inputs.included_hosts) {
      return inputs.included_hosts.split(',').length < 2;
    }
    return false
  },
  /**
   * assign schedule_id of request to null if it's Recommision operation
   * @param request
   * @param requestParams
   */
  assignScheduleId: function (request, requestParams) {
    var oneHost = this.isOneHost(request.Requests.inputs);
    if (request.Requests.request_schedule && oneHost && /Recommission/.test(requestParams.requestContext)) {
      request.Requests.request_schedule.schedule_id = null;
    }
  },

  /**
   * parse request context and if keyword "_PARSE_" is present then format it
   * @param requestContext
   * @return {Object}
   */
  parseRequestContext: function (requestContext) {
    var parsedRequestContext;
    var service;
    var contextCommand;
    if (requestContext) {
      if (requestContext.indexOf(App.BackgroundOperationsController.CommandContexts.PREFIX) !== -1) {
        var contextSplits = requestContext.split('.');
        contextCommand = contextSplits[1];
        service = contextSplits[2];
        switch(contextCommand){
        case "STOP":
        case "START":
          if (service === 'ALL_SERVICES') {
            parsedRequestContext = Em.I18n.t("requestInfo." + contextCommand.toLowerCase()).format(Em.I18n.t('common.allServices'));
          } else {
            parsedRequestContext = Em.I18n.t("requestInfo." + contextCommand.toLowerCase()).format(App.format.role(service));
          }
          break;
        case "ROLLING-RESTART":
          parsedRequestContext = Em.I18n.t("rollingrestart.rest.context").format(App.format.role(service), contextSplits[3], contextSplits[4]);
          break;
        }
      } else {
        parsedRequestContext = requestContext;
      }
    } else {
      parsedRequestContext = Em.I18n.t('requestInfo.unspecified');
    }
    return {
      requestContext: parsedRequestContext,
      dependentService: service,
      contextCommand: contextCommand
    }
  },

  popupView: null,

  /**
   * Onclick handler for background operations number located right to logo
   */
  showPopup: function(){
    // load the checkbox on footer first, then show popup.
    var self = this;
    App.router.get('applicationController').dataLoading().done(function (initValue) {
      App.updater.immediateRun('requestMostRecent');
      if(self.get('popupView') && App.HostPopup.get('isBackgroundOperations')){
        self.set ('popupView.isNotShowBgChecked', !initValue);
        self.set('popupView.isOpen', true);
        $(self.get('popupView.element')).appendTo('#wrapper');
      } else {
        self.set('popupView', App.HostPopup.initPopup("", self, true));
        self.set ('popupView.isNotShowBgChecked', !initValue);
      }
    });
  },

  /**
   * Called on logout
   */
  clear: function () {
    // set operations count to default value
    this.set('operationsCount', 10);
  }

});

/**
 * Each background operation has a context in which it operates.
 * Generally these contexts are fixed messages. However, we might
 * want to associate semantics to this context - like showing, disabling
 * buttons when certain operations are in progress.
 *
 * To make this possible we have command contexts where the context
 * is not a human readable string, but a pattern indicating the command
 * it is running. When UI shows these, they are translated into human
 * readable strings.
 *
 * General pattern of context names is "_PARSE_.{COMMAND}.{ID}[.{Additional-Data}...]"
 */
App.BackgroundOperationsController.CommandContexts = {
  PREFIX : "_PARSE_",
  /**
   * Stops all services
   */
  STOP_ALL_SERVICES : "_PARSE_.STOP.ALL_SERVICES",
  /**
   * Starts all services
   */
  START_ALL_SERVICES : "_PARSE_.START.ALL_SERVICES",
  /**
   * Starts service indicated by serviceID.
   * @param {String} serviceID Parameter {0}. Example: HDFS
   */
  START_SERVICE : "_PARSE_.START.{0}",
  /**
   * Stops service indicated by serviceID.
   * @param {String} serviceID Parameter {0}. Example: HDFS
   */
  STOP_SERVICE : "_PARSE_.STOP.{0}",
  /**
   * Performs rolling restart of componentID in batches.
   * This context is the batchNumber batch out of totalBatchCount batches.
   * @param {String} componentID Parameter {0}. Example "DATANODE"
   * @param {Number} batchNumber Parameter {1}. Batch number of this batch. Example 3.
   * @param {Number} totalBatchCount Parameter {2}. Total number of batches. Example 10.
   */
  ROLLING_RESTART : "_PARSE_.ROLLING-RESTART.{0}.{1}.{2}"
};
});

require.register("controllers/global/cluster_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

App.ClusterController = Em.Controller.extend({
  name: 'clusterController',
  isLoaded: false,
  ambariProperties: null,
  clusterDataLoadedPercent: 'width:0', // 0 to 1

  isGangliaUrlLoaded: false,

  /**
   * Provides the URL to use for Ganglia server. This URL
   * is helpful in populating links in UI.
   *
   * If null is returned, it means GANGLIA service is not installed.
   */
  gangliaUrl: null,

  clusterName: function () {
    return App.get('clusterName');
  }.property('App.clusterName'),

  updateLoadStatus: function (item) {
    var loadList = this.get('dataLoadList');
    var loaded = true;
    var numLoaded = 0;
    var loadListLength = 0;
    loadList.set(item, true);
    for (var i in loadList) {
      if (loadList.hasOwnProperty(i)) {
        loadListLength++;
        if (!loadList[i] && loaded) {
          loaded = false;
        }
      }
      // calculate the number of true
      if (loadList.hasOwnProperty(i) && loadList[i]) {
        numLoaded++;
      }
    }
    this.set('isLoaded', loaded);
    this.set('clusterDataLoadedPercent', 'width:' + (Math.floor(numLoaded / loadListLength * 100)).toString() + '%');
  },

  dataLoadList: Em.Object.create({
    'hosts': false,
    'serviceMetrics': false,
    'stackComponents': false,
    'services': false,
    'cluster': false,
    'clusterStatus': false,
    'racks': false,
    'componentConfigs': false,
    'componentsState': false,
    'rootService': false,
    'alertDefinitions': false,
    'securityStatus': false
  }),

  /**
   * load cluster name
   */
  loadClusterName: function (reload) {
    var dfd = $.Deferred();

    if (App.get('clusterName') && !reload) {
      App.set('clusterName', this.get('clusterName'));
      dfd.resolve();
    } else {
      App.ajax.send({
        name: 'cluster.load_cluster_name',
        sender: this,
        success: 'loadClusterNameSuccessCallback',
        error: 'loadClusterNameErrorCallback'
      }).complete(function () {
        if (!App.get('currentStackVersion')) {
          App.set('currentStackVersion', App.defaultStackVersion);
        }
        dfd.resolve();
      });
    }
    return dfd.promise()
  },

  loadClusterNameSuccessCallback: function (data) {
    if (data.items && data.items.length > 0) {
      App.set('clusterName', data.items[0].Clusters.cluster_name);
      App.set('currentStackVersion', data.items[0].Clusters.version);
    }
  },

  loadClusterNameErrorCallback: function (request, ajaxOptions, error) {
    console.log('failed on loading cluster name');
    this.set('isLoaded', true);
  },

  /**
   * load current server clock in milli-seconds
   */
  loadClientServerClockDistance: function () {
    var dfd = $.Deferred();
    this.getServerClock().done(function () {
      dfd.resolve();
    });
    return dfd.promise();
  },

  getServerClock: function () {
    return App.ajax.send({
      name: 'ambari.service.load_server_clock',
      sender: this,
      success: 'getServerClockSuccessCallback',
      error: 'getServerClockErrorCallback'
    });
  },
  getServerClockSuccessCallback: function (data) {
    var clientClock = new Date().getTime();
    var serverClock = (data.RootServiceComponents.server_clock).toString();
    serverClock = serverClock.length < 13 ? serverClock + '000' : serverClock;
    App.set('clockDistance', serverClock - clientClock);
    App.set('currentServerTime', parseInt(serverClock));
    console.log('loading ambari server clock distance');
  },
  getServerClockErrorCallback: function () {
    console.log('Cannot load ambari server clock');
  },

  getUrl: function (testUrl, url) {
    return (App.get('testMode')) ? testUrl : App.get('apiPrefix') + '/clusters/' + App.get('clusterName') + url;
  },

  setGangliaUrl: function () {
    if (App.get('testMode')) {
      this.set('gangliaUrl', 'http://gangliaserver/ganglia/?t=yes');
      this.set('isGangliaUrlLoaded', true);
    } else {
      // We want live data here
      var gangliaServer = App.HostComponent.find().findProperty('componentName', 'GANGLIA_SERVER');
      if (this.get('isLoaded') && gangliaServer) {
        this.set('isGangliaUrlLoaded', false);
        App.ajax.send({
          name: 'hosts.for_quick_links',
          sender: this,
          data: {
            clusterName: App.get('clusterName'),
            masterHosts: gangliaServer.get('hostName'),
            urlParams: ''
          },
          success: 'setGangliaUrlSuccessCallback'
        });
      }
    }
  }.observes('App.router.updateController.isUpdated', 'dataLoadList.hosts', 'gangliaWebProtocol', 'isLoaded'),

  setGangliaUrlSuccessCallback: function (response) {
    var url = null;
    if (response.items.length > 0) {
      url = this.get('gangliaWebProtocol') + "://" + (App.singleNodeInstall ? App.singleNodeAlias + ":42080" : response.items[0].Hosts.public_host_name) + "/ganglia";
    }
    this.set('gangliaUrl', url);
    this.set('isGangliaUrlLoaded', true);
  },

  gangliaWebProtocol: function () {
    var properties = this.get('ambariProperties');
    if (properties && properties.hasOwnProperty('ganglia.https') && properties['ganglia.https']) {
      return "https";
    } else {
      return "http";
    }
  }.property('ambariProperties'),

  isGangliaInstalled: function () {
    return !!App.Service.find().findProperty('serviceName', 'GANGLIA');
  }.property('App.router.updateController.isUpdated', 'dataLoadList.serviceMetrics'),

  /**
   *  load all data and update load status
   */
  loadClusterData: function () {
    var self = this;
    this.getAllHostNames();
    this.loadAmbariProperties();
    if (!App.get('clusterName')) {
      return;
    }

    if (this.get('isLoaded')) { // do not load data repeatedly
      App.router.get('mainController').startPolling();
      return;
    }

    var clusterUrl = this.getUrl('/data/clusters/cluster.json', '?fields=Clusters');
    var racksUrl = "/data/racks/racks.json";


    var hostsController = App.router.get('mainHostController');
    hostsController.set('isCountersUpdating', true);
    hostsController.updateStatusCounters();

    App.HttpClient.get(racksUrl, App.racksMapper, {
      complete: function (jqXHR, textStatus) {
        self.updateLoadStatus('racks');
      }
    }, function (jqXHR, textStatus) {
      self.updateLoadStatus('racks');
    });

    App.HttpClient.get(clusterUrl, App.clusterMapper, {
      complete: function (jqXHR, textStatus) {
        self.updateLoadStatus('cluster');
      }
    }, function (jqXHR, textStatus) {
      self.updateLoadStatus('cluster');
    });

    if (App.get('testMode')) {
      self.updateLoadStatus('clusterStatus');
    } else {
      App.clusterStatus.updateFromServer().complete(function () {
        self.updateLoadStatus('clusterStatus');
        if (App.get('supports.stackUpgrade')) {
          self.restoreUpgradeState();
        }
      });
    }

    /**
     * Order of loading:
     * 1. request for service components supported by stack
     * 2. load stack components to model
     * 3. request for services
     * 4. put services in cache
     * 5. request for hosts and host-components (single call)
     * 6. request for service metrics
     * 7. load host-components to model
     * 8. load hosts to model
     * 9. load services from cache with metrics to model
     * 10. update stale_configs of host-components (depends on App.supports.hostOverrides)
     * 11. load root service (Ambari)
     * 12. load alert definitions to model
     * 13. load unhealthy alert instances
     * 14. load security status
     */
    self.loadStackServiceComponents(function (data) {
      data.items.forEach(function (service) {
        service.StackServices.is_selected = true;
        service.StackServices.is_installed = false;
      }, self);
      App.stackServiceMapper.mapStackServices(data);
      App.config.setPreDefinedServiceConfigs(true);
      var updater = App.router.get('updateController');
      self.updateLoadStatus('stackComponents');
      updater.updateServices(function () {
        self.updateLoadStatus('services');
        //force clear filters  for hosts page to load all data
        App.db.setFilterConditions('mainHostController', null);

        updater.updateHost(function () {
          self.updateLoadStatus('hosts');
        });

        updater.updateServiceMetric(function () {

          updater.updateComponentConfig(function () {
            self.updateLoadStatus('componentConfigs');
          });

          updater.updateComponentsState(function () {
            self.updateLoadStatus('componentsState');
          });
          self.updateLoadStatus('serviceMetrics');

          updater.updateAlertGroups(function () {
            updater.updateAlertDefinitions(function() {
              updater.updateAlertDefinitionSummary(function () {
                updater.updateUnhealthyAlertInstances(function () {
                  self.updateLoadStatus('alertGroups');
                  self.updateLoadStatus('alertDefinitions');
                  self.updateLoadStatus('alertInstancesUnhealthy');
                });
              });
            });
          });
        });
      });
      self.loadRootService().done(function (data) {
        App.rootServiceMapper.map(data);
        self.updateLoadStatus('rootService');
      });
      // load security status
      App.router.get('mainAdminKerberosController').getSecurityStatus().always(function() {
        self.updateLoadStatus('securityStatus');
      });
    });
  },

  /**
   * restore upgrade status from local storage
   * and make call to get latest status from server
   */
  restoreUpgradeState: function () {
    var dbUpgradeState = App.db.get('MainAdminStackAndUpgrade', 'upgradeState');
    if (!Em.isNone(dbUpgradeState)) {
      App.set('upgradeState', dbUpgradeState);
    }
    App.router.get('mainAdminStackAndUpgradeController').initDBProperties();
    App.router.get('mainAdminStackAndUpgradeController').loadUpgradeData(true);
    App.router.get('mainAdminStackAndUpgradeController').loadStackVersionsToModel(true).done(function () {
      App.set('stackVersionsAvailable', App.StackVersion.find().content.length > 0);
    });
  },

  loadRootService: function () {
    return App.ajax.send({
      name: 'service.ambari',
      sender: this
    });
  },

  requestHosts: function (realUrl, callback) {
    var testHostUrl = '/data/hosts/HDP2/hosts.json';
    var url = this.getUrl(testHostUrl, realUrl);
    App.HttpClient.get(url, App.hostsMapper, {
      complete: callback
    }, callback)
  },

  /**
   *
   * @param callback
   */
  loadStackServiceComponents: function (callback) {
    var callbackObj = {
      loadStackServiceComponentsSuccess: callback
    };
    App.ajax.send({
      name: 'wizard.service_components',
      data: {
        stackUrl: App.get('stackVersionURL'),
        stackVersion: App.get('currentStackVersionNumber')
      },
      sender: callbackObj,
      success: 'loadStackServiceComponentsSuccess'
    });
  },

  loadAmbariProperties: function () {
    return App.ajax.send({
      name: 'ambari.service',
      sender: this,
      success: 'loadAmbariPropertiesSuccess',
      error: 'loadAmbariPropertiesError'
    });
  },

  loadAmbariPropertiesSuccess: function (data) {
    console.log('loading ambari properties');
    this.set('ambariProperties', data.RootServiceComponents.properties);
  },

  loadAmbariPropertiesError: function () {
    console.warn('can\'t get ambari properties');
  },

  updateClusterData: function () {
    var testUrl = '/data/clusters/HDP2/cluster.json';
    var clusterUrl = this.getUrl(testUrl, '?fields=Clusters');
    App.HttpClient.get(clusterUrl, App.clusterMapper, {
      complete: function () {
      }
    });
  },

  /**
   *
   * @returns {*|Transport|$.ajax|boolean|ServerResponse}
   */
  getAllHostNames: function () {
    return App.ajax.send({
      name: 'hosts.all',
      sender: this,
      success: 'getHostNamesSuccess',
      error: 'getHostNamesError'
    });
  },

  getHostNamesSuccess: function (data) {
    App.set("allHostNames", data.items.mapProperty("Hosts.host_name"));
  },

  getHostNamesError: function () {
    console.error('failed to load hostNames');
  },


  /**
   * puts kerberos admin credentials in the live cluster session
   * and resend ajax request
   * @param adminPrincipalValue
   * @param adminPasswordValue
   * @param ajaxOpt
   * @returns {$.ajax}
   */
  createKerberosAdminSession: function (adminPrincipalValue, adminPasswordValue, ajaxOpt) {
    return App.ajax.send({
      name: 'common.cluster.update',
      sender: this,
      data: {
        clusterName: App.get('clusterName'),
        data: [{
          session_attributes: {
            kerberos_admin: {principal: adminPrincipalValue, password: adminPasswordValue}
          }
        }]
      }
    }).success(function () {
      if (ajaxOpt) {
        $.ajax(ajaxOpt);
      }
    });
  },

  //TODO Replace this check with any other which is applicable to non-HDP stack
  /**
   * Check if HDP stack version is more or equal than 2.2.2 to determine if pluggable metrics for Storm are supported
   * @method checkDetailedRepoVersion
   * @returns {promise|*|promise|promise|HTMLElement|promise}
   */
  checkDetailedRepoVersion: function () {
    var dfd;
    var currentStackName = App.get('currentStackName');
    var currentStackVersionNumber = App.get('currentStackVersionNumber');
    if (currentStackName == 'HDP' && currentStackVersionNumber == '2.2') {
      dfd = App.ajax.send({
        name: 'cluster.load_detailed_repo_version',
        sender: this,
        success: 'checkDetailedRepoVersionSuccessCallback',
        error: 'checkDetailedRepoVersionErrorCallback'
      });
    } else {
      dfd = $.Deferred();
      App.set('isStormMetricsSupported', currentStackName != 'HDP' || stringUtils.compareVersions(currentStackVersionNumber, '2.2') == 1);
      dfd.resolve();
    }
    return dfd.promise();
  },

  checkDetailedRepoVersionSuccessCallback: function (data) {
    var items = data.items;
    var version;
    if (items && items.length) {
      var repoVersions = items[0].repository_versions;
      if (repoVersions && repoVersions.length) {
        version = Em.get(repoVersions[0], 'RepositoryVersions.repository_version');
      }
    }
    App.set('isStormMetricsSupported', stringUtils.compareVersions(version, '2.2.2') > -1 || !version);
  },
  checkDetailedRepoVersionErrorCallback: function () {
    App.set('isStormMetricsSupported', true);
  }
});

});

require.register("controllers/global/configuration_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ConfigurationController = Em.Controller.extend({
  name: 'configurationController',

  /**
   * get configs by tags
   * return Deferred object with configs as argument
   * @param tags {Object}
   * ** siteName
   * ** tagName (optional)
   * @return {object}
   */
  getConfigsByTags: function (tags) {
    var storedTags = [];
    App.db.getConfigs().forEach(function (site) {
      storedTags.push({
        siteName: site.type,
        tagName: site.tag
      })
    });
    if (this.checkTagsChanges(tags, storedTags)) {
      return this.loadFromServer(tags);
    } else {
      return this.loadFromDB(tags.mapProperty('siteName'));
    }
  },
  /**
   * check whether tag versions have been changed
   * if they are different then return true
   * otherwise false
   * @param tags
   * @param storedTags
   * @return {Boolean}
   */
  checkTagsChanges: function (tags, storedTags) {
    var isDifferent = false;
    var i = 0;
    while (i < tags.length && !isDifferent) {
      var storedTag = storedTags.findProperty('siteName', tags[i].siteName);
      isDifferent = (!storedTag || storedTag.tagName !== tags[i].tagName);
      i++;
    }
    return isDifferent;
  },
  loadFromDB: function (siteNames) {
    var dfd = $.Deferred();
    var configs = App.db.getConfigs().filter(function (site) {
      return (siteNames.contains(site.type));
    });
    dfd.resolve(configs);
    return dfd.promise()
  },
  /**
   * load configs from server
   * and update them in local DB
   * @param tags
   * @return {Array}
   */
  loadFromServer: function (tags) {
    var self = this;
    var dfd = $.Deferred();
    if (!tags.everyProperty('tagName')) {
      var configTags;
      var jqXhr =  this.loadConfigTags();
      jqXhr.done(function (data) {
        configTags = data.Clusters.desired_configs;
        tags.forEach(function (_tag) {
          if (_tag.siteName && !_tag.tagName) {
            _tag.tagName = configTags[_tag.siteName].tag;
          }
        }, self);
        self.loadConfigsByTags(tags,dfd);
      });
    } else {
      self.loadConfigsByTags(tags,dfd);
    }
    return dfd.promise();
  },

  /**
   *  loadConfigsByTags: Loads properties for a config tag
   *  @params tags
   *  @params dfd jqXhr promise
   */
  loadConfigsByTags: function (tags,dfd) {
    var self = this;
    var loadedConfigs = [];
    App.config.loadConfigsByTags(tags).done(function (data) {
      if (data.items) {
        data.items.forEach(function (item) {
          App.config.loadedConfigurationsCache[item.type + "_" + item.tag] = item.properties;
          loadedConfigs.push(item);
        });
      }
    }).complete(function () {
      self.saveToDB(loadedConfigs);
      dfd.resolve(loadedConfigs);
    });
  },

  /**
   * loadConfigTags: Loads all config tags applied to the cluster
   * @return: jqXhr promise
   */
  loadConfigTags: function () {
    return App.ajax.send({
      name: 'config.tags',
      sender: this
    });
  },

  /**
   * save properties obtained from server to local DB
   * @param loadedConfigs
   */
  saveToDB: function (loadedConfigs) {
    var storedConfigs = App.db.getConfigs();
    loadedConfigs.forEach(function (loadedSite) {
      var storedSite = storedConfigs.findProperty('type', loadedSite.type);
      if (storedSite) {
        storedSite.tag = loadedSite.tag;
        storedSite.properties = loadedSite.properties;
        storedSite.properties_attributes = loadedSite.properties_attributes;
      } else {
        storedConfigs.push(loadedSite);
      }
    });
    App.db.setConfigs(storedConfigs);
  }
});

});

require.register("controllers/global/update_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.UpdateController = Em.Controller.extend({
  name: 'updateController',
  isUpdated: false,
  cluster: null,
  isWorking: false,
  updateAlertInstances: function() {
    return this.get('isWorking') && !App.get('router.mainAlertInstancesController.isUpdating');
  }.property('isWorking', 'App.router.mainAlertInstancesController.isUpdating'),
  timeIntervalId: null,
  clusterName: function () {
    return App.router.get('clusterController.clusterName');
  }.property('App.router.clusterController.clusterName'),
  location: function () {
    return App.router.get('location.lastSetURL');
  }.property('App.router.location.lastSetURL'),

  /**
   * keys which should be preloaded in order to filter hosts by host-components
   */
  hostsPreLoadKeys: ['host_components/HostRoles/component_name', 'host_components/HostRoles/stale_configs', 'host_components/HostRoles/maintenance_state'],

  paginationKeys: ['page_size', 'from'],

  getUrl: function (testUrl, url) {
    return (App.get('testMode')) ? testUrl : App.apiPrefix + '/clusters/' + this.get('clusterName') + url;
  },

  /**
   * construct URL from real URL and query parameters
   * @param testUrl
   * @param realUrl
   * @param queryParams
   * @return {String}
   */
  getComplexUrl: function (testUrl, realUrl, queryParams) {
    var prefix = App.get('apiPrefix') + '/clusters/' + App.get('clusterName'),
      params = '';

    if (App.get('testMode')) {
      return testUrl;
    } else {
      if (queryParams) {
        params = this.computeParameters(queryParams);
      }
      return prefix + realUrl.replace('<parameters>', params);
    }
  },

  /**
   * compute parameters according to their type
   * @param queryParams
   * @return {String}
   */
  computeParameters: function (queryParams) {
    var params = '';

    queryParams.forEach(function (param) {
      switch (param.type) {
        case 'EQUAL':
          params += param.key + '=' + param.value;
          break;
        case 'LESS':
          params += param.key + '<' + param.value;
          break;
        case 'MORE':
          params += param.key + '>' + param.value;
          break;
        case 'MATCH':
          params += param.key + '.matches(' + param.value + ')';
          break;
        case 'MULTIPLE':
          params += param.key + '.in(' + param.value.join(',') + ')';
          break;
        case 'SORT':
          params += 'sortBy=' + param.key + '.' + param.value;
          break;
        case 'CUSTOM':
          param.value.forEach(function(item, index){
            param.key = param.key.replace('{' + index + '}', item);
          }, this);
          params += param.key;
          break;
      }
      params += '&';
    });
    return params;
  },

  /**
   * depict query parameters of table
   */
  queryParams: Em.Object.create({
    'Hosts': []
  }),

  /**
   * Pagination query-parameters for unhealthy alerts request
   * @type {{from: Number, page_size: Number}}
   */
  queryParamsForUnhealthyAlertInstances: {
    from: 0,
    page_size: 10
  },

  /**
   * map describes relations between updater function and table
   */
  tableUpdaterMap: {
    'Hosts': 'updateHost'
  },

  /**
   * Start polling, when <code>isWorking</code> become true
   */
  updateAll: function () {
    if (this.get('isWorking')) {
      App.updater.run(this, 'updateServices', 'isWorking');
      App.updater.run(this, 'updateHost', 'isWorking');
      App.updater.run(this, 'updateServiceMetricConditionally', 'isWorking', App.componentsUpdateInterval);
      App.updater.run(this, 'updateComponentsState', 'isWorking', App.componentsUpdateInterval);
      App.updater.run(this, 'graphsUpdate', 'isWorking');
      App.updater.run(this, 'updateComponentConfig', 'isWorking');

      App.updater.run(this, 'updateAlertGroups', 'isWorking', App.alertGroupsUpdateInterval);
      App.updater.run(this, 'updateAlertDefinitions', 'isWorking', App.alertDefinitionsUpdateInterval);
      App.updater.run(this, 'updateAlertDefinitionSummary', 'isWorking', App.alertDefinitionsUpdateInterval);
      if (!App.get('router.mainAlertInstancesController.isUpdating')) {
        App.updater.run(this, 'updateUnhealthyAlertInstances', 'updateAlertInstances', App.alertInstancesUpdateInterval);
      }
    }
  }.observes('isWorking', 'App.router.mainAlertInstancesController.isUpdating'),
  /**
   * Update service metrics depending on which page is open
   * Make a call only on follow pages:
   * /main/dashboard
   * /main/services/*
   * @param callback
   */
  updateServiceMetricConditionally: function (callback) {
    if (/\/main\/(dashboard|services).*/.test(this.get('location'))) {
      this.updateServiceMetric(callback);
    } else {
      callback();
    }
  },

  updateHost: function (callback, error) {
    var testUrl = '/data/hosts/HDP2/hosts.json',
      self = this,
      hostDetailsFilter = '';
    var realUrl = '/hosts?<parameters>fields=Hosts/host_name,Hosts/maintenance_state,Hosts/public_host_name,Hosts/cpu_count,Hosts/ph_cpu_count,' +
      'alerts_summary,Hosts/host_status,Hosts/last_heartbeat_time,Hosts/ip,host_components/HostRoles/state,host_components/HostRoles/maintenance_state,' +
      'host_components/HostRoles/stale_configs,host_components/HostRoles/service_name,host_components/HostRoles/desired_admin_state,' +
        'metrics/disk,metrics/load/load_one,Hosts/total_mem<hostAuxiliaryInfo><stackVersions>&minimal_response=true';
    var hostAuxiliaryInfo = ',Hosts/os_arch,Hosts/os_type,metrics/cpu/cpu_system,metrics/cpu/cpu_user,metrics/memory/mem_total,metrics/memory/mem_free';
    var stackVersionInfo = ',stack_versions/HostStackVersions,' +
      'stack_versions/repository_versions/RepositoryVersions/repository_version,stack_versions/repository_versions/RepositoryVersions/id,' +
      'stack_versions/repository_versions/RepositoryVersions/display_name';
    realUrl = realUrl.replace("<stackVersions>", (App.get('supports.stackUpgrade') ? stackVersionInfo : ""));

    if (App.router.get('currentState.name') == 'index' && App.router.get('currentState.parentState.name') == 'hosts') {
      App.updater.updateInterval('updateHost', App.get('contentUpdateInterval'));
    }
    else {
      if (App.router.get('currentState.parentState.name') == 'hostDetails' &&
          ['summary', 'alerts', 'stackVersions'].contains(App.router.get('currentState.name'))) {
        hostDetailsFilter = App.router.get('location.lastSetURL').match(/\/hosts\/(.*)\/(summary|alerts|stackVersions)/)[1];
        App.updater.updateInterval('updateHost', App.get('componentsUpdateInterval'));
      }
      else {
        callback();
        // On pages except for hosts/hostDetails, making sure hostsMapper loaded only once on page load, no need to update, but at least once
        if (App.router.get('clusterController.isLoaded')) {
          return;
        }
      }
    }
    var mainHostController = App.router.get('mainHostController'),
      sortProperties = mainHostController.getSortProps();
    if (hostDetailsFilter) {
      //if host details page opened then request info only of one displayed host
      this.get('queryParams').set('Hosts', [
        {
          key: 'Hosts/host_name',
          value: [hostDetailsFilter],
          type: 'MULTIPLE'
        }
      ]);
    } else {
      hostAuxiliaryInfo = '';
      this.get('queryParams').set('Hosts', mainHostController.getQueryParameters(true));
    }
    realUrl = realUrl.replace('<hostAuxiliaryInfo>', hostAuxiliaryInfo);

    var clientCallback = function (skipCall, queryParams) {
      if (skipCall) {
        //no hosts match filter by component
        App.hostsMapper.map({
          items: [],
          itemTotal: '0'
        });
        callback();
      }
      else {
        var params = self.computeParameters(queryParams),
          paginationProps = self.computeParameters(queryParams.filter(function (param) {
            return (this.get('paginationKeys').contains(param.key));
          }, self)),
          sortProps = self.computeParameters(sortProperties);

        if ((params.length + paginationProps.length + sortProps.length) > 0) {
          realUrl = App.get('apiPrefix') + '/clusters/' + App.get('clusterName') +
            realUrl.replace('<parameters>', '') +
            (paginationProps.length > 0 ? '&' + paginationProps.substring(0, paginationProps.length - 1) : '') +
            (sortProps.length > 0 ? '&' + sortProps.substring(0, sortProps.length - 1) : '');
          if (App.get('testMode')) {
            realUrl = testUrl;
          }
          App.HttpClient.get(realUrl, App.hostsMapper, {
            complete: callback,
            doGetAsPost: true,
            params: params.substring(0, params.length - 1),
            error: error
          });
        }
        else {
          var hostsUrl = self.getComplexUrl(testUrl, realUrl, queryParams);
          App.HttpClient.get(hostsUrl, App.hostsMapper, {
            complete: callback,
            doGetAsPost: false,
            error: error
          });
        }
      }
    };

    if (!this.preLoadHosts(clientCallback)) {
      clientCallback(false, self.get('queryParams.Hosts'));
    }
  },

  /**
   * identify if any filter by host-component is active
   * if so run @getHostByHostComponents
   *
   * @param callback
   * @return {Boolean}
   */
  preLoadHosts: function (callback) {
    var preLoadKeys = this.get('hostsPreLoadKeys');

    if (this.get('queryParams.Hosts').length > 0 && this.get('queryParams.Hosts').filter(function (param) {
      return (preLoadKeys.contains(param.key));
    }, this).length > 0) {
      this.getHostByHostComponents(callback);
      return true;
    }
    return false;
  },

  /**
   * get hosts' names which match filter by host-component
   * @param callback
   */
  getHostByHostComponents: function (callback) {
    var testUrl = '/data/hosts/HDP2/hosts.json';
    var realUrl = '/hosts?<parameters>minimal_response=true';

    App.ajax.send({
      name: 'hosts.host_components.pre_load',
      sender: this,
      data: {
        url: this.getComplexUrl(testUrl, realUrl, this.get('queryParams.Hosts')),
        callback: callback
      },
      success: 'getHostByHostComponentsSuccessCallback',
      error: 'getHostByHostComponentsErrorCallback'
    })
  },
  getHostByHostComponentsSuccessCallback: function (data, opt, params) {
    var preLoadKeys = this.get('hostsPreLoadKeys');
    var queryParams = this.get('queryParams.Hosts');
    var hostNames = data.items.mapProperty('Hosts.host_name');
    var skipCall = hostNames.length === 0;

    /**
     * exclude pagination parameters as they were applied in previous call
     * to obtain hostnames of filtered hosts
     */
    preLoadKeys = preLoadKeys.concat(this.get('paginationKeys'));

    var itemTotal = parseInt(data.itemTotal);
    if (!isNaN(itemTotal)) {
      App.router.set('mainHostController.filteredCount', itemTotal);
    }

    if (skipCall) {
      params.callback(skipCall);
    } else {
      queryParams = queryParams.filter(function (param) {
        return !(preLoadKeys.contains(param.key));
      });

      queryParams.push({
        key: 'Hosts/host_name',
        value: hostNames,
        type: 'MULTIPLE'
      });
      params.callback(skipCall, queryParams);
    }
  },
  getHostByHostComponentsErrorCallback: function () {
    console.warn('ERROR: filtering hosts by host-component failed');
  },
  graphs: [],
  graphsUpdate: function (callback) {
    var existedGraphs = [];
    this.get('graphs').forEach(function (_graph) {
      var view = Em.View.views[_graph.id];
      if (view) {
        existedGraphs.push(_graph);
        //console.log('updated graph', _graph.name);
        view.loadData();
        //if graph opened as modal popup update it to
        if ($(".modal-graph-line .modal-body #" + _graph.popupId + "-container-popup").length) {
          view.loadData();
        }
      }
    });
    callback();
    this.set('graphs', existedGraphs);
  },

  /**
   * Updates the services information.
   *
   * @param callback
   */
  updateServiceMetric: function (callback) {
    var self = this;
    self.set('isUpdated', false);
    var isATSPresent = App.StackServiceComponent.find().findProperty('componentName','APP_TIMELINE_SERVER');

    var conditionalFields = this.getConditionalFields(),
      conditionalFieldsString = conditionalFields.length > 0 ? ',' + conditionalFields.join(',') : '',
      testUrl = '/data/dashboard/HDP2/master_components.json',
      isFlumeInstalled = App.cache['services'].mapProperty('ServiceInfo.service_name').contains('FLUME'),
      isATSInstalled = App.cache['services'].mapProperty('ServiceInfo.service_name').contains('YARN') && isATSPresent,
      flumeHandlerParam = isFlumeInstalled ? 'ServiceComponentInfo/component_name=FLUME_HANDLER|' : '',
      atsHandlerParam = isATSInstalled ? 'ServiceComponentInfo/component_name=APP_TIMELINE_SERVER|' : '',
      haComponents = App.get('isHaEnabled') ? 'ServiceComponentInfo/component_name=JOURNALNODE|ServiceComponentInfo/component_name=ZKFC|' : '',
      realUrl = '/components/?' + flumeHandlerParam + atsHandlerParam + haComponents +
        'ServiceComponentInfo/category=MASTER&fields=' +
        'ServiceComponentInfo/Version,' +
        'ServiceComponentInfo/StartTime,' +
        'ServiceComponentInfo/HeapMemoryUsed,' +
        'ServiceComponentInfo/HeapMemoryMax,' +
        'ServiceComponentInfo/service_name,' +
        'host_components/HostRoles/host_name,' +
        'host_components/HostRoles/state,' +
        'host_components/HostRoles/maintenance_state,' +
        'host_components/HostRoles/stale_configs,' +
        'host_components/HostRoles/ha_state,' +
        'host_components/HostRoles/desired_admin_state,' +
        'host_components/metrics/jvm/memHeapUsedM,' +
        'host_components/metrics/jvm/HeapMemoryMax,' +
        'host_components/metrics/jvm/HeapMemoryUsed,' +
        'host_components/metrics/jvm/memHeapCommittedM,' +
        'host_components/metrics/mapred/jobtracker/trackers_decommissioned,' +
        'host_components/metrics/cpu/cpu_wio,' +
        'host_components/metrics/rpc/RpcQueueTime_avg_time,' +
        'host_components/metrics/dfs/FSNamesystem/*,' +
        'host_components/metrics/dfs/namenode/Version,' +
        'host_components/metrics/dfs/namenode/LiveNodes,' +
        'host_components/metrics/dfs/namenode/DeadNodes,' +
        'host_components/metrics/dfs/namenode/DecomNodes,' +
        'host_components/metrics/dfs/namenode/TotalFiles,' +
        'host_components/metrics/dfs/namenode/UpgradeFinalized,' +
        'host_components/metrics/dfs/namenode/Safemode,' +
        'host_components/metrics/runtime/StartTime' +
        conditionalFieldsString +
        '&minimal_response=true';

    var servicesUrl = this.getUrl(testUrl, realUrl);
    callback = callback || function () {
      self.set('isUpdated', true);
    };
    App.HttpClient.get(servicesUrl, App.serviceMetricsMapper, {
      complete: function () {
        callback();
      }
    });
  },
  /**
   * construct conditional parameters of query, depending on which services are installed
   * @return {Array}
   */
  getConditionalFields: function () {
    var conditionalFields = [];
    var serviceSpecificParams = {
      'FLUME': "host_components/processes/HostComponentProcess",
      'YARN': "host_components/metrics/yarn/Queue," +
        "ServiceComponentInfo/rm_metrics/cluster/activeNMcount," +
        "ServiceComponentInfo/rm_metrics/cluster/lostNMcount," +
        "ServiceComponentInfo/rm_metrics/cluster/unhealthyNMcount," +
        "ServiceComponentInfo/rm_metrics/cluster/rebootedNMcount," +
        "ServiceComponentInfo/rm_metrics/cluster/decommissionedNMcount",
      'HBASE': "host_components/metrics/hbase/master/IsActiveMaster," +
        "ServiceComponentInfo/MasterStartTime," +
        "ServiceComponentInfo/MasterActiveTime," +
        "ServiceComponentInfo/AverageLoad," +
        "ServiceComponentInfo/Revision," +
        "ServiceComponentInfo/RegionsInTransition",
      'STORM': /^2.1/.test(App.get('currentStackVersionNumber')) ? 'metrics/api/cluster/summary' : 'metrics/api/v1/cluster/summary,metrics/api/v1/topology/summary'
    };
    var services = App.cache['services'];
    services.forEach(function (service) {
      var urlParams = serviceSpecificParams[service.ServiceInfo.service_name];
      if (urlParams) {
        conditionalFields.push(urlParams);
      }
    });
    return conditionalFields;
  },
  updateServices: function (callback) {
    var testUrl = '/data/services/HDP2/services.json';
    var componentConfigUrl = this.getUrl(testUrl, '/services?fields=ServiceInfo/state,ServiceInfo/maintenance_state&minimal_response=true');
    App.HttpClient.get(componentConfigUrl, App.serviceMapper, {
      complete: callback
    });
  },
  updateComponentConfig: function (callback) {
    var testUrl = '/data/services/host_component_stale_configs.json';
    var componentConfigUrl = this.getUrl(testUrl, '/components?ServiceComponentInfo/category.in(SLAVE,CLIENT)&host_components/HostRoles/stale_configs=true&fields=host_components/HostRoles/service_name,host_components/HostRoles/state,host_components/HostRoles/maintenance_state,host_components/HostRoles/host_name,host_components/HostRoles/stale_configs,host_components/HostRoles/desired_admin_state&minimal_response=true');
    App.HttpClient.get(componentConfigUrl, App.componentConfigMapper, {
      complete: callback
    });
  },
  updateComponentsState: function (callback) {
    var testUrl = '/data/services/HDP2/components_state.json';
    var realUrl = '/components/?ServiceComponentInfo/category.in(SLAVE,CLIENT)&fields=ServiceComponentInfo/service_name,' +
      'ServiceComponentInfo/category,ServiceComponentInfo/installed_count,ServiceComponentInfo/started_count,ServiceComponentInfo/total_count&minimal_response=true';
    var url = this.getUrl(testUrl, realUrl);

    App.HttpClient.get(url, App.componentsStateMapper, {
      complete: callback
    });
  },
  updateAlertDefinitions: function (callback) {
    var testUrl = '/data/alerts/alertDefinitions.json';
    var realUrl = '/alert_definitions?fields=*';
    var url = this.getUrl(testUrl, realUrl);

    App.HttpClient.get(url, App.alertDefinitionsMapper, {
      complete: callback
    });
  },

  updateUnhealthyAlertInstances: function (callback) {
    var testUrl = '/data/alerts/alert_instances.json';
    var queryParams = this.get('queryParamsForUnhealthyAlertInstances');
    var realUrl = '/alerts?fields=*&Alert/state.in(CRITICAL,WARNING)&Alert/maintenance_state.in(OFF)&from=' + queryParams.from + '&page_size=' + queryParams.page_size;
    var url = this.getUrl(testUrl, realUrl);

    App.HttpClient.get(url, App.alertInstanceMapper, {
      complete: callback
    });
  },

  updateAlertDefinitionSummary: function(callback) {
    var testUrl = '/data/alerts/alert_summary.json';
    var realUrl = '/alerts?format=groupedSummary';
    var url = this.getUrl(testUrl, realUrl);

    App.HttpClient.get(url, App.alertDefinitionSummaryMapper, {
      complete: callback
    });
  },

  updateAlertGroups: function (callback) {
    var testUrl = '/data/alerts/alertGroups.json';
    var realUrl = '/alert_groups?fields=*';
    var url = this.getUrl(testUrl, realUrl);

    App.HttpClient.get(url, App.alertGroupsMapper, {
      complete: callback
    });
  },

  updateAlertNotifications: function (callback) {
    App.HttpClient.get(App.get('apiPrefix') + '/alert_targets?fields=*', App.alertNotificationMapper, {
      complete: callback
    });
  }

});

});

require.register("controllers/installer", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.InstallerController = App.WizardController.extend({

  name: 'installerController',

  totalSteps: 11,

  content: Em.Object.create({
    cluster: null,
    installOptions: null,
    hosts: null,
    services: null,
    slaveComponentHosts: null,
    masterComponentHosts: null,
    serviceConfigProperties: null,
    advancedServiceConfig: null,
    configGroups: [],
    slaveGroupProperties: null,
    stacks: null,
    clients: [],
    /**
     * recommendations for host groups loaded from server
     */
    recommendations: null,
    /**
     * recommendationsHostGroups - current component assignment after 5 and 6 steps
     * (uses for host groups validation and to load recommended configs)
     */
    recommendationsHostGroups: null,
    controllerName: 'installerController'
  }),

  /**
   * Wizard properties in local storage, which should be cleaned right after wizard has been finished
   */
  dbPropertiesToClean: [
    'service',
    'hosts',
    'masterComponentHosts',
    'slaveComponentHosts',
    'cluster',
    'allHostNames',
    'installOptions',
    'allHostNamesPattern',
    'serviceComponents',
    'advancedServiceConfig',
    'clientInfo',
    'selectedServiceNames',
    'serviceConfigGroups',
    'serviceConfigProperties',
    'fileNamesToUpdate',
    'bootStatus',
    'stacksVersions',
    'currentStep',
    'serviceInfo',
    'hostInfo',
    'recommendations',
    'recommendationsHostGroups',
    'recommendationsConfigs'
  ],

  init: function () {
    this._super();
    this.get('isStepDisabled').setEach('value', true);
    this.get('isStepDisabled').pushObject(Ember.Object.create({
      step: 0,
      value: false
    }));
  },
  /**
   * redefined connectOutlet method to avoid view loading by unauthorized user
   * @param view
   * @param content
   */
  connectOutlet: function (view, content) {
    if (App.db.getAuthenticated()) {
      this._super(view, content);
    }
  },

  getCluster: function () {
    return jQuery.extend({}, this.get('clusterStatusTemplate'));
  },

  getHosts: function () {
    return [];
  },

  /**
   * Remove host from model. Used at <code>Confirm hosts(step2)</code> step
   * @param hosts Array of hosts, which we want to delete
   */
  removeHosts: function (hosts) {
    var dbHosts = this.getDBProperty('hosts');
    hosts.forEach(function (_hostInfo) {
      var host = _hostInfo.name;
      delete dbHosts[host];
    });
    this.setDBProperty('hosts', dbHosts);
  },

  /**
   * Load confirmed hosts.
   * Will be used at <code>Assign Masters(step5)</code> step
   */
  loadConfirmedHosts: function () {
    this.set('content.hosts', this.getDBProperty('hosts') || {});
  },

  /**
   * Load services data. Will be used at <code>Select services(step4)</code> step
   */
  loadServices: function () {
    var dfd = $.Deferred();
    var self = this;
    var stackServices = App.StackService.find().mapProperty('serviceName');
    if (!(stackServices && !!stackServices.length && App.StackService.find().objectAt(0).get('stackVersion') == App.get('currentStackVersionNumber'))) {
      this.loadServiceComponents().complete(function () {
        self.set('content.services', App.StackService.find());
        dfd.resolve();
      });
    } else {
      dfd.resolve();
    }
    return dfd.promise();
  },

  /**
   * total set of hosts registered to cluster, analog of App.Host model,
   * used in Installer wizard until hosts are installed
   */
  allHosts: function () {
    var rawHosts = this.get('content.hosts');
    var masterComponents = this.get('content.masterComponentHosts');
    var slaveComponents = this.get('content.slaveComponentHosts');
    var hosts = [];
    masterComponents.forEach(function (component) {
      var host = rawHosts[component.hostName];
      if (host.hostComponents) {
        host.hostComponents.push(Em.Object.create({
          componentName: component.component,
          displayName: component.display_name
        }));
      } else {
        rawHosts[component.hostName].hostComponents = [
          Em.Object.create({
            componentName: component.component,
            displayName: component.display_name
          })
        ]
      }
    });
    slaveComponents.forEach(function (component) {
      component.hosts.forEach(function (rawHost) {
        var host = rawHosts[rawHost.hostName];
        if (host.hostComponents) {
          host.hostComponents.push(Em.Object.create({
            componentName: component.componentName,
            displayName: component.displayName
          }));
        } else {
          rawHosts[rawHost.hostName].hostComponents = [
            Em.Object.create({
              componentName: component.componentName,
              displayName: component.displayName
            })
          ]
        }
      });
    });

    for (var hostName in rawHosts) {
      var host = rawHosts[hostName];
      var disksOverallCapacity = 0;
      var diskFree = 0;
      host.disk_info.forEach(function (disk) {
        disksOverallCapacity += parseFloat(disk.size);
        diskFree += parseFloat(disk.available);
      });
      hosts.pushObject(Em.Object.create({
          id: host.name,
          ip: host.ip,
          osType: host.os_type,
          osArch: host.os_arch,
          hostName: host.name,
          publicHostName: host.name,
          cpu: host.cpu,
          memory: host.memory,
          diskInfo: host.disk_info,
          diskTotal: disksOverallCapacity / (1024 * 1024),
          diskFree: diskFree / (1024 * 1024),
          hostComponents: host.hostComponents
        }
      ))
    }
    return hosts;
  }.property('content.hosts'),

  stacks: [],

  /**
   * stack names used as auxiliary data to query stacks by name
   */
  stackNames: [],

  /**
   * Load stacks data from server or take exist data from in memory variable {{content.stacks}}
   * The series of API calls will be called  When landing first time on Select Stacks page
   * or on hitting refresh post select stacks page in installer wizard
   */
  loadStacks: function () {
    var stacks = this.get('content.stacks');
    var dfd = $.Deferred();
    if (stacks && stacks.get('length')) {
      App.set('currentStackVersion', App.Stack.find().findProperty('isSelected').get('id'));
      dfd.resolve(true);
    } else {
      App.ajax.send({
        name: 'wizard.stacks',
        sender: this,
        success: 'loadStacksSuccessCallback',
        error: 'loadStacksErrorCallback'
      }).complete(function () {
        dfd.resolve(false);
      });
    }
    return dfd.promise();
  },

  /**
   * Send queries to load versions for each stack
   */
  loadStacksSuccessCallback: function (data) {
    this.get('stacks').clear();
    this.set('stackNames', data.items.mapProperty('Stacks.stack_name'));
  },

  /**
   * onError callback for loading stacks data
   */
  loadStacksErrorCallback: function () {
    console.log('Error in loading stacks');
  },

  /**
   * query every stack names from server
   * @return {Array}
   */
  loadStacksVersions: function () {
    var requests = [];
    this.get('stackNames').forEach(function (stackName) {
      requests.push(App.ajax.send({
        name: 'wizard.stacks_versions',
        sender: this,
        data: {
          stackName: stackName
        },
        success: 'loadStacksVersionsSuccessCallback',
        error: 'loadStacksVersionsErrorCallback'
      }));
    }, this);
    return requests;
  },

  /**
   * Parse loaded data and create array of stacks objects
   */
  loadStacksVersionsSuccessCallback: function (data) {
    var stacks = App.db.getStacks();
    var isStacksExistInDb = stacks && stacks.length;
    if (isStacksExistInDb) {
      stacks.forEach(function (_stack) {
        var stack = data.items.filterProperty('Versions.stack_name', _stack.stack_name).findProperty('Versions.stack_version', _stack.stack_version);
        if (stack) {
          stack.Versions.is_selected = _stack.is_selected;
        }
      }, this);
    }
    App.stackMapper.map(data);
    if (!isStacksExistInDb) {
      var defaultStackVersion = App.Stack.find().findProperty('id', App.defaultStackVersion);
      if (defaultStackVersion) {
        defaultStackVersion.set('isSelected', true)
      } else {
        App.Stack.find().objectAt(0).set('isSelected', true);
      }
    }
    this.set('content.stacks', App.Stack.find());
    App.set('currentStackVersion', App.Stack.find().findProperty('isSelected').get('id'));
  },

  /**
   * onError callback for loading stacks data
   */
  loadStacksVersionsErrorCallback: function () {
    console.log('Error in loading stacks');
  },

  /**
   * check server version and web client version
   */
  checkServerClientVersion: function () {
    var dfd = $.Deferred();
    var self = this;
    self.getServerVersion().done(function () {
      dfd.resolve();
    });
    return dfd.promise();
  },
  getServerVersion: function () {
    return App.ajax.send({
      name: 'ambari.service.load_server_version',
      sender: this,
      success: 'getServerVersionSuccessCallback',
      error: 'getServerVersionErrorCallback'
    });
  },
  getServerVersionSuccessCallback: function (data) {
    var clientVersion = App.get('version');
    var serverVersion = (data.RootServiceComponents.component_version).toString();
    this.set('ambariServerVersion', serverVersion);
    if (clientVersion) {
      this.set('versionConflictAlertBody', Em.I18n.t('app.versionMismatchAlert.body').format(serverVersion, clientVersion));
      this.set('isServerClientVersionMismatch', clientVersion != serverVersion);
    } else {
      this.set('isServerClientVersionMismatch', false);
    }
    App.set('isManagedMySQLForHiveEnabled', App.config.isManagedMySQLForHiveAllowed(data.RootServiceComponents.properties['server.os_family']));
  },
  getServerVersionErrorCallback: function () {
    console.log('ERROR: Cannot load Ambari server version');
  },

  /**
   * set stacks from server to content and local DB
   */
  setStacks: function () {
    var result = App.Stack.find() || [];
    Em.assert('Stack model is not populated', result.get('length'));
    App.db.setStacks(result.slice());
    this.set('content.stacks', result);
  },

  /**
   * Save data to model
   * @param stepController App.WizardStep4Controller
   */
  saveServices: function (stepController) {
    var selectedServiceNames = [];
    var installedServiceNames = [];
    stepController.filterProperty('isSelected').forEach(function (item) {
      selectedServiceNames.push(item.get('serviceName'));
    });
    stepController.filterProperty('isInstalled').forEach(function (item) {
      installedServiceNames.push(item.get('serviceName'));
    });
    this.set('content.services', App.StackService.find());
    this.set('content.selectedServiceNames', selectedServiceNames);
    this.setDBProperty('selectedServiceNames', selectedServiceNames);
    this.set('content.installedServiceNames', installedServiceNames);
    this.setDBProperty('installedServiceNames', installedServiceNames);
  },

  /**
   * Save Master Component Hosts data to Main Controller
   * @param stepController App.WizardStep5Controller
   */
  saveMasterComponentHosts: function (stepController) {

    var obj = stepController.get('selectedServicesMasters'),
      hosts = this.getDBProperty('hosts');

    var masterComponentHosts = [];
    obj.forEach(function (_component) {
      masterComponentHosts.push({
        display_name: _component.get('display_name'),
        component: _component.get('component_name'),
        serviceId: _component.get('serviceId'),
        isInstalled: false,
        host_id: hosts[_component.get('selectedHost')].id
      });
    });

    console.log("installerController.saveMasterComponentHosts: saved hosts ", masterComponentHosts);
    this.setDBProperty('masterComponentHosts', masterComponentHosts);
    this.set('content.masterComponentHosts', masterComponentHosts);
  },

  /**
   * Load master component hosts data for using in required step controllers
   */
  loadMasterComponentHosts: function () {
    var masterComponentHosts = this.getDBProperty('masterComponentHosts'),
      hosts = this.getDBProperty('hosts'),
      host_names = Em.keys(hosts);
    if (Em.isNone(masterComponentHosts)) {
      masterComponentHosts = [];
    }
    else {
      masterComponentHosts.forEach(function (component) {
        for (var i = 0; i < host_names.length; i++) {
          if (hosts[host_names[i]].id === component.host_id) {
            component.hostName = host_names[i];
            break;
          }
        }
      });
    }
    this.set("content.masterComponentHosts", masterComponentHosts);
  },

  loadRecommendations: function () {
    this.set("content.recommendations", this.getDBProperty('recommendations'));
  },

  loadCurrentHostGroups: function () {
    this.set("content.recommendationsHostGroups", this.getDBProperty('recommendationsHostGroups'));
  },

  loadRecommendationsConfigs: function () {
    App.router.set("wizardStep7Controller.recommendationsConfigs", this.getDBProperty('recommendationsConfigs'));
  },

  /**
   * Load master component hosts data for using in required step controllers
   */
  loadSlaveComponentHosts: function () {
    var slaveComponentHosts = this.getDBProperty('slaveComponentHosts'),
      hosts = this.getDBProperty('hosts'),
      host_names = Em.keys(hosts);
    if (!Em.isNone(slaveComponentHosts)) {
      slaveComponentHosts.forEach(function (component) {
        component.hosts.forEach(function (host) {
          for (var i = 0; i < host_names.length; i++) {
            if (hosts[host_names[i]].id === host.host_id) {
              host.hostName = host_names[i];
              break;
            }
          }
        });
      });
    }
    this.set("content.slaveComponentHosts", slaveComponentHosts);
    console.log("InstallerController.loadSlaveComponentHosts: loaded hosts ", slaveComponentHosts);
  },

  /**
   * Load serviceConfigProperties to model
   */
  loadServiceConfigProperties: function () {
    var serviceConfigProperties = this.getDBProperty('serviceConfigProperties');
    this.set('content.serviceConfigProperties', serviceConfigProperties);
    console.log("InstallerController.loadServiceConfigProperties: loaded config ", serviceConfigProperties);

    this.set('content.advancedServiceConfig', this.getDBProperty('advancedServiceConfig'));
  },
  /**
   * Generate clients list for selected services and save it to model
   * @param stepController step4WizardController
   */
  saveClients: function (stepController) {
    var clients = [];
    stepController.get('content').filterProperty('isSelected', true).forEach(function (_service) {
      var client = _service.get('serviceComponents').filterProperty('isClient', true);
      client.forEach(function (clientComponent) {
        clients.pushObject({
          component_name: clientComponent.get('componentName'),
          display_name: clientComponent.get('displayName'),
          isInstalled: false
        });
      }, this);
    }, this);
    this.setDBProperty('clientInfo', clients);
    this.set('content.clients', clients);
  },

  /**
   * Check validation of the customized local urls
   */
  checkRepoURL: function (wizardStep1Controller) {
    var selectedStack = this.get('content.stacks').findProperty('isSelected', true);
    selectedStack.set('reload', true);
    var nameVersionCombo = selectedStack.get('id');
    var stackName = nameVersionCombo.split('-')[0];
    var stackVersion = nameVersionCombo.split('-')[1];
    var dfd = $.Deferred();
    if (selectedStack && selectedStack.get('operatingSystems')) {
      this.set('validationCnt', selectedStack.get('repositories').filterProperty('isSelected').length);
      var verifyBaseUrl = !wizardStep1Controller.get('skipValidationChecked');
      selectedStack.get('operatingSystems').forEach(function (os) {
        if (os.get('isSelected')) {
          os.get('repositories').forEach(function (repo) {
            repo.set('errorTitle', '');
            repo.set('errorContent', '');
            repo.set('validation', App.Repository.validation['INPROGRESS']);
            App.ajax.send({
              name: 'wizard.advanced_repositories.valid_url',
              sender: this,
              data: {
                stackName: stackName,
                stackVersion: stackVersion,
                repoId: repo.get('repoId'),
                osType: os.get('osType'),
                osId: os.get('id'),
                dfd: dfd,
                data: {
                  'Repositories': {
                    'base_url': repo.get('baseUrl'),
                    "verify_base_url": verifyBaseUrl
                  }
                }
              },
              success: 'checkRepoURLSuccessCallback',
              error: 'checkRepoURLErrorCallback'
            });
          }, this);
        }
      }, this);
    }
    return dfd.promise();
  },
  /**
   * onSuccess callback for check Repo URL.
   */
  checkRepoURLSuccessCallback: function (response, request, data) {
    console.log('Success in check Repo URL. data osType: ' + data.osType);
    var selectedStack = this.get('content.stacks').findProperty('isSelected');
    if (selectedStack && selectedStack.get('operatingSystems')) {
      var os = selectedStack.get('operatingSystems').findProperty('id', data.osId);
      var repo = os.get('repositories').findProperty('repoId', data.repoId);
      if (repo) {
        repo.set('validation', App.Repository.validation['OK']);
      }
    }
    this.set('validationCnt', this.get('validationCnt') - 1);
    if (!this.get('validationCnt')) {
      data.dfd.resolve();
    }
  },

  /**
   * onError callback for check Repo URL.
   */
  checkRepoURLErrorCallback: function (request, ajaxOptions, error, data, params) {
    console.log('Error in check Repo URL. The baseURL sent is:  ' + data.data);
    var selectedStack = this.get('content.stacks').findProperty('isSelected', true);
    if (selectedStack && selectedStack.get('operatingSystems')) {
      var os = selectedStack.get('operatingSystems').findProperty('id', params.osId);
      var repo = os.get('repositories').findProperty('repoId', params.repoId);
      if (repo) {
        repo.set('validation', App.Repository.validation['INVALID']);
        repo.set('errorTitle', request.status + ":" + request.statusText);
        repo.set('errorContent', $.parseJSON(request.responseText) ? $.parseJSON(request.responseText).message : "");
      }
    }
    params.dfd.reject();
  },

  loadMap: {
    '0': [
      {
        type: 'sync',
        callback: function () {
          this.load('cluster');
        }
      }
    ],
    '1': [
      {
        type: 'async',
        callback: function () {
          return this.loadStacks();
        }
      },
      {
        type: 'async',
        callback: function (stacksLoaded) {
          var dfd = $.Deferred();

          if (!stacksLoaded) {
            $.when.apply(this, this.loadStacksVersions()).done(function () {
              dfd.resolve(stacksLoaded);
            });
          } else {
            dfd.resolve(stacksLoaded);
          }

          return dfd.promise();
        }
      }
    ],
    '2': [
      {
        type: 'sync',
        callback: function () {
          this.load('installOptions');
        }
      }
    ],
    '3': [
      {
        type: 'sync',
        callback: function () {
          this.loadConfirmedHosts();
        }
      }
    ],
    '4': [
      {
        type: 'async',
        callback: function () {
          return this.loadServices();
        }
      }
    ],
    '5': [
      {
        type: 'sync',
        callback: function () {
          this.loadMasterComponentHosts();
          this.loadConfirmedHosts();
          this.loadRecommendations();
        }
      }
    ],
    '6': [
      {
        type: 'sync',
        callback: function () {
          this.loadSlaveComponentHosts();
          this.loadClients();
          this.loadRecommendations();
        }
      }
    ],
    '7': [
      {
        type: 'sync',
        callback: function () {
          this.loadServiceConfigGroups();
          this.loadServiceConfigProperties();
          this.loadCurrentHostGroups();
          this.loadRecommendationsConfigs();
        }
      }
    ]
  },
  /**
   * Clear all temporary data
   */
  finish: function () {
    this.setCurrentStep('0');
    this.clearStorageData();
    var persists = App.router.get('applicationController').persistKey();
    App.router.get('applicationController').postUserPref(persists, true);
  },

  /**
   * Save cluster provisioning state to the server
   * @param state cluster provisioning state
   */
  setClusterProvisioningState: function (state) {
    return App.ajax.send({
      name: 'cluster.save_provisioning_state',
      sender: this,
      data: {
        state: state
      }
    });
  },

  setStepsEnable: function () {
    for (var i = 0; i <= this.totalSteps; i++) {
      var step = this.get('isStepDisabled').findProperty('step', i);
      if (i <= this.get('currentStep')) {
        step.set('value', false);
      } else {
        step.set('value', true);
      }
    }
  }.observes('currentStep'),

  setLowerStepsDisable: function (stepNo) {
    for (var i = 0; i < stepNo; i++) {
      var step = this.get('isStepDisabled').findProperty('step', i);
      step.set('value', true);
    }
  }
});


});

require.register("controllers/login_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.LoginController = Em.Object.extend({

  name: 'loginController',

  loginName: '',
  password: '',

  errorMessage: '',

  submit: function (e) {
    this.set('errorMessage', '');

    var self = this;

    App.get('router').login();
  },

  postLogin: function (isConnected, isAuthenticated, responseText) {
    if (!isConnected) {
      console.log('Failed to connect to Ambari Server');
      this.set('errorMessage', Em.I18n.t('login.error.bad.connection'));
    } else if (!isAuthenticated) {
      console.log('Failed to login as: ' + this.get('loginName'));
      var errorMessage = "";
      if( responseText === "User is disabled" ){
        errorMessage = Em.I18n.t('login.error.disabled');
      } else {
        errorMessage = Em.I18n.t('login.error.bad.credentials');
      }
      this.set('errorMessage', errorMessage);
    }
  }

});
});

require.register("controllers/main", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
require('models/background_operation');

App.MainController = Em.Controller.extend({
  name: 'mainController',

  updateTitle: function(){
    var name = App.router.get('clusterController.clusterName');
    if(App.router.get('clusterInstallCompleted')) {
      if (name && App.router.get('clusterController').get('isLoaded')) {
        name = name.length > 13 ? name.substr(0, 10) + "..." : name;
      } else {
        name = Em.I18n.t('common.loading');
      }
      $('title').text(Em.I18n.t('app.name.subtitle').format(name));
    }
  }.observes('App.router.clusterController.clusterName, App.router.clusterInstallCompleted', 'App.router.clusterController.isLoaded'),

  isClusterDataLoaded: function(){
    return App.router.get('clusterController.isLoaded');
  }.property('App.router.clusterController.isLoaded'),

  clusterDataLoadedPercent: function(){
    return App.router.get('clusterController.clusterDataLoadedPercent');
  }.property('App.router.clusterController.clusterDataLoadedPercent'),
  /**
   * run all processes and cluster's data loading
   */
  initialize: function(){
    App.router.get('clusterController').loadClusterData();
  },

  dataLoading: function () {
    var self = this;
    var dfd = $.Deferred();
    if (App.router.get('clusterController.isLoaded')) {
      dfd.resolve();
    } else {
      var interval = setInterval(function () {
        if (self.get('isClusterDataLoaded')) {
          dfd.resolve();
          clearInterval(interval);
        }
      }, 50);
    }
    return dfd.promise();
  },

  startPolling: function () {
    if (App.router.get('applicationController.isExistingClusterDataLoaded')) {
      App.router.get('updateController').set('isWorking', true);
      App.router.get('backgroundOperationsController').set('isWorking', true);
    }
  }.observes('App.router.applicationController.isExistingClusterDataLoaded'),
  stopPolling: function(){
    App.router.get('updateController').set('isWorking', false);
    App.router.get('backgroundOperationsController').set('isWorking', false);
  },

  reloadTimeOut: null,

  pageReload: function () {

    clearTimeout(this.get("reloadTimeOut"));

    this.set('reloadTimeOut',
        setTimeout(function () {
          if (App.clusterStatus.get('isInstalled')) {
            location.reload();
          }
        }, App.pageReloadTime)
    );
  }.observes("App.router.location.lastSetURL", "App.clusterStatus.isInstalled"),

  scRequest: function(request) {
    return App.router.get('mainServiceController').get(request);
  },

  isAllServicesInstalled: function() {
    return this.scRequest('isAllServicesInstalled');
  }.property('App.router.mainServiceController.content.content.@each',
      'App.router.mainServiceController.content.content.length'),

  isStartAllDisabled: function() {
    return this.scRequest('isStartAllDisabled');
  }.property('App.router.mainServiceController.isStartStopAllClicked',
      'App.router.mainServiceController.content.@each.healthStatus'),

  isStopAllDisabled: function() {
    return this.scRequest('isStopAllDisabled');
  }.property('App.router.mainServiceController.isStartStopAllClicked',
      'App.router.mainServiceController.content.@each.healthStatus'),

  gotoAddService: function() {
    App.router.get('mainServiceController').gotoAddService();
  },

  startAllService: function(event){
    App.router.get('mainServiceController').startAllService(event);
  },

  stopAllService: function(event){
    App.router.get('mainServiceController').stopAllService(event);
  },

  /**
   * check server version and web client version
   */
  checkServerClientVersion: function () {
    var dfd = $.Deferred();
    var self = this;
    self.getServerVersion().done(function () {
      dfd.resolve();
    });
    return dfd.promise();
  },
  getServerVersion: function(){
    return App.ajax.send({
      name: 'ambari.service.load_server_version',
      sender: this,
      success: 'getServerVersionSuccessCallback',
      error: 'getServerVersionErrorCallback'
    });
  },
  getServerVersionSuccessCallback: function (data) {
    var clientVersion = App.get('version');
    var serverVersion = (data.RootServiceComponents.component_version).toString();
    this.set('ambariServerVersion', serverVersion);
    if (clientVersion) {
      this.set('versionConflictAlertBody', Em.I18n.t('app.versionMismatchAlert.body').format(serverVersion, clientVersion));
      this.set('isServerClientVersionMismatch', clientVersion != serverVersion);
    } else {
      this.set('isServerClientVersionMismatch', false);
    }
    App.set('isManagedMySQLForHiveEnabled', App.config.isManagedMySQLForHiveAllowed(data.RootServiceComponents.properties['server.os_family']));
  },
  getServerVersionErrorCallback: function () {
    console.log('ERROR: Cannot load Ambari server version');
  }

});

});

require.register("controllers/main/admin", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminController = Em.Controller.extend({
  name: 'mainAdminController',

  /**
   * @type {string}
   * @default null
   */
  category: null,

  /**
   * Check if access page available.
   * Turn on if YARN service is installed with Application Timeline Server component and TEZ installed too.
   *
   * @type {Boolean}
   */
  isAccessAvailable: function () {
    var dependencies = {
      services: ['YARN', 'TEZ']
    };
    var serviceNames = App.Service.find().mapProperty('serviceName')
      .filter(function (serviceName) {
        return dependencies.services.contains(serviceName);
      });
    var isAppTimelineServerAvailable = App.HostComponent.find().someProperty('componentName', 'APP_TIMELINE_SERVER');
    return (dependencies.services.length === serviceNames.length && isAppTimelineServerAvailable);
  }.property()
});
});

require.register("controllers/main/admin/advanced", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminAdvancedController = Em.Controller.extend({
  name:'mainAdminAdvancedController',
  uninstall: function(event){
    var params = event.context;
    App.ModalPopup.show({
      uninstallParams: params,
      header: Em.I18n.t('admin.advanced.popup.header'),
      bodyClass: App.MainAdminAdvancedPasswordView.reopen({}), // layout: Em.Handlebars.compile()
      onPrimary: function(){
        var form = this.getForm();
        if(form) {
          if(form.isValid()) {
            console.warn("TODO: request for cluster uninstall");
          }
        }
        this.onClose();
      },
      onSecondary: function(){
        this.onClose();
      },

      getForm: function(){
        var form = false;
        $.each(this.get('_childViews'), function(){
          if(this.get('path') == "bodyClass") {
            return form = this.get('_childViews')[0];
          }
        });

        return form;
      }
    })
  }
});
});

require.register("controllers/main/admin/authentication", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminAuthenticationController = Em.Controller.extend({
  name:'mainAdminAuthenticationController',
  /**
   * save user form after editing
   * @param event
   */
  save:function (event) {
    var form = event.context;
    if (form.isValid()) {
      form.save();
      App.ModalPopup.show({
        header:Em.I18n.t('admin.authentication.form.testConfiguration'),
        body:form.get('resultText'),
        secondary:false,
        onPrimary:function () {
          this.hide();
        }
      });
    }
  },
  content:App.Authentication.find(1)
});
});

require.register("controllers/main/admin/highAvailability/nameNode/rollbackHA/rollback_wizard_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.RollbackHighAvailabilityWizardController = App.WizardController.extend({

  name: 'rollbackHighAvailabilityWizardController',

  totalSteps: 3,

  /**
   * Used for hiding back button in wizard
   */
  hideBackButton: true,

  content: Em.Object.create({
    controllerName: 'RollbackHighAvailabilityWizardController',
    cluster: null,
    masterComponentHosts: null,
    serviceName: 'MISC',
    hdfsUser:"hdfs",
    nameServiceId: '',
    selectedAddNNHost : null,
    selectedSNNHost : null,
    activeNNHost: null
  }),

  setCurrentStep: function (currentStep, completed) {
    this._super(currentStep, completed);
    App.clusterStatus.setClusterStatus({
      clusterName: this.get('content.cluster.name'),
      clusterState: 'ROLLBACK_HIGH_AVAILABILITY',
      wizardControllerName: 'rollbackHighAvailabilityWizardController',
      localdb: App.db.data
    });
  },

  /**
   * return new object extended from clusterStatusTemplate
   * @return Object
   */
  getCluster: function(){
    return jQuery.extend({}, this.get('clusterStatusTemplate'), {name: App.router.getClusterName()});
  },

  /**
   * save status of the cluster.
   * @param clusterStatus object with status,requestId fields.
   */
  saveClusterStatus: function (clusterStatus) {
    var oldStatus = this.toObject(this.get('content.cluster'));
    clusterStatus = jQuery.extend(oldStatus, clusterStatus);
    if (clusterStatus.requestId) {
      clusterStatus.requestId.forEach(function (requestId) {
        if (clusterStatus.oldRequestsId.indexOf(requestId) === -1) {
          clusterStatus.oldRequestsId.push(requestId)
        }
      }, this);
    }
    this.set('content.cluster', clusterStatus);
    this.save('cluster');
  },

  saveTasksStatuses: function(statuses){
    App.db.setRollbackHighAvailabilityWizardTasksStatuses(statuses);
    this.set('content.tasksStatuses', statuses);
  },

  saveRequestIds: function(requestIds){
    App.db.setRollbackHighAvailabilityWizardRequestIds(requestIds);
    this.set('content.requestIds', requestIds);
  },

  saveSelectedSNN: function(addNN){
    App.db.setRollBackHighAvailabilityWizardSelectedSNN(addNN);
    this.set('content.selectedAddNN', addNN);
  },

  saveSelectedAddNN: function(sNN){
    App.db.setRollBackHighAvailabilityWizardSelectedAddNN(sNN);
    this.set('content.selectedSNN', sNN);
  },

  loadAddNNHost: function () {
    var addNNHost = App.db.getRollBackHighAvailabilityWizardAddNNHost();
    this.set('content.addNNHost', addNNHost);
  },

  loadSNNHost: function () {
    var sNNHost = App.db.getRollBackHighAvailabilityWizardSNNHost();
    this.set('content.sNNHost', sNNHost);
  },

  loadTasksStatuses: function(){
    var sNNHost = App.db.getRollbackHighAvailabilityWizardTasksStatuses();
    this.set('content.tasksStatuses', sNNHost);
  },

  loadRequestIds: function(){
    var requestIds = App.db.getRollbackHighAvailabilityWizardRequestIds();
    this.set('content.requestIds', requestIds);
  },

  /**
   * Load data for all steps until <code>current step</code>
   */
  loadAllPriorSteps: function () {
    var step = this.get('currentStep');
    switch (step) {
      case '3':
      case '2':
      case '1':
        this.loadSNNHost();
        this.loadAddNNHost();
        this.load('cluster');
    }
  },

  /**
   * Remove all loaded data.
   * Created as copy for App.router.clearAllSteps
   */
  clearAllSteps: function () {
    this.clearInstallOptions();
    // clear temporary information stored during the install
    this.set('content.cluster', this.getCluster());
  },

  clearTasksData: function () {
    this.saveTasksStatuses(undefined);
    this.saveRequestIds(undefined);
  },

  /**
   * Clear all temporary data
   */
  finish: function () {
    this.setCurrentStep('1');
    this.clearAllSteps();
    App.router.get('updateController').updateAll();
  }
});

});

require.register("controllers/main/admin/highAvailability/nameNode/rollbackHA/step1_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.RollbackHighAvailabilityWizardStep1Controller = Em.Controller.extend({

  name:"rollbackHighAvailabilityWizardStep1Controller"

});


});

require.register("controllers/main/admin/highAvailability/nameNode/rollbackHA/step2_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.RollbackHighAvailabilityWizardStep2Controller = App.HighAvailabilityWizardStep4Controller.extend({
  name:"rollbackHighAvailabilityWizardStep2Controller",

  pullCheckPointStatus: function () {
    var hostName = this.get('content.activeNNHost');
    App.ajax.send({
      name: 'admin.high_availability.getNnCheckPointStatus',
      sender: this,
      data: {
        hostName: hostName
      },
      success: 'checkNnCheckPointStatus'
    });
  }
});


});

require.register("controllers/main/admin/highAvailability/nameNode/rollbackHA/step3_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.RollbackHighAvailabilityWizardStep3Controller = Em.Controller.extend({
  name:"rollbackHighAvailabilityWizardStep3Controller"
});

});

require.register("controllers/main/admin/highAvailability/nameNode/rollback_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
require('controllers/main/admin/highAvailability/progress_controller');

App.HighAvailabilityRollbackController = App.HighAvailabilityProgressPageController.extend({

  name: "highAvailabilityRollbackController",

  failedTask: null,
  configsSaved: false,
  deletedHdfsClients: 0,
  numOfDelOperations: 0,
  isRollback: true,
  hostsToPerformDel: [],

  commands: [
    'stopAllServices',
    'restoreHBaseConfigs',
    'stopFailoverControllers',
    'deleteFailoverControllers',
    'stopStandbyNameNode',
    'stopNameNode',
    'restoreHDFSConfigs',
    'enableSecondaryNameNode',
    'stopJournalNodes',
    'deleteJournalNodes',
    'deleteAdditionalNameNode',
    'startAllServices'
  ],

  loadStep: function () {
    console.warn('func: loadStep');
    this.initData();
    this.clearStep();
    this.loadTasks();
    this.addObserver('tasks.@each.status', this, 'onTaskStatusChange');
    this.onTaskStatusChange();
  },

  initData: function () {
    console.warn('func: initData');
    this.loadMasterComponentHosts();
    this.loadFailedTask();
    this.loadHdfsClientHosts();
  },

  setCommandsAndTasks: function(tmpTasks) {
    console.warn('func: setCommandsAndTasks');
    var fTask = this.get('failedTask');
    var index = [
      'deleteSNameNode',
      'startAllServices',
      'reconfigureHBase',
      'startZKFC',
      'installZKFC',
      'startSecondNameNode',
      'startNameNode',
      'startZooKeeperServers',
      'reconfigureHDFS',
      'disableSNameNode',
      'startJournalNodes',
      'installJournalNodes',
      'installNameNode',
      'stopAllServices'
    ].indexOf(fTask.command);

    if(index > 6){
      --index;
    }
    var newCommands = this.get('commands').splice(index);
    this.set('commands', newCommands);
    var newTasks = tmpTasks.splice(index);
    for (var i = 0; i < newTasks.length; i++) {
      newTasks[i].id = i;
    }
    this.set('tasks', newTasks);
    var hbaseTask = this.get('tasks').findProperty('command', 'restoreHBaseConfigs');
    if (!App.Service.find().someProperty('serviceName', 'HBASE') && hbaseTask) {
      this.get('tasks').splice(hbaseTask.get('id'), 1);
    }
  },

  clearStep: function () {
    console.warn('func: clearStep');
    this.set('isSubmitDisabled', true);
    this.set('tasks', []);
    this.set('logs', []);
    this.set('currentRequestIds', []);
    var commands = this.get('commands');
    var tmpTasks = [];
    for (var i = 0; i < commands.length; i++) {
      tmpTasks.pushObject(Ember.Object.create({
        title: Em.I18n.t('admin.highAvailability.rollback.task' + i + '.title'),
        status: 'PENDING',
        id: i,
        command: commands[i],
        showRetry: false,
        showRollback: false,
        showSkip: false,
        name: Em.I18n.t('admin.highAvailability.rollback.task' + i + '.title'),
        displayName: Em.I18n.t('admin.highAvailability.rollback.task' + i + '.title'),
        progress: 0,
        isRunning: false,
        hosts: []
      }));
    }
    this.setCommandsAndTasks(tmpTasks);
  },

  onTaskStatusChange: function () {
    console.warn('func: onTaskStatusChange');
    if (!this.get('tasks').someProperty('status', 'IN_PROGRESS') && !this.get('tasks').someProperty('status', 'QUEUED') && !this.get('tasks').someProperty('status', 'FAILED')) {
      var nextTask = this.get('tasks').findProperty('status', 'PENDING');
      if (nextTask) {
        console.warn('func: onTaskStatusChange1');
        this.set('status', 'IN_PROGRESS');
        this.setTaskStatus(nextTask.get('id'), 'QUEUED');
        this.set('currentTaskId', nextTask.get('id'));
        this.runTask(nextTask.get('id'));
      } else {
        console.warn('func: onTaskStatusChange2');
        this.set('status', 'COMPLETED');
        this.set('isSubmitDisabled', false);
      }
    } else if (this.get('tasks').someProperty('status', 'FAILED')) {
      console.warn('func: onTaskStatusChange3');
      this.set('status', 'FAILED');
      this.get('tasks').findProperty('status', 'FAILED').set('showRetry', true);
      this.get('tasks').findProperty('status', 'FAILED').set('showSkip', true);
    }
    this.get('tasks').filterProperty('status','COMPLETED').setEach('showRetry', false);
    this.get('tasks').filterProperty('status','COMPLETED').setEach('showSkip', false);

    var statuses = this.get('tasks').mapProperty('status');
    var logs = this.get('tasks').mapProperty('hosts');
    var requestIds = this.get('currentRequestIds');
    console.warn('func: onTaskStatusChange4',statuses,logs,requestIds);
    this.saveTasksStatuses(statuses);
    this.saveRequestIds(requestIds);
    this.saveLogs(logs);
    App.clusterStatus.setClusterStatus({
      clusterName: this.get('content.cluster.name'),
      clusterState: 'HIGH_AVAILABILITY_ROLLBACK',
      wizardControllerName: 'highAvailabilityRollbackController',
      localdb: App.db.data
    });
  },

  skipTask: function () {
    console.warn('func: skipTask');
    var task = this.get('tasks').findProperty('status', 'FAILED');
    task.set('showRetry', false);
    task.set('showSkip', false);
    task.set('status', 'COMPLETED');
  },

  retryTask: function () {
    console.warn('func: retryTask');
    var task = this.get('tasks').findProperty('status', 'FAILED');
    task.set('showRetry', false);
    task.set('showSkip', false);
    task.set('status', 'PENDING');
  },

  onTaskCompleted: function () {
    console.warn('func: onTaskCompleted');
    var curTaskStatus = this.getTaskStatus(this.get('currentTaskId'));
    if (curTaskStatus != 'FAILED' && curTaskStatus != 'TIMEDOUT' && curTaskStatus != 'ABORTED') {
      this.setTaskStatus(this.get('currentTaskId'), 'COMPLETED');
    }
  },

  getTaskStatus: function (taskId) {
    console.warn('func: getTaskStatus');
    return this.get('tasks').findProperty('id', taskId).get('status');
  },

  loadFailedTask: function(){
    console.warn('func: loadFailedTask');
    var failedTask = App.db.getHighAvailabilityWizardFailedTask();
    this.set('failedTask', failedTask);
  },

  done: function () {
    if (!this.get('isSubmitDisabled')) {
      this.removeObserver('tasks.@each.status', this, 'onTaskStatusChange');
      this.popup.proceedOnClose();
    }
  },

  stopAllServices: function(){
    console.warn('func: stopAllServices');
    App.ajax.send({
      name: 'common.services.update',
      data: {
        context: "Stop all services",
        "ServiceInfo": {
          "state": "INSTALLED"
        }
      },
      sender: this,
      success: 'startPolling',
      error: 'onTaskError'
    });
  },
  restoreHBaseConfigs: function(){
    console.warn('func: restoreHBaseConfigs');
    this.loadConfigTag("hbaseSiteTag");
    var hbaseSiteTag = this.get("content.hbaseSiteTag");
    App.ajax.send({
      name: 'admin.high_availability.load_hbase_configs',
      sender: this,
      data: {
        hbaseSiteTag: hbaseSiteTag
      },
      success: 'onLoadHbaseConfigs',
      error: 'onTaskError'
    });
  },

  stopFailoverControllers: function(){
    console.warn('func: stopFailoverControllers');
    var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'NAMENODE').mapProperty('hostName');
    this.updateComponent('ZKFC', hostNames, "HDFS", "Stop");
  },
  deleteFailoverControllers: function(){
    console.warn('func: deleteFailoverControllers');
    var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'NAMENODE').mapProperty('hostName');
    this.checkBeforeDelete('ZKFC', hostNames);
  },
  stopStandbyNameNode: function(){
    console.warn('func: stopStandbyNameNode');
    var hostName = this.get('content.masterComponentHosts').findProperty('isAddNameNode', true).hostName;
    this.updateComponent('NAMENODE', hostName, "HDFS", "Stop");
  },
  stopNameNode: function(){
    console.warn('func: stopNameNode');
    var hostName = this.get('content.masterComponentHosts').findProperty('isCurNameNode').hostName;
    this.updateComponent('NAMENODE', hostName, "HDFS", "Stop");
  },
  restoreHDFSConfigs: function(){
    console.warn('func: restoreHDFSConfigs');
    this.unInstallHDFSClients();
  },
  enableSecondaryNameNode: function(){
    console.warn('func: enableSecondaryNameNode');
    var hostName = this.get('content.masterComponentHosts').findProperty('component', 'SECONDARY_NAMENODE').hostName;
    this.updateComponent('SECONDARY_NAMENODE', hostName, "HDFS", "Install", hostName.length);
  },
  stopJournalNodes: function(){
    console.warn('func: stopJournalNodes');
    var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'JOURNALNODE').mapProperty('hostName');
    this.updateComponent('JOURNALNODE', hostNames, "HDFS", "Stop");
  },
  deleteJournalNodes: function(){
    console.warn('func: deleteJournalNodes');
    var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'JOURNALNODE').mapProperty('hostName');
    this.unInstallComponent('JOURNALNODE', hostNames);
  },
  deleteAdditionalNameNode: function(){
    console.warn('func: deleteAdditionalNameNode');
    var hostNames = this.get('content.masterComponentHosts').filterProperty('isAddNameNode', true).mapProperty('hostName');
    this.unInstallComponent('NAMENODE', hostNames);
  },
  startAllServices: function(){
    console.warn('func: startAllServices');
    App.ajax.send({
      name: 'common.services.update',
      data: {
        context: "Start all services",
        "ServiceInfo": {
          "state": "STARTED"
        }
      },
      sender: this,
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  onLoadHbaseConfigs: function (data) {
    console.warn('func: onLoadHbaseConfigs');
    var hbaseSiteProperties = data.items.findProperty('type', 'hbase-site').properties;
    App.ajax.send({
      name: 'admin.high_availability.save_configs',
      sender: this,
      data: {
        siteName: 'hbase-site',
        properties: hbaseSiteProperties
      },
      success: 'onTaskCompleted',
      error: 'onTaskError'
    });
  },

  onDeletedHDFSClient: function () {
    console.warn('func: onDeletedHDFSClient');
    var deletedHdfsClients = this.get('deletedHdfsClients');
    var hostName = this.get("content.hdfsClientHostNames");
    var notDeletedHdfsClients = hostName.length - deletedHdfsClients;
    if (notDeletedHdfsClients > 1 && hostName.length != 1 ) {
      this.set('deletedHdfsClients', deletedHdfsClients+1);
      return;
    }
    this.loadConfigTag("hdfsSiteTag");
    this.loadConfigTag("coreSiteTag");
    var hdfsSiteTag = this.get("content.hdfsSiteTag");
    var coreSiteTag = this.get("content.coreSiteTag");
    App.ajax.send({
      name: 'admin.high_availability.load_configs',
      sender: this,
      data: {
        hdfsSiteTag: hdfsSiteTag,
        coreSiteTag: coreSiteTag
      },
      success: 'onLoadConfigs',
      error: 'onTaskError'
    });
  },

  onLoadConfigs: function (data) {
    console.warn('func: onLoadConfigs');
    this.set('configsSaved', false);
    App.ajax.send({
      name: 'admin.high_availability.save_configs',
      sender: this,
      data: {
        siteName: 'hdfs-site',
        properties: data.items.findProperty('type', 'hdfs-site').properties
      },
      success: 'onHdfsConfigsSaved',
      error: 'onTaskError'
    });
    App.ajax.send({
      name: 'admin.high_availability.save_configs',
      sender: this,
      data: {
        siteName: 'core-site',
        properties: data.items.findProperty('type', 'core-site').properties
      },
      success: 'onHdfsConfigsSaved',
      error: 'onTaskError'
    });
  },

  onHdfsConfigsSaved: function () {
    console.warn('func: onHdfsConfigsSaved');
    if (!this.get('configsSaved')) {
      this.set('configsSaved', true);
      return;
    }
    this.onTaskCompleted();
  },

  unInstallHDFSClients: function () {
    console.warn('func: unInstallHDFSClients');
    var hostName = this.get("content.hdfsClientHostNames");
    for (var i = 0; i < hostName.length; i++) {
      App.ajax.send({
        name: 'common.delete.host_component',
        sender: this,
        data: {
          componentName: 'HDFS_CLIENT',
          hostName: hostName[i]
        },
        success: 'onDeletedHDFSClient',
        error: 'onTaskError'
      });
    }
  },

  unInstallComponent: function (componentName, hostName) {
    console.warn('func: unInstallComponent');
    if (!(hostName instanceof Array)) {
      hostName = [hostName];
    }
    for (var i = 0; i < hostName.length; i++) {
      App.ajax.send({
        name: 'common.host.host_component.passive',
        sender: this,
        data: {
          hostName: hostName[i],
          componentName: componentName,
          passive_state: "ON",
          taskNum: hostName.length,
          callback: 'checkBeforeDelete'
        },
        success: 'checkResult',
        error: 'checkResult'
      });
    }
  },

  checkBeforeDelete: function (componentName, hostName){
    console.warn('func: checkBeforeDelete');
    this.set('hostsToPerformDel', []);
    if (!(hostName instanceof Array)) {
      hostName = [hostName];
    }
    for (var i = 0; i < hostName.length; i++) {
      App.ajax.send({
        name: 'admin.high_availability.getHostComponent',
        sender: this,
        data: {
          componentName: componentName,
          hostName: hostName[i],
          taskNum: hostName.length,
          callback: 'deleteComponent'
        },
        success: 'checkResult',
        error: 'checkResult'
      });
    }
  },

  checkResult: function () {
    console.warn('func: checkResult');
    var callback = arguments[2].callback;
    var hostName = arguments[2].hostName;
    var componentName = arguments[2].componentName;
    var taskNum = arguments[2].taskNum;
    var hostsToPerformDel = this.get('hostsToPerformDel');
    if(arguments[1] != 'error'){
      hostsToPerformDel.push({
        hostName: hostName,
        isOnHost: true
      });
    }else{
      hostsToPerformDel.push({
        hostName: 'error',
        isOnHost: false
      });
    }
    if(hostsToPerformDel.length == taskNum){
      var hostsForDel = hostsToPerformDel.filterProperty('isOnHost', true).mapProperty('hostName');
      this.set('hostsToPerformDel', []);
      if(hostsForDel.length == 0){
        this.onTaskCompleted();
        return;
      }
      this[callback](componentName, hostsForDel);
    }
  },

  deleteComponent: function (componentName, hostName) {
    console.warn('func: deleteComponent');
    if (!(hostName instanceof Array)) {
      hostName = [hostName];
    }
    this.set('numOfDelOperations', hostName.length);
    for (var i = 0; i < hostName.length; i++) {
      App.ajax.send({
        name: 'common.delete.host_component',
        sender: this,
        data: {
          componentName: componentName,
          hostName: hostName[i]
        },
        success: 'onDeleteComplete',
        error: 'onTaskError'
      });
    }
  },

  onDeleteComplete: function () {
    console.warn('func: onDeleteComplete');
    var leftOp = this.get('numOfDelOperations');
    if(leftOp > 1){
      this.set('numOfDelOperations', leftOp-1);
      return;
    }
    this.onTaskCompleted();
  }

});

});

require.register("controllers/main/admin/highAvailability/nameNode/step1_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

require('controllers/main/admin/serviceAccounts_controller');

App.HighAvailabilityWizardStep1Controller = Em.Controller.extend({
  name: "highAvailabilityWizardStep1Controller",

  isNameServiceIdValid: function () {
    return /^([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])$/.test(this.get('content.nameServiceId'));
  }.property('content.nameServiceId'),

  next: function () {
    if (this.get('isNameServiceIdValid')) {
      App.router.send('next');
    }
  }

});


});

require.register("controllers/main/admin/highAvailability/nameNode/step2_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

require('controllers/wizard/step5_controller');

App.HighAvailabilityWizardStep2Controller = App.WizardStep5Controller.extend({

  name:"highAvailabilityWizardStep2Controller",

  /**
   * master components which could be assigned to multiple hosts in HA wizard
   */
  multipleComponentsHaWizard: ['NAMENODE', 'JOURNALNODE'],

  /**
   * master components supported by Ambari
   */

  multipleComponents: ['NAMENODE', 'JOURNALNODE','ZOOKEEPER_SERVER','HBASE_MASTER','RESOURCEMANAGER'],

  /**
   * overrides method in wizardStep5Controller
   * it must be empty as it shouldn't be run
   */
  masterHostMappingObserver: function() {},

  /**
   * Load services info to appropriate variable and return masterComponentHosts
   * @return Array
   */
  masterHostMapping:function () {
    var mapping = [], mappingObject, self = this, mappedHosts, hostObj, hostInfo;
    //get the unique assigned hosts and find the master services assigned to them


    mappedHosts = this.get("selectedServicesMasters").mapProperty("selectedHost").uniq();

    mappedHosts.forEach(function (item) {

      hostObj = self.get("hosts").findProperty("host_name", item);
      console.log("Name of the host is: " + hostObj.host_name);

      var masterServices = self.get("selectedServicesMasters").filterProperty("selectedHost", item);
      masterServices.forEach(function(item){
        if(this.get('multipleComponentsHaWizard').contains(item.component_name)){
          item.set('color','green');
        }else{
          item.set('color','grey');
        }
      }, this);

      mappingObject = Ember.Object.create({
        host_name:item,
        hostInfo:hostObj.host_info,
        masterServices:masterServices
      });

      mapping.pushObject(mappingObject);
    }, this);

    return mapping.sortProperty('host_name');
  }.property("selectedServicesMasters.@each.selectedHost"),

  /**
   * Put master components to <code>selectedServicesMasters</code>, which will be automatically rendered in template
   * @param masterComponents
   */
  renderComponents:function (masterComponents) {
    var services = this.get('content.services').filterProperty('isInstalled', true).mapProperty('serviceName'); //list of shown services

    var result = [];

    var curNameNode = masterComponents.findProperty('component_name',"NAMENODE");
    curNameNode.isCurNameNode = true;
    curNameNode.serviceComponentId = 0;

    //Create JOURNALNODE
    for (var index = 0; index < 3; index++) {
      masterComponents.push(
        {
          component_name: "JOURNALNODE",
          display_name: "JournalNode",
          isServiceCoHost: false,
          isInstalled: false,
          selectedHost: this.get("hosts")[index].get("host_name"),
          serviceId: "HDFS",
          serviceComponentId: index
        }
      )
    }
    //Create Additional NameNode
    masterComponents.push(
      {
        component_name: "NAMENODE",
        display_name: "NameNode",
        isServiceCoHost: false,
        isInstalled: false,
        selectedHost: this.get("hosts").mapProperty('host_name').without(curNameNode.selectedHost)[0],
        serviceId: "HDFS",
        isAddNameNode: true,
        serviceComponentId: 1
      }
    );

    masterComponents.forEach(function (item) {
      var componentObj = Ember.Object.create(item);
      result.push(componentObj);
    }, this);

    this.set("selectedServicesMasters", result);

    var components = result.filterProperty('component_name',"NAMENODE");
    components.push.apply(components, result.filterProperty('component_name',"JOURNALNODE"));

    this.set('servicesMasters', components);
    this.set('componentToRebalance', "NAMENODE");
    this.incrementProperty('rebalanceComponentHostsCounter');
    this.set('componentToRebalance', "JOURNALNODE");
    this.incrementProperty('rebalanceComponentHostsCounter');
  }

});


});

require.register("controllers/main/admin/highAvailability/nameNode/step3_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HighAvailabilityWizardStep3Controller = Em.Controller.extend({
  name: "highAvailabilityWizardStep3Controller",
  selectedService: null,
  stepConfigs: [],
  serverConfigData: {},
  haConfig: $.extend(true, {}, require('data/HDP2/ha_properties').haConfig),
  once: false,
  isLoaded: false,
  versionLoaded: true,

  /**
   * Map of sites and properties to delete
   * @type Object
   */
  configsToRemove: {
    'hdfs-site': ['dfs.namenode.secondary.http-address']
  },

  clearStep: function () {
    this.get('stepConfigs').clear();
    this.serverConfigData = {};
  },

  loadStep: function () {
    this.clearStep();
    this.loadConfigsTags();
  },

  loadConfigsTags: function () {
    App.ajax.send({
      name: 'config.tags',
      sender: this,
      success: 'onLoadConfigsTags',
      error: 'onTaskError'
    });
  },


  onLoadConfigsTags: function (data) {
    var urlParams = [];
    var hdfsSiteTag = data.Clusters.desired_configs['hdfs-site'].tag;
    var coreSiteTag = data.Clusters.desired_configs['core-site'].tag;
    urlParams.push('(type=hdfs-site&tag=' + hdfsSiteTag + ')');
    urlParams.push('(type=core-site&tag=' + coreSiteTag + ')');
    this.set("hdfsSiteTag", {name : "hdfsSiteTag", value : hdfsSiteTag});
    this.set("coreSiteTag", {name : "coreSiteTag", value : coreSiteTag});

    if (App.Service.find().someProperty('serviceName', 'HBASE')) {
      var hbaseSiteTag = data.Clusters.desired_configs['hbase-site'].tag;
      urlParams.push('(type=hbase-site&tag=' + hbaseSiteTag + ')');
      this.set("hbaseSiteTag", {name : "hbaseSiteTag", value : hbaseSiteTag});
    }
    App.ajax.send({
      name: 'admin.get.all_configurations',
      sender: this,
      data: {
        urlParams: urlParams.join('|')
      },
      success: 'onLoadConfigs',
      error: 'onTaskError'
    });
  },

  onLoadConfigs: function (data) {
    this.set('serverConfigData',data);
    this.removeConfigs(this.get('configsToRemove'), this.get('serverConfigData'));
    this.tweakServiceConfigs(this.get('haConfig.configs'));
    this.renderServiceConfigs(this.get('haConfig'));
    this.set('isLoaded', true);
  },


  tweakServiceConfigs: function(configs) {
    var nameServiceId = this.get('content.nameServiceId');
    var nameServiceConfig = configs.findProperty('name','dfs.nameservices');
    this.setConfigInitialValue(nameServiceConfig,nameServiceId);
    var defaultFsConfig = configs.findProperty('name','fs.defaultFS');
    this.setConfigInitialValue(defaultFsConfig, "hdfs://" + nameServiceId);
    this.tweakServiceConfigNames(configs,nameServiceId);
    this.tweakServiceConfigValues(configs,nameServiceId);
  },

  tweakServiceConfigNames: function(configs,nameServiceId) {
    var regex = new RegExp("\\$\\{dfs.nameservices\\}","g");
    configs.forEach(function(config) {
      if (config.name.contains("${dfs.nameservices}")) {
        config.name = config.name.replace(regex,nameServiceId);
        config.displayName = config.displayName.replace(regex,nameServiceId);
      }
    }, this);
  },

  tweakServiceConfigValues: function(configs,nameServiceId) {
    var currentNameNodeHost = this.get('content.masterComponentHosts').findProperty('isCurNameNode').hostName;
    var newNameNodeHost = this.get('content.masterComponentHosts').findProperty('isAddNameNode').hostName;
    var journalNodeHosts = this.get('content.masterComponentHosts').filterProperty('component', 'JOURNALNODE').mapProperty('hostName');
    var zooKeeperHosts = this.get('content.masterComponentHosts').filterProperty('component', 'ZOOKEEPER_SERVER').mapProperty('hostName');
    var config = configs.findProperty('name','dfs.namenode.rpc-address.' + nameServiceId + '.nn1');
    this.setConfigInitialValue(config,currentNameNodeHost + ':8020');
    config = configs.findProperty('name','dfs.namenode.rpc-address.' + nameServiceId + '.nn2');
    this.setConfigInitialValue(config,newNameNodeHost + ':8020');
    config = configs.findProperty('name','dfs.namenode.http-address.' + nameServiceId + '.nn1');
    this.setConfigInitialValue(config,currentNameNodeHost + ':50070');
    config = configs.findProperty('name','dfs.namenode.http-address.' + nameServiceId + '.nn2');
    this.setConfigInitialValue(config,newNameNodeHost + ':50070');
    config = configs.findProperty('name','dfs.namenode.https-address.' + nameServiceId + '.nn1');
    this.setConfigInitialValue(config,currentNameNodeHost + ':50470');
    config = configs.findProperty('name','dfs.namenode.https-address.' + nameServiceId + '.nn2');
    this.setConfigInitialValue(config,newNameNodeHost + ':50470');
    config = configs.findProperty('name','dfs.namenode.shared.edits.dir');
    this.setConfigInitialValue(config,'qjournal://' + journalNodeHosts[0] + ':8485;' + journalNodeHosts[1] + ':8485;' + journalNodeHosts[2] + ':8485/' + nameServiceId);
    config = configs.findProperty('name','ha.zookeeper.quorum');
    this.setConfigInitialValue(config,zooKeeperHosts[0] + ':2181,' + zooKeeperHosts[1] + ':2181,' + zooKeeperHosts[2] + ':2181');
    config = configs.findProperty('name','hbase.rootdir');
    if (App.Service.find().someProperty('serviceName', 'HBASE')) {
     var value = this.get('serverConfigData.items').findProperty('type', 'hbase-site').properties['hbase.rootdir'].replace(/\/\/[^\/]*/, '//' + nameServiceId);
     this.setConfigInitialValue(config,value);
    }
    config = configs.findProperty('name','dfs.journalnode.edits.dir');
    if (App.get('isHadoopWindowsStack') && App.Service.find().someProperty('serviceName', 'HDFS')) {
     var value = this.get('serverConfigData.items').findProperty('type', 'hdfs-site').properties['dfs.journalnode.edits.dir'];
     this.setConfigInitialValue(config, value);
    }
  },

  /**
   * Find and remove config properties in <code>serverConfigData</code>
   * @param configsToRemove - map of config sites and properties to remove
   * @param configs - configuration object
   * @returns {Object}
   */
  removeConfigs:function (configsToRemove, configs) {
    Em.keys(configsToRemove).forEach(function(site){
      var siteConfigs = configs.items.findProperty('type', site);
      if (siteConfigs) {
        configsToRemove[site].forEach(function (property) {
          delete siteConfigs.properties[property];
        });
      }
    });
    return configs;
  },

  setConfigInitialValue: function(config,value) {
    config.value = value;
    config.defaultValue = value;
  },

  renderServiceConfigs: function (_serviceConfig) {
    var serviceConfig = App.ServiceConfig.create({
      serviceName: _serviceConfig.serviceName,
      displayName: _serviceConfig.displayName,
      configCategories: [],
      showConfig: true,
      configs: []
    });

    _serviceConfig.configCategories.forEach(function (_configCategory) {
      if (App.Service.find().someProperty('serviceName', _configCategory.name)) {
        serviceConfig.configCategories.pushObject(_configCategory);
      }
    }, this);

    this.loadComponentConfigs(_serviceConfig, serviceConfig);

    this.get('stepConfigs').pushObject(serviceConfig);
    this.set('selectedService', this.get('stepConfigs').objectAt(0));
    this.once = true;
  },

  /**
   * Load child components to service config object
   * @param _componentConfig
   * @param componentConfig
   */
  loadComponentConfigs: function (_componentConfig, componentConfig) {
    _componentConfig.configs.forEach(function (_serviceConfigProperty) {
      var serviceConfigProperty = App.ServiceConfigProperty.create(_serviceConfigProperty);
      componentConfig.configs.pushObject(serviceConfigProperty);
      serviceConfigProperty.set('isEditable', serviceConfigProperty.get('isReconfigurable'));
      serviceConfigProperty.validate();
    }, this);
  },

  isNextDisabled: function () {
    return !this.get('isLoaded');
  }.property('isLoaded')

});


});

require.register("controllers/main/admin/highAvailability/nameNode/step4_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

require('controllers/main/admin/serviceAccounts_controller');

App.HighAvailabilityWizardStep4Controller = Em.Controller.extend({

  name:"highAvailabilityWizardStep4Controller",

  POLL_INTERVAL: 1000,

  isNextEnabled: false,

  isNameNodeStarted: true,

  pullCheckPointStatus: function () {
    var hostName = this.get('content.masterComponentHosts').findProperty('isCurNameNode', true).hostName;
    App.ajax.send({
      name: 'admin.high_availability.getNnCheckPointStatus',
      sender: this,
      data: {
        hostName: hostName
      },
      success: 'checkNnCheckPointStatus'
    });
  },

  checkNnCheckPointStatus: function (data) {
    this.set('isNameNodeStarted', data.HostRoles.desired_state === 'STARTED');
    var self = this;
    var journalTransactionInfo = $.parseJSON(Em.get(data, 'metrics.dfs.namenode.JournalTransactionInfo'));
    var isInSafeMode = !Em.isEmpty(Em.get(data, 'metrics.dfs.namenode.Safemode'));
    // in case when transaction info absent or invalid return 2 which will return false in next `if` statement
    journalTransactionInfo = !!journalTransactionInfo ? (parseInt(journalTransactionInfo.LastAppliedOrWrittenTxId) - parseInt(journalTransactionInfo.MostRecentCheckpointTxId)) : 2;
    if (journalTransactionInfo <= 1 && isInSafeMode) {
      this.set("isNextEnabled", true);
      return;
    }

    window.setTimeout(function () {
      self.pullCheckPointStatus();
    }, self.POLL_INTERVAL);
  },

  done: function () {
    if (this.get('isNextEnabled')) {
      App.get('router.mainAdminKerberosController').getKDCSessionState(function() {
        App.router.send("next");
      });
    }
  }

});


});

require.register("controllers/main/admin/highAvailability/nameNode/step5_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HighAvailabilityWizardStep5Controller = App.HighAvailabilityProgressPageController.extend({

  name:"highAvailabilityWizardStep5Controller",

  isHA: true,

  commands: ['stopAllServices', 'installNameNode', 'installJournalNodes', 'reconfigureHDFS', 'startJournalNodes', 'disableSNameNode'],

  hdfsSiteTag : "",
  coreSiteTag : "",

  stopAllServices: function () {
    App.ajax.send({
      name: 'common.services.update',
      data: {
        context: "Stop all services",
        "ServiceInfo": {
          "state": "INSTALLED"
        }
      },
      sender: this,
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  installNameNode: function () {
    var hostName = this.get('content.masterComponentHosts').findProperty('isAddNameNode').hostName;
    this.createComponent('NAMENODE', hostName, "HDFS");
  },

  installJournalNodes: function () {
    App.ajax.send({
      name: 'admin.high_availability.create_journalnode',
      sender: this,
      success: 'onJournalNodeCreate',
      error: 'onJournalNodeCreate'
    });
  },

  onJournalNodeCreate: function () {
    var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'JOURNALNODE').mapProperty('hostName');
    this.createComponent('JOURNALNODE', hostNames, "HDFS");
  },

  startJournalNodes: function () {
    var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'JOURNALNODE').mapProperty('hostName');
    this.updateComponent('JOURNALNODE', hostNames, "HDFS", "Start");
  },

  disableSNameNode: function () {
    var hostName = this.get('content.masterComponentHosts').findProperty('component', 'SECONDARY_NAMENODE').hostName;
    App.ajax.send({
      name: 'common.host.host_component.passive',
      sender: this,
      data: {
        hostName: hostName,
        passive_state: "ON",
        componentName: 'SECONDARY_NAMENODE'
      },
      success: 'onTaskCompleted',
      error: 'onTaskError'
    });
  },

  reconfigureHDFS: function () {
    var data = this.get('content.serviceConfigProperties');
    if (App.router.get('mainAdminKerberosController.securityEnabled')) {
      this.reconfigureSecureHDFS();
    } else {
      this.updateConfigProperties(data);
    }
  },

  /**
   * Update service configurations
   * @param {Object} data - config object to update
   */
  updateConfigProperties: function(data) {
    var siteNames = ['hdfs-site','core-site'];
    var configData = this.reconfigureSites(siteNames, data);
    App.ajax.send({
      name: 'common.service.configurations',
      sender: this,
      data: {
        desired_config: configData
      },
      error: 'onTaskError',
      success: 'installHDFSClients'
    });
  },

  installHDFSClients: function () {
    var nnHostNames = this.get('content.masterComponentHosts').filterProperty('component', 'NAMENODE').mapProperty('hostName');
    var jnHostNames = this.get('content.masterComponentHosts').filterProperty('component', 'JOURNALNODE').mapProperty('hostName');
    var hostNames = nnHostNames.concat(jnHostNames).uniq();
    this.createComponent('HDFS_CLIENT', hostNames);
    App.router.get(this.get('content.controllerName')).saveHdfsClientHosts(hostNames);
    App.clusterStatus.setClusterStatus({
      clusterName: this.get('content.cluster.name'),
      clusterState: 'HIGH_AVAILABILITY_DEPLOY',
      wizardControllerName: this.get('content.controllerName'),
      localdb: App.db.data
    });
  },

  /**
   * Process configurations for hdfs on kerberized cluster.
   * Secure properties will be applied to hdfs and core site after installing JournalNode.
   * For this case we need to pull updated configurations from API, merge them with configurations created
   * during `Review` step and store new configurations.
   */
  reconfigureSecureHDFS: function() {
    App.router.get('highAvailabilityWizardStep3Controller').loadConfigsTags.call(this);
  },

  onLoadConfigsTags: function() {
    App.router.get('highAvailabilityWizardStep3Controller').onLoadConfigsTags.apply(this, [].slice.call(arguments));
  },

  onLoadConfigs: function(data) {
    var self = this;
    var configController = App.router.get('highAvailabilityWizardStep3Controller');
    var configData = $.extend({}, data);
    var configItems = data.items.map(function(item) {
      var fileName = Em.get(item, 'type');
      var configTypeObject = self.get('content.serviceConfigProperties').items.findProperty('type', fileName);
      if (configTypeObject) {
        $.extend(item.properties, configTypeObject.properties);
      }
      return item;
    });
    configItems = configController.removeConfigs(configController.get('configsToRemove'), {items: configItems});
    this.updateConfigProperties(configItems);
  }
});


});

require.register("controllers/main/admin/highAvailability/nameNode/step6_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HighAvailabilityWizardStep6Controller = Em.Controller.extend({

  name:"highAvailabilityWizardStep6Controller",

  POLL_INTERVAL: 1000,

  isNextEnabled: function(){
    //only 3 JournalNodes could be installed
    return (this.get('initJnCounter') === 3);
  }.property('initJnCounter'),

  initJnCounter: 0,

  pullCheckPointStatus: function () {
    this.set('initJnCounter', 0);
    var hostNames = this.get('content.masterComponentHosts').filterProperty('component', "JOURNALNODE").mapProperty('hostName');
    hostNames.forEach(function (hostName) {
      this.pullEachJnStatus(hostName);
    }, this);
  },

  pullEachJnStatus: function(hostName){
    App.ajax.send({
      name: 'admin.high_availability.getJnCheckPointStatus',
      sender: this,
      data: {
        hostName: hostName
      },
      success: 'checkJnCheckPointStatus'
    });
  },

  checkJnCheckPointStatus: function (data) {
    var self = this;
    var journalStatusInfo;
    if (data.metrics && data.metrics.dfs) {
      journalStatusInfo = $.parseJSON(data.metrics.dfs.journalnode.journalsStatus);
      if (journalStatusInfo[this.get('content.nameServiceId')] && journalStatusInfo[this.get('content.nameServiceId')].Formatted === "true") {
        this.set("initJnCounter", (this.get('initJnCounter') + 1));
        return;
      }
    }

    window.setTimeout(function () {
      self.pullEachJnStatus(data.HostRoles.host_name);
    }, self.POLL_INTERVAL);
  },

  done: function () {
    if (this.get('isNextEnabled')) {
      App.router.send('next');
    }
  }

});


});

require.register("controllers/main/admin/highAvailability/nameNode/step7_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HighAvailabilityWizardStep7Controller = App.HighAvailabilityProgressPageController.extend({

  name:"highAvailabilityWizardStep7Controller",

  isHA: true,

  commands: ['startZooKeeperServers', 'startNameNode'],

  startZooKeeperServers: function () {
    var hostNames = this.get('content.masterComponentHosts').filterProperty('component', 'ZOOKEEPER_SERVER').mapProperty('hostName');
    this.updateComponent('ZOOKEEPER_SERVER', hostNames, "ZOOKEEPER", "Start");
  },

  startNameNode: function () {
    var hostName = this.get('content.masterComponentHosts').findProperty('isCurNameNode').hostName;
    this.updateComponent('NAMENODE', hostName, "HDFS", "Start");
  }
});


});

require.register("controllers/main/admin/highAvailability/nameNode/step8_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HighAvailabilityWizardStep8Controller = Em.Controller.extend({

  name:"highAvailabilityWizardStep8Controller",

  done: function () {
    App.get('router.mainAdminKerberosController').getKDCSessionState(function() {
      App.router.send("next");
    });
  }

});


});

require.register("controllers/main/admin/highAvailability/nameNode/step9_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HighAvailabilityWizardStep9Controller = App.HighAvailabilityProgressPageController.extend({

  name:"highAvailabilityWizardStep9Controller",

  isHA: true,

  commands: ['startSecondNameNode', 'installZKFC', 'startZKFC', 'reconfigureHBase', 'deleteSNameNode', 'startAllServices'],

  hbaseSiteTag: "",

  initializeTasks: function () {
    this._super();
    if (!App.Service.find().someProperty('serviceName', 'HBASE')) {
      this.get('tasks').splice(this.get('tasks').findProperty('command', 'reconfigureHBase').get('id'), 1);
    }
  },

  startSecondNameNode: function () {
    var hostName = this.get('content.masterComponentHosts').findProperty('isAddNameNode', true).hostName;
    this.updateComponent('NAMENODE', hostName, "HDFS", "Start");
  },

  installZKFC: function () {
    App.ajax.send({
      name: 'admin.high_availability.create_zkfc',
      sender: this,
      success: 'onZKFCCreate',
      error: 'onZKFCCreate'
    });
  },

  onZKFCCreate: function () {
    var hostName = this.get('content.masterComponentHosts').filterProperty('component', 'NAMENODE').mapProperty('hostName');
    this.createComponent('ZKFC', hostName, "HDFS");
  },

  startZKFC: function () {
    var hostName = this.get('content.masterComponentHosts').filterProperty('component', 'NAMENODE').mapProperty('hostName');
    this.updateComponent('ZKFC', hostName, "HDFS", "Start");
  },

  reconfigureHBase: function () {
    var data = this.get('content.serviceConfigProperties');
    var configData = this.reconfigureSites(['hbase-site'],data);
    App.ajax.send({
      name: 'common.service.configurations',
      sender: this,
      data: {
        desired_config: configData
      },
      success: 'saveConfigTag',
      error: 'onTaskError'
    });
  },

  saveConfigTag: function () {
    App.clusterStatus.setClusterStatus({
      clusterName: this.get('content.cluster.name'),
      clusterState: 'HIGH_AVAILABILITY_DEPLOY',
      wizardControllerName: this.get('content.controllerName'),
      localdb: App.db.data
    });
    this.onTaskCompleted();
  },

  startAllServices: function () {
    App.ajax.send({
      name: 'common.services.update',
      data: {
        context: "Start all services",
        "ServiceInfo": {
          "state": "STARTED"
        }
      },
      sender: this,
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  deleteSNameNode: function () {
    var hostName = this.get('content.masterComponentHosts').findProperty('component', 'SECONDARY_NAMENODE').hostName;
    App.ajax.send({
      name: 'common.delete.host_component',
      sender: this,
      data: {
        componentName: 'SECONDARY_NAMENODE',
        hostName: hostName
      },
      success: 'onTaskCompleted',
      error: 'onTaskError'
    });
  }

});


});

require.register("controllers/main/admin/highAvailability/nameNode/wizard_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.HighAvailabilityWizardController = App.WizardController.extend({

  name: 'highAvailabilityWizardController',

  totalSteps: 9,

  /**
   * Used for hiding back button in wizard
   */
  hideBackButton: true,

  content: Em.Object.create({
    controllerName: 'highAvailabilityWizardController',
    cluster: null,
    hosts: null,
    services: null,
    slaveComponentHosts: null,
    masterComponentHosts: null,
    serviceConfigProperties: [],
    serviceName: 'MISC',
    hdfsUser:"hdfs",
    nameServiceId: '',
    failedTask : null,
    requestIds: null
  }),

  setCurrentStep: function (currentStep, completed) {
    this._super(currentStep, completed);
    App.clusterStatus.setClusterStatus({
      clusterName: this.get('content.cluster.name'),
      clusterState: 'HIGH_AVAILABILITY_DEPLOY',
      wizardControllerName: 'highAvailabilityWizardController',
      localdb: App.db.data
    });
  },

  /**
   * return new object extended from clusterStatusTemplate
   * @return Object
   */
  getCluster: function(){
    return jQuery.extend({}, this.get('clusterStatusTemplate'), {name: App.router.getClusterName()});
  },

  /**
   * save status of the cluster.
   * @param clusterStatus object with status,requestId fields.
   */
  saveClusterStatus: function (clusterStatus) {
    var oldStatus = this.toObject(this.get('content.cluster'));
    clusterStatus = jQuery.extend(oldStatus, clusterStatus);
    if (clusterStatus.requestId) {
      clusterStatus.requestId.forEach(function (requestId) {
        if (clusterStatus.oldRequestsId.indexOf(requestId) === -1) {
          clusterStatus.oldRequestsId.push(requestId)
        }
      }, this);
    }
    this.set('content.cluster', clusterStatus);
    this.save('cluster');
  },

  /**
   * Save Master Component Hosts data to Main Controller
   * @param stepController App.WizardStep5Controller
   */
  saveMasterComponentHosts: function (stepController) {
    var obj = stepController.get('selectedServicesMasters');
    var masterComponentHosts = [];
    obj.forEach(function (_component) {
      masterComponentHosts.push({
        display_name: _component.get('display_name'),
        component: _component.get('component_name'),
        hostName: _component.get('selectedHost'),
        serviceId: _component.get('serviceId'),
        isCurNameNode: _component.get('isCurNameNode'),
        isAddNameNode: _component.get('isAddNameNode'),
        isInstalled: true
      });
    });
    this.setDBProperty('masterComponentHosts', masterComponentHosts);
    this.set('content.masterComponentHosts', masterComponentHosts);
  },

  saveHdfsUser: function () {
    App.db.setHighAvailabilityWizardHdfsUser(this.get('content.hdfsUser'));
  },

  saveTasksStatuses: function(statuses){
    App.db.setHighAvailabilityWizardTasksStatuses(statuses);
    this.set('content.tasksStatuses', statuses);
  },

  saveConfigTag: function(tag){
    App.db.setHighAvailabilityWizardConfigTag(tag);
    this.set('content.'+[tag.name], tag.value);
  },

  saveHdfsClientHosts: function(hostNames){
    App.db.setHighAvailabilityWizardHdfsClientHosts(hostNames);
    this.set('content.hdfsClientHostNames', hostNames);
  },

    /**
     * Save config properties
     * @param stepController HighAvailabilityWizardStep3Controller
     */
  saveServiceConfigProperties: function(stepController) {
    var serviceConfigProperties = [];
    var data = stepController.get('serverConfigData');

    var _content = stepController.get('stepConfigs')[0];
    _content.get('configs').forEach(function (_configProperties) {
      var siteObj = data.items.findProperty('type', _configProperties.get('filename'));
      if (siteObj) {
        siteObj.properties[_configProperties.get('name')] = _configProperties.get('value');
      }
    }, this);
    this.setDBProperty('serviceConfigProperties', data);
    this.set('content.serviceConfigProperties', data);
  },

  /**
   * load hosts from server
   */
  loadHosts: function () {
    var dfd;
    var hostsFromDb = this.getDBProperty('hosts');
    if (hostsFromDb) {
      this.set('content.hosts', hostsFromDb);
      dfd = $.Deferred();
      dfd.resolve();
    } else {
      dfd = App.ajax.send({
        name: 'hosts.high_availability.wizard',
        data: {},
        sender: this,
        success: 'loadHostsSuccessCallback',
        error: 'loadHostsErrorCallback'
      });
    }
    return dfd.promise();
  },

  /**
   * success callback of <code>loadHosts</code>
   * @param data
   * @param opt
   * @param params
   */
  loadHostsSuccessCallback: function (data, opt, params) {
    var hosts = {};

    data.items.forEach(function (item) {
      hosts[item.Hosts.host_name] = {
        name: item.Hosts.host_name,
        cpu: item.Hosts.cpu_count,
        memory: item.Hosts.total_mem,
        disk_info: item.Hosts.disk_info,
        bootStatus: "REGISTERED",
        isInstalled: true
      };
    });
    this.setDBProperty('hosts', hosts);
    this.set('content.hosts', hosts);
  },

  loadHdfsClientHosts: function(){
    var hostNames = App.db.getHighAvailabilityWizardHdfsClientHosts();
    if (!(hostNames instanceof Array)) {
      hostNames = [hostNames];
    }
    this.set('content.hdfsClientHostNames', hostNames);
  },

  loadConfigTag: function(tag){
    var tagVal = App.db.getHighAvailabilityWizardConfigTag(tag);
    this.set('content.'+tag, tagVal);
  },


  loadHdfsUser: function(){
    var hdfsUser = App.db.getHighAvailabilityWizardHdfsUser();
    this.set('content.hdfsUser', hdfsUser);
  },

  loadTasksStatuses: function(){
    var statuses = App.db.getHighAvailabilityWizardTasksStatuses();
    this.set('content.tasksStatuses', statuses);
  },

  /**
   * Load serviceConfigProperties to model
   */
  loadServiceConfigProperties: function () {
    var serviceConfigProperties = this.getDBProperty('serviceConfigProperties');
    this.set('content.serviceConfigProperties', serviceConfigProperties);
  },

  saveRequestIds: function(requestIds){
    App.db.setHighAvailabilityWizardRequestIds(requestIds);
    this.set('content.requestIds', requestIds);
  },

  loadRequestIds: function(){
    var requestIds = App.db.getHighAvailabilityWizardRequestIds();
    this.set('content.requestIds', requestIds);
  },

  saveNameServiceId: function(nameServiceId){
    App.db.setHighAvailabilityWizardNameServiceId(nameServiceId);
    this.set('content.nameServiceId', nameServiceId);
  },

  loadNameServiceId: function(){
    var nameServiceId = App.db.getHighAvailabilityWizardNameServiceId();
    this.set('content.nameServiceId', nameServiceId);
  },

  saveTasksRequestIds: function (requestIds) {
    App.db.setHighAvailabilityWizardTasksRequestIds(requestIds);
    this.set('content.tasksRequestIds', requestIds);
  },

  loadTasksRequestIds: function () {
    var requestIds = App.db.getHighAvailabilityWizardTasksRequestIds();
    this.set('content.tasksRequestIds', requestIds);
  },

  loadMap: {
    '1': [
      {
        type: 'sync',
        callback: function () {
          this.load('cluster');
        }
      }
    ],
    '2': [
      {
        type: 'async',
        callback: function () {
          var dfd = $.Deferred();
          var self = this;
          this.loadHosts().done(function () {
            self.loadServicesFromServer();
            self.loadMasterComponentHosts();
            self.loadHdfsUser();
            dfd.resolve();
          });
          return dfd.promise();
        }
      }
    ],
    '3': [
      {
        type: 'sync',
        callback: function () {
          this.loadNameServiceId();
          this.loadServiceConfigProperties();
        }
      }
    ],
    '5': [
      {
        type: 'sync',
        callback: function () {
          this.loadTasksStatuses();
          this.loadTasksRequestIds();
          this.loadRequestIds();
        }
      }
    ]
  },

  /**
   * Remove all loaded data.
   * Created as copy for App.router.clearAllSteps
   */
  clearAllSteps: function () {
    this.clearInstallOptions();
    // clear temporary information stored during the install
    this.set('content.cluster', this.getCluster());
  },

  clearTasksData: function () {
    this.saveTasksStatuses(undefined);
    this.saveRequestIds(undefined);
    this.saveTasksRequestIds(undefined);
  },

  /**
   * Clear all temporary data
   */
  finish: function () {
    App.db.data.Installer = {};
    this.resetDbNamespace();
    App.router.get('updateController').updateAll();
  }
});

});

require.register("controllers/main/admin/highAvailability/progress_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var installedComponents = [];

App.HighAvailabilityProgressPageController = App.HighAvailabilityWizardController.extend(App.wizardProgressPageControllerMixin, {

  name: 'highAvailabilityProgressPageController',
  clusterDeployState: 'HIGH_AVAILABILITY_DEPLOY',
  tasksMessagesPrefix: 'admin.highAvailability.wizard.step',
  isRollback: false,

  manualRollback: function () {
    App.ModalPopup.show({
      header: Em.I18n.t('admin.highAvailability.confirmRollbackHeader'),
      primary: Em.I18n.t('yes'),
      showCloseButton: false,
      onPrimary: function () {
        var self = this;
        var controller = App.router.get('highAvailabilityWizardController');
        controller.clearTasksData();
        controller.clearStorageData();
        controller.finish();
        App.router.get('updateController').set('isWorking', true);
        App.clusterStatus.setClusterStatus({
          clusterName: App.router.get('content.cluster.name'),
          clusterState: 'DEFAULT',
          localdb: App.db.data
        },{alwaysCallback: function() {self.hide();App.router.transitionTo('main.index');location.reload();}});
      },
      secondary: Em.I18n.t('no'),
      onSecondary: function () {
        this.hide();
      },
      bodyClass: Ember.View.extend({
        template: Ember.Handlebars.compile(Em.I18n.t('admin.highAvailability.confirmManualRollbackBody'))
      })
    });
  },

  rollback: function () {
    var task = this.get('tasks').findProperty('status', 'FAILED');
    App.router.get(this.get('content.controllerName')).saveFailedTask(task);
    App.ModalPopup.show({
      header: Em.I18n.t('admin.highAvailability.confirmRollbackHeader'),
      primary: Em.I18n.t('common.confirm'),
      showCloseButton: false,
      onPrimary: function () {
        App.router.get('highAvailabilityWizardController').clearTasksData();
        App.router.transitionTo('main.admin.highAvailabilityRollback');
        this.hide();
      },
      secondary: Em.I18n.t('common.cancel'),
      body: Em.I18n.t('admin.highAvailability.confirmRollbackBody')
    });
  },

  /**
   *
   * @param siteNames Array
   */
  reconfigureSites: function(siteNames, data) {
    var tagName = App.get('testMode') ? 'version1' : 'version' + (new Date).getTime();
    var componentName;
    switch (this.get('content.controllerName')) {
      case 'rMHighAvailabilityWizardController':
        componentName =  'RESOURCEMANAGER';
        break;
      default:
        componentName =  'NAMENODE';
    }
    return siteNames.map(function(_siteName) {
      var config = data.items.findProperty('type', _siteName);
      var configToSave = {
        type: _siteName,
        tag: tagName,
        properties: config && config.properties,
        service_config_version_note: Em.I18n.t('admin.highAvailability.step4.save.configuration.note').format(App.format.role(componentName))
      }
      if (config && config.properties_attributes) {
        configToSave.properties_attributes = config.properties_attributes;
      }
      return configToSave;
    });
  }
});

});

require.register("controllers/main/admin/highAvailability/progress_popup_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

App.HighAvailabilityProgressPopupController = Ember.Controller.extend({

  name: 'highAvailabilityProgressPopupController',

  /**
   * Id of current request
   * @type {Array}
   */
  requestIds: [],

  /**
   * Title for popup header
   * @type {String}
   */
  popupTitle: '',

  /**
   * Array with Hosts tasks data used in <code>App.HostPopup</code>
   * @type {Array}
   */
  services: [],

  /**
   * Timestamp used in <code>App.HostPopup</code>
   * @type {Number}
   */
  serviceTimestamp: null,

  /**
   * Progress controller. Used to get tasks data.
   * @type {App.HighAvailabilityProgressPageController}
   */
  progressController: null,

  /**
   * Requests data with tasks
   * @type {Array}
   */
  hostsData: [],

  /**
   *  StageId for the command.
   *  @type {Number}
   */
  stageId: null,

  /**
   * During loading and calculations show popup with spinner
   * @type {Object}
   */
  spinnerPopup: null,

  isTaskPolling: false,

  taskInfo: null,

  /**
   * Get info for <code>requestIds</code> and initialize <code>App.HostPopup</code>
   * @param popupTitle {String}
   * @param requestIds {Array}
   * @param progressController {App.HighAvailabilityProgressPageController}
   * @param showSpinner {Boolean}
   * @param stageId {Number}
   */
  initPopup: function (popupTitle, requestIds, progressController, showSpinner, stageId) {
    if(showSpinner){
      var loadingPopup = App.ModalPopup.show({
        header: Em.I18n.t('jobs.loadingTasks'),
        primary: false,
        secondary: false,
        bodyClass: Ember.View.extend({
          template: Ember.Handlebars.compile('<div class="spinner"></div>')
        })
      });
      this.set('spinnerPopup', loadingPopup);
    }
    this.set('progressController', progressController);
    this.set('popupTitle', popupTitle);
    this.set('requestIds', requestIds);
    this.set('hostsData', []);
    this.set('stageId', stageId);
    this.getHosts();
  },

  /**
   * Send AJAX request to get hosts tasks data
   */
  getHosts: function () {
    var requestIds = this.get('requestIds');
    var stageId = this.get('stageId');
    var name = 'background_operations.get_by_request';
    if (!Em.isNone(stageId)) {
      name = 'common.request.polling';
      if (stageId === 0) {
        stageId = '0';
      }
    }
    requestIds.forEach(function (requestId) {
      App.ajax.send({
        name: name,
        sender: this,
        data: {
          requestId: requestId,
          stageId: stageId
        },
        success: 'onGetHostsSuccess'
      })
    }, this);
  },

  /**
   * Callback for <code>getHosts</code> request
   * @param data
   */
  onGetHostsSuccess: function (data) {
    var hostsData = this.get('hostsData');
    hostsData.push(data);
    if (this.get('requestIds.length') === this.get('hostsData.length')) {
      var popupTitle = this.get('popupTitle');
      this.calculateHostsData(hostsData);
      App.HostPopup.initPopup(popupTitle, this);
      if (this.isRequestRunning(hostsData)) {
        this.addObserver('progressController.logs.length', this, 'getDataFromProgressController');
      }
    }
    if(this.get('spinnerPopup')){
      this.get('spinnerPopup').hide();
      this.set('spinnerPopup', null);
    }
  },

  /**
   * Convert data to format used in <code>App.HostPopup</code>
   * @param data {Array}
   */
  calculateHostsData: function (data) {
    var hosts = [];
    var hostsMap = {};
    var popupTitle = this.get('popupTitle');
    data.forEach(function (request) {
      request.tasks.forEach(function (task) {
        var host = task.Tasks.host_name;
        if (hostsMap[host]) {
          hostsMap[host].logTasks.push(task);
        } else {
          hostsMap[host] = {
            name: task.Tasks.host_name,
            publicName: task.Tasks.host_name,
            logTasks: [task]
          };
        }
      });
    });
    for (var host in hostsMap) {
      hosts.push(hostsMap[host]);
    }
    this.set('services', [
      {name: popupTitle, hosts: hosts}
    ]);
    this.set('serviceTimestamp', App.dateTime());
    if (!this.isRequestRunning(data)) {
      this.removeObserver('progressController.logs.length', this, 'getDataFromProgressController');
    }
  },

  /**
   * Get hosts tasks data from <code>progressController</code>
   */
  getDataFromProgressController: function () {
    var data = this.get('hostsData');
    var tasksData = [];
    var stageId = this.get('stageId');
    // If the progress page is broken into stages then update host with only stage's tasks
    if (!!stageId) {
      tasksData = this.get('progressController.logs').filterProperty('Tasks.stage_id',stageId);
    } else {
      tasksData = this.get('progressController.logs');
    }
    if (tasksData.length) {
      var tasks = [];
      tasksData.forEach(function (logs) {
        tasks.pushObjects(logs);
      }, this);
      data.forEach(function (request) {
        tasks = tasks.filterProperty('Tasks.request_id', request.Requests.id);
        request.tasks = tasks;
      });
      this.calculateHostsData(data);
    }
  },

  /**
   * Identify whether request is running by task counters
   * @param requests {Array}
   * @return {Boolean}
   */
  isRequestRunning: function (requests) {
    var result = false;
    requests.forEach(function (request) {
      if ((request.Requests.task_count -
          (request.Requests.aborted_task_count + request.Requests.completed_task_count + request.Requests.failed_task_count
              + request.Requests.timed_out_task_count - request.Requests.queued_task_count)) > 0) {
        result = true;
      }
    });
    return result;
  },

  startTaskPolling: function (requestId, taskId) {
    this.setProperties({
      isTaskPolling: true,
      taskInfo: {
        id: taskId,
        requestId: requestId
      }
    });
    App.updater.run(this, 'updateTask', 'isTaskPolling', App.bgOperationsUpdateInterval);
    App.updater.immediateRun('updateTask');
  },

  stopTaskPolling: function () {
    this.set('isTaskPolling', false);
  },

  updateTask: function (callback) {
    App.ajax.send({
      name: 'background_operations.get_by_task',
      sender: this,
      data: {
        requestId: this.get('taskInfo.requestId'),
        taskId: this.get('taskInfo.id')
      },
      success: 'updateTaskSuccessCallback',
      callback: callback
    })
  },

  updateTaskSuccessCallback: function (data) {
    this.setProperties({
      'taskInfo.stderr': data.Tasks.stderr,
      'taskInfo.stdout': data.Tasks.stdout,
      'taskInfo.outputLog': data.Tasks.output_log,
      'taskInfo.errorLog': data.Tasks.error_log,
      'isTaskPolling': !['FAILED', 'COMPLETED', 'TIMEDOUT', 'ABORTED'].contains(data.Tasks.status)
    });
  }

});


});

require.register("controllers/main/admin/highAvailability/resourceManager/step1_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.RMHighAvailabilityWizardStep1Controller = Em.Controller.extend({
  name: "rMHighAvailabilityWizardStep1Controller"
});


});

require.register("controllers/main/admin/highAvailability/resourceManager/step2_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.RMHighAvailabilityWizardStep2Controller = App.WizardStep5Controller.extend({
  name: "rMHighAvailabilityWizardStep2Controller",

  loadStepCallback: function (components, self) {
    this._super(components, self);
    self.hideUnusedComponents();
  },

  renderComponents: function (masterComponents) {
    var existedRM = masterComponents.findProperty('component_name', 'RESOURCEMANAGER');
    existedRM.isAdditional = false;
    var additionalRMSelectedHost = this.get('content.rmHosts.additionalRM') ||
        this.get('hosts').mapProperty('host_name').without(existedRM.selectedHost)[0];
    var additionalRM = $.extend({}, existedRM, {
      isInstalled: false,
      isAdditional: true,
      selectedHost: additionalRMSelectedHost
    });
    masterComponents.push(additionalRM);
    this._super(masterComponents);
  },

  /**
   * Remove service masters, that should be hidden in this wizard
   */
  hideUnusedComponents: function () {
    var servicesMasters = this.get('servicesMasters');
    servicesMasters = servicesMasters.filterProperty('component_name', 'RESOURCEMANAGER');
    this.set('servicesMasters', servicesMasters);
  }
});


});

require.register("controllers/main/admin/highAvailability/resourceManager/step3_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.RMHighAvailabilityWizardStep3Controller = Em.Controller.extend({
  name: "rMHighAvailabilityWizardStep3Controller",

  selectedService: null,

  versionLoaded: true,

  isLoaded: false,

  isSubmitDisabled: function () {
    return !this.get('isLoaded');
  }.property('isLoaded'),

  loadStep: function () {
    this.renderConfigs();
  },

  /**
   * Render configs to show them in <code>App.ServiceConfigView</code>
   */
  renderConfigs: function () {

    var configs = $.extend(true, {}, require('data/HDP2/rm_ha_properties').haConfig);

    var serviceConfig = App.ServiceConfig.create({
      serviceName: configs.serviceName,
      displayName: configs.displayName,
      configCategories: [],
      showConfig: true,
      configs: []
    });

    configs.configCategories.forEach(function (configCategory) {
      if (App.Service.find().someProperty('serviceName', configCategory.name)) {
        serviceConfig.configCategories.pushObject(configCategory);
      }
    }, this);

    this.renderConfigProperties(configs, serviceConfig);
    App.ajax.send({
      name: 'config.tags',
      sender: this,
      success: 'loadConfigTagsSuccessCallback',
      error: 'loadConfigsErrorCallback',
      data: {
        serviceConfig: serviceConfig
      }
    });

  },

  loadConfigTagsSuccessCallback: function (data, opt, params) {
    var urlParams = '(type=zoo.cfg&tag=' + data.Clusters.desired_configs['zoo.cfg'].tag + ')|' +
      '(type=yarn-site&tag=' + data.Clusters.desired_configs['yarn-site'].tag + ')';
    App.ajax.send({
      name: 'reassign.load_configs',
      sender: this,
      data: {
        urlParams: urlParams,
        serviceConfig: params.serviceConfig
      },
      success: 'loadConfigsSuccessCallback',
      error: 'loadConfigsSuccessCallback'
    });
  },

  loadConfigsSuccessCallback: function (data, opt, params) {
    var
      zooCfg = data && data.items ? data.items.findProperty('type', 'zoo.cfg') : null,
      yarnSite = data && data.items ? data.items.findProperty('type', 'yarn-site') : null,
      portValue = zooCfg && Em.get(zooCfg, 'properties.clientPort'),
      zkPort = portValue ? portValue : '2181',
      webAddressPort = yarnSite && yarnSite.properties ? yarnSite.properties['yarn.resourcemanager.webapp.address'] : null,
      httpsWebAddressPort = yarnSite && yarnSite.properties ? yarnSite. properties['yarn.resourcemanager.webapp.https.address'] : null;

    webAddressPort = webAddressPort && webAddressPort.match(/:[0-9]*/g) ? webAddressPort.match(/:[0-9]*/g)[0] : ":8088";
    httpsWebAddressPort = httpsWebAddressPort && httpsWebAddressPort.match(/:[0-9]*/g) ? httpsWebAddressPort.match(/:[0-9]*/g)[0] : ":8090";

    params = params.serviceConfig ? params.serviceConfig : arguments[4].serviceConfig;

    this.setDynamicConfigValues(params, zkPort, webAddressPort, httpsWebAddressPort);
    this.setProperties({
      selectedService: params,
      isLoaded: true
    });
  },

  /**
   * Set values dependent on host selection
   * @param configs
   * @param zkPort
   * @param webAddressPort
   * @param httpsWebAddressPort
   */
  setDynamicConfigValues: function (configs, zkPort, webAddressPort, httpsWebAddressPort) {
    var
      configProperties = configs.configs,
      currentRMHost = this.get('content.rmHosts.currentRM'),
      additionalRMHost = this.get('content.rmHosts.additionalRM'),
      zooKeeperHostsWithPort = App.HostComponent.find().filterProperty('componentName', 'ZOOKEEPER_SERVER').map(function (item) {
        return item.get('hostName') + ':' + zkPort;
      }).join(',');

    configProperties.findProperty('name', 'yarn.resourcemanager.hostname.rm1').set('value', currentRMHost).set('defaultValue', currentRMHost);
    configProperties.findProperty('name', 'yarn.resourcemanager.hostname.rm2').set('value', additionalRMHost).set('defaultValue', additionalRMHost);
    configProperties.findProperty('name', 'yarn.resourcemanager.zk-address').set('value', zooKeeperHostsWithPort).set('defaultValue', zooKeeperHostsWithPort);

    configProperties.findProperty('name', 'yarn.resourcemanager.webapp.address.rm1')
      .set('value', currentRMHost + webAddressPort)
      .set('defaultValue', currentRMHost + webAddressPort);

    configProperties.findProperty('name', 'yarn.resourcemanager.webapp.address.rm2')
      .set('value', additionalRMHost + webAddressPort)
      .set('defaultValue', additionalRMHost + webAddressPort);

    configProperties.findProperty('name', 'yarn.resourcemanager.webapp.https.address.rm1')
      .set('value', currentRMHost + httpsWebAddressPort)
      .set('defaultValue', currentRMHost + httpsWebAddressPort);

    configProperties.findProperty('name', 'yarn.resourcemanager.webapp.https.address.rm2')
      .set('value', additionalRMHost + httpsWebAddressPort)
      .set('defaultValue', additionalRMHost + httpsWebAddressPort);
  },

  /**
   * Load child components to service config object
   * @param _componentConfig
   * @param componentConfig
   */
  renderConfigProperties: function (_componentConfig, componentConfig) {
    _componentConfig.configs.forEach(function (_serviceConfigProperty) {
      var serviceConfigProperty = App.ServiceConfigProperty.create(_serviceConfigProperty);
      componentConfig.configs.pushObject(serviceConfigProperty);
      serviceConfigProperty.set('isEditable', serviceConfigProperty.get('isReconfigurable'));
      serviceConfigProperty.validate();
    }, this);
  },

  submit: function () {
    if (!this.get('isSubmitDisabled')) {
      App.get('router.mainAdminKerberosController').getKDCSessionState(function() {
        App.router.send("next");
      });
    }
  }
});


});

require.register("controllers/main/admin/highAvailability/resourceManager/step4_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

require('controllers/main/admin/serviceAccounts_controller');

App.RMHighAvailabilityWizardStep4Controller = App.HighAvailabilityProgressPageController.extend({

  name: "rMHighAvailabilityWizardStep4Controller",

  isRMHA: true,

  clusterDeployState: 'RM_HIGH_AVAILABILITY_DEPLOY',

  commands: ['stopRequiredServices', 'installResourceManager', 'reconfigureYARN', 'startAllServices'],

  tasksMessagesPrefix: 'admin.rm_highAvailability.wizard.step',

  stopRequiredServices: function () {
    var list = App.Service.find().mapProperty("serviceName").without("HDFS").join(',');
    App.ajax.send({
      name: 'common.services.update',
      sender: this,
      data: {
        "context": "Stop without HDFS",
        "ServiceInfo": {
          "state": "INSTALLED"
        },
        urlParams: "ServiceInfo/service_name.in(" + list + ")"},
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  installResourceManager: function () {
    var hostName = this.get('content.rmHosts.additionalRM');
    this.createComponent('RESOURCEMANAGER', hostName, "YARN");
  },

  reconfigureYARN: function () {
    this.loadConfigsTags();
  },

  loadConfigsTags: function () {
    App.ajax.send({
      name: 'config.tags',
      sender: this,
      success: 'onLoadConfigsTags',
      error: 'onTaskError'
    });
  },

  onLoadConfigsTags: function (data) {
    App.ajax.send({
      name: 'reassign.load_configs',
      sender: this,
      data: {
        urlParams: '(type=yarn-site&tag=' + data.Clusters.desired_configs['yarn-site'].tag + ')'
      },
      success: 'onLoadConfigs',
      error: 'onTaskError'
    });
  },

  onLoadConfigs: function (data) {
    var propertiesToAdd = this.get('content.configs');
    propertiesToAdd.forEach(function (property) {
      data.items[0].properties[property.name] = property.value;
    });

    var configData = this.reconfigureSites(['yarn-site'],data);

    App.ajax.send({
      name: 'common.service.configurations',
      sender: this,
      data: {
        desired_config: configData
      },
      success: 'onSaveConfigs',
      error: 'onTaskError'
    });
  },

  onSaveConfigs: function () {
    this.onTaskCompleted();
  },

  startAllServices: function () {
    App.ajax.send({
      name: 'common.services.update',
      sender: this,
      data: {
        "context": "Start all services",
        "ServiceInfo": {
          "state": "STARTED"
        },
        urlParams: "params/run_smoke_test=true"
      },
      success: 'startPolling',
      error: 'onTaskError'
    });
  }
});


});

require.register("controllers/main/admin/highAvailability/resourceManager/wizard_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.RMHighAvailabilityWizardController = App.WizardController.extend({

  name: 'rMHighAvailabilityWizardController',

  totalSteps: 4,

  isFinished: false,

  content: Em.Object.create({
    controllerName: 'rMHighAvailabilityWizardController'
  }),

  init: function () {
    this._super();
    this.clearStep();
  },

  clearStep: function () {
    this.set('isFinished', false);
  },

  setCurrentStep: function (currentStep, completed) {
    this._super(currentStep, completed);
    App.clusterStatus.setClusterStatus({
      clusterName: this.get('content.cluster.name'),
      wizardControllerName: 'rMHighAvailabilityWizardController',
      localdb: App.db.data
    });
  },

  /**
   * Save hosts for additional and current ResourceManagers to local db and <code>controller.content</code>
   * @param rmHosts
   */
  saveRmHosts: function (rmHosts) {
    this.set('content.rmHosts', rmHosts);
    this.setDBProperty('rmHosts', rmHosts);
  },

  /**
   * Load hosts for additional and current ResourceManagers from local db to <code>controller.content</code>
   */
  loadRmHosts: function() {
    var rmHosts = this.getDBProperty('rmHosts');
    this.set('content.rmHosts', rmHosts);
  },

  /**
   * Save configs to load and apply them on Configure Components step
   * @param configs
   */
  saveConfigs: function (configs) {
    this.set('content.configs', configs);
    this.setDBProperty('configs', configs);
  },

  /**
   * Load configs to apply them on Configure Components step
   */
  loadConfigs: function() {
    var configs = this.getDBProperty('configs');
    this.set('content.configs', configs);
  },

  saveTasksStatuses: function (tasksStatuses) {
    this.set('content.tasksStatuses', tasksStatuses);
    this.setDBProperty('tasksStatuses', tasksStatuses);
  },

  loadTasksStatuses: function() {
    var tasksStatuses = this.getDBProperty('tasksStatuses');
    this.set('content.tasksStatuses', tasksStatuses);
  },

  saveTasksRequestIds: function (tasksRequestIds) {
    this.set('content.tasksRequestIds', tasksRequestIds);
    this.setDBProperty('tasksRequestIds', tasksRequestIds);
  },

  loadTasksRequestIds: function() {
    var tasksRequestIds = this.getDBProperty('tasksRequestIds');
    this.set('content.tasksRequestIds', tasksRequestIds);
  },

  saveRequestIds: function (requestIds) {
    this.set('content.requestIds', requestIds);
    this.setDBProperty('requestIds', requestIds);
  },

  loadRequestIds: function() {
    var requestIds = this.getDBProperty('requestIds');
    this.set('content.requestIds', requestIds);
  },

  /**
   * Load data for all steps until <code>current step</code>
   */
  loadAllPriorSteps: function () {
    var step = this.get('currentStep');
    switch (step) {
      case '4':
        this.loadTasksStatuses();
        this.loadTasksRequestIds();
        this.loadRequestIds();
        this.loadConfigs();
      case '3':
      case '2':
        this.loadRmHosts();
        this.loadServicesFromServer();
        this.loadMasterComponentHosts();
        this.loadConfirmedHosts();
      case '1':
        this.load('cluster');
    }
  },

  /**
   * Remove all loaded data.
   * Created as copy for App.router.clearAllSteps
   */
  clearAllSteps: function () {
    this.clearInstallOptions();
    // clear temporary information stored during the install
    this.set('content.cluster', this.getCluster());
  },

  /**
   * Clear all temporary data
   */
  finish: function () {
    this.resetDbNamespace();
    App.router.get('updateController').updateAll();
    this.set('isFinished', true);
  }
});

});

require.register("controllers/main/admin/highAvailability_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminHighAvailabilityController = Em.Controller.extend({
  name: 'mainAdminHighAvailabilityController',

  tag: null,

  dataIsLoaded: false,

  /**
   * enable High Availability
   * @return {Boolean}
   */
  enableHighAvailability: function () {
    var message = [];
    var hostComponents = App.HostComponent.find();

    if (hostComponents.findProperty('componentName', 'NAMENODE').get('workStatus') !== 'STARTED') {
      message.push(Em.I18n.t('admin.highAvailability.error.namenodeStarted'));
    }
    if (hostComponents.filterProperty('componentName', 'ZOOKEEPER_SERVER').length < 3) {
      message.push(Em.I18n.t('admin.highAvailability.error.zooKeeperNum'));
    }

    if (App.router.get('mainHostController.hostsCountMap.TOTAL') < 3) {
      message.push(Em.I18n.t('admin.highAvailability.error.hostsNum'));
    }
    if (message.length > 0) {
      this.showErrorPopup(message);
      return false;
    }
    App.router.transitionTo('main.services.enableHighAvailability');
    return true;
  },

  disableHighAvailability: function () {
    App.router.transitionTo('main.admin.rollbackHighAvailability');
  },

  /**
   * enable ResourceManager High Availability
   * @return {Boolean}
   */
  enableRMHighAvailability: function () {
    //Prerequisite Checks
    var message = [];
    if (App.HostComponent.find().filterProperty('componentName', 'ZOOKEEPER_SERVER').length < 3) {
      message.push(Em.I18n.t('admin.rm_highAvailability.error.zooKeeperNum'));
    }

    if (App.router.get('mainHostController.hostsCountMap.TOTAL') < 3) {
      message.push(Em.I18n.t('admin.rm_highAvailability.error.hostsNum'));
    }
    if (message.length > 0) {
      this.showErrorPopup(message);
      return false;
    }
    App.router.transitionTo('main.services.enableRMHighAvailability');
    return true;
  },

  /**
   * join or wrap message depending on whether it is array or string
   * @param message
   * @return {*}
   */
  joinMessage: function (message) {
    if (Array.isArray(message)) {
      return message.join('<br/>');
    } else {
      return '<p>' + message + '</p>';
    }
  },

  showErrorPopup: function (message) {
    message = this.joinMessage(message);
    App.ModalPopup.show({
      header: Em.I18n.t('common.error'),
      bodyClass: Ember.View.extend({
        template: Ember.Handlebars.compile(message)
      }),
      secondary: false
    });
  }
});

});

require.register("controllers/main/admin/kerberos", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');
require('controllers/main/admin/kerberos/step4_controller');

App.MainAdminKerberosController = App.KerberosWizardStep4Controller.extend({
  name: 'mainAdminKerberosController',
  securityEnabled: false,
  dataIsLoaded: false,
  isRecommendedLoaded: true,
  getAddSecurityWizardStatus: function () {
    return App.db.getSecurityWizardStatus();
  },
  setAddSecurityWizardStatus: function (status) {
    App.db.setSecurityWizardStatus(status);
  },

  setDisableSecurityStatus: function (status) {
    App.db.setDisableSecurityStatus(status);
  },
  getDisableSecurityStatus: function (status) {
    return App.db.getDisableSecurityStatus();
  },

  notifySecurityOff: false,
  notifySecurityAdd: false,

  notifySecurityOffPopup: function () {
    var self = this;
    App.ModalPopup.show({
      header: Em.I18n.t('popup.confirmation.commonHeader'),
      primary: Em.I18n.t('ok'),
      onPrimary: function () {
        App.db.setSecurityDeployCommands(undefined);
        self.setDisableSecurityStatus("RUNNING");
        App.router.transitionTo('disableSecurity');
        this.hide();
      },
      bodyClass: Ember.View.extend({
        templateName: require('templates/main/admin/security/notify_security_off_popup')
      })
    });
  },

  /**
   * Show confirmation popup and after confirmation send request to regenerate keytabs
   * @method regenerateKeytabs
   * @return {App.ModalPopup}
   */
  regenerateKeytabs: function () {
    var self = this;

    return App.ModalPopup.show({

      /**
       * True - regenerate keytabs only for missing hosts and components, false - regenerate for all hosts and components
       * @type {boolean}
       */
      regenerateKeytabsOnlyForMissing: false,

      header: Em.I18n.t('admin.kerberos.button.regenerateKeytabs'),

      bodyClass: Em.View.extend({
        templateName: require('templates/main/admin/kerberos/regenerate_keytabs_popup_body')
      }),

      onPrimary: function () {
        this._super();
        self.regenerateKeytabsRequest(this.get('regenerateKeytabsOnlyForMissing'));
      }
    });
  },

  /**
   * Send request to regenerate keytabs
   * @param {boolean} missingOnly determines type of regeneration - missing|all
   * @returns {$.ajax}
   */
  regenerateKeytabsRequest: function (missingOnly) {
    return App.ajax.send({
      name: "admin.kerberos_security.regenerate_keytabs",
      sender: this,
      data: {
        type: missingOnly ? 'missing' : 'all'
      },
      success: "regenerateKeytabsSuccess"
    });
  },

  /**
   * Success callback of <code>regenerateKeytabs</code>
   * show background operations popup if appropriate option is set
   */
  regenerateKeytabsSuccess: function () {
    App.router.get('applicationController').dataLoading().done(function (initValue) {
      if (initValue) {
        App.router.get('backgroundOperationsController').showPopup();
      }
    });
  },

  getUpdatedSecurityStatus: function () {
    this.getSecurityStatus();
    return this.get('securityEnabled');
  },

  /**
   * performs cluster check before kerbefos security
   * wizard starts if <code>preKerberizeCheck<code> supports is true
   * otherwise runs <code>startKerberosWizard<code>
   * @method checkAndStartKerberosWizard
   */
  checkAndStartKerberosWizard: function() {
    if (App.get('supports.preKerberizeCheck')) {
      App.ajax.send({
        name: "admin.kerberos_security.checks",
        sender: this,
        success: "runSecurityCheckSuccess"
      });
    } else {
      this.startKerberosWizard();
    }
  },

  /**
   * success callback of <code>checkAndStartKerberosWizard()</code>
   * if there are some fails - it shows popup else open security wizard
   * @param data {object}
   * @param opt {object}
   * @param params {object}
   * @returns {App.ModalPopup|undefined}
   */
  runSecurityCheckSuccess: function (data, opt, params) {
    //TODO correct check
    if (data.items.someProperty('UpgradeChecks.status', "FAIL")) {
      var header = Em.I18n.t('popup.clusterCheck.Security.header').format(params.label);
      var title = Em.I18n.t('popup.clusterCheck.Security.title');
      var alert = Em.I18n.t('popup.clusterCheck.Security.alert');
      App.showClusterCheckPopup(data, header, title, alert);
    } else {
      this.startKerberosWizard();
    }
  },

  startKerberosWizard: function () {
    this.setAddSecurityWizardStatus('RUNNING');
    App.router.get('kerberosWizardController').setDBProperty('onClosePath', 'main.admin.adminKerberos.index');
    App.router.transitionTo('adminKerberos.adminAddKerberos');
  },

  /**
   * Loads the security status from server (security_enabled property in cluster-env configuration)
   */
  loadSecurityStatusFromServer: function () {
    if (App.get('testMode')) {
      this.set('securityEnabled', !App.get('testEnableSecurity'));
      this.set('dataIsLoaded', true);
    } else {
      //get Security Status From Server
      return this.getSecurityStatus();
    }
  },

  /**
   * Load security status from server.
   * @returns {$.Deferred}
   */
  getSecurityStatus: function () {
    var self = this;
    var dfd = $.Deferred();
    if (App.get('testMode')) {
      this.set('securityEnabled', !App.get('testEnableSecurity'));
      this.set('dataIsLoaded', true);
      dfd.resolve();
    } else {
      //get Security Status From Server
      App.ajax.send({
        name: 'admin.security_status',
        sender: this,
        success: 'getSecurityStatusSuccessCallback',
        error: 'errorCallback'
      }).always(function() {
        // check for kerberos descriptor artifact
        if (self.get('securityEnabled')) {
          self.loadClusterDescriptorConfigs().then(function() {
            dfd.resolve();
          }, function() {
            // if kerberos descriptor doesn't exist in cluster artifacts we have to kerberize cluster.
            // Show `Enable kerberos` button and set unsecure status.
            self.set('securityEnabled', false);
            dfd.resolve();
          });
        } else {
          dfd.resolve();
        }
      });
    }
    return dfd.promise();
  },

  getSecurityStatusSuccessCallback: function(data) {
    this.set('dataIsLoaded', true);
    var securityType = data.Clusters.security_type;
    this.set('securityEnabled', securityType === 'KERBEROS');
  },

  errorCallback: function (jqXHR) {
    this.set('dataIsLoaded', true);
    // Show the error popup if the API call received a response from the server.
    // jqXHR.status will be empty when browser cancels the request. Refer to AMBARI-5921 for more info
    if (!!jqXHR.status) {
      this.showSecurityErrorPopup();
    }
  },

  showSecurityErrorPopup: function () {
    App.ModalPopup.show({
      header: Em.I18n.t('common.error'),
      secondary: false,
      bodyClass: Ember.View.extend({
        template: Ember.Handlebars.compile('<p>{{t admin.security.status.error}}</p>')
      })
    });
  },

  /**
   * Override <code>App.KerberosWizardStep4Controller</code>
   *
   * @param {App.ServiceConfigProperty[]} properties
   */
  setStepConfigs: function (properties) {
    this.get('stepConfigs').clear();
    this._super(properties);
  },

  /**
   * Override <code>App.KerberosWizardStep4Controller</code>
   *
   * @param {App.ServiceConfigProperty[]} configs
   * @returns {App.ServiceConfigProperty[]}
   */
  prepareConfigProperties: function(configs) {
    var configProperties = configs.slice(0);
    var siteProperties = App.config.get('preDefinedSiteProperties');
    var installedServiceNames = ['Cluster'].concat(App.Service.find().mapProperty('serviceName'));
    configProperties = configProperties.filter(function(item) {
      return installedServiceNames.contains(item.get('serviceName'));
    });
    configProperties.setEach('isSecureConfig', false);
    configProperties.forEach(function(property, item, allConfigs) {
      if (property.get('observesValueFrom')) {
        var observedValue = allConfigs.findProperty('name', property.get('observesValueFrom')).get('value');
        property.set('value', observedValue);
        property.set('defaultValue', observedValue);
      }
      if (property.get('serviceName') == 'Cluster') {
        property.set('category', 'Global');
      } else {
        property.set('category', property.get('serviceName'));
      }
      // All user identity should be grouped under "Ambari Principals" category
      if (property.get('identityType') == 'user') property.set('category', 'Ambari Principals');
      var siteProperty = siteProperties.findProperty('name', property.get('name'));
      if (siteProperty) {
        if (siteProperty.category === property.get('category')) {
          property.set('displayName',siteProperty.displayName);
          if (siteProperty.index) {
            property.set('index', siteProperty.index);
          }
        }
        if (siteProperty.displayType) {
          property.set('displayType', siteProperty.displayType);
        }
      }
    });
    configProperties.setEach('isEditable', false);
    return configProperties;
  },

  getKDCSessionState: function(callback) {
    if (this.get('securityEnabled')) {
      App.ajax.send({
        name: 'kerberos.session.state',
        sender: this,
        data: {
          callback: callback
        },
        success: 'checkState'
      })
    } else {
      callback();
    }
  },

  checkState: function(data, opt, params) {
    var res = Em.get(data, 'Services.attributes.kdc_validation_result');
    var message = Em.get(data, 'Services.attributes.kdc_validation_failure_details');
    if (res.toUpperCase() === "OK") {
      params.callback();
    } else {
      App.showInvalidKDCPopup(opt, App.format.kdcErrorMsg(message, false));
    }
  }
});

});

require.register("controllers/main/admin/kerberos/disable_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
require('controllers/main/admin/kerberos/progress_controller');
require('controllers/main/admin/security/security_progress_controller');

App.KerberosDisableController = App.KerberosProgressPageController.extend({

  name: 'kerberosDisableController',
  clusterDeployState: 'DEFAULT',
  commands: ['stopServices', 'unkerberize', 'deleteKerberos', 'startServices'],

  tasksMessagesPrefix: 'admin.kerberos.disable.step',

  loadStep: function() {
    this.set('content.controllerName', 'kerberosDisableController');
    this.loadTasksStatuses();
    this.loadTasksRequestIds();
    this.loadRequestIds();
    this._super();
  },

  stopServices: function () {
    App.ajax.send({
      name: 'common.services.update',
      sender: this,
      data: {
        context: "Stop services",
        "ServiceInfo": {
          "state": "INSTALLED"
        }
      },
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  unkerberize: function () {
    return App.ajax.send({
      name: 'admin.unkerberize.cluster',
      sender: this,
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  deleteKerberos: function () {
    return App.ajax.send({
      name: 'common.delete.service',
      sender: this,
      data: {
        serviceName: 'KERBEROS'
      },
      success: 'onTaskCompleted',
      error: 'onTaskCompleted'
    });
  },

  startServices: function () {
    App.ajax.send({
      name: 'common.services.update',
      sender: this,
      data: {
        "context": "Start services",
        "ServiceInfo": {
          "state": "STARTED"
        },
        urlParams: "params/run_smoke_test=true"
      },
      success: 'startPolling',
      error: 'onTaskError'
    });
  }

});

});

require.register("controllers/main/admin/kerberos/progress_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

App.KerberosProgressPageController = App.KerberosWizardController.extend(App.wizardProgressPageControllerMixin, {

  name: 'kerberosProgressPageController',
  clusterDeployState: 'KERBEROS__DEPLOY',
  tasksMessagesPrefix: 'admin.kerberos.wizard.step',
  isRollback: false
});

});

require.register("controllers/main/admin/kerberos/step1_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.KerberosWizardStep1Controller = Em.Controller.extend({
  name: "kerberosWizardStep1Controller",

  selectedItem: Em.I18n.t('admin.kerberos.wizard.step1.option.kdc'),

  isSubmitDisabled: function() {
    return this.get('selectedOption.preConditions').someProperty('checked',false);
  }.property('selectedOption', 'selectedOption.preConditions.@each.checked'),

  options: [
    Em.Object.create({
      displayName: Em.I18n.t('admin.kerberos.wizard.step1.option.kdc'),
      value: Em.I18n.t('admin.kerberos.wizard.step1.option.kdc'),
      preConditions: [
        Em.Object.create({
          displayText: Em.I18n.t('admin.kerberos.wizard.step1.option.kdc.condition.1'),
          checked: false
        }),
        Em.Object.create({
          displayText: Em.I18n.t('admin.kerberos.wizard.step1.option.kdc.condition.2'),
          checked: false
        }),
        Em.Object.create({
          displayText: Em.I18n.t('admin.kerberos.wizard.step1.option.kdc.condition.3'),
          checked: false
        })
      ]
    }),
    Em.Object.create({
      displayName: Em.I18n.t('admin.kerberos.wizard.step1.option.ad'),
      value: Em.I18n.t('admin.kerberos.wizard.step1.option.ad'),
      preConditions: [
        Em.Object.create({
          displayText: Em.I18n.t('admin.kerberos.wizard.step1.option.ad.condition.1'),
          checked: false
        }),
        Em.Object.create({
          displayText: Em.I18n.t('admin.kerberos.wizard.step1.option.ad.condition.2'),
          checked: false
        }),
        Em.Object.create({
          displayText: Em.I18n.t('admin.kerberos.wizard.step1.option.ad.condition.3'),
          checked: false
        }),
        Em.Object.create({
          displayText: Em.I18n.t('admin.kerberos.wizard.step1.option.ad.condition.4'),
          checked: false
        }),
        Em.Object.create({
          displayText: Em.I18n.t('admin.kerberos.wizard.step1.option.ad.condition.5'),
          checked: false
        })
      ]
    })
  ],

  /**
   * precondition for the selected KDC option
   */
  selectedOption: function () {
    var options = this.get('options');
    var selectedItem = this.get('selectedItem');
    return options.findProperty('value', selectedItem);
  }.property('selectedItem'),



  loadStep: function () {
    this.set('selectedItem', Em.I18n.t('admin.kerberos.wizard.step1.option.kdc'));
  },

  next: function () {
    if (!this.get('isSubmitDisabled')) {
      App.router.send('next');
    }
  }
});


});

require.register("controllers/main/admin/kerberos/step2_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
require('controllers/wizard/step7_controller');

App.KerberosWizardStep2Controller = App.WizardStep7Controller.extend({
  name: "kerberosWizardStep2Controller",
  isKerberosWizard: true,

  selectedServiceNames: ['KERBEROS'],

  allSelectedServiceNames: ['KERBEROS'],

  componentName: 'KERBEROS_CLIENT',

  installedServiceNames: [],

  servicesInstalled: false,

  addMiscTabToPage: false,

  /**
   * @type {boolean} true if test connection to hosts is in progress
   */
  testConnectionInProgress: false,

  /**
   * Should Back-button be disabled
   * @type {boolean}
   */
  isBackBtnDisabled: function() {
    return this.get('testConnectionInProgress');
  }.property('testConnectionInProgress'),

  /**
   * Should Next-button be disabled
   * @type {boolean}
   */
  isSubmitDisabled: function () {
    if (!this.get('stepConfigs.length') || this.get('testConnectionInProgress') || this.get('submitButtonClicked')) return true;
    return (!this.get('stepConfigs').filterProperty('showConfig', true).everyProperty('errorCount', 0) || this.get("miscModalVisible"));
  }.property('stepConfigs.@each.errorCount', 'miscModalVisible', 'submitButtonClicked', 'testConnectionInProgress'),

  hostNames: function () {
    return App.get('allHostNames');
  }.property('App.allHostNames'),

  serviceConfigTags: [],

  clearStep: function () {
    this._super();
    this.get('serviceConfigTags').clear();
    this.set('servicesInstalled', false);
  },


  /**
   * On load function
   * @method loadStep
   */
  loadStep: function () {
    console.log("TRACE: Loading step7: Configure Services");
    this.clearStep();
    var kerberosStackService = App.StackService.find().findProperty('serviceName', 'KERBEROS');
    if (!kerberosStackService) return;
    //STEP 1: Load advanced configs
    var advancedConfigs = this.get('content.advancedServiceConfig');
    //STEP 2: Load on-site configs by service from local DB
    var storedConfigs = this.get('content.serviceConfigProperties');
    //STEP 3: Merge pre-defined configs with loaded on-site configs
    var configs = App.config.mergePreDefinedWithStored(
      storedConfigs,
      advancedConfigs,
      this.get('selectedServiceNames'));
    App.config.setPreDefinedServiceConfigs(this.get('addMiscTabToPage'));
    //STEP 4: Add advanced configs
    App.config.addAdvancedConfigs(configs, advancedConfigs);
    this.showAdConfigs(configs);
    this.applyServicesConfigs(configs, storedConfigs);
  },

  /**
   * Make Active Directory specific configs visible if user has selected AD option
   * @param configs
   */
  showAdConfigs: function (configs) {
    var kdcType = this.get('content.kerberosOption');
    var configNames = ['ldap_url', 'container_dn'];
    configNames.forEach(function (_configName) {
      var config = configs.findProperty('name', _configName);
      config.isVisible = kdcType === Em.I18n.t('admin.kerberos.wizard.step1.option.ad');
    }, this);
  },

  submit: function () {
    if (this.get('isSubmitDisabled')) return false;
    this.set('isSubmitDisabled', true);
    var self = this;
    this.deleteKerberosService().always(function (data) {
      self.createKerberosResources();
    });
  },

  createKerberosResources: function () {
    var self = this;
    this.createKerberosService().done(function () {
      self.createKerberosComponent().done(function () {
        self.createKerberosHostComponents().done(function () {
          self.createConfigurations().done(function () {
            self.createKerberosAdminSession().done(function () {
              App.router.send('next');
            });
          });
        });
      });
    });
  },

  /**
   * Delete Kerberos service if it exists
   */
  deleteKerberosService: function () {
    var serviceName = this.selectedServiceNames[0];
    return App.ajax.send({
      name: 'common.delete.service',
      sender: this,
      data: {
        serviceName: serviceName
      }
    });
  },

  createKerberosService: function () {
    return App.ajax.send({
      name: 'wizard.step8.create_selected_services',
      sender: this,
      data: {
        data: '{"ServiceInfo": { "service_name": "KERBEROS"}}',
        cluster: App.get('clusterName') || App.clusterStatus.get('clusterName')
      }
    });
  },

  createKerberosComponent: function () {
    return App.ajax.send({
      name: 'common.create_component',
      sender: this,
      data: {
        serviceName: this.selectedServiceNames[0],
        componentName: this.componentName
      }
    });
  },

  createKerberosHostComponents: function () {
    var hostNames = this.get('hostNames');
    var queryStr = '';
    hostNames.forEach(function (hostName) {
      queryStr += 'Hosts/host_name=' + hostName + '|';
    });
    //slice off last symbol '|'
    queryStr = queryStr.slice(0, -1);

    var data = {
      "RequestInfo": {
        "query": queryStr
      },
      "Body": {
        "host_components": [
          {
            "HostRoles": {
              "component_name": this.componentName
            }
          }
        ]
      }
    };

    return App.ajax.send({
      name: 'wizard.step8.register_host_to_component',
      sender: this,
      data: {
        cluster: App.router.getClusterName(),
        data: JSON.stringify(data)
      }
    });
  },

  createConfigurations: function () {
    var service = App.StackService.find().findProperty('serviceName', 'KERBEROS');
    var serviceConfigTags = [];
    var tag = 'version' + (new Date).getTime();
    Object.keys(service.get('configTypes')).forEach(function (type) {
      if (!serviceConfigTags.someProperty('type', type)) {
        var obj = this.createKerberosSiteObj(type, tag);
        obj.service_config_version_note = Em.I18n.t('admin.kerberos.wizard.configuration.note');
        serviceConfigTags.pushObject(obj);
      }
    }, this);
    var allConfigData = [];
    var serviceConfigData = [];
    Object.keys(service.get('configTypesRendered')).forEach(function (type) {
      var serviceConfigTag = serviceConfigTags.findProperty('type', type);
      if (serviceConfigTag) {
        serviceConfigData.pushObject(serviceConfigTag);
      }
    }, this);
    if (serviceConfigData.length) {
      allConfigData.pushObject(JSON.stringify({
        Clusters: {
          desired_config: serviceConfigData
        }
      }));
    }
    return App.ajax.send({
      name: 'common.across.services.configurations',
      sender: this,
      data: {
        data: '[' + allConfigData.toString() + ']'
      }
    });
  },

  createKerberosSiteObj: function (site, tag) {
    var properties = {};
    var content = this.get('stepConfigs')[0].get('configs');
    var configs = content.filterProperty('filename', site + '.xml');
    configs.forEach(function (_configProperty) {
      // do not pass any globals whose name ends with _host or _hosts
      if (_configProperty.isRequiredByAgent !== false) {
        properties[_configProperty.name] = _configProperty.value;
      }
    }, this);
    this.tweakKdcTypeValue(properties);
    return {"type": site, "tag": tag, "properties": properties};
  },

  tweakKdcTypeValue: function (properties) {
    if (properties['kdc_type'] === Em.I18n.t('admin.kerberos.wizard.step1.option.kdc')) {
      properties['kdc_type'] = "mit-kdc";
    } else if (properties['kdc_type'] === Em.I18n.t('admin.kerberos.wizard.step1.option.ad')) {
      properties['kdc_type'] = "active-directory";
    }
  },

  /**
   * puts kerberos admin credentials in the live cluster session
   * @returns {*} jqXHr
   */
  createKerberosAdminSession: function (configs) {
    configs = configs || this.get('stepConfigs')[0].get('configs');
    var adminPrincipalValue = configs.findProperty('name', 'admin_principal').value;
    var adminPasswordValue = configs.findProperty('name', 'admin_password').value;
    return App.ajax.send({
      name: 'common.cluster.update',
      sender: this,
      data: {
        clusterName: App.get('clusterName') || App.clusterStatus.get('clusterName'),
        data: [{
          session_attributes: {
            kerberos_admin: {principal: adminPrincipalValue, password: adminPasswordValue}
          }
        }]
      }
    });
  },

  /**
   * shows popup with to warn user
   * @param primary
   */
  showConnectionInProgressPopup: function(primary) {
    var primaryText = Em.I18n.t('common.exitAnyway');
    var msg = Em.I18n.t('services.service.config.connection.exitPopup.msg');
    App.showConfirmationPopup(primary, msg, null, null, primaryText)
  }
});


});

require.register("controllers/main/admin/kerberos/step3_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

App.KerberosWizardStep3Controller = App.KerberosProgressPageController.extend({
  name: 'kerberosWizardStep3Controller',
  clusterDeployState: 'KERBEROS_DEPLOY',
  serviceName: 'KERBEROS',
  componentName: 'KERBEROS_CLIENT',

  commands: ['installKerberos', 'testKerberos'],

  loadStep: function () {
    this._super();
    this.enableDisablePreviousSteps();
  },

  installKerberos: function() {
    var self = this;
    this.getKerberosClientState().done(function(data) {
      if (data.ServiceComponentInfo.state === 'INIT') {
        App.ajax.send({
          name: 'common.services.update',
          sender: self,
          data: {
            context: Em.I18n.t('requestInfo.kerberosService'),
            ServiceInfo: {"state": "INSTALLED"},
            urlParams: "ServiceInfo/state=INSTALLED&ServiceInfo/service_name=KERBEROS"
          },
          success: 'startPolling',
          error: 'onTaskError'
        });
      } else {
        var hostNames = App.get('allHostNames');
        self.updateComponent('KERBEROS_CLIENT', hostNames, "KERBEROS", "Install");
      }
    });
  },

  onKDCCancel: function() {
    App.router.get(this.get('content.controllerName')).setStepsEnable();
    this.get('tasks').objectAt(this.get('currentTaskId')).set('status', 'FAILED');
  },

  getKerberosClientState: function() {
    return App.ajax.send({
      name: 'common.service_component.info',
      sender: this,
      data: {
        serviceName: this.serviceName,
        componentName: this.componentName,
        urlParams: "fields=ServiceComponentInfo/state"
      }
    });
  },

  testKerberos: function() {
    App.ajax.send({
      'name': 'service.item.smoke',
      'sender': this,
      'success': 'startPolling',
      'error': 'onTestKerberosError',
      'kdcCancelHandler': 'onKDCCancel',
      'data': {
        'serviceName': this.serviceName,
        'displayName': App.format.role(this.serviceName),
        'actionName': this.serviceName + '_SERVICE_CHECK',
        'operationLevel': {
          "level": "CLUSTER",
          "cluster_name": App.get('clusterName')
        }
      }
    });
  },

  onTestKerberosError: function (jqXHR, ajaxOptions, error, opt) {
    App.ajax.defaultErrorHandler(jqXHR, opt.url, opt.type, jqXHR.status);
    this.onTaskError(jqXHR, ajaxOptions, error, opt);
  },

  /**
   * Enable or disable previous steps according to tasks statuses
   */
  enableDisablePreviousSteps: function () {
    var wizardController = App.router.get(this.get('content.controllerName'));
    if (this.get('tasks').someProperty('status', 'FAILED')) {
      wizardController.setStepsEnable();
    } else {
      wizardController.setLowerStepsDisable(3);
    }
  }.observes('tasks.@each.status')
});


});

require.register("controllers/main/admin/kerberos/step4_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
require('controllers/wizard/step7_controller');

App.KerberosWizardStep4Controller = App.WizardStep7Controller.extend(App.AddSecurityConfigs, {
  name: 'kerberosWizardStep4Controller',

  adminPropertyNames: [{name: 'admin_principal', displayName: 'Admin principal'}, {name: 'admin_password', displayName: 'Admin password'}],

  clearStep: function() {
    this.set('isRecommendedLoaded', false);
    this.set('selectedService', null);
    this.set('stepConfigs', []);
  },

  loadStep: function() {
    if (this.get('wizardController.skipConfigureIdentitiesStep')) {
      App.router.send('next');
      return;
    }
    var self = this;
    this.clearStep();
    this.getDescriptorConfigs().then(function(properties) {
      self.setStepConfigs(properties);
      self.set('isRecommendedLoaded', true);
    });
  },

  /**
   * Create service config object for Kerberos service.
   *
   * @param {App.ServiceConfigProperty[]} configs
   * @returns {Em.Object}
   */
  createServiceConfig: function(configs) {
    // Identity configs related to user principal
    var clusterConfigs = configs.filterProperty('serviceName','Cluster');
    // storm user principal is not required for ambari operation
    var userConfigs = configs.filterProperty('identityType','user').rejectProperty('serviceName','STORM');
    var generalConfigs = clusterConfigs.concat(userConfigs).uniq('name');
    var advancedConfigs = configs.filter(function(element){
      return !generalConfigs.findProperty('name', element.get('name'));
    });
    var categoryForGeneralConfigs = [
      App.ServiceConfigCategory.create({ name: 'Global', displayName: 'Global'}),
      App.ServiceConfigCategory.create({ name: 'Ambari Principals', displayName: 'Ambari Principals'})
    ];
    var categoryForAdvancedConfigs = this.createCategoryForServices();
    return [
      App.ServiceConfig.create({
      displayName: 'General',
      name: 'GENERAL',
      serviceName: 'KERBEROS_GENERAL',
      configCategories: categoryForGeneralConfigs,
      configs: generalConfigs,
      showConfig: true
    }),
      App.ServiceConfig.create({
        displayName: 'Advanced',
        name: 'ADVANCED',
        serviceName: 'KERBEROS_ADVANCED',
        configCategories: categoryForAdvancedConfigs,
        configs: advancedConfigs,
        showConfig: true
      })
    ];
  },

  /**
   * creates categories for advanced secure configs
   * @returns {[App.ServiceConfigCategory]}
   */
  createCategoryForServices: function() {
    var services = [];
    if (this.get('wizardController.name') == 'addServiceController') {
      services = App.StackService.find().filter(function(item) {
        return item.get('isInstalled') || item.get('isSelected');
      });
    } else {
      services = App.Service.find();
    }
    return services.map(function(item) {
      return App.ServiceConfigCategory.create({ name: item.get('serviceName'), displayName: item.get('displayName'), collapsedByDefault: true});
    });
  },

  /**
   * Prepare step configs using stack descriptor properties.
   *
   * @param {App.ServiceConfigProperty[]} configs
   */
  setStepConfigs: function(configs) {
    var configProperties = this.prepareConfigProperties(configs);
    if (this.get('wizardController.name') == 'addServiceController') {
      // config properties for installed services should be disabled on Add Service Wizard
      configProperties.forEach(function(item) {
        if (this.get('adminPropertyNames').mapProperty('name').contains(item.get('name'))) return;
        if (this.get('installedServiceNames').contains(item.get('serviceName')) || item.get('serviceName') == 'Cluster') {
          item.set('isEditable', false);
        }
      }, this);
    }
    this.get('stepConfigs').pushObjects(this.createServiceConfig(configProperties));
    this.set('selectedService', this.get('stepConfigs')[0]);
  },

  /**
   * Filter configs by installed services for Kerberos Wizard or by installed + selected services
   * for Add Service Wizard.
   * Set property value observer.
   * Set realm property with value from previous configuration step.
   * Set appropriate category for all configs.
   *
   * @param {App.ServiceConfigProperty[]} configs
   * @returns {App.ServiceConfigProperty[]}
   */
  prepareConfigProperties: function(configs) {
    var self = this;
    var storedServiceConfigs = this.get('wizardController').getDBProperty('serviceConfigProperties');
    var installedServiceNames = ['Cluster'].concat(App.Service.find().mapProperty('serviceName'));
    var adminProps = [];
    var configProperties = configs.slice(0);
    var siteProperties = App.config.get('preDefinedSiteProperties');
    if (this.get('wizardController.name') == 'addServiceController') {
      installedServiceNames = installedServiceNames.concat(this.get('selectedServiceNames'));
      this.get('adminPropertyNames').forEach(function(item) {
        var property = storedServiceConfigs.filterProperty('filename', 'krb5-conf.xml').findProperty('name', item.name);
        if (!!property) {
          var _prop = App.ServiceConfigProperty.create($.extend({}, property, { name: item.name, value: '', defaultValue: '', serviceName: 'Cluster', displayName: item.displayName}));
          _prop.validate();
          adminProps.push(_prop);
        }
      });
      configProperties = adminProps.concat(configProperties);
    }
    configProperties = configProperties.filter(function(item) {
      return installedServiceNames.contains(item.get('serviceName'));
    });
    if (this.get('wizardController.name') != 'addServiceController') {
      var realmValue = storedServiceConfigs.findProperty('name', 'realm').value;
      configProperties.findProperty('name', 'realm').set('value', realmValue);
      configProperties.findProperty('name', 'realm').set('defaultValue', realmValue);
    }

    configProperties.setEach('isSecureConfig', false);
    configProperties.forEach(function(property, item, allConfigs) {
      if (['spnego_keytab', 'spnego_principal'].contains(property.get('name'))) {
        property.addObserver('value', self, 'spnegoPropertiesObserver');
      }
      if (property.get('observesValueFrom')) {
        var observedValue = allConfigs.findProperty('name', property.get('observesValueFrom')).get('value');
        property.set('value', observedValue);
        property.set('defaultValue', observedValue);
      }
      if (property.get('serviceName') == 'Cluster') {
        property.set('category', 'Global');
      }
      else {
        property.set('category', property.get('serviceName'));
      }
      // All user identity except storm should be grouped under "Ambari Principals" category
      if (property.get('identityType') == 'user' && property.get('serviceName') !== 'STORM') property.set('category', 'Ambari Principals');
      var siteProperty = siteProperties.findProperty('name', property.get('name'));
      if (siteProperty) {
        if (siteProperty.category === property.get('category')) {
          property.set('displayName',siteProperty.displayName);
          if (siteProperty.index) {
            property.set('index', siteProperty.index);
          }
        }
        if (siteProperty.displayType) {
          property.set('displayType', siteProperty.displayType);
        }
      }
      this.tweakConfigProperty(property);
    },this);

    return configProperties;
  },

  /**
   * Function to override kerberos descriptor's property values
   */
  tweakConfigProperty: function(config) {
    if (config.name === 'templeton.hive.properties') {
      var defaultHiveMsPort = "9083";
      var hiveMSHosts = App.HostComponent.find().filterProperty('componentName', 'HIVE_METASTORE');
      if (hiveMSHosts.length > 1) {
        var hiveMSHostNames = hiveMSHosts.mapProperty('hostName');
        var port = config.value.match(/:[0-9]{2,4}/);
        port = port ? port[0].slice(1) : defaultHiveMsPort;
        for (var i = 0; i < hiveMSHostNames.length; i++) {
          hiveMSHostNames[i] = "thrift://" + hiveMSHostNames[i] + ":" + port;
        }
        var configValue =  config.value.replace(/thrift.+[0-9]{2,},/i, hiveMSHostNames.join('\\,') + ",");
        config.set('value', configValue);
        config.set('defaultValue', configValue);
      }
    }
  },

  /**
   * Sync up values between inherited property and its reference.
   *
   * @param {App.ServiceConfigProperty} configProperty
   */
  spnegoPropertiesObserver: function(configProperty) {
    var self = this;
    var stepConfig =  this.get('stepConfigs').findProperty('name', 'ADVANCED');
    stepConfig.get('configs').forEach(function(config) {
      if (config.get('observesValueFrom') == configProperty.get('name')) {
        Em.run.once(self, function() {
          config.set('value', configProperty.get('value'));
          config.set('defaultValue', configProperty.get('value'));
        });
      }
    });
  },

  submit: function() {
    this.saveConfigurations();
    App.router.send('next');
  },

  saveConfigurations: function() {
    var kerberosDescriptor = this.get('kerberosDescriptor');
    var configs = [];
    this.get('stepConfigs').forEach(function(_stepConfig){
      configs = configs.concat(_stepConfig.get('configs'));
    });
    this.updateKerberosDescriptor(kerberosDescriptor, configs);
    App.get('router.kerberosWizardController').saveKerberosDescriptorConfigs(kerberosDescriptor);
  }
});

});

require.register("controllers/main/admin/kerberos/step5_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

App.KerberosWizardStep5Controller = App.KerberosProgressPageController.extend({
  name: 'kerberosWizardStep5Controller',
  clusterDeployState: 'KERBEROS_DEPLOY',
  commands: ['stopServices'],

  stopServices: function () {
    App.ajax.send({
      name: 'common.services.update',
      data: {
        context: "Stop services",
        "ServiceInfo": {
          "state": "INSTALLED"
        }
      },
      sender: this,
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  loadStep: function() {
    this.checkComponentsRemoval();
    this._super();
  },

  /**
   * remove Application Timeline Server component if needed.
   */
  checkComponentsRemoval: function() {
    if (App.Service.find().someProperty('serviceName', 'YARN') && !App.get('doesATSSupportKerberos')
      && !this.get('commands').contains('deleteATS') && App.HostComponent.find().findProperty('componentName', 'APP_TIMELINE_SERVER')) {
      this.get('commands').pushObject('deleteATS');
    }
  },

  /**
   * Remove Application Timeline Server from the host.
   * @returns {$.Deferred}
   */
  deleteATS: function() {
    return App.ajax.send({
      name: 'common.delete.host_component',
      sender: this,
      data: {
        componentName: 'APP_TIMELINE_SERVER',
        hostName: App.HostComponent.find().findProperty('componentName', 'APP_TIMELINE_SERVER').get('hostName')
      },
      success: 'onDeleteATSSuccess',
      error: 'onDeleteATSError'
    });
  },

  onDeleteATSSuccess: function() {
    this.onTaskCompleted();
  },

  onDeleteATSError: function(error) {
    if (error.responseText.indexOf('org.apache.ambari.server.controller.spi.NoSuchResourceException') !== -1) {
      this.onDeleteATSSuccess();
    }
  }
});

});

require.register("controllers/main/admin/kerberos/step6_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

App.KerberosWizardStep6Controller = App.KerberosProgressPageController.extend({
  name: 'kerberosWizardStep6Controller',
  clusterDeployState: 'KERBEROS_DEPLOY',
  isSingleRequestPage: true,
  request: {},
  commands: [],
  contextForPollingRequest: Em.I18n.t('requestInfo.kerberizeCluster'),

  /**
   * Define whether show Back button
   * @type {Boolean}
   */
  isBackButtonDisabled: true,

  /**
   * Start cluster kerberization. On retry just unkerberize and kerbrize cluster.
   * @param {bool} isRetry
   */
  setRequest: function (isRetry) {
    var self = this;
    var kerberizeRequest = {
      name: 'KERBERIZE_CLUSTER',
      ajaxName: 'admin.kerberize.cluster',
      ajaxData: {
        data: {
          Clusters: {
            security_type: "KERBEROS"
          }
        }
      }
    };
    if (isRetry) {
      // on retry we have to unkerberize cluster
      this.unkerberizeCluster().always(function() {
        // clear current request object before start of kerberize process
        self.set('request', kerberizeRequest);
        self.clearStage();
        self.loadStep();
      });
    } else {
      this.set('request', kerberizeRequest);
    }
  },

  /**
   * Send request to unkerberisze cluster
   * @returns {$.ajax}
   */
  unkerberizeCluster: function () {
    return App.ajax.send({
      name: 'admin.unkerberize.cluster',
      sender: this,
      success: 'goToNextStep',
      error: 'goToNextStep'
    });
  },

  goToNextStep: function() {
    this.clearStage();
    App.router.transitionTo('step6');
  },

  postKerberosDescriptor: function (kerberosDescriptor) {
    return App.ajax.send({
      name: 'admin.kerberos.cluster.artifact.create',
      sender: this,
      data: {
        artifactName: 'kerberos_descriptor',
        data: {
          artifact_data: kerberosDescriptor
        }
      }
    });
  },

  /**
   * Send request to update kerberos descriptor
   * @param kerberosDescriptor
   * @returns {$.ajax|*}
   */
  putKerberosDescriptor: function (kerberosDescriptor) {
    return App.ajax.send({
      name: 'admin.kerberos.cluster.artifact.update',
      sender: this,
      data: {
        artifactName: 'kerberos_descriptor',
        data: {
          artifact_data: kerberosDescriptor
        }
      },
      success: 'unkerberizeCluster',
      error: 'unkerberizeCluster'
    });
  },

  retry: function () {
    this.set('showRetry', false);
    this.get('tasks').setEach('status', 'PENDING');
    App.router.send('retry');
  },


  /**
   * Enable or disable previous steps according to tasks statuses
   */
  enableDisablePreviousSteps: function () {
    var wizardController = App.router.get(this.get('content.controllerName'));
    if (this.get('tasks').someProperty('status', 'FAILED')) {
      wizardController.enableStep(4);
      this.set('isBackButtonDisabled', false);
    } else {
      wizardController.setLowerStepsDisable(6);
      this.set('isBackButtonDisabled', true);
    }
  }.observes('tasks.@each.status')
});

});

require.register("controllers/main/admin/kerberos/step7_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

App.KerberosWizardStep7Controller = App.KerberosProgressPageController.extend({
  name: 'kerberosWizardStep7Controller',
  clusterDeployState: 'KERBEROS_DEPLOY',
  commands: ['startServices'],

  startServices: function () {
    App.ajax.send({
      name: 'common.services.update',
      sender: this,
      data: {
        "context": "Start services",
        "ServiceInfo": {
          "state": "STARTED"
        },
        urlParams: "params/run_smoke_test=true"
      },
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  isSubmitDisabled: function () {
    return !["COMPLETED", "FAILED"].contains(this.get('status'));
  }.property('status')
});
});

require.register("controllers/main/admin/kerberos/wizard_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.KerberosWizardController = App.WizardController.extend({

  name: 'kerberosWizardController',

  totalSteps: 7,

  isKerberosWizard: true,

  /**
   * Used for hiding back button in wizard
   */
  hideBackButton: true,

  kerberosDescriptorConfigs: null,

  content: Em.Object.create({
    controllerName: 'kerberosWizardController',
    serviceName: 'KERBEROS',
    kerberosOption: null,
    cluster: null,
    services: [],
    advancedServiceConfig: null,
    serviceConfigProperties: [],
    failedTask: null
  }),

  /**
   * set current step
   * @param {string} currentStep
   * @param {boolean} completed
   * @param {boolean} skipStateSave
   */
  setCurrentStep: function (currentStep, completed, skipStateSave) {
    this._super(currentStep, completed);
    if (App.get('testMode') || skipStateSave) {
      return;
    }
    App.clusterStatus.setClusterStatus({
      clusterName: this.get('content.cluster.name'),
      clusterState: 'KERBEROS_DEPLOY',
      wizardControllerName: 'kerberosWizardController',
      localdb: App.db.data
    });
  },

  setStepsEnable: function () {
    for (var i = 1; i <= this.get('totalSteps'); i++) {
      var step = this.get('isStepDisabled').findProperty('step', i);
      if (i <= this.get('currentStep') && App.get('router.clusterController.isLoaded')) {
        step.set('value', false);
      } else {
        step.set('value', i != this.get('currentStep'));
      }
    }
  }.observes('currentStep', 'App.router.clusterController.isLoaded'),

  /**
   * return new object extended from clusterStatusTemplate
   * @return Object
   */
  getCluster: function () {
    return jQuery.extend({}, this.get('clusterStatusTemplate'), {name: App.get('router').getClusterName()});
  },

  /**
   *  Gets the
   * @returns {*} jquery promise
   */
  getClusterEnvData: function () {
    var dfd = $.Deferred();
    var self = this;
    var siteName = 'cluster-env';
    var tags = [{siteName: siteName}];
    App.get('router.configurationController').getConfigsByTags(tags).done(function (data) {
      var properties = self.updateClusterEnvData(data[0].properties);
      var clusterConfig = {"type": siteName, "tag": 'version' + (new Date).getTime(), "properties": properties};
      var clusterConfigData = {
        Clusters: {
          desired_config: clusterConfig
        }
      };
      dfd.resolve(clusterConfigData);
    });
    return dfd;
  },

  updateClusterEnvData: function (configs) {
    var kerberosDescriptor = this.kerberosDescriptorConfigs;
    configs['security_enabled'] = true;
    configs['kerberos_domain'] = kerberosDescriptor.properties.realm;
    return configs;
  },

  /**
   * save status of the cluster.
   * @param clusterStatus object with status,requestId fields.
   */
  saveClusterStatus: function (clusterStatus) {
    var oldStatus = this.toObject(this.get('content.cluster'));
    clusterStatus = jQuery.extend(oldStatus, clusterStatus);
    if (clusterStatus.requestId) {
      clusterStatus.requestId.forEach(function (requestId) {
        if (clusterStatus.oldRequestsId.indexOf(requestId) === -1) {
          clusterStatus.oldRequestsId.push(requestId)
        }
      }, this);
    }
    this.set('content.cluster', clusterStatus);
    this.save('cluster');
  },


  saveTasksStatuses: function (statuses) {
    this.setDBProperty('tasksStatuses',statuses);
    this.set('content.tasksStatuses', statuses);
  },

  saveConfigTag: function (tag) {
    App.db.setKerberosWizardConfigTag(tag);
    this.set('content.' + [tag.name], tag.value);
  },

  saveKerberosOption: function (stepController) {
    this.setDBProperty('kerberosOption', stepController.get('selectedItem'));
    this.set('content.kerberosOption', stepController.get('selectedItem'));
  },

  loadTasksStatuses: function () {
    var statuses = this.getDBProperty('tasksStatuses');
    this.set('content.tasksStatuses', statuses);
  },

  /**
   * Load serviceConfigProperties to model
   */
  loadServiceConfigProperties: function () {
    var serviceConfigProperties = this.getDBProperty('serviceConfigProperties');
    this.set('content.serviceConfigProperties', serviceConfigProperties);
  },

  /**
   * load advanced configs from server
   */
  loadAdvancedConfigs: function (dependentController) {
    var self = this;
    var loadAdvancedConfigResult = [];
    dependentController.set('isAdvancedConfigLoaded', false);
    var serviceName = this.get('content.serviceName');
    App.config.loadAdvancedConfig(serviceName, function (properties) {
      loadAdvancedConfigResult.pushObjects(properties);
      self.set('content.advancedServiceConfig', loadAdvancedConfigResult);
      self.setDBProperty('advancedServiceConfig', loadAdvancedConfigResult);
      dependentController.set('isAdvancedConfigLoaded', true);
    });
  },

  loadKerberosDescriptorConfigs: function () {
    var kerberosDescriptorConfigs = this.getDBProperty('kerberosDescriptorConfigs');
    this.kerberosDescriptorConfigs =  kerberosDescriptorConfigs;
  },


  saveRequestIds: function (requestIds) {
    this.setDBProperty('requestIds',requestIds);
    this.set('content.requestIds', requestIds);
  },

  loadKerberosOption: function () {
    this.set('content.kerberosOption', this.getDBProperty('kerberosOption'));
  },

  loadRequestIds: function () {
    var requestIds = this.getDBProperty('requestIds');
    this.set('content.requestIds', requestIds);
  },

  saveTasksRequestIds: function (requestIds) {
    this.setDBProperty('tasksRequestIds',requestIds);
    this.set('content.tasksRequestIds', requestIds);
  },

  loadTasksRequestIds: function () {
    var requestIds = this.getDBProperty('tasksRequestIds');
    this.set('content.tasksRequestIds', requestIds);
  },

  saveKerberosDescriptorConfigs: function (kerberosDescriptorConfigs) {
    this.setDBProperty('kerberosDescriptorConfigs',kerberosDescriptorConfigs);
    this.kerberosDescriptorConfigs =  kerberosDescriptorConfigs;
  },


  loadMap: {
    '1': [
      {
        type: 'sync',
        callback: function () {
          this.loadKerberosOption();
        }
      }
    ],
    '2': [
      {
        type: 'sync',
        callback: function () {
          var kerberosStep2controller = App.get('router.kerberosWizardStep2Controller');
          this.loadAdvancedConfigs(kerberosStep2controller);
          this.loadServiceConfigProperties();
        }
      }
    ],
    '3': [
      {
        type: 'sync',
        callback: function () {
          this.loadTasksStatuses();
          this.loadTasksRequestIds();
          this.loadRequestIds();
        }
      }
    ],
    '4': [
      {
        type: 'sync',
        callback: function () {
          this.loadKerberosDescriptorConfigs();
        }
      }
    ],
    '5': [
      {
        type: 'sync',
        callback: function () {
          this.loadKerberosDescriptorConfigs();
        }
      }
    ]
  },

  /**
   * Remove all loaded data.
   * Created as copy for App.router.clearAllSteps
   */
  clearAllSteps: function () {
    this.clearInstallOptions();
    // clear temporary information stored during the install
    this.set('content.cluster', this.getCluster());
  },

  clearTasksData: function () {
    this.saveTasksStatuses(undefined);
    this.saveRequestIds(undefined);
    this.saveTasksRequestIds(undefined);
  },
  /**
   * shows popup with to warn user
   * @param primary
   * @param isCritical
   */
  warnBeforeExitPopup: function(primary, isCritical) {
    var primaryText = Em.I18n.t('common.exitAnyway');
    var msg = isCritical ? Em.I18n.t('admin.kerberos.wizard.exit.critical.msg')
      : Em.I18n.t('admin.kerberos.wizard.exit.warning.msg');
    return App.showConfirmationPopup(primary, msg, null, null, primaryText, isCritical);
  },
  /**
   * Clear all temporary data
   */
  finish: function () {
    // The in-memory variable for current step should be reset to 1st step.
    this.setCurrentStep('1', false, true);
    // kerberos wizard namespace in the localStorage should be emptied
    this.resetDbNamespace();
    App.get('router.updateController').updateAll();
  }
});

});

require.register("controllers/main/admin/security", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
App.MainAdminSecurityController = Em.Controller.extend({
  name: 'mainAdminSecurityController',
  isSubmitDisabled: false,
  securityEnabled: false,
  dataIsLoaded: false,
  isRecommendedLoaded: true,
  serviceUsers: [],
  tag: {},
  getAddSecurityWizardStatus: function () {
    return App.db.getSecurityWizardStatus();
  },
  setAddSecurityWizardStatus: function (status) {
    App.db.setSecurityWizardStatus(status);
  },

  setDisableSecurityStatus: function (status) {
    App.db.setDisableSecurityStatus(status);
  },
  getDisableSecurityStatus: function (status) {
    return App.db.getDisableSecurityStatus();
  },
  content: Em.Object.create({
    isATSInstalled: function() {
      // Because the ATS component can be installed/removed at will, the check has to happen every time that security is added.
      var yarnService = App.Service.find().findProperty('serviceName','YARN');
      return !!yarnService && yarnService.get('hostComponents').someProperty('componentName', 'APP_TIMELINE_SERVER');
    }.property('App.router.clusterController.isLoaded')
  }),
  notifySecurityOff: false,
  notifySecurityAdd: false,

  stepConfigs: [],
  desiredConfigs: [],
  securityUsers: [],
  serviceConfigTags: [],
  selectedService: null,
  isNotEditable: true,
  /** need to define <code>filter, filterColumns</code> properties
   * for preventing errors in <code>App.ServiceConfigsByCategoryView</code>
   */
  filter: '',
  filterColumns: function () {
    return [];
  }.property(''),
  services: function () {
    var secureServices = $.extend(true, [], require('data/HDP2/secure_configs'));
    var services = [];

    // Typically, ATS will support Kerberos in HDP 2.2 and higher
    if (this.get('content.isATSInstalled') && App.get('doesATSSupportKerberos')) {
      var yarnConfigCategories = secureServices.findProperty('serviceName', 'YARN').configCategories;
      yarnConfigCategories.push(App.ServiceConfigCategory.create({ name: 'AppTimelineServer', displayName : 'Application Timeline Service'}));
    }

    var installedServices = App.Service.find().mapProperty('serviceName');
    //General (only non service tab) tab is always displayed
    services.push(secureServices.findProperty('serviceName', 'GENERAL'));
    installedServices.forEach(function (_service) {
      var secureService = secureServices.findProperty('serviceName', _service);
      if (secureService) {
        services.push(secureService);
      }
    }, this);
    return services;
  }.property('App.router.clusterController.isLoaded'),

  /**
   * default values of configs, which contains user names
   */
  userNameMap: {
    'hdfs_user': {defaultValue: 'hdfs', siteName: 'hadoop-env', serviceName: 'HDFS'},
    'yarn_user': {defaultValue: 'yarn', siteName: 'yarn-env', serviceName: 'YARN'},
    'mapred_user': {defaultValue: 'mapred', siteName: 'mapred-env', serviceName: 'MAPREDUCE2'},
    'hbase_user': {defaultValue: 'hbase', siteName: 'hbase-env', serviceName: 'HBASE'},
    'hive_user': {defaultValue: 'hive', siteName: 'hive-env', serviceName: 'HIVE'},
    'proxyuser_group': {defaultValue: 'users', siteName: 'hadoop-env', serviceName: 'HDFS'},
    'smokeuser': {defaultValue: 'ambari-qa', siteName: 'cluster-env', serviceName: 'CLUSTER'},
    'zk_user': {defaultValue: 'zookeeper', siteName: 'zookeeper-env', serviceName: 'ZOOKEEPER'},
    'oozie_user': {defaultValue: 'oozie', siteName: 'oozie-env', serviceName: 'OOZIE'},
    'user_group': {defaultValue: 'hadoop', siteName: 'hadoop-env', serviceName: 'HDFS'},
    'storm_user': {defaultValue: 'storm', siteName: 'storm-env', serviceName: 'STORM'},
    'falcon_user': {defaultValue: 'falcon', siteName: 'falcon-env', serviceName: 'FALCON'},
    'knox_user': {defaultValue: 'knox', siteName: 'knox-env', serviceName: 'KNOX'}
  },

  loadStep: function () {
    var step2Controller = App.router.get('mainAdminSecurityAddStep2Controller');
    var services = this.get('services');
    var self = this;
    step2Controller.set('content', Em.Object.create({services: []}));
    step2Controller.set('content.services', services);
    this.get('stepConfigs').clear();
    this.get('securityUsers').clear();
    this.get('serviceConfigTags').clear();
    this.loadSecurityUsers();
    //loadSecurityUsers - desired configs fetched from server
    step2Controller.addUserPrincipals(services, this.get('securityUsers'));
    step2Controller.addMasterHostToConfigs();
    step2Controller.addSlaveHostToConfigs();
    this.renderServiceConfigs(services);
    step2Controller.changeCategoryOnHa(services, this.get('stepConfigs'));

    services.forEach(function (_secureService) {
      this.setServiceTagNames(_secureService, this.get('desiredConfigs'));
    }, this);
    App.router.get('configurationController').getConfigsByTags(this.get('serviceConfigTags')).done(function (serverConfigs) {
      self.setConfigValuesFromServer(self.get('stepConfigs'), serverConfigs);
      self.set('installedServices', App.Service.find().mapProperty('serviceName'));
    });
  },

  /**
   * get actual values of configurations from server
   * @param stepConfigs
   * @param serverConfigs
   */
  setConfigValuesFromServer: function (stepConfigs, serverConfigs) {
    var allConfigs = {};
    serverConfigs.mapProperty('properties').forEach(function (_properties) {
      allConfigs = $.extend(allConfigs, _properties);
    }, this);
    // for all services`
    stepConfigs.forEach(function (_content) {
      //for all components
      _content.get('configs').forEach(function (_config) {

        var componentVal = allConfigs[_config.get('name')];
        //if we have config for specified component
        if (componentVal) {
          //set it
          _config.set('value', componentVal);
        }
      }, this);
    }, this);
  },

  /**
   * set tag names according to installed services and desired configs
   * @param secureService
   * @param configs
   * @return {Object}
   */
  setServiceTagNames: function (secureService, configs) {
    for (var index in configs) {
      if (secureService.sites && secureService.sites.contains(index)) {
        var serviceConfigObj = {
          siteName: index,
          tagName: configs[index].tag,
          newTagName: null,
          configs: {}
        };
        this.get('serviceConfigTags').pushObject(serviceConfigObj);
      }
    }
    return serviceConfigObj;
  },

  loadSecurityUsers: function () {
    var securityUsers = this.get('serviceUsers');
    if (!securityUsers || securityUsers.length < 1) { // Page could be refreshed in middle
      if (App.get('testMode')) {
        securityUsers.pushObject({id: 'puppet var', name: 'hdfs_user', value: 'hdfs'});
        securityUsers.pushObject({id: 'puppet var', name: 'mapred_user', value: 'mapred'});
        securityUsers.pushObject({id: 'puppet var', name: 'hbase_user', value: 'hbase'});
        securityUsers.pushObject({id: 'puppet var', name: 'hive_user', value: 'hive'});
        securityUsers.pushObject({id: 'puppet var', name: 'smokeuser', value: 'ambari-qa'});
      } else {
        this.setSecurityStatus();
        securityUsers = this.get('serviceUsers');
      }
    }
    this.set('securityUsers', securityUsers);
  },
  /**
   * Load child components to service config object
   * @param _componentConfig
   * @param componentConfig
   */
  loadComponentConfigs: function (_componentConfig, componentConfig) {
    _componentConfig.configs.forEach(function (_serviceConfigProperty) {
      var serviceConfigProperty = App.ServiceConfigProperty.create(_serviceConfigProperty);
      componentConfig.configs.pushObject(serviceConfigProperty);
      serviceConfigProperty.set('isEditable', serviceConfigProperty.get('isReconfigurable'));
      serviceConfigProperty.validate();
    }, this);
  },

  /**
   * Render configs for active services
   * @param serviceConfigs
   */
  renderServiceConfigs: function (serviceConfigs) {
    serviceConfigs.forEach(function (_serviceConfig) {

      var serviceConfig = App.ServiceConfig.create({
        filename: _serviceConfig.filename,
        serviceName: _serviceConfig.serviceName,
        displayName: _serviceConfig.displayName,
        configCategories: _serviceConfig.configCategories,
        showConfig: true,
        configs: []
      });

      this.loadComponentConfigs(_serviceConfig, serviceConfig);

      console.log('pushing ' + serviceConfig.serviceName, serviceConfig);

      this.get('stepConfigs').pushObject(serviceConfig);
    }, this);
    this.set('selectedService', this.get('stepConfigs').filterProperty('showConfig', true).objectAt(0));
  },

  notifySecurityOffPopup: function () {
    var self = this;
    if (!this.get('isSubmitDisabled')) {
      App.ModalPopup.show({
        header: Em.I18n.t('popup.confirmation.commonHeader'),
        primary: Em.I18n.t('ok'),
        onPrimary: function () {
          App.db.setSecurityDeployCommands(undefined);
          self.setDisableSecurityStatus("RUNNING");
          App.router.transitionTo('disableSecurity');
          this.hide();
        },
        bodyClass: Ember.View.extend({
          templateName: require('templates/main/admin/security/notify_security_off_popup')
        })
      })
    }
  },

  getUpdatedSecurityStatus: function () {
    this.setSecurityStatus();
    return this.get('securityEnabled');
  },

  setSecurityStatus: function () {
    if (App.get('testMode')) {
      this.set('securityEnabled', !App.get('testEnableSecurity'));
      this.set('dataIsLoaded', true);
    } else {
      //get Security Status From Server
      App.ajax.send({
        name: 'admin.security_status',
        sender: this,
        success: 'getSecurityStatusFromServerSuccessCallback',
        error: 'errorCallback'
      });
    }
  },

  errorCallback: function (jqXHR) {
    this.set('dataIsLoaded', true);
    // Show the error popup if the API call received a response from the server.
    // jqXHR.status will be empty when browser cancels the request. Refer to AMBARI-5921 for more info
    if (!!jqXHR.status) {
      this.showSecurityErrorPopup();
    }
  },

  getSecurityStatusFromServerSuccessCallback: function (data) {
    var configs = data.Clusters.desired_configs;
    var serviceNames = this.get('services').mapProperty('serviceName');
    var configTags = [];
    this.set('desiredConfigs', configs);
    for (var key in this.userNameMap) {
      if (serviceNames.contains(this.userNameMap[key]['serviceName']) || this.userNameMap[key]['serviceName'] === 'CLUSTER')
        configTags.push(this.userNameMap[key]['siteName']);
    }
    configTags = configTags.uniq();

    var errorFlag = false;
    configTags.forEach(function (_tag) {
      if (!configs[_tag]) {
        errorFlag = true;
      }
    }, this);

    if (errorFlag) {
      this.showSecurityErrorPopup();
    }  else {
      var tags = configTags.map(function (_tag) {
        this.set('tag.' + _tag, configs[_tag].tag);
        return {
          siteName: _tag,
          tagName: configs[_tag].tag
        }
      }, this);

      if ('hdfs-site' in configs) {
        this.set('tag.hdfs-site', configs['hdfs-site'].tag);
        tags.pushObject({
          siteName: "hdfs-site",
          tagName: this.get('tag.hdfs-site')
        });
      }
      this.getServiceConfigsFromServer(tags);
    }
  },

  getServiceConfigsFromServer: function (tags) {
    var self = this;

    App.router.get('configurationController').getConfigsByTags(tags).done(function (data) {
      var configs = data.findProperty('tag', self.get('tag.cluster-env')).properties;
      if (configs && (configs['security_enabled'] === 'true' || configs['security_enabled'] === true)) {
        self.set('securityEnabled', true);
      }
      else {
        self.set('securityEnabled', false);
        if (!!self.get('tag.hdfs-site')) {
          var hdfsConfigs = data.findProperty('tag', self.get('tag.hdfs-site')).properties;
          self.setNnHaStatus(hdfsConfigs);
        }
      }
      var userConfigs = {};
      data.forEach(function(_config){
        $.extend(userConfigs, _config.properties);
      });
      self.loadUsers(userConfigs);
      self.set('dataIsLoaded', true);
    });
  },

  setNnHaStatus: function (hdfsConfigs) {
    var nnHaStatus = hdfsConfigs && hdfsConfigs['dfs.nameservices'];
    var namenodesKey;
    if (nnHaStatus) {
      namenodesKey = 'dfs.ha.namenodes.' + hdfsConfigs['dfs.nameservices'];
    }
    if (nnHaStatus && hdfsConfigs[namenodesKey]) {
      App.db.setIsNameNodeHa('true');
    } else {
      App.db.setIsNameNodeHa('false');
    }
  },

  /**
   * load users names,
   * substitute missing values with default
   * @param configs {Object}
   */
  loadUsers: function (configs) {
    var defaultUserNameMap = this.get('userNameMap');
    this.set('serviceUsers',[]);

    for (var configName in defaultUserNameMap) {
      this.get('serviceUsers').push({
        id: 'puppet var',
        name: configName,
        value: configs[configName] || defaultUserNameMap[configName]['defaultValue']
      });
    }
    App.db.setSecureUserInfo(this.get('serviceUsers'));
  },

  showSecurityErrorPopup: function () {
    App.ModalPopup.show({
      header: Em.I18n.t('common.error'),
      secondary: false,
      bodyClass: Ember.View.extend({
        template: Ember.Handlebars.compile('<p>{{t admin.security.status.error}}</p>')
      })
    });
  }
});



});

require.register("controllers/main/admin/security/add/addSecurity_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
App.AddSecurityController = App.WizardController.extend({

  name: 'addSecurityController',
  securityEnabled: false,

  totalSteps: 4,

  content: Em.Object.create({
    services: [],
    isNnHa: 'false',
    serviceConfigProperties: null,
    controllerName: 'addSecurityController',
    isATSInstalled: function() {
      // Because the ATS component can be installed/removed at will, the check has to happen every time that security is added.
      var yarnService = App.Service.find().findProperty('serviceName','YARN');
      return !!yarnService && yarnService.get('hostComponents').someProperty('componentName', 'APP_TIMELINE_SERVER');
    }.property('App.router.clusterController.isLoaded')
  }),

  /**
   * installed services on cluster
   */
  installedServices: function () {
    return App.Service.find().mapProperty('serviceName');
  }.property(),

  /**
   * services with security configurations
   */
  secureServices: function () {
    var configCategories = require('data/HDP2/secure_configs');
    if (this.get('content.isATSInstalled') && App.get('doesATSSupportKerberos')) {
      var yarnConfigCategories = configCategories.findProperty('serviceName', 'YARN').configCategories;
      yarnConfigCategories.push(App.ServiceConfigCategory.create({ name: 'AppTimelineServer', displayName : 'Application Timeline Service'}));
    }
    return configCategories;
  }.property('App.router.clusterController.isLoaded'),

  /**
   * Loads all prior steps on refresh
   */
  loadAllPriorSteps: function () {
    var step = this.get('currentStep');
    switch (step) {
      case '4':
      case '3':
      case '2':
        this.loadServiceConfigs();
      case '1':
        this.loadServices();
        this.loadNnHaStatus();
    }
  },
  /**
   * Load installed services, which match secure services, to content
   */
  loadServices: function () {
    var secureServices = this.get('secureServices');
    var installedServices = this.get('installedServices');

    this.get('content.services').clear();
    //General (only non service tab) tab is always displayed
    this.get('content.services').push(secureServices.findProperty('serviceName', 'GENERAL'));
    installedServices.forEach(function (_service) {
      var secureService = secureServices.findProperty('serviceName', _service);
      if (secureService) {
        this.get('content.services').push(secureService);
      }
    }, this);
  },
  /**
   * identify whether NameNode in high availability mode
   */
  loadNnHaStatus: function () {
    this.set('content.isNnHa', App.db.getIsNameNodeHa());
  },

  /**
   * save service config properties to localStorage
   * @param stepController
   */
  saveServiceConfigProperties: function (stepController) {
    var serviceConfigProperties = [];
    stepController.get('stepConfigs').forEach(function (_content) {
      _content.get('configs').forEach(function (_configProperties) {
        _configProperties.set('value', App.config.trimProperty(_configProperties, true));
        var configProperty = {
          id: _configProperties.get('id'),
          name: _configProperties.get('name'),
          value: _configProperties.get('value'),
          defaultValue: _configProperties.get('defaultValue'),
          serviceName: _configProperties.get('serviceName'),
          domain: _configProperties.get('domain'),
          filename: _configProperties.get('filename'),
          unit: _configProperties.get('unit'),
          components: _configProperties.get('components'),
          component: _configProperties.get('component'),
          overrides: this.getConfigOverrides(_configProperties)
        };
        serviceConfigProperties.push(configProperty);
      }, this);
    }, this);
    App.db.setSecureConfigProperties(serviceConfigProperties);
    this.set('content.serviceConfigProperties', serviceConfigProperties);
  },

  /**
   * get overrides of config
   * @param _configProperties
   * @return {Array}
   */
  getConfigOverrides: function (_configProperties) {
    var overrides = _configProperties.get('overrides');
    var overridesArray = [];
    if (Array.isArray(overrides)) {
      overrides.forEach(function (override) {
        var overrideEntry = {
          value: override.get('value'),
          hosts: []
        };
        override.get('selectedHostOptions').forEach(function (host) {
          overrideEntry.hosts.push(host);
        });
        overridesArray.push(overrideEntry);
      });
    }
    return (overridesArray.length > 0) ? overridesArray : null;
  },

  /**
   * Load service config properties from localStorage
   */
  loadServiceConfigs: function () {
    this.set('content.serviceConfigProperties', App.db.getSecureConfigProperties());
  },

  /**
   * Clear all local storage data for Add security wizard namespace
   */
  finish: function () {
    this.resetDbNamespace();
  }
});


});

require.register("controllers/main/admin/security/add/step1", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
App.MainAdminSecurityAddStep1Controller = Em.Controller.extend({

  name: 'mainAdminSecurityAddStep1Controller',
  /**
   * identify whether ATS(Application Timeline Server) is installed and does not support Kerberization.
   * @return {Boolean}
   */
  shouldRemoveATS: function() {
    var isATSInstalled = this.get('content.isATSInstalled');
    var doesATSSupportKerberos = App.get("doesATSSupportKerberos");
    return isATSInstalled && !doesATSSupportKerberos;
  }
});

});

require.register("controllers/main/admin/security/add/step2", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

App.MainAdminSecurityAddStep2Controller = Em.Controller.extend({

  name: 'mainAdminSecurityAddStep2Controller',
  isRecommendedLoaded: true,
  stepConfigs: [],
  installedServices: [],
  selectedService: null,
  securityUsers: [],
  filter: '',
  filterColumns: [],

  /**
   * map which depict connection between config and slave component
   * in order to set component hosts to config value
   */
  slaveComponentMap: [
    {
      serviceName: 'HDFS',
      configName: 'datanode_hosts',
      component: 'DATANODE'
    },
    {
      serviceName: 'YARN',
      configName: 'nodemanager_host',
      component: 'NODEMANAGER'
    },
    {
      serviceName: 'HBASE',
      configName: 'regionserver_hosts',
      component: 'HBASE_REGIONSERVER'
    },
    {
      serviceName: 'STORM',
      configName: 'supervisor_hosts',
      component: 'SUPERVISOR'
    }
  ],
  /**
   * map which depict connection between config and master component
   * in order to set component hosts to config value
   */
  masterComponentMap: [
    {
      serviceName: 'OOZIE',
      configName: 'oozie_servername',
      components: ['OOZIE_SERVER']
    },
    {
      serviceName: 'HIVE',
      configName: 'hive_metastore',
      components: ['HIVE_METASTORE', 'HIVE_SERVER']
    },
    {
      serviceName: 'HIVE',
      configName: 'webhcatserver_host',
      components: ['WEBHCAT_SERVER']
    },
    {
      serviceName: 'HDFS',
      configName: 'namenode_host',
      components: ['NAMENODE']
    },
    {
      serviceName: 'HDFS',
      configName: 'snamenode_host',
      components: ['SECONDARY_NAMENODE']
    },
    {
      serviceName: 'HDFS',
      configName: 'journalnode_hosts',
      components: ['JOURNALNODE']
    },
    {
      serviceName: 'MAPREDUCE2',
      configName: 'jobhistoryserver_host',
      components: ['HISTORYSERVER']
    },
    {
      serviceName: 'YARN',
      configName: 'resourcemanager_host',
      components: ['RESOURCEMANAGER']
    },
    {
      serviceName: 'YARN',
      configName: 'apptimelineserver_host',
      components: ['APP_TIMELINE_SERVER']
    },
    {
      serviceName: 'HBASE',
      configName: 'hbasemaster_host',
      components: ['HBASE_MASTER']
    },
    {
      serviceName: 'ZOOKEEPER',
      configName: 'zookeeperserver_hosts',
      components: ['ZOOKEEPER_SERVER']
    },
    {
      serviceName: 'FALCON',
      configName: 'falcon_server_host',
      components: ['FALCON_SERVER']
    },
    {
      serviceName: 'STORM',
      configName: 'storm_host',
      components: ['STORM_UI_SERVER', 'NIMBUS', 'SUPERVISOR']
    },
    {
      serviceName: 'STORM',
      configName: 'nimbus_host',
      components: ['NIMBUS']
    },
    {
      serviceName: 'KNOX',
      configName: 'knox_gateway_hosts',
      components: ['KNOX_GATEWAY']
    }
  ],

  hostToPrincipalMap: [
    {
      serviceName: 'OOZIE',
      configName: 'oozie_servername',
      principalName: 'oozie_principal_name',
      primaryName: 'oozie/'
    },
    {
      serviceName: 'OOZIE',
      configName: 'oozie_servername',
      principalName: 'oozie_http_principal_name',
      primaryName: 'HTTP/'
    },
    {
      serviceName: 'FALCON',
      configName: 'falcon_server_host',
      principalName: 'falcon_principal_name',
      primaryName: 'falcon/'
    },
    {
      serviceName: 'FALCON',
      configName: 'falcon_server_host',
      principalName: 'falcon_http_principal_name',
      primaryName: 'HTTP/'
    },
    {
      serviceName: 'HIVE',
      configName: 'webhcatserver_host',
      principalName: 'webHCat_http_principal_name',
      primaryName: 'HTTP/'
    }
  ],

  isSubmitDisabled: function () {
    return !this.get('stepConfigs').filterProperty('showConfig').everyProperty('errorCount', 0);
  }.property('stepConfigs.@each.errorCount'),

  /**
   * clear info of step
   */
  clearStep: function () {
    this.get('stepConfigs').clear();
    this.get('securityUsers').clear();
  },

  /**
   *  Function is called whenever the step is loaded
   */
  loadStep: function () {
    console.log("TRACE: Loading addSecurity step2: Configure Services");
    var versionNumber = App.get('currentStackVersionNumber');
    if( stringUtils.compareVersions(versionNumber, "2.2") >= 0){
      // Add Nimbus config options
      var masterComponentMap = this.get('masterComponentMap');
      masterComponentMap.filterProperty('configName', 'storm_host').components = ["SUPERVISOR", "STORM_UI_SERVER", "DRPC_SERVER", "STORM_REST_API"];
      masterComponentMap.pushObject({
        serviceName: 'STORM',
        configName: 'nimbus_host',
        components: ['NIMBUS']
      });
      this.get('hostToPrincipalMap').pushObject({
        serviceName: 'STORM',
        configName: 'nimbus_host',
        principalName: 'storm_principal_name',
        primaryName: 'storm'
      });
    }
    this.clearStep();
    this.loadUsers();
    this.addUserPrincipals(this.get('content.services'), this.get('securityUsers'));
    this.addMasterHostToConfigs();
    this.addHostPrincipals();
    this.addSlaveHostToConfigs();
    this.renderServiceConfigs(this.get('content.services'));
    this.changeCategoryOnHa(this.get('content.services'), this.get('stepConfigs'));
    this.setStoredConfigsValue(this.get('content.serviceConfigProperties'));
    this.set('installedServices', App.Service.find().mapProperty('serviceName'));
  },

  /**
   * set stored values to service configs
   * @param storedConfigProperties
   * @return {Boolean}
   */
  setStoredConfigsValue: function (storedConfigProperties) {
    if (!storedConfigProperties) return false;

    // for all services`
    this.get('stepConfigs').forEach(function (_content) {
      _content.get('configs').forEach(function (_config) {
        var configProperty = storedConfigProperties.findProperty('name', _config.get('name'));

        if (configProperty) {
          _config.set('value', configProperty.value);
        }
      }, this);
    }, this);
    return true;
  },

  /**
   * Render configs for active services
   * @param serviceConfigs
   */
  renderServiceConfigs: function (serviceConfigs) {
    serviceConfigs.forEach(function (_serviceConfig) {
      var serviceConfig = App.ServiceConfig.create({
        filename: _serviceConfig.filename,
        serviceName: _serviceConfig.serviceName,
        displayName: _serviceConfig.displayName,
        configCategories: _serviceConfig.configCategories,
        showConfig: true,
        configs: this.wrapConfigProperties(_serviceConfig)
      });

      this.get('stepConfigs').pushObject(serviceConfig);
    }, this);
    this.set('selectedService', this.get('stepConfigs').filterProperty('showConfig').objectAt(0));
  },

  /**
   * wrap configs into App.ServiceConfigProperty objects
   * @param _componentConfig
   */
  wrapConfigProperties: function (_componentConfig) {
    var configs = [];
    _componentConfig.configs.forEach(function (_serviceConfigProperty) {
      var serviceConfigProperty = App.ServiceConfigProperty.create(_serviceConfigProperty);
      serviceConfigProperty.set('isEditable', serviceConfigProperty.get('isReconfigurable'));
      serviceConfigProperty.validate();
      configs.pushObject(serviceConfigProperty);
    }, this);
    return configs;
  },

  /**
   * fill config with hosts of component
   * @param serviceName
   * @param configName
   * @param componentNames
   * @return {Boolean}
   */
  setHostsToConfig: function (serviceName, configName, componentNames) {
    var service = this.get('content.services').findProperty('serviceName', serviceName);
    if (service) {
      var hosts = service.configs.findProperty('name', configName);
      if (hosts) {
        hosts.defaultValue = App.Service.find(service.serviceName)
          .get('hostComponents')
          .filter(function (component) {
            return componentNames.contains(component.get('componentName'));
          })
          .mapProperty('hostName')
          .uniq();
        return true;
      }
      return false;
    }
    return false;
  },

  /**
   * set principal default value based on config host and default primary name
   * @param serviceName
   * @param hostConfigName
   * @param principalConfigName
   * @param defaultPrimaryName
   * @return {Boolean}
   */
  setHostToPrincipal: function (serviceName, hostConfigName, principalConfigName, defaultPrimaryName) {
    var service = this.get('content.services').findProperty('serviceName', serviceName);
    if (service) {
      var host = service.configs.findProperty('name', hostConfigName);
      var principal = service.configs.findProperty('name', principalConfigName);
      var versionNumber = App.get('currentStackVersionNumber');
      var special22ConfigsMap = {
        storm_principal_name: defaultPrimaryName,
        oozie_http_principal_name: defaultPrimaryName + '_HOST'
      };
      if (stringUtils.compareVersions(versionNumber, "2.2") >= 0 && special22ConfigsMap[principalConfigName]) {
        principal.defaultValue = special22ConfigsMap[principalConfigName];
        return true;
      }
      if (host && principal) {
        var host_defaultValue = Array.isArray(host.defaultValue) ? host.defaultValue[0] : host.defaultValue;
        principal.defaultValue = defaultPrimaryName + host_defaultValue;
        return true;
       }
      return false;
    }
    return false;
  },

  /**
   * load services users
   */
  loadUsers: function () {
    var securityUsers = App.router.get('mainAdminSecurityController').get('serviceUsers');
    if (Em.isNone(securityUsers) || securityUsers.length === 0) {
      if (App.get('testMode')) {
        securityUsers = securityUsers || [];
        securityUsers.pushObject({id: 'puppet var', name: 'hdfs_user', value: 'hdfs'});
        securityUsers.pushObject({id: 'puppet var', name: 'mapred_user', value: 'mapred'});
        securityUsers.pushObject({id: 'puppet var', name: 'hbase_user', value: 'hbase'});
        securityUsers.pushObject({id: 'puppet var', name: 'hive_user', value: 'hive'});
        securityUsers.pushObject({id: 'puppet var', name: 'smokeuser', value: 'ambari-qa'});
      } else {
        securityUsers = App.db.getSecureUserInfo();
      }
    }
    this.set('securityUsers', securityUsers);
  },

  /**
   * set default values to user principals and control their visibility
   * @param serviceConfigs
   * @param securityUsers
   */
  addUserPrincipals: function (serviceConfigs, securityUsers) {
    var generalService = serviceConfigs.findProperty('serviceName', 'GENERAL').configs;
    this.setUserPrincipalValue(securityUsers.findProperty('name', 'smokeuser'), generalService.findProperty('name', 'smokeuser_principal_name'));
    var servicesWithUserPrincipals = ['HDFS', 'HBASE'];

    servicesWithUserPrincipals.forEach(function (serviceName) {
      var isServiceInstalled = serviceConfigs.someProperty('serviceName', serviceName);
      var userPrincipal = generalService.findProperty('name', serviceName.toLowerCase() + '_principal_name');
      var userKeytab = generalService.findProperty('name', serviceName.toLowerCase() + '_user_keytab');
      var userName = securityUsers.findProperty('name', serviceName.toLowerCase() + '_user');
      if (isServiceInstalled && this.setUserPrincipalValue(userName, userPrincipal)) {
        userPrincipal.isVisible = true;
        userKeytab.isVisible = true;
      }
    }, this);
  },
  /**
   * set default value of user principal
   * @param user
   * @param userPrincipal
   */
  setUserPrincipalValue: function (user, userPrincipal) {
    if (user && userPrincipal) {
      userPrincipal.defaultValue = user.value;
      return true;
    }
    return false;
  },

  /**
   * put hosts of slave component into defaultValue of configs
   */
  addSlaveHostToConfigs: function () {
    this.get('slaveComponentMap').forEach(function (service) {
      this.setHostsToConfig(service.serviceName, service.configName, [service.component]);
    }, this);
  },

  /**
   * put hosts of master component into defaultValue of configs
   */
  addMasterHostToConfigs: function () {
    this.get('masterComponentMap').forEach(function (item) {
      this.setHostsToConfig(item.serviceName, item.configName, item.components);
    }, this);
  },

  /**
   * put hosts to principal default values
   */
  addHostPrincipals: function () {
    this.get('hostToPrincipalMap').forEach(function (item) {
      this.setHostToPrincipal(item.serviceName, item.configName, item.principalName, item.primaryName);
    }, this);
  },

  /**
   * modify config categories depending on whether HA is enabled or not
   * @param serviceConfigs
   * @param stepConfigs
   */
  changeCategoryOnHa: function (serviceConfigs, stepConfigs) {
    var hdfsService = serviceConfigs.findProperty('serviceName', 'HDFS');
    if (hdfsService) {
      var properties = stepConfigs.findProperty('serviceName', 'HDFS').get('configs');
      var configCategories = hdfsService.configCategories;
      if ((App.get('testMode') && App.get('testNameNodeHA')) || (this.get('content.isNnHa') === 'true')) {
        this.removeConfigCategory(properties, configCategories, 'SNameNode');
      } else {
        this.removeConfigCategory(properties, configCategories, 'JournalNode');
      }
      return true;
    }
    return false;
  },
  /**
   * remove config category that belong to component and hide category configs
   * @param properties
   * @param configCategories
   * @param component
   */
  removeConfigCategory: function (properties, configCategories, component) {
    properties.filterProperty('category', component).forEach(function (_snConfig) {
      _snConfig.set('isVisible', false);
    }, this);
    var category = configCategories.findProperty('name', component);
    if (category) {
      configCategories.removeObject(category);
    }
  }
});

});

require.register("controllers/main/admin/security/add/step3", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

App.MainAdminSecurityAddStep3Controller = Em.Controller.extend({
  name: 'mainAdminSecurityAddStep3Controller',
  hostComponents: [],
  hosts: [],
  isLoaded: false,

  componentToUserMap: function() {
    var map = {
      'NAMENODE': 'hdfs_user',
      'SECONDARY_NAMENODE': 'hdfs_user',
      'DATANODE': 'hdfs_user',
      'JOURNALNODE': 'hdfs_user',
      'TASKTRACKER': 'mapred_user',
      'JOBTRACKER': 'mapred_user',
      'HISTORYSERVER': 'mapred_user',
      'RESOURCEMANAGER': 'yarn_user',
      'NODEMANAGER': 'yarn_user',
      'ZOOKEEPER_SERVER': 'zk_user',
      'HIVE_SERVER': 'hive_user',
      'HIVE_METASTORE': 'hive_user',
      'OOZIE_SERVER': 'oozie_user',
      'HBASE_MASTER': 'hbase_user',
      'HBASE_REGIONSERVER': 'hbase_user',
      'SUPERVISOR': 'storm_user',
      'NIMBUS': 'storm_user',
      'STORM_UI_SERVER': 'storm_user',
      'FALCON_SERVER': 'falcon_user',
      'KNOX_GATEWAY': 'knox_user',
      'APP_TIMELINE_SERVER': 'yarn_user'
    };
    if (App.get('isHadoop22Stack')) {
      map['DRPC_SERVER'] = 'storm_user'
    }
    return map;
  }.property('App.isHadoop22Stack'),
  // The componentName, principal, and keytab have to coincide with the values in secure_properties.js
  componentToConfigMap: [
      {
        componentName: 'NAMENODE',
        principal: 'hadoop_http_principal_name',
        keytab: 'hadoop_http_keytab',
        displayName: Em.I18n.t('admin.addSecurity.hdfs.user.httpUser')
      },
      {
        componentName: 'SECONDARY_NAMENODE',
        principal: 'hadoop_http_principal_name',
        keytab: 'hadoop_http_keytab',
        displayName: Em.I18n.t('admin.addSecurity.hdfs.user.httpUser')
      },
      {
        componentName: 'JOURNALNODE',
        principal: 'hadoop_http_principal_name',
        keytab: 'hadoop_http_keytab',
        displayName: Em.I18n.t('admin.addSecurity.hdfs.user.httpUser')
      },
      {
        componentName: 'WEBHCAT_SERVER',
        principal: 'webHCat_http_principal_name',
        keytab: 'webhcat_http_keytab',
        displayName: Em.I18n.t('admin.addSecurity.webhcat.user.httpUser')
      },
      {
        componentName: 'HIVE_SERVER',
        principal: 'hive_metastore_http_principal_name',
        keytab: 'hive_metastore_http_keytab',
        displayName: Em.I18n.t('admin.addSecurity.hive.user.httpUser')
      },
      {
        componentName: 'OOZIE_SERVER',
        principal: 'oozie_http_principal_name',
        keytab: 'oozie_http_keytab',
        displayName: Em.I18n.t('admin.addSecurity.oozie.user.httpUser')
      },
      {
        componentName: 'FALCON_SERVER',
        principal: 'falcon_http_principal_name',
        keytab: 'falcon_http_keytab',
        displayName: Em.I18n.t('admin.addSecurity.falcon.user.httpUser')
      },
      {
        componentName: 'HISTORYSERVER',
        principal: 'jobhistory_http_principal_name',
        keytab: 'jobhistory_http_keytab',
        displayName: Em.I18n.t('admin.addSecurity.historyServer.user.httpUser')
      },
      {
        componentName: 'RESOURCEMANAGER',
        principal: 'resourcemanager_http_principal_name',
        keytab: 'resourcemanager_http_keytab',
        displayName: Em.I18n.t('admin.addSecurity.rm.user.httpUser')
      },
      {
        componentName: 'NODEMANAGER',
        principal: 'nodemanager_http_principal_name',
        keytab: 'nodemanager_http_keytab',
        displayName: Em.I18n.t('admin.addSecurity.nm.user.httpUser')
      },
      {
        componentName: 'APP_TIMELINE_SERVER',
        principal: 'apptimelineserver_http_principal_name',
        keytab: 'apptimelineserver_http_keytab',
        displayName: Em.I18n.t('admin.addSecurity.user.yarn.atsHTTPUser')
      },
      {
        componentName: 'STORM_UI_SERVER',
        principal: 'storm_ui_principal_name',
        keytab: 'storm_ui_keytab',
        displayName: Em.I18n.t('admin.addSecurity.storm.user.httpUser'),
        isHadoop22Stack: true
      }
  ],

  mandatoryConfigs: [
    {
      userConfig: 'smokeuser',
      keytab: 'smokeuser_keytab',
      displayName: Em.I18n.t('admin.addSecurity.user.smokeUser')
    },
    {
      userConfig: 'hdfs_user',
      keytab: 'hdfs_user_keytab',
      displayName: Em.I18n.t('admin.addSecurity.user.hdfsUser'),
      checkService: 'HDFS'
    },
    {
      userConfig: 'hbase_user',
      keytab: 'hbase_user_keytab',
      displayName: Em.I18n.t('admin.addSecurity.user.hbaseUser'),
      checkService: 'HBASE'
    }
  ],

  /**
   * download CSV file
   */
  doDownloadCsv: function () {
    if ($.browser.msie && $.browser.version < 10) {
      this.openInfoInNewTab();
    } else {
      try {
        var blob = new Blob([stringUtils.arrayToCSV(this.get('hostComponents'))], {type: "text/csv;charset=utf-8;"});
        saveAs(blob, "host-principal-keytab-list.csv");
      } catch (e) {
        this.openInfoInNewTab();
      }
    }
  },

  /**
   * open content of CSV file in new window
   */
  openInfoInNewTab: function () {
    var newWindow = window.open('');
    var newDocument = newWindow.document;
    newDocument.write(stringUtils.arrayToCSV(this.get('hostComponents')));
    newWindow.focus();
  },

  /**
   * load hosts from server
   */
  loadHosts: function () {
    App.ajax.send({
      name: 'hosts.security.wizard',
      sender: this,
      data: {},
      error: 'loadHostsErrorCallback',
      success: 'loadHostsSuccessCallback'
    })
  },

  loadHostsSuccessCallback: function (data, opt, params) {
    var hosts = [];

    data.items.forEach(function (item) {
      var hostComponents = [];

      item.host_components.forEach(function (hostComponent) {
        hostComponents.push(Em.Object.create({
          componentName: hostComponent.HostRoles.component_name,
          service: Em.Object.create({
            serviceName: hostComponent.HostRoles.service_name
          }),
          displayName: App.format.role(hostComponent.HostRoles.component_name)
        }));
      });
      hosts.push(Em.Object.create({
        hostName: item.Hosts.host_name,
        hostComponents: hostComponents
      }));
    });
    this.set('isLoaded', true);
    this.set('hosts', hosts);
    this.loadStep();
  },
  loadHostsErrorCallback: function () {
    this.set('isLoaded', true);
    this.set('hosts', []);
    this.loadStep();
  },

  /**
   * load step info
   */
  loadStep: function () {
    var hosts = this.get('hosts');
    var result = [];
    var securityUsers = this.getSecurityUsers();
    var hadoopGroupId = securityUsers.findProperty('name', 'user_group').value;
    var addedPrincipalsHost = {}; //Keys = host_principal, Value = 'true'

    hosts.forEach(function (host) {
      this.setMandatoryConfigs(result, securityUsers, host.get('hostName'), hadoopGroupId);
      this.setComponentsConfig(result, host, hadoopGroupId);
      this.setHostComponentsSecureValue(result, host, addedPrincipalsHost, securityUsers, hadoopGroupId);
    }, this);
    this.set('hostComponents', result);
  },

  /**
   * Returns host name for Nimbus component
   */
  getNimbusHostName: function () {
    var host = this.get('hosts').find(function (host) {
      return !!host.get('hostComponents').findProperty('componentName', 'NIMBUS');
    });
    if (host) {
      return host.get('hostName');
    }
  },

  /**
   * build map of connections between component and user
   * @param securityUsers
   */
  buildComponentToOwnerMap: function (securityUsers) {
    var componentToUserMap = this.get('componentToUserMap');
    var componentToOwnerMap = {};
    for (var component in componentToUserMap) {
      var user = componentToUserMap[component];
      var securityUser = securityUsers.findProperty('name', user);
      componentToOwnerMap[component] = securityUser.value;
    }
    return componentToOwnerMap;
  },

  /**
   * set security settings(principal and keytab) to component depending on whether host has such component
   * @param result
   * @param host
   * @param hadoopGroupId
   */
  setComponentsConfig: function (result, host, hadoopGroupId) {
    var hostComponents = host.get('hostComponents');

    var isATSInstalled = this.get('content.isATSInstalled');
    var doesATSSupportKerberos = App.get("doesATSSupportKerberos");

    this.get('componentToConfigMap').forEach(function (component) {
      //add specific components that supported only in Hadoop2 stack
      if (component.isHadoop22Stack && !App.get('isHadoop22Stack')) return;

      if (hostComponents.someProperty('componentName', component.componentName)) {

        if (component.componentName === "APP_TIMELINE_SERVER" && (!isATSInstalled || !doesATSSupportKerberos)) {
          return;
        }

        var configs = this.get('content.serviceConfigProperties');
        var serviceName = App.StackServiceComponent.find(component.componentName).get('serviceName');
        var serviceConfigs = configs.filterProperty('serviceName', serviceName);
        var servicePrincipal = serviceConfigs.findProperty('name', component.principal);
        var serviceKeytabPath = serviceConfigs.findProperty('name', component.keytab).value;
        result.push({
          host: host.get('hostName'),
          component: component.displayName,
          principal: this.getPrincipal(servicePrincipal, host.get('hostName')),
          keytabfile: stringUtils.getFileFromPath(serviceKeytabPath),
          keytab: stringUtils.getPath(serviceKeytabPath),
          owner: 'root',
          group: hadoopGroupId,
          acl: '440'
        });
      }
    }, this);
  },

  /**
   * set security settings(principal and keytab) to component
   * if checkService is passed then verify that service to his existence in order to set configs to such service
   * @param result
   * @param securityUsers
   * @param hostName
   * @param hadoopGroupId
   */
  setMandatoryConfigs: function (result, securityUsers, hostName, hadoopGroupId) {
    var generalConfigs = this.get('content.serviceConfigProperties').filterProperty('serviceName', 'GENERAL');
    var realm = generalConfigs.findProperty('name', 'kerberos_domain').value;
    var installedServices = App.Service.find().mapProperty('serviceName');

    this.get('mandatoryConfigs').forEach(function (config) {
      if (config.checkService && !installedServices.contains(config.checkService)) return;

      var userId = securityUsers.findProperty('name', config.userConfig).value;
      var userKeytabPath = generalConfigs.findProperty('name', config.keytab).value;
      result.push({
        host: hostName,
        component: config.displayName,
        principal: userId + '@' + realm,
        keytabFile: stringUtils.getFileFromPath(userKeytabPath),
        keytab: stringUtils.getPath(userKeytabPath),
        owner: userId,
        group: hadoopGroupId,
        acl: '440'
      });
    }, this);
  },

  /**
   * set secure properties(keytab and principal) for components, which should be displayed
   * @param result
   * @param host
   * @param addedPrincipalsHost
   * @param securityUsers
   * @param hadoopGroupId
   */
  setHostComponentsSecureValue: function (result, host, addedPrincipalsHost, securityUsers, hadoopGroupId) {
    var componentsToDisplay = ['NAMENODE', 'SECONDARY_NAMENODE', 'DATANODE', 'JOBTRACKER', 'ZOOKEEPER_SERVER', 'HIVE_SERVER', 'HIVE_METASTORE',
      'TASKTRACKER', 'OOZIE_SERVER', 'HBASE_MASTER', 'HBASE_REGIONSERVER', 'HISTORYSERVER', 'RESOURCEMANAGER', 'NODEMANAGER',
      'JOURNALNODE', 'SUPERVISOR', 'NIMBUS', 'STORM_UI_SERVER','FALCON_SERVER', 'KNOX_GATEWAY', 'APP_TIMELINE_SERVER'];
    if (App.get('isHadoop22Stack')) {
      componentsToDisplay.push('DRPC_SERVER');
    }
    var configs = this.get('content.serviceConfigProperties');
    var componentToOwnerMap = this.buildComponentToOwnerMap(securityUsers);
    var hostName = host.get('hostName');

    var isATSInstalled = this.get('content.isATSInstalled');
    var doesATSSupportKerberos = App.get("doesATSSupportKerberos");

    host.get('hostComponents').forEach(function (hostComponent) {
      if (componentsToDisplay.contains(hostComponent.get('componentName'))) {
        var serviceConfigs = configs.filterProperty('serviceName', hostComponent.get('service.serviceName'));
        var targetHost = hostName;
        if (App.get('isHadoop22Stack') && hostComponent.get('componentName') === 'DRPC_SERVER') {
          targetHost = this.getNimbusHostName()
        }
        var secureProperties = this.getSecureProperties(serviceConfigs, hostComponent.get('componentName'), targetHost);
        var displayName = this.changeDisplayName(hostComponent.get('displayName'));
        var key = hostName + "--" + secureProperties.principal;

        if (hostComponent.get('componentName') === "APP_TIMELINE_SERVER" && (!isATSInstalled || !doesATSSupportKerberos)) {
          return;
        }

        if (Em.isNone(addedPrincipalsHost[key])) {
          var owner = componentToOwnerMap[hostComponent.get('componentName')] || '';

          result.push({
            host: hostName,
            component: displayName,
            principal: secureProperties.principal,
            keytabFile: stringUtils.getFileFromPath(secureProperties.keytab),
            keytab: stringUtils.getPath(secureProperties.keytab),
            owner: owner,
            group: hadoopGroupId,
            acl: '400'
          });
          addedPrincipalsHost[key] = true;
        }
      }
    }, this);
  },

  /**
   * get properties (keytab and principle) of secure config that match component
   * @param serviceConfigs
   * @param componentName
   * @param hostName
   * @return {Object}
   */
  getSecureProperties: function (serviceConfigs, componentName, hostName) {
    var secureProperties = {};
    serviceConfigs.forEach(function (config) {
      if ((config.component && config.component === componentName) ||
        (config.components && config.components.contains(componentName))) {
        if (config.name.endsWith('_principal_name')) {
          secureProperties.principal = this.getPrincipal(config, hostName);
        } else if (config.name.endsWith('_keytab') || config.name.endsWith('_keytab_path')) {
          secureProperties.keytab = config.value;
        }
      }
    }, this);
    return secureProperties;
  },

  /**
   * get formatted principal value
   * @param config
   * @param hostName
   * @return {String}
   */
  getPrincipal: function (config, hostName) {
    return config.value.replace('_HOST', hostName.toLowerCase()) + config.unit;
  },

  /**
   * get users from security configs
   * @return {Array}
   */
  getSecurityUsers: function () {
    return App.db.getSecureUserInfo();
  },

  /**
   * format  display names of specific components
   * @param name
   * @return {*}
   */
  changeDisplayName: function (name) {
    if (name === 'HiveServer2' || name === 'Hive Metastore') {
      return 'Hive Metastore and HiveServer2';
    } else {
      return name;
    }
  }
});

});

require.register("controllers/main/admin/security/add/step4", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminSecurityAddStep4Controller = App.MainAdminSecurityProgressController.extend(App.AddSecurityConfigs, {

  name: 'mainAdminSecurityAddStep4Controller',

  isBackBtnDisabled: function () {
    return !this.get('commands').someProperty('isError');
  }.property('commands.@each.isCompleted'),

  isSecurityApplied: function () {
    return this.get('commands').someProperty('name', 'START_SERVICES') && this.get('commands').findProperty('name', 'START_SERVICES').get('isSuccess');
  }.property('commands.@each.isCompleted'),

  /**
   * control disabled property of completion button
   */
  enableSubmit: function () {
    var addSecurityController = App.router.get('addSecurityController');
    if (this.get('commands').someProperty('isError') || this.get('commands').everyProperty('isSuccess')) {
      this.set('isSubmitDisabled', false);
      if (this.get('commands').someProperty('isError')) {
        addSecurityController.setStepsEnable();
      }
    } else {
      this.set('isSubmitDisabled', true);
      addSecurityController.setLowerStepsDisable(4);
    }
  }.observes('commands.@each.isCompleted'),

  /**
   * clear step info
   */
  clearStep: function () {
    this.set('commands', []);
    this.set('isSubmitDisabled', true);
    this.get('serviceConfigTags').clear();
  },

  loadCommands: function () {
    this._super();

    // Determine if ATS Component needs to be removed
    var isATSInstalled = this.get('content.isATSInstalled');
    var doesATSSupportKerberos = App.get("doesATSSupportKerberos");
    if (isATSInstalled && !doesATSSupportKerberos) {
      this.get('commands').splice(2, 0, App.Poll.create({name: 'DELETE_ATS', label: Em.I18n.translations['admin.addSecurity.apply.delete.ats'], isPolling: false}));
    }
  },

  /**
   * load step info
   */
  loadStep: function () {
    this.clearStep();
    this.prepareSecureConfigs();

    if (!this.resumeSavedCommands()) {
      this.loadCommands();
      this.addInfoToCommands();
      this.syncStopServicesOperation();
      this.addObserverToCommands();
      this.moveToNextCommand();
    }
  },

  /**
   * synchronize "STOP_SERVICES" operation from BO with command of step
   * @return {Boolean}
   */
  syncStopServicesOperation: function () {
    var runningOperations = App.router.get('backgroundOperationsController.services').filterProperty('isRunning');
    var stopAllOperation = runningOperations.findProperty('name', 'Stop All Services');
    var stopCommand = this.get('commands').findProperty('name', 'STOP_SERVICES');
    if (stopCommand && stopAllOperation) {
      stopCommand.set('requestId', stopAllOperation.get('id'));
      return true;
    }
    return false;
  },

  /**
   * resume previously saved commands
   * @return {Boolean}
   */
  resumeSavedCommands: function () {
    var commands = App.db.getSecurityDeployCommands();
    if (Em.isNone(commands) || commands.length === 0) return false;

    commands.forEach(function (_command, index) {
      commands[index] = App.Poll.create(_command);
    }, this);
    if (commands.someProperty('isError')) {
      this.get('commands').pushObjects(commands);
      this.addObserverToCommands();
      return true;
    } else if (commands.filterProperty('isStarted').someProperty('isCompleted', false)) {
      var runningCommand = commands.filterProperty('isStarted').findProperty('isCompleted', false);
      runningCommand.set('isStarted', false);
      this.get('commands').pushObjects(commands);
    } else {
      this.get('commands').pushObjects(commands);
    }
    this.addObserverToCommands();
    this.moveToNextCommand();
    return true;
  },

  manageSecureConfigs: function () {
    var serviceConfigTags = this.get('serviceConfigTags');
    var secureConfigs = this.get('secureConfigs');
    var siteProperties = this.get('configs');
    if (serviceConfigTags) {
      secureConfigs.forEach(function (config) {
        this.setPrincipalValue(config.serviceName, config.name);
      }, this);
      serviceConfigTags.forEach(function (_serviceConfigTags) {
        _serviceConfigTags.newTagName = 'version' + (new Date).getTime();
        siteProperties.filterProperty('filename', _serviceConfigTags.siteName + '.xml').forEach(function (_config) {
          if (!/_hosts?$/.test(_config.name)) {
            _serviceConfigTags.configs[_config.name] = _config.value;
          }
        }, this);
      }, this);
      return true;
    } else {
      var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
      command.set('isSuccess', false);
      command.set('isError', true);
      this.onJsError();
      return false;
    }
  },

  /**
   * send DELETE command to server to delete component
   * @param componentName
   * @param hostName
   */
  deleteComponents: function(componentName, hostName) {
    App.ajax.send({
      name: 'common.delete.host_component',
      sender: this,
      data: {
        componentName: componentName,
        hostName: hostName
      },
      success: 'onDeleteComplete',
      error: 'onDeleteError'
    });
  },

  /**
   * callback on successful deletion of component
   */
  onDeleteComplete: function () {
    var deleteAtsCommand = this.get('commands').findProperty('name', 'DELETE_ATS');
    console.warn('APP_TIMELINE_SERVER doesn\'t support security mode in this HDP stack. It has been removed from YARN service ');
    deleteAtsCommand.set('isError', false);
    deleteAtsCommand.set('isSuccess', true);
  },

  /**
   * callback on failed deletion of component
   */
  onDeleteError: function () {
    var deleteAtsCommand = this.get('commands').findProperty('name', 'DELETE_ATS');
    console.warn('Error: Can\'t delete APP_TIMELINE_SERVER');
    deleteAtsCommand.set('isError', true);
    deleteAtsCommand.set('isSuccess', false);
  },

  /**
   * show popup when js error occurred
   */
  onJsError: function () {
    App.ModalPopup.show({
      header: Em.I18n.t('common.error'),
      secondary: false,
      bodyClass: Ember.View.extend({
        template: Ember.Handlebars.compile('<p>{{t admin.security.apply.configuration.error}}</p>')
      })
    });
  }
});

});

require.register("controllers/main/admin/security/disable", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
require('controllers/main/admin/security/security_progress_controller');

App.MainAdminSecurityDisableController = App.MainAdminSecurityProgressController.extend({
  name: 'mainAdminSecurityDisableController',
  secureServices: [],
  /**
   * values of site configs when security disabled.
   * Properties not defined in data/secure_mapping or data/HDP2/secure_mapping and needs to be changed on disabling
   * security should be defined in secureConfigValuesMap Object
   */
  secureConfigValuesMap: {
    'nimbus.childopts': function(value) {
     return value.replace (/-Djava.security.auth.login.config\s*=\s*\S*/g, "");
    },
    'ui.childopts': function(value) {
       return value.replace (/-Djava.security.auth.login.config\s*=\s*\S*/g, "");
    },
    'supervisor.childopts': function(value) {
      return value.replace (/-Djava.security.auth.login.config\s*=\s*\S*/g, "");
    }
  },

  isSubmitDisabled: function () {
    return !(this.get('commands').someProperty('isError') || this.get('commands').everyProperty('isSuccess'));
  }.property('commands.@each.isCompleted'),

  /**
   * clear step info
   */
  clearStep: function () {
    this.get('commands').clear();
    this.get('secureServices').clear();
    this.get('serviceConfigTags').clear();
  },

  /**
   * load info required by current step
   */
  loadStep: function () {
    this.clearStep();
    var commands = App.db.getSecurityDeployCommands();
    if (commands && commands.length > 0) {
      commands.forEach(function (_command, index) {
        commands[index] = App.Poll.create(_command);
      }, this);
      if (commands.someProperty('isError', true)) {
        this.get('commands').pushObjects(commands);
        this.loadSecureServices();
        this.addObserverToCommands();
        return;
      } else if (commands.filterProperty('isStarted', true).someProperty('isCompleted', false)) {
        var runningCommand = commands.filterProperty('isStarted', true).findProperty('isCompleted', false);
        runningCommand.set('isStarted', false);
        this.get('commands').pushObjects(commands);
      } else {
        this.get('commands').pushObjects(commands);
      }
    } else {
      this.loadCommands();
      this.addInfoToCommands();
      this.syncStopServicesCommand();
    }
    this.loadSecureServices();
    this.addObserverToCommands();
    this.moveToNextCommand();
  },

  /**
   * resume info about commands from local storage
   * @return {Boolean}
   */
  resumeCommands: function () {
    var commands = App.db.getSecurityDeployCommands();
    if (!commands || commands.length === 0) return false;

    commands.forEach(function (_command) {
      this.get('commands').pushObject(App.Poll.create(_command));
    }, this);
    var runningCommand = this.get('commands').filterProperty('isStarted').findProperty('isCompleted', false);
    if (runningCommand) {
      runningCommand.set('isStarted', false);
    }
    return true;
  },

  /**
   * synchronize existing background operation "Stop All Services" with command in Security wizard
   */
  syncStopServicesCommand: function () {
    var runningOperations = App.router.get('backgroundOperationsController.services').filterProperty('isRunning');
    var stopAllOperation = runningOperations.findProperty('name', 'Stop All Services');
    var stopCommand = this.get('commands').findProperty('name', 'STOP_SERVICES');
    if (stopCommand && stopAllOperation) {
      stopCommand.set('requestId', stopAllOperation.get('id'));
    }
  },

  /**
   * load secure configs of installed services
   */
  loadSecureServices: function () {
    var secureServices = require('data/HDP2/secure_configs');
    var installedServices = App.Service.find().mapProperty('serviceName');
    this.get('secureServices').pushObject(secureServices.findProperty('serviceName', 'GENERAL'));
    //General (only non service tab) tab is always displayed
    installedServices.forEach(function (_service) {
      var secureService = secureServices.findProperty('serviceName', _service);
      if (secureService) {
        this.get('secureServices').pushObject(secureService);
      }
    }, this);
  },

  /**
   * manage configurations from serviceConfigTags
   * @return {Boolean}
   */
  manageSecureConfigs: function () {
    var serviceConfigTags = this.get('serviceConfigTags');
    var secureProperties = this.get('secureProperties');
    var secureMapping = this.get('secureMapping');
    if (!serviceConfigTags || !secureProperties || !secureMapping) {
      var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
      command.set('isSuccess', false);
      command.set('isError', true);
      return false;
    } else {
      serviceConfigTags.forEach(function (_serviceConfigTags) {
        _serviceConfigTags.newTagName = 'version' + (new Date).getTime();
        if (_serviceConfigTags.siteName.contains('-env')) {
          this.deleteDisabledConfigs(secureProperties, _serviceConfigTags);
          if (_serviceConfigTags.siteName === 'cluster-env') {
            _serviceConfigTags.configs.security_enabled = 'false';
          }
        } else {
          this.modifySiteConfigs(secureMapping, _serviceConfigTags);
        }
      }, this);
      return true;
    }
  },
  /**
   * delete configs, which aren't required when security disabled
   * @param secureProperties
   * @param _serviceConfigTags
   * @return {Boolean}
   */
  deleteDisabledConfigs: function (secureProperties, _serviceConfigTags) {
    if (!secureProperties || !_serviceConfigTags) return false;
    secureProperties.forEach(function (_config) {
      if (_config.name in _serviceConfigTags.configs) {
        delete _serviceConfigTags.configs[_config.name];
      }
    }, this);
    return true;
  },
  /**
   * delete unnecessary site configs and
   * change config values
   * @param secureMapping
   * @param _serviceConfigTags
   * @return {Boolean}
   */
  modifySiteConfigs: function (secureMapping, _serviceConfigTags) {
    var secureConfigValuesMap = this.get('secureConfigValuesMap');
    if (!secureMapping || !_serviceConfigTags) return false;

    // iterate over secureConfigValuesMap to update service-site configProperties not present in secureMapping metadata
    for (var key in secureConfigValuesMap) {
      if (key in _serviceConfigTags.configs) {
        var value = secureConfigValuesMap[key];
        if (typeof value == 'function') {
          _serviceConfigTags.configs[key] = value(_serviceConfigTags.configs[key]);
        }  else if (value) {
          _serviceConfigTags.configs[key] = value;
        }
      }
    }

    secureMapping.filterProperty('filename', _serviceConfigTags.siteName + '.xml').forEach(function (_config) {
      var configName = _config.name;
      var nonSecureConfigValue = _config.nonSecureValue;
      if (configName in _serviceConfigTags.configs) {
        if (nonSecureConfigValue) {
          _serviceConfigTags.configs[configName] = nonSecureConfigValue;
        } else {
          delete _serviceConfigTags.configs[configName]
        }
      }
    }, this);
    return true;
  }
});

});

require.register("controllers/main/admin/security/security_progress_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminSecurityProgressController = Em.Controller.extend({
  name: 'mainAdminSecurityProgressController',

  commands: [],
  configs: [],
  serviceConfigTags: [],
  totalSteps: 3,
  isSubmitDisabled: true,
  hasHostPopup: true,
  services: [],
  serviceTimestamp: null,
  operationsInfo: [
    {
      name: 'STOP_SERVICES',
      realUrl: '/services',
      testUrl: '/data/wizard/deploy/2_hosts/poll_1.json',
      data: '{"RequestInfo": {"context" :"' + Em.I18n.t('requestInfo.stopAllServices') + '"}, "Body": {"ServiceInfo": {"state": "INSTALLED"}}}'
    },
    {
      name: 'START_SERVICES',
      realUrl: '/services?params/run_smoke_test=true',
      testUrl: '/data/wizard/deploy/2_hosts/poll_1.json',
      data: '{"RequestInfo": {"context": "' + Em.I18n.t('requestInfo.startAllServices') + '"}, "Body": {"ServiceInfo": {"state": "STARTED"}}}'
    }
  ],

  secureMapping: require('data/HDP2/secure_mapping'),

  secureProperties: require('data/HDP2/secure_properties').configProperties,

  /**
   * prepare and restart failed command
   */
  retry: function () {
    var failedCommand = this.get('commands').findProperty('isError');
    if (failedCommand) {
      failedCommand.set('requestId', null);
      failedCommand.set('isStarted', false);
      failedCommand.set('isError', false);
      this.startCommand(failedCommand);
    }
  },

  /**
   * start updating current task in parallel
   * @param requestId
   * @param taskId
   * @return {Boolean}
   */
  startUpdatingTask: function (requestId, taskId) {
    if (!requestId || !taskId) return false;
    var command = this.get('commands').findProperty('requestId', requestId);
    command.updateTaskLog(taskId);
    return true;
  },

  /**
   * stop updating current task
   * @param requestId
   * @return {Boolean}
   */
  stopUpdatingTask: function (requestId) {
    if (!requestId) return false;
    var command = this.get('commands').findProperty('requestId', requestId);
    command.set('currentTaskId', null);
    return true;
  },

  /**
   * update info about progress of operation of commands
   */
  updateServices: function () {
    this.services.clear();
    var services = this.get("services");
    this.get("commands").forEach(function (command) {
      var polledData = command.get('polledData');
      var newService = Ember.Object.create({
        name: command.get('label'),
        hosts: []
      });
      if (polledData) {
        var hostNames = polledData.mapProperty('Tasks.host_name').uniq();
        hostNames.forEach(function (name) {
          newService.hosts.push({
            name: name,
            publicName: name,
            logTasks: polledData.filterProperty("Tasks.host_name", name)
          });
        });
        services.push(newService);
      }
    });
    this.set('serviceTimestamp', App.dateTime());
  }.observes('commands.@each.polledData'),
  /**
   * initialize default commands
   */
  loadCommands: function () {
    this.get('commands').pushObjects([
      App.Poll.create({name: 'STOP_SERVICES', label: Em.I18n.translations['admin.addSecurity.apply.stop.services'], isPolling: true }),
      App.Poll.create({name: 'APPLY_CONFIGURATIONS', label: Em.I18n.translations['admin.addSecurity.apply.save.config'], isPolling: false }),
      App.Poll.create({name: 'START_SERVICES', label: Em.I18n.translations['admin.addSecurity.apply.start.services'], isPolling: true })
    ]);
  },

  addObserverToCommands: function () {
    this.setIndex(this.get('commands'));
    this.addObserver('commands.@each.isSuccess', this, 'onCompleteCommand');
  },
  /**
   * set index to each command
   * @param commandArray
   */
  setIndex: function (commandArray) {
    commandArray.forEach(function (command, index) {
      command.set('index', index + 1);
    }, this);
    this.set('totalSteps', commandArray.length);
  },

  startCommand: function (command) {
    if (this.get('commands').length === this.get('totalSteps')) {
      if (!command) {
        var startedCommand = this.get('commands').filterProperty('isStarted', true);
        command = startedCommand.findProperty('isCompleted', false);
      }
      if (command) {
        if (command.get('isPolling')) {
          command.set('isStarted', true);
          command.start();
        } else if (command.get('name') === 'APPLY_CONFIGURATIONS') {
          command.set('isStarted', true);
          if (App.get('testMode')) {
            command.set('isError', false);
            command.set('isSuccess', true);
          } else {
            this.loadClusterConfigs();
          }
        } else if (command.get('name') === 'DELETE_ATS') {
          command.set('isStarted', true);
          if (App.get('testMode')) {
            command.set('isError', false);
            command.set('isSuccess', true);
          } else {
            var timeLineServer = App.HostComponent.find().findProperty('componentName', 'APP_TIMELINE_SERVER');
            if (timeLineServer) {
              this.deleteComponents('APP_TIMELINE_SERVER', timeLineServer.get('hostName'));
            } else {
              this.onDeleteComplete();
            }
          }
        }
        return true;
      }
    }
    return false;
  },
  /**
   * on command completion move to next command
   * @return {Boolean}
   */
  onCompleteCommand: function () {
    if (this.get('commands').length === this.get('totalSteps')) {
      var index = this.get('commands').filterProperty('isSuccess', true).length;
      if (index > 0) {
        var lastCompletedCommandResult = this.get('commands').objectAt(index - 1).get('isSuccess');
        if (lastCompletedCommandResult) {
          var nextCommand = this.get('commands').objectAt(index);
          this.moveToNextCommand(nextCommand);
          return true;
        }
      }
    }
    return false;
  },
  /**
   * move to next command
   * @param nextCommand
   */
  moveToNextCommand: function (nextCommand) {
    nextCommand = nextCommand || this.get('commands').findProperty('isStarted', false);
    if (nextCommand) {
      this.startCommand(nextCommand);
      return true;
    }
    return false;
  },

  /**
   * add query information(url, data) to commands
   */
  addInfoToCommands: function () {
    var operationsInfo = this.get('operationsInfo');
    var urlPrefix = App.apiPrefix + '/clusters/' + App.get('clusterName');
    operationsInfo.forEach(function (operation) {
      var command = this.get('commands').findProperty('name', operation.name);
      var url = (App.get('testMode')) ? operation.testUrl : urlPrefix + operation.realUrl;
      command.set('url', url);
      command.set('data', operation.data);
    }, this);
  },

  loadClusterConfigs: function () {
    App.ajax.send({
      name: 'admin.security.add.cluster_configs',
      sender: this,
      success: 'loadClusterConfigsSuccessCallback',
      error: 'loadClusterConfigsErrorCallback'
    });
  },

  loadClusterConfigsSuccessCallback: function (data) {
    //prepare tags to fetch all configuration for a service
    this.get('secureServices').forEach(function (_secureService) {
      this.setServiceTagNames(_secureService, data.Clusters.desired_configs);
    }, this);
    this.getAllConfigurations();
  },

  loadClusterConfigsErrorCallback: function (request, ajaxOptions, error) {
    var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
    command.set('isSuccess', false);
    command.set('isError', true);
    console.log("TRACE: error code status is: " + request.status);
  },

  /**
   * set tag names according to installed services and desired configs
   * @param secureService
   * @param configs
   * @return {Object}
   */
  setServiceTagNames: function (secureService, configs) {
    //var serviceConfigTags = this.get('serviceConfigTags');
    for (var index in configs) {
      if (secureService.sites && secureService.sites.contains(index)) {
        var serviceConfigObj = {
          siteName: index,
          tagName: configs[index].tag,
          newTagName: null,
          configs: {}
        };
        this.get('serviceConfigTags').pushObject(serviceConfigObj);
      }
    }
    return serviceConfigObj;
  },

  /**
   * form query data and apply security configurations to server
   */
  applyConfigurationsToCluster: function () {
    var configData = this.get('serviceConfigTags').map(function (_serviceConfig) {
      var res = {
        type: _serviceConfig.siteName,
        tag: _serviceConfig.newTagName,
        properties: _serviceConfig.configs,
        service_config_version_note: Em.I18n.t('admin.security.step4.save.configuration.note')
      };
      if (_serviceConfig.properties_attributes) {
        res['properties_attributes'] = _serviceConfig.properties_attributes
      }
      return res;
    }, this);

    var selectedServices = this.get('secureServices');
    var allConfigData = [];
    selectedServices.forEach(function (service) {
      var stackService = App.StackService.find(service.serviceName);
      if (stackService) {
        var serviceConfigData = [];
        Object.keys(stackService.get('configTypesRendered')).forEach(function (type) {
          var serviceConfigTag = configData.findProperty('type', type);
          if (serviceConfigTag) {
            serviceConfigData.pushObject(serviceConfigTag);
          }
        }, this);
        allConfigData.pushObject(JSON.stringify({
          Clusters: {
            desired_config: serviceConfigData
          }
        }));
      }
    }, this);

    var clusterConfig = configData.findProperty('type', 'cluster-env');
    if (clusterConfig) {
      allConfigData.pushObject(JSON.stringify({
        Clusters: {
          desired_config: [clusterConfig]
        }
      }));
    }
    App.ajax.send({
      name: 'common.across.services.configurations',
      sender: this,
      data: {
        data: '[' + allConfigData.toString() + ']'
      },
      success: 'applyConfigurationToClusterSuccessCallback',
      error: 'applyConfigurationToClusterErrorCallback'
    });
  },

  applyConfigurationToClusterSuccessCallback: function (data) {
    var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
    command.set('isSuccess', true);
    command.set('isError', false);
  },

  applyConfigurationToClusterErrorCallback: function (request, ajaxOptions, error) {
    var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
    command.set('isSuccess', false);
    command.set('isError', true);
  },

  /**
   * gets site config properties from server and sets it for every configuration
   */
  getAllConfigurations: function () {
    var urlParams = [];
    this.get('serviceConfigTags').forEach(function (_tag) {
      urlParams.push('(type=' + _tag.siteName + '&tag=' + _tag.tagName + ')');
    }, this);
    if (urlParams.length > 0) {
      App.ajax.send({
        name: 'admin.get.all_configurations',
        sender: this,
        data: {
          urlParams: urlParams.join('|')
        },
        success: 'getAllConfigurationsSuccessCallback',
        error: 'getAllConfigurationsErrorCallback'
      });
    }
  },

  getAllConfigurationsSuccessCallback: function (data) {
    console.log("TRACE: In success function for the GET getServiceConfigsFromServer call");
    var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
    this.get('serviceConfigTags').forEach(function (_tag) {
      if (!data.items.someProperty('type', _tag.siteName)) {
        console.log("Error: Metadata for secure services (secure_configs.js) is having config tags that are not being retrieved from server");
        command.set('isSuccess', false);
        command.set('isError', true);
      }
      var cfg = data.items.findProperty('type', _tag.siteName);
      _tag.configs = this.modifyConfigsForSecure(_tag.siteName, cfg);
      if (cfg.properties_attributes) {
        _tag.properties_attributes = cfg.properties_attributes;
      }
    }, this);
    if (this.manageSecureConfigs()) {
      this.applyConfigurationsToCluster();
    }
  },

  propertiesToUpdate: [
    {
      siteName: 'storm-site',
      name: 'ui.childopts',
      append: ' -Djava.security.auth.login.config=/etc/storm/conf/storm_jaas.conf'
    },
    {
      siteName: 'storm-site',
      name: 'supervisor.childopts',
      append: ' -Djava.security.auth.login.config=/etc/storm/conf/storm_jaas.conf'
    },
    {
      siteName: 'storm-site',
      name: 'nimbus.childopts',
      append: ' -Djava.security.auth.login.config=/etc/storm/conf/storm_jaas.conf'
    }
  ],

  /**
   * updates some configs for correct work in secure mode
   * @method modifyConfigsForSecure
   * @param {String} siteName
   * @param {Object} cfg
   * {
   *   properties: {
   *    'ui.childopts': 'value1'
   *    'property2': 'value2'
   *   }
   * };
   * has other properties but required filed is "properties";
   * @returns {Object}
   *   properties: {
   *    'ui.childopts': 'value1 -Djava.security.auth.login.config=/etc/storm/conf/storm_jaas.conf'
   *    'property2': 'value2'
   *   }
   */
  modifyConfigsForSecure: function(siteName, cfg) {
    var propertiesToUpdate = this.get('propertiesToUpdate').filterProperty('siteName', siteName);
    if (propertiesToUpdate.length) {
      propertiesToUpdate.forEach(function(p) {
        cfg.properties[p.name] += p.append;
      }, this);
    }
    return cfg.properties
  },

  getAllConfigurationsErrorCallback: function (request, ajaxOptions, error) {
    var command = this.get('commands').findProperty('name', 'APPLY_CONFIGURATIONS');
    command.set('isSuccess', false);
    command.set('isError', true);
    console.log("TRACE: In error function for the getServiceConfigsFromServer call");
    console.log("TRACE: error code status is: " + request.status);
  },

  /**
   * save commands to server and local storage
   */
  saveCommands: function () {
    var commands = [];
    if (this.get('commands').length === this.get('totalSteps')) {
      this.get('commands').forEach(function (_command) {
        var command = {
          name: _command.get('name'),
          label: _command.get('label'),
          isPolling: _command.get('isPolling'),
          isVisible: _command.get('isVisible'),
          isStarted: _command.get('isStarted'),
          requestId: _command.get('requestId'),
          isSuccess: _command.get('isSuccess'),
          isError: _command.get('isError'),
          url: _command.get('url'),
          polledData: _command.get('polledData'),
          data: _command.get('data')
        };
        commands.pushObject(command);
      }, this);
      App.db.setSecurityDeployCommands(commands);
      if (!App.get('testMode')) {
        App.clusterStatus.setClusterStatus({
          clusterName: this.get('clusterName'),
          clusterState: 'ADD_SECURITY_STEP_4',
          wizardControllerName: App.router.get('addSecurityController.name'),
          localdb: App.db.data
        });
      }
    }
  }.observes('commands.@each.isCompleted', 'commands.@each.requestId')
});

});

require.register("controllers/main/admin/serviceAccounts_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

require('controllers/main/service/info/configs');

App.MainAdminServiceAccountsController = App.MainServiceInfoConfigsController.extend({
  name: 'mainAdminServiceAccountsController',
  users: null,
  content: Em.Object.create({
    serviceName: 'MISC'
  }),
  loadUsers: function () {
    this.set('selectedService', this.get('content.serviceName') ? this.get('content.serviceName') : "MISC");
    this.loadServiceConfig();
  },
  loadServiceConfig: function () {
    App.ajax.send({
      name: 'config.tags',
      sender: this,
      data: {
        serviceName: this.get('selectedService'),
        serviceConfigsDef: this.get('serviceConfigs').findProperty('serviceName', this.get('selectedService'))
      },
      success: 'loadServiceTagSuccess'
    });
  },
  loadServiceTagSuccess: function (data, opt, params) {
    var self = this;
    var installedServices = App.Service.find().mapProperty("serviceName");
    var serviceConfigsDef = params.serviceConfigsDef;
    var serviceName = this.get('selectedService');
    var loadedClusterSiteToTagMap = {};

    for (var site in Em.get(data, 'Clusters.desired_configs')) {
      if (serviceConfigsDef.configTypes.hasOwnProperty(site)) {
        loadedClusterSiteToTagMap[site] = data.Clusters.desired_configs[site]['tag'];
      }
    }
    this.setServiceConfigTags(loadedClusterSiteToTagMap);
    // load server stored configurations
    App.router.get('configurationController').getConfigsByTags(this.get('serviceConfigTags')).done(function (serverConfigs) {
      // load configurations list for installed services
      App.config.loadAdvancedConfigPartial(installedServices, {
        queryFilter: 'configurations/StackConfigurations/property_type.matches(.*[USER,GROUP].*)'
      }, function(advancedConfigs) {
        // load cluster configs
        App.config.loadClusterConfig(function(clusterConfigs) {
          self.createConfigObject(serverConfigs, advancedConfigs.concat(clusterConfigs));
        });
      });
    });
  },

  /**
   * Generate configuration object that will be rendered
   *
   * @param {Object[]} serverConfigs
   * @param {Object[]} advancedConfigs
   */
  createConfigObject: function(serverConfigs, advancedConfigs) {
    var configSet = App.config.mergePreDefinedWithLoaded(serverConfigs, advancedConfigs, this.get('serviceConfigTags'), this.get('selectedService'));
    var miscConfigs = configSet.configs.filterProperty('serviceName', this.get('selectedService')).filterProperty('category', 'Users and Groups').filterProperty('isVisible', true).rejectProperty('displayType', 'password').rejectProperty('displayType', 'checkbox');

    miscConfigs = App.config.miscConfigVisibleProperty(miscConfigs, App.Service.find().mapProperty('serviceName').concat('MISC'));

    this.set('users', miscConfigs.filterProperty('isVisible'));
    this.set('dataIsLoaded', true);
  },
  /**
   * set config value to property of "content"
   * @param key
   * @param configName
   * @param misc_configs
   * @return {Boolean}
   */
  setContentProperty: function (key, configName, misc_configs) {
    var content = this.get('content');
    if (key && configName && misc_configs.someProperty('name', configName) && content.get(key)) {
      content.set(key, misc_configs.findProperty('name', configName).get("value"));
      return true;
    }
    return false;
  },
  /**
   * sort miscellaneous configs by specific order
   * @param sortOrder
   * @param arrayToSort
   * @return {Array}
   */
  sortByOrder: function (sortOrder, arrayToSort) {
    var sorted = [];
    if (sortOrder && sortOrder.length > 0) {
      sortOrder.forEach(function (name) {
        var user = arrayToSort.findProperty('name', name);
        if (user) {
          sorted.push({
            isVisible: user.get('isVisible'),
            displayName: user.get('displayName'),
            value: user.get('value')
          });
        }
      });
      return sorted;
    } else {
      return arrayToSort;
    }
  },
  /**
   * set displayName of "proxyuser_group" depending on stack version
   * @param misc_configs
   */
  setProxyUserGroupLabel: function (misc_configs) {
    var proxyUserGroup = misc_configs.findProperty('name', 'proxyuser_group');
    //stack, with version lower than 2.1, doesn't have Falcon service
    if (proxyUserGroup) {
      var proxyServices = ['HIVE', 'OOZIE', 'FALCON'];
      var services = Em.A([]);
      proxyServices.forEach(function (serviceName) {
        var stackService = App.StackService.find(serviceName);
        if (stackService) {
          services.push(stackService.get('displayName'));
        }
      }, this);
      proxyUserGroup.set('displayName', "Proxy group for " + stringUtils.getFormattedStringFromArray(services));
    }
  }
});

});

require.register("controllers/main/admin/stack_and_upgrade_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

App.MainAdminStackAndUpgradeController = Em.Controller.extend(App.LocalStorage, {
  name: 'mainAdminStackAndUpgradeController',

  /**
   * @type {boolean}
   */
  isLoaded: false,

  /**
   * @type {object}
   * @default null
   */
  upgradeData: null,

  /**
   * @type {number}
   * @default null
   */
  upgradeId: null,

  /**
   * @type {string}
   * @default null
   */
  upgradeVersion: null,

  /**
   * @type {boolean}
   * @default false
   */
  isDowngrade: false,

  /**
   * version that currently applied to server
   * should be plain object, because stored to localStorage
   * @type {object|null}
   */
  currentVersion: null,

  /**
   * versions to which cluster could be upgraded
   * @type {Array}
   */
  targetVersions: [],

  /**
   * @type {boolean} true if some request that should disable actions is in progress
   */
  requestInProgress: false,
  /**
   * properties that stored to localStorage to resume wizard progress
   */
  wizardStorageProperties: ['upgradeId', 'upgradeVersion', 'currentVersion', 'isDowngrade'],

  /**
   * mutable properties of Upgrade Task
   * @type {Array}
   */
  taskDetailsProperties: ['status', 'stdout', 'stderr', 'error_log', 'host_name', 'output_log'],

  /**
   * path to the mock json
   * @type {String}
   */
  mockRepoUrl: '/data/stack_versions/repo_versions_all.json',

  /**
   * api to get RepoVersions
   * @type {String}
   */
  realRepoUrl: function () {
    return App.get('apiPrefix') + App.get('stackVersionURL') +
      '/repository_versions?fields=*,operating_systems/*,operating_systems/repositories/*';
  }.property('App.stackVersionURL'),

  /**
   * path to the mock json
   * @type {String}
   */
  mockStackUrl: '/data/stack_versions/stack_version_all.json',

  /**
   * api to get ClusterStackVersions with repository_versions (use to init data load)
   * @type {String}
   */
  realStackUrl: function () {
    return App.get('apiPrefix') + '/clusters/' + App.get('clusterName') +
      '/stack_versions?fields=*,repository_versions/*,repository_versions/operating_systems/repositories/*';
  }.property('App.clusterName'),

  /**
   * api to get ClusterStackVersions without repository_versions (use to update data)
   * @type {String}
   */
  realUpdateUrl: function () {
    return App.get('apiPrefix') + '/clusters/' + App.get('clusterName') + '/stack_versions?fields=ClusterStackVersions/*';
  }.property('App.clusterName'),

  init: function () {
    this.initDBProperties();
  },

  /**
   * restore data from localStorage
   */
  initDBProperties: function () {
    this.get('wizardStorageProperties').forEach(function (property) {
      if (this.getDBProperty(property)) {
        this.set(property, this.getDBProperty(property));
      }
    }, this);
  },

  /**
   * load all data:
   * - upgrade data
   * - stack versions
   * - repo versions
   */
  load: function () {
    var dfd = $.Deferred();
    var self = this;

    this.loadUpgradeData(true).done(function() {
      self.loadStackVersionsToModel(true).done(function () {
        self.loadRepoVersionsToModel().done(function() {
          var currentVersion = App.StackVersion.find().findProperty('state', 'CURRENT');
          if (currentVersion) {
            self.set('currentVersion', {
              repository_version: currentVersion.get('repositoryVersion.repositoryVersion'),
              repository_name: currentVersion.get('repositoryVersion.displayName')
            });
          }
          dfd.resolve();
        });
      });
    });
    return dfd.promise();
  },

  /**
   * load upgrade tasks by upgrade id
   * @return {$.Deferred}
   * @param {boolean} onlyState
   */
  loadUpgradeData: function (onlyState) {
    var upgradeId = this.get('upgradeId');
    var deferred = $.Deferred();

    if (Em.isNone(upgradeId)) {
      deferred.resolve();
      console.log('Upgrade in INIT state');
    } else {
      App.ajax.send({
        name: (onlyState) ? 'admin.upgrade.state' : 'admin.upgrade.data',
        sender: this,
        data: {
          id: upgradeId
        },
        success: 'loadUpgradeDataSuccessCallback'
      }).then(deferred.resolve);
    }
    return deferred.promise();
  },

  /**
   * parse and push upgrade tasks to controller
   * @param data
   */
  loadUpgradeDataSuccessCallback: function (data) {
    if (Em.isNone(data)) return;
    App.set('upgradeState', data.Upgrade.request_status);
    this.setDBProperty('upgradeState', data.Upgrade.request_status);
    if (data.upgrade_groups) {
      this.updateUpgradeData(data);
    }
  },

  /**
   * update data of Upgrade
   * @param {object} newData
   */
  updateUpgradeData: function (newData) {
    var oldData = this.get('upgradeData'),
      groupsMap = {},
      itemsMap = {};

    if (Em.isNone(oldData) || (newData.upgrade_groups.length !== oldData.upgradeGroups.length)) {
      this.initUpgradeData(newData);
    } else {
      //create entities maps
      newData.upgrade_groups.forEach(function (newGroup) {
        groupsMap[newGroup.UpgradeGroup.group_id] = newGroup.UpgradeGroup;
        newGroup.upgrade_items.forEach(function (item) {
          itemsMap[item.UpgradeItem.stage_id] = item.UpgradeItem;
        })
      });

      //update existed entities with new data
      oldData.upgradeGroups.forEach(function (oldGroup) {
        oldGroup.set('status', groupsMap[oldGroup.get('group_id')].status);
        oldGroup.set('progress_percent', groupsMap[oldGroup.get('group_id')].progress_percent);
        oldGroup.upgradeItems.forEach(function (item) {
          item.set('status', itemsMap[item.get('stage_id')].status);
          item.set('progress_percent', itemsMap[item.get('stage_id')].progress_percent);
        })
      });
      oldData.set('Upgrade', newData.Upgrade);
    }
  },

  /**
   * change structure of Upgrade
   * In order to maintain nested views in template object should have direct link to its properties, for example
   * item.UpgradeItem.<properties> -> item.<properties>
   * @param {object} newData
   */
  initUpgradeData: function (newData) {
    var upgradeGroups = [];

    //wrap all entities into App.upgradeEntity
    newData.upgrade_groups.forEach(function (newGroup) {
      var oldGroup = App.upgradeEntity.create({type: 'GROUP'}, newGroup.UpgradeGroup);
      var upgradeItems = [];
      newGroup.upgrade_items.forEach(function (item) {
        var oldItem = App.upgradeEntity.create({type: 'ITEM'}, item.UpgradeItem);
        oldItem.set('tasks', []);
        upgradeItems.pushObject(oldItem);
      });
      upgradeItems.reverse();
      oldGroup.set('upgradeItems', upgradeItems);
      upgradeGroups.pushObject(oldGroup);
    });
    upgradeGroups.reverse();
    this.set('upgradeData', Em.Object.create({
      upgradeGroups: upgradeGroups,
      Upgrade: newData.Upgrade
    }));
  },

  /**
   * request Upgrade Item and its tasks from server
   * @return {$.ajax}
   */
  getUpgradeItem: function (item) {
    return App.ajax.send({
      name: 'admin.upgrade.upgrade_item',
      sender: this,
      data: {
        upgradeId: item.get('request_id'),
        groupId: item.get('group_id'),
        stageId: item.get('stage_id')
      },
      success: 'getUpgradeItemSuccessCallback'
    });
  },

  /**
   * success callback of <code>getTasks</code>
   * @param {object} data
   */
  getUpgradeItemSuccessCallback: function (data) {
    this.get('upgradeData.upgradeGroups').forEach(function (group) {
      if (group.get('group_id') === data.UpgradeItem.group_id) {
        group.get('upgradeItems').forEach(function (item) {
          if (item.get('stage_id') === data.UpgradeItem.stage_id) {
            if (item.get('tasks.length')) {
              item.set('isTasksLoaded', true);
              data.tasks.forEach(function (task) {
                var currentTask = item.get('tasks').findProperty('id', task.Tasks.id);
                this.get('taskDetailsProperties').forEach(function (property) {
                  currentTask.set(property, task.Tasks[property]);
                }, this);
              }, this);
            } else {
              var tasks = [];
              data.tasks.forEach(function (task) {
                tasks.pushObject(App.upgradeEntity.create({type: 'TASK'}, task.Tasks));
              });
              item.set('tasks', tasks);
            }
            item.set('isTasksLoaded', true);
          }
        }, this);
      }
    }, this);
  },

  /**
   * downgrade confirmation popup
   * @param {object} event
   */
  confirmDowngrade: function (event) {
    var self = this;
    var currentVersion = this.get('currentVersion');
    return App.showConfirmationPopup(
      function() {
        self.downgrade.call(self, currentVersion, event);
      },
      Em.I18n.t('admin.stackUpgrade.downgrade.body').format(currentVersion.repository_name),
      null,
      Em.I18n.t('admin.stackUpgrade.dialog.downgrade.header').format(currentVersion.repository_name),
      Em.I18n.t('admin.stackUpgrade.downgrade.proceed')
    );
  },

  /**
   * make call to start downgrade process
   * @param {object} currentVersion
   * @param {object} event
   */
  downgrade: function (currentVersion, event) {
    this.set('requestInProgress', true);
    this.abortUpgrade();
    App.ajax.send({
      name: 'admin.downgrade.start',
      sender: this,
      data: {
        value: currentVersion.repository_version,
        label: currentVersion.repository_name,
        isDowngrade: true
      },
      success: 'upgradeSuccessCallback',
      callback: function() {
        this.sender.set('requestInProgress', false);
      }
    });
  },

  /**
   * abort upgrade (in order to start Downgrade)
   */
  abortUpgrade: function () {
    return App.ajax.send({
      name: 'admin.upgrade.abort',
      sender: this,
      data: {
        upgradeId: this.get('upgradeId')
      }
    });
  },

  /**
   * make call to start upgrade process and show popup with current progress
   * @param {object} version
   */
  upgrade: function (version) {
    this.set('requestInProgress', true);
    App.ajax.send({
      name: 'admin.upgrade.start',
      sender: this,
      data: version,
      success: 'upgradeSuccessCallback',
      callback: function() {
        this.sender.set('requestInProgress', false);
      }
    });
    this.setDBProperty('currentVersion', this.get('currentVersion'));
  },

  /**
   * success callback of <code>upgrade()</code>
   * @param {object} data
   */
  upgradeSuccessCallback: function (data, opt, params) {
    this.set('upgradeData', null);
    this.set('upgradeId', data.resources[0].Upgrade.request_id);
    this.set('upgradeVersion', params.label);
    this.set('isDowngrade', !!params.isDowngrade);
    this.setDBProperty('upgradeVersion', params.label);
    this.setDBProperty('upgradeId', data.resources[0].Upgrade.request_id);
    this.setDBProperty('upgradeState', 'PENDING');
    this.setDBProperty('isDowngrade', !!params.isDowngrade);
    App.set('upgradeState', 'PENDING');
    App.clusterStatus.setClusterStatus({
      wizardControllerName: this.get('name'),
      localdb: App.db.data
    });
    this.load();
    this.openUpgradeDialog();
  },

  /**
   * upgrade confirmation popup
   * @param {object} version
   * @return App.ModalPopup
   */
  confirmUpgrade: function (version) {
    var self = this;

    return App.showConfirmationPopup(
      function () {
        self.runPreUpgradeCheck.call(self, version);
      },
      Em.I18n.t('admin.stackUpgrade.upgrade.confirm.body').format(version.get('displayName')),
      null,
      Em.I18n.t('admin.stackUpgrade.dialog.header').format(version.get('displayName'))
    );
  },

  /**
   * send request for pre upgrade check
   * @param version
   */
  runPreUpgradeCheck: function(version) {
    var params = {
      value: version.get('repositoryVersion'),
      label: version.get('displayName')
    };

    if (App.get('supports.preUpgradeCheck')) {
      this.set('requestInProgress', true);
      App.ajax.send({
        name: "admin.rolling_upgrade.pre_upgrade_check",
        sender: this,
        data: params,
        success: "runPreUpgradeCheckSuccess",
        error: "runPreUpgradeCheckError"
      });
    } else {
      this.upgrade(params);
    }
  },

  /**
   * success callback of <code>runPreUpgradeCheckSuccess()</code>
   * if there are some fails - it shows popup else run upgrade
   * @param data {object}
   * @param opt {object}
   * @param params {object}
   * @returns {App.ModalPopup|undefined}
   */
  runPreUpgradeCheckSuccess: function (data, opt, params) {
    if (data.items.someProperty('UpgradeChecks.status', "FAIL")) {
      this.set('requestInProgress', false);
      var header = Em.I18n.t('popup.clusterCheck.Upgrade.header').format(params.label);
      var title = Em.I18n.t('popup.clusterCheck.Upgrade.title');
      var alert = Em.I18n.t('popup.clusterCheck.Upgrade.alert');
      App.showClusterCheckPopup(data, header, title, alert);
    } else {
      this.upgrade(params);
    }
  },

  runPreUpgradeCheckError: function() {
    this.set('requestInProgress', false);
  },

  /**
   * confirmation popup before install repository version
   */
  installRepoVersionConfirmation: function (repo) {
    var self = this;
    return App.showConfirmationPopup(function () {
        self.installRepoVersion(repo);
      },
      Em.I18n.t('admin.stackVersions.version.install.confirm').format(repo.get('displayName'))
    );
  },

  /**
   * sends request to install repoVersion to the cluster
   * and create clusterStackVersion resourse
   * @param {Em.Object} repo
   * @return {$.ajax}
   * @method installRepoVersion
   */
  installRepoVersion: function (repo) {
    this.set('requestInProgress', true);
    var data = {
      ClusterStackVersions: {
        stack: repo.get('stackVersionType'),
        version: repo.get('stackVersionNumber'),
        repository_version: repo.get('repositoryVersion')
      },
      id: repo.get('id')
    };
    return App.ajax.send({
      name: 'admin.stack_version.install.repo_version',
      sender: this,
      data: data,
      success: 'installRepoVersionSuccess',
      callback: function() {
        this.sender.set('requestInProgress', false);
      }
    });
  },

  /**
   * transform repo data into json for
   * saving changes to repository version
   * @param {Em.Object} repo
   * @returns {{operating_systems: Array}}
   */
  prepareRepoForSaving: function(repo) {
    var repoVersion = { "operating_systems": [] };

    repo.get('operatingSystems').forEach(function (os, k) {
      repoVersion.operating_systems.push({
        "OperatingSystems": {
          "os_type": os.get("osType")
        },
        "repositories": []
      });
      os.get('repositories').forEach(function (repository) {
        repoVersion.operating_systems[k].repositories.push({
          "Repositories": {
            "base_url": repository.get('baseUrl'),
            "repo_id": repository.get('repoId'),
            "repo_name": repository.get('repoName')
          }
        });
      });
    });
    return repoVersion;
  },

  /**
   * perform validation if <code>skip<code> is  false and run save if
   * validation successfull or run save without validation is <code>skip<code> is true
   * @param {Em.Object} repo
   * @param {boolean} skip
   * @returns {$.Deferred}
   */
  saveRepoOS: function (repo, skip) {
    var self = this;
    var deferred = $.Deferred();
    this.validateRepoVersions(repo, skip).done(function(data) {
      if (data.length > 0) {
        deferred.resolve(data);
      } else {
        var repoVersion = self.prepareRepoForSaving(repo);

        App.ajax.send({
          name: 'admin.stack_versions.edit.repo',
          sender: this,
          data: {
            stackName: App.get('currentStackName'),
            stackVersion: App.get('currentStackVersionNumber'),
            repoVersionId: repo.get('repoVersionId'),
            repoVersion: repoVersion
          }
        }).success(function() {
          deferred.resolve([]);
        });
      }
    });
    return deferred.promise();
  },

  /**
   * send request for validation for each repository
   * @param {Em.Object} repo
   * @param {boolean} skip
   * @returns {*}
   */
  validateRepoVersions: function(repo, skip) {
    var deferred = $.Deferred(),
      totalCalls = 0,
      invalidUrls = [];

    if (skip) {
      deferred.resolve(invalidUrls);
    } else {
      repo.get('operatingSystems').forEach(function (os) {
        if (os.get('isSelected')) {
          os.get('repositories').forEach(function (repo) {
            totalCalls++;
            App.ajax.send({
              name: 'admin.stack_versions.validate.repo',
              sender: this,
              data: {
                repo: repo,
                repoId: repo.get('repoId'),
                baseUrl: repo.get('baseUrl'),
                osType: os.get('osType'),
                stackName: App.get('currentStackName'),
                stackVersion: App.get('currentStackVersionNumber')
              }
            })
              .success(function () {
                totalCalls--;
                if (totalCalls === 0) deferred.resolve(invalidUrls);
              })
              .error(function () {
                repo.set('hasError', true);
                invalidUrls.push(repo);
                totalCalls--;
                if (totalCalls === 0) deferred.resolve(invalidUrls);
              });
          });
        } else {
          return deferred.resolve(invalidUrls);
        }
      });
    }
    return deferred.promise();
  },

  /**
   * success callback for <code>installRepoVersion()<code>
   * saves request id to the db
   * @param data
   * @param opt
   * @param params
   * @method installStackVersionSuccess
   */
  installRepoVersionSuccess: function (data, opt, params) {
    App.db.set('repoVersionInstall', 'id', [data.Requests.id]);
    App.clusterStatus.setClusterStatus({
      wizardControllerName: this.get('name'),
      localdb: App.db.data
    });
    App.RepositoryVersion.find(params.id).set('defaultStatus', 'INSTALLING');
  },

  /**
   * opens a popup with installations state per host
   * @param {Em.Object} version
   * @method showProgressPopup
   */
  showProgressPopup: function(version) {
    var popupTitle = Em.I18n.t('admin.stackVersions.details.install.hosts.popup.title').format(version.get('displayName'));
    var requestIds = App.get('testMode') ? [1] : App.db.get('repoVersionInstall', 'id');
    var hostProgressPopupController = App.router.get('highAvailabilityProgressPopupController');
    hostProgressPopupController.initPopup(popupTitle, requestIds, this);
  },

  /**
   * reset upgradeState to INIT when upgrade is COMPLETED
   * and clean auxiliary data
   */
  finish: function () {
    if (App.get('upgradeState') === 'COMPLETED') {
      this.setDBProperty('upgradeId', undefined);
      this.setDBProperty('upgradeState', 'INIT');
      this.setDBProperty('upgradeVersion', undefined);
      this.setDBProperty('currentVersion', undefined);
      this.setDBProperty('isDowngrade', undefined);
      App.clusterStatus.setClusterStatus({
        localdb: App.db.data
      });
      App.set('upgradeState', 'INIT');
    }
  }.observes('App.upgradeState'),

  /**
   * show dialog with tasks of upgrade
   * @return {App.ModalPopup}
   */
  openUpgradeDialog: function () {
    App.router.transitionTo('admin.stackUpgrade');
  },

  /**
   * returns url to get data for repoVersion or clusterStackVersion
   * @param {Boolean} stack true if load clusterStackVersion
   * @param {Boolean} fullLoad true if load all data
   * @returns {String}
   * @method getUrl
   */
  getUrl: function(stack, fullLoad) {
    if (App.get('testMode')) {
      return stack ? this.get('mockStackUrl') : this.get('mockRepoUrl')
    } else {
      if (fullLoad) {
        return stack ? this.get('realStackUrl') : this.get('realRepoUrl');
      } else {
        return this.get('realUpdateUrl');
      }
    }
  },

  /**
   * get stack versions from server and push it to model
   * @return {*}
   * @method loadStackVersionsToModel
   */
  loadStackVersionsToModel: function (fullLoad) {
    var dfd = $.Deferred();
    App.HttpClient.get(this.getUrl(true, fullLoad), App.stackVersionMapper, {
      complete: function () {
        dfd.resolve();
      }
    });
    return dfd.promise();
  },

  /**
   * get repo versions from server and push it to model
   * @return {*}
   * @params {Boolean} isUpdate - if true loads part of data that need to be updated
   * @method loadRepoVersionsToModel()
   */
  loadRepoVersionsToModel: function () {
    var dfd = $.Deferred();
    App.HttpClient.get(this.getUrl(false, true), App.repoVersionMapper, {
      complete: function () {
        dfd.resolve();
      }
    });
    return dfd.promise();
  },

  /**
   * set status to Upgrade item
   * @param item
   * @param status
   */
  setUpgradeItemStatus: function(item, status) {
    this.set('requestInProgress', true);
    return App.ajax.send({
      name: 'admin.upgrade.upgradeItem.setState',
      sender: this,
      data: {
        upgradeId: item.get('request_id'),
        itemId: item.get('stage_id'),
        groupId: item.get('group_id'),
        status: status
      },
      callback: function() {
        this.sender.set('requestInProgress', false);
      }
    }).done(function () {
      item.set('status', status);
    });
  },

  currentVersionObserver: function () {
    var versionNumber = this.get('currentVersion.repository_version');
    var currentVersionObject = App.RepositoryVersion.find().findProperty('status', 'CURRENT');
    var versionName = currentVersionObject && currentVersionObject.get('stackVersionType');
    App.set('isStormMetricsSupported', versionName != 'HDP' || stringUtils.compareVersions(versionNumber, '2.2.2') > -1 || !versionNumber);
  }.observes('currentVersion.repository_version'),

  /**
   * get the installed repositories of HDP from server
   */
  loadRepositories: function () {
    if (App.router.get('clusterController.isLoaded')) {
      var nameVersionCombo = App.get('currentStackVersion');
      var stackName = nameVersionCombo.split('-')[0];
      var stackVersion = nameVersionCombo.split('-')[1];
      App.ajax.send({
        name: 'cluster.load_repositories',
        sender: this,
        data: {
          stackName: stackName,
          stackVersion: stackVersion
        },
        success: 'loadRepositoriesSuccessCallback',
        error: 'loadRepositoriesErrorCallback'
      });
    }
  }.observes('App.router.clusterController.isLoaded'),

  loadRepositoriesSuccessCallback: function (data) {
    var allRepos = [];
    data.items.forEach(function (os) {
      os.repositories.forEach(function (repository) {
        var osType = repository.Repositories.os_type;
        var repo = Em.Object.create({
          baseUrl: repository.Repositories.base_url,
          osType: osType,
          repoId: repository.Repositories.repo_id,
          repoName : repository.Repositories.repo_name,
          stackName : repository.Repositories.stack_name,
          stackVersion : repository.Repositories.stack_version,
          isFirst: false
        });
        var group = allRepos.findProperty('name', osType);
        if (!group) {
          group = {
            name: osType,
            repositories: []
          };
          repo.set('isFirst', true);
          allRepos.push(group);
        }
        group.repositories.push(repo);
      });
    }, this);
    allRepos.stackVersion = App.get('currentStackVersionNumber');
    this.set('allRepos', allRepos);
  },

  loadRepositoriesErrorCallback: function (request, ajaxOptions, error) {
    console.log('Error message is: ' + request.responseText);
  }
});

});

require.register("controllers/main/alert_definitions_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAlertDefinitionsController = Em.ArrayController.extend({

  name: 'mainAlertDefinitionsController',

  /**
   * Timestamp when <code>App.alertDefinitionsMapper</code> run last time
   * Current <code>content</code> is updated on when it changed
   * @type {number|null}
   */
  mapperTimestamp: null,

  /**
   * Define whether restore filter conditions from local db
   * @type {Boolean}
   */
  showFilterConditionsFirstLoad: false,

  /**
   * List of all <code>App.AlertDefinition</code>
   * Consists of:
   * <ul>
   *   <li>App.PortAlertDefinition</li>
   *   <li>App.MetricsAlertDefinition</li>
   *   <li>App.WebAlertDefinition</li>
   *   <li>App.AggregateAlertDefinition</li>
   *   <li>App.ScriptAlertDefinition</li>
   * </ul>
   * @type {App.AlertDefinition[]}
   */
  content: App.AlertDefinition.find(),

  /**
   * Enable/disable alertDefinition confirmation popup
   * @param {object} event
   * @method toggleState
   * @return {App.ModalPopup}
   */
  toggleState: function (event) {
    var alertDefinition = event.context;
    var self = this;
    var bodyMessage = Em.Object.create({
      confirmMsg: alertDefinition.get('enabled') ? Em.I18n.t('alerts.table.state.enabled.confirm.msg') : Em.I18n.t('alerts.table.state.disabled.confirm.msg'),
      confirmButton: alertDefinition.get('enabled') ? Em.I18n.t('alerts.table.state.enabled.confirm.btn') : Em.I18n.t('alerts.table.state.disabled.confirm.btn')
    });

    return App.showConfirmationFeedBackPopup(function (query) {
      self.toggleDefinitionState(alertDefinition);
    }, bodyMessage);
  },

  /**
   * Enable/disable alertDefinition
   * @param {object} alertDefinition
   * @returns {$.ajax}
   * @method toggleDefinitionState
   */
  toggleDefinitionState: function (alertDefinition) {
    var newState = !alertDefinition.get('enabled');
    alertDefinition.set('enabled', newState);
    return App.ajax.send({
      name: 'alerts.update_alert_definition',
      sender: this,
      data: {
        id: alertDefinition.get('id'),
        data: {
          "AlertDefinition/enabled": newState
        }
      }
    });
  },

  /**
   * Calculate critical count for each service, to show up the label on services menu
   * @method getCriticalAlertsCountForService
   * @return {Number}
   */
  getCriticalAlertsCountForService: function (service) {
    return this.get('content').filterProperty('service.serviceName', service).
      invoke('getWithDefault', 'summary.CRITICAL.count', 0).
      reduce(Em.sum, 0);
  },

  /**
   * Calculate critical/warning count for each service, to show up the label on services menu
   * @method getCriticalOrWarningAlertsCountForService
   * @return {Number}
   */
  getCriticalOrWarningAlertsCountForService: function (service) {
    return this.get('content').filterProperty('service.serviceName', service).map(function (alertDefinition) {
      return alertDefinition.getWithDefault('summary.CRITICAL.count', 0) + alertDefinition.getWithDefault('summary.WARNING.count', 0);
    }).reduce(Em.sum, 0);
  },

  /**
   *  ========================== alerts popup dialog =========================
   */

  /**
   * Number of all critical and warning alert instances
   * Calculation is based on each <code>alertDefinitions.summary</code>
   * @type {Number}
   */
  unhealthyAlertInstancesCount: function () {
    return this.get('content').map(function (alertDefinition) {
      return alertDefinition.getWithDefault('summary.CRITICAL.count', 0) + alertDefinition.getWithDefault('summary.WARNING.count', 0);
    }).reduce(Em.sum, 0);
  }.property('content.@each.summary'),

  /**
   * if critical alerts exist, the alert badge should be red.
   * @type {Boolean}
   */
  isCriticalAlerts: function () {
    return this.get('content').invoke('getWithDefault', 'summary.CRITICAL.count', 0).reduce(Em.sum, 0) !== 0;
  }.property('content.@each.summary')

});

});

require.register("controllers/main/alerts/add_alert_definition/add_alert_definition_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.AddAlertDefinitionController = App.WizardController.extend({

  name: 'addAlertDefinitionController',

  totalSteps: 3,

  content: {
    selectedType: null
  },

  /**
   * Do request to create new alert definition
   * @param {object} newDefinitionData
   * @returns {$.ajax}
   */
  createNewAlertDefinition: function (newDefinitionData) {
    return App.ajax.send({
      name: 'alerts.create_alert_definition',
      sender: this,
      data: {
        data: newDefinitionData
      }
    });
  },

  finish: function() {
    this.clear();
    this.setCurrentStep('1');
  }

});

});

require.register("controllers/main/alerts/add_alert_definition/step1_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.AddAlertDefinitionStep1Controller = Em.Controller.extend({

  name: 'addAlertDefinitionStep1',

  /**
   * List of available alert definition types
   * @type {{value: string, isActive: boolean}[]}
   */
  alertDefinitionsTypes: [
    Em.Object.create({value: 'PORT', isActive: false, icon: 'icon-signal'}),
    Em.Object.create({value: 'METRIC', isActive: false, icon: 'icon-bolt'}),
    Em.Object.create({value: 'WEB', isActive: false, icon: 'icon-globe'}),
    Em.Object.create({value: 'AGGREGATE', isActive: false, icon: 'icon-plus-sign-alt'}),
    Em.Object.create({value: 'SCRIPT', isActive: false, icon: 'icon-code'})
  ],

  /**
   * "Next"-button is disabled if user doesn't select any alert definition type
   * @type {boolean}
   */
  isSubmitDisabled: function() {
    return this.get('alertDefinitionsTypes').everyProperty('isActive', false);
  }.property('alertDefinitionsTypes.@each.isActive'),

  /**
   * Set selectedType if it exists in the wizard controller
   * @method loadStep
   */
  loadStep: function() {
    this.get('alertDefinitionsTypes').setEach('isActive', false);
    var selectedType = this.get('content.selectedType');
    if(selectedType) {
      this.selectType({context: {value: selectedType}});
    }
  },

  /**
   * Handler for select alert definition type selection
   * @param {object} e
   * @method selectType
   */
  selectType: function(e) {
    var type = e.context,
      types = this.get('alertDefinitionsTypes');
    types.setEach('isActive', false);
    types.findProperty('value', type.value).set('isActive', true);
    this.set('content.selectedType', type.value);
  }

});

});

require.register("controllers/main/alerts/add_alert_definition/step2_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.AddAlertDefinitionStep2Controller = Em.Controller.extend({

  name: 'addAlertDefinitionStep2Controller'

});

});

require.register("controllers/main/alerts/add_alert_definition/step3_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.AddAlertDefinitionStep3Controller = Em.Controller.extend({

  name: 'addAlertDefinitionStep3Controller'

});

});

require.register("controllers/main/alerts/alert_definitions_actions_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

App.MainAlertDefinitionActionsController = Em.ArrayController.extend({

  name: 'mainAlertDefinitionActionsController',

  /**
   * List of available actions for alert definitions
   * @type {{title: string, icon: string, action: string, showDivider: boolean}[]}
   */
  content: [
    /*{
      title: Em.I18n.t('alerts.actions.create'),
      icon: 'icon-plus',
      action: 'createNewAlertDefinition',
      showDivider: true
    },*/
    {
      title: Em.I18n.t('alerts.actions.manageGroups'),
      icon: 'icon-th-large',
      action: 'manageAlertGroups',
      showDivider: false
    },
    {
      title: Em.I18n.t('alerts.actions.manageNotifications'),
      icon: 'icon-envelope-alt',
      action: 'manageNotifications',
      showDivider: false
    }
  ],

  /**
   * Common handler for menu item click
   * Call proper controller's method described in <code>action</code>-field (see <code>content</code>)
   * @param {object} event
   * @method actionHandler
   */
  actionHandler: function(event) {
    var menuElement = event.context,
      action = menuElement.action;
    if ('function' === Em.typeOf(Em.get(this, action))) {
      this[action]();
    }
    else {
      console.error('Invalid action provided - ', action);
    }
  },

  /**
   * Start "Create Alert Definition" wizard
   * @method createNewAlertDefinition
   */
  createNewAlertDefinition: function() {
    App.router.transitionTo('alertAdd');
  },

  /**
   * Handler when clicking on "Manage Alert Groups", a popup will show up
   * @method manageAlertGroups
   * @return {App.ModalPopup}
   */
  manageAlertGroups: function () {

    return App.ModalPopup.show({

      header: Em.I18n.t('alerts.actions.manage_alert_groups_popup.header'),

      bodyClass: App.MainAlertsManageAlertGroupView.extend({
        controllerBinding: 'App.router.manageAlertGroupsController'
      }),

      classNames: ['sixty-percent-width-modal', 'manage-alert-group-popup'],
      primary: Em.I18n.t('common.save'),

      onPrimary: function () {
        var modifiedAlertGroups = this.get('subViewController.defsModifiedAlertGroups');
        var dataForSuccessPopup = {
          created: modifiedAlertGroups.toCreate.length,
          deleted: modifiedAlertGroups.toDelete.length,
          updated: modifiedAlertGroups.toSet.length
        };
        var showSuccessPopup = dataForSuccessPopup.created + dataForSuccessPopup.deleted + dataForSuccessPopup.updated > 0;
        // Save modified Alert-groups
        console.log("manageAlertGroups(): Saving modified Alert groups: ", modifiedAlertGroups);
        var self = this;
        var errors = [];
        var deleteQueriesCounter = modifiedAlertGroups.toDelete.length;
        var createQueriesCounter = modifiedAlertGroups.toCreate.length;
        var deleteQueriesRun = false;
        var createQueriesRun = false;
        var runNextQuery = function () {
          if (!deleteQueriesRun && deleteQueriesCounter > 0) {
            deleteQueriesRun = true;
            modifiedAlertGroups.toDelete.forEach(function (group) {
              self.get('subViewController').removeAlertGroup(group, finishFunction, finishFunction);
            }, this);
          } else if (!createQueriesRun && deleteQueriesCounter < 1) {
            createQueriesRun = true;
            modifiedAlertGroups.toSet.forEach(function (group) {
              self.get('subViewController').updateAlertGroup(group, finishFunction, finishFunction);
            }, this);
            modifiedAlertGroups.toCreate.forEach(function (group) {
              self.get('subViewController').postNewAlertGroup(group, finishFunction);
            }, this);
          }
        };
        var finishFunction = function (xhr, text, errorThrown) {
          if (xhr && errorThrown) {
            var error = xhr.status + "(" + errorThrown + ") ";
            try {
              var json = $.parseJSON(xhr.responseText);
              error += json.message;
            } catch (err) {
            }
            console.error('Error updating Alert Group:', error);
            errors.push(error);
          }
          if (createQueriesRun) {
            createQueriesCounter--;
          } else {
            deleteQueriesCounter--;
          }
          if (deleteQueriesCounter + createQueriesCounter < 1) {
            if (errors.length > 0) {
              console.log(errors);
              self.get('subViewController').set('errorMessage', errors.join(". "));
            }
            else {
              self.hide();
              if (showSuccessPopup) {
                App.ModalPopup.show({
                  secondary: null,
                  header: Em.I18n.t('alerts.groups.successPopup.header'),
                  bodyClass: Em.View.extend({
                    dataForSuccessPopup: dataForSuccessPopup,
                    templateName: require('templates/main/alerts/alert_groups/success_popup_body')
                  })
                });
              }
              App.router.get('updateController').updateAlertGroups(function () {
                App.router.get('manageAlertGroupsController').toggleProperty('changeTrigger');
                App.router.get('updateController').updateAlertDefinitions(function() {
                  App.router.get('updateController').updateAlertNotifications(Em.K);
                });
              });
            }
          } else {
            runNextQuery();
          }
        };
        runNextQuery();
      },

      subViewController: function () {
        return App.router.get('manageAlertGroupsController');
      }.property('App.router.manageAlertGroupsController'),

      updateButtons: function () {
        var modified = this.get('subViewController.isDefsModified');
        this.set('disablePrimary', !modified);
      }.observes('subViewController.isDefsModified'),

      didInsertElement: Em.K

    });
  },

  /**
   * "Manage Alert Notifications" handler
   * @method manageNotifications
   * @return {App.ModalPopup}
   */
  manageNotifications: function () {

    return App.ModalPopup.show({

      header: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.header'),

      bodyClass: App.ManageAlertNotificationsView.extend({
        controllerBinding: 'App.router.manageAlertNotificationsController'
      }),

      classNames: ['sixty-percent-width-modal', 'manage-configuration-group-popup'],

      secondary: null,

      primary: Em.I18n.t('common.close'),

      autoHeight: false

    });
  }

});
});

require.register("controllers/main/alerts/alert_instances_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAlertInstancesController = Em.Controller.extend({

  name: 'mainAlertInstancesController',

  content: App.AlertInstance.find(),

  /**
   * @type {App.AlertInstance[]}
   */
  unhealthyAlertInstances: function () {
    return App.AlertInstance.find().filter(function (item) {
      return ['CRITICAL', 'WARNING'].contains(item.get('state'));
    });
  }.property('content.[]'),

  /**
   * Are alertInstances loaded
   * @type {boolean}
   */
  isLoaded: false,

  /**
   * A flag to reload alert instances table every 10 seconds
   * @type {boolean}
   */
  reload: false,

  /**
   * Causes automatic updates of content if set to true
   * @type {boolean}
   */
  isUpdating: false,

  /**
   * Times for alert instances updater
   * Used in <code>scheduleUpdate</code>
   * @type {number|null}
   */
  updateTimer: null,

  /**
   * @type {string|null} sourceName - hostName or alertDefinitionId
   */
  sourceName: null,

  /**
   * @type {string|null} sourceType - 'HOST'|'ALERT_DEFINITION'
   */
  sourceType: null,

  /**
   * Load alert instances from server (all, for selected host, for selected alert definition)
   * @returns {$.ajax}
   * @method fetchAlertInstances
   */
  fetchAlertInstances: function () {
    var sourceType = this.get('sourceType'),
      sourceName = this.get('sourceName'),
      ajaxData = {
        sender: this,
        success: 'getAlertInstancesSuccessCallback',
        error: 'getAlertInstancesErrorCallback'
      };

    switch (sourceType) {
      case 'HOST':
        $.extend(ajaxData, {
          name: 'alerts.instances.by_host',
          data: {
            hostName: sourceName
          }
        });
        break;

      case 'ALERT_DEFINITION':
        $.extend(ajaxData, {
          name: 'alerts.instances.by_definition',
          data: {
            definitionId: sourceName
          }
        });
        break;

      default:
        $.extend(ajaxData, {
          name: 'alerts.instances'
        });
        break;
    }

    return App.ajax.send(ajaxData);
  },

  /**
   * Pseudo for <code>fetchAlertInstances</code>
   * Used to get all alert instances
   * @method loadAlertInstances
   */
  loadAlertInstances: function () {
    this.set('isLoaded', false);
    this.set('sourceType', null);
    this.set('sourceName', null);
    this.fetchAlertInstances();
  },

  /**
   * Pseudo for <code>fetchAlertInstances</code>
   * Used to get alert instances for some host
   * @param {string} hostName
   * @method loadAlertInstancesByHost
   */
  loadAlertInstancesByHost: function (hostName) {
    this.set('isLoaded', false);
    this.set('sourceType', 'HOST');
    this.set('sourceName', hostName);
    this.fetchAlertInstances();
  },

  /**
   * Pseudo for <code>fetchAlertInstances</code>
   * Used to get alert instances for some alert definition
   * @param {string} definitionId
   * @method loadAlertInstancesByAlertDefinition
   */
  loadAlertInstancesByAlertDefinition: function (definitionId) {
    this.set('isLoaded', false);
    this.set('sourceType', 'ALERT_DEFINITION');
    this.set('sourceName', definitionId);
    this.fetchAlertInstances();
  },

  scheduleUpdate: function () {
    var self = this;
    if (this.get('isUpdating')) {
      this.set('updateTimer', setTimeout(function () {
        self.fetchAlertInstances().complete(function() {
          self.scheduleUpdate();
        });
      }, App.get('alertInstancesUpdateInterval')));
    }
    else {
      clearTimeout(this.get('updateTimer'));
    }
  }.observes('isUpdating'),

  /**
   * Success-callback for alert instances request
   * @param {object} json
   * @method getAlertInstancesSuccessCallback
   */
  getAlertInstancesSuccessCallback: function (json) {
    App.alertInstanceMapper.map(json, true);
    this.set('isLoaded', true);
    this.set('reload', !this.get('reload'));
  },

  /**
   * Error-callback for alert instances request
   * @method getAlertInstancesErrorCallback
   */
  getAlertInstancesErrorCallback: function () {
    this.set('isLoaded', true);
  },

  /**
   * Onclick handler for alerts number located right to bg ops number (see application.hbs)
   * @method showPopup
   * @return {App.ModalPopup}
   */
  showPopup: function () {

    var self = this;

    return App.ModalPopup.show({

      alertsNumberBinding: 'App.router.mainAlertDefinitionsController.unhealthyAlertInstancesCount',

      header: function () {
        return Em.I18n.t('alerts.fastAccess.popup.header').format(this.get('alertsNumber'));
      }.property('alertsNumber'),

      definitionsController: this,

      classNames: ['sixty-percent-width-modal', 'alerts-popup'],

      secondary: Em.I18n.t('alerts.fastAccess.popup.body.showmore'),

      autoHeight: false,

      isHideBodyScroll: true,

      onSecondary: function () {
        this._super();
        App.router.transitionTo('main.alerts.index');
      },

      bodyClass: App.TableView.extend(App.TableServerViewMixin, {

        updaterBinding: 'App.router.updateController',

        templateName: require('templates/common/modal_popups/alerts_popup'),

        controller: self,

        isPaginate: true,

        willInsertElement: function () {
          this._super();
          this.updateAlertInstances();
        },

        /**
         * Number of all critical and warning alert instances
         * @type {Boolean}
         */
        filteredCount: function () {
          return App.router.get('mainAlertDefinitionsController.unhealthyAlertInstancesCount');
        }.property('alertsNumber'),

        content: function () {
          return this.get('controller.unhealthyAlertInstances');
        }.property('controller.unhealthyAlertInstances.@each.state'),

        isLoaded: function () {
          return !!this.get('controller.unhealthyAlertInstances');
        }.property('controller.unhealthyAlertInstances'),

        isAlertEmptyList: function () {
          return !this.get('content.length');
        }.property('content.length'),

        /**
         * Update list of shown alert instances
         * @method updateAlertInstances
         */
        updateAlertInstances: function () {
          var self = this,
            displayLength = this.get('displayLength'),
            startIndex = this.get('startIndex');
          if (!displayLength) return; // wait while table-info is loaded
          this.get('updater').set('queryParamsForUnhealthyAlertInstances', {
            from: startIndex - 1,
            page_size: displayLength
          });
          this.set('filteringComplete', false);
          this.get('updater').updateUnhealthyAlertInstances(function() {
            self.set('filteringComplete', true);
          });
        }.observes('displayLength', 'startIndex'),

        /**
         * Show spinner when filter/sorting request is in processing
         * @method overlayObserver
         */
        overlayObserver: function() {
          var $tbody = this.$('#alert-info'),
            $overlay = this.$('.table-overlay'),
            $spinner = $($overlay).find('.spinner');
          if (!this.get('filteringComplete')) {
            if (!$tbody) return;
            var tbodyPos =  $tbody.position();
            if (!tbodyPos) return;
            $spinner.css('display', 'block');
            $overlay.css({
              top: tbodyPos.top + 1,
              left: tbodyPos.left + 1,
              width: $tbody.width() - 1,
              height: $tbody.height() - 1
            });
          }
        },

        /**
         * No filtering for alert definitions
         * @method filter
         */
        filter: function() {
          this.set('filteredContent', this.get('content'));
        }.observes('content.length'),

        refreshTooltips: function () {
          this.ensureTooltip();
        }.observes('contents.[]'),

        ensureTooltip: function () {
          Em.run.next(this, function () {
            App.tooltip($(".timeago"));
          });
        },
        /**
         * Router transition to alert definition details page
         * @param event
         */
        gotoAlertDetails: function (event) {
          if (event && event.context) {
            this.get('parentView').hide();
            var definition = App.AlertDefinition.find().findProperty('id', event.context.get('definitionId'));
            App.router.transitionTo('main.alerts.alertDetails', definition);
          }
        },

        /**
         * Router transition to service summary page
         * @param event
         */
        gotoService: function (event) {
          if (event && event.context) {
            this.get('parentView').hide();
            App.router.transitionTo('main.services.service.summary', event.context);
          }
        },

        /**
         * Router transition to host level alerts page
         * @param event
         */
        goToHostAlerts: function (event) {
          if (event && event.context) {
            this.get('parentView').hide();
            App.router.transitionTo('main.hosts.hostDetails.alerts', event.context);
          }
        },

        didInsertElement: function () {
          this.filter();
          this.ensureTooltip();
          this.addObserver('filteringComplete', this, this.overlayObserver);
          this.overlayObserver();
          return this._super();
        }
      })
    });
  }

});

});

require.register("controllers/main/alerts/definition_configs_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

App.MainAlertDefinitionConfigsController = Em.Controller.extend({

  name: 'mainAlertDefinitionConfigsController',

  /**
   * All configurable properties of alert definition
   * @type {Array}
   */
  configs: [],

  /**
   * Define whether configs are editable
   * binds to property populated in template
   * @type {Boolean}
   */
  canEdit: true,

  /**
   * Define configs view mode (Wizard or Definition Details page)
   * @type {Boolean}
   */
  isWizard: false,

  /**
   * Alert Definition type
   * binding is set in template
   * @type {String}
   */
  alertDefinitionType: '',

  /**
   * Array of displayNames of all services
   * is used for "Service" config options
   * @type {Array}
   */
  allServices: function () {
    return App.Service.find().mapProperty('displayName');
  }.property(),

  /**
   * All possible values for scope propery
   * @type {Array}
   */
  allScopes: ['Any', 'Host', 'Service'],

  /**
   * Array of all aggregate-alerts names
   * @type {Array}
   */
  aggregateAlertNames: function () {
    return App.AggregateAlertDefinition.find().mapProperty('name');
  }.property(),

  /**
   * Change options of "Component", after changing value of "Service" config
   * @method onServiceSelect
   */
  onServiceSelect: function () {
    var serviceProperty = this.get('configs').findProperty('name', 'service');
    if (serviceProperty && serviceProperty.get('value') !== 'Ambari') {
      var componentsProperty = this.get('configs').findProperty('name', 'component');
      componentsProperty.set('options', ['No component'].concat(App.HostComponent.find().filterProperty('service.displayName', serviceProperty.get('value')).mapProperty('displayName').uniq()));
    }
  }.observes('configs.@each.value'),

  /**
   * OnSelect handler for <code>select_type</code> property
   * disable fields related to definition type and set options to select lists
   */
  changeType: function (selectedType) {
    if (selectedType === 'alert_type_service') {
      this.get('configs').findProperty('name', 'service').setProperties({
        isDisabled: false,
        options: this.get('allServices'),
        value: this.get('allServices')[0]
      });
      this.get('configs').findProperty('name', 'component').setProperties({
        isDisabled: false,
        value: 'No component'
      });
      this.get('configs').findProperty('name', 'scope').setProperties({
        isDisabled: false,
        options: this.get('allScopes'),
        value: this.get('allScopes')[0]
      });
    } else {
      this.get('configs').findProperty('name', 'service').setProperties({
        isDisabled: true,
        options: ['Ambari'],
        value: 'Ambari'
      });
      this.get('configs').findProperty('name', 'component').setProperties({
        isDisabled: true,
        options: ['Ambari Agent'],
        value: 'Ambari Agent'
      });
      this.get('configs').findProperty('name', 'scope').setProperties({
        isDisabled: true,
        options: ['Host'],
        value: 'Host'
      });
    }
  },

  /**
   * @return {string|Null}
   * @method getThresholdsProperty
   */
  getThresholdsProperty: function (type, property) {
    var warning = this.get('content.reporting').findProperty('type', type);
    return warning && warning.get(property) ? warning.get(property) : null;
  },

  /**
   * Render array of configs for appropriate alert definition type
   * @method renderConfigs
   */
  renderConfigs: function () {
    var alertDefinitionType = this.get('alertDefinitionType');
    var configs = [];
    switch (alertDefinitionType) {
      case 'PORT':
        configs = this.renderPortConfigs();
        break;
      case 'METRIC':
        configs = this.renderMetricConfigs();
        break;
      case 'WEB':
        configs = this.renderWebConfigs();
        break;
      case 'SCRIPT':
        configs = this.renderScriptConfigs();
        break;
      case 'AGGREGATE':
        configs = this.renderAggregateConfigs();
        break;
      default:
        console.error('Incorrect Alert Definition Type: ', alertDefinitionType);
    }

    configs.setEach('isDisabled', !this.get('canEdit'));

    this.set('configs', configs);
  },

  /**
   * Render config properties for port-type alert definition
   * @method renderPortConfigs
   * @returns {App.AlertConfigProperty[]}
   */
  renderPortConfigs: function () {
    var result = [];
    var alertDefinition = this.get('content');
    var isWizard = this.get('isWizard');

    if (this.get('isWizard')) {
      result = result.concat(this.renderCommonWizardConfigs());
    }

    result = result.concat([
      App.AlertConfigProperties.Description.create({
        value: isWizard ? '' : alertDefinition.get('description')
      }),
      App.AlertConfigProperties.Interval.create({
        value: isWizard ? '' : alertDefinition.get('interval')
      }),
      App.AlertConfigProperties.Thresholds.OkThreshold.create({
        label: 'Thresholds',
        showInputForValue: false,
        text: isWizard ? '' : this.getThresholdsProperty('ok', 'text'),
        value: isWizard ? '' : this.getThresholdsProperty('ok', 'value')
      }),
      App.AlertConfigProperties.Thresholds.WarningThreshold.create(App.AlertConfigProperties.Thresholds.PositiveMixin, {
        valueMetric: 'Seconds',
        text: isWizard ? '' : this.getThresholdsProperty('warning', 'text'),
        value: isWizard ? '' : this.getThresholdsProperty('warning', 'value')
      }),
      App.AlertConfigProperties.Thresholds.CriticalThreshold.create(App.AlertConfigProperties.Thresholds.PositiveMixin, {
        valueMetric: 'Seconds',
        text: isWizard ? '' : this.getThresholdsProperty('critical', 'text'),
        value: isWizard ? '' : this.getThresholdsProperty('critical', 'value')
      })
    ]);

    return result;
  },

  /**
   * Render config properties for metric-type alert definition
   * @method renderMetricConfigs
   * @returns {App.AlertConfigProperty[]}
   */
  renderMetricConfigs: function () {
    var result = [];
    var alertDefinition = this.get('content');
    var isWizard = this.get('isWizard');
    var units = this.get('content.reporting').findProperty('type','units') ?
      this.get('content.reporting').findProperty('type','units').get('text'): null;

    if (this.get('isWizard')) {
      result = result.concat(this.renderCommonWizardConfigs());
    }

    result = result.concat([
      App.AlertConfigProperties.Description.create({
        value: isWizard ? '' : alertDefinition.get('description')
      }),
      App.AlertConfigProperties.Interval.create({
        value: isWizard ? '' : alertDefinition.get('interval')
      }),
      App.AlertConfigProperties.Thresholds.OkThreshold.create({
        label: 'Thresholds',
        showInputForValue: false,
        text: isWizard ? '' : this.getThresholdsProperty('ok', 'text'),
        value: isWizard ? '' : this.getThresholdsProperty('ok', 'value')
      }),
      App.AlertConfigProperties.Thresholds.WarningThreshold.create({
        valueMetric: units,
        text: isWizard ? '' : this.getThresholdsProperty('warning', 'text'),
        value: isWizard ? '' : this.getThresholdsProperty('warning', 'value')
      }),
      App.AlertConfigProperties.Thresholds.CriticalThreshold.create({
        valueMetric: units,
        text: isWizard ? '' : this.getThresholdsProperty('critical', 'text'),
        value: isWizard ? '' : this.getThresholdsProperty('critical', 'value')
      })
    ]);

    return result;
  },

  /**
   * Render config properties for web-type alert definition
   * @method renderWebConfigs
   * @returns {App.AlertConfigProperty[]}
   */
  renderWebConfigs: function () {
    var result = [];
    var alertDefinition = this.get('content');
    var isWizard = this.get('isWizard');

    if (this.get('isWizard')) {
      result = result.concat(this.renderCommonWizardConfigs());
    }

    result = result.concat([
      App.AlertConfigProperties.Description.create({
        value: isWizard ? '' : alertDefinition.get('description')
      }),
      App.AlertConfigProperties.Interval.create({
        value: isWizard ? '' : alertDefinition.get('interval')
      }),
      App.AlertConfigProperties.Thresholds.OkThreshold.create({
        label: 'Thresholds',
        showInputForValue: false,
        text: isWizard ? '' : this.getThresholdsProperty('ok', 'text'),
        value: isWizard ? '' : this.getThresholdsProperty('ok', 'value')
      }),
      App.AlertConfigProperties.Thresholds.WarningThreshold.create({
        showInputForValue: false,
        text: isWizard ? '' : this.getThresholdsProperty('warning', 'text'),
        value: isWizard ? '' : this.getThresholdsProperty('warning', 'value')
      }),
      App.AlertConfigProperties.Thresholds.CriticalThreshold.create({
        showInputForValue: false,
        text: isWizard ? '' : this.getThresholdsProperty('critical', 'text'),
        value: isWizard ? '' : this.getThresholdsProperty('critical', 'value')
      })
    ]);

    return result;
  },

  /**
   * Render config properties for script-type alert definition
   * @method renderScriptConfigs
   * @returns {App.AlertConfigProperty[]}
   */
  renderScriptConfigs: function () {
    var result = [];
    var alertDefinition = this.get('content');
    var isWizard = this.get('isWizard');

    if (this.get('isWizard')) {
      result = result.concat(this.renderCommonWizardConfigs());
    }

    result = result.concat([
      App.AlertConfigProperties.Description.create({
        value: isWizard ? '' : alertDefinition.get('description')
      }),
      App.AlertConfigProperties.Interval.create({
        value: isWizard ? '' : alertDefinition.get('interval')
      })
    ]);

    return result;
  },

  /**
   * Render config properties for aggregate-type alert definition
   * @method renderAggregateConfigs
   * @returns {App.AlertConfigProperty[]}
   */
  renderAggregateConfigs: function () {
    var isWizard = this.get('isWizard');
    var alertDefinition = this.get('content');
    return [
      App.AlertConfigProperties.Description.create({
        value: isWizard ? '' : alertDefinition.get('description')
      }),
      App.AlertConfigProperties.Interval.create({
        value: isWizard ? '' : alertDefinition.get('interval')
      }),
      App.AlertConfigProperties.Thresholds.OkThreshold.create({
        label: 'Thresholds',
        showInputForValue: false,
        text: isWizard ? '' : this.getThresholdsProperty('ok', 'text'),
        value: isWizard ? '' : this.getThresholdsProperty('ok', 'value')
      }),
      App.AlertConfigProperties.Thresholds.WarningThreshold.create(App.AlertConfigProperties.Thresholds.PercentageMixin, {
        text: isWizard ? '' : this.getThresholdsProperty('warning', 'text'),
        value: isWizard ? '' : this.getThresholdsProperty('warning', 'value'),
        valueMetric: '%'
      }),
      App.AlertConfigProperties.Thresholds.CriticalThreshold.create(App.AlertConfigProperties.Thresholds.PercentageMixin, {
        text: isWizard ? '' : this.getThresholdsProperty('critical', 'text'),
        value: isWizard ? '' : this.getThresholdsProperty('critical', 'value'),
        valueMetric: '%'
      })
    ];
  },

  /**
   * Render common list of configs used in almost all alert types in wizard
   * @returns {App.AlertConfigProperty[]}
   */
  renderCommonWizardConfigs: function () {
    return [
      App.AlertConfigProperties.AlertName.create({
        value: ''
      }),
      App.AlertConfigProperties.ServiceAlertType.create({
        value: true
      }),
      App.AlertConfigProperties.Service.create({
        options: this.get('allServices'),
        value: this.get('allServices')[0],
        isShifted: true
      }),
      App.AlertConfigProperties.Component.create({
        options: this.get('allComponents'),
        value: 'No component',
        isShifted: true
      }),
      App.AlertConfigProperties.Scope.create({
        options: this.get('allScopes'),
        isShifted: true
      }),
      App.AlertConfigProperties.HostAlertType.create({
        value: false
      })
    ];
  },

  /**
   * Edit configs button handler
   * @method editConfigs
   */
  editConfigs: function () {
    this.get('configs').forEach(function (property) {
      property.set('previousValue', property.get('value'));
      property.set('previousText', property.get('text'));
    });
    this.get('configs').setEach('isDisabled', false);
    this.set('canEdit', true);
  },

  /**
   * Cancel edit configs button handler
   * @method cancelEditConfigs
   */
  cancelEditConfigs: function () {
    this.get('configs').forEach(function (property) {
      property.set('value', property.get('previousValue'));
      property.set('text', property.get('previousText'));
    });
    this.get('configs').setEach('isDisabled', true);
    this.set('canEdit', false);
  },

  /**
   * Save edit configs button handler
   * @method saveConfigs
   * @return {$.ajax}
   */
  saveConfigs: function () {
    this.get('configs').setEach('isDisabled', true);
    this.set('canEdit', false);

    return App.ajax.send({
      name: 'alerts.update_alert_definition',
      sender: this,
      data: {
        id: this.get('content.id'),
        data: this.getPropertiesToUpdate(true)
      },
      success: 'saveConfigsSuccessCallback'
    });
  },

  /**
   * Success-callback for saveConfigs-request
   * @method saveConfigsSuccessCallback
   */
  saveConfigsSuccessCallback: function () {
    App.router.get('updateController').updateAlertDefinitions(Em.K);
  },

  /**
   * Create object with new values to put it on server
   * @param {boolean} onlyChanged
   * @method getPropertiesToUpdate
   * @returns {Object}
   */
  getPropertiesToUpdate: function (onlyChanged) {
    var propertiesToUpdate = {};
    var configs = onlyChanged ? this.get('configs').filterProperty('wasChanged') : this.get('configs');
    configs.forEach(function (property) {
      var apiProperties = property.get('apiProperty');
      var apiFormattedValues = property.get('apiFormattedValue');
      if (!Em.isArray(property.get('apiProperty'))) {
        apiProperties = [property.get('apiProperty')];
        apiFormattedValues = [property.get('apiFormattedValue')];
      }
      apiProperties.forEach(function (apiProperty, i) {
        if (apiProperty.contains('source.')) {
          if (!propertiesToUpdate['AlertDefinition/source']) {
            if (this.get('content.rawSourceData')) {
              propertiesToUpdate['AlertDefinition/source'] = this.get('content.rawSourceData');
            }
          }

          if (this.get('content.rawSourceData')) {
            // use rawSourceData to populate propertiesToUpdate
            var sourcePath = propertiesToUpdate['AlertDefinition/source'];
            apiProperty.replace('source.', '').split('.').forEach(function (path, index, array) {
              // check if it is last path
              if (array.length - index === 1) {
                sourcePath[path] = apiFormattedValues[i];
              } else {
                sourcePath = sourcePath[path];
              }
            });
          }
          else {
            if (!propertiesToUpdate['AlertDefinition/source']) {
              propertiesToUpdate['AlertDefinition/source'] = {};
            }
            Ember.setFullPath(propertiesToUpdate['AlertDefinition/source'], apiProperty.replace('source.', ''), apiFormattedValues[i]);
          }

        }
        else {
          if (apiProperty) {
            propertiesToUpdate['AlertDefinition/' + apiProperty] = apiFormattedValues[i];
          }
        }
      }, this);
    }, this);

    return propertiesToUpdate;
  },

  /**
   * Return array of all config values
   * used to save configs to local db in wizard
   * @method getConfigsValues
   * @returns {Array}
   */
  getConfigsValues: function () {
    return this.get('configs').map(function (property) {
      return {
        name: property.get('name'),
        value: property.get('value')
      }
    });
  },

  /**
   * Define whether critical threshold >= critical threshold
   * @type {Boolean}
   */
  hasThresholdsError: function () {
    if (this.get('configs').findProperty('name', 'warning_threshold')) {
      var smallValue = Em.get(this.get('configs').findProperty('name', 'warning_threshold'), 'value');
      var smallValid = Em.get(this.get('configs').findProperty('name', 'warning_threshold'), 'isValid');
    }
    if (this.get('configs').findProperty('name', 'critical_threshold')) {
      var largeValue = Em.get(this.get('configs').findProperty('name', 'critical_threshold'), 'value');
      var largeValid = Em.get(this.get('configs').findProperty('name', 'critical_threshold'), 'isValid');
    }
    return smallValid && largeValid ? Number(smallValue) > Number(largeValue) : false;
  }.property('configs.@each.value'),

  /**
   * Define whether all configs are valid
   * @type {Boolean}
   */
  hasErrors: function () {
    return this.get('configs').someProperty('isValid', false) || this.get('hasThresholdsError');
  }.property('configs.@each.isValid', 'hasThresholdsError')

});

});

require.register("controllers/main/alerts/definition_details_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

App.MainAlertDefinitionDetailsController = Em.Controller.extend({

  name: 'mainAlertDefinitionDetailsController',

  alerts: function () {
    return App.AlertInstance.find().toArray()
        .filterProperty('definitionId', this.get('content.id'));
  }.property('App.router.mainAlertInstancesController.isLoaded', 'App.router.mainAlertInstancesController.reload'),

  // stores object with editing form data (label)
  editing: Em.Object.create({
    label: Em.Object.create({
      name: 'label',
      isEditing: false,
      value: '',
      originalValue: '',
      isError: false,
      bindingValue: 'content.label'
    })
  }),

  /**
   * Host to count of alerts on this host during last day map
   * @type {Object}
   */
  lastDayAlertsCount: null,

  /**
   * Define if let user leave the page
   * @type {Boolean}
   */
  forceTransition: false,

  /**
   * List of all group names related to alert definition
   * @type {Array}
   */
  groupsList: function () {
    return this.get('content.groups').mapProperty('displayName');
  }.property('content.groups.@each'),

  /**
   * Validation function to define if label field populated correctly
   * @method labelValidation
   */
  labelValidation: function () {
    this.set('editing.label.isError', !this.get('editing.label.value').trim());
  }.observes('editing.label.value'),

  /**
   * Set init values for variables
   */
  clearStep: function () {
    var editing = this.get('editing');
    Em.keys(editing).forEach(function (key) {
      editing.get(key).set('isEditing', false);
    });
  },

  /**
   * Load alert instances for current alertDefinition
   * Start updating loaded data
   * @method loadAlertInstances
   */
  loadAlertInstances: function () {
    App.router.get('mainAlertInstancesController').loadAlertInstancesByAlertDefinition(this.get('content.id'));
    App.router.set('mainAlertInstancesController.isUpdating', true);
    this.loadAlertInstancesHistory();
  },

  /**
   * Load alert instances history data
   * used to count instances number of the last 24 hour
   * @method loadAlertInstancesHistory
   */
  loadAlertInstancesHistory: function () {
    this.set('lastDayAlertsCount', null);
    return App.ajax.send({
      name: 'alerts.get_instances_history',
      sender: this,
      data: {
        definitionName: this.get('content.name'),
        timestamp: App.dateTime() - 86400000 // timestamp for time 24-hours ago
      },
      success: 'loadAlertInstancesHistorySuccess'
    });
  },

  /**
   * Success-callback for <code>loadAlertInstancesHistory</code>
   */
  loadAlertInstancesHistorySuccess: function (data) {
    var lastDayAlertsCount = {};
    data.items.forEach(function (alert) {
      if (!lastDayAlertsCount[alert.AlertHistory.host_name]) {
        lastDayAlertsCount[alert.AlertHistory.host_name] = 1;
      } else {
        lastDayAlertsCount[alert.AlertHistory.host_name] += 1;
      }
    });
    this.set('lastDayAlertsCount', lastDayAlertsCount);
  },

  /**
   * Edit button handler
   * @param {object} event
   * @method edit
   */
  edit: function (event) {
    var element = event.context;
    var value = this.get(element.get('bindingValue'));
    element.set('originalValue', value);
    element.set('value', value);
    element.set('isEditing', true);
  },

  /**
   * Cancel button handler
   * @param {object} event
   * @method cancelEdit
   */
  cancelEdit: function (event) {
    var element = event.context;
    element.set('value', element.get('originalValue'));
    element.set('isEditing', false);
  },

  /**
   * Save button handler, could save label of alert definition
   * @param {object} event
   * @returns {$.ajax}
   * @method saveEdit
   */
  saveEdit: function (event) {
    var element = event.context;
    this.set(element.get('bindingValue'), element.get('value'));
    element.set('isEditing', false);

    var data = Em.Object.create({});
    var property_name = "AlertDefinition/" + element.get('name');
    data.set(property_name, element.get('value'));
    var alertDefinition_id = this.get('content.id');
    return App.ajax.send({
      name: 'alerts.update_alert_definition',
      sender: this,
      data: {
        id: alertDefinition_id,
        data: data
      }
    });
  },

  /**
   * Onclick handler for save button on Save popup
   * Save changes of label and configs
   */
  saveLabelAndConfigs: function () {
    var configsController = App.router.get('mainAlertDefinitionConfigsController');
    if (configsController.get('canEdit')) {
      configsController.saveConfigs();
    }
    if (this.get('editing.label.isEditing')) {
      this.saveEdit({
        context: this.get('editing.label')
      });
    }
  },

  /**
   * "Delete" button handler
   * @param {object} event
   * @method deleteAlertDefinition
   */
  deleteAlertDefinition: function (event) {
    var alertDefinition = this.get('content');
    var self = this;
    App.showConfirmationPopup(function () {
      App.ajax.send({
        name: 'alerts.delete_alert_definition',
        sender: self,
        success: 'deleteAlertDefinitionSuccess',
        error: 'deleteAlertDefinitionError',
        data: {
          id: alertDefinition.get('id')
        }
      });
    }, null, function () {
    });
  },

  /**
   * Success-callback for <code>deleteAlertDefinition</code>
   * @method deleteAlertDefinitionSuccess
   */
  deleteAlertDefinitionSuccess: function () {
    App.router.transitionTo('main.alerts.index');
  },

  /**
   * Error-callback for <code>deleteAlertDefinition</code>
   * @method deleteAlertDefinitionError
   */
  deleteAlertDefinitionError: function (xhr, textStatus, errorThrown, opt) {
    console.log(textStatus);
    console.log(errorThrown);
    xhr.responseText = "{\"message\": \"" + xhr.statusText + "\"}";
    App.ajax.defaultErrorHandler(xhr, opt.url, 'DELETE', xhr.status);
  },

  /**
   * "Disable / Enable" button handler
   * @method toggleState
   */
  toggleState: function () {
    var alertDefinition = this.get('content');
    var self = this;
    var bodyMessage = Em.Object.create({
      confirmMsg: alertDefinition.get('enabled') ? Em.I18n.t('alerts.table.state.enabled.confirm.msg') : Em.I18n.t('alerts.table.state.disabled.confirm.msg'),
      confirmButton: alertDefinition.get('enabled') ? Em.I18n.t('alerts.table.state.enabled.confirm.btn') : Em.I18n.t('alerts.table.state.disabled.confirm.btn')
    });

    return App.showConfirmationFeedBackPopup(function (query) {
      self.toggleDefinitionState(alertDefinition);
    }, bodyMessage);
  },

  /**
   * Enable/disable alertDefinition
   * @param {object} alertDefinition
   * @returns {$.ajax}
   * @method toggleDefinitionState
   */
  toggleDefinitionState: function (alertDefinition) {
    var newState = !alertDefinition.get('enabled');
    alertDefinition.set('enabled', newState);
    return App.ajax.send({
      name: 'alerts.update_alert_definition',
      sender: this,
      data: {
        id: alertDefinition.get('id'),
        data: {
          "AlertDefinition/enabled": newState
        }
      }
    });
  },

  /**
   * Router transition to service page
   * @param event
   */
  goToService: function (event) {
    if (event && event.context) {
      App.router.transitionTo('main.services.service.summary', event.context);
    }
  },

  /**
   * Router transition to host level alerts page
   * @param {object} event
   * @method goToHostAlerts
   */
  goToHostAlerts: function (event) {
    if (event && event.context) {
      App.router.get('mainHostDetailsController').set('referer', App.router.location.lastSetURL);
      App.router.transitionTo('main.hosts.hostDetails.alerts', event.context);
    }
  },

  /**
   * Define if label or configs are in edit mode
   * @type {Boolean}
   */
  isEditing: function () {
    return this.get('editing.label.isEditing') || App.router.get('mainAlertDefinitionConfigsController.canEdit');
  }.property('editing.label.isEditing', 'App.router.mainAlertDefinitionConfigsController.canEdit'),

  /**
   * If some configs or label are changed and user navigates away, show this popup with propose to save changes
   * @param {String} path
   * @method showSavePopup
   */
  showSavePopup: function (path) {
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('common.warning'),
      bodyClass: Em.View.extend({
        template: Ember.Handlebars.compile('{{t alerts.saveChanges}}')
      }),
      primary: Em.I18n.t('common.save'),
      secondary: Em.I18n.t('common.discard'),
      third: Em.I18n.t('common.cancel'),
      disablePrimary: function () {
        return App.router.get('mainAlertDefinitionDetailsController.editing.label.isError') || App.router.get('mainAlertDefinitionConfigsController.hasErrors');
      }.property('App.router.mainAlertDefinitionDetailsController.editing.label.isError', 'App.router.mainAlertDefinitionConfigsController.hasErrors'),
      onPrimary: function () {
        self.saveLabelAndConfigs();
        self.set('forceTransition', true);
        App.router.route(path);
        this.hide();
      },
      onSecondary: function () {
        self.set('forceTransition', true);
        App.router.route(path);
        this.hide();
      },
      onThird: function () {
        this.hide();
      }
    });
  }
});
});

require.register("controllers/main/alerts/manage_alert_groups_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

var validator = require('utils/validator');
var numberUtils = require('utils/number_utils');

App.ManageAlertGroupsController = Em.Controller.extend({

  name: 'manageAlertGroupsController',

  /**
   * @type {boolean}
   */
  isLoaded: false,

  /**
   * Property used to trigger Alert Groups Filter content updating
   * @type {Boolean}
   */
  changeTrigger: false,

  /**
   * @type {App.AlertGroup[]}
   */
  alertGroups: [],

  /**
   * @type {App.AlertGroup[]}
   */
  originalAlertGroups: [],

  /**
   * @type {App.AlertGroup}
   */
  selectedAlertGroup: null,

  /**
   * @type {App.AlertDefinition[]}
   */
  selectedDefinitions: [],

  /**
   * List of all Alert Notifications
   * @type {App.AlertNotification[]}
   */
  alertNotifications: function () {
    return this.get('isLoaded') ? App.AlertNotification.find().map(function (target) {
      return Em.Object.create({
        name: target.get('name'),
        id: target.get('id'),
        description: target.get('description'),
        type: target.get('type'),
        global: target.get('global')
      });
    }) : [];
  }.property('isLoaded'),

  /**
   * List of all global Alert Notifications
   * @type {App.AlertNotification[]}
   */
  alertGlobalNotifications: function () {
    return this.get('alertNotifications').filterProperty('global');
  }.property('alertNotifications'),

  /**
   * @type {boolean}
   */
  isRemoveButtonDisabled: true,

  /**
   * @type {boolean}
   */
  isRenameButtonDisabled: true,

  /**
   * @type {boolean}
   */
  isDuplicateButtonDisabled: true,

  /**
   * @type {boolean}
   */
  isDeleteDefinitionsDisabled: function () {
    var selectedGroup = this.get('selectedAlertGroup');
    return selectedGroup ? (selectedGroup.default || this.get('selectedDefinitions').length === 0) : true;
  }.property('selectedAlertGroup', 'selectedAlertGroup.definitions.length', 'selectedDefinitions.length'),

  /**
   * observes if any group changed including: group name, newly created group, deleted group, group with definitions/notifications changed
   * @type {{toDelete: App.AlertGroup[], toSet: App.AlertGroup[], toCreate: App.AlertGroup[]}}
   */
  defsModifiedAlertGroups: function () {
    if (!this.get('isLoaded')) {
      return false;
    }
    var groupsToDelete = [];
    var groupsToSet = [];
    var groupsToCreate = [];
    var groups = this.get('alertGroups'); //current alert groups
    var originalGroups = this.get('originalAlertGroups'); // original alert groups
    var originalGroupsIds = originalGroups.mapProperty('id');
    groups.forEach(function (group) {
      var originalGroup = originalGroups.findProperty('id', group.get('id'));
      if (originalGroup) {
        // should update definitions or notifications
        if (!(JSON.stringify(group.get('definitions').slice().sort()) === JSON.stringify(originalGroup.get('definitions').slice().sort()))
          || !(JSON.stringify(group.get('notifications').slice().sort()) === JSON.stringify(originalGroup.get('notifications').slice().sort()))) {
          groupsToSet.push(group.set('id', originalGroup.get('id')));
        } else if (group.get('name') !== originalGroup.get('name')) {
          // should update name
          groupsToSet.push(group.set('id', originalGroup.get('id')));
        }
        originalGroupsIds = originalGroupsIds.without(group.get('id'));
      } else {
        // should add new group
        groupsToCreate.push(group);
      }
    });
    // should delete groups
    originalGroupsIds.forEach(function (id) {
      groupsToDelete.push(originalGroups.findProperty('id', id));
    }, this);
    return {
      toDelete: groupsToDelete,
      toSet: groupsToSet,
      toCreate: groupsToCreate
    };
  }.property('selectedAlertGroup.definitions.@each', 'selectedAlertGroup.definitions.length', 'selectedAlertGroup.notifications.@each', 'selectedAlertGroup.notifications.length', 'alertGroups', 'isLoaded'),

  /**
   * Determines if some group was edited/created/deleted
   * @type {boolean}
   */
  isDefsModified: function () {
    var modifiedGroups = this.get('defsModifiedAlertGroups');
    if (!this.get('isLoaded')) {
      return false;
    }
    return !!(modifiedGroups.toSet.length || modifiedGroups.toCreate.length || modifiedGroups.toDelete.length);
  }.property('defsModifiedAlertGroups'),

  /**
   * Load all Alert Notifications from server
   * @returns {$.ajax}
   * @method loadAlertNotifications
   */
  loadAlertNotifications: function () {
    this.set('isLoaded', false);
    this.set('alertGroups', []);
    this.set('originalAlertGroups', []);
    this.set('selectedAlertGroup', null);
    this.set('isRemoveButtonDisabled', true);
    this.set('isRenameButtonDisabled', true);
    this.set('isDuplicateButtonDisabled', true);
    return App.ajax.send({
      name: 'alerts.notifications',
      sender: this,
      success: 'getAlertNotificationsSuccessCallback',
      error: 'getAlertNotificationsErrorCallback'
    });
  },

  /**
   * Success-callback for load alert notifications request
   * @param {object} json
   * @method getAlertNotificationsSuccessCallback
   */
  getAlertNotificationsSuccessCallback: function (json) {
    App.alertNotificationMapper.map(json);
    this.loadAlertGroups();
  },

  /**
   * Error-callback for load alert notifications request
   * @method getAlertNotificationsErrorCallback
   */
  getAlertNotificationsErrorCallback: function () {
    this.set('isLoaded', true);
  },

  /**
   * Load all alert groups from alert group model
   * @method loadAlertGroups
   */
  loadAlertGroups: function () {
    var alertGroups = App.AlertGroup.find().map(function (group) {
      var definitions = group.get('definitions').map(function (def) {
        return Em.Object.create({
          name: def.get('name'),
          serviceName: def.get('serviceName'),
          componentName: def.get('componentName'),
          serviceNameDisplay: def.get('service.displayName'),
          componentNameDisplay: def.get('componentNameFormatted'),
          label: def.get('label'),
          id: def.get('id')
        });
      });

      var targets = group.get('targets').map(function (target) {
        return Em.Object.create({
          name: target.get('name'),
          id: target.get('id'),
          description: target.get('description'),
          type: target.get('type'),
          global: target.get('global')
        });
      });

      return Em.Object.create({
        id: group.get('id'),
        name: group.get('name'),
        default: group.get('default'),
        displayName: function () {
          var name = this.get('name');
          if (name && name.length > App.config.CONFIG_GROUP_NAME_MAX_LENGTH) {
            var middle = Math.floor(App.config.CONFIG_GROUP_NAME_MAX_LENGTH / 2);
            name = name.substring(0, middle) + "..." + name.substring(name.length - middle);
          }
          return this.get('default') ? (name + ' Default') : name;
        }.property('name', 'default'),
        label: function () {
          return this.get('displayName') + ' (' + this.get('definitions.length') + ')';
        }.property('displayName', 'definitions.length'),
        definitions: definitions,
        isAddDefinitionsDisabled: group.get('isAddDefinitionsDisabled'),
        notifications: targets
      });
    });
    this.set('alertGroups', alertGroups);
    this.set('isLoaded', true);
    this.set('originalAlertGroups', this.copyAlertGroups(this.get('alertGroups')));
    this.set('selectedAlertGroup', this.get('alertGroups')[0]);
  },

  /**
   * Enable/disable "Remove"/"Rename"/"Duplicate" buttons basing on <code>controller.selectedAlertGroup</code>
   * @method buttonObserver
   */
  buttonObserver: function () {
    var selectedAlertGroup = this.get('selectedAlertGroup');
    var flag = selectedAlertGroup && selectedAlertGroup.get('default');
    this.set('isRemoveButtonDisabled', flag);
    this.set('isRenameButtonDisabled', flag);
    this.set('isDuplicateButtonDisabled', false);
  }.observes('selectedAlertGroup'),

  /**
   * @method resortAlertGroup
   */
  resortAlertGroup: function () {
    var alertGroups = Em.copy(this.get('alertGroups'));
    if (alertGroups.length < 2) {
      return;
    }
    var defaultGroups = alertGroups.filterProperty('default');
    defaultGroups.forEach(function (defaultGroup) {
      alertGroups.removeObject(defaultGroup);
    });
    var sorted = defaultGroups.sortProperty('name').concat(alertGroups.sortProperty('name'));

    this.removeObserver('alertGroups.@each.name', this, 'resortAlertGroup');
    this.set('alertGroups', sorted);
    this.addObserver('alertGroups.@each.name', this, 'resortAlertGroup');
  }.observes('alertGroups.@each.name'),

  /**
   * remove definitions from group
   * @method deleteDefinitions
   */
  deleteDefinitions: function () {
    if (this.get('isDeleteDefinitionsDisabled')) {
      return;
    }
    var groupDefinitions = this.get('selectedAlertGroup.definitions');
    this.get('selectedDefinitions').slice().forEach(function (defObj) {
      groupDefinitions.removeObject(defObj);
    }, this);
    this.set('selectedDefinitions', []);
  },

  /**
   * Provides alert definitions which are available for inclusion in
   * non-default alert groups.
   * @param {App.AlertGroup} selectedAlertGroup
   * @method getAvailableDefinitions
   * @return {{name: string, serviceName: string, componentName: string, serviceNameDisplay: string, componentNameDisplay: string, label: string, id: number}[]}
   */
  getAvailableDefinitions: function (selectedAlertGroup) {
    if (selectedAlertGroup.get('default')) return [];
    var usedDefinitionsMap = {};
    var availableDefinitions = [];
    var sharedDefinitions = App.AlertDefinition.find();

    selectedAlertGroup.get('definitions').forEach(function (def) {
      usedDefinitionsMap[def.name] = true;
    });
    sharedDefinitions.forEach(function (shared_def) {
      if (!usedDefinitionsMap[shared_def.get('name')]) {
        availableDefinitions.pushObject(shared_def);
      }
    });
    return availableDefinitions.map(function (def) {
      return Em.Object.create({
        name: def.get('name'),
        serviceName: def.get('serviceName'),
        componentName: def.get('componentName'),
        serviceNameDisplay: def.get('service.displayName'),
        componentNameDisplay: def.get('componentNameFormatted'),
        label: def.get('label'),
        id: def.get('id')
      });
    });
  },

  /**
   * add alert definitions to a group
   * @method addDefinitions
   */
  addDefinitions: function () {
    if (this.get('selectedAlertGroup.isAddDefinitionsDisabled')) {
      return false;
    }
    var availableDefinitions = this.getAvailableDefinitions(this.get('selectedAlertGroup'));
    var popupDescription = {
      header: Em.I18n.t('alerts.actions.manage_alert_groups_popup.selectDefsDialog.title'),
      dialogMessage: Em.I18n.t('alerts.actions.manage_alert_groups_popup.selectDefsDialog.message').format(this.get('selectedAlertGroup.displayName'))
    };
    var validComponents = App.StackServiceComponent.find().map(function (component) {
      return Em.Object.create({
        componentName: component.get('componentName'),
        displayName: App.format.role(component.get('componentName')),
        selected: false
      });
    });
    var validServices = App.Service.find().map(function (service) {
      return Em.Object.create({
        serviceName: service.get('serviceName'),
        displayName: App.format.role(service.get('serviceName')),
        selected: false
      });
    });
    this.launchDefsSelectionDialog(availableDefinitions, [], validServices, validComponents, this.addDefinitionsCallback.bind(this), popupDescription);
  },

  /**
   * Launch a table view of all available definitions to choose
   * @method launchDefsSelectionDialog
   * @return {App.ModalPopup}
   */
  launchDefsSelectionDialog: function (initialDefs, selectedDefs, validServices, validComponents, callback, popupDescription) {

    return App.ModalPopup.show({

      classNames: [ 'sixty-percent-width-modal' ],

      header: popupDescription.header,

      /**
       * @type {string}
       */
      dialogMessage: popupDescription.dialogMessage,

      /**
       * @type {string|null}
       */
      warningMessage: null,

      /**
       * @type {App.AlertDefinition[]}
       */
      availableDefs: [],

      onPrimary: function () {
        this.set('warningMessage', null);
        var arrayOfSelectedDefs = this.get('availableDefs').filterProperty('selected', true);
        if (arrayOfSelectedDefs.length < 1) {
          this.set('warningMessage', Em.I18n.t('alerts.actions.manage_alert_groups_popup.selectDefsDialog.message.warning'));
          return;
        }
        callback(arrayOfSelectedDefs);
        console.debug('(new-selectedDefs)=', arrayOfSelectedDefs);
        this.hide();
      },

      /**
       * Primary button should be disabled while alert definitions are not loaded
       * @type {boolean}
       */
      disablePrimary: function () {
        return !this.get('isLoaded');
      }.property('isLoaded'),

      onSecondary: function () {
        callback(null);
        this.hide();
      },

      bodyClass: App.SelectDefinitionsPopupBodyView.extend({

        filterComponents: validComponents,

        filterServices: validServices,

        initialDefs: initialDefs

      })

    });
  },

  /**
   * add alert definitions callback
   * @method addDefinitionsCallback
   */
  addDefinitionsCallback: function (selectedDefs) {
    var group = this.get('selectedAlertGroup');
    if (selectedDefs) {
      var alertGroupDefs = group.get('definitions');
      selectedDefs.forEach(function (defObj) {
        alertGroupDefs.pushObject(defObj);
      }, this);
    }
  },

  /**
   * copy alert groups for backup, to compare with current alert groups, so will know if some groups changed/added/deleted
   * @param {App.AlertGroup[]} originGroups
   * @return {App.AlertGroup[]}
   * @method copyAlertGroups
   */
  copyAlertGroups: function (originGroups) {
    var alertGroups = [];
    originGroups.forEach(function (alertGroup) {
      var copiedGroup = Em.Object.create($.extend(true, {}, alertGroup));
      alertGroups.pushObject(copiedGroup);
    });
    return alertGroups;
  },

  /**
   * Create a new alert group
   * @param {Em.Object} newAlertGroupData
   * @param {callback} callback Callback function for Success or Error handling
   * @return {App.AlertGroup} Returns the created alert group
   * @method postNewAlertGroup
   */
  postNewAlertGroup: function (newAlertGroupData, callback) {
    // create a new group with name , definition and notifications
    var data = {
      'name': newAlertGroupData.get('name')
    };
    if (newAlertGroupData.get('definitions').length > 0) {
      data.definitions = newAlertGroupData.get('definitions').mapProperty('id');
    }
    if (newAlertGroupData.get('notifications').length > 0) {
      data.targets = newAlertGroupData.get('notifications').mapProperty('id');
    }
    var sendData = {
      name: 'alert_groups.create',
      data: data,
      success: 'successFunction',
      error: 'errorFunction',
      successFunction: function () {
        if (callback) {
          callback();
        }
      },
      errorFunction: function (xhr, text, errorThrown) {
        if (callback) {
          callback(xhr, text, errorThrown);
        }
        console.error('Error in creating new Alert Group');
      }
    };
    sendData.sender = sendData;
    App.ajax.send(sendData);
    return newAlertGroupData;
  },

  /**
   * PUTs the new alert group information on the server.
   * Changes possible here are the name, definitions, notifications
   *
   * @param {App.AlertGroup} alertGroup
   * @param {Function} successCallback
   * @param {Function} errorCallback
   * @method updateAlertGroup
   */
  updateAlertGroup: function (alertGroup, successCallback, errorCallback) {
    var sendData = {
      name: 'alert_groups.update',
      data: {
        "group_id": alertGroup.id,
        'name': alertGroup.get('name'),
        'definitions': alertGroup.get('definitions').mapProperty('id'),
        'targets': alertGroup.get('notifications').mapProperty('id')
      },
      success: 'successFunction',
      error: 'errorFunction',
      successFunction: function () {
        if (successCallback) {
          successCallback();
        }
      },
      errorFunction: function (xhr, text, errorThrown) {
        if (errorCallback) {
          errorCallback(xhr, text, errorThrown);
        }
      }
    };
    sendData.sender = sendData;
    App.ajax.send(sendData);
  },

  /**
   * Request for deleting alert group
   * @param {App.AlertGroup} alertGroup
   * @param {callback} successCallback
   * @param {callback} errorCallback
   * @method removeAlertGroup
   */
  removeAlertGroup: function (alertGroup, successCallback, errorCallback) {
    var sendData = {
      name: 'alert_groups.delete',
      data: {
        "group_id": alertGroup.id
      },
      success: 'successFunction',
      error: 'errorFunction',
      successFunction: function () {
        if (successCallback) {
          successCallback();
        }
      },
      errorFunction: function (xhr, text, errorThrown) {
        if (errorCallback) {
          errorCallback(xhr, text, errorThrown);
        }
      }
    };
    sendData.sender = sendData;
    App.ajax.send(sendData);
  },

  /**
   * confirm delete alert group
   * @method confirmDelete
   */
  confirmDelete: function () {
    if (this.get('isRemoveButtonDisabled')) return;
    var self = this;
    App.showConfirmationPopup(function () {
      self.deleteAlertGroup();
    });
  },

  /**
   * delete selected alert group
   * @method deleteAlertGroup
   */
  deleteAlertGroup: function () {
    var selectedAlertGroup = this.get('selectedAlertGroup');
    if (this.get('isDeleteAlertDisabled')) {
      return;
    }
    this.get('alertGroups').removeObject(selectedAlertGroup);
    this.set('selectedAlertGroup', this.get('alertGroups')[0]);
  },

  /**
   * Rename non-default alert group
   * @method renameAlertGroup
   */
  renameAlertGroup: function () {

    if (this.get('selectedAlertGroup.default')) {
      return;
    }
    var self = this;
    var popup;
    popup = App.ModalPopup.show({

      header: Em.I18n.t('alerts.actions.manage_alert_groups_popup.renameButton'),

      bodyClass: Ember.View.extend({
        templateName: require('templates/main/alerts/create_new_alert_group')
      }),

      /**
       * @type {string}
       */
      alertGroupName: self.get('selectedAlertGroup.name'),

      /**
       * @type {string|null}
       */
      warningMessage: null,

      /**
       * New group name should be unique and valid
       * @method validate
       */
      validate: function () {
        var warningMessage = '';
        var originalGroup = self.get('selectedAlertGroup');
        var groupName = this.get('alertGroupName').trim();

        if (originalGroup.get('name').trim() === groupName) {
          warningMessage = Em.I18n.t("alerts.actions.manage_alert_groups_popup.addGroup.exist");
        }
        else {
          if (self.get('alertGroups').mapProperty('displayName').contains(groupName)) {
            warningMessage = Em.I18n.t("alerts.actions.manage_alert_groups_popup.addGroup.exist");
          }
          else {
            if (groupName && !validator.isValidAlertGroupName(groupName)) {
              warningMessage = Em.I18n.t("form.validator.alertGroupName");
            }
          }
        }
        this.set('warningMessage', warningMessage);
      }.observes('alertGroupName'),

      /**
       * Primary button is disabled while user doesn't input valid group name
       * @type {boolean}
       */
      disablePrimary: function () {
        return !(this.get('alertGroupName').trim().length > 0 && (this.get('warningMessage') !== null && !this.get('warningMessage')));
      }.property('warningMessage', 'alertGroupName'),

      onPrimary: function () {
        self.set('selectedAlertGroup.name', this.get('alertGroupName'));
        this.hide();
      }

    });
    this.set('renameGroupPopup', popup);
  },

  /**
   * Create new alert group
   * @param {boolean} duplicated is new group a copy of the existing group
   * @method addAlertGroup
   */
  addAlertGroup: function (duplicated) {
    duplicated = (duplicated === true);
    var self = this;
    var popup = App.ModalPopup.show({

      header: Em.I18n.t('alerts.actions.manage_alert_groups_popup.addButton'),

      bodyClass: Em.View.extend({
        templateName: require('templates/main/alerts/create_new_alert_group')
      }),

      /**
       * Name for new alert group
       * @type {string}
       */
      alertGroupName: duplicated ? self.get('selectedAlertGroup.name') + ' Copy' : "",

      /**
       * @type {string}
       */
      warningMessage: '',

      didInsertElement: function () {
        this._super();
        this.validate();
      },

      /**
       * alert group name should be unique and valid
       * @method validate
       */
      validate: function () {
        var warningMessage = '';
        var groupName = this.get('alertGroupName').trim();
        if (self.get('alertGroups').mapProperty('displayName').contains(groupName)) {
          warningMessage = Em.I18n.t("alerts.actions.manage_alert_groups_popup.addGroup.exist");
        }
        else {
          if (groupName && !validator.isValidAlertGroupName(groupName)) {
            warningMessage = Em.I18n.t("form.validator.alertGroupName");
          }
        }
        this.set('warningMessage', warningMessage);
      }.observes('alertGroupName'),

      /**
       * Primary button is disabled while user doesn't input valid group name
       * @type {boolean}
       */
      disablePrimary: function () {
        return !(this.get('alertGroupName').trim().length > 0 && !this.get('warningMessage'));
      }.property('warningMessage', 'alertGroupName'),

      onPrimary: function () {
        var newAlertGroup = Em.Object.create({
          name: this.get('alertGroupName').trim(),
          default: false,
          displayName: function () {
            var name = this.get('name');
            if (name && name.length > App.config.CONFIG_GROUP_NAME_MAX_LENGTH) {
              var middle = Math.floor(App.config.CONFIG_GROUP_NAME_MAX_LENGTH / 2);
              name = name.substring(0, middle) + "..." + name.substring(name.length - middle);
            }
            return this.get('default') ? (name + ' Default') : name;
          }.property('name', 'default'),
          label: function () {
            return this.get('displayName') + ' (' + this.get('definitions.length') + ')';
          }.property('displayName', 'definitions.length'),
          definitions: [],
          notifications: self.get('alertGlobalNotifications'),
          isAddDefinitionsDisabled: false
        });
        self.get('alertGroups').pushObject(newAlertGroup);
        self.set('selectedAlertGroup', newAlertGroup);
        this.hide();
      }

    });
    this.set('addGroupPopup', popup);
  },

  /**
   * @method duplicateAlertGroup
   */
  duplicateAlertGroup: function () {
    this.addAlertGroup(true);
  }

});

});

require.register("controllers/main/alerts/manage_alert_notifications_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var validator = require('utils/validator');

App.ManageAlertNotificationsController = Em.Controller.extend({

  name: 'manageAlertNotificationsController',

  /**
   * Are alert notifications loaded
   * @type {boolean}
   */
  isLoaded: false,

  /**
   * Create/Edit modal popup object
   * used to hide popup
   * @type {App.ModalPopup}
   */
  createEditPopup: null,

  /**
   * Map of edit inputs shown in Create/Edit Notification popup
   * @type {Object}
   */
  inputFields: Em.Object.create({
    name: {
      label: Em.I18n.t('common.name'),
      value: '',
      defaultValue: ''
    },
    groups: {
      label: Em.I18n.t('common.groups'),
      value: [],
      defaultValue: []
    },
    global: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.global'),
      value: false,
      defaultValue: false,
      disabled: false
    },
    allGroups: Em.Object.create({
      value: '',
      defaultValue: 'custom',
      disabled: false,
      isAll: function () {
        return this.get('value') == 'all';
      }.property('value')
    }),
    method: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.method'),
      value: '',
      defaultValue: ''
    },
    email: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.email'),
      value: '',
      defaultValue: ''
    },
    SMTPServer: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.SMTPServer'),
      value: '',
      defaultValue: ''
    },
    SMTPPort: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.SMTPPort'),
      value: '',
      defaultValue: ''
    },
    SMTPUseAuthentication: Em.Object.create({
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.SMTPUseAuthentication'),
      value: false,
      defaultValue: false,
      invertedValue: Em.computed.not('value')
    }),
    SMTPUsername: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.SMTPUsername'),
      value: '',
      defaultValue: ''
    },
    SMTPPassword: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.SMTPPassword'),
      value: '',
      defaultValue: ''
    },
    retypeSMTPPassword: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.retypeSMTPPassword'),
      value: '',
      defaultValue: ''
    },
    SMTPSTARTTLS: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.SMTPSTARTTLS'),
      value: false,
      defaultValue: false
    },
    emailFrom: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.emailFrom'),
      value: '',
      defaultValue: ''
    },
    version: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.version'),
      value: '',
      defaultValue: ''
    },
    OIDs: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.OIDs'),
      value: '',
      defaultValue: ''
    },
    community: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.community'),
      value: '',
      defaultValue: ''
    },
    port: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.port'),
      value: '',
      defaultValue: ''
    },
    severityFilter: {
      label: Em.I18n.t('alerts.actions.manage_alert_notifications_popup.severityFilter'),
      value: [],
      defaultValue: []
    },
    description: {
      label: Em.I18n.t('common.description'),
      value: '',
      defaultValue: ''
    },
    customProperties: Em.A([])
  }),

  /**
   * List of available Notification types
   * used in Type combobox
   * @type {Array}
   */
  methods: ['EMAIL', 'SNMP'],

  /**
   * List of available value for Severity Filter
   * used in Severity Filter combobox
   * @type {Array}
   */
  severities: ['OK', 'WARNING', 'CRITICAL', 'UNKNOWN'],

  /**
   * List of available SNMP versions
   * @type {Array}
   */
  SNMPVersions: ['SNMPv1', 'SNMPv2c', 'SNMPv3'],

  /**
   * List of all Alert Notifications
   * @type {App.AlertNotification[]}
   */
  alertNotifications: function () {
    return this.get('isLoaded') ? App.AlertNotification.find().toArray() : [];
  }.property('isLoaded'),

  /**
   * List of all Alert Groups
   * @type {App.AlertGroup[]}
   */
  allAlertGroups: function () {
    return this.get('isLoaded') ? App.AlertGroup.find().toArray() : [];
  }.property('isLoaded'),

  /**
   * Selected Alert Notification
   * @type {App.AlertNotification}
   */
  selectedAlertNotification: null,

  /**
   * Addable to <code>selectedAlertNotification.properties</code> custom property
   * @type {{name: string, value: string}}
   */
  newCustomProperty: {name: '', value: ''},

  /**
   * List custom property names that shouldn't be displayed on Edit page
   * @type {string[]}
   */
  ignoredCustomProperties: [
    'ambari.dispatch.credential.password',
    'ambari.dispatch.credential.username',
    'ambari.dispatch.recipients',
    'ambari.dispatch.snmp.community',
    'ambari.dispatch.snmp.oids.trap',
    'ambari.dispatch.snmp.port',
    'ambari.dispatch.snmp.version',
    'mail.smtp.auth',
    'mail.smtp.from',
    'mail.smtp.host',
    'mail.smtp.port',
    'mail.smtp.starttls.enable'
  ],

  /**
   * Load all Alert Notifications from server
   * @method loadAlertNotifications
   */
  loadAlertNotifications: function () {
    var self = this;
    this.set('isLoaded', false);
    App.router.get('updateController').updateAlertGroups(function () {
      App.ajax.send({
        name: 'alerts.notifications',
        sender: self,
        success: 'getAlertNotificationsSuccessCallback',
        error: 'getAlertNotificationsErrorCallback'
      });
    });
  },

  /**
   * Success-callback for load alert notifications request
   * @param {object} json
   * @method getAlertNotificationsSuccessCallback
   */
  getAlertNotificationsSuccessCallback: function (json) {
    App.alertNotificationMapper.map(json);
    this.set('isLoaded', true);
  },

  /**
   * Error-callback for load alert notifications request
   * @method getAlertNotificationsErrorCallback
   */
  getAlertNotificationsErrorCallback: function () {
    this.set('isLoaded', true);
  },

  /**
   * Add Notification button handler
   * @method addAlertNotification
   */
  addAlertNotification: function () {
    var inputFields = this.get('inputFields');
    inputFields.setProperties({
      'global.disabled': false
    });
    Em.keys(inputFields).forEach(function (key) {
      inputFields.set(key + '.value', inputFields.get(key + '.defaultValue'));
    });
    inputFields.set('severityFilter.value', ['OK', 'WARNING', 'CRITICAL', 'UNKNOWN']);
    this.showCreateEditPopup(false);
  },

  /**
   * Edit Notification button handler
   * @method editAlertNotification
   */
  editAlertNotification: function () {
    this.fillEditCreateInputs();
    this.showCreateEditPopup(true);
  },

  /**
   * Fill inputs of Create/Edit popup form
   * @param addCopyToName define whether add 'Copy of ' to name
   * @method fillEditCreateInputs
   */
  fillEditCreateInputs: function (addCopyToName) {
    var inputFields = this.get('inputFields');
    var selectedAlertNotification = this.get('selectedAlertNotification');
    inputFields.set('name.value', (addCopyToName ? 'Copy of ' : '') + selectedAlertNotification.get('name'));
    inputFields.set('groups.value', selectedAlertNotification.get('groups').toArray());
    inputFields.set('email.value', selectedAlertNotification.get('properties')['ambari.dispatch.recipients'] ?
      selectedAlertNotification.get('properties')['ambari.dispatch.recipients'].join(', ') : '');
    inputFields.set('SMTPServer.value', selectedAlertNotification.get('properties')['mail.smtp.host']);
    inputFields.set('SMTPPort.value', selectedAlertNotification.get('properties')['mail.smtp.port']);
    inputFields.set('SMTPUseAuthentication.value', selectedAlertNotification.get('properties')['mail.smtp.auth'] !== "false");
    inputFields.set('SMTPUsername.value', selectedAlertNotification.get('properties')['ambari.dispatch.credential.username']);
    inputFields.set('SMTPPassword.value', selectedAlertNotification.get('properties')['ambari.dispatch.credential.password']);
    inputFields.set('retypeSMTPPassword.value', selectedAlertNotification.get('properties')['ambari.dispatch.credential.password']);
    inputFields.set('SMTPSTARTTLS.value', selectedAlertNotification.get('properties')['mail.smtp.starttls.enable'] !== "false");
    inputFields.set('emailFrom.value', selectedAlertNotification.get('properties')['mail.smtp.from']);
    inputFields.set('version.value', selectedAlertNotification.get('properties')['ambari.dispatch.snmp.version']);
    inputFields.set('OIDs.value', selectedAlertNotification.get('properties')['ambari.dispatch.snmp.oids.trap']);
    inputFields.set('community.value', selectedAlertNotification.get('properties')['ambari.dispatch.snmp.community']);
    inputFields.set('port.value', selectedAlertNotification.get('properties')['ambari.dispatch.snmp.port']);
    inputFields.set('severityFilter.value', selectedAlertNotification.get('alertStates'));
    inputFields.set('global.value', selectedAlertNotification.get('global'));
    inputFields.set('allGroups.value', selectedAlertNotification.get('global') ? 'all' : 'custom');
    // not allow to edit global field
    inputFields.set('global.disabled', true);
    inputFields.set('description.value', selectedAlertNotification.get('description'));
    inputFields.set('method.value', selectedAlertNotification.get('type'));
    inputFields.get('customProperties').clear();
    var properties = selectedAlertNotification.get('properties');
    var ignoredCustomProperties = this.get('ignoredCustomProperties');
    Em.keys(properties).forEach(function (k) {
      if (ignoredCustomProperties.contains(k)) return;
      inputFields.get('customProperties').pushObject({
        name: k,
        value: properties[k],
        defaultValue: properties[k]
      });
    });
  },

  /**
   * Show Edit or Create Notification popup
   * @param {boolean} isEdit true - edit, false - create
   * @returns {App.ModalPopup}
   * @method showCreateEditPopup
   */
  showCreateEditPopup: function (isEdit) {
    var self = this;
    var createEditPopup = App.ModalPopup.show({
      header: isEdit ? Em.I18n.t('alerts.actions.manage_alert_notifications_popup.editHeader') : Em.I18n.t('alerts.actions.manage_alert_notifications_popup.addHeader'),
      classNames: ['create-edit-alert-notification-popup'],
      marginBottom: 130,
      bodyClass: Em.View.extend({
        controller: this,
        templateName: require('templates/main/alerts/create_alert_notification'),

        didInsertElement: function () {
          App.tooltip($('.checkbox-tooltip'));
          this.nameValidation();
          this.emailToValidation();
          this.emailFromValidation();
          this.smtpPortValidation();
          this.portValidation();
          this.retypePasswordValidation();
        },

        isEmailMethodSelected: function () {
          return this.get('controller.inputFields.method.value') === 'EMAIL';
        }.property('controller.inputFields.method.value'),

        nameValidation: function () {
          var newName = this.get('controller.inputFields.name.value').trim();
          var errorMessage = '';
          // on editing, save current notification name
          if (newName && !this.get('currentName')) {
            this.set('currentName', newName);
          }
          if (isEdit) {
            // edit current alert notification
            if (!newName) {
              this.set('nameError', true);
              errorMessage = Em.I18n.t('alerts.actions.manage_alert_notifications_popup.error.name.empty');
            } else if (newName && newName != this.get('currentName') && self.get('alertNotifications').mapProperty('name').contains(newName)) {
              this.set('nameError', true);
              errorMessage = Em.I18n.t('alerts.actions.manage_alert_notifications_popup.error.name.existed');
            } else {
              this.set('nameError', false);
            }
          } else {
            // add new alert notification
            if (!newName) {
              this.set('nameError', true);
              errorMessage = Em.I18n.t('alerts.actions.manage_alert_notifications_popup.error.name.empty');
            } else if (newName && self.get('alertNotifications').mapProperty('name').contains(newName)) {
              this.set('nameError', true);
              errorMessage = Em.I18n.t('alerts.actions.manage_alert_notifications_popup.error.name.existed');
            } else {
              this.set('nameError', false);
            }
          }
          this.set('controller.inputFields.name.errorMsg', errorMessage);
        }.observes('controller.inputFields.name.value'),

        emailToValidation: function () {
          var inputValue = this.get('controller.inputFields.email.value').trim(),
              emailsTo = inputValue.split(','),
              emailToError = false,
              i = emailsTo.length,
              emailTo;
          while (i--) {
            emailTo = emailsTo[i];
            if (emailTo && !validator.isValidEmail(emailTo.trim())) {
              emailToError = true;
              break;
            }
          }
          this.set('emailToError', emailToError);
          this.set('controller.inputFields.email.errorMsg', emailToError ? Em.I18n.t('alerts.notifications.error.email') : null);
        }.observes('controller.inputFields.email.value'),

        emailFromValidation: function () {
          var emailFrom = this.get('controller.inputFields.emailFrom.value');
          if (emailFrom && !validator.isValidEmail(emailFrom)) {
            this.set('emailFromError', true);
            this.set('controller.inputFields.emailFrom.errorMsg', Em.I18n.t('alerts.notifications.error.email'));
          } else {
            this.set('emailFromError', false);
            this.set('controller.inputFields.emailFrom.errorMsg', null);
          }
        }.observes('controller.inputFields.emailFrom.value'),

        smtpPortValidation: function () {
          var value = this.get('controller.inputFields.SMTPPort.value');
          if (value && (!validator.isValidInt(value) || value < 0)) {
            this.set('smtpPortError', true);
            this.set('controller.inputFields.SMTPPort.errorMsg', Em.I18n.t('alerts.notifications.error.integer'));
          } else {
            this.set('smtpPortError', false);
            this.set('controller.inputFields.SMTPPort.errorMsg', null);
          }
        }.observes('controller.inputFields.SMTPPort.value'),

        portValidation: function () {
          var value = this.get('controller.inputFields.port.value');
          if (value && (!validator.isValidInt(value) || value < 0)) {
            this.set('portError', true);
            this.set('controller.inputFields.port.errorMsg', Em.I18n.t('alerts.notifications.error.integer'));
          } else {
            this.set('portError', false);
            this.set('controller.inputFields.port.errorMsg', null);
          }
        }.observes('controller.inputFields.port.value'),

        retypePasswordValidation: function () {
          var passwordValue = this.get('controller.inputFields.SMTPPassword.value');
          var retypePasswordValue = this.get('controller.inputFields.retypeSMTPPassword.value');
          if (passwordValue !== retypePasswordValue) {
            this.set('passwordError', true);
            this.set('controller.inputFields.retypeSMTPPassword.errorMsg', Em.I18n.t('alerts.notifications.error.retypePassword'));
          } else {
            this.set('passwordError', false);
            this.set('controller.inputFields.retypeSMTPPassword.errorMsg', null);
          }
        }.observes('controller.inputFields.retypeSMTPPassword.value', 'controller.inputFields.SMTPPassword.value'),

        setParentErrors: function () {
          var hasErrors = this.get('nameError') || this.get('emailToError') || this.get('emailFromError') ||
            this.get('smtpPortError') || this.get('portError') || this.get('passwordError');
          this.set('parentView.hasErrors', hasErrors);
        }.observes('nameError', 'emailToError', 'emailFromError', 'smtpPortError', 'portError', 'passwordError'),


        groupsSelectView: Em.Select.extend({
          attributeBindings: ['disabled'],
          init: function () {
            this._super();
            this.set('parentView.groupSelect', this);
          }
        }),

        groupSelect: null,

        /**
         * Select all alert-groups if <code>allGroups.value</code> is 'custom'
         * @method selectAllGroups
         */
        selectAllGroups: function () {
          if (this.get('controller.inputFields.allGroups.value') == 'custom') {
            this.set('groupSelect.selection', this.get('groupSelect.content').slice());
          }
        },

        /**
         * Deselect all alert-groups if <code>allGroups.value</code> is 'custom'
         * @method clearAllGroups
         */
        clearAllGroups: function () {
          if (this.get('controller.inputFields.allGroups.value') == 'custom') {
            this.set('groupSelect.selection', []);
          }
        },

        severitySelectView: Em.Select.extend({
          init: function () {
            this._super();
            this.set('parentView.severitySelect', this);
          }
        }),

        severitySelect: null,

        /**
         * Determines if all alert-groups are selected
         * @type {boolean}
         */
        allGroupsSelected: function () {
          return this.get('groupSelect.selection.length') === this.get('groupSelect.content.length');
        }.property('groupSelect.selection.length', 'groupSelect.content.length', 'groupSelect.disabled'),

        /**
         * Determines if no one alert-group is selected
         * @type {boolean}
         */
        noneGroupsSelected: function () {
          return this.get('groupSelect.selection.length') === 0;
        }.property('groupSelect.selection.length', 'groupSelect.content.length', 'groupSelect.disabled'),

        /**
         * Determines if all severities are selected
         * @type {boolean}
         */
        allSeveritySelected: function () {
          return this.get('severitySelect.selection.length') === this.get('severitySelect.content.length');
        }.property('severitySelect.selection.length', 'severitySelect.content.length'),

        /**
         * Determines if no one severity is selected
         * @type {boolean}
         */
        noneSeveritySelected: function () {
          return this.get('severitySelect.selection.length') === 0;
        }.property('severitySelect.selection.length', 'severitySelect.content.length'),

        /**
         * Select all severities
         * @method selectAllSeverity
         */
        selectAllSeverity: function () {
          this.set('severitySelect.selection', this.get('severitySelect.content').slice());
        },

        /**
         * Deselect all severities
         * @method clearAllSeverity
         */
        clearAllSeverity: function () {
          this.set('severitySelect.selection', []);
        }
      }),

      isSaving: false,

      hasErrors: false,

      primary: Em.I18n.t('common.save'),

      disablePrimary: function () {
        return this.get('isSaving') || this.get('hasErrors');
      }.property('isSaving', 'hasErrors'),

      onPrimary: function () {
        this.set('isSaving', true);
        var apiObject = self.formatNotificationAPIObject();
        if (isEdit) {
          self.updateAlertNotification(apiObject);
        } else {
          self.createAlertNotification(apiObject);
        }
      }
    });
    this.set('createEditPopup', createEditPopup);
    return createEditPopup;
  },

  /**
   * Create API-formatted object from data populate by user
   * @returns {Object}
   * @method formatNotificationAPIObject
   */
  formatNotificationAPIObject: function () {
    var inputFields = this.get('inputFields');
    var properties = {};
    if (inputFields.get('method.value') === 'EMAIL') {
      properties['ambari.dispatch.recipients'] = inputFields.get('email.value').replace(/\s/g, '').split(',');
      properties['mail.smtp.host'] = inputFields.get('SMTPServer.value');
      properties['mail.smtp.port'] = inputFields.get('SMTPPort.value');
      properties['mail.smtp.from'] = inputFields.get('emailFrom.value');
      properties['mail.smtp.auth'] = inputFields.get('SMTPUseAuthentication.value');
      if (inputFields.get('SMTPUseAuthentication.value')) {
        properties['ambari.dispatch.credential.username'] = inputFields.get('SMTPUsername.value');
        properties['ambari.dispatch.credential.password'] = inputFields.get('SMTPPassword.value');
        properties['mail.smtp.starttls.enable'] = inputFields.get('SMTPSTARTTLS.value');
      }
    } else {
      properties['ambari.dispatch.snmp.version'] = inputFields.get('version.value');
      properties['ambari.dispatch.snmp.oids.trap'] = inputFields.get('OIDs.value');
      properties['ambari.dispatch.snmp.community'] = inputFields.get('community.value');
      properties['ambari.dispatch.snmp.port'] = inputFields.get('port.value');
    }
    inputFields.get('customProperties').forEach(function (customProperty) {
      properties[customProperty.name] = customProperty.value;
    });
    var apiObject = {
      AlertTarget: {
        name: inputFields.get('name.value'),
        description: inputFields.get('description.value'),
        global: inputFields.get('allGroups.value') === 'all',
        notification_type: inputFields.get('method.value'),
        alert_states: inputFields.get('severityFilter.value'),
        properties: properties
      }
    };
    if (inputFields.get('allGroups.value') == 'custom') {
      apiObject.AlertTarget.groups = inputFields.get('groups.value').mapProperty('id');
    }
    return apiObject;
  },

  /**
   * Send request to server to create Alert Notification
   * @param {object} apiObject (@see formatNotificationAPIObject)
   * @returns {$.ajax}
   * @method createAlertNotification
   */
  createAlertNotification: function (apiObject) {
    return App.ajax.send({
      name: 'alerts.create_alert_notification',
      sender: this,
      data: {
        data: apiObject
      },
      success: 'createAlertNotificationSuccessCallback',
      error: 'saveErrorCallback'
    });
  },

  /**
   * Success callback for <code>createAlertNotification</code>
   * @method createAlertNotificationSuccessCallback
   */
  createAlertNotificationSuccessCallback: function () {
    this.loadAlertNotifications();
    var createEditPopup = this.get('createEditPopup');
    if (createEditPopup) {
      createEditPopup.hide();
    }
  },

  /**
   * Send request to server to update Alert Notification
   * @param {object} apiObject (@see formatNotificationAPIObject)
   * @returns {$.ajax}
   * @method updateAlertNotification
   */
  updateAlertNotification: function (apiObject) {
    return App.ajax.send({
      name: 'alerts.update_alert_notification',
      sender: this,
      data: {
        data: apiObject,
        id: this.get('selectedAlertNotification.id')
      },
      success: 'updateAlertNotificationSuccessCallback',
      error: 'saveErrorCallback'
    });
  },

  /**
   * Success callback for <code>updateAlertNotification</code>
   * @method updateAlertNotificationSuccessCallback
   */
  updateAlertNotificationSuccessCallback: function () {
    this.loadAlertNotifications();
    var createEditPopup = this.get('createEditPopup');
    if (createEditPopup) {
      createEditPopup.hide();
    }
  },

  /**
   * Error callback for <code>createAlertNotification</code> and <code>updateAlertNotification</code>
   * @method saveErrorCallback
   */
  saveErrorCallback: function () {
    this.set('createEditPopup.isSaving', false);
  },

  /**
   * Delete Notification button handler
   * @return {App.ModalPopup}
   * @method deleteAlertNotification
   */
  deleteAlertNotification: function () {
    var self = this;
    return App.showConfirmationPopup(function () {
        App.ajax.send({
          name: 'alerts.delete_alert_notification',
          sender: self,
          data: {
            id: self.get('selectedAlertNotification.id')
          },
          success: 'deleteAlertNotificationSuccessCallback'
        });
      }, Em.I18n.t('alerts.actions.manage_alert_notifications_popup.confirmDeleteBody').format(this.get('selectedAlertNotification.name')),
      null, Em.I18n.t('alerts.actions.manage_alert_notifications_popup.confirmDeleteHeader'), Em.I18n.t('common.delete'));
  },

  /**
   * Success callback for <code>deleteAlertNotification</code>
   * @method deleteAlertNotificationSuccessCallback
   */
  deleteAlertNotificationSuccessCallback: function () {
    this.loadAlertNotifications();
    var selectedAlertNotification = this.get('selectedAlertNotification');
    selectedAlertNotification.deleteRecord();
    this.set('selectedAlertNotification', null);
  },

  /**
   * Duplicate Notification button handler
   * @method duplicateAlertNotification
   */
  duplicateAlertNotification: function () {
    this.fillEditCreateInputs(true);
    this.showCreateEditPopup();
  },

  /**
   * Show popup with form for new custom property
   * @method addCustomPropertyHandler
   * @return {App.ModalPopup}
   */
  addCustomPropertyHandler: function () {
    var self = this;

    return App.ModalPopup.show({

      header: Em.I18n.t('alerts.notifications.addCustomPropertyPopup.header'),

      primary: Em.I18n.t('common.add'),

      bodyClass: Em.View.extend({

        /**
         * If some error with new custom property
         * @type {boolean}
         */
        isError: false,

        controller: this,

        /**
         * Error message for new custom property (invalid name, existed name etc)
         * @type {string}
         */
        errorMessage: '',

        /**
         * Check new custom property for errors with its name
         * @method errorHandler
         */
        errorsHandler: function () {
          var name = this.get('controller.newCustomProperty.name');
          var flag = validator.isValidConfigKey(name);
          if (flag) {
            if (this.get('controller.inputFields.customProperties').mapProperty('name').contains(name) ||
              this.get('controller.ignoredCustomProperties').contains(name)) {
              this.set('errorMessage', Em.I18n.t('alerts.notifications.addCustomPropertyPopup.error.propertyExists'));
              flag = false;
            }
          }
          else {
            this.set('errorMessage', Em.I18n.t('alerts.notifications.addCustomPropertyPopup.error.invalidPropertyName'));
          }
          this.set('isError', !flag);
          this.set('parentView.disablePrimary', !flag);
        }.observes('controller.newCustomProperty.name'),

        templateName: require('templates/main/alerts/add_custom_config_to_alert_notification_popup')
      }),

      disablePrimary: true,

      onPrimary: function () {
        self.addCustomProperty();
        self.set('newCustomProperty', {name: '', value: ''}); // cleanup
        this.hide();
      }

    });
  },

  /**
   * Add Custom Property to <code>selectedAlertNotification</code> (push it to <code>properties</code>-field)
   * @method addCustomProperty
   */
  addCustomProperty: function () {
    var newCustomProperty = this.get('newCustomProperty');
    this.get('inputFields.customProperties').pushObject({
      name: newCustomProperty.name,
      value: newCustomProperty.value,
      defaultValue: newCustomProperty.value
    });
  },

  /**
   * "Remove"-button click handler
   * Delete customProperty from <code>inputFields.customProperties</code>
   * @param {object} e
   * @method removeCustomProperty
   */
  removeCustomPropertyHandler: function (e) {
    var customProperties = this.get('inputFields.customProperties');
    this.set('inputFields.customProperties', customProperties.without(e.context));
  }

});

});

require.register("controllers/main/charts", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainChartsController = Em.ArrayController.extend({
  name:'mainChartsController'
});
});

require.register("controllers/main/charts/heatmap", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.MainChartsHeatmapController = Em.Controller.extend({
  name: 'mainChartsHeatmapController',
  modelRacks: App.Rack.find(),
  racks: function () {
    var racks = [];
    this.get('modelRacks').forEach(function (rack) {
      racks.push(Em.Object.create({
        name: rack.get('name'),
        hosts: rack.get('hosts'),
        isLoaded: false
      }));
    });
    return racks;
  }.property('modelRacks.@each.isLoaded'),

  allMetrics: function () {
    var metrics = [];

    // Display host heatmaps if the stack definition has a host metrics service to display it.
    if(App.get('services.hostMetrics').length) {
      metrics.push(
        Em.Object.create({
          label: Em.I18n.t('charts.heatmap.category.host'),
          category: 'host',
          items: [
            App.MainChartHeatmapDiskSpaceUsedMetric.create(),
            App.MainChartHeatmapMemoryUsedMetric.create(),
            App.MainChartHeatmapCpuWaitIOMetric.create()
            /*, App.MainChartHeatmapProcessRunMetric.create()*/
          ]
        })
      );
    }

    if(App.HDFSService.find().get('length')) {
      metrics.push(
        Em.Object.create({
          label: Em.I18n.t('charts.heatmap.category.hdfs'),
          category: 'hdfs',
          items: [
            App.MainChartHeatmapDFSBytesReadMetric.create(),
            App.MainChartHeatmapDFSBytesWrittenMetric.create(),
            App.MainChartHeatmapDFSGCTimeMillisMetric.create(),
            App.MainChartHeatmapDFSMemHeapUsedMetric.create()
          ]
        })
      );
    }

    if (App.YARNService.find().get('length')) {
      metrics.push(
        Em.Object.create({
          label: Em.I18n.t('charts.heatmap.category.yarn'),
          category: 'yarn',
          items: [
            App.MainChartHeatmapYarnGCTimeMillisMetric.create(),
            App.MainChartHeatmapYarnMemHeapUsedMetric.create(),
            App.MainChartHeatmapYarnResourceUsedMetric.create()
          ]
        })
      );
    }

    if (App.HBaseService.find().get('length')) {
      metrics.push(
        Em.Object.create({
          label: Em.I18n.t('charts.heatmap.category.hbase'),
          category: 'hbase',
          items: [
            App.MainChartHeatmapHbaseReadReqCount.create(),
            App.MainChartHeatmapHbaseWriteReqCount.create(),
            App.MainChartHeatmapHbaseCompactionQueueSize.create(),
            App.MainChartHeatmapHbaseRegions.create(),
            App.MainChartHeatmapHbaseMemStoreSize.create()
          ]
        })
      );
    }
    return metrics;
  }.property(),

  selectedMetric: null,

  inputMaximum: '',

  validation: function () {
    if (this.get('selectedMetric')) {
      if (/^\d+$/.test(this.get('inputMaximum'))) {
        $('#inputMaximum').removeClass('error');
        this.set('selectedMetric.maximumValue', this.get('inputMaximum'));
      } else {
        $('#inputMaximum').addClass('error');
      }
    }
  }.observes('inputMaximum'),

  showHeatMapMetric: function (event) {
    var metricItem = event.context;
    if (metricItem) {
      this.set('selectedMetric', metricItem);
    }
  },

  hostToSlotMap: function () {
    return this.get('selectedMetric.hostToSlotMap');
  }.property('selectedMetric.hostToSlotMap'),

  loadMetrics: function () {
    var selectedMetric = this.get('selectedMetric');
    var hostNames = [];

    if (selectedMetric && this.get('racks').everyProperty('isLoaded', true)) {
      this.get('racks').forEach(function (rack) {
        hostNames = hostNames.concat(rack.hosts.mapProperty('hostName'));
      });
      selectedMetric.refreshHostSlots(hostNames);
    }
    this.set('inputMaximum', this.get('selectedMetric.maximumValue'));
  }.observes('selectedMetric'),

  /**
   * return class name for to be used for containing each rack.
   *
   * @this App.MainChartsHeatmapController
   */
  rackClass: function () {
    var rackCount = this.get('racks.length');
    if (rackCount < 2) {
      return "span12";
    } else if (rackCount == 2) {
      return "span6";
    } else {
      return "span4";
    }
  }.property('racks.length')
});

});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
var date = require('utils/date');

var heatmap = require('utils/heatmap');

/**
 * Base class for any heatmap metric.
 *
 * This class basically provides the following for each heatmap metric.
 * <ul>
 * <li> Provides number of slots in which temperature can fall.
 * <li> Maintains the maximum value so as to scale slot ranges.
 * <li> Gets JSON data from server and maps response for all hosts into above
 * slots.
 * </ul>
 *
 */
App.MainChartHeatmapMetric = Em.Object.extend(heatmap.mappers, {
  /**
   * Name of this metric
   */
  name: null,

  /**
   * Number of slots this metric will be mapped into. When changing this value,
   * the color count in 'slotColors' should also be changed.
   */
  numberOfSlots: 5,

  /**
   * Colors for the each of the number of slots defined above. When changing the
   * number of slots, the number of colors also should be updated.
   *
   * @type {Array}
   */
  slotColors: [ {
    r: 0x00,
    g: 0xcc,
    b: 0x00
  }, // Green
  {
    r: 0x9f,
    g: 0xee,
    b: 0x00
  }, {
    r: 0xff,
    g: 0xff,
    b: 0x00
  }, // Yellow
  {
    r: 0xff,
    g: 0xc0,
    b: 0x00
  }, // Orange
  {
    r: 0xff,
    g: 0x00,
    b: 0x00
  } ],// Red

  /**
   * Minimum value of this metric. Default is 0.
   */
  minimumValue: 0,

  /**
   * Maximum value of this metric. This has to be specified by extending classes
   * so that the range from 'minimumValue' to 'maximumValue' can be split among
   * 'numberOfSlots'. It is recommended that this value be a multiple of
   * 'numberOfSlots'.
   */
  maximumValue: 100,

  /**
   * Units of the maximum value which is shown in UI {String}
   */
  units: '',

  /**
   * Indicates whether this metric is currently loading data from the server.
   * {Boolean}
   */
  loading: false,

  /**
   * Provides following information about slots in an array of objects.
   * <ul>
   * <li> from: {number} Slot starts from this value
   * <li> to: {number} Slot ends at this value (inclusive)
   * <li> label: {String} Slot name to be shown
   * <li> cssStyle: {String} style to be embedded on hosts which fall into this
   * slot.
   * </ul>
   *
   * Slot count will be the same as specified in 'numberOfSlots'. Slot
   * definitions will be given in increasing temperature from 'minimumValue' to
   * 'maximumValue'.
   *
   */
  slotDefinitions: function () {
    var max = parseFloat(this.get('maximumValue'));
    var slotCount = this.get('numberOfSlots');
    var labelSuffix = this.get('slotDefinitionLabelSuffix');
    var delta = (max - this.get('minimumValue')) / slotCount;
    var defs = [];
    var slotColors = this.get('slotColors');
    var slotColorIndex = 0;
    for ( var c = 0; c < slotCount - 1; c++) {
      var slotColor = slotColors[slotColorIndex++];
      var start = (c * delta);
      var end = ((c + 1) * delta);
      defs.push(this.generateSlot(start, end, labelSuffix, slotColor));
    }
    slotColor = slotColors[slotColorIndex++];
    start = ((slotCount - 1) * delta);
    defs.push(this.generateSlot(start, max, labelSuffix, slotColor));

    defs.push(Em.Object.create({
      from: NaN,
      to: NaN,
      label: Em.I18n.t('charts.heatmap.label.invalidData'),
      cssStyle: this.getHatchStyle()
    }));
    defs.push(Em.Object.create({
      from: -1,
      to: -1,
      label: Em.I18n.t('charts.heatmap.label.notApplicable'),
      cssStyle: "background-color:rgb(200, 200, 200)"
    }));
    return defs;
  }.property('minimumValue', 'maximumValue', 'numberOfSlots'),
  /**
   * genarate slot by given parameters
   *
   * @param min
   * @param max
   * @param labelSuffix
   * @param slotColor
   * @return {object}
   * @method generateSlot
   */
  generateSlot: function (min, max, labelSuffix, slotColor) {
    var from = this.formatLegendNumber(min);
    var to = this.formatLegendNumber(max);
    var label;
    if ($.trim(labelSuffix) == 'ms') {
      label = date.timingFormat(from, 'zeroValid') + " - " + date.timingFormat(to, 'zeroValid');
    } else {
      label = from + labelSuffix + " - " + to + labelSuffix;
    }
    return Em.Object.create({
      from: from,
      to: to,
      label: label,
      cssStyle: "background-color:rgb(" + slotColor.r + "," + slotColor.g + "," + slotColor.b + ")"
    });
  },

  /**
   * calculate hatch style of slot according to browser version used
   * @return {String}
   */
  getHatchStyle: function () {
    var hatchStyle = "background-color:rgb(135, 206, 250)";
    if (jQuery.browser.webkit) {
      hatchStyle = "background-image:-webkit-repeating-linear-gradient(-45deg, #FF1E10, #FF1E10 3px, #ff6c00 3px, #ff6c00 6px)";
    } else if (jQuery.browser.mozilla) {
      hatchStyle = "background-image:repeating-linear-gradient(-45deg, #FF1E10, #FF1E10 3px, #ff6c00 3px, #ff6c00 6px)";
    } else if (jQuery.browser.msie && jQuery.browser.version) {
      var majorVersion = parseInt(jQuery.browser.version.split('.')[0]);
      if (majorVersion > 9) {
        hatchStyle = "background-image:repeating-linear-gradient(-45deg, #FF1E10, #FF1E10 3px, #ff6c00 3px, #ff6c00 6px)";
      }
    }
    return hatchStyle;
  },

  /**
   * In slot definitions this value is used to construct the label by appending
   * it to slot min-max values. For example giving '%' here would result in slot
   * definition label being '0% - 10%'.
   */
  slotDefinitionLabelSuffix: '',

  defaultMetric: '',

  /**
   * Name in the <code>App.ajax</code>
   * @type {String}
   */
  ajaxIndex: 'hosts.metrics',

  /**
   * Additional data for ajax-request
   * May be redeclared in child-objects
   * @type {Object}
   */
  ajaxData: {},

  /**
   * Maps server JSON into an object where keys are hostnames and values are the
   * true metric values. This function by default will map 'defaultMetric' into
   * its corresponding value.
   *
   * @Function
   */
  metricMapper: function (json) {
    var hostToValueMap = {};
    var metricName = this.get('defaultMetric');
    if (json.items) {
      var props = metricName.split('.');
      json.items.forEach(function (item) {
        var value = item;
        props.forEach(function (prop) {
          if (value != null && prop in value) {
            value = value[prop];
          } else {
            value = null;
          }
        });
        if (value != null) {
          var hostName = item.Hosts.host_name;
          hostToValueMap[hostName] = value;
        }
      });
    }
    return hostToValueMap;
  },

  hostToValueMap: null,

  hostToSlotMap: function () {
    var hostToValueMap = this.get('hostToValueMap');
    var hostNames = this.get('hostNames');
    var hostToSlotMap = {};
    if (hostToValueMap && hostNames) {
      hostNames.forEach(function (hostName) {
        var slot = this.calculateSlot(hostToValueMap, hostName)
        if (slot > -1) {
          hostToSlotMap[hostName] = slot;
        }
      }, this);
    }
    return hostToSlotMap;
  }.property('hostToValueMap', 'slotDefinitions'),

  /**
   * calculate slot position in hostToSlotMap by hostname
   * @param hostToValueMap
   * @param hostName
   * @return {Number}
   */
  calculateSlot: function (hostToValueMap, hostName) {
    var slot = -1;
    var slotDefs = this.get('slotDefinitions');

    if (hostName in hostToValueMap) {
      var value = hostToValueMap[hostName];
      if (isNaN(value)) {
        slot = slotDefs.length - 2;
      } else {
        for (var slotIndex = 0; slotIndex < slotDefs.length - 2; slotIndex++) {
          var slotDef = slotDefs[slotIndex];
          if (value >= slotDef.from && value <= slotDef.to) {
            slot = slotIndex;
          }
        }
        if (slot < 0) {
          // Assign it to the last legend
          slot = slotDefs.length - 3;
        }
      }
    } else {
      slot = slotDefs.length - 1;
    }
    return slot;
  },

  /**
   * Determines which slot each host falls into. This information is given to
   * the callback's #map(hostnameToSlotObject) method. The
   * 'hostnameToSlotObject' has key as hostname, and the slot index as value.
   */
  refreshHostSlots: function (hostNames) {
    this.set('loading', true);
    this.set('hostNames', hostNames);
    var fixedMetricName = this.get('defaultMetric');
    fixedMetricName = fixedMetricName.replace(/\./g, "/");
    var ajaxData = {
      metricName: fixedMetricName
    };
    jQuery.extend(ajaxData, this.get('ajaxData'));

    App.ajax.send({
      name: this.get('ajaxIndex'),
      sender: this,
      data: ajaxData,
      success: 'refreshHostSlotsSuccessCallback',
      error: 'refreshHostSlotsErrorCallback'
    });
  },

  refreshHostSlotsSuccessCallback: function (data) {
    var hostToValueMap = this.metricMapper(data);
    this.set('hostToValueMap', hostToValueMap);
    this.set('loading', false);
  },

  refreshHostSlotsErrorCallback: function () {
    this.set('loading', false);
  },

  /**
   * Turns numbers into displayable values. For example 24.345432425 into 24.3
   * etc.
   *
   * @private
   */
  formatLegendNumber: function (num) {
    var fraction = num % 1;
    if (fraction > 0) {
      return parseFloat(num.toFixed(1));
    }
    return num;
  }
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_cpuWaitIO", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapCpuWaitIOMetric = App.MainChartHeatmapMetric.extend({
  name: Em.I18n.t('charts.heatmap.metrics.cpuWaitIO'),
  maximumValue: 100,
  defaultMetric: 'metrics.cpu.cpu_wio',
  units: '%',
  slotDefinitionLabelSuffix: '%',
  metricMapper: function (json) {
    var map = this._super(json);
    for ( var host in map) {
      if (host in map) {
        var val = map[host];
        map[host] = (val * 100).toFixed(1);
      }
    }
    return map;
  }
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_dfs", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * Base class for any HDFS metric.
 */
App.MainChartHeatmapDFSMetrics = App.MainChartHeatmapMetric.extend({

  ajaxIndex: 'hosts.metrics.host_component',

  ajaxData: {
    serviceName: 'HDFS',
    componentName: 'DATANODE'
  },

  /**
   * Custom mapper for DFS metrics
   */
  metricMapper: function(json) {
    return this.metricMapperWithTransform(json, this.get('defaultMetric'), this.get('transformValue'));
  },
  /**
   * Utility function which allows extending classes to transform the value
   * assigned to a host.
   *
   * @Function
   */
  transformValue: null
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_bytesread", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapDFSBytesReadMetric = App.MainChartHeatmapDFSMetrics.extend({
  name: Em.I18n.t('charts.heatmap.metrics.bytesRead'),
  maximumValue: 1024, // 1GB
  defaultMetric: 'metrics.dfs.datanode.bytes_read',
  units: 'MB',
  slotDefinitionLabelSuffix: 'MB',
  transformValue: function (value) {
    return value / (1024 * 1024); // bytes divided by 1MB.
  }
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_byteswritten", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapDFSBytesWrittenMetric = App.MainChartHeatmapDFSMetrics.extend({
  name: Em.I18n.t('charts.heatmap.metrics.bytesWritten'),
  maximumValue: 1024, // 1GB
  defaultMetric: 'metrics.dfs.datanode.bytes_written',
  units: 'MB',
  slotDefinitionLabelSuffix: 'MB',
  transformValue: function (value) {
    return value / (1024 * 1024); // bytes divided by 1MB.
  }
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_gctime", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapDFSGCTimeMillisMetric = App.MainChartHeatmapDFSMetrics.extend({
  name: Em.I18n.t('charts.heatmap.metrics.DFSGarbageCollection'),
  maximumValue: 10000,
  defaultMetric: 'metrics.jvm.gcTimeMillis',
  units: ' ms',
  slotDefinitionLabelSuffix: ' ms'
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_dfs_memHeapUsed", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapDFSMemHeapUsedMetric = App.MainChartHeatmapDFSMetrics.extend({
  name: Em.I18n.t('charts.heatmap.metrics.DFSMemHeapUsed'),
  maximumValue: 512,
  defaultMetric: 'metrics.jvm.memHeapUsedM',
  units: 'MB',
  slotDefinitionLabelSuffix: 'MB'
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_diskspaceused", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapDiskSpaceUsedMetric = App.MainChartHeatmapMetric.extend({
  name: Em.I18n.t('charts.heatmap.metrics.diskSpaceUsed'),
  maximumValue: 100,
  defaultMetric: 'metrics.disk',
  units: '%',
  slotDefinitionLabelSuffix: '%',
  metricMapper: function (json) {
    var hostToValueMap = {};
    var self = this;
    var metricName = this.get('defaultMetric');
    if (json.items) {
      var props = metricName.split('.');
      json.items.forEach(function (item) {
        var value = item;
        props.forEach(function (prop) {
          if (value != null && prop in value) {
            value = value[prop];
          } else {
            value = null;
          }
        });
        if (value != null) {
          value = self.diskUsageFormatted(value.disk_total - value.disk_free, value.disk_total);
          var hostName = item.Hosts.host_name;
          hostToValueMap[hostName] = value;
        }
      });
    }
    return hostToValueMap;
  },

  /**
   * Format percent disk usage to float with 2 digits
   */
  diskUsageFormatted: function(diskUsed, diskTotal) {
    var diskUsage = (diskUsed) / diskTotal * 100;
    if (isNaN(diskUsage) || diskUsage < 0 || diskUsage > 100) {
      return Em.I18n.t('charts.heatmap.unknown');
    }
    var s = Math.round(diskUsage * Math.pow(10, 2)) / Math.pow(10, 2);
    return isNaN(s) ? 0 : s;
  }
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_hbase", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * Base class for any HDFS metric.
 */
App.MainChartHeatmapHbaseMetrics = App.MainChartHeatmapMetric.extend({

  ajaxIndex: 'hosts.metrics.host_component',

  ajaxData: {
    serviceName: 'HBASE',
    componentName: 'HBASE_REGIONSERVER'
  },

  /**
   * Custom mapper for HBase metrics
   */
  metricMapper: function(json) {
    return this.metricMapperWithTransform(json, this.get('defaultMetric'), this.get('transformValue'));
  },

  /**
   * Utility function which allows extending classes to transform the value
   * assigned to a host.
   *
   * @Function
   */
  transformValue: null
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_compactionqueue", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapHbaseCompactionQueueSize = App.MainChartHeatmapHbaseMetrics.extend({
  name: Em.I18n.t('charts.heatmap.metrics.HbaseRegionServerCompactionQueueSize'),
  maximumValue: 10,
  defaultMetric: 'metrics.hbase.regionserver.compactionQueueSize',
  units: '',
  slotDefinitionLabelSuffix: ''
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_memstoresize", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapHbaseMemStoreSize = App.MainChartHeatmapHbaseMetrics.extend({
  name: Em.I18n.t('charts.heatmap.metrics.HbaseRegionServerMemStoreSize'),
  maximumValue: 100*1024*1024,
  defaultMetric: 'metrics.hbase.regionserver.memstoreSize',
  units: 'B',
  slotDefinitionLabelSuffix: 'B'
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_readrequest", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapHbaseReadReqCount = App.MainChartHeatmapHbaseMetrics.extend({
  name: Em.I18n.t('charts.heatmap.metrics.HbaseRegionServerReadCount'),
  maximumValue: 200,
  defaultMetric: 'metrics.hbase.regionserver.readRequestsCount',
  units: '',
  slotDefinitionLabelSuffix: ''
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_regions", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapHbaseRegions = App.MainChartHeatmapHbaseMetrics.extend({
  name: Em.I18n.t('charts.heatmap.metrics.HbaseRegionServerRegions'),
  maximumValue: 10,
  defaultMetric: 'metrics.hbase.regionserver.regions',
  units: '',
  slotDefinitionLabelSuffix: ''
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_hbase_writerequest", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapHbaseWriteReqCount = App.MainChartHeatmapHbaseMetrics.extend({
  name: Em.I18n.t('charts.heatmap.metrics.HbaseRegionServerWriteCount'),
  maximumValue: 200,
  defaultMetric: 'metrics.hbase.regionserver.writeRequestsCount',
  units: '',
  slotDefinitionLabelSuffix: ''
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_memoryused", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * Base class for any heatmap metric.
 *
 * This class basically provides the following for each heatmap metric.
 * <ul>
 * <li> Provides number of slots in which temperature can fall.
 * <li> Maintains the maximum value so as to scale slot ranges.
 * <li> Gets JSON data from server and maps response for all hosts into above
 * slots.
 * </ul>
 *
 */
App.MainChartHeatmapMemoryUsedMetric = App.MainChartHeatmapMetric.extend({
  name: Em.I18n.t('charts.heatmap.metrics.memoryUsed'),
  maximumValue: 100,
  defaultMetric: 'metrics.memory',
  units: '%',
  slotDefinitionLabelSuffix: '%',
  metricMapper: function (json) {
    var hostToValueMap = {};
    var metricName = this.get('defaultMetric');
    if (json.items) {
      var props = metricName.split('.');
      json.items.forEach(function (item) {
        var value = item;
        props.forEach(function (prop) {
          if (value != null && prop in value) {
            value = value[prop];
          } else {
            value = null;
          }
        });
        if (value != null) {
          var total = value.mem_total;
          var used = value.mem_total - value.mem_free - value.mem_cached;
          value = ((used * 100) / total).toFixed(1);
          var hostName = item.Hosts.host_name;
          hostToValueMap[hostName] = value;
        }
      });
    }
    return hostToValueMap;
  }
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_processrun", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * Base class for any heatmap metric.
 *
 * This class basically provides the following for each heatmap metric.
 * <ul>
 * <li> Provides number of slots in which temperature can fall.
 * <li> Maintains the maximum value so as to scale slot ranges.
 * <li> Gets JSON data from server and maps response for all hosts into above
 * slots.
 * </ul>
 *
 */
App.MainChartHeatmapProcessRunMetric = App.MainChartHeatmapMetric.extend({
  name: Em.I18n.t('charts.heatmap.metrics.processRun'),
  maximumValue: 1,
  defaultMetric: 'metrics.process.proc_run',
  units: 'Processes',
  metricMapper: function (json) {
    var map = this._super(json);
    for (var host in map) {
      var val = map[host];
      map[host] = val.toFixed(1);
    }
    return map;
  }
});
});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_yarn", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * Base class for any YARN metric.
 */
App.MainChartHeatmapYarnMetrics = App.MainChartHeatmapMetric.extend({

  ajaxIndex: 'hosts.metrics.host_component',

  ajaxData: {
    serviceName: 'YARN',
    componentName: 'NODEMANAGER'
  },

  /**
   * Custom mapper for YARN metrics
   */
  metricMapper: function(json) {
    return this.metricMapperWithTransform(json, this.get('defaultMetric'), this.get('transformValue'));
  },

  /**
   * Utility function which allows extending classes to transform the value
   * assigned to a host.
   *
   * @Function
   */
  transformValue: null
});

});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_yarn_ResourceUsed", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapYarnResourceUsedMetric = App.MainChartHeatmapYarnMetrics.extend({
  name: Em.I18n.t('charts.heatmap.metrics.YarnMemoryUsed'),
  maximumValue: 100,
  defaultMetric: 'metrics.yarn',
  units: ' %',
  slotDefinitionLabelSuffix: ' %',

  metricMapper: function (json) {
    var hostToValueMap = {};
    var self = this;
    var metricName = this.get('defaultMetric');
    if (json.host_components) {
      var props = metricName.split('.');
      json.host_components.forEach(function (host) {
        var value = host;
        props.forEach(function (prop) {
          if (value != null && prop in value) {
            value = value[prop];
          } else {
            value = null;
          }
        });
        if (value != null) {
          value = self.memoryUsageFormatted(value.AllocatedGB, value.AvailableGB + value.AllocatedGB);
          var hostName = host.HostRoles.host_name;
          hostToValueMap[hostName] = value;
        }
      });
    }
    return hostToValueMap;
  },

  /**
   * Format percent YARN memory used to float with 2 digits
   */
  memoryUsageFormatted: function(used, total) {
    var usage = (used) / total * 100;
    if (isNaN(usage) || usage < 0 || usage > 100) {
      return Em.I18n.t('charts.heatmap.unknown');
    }
    var s = usage.toFixed(1);
    return isNaN(s) ? 0 : s;
  }
});

});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_yarn_gctime", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapYarnGCTimeMillisMetric = App.MainChartHeatmapYarnMetrics.extend({
  name: Em.I18n.t('charts.heatmap.metrics.YarnGCTime'),
  maximumValue: 10000,
  defaultMetric: 'metrics.jvm.gcTimeMillis',
  units: ' ms',
  slotDefinitionLabelSuffix: ' ms'
});

});

require.register("controllers/main/charts/heatmap_metrics/heatmap_metric_yarn_memHeapUsed", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 *
 */
App.MainChartHeatmapYarnMemHeapUsedMetric = App.MainChartHeatmapYarnMetrics.extend({
  name: Em.I18n.t('charts.heatmap.metrics.YarnMemHeapUsed'),
  maximumValue: 512,
  defaultMetric: 'metrics.jvm.memHeapUsedM',
  units: 'MB',
  slotDefinitionLabelSuffix: 'MB'
});

});

require.register("controllers/main/dashboard", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainDashboardController = Em.Controller.extend({
  name: 'mainDashboardController',
  categorySelected: 'widgets',

  scRequest: function(request) {
    return App.router.get('mainServiceController').get(request);
  },

  isAllServicesInstalled: function() {
    return this.scRequest('isAllServicesInstalled');
  }.property('App.router.mainServiceController.content.content.@each',
    'App.router.mainServiceController.content.content.length'),

  isStartAllDisabled: function() {
    return this.scRequest('isStartAllDisabled');
  }.property('App.router.mainServiceController.isStartStopAllClicked',
    'App.router.mainServiceController.content.@each.healthStatus'),

  isStopAllDisabled: function() {
    return this.scRequest('isStopAllDisabled');
  }.property('App.router.mainServiceController.isStartStopAllClicked',
    'App.router.mainServiceController.content.@each.healthStatus'),

  gotoAddService: function() {
    App.router.get('mainServiceController').gotoAddService();
  },

  startAllService: function(event){
    App.router.get('mainServiceController').startAllService(event);
  },

  stopAllService: function(event){
    App.router.get('mainServiceController').stopAllService(event);
  }
});
});

require.register("controllers/main/dashboard/config_history_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainConfigHistoryController = Em.ArrayController.extend(App.TableServerMixin, {
  name: 'mainConfigHistoryController',

  dataSource: App.ServiceConfigVersion.find(),
  content: function () {
    return this.get('dataSource').filterProperty('isRequested');
  }.property('dataSource.@each.isRequested'),
  isPolling: false,
  totalCount: 0,
  filteredCount: 0,
  timeoutRef: null,
  resetStartIndex: true,
  mockUrl: '/data/configurations/service_versions.json',
  realUrl: function () {
    return App.apiPrefix + '/clusters/' + App.get('clusterName') + '/configurations/service_config_versions?<parameters>fields=service_config_version,user,group_id,group_name,is_current,createtime,service_name,hosts,service_config_version_note&minimal_response=true';
  }.property('App.clusterName'),

  /**
   * associations between host property and column index
   * @type {Array}
   */
  colPropAssoc: function () {
    var associations = [];
    associations[1] = 'serviceVersion';
    associations[2] = 'configGroup';
    associations[3] = 'createTime';
    associations[4] = 'author';
    associations[5] = 'notes';
    return associations;
  }.property(),

  filterProps: [
    {
      name: 'serviceVersion',
      key: 'service_name',
      type: 'EQUAL'
    },
    {
      name: 'configGroup',
      key: 'group_name',
      type: 'EQUAL'
    },
    {
      name: 'createTime',
      key: 'createtime',
      type: 'MORE'
    },
    {
      name: 'author',
      key: 'user',
      type: 'MATCH'
    },
    {
      name: 'notes',
      key: 'service_config_version_note',
      type: 'MATCH'
    }
  ],

  sortProps: [
    {
      name: 'serviceVersion',
      key: 'service_name'
    },
    {
      name: 'configGroup',
      key: 'group_name'
    },
    {
      name: 'createTime',
      key: 'createtime'
    },
    {
      name: 'author',
      key: 'user'
    },
    {
      name: 'notes',
      key: 'service_config_version_note'
    }
  ],

  modifiedFilter: Em.Object.create({
    content: [
      {
        value: 'Any',
        label: Em.I18n.t('any')
      },
      {
        value: 'Past 1 hour',
        label: 'Past 1 hour'
      },
      {
        value: 'Past 1 Day',
        label: 'Past 1 Day'
      },
      {
        value: 'Past 2 Days',
        label: 'Past 2 Days'
      },
      {
        value: 'Past 7 Days',
        label: 'Past 7 Days'
      },
      {
        value: 'Past 14 Days',
        label: 'Past 14 Days'
      },
      {
        value: 'Past 30 Days',
        label: 'Past 30 Days'
      }
    ],
    optionValue: 'Any',
    filterModified: function () {
      var time = "";
      var curTime = new Date().getTime();

      switch (this.get('optionValue.value')) {
        case 'Past 1 hour':
          time = curTime - 3600000;
          break;
        case 'Past 1 Day':
          time = curTime - 86400000;
          break;
        case 'Past 2 Days':
          time = curTime - 172800000;
          break;
        case 'Past 7 Days':
          time = curTime - 604800000;
          break;
        case 'Past 14 Days':
          time = curTime - 1209600000;
          break;
        case 'Past 30 Days':
          time = curTime - 2592000000;
          break;
        case 'Any':
          time = "";
          break;
      }
      this.set("actualValues", {
        endTime: '',
        startTime: time
      });
    }.observes('optionValue'),
    cancel: function () {
      this.set('optionValue', this.get('content').findProperty('value', 'Any'));
    },
    actualValues: Em.Object.create({
      startTime: "",
      endTime: ""
    })
  }),

  /**
   * load all data components required by config history table
   *  - total counter of service config versions(called in parallel)
   *  - current versions
   *  - filtered versions
   * @return {*}
   */
  load: function () {
    var dfd = $.Deferred();
    this.updateTotalCounter();
    this.loadConfigVersionsToModel().done(function () {
      dfd.resolve();
    });
    return dfd.promise();
  },

  /**
   * get filtered service config versions from server and push it to model
   * @return {*}
   */
  loadConfigVersionsToModel: function () {
    var dfd = $.Deferred();
    var queryParams = this.getQueryParameters();

    App.HttpClient.get(this.getUrl(queryParams), App.serviceConfigVersionsMapper, {
      complete: function () {
        dfd.resolve();
      }
    });
    return dfd.promise();
  },

  updateTotalCounter: function () {
    return App.ajax.send({
      name: 'service.serviceConfigVersions.get.total',
      sender: this,
      data: {},
      success: 'updateTotalCounterSuccess'
    })
  },

  updateTotalCounterSuccess: function (data, opt, params) {
    this.set('totalCount', data.itemTotal);
  },

  getUrl: function (queryParams) {
    var params = '';
    if (App.get('testMode')) {
      return this.get('mockUrl');
    } else {
      if (queryParams) {
        params = App.router.get('updateController').computeParameters(queryParams);
      }
      return this.get('realUrl').replace('<parameters>', params);
    }
  },

  /**
   * request latest data from server and update content
   */
  doPolling: function () {
    var self = this;

    this.set('timeoutRef', setTimeout(function () {
      if (self.get('isPolling')) {
        self.load().done(function () {
          self.doPolling();
        })
      }
    }, App.componentsUpdateInterval));
  },

  /**
   * get sort properties from local db. A overridden funtion from parent
   * @return {Array}
   */
  getSortProps: function () {
    var savedSortConditions = App.db.getSortingStatuses(this.get('name')) || [],
      sortProperties = this.get('sortProps'),
      sortParams = [];

    savedSortConditions.forEach(function (sort) {
      var property = sortProperties.findProperty('name', sort.name);
      if (property && (sort.status === 'sorting_asc' || sort.status === 'sorting_desc')) {
        property.value = sort.status.replace('sorting_', '');
        property.type = 'SORT';
        if (property.name == 'serviceVersion'){
          property.key = "service_name." + sort.status.replace('sorting_', '') + ",service_config_version";
          property.value = "desc";
        }
        if (property.name == 'configGroup'){
          property.key = "group_name." + sort.status.replace('sorting_', '') + ",service_config_version";
          property.value = "desc";
        }

        sortParams.push(property);
      }
    });
    return sortParams;
  }
});

});

require.register("controllers/main/host", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var validator = require('utils/validator');
var batchUtils = require('utils/batch_scheduled_requests');

App.MainHostController = Em.ArrayController.extend(App.TableServerMixin, {
  name: 'mainHostController',

  dataSource: App.Host.find(),
  clearFilters: null,

  filteredCount: 0,
  /**
   * total number of installed hosts
   */
  totalCount: function () {
    return this.get('hostsCountMap')['TOTAL'] || 0;
  }.property('hostsCountMap'),
  resetStartIndex: false,
  /**
   * flag responsible for updating status counters of hosts
   */
  isCountersUpdating: false,

  hostsCountMap: {},

  startIndex: 1,

  expandedComponentsSections: [],

  expandedVersionsSections: [],

  /**
   * Components which will be shown in component filter
   * @returns {Array}
   */
  componentsForFilter: function () {
    var installedComponents = App.StackServiceComponent.find().toArray();
    installedComponents.setEach('checkedForHostFilter', false);
    return installedComponents;
  }.property('App.router.clusterController.isLoaded'),

  content: function () {
    return this.get('dataSource').filterProperty('isRequested');
  }.property('dataSource.@each.isRequested'),

  allHostStackVersions: App.HostStackVersion.find(),
  /**
   * filterProperties support follow types of filter:
   * MATCH - match of RegExp
   * EQUAL - equality "="
   * LESS - "<"
   * MORE - ">"
   * MULTIPLE - multiple values to compare
   * CUSTOM - substitute values with keys "{#}" in alias
   */
  filterProperties: [
    {
      name: 'hostName',
      key: 'Hosts/host_name',
      type: 'MATCH'
    },
    {
      name: 'ip',
      key: 'Hosts/ip',
      type: 'MATCH'
    },
    {
      name: 'cpu',
      key: 'Hosts/cpu_count',
      type: 'EQUAL'
    },
    {
      name: 'memoryFormatted',
      key: 'Hosts/total_mem',
      type: 'EQUAL'
    },
    {
      name: 'loadAvg',
      key: 'metrics/load/load_one',
      type: 'EQUAL'
    },
    {
      name: 'hostComponents',
      key: 'host_components/HostRoles/component_name',
      type: 'MULTIPLE'
    },
    {
      name: 'healthClass',
      key: 'Hosts/host_status',
      type: 'EQUAL'
    },
    {
      name: 'criticalWarningAlertsCount',
      key: 'alerts_summary/CRITICAL{0}|alerts_summary/WARNING{1}',
      type: 'CUSTOM'
    },
    {
      name: 'componentsWithStaleConfigsCount',
      key: 'host_components/HostRoles/stale_configs',
      type: 'EQUAL'
    },
    {
      name: 'componentsInPassiveStateCount',
      key: 'host_components/HostRoles/maintenance_state',
      type: 'MULTIPLE'
    },
    {
      name: 'selected',
      key: 'Hosts/host_name',
      type: 'MULTIPLE'
    },
    {
      name: 'hostStackVersion',
      key: 'stack_versions',
      type: 'EQUAL'
    }
  ],

  sortProps: [
    {
      name: 'hostName',
      key: 'Hosts/host_name'
    },
    {
      name: 'ip',
      key: 'Hosts/ip'
    },
    {
      name: 'cpu',
      key: 'Hosts/cpu_count'
    },
    {
      name: 'memoryFormatted',
      key: 'Hosts/total_mem'
    },
    {
      name: 'diskUsage',
      //TODO disk_usage is relative property and need support from API, metrics/disk/disk_free used temporarily
      key: 'metrics/disk/disk_free'
    },
    {
      name: 'loadAvg',
      key: 'metrics/load/load_one'
    }
  ],

  /**
   * Validate and convert input string to valid url parameter.
   * Detect if user have passed string as regular expression or extend
   * string to regexp.
   *
   * @param {String} value
   * @return {String}
   **/
  getRegExp: function (value) {
    value = validator.isValidMatchesRegexp(value) ? value.replace(/(\.+\*?|(\.\*)+)$/, '') + '.*' : '^$';
    value = /^\.\*/.test(value) || value == '^$' ? value : '.*' + value;
    return value;
  },

  /**
   * Sort by host_name by default
   * @method getSortProps
   * @returns {{value: 'asc|desc', name: string, type: 'SORT'}[]}
   */
  getSortProps: function () {
    var controllerName = this.get('name'),
      db = App.db.getSortingStatuses(controllerName);
    if (db && db.everyProperty('status', 'sorting')) {
      App.db.setSortingStatuses(controllerName, {
        name: 'hostName',
        status: 'sorting_asc'
      });
    }
    return this._super();
  },

  /**
   * get query parameters computed from filter properties, sort properties and custom properties of view
   * @param {boolean} [skipNonFilterProperties]
   * @return {Array}
   * @method getQueryParameters
   */
  getQueryParameters: function (skipNonFilterProperties) {
    skipNonFilterProperties = skipNonFilterProperties || false;
    var queryParams = [],
      savedFilterConditions = App.db.getFilterConditions(this.get('name')) || [],
      savedSortConditions = App.db.getSortingStatuses(this.get('name')) || [],
      colPropAssoc = this.get('colPropAssoc'),
      filterProperties = this.get('filterProperties'),
      sortProperties = this.get('sortProps'),
      oldProperties = App.router.get('updateController.queryParams.Hosts');

    this.set('resetStartIndex', false);

    queryParams.pushObjects(this.getPaginationProps());

    savedFilterConditions.forEach(function (filter) {
      var property = filterProperties.findProperty('name', colPropAssoc[filter.iColumn]);
      if (property && filter.value.length > 0 && !filter.skipFilter) {
        var result = {
          key: property.key,
          value: filter.value,
          type: property.type,
          isFilter: true
        };
        if (filter.type === 'string' && sortProperties.someProperty('name', colPropAssoc[filter.iColumn])) {
          result.value = this.getRegExp(filter.value);
        }
        if (filter.type === 'number' || filter.type === 'ambari-bandwidth') {
          result.type = this.getComparisonType(filter.value);
          result.value = this.getProperValue(filter.value);
        }
        // enter an exact number for RAM filter, need to do a range number match for this
        if (filter.type === 'ambari-bandwidth' && result.type == 'EQUAL' && result.value) {
          var valuePair = this.convertMemoryToRange(filter.value);
          queryParams.push({
            key: result.key,
            value: valuePair[0],
            type: 'MORE'
          });
          queryParams.push({
            key: result.key,
            value: valuePair[1],
            type: 'LESS'
          });
        } else if (filter.type === 'ambari-bandwidth' && result.type != 'EQUAL' && result.value){
          // enter a comparison type, eg > 1, just do regular match
          result.value = this.convertMemory(filter.value);
          queryParams.push(result);
        } else if (filter.type === 'sub-resource') {
          filter.value.forEach(function (item) {
            queryParams.push({
              key: result.key + "/" + item.property,
              value: item.value,
              type: 'EQUAL'
            });
          }, this);
        } else {
          queryParams.push(result);
        }

      }
    }, this);

    if (queryParams.filterProperty('isFilter').length !== oldProperties.filterProperty('isFilter').length) {
      queryParams.findProperty('key', 'from').value = 0;
      this.set('resetStartIndex', true);
    } else {
      queryParams.filterProperty('isFilter').forEach(function (queryParam) {
        var oldProperty = oldProperties.filterProperty('isFilter').findProperty('key', queryParam.key);
        if (!oldProperty || JSON.stringify(oldProperty.value) !== JSON.stringify(queryParam.value)) {
          queryParams.findProperty('key', 'from').value = 0;
          this.set('resetStartIndex', true);
        }
      }, this);
    }

    if (!skipNonFilterProperties) {
      queryParams.pushObjects(this.getSortProps());
    }

    return queryParams;
  },

  /**
   * update status counters of hosts
   */
  updateStatusCounters: function () {
    var self = this;

    if (this.get('isCountersUpdating')) {
      App.ajax.send({
        name: 'host.status.counters',
        sender: this,
        data: {},
        success: 'updateStatusCountersSuccessCallback',
        error: 'updateStatusCountersErrorCallback',
        callback: function() {
          setTimeout(function () {
            self.updateStatusCounters();
          }, App.get('hostStatusCountersUpdateInterval'));
        }
      });
    }
  },

  /**
   * success callback on <code>updateStatusCounters()</code>
   * map counters' value to categories
   * @param data
   */
  updateStatusCountersSuccessCallback: function (data) {
    var hostsCountMap = {
      'HEALTHY': data.Clusters.health_report['Host/host_status/HEALTHY'],
      'UNHEALTHY': data.Clusters.health_report['Host/host_status/UNHEALTHY'],
      'ALERT': data.Clusters.health_report['Host/host_status/ALERT'],
      'UNKNOWN': data.Clusters.health_report['Host/host_status/UNKNOWN'],
      'health-status-WITH-ALERTS': (data.alerts_summary_hosts) ? data.alerts_summary_hosts.CRITICAL + data.alerts_summary_hosts.WARNING : 0,
      'health-status-CRITICAL': (data.alerts_summary_hosts) ? data.alerts_summary_hosts.CRITICAL : 0,
      'health-status-RESTART': data.Clusters.health_report['Host/stale_config'],
      'health-status-PASSIVE_STATE': data.Clusters.health_report['Host/maintenance_state'],
      'TOTAL': data.Clusters.total_hosts
    };

    this.set('hostsCountMap', hostsCountMap);
  },

  /**
   * success callback on <code>updateStatusCounters()</code>
   */
  updateStatusCountersErrorCallback: function() {
    console.warn('ERROR: updateStatusCounters failed')
  },

  /**
   * Return value without predicate
   * @param {String} value
   * @return {String}
   */
  getProperValue: function (value) {
    return (value.charAt(0) === '>' || value.charAt(0) === '<' || value.charAt(0) === '=') ? value.substr(1, value.length) : value;
  },

  /**
   * Return value converted to kilobytes
   * @param {String} value
   * @return {*}
   */
  convertMemory: function (value) {
    var scale = value.charAt(value.length - 1);
    // first char may be predicate for comparison
    value = this.getProperValue(value);
    var parsedValue = parseFloat(value);

    if (isNaN(parsedValue)) {
      return value;
    }

    switch (scale) {
      case 'g':
        parsedValue *= 1048576;
        break;
      case 'm':
        parsedValue *= 1024;
        break;
      case 'k':
        break;
      default:
        //default value in GB
        parsedValue *= 1048576;
    }
    return Math.round(parsedValue);
  },

  /**
   * Return value converted to a range of kilobytes
   * @param {String} value
   * @return {Array}
   */
  convertMemoryToRange: function (value) {
    var scale = value.charAt(value.length - 1);
    // first char may be predicate for comparison
    value = this.getProperValue(value);
    var parsedValue = parseFloat(value);
    if (isNaN(parsedValue)) {
      return value;
    }
    var parsedValuePair = this.rangeConvertNumber(parsedValue, scale);
    var multiplyingFactor = 1;
    switch (scale) {
      case 'g':
        multiplyingFactor = 1048576;
        break;
      case 'm':
        multiplyingFactor = 1024;
        break;
      case 'k':
        break;
      default:
        //default value in GB
        multiplyingFactor = 1048576;
    }
    parsedValuePair[0]  = Math.round( parsedValuePair[0] * multiplyingFactor);
    parsedValuePair[1]  = Math.round( parsedValuePair[1] * multiplyingFactor);
    return parsedValuePair;
  },

  /**
   * Return value converted to a range of kilobytes
   * eg, return value 1.83 g will target 1.82500 ~ 1.83499 g
   * eg, return value 1.8 k will target 1.7500 ~ 1.8499 k
   * eg, return value 1.8 m will target 1.7500 ~ 1.8499 m
   * @param {number} value
   * @param {String} scale
   * @return {Array}
   */
  rangeConvertNumber: function (value, scale) {
    if (isNaN(value)) {
      return value;
    }
    var valuePair = [];
    switch (scale) {
      case 'g':
        valuePair = [value - 0.005000, value + 0.004999999];
        break;
      case 'm':
      case 'k':
        valuePair = [value - 0.05000, value + 0.04999];
        break;
      default:
        //default value in GB
        valuePair = [value - 0.005000, value + 0.004999999];
    }
    return valuePair;
  },

  /**
   * Return comparison type depending on populated predicate
   * @param value
   * @return {String}
   */
  getComparisonType: function (value) {
    var comparisonChar = value.charAt(0);
    var result = 'EQUAL';
    if (isNaN(comparisonChar)) {
      switch (comparisonChar) {
        case '>':
          result = 'MORE';
          break;
        case '<':
          result = 'LESS';
          break;
      }
    }
    return result;
  },

  /**
   * Filter hosts by componentName of <code>component</code>
   * @param {App.HostComponent} component
   */
  filterByComponent: function (component) {
    if (!component)
      return;
    var id = component.get('componentName');
    var column = 6;
    this.get('componentsForFilter').setEach('checkedForHostFilter', false);

    var filterForComponent = {
      iColumn: column,
      value: [id],
      type: 'multiple'
    };
    App.db.setFilterConditions(this.get('name'), [filterForComponent]);
  },

  /**
   * Filter hosts by stack version and state
   * @param {String} displayName
   * @param {String} state
   */
  filterByStack: function (displayName, state) {
    if (!displayName || !state)
      return;
    var column = 11;

    var filterForStack = {
      iColumn: column,
      value: [
        {
          property: 'repository_versions/RepositoryVersions/display_name',
          value: displayName
        },
        {
          property: 'HostStackVersions/state',
          value: state.toUpperCase()
        }
      ],
      type: 'sub-resource'
    };
    App.db.setFilterConditions(this.get('name'), [filterForStack]);
  },

  goToHostAlerts: function (event) {
    var host = event && event.context;
    if (host) {
      App.router.transitionTo('main.hosts.hostDetails.alerts', host);
    }
  },

  /**
   * remove selected hosts
   */
  removeHosts: function () {
    var hosts = this.get('content');
    var selectedHosts = hosts.filterProperty('isChecked', true);
    selectedHosts.forEach(function (_hostInfo) {
      console.log('Removing:  ' + _hostInfo.hostName);
    });
    this.get('fullContent').removeObjects(selectedHosts);
  },

  /**
   * remove hosts with id equal host_id
   * @param {String} host_id
   */
  checkRemoved: function (host_id) {
    var hosts = this.get('content');
    var selectedHosts = hosts.filterProperty('id', host_id);
    this.get('fullContent').removeObjects(selectedHosts);
  },

  /**
   * Bulk operation wrapper
   * @param {Object} operationData - data about bulk operation (action, hosts or hostComponents etc)
   * @param {Array} hosts - list of affected hosts
   */
  bulkOperation: function (operationData, hosts) {
    if (operationData.componentNameFormatted) {
      if (operationData.action === 'RESTART') {
        this.bulkOperationForHostComponentsRestart(operationData, hosts);
      }
      else {
        if (operationData.action.indexOf('DECOMMISSION') != -1) {
          this.bulkOperationForHostComponentsDecommission(operationData, hosts);
        }
        else {
          this.bulkOperationForHostComponents(operationData, hosts);
        }
      }
    }
    else {
      if (operationData.action === 'RESTART') {
        this.bulkOperationForHostsRestart(operationData, hosts);
      }
      else {
        if (operationData.action === 'PASSIVE_STATE') {
          this.bulkOperationForHostsPassiveState(operationData, hosts);
        }
        else {
          this.bulkOperationForHosts(operationData, hosts);
        }
      }
    }
  },

  /**
   * Bulk operation (start/stop all) for selected hosts
   * @param {Object} operationData - data about bulk operation (action, hostComponents etc)
   * @param {Array} hosts - list of affected hosts
   */
  bulkOperationForHosts: function (operationData, hosts) {
    var self = this;

    batchUtils.getComponentsFromServer({
      hosts: hosts.mapProperty('hostName'),
      passiveState: 'OFF',
      displayParams: ['host_components/HostRoles/component_name']
    }, function (data) {
      self.bulkOperationForHostsCallback(operationData, data);
    });
  },
  /**
   * run Bulk operation (start/stop all) for selected hosts
   * after host and components are loaded
   * @param operationData
   * @param data
   */
  bulkOperationForHostsCallback: function (operationData, data) {
    var query = [];
    var hostNames = [];
    var hostsMap = {};

    data.items.forEach(function (host) {
      host.host_components.forEach(function (hostComponent) {
        if (!App.components.get('clients').contains((hostComponent.HostRoles.component_name))) {
          if (hostsMap[host.Hosts.host_name]) {
            hostsMap[host.Hosts.host_name].push(hostComponent.HostRoles.component_name);
          } else {
            hostsMap[host.Hosts.host_name] = [hostComponent.HostRoles.component_name];
          }
        }
      });
    });

    for (var hostName in hostsMap) {
      var subQuery = '(HostRoles/component_name.in(%@)&HostRoles/host_name=' + hostName + ')';
      var components = hostsMap[hostName];
      if (components.length) {
        query.push(subQuery.fmt(components.join(',')));
      }
      hostNames.push(hostName);
    }

    hostNames = hostNames.join(",");
    if (query.length) {
      query = query.join('|');
      App.ajax.send({
        name: 'common.host_components.update',
        sender: this,
        data: {
          query: query,
          HostRoles: {
            state: operationData.action
          },
          context: operationData.message,
          hostName: hostNames
        },
        success: 'bulkOperationForHostComponentsSuccessCallback'
      });
    }
    else {
      App.ModalPopup.show({
        header: Em.I18n.t('rolling.nothingToDo.header'),
        body: Em.I18n.t('rolling.nothingToDo.body').format(Em.I18n.t('hosts.host.maintainance.allComponents.context')),
        secondary: false
      });
    }
  },

  /**
   * Bulk restart for selected hosts
   * @param {Object} operationData - data about bulk operation (action, hostComponents etc)
   * @param {Ember.Enumerable} hosts - list of affected hosts
   */
  bulkOperationForHostsRestart: function (operationData, hosts) {
    batchUtils.getComponentsFromServer({
      passiveState: 'OFF',
      hosts: hosts.mapProperty('hostName'),
      displayParams: ['host_components/HostRoles/component_name']
    }, function (data) {
      var hostComponents = [];
      data.items.forEach(function (host) {
        host.host_components.forEach(function (hostComponent) {
          hostComponents.push(Em.Object.create({
            componentName: hostComponent.HostRoles.component_name,
            hostName: host.Hosts.host_name
          }));
        })
      });
      batchUtils.restartHostComponents(hostComponents, Em.I18n.t('rollingrestart.context.allOnSelectedHosts'), "HOST");
    });
  },

  /**
   * Bulk turn on/off passive state for selected hosts
   * @param {Object} operationData - data about bulk operation (action, hostComponents etc)
   * @param {Array} hosts - list of affected hosts
   */
  bulkOperationForHostsPassiveState: function (operationData, hosts) {
    var self = this;

    batchUtils.getComponentsFromServer({
      hosts: hosts.mapProperty('hostName'),
      displayParams: ['Hosts/maintenance_state']
    }, function (data) {
      var hostNames = [];

      data.items.forEach(function (host) {
        if (host.Hosts.maintenance_state !== operationData.state) {
          hostNames.push(host.Hosts.host_name);
        }
      });
      if (hostNames.length) {
        App.ajax.send({
          name: 'bulk_request.hosts.passive_state',
          sender: self,
          data: {
            hostNames: hostNames.join(','),
            passive_state: operationData.state,
            requestInfo: operationData.message
          },
          success: 'updateHostPassiveState'
        });
      } else {
        App.ModalPopup.show({
          header: Em.I18n.t('rolling.nothingToDo.header'),
          body: Em.I18n.t('hosts.bulkOperation.passiveState.nothingToDo.body'),
          secondary: false
        });
      }
    });
  },

  updateHostPassiveState: function (data, opt, params) {
    batchUtils.infoPassiveState(params.passive_state);
  },
  /**
   * Bulk operation for selected hostComponents
   * @param {Object} operationData - data about bulk operation (action, hostComponents etc)
   * @param {Array} hosts - list of affected hosts
   */
  bulkOperationForHostComponents: function (operationData, hosts) {
    var self = this;

    batchUtils.getComponentsFromServer({
      components: [operationData.componentName],
      hosts: hosts.mapProperty('hostName'),
      passiveState: 'OFF'
    }, function (data) {
      if (data.items.length) {
        var hostsWithComponentInProperState = data.items.mapProperty('Hosts.host_name');
        App.ajax.send({
          name: 'common.host_components.update',
          sender: self,
          data: {
            HostRoles: {
              state: operationData.action
            },
            query: 'HostRoles/component_name=' + operationData.componentName + '&HostRoles/host_name.in(' + hostsWithComponentInProperState.join(',') + ')&HostRoles/maintenance_state=OFF',
            context: operationData.message + ' ' + operationData.componentNameFormatted,
            level: 'SERVICE'
          },
          success: 'bulkOperationForHostComponentsSuccessCallback'
        });
      }
      else {
        App.ModalPopup.show({
          header: Em.I18n.t('rolling.nothingToDo.header'),
          body: Em.I18n.t('rolling.nothingToDo.body').format(operationData.componentNameFormatted),
          secondary: false
        });
      }
    });
  },

  /**
   * Bulk decommission/recommission for selected hostComponents
   * @param {Object} operationData
   * @param {Array} hosts
   */
  bulkOperationForHostComponentsDecommission: function (operationData, hosts) {
    var self = this;

    batchUtils.getComponentsFromServer({
      components: [operationData.realComponentName],
      hosts: hosts.mapProperty('hostName'),
      passiveState: 'OFF',
      displayParams: ['host_components/HostRoles/state']
    }, function (data) {
      self.bulkOperationForHostComponentsDecommissionCallBack(operationData, data)
    });
  },

  /**
   * run Bulk decommission/recommission for selected hostComponents
   * after host and components are loaded
   * @param operationData
   * @param data
   */
  bulkOperationForHostComponentsDecommissionCallBack: function (operationData, data) {
    var service = App.Service.find(operationData.serviceName);
    var components = [];

    data.items.forEach(function (host) {
      host.host_components.forEach(function (hostComponent) {
        components.push(Em.Object.create({
          componentName: hostComponent.HostRoles.component_name,
          hostName: host.Hosts.host_name,
          workStatus: hostComponent.HostRoles.state
        }))
      });
    });

    if (components.length) {
      var hostsWithComponentInProperState = components.mapProperty('hostName');
      var turn_off = operationData.action.indexOf('OFF') !== -1;
      var svcName = operationData.serviceName;
      var masterName = operationData.componentName;
      var slaveName = operationData.realComponentName;
      var hostNames = hostsWithComponentInProperState.join(',');
      if (turn_off) {
        // For recommession
        if (svcName === "YARN" || svcName === "HBASE" || svcName === "HDFS") {
          App.router.get('mainHostDetailsController').doRecommissionAndStart(hostNames, svcName, masterName, slaveName);
        }
      } else {
        hostsWithComponentInProperState = components.filterProperty('workStatus', 'STARTED').mapProperty('hostName');
        //For decommession
        if (svcName == "HBASE") {
          // HBASE service, decommission RegionServer in batch requests
          this.warnBeforeDecommission(hostNames);
        } else {
          var parameters = {
            "slave_type": slaveName
          };
          var contextString = turn_off ? 'hosts.host.' + slaveName.toLowerCase() + '.recommission' :
              'hosts.host.' + slaveName.toLowerCase() + '.decommission';
          if (turn_off) {
            parameters['included_hosts'] = hostsWithComponentInProperState.join(',')
          }
          else {
            parameters['excluded_hosts'] = hostsWithComponentInProperState.join(',');
          }
          App.ajax.send({
            name: 'bulk_request.decommission',
            sender: this,
            data: {
              context: Em.I18n.t(contextString),
              serviceName: service.get('serviceName'),
              componentName: operationData.componentName,
              parameters: parameters
            },
            success: 'bulkOperationForHostComponentsSuccessCallback'
          });
        }
      }
    }
    else {
      App.ModalPopup.show({
        header: Em.I18n.t('rolling.nothingToDo.header'),
        body: Em.I18n.t('rolling.nothingToDo.body').format(operationData.componentNameFormatted),
        secondary: false
      });
    }
  },


  /**
   * get info about regionserver passive_state
   * @method warnBeforeDecommission
   * @param {String} hostNames
   * @return {$.ajax}
   */
  warnBeforeDecommission: function (hostNames) {
    return App.ajax.send({
      'name': 'host_components.hbase_regionserver.active',
      'sender': this,
      'data': {
        hostNames: hostNames
      },
      success: 'warnBeforeDecommissionSuccess'
    });
  },

  /**
   * check is hbase regionserver in mm. If so - run decommission
   * otherwise shows warning
   * @method warnBeforeDecommission
   * @param {Object} data
   * @param {Object} opt
   * @param {Object} params
   */
  warnBeforeDecommissionSuccess: function(data, opt, params) {
    if (Em.get(data, 'items.length')) {
      App.router.get('mainHostDetailsController').showHbaseActiveWarning();
    } else {
      App.router.get('mainHostDetailsController').checkRegionServerState(params.hostNames);
    }
  },
  /**
   * Bulk restart for selected hostComponents
   * @param {Object} operationData
   * @param {Array} hosts
   */
  bulkOperationForHostComponentsRestart: function (operationData, hosts) {
    var service = App.Service.find(operationData.serviceName);

    batchUtils.getComponentsFromServer({
      components: [operationData.componentName],
      hosts: hosts.mapProperty('hostName'),
      passiveState: 'OFF',
      displayParams: ['Hosts/maintenance_state', 'host_components/HostRoles/stale_configs', 'host_components/HostRoles/maintenance_state']
    }, function (data) {
      var wrappedHostComponents = [];

      data.items.forEach(function (host) {
        host.host_components.forEach(function (hostComponent) {
          wrappedHostComponents.push(Em.Object.create({
            componentName: hostComponent.HostRoles.component_name,
            serviceName: operationData.serviceName,
            hostName: host.Hosts.host_name,
            hostPassiveState: host.Hosts.maintenance_state,
            staleConfigs: hostComponent.HostRoles.stale_configs,
            passiveState: hostComponent.HostRoles.maintenance_state
          }))
        });
      });

      if (wrappedHostComponents.length) {
        batchUtils.showRollingRestartPopup(wrappedHostComponents.objectAt(0).get('componentName'), service.get('displayName'), service.get('passiveState') === "ON", false, wrappedHostComponents);
      } else {
        App.ModalPopup.show({
          header: Em.I18n.t('rolling.nothingToDo.header'),
          body: Em.I18n.t('rolling.nothingToDo.body').format(operationData.componentNameFormatted),
          secondary: false
        });
      }
    });
  },

  updateHostComponentsPassiveState: function (data, opt, params) {
    batchUtils.infoPassiveState(params.passive_state);
  },
  /**
   * Show BO popup after bulk request
   */
  bulkOperationForHostComponentsSuccessCallback: function () {
    App.router.get('applicationController').dataLoading().done(function (initValue) {
      if (initValue) {
        App.router.get('backgroundOperationsController').showPopup();
      }
    });
  },
  /**
   * associations between host property and column index
   * @type {Array}
   */
  colPropAssoc: function () {
    var associations = [];
    associations[0] = 'healthClass';
    associations[1] = 'hostName';
    associations[2] = 'ip';
    associations[3] = 'cpu';
    associations[4] = 'memoryFormatted';
    associations[5] = 'loadAvg';
    associations[6] = 'hostComponents';
    associations[7] = 'criticalWarningAlertsCount';
    associations[8] = 'componentsWithStaleConfigsCount';
    associations[9] = 'componentsInPassiveStateCount';
    associations[10] = 'selected';
    associations[11] = 'hostStackVersion';
    return associations;
  }.property()

});

});

require.register("controllers/main/host/addHost/step4_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.AddHostStep4Controller= Em.Controller.extend({

  name:"addHostStep4Controller",
  isConfigGroupLoaded: false,

  loadConfigGroups: function () {
    App.ajax.send({
      name: 'config_groups.all_fields',
      sender: this,
      success: 'successLoadingConfigGroup',
      error: 'errorLoadingConfigGroup'
    });
  },

  successLoadingConfigGroup: function (data) {
    App.router.get('addHostController.content').set('configGroups',data.items);
    this.set('isConfigGroupLoaded', true);
  },

  errorLoadingConfigGroup: function(data) {
    App.router.get('addHostController.content').set('configGroups', []);
    this.set('isConfigGroupLoaded', true);
  },

  configGroupsLoading: function () {
    var df = $.Deferred();
    if (this.get('isConfigGroupLoaded')) {
      df.resolve();
    } else {
      this.interval = setInterval(function () {
        if (this.get('isConfigGroupLoaded')) {
          df.resolve();
          clearInterval(this.interval);
        }
      }.bind(this), 55);
    }
    return df.promise();
  }

});


});

require.register("controllers/main/host/add_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.AddHostController = App.WizardController.extend({

  name: 'addHostController',

  totalSteps: 7,

  /**
   * Used for hiding back button in wizard
   */
  hideBackButton: true,

  /**
   * All wizards data will be stored in this variable
   *
   * cluster - cluster name
   * hosts - hosts, ssh key, repo info, etc.
   * services - services list
   * hostsInfo - list of selected hosts
   * slaveComponentHosts, hostSlaveComponents - info about slave hosts
   * masterComponentHosts - info about master hosts
   * serviceConfigGroups - info about selected config group for service
   * configGroups - all config groups
   * config??? - to be described later
   */
  content: Em.Object.create({
    cluster: null,
    hosts: null,
    installOptions: null,
    services: null,
    slaveComponentHosts: null,
    masterComponentHosts: null,
    serviceConfigProperties: null,
    advancedServiceConfig: null,
    controllerName: 'addHostController',
    serviceConfigGroups: null,
    configGroups: null
  }),

  /**
   * save info about wizard progress, particularly current step of wizard
   * @param currentStep
   * @param completed
   */
  setCurrentStep: function (currentStep, completed) {
    this._super(currentStep, completed);
    App.clusterStatus.setClusterStatus({
      wizardControllerName: this.get('name'),
      localdb: App.db.data
    });
  },

  /**
   * return new object extended from clusterStatusTemplate
   * @return Object
   */
  getCluster: function () {
    return jQuery.extend({}, this.get('clusterStatusTemplate'), {name: App.router.getClusterName()});
  },

  /**
   * Remove host from model. Used at <code>Confirm hosts</code> step
   * @param hosts Array of hosts, which we want to delete
   */
  removeHosts: function (hosts) {
    var dbHosts = this.getDBProperty('hosts');
    hosts.forEach(function (_hostInfo) {
      var host = _hostInfo.name;
      delete dbHosts[host];
    });
    this.setDBProperty('hosts', dbHosts);
  },

  /**
   * Load services data. Will be used at <code>Select services(step4)</code> step
   */
  loadServices: function () {
    var services = this.getDBProperty('services');
    if (!services) {
      services = {
        selectedServices: [],
        installedServices: []
      };
      App.StackService.find().forEach(function (item) {
        var isInstalled = App.Service.find().someProperty('serviceName', item.get('serviceName'));
        item.set('isSelected', isInstalled);
        item.set('isInstalled', isInstalled);
        if (isInstalled) {
          services.selectedServices.push(item.get('serviceName'));
          services.installedServices.push(item.get('serviceName'));
        }
      }, this);
      this.setDBProperty('services', services);
    } else {
      App.StackService.find().forEach(function (item) {
        var isSelected = services.selectedServices.contains(item.get('serviceName'));
        var isInstalled = services.installedServices.contains(item.get('serviceName'));
        item.set('isSelected', isSelected);
        item.set('isInstalled', isInstalled);
      }, this);
    }
    this.set('content.services', App.StackService.find());
  },

  /**
   * Load slave component hosts data for using in required step controllers
   * TODO move to mixin
   */
  loadSlaveComponentHosts: function () {
    var slaveComponentHosts = this.getDBProperty('slaveComponentHosts') || [];
    if (slaveComponentHosts.length) {
      var hosts = this.getDBProperty('hosts'),
          host_names = Em.keys(hosts);
      slaveComponentHosts.forEach(function (component) {
        component.hosts.forEach(function (host) {
          //Em.set(host, 'hostName', hosts[host.host_id].name);
          for (var i = 0; i < host_names.length; i++) {
            if (hosts[host_names[i]].id === host.host_id) {
              host.hostName = host_names[i];
              break;
            }
          }
        });
      });
    }
    this.set("content.slaveComponentHosts", slaveComponentHosts);
    console.log("AddHostController.loadSlaveComponentHosts: loaded hosts ", slaveComponentHosts);
  },

  /**
   * Generate clients list for selected services and save it to model
   */
  saveClients: function () {
    var serviceComponents = App.StackServiceComponent.find();
    var services = this.get('content.services').filterProperty('isInstallable').filterProperty('isSelected');
    var clients = this.getClientsToInstall(services, serviceComponents);
    this.setDBProperty('clientInfo', clients);
    this.set('content.clients', clients);
    console.log("AddHostController.saveClients: saved list ", clients);
  },

  /**
   * get list of clients which will be installed on host
   * @param services {Array} of service objects
   * @param components {Array} of component objects
   * @returns {Array} returns array of clients
   * @method getClientsToInstall;
   */
  getClientsToInstall: function(services, components) {
    var clients = [];
    services.forEach(function (_service) {
      var serviceClients = components.filter(function(component) {
        return (component.get('serviceName') == _service.get('serviceName') && component.get('isClient'));
      });
      if (serviceClients.length) {
        serviceClients.forEach(function(client) {
          clients.push({
            component_name: client.get('componentName'),
            display_name: client.get('displayName'),
            isInstalled: false
          });
        });
      }
    }, this);
    return clients;
  },
  /**
   *  Apply config groups from step4 Configurations
   */
  applyConfigGroup: function () {
    var serviceConfigGroups = this.get('content.configGroups');
    serviceConfigGroups.forEach(function (group) {
      if (group.configGroups.someProperty('ConfigGroup.group_name', group.selectedConfigGroup)) {
        var configGroup = group.configGroups.findProperty('ConfigGroup.group_name', group.selectedConfigGroup);
        group.hosts.forEach(function (host) {
          configGroup.ConfigGroup.hosts.push({
            host_name: host
          });
        }, this);
        delete configGroup.href;
        App.ajax.send({
          name: 'config_groups.update_config_group',
          sender: this,
          data: {
            id: configGroup.ConfigGroup.id,
            configGroup: configGroup
          }
        });
      }
    }, this);
  },

  /**
   * Load information about selected config groups
   */
  getServiceConfigGroups: function () {
    var serviceConfigGroups = this.getDBProperty('serviceConfigGroups');
    this.set('content.configGroups', serviceConfigGroups);
  },

  /**
   * Save information about selected config groups
   */
  saveServiceConfigGroups: function () {
    this.setDBProperty('serviceConfigGroups', this.get('content.configGroups'));
  },

  /**
   * Set content.configGroups for step4
   */
  loadServiceConfigGroups: function () {
    var selectedServices = [];
    this.loadServiceConfigGroupsBySlaves(selectedServices);
    this.loadServiceConfigGroupsByClients(selectedServices);
    this.sortServiceConfigGroups(selectedServices);
    this.set('content.configGroups', selectedServices);
  },
  /**
   * sort config groups by name
   * @param selectedServices
   */
  sortServiceConfigGroups: function (selectedServices) {
    selectedServices.forEach(function (selectedService) {
      selectedService.configGroups.sort(function (cfgA, cfgB) {
        if (cfgA.ConfigGroup.group_name < cfgB.ConfigGroup.group_name) return -1;
        if (cfgA.ConfigGroup.group_name > cfgB.ConfigGroup.group_name) return 1;
        return 0;
      });
    });
  },
  /**
   * load service config groups by slave components,
   * push them into selectedServices
   * @param selectedServices
   */
  loadServiceConfigGroupsBySlaves: function (selectedServices) {
    var slaveComponentHosts = this.get('content.slaveComponentHosts');
    if (slaveComponentHosts && slaveComponentHosts.length > 0) {
      slaveComponentHosts.forEach(function (slave) {
        if (slave.hosts.length > 0) {
          if (slave.componentName !== "CLIENT") {
            var service = App.StackServiceComponent.find(slave.componentName).get('stackService');
            var serviceName = service.get('serviceName');
            var configGroups = this.get('content.configGroups').filterProperty('ConfigGroup.tag', serviceName);
            var configGroupsNames = configGroups.mapProperty('ConfigGroup.group_name');
            var defaultGroupName = service.get('displayName') + ' Default';
            configGroupsNames.unshift(defaultGroupName);
            selectedServices.push({
              serviceId: serviceName,
              displayName: service.get('displayName'),
              hosts: slave.hosts.mapProperty('hostName'),
              configGroupsNames: configGroupsNames,
              configGroups: configGroups,
              selectedConfigGroup: defaultGroupName
            });
          }
        }
      }, this);
      return true;
    }
    return false;
  },
  /**
   * load service config groups by clients,
   * push them into selectedServices
   * @param selectedServices
   */
  loadServiceConfigGroupsByClients: function (selectedServices) {
    var slaveComponentHosts = this.get('content.slaveComponentHosts');
    var clients = this.get('content.clients');
    var client = slaveComponentHosts && slaveComponentHosts.findProperty('componentName', 'CLIENT');
    var selectedClientHosts = client && client.hosts.mapProperty('hostName');
    if (clients && selectedClientHosts && clients.length > 0 && selectedClientHosts.length > 0) {
      this.loadClients();
      clients.forEach(function (client) {
        var service = App.StackServiceComponent.find(client.component_name).get('stackService');
        var serviceName = service.get('serviceName');
        var serviceMatch = selectedServices.findProperty('serviceId', serviceName);
        if (serviceMatch) {
          serviceMatch.hosts = serviceMatch.hosts.concat(selectedClientHosts).uniq();
        } else {
          var configGroups = this.get('content.configGroups').filterProperty('ConfigGroup.tag', serviceName);
          var configGroupsNames = configGroups.mapProperty('ConfigGroup.group_name').sort();
          var defaultGroupName = service.get('displayName') + ' Default';
          configGroupsNames.unshift(defaultGroupName);
          selectedServices.push({
            serviceId: serviceName,
            displayName: service.get('displayName'),
            hosts: selectedClientHosts,
            configGroupsNames: configGroupsNames,
            configGroups: configGroups,
            selectedConfigGroup: defaultGroupName
          });
        }
      }, this);
      return true;
    }
    return false;
  },

  loadServiceConfigProperties: function () {
    var serviceConfigProperties = App.db.get('AddService', 'serviceConfigProperties');
    if (!serviceConfigProperties || !serviceConfigProperties.length) {
      serviceConfigProperties = App.db.get('Installer', 'serviceConfigProperties');
    }
    this.set('content.serviceConfigProperties', serviceConfigProperties);
    console.log("AddHostController.loadServiceConfigProperties: loaded config ", serviceConfigProperties);
  },
  /**
   * Load data for all steps until <code>current step</code>
   */
  loadAllPriorSteps: function () {
    var step = this.get('currentStep');
    switch (step) {
      case '7':
      case '6':
      case '5':
        this.loadServiceConfigProperties();
        this.getServiceConfigGroups();
      case '4':
      case '3':
        this.loadClients();
        this.loadServices();
        this.loadMasterComponentHosts();
        this.loadSlaveComponentHosts();
        this.load('hosts');
      case '2':
        this.loadServices();
      case '1':
        this.load('hosts');
        this.load('installOptions');
        this.load('cluster');
    }
  },

  /**
   * Remove all loaded data.
   * Created as copy for App.router.clearAllSteps
   */
  clearAllSteps: function () {
    this.clearInstallOptions();
    // clear temporary information stored during the install
    this.set('content.cluster', this.getCluster());
  },

  clearStorageData: function () {
    this._super();
    this.resetDbNamespace();
  },

  /**
   * Clear all temporary data
   */
  finish: function () {
    this.setCurrentStep('1');
    this.clearAllSteps();
    this.clearStorageData();
    App.router.get('updateController').updateAll();
    App.updater.immediateRun('updateHost');
    App.router.get('clusterController').getAllHostNames();
  },

  /**
   * send request to server in order to install services
   * @param isRetry
   */
  installServices: function (isRetry, callback, errorCallback) {
    callback = callback || Em.K;
    this.set('content.cluster.oldRequestsId', []);
    var clusterName = this.get('content.cluster.name');
    var hostNames = [];
    for (var hostname in this.getDBProperty('hosts')) {
      if(this.getDBProperty('hosts')[hostname].isInstalled == false){
        hostNames.push(hostname);
      }
    }
    if(!clusterName || hostNames.length === 0) return false;

    App.ajax.send({
      name: "common.host_components.update",
      sender: this,
      data: {
        "context": Em.I18n.t('requestInfo.installComponents'),
        "query": "HostRoles/host_name.in(" + hostNames.join(',') + ")",
        "HostRoles": {"state": "INSTALLED"},
        "level": "HOST_COMPONENT"
      },
      success: 'installServicesSuccessCallback',
      error: 'installServicesErrorCallback'
    }).then(callback, errorCallback || callback);
    return true;
  }
});

});

require.register("controllers/main/host/configs_service", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.MainHostServiceConfigsController = App.MainServiceInfoConfigsController.extend({
  name: 'mainHostServiceConfigsController',
  host: null,
  isHostsConfigsPage: true,
  typeTagToHostMap: null,
  configKeyToConfigMap: null,

  /**
   * On load function
   */
  loadStep: function () {
    var content = this.get('content');
    this.set('host', content.host);
    this._super();
  },

  /**
   * Removes categories which are not valid for this host. Ex: Remove JOBTRACKER
   * category on host which does not have it installed.
   */
  renderServiceConfigs: function (serviceConfigs) {
    var newServiceConfigs = jQuery.extend({}, serviceConfigs);
    newServiceConfigs.configCategories = this.filterServiceConfigs(serviceConfigs.configCategories);
    this._super(newServiceConfigs);
  },
  /**
   * filter config categories by host-component of host
   * @param configCategories
   * @return {Array}
   */
  filterServiceConfigs: function (configCategories) {
    var hostComponents = this.get('host.hostComponents');
    var hostHostComponentNames = (hostComponents) ? hostComponents.mapProperty('componentName') : [];

    return configCategories.filter(function (category) {
      var hcNames = category.get('hostComponentNames');
      if (hcNames && hcNames.length > 0) {
        for (var i = 0, l = hcNames.length; i < l; i++) {
          if (hostHostComponentNames.contains(hcNames[i])) {
            return true;
          }
        }
        return false;
      }
      return true;
    });
  },

  /**
   * invoke dialog for switching group of host
   */
  switchHostGroup: function () {
    var self = this;
    App.config.launchSwitchConfigGroupOfHostDialog(this.get('selectedConfigGroup'), this.get('configGroups'), this.get('host.hostName'), function (newGroup) {
      self.set('selectedConfigGroup', newGroup);
    });
  }
});

});

require.register("controllers/main/host/details", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var batchUtils = require('utils/batch_scheduled_requests');
var componentsUtils = require('utils/components');
var stringUtils = require('utils/string_utils');

App.MainHostDetailsController = Em.Controller.extend({

  name: 'mainHostDetailsController',

  /**
   * Viewed host
   * @type {App.Host|null}
   */
  content: null,

  /**
   * Does user come from hosts page
   * @type {bool}
   */
  isFromHosts: false,

  /**
   * path to page visited before
   * @type {string}
   */
  referer: '',

  /**
   *  Host on which Hive Metastore will be added
   * @type {string}
   */
  hiveMetastoreHost: '',

  /**
   * Deferred object will be resolved when Oozie configs are downloaded
   * @type {object}
   */
  isOozieConfigLoaded: $.Deferred(),

  /**
   * @type {bool}
   */
  isOozieServerAddable: true,

  /**
   * Open dashboard page
   * @method routeHome
   */
  routeHome: function () {
    App.router.transitionTo('main.dashboard.index');
  },

  /**
   * List of active (not in passive state) host components
   * @type {Ember.Enumerable}
   */
  serviceActiveComponents: function () {
    return this.get('content.hostComponents').filterProperty('service.isInPassive', false);
  }.property('content.hostComponents'),

  /**
   * List of active host components which aren't clients
   * @type {Ember.Enumerable}
   */
  serviceNonClientActiveComponents: function () {
    return this.get('serviceActiveComponents').filterProperty('isClient', false);
  }.property('serviceActiveComponents'),

  /**
   * send command to server to start selected host component
   * @param {object} event
   * @method startComponent
   */
  startComponent: function (event) {
    var self = this;
    return App.showConfirmationPopup(function () {
      var component = event.context;
      var context = Em.I18n.t('requestInfo.startHostComponent') + " " + component.get('displayName');
      self.sendComponentCommand(component, context, App.HostComponentStatus.started);
    });
  },

  /**
   * send command to server to stop selected host component
   * @param {object} event
   * @method startComponent
   */
  stopComponent: function (event) {
    var self = this;
    return App.showConfirmationPopup(function () {
      var component = event.context;
      var context = Em.I18n.t('requestInfo.stopHostComponent') + " " + component.get('displayName');
      self.sendComponentCommand(component, context, App.HostComponentStatus.stopped);
    });
  },
  /**
   * PUTs a command to server to start/stop a component. If no
   * specific component is provided, all components are started.
   * @param {object} component  When <code>null</code> all startable components are started.
   * @param {String} context  Context under which this command is beign sent.
   * @param {String} state - desired state of component can be 'STARTED' or 'STOPPED'
   * @method sendComponentCommand
   */
  sendComponentCommand: function (component, context, state) {
    var data = {
      hostName: this.get('content.hostName'),
      context: context,
      component: component,
      HostRoles: {
        state: state
      }
    };
    if (Array.isArray(component)) {
      data.query = "HostRoles/component_name.in(" + component.mapProperty('componentName').join(',') + ")";
    } else {
      data.componentName = component.get('componentName');
      data.serviceName = component.get('service.serviceName');
    }
    App.ajax.send({
      name: (Array.isArray(component)) ? 'common.host.host_components.update' : 'common.host.host_component.update',
      sender: this,
      data: data,
      success: 'sendComponentCommandSuccessCallback',
      error: 'ajaxErrorCallback'
    });
  },

  /**
   * Success callback for stop/start host component request
   * @param {object} data
   * @param {object} opt
   * @param {object} params
   * @method stopComponentSuccessCallback
   */
  sendComponentCommandSuccessCallback: function (data, opt, params) {
    var running = (params.HostRoles.state === App.HostComponentStatus.stopped) ? App.HostComponentStatus.stopping : App.HostComponentStatus.starting;
    console.log('Send request for ' + running + ' successfully');
    params.component.set('workStatus', running);
    if (App.get('testMode')) {
      this.mimicWorkStatusChange(params.component, running, params.HostRoles.state);
    }
    this.showBackgroundOperationsPopup();
  },

  /**
   * Default error-callback for ajax-requests in current page
   * @param {object} request
   * @param {object} ajaxOptions
   * @param {string} error
   * @param {object} opt
   * @param {object} params
   * @method ajaxErrorCallback
   */
  ajaxErrorCallback: function (request, ajaxOptions, error, opt, params) {
    return componentsUtils.ajaxErrorCallback(request, ajaxOptions, error, opt, params);
  },
  /**
   * mimic status transition in test mode
   * @param entity
   * @param transitionalState
   * @param finalState
   */
  mimicWorkStatusChange: function (entity, transitionalState, finalState) {
    if (Em.isArray(entity)) {
      entity.forEach(function (item) {
        item.set('workStatus', transitionalState);
        setTimeout(function () {
          item.set('workStatus', finalState);
        }, App.testModeDelayForActions);
      });
    } else {
      entity.set('workStatus', transitionalState);
      setTimeout(function () {
        entity.set('workStatus', finalState);
      }, App.testModeDelayForActions);
    }
  },

  /**
   * load data (if we need to show this background operations popup) from persist
   * @param callback
   */
  showBackgroundOperationsPopup: function (callback) {
    App.router.get('applicationController').dataLoading().done(function (initValue) {
      if (initValue) {
        App.router.get('backgroundOperationsController').showPopup();
      }
      if (typeof callback === 'function') {
        callback();
      }
    });
  },

  /**
   * Send command to server to delete selected host component
   * @param {object} event
   * @method deleteComponent
   */
  deleteComponent: function (event) {
    var self = this;
    var component = event.context;
    var componentName = component.get('componentName');
    var displayName = component.get('displayName');
    var isLastComponent = (this.getTotalComponent(component) === 1);
    return App.ModalPopup.show({
      header: Em.I18n.t('popup.confirmation.commonHeader'),
      primary: Em.I18n.t('hosts.host.deleteComponent.popup.confirm'),
      bodyClass: Em.View.extend({
        templateName: require('templates/main/host/details/deleteComponentPopup')
      }),
      isHiveMetastore: function () {
        return componentName == 'HIVE_METASTORE';
      }.property(),
      deleteHiveMetastoreMsg: Em.View.extend({
        template: Em.Handlebars.compile(Em.I18n.t('hosts.host.deleteComponent.popup.deleteHiveMetastore'))
      }),
      isChecked: false,
      disablePrimary: function () {
        return !this.get('isChecked');
      }.property('isChecked'),
      lastComponent: function () {
        this.set('isChecked', !isLastComponent);
        return isLastComponent;
      }.property(),
      isZkServer: function () {
        return componentName == 'ZOOKEEPER_SERVER';
      }.property(),
      lastComponentError: Em.View.extend({
        template: Em.Handlebars.compile(Em.I18n.t('hosts.host.deleteComponent.popup.warning').format(displayName))
      }),
      deleteComponentMsg: function () {
        return Em.I18n.t('hosts.host.deleteComponent.popup.msg1').format(displayName);
      }.property(),
      deleteZkServerMsg: Em.View.extend({
        template: Em.Handlebars.compile(Em.I18n.t('hosts.host.deleteComponent.popup.deleteZooKeeperServer'))
      }),
      onPrimary: function () {
        var popup = this;
        self._doDeleteHostComponent(component, function () {
          self.set('redrawComponents', true);
          popup.hide();
        });
      }
    });
  },

  /**
   * get total count of host-components
   * @method getTotalComponent
   * @param component
   * @return {Number}
   */
  getTotalComponent: function (component) {
    var count;
    if (component.get('isSlave')) {
      count = App.SlaveComponent.find(component.get('componentName')).get('totalCount');
    } else if (component.get('isClient')) {
      count = App.ClientComponent.find(component.get('componentName')).get('totalCount');
    } else {
      count = App.HostComponent.find().filterProperty('componentName', component.get('componentName')).get('length');
    }
    return count || 0;
  },

  /**
   * Trigger to reset list of master/slaves components on the view
   * @type {bool}
   */
  redrawComponents: false,

  /**
   * Deletes the given host component, or all host components.
   *
   * @param {object|null} component  When <code>null</code> all host components are deleted.
   * @return  <code>null</code> when components get deleted.
   *          <code>{xhr: XhrObj, url: "http://", method: "DELETE"}</code>
   *          when components failed to get deleted.
   * @method _doDeleteHostComponent
   */
  _doDeleteHostComponent: function (component, callback) {
    callback = callback || Em.K;
    App.ajax.send({
      name: (Em.isNone(component)) ? 'common.delete.host' : 'common.delete.host_component',
      sender: this,
      data: {
        componentName: (component) ? component.get('componentName') : '',
        hostName: this.get('content.hostName')
      },
      success: '_doDeleteHostComponentSuccessCallback',
      error: '_doDeleteHostComponentErrorCallback'
    }).then(callback, callback);
  },

  /**
   * Result of delete component(s) request
   * @type {object}
   */
  _deletedHostComponentResult: null,

  /**
   * Success callback for delete host component request
   * @method _doDeleteHostComponentSuccessCallback
   */
  _doDeleteHostComponentSuccessCallback: function (response, request, data) {
    this.set('_deletedHostComponentResult', null);
    this.removeHostComponentModel(data.componentName, data.hostName);
    if (data.componentName == 'ZOOKEEPER_SERVER') {
      this.set('fromDeleteZkServer', true);
      this.loadConfigs();
    } else if (data.componentName == 'HIVE_METASTORE') {
      this.set('deleteHiveMetaStore', true);
      this.loadConfigs('loadHiveConfigs');
    }
  },

  /**
   * Error-callback for delete host component request
   * @param {object} xhr
   * @param {string} textStatus
   * @param {object} errorThrown
   * @method _doDeleteHostComponentErrorCallback
   */
  _doDeleteHostComponentErrorCallback: function (xhr, textStatus, errorThrown, data) {
    console.log('Error deleting host component');
    console.log(textStatus);
    console.log(errorThrown);
    this.set('_deletedHostComponentResult', {xhr: xhr, url: data.url, method: 'DELETE'});
  },

  /**
   * Remove host component data from App.HostComponent model.
   *
   * @param {String} componentName
   * @param {String} hostName
   */
  removeHostComponentModel: function(componentName, hostName) {
    var component = App.HostComponent.find().filterProperty('componentName', componentName).findProperty('hostName', hostName);
    App.serviceMapper.deleteRecord(component);
  },

  /**
   * Send command to server to upgrade selected host component
   * @param {object} event
   * @method upgradeComponent
   */
  upgradeComponent: function (event) {
    var self = this;
    var component = event.context;
    return App.showConfirmationPopup(function () {
      App.ajax.send({
        name: 'host.host_component.upgrade',
        sender: self,
        data: {
          component: component,
          hostName: self.get('content.hostName'),
          componentName: component.get('componentName'),
          data: JSON.stringify({
            RequestInfo: {
              "context": Em.I18n.t('requestInfo.upgradeHostComponent') + " " + component.get('displayName')
            },
            Body: {
              HostRoles: {
                stack_id: 'HDP-1.2.2',
                state: 'INSTALLED'
              }
            }
          })
        },
        success: 'upgradeComponentSuccessCallback',
        error: 'ajaxErrorCallback'
      });
    });
  },

  /**
   * Success callback for upgrade host component request
   * @param {object} data
   * @param {object} opt
   * @param {object} params
   * @method upgradeComponentSuccessCallback
   */
  upgradeComponentSuccessCallback: function (data, opt, params) {
    console.log('Send request for UPGRADE successfully');

    if (App.get('testMode')) {
      this.mimicWorkStatusChange(params.component, App.HostComponentStatus.starting, App.HostComponentStatus.started);
    }
    this.showBackgroundOperationsPopup();
  },

  /**
   * Send command to server to restart selected components
   * @param {object} event
   * @method restartComponent
   */
  restartComponent: function (event) {
    var component = event.context;
    return App.showConfirmationPopup(function () {
      batchUtils.restartHostComponents([component], Em.I18n.t('rollingrestart.context.selectedComponentOnSelectedHost').format(component.get('displayName')), "HOST_COMPONENT");
    });
  },
  /**
   * get current status of security settings,
   * if true security is enabled otherwise disabled
   * @return {Boolean}
   */
  securityEnabled: function () {
    return App.router.get('mainAdminSecurityController.securityEnabled');
  }.property('App.router.mainAdminSecurityController.securityEnabled'),


  /**
   * add component as <code>addComponent<code> method but perform
   * kdc sessionstate if cluster is secure;
   * @param event
   */
  addComponentWithCheck: function(event) {
    App.get('router.mainAdminKerberosController').getKDCSessionState(this.addComponent.bind(this, event));
  },
  /**
   * Send command to server to install selected host component
   * @param {object} event
   * @method addComponent
   */
  addComponent: function (event) {
    var
      returnFunc,
      self = this,
      hiveHost = event.hiveMetastoreHost ? event.hiveMetastoreHost : this.get('content.hostName');
      component = event.context,
      componentName = component.get('componentName'),
      missedComponents = !!hiveHost ? [] : componentsUtils.checkComponentDependencies(componentName, {
        scope: 'host',
        installedComponents: this.get('content.hostComponents').mapProperty('componentName')
      });
    if (!!missedComponents.length) {
      var popupMessage = Em.I18n.t('host.host.addComponent.popup.dependedComponents.body').format(component.get('displayName'),
        stringUtils.getFormattedStringFromArray(missedComponents.map(function(cName) {
          return App.StackServiceComponent.find(cName).get('displayName');
        })));
      return App.showAlertPopup(Em.I18n.t('host.host.addComponent.popup.dependedComponents.header'), popupMessage);
    }

    switch (componentName) {
      case 'ZOOKEEPER_SERVER':
        returnFunc = App.showConfirmationPopup(function () {
          self.primary(component);
        }, Em.I18n.t('hosts.host.addComponent.' + componentName ));
        break;
      case 'HIVE_METASTORE':
        returnFunc = App.showConfirmationPopup(function () {
          self.set('hiveMetastoreHost', hiveHost);
          self.loadConfigs("loadHiveConfigs");
        }, Em.I18n.t('hosts.host.addComponent.' + componentName ));
        break;
      default:
        returnFunc = this.addClientComponent(component);
      }
    return returnFunc;
  },
  /**
   * Send command to server to install client on selected host
   * @param component
   */
  addClientComponent: function (component) {
    var self = this;
    var message = this.formatClientsMessage(component);
    return App.ModalPopup.show({
      primary: Em.I18n.t('hosts.host.addComponent.popup.confirm'),
      header: Em.I18n.t('popup.confirmation.commonHeader'),

      addComponentMsg: function () {
        return Em.I18n.t('hosts.host.addComponent.msg').format(message);
      }.property(),

      bodyClass: Em.View.extend({
        templateName: require('templates/main/host/details/addComponentPopup')
      }),

      onPrimary: function () {
        this.hide();
        self.primary(component);
      }
    });
  },

  /**
   * format message for operation of adding clients
   * @param client
   */
  formatClientsMessage: function (client) {
    var displayName = Em.isNone(client.get('displayName')) ? '' : client.get('displayName');
    var subComponentNames = client.get('subComponentNames');
    if (subComponentNames && subComponentNames.length > 0) {
      var dns = [];
      subComponentNames.forEach(function (scn) {
        dns.push(App.format.role(scn));
      });
      displayName += " (" + dns.join(", ") + ")";
    }
    return displayName;
  },

  /**
   * Send request to add host component
   * @param {App.HostComponent} component
   * @method primary
   */
  primary: function (component) {
    var self = this;
    componentsUtils.installHostComponent(self.get('content.hostName'), component);
  },

  /**
   * Success callback for install host component request (sent in <code>addNewComponentSuccessCallback</code>)
   * @param {object} data
   * @param {object} opt
   * @param {object} params
   * @method installNewComponentSuccessCallbƒack
   */
  installNewComponentSuccessCallback: function (data, opt, params) {
    if (!data.Requests || !data.Requests.id) {
      return false;
    }
    var self = this;
    console.log('Send request for INSTALLING NEW COMPONENT successfully');

    if (App.get('testMode')) {
      this.mimicWorkStatusChange(params.component, App.HostComponentStatus.installing, App.HostComponentStatus.stopped);
    }

    this.showBackgroundOperationsPopup(function () {
      if (params.componentName === 'ZOOKEEPER_SERVER' || params.componentName === 'HIVE_SERVER') {
        self.set(params.componentName === 'ZOOKEEPER_SERVER' ? 'zkRequestId' : 'hiveRequestId', data.Requests.id);
        self.addObserver(
          'App.router.backgroundOperationsController.serviceTimestamp',
          self,
          (params.componentName === 'ZOOKEEPER_SERVER' ? self.checkZkConfigs : self.checkHiveDone)
          );
        params.componentName === 'ZOOKEEPER_SERVER' ? self.checkZkConfigs() : self.checkHiveDone();
      }
    });
    return true;
  },

  /**
   * Call load tags
   * @method checkHiveDone
   */
  checkHiveDone: function () {
    var bg = App.router.get('backgroundOperationsController.services').findProperty('id', this.get('hiveRequestId'));
    if (bg && !bg.get('isRunning')) {
      var self = this;
      this.removeObserver('App.router.backgroundOperationsController.serviceTimestamp', this, this.checkHiveDone);
      setTimeout(function () {
        self.loadConfigs("loadHiveConfigs");
      }, App.get('componentsUpdateInterval'));
    }
  },

  /**
   * Success callback for load configs request
   * @param {object} data
   * @method loadOozieConfigs
   */
  loadOozieConfigs: function (data) {
    return App.ajax.send({
      name: 'admin.get.all_configurations',
      sender: this,
      data: {
        urlParams: '(type=oozie-env&tag=' + data.Clusters.desired_configs['oozie-env'].tag + ')'
      },
      success: 'onLoadOozieConfigs',
      error: 'onLoadConfigsErrorCallback'
    });
  },

  /**
   * get Oozie database config and set databaseType
   * @param {object} data
   * @method onLoadHiveConfigs
   */
  onLoadOozieConfigs: function (data) {
    var configs = {};
    data.items.forEach(function(item) {
      $.extend(configs, item.properties);
    });
    this.set('isOozieServerAddable', !(Em.isEmpty(configs["oozie_database"]) || configs["oozie_database"] === 'New Derby Database'));
    this.get('isOozieConfigLoaded').resolve();
  },

  /**
   * Success callback for load configs request
   * @param {object} data
   * @method loadHiveConfigs
   */
  loadHiveConfigs: function (data) {
    App.ajax.send({
      name: 'admin.get.all_configurations',
      sender: this,
      data: {
        urlParams: '(type=hive-site&tag=' + data.Clusters.desired_configs['hive-site'].tag + ')|(type=webhcat-site&tag=' +
          data.Clusters.desired_configs['webhcat-site'].tag + ')|(type=hive-env&tag=' + data.Clusters.desired_configs['hive-env'].tag +
          ')|(type=core-site&tag=' + data.Clusters.desired_configs['core-site'].tag + ')'
      },
      success: 'onLoadHiveConfigs'
    });
  },

  /**
   * update and save Hive related configs to server
   * @param {object} data
   * @method onLoadHiveConfigs
   */
  onLoadHiveConfigs: function (data) {
    var
      hiveMetastoreHost = this.get('hiveMetastoreHost'),
      hiveMSHosts = this.getHiveHosts(),
      hiveMasterHosts = hiveMSHosts.concat(App.HostComponent.find().filterProperty('componentName', 'HIVE_SERVER').mapProperty('hostName')).uniq().sort().join(','),
      configs = {},
      attributes = {},
      port = "",
      hiveUser = "",
      webhcatUser = "";

    data.items.forEach(function (item) {
      configs[item.type] = item.properties;
      attributes[item.type] = item.properties_attributes || {};
    }, this);

    port = configs['hive-site']['hive.metastore.uris'].match(/:[0-9]{2,4}/);
    port = port ? port[0].slice(1) : "9083";

    hiveUser = configs['hive-env']['hive_user'];
    webhcatUser = configs['hive-env']['webhcat_user'];

    for (var i = 0; i < hiveMSHosts.length; i++) {
      hiveMSHosts[i] = "thrift://" + hiveMSHosts[i] + ":" + port;
    }
    configs['hive-site']['hive.metastore.uris'] = hiveMSHosts.join(',');
    configs['webhcat-site']['templeton.hive.properties'] = configs['webhcat-site']['templeton.hive.properties'].replace(/thrift.+[0-9]{2,},/i, hiveMSHosts.join('\\,') + ",");
    configs['core-site']['hadoop.proxyuser.' + hiveUser + '.hosts'] = hiveMasterHosts;
    configs['core-site']['hadoop.proxyuser.' + webhcatUser + '.hosts'] = hiveMasterHosts;
    var groups = [
      {
        properties: {
          'hive-site': configs['hive-site'],
          'webhcat-site': configs['webhcat-site'],
          'hive-env': configs['hive-env']
        },
        properties_attributes: {
          'hive-site': attributes['hive-site'],
          'webhcat-site': attributes['webhcat-site'],
          'hive-env': attributes['hive-env']
        }
      },
      {
        properties: {
          'core-site': configs['core-site']
        },
        properties_attributes: {
          'core-site': attributes['core-site']
        }
      }
    ];
    this.saveConfigsBatch(groups, hiveMetastoreHost);
  },

  /**
   * save configs' sites in batch
   * @param hiveMetastoreHost
   * @param groups
   */
  saveConfigsBatch: function (groups, hiveMetastoreHost) {
    groups.forEach(function (group) {
      var desiredConfigs = [],
        tag = 'version' + (new Date).getTime(),
        properties = group.properties;

      for (var site in properties) {
        if (!properties.hasOwnProperty(site) || Em.isNone(properties[site])) continue;
        desiredConfigs.push({
          "type": site,
          "tag": tag,
          "properties": properties[site],
          "properties_attributes": group.properties_attributes[site],
          "service_config_version_note": Em.I18n.t('hosts.host.hive.configs.save.note')
        });
      }
      if (desiredConfigs.length > 0) {
        App.ajax.send({
          name: 'common.service.configurations',
          sender: this,
          data: {
            desired_config: desiredConfigs,
            hiveMetastoreHost: hiveMetastoreHost
          },
          success: 'installHiveMetastore'
        });
      }
      //clear hive metastore host not to send second request to install component
      hiveMetastoreHost = null;
    }, this);
  },

  /**
   * success callback for saveConfigsBatch method
   * @param data
   * @param opt
   * @param params
   */
  installHiveMetastore: function(data, opt, params) {
    if (params.hiveMetastoreHost) {
      componentsUtils.installHostComponent(params.hiveMetastoreHost, App.StackServiceComponent.find("HIVE_METASTORE"));
    }
  },
  /**
   * Delete Hive Metastore is performed
   * @type {bool}
   */
  deleteHiveMetaStore: false,

  getHiveHosts: function () {
    var
      hiveHosts = App.HostComponent.find().filterProperty('componentName', 'HIVE_METASTORE').mapProperty('hostName'),
      hiveMetastoreHost = this.get('hiveMetastoreHost');

    if(!!hiveMetastoreHost){
      hiveHosts.push(hiveMetastoreHost);
      this.set('hiveMetastoreHost', '');
    }

    if (this.get('fromDeleteHost') || this.get('deleteHiveMetaStore')) {
      this.set('deleteHiveMetaStore', false);
      this.set('fromDeleteHost', false);
      return hiveHosts.without(this.get('content.hostName'));
    }
    return hiveHosts;
  },

  /**
   * Send command to server to resfresh configs of selected component
   * @param {object} event
   * @method refreshComponentConfigs
   */
  refreshComponentConfigs: function (event) {
    var self = this;
    return App.showConfirmationPopup(function () {
      var component = event.context;
      var context = Em.I18n.t('requestInfo.refreshComponentConfigs').format(component.get('displayName'));
      self.sendRefreshComponentConfigsCommand(component, context);
    });
  },

  /**
   * PUTs a command to server to refresh configs of host component.
   * @param {object} component
   * @param {object} context Context under which this command is beign sent.
   * @method sendRefreshComponentConfigsCommand
   */
  sendRefreshComponentConfigsCommand: function (component, context) {
    var resource_filters = [
      {
        service_name: component.get('service.serviceName'),
        component_name: component.get('componentName'),
        hosts: component.get('host.hostName')
      }
    ];
    App.ajax.send({
      name: 'host.host_component.refresh_configs',
      sender: this,
      data: {
        resource_filters: resource_filters,
        context: context
      },
      success: 'refreshComponentConfigsSuccessCallback'
    });
  },

  /**
   * Success callback for refresh host component configs request
   * @method refreshComponentConfigsSuccessCallback
   */
  refreshComponentConfigsSuccessCallback: function () {
    console.log('Send request for refresh configs successfully');
    this.showBackgroundOperationsPopup();
  },

  /**
   * Load tags
   * @method checkZkConfigs
   */
  checkZkConfigs: function () {
    var bg = App.router.get('backgroundOperationsController.services').findProperty('id', this.get('zkRequestId'));
    if (bg && !bg.get('isRunning')) {
      var self = this;
      this.removeObserver('App.router.backgroundOperationsController.serviceTimestamp', this, this.checkZkConfigs);
      setTimeout(function () {
        self.loadConfigs();
      }, App.get('componentsUpdateInterval'));
    }
  },

  /**
   * Load configs
   * @method loadConfigs
   */
  loadConfigs: function (callback) {
    App.ajax.send({
      name: 'config.tags',
      sender: this,
      success: callback ? callback : 'loadConfigsSuccessCallback',
      error: 'onLoadConfigsErrorCallback'
    });
  },

  /**
   * onLoadConfigsErrorCallback
   * @method onLoadConfigsErrorCallback
   */
  onLoadConfigsErrorCallback: function () {
    this.get('isOozieConfigLoaded').reject();
  },

  /**
   * Success callback for load configs request
   * @param {object} data
   * @method adConfigsSuccessCallback
   */
  loadConfigsSuccessCallback: function (data) {
    var urlParams = this.constructConfigUrlParams(data);
    if (urlParams.length > 0) {
      App.ajax.send({
        name: 'reassign.load_configs',
        sender: this,
        data: {
          urlParams: urlParams.join('|')
        },
        success: 'saveZkConfigs'
      });
      return true;
    }
    return false;
  },

  /**
   * construct URL params for query, that load configs
   * @param data {Object}
   * @return {Array}
   */
  constructConfigUrlParams: function (data) {
    var urlParams = [];
    var services = App.Service.find();
    if (App.get('isHaEnabled')) {
      urlParams.push('(type=core-site&tag=' + data.Clusters.desired_configs['core-site'].tag + ')');
    }
    if (services.someProperty('serviceName', 'HBASE')) {
      urlParams.push('(type=hbase-site&tag=' + data.Clusters.desired_configs['hbase-site'].tag + ')');
    }
    if (services.someProperty('serviceName', 'HIVE')) {
      urlParams.push('(type=webhcat-site&tag=' + data.Clusters.desired_configs['webhcat-site'].tag + ')');
      urlParams.push('(type=hive-site&tag=' + data.Clusters.desired_configs['hive-site'].tag + ')');
    }
    if (services.someProperty('serviceName', 'STORM')) {
      urlParams.push('(type=storm-site&tag=' + data.Clusters.desired_configs['storm-site'].tag + ')');
    }
    if ((services.someProperty('serviceName', 'YARN') && App.get('isHadoop22Stack')) || App.get('isRMHaEnabled')) {
      urlParams.push('(type=yarn-site&tag=' + data.Clusters.desired_configs['yarn-site'].tag + ')');
      urlParams.push('(type=zoo.cfg&tag=' + data.Clusters.desired_configs['zoo.cfg'].tag + ')');
    }
    return urlParams;
  },

  /**
   * save new ZooKeeper configs to server
   * @param {object} data
   * @method saveZkConfigs
   */
  saveZkConfigs: function (data) {
    var configs = {};
    var attributes = {};
    data.items.forEach(function (item) {
      configs[item.type] = item.properties;
      attributes[item.type] = item.properties_attributes || {};
    }, this);

    var zks = this.getZkServerHosts();
    var portValue = configs['zoo.cfg'] && Em.get(configs['zoo.cfg'], 'clientPort');
    var zkPort = typeof portValue === 'udefined' ? '2181' : portValue;
    var zksWithPort = this.concatZkNames(zks, zkPort);
    this.setZKConfigs(configs, zksWithPort, zks);
    var groups = [
      {
        properties: {
          'hive-site': configs['hive-site'],
          'webhcat-site': configs['webhcat-site']
        },
        properties_attributes: {
          'hive-site': attributes['hive-site'],
          'webhcat-site': attributes['webhcat-site']
        }
      }
    ];
    if ((App.Service.find().someProperty('serviceName', 'YARN') && App.get('isHadoop22Stack')) || App.get('isRMHaEnabled')) {
      groups.push(
        {
          properties: {
            'yarn-site': configs['yarn-site']
          },
          properties_attributes: {
            'yarn-site': attributes['yarn-site']
          }
        }
      );
    }
    this.saveConfigsBatch(groups);
  },
  /**
   *
   * Set new values for some configs (based on available ZooKeeper Servers)
   * @param configs {object}
   * @param zksWithPort {string}
   * @param zks {array}
   * @return {Boolean}
   */
  setZKConfigs: function (configs, zksWithPort, zks) {
    if (typeof configs !== 'object' || !Array.isArray(zks)) return false;
    if (App.get('isHaEnabled')) {
      configs['core-site']['ha.zookeeper.quorum'] = zksWithPort;
    }
    if (configs['hbase-site']) {
      configs['hbase-site']['hbase.zookeeper.quorum'] = zks.join(',');
    }
    if (configs['webhcat-site']) {
      configs['webhcat-site']['templeton.zookeeper.hosts'] = zksWithPort;
    }
    if (configs['hive-site']) {
      configs['hive-site']['hive.cluster.delegation.token.store.zookeeper.connectString'] = zksWithPort;
    }
    if (configs['storm-site']) {
      configs['storm-site']['storm.zookeeper.servers'] = JSON.stringify(zks).replace(/"/g, "'");
    }
    if (App.get('isRMHaEnabled')) {
      configs['yarn-site']['yarn.resourcemanager.zk-address'] = zksWithPort;
    }
    if (App.get('isHadoop22Stack')) {
      if (configs['hive-site']) {
        configs['hive-site']['hive.zookeeper.quorum'] = zksWithPort;
      }
      if (configs['yarn-site']) {
        configs['yarn-site']['hadoop.registry.zk.quorum'] = zksWithPort;
      }
    }
    return true;
  },
  /**
   * concatenate URLs to ZOOKEEPER hosts with port "2181",
   * as value of config divided by comma
   * @param zks {array}
   * @param port {string}
   */
  concatZkNames: function (zks, port) {
    var zks_with_port = '';
    zks.forEach(function (zk) {
      zks_with_port += zk + ':' + port + ',';
    });
    return zks_with_port.slice(0, -1);
  },

  /**
   * Is deleteHost action id fired
   * @type {bool}
   */
  fromDeleteHost: false,

  /**
   * Is ZooKeeper Server being deleted from host
   * @type {bool}
   */
  fromDeleteZkServer: false,

  /**
   * Get list of hostnames where ZK Server is installed
   * @returns {string[]}
   * @method getZkServerHosts
   */
  getZkServerHosts: function () {
    var zks = App.HostComponent.find().filterProperty('componentName', 'ZOOKEEPER_SERVER').mapProperty('hostName');
    if (this.get('fromDeleteHost') || this.get('fromDeleteZkServer')) {
      this.set('fromDeleteHost', false);
      this.set('fromDeleteZkServer', false);
      return zks.without(this.get('content.hostName'));
    }
    return zks;
  },

  /**
   * Send command to server to install selected host component
   * @param {Object} event
   * @method installComponent
   */
  installComponent: function (event) {
    var self = this;
    var component = event.context;
    var componentName = component.get('componentName');
    var displayName = component.get('displayName');

    return App.ModalPopup.show({
      primary: Em.I18n.t('hosts.host.installComponent.popup.confirm'),
      header: Em.I18n.t('popup.confirmation.commonHeader'),
      installComponentMessage: function () {
        return Em.I18n.t('hosts.host.installComponent.msg').format(displayName);
      }.property(),
      bodyClass: Em.View.extend({
        templateName: require('templates/main/host/details/installComponentPopup')
      }),
      onPrimary: function () {
        this.hide();

        App.ajax.send({
          name: 'common.host.host_component.update',
          sender: self,
          data: {
            hostName: self.get('content.hostName'),
            serviceName: component.get('service.serviceName'),
            componentName: componentName,
            component: component,
            context: Em.I18n.t('requestInfo.installHostComponent') + " " + displayName,
            HostRoles: {
              state: 'INSTALLED'
            }
          },
          success: 'installComponentSuccessCallback',
          error: 'ajaxErrorCallback'
        });
      }
    });
  },

  /**
   * Success callback for install component request
   * @param {object} data
   * @param {object} opt
   * @param {object} params
   * @method installComponentSuccessCallback
   */
  installComponentSuccessCallback: function (data, opt, params) {
    console.log('Send request for REINSTALL COMPONENT successfully');
    if (App.get('testMode')) {
      this.mimicWorkStatusChange(params.component, App.HostComponentStatus.installing, App.HostComponentStatus.stopped);
    }
    this.showBackgroundOperationsPopup();
  },

  /**
   * Send command to server to run decommission on DATANODE, TASKTRACKER, NODEMANAGER, REGIONSERVER
   * @param {App.HostComponent} component
   * @method decommission
   */
  decommission: function (component) {
    var self = this;
    return App.showConfirmationPopup(function () {
      self.runDecommission.call(self, self.get('content.hostName'), component.get('service.serviceName'));
    });
  },
  /**
   * identify correct component to run decommission on them by service name,
   * in result run proper decommission method
   * @param hostName
   * @param svcName
   */
  runDecommission: function (hostName, svcName) {
    switch (svcName) {
      case 'HDFS':
        this.doDecommission(hostName, svcName, "NAMENODE", "DATANODE");
        break;
      case 'YARN':
        this.doDecommission(hostName, svcName, "RESOURCEMANAGER", "NODEMANAGER");
        break;
      case 'HBASE':
        this.warnBeforeDecommission(hostName);
    }
  },

  /**
   * Send command to server to run recommission on DATANODE, TASKTRACKER, NODEMANAGER
   * @param {App.HostComponent} component
   * @method recommission
   */
  recommission: function (component) {
    var self = this;
    return App.showConfirmationPopup(function () {
      self.runRecommission.call(self, self.get('content.hostName'), component.get('service.serviceName'));
    });
  },
  /**
   * identify correct component to run recommission on them by service name,
   * in result run proper recommission method
   * @param hostName
   * @param svcName
   */
  runRecommission: function (hostName, svcName) {
    switch (svcName) {
      case 'HDFS':
        this.doRecommissionAndStart(hostName, svcName, "NAMENODE", "DATANODE");
        break;
      case 'YARN':
        this.doRecommissionAndStart(hostName, svcName, "RESOURCEMANAGER", "NODEMANAGER");
        break;
      case 'HBASE':
        this.doRecommissionAndStart(hostName, svcName, "HBASE_MASTER", "HBASE_REGIONSERVER");
    }
    this.showBackgroundOperationsPopup();
  },

  /**
   * Performs Decommission (for DN, TT and NM)
   * @param {string} hostName
   * @param {string} serviceName
   * @param {string} componentName
   * @param {string} slaveType
   * @method doDecommission
   */
  doDecommission: function (hostName, serviceName, componentName, slaveType) {
    var contextNameString = 'hosts.host.' + slaveType.toLowerCase() + '.decommission';
    var context = Em.I18n.t(contextNameString);
    App.ajax.send({
      name: 'host.host_component.decommission_slave',
      sender: this,
      data: {
        context: context,
        command: 'DECOMMISSION',
        hostName: hostName,
        serviceName: serviceName,
        componentName: componentName,
        slaveType: slaveType
      },
      success: 'decommissionSuccessCallback',
      error: 'decommissionErrorCallback'
    });
  },

  /**
   * check is hbase regionserver in mm. If so - run decommission
   * otherwise shows warning
   * @method warnBeforeDecommission
   * @param {string} hostNames - list of host when run from bulk operations or current host
   */
  warnBeforeDecommission: function (hostNames) {
    if (this.get('content.hostComponents').findProperty('componentName', 'HBASE_REGIONSERVER').get('passiveState') == "OFF") {
      this.showHbaseActiveWarning();
    } else {
      this.checkRegionServerState(hostNames);
    }
  },

  /**
   *  send call to check is this regionserver last in cluster which has desired_admin_state property "INSERVICE"
   * @method checkRegionServerState
   * @param hostNames
   */
  checkRegionServerState: function (hostNames) {
    return App.ajax.send({
      name: 'host.region_servers.in_inservice',
      sender: this,
      data: {
        hostNames: hostNames
      },
      success: 'checkRegionServerStateSuccessCallback'
    });
  },

  /**
   * check is this regionserver last in cluster which has desired_admin_state property "INSERVICE"
   * @method checkRegionServerStateSuccessCallback
   * @param data
   * @param opt
   * @param params
   */
  checkRegionServerStateSuccessCallback: function (data, opt, params) {
    var hostArray = params.hostNames.split(",");
    var decommissionPossible = (data.items.mapProperty('HostRoles.host_name').filter(function (hostName) {
      return !hostArray.contains(hostName);
    }, this).length >= 1);
    if (decommissionPossible) {
      this.doDecommissionRegionServer(params.hostNames, "HBASE", "HBASE_MASTER", "HBASE_REGIONSERVER");
    } else {
      this.showRegionServerWarning();
    }
  },

  /**
   * show warning that regionserver is last in cluster which has desired_admin_state property "INSERVICE"
   * @method showRegionServerWarning
   * @param hostNames
   */
  showRegionServerWarning: function () {
    return App.ModalPopup.show({
      header: Em.I18n.t('common.warning'),
      message: Em.I18n.t('hosts.host.hbase_regionserver.decommission.warning'),
      bodyClass: Ember.View.extend({
        template: Em.Handlebars.compile('<div class="alert alert-warning">{{message}}</div>')
      }),
      secondary: false
    });
  },

  /**
   * shows warning: put hbase regionserver in passive state
   * @method showHbaseActiveWarning
   * @return {App.ModalPopup}
   */
  showHbaseActiveWarning: function () {
    return App.ModalPopup.show({
      header: Em.I18n.t('common.warning'),
      message: function () {
        return Em.I18n.t('hostPopup.recommendation.beforeDecommission').format(App.format.components["HBASE_REGIONSERVER"]);
      }.property(),
      bodyClass: Ember.View.extend({
        template: Em.Handlebars.compile('<div class="alert alert-warning">{{message}}</div>')
      }),
      secondary: false
    });
  },

  /**
   * Performs Decommission (for RegionServer)
   * @method doDecommissionRegionServer
   * @param {string} hostNames - list of host when run from bulk operations or current host
   * @param {string} serviceName - serviceName
   * @param {string} componentName - master compoent name
   * @param {string} slaveType - slave component name
   */
  doDecommissionRegionServer: function (hostNames, serviceName, componentName, slaveType) {
    var batches = [
      {
        "order_id": 1,
        "type": "POST",
        "uri": App.get('apiPrefix') + "/clusters/" + App.get('clusterName') + "/requests",
        "RequestBodyInfo": {
          "RequestInfo": {
            "context": Em.I18n.t('hosts.host.regionserver.decommission.batch1'),
            "command": "DECOMMISSION",
            "exclusive" :"true",
            "parameters": {
              "slave_type": slaveType,
              "excluded_hosts": hostNames
            },
            'operation_level': {
              level: "HOST_COMPONENT",
              cluster_name: App.get('clusterName'),
              host_name: hostNames,
              service_name: serviceName
            }
          },
          "Requests/resource_filters": [
            {"service_name": serviceName, "component_name": componentName}
          ]
        }
      }];
    var id = 2;
    var hAray = hostNames.split(",");
    for (var i = 0; i < hAray.length; i++) {
      batches.push({
        "order_id": id,
        "type": "PUT",
        "uri": App.get('apiPrefix') + "/clusters/" + App.get('clusterName') + "/hosts/" + hAray[i] + "/host_components/" + slaveType,
        "RequestBodyInfo": {
          "RequestInfo": {
            context: Em.I18n.t('hosts.host.regionserver.decommission.batch2'),
            exclusive: true,
            operation_level: {
              level: "HOST_COMPONENT",
              cluster_name: App.get('clusterName'),
              host_name: hostNames,
              service_name: serviceName || null
            }
          },
          "Body": {
            HostRoles: {
              state: "INSTALLED"
            }
          }
        }
      });
      id++;
    }
    batches.push({
        "order_id": id,
        "type": "POST",
        "uri": App.get('apiPrefix') + "/clusters/" + App.get('clusterName') + "/requests",
        "RequestBodyInfo": {
          "RequestInfo": {
            "context": Em.I18n.t('hosts.host.regionserver.decommission.batch3'),
            "command": "DECOMMISSION",
            "service_name": serviceName,
            "component_name": componentName,
            "parameters": {
              "slave_type": slaveType,
              "excluded_hosts": hostNames,
              "mark_draining_only": true
            },
            'operation_level': {
              level: "HOST_COMPONENT",
              cluster_name: App.get('clusterName'),
              host_name: hostNames,
              service_name: serviceName
            }
          },
          "Requests/resource_filters": [
            {"service_name": serviceName, "component_name": componentName}
          ]
        }
      });
    App.ajax.send({
      name: 'host.host_component.recommission_and_restart',
      sender: this,
      data: {
        intervalTimeSeconds: 1,
        tolerateSize: 0,
        batches: batches
      },
      success: 'decommissionSuccessCallback',
      error: 'decommissionErrorCallback'
    });
  },

  /**
   * Error callback for decommission requests
   * @param {object} request
   * @param {object} ajaxOptions
   * @param {string} error
   * @method decommissionErrorCallback
   */
  decommissionErrorCallback: function (request, ajaxOptions, error) {
    console.log('ERROR: ' + error);
  },

  /**
   * Success ajax response for Recommission/Decommission slaves
   * @param {object} data
   * @method decommissionSuccessCallback
   * @return {Boolean}
   */
  decommissionSuccessCallback: function (data) {
    if (data && (data.Requests || data.resources[0].RequestSchedule)) {
      this.showBackgroundOperationsPopup();
      return true;
    } else {
      console.log('cannot get request id from ', data);
      return false;
    }
  },

  /**
   * Performs Recommission and Start
   * @param {string} hostNames
   * @param {string} serviceName
   * @param {string} componentName
   * @param {string} slaveType
   * @method doRecommissionAndStart
   */
  doRecommissionAndStart: function (hostNames, serviceName, componentName, slaveType) {
    var contextNameString_1 = 'hosts.host.' + slaveType.toLowerCase() + '.recommission';
    var context_1 = Em.I18n.t(contextNameString_1);
    var contextNameString_2 = 'requestInfo.startHostComponent.' + slaveType.toLowerCase();
    var startContext = Em.I18n.t(contextNameString_2);
    var params = {
      "slave_type": slaveType,
      "included_hosts": hostNames
    };
    if (serviceName == "HBASE") {
      params.mark_draining_only = true;
    }
    var batches = [
      {
        "order_id": 1,
        "type": "POST",
        "uri": App.apiPrefix + "/clusters/" + App.get('clusterName') + "/requests",
        "RequestBodyInfo": {
          "RequestInfo": {
            "context": context_1,
            "command": "DECOMMISSION",
            "exclusive":"true",
            "parameters": params,
            'operation_level': {
              level: "HOST_COMPONENT",
              cluster_name: App.get('clusterName'),
              host_name: hostNames,
              service_name: serviceName
            }
          },
          "Requests/resource_filters": [
            {"service_name": serviceName, "component_name": componentName}
          ]
        }
      }];
    var id = 2;
    var hAray = hostNames.split(",");
    for (var i = 0; i < hAray.length; i++) {
      batches.push(    {
        "order_id": id,
        "type": "PUT",
        "uri": App.get('apiPrefix') + "/clusters/" + App.get('clusterName') + "/hosts/" + hAray[i] + "/host_components/" + slaveType,
        "RequestBodyInfo": {
          "RequestInfo": {
            context: startContext,
            operation_level: {
              level: "HOST_COMPONENT",
              cluster_name: App.get('clusterName'),
              host_name: hostNames,
              service_name: serviceName || null
            }
          },
          "Body": {
            HostRoles: {
              state: "STARTED"
            }
          }
        }
      });
      id++;
    }
    App.ajax.send({
      name: 'host.host_component.recommission_and_restart',
      sender: this,
      data: {
        intervalTimeSeconds: 1,
        tolerateSize: 1,
        batches: batches
      },
      success: 'decommissionSuccessCallback',
      error: 'decommissionErrorCallback'
    });
  },

  /**
   * Handler for host-menu items actions
   * @param {object} option
   * @method doAction
   */
  doAction: function (option) {
    switch (option.context.action) {
      case "deleteHost":
        this.validateAndDeleteHost();
        break;
      case "startAllComponents":
        if (!this.get('content.isNotHeartBeating')) this.doStartAllComponents();
        break;
      case "stopAllComponents":
        if (!this.get('content.isNotHeartBeating')) this.doStopAllComponents();
        break;
      case "restartAllComponents":
        if (!this.get('content.isNotHeartBeating')) this.doRestartAllComponents();
        break;
      case "onOffPassiveModeForHost":
        this.onOffPassiveModeForHost(option.context);
        break;
    }
  },

  /**
   * Turn On/Off Passive Mode for host
   * @param {object} context
   * @method onOffPassiveModeForHost
   */
  onOffPassiveModeForHost: function (context) {
    var state = context.active ? 'ON' : 'OFF';
    var self = this;
    var message = Em.I18n.t('hosts.host.details.for.postfix').format(context.label);
    return App.showConfirmationPopup(function () {
        self.hostPassiveModeRequest(state, message);
      },
      Em.I18n.t('hosts.passiveMode.popup').format(context.active ? 'On' : 'Off', this.get('content.hostName'))
    );
  },

  /**
   * Send request to get passive state for host
   * @param {string} state
   * @param {string} message
   * @method hostPassiveModeRequest
   */
  hostPassiveModeRequest: function (state, message) {
    App.ajax.send({
      name: 'bulk_request.hosts.passive_state',
      sender: this,
      data: {
        hostNames: this.get('content.hostName'),
        passive_state: state,
        requestInfo: message
      },
      success: 'updateHost'
    });
  },

  /**
   * Success callback for receiving host passive state
   * @param {object} data
   * @param {object} opt
   * @param {object} params
   * @method updateHost
   */
  updateHost: function (data, opt, params) {
    this.set('content.passiveState', params.passive_state);
    batchUtils.infoPassiveState(params.passive_state);
  },

  /**
   * Send request to get passive state for hostComponent
   * @param {object} component - hostComponentn object
   * @param {string} state
   * @param {string} message
   * @method hostPassiveModeRequest
   */
  updateComponentPassiveState: function (component, state, message) {
    App.ajax.send({
      name: 'common.host.host_component.passive',
      sender: this,
      data: {
        hostName: this.get('content.hostName'),
        componentName: component.get('componentName'),
        component: component,
        passive_state: state,
        context: message
      },
      success: 'updateHostComponent'
    });
  },

  /**
   * Success callback for receiving hostComponent passive state
   * @param {object} data
   * @param {object} opt
   * @param {object} params
   * @method updateHost
   */
  updateHostComponent: function (data, opt, params) {
    params.component.set('passiveState', params.passive_state);
    batchUtils.infoPassiveState(params.passive_state);
  },

  /**
   * Show confirmation popup for action "start all components"
   * @method doStartAllComponents
   */
  doStartAllComponents: function () {
    var self = this;
    var components = this.get('serviceNonClientActiveComponents');
    var componentsLength = Em.isNone(components) ? 0 : components.get('length');
    if (componentsLength > 0) {
      return App.showConfirmationPopup(function () {
        self.sendComponentCommand(components, Em.I18n.t('hosts.host.maintainance.startAllComponents.context'), App.HostComponentStatus.started);
      });
    }
  },

  /**
   * Show confirmation popup for action "stop all components"
   * @method doStopAllComponents
   */
  doStopAllComponents: function () {
    var self = this;
    var components = this.get('serviceNonClientActiveComponents');
    var componentsLength = Em.isNone(components) ? 0 : components.get('length');
    if (componentsLength > 0) {
      return App.showConfirmationPopup(function () {
        self.sendComponentCommand(components, Em.I18n.t('hosts.host.maintainance.stopAllComponents.context'), App.HostComponentStatus.stopped);
      });
    }
  },

  /**
   * Show confirmation popup for action "restart all components"
   * @method doRestartAllComponents
   */
  doRestartAllComponents: function () {
    var self = this;
    var components = this.get('serviceActiveComponents');
    var componentsLength = Em.isNone(components) ? 0 : components.get('length');
    if (componentsLength > 0) {
      return App.showConfirmationPopup(function () {
        batchUtils.restartHostComponents(components, Em.I18n.t('rollingrestart.context.allOnSelectedHost').format(self.get('content.hostName')), "HOST");
      });
    }
  },

  /**
   * get info about host-components, exactly:
   *  - host-components grouped by status, features
   *  - flag, that indicate whether ZooKeeper Server is installed
   * @return {Object}
   */
  getHostComponentsInfo: function () {
    var componentsOnHost = this.get('content.hostComponents');
    var allComponents = App.HostComponent.find();
    var stoppedStates = [App.HostComponentStatus.stopped,
      App.HostComponentStatus.install_failed,
      App.HostComponentStatus.upgrade_failed,
      App.HostComponentStatus.init,
      App.HostComponentStatus.unknown];
    var container = {
      zkServerInstalled: false,
      lastComponents: [],
      masterComponents: [],
      runningComponents: [],
      nonDeletableComponents: [],
      unknownComponents: []
    };

    if (componentsOnHost && componentsOnHost.get('length') > 0) {
      componentsOnHost.forEach(function (cInstance) {
        if (cInstance.get('componentName') === 'ZOOKEEPER_SERVER') {
          container.zkServerInstalled = true;
        }
        if (allComponents.filterProperty('componentName', cInstance.get('componentName')).get('length') === 1) {
          container.lastComponents.push(cInstance.get('displayName'));
        }
        var workStatus = cInstance.get('workStatus');
        if (cInstance.get('isMaster') && !cInstance.get('isDeletable')) {
          container.masterComponents.push(cInstance.get('displayName'));
        }
        if (stoppedStates.indexOf(workStatus) < 0) {
          container.runningComponents.push(cInstance.get('displayName'));
        }
        if (!cInstance.get('isDeletable')) {
          container.nonDeletableComponents.push(cInstance.get('displayName'));
        }
        if (workStatus === App.HostComponentStatus.unknown) {
          container.unknownComponents.push(cInstance.get('displayName'));
        }
      });
    }
    return container;
  },

  /**
   * Deletion of hosts not supported for this version
   * @method validateAndDeleteHost
   */
  validateAndDeleteHost: function () {
    var container = this.getHostComponentsInfo();

    if (container.masterComponents.length > 0) {
      this.raiseDeleteComponentsError(container.masterComponents, 'masterList');
      return;
    } else if (container.nonDeletableComponents.length > 0) {
      this.raiseDeleteComponentsError(container.nonDeletableComponents, 'nonDeletableList');
      return;
    } else if (container.runningComponents.length > 0) {
      this.raiseDeleteComponentsError(container.runningComponents, 'runningList');
      return;
    }
    if (container.zkServerInstalled) {
      var self = this;
      return App.showConfirmationPopup(function () {
        self.confirmDeleteHost(container.unknownComponents, container.lastComponents);
      }, Em.I18n.t('hosts.host.addComponent.deleteHostWithZooKeeper'));
    } else {
      this.confirmDeleteHost(container.unknownComponents, container.lastComponents);
    }
  },

  /**
   * Show popup with info about reasons why host can't be deleted
   * @param {string[]} components
   * @param {string} type
   * @method raiseDeleteComponentsError
   */
  raiseDeleteComponentsError: function (components, type) {
    App.ModalPopup.show({
      header: Em.I18n.t('hosts.cant.do.popup.title'),
      type: type,
      showBodyEnd: function () {
        return this.get('type') === 'runningList' || this.get('type') === 'masterList';
      }.property(),
      components: components,
      componentsStr: function () {
        return this.get('components').join(", ");
      }.property(),
      componentsBody: function () {
        return Em.I18n.t('hosts.cant.do.popup.' + type + '.body').format(this.get('components').length);
      }.property(),
      componentsBodyEnd: function () {
        if (this.get('showBodyEnd')) {
          return Em.I18n.t('hosts.cant.do.popup.' + type + '.body.end');
        }
        return '';
      }.property(),
      bodyClass: Em.View.extend({
        templateName: require('templates/main/host/details/raiseDeleteComponentErrorPopup')
      }),
      secondary: null
    });
  },

  /**
   * Show confirmation popup to delete host
   * @param {string[]} unknownComponents
   * @param {string[]} lastComponents
   * @method confirmDeleteHost
   */
  confirmDeleteHost: function (unknownComponents, lastComponents) {
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('hosts.delete.popup.title'),
      deletePopupBody: function () {
        return Em.I18n.t('hosts.delete.popup.body').format(self.get('content.publicHostName'));
      }.property(),
      lastComponent: function () {
        if (lastComponents && lastComponents.length) {
          this.set('isChecked', false);
          return true;
        } else {
          this.set('isChecked', true);
          return false;
        }
      }.property(),
      disablePrimary: function () {
        return !this.get('isChecked');
      }.property('isChecked'),
      isChecked: false,
      lastComponentError: Em.View.extend({
        template: Em.Handlebars.compile(Em.I18n.t('hosts.delete.popup.body.msg4').format(lastComponents))
      }),
      unknownComponents: function () {
        if (unknownComponents && unknownComponents.length) {
          return unknownComponents.join(", ");
        }
        return '';
      }.property(),
      bodyClass: Em.View.extend({
        templateName: require('templates/main/host/details/doDeleteHostPopup')
      }),
      onPrimary: function () {
        var popup = this;
        var completeCallback = function () {
          popup.hide();
        };
        self.doDeleteHost(completeCallback);
      }
    });
  },

  /**
   * send DELETE calls to components of host and after delete host itself
   * @param completeCallback
   * @method doDeleteHost
   */
  doDeleteHost: function (completeCallback) {
    this.set('fromDeleteHost', true);
    var allComponents = this.get('content.hostComponents');
    var deleteError = null;
    var dfd = $.Deferred();
    var self = this;

    if (allComponents.get('length') > 0) {
      allComponents.forEach(function (component, index) {
        var length = allComponents.get('length');
        if (!deleteError) {
          this._doDeleteHostComponent(component, function () {
            deleteError = self.get('_deletedHostComponentResult');
            if (index == length - 1) {
              dfd.resolve();
            }
          });
        }
      }, this);
    } else {
      dfd.resolve();
    }
    dfd.done(function () {
      if (!deleteError) {
        App.ajax.send({
          name: 'common.delete.host',
          sender: self,
          data: {
            hostName: self.get('content.hostName')
          },
          callback: completeCallback,
          success: 'deleteHostSuccessCallback',
          error: 'deleteHostErrorCallback'
        });
      }
      else {
        completeCallback();
        deleteError.xhr.responseText = "{\"message\": \"" + deleteError.xhr.statusText + "\"}";
        App.ajax.defaultErrorHandler(deleteError.xhr, deleteError.url, deleteError.method, deleteError.xhr.status);
      }
    });
  },
  deleteHostSuccessCallback: function (data, rq, requestBody) {
    var self = this;
    App.router.get('updateController').updateHost(function () {
      self.loadConfigs('loadHiveConfigs');
      self.loadConfigs();
      App.router.transitionTo('hosts.index');
    });
    if(!!(requestBody && requestBody.hostName))
      App.hostsMapper.deleteRecord(App.Host.find().findProperty('hostName', requestBody.hostName));
    App.router.get('clusterController').getAllHostNames();
  },
  deleteHostErrorCallback: function (xhr, textStatus, errorThrown, opt) {
    console.log('Error deleting host.');
    console.log(textStatus);
    console.log(errorThrown);
    xhr.responseText = "{\"message\": \"" + xhr.statusText + "\"}";
    this.loadConfigs();
    App.ajax.defaultErrorHandler(xhr, opt.url, 'DELETE', xhr.status);
  },

  /**
   * Send command to server to restart all host components with stale configs
   * @method restartAllStaleConfigComponents
   */
  restartAllStaleConfigComponents: function () {
    var self = this;
    return App.showConfirmationPopup(function () {
      var staleComponents = self.get('content.componentsWithStaleConfigs');
      batchUtils.restartHostComponents(staleComponents, Em.I18n.t('rollingrestart.context.allWithStaleConfigsOnSelectedHost').format(self.get('content.hostName')), "HOST");
    });
  },

  /**
   * open Reassign Master Wizard with selected component
   * @param {object} event
   * @method moveComponent
   */
  moveComponent: function (event) {
    return App.showConfirmationPopup(function () {
      var component = event.context;
      var reassignMasterController = App.router.get('reassignMasterController');
      reassignMasterController.saveComponentToReassign(component);
      reassignMasterController.getSecurityStatus();
      reassignMasterController.setCurrentStep('1');
      App.router.transitionTo('reassign');
    });
  },

  /**
   * Restart clients host components to apply config changes
   * @param {object} event
   * @method refreshConfigs
   */
  refreshConfigs: function (event) {
    var self = this;
    var components = event.context.filter(function (component) {
      return component.get('staleConfigs');
    });
    if (components.get('length') > 0) {
      return App.showConfirmationPopup(function () {
        batchUtils.restartHostComponents(components, Em.I18n.t('rollingrestart.context.allClientsOnSelectedHost').format(self.get('content.hostName')), "HOST");
      });
    }
  },

  toggleMaintenanceMode: function (event) {
    var self = this;
    var state = event.context.get('passiveState') === "ON" ? "OFF" : "ON";
    var message = Em.I18n.t('passiveState.turn' + state.toCapital() + 'For').format(event.context.get('displayName'));
    return App.showConfirmationPopup(function () {
      self.updateComponentPassiveState(event.context, state, message);
    });
  },

  downloadClientConfigs: function (event) {
    componentsUtils.downloadClientConfigs.call(this, {
      hostName: event.context.get('hostName'),
      componentName: event.context.get('componentName'),
      displayName: event.context.get('displayName')
    });
  },

  reinstallClients: function(event) {
    var clientsToInstall = event.context.filter(function(component) {
      return ['INIT', 'INSTALL_FAILED'].contains(component.get('workStatus'));
    });
    if (!clientsToInstall.length) return;
    this.sendComponentCommand(clientsToInstall, Em.I18n.t('host.host.details.installClients'), 'INSTALLED');
  },

  /**
   * On click handler for custom command from items menu
   * @param context
   */
  executeCustomCommand: function(event) {
    var controller = this;
    var context = event.context;
    return App.showConfirmationPopup(function() {
      App.ajax.send({
        name : 'service.item.executeCustomCommand',
        sender: controller,
        data : {
          command : context.command,
          context : Em.I18n.t('services.service.actions.run.executeCustomCommand.context').format(context.command),
          hosts : context.hosts,
          serviceName : context.service,
          componentName : context.component
        },
        success : 'executeCustomCommandSuccessCallback',
        error : 'executeCustomCommandErrorCallback'
      });
    });
  },

  executeCustomCommandSuccessCallback  : function(data, ajaxOptions, params) {
    if (data.Requests.id) {
      App.router.get('backgroundOperationsController').showPopup();
    } else {
      console.warn('Error during execution of ' + params.command + ' custom command on' + params.componentName);
    }
  },
  executeCustomCommandErrorCallback : function(data) {
    var error = Em.I18n.t('services.service.actions.run.executeCustomCommand.error');
    if(data && data.responseText){
      try {
        var json = $.parseJSON(data.responseText);
        error += json.message;
      } catch (err) {}
    }
    App.showAlertPopup(Em.I18n.t('services.service.actions.run.executeCustomCommand.error'), error);
    console.warn('Error during executing custom command');
  },

  /**
   * show popup confirmation of version installation
   * @param event
   */
  installVersionConfirmation: function (event) {
    var self = this;

    return App.showConfirmationPopup(function () {
        self.installVersion(event);
      },
      Em.I18n.t('hosts.host.stackVersions.install.confirmation').format(event.context.get('displayName'))
    );
  },

  /**
   * install HostStackVersion on host
   * @param {object} event
   */
  installVersion: function (event) {
    App.ajax.send({
      name: 'host.stack_versions.install',
      sender: this,
      data: {
        hostName: this.get('content.hostName'),
        version: event.context
      },
      success: 'installVersionSuccessCallback'
    });
  },

  /**
   * success callback of <code>installVersion</code>
   * on success set version status to INSTALLING
   * @param {object} data
   * @param {object} opt
   * @param {object} params
   */
  installVersionSuccessCallback: function (data, opt, params) {
    App.HostStackVersion.find(params.version.get('id')).set('status', 'INSTALLING');
    App.db.set('repoVersionInstall', 'id', [data.Requests.id]);
    App.clusterStatus.setClusterStatus({
      wizardControllerName: this.get('name'),
      localdb: App.db.data
    });
  }
});

});

require.register("controllers/main/host/host_alerts_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainHostAlertsController = Em.ArrayController.extend({
  name: 'mainHostAlertsController',

  selectedHost: function () {
    return App.get('router.mainHostDetailsController.content');
  }.property('App.router.mainHostDetailsController.content'),

  /**
   * List of all <code>App.AlertInstance</code> by Host
   * @type {App.AlertInstance[]}
   */
  content: function () {
    return App.AlertInstance.find().toArray().filterProperty('host', this.get('selectedHost'));
  }.property('App.router.mainAlertInstancesController.isLoaded', 'selectedHost'),

  /**
   * Open details page of the selected alertDefinition
   * @param {object} event
   * @method routeToAlertDefinition
   */
  routeToAlertDefinition: function (event) {
    var alertDefinition = App.AlertDefinition.find().findProperty('id', event.context);
    App.router.transitionTo('main.alerts.alertDetails', alertDefinition);
  }


});
});

require.register("controllers/main/service", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainServiceController = Em.ArrayController.extend({

  name: 'mainServiceController',

  /**
   * @type {Ember.Object[]}
   */
  content: function () {
    if (!App.router.get('clusterController.isLoaded')) {
      return [];
    }
    return App.Service.find();
  }.property('App.router.clusterController.isLoaded').volatile(),

  /**
   * Current cluster
   * @type {Ember.Object}
   */
  cluster: function () {
    if (!App.router.get('clusterController.isLoaded')) {
      return null;
    }
    return App.Cluster.find().objectAt(0);
  }.property('App.router.clusterController.isLoaded'),

  /**
   * Check if all services are installed
   * true - all installed, false - not all
   * @type {bool}
   */
  isAllServicesInstalled: function () {
    if (!this.get('content.content')) return false;
    var availableServices = App.StackService.find().mapProperty('serviceName');
    return this.get('content.content').length == availableServices.length;
  }.property('content.content.@each', 'content.content.length'),

  /**
   * Should "Start All"-button be disabled
   * @type {bool}
   */
  isStartAllDisabled: function () {
    if (this.get('isStartStopAllClicked') == true) {
      return true;
    }
    var stoppedServices = this.get('content').filter(function (_service) {
      return (_service.get('healthStatus') === 'red' && !App.get('services.clientOnly').contains(_service.get('serviceName')));
    });
    return (stoppedServices.length === 0); // all green status
  }.property('isStartStopAllClicked', 'content.@each.healthStatus'),

  /**
   * Should "Stop All"-button be disabled
   * @type {bool}
   */
  isStopAllDisabled: function () {
    if (this.get('isStartStopAllClicked') == true) {
      return true;
    }
    var startedServiceLength = this.get('content').filterProperty('healthStatus', 'green').length;
    return (startedServiceLength === 0);
  }.property('isStartStopAllClicked', 'content.@each.healthStatus'),

  /**
   * @type {bool}
   */
  isStartStopAllClicked: function () {
    return (App.router.get('backgroundOperationsController').get('allOperationsCount') !== 0);
  }.property('App.router.backgroundOperationsController.allOperationsCount'),

  /**
   * Callback for <code>start all service</code> button
   * @return {App.ModalPopup|null}
   * @method startAllService
   */
  startAllService: function (event) {
    return this.startStopAllService(event, 'STARTED');
  },

  /**
   * Callback for <code>stop all service</code> button
   * @return {App.ModalPopup|null}
   * @method stopAllService
   */
  stopAllService: function (event) {
    return this.startStopAllService(event, 'INSTALLED');
  },

  /**
   * Common method for "start-all", "stop-all" calls
   * @param {object} event
   * @param {string} state 'STARTED|INSTALLED'
   * @returns {App.ModalPopup|null}
   * @method startStopAllService
   */
  startStopAllService: function(event, state) {
    if ($(event.target).hasClass('disabled') || $(event.target.parentElement).hasClass('disabled')) {
      return null;
    }
    var self = this;
    var bodyMessage = Em.Object.create({
      confirmMsg: state == 'INSTALLED' ? Em.I18n.t('services.service.stopAll.confirmMsg') : Em.I18n.t('services.service.startAll.confirmMsg'),
      confirmButton: state == 'INSTALLED' ? Em.I18n.t('services.service.stop.confirmButton') : Em.I18n.t('services.service.start.confirmButton')
    });

    return App.showConfirmationFeedBackPopup(function (query) {
      self.allServicesCall(state, query);
    }, bodyMessage);
  },

  /**
   * Do request to server for "start|stop" all services
   * @param {string} state "STARTED|INSTALLED"
   * @param {object} query
   * @method allServicesCall
   * @return {$.ajax}
   */
  allServicesCall: function (state, query) {
    var context = (state == 'INSTALLED') ? App.BackgroundOperationsController.CommandContexts.STOP_ALL_SERVICES :
      App.BackgroundOperationsController.CommandContexts.START_ALL_SERVICES;
    return App.ajax.send({
      name: 'common.services.update',
      sender: this,
      data: {
        context: context,
        ServiceInfo: {
          state: state
        },
        query: query
      },
      success: 'allServicesCallSuccessCallback',
      error: 'allServicesCallErrorCallback'
    });
  },

  /**
   * Success-callback for all-services request
   * @param {object} data
   * @param {object} xhr
   * @param {object} params
   * @method allServicesCallSuccessCallback
   */
  allServicesCallSuccessCallback: function (data, xhr, params) {
    params.query.set('status', 'SUCCESS');

    // load data (if we need to show this background operations popup) from persist
    App.router.get('applicationController').dataLoading().done(function (initValue) {
      if (initValue) {
        App.router.get('backgroundOperationsController').showPopup();
      }
    });
  },

  /**
   * Error-callback for all-services request
   * @param {object} request
   * @param {object} ajaxOptions
   * @param {string} error
   * @param {object} opt
   * @param {object} params
   * @method allServicesCallErrorCallback
   */
  allServicesCallErrorCallback: function (request, ajaxOptions, error, opt, params) {
    params.query.set('status', 'FAIL');
  },

  /**
   * "Add-service"-click handler
   * @method gotoAddService
   */
  gotoAddService: function () {
    if (this.get('isAllServicesInstalled')) {
      return;
    }
    App.router.get('addServiceController').setDBProperty('onClosePath', 'main.services.index');
    App.router.transitionTo('main.serviceAdd');
  }
});

});

require.register("controllers/main/service/add_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');
App.AddServiceController = App.WizardController.extend(App.AddSecurityConfigs, {

  name: 'addServiceController',
  // @TODO: remove after Kerberos Automation supports
  totalSteps: App.supports.automatedKerberos ? 8 : 7,

  /**
   * Used for hiding back button in wizard
   */
  hideBackButton: true,

  /**
   * @type {string}
   * @default null
   */
  serviceToInstall: null,

  /**
   *
   */
  installClientQueueLength: 0,

  /**
   * All wizards data will be stored in this variable
   *
   * cluster - cluster name
   * installOptions - ssh key, repo info, etc.
   * services - services list
   * hosts - list of selected hosts
   * slaveComponentHosts, - info about slave hosts
   * masterComponentHosts - info about master hosts
   * config??? - to be described later
   */
  content: Em.Object.create({
    cluster: null,
    hosts: null,
    installOptions: null,
    services: null,
    slaveComponentHosts: null,
    masterComponentHosts: null,
    serviceConfigProperties: null,
    advancedServiceConfig: null,
    controllerName: 'addServiceController',
    configGroups: [],
    clients: [],
    additionalClients: [],
    smokeuser: "ambari-qa",
    group: "hadoop"
  }),

  loadMap: {
    '1': [
      {
        type: 'sync',
        callback: function () {
          this.loadServices();
        }
      }
    ],
    '2': [
      {
        type: 'async',
        callback: function () {
          var dfd = $.Deferred();
          var self = this;
          this.loadHosts().done(function () {
            self.loadMasterComponentHosts();
            self.load('hosts');
            dfd.resolve();
          });
          return dfd.promise();
        }
      }
    ],
    '3': [
      {
        type: 'async',
        callback: function () {
          var dfd = $.Deferred();
          var self = this;
          this.loadHosts().done(function () {
            self.loadServices();
            self.loadClients();
            self.loadSlaveComponentHosts();//depends on loadServices
            dfd.resolve();
          });
          return dfd.promise();
        }
      }
    ],
    '4': [
      {
        type: 'async',
        callback: function () {
          var self = this;
          var dfd = $.Deferred();
          this.loadKerberosDescriptorConfigs().done(function() {
            self.loadServiceConfigGroups();
            self.loadServiceConfigProperties();
            dfd.resolve();
          });
          return dfd.promise();
        }
      }
    ],
    '5': [
      {
        type: 'sync',
        callback: function () {
          this.checkSecurityStatus();
          this.load('cluster');
          this.set('content.additionalClients', []);
          this.set('installClientQueueLength', 0);
          this.set('installClietsQueue', App.ajaxQueue.create({abortOnError: false}));
        }
      }
    ]
  },

  setCurrentStep: function (currentStep, completed) {
    this._super(currentStep, completed);
    App.clusterStatus.setClusterStatus({
      wizardControllerName: this.get('name'),
      localdb: App.db.data
    });
  },

  loadHosts: function () {
    var dfd;
    if (this.getDBProperty('hosts')) {
      dfd = $.Deferred();
      dfd.resolve();
    } else {
      dfd = App.ajax.send({
        name: 'hosts.confirmed',
        sender: this,
        data: {},
        success: 'loadHostsSuccessCallback',
        error: 'loadHostsErrorCallback'
      });
    }
    return dfd.promise();
  },

  loadHostsSuccessCallback: function (response) {
    var installedHosts = {};

    response.items.forEach(function (item, indx) {
      installedHosts[item.Hosts.host_name] = {
        name: item.Hosts.host_name,
        cpu: item.Hosts.cpu_count,
        memory: item.Hosts.total_mem,
        disk_info: item.Hosts.disk_info,
        osType: item.Hosts.os_type,
        osArch: item.Hosts.os_arch,
        ip: item.Hosts.ip,
        bootStatus: "REGISTERED",
        isInstalled: true,
        hostComponents: item.host_components,
        id: indx++
      };
    });
    this.setDBProperty('hosts', installedHosts);
    this.set('content.hosts', installedHosts);
  },

  loadHostsErrorCallback: function (jqXHR, ajaxOptions, error, opt) {
    App.ajax.defaultErrorHandler(jqXHR, opt.url, opt.method, jqXHR.status);
    console.log('Loading hosts failed');
  },

  /**
   * Load services data. Will be used at <code>Select services(step4)</code> step
   */
  loadServices: function () {
    var services = this.getDBProperty('services');
    if (!services) {
      services = {
        selectedServices: [],
        installedServices: []
      };
      App.StackService.find().forEach(function (item) {
        var isInstalled = App.Service.find().someProperty('id', item.get('serviceName'));
        var isSelected = (item.get('serviceName') == this.get('serviceToInstall')) || item.get('coSelectedServices').contains(this.get('serviceToInstall'));
        item.set('isSelected', isInstalled || isSelected);
        item.set('isInstalled', isInstalled);
        if (isInstalled) {
          services.selectedServices.push(item.get('serviceName'));
          services.installedServices.push(item.get('serviceName'));
        } else if(isSelected) {
          services.selectedServices.push(item.get('serviceName'));
        }
      }, this);
      this.setDBProperty('services', services);
    } else {
      App.StackService.find().forEach(function (item) {
        var isSelected = services.selectedServices.contains(item.get('serviceName')) || item.get('serviceName') == this.get('serviceToInstall');
        var isInstalled = services.installedServices.contains(item.get('serviceName'));
        item.set('isSelected', isSelected || (this.get("currentStep") == "1" ? isInstalled : isSelected));
        item.set('isInstalled', isInstalled);
      }, this);
      var isServiceWithSlave = App.StackService.find().filterProperty('isSelected').filterProperty('hasSlave').filterProperty('isInstalled', false).length;
      var isServiceWithClient = App.StackService.find().filterProperty('isSelected').filterProperty('hasClient').filterProperty('isInstalled', false).length;
      var isServiceWithCustomAssignedNonMasters = App.StackService.find().filterProperty('isSelected').filterProperty('hasNonMastersWithCustomAssignment').filterProperty('isInstalled', false).length;
      this.set('content.skipSlavesStep', !isServiceWithSlave && !isServiceWithClient || !isServiceWithCustomAssignedNonMasters);
      if (this.get('content.skipSlavesStep')) {
        this.get('isStepDisabled').findProperty('step', 3).set('value', this.get('content.skipSlavesStep'));
      }
    }
    this.set('serviceToInstall', null);
    this.set('content.services', App.StackService.find());
  },

  /**
   * Save data to model
   * @param stepController App.WizardStep4Controller
   */
  saveServices: function (stepController) {
    var services = {
      selectedServices: [],
      installedServices: []
    };
    var selectedServices = stepController.get('content').filterProperty('isSelected', true).filterProperty('isInstalled', false).mapProperty('serviceName');
    services.selectedServices.pushObjects(selectedServices);
    services.installedServices.pushObjects(stepController.get('content').filterProperty('isInstalled', true).mapProperty('serviceName'));
    // save services that already installed but ignored on choose services page
    // these services marked by `isInstallable` flag with value `false`, for example `Kerberos` service
    services.installedServices.pushObjects(App.Service.find().mapProperty('serviceName').filter(function(serviceName) {
      return !services.installedServices.contains(serviceName);
    }));
    this.setDBProperty('services', services);
    console.log('AddServiceController.saveServices: saved data', stepController.get('content'));

    this.set('content.selectedServiceNames', selectedServices);
    this.setDBProperty('selectedServiceNames', selectedServices);
    var isServiceWithSlave = stepController.get('content').filterProperty('isSelected').filterProperty('hasSlave').filterProperty('isInstalled', false).mapProperty('serviceName').length;
    var isServiceWithClient = App.StackService.find().filterProperty('isSelected').filterProperty('hasClient').filterProperty('isInstalled', false).mapProperty('serviceName').length;
    this.set('content.skipSlavesStep', !isServiceWithSlave && !isServiceWithClient);
    if (this.get('content.skipSlavesStep')) {
      this.get('isStepDisabled').findProperty('step', 3).set('value', this.get('content.skipSlavesStep'));
    }
  },

  /**
   * Save Master Component Hosts data to Main Controller
   * @param stepController App.WizardStep5Controller
   */
  saveMasterComponentHosts: function (stepController) {
    var obj = stepController.get('selectedServicesMasters');
    var masterComponentHosts = [];
    var installedComponents = App.HostComponent.find();

    obj.forEach(function (_component) {
      var installedComponent = installedComponents.findProperty('componentName', _component.component_name);
      masterComponentHosts.push({
        display_name: _component.display_name,
        component: _component.component_name,
        hostName: _component.selectedHost,
        serviceId: _component.serviceId,
        isInstalled: !!installedComponent,
        workStatus: installedComponent && installedComponent.get('workStatus')
      });
    });

    console.log("AddServiceController.saveMasterComponentHosts: saved hosts ", masterComponentHosts);
    this.setDBProperty('masterComponentHosts', masterComponentHosts);
    this.set('content.masterComponentHosts', masterComponentHosts);

    this.set('content.skipMasterStep', this.get('content.masterComponentHosts').everyProperty('isInstalled', true));
    this.get('isStepDisabled').findProperty('step', 2).set('value', this.get('content.skipMasterStep'));
  },

  /**
   * Load master component hosts data for using in required step controllers
   */
  loadMasterComponentHosts: function () {
    this._super();
    this.set('content.skipMasterStep', App.StackService.find().filterProperty('isSelected').filterProperty('hasMaster').everyProperty('isInstalled', true));
    this.get('isStepDisabled').findProperty('step', 2).set('value', this.get('content.skipMasterStep'));
  },

  /**
   * Does service have any configs
   * @param {string} serviceName
   * @returns {boolean}
   */
  isServiceNotConfigurable: function (serviceName) {
    return App.get('services.noConfigTypes').contains(serviceName);
  },

  /**
   * Should Config Step be skipped (based on selected services list)
   * @returns {boolean}
   */
  skipConfigStep: function () {
    var skipConfigStep = true;
    var selectedServices = this.get('content.services').filterProperty('isSelected', true).filterProperty('isInstalled', false).mapProperty('serviceName');
    selectedServices.map(function (serviceName) {
      skipConfigStep = skipConfigStep && this.isServiceNotConfigurable(serviceName);
    }, this);
    return skipConfigStep;
  },

  loadServiceConfigProperties: function () {
    this._super();
    if (!this.get('content.services')) {
      this.loadServices();
    }
    if (this.get('currentStep') > 1 && this.get('currentStep') < 6) {
      this.set('content.skipConfigStep', this.skipConfigStep());
      this.get('isStepDisabled').findProperty('step', 4).set('value', this.get('content.skipConfigStep'));
    }
  },

  /**
   * Load kerberos descriptor configuration
   * @returns {$.Deferred}
   */
  loadKerberosDescriptorConfigs: function() {
    var self = this,
        dfd = $.Deferred();
    if (App.router.get('mainAdminKerberosController.securityEnabled')) {
      this.getDescriptorConfigs().then(function(properties) {
        self.set('kerberosDescriptorConfigs', properties);
        dfd.resolve();
      });
    } else {
      dfd.resolve();
    }
    return dfd.promise();
  },

  saveServiceConfigProperties: function (stepController) {
    this._super(stepController);
    if (this.get('currentStep') > 1 && this.get('currentStep') < 6) {
      this.set('content.skipConfigStep', this.skipConfigStep());
      this.get('isStepDisabled').findProperty('step', 4).set('value', this.get('content.skipConfigStep'));
    }
  },

  /**
   * Load master component hosts data for using in required step controllers
   */
  loadSlaveComponentHosts: function () {
    var slaveComponentHosts = this.getDBProperty('slaveComponentHosts'),
      hosts = this.getDBProperty('hosts'),
      host_names = Em.keys(hosts);
    if (!Em.isNone(slaveComponentHosts)) {
      slaveComponentHosts.forEach(function (component) {
        component.hosts.forEach(function (host) {
          //Em.set(host, 'hostName', hosts[host.host_id].name);
          for (var i = 0; i < host_names.length; i++) {
            if (hosts[host_names[i]].id === host.host_id) {
              host.hostName = host_names[i];
              break;
            }
          }
        });
      });
    }
    if (!slaveComponentHosts) {
      slaveComponentHosts = this.getSlaveComponentHosts();
    }
    this.set("content.slaveComponentHosts", slaveComponentHosts);
    console.log("AddServiceController.loadSlaveComponentHosts: loaded hosts ", slaveComponentHosts);
  },

  /**
   * return slaveComponents bound to hosts
   * @return {Array}
   */
  getSlaveComponentHosts: function () {
    var components = this.get('slaveComponents');
    var result = [];
    var installedServices = App.Service.find().mapProperty('serviceName');
    var selectedServices = this.get('content.services').filterProperty('isSelected', true).mapProperty('serviceName');
    var installedComponentsMap = {};
    var uninstalledComponents = [];
    var hosts = this.getDBProperty('hosts') || this.get('content.hosts');
    var masterComponents = App.get('components.masters');
    var nonMasterComponentHosts = [];

    components.forEach(function (component) {
      if (installedServices.contains(component.get('serviceName'))) {
        installedComponentsMap[component.get('componentName')] = [];
      } else if (selectedServices.contains(component.get('serviceName'))) {
        uninstalledComponents.push(component);
      }
    }, this);

    for (var hostName in hosts) {
      if (hosts[hostName].isInstalled) {
        var isMasterComponentHosted = false;
        hosts[hostName].hostComponents.forEach(function (component) {
          if (installedComponentsMap[component.HostRoles.component_name]) {
            installedComponentsMap[component.HostRoles.component_name].push(hostName);
          }
          if (masterComponents.contains(component.HostRoles.component_name)) {
            isMasterComponentHosted = true;
          }
        }, this);
        if (!isMasterComponentHosted) {
          nonMasterComponentHosts.push(hostName);
        }
      }
    }

    for (var componentName in installedComponentsMap) {
      var component = {
        componentName: componentName,
        displayName: App.format.role(componentName),
        hosts: [],
        isInstalled: true
      };
      installedComponentsMap[componentName].forEach(function (hostName) {
        component.hosts.push({
          group: "Default",
          hostName: hostName,
          isInstalled: true
        });
      }, this);
      result.push(component);
    }

    if (!nonMasterComponentHosts.length) {
      nonMasterComponentHosts.push(Object.keys(hosts)[0]);
    }
    var uninstalledComponentHosts =  nonMasterComponentHosts.map(function(_hostName){
      return {
        group: "Default",
        hostName: _hostName,
        isInstalled: false
      }
    });
    uninstalledComponents.forEach(function (component) {
      result.push({
        componentName: component.get('componentName'),
        displayName: App.format.role(component.get('componentName')),
        hosts: uninstalledComponentHosts,
        isInstalled: false
      })
    });

    return result;
  },

  /**
   * Generate clients list for selected services and save it to model
   * @param stepController step4WizardController
   */
  saveClients: function (stepController) {
    var clients = [];
    var serviceComponents = App.StackServiceComponent.find();
    this.get('content.services').filterProperty('isSelected').filterProperty('isInstalled',false).forEach(function (_service) {
      var serviceClients = serviceComponents.filterProperty('serviceName', _service.get('serviceName')).filterProperty('isClient');
      serviceClients.forEach(function (client) {
        clients.push({
          component_name: client.get('componentName'),
          display_name: client.get('displayName'),
          isInstalled: false
        });
      }, this);
    }, this);

    this.setDBProperty('clientInfo', clients);
    this.set('content.clients', clients);
    console.log("AddServiceController.saveClients: saved list ", clients);
  },

  /**
   * Remove all loaded data.
   * Created as copy for App.router.clearAllSteps
   */
  clearAllSteps: function () {
    this.clearInstallOptions();
    // clear temporary information stored during the install
    this.set('content.cluster', this.getCluster());
  },

  /**
   * Clear all temporary data
   */
  finish: function () {
    this.clearAllSteps();
    this.clearStorageData();
    this.resetDbNamespace();
    App.router.get('updateController').updateAll();
  },

  /**
   * genarates data for ajax request to launch install services
   * @method generateDataForInstallServices
   * @param {Array} selectedServices
   * @returns {{context: *, ServiceInfo: {state: string}, urlParams: string}}
   */
  generateDataForInstallServices: function(selectedServices) {
    if (selectedServices.contains('OOZIE')) {
      selectedServices = selectedServices.concat(['HDFS', 'YARN', 'MAPREDUCE2']);
    }
    return {
      "context": Em.I18n.t('requestInfo.installServices'),
      "ServiceInfo": {"state": "INSTALLED"},
      "urlParams": "ServiceInfo/service_name.in(" + selectedServices.join(',')  + ")"
    };
  },

  /**
   * main method for installing additional clients and services
   * @param {function} callback
   * @method installServices
   */
  installServices: function (callback) {
    var self = this;
    this.set('content.cluster.oldRequestsId', []);
    this.installAdditionalClients().done(function () {
      self.installSelectedServices(callback);
    });
  },

  /**
   * method to install added services
   * @param {function} callback
   * @method installSelectedServices
   */
  installSelectedServices: function (callback) {
    var name = 'common.services.update';
    var selectedServices = this.get('content.services').filterProperty('isInstalled', false).filterProperty('isSelected', true).mapProperty('serviceName');
    var data = this.generateDataForInstallServices(selectedServices);
    this.installServicesRequest(name, data, callback.bind(this));
  },

  installServicesRequest: function (name, data, callback) {
    callback = callback || Em.K;
    App.ajax.send({
      name: name,
      sender: this,
      data: data,
      success: 'installServicesSuccessCallback',
      error: 'installServicesErrorCallback'
    }).then(callback, callback);
  },

  /**
   * installs clients before install new services
   * on host where some components require this
   * @method installAdditionalClients
   */
  installAdditionalClients: function () {
    var dfd = $.Deferred();
    if (this.get('content.additionalClients.length') > 0) {
      this.get('content.additionalClients').forEach(function (c, k) {
        if (c.hostNames.length > 0) {
          var queryStr = 'HostRoles/component_name='+ c.componentName + '&HostRoles/host_name.in(' + c.hostNames.join() + ')';
          this.get('installClietsQueue').addRequest({
            name: 'common.host_component.update',
            sender: this,
            data: {
              query: queryStr,
              context: 'Install ' + App.format.role(c.componentName),
              HostRoles: {
                state: 'INSTALLED'
              },
              counter: k,
              deferred: dfd
            },
            success: 'installClientSuccess',
            error: 'installClientError'
          });
        }
      }, this);
      if (this.get('installClietsQueue.queue.length') == 0) {
        return dfd.resolve();
      } else {
        this.set('installClientQueueLength', this.get('installClietsQueue.queue.length'));
        App.get('router.wizardStep8Controller').set('servicesInstalled', true);
        this.get('installClietsQueue').start();
      }
    } else {
      dfd.resolve();
    }
    return dfd.promise();
  },

  /**
   * callback for when install clients success
   * @param data
   * @param opt
   * @param params
   * @method installClientComplete
   */
  installClientSuccess: function(data, opt, params) {
    if (this.get('installClientQueueLength') - 1 == params.counter) {
      params.deferred.resolve();
    }
  },

  /**
   * callback for when install clients fail
   * @param request
   * @param ajaxOptions
   * @param error
   * @param opt
   * @param params
   */
  installClientError: function(request, ajaxOptions, error, opt, params) {
    if (this.get('installClientQueueLength') - 1 == params.counter) {
      params.deferred.resolve();
    }
  },

  checkSecurityStatus: function() {
    if (App.supports.automatedKerberos) {
      if (!App.router.get('mainAdminKerberosController.securityEnabled')) {
        this.set('skipConfigureIdentitiesStep', true);
        this.get('isStepDisabled').findProperty('step', 5).set('value', true);
      }
    }
  }

});

});

require.register("controllers/main/service/info/audit", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainServiceInfoAuditController = Em.Controller.extend({
  name: 'mainServiceInfoAuditController'
});
});

require.register("controllers/main/service/info/configs", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
require('controllers/wizard/slave_component_groups_controller');
var batchUtils = require('utils/batch_scheduled_requests');
var lazyLoading = require('utils/lazy_loading');

App.MainServiceInfoConfigsController = Em.Controller.extend(App.ServerValidatorMixin, {
  name: 'mainServiceInfoConfigsController',
  isHostsConfigsPage: false,
  forceTransition: false,
  isRecommendedLoaded: true,
  dataIsLoaded: false,
  stepConfigs: [], //contains all field properties that are viewed in this service
  selectedService: null,
  serviceConfigTags: null,
  selectedConfigGroup: null,
  configTypesInfo: {
    items: [],
    supportsFinal: []
  },
  requestInProgress: null,
  selectedServiceConfigTypes: [],
  selectedServiceSupportsFinal: [],
  configGroups: [],
  allConfigs: [],
  uiConfigs: [],
  customConfig: [],
  isApplyingChanges: false,
  saveConfigsFlag: true,
  isCompareMode: false,
  compareServiceVersion: null,
  preSelectedConfigVersion: null,
  // contain Service Config Property, when user proceed from Select Config Group dialog
  overrideToAdd: null,
  //version of default config group, configs of which currently applied
  currentDefaultVersion: null,
  //version selected to view
  selectedVersion: null,
  // file names of changed configs
  modifiedFileNames: [],
  // note passed on configs save
  serviceConfigVersionNote: '',
  versionLoaded: false,
  // current cluster-env version
  clusterEnvTagVersion: '',

  isCurrentSelected: function () {
    return App.ServiceConfigVersion.find(this.get('content.serviceName') + "_" + this.get('selectedVersion')).get('isCurrent');
  }.property('selectedVersion', 'content.serviceName', 'dataIsLoaded'),

  /**
   * @type {boolean}
   */
  canEdit: function () {
    return this.get('isCurrentSelected') && !this.get('isCompareMode');
  }.property('isCurrentSelected', 'isCompareMode'),

  serviceConfigs: function () {
    return App.config.get('preDefinedServiceConfigs');
  }.property('App.config.preDefinedServiceConfigs'),
  customConfigs: function () {
    return App.config.get('preDefinedCustomConfigs');
  }.property('App.config.preDefinedCustomConfigs'),
  configMapping: function () {
    return App.config.get('configMapping');
  }.property('App.config.configMapping'),
  configs: function () {
    return  App.config.get('preDefinedSiteProperties');
  }.property('App.config.preDefinedSiteProperties'),

  secureConfigs: require('data/HDP2/secure_mapping'),

  showConfigHistoryFeature: true,
  /**
   * Map, which contains relation between group and site
   * to upload overridden properties
   */
  loadedGroupToOverrideSiteToTagMap: {},
  /**
   * During page load time the cluster level site to tag
   * mapping is stored here.
   *
   * Example:
   * {
   *  'hdfs-site': 'version1',
   *  'core-site': 'version1'
   * }
   */
  loadedClusterSiteToTagMap: {},
  /**
   * Holds the actual base service-config server data uploaded.
   * This is used by the host-override mechanism to update host
   * specific values.
   */
  savedSiteNameToServerServiceConfigDataMap: {},

  isSubmitDisabled: function () {
    return (!(this.get('stepConfigs').everyProperty('errorCount', 0)) || this.get('isApplyingChanges'));
  }.property('stepConfigs.@each.errorCount', 'isApplyingChanges'),

  isPropertiesChanged: function(){
    return this.get('stepConfigs').someProperty('isPropertiesChanged', true);
  }.property('stepConfigs.@each.isPropertiesChanged'),

  slaveComponentGroups: null,

  /**
   * Filter text will be located here
   */
  filter: '',

  /**
   * List of filters for config properties to populate filter combobox
   */
  propertyFilters: [
    {
      attributeName: 'isOverridden',
      attributeValue: true,
      caption: 'common.combobox.dropdown.overridden'
    },
    {
      attributeName: 'isFinal',
      attributeValue: true,
      caption: 'common.combobox.dropdown.final'
    },
    {
      attributeName: 'hasCompareDiffs',
      attributeValue: true,
      caption: 'common.combobox.dropdown.changed',
      dependentOn: 'isCompareMode'
    },
    {
      attributeName: 'isValid',
      attributeValue: false,
      caption: 'common.combobox.dropdown.issues'
    },
    {
      attributeName: 'warn',
      attributeValue: true,
      caption: 'common.combobox.dropdown.warnings'
    }
  ],

  /**
   * List of heapsize properties not to be parsed
   */
  heapsizeException: ['hadoop_heapsize', 'yarn_heapsize', 'nodemanager_heapsize', 'resourcemanager_heapsize', 'apptimelineserver_heapsize', 'jobhistory_heapsize'],

  /**
   * Regular expression for heapsize properties detection
   */
  heapsizeRegExp: /_heapsize|_newsize|_maxnewsize|_permsize|_maxpermsize$/,

  /**
   * Dropdown menu items in filter combobox
   */
  filterColumns: function () {
    var filterColumns = [];

    this.get('propertyFilters').forEach(function(filter) {
      if (Em.isNone(filter.dependentOn) || this.get(filter.dependentOn)) {
        filterColumns.push(Ember.Object.create({
          attributeName: filter.attributeName,
          attributeValue: filter.attributeValue,
          name: this.t(filter.caption),
          selected: filter.dependentOn ? this.get(filter.dependentOn) : false
        }));
      }
    }, this);
    return filterColumns;
  }.property('propertyFilters', 'isCompareMode'),

  /**
   * indicate wtether service config version belongs to default config group
   * @method isVersionDefault
   * @param version
   * @return {Boolean}
   */
  isVersionDefault: function(version) {
    return (App.ServiceConfigVersion.find(this.get('content.serviceName') + "_" + version).get('groupId') == -1);
  },

  /**
   * register request to view to track his progress
   * @param request
   */
  trackRequest: function (request) {
    this.set('requestInProgress', request);
  },

  /**
   * clear and set properties to default value
   */
  clearStep: function () {
    if (this.get('requestInProgress') && this.get('requestInProgress').readyState !== 4) {
      this.get('requestInProgress').abort();
      this.set('requestInProgress', null);
    }
    this.set("isApplyingChanges", false);
    this.set('modifiedFileNames', []);
    this.set('isInit', true);
    this.set('hash', null);
    this.set('forceTransition', false);
    this.set('dataIsLoaded', false);
    this.set('versionLoaded', false);
    this.set('filter', '');
    this.get('filterColumns').setEach('selected', false);
    this.get('stepConfigs').clear();
    this.get('allConfigs').clear();
    this.get('uiConfigs').clear();
    this.get('customConfig').clear();
    this.set('loadedGroupToOverrideSiteToTagMap', {});
    this.set('serviceConfigVersionNote', '');
    this.set('savedSiteNameToServerServiceConfigDataMap', {});
    if (this.get('serviceConfigTags')) {
      this.set('serviceConfigTags', null);
    }
  },

  serviceConfigProperties: function () {
    return App.db.getServiceConfigProperties();
  }.property('content'),

  /**
   * "Finger-print" of the <code>stepConfigs</code>. Filled after first configGroup selecting
   * Used to determine if some changes were made (when user navigates away from this page)
   * {String}
   */
  hash: null,
  /**
   * Is this initial config group changing
   * {Boolean}
   */
  isInit: true,

  restartHosts: Em.A(),

  /**
   * On load function
   */
  loadStep: function () {
    console.log("TRACE: Loading configure for service");
    this.clearStep();
    this.loadClusterEnvSite();
  },

  /**
   * load all tag versions of cluster-env site
   * @returns {$.ajax}
   */
  loadClusterEnvSite: function () {
    var self = this;
    return App.ajax.send({
      name: 'config.cluster_env_site',
      sender: self,
      success: 'loadClusterEnvSiteSuccess'
    });
  },

  loadClusterEnvSiteSuccess: function (data) {
    // find the latest tag version
    var maxVersion = Math.max.apply(this, data.items.mapProperty('version'));
    this.set('clusterEnvTagVersion', data.items.findProperty('version', maxVersion).tag);
    this.loadServiceConfigs();
  },

  getHash: function () {
    if (!this.get('stepConfigs')[0]) {
      return null;
    }
    var hash = {};
    this.get('stepConfigs')[0].configs.forEach(function (config) {
      hash[config.get('name')] = {value: config.get('value'), overrides: [], isFinal: config.get('isFinal')};
      if (!config.get('overrides')) return;
      if (!config.get('overrides.length')) return;

      config.get('overrides').forEach(function (override) {
        hash[config.get('name')].overrides.push(override.get('value'));
      });
    });
    return JSON.stringify(hash);
  },

  /**
   * Loads service configurations
   */
  loadServiceConfigs: function () {
    var advancedConfigs = [];
    var self = this;
    var serviceName = this.get('content.serviceName');

    var stackService = App.StackService.find().findProperty('serviceName', serviceName);
    if (stackService) {
      self.set('configTypesInfo', App.config.getConfigTypesInfoFromService(stackService));
    }

    this.trackRequest(App.config.loadAdvancedConfig(serviceName, function (properties, xhr) {
      if (xhr.statusText === 'abort') return;
      advancedConfigs.pushObjects(properties);
      self.set('advancedConfigs', advancedConfigs);
      self.trackRequest(App.config.loadClusterConfig(function(clusterProperties) {
        self.get('advancedConfigs').pushObjects(clusterProperties);
        self.trackRequest(self.loadServiceConfigVersions());
      }));
    }));
  },

  /**
   * get service config versions of current service
   */
  loadServiceConfigVersions: function () {
    return App.ajax.send({
      name: 'service.serviceConfigVersions.get',
      data: {
        serviceName: this.get('content.serviceName')
      },
      sender: this,
      success: 'loadServiceConfigVersionsSuccess',
      error: 'loadServiceConfigVersionsError'
    })
  },

  /**
   * load service config versions to model
   * set currentDefaultVersion
   * @param data
   * @param opt
   * @param params
   */
  loadServiceConfigVersionsSuccess: function (data, opt, params) {
    App.serviceConfigVersionsMapper.map(data);
    this.set('currentDefaultVersion', data.items.filterProperty('group_id', -1).findProperty('is_current').service_config_version);
    if (this.get('preSelectedConfigVersion')) {
      this.loadSelectedVersion(this.get('preSelectedConfigVersion.version'));
    } else {
      this.loadSelectedVersion();
    }
  },

  /**
   * error callback of loadServiceConfigVersions()
   * override defaultCallback
   * @param request
   * @param ajaxOptions
   * @param error
   * @param opt
   * @param params
   */
  loadServiceConfigVersionsError: Em.K,

  /**
   * get selected service config version
   * In case selected version is undefined then take currentDefaultVersion
   * @param version
   * @param switchToGroup
   */
  loadSelectedVersion: function (version, switchToGroup) {
    var self = this;
    this.set('versionLoaded', false);
    version = version || this.get('currentDefaultVersion');
    //version of non-default group require properties from current version of default group to correctly display page
    var versions = (this.isVersionDefault(version)) ? [version] : [this.get('currentDefaultVersion'), version];
    switchToGroup = (this.isVersionDefault(version) && !switchToGroup) ? this.get('configGroups').findProperty('isDefault') : switchToGroup;

    if (self.get('dataIsLoaded') && switchToGroup) {
      this.set('selectedConfigGroup', switchToGroup);
    }

    this.trackRequest(App.ajax.send({
      name: 'service.serviceConfigVersions.get.multiple',
      sender: this,
      data: {
        serviceName: this.get('content.serviceName'),
        serviceConfigVersions: versions
      },
      success: 'loadSelectedVersionSuccess'
    }).complete(function (xhr) {
        if (xhr.statusText === 'abort') return;
        if (self.get('dataIsLoaded')) {
          self.onConfigGroupChange();
        } else {
          self.loadServiceTagsAndGroups();
        }
      }));
  },

  /**
   * set cluster to site tag map
   * @param data
   * @param opt
   * @param params
   */
  loadSelectedVersionSuccess: function (data, opt, params) {
    var serviceConfigsDef = this.get('serviceConfigs').findProperty('serviceName', this.get('content.serviceName'));
    var siteToTagMap = {};
    var configTypesRendered = Object.keys(serviceConfigsDef.get('configTypesRendered'));
    var selectedVersion = params.serviceConfigVersions.length > 1 ? params.serviceConfigVersions[1] : params.serviceConfigVersions[0];
    var configurations = [];


    configTypesRendered.forEach(function (siteName) {
      data.items.forEach(function (item) {
        if (item.group_id == -1) {
          configurations = item.configurations;
          if (item.configurations.someProperty('type', siteName)) {
            siteToTagMap[siteName] = item.configurations.findProperty('type', siteName).tag;
          } else {
            siteToTagMap[siteName] = 'version1';
          }
        } else {
          //set config tags of non-default config group to load overrides from selected version
          this.loadedGroupToOverrideSiteToTagMap[item.group_name] = {};
          item.configurations.forEach(function (config) {
            this.loadedGroupToOverrideSiteToTagMap[item.group_name][config.type] = config.tag;
          }, this)
        }
      }, this)
    }, this);

    App.router.get('configurationController').saveToDB(configurations);

    // add cluster-env tag
    siteToTagMap['cluster-env'] = this.get('clusterEnvTagVersion');

    this.loadedClusterSiteToTagMap = siteToTagMap;
    this.set('selectedVersion', selectedVersion);
    //reset map if selected current version of default group
    if (this.get('isCurrentSelected') && selectedVersion === this.get('currentDefaultVersion')) {
      this.loadedGroupToOverrideSiteToTagMap = {};
    }
  },

  /**
   * load config groups of service
   */
  loadServiceTagsAndGroups: function () {
    this.trackRequest(App.ajax.send({
      name: 'config.tags_and_groups',
      sender: this,
      data: {
        serviceName: this.get('content.serviceName'),
        urlParams: "&config_groups/ConfigGroup/tag=" + this.get('content.serviceName')
      },
      success: 'loadServiceConfigsSuccess'
    }));
  },

  loadServiceConfigsSuccess: function (data, opt, params) {
    this.setConfigGroups(data, opt, params);
  },

  setConfigTags: function (data, opt, params) {
    var serviceConfigsDef = this.get('serviceConfigs').findProperty('serviceName', this.get('content.serviceName'));
    var siteToTagMap = {};
    var configTypesRendered = Object.keys(serviceConfigsDef.get('configTypesRendered'));
    configTypesRendered.forEach(function (siteName) {
      if (data.Clusters.desired_configs[siteName]) {
        siteToTagMap[siteName] = data.Clusters.desired_configs[siteName].tag;
      } else {
        siteToTagMap[siteName] = 'version1';
      }
    }, this);
    this.loadedClusterSiteToTagMap = siteToTagMap;
  },

  setConfigGroups: function (data, opt, params) {
    var serviceName = this.get('content.serviceName');
    var displayName = this.get('content.displayName');
    var selectedConfigGroup;
    var defaultHosts = App.get('allHostNames');

    //parse loaded config groups
    var configGroups = [];
    if (data && data.config_groups && data.config_groups.length) {
      data.config_groups.forEach(function (item) {
        item = item.ConfigGroup;
        if (item.tag === this.get('content.serviceName')) {
          var groupHosts = item.hosts.mapProperty('host_name');
          var newConfigGroup = App.ConfigGroup.create({
            id: item.id,
            name: item.group_name,
            description: item.description,
            isDefault: false,
            parentConfigGroup: null,
            service: App.Service.find().findProperty('serviceName', item.tag),
            hosts: groupHosts,
            configSiteTags: []
          });
          for (var i = 0; i < groupHosts.length; i++) {
            defaultHosts = defaultHosts.without(groupHosts[i]);
          }
          item.desired_configs.forEach(function (config) {
            newConfigGroup.configSiteTags.push(App.ConfigSiteTag.create({
              site: config.type,
              tag: config.tag
            }));
          }, this);
          // select default selected group for hosts page
          if (!selectedConfigGroup && this.get('isHostsConfigsPage') && newConfigGroup.get('hosts').contains(this.get('host.hostName')) && this.get('content.serviceName') === item.tag) {
            selectedConfigGroup = newConfigGroup;
          }
          configGroups.push(newConfigGroup);
        }
      }, this);
    }
    this.set('configGroups', configGroups);
    var defaultConfigGroup = App.ConfigGroup.create({
      name: displayName + " Default",
      description: "Default cluster level " + serviceName + " configuration",
      isDefault: true,
      hosts: defaultHosts,
      parentConfigGroup: null,
      service: this.get('content'),
      serviceName: serviceName,
      configSiteTags: []
    });
    if (!selectedConfigGroup) {
      selectedConfigGroup = configGroups.findProperty('name', this.get('preSelectedConfigVersion.groupName')) || defaultConfigGroup;
    }

    this.get('configGroups').sort(function (configGroupA, configGroupB) {
      return (configGroupA.name > configGroupB.name);
    });
    this.get('configGroups').unshift(defaultConfigGroup);
    this.set('selectedConfigGroup', selectedConfigGroup);
    this.set('preSelectedConfigVersion', null);
  },

  onConfigGroupChange: function () {
    var self = this;
    this.get('stepConfigs').clear();
    var selectedConfigGroup = this.get('selectedConfigGroup');
    var serviceName = this.get('content.serviceName');
    //STEP 1: handle tags from JSON data for host overrides
    var configGroupsWithOverrides = selectedConfigGroup.get('isDefault') && !this.get('isHostsConfigsPage') ? this.get('configGroups') : [selectedConfigGroup];
    configGroupsWithOverrides.forEach(function (item) {
      var groupName = item.get('name');
      if (Em.isNone(this.loadedGroupToOverrideSiteToTagMap[groupName])) {
        this.loadedGroupToOverrideSiteToTagMap[groupName] = {};
        item.get('configSiteTags').forEach(function (siteTag) {
          var site = siteTag.get('site');
          this.loadedGroupToOverrideSiteToTagMap[groupName][site] = siteTag.get('tag');
        }, this);
      }
    }, this);
    //STEP 2: Create an array of objects defining tag names to be polled and new tag names to be set after submit
    this.setServiceConfigTags(this.loadedClusterSiteToTagMap);
    //STEP 3: Load advanced configs
    var advancedConfigs = this.get('advancedConfigs');
    //STEP 4: Load on-site config by service from server
    App.router.get('configurationController').getConfigsByTags(this.get('serviceConfigTags')).done(function(configGroups){
      //Merge on-site configs with pre-defined
      var configSet = App.config.mergePreDefinedWithLoaded(configGroups, advancedConfigs, self.get('serviceConfigTags'), serviceName);
      configSet = App.config.syncOrderWithPredefined(configSet);

      var configs = configSet.configs;
      //add custom configs
      App.config.addCustomConfigs(configs);
      //put properties from capacity-scheduler.xml into one config with textarea view
      if (self.get('content.serviceName') === 'YARN') {
        configs = App.config.fileConfigsIntoTextarea(configs, 'capacity-scheduler.xml');
      }
      self.set('allConfigs', configs);
      //add configs as names of host components
      self.addHostNamesToConfig();
      //load configs of version being compared against
      self.loadCompareVersionConfigs(self.get('allConfigs')).done(function (isComparison) {
        //Load and add overriden configs of group
        if (!isComparison && (!self.get('selectedConfigGroup').get('isDefault') || self.get('isCurrentSelected'))) {
          App.config.loadServiceConfigGroupOverrides(self.get('allConfigs'), self.get('loadedGroupToOverrideSiteToTagMap'), self.get('configGroups'), self.onLoadOverrides, self);
        } else {
          self.onLoadOverrides(self.get('allConfigs'));
        }
      });
    });
  }.observes('selectedConfigGroup'),

  /**
   * load version configs for comparison
   * @param allConfigs
   * @return {object}
   */
  loadCompareVersionConfigs: function (allConfigs) {
    var dfd = $.Deferred();
    var self = this;
    var compareServiceVersions = [];

    if (this.get('compareServiceVersion')) {
      if (!this.isVersionDefault(this.get('compareServiceVersion').get('version'))) {
        compareServiceVersions = [this.get('compareServiceVersion').get('version'), this.get('selectedVersion')];
      } else {
        compareServiceVersions = [this.get('compareServiceVersion').get('version')];
      }

      this.getCompareVersionConfigs(compareServiceVersions).done(function (json) {
        self.initCompareConfig(allConfigs, json);
        self.set('compareServiceVersion', null);
        self.set('isCompareMode', true);
        dfd.resolve(true);
      }).fail(function () {
          self.set('compareServiceVersion', null);
          dfd.resolve(true);
        });
    } else {
      self.set('isCompareMode', false);
      allConfigs.setEach('isComparison', false);
      dfd.resolve(false);
    }
    return dfd.promise();
  },

  /**
   * attach analogical config to each property for comparison
   * @param allConfigs
   * @param json
   */
  initCompareConfig: function(allConfigs, json) {
    var serviceVersionMap = {};
    var configNamesMap = {};
    var serviceName = this.get('content.serviceName');
    var compareVersionNumber = this.get('compareServiceVersion').get('version');
    //indicate whether compared versions are from non-default group
    var compareNonDefaultVersions = (json.items.length > 1);

    serviceVersionMap[compareVersionNumber] = {};
    if (compareNonDefaultVersions) {
      serviceVersionMap[this.get('selectedVersion')] = {};
    }
    allConfigs.mapProperty('name').forEach(function(name) {
      configNamesMap[name] = true;
    });

    json.items.forEach(function (item) {
      item.configurations.forEach(function (configuration) {
        if (serviceName == 'YARN' && configuration.type == 'capacity-scheduler') {
          // put all properties in a single textarea for capacity-scheduler
          var value = '';
          for (var prop in configuration.properties) {
            value += prop + '=' + configuration.properties[prop] + '\n';
          }
          serviceVersionMap[item.service_config_version][configuration.type + '-' + configuration.type] = {
            name: configuration.type,
            value: value,
            type: configuration.type,
            tag: configuration.tag,
            version: configuration.version,
            service_config_version: item.service_config_version
          };
        } else {
          for (var prop in configuration.properties) {
            serviceVersionMap[item.service_config_version][prop + '-' + configuration.type] = {
              name: prop,
              value: configuration.properties[prop],
              type: configuration.type,
              tag: configuration.tag,
              version: configuration.version,
              service_config_version: item.service_config_version
            };
            if (Em.isNone(configNamesMap[prop])) {
              allConfigs.push(this.getMockConfig(prop, serviceName, App.config.getOriginalFileName(configuration.type)));
            }
          }
        }
        if (configuration.properties_attributes && configuration.properties_attributes.final) {
          for (var final in configuration.properties_attributes.final) {
            serviceVersionMap[item.service_config_version][final + '-' + configuration.type].isFinal = (configuration.properties_attributes.final[final] === 'true');
          }
        }
      }, this);
    }, this);

    if (compareNonDefaultVersions) {
      allConfigs.forEach(function (serviceConfig) {
        this.setCompareConfigs(serviceConfig, serviceVersionMap, compareVersionNumber, this.get('selectedVersion'));
      }, this);
    } else {
      allConfigs.forEach(function (serviceConfig) {
        var serviceCfgVersionMap = serviceVersionMap[this.get('compareServiceVersion').get('version')];
        var compareConfig = serviceCfgVersionMap[serviceConfig.name + '-' + App.config.getConfigTagFromFileName(serviceConfig.filename)]
        this.setCompareDefaultGroupConfig(serviceConfig, compareConfig);
      }, this);
    }
  },


  /**
   * set compare properties to service config of non-default group
   * @param serviceConfig
   * @param serviceVersionMap
   * @param compareVersion
   * @param selectedVersion
   */
  setCompareConfigs: function (serviceConfig, serviceVersionMap, compareVersion, selectedVersion) {
    var compareConfig = serviceVersionMap[compareVersion][serviceConfig.name + '-' + App.config.getConfigTagFromFileName(serviceConfig.filename)];
    var selectedConfig = serviceVersionMap[selectedVersion][serviceConfig.name + '-' + App.config.getConfigTagFromFileName(serviceConfig.filename)];

    serviceConfig.compareConfigs = [];
    serviceConfig.isComparison = true;

    if (compareConfig && selectedConfig) {
      serviceConfig.compareConfigs.push(this.getComparisonConfig(serviceConfig, compareConfig));
      serviceConfig.compareConfigs.push(this.getComparisonConfig(serviceConfig, selectedConfig));
      serviceConfig.hasCompareDiffs = this.hasCompareDiffs(serviceConfig.compareConfigs[0], serviceConfig.compareConfigs[1]);
    } else if (compareConfig && !selectedConfig) {
      serviceConfig.compareConfigs.push(this.getComparisonConfig(serviceConfig, compareConfig));
      serviceConfig.compareConfigs.push(this.getMockComparisonConfig(selectedConfig, selectedVersion));
      serviceConfig.hasCompareDiffs = true;
    } else if (!compareConfig && selectedConfig) {
      serviceConfig.compareConfigs.push(this.getMockComparisonConfig(selectedConfig, compareVersion));
      serviceConfig.compareConfigs.push(this.getComparisonConfig(serviceConfig, selectedConfig));
      serviceConfig.hasCompareDiffs = true;
    }
  },

  /**
   * init attributes and wrap mock compare config into App.ServiceConfigProperty
   * @param serviceConfig
   * @param compareConfig
   * @return {object}
   */
  getMockComparisonConfig: function (serviceConfig, compareServiceVersion) {
    var compareObject = $.extend(true, {isComparison: true},  serviceConfig);
    compareObject.isEditable = false;

    compareObject.serviceVersion = App.ServiceConfigVersion.find(this.get('content.serviceName') + "_" + compareServiceVersion);
    compareObject.isMock = true;
    compareObject.displayType = 'label';
    compareObject = App.ServiceConfigProperty.create(compareObject);
    compareObject.set('value', Em.I18n.t('common.property.undefined'));
    return compareObject;
  },

  /**
   * init attributes and wrap compare config into App.ServiceConfigProperty
   * @param serviceConfig
   * @param compareConfig
   * @return {object}
   */
  getComparisonConfig: function (serviceConfig, compareConfig) {
    var compareObject = $.extend(true, {isComparison: true},  serviceConfig);
    compareObject.isEditable = false;

    if (compareConfig) {
      if (serviceConfig.isMock) {
        compareObject.displayType = 'string';
        compareObject.isMock = false;
      }
      compareObject.serviceVersion = App.ServiceConfigVersion.find(this.get('content.serviceName') + "_" + compareConfig.service_config_version);
      compareObject = App.ServiceConfigProperty.create(compareObject);
      compareObject.set('isFinal', compareConfig.isFinal);
      compareObject.set('value', App.config.formatOverrideValue(serviceConfig, compareConfig.value));
      this.setSupportsFinal(compareObject);
    }
    return compareObject;
  },

  /**
   * set compare properties to service config of default group
   * @param serviceConfig
   * @param compareConfig
   */
  setCompareDefaultGroupConfig: function (serviceConfig, compareConfig) {
    var compareObject = {};

    serviceConfig.compareConfigs = [];
    serviceConfig.isComparison = true;

    //if config isn't reconfigurable then it can't have changed value to compare
    if (compareConfig && (serviceConfig.isReconfigurable || serviceConfig.isUserProperty)) {
      compareObject = this.getComparisonConfig(serviceConfig, compareConfig);
      serviceConfig.hasCompareDiffs = serviceConfig.isMock || this.hasCompareDiffs(serviceConfig, compareObject);
      serviceConfig.compareConfigs.push(compareObject);
    } else if (serviceConfig.isUserProperty) {
      serviceConfig.compareConfigs.push(this.getMockComparisonConfig(serviceConfig, this.get('compareServiceVersion.version')));
      serviceConfig.hasCompareDiffs = true;
    }
    return serviceConfig;
  },

  /**
   * check value and final attribute of original and compare config for differencies
   * @param originalConfig
   * @param compareConfig
   * @return {Boolean}
   */
  hasCompareDiffs: function (originalConfig, compareConfig) {
    return (originalConfig.value !== compareConfig.value) || (!!originalConfig.isFinal !== (compareConfig.isFinal == true));
  },

  /**
   * generate mock config object
   * @param name
   * @param serviceName
   * @param filename
   * @return {Object}
   */
  getMockConfig: function (name, serviceName, filename) {
    var undefinedConfig = {
      description: name,
      displayName: name,
      id: "site property",
      isOverridable: false,
      isReconfigurable: false,
      isRequired: false,
      isRequiredByAgent: false,
      isSecureConfig: false,
      isUserProperty: true,
      isVisible: true,
      name: name,
      filename: filename,
      serviceName: serviceName,
      value: Em.I18n.t('common.property.undefined'),
      isMock: true,
      displayType: 'label'
    };
    var category = App.config.identifyCategory(undefinedConfig);
    undefinedConfig.category = category && category.name;
    return undefinedConfig;
  },

  /**
   * get configs of chosen version from server to compare
   * @param compareServiceVersions
   * @return {$.ajax}
   */
  getCompareVersionConfigs: function (compareServiceVersions) {
    this.set('versionLoaded', false);

    return App.ajax.send({
      name: 'service.serviceConfigVersions.get.multiple',
      sender: this,
      data: {
        serviceName: this.get('content.serviceName'),
        serviceConfigVersions: compareServiceVersions
      }
    });
  },

  checkDatabaseProperties: function (serviceConfig) {
    this.hideHiveDatabaseProperties(serviceConfig.configs);
    this.hideOozieDatabaseProperties(serviceConfig.configs);
  },

  hideHiveDatabaseProperties: function (configs) {
    if (!['HIVE'].contains(this.get('content.serviceName'))) return;
    var property = configs.findProperty('name', 'hive_hostname');
    if (property) property.set('isVisible', false);

    if (configs.someProperty('name', 'hive_database')) {
      var hiveDb = configs.findProperty('name', 'hive_database');
      if (hiveDb.value === 'Existing MSSQL Server database with integrated authentication') {
        configs.findProperty('name', 'javax.jdo.option.ConnectionUserName').setProperties({
          isVisible: false,
          isRequired: false
        });
        configs.findProperty('name', 'javax.jdo.option.ConnectionPassword').setProperties({
          isVisible: false,
          isRequired: false
        });
      }
    }
  },

  hideOozieDatabaseProperties: function (configs) {
    if (!['OOZIE'].contains(this.get('content.serviceName'))) return;
    var property = configs.findProperty('name', 'oozie_hostname');
    if (property) property.set('isVisible', false);

    if (configs.someProperty('name', 'oozie_database')) {
      var oozieDb = configs.findProperty('name', 'oozie_database');
      if (oozieDb.value === 'Existing MSSQL Server database with integrated authentication') {
        configs.findProperty('name', 'oozie.service.JPAService.jdbc.username').setProperties({
          isVisible: false,
          isRequired: false
        });
        configs.findProperty('name', 'oozie.service.JPAService.jdbc.password').setProperties({
          isVisible: false,
          isRequired: false
        });
      }
    }
  },

  onLoadOverrides: function (allConfigs) {
    var serviceName = this.get('content.serviceName');
    var advancedConfigs = this.get('advancedConfigs');
    //STEP 10: creation of serviceConfig object which contains configs for current service
    var serviceConfig = App.config.createServiceConfig(serviceName);
    //STEP11: Make SecondaryNameNode invisible on enabling namenode HA
    if (serviceConfig.get('serviceName') === 'HDFS') {
      App.config.OnNnHAHideSnn(serviceConfig);
    }

    serviceConfig = App.config.createServiceConfig(this.get('content.serviceName'));
    this.loadConfigs(this.get('allConfigs'), serviceConfig);
    this.setVisibilityForRangerProperties(serviceConfig);
    this.checkOverrideProperty(serviceConfig);
    this.checkDatabaseProperties(serviceConfig);
    this.get('stepConfigs').pushObject(serviceConfig);
    this.set('selectedService', this.get('stepConfigs').objectAt(0));
    this.checkForSecureConfig(this.get('selectedService'));
    this.set('dataIsLoaded', true);
    this.set('versionLoaded', true);
    this.set('hash', this.getHash());
    this.set('isInit', false);
  },

  /**
   * Changes format from Object to Array
   *
   * {
   *  'core-site': 'version1',
   *  'hdfs-site': 'version1',
   *  ...
   * }
   *
   * to
   *
   * [
   *  {
   *    siteName: 'core-site',
   *    tagName: 'version1',
   *    newTageName: null
   *  },
   *  ...
   * ]
   *
   * set tagnames for configuration of the *-site.xml
   */
  setServiceConfigTags: function (desiredConfigsSiteTags) {
    console.debug("setServiceConfigTags(): Trying to set ", desiredConfigsSiteTags);
    var newServiceConfigTags = [];
    for (var index in desiredConfigsSiteTags) {
      newServiceConfigTags.pushObject({
        siteName: index,
        tagName: desiredConfigsSiteTags[index],
        newTagName: null
      }, this);
    }
    console.debug("setServiceConfigTags(): Setting 'serviceConfigTags' to ", newServiceConfigTags);
    this.set('serviceConfigTags', newServiceConfigTags);
  },

  /**
   * check whether the config property is a security related knob
   * @param serviceConfig
   */
  checkForSecureConfig: function (serviceConfig) {
    serviceConfig.get('configs').forEach(function (_config) {
      this.get('secureConfigs').forEach(function (_secureConfig) {
        if (_config.get('name') === _secureConfig.name) {
          _config.set('isSecureConfig', true);
        }
      }, this)
    }, this)
  },

  /**
   * Load child components to service config object
   * @param {Array} configs - array of configs
   * @param {Object} componentConfig - component config object
   * @method loadConfigs
   */
  loadConfigs: function (configs, componentConfig) {
    var serviceConfigsData = App.StackService.find().findProperty('serviceName', this.get('content.serviceName'));
    var defaultGroupSelected = this.get('selectedConfigGroup.isDefault');
    configs.forEach(function (_serviceConfigProperty) {
      var serviceConfigProperty = this.createConfigProperty(_serviceConfigProperty, defaultGroupSelected, serviceConfigsData);
      componentConfig.configs.pushObject(serviceConfigProperty);
      serviceConfigProperty.validate();
    }, this);
    componentConfig.set('initConfigsLength', componentConfig.get('configs.length'));
  },

  /**
   * create {Em.Object} service_cfg_property based on {Object}_serviceConfigProperty and additional info
   * @param {Object} _serviceConfigProperty - config object
   * @param {Boolean} defaultGroupSelected - true if selected cfg group is default
   * @param {Object} serviceConfigsData - service cfg object
   * @returns {Ember.Object|null}
   * @method createConfigProperty
   */
  createConfigProperty: function (_serviceConfigProperty, defaultGroupSelected, serviceConfigsData) {
    if (!_serviceConfigProperty) return null;

    var overrides = _serviceConfigProperty.overrides;
    // we will populate the override properties below
    Em.set(_serviceConfigProperty, 'overrides', null);
    _serviceConfigProperty.isOverridable = Em.isNone(_serviceConfigProperty.isOverridable) ? true : _serviceConfigProperty.isOverridable;

    var serviceConfigProperty = App.ServiceConfigProperty.create(_serviceConfigProperty);

    this.setSupportsFinal(serviceConfigProperty);
    this.setValuesForOverrides(overrides, _serviceConfigProperty, serviceConfigProperty, defaultGroupSelected);
    this.setEditability(serviceConfigProperty, defaultGroupSelected);

    return serviceConfigProperty;
  },

  /**
   * hide properties from Advanced ranger category that match pattern
   * if property with dependentConfigPattern is false otherwise don't hide
   * @param serviceConfig
   * @method setVisibilityForRangerProperties
   */
  setVisibilityForRangerProperties: function(serviceConfig) {
    var category = "Advanced ranger-{0}-plugin-properties".format(this.get('content.serviceName').toLowerCase());
    if (serviceConfig.configCategories.findProperty('name', category)) {
      var patternConfig = serviceConfig.configs.findProperty('dependentConfigPattern');
      if (patternConfig) {
        var value = patternConfig.get('value') === true || ["yes", "true"].contains(patternConfig.get('value').toLowerCase());

        serviceConfig.configs.filter(function(c) {
          if (c.get('category') === category && c.get('name').match(patternConfig.get('dependentConfigPattern')) && c.get('name') != patternConfig.get('name'))
            c.set('isVisible', value);
        });
      }
    }
  },
  /**

  /**
   * trigger addOverrideProperty
   * @param {Object} componentConfig
   * @method checkOverrideProperty
   */
  checkOverrideProperty: function (componentConfig) {
    var overrideToAdd = this.get('overrideToAdd');
    if (overrideToAdd) {
      overrideToAdd = componentConfig.configs.filter(function(c){
        return c.name == overrideToAdd.name && c.filename == overrideToAdd.filename;
      });
      if (overrideToAdd[0]) {
        this.addOverrideProperty(overrideToAdd[0], this.get('selectedConfigGroup'));
        this.set('overrideToAdd', null);
      }
    }
  },

  /**
   * set isEditable property of config for admin
   * if default cfg group and not on the host config page
   * @param {Ember.Object} serviceConfigProperty
   * @param {Boolean} defaultGroupSelected
   * @method setEditability
   */
  setEditability: function (serviceConfigProperty, defaultGroupSelected) {
    serviceConfigProperty.set('isEditable', false);
    if (serviceConfigProperty.get('isComparison')) return;
    if (App.isAccessible('ADMIN') && defaultGroupSelected && !this.get('isHostsConfigsPage') && !serviceConfigProperty.get('group')) {
      serviceConfigProperty.set('isEditable', serviceConfigProperty.get('isReconfigurable'));
    } else if (serviceConfigProperty.get('group') && this.get('selectedConfigGroup.name') === serviceConfigProperty.get('group.name')) {
      serviceConfigProperty.set('isEditable', true);
    }
  },

  /**
   * set supportsFinal property of config for admin
   * @param {Ember.Object} serviceConfigProperty
   * @method setSupportsFinal
   */
  setSupportsFinal: function (serviceConfigProperty) {
    if (serviceConfigProperty.get('isMock')) return;
    var fileName = serviceConfigProperty.get('filename');
    var matchingConfigType = this.get('configTypesInfo').supportsFinal.find(function(configType) {
      return fileName.startsWith(configType);
    });
    serviceConfigProperty.set('supportsFinal', !!matchingConfigType);
  },

  /**
   * set override values
   * @param overrides
   * @param _serviceConfigProperty
   * @param serviceConfigProperty
   * @param defaultGroupSelected
   */
  setValuesForOverrides: function (overrides, _serviceConfigProperty, serviceConfigProperty, defaultGroupSelected) {
    if (Em.isNone(overrides)) return;
    overrides.forEach(function (override) {
      var newSCP = this.createNewSCP(override, _serviceConfigProperty, serviceConfigProperty, defaultGroupSelected);
      var parentOverridesArray = serviceConfigProperty.get('overrides');
      if (parentOverridesArray == null) {
        parentOverridesArray = Em.A([]);
        serviceConfigProperty.set('overrides', parentOverridesArray);
      }
      parentOverridesArray.pushObject(newSCP);
      serviceConfigProperty.set('overrideValues', parentOverridesArray.mapProperty('value'));
      serviceConfigProperty.set('overrideIsFinalValues', parentOverridesArray.mapProperty('isFinal'));
      console.debug("createOverrideProperty(): Added override to main-property:", serviceConfigProperty.get('name'));
    }, this);
  },

  /**
   * create new overridden property and set appropriate fields
   * @param override
   * @param _serviceConfigProperty
   * @param serviceConfigProperty
   * @param defaultGroupSelected
   * @returns {*}
   */
  createNewSCP: function (override, _serviceConfigProperty, serviceConfigProperty, defaultGroupSelected) {
    var newSCP = App.ServiceConfigProperty.create(_serviceConfigProperty);
    newSCP.set('value', override.value);
    newSCP.set('isFinal', override.isFinal);
    newSCP.set('supportsFinal', serviceConfigProperty.get('supportsFinal'));
    newSCP.set('isOriginalSCP', false); // indicated this is overridden value,
    newSCP.set('parentSCP', serviceConfigProperty);
    if (defaultGroupSelected) {
      newSCP.set('group', override.group);
      newSCP.set('isEditable', false);
    }
    return newSCP;
  },

  /**
   * Initialize save configs popup
   */
  restartServicePopup: function () {
    if (this.get("isSubmitDisabled")) {
      return;
    }
    this.set("isApplyingChanges", true);
    var self = this;
    var header, message, messageClass, status;
    var serviceName = this.get('content.serviceName'),
      displayName = this.get('content.displayName'),
      urlParams = '';
    this.serverSideValidation().done(function () {
      if (self.isDirChanged()) {
        App.showConfirmationPopup(function () {
          self.saveConfigs();
        }, Em.I18n.t('services.service.config.confirmDirectoryChange').format(displayName), function () {
          self.set('isApplyingChanges', false);
        });
      } else {
        self.saveConfigs();
      }
    }).fail(function () {
      self.set('isApplyingChanges', false);
    });
  },

  /**
   * Define if user has changed some dir properties
   * @return {Boolean}
   */
  isDirChanged: function () {
    var dirChanged = false;
    var serviceName = this.get('content.serviceName');

    if (serviceName === 'HDFS') {
      var hdfsConfigs = this.get('stepConfigs').findProperty('serviceName', 'HDFS').get('configs');
      if ((hdfsConfigs.findProperty('name', 'dfs.namenode.name.dir') && hdfsConfigs.findProperty('name', 'dfs.namenode.name.dir').get('isNotDefaultValue')) ||
          (hdfsConfigs.findProperty('name', 'dfs.namenode.checkpoint.dir') && hdfsConfigs.findProperty('name', 'dfs.namenode.checkpoint.dir').get('isNotDefaultValue')) ||
          (hdfsConfigs.findProperty('name', 'dfs.datanode.data.dir') && hdfsConfigs.findProperty('name', 'dfs.datanode.data.dir').get('isNotDefaultValue'))) {
          dirChanged = true;
      }
    }
    return dirChanged;
  },

  /**
   * Save changed configs and config groups
   */
  saveConfigs: function () {
    var selectedConfigGroup = this.get('selectedConfigGroup');
    var configs = this.get('stepConfigs').findProperty('serviceName', this.get('content.serviceName')).get('configs');
    var self = this;

    if (selectedConfigGroup.get('isDefault')) {
      if (this.get('content.serviceName') === 'YARN') {
        configs = App.config.textareaIntoFileConfigs(configs, 'capacity-scheduler.xml');
      }
      var modifiedConfigs = configs
        // get only modified and created configs
        .filter(function(config) { return config.get('isNotDefaultValue') || config.get('isNotSaved'); })
        // get file names and add file names that was modified, for example after property removing
        .mapProperty('filename').concat(this.get('modifiedFileNames')).uniq()
        // get configs by filename
        .map(function(fileName) {
          return configs.filterProperty('filename', fileName);
        });
      if (!!modifiedConfigs.length) {
        // concatenate results
        modifiedConfigs = modifiedConfigs.reduce(function(current, prev) { return current.concat(prev); });
      }

      // save modified original configs that have no group
      this.saveSiteConfigs(modifiedConfigs.filter(function(config) { return !config.get('group'); }));

      /**
       * First we put cluster configurations, which automatically creates /configurations
       * resources. Next we update host level overrides.
       */
      this.doPUTClusterConfigurations();
    } else {
      var overridenConfigs = [];
      var groupHosts = [];
      configs.filterProperty('isOverridden', true).forEach(function (config) {
        overridenConfigs = overridenConfigs.concat(config.get('overrides'));
      });
      // find custom original properties that assigned to selected config group
      overridenConfigs = overridenConfigs.concat(configs.filterProperty('group')
        .filter(function(config) { return config.get('group.name') == self.get('selectedConfigGroup.name'); }));
      this.formatConfigValues(overridenConfigs);
      selectedConfigGroup.get('hosts').forEach(function (hostName) {
        groupHosts.push({"host_name": hostName});
      });

      this.putConfigGroupChanges({
        ConfigGroup: {
          "id": selectedConfigGroup.get('id'),
          "cluster_name": App.get('clusterName'),
          "group_name": selectedConfigGroup.get('name'),
          "tag": selectedConfigGroup.get('service.id'),
          "description": selectedConfigGroup.get('description'),
          "hosts": groupHosts,
          "service_config_version_note": this.get('serviceConfigVersionNote'),
          "desired_configs": this.buildGroupDesiredConfigs(overridenConfigs)
        }
      }, true);
    }
  },

  /**
   * On save configs handler. Open save configs popup with appropriate message.
   */
  onDoPUTClusterConfigurations: function () {
    var header, message, messageClass, value, status = 'unknown', urlParams = '',
    result = {
      flag: this.get('saveConfigsFlag'),
      message: null,
      value: null
    },
    extendedModel = App.Service.extendedModel[this.get('content.serviceName')],
    currentService = extendedModel ? App[extendedModel].find(this.get('content.serviceName')) : App.Service.find(this.get('content.serviceName'));

    if (!result.flag) {
      result.message = Em.I18n.t('services.service.config.failSaveConfig');
    }

    App.router.get('clusterController').updateClusterData();
    App.router.get('updateController').updateComponentConfig(function () {
    });
    var flag = result.flag;
    if (result.flag === true) {
      header = Em.I18n.t('services.service.config.saved');
      message = Em.I18n.t('services.service.config.saved.message');
      messageClass = 'alert alert-success';
      // warn the user if any of the components are in UNKNOWN state
      urlParams += ',ServiceComponentInfo/installed_count,ServiceComponentInfo/total_count';
      if (this.get('content.serviceName') === 'HDFS') {
        urlParams += '&ServiceComponentInfo/service_name.in(HDFS)'
      }
    } else {
      header = Em.I18n.t('common.failure');
      message = result.message;
      messageClass = 'alert alert-error';
      value = result.value;
    }
    if(currentService){
      App.QuickViewLinks.proto().set('content', currentService);
      App.QuickViewLinks.proto().loadTags();
    }
    this.showSaveConfigsPopup(header, flag, message, messageClass, value, status, urlParams);
  },

  /**
   * Show save configs popup
   * @method showSaveConfigsPopup
   * @return {App.ModalPopup}
   */
  showSaveConfigsPopup: function (header, flag, message, messageClass, value, status, urlParams) {
    var self = this;
    if (flag) {
      this.set('forceTransition', flag);
      self.loadStep();
    }
    return App.ModalPopup.show({
      header: header,
      primary: Em.I18n.t('ok'),
      secondary: null,
      onPrimary: function () {
        this.hide();
        if (!flag) {
          self.set('isApplyingChanges', false);
        }
      },
      onClose: function () {
        this.hide();
        self.set('isApplyingChanges', false);
      },
      disablePrimary: true,
      bodyClass: Ember.View.extend({
        flag: flag,
        message: function () {
          return this.get('isLoaded') ? message : Em.I18n.t('services.service.config.saving.message');
        }.property('isLoaded'),
        messageClass: function () {
          return this.get('isLoaded') ? messageClass : 'alert alert-info';
        }.property('isLoaded'),
        setDisablePrimary: function () {
          this.get('parentView').set('disablePrimary', !this.get('isLoaded'));
        }.observes('isLoaded'),
        runningHosts: [],
        runningComponentCount: 0,
        unknownHosts: [],
        unknownComponentCount: 0,
        siteProperties: value,
        isLoaded: false,
        componentsFilterSuccessCallback: function (response) {
          var count = 0,
            view = this,
            lazyLoadHosts = function (dest) {
              lazyLoading.run({
                initSize: 20,
                chunkSize: 50,
                delay: 50,
                destination: dest,
                source: hosts,
                context: view
              });
            },
            /**
             * Map components for their hosts
             * Return format:
             * <code>
             *   {
             *    host1: [component1, component2, ...],
             *    host2: [component3, component4, ...]
             *   }
             * </code>
             * @return {object}
             */
            setComponents = function (item, components) {
              item.host_components.forEach(function (c) {
                var name = c.HostRoles.host_name;
                if (!components[name]) {
                  components[name] = [];
                }
                components[name].push(App.format.role(item.ServiceComponentInfo.component_name));
              });
              return components;
            },
            /**
             * Map result of <code>setComponents</code> to array
             * @return {{name: string, components: string}[]}
             */
            setHosts = function (components) {
              var hosts = [];
              Em.keys(components).forEach(function (key) {
                hosts.push({
                  name: key,
                  components: components[key].join(', ')
                });
              });
              return hosts;
            },
            components = {},
            hosts = [];
          switch (status) {
            case 'unknown':
              response.items.filter(function (item) {
                return (item.ServiceComponentInfo.total_count > item.ServiceComponentInfo.started_count + item.ServiceComponentInfo.installed_count);
              }).forEach(function (item) {
                  var total = item.ServiceComponentInfo.total_count,
                    started = item.ServiceComponentInfo.started_count,
                    installed = item.ServiceComponentInfo.installed_count,
                    unknown = total - started + installed;
                  components = setComponents(item, components);
                  count += unknown;
                });
              hosts = setHosts(components);
              this.set('unknownComponentCount', count);
              lazyLoadHosts(this.get('unknownHosts'));
              break;
            case 'started':
              response.items.filterProperty('ServiceComponentInfo.started_count').forEach(function (item) {
                var started = item.ServiceComponentInfo.started_count;
                components = setComponents(item, components);
                count += started;
                hosts = setHosts(components);
              });
              this.set('runningComponentCount', count);
              lazyLoadHosts(this.get('runningHosts'));
              break;
          }
        },
        componentsFilterErrorCallback: function () {
          this.set('isLoaded', true);
        },
        didInsertElement: function () {
          return App.ajax.send({
            name: 'components.filter_by_status',
            sender: this,
            data: {
              clusterName: App.get('clusterName'),
              urlParams: urlParams
            },
            success: 'componentsFilterSuccessCallback',
            error: 'componentsFilterErrorCallback'
          });
        },
        getDisplayMessage: function () {
          var displayMsg = [];
          var siteProperties = this.get('siteProperties');
          if (siteProperties) {
            siteProperties.forEach(function (_siteProperty) {
              var displayProperty = _siteProperty.siteProperty;
              var displayNames = _siteProperty.displayNames;

              if (displayNames && displayNames.length) {
                if (displayNames.length === 1) {
                  displayMsg.push(displayProperty + Em.I18n.t('as') + displayNames[0]);
                } else {
                  var name;
                  displayNames.forEach(function (_name, index) {
                    if (index === 0) {
                      name = _name;
                    } else if (index === siteProperties.length - 1) {
                      name = name + Em.I18n.t('and') + _name;
                    } else {
                      name = name + ', ' + _name;
                    }
                  }, this);
                  displayMsg.push(displayProperty + Em.I18n.t('as') + name);

                }
              } else {
                displayMsg.push(displayProperty);
              }
            }, this);
          }
          return displayMsg;

        }.property('siteProperties'),

        runningHostsMessage: function () {
          return Em.I18n.t('services.service.config.stopService.runningHostComponents').format(this.get('runningComponentCount'), this.get('runningHosts.length'));
        }.property('runningComponentCount', 'runningHosts.length'),

        unknownHostsMessage: function () {
          return Em.I18n.t('services.service.config.stopService.unknownHostComponents').format(this.get('unknownComponentCount'), this.get('unknownHosts.length'));
        }.property('unknownComponentCount', 'unknownHosts.length'),

        templateName: require('templates/main/service/info/configs_save_popup')
      })
    })
  },

  /**
   * construct desired_configs for config groups from overriden properties
   * @param configs
   * @param timeTag
   * @return {Array}
   */
  buildGroupDesiredConfigs: function (configs, timeTag) {
    var sites = [];
    var time = timeTag || (new Date).getTime();
    var siteFileNames = configs.mapProperty('filename').uniq();
    sites = siteFileNames.map(function (filename) {
      return {
        type: filename.replace('.xml', ''),
        tag: 'version' + time,
        properties: []
      };
    });

    configs.forEach(function (config) {
      var type = config.get('filename').replace('.xml', '');
      var site = sites.findProperty('type', type);
      site.properties.push(config);
    });

    return sites.map(function (site) {
      return this.createSiteObj(site.type, site.tag, site.properties);
    }, this);
  },
  /**
   * persist properties of config groups to server
   * show result popup if <code>showPopup</code> is true
   * @param data {Object}
   * @param showPopup {Boolean}
   */
  putConfigGroupChanges: function (data, showPopup) {
    var ajaxOptions = {
      name: 'config_groups.update_config_group',
      sender: this,
      data: {
        id: data.ConfigGroup.id,
        configGroup: data
      }
    };
    if (showPopup) {
      ajaxOptions.success = "putConfigGroupChangesSuccess";
    }
    App.ajax.send(ajaxOptions);
  },

  putConfigGroupChangesSuccess: function () {
    this.set('saveConfigsFlag', true);
    this.onDoPUTClusterConfigurations();
  },

  /**
   * set hive hostnames in configs
   * @param configs
   */
  setHiveHostName: function (configs) {
    var dbHostPropertyName = null;
    if (configs.someProperty('name', 'hive_database')) {
      var hiveDb = configs.findProperty('name', 'hive_database');
      if (hiveDb.value === 'New MySQL Database' || hiveDb.value === 'New PostgreSQL Database') {
        var ambariHost = configs.findProperty('name', 'hive_ambari_host');
        if (ambariHost) {
          dbHostPropertyName = 'hive_ambari_host';
        }
        configs = configs.without(configs.findProperty('name', 'hive_existing_mysql_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mysql_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_oracle_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_oracle_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_postgresql_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_postgresql_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_2_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_2_host'));
      } else if (hiveDb.value === 'Existing MySQL Database') {
        var existingMySqlHost = configs.findProperty('name', 'hive_existing_mysql_host');
        if (existingMySqlHost) {
          dbHostPropertyName = 'hive_existing_mysql_host';
        }
        configs = configs.without(configs.findProperty('name', 'hive_ambari_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_oracle_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_oracle_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_postgresql_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_postgresql_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_2_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_2_host'));
      } else if (hiveDb.value === Em.I18n.t('services.service.config.hive.oozie.postgresql')) {
        var existingPostgreSqlHost = configs.findProperty('name', 'hive_existing_postgresql_host');
        if (existingPostgreSqlHost) {
          dbHostPropertyName = 'hive_existing_postgresql_host';
        }
        configs = configs.without(configs.findProperty('name', 'hive_ambari_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mysql_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mysql_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_oracle_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_oracle_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_2_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_2_host'));
      } else if (hiveDb.value === 'Existing Oracle Database') {
        var existingOracleHost = configs.findProperty('name', 'hive_existing_oracle_host');
        if (existingOracleHost) {
          dbHostPropertyName = 'hive_existing_oracle_host';
        }
        configs = configs.without(configs.findProperty('name', 'hive_ambari_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mysql_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mysql_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_postgresql_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_postgresql_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_2_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_2_host'));
      } else if (hiveDb.value === 'Existing MSSQL Server database with SQL authentication') {
        var existingMSSQLServerHost = configs.findProperty('name', 'hive_existing_mssql_server_host');
        if (existingMSSQLServerHost) {
          dbHostPropertyName = 'hive_existing_mssql_server_host';
        }
        configs = configs.without(configs.findProperty('name', 'hive_ambari_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mysql_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mysql_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_postgresql_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_postgresql_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_oracle_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_oracle_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_2_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_2_host'));
      } else if (hiveDb.value === 'Existing MSSQL Server database with integrated authentication') {
        var existingMSSQL2ServerHost = configs.findProperty('name', 'hive_existing_mssql_server_2_host');
        if (existingMSSQL2ServerHost) {
          dbHostPropertyName = 'hive_existing_mssql_server_2_host';
        }
        configs = configs.without(configs.findProperty('name', 'hive_ambari_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mysql_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mysql_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_postgresql_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_postgresql_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_oracle_host'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_oracle_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_database'));
        configs = configs.without(configs.findProperty('name', 'hive_existing_mssql_server_host'));
      }
    }
    if (dbHostPropertyName) {
      var hiveHostNameProperty = App.ServiceConfigProperty.create(App.config.get('preDefinedSiteProperties').findProperty('name', 'hive_hostname'));
      hiveHostNameProperty.set('value', configs.findProperty('name', dbHostPropertyName).get('value'));
      configs.pushObject(hiveHostNameProperty);
    }
  },

  /**
   * set oozie hostnames in configs
   * @param configs
   */
  setOozieHostName: function (configs) {
    var dbHostPropertyName = null;
    if (configs.someProperty('name', 'oozie_database')) {
      var oozieDb = configs.findProperty('name', 'oozie_database');
      if (oozieDb.value === 'New Derby Database') {
        configs = configs.without(configs.findProperty('name', 'oozie_ambari_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_ambari_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mysql_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mysql_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_oracle_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_oracle_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_postgresql_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_postgresql_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_2_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_2_host'));
      } else if (oozieDb.value === 'New MySQL Database') {
        var ambariHost = configs.findProperty('name', 'oozie_ambari_host');
        if (ambariHost) {
          ambariHost.name = 'oozie_hostname';
        }
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mysql_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mysql_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_oracle_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_oracle_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_derby_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_postgresql_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_postgresql_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_2_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_2_host'));

      } else if (oozieDb.value === 'Existing MySQL Database') {
        var existingMySqlHost = configs.findProperty('name', 'oozie_existing_mysql_host');
        if (existingMySqlHost) {
          dbHostPropertyName = 'oozie_existing_mysql_host';
        }
        configs = configs.without(configs.findProperty('name', 'oozie_ambari_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_ambari_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_oracle_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_oracle_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_derby_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_postgresql_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_postgresql_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_2_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_2_host'));
      } else if (oozieDb.value === Em.I18n.t('services.service.config.hive.oozie.postgresql')) {
        var existingPostgreSqlHost = configs.findProperty('name', 'oozie_existing_postgresql_host');


        if (existingPostgreSqlHost) {
          dbHostPropertyName = 'oozie_existing_postgresql_host';
        }
        configs = configs.without(configs.findProperty('name', 'oozie_ambari_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_ambari_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mysql_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mysql_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_oracle_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_oracle_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_2_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_2_host'));
      }
      else if (oozieDb.value === 'Existing Oracle Database') {
        var existingOracleHost = configs.findProperty('name', 'oozie_existing_oracle_host');
        if (existingOracleHost) {
          dbHostPropertyName = 'oozie_existing_oracle_host';
        }
        configs = configs.without(configs.findProperty('name', 'oozie_ambari_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_ambari_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mysql_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mysql_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_derby_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_2_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_2_host'));
      } else if (oozieDb.value === 'Existing MSSQL Server database with SQL authentication') {
        var existingMySqlServerHost = configs.findProperty('name', 'oozie_existing_mssql_server_host');
        if (existingMySqlServerHost) {
          dbHostPropertyName = 'oozie_existing_mssql_server_host';
        }
        configs = configs.without(configs.findProperty('name', 'oozie_ambari_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_ambari_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_oracle_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_oracle_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_derby_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_postgresql_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_postgresql_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mysql_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mysql_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_2_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_2_host'));
      } else if (oozieDb.value === 'Existing MSSQL Server database with integrated authentication') {
        var existingMySql2ServerHost = configs.findProperty('name', 'oozie_existing_mssql_server_2_host');
        if (existingMySql2ServerHost) {
          dbHostPropertyName = 'oozie_existing_mssql_server_2_host';
        }
        configs = configs.without(configs.findProperty('name', 'oozie_ambari_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_ambari_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_oracle_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_oracle_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_derby_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_postgresql_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_postgresql_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mysql_host'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mysql_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_database'));
        configs = configs.without(configs.findProperty('name', 'oozie_existing_mssql_server_host'));
      }
    }

    if (dbHostPropertyName) {
      var oozieHostNameProperty = App.ServiceConfigProperty.create(App.config.get('preDefinedSiteProperties').findProperty('name', 'oozie_hostname'));
      oozieHostNameProperty.set('value', configs.findProperty('name', dbHostPropertyName).get('value'));
      configs.pushObject(oozieHostNameProperty);
    }
  },

  /**
   * save site configs
   * @param configs
   */
  saveSiteConfigs: function (configs) {
    //storedConfigs contains custom configs as well
    this.setHiveHostName(configs);
    this.setOozieHostName(configs);
    this.formatConfigValues(configs);
    var mappedConfigs = App.config.excludeUnsupportedConfigs(this.get('configMapping').all(), App.Service.find().mapProperty('serviceName'));
    var allUiConfigs = this.loadUiSideConfigs(mappedConfigs);
    this.set('uiConfigs', configs.concat(allUiConfigs));
  },

  formatConfigValues: function (serviceConfigProperties) {
    serviceConfigProperties.forEach(function (_config) {
      if (typeof _config.get('value') === "boolean") _config.set('value', _config.value.toString());
      _config.set('value', App.config.trimProperty(_config, true));
    });
  },

  /**
   * return configs from the UI side
   * @param configMapping array with configs
   * @return {Array}
   */
  loadUiSideConfigs: function (configMapping) {
    var uiConfig = [];
    var configs = configMapping.filterProperty('foreignKey', null);
    this.addDynamicProperties(configs);
    configs.forEach(function (_config) {
      var valueWithOverrides = this.getGlobConfigValueWithOverrides(_config.templateName, _config.value, _config.name);
      if (valueWithOverrides !== null) {
        uiConfig.pushObject({
          "id": "site property",
          "name": _config.name,
          "value": valueWithOverrides.value,
          "filename": _config.filename,
          "overrides": valueWithOverrides.overrides
        });
      }
    }, this);
    return uiConfig;
  },


  addDynamicProperties: function (configs) {
    var allConfigs = this.get('stepConfigs').findProperty('serviceName', this.get('content.serviceName')).get('configs');
    var templetonHiveProperty = allConfigs.someProperty('name', 'templeton.hive.properties');
    if (!templetonHiveProperty && this.get('content.serviceName') === 'HIVE') {
      configs.pushObject({
        "name": "templeton.hive.properties",
        "templateName": ["hive.metastore.uris"],
        "foreignKey": null,
        "value": "hive.metastore.local=false,hive.metastore.uris=<templateName[0]>,hive.metastore.sasl.enabled=yes,hive.metastore.execute.setugi=true,hive.metastore.warehouse.dir=/apps/hive/warehouse",
        "filename": "webhcat-site.xml"
      });
    }
  },

  /**
   * return config value
   * @param templateName
   * @param expression
   * @param name
   * @return {Object}
   * example: <code>{
   *   value: '...',
   *   overrides: {
   *    'value1': [h1, h2],
   *    'value2': [h3]
   *   }
   * }</code>
   */
  getGlobConfigValueWithOverrides: function (templateName, expression, name) {
    var express = expression.match(/<(.*?)>/g);
    var value = expression;
    var overrideHostToValue = {};
    if (express != null) {
      express.forEach(function (_express) {
        var index = parseInt(_express.match(/\[([\d]*)(?=\])/)[1]);
        var globalObj = this.get('allConfigs').findProperty('name', templateName[index]);
        if (globalObj) {
          var globOverride = globalObj.overrides;
          if (globOverride != null) {
            for (var ov in globOverride) {
              globOverride[ov].forEach(function (host) {
                var replacedVal = (host in overrideHostToValue) ? overrideHostToValue[host] : expression;
                overrideHostToValue[host] = App.config.replaceConfigValues(name, _express, replacedVal, ov);
              }, this);
            }
          }
          value = App.config.replaceConfigValues(name, _express, expression, globalObj.value);
        } else {
          value = null;
        }
      }, this);
    }
    return this.getValueWithOverrides(value, overrideHostToValue)
  },

  getValueWithOverrides: function (value, overrideHostToValue) {
    var valueWithOverrides = {
      value: value,
      overrides: {}
    };
    if (!jQuery.isEmptyObject(overrideHostToValue)) {
      for (var host in overrideHostToValue) {
        var hostVal = overrideHostToValue[host];
        if (!(hostVal in valueWithOverrides.overrides)) {
          valueWithOverrides.overrides[hostVal] = [];
        }
        valueWithOverrides.overrides[hostVal].push(host);
      }
    }
    return valueWithOverrides;
  },

  /**
   * Saves cluster level configurations for all necessary sites.
   * PUT calls are made to /api/v1/clusters/clusterName for each site.
   * @return {Boolean}
   * @method doPUTClusterConfigurations
   */
  doPUTClusterConfigurations: function () {
    this.set('saveConfigsFlag', true);
    var serviceConfigTags = this.get('serviceConfigTags');
    this.setNewTagNames(serviceConfigTags);
    var siteNameToServerDataMap = {};
    var configsToSave = [];
    serviceConfigTags.forEach(function (_serviceTags) {
      var configs = this.createConfigObject(_serviceTags.siteName, _serviceTags.newTagName);
      if (configs) {
        configsToSave.push(configs);
        siteNameToServerDataMap[_serviceTags.siteName] = configs;
      }
    }, this);
    configsToSave = this.filterChangedConfiguration(configsToSave);
    if (configsToSave.length > 0) {
      this.doPUTClusterConfigurationSites(configsToSave);
    } else {
      this.onDoPUTClusterConfigurations();
    }
    this.set("savedSiteNameToServerServiceConfigDataMap", siteNameToServerDataMap);
  },

  /**
   * create different config object depending on siteName
   * @param {String} siteName
   * @param {String} tagName
   * @returns {Object|null}
   * @method createConfigObject
   */
  createConfigObject: function (siteName, tagName) {
    console.log("TRACE: Inside " + siteName);
    var configObject = {};
    switch (siteName) {
      case 'core-site':
        if (this.get('content.serviceName') === 'HDFS' || this.get('content.serviceName') === 'GLUSTERFS') {
          configObject = this.createCoreSiteObj(tagName);
        } else {
          return null;
        }
        break;
      default:
        var filename = (App.config.get('filenameExceptions').contains(siteName)) ? siteName : siteName + '.xml';
        if (filename === 'mapred-queue-acls.xml') {
          return null;
        }
        configObject = this.createSiteObj(siteName, tagName, this.get('uiConfigs').filterProperty('filename', filename));
       break;
    }
    configObject.service_config_version_note = this.get('serviceConfigVersionNote');
    return configObject;
  },

  /**
   * filter out unchanged configurations
   * @param {Array} configsToSave
   * @method filterChangedConfiguration
   */
  filterChangedConfiguration: function (configsToSave) {
    var changedConfigs = [];

    configsToSave.forEach(function (configSite) {
      var oldConfig = App.router.get('configurationController').getConfigsByTags([
        {siteName: configSite.type, tagName: this.loadedClusterSiteToTagMap[configSite.type]}
      ]);
      oldConfig = oldConfig[0] || {};
      var oldProperties = oldConfig.properties || {};
      var oldAttributes = oldConfig["properties_attributes"] || {};
      var newProperties = configSite.properties || {};
      var newAttributes = configSite["properties_attributes"] || {};
      if (this.isAttributesChanged(oldAttributes, newAttributes) || this.isConfigChanged(oldProperties, newProperties) || this.get('modifiedFileNames').contains(App.config.getOriginalFileName(configSite.type))) {
        changedConfigs.push(configSite);
      }
    }, this);
    return changedConfigs;
  },

  /**
   * Compares the loaded config values with the saving config values.
   * @param {Object} loadedConfig -
   * loadedConfig: {
   *      configName1: "configValue1",
   *      configName2: "configValue2"
   *   }
   * @param {Object} savingConfig
   * savingConfig: {
   *      configName1: "configValue1",
   *      configName2: "configValue2"
   *   }
   * @returns {boolean}
   * @method isConfigChanged
   */
  isConfigChanged: function (loadedConfig, savingConfig) {
    if (loadedConfig != null && savingConfig != null) {
      var seenLoadKeys = [];
      for (var loadKey in loadedConfig) {
        if (!loadedConfig.hasOwnProperty(loadKey)) continue;
        seenLoadKeys.push(loadKey);
        var loadValue = loadedConfig[loadKey];
        var saveValue = savingConfig[loadKey];
        if ("boolean" == typeof(saveValue)) {
          saveValue = saveValue.toString();
        }
        if (saveValue == null) {
          saveValue = "null";
        }
        if (loadValue !== saveValue) {
          return true;
        }
      }
      for (var saveKey in savingConfig) {
        if (seenLoadKeys.indexOf(saveKey) < 0) {
          return true;
        }
      }
    }
    return false;
  },

  /**
   * Compares the loaded config properties attributes with the saving config properties attributes.
   * @param {Object} oldAttributes -
   * oldAttributes: {
   *   supports: {
   *     final: {
   *       "configValue1" : "true",
   *       "configValue2" : "true"
   *     }
   *   }
   * }
   * @param {Object} newAttributes
   * newAttributes: {
   *   supports: {
   *     final: {
   *       "configValue1" : "true",
   *       "configValue2" : "true"
   *     }
   *   }
   * }
   * @returns {boolean}
   * @method isConfigChanged
   */
  isAttributesChanged: function (oldAttributes, newAttributes) {
    oldAttributes = oldAttributes.final || {};
    newAttributes = newAttributes.final || {};

    var key;
    for (key in oldAttributes) {
      if (oldAttributes.hasOwnProperty(key)
        && (!newAttributes.hasOwnProperty(key) || newAttributes[key] !== oldAttributes[key])) {
        return true;
      }
    }
    for (key in newAttributes) {
      if (newAttributes.hasOwnProperty(key)
        && (!oldAttributes.hasOwnProperty(key) || newAttributes[key] !== oldAttributes[key])) {
        return true;
      }
    }
    return false;
  },

  /**
   * Saves configuration of set of sites. The provided data
   * contains the site name and tag to be used.
   * @param {Object} sites
   * @method doPUTClusterConfigurationSites
   */
  doPUTClusterConfigurationSites: function (sites) {
    App.ajax.send({
      name: 'common.service.configurations',
      sender: this,
      data: {
        desired_config: sites
      },
      success: 'doPUTClusterConfigurationSiteSuccessCallback',
      error: 'doPUTClusterConfigurationSiteErrorCallback'
    });
  },

  doPUTClusterConfigurationSiteSuccessCallback: function () {
    this.onDoPUTClusterConfigurations();
  },

  doPUTClusterConfigurationSiteErrorCallback: function () {
    this.set('saveConfigsFlag', false);
    this.doPUTClusterConfigurationSiteSuccessCallback();
  },

  /**
   * add newTagName property to each config in serviceConfigs
   * @param serviceConfigs
   */
  setNewTagNames: function (serviceConfigs) {
    var time = (new Date).getTime();
    serviceConfigs.forEach(function (_serviceConfigs) {
      _serviceConfigs.newTagName = 'version' + time;
    }, this);
  },

  /**
   * Save "final" attribute for properties
   * @param {Array} properties - array of properties
   * @returns {Object|null}
   * */
  getConfigAttributes: function(properties) {
    var attributes = {
      final: {}
    };
    var finalAttributes = attributes.final;
    var hasAttributes = false;
    properties.forEach(function (property) {
      if (property.isRequiredByAgent !== false && property.isFinal) {
        hasAttributes = true;
        finalAttributes[property.name] = "true";
      }
    });
    if (hasAttributes) {
      return attributes;
    }
    return null;
  },

  /**
   * create core site object
   * @param tagName
   * @return {{"type": String, "tag": String, "properties": Object}}
   * @method createCoreSiteObj
   */
  createCoreSiteObj: function (tagName) {
    var coreSiteObj = this.get('uiConfigs').filterProperty('filename', 'core-site.xml');
    var coreSiteProperties = {};
    coreSiteObj.forEach(function (_coreSiteObj) {
      coreSiteProperties[_coreSiteObj.name] = _coreSiteObj.value;
      //this.recordHostOverride(_coreSiteObj, 'core-site', tagName, this);
    }, this);
    var result = {"type": "core-site", "tag": tagName, "properties": coreSiteProperties};
    var attributes = this.getConfigAttributes(coreSiteObj);
    if (attributes) {
      result['properties_attributes'] = attributes;
    }
    return result;
  },

  /**
   * create site object
   * @param siteName
   * @param tagName
   * @param siteObj
   * @return {Object}
   */
  createSiteObj: function (siteName, tagName, siteObj) {
    var heapsizeException = this.get('heapsizeException');
    var heapsizeRegExp = this.get('heapsizeRegExp');
    var siteProperties = {};
    siteObj.forEach(function (_siteObj) {
      var value = _siteObj.value;
      if (_siteObj.isRequiredByAgent == false) return;
      // site object name follow the format *permsize/*heapsize and the value NOT ends with "m"
      if (heapsizeRegExp.test(_siteObj.name) && !heapsizeException.contains(_siteObj.name) && !(_siteObj.value).endsWith("m")) {
        value += "m";
      }
      siteProperties[_siteObj.name] = value;
      switch (siteName) {
        case 'falcon-startup.properties':
        case 'falcon-runtime.properties':
        case 'pig-properties':
          siteProperties[_siteObj.name] = value;
          break;
        default:
          siteProperties[_siteObj.name] = this.setServerConfigValue(_siteObj.name, value);
      }
    }, this);
    var result = {"type": siteName, "tag": tagName, "properties": siteProperties};
    var attributes = this.getConfigAttributes(siteObj);
    if (attributes) {
      result['properties_attributes'] = attributes;
    }
    return result;
  },
  /**
   * This method will be moved to config's decorators class.
   *
   * For now, provide handling for special properties that need
   * be specified in special format required for server.
   *
   * @param configName {String} - name of config property
   * @param value {Mixed} - value of config property
   *
   * @return {String} - formated value
   */
  setServerConfigValue: function (configName, value) {
    switch (configName) {
      case 'storm.zookeeper.servers':
        if( Object.prototype.toString.call( value ) === '[object Array]' ) {
          return JSON.stringify(value).replace(/"/g, "'");
        } else {
          return value;
        }
        break;
      default:
        return value;
    }
  },

  /**
   * Array of Objects
   * {
   *  hostProperty - hostName property name for current component
   *  componentName - master componentName
   *  serviceName - serviceName of component
   *  serviceUseThis - services that use hostname property of component(componentName)
   *  m(multiple) - true if can be more than one components installed on cluster
   * }
   */

  hostComponentsmapping: [
    {
      hostProperty: 'snamenode_host',
      componentName: 'SECONDARY_NAMENODE',
      serviceName: 'HDFS',
      serviceUseThis: []
    },
    {
      hostProperty: 'jobtracker_host',
      componentName: 'JOBTRACKER',
      serviceName: 'MAPREDUCE2',
      serviceUseThis: []
    },
    {
      hostProperty: 'hs_host',
      componentName: 'HISTORYSERVER',
      serviceName: 'MAPREDUCE2',
      serviceUseThis: ['YARN']
    },
    {
      hostProperty: 'ats_host',
      componentName: 'APP_TIMELINE_SERVER',
      serviceName: 'YARN',
      serviceUseThis: []
    },
    {
      hostProperty: 'rm_host',
      componentName: 'RESOURCEMANAGER',
      serviceName: 'YARN',
      serviceUseThis: []
    },
    {
      hostProperty: 'hivemetastore_host',
      componentName: 'HIVE_METASTORE',
      serviceName: 'HIVE',
      serviceUseThis: ['HIVE'],
      m: true
    },
    {
      hostProperty: 'oozieserver_host',
      componentName: 'OOZIE_SERVER',
      serviceName: 'OOZIE',
      serviceUseThis: []
    },
    {
      hostProperty: 'hbasemaster_host',
      componentName: 'HBASE_MASTER',
      serviceName: 'HBASE',
      serviceUseThis: [],
      m: true
    },
    {
      hostProperty: 'webhcatserver_host',
      componentName: 'WEBHCAT_SERVER',
      serviceName: 'HIVE',
      serviceUseThis: [],
      m: true
    },
    {
      hostProperty: 'zookeeperserver_hosts',
      componentName: 'ZOOKEEPER_SERVER',
      serviceName: 'ZOOKEEPER',
      serviceUseThis: ['HBASE', 'HIVE'],
      m: true
    },
    {
      hostProperty: 'stormuiserver_host',
      componentName: 'STORM_UI_SERVER',
      serviceName: 'STORM',
      serviceUseThis: []
    },
    {
      hostProperty: 'drpcserver_host',
      componentName: 'DRPC_SERVER',
      serviceName: 'STORM',
      serviceUseThis: []
    },
    {
      hostProperty: 'storm_rest_api_host',
      componentName: 'STORM_REST_API',
      serviceName: 'STORM',
      serviceUseThis: []
    },
    {
      hostProperty: 'supervisor_hosts',
      componentName: 'SUPERVISOR',
      serviceName: 'STORM',
      serviceUseThis: [],
      m: true
    }
  ],

  /**
   * Adds host name of master component to config;
   * @method addHostNamesToGlobalConfig
   */
  addHostNamesToConfig: function () {
    var serviceName = this.get('content.serviceName');
    var configs = this.get('allConfigs');
    //namenode_host is required to derive "fs.default.name" a property of core-site
    try {
      this.setHostForService('HDFS', 'NAMENODE', 'namenode_host', true);
    } catch (err) {
      console.log("No NameNode Host available.  This is expected if you're using GLUSTERFS rather than HDFS.");
    }

    var hostProperties = this.get('hostComponentsmapping').filter(function (h) {
      return h.serviceUseThis.contains(serviceName) || h.serviceName == serviceName;
    });
    hostProperties.forEach(function (h) {
      this.setHostForService(h.serviceName, h.componentName, h.hostProperty, h.m);
    }, this);

    if (serviceName === 'HIVE') {
      var hiveDb = configs.findProperty('name', 'hive_database').value;
      if (['Existing MySQL Database', 'Existing Oracle Database', 'Existing PostgreSQL Database', 'Existing MSSQL Server database with SQL authentication', 'Existing MSSQL Server database with integrated authentication'].contains(hiveDb)) {
        configs.findProperty('name', 'hive_hostname').isVisible = true;
      }
    }
    if (serviceName === 'OOZIE') {
      var oozieDb = configs.findProperty('name', 'oozie_database').value;
      if (['Existing MySQL Database', 'Existing Oracle Database', 'Existing PostgreSQL Database', 'Existing MSSQL Server database with SQL authentication', 'Existing MSSQL Server database with integrated authentication'].contains(oozieDb)) {
        configs.findProperty('name', 'oozie_hostname').isVisible = true;
      }
    }
  },
  /**
   * set host name(s) property for component
   * @param {String} serviceName - service name of component
   * @param {String} componentName - component name which host we want to know
   * @param {String} hostProperty - name of host property for current component
   * @param {Boolean} multiple - true if can be more than one component
   * @method setHostForService
   */
  setHostForService: function (serviceName, componentName, hostProperty, multiple) {
    var configs = this.get('allConfigs');
    var serviceConfigs = this.get('serviceConfigs').findProperty('serviceName', serviceName).get('configs');
    var hostConfig = serviceConfigs.findProperty('name', hostProperty);
    if (hostConfig) {
      hostConfig.defaultValue = this.getMasterComponentHostValue(componentName, multiple);
      configs.push(hostConfig);
    }
  },

  /**
   * get hostName of component
   * @param {String} componentName
   * @param {Boolean} multiple - true if can be more than one component installed on cluster
   * @return {String|Array|Boolean} hostName|hostNames|false if missing component
   * @method getMasterComponentHostValue
   */
  getMasterComponentHostValue: function (componentName, multiple) {
    var components = this.get('content.hostComponents').filterProperty('componentName', componentName);
    if (components.length > 0) {
      return multiple ? components.mapProperty('hostName') : components[0].get('hostName');
    }
    return false;
  },
  /**
   * Provides service component name and display-name information for
   * the current selected service.
   * @return {Em.Array} validComponents - array of valid components
   * @method getCurrentServiceComponents
   */
  getCurrentServiceComponents: function () {
    var components = this.get('content.hostComponents');
    var validComponents = Ember.A([]);
    var seenComponents = {};
    components.forEach(function (component) {
      var cn = component.get('componentName');
      var cdn = component.get('displayName');
      if (!seenComponents[cn]) {
        validComponents.push(Ember.Object.create({
          componentName: cn,
          displayName: cdn,
          selected: false
        }));
        seenComponents[cn] = cn;
      }
    });
    return validComponents;
  }.property('content'),

  /**
   * trigger loadStep
   * @method loadStep
   */
  doCancel: function () {
    this.set('preSelectedConfigVersion', null);
    Em.run.once(this, 'onConfigGroupChange');
  },

  /**
   * trigger restartAllServiceHostComponents(batchUtils) if confirmed in popup
   * @method restartAllStaleConfigComponents
   * @return App.showConfirmationFeedBackPopup
   */
  restartAllStaleConfigComponents: function () {
    var self = this;
    var serviceDisplayName = this.get('content.displayName');
    var bodyMessage = Em.Object.create({
      confirmMsg: Em.I18n.t('services.service.restartAll.confirmMsg').format(serviceDisplayName),
      confirmButton: Em.I18n.t('services.service.restartAll.confirmButton'),
      additionalWarningMsg: this.get('content.passiveState') === 'OFF' ? Em.I18n.t('services.service.restartAll.warningMsg.turnOnMM').format(serviceDisplayName) : null
    });
    return App.showConfirmationFeedBackPopup(function (query) {
      var selectedService = self.get('content.id');
      batchUtils.restartAllServiceHostComponents(selectedService, true, query);
    }, bodyMessage);
  },

  /**
   * trigger launchHostComponentRollingRestart(batchUtils)
   * @method rollingRestartStaleConfigSlaveComponents
   */
  rollingRestartStaleConfigSlaveComponents: function (componentName) {
    batchUtils.launchHostComponentRollingRestart(componentName.context, this.get('content.displayName'), this.get('content.passiveState') === "ON", true);
  },

  /**
   * trigger showItemsShouldBeRestarted popup with hosts that requires resetart
   * @param event
   * @method showHostsShouldBeRestarted
   */
  showHostsShouldBeRestarted: function (event) {
    var restartRequiredHostsAndComponents = event.context;
    var hosts = [];
    for (var hostName in restartRequiredHostsAndComponents) {
      hosts.push(hostName);
    }
    var hostsText = hosts.length == 1 ? Em.I18n.t('common.host') : Em.I18n.t('common.hosts');
    hosts = hosts.join(', ');
    this.showItemsShouldBeRestarted(hosts, Em.I18n.t('service.service.config.restartService.shouldBeRestarted').format(hostsText));
  },

  /**
   * trigger showItemsShouldBeRestarted popup with components that requires resetart
   * @param event
   * @method showComponentsShouldBeRestarted
   */
  showComponentsShouldBeRestarted: function (event) {
    var restartRequiredHostsAndComponents = event.context;
    var hostsComponets = [];
    var componentsObject = {};
    for (var hostName in restartRequiredHostsAndComponents) {
      restartRequiredHostsAndComponents[hostName].forEach(function (hostComponent) {
        hostsComponets.push(hostComponent);
        if (componentsObject[hostComponent] != undefined) {
          componentsObject[hostComponent]++;
        } else {
          componentsObject[hostComponent] = 1;
        }
      })
    }
    var componentsList = [];
    for (var obj in componentsObject) {
      var componentDisplayName = (componentsObject[obj] > 1) ? obj + 's' : obj;
      componentsList.push(componentsObject[obj] + ' ' + componentDisplayName);
    }
    var componentsText = componentsList.length == 1 ? Em.I18n.t('common.component') : Em.I18n.t('common.components');
    hostsComponets = componentsList.join(', ');
    this.showItemsShouldBeRestarted(hostsComponets, Em.I18n.t('service.service.config.restartService.shouldBeRestarted').format(componentsText));
  },

  showItemsShouldBeRestarted: function (content, header) {
    App.ModalPopup.show({
      content: content,
      header: header,
      bodyClass: Em.View.extend({
        templateName: require('templates/common/selectable_popup'),
        textareaVisible: false,
        textTrigger: function () {
          this.set('textareaVisible', !this.get('textareaVisible'));
        },
        putContentToTextarea: function () {
          var content = this.get('parentView.content');
          if (this.get('textareaVisible')) {
            var wrapper = $(".task-detail-log-maintext");
            $('.task-detail-log-clipboard').html(content).width(wrapper.width()).height(wrapper.height());
            Em.run.next(function () {
              $('.task-detail-log-clipboard').select();
            });
          }
        }.observes('textareaVisible')
      }),
      secondary: null
    });
  },

  /**
   * add new overridden property to config property object
   * @param {object} serviceConfigProperty - config property object
   * @param {App.ConfigGroup} group - config group for new property
   * @method addOverrideProperty
   */
  addOverrideProperty: function (serviceConfigProperty, group) {
    var overrides = serviceConfigProperty.get('overrides');
    if (!overrides) {
      overrides = [];
      serviceConfigProperty.set('overrides', overrides);
    }
    // create new override with new value
    var newSCP = App.ServiceConfigProperty.create(serviceConfigProperty);
    newSCP.set('value', '');
    newSCP.set('isOriginalSCP', false); // indicated this is overridden value,
    newSCP.set('parentSCP', serviceConfigProperty);
    newSCP.set('isEditable', true);
    newSCP.set('group', group);
    console.debug("createOverrideProperty(): Added:", newSCP, " to main-property:", serviceConfigProperty);
    overrides.pushObject(newSCP);
  },

  /**
   * trigger manageConfigurationGroups
   * @method manageConfigurationGroup
   */
  manageConfigurationGroup: function () {
    this.manageConfigurationGroups();
  },

  manageConfigurationGroups: function (controller) {
    var serviceData = (controller && controller.get('selectedService')) || this.get('content');
    var serviceName = serviceData.get('serviceName');
    var displayName = serviceData.get('displayName');
    App.router.get('manageConfigGroupsController').set('isInstaller', !!controller);
    App.router.get('manageConfigGroupsController').set('serviceName', serviceName);
    if (controller) {
      App.router.get('manageConfigGroupsController').set('isAddService', controller.get('content.controllerName') == 'addServiceController');
    }
    return App.ModalPopup.show({
      header: Em.I18n.t('services.service.config_groups_popup.header').format(displayName),
      bodyClass: App.MainServiceManageConfigGroupView.extend({
        serviceName: serviceName,
        displayName: displayName,
        controllerBinding: 'App.router.manageConfigGroupsController'
      }),
      classNames: ['sixty-percent-width-modal', 'manage-configuration-group-popup'],
      primary: Em.I18n.t('common.save'),
      onPrimary: function () {
        var modifiedConfigGroups = this.get('subViewController.hostsModifiedConfigGroups');
        // Save modified config-groups
        if (!!controller) {
          controller.set('selectedService.configGroups', App.router.get('manageConfigGroupsController.configGroups'));
          controller.selectedServiceObserver();
          if (controller.get('name') == "wizardStep7Controller") {
            if (controller.get('selectedService.selected') === false && modifiedConfigGroups.toDelete.length > 0) {
              controller.setGroupsToDelete(modifiedConfigGroups.toDelete);
            }
            App.config.persistWizardStep7ConfigGroups();
            this.updateConfigGroupOnServicePage();
          }
          this.hide();
          return;
        }
        console.log("manageConfigurationGroups(): Saving modified config-groups: ", modifiedConfigGroups);
        var self = this;
        var errors = [];
        var deleteQueriesCounter = modifiedConfigGroups.toClearHosts.length + modifiedConfigGroups.toDelete.length;
        var createQueriesCounter = modifiedConfigGroups.toSetHosts.length + modifiedConfigGroups.toCreate.length;
        var deleteQueriesRun = false;
        var createQueriesRun = false;
        var runNextQuery = function () {
          if (!deleteQueriesRun && deleteQueriesCounter > 0) {
            deleteQueriesRun = true;
            modifiedConfigGroups.toClearHosts.forEach(function (cg) {
              App.config.clearConfigurationGroupHosts(cg, finishFunction, finishFunction);
            }, this);
            modifiedConfigGroups.toDelete.forEach(function (cg) {
              App.config.deleteConfigGroup(cg, finishFunction, finishFunction);
            }, this);
          } else if (!createQueriesRun && deleteQueriesCounter < 1) {
            createQueriesRun = true;
            modifiedConfigGroups.toSetHosts.forEach(function (cg) {
              App.config.updateConfigurationGroup(cg, finishFunction, finishFunction);
            }, this);
            modifiedConfigGroups.toCreate.forEach(function (cg) {
              App.config.postNewConfigurationGroup(cg, finishFunction);
            }, this);
          }
        };
        var finishFunction = function (xhr, text, errorThrown) {
          if (xhr && errorThrown) {
            var error = xhr.status + "(" + errorThrown + ") ";
            try {
              var json = $.parseJSON(xhr.responseText);
              error += json.message;
            } catch (err) {
            }
            console.error('Error updating Config Group:', error);
            errors.push(error);
          }
          if (createQueriesRun) {
            createQueriesCounter--;
          } else {
            deleteQueriesCounter--;
          }
          if (deleteQueriesCounter + createQueriesCounter < 1) {
            if (errors.length > 0) {
              console.log(errors);
              self.get('subViewController').set('errorMessage', errors.join(". "));
            } else {
              self.updateConfigGroupOnServicePage();
              self.hide();
            }
          } else {
            runNextQuery();
          }
        };
        runNextQuery();
      },
      onSecondary: function () {
        this.hide();
      },
      onClose: function () {
        this.hide();
      },
      subViewController: function () {
        return App.router.get('manageConfigGroupsController');
      }.property('App.router.manageConfigGroupsController'),

      updateConfigGroupOnServicePage: function () {
        var subViewController = this.get('subViewController');
        var selectedConfigGroup = subViewController.get('selectedConfigGroup');
        var managedConfigGroups = subViewController.get('configGroups');
        if (!controller) {
          controller = App.router.get('mainServiceInfoConfigsController');
          controller.set('configGroups', managedConfigGroups);
        } else {
          controller.set('selectedService.configGroups', managedConfigGroups);
        }

        var selectEventObject = {};
        //check whether selectedConfigGroup exists
        if (selectedConfigGroup && controller.get('configGroups').someProperty('name', selectedConfigGroup.get('name'))) {
          selectEventObject.context = selectedConfigGroup;
        } else {
          selectEventObject.context = managedConfigGroups.findProperty('isDefault', true);
        }
        controller.selectConfigGroup(selectEventObject);
      },

      updateButtons: function () {
        var modified = this.get('subViewController.isHostsModified');
        this.set('disablePrimary', !modified);
      }.observes('subViewController.isHostsModified'),
      secondary: Em.I18n.t('common.cancel'),
      didInsertElement: Em.K
    });
  },

  /**
   * If user changes cfg group if some configs was changed popup with propose to save changes must be shown
   * @param {object} event - triggered event for selecting another config-group
   * @method selectConfigGroup
   */
  selectConfigGroup: function (event) {
    var self = this;

    function callback() {
      self.doSelectConfigGroup(event);
    }

    if (!this.get('isInit')) {
      if (this.hasUnsavedChanges()) {
        this.showSavePopup(null, callback);
        return;
      }
    }
    callback();
  },
  /**
   * switch view to selected group
   * @param event
   * @method selectConfigGroup
   */
  doSelectConfigGroup: function (event) {
    //clean when switch config group
    this.loadedGroupToOverrideSiteToTagMap = {};
    var configGroupVersions = App.ServiceConfigVersion.find().filterProperty('groupId', event.context.get('id'));
    //check whether config group has config versions
    if (configGroupVersions.length > 0) {
      this.loadSelectedVersion(configGroupVersions.findProperty('isCurrent').get('version'), event.context);
    } else {
      this.loadSelectedVersion(null, event.context);
    }
  },

  /**
   * Are some unsaved changes available
   * @returns {boolean}
   */
  hasUnsavedChanges: function () {
    return this.get('hash') != this.getHash();
  },

  /**
   * If some configs are changed and user navigates away or select another config-group, show this popup with propose to save changes
   * @param {String} path
   * @param {object} callback - callback with action to change configs view(change group or version)
   * @method showSavePopup
   */
  showSavePopup: function (path, callback) {
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('common.warning'),
      bodyClass: Em.View.extend({
        templateName: require('templates/common/configs/save_configuration'),
        showSaveWarning: true,
        notesArea: Em.TextArea.extend({
          classNames: ['full-width'],
          placeholder: Em.I18n.t('dashboard.configHistory.info-bar.save.popup.placeholder'),
          onChangeValue: function() {
            this.get('parentView.parentView').set('serviceConfigNote', this.get('value'));
          }.observes('value')
        })
      }),
      footerClass: Ember.View.extend({
        templateName: require('templates/main/service/info/save_popup_footer')
      }),
      primary: Em.I18n.t('common.save'),
      secondary: Em.I18n.t('common.cancel'),
      onSave: function () {
        self.set('serviceConfigVersionNote', this.get('serviceConfigNote'));
        self.restartServicePopup();
        this.hide();
      },
      onDiscard: function () {
        self.set('preSelectedConfigVersion', null);
        if (path) {
          self.set('forceTransition', true);
          App.router.route(path);
        } else if (callback) {
          // Prevent multiple popups
          self.set('hash', self.getHash());
          callback();
        }
        this.hide();
      },
      onCancel: function () {
        this.hide();
      }
    });
  }
});

});

require.register("controllers/main/service/info/summary", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.MainServiceInfoSummaryController = Em.Controller.extend({
  name: 'mainServiceInfoSummaryController',

  selectedFlumeAgent: null,

  /**
   * Indicates whether Ranger plugins status update polling is active
   * @type {boolean}
   */
  isRangerUpdateWorking: false,

  /**
   * Indicates whether array with initial Ranger plugins data is set
   * @type {boolean}
   */
  isRangerPluginsArraySet: false,

  /**
   * Indicates whether previous AJAX request for Ranger plugins config properties has failed
   * @type {boolean}
   */
  isPreviousRangerConfigsCallFailed: false,

  /**
   * Ranger plugins data
   * @type {array}
   */
  rangerPlugins: [
    {
      serviceName: 'HDFS',
      type: 'ranger-hdfs-plugin-properties',
      propertyName: 'ranger-hdfs-plugin-enabled'
    },
    {
      serviceName: 'HIVE',
      type: 'ranger-hive-plugin-properties',
      propertyName: 'ranger-hive-plugin-enabled'
    },
    {
      serviceName: 'HBASE',
      type: 'ranger-hbase-plugin-properties',
      propertyName: 'ranger-hbase-plugin-enabled'
    },
    {
      serviceName: 'KNOX',
      type: 'ranger-knox-plugin-properties',
      propertyName: 'ranger-knox-plugin-enabled'
    },
    {
      serviceName: 'STORM',
      type: 'ranger-storm-plugin-properties',
      propertyName: 'ranger-storm-plugin-enabled'
    }
  ],

  /**
   * Set initial Ranger plugins data
   * @method setRangerPlugins
   */
  setRangerPlugins: function () {
    if (App.get('router.clusterController.isLoaded') && !this.get('isRangerPluginsArraySet')) {
      this.setProperties({
        rangerPlugins: this.get('rangerPlugins').map(function (item) {
          var stackService = App.StackService.find().findProperty('serviceName', item.serviceName);
          var displayName = (stackService) ? stackService.get('displayName') : item.serviceName;
          return $.extend(item, {
            pluginTitle: Em.I18n.t('services.service.summary.ranger.plugin.title').format(displayName),
            isDisplayed: App.Service.find().someProperty('serviceName', item.serviceName),
            status: Em.I18n.t('services.service.summary.ranger.plugin.loadingStatus')
          });
        }),
        isRangerPluginsArraySet: true
      });
    }
  }.observes('App.router.clusterController.isLoaded'),

  /**
   * Get latest config tags
   * @method updateRangerPluginsStatus
   * @param callback
   */
  updateRangerPluginsStatus: function (callback) {
    App.ajax.send({
      name: 'config.tags',
      sender: this,
      success: 'getRangerPluginsStatus',
      callback: callback
    });
  },

  /**
   * Get latest Ranger plugins config properties
   * @method getRangerPluginsStatus
   * @param data
   */
  getRangerPluginsStatus: function (data) {
    var urlParams = [];
    this.get('rangerPlugins').forEach(function (item) {
      if (App.Service.find().someProperty('serviceName', item.serviceName)) {
        var currentTag = data.Clusters.desired_configs[item.type].tag;
        var isTagChanged = item.tag != currentTag;
        Em.set(item, 'isDisplayed', true);
        //Request for properties should be sent either if configs have changed or if previous Ranger plugins config properties has failed
        if (isTagChanged || this.get('isPreviousRangerConfigsCallFailed')) {
          Em.set(item, 'tag', currentTag);
          urlParams.push('(type=' + item.type + '&tag=' + currentTag + ')');
        }
      } else {
        Em.set(item, 'isDisplayed', false);
      }
    }, this);
    if (urlParams.length) {
      App.ajax.send({
        name: 'reassign.load_configs',
        sender: this,
        data: {
          urlParams: urlParams.join('|')
        },
        success: 'getRangerPluginsStatusSuccess',
        error: 'getRangerPluginsStatusError'
      });
    }
  },

  /**
   * Set Ranger plugins statuses
   * @method getRangerPluginsStatusSuccess
   * @param data
   */
  getRangerPluginsStatusSuccess: function (data) {
    this.set('isPreviousRangerConfigsCallFailed', false);
    data.items.forEach(function (item) {
      var serviceName = this.get('rangerPlugins').findProperty('type', item.type).serviceName;
      var propertyName = this.get('rangerPlugins').findProperty('type', item.type).propertyName;
      var statusMap = {
        Yes: 'alerts.table.state.enabled',
        No: 'alerts.table.state.disabled'
      };
      var statusString = statusMap[item.properties[propertyName]] || 'common.unknown';
      Em.set(this.get('rangerPlugins').findProperty('serviceName', serviceName), 'status', Em.I18n.t(statusString));
    }, this);
  },

  /**
   * Method executed if Ranger plugins config properties request has failed
   * @method getRangerPluginsStatusError
   */
  getRangerPluginsStatusError: function () {
    this.set('isPreviousRangerConfigsCallFailed', true);
  },

  /**
   * Send start command for selected Flume Agent
   * @method startFlumeAgent
   */
  startFlumeAgent: function () {
    var selectedFlumeAgent = arguments[0].context;
    if (selectedFlumeAgent && selectedFlumeAgent.get('status') === 'NOT_RUNNING') {
      var self = this;
      App.showConfirmationPopup(function () {
        var state = 'STARTED';
        var context = Em.I18n.t('services.service.summary.flume.start.context').format(selectedFlumeAgent.get('name'));
        self.sendFlumeAgentCommandToServer(state, context, selectedFlumeAgent);
      });
    }
  },

  /**
   * Send stop command for selected Flume Agent
   * @method stopFlumeAgent
   */
  stopFlumeAgent: function () {
    var selectedFlumeAgent = arguments[0].context;
    if (selectedFlumeAgent && selectedFlumeAgent.get('status') === 'RUNNING') {
      var self = this;
      App.showConfirmationPopup(function () {
        var state = 'INSTALLED';
        var context = Em.I18n.t('services.service.summary.flume.stop.context').format(selectedFlumeAgent.get('name'));
        self.sendFlumeAgentCommandToServer(state, context, selectedFlumeAgent);
      });
    }
  },

  /**
   * Send command for Flume Agent to server
   * @param {string} state
   * @param {string} context
   * @param {Object} agent
   * @method sendFlumeAgentCommandToServer
   */
  sendFlumeAgentCommandToServer: function (state, context, agent) {
    App.ajax.send({
      name: 'service.flume.agent.command',
      sender: this,
      data: {
        state: state,
        context: context,
        agentName: agent.get('name'),
        host: agent.get('hostName')
      },
      success: 'commandSuccessCallback'
    });
  },

  /**
   * Callback, that shows Background operations popup if request was successful
   */
  commandSuccessCallback: function () {
    console.log('Send request for refresh configs successfully');
    // load data (if we need to show this background operations popup) from persist
    App.router.get('applicationController').dataLoading().done(function (showPopup) {
      if (showPopup) {
        App.router.get('backgroundOperationsController').showPopup();
      }
    });
  },

  gotoConfigs: function () {
    App.router.get('mainServiceItemController').set('routeToConfigs', true);
    App.router.transitionTo('main.services.service.configs', this.get('content'));
    App.router.get('mainServiceItemController').set('routeToConfigs', false);
  },

  showServiceAlertsPopup: function (event) {
    var service = event.context;
    return App.ModalPopup.show({
      header: Em.I18n.t('services.service.summary.alerts.popup.header').format(service.get('displayName')),
      autoHeight: false,
      bodyClass: Em.View.extend({
        templateName: require('templates/main/service/info/service_alert_popup'),
        classNames: ['service-alerts'],
        controllerBinding: 'App.router.mainAlertDefinitionsController',
        didInsertElement: function () {
          Em.run.next(this, function () {
            App.tooltip($(".timeago"));
          });
        },
        alerts: function () {
          var serviceDefinitions = this.get('controller.content').filterProperty('service', service);
          // definitions should be sorted in order: critical, warning, ok, unknown, other
          var criticalDefinitions = [], warningDefinitions = [], okDefinitions = [], unknownDefinitions = [];
          serviceDefinitions.forEach(function (definition) {
            if (definition.get('isCritical')) {
              criticalDefinitions.push(definition);
              serviceDefinitions = serviceDefinitions.without(definition);
            } else if (definition.get('isWarning')) {
              warningDefinitions.push(definition);
              serviceDefinitions = serviceDefinitions.without(definition);
            } else if (definition.get('isOK')) {
              okDefinitions.push(definition);
              serviceDefinitions = serviceDefinitions.without(definition);
            } else if (definition.get('isUnknown')) {
              unknownDefinitions.push(definition);
              serviceDefinitions = serviceDefinitions.without(definition);
            }
          });
          serviceDefinitions = criticalDefinitions.concat(warningDefinitions, okDefinitions, unknownDefinitions, serviceDefinitions);
          return serviceDefinitions;
        }.property('controller.content'),
        gotoAlertDetails: function (event) {
          if (event && event.context) {
            this.get('parentView').hide();
            App.router.transitionTo('main.alerts.alertDetails', event.context);
          }
        },
        closePopup: function () {
          this.get('parentView').hide();
        }
      }),
      isHideBodyScroll: false,
      primary: Em.I18n.t('common.close'),
      secondary: null
    });
  }

});
});

require.register("controllers/main/service/item", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var batchUtils = require('utils/batch_scheduled_requests');
var componentsUtils = require('utils/components');

App.MainServiceItemController = Em.Controller.extend({
  name: 'mainServiceItemController',

  /**
   * Callback functions for start and stop service have few differences
   *
   * Used with currentCallBack property
   */
  callBackConfig: {
    'STARTED': {
      'c': 'STARTING',
      'f': 'starting',
      'c2': 'live',
      'hs': 'started',
      's': 'start'
    },
    'INSTALLED': {
      'c': 'STOPPING',
      'f': 'stopping',
      'c2': 'dead',
      'hs': 'stopped',
      's': 'stop'
    }
  },

  initHosts: function() {
    if (App.get('components.masters').length !== 0) {
      var self = this;

      var hostNames = App.Host.find().mapProperty('hostName');
      this.set('allHosts', hostNames);

      ['HBASE_MASTER', 'HIVE_METASTORE', 'ZOOKEEPER_SERVER', 'FLUME_HANDLER', 'HIVE_SERVER'].forEach(function(componentName) {
        self.loadHostsWithoutComponent(componentName);
      });
    }
  }.observes('App.components.masters', 'content.hostComponents.length'),

  loadHostsWithoutComponent: function (componentName) {
    var self = this;
    var hostsWithComponent = App.HostComponent.find().filterProperty('componentName', componentName).mapProperty('hostName');

    var hostsWithoutComponent = this.get('allHosts').filter(function(hostName) {
      return !hostsWithComponent.contains(hostName);
    });

    self.set('add' + componentName, function() {
      App.get('router.mainAdminKerberosController').getKDCSessionState(function() {
        self.addComponent(componentName);
      });
    });

    Em.defineProperty(self, 'addDisabledTooltip-' + componentName, Em.computed('isAddDisabled-' + componentName, 'addDisabledMsg-' + componentName, function() {
      if (self.get('isAddDisabled-' + componentName)) {
        return self.get('addDisabledMsg-' + componentName);
      }
    }));

    Em.defineProperty(self, 'isAddDisabled-' + componentName, Em.computed('hostsWithoutComponent-' + componentName, function() {
      return self.get('hostsWithoutComponent-' + componentName).length === 0 ? 'disabled' : '';
    }));

    var disabledMsg = Em.I18n.t('services.summary.allHostsAlreadyRunComponent').format(componentName);
    self.set('hostsWithoutComponent-' + componentName, hostsWithoutComponent);
    self.set('addDisabledMsg-' + componentName, disabledMsg);
  },

  /**
   * flag to control router switch between service summary and configs
   * @type {boolean}
   */
  routeToConfigs: false,

  isClientsOnlyService: function() {
    return App.get('services.clientOnly').contains(this.get('content.serviceName'));
  }.property('content.serviceName'),

  isConfigurable: function () {
    return !App.get('services.noConfigTypes').contains(this.get('content.serviceName'));
  }.property('App.services.noConfigTypes','content.serviceName'),

  allHosts: [],

  clientComponents: function () {
    var clientNames = [];
    var clients = App.StackServiceComponent.find().filterProperty('serviceName', this.get('content.serviceName')).filterProperty('isClient');
    clients.forEach(function (item) {
      clientNames.push({
        action: 'downloadClientConfigs',
        context: {
          name: item.get('componentName'),
          label: item.get('displayName')
        }
      });
    });
    return clientNames;
  }.property('content.serviceName'),

  /**
   * Common method for ajax (start/stop service) responses
   * @param data
   * @param ajaxOptions
   * @param params
   */
  startStopPopupSuccessCallback: function (data, ajaxOptions, params) {
    if (data && data.Requests) {
      params.query.set('status', 'SUCCESS');
      var config = this.get('callBackConfig')[(JSON.parse(ajaxOptions.data)).Body.ServiceInfo.state];
      var self = this;
      console.log('Send request for ' + config.c + ' successfully');
      if (App.get('testMode')) {
        self.set('content.workStatus', App.Service.Health[config.f]);
        self.get('content.hostComponents').setEach('workStatus', App.HostComponentStatus[config.f]);
        setTimeout(function () {
          self.set('content.workStatus', App.Service.Health[config.c2]);
          self.get('content.hostComponents').setEach('workStatus', App.HostComponentStatus[config.hs]);
        }, App.get('testModeDelayForActions'));
      }
      // load data (if we need to show this background operations popup) from persist
      App.router.get('applicationController').dataLoading().done(function (initValue) {
        if (initValue) {
          App.router.get('backgroundOperationsController').showPopup();
        }
      });
    } else {
      params.query.set('status', 'FAIL');
      console.log('cannot get request id from ', data);
    }
  },
  startStopPopupErrorCallback: function(request, ajaxOptions, error, opt, params){
    params.query.set('status', 'FAIL');
  },
  /**
   * Confirmation popup for start/stop services
   * @param event
   * @param serviceHealth - 'STARTED' or 'INSTALLED'
   */
  startStopPopup: function(event, serviceHealth) {
    if ($(event.target).hasClass('disabled') || $(event.target.parentElement).hasClass('disabled')) {
      return;
    }
    var self = this;
    var serviceDisplayName = this.get('content.displayName');
    var isMaintenanceOFF = this.get('content.passiveState') === 'OFF';
    var bodyMessage = Em.Object.create({
      putInMaintenance: (serviceHealth == 'INSTALLED' && isMaintenanceOFF) || (serviceHealth == 'STARTED' && !isMaintenanceOFF),
      turnOnMmMsg: serviceHealth == 'INSTALLED' ? Em.I18n.t('passiveState.turnOnFor').format(serviceDisplayName) : Em.I18n.t('passiveState.turnOffFor').format(serviceDisplayName),
      confirmMsg: serviceHealth == 'INSTALLED'? Em.I18n.t('services.service.stop.confirmMsg').format(serviceDisplayName) : Em.I18n.t('services.service.start.confirmMsg').format(serviceDisplayName),
      confirmButton: serviceHealth == 'INSTALLED'? Em.I18n.t('services.service.stop.confirmButton') : Em.I18n.t('services.service.start.confirmButton'),
      additionalWarningMsg:  isMaintenanceOFF && serviceHealth == 'INSTALLED'? Em.I18n.t('services.service.stop.warningMsg.turnOnMM').format(serviceDisplayName) : null
    });

    return App.showConfirmationFeedBackPopup(function(query, runMmOperation) {
      self.set('isPending', true);
      self.startStopWithMmode(serviceHealth, query, runMmOperation);
    }, bodyMessage);
  },


  startStopWithMmode: function(serviceHealth, query, runMmOperation) {
    var self = this;
    if (runMmOperation) {
      if (serviceHealth == "STARTED") {
        this.startStopPopupPrimary(serviceHealth, query).complete(function() {
          batchUtils.turnOnOffPassiveRequest("OFF", Em.I18n.t('passiveState.turnOff'), self.get('content.serviceName').toUpperCase());
        });
      } else {
        batchUtils.turnOnOffPassiveRequest("ON", Em.I18n.t('passiveState.turnOn'), this.get('content.serviceName').toUpperCase()).complete(function() {
          self.startStopPopupPrimary(serviceHealth, query);
        })
      }
    } else {
      this.startStopPopupPrimary(serviceHealth, query);
    }

  },

  startStopPopupPrimary: function (serviceHealth, query) {
    var requestInfo = (serviceHealth == "STARTED")
        ? App.BackgroundOperationsController.CommandContexts.START_SERVICE.format(this.get('content.serviceName'))
        : App.BackgroundOperationsController.CommandContexts.STOP_SERVICE.format(this.get('content.serviceName'));

    var data = {
      'context': requestInfo,
      'serviceName': this.get('content.serviceName').toUpperCase(),
      'ServiceInfo': {
        'state': serviceHealth
      },
      'query': query
    };

    return App.ajax.send({
      'name': 'common.service.update',
      'sender': this,
      'success': 'startStopPopupSuccessCallback',
      'error': 'startStopPopupErrorCallback',
      'data': data
    });
  },

  /**
   * On click callback for <code>start service</code> button
   * @param event
   */
  startService: function (event) {
    this.startStopPopup(event, App.HostComponentStatus.started);
  },

  /**
   * On click callback for <code>stop service</code> button
   * @param event
   */
  stopService: function (event) {
    this.startStopPopup(event, App.HostComponentStatus.stopped);
  },

  /**
   * On click callback for <code>run rebalancer</code> button
   * @param event
   */
  runRebalancer: function (event) {
    var self = this;
    return App.showConfirmationPopup(function() {
      self.set("content.runRebalancer", true);
      // load data (if we need to show this background operations popup) from persist
      App.router.get('applicationController').dataLoading().done(function (initValue) {
        if (initValue) {
          App.router.get('backgroundOperationsController').showPopup();
        }
      });
    });
  },
   /**
   * On click handler for Yarn Refresh Queues command from items menu
   * @param event
   */
  refreshYarnQueues : function (event) {
    var controller = this;
    var hosts = App.Service.find('YARN').get('hostComponents').filterProperty('componentName', 'RESOURCEMANAGER').mapProperty('hostName');
    return App.showConfirmationPopup(function() {
    App.ajax.send({
      name : 'service.item.refreshQueueYarnRequest',
        sender: controller,
      data : {
        command : "REFRESHQUEUES",
        context : Em.I18n.t('services.service.actions.run.yarnRefreshQueues.context') ,
        hosts : hosts.join(','),
        serviceName : "YARN",
        componentName : "RESOURCEMANAGER",
        forceRefreshConfigTags : "capacity-scheduler"
      },
      success : 'refreshYarnQueuesSuccessCallback',
      error : 'refreshYarnQueuesErrorCallback'
    });
    });
  },
  refreshYarnQueuesSuccessCallback  : function(data, ajaxOptions, params) {
    if (data.Requests.id) {
      App.router.get('backgroundOperationsController').showPopup();
    } else {
      console.warn('Error during refreshYarnQueues');
    }
  },
  refreshYarnQueuesErrorCallback : function(data) {
    var error = Em.I18n.t('services.service.actions.run.yarnRefreshQueues.error');
    if(data && data.responseText){
      try {
        var json = $.parseJSON(data.responseText);
        error += json.message;
      } catch (err) {}
    }
    App.showAlertPopup(Em.I18n.t('services.service.actions.run.yarnRefreshQueues.error'), error);
    console.warn('Error during refreshYarnQueues:'+error);
  },

  startLdapKnox: function(event) {
    var context =  Em.I18n.t('services.service.actions.run.startLdapKnox.context');
    this.startStopLdapKnox('STARTDEMOLDAP',context);
  },

  stopLdapKnox: function(event) {
    var context = Em.I18n.t('services.service.actions.run.stopLdapKnox.context');
    this.startStopLdapKnox('STOPDEMOLDAP',context);
  },

  startStopLdapKnox: function(command,context) {
    var controller = this;
    var host = App.HostComponent.find().findProperty('componentName', 'KNOX_GATEWAY').get('hostName');
    return App.showConfirmationPopup(function() {
      App.ajax.send({
        name: 'service.item.startStopLdapKnox',
        sender: controller,
        data: {
          command: command,
          context: context,
          host: host,
          serviceName: "KNOX",
          componentName: "KNOX_GATEWAY"
        },
        success: 'startStopLdapKnoxSuccessCallback',
        error: 'startStopLdapKnoxErrorCallback'
      });
    });
  },

  startStopLdapKnoxSuccessCallback  : function(data, ajaxOptions, params) {
    if (data.Requests.id) {
      App.router.get('backgroundOperationsController').showPopup();
    } else {
      console.warn('Error during startStopLdapKnox');
    }
  },
  startStopLdapKnoxErrorCallback : function(data) {
    var error = Em.I18n.t('services.service.actions.run.startStopLdapKnox.error');
    if(data && data.responseText){
      try {
        var json = $.parseJSON(data.responseText);
        error += json.message;
      } catch (err) {}
    }
    App.showAlertPopup(Em.I18n.t('services.service.actions.run.yarnRefreshQueues.error'), error);
    console.warn('Error during refreshYarnQueues:'+ error);
  },

  /**
   * On click handler for rebalance Hdfs command from items menu
   */
  rebalanceHdfsNodes: function () {
    var controller = this;
    App.ModalPopup.show({
      classNames: ['fourty-percent-width-modal'],
      header: Em.I18n.t('services.service.actions.run.rebalanceHdfsNodes.context'),
      primary: Em.I18n.t('common.start'),
      secondary: Em.I18n.t('common.cancel'),
      inputValue: 10,
      errorMessage: Em.I18n.t('services.service.actions.run.rebalanceHdfsNodes.promptError'),
      isInvalid: function () {
        var intValue = Number(this.get('inputValue'));
        return this.get('inputValue')!=='DEBUG' && (isNaN(intValue) || intValue < 1 || intValue > 100);
      }.property('inputValue'),
      disablePrimary : function() {
        return this.get('isInvalid');
      }.property('isInvalid'),
      onPrimary: function () {
        if (this.get('isInvalid')) {
          return;
        }
        App.ajax.send({
          name : 'service.item.rebalanceHdfsNodes',
          sender: controller,
          data : {
            hosts : App.Service.find('HDFS').get('hostComponents').findProperty('componentName', 'NAMENODE').get('hostName'),
            threshold: this.get('inputValue')
          },
          success : 'rebalanceHdfsNodesSuccessCallback',
          error : 'rebalanceHdfsNodesErrorCallback'
        });
        this.hide();
      },
      bodyClass: Ember.View.extend({
        templateName: require('templates/common/modal_popups/prompt_popup'),
        text: Em.I18n.t('services.service.actions.run.rebalanceHdfsNodes.prompt'),
        didInsertElement: function () {
          App.tooltip(this.$(".prompt-input"), {
            placement: "bottom",
            title: Em.I18n.t('services.service.actions.run.rebalanceHdfsNodes.promptTooltip')
          });
        }
      })
    });
  },
  rebalanceHdfsNodesSuccessCallback: function (data) {
    if (data.Requests.id) {
      App.router.get('backgroundOperationsController').showPopup();
    } else {
      console.warn('Error during runRebalanceHdfsNodes');
    }
  },
  rebalanceHdfsNodesErrorCallback : function(data) {
    var error = Em.I18n.t('services.service.actions.run.rebalanceHdfsNodes.error');
    if(data && data.responseText){
      try {
        var json = $.parseJSON(data.responseText);
        error += json.message;
      } catch (err) {
      }
    }
    App.showAlertPopup(Em.I18n.t('services.service.actions.run.rebalanceHdfsNodes.error'), error);
    console.warn('Error during runRebalanceHdfsNodes:'+error);
  },

  /**
   * On click callback for <code>run compaction</code> button
   * @param event
   */
  runCompaction: function (event) {
    var self = this;
    return App.showConfirmationPopup(function() {
      self.set("content.runCompaction", true);
      // load data (if we need to show this background operations popup) from persist
      App.router.get('applicationController').dataLoading().done(function (initValue) {
        if (initValue) {
          App.router.get('backgroundOperationsController').showPopup();
        }
      });
    });
  },

  /**
   * On click callback for <code>run smoke test</code> button
   * @param event
   */
  runSmokeTest: function (event) {
    var self = this;
    if (this.get('content.serviceName') === 'MAPREDUCE2' && !App.Service.find('YARN').get('isStarted')) {
      return App.showAlertPopup(Em.I18n.t('common.error'), Em.I18n.t('services.mapreduce2.smokeTest.requirement'));
    }
    return App.showConfirmationFeedBackPopup(function(query) {
      self.runSmokeTestPrimary(query);
    });
  },

  restartAllHostComponents : function(serviceName) {
    var serviceDisplayName = this.get('content.displayName');
    var bodyMessage = Em.Object.create({
      putInMaintenance: this.get('content.passiveState') === 'OFF',
      turnOnMmMsg: Em.I18n.t('passiveState.turnOnFor').format(serviceDisplayName),
      confirmMsg: Em.I18n.t('services.service.restartAll.confirmMsg').format(serviceDisplayName),
      confirmButton: Em.I18n.t('services.service.restartAll.confirmButton'),
      additionalWarningMsg: this.get('content.passiveState') === 'OFF' ? Em.I18n.t('services.service.restartAll.warningMsg.turnOnMM').format(serviceDisplayName): null
     });
    return App.showConfirmationFeedBackPopup(function(query, runMmOperation) {
      batchUtils.restartAllServiceHostComponents(serviceName, false, query, runMmOperation);
    }, bodyMessage);
  },

  turnOnOffPassive: function(label) {
    var self = this;
    var state = this.get('content.passiveState') == 'OFF' ? 'ON' : 'OFF';
    var onOff = state === 'ON' ? "On" : "Off";
    return App.showConfirmationPopup(function() {
          batchUtils.turnOnOffPassiveRequest(state, label, self.get('content.serviceName').toUpperCase(), function(data, opt, params) {
            self.set('content.passiveState', params.passive_state);
            batchUtils.infoPassiveState(params.passive_state);})
        },
        Em.I18n.t('hosts.passiveMode.popup').format(onOff,self.get('content.displayName'))
    );
  },

  rollingRestart: function(hostComponentName) {
    batchUtils.launchHostComponentRollingRestart(hostComponentName, this.get('content.displayName'), this.get('content.passiveState') === "ON", false, this.get('content.passiveState') === "ON");
  },

  runSmokeTestPrimary: function(query) {
    var clusterLevelRequired = ['KERBEROS'];
    var requestData = {
        'serviceName': this.get('content.serviceName'),
        'displayName': this.get('content.displayName'),
        'actionName': this.get('content.serviceName') === 'ZOOKEEPER' ? 'ZOOKEEPER_QUORUM_SERVICE_CHECK' : this.get('content.serviceName') + '_SERVICE_CHECK',
        'query': query
    };
    if (clusterLevelRequired.contains(this.get('content.serviceName'))) {
      requestData.operationLevel = {
        "level": "CLUSTER",
        "cluster_name": App.get('clusterName')
      };
    }
    App.ajax.send({
      'name': 'service.item.smoke',
      'sender': this,
      'success':'runSmokeTestSuccessCallBack',
      'error':'runSmokeTestErrorCallBack',
      'data': requestData
    });
  },

  runSmokeTestSuccessCallBack: function (data, ajaxOptions, params) {
    if (data.Requests.id) {
      // load data (if we need to show this background operations popup) from persist
      App.router.get('applicationController').dataLoading().done(function (initValue) {
        params.query.set('status', 'SUCCESS');
        if (initValue) {
          App.router.get('backgroundOperationsController').showPopup();
        }
      });
    }
    else {
      params.query.set('status', 'FAIL');
      console.warn('error during runSmokeTestSuccessCallBack');
    }
  },
  runSmokeTestErrorCallBack: function (request, ajaxOptions, error, opt, params) {
    params.query.set('status', 'FAIL');
  },

  /**
   * On click callback for <code>Reassign <master component></code> button
   * @param hostComponent
   */
  reassignMaster: function (hostComponent) {
    var component = App.HostComponent.find().findProperty('componentName', hostComponent);
    console.log('In Reassign Master', hostComponent);
    if (component) {
      var reassignMasterController = App.router.get('reassignMasterController');
      reassignMasterController.saveComponentToReassign(component);
      reassignMasterController.getSecurityStatus();
      reassignMasterController.setCurrentStep('1');
      App.router.transitionTo('reassign');
    }
  },

  /**
   * On click callback for <code>action</code> dropdown menu
   * Calls runSmokeTest, runRebalancer, runCompaction or reassignMaster depending on context
   * @param event
   */
  doAction: function (event) {
    if ($(event.target).hasClass('disabled') || $(event.target.parentElement).hasClass('disabled')) {
      return;
    }
    var methodName = event.context.action;
    var context = event.context.context;
    if (methodName) {
      this[methodName](context);
    }
  },

  /**
   * Restart clients host components to apply config changes
   */
  refreshConfigs: function () {
    var self = this;
    if (this.get('isClientsOnlyService') || this.get('content.serviceName') == "FLUME") {
      return App.showConfirmationFeedBackPopup(function (query) {
        batchUtils.getComponentsFromServer({
          services: [self.get('content.serviceName')]
        }, function (data) {
          var hostComponents = [];

          data.items.forEach(function (host) {
            host.host_components.forEach(function (hostComponent) {
              hostComponents.push(Em.Object.create({
                componentName: hostComponent.HostRoles.component_name,
                hostName: host.Hosts.host_name
              }))
            });
          });
          batchUtils.restartHostComponents(hostComponents, Em.I18n.t('rollingrestart.context.allForSelectedService').format(self.get('content.serviceName')), "SERVICE", query);
        })
      });
    }
  },

  /**
   * Send command to server to install client on selected host
   * @param componentName
   */
  addComponent: function (componentName) {
    var self = this;
    var component = App.StackServiceComponent.find().findProperty('componentName', componentName);
    var componentDisplayName = component.get('displayName');

    self.loadHostsWithoutComponent(componentName);

    return App.ModalPopup.show({
      primary: function() {
        if (this.get('anyHostsWithoutComponent')) {
          return Em.I18n.t('hosts.host.addComponent.popup.confirm')
        } else {
          return undefined;
        }
      }.property('anyHostsWithoutComponent'),

      header: Em.I18n.t('popup.confirmation.commonHeader'),

      addComponentMsg: function () {
        return Em.I18n.t('hosts.host.addComponent.msg').format(componentDisplayName);
      }.property(),

      selectHostMsg: function () {
        return Em.I18n.t('services.summary.selectHostForComponent').format(this.get('componentDisplayName'))
      }.property('componentDisplayName'),

      thereIsNoHostsMsg: function () {
        return Em.I18n.t('services.summary.allHostsAlreadyRunComponent').format(this.get('componentDisplayName'))
      }.property('componentDisplayName'),

      hostsWithoutComponent: function() {
        return self.get("hostsWithoutComponent-" + this.get('componentName'));
      }.property('componentName', 'self.hostsWithoutComponent-' + this.get('componentName')),

      anyHostsWithoutComponent: function() {
        return this.get('hostsWithoutComponent').length > 0
      }.property('hostsWithoutComponent'),

      selectedHost: null,

      componentName: function() {
        return componentName;
      }.property(),

      componentDisplayName: function() {
        return componentDisplayName;
      }.property(),

      bodyClass: Em.View.extend({
        templateName: require('templates/main/service/add_host_popup')
      }),

      onPrimary: function () {
        var selectedHost = this.get('selectedHost');

        // Install
        if(component.get('componentName') == "HIVE_METASTORE" && !!selectedHost){
          App.router.get('mainHostDetailsController').addComponent(
            {
              context: component,
              hiveMetastoreHost: selectedHost
            }
          );
        } else {
          componentsUtils.installHostComponent(selectedHost, component);
        }

        // Remove host from 'without' collection to immediate recalculate add menu item state
        var hostsWithoutComponent = this.get('hostsWithoutComponent');
        var index = hostsWithoutComponent.indexOf(this.get('selectedHost'));
        if (index > -1) {
          hostsWithoutComponent.splice(index, 1);
        }

        self.set('hostsWithoutComponent-' + this.get('componentName'), hostsWithoutComponent);
        this.hide();
      }
    });
  },

  /**
   * set property isPending (if this property is true - means that service has task in BGO)
   * and this makes start/stop button disabled
   */
  setStartStopState: function () {
    var serviceName = this.get('content.serviceName');
    var backgroundOperations = App.router.get('backgroundOperationsController.services');
    if (backgroundOperations && backgroundOperations.length > 0) {
      for (var i = 0; i < backgroundOperations.length; i++) {
        if (backgroundOperations[i].isRunning &&
            (backgroundOperations[i].dependentService === "ALL_SERVICES" ||
             backgroundOperations[i].dependentService === serviceName)) {
          this.set('isPending', true);
          return;
        }
      }
      this.set('isPending', false);
    } else {
      this.set('isPending', true);
    }
  }.observes('App.router.backgroundOperationsController.serviceTimestamp'),

  isStartDisabled: function () {
    if(this.get('isPending')) return true;
    return !(this.get('content.healthStatus') == 'red');
  }.property('content.healthStatus','isPending'),

  isStopDisabled: function () {
    if(this.get('isPending')) return true;
    if (App.get('isHaEnabled') && this.get('content.serviceName') == 'HDFS' && this.get('content.hostComponents').filterProperty('componentName', 'NAMENODE').someProperty('workStatus', App.HostComponentStatus.started)) {
      return false;
    }
    return (this.get('content.healthStatus') != 'green');
  }.property('content.healthStatus','isPending'),

  /**
   * Determine if service has than one service client components
   */
  isSeveralClients: function () {
    return App.StackServiceComponent.find().filterProperty('serviceName', this.get('content.serviceName')).filterProperty('isClient').length > 1;
  }.property('content.serviceName'),

  enableHighAvailability: function() {
    var ability_controller = App.router.get('mainAdminHighAvailabilityController');
    ability_controller.enableHighAvailability();
  },

  disableHighAvailability: function() {
    var ability_controller = App.router.get('mainAdminHighAvailabilityController');
    ability_controller.disableHighAvailability();
  },

  enableRMHighAvailability: function() {
    var ability_controller = App.router.get('mainAdminHighAvailabilityController');
    ability_controller.enableRMHighAvailability();
  },

  downloadClientConfigs: function (event) {
    var component = this.get('content.clientComponents').rejectProperty('totalCount', 0)[0];
    componentsUtils.downloadClientConfigs.call(this, {
      serviceName: this.get('content.serviceName'),
      componentName: (event && event.name) || component.get('componentName'),
      displayName: (event && event.label) || component.get('displayName')
    });
  },

  /**
   * On click handler for custom command from items menu
   * @param context
   */
  executeCustomCommand: function(context) {
    var controller = this;
    return App.showConfirmationPopup(function() {
      App.ajax.send({
        name : 'service.item.executeCustomCommand',
        sender: controller,
        data : {
          command : context.command,
          context : 'Execute ' + context.command,
          hosts : App.Service.find(context.service).get('hostComponents').findProperty('componentName', context.component).get('hostName'),
          serviceName : context.service,
          componentName : context.component,
          forceRefreshConfigTags : "capacity-scheduler"
        },
        success : 'executeCustomCommandSuccessCallback',
        error : 'executeCustomCommandErrorCallback'
      });
    });
  },

  executeCustomCommandSuccessCallback  : function(data, ajaxOptions, params) {
    if (data.Requests.id) {
      App.router.get('backgroundOperationsController').showPopup();
    } else {
      console.warn('Error during execution of ' + params.command + ' custom command on' + params.componentName);
    }
  },

  executeCustomCommandErrorCallback : function(data) {
    var error = Em.I18n.t('services.service.actions.run.executeCustomCommand.error');
    if(data && data.responseText){
      try {
        var json = $.parseJSON(data.responseText);
        error += json.message;
      } catch (err) {}
    }
    App.showAlertPopup(Em.I18n.t('services.service.actions.run.executeCustomCommand.error'), error);
    console.warn('Error during executing custom command');
  },

  isPending:true

});

});

require.register("controllers/main/service/manage_config_groups_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');
var validator = require('utils/validator');
var hostsManagement = require('utils/hosts');
var numberUtils = require('utils/number_utils');

App.ManageConfigGroupsController = Em.Controller.extend({
  name: 'manageConfigGroupsController',

  isLoaded: false,

  isInstaller: false,

  isAddService: false,

  serviceName: null,

  configGroups: [],

  originalConfigGroups: [],

  selectedConfigGroup: null,

  selectedHosts: [],

  clusterHosts: [],

  resortConfigGroup: function() {
    var configGroups = Ember.copy(this.get('configGroups'));
    if(configGroups.length < 2){
      return;
    }
    var defaultConfigGroup = configGroups.findProperty('isDefault');
    configGroups.removeObject(defaultConfigGroup);
    var sorted = [defaultConfigGroup].concat(configGroups.sortProperty('name'));

    this.removeObserver('configGroups.@each.name', this, 'resortConfigGroup');
    this.set('configGroups', sorted);
    this.addObserver('configGroups.@each.name', this, 'resortConfigGroup');
  }.observes('configGroups.@each.name'),

  loadHosts: function() {
    this.set('isLoaded', false);
    if (this.get('isInstaller')) {
      var allHosts = this.get('isAddService') ? App.router.get('addServiceController').get('allHosts') : App.router.get('installerController').get('allHosts');
      this.set('clusterHosts', allHosts);
      this.loadConfigGroups(this.get('serviceName'));
    } else {
      this.loadHostsFromServer();
    }
  },

  /**
   * request all hosts directly from server
   */
  loadHostsFromServer: function() {
    App.ajax.send({
      name: 'hosts.config_groups',
      sender: this,
      data: {},
      success: 'loadHostsFromServerSuccessCallback',
      error: 'loadHostsFromServerErrorCallback'
    });
  },

  /**
   * parse hosts response and wrap them into Ember.Object
   * @param data
   */
  loadHostsFromServerSuccessCallback: function (data) {
    var wrappedHosts = [];

    data.items.forEach(function (host) {
      var hostComponents = [];
      var diskInfo = host.Hosts.disk_info.filter(function(item) {
        return /^ext|^ntfs|^fat|^xfs/i.test(item.type);
      });
      if (diskInfo.length) {
        diskInfo = diskInfo.reduce(function(a, b) {
          return {
            available: parseInt(a.available) + parseInt(b.available),
            size: parseInt(a.size) + parseInt(b.size)
          };
        });
      }
      host.host_components.forEach(function (hostComponent) {
        hostComponents.push(Em.Object.create({
          componentName: hostComponent.HostRoles.component_name,
          displayName: App.format.role(hostComponent.HostRoles.component_name)
        }));
      }, this);
      wrappedHosts.pushObject(Em.Object.create({
          id: host.Hosts.host_name,
          ip: host.Hosts.ip,
          osType: host.Hosts.os_type,
          osArch: host.Hosts.os_arch,
          hostName: host.Hosts.host_name,
          publicHostName: host.Hosts.public_host_name,
          cpu: host.Hosts.cpu_count,
          memory: host.Hosts.total_mem,
          diskTotal: numberUtils.bytesToSize(diskInfo.size, 0, undefined, 1024),
          diskFree: numberUtils.bytesToSize(diskInfo.available, 0, undefined, 1024),
          disksMounted: host.Hosts.disk_info.length,
          hostComponents: hostComponents
        }
      ));
    }, this);

    this.set('clusterHosts', wrappedHosts);
    this.loadConfigGroups(this.get('serviceName'));
  },

  loadHostsFromServerErrorCallback: function () {
    console.warn('ERROR: request to fetch all hosts failed');
    this.set('clusterHosts', []);
    this.loadConfigGroups(this.get('serviceName'));
  },

  loadConfigGroups: function (serviceName) {
    if (this.get('isInstaller')) {
      this.set('serviceName', serviceName);
      var configGroups = this.copyConfigGroups(App.router.get('wizardStep7Controller.selectedService.configGroups'));
      var originalConfigGroups = this.copyConfigGroups(configGroups);
      this.set('configGroups', configGroups);
      this.set('originalConfigGroups', originalConfigGroups);
      this.set('isLoaded', true);
    } else {
      this.set('serviceName', serviceName);
      App.ajax.send({
        name: 'service.load_config_groups',
        data: {
          serviceName: serviceName
        },
        sender: this,
        success: 'onLoadConfigGroupsSuccess',
        error: 'onLoadConfigGroupsError'
      });
    }
  },

  onLoadConfigGroupsSuccess: function (data) {
    var usedHosts = [];
    var unusedHosts = [];
    var serviceName = this.get('serviceName');
    var serviceDisplayName =  App.StackService.find().findProperty('serviceName', this.get('serviceName')).get('displayName');
    var defaultConfigGroup = App.ConfigGroup.create({
      name: serviceDisplayName + " Default",
      description: "Default cluster level " + this.get('serviceName') + " configuration",
      isDefault: true,
      parentConfigGroup: null,
      service: this.get('content'),
      configSiteTags: [],
      serviceName: serviceName
    });

    if (data && data.items) {
      var groupToTypeToTagMap = {};
      var configGroups = [];
      data.items.forEach(function (configGroup) {
        configGroup = configGroup.ConfigGroup;
        var hostNames = configGroup.hosts.mapProperty('host_name');
        var newConfigGroup = App.ConfigGroup.create({
          id: configGroup.id,
          name: configGroup.group_name,
          description: configGroup.description,
          isDefault: false,
          parentConfigGroup: defaultConfigGroup,
          service: App.Service.find().findProperty('serviceName', configGroup.tag),
          hosts: hostNames,
          configSiteTags: [],
          properties: [],
          apiResponse: configGroup
        });
        usedHosts = usedHosts.concat(newConfigGroup.get('hosts'));
        configGroups.push(newConfigGroup);
        var newConfigGroupSiteTags = newConfigGroup.get('configSiteTags');
        configGroup.desired_configs.forEach(function (config) {
          newConfigGroupSiteTags.push(App.ConfigSiteTag.create({
            site: config.type,
            tag: config.tag
          }));
          if (!groupToTypeToTagMap[configGroup.group_name]) {
            groupToTypeToTagMap[configGroup.group_name] = {}
          }
          groupToTypeToTagMap[configGroup.group_name][config.type] = config.tag;
        });
      }, this);
      unusedHosts = this.get('clusterHosts').mapProperty('hostName');
      usedHosts.uniq().forEach(function (host) {
        unusedHosts = unusedHosts.without(host);
      }, this);
      defaultConfigGroup.set('childConfigGroups', configGroups);
      defaultConfigGroup.set('hosts', unusedHosts);
      var allGroups = [defaultConfigGroup].concat(configGroups);
      this.set('configGroups', allGroups);
      var originalGroups = this.copyConfigGroups(allGroups);
      this.set('originalConfigGroups', originalGroups);
      this.loadProperties(groupToTypeToTagMap);
      this.set('isLoaded', true);
    }
  },

  onLoadConfigGroupsError: function () {
    console.error('Unable to load config groups for service.');
  },

  loadProperties: function (groupToTypeToTagMap) {
    var typeTagToGroupMap = {};
    var urlParams = [];
    for (var group in groupToTypeToTagMap) {
      var overrideTypeTags = groupToTypeToTagMap[group];
      for (var type in overrideTypeTags) {
        var tag = overrideTypeTags[type];
        typeTagToGroupMap[type + "///" + tag] = group;
        urlParams.push('(type=' + type + '&tag=' + tag + ')');
      }
    }
    var params = urlParams.join('|');
    if (urlParams.length) {
      App.ajax.send({
        name: 'config.host_overrides',
        sender: this,
        data: {
          params: params,
          typeTagToGroupMap: typeTagToGroupMap
        },
        success: 'onLoadPropertiesSuccess'
      });
    }
  },

  onLoadPropertiesSuccess: function (data, opt, params) {
    data.items.forEach(function (configs) {
      var typeTagConfigs = [];
      App.config.loadedConfigurationsCache[configs.type + "_" + configs.tag] = configs.properties;
      var group = params.typeTagToGroupMap[configs.type + "///" + configs.tag];
      for (var config in configs.properties) {
        typeTagConfigs.push(Em.Object.create({
          name: config,
          value: configs.properties[config]
        }));
      }
      this.get('configGroups').findProperty('name', group).get('properties').pushObjects(typeTagConfigs);
    }, this);
  },

  showProperties: function () {
    var properies = this.get('selectedConfigGroup.propertiesList').htmlSafe();
    if (properies) {
      App.showAlertPopup(Em.I18n.t('services.service.config_groups_popup.properties'), properies);
    }
  },
  addHosts: function () {
    if (this.get('selectedConfigGroup.isAddHostsDisabled')){
      return false;
    }
    var availableHosts = this.get('selectedConfigGroup.availableHosts');
    var popupDescription = {
      header: Em.I18n.t('hosts.selectHostsDialog.title'),
      dialogMessage: Em.I18n.t('hosts.selectHostsDialog.message').format(this.get('selectedConfigGroup.displayName'))
    };
    hostsManagement.launchHostsSelectionDialog(availableHosts, [], false, this.get('componentsForFilter'), this.addHostsCallback.bind(this), popupDescription);
  },

  /**
   * add hosts callback
   * @param {string[]} selectedHosts
   * @method addHostsCallback
   */
  addHostsCallback: function (selectedHosts) {
    var group = this.get('selectedConfigGroup');
    if (selectedHosts) {
      selectedHosts.forEach(function (hostName) {
        group.get('hosts').pushObject(hostName);
        group.get('parentConfigGroup.hosts').removeObject(hostName);
      }, this);
    }
  },

  /**
   * delete hosts from group
   * @method deleteHosts
   */
  deleteHosts: function () {
    if (this.get('isDeleteHostsDisabled')) {
      return;
    }
    this.get('selectedHosts').slice().forEach(function (hostName) {
      this.get('selectedConfigGroup.parentConfigGroup.hosts').pushObject(hostName);
      this.get('selectedConfigGroup.hosts').removeObject(hostName);
    }, this);
    this.set('selectedHosts', []);
  },

  isDeleteHostsDisabled: function () {
    var selectedConfigGroup = this.get('selectedConfigGroup');
    if (selectedConfigGroup) {
      return selectedConfigGroup.isDefault || this.get('selectedHosts').length === 0;
    }
    return true;
  }.property('selectedConfigGroup', 'selectedConfigGroup.hosts.length', 'selectedHosts.length'),

  /**
   * confirm delete config group
   */
  confirmDelete : function () {
    var self = this;
    App.showConfirmationPopup(function() {
      self.deleteConfigGroup();
    });
  },
  /**
   * add hosts to group
   * @return {Array}
   */
  componentsForFilter: function () {
    return App.StackServiceComponent.find().filterProperty('serviceName', this.get('serviceName')).map(function (component) {
      return Em.Object.create({
        componentName: component.get('componentName'),
        displayName: App.format.role(component.get('componentName')),
        selected: false
      });
    });
  }.property('serviceName'),

  /**
   * delete selected config group
   */
  deleteConfigGroup: function () {
    var selectedConfigGroup = this.get('selectedConfigGroup');
    if (this.get('isDeleteGroupDisabled')) {
      return;
    }
    //move hosts of group to default group (available hosts)
    this.set('selectedHosts', selectedConfigGroup.get('hosts'));
    this.deleteHosts();
    this.get('configGroups').removeObject(selectedConfigGroup);
    this.set('selectedConfigGroup', this.get('configGroups').findProperty('isDefault'));
  },
  /**
   * rename new config group
   */
  renameConfigGroup: function () {
    if(this.get('selectedConfigGroup.isDefault')) {
      return;
    }
    var self = this;
    this.renameGroupPopup = App.ModalPopup.show({
      primary: Em.I18n.t('ok'),
      secondary: Em.I18n.t('common.cancel'),
      header: Em.I18n.t('services.service.config_groups.rename_config_group_popup.header'),
      bodyClass: Ember.View.extend({
        templateName: require('templates/main/service/new_config_group')
      }),
      configGroupName: self.get('selectedConfigGroup.name'),
      configGroupDesc: self.get('selectedConfigGroup.description'),
      warningMessage: null,
      isDescriptionDirty: false,
      validate: function () {
        var warningMessage = '';
        var originalGroup = self.get('selectedConfigGroup');
        var groupName = this.get('configGroupName').trim();
        if (originalGroup.get('description') !== this.get('configGroupDesc') && !this.get('isDescriptionDirty')) {
          this.set('isDescriptionDirty', true);
        }
        if (originalGroup.get('name').trim() === groupName) {
          if (this.get('isDescriptionDirty')) {
            warningMessage = '';
          } else {
            warningMessage = Em.I18n.t("config.group.selection.dialog.err.name.exists");
          }
        } else {
          if (self.get('configGroups').mapProperty('name').contains(groupName)) {
            warningMessage = Em.I18n.t("config.group.selection.dialog.err.name.exists");
          }
          else if (groupName && !validator.isValidConfigGroupName(groupName)) {
            warningMessage = Em.I18n.t("form.validator.configGroupName");
          }
        }
        this.set('warningMessage', warningMessage);
      }.observes('configGroupName', 'configGroupDesc'),
      disablePrimary: function () {
        return !(this.get('configGroupName').trim().length > 0 && (this.get('warningMessage') !== null && !this.get('warningMessage')));
      }.property('warningMessage', 'configGroupName', 'configGroupDesc'),
      onPrimary: function () {
        self.set('selectedConfigGroup.name', this.get('configGroupName'));
        self.set('selectedConfigGroup.description', this.get('configGroupDesc'));
        self.get('selectedConfigGroup.properties').forEach(function(property){
          property.set('group', self.get('selectedConfigGroup'));
        });
        this.hide();
      }
    });
  },

  /**
   * add new config group
   */
  addConfigGroup: function (duplicated) {
    duplicated = (duplicated === true);
    var self = this;
    this.addGroupPopup = App.ModalPopup.show({
      primary: Em.I18n.t('ok'),
      secondary: Em.I18n.t('common.cancel'),
      header: Em.I18n.t('services.service.config_groups.add_config_group_popup.header'),
      bodyClass: Ember.View.extend({
        templateName: require('templates/main/service/new_config_group')
      }),
      configGroupName: duplicated ? self.get('selectedConfigGroup.name') + ' Copy' : "",
      configGroupDesc: duplicated ? self.get('selectedConfigGroup.description') + ' (Copy)' : "",
      warningMessage: '',
      didInsertElement: function(){
        this.validate();
        this.$('input').focus();
      },
      validate: function () {
        var warningMessage = '';
        var groupName = this.get('configGroupName').trim();
        if (self.get('configGroups').mapProperty('name').contains(groupName)) {
          warningMessage = Em.I18n.t("config.group.selection.dialog.err.name.exists");
        }
        else if (groupName && !validator.isValidConfigGroupName(groupName)) {
          warningMessage = Em.I18n.t("form.validator.configGroupName");
        }
        this.set('warningMessage', warningMessage);
      }.observes('configGroupName'),
      disablePrimary: function () {
        return !(this.get('configGroupName').trim().length > 0 && !this.get('warningMessage'));
      }.property('warningMessage', 'configGroupName'),
      onPrimary: function () {
        var defaultConfigGroup = self.get('configGroups').findProperty('isDefault');
        var properties = [];
        var newConfigGroupData = App.ConfigGroup.create({
          id: null,
          name: this.get('configGroupName').trim(),
          description: this.get('configGroupDesc'),
          isDefault: false,
          parentConfigGroup: defaultConfigGroup,
          service: Em.Object.create({id: self.get('serviceName')}),
          hosts: [],
          configSiteTags: [],
          properties: []
        });
        if (duplicated) {
          self.get('selectedConfigGroup.properties').forEach(function(property) {
            var property = App.ServiceConfigProperty.create($.extend(false, {}, property));
            property.set('group', newConfigGroupData);
            properties.push(property);
          });
          newConfigGroupData.set('properties', properties);
        } else {
          newConfigGroupData.set('properties', []);
        }
        self.get('configGroups').pushObject(newConfigGroupData);
        defaultConfigGroup.get('childConfigGroups').pushObject(newConfigGroupData);
        this.hide();
      }
    });
  },

  duplicateConfigGroup: function() {
    this.addConfigGroup(true);
  },

  hostsModifiedConfigGroups: function () {
    if (!this.get('isLoaded')) {
      return false;
    }
    var groupsToClearHosts = [];
    var groupsToDelete = [];
    var groupsToSetHosts = [];
    var groupsToCreate = [];
    var groups = this.get('configGroups');
    var originalGroups = this.get('originalConfigGroups');
    // remove default group
    originalGroups = originalGroups.without(originalGroups.findProperty('isDefault'));
    var originalGroupsIds = originalGroups.mapProperty('id');
    groups.forEach(function (group) {
      if (!group.get('isDefault')) {
        var originalGroup = originalGroups.findProperty('id', group.get('id'));
        if (originalGroup) {
          if (!(JSON.stringify(group.get('hosts').slice().sort()) === JSON.stringify(originalGroup.get('hosts').sort()))) {
            groupsToClearHosts.push(group.set('id', originalGroup.get('id')));
            if (group.get('hosts').length) {
              groupsToSetHosts.push(group.set('id', originalGroup.get('id')));
            }
          // should update name or description
          } else if (group.get('description') !== originalGroup.get('description') || group.get('name') !== originalGroup.get('name') ) {
            groupsToSetHosts.push(group.set('id', originalGroup.get('id')));
          }
          originalGroupsIds = originalGroupsIds.without(group.get('id'));
        } else {
          groupsToCreate.push(group);
        }
      }
    });
    originalGroupsIds.forEach(function (id) {
      groupsToDelete.push(originalGroups.findProperty('id', id));
    }, this);
    return {
      toClearHosts: groupsToClearHosts,
      toDelete: groupsToDelete,
      toSetHosts: groupsToSetHosts,
      toCreate: groupsToCreate
    };
  }.property('selectedConfigGroup.hosts.@each', 'selectedConfigGroup.hosts.length', 'selectedConfigGroup.description', 'configGroups', 'isLoaded'),

  isHostsModified: function () {
    var modifiedGroups = this.get('hostsModifiedConfigGroups');
    if (!this.get('isLoaded')) {
      return false;
    }
    return !!(modifiedGroups.toClearHosts.length || modifiedGroups.toSetHosts.length || modifiedGroups.toCreate.length || modifiedGroups.toDelete.length);
  }.property('hostsModifiedConfigGroups'),

  /**
   * copy config groups to manage popup to give user choice whether or not save changes
   * @param originGroups
   * @return {Array}
   */
  copyConfigGroups: function (originGroups) {
    var configGroups = [];
    var result = [];
    var defaultConfigGroup = App.ConfigGroup.create($.extend(true, {}, originGroups.findProperty('isDefault')));
    originGroups.forEach(function (configGroup) {
      if (!configGroup.get('isDefault')) {
        var copiedGroup = App.ConfigGroup.create($.extend(true, {}, configGroup));
        copiedGroup.set('parentConfigGroup', defaultConfigGroup);
        configGroups.pushObject(copiedGroup);
      }
    });
    defaultConfigGroup.set('childConfigGroups', configGroups.slice());
    configGroups.pushObject(defaultConfigGroup);
    configGroups.forEach(function (group) {
      var groupCopy = {};
      for (var prop in group) {
        if (group.hasOwnProperty(prop)) {
          groupCopy[prop] = group[prop];
        }
      }
      groupCopy.properties.forEach(function(property){
        property.set('group', group);
      });
      result.push(App.ConfigGroup.create(groupCopy));
    }, this);
    return result;
  }
});

});

require.register("controllers/main/service/reassign/step1_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ReassignMasterWizardStep1Controller = Em.Controller.extend({
  name: 'reassignMasterWizardStep1Controller',
  databaseType: null,

  dbProperty: function() {
    var componentName = this.get('content.reassign.component_name');

    var property = null;
    switch(componentName) {
      case 'HIVE_SERVER':
      case 'HIVE_METASTORE':
        property = 'javax.jdo.option.ConnectionDriverName';
        break;
      case 'OOZIE_SERVER':
        property = 'oozie.service.JPAService.jdbc.driver';
        break;
    }

    return property;
  },

  loadConfigsTags: function () {
      App.ajax.send({
        name: 'config.tags',
        sender: this,
        success: 'onLoadConfigsTags',
        error: ''
      });
  },

  /**
   * construct URL parameters for config call
   * @param componentName
   * @param data
   * @return {Array}
   */
  getConfigUrlParams: function (componentName, data) {
    var urlParams = [];
    switch (componentName) {
      case 'OOZIE_SERVER':
        urlParams.push('(type=oozie-site&tag=' + data.Clusters.desired_configs['oozie-site'].tag + ')');
        urlParams.push('(type=oozie-env&tag=' + data.Clusters.desired_configs['oozie-env'].tag + ')');
        break;
      case 'HIVE_SERVER':
      case 'HIVE_METASTORE':
        urlParams.push('(type=hive-site&tag=' + data.Clusters.desired_configs['hive-site'].tag + ')');
        urlParams.push('(type=hive-env&tag=' + data.Clusters.desired_configs['hive-env'].tag + ')');
        break;
    }
    return urlParams;
  },

  onLoadConfigsTags: function (data) {
    var urlParams = this.getConfigUrlParams(this.get('content.reassign.component_name'), data);

    if (urlParams.length > 0) {
      App.ajax.send({
        name: 'reassign.load_configs',
        sender: this,
        data: {
          urlParams: urlParams.join('|')
        },
        success: 'onLoadConfigs',
        error: ''
      });
    }
  },

  onLoadConfigs: function (data) {
    var databaseProperty = null,
        databaseType = null,
        properties = {},
        isRemoteDB = null;

    data.items.forEach(function(item) {
      $.extend(properties, item.properties);
    });

    this.set('content.serviceProperties', properties);

    databaseProperty = properties[ this.dbProperty() ];
    databaseType = databaseProperty.match(/MySQL|PostgreS|Oracle|Derby|MSSQL/gi)[0];
    this.set('databaseType', databaseType);

    if (this.get('content.reassign.component_name') == 'OOZIE_SERVER' && databaseType !== 'derby') {
      App.router.reassignMasterController.set('content.hasManualSteps', false);
    }

    var serviceDbProp = this.get('content.reassign.service_id').toLowerCase() + "_database";
    properties['is_remote_db'] = /Existing/ig.test( properties[serviceDbProp] );

    properties['database_hostname'] = this.getDatabaseHost();

    this.saveDatabaseType(databaseType);
    this.saveServiceProperties(properties);
  },

  saveDatabaseType: function(type) {
    if(type) {
      App.router.get(this.get('content.controllerName')).saveDatabaseType(type);
    }
  },

  saveServiceProperties: function(properties) {
    if(properties) {
      App.router.get(this.get('content.controllerName')).saveServiceProperties(properties);
    }
  },

  getDatabaseHost: function() {
    var db_type = this.get('databaseType');
    var connectionURLPRops = {
      'HIVE': 'javax.jdo.option.ConnectionURL',
      'OOZIE': 'oozie.service.JPAService.jdbc.url'
    };

    var service = this.get('content.reassign.service_id');
    var connectionURL = this.get('content.serviceProperties')[connectionURLPRops[service]];

    connectionURL = connectionURL.replace("jdbc:" + db_type + "://", "");
    connectionURL = connectionURL.replace("/hive?createDatabaseIfNotExist=true", "");
    connectionURL = connectionURL.replace("/oozie", "");

    return connectionURL;
  }
});

});

require.register("controllers/main/service/reassign/step2_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ReassignMasterWizardStep2Controller = App.WizardStep5Controller.extend({

  currentHostId: null,
  showCurrentHost: true,
  useServerValidation: false,

  loadStep: function() {
    // If High Availability is enabled NameNode became a multiple component
    if (App.get('isHaEnabled')) {
      this.get('multipleComponents').push('NAMENODE');
    }
    this.clearStep();
    this.renderHostInfo();
    this.loadStepCallback(this.loadComponents(), this);

    // if moving NameNode with HA enabled
    if (this.get('content.reassign.component_name') === "NAMENODE" && App.get('isHaEnabled')) {
      this.set('showCurrentHost', false);
      this.set('componentToRebalance', 'NAMENODE');
      this.incrementProperty('rebalanceComponentHostsCounter');

    // if moving ResourceManager with HA enabled
    } else if (this.get('content.reassign.component_name') === "RESOURCEMANAGER" && App.get('isRMHaEnabled')) {
      this.set('showCurrentHost', false);
      this.set('componentToRebalance', 'RESOURCEMANAGER');
      this.incrementProperty('rebalanceComponentHostsCounter');
    } else {
      this.set('showCurrentHost', true);
      this.rebalanceSingleComponentHosts(this.get('content.reassign.component_name'));
    }
  },

  /**
   * load master components
   * @return {Array}
   */
  loadComponents: function () {
    var masterComponents = this.get('content.masterComponentHosts');
    this.set('currentHostId', this.get('content').get('reassign').host_id);
    var componentNameToReassign = this.get('content').get('reassign').component_name;
    var result = [];

    masterComponents.forEach(function (master) {
      var color = "grey";
      if (master.component == componentNameToReassign) {
        color = 'green';
      }
      result.push({
        component_name: master.component,
        display_name: App.format.role(master.component),
        selectedHost: master.hostName,
        isInstalled: true,
        serviceId: App.HostComponent.find().findProperty('componentName', master.component).get('serviceName'),
        isServiceCoHost: ['HIVE_METASTORE', 'WEBHCAT_SERVER'].contains(master.component) && !this.get('this.isReassignWizard'),
        color: color
      });
    }, this);
    return result;
  },

  /**
   * rebalance single component among available hosts
   * @param componentName
   * @return {Boolean}
   */
  rebalanceSingleComponentHosts: function (componentName) {
    var currentComponents = this.get("selectedServicesMasters").filterProperty("component_name", componentName),
      availableComponentHosts = [];

    this.get("hosts").forEach(function (item) {
      if ( (this.get('currentHostId') !== item.get('host_name')) && !(currentComponents.mapProperty("selectedHost").contains(item.get('host_name')))) {
        availableComponentHosts.pushObject(item);
      }
    }, this);

    if (availableComponentHosts.length > 0) {
      var preparedAvailableHosts = availableComponentHosts.slice(0);

      currentComponents.forEach(function (item) {
        if (item.get('selectedHost') === this.get('currentHostId') && item.get('component_name') === componentName) {
          item.set('selectedHost', preparedAvailableHosts.objectAt(0).host_name);
        }
        item.set("availableHosts", preparedAvailableHosts.sortProperty('host_name'));
      }, this);
      return true;
    }
    return false;
  },

  updateIsSubmitDisabled: function () {
    var isSubmitDisabled = this._super();
    if (!isSubmitDisabled) {
      var reassigned = 0;
      var existedComponents = App.HostComponent.find().filterProperty('componentName', this.get('content.reassign.component_name')).mapProperty('hostName');
      var newComponents = this.get('servicesMasters').mapProperty('selectedHost');
      existedComponents.forEach(function (host) {
        if (!newComponents.contains(host)) {
          reassigned++;
        }
      }, this);
      isSubmitDisabled = reassigned !== 1;
    }
    this.set('submitDisabled', isSubmitDisabled);
    return isSubmitDisabled;
  }.observes('servicesMasters.@each.selectedHost', 'servicesMasters.@each.isHostNameValid')
});


});

require.register("controllers/main/service/reassign/step3_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ReassignMasterWizardStep3Controller = Em.Controller.extend({
  name: 'reassignMasterWizardStep3Controller',

  submit: function() {
    App.get('router.mainAdminKerberosController').getKDCSessionState(function() {
      App.router.send("next");
    });
  }
});

});

require.register("controllers/main/service/reassign/step4_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ReassignMasterWizardStep4Controller = App.HighAvailabilityProgressPageController.extend({

  isReassign: true,

  commands: ['stopServices', 'cleanMySqlServer', 'createHostComponents', 'putHostComponentsInMaintenanceMode', 'reconfigure', 'installHostComponents', 'startZooKeeperServers', 'startNameNode', 'deleteHostComponents', 'configureMySqlServer',
  'startMySqlServer', 'startNewMySqlServer' , 'startServices'],
  // custom commands for Components with DB Configuration and Check
  commandsForDB: ['createHostComponents', 'installHostComponents', 'configureMySqlServer', 'startMySqlServer', 'testDBConnection', 'stopServices', 'cleanMySqlServer', 'putHostComponentsInMaintenanceMode', 'reconfigure', 'deleteHostComponents', 'configureMySqlServer', 'startServices'],

  clusterDeployState: 'REASSIGN_MASTER_INSTALLING',

  multiTaskCounter: 0,

  hostComponents: [],

  /**
   * Map with lists of unrelated services.
   * Used to define list of services to stop/start.
   */
  unrelatedServicesMap: {
    'JOBTRACKER': ['HDFS', 'ZOOKEEPER', 'HBASE', 'FLUME', 'SQOOP', 'STORM'],
    'RESOURCEMANAGER': ['HDFS', 'ZOOKEEPER', 'HBASE', 'FLUME', 'SQOOP', 'STORM'],
    'APP_TIMELINE_SERVER': ['HDFS', 'ZOOKEEPER', 'HBASE', 'FLUME', 'SQOOP', 'STORM'],
    'OOZIE_SERVER': ['HDFS', 'ZOOKEEPER', 'HBASE', 'FLUME', 'SQOOP', 'STORM', 'HIVE'],
    'WEBHCAT_SERVER': ['HDFS', 'ZOOKEEPER', 'HBASE', 'FLUME', 'SQOOP', 'STORM'],
    'HIVE_SERVER': ['HDFS', 'ZOOKEEPER', 'HBASE', 'FLUME', 'SQOOP', 'STORM'],
    'HIVE_METASTORE': ['HDFS', 'ZOOKEEPER', 'HBASE', 'FLUME', 'SQOOP', 'STORM'],
    'MYSQL_SERVER': ['HDFS', 'ZOOKEEPER', 'HBASE', 'FLUME', 'SQOOP', 'STORM']
  },

  /**
   * additional configs with template values
   * Part of value to substitute has following format: "<replace-value>"
   */
  additionalConfigsMap: [
    {
      componentName: 'RESOURCEMANAGER',
      configs: {
        'yarn-site': {
          'yarn.resourcemanager.address': '<replace-value>:8050',
          'yarn.resourcemanager.admin.address': '<replace-value>:8141',
          'yarn.resourcemanager.resource-tracker.address': '<replace-value>:8025',
          'yarn.resourcemanager.scheduler.address': '<replace-value>:8030',
          'yarn.resourcemanager.webapp.address': '<replace-value>:8088',
          'yarn.resourcemanager.hostname': '<replace-value>'
        }
      }
    },
    {
      componentName: 'JOBTRACKER',
      configs: {
        'mapred-site': {
          'mapred.job.tracker.http.address': '<replace-value>:50030',
          'mapred.job.tracker': '<replace-value>:50300'
        }
      }
    },
    {
      componentName: 'SECONDARY_NAMENODE',
      configs: {
        'hdfs-site': {
          'dfs.secondary.http.address': '<replace-value>:50090'
        }
      },
      configs_Hadoop2: {
        'hdfs-site': {
          'dfs.namenode.secondary.http-address': '<replace-value>:50090'
        }
      }
    },
    {
      componentName: 'NAMENODE',
      configs: {
        'hdfs-site': {
          'dfs.http.address': '<replace-value>:50070',
          'dfs.https.address': '<replace-value>:50470'
        },
        'core-site': {
          'fs.default.name': 'hdfs://<replace-value>:8020'
        }
      },
      configs_Hadoop2: {
        'hdfs-site': {
          'dfs.namenode.http-address': '<replace-value>:50070',
          'dfs.namenode.https-address': '<replace-value>:50470'
        },
        'core-site': {
          'fs.defaultFS': 'hdfs://<replace-value>:8020'
        }
      }
    },
    {
      componentName: 'APP_TIMELINE_SERVER',
      configs: {
        'yarn-site': {
          'yarn.timeline-service.webapp.address': '<replace-value>:8188',
          'yarn.timeline-service.webapp.https.address': '<replace-value>:8190',
          'yarn.timeline-service.address': '<replace-value>:10200'
        }
      }
    },
    {
      componentName: 'OOZIE_SERVER',
        configs: {
          'oozie-site': {
            'oozie.base.url': 'http://<replace-value>:11000/oozie'
          },
          'core-site': {
            'hadoop.proxyuser.oozie.hosts': '<replace-value>'
          }
        }
    },
    {
      componentName: 'HIVE_METASTORE',
      configs: {
        'hive-site': {}
      }
    },
    {
      componentName: 'MYSQL_SERVER',
      configs: {
        'hive-site': {
          'javax.jdo.option.ConnectionURL': 'jdbc:mysql://<replace-value>/hive?createDatabaseIfNotExist=true'
        }
      }
    }
  ],

  secureConfigsMap: [
    {
      componentName: 'NAMENODE',
      configs: [
        {
          site: 'hdfs-site',
          keytab: 'dfs.namenode.keytab.file',
          principal: 'dfs.namenode.kerberos.principal'
        },
        {
          site: 'hdfs-site',
          keytab: 'dfs.web.authentication.kerberos.keytab',
          principal: 'dfs.web.authentication.kerberos.principal'
        }
      ]
    },
    {
      componentName: 'SECONDARY_NAMENODE',
      configs: [
        {
          site: 'hdfs-site',
          keytab: 'dfs.secondary.namenode.keytab.file',
          principal: 'dfs.secondary.namenode.kerberos.principal'
        },
        {
          site: 'hdfs-site',
          keytab: 'dfs.web.authentication.kerberos.keytab',
          principal: 'dfs.web.authentication.kerberos.principal'
        }
      ]
    },
    {
      componentName: 'RESOURCEMANAGER',
      configs: [
        {
          site: 'yarn-site',
          keytab: 'yarn.resourcemanager.keytab',
          principal: 'yarn.resourcemanager.principal'
        },
        {
          site: 'yarn-site',
          keytab: 'yarn.resourcemanager.webapp.spnego-keytab-file',
          principal: 'yarn.resourcemanager.webapp.spnego-principal'
        }
      ]
    },
    {
      componentName: 'OOZIE_SERVER',
      configs: [
        {
          site: 'oozie-site',
          keytab: 'oozie.authentication.kerberos.keytab',
          principal: 'oozie.authentication.kerberos.principal'
        },
        {
          site: 'oozie-site',
          keytab: 'oozie.service.HadoopAccessorService.keytab.file',
          principal: 'oozie.service.HadoopAccessorService.kerberos.principal'
        }
      ]
    },
    {
      componentName: 'WEBHCAT_SERVER',
      configs: [
        {
          site: 'webhcat-site',
          keytab: 'templeton.kerberos.keytab',
          principal: 'templeton.kerberos.principal'
        }
      ]
    },
    {
      componentName: 'HIVE_SERVER',
      configs: [
        {
          site: 'hive-site',
          keytab: 'hive.server2.authentication.kerberos.keytab',
          principal: 'hive.server2.authentication.kerberos.principal'
        },
        {
          site: 'hive-site',
          keytab: 'hive.server2.authentication.spnego.keytab',
          principal: 'hive.server2.authentication.spnego.principal'
        }
      ]
    },
    {
      componentName: 'HIVE_METASTORE',
      configs: [
        {
          site: 'hive-site',
          keytab: 'hive.metastore.kerberos.keytab.file',
          principal: 'hive.metastore.kerberos.principal'
        }
      ]
    }

  ],

  /**
   * set additional configs
   * configs_Hadoop2 - configs which belongs to Hadoop 2 stack only
   * @param configs
   * @param componentName
   * @param replaceValue
   * @return {Boolean}
   */
  setAdditionalConfigs: function (configs, componentName, replaceValue) {
    var component = this.get('additionalConfigsMap').findProperty('componentName', componentName);

    if (Em.isNone(component)) return false;
    var additionalConfigs = (component.configs_Hadoop2) ? component.configs_Hadoop2 : component.configs;

    for (var site in additionalConfigs) {
      if (additionalConfigs.hasOwnProperty(site)) {
        for (var property in additionalConfigs[site]) {
          if (additionalConfigs[site].hasOwnProperty(property)) {
            if (App.get('isHaEnabled') && componentName === 'NAMENODE' && property === 'fs.defaultFS') continue;

            configs[site][property] = additionalConfigs[site][property].replace('<replace-value>', replaceValue);
          }
        }
      }
    }
    return true;
  },

  /**
   * load step info
   */
  loadStep: function () {
    if (this.get('content.reassign.component_name') === 'NAMENODE' && App.get('isHaEnabled')) {
      this.set('hostComponents', ['NAMENODE', 'ZKFC']);
    } else {
      this.set('hostComponents', [this.get('content.reassign.component_name')]);
    }
    this.set('serviceName', [this.get('content.reassign.service_id')]);
    this._super();
  },

  /**
   * concat host-component names into string
   * @return {String}
   */
  getHostComponentsNames: function () {
    var hostComponentsNames = '';
    this.get('hostComponents').forEach(function (comp, index) {
      hostComponentsNames += index ? '+' : '';
      hostComponentsNames += comp === 'ZKFC' ? comp : App.format.role(comp);
    }, this);
    return hostComponentsNames;
  },

  /**
   * remove unneeded tasks
   */
  removeUnneededTasks: function () {
    if(this.isComponentWithDB()) {
      var db_type = this.get('content.databaseType');
      var is_remote_db = this.get('content.serviceProperties.is_remote_db');


      if(is_remote_db || db_type !== 'mysql') {
        this.removeTasks(['configureMySqlServer', 'startMySqlServer', 'cleanMySqlServer', 'configureMySqlServer']);
      }

      if (db_type === 'derby') {
        this.removeTasks(['testDBConnection']);
      }
    }

    if ( this.get('content.reassign.component_name') !== 'MYSQL_SERVER' && !this.isComponentWithDB()) {
      this.removeTasks(['configureMySqlServer', 'startMySqlServer', 'cleanMySqlServer', 'startNewMySqlServer', 'configureMySqlServer']);
    }

    if ( this.get('content.reassign.component_name') === 'MYSQL_SERVER' ) {
      this.removeTasks(['cleanMySqlServer']);
    }

    if (this.get('content.hasManualSteps')) {
      if (this.get('content.reassign.component_name') === 'NAMENODE' && App.get('isHaEnabled')) {
        // Only for reassign NameNode with HA enabled
        this.removeTasks(['deleteHostComponents', 'startServices']);
      } else {
        this.removeTasks(['startZooKeeperServers', 'startNameNode', 'deleteHostComponents', 'startServices']);
      }
    } else {
      this.removeTasks(['startZooKeeperServers', 'startNameNode']);
    }
  },

  /**
   * remove tasks by command name
   */
  removeTasks: function(commands) {
    var tasks = this.get('tasks');

    commands.forEach(function(command) {
      var cmd = tasks.filterProperty('command', command);
      var index = null;

      if (cmd.length === 0) {
        return false;
      } else {
        index = tasks.indexOf( cmd[0] );
      }

      tasks.splice( index, 1 );
    });
  },

  /**
   * initialize tasks
   */
  initializeTasks: function () {
    var commands = this.get('commands');
    var currentStep = App.router.get('reassignMasterController.currentStep');
    var hostComponentsNames = this.getHostComponentsNames();

    if (this.isComponentWithDB()) {
      commands = this.get('commandsForDB');
    }

    for (var i = 0; i < commands.length; i++) {
      var TaskLabel = i === 3 ? this.get('serviceName') : hostComponentsNames; //For Reconfigure task, show serviceName
      var title = Em.I18n.t('services.reassign.step4.tasks.' + commands[i] + '.title').format(TaskLabel);
      this.get('tasks').pushObject(Ember.Object.create({
        title: title,
        status: 'PENDING',
        id: i,
        command: commands[i],
        showRetry: false,
        showRollback: false,
        name: title,
        displayName: title,
        progress: 0,
        isRunning: false,
        hosts: []
      }));
    }
    this.removeUnneededTasks();
    this.set('isLoaded', true);
  },

  hideRollbackButton: function () {
    var failedTask = this.get('tasks').findProperty('showRollback');
    if (failedTask) {
      failedTask.set('showRollback', false);
    }
  }.observes('tasks.@each.showRollback'),

  onComponentsTasksSuccess: function () {
    this.incrementProperty('multiTaskCounter');
    if (this.get('multiTaskCounter') >= this.get('hostComponents').length) {
      this.onTaskCompleted();
    }
  },

  /**
   * compute data for call to stop services
   * @return {Object}
   */
  getStopServicesData: function () {
    var data = {
      "ServiceInfo": {
        "state": "INSTALLED"
      }
    };
    var unrelatedServices = this.get('unrelatedServicesMap')[this.get('content.reassign.component_name')];
    if (unrelatedServices) {
      var list = App.Service.find().mapProperty("serviceName").filter(function (s) {
        return !unrelatedServices.contains(s)
      }).join(',');
      data.context = "Stop required services";
      data.urlParams = "ServiceInfo/service_name.in(" + list + ")";
    } else {
      data.context = "Stop all services";
    }
    return data;
  },

  /**
   * make server call to stop services
   */
  stopServices: function () {
    App.ajax.send({
      name: 'common.services.update',
      sender: this,
      data: this.getStopServicesData(),
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  createHostComponents: function () {
    this.set('multiTaskCounter', 0);
    var hostComponents = this.get('hostComponents');
    var hostName = this.get('content.reassignHosts.target');
    for (var i = 0; i < hostComponents.length; i++) {
      this.createComponent(hostComponents[i], hostName, this.get('content.reassign.service_id'));
    }
  },

  onCreateComponent: function () {
    this.onComponentsTasksSuccess();
  },

  putHostComponentsInMaintenanceMode: function () {
    this.set('multiTaskCounter', 0);
    var hostComponents = this.get('hostComponents');
    var hostName = this.get('content.reassignHosts.source');
    for (var i = 0; i < hostComponents.length; i++) {
      App.ajax.send({
        name: 'common.host.host_component.passive',
        sender: this,
        data: {
          hostName: hostName,
          passive_state: "ON",
          componentName: hostComponents[i]
        },
        success: 'onComponentsTasksSuccess',
        error: 'onTaskError'
      });
    }
  },

  installHostComponents: function () {
    this.set('multiTaskCounter', 0);
    var hostComponents = this.get('hostComponents');
    var hostName = this.get('content.reassignHosts.target');
    for (var i = 0; i < hostComponents.length; i++) {
      this.updateComponent(hostComponents[i], hostName, this.get('content.reassign.service_id'), "Install", hostComponents.length);
    }
  },

  reconfigure: function () {
    this.loadConfigsTags();
  },

  loadConfigsTags: function () {
    App.ajax.send({
      name: 'config.tags',
      sender: this,
      success: 'onLoadConfigsTags',
      error: 'onTaskError'
    });
  },

  /**
   * construct URL parameters for config call
   * @param componentName
   * @param data
   * @return {Array}
   */
  getConfigUrlParams: function (componentName, data) {
    var urlParams = [];
    switch (componentName) {
      case 'NAMENODE':
        urlParams.push('(type=hdfs-site&tag=' + data.Clusters.desired_configs['hdfs-site'].tag + ')');
        urlParams.push('(type=core-site&tag=' + data.Clusters.desired_configs['core-site'].tag + ')');
        if (App.Service.find().someProperty('serviceName', 'HBASE')) {
          urlParams.push('(type=hbase-site&tag=' + data.Clusters.desired_configs['hbase-site'].tag + ')');
        }
        break;
      case 'SECONDARY_NAMENODE':
        urlParams.push('(type=hdfs-site&tag=' + data.Clusters.desired_configs['hdfs-site'].tag + ')');
        urlParams.push('(type=core-site&tag=' + data.Clusters.desired_configs['core-site'].tag + ')');
        break;
      case 'JOBTRACKER':
        urlParams.push('(type=mapred-site&tag=' + data.Clusters.desired_configs['mapred-site'].tag + ')');
        break;
      case 'RESOURCEMANAGER':
        urlParams.push('(type=yarn-site&tag=' + data.Clusters.desired_configs['yarn-site'].tag + ')');
        break;
      case 'WEBHCAT_SERVER':
        urlParams.push('(type=webhcat-site&tag=' + data.Clusters.desired_configs['webhcat-site'].tag + ')');
        break;
      case 'APP_TIMELINE_SERVER':
        urlParams.push('(type=yarn-site&tag=' + data.Clusters.desired_configs['yarn-site'].tag + ')');
        urlParams.push('(type=yarn-env&tag=' + data.Clusters.desired_configs['yarn-env'].tag + ')');
        break;
      case 'OOZIE_SERVER':
        urlParams.push('(type=oozie-site&tag=' + data.Clusters.desired_configs['oozie-site'].tag + ')');
        urlParams.push('(type=core-site&tag=' + data.Clusters.desired_configs['core-site'].tag + ')');
        break;
      case 'HIVE_SERVER':
      case 'HIVE_METASTORE':
        urlParams.push('(type=hive-site&tag=' + data.Clusters.desired_configs['hive-site'].tag + ')');
        urlParams.push('(type=webhcat-site&tag=' + data.Clusters.desired_configs['webhcat-site'].tag + ')');
        urlParams.push('(type=hive-env&tag=' + data.Clusters.desired_configs['hive-env'].tag + ')');
        urlParams.push('(type=core-site&tag=' + data.Clusters.desired_configs['core-site'].tag + ')');
        break;
      case 'MYSQL_SERVER':
        urlParams.push('(type=hive-site&tag=' + data.Clusters.desired_configs['hive-site'].tag + ')');
        break;
    }
    return urlParams;
  },

  onLoadConfigsTags: function (data) {
    var urlParams = this.getConfigUrlParams(this.get('content.reassign.component_name'), data);

    App.ajax.send({
      name: 'reassign.load_configs',
      sender: this,
      data: {
        urlParams: urlParams.join('|')
      },
      success: 'onLoadConfigs',
      error: 'onTaskError'
    });
  },

  onLoadConfigs: function (data) {
    var componentName = this.get('content.reassign.component_name');
    var targetHostName = this.get('content.reassignHosts.target');
    var configs = {};
    var secureConfigs = [];

    data.items.forEach(function (item) {
      configs[item.type] = item.properties;
    }, this);

    this.setAdditionalConfigs(configs, componentName, targetHostName);
    this.setSecureConfigs(secureConfigs, configs, componentName);

    if (componentName === 'NAMENODE') {
      this.setSpecificNamenodeConfigs(configs, targetHostName);
    }

    if (componentName === 'RESOURCEMANAGER') {
      this.setSpecificResourceMangerConfigs(configs, targetHostName);
    }

    if (componentName === 'HIVE_METASTORE' || componentName === 'HIVE_SERVER') {
      this.setSpecificHiveConfigs(configs, targetHostName);
    }

    this.saveClusterStatus(secureConfigs, this.getComponentDir(configs, componentName));
    this.saveConfigsToServer(configs);
    this.saveServiceProperties(configs);
  },

  /**
   * make PUT call to save configs to server
   * @param configs
   */
  saveConfigsToServer: function (configs) {
    App.ajax.send({
      name: 'common.across.services.configurations',
      sender: this,
      data: {
        data: '[' + this.getServiceConfigData(configs).toString() + ']'
      },
      success: 'onSaveConfigs',
      error: 'onTaskError'
    });
  },
  /**
   * gather and format config data before sending to server
   * @param configs
   * @return {Array}
   * @method getServiceConfigData
   */
  getServiceConfigData: function (configs) {
    var componentName = this.get('content.reassign.component_name');
    var tagName = 'version' + (new Date).getTime();
    var configData = Object.keys(configs).map(function (_siteName) {
      return {
        type: _siteName,
        tag: tagName,
        properties: configs[_siteName],
        service_config_version_note: Em.I18n.t('services.reassign.step4.save.configuration.note').format(App.format.role(componentName))
      }
    });
    var allConfigData = [];

    App.Service.find().forEach(function (service) {
      var stackService = App.StackService.find().findProperty('serviceName', service.get('serviceName'));
      if (stackService) {
        var serviceConfigData = [];
        Object.keys(stackService.get('configTypesRendered')).forEach(function (type) {
          var serviceConfigTag = configData.findProperty('type', type);
          if (serviceConfigTag) {
            serviceConfigData.pushObject(serviceConfigTag);
          }
        }, this);
        allConfigData.pushObject(JSON.stringify({
          Clusters: {
            desired_config: serviceConfigData
          }
        }));
      }
    }, this);
    return allConfigData;
  },

  /**
   * set specific configs which applies only to NameNode component
   * @param configs
   * @param targetHostName
   */
  setSpecificNamenodeConfigs: function (configs, targetHostName) {
    var sourceHostName = this.get('content.reassignHosts.source');

    if (App.get('isHaEnabled')) {
      var nameServices = configs['hdfs-site']['dfs.nameservices'];
      var suffix = (configs['hdfs-site']['dfs.namenode.http-address.' + nameServices + '.nn1'] === sourceHostName + ':50070') ? '.nn1' : '.nn2';
      configs['hdfs-site']['dfs.namenode.http-address.' + nameServices + suffix] = targetHostName + ':50070';
      configs['hdfs-site']['dfs.namenode.https-address.' + nameServices + suffix] = targetHostName + ':50470';
      configs['hdfs-site']['dfs.namenode.rpc-address.' + nameServices + suffix] = targetHostName + ':8020';
    }
    if (!App.get('isHaEnabled') && App.Service.find('HBASE').get('isLoaded')) {
      configs['hbase-site']['hbase.rootdir'] = configs['hbase-site']['hbase.rootdir'].replace(/\/\/[^\/]*/, '//' + targetHostName + ':8020');
    }
  },

  /**
   * set specific configs which applies only to ResourceManager component
   * @param configs
   * @param targetHostName
   */
  setSpecificResourceMangerConfigs: function (configs, targetHostName) {
    var sourceHostName = this.get('content.reassignHosts.source');

    if (App.get('isRMHaEnabled')) {
      if (configs['yarn-site']['yarn.resourcemanager.hostname.rm1'] === sourceHostName) {
        configs['yarn-site']['yarn.resourcemanager.hostname.rm1'] = targetHostName;
      } else {
        configs['yarn-site']['yarn.resourcemanager.hostname.rm2'] = targetHostName;
      }
    }
  },

  /**
   * set specific configs which applies only to Hive related configs
   * @param configs
   * @param targetHostName
   */
  setSpecificHiveConfigs: function (configs, targetHostName) {
    var sourceHostName = this.get('content.reassignHosts.source');
    var hiveMSHosts = App.HostComponent.find().filterProperty('componentName', 'HIVE_METASTORE').mapProperty('hostName');
    if (this.get('content.reassign.component_name') === 'HIVE_METASTORE') hiveMSHosts = hiveMSHosts.removeObject(sourceHostName).addObject(targetHostName);
    var hiveServerHosts = App.HostComponent.find().filterProperty('componentName', 'HIVE_SERVER').mapProperty('hostName');
    if (this.get('content.reassign.component_name') === 'HIVE_SERVER') hiveServerHosts = hiveServerHosts.removeObject(sourceHostName).addObject(targetHostName);
    var hiveMasterHosts = hiveMSHosts.concat(hiveServerHosts).uniq().join(',');
    var hiveUser = configs['hive-env']['hive_user'];
    var webhcatUser = configs['hive-env']['webhcat_user'];

    var port = configs['hive-site']['hive.metastore.uris'].match(/:[0-9]{2,4}/);
    port = port ? port[0].slice(1) : "9083";

    for (var i = 0; i < hiveMSHosts.length; i++) {
      hiveMSHosts[i] = "thrift://" + hiveMSHosts[i] + ":" + port;
    }

    configs['hive-site']['hive.metastore.uris'] = hiveMSHosts.join(',');
    configs['webhcat-site']['templeton.hive.properties'] = configs['webhcat-site']['templeton.hive.properties'].replace(/thrift.+[0-9]{2,},/i, hiveMSHosts.join('\\,') + ",");
    configs['core-site']['hadoop.proxyuser.' + hiveUser + '.hosts'] = hiveMasterHosts;
    configs['core-site']['hadoop.proxyuser.' + webhcatUser + '.hosts'] = hiveMasterHosts;
  },

  /**
   * set secure configs for component
   * @param secureConfigs
   * @param configs
   * @param componentName
   * @return {Boolean}
   */
  setSecureConfigs: function (secureConfigs, configs, componentName) {
    var securityEnabled = this.get('content.securityEnabled');
    var component = this.get('secureConfigsMap').findProperty('componentName', componentName);
    if (Em.isNone(component) || !securityEnabled) return false;

    component.configs.forEach(function (config) {
      secureConfigs.push({
        keytab: configs[config.site][config.keytab],
        principal: configs[config.site][config.principal]
      });
    });
    return true;
  },

  /**
   * derive component directory from configurations
   * @param configs
   * @param componentName
   * @return {String}
   */
  getComponentDir: function (configs, componentName) {
    if (componentName === 'NAMENODE') {
      return configs['hdfs-site']['dfs.namenode.name.dir'];
    } else if (componentName === 'SECONDARY_NAMENODE') {
      return configs['hdfs-site']['dfs.namenode.checkpoint.dir'];
    }
    return '';
  },

  /**
   * save cluster status to server
   *
   * @param secureConfigs
   * @param componentDir
   * @return {Boolean}
   */
  saveClusterStatus: function (secureConfigs, componentDir) {
    if (componentDir || secureConfigs.length) {
      App.router.get(this.get('content.controllerName')).saveComponentDir(componentDir);
      App.router.get(this.get('content.controllerName')).saveSecureConfigs(secureConfigs);
      App.clusterStatus.setClusterStatus({
        clusterName: this.get('content.cluster.name'),
        clusterState: this.get('clusterDeployState'),
        wizardControllerName: this.get('content.controllerName'),
        localdb: App.db.data
      });
      return true;
    }
    return false;
  },

  onSaveConfigs: function () {
    this.onTaskCompleted();
  },

  startZooKeeperServers: function () {
    var components = this.get('content.masterComponentHosts').filterProperty('component', 'ZOOKEEPER_SERVER');
    this.updateComponent('ZOOKEEPER_SERVER', components.mapProperty('hostName'), "ZOOKEEPER", "Start");
  },

  startNameNode: function () {
    var components = this.get('content.masterComponentHosts').filterProperty('component', 'NAMENODE');
    this.updateComponent('NAMENODE', components.mapProperty('hostName').without(this.get('content.reassignHosts.source')), "HDFS", "Start");
  },

  /**
   * make server call to start services
   */
  startServices: function () {
    App.ajax.send({
      name: 'common.services.update',
      sender: this,
      data: this.getStartServicesData(),
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  /**
   * compute data for call to start services
   * @return {Object}
   */
  getStartServicesData: function () {
    var unrelatedServices = this.get('unrelatedServicesMap')[this.get('content.reassign.component_name')];
    var data = {};

    if (unrelatedServices) {
      var list = App.Service.find().mapProperty("serviceName").filter(function (s) {
        return !unrelatedServices.contains(s)
      }).join(',');
      data = {
        "context": "Start required services",
        "ServiceInfo": {
          "state": "STARTED"
        },
        "urlParams": "ServiceInfo/service_name.in(" + list + ")"
      };
    } else {
      data = {
        "context": "Start all services",
        "ServiceInfo": {
          "state": "STARTED"
        },
        "urlParams": "params/run_smoke_test=true"
      };
    }
    return data;
  },

  /**
   * make DELETE call for each host component on host
   */
  deleteHostComponents: function () {
    this.set('multiTaskCounter', 0);
    var hostComponents = this.get('hostComponents');
    var hostName = this.get('content.reassignHosts.source');
    for (var i = 0; i < hostComponents.length; i++) {
      App.ajax.send({
        name: 'common.delete.host_component',
        sender: this,
        data: {
          hostName: hostName,
          componentName: hostComponents[i]
        },
        success: 'onComponentsTasksSuccess',
        error: 'onDeleteHostComponentsError'
      });
    }
  },

  onDeleteHostComponentsError: function (error) {
    if (error.responseText.indexOf('org.apache.ambari.server.controller.spi.NoSuchResourceException') !== -1) {
      this.onComponentsTasksSuccess();
    } else {
      this.onTaskError();
    }
  },

  done: function () {
    if (!this.get('isSubmitDisabled')) {
      this.removeObserver('tasks.@each.status', this, 'onTaskStatusChange');
      if (this.get('content.hasManualSteps')) {
        App.router.send('next');
      } else {
        App.router.send('complete');
      }
    }
  },

  /**
   * make server call to clean MYSQL
   */
  cleanMySqlServer: function () {
    var hostname = App.HostComponent.find().filterProperty('componentName', 'MYSQL_SERVER').get('firstObject.hostName');

    if (this.get('content.reassign.component_name') === 'MYSQL_SERVER') {
      hostname = this.get('content.reassignHosts.target');
    }

    App.ajax.send({
      name: 'service.mysql.clean',
      sender: this,
      data: {
        host: hostname
      },
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  /**
   * make server call to configure MYSQL
   */
  configureMySqlServer : function () {
    var hostname = App.HostComponent.find().filterProperty('componentName', 'MYSQL_SERVER').get('firstObject.hostName');

    if (this.get('content.reassign.component_name') === 'MYSQL_SERVER') {
      hostname = this.get('content.reassignHosts.target');
    }

    App.ajax.send({
      name: 'service.mysql.configure',
      sender: this,
      data: {
        host: hostname
      },
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  startMySqlServer: function() {
    App.ajax.send({
      name: 'common.host.host_component.update',
      sender: this,
      data: {
        context: "Start MySQL Server",
        hostName: App.HostComponent.find().filterProperty('componentName', 'MYSQL_SERVER').get('firstObject.hostName'),
        serviceName: "HIVE",
        componentName: "MYSQL_SERVER",
        HostRoles: {
          state: "STARTED"
        }
      },
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  startNewMySqlServer: function() {
    App.ajax.send({
      name: 'common.host.host_component.update',
      sender: this,
      data: {
        context: "Start MySQL Server",
        hostName: this.get('content.reassignHosts.target'),
        serviceName: "HIVE",
        componentName: "MYSQL_SERVER",
        HostRoles: {
          state: "STARTED"
        }
      },
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  testDBConnection: function() {
    this.prepareDBCheckAction();
    // this.onTaskCompleted();
  },

  isComponentWithDB: function() {
    return ['HIVE_SERVER', 'HIVE_METASTORE', 'OOZIE_SERVER'].contains(this.get('content.reassign.component_name'));
  },

  dbProperty: function() {
    var componentName = this.get('content.reassign.component_name');

    var property = null;
    switch(componentName) {
      case 'HIVE_SERVER':
      case 'HIVE_METASTORE':
        property = 'javax.jdo.option.ConnectionDriverName';
        break;
      case 'OOZIE_SERVER':
        property = 'oozie.service.JPAService.jdbc.url';
        break;
    }

    return property;
  }.property(),

  /** @property {Object} propertiesPattern - check pattern according to type of connection properties **/
  propertiesPattern: function() {
    return {
      user_name: /(username|dblogin)$/ig,
      user_passwd: /(dbpassword|password)$/ig,
      db_connection_url: /jdbc\.url|connectionurl/ig,
      driver_class: /ConnectionDriverName|jdbc\.driver/ig,
      schema_name: /db\.schema\.name/ig
    };
  }.property(),

  /** @property {Object} connectionProperties - service specific config values mapped for custom action request **/
  connectionProperties: function() {
    var propObj = {};
    for (var key in this.get('propertiesPattern')) {
      propObj[key] = this.getConnectionProperty(this.get('propertiesPattern')[key]);
    }
    return propObj;
  }.property('propertiesPattern'),

  getConnectionProperty: function(regexp) {
    var propertyName = this.get('requiredProperties').filter(function(item) {
      return regexp.test(item);
    })[0];
    return this.get('content.serviceProperties')[propertyName];
  },

  /**
   * Properties that stores in local storage used for handling
   * last success connection.
   *
   * @property {Object} preparedDBProperties
   **/
  preparedDBProperties: function() {
    var propObj = {};
    for (var key in this.get('propertiesPattern')) {
      var propValue = this.getConnectionProperty(this.get('propertiesPattern')[key]);
      propObj[key] = propValue;
    }
    return propObj;
  }.property(),

  /** @property {object} requiredProperties - properties that necessary for database connection **/
  requiredProperties: function() {
    var propertiesMap = {
      OOZIE: ['oozie.db.schema.name','oozie.service.JPAService.jdbc.username','oozie.service.JPAService.jdbc.password','oozie.service.JPAService.jdbc.driver','oozie.service.JPAService.jdbc.url'],
      HIVE: ['ambari.hive.db.schema.name','javax.jdo.option.ConnectionUserName','javax.jdo.option.ConnectionPassword','javax.jdo.option.ConnectionDriverName','javax.jdo.option.ConnectionURL']
    };

    return propertiesMap[this.get('content.reassign.service_id')];
  }.property(),

  dbType: function() {
    var databaseTypes = /MySQL|PostgreS|Oracle|Derby|MSSQL/gi;
    var databaseProp = this.get('content.serviceProperties')[this.get('dbProperty')];

    return databaseProp.match(databaseTypes)[0];
  }.property('dbProperty'),

  prepareDBCheckAction: function() {
    var ambariProperties = null;
    var properties = this.get('content.serviceProperties');
    var params = this.get('preparedDBProperties');

    ambariProperties = App.router.get('clusterController.ambariProperties');

    params['db_name'] = this.get('dbType');
    params['jdk_location'] = ambariProperties['jdk_location'];
    params['jdk_name'] = ambariProperties['jdk.name'];
    params['java_home'] = ambariProperties['java.home'];

    params['threshold'] = 60;
    params['ambari_server_host'] = location.hostname;
    params['check_execute_list'] = "db_connection_check";

    App.ajax.send({
      name: 'custom_action.create',
      sender: this,
      data: {
        requestInfo: {
          "context": "Check host",
          "action": "check_host",
          "parameters": params
        },
        filteredHosts: [this.get('content.reassignHosts.target')]
      },
      success: 'onCreateActionSuccess',
      error: 'onTaskError'
    });
  },

  onCreateActionSuccess: function(data) {
    this.set('checkDBRequestId', data.Requests.id);
    App.ajax.send({
      name: 'custom_action.request',
      sender: this,
      data: {
        requestId: this.get('checkDBRequestId')
      },
      success: 'setCheckDBTaskId'
    });
  },

  setCheckDBTaskId: function(data) {
    this.set('checkDBTaskId', data.items[0].Tasks.id);
    this.startDBCheckPolling();
  },

  startDBCheckPolling: function() {
      this.getDBConnTaskInfo();
  },

  getDBConnTaskInfo: function() {
    this.setTaskStatus(this.get('currentTaskId'), 'IN_PROGRESS');
    this.get('tasks').findProperty('id', this.get('currentTaskId')).set('progress', 100);

    this.set('logs', []);
    App.ajax.send({
      name: 'custom_action.request',
      sender: this,
      data: {
        requestId: this.get('checkDBRequestId'),
        taskId: this.get('checkDBTaskId')
      },
      success: 'getDBConnTaskInfoSuccess'
    });
  },

  getDBConnTaskInfoSuccess: function(data) {
    var task = data.Tasks;
    if (task.status === 'COMPLETED') {
      var structuredOut = task.structured_out.db_connection_check;
      if (structuredOut.exit_code != 0) {
        this.showConnectionErrorPopup(structuredOut.message);
        this.onTaskError();
      } else {
        this.onTaskCompleted();
      }
    }

    if (task.status === 'FAILED') {
      this.onTaskError();
    }

    if (/PENDING|QUEUED|IN_PROGRESS/.test(task.status)) {
      Em.run.later(this, function() {
        this.startDBCheckPolling();
      }, 3000);
    }
  },

  showConnectionErrorPopup: function(error) {
    var popup = App.showAlertPopup('Database Connection Error');
    popup.set('body', error);
  },

  testDBRetryTooltip: function() {
    var db_host = this.get('content.serviceProperties.database_hostname');
    var db_type = this.get('dbType');
    var db_props = this.get('preparedDBProperties');

    return Em.I18n.t('services.reassign.step4.tasks.testDBConnection.tooltip').format(
      db_host, db_type, db_props['schema_name'], db_props['user_name'],
      db_props['user_passwd'], db_props['driver_class'], db_props['db_connection_url']
    );
  }.property('dbProperties'),

  saveServiceProperties: function(configs) {
    App.router.get(this.get('content.controllerName')).saveServiceProperties(configs);
  }

});

});

require.register("controllers/main/service/reassign/step5_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ReassignMasterWizardStep5Controller = Em.Controller.extend({

});

});

require.register("controllers/main/service/reassign/step6_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ReassignMasterWizardStep6Controller = App.HighAvailabilityProgressPageController.extend({

  isReassign: true,

  commands: ['stopMysqlService', 'putHostComponentsInMaintenanceMode', 'deleteHostComponents', 'startServices'],

  clusterDeployState: 'REASSIGN_MASTER_INSTALLING',

  multiTaskCounter: 0,

  hostComponents: [],

  loadStep: function () {
    if (this.get('content.reassign.component_name') === 'NAMENODE' && App.get('isHaEnabled')) {
      this.set('hostComponents', ['NAMENODE', 'ZKFC']);
    } else {
      this.set('hostComponents', [this.get('content.reassign.component_name')]);
    }
    this._super();
  },

  initializeTasks: function () {
    var commands = this.get('commands');
    var hostComponentsNames = '';
    this.get('hostComponents').forEach(function (comp, index) {
      hostComponentsNames += index ? '+' : '';
      hostComponentsNames += comp === 'ZKFC' ? comp : App.format.role(comp);
    }, this);
    var currentStep = App.router.get('reassignMasterController.currentStep');
    for (var i = 0; i < commands.length; i++) {
      var title =  Em.I18n.t('services.reassign.step6.tasks.' + commands[i] + '.title').format(hostComponentsNames);
      this.get('tasks').pushObject(Ember.Object.create({
        title: title,
        status: 'PENDING',
        id: i,
        command: commands[i],
        showRetry: false,
        showRollback: false,
        name: title,
        displayName: title,
        progress: 0,
        isRunning: false,
        hosts: []
      }));
    }

    this.removeUnneededTasks();
    this.set('isLoaded', true);
  },

  removeUnneededTasks: function () {
    if ( this.get('content.reassign.component_name') !== 'MYSQL_SERVER' ) {
      this.removeTasks(['putHostComponentsInMaintenanceMode', 'stopMysqlService']);
    }
  },

  /**
   * remove tasks by command name
   */
  removeTasks: function(commands) {
    var tasks = this.get('tasks'),
        index = null
        cmd = null;

    commands.forEach(function(command) {
      cmd = tasks.filterProperty('command', command);

      if (cmd.length === 0) {
        return false;
      } else {
        index = tasks.indexOf( cmd[0] );
      }

      tasks.splice( index, 1 );
    });
  },

  hideRollbackButton: function () {
    var failedTask = this.get('tasks').findProperty('showRollback');
    if (failedTask) {
      failedTask.set('showRollback', false)
    }
  }.observes('tasks.@each.showRollback'),

  onComponentsTasksSuccess: function () {
    this.set('multiTaskCounter', this.get('multiTaskCounter') + 1);
    if (this.get('multiTaskCounter') >= this.get('hostComponents').length) {
      this.onTaskCompleted();
    }
  },

  startServices: function () {
    App.ajax.send({
      name: 'common.services.update',
      sender: this,
      data: {
        "context": "Start all services",
        "ServiceInfo": {
          "state": "STARTED"
        },
        urlParams: "params/run_smoke_test=true"
      },
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  deleteHostComponents: function () {
    this.set('multiTaskCounter', 0);
    var hostComponents = this.get('hostComponents');
    var hostName = this.get('content.reassignHosts.source');
    for (var i = 0; i < hostComponents.length; i++) {
      App.ajax.send({
        name: 'common.delete.host_component',
        sender: this,
        data: {
          hostName: hostName,
          componentName: hostComponents[i]
        },
        success: 'onComponentsTasksSuccess',
        error: 'onDeleteHostComponentsError'
      });
    }
  },

  onDeleteHostComponentsError: function (error) {
    if (error.responseText.indexOf('org.apache.ambari.server.controller.spi.NoSuchResourceException') !== -1) {
      this.onComponentsTasksSuccess();
    } else {
      this.onTaskError();
    }
  },

  putHostComponentsInMaintenanceMode: function () {
    this.set('multiTaskCounter', 0);
    var hostComponents = this.get('hostComponents');
    var hostName = this.get('content.reassignHosts.source');
    for (var i = 0; i < hostComponents.length; i++) {
      App.ajax.send({
        name: 'common.host.host_component.passive',
        sender: this,
        data: {
          hostName: hostName,
          passive_state: "ON",
          componentName: hostComponents[i]
        },
        success: 'onComponentsTasksSuccess',
        error: 'onTaskError'
      });
    }
  },

  /**
   * make server call to stop services
   */
  stopMysqlService: function () {
    var data = {};

    data.context = "Stop required services";
    data.hostName = this.get('content.reassignHosts.source');
    data.serviceName = 'HIVE';
    data.HostRoles = { "state": "INSTALLED" };
    data.componentName = "MYSQL_SERVER";

    App.ajax.send({
      name: 'common.host.host_component.update',
      sender: this,
      data: data,
      success: 'startPolling',
      error: 'onTaskError'
    });
  }
});

});

require.register("controllers/main/service/reassign/step7_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ReassignMasterWizardStep7Controller = App.ReassignMasterWizardStep4Controller.extend({
  name: 'reassignMasterWizardStep7Controller',
  isReassign: true,
  commands: ['putHostComponentsInMaintenanceMode', 'deleteHostComponents', 'cleanMySqlServer', 'configureMySqlServer', 'startServices'],
  clusterDeployState: 'REASSIGN_MASTER_INSTALLING',
  multiTaskCounter: 0,

  initializeTasks: function () {

    var commands = this.get('commands');
    var hostComponentsNames = this.getHostComponentsNames();

    for (var i = 0; i < commands.length; i++) {
      var TaskLabel = i === 3 ? this.get('serviceName') : hostComponentsNames; //For Reconfigure task, show serviceName
      var title = Em.I18n.t('services.reassign.step4.tasks.' + commands[i] + '.title').format(TaskLabel);
      this.get('tasks').pushObject(Ember.Object.create({
        title: title,
        status: 'PENDING',
        id: i,
        command: commands[i],
        showRetry: false,
        showRollback: false,
        name: title,
        displayName: title,
        progress: 0,
        isRunning: false,
        hosts: []
      }));
    }

    this.removeUnneededTasks();
    this.set('isLoaded', true);
  },

  putHostComponentsInMaintenanceMode: function () {
    this.set('multiTaskCounter', 0);
    var hostComponents = this.get('hostComponents');
    var hostName = this.get('content.reassignHosts.target');
    for (var i = 0; i < hostComponents.length; i++) {
      App.ajax.send({
        name: 'common.host.host_component.passive',
        sender: this,
        data: {
          hostName: hostName,
          passive_state: "ON",
          componentName: hostComponents[i]
        },
        success: 'onComponentsTasksSuccess',
        error: 'onTaskError'
      });
    }
  },

  deleteHostComponents: function () {
    this.set('multiTaskCounter', 0);
    var hostComponents = this.get('hostComponents');
    var hostName = this.get('content.reassignHosts.target');
    for (var i = 0; i < hostComponents.length; i++) {
      App.ajax.send({
        name: 'common.delete.host_component',
        sender: this,
        data: {
          hostName: hostName,
          componentName: hostComponents[i]
        },
        success: 'onComponentsTasksSuccess',
        error: 'onDeleteHostComponentsError'
      });
    }
  }
});

});

require.register("controllers/main/service/reassign_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.ReassignMasterController = App.WizardController.extend({

  name: 'reassignMasterController',

  totalSteps: 7,

  /**
   * Used for hiding back button in wizard
   */
  hideBackButton: true,

  /**
   * All wizards data will be stored in this variable
   *
   * cluster - cluster name
   * installOptions - ssh key, repo info, etc.
   * services - services list
   * hosts - list of selected hosts
   * slaveComponentHosts, - info about slave hosts
   * masterComponentHosts - info about master hosts
   * config??? - to be described later
   */
  content: Em.Object.create({
    cluster: null,
    hosts: null,
    installOptions: null,
    services: null,
    slaveComponentHosts: null,
    masterComponentHosts: null,
    serviceConfigProperties: null,
    advancedServiceConfig: null,
    controllerName: 'reassignMasterController',
    serviceName: 'MISC',
    hdfsUser: "hdfs",
    group: "hadoop",
    reassign: null,
    componentsWithManualCommands: ['NAMENODE', 'SECONDARY_NAMENODE', 'OOZIE_SERVER', 'MYSQL_SERVER', 'APP_TIMELINE_SERVER'],
    hasManualSteps: false,
    hasCheckDBStep: false,
    componentsWithCheckDBStep: ['HIVE_METASTORE', 'HIVE_SERVER', 'OOZIE_SERVER'],
    securityEnabled: false,
    componentsWithoutSecurityConfigs: ['MYSQL_SERVER']
  }),

  /**
   * Wizard properties in local storage, which should be cleaned right after wizard has been finished
   */
  dbPropertiesToClean: [
    'cluster',
    'hosts',
    'installOptions',
    'masterComponentHosts',
    'serviceComponents',
    'masterComponent',
    'securityEnabled',
    'currentStep',
    'reassignHosts',
    'tasksStatuses',
    'tasksRequestIds',
    'requestIds'
  ],

  addManualSteps: function () {
    this.set('content.hasManualSteps', this.get('content.componentsWithManualCommands').contains(this.get('content.reassign.component_name')));
  }.observes('content.reassign.component_name'),

  addCheckDBStep: function () {
    this.set('content.hasCheckDBStep', this.get('content.componentsWithCheckDBStep').contains(this.get('content.reassign.component_name')));
  }.observes('content.reassign.component_name'),

  getSecurityStatus: function () {
    if (App.get('testMode')) {
      this.set('securityEnabled', !App.get('testEnableSecurity'));
    } else {
      //get Security Status From Server
      App.ajax.send({
        name: 'config.tags',
        sender: this,
        success: 'getSecurityStatusSuccessCallback',
        error: 'errorCallback'
      });
    }
  },

  errorCallback: function () {
    console.error('Cannot get security status from server');
  },

  getSecurityStatusSuccessCallback: function (data) {
    var configs = data.Clusters.desired_configs;
    if ('cluster-env' in configs) {
      this.getServiceConfigsFromServer(configs['cluster-env'].tag);
    }
    else {
      console.error('Cannot get security status from server');
    }
  },

  getServiceConfigsFromServer: function (tag) {
    var self = this;
    var tags = [
      {
        siteName: "cluster-env",
        tagName: tag
      }
    ];
    App.router.get('configurationController').getConfigsByTags(tags).done(function (data) {
      var configs = data.findProperty('tag', tag).properties;
      var result = configs && (configs['security_enabled'] === 'true' || configs['security_enabled'] === true);
      self.saveSecurityEnabled(result);
      App.clusterStatus.setClusterStatus({
        clusterName: self.get('content.cluster.name'),
        clusterState: 'DEFAULT',
        wizardControllerName: 'reassignMasterController',
        localdb: App.db.data
      });
    });
  },

  /**
   * Load tasks statuses for step5 of Reassign Master Wizard to restore installation
   */
  loadTasksStatuses: function () {
    var statuses = App.db.getReassignTasksStatuses();
    this.set('content.tasksStatuses', statuses);
    console.log('ReassignMasterController.loadTasksStatuses: loaded statuses', statuses);
  },

  /**
   * save status of the cluster.
   * @param clusterStatus object with status,requestId fields.
   */
  saveClusterStatus: function (clusterStatus) {
    var oldStatus = this.toObject(this.get('content.cluster'));
    clusterStatus = jQuery.extend(oldStatus, clusterStatus);
    if (clusterStatus.requestId) {
      clusterStatus.requestId.forEach(function (requestId) {
        if (clusterStatus.oldRequestsId.indexOf(requestId) === -1) {
          clusterStatus.oldRequestsId.push(requestId)
        }
      }, this);
    }
    this.set('content.cluster', clusterStatus);
    this.save('cluster');
  },

  /**
   * Save Master Component Hosts data to Main Controller
   * @param stepController App.WizardStep5Controller
   */
  saveMasterComponentHosts: function (stepController) {
    var obj = stepController.get('selectedServicesMasters');

    var masterComponentHosts = [];
    obj.forEach(function (_component) {
      masterComponentHosts.push({
        display_name: _component.get('display_name'),
        component: _component.get('component_name'),
        hostName: _component.get('selectedHost'),
        serviceId: _component.get('serviceId'),
        isInstalled: true
      });
    });
    App.db.setMasterComponentHosts(masterComponentHosts);
    this.set('content.masterComponentHosts', masterComponentHosts);
  },

  loadComponentToReassign: function () {
    var masterComponent = App.db.getMasterToReassign();
    if (masterComponent) {
      this.set('content.reassign', masterComponent);
    }
  },

  saveComponentToReassign: function (masterComponent) {
    var component = {
      component_name: masterComponent.get('componentName'),
      display_name: masterComponent.get('displayName'),
      service_id: masterComponent.get('service.serviceName'),
      host_id: masterComponent.get('hostName')
    };
    App.db.setMasterToReassign(component);
  },

  saveTasksStatuses: function (statuses) {
    App.db.setReassignTasksStatuses(statuses);
    this.set('content.tasksStatuses', statuses);
    console.log('ReassignMasterController.saveTasksStatuses: saved statuses', statuses);
  },

  loadTasksRequestIds: function () {
    var requestIds = App.db.getReassignTasksRequestIds();
    this.set('content.tasksRequestIds', requestIds);
  },

  saveTasksRequestIds: function (requestIds) {
    App.db.setReassignTasksRequestIds(requestIds);
    this.set('content.tasksRequestIds', requestIds);
  },

  loadRequestIds: function () {
    var requestIds = App.db.getReassignMasterWizardRequestIds();
    this.set('content.requestIds', requestIds);
  },

  saveRequestIds: function (requestIds) {
    App.db.setReassignMasterWizardRequestIds(requestIds);
    this.set('content.requestIds', requestIds);
  },

  saveComponentDir: function (componentDir) {
    App.db.setReassignMasterWizardComponentDir(componentDir);
    this.set('content.componentDir', componentDir);
  },

  loadComponentDir: function () {
    var componentDir = App.db.getReassignMasterWizardComponentDir();
    this.set('content.componentDir', componentDir);
  },

  saveReassignHosts: function (reassignHosts) {
    App.db.setReassignMasterWizardReassignHosts(reassignHosts);
    this.set('content.reassignHosts', reassignHosts);
  },

  loadReassignHosts: function () {
    var reassignHosts = App.db.getReassignMasterWizardReassignHosts();
    this.set('content.reassignHosts', reassignHosts);
  },


  saveSecurityEnabled: function (securityEnabled) {
    this.setDBProperty('securityEnabled', securityEnabled);
    this.set('content.securityEnabled', securityEnabled);
  },

  loadSecurityEnabled: function () {
    var securityEnabled = this.getDBProperty('securityEnabled');
    this.set('content.securityEnabled', securityEnabled);
  },

  saveSecureConfigs: function (secureConfigs) {
    this.setDBProperty('secureConfigs', secureConfigs);
    this.set('content.secureConfigs', secureConfigs);
  },

  loadSecureConfigs: function () {
    var secureConfigs = this.getDBProperty('secureConfigs');
    this.set('content.secureConfigs', secureConfigs);
  },

  saveServiceProperties: function (properties) {
    this.setDBProperty('serviceProperties', properties);
    this.set('content.serviceProperties', properties);
  },

  loadServiceProperties: function () {
    var serviceProperties = this.getDBProperty('serviceProperties');
    this.set('content.serviceProperties', serviceProperties);
  },

  saveDatabaseType: function (type) {
    this.setDBProperty('databaseType', type);
    this.set('content.databaseType', type);
  },

  loadDatabaseType: function () {
    var databaseType = this.getDBProperty('databaseType');
    this.set('content.databaseType', databaseType);
    var component = this.get('content.reassign.component_name');

    if (component === 'OOZIE_SERVER') {
      if (this.get('content.hasCheckDBStep') && databaseType && databaseType !== 'derby') {
        // components with manual commands
        var manual = App.router.reassignMasterController.get('content.componentsWithManualCommands').without('OOZIE_SERVER');
        App.router.reassignMasterController.set('content.hasManualSteps', false);
        App.router.reassignMasterController.set('content.componentsWithManualCommands', manual);
      }
    }
  },

  /**
   * Load data for all steps until <code>current step</code>
   */
  loadAllPriorSteps: function () {
    var step = this.get('currentStep');
    switch (step) {
      case '7':
      case '6':
      case '5':
        this.loadSecureConfigs();
        this.loadComponentDir();
      case '4':
        this.loadTasksStatuses();
        this.loadTasksRequestIds();
        this.loadRequestIds();
      case '3':
        this.loadReassignHosts();
      case '2':
        this.loadServicesFromServer();
        this.loadMasterComponentHosts();
        this.loadConfirmedHosts();
      case '1':
        this.loadComponentToReassign();
        this.loadDatabaseType();
        this.loadServiceProperties();
        this.load('cluster');
    }
  },

  /**
   * Remove all loaded data.
   * Created as copy for App.router.clearAllSteps
   */
  clearAllSteps: function () {
    this.clearInstallOptions();
    // clear temporary information stored during the install
    this.set('content.cluster', this.getCluster());
  },

  setCurrentStep: function (currentStep, completed) {
    this._super(currentStep, completed);
    App.clusterStatus.setClusterStatus({
      clusterName: this.get('content.cluster.name'),
      wizardControllerName: 'reassignMasterController',
      localdb: App.db.data
    });
  },

  /**
   * Clear all temporary data
   */
  finish: function () {
    this.setCurrentStep('1');
    this.clearAllSteps();
    this.clearStorageData();
    this.resetDbNamespace();
    App.router.get('updateController').updateAll();
  }

});

});

require.register("controllers/main/views/details_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

App.MainViewsDetailsController = Em.Controller.extend({
  name:'mainViewsDetailsController'
});

});

require.register("controllers/main/views_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainViewsController = Em.Controller.extend({
  name:'mainViewsController',

  isDataLoaded: false,

  ambariViews: [],

  dataLoading: function () {
    var viewsController = this;
    var dfd = $.Deferred();
    if (this.get('isDataLoaded')) {
      dfd.resolve(this.get('ambariViews'));
    } else {
      var interval = setInterval(function () {
        if (viewsController.get('isDataLoaded')) {
          dfd.resolve(viewsController.get('ambariViews'));
          clearInterval(interval);
        }
      }, 50);
    }
    return dfd.promise();
  },


  loadAmbariViews: function () {
    if (!App.router.get('loggedIn')) {
      return;
    }
    App.ajax.send({
      name: 'views.info',
      sender: this,
      success: 'loadAmbariViewsSuccess',
      error: 'loadAmbariViewsError'
    });
  },

  loadAmbariViewsSuccess: function (data, opt, params) {
    if (data.items.length) {
      App.ajax.send({
        name: 'views.instances',
        sender: this,
        success: 'loadViewInstancesSuccess',
        error: 'loadViewInstancesError'
      });
    } else {
      this.set('ambariViews', []);
      this.set('isDataLoaded', true);
    }
  },

  loadAmbariViewsError: function () {
    this.set('ambariViews', []);
    this.set('isDataLoaded', true);
  },

  loadViewInstancesSuccess: function (data, opt, params) {
    this.set('ambariViews', []);
    var instances = [];
    data.items.forEach(function (view) {
      view.versions.forEach(function (version) {
        version.instances.forEach(function (instance) {
          var current_instance = Em.Object.create({
            iconPath: instance.ViewInstanceInfo.icon_path || "/img/ambari-view-default.png",
            label: instance.ViewInstanceInfo.label || version.ViewVersionInfo.label || instance.ViewInstanceInfo.view_name,
            visible: instance.ViewInstanceInfo.visible || false,
            version: instance.ViewInstanceInfo.version,
            description: instance.ViewInstanceInfo.description || Em.I18n.t('views.main.instance.noDescription'),
            viewName: instance.ViewInstanceInfo.view_name,
            instanceName: instance.ViewInstanceInfo.instance_name,
            href: instance.ViewInstanceInfo.context_path
          });
          if( current_instance.visible ){
            instances.push(current_instance);
          }
        }, this);
      }, this);
    }, this);
    this.get('ambariViews').pushObjects(instances);
    this.set('isDataLoaded', true);
  },

  loadViewInstancesError: function () {
    this.set('ambariViews', []);
    this.set('isDataLoaded', true);
  },

  setView: function(event) {
    if(event.context){
      App.router.route('main/views/' + event.context.viewName + '/' + event.context.version + '/' + event.context.instanceName);
    }
  }
});
});

require.register("controllers/wizard", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');
require('models/host');

App.WizardController = Em.Controller.extend(App.LocalStorage, {

  isStepDisabled: null,

  /**
   * map of actions which load data required by which step
   * used by <code>loadAllPriorSteps</code>
   */
  loadMap: {},

  /**
   * Wizard properties in local storage, which should be cleaned right after wizard has been finished
   */
  dbPropertiesToClean: [
    'service',
    'hosts',
    'masterComponentHosts',
    'slaveComponentHosts',
    'cluster',
    'allHostNames',
    'installOptions',
    'allHostNamesPattern',
    'serviceComponents'
  ],

  init: function () {
    this.set('isStepDisabled', []);
    this.clusters = App.Cluster.find();
    this.get('isStepDisabled').pushObject(Ember.Object.create({
      step: 1,
      value: false
    }));
    for (var i = 2; i <= this.get('totalSteps'); i++) {
      this.get('isStepDisabled').pushObject(Ember.Object.create({
        step: i,
        value: true
      }));
    }
  },

  slaveComponents: function () {
    return App.StackServiceComponent.find().filterProperty('isSlave', true);
  }.property('App.router.clusterController.isLoaded'),

  allHosts: function () {
    var dbHosts = this.get('content.hosts');
    var hosts = [];
    var hostComponents = [];

    for (var hostName in dbHosts) {
      hostComponents = [];
      var disksOverallCapacity = 0;
      var diskFree = 0;
      dbHosts[hostName].hostComponents.forEach(function (componentName) {
        hostComponents.push(Em.Object.create({
          componentName: componentName,
          displayName: App.format.role(componentName)
        }));
      });
      dbHosts[hostName].disk_info.forEach(function (disk) {
        disksOverallCapacity += parseFloat(disk.size);
        diskFree += parseFloat(disk.available);
      });

      hosts.push(Em.Object.create({
        id: hostName,
        hostName: hostName,
        publicHostName: hostName,
        diskInfo: dbHosts[hostName].disk_info,
        diskTotal: disksOverallCapacity / (1024 * 1024),
        diskFree: diskFree / (1024 * 1024),
        disksMounted: dbHosts[hostName].disk_info.length,
        cpu: dbHosts[hostName].cpu,
        memory: dbHosts[hostName].memory,
        osType: dbHosts[hostName].osType ? dbHosts[hostName].osType: 0,
        osArch: dbHosts[hostName].osArch ? dbHosts[hostName].osArch : 0,
        ip: dbHosts[hostName].ip ? dbHosts[hostName].ip: 0,
        hostComponents: hostComponents
      }))
    }
    return hosts;
  }.property('content.hosts'),

  setStepsEnable: function () {
    for (var i = 1; i <= this.totalSteps; i++) {
      var step = this.get('isStepDisabled').findProperty('step', i);
      if (i <= this.get('currentStep')) {
        step.set('value', false);
      } else {
        step.set('value', true);
      }
    }
  }.observes('currentStep'),

  /**
   * Enable step link in left nav menu
   * @param step - step number
   */
  enableStep: function (step) {
    this.get('isStepDisabled').findProperty('step', step).set('value', false);
  },

  setLowerStepsDisable: function (stepNo) {
    for (var i = 1; i < stepNo; i++) {
      var step = this.get('isStepDisabled').findProperty('step', i);
      step.set('value', true);
    }
  },

  /**
   * Set current step to new value.
   * Method moved from App.router.setInstallerCurrentStep
   * @param currentStep
   * @param completed
   */
  currentStep: function () {
    return App.get('router').getWizardCurrentStep(this.get('name').substr(0, this.get('name').length - 10));
  }.property(),

  /**
   * Set current step to new value.
   * Method moved from App.router.setInstallerCurrentStep
   * @param currentStep
   * @param completed
   */
  setCurrentStep: function (currentStep, completed) {
    App.db.setWizardCurrentStep(this.get('name').substr(0, this.get('name').length - 10), currentStep, completed);
    this.set('currentStep', currentStep);
  },

  clusters: null,

  isStep0: function () {
    return this.get('currentStep') == 0;
  }.property('currentStep'),

  isStep1: function () {
    return this.get('currentStep') == 1;
  }.property('currentStep'),

  isStep2: function () {
    return this.get('currentStep') == 2;
  }.property('currentStep'),

  isStep3: function () {
    return this.get('currentStep') == 3;
  }.property('currentStep'),

  isStep4: function () {
    return this.get('currentStep') == 4;
  }.property('currentStep'),

  isStep5: function () {
    return this.get('currentStep') == 5;
  }.property('currentStep'),

  isStep6: function () {
    return this.get('currentStep') == 6;
  }.property('currentStep'),

  isStep7: function () {
    return this.get('currentStep') == 7;
  }.property('currentStep'),

  isStep8: function () {
    return this.get('currentStep') == 8;
  }.property('currentStep'),

  isStep9: function () {
    return this.get('currentStep') == 9;
  }.property('currentStep'),

  isStep10: function () {
    return this.get('currentStep') == 10;
  }.property('currentStep'),

  gotoStep: function (step, disableNaviWarning) {
    if (this.get('isStepDisabled').findProperty('step', step).get('value') !== false) {
      return false;
    }
    // if going back from Step 9 in Install Wizard, delete the checkpoint so that the user is not redirected
    // to Step 9
    if (this.get('content.controllerName') == 'installerController' && this.get('currentStep') === '9' && step < 9) {
      App.clusterStatus.setClusterStatus({
        clusterName: this.get('clusterName'),
        clusterState: 'CLUSTER_NOT_CREATED_1',
        wizardControllerName: 'installerController',
        localdb: App.db.data
      });
    }
    if ((this.get('currentStep') - step) > 1 && !disableNaviWarning) {
      App.ModalPopup.show({
        header: Em.I18n.t('installer.navigation.warning.header'),
        onPrimary: function () {
          App.router.send('gotoStep' + step);
          this.hide();
        },
        body: "If you proceed to go back to Step " + step + ", you will lose any changes you have made beyond this step"
      });
    } else {
      App.router.send('gotoStep' + step);
    }
    return true;
  },

  gotoStep0: function () {
    this.gotoStep(0);
  },

  gotoStep1: function () {
    this.gotoStep(1);
  },

  gotoStep2: function () {
    this.gotoStep(2);
  },

  gotoStep3: function () {
    this.gotoStep(3);
  },

  gotoStep4: function () {
    this.gotoStep(4);
  },

  gotoStep5: function () {
    this.gotoStep(5);
  },

  gotoStep6: function () {
    this.gotoStep(6);
  },

  gotoStep7: function () {
    this.gotoStep(7);
  },

  gotoStep8: function () {
    this.gotoStep(8);
  },

  gotoStep9: function () {
    this.gotoStep(9);
  },

  gotoStep10: function () {
    this.gotoStep(10);
  },

  /**
   * Initialize host status info for step9
   */
  setInfoForStep9: function () {

    var hostInfo = this.getDBProperty('hosts');
    for (var index in hostInfo) {
      hostInfo[index].status = "pending";
      hostInfo[index].message = 'Waiting';
      hostInfo[index].logTasks = [];
      hostInfo[index].tasks = [];
      hostInfo[index].progress = '0';
    }
    this.setDBProperty('hosts', hostInfo);
  },

  /**
   * Remove all data for installOptions step
   */
  clearInstallOptions: function () {
    var installOptions = this.getInstallOptions();
    this.set('content.installOptions', installOptions);
    this.setDBProperty('installOptions', installOptions);
    this.set('content.hosts', {});
    this.setDBProperty('hosts', {});
  },

  toObject: function (object) {
    var result = {};
    for (var i in object) {
      if (object.hasOwnProperty(i)) {
        result[i] = object[i];
      }
    }
    return result;
  },

  /**
   * save status of the cluster. This is called from step8 and step9 to persist install and start requestId
   * @param clusterStatus object with status, isCompleted, requestId, isInstallError and isStartError field.
   */
  saveClusterStatus: function (clusterStatus) {
    var oldStatus = this.toObject(this.get('content.cluster'));
    clusterStatus = jQuery.extend(oldStatus, clusterStatus);
    if (clusterStatus.requestId &&
      clusterStatus.oldRequestsId.indexOf(clusterStatus.requestId) === -1) {
      clusterStatus.oldRequestsId.push(clusterStatus.requestId);
    }
    this.set('content.cluster', clusterStatus);
    this.setDBProperty('cluster', clusterStatus);
  },

  /**
   * Invoke installation of selected services to the server and saves the request id returned by the server.
   * @param isRetry
   */
  installServices: function (isRetry, callback) {
    // clear requests since we are installing services
    // and we don't want to get tasks for previous install attempts
    this.set('content.cluster.oldRequestsId', []);
    var data;
    callback = callback || Em.K;
    if (isRetry) {
      data = {
        context: Em.I18n.t('requestInfo.installComponents'),
        HostRoles: {"state": "INSTALLED"},
        urlParams: "HostRoles/state=INSTALLED"
      }
    } else {
      data = {
        context: Em.I18n.t('requestInfo.installServices'),
        ServiceInfo: {"state": "INSTALLED"},
        urlParams: "ServiceInfo/state=INIT"
      }
    }

    App.ajax.send({
      name: isRetry ? 'common.host_components.update' : 'common.services.update',
      sender: this,
      data: data,
      success: 'installServicesSuccessCallback',
      error: 'installServicesErrorCallback'
    }).then(callback, callback);
  },

  installServicesSuccessCallback: function (jsonData) {
    var installStartTime = App.dateTime();
    console.log("TRACE: In success function for the installService call");
    if (jsonData) {
      var requestId = jsonData.Requests.id;
      console.log('requestId is: ' + requestId);
      var clusterStatus = {
        status: 'PENDING',
        requestId: requestId,
        isInstallError: false,
        isCompleted: false,
        installStartTime: installStartTime
      };
      this.saveClusterStatus(clusterStatus);
    } else {
      console.log('ERROR: Error occurred in parsing JSON data');
    }
  },

  installServicesErrorCallback: function (request, ajaxOptions, error) {
    console.log("TRACE: In error function for the installService call");
    console.log("TRACE: error code status is: " + request.status);
    console.log('Error message is: ' + request.responseText);
    var clusterStatus = {
      status: 'PENDING',
      requestId: this.get('content.cluster.requestId'),
      isInstallError: true,
      isCompleted: false
    };
    this.saveClusterStatus(clusterStatus);
    App.showAlertPopup(Em.I18n.t('common.errorPopup.header'), request.responseText);
  },
  /**
   * show popup, that display status of bootstrap launching
   * @param callback
   * @return {Object}
   */
  showLaunchBootstrapPopup: function (callback) {
    return App.ModalPopup.show({
      header: Em.I18n.t('installer.step2.bootStrap.header'),
      isError: false,
      serverError: null,
      bodyClass: Em.View.extend({
        templateName: require('templates/wizard/bootstrap_call_popup')
      }),
      showFooter: false,
      showCloseButton: false,
      secondary: null,
      /**
       * handle requestId when call is completed,
       * if it's correct call callback and hide popup
       * otherwise notify error and enable buttons to close popup
       * @param requestId
       * @param serverError
       */
      finishLoading: function (requestId, serverError) {
        if (Em.isNone(requestId)) {
          this.set('isError', true);
          this.set('showFooter', true);
          this.set('showCloseButton', true);
          this.set('serverError', serverError);
        } else {
          callback(requestId);
          this.hide();
        }
      }
    });
  },
  /**
   * Bootstrap selected hosts.
   * @param bootStrapData
   * @param callback
   * @return {Object}
   */
  launchBootstrap: function (bootStrapData, callback) {
    var popup = this.showLaunchBootstrapPopup(callback);
    App.ajax.send({
      name: 'wizard.launch_bootstrap',
      sender: this,
      data: {
        bootStrapData: bootStrapData,
        popup: popup
      },
      success: 'launchBootstrapSuccessCallback',
      error: 'launchBootstrapErrorCallback'
    });
    return popup;
  },

  launchBootstrapSuccessCallback: function (data, opt, params) {
    console.log("TRACE: POST bootstrap succeeded");
    params.popup.finishLoading(data.requestId, null);
  },

  launchBootstrapErrorCallback: function (request, ajaxOptions, error, opt, params) {
    console.log("ERROR: POST bootstrap failed");
    params.popup.finishLoading(null, error);
  },

  /**
   * Load <code>content.<name></code> variable from localStorage, if wasn't loaded before.
   * If you specify <code>reload</code> to true - it will reload it.
   * @param name
   * @param reload
   * @return {Boolean}
   */
  load: function (name, reload) {
    if (this.get('content.' + name) && !reload) {
      return false;
    }
    var result = this.getDBProperty(name);
    if (!result) {
      if (this['get' + name.capitalize()]) {
        result = this['get' + name.capitalize()]();
        this.setDBProperty(name, result);
        console.log(this.get('name') + ": created " + name, result);
      }
      else {
        console.debug('get' + name.capitalize(), ' not defined in the ' + this.get('name'));
      }
    }
    this.set('content.' + name, result);
    console.log(this.get('name') + ": loaded " + name, result);
  },

  save: function (name) {
    var value = this.toObject(this.get('content.' + name));
    this.setDBProperty(name, value);
    console.log(this.get('name') + ": saved " + name, value);
  },

  clear: function () {
    this.set('content', Ember.Object.create({
      'controllerName': this.get('content.controllerName')
    }));
    this.set('currentStep', 0);
    this.clearStorageData();
  },

  clusterStatusTemplate: {
    name: "",
    status: "PENDING",
    isCompleted: false,
    requestId: null,
    installStartTime: null,
    installTime: null,
    isInstallError: false,
    isStartError: false,
    oldRequestsId: []
  },

  clearStorageData: function () {
    this.get('dbPropertiesToClean').forEach(function (key) {
      this.setDBProperty(key, undefined);
    }, this);
  },

  getInstallOptions: function() {
    return jQuery.extend({}, App.get('isHadoopWindowsStack') ? this.get('installWindowsOptionsTemplate') : this.get('installOptionsTemplate'));
  },

  installOptionsTemplate: {
    hostNames: "", //string
    manualInstall: false, //true, false
    useSsh: true, //bool
    javaHome: App.defaultJavaHome, //string
    localRepo: false, //true, false
    sshKey: "", //string
    bootRequestId: null, //string
    sshUser: "root", //string
    agentUser: "root" //string
  },

  installWindowsOptionsTemplate: {
    hostNames: "", //string
    manualInstall: true, //true, false
    useSsh: false, //bool
    javaHome: App.defaultJavaHome, //string
    localRepo: false, //true, false
    sshKey: "", //string
    bootRequestId: null, //string
    sshUser: "", //string
    agentUser: "" //string
  },

  loadedServiceComponents: null,

  /**
   * Generate serviceComponents as pr the stack definition  and save it to localdata
   * called form stepController step4WizardController
   */
  loadServiceComponents: function () {
    return App.ajax.send({
      name: 'wizard.service_components',
      sender: this,
      data: {
        stackUrl: App.get('stackVersionURL'),
        stackVersion: App.get('currentStackVersionNumber')
      },
      success: 'loadServiceComponentsSuccessCallback',
      error: 'loadServiceComponentsErrorCallback'
    });
  },

  loadServiceComponentsSuccessCallback: function (jsonData) {
    var savedSelectedServices = this.getDBProperty('selectedServiceNames');
    var savedInstalledServices = this.getDBProperty('installedServiceNames');
    this.set('content.selectedServiceNames', savedSelectedServices);
    this.set('content.installedServiceNames', savedInstalledServices);
    if (!savedSelectedServices) {
      jsonData.items.forEach(function (service) {
        service.StackServices.is_selected = true;
      }, this);
    } else {
      jsonData.items.forEach(function (service) {
        if (savedSelectedServices.contains(service.StackServices.service_name))
          service.StackServices.is_selected = true;
        else
          service.StackServices.is_selected = false;
      }, this);
    }

    if (!savedInstalledServices) {
      jsonData.items.forEach(function (service) {
        service.StackServices.is_installed = false;
      }, this);
    } else {
      jsonData.items.forEach(function (service) {
        if (savedInstalledServices.contains(service.StackServices.service_name))
          service.StackServices.is_installed = true;
        else
          service.StackServices.is_installed = false;
      }, this);
    }

    App.stackServiceMapper.mapStackServices(jsonData);
  },

  loadServiceComponentsErrorCallback: function (request, ajaxOptions, error) {
    console.log("TRACE: STep5 -> In error function for the getServiceComponents call");
    console.log("TRACE: STep5 -> error code status is: " + request.status);
    console.log('Step8: Error message is: ' + request.responseText);
  },

  /**
   * Load config groups from local DB
   */
  loadServiceConfigGroups: function () {
    var serviceConfigGroups = this.getDBProperty('serviceConfigGroups'),
      hosts = this.getDBProperty('hosts'),
      host_names = Em.keys(hosts);
    if (Em.isNone(serviceConfigGroups)) {
      serviceConfigGroups = [];
    }
    else {
      serviceConfigGroups.forEach(function(group) {
        var hostNames = group.hosts.map(function(host_id) {
          for (var i = 0; i < host_names.length; i++) {
            if (hosts[host_names[i]].id === host_id) {
              return host_names[i];
            }
          }
          Em.assert('host is missing!!!!', false);
        });
        Em.set(group, 'hosts', hostNames);
      });
    }
    this.set('content.configGroups', serviceConfigGroups);
    console.log("InstallerController.configGroups: loaded config ", serviceConfigGroups);
  },

  registerErrPopup: function (header, message) {
    App.ModalPopup.show({
      header: header,
      secondary: false,
      bodyClass: Ember.View.extend({
        template: Ember.Handlebars.compile('<p>{{view.message}}</p>'),
        message: message
      })
    });
  },

  /**
   * Save hosts that the user confirmed to proceed with from step 3
   * @param stepController App.WizardStep3Controller
   */
  saveConfirmedHosts: function (stepController) {
    var hosts = this.get('content.hosts'),
      indx = 1;

    //add previously installed hosts
    for (var hostName in hosts) {
      if (!hosts[hostName].isInstalled) {
        delete hosts[hostName];
      }
    }

    stepController.get('confirmedHosts').forEach(function (_host) {
      if (_host.bootStatus == 'REGISTERED') {
        hosts[_host.name] = {
          name: _host.name,
          cpu: _host.cpu,
          memory: _host.memory,
          disk_info: _host.disk_info,
          os_type: _host.os_type,
          os_arch: _host.os_arch,
          ip: _host.ip,
          bootStatus: _host.bootStatus,
          isInstalled: false,
          id: indx++
        };
      }
    });
    console.log('wizardController:saveConfirmedHosts: save hosts ', hosts);
    this.setDBProperty('hosts', hosts);
    this.set('content.hosts', hosts);
  },

  /**
   * Save data after installation to main controller
   * @param stepController App.WizardStep9Controller
   */
  saveInstalledHosts: function (stepController) {
    var hosts = stepController.get('hosts');
    var hostInfo = this.getDBProperty('hosts');

    for (var index in hostInfo) {
      hostInfo[index].status = "pending";
      var host = hosts.findProperty('name', hostInfo[index].name);
      if (host) {
        hostInfo[index].status = host.status;
        hostInfo[index].message = host.message;
        hostInfo[index].progress = host.progress;
      }
    }
    this.set('content.hosts', hostInfo);
    this.setDBProperty('hosts', hostInfo);
    console.log('wizardController:saveInstalledHosts: save hosts ', hostInfo);
  },

  /**
   * Save slaveHostComponents to main controller
   * @param stepController
   */
  saveSlaveComponentHosts: function (stepController) {
    var hosts = stepController.get('hosts'),
      dbHosts = this.getDBProperty('hosts'),
      headers = stepController.get('headers');

    var formattedHosts = Ember.Object.create();
    headers.forEach(function (header) {
      formattedHosts.set(header.get('name'), []);
    });

    hosts.forEach(function (host) {

      var checkboxes = host.get('checkboxes');
      headers.forEach(function (header) {
        var cb = checkboxes.findProperty('title', header.get('label'));
        if (cb.get('checked')) {
          formattedHosts.get(header.get('name')).push({
            group: 'Default',
            isInstalled: cb.get('isInstalled'),
            host_id: dbHosts[host.hostName].id
          });
        }
      });
    });

    var slaveComponentHosts = [];

    headers.forEach(function (header) {
      slaveComponentHosts.push({
        componentName: header.get('name'),
        displayName: header.get('label').replace(/\s/g, ''),
        hosts: formattedHosts.get(header.get('name'))
      });
    });

    this.setDBProperty('slaveComponentHosts', slaveComponentHosts);
    console.log('wizardController.slaveComponentHosts: saved hosts', slaveComponentHosts);
    this.set('content.slaveComponentHosts', slaveComponentHosts);
  },

  /**
   * Return true if cluster data is loaded and false otherwise.
   * This is used for all wizard controllers except for installer wizard.
   */
  dataLoading: function () {
    var dfd = $.Deferred();
    this.connectOutlet('loading');
    if (App.router.get('clusterController.isLoaded')) {
      dfd.resolve();
    } else {
      var interval = setInterval(function () {
        if (App.router.get('clusterController.isLoaded')) {
          dfd.resolve();
          clearInterval(interval);
        }
      }, 50);
    }
    return dfd.promise();
  },

  /**
   * Return true if user data is loaded via App.MainServiceInfoConfigsController
   * This function is used in reassign master wizard right now.
   */

  usersLoading: function () {
    var self = this;
    var dfd = $.Deferred();
    var miscController = App.MainAdminServiceAccountsController.create({content: self.get('content')});
    miscController.loadUsers();
    var interval = setInterval(function () {
      if (miscController.get('dataIsLoaded')) {
        if (self.get("content.hdfsUser")) {
          self.set('content.hdfsUser', miscController.get('content.hdfsUser'));
        }
        dfd.resolve();
        clearInterval(interval);
      }
    }, 10);
    return dfd.promise();
  },

  /**
   * Save cluster status before going to deploy step
   * @param name cluster state. Unique for every wizard
   */
  saveClusterState: function (name) {
    App.clusterStatus.setClusterStatus({
      clusterName: this.get('content.cluster.name'),
      clusterState: name,
      wizardControllerName: this.get('content.controllerName'),
      localdb: App.db.data
    });
  },

  /**
   * load advanced configs from server
   */
  loadAdvancedConfigs: function (dependentController) {
    var self = this;
    var loadServiceConfigsFn = function(clusterProperties) {
      var stackServices = self.get('content.services').filter(function (service) {
        return service.get('isInstalled') || service.get('isSelected');
      });
      var counter = stackServices.length;
      var loadAdvancedConfigResult = [];
      dependentController.set('isAdvancedConfigLoaded', false);
      stackServices.forEach(function (service) {
        var serviceName = service.get('serviceName');
        App.config.loadAdvancedConfig(serviceName, function (properties) {
          var supportsFinal = App.config.getConfigTypesInfoFromService(service).supportsFinal;

          function shouldSupportFinal(filename) {
            var matchingConfigType = supportsFinal.find(function (configType) {
              return filename.startsWith(configType);
            });
            return !!matchingConfigType;
          }

          properties.forEach(function (property) {
            property.supportsFinal = shouldSupportFinal(property.filename);
          });
          loadAdvancedConfigResult.pushObjects(properties);
          counter--;
          //pass configs to controller after last call is completed
          if (counter === 0) {
            loadAdvancedConfigResult.pushObjects(clusterProperties);
            self.set('content.advancedServiceConfig', loadAdvancedConfigResult);
            self.setDBProperty('advancedServiceConfig', loadAdvancedConfigResult);
            dependentController.set('isAdvancedConfigLoaded', true);
          }
        });
      }, this);
    };
    App.config.loadClusterConfig(loadServiceConfigsFn);
  },


  /**
   * Load serviceConfigProperties to model
   */
  loadServiceConfigProperties: function () {
    var serviceConfigProperties = this.getDBProperty('serviceConfigProperties');
    this.set('content.serviceConfigProperties', serviceConfigProperties);
    console.log("AddHostController.loadServiceConfigProperties: loaded config ", serviceConfigProperties);
  },
  /**
   * Save config properties
   * @param stepController Step7WizardController
   */
  saveServiceConfigProperties: function (stepController) {
    var serviceConfigProperties = [];
    var fileNamesToUpdate = [];
    stepController.get('stepConfigs').forEach(function (_content) {

      if (_content.serviceName === 'YARN') {
        _content.set('configs', App.config.textareaIntoFileConfigs(_content.get('configs'), 'capacity-scheduler.xml'));
      }

      _content.get('configs').forEach(function (_configProperties) {
        var configProperty = {
          id: _configProperties.get('id'),
          name: _configProperties.get('name'),
          value: _configProperties.get('value'),
          defaultValue: _configProperties.get('defaultValue'),
          description: _configProperties.get('description'),
          serviceName: _configProperties.get('serviceName'),
          domain: _configProperties.get('domain'),
          isVisible: _configProperties.get('isVisible'),
          isFinal: _configProperties.get('isFinal'),
          defaultIsFinal: _configProperties.get('isFinal'),
          supportsFinal: _configProperties.get('supportsFinal'),
          filename: _configProperties.get('filename'),
          displayType: _configProperties.get('displayType'),
          isRequiredByAgent: _configProperties.get('isRequiredByAgent'),
          hasInitialValue: !!_configProperties.get('hasInitialValue'),
          isRequired: _configProperties.get('isRequired'), // flag that allow saving property with empty value
          group: !!_configProperties.get('group') ? _configProperties.get('group.name') : null,
          showLabel: _configProperties.get('showLabel'),
          category: _configProperties.get('category')
        };
        serviceConfigProperties.push(configProperty);
      }, this);
      // check for configs that need to update for installed services
      if (stepController.get('installedServiceNames') && stepController.get('installedServiceNames').contains(_content.get('serviceName'))) {
        // get only modified configs
        var configs = _content.get('configs').filter(function (config) {
          if (config.get('isNotDefaultValue') || (config.get('defaultValue') === null)) {
            var notAllowed = ['masterHost', 'masterHosts', 'slaveHosts', 'slaveHost'];
            return !notAllowed.contains(config.get('displayType')) && !!config.filename;
          }
          return false;
        });
        // if modified configs detected push all service's configs for update
        if (configs.length) {
          fileNamesToUpdate = fileNamesToUpdate.concat(configs.mapProperty('filename').uniq());
        }
        // watch for properties that are not modified but have to be updated
        if (_content.get('configs').someProperty('forceUpdate')) {
          // check for already added modified properties
          var forceUpdatedFileNames = _content.get('configs').filterProperty('forceUpdate', true).mapProperty('filename').uniq();
          fileNamesToUpdate = fileNamesToUpdate.concat(forceUpdatedFileNames).uniq();
        }
      }
    }, this);
    this.setDBProperty('serviceConfigProperties', serviceConfigProperties);
    this.set('content.serviceConfigProperties', serviceConfigProperties);
    this.setDBProperty('fileNamesToUpdate', fileNamesToUpdate);
  },
  /**
   * save Config groups
   * @param stepController
   * @param isAddService
   */
  saveServiceConfigGroups: function (stepController, isAddService) {
    var serviceConfigGroups = [],
      isForInstalledService = false,
      hosts = isAddService ? App.router.get('addServiceController').getDBProperty('hosts') : this.getDBProperty('hosts');
    stepController.get('stepConfigs').forEach(function (service) {
      // mark group of installed service
      if (service.get('selected') === false) isForInstalledService = true;
      service.get('configGroups').forEach(function (configGroup) {
        var properties = [];
        configGroup.get('properties').forEach(function (property) {
          properties.push({
            isRequiredByAgent: property.get('isRequiredByAgent'),
            name: property.get('name'),
            value: property.get('value'),
            isFinal: property.get('isFinal'),
            filename: property.get('filename')
          })
        });
        //configGroup copied into plain JS object to avoid Converting circular structure to JSON
        var hostNames = configGroup.get('hosts').map(function(host_name) {return hosts[host_name].id;});
        serviceConfigGroups.push({
          id: configGroup.get('id'),
          name: configGroup.get('name'),
          description: configGroup.get('description'),
          hosts: hostNames,
          properties: properties,
          isDefault: configGroup.get('isDefault'),
          isForInstalledService: isForInstalledService,
          isForUpdate: configGroup.isForUpdate || configGroup.get('hash') != this.getConfigGroupHash(configGroup, hostNames),
          service: {id: configGroup.get('service.id')}
        });
      }, this)
    }, this);
    this.setDBProperty('serviceConfigGroups', serviceConfigGroups);
    this.set('content.configGroups', serviceConfigGroups);
  },

  /**
   * generate string hash for config group
   * @param {Object} configGroup
   * @param {Array|undefined} hosts
   * @returns {String|null}
   * @method getConfigGroupHash
   */
  getConfigGroupHash: function(configGroup,  hosts) {
    if (!Em.get(configGroup, 'properties.length') && !Em.get(configGroup, 'hosts.length') && !hosts) {
      return null;
    }
    var hash = {};
    Em.get(configGroup, 'properties').forEach(function (config) {
      hash[Em.get(config, 'name')] = {value: Em.get(config, 'value'), isFinal: Em.get(config, 'isFinal')};
    });
    hash['hosts'] = hosts || Em.get(configGroup, 'hosts');
    return JSON.stringify(hash);
  },

  /**
   * return slaveComponents bound to hosts
   * @return {Array}
   */
  getSlaveComponentHosts: function () {
    var components = this.get('slaveComponents');
    var result = [];
    var installedServices = App.Service.find().mapProperty('serviceName');
    var selectedServices = App.StackService.find().filterProperty('isSelected', true).mapProperty('serviceName');
    var installedComponentsMap = {};
    var uninstalledComponents = [];

    components.forEach(function (component) {
      if (installedServices.contains(component.get('serviceName'))) {
        installedComponentsMap[component.get('componentName')] = [];
      } else if (selectedServices.contains(component.get('serviceName'))) {
        uninstalledComponents.push(component);
      }
    }, this);
    installedComponentsMap['HDFS_CLIENT'] = [];

    App.HostComponent.find().forEach(function (hostComponent) {
      if (installedComponentsMap[hostComponent.get('componentName')]) {
        installedComponentsMap[hostComponent.get('componentName')].push(hostComponent.get('hostName'));
      }
    }, this);

    for (var componentName in installedComponentsMap) {
      var name = (componentName === 'HDFS_CLIENT') ? 'CLIENT' : componentName;
      var component = {
        componentName: name,
        displayName: App.format.role(name),
        hosts: [],
        isInstalled: true
      };
      installedComponentsMap[componentName].forEach(function (hostName) {
        component.hosts.push({
          group: "Default",
          hostName: hostName,
          isInstalled: true
        });
      }, this);
      result.push(component);
    }

    uninstalledComponents.forEach(function (component) {
      var hosts = jQuery.extend(true, [], result.findProperty('componentName', 'DATANODE').hosts);
      hosts.setEach('isInstalled', false);
      result.push({
        componentName: component.get('componentName'),
        displayName: App.format.role(component.get('componentName')),
        hosts: hosts,
        isInstalled: false
      })
    });

    return result;
  },
  /**
   * Load master component hosts data for using in required step controllers
   */
  loadMasterComponentHosts: function () {
    var masterComponentHosts = this.getDBProperty('masterComponentHosts');
    var stackMasterComponents = App.get('components.masters').uniq();
    if (!masterComponentHosts) {
      masterComponentHosts = [];
      App.HostComponent.find().filter(function(component) {
        return stackMasterComponents.contains(component.get('componentName'));
      }).forEach(function (item) {
        masterComponentHosts.push({
          component: item.get('componentName'),
          hostName: item.get('hostName'),
          isInstalled: true,
          serviceId: item.get('service.id'),
          display_name: item.get('displayName')
        })
      });
      this.setDBProperty('masterComponentHosts', masterComponentHosts);
    }
    this.set("content.masterComponentHosts", masterComponentHosts);
  },
  /**
   * Load information about hosts with clients components
   */
  loadClients: function () {
    var clients = this.getDBProperty('clientInfo');
    this.set('content.clients', clients);
    console.log(this.get('content.controllerName') + ".loadClients: loaded list ", clients);
  },

  /**
   * load methods assigned to each step
   * methods executed in exact order as they described in map
   * @return {object}
   */
  loadAllPriorSteps: function () {
    var currentStep = this.get('currentStep');
    var loadMap = this.get('loadMap');
    var operationStack = [];
    var dfd = $.Deferred();

    for (var s in loadMap) {
      if (parseInt(s) <= parseInt(currentStep)) {
        operationStack.pushObjects(loadMap[s]);
      }
    }

    var sequence = App.actionSequence.create({context: this});
    sequence.setSequence(operationStack).onFinish(function () {
      dfd.resolve();
    }).start();

    return dfd.promise();
  },

  /**
   * return new object extended from clusterStatusTemplate
   * @return Object
   */
  getCluster: function () {
    return jQuery.extend({}, this.get('clusterStatusTemplate'), {name: App.router.getClusterName()});
  },

  /**
   * Load services data from server.
   */
  loadServicesFromServer: function () {
    var services = this.getDBProperty('services');
    if (!services) {
      services = {
        selectedServices: [],
        installedServices: []
      };
      App.StackService.find().forEach(function(item){
        var isInstalled = App.Service.find().someProperty('id', item.get('serviceName'));
        item.set('isSelected', isInstalled);
        item.set('isInstalled', isInstalled);
        if (isInstalled) {
          services.selectedServices.push(item.get('serviceName'));
          services.installedServices.push(item.get('serviceName'));
        }
      },this);
      this.setDBProperty('services',services);
    } else {
      App.StackService.find().forEach(function(item) {
        var isSelected =   services.selectedServices.contains(item.get('serviceName'));
        var isInstalled = services.installedServices.contains(item.get('serviceName'));
        item.set('isSelected', isSelected);
        item.set('isInstalled', isInstalled);
      },this);
    }
    this.set('content.services', App.StackService.find());
  },

  /**
   * Load confirmed hosts.
   * Will be used at <code>Assign Masters(step5)</code> step
   */
  loadConfirmedHosts: function () {
    var hosts = App.db.getHosts();

    if (hosts) {
      this.set('content.hosts', hosts);
    }
  }
});

});

require.register("controllers/wizard/slave_component_groups_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');
/**
 * Used to manage slave component config. User could create different settings for separate group
 * @type {*}
 */
App.SlaveComponentGroupsController = Em.ArrayController.extend({

  name: 'slaveComponentGroupsController',

  contentBinding: 'App.router.wizardStep7Controller.slaveComponentHosts',

  stepConfigsBinding: 'App.router.wizardStep7Controller.stepConfigs',

  serviceBinding: 'App.router.wizardStep7Controller.selectedService',

  servicesBinding: 'App.router.wizardStep7Controller.content.services',

  clearStep: function () {

  },

  loadStep: function () {
    this.clearStep();
    this.loadGroups();
  },

  loadGroups: function () {

    this.get('stepConfigs').forEach(function (_serviceConfig) {
      var categoryConfig = _serviceConfig.get('configCategories');
      if (categoryConfig.someProperty('isForSlaveComponent', true)) {
        var slaveCategory = categoryConfig.findProperty('isForSlaveComponent', true);
        // this.get('content') -> Output of Step 6: Mapping of each slave component and set of hosts it runs on
        if (this.get('content')) {
          if (this.get('content').someProperty('componentName', slaveCategory.get('primaryName'))) {
            // component --> each column in Step 6 is a component ( slave component )
            var component = this.get('content').findProperty('componentName', slaveCategory.get('primaryName'));
            // slaveConfigs --> originally set as null in the class App.SlaveCategory in model/service_config.js
            var slaveConfigs = slaveCategory.get('slaveConfigs');

            slaveCategory.set('slaveConfigs', App.SlaveConfigs.create(component));
            var slaveGroups = [];
            if (component.groups) {
              component.groups.forEach(function (_group) {
                slaveGroups.pushObject(_group);
              }, this);
              slaveCategory.set('slaveConfigs.groups', slaveGroups);
            }
            slaveCategory.set('slaveConfigs.componentName', component.componentName);
            slaveCategory.set('slaveConfigs.displayName', component.displayName);
            /*slaveCategory.set('slaveConfigs.groups.name', component.get('name'));
             slaveCategory.set('slaveConfigs.groups.index', component.get('index'));
             slaveCategory.set('slaveConfigs.groups.type', component.get('type'));
             slaveCategory.set('slaveConfigs.groups.active', component.get('active'));*/
            if (!slaveCategory.get('slaveConfigs.groups')) {
              slaveCategory.set('slaveConfigs.groups', []);
              var componentProperties = this.componentProperties(_serviceConfig.serviceName);
              var defaultGroup = {name: 'Default', index: 'default', type: 'default', active: true, properties: componentProperties};
              slaveCategory.get('slaveConfigs.groups').pushObject(App.Group.create(defaultGroup));
            }
          }
        }
      }
    }, this);
  },

  // returns key-value pairs i.e. all fields for slave component for this specific service.
  componentProperties: function (serviceName) {

    var serviceConfigs = App.config.get('preDefinedServiceConfigs').findProperty('serviceName', serviceName);

    var configs = [];
    var componentName = null;
    switch (serviceName) {
      case 'HDFS':
        componentName = 'DataNode';
        break;
      case 'HBASE':
        componentName = 'RegionServer';
    }
    var slaveConfigs = serviceConfigs.configs.filterProperty('category', componentName);
    slaveConfigs.forEach(function (_serviceConfigProperty) {
      var serviceConfigProperty = App.ServiceConfigProperty.create(_serviceConfigProperty);

      switch(serviceConfigProperty.name){
        case 'dfs_data_dir' :
          serviceConfigProperty.initialValue();
          break;
        case 'mapred_local_dir' :
          serviceConfigProperty.initialValue();
          break;
      }
      configs.pushObject(serviceConfigProperty);
      serviceConfigProperty.validate();
    }, this);
    return configs;
  },

  selectedComponentName: function () {
    switch (App.router.get('wizardStep7Controller.selectedService.serviceName')) {
      case 'HDFS':
        return { name: 'DATANODE',
          displayName: 'DataNode'};
      case 'HBASE':
        return { name: 'HBASE_REGIONSERVER',
          displayName: 'RegionServer'};
      default:
        return null;
    }

  }.property('service'),

  selectedComponentDisplayName: function() {
    return App.format.role(this.get('selectedComponentName'));
  }.property('selectedComponentName'),

  selectedSlaveComponent: function () {
    var selectedComponentName = this.get('selectedComponentName') ? this.get('selectedComponentName').displayName : null;
    var configs = null;
    if (selectedComponentName) {
      App.router.get('wizardStep7Controller.stepConfigs').forEach(function (_serviceConfig) {
        var categoryConfig = _serviceConfig.get('configCategories');
        if (categoryConfig.someProperty('name', selectedComponentName)) {
          configs = categoryConfig.findProperty('name', selectedComponentName).get('slaveConfigs');
        }
      }, this);
    }
    return configs;
  }.property('selectedComponentName', 'stepConfigs.@each.configCategories', 'stepConfigs.@each.configCategories.@each.slaveConfigs'),

  hosts: function () {
    if (this.get('selectedSlaveComponent')) {
      return this.get('selectedSlaveComponent').get('hosts');
    }
  }.property('selectedSlaveComponent'),

  groups: function () {
    var hosts = this.get('hosts');
    if(hosts){
      return hosts.mapProperty('group').uniq();
    }
    return [];
  }.property('hosts'),

  componentGroups: function () {
    var component = this.get('selectedSlaveComponent');
    if (component && component.get('groups')) {
      return component.get('groups');
    }
    return [];
  }.property('selectedSlaveComponent', 'selectedSlaveComponent.groups', 'stepConfigs.@each.configCategories.@each.slaveConfigs.groups.@each.properties.@each.value'),


  getGroupsForDropDown: function () {
    return this.get('componentGroups').getEach('name');
  }.property('selectedComponentName', 'componentGroups.@each.name'),

  activeGroup: function () {
    var componentGroups = this.get('componentGroups');
    if (componentGroups) {
      var active = componentGroups.findProperty('active', true);
      if (active){
        return active;
      }
    }
    return null;
  }.property('componentGroups.@each.active', 'componentGroups.@each.name', 'componentGroups.@each.properties.@each.value'),


  /**
   * Show slave hosts to groups popup
   * @param event
   */
  showAddSlaveComponentGroup: function (event) {
    var componentName = event.context;
    var component = this.get('selectedSlaveComponent');
    App.ModalPopup.show({
      header: componentName + Em.I18n.t('installer.controls.slaveComponentGroups'),
      bodyClass: Ember.View.extend({
        controllerBinding: 'App.router.slaveComponentGroupsController',
        header: Em.I18n.t('installer.slaveComponentHostsPopup.header').format(this.get('selectedComponentDisplayName')),
        templateName: require('templates/wizard/slave_component_hosts_popup')
      }),
      onPrimary: function (event) {
        if (component.tempSelectedGroups && component.tempSelectedGroups.length) {
          component.tempSelectedGroups.forEach(function (item) {
            var changed = component.get('hosts').filterProperty('hostName', item.hostName);
            changed.setEach('group', item.groupName);
          })
        }
        delete component.tempSelectedGroups;
        this.hide();
      },
      onSecondary: function (event) {
        delete component.tempSelectedGroups;
        this.hide();
      },
      onClose: function (event) {
        delete component.tempSelectedGroups;
        this.hide();
      }
    });
  },

  /**
   * Utility method. Save temporary info about changes in <code>slave hosts to groups</code> popup
   * @param host
   * @param groupName
   */
  changeHostGroup: function (host, groupName) {
    var component = this.get('selectedSlaveComponent');
    if (component.tempSelectedGroups === undefined) {
      component.tempSelectedGroups = [];
    }
    var values = component.tempSelectedGroups.filterProperty('hostName', host.hostName);
    if (values.length === 0)
      component.tempSelectedGroups.pushObject({hostName: host.hostName, groupName: groupName});
    else
      values.setEach('groupName', groupName);

  },

  /**
   * add new group to component(click on button)
   */
  addSlaveComponentGroup: function () {
    var component = this.get('selectedSlaveComponent');
    var newGroupName = 'New Group';
    component.get('groups').setEach('active', false);
    var newGroups = component.get('groups').filterProperty('name', newGroupName);
    if (newGroups.length === 0){
      component.newGroupIndex = 0;
    }
    else {
      component.newGroupIndex = component.newGroupIndex || 0;
      this.checkGroupName();
      newGroupName = 'New Group ' + component.newGroupIndex;
    }
    var newGroup = {name: newGroupName, index: component.newGroupIndex, type: 'new', active: true, properties: this.componentProperties(App.router.get('wizardStep7Controller.selectedService.serviceName'))};
    component.groups.pushObject(App.Group.create(newGroup));
    $('.remove-group-error').hide();
  },

  checkGroupName: function () {
    var component = this.get('selectedSlaveComponent');
    component.newGroupIndex++;
    var newGroupName = 'New Group ' + component.newGroupIndex;
    var groups = component.get('groups').filterProperty('name', newGroupName);
    if (groups.length !== 0) {
      this.checkGroupName();
    }
  },

  /**
   * Onclick handler for <code>choose hosts for slave group</code> link
   * @param event
   */
  showEditSlaveComponentGroups: function (event) {
    this.showAddSlaveComponentGroup(event);
  },

  getHostsByGroup: function (group) {
    var hosts = this.get('hosts');
    if(hosts){
      return hosts.filterProperty('group', group.name);
    }
    return [];
  },

  /**
   * Change tab
   * @param event
   */
  showSlaveComponentGroup: function (event) {
    var component = this.get('selectedSlaveComponent');
    if(!component.groups){

    }
    component.get('groups').setEach('active', false);
    var group = component.get('groups').filterProperty('name', event.context.name);
    group.setEach('active', true);
    var assignedHosts = component.get('hosts').filterProperty('group', event.context.name);
    if (assignedHosts.length === 0) {
      $('.remove-group-error').hide();
    }
  },

  /**
   * Remove tab
   * @param event
   */
  removeSlaveComponentGroup: function (event) {
    var group = event.context;
    var component = this.get('selectedSlaveComponent');
    var assignedHosts = component.get('hosts').filterProperty('group', group.name);
    if (assignedHosts.length !== 0) {
      $('.remove-group-error').show();
    } else {
      $('.remove-group-error').hide();
      var key = component.groups.indexOf(group);
      component.groups.removeObject(component.groups[key]);

      var newGroups = component.groups.filterProperty('type', 'new');
      if (newGroups.length == 0)
        component.newGroupIndex = 0;
      else {
        var lastNewGroup = newGroups[newGroups.length - 1];
        component.newGroupIndex = lastNewGroup.index;
      }
      if (group.active) {
        var lastGroup;
        if (key === component.groups.length)
          lastGroup = component.groups.slice(key - 1, key);
        else lastGroup = component.groups.slice(key, key + 1);
        lastGroup.setEach('active', true);
      }
    }
  },

  /**
   * change group name of slave component
   * @param group
   * @param newGroupName
   * @return {Boolean}
   */
  changeSlaveGroupName: function (group, newGroupName) {
    var component = this.get('selectedSlaveComponent');
    var isExist = component.get('groups').filterProperty('name', newGroupName);
    if (isExist.length !== 0)
      return true;
    else {
      var assignedHosts = component.get('hosts').filterProperty('group', group.name);
      if (assignedHosts.length !== 0){
        assignedHosts.setEach('group', newGroupName);
      }
      var groupFilter = component.get('groups').filterProperty('name', group.name);
      groupFilter.setEach('name', newGroupName);
    }
    return false;
  }

});

});

require.register("controllers/wizard/step0_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.WizardStep0Controller = Em.Controller.extend({

  name: 'wizardStep0Controller',

  /**
   * Is step submitted
   * @type {bool}
   */
  hasSubmitted: false,

  /**
   * validate cluster name
   * set <code>clusterNameError</code> if validation fails
   */
  invalidClusterName: function () {
    var clusterName = this.get('content.cluster.name');
    if (clusterName == '' && this.get('hasSubmitted')) {
      this.set('clusterNameError', Em.I18n.t('installer.step0.clusterName.error.required'));
      return true;
    } else if (/\s/.test(clusterName)) {
      this.set('clusterNameError', Em.I18n.t('installer.step0.clusterName.error.whitespace'));
      return true;
    } else if (/[^\w\s]/gi.test(clusterName)) {
      this.set('clusterNameError', Em.I18n.t('installer.step0.clusterName.error.specialChar'));
      return true;
    } else {
      this.set('clusterNameError', '');
      return false;
    }
  }.property('hasSubmitted', 'content.cluster.name'),

  /**
   * calculates by <code>invalidClusterName</code> property
   * todo: mix this and previous variables in one
   */
  clusterNameError: '',

  loadStep: function () {
    this.set('hasSubmitted', false);
    this.set('clusterNameError', '');
  },

  /**
   * Onclick handler for <code>next</code> button
   * @method submit
   */
  submit: function () {
    this.set('hasSubmitted', true);
    if (!this.get('invalidClusterName')) {
      App.clusterStatus.set('clusterName', this.get('content.cluster.name'));
      this.set('content.cluster.status', 'PENDING');
      this.set('content.cluster.isCompleted', false);
      App.router.send('next');
    }
  }

});

});

require.register("controllers/wizard/step10_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.WizardStep10Controller = Em.Controller.extend({

  /**
   * List of data about installed cluster components (hosts, services etc)
   * @type {Ember.Object[]}
   */
  clusterInfo: [],

  /**
   * is Add service wizard the ongoing wizard
   * @type {bool}
   */
  isAddServiceWizard: function () {
    return this.get('content.controllerName') === 'addServiceController';
  }.property('content.controllerName'),

  /**
   * Clear <code>clusterInfo</code>
   * @method clearStep
   */
  clearStep: function () {
    this.get('clusterInfo').clear();
  },

  /**
   * @method loadStep
   */
  loadStep: function () {
    console.log("TRACE: Loading step10: Summary Page");
    this.clearStep();
    this.loadInstalledHosts(this.loadRegisteredHosts());
    var installFlag = true;
    var startFlag = true;
    if (this.get('content.controllerName') == 'installerController') {
      installFlag = this.loadMasterComponents();
      startFlag = this.loadStartedServices();
    }
    if (installFlag && startFlag) {
      this.loadInstallTime();
    }
  },

  /**
   * Get list of clusterInfo object about registered hosts
   * @returns {{id: number, color: string, displayStatement: string, status: array}}
   * @method loadRegisteredHosts
   */
  loadRegisteredHosts: function () {
    var masterHosts = this.get('content.masterComponentHosts').mapProperty('hostName').uniq();
    var slaveHosts = this.get('content.slaveComponentHosts');
    var hostObj = [];
    slaveHosts.forEach(function (_hosts) {
      hostObj = hostObj.concat(_hosts.hosts);
    }, this);
    slaveHosts = hostObj.mapProperty('hostName').uniq();
    var registeredHosts = App.Host.find().mapProperty('hostName').concat(masterHosts.concat(slaveHosts)).uniq();
    var registerHostsStatement = Em.I18n.t('installer.step10.hostsSummary').format(registeredHosts.length);
    var registerHostsObj = Em.Object.create({
      id: 1,
      color: 'text-info',
      displayStatement: registerHostsStatement,
      status: []
    });
    this.get('clusterInfo').pushObject(registerHostsObj);

    return registerHostsObj;
  },

  /**
   * Push info about installed hosts to <code>clusterInfo</code>
   * @method loadInstalledHosts
   */
  loadInstalledHosts: function () {
    var hosts = this.get('content.hosts');
    var hostsInfo = [];
    for (var index in hosts) {
      if (hosts.hasOwnProperty(index)) {
        hostsInfo.pushObject(hosts[index]);
      }
    }
    var succeededHosts = hostsInfo.filterProperty('status', 'success');
    var warnedHosts = hostsInfo.filter(function(host) {
      return ['warning', 'failed'].contains(host.status);
    });
    if (succeededHosts.length) {
      var successStatement = Em.I18n.t('installer.step10.servicesSummary').format(succeededHosts.length) + ((succeededHosts.length > 1) ? Em.I18n.t('installer.step8.hosts') : Em.I18n.t('installer.step8.host'));
      this.get('clusterInfo').findProperty('id', 1).get('status').pushObject(Em.Object.create({
        id: 1,
        color: 'text-success',
        displayStatement: successStatement
      }));
    }

    if (warnedHosts.length) {
      var warnStatement = warnedHosts.length + Em.I18n.t('installer.step10.warnings');
      this.get('clusterInfo').findProperty('id', 1).get('status').pushObject(Em.Object.create({
        id: 2,
        color: 'text-warning',
        displayStatement: warnStatement,
        statements: []
      }));
      warnedHosts.forEach(function (_host) {
        var clusterState = '';
        if (this.get('content.cluster.status') === 'INSTALL FAILED') {
          clusterState = Em.I18n.t('installer.step10.clusterState.installing');
        }
        else {
          if (this.get('content.cluster.status') === 'START FAILED') {
            clusterState = Em.I18n.t('installer.step10.clusterState.starting');
          }
        }

        var self = this;
        Em.A([
            {Tst: 'FAILED', st: 'failed'},
            {Tst: 'ABORTED', st: 'aborted'},
            {Tst: 'TIMEDOUT', st: 'timedout'}
          ]).forEach(function (s) {
            _host.tasks.filterProperty('Tasks.status', s.Tst).forEach(function (_task) {
              var statement = clusterState + App.format.role(_task.Tasks.role) + Em.I18n.t('installer.step10.taskStatus.failed') + _host.name;
              self.get('clusterInfo').findProperty('id', 1).get('status').findProperty('id', 2).get('statements').pushObject(Em.Object.create({
                status: s.st,
                color: 'text-info',
                displayStatement: statement
              }));
            });
          });
      }, this);
    }
  },

  /**
   * Push info about installed/failed master components to <code>clusterInfo</code>
   * @returns {bool}
   * @method loadMasterComponents
   */
  loadMasterComponents: function () {
    var components = this.get('content.masterComponentHosts');
    if (this.get('content.cluster.status') === 'INSTALL FAILED') {
      this.get('clusterInfo').pushObject(Em.Object.create({
        id: 2,
        displayStatement: Em.I18n.t('installer.step10.installStatus.failed'),
        color: 'text-error',
        status: []
      }));
      return false;
    }
    else {
      this.get('clusterInfo').pushObject(Em.Object.create({
        id: 2,
        displayStatement: Em.I18n.t('installer.step10.installStatus.installed'),
        color: 'text-success',
        status: []
      }));
    }

    console.log('STEP10 master components:  ' + JSON.stringify(components));
    components.forEach(function (_component) {
      var component = Em.Object.create(_component);
      if (['NAMENODE', 'SECONDARY_NAMENODE', 'JOBTRACKER', 'HISTORYSERVER', 'RESOURCEMANAGER', 'HBASE_MASTER',
        'HIVE_SERVER', 'OOZIE_SERVER', 'GANGLIA_SERVER'].contains(component.component)) {
        this.loadMasterComponent(component);
      }
    }, this);
    return true;
  },

  /**
   * Push component info to <code>clusterInfo</code>
   * @param {Ember.Object} component
   * @method loadMasterComponent
   */
  loadMasterComponent: function (component) {
    if (component.get('hostName')) {
      var statement = Em.I18n.t('installer.step10.master.installedOn').format(component.get('display_name'), component.get('hostName'));
      this.get('clusterInfo').findProperty('id', 2).get('status').pushObject(Em.Object.create({
        id: 1,
        color: 'text-info',
        displayStatement: statement
      }));
    }
  },

  /**
   * Push info about installed/started/failed services to <code>clusterInfo</code>
   * @returns {bool}
   * @method loadStartedServices
   */
  loadStartedServices: function () {
    if (this.get('content.cluster.status') === 'STARTED') {
      this.get('clusterInfo').pushObject(Em.Object.create({
        id: 3,
        color: 'text-success',
        displayStatement: Em.I18n.t('installer.step10.startStatus.started'),
        status: []
      }));
      this.get('clusterInfo').pushObject(Em.Object.create({
        id: 4,
        color: 'text-success',
        displayStatement: Em.I18n.t('installer.step10.startStatus.passed'),
        status: []
      }));
      return true;
    } else {
      this.get('clusterInfo').pushObject(Em.Object.create({
        id: 3,
        color: 'text-error',
        displayStatement: Em.I18n.t('installer.step10.startStatus.failed'),
        status: []
      }));
      return false;
    }
  },

  /**
   * Push install time to <code>clusterInfo</code>
   * @method loadInstallTime
   */
  loadInstallTime: function () {
    var statement, time;
    if (this.get('content.cluster.installTime')) {
      time = this.calculateInstallTime(this.get('content.cluster.installTime'));
      if (time.minutes !== 0) {
        statement = Em.I18n.t('installer.step10.installTime.minutes').format(time.minutes, time.seconds);
      } else {
        statement = Em.I18n.t('installer.step10.installTime.seconds').format(time.seconds);
      }
      this.get('clusterInfo').pushObject(Em.Object.create({
        id: 5,
        color: 'text-info',
        displayStatement: statement,
        status: []
      }));
    }
  },

  /**
   * Get used time for install process
   * @param {number} installTime
   * @returns {{minutes: *, seconds: *}}
   * @method calculateInstallTime
   */
  calculateInstallTime: function (installTime) {
    var secondsPerMinute = 60;
    var minutes = Math.floor(installTime);
    var seconds = Math.floor((installTime - minutes) * secondsPerMinute);
    return {
      minutes: minutes,
      seconds: seconds
    }
  }

});

});

require.register("controllers/wizard/step1_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.WizardStep1Controller = Em.Controller.extend({

  name: 'wizardStep1Controller',
  /**
   * Skip repo-validation
   * @type {bool}
   */
  skipValidationChecked: false,

  selectedStack: function() {
    return App.Stack.find().findProperty('isSelected');
  }.property('content.stacks.@each.isSelected')


});

});

require.register("controllers/wizard/step2_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var validator = require('utils/validator');
var lazyloading = require('utils/lazy_loading');

App.WizardStep2Controller = Em.Controller.extend({

  name: 'wizardStep2Controller',

  /**
   * List of not installed hostnames
   * @type {string[]}
   */
  hostNameArr: [],

  /**
   * Does pattern-expression for hostnames contains some errors
   * @type {bool}
   */
  isPattern: false,

  /**
   * Don't know if it used any more
   */
  bootRequestId: null,

  /**
   * Is step submitted
   * @type {bool}
   */
  hasSubmitted: false,

  /**
   * @type {string[]}
   */
  inputtedAgainHostNames: [],

  /**
   * Is Installer Controller used
   * @type {bool}
   */
  isInstaller: function () {
    return this.get('content.controllerName') == 'installerController';
  }.property('content.controllerName'),

  /**
   * "Shortcut" to <code>content.installOptions.hostNames</code>
   * @type {string}
   */
  hostNames: function () {
    return this.get('content.installOptions.hostNames').toLowerCase();
  }.property('content.installOptions.hostNames'),

  /**
   * Is manual install selected
   * "Shortcut" to <code>content.installOptions.manualInstall</code>
   * @type {bool}
   */
  manualInstall: function () {
    return this.get('content.installOptions.manualInstall');
  }.property('content.installOptions.manualInstall'),

  /**
   * "Shortcut" to <code>content.installOptions.sshKey</code>
   * @type {string}
   */
  sshKey: function () {
    return this.get('content.installOptions.sshKey');
  }.property('content.installOptions.sshKey'),

  /**
   * "Shortcut" to <code>content.installOptions.sshUser</code>
   * @type {string}
   */
  sshUser: function () {
    return this.get('content.installOptions.sshUser');
  }.property('content.installOptions.sshUser'),

  /**
   * "Shortcut" to <code>content.installOptions.agentUser</code>
   * @type {string}
   */
  agentUser: function () {
    return this.get('content.installOptions.agentUser');
  }.property('content.installOptions.agentUser'),

  /**
   * Installed type based on <code>manualInstall</code>
   * @type {string}
   */
  installType: function () {
    return this.get('manualInstall') ? 'manualDriven' : 'ambariDriven';
  }.property('manualInstall'),

  /**
   * List of invalid hostnames
   * @type {string[]}
   */
  invalidHostNames: [],

  /**
   * Error-message if <code>hostNames</code> is empty, null otherwise
   * @type {string|null}
   */
  hostsError: null,

  isSSHRegistrationEnabled: function () {
    return !App.get('isHadoopWindowsStack');
  }.property('App.isHadoopWindowsStack'),

  /**
   * Error-message if <code>sshKey</code> is empty, null otherwise
   * @type {string|null}
   */
  sshKeyError: function () {
    if (this.get('hasSubmitted') && this.get('manualInstall') === false && Em.isEmpty(this.get('sshKey').trim())) {
      return Em.I18n.t('installer.step2.sshKey.error.required');
    }
    return null;
  }.property('sshKey', 'manualInstall', 'hasSubmitted'),

  /**
   * Error-message if <code>sshUser</code> is empty, null otherwise
   * @type {string|null}
   */
  sshUserError: function () {
    if (this.get('manualInstall') === false && Em.isEmpty(this.get('sshUser').trim())) {
      return Em.I18n.t('installer.step2.sshUser.required');
    }
    return null;
  }.property('sshUser', 'hasSubmitted', 'manualInstall'),

  /**
   * Error-message if <code>agentUser</code> is empty, null otherwise
   * @type {string|null}
   */
  agentUserError: function () {
    if (App.get('supports.customizeAgentUserAccount') && this.get('manualInstall') === false && Em.isEmpty(this.get('agentUser').trim())) {
      return Em.I18n.t('installer.step2.sshUser.required');
    }
    return null;
  }.property('agentUser', 'hasSubmitted', 'manualInstall'),

  /**
   * is Submit button disabled
   * @type {bool}
   */
  isSubmitDisabled: function () {
    return (this.get('hostsError') || this.get('sshKeyError') || this.get('sshUserError') || this.get('agentUserError'));
  }.property('hostsError', 'sshKeyError', 'sshUserError', 'agentUserError'),

  installedHostNames: function () {
    var installedHostsName = [];
    var hosts = this.get('content.hosts');

    for (var hostName in hosts) {
      if (hosts[hostName].isInstalled) {
        installedHostsName.push(hostName);
      }
    }
    return installedHostsName;
  }.property('content.hosts'),

  /**
   * Set not installed hosts to the hostNameArr
   * @method updateHostNameArr
   */
  updateHostNameArr: function () {
    this.set('hostNameArr', this.get('hostNames').trim().split(new RegExp("\\s+", "g")));
    this.parseHostNamesAsPatternExpression();
    this.get('inputtedAgainHostNames').clear();
    var tempArr = [],
        hostNameArr = this.get('hostNameArr');
    for (var i = 0; i < hostNameArr.length; i++) {
      if (!this.get('installedHostNames').contains(hostNameArr[i])) {
        tempArr.push(hostNameArr[i]);
      }
      else {
        this.get('inputtedAgainHostNames').push(hostNameArr[i]);
      }
    }
    this.set('hostNameArr', tempArr);
  },

  /**
   * Validate host names
   * @method isAllHostNamesValid
   * @return {bool}
   */
  isAllHostNamesValid: function () {
    var result = true;
    this.updateHostNameArr();
    this.get('invalidHostNames').clear();
    this.get('hostNameArr').forEach(function (hostName) {
      if (!validator.isHostname(hostName)) {
        this.get('invalidHostNames').push(hostName);
        result = false;
      }
    }, this);

    return result;
  },

  /**
   * Set hostsError if host names don't pass validation
   * @method checkHostError
   */
  checkHostError: function () {
    if (Em.isEmpty(this.get('hostNames').trim())) {
      this.set('hostsError', Em.I18n.t('installer.step2.hostName.error.required'));
    }
    else {
      this.set('hostsError', null);
    }
  },

  /**
   * Check hostnames after Submit was clicked or <code>hostNames</code> were changed
   * @method checkHostAfterSubmitHandler
   */
  checkHostAfterSubmitHandler: function () {
    if (this.get('hasSubmitted')) {
      this.checkHostError();
    }
  }.observes('hasSubmitted', 'hostNames'),

  /**
   * Get host info, which will be saved in parent controller
   * @method getHostInfo
   */
  getHostInfo: function () {

    var hostNameArr = this.get('hostNameArr');
    var hostInfo = {};
    for (var i = 0; i < hostNameArr.length; i++) {
      hostInfo[hostNameArr[i]] = {
        name: hostNameArr[i],
        installType: this.get('installType'),
        bootStatus: 'PENDING',
        isInstalled: false
      };
    }

    return hostInfo;
  },

  /**
   * Used to set sshKey from FileUploader
   * @method setSshKey
   * @param {string} sshKey
   */
  setSshKey: function (sshKey) {
    this.set("content.installOptions.sshKey", sshKey);
  },

  /**
   * Onclick handler for <code>next button</code>. Do all UI work except data saving.
   * This work is doing by router.
   * @method evaluateStep
   * @return {bool}
   */
  evaluateStep: function () {
    console.log('TRACE: Entering controller:WizardStep2:evaluateStep function');

    if (this.get('isSubmitDisabled')) {
      return false;
    }

    this.set('hasSubmitted', true);

    this.checkHostError();
    this.updateHostNameArr();

    if (!this.get('hostNameArr.length')) {
      this.set('hostsError', Em.I18n.t('installer.step2.hostName.error.already_installed'));
    }

    if (this.get('hostsError') || this.get('sshUserError') || this.get('agentUserError') || this.get('sshKeyError')) {
      return false;
    }

    if (this.get('isPattern')) {
      this.hostNamePatternPopup(this.get('hostNameArr'));
      return false;
    }
    if (this.get('inputtedAgainHostNames.length')) {
      this.installedHostsPopup();
    }
    else {
      this.proceedNext();
    }
    return true;
  },

  /**
   * check is there a pattern expression in host name textarea
   * push hosts that match pattern in hostNamesArr
   * @method parseHostNamesAsPatternExpression
   */
  parseHostNamesAsPatternExpression: function () {
    this.set('isPattern', false);
    var self = this;
    var hostNames = [];
    $.each(this.get('hostNameArr'), function (e, a) {
      var start, end, extra = {0: ""};
      if (/\[\d*\-\d*\]/.test(a)) {
        start = a.match(/\[\d*/);
        end = a.match(/\-\d*]/);

        start = start[0].substr(1);
        end = end[0].substr(1);

        if (parseInt(start) <= parseInt(end, 10) && parseInt(start, 10) >= 0) {
          self.set('isPattern', true);

          if (start[0] == "0" && start.length > 1) {
            extra = start.match(/0*/);
          }

          for (var i = parseInt(start, 10); i < parseInt(end, 10) + 1; i++) {
            hostNames.push(a.replace(/\[\d*\-\d*\]/, extra[0].substring(0, start.length - i.toString().length) + i))
          }

        } else {
          hostNames.push(a);
        }
      } else {
        hostNames.push(a);
      }
    });
    this.set('hostNameArr', hostNames);
  },

  /**
   * launch hosts to bootstrap
   * and save already registered hosts
   * @method proceedNext
   * @return {bool}
   */
  proceedNext: function (warningConfirmed) {
    if (this.isAllHostNamesValid() !== true && !warningConfirmed) {
      this.warningPopup();
      return false;
    }

    if (this.get('manualInstall') === true) {
      this.manualInstallPopup();
      return false;
    }

    this.saveHosts();
    App.router.send('next');
    return true;
  },

  /**
   * show warning for host names without dots or IP addresses
   * @return {App.ModalPopup}
   * @method warningPopup
   */
  warningPopup: function () {
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('common.warning'),
      onPrimary: function () {
        this.hide();
        self.proceedNext(true);
      },
      bodyClass: Em.View.extend({
        template: Em.Handlebars.compile(Em.I18n.t('installer.step2.warning.popup.body').format(self.get('invalidHostNames').join(', ')))
      })
    });
  },

  /**
   * show popup with the list of hosts that are already part of the cluster
   * @return {App.ModalPopup}
   * @method installedHostsPopup
   */
  installedHostsPopup: function () {
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('common.warning'),
      onPrimary: function () {
        self.proceedNext();
        this.hide();
      },
      bodyClass: Em.View.extend({
        inputtedAgainHostNames: function () {
          return self.get('inputtedAgainHostNames').join(', ');
        }.property(),
        templateName: require('templates/wizard/step2_installed_hosts_popup')
      })
    });
  },

  /**
   * Show popup with hosts generated by pattern
   * @method hostNamePatternPopup
   * @param {string[]} hostNames
   * @return {App.ModalPopup}
   */
  hostNamePatternPopup: function (hostNames) {
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('installer.step2.hostName.pattern.header'),
      onPrimary: function () {
        self.proceedNext();
        this.hide();
      },
      bodyClass: Em.View.extend({
        templateName: require('templates/common/items_list_popup'),
        items: hostNames,
        insertedItems: [],
        didInsertElement: function () {
          lazyloading.run({
            destination: this.get('insertedItems'),
            source: this.get('items'),
            context: this,
            initSize: 100,
            chunkSize: 500,
            delay: 100
          });
        }
      })
    });
  },

  /**
   * Show notify that installation is manual
   * save hosts
   * @return {App.ModalPopup}
   * @method manualInstallPopup
   */
  manualInstallPopup: function () {
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('installer.step2.manualInstall.popup.header'),
      onPrimary: function () {
        this.hide();
        self.saveHosts();
        App.router.send('next');
      },
      bodyClass: Em.View.extend({
        templateName: require('templates/wizard/step2ManualInstallPopup')
      })
    });
  },

  /**
   * Warn to manually install ambari-agent on each host
   * @method manualInstallWarningPopup
   */
  manualInstallWarningPopup: function () {
    if (this.get('isSSHRegistrationEnabled') && !this.get('content.installOptions.useSsh')) {
      App.ModalPopup.show({
        header: Em.I18n.t('common.warning'),
        body: Em.I18n.t('installer.step2.manualInstall.info'),
        encodeBody: false,
        secondary: null
      });
    }
    this.set('content.installOptions.manualInstall', !this.get('content.installOptions.useSsh'));
  }.observes('content.installOptions.useSsh'),

  /**
   * Load java.home value frin server
   * @method setAmbariJavaHome
   */
  setAmbariJavaHome: function () {
    App.ajax.send({
      name: 'ambari.service',
      sender: this,
      success: 'onGetAmbariJavaHomeSuccess',
      error: 'onGetAmbariJavaHomeError'
    });
  },

  /**
   * Set received java.home value
   * @method onGetAmbariJavaHomeSuccess
   * @param {Object} data
   */
  onGetAmbariJavaHomeSuccess: function (data) {
    this.set('content.installOptions.javaHome', data.RootServiceComponents.properties['java.home']);
  },

  /**
   * Set default java.home value
   * @method onGetAmbariJavaHomeError
   */
  onGetAmbariJavaHomeError: function () {
    console.warn('can\'t get java.home value from server');
    this.set('content.installOptions.javaHome', App.get('defaultJavaHome'));
  },

  /**
   * Save hosts info and proceed to the next step
   * @method saveHosts
   */
  saveHosts: function () {
    var hosts = this.get('content.hosts');

    //add previously installed hosts
    for (var hostName in hosts) {
      if (!hosts[hostName].isInstalled) {
        delete hosts[hostName];
      }
    }

    this.set('content.hosts', $.extend(hosts, this.getHostInfo()));
    this.setAmbariJavaHome();
  }

});

});

require.register("controllers/wizard/step3_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var lazyloading = require('utils/lazy_loading');
var numberUtils = require('utils/number_utils');

App.WizardStep3Controller = Em.Controller.extend({

  name: 'wizardStep3Controller',

  hosts: [],

  content: [],

  bootHosts: [],

  registeredHosts: [],

  hostCheckWarnings: [],
  repoCategoryWarnings: [],
  diskCategoryWarnings: [],
  thpCategoryWarnings: [],
  jdkCategoryWarnings: null,
  jdkRequestIndex: null,

  registrationStartedAt: null,

  requestId: 0,

  /**
   * Timeout for registration
   * Based on <code>installOptions.manualInstall</code>
   * @type {number}
   */
  registrationTimeoutSecs: function () {
    return this.get('content.installOptions.manualInstall') ? 15 : 120;
  }.property('content.installOptions.manualInstall'),

  /**
   * Bootstrap calls are stopped
   * @type {bool}
   */
  stopBootstrap: false,

  /**
   * is Submit button disabled
   * @type {bool}
   */
  isSubmitDisabled: true,

  /**
   * is Retry button disabled
   * @type {bool}
   */
  isRetryDisabled: true,

  /**
   * Is Back button disabled
   * @return {bool}
   */
  isBackDisabled: function () {
    return this.get('isRegistrationInProgress') || !this.get('isWarningsLoaded');
  }.property('isRegistrationInProgress', 'isWarningsLoaded'),

  /**
   * @type {bool}
   */
  isLoaded: false,

  /**
   * Polls count
   * @type {number}
   */
  numPolls: 0,

  /**
   * Is hosts registration in progress
   * @type {bool}
   */
  isRegistrationInProgress: true,

  /**
   * Are some registered hosts which are not added by user
   * @type {bool}
   */
  hasMoreRegisteredHosts: false,

  /**
   * Contain data about installed packages on hosts
   * @type {Array}
   */
  hostsPackagesData: [],

  /**
   * List of installed hostnames
   * @type {string[]}
   */
  hostsInCluster: function () {
    var installedHostsName = [];
    var hosts = this.get('content.hosts');

    for (var hostName in hosts) {
      if (hosts[hostName].isInstalled) {
        installedHostsName.push(hostName);
      }
    }
    return installedHostsName;
  }.property('content.hosts'),

  /**
   * All hosts warnings
   * @type {object[]}
   */
  warnings: [],

  /**
   * Warnings grouped by host
   * @type {Ember.Enumerable}
   */
  warningsByHost: [],

  /**
   * Timeout for "warning"-requests
   * @type {number}
   */
  warningsTimeInterval: 60000,

  /**
   * Are hosts warnings loaded
   * @type {bool}
   */
  isWarningsLoaded: function () {
    return this.get('isJDKWarningsLoaded') && this.get('isHostsWarningsLoaded');
  }.property('isJDKWarningsLoaded', 'isHostsWarningsLoaded'),

  /**
   * Check are hosts have any warnings
   * @type {bool}
   */
  isHostHaveWarnings: function () {
    return this.get('warnings.length') > 0;
  }.property('warnings'),

  /**
   * Should warnings-box be visible
   * @type {bool}
   */
  isWarningsBoxVisible: function () {
    return (App.get('testMode')) ? true : !this.get('isRegistrationInProgress');
  }.property('isRegistrationInProgress'),

  /**
   * Progress value for "update hosts status" process
   * @type {number}
   */
  checksUpdateProgress: 0,

  /**
   *
   * @type {object}
   */
  checksUpdateStatus: null,

  /**
   *
   * @method navigateStep
   */
  navigateStep: function () {
    if (this.get('isLoaded')) {
      if (!this.get('content.installOptions.manualInstall')) {
        if (!this.get('wizardController').getDBProperty('bootStatus')) {
          this.setupBootStrap();
        }
      } else {
        this.set('bootHosts', this.get('hosts'));
        if (App.get('testMode')) {
          this.startHostcheck();
          this.get('bootHosts').setEach('cpu', '2');
          this.get('bootHosts').setEach('memory', '2000000');
          this.set('isSubmitDisabled', false);
        } else {
          this.set('registrationStartedAt', null);
          this.startRegistration();
        }
      }
    }
  }.observes('isLoaded'),

  /**
   * Clear controller data
   * @method clearStep
   */
  clearStep: function () {
    this.set('stopBootstrap', false);
    this.set('hosts', []);
    this.get('bootHosts').clear();
    this.get('wizardController').setDBProperty('bootStatus', false);
    this.set('isHostsWarningsLoaded', false);
    this.set('isJDKWarningsLoaded', false);
    this.set('registrationStartedAt', null);
    this.set('isLoaded', false);
    this.set('isSubmitDisabled', true);
    this.set('isRetryDisabled', true);
    this.set('stopChecking', false);
  },

  /**
   * setup bootstrap data and completion callback for bootstrap call
   * @method setupBootStrap
   */
  setupBootStrap: function () {
    var self = this;
    var bootStrapData = JSON.stringify({
        'verbose': true,
        'sshKey': this.get('content.installOptions.sshKey'),
        'hosts': Em.keys(this.get('content.hosts')),
        'user': this.get('content.installOptions.sshUser'),
        'userRunAs': App.get('supports.customizeAgentUserAccount') ? this.get('content.installOptions.agentUser') : 'root'
    });
    App.router.get(this.get('content.controllerName')).launchBootstrap(bootStrapData, function (requestId) {
      if (requestId == '0') {
        var controller = App.router.get(App.clusterStatus.wizardControllerName);
        controller.registerErrPopup(Em.I18n.t('common.information'), Em.I18n.t('installer.step2.evaluateStep.hostRegInProgress'));
      } else if (requestId) {
        self.set('content.installOptions.bootRequestId', requestId);
        self.startBootstrap();
      }
    });
  },

  /**
   * Make basic init steps
   * @method loadStep
   */
  loadStep: function () {
    console.log("TRACE: Loading step3: Confirm Hosts");
    this.disablePreviousSteps();
    this.clearStep();
    App.router.get('clusterController').loadAmbariProperties();
    this.loadHosts();
  },

  /**
   * Loads the hostinfo from localStorage on the insertion of view. It's being called from view
   * @method loadHosts
   */
  loadHosts: function () {
    var hostsInfo = this.get('content.hosts');
    var hosts = [];
    var bootStatus = (this.get('content.installOptions.manualInstall')) ? 'DONE' : 'PENDING';
    if (App.get('testMode')) {
      bootStatus = 'REGISTERED';
    }

    for (var index in hostsInfo) {
      if (hostsInfo.hasOwnProperty(index) && !hostsInfo[index].isInstalled) {
        hosts.pushObject(App.HostInfo.create({
          name: hostsInfo[index].name,
          bootStatus: bootStatus,
          isChecked: false
        }));
      }
    }
    this.set('hosts', hosts);
    this.set('isLoaded', true);
  },

  /**
   * Parses and updates the content based on bootstrap API response.
   * @return {bool} true if polling should continue (some hosts are in "RUNNING" state); false otherwise
   * @method parseHostInfo
   */
  parseHostInfo: function (hostsStatusFromServer) {
    hostsStatusFromServer.forEach(function (_hostStatus) {
      var host = this.get('bootHosts').findProperty('name', _hostStatus.hostName);
      // check if hostname extracted from REST API data matches any hostname in content
      // also, make sure that bootStatus modified by isHostsRegistered call does not get overwritten
      // since these calls are being made in parallel
      if (host && !['REGISTERED', 'REGISTERING'].contains(host.get('bootStatus'))) {
        host.set('bootStatus', _hostStatus.status);
        host.set('bootLog', _hostStatus.log);
      }
    }, this);
    // if the data rendered by REST API has hosts in "RUNNING" state, polling will continue
    return this.get('bootHosts').length != 0 && this.get('bootHosts').someProperty('bootStatus', 'RUNNING');
  },

  /**
   * Remove list of hosts
   * @param {Ember.Enumerable} hosts
   * @return {App.ModalPopup}
   * @method removeHosts
   */
  removeHosts: function (hosts) {
    var self = this;
    return App.showConfirmationPopup(function () {
      App.router.send('removeHosts', hosts);
      self.hosts.removeObjects(hosts);
      if (!self.hosts.length) {
        self.set('isSubmitDisabled', true);
      }
    }, Em.I18n.t('installer.step3.hosts.remove.popup.body'));
  },

  /**
   * Removes a single element on the trash icon click. Called from View
   * @param {object} hostInfo
   * @method removeHost
   */
  removeHost: function (hostInfo) {
    this.removeHosts([hostInfo]);
  },

  /**
   * Remove selected hosts (click-handler)
   * @return App.ModalPopup
   * @method removeSelectedHosts
   */
  removeSelectedHosts: function () {
    var selectedHosts = this.get('hosts').filterProperty('isChecked', true);
    selectedHosts.forEach(function (_hostInfo) {
      console.log('Removing:  ' + _hostInfo.name);
    });
    return this.removeHosts(selectedHosts);
  },

  /**
   * Show popup with the list of hosts which are selected
   * @return App.ModalPopup
   * @method selectedHostsPopup
   */
  selectedHostsPopup: function () {
    var selectedHosts = this.get('hosts').filterProperty('isChecked').mapProperty('name');
    return App.ModalPopup.show({
      header: Em.I18n.t('installer.step3.selectedHosts.popup.header'),
      secondary: null,
      bodyClass: Em.View.extend({
        templateName: require('templates/common/items_list_popup'),
        items: selectedHosts,
        insertedItems: [],
        didInsertElement: function () {
          lazyloading.run({
            destination: this.get('insertedItems'),
            source: this.get('items'),
            context: this,
            initSize: 100,
            chunkSize: 500,
            delay: 100
          });
        }
      })
    });
  },

  /**
   * Retry one host {click-handler}
   * @param {object} hostInfo
   * @method retryHost
   */
  retryHost: function (hostInfo) {
    this.retryHosts([hostInfo]);
  },

  /**
   * Retry list of hosts
   * @param {object[]} hosts
   * @method retryHosts
   */
  retryHosts: function (hosts) {
    var self = this;
    var bootStrapData = JSON.stringify({
        'verbose': true,
        'sshKey': this.get('content.installOptions.sshKey'),
        'hosts': hosts.mapProperty('name'),
        'user': this.get('content.installOptions.sshUser'),
        'userRunAs': App.get('supports.customizeAgentUserAccount') ? this.get('content.installOptions.agentUser') : 'root'
      });
    this.set('numPolls', 0);
    this.set('registrationStartedAt', null);
    this.set('isHostsWarningsLoaded', false);
    this.set('stopChecking', false);
    if (this.get('content.installOptions.manualInstall')) {
      this.startRegistration();
    } else {
      App.router.get(this.get('content.controllerName')).launchBootstrap(bootStrapData, function (requestId) {
        self.set('content.installOptions.bootRequestId', requestId);
        self.doBootstrap();
      });
    }
  },

  /**
   * Retry selected hosts (click-handler)
   * @method retrySelectedHosts
   */
  retrySelectedHosts: function () {
    if (!this.get('isRetryDisabled')) {
      this.set('isRetryDisabled', true);
      var selectedHosts = this.get('bootHosts').filterProperty('bootStatus', 'FAILED');
      selectedHosts.forEach(function (_host) {
        _host.set('bootStatus', 'DONE');
        _host.set('bootLog', 'Retrying ...');
      }, this);
      this.retryHosts(selectedHosts);
    }
  },

  /**
   * Init bootstrap settings and start it
   * @method startBootstrap
   */
  startBootstrap: function () {
    //this.set('isSubmitDisabled', true);    //TODO: uncomment after actual hookup
    this.set('numPolls', 0);
    this.set('registrationStartedAt', null);
    this.set('bootHosts', this.get('hosts'));
    this.doBootstrap();
  },

  /**
   * Update <code>isRegistrationInProgress</code> once
   * @method setRegistrationInProgressOnce
   */
  setRegistrationInProgressOnce: function () {
    Em.run.once(this, 'setRegistrationInProgress');
  }.observes('bootHosts.@each.bootStatus'),

  /**
   * Set <code>isRegistrationInProgress</code> value based on each host boot status
   * @method setRegistrationInProgress
   */
  setRegistrationInProgress: function () {
    var bootHosts = this.get('bootHosts');
    //if hosts aren't loaded yet then registration should be in progress
    var result = (bootHosts.length === 0 && !this.get('isLoaded'));
    for (var i = 0, l = bootHosts.length; i < l; i++) {
      if (bootHosts[i].get('bootStatus') !== 'REGISTERED' && bootHosts[i].get('bootStatus') !== 'FAILED') {
        result = true;
        break;
      }
    }
    this.set('isRegistrationInProgress', result);
  },

  /**
   * Disable wizard's previous steps (while registering)
   * @method disablePreviousSteps
   */
  disablePreviousSteps: function () {
    App.router.get('installerController.isStepDisabled').filter(function (step) {
      return step.step >= 0 && step.step <= 2;
    }).setEach('value', this.get('isBackDisabled'));
    if (this.get('isBackDisabled')) {
      this.set('isSubmitDisabled', true);
    }
  }.observes('isBackDisabled'),

  /**
   * Do bootstrap calls
   * @method doBootstrap
   * @return {$.ajax|null}
   */
  doBootstrap: function () {
    if (this.get('stopBootstrap')) {
      return null;
    }
    this.incrementProperty('numPolls');

    return App.ajax.send({
      name: 'wizard.step3.bootstrap',
      sender: this,
      data: {
        bootRequestId: this.get('content.installOptions.bootRequestId'),
        numPolls: this.get('numPolls')
      },
      success: 'doBootstrapSuccessCallback'
    }).
      retry({
        times: App.maxRetries,
        timeout: App.timeout
      }).
      then(
      null,
      function () {
        App.showReloadPopup();
        console.log('Bootstrap failed');
      }
    );
  },

  /**
   * Success-callback for each boostrap request
   * @param {object} data
   * @method doBootstrapSuccessCallback
   */
  doBootstrapSuccessCallback: function (data) {
    var self = this;
    var pollingInterval = 3000;
    if (Em.isNone(data.hostsStatus)) {
      console.log('Invalid response, setting timeout');
      window.setTimeout(function () {
        self.doBootstrap()
      }, pollingInterval);
    } else {
      // in case of bootstrapping just one host, the server returns an object rather than an array, so
      // force into an array
      if (!(data.hostsStatus instanceof Array)) {
        data.hostsStatus = [ data.hostsStatus ];
      }
      console.log("TRACE: In success function for the GET bootstrap call");
      var keepPolling = this.parseHostInfo(data.hostsStatus);

      // Single host : if the only hostname is invalid (data.status == 'ERROR')
      // Multiple hosts : if one or more hostnames are invalid
      // following check will mark the bootStatus as 'FAILED' for the invalid hostname
      if (data.status == 'ERROR' || data.hostsStatus.length != this.get('bootHosts').length) {

        var hosts = this.get('bootHosts');

        for (var i = 0; i < hosts.length; i++) {

          var isValidHost = data.hostsStatus.someProperty('hostName', hosts[i].get('name'));
          if (hosts[i].get('bootStatus') !== 'REGISTERED') {
            if (!isValidHost) {
              hosts[i].set('bootStatus', 'FAILED');
              hosts[i].set('bootLog', Em.I18n.t('installer.step3.hosts.bootLog.failed'));
            }
          }
        }
      }

      if (data.status == 'ERROR' || data.hostsStatus.someProperty('status', 'DONE') || data.hostsStatus.someProperty('status', 'FAILED')) {
        // kicking off registration polls after at least one host has succeeded
        this.startRegistration();
      }
      if (keepPolling) {
        window.setTimeout(function () {
          self.doBootstrap()
        }, pollingInterval);
      }
    }
  },

  /**
   * Start hosts registration
   * @method startRegistration
   */
  startRegistration: function () {
    if (Em.isNone(this.get('registrationStartedAt'))) {
      this.set('registrationStartedAt', App.dateTime());
      console.log('registration started at ' + this.get('registrationStartedAt'));
      this.isHostsRegistered();
    }
  },

  /**
   * Do requests to check if hosts are already registered
   * @return {$.ajax|null}
   * @method isHostsRegistered
   */
  isHostsRegistered: function () {
    if (this.get('stopBootstrap')) {
      return null;
    }
    return App.ajax.send({
      name: 'wizard.step3.is_hosts_registered',
      sender: this,
      success: 'isHostsRegisteredSuccessCallback'
    }).
      retry({
        times: App.maxRetries,
        timeout: App.timeout
      }).
      then(
      null,
      function () {
        App.showReloadPopup();
        console.log('Error: Getting registered host information from the server');
      }
    );
  },

  /**
   * Success-callback for registered hosts request
   * @param {object} data
   * @method isHostsRegisteredSuccessCallback
   */
  isHostsRegisteredSuccessCallback: function (data) {
    console.log('registration attempt...');
    var hosts = this.get('bootHosts');
    var jsonData = data;
    if (!jsonData) {
      console.warn("Error: jsonData is null");
      return;
    }

    // keep polling until all hosts have registered/failed, or registrationTimeout seconds after the last host finished bootstrapping
    var stopPolling = true;
    hosts.forEach(function (_host, index) {
      // Change name of first host for test mode.
      if (App.get('testMode')) {
        if (index == 0) {
          _host.set('name', 'localhost.localdomain');
        }
      }
      // actions to take depending on the host's current bootStatus
      // RUNNING - bootstrap is running; leave it alone
      // DONE - bootstrap is done; transition to REGISTERING
      // REGISTERING - bootstrap is done but has not registered; transition to REGISTERED if host found in polling API result
      // REGISTERED - bootstrap and registration is done; leave it alone
      // FAILED - either bootstrap or registration failed; leave it alone
      switch (_host.get('bootStatus')) {
        case 'DONE':
          _host.set('bootStatus', 'REGISTERING');
          _host.set('bootLog', (_host.get('bootLog') != null ? _host.get('bootLog') : '') + Em.I18n.t('installer.step3.hosts.bootLog.registering'));
          // update registration timestamp so that the timeout is computed from the last host that finished bootstrapping
          this.set('registrationStartedAt', App.dateTime());
          stopPolling = false;
          break;
        case 'REGISTERING':
          if (jsonData.items.someProperty('Hosts.host_name', _host.name) && !jsonData.items.filterProperty('Hosts.host_name', _host.name).someProperty('Hosts.host_status', 'UNKNOWN')) {
            _host.set('bootStatus', 'REGISTERED');
            _host.set('bootLog', (_host.get('bootLog') != null ? _host.get('bootLog') : '') + Em.I18n.t('installer.step3.hosts.bootLog.registering'));
          } else {
            stopPolling = false;
          }
          break;
        case 'RUNNING':
          stopPolling = false;
          break;
        case 'REGISTERED':
        case 'FAILED':
        default:
          break;
      }
    }, this);

    if (stopPolling) {
      this.startHostcheck();
    }
    else {
      if (hosts.someProperty('bootStatus', 'RUNNING') || App.dateTime() - this.get('registrationStartedAt') < this.get('registrationTimeoutSecs') * 1000) {
        // we want to keep polling for registration status if any of the hosts are still bootstrapping (so we check for RUNNING).
        var self = this;
        window.setTimeout(function () {
          self.isHostsRegistered();
        }, 3000);
      }
      else {
        // registration timed out.  mark all REGISTERING hosts to FAILED
        console.log('registration timed out');
        hosts.filterProperty('bootStatus', 'REGISTERING').forEach(function (_host) {
          _host.set('bootStatus', 'FAILED');
          _host.set('bootLog', (_host.get('bootLog') != null ? _host.get('bootLog') : '') + Em.I18n.t('installer.step3.hosts.bootLog.failed'));
        });
        this.startHostcheck();
      }
    }
  },

  /**
   * Do request for all registered hosts
   * @return {$.ajax}
   * @method getAllRegisteredHosts
   */
  getAllRegisteredHosts: function () {
    return App.ajax.send({
      name: 'wizard.step3.is_hosts_registered',
      sender: this,
      success: 'getAllRegisteredHostsCallback'
    });
  }.observes('bootHosts'),

  /**
   * Success-callback for all registered hosts request
   * @param {object} hosts
   * @method getAllRegisteredHostsCallback
   */
  getAllRegisteredHostsCallback: function (hosts) {
    var registeredHosts = [];
    var hostsInCluster = this.get('hostsInCluster');
    var addedHosts = this.get('bootHosts').getEach('name');
    hosts.items.forEach(function (host) {
      if (!hostsInCluster.contains(host.Hosts.host_name) && !addedHosts.contains(host.Hosts.host_name)) {
        registeredHosts.push(host.Hosts.host_name);
      }
    });
    if (registeredHosts.length) {
      this.set('hasMoreRegisteredHosts', true);
      this.set('registeredHosts', registeredHosts);
    } else {
      this.set('hasMoreRegisteredHosts', false);
      this.set('registeredHosts', '');
    }
  },

  /**
   * Show popup with regitration error-message
   * @param {string} header
   * @param {string} message
   * @return {App.ModalPopup}
   * @method registerErrPopup
   */
  registerErrPopup: function (header, message) {
    return App.ModalPopup.show({
      header: header,
      secondary: false,
      bodyClass: Em.View.extend({
        template: Em.Handlebars.compile('<p>{{view.message}}</p>'),
        message: message
      })
    });
  },

  /**
   * Get JDK name from server to determine if user had setup a customized JDK path when doing 'ambari-server setup'.
   * The Ambari properties are different from default ambari-server setup, property 'jdk.name' will be missing if a customized jdk path is applied.
   * @return {$.ajax}
   * @method getJDKName
   */
  getJDKName: function () {
    return App.ajax.send({
      name: 'ambari.service.load_jdk_name',
      sender: this,
      success: 'getJDKNameSuccessCallback'
    });
  },

  /**
    * Success callback for JDK name, property 'jdk.name' will be missing if a customized jdk path is applied
    * @param {object} data
    * @method getJDKNameSuccessCallback
    */
  getJDKNameSuccessCallback: function (data) {
    this.set('needJDKCheckOnHosts', !data.RootServiceComponents.properties["jdk.name"]);
    this.set('jdkLocation', Em.get(data, "RootServiceComponents.properties.jdk_location"));
    this.set('javaHome', data.RootServiceComponents.properties["java.home"]);
  },

  doCheckJDK: function () {
    var hostsNames = (!this.get('content.installOptions.manualInstall')) ? this.get('bootHosts').filterProperty('bootStatus', 'REGISTERED').getEach('name').join(",") : this.get('bootHosts').getEach('name').join(",");
    var javaHome = this.get('javaHome');
    var jdkLocation = this.get('jdkLocation');
    App.ajax.send({
      name: 'wizard.step3.jdk_check',
      sender: this,
      data: {
        host_names: hostsNames,
        java_home: javaHome,
        jdk_location: jdkLocation
      },
      success: 'doCheckJDKsuccessCallback',
      error: 'doCheckJDKerrorCallback'
    });
  },

  doCheckJDKsuccessCallback: function (data) {
    if(data){
      this.set('jdkRequestIndex', data.href.split('/')[data.href.split('/').length - 1]);
    }
    if (this.get('jdkCategoryWarnings') == null) {
      // get jdk check results for all hosts
      App.ajax.send({
        name: 'wizard.step3.jdk_check.get_results',
        sender: this,
        data: {
          requestIndex: this.get('jdkRequestIndex')
        },
        success: 'parseJDKCheckResults'
      })
    } else {
      this.set('isJDKWarningsLoaded', true);
    }
  },
  doCheckJDKerrorCallback: function () {
    console.log('INFO: Doing JDK check for host failed');
    this.set('isJDKWarningsLoaded', true);
  },
  parseJDKCheckResults: function (data) {
    var jdkWarnings = [], hostsJDKContext = [], hostsJDKNames = [];
    // check if the request ended
    if (data.Requests.end_time > 0 && data.tasks) {
      data.tasks.forEach( function(task) {
        // generate warning context
        if (Em.get(task, "Tasks.structured_out.java_home_check.exit_code") == 1){
          var jdkContext = Em.I18n.t('installer.step3.hostWarningsPopup.jdk.context').format(task.Tasks.host_name);
          hostsJDKContext.push(jdkContext);
          hostsJDKNames.push(task.Tasks.host_name);
        }
      });
      if (hostsJDKContext.length > 0) { // java jdk warning exist
        var invalidJavaHome = this.get('javaHome');
        jdkWarnings.push({
          name: Em.I18n.t('installer.step3.hostWarningsPopup.jdk.name').format(invalidJavaHome),
          hosts: hostsJDKContext,
          hostsNames: hostsJDKNames,
          category: 'jdk',
          onSingleHost: false
        });
      }
      this.set('jdkCategoryWarnings', jdkWarnings);
    } else {
      // still doing JDK check, data not ready to be parsed
      this.set('jdkCategoryWarnings', null);
    }
    this.doCheckJDKsuccessCallback();
  },

  /**
   * Check JDK issues on registered hosts.
   */
  checkHostJDK: function () {
    this.set('isJDKWarningsLoaded', false);
    this.set('jdkCategoryWarnings', null);
    var self = this;
    this.getJDKName().done( function() {
      if (self.get('needJDKCheckOnHosts')) {
        // need to do JDK check on each host
       self.doCheckJDK();
      } else {
        // no customized JDK path, so no need to check jdk
        self.set('jdkCategoryWarnings', []);
        self.set('isJDKWarningsLoaded', true);
      }
    });
  },

  /**
   * Get disk info and cpu count of booted hosts from server
   * @return {$.ajax}
   * @method getHostInfo
   */
  getHostInfo: function () {
    this.set('isHostsWarningsLoaded', false);
    // begin JDK check for each host
    return App.ajax.send({
      name: 'wizard.step3.host_info',
      sender: this,
      success: 'getHostInfoSuccessCallback',
      error: 'getHostInfoErrorCallback'
    });
  },


  startHostcheck: function() {
    this.set('isWarningsLoaded', false);
    this.getHostNameResolution();
    this.checkHostJDK();
  },

  getHostNameResolution: function () {
    if (App.get('testMode')) {
      this.getHostCheckSuccess();
    } else {
      var data = this.getDataForCheckRequest("host_resolution_check", true);
      data ? this.requestToPerformHostCheck(data) : this.stopHostCheck();
    }
  },

  getGeneralHostCheck: function () {
    if (App.get('testMode')) {
      this.getHostInfo();
    } else {
      var data = this.getDataForCheckRequest("last_agent_env_check,installed_packages,existing_repos", false);
      data ? this.requestToPerformHostCheck(data) : this.stopHostCheck();
    }
  },

  /**
   * set all fields from which depends running host check to true value
   * which force finish checking;
   */
  stopHostCheck: function() {
    this.set('stopChecking', true);
    this.set('isJDKWarningsLoaded', true);
    this.set('isHostsWarningsLoaded', true);
  },

  getHostCheckSuccess: function(response) {
    if (!App.get('testMode')) {
      this.set("requestId", response.Requests.id);
    }
    this.getHostCheckTasks();
  },

  /**
   * generates data for reuest to perform check
   * @param {string} checkExecuteList - for now supported:
   *  <code>"last_agent_env_check"<code>
   *  <code>"host_resolution_check"<code>
   * @param {boolean} addHosts - true
   * @return {object|null}
   * @method getDataForCheckRequest
   */
  getDataForCheckRequest: function (checkExecuteList, addHosts) {
    var hosts = this.get('bootHosts').filterProperty('bootStatus', 'REGISTERED').getEach('name').join(",");
    if (hosts.length == 0) return null;
    var jdk_location = App.router.get('clusterController.ambariProperties.jdk_location');
    var RequestInfo = {
      "action": "check_host",
      "context": "Check host",
      "parameters": {
        "check_execute_list": checkExecuteList,
        "jdk_location" : jdk_location,
        "threshold": "20"
      }
    };
    if (addHosts) {
      RequestInfo.parameters.hosts = hosts;
    }
    var resource_filters = {
      "hosts": hosts
    };
    return {
      RequestInfo: RequestInfo,
      resource_filters: resource_filters
    }
  },

  /**
   * send request to ceate tasks for performing hosts checks
   * @params {object} data
   *    {
   *       RequestInfo: {
   *           "action": {string},
   *           "context": {string},
   *           "parameters": {
   *             "check_execute_list": {string},
   *             "jdk_location" : {string},
   *             "threshold": {string}
   *             "hosts": {string|undefined}
   *       },
   *       resource_filters: {
   *         "hosts": {string}
   *       }
   *    }
   * @returns {$.ajax}
   * @method requestToPerformHostCheck
   */
  requestToPerformHostCheck: function(data) {
    return App.ajax.send({
      name: 'preinstalled.checks',
      sender: this,
      data: {
        RequestInfo: data.RequestInfo,
        resource_filters: data.resource_filters
      },
      success: "getHostCheckSuccess",
      error: "getHostCheckError"
    })
  },

  /**
   * send ajax request to get all tasks
   * @method getHostCheckTasks
   */
  getHostCheckTasks: function () {
    var self = this;
    var requestId = this.get("requestId");
    var checker = setTimeout(function () {
      if (self.get('stopChecking') == true) {
        clearTimeout(checker);
      } else {
        App.ajax.send({
          name: 'preinstalled.checks.tasks',
          sender: self,
          data: {
            requestId: requestId
          },
          success: 'getHostCheckTasksSuccess',
          error: 'getHostCheckTasksError'
        });
      }
    }, 1000);
  },

  /**
   * add warnings to host warning popup if needed
   * @param data {Object} - json
   * @method getHostCheckTasksSuccess
   */
  getHostCheckTasksSuccess: function (data) {
    console.log('checking attempt...');
    if (!data) {
      console.warn("Error: jsonData is null");
      return;
    }
    if (["FAILED", "COMPLETED", "TIMEDOUT"].contains(data.Requests.request_status)) {
      if (data.Requests.inputs.indexOf("last_agent_env_check") != -1) {
        this.set('stopChecking', true);
        this.set('hostsPackagesData', data.tasks.map(function (task) {
          var installed_packages = Em.get(task, 'Tasks.structured_out.installed_packages');
          return {
            hostName: Em.get(task, 'Tasks.host_name'),
            installedPackages: installed_packages ? installed_packages : []
          }
        }));
        this.getHostInfo();
      } else if (data.Requests.inputs.indexOf("host_resolution_check") != -1) {
        this.parseHostNameResolution(data);
        this.getGeneralHostCheck();
       }
    } else {
      this.getHostCheckTasks();
    }
  },

  /**
   * parse warnings for host names resolution only
   * @param {object} data
   * @method parseHostNameResolution
   */
  parseHostNameResolution: function (data) {
    if (!data) {
      console.warn("Error: jsonData is null");
      return;
    }
    data.tasks.forEach(function (task) {
      var name = Em.I18n.t('installer.step3.hostWarningsPopup.resolution.validation.error');
      var hostInfo = this.get("hostCheckWarnings").findProperty('name', name);
      if (["FAILED", "COMPLETED", "TIMEDOUT"].contains(task.Tasks.status)) {
        if (!(task.Tasks.status == "COMPLETED" && Em.get(task, "Tasks.structured_out.host_resolution_check.failed_count") == 0)) {
          var targetHostName = Em.get(task, "Tasks.host_name");
          var relatedHostNames = Em.get(task, "Tasks.structured_out.host_resolution_check.failures")
            ? Em.get(task, "Tasks.structured_out.host_resolution_check.failures").mapProperty('host') : [];
          var contextMessage = Em.I18n.t('installer.step3.hostWarningsPopup.resolution.validation.context').format(targetHostName, relatedHostNames.join(', '));
          if (!hostInfo) {
            hostInfo = {
              name: name,
              hosts: [contextMessage],
              hostsNames: [targetHostName],
              onSingleHost: true
            };
            this.get("hostCheckWarnings").push(hostInfo);
          } else {
            if (!hostInfo.hostsNames.contains(targetHostName)) {
              hostInfo.hosts.push(contextMessage);
              hostInfo.hostsNames.push(targetHostName);
            }
          }
        }
      }
    }, this);
  },

  getHostCheckError: function() {
    this.getHostInfo();
  },

  stopChecking: false,

  /**
   * @method getHostCheckTasksError
   */
  getHostCheckTasksError: function() {
    console.warn("failed to cheek hostName resolution");
    this.set('stopChecking', true);
  },

  /**
   * Success-callback for hosts info request
   * @param {object} jsonData
   * @method getHostInfoSuccessCallback
   */
  getHostInfoSuccessCallback: function (jsonData) {
    var hosts = this.get('bootHosts'),
      self = this,
      repoWarnings = [], hostsRepoNames = [], hostsContext = [],
      diskWarnings = [], hostsDiskContext = [], hostsDiskNames = [],
      thpWarnings = [], thpContext = [], thpHostsNames = [];

    // parse host checks warning
    this.parseWarnings(jsonData);
    this.set('isHostsWarningsLoaded', true);
    hosts.forEach(function (_host) {
      var host = (App.get('testMode')) ? jsonData.items[0] : jsonData.items.findProperty('Hosts.host_name', _host.name);
      if (App.get('skipBootstrap')) {
        self._setHostDataWithSkipBootstrap(_host);
      }
      else {
        if (host) {
          self._setHostDataFromLoadedHostInfo(_host, host);
          var host_name = Em.get(host, 'Hosts.host_name');

          var context = self.checkHostOSType(host.Hosts.os_type, host_name);
          if (context) {
            hostsContext.push(context);
            hostsRepoNames.push(host_name);
          }
          var diskContext = self.checkHostDiskSpace(host_name, host.Hosts.disk_info);
          if (diskContext) {
            hostsDiskContext.push(diskContext);
            hostsDiskNames.push(host_name);
          }
          // "Transparent Huge Pages" check
          context = self.checkTHP(host_name, Em.get(host, 'Hosts.last_agent_env.transparentHugePage'));
          if (context) {
            thpContext.push(context);
            thpHostsNames.push(host_name);
          }
        }
      }
    });
    if (hostsContext.length > 0) { // repository warning exist
      repoWarnings.push({
        name: Em.I18n.t('installer.step3.hostWarningsPopup.repositories.name'),
        hosts: hostsContext,
        hostsNames: hostsRepoNames,
        category: 'repositories',
        onSingleHost: false
      });
    }
    if (hostsDiskContext.length > 0) { // disk space warning exist
      diskWarnings.push({
        name: Em.I18n.t('installer.step3.hostWarningsPopup.disk.name'),
        hosts: hostsDiskContext,
        hostsNames: hostsDiskNames,
        category: 'disk',
        onSingleHost: false
      });
    }
    if (thpContext.length > 0) { // THP warning existed
      thpWarnings.push({
        name: Em.I18n.t('installer.step3.hostWarningsPopup.thp.name'),
        hosts: thpContext,
        hostsNames: thpHostsNames,
        category: 'thp',
        onSingleHost: false
      });
    }

    this.set('repoCategoryWarnings', repoWarnings);
    this.set('diskCategoryWarnings', diskWarnings);
    this.set('thpCategoryWarnings', thpWarnings);
    this.stopRegistration();
  },

  /**
   * Set metrics to host object
   * Used when <code>App.skipBootstrap</code> is true
   * @param {Ember.Object} host
   * @returns {object}
   * @private
   * @methos _setHostDataWithSkipBootstrap
   */
  _setHostDataWithSkipBootstrap: function(host) {
    host.set('cpu', 2);
    host.set('memory', ((parseInt(2000000))).toFixed(2));
    host.set('disk_info', [
      {"mountpoint": "/", "type": "ext4"},
      {"mountpoint": "/grid/0", "type": "ext4"},
      {"mountpoint": "/grid/1", "type": "ext4"},
      {"mountpoint": "/grid/2", "type": "ext4"}
    ]);
    return host;
  },

  /**
   * Set loaded metrics to host object
   * @param {object} host
   * @param {object} hostInfo
   * @returns {object}
   * @method _setHostDataFromLoadedHostInfo
   * @private
   */
  _setHostDataFromLoadedHostInfo: function(host, hostInfo) {
    host.set('cpu', Em.get(hostInfo, 'Hosts.cpu_count'));
    host.set('memory', ((parseInt(Em.get(hostInfo, 'Hosts.total_mem')))).toFixed(2));
    host.set('disk_info', Em.get(hostInfo, 'Hosts.disk_info').filter(function (h) {
      return h.mountpoint != "/boot"
    }));
    host.set('os_type', Em.get(hostInfo, 'Hosts.os_type'));
    host.set('os_arch', Em.get(hostInfo, 'Hosts.os_arch'));
    host.set('ip', Em.get(hostInfo, 'Hosts.ip'));
    return host;
  },

  /**
   * Error-callback for hosts info request
   * @method getHostInfoErrorCallback
   */
  getHostInfoErrorCallback: function () {
    console.log('INFO: Getting host information(cpu_count and total_mem) from the server failed');
    this.set('isHostsWarningsLoaded', true);
    this.registerErrPopup(Em.I18n.t('installer.step3.hostInformation.popup.header'), Em.I18n.t('installer.step3.hostInformation.popup.body'));
  },

  /**
   * Enable or disable submit/retry buttons according to hosts boot statuses
   * @method stopRegistration
   */
  stopRegistration: function () {
    this.set('isSubmitDisabled', !this.get('bootHosts').someProperty('bootStatus', 'REGISTERED'));
    this.set('isRetryDisabled', !this.get('bootHosts').someProperty('bootStatus', 'FAILED'));
  },

  /**
   * Check if the 'Transparent Huge Pages' enabled.
   * @param {string} transparentHugePage
   * @param {string} hostName
   * @return {string} error-message or empty string
   * @method checkTHP
   */
  checkTHP: function (hostName, transparentHugePage) {
    if (transparentHugePage == "always") {
      console.log('WARNING: Transparent Huge Page enabled on host: '+ hostName);
      return Em.I18n.t('installer.step3.hostWarningsPopup.thp.context').format(hostName);
    } else {
      return '';
    }
  },

  /**
   * Check if the customized os group contains the registered host os type. If not the repo on that host is invalid.
   * @param {string} osType
   * @param {string} hostName
   * @return {string} error-message or empty string
   * @method checkHostOSType
   */
  checkHostOSType: function (osType, hostName) {
    if (this.get('content.stacks')) {
      var selectedStack = this.get('content.stacks').findProperty('isSelected', true);
      var selectedOS = [];
      var self = this;
      var isValid = false;
      if (selectedStack && selectedStack.get('operatingSystems')) {
        selectedStack.get('operatingSystems').filterProperty('isSelected', true).forEach(function (os) {
          selectedOS.pushObject(os.get('osType'));
          if (self.repoToAgentOsType(os.get('osType')).indexOf(osType) >= 0) {
            isValid = true;
          }
        });
      }
      if (isValid) {
        return '';
      } else {
        console.log('WARNING: Getting host os type does NOT match the user selected os group in step1. ' +
          'Host Name: ' + hostName + '. Host os type:' + osType + '. Selected group:' + selectedOS.uniq());
        return Em.I18n.t('installer.step3.hostWarningsPopup.repositories.context').format(hostName, osType, selectedOS.uniq());
      }
    } else {
      return '';
    }
  },

  /**
   * return the supported agent os types for a repo os type
   * @param {String} repoType
   * @return {Array} supported agent os type array
   * @method repoToAgentOsType
   */
  repoToAgentOsType : function (repoType) {
    /* istanbul ignore next */
    switch (repoType) {
      case "redhat6":
        return ["redhat6", "centos6", "oraclelinux6", "rhel6"];
      case "redhat5":
        return ["redhat5", "centos5", "oraclelinux5", "rhel5"];
      case "suse11":
        return ["suse11", "sles11", "opensuse11"];
      case "ubuntu12":
        return ["debian12", "ubuntu12"];
      default:
        return [];
    }
  },

  /**
   * Check if current host has enough free disk usage.
   * @param {string} hostName
   * @param {object} diskInfo
   * @return {string} error-message or empty string
   * @method checkHostDiskSpace
   */
  checkHostDiskSpace: function (hostName, diskInfo) {
    var minFreeRootSpace = App.minDiskSpace * 1024 * 1024; //in kilobyte
    var minFreeUsrLibSpace = App.minDiskSpaceUsrLib * 1024 * 1024; //in kilobyte
    var warningString = '';

    diskInfo.forEach(function (info) {
      switch (info.mountpoint) {
        case '/':
          warningString = info.available < minFreeRootSpace ?
            Em.I18n.t('installer.step3.hostWarningsPopup.disk.context2').format(App.minDiskSpace + 'GB', info.mountpoint) + ' ' + warningString :
            warningString;
          break;
        case '/usr':
        case '/usr/lib':
          warningString = info.available < minFreeUsrLibSpace ?
            Em.I18n.t('installer.step3.hostWarningsPopup.disk.context2').format(App.minDiskSpaceUsrLib + 'GB', info.mountpoint) + ' ' + warningString :
            warningString;
          break;
        default:
          break;
      }
    });
    if (warningString) {
      console.log('WARNING: Getting host free disk space. ' + 'Host Name: ' + hostName);
      return Em.I18n.t('installer.step3.hostWarningsPopup.disk.context1').format(hostName) + ' ' + warningString;
    } else {
      return '';
    }
  },

  /**
   * Submit-click handler
   * @return {App.ModalPopup|null}
   * @method submit
   */
  submit: function () {
    var self = this;

    if (this.get('isHostHaveWarnings')) {
      return App.showConfirmationPopup(
        function () {
          self.set('confirmedHosts', self.get('bootHosts'));
          App.router.send('next');
        },
        Em.I18n.t('installer.step3.hostWarningsPopup.hostHasWarnings'));
    }
    else {
      this.set('confirmedHosts', this.get('bootHosts'));
      App.router.send('next');
    }
    return null;
  },

  /**
   * Show popup with host log
   * @param {object} event
   * @return {App.ModalPopup}
   */
  hostLogPopup: function (event) {
    var host = event.context;

    return App.ModalPopup.show({
      header: Em.I18n.t('installer.step3.hostLog.popup.header').format(host.get('name')),
      secondary: null,
      host: host,
      bodyClass: App.WizardStep3HostLogPopupBody
    });
  },

  /**
   * Check warnings from server and put it in parsing
   * @method rerunChecks
   */
  rerunChecks: function () {
    var self = this;
    var currentProgress = 0;
    this.getHostNameResolution();
    this.checkHostJDK();
    var interval = setInterval(function () {
      currentProgress += 100000 / self.get('warningsTimeInterval');
      if (currentProgress < 100) {
        self.set('checksUpdateProgress', currentProgress);
      } else {
        clearInterval(interval);
        App.ajax.send({
          name: 'wizard.step3.rerun_checks',
          sender: self,
          success: 'rerunChecksSuccessCallback',
          error: 'rerunChecksErrorCallback'
        });
      }
    }, 1000);
  },

  /**
   * Success-callback for rerun request
   * @param {object} data
   * @method rerunChecksSuccessCallback
   */
  rerunChecksSuccessCallback: function (data) {
    this.set('checksUpdateProgress', 100);
    this.set('checksUpdateStatus', 'SUCCESS');
    this.parseWarnings(data);
  },

  /**
   * Error-callback for rerun request
   * @method rerunChecksErrorCallback
   */
  rerunChecksErrorCallback: function () {
    this.set('checksUpdateProgress', 100);
    this.set('checksUpdateStatus', 'FAILED');
    console.log('INFO: Getting host information(last_agent_env) from the server failed');
  },

  /**
   * Filter data for warnings parse
   * is data from host in bootStrap
   * @param {object} data
   * @return {Object}
   * @method filterBootHosts
   */
  filterBootHosts: function (data) {
    var bootHostNames = {};
    this.get('bootHosts').forEach(function (bootHost) {
      bootHostNames[bootHost.get('name')] = true;
    });
    var filteredData = {
      href: data.href,
      items: []
    };
    data.items.forEach(function (host) {
      if (bootHostNames[host.Hosts.host_name]) {
        filteredData.items.push(host);
      }
    });
    return filteredData;
  },

  /**
   * Parse warnings data for each host and total
   * @param {object} data
   * @method parseWarnings
   */
  parseWarnings: function (data) {
    data = App.get('testMode') ? data : this.filterBootHosts(data);
    var warnings = [];
    var warning;
    var hosts = [];
    var warningCategories = {
      fileFoldersWarnings: {},
      packagesWarnings: {},
      processesWarnings: {},
      servicesWarnings: {},
      usersWarnings: {},
      alternativeWarnings: {}
    };
    var hostsPackagesData = this.get('hostsPackagesData');

    data.items.sortPropertyLight('Hosts.host_name').forEach(function (_host) {
      var host = {
        name: _host.Hosts.host_name,
        warnings: []
      };
      if (!_host.Hosts.last_agent_env) {
        // in some unusual circumstances when last_agent_env is not available from the _host,
        // skip the _host and proceed to process the rest of the hosts.
        console.log("last_agent_env is missing for " + _host.Hosts.host_name + ".  Skipping _host check.");
        return;
      }

      //parse all directories and files warnings for host

      //todo: to be removed after check in new API
      var stackFoldersAndFiles = _host.Hosts.last_agent_env.stackFoldersAndFiles || [];
      stackFoldersAndFiles.forEach(function (path) {
        warning = warningCategories.fileFoldersWarnings[path.name];
        if (warning) {
          warning.hosts.push(_host.Hosts.host_name);
          warning.onSingleHost = false;
        } else {
          warningCategories.fileFoldersWarnings[path.name] = warning = {
            name: path.name,
            hosts: [_host.Hosts.host_name],
            category: 'fileFolders',
            onSingleHost: true
          };
        }
        host.warnings.push(warning);
      }, this);

      //parse all package warnings for host
      var _hostPackagesData = hostsPackagesData.findProperty('hostName', _host.Hosts.host_name);
      if (_hostPackagesData) {
        _hostPackagesData.installedPackages.forEach(function (_package) {
          warning = warningCategories.packagesWarnings[_package.name];
          if (warning) {
            warning.hosts.push(_host.Hosts.host_name);
            warning.version = _package.version;
            warning.onSingleHost = false;
          } else {
            warningCategories.packagesWarnings[_package.name] = warning = {
              name: _package.name,
              version: _package.version,
              hosts: [_host.Hosts.host_name],
              category: 'packages',
              onSingleHost: true
            };
          }
          host.warnings.push(warning);
        }, this);
      }

      //parse all process warnings for host

      //todo: to be removed after check in new API
      var javaProcs = _host.Hosts.last_agent_env.hostHealth ? _host.Hosts.last_agent_env.hostHealth.activeJavaProcs : _host.Hosts.last_agent_env.javaProcs;
      if (javaProcs) {
        javaProcs.forEach(function (process) {
          warning = warningCategories.processesWarnings[process.pid];
          if (warning) {
            warning.hosts.push(_host.Hosts.host_name);
            warning.onSingleHost = false;
          } else {
            warningCategories.processesWarnings[process.pid] = warning = {
              name: (process.command.substr(0, 35) + '...'),
              hosts: [_host.Hosts.host_name],
              category: 'processes',
              user: process.user,
              pid: process.pid,
              command: '<table><tr><td style="word-break: break-all;">' +
                ((process.command.length < 500) ? process.command : process.command.substr(0, 230) + '...' +
                  '<p style="text-align: center">................</p>' +
                  '...' + process.command.substr(-230)) + '</td></tr></table>',
              onSingleHost: true
            };
          }
          host.warnings.push(warning);
        }, this);
      }

      //parse all service warnings for host

      //todo: to be removed after check in new API
      if (_host.Hosts.last_agent_env.hostHealth && _host.Hosts.last_agent_env.hostHealth.liveServices) {
        _host.Hosts.last_agent_env.hostHealth.liveServices.forEach(function (service) {
          if (service.status === 'Unhealthy') {
            warning = warningCategories.servicesWarnings[service.name];
            if (warning) {
              warning.hosts.push(_host.Hosts.host_name);
              warning.onSingleHost = false;
            } else {
              warningCategories.servicesWarnings[service.name] = warning = {
                name: service.name,
                hosts: [_host.Hosts.host_name],
                category: 'services',
                onSingleHost: true
              };
            }
            host.warnings.push(warning);
          }
        }, this);
      }
      //parse all user warnings for host

      //todo: to be removed after check in new API
      if (_host.Hosts.last_agent_env.existingUsers) {
        _host.Hosts.last_agent_env.existingUsers.forEach(function (user) {
          warning = warningCategories.usersWarnings[user.userName];
          if (warning) {
            warning.hosts.push(_host.Hosts.host_name);
            warning.onSingleHost = false;
          } else {
            warningCategories.usersWarnings[user.userName] = warning = {
              name: user.userName,
              hosts: [_host.Hosts.host_name],
              category: 'users',
              onSingleHost: true
            };
          }
          host.warnings.push(warning);
        }, this);
      }

      //parse misc warnings for host
      var umask = _host.Hosts.last_agent_env.umask;
      if (umask && umask > 23) {
        warning = warnings.filterProperty('category', 'misc').findProperty('name', umask);
        if (warning) {
          warning.hosts.push(_host.Hosts.host_name);
          warning.onSingleHost = false;
        } else {
          warning = {
            name: umask,
            hosts: [_host.Hosts.host_name],
            category: 'misc',
            onSingleHost: true
          };
          warnings.push(warning);
        }
        host.warnings.push(warning);
      }

      var firewallRunning = _host.Hosts.last_agent_env.iptablesIsRunning;
      if (firewallRunning !== null && firewallRunning) {
        var name = Em.I18n.t('installer.step3.hostWarningsPopup.firewall.name');
        warning = warnings.filterProperty('category', 'firewall').findProperty('name', name);
        if (warning) {
          warning.hosts.push(_host.Hosts.host_name);
          warning.onSingleHost = false;
        } else {
          warning = {
            name: name,
            hosts: [_host.Hosts.host_name],
            category: 'firewall',
            onSingleHost: true
          };
          warnings.push(warning);
        }
        host.warnings.push(warning);
      }

      if (_host.Hosts.last_agent_env.alternatives) {
        _host.Hosts.last_agent_env.alternatives.forEach(function (alternative) {
          warning = warningCategories.alternativeWarnings[alternative.name];
          if (warning) {
            warning.hosts.push(_host.Hosts.host_name);
            warning.onSingleHost = false;
          } else {
            warningCategories.alternativeWarnings[alternative.name] = warning = {
              name: alternative.name,
              target: alternative.target,
              hosts: [_host.Hosts.host_name],
              category: 'alternatives',
              onSingleHost: true
            };
          }
          host.warnings.push(warning);
        }, this);
      }

      if (_host.Hosts.last_agent_env.reverseLookup === false) {
        var name = Em.I18n.t('installer.step3.hostWarningsPopup.reverseLookup.name');
        warning = warnings.filterProperty('category', 'reverseLookup').findProperty('name', name);
        if (warning) {
          warning.hosts.push(_host.Hosts.host_name);
          warning.onSingleHost = false;
        } else {
          warning = {
            name: name,
            hosts: [_host.Hosts.host_name],
            category: 'reverseLookup',
            onSingleHost: true
          };
          warnings.push(warning);
        }
        host.warnings.push(warning);
      }
      hosts.push(host);
    }, this);

    for (var categoryId in warningCategories) {
      var category = warningCategories[categoryId]
      for (var warningId in category) {
        warnings.push(category[warningId]);
      }
    }

    hosts.unshift({
      name: 'All Hosts',
      warnings: warnings
    });
    this.set('warnings', warnings);
    this.set('warningsByHost', hosts);
  },

  /**
   * Open popup that contain hosts' warnings
   * @return {App.ModalPopup}
   * @method hostWarningsPopup
   */
  hostWarningsPopup: function () {
    var self = this;
    return App.ModalPopup.show({

      header: Em.I18n.t('installer.step3.warnings.popup.header'),

      secondary: Em.I18n.t('installer.step3.hostWarningsPopup.rerunChecks'),

      primary: Em.I18n.t('common.close'),

      onPrimary: function () {
        self.set('checksUpdateStatus', null);
        this.hide();
      },

      onClose: function () {
        self.set('checksUpdateStatus', null);
        this.hide();
      },

      onSecondary: function () {
        self.rerunChecks();
      },

      didInsertElement: function () {
        this.fitHeight();
      },

      footerClass: App.WizardStep3HostWarningPopupFooter,

      bodyClass: App.WizardStep3HostWarningPopupBody
    });
  },

  /**
   * Show popup with registered hosts
   * @return {App.ModalPopup}
   * @method registeredHostsPopup
   */
  registeredHostsPopup: function () {
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('installer.step3.warning.registeredHosts').format(this.get('registeredHosts').length),
      secondary: null,
      bodyClass: Em.View.extend({
        templateName: require('templates/wizard/step3/step3_registered_hosts_popup'),
        message: Em.I18n.t('installer.step3.registeredHostsPopup'),
        registeredHosts: self.get('registeredHosts')
      })
    })
  }

});

});

require.register("controllers/wizard/step4_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

App.WizardStep4Controller = Em.ArrayController.extend({

  name: 'wizardStep4Controller',

  /**
   * List of Services
   * @type {Object[]}
   */
  content: [],

  /**
   * Check / Uncheck 'Select All' checkbox with one argument; Check / Uncheck all other checkboxes with more arguments
   * @type {bool}
   */
  isAllChecked: function(key, value) {
    if (arguments.length > 1) {
      this.filterProperty('isInstalled', false).setEach('isSelected', value);
      return value;
    } else {
      return this.filterProperty('isInstalled', false).
        filterProperty('isHiddenOnSelectServicePage', false).
        everyProperty('isSelected', true);
    }
  }.property('@each.isSelected'),

  /**
   * Is Submit button disabled
   * @type {bool}
   */
  isSubmitDisabled: function () {
    return this.filterProperty('isSelected', true).filterProperty('isInstalled', false).length === 0;
  }.property("@each.isSelected"),

  /**
   * List of validation errors. Look to #createError method for information
   * regarding object structure.
   *
   * @type {Object[]}
   */
  errorStack: [],

  /**
   * Drop errorStack content on selected state changes.
   **/
  clearErrors: function() {
    this.set('errorStack', []);
  }.observes('@each.isSelected'),

  /**
   * Check if multiple distributed file systems were selected
   * @return {bool}
   * @method multipleDFSs
   */
  multipleDFSs: function () {
    var dfsServices = this.filterProperty('isDFS',true).filterProperty('isSelected',true);
	  return  dfsServices.length > 1;
  },

  /**
   * Check whether user selected Ambari Metrics service to install and go to next step
   * @method ambariMetricsValidation
   */
  ambariMetricsValidation: function () {
    var ambariMetricsService = this.findProperty('serviceName', 'AMBARI_METRICS');
    if (ambariMetricsService && !ambariMetricsService.get('isSelected')) {
      this.addValidationError({
        id: 'ambariMetricsCheck',
        type: 'WARNING',
        callback: this.ambariMetricsCheckPopup
      });
    }
  },

  /**
   * Check whether Ranger is selected and show installation requirements if yes
   * @method rangerValidation
   */
  rangerValidation: function () {
    var rangerService = this.findProperty('serviceName', 'RANGER');
    if (rangerService && rangerService.get('isSelected') && !rangerService.get('isInstalled')) {
      this.addValidationError({
        id: 'rangerRequirements',
        type: 'WARNING',
        callback: this.rangerRequirementsPopup
      });
    }
  },

  /**
   * Warn user if he tries to install Spark with HDP 2.2
   * @method sparkValidation
   */
  sparkValidation: function () {
    var sparkService = this.findProperty('serviceName', 'SPARK');
    if (sparkService && sparkService.get('isSelected') && !sparkService.get('isInstalled') &&
      App.get('currentStackName') == 'HDP' && App.get('currentStackVersionNumber') == '2.2') {
      this.addValidationError({
        id: 'sparkWarning',
        type: 'WARNING',
        callback: this.sparkWarningPopup
      });
    }
  },

  /**
   * Onclick handler for <code>Next</code> button.
   * @method submit
   */
  submit: function () {
    if (!this.get('isSubmitDisabled')) {
      this.unSelectServices();
      this.setGroupedServices();
      if (this.validate()) {
        this.set('errorStack', []);
        App.router.send('next');
      }
    }
  },

  /**
   * Set isSelected based on property doNotShowAndInstall
   */
  unSelectServices: function () {
    this.filterProperty('isSelected',true).filterProperty('doNotShowAndInstall', true).setEach('isSelected', false);
  },

  /**
   * Check if validation passed:
   *  - required file system services selected
   *  - dependencies between services
   *  - monitoring services selected (not required)
   *
   * @return {Boolean}
   * @method validate
   **/
  validate: function() {
    this.serviceDependencyValidation();
    this.fileSystemServiceValidation();
    if (this.get('wizardController.name') == 'installerController') {
      this.ambariMetricsValidation();
    }
    this.rangerValidation();
    this.sparkValidation();
    if (!!this.get('errorStack').filterProperty('isShown', false).length) {
      this.showError(this.get('errorStack').findProperty('isShown', false));
      return false;
    }
    return true;
  },

  /**
   * Create error and push it to stack.
   *
   * @param {Object} errorObject - look to #createError
   * @return {Boolean}
   * @method addValidationError
   **/
  addValidationError: function(errorObject) {
    if (!this.get('errorStack').mapProperty('id').contains(errorObject.id)) {
      this.get('errorStack').push(this.createError(errorObject));
      return true;
    } else {
      return false;
    }
  },

  /**
   * Show current error by passed error object.
   *
   * @param {Object} errorObject
   * @method showError
   **/
  showError: function(errorObject) {
    return errorObject.callback.apply(errorObject.callbackContext, errorObject.callbackParams);
  },

  /**
   * Default primary button("Ok") callback for warning popups.
   *  Change isShown state for last shown error.
   *  Call #submit() method.
   *
   *  @method onPrimaryPopupCallback
   **/
  onPrimaryPopupCallback: function() {
    if (this.get('errorStack').someProperty('isShown', false)) {
      this.get('errorStack').findProperty('isShown', false).isShown = true;
    }
    this.submit();
  },

  /**
   * Create error object with passed options.
   * Available options:
   *  id - {String}
   *  type - {String}
   *  isShowed - {Boolean}
   *  callback - {Function}
   *  callbackContext
   *  callbackParams - {Array}
   *
   * @param {Object} opt
   * @return {Object}
   * @method createError
   **/
  createError: function(opt) {
    var options = {
      // {String} error identifier
      id: '',
      // {String} type of error CRITICAL|WARNING
      type: 'CRITICAL',
      // {Boolean} error was shown
      isShown: false,
      // {Function} callback to execute
      callback: null,
      // context which execute from
      callbackContext: this,
      // {Array} params applied to callback
      callbackParams: []
    };
    $.extend(options, opt);
    return options;
  },

  /**
   * Checks if a filesystem is present in the Stack
   *
   * @method isDFSStack
   */
  isDFSStack: function () {
	  var bDFSStack = false;
    var dfsServices = ['HDFS', 'GLUSTERFS'];
    var availableServices = this.filterProperty('isInstalled',false);
    availableServices.forEach(function(service){
      if (dfsServices.contains(service.get('serviceName'))) {
        console.log("found DFS " + service.get('serviceName'));
        bDFSStack=true;
      }
    },this);
    return bDFSStack;
  },

  /**
   * Checks if a filesystem is selected and only one filesystem is selected
   *
   * @method isFileSystemCheckFailed
   */
  fileSystemServiceValidation: function() {
    if(this.isDFSStack()){
      var primaryDFS = this.findProperty('isPrimaryDFS',true);
      var primaryDfsDisplayName = primaryDFS.get('displayNameOnSelectServicePage');
      var primaryDfsServiceName = primaryDFS.get('serviceName');
      if (this.multipleDFSs()) {
        var dfsServices = this.filterProperty('isDFS',true).filterProperty('isSelected',true).mapProperty('serviceName');
        var services = dfsServices.map(function (item){
          return  {
            serviceName: item,
            selected: item === primaryDfsServiceName
          };
        });
        this.addValidationError({
          id: 'multipleDFS',
          callback: this.needToAddServicePopup,
          callbackParams: [services, 'multipleDFS', primaryDfsDisplayName]
        });
      }
    }
  },

  /**
   * Checks if a dependent service is selected without selecting the main service.
   *
   * @method serviceDependencyValidation
   */
  serviceDependencyValidation: function() {
    var selectedServices = this.filterProperty('isSelected',true);
    var missingDependencies = [];
    var missingDependenciesDisplayName = [];
    selectedServices.forEach(function(service){
      var requiredServices =  service.get('requiredServices');
      if (!!requiredServices && requiredServices.length) {
        requiredServices.forEach(function(_requiredService){
          var requiredService = this.findProperty('serviceName', _requiredService);
          if (requiredService && requiredService.get('isSelected') === false) {
            if(missingDependencies.indexOf(_requiredService) == -1 ) {
              missingDependencies.push(_requiredService);
              missingDependenciesDisplayName.push(requiredService.get('displayNameOnSelectServicePage'));
            }
          }
        },this);
      }
    },this);

    if (missingDependencies.length > 0) {
      for(var i = 0; i < missingDependencies.length; i++) {
        this.addValidationError({
          id: 'serviceCheck_' + missingDependencies[i],
          callback: this.needToAddServicePopup,
          callbackParams: [{serviceName: missingDependencies[i], selected: true}, 'serviceCheck', missingDependenciesDisplayName[i]]
        });
      }
    }
  },

  /**
   * Select co hosted services which not showed on UI.
   *
   * @method setGroupedServices
   **/
  setGroupedServices: function() {
    this.forEach(function(service){
      var coSelectedServices = service.get('coSelectedServices');
      coSelectedServices.forEach(function(groupedServiceName) {
        var groupedService = this.findProperty('serviceName', groupedServiceName);
        if (groupedService.get('isSelected') !== service.get('isSelected')) {
          groupedService.set('isSelected',service.get('isSelected'));
        }
      },this);
    },this);
  },

  /**
   * Select/deselect services
   * @param services array of objects
   *  <code>
   *    [
   *      {
   *        service: 'HDFS',
   *        selected: true
   *      },
   *      ....
   *    ]
   *  </code>
   * @param {string} i18nSuffix
   * @param {string} serviceName
   * @return {App.ModalPopup}
   * @method needToAddServicePopup
   */

  needToAddServicePopup: function(services, i18nSuffix, serviceName) {
    if (!(services instanceof Array)) {
      services = [services];
    }
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('installer.step4.' + i18nSuffix + '.popup.header').format(serviceName),
      body: Em.I18n.t('installer.step4.' + i18nSuffix + '.popup.body').format(serviceName),
      onPrimary: function () {
        services.forEach(function (service) {
          self.findProperty('serviceName', service.serviceName).set('isSelected', service.selected);
        });
        self.onPrimaryPopupCallback();
        this.hide();
      }
    });
  },

  /**
   * Show popup with info about not selected Ambari Metrics service
   * @return {App.ModalPopup}
   * @method ambariMetricsCheckPopup
   */
  ambariMetricsCheckPopup: function () {
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('installer.step4.ambariMetricsCheck.popup.header'),
      body: Em.I18n.t('installer.step4.ambariMetricsCheck.popup.body'),
      primary: Em.I18n.t('common.proceedAnyway'),
      onPrimary: function () {
        self.onPrimaryPopupCallback();
        this.hide();
      }
    });
  },

  /**
   * Show popup with installation requirements for Ranger service
   * @return {App.ModalPopup}
   * @method rangerRequirementsPopup
   */
  rangerRequirementsPopup: function () {
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('installer.step4.rangerRequirements.popup.header'),
      bodyClass: Em.View.extend({
        templateName: require('templates/wizard/step4/step4_ranger_requirements_popup')
      }),
      primary: Em.I18n.t('common.proceed'),
      isChecked: false,
      disablePrimary: function () {
        return !this.get('isChecked');
      }.property('isChecked'),
      onPrimary: function () {
        self.onPrimaryPopupCallback();
        this.hide();
      }
    });
  },

  /**
   * Show popup with Spark installation warning
   * @return {App.ModalPopup}
   * @method sparkWarningPopup
   */
  sparkWarningPopup: function () {
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('common.warning'),
      body: Em.I18n.t('installer.step4.sparkWarning.popup.body'),
      primary: Em.I18n.t('common.proceed'),
      onPrimary: function () {
        self.onPrimaryPopupCallback();
        this.hide();
      }
    });
  }
});

});

require.register("controllers/wizard/step5_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var blueprintUtils = require('utils/blueprint');
var numberUtils = require('utils/number_utils');
var validationUtils = require('utils/validator');

App.WizardStep5Controller = Em.Controller.extend(App.BlueprintMixin, {

  name: "wizardStep5Controller",

  /**
   * Step title
   * Custom if <code>App.ReassignMasterController</code> is used
   * @type {string}
   */
  title: function () {
    if (this.get('content.controllerName') == 'reassignMasterController') {
      return Em.I18n.t('installer.step5.reassign.header');
    }
    return Em.I18n.t('installer.step5.header');
  }.property('content.controllerName'),

  /**
   * Is ReassignWizard used
   * @type {bool}
   */
  isReassignWizard: function () {
    return this.get('content.controllerName') == 'reassignMasterController';
  }.property('content.controllerName'),

  /**
   * Is isHighAvailabilityWizard used
   * @type {bool}
   */
  isHighAvailabilityWizard: function () {
    return this.get('content.controllerName') == 'highAvailabilityWizardController';
  }.property('content.controllerName'),

  /**
   * Check if <code>installerWizard</code> used
   * @type {bool}
   */
  isInstallerWizard: function () {
    return this.get('content.controllerName') === 'installerController';
  }.property('content.controllerName'),

  /**
   * Is AddServiceWizard used
   * @type {bool}
   */
  isAddServiceWizard: function () {
    return this.get('content.controllerName') == 'addServiceController';
  }.property('content.controllerName'),

  /**
   * Master components which could be assigned to multiple hosts
   * @type {string[]}
   */
  multipleComponents: function () {
    return App.get('components.multipleMasters');
  }.property('App.components.multipleMasters'),

  /**
   * Master components which could be assigned to multiple hosts
   * @type {string[]}
   */
  addableComponents: function () {
    return App.get('components.addableMasterInstallerWizard');
  }.property('App.components.addableMasterInstallerWizard'),

  /**
   * Define state for submit button
   * @type {bool}
   */
  submitDisabled: false,

  /**
   * Is Submit-click processing now
   * @type {bool}
   */
  submitButtonClicked: false,

  /**
   * Either use or not use server validation in this controller
   * @type {bool}
   */
  useServerValidation: true,

  /**
   * Trigger for executing host names check for components
   * Should de "triggered" when host changed for some component and when new multiple component is added/removed
   * @type {bool}
   */
  hostNameCheckTrigger: false,

  /**
   * List of hosts
   * @type {Array}
   */
  hosts: [],

  /**
   * Name of multiple component which host name was changed last
   * @type {Object|null}
   */
  componentToRebalance: null,

  /**
   * Name of component which host was changed last
   * @type {string}
   */
  lastChangedComponent: null,

  /**
   * Flag for rebalance multiple components
   * @type {number}
   */
  rebalanceComponentHostsCounter: 0,

  /**
   * @type {Ember.Enumerable}
   */
  servicesMasters: [],

  /**
   * @type {Ember.Enumerable}
   */
  selectedServicesMasters: [],


  /**
   * Is data for current step loaded
   * @type {bool}
   */
  isLoaded: false,

  /**
   * Validation error messages which don't related with any master
   */
  generalErrorMessages: [],

  /**
   * Validation warning messages which don't related with any master
   */
  generalWarningMessages: [],

  /**
   * Is masters-hosts layout initial one
   * @type {bool}
   */
  isInitialLayout: true,

  /**
   * true if any error exists
   */
  anyError: function() {
    return this.get('servicesMasters').some(function(m) { return m.get('errorMessage'); }) || this.get('generalErrorMessages').some(function(m) { return m; });
  }.property('servicesMasters.@each.errorMessage', 'generalErrorMessages'),

  /**
   * true if any warning exists
   */
  anyWarning: function() {
    return this.get('servicesMasters').some(function(m) { return m.get('warnMessage'); }) || this.get('generalWarningMessages').some(function(m) { return m; });
  }.property('servicesMasters.@each.warnMessage', 'generalWarningMessages'),

    /**
   * Clear loaded recommendations
   */
  clearRecommendations: function() {
    if (this.get('content.recommendations')) {
      this.set('content.recommendations', null);
    }
  },

  /**
   * List of host with assigned masters
   * Format:
   * <code>
   *   [
   *     {
   *       host_name: '',
   *       hostInfo: {},
   *       masterServices: [],
   *       masterServicesToDisplay: [] // used only in template
   *    },
   *    ....
   *   ]
   * </code>
   * @type {Ember.Enumerable}
   */
  masterHostMapping: function () {
    var mapping = [], mappingObject, mappedHosts, hostObj;
    //get the unique assigned hosts and find the master services assigned to them
    mappedHosts = this.get("selectedServicesMasters").mapProperty("selectedHost").uniq();
    mappedHosts.forEach(function (item) {
      hostObj = this.get("hosts").findProperty("host_name", item);
      // User may input invalid host name (this is handled in hostname checker). Here we just skip it
      if (!hostObj) return;
      var masterServices = this.get("selectedServicesMasters").filterProperty("selectedHost", item),
        masterServicesToDisplay = [];
      masterServices.mapProperty('display_name').uniq().forEach(function (n) {
        masterServicesToDisplay.pushObject(masterServices.findProperty('display_name', n));
      });
      mappingObject = Em.Object.create({
        host_name: item,
        hostInfo: hostObj.host_info,
        masterServices: masterServices,
        masterServicesToDisplay: masterServicesToDisplay
      });

      mapping.pushObject(mappingObject);
    }, this);

    return mapping.sortProperty('host_name');
  }.property("selectedServicesMasters.@each.selectedHost", 'selectedServicesMasters.@each.isHostNameValid'),

  /**
   * Count of hosts without masters
   * @type {number}
   */
  remainingHosts: function () {
    if (this.get('content.controllerName') === 'installerController') {
      return 0;
    } else {
      return (this.get("hosts.length") - this.get("masterHostMapping.length"));
    }
  }.property('masterHostMapping.length', 'selectedServicesMasters.@each.selectedHost'),

  /**
   * Update submit button status
   * @metohd updateIsSubmitDisabled
   */
  updateIsSubmitDisabled: function () {

    if (this.thereIsNoMasters()) {
      return false;
    }

    var isSubmitDisabled = this.get('servicesMasters').someProperty('isHostNameValid', false);

    if (this.get('useServerValidation')) {
      this.set('submitDisabled', true);

      if (this.get('servicesMasters').length === 0) {
        return;
      }

      if (!isSubmitDisabled) {
        if (!this.get('isInitialLayout')) {
          this.clearRecommendations(); // reset previous recommendations
        } else {
          this.set('isInitialLayout', false);
        }
        this.recommendAndValidate();
      }
    } else {
      this.set('submitDisabled', isSubmitDisabled);
      return isSubmitDisabled;
    }
  }.observes('servicesMasters.@each.selectedHost'),

  /**
   * Send AJAX request to validate current host layout
   * @param blueprint - blueprint for validation (can be with/withour slave/client components)
   */
  validate: function(blueprint, callback) {
    var self = this;

    var selectedServices = App.StackService.find().filterProperty('isSelected').mapProperty('serviceName');
    var installedServices = App.StackService.find().filterProperty('isInstalled').mapProperty('serviceName');
    var services = installedServices.concat(selectedServices).uniq();

    var hostNames = self.get('hosts').mapProperty('host_name');

    App.ajax.send({
      name: 'config.validations',
      sender: self,
      data: {
        stackVersionUrl: App.get('stackVersionURL'),
        hosts: hostNames,
        services: services,
        validate: 'host_groups',
        recommendations: blueprint
      },
      success: 'updateValidationsSuccessCallback',
      error: 'updateValidationsErrorCallback'
    }).
      then(function() {
        if (callback) {
          callback();
        }
      }
    );
  },

/**
  * Success-callback for validations request
  * @param {object} data
  * @method updateValidationsSuccessCallback
  */
  updateValidationsSuccessCallback: function (data) {
    var self = this;

    var generalErrorMessages = [];
    var generalWarningMessages = [];
    this.get('servicesMasters').setEach('warnMessage', null);
    this.get('servicesMasters').setEach('errorMessage', null);
    var anyErrors = false;

    var validationData = validationUtils.filterNotInstalledComponents(data);
    validationData.filterProperty('type', 'host-component').forEach(function(item) {
      var master = self.get('servicesMasters').find(function(m) {
        return m.component_name === item['component-name'] && m.selectedHost === item.host;
      });
      if (master) {
        if (item.level === 'ERROR') {
          anyErrors = true;
          master.set('errorMessage', item.message);
        } else if (item.level === 'WARN') {
          master.set('warnMessage', item.message);
        }
      }
    });

    this.set('generalErrorMessages', generalErrorMessages);
    this.set('generalWarningMessages', generalWarningMessages);

    // use this.set('submitDisabled', anyErrors); is validation results should block next button
    // It's because showValidationIssuesAcceptBox allow use accept validation issues and continue
    this.set('submitDisabled', false); //this.set('submitDisabled', anyErrors);
  },

  /**
   * Error-callback for validations request
   * @param {object} jqXHR
   * @param {object} ajaxOptions
   * @param {string} error
   * @param {object} opt
   * @method updateValidationsErrorCallback
   */
  updateValidationsErrorCallback: function (jqXHR, ajaxOptions, error, opt) {
    App.ajax.defaultErrorHandler(jqXHR, opt.url, opt.method, jqXHR.status);
    console.log('Load validations failed');
  },

  /**
   * Composes selected values of comboboxes into master blueprint + merge it with currenlty installed slave blueprint
   */
  getCurrentBlueprint: function() {
    var self = this;

    var res = {
      blueprint: { host_groups: [] },
      blueprint_cluster_binding: { host_groups: [] }
    };

    var mapping = self.get('masterHostMapping');

    mapping.forEach(function(item, i) {
      var group_name = 'host-group-' + (i+1);

      var host_group = {
        name: group_name,
        components: item.masterServices.map(function(master) {
          return { name: master.component_name };
        })
      };

      var binding = {
        name: group_name,
        hosts: [ { fqdn: item.host_name } ]
      };

      res.blueprint.host_groups.push(host_group);
      res.blueprint_cluster_binding.host_groups.push(binding);
    });

    return blueprintUtils.mergeBlueprints(res, self.getCurrentSlaveBlueprint());
  },

/**
   * Clear controller data (hosts, masters etc)
   * @method clearStep
   */
  clearStep: function () {
    this.set('hosts', []);
    this.set('selectedServicesMasters', []);
    this.set('servicesMasters', []);
    App.StackServiceComponent.find().forEach(function (stackComponent) {
      stackComponent.set('serviceComponentId', 1);
    }, this);

  },

  /**
   * Load controller data (hosts, host components etc)
   * @method loadStep
   */
  loadStep: function () {
    console.log("WizardStep5Controller: Loading step5: Assign Masters");
    this.clearStep();
    this.renderHostInfo();
    this.loadComponentsRecommendationsFromServer(this.loadStepCallback);
  },

  /**
   * Callback after load controller data (hosts, host components etc)
   * @method loadStepCallback
   */
  loadStepCallback: function(components, self) {
    self.renderComponents(components);

    self.get('addableComponents').forEach(function (componentName) {
      self.updateComponent(componentName);
    }, self);
    if (self.thereIsNoMasters()) {
      console.log('no master components to add');
      App.router.send('next');
    }
  },

  /**
  * Returns true if there is no new master components which need assigment to host
  */
  thereIsNoMasters: function() {
    return !this.get("selectedServicesMasters").filterProperty('isInstalled', false).length;
  },

  /**
   * Used to set showAddControl flag for installer wizard
   * @method updateComponent
   */
  updateComponent: function (componentName) {
    var component = this.last(componentName);
    if (!component) {
      return;
    }

    var showControl = !App.StackServiceComponent.find().findProperty('componentName', componentName).get('stackService').get('isInstalled');

    if (showControl) {
      var mastersLength = this.get("selectedServicesMasters").filterProperty("component_name", componentName).length;
      if (mastersLength < this.get("hosts.length") && !this.get('isReassignWizard') && !this.get('isHighAvailabilityWizard')) {
        component.set('showAddControl', true);
      } else if (mastersLength == 1 || this.get('isReassignWizard') || this.get('isHighAvailabilityWizard')) {
        component.set('showRemoveControl', false);
      }
    }
  },

  /**
   * Load active host list to <code>hosts</code> variable
   * @method renderHostInfo
   */
  renderHostInfo: function () {
    var hostInfo = this.get('content.hosts');
    var result = [];

    for (var index in hostInfo) {
      var _host = hostInfo[index];
      if (_host.bootStatus === 'REGISTERED') {
        result.push(Em.Object.create({
          host_name: _host.name,
          cpu: _host.cpu,
          memory: _host.memory,
          disk_info: _host.disk_info,
          host_info: Em.I18n.t('installer.step5.hostInfo').fmt(_host.name, numberUtils.bytesToSize(_host.memory, 1, 'parseFloat', 1024), _host.cpu)
        }));
      }
    }
    this.set("hosts", result);
    this.sortHosts(this.get('hosts'));
    this.set('isLoaded', true);
  },

  /**
   * Sort list of host-objects by properties (memory - desc, cpu - desc, hostname - asc)
   * @param {object[]} hosts
   */
  sortHosts: function (hosts) {
    hosts.sort(function (a, b) {
      if (a.get('memory') == b.get('memory')) {
        if (a.get('cpu') == b.get('cpu')) {
          return a.get('host_name').localeCompare(b.get('host_name')); // hostname asc
        }
        return b.get('cpu') - a.get('cpu'); // cores desc
      }
      return b.get('memory') - a.get('memory'); // ram desc
    });
  },

  /**
   * Get recommendations info from API
   * @return {undefined}
   * @param function(componentInstallationobjects, this) callback
   * @param bool includeMasters
   */
  loadComponentsRecommendationsFromServer: function(callback, includeMasters) {
    var self = this;

    if (this.get('content.recommendations')) {
      // Don't do AJAX call if recommendations has been already received
      // But if user returns to previous step (selecting services), stored recommendations will be cleared in routers' next handler and AJAX call will be made again
      callback(self.createComponentInstallationObjects(), self);
    } else {
      var selectedServices = App.StackService.find().filterProperty('isSelected').mapProperty('serviceName');
      var installedServices = App.StackService.find().filterProperty('isInstalled').mapProperty('serviceName');
      var services = installedServices.concat(selectedServices).uniq();

      var hostNames = self.get('hosts').mapProperty('host_name');

      var data = {
        stackVersionUrl: App.get('stackVersionURL'),
        hosts: hostNames,
        services: services,
        recommend: 'host_groups'
      };

      if (includeMasters) {
        // Made partial recommendation request for reflect in blueprint host-layout changes which were made by user in UI
        data.recommendations = self.getCurrentBlueprint();
      } else if (!self.get('isInstallerWizard')) {
        data.recommendations = self.getCurrentMasterSlaveBlueprint();
      }

      return App.ajax.send({
        name: 'wizard.loadrecommendations',
        sender: self,
        data: data,
        success: 'loadRecommendationsSuccessCallback',
        error: 'loadRecommendationsErrorCallback'
      }).
        then(function () {
          callback(self.createComponentInstallationObjects(), self);
        });
    }
  },

  /**
   * Create components for displaying component-host comboboxes in UI assign dialog
   * expects content.recommendations will be filled with recommendations API call result
   * @return {Object[]}
   */
  createComponentInstallationObjects: function() {
    var self = this;

    var masterComponents = [];
    if (self.get('isAddServiceWizard')) {
      masterComponents = App.StackServiceComponent.find().filterProperty('isShownOnAddServiceAssignMasterPage');
    } else {
      masterComponents = App.StackServiceComponent.find().filterProperty('isShownOnInstallerAssignMasterPage');
    }

    var masterHosts = self.get('content.masterComponentHosts'); //saved to local storage info
    var selectedNotInstalledServices = self.get('content.services').filterProperty('isSelected').filterProperty('isInstalled', false).mapProperty('serviceName');
    var recommendations = this.get('content.recommendations');

    var resultComponents = [];
    var multipleComponentHasBeenAdded = {};

    recommendations.blueprint.host_groups.forEach(function(host_group) {
      var hosts = recommendations.blueprint_cluster_binding.host_groups.findProperty('name', host_group.name).hosts;

      hosts.forEach(function(host) {
        host_group.components.forEach(function(component) {
          var willBeAdded = true;
          var fullComponent = masterComponents.findProperty('componentName', component.name);
          // If it's master component which should be shown
          if (fullComponent) {
            // If service is already installed and not being added as a new service then render on UI only those master components
            // that have already installed hostComponents.
            // NOTE: On upgrade there might be a prior installed service with non-installed newly introduced serviceComponent
            var isNotSelectedService = !selectedNotInstalledServices.contains(fullComponent.get('serviceName'));
            if (isNotSelectedService) {
              willBeAdded = App.HostComponent.find().someProperty('componentName', component.name);
            }

            if (willBeAdded) {
              var savedComponents = masterHosts.filterProperty('component', component.name);

              if (self.get('multipleComponents').contains(component.name) && savedComponents.length > 0) {
                if (!multipleComponentHasBeenAdded[component.name]) {
                  multipleComponentHasBeenAdded[component.name] = true;

                  savedComponents.forEach(function(saved) {
                    resultComponents.push(self.createComponentInstallationObject(fullComponent, host.fqdn.toLowerCase(), saved));
                  });
                }
              } else {
                var savedComponent = masterHosts.findProperty('component', component.name);
                resultComponents.push(self.createComponentInstallationObject(fullComponent, host.fqdn.toLowerCase(), savedComponent));
              }
            }
          }
        });
      });
    });
    return resultComponents;
  },

  /**
   * Create component for displaying component-host comboboxes in UI assign dialog
   * @param fullComponent - full component description
   * @param hostName - host fqdn where component will be installed
   * @param savedComponent - the same object which function returns but created before
   * @return {Object}
   */
  createComponentInstallationObject: function(fullComponent, hostName, savedComponent) {
    var componentName = fullComponent.get('componentName');

    var componentObj = {};
    componentObj.component_name = componentName;
    componentObj.display_name = App.format.role(fullComponent.get('componentName'));
    componentObj.serviceId = fullComponent.get('serviceName');
    componentObj.isServiceCoHost = App.StackServiceComponent.find().findProperty('componentName', componentName).get('isCoHostedComponent') && !this.get('isReassignWizard');

    if (savedComponent) {
      componentObj.selectedHost = savedComponent.hostName;
      componentObj.isInstalled = savedComponent.isInstalled;
    } else {
      componentObj.selectedHost = hostName;
      componentObj.isInstalled = false;
    }

    return componentObj;
  },

  /**
   * Success-callback for recommendations request
   * @param {object} data
   * @method loadRecommendationsSuccessCallback
   */
  loadRecommendationsSuccessCallback: function (data) {
    this.set('content.recommendations', data.resources[0].recommendations);
  },

  /**
   * Error-callback for recommendations request
   * @param {object} jqXHR
   * @param {object} ajaxOptions
   * @param {string} error
   * @param {object} opt
   * @method loadRecommendationsErrorCallback
   */
  loadRecommendationsErrorCallback: function (jqXHR, ajaxOptions, error, opt) {
    App.ajax.defaultErrorHandler(jqXHR, opt.url, opt.method, jqXHR.status);
    console.log('Load recommendations failed');
  },

  /**
   * Put master components to <code>selectedServicesMasters</code>, which will be automatically rendered in template
   * @param {Ember.Enumerable} masterComponents
   * @method renderComponents
   */
  renderComponents: function (masterComponents) {
    var installedServices = App.StackService.find().filterProperty('isSelected').filterProperty('isInstalled', false).mapProperty('serviceName'); //list of shown services
    var result = [];
    var serviceComponentId, previousComponentName;

    masterComponents.forEach(function (item) {
      var serviceComponent = App.StackServiceComponent.find().findProperty('componentName', item.component_name);
      var showRemoveControl = installedServices.contains(serviceComponent.get('stackService.serviceName')) &&
        (masterComponents.filterProperty('component_name', item.component_name).length > 1);
      var componentObj = Em.Object.create(item);
      console.log("TRACE: render master component name is: " + item.component_name);
      var masterComponent = App.StackServiceComponent.find().findProperty('componentName', item.component_name);
      if (masterComponent.get('isMasterWithMultipleInstances')) {
        previousComponentName = item.component_name;
        componentObj.set('serviceComponentId', result.filterProperty('component_name', item.component_name).length + 1);
        componentObj.set("showRemoveControl", showRemoveControl);
      }
      componentObj.set('isHostNameValid', true);

      result.push(componentObj);
    }, this);
    result = this.sortComponentsByServiceName(result);
    this.set("selectedServicesMasters", result);
    if (this.get('isReassignWizard')) {
      var components = result.filterProperty('component_name', this.get('content.reassign.component_name'));
      components.setEach('isInstalled', false);
      this.set('servicesMasters', components);
    } else {
      this.set('servicesMasters', result);
    }
  },

  sortComponentsByServiceName: function(components) {
    var displayOrder = App.StackService.displayOrder;
    return components.sort(function (a, b) {
      var aValue = displayOrder.indexOf(a.serviceId) != -1 ? displayOrder.indexOf(a.serviceId) : components.length;
      var bValue = displayOrder.indexOf(b.serviceId) != -1 ? displayOrder.indexOf(b.serviceId) : components.length;
      return aValue - bValue;
    });
  },
  /**
   * Update dependent co-hosted components according to the change in the component host
   * @method updateCoHosts
   */
  updateCoHosts: function () {
    // reassign wizard has no co-host constraints
    if (this.get('isReassignWizard')) {
      return false;
    }

    var components = App.StackServiceComponent.find().filterProperty('isOtherComponentCoHosted');
    var selectedServicesMasters = this.get('selectedServicesMasters');
    components.forEach(function (component) {
      var componentName = component.get('componentName');
      var hostComponent = selectedServicesMasters.findProperty('component_name', componentName);
      var dependentCoHosts = component.get('coHostedComponents');
      dependentCoHosts.forEach(function (coHostedComponent) {
        var dependentHostComponent = selectedServicesMasters.findProperty('component_name', coHostedComponent);
        if (hostComponent && dependentHostComponent) dependentHostComponent.set('selectedHost', hostComponent.get('selectedHost'));
      }, this);
    }, this);
  }.observes('selectedServicesMasters.@each.selectedHost'),


  /**
   * On change callback for inputs
   * @param {string} componentName
   * @param {string} selectedHost
   * @param {number} serviceComponentId
   * @method assignHostToMaster
   */
  assignHostToMaster: function (componentName, selectedHost, serviceComponentId) {
    var flag = this.isHostNameValid(componentName, selectedHost);
    this.updateIsHostNameValidFlag(componentName, serviceComponentId, flag);
    if (serviceComponentId) {
      this.get('selectedServicesMasters').filterProperty('component_name', componentName).findProperty("serviceComponentId", serviceComponentId).set("selectedHost", selectedHost);
    }
    else {
      this.get('selectedServicesMasters').findProperty("component_name", componentName).set("selectedHost", selectedHost);
    }
  },

  /**
   * Determines if hostName is valid for component:
   * <ul>
   *  <li>host name shouldn't be empty</li>
   *  <li>host should exist</li>
   *  <li>host should have only one component with <code>componentName</code></li>
   * </ul>
   * @param {string} componentName
   * @param {string} selectedHost
   * @returns {boolean} true - valid, false - invalid
   * @method isHostNameValid
   */
  isHostNameValid: function (componentName, selectedHost) {
    return (selectedHost.trim() !== '') &&
      this.get('hosts').mapProperty('host_name').contains(selectedHost) &&
      (this.get('selectedServicesMasters').
        filterProperty('component_name', componentName).
        mapProperty('selectedHost').
        filter(function (h) {
          return h === selectedHost;
        }).length <= 1);
  },

  /**
   * Update <code>isHostNameValid</code> property with <code>flag</code> value
   * for component with name <code>componentName</code> and
   * <code>serviceComponentId</code>-property equal to <code>serviceComponentId</code>-parameter value
   * @param {string} componentName
   * @param {number} serviceComponentId
   * @param {bool} flag
   * @method updateIsHostNameValidFlag
   */
  updateIsHostNameValidFlag: function (componentName, serviceComponentId, flag) {
    if (componentName) {
      if (serviceComponentId) {
        this.get('selectedServicesMasters').filterProperty('component_name', componentName).findProperty("serviceComponentId", serviceComponentId).set("isHostNameValid", flag);
      } else {
        this.get('selectedServicesMasters').findProperty("component_name", componentName).set("isHostNameValid", flag);
      }
    }
  },

  /**
   * Returns last component of selected type
   * @param {string} componentName
   * @return {Em.Object|null}
   * @method last
   */
  last: function (componentName) {
    return this.get("selectedServicesMasters").filterProperty("component_name", componentName).get("lastObject");
  },

  /**
   * Add new component to ZooKeeper Server and Hbase master
   * @param {string} componentName
   * @return {bool} true - added, false - not added
   * @method addComponent
   */
  addComponent: function (componentName) {
    /*
     * Logic: If ZooKeeper or Hbase service is selected then there can be
     * minimum 1 ZooKeeper or Hbase master in total, and
     * maximum 1 ZooKeeper or Hbase on every host
     */

    var maxNumMasters = this.get("hosts.length"),
      currentMasters = this.get("selectedServicesMasters").filterProperty("component_name", componentName),
      newMaster = null,
      masterHosts = null,
      suggestedHost = null,
      i = 0,
      lastMaster = null;

    if (!currentMasters.length) {
      console.log('ALERT: Zookeeper service was not selected');
      return false;
    }

    if (currentMasters.get("length") < maxNumMasters) {

      currentMasters.set("lastObject.showAddControl", false);
      currentMasters.set("lastObject.showRemoveControl", true);

      //create a new master component host based on an existing one
      newMaster = Em.Object.create({});
      lastMaster = currentMasters.get("lastObject");
      newMaster.set("display_name", lastMaster.get("display_name"));
      newMaster.set("component_name", lastMaster.get("component_name"));
      newMaster.set("selectedHost", lastMaster.get("selectedHost"));
      newMaster.set("serviceId", lastMaster.get("serviceId"));
      newMaster.set("isInstalled", false);

      if (currentMasters.get("length") === (maxNumMasters - 1)) {
        newMaster.set("showAddControl", false);
      } else {
        newMaster.set("showAddControl", true);
      }
      newMaster.set("showRemoveControl", true);

      //get recommended host for the new Zookeeper server
      masterHosts = currentMasters.mapProperty("selectedHost").uniq();

      for (i = 0; i < this.get("hosts.length"); i++) {
        if (!(masterHosts.contains(this.get("hosts")[i].get("host_name")))) {
          suggestedHost = this.get("hosts")[i].get("host_name");
          break;
        }
      }

      newMaster.set("selectedHost", suggestedHost);
      newMaster.set("serviceComponentId", (currentMasters.get("lastObject.serviceComponentId") + 1));

      this.get("selectedServicesMasters").insertAt(this.get("selectedServicesMasters").indexOf(lastMaster) + 1, newMaster);

      this.set('componentToRebalance', componentName);
      this.incrementProperty('rebalanceComponentHostsCounter');
      this.toggleProperty('hostNameCheckTrigger');
      return true;
    }
    return false;//if no more zookeepers can be added
  },

  /**
   * Remove component from ZooKeeper server or Hbase Master
   * @param {string} componentName
   * @param {number} serviceComponentId
   * @return {bool} true - removed, false - no
   * @method removeComponent
   */
  removeComponent: function (componentName, serviceComponentId) {
    var currentMasters = this.get("selectedServicesMasters").filterProperty("component_name", componentName);

    //work only if the multiple master service is selected in previous step
    if (currentMasters.length <= 1) {
      return false;
    }

    this.get("selectedServicesMasters").removeAt(this.get("selectedServicesMasters").indexOf(currentMasters.findProperty("serviceComponentId", serviceComponentId)));

    currentMasters = this.get("selectedServicesMasters").filterProperty("component_name", componentName);
    if (currentMasters.get("length") < this.get("hosts.length")) {
      currentMasters.set("lastObject.showAddControl", true);
    }

    if (currentMasters.get("length") === 1) {
      currentMasters.set("lastObject.showRemoveControl", false);
    }

    this.set('componentToRebalance', componentName);
    this.incrementProperty('rebalanceComponentHostsCounter');
    this.toggleProperty('hostNameCheckTrigger');
    return true;
  },

  recommendAndValidate: function(callback) {
    var self = this;

    // load recommendations with partial request
    self.loadComponentsRecommendationsFromServer(function() {
      // For validation use latest received recommendations because ir contains current master layout and recommended slave/client layout
      self.validate(self.get('content.recommendations'), function() {
        if (callback) {
          callback();
        }
      });
    }, true);
  },

  /**
   * Submit button click handler
   * @method submit
   */
  submit: function () {
    var self = this;
    if (!this.get('submitButtonClicked')) {
      this.set('submitButtonClicked', true);

      var goNextStepIfValid = function () {
        if (!self.get('submitDisabled')) {
          App.router.send('next');
        }
        self.set('submitButtonClicked', false);
      };

      if (this.get('useServerValidation')) {
        self.recommendAndValidate(function () {
          self.showValidationIssuesAcceptBox(goNextStepIfValid);
        });
      } else {
        self.updateIsSubmitDisabled();
        goNextStepIfValid();
      }
    }
  },

  /**
   * In case of any validation issues shows accept dialog box for user which allow cancel and fix issues or continue anyway
   * @method showValidationIssuesAcceptBox
   */
  showValidationIssuesAcceptBox: function(callback) {
    var self = this;
    if (self.get('anyWarning') || self.get('anyError')) {
      App.ModalPopup.show({
        primary: Em.I18n.t('common.continueAnyway'),
        header: Em.I18n.t('installer.step5.validationIssuesAttention.header'),
        body: Em.I18n.t('installer.step5.validationIssuesAttention'),
        onPrimary: function () {
          this.hide();
          callback();
        }
      });
    } else {
      callback();
    }
  }
});

});

require.register("controllers/wizard/step6_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var db = require('utils/db');
var stringUtils = require('utils/string_utils');
var blueprintUtils = require('utils/blueprint');
var validationUtils = require('utils/validator');

/**
 * By Step 6, we have the following information stored in App.db and set on this
 * controller by the router:
 *
 *   hosts: App.db.hosts (list of all hosts the user selected in Step 3)
 *   selectedServiceNames: App.db.selectedServiceNames (the services that the user selected in Step 4)
 *   masterComponentHosts: App.db.masterComponentHosts (master-components-to-hosts mapping the user selected in Step 5)
 *
 * Step 6 will set the following information in App.db:
 *   slaveComponentHosts: App.db.slaveComponentHosts (slave-components-to-hosts mapping the user selected in Step 6)
 *
 */
App.WizardStep6Controller = Em.Controller.extend(App.BlueprintMixin, {

  /**
   * List of hosts
   * @type {object[]}
   */
  hosts: [],

  /**
   * List of components info about selecting/deselecting status for components.
   *
   * @type {Array}
   *  @item {Em.Object}
   *    @property name {String} - component name
   *    @property label {String} - component display name
   *    @property allChecked {bool} - all checkboxes are checked
   *    @property noChecked {bool} - no checkboxes checked
   */
  headers: [],

  /**
   * @type {bool}
   */
  isLoaded: false,

  /**
   * Indication if user has chosen hosts to install clients
   * @type {bool}
   */
  isClientsSet: false,

  /**
   * Define state for submit button
   * @type {bool}
   */
  submitDisabled: false,

  /**
   * Check if <code>addHostWizard</code> used
   * @type {bool}
   */
  isAddHostWizard: function () {
    return this.get('content.controllerName') === 'addHostController';
  }.property('content.controllerName'),

  /**
   * Check if <code>installerWizard</code> used
   * @type {bool}
   */
  isInstallerWizard: function () {
    return this.get('content.controllerName') === 'installerController';
  }.property('content.controllerName'),

  /**
   * Check if <code>addServiceWizard</code> used
   * @type {bool}
   */
  isAddServiceWizard: function () {
    return this.get('content.controllerName') === 'addServiceController';
  }.property('content.controllerName'),

  installedServiceNames: function () {
    return this.get('content.services').filterProperty('isInstalled').mapProperty('serviceName');
  }.property('content.services').cacheable(),

  /**
   * Validation error messages which don't related with any master
   */
  generalErrorMessages: [],

  /**
   * Validation warning messages which don't related with any master
   */
  generalWarningMessages: [],

  /**
   * true if validation has any general (which is not related with concrete host) error message
   */
  anyGeneralErrors: function() {
    var messages = this.get('generalErrorMessages');
    return this.get('errorMessage') || (messages && messages.length > 0);
  }.property('generalErrorMessages', 'generalErrorMessages.@each', 'errorMessage'),

  /**
   * true if validation has any general (which is not related with concrete host) warning message
   */
  anyGeneralWarnings: function() {
    var messages = this.get('generalWarningMessages');
    return messages && messages.length > 0;
  }.property('generalWarningMessages', 'generalWarningMessages.@each'),

  /**
   * true if validation has any general (which is not related with concrete host) error or warning message
   */
  anyGeneralIssues: function () {
    return this.get('anyGeneralErrors') || this.get('anyGeneralWarnings');
  }.property('anyGeneralErrors', 'anyGeneralWarnings'),

  /**
   * true if validation has any error message (general or host specific)
   */
  anyErrors: function() {
    return this.get('anyGeneralErrors') || this.get('hosts').some(function(h) { return h.get('errorMessages').length > 0; });
  }.property('anyGeneralErrors', 'hosts.@each.errorMessages'),

  /**
   * true if validation has any warning message (general or host specific)
   */
  anyWarnings: function() {
    return this.get('anyGeneralWarnings') || this.get('hosts').some(function(h) { return h.get('warnMessages').length > 0; });
  }.property('anyGeneralWarnings', 'hosts.@each.warnMessages'),

  openSlavesAndClientsIssues: function () {
    App.ModalPopup.show({
      header: Em.I18n.t('installer.step6.validationSlavesAndClients.popup.header'),
      bodyClass: Em.View.extend({
        controller: this,
        templateName: require('templates/wizard/step6/step6_issues_popup')
      }),
      secondary: null
    });
  },

  /**
   * Verify condition that at least one checkbox of each component was checked
   * @method clearError
   */
  clearError: function () {
    var self = this;
    var isError = false;
    var err = false;
    var hosts = this.get('hosts');
    var headers = this.get('headers');
    var headersMap = {};

    headers.forEach(function (header) {
      headersMap[header.name] = true;
    });
    hosts.forEach(function (host) {
      host.get('checkboxes').forEach(function (checkbox) {
        if (headersMap[checkbox.get('component')]) {
          headersMap[checkbox.get('component')] = !checkbox.get('checked');
        }
      });
    });
    for (var i in headersMap) {
      err |= headersMap[i];
    }

    if (!err) {
      this.set('errorMessage', '');
    }

    if (this.get('isAddHostWizard')) {
      hosts.forEach(function (host) {
        isError = false;
        headers.forEach(function (header) {
          isError |= host.get('checkboxes').findProperty('title', header.get('label')).checked;
        });
        isError = !isError;
        if (!isError) {
          self.set('errorMessage', '');
        }
      });
    }
  },

  /**
   * Clear Step6 data like <code>hosts</code>, <code>headers</code> etc
   * @method clearStep
   */
  clearStep: function () {
    this.set('hosts', []);
    this.set('headers', []);
    this.clearError();
    this.set('isLoaded', false);
  },

  /**
   * Enable some service for all hosts
   * @param {object} event
   * @method selectAllNodes
   */
  selectAllNodes: function (event) {
    var name = Em.get(event, 'context.name');
    if (name) {
      this.setAllNodes(name, true);
      this.callValidation();
    }
  },

  /**
   * Disable some services for all hosts
   * @param {object} event
   * @method deselectAllNodes
   */
  deselectAllNodes: function (event) {
    var name = Em.get(event, 'context.name');
    if (name) {
      this.setAllNodes(name, false);
      this.callValidation();
    }
  },

  /**
   * Enable/disable some service for all hosts
   * @param {String} component - component name
   * @param {bool} checked - true - enable, false - disable
   * @method setAllNodes
   */
  setAllNodes: function (component, checked) {
    this.get('hosts').forEach(function (host) {
      host.get('checkboxes').filterProperty('isInstalled', false).forEach(function (checkbox) {
        if (checkbox.get('component') === component) {
          checkbox.set('checked', checked);
        }
      });
    });
    this.checkCallback(component);
  },

  /**
   * Checkbox check callback
   * Verify if all/none checkboxes for current component are checked
   * @param {String} component
   * @method checkCallback
   */
  checkCallback: function (component) {
    var header = this.get('headers').findProperty('name', component);
    if (header) {
      var hosts = this.get('hosts');
      var allTrue = true;
      var allFalse = true;
      hosts.forEach(function (host) {
        host.get('checkboxes').forEach(function (checkbox) {
          if (checkbox.get('component') === component && !checkbox.get('isInstalled')) {
            allTrue = allTrue && checkbox.get('checked');
            allFalse = allFalse && !checkbox.get('checked');
          }
        });
      });
      header.set('allChecked', allTrue);
      header.set('noChecked', allFalse);
    }
    this.clearError();
  },

  /**
   * Init step6 data
   * @method loadStep
   */
  loadStep: function () {

    console.log("WizardStep6Controller: Loading step6: Assign Slaves");
    this.clearStep();
    var selectedServices = App.StackService.find().filterProperty('isSelected');
    var installedServices = App.StackService.find().filterProperty('isInstalled');
    var services;
    if (this.get('isInstallerWizard')) services = selectedServices;
    else if (this.get('isAddHostWizard')) services = installedServices;
    else if (this.get('isAddServiceWizard')) services = installedServices.concat(selectedServices);

    var headers = Em.A([]);
    services.forEach(function (stackService) {
      stackService.get('serviceComponents').forEach(function (serviceComponent) {
        if (serviceComponent.get('isShownOnInstallerSlaveClientPage')) {
          headers.pushObject(Em.Object.create({
            name: serviceComponent.get('componentName'),
            label: App.format.role(serviceComponent.get('componentName')),
            allChecked: false,
            isRequired: serviceComponent.get('isRequired'),
            noChecked: true,
            isDisabled: installedServices.someProperty('serviceName', stackService.get('serviceName')) && this.get('isAddServiceWizard')
          }));
        }
      }, this);
    }, this);
    if (this.get('content.clients') && !!this.get('content.clients').length) {
      headers.pushObject(Em.Object.create({
        name: 'CLIENT',
        label: App.format.role('CLIENT'),
        allChecked: false,
        noChecked: true,
        isDisabled: false
      }));
    }
    this.get('headers').pushObjects(headers);

    this.render();
    if (this.get('content.skipSlavesStep')) {
      App.router.send('next');
    } else {
      this.callValidation();
    }
  },

  /**
   * Get active host names
   * @return {string[]}
   * @method getHostNames
   */
  getHostNames: function () {
    var hostInfo = this.get('content.hosts');
    var hostNames = [];
    //flag identify whether get all hosts or only uninstalled(newly added) hosts
    var getUninstalledHosts = (this.get('content.controllerName') !== 'addServiceController');

    for (var index in hostInfo) {
      if (hostInfo.hasOwnProperty(index)) {
        if (hostInfo[index].bootStatus === 'REGISTERED') {
          if (!getUninstalledHosts || !hostInfo[index].isInstalled) {
            hostNames.push(hostInfo[index].name);
          }
        }
      }
    }
    return hostNames;
  },

  /**
   * Load all data needed for this module. Then it automatically renders in template
   * @method render
   */
  render: function () {
    var hostsObj = [],
      masterHosts = [],
      headers = this.get('headers'),
      masterHostNames = this.get('content.masterComponentHosts').mapProperty('hostName').uniq();

    this.getHostNames().forEach(function (_hostName) {
      var hasMaster = masterHostNames.contains(_hostName);

      var obj = Em.Object.create({
        hostName: _hostName,
        hasMaster: hasMaster,
        checkboxes: []
      });

      headers.forEach(function (header) {
        obj.checkboxes.pushObject(Em.Object.create({
          component: header.name,
          title: header.label,
          checked: false,
          isInstalled: false,
          isDisabled: header.get('isDisabled')
        }));
      });

      if (hasMaster) {
        masterHosts.pushObject(obj)
      } else {
        hostsObj.pushObject(obj);
      }
    });
    //hosts with master components should be in the beginning of list
    hostsObj.unshift.apply(hostsObj, masterHosts);
    hostsObj = this.renderSlaves(hostsObj);
    this.set('hosts', hostsObj);
    headers.forEach(function (header) {
      this.checkCallback(header.get('name'));
    }, this);
    this.set('isLoaded', true);
  },

  /**
   * Set checked values for slaves checkboxes
   * @param {Array} hostsObj
   * @return {Array}
   * @method renderSlaves
   */
  renderSlaves: function (hostsObj) {
    var headers = this.get('headers');
    var clientHeaders = headers.findProperty('name', 'CLIENT');
    var slaveComponents = this.get('content.slaveComponentHosts');
    if (!slaveComponents) { // we are at this page for the first time
      var recommendations = this.get('content.recommendations');
      // Get all host-component pairs from recommendations
      var componentHostPairs = recommendations.blueprint.host_groups.map(function (group) {
        return group.components.map(function (component) {
          return recommendations.blueprint_cluster_binding.host_groups.findProperty('name', group.name).hosts.map(function (host) {
            return { component: component.name, host: host.fqdn};
          });
        });
      });

      // Flatten results twice because of two map() call before
      componentHostPairs = [].concat.apply([], componentHostPairs);
      componentHostPairs = [].concat.apply([], componentHostPairs);

      var clientComponents = App.get('components.clients');

      hostsObj.forEach(function (host) {
        var checkboxes = host.get('checkboxes');
        checkboxes.forEach(function (checkbox) {
          var recommended = componentHostPairs.some(function (pair) {
            var componentMatch = pair.component === checkbox.component;
            if (checkbox.component === 'CLIENT' && !componentMatch) {
              componentMatch = clientComponents.contains(pair.component);
            }
            return pair.host === host.hostName && componentMatch;
          });
          checkbox.checked = recommended;
        });
      });
    } else {
      this.get('headers').forEach(function (header) {
        var nodes = slaveComponents.findProperty('componentName', header.get('name'));
        if (nodes) {
          nodes.hosts.forEach(function (_node) {
            var node = hostsObj.findProperty('hostName', _node.hostName);
            if (node) {
              node.get('checkboxes').findProperty('title', header.get('label')).set('checked', true);
              node.get('checkboxes').findProperty('title', header.get('label')).set('isInstalled', _node.isInstalled);
            }
          });
        }
      });
    }
    this.selectClientHost(hostsObj);
    return hostsObj;
  },


  /**
   *
   * @param hostsObj
   */
  selectClientHost: function (hostsObj) {
    if (!this.get('isClientsSet')) {
      var nonMasterHost = hostsObj.findProperty('hasMaster', false);
      var clientHost = !!nonMasterHost ? nonMasterHost : hostsObj[hostsObj.length - 1]; // last host
      var clientCheckBox = clientHost.get('checkboxes').findProperty('component', 'CLIENT');
      if (clientCheckBox) {
        clientCheckBox.set('checked', true);
      }
      this.set('isClientsSet', true);
    }
  },

  /**
   * Select checkboxes which correspond to master components
   *
   * @param {Array} hostsObj
   * @return {Array}
   * @method selectMasterComponents
   */
  selectMasterComponents: function (hostsObj) {
    var masterComponentHosts = this.get('content.masterComponentHosts');
    console.log('Master components selected on:', masterComponentHosts.mapProperty('hostName').uniq().join(", "));

    if (masterComponentHosts) {
      masterComponentHosts.forEach(function (item) {
        var host = hostsObj.findProperty('hostName', item.hostName);
        if (host) {
          var checkbox = host.get('checkboxes').findProperty('component', item.component);
          if (checkbox) {
            checkbox.set('checked', true);
          }
        }
      });
    }
    return hostsObj;
  },

  /**
   * Return list of master components for specified <code>hostname</code>
   * @param {string} hostName
   * @return {string[]}
   * @method getMasterComponentsForHost
   */
  getMasterComponentsForHost: function (hostName) {
    return this.get('content.masterComponentHosts').filterProperty('hostName', hostName).mapProperty('component');
  },

  callValidation: function (successCallback) {
    this.callServerSideValidation(successCallback);
  },

  /**
   * Update submit button status
   * @metohd callServerSideValidation
   */
  callServerSideValidation: function (successCallback) {
    var self = this;

    // We do not want to disable Next due to server validation issues - hence commented out line below
    // self.set('submitDisabled', true);

    var selectedServices = App.StackService.find().filterProperty('isSelected').mapProperty('serviceName');
    var installedServices = App.StackService.find().filterProperty('isInstalled').mapProperty('serviceName');
    var services = installedServices.concat(selectedServices).uniq();

    var hostNames = self.get('hosts').mapProperty('hostName');
    var slaveBlueprint = self.getCurrentBlueprint();
    var masterBlueprint = null;
    var invisibleSlavesAndClients = App.StackServiceComponent.find().filter(function (component) {
      return component.get("isSlave") && component.get("isShownOnInstallerSlaveClientPage") === false ||
        component.get("isClient") && component.get("isRequiredOnAllHosts");
    }).mapProperty("componentName");

    if (this.get('isInstallerWizard') || this.get('isAddServiceWizard')) {
      masterBlueprint = self.getCurrentMastersBlueprint();

      var invisibleMasters = [];
      if (this.get('isInstallerWizard')) {
        invisibleMasters = App.StackServiceComponent.find().filterProperty("isMaster").filterProperty("isShownOnInstallerAssignMasterPage", false).mapProperty("componentName");
      } else if (this.get('isAddServiceWizard')) {
        invisibleMasters = App.StackServiceComponent.find().filterProperty("isMaster").filterProperty("isShownOnAddServiceAssignMasterPage", false).mapProperty("componentName");
      }

      var selectedClientComponents = self.get('content.clients').mapProperty('component_name');
      var alreadyInstalledClients = App.get('components.clients').reject(function (c) {
        return selectedClientComponents.contains(c);
      });

      var invisibleComponents = invisibleMasters.concat(invisibleSlavesAndClients).concat(alreadyInstalledClients);

      var invisibleBlueprint = blueprintUtils.filterByComponents(this.get('content.recommendations'), invisibleComponents);
      masterBlueprint = blueprintUtils.mergeBlueprints(masterBlueprint, invisibleBlueprint);
    } else if (this.get('isAddHostWizard')) {
      masterBlueprint = self.getCurrentMasterSlaveBlueprint();
      hostNames = hostNames.concat(App.Host.find().mapProperty("hostName")).uniq();
      slaveBlueprint = blueprintUtils.addComponentsToBlueprint(slaveBlueprint, invisibleSlavesAndClients);
    }

    var bluePrintsForValidation = blueprintUtils.mergeBlueprints(masterBlueprint, slaveBlueprint);
    this.set('content.recommendationsHostGroups', bluePrintsForValidation);

    App.ajax.send({
      name: 'config.validations',
      sender: self,
      data: {
        stackVersionUrl: App.get('stackVersionURL'),
        hosts: hostNames,
        services: services,
        validate: 'host_groups',
        recommendations: bluePrintsForValidation
      },
      success: 'updateValidationsSuccessCallback',
      error: 'updateValidationsErrorCallback'
    }).
      then(function () {
        if (!self.get('submitDisabled') && successCallback) {
          successCallback();
        }
      }
    );
  },

  /**
   * Success-callback for validations request
   * @param {object} data
   * @method updateValidationsSuccessCallback
   */
  updateValidationsSuccessCallback: function (data) {
    var self = this;
    //data = JSON.parse(data); // temporary fix

    var clientComponents = App.get('components.clients');

    this.set('generalErrorMessages', []);
    this.set('generalWarningMessages', []);
    this.get('hosts').setEach('warnMessages', []);
    this.get('hosts').setEach('errorMessages', []);
    this.get('hosts').setEach('anyMessage', false);
    this.get('hosts').forEach(function (host) {
      host.checkboxes.setEach('hasWarnMessage', false);
      host.checkboxes.setEach('hasErrorMessage', false);
    });
    var anyErrors = false;
    var anyGeneralClientErrors = false; // any error/warning for any client component (under "CLIENT" alias)

    var validationData = validationUtils.filterNotInstalledComponents(data);
    validationData.filterProperty('type', 'host-component').filter(function (i) {
      return !(i['component-name'] && App.StackServiceComponent.find().findProperty('componentName', i['component-name']).get('isMaster'));
    }).forEach(function (item) {
      var checkboxWithIssue = null;
      var isGeneralClientValidationItem = clientComponents.contains(item['component-name']); // it is an error/warning for any client component (under "CLIENT" alias)
      var host = self.get('hosts').find(function (h) {
        return h.hostName === item.host && h.checkboxes.some(function (checkbox) {
          var isClientComponent = checkbox.component === "CLIENT" && isGeneralClientValidationItem;
          if (checkbox.component === item['component-name'] || isClientComponent) {
            checkboxWithIssue = checkbox;
            return true;
          } else {
            return false;
          }
        });
      });
      if (host) {
        host.set('anyMessage', true);

        if (item.level === 'ERROR') {
          anyErrors = true;
          host.get('errorMessages').push(item.message);
          checkboxWithIssue.set('hasErrorMessage', true);
        } else if (item.level === 'WARN') {
          host.get('warnMessages').push(item.message);
          checkboxWithIssue.set('hasWarnMessage', true);
        }
      } else {
        var component;
        if (isGeneralClientValidationItem) {
          if (!anyGeneralClientErrors) {
            anyGeneralClientErrors = true;
            component = "Client";
          }
        } else {
          component = item['component-name'];
        }

        if (component || !item['component-name']) {
          var details = "";
          if (item.host) {
            details += " (" + item.host + ")";
          }

          if (item.level === 'ERROR') {
            anyErrors = true;
            self.get('generalErrorMessages').push(item.message + details);
          } else if (item.level === 'WARN') {
            self.get('generalWarningMessages').push(item.message + details);
          }
        }
      }
    });

    // use this.set('submitDisabled', anyErrors); is validation results should block next button
    // It's because showValidationIssuesAcceptBox allow use accept validation issues and continue
    // this.set('submitDisabled', false);
  },

  /**
   * Error-callback for validations request
   * @param {object} jqXHR
   * @param {object} ajaxOptions
   * @param {string} error
   * @param {object} opt
   * @method updateValidationsErrorCallback
   */
  updateValidationsErrorCallback: function (jqXHR, ajaxOptions, error, opt) {
    App.ajax.defaultErrorHandler(jqXHR, opt.url, opt.method, jqXHR.status);
    console.log('Load validations failed');
  },

  /**
   * Composes selected values of comboboxes into blueprint format
   */
  getCurrentBlueprint: function () {
    var self = this;

    var res = {
      blueprint: { host_groups: [] },
      blueprint_cluster_binding: { host_groups: [] }
    };

    var clientComponents = self.get('content.clients').mapProperty('component_name');
    var mapping = self.get('hosts');

    mapping.forEach(function (item, i) {
      var group_name = 'host-group-' + (i+1);

      var host_group = {
        name: group_name,
        components: item.checkboxes.filterProperty('checked', true).map(function (checkbox) {
          if (checkbox.component === "CLIENT") {
            return clientComponents.map(function (client) {
              return { name: client };
            });
          } else {
            return { name: checkbox.component };
          }
        })
      };

      host_group.components = [].concat.apply([], host_group.components);

      var binding = {
        name: group_name,
        hosts: [
          { fqdn: item.hostName }
        ]
      };

      res.blueprint.host_groups.push(host_group);
      res.blueprint_cluster_binding.host_groups.push(binding);
    });

    return res;
  },

  /**
   * Create blueprint from assigned master components to appropriate hosts
   * @returns {Object}
   * @method getCurrentMastersBlueprint
   */
  getCurrentMastersBlueprint: function () {
    var res = {
      blueprint: { host_groups: [] },
      blueprint_cluster_binding: { host_groups: [] }
    };

    var masters = this.get('content.masterComponentHosts');
    var hosts = this.get('content.hosts');

    Em.keys(hosts).forEach(function (host, i) {
      var group_name = 'host-group-' + (i + 1);
      var components = [];
      masters.forEach(function (master) {
        if (master.hostName === host) {
          components.push({
            name: master.component
          });
        }
      });
      res.blueprint.host_groups.push({
        name: group_name,
        components: components
      });
      res.blueprint_cluster_binding.host_groups.push({
        name: group_name,
        hosts: [
          {
            fqdn: host
          }
        ]
      });
    }, this);
    return blueprintUtils.mergeBlueprints(res, this.getCurrentSlaveBlueprint());
  },

  /**
   * In case of any validation issues shows accept dialog box for user which allow cancel and fix issues or continue anyway
   * @metohd submit
   */
  showValidationIssuesAcceptBox: function(callback) {
    var self = this;

    if (self.get('anyWarnings') || self.get('anyErrors')) {
      App.ModalPopup.show({
        primary: Em.I18n.t('common.continueAnyway'),
        header: Em.I18n.t('installer.step6.validationIssuesAttention.header'),
        body: Em.I18n.t('installer.step6.validationIssuesAttention'),
        onPrimary: function () {
          this.hide();
          callback();
        }
      });
    } else {
      callback();
    }
  }
});

});

require.register("controllers/wizard/step7_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
/**
 * By Step 7, we have the following information stored in App.db and set on this
 * controller by the router.
 *
 *   selectedServices: App.db.selectedServices (the services that the user selected in Step 4)
 *   masterComponentHosts: App.db.masterComponentHosts (master-components-to-hosts mapping the user selected in Step 5)
 *   slaveComponentHosts: App.db.slaveComponentHosts (slave-components-to-hosts mapping the user selected in Step 6)
 *
 */

App.WizardStep7Controller = Em.Controller.extend(App.ServerValidatorMixin, {

  name: 'wizardStep7Controller',

  /**
   * Contains all field properties that are viewed in this step
   * @type {object[]}
   */
  stepConfigs: [],

  selectedService: null,

  slaveHostToGroup: null,

  addMiscTabToPage: true,

  /**
   * Is Submit-click processing now
   * @type {bool}
   */
  submitButtonClicked: false,

  isRecommendedLoaded: false,
  /**
   * used in services_config.js view to mark a config with security icon
   */
  secureConfigs: require('data/HDP2/secure_mapping'),

  /**
   * config categories with secure properties
   * use only for add service wizard when security is enabled;
   */
  secureServices: function () {
    return $.extend(true, [], require('data/HDP2/secure_configs'));
  }.property(),

  /**
   * uses for add service - find out is security is enabled
   */
  securityEnabled: function () {
    return App.router.get('mainAdminKerberosController.securityEnabled');
  }.property('App.router.mainAdminKerberosController.securityEnabled'),
  /**
   * If configChangeObserver Modal is shown
   * @type {bool}
   */
  miscModalVisible: false,

  gangliaAvailableSpace: null,

  /**
   * @type {string}
   */
  gangliaMoutDir: '/',

  overrideToAdd: null,

  /**
   * Is installer controller used
   * @type {bool}
   */
  isInstaller: true,

  /**
   * List of config groups
   * @type {object[]}
   */
  configGroups: [],

  /**
   * List of config group to be deleted
   * @type {object[]}
   */
  groupsToDelete: [],

  preSelectedConfigGroup: null,

  /**
   * Currently selected config group
   * @type {object}
   */
  selectedConfigGroup: null,

  /**
   * Config tags of actually installed services
   * @type {array}
   */
  serviceConfigTags: [],

  /**
   * Are advanced configs loaded
   * @type {bool}
   */
  isAdvancedConfigLoaded: true,

  /**
   * Are applied to service configs loaded
   * @type {bool}
   */
  isAppliedConfigLoaded: true,

  isConfigsLoaded: function () {
    return (this.get('isAdvancedConfigLoaded') && this.get('isAppliedConfigLoaded'));
  }.property('isAdvancedConfigLoaded', 'isAppliedConfigLoaded'),

  /**
   * Should Next-button be disabled
   * @type {bool}
   */
  isSubmitDisabled: function () {
    if (!this.get('stepConfigs.length')) return true;
    if (this.get('submitButtonClicked')) return true;
    return (!this.get('stepConfigs').filterProperty('showConfig', true).everyProperty('errorCount', 0) || this.get("miscModalVisible"));
  }.property('stepConfigs.@each.errorCount', 'miscModalVisible', 'submitButtonClicked'),

  /**
   * List of selected to install service names
   * @type {string[]}
   */
  selectedServiceNames: function () {
    return this.get('content.services').filterProperty('isSelected', true).filterProperty('isInstalled', false).mapProperty('serviceName');
  }.property('content.services', 'content.services.@each.isSelected', 'content.services.@each.isInstalled', 'content.stacks.@each.isSelected').cacheable(),

  /**
   * List of installed and selected to install service names
   * @type {string[]}
   */
  allSelectedServiceNames: function () {
    return this.get('content.services').filter(function (service) {
      return service.get('isInstalled') || service.get('isSelected');
    }).mapProperty('serviceName');
  }.property('content.services', 'content.services.@each.isSelected', 'content.services.@each.isInstalled', 'content.stacks.@each.isSelected').cacheable(),

  /**
   * List of installed service names
   * @type {string[]}
   */
  installedServiceNames: function () {
    var serviceNames = this.get('content.services').filterProperty('isInstalled').mapProperty('serviceName');
    if (this.get('content.controllerName') !== 'installerController') {
      serviceNames = serviceNames.filter(function (_serviceName) {
        return !App.get('services.noConfigTypes').contains(_serviceName);
      });
    }
    return serviceNames;
  }.property('content.services').cacheable(),

  /**
   * List of master components
   * @type {Ember.Enumerable}
   */
  masterComponentHosts: function () {
    return this.get('content.masterComponentHosts');
  }.property('content.masterComponentHosts'),

  /**
   * List of slave components
   * @type {Ember.Enumerable}
   */
  slaveComponentHosts: function () {
    return this.get('content.slaveGroupProperties');
  }.property('content.slaveGroupProperties', 'content.slaveComponentHosts'),

  customData: [],

  /**
   * Filter text will be located here
   * @type {string}
   */
  filter: '',

  /**
   * List of filters for config properties to populate filter combobox
   */
  propertyFilters: [
    {
      attributeName: 'isOverridden',
      attributeValue: true,
      caption: 'common.combobox.dropdown.overridden'
    },
    {
      attributeName: 'isFinal',
      attributeValue: true,
      caption: 'common.combobox.dropdown.final'
    },
    {
      attributeName: 'isValid',
      attributeValue: false,
      caption: 'common.combobox.dropdown.issues'
    },
    {
      attributeName: 'warn',
      attributeValue: true,
      caption: 'common.combobox.dropdown.warnings'
    }
  ],

  /**
   * Dropdown menu items in filter combobox
   */
  filterColumns: function () {
    return this.get('propertyFilters').map(function (filter) {
      return Ember.Object.create({
        attributeName: filter.attributeName,
        attributeValue: filter.attributeValue,
        name: this.t(filter.caption),
        selected: false
      });
    }, this);
  }.property('propertyFilters'),

  /**
   * Clear controller's properties:
   *  <ul>
   *    <li>stepConfigs</li>
   *    <li>filter</li>
   *  </ul>
   *  and desect all <code>filterColumns</code>
   * @method clearStep
   */
  clearStep: function () {
    this.set('configValidationGlobalMessage', []);
    this.set('submitButtonClicked', false);
    this.set('isSubmitDisabled', true);
    this.set('isRecommendedLoaded', false);
    this.get('stepConfigs').clear();
    this.set('filter', '');
    this.get('filterColumns').setEach('selected', false);
  },

  /**
   * Load config groups for installed services
   * One ajax-request for each service
   * @param {string[]} servicesNames
   * @method loadInstalledServicesConfigGroups
   */
  loadInstalledServicesConfigGroups: function (servicesNames) {
    servicesNames.forEach(function (serviceName) {
      App.ajax.send({
        name: 'config.tags_and_groups',
        sender: this,
        data: {
          serviceName: serviceName,
          serviceConfigsDef: App.config.get('preDefinedServiceConfigs').findProperty('serviceName', serviceName)
        },
        success: 'loadServiceTagsSuccess'
      });
    }, this);
  },

  /**
   * Create site to tag map. Format:
   * <code>
   *   {
   *    site1: tag1,
   *    site1: tag2,
   *    site2: tag3
   *    ...
   *   }
   * </code>
   * @param {object} desired_configs
   * @param {string[]} sites
   * @returns {object}
   * @private
   * @method _createSiteToTagMap
   */
  _createSiteToTagMap: function (desired_configs, sites) {
    var siteToTagMap = {};
    for (var site in desired_configs) {
      if (desired_configs.hasOwnProperty(site)) {
        if (!!sites[site]) {
          siteToTagMap[site] = desired_configs[site].tag;
        }
      }
    }
    return siteToTagMap;
  },

  /**
   * Load config groups success callback
   * @param {object} data
   * @param {object} opt
   * @param {object} params
   * @method loadServiceTagsSuccess
   */
  loadServiceTagsSuccess: function (data, opt, params) {
    var serviceName = params.serviceName,
      service = this.get('stepConfigs').findProperty('serviceName', serviceName),
      defaultConfigGroupHosts = this.get('wizardController.allHosts').mapProperty('hostName'),
      siteToTagMap = this._createSiteToTagMap(data.Clusters.desired_configs, params.serviceConfigsDef.get('configTypes')),
      selectedConfigGroup;
    this.set('loadedClusterSiteToTagMap', siteToTagMap);

    //parse loaded config groups
    var configGroups = [];
    if (data.config_groups.length) {
      data.config_groups.forEach(function (item) {
        item = item.ConfigGroup;
        if (item.tag === serviceName) {
          var groupHosts = item.hosts.mapProperty('host_name');
          var newConfigGroup = App.ConfigGroup.create({
            id: item.id,
            name: item.group_name,
            description: item.description,
            isDefault: false,
            parentConfigGroup: null,
            service: App.Service.find().findProperty('serviceName', item.tag),
            hosts: groupHosts,
            configSiteTags: []
          });
          groupHosts.forEach(function (host) {
            defaultConfigGroupHosts = defaultConfigGroupHosts.without(host);
          }, this);
          item.desired_configs.forEach(function (config) {
            newConfigGroup.configSiteTags.push(App.ConfigSiteTag.create({
              site: config.type,
              tag: config.tag
            }));
          }, this);
          configGroups.push(newConfigGroup);
        }
      }, this);
    }
    var defaultConfigGroup = App.ConfigGroup.create({
      name: App.format.role(serviceName) + " Default",
      description: "Default cluster level " + serviceName + " configuration",
      isDefault: true,
      hosts: defaultConfigGroupHosts,
      parentConfigGroup: null,
      service: Em.Object.create({
        id: serviceName
      }),
      serviceName: serviceName,
      configSiteTags: []
    });
    if (!selectedConfigGroup) {
      selectedConfigGroup = defaultConfigGroup;
    }
    configGroups = configGroups.sortProperty('name');
    configGroups.unshift(defaultConfigGroup);
    service.set('configGroups', configGroups);
    var loadedGroupToOverrideSiteToTagMap = {};
    var configGroupsWithOverrides = selectedConfigGroup.get('isDefault') ? service.get('configGroups') : [selectedConfigGroup];
    configGroupsWithOverrides.forEach(function (item) {
      var groupName = item.get('name');
      loadedGroupToOverrideSiteToTagMap[groupName] = {};
      item.get('configSiteTags').forEach(function (siteTag) {
        var site = siteTag.get('site');
        loadedGroupToOverrideSiteToTagMap[groupName][site] = siteTag.get('tag');
      }, this);
    }, this);
    this.set('preSelectedConfigGroup', selectedConfigGroup);
    App.config.loadServiceConfigGroupOverrides(service.get('configs'), loadedGroupToOverrideSiteToTagMap, service.get('configGroups'), this.onLoadOverrides, this);
  },

  onLoadOverrides: function (configs) {
    var serviceName = configs[0].serviceName,
      service = this.get('stepConfigs').findProperty('serviceName', serviceName);
    var serviceConfig = App.config.createServiceConfig(serviceName);
    if (serviceConfig.get('serviceName') === 'HDFS') {
      App.config.OnNnHAHideSnn(serviceConfig);
    }
    service.set('selectedConfigGroup', this.get('preSelectedConfigGroup'));
    this.loadComponentConfigs(service.get('configs'), serviceConfig, service);
    service.set('configs', serviceConfig.get('configs'));
  },

  /**
   * Set <code>isEditable</code>-property to <code>serviceConfigProperty</code>
   * Based on user's permissions and selected config group
   * @param {Ember.Object} serviceConfigProperty
   * @param {bool} defaultGroupSelected
   * @returns {Ember.Object} Updated config-object
   * @method _updateIsEditableFlagForConfig
   */
  _updateIsEditableFlagForConfig: function (serviceConfigProperty, defaultGroupSelected) {
    if (App.isAccessible('ADMIN')) {
      if (defaultGroupSelected && !this.get('isHostsConfigsPage') && !Em.get(serviceConfigProperty, 'group')) {
        serviceConfigProperty.set('isEditable', serviceConfigProperty.get('isReconfigurable'));
      } else if (Em.get(serviceConfigProperty, 'group') && Em.get(serviceConfigProperty, 'group.name') == this.get('selectedConfigGroup.name')) {
        serviceConfigProperty.set('isEditable', true);
      } else {
        serviceConfigProperty.set('isEditable', false);
      }
    }
    else {
      serviceConfigProperty.set('isEditable', false);
    }
    return serviceConfigProperty;
  },

  /**
   * Set <code>overrides</code>-property to <code>serviceConfigProperty<code>
   * @param {Ember.Object} serviceConfigProperty
   * @param {Ember.Object} component
   * @return {Ember.Object} Updated config-object
   * @method _updateOverridesForConfig
   */
  _updateOverridesForConfig: function (serviceConfigProperty, component) {

    var overrides = serviceConfigProperty.get('overrides');

    if (Em.isNone(overrides)) {
      serviceConfigProperty.set('overrides', Em.A([]));
      return serviceConfigProperty;
    }
    serviceConfigProperty.set('overrides', null);
    var defaultGroupSelected = component.get('selectedConfigGroup.isDefault');

    // Wrap each override to App.ServiceConfigProperty
    overrides.forEach(function (override) {
      var newSCP = App.ServiceConfigProperty.create(serviceConfigProperty);
      newSCP.set('value', override.value);
      newSCP.set('isOriginalSCP', false); // indicated this is overridden value,
      newSCP.set('parentSCP', serviceConfigProperty);
      if (defaultGroupSelected) {
        var group = component.get('configGroups').findProperty('name', override.group.get('name'));
        // prevent cycle in proto object, clean link
        if (group.get('properties').length == 0) {
          group.set('properties', Em.A([]));
        }
        group.get('properties').push(newSCP);
        newSCP.set('group', override.group);
        newSCP.set('isEditable', false);
      }
      var parentOverridesArray = serviceConfigProperty.get('overrides');
      if (Em.isNone(parentOverridesArray)) {
        parentOverridesArray = Em.A([]);
        serviceConfigProperty.set('overrides', parentOverridesArray);
      }
      serviceConfigProperty.get('overrides').pushObject(newSCP);
    }, this);
    return serviceConfigProperty;
  },

  /**
   * Set configs with overrides, recommended defaults to component
   * @param {Ember.Object[]} configs
   * @param {Ember.Object} componentConfig
   * @param {Ember.Object} component
   * @method loadComponentConfigs
   */
  loadComponentConfigs: function (configs, componentConfig, component) {
    var defaultGroupSelected = component.get('selectedConfigGroup.isDefault');

    configs.forEach(function (serviceConfigProperty) {
      if (!serviceConfigProperty) return;

      if (Em.isNone(serviceConfigProperty.get('isOverridable'))) {
        serviceConfigProperty.set('isOverridable', true);
      }
      this._updateOverridesForConfig(serviceConfigProperty, component);
      this._updateIsEditableFlagForConfig(serviceConfigProperty, defaultGroupSelected);

      componentConfig.get('configs').pushObject(serviceConfigProperty);
      serviceConfigProperty.validate();

    }, this);
    component.get('configGroups').filterProperty('isDefault', false).forEach(function (configGroup) {
      configGroup.set('hash', this.get('wizardController').getConfigGroupHash(configGroup));
    }, this);
    var overrideToAdd = this.get('overrideToAdd');
    if (overrideToAdd) {
      overrideToAdd = componentConfig.get('configs').findProperty('name', overrideToAdd.name);
      if (overrideToAdd) {
        this.addOverrideProperty(overrideToAdd);
        component.set('overrideToAdd', null);
      }
    }
  },

  /**
   *  Resolve dependency between configs.
   *  @param serviceName {String}
   *  @param configs {Ember.Enumerable}
   */
  resolveServiceDependencyConfigs: function (serviceName, configs) {
    switch (serviceName) {
      case 'STORM':
        this.resolveStormConfigs(configs);
        break;
      case 'YARN':
        this.resolveYarnConfigs(configs);
      default:
        break;
    }
  },

  /**
   * Update some Storm configs
   * If Ganglia is selected to install or already installed, Ganglia host should be added to configs
   * @param {Ember.Enumerable} configs
   * @method resolveStormConfigs
   */
  resolveStormConfigs: function (configs) {
    var dependentConfigs, gangliaServerHost, gangliaHostId, hosts;
    dependentConfigs = ['nimbus.childopts', 'supervisor.childopts', 'worker.childopts'];
    // if Ganglia selected or installed, set ganglia host to configs
    if (this.get('installedServiceNames').contains('STORM') && this.get('installedServiceNames').contains('GANGLIA')) return;
    if (this.get('allSelectedServiceNames').contains('GANGLIA') || this.get('installedServiceNames').contains('GANGLIA')) {
      if (this.get('wizardController.name') === 'addServiceController') {
        gangliaServerHost = this.get('wizardController').getDBProperty('masterComponentHosts').findProperty('component', 'GANGLIA_SERVER').hostName;
      } else {
        hosts = this.get('wizardController').getDBProperty('hosts');
        gangliaHostId = this.get('wizardController').getDBProperty('masterComponentHosts').findProperty('component', 'GANGLIA_SERVER').host_id;
        for (var hostName in hosts) {
          if (hosts[hostName].id == gangliaHostId) gangliaServerHost = hosts[hostName].name;
        }
      }
      dependentConfigs.forEach(function (configName) {
        var config = configs.findProperty('name', configName);
        if (!Em.isNone(config.value)) {
          var replaceStr = config.value.match(/.jar=host[^,]+/)[0];
          var replaceWith = replaceStr.slice(0, replaceStr.lastIndexOf('=') - replaceStr.length + 1) + gangliaServerHost;
          config.value = config.defaultValue = config.value.replace(replaceStr, replaceWith);
        }
        config.forceUpdate = true;
      }, this);
    }
  },

  /**
   * Update some Storm configs
   * If SLIDER is selected to install or already installed,
   * some Yarn properties must be changed
   * @param {Ember.Enumerable} configs
   * @method resolveYarnConfigs
   */
  resolveYarnConfigs: function (configs) {
    var cfgToChange = configs.findProperty('name', 'hadoop.registry.rm.enabled');
    if (cfgToChange) {
      var res = this.get('allSelectedServiceNames').contains('SLIDER');
      if (Em.get(cfgToChange, 'value') !== res) {
        Em.set(cfgToChange, 'defaultValue', res);
        Em.set(cfgToChange, 'value', res);
        Em.set(cfgToChange, 'forceUpdate', true);
      }
    }
  },
  /**
   * On load function
   * @method loadStep
   */
  loadStep: function () {
    console.log("TRACE: Loading step7: Configure Services");
    if (!this.get('isConfigsLoaded')) {
      return;
    }
    this.clearStep();

    var self = this;
    //STEP 1: Load advanced configs
    var advancedConfigs = this.get('content.advancedServiceConfig');
    //STEP 2: Load on-site configs by service from local DB
    var storedConfigs = this.get('content.serviceConfigProperties');
    //STEP 3: Merge pre-defined configs with loaded on-site configs
    var configs = App.config.mergePreDefinedWithStored(
      storedConfigs,
      advancedConfigs,
      this.get('selectedServiceNames').concat(this.get('installedServiceNames'))
    );
    App.config.setPreDefinedServiceConfigs(this.get('addMiscTabToPage'));
    //STEP 4: Add advanced configs
    App.config.addAdvancedConfigs(configs, advancedConfigs);
    //STEP 5: Add custom configs
    App.config.addCustomConfigs(configs);

    this.set('groupsToDelete', this.get('wizardController').getDBProperty('groupsToDelete') || []);

    if (this.get('wizardController.name') === 'addServiceController') {
      App.router.get('configurationController').getConfigsByTags(this.get('serviceConfigTags')).done(function (loadedConfigs) {
        self.setInstalledServiceConfigs(self.get('serviceConfigTags'), configs, loadedConfigs, self.get('installedServiceNames'));
        self.applyServicesConfigs(configs, storedConfigs);
      });
    } else {
      this.applyServicesConfigs(configs, storedConfigs);
    }
  },

  applyServicesConfigs: function (configs, storedConfigs) {
    if (this.get('allSelectedServiceNames').contains('YARN')) {
      configs = App.config.fileConfigsIntoTextarea(configs, 'capacity-scheduler.xml');
    }
    var dependendServices = ["STORM", "YARN"];
    dependendServices.forEach(function (serviceName) {
      if (this.get('allSelectedServiceNames').contains(serviceName)) {
        this.resolveServiceDependencyConfigs(serviceName, configs);
      }
    }, this);
    //STEP 6: Distribute configs by service and wrap each one in App.ServiceConfigProperty (configs -> serviceConfigs)
    var self = this;
    this.loadServerSideConfigsRecommendations().always(function () {
      self.set('isRecommendedLoaded', true);
      // format descriptor configs
      if (self.get('securityEnabled') && self.get('wizardController.name') == 'addServiceController') {
        App.config.addKerberosDescriptorConfigs(configs, self.get('wizardController.kerberosDescriptorConfigs') || []);
      }
      self.setStepConfigs(configs, storedConfigs);
      self.checkHostOverrideInstaller();
      self.activateSpecialConfigs();
      self.selectProperService();
      if (self.get('content.skipConfigStep')) {
        App.router.send('next');
      }
    });
  },
  /**
   * Load config groups
   * and (if some services are already installed) load config groups for installed services
   * @method checkHostOverrideInstaller
   */
  checkHostOverrideInstaller: function () {
    if (this.get('wizardController.name') !== 'kerberosWizardController') {
      this.loadConfigGroups(this.get('content.configGroups'));
    }
    if (this.get('installedServiceNames').length > 0) {
      this.loadInstalledServicesConfigGroups(this.get('installedServiceNames'));
    }
  },

  /**
   * Set init <code>stepConfigs</code> value
   * Set <code>selected</code> for addable services if addServiceController is used
   * Remove SNameNode if HA is enabled (and if addServiceController is used)
   * @param {Ember.Object[]} configs
   * @param {Ember.Object[]} storedConfigs
   * @method setStepConfigs
   */
  setStepConfigs: function (configs, storedConfigs) {
    var localDB = {
      hosts: this.get('wizardController.content.hosts'),
      masterComponentHosts: this.get('wizardController.content.masterComponentHosts'),
      slaveComponentHosts: this.get('wizardController.content.slaveComponentHosts')
    };
    var serviceConfigs = App.config.renderConfigs(configs, storedConfigs, this.get('allSelectedServiceNames'), this.get('installedServiceNames'), localDB, this.get('recommendationsConfigs'));
    if (this.get('wizardController.name') === 'addServiceController') {
      serviceConfigs.setEach('showConfig', true);
      serviceConfigs.setEach('selected', false);
      this.get('selectedServiceNames').forEach(function (serviceName) {
        if (!serviceConfigs.findProperty('serviceName', serviceName)) return;
        var selectedService = serviceConfigs.findProperty('serviceName', serviceName).set('selected', true);
      }, this);
      this.get('installedServiceNames').forEach(function (serviceName) {
        var serviceConfigObj = serviceConfigs.findProperty('serviceName', serviceName);
        var isInstallableService = App.StackService.find(serviceName).get('isInstallable');
        if (!isInstallableService) serviceConfigObj.set('showConfig', false);
      }, this);
      // if HA is enabled -> Remove SNameNode, hbase.rootdir should use Name Service ID
      if (App.get('isHaEnabled')) {
        var c = serviceConfigs.findProperty('serviceName', 'HDFS').configs,
          nameServiceId = c.findProperty('name', 'dfs.nameservices'),
          removedConfigs = c.filterProperty('category', 'SECONDARY_NAMENODE');
        removedConfigs.map(function (config) {
          c = c.without(config);
        });
        serviceConfigs.findProperty('serviceName', 'HDFS').configs = c;

        if(this.get('selectedServiceNames').contains('HBASE') && nameServiceId){
          var hRootDir = serviceConfigs.findProperty('serviceName', 'HBASE').configs.findProperty('name','hbase.rootdir'),
            valueToChange = hRootDir.get('value').replace(/\/\/.*:/i, '//' + nameServiceId.get('value') + ':');

          hRootDir.setProperties({
            'value':  valueToChange,
            'defaultValue' : valueToChange
          });
        }
      }
    }

    // Remove Notifications from MISC if it isn't Installer Controller
    if (this.get('wizardController.name') !== 'installerController') {
      var miscService = serviceConfigs.findProperty('serviceName', 'MISC');
      if (miscService) {
        c = miscService.configs;
        removedConfigs = c.filterProperty('category', 'Notifications');
        removedConfigs.map(function (config) {
          c = c.without(config);
        });
        miscService.configs = c;
      }
    }

    this.set('stepConfigs', serviceConfigs);
  },

  /**
   * Select first addable service for <code>addServiceWizard</code>
   * Select first service at all in other cases
   * @method selectProperService
   */
  selectProperService: function () {
    if (this.get('wizardController.name') === 'addServiceController') {
      this.set('selectedService', this.get('stepConfigs').filterProperty('selected', true).get('firstObject'));
    } else {
      this.set('selectedService', this.get('stepConfigs').filterProperty('showConfig', true).objectAt(0));
    }
  },

  /**
   * Load config tags
   * @return {$.ajax|null}
   * @method getConfigTags
   */
  getConfigTags: function () {
    this.set('isAppliedConfigLoaded', false);
    return App.ajax.send({
      name: 'config.tags',
      sender: this,
      success: 'getConfigTagsSuccess'
    });
  },

  /**
   * Success callback for config tags request
   * Updates <code>serviceConfigTags</code> with tags received from server
   * @param {object} data
   * @method getConfigTagsSuccess
   */
  getConfigTagsSuccess: function (data) {
    var installedServiceSites = [];
    App.StackService.find().filterProperty('isInstalled').forEach(function (service) {
      if (!service.get('configTypes')) return;
      var configTypes = Object.keys(service.get('configTypes'));
      installedServiceSites = installedServiceSites.concat(configTypes);
    }, this);
    installedServiceSites = installedServiceSites.uniq();
    var serviceConfigTags = [];
    for (var site in data.Clusters.desired_configs) {
      if (data.Clusters.desired_configs.hasOwnProperty(site)) {
        if (installedServiceSites.contains(site)) {
          serviceConfigTags.push({
            siteName: site,
            tagName: data.Clusters.desired_configs[site].tag,
            newTagName: null
          });
        }
      }
    }
    this.set('serviceConfigTags', serviceConfigTags);
    this.set('isAppliedConfigLoaded', true);
  },

  /**
   * set configs actual values from server
   * @param serviceConfigTags
   * @param configs
   * @param configsByTags
   * @param installedServiceNames
   * @method setInstalledServiceConfigs
   */
  setInstalledServiceConfigs: function (serviceConfigTags, configs, configsByTags, installedServiceNames) {
    var configsMap = {};
    var configTypeMap = {};
    var configMixin = App.get('config');
    var self = this;

    configsByTags.forEach(function (configSite) {
      configsMap[configSite.type] = configSite.properties || {};
    });
    configs.forEach(function (_config) {
      var nonServiceTab = require('data/service_configs');
      var type = _config.filename ? App.config.getConfigTagFromFileName(_config.filename) : null;
      var mappedConfigValue = type && configsMap[type] ? configsMap[type][_config.name] : null;
      if (!Em.isNone(mappedConfigValue) && ((installedServiceNames && installedServiceNames.contains(_config.serviceName) || nonServiceTab.someProperty('serviceName', _config.serviceName)))) {
        // prevent overriding already edited properties
        if (_config.defaultValue != mappedConfigValue) {
          _config.value = mappedConfigValue;
        }
        _config.defaultValue = mappedConfigValue;
        _config.hasInitialValue = true;
        App.config.handleSpecialProperties(_config);
        delete configsMap[type][_config.name];
      }
    });
    self.setServiceDatabaseConfigs(configs);
    //add user properties

    Em.keys(configsMap).forEach(function (filename) {
      Em.keys(configsMap[filename]).forEach(function (propertyName) {
        configs.push(configMixin.addUserProperty({
          id: 'site property',
          name: propertyName,
          serviceName: configMixin.getServiceNameByConfigType(filename),
          value: configsMap[filename][propertyName],
          defaultValue: configsMap[filename][propertyName],
          filename: configMixin.get('filenameExceptions').contains(filename) ? filename : filename + '.xml',
          category: 'Advanced',
          hasInitialValue: true,
          isUserProperty: true,
          isOverridable: true,
          overrides: [],
          isRequired: true,
          isVisible: true,
          showLabel: true
        }, false, []));
      });
    });
  },

  /**
   * Check if Oozie or Hive use existing database then need
   * to restore missed properties
   *
   * @param {Object[]} configs
   **/
  setServiceDatabaseConfigs: function (configs) {
    var serviceNames = this.get('installedServiceNames').filter(function (serviceName) {
      return ['OOZIE', 'HIVE'].contains(serviceName);
    });
    serviceNames.forEach(function (serviceName) {
      var propertyPrefix = serviceName.toLowerCase();
      var dbTypeConfig = configs.findProperty('name', propertyPrefix + '_database');
      if (!/existing/gi.test(dbTypeConfig.value)) return;
      var dbHostName = propertyPrefix + '_hostname';
      var database = dbTypeConfig.value.match(/MySQL|PostgreSQL|Oracle|Derby|MSSQL/gi)[0];
      var dbPrefix = database.toLowerCase();
      if (database.toLowerCase() == 'mssql') {
        if (/integrated/gi.test(dbTypeConfig.value)) {
          dbPrefix = 'mssql_server_2';
        } else {
          dbPrefix = 'mssql_server';
        }
      }
      var propertyName = propertyPrefix + '_existing_' + dbPrefix + '_host';
      var existingDBConfig = configs.findProperty('name', propertyName);
      if (!existingDBConfig.value)
        existingDBConfig.value = existingDBConfig.defaultValue = configs.findProperty('name', dbHostName).value;
    }, this);
  },
  /**
   * Add group ids to <code>groupsToDelete</code>
   * Also save <code>groupsToDelete</code> to local storage
   * @param {Ember.Object[]} groups
   * @method setGroupsToDelete
   */
  setGroupsToDelete: function (groups) {
    var groupsToDelete = this.get('groupsToDelete');
    groups.forEach(function (group) {
      if (group.get('id'))
        groupsToDelete.push({
          id: group.get('id')
        });
    });
    this.get('wizardController').setDBProperty('groupsToDelete', groupsToDelete);
  },

  /**
   * Update <code>configGroups</code> with selected service configGroups
   * Also set default group to first position
   * Update <code>selectedConfigGroup</code> with new default group
   * @method selectedServiceObserver
   */
  selectedServiceObserver: function () {
    if (this.get('selectedService') && (this.get('selectedService.serviceName') !== 'MISC')) {
      var serviceGroups = this.get('selectedService.configGroups');
      serviceGroups.forEach(function (item, index, array) {
        if (item.isDefault) {
          array.unshift(item);
          array.splice(index + 1, 1);
        }
      });
      this.set('configGroups', serviceGroups);
      this.set('selectedConfigGroup', serviceGroups.findProperty('isDefault'));
    }
  }.observes('selectedService.configGroups.@each'),

  /**
   * load default groups for each service in case of initial load
   * @param serviceConfigGroups
   * @method loadConfigGroups
   */
  loadConfigGroups: function (serviceConfigGroups) {
    var services = this.get('stepConfigs');
    var hosts = this.get('wizardController.allHosts').mapProperty('hostName');
    services.forEach(function (service) {
      if (service.get('serviceName') === 'MISC') return;
      var serviceRawGroups = serviceConfigGroups.filterProperty('service.id', service.serviceName);
      if (!serviceRawGroups.length) {
        service.set('configGroups', [
          App.ConfigGroup.create({
            name: service.displayName + " Default",
            description: "Default cluster level " + service.serviceName + " configuration",
            isDefault: true,
            hosts: Em.copy(hosts),
            service: Em.Object.create({
              id: service.serviceName
            }),
            serviceName: service.serviceName
          })
        ]);
      }
      else {
        var defaultGroup = App.ConfigGroup.create(serviceRawGroups.findProperty('isDefault'));
        var serviceGroups = service.get('configGroups');
        serviceRawGroups.filterProperty('isDefault', false).forEach(function (configGroup) {
          var readyGroup = App.ConfigGroup.create(configGroup);
          var wrappedProperties = [];
          readyGroup.get('properties').forEach(function (propertyData) {
            var parentSCP = service.configs.filterProperty('filename', propertyData.filename).findProperty('name', propertyData.name);
            var overriddenSCP = App.ServiceConfigProperty.create(parentSCP);
            overriddenSCP.set('isOriginalSCP', false);
            overriddenSCP.set('parentSCP', parentSCP);
            overriddenSCP.set('group', readyGroup);
            overriddenSCP.setProperties(propertyData);
            wrappedProperties.pushObject(App.ServiceConfigProperty.create(overriddenSCP));
          });
          wrappedProperties.setEach('group', readyGroup);
          readyGroup.set('properties', wrappedProperties);
          readyGroup.set('parentConfigGroup', defaultGroup);
          serviceGroups.pushObject(readyGroup);
        });
        defaultGroup.set('childConfigGroups', serviceGroups);
        serviceGroups.pushObject(defaultGroup);
      }
    });
  },

  /**
   * Click-handler on config-group to make it selected
   * @param {object} event
   * @method selectConfigGroup
   */
  selectConfigGroup: function (event) {
    this.set('selectedConfigGroup', event.context);
  },

  /**
   * Rebuild list of configs switch of config group:
   * on default - display all configs from default group and configs from non-default groups as disabled
   * on non-default - display all from default group as disabled and configs from selected non-default group
   * @method switchConfigGroupConfigs
   */
  switchConfigGroupConfigs: function () {
    var serviceConfigs = this.get('selectedService.configs'),
      selectedGroup = this.get('selectedConfigGroup'),
      overrideToAdd = this.get('overrideToAdd'),
      overrides = [];
    if (!selectedGroup) return;

    var displayedConfigGroups = this._getDisplayedConfigGroups();
    displayedConfigGroups.forEach(function (group) {
      overrides.pushObjects(group.get('properties'));
    });
    serviceConfigs.forEach(function (config) {
      this._setEditableValue(config);
      this._setOverrides(config, overrides);
    }, this);
  }.observes('selectedConfigGroup'),

  /**
   * Get list of config groups to display
   * Returns empty array if no <code>selectedConfigGroup</code>
   * @return {Array}
   * @method _getDisplayedConfigGroups
   */
  _getDisplayedConfigGroups: function () {
    var selectedGroup = this.get('selectedConfigGroup');
    if (!selectedGroup) return [];
    return (selectedGroup.get('isDefault')) ?
      this.get('selectedService.configGroups').filterProperty('isDefault', false) :
      [this.get('selectedConfigGroup')];
  },

  /**
   * Set <code>isEditable</code> property to <code>config</code>
   * @param {Ember.Object} config
   * @return {Ember.Object} updated config-object
   * @method _setEditableValue
   */
  _setEditableValue: function (config) {
    var selectedGroup = this.get('selectedConfigGroup');
    if (!selectedGroup) return config;
    var isEditable = config.get('isEditable'),
      isServiceInstalled = this.get('installedServiceNames').contains(this.get('selectedService.serviceName'));
    if (isServiceInstalled) {
      isEditable = (!isEditable && !config.get('isReconfigurable')) ? false : selectedGroup.get('isDefault');
    }
    else {
      isEditable = selectedGroup.get('isDefault');
    }
    if (config.get('group')) {
      isEditable = config.get('group.name') == this.get('selectedConfigGroup.name');
    }
    config.set('isEditable', isEditable);
    return config;
  },

  /**
   * Set <code>overrides</code> property to <code>config</code>
   * @param {Ember.Object} config
   * @param {Ember.Enumerable} overrides
   * @returns {Ember.Object}
   * @method _setOverrides
   */
  _setOverrides: function (config, overrides) {
    var selectedGroup = this.get('selectedConfigGroup'),
      overrideToAdd = this.get('overrideToAdd'),
      configOverrides = overrides.filterProperty('name', config.get('name'));
    if (!selectedGroup) return config;
    if (overrideToAdd && overrideToAdd.get('name') === config.get('name')) {
      configOverrides.push(this.addOverrideProperty(config));
      this.set('overrideToAdd', null);
    }
    configOverrides.setEach('isEditable', !selectedGroup.get('isDefault'));
    configOverrides.setEach('parentSCP', config);
    config.set('overrides', configOverrides);
    return config;
  },

  /**
   * create overriden property and push it into Config group
   * @param {App.ServiceConfigProperty} serviceConfigProperty
   * @return {App.ServiceConfigProperty}
   * @method addOverrideProperty
   */
  addOverrideProperty: function (serviceConfigProperty) {
    var overrides = serviceConfigProperty.get('overrides') || [];
    var newSCP = App.ServiceConfigProperty.create(serviceConfigProperty);
    var group = this.get('selectedService.configGroups').findProperty('name', this.get('selectedConfigGroup.name'));
    newSCP.set('group', group);
    newSCP.set('value', '');
    newSCP.set('isOriginalSCP', false); // indicated this is overridden value,
    newSCP.set('parentSCP', serviceConfigProperty);
    newSCP.set('isEditable', true);
    group.get('properties').pushObject(newSCP);
    overrides.pushObject(newSCP);
    return newSCP;
  },

  /**
   * @method manageConfigurationGroup
   */
  manageConfigurationGroup: function () {
    App.router.get('mainServiceInfoConfigsController').manageConfigurationGroups(this);
  },

  /**
   * Make some configs visible depending on active services
   * @method activateSpecialConfigs
   */
  activateSpecialConfigs: function () {
    if (this.get('addMiscTabToPage')) {
      var serviceToShow = this.get('selectedServiceNames').concat('MISC');
      var miscConfigs = this.get('stepConfigs').findProperty('serviceName', 'MISC').configs;
      if (this.get('wizardController.name') == "addServiceController") {
        miscConfigs.findProperty('name', 'smokeuser').set('value', this.get('content.smokeuser')).set('isEditable', false);
        miscConfigs.findProperty('name', 'user_group').set('value', this.get('content.group')).set('isEditable', false);
      }
      App.config.miscConfigVisibleProperty(miscConfigs, serviceToShow);
    }
    var wizardController = this.get('wizardController');
    if (wizardController.get('name') === "kerberosWizardController")  {
      var kerberosConfigs =  this.get('stepConfigs').findProperty('serviceName', 'KERBEROS').configs;
      kerberosConfigs.findProperty('name', 'kdc_type').set('value', wizardController.get('content.kerberosOption'));
    }
  },

  /**
   * Check whether hive New MySQL database is on the same host as Ambari server MySQL server
   * @return {$.ajax|null}
   * @method checkMySQLHost
   */
  checkMySQLHost: function () {
    // get ambari database type and hostname
    return App.ajax.send({
      name: 'config.ambari.database.info',
      sender: this,
      success: 'getAmbariDatabaseSuccess'
    });
  },

  /**
   * Success callback for ambari database, get Ambari DB type and DB server hostname, then
   * Check whether hive New MySQL database is on the same host as Ambari server MySQL server
   * @param {object} data
   * @method getAmbariDatabaseSuccess
   */
  getAmbariDatabaseSuccess: function (data) {
    var hiveDBHostname = this.get('stepConfigs').findProperty('serviceName', 'HIVE').configs.findProperty('name', 'hive_ambari_host').value;
    var ambariServiceHostComponents = data.hostComponents;
    if (!!ambariServiceHostComponents.length) {
      var ambariDBInfo = JSON.stringify(ambariServiceHostComponents[0].RootServiceHostComponents.properties);
      this.set('mySQLServerConflict', ambariDBInfo.indexOf('mysql') > 0 && ambariDBInfo.indexOf(hiveDBHostname) > 0);
    } else {
      this.set('mySQLServerConflict', false);
    }
  },

  /**
   * Check if new MySql database was chosen for Hive service
   * and it is not located on the same host as Ambari server
   * that using MySql database too.
   *
   * @method resolveHiveMysqlDatabase
   **/
  resolveHiveMysqlDatabase: function () {
    var hiveService = this.get('content.services').findProperty('serviceName', 'HIVE');
    if (!hiveService || !hiveService.get('isSelected') || hiveService.get('isInstalled')) {
      this.moveNext();
      return;
    }
    var hiveDBType = this.get('stepConfigs').findProperty('serviceName', 'HIVE').configs.findProperty('name', 'hive_database').value;
    if (hiveDBType == 'New MySQL Database') {
      var self = this;
      this.checkMySQLHost().done(function () {
        if (self.get('mySQLServerConflict')) {
          // error popup before you can proceed
          return App.ModalPopup.show({
            header: Em.I18n.t('installer.step7.popup.mySQLWarning.header'),
            bodyClass: Ember.View.extend({
              template: Ember.Handlebars.compile(Em.I18n.t('installer.step7.popup.mySQLWarning.body'))
            }),
            secondary: Em.I18n.t('installer.step7.popup.mySQLWarning.button.gotostep5'),
            primary: Em.I18n.t('installer.step7.popup.mySQLWarning.button.dismiss'),
            onSecondary: function () {
              var parent = this;
              return App.ModalPopup.show({
                header: Em.I18n.t('installer.step7.popup.mySQLWarning.confirmation.header'),
                bodyClass: Ember.View.extend({
                  template: Ember.Handlebars.compile(Em.I18n.t('installer.step7.popup.mySQLWarning.confirmation.body'))
                }),
                onPrimary: function () {
                  this.hide();
                  parent.hide();
                  // go back to step 5: assign masters and disable default navigation warning
                  App.router.get('installerController').gotoStep(5, true);
                }
              });
            }
          });
        } else {
          self.moveNext();
        }
      });
    } else {
      this.moveNext();
    }
  },

  checkDatabaseConnectionTest: function () {
    var deferred = $.Deferred();
    var configMap = [
      {
        serviceName: 'OOZIE',
        ignored: [Em.I18n.t('installer.step7.oozie.database.new')]
      },
      {
        serviceName: 'HIVE',
        ignored: [Em.I18n.t('installer.step7.hive.database.new.mysql'), Em.I18n.t('installer.step7.hive.database.new.postgres')]
      }
    ];
    configMap.forEach(function (config) {
      var isConnectionNotTested = false;
      var service = this.get('content.services').findProperty('serviceName', config.serviceName);
      if (service && service.get('isSelected') && !service.get('isInstalled')) {
        var serviceConfigs = this.get('stepConfigs').findProperty('serviceName', config.serviceName).configs;
        var serviceDatabase = serviceConfigs.findProperty('name', config.serviceName.toLowerCase() + '_database').get('value');
        if (!config.ignored.contains(serviceDatabase)) {
          var filledProperties = App.db.get('tmp', config.serviceName + '_connection');
          if (!filledProperties || App.isEmptyObject(filledProperties)) {
            isConnectionNotTested = true;
          } else {
            for (var key in filledProperties) {
              if (serviceConfigs.findProperty('name', key).get('value') !== filledProperties[key])
                isConnectionNotTested = true;
            }
          }
        }
      }
      config.isCheckIgnored = isConnectionNotTested;
    }, this);
    var ignoredServices = configMap.filterProperty('isCheckIgnored', true);
    if (ignoredServices.length) {
      var displayedServiceNames = ignoredServices.mapProperty('serviceName').map(function (serviceName) {
        return this.get('content.services').findProperty('serviceName', serviceName).get('displayName');
      }, this);
      this.showDatabaseConnectionWarningPopup(displayedServiceNames, deferred);
    }
    else {
      deferred.resolve();
    }
    return deferred;
  },

  showDatabaseConnectionWarningPopup: function (serviceNames, deferred) {
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('installer.step7.popup.database.connection.header'),
      body: Em.I18n.t('installer.step7.popup.database.connection.body').format(serviceNames.join(', ')),
      secondary: Em.I18n.t('common.cancel'),
      primary: Em.I18n.t('common.proceedAnyway'),
      onPrimary: function () {
        deferred.resolve();
        this._super();
      },
      onSecondary: function () {
        self.set('submitButtonClicked', false);
        deferred.reject();
        this._super();
      }
    })
  },
  /**
   * Proceed to the next step
   **/
  moveNext: function () {
    App.router.send('next');
  },

  /**
   * Click-handler on Next button
   * Disable "Submit"-button while server-side processes are running
   * @method submit
   */
  submit: function () {
    if (this.get('isSubmitDisabled')) {
      return;
    }
    var self = this;
    this.set('submitButtonClicked', true);
    this.serverSideValidation().done(function () {
      self.checkDatabaseConnectionTest().done(function () {
        self.resolveHiveMysqlDatabase();
        self.set('submitButtonClicked', false);
      });
    }).fail(function (value) {
      if ("invalid_configs" == value) {
        self.set('submitButtonClicked', false);
      } else {
        // Failed due to validation mechanism failure.
        // Should proceed with other checks
        self.checkDatabaseConnectionTest().done(function () {
          self.resolveHiveMysqlDatabase();
          self.set('submitButtonClicked', false);
        });
      }
    });
  }

});

});

require.register("controllers/wizard/step8_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

App.WizardStep8Controller = Em.Controller.extend(App.AddSecurityConfigs, App.wizardDeployProgressControllerMixin, {

  name: 'wizardStep8Controller',

  /**
   * List of raw data about cluster that should be displayed
   * @type {Array}
   */
  rawContent: [
    {
      config_name: 'Admin',
      display_name: 'Admin Name',
      config_value: ''
    },
    {
      config_name: 'cluster',
      display_name: 'Cluster Name',
      config_value: ''
    },
    {
      config_name: 'hosts',
      display_name: 'Total Hosts',
      config_value: ''
    },
    {
      config_name: 'Repo',
      display_name: 'Local Repository',
      config_value: ''
    }
  ],

  /**
   * List of data about cluster (based on formatted <code>rawContent</code>)
   * @type {Object[]}
   */
  clusterInfo: [],

  /**
   * List of services with components assigned to hosts
   * @type {Object[]}
   */
  services: [],

  /**
   * @type {Object[]}
   */
  configs: [],

  /**
   * All configs
   * @type {Array}
   */
  configMapping: function () {
    return App.config.get('configMapping').all(true);
  }.property('App.config.configMapping'),

  /**
   *
   */
  slaveComponentConfig: null,

  /**
   * Should Submit button be disabled
   * @type {bool}
   */
  isSubmitDisabled: false,

  /**
   * Should Back button be disabled
   * @type {bool}
   */
  isBackBtnDisabled: false,

  /**
   * This flag when turned to true launches deploy progress bar
   */
  isDeployStarted: function() {
    this.get('isSubmitDisabled');
  }.property('isSubmitDisabled'),

  /**
   * Is error appears while <code>ajaxQueue</code> executes
   * @type {bool}
   */
  hasErrorOccurred: false,

  /**
   * Are services installed
   * Used to hide Deploy Progress Bar
   * @type {bool}
   */
  servicesInstalled: false,

  /**
   * List of service config tags
   * @type {Object[]}
   */
  serviceConfigTags: [],

  /**
   * Is cluster security enabled
   * @type {bool}
   */
  securityEnabled: function () {
    return App.router.get('mainAdminKerberosController.securityEnabled');
  }.property('App.router.mainAdminKerberosController.securityEnabled'),

  /**
   * Selected config group
   * @type {Object}
   */
  selectedConfigGroup: null,

  /**
   * List of config groups
   * @type {Object[]}
   */
  configGroups: [],

  /**
   * List of selected but not installed services
   * @type {Object[]}
   */
  selectedServices: function () {
    return this.get('content.services').filterProperty('isSelected', true).filterProperty('isInstalled', false);
  }.property('content.services.@each.isSelected','content.services.@each.isInstalled').cacheable(),

  /**
   * List of installed services
   * @type {Object[]}
   */
  installedServices: function () {
    return this.get('content.services').filterProperty('isInstalled');
  }.property('content.services').cacheable(),

  /**
   * Current cluster name
   * @type {string}
   */
  clusterName: function () {
    return this.get('content.cluster.name');
  }.property('content.cluster.name'),

  /**
   * List of existing cluster names
   * @type {string[]}
   */
  clusterNames: [],

  /**
   * Number of completed cluster delete requests
   * @type {number}
   */
  clusterDeleteRequestsCompleted: 0,

  /**
   * Indicates if all cluster delete requests are completed
   * @type {boolean}
   */
  isAllClusterDeleteRequestsCompleted: function () {
    return this.get('clusterDeleteRequestsCompleted') == this.get('clusterNames.length');
  }.property('clusterDeleteRequestsCompleted'),

  /**
   * Error popup body views for clusters that couldn't be deleted
   * @type {App.AjaxDefaultErrorPopupBodyView[]}
   */
  clusterDeleteErrorViews: [],

  /**
   * Clear current step data
   * @method clearStep
   */
  clearStep: function () {
    this.get('services').clear();
    this.get('configs').clear();
    this.get('clusterInfo').clear();
    this.get('serviceConfigTags').clear();
    this.set('servicesInstalled', false);
    this.set('ajaxQueueLength', 0);
    this.set('ajaxRequestsQueue', App.ajaxQueue.create());
    this.set('ajaxRequestsQueue.finishedCallback', this.ajaxQueueFinished);
    this.get('clusterDeleteErrorViews').clear();
    this.set('clusterDeleteRequestsCompleted', 0);
  },

  /**
   * Load current step data
   * @method loadStep
   */
  loadStep: function () {
    console.log("TRACE: Loading step8: Review Page");
    this.clearStep();
    if (this.get('content.serviceConfigProperties')) {
      this.formatProperties();
      this.loadConfigs();
    }
    this.loadClusterInfo();
    this.loadServices();
    this.set('isSubmitDisabled', false);
    this.set('isBackBtnDisabled', false);
  },

  /**
   * replace whitespace character with coma between directories
   * @method formatProperties
   */
  formatProperties: function () {
    this.get('content.serviceConfigProperties').forEach(function (_configProperty) {
      _configProperty.value = App.config.trimProperty(_configProperty, false);
    });
  },

  /**
   * Remove unused Hive configs
   * @param {Ember.Enumerable} configs
   * @returns {Ember.Enumerable}
   * @method removeHiveConfigs
   */
  removeHiveConfigs: function (configs) {
    var hiveDb = configs.findProperty('name', 'hive_database');
    var hiveDbType = configs.findProperty('name', 'hive_database_type');
    if (hiveDbType) {
      var hive_properties = Em.A([]);

      switch (hiveDb.value) {
        case 'New MySQL Database':
          if (configs.someProperty('name', 'hive_ambari_host')) {
            configs.findProperty('name', 'hive_hostname').value = configs.findProperty('name', 'hive_ambari_host').value;
          }
          hive_properties = Em.A(['hive_existing_mysql_database', 'hive_existing_oracle_database', 'hive_existing_postgresql_database',
            'hive_existing_mssql_server_database', 'hive_existing_mssql_server_2_database']);
          break;
        case 'New PostgreSQL Database':
          if (configs.someProperty('name', 'hive_ambari_host')) {
            configs.findProperty('name', 'hive_hostname').value = configs.findProperty('name', 'hive_ambari_host').value;
          }
          hive_properties = Em.A(['hive_existing_mysql_database', 'hive_existing_oracle_database', 'hive_existing_postgresql_database']);
          break;
        case 'Existing MySQL Database':
          configs.findProperty('name', 'hive_hostname').value = configs.findProperty('name', 'hive_existing_mysql_host').value;
          hive_properties = Em.A(['hive_ambari_database', 'hive_existing_oracle_database', 'hive_existing_postgresql_database',
            'hive_existing_mssql_server_database', 'hive_existing_mssql_server_2_database']);
          break;
        case Em.I18n.t('services.service.config.hive.oozie.postgresql'):
          configs.findProperty('name', 'hive_hostname').value = configs.findProperty('name', 'hive_existing_postgresql_host').value;
          hive_properties = Em.A(['hive_ambari_database', 'hive_existing_oracle_database', 'hive_existing_mysql_database',
            'hive_existing_mssql_server_database', 'hive_existing_mssql_server_2_database']);
          break;
        case 'Existing MSSQL Server database with SQL authentication':
          configs.findProperty('name', 'hive_hostname').value = configs.findProperty('name', 'hive_existing_mssql_server_host').value;
          hive_properties = Em.A(['hive_ambari_database', 'hive_existing_oracle_database', 'hive_existing_postgresql_database',
            'hive_existing_mysql_database', 'hive_existing_mssql_server_database', 'hive_existing_mssql_server_database']);
          break;
        case 'Existing MSSQL Server database with integrated authentication':
          configs.findProperty('name', 'hive_hostname').value = configs.findProperty('name', 'hive_existing_mssql_server_2_host').value;
          hive_properties = Em.A(['hive_ambari_database', 'hive_existing_oracle_database', 'hive_existing_postgresql_database',
            'hive_existing_mysql_database', 'hive_existing_mssql_server_database', 'hive_existing_mssql_server_2_database']);
          break;
        default:
          configs.findProperty('name', 'hive_hostname').value = configs.findProperty('name', 'hive_existing_oracle_host').value;
          hive_properties = Em.A(['hive_ambari_database',  'hive_existing_mysql_database', 'hive_existing_postgresql_database',
            'hive_existing_mssql_server_database', 'hive_existing_mssql_server_2_database']);
          break;
      }

      hive_properties.push('hive_master_hosts');
      hive_properties.forEach(function (property) {
        configs = configs.without(configs.findProperty('name', property));
      });
    }
    return configs;
  },

  /**
   * Remove unused Oozie configs
   * @param {Ember.Enumerable} configs
   * @returns {Ember.Enumerable}
   * @method removeOozieConfigs
   */
  removeOozieConfigs: function (configs) {
    var oozieDb = configs.findProperty('name', 'oozie_database');
    if (oozieDb) {
      var oozie_properties = Em.A(['oozie_ambari_host', 'oozie_ambari_database']);

      switch (oozieDb.value) {
        case 'New Derby Database':
          configs.findProperty('name', 'oozie_hostname').value = configs.findProperty('name', 'oozie_ambari_host').value;
          oozie_properties = Em.A(['oozie_ambari_database', 'oozie_existing_mysql_database', 'oozie_existing_oracle_database',
            'oozie_existing_postgresql_database', 'oozie_existing_mssql_server_database', 'oozie_existing_mssql_server_2_database']);
          break;
        case 'Existing MySQL Database':
          configs.findProperty('name', 'oozie_hostname').value = configs.findProperty('name', 'oozie_existing_mysql_host').value;
          oozie_properties = Em.A(['oozie_ambari_database', 'oozie_existing_oracle_database', 'oozie_derby_database',
            'oozie_existing_postgresql_database', 'oozie_existing_mssql_server_database', 'oozie_existing_mssql_server_2_database']);
          break;
        case Em.I18n.t('services.service.config.hive.oozie.postgresql'):
          configs.findProperty('name', 'oozie_hostname').value = configs.findProperty('name', 'oozie_existing_postgresql_host').value;
          oozie_properties = Em.A(['oozie_ambari_database', 'oozie_existing_oracle_database', 'oozie_existing_mysql_database',
            'oozie_existing_mssql_server_database', 'oozie_existing_mssql_server_2_database']);
          break;
        case 'Existing MSSQL Server database with SQL authentication':
          configs.findProperty('name', 'oozie_hostname').value = configs.findProperty('name', 'oozie_existing_mssql_server_host').value;
          oozie_properties = Em.A(['oozie_existing_oracle_database', 'oozie_existing_postgresql_database',
            'oozie_existing_mysql_database', 'oozie_existing_mssql_server_database', 'oozie_existing_mssql_server_2_database']);
          break;
        case 'Existing MSSQL Server database with integrated authentication':
          configs.findProperty('name', 'oozie_hostname').value = configs.findProperty('name', 'oozie_existing_mssql_server_2_host').value;
          oozie_properties = Em.A(['oozie_existing_oracle_database', 'oozie_existing_postgresql_database',
            'oozie_existing_mysql_database', 'oozie_existing_mssql_server_database', 'oozie_existing_mssql_server_database']);
          break;
        default:
          configs.findProperty('name', 'oozie_hostname').value = configs.findProperty('name', 'oozie_existing_oracle_host').value;
          oozie_properties = Em.A(['oozie_ambari_database', 'oozie_existing_mysql_database', 'oozie_derby_database',
            'oozie_existing_postgresql_database', 'oozie_existing_mssql_server_database', 'oozie_existing_mssql_server_2_database']);
          break;
      }
      oozie_properties.forEach(function (property) {
        configs = configs.without(configs.findProperty('name', property));
      });
    }
    return configs;
  },

  /**
   * Load all site properties
   * @method loadConfigs
   */
  loadConfigs: function () {
    //storedConfigs contains custom configs as well
    var configs = this.get('content.serviceConfigProperties');
    if (configs.someProperty('name', 'hive_database')) {
      configs = this.removeHiveConfigs(configs);
    }
    if (configs.someProperty('name', 'oozie_database')) {
      configs = this.removeOozieConfigs(configs);
    }
    configs.forEach(function (_config) {
      _config.value = (typeof _config.value === "boolean") ? _config.value.toString() : _config.value;
    });
    var mappedConfigs = App.config.excludeUnsupportedConfigs(this.get('configMapping'), this.get('selectedServices').mapProperty('serviceName'));
    var uiConfigs = this.loadUiSideConfigs(mappedConfigs);
    var customGroupConfigs = [];
    var allConfigs = configs.concat(uiConfigs).filter(function (config) {
      if (config.group) {
        customGroupConfigs.push(config);
        return false;
      } else {
        return true;
      }
    });
    this.set('customNonDefaultGroupConfigs', customGroupConfigs);
    this.set('configs', allConfigs);
  },

  /**
   * Load UI configs
   * @param {Array} configMapping
   * @return {Array}
   * @method loadUiSideConfigs
   */
  loadUiSideConfigs: function (configMapping) {
    var uiConfig = [];
    var configs = configMapping.filterProperty('foreignKey', null);
    this.addDynamicProperties(configs);
    configs.forEach(function (_config) {
      var valueWithOverrides = this.getGlobConfigValueWithOverrides(_config.templateName, _config.value, _config.name);
      uiConfig.pushObject({
        "id": "site property",
        "name": _config.name,
        "value": valueWithOverrides.value,
        "filename": _config.filename,
        "overrides": valueWithOverrides.overrides
      });
    }, this);
    var dependentConfig = $.extend(true, [], configMapping.filterProperty('foreignKey'));
    dependentConfig.forEach(function (_config) {
      App.config.setConfigValue(uiConfig, this.get('content.serviceConfigProperties'), _config);
      if(!_config.noMatchSoSkipThisConfig)
        uiConfig.pushObject({
          "id": "site property",
          "name": _config._name || _config.name,
          "value": _config.value,
          "filename": _config.filename
        });
    }, this);
    return uiConfig;
  },

  /**
   * Add dynamic properties to configs
   * @param {Array} configs
   * @method addDynamicProperties
   */
  addDynamicProperties: function (configs) {
    var templetonHiveProperty = this.get('content.serviceConfigProperties').someProperty('name', 'templeton.hive.properties');
    if (!templetonHiveProperty) {
      configs.pushObject({
        "name": "templeton.hive.properties",
        "templateName": ["hive.metastore.uris"],
        "foreignKey": null,
        "value": "hive.metastore.local=false,hive.metastore.uris=<templateName[0]>,hive.metastore.sasl.enabled=yes,hive.metastore.execute.setugi=true,hive.metastore.warehouse.dir=/apps/hive/warehouse",
        "filename": "webhcat-site.xml"
      });
    }
  },

  /**
   * Format <code>content.hosts</code> from Object to Array
   * @returns {Array}
   * @method getRegisteredHosts
   */
  getRegisteredHosts: function () {
    var allHosts = this.get('content.hosts');
    var hosts = [];
    for (var hostName in allHosts) {
      if (allHosts.hasOwnProperty(hostName)) {
        if (allHosts[hostName].bootStatus == 'REGISTERED') {
          allHosts[hostName].hostName = allHosts[hostName].name;
          hosts.pushObject(allHosts[hostName]);
        }
      }
    }
    return hosts;
  },

  /**
   * Set all site property that are derived from other puppet-variable
   * @param {String} templateName
   * @param {String} expression
   * @param {String} name
   * @return {Object}
   * example: <code>{
   *   value: '...',
   *   overrides: [
   *    {
   *      value: 'v1',
   *      hosts: ['h1', 'h2']
   *    },
   *    {
   *      value: 'v2',
   *      hosts: ['h2', 'h3']
   *    },
   *    ....
   *   ]
   * }</code>
   * @method getGlobConfigValueWithOverrides
   */
  getGlobConfigValueWithOverrides: function (templateName, expression, name) {
    var express = expression.match(/<(.*?)>/g);
    var value = expression;
    if (express == null) {
      return { value: expression, overrides: []};      // if site property do not map any global property then return the value
    }
    var overrideHostToValue = {};
    express.forEach(function (_express) {
      //console.log("The value of template is: " + _express);
      var index = parseInt(_express.match(/\[([\d]*)(?=\])/)[1]);
      if (this.get('configs').someProperty('name', templateName[index])) {
        //console.log("The name of the variable is: " + this.get('content.serviceConfigProperties').findProperty('name', templateName[index]).name);
        var globalObj = this.get('configs').findProperty('name', templateName[index]);
        var globValue = globalObj.value;
        // Hack for templeton.zookeeper.hosts
        var preReplaceValue = null;
        if (value !== null) {   // if the property depends on more than one template name like <templateName[0]>/<templateName[1]> then don't proceed to the next if the prior is null or not found in the global configs
          preReplaceValue = value;
          value = App.config.replaceConfigValues(name, _express, value, globValue);
        }
        if (globalObj.overrides != null) {
          globalObj.overrides.forEach(function (override) {
            var ov = override.value;
            var hostsArray = override.hosts;
            hostsArray.forEach(function (host) {
              if (!(host in overrideHostToValue)) {
                overrideHostToValue[host] = App.config.replaceConfigValues(name, _express, preReplaceValue, ov);
              } else {
                overrideHostToValue[host] = App.config.replaceConfigValues(name, _express, overrideHostToValue[host], ov);
              }
            }, this);
          }, this);
        }
      } else {
        /*
         console.log("ERROR: The variable name is: " + templateName[index]);
         console.log("ERROR: mapped config from configMapping file has no corresponding variable in " +
         "content.serviceConfigProperties. Two possible reasons for the error could be: 1) The service is not selected. " +
         "and/OR 2) The service_config metadata file has no corresponding global var for the site property variable");
         */
        value = null;
      }
    }, this);

    var valueWithOverrides = {
      value: value,
      overrides: []
    };
    var overrideValueToHostMap = {};
    if (!jQuery.isEmptyObject(overrideHostToValue)) {
      for (var host in overrideHostToValue) {
        var hostVal = overrideHostToValue[host];
        if (!(hostVal in overrideValueToHostMap)) {
          overrideValueToHostMap[hostVal] = [];
        }
        overrideValueToHostMap[hostVal].push(host);
      }
    }
    for (var val in overrideValueToHostMap) {
      valueWithOverrides.overrides.push({
        value: val,
        hosts: overrideValueToHostMap[val]
      });
    }
    return valueWithOverrides;
  },

  /**
   * Load all info about cluster to <code>clusterInfo</code> variable
   * @method loadClusterInfo
   */
  loadClusterInfo: function () {

    //Admin name
    var admin = this.rawContent.findProperty('config_name', 'Admin');
    admin.config_value = App.db.getLoginName();
    console.log("STEP8: the value of content cluster name: " + App.db.getLoginName());
    if (admin.config_value) {
      this.get('clusterInfo').pushObject(Ember.Object.create(admin));
    }

    // cluster name
    var cluster = this.rawContent.findProperty('config_name', 'cluster');
    cluster.config_value = this.get('content.cluster.name');
    console.log("STEP8: the value of content cluster name: " + this.get('content.cluster.name'));
    this.get('clusterInfo').pushObject(Ember.Object.create(cluster));

    //hosts
    var newHostsCount = 0;
    var totalHostsCount = 0;
    var hosts = this.get('content.hosts');
    for (var hostName in hosts) {
      newHostsCount += ~~(!hosts[hostName].isInstalled);
      totalHostsCount++;
    }

    var totalHostsObj = this.rawContent.findProperty('config_name', 'hosts');
    totalHostsObj.config_value = totalHostsCount + ' (' + newHostsCount + ' new)';
    this.get('clusterInfo').pushObject(Em.Object.create(totalHostsObj));

    //repo
    if (['addHostController', 'addServiceController'].contains(this.get('content.controllerName'))) {
      this.loadRepoInfo();
    } else {
      // from install wizard
      var selectedStack = App.Stack.find().findProperty('isSelected');
      var allRepos = [];
      if (selectedStack && selectedStack.get('operatingSystems')) {
        selectedStack.get('operatingSystems').forEach(function (os) {
          if (os.get('isSelected')) {
            os.get('repositories').forEach(function(repo) {
              allRepos.push(Em.Object.create({
                base_url: repo.get('baseUrl'),
                os_type: repo.get('osType'),
                repo_id: repo.get('repoId')
              }));
            }, this);
          }
        }, this);
      }
      allRepos.set('display_name', Em.I18n.t("installer.step8.repoInfo.displayName"));
      this.get('clusterInfo').set('repoInfo', allRepos);
    }
  },

  /**
   * Load repo info for add Service/Host wizard review page
   * @return {$.ajax|null}
   * @method loadRepoInfo
   */
  loadRepoInfo: function () {
    var nameVersionCombo = App.get('currentStackVersion').split('-');
    return App.ajax.send({
      name: 'cluster.load_repositories',
      sender: this,
      data: {
        stackName: nameVersionCombo[0],
        stackVersion: nameVersionCombo[1]
      },
      success: 'loadRepoInfoSuccessCallback',
      error: 'loadRepoInfoErrorCallback'
    });
  },

  /**
   * Save all repo base URL of all OS type to <code>repoInfo<code>
   * @param {object} data
   * @method loadRepoInfoSuccessCallback
   */
  loadRepoInfoSuccessCallback: function (data) {
    var allRepos = [];
    data.items.forEach(function (os) {
      os.repositories.forEach(function (repository) {
        allRepos.push(Em.Object.create({
          base_url: repository.Repositories.base_url,
          os_type: repository.Repositories.os_type,
          repo_id: repository.Repositories.repo_id
        }));
      });
    }, this);
    allRepos.set('display_name', Em.I18n.t("installer.step8.repoInfo.displayName"));
    this.get('clusterInfo').set('repoInfo', allRepos);
  },

  /**
   * @param {object} request
   * @method loadRepoInfoErrorCallback
   */
  loadRepoInfoErrorCallback: function (request) {
    console.log('Error message is: ' + request.responseText);
    var allRepos = [];
    allRepos.set('display_name', Em.I18n.t("installer.step8.repoInfo.displayName"));
    this.get('clusterInfo').set('repoInfo', allRepos);
  },

  /**
   * Load all info about services to <code>services</code> variable
   * @method loadServices
   */
  loadServices: function () {
    this.get('selectedServices').filterProperty('isHiddenOnSelectServicePage', false).forEach(function (service) {
      console.log('INFO: step8: Name of the service from getService function: ' + service.get('serviceName'));
      var serviceObj = Em.Object.create({
        service_name: service.get('serviceName'),
        display_name: service.get('displayNameOnSelectServicePage'),
        service_components: Em.A([])
      });
      service.get('serviceComponents').forEach(function (component) {
        // show clients for services that have only clients components
        if ((component.get('isClient') || component.get('isRequiredOnAllHosts')) && !service.get('isClientOnlyService')) return;
        // skip components that was hide on assign master page
        if (component.get('isMaster') && !component.get('isShownOnInstallerAssignMasterPage')) return;
        // no HA component
        if (component.get('isHAComponentOnly')) return;
        var displayName;
        if (component.get('isClient')) {
          displayName = Em.I18n.t('common.clients')
        } else {
          // remove service name from component display name
          displayName = App.format.role(component.get('componentName')).replace(new RegExp('^' + service.get('serviceName') + '\\s', 'i'), '');
        }
        serviceObj.get('service_components').pushObject(Em.Object.create({
          component_name: component.get('isClient') ? Em.I18n.t('common.client').toUpperCase() : component.get('componentName'),
          display_name: displayName,
          component_value: this.assignComponentHosts(component)
        }));
      }, this);
      if (service.get('customReviewHandler')) {
        for (var displayName in service.get('customReviewHandler')) {
          serviceObj.get('service_components').pushObject(Em.Object.create({
            display_name: displayName,
            component_value: this.assignComponentHosts(Em.Object.create({
              customHandler: service.get('customReviewHandler.' + displayName)
            }))
          }));
        }
      }
      this.get('services').pushObject(serviceObj);
    }, this);
  },

  /**
   * Set <code>component_value</code> property to <code>component</code>
   * @param {Em.Object} component
   * @return {String}
   * @method assignComponentHosts
   */
  assignComponentHosts: function (component) {
    var componentValue;
    if (component.get('customHandler')) {
      componentValue = this[component.get('customHandler')].call(this, component);
    }
    else {
      if (component.get('isMaster')) {
        componentValue = this.getMasterComponentValue(component.get('componentName'));
      }
      else {
        console.log(' --- ---INFO: step8: NOT component isMaster');
        var componentName = component.get('isClient') ? Em.I18n.t('common.client').toUpperCase() : component.get('componentName');
        var hostsLength = this.get('content.slaveComponentHosts')
          .findProperty('componentName', componentName).hosts.length;
        componentValue = hostsLength + Em.I18n.t('installer.step8.host' + ((hostsLength > 1) ? 's' : ''));
      }
    }
    return componentValue;
  },

  getMasterComponentValue: function (componentName) {
    var masterComponents = this.get('content.masterComponentHosts');
    var hostsCount = masterComponents.filterProperty('component', componentName).length;
    return stringUtils.pluralize(hostsCount,
      masterComponents.findProperty('component', componentName).hostName,
        hostsCount + ' ' + Em.I18n.t('installer.step8.hosts'));
  },

  /**
   * Set displayed Hive DB value based on DB type
   * @method loadHiveDbValue
   */
  loadHiveDbValue: function () {
    var db, serviceConfigProperties = this.get('wizardController').getDBProperty('serviceConfigProperties'),
      hiveDb = serviceConfigProperties.findProperty('name', 'hive_database');
    if (hiveDb.value === 'New MySQL Database') {
      return 'MySQL (New Database)';
    } else if (hiveDb.value === 'New PostgreSQL Database') {
      return 'Postgres (New Database)';
    }
    else {
      if (hiveDb.value === 'Existing MySQL Database') {
        db = serviceConfigProperties.findProperty('name', 'hive_existing_mysql_database');
        return db.value + ' (' + hiveDb.value + ')';
      }
      else {
        if (hiveDb.value === Em.I18n.t('services.service.config.hive.oozie.postgresql')) {
          db = serviceConfigProperties.findProperty('name', 'hive_existing_postgresql_database');
          return db.value + ' (' + hiveDb.value + ')';
        }
        else {
          if (hiveDb.value === 'Existing MSSQL Server database with SQL authentication') {
            db = serviceConfigProperties.findProperty('name', 'hive_existing_mssql_server_database');
            return db.value + ' (' + hiveDb.value + ')';
          }
          else {
            if (hiveDb.value === 'Existing MSSQL Server database with integrated authentication') {
              db = serviceConfigProperties.findProperty('name', 'hive_existing_mssql_server_2_database');
              return db.value + ' (' + hiveDb.value + ')';
            }
            else {
              // existing oracle database
              db = serviceConfigProperties.findProperty('name', 'hive_existing_oracle_database');
              return db.value + ' (' + hiveDb.value + ')';
            }
          }
        }
      }
    }
  },

  /**
   * Set displayed HBase master value
   * @param {Object} hbaseMaster
   * @method loadHbaseMasterValue
   */
  loadHbaseMasterValue: function (hbaseMaster) {
    var hbaseHostName = this.get('content.masterComponentHosts').filterProperty('component', hbaseMaster.component_name);
    if (hbaseHostName.length == 1) {
      hbaseMaster.set('component_value', hbaseHostName[0].hostName);
    } else {
      hbaseMaster.set('component_value', hbaseHostName[0].hostName + " " + Em.I18n.t('installer.step8.other').format(hbaseHostName.length - 1));
    }
  },

  /**
   * Set displayed ZooKeeper Server value
   * @param {Object} serverComponent
   * @method loadZkServerValue
   */
  loadZkServerValue: function (serverComponent) {
    var zkHostNames = this.get('content.masterComponentHosts').filterProperty('component', serverComponent.component_name).length;
    var hostSuffix;
    if (zkHostNames === 1) {
      hostSuffix = Em.I18n.t('installer.step8.host');
    } else {
      hostSuffix = Em.I18n.t('installer.step8.hosts');
    }
    serverComponent.set('component_value', zkHostNames + hostSuffix);
  },

  /**
   * Set displayed Oozie DB value based on DB type
   * @method loadOozieDbValue
   */
  loadOozieDbValue: function () {
    var db, oozieDb = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'oozie_database');
    if (oozieDb.value === 'New Derby Database') {
      db = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'oozie_derby_database');
      return db.value + ' (' + oozieDb.value + ')';
    }
    else {
      if (oozieDb.value === 'Existing MySQL Database') {
        db = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'oozie_existing_mysql_database');
        return db.value + ' (' + oozieDb.value + ')';
      }
      else {
        if (oozieDb.value === Em.I18n.t('services.service.config.hive.oozie.postgresql')) {
          db = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'oozie_existing_postgresql_database');
          return db.value + ' (' + oozieDb.value + ')';
        }
        else {
          if (oozieDb.value === 'Existing MSSQL Server database with SQL authentication') {
            db = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'oozie_existing_mssql_server_database');
            return db.value + ' (' + oozieDb.value + ')';
          }
          else {
            if (oozieDb.value === 'Existing MSSQL Server database with integrated authentication') {
              db = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'oozie_existing_mssql_server_2_database');
              return db.value + ' (' + oozieDb.value + ')';
            }
            else {
              // existing oracle database
              db = this.get('wizardController').getDBProperty('serviceConfigProperties').findProperty('name', 'oozie_existing_oracle_database');
              return db.value + ' (' + oozieDb.value + ')';
            }
          }
        }
      }
    }
  },

  /**
   * Onclick handler for <code>next</code> button
   * @method submit
   * @return {void}
   */
  submit: function () {
    if (!this.get('isSubmitDisabled')) {
      if (this.get('content.controllerName') != 'installerController' && this.get('securityEnabled')) {
        App.get('router.mainAdminKerberosController').getKDCSessionState(this.submitProceed.bind(this));
      } else {
        this.submitProceed();
      }
    }
  },
  /**
   * Update configurations for installed services.
   * Do separated PUT-request for each siteName for each service
   *
   * @param {Array} fileNamesToUpdate - file names that should be updated
   * @method updateConfigurations
   */
  updateConfigurations: function (fileNamesToUpdate) {
    var configurationController = App.router.get('mainServiceInfoConfigsController');
    var configs = this.get('configs').slice(0);
    var configsMap = [];
    fileNamesToUpdate.forEach(function (fileName) {
      if (!fileName || /^(core)/.test(fileName)) return;
      var tagName = 'version' + (new Date).getTime();
      var configsToSave = configs.filterProperty('filename', fileName);
      configsToSave.forEach(function (item) {
        item.value = App.config.trimProperty(item, false);
      });
      configsMap.push(configurationController.createSiteObj(fileName.replace(".xml", ""), tagName, configsToSave));
    }, this);

    if (!configsMap.length) return;
    this.applyConfigurationsToCluster(configsMap);
  },
  /**
   * Prepare <code>ajaxQueue</code> and start to execute it
   * @method submitProceed
   */
  submitProceed: function () {
    var self = this;
    this.setProperties({
      isSubmitDisabled: true,
      isBackBtnDisabled: true,
      clusterDeleteRequestsCompleted: 0
    });
    this.get('clusterDeleteErrorViews').clear();
    if (this.get('content.controllerName') == 'addHostController') {
      App.router.get('addHostController').setLowerStepsDisable(4);
    }

    // checkpoint the cluster status on the server so that the user can resume from where they left off
    switch (this.get('content.controllerName')) {
      case 'installerController':
        App.clusterStatus.setClusterStatus({
          clusterName: this.get('clusterName'),
          clusterState: 'CLUSTER_DEPLOY_PREP_2',
          wizardControllerName: this.get('content.controllerName'),
          localdb: App.db.data
        });
        break;
      case 'addHostController':
        App.clusterStatus.setClusterStatus({
          clusterName: this.get('clusterName'),
          clusterState: 'ADD_HOSTS_DEPLOY_PREP_2',
          wizardControllerName: this.get('content.controllerName'),
          localdb: App.db.data
        });
        break;
      case 'addServiceController':
        App.clusterStatus.setClusterStatus({
          clusterName: this.get('clusterName'),
          clusterState: 'ADD_SERVICES_DEPLOY_PREP_2',
          wizardControllerName: this.get('content.controllerName'),
          localdb: App.db.data
        });
        break;
      default:
        break;
    }
    // delete any existing clusters to start from a clean slate
    // before creating a new cluster in install wizard
    // TODO: modify for multi-cluster support
    this.getExistingClusterNames().complete(function () {
      var clusterNames = self.get('clusterNames');
      if (self.get('content.controllerName') == 'installerController' && (!App.get('testMode')) && clusterNames.length) {
        self.deleteClusters(clusterNames);
      } else {
        self.startDeploy();
      }
    });
  },

  /**
   * Get list of existing cluster names
   * @returns {object|null}
   * returns an array of existing cluster names.
   * returns an empty array if there are no existing clusters.
   * @method getExistingClusterNames
   */
  getExistingClusterNames: function () {
    return App.ajax.send({
      name: 'wizard.step8.existing_cluster_names',
      sender: this,
      success: 'getExistingClusterNamesSuccessCallBack',
      error: 'getExistingClusterNamesErrorCallback'
    });
  },

  /**
   * Save received list to <code>clusterNames</code>
   * @param {Object} data
   * @method getExistingClusterNamesSuccessCallBack
   */
  getExistingClusterNamesSuccessCallBack: function (data) {
    var clusterNames = data.items.mapProperty('Clusters.cluster_name');
    console.log("Got existing cluster names: " + clusterNames);
    this.set('clusterNames', clusterNames);
  },

  /**
   * If error appears, set <code>clusterNames</code> to <code>[]</code>
   * @method getExistingClusterNamesErrorCallback
   */
  getExistingClusterNamesErrorCallback: function () {
    console.log("Failed to get existing cluster names");
    this.set('clusterNames', []);
  },

  /**
   * Delete cluster by name
   * One request for one cluster!
   * @param {string[]} clusterNames
   * @method deleteClusters
   */
  deleteClusters: function (clusterNames) {
    this.get('clusterDeleteErrorViews').clear();
    clusterNames.forEach(function (clusterName, index) {
      App.ajax.send({
        name: 'common.delete.cluster',
        sender: this,
        data: {
          name: clusterName,
          isLast: index == clusterNames.length - 1
        },
        success: 'deleteClusterSuccessCallback',
        error: 'deleteClusterErrorCallback'
      });
    }, this);

  },

  /**
   * Method to execute after successful cluster deletion
   * @method deleteClusterSuccessCallback
   */
  deleteClusterSuccessCallback: function () {
    this.incrementProperty('clusterDeleteRequestsCompleted');
    if (this.get('isAllClusterDeleteRequestsCompleted')) {
      if (this.get('clusterDeleteErrorViews.length')) {
        this.showDeleteClustersErrorPopup();
      } else {
        this.startDeploy();
      }
    }
  },

  /**
   * Method to execute after failed cluster deletion
   * @param {object} request
   * @param {string} ajaxOptions
   * @param {string} error
   * @param {object} opt
   * @method deleteClusterErrorCallback
   */
  deleteClusterErrorCallback: function (request, ajaxOptions, error, opt) {
    this.incrementProperty('clusterDeleteRequestsCompleted');
    try {
      var json = $.parseJSON(request.responseText);
      var message = json.message;
    } catch (err) {
    }
    this.get('clusterDeleteErrorViews').pushObject(App.AjaxDefaultErrorPopupBodyView.create({
      url: opt.url,
      type: opt.type,
      status: request.status,
      message: message
    }));
    if (this.get('isAllClusterDeleteRequestsCompleted')) {
      this.showDeleteClustersErrorPopup();
    }
  },

  /**
   * Show error popup if cluster deletion failed
   * @method showDeleteClustersErrorPopup
   */
  showDeleteClustersErrorPopup: function () {
    var self = this;
    this.setProperties({
      isSubmitDisabled: false,
      isBackBtnDisabled: false
    });
    App.ModalPopup.show({
      header: Em.I18n.t('common.error'),
      secondary: false,
      onPrimary: function () {
        this.hide();
      },
      bodyClass: Em.ContainerView.extend({
        childViews: self.get('clusterDeleteErrorViews')
      })
    });
  },

  /**
   * updates kerberosDescriptorConfigs
   * @method updateKerberosDescriptor
   */
  updateKerberosDescriptor: function() {
    var kerberosDescriptor = App.db.get('KerberosWizard', 'kerberosDescriptorConfigs');
    this.addRequestToAjaxQueue({
      name: 'admin.kerberos.cluster.artifact.update',
      data: {
        artifactName: 'kerberos_descriptor',
        data: {
          artifact_data: kerberosDescriptor
        }
      }
    });
  },
  /**
   * Start deploy process
   * @method startDeploy
   */
  startDeploy: function () {
    this.createCluster();
    this.createSelectedServices();
    if (this.get('content.controllerName') !== 'addHostController') {
      if (this.get('content.controllerName') === 'addServiceController') {
        if (this.get('securityEnabled')) {
          this.updateKerberosDescriptor();
        }
      }
      if (this.get('wizardController').getDBProperty('fileNamesToUpdate') && this.get('wizardController').getDBProperty('fileNamesToUpdate').length) {
        this.updateConfigurations(this.get('wizardController').getDBProperty('fileNamesToUpdate'));
      }
      this.createConfigurations();
      this.applyConfigurationsToCluster(this.get('serviceConfigTags'));
    }
    this.createComponents();
    this.registerHostsToCluster();
    this.createConfigurationGroups();
    this.createMasterHostComponents();
    this.createSlaveAndClientsHostComponents();
    if (this.get('content.controllerName') === 'addServiceController') {
      this.createAdditionalClientComponents();
    }
    this.createAdditionalHostComponents();

    this.set('ajaxQueueLength', this.get('ajaxRequestsQueue.queue.length'));
    this.get('ajaxRequestsQueue').start();
  },

  /**
   * *******************************************************************
   * The following create* functions are called upon submitting Step 8.
   * *******************************************************************
   */

  /**
   * Create cluster using selected stack version
   * Queued request
   * @method createCluster
   */
  createCluster: function () {
    if (this.get('content.controllerName') !== 'installerController') return;
    var stackVersion = (this.get('content.installOptions.localRepo')) ? App.currentStackVersion.replace(/(-\d+(\.\d)*)/ig, "Local$&") : App.currentStackVersion;
    this.addRequestToAjaxQueue({
      name: 'wizard.step8.create_cluster',
      data: {
        data: JSON.stringify({ "Clusters": {"version": stackVersion }})
      },
      success: 'createClusterSuccess'
    });
  },

  createClusterSuccess: function (data, xhr, params) {
    App.set('clusterName', params.cluster)
  },

  /**
   * Create selected to install services
   * Queued request
   * Skipped if no services where selected!
   * @method createSelectedServices
   */
  createSelectedServices: function () {
    var data = this.createSelectedServicesData();
    if (!data.length) return;
    this.addRequestToAjaxQueue({
      name: 'wizard.step8.create_selected_services',
      data: {
        data: JSON.stringify(data)
      }
    });
  },

  /**
   * Format data for <code>createSelectedServices</code> request
   * @returns {Object[]}
   * @method createSelectedServicesData
   */
  createSelectedServicesData: function () {
    return this.get('selectedServices').map(function (_service) {
      return {"ServiceInfo": { "service_name": _service.get('serviceName') }};
    });
  },

  /**
   * Create components for selected services
   * Queued requests
   * One request for each service!
   * @method createComponents
   */
  createComponents: function () {
    var serviceComponents = App.StackServiceComponent.find();
    this.get('selectedServices').forEach(function (_service) {
      var serviceName = _service.get('serviceName');
      var componentsData = serviceComponents.filterProperty('serviceName', serviceName).map(function (_component) {
        return { "ServiceComponentInfo": { "component_name": _component.get('componentName') } };
      });

      // Service must be specified in terms of a query for creating multiple components at the same time.
      // See AMBARI-1018.
      this.addRequestToAjaxQueue({
        name: 'wizard.step8.create_components',
        data: {
          data: JSON.stringify({"components": componentsData}),
          serviceName: serviceName
        }
      });
    }, this);
  },

  /**
   * Error callback for new service component request
   * So, if component doesn't exist we should create it
   * @param {object} request
   * @param {object} ajaxOptions
   * @param {string} error
   * @param {object} opt
   * @param {object} params
   * @method newServiceComponentErrorCallback
   */
  newServiceComponentErrorCallback: function (request, ajaxOptions, error, opt, params) {
    this.addRequestToAjaxQueue({
      name: 'wizard.step8.create_components',
      data: {
        serviceName: params.serviceName,
        data: JSON.stringify({
          "components": [
            {
              "ServiceComponentInfo": {
                "component_name": params.componentName
              }
            }
          ]
        })
      }
    });
  },

  /**
   * Register hosts
   * Queued request
   * @method registerHostsToCluster
   */
  registerHostsToCluster: function () {
    var data = this.createRegisterHostData();
    if (!data.length) return;
    this.addRequestToAjaxQueue({
      name: 'wizard.step8.register_host_to_cluster',
      data: {
        data: JSON.stringify(data)
      }
    });
  },

  /**
   * Format request-data for <code>registerHostsToCluster</code>
   * @returns {Object}
   * @method createRegisterHostData
   */
  createRegisterHostData: function () {
    return this.getRegisteredHosts().filterProperty('isInstalled', false).map(function (host) {
      return {"Hosts": { "host_name": host.hostName}};
    });
  },

  /**
   * Register new master components
   * @uses registerHostsToComponent
   * @method createMasterHostComponents
   */
  createMasterHostComponents: function () {
    // create master components for only selected services.
    var selectedMasterComponents = this.get('content.masterComponentHosts').filter(function (_component) {
      return this.get('selectedServices').mapProperty('serviceName').contains(_component.serviceId)
    }, this);
    selectedMasterComponents.mapProperty('component').uniq().forEach(function (component) {
      var hostNames = selectedMasterComponents.filterProperty('component', component).filterProperty('isInstalled', false).mapProperty('hostName');
      this.registerHostsToComponent(hostNames, component);
    }, this);
  },

  getClientsToMasterMap: function () {
    var clientNames = App.StackServiceComponent.find().filterProperty('isClient').mapProperty('componentName'),
      clientsMap = {},
      dependedComponents = App.StackServiceComponent.find().filterProperty('isMaster');
    clientNames.forEach(function (clientName) {
      clientsMap[clientName] = Em.A([]);
      dependedComponents.forEach(function (component) {
        if (component.get('dependencies').mapProperty('componentName').contains(clientName)) clientsMap[clientName].push(component.get('componentName'));
      });
      if (!clientsMap[clientName].length) delete clientsMap[clientName];
    });
    return clientsMap;
  },
  /**
   * Register slave components and clients
   * @uses registerHostsToComponent
   * @method createSlaveAndClientsHostComponents
   */
  createSlaveAndClientsHostComponents: function () {
    var masterHosts = this.get('content.masterComponentHosts');
    var slaveHosts = this.get('content.slaveComponentHosts');
    var clients = this.get('content.clients').filterProperty('isInstalled', false);

    /**
     * Determines on which hosts client should be installed (based on availability of master components on hosts)
     * @type {Object}
     * Format:
     * <code>
     *  {
     *    CLIENT1: Em.A([MASTER1, MASTER2, ...]),
     *    CLIENT2: Em.A([MASTER3, MASTER1, ...])
     *    ...
     *  }
     * </code>
     */
    var clientsToMasterMap = this.getClientsToMasterMap();

    slaveHosts.forEach(function (_slave) {
      if (_slave.componentName !== 'CLIENT') {
        var hostNames = _slave.hosts.filterProperty('isInstalled', false).mapProperty('hostName');
        this.registerHostsToComponent(hostNames, _slave.componentName);
      }
      else {
        clients.forEach(function (_client) {
          var hostNames = _slave.hosts.mapProperty('hostName');
          // The below logic to install clients to existing/New master hosts should not be applied to Add Host wizard.
          // This is with the presumption that Add Host controller does not add any new Master component to the cluster
          if (this.get('content.controllerName') !== 'addHostController') {
            if (clientsToMasterMap[_client.component_name]) {
              clientsToMasterMap[_client.component_name].forEach(function (componentName) {
                masterHosts.filterProperty('component', componentName).forEach(function (_masterHost) {
                  hostNames.pushObject(_masterHost.hostName);
                });
              });
            }
          }
          hostNames = hostNames.uniq();
          this.registerHostsToComponent(hostNames, _client.component_name);
        }, this);
      }
    }, this);
  },

  /**
   * This function is specific to addServiceController
   * Newly introduced master components requires some existing client components to be hosted along with them
   */
  createAdditionalClientComponents: function () {
    var masterHosts = this.get('content.masterComponentHosts');
    var clientsToMasterMap = this.getClientsToMasterMap();
    var installedClients = [];

    // Get all the installed Client components
    this.get('content.services').filterProperty('isInstalled').forEach(function (_service) {
      var serviceClients = App.StackServiceComponent.find().filterProperty('serviceName', _service.get('serviceName')).filterProperty('isClient');
      serviceClients.forEach(function (client) {
        installedClients.push(client.get('componentName'));
      }, this);
    }, this);

    // Check if there is a dependency for being co-hosted between existing client and selected new master
    installedClients.forEach(function (_clientName) {
      if (clientsToMasterMap[_clientName]) {
        var hostNames = [];
        clientsToMasterMap[_clientName].forEach(function (componentName) {
          masterHosts.filterProperty('component', componentName).filterProperty('isInstalled', false).forEach(function (_masterHost) {
            hostNames.pushObject(_masterHost.hostName);
          }, this);
        }, this);
        hostNames = hostNames.uniq();
        if (hostNames.length > 0) {
          this.get('content.additionalClients').pushObject({hostNames: hostNames, componentName: _clientName});
          // If a dependency for being co-hosted is derived between existing client and selected new master but that
          // dependency is already satisfied in the cluster then disregard the derived dependency
          this.removeClientsFromList(_clientName, hostNames);
          this.registerHostsToComponent(hostNames, _clientName);
        }
      }
    }, this);
  },

  /**
   *
   * @param clientName
   * @param hostList
   */
  removeClientsFromList: function (clientName, hostList) {
    var clientHosts = [];
    var installedHosts = this.get('content.hosts');
    for (var hostName in installedHosts) {
      if (installedHosts[hostName].isInstalled) {
        if (installedHosts[hostName].hostComponents.mapProperty('HostRoles.component_name').contains(clientName)) {
          clientHosts.push(hostName);
        }
      }
    }

    if (clientHosts.length > 0) {
      clientHosts.forEach(function (hostName) {
        if (hostList.contains(hostName)) {
          hostList.splice(hostList.indexOf(hostName), 1);
        }
      }, this);
    }
  },

  /**
   * Register additional components
   * Based on availability of some services
   * @uses registerHostsToComponent
   * @method createAdditionalHostComponents
   */
  createAdditionalHostComponents: function () {
    var masterHosts = this.get('content.masterComponentHosts');

    // add all components with cardinality == ALL of selected services
    var registeredHosts = this.getRegisteredHosts();
    var notInstalledHosts = registeredHosts.filterProperty('isInstalled', false);
    this.get('content.services').filterProperty('isSelected').forEach(function (service) {
      service.get('serviceComponents').filterProperty('isRequiredOnAllHosts').forEach(function (component) {
        if (service.get('isInstalled') && notInstalledHosts.length) {
          this.registerHostsToComponent(notInstalledHosts.mapProperty('hostName'), component.get('componentName'));
        } else if (!service.get('isInstalled') && registeredHosts.length) {
          this.registerHostsToComponent(registeredHosts.mapProperty('hostName'), component.get('componentName'));
        }
      }, this);
    }, this);

    // add MySQL Server if Hive is selected
    var hiveService = this.get('content.services').filterProperty('isSelected', true).filterProperty('isInstalled', false).findProperty('serviceName', 'HIVE');
    if (hiveService) {
      var hiveDb = this.get('content.serviceConfigProperties').findProperty('name', 'hive_database');
      if (hiveDb.value == "New MySQL Database") {
        this.registerHostsToComponent(masterHosts.filterProperty('component', 'HIVE_SERVER').mapProperty('hostName'), 'MYSQL_SERVER');
      } else if (hiveDb.value === "New PostgreSQL Database") {
        this.registerHostsToComponent(masterHosts.filterProperty('component', 'HIVE_SERVER').mapProperty('hostName'), 'POSTGRESQL_SERVER');
      }
    }
  },

  /**
   * Register component to hosts
   * Queued request
   * @param {String[]} hostNames
   * @param {String} componentName
   * @method registerHostsToComponent
   */
  registerHostsToComponent: function (hostNames, componentName) {
    if (!hostNames.length) return;

    var queryStr = '';
    hostNames.forEach(function (hostName) {
      queryStr += 'Hosts/host_name=' + hostName + '|';
    });
    //slice off last symbol '|'
    queryStr = queryStr.slice(0, -1);

    var data = {
      "RequestInfo": {
        "query": queryStr
      },
      "Body": {
        "host_components": [
          {
            "HostRoles": {
              "component_name": componentName
            }
          }
        ]
      }
    };

    this.addRequestToAjaxQueue({
      name: 'wizard.step8.register_host_to_component',
      data: {
        data: JSON.stringify(data)
      }
    });
  },

  /**
   * Compare generated config object with current configs that were filled
   * on "Customize Services" page.
   *
   * @param {Object} properties - generated by createSiteObj|createCoreSiteObj
   * @param {Array} configs - current configs to compare
   * @return {Boolean}
   * @method isConfigsChanged
   **/
  isConfigsChanged: function (properties, configs) {
    var isChanged = false;
    for (var property in properties) {
      var config = configs.findProperty('name', property);
      // if config not found then it's looks like a new config
      if (!config) {
        isChanged = true;
      } else {
        if (!config.hasInitialValue || config.isNotDefaultValue) {
          isChanged = true;
        }
      }
    }
    return isChanged;
  },

  /**
   * Create config objects for cluster and services
   * @method createConfigurations
   */
  createConfigurations: function () {
    var selectedServices = this.get('selectedServices');
    var coreSiteObject = this.createCoreSiteObj();
    var tag = 'version1';
    var clusterSiteObj = this.createSiteObj('cluster-env', tag);

    if (this.get('content.controllerName') == 'installerController') {
      this.get('serviceConfigTags').pushObject(clusterSiteObj);
    }

    if (this.get('content.controllerName') == 'addServiceController') {
      tag = 'version' + (new Date).getTime();
      coreSiteObject.tag = tag;
      var coreSiteConfigs = this.get('configs').filterProperty('filename', 'core-site.xml');
      if (this.isConfigsChanged(coreSiteObject.properties, coreSiteConfigs)) {
        coreSiteObject.service_config_version_note = Em.I18n.t('dashboard.configHistory.table.notes.addService');
        this.get('serviceConfigTags').pushObject(coreSiteObject);
      }
    }

    selectedServices.forEach(function (service) {
      Object.keys(service.get('configTypes')).forEach(function (type) {
        if (!this.get('serviceConfigTags').someProperty('type', type)) {
          var serviceVersionNotes = Em.I18n.t('dashboard.configHistory.table.notes.default').format(service.get('displayName'));
          if (type === 'core-site') {
            coreSiteObject.service_config_version_note = serviceVersionNotes;
            this.get('serviceConfigTags').pushObject(coreSiteObject);
          } else if (type === 'storm-site') {
            var obj = this.createStormSiteObj(tag);
            obj.service_config_version_note = serviceVersionNotes;
            this.get('serviceConfigTags').pushObject(obj);
          }  else {
            var obj = this.createSiteObj(type, tag);
            obj.service_config_version_note = serviceVersionNotes;
            this.get('serviceConfigTags').pushObject(obj);
          }
        }
      }, this);
    }, this);
    this.createNotification();
  },

  /**
   * Send <code>serviceConfigTags</code> to server
   * Queued request
   * One request for each service config tag
   * @param serviceConfigTags
   * @method applyConfigurationsToCluster
   */
  applyConfigurationsToCluster: function (serviceConfigTags) {
    var allServices = this.get('installedServices').concat(this.get('selectedServices'));
    var allConfigData = [];
    allServices.forEach(function (service) {
      var serviceConfigData = [];
      Object.keys(service.get('configTypesRendered')).forEach(function (type) {
        var serviceConfigTag = serviceConfigTags.findProperty('type', type);
        if (serviceConfigTag) {
          serviceConfigData.pushObject(serviceConfigTag);
        }
      }, this);
      if (serviceConfigData.length) {
        allConfigData.pushObject(JSON.stringify({
          Clusters: {
            desired_config: serviceConfigData
          }
        }));
      }
    }, this);
    var clusterConfig = serviceConfigTags.findProperty('type', 'cluster-env');
    if (clusterConfig) {
      allConfigData.pushObject(JSON.stringify({
        Clusters: {
          desired_config: [clusterConfig]
        }
      }));
    }

    this.addRequestToAjaxQueue({
      name: 'common.across.services.configurations',
      data: {
        data: '[' + allConfigData.toString() + ']'
      }
    });
  },

  /**
   * Create and update config groups
   * @method createConfigurationGroups
   */
  createConfigurationGroups: function () {
    var configGroups = this.get('content.configGroups').filterProperty('isDefault', false);
    var clusterName = this.get('clusterName');
    var sendData = [];
    var updateData = [];
    var serviceConfigController = App.router.get('mainServiceInfoConfigsController');
    var timeTag = (new Date).getTime();
    var groupsToDelete = App.router.get(this.get('content.controllerName')).getDBProperty('groupsToDelete');
    if (groupsToDelete && groupsToDelete.length > 0) {
      this.removeInstalledServicesConfigurationGroups(groupsToDelete);
    }
    configGroups.forEach(function (configGroup) {
      var groupConfigs = [];
      var groupData = {
        "cluster_name": clusterName,
        "group_name": configGroup.name,
        "tag": configGroup.service.id,
        "description": configGroup.description,
        "hosts": [],
        "desired_configs": []
      };
      configGroup.hosts.forEach(function (hostName) {
        groupData.hosts.push({"host_name": hostName});
      });
      // get properties that was created for non-default config group
      configGroup.properties = configGroup.properties.concat(this.get('customNonDefaultGroupConfigs').filterProperty('group', configGroup.name));
      //wrap properties into Em.Object to make them compatible with buildGroupDesiredConfigs method
      configGroup.properties.forEach(function (property) {
        groupConfigs.push(Em.Object.create(property));
      });
      groupData.desired_configs = serviceConfigController.buildGroupDesiredConfigs.call(serviceConfigController, groupConfigs, timeTag);
      // check for group from installed service
      if (configGroup.isForInstalledService === true) {
        // if group is a new one, create it
        if (!configGroup.id) {
          sendData.push({"ConfigGroup": groupData});
        } else if (configGroup.isForUpdate){
          // update an existing group
          groupData.id = configGroup.id;
          updateData.push({"ConfigGroup": groupData});
        }
      } else {
        sendData.push({"ConfigGroup": groupData});
      }
      //each group should have unique tag to prevent overriding configs from common sites
      timeTag++;
    }, this);
    if (sendData.length > 0) {
      this.applyConfigurationGroups(sendData);
    }
    if (updateData.length > 0) {
      this.applyInstalledServicesConfigurationGroup(updateData);
    }
  },

  /**
   * Create new config groups request
   * Queued request
   * @param {Object[]} sendData
   * @method applyConfigurationGroups
   */
  applyConfigurationGroups: function (sendData) {
    this.addRequestToAjaxQueue({
      name: 'wizard.step8.apply_configuration_groups',
      data: {
        data: JSON.stringify(sendData)
      }
    });
  },

  /**
   * Update existed config groups
   * Separated request for each group
   * @param {Object[]} updateData
   * @method applyInstalledServicesConfigurationGroup
   */
  applyInstalledServicesConfigurationGroup: function (updateData) {
    updateData.forEach(function (item) {
      App.router.get('mainServiceInfoConfigsController').putConfigGroupChanges(item);
    });
  },

  /**
   * Delete selected config groups
   * @param {Object[]} groupsToDelete
   * @method removeInstalledServicesConfigurationGroups
   */
  removeInstalledServicesConfigurationGroups: function (groupsToDelete) {
    groupsToDelete.forEach(function (item) {
      App.config.deleteConfigGroup(Em.Object.create(item));
    });
  },

  /**
   * proxyuser configs which depend on service
   * @type {{serviceName: string, user: string}[]}
   */
  optionalCoreSiteConfigs: [
    {
      serviceName: 'OOZIE',
      user: 'oozie_user'
    },
    {
      serviceName: 'HIVE',
      user: 'hive_user'
    },
    {
      serviceName: 'HIVE',
      user: 'hcat_user'
    },
    {
      serviceName: 'FALCON',
      user: 'falcon_user'
    }
  ],

  /**
   * push proxyuser properties to core-site if they required by dependencies
   * @param coreSiteObj
   * @param installedAndSelectedServices
   * @return {Object}
   */
  resolveProxyuserDependecies: function (coreSiteObj, installedAndSelectedServices) {
    var coreSiteProperties = {};
    var optionalCoreSiteConfigs = this.get('optionalCoreSiteConfigs');
    var proxyuserGroup = this.get('configs').findProperty('name', 'proxyuser_group');

    coreSiteObj.forEach(function (_coreSiteObj) {
      //proxyuser_group property should be added only if proxyuser properties are used
      if (proxyuserGroup && _coreSiteObj.name === proxyuserGroup.name) return;

      // exclude some configs if service wasn't selected
      var addProperty = optionalCoreSiteConfigs.every(function (config) {
        var userValue = this.get('configs').someProperty('name', config.user) ? this.get('configs').findProperty('name', config.user).value : null
        return (installedAndSelectedServices.someProperty('serviceName', config.serviceName) ||
          (_coreSiteObj.name != 'hadoop.proxyuser.' + userValue + '.hosts' && _coreSiteObj.name != 'hadoop.proxyuser.' + userValue + '.groups'))
      }, this);
      if (addProperty) {
        coreSiteProperties[_coreSiteObj.name] = _coreSiteObj.value;
      }
    }, this);

    if (!App.isEmptyObject(coreSiteProperties) && proxyuserGroup) {
      coreSiteProperties[proxyuserGroup.name] = proxyuserGroup.value;
    }
    return coreSiteProperties;
  },

  /**
   * Create Core Site object
   * @returns {{type: string, tag: string, properties: {}}}
   * @method createCoreSiteObj
   */
  createCoreSiteObj: function () {
    var installedAndSelectedServices = Em.A([]);
    installedAndSelectedServices.pushObjects(this.get('installedServices'));
    installedAndSelectedServices.pushObjects(this.get('selectedServices'));
    var coreSiteObj = this.get('configs').filterProperty('filename', 'core-site.xml'),
      coreSiteProperties = this.resolveProxyuserDependecies(coreSiteObj, installedAndSelectedServices),
      isGLUSTERFSSelected = installedAndSelectedServices.someProperty('serviceName', 'GLUSTERFS');

    coreSiteObj.forEach(function (_coreSiteObj) {
      if (isGLUSTERFSSelected && _coreSiteObj.name == "fs.default.name") {
        coreSiteProperties[_coreSiteObj.name] =
          this.get('configs').someProperty('name', 'fs_glusterfs_default_name') ?
            this.get('configs').findProperty('name', 'fs_glusterfs_default_name').value : null;
      }
      if (isGLUSTERFSSelected && _coreSiteObj.name == "fs.defaultFS") {
        coreSiteProperties[_coreSiteObj.name] =
          this.get('configs').someProperty('name', 'glusterfs_defaultFS_name') ?
            this.get('configs').findProperty('name', 'glusterfs_defaultFS_name').value : null;
      }
    }, this);
    var attributes = App.router.get('mainServiceInfoConfigsController').getConfigAttributes(coreSiteObj);
    var configObj = {"type": "core-site", "tag": "version1", "properties": coreSiteProperties};
    if (attributes) {
      configObj['properties_attributes'] = attributes;
    }
    return  configObj;
  },

  /**
   * Create siteObj for custom service with it own configs
   * @param {string} site
   * @param tag
   * @returns {{type: string, tag: string, properties: {}}}
   * @method createSiteObj
   */
  createSiteObj: function (site,  tag) {
    var properties = {};
    var configs = this.get('configs').filterProperty('filename', site + '.xml');
    var attributes = App.router.get('mainServiceInfoConfigsController').getConfigAttributes(configs);
    configs.forEach(function (_configProperty) {
        var heapsizeExceptions = ['hadoop_heapsize', 'yarn_heapsize', 'nodemanager_heapsize', 'resourcemanager_heapsize', 'apptimelineserver_heapsize', 'jobhistory_heapsize'];
        // do not pass any globals whose name ends with _host or _hosts
        if (_configProperty.isRequiredByAgent !== false) {
          // append "m" to JVM memory options except for heapsizeExtensions
          if (/_heapsize|_newsize|_maxnewsize|_permsize|_maxpermsize$/.test(_configProperty.name) && !heapsizeExceptions.contains(_configProperty.name) && !(_configProperty.value).endsWith("m")) {
            properties[_configProperty.name] = _configProperty.value + "m";
          } else {
            properties[_configProperty.name] = _configProperty.value;
          }
        }
    }, this);
    var configObj = {"type": site, "tag": tag, "properties": properties };
    if (attributes) {
      configObj['properties_attributes'] = attributes;
    }
    return configObj;
  },

  /**
   * Create site obj for Storm
   * Some config-properties should be modified in custom way
   * @param tag
   * @returns {{type: string, tag: string, properties: {}}}
   * @method createStormSiteObj
   */
  createStormSiteObj: function (tag) {
    var configs = this.get('configs').filterProperty('filename', 'storm-site.xml');
    var stormProperties = {};
    configs.forEach(function (_configProperty) {
      if (_configProperty.name == "storm.zookeeper.servers") {
        stormProperties[_configProperty.name] = JSON.stringify(_configProperty.value).replace(/"/g, "'");
      } else {
        stormProperties[_configProperty.name] = _configProperty.value;
      }
    }, this);
    return {type: 'storm-site', tag: tag, properties: stormProperties};
  },

  /**
   * Create one Alert Notification (if user select this on step7)
   * Only for Install Wizard and stack
   * @method createNotification
   */
  createNotification: function () {
    if (this.get('content.controllerName') !== 'installerController') return;
    var miscConfigs = this.get('configs').filterProperty('serviceName', 'MISC'),
      createNotification = miscConfigs.findProperty('name', 'create_notification').value;
    if (createNotification !== 'yes') return;
      var predefinedNotificationConfigNames = require('data/HDP2/site_properties').configProperties.filterProperty('filename', 'alert_notification').mapProperty('name'),
      configsForNotification = this.get('configs').filterProperty('filename', 'alert_notification');
    var properties = {},
      names = [
        'ambari.dispatch.recipients',
        'mail.smtp.host',
        'mail.smtp.port',
        'mail.smtp.from',
        'mail.smtp.starttls.enable',
        'mail.smtp.startssl.enable'
      ];
    if (miscConfigs.findProperty('name', 'smtp_use_auth').value == 'true') { // yes, it's not converted to boolean
      names.pushObjects(['ambari.dispatch.credential.username', 'ambari.dispatch.credential.password']);
    }

    names.forEach(function (name) {
      properties[name] = miscConfigs.findProperty('name', name).value;
    });

    properties['ambari.dispatch.recipients'] = properties['ambari.dispatch.recipients'].replace(/\s/g, '').split(',');

    configsForNotification.forEach(function (config) {
      if (predefinedNotificationConfigNames.contains(config.name)) return;
      properties[config.name] = config.value;
    });

    var apiObject = {
      AlertTarget: {
        name: 'Initial Notification',
        description: 'Notification created during cluster installing',
        global: true,
        notification_type: 'EMAIL',
        alert_states: ['OK', 'WARNING', 'CRITICAL', 'UNKNOWN'],
        properties: properties
      }
    };
    this.addRequestToAjaxQueue({
      name: 'alerts.create_alert_notification',
      data: {
        urlParams: 'overwrite_existing=true',
        data: apiObject
      }
    });
  }
});

});

require.register("controllers/wizard/step9_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');
var stringUtils = require('utils/string_utils');

App.WizardStep9Controller = Em.Controller.extend({

  name: 'wizardStep9Controller',

  /**
   *  Array of host Objects that are successfully registered on "Confirm Host Options" page
   *  <code>
   *    {
   *      name: {String} host name.
   *      status: {String} Current status of the host. This field is used in the step 9 view to set the css class of the
   *             host progress bar and set the appropriate message. Possible values: info, warning, failed, heartbeat_lost and success
   *      logTasks: {Array} Tasks that are scheduled on the host for the current request.
   *      message: {String} Message to be shown in front of the host name.
   *      progress: {Int} Progress of the tasks on the host. Amount of tasks completed for all the tasks scheduled on the host.
   *      isNoTasksForInstall: {Boolean} Gets set when no task is scheduled for the host on install phase.
   *    }
   *  </code>
   *  @type {Array.<{name: string, status: string, logTasks: object[], message: string, progress: number, isNoTasksForInstall: bool}>}
   */
  hosts: [],

  /**
   * Overall progress of <Install,Start and Test> page shown as progress bar on the top of the page
   * @type {string}
   */
  progress: '0',

  /*
   * Json file for the mock data to be used in mock mode
   * @type {string}
   */
  mockDataPrefix: '/data/wizard/deploy/5_hosts',

  /*
   * Current Request data polled from the API: api/v1/clusters/{clusterName}/requests/{RequestId}?fields=tasks/Tasks/command,
   * tasks/Tasks/exit_code,tasks/Tasks/start_time,tasks/Tasks/end_time,tasks/Tasks/host_name,tasks/Tasks/id,tasks/Tasks/role,
   * tasks/Tasks/status&minimal_response=true
   * @type {Object[]}
   */
  polledData: [],

  /*
   * This flag is only used in UI mock mode as a counter for number of polls.
   * @type {number}
   */
  numPolls: 1,

  /**
   * Interval in milliseconds between API calls While polling the request status for Install and, Start and Test tasks
   * @type {number}
   */
  POLL_INTERVAL: 4000,

  /**
   * Array of objects
   * <code>
   *   {
   *     hostName: {String} Name of host that has stopped heartbeating to ambari-server
   *     componentNames: {Sting} Name of all components that are on the host
   *   }
   * </code>
   * @type {Object[]}
   */
  hostsWithHeartbeatLost: [],

  /**
   * Flag is set in the start all services error callback function
   * @type {bool}
   */
  startCallFailed: false,

  /*
   * Status of the page. Possible values: <info, warning, failed and success>.
   * This property is used in the step-9 view for displaying the appropriate color of the overall progress bar and
   * the appropriate result message at the bottom of the page
   * @type {string}
   */
  status: 'info',

  /**
   * This computed property is used to determine if the Next button and back button on the page should be disabled. The property is
   * used in the template to grey out Next and Back buttons. Although clicking on the greyed out button do trigger the event and
   * calls submit and back function of the controller.
   * @type {bool}
   */
  isSubmitDisabled: function () {
    var validStates = ['STARTED', 'START FAILED'];
    var controllerName = this.get('content.controllerName');
    if (controllerName == 'addHostController' || controllerName == 'addServiceController') {
      validStates.push('INSTALL FAILED');
    }
    return !validStates.contains(this.get('content.cluster.status'));
  }.property('content.cluster.status'),

  /**
   * Observer function: Enables previous steps link if install task failed in installer wizard.
   * @method togglePreviousSteps
   */
  togglePreviousSteps: function () {
    if (App.get('testMode')) {
      return;
    }
    var installerController = App.router.get('installerController');
    if ('INSTALL FAILED' === this.get('content.cluster.status') && this.get('content.controllerName') == 'installerController') {
      installerController.setStepsEnable();
    } else {
      installerController.setLowerStepsDisable(9);
    }
  }.observes('content.cluster.status', 'content.controllerName'),

  /*
   * Computed property to determine if the Retry button should be made visible on the page.
   * @type {bool}
   */
  showRetry: function () {
    return this.get('content.cluster.status') == 'INSTALL FAILED';
  }.property('content.cluster.status'),

  /**
   * Observer function: Calls {hostStatusUpdates} function once with change in a host status from any registered hosts.
   * @method hostStatusObserver
   */
  hostStatusObserver: function () {
    Em.run.once(this, 'updateStatus');
  }.observes('hosts.@each.status'),

  /**
   * A flag that gets set with installation failure.
   * @type {bool}
   */
  installFailed: false,

  /**
   * Incremental flag that triggers an event in step 9 view to change the tasks related data and icons of hosts.
   * @type {number}
   */
  logTasksChangesCounter: 0,

  /**
   * <code>taskId</code> of current open task
   * @type {number}
   */
  currentOpenTaskId: 0,

  /**
   * <code>requestId</code> of current open task
   * @type {number}
   */
  currentOpenTaskRequestId: 0,

  /**
   * True if stage transition is completed.
   * On true, polling will be stopped.
   * @type {bool}s
   */
  parseHostInfo: false,

  isPolling: true,

  changeParseHostInfo: function (value) {
    this.set('parseHostInfo', value);
    this.parseHostInfoPolling();
  },

  parseHostInfoPolling: function () {
    var self = this;
    var result = this.get('parseHostInfo');
    if (!this.get('isPolling')) {
      if (this.get('content.cluster.status') === 'INSTALL FAILED') {
        this.isAllComponentsInstalled();
      }
      return;
    }
    if (result !== true) {
      window.setTimeout(function () {
        if (self.get('currentOpenTaskId')) {
          self.loadCurrentTaskLog();
        }
        self.doPolling();
      }, this.get('POLL_INTERVAL'));
    }
  },

  /**
   * Observer function: Updates {status} field of the controller.
   * @method updateStatus
   */
  updateStatus: function () {
    var status = 'info';
    if (this.get('hosts').someProperty('status', 'failed')
      || this.get('hosts').someProperty('status', 'heartbeat_lost')
      || this.get('startCallFailed')) {
      status = 'failed';
    } else if (this.get('hosts').someProperty('status', 'warning')) {
      if (this.isStepFailed()) {
        status = 'failed';
      } else {
        status = 'warning';
      }
    } else if (this.get('progress') == '100' && this.get('content.cluster.status') !== 'INSTALL FAILED') {
      status = 'success';
    }
    this.set('status', status);
  }.observes('progress'),

  /**
   * This function is called when a click event happens on Next button of "Install, Start and Test" page
   * @method submit
   */
  submit: function () {
    App.router.send('next');
  },

  /**
   * This function is called when a click event happens on back button of "Install, Start and Test" page
   * @method back
   */
  back: function () {
    if (!this.get('isSubmitDisabled')) {
      App.router.send('back');
    }
  },

  /**
   * navigateStep is called by App.WizardStep9View's didInsertElement and "retry" from router.
   * content.cluster.status can be:
   * PENDING: set upon successful transition from step 1 to step 2
   * INSTALLED: set upon successful completion of install phase as well as successful invocation of start services API
   * STARTED: set up on successful completion of start phase
   * INSTALL FAILED: set up upon encountering a failure in install phase
   * START FAILED: set upon unsuccessful invocation of start services API and also upon encountering a failure
   * during start phase

   * content.cluster.isCompleted
   * set to false upon successful transition from step 1 to step 2
   * set to true upon successful start of services in this step
   * note: looks like this is the same thing as checking content.cluster.status == 'STARTED'
   * @method navigateStep
   */
  navigateStep: function () {
    if (App.get('testMode')) {
      // this is for repeatedly testing out installs in test mode
      this.set('content.cluster.status', 'PENDING');
      this.set('content.cluster.isCompleted', false);
      this.set('content.cluster.requestId', 1);
    }
    var needPolling = false;
    var clusterStatus = this.get('content.cluster.status');
    console.log('navigateStep: clusterStatus = ' + clusterStatus);
    if (this.get('content.cluster.isCompleted') === false) {
      if (clusterStatus !== 'INSTALL FAILED' && clusterStatus !== 'START FAILED') {
        needPolling = true;
      }
    } else {
      // handle STARTED
      // the cluster has successfully installed and started
      this.set('progress', '100');
    }
    this.loadStep();
    this.loadLogData(needPolling);
  },

  /**
   * This is called on initial page load, refreshes and retry event.
   * clears all in memory stale data for retry event.
   * @method clearStep
   */
  clearStep: function () {
    this.get('hosts').clear();
    this.set('hostsWithHeartbeatLost', []);
    this.set('startCallFailed', false);
    this.set('status', 'info');
    this.set('progress', '0');
    this.set('numPolls', 1);
  },

  /**
   * This is called on initial page load, refreshes and retry event.
   * @method loadStep
   */
  loadStep: function () {
    console.log("TRACE: Loading step9: Install, Start and Test");
    this.clearStep();
    this.loadHosts();
  },

  /**
   * Reset status and message of all hosts when retry install
   * @method resetHostsForRetry
   */
  resetHostsForRetry: function () {
    var hosts = this.get('content.hosts');
    for (var name in hosts) {
      if (hosts.hasOwnProperty(name)) {
        hosts[name].status = "pending";
        hosts[name].message = 'Waiting';
        hosts[name].isNoTasksForInstall = false;
      }
    }
    this.set('content.hosts', hosts);
  },

  /**
   * Sets the <code>hosts</code> array for the controller
   * @method loadHosts
   */
  loadHosts: function () {
    var hosts = this.get('content.hosts');

    //flag identify whether get all hosts or only NewHosts(newly added) hosts
    var getOnlyNewHosts = (this.get('content.controllerName') !== 'addServiceController');

    for (var index in hosts) {
      if (hosts[index].bootStatus === 'REGISTERED' && (!getOnlyNewHosts || !hosts[index].isInstalled)) {
        var hostInfo = App.HostInfo.create({
          name: hosts[index].name,
          status: (hosts[index].status) ? hosts[index].status : 'info',
          logTasks: [],
          message: (hosts[index].message) ? hosts[index].message : 'Waiting',
          progress: 0,
          isNoTasksForInstall: false
        });
        this.get('hosts').pushObject(hostInfo);
      }
    }
  },

  /**
   * Set new polled data
   * @param polledData sets the <code>polledData</code> object of the controller
   * @method replacePolledData
   */
  replacePolledData: function (polledData) {
    this.get('polledData').clear();
    this.set('polledData', polledData);
  },

  /**
   * Get the appropriate message for the host as per the running task
   * @param {object} task
   * @returns {String}
   * @method displayMessage
   */
  displayMessage: function (task) {
    var role = App.format.role(task.role);
    /* istanbul ignore next */
    switch (task.command) {
      case 'INSTALL':
        switch (task.status) {
          case 'PENDING':
            return Em.I18n.t('installer.step9.serviceStatus.install.pending') + role;
          case 'QUEUED' :
            return Em.I18n.t('installer.step9.serviceStatus.install.queued') + role;
          case 'IN_PROGRESS':
            return Em.I18n.t('installer.step9.serviceStatus.install.inProgress') + role;
          case 'COMPLETED' :
            return Em.I18n.t('installer.step9.serviceStatus.install.completed') + role;
          case 'FAILED':
            return Em.I18n.t('installer.step9.serviceStatus.install.failed') + role;
        }
        break;
      case 'UNINSTALL':
        switch (task.status) {
          case 'PENDING':
            return Em.I18n.t('installer.step9.serviceStatus.uninstall.pending') + role;
          case 'QUEUED' :
            return Em.I18n.t('installer.step9.serviceStatus.uninstall.queued') + role;
          case 'IN_PROGRESS':
            return Em.I18n.t('installer.step9.serviceStatus.uninstall.inProgress') + role;
          case 'COMPLETED' :
            return Em.I18n.t('installer.step9.serviceStatus.uninstall.completed') + role;
          case 'FAILED':
            return Em.I18n.t('installer.step9.serviceStatus.uninstall.failed') + role;
        }
        break;
      case 'START' :
        switch (task.status) {
          case 'PENDING':
            return Em.I18n.t('installer.step9.serviceStatus.start.pending') + role;
          case 'QUEUED' :
            return Em.I18n.t('installer.step9.serviceStatus.start.queued') + role;
          case 'IN_PROGRESS':
            return Em.I18n.t('installer.step9.serviceStatus.start.inProgress') + role;
          case 'COMPLETED' :
            return role + Em.I18n.t('installer.step9.serviceStatus.start.completed');
          case 'FAILED':
            return role + Em.I18n.t('installer.step9.serviceStatus.start.failed');
        }
        break;
      case 'STOP' :
        switch (task.status) {
          case 'PENDING':
            return Em.I18n.t('installer.step9.serviceStatus.stop.pending') + role;
          case 'QUEUED' :
            return Em.I18n.t('installer.step9.serviceStatus.stop.queued') + role;
          case 'IN_PROGRESS':
            return Em.I18n.t('installer.step9.serviceStatus.stop.inProgress') + role;
          case 'COMPLETED' :
            return role + Em.I18n.t('installer.step9.serviceStatus.stop.completed');
          case 'FAILED':
            return role + Em.I18n.t('installer.step9.serviceStatus.stop.failed');
        }
        break;
      case 'EXECUTE' :
      case 'SERVICE_CHECK' :
        switch (task.status) {
          case 'PENDING':
            return Em.I18n.t('installer.step9.serviceStatus.execute.pending') + role;
          case 'QUEUED' :
            return Em.I18n.t('installer.step9.serviceStatus.execute.queued') + role;
          case 'IN_PROGRESS':
            return Em.I18n.t('installer.step9.serviceStatus.execute.inProgress') + role;
          case 'COMPLETED' :
            return role + Em.I18n.t('installer.step9.serviceStatus.execute.completed');
          case 'FAILED':
            return role + Em.I18n.t('installer.step9.serviceStatus.execute.failed');
        }
        break;
      case 'ABORT' :
        switch (task.status) {
          case 'PENDING':
            return Em.I18n.t('installer.step9.serviceStatus.abort.pending') + role;
          case 'QUEUED' :
            return Em.I18n.t('installer.step9.serviceStatus.abort.queued') + role;
          case 'IN_PROGRESS':
            return Em.I18n.t('installer.step9.serviceStatus.abort.inProgress') + role;
          case 'COMPLETED' :
            return role + Em.I18n.t('installer.step9.serviceStatus.abort.completed');
          case 'FAILED':
            return role + Em.I18n.t('installer.step9.serviceStatus.abort.failed');
        }
    }
    return '';
  },

  /**
   * Run start/check services after installation phase.
   * Does Ajax call to start all services
   * @return {$.ajax|null}
   * @method launchStartServices
   */
  launchStartServices: function (callback) {
    var data = {};
    var name = '';
    callback = callback || Em.K;
    switch(this.get('content.controllerName')) {
      case 'addHostController':
        name = 'common.host_components.update';
        var hostnames = [];
        for (var hostname in this.get('wizardController').getDBProperty('hosts')) {
          if(this.get('hosts').findProperty('name', hostname)){
            hostnames.push(hostname);
          }
        }
        data = {
          "context": Em.I18n.t("requestInfo.startHostComponents"),
          "query": "HostRoles/component_name.in(" + App.get('components.slaves').join(',') + ")&HostRoles/state=INSTALLED&HostRoles/host_name.in(" + hostnames.join(',') + ")",
          "HostRoles": { "state": "STARTED"}
        };
        break;
      case 'addServiceController':
        var servicesList = this.get('content.services').filterProperty('isSelected').filterProperty('isInstalled', false).mapProperty('serviceName');
        if (servicesList.contains('OOZIE')) {
          servicesList = servicesList.concat(['HDFS', 'YARN', 'MAPREDUCE2']);
        }
        name = 'common.services.update';
        data = {
          "context": Em.I18n.t("requestInfo.startAddedServices"),
          "ServiceInfo": { "state": "STARTED" },
          "urlParams": "ServiceInfo/state=INSTALLED&ServiceInfo/service_name.in(" + servicesList.join(",") + ")&params/run_smoke_test=true&params/reconfigure_client=false"
        };
        break;
      default:
        name = 'common.services.update';
        data = {
          "context": Em.I18n.t("requestInfo.startServices"),
          "ServiceInfo": { "state": "STARTED" },
          "urlParams": "ServiceInfo/state=INSTALLED&params/run_smoke_test=true&params/reconfigure_client=false"
        };
    }

    if (App.get('testMode')) {
      this.set('numPolls', 6);
    }

    return App.ajax.send({
      name: name,
      sender: this,
      data: data,
      success: 'launchStartServicesSuccessCallback',
      error: 'launchStartServicesErrorCallback'
    }).then(callback, callback);
  },

  /**
   * Success callback function for start services task.
   * @param {object} jsonData Contains Request id to poll.
   * @method launchStartServicesSuccessCallback
   */
  launchStartServicesSuccessCallback: function (jsonData) {
    var clusterStatus = {};
    if (jsonData) {
      console.log("TRACE: Step9 -> In success function for the startService call");
      console.log("TRACE: Step9 -> value of the received data is: " + jsonData);
      var requestId = jsonData.Requests.id;
      console.log('requestId is: ' + requestId);
      clusterStatus = {
        status: 'INSTALLED',
        requestId: requestId,
        isStartError: false,
        isCompleted: false
      };
      this.hostHasClientsOnly(false);
      this.saveClusterStatus(clusterStatus);
    }
    else {
      console.log('ERROR: Error occurred in parsing JSON data');
      this.hostHasClientsOnly(true);
      clusterStatus = {
        status: 'STARTED',
        isStartError: false,
        isCompleted: true
      };
      this.saveClusterStatus(clusterStatus);
      this.set('status', 'success');
      this.set('progress', '100');
    }
    // We need to do recovery if there is a browser crash
    App.clusterStatus.setClusterStatus({
      clusterState: 'SERVICE_STARTING_3',
      wizardControllerName: this.get('content.controllerName'),
      localdb: App.db.data
    });

    if (jsonData) {
      this.startPolling();
    }
  },

  /**
   * This function will be called for Add host wizard only.
   * @param {bool} jsonError Boolean is true when API to start services returns 200 ok and no json data
   * @method hostHasClientsOnly
   */
  hostHasClientsOnly: function (jsonError) {
    this.get('hosts').forEach(function (host) {
      var OnlyClients = true;
      var tasks = host.get('logTasks');
      tasks.forEach(function (task) {
        var component = App.StackServiceComponent.find().findProperty('componentName', task.Tasks.role);
        if (!(component && component.get('isClient'))) {
          OnlyClients = false;
        }
      });
      if (OnlyClients || jsonError) {
        host.set('status', 'success');
        host.set('progress', '100');
      }
    });
  },

  /**
   * Error callback function for start services task.
   * @param {object} jqXHR
   * @param {object} ajaxOptions
   * @param {string} error
   * @param {object} opt
   * @method launchStartServicesErrorCallback
   */
  launchStartServicesErrorCallback: function (jqXHR, ajaxOptions, error, opt) {
    console.log("ERROR");
    this.set('startCallFailed', true);
    var clusterStatus = {
      status: 'INSTALL FAILED',
      isStartError: false,
      isCompleted: false
    };
    this.saveClusterStatus(clusterStatus);
    this.get('hosts').forEach(function (host) {
      host.set('progress', '100');
    });
    this.set('progress', '100');

    App.ajax.defaultErrorHandler(jqXHR, opt.url, opt.method, jqXHR.status);
  },

  /**
   * Marks a host's status as "success" if all tasks are in COMPLETED state
   * @method onSuccessPerHost
   */
  onSuccessPerHost: function (actions, contentHost) {
    if (actions.everyProperty('Tasks.status', 'COMPLETED') && this.get('content.cluster.status') === 'INSTALLED') {
      contentHost.set('status', 'success');
    }
  },

  /**
   * marks a host's status as "warning" if at least one of the tasks is FAILED, ABORTED, or TIMEDOUT and marks host's status as "failed" if at least one master component install task is FAILED.
   * note that if the master failed to install because of ABORTED or TIMEDOUT, we don't mark it as failed, because this would mark all hosts as "failed" and makes it difficult for the user
   * to find which host FAILED occurred on, if any
   * @param actions {Array} of tasks retrieved from polled data
   * @param contentHost {Object} A host object
   * @method onErrorPerHost
   */
  onErrorPerHost: function (actions, contentHost) {
    if (!actions) return;
    if (actions.someProperty('Tasks.status', 'FAILED') || actions.someProperty('Tasks.status', 'ABORTED') || actions.someProperty('Tasks.status', 'TIMEDOUT')) {
      contentHost.set('status', 'warning');
    }
    if ((this.get('content.cluster.status') === 'PENDING' && actions.someProperty('Tasks.status', 'FAILED')) || (this.isMasterFailed(actions))) {
      contentHost.get('status') !== 'heartbeat_lost' ? contentHost.set('status', 'failed') : '';
    }
  },

  /**
   * Check if some master component is failed
   * @param {object} polledData data polled from API.
   * @returns {bool}  true if there is at least one FAILED task of master component install
   * @method isMasterFailed
   */
  isMasterFailed: function (polledData) {
    var result = false;
    polledData.filterProperty('Tasks.command', 'INSTALL').filterProperty('Tasks.status', 'FAILED').mapProperty('Tasks.role').forEach(
      function (role) {
        if (!App.get('components.slaves').contains(role)) {
          result = true;
        }
      }
    );
    return result;
  },

  /**
   * Mark a host status as in_progress if the any task on the host if either in IN_PROGRESS, QUEUED or PENDONG state.
   * @param actions {Array} of tasks retrieved from polled data
   * @param contentHost {Object} A host object
   * @method onInProgressPerHost
   */
  onInProgressPerHost: function (actions, contentHost) {
    var runningAction = actions.findProperty('Tasks.status', 'IN_PROGRESS');
    if (runningAction === undefined || runningAction === null) {
      runningAction = actions.findProperty('Tasks.status', 'QUEUED');
    }
    if (runningAction === undefined || runningAction === null) {
      runningAction = actions.findProperty('Tasks.status', 'PENDING');
    }
    if (runningAction !== null && runningAction !== undefined) {
      contentHost.set('status', 'in_progress');
      contentHost.set('message', this.displayMessage(runningAction.Tasks));
    }
  },

  /**
   * Calculate progress of tasks per host
   * @param {object} actions
   * @param {object} contentHost
   * @return {Number}
   * @method progressPerHost
   */
  progressPerHost: function (actions, contentHost) {
    var progress = 0;
    var actionsPerHost = actions.length;
    var completedActions = 0;
    var queuedActions = 0;
    var inProgressActions = 0;
    actions.forEach(function (action) {
      completedActions += +(['COMPLETED', 'FAILED', 'ABORTED', 'TIMEDOUT'].contains(action.Tasks.status));
      queuedActions += +(action.Tasks.status === 'QUEUED');
      inProgressActions += +(action.Tasks.status === 'IN_PROGRESS');
    }, this);
    /** for the install phase (PENDING), % completed per host goes up to 33%; floor(100 / 3)
     * for the start phase (INSTALLED), % completed starts from 34%
     * when task in queued state means it's completed on 9%
     * in progress - 35%
     * completed - 100%
     */
    switch (this.get('content.cluster.status')) {
      case 'PENDING':
        progress = actionsPerHost ? (Math.ceil(((queuedActions * 0.09) + (inProgressActions * 0.35) + completedActions ) / actionsPerHost * 33)) : 33;
        break;
      case 'INSTALLED':
        progress = actionsPerHost ? (34 + Math.ceil(((queuedActions * 0.09) + (inProgressActions * 0.35) + completedActions ) / actionsPerHost * 66)) : 100;
        break;
      default:
        progress = 100;
        break;
    }
    contentHost.set('progress', progress.toString());
    return progress;
  },

  /**
   * Check if step completed successfully
   * @param {object} polledData data retrieved from API
   * @returns {bool}
   * @method isSuccess
   */
  isSuccess: function (polledData) {
    return polledData.everyProperty('Tasks.status', 'COMPLETED');
  },

  /**
   * Check if step is failed
   * Return true if:
   *  1. any of the master/client components failed to install
   *  OR
   *  2. at least 50% of the slave host components for the particular service component fails to install
   *  @method isStepFailed
   */
  isStepFailed: function () {
    var failed = false;
    var polledData = this.get('polledData');
    polledData.filterProperty('Tasks.command', 'INSTALL').mapProperty('Tasks.role').uniq().forEach(function (role) {
      if (failed) {
        return;
      }
      var actionsPerRole = polledData.filterProperty('Tasks.role', role);
      if (App.get('components.slaves').contains(role)) {
        // check slave components for success factor.
        // partial failure for slave components are allowed.
        var actionsFailed = actionsPerRole.filterProperty('Tasks.status', 'FAILED');
        var actionsAborted = actionsPerRole.filterProperty('Tasks.status', 'ABORTED');
        var actionsTimedOut = actionsPerRole.filterProperty('Tasks.status', 'TIMEDOUT');
        if ((((actionsFailed.length + actionsAborted.length + actionsTimedOut.length) / actionsPerRole.length) * 100) > 50) {
          failed = true;
        }
      } else if (actionsPerRole.someProperty('Tasks.status', 'FAILED') || actionsPerRole.someProperty('Tasks.status', 'ABORTED') ||
        actionsPerRole.someProperty('Tasks.status', 'TIMEDOUT')) {
        // check non-salve components (i.e., masters and clients).  all of these must be successfully installed.
        failed = true;
      }
    }, this);
    return failed;
  },

  /**
   * polling from ui stops only when no action has 'PENDING', 'QUEUED' or 'IN_PROGRESS' status
   * Makes a state transition
   * PENDING -> INSTALLED
   * PENDING -> INSTALL FAILED
   * INSTALLED -> STARTED
   * INSTALLED -> START_FAILED
   * @param polledData json data retrieved from API
   * @method setFinishState
   */
  setFinishState: function (polledData) {
    if (this.get('content.cluster.status') === 'INSTALLED') {
      this.changeParseHostInfo(this.isServicesStarted(polledData));
      return;
    } else if (this.get('content.cluster.status') === 'PENDING') {
      this.setIsServicesInstalled(polledData);
      return;
    } else if (this.get('content.cluster.status') === 'INSTALL FAILED' || this.get('content.cluster.status') === 'START FAILED'
      || this.get('content.cluster.status') === 'STARTED') {
      this.set('progress', '100');
      this.changeParseHostInfo(true);
      return;
    }
    this.changeParseHostInfo(true);
  },

  /**
   * @param polledData Josn data retrieved from API
   * @returns {bool} Has "Start All Services" request completed successfully
   * @method isServicesStarted
   */
  isServicesStarted: function (polledData) {
    var clusterStatus = {};
    if (!polledData.someProperty('Tasks.status', 'PENDING') && !polledData.someProperty('Tasks.status', 'QUEUED') && !polledData.someProperty('Tasks.status', 'IN_PROGRESS')) {
      this.set('progress', '100');
      clusterStatus = {
        status: 'INSTALLED',
        requestId: this.get('content.cluster.requestId'),
        isCompleted: true
      };
      if (this.isSuccess(polledData)) {
        clusterStatus.status = 'STARTED';
        var serviceStartTime = App.dateTime();
        clusterStatus.installTime = ((parseInt(serviceStartTime) - parseInt(this.get('content.cluster.installStartTime'))) / 60000).toFixed(2);
      }
      else {
        clusterStatus.status = 'START FAILED'; // 'START FAILED' implies to step10 that installation was successful but start failed
      }
      this.saveClusterStatus(clusterStatus);
      this.saveInstalledHosts(this);
      return true;
    }
    return false;
  },

  /**
   * @param polledData Json data retrieved from API
   * @method setIsServicesInstalled
   */
  setIsServicesInstalled: function (polledData) {
    var clusterStatus = {};
    var self = this;

    if (!polledData.someProperty('Tasks.status', 'PENDING') && !polledData.someProperty('Tasks.status', 'QUEUED') && !polledData.someProperty('Tasks.status', 'IN_PROGRESS')) {
      clusterStatus = {
        status: 'PENDING',
        requestId: this.get('content.cluster.requestId'),
        isCompleted: false
      };
      if (this.get('status') === 'failed') {
        clusterStatus.status = 'INSTALL FAILED';
        this.saveClusterStatus(clusterStatus);
        this.set('progress', '100');
        this.get('hosts').forEach(function (host) {
          host.set('progress', '100');
        });
        this.isAllComponentsInstalled().done(function(){
          self.changeParseHostInfo(false);
        });
        return;
      } else {
        this.set('progress', '34');
        if (this.get('content.controllerName') === 'installerController') {
          this.isAllComponentsInstalled().done(function(){
            self.saveInstalledHosts(self);
            self.changeParseHostInfo(true);
          });
          return;
        } else {
          this.launchStartServices(function () {
            self.saveInstalledHosts(self);
            self.changeParseHostInfo(true);
          });
          return;
        }
      }
    }
    this.changeParseHostInfo(false);
  },

  /**
   * This is done at HostRole level.
   * @param {Ember.Enumerable} tasksPerHost
   * @param {Object} host
   * @method setLogTasksStatePerHost
   */
  setLogTasksStatePerHost: function (tasksPerHost, host) {
    tasksPerHost.forEach(function (_task) {
      var task = host.logTasks.findProperty('Tasks.id', _task.Tasks.id);
      if (task) {
        task.Tasks.status = _task.Tasks.status;
        task.Tasks.start_time = _task.Tasks.start_time;
        task.Tasks.end_time = _task.Tasks.end_time;
        task.Tasks.exit_code = _task.Tasks.exit_code;
      } else {
        host.logTasks.pushObject(_task);
      }
    }, this);
  },

  /**
   * Parses the Json data retrieved from API and sets the task on the host of {hosts} array binded to template
   * @param polledData Json data retrieved from API
   * @method setParseHostInfo
   */
  setParseHostInfo: function (polledData) {
    console.log('TRACE: Entering host info function');
    var self = this;
    var totalProgress = 0;
    var tasksData = polledData.tasks;
    console.log("The value of tasksData is: ", tasksData);
    var requestId = this.get('content.cluster.requestId');
    tasksData.setEach('Tasks.request_id', requestId);
    if (polledData.Requests && polledData.Requests.id && polledData.Requests.id != requestId) {
      // We don't want to use non-current requestId's tasks data to
      // determine the current install status.
      // Also, we don't want to keep polling if it is not the
      // current requestId.
      this.changeParseHostInfo(false);
      return;
    }
    this.replacePolledData(tasksData);
    var tasksHostMap = {};
    tasksData.forEach(function (task) {
      if (tasksHostMap[task.Tasks.host_name]) {
        tasksHostMap[task.Tasks.host_name].push(task);
      } else {
        tasksHostMap[task.Tasks.host_name] = [task];
      }
    });

    this.get('hosts').forEach(function (_host) {
      var actionsPerHost = tasksHostMap[_host.name] || []; // retrieved from polled Data
      if (actionsPerHost.length === 0) {
        if (this.get('content.cluster.status') === 'PENDING' || this.get('content.cluster.status') === 'INSTALL FAILED') {
          _host.set('progress', '33');
          _host.set('isNoTasksForInstall', true);
          _host.set('status', 'pending');
        }
        if (this.get('content.cluster.status') === 'INSTALLED' || this.get('content.cluster.status') === 'FAILED') {
          _host.set('progress', '100');
          _host.set('status', 'success');
        }
        console.log("INFO: No task is hosted on the host");
      } else {
        _host.set('isNoTasksForInstall', false);
      }
      this.setLogTasksStatePerHost(actionsPerHost, _host);
      this.onSuccessPerHost(actionsPerHost, _host);     // every action should be a success
      this.onErrorPerHost(actionsPerHost, _host);     // any action should be a failure
      this.onInProgressPerHost(actionsPerHost, _host);  // current running action for a host
      totalProgress += self.progressPerHost(actionsPerHost, _host);
      if (_host.get('progress') == '33' && _host.get('status') != 'failed' && _host.get('status') != 'warning') {
        _host.set('message', this.t('installer.step9.host.status.nothingToInstall'));
        _host.set('status', 'pending');
      }
    }, this);
    this.set('logTasksChangesCounter', this.get('logTasksChangesCounter') + 1);
    totalProgress = Math.floor(totalProgress / this.get('hosts.length'));
    this.set('progress', totalProgress.toString());
    console.log("INFO: right now the progress is: " + this.get('progress'));
    this.setFinishState(tasksData);
  },

  /**
   * Starts polling to the API
   * @method startPolling
   */
  startPolling: function () {
    this.set('isSubmitDisabled', true);
    this.doPolling();
  },

  /**
   * This function calls API just once to fetch log data of all tasks.
   * @method loadLogData
   */
  loadLogData: function (startPolling) {
    var requestsId = this.get('wizardController').getDBProperty('cluster').oldRequestsId;
    if (App.get('testMode')) {
      this.set('POLL_INTERVAL', 1);
    }
    this.getLogsByRequest(!!startPolling, requestsId[requestsId.length-1]);
  },

  /**
   * Load form server <code>stderr, stdout</code> of current open task
   * @method loadCurrentTaskLog
   */
  loadCurrentTaskLog: function () {
    var taskId = this.get('currentOpenTaskId');
    var requestId = this.get('currentOpenTaskRequestId');
    var clusterName = this.get('content.cluster.name');
    if (!taskId) {
      console.log('taskId is null.');
      return;
    }
    App.ajax.send({
      name: 'background_operations.get_by_task',
      sender: this,
      data: {
        'taskId': taskId,
        'requestId': requestId,
        'clusterName': clusterName
      },
      success: 'loadCurrentTaskLogSuccessCallback',
      error: 'loadCurrentTaskLogErrorCallback'
    });
  },

  /**
   * success callback function for getting log data of the opened task
   * @param {object} data
   * @method loadCurrentTaskLogSuccessCallback
   */
  loadCurrentTaskLogSuccessCallback: function (data) {
    var taskId = this.get('currentOpenTaskId');
    if (taskId) {
      var host = this.get('hosts').findProperty('name', data.Tasks.host_name);
      if (host) {
        var currentTask = host.get('logTasks').findProperty('Tasks.id', data.Tasks.id);
        if (currentTask) {
          currentTask.Tasks.stderr = data.Tasks.stderr;
          currentTask.Tasks.stdout = data.Tasks.stdout;
          currentTask.Tasks.output_log = data.Tasks.output_log;
          currentTask.Tasks.error_log = data.Tasks.error_log;
        }
      }
    }
    this.set('logTasksChangesCounter', this.get('logTasksChangesCounter') + 1);
  },

  /**
   * Error callback function for getting log data of the opened task
   * @method loadCurrentTaskLogErrorCallback
   */
  loadCurrentTaskLogErrorCallback: function () {
    this.set('currentOpenTaskId', 0);
  },

  /**
   * Function polls the API to retrieve data for the request.
   * @param {bool} polling whether to continue polling for status or not
   * @param {number} requestId
   * @method getLogsByRequest
   * @return {$.ajax|null}
   */
  getLogsByRequest: function (polling, requestId) {
    return App.ajax.send({
      name: 'wizard.step9.load_log',
      sender: this,
      data: {
        polling: polling,
        cluster: this.get('content.cluster.name'),
        requestId: requestId,
        numPolls: this.get('numPolls')
      },
      success: 'getLogsByRequestSuccessCallback',
      error: 'getLogsByRequestErrorCallback'
    }).retry({times: App.maxRetries, timeout: 3000}).then(null,
      function () {
        App.showReloadPopup();
        console.log('Install services all retries failed');
      }
    );
  },

  /**
   * Success callback for get log by request
   * @param {object} data
   * @param {object} opt
   * @param {object} params
   * @method getLogsByRequestSuccessCallback
   */
  getLogsByRequestSuccessCallback: function (data, opt, params) {
    var parsedData = jQuery.parseJSON(data);
    console.log("TRACE: In success function for the GET logs data");
    console.log("TRACE: Step9 -> The value is: ", parsedData);
    this.set('isPolling', params.polling);
    this.setParseHostInfo(parsedData);
  },

  /**
   * Error-callback for get log by request
   * @method getLogsByRequestErrorCallback
   */
  getLogsByRequestErrorCallback: function () {
    this.loadLogData(true);
  },

  /**
   * Delegates the function call to {getLogsByRequest} with appropriate params
   * @method doPolling
   */
  doPolling: function () {
    var requestId = this.get('content.cluster.requestId');
    if (App.get('testMode')) {
      this.incrementProperty('numPolls');
    }
    this.getLogsByRequest(true, requestId);
  },

  /**
   * Check that all components are in INSTALLED state before issuing start command
   * @method isAllComponentsInstalled
   * @return {$.ajax|null}
   */
  isAllComponentsInstalled: function () {
    var dfd = $.Deferred();
    if (this.get('content.controllerName') !== 'installerController' && this.get('content.controllerName') !== 'addHostController') {
      dfd.resolve();
    } else {
      App.ajax.send({
        name: 'wizard.step9.installer.get_host_status',
        sender: this,
        data: {
          cluster: this.get('content.cluster.name')
        },
        success: 'isAllComponentsInstalledSuccessCallback',
        error: 'isAllComponentsInstalledErrorCallback'
      }).complete(function(){
          dfd.resolve();
        });
    }
    return dfd.promise();
  },

  /**
   * Success callback function for API checking host state and host_components state.
   * @param {Object} jsonData
   * @method isAllComponentsInstalledSuccessCallback
   */
  isAllComponentsInstalledSuccessCallback: function (jsonData) {
    var clusterStatus = {
      status: 'INSTALL FAILED',
      isStartError: true,
      isCompleted: false
    };
    var hostsWithHeartbeatLost = [];
    jsonData.items.filterProperty('Hosts.host_state', 'HEARTBEAT_LOST').forEach(function (host) {
      var hostComponentObj = {hostName: host.Hosts.host_name};
      var componentArr = [];
      host.host_components.forEach(function (_hostComponent) {
        var componentName = App.format.role(_hostComponent.HostRoles.component_name);
        componentArr.pushObject(componentName);
      }, this);
      hostComponentObj.componentNames = stringUtils.getFormattedStringFromArray(componentArr);
      hostsWithHeartbeatLost.pushObject(hostComponentObj);
    }, this);
    this.set('hostsWithHeartbeatLost', hostsWithHeartbeatLost);
    if (hostsWithHeartbeatLost.length) {
      this.get('hosts').forEach(function (host) {
        if (hostsWithHeartbeatLost.someProperty(('hostName'), host.get('name'))) {
          host.set('status', 'heartbeat_lost');
        } else if (host.get('status') !== 'failed' && host.get('status') !== 'warning') {
          host.set('message', Em.I18n.t('installer.step9.host.status.startAborted'));
        }
        host.set('progress', '100');
      });
      this.set('progress', '100');
      this.saveClusterStatus(clusterStatus);
    } else if (this.get('content.cluster.status') === 'PENDING') {
      this.launchStartServices();
    }

  },

  /**
   * Error callback function for API checking host state and host_components state
   * @method isAllComponentsInstalledErrorCallback
   */
  isAllComponentsInstalledErrorCallback: function () {
    var clusterStatus = {
      status: 'INSTALL FAILED',
      isStartError: true,
      isCompleted: false
    };
    this.set('progress', '100');
    this.get('hosts').forEach(function (host) {
      if (host.get('status') !== 'failed' && host.get('status') !== 'warning') {
        host.set('message', Em.I18n.t('installer.step9.host.status.startAborted'));
        host.set('progress', '100');
      }
    });
    this.saveClusterStatus(clusterStatus);
  },

  /**
   * save cluster status in the parentController and localdb
   * @param {object} clusterStatus
   * @method saveClusterStatus
   */
  saveClusterStatus: function (clusterStatus) {
    if (!App.get('testMode')) {
      App.router.get(this.get('content.controllerName')).saveClusterStatus(clusterStatus);
    }
    else {
      this.set('content.cluster', clusterStatus);
    }
  },

  /**
   * save cluster status in the parentController and localdb
   * @param {object} context
   * @method saveInstalledHosts
   */
  saveInstalledHosts: function (context) {
    if (!App.get('testMode')) {
      App.router.get(this.get('content.controllerName')).saveInstalledHosts(context);
    }
  }

});

});

require.register("data/BIGTOP/site_properties", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

module.exports =
{
  "configProperties": [
  //***************************************** BIGTOP stack **************************************
  /**********************************************HDFS***************************************/
    {
      "id": "site property",
      "name": "dfs.namenode.checkpoint.dir",
      "displayName": "SecondaryNameNode Checkpoint directories",
      "defaultDirectory": "/hadoop/hdfs/namesecondary",
      "displayType": "directories",
      "isOverridable": false,
      "serviceName": "HDFS",
      "filename": "hdfs-site.xml",
      "category": "SECONDARY_NAMENODE",
      "index": 1
    },
    {
      "id": "site property",
      "name": "dfs.namenode.checkpoint.period",
      "displayName": "HDFS Maximum Checkpoint Delay",
      "displayType": "int",
      "unit": "seconds",
      "category": "General",
      "serviceName": "HDFS",
      "filename": "hdfs-site.xml",
      "index": 3
    },
    {
      "id": "site property",
      "name": "dfs.namenode.name.dir",
      "displayName": "NameNode directories",
      "defaultDirectory": "/hadoop/hdfs/namenode",
      "displayType": "directories",
      "isOverridable": false,
      "serviceName": "HDFS",
      "filename": "hdfs-site.xml",
      "category": "NAMENODE",
      "index": 1
    },
    {
      "id": "site property",
      "name": "dfs.webhdfs.enabled",
      "displayName": "WebHDFS enabled",
      "displayType": "checkbox",
      "isOverridable": false,
      "category": "General",
      "serviceName": "HDFS",
      "filename": "hdfs-site.xml",
      "index": 0
    },
    {
      "id": "site property",
      "name": "dfs.datanode.failed.volumes.tolerated",
      "displayName": "DataNode volumes failure toleration",
      "displayType": "int",
      "category": "DATANODE",
      "serviceName": "HDFS",
      "filename": "hdfs-site.xml",
      "index": 3
    },
    {
      "id": "site property",
      "name": "dfs.datanode.data.dir.mount.file",
      "displayName": "File that stores mount point for each data dir",
      "description": "File path that contains the last known mount point for each data dir. This file is used to avoid creating a DFS data dir on the root drive (and filling it up) if a path was previously mounted on a drive.",
      "defaultValue": "/etc/hadoop/conf/dfs_data_dir_mount.hist",
      "displayType": "directory",
      "isVisible": true,
      "category": "DATANODE",
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "index": 4
    },
    {
      "id": "site property",
      "name": "dfs.datanode.data.dir",
      "displayName": "DataNode directories",
      "defaultDirectory": "/hadoop/hdfs/data",
      "displayType": "directories",
      "category": "DATANODE",
      "serviceName": "HDFS",
      "filename": "hdfs-site.xml",
      "index": 1
    },
    {
      "id": "site property",
      "name": "dfs.datanode.data.dir.perm",
      "displayName": "DataNode directories permission",
      "displayType": "int",
      "category": "DATANODE",
      "serviceName": "HDFS",
      "filename": "hdfs-site.xml"
    },
    {
      "id": "site property",
      "name": "dfs.replication",
      "displayName": "Block replication",
      "displayType": "int",
      "category": "General",
      "serviceName": "HDFS",
      "filename": "hdfs-site.xml"
    },
    {
      "id": "site property",
      "name": "dfs.datanode.du.reserved",
      "displayName": "Reserved space for HDFS",
      "displayType": "int",
      "unit": "bytes",
      "category": "General",
      "serviceName": "HDFS",
      "filename": "hdfs-site.xml",
      "index": 2
    },
    {
      "id": "site property",
      "name": "dfs.client.read.shortcircuit",
      "displayName": "HDFS Short-circuit read",
      "displayType": "checkbox",
      "category": "Advanced hdfs-site",
      "serviceName": "HDFS",
      "filename": "hdfs-site.xml"
    },
    {
      "id": "site property",
      "name": "apache_artifacts_download_url",
      "displayName": "apache_artifacts_download_url",
      "description": "",
      "isRequired": false,
      "isRequiredByAgent": false,
      "isVisible": false,
      "category": "Advanced hdfs-site",
      "serviceName": "HDFS"
    },

  /**********************************************YARN***************************************/
    {
      "id": "site property",
      "name": "yarn.acl.enable",
      "displayName": "yarn.acl.enable",
      "displayType": "checkbox",
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "RESOURCEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.admin.acl",
      "displayName": "yarn.admin.acl",
      "isRequired": false,
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "RESOURCEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.log-aggregation-enable",
      "displayName": "yarn.log-aggregation-enable",
      "displayType": "checkbox",
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "RESOURCEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.resourcemanager.scheduler.class",
      "displayName": "yarn.resourcemanager.scheduler.class",
      "serviceName": "YARN",
      "category": "CapacityScheduler",
      "filename": "yarn-site.xml"
    },
    {
      "id": "site property",
      "name": "yarn.scheduler.minimum-allocation-mb",
      "displayName": "yarn.scheduler.minimum-allocation-mb",
      "displayType": "int",
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "CapacityScheduler"
    },
    {
      "id": "site property",
      "name": "yarn.scheduler.maximum-allocation-mb",
      "displayName": "yarn.scheduler.maximum-allocation-mb",
      "displayType": "int",
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "CapacityScheduler"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.resource.memory-mb",
      "displayName": "yarn.nodemanager.resource.memory-mb",
      "displayType": "int",
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.vmem-pmem-ratio",
      "displayName": "yarn.nodemanager.vmem-pmem-ratio",
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.linux-container-executor.group",
      "displayName": "yarn.nodemanager.linux-container-executor.group",
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.log-dirs",
      "displayName": "yarn.nodemanager.log-dirs",
      "defaultDirectory": "/hadoop/yarn/log",
      "displayType": "directories",
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.local-dirs",
      "displayName": "yarn.nodemanager.local-dirs",
      "defaultDirectory": "/hadoop/yarn/local",
      "displayType": "directories",
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.remote-app-log-dir",
      "displayName": "yarn.nodemanager.remote-app-log-dir",
      "displayType": "directory",
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.remote-app-log-dir-suffix",
      "displayName": "yarn.nodemanager.remote-app-log-dir-suffix",
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.aux-services",
      "displayName": "yarn.nodemanager.aux-services",
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.log.retain-second",
      "displayName": "yarn.nodemanager.log.retain-second",
      "serviceName": "YARN",
      "filename": "yarn-site.xml",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.log.server.url",
      "displayName": "yarn.log.server.url",
      "category": "Advanced yarn-site",
      "serviceName": "YARN",
      "filename": "yarn-site.xml"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.enabled",
      "displayName": "yarn.timeline-service.enabled",
      "category": "APP_TIMELINE_SERVER",
      "displayType": "checkbox",
      "serviceName": "YARN",
      "filename": "yarn-site.xml"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.leveldb-timeline-store.path",
      "displayName": "yarn.timeline-service.leveldb-timeline-store.path",
      "defaultDirectory": "/hadoop/yarn/timeline",
      "category": "APP_TIMELINE_SERVER",
      "displayType": "directory",
      "serviceName": "YARN",
      "filename": "yarn-site.xml"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.leveldb-timeline-store.ttl-interval-ms",
      "displayName": "yarn.timeline-service.leveldb-timeline-store.ttl-interval-ms",
      "displayType": "int",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN",
      "filename": "yarn-site.xml"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.store-class",
      "displayName": "yarn.timeline-service.store-class",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN",
      "filename": "yarn-site.xml"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.ttl-enable",
      "displayName": "yarn.timeline-service.ttl-enable",
      "displayType": "checkbox",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN",
      "filename": "yarn-site.xml"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.ttl-ms",
      "displayName": "yarn.timeline-service.ttl-ms",
      "displayType": "int",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN",
      "filename": "yarn-site.xml"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.generic-application-history.store-class",
      "displayName": "yarn.timeline-service.generic-application-history.store-class",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN",
      "filename": "yarn-site.xml"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.webapp.address",
      "displayName": "yarn.timeline-service.webapp.address",
      "displayType": "string",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN",
      "filename": "yarn-site.xml"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.webapp.https.address",
      "displayName": "yarn.timeline-service.webapp.https.address",
      "displayType": "string",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN",
      "filename": "yarn-site.xml"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.address",
      "displayName": "yarn.timeline-service.address",
      "displayType": "string",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN",
      "filename": "yarn-site.xml"
    },
  /**********************************************MAPREDUCE2***************************************/
    {
      "id": "site property",
      "name": "mapreduce.map.memory.mb",
      "displayName": "Default virtual memory for a job's map-task",
      "displayType": "int",
      "unit": "MB",
      "category": "General",
      "serviceName": "MAPREDUCE2",
      "filename": "mapred-site.xml"
    },
    {
      "id": "site property",
      "name": "mapreduce.reduce.memory.mb",
      "displayName": "Default virtual memory for a job's reduce-task",
      "displayType": "int",
      "unit": "MB",
      "category": "General",
      "serviceName": "MAPREDUCE2",
      "filename": "mapred-site.xml"
    },
    {
      "id": "site property",
      "name": "mapreduce.task.io.sort.mb",
      "displayName": "Map-side sort buffer memory",
      "displayType": "int",
      "unit": "MB",
      "category": "General",
      "serviceName": "MAPREDUCE2",
      "filename": "mapred-site.xml"
    },
    {
      "id": "site property",
      "name": "hadoop.security.auth_to_local",
      "displayName": "hadoop.security.auth_to_local",
      "displayType": "multiLine",
      "serviceName": "HDFS",
      "filename": "core-site.xml",
      "category": "Advanced core-site"
    },
    {
      "id": "site property",
      "name": "yarn.app.mapreduce.am.resource.mb",
      "displayName": "yarn.app.mapreduce.am.resource.mb",
      "displayType": "int",
      "category": "Advanced mapred-site",
      "serviceName": "MAPREDUCE2",
      "filename": "mapred-site.xml"
    },

  /**********************************************oozie-site***************************************/
    {
      "id": "site property",
      "name": "oozie.db.schema.name",
      "displayName": "Database Name",
      "isOverridable": false,
      "displayType": "host",
      "isObserved": true,
      "category": "OOZIE_SERVER",
      "serviceName": "OOZIE",
      "filename": "oozie-site.xml",
      "index": 4
    },
    {
      "id": "site property",
      "name": "oozie.service.JPAService.jdbc.username",
      "displayName": "Database Username",
      "isOverridable": false,
      "displayType": "host",
      "category": "OOZIE_SERVER",
      "serviceName": "OOZIE",
      "filename": "oozie-site.xml",
      "index": 5
    },
    {
      "id": "site property",
      "name": "oozie.service.JPAService.jdbc.password",
      "displayName": "Database Password",
      "isOverridable": false,
      "displayType": "password",
      "category": "OOZIE_SERVER",
      "serviceName": "OOZIE",
      "filename": "oozie-site.xml",
      "index": 6
    },
    {
      "id": "site property",
      "name": "oozie.service.JPAService.jdbc.driver", // the default value of this property is overriden in code
      "displayName": "JDBC Driver Class",
      "isOverridable": false,
      "category": "OOZIE_SERVER",
      "serviceName": "OOZIE",
      "filename": "oozie-site.xml",
      "index": 7
    },
    {
      "id": "site property",
      "name": "oozie.service.JPAService.jdbc.url",
      "displayName": "Database URL",
      "isOverridable": false,
      "displayType": "advanced",
      "category": "OOZIE_SERVER",
      "serviceName": "OOZIE",
      "filename": "oozie-site.xml",
      "index": 8
    },

  /**********************************************hive-site***************************************/
    {
      "id": "site property",
      "name": "javax.jdo.option.ConnectionDriverName",  // the default value is overwritten in code
      "displayName": "JDBC Driver Class",
      "isOverridable": false,
      "category": "HIVE_METASTORE",
      "serviceName": "HIVE",
      "filename": "hive-site.xml",
      "index": 7
    },
    {
      "id": "site property",
      "name": "hive.heapsize",
      "displayName": "Hive heap size",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "serviceName": "HIVE",
      "filename": "hive-site.xml",
      "category": "General",
      "index": 9
    },
    {
      "id": "site property",
      "name": "javax.jdo.option.ConnectionUserName",
      "displayName": "Database Username",
      "displayType": "host",
      "isOverridable": false,
      "category": "HIVE_METASTORE",
      "serviceName": "HIVE",
      "filename": "hive-site.xml",
      "index": 5
    },
    {
      "id": "site property",
      "name": "javax.jdo.option.ConnectionPassword",
      "displayName": "Database Password",
      "displayType": "password",
      "isOverridable": false,
      "category": "HIVE_METASTORE",
      "serviceName": "HIVE",
      "filename": "hive-site.xml",
      "index": 6
    },
    {
      "id": "site property",
      "name": "javax.jdo.option.ConnectionURL",
      "displayName": "Database URL",
      "displayType": "advanced",
      "isOverridable": false,
      "category": "HIVE_METASTORE",
      "serviceName": "HIVE",
      "filename": "hive-site.xml",
      "index": 8
    },
    {
      "id": "site property",
      "name": "ambari.hive.db.schema.name",
      "displayName": "Database Name",
      "displayType": "host",
      "isOverridable": false,
      "isObserved": true,
      "serviceName": "HIVE",
      "filename": "hive-site.xml",
      "category": "HIVE_METASTORE",
      "index": 4
    },
    {
      "id": "site property",
      "name": "hive.server2.tez.default.queues",
      "displayName": "hive.server2.tez.default.queues",
      "isRequired": false,
      "serviceName": "HIVE",
      "category": "Advanced hive-site"
    },
    {
      "id": "site property",
      "name": "hive.server2.thrift.port",
      "displayName": "Hive Server Port",
      "description": "TCP port number to listen on, default 10000.",
      "defaultValue": "10000",
      "displayType": "int",
      "isReconfigurable": true,
      "isOverridable": false,
      "isVisible": true,
      "category": "Advanced hive-site",
      "serviceName": "HIVE",
      "filename": "hive-site.xml"
    },
    {
      "id": "site property",
      "name": "hive.server2.support.dynamic.service.discovery",
      "displayName": "hive.server2.support.dynamic.service.discovery",
      "defaultValue": true,
      "displayType": "checkbox",
      "category": "Advanced hive-site",
      "serviceName": "HIVE"
    },
    {
      "id": "site property",
      "name": "hive.zookeeper.quorum",
      "displayName": "hive.zookeeper.quorum",
      "defaultValue": "localhost:2181",
      "displayType": "multiLine",
      "isVisible": true,
      "serviceName": "HIVE",
      "category": "Advanced hive-site"
    },
  /**********************************************tez-site*****************************************/
    {
      "id": "site property",
      "name": "tez.am.resource.memory.mb",
      "displayName": "tez.am.resource.memory.mb",
      "displayType": "int",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.am.java.opts",
      "displayName": "tez.am.java.opts",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.am.grouping.split-waves",
      "displayName": "tez.am.grouping.split-waves",
      "displayType": "float",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.am.grouping.min-size",
      "displayName": "tez.am.grouping.min-size",
      "displayType": "int",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.am.grouping.max-size",
      "displayName": "tez.am.grouping.max-size",
      "displayType": "int",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.am.log.level",
      "displayName": "tez.am.log.level",
      "displayType": "string",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.runtime.intermediate-input.compress.codec",
      "displayName": "tez.runtime.intermediate-input.compress.codec",
      "displayType": "string",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.runtime.intermediate-input.is-compressed",
      "displayName": "tez.runtime.intermediate-input.is-compressed",
      "displayType": "checkbox",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.runtime.intermediate-output.compress.codec",
      "displayName": "tez.runtime.intermediate-output.compress.codec",
      "displayType": "string",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.runtime.intermediate-output.should-compress",
      "displayName": "tez.runtime.intermediate-output.should-compress",
      "displayType": "checkbox",
      "category": "General",
      "serviceName": "TEZ"
    },

  /**********************************************hbase-site***************************************/
    {
      "id": "site property",
      "name": "hbase.tmp.dir",
      "displayName": "HBase local directory",
      "defaultDirectory": "/hadoop/hbase",
      "displayType": "directory",
      "category": "Advanced hbase-site",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml"
    },
    {
      "id": "site property",
      "name": "hbase.master.port",
      "displayName": "HBase Master Port",
      "isReconfigurable": true,
      "displayType": "int",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HBASE",
      "filename": "hbase-site.xml",
      "category": "Advanced hbase-site"
    },
    {
      "id": "site property",
      "name": "hbase.regionserver.global.memstore.upperLimit",
      "displayName": "hbase.regionserver.global.memstore.upperLimit",
      "displayType": "float",
      "category": "Advanced hbase-site",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml"
    },
    {
      "id": "site property",
      "name": "hbase.regionserver.global.memstore.lowerLimit",
      "displayName": "hbase.regionserver.global.memstore.lowerLimit",
      "displayType": "float",
      "category": "Advanced hbase-site",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml"
    },
    {
      "id": "site property",
      "name": "hbase.hstore.blockingStoreFiles",
      "displayName": "hstore blocking storefiles",
      "displayType": "int",
      "category": "Advanced hbase-site",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml"
    },
    {
      "id": "site property",
      "name": "hbase.hstore.compactionThreshold",
      "displayName": "HBase HStore compaction threshold",
      "displayType": "int",
      "category": "General",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml",
      "index": 0
    },
    {
      "id": "site property",
      "name": "hfile.block.cache.size",
      "displayName": "HFile block cache size ",
      "displayType": "float",
      "category": "General",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml",
      "index": 1
    },
    {
      "id": "site property",
      "name": "hbase.hregion.max.filesize",
      "displayName": "Maximum HStoreFile Size",
      "displayType": "int",
      "unit": "bytes",
      "category": "General",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml",
      "index": 2
    },
    {
      "id": "site property",
      "name": "hbase.regionserver.handler.count",
      "displayName": "RegionServer Handler",
      "displayType": "int",
      "category": "HBASE_REGIONSERVER",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml",
      "index": 2
    },
    {
      "id": "site property",
      "name": "hbase.hregion.majorcompaction",
      "displayName": "HBase Region Major Compaction",
      "displayType": "int",
      "unit": "ms",
      "category": "HBASE_REGIONSERVER",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml",
      "index": 3
    },
    {
      "id": "site property",
      "name": "hbase.hregion.memstore.block.multiplier",
      "displayName": "HBase Region Block Multiplier",
      "displayType": "int",
      "category": "HBASE_REGIONSERVER",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml",
      "index": 4
    },
    {
      "id": "site property",
      "name": "hbase.hregion.memstore.mslab.enabled",
      "displayName": "hbase.hregion.memstore.mslab.enabled",
      "displayType": "checkbox",
      "category": "Advanced hbase-site",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml"
    },
    {
      "id": "site property",
      "name": "hbase.hregion.memstore.flush.size",
      "displayName": "HBase Region Memstore Flush Size",
      "displayType": "int",
      "unit": "bytes",
      "category": "HBASE_REGIONSERVER",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml",
      "index": 5
    },
    {
      "id": "site property",
      "name": "hbase.client.scanner.caching",
      "displayName": "HBase Client Scanner Caching",
      "displayType": "int",
      "unit": "rows",
      "category": "General",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml",
      "index": 3
    },
    {
      "id": "site property",
      "name": "zookeeper.session.timeout",
      "displayName": "Zookeeper timeout for HBase Session",
      "displayType": "int",
      "unit": "ms",
      "category": "General",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml",
      "index": 4
    },
    {
      "id": "site property",
      "name": "hbase.client.keyvalue.maxsize",
      "displayName": "HBase Client Maximum key-value Size",
      "displayType": "int",
      "unit": "bytes",
      "category": "General",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml",
      "index": 5
    },
    {
      "id": "site property",
      "name": "hbase.coprocessor.region.classes",
      "displayName": "hbase.coprocessor.region.classes",
      "category": "Advanced hbase-site",
      "isRequired": false,
      "serviceName": "HBASE",
      "filename": "hbase-site.xml"
    },
    {
      "id": "site property",
      "name": "hbase.coprocessor.master.classes",
      "displayName": "hbase.coprocessor.master.classes",
      "category": "Advanced hbase-site",
      "isRequired": false,
      "serviceName": "HBASE",
      "filename": "hbase-site.xml"
    },
    {
      "id": "site property",
      "name": "hbase.zookeeper.quorum",
      "displayName": "hbase.zookeeper.quorum",
      "displayType": "multiLine",
      "serviceName": "HBASE",
      "filename": "hbase-site.xml",
      "category": "Advanced hbase-site"
    },

  /**********************************************storm-site***************************************/
    {
      "id": "site property",
      "name": "storm.zookeeper.root",
      "displayName": "storm.zookeeper.root",
      "displayType": "directory",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.local.dir",
      "displayName": "storm.local.dir",
      "defaultDirectory": "/hadoop/storm",
      "displayType": "directory",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.servers",
      "displayName": "storm.zookeeper.servers",
      "displayType": "masterHosts",
      "isOverridable": false,
      "isReconfigurable": false,
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.port",
      "displayName": "storm.zookeeper.port",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.session.timeout",
      "displayName": "storm.zookeeper.session.timeout",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.connection.timeout",
      "displayName": "storm.zookeeper.connection.timeout",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.retry.times",
      "displayName": "storm.zookeeper.retry.times",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.retry.interval",
      "displayName": "storm.zookeeper.retry.interval",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "General",
      "unit": "ms"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.retry.intervalceiling.millis",
      "displayName": "storm.zookeeper.retry.intervalceiling.millis",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "General",
      "unit": "ms"
    },
    {
      "id": "site property",
      "name": "storm.cluster.mode",
      "displayName": "storm.cluster.mode",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.local.mode.zmq",
      "displayName": "storm.local.mode.zmq",
      "displayType": "checkbox",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.thrift.transport",
      "displayName": "storm.thrift.transport",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.messaging.transport",
      "displayName": "storm.messaging.transport",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "General",
      "displayName": "storm.messaging.netty.buffer_size",
      "name": "storm.messaging.netty.buffer_size",
      "displayType": "int",
      "unit": "bytes"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "General",
      "displayName": "storm.messaging.netty.max_retries",
      "name": "storm.messaging.netty.max_retries",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "General",
      "displayName": "storm.messaging.netty.max_wait_ms",
      "name": "storm.messaging.netty.max_wait_ms",
      "displayType": "int",
      "unit": "ms"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "General",
      "displayName": "storm.messaging.netty.min_wait_ms",
      "name": "storm.messaging.netty.min_wait_ms",
      "displayType": "int",
      "unit": "ms"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "General",
      "displayName": "storm.messaging.netty.server_worker_threads",
      "name": "storm.messaging.netty.server_worker_threads",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "General",
      "displayName": "storm.messaging.netty.client_worker_threads",
      "name": "storm.messaging.netty.client_worker_threads",
      "displayType": "int"
    },
    {
      "id": "site property",
      "name": "nimbus.host",
      "displayName": "nimbus.host",
      "displayType": "masterHost",
      "isOverridable": false,
      "isReconfigurable": false,
      "serviceName": "STORM",
      "category": "NIMBUS"
    },
    {
      "id": "site property",
      "name": "nimbus.thrift.port",
      "displayName": "nimbus.thrift.port",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS"
    },
    {
      "id": "site property",
      "name": "nimbus.thrift.max_buffer_size",
      "displayName": "nimbus.thrift.max_buffer_size",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "bytes"
    },
    {
      "id": "site property",
      "name": "nimbus.childopts",
      "displayName": "nimbus.childopts",
      "displayType": "multiLine",
      "isOverridable": false,
      "serviceName": "STORM",
      "category": "NIMBUS",
      "filename": "storm-site.xml"
    },
    {
      "id": "site property",
      "name": "nimbus.task.timeout.secs",
      "displayName": "nimbus.task.timeout.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.supervisor.timeout.secs",
      "displayName": "nimbus.supervisor.timeout.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.monitor.freq.secs",
      "displayName": "nimbus.monitor.freq.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.cleanup.inbox.freq.secs",
      "displayName": "nimbus.cleanup.inbox.freq.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.inbox.jar.expiration.secs",
      "displayName": "nimbus.inbox.jar.expiration.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.task.launch.secs",
      "displayName": "nimbus.task.launch.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.reassign",
      "displayName": "nimbus.reassign",
      "displayType": "checkbox",
      "isReconfigurable": true,
      "serviceName": "STORM",
      "category": "NIMBUS"
    },
    {
      "id": "site property",
      "name": "nimbus.file.copy.expiration.secs",
      "displayName": "nimbus.file.copy.expiration.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.topology.validator",
      "displayName": "nimbus.topology.validator",
      "serviceName": "STORM",
      "category": "NIMBUS"
    },
    {
      "id": "site property",
      "name": "supervisor.slots.ports",
      "displayName": "supervisor.slots.ports",
      "displayType": "string",
      "serviceName": "STORM",
      "category": "SUPERVISOR"
    },
    {
      "id": "site property",
      "isOverrideable": false,
      "serviceName": "STORM",
      "category": "SUPERVISOR",
      "displayName": "supervisor.childopts",
      "name": "supervisor.childopts",
      "displayType": "multiLine",
      "filename": "storm-site.xml"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "SUPERVISOR",
      "displayName": "supervisor.worker.start.timeout.secs",
      "name": "supervisor.worker.start.timeout.secs",
      "displayType": "int",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "SUPERVISOR",
      "displayName": "supervisor.worker.timeout.secs",
      "name": "supervisor.worker.timeout.secs",
      "displayType": "int",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "SUPERVISOR",
      "displayName": "supervisor.monitor.frequency.secs",
      "name": "supervisor.monitor.frequency.secs",
      "displayType": "int",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "SUPERVISOR",
      "displayName": "supervisor.heartbeat.frequency.secs",
      "name": "supervisor.heartbeat.frequency.secs",
      "displayType": "int",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "DRPC_SERVER",
      "displayName": "drpc.port",
      "name": "drpc.port",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "DRPC_SERVER",
      "displayName": "drpc.worker.threads",
      "name": "drpc.worker.threads",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "DRPC_SERVER",
      "displayName": "drpc.queue.size",
      "name": "drpc.queue.size",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "DRPC_SERVER",
      "displayName": "drpc.invocations.port",
      "name": "drpc.invocations.port",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "DRPC_SERVER",
      "displayName": "drpc.request.timeout.secs",
      "name": "drpc.request.timeout.secs",
      "displayType": "int",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "DRPC_SERVER",
      "displayName": "drpc.childopts",
      "name": "drpc.childopts",
      "displayType": "string"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "STORM_UI_SERVER",
      "displayName": "ui.port",
      "name": "ui.port",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "STORM_UI_SERVER",
      "displayName": "ui.childopts",
      "name": "ui.childopts",
      "displayType": "string"
    },
    //@Todo: uncomment following properties when logviewer is treated as different section on storm service page
    /*
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "LogviewerServer",
      "displayName": "logviewer.port",
      "name": "logviewer.port",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "LogviewerServer",
      "displayName": "logviewer.childopts",
      "name": "logviewer.childopts",
      "displayType": "string"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "LogviewerServer",
      "displayName": "logviewer.appender.name",
      "name": "logviewer.appender.name",
      "displayType": "string"
    },
    */
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "Advanced storm-site",
      "displayName": "worker.childopts",
      "name": "worker.childopts",
      "displayType": "multiLine",
      "filename": "storm-site.xml"
    },
  /*********************************************oozie-site for Falcon*****************************/
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-job-submit-instances",
      "name": "oozie.service.ELService.ext.functions.coord-job-submit-instances",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-action-create-inst",
      "name": "oozie.service.ELService.ext.functions.coord-action-create-inst",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-action-create",
      "name": "oozie.service.ELService.ext.functions.coord-action-create",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-job-submit-data",
      "name": "oozie.service.ELService.ext.functions.coord-job-submit-data",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-action-start",
      "name": "oozie.service.ELService.ext.functions.coord-action-start",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-sla-submit",
      "name": "oozie.service.ELService.ext.functions.coord-sla-submit",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-sla-create",
      "name": "oozie.service.ELService.ext.functions.coord-sla-create",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },

    // Runtime properties
    {
      "id": "site property",
      "name": "*.domain",
      "displayName": "*.domain",
      "category": "FalconRuntimeSite",
      "serviceName": "FALCON",
      "filename": "falcon-runtime.properties.xml"

    },
    {
      "id": "site property",
      "name": "*.log.cleanup.frequency.minutes.retention",
      "displayName": "*.log.cleanup.frequency.minutes.retention",
      "category": "FalconRuntimeSite",
      "serviceName": "FALCON",
      "filename": "falcon-runtime.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.log.cleanup.frequency.hours.retention",
      "displayName": "*.log.cleanup.frequency.hours.retention",
      "category": "FalconRuntimeSite",
      "serviceName": "FALCON",
      "filename": "falcon-runtime.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.log.cleanup.frequency.days.retention",
      "displayName": "*.log.cleanup.frequency.days.retention",
      "category": "FalconRuntimeSite",
      "serviceName": "FALCON",
      "filename": "falcon-runtime.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.log.cleanup.frequency.months.retention",
      "displayName": "*.log.cleanup.frequency.months.retention",
      "category": "FalconRuntimeSite",
      "serviceName": "FALCON",
      "filename": "falcon-runtime.properties.xml"
    },

    //  Startup properties

    {
      "id": "site property",
      "name": "*.domain",
      "displayName": "*.domain",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.workflow.engine.impl",
      "displayName": "*.workflow.engine.impl",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.oozie.process.workflow.builder",
      "displayName": "*.oozie.process.workflow.builder",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.oozie.feed.workflow.builder",
      "displayName": "*.oozie.feed.workflow.builder",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.SchedulableEntityManager.impl",
      "displayName": "*.SchedulableEntityManager.impl",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.ConfigSyncService.impl",
      "displayName": "*.ConfigSyncService.impl",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.ProcessInstanceManager.impl",
      "displayName": "*.ProcessInstanceManager.impl",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.catalog.service.impl",
      "displayName": "*.catalog.service.impl",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.application.services",
      "displayName": "*.application.services",
      "displayType": "multiLine",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.configstore.listeners",
      "displayName": "*.configstore.listeners",
      "displayType": "multiLine",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.broker.impl.class",
      "displayName": "*.broker.impl.class",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.shared.libs",
      "displayName": "*.shared.libs",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.config.store.uri",
      "displayName": "*.config.store.uri",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.system.lib.location",
      "displayName": "*.system.lib.location",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.retry.recorder.path",
      "displayName": "*.retry.recorder.path",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.cleanup.service.frequency",
      "displayName": "*.falcon.cleanup.service.frequency",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.broker.url",
      "displayName": "*.broker.url",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.broker.ttlInMins",
      "displayName": "*.broker.ttlInMins",
      "displayType": "int",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.entity.topic",
      "displayName": "*.entity.topic",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.max.retry.failure.count",
      "displayName": "*.max.retry.failure.count",
      "displayType": "int",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.internal.queue.size",
      "displayName": "*.internal.queue.size",
      "displayType": "int",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.authentication.type",
      "displayName": "*.falcon.authentication.type",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.http.authentication.type",
      "displayName": "*.falcon.http.authentication.type",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.http.authentication.token.validity",
      "displayName": "*.falcon.http.authentication.token.validity",
      "displayType": "int",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.http.authentication.signature.secret",
      "displayName": "*.falcon.http.authentication.signature.secret",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.http.authentication.simple.anonymous.allowed",
      "displayName": "*.falcon.http.authentication.simple.anonymous.allowed",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.http.authentication.kerberos.name.rules",
      "displayName": "*.falcon.http.authentication.kerberos.name.rules",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.http.authentication.blacklisted.users",
      "displayName": "*.falcon.http.authentication.blacklisted.users",
      "isRequired": false,
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },

  /**********************************************webhcat-site***************************************/
    {
      "id": "site property",
      "name": "templeton.hive.archive",
      "displayName": "templeton.hive.archive",
      "isRequired": false,
      "serviceName": "HIVE",
      "filename": "webhcat-site.xml",
      "category": "Advanced webhcat-site"
    },
    {
      "id": "site property",
      "name": "templeton.pig.archive",
      "displayName": "templeton.pig.archive",
      "isRequired": false,
      "serviceName": "HIVE",
      "filename": "webhcat-site.xml",
      "category": "Advanced webhcat-site"
    },
    {
      "id": "site property",
      "name": "templeton.zookeeper.hosts",
      "displayName": "templeton.zookeeper.hosts",
      "displayType": "multiLine",
      "serviceName": "HIVE",
      "filename": "webhcat-site.xml",
      "category": "Advanced webhcat-site"
    },
  /**********************************************pig.properties*****************************************/
    {
      "id": "site property",
      "name": "content",
      "displayName": "content",
      "value": "",
      "defaultValue": "",
      "description": "pig properties",
      "displayType": "content",
      "isRequired": false,
      "showLabel": false,
      "serviceName": "PIG",
      "filename": "pig-properties.xml",
      "category": "Advanced pig-properties"
    },
  /********************************************* flume-agent *****************************/
    {
      "id": "site property",
      "name": "content",
      "displayName": "content",
      "showLabel": false,
      "isRequired": false,
      "displayType": "content",
      "serviceName": "FLUME",
      "category": "FLUME_HANDLER",
      "filename": "flume-conf.xml"
    },
    {
      "id": "puppet var",
      "name": "flume_conf_dir",
      "displayName": "Flume Conf Dir",
      "description": "Location to save configuration files",
      "defaultValue": "/etc/flume/conf",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "FLUME",
      "filename": "flume-env.xml",
      "category": "Advanced flume-env",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "flume_log_dir",
      "displayName": "Flume Log Dir",
      "description": "Location to save log files",
      "defaultValue": "/var/log/flume",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "FLUME",
      "filename": "flume-env.xml",
      "category": "Advanced flume-env",
      "index": 1
    },
  /**********************************************HDFS***************************************/
    {
      "id": "puppet var",
      "name": "namenode_host",
      "displayName": "NameNode hosts",
      "value": "",
      "defaultValue": "",
      "description": "The hosts that has been assigned to run NameNode",
      "displayType": "masterHosts",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "NAMENODE",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "namenode_heapsize",
      "displayName": "NameNode Java heap size",
      "description": "Initial and maximum Java heap size for NameNode (Java options -Xms and -Xmx).  This also applies to the Secondary NameNode.",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "NAMENODE",
      "index": 2
    },
    {
      "id": "puppet var",
      "name": "namenode_opt_newsize",
      "displayName": "NameNode new generation size",
      "description": "Default size of Java new generation for NameNode (Java option -XX:NewSize).  This also applies to the Secondary NameNode.",
      "defaultValue": "200",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "NAMENODE",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "namenode_opt_maxnewsize",
      "displayName": "NameNode maximum new generation size",
      "description": "Maximum size of Java new generation for NameNode (Java option -XX:MaxnewSize).",
      "defaultValue": "200",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "NAMENODE",
      "index": 4
    },
    {
      "id": "puppet var",
      "name": "namenode_opt_permsize",
      "displayName": "NameNode permanent generation size",
      "description": "Default size of Java new generation for NameNode (Java option -XX:PermSize).  This also applies to the Secondary NameNode.",
      "defaultValue": "128",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "NAMENODE",
      "index": 5
    },
    {
      "id": "puppet var",
      "name": "namenode_opt_maxpermsize",
      "displayName": "NameNode maximum permanent generation size",
      "description": "Maximum size of Java permanent generation for NameNode (Java option -XX:MaxPermSize).",
      "defaultValue": "256",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "NAMENODE",
      "index": 6
    },
    {
      "id": "puppet var",
      "name": "snamenode_host",
      "displayName": "SNameNode host",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run SecondaryNameNode",
      "displayType": "masterHost",
      "isOverridable": false,
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "SECONDARY_NAMENODE",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "datanode_hosts", //not in the schema. For UI purpose
      "displayName": "DataNode hosts",
      "value": "",
      "defaultValue": "",
      "description": "The hosts that have been assigned to run DataNode",
      "displayType": "slaveHosts",
      "isRequired": false,
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "DATANODE",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "dtnode_heapsize",
      "displayName": "DataNode maximum Java heap size",
      "description": "Maximum Java heap size for DataNode (Java option -Xmx)",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "DATANODE",
      "index": 2
    },
    {
      "id": "puppet var",
      "name": "hadoop_heapsize",
      "displayName": "Hadoop maximum Java heap size",
      "description": "Maximum Java heap size for daemons such as Balancer (Java option -Xmx)",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "General",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "hdfs_log_dir_prefix",
      "displayName": "Hadoop Log Dir Prefix",
      "description": "The parent directory for Hadoop log files.  The HDFS log directory will be ${hadoop_log_dir_prefix} / ${hdfs_user} and the MapReduce log directory will be ${hadoop_log_dir_prefix} / ${mapred_user}.",
      "defaultValue": "/var/log/hadoop",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "Advanced hadoop-env"
    },
    {
      "id": "puppet var",
      "name": "hadoop_pid_dir_prefix",
      "displayName": "Hadoop PID Dir Prefix",
      "description": "The parent directory in which the PID files for Hadoop processes will be created.  The HDFS PID directory will be ${hadoop_pid_dir_prefix} / ${hdfs_user} and the MapReduce PID directory will be ${hadoop_pid_dir_prefix} / ${mapred_user}.",
      "defaultValue": "/var/run/hadoop",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "Advanced hadoop-env"
    },
    {
      "id": "puppet var",
      "name": "hadoop_root_logger",
      "displayName": "Hadoop Root Logger",
      "description": "Hadoop logging options",
      "defaultValue": "INFO,RFA",
      "displayType": "string",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "Advanced hadoop-env"
    },

  /**********************************************MAPREDUCE2***************************************/
    {
      "id": "puppet var",
      "name": "hs_host",
      "displayName": "History Server",
      "description": "History Server",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "masterHost",
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "MAPREDUCE2",
      "filename": "mapred-env.xml",
      "category": "HISTORYSERVER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "jobhistory_heapsize",
      "displayName": "History Server heap size",
      "description": "History Server heap size",
      "defaultValue": "900",
      "unit": "MB",
      "isOverridable": true,
      "displayType": "int",
      "isVisible": true,
      "serviceName": "MAPREDUCE2",
      "filename": "mapred-env.xml",
      "category": "HISTORYSERVER",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "mapred_log_dir_prefix",
      "displayName": "Mapreduce Log Dir Prefix",
      "description": "",
      "defaultValue": "/var/log/hadoop-mapreduce",
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isReconfigurable": false,
      "serviceName": "MAPREDUCE2",
      "filename": "mapred-env.xml",
      "category": "Advanced mapred-env"
    },
    {
      "id": "puppet var",
      "name": "mapred_pid_dir_prefix",
      "displayName": "Mapreduce PID Dir Prefix",
      "description": "",
      "defaultValue": "/var/run/hadoop-mapreduce",
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isReconfigurable": false,
      "serviceName": "MAPREDUCE2",
      "filename": "mapred-env.xml",
      "category": "Advanced mapred-env"
    },
  /**********************************************YARN***************************************/
    {
      "id": "puppet var",
      "name": "yarn_heapsize",
      "displayName": "YARN Java heap size",
      "description": "Max heapsize for all YARN components",
      "defaultValue": "1024",
      "isOverridable": true,
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "General",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "rm_host",
      "displayName": "ResourceManager",
      "description": "ResourceManager",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "masterHost",
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "RESOURCEMANAGER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "resourcemanager_heapsize",
      "displayName": "ResourceManager Java heap size",
      "description": "Max heapsize for ResourceManager",
      "defaultValue": "1024",
      "isOverridable": false,
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "RESOURCEMANAGER",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "nm_hosts",
      "displayName": "NodeManager",
      "description": "List of NodeManager Hosts.",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "slaveHosts",
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "NODEMANAGER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "nodemanager_heapsize",
      "displayName": "NodeManager Java heap size",
      "description": "Max heapsize for NodeManager",
      "defaultValue": "1024",
      "isOverridable": true,
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "NODEMANAGER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "yarn_log_dir_prefix",
      "displayName": "YARN Log Dir Prefix",
      "description": "",
      "defaultValue": "/var/log/hadoop-yarn",
      "displayType": "directory",
      "isOverridable": false,
      "isReconfigurable": false,
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "Advanced yarn-env"
    },
    {
      "id": "puppet var",
      "name": "yarn_pid_dir_prefix",
      "displayName": "YARN PID Dir Prefix",
      "description": "",
      "defaultValue": "/var/run/hadoop-yarn",
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isReconfigurable": false,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "Advanced yarn-env"
    },
    {
      "id": "puppet var",
      "name": "min_user_id",
      "displayName": "Minimum user ID for submitting job",
      "isOverridable": true,
      "displayType": "int",
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "Advanced yarn-env"
    },
    {
      "id": "puppet var",
      "name": "ats_host",
      "displayName": "App Timeline Server",
      "description": "Application Timeline Server Host",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "masterHost",
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "APP_TIMELINE_SERVER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "apptimelineserver_heapsize",
      "displayName": "AppTimelineServer Java heap size",
      "description": "AppTimelineServer Java heap size",
      "defaultValue": "1024",
      "isOverridable": false,
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "APP_TIMELINE_SERVER",
      "index": 1
    },
  /**********************************************HBASE***************************************/
    {
      "id": "puppet var",
      "name": "hbasemaster_host",
      "displayName": "HBase Master hosts",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run HBase Master",
      "displayType": "masterHosts",
      "isOverridable": false,
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "HBASE_MASTER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "hbase_master_heapsize",
      "displayName": "HBase Master Maximum Java heap size",
      "description": "Maximum Java heap size for HBase master (Java option -Xmx)",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": true,
      "isVisible": true,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "HBASE_MASTER",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "regionserver_hosts",
      "displayName": "RegionServer hosts",
      "value": "",
      "defaultValue": "",
      "description": "The hosts that have been assigned to run RegionServer",
      "displayType": "slaveHosts",
      "isOverridable": false,
      "isRequiredByAgent": false,
      "isVisible": true,
      "isRequired": false,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "HBASE_REGIONSERVER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "hbase_regionserver_heapsize",
      "displayName": "RegionServers maximum Java heap size",
      "description": "Maximum Java heap size for RegionServers (Java option -Xmx)",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "HBASE_REGIONSERVER",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "hbase_regionserver_xmn_max",
      "displayName": "RegionServers maximum value for -Xmn",
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "HBASE_REGIONSERVER",
      "index": 6
    },
    {
      "id": "puppet var",
      "name": "hbase_regionserver_xmn_ratio",
      "displayName": "RegionServers -Xmn in -Xmx ratio",
      "displayType": "float",
      "isVisible": true,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "HBASE_REGIONSERVER",
      "index": 7
    },
    {
      "id": "puppet var",
      "name": "hbase_log_dir",
      "displayName": "HBase Log Dir",
      "description": "Directory for HBase logs",
      "defaultValue": "/var/log/hbase",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "Advanced hbase-env"
    },
    {
      "id": "puppet var",
      "name": "hbase_pid_dir",
      "displayName": "HBase PID Dir",
      "description": "Directory in which the pid files for HBase processes will be created",
      "defaultValue": "/var/run/hbase",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "Advanced hbase-env"
    },
     //***************************************** GLUSTERFS stack********************************************
    {
      "id": "site property",
      "name": "fs.glusterfs.impl",
      "displayName": "GlusterFS fs impl",
      "displayType": "string",
      "filename": "core-site.xml",
      "serviceName": "GLUSTERFS",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "fs.AbstractFileSystem.glusterfs.impl",
      "displayName": "GlusterFS Abstract File System Implementation",
      "displayType": "string",
      "filename": "core-site.xml",
      "serviceName": "GLUSTERFS",
      "category": "General"
    },
  /**********************************************GLUSTERFS***************************************/
    {
      "id": "puppet var",
      "name": "fs_glusterfs_default_name",
      "displayName": "GlusterFS default fs name 1.x Hadoop",
      "description": "GlusterFS default filesystem name (glusterfs://{MasterFQDN}:9000)",
      "defaultValue": "glusterfs:///localhost:8020",
      "displayType": "string",
      "isVisible": true,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "glusterfs_defaultFS_name",
      "displayName": "GlusterFS default fs name 2.x Hadoop",
      "description": "GlusterFS default filesystem name (glusterfs:///)",
      "defaultValue": "glusterfs:///localhost:8020",
      "displayType": "string",
      "isVisible": true,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "hadoop_heapsize",
      "displayName": "Hadoop maximum Java heap size",
      "description": "Maximum Java heap size for daemons such as Balancer (Java option -Xmx)",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "hdfs_log_dir_prefix",
      "displayName": "Hadoop Log Dir Prefix",
      "description": "The parent directory for Hadoop log files.  The HDFS log directory will be ${hadoop_log_dir_prefix} / ${hdfs_user} and the MapReduce log directory will be ${hadoop_log_dir_prefix} / ${mapred_user}.",
      "defaultValue": "/var/log/hadoop",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "hadoop_pid_dir_prefix",
      "displayName": "Hadoop PID Dir Prefix",
      "description": "The parent directory in which the PID files for Hadoop processes will be created.  The HDFS PID directory will be ${hadoop_pid_dir_prefix} / ${hdfs_user} and the MapReduce PID directory will be ${hadoop_pid_dir_prefix} / ${mapred_user}.",
      "defaultValue": "/var/run/hadoop",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "namenode_heapsize",
      "displayName": "Name Node Heap Size",
      "description": "Name Node Heap Size, default jvm memory setting",
      "defaultValue": "1024",
      "isReconfigurable": false,
      "displayType": "int",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "namenode_opt_newsize",
      "displayName": "NameNode new generation size",
      "description": "Default size of Java new generation for NameNode (Java option -XX:NewSize).  This also applies to the Secondary NameNode.",
      "defaultValue": "200",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "namenode_opt_maxnewsize",
      "displayName": "NameNode maximum new generation size",
      "description": "Maximum size of Java new generation for NameNode (Java option -XX:MaxnewSize).",
      "defaultValue": "200",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "namenode_opt_permsize",
      "displayName": "NameNode permanent generation size",
      "description": "Default size of Java permanent generation for NameNode (Java option -XX:PermSize).  This also applies to the Secondary NameNode.",
      "defaultValue": "128",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "namenode_opt_maxpermsize",
      "displayName": "NameNode maximum permanent generation size",
      "description": "Maximum size of Java permanent generation for NameNode (Java option -XX:MaxPermSize).",
      "defaultValue": "256",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "dtnode_heapsize",
      "displayName": "DataNode maximum Java heap size",
      "description": "Maximum Java heap size for DataNode (Java option -Xmx)",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "glusterfs_user",
      "displayName": "glusterfs user",
      "description": "glusterfs user",
      "defaultValue": "root",
      "displayType": "string",
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "namenode_host",
      "displayName": "NameNode Host",
      "description": "NameNode Host.",
      "defaultValue": "",
      "displayType": "string",
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "snamenode_host",
      "displayName": "Secondary NameNode Host",
      "description": "Secondary NameNode Host.",
      "defaultValue": "",
      "displayType": "string",
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
  /**********************************************HIVE***************************************/
    {
      "id": "puppet var",
      "name": "hivemetastore_host",
      "displayName": "Hive Metastore hosts",
      "value": "",
      "defaultValue": "",
      "description": "The hosts that have been assigned to run Hive Metastore",
      "displayType": "masterHosts",
      "isOverridable": false,
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 0
    },
    {
      "name": "hive_master_hosts",
      "value": "",
      "defaultValue": "",
      "isOverridable": false,
      "isRequiredByAgent": false,
      "isVisible": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml"
    },
    // for existing MySQL
    {
      "id": "puppet var",
      "name": "hive_existing_mysql_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "MySQL",
      "description": "Using an existing MySQL database for Hive Metastore",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": false,
      "isReconfigurable": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 1
    },
    // for existing PostgreSQL
    {
      "id": "puppet var",
      "name": "hive_existing_postgresql_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "PostgreSQL",
      "description": "Using an existing PostgreSQL database for Hive Metastore",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": false,
      "isReconfigurable": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 1
    },
    // for existing Oracle
    {
      "id": "puppet var",
      "name": "hive_existing_oracle_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "Oracle",
      "description": "Using an existing Oracle database for Hive Metastore",
      "displayType": "masterHost",
      "isVisible": false,
      "isOverridable": false,
      "isReconfigurable": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 1
    },
    // for new PostgreSQL
    {
      "id": "puppet var",
      "name": "hive_ambari_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "PostgreSQL",
      "description": "PostgreSQL will be installed by Ambari",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "hive_database",
      "displayName": "Hive Database",
      "value": "",
      "defaultValue": "New PostgreSQL Database",
      "options": [
        {
          displayName: 'New PostgreSQL Database',
          foreignKeys: ['hive_ambari_database', 'hive_ambari_host']
        },
        {
          displayName: 'Existing MySQL Database',
          foreignKeys: ['hive_existing_mysql_database', 'hive_existing_mysql_host']
        },
        {
          displayName: 'Existing PostgreSQL Database',
          foreignKeys: ['hive_existing_postgresql_database', 'hive_existing_postgresql_host']
        },
        {
          displayName: 'Existing Oracle Database',
          foreignKeys: ['hive_existing_oracle_database', 'hive_existing_oracle_host']
        }
      ],
      "description": "PostgreSQL will be installed by Ambari",
      "displayType": "radio button",
      "isReconfigurable": true,
      "radioName": "hive-database",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 2
    },
    {
      "id": "puppet var",
      "name": "hive_hostname",
      "displayName": "Database Host",
      "description": "Specify the host on which the database is hosted",
      "defaultValue": "",
      "isReconfigurable": true,
      "displayType": "host",
      "isOverridable": false,
      "isVisible": false,
      "isObserved": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "hive_existing_mysql_host",
      "displayName": "Database Host",
      "description": "Specify the host on which the existing database is hosted",
      "defaultValue": "",
      "displayType": "host",
      "isOverridable": false,
      "isVisible": false,
      "isObserved": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "hive_existing_postgresql_host",
      "displayName": "Database Host",
      "description": "Specify the host on which the existing database is hosted",
      "defaultValue": "",
      "displayType": "host",
      "isOverridable": false,
      "isVisible": false,
      "isObserved": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "hive_existing_oracle_host",
      "displayName": "Database Host",
      "description": "Specify the host on which the existing database is hosted",
      "defaultValue": "",
      "displayType": "host",
      "isOverridable": false,
      "isVisible": false,
      "isObserved": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "hive_ambari_host",
      "value": "",
      "defaultValue": "",
      "displayName": "Database Host",
      "description": "Host on which the database will be created by Ambari",
      "isReconfigurable": false,
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "hive_metastore_port",
      "displayName": "Hive metastore port",
      "description": "",
      "defaultValue": "9083",
      "isReconfigurable": false,
      "displayType": "int",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "Advanced hive-env"
    },
    {
      "id": "puppet var",
      "name": "hive_lib",
      "displayName": "Hive library",
      "description": "",
      "defaultValue": "/usr/lib/hive/lib/",
      "isReconfigurable": false,
      "displayType": "directory",
      "isVisible": false,
      "isRequiredByAgent": false, // Make this to true when we expose the property on ui by making "isVisible": true
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "Advanced hive-env"
    },
    {
      "id": "puppet var",
      "name": "hive_dbroot",
      "displayName": "Hive db directory",
      "description": "",
      "defaultValue": "/usr/lib/hive/lib",
      "isReconfigurable": false,
      "displayType": "directory",
      "isVisible": false,
      "isRequiredByAgent": false, // Make this to true when we expose the property on ui by making "isVisible": true
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "Advanced hive-env"
    },
    {
      "id": "puppet var",
      "name": "hive_log_dir",
      "displayName": "Hive Log Dir",
      "description": "Directory for Hive log files",
      "defaultValue": "/var/log/hive",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "Advanced hive-env"
    },
    {
      "id": "puppet var",
      "name": "hive_pid_dir",
      "displayName": "Hive PID Dir",
      "description": "Directory in which the PID files for Hive processes will be created",
      "defaultValue": "/var/run/hive",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "Advanced hive-env"
    },
  /**********************************************HIVE***************************************/
    {
      "id": "puppet var",
      "name": "webhcatserver_host",
      "displayName": "WebHCat Server host",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run WebHCat Server",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "HIVE",
      "filename": "webhcat-env.xml",
      "category": "WEBHCAT_SERVER"
    },
    {
      "id": "puppet var",
      "name": "hcat_log_dir",
      "displayName": "WebHCat Log Dir",
      "description": "Directory for WebHCat log files",
      "defaultValue": "/var/log/webhcat",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "Advanced webhcat-env"
    },
    {
      "id": "puppet var",
      "name": "hcat_pid_dir",
      "displayName": "WebHCat PID Dir",
      "description": "Directory in which the PID files for WebHCat processes will be created",
      "defaultValue": "/var/run/webhcat",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "Advanced webhcat-env"
    },
  /**********************************************OOZIE***************************************/
    {
      "id": "puppet var",
      "name": "oozieserver_host",
      "displayName": "Oozie Server host",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run Oozie Server",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 0
    },
    // for existing Oracle
    {
      "id": "puppet var",
      "name": "oozie_existing_oracle_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "Oracle",
      "description": "Using an existing Oracle database for Oozie Metastore",
      "displayType": "masterHost",
      "isVisible": false,
      "isObserved": true,
      "isReconfigurable": false,
      "isOverridable": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 1
    },
    // for existing PostgreSQL
    {
      "id": "puppet var",
      "name": "oozie_existing_postgresql_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "PostgreSQL",
      "description": "Using an existing PostgreSQL database for Oozie Metastore",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": false,
      "isReconfigurable": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 1
    },
    // for new MySQL
    {
      "id": "puppet var",
      "name": "oozie_ambari_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "MySQL",
      "description": "MySQL will be installed by Ambari",
      "displayType": "masterHost",
      "isVisible": false,
      "isOverridable": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 1
    },
    // for current derby
    {
      "id": "puppet var",
      "name": "oozie_derby_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "Derby",
      "description": "Using current Derby database for Oozie Metastore",
      "displayType": "masterHost",
      "isVisible": false,
      "isReconfigurable": false,
      "isOverridable": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 1
    },
    // for existing MySQL oozie
    {
      "id": "puppet var",
      "name": "oozie_existing_mysql_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "MySQL",
      "description": "Using an existing MySQL database for Oozie Metastore",
      "displayType": "masterHost",
      "isVisible": false,
      "isReconfigurable": false,
      "isOverridable": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "oozie_database",
      "displayName": "Oozie Database",
      "value": "",
      "defaultValue": "New Derby Database",
      "options": [
        {
          displayName: 'New Derby Database',
          foreignKeys: ['oozie_derby_database']
        },
        {
          displayName: 'Existing MySQL Database',
          foreignKeys: ['oozie_existing_mysql_database', 'oozie_existing_mysql_host']
        },
        {
          displayName: 'Existing PostgreSQL Database',
          foreignKeys: ['oozie_existing_postgresql_database', 'oozie_existing_postgresql_host']
        },
        {
          displayName: 'Existing Oracle Database',
          foreignKeys: ['oozie_existing_oracle_database', 'oozie_existing_oracle_host']
        }
      ],
      "description": "Current Derby Database will be installed by Ambari",
      "displayType": "radio button",
      "isReconfigurable": true,
      "isOverridable": false,
      "radioName": "oozie-database",
      "isVisible": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 2
    },
    {
      "id": "puppet var",
      "name": "oozie_data_dir",
      "displayName": "Oozie Data Dir",
      "description": "Data directory in which the Oozie DB exists",
      "defaultValue": "",
      "defaultDirectory": "/hadoop/oozie/data",
      "isReconfigurable": true,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isRequired": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 9
    },
    {
      "id": "puppet var",
      "name": "oozie_hostname",
      "defaultValue": "",
      "displayName": "Database Host",
      "description": "The host where the Oozie database is located",
      "isReconfigurable": true,
      "isOverridable": false,
      "displayType": "host",
      "isVisible": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "oozie_existing_mysql_host",
      "displayName": "Database Host",
      "description": "Specify the host on which the existing database is hosted",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "host",
      "isVisible": false,
      "isObserved": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "oozie_existing_postgresql_host",
      "displayName": "Database Host",
      "description": "Specify the host on which the existing database is hosted",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "host",
      "isVisible": false,
      "isObserved": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "oozie_existing_oracle_host",
      "displayName": "Database Host",
      "description": "Specify the host on which the existing database is hosted",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "host",
      "isVisible": false,
      "isObserved": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "oozie_ambari_host",
      "value": "",
      "defaultValue": "",
      "displayName": "Database Host",
      "description": "Host on which the database will be created by Ambari",
      "isReconfigurable": false,
      "isOverridable": false,
      "displayType": "masterHost",
      "isVisible": false,
      "isRequiredByAgent": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER"
    },
    {
      "id": "puppet var",
      "name": "oozie_log_dir",
      "displayName": "Oozie Log Dir",
      "description": "Directory for oozie logs",
      "defaultValue": "/var/log/oozie",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "Advanced oozie-env"
    },
    {
      "id": "puppet var",
      "name": "oozie_pid_dir",
      "displayName": "Oozie PID Dir",
      "description": "Directory in which the pid files for oozie processes will be created",
      "defaultValue": "/var/run/oozie",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "Advanced oozie-env"
    },
    {
      "id": "puppet var",
      "name": "oozie_admin_port",
      "displayName": "Oozie Server Admin Port",
      "isReconfigurable": true,
      "displayType": "int",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "Advanced oozie-env"
    },
  /**********************************************ZOOKEEPER***************************************/
    {
      "id": "puppet var",
      "name": "zookeeperserver_hosts",
      "displayName": "ZooKeeper Server hosts",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run ZooKeeper Server",
      "displayType": "masterHosts",
      "isVisible": true,
      "isRequiredByAgent": false,
      "isOverridable": false,
      "isRequired": false,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZOOKEEPER_SERVER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "zk_data_dir",
      "displayName": "ZooKeeper directory",
      "description": "Data directory for ZooKeeper",
      "defaultValue": "",
      "defaultDirectory": "/hadoop/zookeeper",
      "isReconfigurable": true,
      "displayType": "directory",
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZOOKEEPER_SERVER",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "tickTime",
      "displayName": "Length of single Tick",
      "description": "The length of a single tick in milliseconds, which is the basic time unit used by ZooKeeper",
      "defaultValue": "2000",
      "displayType": "int",
      "unit": "ms",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZOOKEEPER_SERVER",
      "index": 2
    },
    {
      "id": "puppet var",
      "name": "initLimit",
      "displayName": "Ticks to allow for sync at Init",
      "description": "Amount of time, in ticks to allow followers to connect and sync to a leader",
      "defaultValue": "10",
      "displayType": "int",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZOOKEEPER_SERVER",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "syncLimit",
      "displayName": "Ticks to allow for sync at Runtime",
      "description": "Amount of time, in ticks to allow followers to connect",
      "defaultValue": "5",
      "displayType": "int",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZOOKEEPER_SERVER",
      "index": 4
    },
    {
      "id": "puppet var",
      "name": "clientPort",
      "displayName": "Port for running ZK Server",
      "description": "Port for running ZooKeeper server",
      "defaultValue": "2181",
      "displayType": "int",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZOOKEEPER_SERVER",
      "index": 5
    },
    {
      "id": "puppet var",
      "name": "zk_log_dir",
      "displayName": "ZooKeeper Log Dir",
      "description": "Directory for ZooKeeper log files",
      "defaultValue": "/var/log/zookeeper",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "Advanced zookeeper-env",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "zk_pid_dir",
      "displayName": "ZooKeeper PID Dir",
      "description": "Directory in which the pid files for zookeeper processes will be created",
      "defaultValue": "/var/run/zookeeper",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "Advanced zookeeper-env",
      "index": 1
    },
  /**********************************************GANGLIA***************************************/
    {
      "id": "puppet var",
      "name": "ganglia_conf_dir",
      "displayName": "Ganglia conf directory",
      "description": "",
      "defaultValue": "/etc/ganglia/hdp",
      "isReconfigurable": false,
      "displayType": "directory",
      "isVisible": false,
      "isRequiredByAgent": false,
      "serviceName": "GANGLIA",
      "filename": "ganglia-env.xml",
      "category": "Advanced ganglia-env"
    },
  /**********************************************FALCON***************************************/
    {
      "id": "puppet var",
      "name": "falconserver_host",
      "displayName": "Falcon Server",
      "description": "The host that has been assigned to run Falcon Server",
      "defaultValue": "falcon",
      "displayType": "masterHost",
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "FALCON_SERVER"
    },
    {
      "id": "puppet var",
      "name": "falcon_port",
      "displayName": "Falcon server port",
      "description": "Port the Falcon Server listens on",
      "defaultValue": "15000",
      "isReconfigurable": true,
      "displayType": "int",
      "isVisible": true,
      "isRequiredByAgent": true,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "FALCON_SERVER"
    },
    {
      "id": "puppet var",
      "name": "falcon_local_dir",
      "displayName": "Falcon data directory",
      "description": "Directory where Falcon data, such as activemq data, is stored",
      "defaultValue": "/hadoop/falcon",
      "isReconfigurable": true,
      "displayType": "directory",
      "isVisible": true,
      "isRequiredByAgent": true,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "FALCON_SERVER"
    },
    {
      "id": "puppet var",
      "name": "falcon_store_uri",
      "displayName": "Falcon store URI",
      "description": "Directory where entity definitions are stored",
      "defaultValue": "file:///hadoop/falcon/store",
      "isReconfigurable": true,
      "displayType": "string",
      "isVisible": true,
      "isRequiredByAgent": true,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "FALCON_SERVER"
    },
    {
      "id": "puppet var",
      "name": "falcon_log_dir",
      "displayName": "Falcon Log Dir",
      "description": "Directory for Falcon logs",
      "defaultValue": "/var/log/falcon",
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": true,
      "isReconfigurable": false,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "Advanced falcon-env"
    },
    {
      "id": "puppet var",
      "name": "falcon_pid_dir",
      "displayName": "Falcon PID Dir",
      "description": "Directory in which the pid files for Falcon processes will be created",
      "defaultValue": "/var/run/falcon",
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": true,
      "isReconfigurable": false,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "Advanced falcon-env"
    },
    {
      "id": "puppet var",
      "name": "falcon.embeddedmq",
      "displayName": "falcon.embeddedmq",
      "description": "Whether embeddedmq is enabled or not.",
      "defaultValue": "true",
      "displayType": "string",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": true,
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "Advanced falcon-env"
    },
    {
      "id": "puppet var",
      "name": "falcon.embeddedmq.data",
      "displayName": "falcon.embeddedmq.data",
      "description": "Directory in which embeddedmq data is stored.",
      "defaultValue": "/hadoop/falcon/embeddedmq/data",
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": true,
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "Advanced falcon-env"
    },
    {
      "id": "puppet var",
      "name": "falcon.emeddedmq.port",
      "displayName": "falcon.emeddedmq.port",
      "description": "Port that embeddedmq will listen on.",
      "defaultValue": "61616",
      "displayType": "string",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": true,
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "Advanced falcon-env"
    },
  /**********************************************STORM***************************************/
    {
      "id": "puppet var",
      "name": "storm_log_dir",
      "displayName": "storm_log_dir",
      "description": "Storm log directory",
      "defaultValue": "/var/log/storm",
      "displayType": "directory",
      "isReconfigurable": true,
      "isVisible": true,
      "isRequiredByAgent": true,
      "serviceName": "STORM",
      "filename": "storm-env.xml",
      "category": "Advanced storm-env"
    },
    {
      "id": "puppet var",
      "name": "storm_pid_dir",
      "displayName": "storm_pid_dir",
      "description": "Storm PID directory",
      "defaultValue": "/var/run/storm",
      "displayType": "directory",
      "isReconfigurable": true,
      "isVisible": true,
      "isRequiredByAgent": true,
      "serviceName": "STORM",
      "filename": "storm-env.xml",
      "category": "Advanced storm-env"
    },
    {
      "id": "puppet var",
      "name": "stormuiserver_host",
      "displayName": "Storm UI Server host",
      "description": "The host that has been assigned to run Storm UI Server",
      "defaultValue": "",
      "displayType": "masterHost",
      "isReconfigurable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "STORM",
      "filename": "storm-env.xml",
      "category": "STORM_UI_SERVER"
    },
    {
      "id": "puppet var",
      "name": "drpcserver_host",
      "displayName": "DRPC Server host",
      "description": "The host that has been assigned to run DRPC Server",
      "defaultValue": "",
      "displayType": "masterHost",
      "isReconfigurable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "STORM",
      "filename": "storm-env.xml",
      "category": "DRPC_SERVER"
    },
    {
      "id": "puppet var",
      "name": "supervisor_hosts",
      "displayName": "Supervisor hosts",
      "description": "The host that has been assigned to run Supervisor",
      "defaultValue": "",
      "displayType": "slaveHosts",
      "isReconfigurable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "STORM",
      "filename": "storm-env.xml",
      "isOverridable": false,
      "category": "SUPERVISOR"
    },
    {
      "id": "puppet var",
      "name": "storm_rest_api_host",
      "displayName": "Storm REST API host",
      "description": "The host that has been assigned to run Storm REST API Server",
      "defaultValue": "",
      "displayType": "masterHost",
      "isReconfigurable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "STORM",
      "filename": "storm-env.xml",
      "isOverridable": false,
      "category": "STORM_REST_API"
    },
  /**********************************************MISC***************************************/
    {
      "id": "puppet var",
      "name": "hbase_conf_dir",
      "displayName": "HBase conf dir",
      "description": "",
      "defaultValue": "/etc/hbase",
      "isRequired": false,
      "displayType": "directory",
      "isVisible": false,
      "isRequiredByAgent": false,
      "serviceName": "MISC",
      "filename": "hbase-env.xml",
      "category": "General",
      "belongsToService": []
    },
    {
      "id": "puppet var",
      "name": "proxyuser_group",
      "displayName": "Proxy group for Hive, WebHCat, Oozie and Falcon",
      "description": "",
      "defaultValue": "users",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "filename": "hadoop-env.xml",
      "serviceName": "MISC",
      "category": "Users and Groups",
      "belongsToService": ["HIVE", "OOZIE", "FALCON"],
      "index": 18
    },
    {
      "id": "puppet var",
      "name": "ganglia_runtime_dir",
      "displayName": "Ganglia runtime directory",
      "description": "",
      "defaultValue": "/var/run/ganglia/hdp",
      "isReconfigurable": false,
      "displayType": "directory",
      "isVisible": false,
      "serviceName": "MISC",
      "filename": "ganglia-env.xml",
      "category": "General",
      "belongsToService": []
    },
    {
      "id": "puppet var",
      "name": "hdfs_user",
      "displayName": "HDFS User",
      "description": "User to run HDFS as",
      "defaultValue": "hdfs",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "hadoop-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["HDFS"],
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "mapred_user",
      "displayName": "MapReduce User",
      "description": "User to run MapReduce as",
      "defaultValue": "mapred",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "mapred-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["MAPREDUCE2"],
      "index": 2
    },
    {
      "id": "puppet var",
      "name": "yarn_user",
      "displayName": "YARN User",
      "description": "User to run YARN as",
      "defaultValue": "yarn",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "yarn-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["YARN"],
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "hbase_user",
      "displayName": "HBase User",
      "description": "User to run HBase as",
      "defaultValue": "hbase",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "hbase-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["HBASE"],
      "index": 4
    },
    {
      "id": "puppet var",
      "name": "hive_user",
      "displayName": "Hive User",
      "description": "User to run Hive as",
      "defaultValue": "hive",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "hive-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["HIVE"],
      "index": 5
    },
    {
      "id": "puppet var",
      "name": "hcat_user",
      "displayName": "HCat User",
      "description": "User to run HCatalog as",
      "defaultValue": "hcat",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "hive-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["HIVE"],
      "index": 6
    },
    {
      "id": "puppet var",
      "name": "webhcat_user",
      "displayName": "WebHCat User",
      "description": "User to run WebHCat as",
      "defaultValue": "hcat",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "hive-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["HIVE"],
      "index": 7
    },
    {
      "id": "puppet var",
      "name": "oozie_user",
      "displayName": "Oozie User",
      "description": "User to run Oozie as",
      "defaultValue": "oozie",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "oozie-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["OOZIE"],
      "index": 8
    },
    {
      "id": "puppet var",
      "name": "falcon_user",
      "displayName": "Falcon User",
      "description": "User to run Falcon as",
      "defaultValue": "falcon",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "falcon-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["FALCON"],
      "index": 10
    },
    {
      "id": "puppet var",
      "name": "storm_user",
      "displayName": "Storm User",
      "description": "User to run Storm as",
      "defaultValue": "storm",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "storm-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["STORM"],
      "index": 9
    },
    {
      "id": "puppet var",
      "name": "zk_user",
      "displayName": "ZooKeeper User",
      "description": "User to run ZooKeeper as",
      "defaultValue": "zookeeper",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "zookeeper-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["ZOOKEEPER"],
      "index": 11
    },
    {
      "id": "puppet var",
      "name": "flume_user",
      "displayName": "Flume User",
      "description": "User to run Flume as",
      "defaultValue": "flume",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "flume-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["FLUME"],
      "index": 12
    },
    {
      "id": "puppet var",
      "name": "gmetad_user",
      "displayName": "Ganglia User",
      "description": "The user used to run Ganglia",
      "defaultValue": "nobody",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "ganglia-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["GANGLIA"],
      "index": 13
    },
    {
      "id": "puppet var",
      "name": "gmond_user",
      "displayName": "Gmond User",
      "description": "The user used to run gmond for Ganglia",
      "defaultValue": "nobody",
      "isReconfigurable": false,
      "displayType": "advanced",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "MISC",
      "filename": "ganglia-env.xml",
      "category": "Users and Groups",
      "belongsToService": []
    },
    {
      "id": "puppet var",
      "name": "tez_user",
      "displayName": "Tez User",
      "description": "User to run Tez as",
      "defaultValue": "tez",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "tez-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["TEZ"],
      "index": 15
    },
    {
      "id": "puppet var",
      "name": "sqoop_user",
      "displayName": "Sqoop User",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "sqoop-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["SQOOP"],
      "index": 17
    },
    {
      "id": "puppet var",
      "name": "rrdcached_base_dir",
      "displayName": "Ganglia rrdcached base directory",
      "description": "Default directory for saving the rrd files on ganglia server",
      "defaultValue": "/var/lib/ganglia/rrds",
      "displayType": "directory",
      "isReconfigurable": true,
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "GANGLIA",
      "filename": "ganglia-env.xml",
      "category": "General",
      "belongsToService": ["GANGLIA"]
    },
    {
      "id": "puppet var",
      "name": "ignore_groupsusers_create",
      "displayName": "Skip group modifications during install",
      "displayType": "checkbox",
      "isReconfigurable": true,
      "isOverridable": false,
      "isVisible": true,
      "filename": "cluster-env.xml",
      "category": "Users and Groups"
    }
  ]
};

});

require.register("data/HDP2.2/hive_properties", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

var hivePropsByCategory = {

  'Performance': [
    'hive.cbo.enable',
    'hive.optimize.reducededuplication.min.reducer',
    'hive.optimize.reducededuplication',
    'hive.orc.splits.include.file.footer',
    'hive.merge.mapfiles',
    'hive.merge.mapredfiles',
    'hive.merge.tezfiles',
    'hive.merge.smallfiles.avgsize',
    'hive.merge.size.per.task',
    'hive.merge.orcfile.stripe.level',
    'hive.auto.convert.join',
    'hive.auto.convert.join.noconditionaltask',
    'hive.auto.convert.join.noconditionaltask.size',
    'hive.optimize.bucketmapjoin.sortedmerge',
    'hive.tez.smb.number.waves',
    'hive.map.aggr.hash.percentmemory',
    'hive.map.aggr',
    'hive.optimize.sort.dynamic.partition',
    'hive.stats.autogather',
    'hive.stats.fetch.column.stats',
    'hive.vectorized.execution.enabled',
    'hive.vectorized.execution.reduce.enabled',
    'hive.vectorized.groupby.checkinterval',
    'hive.vectorized.groupby.flush.percent',
    'hive.compute.query.using.stats',
    'hive.limit.pushdown.memory.usage',
    'hive.optimize.index.filter',
    'hive.exec.reducers.bytes.per.reducer',
    'hive.smbjoin.cache.rows',
    'hive.exec.orc.default.stripe.size',
    'hive.fetch.task.conversion',
    'hive.fetch.task.conversion.threshold',
    'hive.fetch.task.aggr',
    'hive.compute.query.using.stats',
    'hive.tez.auto.reducer.parallelism',
    'hive.tez.max.partition.factor',
    'hive.tez.min.partition.factor',
    'hive.tez.dynamic.partition.pruning',
    'hive.tez.dynamic.partition.pruning.max.event.size',
    'hive.tez.dynamic.partition.pruning.max.data.size'
  ],

  'General': [
    'hive.exec.pre.hooks',
    'hive.exec.post.hooks',
    'hive.exec.failure.hooks',
    'hive.execution.engine',
    'hive.exec.dynamic.partition',
    'hive.exec.dynamic.partition.mode',
    'hive.exec.max.dynamic.partitions',
    'hive.exec.max.dynamic.partitions.pernode',
    'hive.exec.max.created.files',
    'hive.enforce.bucketing',
    'datanucleus.cache.level2.type',
    'hive.metastore.uris',
    'hive.metastore.warehouse.dir',
    'hive.exec.parallel.thread.number',
    'hive.security.authorization.enabled',
    'hive.security.authorization.manager',
    'hive.security.metastore.authenticator.manager',
    'hive.security.metastore.authorization.manager',
    'hive.server2.authentication',
    'hive.server2.enable.doAs',
    'hive.server2.tez.default.queues',
    'hive.server2.tez.initialize.default.sessions',
    'hive.server2.tez.sessions.per.default.queue',
    'hive.server2.thrift.http.path',
    'hive.server2.thrift.http.port',
    'hive.server2.thrift.max.worker.threads',
    'hive.server2.thrift.port',
    'hive.server2.thrift.sasl.qop',
    'hive.server2.transport.mode',
    'hive.server2.use.SSL',
    'hive.tez.container.size',
    'hive.tez.java.opts',
    'hive.tez.log.level',
    'hive.txn.manager',
    'hive.txn.timeout',
    'hive.txn.max.open.batch',
    'hive.compactor.initiator.on',
    'hive.compactor.worker.threads',
    'hive.compactor.worker.timeout',
    'hive.compactor.check.interval',
    'hive.compactor.delta.num.threshold',
    'hive.compactor.delta.pct.threshold'
  ]
};

var hiveProps = [];

for (var category in hivePropsByCategory) {
  hiveProps = hiveProps.concat(App.config.generateConfigPropertiesByName(hivePropsByCategory[category],
    { category: category, serviceName: 'HIVE', filename: 'hive-site.xml'}));
}

module.exports = hiveProps;


});

require.register("data/HDP2.2/site_properties", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

var hdp22SepcificProperties = [
  require('data/HDP2.2/yarn_properties'),
  require('data/HDP2.2/tez_properties'),
  require('data/HDP2.2/hive_properties')
].reduce(function(p, c) { return c.concat(p); });

var hdp2properties = require('data/HDP2/site_properties').configProperties;
var excludedConfigs = [
  'storm.thrift.transport', //In HDP2.2 storm.thrift.transport property is computed on server
  'storm_rest_api_host',
  'tez.am.container.session.delay-allocation-millis',
  'tez.am.grouping.max-size',
  'tez.am.grouping.min-size',
  'tez.am.grouping.split-waves',
  'tez.am.java.opts',
  'tez.runtime.intermediate-input.compress.codec',
  'tez.runtime.intermediate-input.is-compressed',
  'tez.runtime.intermediate-output.compress.codec',
  'tez.runtime.intermediate-output.should-compress',
  'dfs.datanode.data.dir',
  'hive.heapsize'
];
var hdp22properties = hdp2properties.filter(function (item) {
  return !excludedConfigs.contains(item.name);
});

hdp22properties.push(
  {
    "id": "site property",
    "name": "hive.zookeeper.quorum",
    "displayName": "hive.zookeeper.quorum",
    "defaultValue": "localhost:2181",
    "displayType": "multiLine",
    "isRequired": false,
    "isVisible": true,
    "serviceName": "HIVE",
    "filename": "hive-site.xml",
    "category": "Advanced hive-site"
  },
  {
    "id": "site property",
    "name": "hadoop.registry.rm.enabled",
    "displayName": "hadoop.registry.rm.enabled",
    "defaultValue": "false",
    "displayType": "checkbox",
    "isVisible": true,
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "Advanced yarn-site"
  },
  {
    "id": "site property",
    "name": "hadoop.registry.zk.quorum",
    "displayName": "hadoop.registry.zk.quorum",
    "defaultValue": "localhost:2181",
    "isVisible": true,
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "Advanced yarn-site"
  },
  {
    "id": "site property",
    "name": "dfs.datanode.data.dir",
    "displayName": "DataNode directories",
    "defaultDirectory": "/hadoop/hdfs/data",
    "displayType": "datanodedirs",
    "category": "DATANODE",
    "serviceName": "HDFS",
    "filename": "hdfs-site.xml",
    "index": 1
  },
  {
    "id": "site property",
    "name": "*.falcon.graph.blueprints.graph",
    "displayName": "*.falcon.graph.blueprints.graph",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.falcon.graph.storage.backend",
    "displayName": "*.falcon.graph.storage.backend",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  /*********RANGER FOR HDFS************/
  {
    "id": "site property",
    "name": "XAAUDIT.HDFS.IS_ENABLED",
    "displayName": "Audit to HDFS",
    "displayType": "checkbox",
    "dependentConfigPattern": "^XAAUDIT.HDFS",
    "filename": "ranger-hdfs-plugin-properties.xml",
    "category": "Advanced ranger-hdfs-plugin-properties",
    "serviceName": "HDFS"
  },
  {
    "id": "site property",
    "name": "XAAUDIT.DB.IS_ENABLED",
    "displayName": "Audit to DB",
    "displayType": "checkbox",
    "filename": "ranger-hdfs-plugin-properties.xml",
    "category": "Advanced ranger-hdfs-plugin-properties",
    "serviceName": "HDFS"
  },
  {
    "id": "site property",
    "name": "ranger-hdfs-plugin-enabled",
    "displayType": "checkbox",
    "displayName": "Enable Ranger for HDFS",
    "isOverridable": false,
    "filename": "ranger-hdfs-plugin-properties.xml",
    "category": "Advanced ranger-hdfs-plugin-properties",
    "serviceName": "HDFS",
    "index": 1
  },
  {
    "id": "site property",
    "name": "policy_user",
    "value": "ambari-qa",
    "defaultValue": "ambari-qa",
    "displayName": "policy User for HDFS",
    "filename": "ranger-hdfs-plugin-properties.xml",
    "category": "Advanced ranger-hdfs-plugin-properties",
    "serviceName": "HDFS"
  },
  {
    "id": "site property",
    "name": "REPOSITORY_CONFIG_PASSWORD",
    "displayName": "Ranger repository config password",
    "filename": "ranger-hdfs-plugin-properties.xml",
    "category": "Advanced ranger-hdfs-plugin-properties",
    "serviceName": "HDFS"
  },
  {
    "id": "site property",
    "name": "REPOSITORY_CONFIG_USERNAME",
    "displayName": "Ranger repository config user",
    "filename": "ranger-hdfs-plugin-properties.xml",
    "category": "Advanced ranger-hdfs-plugin-properties",
    "serviceName": "HDFS"
  },
  /*********RANGER FOR HIVE************/
  {
    "id": "site property",
    "name": "XAAUDIT.HDFS.IS_ENABLED",
    "displayName": "Audit to HDFS",
    "displayType": "checkbox",
    "dependentConfigPattern": "^XAAUDIT.HDFS",
    "filename": "ranger-hive-plugin-properties.xml",
    "category": "Advanced ranger-hive-plugin-properties",
    "serviceName": "HIVE"
  },
  {
    "id": "site property",
    "name": "XAAUDIT.DB.IS_ENABLED",
    "displayName": "Audit to DB",
    "displayType": "checkbox",
    "filename": "ranger-hive-plugin-properties.xml",
    "category": "Advanced ranger-hive-plugin-properties",
    "serviceName": "HIVE"
  },
  {
    "id": "site property",
    "name": "ranger-hive-plugin-enabled",
    "displayType": "checkbox",
    "displayName": "Enable Ranger for HIVE",
    "isOverridable": false,
    "filename": "ranger-hive-plugin-properties.xml",
    "category": "Advanced ranger-hive-plugin-properties",
    "serviceName": "HIVE",
    "index": 1
  },
  {
    "id": "site property",
    "name": "policy_user",
    "value": "ambari-qa",
    "defaultValue": "ambari-qa",
    "displayName": "policy User for HIVE",
    "filename": "ranger-hive-plugin-properties.xml",
    "category": "Advanced ranger-hive-plugin-properties",
    "serviceName": "HIVE"
  },
  {
    "id": "site property",
    "name": "REPOSITORY_CONFIG_PASSWORD",
    "displayName": "Ranger repository config password",
    "filename": "ranger-hive-plugin-properties.xml",
    "category": "Advanced ranger-hive-plugin-properties",
    "serviceName": "HIVE"
  },
  {
    "id": "site property",
    "name": "REPOSITORY_CONFIG_USERNAME",
    "displayName": "Ranger repository config user",
    "filename": "ranger-hive-plugin-properties.xml",
    "category": "Advanced ranger-hive-plugin-properties",
    "serviceName": "HIVE"
  },
  {
    "id": "site property",
    "name": "UPDATE_XAPOLICIES_ON_GRANT_REVOKE",
    "defaultValue": true,
    "displayType": "checkbox",
    "displayName": "Should Hive GRANT/REVOKE update XA policies?",
    "filename": "ranger-hive-plugin-properties.xml",
    "category": "Advanced ranger-hive-plugin-properties",
    "serviceName": "HIVE"
  },
  /*********RANGER FOR HBASE************/
  {
    "id": "site property",
    "name": "XAAUDIT.HDFS.IS_ENABLED",
    "displayName": "Audit to HDFS",
    "displayType": "checkbox",
    "dependentConfigPattern": "^XAAUDIT.HDFS",
    "filename": "ranger-hbase-plugin-properties.xml",
    "category": "Advanced ranger-hbase-plugin-properties",
    "serviceName": "HBASE"
  },
  {
    "id": "site property",
    "name": "XAAUDIT.DB.IS_ENABLED",
    "displayName": "Audit to DB",
    "displayType": "checkbox",
    "filename": "ranger-hbase-plugin-properties.xml",
    "category": "Advanced ranger-hbase-plugin-properties",
    "serviceName": "HBASE"
  },
  {
    "id": "site property",
    "name": "ranger-hbase-plugin-enabled",
    "displayType": "checkbox",
    "displayName": "Enable Ranger for HBASE",
    "isOverridable": false,
    "filename": "ranger-hbase-plugin-properties.xml",
    "category": "Advanced ranger-hbase-plugin-properties",
    "serviceName": "HBASE",
    "index": 1
  },
  {
    "id": "site property",
    "name": "policy_user",
    "value": "ambari-qa",
    "defaultValue": "ambari-qa",
    "displayName": "policy User for HBASE",
    "filename": "ranger-hbase-plugin-properties.xml",
    "category": "Advanced ranger-hbase-plugin-properties",
    "serviceName": "HBASE"
  },
  {
    "id": "site property",
    "name": "REPOSITORY_CONFIG_PASSWORD",
    "displayName": "Ranger repository config password",
    "filename": "ranger-hbase-plugin-properties.xml",
    "category": "Advanced ranger-hbase-plugin-properties",
    "serviceName": "HBASE"
  },
  {
    "id": "site property",
    "name": "REPOSITORY_CONFIG_USERNAME",
    "displayName": "Ranger repository config user",
    "filename": "ranger-hbase-plugin-properties.xml",
    "category": "Advanced ranger-hbase-plugin-properties",
    "serviceName": "HBASE"
  },
  {
    "id": "site property",
    "name": "UPDATE_XAPOLICIES_ON_GRANT_REVOKE",
    "defaultValue": true,
    "displayName": "Should HBase GRANT/REVOKE update XA policies?",
    "displayType": "checkbox",
    "filename": "ranger-hbase-plugin-properties.xml",
    "category": "Advanced ranger-hbase-plugin-properties",
    "serviceName": "HBASE"
  },
  /*********RANGER FOR STORM************/
  {
    "id": "site property",
    "name": "XAAUDIT.HDFS.IS_ENABLED",
    "displayName": "Audit to HDFS",
    "displayType": "checkbox",
    "dependentConfigPattern": "^XAAUDIT.HDFS",
    "filename": "ranger-storm-plugin-properties.xml",
    "category": "Advanced ranger-storm-plugin-properties",
    "serviceName": "STORM"
  },
  {
    "id": "site property",
    "name": "XAAUDIT.DB.IS_ENABLED",
    "displayName": "Audit to DB",
    "displayType": "checkbox",
    "filename": "ranger-storm-plugin-properties.xml",
    "category": "Advanced ranger-storm-plugin-properties",
    "serviceName": "STORM"
  },
  {
    "id": "site property",
    "name": "ranger-storm-plugin-enabled",
    "displayType": "checkbox",
    "displayName": "Enable Ranger for STORM",
    "isOverridable": false,
    "filename": "ranger-storm-plugin-properties.xml",
    "category": "Advanced ranger-storm-plugin-properties",
    "serviceName": "STORM",
    "index": 1
  },
  {
    "id": "site property",
    "name": "policy_user",
    "value": "ambari-qa",
    "defaultValue": "ambari-qa",
    "displayName": "policy User for STORM",
    "filename": "ranger-storm-plugin-properties.xml",
    "category": "Advanced ranger-storm-plugin-properties",
    "serviceName": "STORM"
  },
  {
    "id": "site property",
    "name": "REPOSITORY_CONFIG_PASSWORD",
    "displayName": "Ranger repository config password",
    "filename": "ranger-storm-plugin-properties.xml",
    "category": "Advanced ranger-storm-plugin-properties",
    "serviceName": "STORM"
  },
  {
    "id": "site property",
    "name": "REPOSITORY_CONFIG_USERNAME",
    "displayName": "Ranger repository config user",
    "filename": "ranger-storm-plugin-properties.xml",
    "category": "Advanced ranger-storm-plugin-properties",
    "serviceName": "STORM"
  },
  /*********RANGER FOR KNOX************/
  {
    "id": "site property",
    "name": "XAAUDIT.HDFS.IS_ENABLED",
    "displayName": "Audit to HDFS",
    "displayType": "checkbox",
    "dependentConfigPattern": "^XAAUDIT.HDFS",
    "filename": "ranger-knox-plugin-properties.xml",
    "category": "Advanced ranger-knox-plugin-properties",
    "serviceName": "KNOX"
  },
  {
    "id": "site property",
    "name": "XAAUDIT.DB.IS_ENABLED",
    "displayName": "Audit to DB",
    "displayType": "checkbox",
    "filename": "ranger-knox-plugin-properties.xml",
    "category": "Advanced ranger-knox-plugin-properties",
    "serviceName": "KNOX"
  },
  {
    "id": "site property",
    "name": "ranger-knox-plugin-enabled",
    "displayType": "checkbox",
    "displayName": "Enable Ranger for KNOX",
    "isOverridable": false,
    "filename": "ranger-knox-plugin-properties.xml",
    "category": "Advanced ranger-knox-plugin-properties",
    "serviceName": "KNOX",
    "index": 1
  },
  {
    "id": "site property",
    "name": "policy_user",
    "value": "ambari-qa",
    "defaultValue": "ambari-qa",
    "displayName": "policy User for KNOX",
    "filename": "ranger-knox-plugin-properties.xml",
    "category": "Advanced ranger-knox-plugin-properties",
    "serviceName": "KNOX"
  },
  {
    "id": "site property",
    "name": "REPOSITORY_CONFIG_PASSWORD",
    "displayName": "Ranger repository config password",
    "filename": "ranger-knox-plugin-properties.xml",
    "category": "Advanced ranger-knox-plugin-properties",
    "serviceName": "KNOX"
  },
  {
    "id": "site property",
    "name": "REPOSITORY_CONFIG_USERNAME",
    "displayName": "Ranger repository config user",
    "filename": "ranger-knox-plugin-properties.xml",
    "category": "Advanced ranger-knox-plugin-properties",
    "serviceName": "KNOX"
  },
  /**********************************************SPARK***************************************/
  {
    "id": "site property",
    "name": "spark.driver.extraJavaOptions",
    "displayName": "spark.driver.extraJavaOptions",
    "defaultValue": "",
    "isRequired": false,
    "category": "Advanced spark-defaults",
    "serviceName": "SPARK",
    "filename": "spark-defaults.xml"
  },
  {
    "id": "site property",
    "name": "spark.yarn.am.extraJavaOptions",
    "displayName": "spark.yarn.am.extraJavaOptions",
    "defaultValue": "",
    "isRequired": false,
    "category": "Advanced spark-defaults",
    "serviceName": "SPARK",
    "filename": "spark-defaults.xml"
  },
  /**********************************************RANGER***************************************/
  {
    "id": "site property",
    "name": "ranger_admin_password",
    "displayName": "Ranger Admin user's password for Ambari",
    "defaultValue": "ambari",
    "isReconfigurable": true,
    "displayType": "password",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "ranger-env.xml",
    "category": "AdminSettings"
  },
  {
    "id": "site property",
    "name": "SQL_CONNECTOR_JAR",
    "displayName": "Location of Sql Connector Jar",
    "defaultValue": "/usr/share/java/mysql-connector-java.jar",
    "isReconfigurable": true,
    "displayType": "",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "AdminSettings"
  },
  {
    "id": "site property",
    "name": "DB_FLAVOR",
    "displayName": "DB FLAVOR",
    "value": "MYSQL",
    "defaultValue": "MYSQL",
    "isReconfigurable": true,
    "options": [
      {
        displayName: 'MYSQL'
      },
      {
        displayName: 'ORACLE'
      }
    ],
    "displayType": "radio button",
    "radioName": "RANGER DB_FLAVOR",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "DBSettings"
  },
  {
    "id": "site property",
    "name": "SQL_COMMAND_INVOKER",
    "displayName": "SQL Command Invoker",
    "defaultValue": "mysql",
    "isReconfigurable": true,
    "displayType": "",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "DBSettings"
  },
  {
    "id": "site property",
    "name": "db_host",
    "displayName": "Ranger DB host",
    "defaultValue": "",
    "isReconfigurable": true,
    "displayType": "",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "DBSettings"
  },
  {
    "id": "site property",
    "name": "db_root_user",
    "displayName": "Ranger DB root user",
    "defaultValue": "",
    "isReconfigurable": true,
    "displayType": "",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "DBSettings"
  },
  {
    "id": "site property",
    "name": "db_root_password",
    "displayName": "Ranger DB root password",
    "defaultValue": "",
    "isReconfigurable": true,
    "displayType": "password",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "DBSettings"
  },
  {
    "id": "site property",
    "name": "db_name",
    "displayName": "Ranger DB name",
    "defaultValue": "",
    "isReconfigurable": true,
    "displayType": "",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "DBSettings"
  },

  {
    "id": "site property",
    "name": "db_user",
    "displayName": "Ranger DB username",
    "defaultValue": "",
    "isReconfigurable": true,
    "displayType": "",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "DBSettings"
  },
  {
    "id": "site property",
    "name": "db_password",
    "displayName": "Ranger DB password",
    "defaultValue": "",
    "isReconfigurable": true,
    "displayType": "password",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "DBSettings"
  },
  {
    "id": "site property",
    "name": "audit_db_name",
    "displayName": "Ranger Audit DB name",
    "defaultValue": "",
    "isReconfigurable": true,
    "displayType": "",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "DBSettings"
  },
  {
    "id": "site property",
    "name": "audit_db_user",
    "displayName": "Ranger Audit DB username",
    "defaultValue": "",
    "isReconfigurable": true,
    "displayType": "",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "DBSettings"
  },
  {
    "id": "site property",
    "name": "audit_db_password",
    "displayName": "Ranger Audit DB password",
    "defaultValue": "",
    "isReconfigurable": true,
    "displayType": "password",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "DBSettings"
  },
  {
    "id": "puppet var",
    "name": "policymgr_external_url",
    "displayName": "External URL",
    "defaultValue": "http://localhost:6080",
    "isReconfigurable": true,
    "displayType": "",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "RangerSettings"
  },
  {
    "id": "puppet var",
    "name": "policymgr_http_enabled",
    "displayName": "HTTP enabled",
    "defaultValue": true,
    "isReconfigurable": true,
    "displayType": "checkbox",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "RangerSettings"
  },
  {
    "id": "puppet var",
    "name": "ranger_user",
    "displayName": "Used to create user and assign permission",
    "defaultValue": "ranger",
    "isReconfigurable": true,
    "displayType": "",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "ranger-env.xml",
    "category": "RangerSettings"
  },
  {
    "id": "puppet var",
    "name": "ranger_group",
    "displayName": "Used to create group and assign permission",
    "defaultValue": "ranger",
    "isReconfigurable": true,
    "displayType": "",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "ranger-env.xml",
    "category": "RangerSettings"
  },
  {
    "id": "site property",
    "name": "authentication_method",
    "displayName": "Authentication method",
    "defaultValue": "NONE",
    "options": [
      {
        displayName: 'LDAP',
        foreignKeys: ['xa_ldap_userDNpattern', 'xa_ldap_groupRoleAttribute', 'xa_ldap_url', 'xa_ldap_groupSearchBase', 'xa_ldap_groupSearchFilter']
      },
      {
        displayName: 'ACTIVE_DIRECTORY',
        foreignKeys: ['xa_ldap_ad_domain', 'xa_ldap_ad_url']
      },
      {
        displayName: 'UNIX',
        foreignKeys: ['remoteLoginEnabled', 'authServiceHostName', 'authServicePort']
      },
      {
        displayName: 'NONE'
      }
    ],
    "displayType": "radio button",
    "radioName": "authentication-method",
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "RangerSettings"
  },
  {
    "id": "site property",
    "name": "remoteLoginEnabled",
    "displayName": "Allow remote Login",
    "defaultValue": true,
    "isReconfigurable": true,
    "displayType": "checkbox",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "UnixAuthenticationSettings"
  },
  {
    "id": "site property",
    "name": "authServiceHostName",
    "displayName": "authServiceHostName",
    "defaultValue": 'localhost',
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "UnixAuthenticationSettings"
  },
  {
    "id": "site property",
    "name": "authServicePort",
    "displayName": "authServicePort",
    "defaultValue": '5151',
    "isReconfigurable": true,
    "displayType": "int",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "UnixAuthenticationSettings"
  },
  {
    "id": "site property",
    "name": "xa_ldap_url",
    "displayName": "xa_ldap_url",
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "LDAPSettings"
  },
  {
    "id": "site property",
    "name": "xa_ldap_userDNpattern",
    "displayName": "xa_ldap_userDNpattern",
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "LDAPSettings"
  },
  {
    "id": "site property",
    "name": "xa_ldap_groupRoleAttribute",
    "displayName": "xa_ldap_groupRoleAttribute",
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "LDAPSettings"
  },
  {
    "id": "site property",
    "name": "xa_ldap_ad_domain",
    "displayName": "xa_ldap_ad_domain",
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "ADSettings"
  },
  {
    "id": "site property",
    "name": "xa_ldap_ad_url",
    "displayName": "xa_ldap_ad_url",
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "RANGER",
    "filename": "admin-properties.xml",
    "category": "ADSettings"
  }
);

var additionalProperties = [];

hdp22SepcificProperties.forEach(function (config) {
  if (!hdp22properties.findProperty('name', config.name)) additionalProperties.push(config);
  else {
    hdp22properties.findProperty('name', config.name).category = config.category;
  }
});

module.exports =
{
  "configProperties": hdp22properties.concat(additionalProperties)
};

});

require.register("data/HDP2.2/tez_properties", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');

var tezPropsToCategory = {
  'General': [
    'tez.am.launch.cmd-opts',
    'tez.am.launch.env',
    'tez.task.resource.memory.mb',
    'tez.task.launch.cmd-opts',
    'tez.task.launch.env',
    'tez.grouping.split-waves',
    'tez.grouping.min-size',
    'tez.grouping.max-size',
    'tez.runtime.compress',
    'tez.runtime.compress.codec'
  ]
};

var tezProps = [];

for (var category in tezPropsToCategory) {
  tezProps = tezProps.concat(App.config.generateConfigPropertiesByName(tezPropsToCategory[category],
    { category: category, serviceName: 'TEZ', filename: 'tez-site.xml'}));
}

module.exports = tezProps;


});

require.register("data/HDP2.2/yarn_properties", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');

var yarnPropsToCategory = {

  'FaultTolerance': [
    'yarn.nodemanager.recovery.enabled',
    'yarn.resourcemanager.recovery.enabled',
    'yarn.resourcemanager.work-preserving-recovery.enabled',
    'yarn.resourcemanager.zk-address',
    'yarn.resourcemanager.connect.retry-interval.ms',
    'yarn.resourcemanager.connect.max-wait.ms',
    'yarn.resourcemanager.ha.enabled'
  ],

  'Isolation': [
    'yarn.nodemanager.linux-container-executor.group',
    'yarn.nodemanager.container-executor.class',
    'yarn.nodemanager.linux-container-executor.resources-handler.class',
    'yarn.nodemanager.linux-container-executor.cgroups.hierarchy',
    'yarn.nodemanager.linux-container-executor.cgroups.mount',
    'yarn.nodemanager.linux-container-executor.cgroups.strict-resource-usage'
  ],

  'CapacityScheduler': [
    'yarn.nodemanager.resource.cpu-vcores',
    'yarn.nodemanager.resource.percentage-physical-cpu-limit'
  ]
};

var yarnProps = [];

for (var category in yarnPropsToCategory) {
  yarnProps = yarnProps.concat(App.config.generateConfigPropertiesByName(yarnPropsToCategory[category],
    { category: category, serviceName: 'YARN', filename: 'yarn-site.xml'}));
}

module.exports = yarnProps;

});

require.register("data/HDP2/config_mapping", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var configs = [
/**********************************************core-site***************************************/
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
    "templateName": ["proxyuser_group"],
    "foreignKey": ["hive_user"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "isOverridable": true
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
    "templateName": ["hive_master_hosts"],
    "foreignKey": ["hive_user"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "isOverridable": true
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
    "templateName": [],
    "foreignKey": ["oozie_user"],
    "value": "*",
    "filename": "core-site.xml",
    "isOverridable": true
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
    "templateName": ["oozieserver_host"],
    "foreignKey": ["oozie_user"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "isOverridable": true
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
    "templateName": ["proxyuser_group"],
    "foreignKey": ["webhcat_user"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "isOverridable": true
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
    "templateName": ["hive_master_hosts"],
    "foreignKey": ["webhcat_user"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "isOverridable": true
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
    "templateName": ["proxyuser_group"],
    "foreignKey": ["falcon_user"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "isOverridable": true
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
    "templateName": [],
    "foreignKey": ["falcon_user"],
    "value": "*",
    "filename": "core-site.xml",
    "isOverridable": true
  },

/**********************************Oozie******************************/

  {
    "name": "oozie.service.ProxyUserService.proxyuser.<foreignKey[0]>.groups",
    "templateName": [],
    "foreignKey": ["falcon_user"],
    "value": "*",
    "filename": "oozie-site.xml",
    "isOverridable": true,
    "serviceName": "OOZIE"
  },
  {
    "name": "oozie.service.ProxyUserService.proxyuser.<foreignKey[0]>.hosts",
    "templateName": [],
    "foreignKey": ["falcon_user"],
    "value": "*",
    "filename": "oozie-site.xml",
    "isOverridable": true,
    "serviceName": "OOZIE"
  }
];

/**
 * Configs consists of 2 types: Computed values, which cannot be modified by user
 * and overridable values, which user can modify. We provide interface how to get all of this
 * configs separately
 * @type {Object}
 */
module.exports = {

  checkPrecondition: function () {
    return configs.filter(function (config) {
      return ((!config.precondition) || (config.precondition()));
    });
  },
  all: function (skipPreconditionCheck) {
    if (skipPreconditionCheck) {
      return configs.slice(0);
    } else {
      return this.checkPrecondition().slice(0);
    }
  },
  overridable: function (skipPreconditionCheck) {
    if (skipPreconditionCheck) {
      return configs.filterProperty('foreignKey');
    } else {
      return this.checkPrecondition().filterProperty('foreignKey');
    }
  },
  computed: function (skipPreconditionCheck) {
    if (skipPreconditionCheck) {
      return configs.filterProperty('foreignKey', null);
    } else {
      return this.checkPrecondition().filterProperty('foreignKey', null);
    }
  }
};

});

require.register("data/HDP2/custom_configs", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


module.exports =
  [
    {
      "id": "site property",
      "name": "yarn.scheduler.capacity.root.<queue-name>.capacity",
      "displayName": "Capacity",
      "value": '',
      "defaultValue": '',
      "description": "Percentage of the number of slots in the cluster that are made to be available for jobs in this queue. The sum of capacities for all queues should be less than or equal 100.",
      "isVisible": true,
      "isRequired": true,
      "serviceName": "",
      "category": "CapacityScheduler",
      "unit": "%",
      "isQueue": true,
      "filename": 'capacity-scheduler.xml',
      "inTable": true,
      "index": 1
    },
    {
      "id": "site property",
      "name": "yarn.scheduler.capacity.root.<queue-name>.maximum-capacity",
      "displayName": "Max Capacity",
      "value": '',
      "defaultValue": '100',
      "displayType": "int",
      "description": "Defines a limit beyond which a queue cannot use the capacity of the cluster." +
        "This provides a means to limit how much excess capacity a queue can use. By default, there is no limit." +
        "The Max Capacity of a queue can only be greater than or equal to its minimum capacity. " +
        "This property could be to curtail certain jobs which are long running in nature from occupying more than a certain " +
        "percentage of the cluster, which in the absence of pre-emption, could lead to capacity guarantees of other queues being affected. " +
        "One important thing to note is that maximum-capacity is a percentage , so based on the cluster's capacity it would change. " +
        "So if large no of nodes or racks get added to the cluster, Max Capacity in absolute terms would increase accordingly. ",
      "isVisible": true,
      "isRequired": true,
      "serviceName": "YARN",
      "category": "CapacityScheduler",
      "unit": "%",
      "valueRange": [0, 100],
      "isQueue": true,
      "filename": 'capacity-scheduler.xml',
      "inTable": true,
      "index": 2
    },
    {
      "id": "site property",
      "name": "yarn.scheduler.capacity.root.<queue-name>.user-limit-factor",
      "displayName": "User Limit Factor",
      "value": '',
      "defaultValue": '1',
      "displayType": "int",
      "description": "The multiple of the queue capacity which can be configured to allow a single user to acquire more slots. " +
        "By default this is set to 1 which ensure that a single user can never take more than the queue's configured capacity " +
        "irrespective of how idle the cluster is.",
      "isVisible": true,
      "isRequired": true,
      "serviceName": "YARN",
      "category": "CapacityScheduler",
      "isQueue": true,
      "filename": 'capacity-scheduler.xml',
      "inTable": true,
      "index": 7
    },
    {
      "id": "site property",
      "name": "yarn.scheduler.capacity.root.<queue-name>.state",
      "displayName": "State",
      "value": '',
      "defaultValue": 'RUNNING',
      "description": "state",
      "isVisible": true,
      "isRequired": true,
      "serviceName": "YARN",
      "category": "CapacityScheduler",
      "isQueue": true,
      "filename": 'capacity-scheduler.xml',
      "inTable": true,
      "index": 8
    },
    {
      "id": "site property",
      "name": "yarn.scheduler.capacity.root.<queue-name>.acl_submit_jobs",
      "displayName": "",
      "value": '',
      "defaultValue": '*',
      "description": "",
      "isVisible": false,
      "isRequired": true,
      "serviceName": "YARN",
      "category": "CapacityScheduler",
      "isQueue": true,
      "filename": 'capacity-scheduler.xml'
    },
    {
      "id": "site property",
      "name": "yarn.scheduler.capacity.root.<queue-name>.acl_administer_jobs",
      "displayName": "",
      "value": '',
      "defaultValue": '*',
      "description": "",
      "isVisible": false,
      "isRequired": true,
      "serviceName": "YARN",
      "category": "CapacityScheduler",
      "isQueue": true,
      "filename": 'capacity-scheduler.xml'
    }
  ];

});

require.register("data/HDP2/ha_properties", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

module.exports =
{
  "haConfig": {
    serviceName: 'MISC',
    displayName: 'MISC',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'HDFS', displayName: 'HDFS'}),
      App.ServiceConfigCategory.create({ name: 'HBASE', displayName: 'HBase'})
    ],
    sites: ['core-site', 'hdfs-site','hbase-site'],
    configs: [
    /**********************************************HDFS***************************************/
      {
        "id": "site property",
        "name": "dfs.journalnode.edits.dir",
        "displayName": "dfs.journalnode.edits.dir",
        "description": "The Directory where the JournalNode will store its local state.",
        "isReconfigurable": true,
        "defaultValue": "/hadoop/hdfs/journal",
        "value": "/hadoop/hdfs/journal",
        "defaultDirectory": "/hadoop/hdfs/journal",
        "displayType": "directory",
        "category": "HDFS",
        "filename": "hdfs-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "fs.defaultFS",
        "displayName": "fs.defaultFS",
        "description": "The default path prefix used by the Hadoop FS client when none is given.",
        "defaultValue": "hdfs://haCluster",
        "isReconfigurable": false,
        "value": "hdfs://haCluster",
        "category": "HDFS",
        "filename": "core-site",
        serviceName: 'MISC'
      },
      {
        "id": "site property",
        "name": "ha.zookeeper.quorum",
        "displayName": "ha.zookeeper.quorum",
        "isReconfigurable": false,
        "description": "This lists the host-port pairs running the ZooKeeper service.",
        "defaultValue": "zk1.example.com:2181,zk2.example.com:2181,zk3.example.com:2181",
        "value": "zk1.example.com:2181,zk2.example.com:2181,zk3.example.com:2181",
        "category": "HDFS",
        "filename": "core-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "dfs.nameservices",
        "displayName": "dfs.nameservices",
        "description": "Comma-separated list of nameservices.",
        "isReconfigurable": false,
        "defaultValue": "haCluster",
        "value": "haCluster",
        "category": "HDFS",
        "filename": "hdfs-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "dfs.ha.namenodes.${dfs.nameservices}",
        "displayName": "dfs.ha.namenodes.${dfs.nameservices}",
        "description": "The prefix for a given nameservice, contains a comma-separated list of namenodes for a given nameservice.",
        "isReconfigurable": false,
        "defaultValue": "nn1,nn2",
        "value": "nn1,nn2",
        "category": "HDFS",
        "filename": "hdfs-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "dfs.namenode.rpc-address.${dfs.nameservices}.nn1",
        "displayName": "dfs.namenode.rpc-address.${dfs.nameservices}.nn1",
        "description": "RPC address that handles all clients requests for nn1.",
        "isReconfigurable": false,
        "defaultValue": "0.0.0.0:8020",
        "value": "0.0.0.0:8020",
        "category": "HDFS",
        "filename": "hdfs-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "dfs.namenode.rpc-address.${dfs.nameservices}.nn2",
        "displayName": "dfs.namenode.rpc-address.${dfs.nameservices}.nn2",
        "description": "RPC address that handles all clients requests for nn2.",
        "isReconfigurable": false,
        "defaultValue": "0.0.0.0:8020",
        "value": "0.0.0.0:8020",
        "category": "HDFS",
        "filename": "hdfs-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "dfs.namenode.http-address.${dfs.nameservices}.nn1",
        "displayName": "dfs.namenode.http-address.${dfs.nameservices}.nn1",
        "description": "The fully-qualified HTTP address for nn1 NameNode.",
        "isReconfigurable": false,
        "defaultValue": "0.0.0.0:50070",
        "value": "0.0.0.0:50070",
        "category": "HDFS",
        "filename": "hdfs-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "dfs.namenode.http-address.${dfs.nameservices}.nn2",
        "displayName": "dfs.namenode.http-address.${dfs.nameservices}.nn2",
        "description": "The fully-qualified HTTP address for nn2 NameNode.",
        "isReconfigurable": false,
        "defaultValue": "0.0.0.0:50070",
        "value": "0.0.0.0:50070",
        "category": "HDFS",
        "filename": "hdfs-site",
        serviceName: 'MISC'
      },
      {
        "id": "site property",
        "name": "dfs.namenode.https-address.${dfs.nameservices}.nn1",
        "displayName": "dfs.namenode.https-address.${dfs.nameservices}.nn1",
        "description": "The fully-qualified HTTP address for nn1 NameNode.",
        "isReconfigurable": false,
        "defaultValue": "0.0.0.0:50470",
        "value": "0.0.0.0:50470",
        "category": "HDFS",
        "filename": "hdfs-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "dfs.namenode.https-address.${dfs.nameservices}.nn2",
        "displayName": "dfs.namenode.https-address.${dfs.nameservices}.nn2",
        "description": "The fully-qualified HTTP address for nn2 NameNode.",
        "isReconfigurable": false,
        "defaultValue": "0.0.0.0:50470",
        "value": "0.0.0.0:50470",
        "category": "HDFS",
        "filename": "hdfs-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "dfs.client.failover.proxy.provider.${dfs.nameservices}",
        "displayName": "dfs.client.failover.proxy.provider.${dfs.nameservices}",
        "description": "The Java class that HDFS clients use to contact the Active NameNode.",
        "defaultValue": "org.apache.hadoop.hdfs.server.namenode.ha.ConfiguredFailoverProxyProvider",
        "isReconfigurable": false,
        "value": "org.apache.hadoop.hdfs.server.namenode.ha.ConfiguredFailoverProxyProvider",
        "category": "HDFS",
        "filename": "hdfs-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "dfs.namenode.shared.edits.dir",
        "displayName": "dfs.namenode.shared.edits.dir",
        "description": " The URI which identifies the group of JNs where the NameNodes will write/read edits.",
        "isReconfigurable": false,
        "defaultValue": "qjournal://node1.example.com:8485;node2.example.com:8485;node3.example.com:8485/mycluster",
        "value": "qjournal://node1.example.com:8485;node2.example.com:8485;node3.example.com:8485/mycluster",
        "category": "HDFS",
        "filename": "hdfs-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "dfs.ha.fencing.methods",
        "displayName": "dfs.ha.fencing.methods",
        "description": "A list of scripts or Java classes which will be used to fence the Active NameNode during a failover.",
        "isReconfigurable": false,
        "defaultValue": "shell(/bin/true)",
        "value": "shell(/bin/true)",
        "category": "HDFS",
        "filename": "hdfs-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "dfs.ha.automatic-failover.enabled",
        "displayName": "dfs.ha.automatic-failover.enabled",
        "description": "Enable Automatic failover.",
        "isReconfigurable": false,
        "defaultValue": true,
        "value": true,
        "displayType": "checkbox",
        "category": "HDFS",
        "filename": "hdfs-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "dfs.namenode.safemode.threshold-pct",
        "displayName": "dfs.namenode.safemode.threshold-pct",
        "description": "Specifies the percentage of blocks that should satisfy\n        the minimal replication requirement defined by dfs.namenode.replication.min.\n        Values less than or equal to 0 mean not to start in safe mode.\n        Values greater than 1 will make safe mode permanent.\n ",
        "isReconfigurable": false,
        "defaultValue": "0.99f",
        "value": "0.99f",
        "category": "HDFS",
        "filename": "hdfs-site",
        "serviceName": 'MISC'
      },
      {
        "id": "site property",
        "name": "hbase.rootdir",
        "displayName": "hbase.rootdir",
        "description": "The directory shared by region servers and into which HBase persists.",
        "isReconfigurable": false,
        "defaultValue": "/hadoop/hdfs/journal",
        "value": "/hadoop/hdfs/journal",
        "category": "HBASE",
        "filename": "hbase-site",
        "serviceName": 'MISC'
      }
    ]
  }
};

});

require.register("data/HDP2/rm_ha_properties", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

module.exports =
{
  "haConfig": {
    serviceName: 'MISC',
    displayName: 'MISC',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'YARN', displayName: 'YARN'})
    ],
    sites: ['yarn-site'],
    configs: [
    /**********************************************HDFS***************************************/
      {
        "id": "site property",
        "name": "yarn.resourcemanager.ha.enabled",
        "displayName": "yarn.resourcemanager.ha.enabled",
        "isReconfigurable": false,
        "defaultValue": true,
        "value": true,
        "displayType": "checkbox",
        "category": "YARN",
        "filename": "yarn-site",
        serviceName: 'MISC'
      },
      {
        "id": "site property",
        "name": "yarn.resourcemanager.ha.rm-ids",
        "displayName": "yarn.resourcemanager.ha.rm-ids",
        "isReconfigurable": false,
        "defaultValue": "rm1,rm2",
        "value": "rm1,rm2",
        "category": "YARN",
        "filename": "yarn-site",
        serviceName: 'MISC'
      },
      {
        "id": "site property",
        "name": "yarn.resourcemanager.hostname.rm1",
        "displayName": "yarn.resourcemanager.hostname.rm1",
        "isReconfigurable": false,
        "defaultValue": "",
        "value": "",
        "category": "YARN",
        "filename": "yarn-site",
        serviceName: 'MISC'
      },

      {
        "id": "site property",
        "name": "yarn.resourcemanager.webapp.address.rm1",
        "displayName": "yarn.resourcemanager.webapp.address.rm1",
        "isReconfigurable": false,
        "defaultValue": "",
        "value": "",
        "category": "YARN",
        "filename": "yarn-site",
        serviceName: 'MISC'
      },
      {
        "id": "site property",
        "name": "yarn.resourcemanager.webapp.address.rm2",
        "displayName": "yarn.resourcemanager.webapp.address.rm2",
        "isReconfigurable": false,
        "defaultValue": "",
        "value": "",
        "category": "YARN",
        "filename": "yarn-site",
        serviceName: 'MISC'
      },
      {
        "id": "site property",
        "name": "yarn.resourcemanager.webapp.https.address.rm1",
        "displayName": "yarn.resourcemanager.webapp.https.address.rm1",
        "isReconfigurable": false,
        "defaultValue": "",
        "value": "",
        "category": "YARN",
        "filename": "yarn-site",
        serviceName: 'MISC'
      },
      {
        "id": "site property",
        "name": "yarn.resourcemanager.webapp.https.address.rm2",
        "displayName": "yarn.resourcemanager.webapp.https.address.rm2",
        "isReconfigurable": false,
        "defaultValue": "",
        "value": "",
        "category": "YARN",
        "filename": "yarn-site",
        serviceName: 'MISC'
      },

      {
        "id": "site property",
        "name": "yarn.resourcemanager.hostname.rm2",
        "displayName": "yarn.resourcemanager.hostname.rm2",
        "isReconfigurable": false,
        "defaultValue": "",
        "value": "",
        "category": "YARN",
        "filename": "yarn-site",
        serviceName: 'MISC'
      },
      {
        "id": "site property",
        "name": "yarn.resourcemanager.recovery.enabled",
        "displayName": "yarn.resourcemanager.recovery.enabled",
        "isReconfigurable": false,
        "defaultValue": true,
        "value": true,
        "displayType": "checkbox",
        "category": "YARN",
        "filename": "yarn-site",
        serviceName: 'MISC'
      },
      {
        "id": "site property",
        "name": "yarn.resourcemanager.store.class",
        "displayName": "yarn.resourcemanager.store.class",
        "isReconfigurable": false,
        "defaultValue": "org.apache.hadoop.yarn.server.resourcemanager.recovery.ZKRMStateStore",
        "value": "org.apache.hadoop.yarn.server.resourcemanager.recovery.ZKRMStateStore",
        "category": "YARN",
        "filename": "yarn-site",
        serviceName: 'MISC'
      },
      {
        "id": "site property",
        "name": "yarn.resourcemanager.zk-address",
        "displayName": "yarn.resourcemanager.zk-address",
        "isReconfigurable": false,
        "defaultValue": "",
        "value": "",
        "category": "YARN",
        "filename": "yarn-site",
        serviceName: 'MISC'
      },
      {
        "id": "site property",
        "name": "yarn.resourcemanager.cluster-id",
        "displayName": "yarn.resourcemanager.cluster-id",
        "isReconfigurable": false,
        "defaultValue": "yarn-cluster",
        "value": "yarn-cluster",
        "category": "YARN",
        "filename": "yarn-site",
        serviceName: 'MISC'
      },
      {
        "id": "site property",
        "name": "yarn.resourcemanager.ha.automatic-failover.zk-base-path",
        "displayName": "yarn.resourcemanager.ha.automatic-failover.zk-base-path",
        "isReconfigurable": false,
        "defaultValue": "/yarn-leader-election",
        "value": "/yarn-leader-election",
        "category": "YARN",
        "filename": "yarn-site",
        serviceName: 'MISC'
      }
    ]
  }
};

});

require.register("data/HDP2/secure_configs", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');
require('models/service_config');
require('models/service');

App.SecureConfigProperties = Ember.ArrayProxy.extend({
  content: require('data/HDP2/secure_properties').configProperties
});

var configProperties = App.SecureConfigProperties.create();

var configs = [
  {
    serviceName: 'GENERAL',
    displayName: 'General',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'KERBEROS', displayName: 'Kerberos'}),
      App.ServiceConfigCategory.create({ name: 'AMBARI', displayName: 'Ambari'})
    ],
    sites: ['cluster-env'],
    configs: configProperties.filterProperty('serviceName', 'GENERAL')
  },
  {
    serviceName: 'HDFS',
    displayName: 'HDFS',
    filename: 'hdfs-site',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'General', displayName: 'General', exclude:'HA'}),
      App.ServiceConfigCategory.create({ name: 'NameNode', displayName: 'NameNode'}),
      App.ServiceConfigCategory.create({ name: 'SNameNode', displayName: 'Secondary NameNode',exclude:'HA'}),
      App.ServiceConfigCategory.create({ name: 'JournalNode', displayName: 'JournalNode',exclude:'non-HA'}),
      App.ServiceConfigCategory.create({ name: 'DataNode', displayName: 'DataNode'})
    ],
    sites: ['hadoop-env','core-site', 'hdfs-site'],
    configs: configProperties.filterProperty('serviceName', 'HDFS')
  },
  {
    serviceName: 'MAPREDUCE2',
    displayName: 'MapReduce2',
    filename: 'mapred-site',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'JobHistoryServer', displayName: 'History Server'})
    ],
    sites: ['mapred-site'],
    configs: configProperties.filterProperty('serviceName', 'MAPREDUCE2')
  },
  {
    serviceName: 'YARN',
    displayName: 'YARN',
    filename: 'yarn-site',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'ResourceManager', displayName : 'ResourceManager'}),
      App.ServiceConfigCategory.create({ name: 'NodeManager', displayName : 'NodeManager'})
    ], // these properties can be dynamic
    sites: ['yarn-site', 'core-site'],
    configs: configProperties.filterProperty('serviceName', 'YARN')
  },
  {
    serviceName: 'HIVE',
    displayName: 'Hive',
    filename: 'hive-site',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'Hive Metastore', displayName: 'Hive Metastore and Hive Server 2'}),
      App.ServiceConfigCategory.create({ name: 'WebHCat Server', displayName : 'WebHCat Server'})
    ],
    sites: ['hive-site','webhcat-site'],
    configs: configProperties.filterProperty('serviceName', 'HIVE')
  },
  {
    serviceName: 'HBASE',
    displayName: 'HBase',
    filename: 'hbase-site',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'HBase Master', displayName: 'HBase Master'}),
      App.ServiceConfigCategory.create({ name: 'RegionServer', displayName: 'RegionServer'})
    ],
    sites: ['hbase-env','hbase-site'],
    configs: configProperties.filterProperty('serviceName', 'HBASE')
  },
  {
    serviceName: 'ZOOKEEPER',
    displayName: 'ZooKeeper',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'ZooKeeper Server', displayName: 'ZooKeeper Server'})
    ],
    sites: ['zookeeper-env'],
    configs: configProperties.filterProperty('serviceName', 'ZOOKEEPER')

  },
  {
    serviceName: 'OOZIE',
    displayName: 'Oozie',
    filename: 'oozie-site',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'Oozie Server', displayName:  'Oozie Server'})
    ],
    sites: ['oozie-env','oozie-site'],
    configs: configProperties.filterProperty('serviceName', 'OOZIE')
  },
  {
    serviceName: 'STORM',
    displayName: 'Storm',
    filename: 'storm-site',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'Storm Topology', displayName:  'Storm Client'})
    ],
    sites: ['storm-env','storm-site'],
    configs: configProperties.filterProperty('serviceName', 'STORM')
  },
  {
    serviceName: 'FALCON',
    displayName: 'Falcon',
    filename: 'falcon-startup.properties',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'Falcon Server', displayName:  'Falcon Server startup properties'})
    ],
    sites: ['falcon-startup.properties'],
    configs: configProperties.filterProperty('serviceName', 'FALCON')
  },
  {
    serviceName: 'KNOX',
    displayName: 'Knox',
    filename: 'gateway-site.xml',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'Knox Gateway', displayName:  'Knox Gateway'})
    ],
    sites: ['gateway-site','knox-env'],
    configs: configProperties.filterProperty('serviceName', 'KNOX')
  }
];

if(App.get('isHadoop22Stack')){
  for(var i = 0; i < configs.length; i++){
    var config = configs[i];
    if( config.serviceName === 'STORM' ){
      config.configCategories.unshift(App.ServiceConfigCategory.create({ name: 'Nimbus', displayName:  'Nimbus'}));
    }
  }
}


module.exports = configs;



});

require.register("data/HDP2/secure_mapping", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
// All of the "name" properties have to coincide with how they will appear in the *-site.xml file
// The "template" properties can come from the config properties in site_properties.js or secure_properties.js .
var props = [
  {
    "name": "hadoop.security.authentication",
    "templateName": [],
    "foreignKey": null,
    "value": "kerberos",
    "nonSecureValue": "simple",
    "filename": "core-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "hadoop.security.authorization",
    "templateName": [],
    "foreignKey": null,
    "value": "true",
    "nonSecureValue": "false",
    "filename": "core-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "hadoop.security.auth_to_local",
    "templateName": ["resourcemanager_primary_name", "kerberos_domain", "yarn_user", "nodemanager_primary_name", "namenode_primary_name", "hdfs_user", "datanode_primary_name", "hbase_master_primary_name", "hbase_user","hbase_regionserver_primary_name","oozie_primary_name","oozie_user","jobhistory_primary_name","mapred_user","journalnode_principal_name","falcon_primary_name","falcon_user"],
    "foreignKey": null,
    "value": "RULE:[2:$1@$0](<templateName[0]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nRULE:[2:$1@$0](<templateName[3]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nRULE:[2:$1@$0](<templateName[4]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[6]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[7]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[9]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[10]>@.*<templateName[1]>)s/.*/<templateName[11]>/\nRULE:[2:$1@$0](<templateName[12]>@.*<templateName[1]>)s/.*/<templateName[13]>/\nRULE:[2:$1@$0](<templateName[14]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[15]>@.*<templateName[1]>)s/.*/<templateName[16]>/\nDEFAULT",
    "filename": "core-site.xml",
    "serviceName": "HDFS",
    "dependedServiceName": [{name: "HBASE", replace: "\nRULE:[2:$1@$0](<templateName[7]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[9]>@.*<templateName[1]>)s/.*/<templateName[8]>/"},{name: "OOZIE",replace: "\nRULE:[2:$1@$0](<templateName[10]>@.*<templateName[1]>)s/.*/<templateName[11]>/"},{name: "MAPREDUCE2",replace: "\nRULE:[2:$1@$0](<templateName[12]>@.*<templateName[1]>)s/.*/<templateName[13]>/"}]
  },
  {
    "name": "dfs.namenode.kerberos.principal",
    "templateName": ["namenode_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.namenode.keytab.file",
    "templateName": ["namenode_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.secondary.namenode.kerberos.principal",
    "templateName": ["snamenode_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.secondary.namenode.keytab.file",
    "templateName": ["snamenode_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.web.authentication.kerberos.principal",
    "templateName": ["hadoop_http_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.web.authentication.kerberos.keytab",
    "templateName": ["hadoop_http_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.datanode.kerberos.principal",
    "templateName": ["datanode_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.datanode.keytab.file",
    "templateName": ["datanode_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.namenode.kerberos.internal.spnego.principal",
    "templateName": [],
    "foreignKey": null,
    "value": "${dfs.web.authentication.kerberos.principal}",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.secondary.namenode.kerberos.internal.spnego.principal",
    "templateName": [],
    "foreignKey": null,
    "value": "${dfs.web.authentication.kerberos.principal}",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.journalnode.kerberos.principal",
    "templateName": ["journalnode_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.journalnode.kerberos.internal.spnego.principal",
    "templateName": ["hadoop_http_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.journalnode.keytab.file",
    "templateName": ["journalnode_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.datanode.address",
    "templateName": ["dfs_datanode_address"],
    "foreignKey": null,
    "value": "0.0.0.0:<templateName[0]>",
    "nonSecureValue": "0.0.0.0:50010",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "dfs.datanode.http.address",
    "templateName": ["dfs_datanode_http_address"],
    "foreignKey": null,
    "value": "0.0.0.0:<templateName[0]>",
    "nonSecureValue": "0.0.0.0:50075",
    "filename": "hdfs-site.xml",
    "serviceName": "HDFS"
  },
  {
    "name": "mapreduce.jobhistory.principal",
    "templateName": ["jobhistory_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "mapred-site.xml",
    "serviceName": "MAPREDUCE2"
  },
  {
    "name": "mapreduce.jobhistory.keytab",
    "templateName": ["jobhistory_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "mapred-site.xml",
    "serviceName": "MAPREDUCE2"
  },
  {
    "name": "mapreduce.jobhistory.webapp.spnego-principal",
    "templateName": ["jobhistory_http_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "mapred-site.xml",
    "serviceName": "MAPREDUCE2"
  },
  {
    "name": "mapreduce.jobhistory.webapp.spnego-keytab-file",
    "templateName": ["jobhistory_http_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "mapred-site.xml",
    "serviceName": "MAPREDUCE2"
  },
  {
    "name": "yarn.timeline-service.enabled",
    "templateName": [],
    "foreignKey": null,
    "value": App.get('isHadoop22Stack') ? "true" : "false",
    "nonSecureValue": "false",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "yarn.acl.enable",
    "templateName": [],
    "foreignKey": null,
    "value": "true",
    "nonSecureValue": "false",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  // YARN Timeline Service
  // These "http-authentication" properties are supported in HDP Champlain
  {
    "name": "yarn.timeline-service.principal",
    "templateName": ["apptimelineserver_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "yarn.timeline-service.keytab",
    "templateName": ["apptimelineserver_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "yarn.timeline-service.http-authentication.type",
    "templateName": [],
    "foreignKey": null,
    "value": "kerberos",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "yarn.timeline-service.http-authentication.kerberos.principal",
    "templateName": ["apptimelineserver_http_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "yarn.timeline-service.http-authentication.kerberos.keytab",
    "templateName": ["apptimelineserver_http_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  // YARN Resource Manager
  {
    "name": "yarn.resourcemanager.principal",
    "templateName": ["resourcemanager_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "yarn.resourcemanager.keytab",
    "templateName": ["resourcemanager_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "yarn.nodemanager.principal",
    "templateName": ["nodemanager_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "yarn.nodemanager.keytab",
    "templateName": ["nodemanager_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "yarn.resourcemanager.webapp.spnego-principal",
    "templateName": ["resourcemanager_http_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "yarn.resourcemanager.webapp.spnego-keytab-file",
    "templateName": ["resourcemanager_http_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "yarn.nodemanager.webapp.spnego-principal",
    "templateName": ["nodemanager_http_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "yarn.nodemanager.webapp.spnego-keytab-file",
    "templateName": ["nodemanager_http_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "yarn.nodemanager.container-executor.class",
    "templateName": ["yarn_nodemanager_container-executor_class"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "nonSecureValue": "org.apache.hadoop.yarn.server.nodemanager.DefaultContainerExecutor",
    "filename": "yarn-site.xml",
    "serviceName": "YARN"
  },
  {
    "name": "hbase.master.kerberos.principal",
    "templateName": ["hbase_master_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "hbase-site.xml",
    "serviceName": "HBASE"
  },
  {
    "name": "hbase.master.keytab.file",
    "templateName": ["hbase_master_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "hbase-site.xml",
    "serviceName": "HBASE"
  },
  {
    "name": "hbase.regionserver.kerberos.principal",
    "templateName": ["hbase_regionserver_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "hbase-site.xml",
    "serviceName": "HBASE"
  },
  {
    "name": "hbase.regionserver.keytab.file",
    "templateName": ["hbase_regionserver_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "hbase-site.xml",
    "serviceName": "HBASE"
  },
  {
    "name": "hive.metastore.sasl.enabled",
    "templateName": [],
    "foreignKey": null,
    "value": "true",
    "nonSecureValue": "false",
    "filename": "hive-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "hive.security.authorization.enabled",
    "templateName": [],
    "foreignKey": null,
    "value": "true",
    "nonSecureValue": "false",
    "filename": "hive-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "hive.server2.authentication",
    "templateName": [],
    "foreignKey": null,
    "value": "KERBEROS",
    "nonSecureValue": "NONE",
    "filename": "hive-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "hive.metastore.kerberos.principal",
    "templateName": ["hive_metastore_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "hive-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "hive.metastore.kerberos.keytab.file",
    "templateName": ["hive_metastore_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "hive-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "hive.server2.authentication.kerberos.principal",
    "templateName": ["hive_metastore_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "hive-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "hive.server2.authentication.kerberos.keytab",
    "templateName": ["hive_metastore_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "hive-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "hive.server2.authentication.spnego.principal",
    "templateName": ["hive_metastore_http_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "hive-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "hive.server2.authentication.spnego.keytab",
    "templateName": ["hive_metastore_http_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "hive-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "oozie.service.AuthorizationService.authorization.enabled",
    "templateName": [],
    "foreignKey": null,
    "value": "true",
    "filename": "oozie-site.xml",
    "serviceName": "OOZIE"
  },
  {
    "name": "oozie.service.HadoopAccessorService.kerberos.enabled",
    "templateName": [],
    "foreignKey": null,
    "value": "true",
    "nonSecureValue": "false",
    "filename": "oozie-site.xml",
    "serviceName": "OOZIE"
  },
  {
    "name": "local.realm",
    "templateName": ["kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "oozie-site.xml",
    "serviceName": "OOZIE"
  },
  {
    "name": "oozie.service.HadoopAccessorService.keytab.file",
    "templateName": ["oozie_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "oozie-site.xml",
    "serviceName": "OOZIE"
  },
  {
    "name": "oozie.service.HadoopAccessorService.kerberos.principal",
    "templateName": ["oozie_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "oozie-site.xml",
    "serviceName": "OOZIE"
  },
  {
    "name": "oozie.authentication.type",
    "templateName": [],
    "foreignKey": null,
    "value": "kerberos",
    "nonSecureValue": "simple",
    "filename": "oozie-site.xml",
    "serviceName": "OOZIE"
  },
  {
    "name": "oozie.authentication.kerberos.principal",
    "templateName": ["oozie_http_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "oozie-site.xml",
    "serviceName": "OOZIE"
  },
  {
    "name": "oozie.authentication.kerberos.keytab",
    "templateName": ["oozie_http_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "oozie-site.xml",
    "serviceName": "OOZIE"
  },
  {
    "name": "oozie.authentication.kerberos.name.rules",
    "templateName": ["resourcemanager_primary_name", "kerberos_domain", "yarn_user", "nodemanager_primary_name", "namenode_primary_name", "hdfs_user", "datanode_primary_name", "hbase_master_primary_name", "hbase_user","hbase_regionserver_primary_name"],
    "foreignKey": null,
    "value": "RULE:[2:$1@$0](<templateName[0]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nRULE:[2:$1@$0](<templateName[3]>@.*<templateName[1]>)s/.*/<templateName[2]>/\nRULE:[2:$1@$0](<templateName[4]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[6]>@.*<templateName[1]>)s/.*/<templateName[5]>/\nRULE:[2:$1@$0](<templateName[7]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[9]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nDEFAULT",
    "filename": "oozie-site.xml",
    "serviceName": "OOZIE",
    "dependedServiceName": [{name: "HBASE", replace: "\nRULE:[2:$1@$0](<templateName[7]>@.*<templateName[1]>)s/.*/<templateName[8]>/\nRULE:[2:$1@$0](<templateName[9]>@.*<templateName[1]>)s/.*/<templateName[8]>/"}]
  },
  {
    "name": "templeton.kerberos.principal",
    "templateName": ["webHCat_http_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "webhcat-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "templeton.kerberos.keytab",
    "templateName": ["webhcat_http_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "webhcat-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "templeton.kerberos.secret",
    "templateName": [""],
    "foreignKey": null,
    "value": "secret",
    "filename": "webhcat-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "templeton.hive.properties",
    "templateName": ["hive_metastore","hive_metastore_principal_name","kerberos_domain"],
    "foreignKey": null,
    "value": "hive.metastore.local=false,hive.metastore.uris=<templateName[0]>,hive." +
      "metastore.sasl.enabled=true,hive.metastore.execute.setugi=true,hive.metastore.warehouse.dir=/apps/hive/warehouse,hive.exec.mode.local.auto=false,hive.metastore.kerberos.principal=<templateName[1]>@<templateName[2]>",
    "filename": "webhcat-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "hbase.security.authentication",
    "templateName": [],
    "foreignKey": null,
    "value": "kerberos",
    "nonSecureValue": "simple",
    "filename": "hbase-site.xml",
    "serviceName": "HBASE"
  },
  {
    "name": "hbase.security.authorization",
    "templateName": [],
    "foreignKey": null,
    "value": "true",
    "nonSecureValue": "false",
    "filename": "hbase-site.xml",
    "serviceName": "HBASE"
  },
  {
    "name": "hbase.bulkload.staging.dir",
    "templateName": [],
    "foreignKey": null,
    "value": "/apps/hbase/staging",
    "filename": "hbase-site.xml",
    "serviceName": "HBASE"
  },
  {
    "name": "zookeeper.znode.parent",
    "templateName": [],
    "foreignKey": null,
    "value": "/hbase-secure",
    "nonSecureValue": "/hbase-unsecure",
    "filename": "hbase-site.xml",
    "serviceName": "HBASE"
  },

  /***************************************FALCON***********************************************/
  {
    "name": "*.falcon.authentication.type",
    "templateName": [],
    "foreignKey": null,
    "value": "kerberos",
    "nonSecureValue": "simple",
    "filename": "falcon-startup.properties.xml",
    "serviceName": "FALCON"
  },
  {
    "name": "*.falcon.http.authentication.type",
    "templateName": [],
    "foreignKey": null,
    "value": "kerberos",
    "nonSecureValue": "simple",
    "filename": "falcon-startup.properties.xml",
    "serviceName": "FALCON"
  },
  {
    "name": "*.falcon.service.authentication.kerberos.principal",
    "templateName": ["falcon_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "falcon-startup.properties.xml",
    "serviceName": "FALCON"
  },
  {
    "name": "*.falcon.service.authentication.kerberos.keytab",
    "templateName": ["falcon_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "falcon-startup.properties.xml",
    "serviceName": "FALCON"
  },
  {
    "name": "*.falcon.http.authentication.kerberos.principal",
    "templateName": ["falcon_http_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "falcon-startup.properties.xml",
    "serviceName": "FALCON"
  },
  {
    "name": "*.falcon.http.authentication.kerberos.keytab",
    "templateName": ["falcon_http_keytab"],
    "foreignKey": null,
    "value": "<templateName[0]>",
    "filename": "falcon-startup.properties.xml",
    "serviceName": "FALCON"
  },
  {
    "name": "*.dfs.namenode.kerberos.principal",
    "templateName": ["namenode_principal_name_falcon", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "filename": "falcon-startup.properties.xml",
    "serviceName": "FALCON"
  },

  /***************************************KNOX***********************************************/
  {
    "name": "gateway.hadoop.kerberos.secured",
    "templateName": [],
    "foreignKey": null,
    "value": "true",
    "nonSecureValue": "false",
    "filename": "gateway-site.xml",
    "serviceName": "KNOX"
  },
  {
    "name": "java.security.krb5.conf",
    "templateName": [],
    "foreignKey": null,
    "value": "/etc/krb5.conf",
    "filename": "gateway-site.xml",
    "serviceName": "KNOX"
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
    "templateName": ["proxyuser_group"],
    "foreignKey": ["knox_primary_name"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "serviceName": "KNOX"
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
    "templateName": ["knox_gateway_hosts"],
    "foreignKey": ["knox_primary_name"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "serviceName": "KNOX"
  },
  {
    "name": "webhcat.proxyuser.<foreignKey[0]>.groups",
    "templateName": ["proxyuser_group"],
    "foreignKey": ["knox_primary_name"],
    "value": "<templateName[0]>",
    "filename": "webhcat-site.xml",
    "serviceName": "KNOX"
  },
  {
    "name": "webhcat.proxyuser.<foreignKey[0]>.hosts",
    "templateName": ["knox_gateway_hosts"],
    "foreignKey": ["knox_primary_name"],
    "value": "<templateName[0]>",
    "filename": "webhcat-site.xml",
    "serviceName": "KNOX"
  },
  {
    "name": "oozie.service.ProxyUserService.proxyuser.<foreignKey[0]>.groups",
    "templateName": ["proxyuser_group"],
    "foreignKey": ["knox_primary_name"],
    "value": "<templateName[0]>",
    "filename": "oozie-site.xml",
    "serviceName": "KNOX"
  },
  {
    "name": "oozie.service.ProxyUserService.proxyuser.<foreignKey[0]>.hosts",
    "templateName": ["knox_gateway_hosts"],
    "foreignKey": ["knox_primary_name"],
    "value": "<templateName[0]>",
    "filename": "oozie-site.xml",
    "serviceName": "KNOX"
  },
/***************************************core-site***************************************************/
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
    "templateName": ["proxyuser_group"],
    "foreignKey": ["hive_metastore_primary_name"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
    "templateName": ["hive_metastore"],
    "foreignKey": ["hive_metastore_primary_name"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
    "templateName": ["proxyuser_group"],
    "foreignKey": ["oozie_primary_name"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "serviceName": "OOZIE"
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
    "templateName": ["oozieserver_host"],
    "foreignKey": ["oozie_primary_name"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "serviceName": "OOZIE"
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.groups",
    "templateName": ["proxyuser_group"],
    "foreignKey": ["webHCat_http_primary_name"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "serviceName": "HIVE"
  },
  {
    "name": "hadoop.proxyuser.<foreignKey[0]>.hosts",
    "templateName": ["webhcat_server"],
    "foreignKey": ["webHCat_http_primary_name"],
    "value": "<templateName[0]>",
    "filename": "core-site.xml",
    "serviceName": "HIVE"
  }
];

var yarn22Mapping = [
  {
    "name": 'hadoop.http.authentication.kerberos.principal',
    "templateName": ["hadoop_http_principal_name", "kerberos_domain"],
    "foreignKey": null,
    "value": "<templateName[0]>@<templateName[1]>",
    "serviceName": "HDFS",
    "filename": "core-site.xml"
  },
  {
    "name": 'hadoop.http.authentication.kerberos.keytab',
    "foreignKey": null,
    "templateName": ["hadoop_http_keytab"],
    "value": "<templateName[0]>",
    "serviceName": "HDFS",
    "filename": "core-site.xml"
  },
  {
    name: 'hadoop.http.authentication.kerberos.name.rules',
    templateName: [],
    foreignKey: null,
    value: "",
    "serviceName": "HDFS",
    "filename": "core-site.xml"
  },
  {
    "name": 'yarn.timeline-service.http-authentication.signature.secret',
    "templateName": [],
    "foreignKey": null,
    "value": "",
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.timeline-service.http-authentication.signature.secret.file',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.timeline-service.http-authentication.signer.secret.provider',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.timeline-service.http-authentication.signer.secret.provider.object',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {

    "name": 'yarn.timeline-service.http-authentication.token.validity',
    "templateName": [],
    "foreignKey": null,
    "value": "",
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.timeline-service.http-authentication.cookie.domain',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.timeline-service.http-authentication.cookie.path',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.timeline-service.http-authentication.simple.anonymous.allowed',
    "value": "true",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.timeline-service.http-authentication.proxyusers.*.hosts',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.timeline-service.http-authentication.proxyusers.*.users',
    "value": "",
    "serviceName": "YARN",
    "templateName": [],
    "foreignKey": null,
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.timeline-service.http-authentication.proxyusers.*.groups',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'hadoop.http.filter.initializers',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "HDFS",
    "filename": "core-site.xml"
  },
  {
    "name": 'hadoop.http.authentication.type',
    "value": "simple",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "HDFS",
    "filename": "core-site.xml"
  },
  {
    "name": 'hadoop.http.authentication.signature.secret',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "HDFS",
    "filename": "core-site.xml"
  },
  {
    "name": 'hadoop.http.authentication.signature.secret.file',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "HDFS",
    "filename": "core-site.xml"
  },
  {
    "name": 'hadoop.http.authentication.signer.secret.provider',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "HDFS",
    "filename": "core-site.xml"
  },
  {
    "name": 'hadoop.http.authentication.signer.secret.provider.object',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "HDFS",
    "filename": "core-site.xml"
  },
  {
    "name": 'hadoop.http.authentication.token.validity',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "HDFS",
    "filename": "core-site.xml"
  },
  {
    "name": 'hadoop.http.authentication.cookie.domain',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "HDFS",
    "filename": "core-site.xml"
  },
  {
    "name": 'hadoop.http.authentication.cookie.path',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "HDFS",
    "filename": "core-site.xml"
  },
  {
    "name": 'yarn.timeline-service.http-authentication.kerberos.name.rules',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.resourcemanager.proxyusers.*.hosts',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.resourcemanager.proxyusers.*.users',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.resourcemanager.proxyusers.*.groups',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.resourcemanager.proxy-user-privileges.enabled',
    "value": "true",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "name": 'yarn.nodemanager.linux-container-executor.cgroups.mount-path',
    "value": "",
    "templateName": [],
    "foreignKey": null,
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  }
];

if (App.get('isHadoop22Stack')) {
  props.pushObjects(yarn22Mapping);
}

module.exports = props;

});

require.register("data/HDP2/secure_properties", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/*
 * Description of fields:
 * name: has to coincide with how the property will be saved to the file
 * displayName: how it will be shown in the Admin Security Wizard
 * serviceName: the tab in which it will appear in the Admin Security Wizard
 * filename: the file it is saved to, and the section of the command-#.json file
 * category: the accordion name in the tab shown in the Admin Security Wizard
 * component: Ambari component name
 */

var props = {
  "configProperties": [
    {
      "id": "puppet var",
      "name": "security_enabled",
      "displayName": "Enable security",
      "value": "",
      "defaultValue": 'true',
      "description": "Enable kerberos security for the cluster",
      "isVisible": false,
      "isOverridable": false,
      "serviceName": "GENERAL",
      "filename": "cluster-env.xml",
      "category": "KERBEROS"
    },
    {
      "id": "puppet var",
      "name": "kerberos_install_type",
      "displayName": "Type of security",
      "value": "",
      "defaultValue": "MANUALLY_SET_KERBEROS",
      "description": "Type of kerberos security for the cluster",
      "isVisible": false,
      "isOverridable": false,
      "serviceName": "GENERAL",
      "category": "KERBEROS"
    },
    {
      "id": "puppet var",
      "name": "kerberos_domain",
      "displayName": "Realm name",
      "value": "",
      "defaultValue": "EXAMPLE.COM",
      "description": "Realm name to be used for all principal names",
      "displayType": "advanced",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "GENERAL",
      "filename": "cluster-env.xml",
      "category": "KERBEROS"
    },
    {
      "id": "puppet var",
      "name": "kinit_path_local",
      "displayName": "Kerberos tool path",
      "value": "",
      "defaultValue": "/usr/bin",
      "description": "Directoy path to installed kerberos tools like kinit, kdestroy etc. This can have multiple comma delimited paths",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "GENERAL",
      "filename": "cluster-env.xml",
      "category": "KERBEROS"
    },
    {
      "id": "puppet var",
      "name": "smokeuser_principal_name",
      "displayName": "Smoke test user principal",
      "value": "",
      "defaultValue": "ambari-qa",
      "description": "This is the principal name for Smoke test user",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "isReconfigurable": false,
      "serviceName": "GENERAL",
      "category": "AMBARI"
    },
    {
      "id": "puppet var",
      "name": "smokeuser_keytab",
      "displayName": "Path to smoke test user keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/smokeuser.headless.keytab",
      "description": "Path to keytab file for smoke test user",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "GENERAL",
      "filename": "cluster-env.xml",
      "category": "AMBARI"
    },
    {
      "id": "puppet var",
      "name": "hdfs_principal_name",
      "displayName": "HDFS user principal",
      "value": "",
      "defaultValue": "hdfs",
      "description": "This is the principal name for HDFS user",
      "displayType": "principal",
      "isVisible": false,
      "isOverridable": false,
      "isReconfigurable": false,
      "serviceName": "GENERAL",
      "filename": "hadoop-env.xml",
      "category": "AMBARI"
    },
    {
      "id": "puppet var",
      "name": "hdfs_user_keytab",
      "displayName": "Path to HDFS user keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/hdfs.headless.keytab",
      "description": "Path to keytab file for HDFS user",
      "displayType": "directory",
      "isVisible": false,
      "isOverridable": false,
      "serviceName": "GENERAL",
      "filename": "hadoop-env.xml",
      "category": "AMBARI"
    },
    {
      "id": "puppet var",
      "name": "hbase_principal_name",
      "displayName": "HBase user principal",
      "value": "",
      "defaultValue": "hbase",
      "description": "This is the principal name for HBase user",
      "displayType": "principal",
      "isVisible": false,
      "isOverridable": false,
      "isReconfigurable": false,
      "serviceName": "GENERAL",
      "filename": "hbase-env.xml",
      "category": "AMBARI"
    },
    {
      "id": "puppet var",
      "name": "hbase_user_keytab",
      "displayName": "Path to HBase user keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/hbase.headless.keytab",
      "description": "Path to keytab file for Hbase user",
      "displayType": "directory",
      "isVisible": false,
      "isOverridable": false,
      "serviceName": "GENERAL",
      "filename": "hbase-env.xml",
      "category": "AMBARI"
    },

  /**********************************************HDFS***************************************/
    {
      "id": "puppet var",
      "name": "namenode_host",
      "displayName": "NameNode hosts",
      "value": "",
      "defaultValue": "",
      "description": "The hosts that has been assigned to run NameNode",
      "displayType": "masterHosts",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "category": "NameNode"
    },
    {
      "id": "puppet var",
      "name": "namenode_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "nn/_HOST",
      "description": "Principal name for NameNode. _HOST will get automatically replaced with actual hostname at an instance of NameNode",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "NameNode",
      "components": ["NAMENODE"]
    },
    {
      "id": "puppet var",
      "name": "namenode_keytab",
      "displayName": "Path to Keytab File",
      "value": "",
      "defaultValue": "/etc/security/keytabs/nn.service.keytab",
      "description": "Path to NameNode keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "NameNode",
      "components": ["NAMENODE"]
    },
    {
      "id": "puppet var",
      "name": "snamenode_host",
      "displayName": "SNameNode host",
      "value": "",
      "defaultValue": "localhost",
      "description": "The host that has been assigned to run SecondaryNameNode",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "category": "SNameNode"
    },
    {
      "id": "puppet var",
      "name": "snamenode_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "nn/_HOST",
      "description": "Principal name for SNameNode. _HOST will get automatically replaced with actual hostname at an instance of SNameNode",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "SNameNode",
      "components": ["SECONDARY_NAMENODE"]
    },
    {
      "id": "puppet var",
      "name": "snamenode_keytab",
      "displayName": "Path to Keytab File",
      "value": "",
      "defaultValue": "/etc/security/keytabs/nn.service.keytab",
      "description": "Path to SNameNode keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "SNameNode",
      "components": ["SECONDARY_NAMENODE"]
    },
    {
      "id": "puppet var",
      "name": "journalnode_hosts",
      "displayName": "JournalNode hosts",
      "value": "",
      "defaultValue": "localhost",
      "description": "The hosts that have been assigned to run JournalNodes",
      "displayType": "masterHosts",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "category": "JournalNode"
    },
    {
      "id": "puppet var",
      "name": "journalnode_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "jn/_HOST",
      "description": "Principal name for JournalNode. _HOST will get automatically replaced with actual hostname at every instance of JournalNode",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "JournalNode",
      "component": "JOURNALNODE"
    },
    {
      "id": "puppet var",
      "name": "journalnode_keytab",
      "displayName": "Path to keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/jn.service.keytab",
      "description": "Path to JournalNode keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "JournalNode",
      "component": "JOURNALNODE"
    },
    {
      "id": "puppet var",
      "name": "datanode_hosts", //not in the schema. For UI purpose
      "displayName": "DataNode hosts",
      "value": "",
      "defaultValue": "",
      "description": "The hosts that have been assigned to run DataNode",
      "displayType": "slaveHosts",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "category": "DataNode"
    },
    {
      "id": "puppet var",
      "name": "dfs_datanode_address",
      "displayName": "Datanode address",
      "value": "",
      "defaultValue": "1019",
      "description": "Address for DataNode",
      "displayType": "principal",
      "isVisible": false,
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "DataNode"
    },
    {
      "id": "puppet var",
      "name": "dfs_datanode_http_address",
      "displayName": "Datanode HTTP address",
      "value": "",
      "defaultValue": "1022",
      "description": "Address for DataNode",
      "displayType": "principal",
      "isVisible": false,
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "DataNode"
    },
    {
      "id": "puppet var",
      "name": "datanode_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "dn/_HOST",
      "description": "Principal name for DataNode. _HOST will get automatically replaced with actual hostname at every instance of DataNode",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "DataNode",
      "component": "DATANODE"
    },
    {
      "id": "puppet var",
      "name": "datanode_keytab",
      "displayName": "Path to keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/dn.service.keytab",
      "description": "Path to DataNode keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "DataNode",
      "component": "DATANODE"
    },
    {
      "id": "puppet var",
      "name": "hadoop_http_principal_name",
      "displayName": "DFS web principal name",
      "value": "",
      "defaultValue": "HTTP/_HOST",
      "description": "Principal name for SPNEGO access for HDFS components. _HOST will get automatically replaced with actual hostname at instance of HDFS component",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "General"
    },
    {
      "id": "puppet var",
      "name": "hadoop_http_keytab",
      "displayName": "Path to SPNEGO keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
      "description": "Path to SPNEGO keytab file for NameNode and SNameNode",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "General"
    },

  /**********************************************MAPREDUCE2***************************************/
    {
      "id": "puppet var",
      "name": "jobhistoryserver_host",
      "displayName": "History Server host",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run History Server",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MAPREDUCE2",
      "category": "JobHistoryServer"
    },
    {
      "id": "puppet var",
      "name": "jobhistory_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "jhs/_HOST",
      "description": "Principal name for History Server. _HOST will get automatically replaced with actual hostname at an instance of History Server",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "MAPREDUCE2",
      "category": "JobHistoryServer",
      "component": "HISTORYSERVER"
    },
    {
      "id": "puppet var",
      "name": "jobhistory_keytab",
      "displayName": "Path to keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/jhs.service.keytab",
      "description": "Path to History Server keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "MAPREDUCE2",
      "category": "JobHistoryServer",
      "component": "HISTORYSERVER"
    },
    {
      "id": "puppet var",
      "name": "jobhistory_http_principal_name",
      "displayName": "Web principal name",
      "value": "",
      "defaultValue": "HTTP/_HOST",
      "description": "Principal name for SPNEGO access to Job History Server. _HOST will get automatically replaced with actual hostname at an instance of Job History Server",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "MAPREDUCE2",
      "category": "JobHistoryServer"
    },
    {
      "id": "puppet var",
      "name": "jobhistory_http_keytab",
      "displayName": "Path to SPNEGO keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
      "description": "Path to SPNEGO keytab file for Job History Server",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "MAPREDUCE2",
      "category": "JobHistoryServer"
    },

  /**********************************************YARN***************************************/
    {
      "id": "puppet var",
      "name": "resourcemanager_host",
      "displayName": "ResourceManager host",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run ResourceManager",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "YARN",
      "category": "ResourceManager"
    },
    // YARN Application Timeline Server
    {
      "id": "puppet var",
      "name": "apptimelineserver_host",
      "displayName": "Application Timeline Server host",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run AppTimelineServer",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "YARN",
      "category": "AppTimelineServer"
    },
    {
      "id": "puppet var",
      "name": "apptimelineserver_principal_name",
      "displayName": "App Timeline Server Principal name",
      "value": "",
      "defaultValue": "yarn/_HOST",
      "description": "Principal name for App Timeline Server. _HOST will get automatically replaced with actual hostname at an instance of App Timeline Server",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": true,
      "serviceName": "YARN",
      "category": "AppTimelineServer",
      "component": "APP_TIMELINE_SERVER"
    },
    {
      "id": "puppet var",
      "name": "apptimelineserver_keytab",
      "displayName": "Path to App Timeline Server keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/yarn.service.keytab",
      "description": "Path to App Timeline Server keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": true,
      "serviceName": "YARN",
      "category": "AppTimelineServer",
      "component": "APP_TIMELINE_SERVER"
    },
    {
      "id": "puppet var",
      "name": "apptimelineserver_http_principal_name",
      "displayName": "App Timeline Server HTTP Principal name",
      "value": "",
      "defaultValue": "HTTP/_HOST",
      "description": "Principal name for App Timeline Server HTTP. _HOST will get automatically replaced with actual hostname at an instance of App Timeline Server",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": true,
      "serviceName": "YARN",
      "category": "AppTimelineServer"
    },
    {
      "id": "puppet var",
      "name": "apptimelineserver_http_keytab",
      "displayName": "Path to App Timeline Server SPNEGO HTTP keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
      "description": "Path to App Timeline Server SPNEGO HTTP keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": true,
      "serviceName": "YARN",
      "category": "AppTimelineServer"
    },

    // YARN Resource Manager
    {
      "id": "puppet var",
      "name": "resourcemanager_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "rm/_HOST",
      "description": "Principal name for ResourceManager. _HOST will get automatically replaced with actual hostname at an instance of ResourceManager",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "YARN",
      "category": "ResourceManager",
      "component": "RESOURCEMANAGER"
    },
    {
      "id": "puppet var",
      "name": "resourcemanager_keytab",
      "displayName": "Path to keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/rm.service.keytab",
      "description": "Path to ResourceManager keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "YARN",
      "category": "ResourceManager",
      "component": "RESOURCEMANAGER"
    },
    {
      "id": "puppet var",
      "name": "resourcemanager_http_principal_name",
      "displayName": "Web principal name",
      "value": "",
      "defaultValue": "HTTP/_HOST",
      "description": "Principal name for SPNEGO access to ResourceManager. _HOST will get automatically replaced with actual hostname at an instance of ResourceManager",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "YARN",
      "category": "ResourceManager"
    },
    {
      "id": "puppet var",
      "name": "resourcemanager_http_keytab",
      "displayName": "Path to SPNEGO keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
      "description": "Path to SPNEGO keytab file for ResourceManager",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "YARN",
      "category": "ResourceManager"
    },
    {
      "id": "puppet var",
      "name": "nodemanager_host",
      "displayName": "NodeManager hosts",
      "value": "",
      "defaultValue": "",
      "description": "The hosts that has been assigned to run NodeManager",
      "displayType": "slaveHosts",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "YARN",
      "category": "NodeManager"
    },
    {
      "id": "puppet var",
      "name": "nodemanager_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "nm/_HOST",
      "description": "Principal name for NodeManager. _HOST will get automatically replaced with actual hostname at all instances of NodeManager",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "YARN",
      "category": "NodeManager",
      "component": "NODEMANAGER"
    },
    {
      "id": "puppet var",
      "name": "nodemanager_keytab",
      "displayName": "Path to keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/nm.service.keytab",
      "description": "Path to NodeManager keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "YARN",
      "category": "NodeManager",
      "component": "NODEMANAGER"
    },
    {
      "id": "puppet var",
      "name": "nodemanager_http_principal_name",
      "displayName": "Web principal name",
      "value": "",
      "defaultValue": "HTTP/_HOST",
      "description": "Principal name for SPNEGO access to NodeManager. _HOST will get automatically replaced with actual hostname at all instances of NodeManager",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "YARN",
      "category": "NodeManager"
    },
    {
      "id": "puppet var",
      "name": "nodemanager_http_keytab",
      "displayName": "Path to SPNEGO keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
      "description": "Path to SPNEGO keytab file for NodeManager",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "YARN",
      "category": "NodeManager"
    },
    {
      "id": "puppet var",
      "name": "yarn_nodemanager_container-executor_class",
      "displayName": "yarn.nodemanager.container-executor.class",
      "value": "",
      "defaultValue": "org.apache.hadoop.yarn.server.nodemanager.LinuxContainerExecutor",
      "description": "Executor(launcher) of the containers",
      "displayType": "advanced",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "YARN",
      "category": "NodeManager"
    },
    {
      "id": "puppet var",
      "name": "webhcatserver_host",
      "displayName": "WebHCat Server host",
      "value": "",
      "defaultValue": "localhost",
      "description": "The host that has been assigned to run WebHCat Server",
      "displayType": "masterHost",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HIVE",
      "category": "WebHCat Server"
    },
    {
      "id": "puppet var",
      "name": "webHCat_http_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "HTTP/_HOST",
      "description": "Principal name for SPNEGO access for WebHCat",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HIVE",
      "category": "WebHCat Server"
    },
    {
      "id": "puppet var",
      "name": "webhcat_http_keytab",
      "displayName": "Path to keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
      "description": "Path to SPNEGO keytab file for WebHCat",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HIVE",
      "category": "WebHCat Server"
    },

  /**********************************************HBASE***************************************/
    {
      "id": "puppet var",
      "name": "hbasemaster_host",
      "displayName": "HBase Master hosts",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run HBase Master",
      "displayType": "masterHosts",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HBASE",
      "category": "HBase Master"
    },
    {
      "id": "puppet var",
      "name": "hbase_master_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "hbase/_HOST",
      "description": "Principal name for HBase master. _HOST will get automatically replaced with actual hostname at an instance of HBase Master",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HBASE",
      "category": "HBase Master",
      "components": ["HBASE_MASTER"]
    },
    {
      "id": "puppet var",
      "name": "hbase_master_keytab",
      "displayName": "Path to keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/hbase.service.keytab",
      "description": "Path to HBase master keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HBASE",
      "category": "HBase Master",
      "components": ["HBASE_MASTER"]
    },
    {
      "id": "puppet var",
      "name": "regionserver_hosts",
      "displayName": "RegionServer hosts",
      "value": "",
      "defaultValue": "",
      "description": "The hosts that have been assigned to run RegionServer",
      "displayType": "slaveHosts",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HBASE",
      "category": "RegionServer"
    },
    {
      "id": "puppet var",
      "name": "hbase_regionserver_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "hbase/_HOST",
      "description": "Principal name for RegionServer. _HOST will get automatically replaced with actual hostname at every instance of RegionServer",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HBASE",
      "category": "RegionServer",
      "components": ["HBASE_REGIONSERVER"]
    },
    {
      "id": "puppet var",
      "name": "hbase_regionserver_keytab",
      "displayName": "Path to keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/hbase.service.keytab",
      "description": "Path to RegionServer keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HBASE",
      "category": "RegionServer",
      "components": ["HBASE_REGIONSERVER"]
    },

  /**********************************************HIVE***************************************/
    {
      "id": "puppet var",
      "name": "hive_metastore",
      "displayName": "Hive Metastore hosts",
      "value": "",
      "defaultValue": "localhost",
      "description": "The hosts that have been assigned to run Hive Metastore and HiveServer2",
      "displayType": "masterHosts",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HIVE",
      "category": "Hive Metastore"
    },
    {
      "id": "puppet var",
      "name": "hive_metastore_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "hive/_HOST",
      "description": "Principal name for Hive Metastore and HiveServer2. _HOST will get automatically replaced with actual hostname at an instance of Hive Metastore and HiveServer2",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HIVE",
      "category": "Hive Metastore",
      "components": ["HIVE_SERVER", "HIVE_METASTORE"]
    },
    {
      "id": "puppet var",
      "name": "hive_metastore_keytab",
      "displayName": "Path to keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/hive.service.keytab",
      "description": "Path to Hive Metastore and HiveServer2 keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HIVE",
      "category": "Hive Metastore",
      "components": ["HIVE_SERVER", "HIVE_METASTORE"]
    },
    {
      "id": "puppet var",
      "name": "hive_metastore_http_principal_name",
      "displayName": "Web principal name",
      "value": "",
      "defaultValue": "HTTP/_HOST",
      "description": "Principal name for SPNEGO access to Hive Metastore and HiveServer2. _HOST will get automatically replaced with actual hostname at an instance of Hive Metastore and HiveServer2",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HIVE",
      "category": "Hive Metastore"
    },
    {
      "id": "puppet var",
      "name": "hive_metastore_http_keytab",
      "displayName": "Path to SPNEGO keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
      "description": "Path to SPNEGO keytab file for  Hive Metastore and HiveServer2",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "HIVE",
      "category": "Hive Metastore"
    },

  /**********************************************OOZIE***************************************/
    {
      "id": "puppet var",
      "name": "oozie_servername",
      "displayName": "Oozie Server host",
      "value": "",
      "defaultValue": "localhost",
      "description": "Oozie server host name",
      "displayType": "masterHost",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "OOZIE",
      "category": "Oozie Server"
    },
    {
      "id": "puppet var",
      "name": "oozie_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "oozie/_HOST",
      "description": "Principal name for Oozie server",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "OOZIE",
      "category": "Oozie Server",
      "component": "OOZIE_SERVER"
    },
    {
      "id": "puppet var",
      "name": "oozie_keytab",
      "displayName": "Path to keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/oozie.service.keytab",
      "description": "Path to Oozie server keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "Oozie Server",
      "component": "OOZIE_SERVER"
    },
    {
      "id": "puppet var",
      "name": "oozie_http_principal_name",
      "displayName": "Web principal name",
      "value": "",
      "defaultValue": "HTTP/_HOST",
      "description": "Principal name for SPNEGO access to Oozie",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "OOZIE",
      "category": "Oozie Server"
    },
    {
      "id": "puppet var",
      "name": "oozie_http_keytab",
      "displayName": "Path to SPNEGO keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
      "description": "Path to SPNEGO keytab file for oozie",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "OOZIE",
      "category": "Oozie Server"
    },

  /**********************************************ZOOKEEPER***************************************/
    {
      "id": "puppet var",
      "name": "zookeeperserver_hosts",
      "displayName": "ZooKeeper Server hosts",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run ZooKeeper Server",
      "displayType": "masterHosts",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "ZOOKEEPER",
      "category": "ZooKeeper Server"
    },
    {
      "id": "puppet var",
      "name": "zookeeper_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "zookeeper/_HOST",
      "description": "Principal name for ZooKeeper. _HOST will get automatically replaced with actual hostname at every instance of zookeeper server",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZooKeeper Server",
      "component": "ZOOKEEPER_SERVER"
    },
    {
      "id": "puppet var",
      "name": "zookeeper_keytab_path",
      "displayName": "Path to keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/zk.service.keytab",
      "description": "Path to ZooKeeper keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZooKeeper Server",
      "component": "ZOOKEEPER_SERVER"
    },
  /**********************************************Falcon***************************************/
    {
      "id": "puppet var",
      "name": "falcon_server_host",
      "displayName": "Falcon server host",
      "value": "",
      "defaultValue": "",
      "description": "Falcon Server host",
      "displayType": "masterHost",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "FALCON",
      "category": "Falcon Server"
    },
    {
      "id": "puppet var",
      "name": "falcon_principal_name",
      "displayName": "Falcon principal name",
      "value": "",
      "defaultValue": "falcon/_HOST",
      "description": "This is the principal name for Falcon Server",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "FALCON",
      "category": "Falcon Server",
      "component": "FALCON_SERVER"
    },
    {
      "id": "puppet var",
      "name": "falcon_keytab",
      "displayName": "Path to Falcon server keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/falcon.service.keytab",
      "description": "Path to the Falcon Server keytab file",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "FALCON",
      "category": "Falcon Server",
      "component": "FALCON_SERVER"
    },
    {
      "id": "puppet var",
      "name": "falcon_http_principal_name",
      "displayName": "Web principal name",
      "value": "",
      "defaultValue": "HTTP/_HOST",
      "description": "Principal name for SPNEGO access to Falcon",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "FALCON",
      "category": "Falcon Server"
    },
    {
      "id": "puppet var",
      "name": "falcon_http_keytab",
      "displayName": "Path to SPNEGO keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
      "description": "Path to SPNEGO keytab file for Falcon",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "FALCON",
      "category": "Falcon Server"
    },
    {
      "id": "puppet var",
      "name": "namenode_principal_name_falcon",
      "displayName": "NameNode principal name",
      "value": "",
      "defaultValue": "nn/_HOST",
      "description": "NameNode principal to talk to config store",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "FALCON",
      "category": "Falcon Server"
    },

  /**********************************************Knox***************************************/
    {
      "id": "puppet var",
      "name": "knox_gateway_hosts",
      "displayName": "Knox Gateway hosts",
      "value": "",
      "defaultValue": "",
      "description": "The hosts that has been assigned to run Knox Gateway",
      "displayType": "masterHosts",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "KNOX",
      "category": "Knox Gateway"
    },
    {
      "id": "puppet var",
      "name": "knox_principal_name",
      "displayName": "Principal name",
      "value": "",
      "defaultValue": "knox/_HOST",
      "description": "This is the principal name for Knox Gateway",
      "displayType": "principal",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "KNOX",
      "filename": "knox-env.xml",
      "category": "Knox Gateway",
      "component": "KNOX_GATEWAY"
    },
    {
      "id": "puppet var",
      "name": "knox_keytab_path",
      "displayName": "Path to keytab file",
      "value": "",
      "defaultValue": "/etc/security/keytabs/knox.service.keytab",
      "description": "This is the keytab file for Knox Gateway",
      "displayType": "directory",
      "isVisible": true,
      "isOverridable": false,
      "serviceName": "KNOX",
      "filename": "knox-env.xml",
      "category": "Knox Gateway",
      "component": "KNOX_GATEWAY"
    }
  ]
};

var stormProperties = [
  {
    "id": "puppet var",
    "name": "storm_host",
    "displayName": "Storm component hosts",
    "value": "",
    "defaultValue": "",
    "description": "Storm component hosts",
    "displayType": "slaveHosts",
    "isVisible": true,
    "isOverridable": false,
    "serviceName": "STORM",
    "category": "Storm Topology"
  },
  {
    "id": "puppet var",
    "name": "storm_principal_name",
    "displayName": " Storm principal name",
    "value": "",
    "defaultValue": "storm/_HOST",
    "description": "Principal name for Supervisor. _HOST will get automatically replaced with actual hostname at an instance of every storm component.",
    "displayType": "principal",
    "isVisible": true,
    "isOverridable": false,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "category": "Storm Topology",
    "components": ["SUPERVISOR", "NIMBUS", "STORM_UI_SERVER"]
  },
  {
    "id": "puppet var",
    "name": "storm_keytab",
    "displayName": "Path to Storm keytab file",
    "value": "",
    "defaultValue": "/etc/security/keytabs/storm.service.keytab",
    "description": "Path to the storm keytab file",
    "displayType": "directory",
    "isVisible": true,
    "isOverridable": false,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "category": "Storm Topology",
    "components": ["SUPERVISOR", "NIMBUS"]
  }
];
var storm22Properties = [
  {
    "id": "puppet var",
    "name": "storm_principal_name",
    "displayName": " Storm principal name",
    "value": "",
    "defaultValue": "storm/_HOST",
    "description": "Principal name for Storm components. _HOST will get automatically replaced with actual hostname at an instance of every storm component.",
    "displayType": "principal",
    "isVisible": true,
    "isOverridable": false,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "category": "Storm Topology",
    "components": ["SUPERVISOR", "STORM_UI_SERVER", "DRPC_SERVER", "STORM_REST_API"]
  },
  {
    "id": "puppet var",
    "name": "storm_keytab",
    "displayName": "Path to Storm keytab file",
    "value": "",
    "defaultValue": "/etc/security/keytabs/storm.service.keytab",
    "description": "Path to the storm keytab file",
    "displayType": "directory",
    "isVisible": true,
    "isOverridable": false,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "category": "Storm Topology",
    "components": ["SUPERVISOR", "STORM_UI_SERVER", "DRPC_SERVER", "STORM_REST_API"]
  },
  {
    "id": "puppet var",
    "name": "nimbus_host",
    "displayName": "Nimbus hosts",
    "value": "",
    "defaultValue": "",
    "description": "Nimbus component hosts",
    "displayType": "slaveHosts",
    "isVisible": true,
    "isOverridable": false,
    "serviceName": "STORM",
    "category": "Nimbus"
  },
  {
    "id": "puppet var",
    "name": "nimbus_principal_name",
    "displayName": " Nimbus principal name",
    "value": "",
    "defaultValue": "nimbus/_HOST",
    "description": "Nimbus Principal name",
    "displayType": "principal",
    "isVisible": true,
    "isOverridable": false,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "category": "Nimbus",
    "components": ["NIMBUS","DRPC_SERVER"]
  },
  {
    "id": "puppet var",
    "name": "nimbus_keytab",
    "displayName": "Path to Nimbus keytab file",
    "value": "",
    "defaultValue": "/etc/security/keytabs/nimbus.service.keytab",
    "description": "Path to the nimbus keytab file",
    "displayType": "directory",
    "isVisible": true,
    "isOverridable": false,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "category": "Nimbus",
    "components": ["NIMBUS","DRPC_SERVER"]
  },
  {
    "id": "puppet var",
    "name": "storm_ui_principal_name",
    "displayName": "Storm UI principal name",
    "value": "",
    "defaultValue": "HTTP/_HOST",
    "description": "Principal name for Storm UI",
    "displayType": "principal",
    "isVisible": false,
    "isOverridable": false,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "category": "Nimbus"
  },
  {
    "id": "puppet var",
    "name": "storm_ui_keytab",
    "displayName": "Path to Nimbus UI keytab file",
    "value": "",
    "defaultValue": "/etc/security/keytabs/spnego.service.keytab",
    "description": "Path to the Storm UI keytab file",
    "displayType": "directory",
    "isVisible": false,
    "isOverridable": false,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "category": "Nimbus"
  }
];

if(App.get('isHadoop22Stack')) {
  props.configProperties.pushObjects(storm22Properties);
} else {
  props.configProperties.pushObjects(stormProperties);
}

module.exports = props;

});

require.register("data/HDP2/site_properties", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

var hdp2properties = [
  //***************************************** HDP stack **************************************
/**********************************************HDFS***************************************/
  {
    "id": "site property",
    "name": "dfs.namenode.checkpoint.dir",
    "displayName": "SecondaryNameNode Checkpoint directories",
    "defaultDirectory": "/hadoop/hdfs/namesecondary",
    "displayType": "directories",
    "isOverridable": false,
    "serviceName": "HDFS",
    "filename": "hdfs-site.xml",
    "category": "SECONDARY_NAMENODE",
    "index": 1
  },
  {
    "id": "site property",
    "name": "dfs.namenode.checkpoint.period",
    "displayName": "HDFS Maximum Checkpoint Delay",
    "displayType": "int",
    "unit": "seconds",
    "category": "General",
    "serviceName": "HDFS",
    "filename": "hdfs-site.xml",
    "index": 3
  },
  {
    "id": "site property",
    "name": "dfs.namenode.name.dir",
    "displayName": "NameNode directories",
    "defaultDirectory": "/hadoop/hdfs/namenode",
    "displayType": "directories",
    "isOverridable": false,
    "serviceName": "HDFS",
    "filename": "hdfs-site.xml",
    "category": "NAMENODE",
    "index": 1
  },
  {
    "id": "site property",
    "name": "dfs.webhdfs.enabled",
    "displayName": "WebHDFS enabled",
    "displayType": "checkbox",
    "isOverridable": false,
    "category": "General",
    "serviceName": "HDFS",
    "filename": "hdfs-site.xml",
    "index": 0
  },
  {
    "id": "site property",
    "name": "dfs.datanode.failed.volumes.tolerated",
    "displayName": "DataNode volumes failure toleration",
    "displayType": "int",
    "category": "DATANODE",
    "serviceName": "HDFS",
    "filename": "hdfs-site.xml",
    "index": 3
  },
  {
    "id": "site property",
    "name": "dfs.datanode.data.dir.mount.file",
    "displayName": "File that stores mount point for each data dir",
    "description": "File path that contains the last known mount point for each data dir. This file is used to avoid creating a DFS data dir on the root drive (and filling it up) if a path was previously mounted on a drive.",
    "defaultValue": "/etc/hadoop/conf/dfs_data_dir_mount.hist",
    "displayType": "directory",
    "isVisible": true,
    "category": "DATANODE",
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "index": 4
  },
  {
    "id": "site property",
    "name": "dfs.datanode.data.dir",
    "displayName": "DataNode directories",
    "defaultDirectory": "/hadoop/hdfs/data",
    "displayType": "directories",
    "category": "DATANODE",
    "serviceName": "HDFS",
    "filename": "hdfs-site.xml",
    "index": 1
  },
  {
    "id": "site property",
    "name": "dfs.datanode.data.dir.perm",
    "displayName": "DataNode directories permission",
    "displayType": "int",
    "category": "DATANODE",
    "serviceName": "HDFS",
    "filename": "hdfs-site.xml"
  },
  {
    "id": "site property",
    "name": "dfs.replication",
    "displayName": "Block replication",
    "displayType": "int",
    "category": "General",
    "serviceName": "HDFS",
    "filename": "hdfs-site.xml"
  },
  {
    "id": "site property",
    "name": "dfs.datanode.du.reserved",
    "displayName": "Reserved space for HDFS",
    "displayType": "int",
    "unit": "bytes",
    "category": "General",
    "serviceName": "HDFS",
    "filename": "hdfs-site.xml",
    "index": 2
  },
  {
    "id": "site property",
    "name": "dfs.client.read.shortcircuit",
    "displayName": "HDFS Short-circuit read",
    "displayType": "checkbox",
    "category": "Advanced hdfs-site",
    "serviceName": "HDFS",
    "filename": "hdfs-site.xml"
  },
  {
    "id": "site property",
    "name": "apache_artifacts_download_url",
    "displayName": "apache_artifacts_download_url",
    "description": "",
    "isRequired": false,
    "isRequiredByAgent": false,
    "isVisible": false,
    "category": "Advanced hdfs-site",
    "serviceName": "HDFS"
  },
  {
    "id": "site property",
    "name": "dfs.cluster.administrators",
    "displayName": "dfs.cluster.administrators",
    "description": "ACL for who all can view the default servlets in the HDFS",
    "isVisible": true,
    "category": "Advanced hdfs-site",
    "serviceName": "HDFS",
    "filename": "hdfs-site.xml"
  },

/**********************************************YARN***************************************/
  {
    "id": "site property",
    "name": "yarn.acl.enable",
    "displayName": "yarn.acl.enable",
    "displayType": "checkbox",
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "RESOURCEMANAGER"
  },
  {
    "id": "site property",
    "name": "yarn.admin.acl",
    "displayName": "yarn.admin.acl",
    "isRequired": false,
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "RESOURCEMANAGER"
  },
  {
    "id": "site property",
    "name": "yarn.log-aggregation-enable",
    "displayName": "yarn.log-aggregation-enable",
    "displayType": "checkbox",
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "RESOURCEMANAGER"
  },
  {
    "id": "site property",
    "name": "yarn.resourcemanager.scheduler.class",
    "displayName": "yarn.resourcemanager.scheduler.class",
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "CapacityScheduler"
  },
  {
    "id": "site property",
    "name": "yarn.scheduler.minimum-allocation-mb",
    "displayName": "yarn.scheduler.minimum-allocation-mb",
    "displayType": "int",
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "CapacityScheduler"
  },
  {
    "id": "site property",
    "name": "yarn.scheduler.maximum-allocation-mb",
    "displayName": "yarn.scheduler.maximum-allocation-mb",
    "displayType": "int",
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "CapacityScheduler"
  },
  {
    "id": "site property",
    "name": "yarn.nodemanager.resource.memory-mb",
    "displayName": "yarn.nodemanager.resource.memory-mb",
    "displayType": "int",
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "NODEMANAGER"
  },
  {
    "id": "site property",
    "name": "yarn.nodemanager.vmem-pmem-ratio",
    "displayName": "yarn.nodemanager.vmem-pmem-ratio",
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "NODEMANAGER"
  },
  {
    "id": "site property",
    "name": "yarn.nodemanager.linux-container-executor.group",
    "displayName": "yarn.nodemanager.linux-container-executor.group",
    "serviceName": "YARN",
    "category": "NODEMANAGER",
    "filename": "yarn-site.xml"
  },
  {
    "id": "site property",
    "name": "yarn.nodemanager.log-dirs",
    "displayName": "yarn.nodemanager.log-dirs",
    "defaultDirectory": "/hadoop/yarn/log",
    "displayType": "directories",
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "NODEMANAGER"
  },
  {
    "id": "site property",
    "name": "yarn.nodemanager.local-dirs",
    "displayName": "yarn.nodemanager.local-dirs",
    "defaultDirectory": "/hadoop/yarn/local",
    "displayType": "directories",
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "NODEMANAGER"
  },
  {
    "id": "site property",
    "name": "yarn.nodemanager.remote-app-log-dir",
    "displayName": "yarn.nodemanager.remote-app-log-dir",
    "displayType": "directory",
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "NODEMANAGER"
  },
  {
    "id": "site property",
    "name": "yarn.nodemanager.remote-app-log-dir-suffix",
    "displayName": "yarn.nodemanager.remote-app-log-dir-suffix",
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "NODEMANAGER"
  },
  {
    "id": "site property",
    "name": "yarn.nodemanager.aux-services",
    "displayName": "yarn.nodemanager.aux-services",
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "NODEMANAGER"
  },
  {
    "id": "site property",
    "name": "yarn.nodemanager.log.retain-second",
    "displayName": "yarn.nodemanager.log.retain-second",
    "serviceName": "YARN",
    "filename": "yarn-site.xml",
    "category": "NODEMANAGER"
  },
  {
    "id": "site property",
    "name": "yarn.log.server.url",
    "displayName": "yarn.log.server.url",
    "category": "Advanced yarn-site",
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
/**********************************************MAPREDUCE2***************************************/
  {
    "id": "site property",
    "name": "mapreduce.map.memory.mb",
    "displayName": "Default virtual memory for a job's map-task",
    "displayType": "int",
    "unit": "MB",
    "category": "General",
    "serviceName": "MAPREDUCE2",
    "filename": "mapred-site.xml"
  },
  {
    "id": "site property",
    "name": "mapreduce.reduce.memory.mb",
    "displayName": "Default virtual memory for a job's reduce-task",
    "displayType": "int",
    "unit": "MB",
    "category": "General",
    "serviceName": "MAPREDUCE2",
    "filename": "mapred-site.xml"
  },
  {
    "id": "site property",
    "name": "mapreduce.task.io.sort.mb",
    "displayName": "Map-side sort buffer memory",
    "displayType": "int",
    "unit": "MB",
    "category": "General",
    "serviceName": "MAPREDUCE2",
    "filename": "mapred-site.xml"
  },
  {
    "id": "site property",
    "name": "hadoop.security.auth_to_local",
    "displayName": "hadoop.security.auth_to_local",
    "displayType": "multiLine",
    "serviceName": "HDFS",
    "filename": "core-site.xml",
    "category": "Advanced core-site"
  },
  {
    "id": "site property",
    "name": "yarn.app.mapreduce.am.resource.mb",
    "displayName": "yarn.app.mapreduce.am.resource.mb",
    "displayType": "int",
    "category": "Advanced mapred-site",
    "serviceName": "MAPREDUCE2",
    "filename": "mapred-site.xml"
  },

/**********************************************oozie-site***************************************/
  {
    "id": "site property",
    "name": "oozie.db.schema.name",
    "displayName": "Database Name",
    "isOverridable": false,
    "displayType": "host",
    "isObserved": true,
    "category": "OOZIE_SERVER",
    "serviceName": "OOZIE",
    "filename": "oozie-site.xml",
    "index": 4
  },
  {
    "id": "site property",
    "name": "oozie.service.JPAService.jdbc.username",
    "displayName": "Database Username",
    "isOverridable": false,
    "displayType": "user",
    "category": "OOZIE_SERVER",
    "serviceName": "OOZIE",
    "filename": "oozie-site.xml",
    "index": 5
  },
  {
    "id": "site property",
    "name": "oozie.service.JPAService.jdbc.password",
    "displayName": "Database Password",
    "isOverridable": false,
    "displayType": "password",
    "category": "OOZIE_SERVER",
    "serviceName": "OOZIE",
    "filename": "oozie-site.xml",
    "index": 6
  },
  {
    "id": "site property",
    "name": "oozie.service.JPAService.jdbc.driver", // the default value of this property is overriden in code
    "displayName": "JDBC Driver Class",
    "isOverridable": false,
    "category": "OOZIE_SERVER",
    "serviceName": "OOZIE",
    "filename": "oozie-site.xml",
    "index": 7
  },
  {
    "id": "site property",
    "name": "oozie.service.JPAService.jdbc.url",
    "displayName": "Database URL",
    "isOverridable": false,
    "displayType": "advanced",
    "category": "OOZIE_SERVER",
    "serviceName": "OOZIE",
    "filename": "oozie-site.xml",
    "index": 8
  },

/**********************************************hive-site***************************************/
  {
    "id": "site property",
    "name": "javax.jdo.option.ConnectionDriverName",  // the default value is overwritten in code
    "displayName": "JDBC Driver Class",
    "isOverridable": false,
    "category": "HIVE_METASTORE",
    "serviceName": "HIVE",
    "filename": "hive-site.xml",
    "index": 7
  },
  {
    "id": "site property",
    "name": "hive.heapsize",
    "displayName": "Hive heap size",
    "displayType": "int",
    "unit": "MB",
    "isOverridable": false,
    "serviceName": "HIVE",
    "filename": "hive-site.xml",
    "category": "General",
    "index": 9
  },
  {
    "id": "site property",
    "name": "javax.jdo.option.ConnectionUserName",
    "displayName": "Database Username",
    "displayType": "user",
    "isOverridable": false,
    "category": "HIVE_METASTORE",
    "serviceName": "HIVE",
    "filename": "hive-site.xml",
    "index": 5
  },
  {
    "id": "site property",
    "name": "javax.jdo.option.ConnectionPassword",
    "displayName": "Database Password",
    "displayType": "password",
    "isOverridable": false,
    "category": "HIVE_METASTORE",
    "serviceName": "HIVE",
    "filename": "hive-site.xml",
    "index": 6
  },
  {
    "id": "site property",
    "name": "javax.jdo.option.ConnectionURL",
    "displayName": "Database URL",
    "displayType": "advanced",
    "isOverridable": false,
    "category": "HIVE_METASTORE",
    "serviceName": "HIVE",
    "filename": "hive-site.xml",
    "index": 8
  },
  {
    "id": "site property",
    "name": "ambari.hive.db.schema.name",
    "displayName": "Database Name",
    "displayType": "host",
    "isOverridable": false,
    "isObserved": true,
    "serviceName": "HIVE",
    "filename": "hive-site.xml",
    "category": "HIVE_METASTORE",
    "index": 4
  },
  {
    "id": "site property",
    "name": "hive.server2.tez.default.queues",
    "displayName": "hive.server2.tez.default.queues",
    "isRequired": false,
    "serviceName": "HIVE",
    "filename": "hive-site.xml",
    "category": "Advanced hive-site"
  },
  {
    "id": "site property",
    "name": "hive.server2.thrift.port",
    "displayName": "Hive Server Port",
    "description": "TCP port number to listen on, default 10000.",
    "defaultValue": "10000",
    "displayType": "int",
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "category": "Advanced hive-site",
    "serviceName": "HIVE",
    "filename": "hive-site.xml"
  },
  {
    "id": "site property",
    "name": "hive.server2.support.dynamic.service.discovery",
    "displayName": "hive.server2.support.dynamic.service.discovery",
    "defaultValue": true,
    "displayType": "checkbox",
    "category": "Advanced hive-site",
    "serviceName": "HIVE",
    "filename": "hive-site.xml"
  },
  {
    "id": "site property",
    "name": "hive.security.authorization.enabled",
    "displayName": "hive.security.authorization.enabled",
    "defaultValue": false,
    "displayType": "checkbox",
    "category": "Advanced hive-site",
    "serviceName": "HIVE",
    "filename": "hive-site.xml"
  },
/**********************************************tez-site*****************************************/
  {
    "id": "site property",
    "name": "tez.am.resource.memory.mb",
    "displayName": "tez.am.resource.memory.mb",
    "displayType": "int",
    "category": "General",
    "serviceName": "TEZ",
    "filename": "tez-site.xml"
  },
  {
    "id": "site property",
    "name": "tez.am.java.opts",
    "displayName": "tez.am.java.opts",
    "category": "General",
    "serviceName": "TEZ",
    "filename": "tez-site.xml"
  },
  {
    "id": "site property",
    "name": "tez.am.grouping.split-waves",
    "displayName": "tez.am.grouping.split-waves",
    "displayType": "float",
    "category": "General",
    "serviceName": "TEZ",
    "filename": "tez-site.xml"
  },
  {
    "id": "site property",
    "name": "tez.am.grouping.min-size",
    "displayName": "tez.am.grouping.min-size",
    "displayType": "int",
    "category": "General",
    "serviceName": "TEZ",
    "filename": "tez-site.xml"
  },
  {
    "id": "site property",
    "name": "tez.am.grouping.max-size",
    "displayName": "tez.am.grouping.max-size",
    "displayType": "int",
    "category": "General",
    "serviceName": "TEZ",
    "filename": "tez-site.xml"
  },
  {
    "id": "site property",
    "name": "tez.am.log.level",
    "displayName": "tez.am.log.level",
    "displayType": "string",
    "category": "General",
    "serviceName": "TEZ",
    "filename": "tez-site.xml"
  },
  {
    "id": "site property",
    "name": "tez.runtime.intermediate-input.compress.codec",
    "displayName": "tez.runtime.intermediate-input.compress.codec",
    "displayType": "string",
    "category": "General",
    "serviceName": "TEZ",
    "filename": "tez-site.xml"
  },
  {
    "id": "site property",
    "name": "tez.runtime.intermediate-input.is-compressed",
    "displayName": "tez.runtime.intermediate-input.is-compressed",
    "displayType": "checkbox",
    "category": "General",
    "serviceName": "TEZ",
    "filename": "tez-site.xml"
  },
  {
    "id": "site property",
    "name": "tez.runtime.intermediate-output.compress.codec",
    "displayName": "tez.runtime.intermediate-output.compress.codec",
    "displayType": "string",
    "category": "General",
    "serviceName": "TEZ",
    "filename": "tez-site.xml"
  },
  {
    "id": "site property",
    "name": "tez.runtime.intermediate-output.should-compress",
    "displayName": "tez.runtime.intermediate-output.should-compress",
    "displayType": "checkbox",
    "category": "General",
    "serviceName": "TEZ",
    "filename": "tez-site.xml"
  },

/**********************************************hbase-site***************************************/
  {
    "id": "site property",
    "name": "hbase.tmp.dir",
    "displayName": "HBase local directory",
    "defaultDirectory": "/hadoop/hbase",
    "displayType": "directory",
    "category": "Advanced hbase-site",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.master.port",
    "displayName": "HBase Master Port",
    "isReconfigurable": true,
    "displayType": "int",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HBASE",
    "filename": "hbase-site.xml",
    "category": "Advanced hbase-site"
  },
  {
    "id": "site property",
    "name": "hbase.regionserver.global.memstore.upperLimit",
    "displayName": "hbase.regionserver.global.memstore.upperLimit",
    "displayType": "float",
    "category": "Advanced hbase-site",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.regionserver.global.memstore.lowerLimit",
    "displayName": "hbase.regionserver.global.memstore.lowerLimit",
    "displayType": "float",
    "category": "Advanced hbase-site",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.hstore.blockingStoreFiles",
    "displayName": "hstore blocking storefiles",
    "displayType": "int",
    "category": "Advanced hbase-site",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.hstore.compactionThreshold",
    "displayName": "HBase HStore compaction threshold",
    "displayType": "int",
    "category": "General",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml",
    "index": 0
  },
  {
    "id": "site property",
    "name": "hfile.block.cache.size",
    "displayName": "HFile block cache size ",
    "displayType": "float",
    "category": "General",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml",
    "index": 1
  },
  {
    "id": "site property",
    "name": "hbase.hregion.max.filesize",
    "displayName": "Maximum HStoreFile Size",
    "displayType": "int",
    "unit": "bytes",
    "category": "General",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml",
    "index": 2
  },
  {
    "id": "site property",
    "name": "hbase.regionserver.handler.count",
    "displayName": "RegionServer Handler",
    "displayType": "int",
    "category": "HBASE_REGIONSERVER",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml",
    "index": 2
  },
  {
    "id": "site property",
    "name": "hbase.hregion.majorcompaction",
    "displayName": "HBase Region Major Compaction",
    "displayType": "int",
    "unit": "ms",
    "category": "HBASE_REGIONSERVER",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml",
    "index": 3
  },
  {
    "id": "site property",
    "name": "hbase.hregion.memstore.block.multiplier",
    "displayName": "HBase Region Block Multiplier",
    "displayType": "int",
    "category": "HBASE_REGIONSERVER",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml",
    "index": 4
  },
  {
    "id": "site property",
    "name": "hbase.hregion.memstore.mslab.enabled",
    "displayName": "hbase.hregion.memstore.mslab.enabled",
    "displayType": "checkbox",
    "category": "Advanced hbase-site",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.hregion.memstore.flush.size",
    "displayName": "HBase Region Memstore Flush Size",
    "displayType": "int",
    "unit": "bytes",
    "category": "HBASE_REGIONSERVER",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml",
    "index": 5
  },
  {
    "id": "site property",
    "name": "hbase.client.scanner.caching",
    "displayName": "HBase Client Scanner Caching",
    "displayType": "int",
    "unit": "rows",
    "category": "General",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml",
    "index": 3
  },
  {
    "id": "site property",
    "name": "zookeeper.session.timeout",
    "displayName": "Zookeeper timeout for HBase Session",
    "displayType": "int",
    "unit": "ms",
    "category": "General",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml",
    "index": 4
  },
  {
    "id": "site property",
    "name": "hbase.client.keyvalue.maxsize",
    "displayName": "HBase Client Maximum key-value Size",
    "displayType": "int",
    "unit": "bytes",
    "category": "General",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml",
    "index": 5
  },
  {
    "id": "site property",
    "name": "hbase.coprocessor.region.classes",
    "displayName": "hbase.coprocessor.region.classes",
    "category": "Advanced hbase-site",
    "isRequired": false,
    "serviceName": "HBASE",
    "filename": "hbase-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.coprocessor.master.classes",
    "displayName": "hbase.coprocessor.master.classes",
    "category": "Advanced hbase-site",
    "isRequired": false,
    "serviceName": "HBASE",
    "filename": "hbase-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.zookeeper.quorum",
    "displayName": "hbase.zookeeper.quorum",
    "displayType": "multiLine",
    "serviceName": "HBASE",
    "filename": "hbase-site.xml",
    "category": "Advanced hbase-site"
  },

/**********************************************storm-site***************************************/
  {
    "id": "site property",
    "name": "storm.zookeeper.root",
    "displayName": "storm.zookeeper.root",
    "displayType": "directory",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "General"
  },
  {
    "id": "site property",
    "name": "storm.local.dir",
    "displayName": "storm.local.dir",
    "defaultDirectory": "/hadoop/storm",
    "displayType": "directory",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "General"
  },
  {
    "id": "site property",
    "name": "storm.zookeeper.servers",
    "displayName": "storm.zookeeper.servers",
    "displayType": "masterHosts",
    "isOverridable": false,
    "isReconfigurable": false,
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "General"
  },
  {
    "id": "site property",
    "name": "storm.zookeeper.port",
    "displayName": "storm.zookeeper.port",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "General"
  },
  {
    "id": "site property",
    "name": "storm.zookeeper.session.timeout",
    "displayName": "storm.zookeeper.session.timeout",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "General"
  },
  {
    "id": "site property",
    "name": "storm.zookeeper.connection.timeout",
    "displayName": "storm.zookeeper.connection.timeout",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "General"
  },
  {
    "id": "site property",
    "name": "storm.zookeeper.retry.times",
    "displayName": "storm.zookeeper.retry.times",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "General"
  },
  {
    "id": "site property",
    "name": "storm.zookeeper.retry.interval",
    "displayName": "storm.zookeeper.retry.interval",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "General",
    "unit": "ms"
  },
  {
    "id": "site property",
    "name": "storm.zookeeper.retry.intervalceiling.millis",
    "displayName": "storm.zookeeper.retry.intervalceiling.millis",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "General",
    "unit": "ms"
  },
  {
    "id": "site property",
    "name": "storm.cluster.mode",
    "displayName": "storm.cluster.mode",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "General"
  },
  {
    "id": "site property",
    "name": "storm.local.mode.zmq",
    "displayName": "storm.local.mode.zmq",
    "displayType": "checkbox",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "General"
  },
  {
    "id": "site property",
    "name": "storm.thrift.transport",
    "displayName": "storm.thrift.transport",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "General"
  },
  {
    "id": "site property",
    "name": "storm.messaging.transport",
    "displayName": "storm.messaging.transport",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "General"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "General",
    "displayName": "storm.messaging.netty.buffer_size",
    "name": "storm.messaging.netty.buffer_size",
    "displayType": "int",
    "unit": "bytes",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "General",
    "displayName": "storm.messaging.netty.max_retries",
    "name": "storm.messaging.netty.max_retries",
    "displayType": "int",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "General",
    "displayName": "storm.messaging.netty.max_wait_ms",
    "name": "storm.messaging.netty.max_wait_ms",
    "displayType": "int",
    "unit": "ms",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "General",
    "displayName": "storm.messaging.netty.min_wait_ms",
    "name": "storm.messaging.netty.min_wait_ms",
    "displayType": "int",
    "unit": "ms",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "General",
    "displayName": "storm.messaging.netty.server_worker_threads",
    "name": "storm.messaging.netty.server_worker_threads",
    "displayType": "int",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "General",
    "displayName": "storm.messaging.netty.client_worker_threads",
    "name": "storm.messaging.netty.client_worker_threads",
    "displayType": "int",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "name": "nimbus.host",
    "displayName": "nimbus.host",
    "displayType": "masterHost",
    "isOverridable": false,
    "isReconfigurable": false,
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "NIMBUS"
  },
  {
    "id": "site property",
    "name": "nimbus.thrift.port",
    "displayName": "nimbus.thrift.port",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "NIMBUS"
  },
  {
    "id": "site property",
    "name": "nimbus.thrift.max_buffer_size",
    "displayName": "nimbus.thrift.max_buffer_size",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "NIMBUS",
    "unit": "bytes"
  },
  {
    "id": "site property",
    "name": "nimbus.childopts",
    "displayName": "nimbus.childopts",
    "displayType": "multiLine",
    "isOverridable": false,
    "serviceName": "STORM",
    "category": "NIMBUS",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "name": "nimbus.task.timeout.secs",
    "displayName": "nimbus.task.timeout.secs",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "NIMBUS",
    "unit": "seconds"
  },
  {
    "id": "site property",
    "name": "nimbus.supervisor.timeout.secs",
    "displayName": "nimbus.supervisor.timeout.secs",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "NIMBUS",
    "unit": "seconds"
  },
  {
    "id": "site property",
    "name": "nimbus.monitor.freq.secs",
    "displayName": "nimbus.monitor.freq.secs",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "NIMBUS",
    "unit": "seconds"
  },
  {
    "id": "site property",
    "name": "nimbus.cleanup.inbox.freq.secs",
    "displayName": "nimbus.cleanup.inbox.freq.secs",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "NIMBUS",
    "unit": "seconds"
  },
  {
    "id": "site property",
    "name": "nimbus.inbox.jar.expiration.secs",
    "displayName": "nimbus.inbox.jar.expiration.secs",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "NIMBUS",
    "unit": "seconds"
  },
  {
    "id": "site property",
    "name": "nimbus.task.launch.secs",
    "displayName": "nimbus.task.launch.secs",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "NIMBUS",
    "unit": "seconds"
  },
  {
    "id": "site property",
    "name": "nimbus.reassign",
    "displayName": "nimbus.reassign",
    "displayType": "checkbox",
    "isReconfigurable": true,
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "NIMBUS"
  },
  {
    "id": "site property",
    "name": "nimbus.file.copy.expiration.secs",
    "displayName": "nimbus.file.copy.expiration.secs",
    "displayType": "int",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "NIMBUS",
    "unit": "seconds"
  },
  {
    "id": "site property",
    "name": "nimbus.topology.validator",
    "displayName": "nimbus.topology.validator",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "NIMBUS"
  },
  {
    "id": "site property",
    "name": "supervisor.slots.ports",
    "displayName": "supervisor.slots.ports",
    "displayType": "string",
    "serviceName": "STORM",
    "filename": "storm-site.xml",
    "category": "SUPERVISOR"
  },
  {
    "id": "site property",
    "isOverrideable": false,
    "serviceName": "STORM",
    "category": "SUPERVISOR",
    "displayName": "supervisor.childopts",
    "name": "supervisor.childopts",
    "displayType": "multiLine",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "SUPERVISOR",
    "displayName": "supervisor.worker.start.timeout.secs",
    "name": "supervisor.worker.start.timeout.secs",
    "displayType": "int",
    "unit": "seconds",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "SUPERVISOR",
    "displayName": "supervisor.worker.timeout.secs",
    "name": "supervisor.worker.timeout.secs",
    "displayType": "int",
    "unit": "seconds",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "SUPERVISOR",
    "displayName": "supervisor.monitor.frequency.secs",
    "name": "supervisor.monitor.frequency.secs",
    "displayType": "int",
    "unit": "seconds",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "SUPERVISOR",
    "displayName": "supervisor.heartbeat.frequency.secs",
    "name": "supervisor.heartbeat.frequency.secs",
    "displayType": "int",
    "unit": "seconds",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "DRPC_SERVER",
    "displayName": "drpc.port",
    "name": "drpc.port",
    "displayType": "int",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "DRPC_SERVER",
    "displayName": "drpc.worker.threads",
    "name": "drpc.worker.threads",
    "displayType": "int",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "DRPC_SERVER",
    "displayName": "drpc.queue.size",
    "name": "drpc.queue.size",
    "displayType": "int",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "DRPC_SERVER",
    "displayName": "drpc.invocations.port",
    "name": "drpc.invocations.port",
    "displayType": "int",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "DRPC_SERVER",
    "displayName": "drpc.request.timeout.secs",
    "name": "drpc.request.timeout.secs",
    "displayType": "int",
    "unit": "seconds",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "DRPC_SERVER",
    "displayName": "drpc.childopts",
    "name": "drpc.childopts",
    "displayType": "string",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "STORM_UI_SERVER",
    "displayName": "ui.port",
    "name": "ui.port",
    "displayType": "int",
    "filename": "storm-site.xml"
  },
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "STORM_UI_SERVER",
    "displayName": "ui.childopts",
    "name": "ui.childopts",
    "displayType": "string",
    "filename": "storm-site.xml"
  },
  //@Todo: uncomment following properties when logviewer is treated as different section on storm service page
  /*
   {
   "id": "site property",
   "serviceName": "STORM",
   "category": "LogviewerServer",
   "displayName": "logviewer.port",
   "name": "logviewer.port",
   "displayType": "int"
   },
   {
   "id": "site property",
   "serviceName": "STORM",
   "category": "LogviewerServer",
   "displayName": "logviewer.childopts",
   "name": "logviewer.childopts",
   "displayType": "string"
   },
   {
   "id": "site property",
   "serviceName": "STORM",
   "category": "LogviewerServer",
   "displayName": "logviewer.appender.name",
   "name": "logviewer.appender.name",
   "displayType": "string"
   },
   */
  {
    "id": "site property",
    "serviceName": "STORM",
    "category": "Advanced storm-site",
    "displayName": "worker.childopts",
    "name": "worker.childopts",
    "displayType": "multiLine",
    "filename": "storm-site.xml"
  },
/*********************************************oozie-site for Falcon*****************************/
  {
    "id": "site property",
    "isReconfigurable": true,
    "serviceName": "FALCON",
    "category": "Falcon - Oozie integration",
    "displayName": "oozie.service.ELService.ext.functions.coord-job-submit-instances",
    "name": "oozie.service.ELService.ext.functions.coord-job-submit-instances",
    "displayType": "custom",
    "filename": "oozie-site.xml"
  },
  {
    "id": "site property",
    "isReconfigurable": true,
    "serviceName": "FALCON",
    "category": "Falcon - Oozie integration",
    "displayName": "oozie.service.ELService.ext.functions.coord-action-create-inst",
    "name": "oozie.service.ELService.ext.functions.coord-action-create-inst",
    "displayType": "custom",
    "filename": "oozie-site.xml"
  },
  {
    "id": "site property",
    "isReconfigurable": true,
    "serviceName": "FALCON",
    "category": "Falcon - Oozie integration",
    "displayName": "oozie.service.ELService.ext.functions.coord-action-create",
    "name": "oozie.service.ELService.ext.functions.coord-action-create",
    "displayType": "custom",
    "filename": "oozie-site.xml"
  },
  {
    "id": "site property",
    "isReconfigurable": true,
    "serviceName": "FALCON",
    "category": "Falcon - Oozie integration",
    "displayName": "oozie.service.ELService.ext.functions.coord-job-submit-data",
    "name": "oozie.service.ELService.ext.functions.coord-job-submit-data",
    "displayType": "custom",
    "filename": "oozie-site.xml"
  },
  {
    "id": "site property",
    "isReconfigurable": true,
    "serviceName": "FALCON",
    "category": "Falcon - Oozie integration",
    "displayName": "oozie.service.ELService.ext.functions.coord-action-start",
    "name": "oozie.service.ELService.ext.functions.coord-action-start",
    "displayType": "custom",
    "filename": "oozie-site.xml"
  },
  {
    "id": "site property",
    "isReconfigurable": true,
    "serviceName": "FALCON",
    "category": "Falcon - Oozie integration",
    "displayName": "oozie.service.ELService.ext.functions.coord-sla-submit",
    "name": "oozie.service.ELService.ext.functions.coord-sla-submit",
    "displayType": "custom",
    "filename": "oozie-site.xml"
  },
  {
    "id": "site property",
    "isReconfigurable": true,
    "serviceName": "FALCON",
    "category": "Falcon - Oozie integration",
    "displayName": "oozie.service.ELService.ext.functions.coord-sla-create",
    "name": "oozie.service.ELService.ext.functions.coord-sla-create",
    "displayType": "custom",
    "filename": "oozie-site.xml"
  },

  // Runtime properties
  {
    "id": "site property",
    "name": "*.domain",
    "displayName": "*.domain",
    "category": "FalconRuntimeSite",
    "serviceName": "FALCON",
    "filename": "falcon-runtime.properties.xml"

  },
  {
    "id": "site property",
    "name": "*.log.cleanup.frequency.minutes.retention",
    "displayName": "*.log.cleanup.frequency.minutes.retention",
    "category": "FalconRuntimeSite",
    "serviceName": "FALCON",
    "filename": "falcon-runtime.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.log.cleanup.frequency.hours.retention",
    "displayName": "*.log.cleanup.frequency.hours.retention",
    "category": "FalconRuntimeSite",
    "serviceName": "FALCON",
    "filename": "falcon-runtime.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.log.cleanup.frequency.days.retention",
    "displayName": "*.log.cleanup.frequency.days.retention",
    "category": "FalconRuntimeSite",
    "serviceName": "FALCON",
    "filename": "falcon-runtime.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.log.cleanup.frequency.months.retention",
    "displayName": "*.log.cleanup.frequency.months.retention",
    "category": "FalconRuntimeSite",
    "serviceName": "FALCON",
    "filename": "falcon-runtime.properties.xml"
  },

  //  Startup properties

  {
    "id": "site property",
    "name": "*.domain",
    "displayName": "*.domain",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.workflow.engine.impl",
    "displayName": "*.workflow.engine.impl",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.oozie.process.workflow.builder",
    "displayName": "*.oozie.process.workflow.builder",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.oozie.feed.workflow.builder",
    "displayName": "*.oozie.feed.workflow.builder",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.SchedulableEntityManager.impl",
    "displayName": "*.SchedulableEntityManager.impl",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.ConfigSyncService.impl",
    "displayName": "*.ConfigSyncService.impl",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.ProcessInstanceManager.impl",
    "displayName": "*.ProcessInstanceManager.impl",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.catalog.service.impl",
    "displayName": "*.catalog.service.impl",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.application.services",
    "displayName": "*.application.services",
    "displayType": "multiLine",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.configstore.listeners",
    "displayName": "*.configstore.listeners",
    "displayType": "multiLine",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.broker.impl.class",
    "displayName": "*.broker.impl.class",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.shared.libs",
    "displayName": "*.shared.libs",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.config.store.uri",
    "displayName": "*.config.store.uri",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.system.lib.location",
    "displayName": "*.system.lib.location",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.retry.recorder.path",
    "displayName": "*.retry.recorder.path",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.falcon.cleanup.service.frequency",
    "displayName": "*.falcon.cleanup.service.frequency",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.broker.url",
    "displayName": "*.broker.url",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.broker.ttlInMins",
    "displayName": "*.broker.ttlInMins",
    "displayType": "int",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.entity.topic",
    "displayName": "*.entity.topic",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.max.retry.failure.count",
    "displayName": "*.max.retry.failure.count",
    "displayType": "int",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.internal.queue.size",
    "displayName": "*.internal.queue.size",
    "displayType": "int",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.falcon.authentication.type",
    "displayName": "*.falcon.authentication.type",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.falcon.http.authentication.type",
    "displayName": "*.falcon.http.authentication.type",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.falcon.http.authentication.token.validity",
    "displayName": "*.falcon.http.authentication.token.validity",
    "displayType": "int",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.falcon.http.authentication.signature.secret",
    "displayName": "*.falcon.http.authentication.signature.secret",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.falcon.http.authentication.simple.anonymous.allowed",
    "displayName": "*.falcon.http.authentication.simple.anonymous.allowed",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.falcon.http.authentication.kerberos.name.rules",
    "displayName": "*.falcon.http.authentication.kerberos.name.rules",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.falcon.http.authentication.blacklisted.users",
    "displayName": "*.falcon.http.authentication.blacklisted.users",
    "isRequired": false,
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },

  // Falcon Graph and Storage
  {
    "id": "site property",
    "name": "*.falcon.graph.storage.directory",
    "displayName": "*.falcon.graph.storage.directory",
    "defaultDirectory": "/hadoop/falcon/data/lineage/graphdb",
    "displayType": "directory",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.falcon.graph.serialize.path",
    "displayName": "*.falcon.graph.serialize.path",
    "defaultDirectory": "/hadoop/falcon/data/lineage",
    "displayType": "directory",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },
  {
    "id": "site property",
    "name": "*.falcon.graph.preserve.history",
    "displayName": "*.falcon.graph.preserve.history",
    "defaultValue": false,
    "displayType": "checkbox",
    "category": "FalconStartupSite",
    "serviceName": "FALCON",
    "filename": "falcon-startup.properties.xml"
  },

/**********************************************webhcat-site***************************************/
  {
    "id": "site property",
    "name": "templeton.hive.archive",
    "displayName": "templeton.hive.archive",
    "isRequired": false,
    "serviceName": "HIVE",
    "filename": "webhcat-site.xml",
    "category": "Advanced webhcat-site"
  },
  {
    "id": "site property",
    "name": "templeton.pig.archive",
    "displayName": "templeton.pig.archive",
    "isRequired": false,
    "serviceName": "HIVE",
    "filename": "webhcat-site.xml",
    "category": "Advanced webhcat-site"
  },
  {
    "id": "site property",
    "name": "templeton.zookeeper.hosts",
    "displayName": "templeton.zookeeper.hosts",
    "displayType": "multiLine",
    "serviceName": "HIVE",
    "filename": "webhcat-site.xml",
    "category": "Advanced webhcat-site"
  },
/**********************************************pig.properties*****************************************/
  {
    "id": "site property",
    "name": "content",
    "displayName": "content",
    "value": "",
    "defaultValue": "",
    "description": "pig properties",
    "displayType": "content",
    "isRequired": false,
    "showLabel": false,
    "serviceName": "PIG",
    "filename": "pig-properties.xml",
    "category": "Advanced pig-properties"
  },

/**********************************************KNOX*****************************************/
  {
    "id": "site property",
    "name": "content",
    "displayName": "content",
    "value": "",
    "defaultValue": "",
    "displayType": "content",
    "isRequired": false,
    "showLabel": false,
    "serviceName": "KNOX",
    "filename": "topology.xml",
    "category": "Advanced topology"
  },

  {
    "id": "site property",
    "name": "content",
    "displayName": "content",
    "value": "",
    "defaultValue": "",
    "displayType": "content",
    "isRequired": false,
    "showLabel": false,
    "serviceName": "KNOX",
    "filename": "users-ldif.xml",
    "category": "Advanced users-ldif"
  },
  {
    "id": "puppet var",
    "name": "knox_gateway_host",
    "displayName": "Knox Gateway host",
    "value": "",
    "defaultValue": "",
    "description": "The host that has been assigned to run Knox Gateway",
    "displayType": "masterHost",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": false,
    "serviceName": "KNOX",
    "filename": "knox-env.xml",
    "category": "KNOX_GATEWAY",
    "index": 0
  },
  {
    "id": "site property",
    "name": "knox_master_secret",
    "displayName": "Knox Master Secret",
    "value": "",
    "defaultValue": "",
    "displayType": "password",
    "isReconfigurable": false,
    "isOverridable": false,
    "isRequired": true,
    "serviceName": "KNOX",
    "filename": "knox-env.xml",
    "category": "KNOX_GATEWAY"
  },
  {
    "id": "puppet var",
    "name": "knox_pid_dir",
    "displayName": "Knox PID dir",
    "value": "",
    "displayType": "directory",
    "isReconfigurable": false,
    "isOverridable": false,
    "serviceName": "KNOX",
    "filename": "knox-env.xml",
    "category": "Advanced knox-env"
  },

/********************************************* KAFKA *****************************/
  {
    "id": "puppet var",
    "name": "kafka_broker_hosts",
    "displayName": "Kafka Broker host",
    "value": "",
    "defaultValue": "",
    "description": "The host that has been assigned to run Kafka Broker",
    "displayType": "masterHosts",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": false,
    "serviceName": "KAFKA",
    "filename": "kafka-broker.xml",
    "category": "KAFKA_BROKER",
    "index": 0
  },
  {
    "id": "site property",
    "name": "log.dirs",
    "displayName": "log.dirs",
    "value": "",
    "defaultValue": "",
    "defaultDirectory": "/kafka-logs",
    "displayType": "directories",
    "serviceName": "KAFKA",
    "filename": "kafka-broker.xml",
    "category": "KAFKA_BROKER",
    "index": 0
  },
  {
    "id": "site property",
    "name": "port",
    "displayName": "port",
    "value": "",
    "defaultValue": "",
    "displayType": "int",
    "serviceName": "KAFKA",
    "filename": "kafka-broker.xml",
    "category": "KAFKA_BROKER",
    "index": 0
  },
  {
    "id": "site property",
    "name": "log.roll.hours",
    "displayName": "log.roll.hours",
    "value": "",
    "defaultValue": "",
    "displayType": "advanced",
    "serviceName": "KAFKA",
    "filename": "kafka-broker.xml",
    "category": "KAFKA_BROKER",
    "index": 0
  },
  {
    "id": "site property",
    "name": "log.retention.hours",
    "displayName": "log.retention.hours",
    "value": "",
    "defaultValue": "",
    "displayType": "advanced",
    "serviceName": "KAFKA",
    "filename": "kafka-broker.xml",
    "category": "KAFKA_BROKER",
    "index": 0
  },
  {
    "id": "site property",
    "name": "zookeeper.connect",
    "displayName": "zookeeper.connect",
    "value": "",
    "defaultValue": "",
    "displayType": "advanced",
    "serviceName": "KAFKA",
    "filename": "kafka-broker.xml",
    "category": "KAFKA_BROKER",
    "index": 0
  },
  {
    "id": "site property",
    "name": "kafka_pid_dir",
    "displayName": "Kafka PID dir",
    "value": "",
    "defaultValue": "",
    "isReconfigurable": false,
    "isOverridable": false,
    "displayType": "directory",
    "serviceName": "KAFKA",
    "filename": "kafka-env.xml",
    "category": "Advanced kafka-env",
    "index": 0
  },

/*******************************************kerberos***********************************/
  {
    "id": "puppet var",
    "name": "kdc_type",
    "displayName": "KDC type",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": true,
    "displayType": "masterHost",
    "serviceName": "KERBEROS",
    "filename": "kerberos-env.xml",
    "category": "KDC",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "kdc_host",
    "displayName": "KDC host",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": true,
    "displayType": "supportTextConnection",
    "serviceName": "KERBEROS",
    "filename": "kerberos-env.xml",
    "category": "KDC",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "realm",
    "displayName": "Realm name",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": true,
    "isReconfigurable": false,
    "serviceName": "KERBEROS",
    "filename": "kerberos-env.xml",
    "category": "KDC",
    "index": 2
  },
  {
    "id": "puppet var",
    "name": "ldap_url",
    "displayName": "LDAP url",
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "KERBEROS",
    "filename": "kerberos-env.xml",
    "category": "KDC",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "container_dn",
    "displayName": "Container DN",
    "isVisible": false,
    "serviceName": "KERBEROS",
    "filename": "kerberos-env.xml",
    "category": "KDC",
    "index": 4
  },
  {
    "id": "puppet var",
    "name": "create_attributes_template",
    "displayName": "Attribute template",
    "serviceName": "KERBEROS",
    "filename": "kerberos-env.xml",
    "displayType": "content",
    "category": "Advanced kerberos-env"
  },
  {
    "id": "puppet var",
    "name": "encryption_types",
    "displayName": "Encryption Types",
    "serviceName": "KERBEROS",
    "filename": "kerberos-env.xml",
    "displayType": "multiLine",
    "category": "Advanced kerberos-env"
  },
  {
    "id": "puppet var",
    "name": "domains",
    "displayName": "Domains",
    "isRequired": false,
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": true,
    "serviceName": "KERBEROS",
    "filename": "krb5-conf.xml",
    "category": "KDC",
    "index": 5
  },
  {
    "id": "puppet var",
    "name": "admin_server_host",
    "displayName": "Kadmin host",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": true,
    "serviceName": "KERBEROS",
    "filename": "kerberos-env.xml",
    "category": "Kadmin",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "admin_principal",
    "displayName": "Admin principal",
    "description": "Admin principal used to create principals and export key tabs (e.g. admin/admin@EXAMPLE.COM).",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": false,
    "serviceName": "KERBEROS",
    "filename": "krb5-conf.xml",
    "category": "Kadmin",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "admin_password",
    "displayName": "Admin password",
    "displayType": "password",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": false,
    "serviceName": "KERBEROS",
    "filename": "krb5-conf.xml",
    "category": "Kadmin",
    "index": 2
  },
  {
    "id": "puppet var",
    "name": "manage_krb5_conf",
    "displayName": "Manage Kerberos client krb5.conf",
    "displayType": "checkbox",
    "dependentConfigPattern": "CATEGORY",
    "serviceName": "KERBEROS",
    "filename": "krb5-conf.xml",
    "category": "Advanced krb5-conf",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "conf_dir",
    "displayName": "krb5-conf directory path",
    "value": "",
    "defaultValue": "",
    "description": "",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": true,
    "displayType": "directory",
    "serviceName": "KERBEROS",
    "filename": "krb5-conf.xml",
    "category": "Advanced krb5-conf",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "content",
    "displayName": "krb5-conf template",
    "value": "",
    "defaultValue": "",
    "description": "",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": true,
    "displayType": "content",
    "serviceName": "KERBEROS",
    "filename": "krb5-conf.xml",
    "category": "Advanced krb5-conf",
    "index": 2
  },
/********************************************* flume-agent *****************************/
  {
    "id": "site property",
    "name": "content",
    "displayName": "content",
    "showLabel": false,
    "isRequired": false,
    "displayType": "content",
    "serviceName": "FLUME",
    "category": "FLUME_HANDLER",
    "filename": "flume-conf.xml"
  },
  {
    "id": "puppet var",
    "name": "flume_conf_dir",
    "displayName": "Flume Conf Dir",
    "description": "Location to save configuration files",
    "defaultValue": "/etc/flume/conf",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "FLUME",
    "filename": "flume-env.xml",
    "category": "Advanced flume-env",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "flume_log_dir",
    "displayName": "Flume Log Dir",
    "description": "Location to save log files",
    "defaultValue": "/var/log/flume",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "FLUME",
    "filename": "flume-env.xml",
    "category": "Advanced flume-env",
    "index": 1
  },
/**********************************************HDFS***************************************/
  {
    "id": "puppet var",
    "name": "namenode_host",
    "displayName": "NameNode hosts",
    "value": "",
    "defaultValue": "",
    "description": "The hosts that has been assigned to run NameNode",
    "displayType": "masterHosts",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": false,
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "category": "NAMENODE",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "namenode_heapsize",
    "displayName": "NameNode Java heap size",
    "description": "Initial and maximum Java heap size for NameNode (Java options -Xms and -Xmx).  This also applies to the Secondary NameNode.",
    "defaultValue": "1024",
    "displayType": "int",
    "unit": "MB",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "category": "NAMENODE",
    "index": 2
  },
  {
    "id": "puppet var",
    "name": "namenode_opt_newsize",
    "displayName": "NameNode new generation size",
    "description": "Default size of Java new generation for NameNode (Java option -XX:NewSize).  This also applies to the Secondary NameNode.",
    "defaultValue": "200",
    "displayType": "int",
    "unit": "MB",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "category": "NAMENODE",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "namenode_opt_permsize",
    "displayName": "NameNode permanent generation size",
    "description": "Default size of Java permanent generation for NameNode (Java option -XX:PermSize).  This also applies to the Secondary NameNode.",
    "defaultValue": "128",
    "displayType": "int",
    "unit": "MB",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "category": "NAMENODE",
    "index": 5
  },
  {
    "id": "puppet var",
    "name": "namenode_opt_maxpermsize",
    "displayName": "NameNode maximum permanent generation size",
    "description": "Maximum size of Java permanent generation for NameNode (Java option -XX:MaxPermSize).",
    "defaultValue": "256",
    "displayType": "int",
    "unit": "MB",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "category": "NAMENODE",
    "index": 6
  },
  {
    "id": "puppet var",
    "name": "namenode_opt_maxnewsize",
    "displayName": "NameNode maximum new generation size",
    "description": "Maximum size of Java new generation for NameNode (Java option -XX:MaxnewSize).",
    "defaultValue": "200",
    "displayType": "int",
    "unit": "MB",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "category": "NAMENODE",
    "index": 4
  },
  {
    "id": "puppet var",
    "name": "snamenode_host",
    "displayName": "SNameNode host",
    "value": "",
    "defaultValue": "",
    "description": "The host that has been assigned to run SecondaryNameNode",
    "displayType": "masterHost",
    "isOverridable": false,
    "isRequiredByAgent": false,
    "isVisible": true,
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "category": "SECONDARY_NAMENODE",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "datanode_hosts", //not in the schema. For UI purpose
    "displayName": "DataNode hosts",
    "value": "",
    "defaultValue": "",
    "description": "The hosts that have been assigned to run DataNode",
    "displayType": "slaveHosts",
    "isRequired": false,
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": false,
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "category": "DATANODE",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "dtnode_heapsize",
    "displayName": "DataNode maximum Java heap size",
    "description": "Maximum Java heap size for DataNode (Java option -Xmx)",
    "defaultValue": "1024",
    "displayType": "int",
    "unit": "MB",
    "isVisible": true,
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "category": "DATANODE",
    "index": 2
  },
  {
    "id": "puppet var",
    "name": "hadoop_heapsize",
    "displayName": "Hadoop maximum Java heap size",
    "description": "Maximum Java heap size for daemons such as Balancer (Java option -Xmx)",
    "defaultValue": "1024",
    "displayType": "int",
    "unit": "MB",
    "isVisible": true,
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "category": "General",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "hdfs_log_dir_prefix",
    "displayName": "Hadoop Log Dir Prefix",
    "description": "The parent directory for Hadoop log files.  The HDFS log directory will be ${hadoop_log_dir_prefix} / ${hdfs_user} and the MapReduce log directory will be ${hadoop_log_dir_prefix} / ${mapred_user}.",
    "defaultValue": "/var/log/hadoop",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "category": "Advanced hadoop-env"
  },
  {
    "id": "puppet var",
    "name": "hadoop_pid_dir_prefix",
    "displayName": "Hadoop PID Dir Prefix",
    "description": "The parent directory in which the PID files for Hadoop processes will be created.  The HDFS PID directory will be ${hadoop_pid_dir_prefix} / ${hdfs_user} and the MapReduce PID directory will be ${hadoop_pid_dir_prefix} / ${mapred_user}.",
    "defaultValue": "/var/run/hadoop",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "category": "Advanced hadoop-env"
  },
  {
    "id": "puppet var",
    "name": "hadoop_root_logger",
    "displayName": "Hadoop Root Logger",
    "description": "Hadoop logging options",
    "defaultValue": "INFO,RFA",
    "displayType": "string",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HDFS",
    "filename": "hadoop-env.xml",
    "category": "Advanced hadoop-env"
  },

/**********************************************MAPREDUCE2***************************************/
  {
    "id": "puppet var",
    "name": "hs_host",
    "displayName": "History Server",
    "description": "History Server",
    "defaultValue": "",
    "isOverridable": false,
    "displayType": "masterHost",
    "isRequiredByAgent": false,
    "isVisible": true,
    "serviceName": "MAPREDUCE2",
    "filename": "mapred-env.xml",
    "category": "HISTORYSERVER",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "jobhistory_heapsize",
    "displayName": "History Server heap size",
    "description": "History Server heap size",
    "defaultValue": "900",
    "unit": "MB",
    "isOverridable": true,
    "displayType": "int",
    "isVisible": true,
    "serviceName": "MAPREDUCE2",
    "filename": "mapred-env.xml",
    "category": "HISTORYSERVER",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "mapred_log_dir_prefix",
    "displayName": "Mapreduce Log Dir Prefix",
    "description": "",
    "defaultValue": "/var/log/hadoop-mapreduce",
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "isReconfigurable": false,
    "serviceName": "MAPREDUCE2",
    "filename": "mapred-env.xml",
    "category": "Advanced mapred-env"
  },
  {
    "id": "puppet var",
    "name": "mapred_pid_dir_prefix",
    "displayName": "Mapreduce PID Dir Prefix",
    "description": "",
    "defaultValue": "/var/run/hadoop-mapreduce",
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "isReconfigurable": false,
    "serviceName": "MAPREDUCE2",
    "filename": "mapred-env.xml",
    "category": "Advanced mapred-env"
  },
/**********************************************YARN***************************************/
  {
    "id": "puppet var",
    "name": "yarn_heapsize",
    "displayName": "YARN Java heap size",
    "description": "Max heapsize for all YARN components",
    "defaultValue": "1024",
    "isOverridable": true,
    "displayType": "int",
    "unit": "MB",
    "isVisible": true,
    "serviceName": "YARN",
    "filename": "yarn-env.xml",
    "category": "General",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "rm_host",
    "displayName": "ResourceManager",
    "description": "ResourceManager",
    "defaultValue": "",
    "isOverridable": false,
    "displayType": "masterHost",
    "isRequiredByAgent": false,
    "isVisible": true,
    "serviceName": "YARN",
    "filename": "yarn-env.xml",
    "category": "RESOURCEMANAGER",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "resourcemanager_heapsize",
    "displayName": "ResourceManager Java heap size",
    "description": "Max heapsize for ResourceManager",
    "defaultValue": "1024",
    "isOverridable": false,
    "displayType": "int",
    "unit": "MB",
    "isVisible": true,
    "serviceName": "YARN",
    "filename": "yarn-env.xml",
    "category": "RESOURCEMANAGER",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "nm_hosts",
    "displayName": "NodeManager",
    "description": "List of NodeManager Hosts.",
    "defaultValue": "",
    "isOverridable": false,
    "displayType": "slaveHosts",
    "isRequiredByAgent": false,
    "isVisible": true,
    "serviceName": "YARN",
    "filename": "yarn-env.xml",
    "category": "NODEMANAGER",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "nodemanager_heapsize",
    "displayName": "NodeManager Java heap size",
    "description": "Max heapsize for NodeManager",
    "defaultValue": "1024",
    "isOverridable": true,
    "displayType": "int",
    "unit": "MB",
    "isVisible": true,
    "serviceName": "YARN",
    "filename": "yarn-env.xml",
    "category": "NODEMANAGER",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "yarn_log_dir_prefix",
    "displayName": "YARN Log Dir Prefix",
    "description": "",
    "defaultValue": "/var/log/hadoop-yarn",
    "displayType": "directory",
    "isOverridable": false,
    "isReconfigurable": false,
    "isVisible": true,
    "serviceName": "YARN",
    "filename": "yarn-env.xml",
    "category": "Advanced yarn-env"
  },
  {
    "id": "puppet var",
    "name": "yarn_pid_dir_prefix",
    "displayName": "YARN PID Dir Prefix",
    "description": "",
    "defaultValue": "/var/run/hadoop-yarn",
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "isReconfigurable": false,
    "serviceName": "YARN",
    "filename": "yarn-env.xml",
    "category": "Advanced yarn-env"
  },
  {
    "id": "puppet var",
    "name": "min_user_id",
    "displayName": "Minimum user ID for submitting job",
    "isOverridable": true,
    "displayType": "int",
    "isVisible": true,
    "serviceName": "YARN",
    "filename": "yarn-env.xml",
    "category": "Advanced yarn-env"
  },
  {
    "id": "puppet var",
    "name": "ats_host",
    "displayName": "App Timeline Server",
    "description": "Application Timeline Server Host",
    "defaultValue": "",
    "isOverridable": false,
    "displayType": "masterHost",
    "isRequiredByAgent": false,
    "isVisible": true,
    "serviceName": "YARN",
    "filename": "yarn-env.xml",
    "category": "APP_TIMELINE_SERVER",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "apptimelineserver_heapsize",
    "displayName": "AppTimelineServer Java heap size",
    "description": "AppTimelineServer Java heap size",
    "defaultValue": "1024",
    "isOverridable": false,
    "displayType": "int",
    "unit": "MB",
    "isVisible": true,
    "serviceName": "YARN",
    "filename": "yarn-env.xml",
    "category": "APP_TIMELINE_SERVER",
    "index": 1
  },
/**********************************************HBASE***************************************/
  {
    "id": "puppet var",
    "name": "hbasemaster_host",
    "displayName": "HBase Master hosts",
    "value": "",
    "defaultValue": "",
    "description": "The host that has been assigned to run HBase Master",
    "displayType": "masterHosts",
    "isOverridable": false,
    "isRequiredByAgent": false,
    "isVisible": true,
    "serviceName": "HBASE",
    "filename": "hbase-env.xml",
    "category": "HBASE_MASTER",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "hbase_master_heapsize",
    "displayName": "HBase Master Maximum Java heap size",
    "description": "Maximum Java heap size for HBase master (Java option -Xmx)",
    "defaultValue": "1024",
    "displayType": "int",
    "unit": "MB",
    "isOverridable": true,
    "isVisible": true,
    "serviceName": "HBASE",
    "filename": "hbase-env.xml",
    "category": "HBASE_MASTER",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "regionserver_hosts",
    "displayName": "RegionServer hosts",
    "value": "",
    "defaultValue": "",
    "description": "The hosts that have been assigned to run RegionServer",
    "displayType": "slaveHosts",
    "isOverridable": false,
    "isRequiredByAgent": false,
    "isVisible": true,
    "isRequired": false,
    "serviceName": "HBASE",
    "filename": "hbase-env.xml",
    "category": "HBASE_REGIONSERVER",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "hbase_regionserver_heapsize",
    "displayName": "RegionServers maximum Java heap size",
    "description": "Maximum Java heap size for RegionServers (Java option -Xmx)",
    "defaultValue": "1024",
    "displayType": "int",
    "unit": "MB",
    "isVisible": true,
    "serviceName": "HBASE",
    "filename": "hbase-env.xml",
    "category": "HBASE_REGIONSERVER",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "hbase_regionserver_xmn_max",
    "displayName": "RegionServers maximum value for -Xmn",
    "displayType": "int",
    "unit": "MB",
    "isVisible": true,
    "serviceName": "HBASE",
    "filename": "hbase-env.xml",
    "category": "HBASE_REGIONSERVER",
    "index": 6
  },
  {
    "id": "puppet var",
    "name": "hbase_regionserver_xmn_ratio",
    "displayName": "RegionServers -Xmn in -Xmx ratio",
    "displayType": "float",
    "isVisible": true,
    "serviceName": "HBASE",
    "filename": "hbase-env.xml",
    "category": "HBASE_REGIONSERVER",
    "index": 7
  },
  {
    "id": "puppet var",
    "name": "hbase_log_dir",
    "displayName": "HBase Log Dir",
    "description": "Directory for HBase logs",
    "defaultValue": "/var/log/hbase",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HBASE",
    "filename": "hbase-env.xml",
    "category": "Advanced hbase-env"
  },
  {
    "id": "puppet var",
    "name": "hbase_pid_dir",
    "displayName": "HBase PID Dir",
    "description": "Directory in which the pid files for HBase processes will be created",
    "defaultValue": "/var/run/hbase",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HBASE",
    "filename": "hbase-env.xml",
    "category": "Advanced hbase-env"
  },
  //***************************************** GLUSTERFS stack********************************************
  {
    "id": "site property",
    "name": "fs.glusterfs.impl",
    "displayName": "GlusterFS fs impl",
    "displayType": "string",
    "filename": "core-site.xml",
    "serviceName": "GLUSTERFS",
    "category": "General"
  },
  {
    "id": "site property",
    "name": "fs.AbstractFileSystem.glusterfs.impl",
    "displayName": "GlusterFS Abstract File System Implementation",
    "displayType": "string",
    "filename": "core-site.xml",
    "serviceName": "GLUSTERFS",
    "category": "General"
  },
/**********************************************GLUSTERFS***************************************/
  {
    "id": "puppet var",
    "name": "fs_glusterfs_default_name",
    "displayName": "GlusterFS default fs name 1.x Hadoop",
    "description": "GlusterFS default filesystem name (glusterfs://{MasterFQDN}:9000)",
    "defaultValue": "glusterfs:///localhost:8020",
    "displayType": "string",
    "isVisible": true,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop"
  },
  {
    "id": "puppet var",
    "name": "glusterfs_defaultFS_name",
    "displayName": "GlusterFS default fs name 2.x Hadoop",
    "description": "GlusterFS default filesystem name (glusterfs:///)",
    "defaultValue": "glusterfs:///localhost:8020",
    "displayType": "string",
    "isVisible": true,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop"
  },
  {
    "id": "puppet var",
    "name": "hadoop_heapsize",
    "displayName": "Hadoop maximum Java heap size",
    "description": "Maximum Java heap size for daemons such as Balancer (Java option -Xmx)",
    "defaultValue": "1024",
    "displayType": "int",
    "unit": "MB",
    "isVisible": true,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "hdfs_log_dir_prefix",
    "displayName": "Hadoop Log Dir Prefix",
    "description": "The parent directory for Hadoop log files.  The HDFS log directory will be ${hadoop_log_dir_prefix} / ${hdfs_user} and the MapReduce log directory will be ${hadoop_log_dir_prefix} / ${mapred_user}.",
    "defaultValue": "/var/log/hadoop",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop"
  },
  {
    "id": "puppet var",
    "name": "hadoop_pid_dir_prefix",
    "displayName": "Hadoop PID Dir Prefix",
    "description": "The parent directory in which the PID files for Hadoop processes will be created.  The HDFS PID directory will be ${hadoop_pid_dir_prefix} / ${hdfs_user} and the MapReduce PID directory will be ${hadoop_pid_dir_prefix} / ${mapred_user}.",
    "defaultValue": "/var/run/hadoop",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop"
  },
  {
    "id": "puppet var",
    "name": "namenode_heapsize",
    "displayName": "Name Node Heap Size",
    "description": "Name Node Heap Size, default jvm memory setting",
    "defaultValue": "1024",
    "isReconfigurable": false,
    "displayType": "int",
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop"
  },
  {
    "id": "puppet var",
    "name": "namenode_opt_newsize",
    "displayName": "NameNode new generation size",
    "description": "Default size of Java new generation for NameNode (Java option -XX:NewSize).  This also applies to the Secondary NameNode.",
    "defaultValue": "200",
    "displayType": "int",
    "unit": "MB",
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop"
  },
  {
    "id": "puppet var",
    "name": "namenode_opt_maxnewsize",
    "displayName": "NameNode maximum new generation size",
    "description": "Maximum size of Java new generation for NameNode (Java option -XX:MaxnewSize).",
    "defaultValue": "200",
    "displayType": "int",
    "unit": "MB",
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop"
  },
  {
    "id": "puppet var",
    "name": "namenode_opt_permsize",
    "displayName": "NameNode permanent generation size",
    "description": "Default size of Java permanent generation for NameNode (Java option -XX:PermSize).  This also applies to the Secondary NameNode.",
    "defaultValue": "128",
    "displayType": "int",
    "unit": "MB",
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop"
  },
  {
    "id": "puppet var",
    "name": "namenode_opt_maxpermsize",
    "displayName": "NameNode maximum permanent generation size",
    "description": "Maximum size of Java permanent generation for NameNode (Java option -XX:MaxPermSize).",
    "defaultValue": "256",
    "displayType": "int",
    "unit": "MB",
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop"
  },
  {
    "id": "puppet var",
    "name": "dtnode_heapsize",
    "displayName": "DataNode maximum Java heap size",
    "description": "Maximum Java heap size for DataNode (Java option -Xmx)",
    "defaultValue": "1024",
    "displayType": "int",
    "unit": "MB",
    "isVisible": false,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop"
  },
  {
    "id": "puppet var",
    "name": "glusterfs_user",
    "displayName": "glusterfs user",
    "description": "glusterfs user",
    "defaultValue": "root",
    "displayType": "string",
    "isVisible": false,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop"
  },
  {
    "id": "puppet var",
    "name": "namenode_host",
    "displayName": "NameNode Host",
    "description": "NameNode Host.",
    "defaultValue": "",
    "displayType": "string",
    "isVisible": false,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop"
  },
  {
    "id": "puppet var",
    "name": "snamenode_host",
    "displayName": "Secondary NameNode Host",
    "description": "Secondary NameNode Host.",
    "defaultValue": "",
    "displayType": "string",
    "isVisible": false,
    "serviceName": "GLUSTERFS",
    "filename": "hadoop-env.xml",
    "category": "General Hadoop"
  },
/**********************************************HIVE***************************************/
  {
    "id": "puppet var",
    "name": "hivemetastore_host",
    "displayName": "Hive Metastore hosts",
    "value": "",
    "defaultValue": "",
    "description": "The hosts that have been assigned to run Hive Metastore",
    "displayType": "masterHosts",
    "isOverridable": false,
    "isRequiredByAgent": false,
    "isVisible": true,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 0
  },
  {
    "name": "hive_master_hosts",
    "value": "",
    "defaultValue": "",
    "isOverridable": false,
    "isRequiredByAgent": false,
    "isVisible": false,
    "serviceName": "HIVE",
    "filename": "hive-env.xml"
  },
  // for existing MySQL
  {
    "id": "puppet var",
    "name": "hive_existing_mysql_database",
    "displayName": "Database Type",
    "value": "",
    "defaultValue": "MySQL",
    "description": "Using an existing MySQL database for Hive Metastore",
    "displayType": "masterHost",
    "isOverridable": false,
    "isVisible": false,
    "isReconfigurable": false,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "hive_existing_mssql_server_database",
    "displayName": "Database Type",
    "value": "",
    "defaultValue": "MSSQL",
    "description": "Using an existing MSSQL database for Hive Metastore",
    "displayType": "masterHost",
    "isOverridable": false,
    "isVisible": false,
    "isReconfigurable": false,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "hive_existing_mssql_server_2_database",
    "displayName": "Database Type",
    "value": "",
    "defaultValue": "MSSQL",
    "description": "Using an existing MSSQL database for Hive Metastore",
    "displayType": "masterHost",
    "isOverridable": false,
    "isVisible": false,
    "isReconfigurable": false,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 1
  },
  // for existing PostgreSQL
  {
    "id": "puppet var",
    "name": "hive_existing_postgresql_database",
    "displayName": "Database Type",
    "value": "",
    "defaultValue": "PostgreSQL",
    "description": "Using an existing PostgreSQL database for Hive Metastore",
    "displayType": "masterHost",
    "isOverridable": false,
    "isVisible": false,
    "isReconfigurable": false,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 1
  },
  // for existing Oracle
  {
    "id": "puppet var",
    "name": "hive_existing_oracle_database",
    "displayName": "Database Type",
    "value": "",
    "defaultValue": "Oracle",
    "description": "Using an existing Oracle database for Hive Metastore",
    "displayType": "masterHost",
    "isVisible": false,
    "isOverridable": false,
    "isReconfigurable": false,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 1
  },
  // for new MySQL
  {
    "id": "puppet var",
    "name": "hive_ambari_database",
    "displayName": "Database Type",
    "value": "",
    "defaultValue": "MySQL",
    "description": "MySQL will be installed by Ambari",
    "displayType": "masterHost",
    "isOverridable": false,
    "isReconfigurable": false,
    "isVisible": false,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "hive_database",
    "displayName": "Hive Database",
    "value": "",
    "defaultValue": "New MySQL Database",
    "options": [
      {
        displayName: 'New MySQL Database',
        foreignKeys: ['hive_ambari_database', 'hive_ambari_host'],
        hidden: App.get('isHadoopWindowsStack')
      },
      {
        displayName: 'Existing MySQL Database',
        foreignKeys: ['hive_existing_mysql_database', 'hive_existing_mysql_host'],
        hidden: App.get('isHadoopWindowsStack')
      },
      {
        displayName: 'Existing PostgreSQL Database',
        foreignKeys: ['hive_existing_postgresql_database', 'hive_existing_postgresql_host'],
        hidden: App.get('isHadoopWindowsStack')
      },
      {
        displayName: 'Existing Oracle Database',
        foreignKeys: ['hive_existing_oracle_database', 'hive_existing_oracle_host'],
        hidden: App.get('isHadoopWindowsStack')
      },
      {
        displayName: 'Existing MSSQL Server database with SQL authentication',
        foreignKeys: ['hive_existing_mssql_server_database', 'hive_existing_mssql_server_host'],
        hidden: !App.get('isHadoopWindowsStack')
      },
      {
        displayName: 'Existing MSSQL Server database with integrated authentication',
        foreignKeys: ['hive_existing_mssql_server_2_database', 'hive_existing_mssql_server_2_host'],
        hidden: !App.get('isHadoopWindowsStack')
      }
    ],
    "description": "MySQL will be installed by Ambari",
    "displayType": "radio button",
    "isReconfigurable": true,
    "radioName": "hive-database",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 2
  },
  {
    "id": "puppet var",
    "name": "hive_hostname",
    "displayName": "Database Host",
    "description": "Specify the host on which the database is hosted",
    "defaultValue": "",
    "isReconfigurable": true,
    "displayType": "host",
    "isOverridable": false,
    "isVisible": false,
    "isObserved": true,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "hive_existing_mysql_host",
    "displayName": "Database Host",
    "description": "Specify the host on which the existing database is hosted",
    "defaultValue": "",
    "displayType": "host",
    "isOverridable": false,
    "isVisible": false,
    "isObserved": true,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "hive_existing_postgresql_host",
    "displayName": "Database Host",
    "description": "Specify the host on which the existing database is hosted",
    "defaultValue": "",
    "displayType": "host",
    "isOverridable": false,
    "isVisible": false,
    "isObserved": true,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "hive_existing_oracle_host",
    "displayName": "Database Host",
    "description": "Specify the host on which the existing database is hosted",
    "defaultValue": "",
    "displayType": "host",
    "isOverridable": false,
    "isVisible": false,
    "isObserved": true,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "hive_ambari_host",
    "value": "",
    "defaultValue": "",
    "displayName": "Database Host",
    "description": "Host on which the database will be created by Ambari",
    "isReconfigurable": false,
    "displayType": "masterHost",
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "hive_existing_mssql_server_host",
    "displayName": "Database Host",
    "description": "Specify the host on which the existing database is hosted",
    "defaultValue": "",
    "displayType": "host",
    "isOverridable": false,
    "isVisible": false,
    "isObserved": true,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "hive_existing_mssql_server_2_host",
    "displayName": "Database Host",
    "description": "Specify the host on which the existing database is hosted",
    "defaultValue": "",
    "displayType": "host",
    "isOverridable": false,
    "isVisible": false,
    "isObserved": true,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "HIVE_METASTORE",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "hive_metastore_port",
    "displayName": "Hive metastore port",
    "description": "",
    "defaultValue": "9083",
    "isReconfigurable": false,
    "displayType": "int",
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "Advanced hive-env"
  },
  {
    "id": "puppet var",
    "name": "hive_lib",
    "displayName": "Hive library",
    "description": "",
    "defaultValue": "/usr/lib/hive/lib/",
    "isReconfigurable": false,
    "displayType": "directory",
    "isVisible": false,
    "isRequiredByAgent": false, // Make this to true when we expose the property on ui by making "isVisible": true
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "Advanced hive-env"
  },
  {
    "id": "puppet var",
    "name": "hive_dbroot",
    "displayName": "Hive db directory",
    "description": "",
    "defaultValue": "/usr/lib/hive/lib",
    "isReconfigurable": false,
    "displayType": "directory",
    "isVisible": false,
    "isRequiredByAgent": false, // Make this to true when we expose the property on ui by making "isVisible": true
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "Advanced hive-env"
  },
  {
    "id": "puppet var",
    "name": "hive_log_dir",
    "displayName": "Hive Log Dir",
    "description": "Directory for Hive log files",
    "defaultValue": "/var/log/hive",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "Advanced hive-env"
  },
  {
    "id": "puppet var",
    "name": "hive_pid_dir",
    "displayName": "Hive PID Dir",
    "description": "Directory in which the PID files for Hive processes will be created",
    "defaultValue": "/var/run/hive",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "Advanced hive-env"
  },
/**********************************************HIVE***************************************/
  {
    "id": "puppet var",
    "name": "webhcatserver_host",
    "displayName": "WebHCat Server host",
    "value": "",
    "defaultValue": "",
    "description": "The host that has been assigned to run WebHCat Server",
    "displayType": "masterHost",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": false,
    "serviceName": "HIVE",
    "filename": "webhcat-env.xml",
    "category": "WEBHCAT_SERVER"
  },
  {
    "id": "puppet var",
    "name": "hcat_log_dir",
    "displayName": "WebHCat Log Dir",
    "description": "Directory for WebHCat log files",
    "defaultValue": "/var/log/webhcat",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "Advanced webhcat-env"
  },
  {
    "id": "puppet var",
    "name": "hcat_pid_dir",
    "displayName": "WebHCat PID Dir",
    "description": "Directory in which the PID files for WebHCat processes will be created",
    "defaultValue": "/var/run/webhcat",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "Advanced webhcat-env"
  },
  {
    "id": "puppet var",
    "name": "hive_database_name",
    "displayName": "hive_database_name",
    "defaultValue": "",
    "isReconfigurable": false,
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "Advanced hive-env"
  },
  {
    "id": "puppet var",
    "name": "hive_database_type",
    "displayName": "hive_database_type",
    "defaultValue": "",
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "HIVE",
    "filename": "hive-env.xml",
    "category": "Advanced hive-env"
  },
/**********************************************OOZIE***************************************/
  {
    "id": "puppet var",
    "name": "oozieserver_host",
    "displayName": "Oozie Server host",
    "value": "",
    "defaultValue": "",
    "description": "The host that has been assigned to run Oozie Server",
    "displayType": "masterHost",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": false,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 0
  },
  // for existing Oracle
  {
    "id": "puppet var",
    "name": "oozie_existing_oracle_database",
    "displayName": "Database Type",
    "value": "",
    "defaultValue": "Oracle",
    "description": "Using an existing Oracle database for Oozie Metastore",
    "displayType": "masterHost",
    "isVisible": false,
    "isObserved": true,
    "isReconfigurable": false,
    "isOverridable": false,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 1
  },
  // for existing PostgreSQL
  {
    "id": "puppet var",
    "name": "oozie_existing_postgresql_database",
    "displayName": "Database Type",
    "value": "",
    "defaultValue": "PostgreSQL",
    "description": "Using an existing PostgreSQL database for Oozie Metastore",
    "displayType": "masterHost",
    "isOverridable": false,
    "isVisible": false,
    "isReconfigurable": false,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 1
  },
  // for new MySQL
  {
    "id": "puppet var",
    "name": "oozie_ambari_database",
    "displayName": "Database Type",
    "value": "",
    "defaultValue": "MySQL",
    "description": "MySQL will be installed by Ambari",
    "displayType": "masterHost",
    "isVisible": false,
    "isReconfigurable": false,
    "isOverridable": false,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 1
  },
  // for current derby
  {
    "id": "puppet var",
    "name": "oozie_derby_database",
    "displayName": "Database Type",
    "value": "",
    "defaultValue": "Derby",
    "description": "Using current Derby database for Oozie Metastore",
    "displayType": "masterHost",
    "isVisible": false,
    "isReconfigurable": false,
    "isOverridable": false,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 1
  },
  // for existing MySQL oozie
  {
    "id": "puppet var",
    "name": "oozie_existing_mysql_database",
    "displayName": "Database Type",
    "value": "",
    "defaultValue": "MySQL",
    "description": "Using an existing MySQL database for Oozie Metastore",
    "displayType": "masterHost",
    "isVisible": false,
    "isReconfigurable": false,
    "isOverridable": false,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "oozie_existing_mssql_server_database",
    "displayName": "Database Type",
    "value": "",
    "defaultValue": "MSSQL",
    "description": "Using an existing MSSQL database with SQL authentication for Oozie Metastore",
    "displayType": "masterHost",
    "isVisible": false,
    "isReconfigurable": false,
    "isOverridable": false,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "oozie_existing_mssql_server_2_database",
    "displayName": "Database Type",
    "value": "",
    "defaultValue": "MSSQL",
    "description": "Using an existing MSSQL database with integrated authentication for Oozie Metastore",
    "displayType": "masterHost",
    "isVisible": false,
    "isReconfigurable": false,
    "isOverridable": false,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "oozie_database",
    "displayName": "Oozie Database",
    "value": "",
    "defaultValue": "New Derby Database",
    "options": [
      {
        displayName: 'New Derby Database',
        foreignKeys: ['oozie_derby_database'],
        hidden: App.get('isHadoopWindowsStack')
      },
      {
        displayName: 'Existing MySQL Database',
        foreignKeys: ['oozie_existing_mysql_database', 'oozie_existing_mysql_host'],
        hidden: App.get('isHadoopWindowsStack')
      },
      {
        displayName: 'Existing PostgreSQL Database',
        foreignKeys: ['oozie_existing_postgresql_database', 'oozie_existing_postgresql_host'],
        hidden: App.get('isHadoopWindowsStack')
      },
      {
        displayName: 'Existing Oracle Database',
        foreignKeys: ['oozie_existing_oracle_database', 'oozie_existing_oracle_host'],
        hidden: App.get('isHadoopWindowsStack')
      },
      {
        displayName: 'Existing MSSQL Server database with SQL authentication',
        foreignKeys: ['oozie_existing_mssql_server_database', 'oozie_existing_mssql_server_host'],
        hidden: !App.get('isHadoopWindowsStack')
      },
      {
        displayName: 'Existing MSSQL Server database with integrated authentication',
        foreignKeys: ['oozie_existing_mssql_server_2_database', 'oozie_existing_mssql_server_2_host'],
        hidden: !App.get('isHadoopWindowsStack')
      }
    ],
    "description": "Current Derby Database will be installed by Ambari",
    "displayType": "radio button",
    "isReconfigurable": true,
    "isOverridable": false,
    "radioName": "oozie-database",
    "isVisible": true,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 2
  },
  {
    "id": "puppet var",
    "name": "oozie_data_dir",
    "displayName": "Oozie Data Dir",
    "description": "Data directory in which the Oozie DB exists",
    "defaultValue": "",
    "defaultDirectory": "/hadoop/oozie/data",
    "isReconfigurable": true,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "isRequired": false,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 9
  },
  {
    "id": "puppet var",
    "name": "oozie_hostname",
    "defaultValue": "",
    "displayName": "Database Host",
    "description": "The host where the Oozie database is located",
    "isReconfigurable": true,
    "isOverridable": false,
    "displayType": "host",
    "isVisible": false,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "oozie_existing_mysql_host",
    "displayName": "Database Host",
    "description": "Specify the host on which the existing database is hosted",
    "defaultValue": "",
    "isOverridable": false,
    "displayType": "host",
    "isVisible": false,
    "isObserved": true,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "oozie_existing_postgresql_host",
    "displayName": "Database Host",
    "description": "Specify the host on which the existing database is hosted",
    "defaultValue": "",
    "isOverridable": false,
    "displayType": "host",
    "isVisible": false,
    "isObserved": true,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "oozie_existing_oracle_host",
    "displayName": "Database Host",
    "description": "Specify the host on which the existing database is hosted",
    "defaultValue": "",
    "isOverridable": false,
    "displayType": "host",
    "isVisible": false,
    "isObserved": true,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "oozie_ambari_host",
    "value": "",
    "defaultValue": "",
    "displayName": "Database Host",
    "description": "Host on which the database will be created by Ambari",
    "isReconfigurable": false,
    "isOverridable": false,
    "displayType": "masterHost",
    "isVisible": false,
    "isRequiredByAgent": false,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER"
  },
  {
    "id": "puppet var",
    "name": "oozie_existing_mssql_server_host",
    "displayName": "Database Host",
    "description": "Specify the host on which the existing database is hosted",
    "defaultValue": "",
    "displayType": "host",
    "isOverridable": false,
    "isRequiredByAgent": false,
    "isVisible": false,
    "isObserved": true,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "oozie_existing_mssql_server_2_host",
    "displayName": "Database Host",
    "description": "Specify the host on which the existing database is hosted",
    "defaultValue": "",
    "displayType": "host",
    "isOverridable": false,
    "isRequiredByAgent": false,
    "isVisible": false,
    "isObserved": true,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "OOZIE_SERVER",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "oozie_log_dir",
    "displayName": "Oozie Log Dir",
    "description": "Directory for oozie logs",
    "defaultValue": "/var/log/oozie",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "Advanced oozie-env"
  },
  {
    "id": "puppet var",
    "name": "oozie_pid_dir",
    "displayName": "Oozie PID Dir",
    "description": "Directory in which the pid files for oozie processes will be created",
    "defaultValue": "/var/run/oozie",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "Advanced oozie-env"
  },
  {
    "id": "puppet var",
    "name": "oozie_admin_port",
    "displayName": "Oozie Server Admin Port",
    "isReconfigurable": true,
    "displayType": "int",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "OOZIE",
    "filename": "oozie-env.xml",
    "category": "Advanced oozie-env"
  },
/**********************************************ZOOKEEPER***************************************/
  {
    "id": "puppet var",
    "name": "zookeeperserver_hosts",
    "displayName": "ZooKeeper Server hosts",
    "value": "",
    "defaultValue": "",
    "description": "The host that has been assigned to run ZooKeeper Server",
    "displayType": "masterHosts",
    "isVisible": true,
    "isRequiredByAgent": false,
    "isOverridable": false,
    "isRequired": false,
    "serviceName": "ZOOKEEPER",
    "filename": "zookeeper-env.xml",
    "category": "ZOOKEEPER_SERVER",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "dataDir",
    "displayName": "ZooKeeper directory",
    "description": "Data directory for ZooKeeper",
    "defaultValue": "",
    "defaultDirectory": "/hadoop/zookeeper",
    "isReconfigurable": true,
    "displayType": "directory",
    "isVisible": true,
    "serviceName": "ZOOKEEPER",
    "filename": "zoo.cfg.xml",
    "category": "ZOOKEEPER_SERVER",
    "index": 1
  },
  {
    "id": "puppet var",
    "name": "tickTime",
    "displayName": "Length of single Tick",
    "description": "The length of a single tick in milliseconds, which is the basic time unit used by ZooKeeper",
    "defaultValue": "2000",
    "displayType": "int",
    "unit": "ms",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "ZOOKEEPER",
    "filename": "zoo.cfg.xml",
    "category": "ZOOKEEPER_SERVER",
    "index": 2
  },
  {
    "id": "puppet var",
    "name": "initLimit",
    "displayName": "Ticks to allow for sync at Init",
    "description": "Amount of time, in ticks to allow followers to connect and sync to a leader",
    "defaultValue": "10",
    "displayType": "int",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "ZOOKEEPER",
    "filename": "zoo.cfg.xml",
    "category": "ZOOKEEPER_SERVER",
    "index": 3
  },
  {
    "id": "puppet var",
    "name": "syncLimit",
    "displayName": "Ticks to allow for sync at Runtime",
    "description": "Amount of time, in ticks to allow followers to connect",
    "defaultValue": "5",
    "displayType": "int",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "ZOOKEEPER",
    "filename": "zoo.cfg.xml",
    "category": "ZOOKEEPER_SERVER",
    "index": 4
  },
  {
    "id": "puppet var",
    "name": "clientPort",
    "displayName": "Port for running ZK Server",
    "description": "Port for running ZooKeeper server",
    "defaultValue": "2181",
    "displayType": "int",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "ZOOKEEPER",
    "filename": "zoo.cfg.xml",
    "category": "ZOOKEEPER_SERVER",
    "index": 5
  },
  {
    "id": "puppet var",
    "name": "zk_log_dir",
    "displayName": "ZooKeeper Log Dir",
    "description": "Directory for ZooKeeper log files",
    "defaultValue": "/var/log/zookeeper",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "ZOOKEEPER",
    "filename": "zookeeper-env.xml",
    "category": "Advanced zookeeper-env",
    "index": 0
  },
  {
    "id": "puppet var",
    "name": "zk_pid_dir",
    "displayName": "ZooKeeper PID Dir",
    "description": "Directory in which the pid files for zookeeper processes will be created",
    "defaultValue": "/var/run/zookeeper",
    "isReconfigurable": false,
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "ZOOKEEPER",
    "filename": "zookeeper-env.xml",
    "category": "Advanced zookeeper-env",
    "index": 1
  },
/**********************************************GANGLIA***************************************/
  {
    "id": "puppet var",
    "name": "ganglia_conf_dir",
    "displayName": "Ganglia conf directory",
    "description": "",
    "defaultValue": "/etc/ganglia/hdp",
    "isReconfigurable": false,
    "displayType": "directory",
    "isVisible": false,
    "isRequiredByAgent": false,
    "serviceName": "GANGLIA",
    "filename": "ganglia-env.xml",
    "category": "Advanced ganglia-env"
  },
/**********************************************FALCON***************************************/
  {
    "id": "puppet var",
    "name": "falconserver_host",
    "displayName": "Falcon Server",
    "description": "The host that has been assigned to run Falcon Server",
    "defaultValue": "falcon",
    "displayType": "masterHost",
    "isVisible": true,
    "isRequiredByAgent": false,
    "serviceName": "FALCON",
    "filename": "falcon-env.xml",
    "category": "FALCON_SERVER"
  },
  {
    "id": "puppet var",
    "name": "falcon_port",
    "displayName": "Falcon server port",
    "description": "Port the Falcon Server listens on",
    "defaultValue": "15000",
    "isReconfigurable": true,
    "displayType": "int",
    "isVisible": true,
    "isRequiredByAgent": true,
    "serviceName": "FALCON",
    "filename": "falcon-env.xml",
    "category": "FALCON_SERVER"
  },
  {
    "id": "puppet var",
    "name": "falcon_local_dir",
    "displayName": "Falcon data directory",
    "description": "Directory where Falcon data, such as activemq data, is stored",
    "defaultValue": "/hadoop/falcon",
    "isReconfigurable": true,
    "displayType": "directory",
    "isVisible": true,
    "isRequiredByAgent": true,
    "serviceName": "FALCON",
    "filename": "falcon-env.xml",
    "category": "FALCON_SERVER"
  },
  {
    "id": "puppet var",
    "name": "falcon_store_uri",
    "displayName": "Falcon store URI",
    "description": "Directory where entity definitions are stored",
    "defaultValue": "file:///hadoop/falcon/store",
    "isReconfigurable": true,
    "displayType": "string",
    "isVisible": true,
    "isRequiredByAgent": true,
    "serviceName": "FALCON",
    "filename": "falcon-env.xml",
    "category": "FALCON_SERVER"
  },
  {
    "id": "puppet var",
    "name": "falcon_log_dir",
    "displayName": "Falcon Log Dir",
    "description": "Directory for Falcon logs",
    "defaultValue": "/var/log/falcon",
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": true,
    "isReconfigurable": false,
    "serviceName": "FALCON",
    "filename": "falcon-env.xml",
    "category": "Advanced falcon-env"
  },
  {
    "id": "puppet var",
    "name": "falcon_pid_dir",
    "displayName": "Falcon PID Dir",
    "description": "Directory in which the pid files for Falcon processes will be created",
    "defaultValue": "/var/run/falcon",
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": true,
    "isReconfigurable": false,
    "serviceName": "FALCON",
    "filename": "falcon-env.xml",
    "category": "Advanced falcon-env"
  },
  {
    "id": "puppet var",
    "name": "falcon.embeddedmq",
    "displayName": "falcon.embeddedmq",
    "description": "Whether embeddedmq is enabled or not.",
    "defaultValue": "true",
    "displayType": "string",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": true,
    "isReconfigurable": true,
    "serviceName": "FALCON",
    "filename": "falcon-env.xml",
    "category": "Advanced falcon-env"
  },
  {
    "id": "puppet var",
    "name": "falcon.embeddedmq.data",
    "displayName": "falcon.embeddedmq.data",
    "description": "Directory in which embeddedmq data is stored.",
    "defaultValue": "/hadoop/falcon/embeddedmq/data",
    "displayType": "directory",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": true,
    "isReconfigurable": true,
    "serviceName": "FALCON",
    "filename": "falcon-env.xml",
    "category": "Advanced falcon-env"
  },
  {
    "id": "puppet var",
    "name": "falcon.emeddedmq.port",
    "displayName": "falcon.emeddedmq.port",
    "description": "Port that embeddedmq will listen on.",
    "defaultValue": "61616",
    "displayType": "string",
    "isOverridable": false,
    "isVisible": true,
    "isRequiredByAgent": true,
    "isReconfigurable": true,
    "serviceName": "FALCON",
    "filename": "falcon-env.xml",
    "category": "Advanced falcon-env"
  },
/**********************************************STORM***************************************/
  {
    "id": "puppet var",
    "name": "storm_log_dir",
    "displayName": "storm_log_dir",
    "description": "Storm log directory",
    "defaultValue": "/var/log/storm",
    "displayType": "directory",
    "isReconfigurable": true,
    "isVisible": true,
    "isRequiredByAgent": true,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "category": "Advanced storm-env"
  },
  {
    "id": "puppet var",
    "name": "storm_pid_dir",
    "displayName": "storm_pid_dir",
    "description": "Storm PID directory",
    "defaultValue": "/var/run/storm",
    "displayType": "directory",
    "isReconfigurable": true,
    "isVisible": true,
    "isRequiredByAgent": true,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "category": "Advanced storm-env"
  },
  {
    "id": "puppet var",
    "name": "stormuiserver_host",
    "displayName": "Storm UI Server host",
    "description": "The host that has been assigned to run Storm UI Server",
    "defaultValue": "",
    "displayType": "masterHost",
    "isReconfigurable": false,
    "isVisible": true,
    "isRequiredByAgent": false,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "category": "STORM_UI_SERVER"
  },
  {
    "id": "puppet var",
    "name": "drpcserver_host",
    "displayName": "DRPC Server host",
    "description": "The host that has been assigned to run DRPC Server",
    "defaultValue": "",
    "displayType": "masterHost",
    "isReconfigurable": false,
    "isVisible": true,
    "isRequiredByAgent": false,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "category": "DRPC_SERVER"
  },
  {
    "id": "puppet var",
    "name": "supervisor_hosts",
    "displayName": "Supervisor hosts",
    "description": "The host that has been assigned to run Supervisor",
    "defaultValue": "",
    "displayType": "slaveHosts",
    "isReconfigurable": false,
    "isVisible": true,
    "isRequiredByAgent": false,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "isOverridable": false,
    "category": "SUPERVISOR"
  },
  {
    "id": "puppet var",
    "name": "storm_rest_api_host",
    "displayName": "Storm REST API host",
    "description": "The host that has been assigned to run Storm REST API Server",
    "defaultValue": "",
    "displayType": "masterHost",
    "isReconfigurable": false,
    "isVisible": true,
    "isRequiredByAgent": false,
    "serviceName": "STORM",
    "filename": "storm-env.xml",
    "isOverridable": false,
    "category": "STORM_REST_API"
  },
/**********************************************MISC***************************************/
  {
    "id": "puppet var",
    "name": "hbase_conf_dir",
    "displayName": "HBase conf dir",
    "description": "",
    "defaultValue": "/etc/hbase",
    "isRequired": false,
    "displayType": "directory",
    "isVisible": false,
    "isRequiredByAgent": false,
    "serviceName": "MISC",
    "filename": "hbase-env.xml",
    "category": "General",
    "belongsToService": []
  },
  {
    "id": "puppet var",
    "name": "ganglia_runtime_dir",
    "displayName": "Ganglia runtime directory",
    "description": "",
    "defaultValue": "/var/run/ganglia/hdp",
    "isReconfigurable": false,
    "displayType": "directory",
    "isVisible": false,
    "serviceName": "MISC",
    "filename": "ganglia-env.xml",
    "category": "General",
    "belongsToService": []
  },
  {
    "id": "puppet var",
    "name": "rrdcached_base_dir",
    "displayName": "Ganglia rrdcached base directory",
    "description": "Default directory for saving the rrd files on ganglia server",
    "defaultValue": "/var/lib/ganglia/rrds",
    "displayType": "directory",
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "GANGLIA",
    "filename": "ganglia-env.xml",
    "category": "General",
    "belongsToService": ["GANGLIA"]
  },
  {
    "id": "puppet var",
    "name": "ignore_groupsusers_create",
    "displayName": "Skip group modifications during install",
    "displayType": "checkbox",
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": !App.get('isHadoopWindowsStack'),
    "filename": "cluster-env.xml",
    "category": "Users and Groups"
  },
  {
    "id": "puppet var",
    "name": "create_notification",
    "displayName": "Create Notification",
    "isRequired": true,
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "MISC",
    "category": "Notifications",
    "defaultValue": "no",
    "filename": "alert_notification"
  },
  {
    "id": "puppet var",
    "name": "mail.smtp.host",
    "displayName": "SMTP Host",
    "displayType": "host",
    "isRequired": false,
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "MISC",
    "category": "Notifications",
    "rowStyleClass": "indent-1",
    "filename": "alert_notification"
  },
  {
    "id": "puppet var",
    "name": "mail.smtp.port",
    "displayName": "SMTP Port",
    "displayType": "int",
    "isRequired": false,
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "MISC",
    "category": "Notifications",
    "rowStyleClass": "indent-1",
    "filename": "alert_notification"
  },
  {
    "id": "puppet var",
    "name": "mail.smtp.from",
    "displayName": "FROM Email Address",
    "displayType": "email",
    "isRequired": false,
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "MISC",
    "category": "Notifications",
    "rowStyleClass": "indent-1",
    "filename": "alert_notification"
  },
  {
    "id": "puppet var",
    "name": "ambari.dispatch.recipients",
    "displayName": " TO Email Address",
    "displayType": "email",
    "isRequired": false,
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "MISC",
    "category": "Notifications",
    "rowStyleClass": "indent-1",
    "filename": "alert_notification"
  },
  {
    "id": "puppet var",
    "name": "smtp_use_auth",
    "displayName": "SMTP server requires authentication",
    "displayType": "checkbox",
    "isRequired": false,
    "isReconfigurable": true,
    "isOverridable": false,
    "defaultValue": true,
    "isVisible": true,
    "serviceName": "MISC",
    "category": "Notifications",
    "rowStyleClass": "indent-1",
    "filename": "alert_notification"
  },
  {
    "id": "puppet var",
    "name": "ambari.dispatch.credential.username",
    "displayName": "SMTP Username",
    "displayType": "string",
    "isRequired": false,
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "MISC",
    "category": "Notifications",
    "rowStyleClass": "indent-2",
    "filename": "alert_notification"
  },
  {
    "id": "puppet var",
    "name": "ambari.dispatch.credential.password",
    "displayName": "SMTP Password",
    "displayType": "string",
    "isRequired": false,
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": true,
    "serviceName": "MISC",
    "category": "Notifications",
    "rowStyleClass": "indent-2",
    "filename": "alert_notification"
  },
  {
    "id": "puppet var",
    "name": "mail.smtp.starttls.enable",
    "isRequired": false,
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "MISC",
    "category": "Notifications",
    "rowStyleClass": "indent-1",
    "filename": "alert_notification"
  },
  {
    "id": "puppet var",
    "name": "mail.smtp.startssl.enable",
    "isRequired": false,
    "isReconfigurable": true,
    "isOverridable": false,
    "isVisible": false,
    "serviceName": "MISC",
    "category": "Notifications",
    "rowStyleClass": "indent-1",
    "filename": "alert_notification"
  },
/************************************************AMBARI_METRICS******************************************/
  {
    "id": "site property",
    "name": "timeline.metrics.service.operation.mode",
    "displayName": "Metrics Service operation mode",
    "description": "\n      Service Operation modes:\n      1) embedded: Metrics stored on local FS, HBase in Standalone mode\n      2) distributed: HBase daemons writing to HDFS\n      3) external: External HBase storage backend\n    ",
    "defaultValue": "embedded",
    "displayType": "string",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "General",
    "index": 1
  },
  {
    "id": "site property",
    "name": "metrics_collector_log_dir",
    "displayName": "Metrics Collector log dir",
    "description": "\n      Log location for collector logs\n    ",
    "defaultValue": "embedded",
    "displayType": "string",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-env.xml",
    "category": "General",
    "index": 2
  },
  {
    "id": "site property",
    "name": "metrics_collector_pid_dir",
    "displayName": "Metrics Collector pid dir",
    "description": "\n      pid location for collector\n    ",
    "defaultValue": "embedded",
    "displayType": "string",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-env.xml",
    "category": "General",
    "index": 3
  },
  {
    "id": "site property",
    "name": "metrics_monitor_log_dir",
    "displayName": "Metrics Monitor log dir",
    "description": "\n      Log location for monitor logs\n    ",
    "defaultValue": "embedded",
    "displayType": "string",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-env.xml",
    "category": "General",
    "index": 4
  },
  {
    "id": "site property",
    "name": "metrics_monitor_pid_dir",
    "displayName": "Metrics Monitor pid dir",
    "description": "\n      pid location for monitor\n    ",
    "defaultValue": "embedded",
    "displayType": "string",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-env.xml",
    "category": "General",
    "index": 5
  },
  {
    "id": "site property",
    "name": "timeline.metrics.aggregator.checkpoint.dir",
    "displayName": "Aggregator checkpoint directory",
    "description": "\n      Directory to store aggregator checkpoints\n    ",
    "defaultValue": "/tmp",
    "displayType": "directory",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 17
  },
  {
    "id": "site property",
    "name": "timeline.metrics.cluster.aggregator.hourly.checkpointCutOffMultiplier",
    "displayName": "Hourly cluster aggregator checkpoint cutOff multiplier",
    "description": "\n      Multiplier value * interval = Max allowed checkpoint lag. Effectively\n      if aggregator checkpoint is greater than max allowed checkpoint delay,\n      the checkpoint will be discarded by the aggregator.\n    ",
    "defaultValue": "2",
    "displayType": "int",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 16
  },
  {
    "id": "site property",
    "name": "timeline.metrics.cluster.aggregator.hourly.disabled",
    "displayName": "Disable Hourly cluster aggregator",
    "description": "\n      Disable cluster based hourly aggregations.\n    ",
    "defaultValue": "false",
    "displayType": "string",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 14
  },
  {
    "id": "site property",
    "name": "timeline.metrics.cluster.aggregator.hourly.interval",
    "displayName": "Hourly cluster aggregator Interval",
    "description": "\n      Time in seconds to sleep for the hourly resolution cluster wide\n      aggregator. Default is 1 hour.\n    ",
    "defaultValue": "3600",
    "displayType": "int",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 15
  },
  {
    "id": "site property",
    "name": "timeline.metrics.cluster.aggregator.minute.checkpointCutOffMultiplier",
    "displayName": "Minute cluster aggregator checkpoint cutOff multiplier",
    "description": "\n      Multiplier value * interval = Max allowed checkpoint lag. Effectively\n      if aggregator checkpoint is greater than max allowed checkpoint delay,\n      the checkpoint will be discarded by the aggregator.\n    ",
    "defaultValue": "2",
    "displayType": "int",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 13
  },
  {
    "id": "site property",
    "name": "timeline.metrics.cluster.aggregator.minute.disabled",
    "displayName": "Disable minute cluster aggregator",
    "description": "\n      Disable cluster based minute aggregations.\n    ",
    "defaultValue": "false",
    "displayType": "string",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 10
  },
  {
    "id": "site property",
    "name": "timeline.metrics.cluster.aggregator.minute.interval",
    "displayName": "Minute cluster aggregator interval",
    "description": "\n      Time in seconds to sleep for the minute resolution cluster wide\n      aggregator. Default resolution is 2 minutes.\n    ",
    "defaultValue": "120",
    "displayType": "int",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 11
  },
  {
    "id": "site property",
    "name": "timeline.metrics.cluster.aggregator.minute.timeslice.interval",
    "displayName": "Minute cluster aggregator timeslice interval",
    "description": "\n      Lowest resolution of desired data for cluster level minute aggregates.\n    ",
    "defaultValue": "15",
    "displayType": "int",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 12
  },
  {
    "id": "site property",
    "name": "timeline.metrics.host.aggregator.hourly.checkpointCutOffMultiplier",
    "displayName": "Hourly host aggregator checkpoint cutOff multiplier",
    "description": "\n      Multiplier value * interval = Max allowed checkpoint lag. Effectively\n      if aggregator checkpoint is greater than max allowed checkpoint delay,\n      the checkpoint will be discarded by the aggregator.\n    ",
    "defaultValue": "2",
    "displayType": "int",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 9
  },
  {
    "id": "site property",
    "name": "timeline.metrics.host.aggregator.hourly.disabled",
    "displayName": "Disable Hourly host aggregator",
    "description": "\n      Disable host based hourly aggregations\n    ",
    "defaultValue": "false",
    "displayType": "string",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 7
  },
  {
    "id": "site property",
    "name": "timeline.metrics.host.aggregator.hourly.interval",
    "displayName": "Hourly host aggregator interval",
    "description": "\n      Time in seconds to sleep for the hourly resolution host based\n      aggregator. Default resolution is 1 hour.\n    ",
    "defaultValue": "3600",
    "displayType": "int",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 8
  },
  {
    "id": "site property",
    "name": "timeline.metrics.host.aggregator.minute.checkpointCutOffMultiplier",
    "displayName": "Minute host aggregator checkpoint cutOff multiplier",
    "description": "\n      Multiplier value * interval = Max allowed checkpoint lag. Effectively\n      if aggregator checkpoint is greater than max allowed checkpoint delay,\n      the checkpoint will be discarded by the aggregator.\n    ",
    "defaultValue": "2",
    "displayType": "int",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 6
  },
  {
    "id": "site property",
    "name": "timeline.metrics.host.aggregator.minute.disabled",
    "displayName": "Disable Minute host aggregator",
    "description": "\n      Disable host based minute aggregations.\n    ",
    "defaultValue": "false",
    "displayType": "string",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 4
  },
  {
    "id": "site property",
    "name": "timeline.metrics.host.aggregator.minute.interval",
    "displayName": "Minute host aggregator interval",
    "description": "\n      Time in seconds to sleep for the minute resolution host based\n      aggregator. Default resolution is 5 minutes.\n    ",
    "defaultValue": "300",
    "displayType": "int",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 5
  },
  {
    "id": "site property",
    "name": "timeline.metrics.service.checkpointDelay",
    "displayName": "Metrics service checkpoint delay",
    "description": "\n      Time in seconds to sleep on the first run or when the checkpoint is\n      too old.\n    ",
    "defaultValue": "120",
    "displayType": "int",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 1
  },
  {
    "id": "site property",
    "name": "timeline.metrics.service.default.result.limit",
    "displayName": "Metrics service default result limit",
    "description": "\n      Max result limit on number of rows returned. Calculated as follows:\n      4 aggregate metrics/min * 60 * 24: Retrieve aggregate data for 1 day.\n    ",
    "defaultValue": "5760",
    "displayType": "int",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 2
  },
  {
    "id": "site property",
    "name": "timeline.metrics.service.resultset.fetchSize",
    "displayName": "Metrics service resultset fetchSize",
    "description": "\n      JDBC resultset prefect size for aggregator queries.\n    ",
    "defaultValue": "2000",
    "displayType": "int",
    "serviceName": "AMBARI_METRICS",
    "filename": "ams-site.xml",
    "category": "MetricCollector",
    "index": 3
  },
  {
    "id": "site property",
    "name": "ams.zookeeper.keytab",
    "displayName": "ams.zookeeper.keytab",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "ams.zookeeper.principal",
    "displayName": "ams.zookeeper.principal",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hadoop.security.authentication",
    "displayName": "hadoop.security.authentication",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.coprocessor.master.classes",
    "displayName": "hbase.coprocessor.master.classes",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.coprocessor.region.classes",
    "displayName": "hbase.coprocessor.region.classes",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.master.kerberos.principal",
    "displayName": "hbase.master.kerberos.principal",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.master.keytab.file",
    "displayName": "hbase.master.keytab.file",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.myclient.keytab",
    "displayName": "hbase.myclient.keytab",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.myclient.principal",
    "displayName": "hbase.myclient.principal",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.regionserver.kerberos.principal",
    "displayName": "hbase.regionserver.kerberos.principal",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.regionserver.keytab.file",
    "displayName": "hbase.regionserver.keytab.file",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.security.authentication",
    "displayName": "hbase.security.authentication",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.security.authorization",
    "displayName": "hbase.security.authorization",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.zookeeper.property.authProvider.1",
    "displayName": "hbase.zookeeper.property.authProvider.1",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.zookeeper.property.jaasLoginRenew",
    "displayName": "hbase.zookeeper.property.jaasLoginRenew",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.zookeeper.property.kerberos.removeHostFromPrincipal",
    "displayName": "hbase.zookeeper.property.kerberos.removeHostFromPrincipal",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "hbase.zookeeper.property.kerberos.removeRealmFromPrincipal",
    "displayName": "hbase.zookeeper.property.kerberos.removeRealmFromPrincipal",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
  {
    "id": "site property",
    "name": "zookeeper.znode.parent",
    "displayName": "zookeeper.znode.parent",
    "isRequired": false,
    "serviceName": "AMBARI_METRICS",
    "category": "Advanced ams-hbase-security-site",
    "filename": "ams-hbase-security-site.xml"
  },
/************************************************Kerberos Descriptor******************************************/
  {
    "name": "smokeuser_principal_name",
    "displayName": "Smoke user principal",
    "category": "Ambari Principals",
    "index": 1
  },
  {
    "name": "smokeuser_keytab",
    "displayName": "Smoke user keytab",
    "category": "Ambari Principals",
    "index": 2
  },
  {
    "name": "hdfs_principal_name",
    "displayName": "HDFS user principal",
    "category": "Ambari Principals",
    "index": 3
  },
  {
    "name": "hdfs_user_keytab",
    "displayName": "HDFS user keytab",
    "category": "Ambari Principals",
    "index": 4
  },
  {
    "name": "hbase_principal_name",
    "displayName": "HBase user principal",
    "category": "Ambari Principals",
    "index": 5
  },
  {
    "name": "hbase_user_keytab",
    "displayName": "HBase user keytab",
    "category": "Ambari Principals",
    "index": 6
  }
];

if (App.get('isHadoopWindowsStack')) {
  hdp2properties.push(
    {
      "id": "puppet var",
      "name": "hadoop.user.name",
      "displayName": "Hadoop User Name",
      "description": "User to run Hadoop services under",
      "defaultValue": "hadoop",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": App.get('isHadoopWindowsStack'),
      "serviceName": "MISC",
      "filename": "cluster-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["HDFS"],
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "hadoop.user.password",
      "displayName": "Hadoop User Password",
      "description": "Password for hadoop user",
      "isReconfigurable": false,
      "displayType": "password",
      "isOverridable": false,
      "isVisible": App.get('isHadoopWindowsStack'),
      "serviceName": "MISC",
      "filename": "cluster-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["HDFS"],
      "index": 1
    }
  );
}

var atsProperties = [
  {
    "id": "site property",
    "name": "yarn.timeline-service.enabled",
    "displayName": "yarn.timeline-service.enabled",
    "category": "APP_TIMELINE_SERVER",
    "displayType": "checkbox",
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "id": "site property",
    "name": "yarn.timeline-service.leveldb-timeline-store.path",
    "displayName": "yarn.timeline-service.leveldb-timeline-store.path",
    "defaultDirectory": "/hadoop/yarn/timeline",
    "category": "APP_TIMELINE_SERVER",
    "displayType": "directory",
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "id": "site property",
    "name": "yarn.timeline-service.leveldb-timeline-store.ttl-interval-ms",
    "displayName": "yarn.timeline-service.leveldb-timeline-store.ttl-interval-ms",
    "displayType": "int",
    "category": "APP_TIMELINE_SERVER",
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "id": "site property",
    "name": "yarn.timeline-service.store-class",
    "displayName": "yarn.timeline-service.store-class",
    "category": "APP_TIMELINE_SERVER",
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "id": "site property",
    "name": "yarn.timeline-service.ttl-enable",
    "displayName": "yarn.timeline-service.ttl-enable",
    "displayType": "checkbox",
    "category": "APP_TIMELINE_SERVER",
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "id": "site property",
    "name": "yarn.timeline-service.ttl-ms",
    "displayName": "yarn.timeline-service.ttl-ms",
    "displayType": "int",
    "category": "APP_TIMELINE_SERVER",
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "id": "site property",
    "name": "yarn.timeline-service.generic-application-history.store-class",
    "displayName": "yarn.timeline-service.generic-application-history.store-class",
    "category": "APP_TIMELINE_SERVER",
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "id": "site property",
    "name": "yarn.timeline-service.webapp.address",
    "displayName": "yarn.timeline-service.webapp.address",
    "displayType": "string",
    "category": "APP_TIMELINE_SERVER",
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "id": "site property",
    "name": "yarn.timeline-service.webapp.https.address",
    "displayName": "yarn.timeline-service.webapp.https.address",
    "displayType": "string",
    "category": "APP_TIMELINE_SERVER",
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  },
  {
    "id": "site property",
    "name": "yarn.timeline-service.address",
    "displayName": "yarn.timeline-service.address",
    "displayType": "string",
    "category": "APP_TIMELINE_SERVER",
    "serviceName": "YARN",
    "filename": "yarn-site.xml"
  }
];

if (!App.get('isHadoop20Stack')) {
  hdp2properties.pushObjects(atsProperties);
}

module.exports =
{
  "configProperties": hdp2properties
};

});

require.register("data/PHD/site_properties", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

module.exports =
{
  "configProperties": [
  //***************************************** PHD stack **************************************
  /**********************************************HDFS***************************************/
    {
      "id": "site property",
      "name": "dfs.namenode.checkpoint.dir",
      "displayName": "SecondaryNameNode Checkpoint directories",
      "defaultDirectory": "/hadoop/hdfs/namesecondary",
      "displayType": "directories",
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "SECONDARY_NAMENODE",
      "index": 1
    },
    {
      "id": "site property",
      "name": "dfs.namenode.checkpoint.period",
      "displayName": "HDFS Maximum Checkpoint Delay",
      "displayType": "int",
      "unit": "seconds",
      "category": "General",
      "serviceName": "HDFS",
      "index": 3
    },
    {
      "id": "site property",
      "name": "dfs.namenode.name.dir",
      "displayName": "NameNode directories",
      "defaultDirectory": "/hadoop/hdfs/namenode",
      "displayType": "directories",
      "isOverridable": false,
      "serviceName": "HDFS",
      "category": "NAMENODE",
      "index": 1
    },
    {
      "id": "site property",
      "name": "dfs.webhdfs.enabled",
      "displayName": "WebHDFS enabled",
      "displayType": "checkbox",
      "isOverridable": false,
      "category": "General",
      "serviceName": "HDFS",
      "index": 0
    },
    {
      "id": "site property",
      "name": "dfs.datanode.failed.volumes.tolerated",
      "displayName": "DataNode volumes failure toleration",
      "displayType": "int",
      "category": "DATANODE",
      "serviceName": "HDFS",
      "index": 3
    },
    {
      "id": "site property",
      "name": "dfs.datanode.data.dir.mount.file",
      "displayName": "File that stores mount point for each data dir",
      "description": "File path that contains the last known mount point for each data dir. This file is used to avoid creating a DFS data dir on the root drive (and filling it up) if a path was previously mounted on a drive.",
      "defaultValue": "/etc/hadoop/conf/dfs_data_dir_mount.hist",
      "displayType": "directory",
      "isVisible": true,
      "category": "DATANODE",
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "index": 4
    },
    {
      "id": "site property",
      "name": "dfs.datanode.data.dir",
      "displayName": "DataNode directories",
      "defaultDirectory": "/hadoop/hdfs/data",
      "displayType": "directories",
      "category": "DATANODE",
      "serviceName": "HDFS",
      "index": 1
    },
    {
      "id": "site property",
      "name": "dfs.datanode.data.dir.perm",
      "displayName": "DataNode directories permission",
      "displayType": "int",
      "category": "DATANODE",
      "serviceName": "HDFS"
    },
    {
      "id": "site property",
      "name": "dfs.replication",
      "displayName": "Block replication",
      "displayType": "int",
      "category": "General",
      "serviceName": "HDFS"
    },
    {
      "id": "site property",
      "name": "dfs.datanode.du.reserved",
      "displayName": "Reserved space for HDFS",
      "displayType": "int",
      "unit": "bytes",
      "category": "General",
      "serviceName": "HDFS",
      "index": 2
    },
    {
      "id": "site property",
      "name": "dfs.client.read.shortcircuit",
      "displayName": "HDFS Short-circuit read",
      "displayType": "checkbox",
      "category": "Advanced hdfs-site",
      "serviceName": "HDFS"
    },
    {
      "id": "site property",
      "name": "apache_artifacts_download_url",
      "displayName": "apache_artifacts_download_url",
      "description": "",
      "isRequired": false,
      "isRequiredByAgent": false,
      "isVisible": false,
      "category": "Advanced hdfs-site",
      "serviceName": "HDFS"
    },
    {
      "id": "site property",
      "name": "gpl_artifacts_download_url",
      "displayName": "gpl_artifacts_download_url",
      "description": "",
      "isRequiredByAgent": false,
      "isRequired": false,
      "isVisible": false,
      "category": "Advanced hdfs-site",
      "serviceName": "HDFS"
    },

  /**********************************************YARN***************************************/
    {
      "id": "site property",
      "name": "yarn.acl.enable",
      "displayName": "yarn.acl.enable",
      "displayType": "checkbox",
      "serviceName": "YARN",
      "category": "RESOURCEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.admin.acl",
      "displayName": "yarn.admin.acl",
      "isRequired": false,
      "serviceName": "YARN",
      "category": "RESOURCEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.log-aggregation-enable",
      "displayName": "yarn.log-aggregation-enable",
      "displayType": "checkbox",
      "serviceName": "YARN",
      "category": "RESOURCEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.resourcemanager.scheduler.class",
      "displayName": "yarn.resourcemanager.scheduler.class",
      "serviceName": "YARN",
      "category": "CapacityScheduler"
    },
    {
      "id": "site property",
      "name": "yarn.scheduler.minimum-allocation-mb",
      "displayName": "yarn.scheduler.minimum-allocation-mb",
      "displayType": "int",
      "serviceName": "YARN",
      "category": "CapacityScheduler"
    },
    {
      "id": "site property",
      "name": "yarn.scheduler.maximum-allocation-mb",
      "displayName": "yarn.scheduler.maximum-allocation-mb",
      "displayType": "int",
      "serviceName": "YARN",
      "category": "CapacityScheduler"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.resource.memory-mb",
      "displayName": "yarn.nodemanager.resource.memory-mb",
      "displayType": "int",
      "serviceName": "YARN",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.vmem-pmem-ratio",
      "displayName": "yarn.nodemanager.vmem-pmem-ratio",
      "serviceName": "YARN",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.linux-container-executor.group",
      "displayName": "yarn.nodemanager.linux-container-executor.group",
      "serviceName": "YARN",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.log-dirs",
      "displayName": "yarn.nodemanager.log-dirs",
      "defaultDirectory": "/hadoop/yarn/log",
      "displayType": "directories",
      "serviceName": "YARN",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.local-dirs",
      "displayName": "yarn.nodemanager.local-dirs",
      "defaultDirectory": "/hadoop/yarn/local",
      "displayType": "directories",
      "serviceName": "YARN",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.remote-app-log-dir",
      "displayName": "yarn.nodemanager.remote-app-log-dir",
      "displayType": "directory",
      "serviceName": "YARN",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.remote-app-log-dir-suffix",
      "displayName": "yarn.nodemanager.remote-app-log-dir-suffix",
      "serviceName": "YARN",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.aux-services",
      "displayName": "yarn.nodemanager.aux-services",
      "serviceName": "YARN",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.nodemanager.log.retain-second",
      "displayName": "yarn.nodemanager.log.retain-second",
      "serviceName": "YARN",
      "category": "NODEMANAGER"
    },
    {
      "id": "site property",
      "name": "yarn.log.server.url",
      "displayName": "yarn.log.server.url",
      "category": "Advanced yarn-site",
      "serviceName": "YARN"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.enabled",
      "displayName": "yarn.timeline-service.enabled",
      "category": "APP_TIMELINE_SERVER",
      "displayType": "checkbox",
      "serviceName": "YARN"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.leveldb-timeline-store.path",
      "displayName": "yarn.timeline-service.leveldb-timeline-store.path",
      "defaultDirectory": "/hadoop/yarn/timeline",
      "category": "APP_TIMELINE_SERVER",
      "displayType": "directory",
      "serviceName": "YARN"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.leveldb-timeline-store.ttl-interval-ms",
      "displayName": "yarn.timeline-service.leveldb-timeline-store.ttl-interval-ms",
      "displayType": "int",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.store-class",
      "displayName": "yarn.timeline-service.store-class",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.ttl-enable",
      "displayName": "yarn.timeline-service.ttl-enable",
      "displayType": "checkbox",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.ttl-ms",
      "displayName": "yarn.timeline-service.ttl-ms",
      "displayType": "int",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.generic-application-history.store-class",
      "displayName": "yarn.timeline-service.generic-application-history.store-class",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.webapp.address",
      "displayName": "yarn.timeline-service.webapp.address",
      "displayType": "string",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.webapp.https.address",
      "displayName": "yarn.timeline-service.webapp.https.address",
      "displayType": "string",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN"
    },
    {
      "id": "site property",
      "name": "yarn.timeline-service.address",
      "displayName": "yarn.timeline-service.address",
      "displayType": "string",
      "category": "APP_TIMELINE_SERVER",
      "serviceName": "YARN"
    },
  /**********************************************MAPREDUCE2***************************************/
    {
      "id": "site property",
      "name": "mapreduce.map.memory.mb",
      "displayName": "Default virtual memory for a job's map-task",
      "displayType": "int",
      "unit": "MB",
      "category": "General",
      "serviceName": "MAPREDUCE2"
    },
    {
      "id": "site property",
      "name": "mapreduce.reduce.memory.mb",
      "displayName": "Default virtual memory for a job's reduce-task",
      "displayType": "int",
      "unit": "MB",
      "category": "General",
      "serviceName": "MAPREDUCE2"
    },
    {
      "id": "site property",
      "name": "mapreduce.task.io.sort.mb",
      "displayName": "Map-side sort buffer memory",
      "displayType": "int",
      "unit": "MB",
      "category": "General",
      "serviceName": "MAPREDUCE2"
    },
    {
      "id": "site property",
      "name": "hadoop.security.auth_to_local",
      "displayName": "hadoop.security.auth_to_local",
      "displayType": "multiLine",
      "serviceName": "HDFS",
      "category": "Advanced core-site"
    },
    {
      "id": "site property",
      "name": "yarn.app.mapreduce.am.resource.mb",
      "displayName": "yarn.app.mapreduce.am.resource.mb",
      "displayType": "int",
      "category": "Advanced mapred-site",
      "serviceName": "MAPREDUCE2"
    },

  /**********************************************oozie-site***************************************/
    {
      "id": "site property",
      "name": "oozie.db.schema.name",
      "displayName": "Database Name",
      "isOverridable": false,
      "displayType": "host",
      "isObserved": true,
      "category": "OOZIE_SERVER",
      "serviceName": "OOZIE",
      "index": 4
    },
    {
      "id": "site property",
      "name": "oozie.service.JPAService.jdbc.username",
      "displayName": "Database Username",
      "isOverridable": false,
      "displayType": "host",
      "category": "OOZIE_SERVER",
      "serviceName": "OOZIE",
      "index": 5
    },
    {
      "id": "site property",
      "name": "oozie.service.JPAService.jdbc.password",
      "displayName": "Database Password",
      "isOverridable": false,
      "displayType": "password",
      "category": "OOZIE_SERVER",
      "serviceName": "OOZIE",
      "filename": "oozie-site.xml",
      "index": 6
    },
    {
      "id": "site property",
      "name": "oozie.service.JPAService.jdbc.driver", // the default value of this property is overriden in code
      "displayName": "JDBC Driver Class",
      "isOverridable": false,
      "category": "OOZIE_SERVER",
      "serviceName": "OOZIE",
      "index": 7
    },
    {
      "id": "site property",
      "name": "oozie.service.JPAService.jdbc.url",
      "displayName": "Database URL",
      "isOverridable": false,
      "displayType": "advanced",
      "category": "OOZIE_SERVER",
      "serviceName": "OOZIE",
      "index": 8
    },

  /**********************************************hive-site***************************************/
    {
      "id": "site property",
      "name": "javax.jdo.option.ConnectionDriverName",  // the default value is overwritten in code
      "displayName": "JDBC Driver Class",
      "isOverridable": false,
      "category": "HIVE_METASTORE",
      "serviceName": "HIVE",
      "index": 7
    },
    {
      "id": "site property",
      "name": "hive.heapsize",
      "displayName": "Hive heap size",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "serviceName": "HIVE",
      "category": "General",
      "index": 9
    },
    {
      "id": "site property",
      "name": "javax.jdo.option.ConnectionUserName",
      "displayName": "Database Username",
      "displayType": "host",
      "isOverridable": false,
      "category": "HIVE_METASTORE",
      "serviceName": "HIVE",
      "index": 5
    },
    {
      "id": "site property",
      "name": "javax.jdo.option.ConnectionPassword",
      "displayName": "Database Password",
      "displayType": "password",
      "isOverridable": false,
      "category": "HIVE_METASTORE",
      "serviceName": "HIVE",
      "index": 6
    },
    {
      "id": "site property",
      "name": "javax.jdo.option.ConnectionURL",
      "displayName": "Database URL",
      "displayType": "advanced",
      "isOverridable": false,
      "category": "HIVE_METASTORE",
      "serviceName": "HIVE",
      "index": 8
    },
    {
      "id": "site property",
      "name": "ambari.hive.db.schema.name",
      "displayName": "Database Name",
      "displayType": "host",
      "isOverridable": false,
      "isObserved": true,
      "serviceName": "HIVE",
      "category": "HIVE_METASTORE",
      "index": 4
    },
    {
      "id": "site property",
      "name": "hive.server2.tez.default.queues",
      "displayName": "hive.server2.tez.default.queues",
      "isRequired": false,
      "serviceName": "HIVE",
      "category": "Advanced hive-site"
    },
    {
      "id": "site property",
      "name": "hive.server2.thrift.port",
      "displayName": "Hive Server Port",
      "description": "TCP port number to listen on, default 10000.",
      "defaultValue": "10000",
      "displayType": "int",
      "isReconfigurable": true,
      "isOverridable": false,
      "isVisible": true,
      "category": "Advanced hive-site",
      "serviceName": "HIVE"
    },
    {
      "id": "site property",
      "name": "hive.server2.support.dynamic.service.discovery",
      "displayName": "hive.server2.support.dynamic.service.discovery",
      "defaultValue": true,
      "displayType": "checkbox",
      "category": "Advanced hive-site",
      "serviceName": "HIVE"
    },
    {
      "id": "site property",
      "name": "hive.zookeeper.quorum",
      "displayName": "hive.zookeeper.quorum",
      "defaultValue": "localhost:2181",
      "displayType": "multiLine",
      "isVisible": true,
      "serviceName": "HIVE",
      "category": "Advanced hive-site"
    },
  /**********************************************tez-site*****************************************/
    {
      "id": "site property",
      "name": "tez.am.resource.memory.mb",
      "displayName": "tez.am.resource.memory.mb",
      "displayType": "int",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.am.java.opts",
      "displayName": "tez.am.java.opts",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.am.grouping.split-waves",
      "displayName": "tez.am.grouping.split-waves",
      "displayType": "float",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.am.grouping.min-size",
      "displayName": "tez.am.grouping.min-size",
      "displayType": "int",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.am.grouping.max-size",
      "displayName": "tez.am.grouping.max-size",
      "displayType": "int",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.am.log.level",
      "displayName": "tez.am.log.level",
      "displayType": "string",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.runtime.intermediate-input.compress.codec",
      "displayName": "tez.runtime.intermediate-input.compress.codec",
      "displayType": "string",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.runtime.intermediate-input.is-compressed",
      "displayName": "tez.runtime.intermediate-input.is-compressed",
      "displayType": "checkbox",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.runtime.intermediate-output.compress.codec",
      "displayName": "tez.runtime.intermediate-output.compress.codec",
      "displayType": "string",
      "category": "General",
      "serviceName": "TEZ"
    },
    {
      "id": "site property",
      "name": "tez.runtime.intermediate-output.should-compress",
      "displayName": "tez.runtime.intermediate-output.should-compress",
      "displayType": "checkbox",
      "category": "General",
      "serviceName": "TEZ"
    },

  /**********************************************hbase-site***************************************/
    {
      "id": "site property",
      "name": "hbase.tmp.dir",
      "displayName": "HBase local directory",
      "defaultDirectory": "/hadoop/hbase",
      "displayType": "directory",
      "category": "Advanced hbase-site",
      "serviceName": "HBASE"

    },
    {
      "id": "site property",
      "name": "hbase.master.port",
      "displayName": "HBase Master Port",
      "isReconfigurable": true,
      "displayType": "int",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HBASE",
      "category": "Advanced hbase-site"
    },
    {
      "id": "site property",
      "name": "hbase.regionserver.global.memstore.upperLimit",
      "displayName": "hbase.regionserver.global.memstore.upperLimit",
      "displayType": "float",
      "category": "Advanced hbase-site",
      "serviceName": "HBASE"
    },
    {
      "id": "site property",
      "name": "hbase.regionserver.global.memstore.lowerLimit",
      "displayName": "hbase.regionserver.global.memstore.lowerLimit",
      "displayType": "float",
      "category": "Advanced hbase-site",
      "serviceName": "HBASE"
    },
    {
      "id": "site property",
      "name": "hbase.hstore.blockingStoreFiles",
      "displayName": "hstore blocking storefiles",
      "displayType": "int",
      "category": "Advanced hbase-site",
      "serviceName": "HBASE"
    },
    {
      "id": "site property",
      "name": "hbase.hstore.compactionThreshold",
      "displayName": "HBase HStore compaction threshold",
      "displayType": "int",
      "category": "General",
      "serviceName": "HBASE",
      "index": 0
    },
    {
      "id": "site property",
      "name": "hfile.block.cache.size",
      "displayName": "HFile block cache size ",
      "displayType": "float",
      "category": "General",
      "serviceName": "HBASE",
      "index": 1
    },
    {
      "id": "site property",
      "name": "hbase.hregion.max.filesize",
      "displayName": "Maximum HStoreFile Size",
      "displayType": "int",
      "unit": "bytes",
      "category": "General",
      "serviceName": "HBASE",
      "index": 2
    },
    {
      "id": "site property",
      "name": "hbase.regionserver.handler.count",
      "displayName": "RegionServer Handler",
      "displayType": "int",
      "category": "HBASE_REGIONSERVER",
      "serviceName": "HBASE",
      "index": 2
    },
    {
      "id": "site property",
      "name": "hbase.hregion.majorcompaction",
      "displayName": "HBase Region Major Compaction",
      "displayType": "int",
      "unit": "ms",
      "category": "HBASE_REGIONSERVER",
      "serviceName": "HBASE",
      "index": 3
    },
    {
      "id": "site property",
      "name": "hbase.hregion.memstore.block.multiplier",
      "displayName": "HBase Region Block Multiplier",
      "displayType": "int",
      "category": "HBASE_REGIONSERVER",
      "serviceName": "HBASE",
      "index": 4
    },
    {
      "id": "site property",
      "name": "hbase.hregion.memstore.mslab.enabled",
      "displayName": "hbase.hregion.memstore.mslab.enabled",
      "displayType": "checkbox",
      "category": "Advanced hbase-site",
      "serviceName": "HBASE"
    },
    {
      "id": "site property",
      "name": "hbase.hregion.memstore.flush.size",
      "displayName": "HBase Region Memstore Flush Size",
      "displayType": "int",
      "unit": "bytes",
      "category": "HBASE_REGIONSERVER",
      "serviceName": "HBASE",
      "index": 5
    },
    {
      "id": "site property",
      "name": "hbase.client.scanner.caching",
      "displayName": "HBase Client Scanner Caching",
      "displayType": "int",
      "unit": "rows",
      "category": "General",
      "serviceName": "HBASE",
      "index": 3
    },
    {
      "id": "site property",
      "name": "zookeeper.session.timeout",
      "displayName": "Zookeeper timeout for HBase Session",
      "displayType": "int",
      "unit": "ms",
      "category": "General",
      "serviceName": "HBASE",
      "index": 4
    },
    {
      "id": "site property",
      "name": "hbase.client.keyvalue.maxsize",
      "displayName": "HBase Client Maximum key-value Size",
      "displayType": "int",
      "unit": "bytes",
      "category": "General",
      "serviceName": "HBASE",
      "index": 5
    },
    {
      "id": "site property",
      "name": "hbase.coprocessor.region.classes",
      "displayName": "hbase.coprocessor.region.classes",
      "category": "Advanced hbase-site",
      "isRequired": false,
      "serviceName": "HBASE"
    },
    {
      "id": "site property",
      "name": "hbase.coprocessor.master.classes",
      "displayName": "hbase.coprocessor.master.classes",
      "category": "Advanced hbase-site",
      "isRequired": false,
      "serviceName": "HBASE"
    },
    {
      "id": "site property",
      "name": "hbase.zookeeper.quorum",
      "displayName": "hbase.zookeeper.quorum",
      "displayType": "multiLine",
      "serviceName": "HBASE",
      "category": "Advanced hbase-site"
    },

  /**********************************************storm-site***************************************/
    {
      "id": "site property",
      "name": "storm.zookeeper.root",
      "displayName": "storm.zookeeper.root",
      "displayType": "directory",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.local.dir",
      "displayName": "storm.local.dir",
      "defaultDirectory": "/hadoop/storm",
      "displayType": "directory",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.servers",
      "displayName": "storm.zookeeper.servers",
      "displayType": "masterHosts",
      "isOverridable": false,
      "isReconfigurable": false,
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.port",
      "displayName": "storm.zookeeper.port",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.session.timeout",
      "displayName": "storm.zookeeper.session.timeout",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.connection.timeout",
      "displayName": "storm.zookeeper.connection.timeout",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.retry.times",
      "displayName": "storm.zookeeper.retry.times",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.retry.interval",
      "displayName": "storm.zookeeper.retry.interval",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "General",
      "unit": "ms"
    },
    {
      "id": "site property",
      "name": "storm.zookeeper.retry.intervalceiling.millis",
      "displayName": "storm.zookeeper.retry.intervalceiling.millis",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "General",
      "unit": "ms"
    },
    {
      "id": "site property",
      "name": "storm.cluster.mode",
      "displayName": "storm.cluster.mode",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.local.mode.zmq",
      "displayName": "storm.local.mode.zmq",
      "displayType": "checkbox",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.thrift.transport",
      "displayName": "storm.thrift.transport",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "storm.messaging.transport",
      "displayName": "storm.messaging.transport",
      "serviceName": "STORM",
      "category": "General"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "General",
      "displayName": "storm.messaging.netty.buffer_size",
      "name": "storm.messaging.netty.buffer_size",
      "displayType": "int",
      "unit": "bytes"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "General",
      "displayName": "storm.messaging.netty.max_retries",
      "name": "storm.messaging.netty.max_retries",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "General",
      "displayName": "storm.messaging.netty.max_wait_ms",
      "name": "storm.messaging.netty.max_wait_ms",
      "displayType": "int",
      "unit": "ms"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "General",
      "displayName": "storm.messaging.netty.min_wait_ms",
      "name": "storm.messaging.netty.min_wait_ms",
      "displayType": "int",
      "unit": "ms"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "General",
      "displayName": "storm.messaging.netty.server_worker_threads",
      "name": "storm.messaging.netty.server_worker_threads",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "General",
      "displayName": "storm.messaging.netty.client_worker_threads",
      "name": "storm.messaging.netty.client_worker_threads",
      "displayType": "int"
    },
    {
      "id": "site property",
      "name": "nimbus.host",
      "displayName": "nimbus.host",
      "displayType": "masterHost",
      "isOverridable": false,
      "isReconfigurable": false,
      "serviceName": "STORM",
      "category": "NIMBUS"
    },
    {
      "id": "site property",
      "name": "nimbus.thrift.port",
      "displayName": "nimbus.thrift.port",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS"
    },
    {
      "id": "site property",
      "name": "nimbus.thrift.max_buffer_size",
      "displayName": "nimbus.thrift.max_buffer_size",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "bytes"
    },
    {
      "id": "site property",
      "name": "nimbus.childopts",
      "displayName": "nimbus.childopts",
      "displayType": "multiLine",
      "isOverridable": false,
      "serviceName": "STORM",
      "category": "NIMBUS",
      "filename": "storm-site.xml"
    },
    {
      "id": "site property",
      "name": "nimbus.task.timeout.secs",
      "displayName": "nimbus.task.timeout.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.supervisor.timeout.secs",
      "displayName": "nimbus.supervisor.timeout.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.monitor.freq.secs",
      "displayName": "nimbus.monitor.freq.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.cleanup.inbox.freq.secs",
      "displayName": "nimbus.cleanup.inbox.freq.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.inbox.jar.expiration.secs",
      "displayName": "nimbus.inbox.jar.expiration.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.task.launch.secs",
      "displayName": "nimbus.task.launch.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.reassign",
      "displayName": "nimbus.reassign",
      "displayType": "checkbox",
      "isReconfigurable": true,
      "serviceName": "STORM",
      "category": "NIMBUS"
    },
    {
      "id": "site property",
      "name": "nimbus.file.copy.expiration.secs",
      "displayName": "nimbus.file.copy.expiration.secs",
      "displayType": "int",
      "serviceName": "STORM",
      "category": "NIMBUS",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "name": "nimbus.topology.validator",
      "displayName": "nimbus.topology.validator",
      "serviceName": "STORM",
      "category": "NIMBUS"
    },
    {
      "id": "site property",
      "name": "supervisor.slots.ports",
      "displayName": "supervisor.slots.ports",
      "displayType": "string",
      "serviceName": "STORM",
      "category": "SUPERVISOR"
    },
    {
      "id": "site property",
      "isOverrideable": false,
      "serviceName": "STORM",
      "category": "SUPERVISOR",
      "displayName": "supervisor.childopts",
      "name": "supervisor.childopts",
      "displayType": "multiLine",
      "filename": "storm-site.xml"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "SUPERVISOR",
      "displayName": "supervisor.worker.start.timeout.secs",
      "name": "supervisor.worker.start.timeout.secs",
      "displayType": "int",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "SUPERVISOR",
      "displayName": "supervisor.worker.timeout.secs",
      "name": "supervisor.worker.timeout.secs",
      "displayType": "int",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "SUPERVISOR",
      "displayName": "supervisor.monitor.frequency.secs",
      "name": "supervisor.monitor.frequency.secs",
      "displayType": "int",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "SUPERVISOR",
      "displayName": "supervisor.heartbeat.frequency.secs",
      "name": "supervisor.heartbeat.frequency.secs",
      "displayType": "int",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "DRPC_SERVER",
      "displayName": "drpc.port",
      "name": "drpc.port",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "DRPC_SERVER",
      "displayName": "drpc.worker.threads",
      "name": "drpc.worker.threads",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "DRPC_SERVER",
      "displayName": "drpc.queue.size",
      "name": "drpc.queue.size",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "DRPC_SERVER",
      "displayName": "drpc.invocations.port",
      "name": "drpc.invocations.port",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "DRPC_SERVER",
      "displayName": "drpc.request.timeout.secs",
      "name": "drpc.request.timeout.secs",
      "displayType": "int",
      "unit": "seconds"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "DRPC_SERVER",
      "displayName": "drpc.childopts",
      "name": "drpc.childopts",
      "displayType": "string"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "STORM_UI_SERVER",
      "displayName": "ui.port",
      "name": "ui.port",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "STORM_UI_SERVER",
      "displayName": "ui.childopts",
      "name": "ui.childopts",
      "displayType": "string"
    },
    //@Todo: uncomment following properties when logviewer is treated as different section on storm service page
    /*
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "LogviewerServer",
      "displayName": "logviewer.port",
      "name": "logviewer.port",
      "displayType": "int"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "LogviewerServer",
      "displayName": "logviewer.childopts",
      "name": "logviewer.childopts",
      "displayType": "string"
    },
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "LogviewerServer",
      "displayName": "logviewer.appender.name",
      "name": "logviewer.appender.name",
      "displayType": "string"
    },
    */
    {
      "id": "site property",
      "serviceName": "STORM",
      "category": "Advanced storm-site",
      "displayName": "worker.childopts",
      "name": "worker.childopts",
      "displayType": "multiLine",
      "filename": "storm-site.xml"
    },
  /*********************************************oozie-site for Falcon*****************************/
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-job-submit-instances",
      "name": "oozie.service.ELService.ext.functions.coord-job-submit-instances",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-action-create-inst",
      "name": "oozie.service.ELService.ext.functions.coord-action-create-inst",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-action-create",
      "name": "oozie.service.ELService.ext.functions.coord-action-create",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-job-submit-data",
      "name": "oozie.service.ELService.ext.functions.coord-job-submit-data",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-action-start",
      "name": "oozie.service.ELService.ext.functions.coord-action-start",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-sla-submit",
      "name": "oozie.service.ELService.ext.functions.coord-sla-submit",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },
    {
      "id": "site property",
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "category": "Falcon - Oozie integration",
      "displayName": "oozie.service.ELService.ext.functions.coord-sla-create",
      "name": "oozie.service.ELService.ext.functions.coord-sla-create",
      "displayType": "custom",
      "filename": "oozie-site.xml"
    },

    // Runtime properties
    {
      "id": "site property",
      "name": "*.domain",
      "displayName": "*.domain",
      "category": "FalconRuntimeSite",
      "serviceName": "FALCON",
      "filename": "falcon-runtime.properties.xml"

    },
    {
      "id": "site property",
      "name": "*.log.cleanup.frequency.minutes.retention",
      "displayName": "*.log.cleanup.frequency.minutes.retention",
      "category": "FalconRuntimeSite",
      "serviceName": "FALCON",
      "filename": "falcon-runtime.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.log.cleanup.frequency.hours.retention",
      "displayName": "*.log.cleanup.frequency.hours.retention",
      "category": "FalconRuntimeSite",
      "serviceName": "FALCON",
      "filename": "falcon-runtime.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.log.cleanup.frequency.days.retention",
      "displayName": "*.log.cleanup.frequency.days.retention",
      "category": "FalconRuntimeSite",
      "serviceName": "FALCON",
      "filename": "falcon-runtime.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.log.cleanup.frequency.months.retention",
      "displayName": "*.log.cleanup.frequency.months.retention",
      "category": "FalconRuntimeSite",
      "serviceName": "FALCON",
      "filename": "falcon-runtime.properties.xml"
    },

    //  Startup properties

    {
      "id": "site property",
      "name": "*.domain",
      "displayName": "*.domain",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.workflow.engine.impl",
      "displayName": "*.workflow.engine.impl",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.oozie.process.workflow.builder",
      "displayName": "*.oozie.process.workflow.builder",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.oozie.feed.workflow.builder",
      "displayName": "*.oozie.feed.workflow.builder",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.SchedulableEntityManager.impl",
      "displayName": "*.SchedulableEntityManager.impl",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.ConfigSyncService.impl",
      "displayName": "*.ConfigSyncService.impl",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.ProcessInstanceManager.impl",
      "displayName": "*.ProcessInstanceManager.impl",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.catalog.service.impl",
      "displayName": "*.catalog.service.impl",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.application.services",
      "displayName": "*.application.services",
      "displayType": "multiLine",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.configstore.listeners",
      "displayName": "*.configstore.listeners",
      "displayType": "multiLine",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.broker.impl.class",
      "displayName": "*.broker.impl.class",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.shared.libs",
      "displayName": "*.shared.libs",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.config.store.uri",
      "displayName": "*.config.store.uri",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.system.lib.location",
      "displayName": "*.system.lib.location",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.retry.recorder.path",
      "displayName": "*.retry.recorder.path",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.cleanup.service.frequency",
      "displayName": "*.falcon.cleanup.service.frequency",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.broker.url",
      "displayName": "*.broker.url",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.broker.ttlInMins",
      "displayName": "*.broker.ttlInMins",
      "displayType": "int",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.entity.topic",
      "displayName": "*.entity.topic",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.max.retry.failure.count",
      "displayName": "*.max.retry.failure.count",
      "displayType": "int",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.internal.queue.size",
      "displayName": "*.internal.queue.size",
      "displayType": "int",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.authentication.type",
      "displayName": "*.falcon.authentication.type",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.http.authentication.type",
      "displayName": "*.falcon.http.authentication.type",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.http.authentication.token.validity",
      "displayName": "*.falcon.http.authentication.token.validity",
      "displayType": "int",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.http.authentication.signature.secret",
      "displayName": "*.falcon.http.authentication.signature.secret",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.http.authentication.simple.anonymous.allowed",
      "displayName": "*.falcon.http.authentication.simple.anonymous.allowed",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.http.authentication.kerberos.name.rules",
      "displayName": "*.falcon.http.authentication.kerberos.name.rules",
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },
    {
      "id": "site property",
      "name": "*.falcon.http.authentication.blacklisted.users",
      "displayName": "*.falcon.http.authentication.blacklisted.users",
      "isRequired": false,
      "category": "FalconStartupSite",
      "serviceName": "FALCON",
      "filename": "falcon-startup.properties.xml"
    },

  /**********************************************webhcat-site***************************************/
    {
      "id": "site property",
      "name": "templeton.hive.archive",
      "displayName": "templeton.hive.archive",
      "isRequired": false,
      "serviceName": "HIVE",
      "category": "Advanced webhcat-site"
    },
    {
      "id": "site property",
      "name": "templeton.pig.archive",
      "displayName": "templeton.pig.archive",
      "isRequired": false,
      "serviceName": "HIVE",
      "category": "Advanced webhcat-site"
    },
    {
      "id": "site property",
      "name": "templeton.zookeeper.hosts",
      "displayName": "templeton.zookeeper.hosts",
      "displayType": "multiLine",
      "serviceName": "HIVE",
      "category": "Advanced webhcat-site"
    },
  /**********************************************pig.properties*****************************************/
    {
      "id": "site property",
      "name": "content",
      "displayName": "content",
      "value": "",
      "defaultValue": "",
      "description": "pig properties",
      "displayType": "content",
      "isRequired": false,
      "showLabel": false,
      "serviceName": "PIG",
      "filename": "pig-properties.xml",
      "category": "Advanced pig-properties"
    },
  /********************************************* flume-agent *****************************/
    {
      "id": "site property",
      "name": "content",
      "displayName": "content",
      "showLabel": false,
      "isRequired": false,
      "displayType": "content",
      "serviceName": "FLUME",
      "category": "FLUME_HANDLER",
      "filename": "flume-conf.xml"
    },
  /**********************************************HDFS***************************************/
    {
      "id": "puppet var",
      "name": "namenode_host",
      "displayName": "NameNode hosts",
      "value": "",
      "defaultValue": "",
      "description": "The hosts that has been assigned to run NameNode",
      "displayType": "masterHosts",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "NAMENODE",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "namenode_heapsize",
      "displayName": "NameNode Java heap size",
      "description": "Initial and maximum Java heap size for NameNode (Java options -Xms and -Xmx).  This also applies to the Secondary NameNode.",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "NAMENODE",
      "index": 2
    },
    {
      "id": "puppet var",
      "name": "namenode_opt_newsize",
      "displayName": "NameNode new generation size",
      "description": "Default size of Java new generation for NameNode (Java option -XX:NewSize).  This also applies to the Secondary NameNode.",
      "defaultValue": "200",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "NAMENODE",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "namenode_opt_maxnewsize",
      "displayName": "NameNode maximum new generation size",
      "description": "Maximum size of Java new generation for NameNode (Java option -XX:MaxnewSize).",
      "defaultValue": "200",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "NAMENODE",
      "index": 4
    },
    {
      "id": "puppet var",
      "name": "snamenode_host",
      "displayName": "SNameNode host",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run SecondaryNameNode",
      "displayType": "masterHost",
      "isOverridable": false,
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "SECONDARY_NAMENODE",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "datanode_hosts", //not in the schema. For UI purpose
      "displayName": "DataNode hosts",
      "value": "",
      "defaultValue": "",
      "description": "The hosts that have been assigned to run DataNode",
      "displayType": "slaveHosts",
      "isRequired": false,
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "DATANODE",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "dtnode_heapsize",
      "displayName": "DataNode maximum Java heap size",
      "description": "Maximum Java heap size for DataNode (Java option -Xmx)",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "DATANODE",
      "index": 2
    },
    {
      "id": "puppet var",
      "name": "hadoop_heapsize",
      "displayName": "Hadoop maximum Java heap size",
      "description": "Maximum Java heap size for daemons such as Balancer (Java option -Xmx)",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "General",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "hdfs_log_dir_prefix",
      "displayName": "Hadoop Log Dir Prefix",
      "description": "The parent directory for Hadoop log files.  The HDFS log directory will be ${hadoop_log_dir_prefix} / ${hdfs_user} and the MapReduce log directory will be ${hadoop_log_dir_prefix} / ${mapred_user}.",
      "defaultValue": "/var/log/hadoop",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "Advanced hadoop-env"
    },
    {
      "id": "puppet var",
      "name": "hadoop_pid_dir_prefix",
      "displayName": "Hadoop PID Dir Prefix",
      "description": "The parent directory in which the PID files for Hadoop processes will be created.  The HDFS PID directory will be ${hadoop_pid_dir_prefix} / ${hdfs_user} and the MapReduce PID directory will be ${hadoop_pid_dir_prefix} / ${mapred_user}.",
      "defaultValue": "/var/run/hadoop",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "Advanced hadoop-env"
    },
    {
      "id": "puppet var",
      "name": "hadoop_root_logger",
      "displayName": "Hadoop Root Logger",
      "description": "Hadoop logging options",
      "defaultValue": "INFO,RFA",
      "displayType": "string",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HDFS",
      "filename": "hadoop-env.xml",
      "category": "Advanced hadoop-env"
    },

  /**********************************************MAPREDUCE2***************************************/
    {
      "id": "puppet var",
      "name": "hs_host",
      "displayName": "History Server",
      "description": "History Server",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "masterHost",
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "MAPREDUCE2",
      "filename": "mapred-env.xml",
      "category": "HISTORYSERVER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "jobhistory_heapsize",
      "displayName": "History Server heap size",
      "description": "History Server heap size",
      "defaultValue": "900",
      "unit": "MB",
      "isOverridable": true,
      "displayType": "int",
      "isVisible": true,
      "serviceName": "MAPREDUCE2",
      "filename": "mapred-env.xml",
      "category": "HISTORYSERVER",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "mapred_log_dir_prefix",
      "displayName": "Mapreduce Log Dir Prefix",
      "description": "",
      "defaultValue": "/var/log/hadoop-mapreduce",
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isReconfigurable": false,
      "serviceName": "MAPREDUCE2",
      "filename": "mapred-env.xml",
      "category": "Advanced mapred-env"
    },
    {
      "id": "puppet var",
      "name": "mapred_pid_dir_prefix",
      "displayName": "Mapreduce PID Dir Prefix",
      "description": "",
      "defaultValue": "/var/run/hadoop-mapreduce",
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isReconfigurable": false,
      "serviceName": "MAPREDUCE2",
      "filename": "mapred-env.xml",
      "category": "Advanced mapred-env"
    },
  /**********************************************YARN***************************************/
    {
      "id": "puppet var",
      "name": "yarn_heapsize",
      "displayName": "YARN Java heap size",
      "description": "Max heapsize for all YARN components",
      "defaultValue": "1024",
      "isOverridable": true,
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "General",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "rm_host",
      "displayName": "ResourceManager",
      "description": "ResourceManager",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "masterHost",
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "RESOURCEMANAGER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "resourcemanager_heapsize",
      "displayName": "ResourceManager Java heap size",
      "description": "Max heapsize for ResourceManager",
      "defaultValue": "1024",
      "isOverridable": false,
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "RESOURCEMANAGER",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "nm_hosts",
      "displayName": "NodeManager",
      "description": "List of NodeManager Hosts.",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "slaveHosts",
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "NODEMANAGER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "nodemanager_heapsize",
      "displayName": "NodeManager Java heap size",
      "description": "Max heapsize for NodeManager",
      "defaultValue": "1024",
      "isOverridable": true,
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "NODEMANAGER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "yarn_log_dir_prefix",
      "displayName": "YARN Log Dir Prefix",
      "description": "",
      "defaultValue": "/var/log/hadoop-yarn",
      "displayType": "directory",
      "isOverridable": false,
      "isReconfigurable": false,
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "Advanced yarn-env"
    },
    {
      "id": "puppet var",
      "name": "yarn_pid_dir_prefix",
      "displayName": "YARN PID Dir Prefix",
      "description": "",
      "defaultValue": "/var/run/hadoop-yarn",
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isReconfigurable": false,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "Advanced yarn-env"
    },
    {
      "id": "puppet var",
      "name": "min_user_id",
      "displayName": "Minimum user ID for submitting job",
      "isOverridable": true,
      "displayType": "int",
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "Advanced yarn-env"
    },
    {
      "id": "puppet var",
      "name": "ats_host",
      "displayName": "App Timeline Server",
      "description": "Application Timeline Server Host",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "masterHost",
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "APP_TIMELINE_SERVER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "apptimelineserver_heapsize",
      "displayName": "AppTimelineServer Java heap size",
      "description": "AppTimelineServer Java heap size",
      "defaultValue": "1024",
      "isOverridable": false,
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "YARN",
      "filename": "yarn-env.xml",
      "category": "APP_TIMELINE_SERVER",
      "index": 1
    },
  /**********************************************HBASE***************************************/
    {
      "id": "puppet var",
      "name": "hbasemaster_host",
      "displayName": "HBase Master hosts",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run HBase Master",
      "displayType": "masterHosts",
      "isOverridable": false,
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "HBASE_MASTER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "hbase_master_heapsize",
      "displayName": "HBase Master Maximum Java heap size",
      "description": "Maximum Java heap size for HBase master (Java option -Xmx)",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": true,
      "isVisible": true,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "HBASE_MASTER",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "regionserver_hosts",
      "displayName": "RegionServer hosts",
      "value": "",
      "defaultValue": "",
      "description": "The hosts that have been assigned to run RegionServer",
      "displayType": "slaveHosts",
      "isOverridable": false,
      "isRequiredByAgent": false,
      "isVisible": true,
      "isRequired": false,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "HBASE_REGIONSERVER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "hbase_regionserver_heapsize",
      "displayName": "RegionServers maximum Java heap size",
      "description": "Maximum Java heap size for RegionServers (Java option -Xmx)",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "HBASE_REGIONSERVER",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "hbase_regionserver_xmn_max",
      "displayName": "RegionServers maximum value for -Xmn",
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "HBASE",
      "category": "HBASE_REGIONSERVER",
      "index": 6
    },
    {
      "id": "puppet var",
      "name": "hbase_regionserver_xmn_ratio",
      "displayName": "RegionServers -Xmn in -Xmx ratio",
      "displayType": "float",
      "isVisible": true,
      "serviceName": "HBASE",
      "category": "HBASE_REGIONSERVER",
      "index": 7
    },
    {
      "id": "puppet var",
      "name": "hbase_log_dir",
      "displayName": "HBase Log Dir",
      "description": "Directory for HBase logs",
      "defaultValue": "/var/log/hbase",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "Advanced hbase-env"
    },
    {
      "id": "puppet var",
      "name": "hbase_pid_dir",
      "displayName": "HBase PID Dir",
      "description": "Directory in which the pid files for HBase processes will be created",
      "defaultValue": "/var/run/hbase",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HBASE",
      "filename": "hbase-env.xml",
      "category": "Advanced hbase-env"
    },
     //***************************************** GLUSTERFS stack********************************************
    {
      "id": "site property",
      "name": "fs.glusterfs.impl",
      "displayName": "GlusterFS fs impl",
      "displayType": "string",
      "filename": "core-site.xml",
      "serviceName": "GLUSTERFS",
      "category": "General"
    },
    {
      "id": "site property",
      "name": "fs.AbstractFileSystem.glusterfs.impl",
      "displayName": "GlusterFS Abstract File System Implementation",
      "displayType": "string",
      "filename": "core-site.xml",
      "serviceName": "GLUSTERFS",
      "category": "General"
    },
  /**********************************************GLUSTERFS***************************************/
    {
      "id": "puppet var",
      "name": "fs_glusterfs_default_name",
      "displayName": "GlusterFS default fs name 1.x Hadoop",
      "description": "GlusterFS default filesystem name (glusterfs://{MasterFQDN}:9000)",
      "defaultValue": "glusterfs:///localhost:8020",
      "displayType": "string",
      "isVisible": true,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "glusterfs_defaultFS_name",
      "displayName": "GlusterFS default fs name 2.x Hadoop",
      "description": "GlusterFS default filesystem name (glusterfs:///)",
      "defaultValue": "glusterfs:///localhost:8020",
      "displayType": "string",
      "isVisible": true,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "hadoop_heapsize",
      "displayName": "Hadoop maximum Java heap size",
      "description": "Maximum Java heap size for daemons such as Balancer (Java option -Xmx)",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isVisible": true,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "hdfs_log_dir_prefix",
      "displayName": "Hadoop Log Dir Prefix",
      "description": "The parent directory for Hadoop log files.  The HDFS log directory will be ${hadoop_log_dir_prefix} / ${hdfs_user} and the MapReduce log directory will be ${hadoop_log_dir_prefix} / ${mapred_user}.",
      "defaultValue": "/var/log/hadoop",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "hadoop_pid_dir_prefix",
      "displayName": "Hadoop PID Dir Prefix",
      "description": "The parent directory in which the PID files for Hadoop processes will be created.  The HDFS PID directory will be ${hadoop_pid_dir_prefix} / ${hdfs_user} and the MapReduce PID directory will be ${hadoop_pid_dir_prefix} / ${mapred_user}.",
      "defaultValue": "/var/run/hadoop",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "namenode_heapsize",
      "displayName": "Name Node Heap Size",
      "description": "Name Node Heap Size, default jvm memory setting",
      "defaultValue": "1024",
      "isReconfigurable": false,
      "displayType": "int",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "namenode_opt_newsize",
      "displayName": "NameNode new generation size",
      "description": "Default size of Java new generation for NameNode (Java option -XX:NewSize).  This also applies to the Secondary NameNode.",
      "defaultValue": "200",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "namenode_opt_maxnewsize",
      "displayName": "NameNode maximum new generation size",
      "description": "Maximum size of Java new generation for NameNode (Java option -XX:MaxnewSize).",
      "defaultValue": "200",
      "displayType": "int",
      "unit": "MB",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "dtnode_heapsize",
      "displayName": "DataNode maximum Java heap size",
      "description": "Maximum Java heap size for DataNode (Java option -Xmx)",
      "defaultValue": "1024",
      "displayType": "int",
      "unit": "MB",
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "glusterfs_user",
      "displayName": "glusterfs user",
      "description": "glusterfs user",
      "defaultValue": "root",
      "displayType": "string",
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "namenode_host",
      "displayName": "NameNode Host",
      "description": "NameNode Host.",
      "defaultValue": "",
      "displayType": "string",
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
    {
      "id": "puppet var",
      "name": "snamenode_host",
      "displayName": "Secondary NameNode Host",
      "description": "Secondary NameNode Host.",
      "defaultValue": "",
      "displayType": "string",
      "isVisible": false,
      "serviceName": "GLUSTERFS",
      "filename": "hadoop-env.xml",
      "category": "General Hadoop"
    },
  /**********************************************HIVE***************************************/
    {
      "id": "puppet var",
      "name": "hivemetastore_host",
      "displayName": "Hive Metastore hosts",
      "value": "",
      "defaultValue": "",
      "description": "The hosts that have been assigned to run Hive Metastore",
      "displayType": "masterHosts",
      "isOverridable": false,
      "isRequiredByAgent": false,
      "isVisible": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 0
    },
    {
      "name": "hive_master_hosts",
      "value": "",
      "defaultValue": "",
      "isOverridable": false,
      "isRequiredByAgent": false,
      "isVisible": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml"
    },
    // for existing MySQL
    {
      "id": "puppet var",
      "name": "hive_existing_mysql_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "MySQL",
      "description": "Using an existing MySQL database for Hive Metastore",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": false,
      "isReconfigurable": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 1
    },
    // for existing PostgreSQL
    {
      "id": "puppet var",
      "name": "hive_existing_postgresql_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "PostgreSQL",
      "description": "Using an existing PostgreSQL database for Hive Metastore",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": false,
      "isReconfigurable": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 1
    },
    // for existing Oracle
    {
      "id": "puppet var",
      "name": "hive_existing_oracle_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "Oracle",
      "description": "Using an existing Oracle database for Hive Metastore",
      "displayType": "masterHost",
      "isVisible": false,
      "isOverridable": false,
      "isReconfigurable": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 1
    },
    // for new PostgreSQL
    {
      "id": "puppet var",
      "name": "hive_ambari_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "PostgreSQL",
      "description": "PostgreSQL will be installed by Ambari",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "hive_database",
      "displayName": "Hive Database",
      "value": "",
      "defaultValue": "New PostgreSQL Database",
      "options": [
        {
          displayName: 'New PostgreSQL Database',
          foreignKeys: ['hive_ambari_database', 'hive_ambari_host']
        },
        {
          displayName: 'Existing MySQL Database',
          foreignKeys: ['hive_existing_mysql_database', 'hive_existing_mysql_host']
        },
        {
          displayName: 'Existing PostgreSQL Database',
          foreignKeys: ['hive_existing_postgresql_database', 'hive_existing_postgresql_host']
        },
        {
          displayName: 'Existing Oracle Database',
          foreignKeys: ['hive_existing_oracle_database', 'hive_existing_oracle_host'],
          hidden: !App.supports.hiveOozieExtraDatabases
        }
      ],
      "description": "PostgreSQL will be installed by Ambari",
      "displayType": "radio button",
      "isReconfigurable": true,
      "radioName": "hive-database",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 2
    },
    {
      "id": "puppet var",
      "name": "hive_hostname",
      "displayName": "Database Host",
      "description": "Specify the host on which the database is hosted",
      "defaultValue": "",
      "isReconfigurable": true,
      "displayType": "host",
      "isOverridable": false,
      "isVisible": false,
      "isObserved": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "hive_existing_mysql_host",
      "displayName": "Database Host",
      "description": "Specify the host on which the existing database is hosted",
      "defaultValue": "",
      "displayType": "host",
      "isOverridable": false,
      "isVisible": false,
      "isObserved": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "hive_existing_postgresql_host",
      "displayName": "Database Host",
      "description": "Specify the host on which the existing database is hosted",
      "defaultValue": "",
      "displayType": "host",
      "isOverridable": false,
      "isVisible": false,
      "isObserved": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "hive_existing_oracle_host",
      "displayName": "Database Host",
      "description": "Specify the host on which the existing database is hosted",
      "defaultValue": "",
      "displayType": "host",
      "isOverridable": false,
      "isVisible": false,
      "isObserved": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "hive_ambari_host",
      "value": "",
      "defaultValue": "",
      "displayName": "Database Host",
      "description": "Host on which the database will be created by Ambari",
      "isReconfigurable": false,
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "HIVE_METASTORE",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "hive_metastore_port",
      "displayName": "Hive metastore port",
      "description": "",
      "defaultValue": "9083",
      "isReconfigurable": false,
      "displayType": "int",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "Advanced hive-env"
    },
    {
      "id": "puppet var",
      "name": "hive_lib",
      "displayName": "Hive library",
      "description": "",
      "defaultValue": "/usr/lib/hive/lib/",
      "isReconfigurable": false,
      "displayType": "directory",
      "isVisible": false,
      "isRequiredByAgent": false, // Make this to true when we expose the property on ui by making "isVisible": true
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "Advanced hive-env"
    },
    {
      "id": "puppet var",
      "name": "hive_dbroot",
      "displayName": "Hive db directory",
      "description": "",
      "defaultValue": "/usr/lib/hive/lib",
      "isReconfigurable": false,
      "displayType": "directory",
      "isVisible": false,
      "isRequiredByAgent": false, // Make this to true when we expose the property on ui by making "isVisible": true
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "Advanced hive-env"
    },
    {
      "id": "puppet var",
      "name": "hive_log_dir",
      "displayName": "Hive Log Dir",
      "description": "Directory for Hive log files",
      "defaultValue": "/var/log/hive",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "Advanced hive-env"
    },
    {
      "id": "puppet var",
      "name": "hive_pid_dir",
      "displayName": "Hive PID Dir",
      "description": "Directory in which the PID files for Hive processes will be created",
      "defaultValue": "/var/run/hive",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HIVE",
      "filename": "hive-env.xml",
      "category": "Advanced hive-env"
    },
  /**********************************************HIVE***************************************/
    {
      "id": "puppet var",
      "name": "webhcatserver_host",
      "displayName": "WebHCat Server host",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run WebHCat Server",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "HIVE",
      "filename": "webhcat-env.xml",
      "category": "WEBHCAT_SERVER"
    },
    {
      "id": "puppet var",
      "name": "hcat_log_dir",
      "displayName": "WebHCat Log Dir",
      "description": "Directory for WebHCat log files",
      "defaultValue": "/var/log/webhcat",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HIVE",
      "filename": "webhcat-env.xml",
      "category": "Advanced webhcat-env"
    },
    {
      "id": "puppet var",
      "name": "hcat_pid_dir",
      "displayName": "WebHCat PID Dir",
      "description": "Directory in which the PID files for WebHCat processes will be created",
      "defaultValue": "/var/run/webhcat",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "HIVE",
      "filename": "webhcat-env.xml",
      "category": "Advanced webhcat-env"
    },
  /**********************************************OOZIE***************************************/
    {
      "id": "puppet var",
      "name": "oozieserver_host",
      "displayName": "Oozie Server host",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run Oozie Server",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 0
    },
    // for existing Oracle
    {
      "id": "puppet var",
      "name": "oozie_existing_oracle_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "Oracle",
      "description": "Using an existing Oracle database for Oozie Metastore",
      "displayType": "masterHost",
      "isVisible": false,
      "isObserved": true,
      "isReconfigurable": false,
      "isOverridable": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 1
    },
    // for existing PostgreSQL
    {
      "id": "puppet var",
      "name": "oozie_existing_postgresql_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "PostgreSQL",
      "description": "Using an existing PostgreSQL database for Oozie Metastore",
      "displayType": "masterHost",
      "isOverridable": false,
      "isVisible": false,
      "isReconfigurable": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 1
    },
    // for new MySQL
    {
      "id": "puppet var",
      "name": "oozie_ambari_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "MySQL",
      "description": "MySQL will be installed by Ambari",
      "displayType": "masterHost",
      "isVisible": false,
      "isOverridable": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 1
    },
    // for current derby
    {
      "id": "puppet var",
      "name": "oozie_derby_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "Derby",
      "description": "Using current Derby database for Oozie Metastore",
      "displayType": "masterHost",
      "isVisible": false,
      "isReconfigurable": false,
      "isOverridable": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 1
    },
    // for existing MySQL oozie
    {
      "id": "puppet var",
      "name": "oozie_existing_mysql_database",
      "displayName": "Database Type",
      "value": "",
      "defaultValue": "MySQL",
      "description": "Using an existing MySQL database for Oozie Metastore",
      "displayType": "masterHost",
      "isVisible": false,
      "isReconfigurable": false,
      "isOverridable": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "oozie_database",
      "displayName": "Oozie Database",
      "value": "",
      "defaultValue": "New Derby Database",
      "options": [
        {
          displayName: 'New Derby Database',
          foreignKeys: ['oozie_derby_database']
        },
        /*{
         displayName: 'New MySQL Database',
         foreignKeys: ['oozie_ambari_database', 'oozie_ambari_host'],
         hidden: !App.supports.hiveOozieExtraDatabases
         },*/
        {
          displayName: 'Existing MySQL Database',
          foreignKeys: ['oozie_existing_mysql_database', 'oozie_existing_mysql_host'],
          hidden: !App.supports.hiveOozieExtraDatabases
        },
        {
          displayName: 'Existing PostgreSQL Database',
          foreignKeys: ['oozie_existing_postgresql_database', 'oozie_existing_postgresql_host'],
          hidden: !App.supports.hiveOozieExtraDatabases
        },
        {
          displayName: 'Existing Oracle Database',
          foreignKeys: ['oozie_existing_oracle_database', 'oozie_existing_oracle_host'],
          hidden: !App.supports.hiveOozieExtraDatabases
        }
      ],
      "description": "Current Derby Database will be installed by Ambari",
      "displayType": "radio button",
      "isReconfigurable": true,
      "isOverridable": false,
      "radioName": "oozie-database",
      "isVisible": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 2
    },
    {
      "id": "puppet var",
      "name": "oozie_data_dir",
      "displayName": "Oozie Data Dir",
      "description": "Data directory in which the Oozie DB exists",
      "defaultValue": "",
      "defaultDirectory": "/hadoop/oozie/data",
      "isReconfigurable": true,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isRequired": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 9
    },
    {
      "id": "puppet var",
      "name": "oozie_hostname",
      "defaultValue": "",
      "displayName": "Database Host",
      "description": "The host where the Oozie database is located",
      "isReconfigurable": true,
      "isOverridable": false,
      "displayType": "host",
      "isVisible": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "oozie_existing_mysql_host",
      "displayName": "Database Host",
      "description": "Specify the host on which the existing database is hosted",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "host",
      "isVisible": false,
      "isObserved": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "oozie_existing_postgresql_host",
      "displayName": "Database Host",
      "description": "Specify the host on which the existing database is hosted",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "host",
      "isVisible": false,
      "isObserved": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "oozie_existing_oracle_host",
      "displayName": "Database Host",
      "description": "Specify the host on which the existing database is hosted",
      "defaultValue": "",
      "isOverridable": false,
      "displayType": "host",
      "isVisible": false,
      "isObserved": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "oozie_ambari_host",
      "value": "",
      "defaultValue": "",
      "displayName": "Database Host",
      "description": "Host on which the database will be created by Ambari",
      "isReconfigurable": false,
      "isOverridable": false,
      "displayType": "masterHost",
      "isVisible": false,
      "isRequiredByAgent": false,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "OOZIE_SERVER"
    },
    {
      "id": "puppet var",
      "name": "oozie_log_dir",
      "displayName": "Oozie Log Dir",
      "description": "Directory for oozie logs",
      "defaultValue": "/var/log/oozie",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "Advanced oozie-env"
    },
    {
      "id": "puppet var",
      "name": "oozie_pid_dir",
      "displayName": "Oozie PID Dir",
      "description": "Directory in which the pid files for oozie processes will be created",
      "defaultValue": "/var/run/oozie",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "Advanced oozie-env"
    },
    {
      "id": "puppet var",
      "name": "oozie_admin_port",
      "displayName": "Oozie Server Admin Port",
      "isReconfigurable": true,
      "displayType": "int",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "OOZIE",
      "filename": "oozie-env.xml",
      "category": "Advanced oozie-env"
    },
  /**********************************************ZOOKEEPER***************************************/
    {
      "id": "puppet var",
      "name": "zookeeperserver_hosts",
      "displayName": "ZooKeeper Server hosts",
      "value": "",
      "defaultValue": "",
      "description": "The host that has been assigned to run ZooKeeper Server",
      "displayType": "masterHosts",
      "isVisible": true,
      "isRequiredByAgent": false,
      "isOverridable": false,
      "isRequired": false,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZOOKEEPER_SERVER",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "zk_data_dir",
      "displayName": "ZooKeeper directory",
      "description": "Data directory for ZooKeeper",
      "defaultValue": "",
      "defaultDirectory": "/hadoop/zookeeper",
      "isReconfigurable": true,
      "displayType": "directory",
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZOOKEEPER_SERVER",
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "tickTime",
      "displayName": "Length of single Tick",
      "description": "The length of a single tick in milliseconds, which is the basic time unit used by ZooKeeper",
      "defaultValue": "2000",
      "displayType": "int",
      "unit": "ms",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZOOKEEPER_SERVER",
      "index": 2
    },
    {
      "id": "puppet var",
      "name": "initLimit",
      "displayName": "Ticks to allow for sync at Init",
      "description": "Amount of time, in ticks to allow followers to connect and sync to a leader",
      "defaultValue": "10",
      "displayType": "int",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZOOKEEPER_SERVER",
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "syncLimit",
      "displayName": "Ticks to allow for sync at Runtime",
      "description": "Amount of time, in ticks to allow followers to connect",
      "defaultValue": "5",
      "displayType": "int",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZOOKEEPER_SERVER",
      "index": 4
    },
    {
      "id": "puppet var",
      "name": "clientPort",
      "displayName": "Port for running ZK Server",
      "description": "Port for running ZooKeeper server",
      "defaultValue": "2181",
      "displayType": "int",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "ZOOKEEPER_SERVER",
      "index": 5
    },
    {
      "id": "puppet var",
      "name": "zk_log_dir",
      "displayName": "ZooKeeper Log Dir",
      "description": "Directory for ZooKeeper log files",
      "defaultValue": "/var/log/zookeeper",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "Advanced zookeeper-env",
      "index": 0
    },
    {
      "id": "puppet var",
      "name": "zk_pid_dir",
      "displayName": "ZooKeeper PID Dir",
      "description": "Directory in which the pid files for zookeeper processes will be created",
      "defaultValue": "/var/run/zookeeper",
      "isReconfigurable": false,
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "ZOOKEEPER",
      "filename": "zookeeper-env.xml",
      "category": "Advanced zookeeper-env",
      "index": 1
    },
  /**********************************************GANGLIA***************************************/
    {
      "id": "puppet var",
      "name": "ganglia_conf_dir",
      "displayName": "Ganglia conf directory",
      "description": "",
      "defaultValue": "/etc/ganglia/hdp",
      "isReconfigurable": false,
      "displayType": "directory",
      "isVisible": false,
      "isRequiredByAgent": false,
      "serviceName": "GANGLIA",
      "filename": "ganglia-env.xml",
      "category": "Advanced ganglia-env"
    },
  /**********************************************FALCON***************************************/
    {
      "id": "puppet var",
      "name": "falconserver_host",
      "displayName": "Falcon Server",
      "description": "The host that has been assigned to run Falcon Server",
      "defaultValue": "falcon",
      "displayType": "masterHost",
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "FALCON_SERVER"
    },
    {
      "id": "puppet var",
      "name": "falcon_port",
      "displayName": "Falcon server port",
      "description": "Port the Falcon Server listens on",
      "defaultValue": "15000",
      "isReconfigurable": true,
      "displayType": "int",
      "isVisible": true,
      "isRequiredByAgent": true,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "FALCON_SERVER"
    },
    {
      "id": "puppet var",
      "name": "falcon_local_dir",
      "displayName": "Falcon data directory",
      "description": "Directory where Falcon data, such as activemq data, is stored",
      "defaultValue": "/hadoop/falcon",
      "isReconfigurable": true,
      "displayType": "directory",
      "isVisible": true,
      "isRequiredByAgent": true,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "FALCON_SERVER"
    },
    {
      "id": "puppet var",
      "name": "falcon_store_uri",
      "displayName": "Falcon store URI",
      "description": "Directory where entity definitions are stored",
      "defaultValue": "file:///hadoop/falcon/store",
      "isReconfigurable": true,
      "displayType": "string",
      "isVisible": true,
      "isRequiredByAgent": true,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "FALCON_SERVER"
    },
    {
      "id": "puppet var",
      "name": "falcon_log_dir",
      "displayName": "Falcon Log Dir",
      "description": "Directory for Falcon logs",
      "defaultValue": "/var/log/falcon",
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": true,
      "isReconfigurable": false,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "Advanced falcon-env"
    },
    {
      "id": "puppet var",
      "name": "falcon_pid_dir",
      "displayName": "Falcon PID Dir",
      "description": "Directory in which the pid files for Falcon processes will be created",
      "defaultValue": "/var/run/falcon",
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": true,
      "isReconfigurable": false,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "Advanced falcon-env"
    },
    {
      "id": "puppet var",
      "name": "falcon.embeddedmq",
      "displayName": "falcon.embeddedmq",
      "description": "Whether embeddedmq is enabled or not.",
      "defaultValue": "true",
      "displayType": "string",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": true,
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "Advanced falcon-env"
    },
    {
      "id": "puppet var",
      "name": "falcon.embeddedmq.data",
      "displayName": "falcon.embeddedmq.data",
      "description": "Directory in which embeddedmq data is stored.",
      "defaultValue": "/hadoop/falcon/embeddedmq/data",
      "displayType": "directory",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": true,
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "Advanced falcon-env"
    },
    {
      "id": "puppet var",
      "name": "falcon.emeddedmq.port",
      "displayName": "falcon.emeddedmq.port",
      "description": "Port that embeddedmq will listen on.",
      "defaultValue": "61616",
      "displayType": "string",
      "isOverridable": false,
      "isVisible": true,
      "isRequiredByAgent": true,
      "isReconfigurable": true,
      "serviceName": "FALCON",
      "filename": "falcon-env.xml",
      "category": "Advanced falcon-env"
    },
  /**********************************************STORM***************************************/
    {
      "id": "puppet var",
      "name": "storm_log_dir",
      "displayName": "storm_log_dir",
      "description": "Storm log directory",
      "defaultValue": "/var/log/storm",
      "displayType": "directory",
      "isReconfigurable": true,
      "isVisible": true,
      "isRequiredByAgent": true,
      "serviceName": "STORM",
      "filename": "storm-env.xml",
      "category": "Advanced storm-env"
    },
    {
      "id": "puppet var",
      "name": "storm_pid_dir",
      "displayName": "storm_pid_dir",
      "description": "Storm PID directory",
      "defaultValue": "/var/run/storm",
      "displayType": "directory",
      "isReconfigurable": true,
      "isVisible": true,
      "isRequiredByAgent": true,
      "serviceName": "STORM",
      "filename": "storm-env.xml",
      "category": "Advanced storm-env"
    },
    {
      "id": "puppet var",
      "name": "stormuiserver_host",
      "displayName": "Storm UI Server host",
      "description": "The host that has been assigned to run Storm UI Server",
      "defaultValue": "",
      "displayType": "masterHost",
      "isReconfigurable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "STORM",
      "filename": "storm-env.xml",
      "category": "STORM_UI_SERVER"
    },
    {
      "id": "puppet var",
      "name": "drpcserver_host",
      "displayName": "DRPC Server host",
      "description": "The host that has been assigned to run DRPC Server",
      "defaultValue": "",
      "displayType": "masterHost",
      "isReconfigurable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "STORM",
      "filename": "storm-env.xml",
      "category": "DRPC_SERVER"
    },
    {
      "id": "puppet var",
      "name": "supervisor_hosts",
      "displayName": "Supervisor hosts",
      "description": "The host that has been assigned to run Supervisor",
      "defaultValue": "",
      "displayType": "slaveHosts",
      "isReconfigurable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "STORM",
      "filename": "storm-env.xml",
      "isOverridable": false,
      "category": "SUPERVISOR"
    },
    {
      "id": "puppet var",
      "name": "storm_rest_api_host",
      "displayName": "Storm REST API host",
      "description": "The host that has been assigned to run Storm REST API Server",
      "defaultValue": "",
      "displayType": "masterHost",
      "isReconfigurable": false,
      "isVisible": true,
      "isRequiredByAgent": false,
      "serviceName": "STORM",
      "filename": "storm-env.xml",
      "isOverridable": false,
      "category": "STORM_REST_API"
    },
  /**********************************************MISC***************************************/
    {
      "id": "puppet var",
      "name": "hbase_conf_dir",
      "displayName": "HBase conf dir",
      "description": "",
      "defaultValue": "/etc/hbase",
      "isRequired": false,
      "displayType": "directory",
      "isVisible": false,
      "isRequiredByAgent": false,
      "serviceName": "MISC",
      "filename": "hbase-env.xml",
      "category": "General",
      "belongsToService": []
    },
    {
      "id": "puppet var",
      "name": "proxyuser_group",
      "displayName": "Proxy group for Hive, WebHCat, Oozie and Falcon",
      "description": "",
      "defaultValue": "users",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "filename": "core-site.xml",
      "serviceName": "MISC",
      "category": "Users and Groups",
      "belongsToService": ["HIVE", "OOZIE", "FALCON"],
      "index": 18
    },
    {
      "id": "puppet var",
      "name": "ganglia_runtime_dir",
      "displayName": "Ganglia runtime directory",
      "description": "",
      "defaultValue": "/var/run/ganglia/hdp",
      "isReconfigurable": false,
      "displayType": "directory",
      "isVisible": false,
      "serviceName": "MISC",
      "filename": "ganglia-env.xml",
      "category": "General",
      "belongsToService": []
    },
    {
      "id": "puppet var",
      "name": "hdfs_user",
      "displayName": "HDFS User",
      "description": "User to run HDFS as",
      "defaultValue": "hdfs",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "hadoop-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["HDFS"],
      "index": 1
    },
    {
      "id": "puppet var",
      "name": "mapred_user",
      "displayName": "MapReduce User",
      "description": "User to run MapReduce as",
      "defaultValue": "mapred",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "mapred-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["MAPREDUCE2"],
      "index": 2
    },
    {
      "id": "puppet var",
      "name": "yarn_user",
      "displayName": "YARN User",
      "description": "User to run YARN as",
      "defaultValue": "yarn",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "yarn-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["YARN"],
      "index": 3
    },
    {
      "id": "puppet var",
      "name": "hbase_user",
      "displayName": "HBase User",
      "description": "User to run HBase as",
      "defaultValue": "hbase",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "hbase-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["HBASE"],
      "index": 4
    },
    {
      "id": "puppet var",
      "name": "hive_user",
      "displayName": "Hive User",
      "description": "User to run Hive as",
      "defaultValue": "hive",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "hive-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["HIVE"],
      "index": 5
    },
    {
      "id": "puppet var",
      "name": "hcat_user",
      "displayName": "HCat User",
      "description": "User to run HCatalog as",
      "defaultValue": "hcat",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "hive-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["HIVE"],
      "index": 6
    },
    {
      "id": "puppet var",
      "name": "webhcat_user",
      "displayName": "WebHCat User",
      "description": "User to run WebHCat as",
      "defaultValue": "hcat",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "webhcat-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["HIVE"],
      "index": 7
    },
    {
      "id": "puppet var",
      "name": "oozie_user",
      "displayName": "Oozie User",
      "description": "User to run Oozie as",
      "defaultValue": "oozie",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "oozie-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["OOZIE"],
      "index": 8
    },
    {
      "id": "puppet var",
      "name": "falcon_user",
      "displayName": "Falcon User",
      "description": "User to run Falcon as",
      "defaultValue": "falcon",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "falcon-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["FALCON"],
      "index": 10
    },
    {
      "id": "puppet var",
      "name": "storm_user",
      "displayName": "Storm User",
      "description": "User to run Storm as",
      "defaultValue": "storm",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "storm-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["STORM"],
      "index": 9
    },
    {
      "id": "puppet var",
      "name": "zk_user",
      "displayName": "ZooKeeper User",
      "description": "User to run ZooKeeper as",
      "defaultValue": "zookeeper",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "zookeeper-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["ZOOKEEPER"],
      "index": 11
    },
    {
      "id": "puppet var",
      "name": "flume_user",
      "displayName": "Flume User",
      "description": "User to run Flume as",
      "defaultValue": "flume",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "flume-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["FLUME"],
      "index": 12
    },
    {
      "id": "puppet var",
      "name": "gmetad_user",
      "displayName": "Ganglia User",
      "description": "The user used to run Ganglia",
      "defaultValue": "nobody",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "ganglia-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["GANGLIA"],
      "index": 13
    },
    {
      "id": "puppet var",
      "name": "gmond_user",
      "displayName": "Gmond User",
      "description": "The user used to run gmond for Ganglia",
      "defaultValue": "nobody",
      "isReconfigurable": false,
      "displayType": "advanced",
      "isOverridable": false,
      "isVisible": false,
      "serviceName": "MISC",
      "filename": "ganglia-env.xml",
      "category": "Users and Groups",
      "belongsToService": []
    },
    {
      "id": "puppet var",
      "name": "tez_user",
      "displayName": "Tez User",
      "description": "User to run Tez as",
      "defaultValue": "tez",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "tez-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["TEZ"],
      "index": 15
    },
    {
      "id": "puppet var",
      "name": "sqoop_user",
      "displayName": "Sqoop User",
      "isReconfigurable": false,
      "displayType": "user",
      "isOverridable": false,
      "isVisible": true,
      "serviceName": "MISC",
      "filename": "sqoop-env.xml",
      "category": "Users and Groups",
      "belongsToService": ["SQOOP"],
      "index": 17
    },
    {
      "id": "puppet var",
      "name": "ignore_groupsusers_create",
      "displayName": "Skip group modifications during install",
      "displayType": "checkbox",
      "isReconfigurable": true,
      "isOverridable": false,
      "isVisible": true,
      "filename": "cluster-env.xml",
      "category": "Users and Groups"
    }
  ]
};


});

require.register("data/host/categories", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');
module.exports = [
  {
    value: Em.I18n.t('common.all'),
    isHealthStatus: true,
    healthStatus: '',
    healthClass: '',
    isActive: true,
    isVisible: false
  },
  {
    value: Em.I18n.t('hosts.host.healthStatusCategory.green'),
    isHealthStatus: true,
    class: App.healthIconClassGreen,
    healthStatus: 'HEALTHY',
    healthClass: 'health-status-LIVE'
  },
  {
    value: Em.I18n.t('hosts.host.healthStatusCategory.red'),
    isHealthStatus: true,
    class: App.healthIconClassRed,
    healthStatus: 'UNHEALTHY',
    healthClass: 'health-status-DEAD-RED'
  },
  {
    value: Em.I18n.t('hosts.host.healthStatusCategory.orange'),
    isHealthStatus: true,
    class: App.healthIconClassOrange,
    healthStatus: 'ALERT',
    healthClass: 'health-status-DEAD-ORANGE'
  },
  {
    value: Em.I18n.t('hosts.host.healthStatusCategory.yellow'),
    isHealthStatus: true,
    class: App.healthIconClassYellow,
    healthStatus: 'UNKNOWN',
    healthClass: 'health-status-DEAD-YELLOW'
  },
  {
    value: Em.I18n.t('hosts.host.alerts.label'),
    hostProperty: 'criticalWarningAlertsCount',
    class: 'icon-exclamation-sign',
    isHealthStatus: false,
    healthClass: 'health-status-WITH-ALERTS',
    healthStatus: 'health-status-WITH-ALERTS',
    column: 7,
    type: 'custom',
    filterValue: ['>0', '>0']
  },
  {
    value: Em.I18n.t('common.restart'),
    hostProperty: 'componentsWithStaleConfigsCount',
    class: 'icon-refresh',
    isHealthStatus: false,
    healthClass: 'health-status-RESTART',
    healthStatus: 'health-status-RESTART',
    column: 8,
    type: 'string',
    filterValue: 'true'
  },
  {
    value: Em.I18n.t('common.passive_state'),
    hostProperty: 'componentsInPassiveStateCount',
    class: 'passive-state icon-medkit',
    isHealthStatus: false,
    healthClass: 'health-status-PASSIVE_STATE',
    healthStatus: 'health-status-PASSIVE_STATE',
    column: 9,
    type: 'string',
    filterValue: ['ON','IMPLIED_FROM_HOST','IMPLIED_FROM_SERVICE','IMPLIED_FROM_SERVICE_AND_HOST']
  }
];

});

require.register("data/service_configs", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
require('models/service_config');

/**
 * This
 * @type {Array} Array of non-service tabs that will appears on Customizes services page
 */
module.exports = [
  Em.Object.create({
    serviceName: 'MISC',
    displayName: 'Misc',
    configCategories: [
      App.ServiceConfigCategory.create({ name: 'Users and Groups', displayName : 'Services Accounts'}),
      App.ServiceConfigCategory.create({ name: 'Notifications', displayName : 'Notifications', isCustomView: true, customView: App.NotificationsConfigsView, siteFileName: 'alert_notification'})
    ],
    configTypes: {
      "cluster-env": {supports: {final: false}}
    },
    configs: []
  })
];

});

require.register("data/service_graph_config", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
This determines the graphs to display on the service page under each service.

This is based on the name of the object associated with it.

The name of the object is of the format: 'App.ChartServiceMetrics<name>' where <name>
is one of the items below.
**/
App.service_graph_config = {
	'hdfs': [
		'HDFS_SpaceUtilization',
		'HDFS_FileOperations',
		'HDFS_BlockStatus',
		'HDFS_IO',
		'HDFS_RPC',
		'HDFS_GC',
		'HDFS_JVMHeap',
		'HDFS_JVMThreads'
	],

	'yarn': [
		'YARN_AllocatedMemory',
		'YARN_QMR',
		'YARN_AllocatedContainer',
		'YARN_NMS',
		'YARN_ApplicationCurrentStates',
		'YARN_ApplicationFinishedStates',
		'YARN_RPC',
		'YARN_GC',
		'YARN_JVMThreads',
		'YARN_JVMHeap'
	],

	'hbase': [
		'HBASE_ClusterRequests',
		'HBASE_RegionServerReadWriteRequests',
		'HBASE_RegionServerRegions',
		'HBASE_RegionServerQueueSize',
		'HBASE_HlogSplitTime',
		'HBASE_HlogSplitSize'
	],

	'flume': [
		'Flume_ChannelSizeMMA',
		'Flume_ChannelSizeSum',
		'Flume_IncommingMMA',
		'Flume_IncommingSum',
		'Flume_OutgoingMMA',
		'Flume_OutgoingSum'
	],

	'storm': [
		'STORM_SlotsNumber',
		'STORM_Executors',
		'STORM_Topologies',
		'STORM_Tasks'
	],

  'kafka': [
    'Kafka_BrokerTopicMetrics',
    'Kafka_Controller',
    'Kafka_ControllerStatus',
    'Kafka_ReplicaManager',
    'Kafka_LogFlush',
    'Kafka_ReplicaFetcher'
  ]

};
});

require.register("data/statusCodes", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


module.exports = {
	200: function () {
		console.log("Status code 200: Success.");
  },
  202: function () {
    console.log("Status code 202: Success for creation.");
  },
	400: function () {
		console.log("Error code 400: Bad Request.");
	},
	401: function () {
		console.log("Error code 401: Unauthorized.");
	},
	402: function () {
		console.log("Error code 402: Payment Required.");
	},
	403: function () {
		console.log("Error code 403: Forbidden.");
    App.router.logOff();
	},
	404: function () {
		console.log("Error code 404: URI not found.");
	},
	500: function () {
		console.log("Error code 500: Internal Error on server side.");
	},
	501: function () {
		console.log("Error code 501: Not implemented yet.");
	},
	502: function () {
		console.log("Error code 502: Services temporarily overloaded.");
	},
	503: function () {
		console.log("Error code 503: Gateway timeout.");
	}
};

});

require.register("ember", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

module.exports=Ember;
});

require.register("initialize", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


window.App = require('app');

require('config');

require('messages');
require('utils/base64');
require('utils/db');
require('utils/helper');
require('utils/config');
require('mixins');
require('models');
require('controllers');
require('templates');
require('views');
require('router');

require('utils/ajax/ajax');
require('utils/ajax/ajax_queue');
require('utils/updater');
require('utils/action_sequence');

require('mappers');

require('utils/http_client');
require('utils/host_progress_popup');

App.initialize();

console.log('after initialize');
console.log('TRACE: app.js-> localStorage:Ambari.authenticated=' + localStorage.getItem('Ambari' + 'authenticated'));
console.log('TRACE: app.js-> localStorage:currentStep=' + localStorage.getItem(App.get('router').getLoginName() + 'Installer' + 'currentStep'));
console.log('TRACE: app.js-> router.authenticated=' + App.get('router.loggedIn'));

});

require.register("mappers", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

//load all mappers
require('mappers/server_data_mapper');
require('mappers/stack_service_mapper');
require('mappers/stack_mapper');
require('mappers/stack_version_mapper');
require('mappers/repository_version_mapper');
require('mappers/hosts_mapper');
require('mappers/cluster_mapper');
require('mappers/racks_mapper');
require('mappers/users_mapper');
require('mappers/service_mapper');
require('mappers/service_metrics_mapper');
require('mappers/target_cluster_mapper');
require('mappers/component_config_mapper');
require('mappers/components_state_mapper');
require('mappers/service_config_version_mapper');
require('mappers/alert_definitions_mapper');
require('mappers/alert_definition_summary_mapper');
require('mappers/alert_instances_mapper');
require('mappers/alert_groups_mapper');
require('mappers/alert_notification_mapper');
require('mappers/root_service_mapper');

});

require.register("mappers/alert_definition_summary_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.alertDefinitionSummaryMapper = App.QuickDataMapper.create({

  config: {},

  map: function(data) {
    if (!data.alerts_summary_grouped) return;
    var alertDefinitions = App.AlertDefinition.find();
    data.alerts_summary_grouped.forEach(function(alertDefinitionSummary) {
      var alertDefinition = alertDefinitions.findProperty('id', alertDefinitionSummary.definition_id);
      if (alertDefinition) {
        var summary = {},
          timestamp = 0;
        Em.keys(alertDefinitionSummary.summary).forEach(function(status) {
          summary[status] = {
            count: alertDefinitionSummary.summary[status].count,
            maintenanceCount: alertDefinitionSummary.summary[status].maintenance_count
          };
          if (alertDefinitionSummary.summary[status].latest_text) {
            summary[status].latestText = alertDefinitionSummary.summary[status].latest_text;
          }
          if (alertDefinitionSummary.summary[status].original_timestamp > timestamp) {
            timestamp = alertDefinitionSummary.summary[status].original_timestamp;
          }
        });
        alertDefinition.setProperties({
          summary: summary,
          lastTriggered: parseInt(timestamp)
        });
      }
    });

    // clear summary for disabled alert definitions
    alertDefinitions.forEach(function (definition) {
      if (!definition.get('enabled')) {
        definition.set('summary', {});
      }
    });

    if (App.router.get('mainAlertDefinitionsController')) {
      App.router.set('mainAlertDefinitionsController.mapperTimestamp', (new Date()).getTime());
    }

  }
});

});

require.register("mappers/alert_definitions_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

var stringUtils = require('utils/string_utils');

App.alertDefinitionsMapper = App.QuickDataMapper.create({

  model: App.AlertDefinition,
  reportModel: App.AlertReportDefinition,
  metricsSourceModel: App.AlertMetricsSourceDefinition,
  metricsUriModel: App.AlertMetricsUriDefinition,

  config: {
    id: 'AlertDefinition.id',
    name: 'AlertDefinition.name',
    description: 'AlertDefinition.description',
    label: 'AlertDefinition.label',
    service_id: 'AlertDefinition.service_name',
    service_name: 'AlertDefinition.service_name',
    component_name: 'AlertDefinition.component_name',
    enabled: 'AlertDefinition.enabled',
    scope: 'AlertDefinition.scope',
    interval: 'AlertDefinition.interval',
    type: 'AlertDefinition.source.type',
    reporting_key: 'reporting',
    reporting_type: 'array',
    reporting: {
      item: 'id'
    }
  },

  portConfig: {
    default_port: 'AlertDefinition.source.default_port',
    port_uri: 'AlertDefinition.source.uri'
  },

  aggregateConfig: {
    alert_name: 'AlertDefinition.source.alert_name'
  },

  scriptConfig: {
    location: 'AlertDefinition.source.path'
  },

  uriConfig: {
    id: 'AlertDefinition.source.uri.id',
    http: 'AlertDefinition.source.uri.http',
    https: 'AlertDefinition.source.uri.https',
    https_property: 'AlertDefinition.source.uri.https_property',
    https_property_value: 'AlertDefinition.source.uri.https_property_value'
  },

  map: function (json) {
    if (json && json.items) {

      var self = this,
          alertDefinitions = [],
          alertReportDefinitions = [],
          alertMetricsSourceDefinitions = [],
          alertMetricsUriDefinitions = [],
          alertGroupsMap = App.cache['previousAlertGroupsMap'],
          existingAlertDefinitions = App.AlertDefinition.find(),
          alertDefinitionsToDelete = existingAlertDefinitions.mapProperty('id'),
          rawSourceData = {};

      json.items.forEach(function (item) {
        var convertedReportDefinitions = [];
        var reporting = item.AlertDefinition.source.reporting;
        for (var report in reporting) {
          if (reporting.hasOwnProperty(report)) {
            if (report == "units") {
              convertedReportDefinitions.push({
                id: item.AlertDefinition.id + report,
                type: report,
                text: reporting[report]
              });
            } else {
              convertedReportDefinitions.push({
                id: item.AlertDefinition.id + report,
                type: report,
                text: reporting[report].text,
                value: reporting[report].value
              });
            }
          }
        }

        alertReportDefinitions = alertReportDefinitions.concat(convertedReportDefinitions);
        item.reporting = convertedReportDefinitions;

        rawSourceData[item.AlertDefinition.id] = item.AlertDefinition.source;
        item.AlertDefinition.description = item.AlertDefinition.description || '';

        var alertDefinition = this.parseIt(item, this.get('config'));

        if (alertGroupsMap[alertDefinition.id]) {
          alertDefinition.groups = alertGroupsMap[alertDefinition.id];
        }

        var oldAlertDefinition = existingAlertDefinitions.findProperty('id', alertDefinition.id);
        if (oldAlertDefinition) {
          // new values will be parsed in the another mapper, so for now just use old values
          alertDefinition.summary = oldAlertDefinition.get('summary');
          alertDefinition.last_triggered = oldAlertDefinition.get('lastTriggered');
        }

        alertDefinitionsToDelete = alertDefinitionsToDelete.without(alertDefinition.id);

        // map properties dependent on Alert Definition type
        switch (item.AlertDefinition.source.type) {
          case 'PORT':
            alertDefinitions.push($.extend(alertDefinition, this.parseIt(item, this.get('portConfig'))));
            break;
          case 'METRIC':
            // map App.AlertMetricsSourceDefinition's
            var jmxMetric = item.AlertDefinition.source.jmx;
            var gangliaMetric = item.AlertDefinition.source.ganglia;
            if (jmxMetric) {
              alertDefinition.jmx_id = item.AlertDefinition.id + 'jmx';
              alertMetricsSourceDefinitions.push({
                id: alertDefinition.jmx_id,
                value: jmxMetric.value,
                property_list: jmxMetric.property_list
              });
            }
            if (gangliaMetric) {
              alertDefinition.ganglia_id = item.AlertDefinition.id + 'ganglia';
              alertMetricsSourceDefinitions.push({
                id: alertDefinition.ganglia_id,
                value: gangliaMetric.value,
                property_list: gangliaMetric.property_list
              });
            }

            // map App.AlertMetricsUriDefinition
            alertDefinition.uri_id = item.AlertDefinition.id + 'uri';
            item.AlertDefinition.source.uri.id = alertDefinition.uri_id;
            alertMetricsUriDefinitions.push(this.parseIt(item, this.get('uriConfig')));
            alertDefinitions.push(alertDefinition);
            break;
          case 'WEB':
            // map App.AlertMetricsUriDefinition
            alertDefinition.uri_id = item.AlertDefinition.id + 'uri';
            item.AlertDefinition.source.uri.id = alertDefinition.uri_id;
            alertMetricsUriDefinitions.push(this.parseIt(item, this.get('uriConfig')));
            alertDefinitions.push(alertDefinition);
            break;
          case 'AGGREGATE':
            alertDefinitions.push($.extend(alertDefinition, this.parseIt(item, this.get('aggregateConfig'))));
            break;
          case 'SCRIPT':
            alertDefinitions.push($.extend(alertDefinition, this.parseIt(item, this.get('scriptConfig'))));
            break;
          default:
            console.error('Incorrect Alert Definition type:', item.AlertDefinition);
        }
      }, this);

      alertDefinitionsToDelete.forEach(function(definitionId) {
        self.deleteRecord(existingAlertDefinitions.findProperty('id', definitionId));
      });

      // load all mapped data to model
      App.store.loadMany(this.get('reportModel'), alertReportDefinitions);
      App.store.loadMany(this.get('metricsSourceModel'), alertMetricsSourceDefinitions);
      this.setMetricsSourcePropertyLists(this.get('metricsSourceModel'), alertMetricsSourceDefinitions);
      App.store.loadMany(this.get('metricsUriModel'), alertMetricsUriDefinitions);
      App.store.loadMany(this.get('model'), alertDefinitions);
      this.setAlertDefinitionsRawSourceData(rawSourceData);
      if (App.router.get('mainAlertDefinitionsController')) {
        App.router.set('mainAlertDefinitionsController.mapperTimestamp', (new Date()).getTime());
      }
      App.store.commit();
    }
  },

  /**
   * set propertyList properties from <code>data</code> for records in <code>model</code>
   * @param model
   * @param data
   */
  setMetricsSourcePropertyLists: function (model, data) {
    data.forEach(function (record) {
      model.find().findProperty('id', record.id).set('propertyList', record.property_list);
    });
  },

  /**
   * set rawSourceDate properties for <code>App.AlertDefinition</code> records
   * @param rawSourceData
   */
  setAlertDefinitionsRawSourceData: function (rawSourceData) {
    var allDefinitions = App.AlertDefinition.find();
    for (var alertDefinitionId in rawSourceData) {
      if (rawSourceData.hasOwnProperty(alertDefinitionId)) {
        allDefinitions.findProperty('id', +alertDefinitionId).set('rawSourceData', rawSourceData[alertDefinitionId]);
      }
    }
  }
});

});

require.register("mappers/alert_groups_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * Mapper for <code>App.AlertGroup</code>
 * Save general information
 * Use <code>App.cache['previousAlertGroupsMap']</code> to store map alertDefinitions-alertGroups. This map is used
 * in the <code>App.AlertDefinitionsMapper</code> to correctly link alertDefinitions and alertGroups
 */
App.alertGroupsMapper = App.QuickDataMapper.create({

  model: App.AlertGroup,

  config: {
    id: 'AlertGroup.id',
    name: 'AlertGroup.name',
    default: 'AlertGroup.default',
    targets_key: 'AlertGroup.targets',
    targets_type: 'array',
    targets: {
      item: 'id'
    }
  },

  map: function (json) {
    if (!Em.isNone(json, 'items')) {

      var alertGroups = [],
        self = this,
        groupsToDelete = App.AlertGroup.find().mapProperty('id'),
        /**
         * AlertGroups-map for <code>App.AlertDefinitionsMappers</code>
         * Format:
         * <code>
         *   {
         *    alert_definition1_id: [alert_group1_id, alert_group2_id],
         *    alert_definition2_id: [alert_group3_id, alert_group1_id],
         *    ...
         *   }
         * </code>
         * @type {object}
         */
        alertDefinitionsGroupsMap = {},
        alertNotificationsGroupsMap = {};

      json.items.forEach(function(item) {
        var group = self.parseIt(item, self.get('config'));
        groupsToDelete = groupsToDelete.without(group.id);
        group.targets = [];
        group.definitions = [];
        if (item.AlertGroup.definitions) {
          item.AlertGroup.definitions.forEach(function(definition) {
            if (!group.definitions.contains(definition.id)) {
              group.definitions.push(definition.id);
            }
            if (Em.isNone(alertDefinitionsGroupsMap[definition.id])) {
              alertDefinitionsGroupsMap[definition.id] = [];
            }
            alertDefinitionsGroupsMap[definition.id].push(group.id);
          });
        }
        if (item.AlertGroup.targets) {
          item.AlertGroup.targets.forEach(function(target) {
            if (!group.targets.contains(target.id)) {
              group.targets.push(target.id);
            }
            if (Em.isNone(alertNotificationsGroupsMap[target.id])) {
              alertNotificationsGroupsMap[target.id] = [];
            }
            alertNotificationsGroupsMap[target.id].push(group.id);
          });
        }
        alertGroups.push(group);
      }, this);

      groupsToDelete.forEach(function(groupId) {
        self.deleteRecord(App.AlertGroup.find(groupId));
      });

      App.cache['previousAlertGroupsMap'] = alertDefinitionsGroupsMap;
      App.cache['alertNotificationsGroupsMap'] = alertNotificationsGroupsMap;
      App.store.loadMany(this.get('model'), alertGroups);
      App.store.commit();
    }
  }
});

});

require.register("mappers/alert_instances_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.alertInstanceMapper = App.QuickDataMapper.create({

  model : App.AlertInstance,

  config : {
    id: 'Alert.id',
    label: 'Alert.label',
    definition_name: 'Alert.definition_name',
    definition_id: 'Alert.definition_id',
    service_id: 'Alert.service_name',
    service_name: 'Alert.service_name',
    component_name: 'Alert.component_name',
    host_id: 'Alert.host_name',
    host_name: 'Alert.host_name',
    scope: 'Alert.scope',
    original_timestamp: 'Alert.original_timestamp',
    latest_timestamp: 'Alert.latest_timestamp',
    maintenance_state: 'Alert.maintenance_state',
    instance: 'Alert.instance',
    state: 'Alert.state',
    text: 'Alert.text'
  },

  map: function(json, skipDelete) {
    if (json.items) {
      var alertInstances = (skipDelete) ? this.mapWithoutDelete(json) : this.mapAndDelete(json);

      App.store.loadMany(this.get('model'), alertInstances);
    }
  },

  /**
   * method that used when we not on alert definition state
   * in this case we need to delete alerts that is not critical and not warning
   * @param json
   * @returns {Array}
   */
  mapAndDelete: function(json) {
    var self = this,
      alertInstances = [],
      model = this.get('model'),
      alertsToDelete = model.find().mapProperty('id');

    json.items.forEach(function (item) {
      var alert = this.parseIt(item, this.get('config'));
      alertInstances.push(alert);
      alertsToDelete = alertsToDelete.without(alert.id);
    }, this);


    alertsToDelete.forEach(function(alertId) {
      var item = model.find(alertId);
      self.deleteRecord(item);
    });

    return alertInstances;
  },

  /**
   * this method is used on alert definition page
   * @param json
   * @returns {Array}
   */
  mapWithoutDelete: function(json) {
    var alertInstances = [];
    json.items.forEach(function (item) {
      var alert = this.parseIt(item, this.get('config'));
      alertInstances.push(alert);
    }, this);
    return alertInstances;
  }

});

});

require.register("mappers/alert_notification_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.alertNotificationMapper = App.QuickDataMapper.create({
  model: App.AlertNotification,
  config: {
    id: 'AlertTarget.id',
    name: 'AlertTarget.name',
    type: 'AlertTarget.notification_type',
    description: 'AlertTarget.description',
    global: 'AlertTarget.global'
  },

  map: function (json) {
    if (json.items) {
      var result = [];
      var notificationsProperties = {};
      var notificationsAlertStates = {};
      var groupsMap = App.cache['alertNotificationsGroupsMap'];

      json.items.forEach(function (item) {
        var notification = this.parseIt(item, this.config);
        var groups = groupsMap && groupsMap[notification.id];
        if (groups) {
          notification.groups = groups;
        }
        result.push(notification);
        notificationsProperties[item.AlertTarget.id] = item.AlertTarget.properties;
        notificationsAlertStates[item.AlertTarget.id] = item.AlertTarget.alert_states;
      }, this);

      App.store.loadMany(this.get('model'), result);
      this.setProperties('properties', notificationsProperties);
      this.setProperties('alertStates', notificationsAlertStates);
    }
  },

  /**
   * Set values from <code>propertyMap</code> for <code>propertyName</code> for each record in model
   * @param propertyName
   * @param propertiesMap record_id to value map
   */
  setProperties: function (propertyName, propertiesMap) {
    var modelRecords = this.get('model').find();
    for (var recordId in propertiesMap) {
      if (propertiesMap.hasOwnProperty(recordId)) {
        modelRecords.findProperty('id', +recordId).set(propertyName, propertiesMap[recordId]);
      }
    }
  }
});
});

require.register("mappers/cluster_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


App.clusterMapper = App.QuickDataMapper.create({
    model : App.Cluster,
    map:function(json){
      if(!this.get('model')) {return;}
      if(json){
        var result = json;
        result = this.parseIt(result, this.config);
        App.store.load(this.get('model'), result);

        var cluster = App.Cluster.find(result.id);
        var clusterDesiredConfigs = [];
        // Create desired_configs_array
        if(json.Clusters.desired_configs){
          for(var site in json.Clusters.desired_configs){
            var tag = json.Clusters.desired_configs[site].tag;
            var configObj = App.ConfigSiteTag.create({
              site: site,
              tag: tag,
              hostOverrides: {}
            });
            if(json.Clusters.desired_configs[site].host_overrides!=null){
              var hostOverridesArray = {};
              json.Clusters.desired_configs[site].host_overrides.forEach(function(override){
                var hostname = override.host_name;
                hostOverridesArray[hostname] = override.tag;
              });
              configObj.set('hostOverrides', hostOverridesArray);
            }
            clusterDesiredConfigs.push(configObj);
          }
        }
        cluster.set('desiredConfigs', clusterDesiredConfigs);
      }
    },
    config : {
      id:'Clusters.cluster_id',
      cluster_name: 'Clusters.cluster_name',
      stack_name: 'Clusters.stack_name',
      version: 'Clusters.version',
      //$hosts: [1, 2, 3, 4],
      $racks: [1],
      max_hosts_per_rack: 'Clusters.max_hosts_per_rack'
    }
});

});

require.register("mappers/component_config_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.componentConfigMapper = App.QuickDataMapper.create({
  model: App.HostComponent,
  config: {
    id: 'id',
    work_status: 'HostRoles.state',
    passive_state: 'HostRoles.maintenance_state',
    component_name: 'HostRoles.component_name',
    host_name: 'HostRoles.host_name',
    $ha_status: '',
    $display_name_advanced: '',
    stale_configs: 'HostRoles.stale_configs',
    host_id: 'HostRoles.host_name',
    service_id: 'HostRoles.service_name',
    admin_state: 'HostRoles.desired_admin_state'
  },
  map: function (json) {
    console.time('App.componentConfigMapper execution time');
    var hostComponents = [];
    var serviceToHostComponentIdMap = {};
    var cacheServices = App.cache['services'];
    var loadedServiceComponentsMap = this.buildServiceComponentMap(cacheServices);
    var mapConfig = this.get('config');
    // We do not want to parse JSON if there is no need to
    var hostComponentJsonMap = {};
    var hostComponentJsonIds = [];
    var hostComponentJsonsToRemove = {};
    json.items.forEach(function (item) {
      item.host_components.forEach(function (host_component) {
        host_component.id = host_component.HostRoles.component_name + '_' + host_component.HostRoles.host_name;
        hostComponentJsonIds.push(host_component.id);
        hostComponentJsonMap[host_component.id] = host_component;
      });
    });
    this.get('model').find().forEach(function (hostComponent) {
      var hostComponentJson = hostComponentJsonMap[hostComponent.get('id')];
      if (!hostComponentJson &&
        !App.StackServiceComponent.find().findProperty('componentName', hostComponent.get('componentName')).get('isMaster')) {
        hostComponent.set('staleConfigs', false);
      }
      if (hostComponentJson!=null && hostComponent.get('staleConfigs') &&
          hostComponentJson.HostRoles.state == hostComponent.get('workStatus') &&
          hostComponentJson.HostRoles.maintenance_state == hostComponent.get('passiveState')) {
        // A component already exists with correct stale_configs flag and other values - no need to load again
        hostComponentJsonsToRemove[hostComponentJson.id] = hostComponentJson;
      }
    });
    hostComponentJsonIds.forEach(function (hcId) {
      if(!hostComponentJsonsToRemove[hcId]){
        var host_component = hostComponentJsonMap[hcId];
        var serviceName = host_component.HostRoles.service_name;
        hostComponents.push(this.parseIt(host_component, mapConfig));
        if (!serviceToHostComponentIdMap[serviceName]) {
          serviceToHostComponentIdMap[serviceName] = [];
        }
        serviceToHostComponentIdMap[serviceName].push(host_component.id);
      }
    }, this);
    App.store.loadMany(this.get('model'), hostComponents);
    this.addNewHostComponents(loadedServiceComponentsMap, serviceToHostComponentIdMap, cacheServices);
    console.timeEnd('App.componentConfigMapper execution time');
  },

  /**
   * build map that include loaded host-components to avoid duplicate loading
   * @param cacheServices
   * @return {Object}
   */
  buildServiceComponentMap: function (cacheServices) {
    var loadedServiceComponentsMap = {};

    cacheServices.forEach(function (cacheService) {
      var componentsMap = {};

      cacheService.host_components.forEach(function (componentId) {
        componentsMap[componentId] = true;
      });
      loadedServiceComponentsMap[cacheService.ServiceInfo.service_name] = componentsMap;
    });
    return loadedServiceComponentsMap;
  },

  /**
   * add only new host-components to every service
   * to update service - host-component relations in model
   * @param loadedServiceComponentsMap
   * @param serviceToHostComponentIdMap
   * @param cacheServices
   * @return {boolean}
   */
  addNewHostComponents: function (loadedServiceComponentsMap, serviceToHostComponentIdMap, cacheServices) {
    if (!loadedServiceComponentsMap || !serviceToHostComponentIdMap || !cacheServices) return false;

    for (var serviceName in serviceToHostComponentIdMap) {
      var loadedService = cacheServices.findProperty('ServiceInfo.service_name', serviceName);

      if (serviceToHostComponentIdMap[serviceName] && loadedService) {
        serviceToHostComponentIdMap[serviceName].forEach(function (componentId) {
          if (!loadedServiceComponentsMap[serviceName][componentId]) {
            loadedService.host_components.push(componentId)
          }
        });
      }
    }
    return true;
  }
});

});

require.register("mappers/components_state_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

App.componentsStateMapper = App.QuickDataMapper.create({

  model: App.Service,

  clientModel: App.ClientComponent,
  clientMap: {
    id: 'ServiceComponentInfo.component_name',
    service_id: 'ServiceComponentInfo.service_name',
    stack_info_id: 'ServiceComponentInfo.component_name',
    component_name: 'ServiceComponentInfo.component_name',
    service_name: 'ServiceComponentInfo.service_name',
    installed_count: 'ServiceComponentInfo.installed_count',
    started_count: 'ServiceComponentInfo.started_count',
    total_count: 'ServiceComponentInfo.total_count'
  },

  slaveModel: App.SlaveComponent,

  paths: {
    INSTALLED_PATH: 'ServiceComponentInfo.installed_count',
    STARTED_PATH: 'ServiceComponentInfo.started_count',
    TOTAL_PATH: 'ServiceComponentInfo.total_count'
  },
  configMap: {
    'DATANODE': {
      data_nodes_started: 'STARTED_PATH',
      data_nodes_installed: 'INSTALLED_PATH',
      data_nodes_total: 'TOTAL_PATH'
    },
    'NODEMANAGER': {
      node_managers_started: 'STARTED_PATH',
      node_managers_installed: 'INSTALLED_PATH',
      node_managers_total: 'TOTAL_PATH'
    },
    'HBASE_REGIONSERVER': {
      region_servers_started: 'STARTED_PATH',
      region_servers_installed: 'INSTALLED_PATH',
      region_servers_total: 'TOTAL_PATH'
    },
    'GANGLIA_MONITOR': {
      ganglia_monitors_started: 'STARTED_PATH',
      ganglia_monitors_installed: 'INSTALLED_PATH',
      ganglia_monitors_total: 'TOTAL_PATH'
    },
    'SUPERVISOR': {
      super_visors_started: 'STARTED_PATH',
      super_visors_installed: 'INSTALLED_PATH',
      super_visors_total: 'TOTAL_PATH'
    },
    'MAPREDUCE2_CLIENT': {
      map_reduce2_clients: 'INSTALLED_PATH'
    },
    'TEZ_CLIENT': {
      installed_clients: 'INSTALLED_PATH'
    },
    'HIVE_CLIENT': {
      installed_clients: 'INSTALLED_PATH'
    },
    'FALCON_CLIENT': {
      installed_clients: 'INSTALLED_PATH'
    },
    'OOZIE_CLIENT': {
      installed_clients: 'INSTALLED_PATH'
    },
    'ZOOKEEPER_CLIENT': {
      installed_clients: 'INSTALLED_PATH'
    },
    'PIG': {
      installed_clients: 'INSTALLED_PATH'
    },
    'SQOOP': {
      installed_clients: 'INSTALLED_PATH'
    },
    'YARN_CLIENT': {
      installed_clients: 'INSTALLED_PATH'
    },
    'HDFS_CLIENT': {
      installed_clients: 'INSTALLED_PATH'
    },
    'FLUME_HANDLER': {
      flume_handlers_total: 'TOTAL_PATH'
    }
  },
  /**
   * get formatted component config
   * @param componentName
   * @return {Object}
   */
  getComponentConfig: function (componentName) {
    var config = {};
    var componentConfig = this.get('configMap')[componentName];
    var paths = this.get('paths');

    for (var property in componentConfig) {
      if (paths[componentConfig[property]]) {
        config[property] = paths[componentConfig[property]];
      }
    }
    return config;
  },
  /**
   * get service extended model if it has one
   * @param serviceName
   * @return {Object|null}
   */
  getExtendedModel: function (serviceName) {
    if (App[App.Service.extendedModel[serviceName]]) {
      return App[App.Service.extendedModel[serviceName]].find(serviceName);
    }
    return null;
  },

  map: function (json) {
    console.time('App.componentsStateMapper execution time');

    var clients = [];
    var slaves = [];

    if (json.items) {
      json.items.forEach(function (item) {
        var componentConfig = this.getComponentConfig(item.ServiceComponentInfo.component_name);
        var parsedItem = this.parseIt(item, componentConfig);
        var service = App.Service.find(item.ServiceComponentInfo.service_name);
        var extendedModel = this.getExtendedModel(item.ServiceComponentInfo.service_name);
        var cacheService = App.cache['services'].findProperty('ServiceInfo.service_name', item.ServiceComponentInfo.service_name);

        if (item.ServiceComponentInfo.category === 'CLIENT') {
          clients.push(this.parseIt(item, this.clientMap));
        }
        if (item.ServiceComponentInfo.category === 'SLAVE') {
          // for now map for slaves and clients are equal but it may vary in future.
          slaves.push(this.parseIt(item, this.clientMap));
        }

        cacheService.client_components = clients.filterProperty('service_name', cacheService.ServiceInfo.service_name).mapProperty('component_name');
        cacheService.slave_components = slaves.filterProperty('service_name', cacheService.ServiceInfo.service_name).mapProperty('component_name');

        for (var i in parsedItem) {
          if (service.get('isLoaded')) {
            cacheService[i] = parsedItem[i];
            service.set(stringUtils.underScoreToCamelCase(i), parsedItem[i]);
            if (extendedModel) {
              extendedModel.set(stringUtils.underScoreToCamelCase(i), parsedItem[i]);
            }
          }
        }
      }, this)
    }
    App.store.loadMany(this.clientModel, clients);
    App.store.loadMany(this.slaveModel, slaves);

    console.timeEnd('App.componentsStateMapper execution time');
  }
});

});

require.register("mappers/hosts_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

var stringUtils = require('utils/string_utils');

App.hostsMapper = App.QuickDataMapper.create({

  model: App.Host,
  config: {
    id: 'Hosts.host_name',
    host_name: 'Hosts.host_name',
    public_host_name: 'Hosts.public_host_name',
    cluster_id: 'cluster_id',// Hosts.cluster_name
    rack: 'Hosts.rack_info',
    host_components_key: 'host_components',
    host_components_type: 'array',
    host_components: {
      item: 'id'
    },
    alerts_summary: 'alerts_summary',
    critical_warning_alerts_count: 'critical_warning_alerts_count',
    cpu: 'Hosts.cpu_count',
    cpu_physical: 'Hosts.ph_cpu_count',
    memory: 'Hosts.total_mem',
    disk_info: 'Hosts.disk_info',
    disk_total: 'metrics.disk.disk_total',
    disk_free: 'metrics.disk.disk_free',
    health_status: 'Hosts.host_status',
    load_one: 'metrics.load.load_one',
    load_five: 'metrics.load.load_five',
    load_fifteen: 'metrics.load.load_fifteen',
    cpu_system: 'metrics.cpu.cpu_system',
    cpu_user: 'metrics.cpu.cpu_user',
    mem_total: 'metrics.memory.mem_total',
    mem_free: 'metrics.memory.mem_free',
    last_heart_beat_time: "Hosts.last_heartbeat_time",
    os_arch: 'Hosts.os_arch',
    os_type: 'Hosts.os_type',
    ip: 'Hosts.ip',
    passive_state: 'Hosts.maintenance_state',
    index: 'index'
  },
  hostComponentConfig: {
    component_name: 'HostRoles.component_name',
    service_id: 'HostRoles.service_name',
    passive_state: 'HostRoles.maintenance_state',
    work_status: 'HostRoles.state',
    stale_configs: 'HostRoles.stale_configs',
    host_name: 'host_name',
    admin_state: 'HostRoles.desired_admin_state'
  },
  stackVersionConfig: {
    id: 'HostStackVersions.id',
    stack: 'HostStackVersions.stack',
    repo_id: 'repository_versions[0].RepositoryVersions.id',
    repo_version: 'repository_versions[0].RepositoryVersions.repository_version',
    display_name: 'repository_versions[0].RepositoryVersions.display_name',
    version: 'HostStackVersions.version',
    status: 'HostStackVersions.state',
    host_name: 'host_name',
    host_id: 'host_name',
    is_visible: 'is_visible'
  },
  map: function (json, returnMapped) {
    returnMapped = !!returnMapped;
    console.time('App.hostsMapper execution time');
    if (json.items) {
      var hostsWithFullInfo = [];
      var hostIds = {};
      var components = [];
      var stackVersions = [];
      var componentsIdMap = {};
      var cacheServices = App.cache['services'];
      var loadedServiceComponentsMap = App.get('componentConfigMapper').buildServiceComponentMap(cacheServices);
      var serviceToHostComponentIdMap = {};
      var selectedHosts = App.db.getSelectedHosts('mainHostController');

      json.items.forEach(function (item, index) {
        item.host_components = item.host_components || [];
        item.host_components.forEach(function (host_component) {
          host_component.id = host_component.HostRoles.component_name + "_" + item.Hosts.host_name;
          var component = this.parseIt(host_component, this.hostComponentConfig);
          var serviceName = host_component.HostRoles.service_name;

          component.id = host_component.HostRoles.component_name + "_" + item.Hosts.host_name;
          component.host_id = item.Hosts.host_name;
          component.host_name = item.Hosts.host_name;
          components.push(component);
          componentsIdMap[component.id] = component;
          if (!serviceToHostComponentIdMap[serviceName]) {
            serviceToHostComponentIdMap[serviceName] = [];
          }
          serviceToHostComponentIdMap[serviceName].push(component.id);
        }, this);

        if (App.get('supports.stackUpgrade')) {
          var currentVersion = item.stack_versions.findProperty('HostStackVersions.state', 'CURRENT');
          var currentVersionNumber = currentVersion && currentVersion.repository_versions
            ? Em.get(currentVersion.repository_versions[0], 'RepositoryVersions.repository_version') : '';
          item.stack_versions.forEach(function (stackVersion) {
            stackVersion.host_name = item.Hosts.host_name;
            stackVersion.is_visible = stringUtils.compareVersions(Em.get(stackVersion.repository_versions[0], 'RepositoryVersions.repository_version'), currentVersionNumber) >= 0
              || App.get('supports.displayOlderVersions') || !currentVersionNumber;
            stackVersions.push(this.parseIt(stackVersion, this.stackVersionConfig));
          }, this);
        }

        var alertsSummary = item.alerts_summary;
        item.critical_warning_alerts_count = alertsSummary ? (alertsSummary.CRITICAL || 0) + (alertsSummary.WARNING || 0) : 0;
        item.cluster_id = App.get('clusterName');
        item.index = index;

        if (App.get('supports.stackUpgrade')) {
          this.config = $.extend(this.config, {
            stack_versions_key: 'stack_versions',
            stack_versions_type: 'array',
            stack_versions: {
              item: 'HostStackVersions.id'
            }
          })
        }
        var parsedItem = this.parseIt(item, this.config);
        parsedItem.is_requested = true;
        parsedItem.selected = selectedHosts.contains(parsedItem.host_name);

        hostIds[item.Hosts.host_name] = parsedItem;

        hostsWithFullInfo.push(parsedItem);
      }, this);

      if(returnMapped){
        return hostsWithFullInfo;
      }


      App.Host.find().forEach(function (host) {
        if (!hostIds[host.get('hostName')]) {
          host.set('isRequested', false);
        }
      });
      App.HostComponent.find().filterProperty('isMaster').forEach(function(component) {
        if (componentsIdMap[component.get('id')]) componentsIdMap[component.get('id')].display_name_advanced = component.get('displayNameAdvanced');
      });
      App.store.commit();
      if (App.get('supports.stackUpgrade')) {
        App.store.loadMany(App.HostStackVersion, stackVersions);
      }
      App.store.loadMany(App.HostComponent, components);
      App.store.loadMany(App.Host, hostsWithFullInfo);
      var itemTotal = parseInt(json.itemTotal);
      if (!isNaN(itemTotal)) {
        App.router.set('mainHostController.filteredCount', itemTotal);
      }
      //bind host-components with service records
      App.get('componentConfigMapper').addNewHostComponents(loadedServiceComponentsMap, serviceToHostComponentIdMap, cacheServices);
    }
    console.timeEnd('App.hostsMapper execution time');
  }
});

});

require.register("mappers/racks_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */


App.racksMapper = App.QuickDataMapper.create({
  model: App.Rack,
  config: {
    id: "Racks.id",
    name: "Racks.name"
    //$hosts: ["host01", "host06", "host05"],
    //status: "Racks.status",
    //live_hosts_count: "Racks.live_hosts_count",
    //critical_hosts_count: "Racks.critical_hosts_count",
    //dead_hosts_count: "Racks.dead_hosts_count"
  }
});

});

require.register("mappers/repository_version_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * 'License'); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an 'AS IS' BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');

App.repoVersionMapper = App.QuickDataMapper.create({
  modelRepoVersions: App.RepositoryVersion,
  modelOperatingSystems: App.OS,
  modelRepositories: App.Repository,

  modelRepoVersion: {
    id: 'RepositoryVersions.id',
    stack_version_id: 'RepositoryVersions.stackVersionId',
    display_name: 'RepositoryVersions.display_name',
    repository_version: 'RepositoryVersions.repository_version',
    upgrade_pack: 'RepositoryVersions.upgrade_pack',
    stack_version_type: 'RepositoryVersions.stack_name',
    stack_version_number: 'RepositoryVersions.stack_version',
    operating_systems_key: 'operating_systems',
    operating_systems_type: 'array',
    operating_systems: {
      item: 'id'
    }
  },

  modelOS: {
    id: 'id',
    repository_version_id: 'repository_version_id',
    os_type: 'OperatingSystems.os_type',
    stack_name: 'OperatingSystems.stack_name',
    stack_version: 'OperatingSystems.stack_version',
    repositories_key: 'repositories',
    repositories_type: 'array',
    repositories: {
      item: 'id'
    }
  },

  modelRepository: {
    id: 'id',
    operating_system_id: 'Repositories.operating_system_id',
    base_url : 'Repositories.base_url',
    default_base_url : 'Repositories.default_base_url',
    latest_base_url : 'Repositories.latest_base_url',
    mirrors_list : 'Repositories.mirrors_list',
    os_type : 'Repositories.os_type',
    repo_id : 'Repositories.repo_id',
    repo_name : 'Repositories.repo_name',
    stack_name : 'Repositories.stack_name',
    stack_version : 'Repositories.stack_version'
  },

  map: function (json, loadAll) {
    var modelRepoVersions = this.get('modelRepoVersions');
    var modelOperatingSystems = this.get('modelOperatingSystems');
    var modelRepositories = this.get('modelRepositories');

    var resultRepoVersion = [];
    var resultOS = [];
    var resultRepo = [];

    if (json && json.items) {
      json.items.forEach(function (item) {
        if (loadAll || (item.RepositoryVersions && !App.StackVersion.find().someProperty('repositoryVersion.id', item.RepositoryVersions.id))) {
          var repo = item;
          var osArray = [];
          if (item.operating_systems) {
            item.operating_systems.forEach(function (os) {
              os.id = item.RepositoryVersions.repository_version + os.OperatingSystems.os_type;
              os.repository_version_id = repo.id;
              var repoArray = [];
              if (Em.get(os, 'repositories')) {
                os.repositories.forEach(function (repo) {
                  repo.id = repo.Repositories.repo_id + os.id;
                  repo.operating_system_id = os.id;
                  repoArray.pushObject(repo);
                  resultRepo.push(this.parseIt(repo, this.get('modelRepository')));
                }, this);
              }
              os.repositories = repoArray;
              osArray.pushObject(os);
              resultOS.push(this.parseIt(os, this.get('modelOS')));
            }, this);
          }
          repo.operating_systems = osArray;
          resultRepoVersion.push(this.parseIt(repo, this.get('modelRepoVersion')));
        }
      }, this);
    }
    App.store.commit();
    App.store.loadMany(modelRepositories, resultRepo);
    App.store.loadMany(modelOperatingSystems, resultOS);
    App.store.loadMany(modelRepoVersions, resultRepoVersion);
  }
});

});

require.register("mappers/root_service_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.rootServiceMapper = App.QuickDataMapper.create({
  root_service_model: App.RootService,
  root_service_component_model: App.RootServiceComponents,

  configRootService: {
    id: 'service_name',
    service_name: 'service_name',
    components_key: 'components',
    components_type: 'array',
    components: {
      item: 'id'
    }
  },

  configRootServiceComponents: {
    id: 'component_name',
    component_name: 'component_name',
    component_version: 'component_version',
    server_clock: 'server_clock',
    service_name: 'service_name',
    properties: 'properties'
  },

  map: function (data) {
    var rootServiceModel = this.get('root_service_model');
    var rootServiceComponentModel = this.get('root_service_component_model');
    var rootService = {}, rootServiceComponents = [];
    rootService.id = rootService.service_name = data.RootService.service_name;
    data.components.forEach(function (item) {
      item.RootServiceComponents.id = item.RootServiceComponents.component_name;
      rootServiceComponents.push(this.parseIt(item.RootServiceComponents, this.configRootServiceComponents));
    }, this);
    rootService.components =  rootServiceComponents;
    App.store.commit();
    App.store.loadMany(rootServiceComponentModel, rootServiceComponents);
    App.store.load(rootServiceModel, this.parseIt(rootService, this.configRootService));
  }
});

});

require.register("mappers/server_data_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * initialize common cache container for mappers
 * App.cache contains shared data, used for syncronizing incoming server data among mappers
 */
App.cache = {
  'previousHostStatuses': {},
  'previousComponentStatuses': {},
  'previousComponentPassiveStates': {},
  'services': [],
  'currentConfigVersions': {}
};

App.ServerDataMapper = Em.Object.extend({
  jsonKey: false,
  map: function (json) {
    if (json) {
      var model = this.get('model');
      var jsonKey = this.get('jsonKey');

      if (jsonKey && json[jsonKey]) { // if data come as { hdfs: {...} }
        json = json[jsonKey];
      }

      $.each(json, function (field, value) {
        model.set(field, value);
      })
    }
  }
});


App.QuickDataMapper = App.ServerDataMapper.extend({
  config: {},
  model: null,
  map: function (json) {
    if (!this.get('model')) {
      return;
    }

    if (json.items) {
      var result = [];

      json.items.forEach(function (item) {
        result.push(this.parseIt(item, this.config));
      }, this);

      App.store.loadMany(this.get('model'), result);
    }
  },

  parseIt: function (data, config) {
    var result = {};
    for ( var i in config) {
      if (i.substr(0, 1) === '$') {
        i = i.substr(1, i.length);
        result[i] = config['$' + i];
      } else {
        var isSpecial = false;
        if (i.substr(-5) == '_type') {
          var prefix = i.substr(0, i.length - 5);
          isSpecial = config[prefix + '_key'] != null;
        } else if (i.substr(-4) == '_key') {
          var prefix = i.substr(0, i.length - 4);
          isSpecial = config[prefix + '_type'] != null;
        }
        if (!isSpecial && typeof config[i] == 'string') {
          result[i] = this.getJsonProperty(data, config[i]);
        } else if (typeof config[i] == 'object') {
          result[i] = [];
          var _data = this.getJsonProperty(data, config[i + '_key']);
          var _type = config[i + '_type'];
          var l = _data.length;
          for ( var index = 0; index < l; index++) {
            if (_type == 'array') {
              result[i].push(this.getJsonProperty(_data[index], config[i].item));
            } else {
              result[i].push(this.parseIt(_data[index], config[i]));
            }
          }
          if(_type == 'array'){
            result[i] = result[i].sort();
          }
        }
      }
    }
    return result;
  },

  getJsonProperty: function (json, path) {
    var pathArr = path.split('.');
    var current = json;
    pathArr = this.filterDotted(pathArr);
    while (pathArr.length && current) {
      if (pathArr[0].substr(-1) == ']') {
        var index = parseInt(pathArr[0].substr(-2, 1));
        var attr = pathArr[0].substr(0, pathArr[0].length - 3);
        if (attr in current) {
          current = current[attr][index];
        }
      } else {
        current = current[pathArr[0]];
      }
      pathArr.splice(0, 1);
    }
    return current;
  },

  filterDotted: function(arr) {
    var buffer = [];
    var dottedBuffer = [];
    var dotted = false;
    arr.forEach(function(item) {
      if(/\['|\["/.test(item)) {
        dottedBuffer.push(item.substr(2, item.length));
        dotted = true;
      } else if (dotted && !/\]'|"\]/.test(item)) {
        dottedBuffer.push(item);
      } else if (/']|"]/.test(item)) {
        dottedBuffer.push(item.substr(0, item.length - 2));
        buffer.push(dottedBuffer.join('.'));
        dotted = false;
        dottedBuffer = [];
      } else {
        buffer.push(item);
      }
    });

    return buffer;
  },

  /**
   * properly delete record from model
   * @param item
   */
  deleteRecord: function (item) {
    item.deleteRecord();
    App.store.commit();
    item.get('stateManager').transitionTo('loading');
    console.log('Record with id:' + item.get('id') + ' was deleted from model');
  },
  /**
   * check mutable fields whether they have been changed and if positive
   * return host object only with properties, that contains new value
   * @param current
   * @param previous
   * @param fields
   * @return {*}
   */
  getDiscrepancies: function (current, previous, fields) {
    var result = {};
    if (previous) {
      fields.forEach(function (field) {
        if (Array.isArray(current[field])) {
          if (JSON.stringify(current[field]) !== JSON.stringify(previous[field])) {
            result[field] = current[field];
            result.isLoadNeeded = true;
          }
        } else {
          if (current[field] != previous[field]) result[field] = current[field];
        }
      });
      return result;
    }
    return current;
  },

  calculateState: function (json) {
//    var stateEqual = (json.desired_status != json.work_status);
//    if (stateEqual) {
//      if (json.desired_status == 'STARTED' && json.work_status == 'INSTALLED') {
//        json.work_status = 'STARTING';
//      } else if (json.desired_status == 'INSTALLED' && json.work_status == 'STARTED') {
//        json.work_status = 'STOPPING';
//      }
//    }
    return json;
  }
});

});

require.register("mappers/service_config_version_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.serviceConfigVersionsMapper = App.QuickDataMapper.create({
  model: App.ServiceConfigVersion,
  config: {
    service_name: 'service_name',
    service_id: 'service_name',
    version: "service_config_version",
    create_time: 'createtime',
    group_id: 'group_id',
    group_name: 'group_name',
    hosts: 'hosts',
    author: 'user',
    notes: 'service_config_version_note',
    is_current: 'is_current',
    index: 'index'
  },
  map: function (json) {
    var result = [];
    var itemIds = {};
    var serviceToHostMap = {};

    if (json && json.items) {
      json.items.forEach(function (item, index) {
        var parsedItem = this.parseIt(item, this.get('config'));
        parsedItem.id = parsedItem.service_name + '_' + parsedItem.version;
        parsedItem.is_requested = true;
        itemIds[parsedItem.id] = true;
        parsedItem.index = index;
        if (serviceToHostMap[item.service_name]) {
          serviceToHostMap[item.service_name] = serviceToHostMap[item.service_name].concat(item.hosts);
        } else {
          serviceToHostMap[item.service_name] = item.hosts;
        }
        result.push(parsedItem);
      }, this);

      this.get('model').find().forEach(function (item) {
        if (!itemIds[item.get('id')]) {
          item.set('isRequested', false);
        }
      });
      var itemTotal = parseInt(json.itemTotal);
      if (!isNaN(itemTotal)) {
        App.router.set('mainConfigHistoryController.filteredCount', itemTotal);
      }
      /**
       * this code sets hostNames for default confg group
       * by excluding hostNames that belongs to not default groups
       * from list of all hosts
       */
      Object.keys(serviceToHostMap).forEach(function(sName) {
        var defaultHostNames = App.get('allHostNames');
        for (var i = 0; i < serviceToHostMap[sName].length; i++) {
          defaultHostNames = defaultHostNames.without(serviceToHostMap[sName][i]);
        }
        var defVer = result.find(function(v) {
          return v.is_current && v.group_id == -1 && v.service_name == sName;
        });
        if (defVer) {
          defVer.hosts = defaultHostNames;
        }
      });
      App.store.commit();
      App.store.loadMany(this.get('model'), result);
    }
  }
});

});

require.register("mappers/service_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.serviceMapper = App.QuickDataMapper.create({
  map: function (json) {
    console.time("App.serviceMapper execution time");

    json.items.forEach(function (service) {
      var cachedService = App.cache['services'].findProperty('ServiceInfo.service_name', service.ServiceInfo.service_name);
      if (cachedService) {
        // restore service workStatus
        App.Service.find(cachedService.ServiceInfo.service_name).set('workStatus', service.ServiceInfo.state);
        cachedService.ServiceInfo.state = service.ServiceInfo.state;
        cachedService.ServiceInfo.passive_state = service.ServiceInfo.maintenance_state;

        //check whether Alerts installed and started
        if (service.legacy_alerts) {
          cachedService.ServiceInfo.critical_alerts_count = service.legacy_alerts.summary.CRITICAL + service.legacy_alerts.summary.WARNING;
        }
      } else {
        var serviceData = {
          ServiceInfo: {
            service_name: service.ServiceInfo.service_name,
            state: service.ServiceInfo.state,
            passive_state: service.ServiceInfo.maintenance_state
          },
          host_components: [],
          components: []
        };

        //check whether Alerts installed and started
        if (service.legacy_alerts) {
          serviceData.ServiceInfo.critical_alerts_count = service.legacy_alerts.summary.CRITICAL + service.legacy_alerts.summary.WARNING;
        }
        App.cache['services'].push(serviceData);
      }
    });

    console.timeEnd("App.serviceMapper execution time");
  }
});

});

require.register("mappers/service_metrics_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
var misc = require('utils/misc');
var stringUtils = require('utils/string_utils');
var dateUtils = require('utils/date');
var previousResponse = [];

App.serviceMetricsMapper = App.QuickDataMapper.create({

  model: App.Service,
  config: {
    id: 'ServiceInfo.service_name',
    service_name: 'ServiceInfo.service_name',
    work_status: 'ServiceInfo.state',
    passive_state: 'ServiceInfo.passive_state',
    critical_alerts_count: 'ServiceInfo.critical_alerts_count',
    $rand: Math.random(),
    $alerts: [ 1, 2, 3 ],
    host_components: 'host_components',
    tool_tip_content: 'tool_tip_content',
    installed_clients: 'installed_clients',
    client_components: 'client_components',
    slave_components: 'slave_components'
  },
  hdfsConfig: {
    version: 'nameNodeComponent.host_components[0].metrics.dfs.namenode.Version',
    active_name_node_id: 'active_name_node_id',
    standby_name_node_id: 'standby_name_node_id',
    standby_name_node2_id: 'standby_name_node2_id',
    journal_nodes: 'journal_nodes',
    metrics_not_available: 'metrics_not_available',
    name_node_start_time: 'nameNodeComponent.host_components[0].metrics.runtime.StartTime',
    jvm_memory_heap_used: 'nameNodeComponent.host_components[0].metrics.jvm.HeapMemoryUsed',
    jvm_memory_heap_max: 'nameNodeComponent.host_components[0].metrics.jvm.HeapMemoryMax',
    live_data_nodes: 'live_data_nodes',
    dead_data_nodes: 'dead_data_nodes',
    decommission_data_nodes: 'decommission_data_nodes',
    capacity_used: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.CapacityUsed',
    capacity_total: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.CapacityTotal',
    capacity_remaining: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.CapacityRemaining',
    dfs_total_blocks: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.BlocksTotal',
    dfs_corrupt_blocks: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.CorruptBlocks',
    dfs_missing_blocks: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.MissingBlocks',
    dfs_under_replicated_blocks: 'nameNodeComponent.host_components[0].metrics.dfs.FSNamesystem.UnderReplicatedBlocks',
    dfs_total_files: 'nameNodeComponent.host_components[0].metrics.dfs.namenode.TotalFiles',
    upgrade_status: 'nameNodeComponent.host_components[0].metrics.dfs.namenode.UpgradeFinalized',
    safe_mode_status: 'nameNodeComponent.host_components[0].metrics.dfs.namenode.Safemode',
    name_node_cpu: 'nameNodeComponent.host_components[0].metrics.cpu.cpu_wio',
    name_node_rpc: 'nameNodeComponent.host_components[0].metrics.rpc.RpcQueueTime_avg_time',
    data_nodes_started: 'data_nodes_started',
    data_nodes_installed: 'data_nodes_installed',
    data_nodes_total: 'data_nodes_total'
  },
  yarnConfig: {
    version: 'resourceManagerComponent.ServiceComponentInfo.Version',
    resource_manager_start_time: 'resourceManagerComponent.ServiceComponentInfo.StartTime',
    jvm_memory_heap_used: 'resourceManagerComponent.host_components[0].metrics.jvm.HeapMemoryUsed',
    jvm_memory_heap_max: 'resourceManagerComponent.host_components[0].metrics.jvm.HeapMemoryMax',
    containers_allocated: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AllocatedContainers',
    containers_pending: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.PendingContainers',
    containers_reserved: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.ReservedContainers',
    apps_submitted: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AppsSubmitted',
    apps_running: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AppsRunning',
    apps_pending: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AppsPending',
    apps_completed: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AppsCompleted',
    apps_killed: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AppsKilled',
    apps_failed: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AppsFailed',
    node_managers_count_active: 'resourceManagerComponent.ServiceComponentInfo.rm_metrics.cluster.activeNMcount',
    node_managers_count_lost: 'resourceManagerComponent.ServiceComponentInfo.rm_metrics.cluster.lostNMcount',
    node_managers_count_unhealthy: 'resourceManagerComponent.ServiceComponentInfo.rm_metrics.cluster.unhealthyNMcount',
    node_managers_count_rebooted: 'resourceManagerComponent.ServiceComponentInfo.rm_metrics.cluster.rebootedNMcount',
    node_managers_count_decommissioned: 'resourceManagerComponent.ServiceComponentInfo.rm_metrics.cluster.decommissionedNMcount',
    allocated_memory: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AllocatedMB',
    available_memory: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.AvailableMB',
    reserved_memory: 'resourceManagerComponent.host_components[0].metrics.yarn.Queue.root.ReservedMB',
    queue: 'resourceManagerComponent.queue',
    node_managers_started: 'node_managers_started',
    node_managers_installed: 'node_managers_installed',
    node_managers_total: 'node_managers_total'
  },
  mapReduce2Config: {
    version: 'jobHistoryServerComponent.ServiceComponentInfo.Version',
    map_reduce2_clients: 'map_reduce2_clients'
  },
  hbaseConfig: {
    version: 'masterComponent.ServiceComponentInfo.Version',
    master_start_time: 'masterComponent.ServiceComponentInfo.MasterStartTime',
    master_active_time: 'masterComponent.ServiceComponentInfo.MasterActiveTime',
    average_load: 'masterComponent.ServiceComponentInfo.AverageLoad',
    regions_in_transition: 'regions_in_transition',
    revision: 'masterComponent.ServiceComponentInfo.Revision',
    heap_memory_used: 'masterComponent.ServiceComponentInfo.HeapMemoryUsed',
    heap_memory_max: 'masterComponent.ServiceComponentInfo.HeapMemoryMax',
    region_servers_started: 'region_servers_started',
    region_servers_installed: 'region_servers_installed',
    region_servers_total: 'region_servers_total'
  },
  stormConfig: {
    total_tasks: 'restApiComponent.tasksTotal',
    total_slots: 'restApiComponent.slotsTotal',
    free_slots: 'restApiComponent.slotsFree',
    used_slots: 'restApiComponent.slotsUsed',
    topologies: 'restApiComponent.topologies',
    total_executors: 'restApiComponent.executorsTotal',
    nimbus_uptime: 'restApiComponent.nimbusUptime',
    super_visors_started: 'super_visors_started',
    super_visors_installed: 'super_visors_installed',
    super_visors_total: 'super_visors_total'
  },
  flumeConfig: {
    flume_handlers_total: 'flume_handlers_total'
  },
  flumeAgentConfig: {
    name: 'HostComponentProcess.name',
    status: 'HostComponentProcess.status',
    host_id: 'HostComponentProcess.host_name',
    host_name: 'HostComponentProcess.host_name',
    channels_count: 'HostComponentProcess.channels_count',
    sources_count: 'HostComponentProcess.sources_count',
    sinks_count: 'HostComponentProcess.sinks_count'
  },

  model3: App.HostComponent,
  config3: {
    id: 'id',
    work_status: 'HostRoles.state',
    passive_state:'HostRoles.maintenance_state',
    desired_status: 'HostRoles.desired_state',
    component_name: 'HostRoles.component_name',
    host_id: 'HostRoles.host_name',
    host_name: 'HostRoles.host_name',
    stale_configs: 'HostRoles.stale_configs',
    ha_status: 'HostRoles.ha_state',
    display_name_advanced: 'display_name_advanced',
    $service_id: 'none', /* will be set outside of parse function */
    admin_state: 'HostRoles.desired_admin_state'
  },

  map: function (json) {
    console.time('App.serviceMetricsMapper execution time');
    if (json.items) {

      // Host components
      var hostComponents = [];
      var services = App.cache['services'];
      var previousComponentStatuses = App.cache['previousComponentStatuses'];
      var previousComponentPassiveStates = App.cache['previousComponentPassiveStates'];
      var result = [];
      /**
       * services contains constructed service-components structure from components array
       */

      services.forEach(function(service) {
        service.host_components = [];
        service.components = [];
      });

      json.items.forEach(function (component) {
        var serviceName = component.ServiceComponentInfo.service_name;
        var service = services.findProperty('ServiceInfo.service_name', serviceName);
        if (service) {
          service.components.push(component);
        }
        component.host_components.forEach(function (host_component) {
          host_component.id = host_component.HostRoles.component_name + "_" + host_component.HostRoles.host_name;
          previousComponentStatuses[host_component.id] = host_component.HostRoles.state;
          previousComponentPassiveStates[host_component.id] = host_component.HostRoles.maintenance_state;
          this.config3.ha_status = host_component.HostRoles.component_name == "HBASE_MASTER" ?
            'metrics.hbase.master.IsActiveMaster' : 'HostRoles.ha_state';
          var comp = this.parseIt(host_component, this.config3);
          comp.service_id = serviceName;
          hostComponents.push(comp);
        }, this);
      }, this);

      this.computeAdditionalRelations(hostComponents, services);
      //load master components to model
      App.HostComponent.find().filterProperty('isMaster').forEach(function (hostComponent) {
        if (hostComponent && !hostComponents.someProperty('id', hostComponent.get('id'))) {
          this.deleteRecord(hostComponent);
          var serviceCache = services.findProperty('ServiceInfo.service_name', hostComponent.get('service.serviceName'));
          if (serviceCache) {
            serviceCache.host_components = serviceCache.host_components.without(hostComponent.get('id'));
          }
        }
      }, this);
      App.store.loadMany(this.get('model3'), hostComponents);

      //parse service metrics from components
      services.forEach(function (item) {
        var finalJson = [];

        hostComponents.filterProperty('service_id', item.ServiceInfo.service_name).mapProperty('id').forEach(function (hostComponent) {
          if (!item.host_components.contains(hostComponent)) {
            item.host_components.push(hostComponent);
          }
        }, this);
        item.host_components.sort();

        if (item && item.ServiceInfo && item.ServiceInfo.service_name == "HDFS") {
          finalJson = this.hdfsMapper(item);
          finalJson.rand = Math.random();
          result.push(finalJson);
          App.store.load(App.HDFSService, finalJson);
        } else if (item && item.ServiceInfo && item.ServiceInfo.service_name == "HBASE") {
          finalJson = this.hbaseMapper(item);
          finalJson.rand = Math.random();
          result.push(finalJson);
          App.store.load(App.HBaseService, finalJson);
        } else if (item && item.ServiceInfo && item.ServiceInfo.service_name == "FLUME") {
          finalJson = this.flumeMapper(item);
          finalJson.rand = Math.random();
          result.push(finalJson);
          App.store.load(App.FlumeService, finalJson);
          App.store.loadMany(App.FlumeAgent, finalJson.agentJsons);
        } else if (item && item.ServiceInfo && item.ServiceInfo.service_name == "YARN") {
          finalJson = this.yarnMapper(item);
          finalJson.rand = Math.random();
          result.push(finalJson);
          App.store.load(App.YARNService, finalJson);
        } else if (item && item.ServiceInfo && item.ServiceInfo.service_name == "MAPREDUCE2") {
          finalJson = this.mapreduce2Mapper(item);
          finalJson.rand = Math.random();
          result.push(finalJson);
          App.store.load(App.MapReduce2Service, finalJson);
        } else if (item && item.ServiceInfo && item.ServiceInfo.service_name == "STORM") {
          finalJson = this.stormMapper(item);
          finalJson.rand = Math.random();
          this.mapQuickLinks(finalJson, item);
          result.push(finalJson);
          App.store.load(App.StormService, finalJson);
        } else {
          finalJson = this.parseIt(item, this.config);
          finalJson.rand = Math.random();
          this.mapQuickLinks(finalJson, item);
          result.push(finalJson);
        }
      }, this);

      var stackServices = App.StackService.find().mapProperty('serviceName');
      result = misc.sortByOrder(stackServices, result);

      //load services to model
      App.store.loadMany(this.get('model'), result);
      /*if (previousResponse.length !== result.length) {
        App.store.loadMany(this.get('model'), result);
      } else {
        result.forEach(function (serviceJson) {
          var fields = ['passive_state','work_status', 'rand', 'alerts', 'quick_links', 'host_components', 'tool_tip_content', 'critical_alerts_count'];
          var service = this.get('model').find(serviceJson.id);
          var modifiedData = this.getDiscrepancies(serviceJson, previousResponse.findProperty('id', serviceJson.id), fields);
          if (modifiedData.isLoadNeeded) {
            App.store.load(this.get('model'), serviceJson);
          } else {
            for (var property in modifiedData) {
              service.set(stringUtils.underScoreToCamelCase(property), modifiedData[property]);
            }
          }
        }, this)
      }

      previousResponse = result;*/
    }
    console.timeEnd('App.serviceMetricsMapper execution time');
  },
  /**
   * compute display name of host-components
   * compute tooltip content of services by host-components
   * @param hostComponents
   * @param services
   */
  computeAdditionalRelations: function (hostComponents, services) {
    var isSecondaryNamenode = hostComponents.findProperty('component_name', 'SECONDARY_NAMENODE');
    var isRMHAEnabled = hostComponents.filterProperty('component_name', 'RESOURCEMANAGER').length > 1;
    services.setEach('tool_tip_content', '');
    // set tooltip for client-only services
    var clientOnlyServiceNames = App.get('services.clientOnly');
    clientOnlyServiceNames.forEach(function(serviceName) {
      var service = services.findProperty('ServiceInfo.service_name', serviceName);
      if (service) {
        service.tool_tip_content = Em.I18n.t('services.service.summary.clientOnlyService.ToolTip');
      }
    });
    hostComponents.forEach(function (hostComponent) {
      var service = services.findProperty('ServiceInfo.service_name', hostComponent.service_id);
      if (hostComponent) {
        // set advanced nameNode display name for HA, Active NameNode or Standby NameNode
        // this is useful on three places: 1) HDFS health status hover tooltip, 2) HDFS service summary 3) NameNode component on host detail page
        if (hostComponent.component_name === 'NAMENODE' && !isSecondaryNamenode) {
          var hdfs = this.hdfsMapper(service);
          var hostName = hostComponent.host_id;
          var activeNNText = Em.I18n.t('services.service.summary.nameNode.active');
          var standbyNNText = Em.I18n.t('services.service.summary.nameNode.standby');
          if (hdfs) {
            // active_name_node_id format : NAMENODE_c6401.ambari.apache.org
            if (hdfs.active_name_node_id && hdfs.active_name_node_id.contains(hostName)) {
              hostComponent.display_name_advanced = activeNNText;
            } else if ((hdfs.standby_name_node_id && hdfs.standby_name_node_id.contains(hostName)) || ( hdfs.standby_name_node2_id && hdfs.standby_name_node2_id.contains(hostName))) {
              hostComponent.display_name_advanced = standbyNNText;
            } else {
              hostComponent.display_name_advanced = null;
            }
          }
        } else if (hostComponent.component_name === 'HBASE_MASTER') {
          if (hostComponent.work_status === 'STARTED') {
            (hostComponent.ha_status === true || hostComponent.ha_status == 'true') ?
              hostComponent.display_name_advanced = this.t('dashboard.services.hbase.masterServer.active') :
              hostComponent.display_name_advanced = this.t('dashboard.services.hbase.masterServer.standby');
          } else {
            hostComponent.display_name_advanced = null;
          }
        } else if (hostComponent.component_name === 'RESOURCEMANAGER' && isRMHAEnabled && hostComponent.work_status === 'STARTED') {
          switch (hostComponent.ha_status) {
            case 'ACTIVE':
              hostComponent.display_name_advanced = Em.I18n.t('dashboard.services.yarn.resourceManager.active');
              break;
            case 'STANDBY':
              hostComponent.display_name_advanced = Em.I18n.t('dashboard.services.yarn.resourceManager.standby');
              break;
          }
        }
        if (service) {
          if (hostComponent.display_name_advanced) {
            service.tool_tip_content += hostComponent.display_name_advanced + " " + App.HostComponentStatus.getTextStatus(hostComponent.work_status) + "<br/>";
          } else {
            service.tool_tip_content += App.format.role(hostComponent.component_name) + " " + App.HostComponentStatus.getTextStatus(hostComponent.work_status) + "<br/>";
          }
        }
      }
    }, this)
  },
  /**
   * Map quick links to services:OOZIE,GANGLIA
   * @param finalJson
   * @param item
   */
  mapQuickLinks: function (finalJson, item){
    if(!(item && item.ServiceInfo)) return;
    var quickLinks = {
      OOZIE: [19],
      GANGLIA: [20],
      STORM: [31],
      FALCON: [32],
      RANGER: [33],
      SPARK: [34],
      EAGLE: [35,36]
    };
    if (quickLinks[item.ServiceInfo.service_name])
      finalJson.quick_links = quickLinks[item.ServiceInfo.service_name];
  },

  hdfsMapper: function (item) {
    var finalConfig = jQuery.extend({}, this.config);
    // Change the JSON so that it is easy to map
    var hdfsConfig = this.hdfsConfig;
    item.components.forEach(function (component) {
      if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "NAMENODE") {
        //enabled HA
        if ( component.host_components.length == 2) {
          var haState1;
          var haState2;
          if (component.host_components[1].metrics && component.host_components[1].metrics.dfs) {
            haState2 = component.host_components[1].metrics.dfs.FSNamesystem.HAState;
          }
          if (component.host_components[0].metrics && component.host_components[0].metrics.dfs) {
            haState1 = component.host_components[0].metrics.dfs.FSNamesystem.HAState;
          }
          var active_name_node = [];
          var standby_name_nodes = [];
          switch (haState1) {
            case "active":
              active_name_node.push(component.host_components[0].HostRoles.host_name);
              break;
            case "standby":
              standby_name_nodes.push(component.host_components[0].HostRoles.host_name);
              break;
          }
          switch (haState2) {
            case "active":
              active_name_node.push(component.host_components[1].HostRoles.host_name);
              break;
            case "standby":
              standby_name_nodes.push(component.host_components[1].HostRoles.host_name);
              break;
          }
          item.active_name_node_id = null;
          item.standby_name_node_id = null;
          item.standby_name_node2_id = null;
          switch (active_name_node.length) {
            case 1:
              item.active_name_node_id = 'NAMENODE' + '_' + active_name_node[0];
              break;
          }
          switch (standby_name_nodes.length) {
            case 1:
              item.standby_name_node_id = 'NAMENODE' + '_' + standby_name_nodes[0];
              break;
            case 2:
              item.standby_name_node_id = 'NAMENODE' + '_' + standby_name_nodes[0];
              item.standby_name_node2_id = 'NAMENODE' + '_' + standby_name_nodes[1];
              break;
          }
          // important: active nameNode always at host_components[0]; if no active, then any nameNode could work.
          if (haState2 == "active") { // change places for all model bind with host_component[0]
            var tmp = component.host_components[1];
            component.host_components[1] = component.host_components[0];
            component.host_components[0] = tmp;
          }
        }
        item.nameNodeComponent = component;
        finalConfig = jQuery.extend(finalConfig, hdfsConfig);
        // Get the live, dead & decommission nodes from string json
        if (component.host_components[0].metrics && component.host_components[0].metrics.dfs && component.host_components[0].metrics.dfs.namenode) {
          item.metrics_not_available = false;
          var decommissionNodesJson = App.parseJSON(component.host_components[0].metrics.dfs.namenode.DecomNodes);
          var deadNodesJson = App.parseJSON(component.host_components[0].metrics.dfs.namenode.DeadNodes);
          var liveNodesJson = App.parseJSON(component.host_components[0].metrics.dfs.namenode.LiveNodes);
        } else {
          item.metrics_not_available = true;
        }
        item.decommission_data_nodes = [];
        item.dead_data_nodes = [];
        item.live_data_nodes = [];
        for (var host in decommissionNodesJson) {
          item.decommission_data_nodes.push('DATANODE'+ '_' + host);
        }
        for (var host in deadNodesJson) {
          item.dead_data_nodes.push('DATANODE'+ '_' + host);
        }
        for (var host in liveNodesJson) {
          item.live_data_nodes.push('DATANODE'+ '_' + host);
        }
      }
      if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "JOURNALNODE") {
        if (!item.journal_nodes) {
          item.journal_nodes = [];
        }
        if (component.host_components) {
          component.host_components.forEach(function (hc) {
            item.journal_nodes.push(hc.HostRoles.host_name);
          });
        }
      }
    });
    // Map
    var finalJson = this.parseIt(item, finalConfig);
    finalJson.quick_links = [1, 2, 3, 4];

    return finalJson;
  },
  yarnMapper: function (item) {
    var result = [];
    var self = this;
    var finalConfig = jQuery.extend({}, this.config);
    // Change the JSON so that it is easy to map
    var yarnConfig = this.yarnConfig;
    item.components.forEach(function (component) {
      if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "RESOURCEMANAGER") {
        item.resourceManagerComponent = component;

        // if YARN has two host components, ACTIVE one should be first in component.host_components array for proper metrics mapping
        if (component.host_components.length === 2) {
          var activeRM = component.host_components.findProperty('HostRoles.ha_state', 'ACTIVE');
          // if "second" RM isn't STARTED his ha_status is null (not STANDBY)
          var standbyRM = component.host_components.filter(function(host_component) {return host_component.HostRoles.ha_state !== 'ACTIVE';})[0];
          if (activeRM && standbyRM) {
            component.host_components = [activeRM, standbyRM];
          }
        }

        if (component.host_components[0].metrics && component.host_components[0].metrics.yarn) {
          var root = component.host_components[0].metrics.yarn.Queue.root;
          component.queue = JSON.stringify({
            'root': self.parseObject(root)
          });
        }
        // extend config
        finalConfig = jQuery.extend(finalConfig, yarnConfig);
      }
    });
    // Map
    var finalJson = this.parseIt(item, finalConfig);
    finalJson.quick_links = [ 23, 24, 25, 26 ];
    return finalJson;
  },

  parseObject: function(obj) {
    var res = {};
    for (var p in obj) {
      if (obj.hasOwnProperty(p)) {
        if (obj[p] instanceof Object) {
          res[p] = this.parseObject(obj[p]);
        }
      }
    }
    return res;
  },

  mapreduce2Mapper: function (item) {
    var result = [];
    var finalConfig = jQuery.extend({}, this.config);
    // Change the JSON so that it is easy to map
    var mapReduce2Config = this.mapReduce2Config;
    item.components.forEach(function (component) {
      if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "HISTORYSERVER") {
        item.jobHistoryServerComponent = component;
        finalConfig = jQuery.extend(finalConfig, mapReduce2Config);
      }
    });
    // Map
    var finalJson = this.parseIt(item, finalConfig);
    finalJson.quick_links = [27, 28, 29, 30];

    return finalJson;
  },
  hbaseMapper: function (item) {
    // Change the JSON so that it is easy to map
    var finalConfig = jQuery.extend({}, this.config);
    var hbaseConfig = this.hbaseConfig;
    item.components.forEach(function (component) {
      if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == "HBASE_MASTER") {
        item.masterComponent = component;
        finalConfig = jQuery.extend(finalConfig, hbaseConfig);
        var regionsArray = App.parseJSON(component.ServiceComponentInfo.RegionsInTransition);
        //regions_in_transition can have various type of value: null, array or int
        if (Array.isArray(regionsArray)) {
          item.regions_in_transition = regionsArray.length;
        } else {
          item.regions_in_transition = regionsArray == null ? 0 : regionsArray;
        }
      }
    });
    // Map
    var finalJson = this.parseIt(item, finalConfig);
    finalJson.average_load = parseFloat(finalJson.average_load).toFixed(2);
    finalJson.quick_links = [13, 14, 15, 16, 17, 18];
    return finalJson;
  },

  /**
   * Flume is different from other services, in that the important
   * data is in customizable channels. Hence we directly transfer
   * data into the JSON object.
   */
  flumeMapper: function (item) {
    var self = this;
    var finalConfig = jQuery.extend({}, this.config);
    var flumeConfig = this.flumeConfig;
    finalConfig = jQuery.extend(finalConfig, flumeConfig);
    var finalJson = self.parseIt(item, finalConfig);
    var flumeHandlers = item.components.findProperty('ServiceComponentInfo.component_name', "FLUME_HANDLER");
    flumeHandlers = flumeHandlers ? flumeHandlers.host_components : [];
    finalJson.agents = [];
    finalJson.agentJsons = [];
    flumeHandlers.forEach(function(flumeHandler){
      var hostName = flumeHandler.HostRoles.host_name;
      flumeHandler.processes.forEach(function(process){
        var agentJson = self.parseIt(process, self.flumeAgentConfig);
        var agentId = agentJson.name + "-" + hostName;
        finalJson.agents.push(agentId);
        agentJson.id = agentId;
        finalJson.agentJsons.push(agentJson);
      });
    });
    return finalJson;
  },

  /**
   * Storm mapper
   */
  stormMapper: function(item) {
    var finalConfig = jQuery.extend({}, this.config);
    var stormConfig = this.stormConfig;
    var metricsInfoComponent = /^2.1/.test(App.get('currentStackVersionNumber')) ? 'STORM_REST_API' : 'STORM_UI_SERVER';
    var metricsPath = {
      STORM_REST_API: 'metrics.api.cluster.summary',
      STORM_UI_SERVER: 'metrics.api.v1.cluster.summary'
    }[metricsInfoComponent];

    item.components.forEach(function(component) {
      if (component.ServiceComponentInfo && component.ServiceComponentInfo.component_name == metricsInfoComponent) {
        if (Em.get(component, metricsPath)) {
          item.restApiComponent = App.keysDottedToCamelCase(Em.get(component, metricsPath));
          if (metricsInfoComponent == 'STORM_UI_SERVER') {
            item.restApiComponent.topologies = Em.get(component, 'metrics.api.v1.topology.summary.length');
          } else {
            item.restApiComponent.nimbusUptime = dateUtils.timingFormat(item.restApiComponent.nimbusUptime * 1000);
          }
        }
        finalConfig = jQuery.extend({}, finalConfig, stormConfig);
      }
    });
    return this.parseIt(item, finalConfig);
  }
});

});

require.register("mappers/stack_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');

App.stackMapper = App.QuickDataMapper.create({
  modelStack: App.Stack,
  modelOS: App.OperatingSystem,
  modelRepo: App.Repository,

  configStack: {
    id: 'id',
    stack_name: 'stack_name',
    stack_version: 'stack_version',
    active: 'active',
    parent_stack_version: 'parent_stack_version',
    min_upgrade_version: 'min_upgrade_version',
    is_selected: 'is_selected',
    config_types: 'config_types',
    operating_systems_key: 'operating_systems',
    operating_systems_type: 'array',
    operating_systems: {
      item: 'id'
    }
  },

  configOS: {
    id: 'id',
    os_type: 'os_type',
    stack_name: 'stack_name',
    stack_version: 'stack_version',
    stack_id: 'stack_id',
    repositories_key: 'repositories',
    repositories_type: 'array',
    repositories: {
      item: 'id'
    }
  },

  configRepository: {
    id: 'id',
    base_url: 'base_url',
    default_base_url: 'default_base_url',
    latest_base_url: 'latest_base_url',
    mirrors_list: 'mirrors_list',
    os_type: 'os_type',
    repo_id: 'repo_id',
    repo_name: 'repo_name',
    stack_name: 'stack_name',
    stack_version: 'stack_version',
    operating_system_id: 'os_id'
  },

  map: function(json) {
    var modelStack = this.get('modelStack');
    var modelOS = this.get('modelOS');
    var modelRepo = this.get('modelRepo');
    var resultStack = [];
    var resultOS = [];
    var resultRepo = [];

    var stackVersions = json.items.filterProperty('Versions.active');
    stackVersions.sortProperty('Versions.stack_version').reverse().forEach(function(item) {
      var stack = item.Versions;
      var operatingSystemsArray = [];

      stack.id = stack.stack_name + "-" + stack.stack_version;

      item.operating_systems.forEach(function(ops) {
        var operatingSystems = ops.OperatingSystems;

        var repositoriesArray = [];
        ops.repositories.forEach(function(repo) {
          repo.Repositories.id = repo.Repositories.os_type + "-" + repo.Repositories.repo_id;
          repo.Repositories.os_id = repo.Repositories.stack_name + "-" + repo.Repositories.stack_version + "-" + repo.Repositories.os_type;
          resultRepo.push(this.parseIt(repo.Repositories, this.get('configRepository')));
          repositoriesArray.pushObject(repo.Repositories);
        }, this);


        operatingSystems.id = operatingSystems.stack_name + "-" + operatingSystems.stack_version + "-" + operatingSystems.os_type;
        operatingSystems.stack_id = operatingSystems.stack_name + "-" + operatingSystems.stack_version;
        operatingSystems.repositories = repositoriesArray;
        resultOS.push(this.parseIt(operatingSystems, this.get('configOS')));
        operatingSystemsArray.pushObject(operatingSystems);

      }, this);


      stack.operating_systems = operatingSystemsArray;
      resultStack.push(this.parseIt(stack, this.get('configStack')));

    }, this);

    App.store.commit();
    App.store.loadMany(modelRepo, resultRepo);
    App.store.loadMany(modelOS, resultOS);
    App.store.loadMany(modelStack, resultStack);
  }
});
});

require.register("mappers/stack_service_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');

App.stackServiceMapper = App.QuickDataMapper.create({
  model: App.StackService,
  component_model: App.StackServiceComponent,

  config: {
    id: 'service_name',
    stack_id: 'stack_id',
    service_name: 'service_name',
    display_name: 'display_name',
    config_types: 'config_types',
    comments: 'comments',
    service_version: 'service_version',
    stack_name: 'stack_name',
    stack_version: 'stack_version',
    is_selected: 'is_selected',
    is_installed: 'is_installed',
    is_installable: 'is_installable',
    required_services: 'required_services',
    service_check_supported: 'service_check_supported',
    service_components_key: 'service_components',
    service_components_type: 'array',
    service_components: {
      item: 'id'
    }
  },

  component_config: {
    id: 'component_name',
    component_name: 'component_name',
    display_name: 'display_name',
    cardinality: 'cardinality',
    custom_commands: 'custom_commands',
    service_name: 'service_name',
    component_category: 'component_category',
    is_master: 'is_master',
    is_client: 'is_client',
    stack_name: 'stack_name',
    stack_version: 'stack_version',
    stack_service_id: 'service_name',
    dependencies_key: 'dependencies',
    dependencies_type: 'array',
    dependencies: {
      item: 'Dependencies'
    }
  },

  mapStackServices: function(json) {
    App.set('isStackServicesLoaded',false);
    this.clearStackModels();
    App.resetDsStoreTypeMap(App.StackServiceComponent);
    App.resetDsStoreTypeMap(App.StackService);
    this.map(json);
    App.set('isStackServicesLoaded',true);
  },

  map: function (json) {
    var model = this.get('model');
    var result = [];
    var stackServiceComponents = [];
    var nonInstallableServices = ['KERBEROS'];
    this.rearrangeServicesForDisplayOrder(json.items, App.StackService.displayOrder);
    json.items.forEach(function (item) {
      var stackService = item.StackServices;
      var serviceComponents = [];
      item.components.forEach(function (serviceComponent) {
        var dependencies = serviceComponent.dependencies.map(function (dependecy) {
          return { Dependencies: App.keysUnderscoreToCamelCase(App.permit(dependecy.Dependencies, ['component_name', 'scope'])) };
        });
        serviceComponent.StackServiceComponents.id = serviceComponent.StackServiceComponents.component_name;
        serviceComponent.StackServiceComponents.dependencies = dependencies;
        serviceComponents.push(serviceComponent.StackServiceComponents);
        var parsedResult = this.parseIt(serviceComponent.StackServiceComponents, this.get('component_config'));
        if (parsedResult.id == 'MYSQL_SERVER') {
          parsedResult.custom_commands = parsedResult.custom_commands.without('CLEAN');
        }
        stackServiceComponents.push(parsedResult);
      }, this);
      stackService.stack_id = stackService.stack_name + '-' + stackService.stack_version;
      stackService.service_components = serviceComponents;
      // @todo: replace with server response value after API implementation
      if (nonInstallableServices.contains(stackService.service_name)) {
        stackService.is_installable = false;
        stackService.is_selected = false;
      }
      result.push(this.parseIt(stackService, this.get('config')));
    }, this);
    App.store.loadMany(this.get('component_model'), stackServiceComponents);
    App.store.loadMany(model, result);
  },

  /**
   * Clean store from already loaded data.
   **/
  clearStackModels: function () {
    var models = [App.StackServiceComponent, App.StackService];
    models.forEach(function (model) {
      var records = App.get('store').findAll(model).filterProperty('id');
      records.forEach(function (rec) {
        Ember.run(this, function () {
          rec.deleteRecord();
          App.store.commit();
        });
      }, this);
    }, this);
  },

  rearrangeServicesForDisplayOrder: function (array, displayOrderArray) {
    return array.sort(function (a, b) {
      var aValue = displayOrderArray.indexOf(a.StackServices.service_name) != -1 ? displayOrderArray.indexOf(a.StackServices.service_name) : array.length;
      var bValue = displayOrderArray.indexOf(b.StackServices.service_name) != -1 ? displayOrderArray.indexOf(b.StackServices.service_name) : array.length;
      return aValue - bValue;
    });
  }
});


});

require.register("mappers/stack_version_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');
var stringUtils = require('utils/string_utils');

App.stackVersionMapper = App.QuickDataMapper.create({
  modelStackVerion: App.StackVersion,

  modelStack: {
    "id": "id",
    "cluster_name": "cluster_name",
    "stack": "stack",
    "version": "version",
    "repository_version_id": "repository_version_id",
    "state": "state",
    "not_installed_hosts": "host_states.NOT_INSTALLED",
    "installing_hosts": "host_states.INSTALLING",
    "installed_hosts": "host_states.INSTALLED",
    "install_failed_hosts": "host_states.INSTALL_FAILED",
    "out_of_sync_hosts": "host_states.OUT_OF_SYNC",
    "upgrading_hosts": "host_states.UPGRADING",
    "upgraded_hosts": "host_states.UPGRADED",
    "upgrade_failed_hosts": "host_states.UPGRADE_FAILED",
    "current_hosts": "host_states.CURRENT"
  },

  map: function (json) {
    var modelStackVerion = this.get('modelStackVerion');
    var resultStack = [];

    if (json && json.items) {
      json.items.sort(function (a, b) {
        return stringUtils.compareVersions(a.repository_versions[0].RepositoryVersions.repository_version, b.repository_versions[0].RepositoryVersions.repository_version);
      });
      json.items.forEach(function (item) {
        var stack = item.ClusterStackVersions;
        stack.repository_version_id = item.ClusterStackVersions.repository_version;
        /**
         * this property contains array of hosts on which repoversion wasn't installed
         * possible states:
         * <code>INSTALLING<code>
         * <code>INSTALL_FAILED<code>
         * <code>OUT_OF_SYNC<code>
         */
        stack.host_states.NOT_INSTALLED = item.ClusterStackVersions.host_states.INSTALLING
          .concat(item.ClusterStackVersions.host_states.INSTALL_FAILED)
          .concat(item.ClusterStackVersions.host_states.OUT_OF_SYNC);

        /**
         * this property contains array of hosts on which repoversion was installed
         * but state of repoveriosn for this hosts can be any postinstalled state
         * possible states:
         * <code>INSTALLED<code>
         * <code>UPGRADING<code>
         * <code>UPGRADED<code>
         * <code>UPGRADE_FAILED<code>
         */
        stack.host_states.INSTALLED = item.ClusterStackVersions.host_states.INSTALLED
          .concat(item.ClusterStackVersions.host_states.UPGRADING)
          .concat(item.ClusterStackVersions.host_states.UPGRADED)
          .concat(item.ClusterStackVersions.host_states.UPGRADE_FAILED);

        if (item.repository_versions && item.repository_versions[0]) {
          item.repository_versions[0].RepositoryVersions.stackVersionId = item.ClusterStackVersions.id;
          App.repoVersionMapper.map({"items": item.repository_versions }, true);
        }
        resultStack.push(this.parseIt(stack, this.get('modelStack')));
      }, this);
    }
    App.store.commit();
    App.store.loadMany(modelStackVerion, resultStack);
  }
});

});

require.register("mappers/target_cluster_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


App.targetClusterMapper = App.QuickDataMapper.create({
  model: App.TargetCluster,
  config: {
    id: 'name',
    name: 'name',
    execute: 'execute',
    workflow: 'workflow',
    write: 'write',
    readonly: 'readonly',
    staging: 'staging',
    working: 'working',
    temp: 'temp'
  },

  map: function (json) {
    var model = this.get('model');
    if (!model) {
      return;
    }

    if (json.items) {
      var result = [];
      var clustersToDelete = model.find().mapProperty('name');
      json.items.forEach(function (item) {
        result.push(this.parseIt(item, this.config));
        clustersToDelete = clustersToDelete.without(item.name);
      }, this);
      clustersToDelete.forEach(function (name) {
        this.deleteRecord(model.find().findProperty('name', name));
      }, this);
      App.store.loadMany(model, result);
    }
  }
});

});

require.register("mappers/users_mapper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.usersMapper = App.QuickDataMapper.create({
  model : App.User,
  config : {
    id : 'Users.user_name',
    user_name : 'Users.user_name',
    is_ldap: 'Users.ldap_user',
    admin: 'Users.admin',
    operator: 'Users.operator',
    permissions: 'permissions'
  },
  map: function (json) {
    var self = this;
    json.items.forEach(function (item) {
      var result= [];
      if(!App.User.find().someProperty("userName", item.Users.user_name)) {
        item.permissions = [];
        var privileges = item.privileges || [];
        if (privileges.length) {
          item.permissions = privileges.mapProperty('PrivilegeInfo.permission_name');
        }
        item.Users.admin = self.isAdmin(item.permissions);
        item.Users.operator = self.isOperator(item.permissions);
        result.push(self.parseIt(item, self.config));
        App.store.loadMany(self.get('model'), result);
      }
    });
  },

  /**
   * Check if user is admin.
   * @param {Array} permissionList
   * @return {Boolean}
   **/
  isAdmin: function(permissionList) {
    //TODO: Separate cluster operator from admin
    return permissionList.indexOf('AMBARI.ADMIN') > -1 || permissionList.indexOf('CLUSTER.OPERATE') > -1;
  },

  /**
   * Check if user is operator.
   * @param {Array} permissionList
   * @return {Boolean}
   **/
  isOperator: function(permissionList) {
    return permissionList.indexOf('CLUSTER.OPERATE') > -1 && !(permissionList.indexOf('AMBARI.ADMIN') > -1);
  }
});

});

require.register("messages", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

Em.I18n.translations = {

  'app.name':'Ambari',
  'app.name.subtitle':'Ambari - {0}',
  'app.name.subtitle.experimental':'Ambari Experimental',
  'app.reloadPopup.link': 'Reload Page',
  'app.reloadPopup.text': 'Trying to connect to server...',
  'app.reloadPopup.header': 'Reload Page',

  'app.loadingPlaceholder': 'Loading...',
  'app.versionMismatchAlert.title': 'Ambari Server / Web Client Version Mismatch',
  'app.versionMismatchAlert.body': 'Ambari Server and Web Client versions do not match:<br> ' +
    '<br>Ambari Server: <strong>{0}</strong>' +
    '<br>Ambari Web Client: <strong>{1}</strong><br>' +
    '<br>This typically happens after upgrading Ambari due to Ambari Web Client code cached in the browser.' +
    '<br>Perform a hard browser cache refresh, typically \'Ctrl+Shift+R\' (\'Cmd+Shift+R\' on OS X), to see if this message disappears.' +
    '<br>If you keep seeing this message, clear the browser cache completely and hit this URL again.' +
    '<br>You must resolve this error in order to continue.',
  'app.signout':'Sign out',
  'app.settings':'Settings',
  'app.manageAmbari': 'Manage Ambari',
  'app.aboutAmbari':'About',
  'app.settings.notshowBgOperationsPopup': 'Do not show the Background Operations dialog when starting an operation',
  'app.settings.notShowBgOperations': 'Do not show this dialog again when starting a background operation',

  'app.aboutAmbari.getInvolved': 'Get involved!',
  'app.aboutAmbari.version': 'Version',
  'app.aboutAmbari.licensed': 'Licensed under the Apache License, Version 2.0',

  'apply':'apply',
  'and':'and',
  'none':'none',
  'all':'all',
  'minimum':'minimum',
  'from':'From',
  'to':'To',
  'ok':'OK',
  'as':'as',
  'on':'on',
  'any': 'Any',
  'more':'more',
  'yes':'Yes',
  'no':'No',
  'add': 'Add',
  'op': 'op',
  'ops': 'ops',


  'common.access':'Access',
  'common.learnMore':'Learn more',
  'common.showDetails':'Show Details',
  'common.back':'Back',
  'common.prev':'Prev',
  'common.next':'Next',
  'common.host':'Host',
  'common.hosts':'Hosts',
  'common.services':'Services',
  'common.group':'Group',
  'common.groups':'Groups',
  'common.progress':'Progress',
  'common.status':'Status',
  'common.action':'Action',
  'common.remove':'Remove',
  'common.retry':'Retry',
  'common.skip':'Skip',
  'common.filter': 'Filter',
  'common.rollBack':'Rollback',
  'common.show':'Show',
  'common.hide':'Hide',
  'common.cancel':'Cancel',
  'common.apply':'Apply',
  'common.done':'Done',
  'common.failed':'Failed',
  'common.service': 'Service',
  'common.version':'Version',
  'common.downgrade':'Downgrade',
  'common.description':'Description',
  'common.default':'Default',
  'common.client':'Client',
  'common.zookeeper':'ZooKeeper',
  'common.hbase':'HBase',
  'common.regionServer':'RegionServer',
  'common.taskTracker':'TaskTracker',
  'common.dataNode':'DataNode',
  'common.more': 'More...',
  'common.print':'Print',
  'common.deploy':'Deploy',
  'common.message':'Message',
  'common.tasks':'Tasks',
  'common.open':'Open',
  'common.copy':'Copy',
  'common.complete':'Complete',
  'common.metrics':'Metrics',
  'common.timeRange':'Time Range',
  'common.name':'Name',
  'common.key':'Key',
  'common.value':'Value',
  'common.ipAddress':'IP Address',
  'common.cpu':'CPU',
  'common.cores': 'Cores (CPU)',
  'common.ram':'RAM',
  'common.disk':'Disk',
  'common.diskUsage':'Disk Usage',
  'common.loadAvg':'Load Avg',
  'common.components':'Components',
  'common.component':'Component',
  'common.quickLinks':'Quick Links',
  'common.save':'Save',
  'common.saveAnyway':'Save Anyway',
  'common.servers':'Servers',
  'common.clients':'Clients',
  'common.user': 'User',
  'common.users': 'Users',
  'common.issues': 'Issues',
  'common.os':'OS',
  'common.oss':'OSs',
  'common.memory':'Memory',
  'common.maximum':'Maximum',
  'common.started':'Started',
  'common.start':'Start',
  'common.stop':'Stop',
  'common.pause':'Pause',
  'common.decommission':'Decommission',
  'common.recommission':'Recommission',
  'common.failure': 'Failure',
  'common.type': 'Type',
  'common.close': 'Close',
  'common.warning': 'Warning',
  'common.information': 'Information',
  'common.all':'All',
  'common.success': 'Success',
  'common.fail':'Fail',
  'common.error': 'Error',
  'common.loading': 'Loading',
  'common.search': 'Search',
  'common.confirm': 'Confirm',
  'common.upgrade': 'Upgrade',
  'common.reUpgrade': 'Retry Upgrade',
  'common.security':'Security',
  'common.kerberos':'Kerberos',
  'common.cluster':'Cluster',
  'common.repositories':'Repositories',
  'common.stack.versions':'Stack Versions',
  'common.versions':'Versions',
  'common.serviceAccounts': 'Service Accounts',
  'common.add': 'Add',
  'common.edit': 'Edit',
  'common.delete': 'Delete',
  'common.duplicate': 'Duplicate',
  'common.empty': 'Empty',
  'common.override':'Override',
  'common.undo':'Undo',
  'common.details':'Details',
  'common.stats':'Stats',
  'common.abort': 'Abort',
  'common.misc': 'Misc',
  'common.userSettings': 'User Settings',
  'common.aboutAmbari': 'About',
  'common.notAvailable': 'Not Available',
  'common.na': 'n/a',
  'common.operations': 'Operations',
  'common.startTime': 'Start Time',
  'common.duration': 'Duration',
  'common.reinstall': 'Re-Install',
  'common.revert': 'Revert',
  'common.errorPopup.header': 'An error has been encountered',
  'common.use': 'Use',
  'common.stacks': 'Stacks',
  'common.stack': 'Stack',
  'common.reset': 'Reset',
  'common.path': 'Path',
  'common.package': 'Package',
  'common.proceed': 'Proceed',
  'common.proceedAnyway': 'Proceed Anyway',
  'common.exitAnyway': 'Exit Anyway',
  'common.process': 'Process',
  'common.property': 'Property',
  'common.installed': 'Installed',
  'common.persist.error' : 'Error in persisting web client state at ambari server:',
  'common.update.error' : 'Error in retrieving web client state from ambari server',
  'common.tags': 'Tags',
  'common.important': '<strong>Important:</strong>',
  'common.allServices':'All Services',
  'common.move':'Move',
  'common.change': 'Change',
  'common.overrides': 'Overrides',
  'common.properties': 'properties',
  'common.conf.group': 'Configuration Group',
  'common.ignore': 'Ignore',
  'common.restart': 'Restart',
  'common.discard': 'Discard',
  'common.actions': 'Actions',
  'common.maintenance': 'Maintenance',
  'common.passive_state': 'Maintenance Mode',
  'common.selected': 'Selected',
  'common.password': 'Password',
  'common.url': 'URL',
  'common.advanced': 'Advanced',
  'common.download': 'Download',
  'common.current': 'Current',
  'common.additional': 'Additional',
  'common.time.start': 'Start Time',
  'common.time.end': 'End Time',
  'common.hostLog.popup.logDir.path':'/var/lib/ambari-agent/data/',   // TODO, this hardcoded path needs to be removed.
  'common.hostLog.popup.outputLog.value': 'output-{0}.txt',
  'common.hostLog.popup.errorLog.value': 'errors-{0}.txt',
  'common.maintenance.task': ' Toggle Maintenance Mode',
  'common.installRepo.task': ' Install Packages',
  'common.used': 'used',
  'common.free': 'free',
  'common.type.string': 'string',
  'common.type.number': 'number',
  'common.author': 'Author',
  'common.notes': 'Notes',
  'common.view': 'View',
  'common.compare': 'Compare',
  'common.latest': 'Latest',
  'common.custom': 'Custom',
  'common.continueAnyway': 'Continue Anyway',
  'common.property.undefined': "Undefined",
  'common.summary': "Summary",
  'common.configs': "Configs",
  'common.configuration': "Configuration",
  'common.unknown': "Unknown",
  'common.install': "Install",
  'common.alertDefinition': "Alert Definition",
  'common.prerequisites': 'Prerequisites',
  'common.finalize': "Finalize",
  'common.severity': "Severity",
  'common.dismiss': "Dismiss",
  'common.stdout': "stdout",
  'common.stderr': "stderr",

  'models.alert_instance.tiggered.verbose': "Occured on {0} <br> Checked on {1}",
  'models.alert_definition.triggered.verbose': "Occured on {0}",
  'models.alert_definition.triggered.checked': "Status Changed: {0}\nLast Checked: {1}",

  'passiveState.turnOn':'Turn On Maintenance Mode',
  'passiveState.turnOff':'Turn Off Maintenance Mode',
  'passiveState.turnOnFor':'Turn On Maintenance Mode for {0}',
  'passiveState.turnOffFor':'Turn Off Maintenance Mode for {0}',

  'requestInfo.installComponents':'Install Components',
  'requestInfo.installKerbeorosComponents':'Install Kerberos Components',
  'requestInfo.installServices':'Install Services',
  'requestInfo.kerberosService': 'Install Kerberos Service',
  'requestInfo.startServices':'Start Services',
  'requestInfo.startAddedServices':'Start Added Services',
  'requestInfo.stopAllServices':'Stop All Services',
  'requestInfo.startAllServices':'Start All Services',
  'requestInfo.startHostComponent':'Start',
  'requestInfo.startHostComponent.datanode':'Start DataNode',
  'requestInfo.startHostComponent.nodemanager':'Start NodeManager',
  'requestInfo.startHostComponent.hbase_regionserver':'Start RegionServer',
  'requestInfo.startHostComponents':'Start Components',
  'requestInfo.upgradeHostComponent':'Upgrade',
  'requestInfo.stopHostComponent':'Stop',
  'requestInfo.installHostComponent':'Install',
  'requestInfo.installNewHostComponent':'Install',
  'requestInfo.refreshComponentConfigs':'Refresh {0} configs',
  'requestInfo.stop':'Stop {0}',
  'requestInfo.start':'Start {0}',
  'requestInfo.unspecified':'Request name not specified',
  'requestInfo.kerberizeCluster': 'Kerberize Cluster',

  'hostPopup.noServicesToShow':'No services to show',
  'hostPopup.noHostsToShow':'No hosts to show',
  'hostPopup.noTasksToShow':'No tasks to show',
  'hostPopup.status.category.all':'All ({0})',
  'hostPopup.status.category.pending':'Pending ({0})',
  'hostPopup.status.category.inProgress':'In Progress ({0})',
  'hostPopup.status.category.failed':'Failed ({0})',
  'hostPopup.status.category.success':'Success ({0})',
  'hostPopup.status.category.aborted':'Aborted ({0})',
  'hostPopup.status.category.timedout':'Timedout ({0})',
  'hostPopup.header.postFix':' Background Operation{0} Running',
  'hostPopup.serviceInfo.showMore':'Show more...',
  'hostPopup.bgop.abortRequest.title': 'Abort operation',
  'hostPopup.bgop.abortRequest.confirmation.body': 'Are you sure you want to abort \'{0}\' operation?',
  'hostPopup.bgop.abortRequest.reason': 'Aborted by user',
  'hostPopup.bgop.abortRequest.modal.header': 'Abort request sent',
  'hostPopup.bgop.abortRequest.modal.body': 'The abort request for <strong>{0}</strong> has been sent to the server. Note: some tasks that are already running may have time to complete or fail before the abort request is applied.',
  'hostPopup.bgop.sourceRequestSchedule.running': 'Future operations of this batch request can be aborted',
  'hostPopup.bgop.sourceRequestSchedule.aborted': 'Future operations of this batch request have been aborted',
  'hostPopup.bgop.abort.rollingRestart': 'Abort Rolling Restart',
  'hostPopup.warning.alertsTimeOut': 'Maintenance Mode has been turned {0}. It may take a few minutes for the alerts to be {1}.',
  'hostPopup.recommendation.beforeDecommission': '{0} Maintenance Mode is pre required for decommissioning.',

  'question.sure':'Are you sure?',

  'popup.highlight':'click to highlight',
  'popup.confirmation.commonHeader':'Confirmation',
  'popup.prompt.commonHeader':'Prompt',
  'popup.confirmationFeedBack.sending':'Sending...',
  'popup.confirmationFeedBack.query.fail':'Request failed',

  'popup.clusterCheck.failedOn': 'Failed on: ',
  'popup.clusterCheck.reason': 'Reason: ',
  'popup.clusterCheck.Upgrade.header': 'Upgrade to {0}',
  'popup.clusterCheck.Upgrade.title': 'Upgrade Requirements Not Met',
  'popup.clusterCheck.Upgrade.alert': 'You must meet the following requirements before you can proceed with rolling upgrade.',
  'popup.clusterCheck.Security.header': 'Enable Security',
  'popup.clusterCheck.Security.title': 'Security Requirements Not Met',
  'popup.clusterCheck.Security.alert': 'You must meet the following requirements before you can enable security.',

  'popup.invalid.KDC.header': 'Admin session expiration error',
  'popup.invalid.KDC.msg': ' Please enter admin principal and password.',
  'popup.invalid.KDC.admin.principal': 'Admin principal',
  'popup.invalid.KDC.admin.password': 'Admin password',

  'login.header':'Sign in',
  'login.username':'Username',
  'login.loginButton':'Sign in',
  'login.error.bad.credentials':'Unable to sign in. Invalid username/password combination.',
  'login.error.disabled':'Unable to sign in. Invalid username/password combination.',
  'login.error.bad.connection':'Unable to connect to Ambari Server. Confirm Ambari Server is running and you can reach Ambari Server from this machine.',

  'titlebar.alerts.noAlerts': 'No Alerts',

  'graphs.noData.title': 'No Data',
  'graphs.noData.message': 'There was no data available. Possible reasons include inaccessible/unsupported metrics service.',
  'graphs.noDataAtTime.message': 'No available data for the time period.',
  'graphs.error.title': 'Error',
  'graphs.error.message': 'There was a problem getting data for the chart ({0}: {1})',
  'graphs.timeRange.hour': 'Last 1 hour',
  'graphs.timeRange.twoHours': 'Last 2 hours',
  'graphs.timeRange.fourHours': 'Last 4 hours',
  'graphs.timeRange.twelveHours': 'Last 12 hours',
  'graphs.timeRange.day': 'Last 24 hours',
  'graphs.timeRange.week': 'Last 1 week',
  'graphs.timeRange.month': 'Last 1 month',
  'graphs.timeRange.year': 'Last 1 year',

  'users.userName.validationFail': 'Only lowercase letters and numbers are recommended; must start with a letter',
  'host.spacesValidation': 'Cannot contain whitespace',
  'host.trimspacesValidation': 'Cannot contain leading or trailing whitespace',

  'services.hdfs.rebalance.title' : 'HDFS Rebalance',
  'services.ganglia.description':'Ganglia Metrics Collection system',
  'services.hdfs.description':'Apache Hadoop Distributed File System',
  'services.glusterfs.description':'Apache Hadoop Compatible File System (must be installed manually)',
  'services.sqoop.description':'Tool for transferring bulk data between Apache Hadoop and structured data stores such as relational databases',
  'services.pig.description':'Scripting platform for analyzing large datasets',
  'services.hive.description':'Data warehouse system for ad-hoc queries & analysis of large datasets and table & storage management service',
  'services.oozie.description':'System for workflow coordination and execution of Apache Hadoop jobs',
  'services.zookeeper.description':'Centralized service which provides highly reliable distributed coordination',
  'services.hbase.description':'Non-relational distributed database and centralized service for configuration management & synchronization',
  'services.hive.databaseComponent':'Database Server',
  'services.mapreduce2.description':'Apache Hadoop NextGen MapReduce (client libraries)',
  'services.yarn.description':'Apache Hadoop NextGen MapReduce (YARN)',
  'services.tez.description':'Tez is the next generation Hadoop Query Processing framework written on top of YARN',
  'services.falcon.description': 'Falcon mirroring engine',
  'services.storm.description': 'Apache Hadoop Stream processing framework',
  'services.storm.slots.metrics.title': 'Number of slots',
  'services.storm.slots.metrics.free': 'Free slots',
  'services.storm.slots.metrics.total': 'Total slots',
  'services.storm.slots.metrics.used': 'Used slots',
  'services.storm.executors.metrics.title': 'Number of executors',
  'services.storm.executors.metrics.total': 'Total executors',
  'services.storm.topology.metrics.title': 'Number of topologies',
  'services.storm.topology.metrics.total': 'Total topologies',
  'services.storm.tasks.metrics.title': 'Number of tasks',
  'services.storm.tasks.metrics.total': 'Total tasks',
  'services.storm.configs.range-plugin-enable.dialog.title': 'Enable Ranger for STORM',
  'services.storm.configs.range-plugin-enable.dialog.message': 'Enabling Ranger plugin for STORM is effective only on a secured cluster.',


  'services.alerts.head':'You have {0} critical alert notification(s).',
  'services.alerts.OK.timePrefixShort': 'OK',
  'services.alerts.WARN.timePrefixShort': 'WARN',
  'services.alerts.CRIT.timePrefixShort': 'CRIT',
  'services.alerts.MAINT.timePrefixShort': 'MAINT',
  'services.alerts.UNKNOWN.timePrefixShort': 'UNKNOWN',
  'services.alerts.OK.timePrefix': 'OK for {0}',
  'services.alerts.WARN.timePrefix': 'WARN for {0}',
  'services.alerts.CRIT.timePrefix': 'CRIT for {0}',
  'services.alerts.MAINT.timePrefix': 'MAINT for {0}',
  'services.alerts.UNKNOWN.timePrefix': 'UNKNOWN for {0}',
  'services.alerts.lastCheck': 'Last Checked {0}',
  'services.alerts.brLastCheck': '<br>Last Checked {0}',
  'services.alerts.occurredOn': 'Occurred on {0}, {1}',
  'services.alerts.goToService': 'Go to Service',

  'services.summary.selectHostForComponent': 'Select the host to add {0} component',
  'services.summary.allHostsAlreadyRunComponent': 'All hosts already running {0} component',

  'topnav.logo.href':'/#/main/dashboard',
  'topnav.help.href':'https://cwiki.apache.org/confluence/display/AMBARI/Ambari',

  'installer.header':'Cluster Install Wizard',
  'installer.navigation.warning.header':'Navigation Warning',

  'installer.noHostsAssigned':'No host assigned',
  'installer.slaveComponentHosts.selectHosts':'select hosts for this group',
  'installer.slaveComponentHostsPopup.header':'Select which hosts should belong to which {0} group',

  'installer.controls.slaveComponentGroups':' Groups',
  'installer.controls.serviceConfigPopover.title':'{0}<br><small>{1}</small>',
  'installer.controls.checkConnection.popover':'This action will check accessibility of {0} host and port from Ambari Server host',
  'installer.controls.serviceConfigMultipleHosts.other':'1 other',
  'installer.controls.serviceConfigMultipleHosts.others':'{0} others',
  'installer.controls.serviceConfigMasterHosts.header':'{0} Hosts',
  'installer.controls.addSlaveComponentGroupButton.title':'Add a {0} Group',
  'installer.controls.addSlaveComponentGroupButton.content':'If you need different settings on certain {0}s, you can add a {1} group.<br>All {2}s within the same group will have the same set of settings.  You can create multiple groups.',
  'installer.controls.slaveComponentChangeGroupName.error':'group with this name already exist',

  'installer.step0.header':'Get Started',
  'installer.step0.body.header':'Get Started',
  'installer.step0.body':'This wizard will walk you through the cluster installation process.  First, start by naming your new cluster.',
  'installer.step0.clusterName':'Name your cluster',
  'installer.step0.clusterName.tooltip.title':'Cluster Name',
  'installer.step0.clusterName.tooltip.content':'Enter a unique cluster name.',
  'installer.step0.clusterName.error.required':'Cluster Name is required',
  'installer.step0.clusterName.error.whitespace':'Cluster Name cannot contain whitespace',
  'installer.step0.clusterName.error.specialChar':'Cluster Name cannot contain special characters',

  'installer.step1.header':'Select Stack',
  'installer.step1.body':'Please select the service stack that you want to use to install your Hadoop cluster.',
  'installer.step1.advancedRepo.title':'Advanced Repository Options',
  'installer.step1.advancedRepo.message':'Customize the repository Base URLs for downloading the Stack software packages. If your hosts do not have access to the internet, you will have to create a local mirror of the Stack repository that is accessible by all hosts and use those Base URLs here.',
  'installer.step1.advancedRepo.importantMassage':'<b>Important:</b> When using local mirror repositories, you only need to provide Base URLs for the Operating System you are installing for your Stack. Uncheck all other repositories.',
  'installer.step1.advancedRepo.localRepo.error.modifyUrl':'Local repository URL must be modified',
  'installer.step1.advancedRepo.localRepo.error.noUrl':'Base URL required for a local repository',
  'installer.step1.advancedRepo.localRepo.column.baseUrl':'Base URL',
  'installer.step1.advancedRepo.localRepo.label.os':'Operating System',
  'installer.step1.advancedRepo.localRepo.label.baseUrl':'Repository Base URL',
  'installer.step1.advancedRepo.localRepo.label.stack':'Stack',
  'installer.step1.advancedRepo.skipValidation.tooltip':'<b>Warning:</b> This is for advanced users only. Use this option if you want to skip validation for Repository Base URLs.',
  'installer.step1.advancedRepo.skipValidation.message':'Skip Repository Base URL validation (Advanced)',
  'installer.step1.attentionNeeded':'<b>Attention:</b> Repository URLs are REQUIRED before you can proceed.',
  'installer.step1.invalidURLAttention': '<b>Attention:</b> Please make sure all repository URLs are valid before proceeding.',
  'installer.step1.checkAtLeastOneAttention': '<b>Attention:</b> Please check at least one repository.',
  'installer.step1.retryRepoUrls': 'Click <b>here</b> to retry.',

  'installer.step2.header':'Install Options',
  'installer.step2.body':'Enter the list of hosts to be included in the cluster and provide your SSH key.',
  'installer.step2.targetHosts':'Target Hosts',
  'installer.step2.targetHosts.info':'Enter a list of hosts using the Fully Qualified Domain Name (FQDN), one per line',
  'installer.step2.hostPattern.tooltip.title':'Pattern Expressions',
  'installer.step2.hostPattern.tooltip.content':'You can use pattern expressions to specify a number of target hosts. For example, to specify host01.domain thru host10.domain, enter host[01-10].domain in the target hosts textarea.',
  'installer.step2.hostName.error.required':'You must specify at least one host name',
  'installer.step2.hostName.error.already_installed':'All these hosts are already part of the cluster',
  'installer.step2.hostName.error.notRequired':'Host Names will be ignored if not using SSH to automatically configure hosts',
  'installer.step2.hostName.error.invalid':'Invalid Host Name(s)',
  'installer.step2.hostName.pattern.header':'Host name pattern expressions',
  'installer.step2.sshKey':'Host Registration Information',
  'installer.step2.sshKey.error.required':'SSH Private Key is required',
  'installer.step2.passphrase.error.match':'Passphrases do not match',
  'installer.step2.manualInstall.label':'Do not use SSH to automatically configure hosts ',
  'installer.step2.manualInstall.info':'By not using SSH to connect to the target hosts, you must manually install and' +
    ' start the Ambari Agent on each host in order for the wizard to perform the necessary configurations and' +
    ' software installs.',
  'installer.step2.advancedOption':'Advanced Options',
  'installer.step2.repoConf':'Software Repository Configuration File Path',
  'installer.step2.advancedOptions.header':'Advanced Options',
  'installer.step2.localRepo.label_use':'Use a',
  'installer.step2.localRepo.label_instead':'instead of downloading software packages from the Internet',
  'installer.step2.localRepo.error.required':'Local repository file path is required',
  'installer.step2.localRepo.tooltip.title':'Local Software Repository',
  'installer.step2.localRepo.tooltip.content': 'The cluster install requires access to the Internet to fetch software ' +
    'from a remote repository. In some cases, adequate bandwidth is not available and you want to prevent downloading ' +
    'packages from the remote repository over and over again. Other times, Internet access is not available from the ' +
    'hosts in your cluster. In these situations, you must set up a version of the repository that your machines can ' +
    'access locally and this is called a <b>Local Software Repository</b>',
  'installer.step2.javaHome.label' : 'Path to 64-bit JDK',
  'installer.step2.javaHome.tooltip.title' : 'JAVA_HOME',
  'installer.step2.javaHome.tooltip.content' : 'Path to 64-bit JAVA_HOME. /usr/jdk/jdk1.6.0_31 is the default used by Ambari. You can override this to a specific path that contains the JDK. <br/> Note: the path must be valid on <b>ALL</b> hosts in your cluster.',
  'installer.step2.javaHome.tooltip.placeholder' : '/usr/jdk/jdk1.6.0_31',
  'installer.step2.useSsh.provide' : 'Provide your',
  'installer.step2.useSsh.provide_id_rsa' : ' to automatically register hosts',
  'installer.step2.useSsh.tooltip.title':'SSH Private Key',
  'installer.step2.useSsh.tooltip.content':'The <b>SSH Private Key File</b> is used to connect to the target hosts in your cluster to install the Ambari Agent.',
  'installer.step2.manualInstall.perform':'Perform',
  'installer.step2.manualInstall.perform_on_hosts':'on hosts and do not use SSH',
  'installer.step2.manualInstall.tooltip.title':'manual registration',
  'installer.step2.manualInstall.tooltip.content':'Manually registering the Ambari Agent on each host eliminates the need for SSH and should be performed prior to continuing cluster installation.',
  'installer.step2.manualInstall.popup.header':'Before You Proceed',
  'installer.step2.manualInstall.popup.body':'You must install Ambari Agents on each host you want to manage before you proceed.',
  'installer.step2.warning.popup.body':'<p>The following hostnames are not valid FQDNs:</p><p> {0} </p><p>This may cause problems during installation. Do you want to continue?</p>',
  'installer.step2.orUse':'Or use',
  'installer.step2.registerAndConfirm':'Register and Confirm',
  'installer.step2.evaluateStep.installedHosts':'These hosts are already installed on the cluster and will be ignored:',
  'installer.step2.evaluateStep.continueConfirm':'Do you want to continue?',
  'installer.step2.evaluateStep.hostRegInProgress':'Host Registration is currently in progress.  Please try again later.',
  'installer.step2.sshUser':'SSH User Account',
  'installer.step2.sshUser.toolTip':'The user account used to install the Ambari Agent on the target host(s) via SSH. This user must be set up with passwordless SSH and sudo access on all the target host(s)',
  'installer.step2.sshUser.placeholder':'Enter user name',
  'installer.step2.sshUser.required':'User name is required',
  'installer.step2.agentUser':'Ambari Agent User Account',
  'installer.step2.agentUser.toolTip':'The user account used to run the Ambari Agent daemon on the target host(s). This user must be set up with passwordless sudo access on all the target host(s)',
  'installer.step2.bootStrap.error':'Errors were encountered while setting up Ambari Agents on the hosts.',
  'installer.step2.bootStrap.inProgress':'Please wait while Ambari Agents are being set up on the hosts. This can take several minutes depending on the number of hosts.',
  'installer.step2.bootStrap.header':'Setting Up Ambari Agents',

  'installer.step3.header':'Confirm Hosts',
  'installer.step3.body':'Registering your hosts.<br>' +
    'Please confirm the host list and remove any hosts that you do not want to include in the cluster.',
  'installer.step3.hostLog.popup.header':'Registration log for {0}',
  'installer.step3.hosts.remove.popup.header':'Remove Hosts',
  'installer.step3.hosts.remove.popup.body':'Are you sure you want to remove the selected host(s)?',
  'installer.step3.hostInformation.popup.header':'Error in retrieving host Information',
  'installer.step3.hostInformation.popup.body' : 'All bootstrapped hosts registered but unable to retrieve cpu and memory related information',
  'installer.step3.hostOsTypeCheck.popup.body' : 'Host registered successfully, but the host operating system type NOT match the selected group in "Select Stack" step: Advanced Repository Option. You can go back to "Select Stack" step OR remove this host.' +
    'The host type is {0}, but you selected group {1} in step 1.',
  'installer.step3.hostWarningsPopup.report':'Show Report',
  'installer.step3.hostWarningsPopup.report.header': '<p style="font-family: monospace">######################################<br># Host Checks Report<br>#<br># Generated: ',
  'installer.step3.hostWarningsPopup.report.hosts': '<br>######################################<br><br>######################################<br># Hosts<br>#<br># A space delimited list of hosts which have issues.<br># Provided so that administrators can easily copy hostnames into scripts, email etc.<br>######################################<br>HOSTS<br>',
  'installer.step3.hostWarningsPopup.report.jdk': '<br><br>######################################<br># JDK Check <br>#<br># A newline delimited list of JDK issues.<br>######################################<br>JDK ISSUES<br>',
  'installer.step3.hostWarningsPopup.report.disk': '<br><br>######################################<br># Disk <br>#<br># A newline delimited list of disk issues.<br>######################################<br>DISK ISSUES<br>',
  'installer.step3.hostWarningsPopup.report.repositories': '<br><br>######################################<br># Repositories <br>#<br># A newline delimited list of repositories issues.<br>######################################<br>REPOSITORIES ISSUES<br>',
  'installer.step3.hostWarningsPopup.report.hostNameResolution': '<br><br>######################################<br># Hostname Resolution<br>#<br># A newline delimited list of hostname resolution issues.<br>######################################<br>HOSTNAME RESOLUTION ISSUES<br>',
  'installer.step3.hostWarningsPopup.report.thp': '<br><br>######################################<br># Transparent Huge Pages(THP) <br>#<br># A space delimited list of hostnames on which Transparent Huge Pages are enabled.<br>######################################<br>THP ISSUES HOSTS<br>',
  'installer.step3.hostWarningsPopup.report.firewall': '<br><br>######################################<br># Firewall<br>#<br># A newline delimited list of firewall issues.<br>######################################<br>FIREWALL<br>',
  'installer.step3.hostWarningsPopup.report.fileFolders': '<br><br>######################################<br># Files and Folders<br>#<br># A space delimited list of files and folders which should not exist.<br># Provided so that administrators can easily copy paths into scripts, email etc.<br># Example: rm -r /etc/hadoop /etc/hbase<br>######################################<br>FILES AND FOLDERS<br>',
  'installer.step3.hostWarningsPopup.report.reverseLookup': '<br><br>######################################<br># Reverse Lookup<br># <br># The hostname was not found in the reverse DNS lookup. This may result in incorrect behavior. <br># Please check the DNS setup and fix the issue.<br>######################################<br>REVERSE LOOKUP<br>',
  'installer.step3.hostWarningsPopup.report.process': '<br><br>######################################<br># Processes<br>#<br># A comma separated list of process tuples which should not be running.<br># Provided so that administrators can easily copy paths into scripts, email etc.<br>######################################<br>PROCESSES<br>',
  'installer.step3.hostWarningsPopup.report.package': '<br><br>######################################<br># Packages<br>#<br># A space delimited list of software packages which should be uninstalled.<br># Provided so that administrators can easily copy paths into scripts, email etc.<br># Example: yum remove hadoop-hdfs yarn<br>######################################<br>PACKAGES<br>',
  'installer.step3.hostWarningsPopup.report.service': '<br><br>######################################<br># Services<br>#<br># A space delimited list of services which should be up and running.<br># Provided so that administrators can easily copy paths into scripts, email etc.<br># Example: services start ntpd httpd<br>######################################<br>SERVICES<br>',
  'installer.step3.hostWarningsPopup.report.user': '<br><br>######################################<br># Users<br>#<br># A space delimited list of users who should not exist.<br># Provided so that administrators can easily copy paths into scripts, email etc.<br># Example: userdel hdfs<br>######################################<br>USERS<br>',
  'installer.step3.hostWarningsPopup.report.folder': '\\ /folder',
  'installer.step3.hostWarningsPopup.checks': 'Host Checks found',
  'installer.step3.hostWarningsPopup.notice':'After manually resolving the issues, click <b>Rerun Checks</b>.'+
    '<br>To manually resolve issues on <b>each host</b> run the HostCleanup script (Python 2.6 or greater is required):'+
    '<br><div class="code-snippet">python /usr/lib/python2.6/site-packages/ambari_agent/HostCleanup.py --silent --skip=users</div>' +
    '<div class="alert alert-warn"><b>Note</b>: To clean up in interactive mode, remove <b>--silent</b> option. To clean up all resources, including <i>users</i>, remove <b>--skip=users</b> option. Use <b>--help</b> for a list of available options.</div>',
  'installer.step3.hostWarningsPopup.summary':'{0} on {1}',
  'installer.step3.hostWarningsPopup.jdk':'JDK Issues',
  'installer.step3.hostWarningsPopup.jdk.name':'JDK not found at <i>{0}</i>',
  'installer.step3.hostWarningsPopup.jdk.context':'{0}',
  'installer.step3.hostWarningsPopup.jdk.message':'The following registered hosts have issues related to JDK',
  'installer.step3.hostWarningsPopup.repositories':'Repository Issues',
  'installer.step3.hostWarningsPopup.repositories.name':'Repository for OS not available',
  'installer.step3.hostWarningsPopup.repositories.context':'Host ({0}) is {1} OS type, but the repositories chosen in "Select Stack" step was {2}. Selected repositories do not support this host OS type.',
  'installer.step3.hostWarningsPopup.repositories.message': 'The following registered hosts have different Operating System types from the available Repositories chosen in "Select Stack" step. You can go back to "Select Stack" step to select another OS repository <b>or</b> remove the host.',
  'installer.step3.hostWarningsPopup.disk':'Disk Issues',
  'installer.step3.hostWarningsPopup.disk.name':'Not enough disk space ',
  'installer.step3.hostWarningsPopup.disk.context1':'Not enough disk space on host ({0}).',
  'installer.step3.hostWarningsPopup.disk.context2':'A minimum of {0} is required for "{1}" mount.',
  'installer.step3.hostWarningsPopup.disk.message':'The following registered hosts have issues related to disk space',
  'installer.step3.hostWarningsPopup.thp':'Transparent Huge Pages Issues',
  'installer.step3.hostWarningsPopup.thp.name':'Transparent Huge Pages',
  'installer.step3.hostWarningsPopup.thp.context':'{0}',
  'installer.step3.hostWarningsPopup.thp.message':'The following hosts have Transparent Huge Pages (THP) enabled. THP should be disabled to avoid potential Hadoop performance issues.',
  'installer.step3.hostWarningsPopup.firewall':'Firewall Issues',
  'installer.step3.hostWarningsPopup.firewall.message':'Firewall is running on the following hosts. Please configure the firewall to allow communications on the ports documented in the <i>Configuring Ports</i> section of  the <a target=\"_blank\" href=\"http://ambari.apache.org/current/installing-hadoop-using-ambari/content/\">Ambari documentation</a>',
  'installer.step3.hostWarningsPopup.firewall.name':'<i>iptables</i> Running',
  'installer.step3.hostWarningsPopup.process':'Process Issues',
  'installer.step3.hostWarningsPopup.processes.message':'The following processes should not be running',
  'installer.step3.hostWarningsPopup.fileAndFolder':'File and Folder Issues',
  'installer.step3.hostWarningsPopup.fileFolders.message':'The following files and folders should not exist',
  'installer.step3.hostWarningsPopup.package':'Package Issues',
  'installer.step3.hostWarningsPopup.packages.message':'The following packages should be uninstalled',
  'installer.step3.hostWarningsPopup.user':'User Issues',
  'installer.step3.hostWarningsPopup.users.message':'The following users should be removed',
  'installer.step3.hostWarningsPopup.service':'Service Issues',
  'installer.step3.hostWarningsPopup.services.message':'The following services should be up',
  'installer.step3.hostWarningsPopup.misc':'Misc Issues',
  'installer.step3.hostWarningsPopup.misc.message':'The following umasks should be changed',
  'installer.step3.hostWarningsPopup.misc.umask':'Umask',
  'installer.step3.hostWarningsPopup.alternatives':'Alternatives Issues',
  'installer.step3.hostWarningsPopup.alternatives.umask' : 'Alternatives',
  'installer.step3.hostWarningsPopup.alternatives.message':'The following alternatives should be removed',
  'installer.step3.hostWarningsPopup.alternatives.noIssues':'There was no alternative issue following alternatives should be removed',
  'installer.step3.hostWarningsPopup.alternatives.empty': 'alternative issues',
  'installer.step3.hostWarningsPopup.issue':'issue',
  'installer.step3.hostWarningsPopup.issues':'issues',
  'installer.step3.hostWarningsPopup.emptyMessage':'There were no',
  'installer.step3.hostWarningsPopup.empty.filesAndFolders':'unwanted files and folders',
  'installer.step3.hostWarningsPopup.empty.processes':'unwanted processes',
  'installer.step3.hostWarningsPopup.empty.packages':'unwanted packages',
  'installer.step3.hostWarningsPopup.empty.users':'unwanted users',
  'installer.step3.hostWarningsPopup.empty.services':'unwanted services',
  'installer.step3.hostWarningsPopup.empty.misc':'issues',
  'installer.step3.hostWarningsPopup.empty.firewall':'firewalls running',
  'installer.step3.hostWarningsPopup.empty.repositories':'repositories OS type mis-match with registered hosts',
  'installer.step3.hostWarningsPopup.empty.disk':'disk space issues',
  'installer.step3.hostWarningsPopup.empty.jdk':'JDK issues',
  'installer.step3.hostWarningsPopup.empty.thp':'THP issues',
  'installer.step3.hostWarningsPopup.reverseLookup.name': 'Reverse Lookup validation failed on',
  'installer.step3.hostWarningsPopup.reverseLookup': 'Reverse Lookup Issues',
  'installer.step3.hostWarningsPopup.reverseLookup.message': 'The hostname was not found in the reverse DNS lookup. This may result in incorrect behavior. Please check the DNS setup and fix the issue.',
  'installer.step3.hostWarningsPopup.reverseLookup.empty': 'reverse DNS lookup issues.',
  'installer.step3.hostWarningsPopup.resolution.validation.name': 'Hostname Resolution Issues',
  'installer.step3.hostWarningsPopup.resolution.validation.error': 'Hostname resolution',
  'installer.step3.hostWarningsPopup.resolution.validation': 'Hostname resolution validation',
  'installer.step3.hostWarningsPopup.resolution.validation.message': 'Not all hosts could resolve hostnames of other hosts. Make sure that host resolution works properly on all hosts before continuing.',
  'installer.step3.hostWarningsPopup.resolution.validation.empty': 'hostname resolution issues',
  'installer.step3.hostWarningsPopup.resolution.validation.context': '{0} could not resolve: {1}.',
  'installer.step3.hostWarningsPopup.action.exists':'Exists on',
  'installer.step3.hostWarningsPopup.action.notRunning':'Not running on',
  'installer.step3.hostWarningsPopup.action.installed':'Installed on',
  'installer.step3.hostWarningsPopup.action.running':'Running on',
  'installer.step3.hostWarningsPopup.action.invalid':'Invalid on',
  'installer.step3.hostWarningsPopup.action.failed':'Failed on',
  'installer.step3.hostWarningsPopup.action.enabled':'Enabled on',
  'installer.step3.hostWarningsPopup.host':'host',
  'installer.step3.hostWarningsPopup.hosts':'hosts',
  'installer.step3.hostWarningsPopup.moreHosts':'{0} more hosts...<br>Click on link to view all hosts.',
  'installer.step3.hostWarningsPopup.allHosts':'List of hosts',
  'installer.step3.hostWarningsPopup.rerunChecks':'Rerun Checks',
  'installer.step3.hostWarningsPopup.hostHasWarnings':'Warning: Host checks failed on some of your hosts. It is highly recommended that you fix these problems first before proceeding to prevent potentially major problems with cluster installation. Are you sure you want to ignore these warnings and proceed?',
  'installer.step3.warningsWindow.allHosts':'Warnings across all hosts',
  'installer.step3.warningsWindow.warningsOn':'Warnings on ',
  'installer.step3.warningsWindow.directoriesAndFiles':'DIRECTORIES AND FILES',
  'installer.step3.warningsWindow.noWarnings':'No warnings',
  'installer.step3.hosts.noHosts':'No hosts to display',
  'installer.step3.warnings.popup.header':'Host Checks',
  'installer.step3.warnings.description':'Some warnings were encountered while performing checks against the above hosts.',
  'installer.step3.warnings.linkText':'Click here to see the warnings.',
  'installer.step3.noWarnings.linkText':'Click here to see the check results.',
  'installer.step3.warnings.noWarnings':'All host checks passed on {0} registered hosts.',
  'installer.step3.warnings.fails':'Some warnings were encountered while performing checks against the {0} registered hosts above',
  'installer.step3.warnings.someWarnings':'All host checks passed on {0} registered hosts. Note: Host checks were skipped on {1} hosts that failed to register.',
  'installer.step3.warnings.allFailed':'Host checks were skipped on {0} hosts that failed to register.',
  'installer.step3.warnings.updateChecks.success':'Host Checks successfully updated',
  'installer.step3.warnings.updateChecks.failed':'Host Checks update failed',
  'installer.step3.warnings.missingHosts':'There is no registered host',
  'installer.step3.warning.registeredHosts': '{0} Other Registered Hosts',
  'installer.step3.warning.loading':'Please wait while the hosts are being checked for potential problems...',
  'installer.step3.registeredHostsPopup': 'These are the hosts that have registered with the server, but do not appear in the list of hosts that you are adding.',
  'installer.step3.removeSelected':'Remove Selected',
  'installer.step3.retryFailed':'Retry Failed',
  'installer.step3.hosts.status.registering':'Registering',
  'installer.step3.hosts.status.installing':'Installing',
  'installer.step3.hosts.bootLog.failed':'\nRegistration with the server failed.',
  'installer.step3.hosts.bootLog.registering':'\nRegistering with the server...',
  'installer.step3.hostLogPopup.highlight':'click to highlight',
  'installer.step3.hostLogPopup.copy':'press CTRL+C',
  'installer.step3.hostsTable.selectAll':'Select All Hosts',
  'installer.step3.selectedHosts.popup.header':'Selected Hosts',

  'installer.step4.header':'Choose Services',
  'installer.step4.body':'Choose which services you want to install on your cluster.',
  'installer.step4.fsCheck.popup.header':'File System Required',
  'installer.step4.fsCheck.popup.body':'You did not select a File System but one is required. We will automatically add {0}. Is this OK?',
  'installer.step4.multipleDFS.popup.header':'Multiple File Systems Selected',
  'installer.step4.multipleDFS.popup.body':'You selected more than one file system. We will automatically select only {0}. Is this OK?',
  'installer.step4.serviceCheck.popup.header':'{0} Needed',
  'installer.step4.serviceCheck.popup.body':'You did not select {0}, but it is needed by other services you selected. We will automatically add {0}. Is this OK?',
  'installer.step4.ambariMetricsCheck.popup.header':'Limited Functionality Warning',
  'installer.step4.ambariMetricsCheck.popup.body':'Ambari Metrics collects metrics from the cluster and makes them available to Ambari.  If you do not install Ambari Metrics service, metrics will not be accessible from Ambari.  Are you sure you want to proceed without Ambari Metrics?',
  'installer.step4.rangerRequirements.popup.header': 'Ranger Requirements',
  'installer.step4.rangerRequirements.popup.body.requirements': '<ol><li>You must have an <strong>existing MySQL Server or Oracle Server</strong> database instance running to be used by Ranger.</li>' +
    '<li>In Assign Masters step of this wizard, you will be prompted to specify which host for the Ranger Admin. On that host, you <strong>must have MySQL Client or Oracle Client installed</strong> for Ranger to access to the database.</li>' +
    '<li>Ensure that the access for the DB Admin user (<strong>root</strong> in case of MySQL or <strong>SYS</strong> in case of Oracle) is enabled in DB server from any host.</li>' +
    '<li>Execute the following command on the Ambari Server host. Replace <code>database-type</code> with <strong>mysql</strong> or <strong>oracle</strong> and <code>/jdbc/driver/path</code> based on the location of the MySQL or Oracle JDBC driver:' +
    '<pre>ambari-server setup --jdbc-db={database-type} --jdbc-driver={/jdbc/driver/path}</pre></li></ol>',
  'installer.step4.rangerRequirements.popup.body.confirmation': 'I have met all the requirements above.',
  'installer.step4.sparkWarning.popup.body': 'Spark requires HDP 2.2.2 or later. Attempting to install Spark to a HDP 2.2.0 cluster will fail. Confirm you are using HDP 2.2.2 or later packages. Are you sure you want to proceed?',

  'installer.step5.header':'Assign Masters',
  'installer.step5.reassign.header':'Select Target Host',
  'installer.step5.attention':' hosts not running master services',
  'installer.step5.body':'Assign master components to hosts you want to run them on.',
  'installer.step5.body.coHostedComponents':'<i class="icon-asterisks">&#10037;</i> {0} will be hosted on the same host.',
  'installer.step5.hostInfo':'%@ (%@, %@ cores)',
  'installer.step5.hiveGroup':'HiveServer2, WebHCat Server, MySQL Server',
  'installer.step5.validationIssuesAttention.header': 'Validation Issues',
  'installer.step5.validationIssuesAttention': 'Master component assignments have issues that need attention.',

  'installer.step6.header':'Assign Slaves and Clients',
  'installer.step6.body':'Assign slave and client components to hosts you want to run them on.<br/>Hosts that are assigned master components are shown with <i class=icon-asterisks>&#10037;</i>.',
  'installer.step6.body.clientText': ' <br/>&quot;Client&quot; will install ',
  'installer.step6.error.mustSelectOne':'You must assign at least one host to each component.',
  'installer.step6.error.mustSelectOneForHost':'You must assign at least one slave/client component to each.',
  'installer.step6.error.mustSelectComponents': 'You must assign at least: {0}',
  'installer.step6.wizardStep6Host.title':'master components hosted on {0}',
  'installer.step6.addHostWizard.body':'Assign HBase master and ZooKeeper server.',
  'installer.step6.error.mustSelectOneForSlaveHost': 'You must assign at least one slave/client component to each host with no master component',
  'installer.step6.validationSlavesAndClients.hasIssues': 'Your slave and client assignment has issues. ',
  'installer.step6.validationSlavesAndClients.click': 'Click',
  'installer.step6.validationSlavesAndClients.forDetails': ' for details.',
  'installer.step6.validationSlavesAndClients.popup.header': 'Assign Slaves and Clients Issues',
  'installer.step6.validationSlavesAndClients.popup.body': 'Assignment of slave and client components has the following issues',
  'installer.step6.validationIssuesAttention.header': 'Validation Issues',
  'installer.step6.validationIssuesAttention': 'Slave and Client component assignments have issues that need attention.',

  'installer.step7.header':'Customize Services',
  'installer.step7.body':'We have come up with recommended configurations for the services you selected. Customize them as you see fit.',
  'installer.step7.attentionNeeded':'<b>Attention:</b> Some configurations need your attention before you can proceed.',
  'installer.step7.config.addProperty':'Add Property',
  'installer.step7.ConfigErrMsg.header':'Custom configuration error: ',
  'installer.step7.ConfigErrMsg.message':'Error in custom configuration. Some properties entered in the box are already exposed on this page',
  'installer.step7.popup.currentValue':'Current Value',
  'installer.step7.popup.adjustedValue':'Adjusted Value',
  'installer.step7.popup.rddWarning.header':'Warning: disk space low on {0}',
  'installer.step7.popup.rddWarning.body':'A minimum of 16GB is recommended for the Ganglia Server logs but the disk mount "{0}" on {1} does not have enough space ({2} free). Go to the Misc tab and change Ganglia rrdcached base directory with more than 16GB of disk space. If you proceed without changing it, {1} will likely run out of disk space and become inoperable.',
  'installer.step7.popup.mySQLWarning.header':'Error: New MySQL Database for Hive Conflict',
  'installer.step7.popup.mySQLWarning.body':'You cannot install a \"New MySQL Database\" for Hive on the same host as the Ambari Server MySQL Server. Please go back to <b>Assign Masters</b> and reassign Hive to another host <b>OR</b> choose \"Existing MySQL Database\" option to specify the Database Credentials and URL for the Ambari Server MySQL Server instance. If you choose \"Existing MySQL Database\" option, you need to perform the Hive prerequisite steps to prepare the MySQL Server instance for Hive.',
  'installer.step7.popup.mySQLWarning.button.gotostep5': 'Go to Assign Masters',
  'installer.step7.popup.mySQLWarning.button.dismiss': 'Dismiss',
  'installer.step7.popup.mySQLWarning.confirmation.header': 'Confirmation: Go to Assign Masters',
  'installer.step7.popup.mySQLWarning.confirmation.body': 'You will be brought back to the \"Assign Masters\" step and will lose all your current customizations. Are you sure?',
  'installer.step7.popup.database.connection.header': 'Database Connectivity Warning',
  'installer.step7.popup.database.connection.body': 'You have not run or passed the database connection test for: {0}. It is highly recommended that you pass the connection test before proceeding to prevent failures during deployment.',
  'installer.step7.popup.validation.failed.header': 'Validation failed.',
  'installer.step7.popup.validation.failed.body': 'Some services are not properly configured. You have to change the highlighted configs according to the recommended values.',
  'installer.step7.popup.validation.request.failed.body': 'Config validation failed.',
  'installer.step7.popup.validation.warning.header': 'Configurations',
  'installer.step7.popup.validation.warning.body': 'Some service configurations are not configured properly. We recommend you review and change the highlighted configuration values. Are you sure you want to proceed without correcting configurations?',
  'installer.step7.oozie.database.new': 'New Derby Database',
  'installer.step7.hive.database.new.mysql': 'New MySQL Database',
  'installer.step7.hive.database.new.postgres': 'New PostgreSQL Database',
  'installer.step7.misc.notification.configure':'Configure email notifications',
  'installer.step7.misc.notification.configure.later':'Configure email notifications later',
  'installer.step7.misc.notification.use_tls':'Use TLS',
  'installer.step7.misc.notification.use_ssl':'Use SSL',


  'installer.step8.header':'Review',
  'installer.step8.body':'Please review the configuration before installation',
  'installer.step8.deployPopup.message':'Preparing to Deploy: {0} of {1} tasks completed.',
  'installer.step8.hosts':' hosts',
  'installer.step8.host':' host',
  'installer.step8.other':'and {0} other hosts',
  'installer.step8.repoInfo.osType.redhat6':'RHEL 6/CentOS 6/Oracle Linux 6',
  'installer.step8.repoInfo.osType.redhat5':'RHEL 5/CentOS 5/Oracle Linux 5',
  'installer.step8.repoInfo.osType.sles11':'SLES 11/SUSE 11',
  'installer.step8.repoInfo.displayName':'Repositories',
  'installer.step9.header':'Install, Start and Test',
  'installer.step9.body':'Please wait while the selected services are installed and started.',
  'installer.step9.status.success':'Successfully installed and started the services.',
  'installer.step9.status.warning':'Installed and started the services with some warnings.',
  'installer.step9.status.failed':'Failed to install/start the services.',
  'installer.step9.status.start.services.failed':'Start all services API call failed.',
  'installer.step9.status.install.components.failed': 'Some service components are still not known to have installed successfully. Please Retry',
  'installer.step9.status.hosts.heartbeat_lost': 'Ambari agent is not running on {0} hosts.',
  'installer.step9.host.heartbeat_lost': 'Heartbeat lost for the host',
  'installer.step9.host.heartbeat_lost_popup': 'Ambari agent process is not heartbeating on the host',
  'installer.step9.host.heartbeat_lost.header' : 'Hosts with heartbeat lost',
  'installer.step9.host.status.success':'Success',
  'installer.step9.host.status.startAborted':'Install completed. Start aborted',
  'installer.step9.host.status.warning':'Warnings encountered',
  'installer.step9.host.status.failed':'Failures encountered',
  'installer.step9.host.status.nothingToInstall':'Install complete (Waiting to start)',
  'installer.step9.hosts.status.label.inProgress':'In Progress',
  'installer.step9.hosts.status.label.warning':'Warning',
  'installer.step9.hosts.filteredHostsInfo': '{0} of {1} hosts showing',
  'installer.step9.hostLog.popup.header':'Tasks. executed on ',
  'installer.step9.hostLog.popup.categories.all':'All',
  'installer.step9.hostLog.popup.categories.pending':'Queued / Pending',
  'installer.step9.hostLog.popup.categories.in_progress':'In Progress',
  'installer.step9.hostLog.popup.categories.failed':'Failed',
  'installer.step9.hostLog.popup.categories.completed':'Success',
  'installer.step9.hostLog.popup.categories.aborted':'Aborted',
  'installer.step9.hostLog.popup.categories.timedout':'Timed Out',
  'installer.step9.hostLog.popup.noTasksToShow':'No tasks to show',
  'installer.step9.host.status.noTasks': 'Ambari server did not schedule any tasks on the host. Either the service component on the host is already in installed state <b>OR</b> the pre-check of host reachability failed.',
  'installer.step9.overallProgress':'{0} % overall',
  'installer.step9.serviceStatus.install.pending':'Preparing to install ',
  'installer.step9.serviceStatus.install.queued':'Waiting to install ',
  'installer.step9.serviceStatus.install.inProgress':'Installing ',
  'installer.step9.serviceStatus.install.completed':'Successfully installed ',
  'installer.step9.serviceStatus.install.failed':'Failed to install ',
  'installer.step9.serviceStatus.uninstall.pending':'Preparing to uninstall ',
  'installer.step9.serviceStatus.uninstall.queued':'Waiting to uninstall ',
  'installer.step9.serviceStatus.uninstall.inProgress':'Uninstalling ',
  'installer.step9.serviceStatus.uninstall.completed':'Successfully uninstalled ',
  'installer.step9.serviceStatus.uninstall.failed':'Failed to uninstall ',
  'installer.step9.serviceStatus.start.pending':'Preparing to start ',
  'installer.step9.serviceStatus.start.queued':'Waiting to start ',
  'installer.step9.serviceStatus.start.inProgress':'Starting ',
  'installer.step9.serviceStatus.start.completed':' started successfully',
  'installer.step9.serviceStatus.start.failed':' failed to start',
  'installer.step9.serviceStatus.stop.pending':'Preparing to stop ',
  'installer.step9.serviceStatus.stop.queued':'Waiting to stop ',
  'installer.step9.serviceStatus.stop.inProgress':'Stopping ',
  'installer.step9.serviceStatus.stop.completed':' stopped successfully',
  'installer.step9.serviceStatus.stop.failed':' failed to stop',
  'installer.step9.serviceStatus.execute.pending':'Preparing to execute ',
  'installer.step9.serviceStatus.execute.queued':'Waiting to execute ',
  'installer.step9.serviceStatus.execute.inProgress':'Executing ',
  'installer.step9.serviceStatus.execute.completed':' executed successfully',
  'installer.step9.serviceStatus.execute.failed':' failed to execute',
  'installer.step9.serviceStatus.abort.pending':'Preparing to abort ',
  'installer.step9.serviceStatus.abort.queued':'Waiting to abort ',
  'installer.step9.serviceStatus.abort.inProgress':'Aborting ',
  'installer.step9.serviceStatus.abort.completed':' aborted successfully',
  'installer.step9.serviceStatus.abort.failed':' failed to abort',
  'installer.step9.components.install.failed': 'Installation Failure',

  'installer.step10.header':'Summary',
  'installer.step10.body':'Here is the summary of the install process.',
  'installer.step10.staleServicesRestartRequired':' You may also need to restart other services for the newly added ' +
    'services to function properly (for example, HDFS and YARN/MapReduce need to be restarted after adding Oozie). After closing this ' +
    'wizard, please restart all services that have the restart indicator <i class="icon-refresh"></i> next to the service name.',
  'installer.step10.hostsSummary':'The cluster consists of {0} hosts',
  'installer.step10.servicesSummary':'Installed and started services successfully on {0} new ',
  'installer.step10.warnings':' warnings',
  'installer.step10.clusterState.installing':'Installing ',
  'installer.step10.clusterState.starting':'Starting ',
  'installer.step10.taskStatus.failed':' failed on ',
  'installer.step10.taskStatus.aborted':' aborted on ',
  'installer.step10.taskStatus.timedOut':' timed out on ',
  'installer.step10.installStatus.failed':'Installing master services failed',
  'installer.step10.installStatus.installed':'Master services installed',
  'installer.step10.master.installedOn':'{0} installed on {1}',
  'installer.step10.startStatus.failed':'Starting services failed',
  'installer.step10.startStatus.passed':'All tests passed',
  'installer.step10.startStatus.started':'All services started',
  'installer.step10.installTime.seconds':'Install and start completed in {0} seconds',
  'installer.step10.installTime.minutes':'Install and start completed in {0} minutes and {1} seconds',

  'addHost.step4.header':'Configurations',
  'addHost.step4.title':'Select the configuration groups to which the added hosts will belong to.',

  'form.create':'Create',
  'form.save':'Save',
  'form.cancel':'Cancel',
  'form.passwordRetype':'Retype Password',
  'form.saveSuccess':'Successfully saved.',
  'form.saveError':'Sorry, errors occurred.',
  'form.item.placeholders.typePassword':'Type password',

  'form.validator.invalidIp':'Please enter valid ip address',
  'form.validator.configKey':'Invalid Key. Only alphanumerics, hyphens, underscores, asterisks and periods are allowed.',
  'form.validator.configGroupName':'Invalid Group Name. Only alphanumerics, hyphens, spaces and underscores are allowed.',
  'form.validator.alertGroupName':'Invalid Alert Group Name. Only alphanumerics, hyphens, spaces and underscores are allowed.',
  'form.validator.configKey.specific':'"{0}" is invalid Key. Only alphanumerics, hyphens, underscores, asterisks and periods are allowed.',

  'alerts.add.header': 'Create Alert Definition',
  'alerts.add.step1.header': 'Choose Type',
  'alerts.add.step2.header': 'Configure',
  'alerts.add.step3.header': 'Review',
  'alerts.add.step3.selectedType': 'Selected Type',

  'alerts.fastAccess.popup.header': '{0} Critical or Warning Alerts',
  'alerts.fastAccess.popup.body.name': 'Alert Definition Name',
  'alerts.fastAccess.popup.body.showmore': 'Go to Alerts Definitions',
  'alerts.fastAccess.popup.body.noalerts': 'No critical or warning alerts to display',

  'alerts.actions.create': 'Create Alert',
  'alerts.actions.manageGroups': 'Manage Alert Groups',
  'alerts.actions.manageNotifications': 'Manage Notifications',
  'alerts.actions.manageNotifications.info': 'You can manage notification methods and recipients.',

  'alerts.groups.successPopup.header': 'Alert Groups processing results',
  'alerts.groups.successPopup.body.created': 'New Alert Groups',
  'alerts.groups.successPopup.body.updated': 'Updated Alert Groups',
  'alerts.groups.successPopup.body.deleted': 'Removed Alert Groups',

  'alerts.table.noAlerts': 'No Alerts to display',
  'alerts.table.header.lastTriggered': 'Last Status Changed',
  'alerts.table.header.lastChecked': 'Last Checked',
  'alerts.table.header.lastTrigger': 'Last Changed',
  'alerts.table.header.check.response': 'Response',
  'alerts.table.header.definitionName': 'Alert Definition Name',
  'alerts.table.header.notification': 'Notification',
  'alerts.table.state': 'State',
  'alerts.table.state.enabled': 'Enabled',
  'alerts.table.state.disabled': 'Disabled',
  'alerts.table.state.enabled.tooltip': 'Click to disable this alert definition',
  'alerts.table.state.disabled.tooltip': 'Click to enable this alert definition',
  'alerts.table.state.enabled.confirm.msg': 'You are about to Disable this alert definition.',
  'alerts.table.state.disabled.confirm.msg': 'You are about to Enable this alert definition.',
  'alerts.table.state.enabled.confirm.btn': 'Confirm Disable',
  'alerts.table.state.disabled.confirm.btn': 'Confirm Enable',
  'alerts.filters.filteredAlertsInfo': '{0} of {1} definitions showing',
  'alerts.definition.name': 'Alert Definition Name',
  'alerts.saveChanges': 'You have unsaved changes',

  'alerts.definition.details.enable': 'Enable',
  'alerts.definition.details.disable': 'Disable',
  'alerts.definition.details.enableDisable': 'Enable / Disable',
  'alerts.definition.details.groups': 'Groups',
  'alerts.definition.details.instances': 'Instances',
  'alerts.definition.details.serviceHost': 'Service / Host',
  'alerts.definition.details.24-hour': '24-Hour',
  'alerts.definition.details.notification': 'Notification',
  'alerts.definition.details.noAlerts': 'No alert instances to show',
  'alerts.definition.details.configs.thresholdsErrorMsg': 'Critical threshold should be larger than warning threshold',

  'alerts.notifications.error.email': 'Must be a valid email address',
  'alerts.notifications.error.integer': 'Must be an integer',
  'alerts.notifications.error.retypePassword': 'Password confirmation must match password',

  'alerts.notifications.addCustomPropertyPopup.header': 'Add Property',
  'alerts.notifications.addCustomPropertyPopup.error.propertyExists': 'Custom Property with current name already exists',
  'alerts.notifications.addCustomPropertyPopup.error.invalidPropertyName': 'Property name can only contain letters, numbers or . -_* characters',


  'wizard.progressPage.notice.completed':'Please proceed to the next step.',
  'wizard.progressPage.notice.failed':'You can click on the Retry button to retry failed tasks.',
  'wizard.singleRequest.progressPage.notice.failed': 'Please click on the Retry link to retry the failed request.',

  'admin.advanced.caution':'This section is for advanced user only.<br/>Proceed with caution.',
  'admin.advanced.button.uninstallIncludingData':'Uninstall cluster including all data.',
  'admin.advanced.button.uninstallKeepData':'Uninstall cluster but keep data.',

  'admin.advanced.popup.header':'Uninstall Cluster',

  /*'admin.advanced.popup.text':'Uninstall Cluster',*/

  'admin.audit.grid.date':"Date/Time",
  'admin.audit.grid.category':"Category",
  'admin.audit.grid.operationName':"Operation",
  'admin.audit.grid.performedBy':"Performed By",
  'admin.audit.grid.service':"Category",

  'admin.authentication.form.method.database':'Use Ambari Database to authenticate users',
  'admin.authentication.form.method.ldap':'Use LDAP/Active Directory to authenticate',
  'admin.authentication.form.primaryServer':'Primary Server',
  'admin.authentication.form.secondaryServer':'Secondary Server',
  'admin.authentication.form.useSsl':'Use SSL',
  'admin.authentication.form.bind.anonymously':"Bind Anonymously",
  'admin.authentication.form.bind.useCrenedtials':"Use Credentials To Bind",
  'admin.authentication.form.bindUserDN':'Bind User DN',
  'admin.authentication.form.searchBaseDN':'Search Base DN',
  'admin.authentication.form.usernameAttribute':'Username Attribute',

  'admin.authentication.form.userDN':'User DN',
  'admin.authentication.form.configurationTest':'Configuration Test',
  'admin.authentication.form.testConfiguration':'Test Configuration',

  'admin.authentication.form.test.success':'The configuration passes the test',
  'admin.authentication.form.test.fail':'The configuration fails the test',

  'admin.kerberos.wizard.configuration.note': 'This is the initial configuration created by Enable Kerberos wizard.',
  'admin.kerberos.wizard.header':'Enable Kerberos Wizard',
  'admin.kerberos.button.enable': 'Enable Kerberos',
  'admin.kerberos.button.disable': 'Disable Kerberos',
  'admin.kerberos.button.regenerateKeytabs': 'Regenerate Keytabs',
  'admin.kerberos.wizard.exit.warning.msg': 'Configuring Kerberos is in progress. Do you really want to exit the Enable Kerberos Wizard?',
  'admin.kerberos.wizard.exit.critical.msg': 'Configuring Kerberos is in progress. <strong>Before dismissing, you should complete the wizard.</strong> Do you really want to exit the Enable Kerberos Wizard?',
  'admin.kerberos.wizard.step1.header': 'Get Started',
  'admin.kerberos.wizard.step2.header': 'Configure Kerberos',
  'admin.kerberos.wizard.step3.header': 'Install and Test Kerberos Client',
  'admin.kerberos.wizard.step4.header': 'Configure Identities',
  'admin.kerberos.wizard.step5.header': 'Stop Services',
  'admin.kerberos.wizard.step6.header': 'Kerberize Cluster',
  'admin.kerberos.wizard.step7.header': 'Start and Test Services',
  'admin.kerberos.wizard.step1.info.body': 'Welcome to the Ambari Security Wizard. Use this wizard to enable kerberos security in your cluster. </br>Let\'s get started.',
  'admin.kerberos.wizard.step1.alert.body': 'Note: This process requires services to be restarted and cluster downtime. As well, depending on the options you select, might require support from your Security administrators. Please plan accordingly.',
  'admin.kerberos.wizard.step1.body.text': 'What type of KDC do you plan on using?',
  'admin.kerberos.wizard.step1.option.kdc': 'Existing MIT KDC',
  'admin.kerberos.wizard.step1.option.kdc.condition.1': 'Ambari Server and cluster hosts have network access to both the KDC and KDC admin hosts.',
  'admin.kerberos.wizard.step1.option.kdc.condition.2': 'KDC administrative credentials are on-hand.',
  'admin.kerberos.wizard.step1.option.kdc.condition.3': 'The Java Cryptography Extensions (JCE) have been setup on the Ambari Server host and all hosts in the cluster.',
  'admin.kerberos.wizard.step1.option.ad': 'Existing Active Directory',
  'admin.kerberos.wizard.step1.option.ad.condition.1': 'Ambari Server and cluster hosts have network access to the Domain Controllers.',
  'admin.kerberos.wizard.step1.option.ad.condition.2': 'Active Directory secure LDAP (LDAPS) connectivity has been configured.',
  'admin.kerberos.wizard.step1.option.ad.condition.3': 'Active Directory User container for principals has been created and is on-hand (e.g. OU=Hadoop,OU=People,dc=apache,dc=org)',
  'admin.kerberos.wizard.step1.option.ad.condition.4': 'Active Directory administrative credentials with delegated control of “Create, delete, and manage user accounts” on the previously mentioned User container are on-hand.',
  'admin.kerberos.wizard.step1.option.ad.condition.5': 'The Java Cryptography Extensions (JCE) have been setup on the Ambari Server host and all hosts in the cluster.',
  'admin.kerberos.wizard.step1.prerequisites.label': 'Following prerequisites needs to be checked to progress ahead in the wizard.',
  'admin.kerberos.wizard.step2.info.body': 'Please configure kerberos related properties.',
  'admin.kerberos.wizard.step3.task0.title': 'Install Kerberos Client',
  'admin.kerberos.wizard.step3.task1.title': 'Test Kerberos Client',
  'admin.kerberos.wizard.step3.notice.inProgress': 'Please wait while the Kerberos clients are being installed and tested.',
  'admin.kerberos.wizard.step3.notice.completed': 'Kerberos service has been installed and tested successfully.',
  'admin.kerberos.wizard.progressPage.notice.inProgress': 'Please wait while cluster is being kerberized',
  'admin.kerberos.wizard.step4.info.body': 'Configure principal name and keytab location for service users and hadoop service components.',
  'admin.kerberos.wizard.step5.task0.title' : 'Stop Services',
  'admin.kerberos.wizard.step5.task1.title' : 'Delete ATS',
  'admin.kerberos.wizard.step5.notice.inProgress': 'Please wait while services are being stopped.',
  'admin.kerberos.wizard.step5.notice.completed': 'Services have been successfully stopped.',
  'admin.kerberos.wizard.step6.notice.inProgress': 'Please wait while cluster is being kerberized.',
  'admin.kerberos.wizard.step6.notice.completed': 'Kerberos has successfully been enabled on the cluster.',
  'admin.kerberos.wizard.step7.notice.inProgress': 'Please wait while services are being started and tested.',
  'admin.kerberos.wizard.step7.notice.completed': 'Services have been successfully tested with kerberos environment.',
  'admin.kerberos.wizard.step7.notice.failed': 'Some services failed to start and execute tests successfully. Click Retry to attempt again or click Complete to dismiss the wizard and fix manually.',
  'admin.kerberos.wizard.step7.task0.title' : 'Start and Test Services',

  'admin.kerberos.regenerate_keytabs.popup.body': 'You are about to regenerate keytabs for the hosts in the cluster. This will stop all the services in your cluster and regenerate the Kerberos keytabs. <strong>Are you sure you wish to proceed with keytab regeneration?</strong>',
  'admin.kerberos.regenerate_keytabs.checkbox.label': ' Only regenerate keytabs for missing hosts and components',

  'admin.kerberos.disable.step1.task0.title': 'Stop Services',
  'admin.kerberos.disable.step1.task1.title': 'Unkerberize Cluster',
  'admin.kerberos.disable.step1.task2.title': 'Remove Kerberos',
  'admin.kerberos.disable.step1.task3.title': 'Start Services',
  'admin.kerberos.disable.unkerberize.header': 'Unkerberize cluster',
  'admin.kerberos.disable.unkerberize.message': 'You cannot quit wizard while cluster is being unkerberized',
  'admin.kerberos.disable.inProgress': 'Please wait while cluster is being unkerberized',
  'admin.kerberos.disable.notice.completed': 'Services have been successfully tested without kerberos environment.',
  'admin.kerberos.wizard.step1.notice.inProgress': 'Please wait while cluster is being unkerberized',

  'admin.highAvailability':' High Availability',
  'admin.highAvailability.button.enable':'Enable NameNode HA',
  'admin.highAvailability.button.disable':'Disable NameNode HA',
  'admin.rm_highAvailability.button.enable':'Enable ResourceManager HA',
  'admin.rm_highAvailability.button.disable':'Disable ResourceManager HA',
  'admin.highAvailability.disabled':'NameNode HA is disabled',
  'admin.highAvailability.enabled':'NameNode HA is enabled',
  'admin.rm_highAvailability.disabled':'ResourceManager HA is disabled',
  'admin.rm_highAvailability.enabled':'ResourceManager HA is enabled',
  'admin.highAvailability.confirmRollbackHeader':'Confirm Rollback',
  'admin.highAvailability.confirmRollbackBody':'This will rollback all operations that were done in HA wizard',
  'admin.highAvailability.confirmManualRollbackBody':'You are in the process of enabling NameNode HA. If you exit now, you must follow manual instructions to revert back to the non-HA setup as documented in the Ambari User Guide\'s <i>Rolling Back NameNode HA</i> section.  Are you sure you want to exit the wizard?',
  'admin.highAvailability.error.hostsNum':'You must have at least 3 hosts in your cluster to enable NameNode HA.',
  'admin.highAvailability.error.namenodeStarted':'NameNode must be running before you enable NameNode HA.',
  'admin.highAvailability.error.zooKeeperNum':'You must have at least 3 ZooKeeper Servers in your cluster to enable NameNode HA.',
  'admin.rm_highAvailability.error.hostsNum':'You must have at least 3 hosts in your cluster to enable ResourceManager HA.',
  'admin.rm_highAvailability.error.zooKeeperNum':'You must have at least 3 ZooKeeper Servers in your cluster to enable ResourceManager HA.',
  'admin.rm_highAvailability.closePopup':'Enable ResourceManager HA Wizard is in progress. You must allow the wizard to complete for Ambari to be in usable state. If you choose to quit, you must follow manual instructions to complete or revert enabling ResourceManager HA as documented in the Ambari User Guide. Are you sure you want to exit the wizard?',

  'admin.highAvailability.wizard.header':'Enable NameNode HA Wizard',
  'admin.highAvailability.wizard.progressPage.notice.inProgress':'Please wait while NameNode HA is being deployed.',
  'admin.highAvailability.wizard.progressPage.header':'Deploy',
  'admin.highAvailability.wizard.step1.header':'Get Started',
  'admin.highAvailability.wizard.step1.nameserviceid.tooltip.title':'Nameservice ID',
  'admin.highAvailability.wizard.step1.nameserviceid.tooltip.content':'This will be the ID for the NameNode HA cluster. For example, if you set Nameservice ID to <b>mycluster</b>, the logical URI for HDFS will be <b>hdfs://mycluster</b>.',
  'admin.highAvailability.wizard.step1.nameserviceid':'Nameservice ID',
  'admin.highAvailability.wizard.step1.nameserviceid.error':'Must consist of letters, numbers, and hyphens. Cannot begin or end with a hyphen.',
  'admin.highAvailability.wizard.step2.header':'Select Hosts',
  'admin.highAvailability.wizard.step3.header':'Review',
  'admin.highAvailability.wizard.step4.header':'Create Checkpoint',
  'admin.highAvailability.wizard.step4.error.nameNode':'NameNode is in the process of being stopped. Please make sure that namenode is running to create checkpoint successfully.',
  'admin.highAvailability.wizard.step5.header':'Configure Components',
  'admin.highAvailability.wizard.step6.header':'Initialize JournalNodes',
  'admin.highAvailability.wizard.step7.header':'Start Components',
  'admin.highAvailability.wizard.step8.header':'Initialize Metadata',
  'admin.highAvailability.wizard.step9.header':'Finalize HA Setup',
  'admin.highAvailability.wizard.step4.bodyHeader':'Manual Steps Required: Create Checkpoint on NameNode',
  'admin.highAvailability.wizard.step6.bodyHeader':'Manual Steps Required: Initialize JournalNodes',
  'admin.highAvailability.wizard.step8.bodyHeader':'Manual Steps Required: Initialize NameNode HA Metadata',

  'admin.rollbackHighAvailability.wizard.step1.header':'Select Hosts page',
  'admin.rollbackHighAvailability.wizard.step2.header':'Create Checkpoint page',
  'admin.rollbackHighAvailability.wizard.step3.header':'Progress page',

  'admin.highAvailability.wizard.step5.notice.inProgress':'Please wait while the wizard configures the components.',
  'admin.highAvailability.wizard.step7.notice.inProgress':'Please wait while the wizard starts the components.',
  'admin.highAvailability.wizard.step9.notice.inProgress':'Please wait while the wizard finalizes the HA setup.',
  'admin.highAvailability.wizard.rollback.notice.inProgres':'Reverting Back to Non-HA Setup',

  'admin.highAvailability.wizard.step5.header.title':'Configure Components',
  'admin.highAvailability.wizard.step7.header.title':'Start Components',
  'admin.highAvailability.wizard.step9.header.title':'Finalize HA Setup',
  'admin.highAvailability.wizard.rollback.header.title':'Reverting Back to Non-HA Setup.',

  'admin.highAvailability.wizard.step5.task0.title':'Stop All Services',
  'admin.highAvailability.wizard.step5.task1.title':'Install Additional NameNode',
  'admin.highAvailability.wizard.step5.task2.title':'Install JournalNodes',
  'admin.highAvailability.wizard.step5.task3.title':'Reconfigure HDFS',
  'admin.highAvailability.wizard.step5.task4.title':'Start JournalNodes',
  'admin.highAvailability.wizard.step5.task5.title':'Disable Secondary NameNode',

  'admin.highAvailability.wizard.step7.task0.title':'Start ZooKeeper Servers',
  'admin.highAvailability.wizard.step7.task1.title':'Start NameNode',

  'admin.highAvailability.wizard.step9.task0.title':'Start Additional NameNode',
  'admin.highAvailability.wizard.step9.task1.title':'Install Failover Controllers',
  'admin.highAvailability.wizard.step9.task2.title':'Start Failover Controllers',
  'admin.highAvailability.wizard.step9.task3.title':'Reconfigure HBase',
  'admin.highAvailability.wizard.step9.task4.title':'Delete Secondary NameNode',
  'admin.highAvailability.wizard.step9.task5.title':'Start All Services',
  'admin.highAvailability.wizard.step9.notice.completed':'NameNode HA has been enabled successfully.',

  'admin.highAvailability.wizard.step3.curNameNode': '<b>Current NameNode:</b> ',
  'admin.highAvailability.wizard.step3.addNameNode': '<b>Additional NameNode:</b> ',
  'admin.highAvailability.wizard.step3.secNameNode': '<b>Secondary NameNode:</b> ',
  'admin.highAvailability.wizard.step3.journalNode': '<b>JournalNode:</b> ',
  'admin.highAvailability.wizard.step3.toBeInstalled': 'TO BE INSTALLED',
  'admin.highAvailability.wizard.step3.toBeDeleted': 'TO BE DELETED',
  'admin.highAvailability.wizard.step4.ckNotCreated':'Checkpoint not created yet',
  'admin.highAvailability.wizard.step4.ckCreated':'Checkpoint created',
  'admin.highAvailability.step4.save.configuration.note':'This configuration is created by Enable {0} HA wizard',
  'admin.highAvailability.wizard.step6.jsNoInit':'JournalNodes not initialized yet',
  'admin.highAvailability.wizard.step6.jsInit':'JournalNodes initialized',
  'admin.highAvailability.wizard.step8.metaNoInit':'Metadata not initialized yet',
  'admin.highAvailability.wizard.step8.confirmPopup.body':'Please confirm that you have run the manual steps before continuing.',

  'admin.highAvailability.rollback.header':'Disable NameNode HA Wizard',
  'admin.highAvailability.rollback.task0.title':'Stop All Services',
  'admin.highAvailability.rollback.task1.title':'Restore HBase Configurations',
  'admin.highAvailability.rollback.task2.title':'Stop Failover Controllers',
  'admin.highAvailability.rollback.task3.title':'Delete Failover Controllers',
  'admin.highAvailability.rollback.task4.title':'Stop Additional NameNode',
  'admin.highAvailability.rollback.task5.title':'Stop NameNode',
  'admin.highAvailability.rollback.task6.title':'Restore HDFS Configurations',
  'admin.highAvailability.rollback.task7.title':'Enable Secondary NameNode',
  'admin.highAvailability.rollback.task8.title':'Stop JournalNodes',
  'admin.highAvailability.rollback.task9.title':'Delete JournalNodes',
  'admin.highAvailability.rollback.task10.title':'Delete Additional NameNode',
  'admin.highAvailability.rollback.task11.title':'Start All Services',
  'admin.highAvailability.rollback.notice.inProgress':'Please wait while the wizard reverts back to the non-HA setup.',

  'admin.highAvailability.rollback.step2.body':
    '<ol>' +
      '<li>Login to the NameNode host <b>{1}</b>.</li>' +
      '<li>Put the NameNode in Safe Mode (read-only mode):' +
      '<div class="code-snippet">sudo su {0} -l -c \'hdfs dfsadmin -safemode enter\'</div></li>' +
      '<li>Once in Safe Mode, create a Checkpoint:' +
      '<div class="code-snippet">sudo su {0} -l -c \'hdfs dfsadmin -saveNamespace\'</div></li>' +
      '</ol>',

  'admin.highAvailability.wizard.step8.body':
    '<div class="alert alert-info">' +
    '<ol>' +
    '<li>Login to the NameNode host <b>{1}</b>.</li>' +
    '<li>Initialize the metadata for NameNode automatic failover by running:' +
    '<div class="code-snippet">sudo su {0} -l -c \'hdfs zkfc -formatZK\'</div></li>' +
    '</div>' +
    '<div class="alert alert-info">' +
    '<ol start="3">' +
    '<li>Login to the Additional NameNode host <b>{2}</b>.<br>' +
    '<div class="alert alert-warn"><strong>Important!</strong> Be sure to login to the Additional NameNode host.<br>This is a different host from the Steps 1 and 2 above.</div>' +
    '</li>' +
    '<li>Initialize the metadata for the Additional NameNode by running:' +
    '<div class="code-snippet">sudo su {0} -l -c \'hdfs namenode -bootstrapStandby\'</div></li>' +
    '</ol>' +
    '</div>' +
    'Please proceed once you have completed the steps above.',
  'admin.highAvailability.wizard.step6.body':
    '<ol>' +
    '<li>Login to the NameNode host <b>{1}</b>.</li>' +
    '<li>Initialize the JournalNodes by running:' +
    '<div class="code-snippet">sudo su {0} -l -c \'hdfs namenode -initializeSharedEdits\'</div></li>' +
    '<li>You will be able to proceed once Ambari detects that the JournalNodes have been initialized successfully.</li>' +
    '</ol>',
  'admin.highAvailability.wizard.step4.body':
    '<ol>' +
    '<li>Login to the NameNode host <b>{1}</b>.</li>' +
    '<li>Put the NameNode in Safe Mode (read-only mode):' +
    '<div class="code-snippet">sudo su {0} -l -c \'hdfs dfsadmin -safemode enter\'</div></li>' +
    '<li>Once in Safe Mode, create a Checkpoint:' +
    '<div class="code-snippet">sudo su {0} -l -c \'hdfs dfsadmin -saveNamespace\'</div></li>' +
    '<li>You will be able to proceed once Ambari detects that the NameNode is in Safe Mode and the Checkpoint has been created successfully.</li>'+
    '<div class="alert alert-warn">If the <b>Next</b> button is enabled before you run the <b>"Step 4: Create a Checkpoint"</b> command, it means there is a recent Checkpoint already and you may proceed without running the <b>"Step 4: Create a Checkpoint"</b> command.</div>' +
    '</ol>',
  'admin.highAvailability.wizard.step3.confirm.host.body':'<b>Confirm your host selections.</b>',
  'admin.highAvailability.wizard.step3.confirm.config.body':'<div class="alert alert-info">' +
    '<b>Review Configuration Changes.</b></br>' +
    'The following lists the configuration changes that will be made by the Wizard to enable NameNode HA. This information is for <b> review only </b> and is not editable except for the  <b>dfs.journalnode.edits.dir</b> property' +
    '</div>',
  'admin.highAvailability.wizard.step2.body':'Select a host that will be running the additional NameNode.<br/> In addition,' +
    ' select the hosts to run JournalNodes, which store NameNode edit logs in a fault tolerant manner.',
  'admin.highAvailability.wizard.step1.body':'This wizard will walk you through enabling NameNode HA on your cluster.<br/>' +
    'Once enabled, you will be running a Standby NameNode in addition to your Active NameNode.<br/>' +
    'This allows for an Active-Standby NameNode configuration that automatically performs failover.<br/><br/>' +
    'The process to enable HA involves a combination of <b>automated steps</b> (that will be handled by the wizard) and ' +
    '<b>manual steps</b> (that you must perform in sequence as instructed by the wizard).<br/><br/>' +
    '<b>You should plan a cluster maintenance window and prepare for cluster downtime when enabling NameNode HA.</b>',
  'admin.highAvailability.wizard.step1.alert':'If you have HBase running, please exit this wizard and stop HBase first.',

  'admin.rm_highAvailability.wizard.header': 'Enable ResourceManager HA Wizard',
  'admin.rm_highAvailability.wizard.step1.header': 'Get Started',
  'admin.rm_highAvailability.wizard.step1.body':'This wizard will walk you through enabling ResourceManager HA on your cluster.<br/>' +
      'Once enabled, you will be running a Standby ResourceManager in addition to your Active ResourceManager.<br/>' +
      'This allows for an Active-Standby ResourceManager configuration that automatically performs failover.<br/><br/>' +
      '<b>You should plan a cluster maintenance window and prepare for cluster downtime when enabling ResourceManager HA.</b>',
  'admin.rm_highAvailability.wizard.step2.header': 'Select Host',
  'admin.rm_highAvailability.wizard.step2.body': 'Select a host that will be running the additional ResourceManager',
  'admin.rm_highAvailability.wizard.step3.header': 'Review',
  'admin.rm_highAvailability.wizard.step3.confirm.host.body':'<b>Confirm your host selections.</b>',
  'admin.rm_highAvailability.wizard.step3.confirm.config.body':'<div class="alert alert-info">' +
      '<b>Review Configuration Changes.</b></br>' +
      'The following lists the configuration changes that will be made by the Wizard to enable ResourceManager HA. This information is for <b> review only </b> and is not editable.' +
      '</div>',
  'admin.rm_highAvailability.wizard.step3.currentRM': 'Current ResourceManager',
  'admin.rm_highAvailability.wizard.step3.additionalRM': 'Additional ResourceManager',
  'admin.rm_highAvailability.wizard.step4.header': 'Configure Components',
  'admin.rm_highAvailability.wizard.step4.task0.title': 'Stop Required Services',
  'admin.rm_highAvailability.wizard.step4.task1.title': 'Install Additional ResourceManager',
  'admin.rm_highAvailability.wizard.step4.task2.title': 'Reconfigure YARN',
  'admin.rm_highAvailability.wizard.step4.task3.title': 'Start All Services',
  'admin.rm_highAvailability.wizard.step4.notice.inProgress':'Please wait while ResourceManager HA is being deployed.',
  'admin.rm_highAvailability.wizard.step4.notice.completed':'ResourceManager HA has been enabled successfully.',

  'admin.security.title':'Kerberos security has not been enabled',
  'admin.security.enabled': 'Kerberos security is enabled',
  'admin.security.disabled': 'Kerberos security is disabled',
  'admin.security.button.enable':'Enable Security',
  'admin.security.button.disable':'Disable Security',
  'admin.security.enable.popup.body': 'We will walk you through add security wizard',
  'admin.security.enable.popup.header': 'Add security',
  'admin.security.disable.popup.header': 'Remove security',
  'admin.security.disable.popup.body': 'You are about to disable Kerberos on the cluster. This will stop all the services in your cluster and remove the Kerberos configurations. <strong>Are you sure you wish to proceed with disabling Kerberos?</strong>',
  'admin.addSecurity.header': 'Enable Security Wizard',
  'admin.security.step1.header': 'Get Started',
  'admin.security.step2.header': 'Configure Services',
  'admin.security.step3.header': 'Create Principals and Keytabs',
  'admin.security.step4.header': 'Save and Apply Configuration',
  'admin.security.step1.body.header': 'Important: Before configuring Ambari to manage your Kerberos-enabled cluster, ' +
    'you must perform the following manual steps on your cluster. Be sure to record the location of the keytab files ' +
    'for each host and the principals for each Hadoop service. This information is required in order to use the wizard.',
  'admin.security.step1.body.instruction1': 'Install, configure and start your Kerberos KDC',
  'admin.security.step1.body.instruction2': 'Install and configure the Kerberos client on every host in the cluster',
  'admin.security.step1.body.instruction3': 'Create Kerberos principals for Hadoop services and hosts',
  'admin.security.step1.body.instruction4': 'Generate keytabs for each principal and place on the appropriate hosts',
  'admin.security.step1.body.instruction5': '<b>Application Timeline Server</b> component of YARN service will be <span class="text-error"><b>deleted</b></span> as part of enabling security in this HDP stack version',
  'admin.security.step2.body.header': 'Configure Kerberos security properties',
  'admin.security.step3.notice': 'You need to create Kerberos principals and keytabs before proceeding.<br />'+
  'Download the CSV file and use it to create a script to generate the principals and keytabs on specified hosts. ' +
  'Once the principals and keytabs have been created, click on <i>Apply</i> to continue. If you need to make configuration changes, click <i>Back</i>.',
  'admin.security.step3.table.principal': 'Principal',
  'admin.security.step3.table.keytab': 'Keytab',
  'admin.security.step3.downloadCSV': 'Download CSV',
  'admin.security.step4.body.header': 'Applying kerberos security to the cluster',
  'admin.security.step4.body.success.header' : 'Kerberos-based security has been enabled on your cluster. Please wait while services are started in secure mode.',
  'admin.security.step4.body.failure.header' : 'Failed to enable Kerberos-based security on your cluster. Your cluster will keep running in non-secure mode.',
  'admin.security.step4.save.configuration.note':'This configuration is created by Enable/Disable security wizard',
  'admin.security.apply.configuration.error' : 'Failed to apply secure configurations to the cluster. Please navigate to "Configure Services" step and make sure all services are configured with appropriate values.',
  'admin.security.disable.body.header' : 'Disabling kerberos security on the cluster',
  'admin.security.disable.body.success.header': 'Kerberos-based security has been disabled on your cluster. Please wait while services are started in non-secure mode.',
  'admin.security.disable.body.failure.header': 'Failed to disable Kerberos-based security on your cluster. Your cluster will keep running in secure mode.',
  'admin.addSecurity.apply.stop.services': 'Stop Services',
  'admin.addSecurity.apply.save.config': 'Save Configurations',
  'admin.addSecurity.apply.start.services': 'Start Services',
  'admin.addSecurity.apply.delete.ats': 'Delete ATS',
  'admin.addSecurity.user.smokeUser': 'Ambari Smoke Test User',
  'admin.addSecurity.user.hdfsUser': 'HDFS User',
  'admin.addSecurity.user.hbaseUser': 'HBase User',
  'admin.addSecurity.user.stormUser': 'Storm User',
  'admin.addSecurity.hdfs.user.httpUser': 'HDFS SPNEGO User',
  'admin.addSecurity.rm.user.httpUser': 'ResourceManager SPNEGO User',
  'admin.addSecurity.nm.user.httpUser': 'NodeManager SPNEGO User',
  'admin.addSecurity.historyServer.user.httpUser': 'History server SPNEGO User',
  'admin.addSecurity.webhcat.user.httpUser': 'WebHCat SPNEGO User',
  'admin.addSecurity.hive.user.httpUser': 'Hive SPNEGO User',
  'admin.addSecurity.oozie.user.httpUser': 'Oozie SPNEGO User',
  'admin.addSecurity.falcon.user.httpUser': 'Falcon SPNEGO User',
  'admin.addSecurity.storm.user.httpUser': 'Storm UI Server',
  'admin.addSecurity.user.yarn.atsHTTPUser': 'YARN ATS HTTP User',
  'admin.addSecurity.knox.user': 'Knox Gateway',
  'admin.addSecurity.enable.onClose': 'You are in the process of enabling security on your cluster. ' +
    'Are you sure you want to quit? If you quit, ' +
    'you may have to re-run the security wizard from the beginning to enable security.',
  'admin.addSecurity.enable.after.stage2.onClose': 'Services are being started with the Kerberos settings you specified. '+
    'It is recommended that you wait until all the services are started to ensure that they are set up properly. ' +
    'Are you sure you want to quit?',
  'admin.addSecurity.disable.onClose': 'You are in the process of disabling security on your cluster. ' +
    'Are you sure you want to quit?',
  'admin.removeSecurity.header': 'Disable Security',
  'admin.security.applying.config.header': 'Applying Configurations',
  'admin.security.applying.config.body':'You cannot quit wizard while configurations are being applied',
  'admin.security.status.error' : 'Error in retrieving cluster security status from Ambari server',
  'admin.users.ldapAuthUsed':'LDAP Authentication is being used to authenticate users',
  'admin.users.delete.yourself.message':'You cannot delete yourself',
  'admin.users.delete.yourself.header':'Deleting warning',

  'admin.users.delete.header':'Delete {0}',

  'admin.users.addButton':'Add Local User',
  'admin.users.editButton': 'Edit Local User',
  'admin.users.delete':'delete',
  'admin.users.edit':'edit',
  'admin.users.privileges':'Admin',
  'admin.users.type':'Type',
  'admin.users.action':'Action',
  'admin.users.passwordRetype':'Retype Password',
  'admin.users.username':'Username',
  'admin.users.createSuccess': 'User successfully created.',
  'admin.users.createError': 'Error occurred while user creating.',
  'admin.users.createError.passwordValidation': 'Passwords are different',
  'admin.users.createError.userNameExists': 'User with the same name is already exists',
  'admin.users.editError.requiredField': 'This is required',

  'admin.access.showJobs':'Enable Jobs tab for non-admin users',

  'admin.confirmUninstall':'Confirm Uninstall',
  'admin.cluster.stackVersion':'Cluster Stack Version',
  'admin.cluster.upgradeAvailable':'Upgrade available',
  'admin.cluster.upgradeUnavailable':'Upgrade unavailable',
  'admin.cluster.repositories.repositories':'Repositories',
  'admin.cluster.repositories.os':'OS',
  'admin.cluster.repositories.baseUrl':'Base URL',
  'admin.cluster.repositories.popup.header.success':'Repo Base URLs Saved',
  'admin.cluster.repositories.popup.header.fail':'Base URL Validation Failed',
  'admin.cluster.repositories.popup.body.success':'The Repo Base URL has been saved successfully.',
  'admin.cluster.repositories.popup.body.fail':'The Base URL failed validation and has not been saved.',
  'admin.cluster.repositories.invalidURLAttention': '<b>Attention:</b> Please make sure all repository URLs are valid before saving.',
  'admin.cluster.repositories.emptyURLattention':'<b>Attention:</b> Repository URLs are REQUIRED before you can save.',
  'admin.cluster.repositories.skipValidation.tooltip':'<b>Warning:</b> This is for advanced users only. Use this option if you want to skip validation for Repository Base URLs.',

  'admin.misc.header': 'Service Users and Groups',
  'admin.misc.nothingToShow': 'No user accounts to display',

  'admin.stackVersions.filter.notInstalled': "Not Installed ({0})",
  'admin.stackVersions.filter.all': "All ({0})",
  'admin.stackVersions.filter.upgradeReady': "Upgrade Ready ({0})",
  'admin.stackVersions.filter.installed': "Installed ({0})",
  'admin.stackVersions.filter.current': "Current ({0})",
  'admin.stackVersions.filter.upgrading': "Upgrade/Downgrade In Process ({0})",
  'admin.stackVersions.filter.upgraded': "Ready to Finalize ({0})",

  'admin.stackVersions.editRepositories.info': 'Provide Base URLs for the Operating Systems you are configuring. Uncheck all other Operating Systems.',
  'admin.stackVersions.editRepositories.validation.warning': 'Some of the repositories failed validation. Make changes to the base url or skip validation if you are sure that urls are correct',
  'admin.stackVersions.version.install.confirm': 'You are about to install packages for version <strong>{0}</strong> on all hosts.',
  'admin.stackVersions.version.linkTooltip': 'Click to Edit Repositories',
  'admin.stackVersions.version.hostsTooltip': 'Click to List Hosts',
  'admin.stackVersions.version.emptyHostsTooltip': 'No Hosts to List',
  'admin.stackVersions.version.notInstalled': "Not Installed",
  'admin.stackVersions.manageVersions': "Manage Versions",
  'admin.stackVersions.manageVersions.popup.body': 'You are about to leave the <b>Cluster Management</b> interface' +
    ' and go to the <b>Ambari Administration</b> interface. You can return to cluster management by using the' +
    ' “Go to Dashboard” link in the Ambari Administration > Clusters section.',
  'admin.stackVersions.version.installNow': "Install Packages",
  'admin.stackVersions.version.performUpgrade': "Perform Upgrade",
  'admin.stackVersions.version.upgrade.pause': "Upgrade: Action Required",
  'admin.stackVersions.version.upgrade.running': "Upgrade: In Process",
  'admin.stackVersions.version.downgrade.pause': "Downgrade: Action Required",
  'admin.stackVersions.version.downgrade.running': "Downgrade: In Process",

  'admin.stackVersions.hosts.popup.header.current': "Current",
  'admin.stackVersions.hosts.popup.header.installed': "Installed",
  'admin.stackVersions.hosts.popup.header.not_installed': "Not installed",
  'admin.stackVersions.hosts.popup.header': "Version Status: {0}",
  'admin.stackVersions.hosts.popup.title': "{0} Version is {1} on {2} hosts:",
  'admin.stackVersions.hosts.popup.primary': "Go to Hosts",

  'admin.stackVersions.details.install.hosts.popup.title': "Install {0} version",

  'admin.stackUpgrade.finalize.later': "Finalize Later",
  'admin.stackUpgrade.finalize.message.upgrade': "Your cluster version has been upgraded. " +
  "Click on <b>Finalize</b> when you are ready to finalize the upgrade and commit to the new version." +
  " You are strongly encouraged to run tests on your cluster to ensure it is fully operational before finalizing." +
  " <b>You cannot go back to the original version once the upgrade is finalized.</b>",
  'admin.stackUpgrade.finalize.message.downgrade': "Your cluster version has been downgraded. " +
    "Click on <b>Finalize</b> when you are ready to finalize the downgrade and commit to the new version." +
    " You are strongly encouraged to run tests on your cluster to ensure it is fully operational before finalizing." +
    " <b>You cannot go back to the original version once the downgrade is finalized.</b>",
  'admin.stackUpgrade.doThisLater': "Do This Later",
  'admin.stackUpgrade.pauseUpgrade': "Pause Upgrade",
  'admin.stackUpgrade.downgrade.proceed': "Proceed with Downgrade",
  'admin.stackUpgrade.downgrade.body': "Are you sure you wish to abort the upgrade process and downgrade to <b>{0}</b>?",
  'admin.stackUpgrade.upgrade.confirm.body': "You are about to perform an upgrade to {0}.",
  'admin.stackUpgrade.title': "Stack and Versions",
  'admin.stackUpgrade.state.inProgress': "Upgrade in Progress",
  'admin.stackUpgrade.state.paused': "Upgrade Paused",
  'admin.stackUpgrade.state.completed': "Upgrade Finished",
  'admin.stackUpgrade.state.inProgress.downgrade': "Downgrade in Progress",
  'admin.stackUpgrade.state.paused.downgrade': "Downgrade Paused",
  'admin.stackUpgrade.state.completed.downgrade': "Downgrade Finished",
  'admin.stackUpgrade.dialog.header': "Rolling Upgrade to {0}",
  'admin.stackUpgrade.dialog.downgrade.header': "Downgrade to {0}",
  'admin.stackUpgrade.dialog.operationFailed': "This operation failed.",
  'admin.stackUpgrade.dialog.stop': "Stop Upgrade",
  'admin.stackUpgrade.dialog.continue': "Ignore and Proceed",
  'admin.stackUpgrade.dialog.cancel': "Cancel Upgrade",
  'admin.stackUpgrade.dialog.inProgress': "Now Running:",
  'admin.stackUpgrade.dialog.keepRunning': "Keep running Upgrade in background",
  'admin.stackUpgrade.dialog.failed': "Failed on:",
  'admin.stackUpgrade.dialog.manual': "Manual steps required",
  'admin.stackUpgrade.dialog.manualDone': "I have performed the manual steps above.",
  'admin.stackUpgrade.dialog.closeProgress': "Upgrade is in progress. \n If you dismiss this window, Upgrade will keep running in background.",
  'admin.stackUpgrade.dialog.closePause': "Upgrade is paused. \n If you dismiss this window, you can resume Upgrade later.",
  'admin.stackUpgrade.dialog.details.open': "show details",
  'admin.stackUpgrade.dialog.details.hide': "hide details",
  'admin.stackUpgrade.dialog.notActive': "Waiting to execute the next task...",
  'services.service.start':'Start',
  'services.service.stop':'Stop',
  'services.service.metrics':'Metrics',
  'services.nothingToAdd':'Nothing to add',
  'services.service.summary.version':'Version',
  'services.service.summary.viewHost':'View Host',
  'services.service.summary.viewHosts':'View Hosts',
  'services.service.summary.clientOnlyService.ToolTip':'Client-only service',
  'services.service.summary.DataNodesLive':'DataNodes Live',
  'services.service.summary.JournalNodesLive':'JournalNodes Live',
  'services.service.summary.mapreduce2.client':'MapReduce2 Client',
  'services.service.summary.mapreduce2.clients':'MapReduce2 Clients',
  'services.service.summary.nodeManagersLive':'NodeManagers Live',
  'services.service.summary.TrackersLive':'Trackers Live',
  'services.service.summary.RegionServersLIVE':'RegionServers Live',
  'services.service.summary.GangliaMonitorsLIVE':'Ganglia Monitors Live',
  'services.service.summary.SupervisorsLIVE':'Supervisors Live',
  'services.service.summary.nameNode':'NameNode Web UI',
  'services.service.summary.nameNodeUptime':'NameNode Uptime',
  'services.service.summary.nameNodeHeap':'NameNode Heap',
  'services.service.summary.nameNode.active':'Active NameNode',
  'services.service.summary.nameNode.standby':'Standby NameNode',
  'services.service.summary.pendingUpgradeStatus':'Upgrade Status',
  'services.service.summary.pendingUpgradeStatus.notFinalized':'Upgrade not finalized',
  'services.service.summary.pendingUpgradeStatus.notPending':'No pending upgrade',
  'services.service.summary.safeModeStatus':'Safe Mode Status',
  'services.service.summary.safeModeStatus.inSafeMode':'In safe mode',
  'services.service.summary.safeModeStatus.notInSafeMode':'Not in safe mode',
  'services.service.summary.dataNodes':'DataNodes',
  'services.service.summary.diskCapacity':'HDFS Disk Capacity',
  'services.service.summary.blocksTotal':'Blocks (total)',
  'services.service.summary.blockErrors':'Block Errors',
  'services.service.summary.totalFiles':'Total Files + Dirs',
  'services.service.summary.jobTracker':'JobTracker',
  'services.service.summary.jobTrackerWebUI':'JobTracker Web UI',
  'services.service.summary.jobTrackerUptime':'JobTracker Uptime',
  'services.service.summary.trackersLiveTotal':'Trackers',
  'services.service.summary.trackersBlacklistGraylist':'Trackers',
  'services.service.summary.jobTrackerHeap':'JobTracker Heap',
  'services.service.summary.totalSlotsCapacity':'Total Slots Capacity',
  'services.service.summary.totalJobs':'Total Jobs',
  'services.service.summary.currentSlotUtiliMaps':'Map Slots',
  'services.service.summary.currentSlotUtiliReduces':'Reduce Slots',
  'services.service.summary.tasksMaps':'Tasks: Maps',
  'services.service.summary.tasksReduces':'Tasks: Reduces',
  'services.service.summary.hbaseMaster':'HBase Master Web UI',
  'services.service.summary.regionServerCount':'RegionServer Count',
  'services.service.summary.regionInTransition':'Region In Transition',
  'services.service.summary.masterStarted':'Master Started',
  'services.service.summary.masterActivated':'Master Activated',
  'services.service.summary.averageLoad':'Average Load',
  'services.service.summary.masterHeap':'Master Heap',
  'services.service.summary.moreStats':'more stats here',
  'services.service.summary.clientCount': '{0} Client Hosts',
  'services.service.summary.historyServer': 'History Server Web UI',
  'services.service.actions.downloadClientConfigs':'Download Client Configs',
  'services.service.actions.downloadClientConfigs.fail.noConfigFile':'No configuration files defined for the component',
  'services.service.actions.downloadClientConfigs.fail.popup.header':'{0} Configs',
  'services.service.actions.downloadClientConfigs.fail.popup.body.noErrorMessage':'Generation of {0} configurations file has failed. ',
  'services.service.actions.downloadClientConfigs.fail.popup.body.errorMessage':'Generation of {0} configurations file has failed with {1} error: <br /><pre><span class="text-error">{2}</span></pre>',
  'services.service.actions.downloadClientConfigs.fail.popup.body.question':'Do you want to try again?',
  'services.service.actions.run.rebalancer':'Run Rebalancer',
  'services.service.actions.run.rebalanceHdfsNodes':'Rebalance HDFS',
  'services.service.actions.run.rebalanceHdfsNodes.title':'HDFS Rebalance',
  'services.service.actions.run.rebalanceHdfsNodes.prompt':'Balancer threshold (percentage of disk capacity):',
  'services.service.actions.run.rebalanceHdfsNodes.promptTooltip':'Percentage of disk capacity. This overwrites the default threshold',
  'services.service.actions.run.rebalanceHdfsNodes.promptError':'Value should be number between 1 and 100',
  'services.service.actions.run.rebalanceHdfsNodes.context':'Rebalance HDFS',
  'services.service.actions.run.rebalanceHdfsNodes.error':'Error during remote command: ',
  'services.service.actions.run.yarnRefreshQueues.menu':'Refresh YARN Capacity Scheduler',
  'services.service.actions.run.yarnRefreshQueues.context':'Refresh YARN Capacity Scheduler',
  'services.service.actions.run.yarnRefreshQueues.error':'Error during remote command: ',
  'services.service.actions.run.executeCustomCommand.menu':'{0}',
  'services.service.actions.run.executeCustomCommand.context':'Execute {0}',
  'services.service.actions.run.executeCustomCommand.error':'Error during remote command: ',
  'services.service.actions.run.compaction':'Run Compaction',
  'services.service.actions.run.smoke':'Run Service Check',
  'services.service.actions.reassign.master':'Move {0}',
  'services.service.actions.reassign.master.hive':'Move HiveServer2, WebHCat Server, MySQL Server',
  'services.service.actions.manage_configuration_groups':'Manage Configuration Groups...',
  'services.service.actions.run.startLdapKnox.context':'Start Demo LDAP',
  'services.service.actions.run.stopLdapKnox.context':'Stop Demo LDAP',
  'services.service.actions.run.startStopLdapKnox.error': 'Error during remote command: ',
  'services.service.actions.manage_configuration_groups.short':'Manage Config Groups',
  'services.service.actions.serviceActions':'Service Actions',
  'services.service.summary.unknown':'unknown',
  'services.service.summary.notRunning':'Not Running',
  'services.service.summary.notAvailable':'n/a',
  'services.service.summary.diskInfoBar.used':'used',
  'services.service.summary.storm.freeslots': 'Free slots',
  'services.service.summary.storm.executors': 'Executors',
  'services.service.summary.storm.tasks': 'Tasks',
  'services.service.summary.storm.nimbus.uptime': 'Nimbus uptime',
  'services.service.summary.storm.topologies': 'Topologies',
  'services.service.summary.flume.startAgent': 'Start Agent',
  'services.service.summary.flume.stopAgent': 'Stop Agent',
  'services.service.summary.flume.stop.context': 'Stop Flume {0}',
  'services.service.summary.flume.start.context': 'Start Flume {0}',
  'services.service.summary.flume.noAgents': 'No Flume to display',

  'services.service.summary.ranger.plugin.title': 'Ranger {0} plugin',
  'services.service.summary.ranger.plugin.loadingStatus': 'Loading status...',

  'services.service.summary.alerts.noAlerts': 'No alerts',
  'services.service.summary.alerts.alertsExist': '{0} alerts',
  'services.service.summary.alerts.popup.header': 'Alerts for {0}',

  'services.service.info.metrics.flume.channelFillPercent':'Channel Fill Percentage',
  'services.service.info.metrics.flume.channelSize':'Channel Size',
  'services.service.info.metrics.flume.sinkDrainSuccess':'Sink Event Drain Count',
  'services.service.info.metrics.flume.sourceAccepted':'Source Event Accepted Count',
  'services.service.info.metrics.flume.sinkConnectionFailed':'Sink Connection Failed Count',
  'services.service.info.metrics.flume.channelSizeMMA':'Channel Size Events',
  'services.service.info.metrics.flume.channelSizeSum':'Channel Size Event Sum',
  'services.service.info.metrics.flume.incoming.mma':'Incoming Event Rate',
  'services.service.info.metrics.flume.incoming.sum':'Incoming Event Sum',
  'services.service.info.metrics.flume.outgoing.mma':'Outgoing Event Rate',
  'services.service.info.metrics.flume.outgoing.sum':'Outgoing Event Sum',
  'services.service.info.metrics.flume.gc':'Garbage Collection Time',
  'services.service.info.metrics.flume.cpu.user':'CPU (User)',
  'services.service.info.metrics.flume.jvmThreadsRunnable':'JVM Runnable Threads',
  'services.service.info.metrics.flume.jvmHeapUsed':'JVM Heap Memory Used',
  'services.service.info.metrics.flume.channelType':'Channel size {0}',
  'services.service.info.metrics.flume.incoming_mma':'Incoming {0}',
  'services.service.info.metrics.flume.outgoing_mma':'Outgoing {0}',
  'services.service.info.metrics.flume.sourceName':'Source {0}',
  'services.service.info.metrics.flume.hostName':'Host: {0}',
  'services.service.info.metrics.flume.channelName':'Channel {0}',
  'services.service.info.metrics.flume.sinkName':'Sink {0}',

  'services.service.info.metrics.hbase.clusterRequests':'Cluster Requests',
  'services.service.info.metrics.hbase.clusterRequests.displayNames.requestCount':'Request Count',
  'services.service.info.metrics.hbase.hlogSplitSize':'HLog Split Size',
  'services.service.info.metrics.hbase.hlogSplitSize.displayNames.splitSize':'Split Size',
  'services.service.info.metrics.hbase.hlogSplitTime':'HLog Split Time',
  'services.service.info.metrics.hbase.hlogSplitTime.displayNames.splitTime':'Split Time',
  'services.service.info.metrics.hbase.regionServerQueueSize':'RegionServer Queue Size',
  'services.service.info.metrics.hbase.regionServerQueueSize.displayNames.compactionQueueSize':'Compaction Queue Size',
  'services.service.info.metrics.hbase.regionServerQueueSize.displayNames.flushQueueSize':'Flush Queue Size',
  'services.service.info.metrics.hbase.regionServerRegions':'RegionServer Regions',
  'services.service.info.metrics.hbase.regionServerRegions.displayNames.regions':'Regions',
  'services.service.info.metrics.hbase.regionServerRequests':'Cumulative Requests',
  'services.service.info.metrics.hbase.regionServerRequests.2':'RegionServer Requests',
  'services.service.info.metrics.hbase.regionServerRequests.displayNames.writeRequests':'Write Requests',
  'services.service.info.metrics.hbase.regionServerRequests.displayNames.readRequests':'Read Requests',

  'services.service.info.metrics.hdfs.blockStatus':'Block Status',
  'services.service.info.metrics.hdfs.blockStatus.displayNames.pendingReplicationBlocks':'Pending Replication Blocks',
  'services.service.info.metrics.hdfs.blockStatus.displayNames.underReplicatedBlocks':'Under Replicated Blocks',
  'services.service.info.metrics.hdfs.fileOperations':'File Operations',
  'services.service.info.metrics.hdfs.fileOperations.displayNames.fileInformationOperations':'File Information Operations',
  'services.service.info.metrics.hdfs.fileOperations.displayNames.deleteFileOperations':'Delete File Operations',
  'services.service.info.metrics.hdfs.fileOperations.displayNames.createFileOperations':'Create File Operations',
  'services.service.info.metrics.hdfs.gc':'Garbage Collection',
  'services.service.info.metrics.hdfs.gc.displayNames.gcTimeMillis':'Time',
  'services.service.info.metrics.hdfs.io':'HDFS I/O',
  'services.service.info.metrics.hdfs.io.displayNames.bytesWritten':'Bytes Written',
  'services.service.info.metrics.hdfs.io.displayNames.bytesRead':'Bytes Read',
  'services.service.info.metrics.hdfs.jvmHeap':'JVM Memory Status',
  'services.service.info.metrics.hdfs.jvmHeap.displayNames.memHeapCommittedM':'Heap Memory Committed',
  'services.service.info.metrics.hdfs.jvmHeap.displayNames.memNonHeapUsedM':'Non Heap Memory Used',
  'services.service.info.metrics.hdfs.jvmHeap.displayNames.memHeapUsedM':'Heap Memory Used',
  'services.service.info.metrics.hdfs.jvmHeap.displayNames.memNonHeapCommittedM':'Non Heap Memory Committed',
  'services.service.info.metrics.hdfs.jvmThreads':'JVM Thread Status',
  'services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsBlocked':'Threads Blocked',
  'services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsWaiting':'Threads Waiting',
  'services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsTimedWaiting':'Threads Timed Waiting',
  'services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsRunnable':'Threads Runnable',
  'services.service.info.metrics.hdfs.rpc':'RPC',
  'services.service.info.metrics.hdfs.rpc.displayNames.rpcQueueTimeAvgTime':'Queue Average Wait Time',
  'services.service.info.metrics.hdfs.spaceUtilization':'Total Space Utilization',
  'services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityRemainingGB':'Capacity Remaining',
  'services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityUsedGB':'Capacity Used',
  'services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityTotalGB':'Capacity Total',
  'services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityRemaining':'Capacity Remaining',
  'services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityUsed':'Capacity Used',
  'services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityTotal':'Capacity Total',
  'services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityNonDFSUsed':'Non DFS Capacity Used',

  'services.service.info.metrics.yarn.jvmHeap':'JVM Memory Status',
  'services.service.info.metrics.yarn.jvmHeap.displayNames.memHeapCommittedM':'Heap Memory Committed',
  'services.service.info.metrics.yarn.jvmHeap.displayNames.memNonHeapUsedM':'Non Heap Memory Used',
  'services.service.info.metrics.yarn.jvmHeap.displayNames.memHeapUsedM':'Heap Memory Used',
  'services.service.info.metrics.yarn.jvmHeap.displayNames.memNonHeapCommittedM':'Non Heap Memory Committed',
  'services.service.info.metrics.yarn.jvmThreads':'JVM Thread Status',
  'services.service.info.metrics.yarn.jvmThreads.displayNames.threadsBlocked':'Threads Blocked',
  'services.service.info.metrics.yarn.jvmThreads.displayNames.threadsWaiting':'Threads Waiting',
  'services.service.info.metrics.yarn.jvmThreads.displayNames.threadsTimedWaiting':'Threads Timed Waiting',
  'services.service.info.metrics.yarn.jvmThreads.displayNames.threadsRunnable':'Threads Runnable',
  'services.service.info.metrics.yarn.rpc':'RPC',
  'services.service.info.metrics.yarn.rpc.displayNames.RpcQueueTimeAvgTime':'Queue Average Wait Time',
  'services.service.info.metrics.yarn.gc': 'Garbage Collection',
  'services.service.info.metrics.yarn.gc.displayNames.gcTimeMillis':'Time',
  'services.service.info.metrics.yarn.allocated.memory': 'Cluster Memory',
  'services.service.info.metrics.yarn.allocated.memory.displayNames.allocated': 'Allocated',
  'services.service.info.metrics.yarn.allocated.memory.displayNames.available': 'Available',
  'services.service.info.metrics.yarn.allocated.memory.displayNames.pending': 'Pending',
  'services.service.info.metrics.yarn.allocated.container': 'Containers',
  'services.service.info.metrics.yarn.allocated.container.displayNames.allocated': 'Allocated',
  'services.service.info.metrics.yarn.allocated.container.displayNames.reserved': 'Reserved',
  'services.service.info.metrics.yarn.allocated.container.displayNames.pending': 'Pending',
  'services.service.info.metrics.yarn.nodemanager.statuses':'NodeManagers',
  'services.service.info.metrics.yarn.nodemanager.statuses.displayNames.active':'Active NodeManagers',
  'services.service.info.metrics.yarn.nodemanager.statuses.displayNames.decommissioned':'Decommissioned NodeManagers',
  'services.service.info.metrics.yarn.nodemanager.statuses.displayNames.lost':'Lost NodeManagers',
  'services.service.info.metrics.yarn.nodemanager.statuses.displayNames.rebooted':'Rebooted NodeManagers',
  'services.service.info.metrics.yarn.nodemanager.statuses.displayNames.unhealthy':'Unhealthy NodeManagers',
  'services.service.info.metrics.yarn.queueMemoryResource':'Queue Memory Used',
  'services.service.info.metrics.yarn.queueMemoryResource.displayNames.allocated':'Allocated ({0})',
  'services.service.info.metrics.yarn.queueMemoryResource.displayNames.available':'Available ({0})',
  'services.service.info.metrics.yarn.queueMemoryResource.displayName':'Queue: {0}',
  'services.service.info.metrics.yarn.apps.states.current.title': 'Current Applications',
  'services.service.info.metrics.yarn.apps.states.finished.title': 'Finished Applications',
  'services.service.info.metrics.yarn.apps.states.completed': 'Completed',
  'services.service.info.metrics.yarn.apps.states.failed': 'Failed',
  'services.service.info.metrics.yarn.apps.states.killed': 'Killed',
  'services.service.info.metrics.yarn.apps.states.pending': 'Pending',
  'services.service.info.metrics.yarn.apps.states.running': 'Running',
  'services.service.info.metrics.yarn.apps.states.submitted': 'Submitted',

  'services.service.info.metrics.kafka.server.brokerTopic.title': 'Broker Topics',
  'services.service.info.metrics.kafka.server.brokerTopic.displayNames.AllTopicsBytesOutPerSec': 'Bytes Out',
  'services.service.info.metrics.kafka.server.brokerTopic.displayNames.AllTopicsBytesInPerSec': 'Bytes In',
  'services.service.info.metrics.kafka.server.brokerTopic.displayNames.AllTopicsMessagesInPerSec': 'Messages In',
  'services.service.info.metrics.kafka.server.ReplicaManager.title': 'Replica Manager',
  'services.service.info.metrics.kafka.server.ReplicaManager.displayNames.PartitionCount': 'Partiotions count',
  'services.service.info.metrics.kafka.server.ReplicaManager.displayNames.UnderReplicatedPartitions': 'Under Replicated Partiotions',
  'services.service.info.metrics.kafka.server.ReplicaManager.displayNames.LeaderCount': 'Leader Count',
  'services.service.info.metrics.kafka.controller.ControllerStats.title': 'Controller Status',
  'services.service.info.metrics.kafka.controller.ControllerStats.displayNames.LeaderElectionRateAndTimeMs': 'Leader Election Rate And Time',
  'services.service.info.metrics.kafka.controller.ControllerStats.displayNames.UncleanLeaderElectionsPerSec': 'Unclean Leader Election',
  'services.service.info.metrics.kafka.controller.KafkaController.title': 'Active Controller Count',
  'services.service.info.metrics.kafka.controller.KafkaController.displayNames.ActiveControllerCount': 'Active Controller Count',
  'services.service.info.metrics.kafka.log.LogFlushStats.title': 'Log Flush Status',
  'services.service.info.metrics.kafka.log.LogFlushStats.displayNames.LogFlushRateAndTimeMs': 'Log Flush Rate amd Time',
  'services.service.info.metrics.kafka.server.ReplicaFetcherManager.title': 'Replica MaxLag',
  'services.service.info.metrics.kafka.server.ReplicaFetcherManager.displayNames.Replica-MaxLag': 'Replica MaxLag',

  'services.service.info.menu.summary':'Summary',
  'services.service.info.menu.configs':'Configs',
  'services.service.info.summary.hostsRunningMonitor':'{0}/{1}',
  'services.service.info.summary.serversHostCount':'{0} more',

  'services.service.config.final':'Final',
  'services.service.config.saved':'Save Configuration Changes',
  'services.service.config.notSaved':'Unable to Save Configuration Changes',
  'services.service.config.restartService.TooltipMessage':'<b>Restart Service</b><br>Stale configuration used by {0} components on {1} hosts:{2}',
  'services.service.config.restartService.needToRestart':'<strong>Restart Required:</strong> ',
  'services.service.config.restartService.needToRestartEnd':'should be restarted',
  'service.service.config.restartService.shouldBeRestarted':'{0} Requiring Restart',
  'services.service.config.saved.message':'Service configuration changes saved successfully.',
  'services.service.config.saving.message':'Configuration changes are being saved...',
  'services.service.config.msgServiceStop':'Could not save configuration changes.  Please stop the service first. You will be able to save configuration changes after all of its components are stopped.',
  'services.service.config.failCreateConfig' : 'Failure in creating service configuration',
  'services.service.config.failSaveConfig':'Failure in saving service configuration',
  'services.service.config.failSaveConfigHostOverrides':'Failure in saving service configuration overrides',
  'services.service.config.addPropertyWindow.error.required':'This is required',
  'services.service.config.addPropertyWindow.error.derivedKey':'This property is already defined',
  'services.service.config.addPropertyWindow.error.derivedKey.specific':'Property "{0}" is already defined',
  'services.service.config.addPropertyWindow.error.format':'Key and value should be separated by "=" sign',
  'services.service.config.addPropertyWindow.error.lineNumber':'Line {0}: ',
  'services.service.config.addPropertyWindow.filterKeyLink' : 'Find property',
  'services.service.config.addPropertyWindow.properties' : 'Properties',
  'services.service.config.addPropertyWindow.propertiesHelper' : 'key=value (one per line)',
  'services.service.config.addPropertyWindow.propertiesPlaceholder' : 'Enter key=value (one per line)',
  'services.service.config.addPropertyWindow.bulkMode' : 'Bulk property add mode',
  'services.service.config.addPropertyWindow.singleMode' : 'Single property add mode',
  'services.service.config.stopService.runningHostComponents':'{0} components on {1} hosts are running',
  'services.service.config.stopService.unknownHostComponents':'{0} components on {1} hosts are in unknown state.  These components might actually be running and need restarting later to make the changes effective.',
  'services.service.config.confirmDirectoryChange':'You are about to make changes to service directories that are core to {0}. Before you proceed, be absolutely certain of the implications and that you have taken necessary manual steps, if any, for the changes. Are you sure you want to proceed?',
  'services.service.config.configOverride.head':'Config Override',
  'services.service.config.configOverride.body':'Cannot override a config that has not been saved yet.',
  'services.service.config.exitPopup.body':'You have unsaved changes. Save changes or discard?',
  'services.service.config.propertyFilterPopover.title':'Properties filter',
  'services.service.config.propertyFilterPopover.content':'Enter keywords to filter properties by property name, value, or description.',
  'services.service.config.hive.oozie.postgresql': 'Existing PostgreSQL Database',
  'services.service.config.database.connection.success': 'Connection OK',
  'services.service.config.database.connection.inProgress': 'Checking connectivity',
  'services.service.config.database.connection.failed': 'Connection Failed',
  'services.service.config.connection.exitPopup.msg': 'Test connection is in progress. It\'s recommended to wait until it wil be complete. Are you sure you want to exit Enable Kerberos Wizard?',
  'services.service.config.database.btn.idle': 'Test Connection',
  'services.service.config.kdc.btn.idle': 'Test KDC Connection',
  'services.service.config.database.btn.connecting': 'Connecting...',
  'services.service.config.database.msg.jdbcSetup': 'Be sure you have run:<br/>' +
    '<b>ambari-server setup --jdbc-db={0} --jdbc-driver=/path/to/{1}/driver.jar</b> ' +
    'on the Ambari Server host to make the JDBC driver available and to enable testing the database connection.',

  'services.service.config.configHistory.configGroup': 'Config Group',
  'services.service.config.configHistory.rightArrow.tooltip': 'Show earlier versions',
  'services.service.config.configHistory.leftArrow.tooltip': 'Show later versions',
  'services.service.config.configHistory.dismissIcon.tooltip': 'Dismiss',
  'services.service.config.configHistory.makeCurrent.message': 'Created from service config version {0}',
  'services.service.config.configHistory.comparing': 'Comparing',

  'services.add.header':'Add Service Wizard',
  'services.reassign.header':'Move Master Wizard',
  'services.service.add':'Add Service',
  'services.service.startAll':'Start All',
  'services.service.stopAll':'Stop All',
  'services.service.startAll.confirmMsg' : 'You are about to start all services',
  'services.service.stopAll.confirmMsg' : 'You are about to stop all services',
  'services.service.start.confirmMsg' : 'You are about to start {0}',
  'services.service.stop.confirmMsg' : 'You are about to stop {0}',
  'services.service.stop.confirmButton': 'Confirm Stop',
  'services.service.start.confirmButton' : 'Confirm Start',
  'services.service.stop.warningMsg.turnOnMM': 'This will generate alerts as the service is stopped. To suppress alerts, turn on Maintenance Mode for {0} prior to stopping',
  'services.service.restartAll.confirmButton': 'Confirm Restart All',
  'services.service.restartAll.confirmMsg': 'You are about to restart {0}',
  'services.service.restartAll.warningMsg.turnOnMM': 'This will trigger alerts as the service is restarted. To suppress alerts, turn on Maintenance Mode for {0} prior to running restart all',
  'services.service.config_groups_popup.header':'Manage {0} Configuration Groups',
  'services.service.config_groups_popup.notice':'You can apply different sets of {{serviceName}} configurations to groups of hosts by managing {{serviceName}} Configuration Groups and their host membership.  Hosts belonging to a {{serviceName}} Configuration Group have the same set of configurations for {{serviceName}}. Each host belongs to one {{serviceName}} Configuration Group.',
  'services.service.config_groups_popup.rename':'Rename',
  'services.service.config_groups_popup.duplicate':'Duplicate',
  'services.service.config_groups_popup.group_name_lable':'Name',
  'services.service.config_groups_popup.group_desc_lable':'Description',
  'services.service.config_groups_popup.properties':'Properties',
  'services.service.config_groups_popup.addButton':'Create new Configuration Group',
  'services.service.config_groups_popup.removeButton':'Delete Configuration Group',
  'services.service.config_groups_popup.renameButton':'Rename Configuration Group',
  'services.service.config_groups_popup.addHost':'Add hosts to selected Configuration Group',
  'services.service.config_groups_popup.addHostDisabled':'There are no available hosts to add.',
  'services.service.config_groups_popup.removeHost':'Remove hosts from selected Configuration Group',
  'services.service.config_groups_popup.duplicateButton':'Duplicate Configuration Group',
  'services.service.config_groups.add_config_group_popup.header':'Create New Configuration Group',
  'services.service.config_groups.duplicate_config_group_popup.header':'Duplicate Configuration Group',
  'services.service.config_groups.rename_config_group_popup.header':'Rename Configuration Group',
  'services.service.config_groups.switchGroupTextFull':'Switch to \'{0}\' host config group',
  'services.service.config_groups.switchGroupTextShort':'Switch to \'{0}\'',
  'services.reassign.closePopup':'Move {0} wizard is in progress. You must allow the wizard to complete for Ambari to be in usable state. If you choose to quit, you must follow manual instructions to complete or revert move {0} wizard as documented in the Ambari User Guide. Are you sure you want to exit the wizard?',
  'services.reassign.error.fewHosts':'You must have at least 2 hosts in your cluster to run Move Wizard.',

  'services.reassign.step1.header':'Get Started',
  'services.reassign.step1.message1': 'This wizard will walk you through moving {0}.<br/>',
  'services.reassign.step1.message2': 'The process to reassign {0} involves a combination of <b>automated steps</b> (that will be handled by the wizard) and ' +
      '<b>manual steps</b> (that you must perform in sequence as instructed by the wizard).<br/><br/>',
  'services.reassign.step1.message3': '<br/><b>All required services will be restarted as part of the wizard. You should plan a cluster maintenance window and prepare ' +
    'for cluster downtime when moving {0}.</b>',

  'services.reassign.step2.header':'Assign Master',
  'services.reassign.step2.currentHost':'Current:',
  'services.reassign.step2.body':'Assign {0} to new host.',
  'services.reassign.step2.body.namenodeHA':'Move {0} to new host. You can move only one master component at a time.',
  'services.reassign.step3.header':'Review',
  'services.reassign.step3.body':'Please review the changes you made',
  'services.reassign.step3.targetHost':'Target Host:',
  'services.reassign.step3.sourceHost':'Source Host:',
  'services.reassign.step3.component':'Component name:',
  'services.reassign.step4.header':'Configure Component',

  'services.reassign.step4.tasks.stopServices.title':'Stop Required Services',
  'services.reassign.step4.tasks.createHostComponents.title':'Create {0}',
  'services.reassign.step4.tasks.putHostComponentsInMaintenanceMode.title':'Disable {0}',
  'services.reassign.step4.tasks.reconfigure.title':'Reconfigure {0}',
  'services.reassign.step4.tasks.save.configuration.note':'This configuration is created by Move {0} wizard',
  'services.reassign.step4.tasks.installHostComponents.title':'Install {0}',
  'services.reassign.step4.tasks.startZooKeeperServers.title':'Start ZooKeeper Servers',
  'services.reassign.step4.tasks.startNameNode.title':'Start NameNode',
  'services.reassign.step4.tasks.deleteHostComponents.title':'Delete disabled {0}',
  'services.reassign.step4.tasks.startServices.title':'Start Required Services',
  'services.reassign.step4.tasks.cleanMySqlServer.title':'Clean MYSQL Server',
  'services.reassign.step4.tasks.testHiveMysqlConnection.title':'Test MYSQL Connection',
  'services.reassign.step4.tasks.configureMySqlServer.title':'Configure MYSQL Server',
  'services.reassign.step4.tasks.startMySqlServer.title':'Start MYSQL Server',
  'services.reassign.step4.tasks.testDBConnection.title':'Test DB Connection',
  'services.reassign.step4.tasks.testDBConnection.tooltip':'Database Host: {0}\n' +
  'Database Type: {1}\n' +
  'Database Name: {2}\n' +
  'Username: {3}\n' +
  'Password: {4}\n' +
  'JDBC Driver Class: {5}\n' +
  'Database URL: {6}',
  'services.reassign.rollback.confirm':'Are you sure?',


  'services.reassign.step4.task0.title':'Stop Required Services',
  'services.reassign.step4.task1.title':'Create {0}',
  'services.reassign.step4.task2.title':'Disable {0}',
  'services.reassign.step4.task3.title':'Reconfigure {0}',
  'services.reassign.step4.save.configuration.note':'This configuration is created by Move {0} wizard',
  'services.reassign.step4.task4.title':'Install {0}',
  'services.reassign.step4.task5.title':'Start ZooKeeper Servers',
  'services.reassign.step4.task6.title':'Start NameNode',
  'services.reassign.step4.task7.title':'Delete disabled {0}',
  'services.reassign.step4.task8.title':'Start Required Services',
  'services.reassign.step4.tasks.startNewMySqlServer.title':'Start New MYSQL Server',
  'services.reassign.step4.status.success': 'Successfully moved <b>{0}</b> from <b>{1}</b> host to <b>{2}</b> host',
  'services.reassign.step4.status.success.withManualSteps': 'Proceed to the next step',
  'services.reassign.step4.status.failed': 'Failed to move <b>{0}</b> from <b>{1}</b> host to <b>{2}</b> host',
  'services.reassign.step4.status.info': 'Reassigning {0}. \nPlease wait for all tasks to be completed.',
  'services.reassign.step4.retry': 'You can click on the Retry or Abort button to retry failed task or abort changes',
  'services.reassign.step4.abortError': 'Error in aborting changes.',
  'services.reassign.step5.header': 'Manual commands',
  'services.reassign.step5.body.namenode':
      '<div class="alert alert-info">' +
      '<ol>' +
      '<li>Copy the contents of <b>{0}</b> on the source host <b>{1}</b> to <b>{0}</b> on the target host <b>{2}</b>.</li>' +
      '<li>Login to the target host <b>{2}</b> and change permissions for the NameNode dirs by running:' +
      '<div class="code-snippet">chown -R {3}:{5} {6}</div></li>' +
      '<li>Create marker directory by running:' +
      '<div class="code-snippet">mkdir -p /var/lib/hdfs/namenode/formatted</div></li>' +
      '</ol>' +
      '</div>',
  'services.reassign.step5.body.namenode_ha':
      '<div class="alert alert-info">' +
      '<ol>' +
      '<li>Login to the NameNode host <b>{4}</b>.</li>' +
      '<li>Reset automatic failover information in ZooKeeper by running:' +
      '<div class="code-snippet">sudo su {3} -l -c \'hdfs zkfc -formatZK\'</div></li>' +
      '</ol>' +
      '</div>' +
      '<div class="alert alert-info">' +
      '<ol start="3">' +
      '<li>Login to the newly installed NameNode host <b>{2}</b>.<br>' +
      '<div class="alert alert-warn"><strong>Important!</strong> Be sure to login to the newly installed NameNode host.<br>This is a different host from the Steps 1 and 2 above.</div>' +
      '</li>' +
      '<li>Initialize the metadata by running:' +
      "<div class='code-snippet'>sudo su {3} -l -c 'hdfs namenode -bootstrapStandby'</div></li>" +
      '</ol>' +
      '</div>',
  'services.reassign.step5.body.secondary_namenode':
      '<div class="alert alert-info">' +
      '<ol>' +
      '<li>Copy the contents of <b>{0}</b> on the source host <b>{1}</b> to <b>{0}</b> on the target host <b>{2}</b>.</li>' +
      '<li>Login to the target host <b>{2}</b> and change permissions for the SNameNode dirs by running:' +
      '<div class="code-snippet">chown -R {3}:{5} {6}</div></li>' +
      '</ol>' +
      '</div>',
  'services.reassign.step5.body.oozie_server':
    '<div class="alert alert-info">' +
    '<ol>' +
    '<li>On <b>{1}</b> copy the contents of' +
    '<div class="code-snippet"><b>/hadoop/oozie/data/oozie-db</b></div></li>' +
    '<li>To the target host <b>{2}</b></li>' +
    '<li>If the directory doesn\'t exists you can create by running' +
    '<div class="code-snippet">mkdir -p /hadoop/oozie/data/oozie-db</div></li>' +
    '<li>Update directory permissions by running' +
    '<div class="code-snippet">chown -R oozie:{5} /hadoop/oozie/data</div></li>' +
    '</ol>' +
    '</div>',
  'services.reassign.step5.body.mysql_server':
    '<div class="alert alert-info">' +
    '<ol>' +
    '<li>On <b>{1}</b> using a terminal you can export your Metastore DB (MYSQL) using:' +
    '<div class="code-snippet">mysqldump db_name > backup-file.sql</div></li>' +
    '<li>Copy the file to the target host <b>{2}</b> hosting the MySQL DB</li>' +
    '<li>Execute this SQL inside <b>mysql<b>' +
    '<div class="code-snippet">CREATE DATABASE db_name;</div></li>' +
    '<li>Import the database using' +
    '<div class="code-snippet">mysql db_name < backup-file.sql</div></li>' +
    '</ol>' +
    '</div>',
  'services.reassign.step5.body.app_timeline_server': '<div class="alert alert-info">' +
  '<ol>' +
  '<li>Copy <b>/hadoop/yarn/timeline/leveldb-timeline-store.ldb</b> from the source host <b>{1}</b> to <b>/hadoop/yarn/timeline/leveldb-timeline-store.ldb</b> on the target host <b>{2}</b>.</li>' +
  '<li>Login to the target host <b>{2}</b> and change permissions by running:' +
  '<div class="code-snippet">chown -R {3}:{5} /hadoop/yarn/timeline/leveldb-timeline-store.ldb</div></li>' +
  '<div class="code-snippet">chmod -R 700 /hadoop/yarn/timeline/leveldb-timeline-store.ldb</div></li>' +
  '</ol>' +
  '</div>',
  'services.reassign.step5.body.securityNotice': '<div class="alert alert-info"> <div class="alert alert-warn"> <strong>Note: </strong> Secure cluster' +
    ' requires generating necessary principals for reassigned component and creating keytab files with the principal on ' +
    'the target host. The keytab file should be accessible to the service user.</div> {0} </div>',
  'services.reassign.step5.body.securityConfigsList': 'Create keytab file <b>{0}</b> with principal <b>{1}</b> on <b>{2}</b> host.',
  'services.reassign.step5.body.proceedMsg': 'Please proceed once you have completed the steps above',
  'services.reassign.step5.confirmPopup.body': 'Please confirm that you have run the manual steps before continuing.',
  'services.reassign.step6.header': 'Start and Test services',
  'services.reassign.step6.tasks.putHostComponentsInMaintenanceMode.title':'Disable {0}',
  'services.reassign.step6.tasks.deleteHostComponents.title': 'Delete disabled {0}',
  'services.reassign.step6.tasks.startServices.title': 'Start All Services',
  'services.reassign.step6.tasks.stopMysqlService.title': 'Stop Mysql Server',
  'services.reassign.step6.status.success': 'Successfully moved <b>{0}</b> from <b>{1}</b> host to <b>{2}</b> host.',
  'services.reassign.step6.status.failed': 'Failed to move <b>{0}</b> from <b>{1}</b> host to <b>{2}</b> host.',
  'services.reassign.step6.status.info': 'Reassigning {0}. \nPlease wait for all tasks to be completed.',
  'services.reassign.step7.header': 'Rollback',
  'services.reassign.step7.info': 'Rollback in progress',
  'services.reassign.step7.failed': 'Rollback failed',
  'services.reassign.step7.success': 'Rollback successfully completed',

  /** services page constants **/

  'service.hbase.activeMaster': 'Active Master',

  'services.hive.client': 'Hive Client',
  'services.hive.clients': 'Hive Clients',

  'services.falcon.client': 'Falcon Client',
  'services.falcon.clients': 'Falcon Clients',

  'services.oozie.client': 'Oozie Client',
  'services.oozie.clients': 'Oozie Clients',
  'services.oozie.webUi': 'Oozie Web UI',

  'services.ganglia.webUi': 'Ganglia Web UI',
  'services.ganglia.monitors': 'Ganglia Monitors',

  'services.mapreduce2.webUi': 'History Server UI',

  'services.zookeeper.prefix': '{0} of',
  'services.zookeeper.title': '{0} ZooKeepers',
  'services.zookeeper.postfix': 'running',
  'services.zookeeper.clients': 'ZooKeeper Clients',
  'services.zookeeper.client': 'ZooKeeper Client',

  'services.mapreduce2.history.running': 'History server is running',
  'services.mapreduce2.history.stopped': 'History server is stopped',
  'services.mapreduce2.history.unknown': 'History server status is unknown',
  'services.mapreduce2.smokeTest.requirement': 'MapReduce2 smoke test requires YARN service be started',

  'services.tez.client': 'Tez client',
  'services.tez.clients': 'Tez clients',
  'services.pig.client': 'Pig client',
  'services.pig.clients': 'Pig clients',
  'services.glusterfs.client': 'GLUSTERFS client',
  'services.glusterfs.clients': 'GLUSTERFS clients',
  'services.sqoop.client': 'Sqoop client',
  'services.sqoop.clients': 'Sqoop clients',

  'services.hbase.master.error':'None of the HBase masters is active',

  'alerts.actions.manage_alert_groups_popup.header':'Manage Alert Groups',
  'alerts.actions.manage_alert_groups_popup.notice':'You can manage alert groups for each service in this dialog. View the list of alert groups and the alert definitions configured in them. ' +
    'You can also add/remove alert definitions, and pick notification for that alert group.',
  'alerts.actions.manage_alert_groups_popup.rename':'Rename',
  'alerts.actions.manage_alert_groups_popup.duplicate':'Duplicate',
  'alerts.actions.manage_alert_groups_popup.group_name_lable':'Name',
  'alerts.actions.manage_alert_groups_popup.group_desc_lable':'Description',
  'alerts.actions.manage_alert_groups_popup.notifications':'Notifications',
  'alerts.actions.manage_alert_groups_popup.addButton':'Create Alert Group',
  'alerts.actions.manage_alert_groups_popup.addGroup.exist': 'Alert Group with given name already exists',
  'alerts.actions.manage_alert_groups_popup.removeButton':'Delete Alert Group',
  'alerts.actions.manage_alert_groups_popup.removeButtonDisabled':'Cannot delete default alert group',
  'alerts.actions.manage_alert_groups_popup.renameButton':'Rename Alert Group',
  'alerts.actions.manage_alert_groups_popup.duplicateButton':'Duplicate Alert Group',
  'alerts.actions.manage_alert_groups_popup.addDefinition.noDefinitions':'No alert definitions to display',
  'alerts.actions.manage_alert_groups_popup.addDefinition':'Add alert definitions to selected Alert Group',
  'alerts.actions.manage_alert_groups_popup.addDefinitionDisabled':'There are no available alert definitions to add',
  'alerts.actions.manage_alert_groups_popup.addDefinitionToDefault': 'Cannot modify default alert group',
  'alerts.actions.manage_alert_groups_popup.removeDefinition':'Remove alert definitions from selected Alert Group',
  'alerts.actions.manage_alert_groups_popup.removeDefinitionDisabled':'Cannot modify default alert group',
  'alerts.actions.manage_alert_groups_popup.selectDefsDialog.title': 'Select Alert Group Definitions',
  'alerts.actions.manage_alert_groups_popup.selectDefsDialog.message': 'Select alert definitions to be added to this "{0}" Alert Group.',
  'alerts.actions.manage_alert_groups_popup.selectDefsDialog.filter.placeHolder': 'All',
  'alerts.actions.manage_alert_groups_popup.selectDefsDialog.selectedDefsLink': '{0} out of {1} alert definitions selected',
  'alerts.actions.manage_alert_groups_popup.selectDefsDialog.message.warning': 'At least one alert definition needs to be selected.',

  'alerts.actions.manage_alert_notifications_popup.header':'Manage Alert Notifications',
  'alerts.actions.manage_alert_notifications_popup.noAlertNotification':'No alert notifications defined',
  'alerts.actions.manage_alert_notifications_popup.addButton':'Create new Alert Notification',
  'alerts.actions.manage_alert_notifications_popup.addHeader':'Create Alert Notification',
  'alerts.actions.manage_alert_notifications_popup.removeButton':'Delete Alert Notification',
  'alerts.actions.manage_alert_notifications_popup.editButton':'Edit Alert Notification',
  'alerts.actions.manage_alert_notifications_popup.editHeader':'Edit Notification',
  'alerts.actions.manage_alert_notifications_popup.duplicateButton':'Duplicate Alert Notification',
  'alerts.actions.manage_alert_notifications_popup.method':'Method',
  'alerts.actions.manage_alert_notifications_popup.email':'Email To',
  'alerts.actions.manage_alert_notifications_popup.SMTPServer':'SMTP Server',
  'alerts.actions.manage_alert_notifications_popup.SMTPPort':'SMTP Port',
  'alerts.actions.manage_alert_notifications_popup.SMTPUseAuthentication':'Use authentication',
  'alerts.actions.manage_alert_notifications_popup.SMTPUsername':'Username',
  'alerts.actions.manage_alert_notifications_popup.SMTPPassword':'Password',
  'alerts.actions.manage_alert_notifications_popup.retypeSMTPPassword':'Password Confirmation',
  'alerts.actions.manage_alert_notifications_popup.SMTPSTARTTLS':'Start TLS',
  'alerts.actions.manage_alert_notifications_popup.emailFrom':'Email From',
  'alerts.actions.manage_alert_notifications_popup.version':'Version',
  'alerts.actions.manage_alert_notifications_popup.OIDs':'OIDs',
  'alerts.actions.manage_alert_notifications_popup.community':'Community',
  'alerts.actions.manage_alert_notifications_popup.port':'Port',
  'alerts.actions.manage_alert_notifications_popup.global':'Global',
  'alerts.actions.manage_alert_notifications_popup.noDescription':'<i>No description</i>',
  'alerts.actions.manage_alert_notifications_popup.noGroup':'<i>None selected</i>',
  'alerts.actions.manage_alert_notifications_popup.severityFilter':'Severity',
  'alerts.actions.manage_alert_notifications_popup.clearAll':'Clear All',
  'alerts.actions.manage_alert_notifications_popup.selectAll':'Select All',
  'alerts.actions.manage_alert_notifications_popup.confirmDeleteHeader':'Confirm Delete',
  'alerts.actions.manage_alert_notifications_popup.confirmDeleteBody':'Are you sure you want to delete {0} notification?',
  'alerts.actions.manage_alert_notifications_popup.error.name.empty': 'Notification name is required',
  'alerts.actions.manage_alert_notifications_popup.error.name.existed': 'Notification name already exists',

  'hosts.host.add':'Add New Hosts',
  'hosts.table.noHosts':'No hosts to display',

  'hosts.filters.filteredHostsInfo': '{0} of {1} hosts showing',

  'hosts.filters.selectedHostInfo': 'host selected',
  'hosts.filters.selectedHostsInfo': 'hosts selected',

  'hosts.filters.clearSelection': 'clear selection',

  'hosts.filters.filterComponents': 'Filter by <strong>Component</strong>',

  'hosts.table.restartComponents.withNames':'Restart {0}',
  'hosts.table.restartComponents.withoutNames':'{0} components should be restarted',

  'hosts.table.componentsInPassiveState.withNames':'{0} in Maintenance Mode',
  'hosts.table.componentsInPassiveState.withoutNames':'{0} components in Maintenance Mode',

  'hosts.table.menu.l1.selectedHosts':'Selected Hosts',
  'hosts.table.menu.l1.filteredHosts':'Filtered Hosts',
  'hosts.table.menu.l1.allHosts':'All Hosts',
  'hosts.table.menu.l2.allComponents':'All Components',
  'hosts.table.menu.l2.restartAllComponents':'Restart All Components',

  'hosts.bulkOperation.confirmation.header':'Confirm Bulk Operation',
  'hosts.bulkOperation.confirmation.hosts':'Are you sure you want to <strong>{0}</strong> on the following {1} hosts?',
  'hosts.bulkOperation.confirmation.hostComponents':'Are you sure you want to <strong>{0} {1}</strong> on the following {2} hosts?',
  'hosts.bulkOperation.passiveState.nothingToDo.body':'All hosts that you selected are already in Maintenance Mode.',
  'hosts.bulkOperation.warningInfo.body':'Components on these hosts are stopped so decommission will be skipped.',
  'hosts.bulkOperation.host_components.passiveState.nothingToDo.body':'All host components that you selected are already in Maintenance Mode',

  'hosts.selectHostsDialog.title': 'Select Configuration Group Hosts',
  'hosts.selectHostsDialog.message': 'Select hosts that should belong to this {0} Configuration Group. All hosts belonging to this group will have the same set of configurations.',
  'hosts.selectHostsDialog.filter.placeHolder': 'Filter...',
  'hosts.selectHostsDialog.selectedHostsLink': '{0} out of {1} hosts selected',
  'hosts.selectHostsDialog.message.warning': 'At least one host needs to be selected.',

  'hosts.host.serviceNotAvailable': 'Service not available on this host',

  'hosts.host.menu.stackVersions': 'Versions',
  'hosts.host.stackVersions.table.allVersions': 'All Versions',
  'hosts.host.stackVersions.table.allNames': 'All Names',
  'hosts.host.stackVersions.table.noVersions': 'No versions',
  'hosts.host.stackVersions.table.filteredInfo': '{0} of {1} versions showing',
  'hosts.host.stackVersions.status.init': 'Uninstalled',
  'hosts.host.stackVersions.status.installed': 'Installed',
  'hosts.host.stackVersions.status.install_failed': 'Install Failed',
  'hosts.host.stackVersions.status.installing': 'Installing',
  'hosts.host.stackVersions.status.current': 'Current',
  'hosts.host.stackVersions.status.out_of_sync': 'Out of Sync',
  'hosts.host.stackVersions.status.upgrading': 'Upgrading',
  'hosts.host.stackVersions.status.upgrade_failed': 'Upgrade Failed',
  'hosts.host.stackVersions.outOfSync.tooltip': 'This version is Out of Sync on this host. Out of Sync can occur ' +
    'the components on a host change after installing a version not in use. ' +
    'Click Install to have Ambari install the packages for this version to get this host in sync.',
  'hosts.host.stackVersions.install.confirmation': 'You are about to install version <b>{0}</b> on this host.',

  'hosts.host.metrics.dataUnavailable':'Data Unavailable',
  'hosts.host.metrics.cpu':'CPU Usage',
  'hosts.host.metrics.cpu.displayNames.cpu_wio':'CPU I/O Idle',
  'hosts.host.metrics.cpu.displayNames.cpu_idle':'CPU Idle',
  'hosts.host.metrics.cpu.displayNames.cpu_nice':'CPU Nice',
  'hosts.host.metrics.cpu.displayNames.cpu_aidle':'CPU Boot Idle',
  'hosts.host.metrics.cpu.displayNames.cpu_system':'CPU System',
  'hosts.host.metrics.cpu.displayNames.cpu_user':'CPU User',
  'hosts.host.metrics.disk':'Disk Usage',
  'hosts.host.metrics.disk.displayNames.disk_total':'Total',
  'hosts.host.metrics.disk.displayNames.disk_free':'Available',
  'hosts.host.metrics.load':'Load',
  'hosts.host.metrics.load.displayNames.load_fifteen':'15 Minute Load',
  'hosts.host.metrics.load.displayNames.load_one':'1 Minute Load',
  'hosts.host.metrics.load.displayNames.load_five':'5 Minute Load',
  'hosts.host.metrics.memory':'Memory Usage',
  'hosts.host.metrics.memory.displayNames.mem_shared':'Shared',
  'hosts.host.metrics.memory.displayNames.swap_free':'Swap',
  'hosts.host.metrics.memory.displayNames.mem_buffers':'Buffers',
  'hosts.host.metrics.memory.displayNames.mem_free':'Free',
  'hosts.host.metrics.memory.displayNames.mem_cached':'Cached',
  'hosts.host.metrics.network':'Network Usage',
  'hosts.host.metrics.network.displayNames.pkts_out':'Packets Out',
  'hosts.host.metrics.network.displayNames.bytes_in':'Bytes In',
  'hosts.host.metrics.network.displayNames.bytes_out':'Bytes Out',
  'hosts.host.metrics.network.displayNames.pkts_in':'Packets In',
  'hosts.host.metrics.processes':'Processes',
  'hosts.host.metrics.processes.displayNames.proc_total':'Total Processes',
  'hosts.host.metrics.processes.displayNames.proc_run':'Processes Run',

  'hosts.host.summary.header':'Summary',
  'hosts.host.summary.hostname':'Hostname',
  'hosts.host.summary.agentHeartbeat':'Heartbeat',
  'hosts.host.summary.hostMetrics':'Host Metrics',
  'hosts.host.summary.addComponent':'Add Component',
  'hosts.host.summary.currentVersion':'Current Version',

  'hosts.host.details.hostActions':'Host Actions',
  'hosts.host.details.needToRestart':'Host needs {0} {1} restarted',
  'hosts.host.details.needToRestart.button':'Restart',
  'hosts.host.details.deleteHost':'Delete Host',
  'hosts.host.details.startAllComponents':'Start All Components',
  'hosts.host.details.stopAllComponents':'Stop All Components',
  'hosts.host.details.restartAllComponents':'Restart All Components',
  'hosts.host.details.refreshConfigs':'Refresh configs',
  'hosts.host.details.for.postfix':'{0} for host',
  'host.host.details.installClients': 'Install clients',

  'host.host.componentFilter.master':'Master Components',
  'host.host.componentFilter.slave':'Slave Components',
  'host.host.componentFilter.client':'Client Components',
  'hosts.host.deleteComponent.popup.msg1':'Are you sure you want to delete {0}?',
  'hosts.host.deleteComponent.popup.deleteZooKeeperServer':'Deleting <i>ZooKeeper Server</i> may reconfigure such properties:<ul><li>ha.zookeeper.quorum</li><li>hbase.zookeeper.quorum</li><li>templeton.zookeeper.hosts</li><li>yarn.resourcemanager.zk-address</li><li>hive.zookeeper.quorum</li><li>hive.cluster.delegation.token.store.zookeeper.connectString</li></ul>',
  'hosts.host.deleteComponent.popup.warning':'<b>WARNING!</b> Delete the last <i>{0}</i> component in the cluster?</br>Deleting the last component in the cluster could result in permanent loss of service data.',
  'hosts.host.deleteComponent.popup.confirm':'Confirm Delete',
  'hosts.host.installComponent.popup.confirm':'Confirm Install',
  'hosts.host.installComponent.msg':'Are you sure you want to install {0}?',
  'hosts.host.addComponent.msg':'Are you sure you want to add {0}?',
  'hosts.host.addComponent.ZOOKEEPER_SERVER':'Adding ZooKeeper Server may reconfigure such properties:<ul><li>ha.zookeeper.quorum</li><li>hbase.zookeeper.quorum</li><li>templeton.zookeeper.hosts</li><li>yarn.resourcemanager.zk-address</li><li>hive.zookeeper.quorum</li><li>hive.cluster.delegation.token.store.zookeeper.connectString</li></ul>',
  'hosts.host.addComponent.deleteHostWithZooKeeper':'Deleting host with ZooKeeper Server may reconfigure such properties:<ul><li>ha.zookeeper.quorum</li><li>hbase.zookeeper.quorum</li><li>templeton.zookeeper.hosts</li><li>yarn.resourcemanager.zk-address</li><li>hive.zookeeper.quorum</li><li>hive.cluster.delegation.token.store.zookeeper.connectString</li></ul>',
  'host.host.addComponent.popup.dependedComponents.body': '{0} requires {1} to be installed along with it on the same host. Please add them first and then try adding {0}',
  'host.host.addComponent.popup.dependedComponents.header': 'Component dependencies',
  'hosts.host.zooKeeper.configs.save.note': 'This configuration is created by ambari while installing/deleting zookeeper component on a host',
  'hosts.host.addComponent.securityNote':'You are running your cluster in secure mode. You must set up the keytab for {0} on {1} before you proceed. Otherwise, the component will not be able to start properly.',
  'hosts.host.addComponent.popup.confirm':'Confirm Add',
  'hosts.host.datanode.decommission':'Decommission DataNode',
  'hosts.host.datanode.recommission':'Recommission DataNode',
  'hosts.host.nodemanager.decommission':'Decommission NodeManager',
  'hosts.host.nodemanager.recommission':'Recommission NodeManager',
  'hosts.host.tasktracker.decommission':'Decommission TaskTracker',
  'hosts.host.tasktracker.recommission':'Recommission TaskTracker',
  'hosts.host.tasktracker.restart':'Restart TaskTracker',
  'hosts.host.regionserver.decommission.batch1':'Decommission RegionServer - Turn drain mode on',
  'hosts.host.regionserver.decommission.batch2':'Decommission RegionServer - Stop RegionServer',
  'hosts.host.regionserver.decommission.batch3':'Decommission RegionServer - Turn drain mode off',
  'hosts.host.hbase_regionserver.recommission':'Recommission RegionServer',
  'hosts.host.hbase_regionserver.decommission':'Decommission RegionServer',
  'hosts.host.hbase_regionserver.decommission.warning':'Last RegionServer can\'t be decommissioned',
  'hosts.host.decommissioned':'Decommissioned',
  'hosts.host.decommissioning':'Decommissioning',
  'hosts.host.addComponent.HIVE_METASTORE':'Adding <i>Hive Metastore</i> will reconfigure such properties:<ul><li>hive.metastore.uris</li><li>templeton.hive.properties</li></ul>',
  'hosts.host.deleteComponent.popup.deleteHiveMetastore':'Deleting <i>Hive Metastore</i> will reconfigure such properties:<ul><li>hive.metastore.uris</li><li>templeton.hive.properties</li></ul>',
  'hosts.host.hive.configs.save.note': 'This configuration is created by ambari while installing/deleting hive component on a host',

  'hosts.component.passive.implied.host.mode.tooltip':'Cannot Turn Off Maintenance Mode because Host is in Maintenance Mode',
  'hosts.component.passive.implied.service.mode.tooltip':'Cannot Turn Off Maintenance Mode because {0} is in Maintenance Mode',
  'hosts.component.passive.mode':'Component is in Maintenance Mode',
  'hosts.component.passive.short.mode':'Maintenance Mode',
  'hosts.host.passive.mode':'Host is in Maintenance Mode',
  'hosts.host.alert.noAlerts':'No alerts',
  'hosts.host.alert.noAlerts.message':'There are no alerts for this host.',
  'hosts.host.healthStatus.heartBeatNotReceived':'The server has not received a heartbeat from this host for more than 3 minutes.',
  'hosts.host.healthStatus.mastersDown':"The following master components are down:\n",
  'hosts.host.healthStatus.slavesDown':"The following slave components are down:\n",
  'hosts.host.healthStatus.allUp':'All components are up',
  'hosts.host.healthStatusCategory.green': "Healthy",
  'hosts.host.healthStatusCategory.red': "Master Down",
  'hosts.host.healthStatusCategory.orange': "Slave Down",
  'hosts.host.healthStatusCategory.yellow': "Lost Heartbeat",
  'hosts.host.alerts.label': 'Alerts',
  'hosts.host.maintainance.allComponents.context': 'All Host Components',
  'hosts.host.maintainance.stopAllComponents.context': 'Stop All Host Components',
  'hosts.host.maintainance.startAllComponents.context': 'Start All Host Components',
  'hosts.host.alerts.st':'&nbsp;!&nbsp;',
  'hosts.decommission.popup.body':'Are you sure?',
  'hosts.decommission.popup.header':'Confirmation',
  'hosts.decommission.tooltip.warning':'Cannot {0} since {1} is not running',
  'hosts.delete.popup.body':'Are you sure you want to delete host <i>{0}</i>?',
  'hosts.delete.popup.body.msg1':'By removing this host, Ambari will ignore future communications from this host. Software packages will not be removed from the host. The components on the host should not be restarted. If you wish to readd this host to the cluster, be sure to clean the host.',
  'hosts.delete.popup.body.msg3':'If this host was hosting a Zookeeper Server, the Zookeeper Service should be restarted. Go to the <i>Services</i> page.',
  'hosts.delete.popup.body.msg4':'<b>WARNING!</b> Delete the last <i>{0}</i> component[s] in the cluster?</br>Deleting the last components in the cluster could result in permanent loss of service data.',
  'hosts.delete.popup.body.msg.unknownComponents':'This host does not appear to be online and Ambari communication with the Agent has been lost.',
  'hosts.delete.popup.header':'Confirmation',
  'hosts.delete.popup.title':'Delete Host',
  'hosts.delete.popup.unknownComponents':'The following components have unknown status:',
  'hosts.cant.do.popup.title':'Unable to Delete Host',
  'hosts.cant.do.popup.masterList.body':'This host cannot be deleted since it has the following master components:',
  'hosts.cant.do.popup.masterList.body.end':'To delete this host, you must first move all the master components listed above to another host.',
  'hosts.cant.do.popup.nonDeletableList.body':'Deletion of the following {0} components is not supported. ',
  'hosts.cant.do.popup.runningList.body':'This host cannot be deleted since the following components are running:',
  'hosts.cant.do.popup.runningList.body.end':'To delete this host, you must first stop all the running components listed above. ' +
    'If this host has a DataNode, it should be decommissioned first to prevent data loss.',
  'hosts.add.header':'Add Host Wizard',
  'hosts.add.exit.header':'Exit',
  'hosts.add.exit.body':'Do you really want to exit Add Host Wizard?',
  'hosts.assignRack':'Assign Rack',
  'hosts.passiveMode.popup':'Are you sure you want to <b>Turn {0} Maintenance Mode</b> for {1}?',

  'charts.horizon.chart.showText':'show',
  'charts.horizon.chart.hideText':'hide',
  'charts.horizon.chart.attributes.cpu':'CPU',
  'charts.horizon.chart.attributes.memory':'Memory',
  'charts.horizon.chart.attributes.network':'Network',
  'charts.horizon.chart.attributes.io':'I/O',

  'charts.heatmap.selectMetric':'Select Metric...',

  'charts.heatmap.category.host':'Host',
  'charts.heatmap.item.host.memory':'Memory Used',
  'charts.heatmap.item.host.disk':'Disk Space Used',
  'charts.heatmap.item.host.process':'Total Running Processes',
  'charts.heatmap.category.hdfs':'HDFS',
  'charts.heatmap.category.yarn': 'YARN',
  'charts.heatmap.category.hbase': 'HBase',
  'charts.heatmap.unknown': 'Unknown',
  'charts.heatmap.label.notApplicable' :'Not Applicable',
  'charts.heatmap.label.invalidData' :'Invalid data',
  'charts.heatmap.metrics.bytesRead' :'HDFS Bytes Read',
  'charts.heatmap.metrics.bytesWritten' :'HDFS Bytes Written',
  'charts.heatmap.metrics.DFSGarbageCollection' :'HDFS Garbage Collection Time',
  'charts.heatmap.metrics.DFSMemHeapUsed' :'HDFS JVM Heap Memory Used',
  'charts.heatmap.metrics.diskSpaceUsed' :'Host Disk Space Used %',
  'charts.heatmap.metrics.YarnGCTime' :'YARN Garbage Collection Time',
  'charts.heatmap.metrics.YarnMemHeapUsed' :'YARN JVM Heap Memory Used',

  'charts.heatmap.metrics.memoryUsed' :'Host Memory Used %',
  'charts.heatmap.metrics.processRun' :'Total Running Processes',
  'charts.heatmap.metrics.YarnMemoryUsed' :'YARN Memory used %',
  'charts.heatmap.metrics.cpuWaitIO':'Host CPU Wait I/O %',
  'charts.heatmap.metrics.HbaseRegionServerReadCount': 'HBase Read Request Count',
  'charts.heatmap.metrics.HbaseRegionServerWriteCount': 'HBase Write Request Count',
  'charts.heatmap.metrics.HbaseRegionServerCompactionQueueSize': 'HBase Compaction Queue Size',
  'charts.heatmap.metrics.HbaseRegionServerRegions': 'HBase Regions',
  'charts.heatmap.metrics.HbaseRegionServerMemStoreSize': 'HBase Memstore Sizes',
  'metric.notFound':'no items found',
  'metric.default':'combined',
  'metric.cpu':'cpu',
  'metric.memory':'disk used',
  'metric.network':'network',
  'metric.io':'io',
  'metric.more':'more',
  'metric.more.cpu':'CPU',
  'metric.more.disk':'Disk',
  'metric.more.load':'Load',
  'metric.more.memory':'Memory',
  'metric.more.network':'Network',
  'metric.more.process':'Process',

  'dashboard.clusterMetrics':'Cluster Metrics',

  'dashboard.clusterMetrics.cpu':'CPU Usage',
  'dashboard.clusterMetrics.cpu.displayNames.idle':'Idle',
  'dashboard.clusterMetrics.load':'Cluster Load',
  'dashboard.clusterMetrics.memory':'Memory Usage',
  'dashboard.clusterMetrics.network':'Network Usage',

  'dashboard.widgets.title': 'Metrics',
  'dashboard.heatmaps.title': 'Heatmaps',
  'dashboard.button.switch': 'Switch to classic dashboard',
  'dashboard.button.switchShort': 'Switch',
  'dashboard.button.reset': 'Reset all widgets to default ',
  'dashboard.button.gangliaLink': 'View metrics in Ganglia',

  'dashboard.widgets.NameNodeHeap': 'NameNode Heap',
  'dashboard.widgets.NameNodeCpu': 'NameNode CPU WIO',
  'dashboard.widgets.HDFSDiskUsage': 'HDFS Disk Usage',
  'dashboard.widgets.HDFSDiskUsage.DFSused': 'DFS used',
  'dashboard.widgets.HDFSDiskUsage.nonDFSused': 'non DFS used',
  'dashboard.widgets.HDFSDiskUsage.remaining': 'remaining',
  'dashboard.widgets.HDFSDiskUsage.info': '{0} ({1}%)',
  'dashboard.widgets.DataNodeUp': 'DataNodes Live',
  'dashboard.widgets.SuperVisorUp': 'Supervisors Live',
  'dashboard.widgets.FlumeAgentUp': 'Flume Live',
  'dashboard.widgets.NameNodeRpc': 'NameNode RPC',
  'dashboard.widgets.nothing': 'No Widget to Add',
  'dashboard.widgets.NameNodeUptime': 'NameNode Uptime',
  'dashboard.widgets.HDFSLinks': 'HDFS Links',
  'dashboard.widgets.HDFSLinks.activeNameNode': 'Active NameNode',
  'dashboard.widgets.HDFSLinks.standbyNameNode': 'Standby NameNode',
  'dashboard.widgets.HDFSLinks.standbyNameNodes': '2 Standby NameNodes',
  'dashboard.widgets.HBaseLinks': 'HBase Links',
  'dashboard.widgets.HBaseAverageLoad': 'HBase Ave Load',
  'dashboard.widgets.HBaseMasterHeap': 'HBase Master Heap',
  'dashboard.widgets.HBaseRegionsInTransition': 'Region In Transition',
  'dashboard.widgets.HBaseMasterUptime': 'HBase Master Uptime',
  'dashboard.widgets.ResourceManagerHeap': 'ResourceManager Heap',
  'dashboard.widgets.ResourceManagerUptime': 'ResourceManager Uptime',
  'dashboard.widgets.NodeManagersLive': 'NodeManagers Live',
  'dashboard.widgets.YARNMemory': 'YARN Memory',
  'dashboard.widgets.YARNLinks': 'YARN Links',
  'dashboard.widgets.error.invalid': 'Invalid! Enter a number between 0 - {0}',
  'dashboard.widgets.error.smaller': 'Threshold 1 should be smaller than threshold 2!',

  'dashboard': {
    'widgets': {
      'popupHeader': 'Customize Widget',
      'hintInfo': {
        'common': 'Edit the percentage thresholds to change the color of current pie chart. <br />Enter two numbers between 0 to {0}',
        'hint1': 'Edit the percentage of thresholds to change the color of current widget. <br />Assume all components UP is 100, and all DOWN is 0. <br /> So enter two numbers between 0 to {0}',
        'hint2': 'Edit the thresholds to change the color of current widget.<br /><br />So enter two numbers larger than 0.',
        'hint3': 'Edit the thresholds to change the color of current widget.<br />The unit is milli-second. <br />So enter two numbers larger than 0. '
      }
    }
  },

  'dashboard.services':'Services',
  'dashboard.services.hosts':'Hosts',
  'dashboard.services.uptime':'{0}',
  'dashboard.services.summary.slaves.live': 'Live',
  'dashboard.services.hdfs.summary':'{0} of {1} nodes live, {2}% capacity used',
  'dashboard.services.hdfs.nanmenode':'NameNode',
  'dashboard.services.hdfs.snanmenode':'Secondary NameNode',
  'dashboard.services.hdfs.journalnodes':'JournalNodes',
  'dashboard.services.hdfs.capacity':'HDFS Disk Usage',
  'dashboard.services.hdfs.capacity.dfsUsed':'Disk Usage (DFS Used)',
  'dashboard.services.hdfs.capacity.nonDfsUsed':'Disk Usage (Non DFS Used)',
  'dashboard.services.hdfs.capacity.remaining':'Disk Usage (Remaining)',
  'dashboard.services.hdfs.capacityUsed':'{0} / {1} ({2}%)',
  'dashboard.services.hdfs.totalFilesAndDirs':'Total Files + Directories',
  'dashboard.services.hdfs.datanodes':'DataNodes',
  'dashboard.services.hdfs.datanodecounts':'DataNodes Status',
  'dashboard.services.hdfs.version':'Version',
  'dashboard.services.hdfs.nameNodeWebUI':'NameNode Web UI',
  'dashboard.services.hdfs.nodes.live':'live',
  'dashboard.services.hdfs.nodes.dead':'dead',
  'dashboard.services.hdfs.nodes.decom':'decommissioning',
  'dashboard.services.hdfs.nodes.uptime':'NameNode Uptime',
  'dashboard.services.hdfs.nodes.heap':'NameNode Heap',
  'dashboard.services.hdfs.nodes.heapUsed':'{0} / {1} ({2}% used)',
  'dashboard.services.hdfs.chart.label':'Capacity (Used/Total)',
  'dashboard.services.hdfs.blockErrors':'{0} corrupt / {1} missing / {2} under replicated',
  'dashboard.services.hdfs.datanode.status.tooltip.live': 'This is the number of DataNodes that are live as reported from ' +
    'the NameNode. Even if a DataNode process is up, NameNode might see the status as dead ' +
    'if the DataNode is not communicating with the NameNode as expected. This can be due situations ' +
    'such as a network issue or a hanging DataNode process due to excessive garbage collection.',
  'dashboard.services.hdfs.datanode.status.tooltip.dead': 'This is the number of DataNodes that are dead as reported from ' +
  'the NameNode. Even if a DataNode process is up, NameNode might see the status as dead ' +
  'if the DataNode is not communicating with the NameNode as expected. This can be due situations ' +
  'such as a network issue or a hanging DataNode process due to excessive garbage collection.',
  'dashboard.services.hdfs.datanode.status.tooltip.decommission': 'This is the number of DataNodes that are currently ' +
  'Decommissioning as reported from the NameNode. If there are not enough other DataNodes ' +
  'in the cluster to create the configured number of block replicas based on the dfs.replication ' +
  'property (typically 3), a DataNode can get stuck in decommissioning state until ' +
  'more DataNodes become available to the NameNode.',

  'dashboard.services.yarn.summary':'{0} of {1} nodes live',
  'dashboard.services.yarn.resourceManager':'ResourceManager',
  'dashboard.services.yarn.nodeManagers':'NodeManagers',
  'dashboard.services.yarn.nodeManager':'NodeManager',
  'dashboard.services.yarn.clients':'YARN Clients',
  'dashboard.services.yarn.client':'YARN Client',
  'dashboard.services.yarn.resourceManager.uptime':'ResourceManager Uptime',
  'dashboard.services.yarn.resourceManager.active':'Active ResourceManager',
  'dashboard.services.yarn.resourceManager.standby':'Standby ResourceManager',
  'dashboard.services.resourceManager.nodes.heap':'ResourceManager Heap',
  'dashboard.services.yarn.nodeManagers.status': 'NodeManagers Status',
  'dashboard.services.yarn.nodeManagers.status.msg': '{0} active / {1} lost / {2} unhealthy / {3} rebooted / {4} decommissioned',
  'dashboard.services.yarn.containers': 'Containers',
  'dashboard.services.yarn.containers.msg': '{0} allocated / {1} pending / {2} reserved',
  'dashboard.services.yarn.apps': 'Applications',
  'dashboard.services.yarn.apps.msg': '{0} submitted / {1} running / {2} pending / {3} completed / {4} killed / {5} failed',
  'dashboard.services.yarn.memory': 'Cluster Memory',
  'dashboard.services.yarn.memory.msg': '{0} used / {1} reserved / {2} available',
  'dashboard.services.yarn.queues': 'Queues',
  'dashboard.services.yarn.queues.msg': '{0} Queues',

  'dashboard.services.flume.summary.title':'Flume installed on {0} host{1} ({2} agent{3})',
  'dashboard.services.flume.summary.configure':'Configure Agents',
  'dashboard.services.flume.agentsLabel': 'Flume',
  'dashboard.services.flume.agentLabel': 'Flume Component',
  'dashboard.services.flume.channels': 'Channels',
  'dashboard.services.flume.sources': 'Sources',
  'dashboard.services.flume.sinks': 'Sinks',
  'dashboard.services.flume.agent': 'Agent',

  'dashboard.services.hbase.summary':'{0} region servers with {1} average load',
  'dashboard.services.hbase.masterServer':'HBase Master',
  'dashboard.services.hbase.masterServer.active':'Active HBase Master',
  'dashboard.services.hbase.masterServer.standby':'Standby HBase Master',
  'dashboard.services.hbase.noMasterServer':'No Active Master',
  'dashboard.services.hbase.masterServerHeap':'Master Heap',
  'dashboard.services.hbase.masterServerHeap.summary':'{0} / {1} ({2}% used)',
  'dashboard.services.hbase.masterServerUptime':'Master Server Uptime',
  'dashboard.services.hbase.averageLoad':'Average Load',
  'dashboard.services.hbase.averageLoadPerServer':'{0} regions per RegionServer',
  'dashboard.services.hbase.regionServers':'RegionServers',
  'dashboard.services.hbase.regionServersSummary':'{0} live / {1} total',
  'dashboard.services.hbase.chart.label':'Request Count',
  'dashboard.services.hbase.version':'Version',
  'dashboard.services.hbase.masterWebUI':'Master Web UI',
  'dashboard.services.hbase.regions.transition':'Regions In Transition',
  'dashboard.services.hbase.masterStarted':'Master Started',
  'dashboard.services.hbase.masterActivated':'Master Activated',

  'dashboard.services.hive.clients':'Hive Clients',
  'dashboard.services.hive.client':'Hive Client',
  'dashboard.services.hive.metastore':'Hive Metastore',
  'dashboard.services.hive.server2':'HiveServer2',

  'dashboard.services.oozie.clients':'Oozie Clients',
  'dashboard.services.oozie.client':'Oozie Client',
  'dashboard.services.storm.supervisor': 'Supervisor',
  'dashboard.services.storm.supervisors': 'Supervisors',

  'dashboard.services.configs.popup.stopService.header':'Stop service',
  'dashboard.services.configs.popup.stopService.body' : 'Service needs to be stopped for reconfiguration',
  'dashboard.services.configs.popup.restartService.header' : 'Restart service',
  'dashboard.services.configs.popup.restartService.body' : 'Service needs to be restarted for reconfiguration',

  'dashboard.services.zookeeper.server' : 'ZooKeeper Server',

  'dashboard.configHistory.title': 'Config History',
  'dashboard.configHistory.table.version.title' : 'Service',
  'dashboard.configHistory.table.configGroup.title' : 'Config Group',
  'dashboard.configHistory.table.created.title' : 'Created',
  'dashboard.configHistory.table.configGroup.default' : 'default',
  'dashboard.configHistory.table.empty' : 'No history to display',
  'dashboard.configHistory.table.notes.default': 'Initial configurations for {0}',
  'dashboard.configHistory.table.notes.addService': 'Configuration updated while adding service',
  'dashboard.configHistory.table.notes.no': '<i>No notes</i>',
  'dashboard.configHistory.table.version.versionText' : 'V{0}',
  'dashboard.configHistory.table.version.prefix' : 'V',
  'dashboard.configHistory.table.current.tooltip' : 'Current config for {0}:{1}',
  'dashboard.configHistory.table.restart.tooltip' : 'Restart required',
  'dashboard.configHistory.table.filteredHostsInfo': '{0} of {1} versions showing',
  'dashboard.configHistory.info-bar.authoredOn': 'authored on',
  'dashboard.configHistory.info-bar.changesToHandle': 'Changes to handle',
  'dashboard.configHistory.info-bar.showMore': 'Show more',
  'dashboard.configHistory.info-bar.save.popup.title': 'Save Configuration',
  'dashboard.configHistory.info-bar.makeCurrent.popup.title': 'Make Current Confirmation',
  'dashboard.configHistory.info-bar.save.popup.placeholder': 'What did you change?',
  'dashboard.configHistory.info-bar.revert.button': 'Make Current',
  'dashboard.configHistory.info-bar.revert.versionButton': 'Make {0} Current',
  'dashboard.configHistory.info-bar.view.button.disabled': 'You are currently viewing this version.',
  'dashboard.configHistory.info-bar.compare.button.disabled': 'You cannot compare against the same version.',
  'dashboard.configHistory.info-bar.revert.button.disabled': 'This is the current version.',

  'timeRange.presets.1hour':'1h',
  'timeRange.presets.12hour':'12h',
  'timeRange.presets.1day':'1d',
  'timeRange.presets.1week':'1wk',
  'timeRange.presets.1month':'1mo',
  'timeRange.presets.1year':'1yr',

  'tableView.filters.all': 'All',
  'tableView.filters.filtered': 'Filtered',
  'tableView.filters.clearFilters': 'Clear filters',
  'tableView.filters.paginationInfo': '{0} - {1} of {2}',
  'tableView.filters.clearAllFilters': 'clear filters',
  'tableView.filters.showAll': 'Show All',
  'tableView.filters.filteredConfigVersionInfo': '{0} of {1} versions showing',

  'rollingrestart.dialog.title': 'Restart {0}s',
  'rollingrestart.dialog.primary': 'Trigger Rolling Restart',
  'rollingrestart.notsupported.hostComponent': 'Rolling restart not supported for {0} components',
  'rollingrestart.dialog.msg.restart': 'This will restart a specified number of {0}s at a time.',
  'rollingrestart.dialog.msg.noRestartHosts': 'There are no {0}s to do rolling restarts',
  'rollingrestart.dialog.msg.maintainance': 'Note: {0} {1} in Maintenance Mode will not be restarted',
  'rollingrestart.dialog.msg.maintainance.plural': 'Note: {0} {1}s in Maintenance Mode will not be restarted',
  'rollingrestart.dialog.msg.componentsAtATime': '{0}s at a time',
  'rollingrestart.dialog.msg.timegap.prefix': 'Wait ',
  'rollingrestart.dialog.msg.timegap.suffix': 'seconds between batches ',
  'rollingrestart.dialog.msg.toleration.prefix': 'Tolerate up to ',
  'rollingrestart.dialog.msg.toleration.suffix': 'restart failures',
  'rollingrestart.dialog.err.invalid.batchsize': 'Invalid restart batch size: {0}',
  'rollingrestart.dialog.err.invalid.waitTime': 'Invalid wait time between batches: {0}',
  'rollingrestart.dialog.err.invalid.toleratesize': 'Invalid failure toleration count: {0}',
  'rollingrestart.dialog.msg.serviceNotInMM':'Note: This will trigger alerts. To suppress alerts, turn on Maintenance Mode for {0} prior to triggering a rolling restart',
  'rollingrestart.dialog.msg.staleConfigsOnly': 'Only restart {0}s with stale configs',
  'rollingrestart.rest.context': 'Rolling Restart of {0}s - batch {1} of {2}',
  'rollingrestart.context.allOnSelectedHosts':'Restart all components on the selected hosts',
  'rollingrestart.context.allForSelectedService':'Restart all components for {0}',
  'rollingrestart.context.allWithStaleConfigsForSelectedService':'Restart all components with Stale Configs for {0}',
  'rollingrestart.context.allClientsOnSelectedHost':'Restart all clients on {0}',
  'rollingrestart.context.allWithStaleConfigsOnSelectedHost':'Restart components with Stale Configs on {0}',
  'rollingrestart.context.allOnSelectedHost':'Restart all components on {0}',
  'rollingrestart.context.selectedComponentOnSelectedHost':'Restart {0}',
  'rollingrestart.context.default':'Restart components',

  'rolling.command.context': 'Rolling set {0} to state "{1}" - batch {2} of {3}',
  'rolling.nothingToDo.header': 'Nothing to do',
  'rolling.nothingToDo.body': '{0} on selected hosts are already in selected state or in Maintenance Mode.',

  'restart.service.all': 'Restart All',
  'restart.service.rest.context': 'Restart {0}s',

  'menu.item.dashboard':'Dashboard',
  'menu.item.services':'Services',
  'menu.item.hosts':'Hosts',
  'menu.item.admin':'Admin',
  'menu.item.alerts': 'Alerts',
  'menu.item.views':'<i class="icon-th"></i>',
  'menu.item.views.noViews':'No Views',

  'bulkOperation.loading': 'Loading...',
  'jobs.nothingToShow': 'No jobs to display',
  'jobs.loadingTasks': 'Loading...',
  'jobs.error.ats.down': 'Jobs data cannot be shown since YARN App Timeline Server is not running.',
  'jobs.error.400': 'Unable to load data.',
  'jobs.table.custom.date.am':'AM',
  'jobs.table.custom.date.pm':'PM',
  'jobs.table.custom.date.header':'Select Custom Dates',
  'jobs.table.job.fail':'Job failed to run',
  'jobs.customDateFilter.error.required':'This field is required',
  'jobs.customDateFilter.error.date.order':'End Date must be after Start Date',
  'jobs.customDateFilter.startTime':'Start Time',
  'jobs.customDateFilter.endTime':'End Time',

  'views.main.yourViews': 'Your Views',
  'views.main.noViews': 'No views',
  'views.main.instance.noDescription': 'No description',

  'number.validate.empty': 'cannot be empty',
  'number.validate.notValidNumber': 'not a valid number',
  'number.validate.lessThanMinumum': 'value less than {0}',
  'number.validate.moreThanMaximum': 'value greater than {0}',

  'common.combobox.placeholder': 'Filter...',
  'common.combobox.dropdown.overridden': 'Overridden properties',
  'common.combobox.dropdown.final': 'Final properties',
  'common.combobox.dropdown.changed': 'Changed properties',
  'common.combobox.dropdown.issues': 'Show property issues',
  'common.combobox.dropdown.warnings': 'Show property warnings',

  'quick.links.error.label': 'Hostname is undefined',
  'quick.links.publicHostName': '{0} ({1})',
  'quick.links.label.active': 'Active',
  'quick.links.label.standby': 'Standby',

  'contact.administrator': 'Contact System Administrator for more information!',

  'config.group.selection.dialog.title': '{0} Configuration Group',
  'config.group.selection.dialog.subtitle': 'Select or create a {0} Configuration Group where the configuration value will be overridden.',
  'config.group.selection.dialog.option.select': 'Select an existing {0} Configuration Group',
  'config.group.selection.dialog.option.select.msg': 'Overridden property will be changed for hosts belonging to the selected group.',
  'config.group.selection.dialog.option.create': 'Create a new {0} Configuration Group',
  'config.group.selection.dialog.option.create.msg': 'A new {0} Configuration Group will be created with the given name. Initially there will be no hosts in the group, with only the selected property overridden.',
  'config.group.selection.dialog.err.name.exists': 'Configuration Group with given name already exists',
  'config.group.selection.dialog.err.create': 'Error creating new Configuration Group [{0}]',
  'config.group.selection.dialog.no.groups': 'There are no existing {0} Configuration Groups.',
  'config.group.host.switch.dialog.title': 'Change Group',

  'config.group.save.confirmation.header': 'Save Configuration Group',
  'config.group.save.confirmation.msg': 'Click <em>Manage Hosts</em> to manage host membership to the configuration group',
  'config.group.save.confirmation.configGroup': 'Configuration Group',
  'config.group.save.confirmation.saved': 'has been successfully saved',
  'config.group.save.confirmation.manage.button': 'Manage Hosts',
  'config.group.description.default': 'New configuration group created on {0}',

  'utils.ajax.errorMessage': 'Error message',
  'utils.ajax.defaultErrorPopupBody.message': 'received on {0} method for API: {1}',
  'utils.ajax.defaultErrorPopupBody.statusCode': '{0} status code'
};

});

require.register("mixins", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


// load all mixins here

require('mixins/common/blueprint');
require('mixins/common/localStorage');
require('mixins/common/userPref');
require('mixins/common/serverValidator');
require('mixins/common/table_server_view_mixin');
require('mixins/common/table_server_mixin');
require('mixins/main/host/details/host_components/decommissionable');
require('mixins/routers/redirections');
require('mixins/wizard/wizardProgressPageController');
require('mixins/wizard/wizardDeployProgressController');
require('mixins/wizard/wizardProgressPageView');
require('mixins/wizard/wizardDeployProgressView');
require('mixins/wizard/selectHost');
require('mixins/wizard/addSecurityConfigs');
require('mixins/wizard/wizard_menu_view');

});

require.register("mixins/common/blueprint", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var blueprintUtils = require('utils/blueprint');

App.BlueprintMixin = Em.Mixin.create({
  /**
   * returns blueprint for all currenlty installed master, slave and client components
   */
  getCurrentMasterSlaveBlueprint: function () {
    var components = App.HostComponent.find();
    var hosts = components.mapProperty("hostName").uniq();

    var res = {
      blueprint: { host_groups: [] },
      blueprint_cluster_binding: { host_groups: [] }
    };

    hosts.forEach(function (host, i) {
      var group_name = 'host-group-' + (i+1);

      res.blueprint.host_groups.push({
        name: group_name,
        components: components.filterProperty("hostName", host).mapProperty("componentName").map(function (c) {
          return { name: c };
        })
      });

      res.blueprint_cluster_binding.host_groups.push({
        name: group_name,
        hosts: [
          { fqdn: host }
        ]
      });
    });
    return res;
  },

  /**
   * returns blueprint for all currenlty installed slave and client components
   */
  getCurrentSlaveBlueprint: function () {
    var self = this;
    var fullBlueprint = self.getCurrentMasterSlaveBlueprint();

    var slaveComponents = App.StackServiceComponent.find().filterProperty("isSlave").mapProperty("componentName");
    var clientComponents = App.StackServiceComponent.find().filterProperty("isClient").mapProperty("componentName");
    var slaveAndClientComponents = slaveComponents.concat(clientComponents);

    return blueprintUtils.filterByComponents(fullBlueprint, slaveAndClientComponents);
  }
});
});

require.register("mixins/common/chart/storm_linear_time", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.StormLinearTimeChartMixin = Em.Mixin.create({
  ajaxIndex: 'service.metrics.storm.nimbus',
  metricsTemplate: 'metrics/storm/nimbus/{0}[{1},{2},{3}]',

  getDataForAjaxRequest: function() {
    var currentTime = Math.round(App.dateTime() / 1000);
    var metricTemplate = [];
    this.get('stormChartDefinition').forEach(function(chartInfo) {
      metricTemplate.push(
        this.get('metricsTemplate').format(chartInfo.field, currentTime - this.get('timeUnitSeconds'), currentTime, 15)
      );
    }, this);
    return {
      metricsTemplate: metricTemplate.join(',')
    };
  },

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var pathKeys = ['metrics','storm','nimbus'];
    var validPath = true;
    pathKeys.forEach(function(key) {
      if (!jsonData[key]) {
        validPath = false;
      } else {
        jsonData = jsonData[key];
      }
    });
    if (!validPath) {
      return seriesArray;
    }
    this.get('stormChartDefinition').forEach(function(chart){
      seriesArray.push(this.transformData(jsonData[chart.field], chart.name));
    }, this);
    return seriesArray;
  }
});

});

require.register("mixins/common/localStorage", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Fast save/load value to Local Storage
 * Object that implements it should have property <code>name</code> defined
 * @type {Ember.Mixin}
 */
App.LocalStorage = Em.Mixin.create({

  /**
   * <code>App.db.data</code> key
   * @returns {String}
   */
  dbNamespace: function() {
    var name = this.get('name');
    if (Em.isNone(name)) {
      name = this.get('controller.name');
    }
    return name.capitalize().replace('Controller', '');
  }.property('name'),

  /**
   * get property from local storage
   * @param {String} key
   * @return {*}
   */
  getDBProperty: function(key){
    return App.db.get(this.get('dbNamespace'), key);
  },

  /**
   * set property to local storage
   * @param {String} key
   * @param {*} value
   */
  setDBProperty: function(key, value){
    App.db.set(this.get('dbNamespace'), key, value);
  },

  /**
   * Delete the dbNamespace from the localStorage
   * Usually this function is called on quiting/completing the wizard
   */
  resetDbNamespace: function() {
    App.db.data[this.get('dbNamespace')] =  {};
    localStorage.setObject('ambari', App.db.data);
  }

});

});

require.register("mixins/common/serverValidator", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var blueprintUtils = require('utils/blueprint');

App.ServerValidatorMixin = Em.Mixin.create({

  /**
   * @type {bool} set true if at leasst one config has error
   */
  configValidationError: false,

  /**
   * @type {bool} set true if at leasst one config has warning
   */
  configValidationWarning: false,

  /**
   * @type {bool} set true if at leasst one config has warning
   */
  configValidationFailed: false,

  /**
   * @type {object[]} contains additional message about validation errors
   */
  configValidationGlobalMessage: [],

  /**
   * recommendation configs loaded from server
   * (used only during install)
   * @type {Object}
   */
  recommendationsConfigs: null,

  /**
   * by default loads data from model otherwise must be overridden as computed property
   * refer to \assets\data\stacks\HDP-2.1\recommendations_configs.json to learn structure
   * (shouldn't contain configurations filed)
   * @type {Object}
   */
  hostNames: function() {
    return this.get('content.hosts')
        ? Object.keys(this.get('content.hosts'))
        : App.get('allHostNames');
  }.property('content.hosts', 'App.allHostNames'),

  allHostNames: [],
  /**
   * by default loads data from model otherwise must be overridden as computed property
   * @type {Array} - of strings (serviceNames)
   */
  serviceNames: function() {
    return this.get('content.serviceName') ? [this.get('content.serviceName')] : this.get('allSelectedServiceNames');
  }.property('content.serviceName', 'allSelectedServiceNames.@each'),

  /**
   * by default loads data from model otherwise must be overridden as computed property
   * filter services that support server validation and concat with misc configs if Installer or current service
   * @type {Array} - of objects (services)
   */
  services: function() {
    return this.get('content.serviceName')
        ? [App.StackService.find(this.get('content.serviceName'))]
        : this.get('content.services').filter(function(s){
          return (s.get('isSelected') || s.get('isInstalled'));
        }).concat(require("data/service_configs"));
  }.property('content.serviceName', 'content.services', 'content.services.@each.isSelected', 'content.services.@each.isInstalled', 'content.stacks.@each.isSelected'),

  /**
   * by default loads data from model otherwise must be overridden as computed property
   * can be used for service|host configs pages
   * @type {Array} of strings (hostNames)
   */
  hostGroups: function() {
    return this.get('content.recommendationsHostGroups') || blueprintUtils.generateHostGroups(this.get('hostNames'), App.HostComponent.find());
  }.property('content.recommendationsHostGroups', 'hostNames'),

  /**
   * controller that is child of this mixis has to contain stepConfigs
   * @type {Array}
   */
  stepConfigs: null,

  /**
   * @method loadServerSideConfigsRecommendations
   * laod recommendations from server
   * (used only during install)
   * @returns {*}
   */
  loadServerSideConfigsRecommendations: function() {
    if (this.get('recommendationsConfigs')) {
      return $.Deferred().resolve();
    }
    return App.ajax.send({
      'name': 'wizard.step7.loadrecommendations.configs',
      'sender': this,
      'data': {
        stackVersionUrl: App.get('stackVersionURL'),
        hosts: this.get('hostNames'),
        services: this.get('serviceNames'),
        recommendations: this.get('hostGroups')
      },
      'success': 'loadRecommendationsSuccess',
      'error': 'loadRecommendationsError'
    });
  },

  /**
   * @method loadRecommendationsSuccess
   * success callback after loading recommendations
   * (used only during install)
   * @param data
   */
  loadRecommendationsSuccess: function(data) {
    if (!data) {
      console.warn('error while loading default config values');
    }
    this.set("recommendationsConfigs", Em.get(data.resources[0] , "recommendations.blueprint.configurations"));
  },

  loadRecommendationsError: function(jqXHR, ajaxOptions, error, opt) {
    App.ajax.defaultErrorHandler(jqXHR, opt.url, opt.method, jqXHR.status);
    console.error('Load recommendations failed');
  },

  serverSideValidation: function () {
    var deferred = $.Deferred();
    this.set('configValidationFailed', false);
    this.set('configValidationGlobalMessage', []);
    if (this.get('configValidationFailed')) {
      this.warnUser(deferred);
    } else {
      this.runServerSideValidation(deferred);
    }
    return deferred;
  },

  /**
   * @method serverSideValidation
   * send request to validate configs
   * @returns {*}
   */
  runServerSideValidation: function(deferred) {
    var self = this;
    var recommendations = this.get('hostGroups');
    recommendations.blueprint.configurations = blueprintUtils.buildConfisJSON(this.get('services'), this.get('stepConfigs'));

    var serviceNames = this.get('serviceNames');
    if (!self.get('isInstaller')) {
      // When editing a service we validate only that service's configs.
      // However, we should pass the IDs of services installed, or else,
      // default value calculations will alter.
      serviceNames = App.Service.find().mapProperty('serviceName');
    }

    return App.ajax.send({
      name: 'config.validations',
      sender: this,
      data: {
        stackVersionUrl: App.get('stackVersionURL'),
        hosts: this.get('hostNames'),
        services: serviceNames,
        validate: 'configurations',
        recommendations: recommendations
      },
      success: 'validationSuccess',
      error: 'validationError'
    }).complete(function() {
      self.warnUser(deferred);
    });
  },


  /**
   * @method validationSuccess
   * success callback after getting responce from server
   * go through the step configs and set warn and error messages
   * @param data
   */
  validationSuccess: function(data) {
    var self = this;
    var checkedProperties = [];
    var globalWarning = [];
    self.set('configValidationError', false);
    self.set('configValidationWarning', false);
    data.resources.forEach(function(r) {
      r.items.forEach(function(item){
        if (item.type == "configuration") {
          self.get('stepConfigs').forEach(function(service) {
            service.get('configs').forEach(function(property) {
              if ((property.get('filename') == item['config-type'] + '.xml') && (property.get('name') == item['config-name'])) {
                if (item.level == "ERROR") {
                  self.set('configValidationError', true);
                  property.set('errorMessage', item.message);
                  property.set('error', true);
                } else if (item.level == "WARN") {
                  self.set('configValidationWarning', true);
                  property.set('warnMessage', item.message);
                  property.set('warn', true);
                }
                // store property data to detect WARN or ERROR messages for missed property
                if (["ERROR", "WARN"].contains(item.level)) checkedProperties.push(item['config-type'] + '/' + item['config-name']);
              }
            });
          });
          // check if error or warn message detected for property that absent in step configs
          if (["ERROR", "WARN"].contains(item.level) && !checkedProperties.contains(item['config-type'] + '/' + item['config-name'])) {
            var message = {
              propertyName: item['config-name'],
              filename: item['config-type'],
              warnMessage: item.message,
              serviceName: App.StackService.find().filter(function(service) {
                return !!service.get('configTypes')[item['config-type']];
              })[0].get('displayName')
            };
            self.set(item.level == 'WARN' ? 'configValidationWarning' : 'configValidationError', true);
            globalWarning.push(message);
          }
        }
      });
    });
    self.set('configValidationGlobalMessage', globalWarning);
  },

  validationError: function (jqXHR, ajaxOptions, error, opt) {
    this.set('configValidationFailed', true);
    App.ajax.defaultErrorHandler(jqXHR, opt.url, opt.method, jqXHR.status);
    console.error('config validation failed');
  },


  /**
   * warn user if some errors or warning were
   * in seting up configs otherwise go to the nex operation
   * @param deferred
   * @returns {*}
   */
  warnUser: function(deferred) {
    var self = this;
    if (this.get('configValidationFailed')) {
      deferred.reject();
      return App.showAlertPopup(Em.I18n.t('installer.step7.popup.validation.failed.header'), Em.I18n.t('installer.step7.popup.validation.request.failed.body'));
    } else if (this.get('configValidationWarning') || this.get('configValidationError')) {
      // Motivation: for server-side validation warnings and EVEN errors allow user to continue wizard
      return App.ModalPopup.show({
        header: Em. I18n.t('installer.step7.popup.validation.warning.header'),
        classNames: ['sixty-percent-width-modal','modal-full-width'],
        primary: Em.I18n.t('common.proceedAnyway'),
        primaryClass: 'btn-danger',
        marginBottom: 200,
        onPrimary: function () {
          this.hide();
          deferred.resolve();
        },
        onSecondary: function () {
          this.hide();
          deferred.reject("invalid_configs"); // message used to differentiate types of rejections.
        },
        onClose: function () {
          this.hide();
          deferred.reject("invalid_configs"); // message used to differentiate types of rejections.
        },
        bodyClass: Em.View.extend({
          controller: self,
          templateName: require('templates/common/configs/config_recommendation_popup')
        })
      });
    } else {
      deferred.resolve();
    }
  }
});

});

require.register("mixins/common/table_server_mixin", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var validator = require('utils/validator');

App.TableServerMixin = Em.Mixin.create({
  queryParams: [],
  resetStartIndex: false,

  /**
   * filterProps support follow types of filter:
   * MATCH - match of RegExp
   * EQUAL - equality "="
   * LESS - "<"
   * MORE - ">"
   * MULTIPLE - multiple values to compare
   * CUSTOM - substitute values with keys "{#}" in alias
   */
  filterProps: [],

  /**
   * include "from" and "page_size"
   */
  paginationProps: [
    {
      name: 'displayLength',
      key: 'page_size',
      value: '25',
      type: 'EQUAL'
    },
    {
      name: 'startIndex',
      key: 'from',
      value: 0,
      type: 'EQUAL'
    }
  ],

  sortProps: [],

  /**
   * update values of pagination properties from local db and return them
   * @return {array}
   */
  getPaginationProps: function () {
    var displayLength = App.db.getDisplayLength(this.get('name'));
    if (displayLength) {
      this.get('paginationProps').findProperty('name', 'displayLength').value = displayLength;
    }

    var startIndex = App.db.getStartIndex(this.get('name'));
    if (!Em.isNone(startIndex)) {
      startIndex = (startIndex > 0) ? startIndex - 1 : startIndex;
      this.get('paginationProps').findProperty('name', 'startIndex').value = startIndex;
    }
    return this.get('paginationProps');
  },

  /**
   * get sort properties from local db
   * @return {Array}
   */
  getSortProps: function () {
    var savedSortConditions = App.db.getSortingStatuses(this.get('name')) || [],
      sortProperties = this.get('sortProps'),
      sortParams = [];
    savedSortConditions.forEach(function (sort) {
      var property = sortProperties.findProperty('name', sort.name);
      if (property && (sort.status === 'sorting_asc' || sort.status === 'sorting_desc')) {
        property.value = sort.status.replace('sorting_', '');
        property.type = 'SORT';
        sortParams.push(property);
      }
    });
    return sortParams;
  },
  /**
   * get filter properties from local db
   * @return {Array}
   */
  getFilterProps: function () {
    var savedFilterConditions = App.db.getFilterConditions(this.get('name')) || [],
      filterProperties = this.get('filterProps'),
      filterParams = [],
      colPropAssoc = this.get('colPropAssoc');

    savedFilterConditions.forEach(function (filter) {
      var property = filterProperties.findProperty('name', colPropAssoc[filter.iColumn]);

      if (property && filter.value && filter.value.length > 0 && !filter.skipFilter) {
        property.isFilter = true;

        if (filter.type === 'range') {
          //range value should contain array of two element with start and end values accordingly
          filter.value.forEach(function (val, index) {
            if (val) {
              property.type = (index === 0) ? "MORE" : "LESS";
              property.value = val;
              filterParams.push(property);
            }
          });
        } else if (filter.type === 'string') {
          property.value = this.getRegExp(filter.value);
          filterParams.push(property);
        } else {
          property.value = filter.value;
          filterParams.push(property);
        }
      }
    }, this);
    return filterParams;
  },

  getQueryParameters: function () {
    var queryParams = [];

    queryParams.pushObjects(this.getPaginationProps());
    queryParams.pushObjects(this.getSortProps());
    queryParams.pushObjects(this.getFilterProps());
    this.set('queryParams', queryParams);
    return queryParams;
  },

  getRegExp: function (value) {
    value = validator.isValidMatchesRegexp(value) ? value.replace(/(\.+\*?|(\.\*)+)$/, '') + '.*' : '^$';
    value = /^\.\*/.test(value) || value == '^$' ? value : '.*' + value;
    return value;
  }

});

});

require.register("mixins/common/table_server_view_mixin", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Mixin should be used for Em.View of table that uses server to filter, sort, paginate content
 */
App.TableServerViewMixin = Em.Mixin.create({
  filteringComplete: true,
  timeOut: null,
  /**
   * filter delay time, used to combine filters change into one content update
   */
  filterWaitingTime: 500,

  /**
   * count of filtered items
   */
  filteredCount: function () {
    return this.get('controller.filteredCount');
  }.property('controller.filteredCount'),
  /**
   * total count of items
   */
  totalCount: function () {
    return this.get('controller.totalCount');
  }.property('controller.totalCount'),
  /**
   * data requested from server
   */
  content: function () {
    return this.get('controller.content');
  }.property('controller.content'),
  /**
   * content already filtered on server-side
   */
  filteredContent: function () {
    return this.get('content');
  }.property('content'),
  /**
   * sort and slice recieved content by pagination parameters
   */
  pageContent: function () {
    var content = this.get('filteredContent');
    if (content.length > ((this.get('endIndex') - this.get('startIndex')) + 1)) {
      content = content.slice(0, (this.get('endIndex') - this.get('startIndex')) + 1);
    }
    return content.sort(function (a, b) {
      return a.get('index') - b.get('index');
    });
  }.property('filteredContent'),

  /**
   * compute applied filter and run content update from server
   * @param iColumn
   * @param value
   * @param type
   */
  updateFilter: function (iColumn, value, type) {
    var self = this;
    this.set('controller.resetStartIndex', false);
    this.saveFilterConditions(iColumn, value, type, false);
    if (!this.get('filteringComplete')) {
      clearTimeout(this.get('timeOut'));
      this.set('timeOut', setTimeout(function () {
        self.updateFilter(iColumn, value, type);
      }, this.get('filterWaitingTime')));
    } else {
      clearTimeout(this.get('timeOut'));
      this.set('controller.resetStartIndex', true);
      this.refresh();
    }
  },
  /**
   * synchronize properties of view with controller to generate query parameters
   */
  updatePagination: function (key) {
    if (!Em.isNone(this.get('displayLength'))) {
      App.db.setDisplayLength(this.get('controller.name'), this.get('displayLength'));
      this.get('controller.paginationProps').findProperty('name', 'displayLength').value = this.get('displayLength');
    }
    if (!Em.isNone(this.get('startIndex'))) {
      App.db.setStartIndex(this.get('controller.name'), this.get('startIndex'));
      this.get('controller.paginationProps').findProperty('name', 'startIndex').value = this.get('startIndex');
    }

    if (key !== 'SKIP_REFRESH') {
      this.refresh();
    }
  },
  /**
   * when new filter applied page index should be reset to first page
   */
  resetStartIndex: function () {
    if (this.get('controller.resetStartIndex') && this.get('filteredCount') > 0) {
      this.set('startIndex', 1);
    }
  }.observes('controller.resetStartIndex')
});

});

require.register("mixins/common/userPref", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Small mixin for processing user preferences
 * Provide methods to save/load some values in <code>persist</code> storage
 * Save available only for admin users!
 * When using this mixin you should redeclare methods:
 * <ul>
 *   <li>getUserPrefSuccessCallback</li>
 *   <li>getUserPrefErrorCallback</li>
 *   <li>postuserPrefSuccessCallback</li>
 *   <li>postuserPrefErrorCallback</li>
 * </ul>
 * @type {Em.Mixin}
 */
App.UserPref = Em.Mixin.create({

  /**
   * Additional to request data
   * @type {object}
   */
  additionalData: {},

  /**
   * Get persist value from server with persistKey
   * @param {String} key
   */
  getUserPref: function(key) {
    return App.ajax.send({
      name: 'settings.get.user_pref',
      sender: this,
      data: {
        key: key,
        data: this.get('additionalData')
      },
      success: 'getUserPrefSuccessCallback',
      error: 'getUserPrefErrorCallback'
    });
  },

  /**
   * Should be redeclared in objects that use this mixin
   * @param {*} response
   * @param {Object} request
   * @param {Object} data
   * @returns {*}
   */
  getUserPrefSuccessCallback: function (response, request, data) {},

  /**
   * Should be redeclared in objects that use this mixin
   * @param {Object} request
   * @param {Object} ajaxOptions
   * @param {String} error
   */
  getUserPrefErrorCallback: function (request, ajaxOptions, error) {},

  /**
   * Post persist key/value to server, value is object
   * Only for admin users!
   * @param {String} key
   * @param {Object} value
   */
  postUserPref: function (key, value) {
    if (!App.isAccessible('upgrade_ADMIN')) {
      return $.Deferred().reject().promise();
    }
    var keyValuePair = {};
    keyValuePair[key] = JSON.stringify(value);
    return App.ajax.send({
      'name': 'settings.post.user_pref',
      'sender': this,
      'beforeSend': 'postUserPrefBeforeSend',
      'data': {
        'keyValuePair': keyValuePair
      },
      'success': 'postUserPrefSuccessCallback',
      'error': 'postUserPrefErrorCallback'
    });
  },

  /**
   * Should be redeclared in objects that use this mixin
   * @param {*} response
   * @param {Object} request
   * @param {Object} data
   * @returns {*}
   */
  postUserPrefSuccessCallback: function (response, request, data) {},

  /**
   * Should be redeclared in objects that use this mixin
   * @param {Object} request
   * @param {Object} ajaxOptions
   * @param {String} error
   */
  postUserPrefErrorCallback: function(request, ajaxOptions, error) {},

  /**
   * Little log before post request
   * @param {Object} request
   * @param {Object} ajaxOptions
   * @param {Object} data
   */
  postUserPrefBeforeSend: function(request, ajaxOptions, data){
    console.log('BeforeSend to persist: persistKeyValues', data.keyValuePair);
  }

});

});

require.register("mixins/main/host/details/host_components/decommissionable", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var uiEffects = require('utils/ui_effects');

/**
 * Mixin for <code>App.HostComponentView</code>
 * Contains code for processing components with allowed decommission
 * @type {Em.Mixin}
 */
App.Decommissionable = Em.Mixin.create({

  /**
   * Should be redeclared in views that use this mixin
   * @type {String}
   */
  componentForCheckDecommission: '',

  /**
   * Is component in decommission process right know
   * @type {bool}
   */
  isComponentDecommissioning: null,

  /**
   * May conponent be decommissioned
   * @type {bool}
   */
  isComponentDecommissionAvailable: null,

  /**
   * May component be recommissioned
   * @type {bool}
   */
  isComponentRecommissionAvailable: null,

  /**
   * Component with stopped masters can't be docommissioned
   * @type {bool}
   */
  isComponentDecommissionDisable: function() {
    var masterComponent = this.get('content.service.hostComponents').findProperty('componentName', this.get('componentForCheckDecommission'));
    if (masterComponent && masterComponent.get('workStatus') != App.HostComponentStatus.started) return true;
    return this.get('content.service.workStatus') != App.HostComponentStatus.started;
  }.property('content.service.workStatus', 'content.service.hostComponents.@each.workStatus'),

  /**
   * @override App.HostComponentView.isRestartableComponent
   */
  isRestartableComponent: function() {
    return this.get('isComponentDecommissionAvailable') && App.get('components.restartable').contains(this.get('content.componentName'));
  }.property('isComponentDecommissionAvailable'),

  /**
   * Tooltip message shows if decommission/recommission is disabled
   * when masters for current component is down
   */
  decommissionTooltipMessage: function() {
    if (this.get('isComponentDecommissionDisable') && (this.get('isComponentRecommissionAvailable') || this.get('isComponentDecommissionAvailable'))) {
      var decom = this.get('isComponentRecommissionAvailable') ? Em.I18n.t('common.recommission') : Em.I18n.t('common.decommission');
      return Em.I18n.t('hosts.decommission.tooltip.warning').format(decom, App.format.role(this.get('componentForCheckDecommission')));
    }
  }.property('isComponentDecommissionDisable', 'isComponentRecommissionAvailable', 'isComponentDecommissionAvailable', 'componentForCheckDecommission'),
  /**
   * Recalculated component status based on decommission
   * @type {string}
   */
  statusClass: function () {

    //Class when install failed
    if (this.get('workStatus') === App.HostComponentStatus.install_failed) {
      return 'health-status-color-red icon-cog';
    }

    //Class when installing
    if (this.get('workStatus') === App.HostComponentStatus.installing) {
      return 'health-status-color-blue icon-cog';
    }

    if (this.get('isComponentRecommissionAvailable') && (this.get('isStart') || this.get('workStatus') == 'INSTALLED')) {
      return 'health-status-DEAD-ORANGE';
    }

    //For all other cases
    return 'health-status-' + App.HostComponentStatus.getKeyName(this.get('workStatus'));

  }.property('workStatus', 'isComponentRecommissionAvailable', 'isComponentDecommissioning'),

  /**
   * Return host component text status
   * @type {String}
   */
  componentTextStatus: function () {
    var componentTextStatus = this.get('content.componentTextStatus');
    var hostComponent = this.get('hostComponent');
    if (hostComponent) {
      componentTextStatus = hostComponent.get('componentTextStatus');
      if(this.get('isComponentRecommissionAvailable')){
        if(this.get('isComponentDecommissioning')){
          componentTextStatus = Em.I18n.t('hosts.host.decommissioning');
        } else {
          componentTextStatus = Em.I18n.t('hosts.host.decommissioned');
        }
      }
    }
    return componentTextStatus;
  }.property('workStatus','isComponentRecommissionAvailable','isComponentDecommissioning'),

  /**
   * For Stopping or Starting states, also for decommissioning
   * @type {bool}
   */
  isInProgress: function () {
    return (this.get('workStatus') === App.HostComponentStatus.stopping ||
      this.get('workStatus') === App.HostComponentStatus.starting) ||
      this.get('isDecommissioning');
  }.property('workStatus', 'isDecommissioning'),

  /**
   * load Recommission/Decommission status of component
   */
  loadComponentDecommissionStatus: function () {
    return this.getDesiredAdminState();
  },

  /**
   * Get desired_admin_state status from server
   */
  getDesiredAdminState: function(){
    return App.ajax.send({
      name: 'host.host_component.slave_desired_admin_state',
      sender: this,
      data: {
        hostName: this.get('content.hostName'),
        componentName: this.get('content.componentName')
      },
      success: 'getDesiredAdminStateSuccessCallback',
      error: 'getDesiredAdminStateErrorCallback'
    });
  },

  /**
   * pass received value or null to <code>setDesiredAdminState</code>
   * @param {Object} response
   * @returns {String|null}
   */
  getDesiredAdminStateSuccessCallback: function (response) {
    var status = response.HostRoles.desired_admin_state;
    if (status != null) {
      this.setDesiredAdminState(status);
      return status;
    }
    return null;
  },

  /**
   * error callback of <code>getDesiredAdminState</code>
   */
  getDesiredAdminStateErrorCallback: Em.K,

  /**
   * compute decommission state by desiredAdminState
   * @param {Object} status
   */
  setDesiredAdminState: Em.K,

  /**
   * Get component decommission status from server
   * @returns {$.ajax}
   */
  getDecommissionStatus: function() {
    return App.ajax.send({
      name: 'host.host_component.decommission_status',
      sender: this,
      data: {
        hostName: this.get('content.hostName'),
        componentName: this.get('componentForCheckDecommission'),
        serviceName: this.get('content.service.serviceName')
      },
      success: 'getDecommissionStatusSuccessCallback',
      error: 'getDecommissionStatusErrorCallback'
    });
  },

  /**
   * pass received value or null to <code>setDecommissionStatus</code>
   * @param {Object} response
   * @returns {Object|null}
   */
  getDecommissionStatusSuccessCallback: function (response) {
    var statusObject = response.ServiceComponentInfo;
    if ( statusObject != null) {
      statusObject.component_state = response.host_components[0].HostRoles.state;
      this.setDecommissionStatus(statusObject);
      return statusObject;
    }
    return null;
  },

  /**
   * Set null to <code>decommissionedStatusObject</code> if server returns error
   * @returns {null}
   */
  getDecommissionStatusErrorCallback: Em.K,

  /**
   * compute decommission state by component info
   * @param {Object} status
   */
  setDecommissionStatus: Em.K,

  /**
   * set decommission and recommission flags according to status
   * @param status
   */
  setStatusAs: function (status) {
    switch (status) {
      case "INSERVICE":
        this.set('isComponentRecommissionAvailable', false);
        this.set('isComponentDecommissioning', false);
        this.set('isComponentDecommissionAvailable', this.get('isStart'));
        break;
      case "DECOMMISSIONING":
        this.set('isComponentRecommissionAvailable', true);
        this.set('isComponentDecommissioning', true);
        this.set('isComponentDecommissionAvailable', false);
        break;
      case "DECOMMISSIONED":
        this.set('isComponentRecommissionAvailable', true);
        this.set('isComponentDecommissioning', false);
        this.set('isComponentDecommissionAvailable', false);
        break;
      case "RS_DECOMMISSIONED":
        this.set('isComponentRecommissionAvailable', true);
        this.set('isComponentDecommissioning', this.get('isStart'));
        this.set('isComponentDecommissionAvailable', false);
        break;
    }
  },

  /**
   * Do blinking for 1 minute
   */
  doBlinking: function () {
    var workStatus = this.get('workStatus');
    var self = this;
    var pulsate = [App.HostComponentStatus.starting, App.HostComponentStatus.stopping, App.HostComponentStatus.installing].contains(workStatus);
    if (!pulsate) {
      var component = this.get('content');
      if (component && workStatus != "INSTALLED") {
        pulsate = this.get('isDecommissioning');
      }
    }
    if (pulsate && !self.get('isBlinking')) {
      self.set('isBlinking', true);
      uiEffects.pulsate(self.$('.components-health'), 1000, function () {
        self.set('isBlinking', false);
        self.doBlinking();
      });
    }
  },

  /**
   * Start blinking when host component is starting/stopping/decommissioning
   */
  startBlinking: function () {
    this.$('.components-health').stop(true, true);
    this.$('.components-health').css({opacity: 1.0});
    this.doBlinking();
  }.observes('workStatus','isComponentRecommissionAvailable', 'isDecommissioning'),

  didInsertElement: function() {
    this._super();
    this.loadComponentDecommissionStatus();
  },

  /**
   * Update Decommission status only one time when component was changed
   */
  updateDecommissionStatus: function() {
    Em.run.once(this, 'loadComponentDecommissionStatus');
  }.observes('content.workStatus', 'content.passiveState'),


  decommissionView: Em.View.extend({

    templateName: require('templates/main/host/decommission'),

    text: function() {
      return this.get('parentView.isComponentDecommissionAvailable') ? Em.I18n.t('common.decommission') : Em.I18n.t('common.recommission');
    }.property('parentView.isComponentDecommissionAvailable'),

    didInsertElement: function() {
      this._super();
      App.tooltip($("[rel='decommissionTooltip']"));
    },

    click: function() {
      if (!this.get('parentView.isComponentDecommissionDisable')) {
        if (this.get('parentView.isComponentDecommissionAvailable')) {
          this.get('controller').decommission(this.get('parentView.content'));
        } else {
          this.get('controller').recommission(this.get('parentView.content'));
        }
      }
    }
  })
});

});

require.register("mixins/routers/redirections", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Mixin for routers which may redirect user to the some cluster installer steps
 * Based on cluster state
 * @type {Em.Mixin}
 */
App.RouterRedirections = Em.Mixin.create({

  /**
   * List of cluster statuses when it's not installed
   * @type {string[]}
   */
  installerStatuses: ['CLUSTER_NOT_CREATED_1', 'CLUSTER_DEPLOY_PREP_2', 'CLUSTER_INSTALLING_3', 'SERVICE_STARTING_3', 'CLUSTER_INSTALLED_4'],

  /**
   * Redirect user to the proper installer step if needed
   * Based on cluster state:
   *  1. CLUSTER_NOT_CREATED_1 - to current step
   *  2. CLUSTER_DEPLOY_PREP_2 - to step 8
   *  3. CLUSTER_INSTALLING_3/SERVICE_STARTING_3 - to step 9
   *  4. CLUSTER_INSTALLED_4 - to step 10
   * @param {Em.Router} router
   * @param {Object} currentClusterStatus
   * @param {Boolean} isOnInstaller true - user is on the cluster installer, false - something else
   * @method redirectToInstaller
   */
  redirectToInstaller: function (router, currentClusterStatus, isOnInstaller) {

    var installerController = router.get('installerController');
    var path = isOnInstaller ? '' : 'installer.';
    switch (currentClusterStatus.clusterState) {
      case 'CLUSTER_NOT_CREATED_1' :
        var localDb = currentClusterStatus.localdb;
        if (localDb && localDb.Installer && localDb.Installer.currentStep) {
          App.db.data = currentClusterStatus.localdb;
          App.get('router').setAuthenticated(true);
          var controllerName = installerController.get('name');
          var suffixLength = 10;
          var currentStep = App.get('router').getWizardCurrentStep(controllerName.substr(0, controllerName.length - suffixLength));
          installerController.setCurrentStep(currentStep);
        }
        router.transitionTo(path + 'step' + installerController.get('currentStep'));
        break;
      case 'CLUSTER_DEPLOY_PREP_2' :
        installerController.setCurrentStep('8');
        App.db.data = currentClusterStatus.localdb;
        App.get('router').setAuthenticated(true);
        router.transitionTo(path + 'step' + installerController.get('currentStep'));
        break;
      case 'CLUSTER_INSTALLING_3' :
      case 'SERVICE_STARTING_3' :
        if (!installerController.get('isStep9')) {
          installerController.setCurrentStep('9');
        }
        router.transitionTo(path + 'step' + installerController.get('currentStep'));
        break;
      case 'CLUSTER_INSTALLED_4' :
        if (!installerController.get('isStep10')) {
          installerController.setCurrentStep('10');
        }
        App.db.data = currentClusterStatus.localdb;
        App.get('router').setAuthenticated(true);
        router.transitionTo(path + 'step' + installerController.get('currentStep'));
        break;
    }
  }

});
});

require.register("mixins/wizard/addSecurityConfigs", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Mixin for loading and setting secure configs
 *
 * @type {Ember.Mixin}
 */
App.AddSecurityConfigs = Em.Mixin.create({

  kerberosDescriptor: {},

  secureProperties: require('data/HDP2/secure_properties').configProperties,

  secureMapping: require('data/HDP2/secure_mapping'),

  serviceUsersBinding: 'App.router.mainAdminSecurityController.serviceUsers',

  componentsConfig: [
    {
      serviceName: 'OOZIE',
      componentName: 'OOZIE_SERVER',
      configName: 'oozieserver_host'
    },
    {
      serviceName: 'HIVE',
      componentName: 'WEBHCAT_SERVER',
      configName: 'webhcat_server'
    }
  ],

  /**
   * mock users used in testMode
   */
  testModeUsers: [
    {
      name: 'hdfs_user',
      value: 'hdfs'
    },
    {
      name: 'mapred_user',
      value: 'mapred'
    },
    {
      name: 'hbase_user',
      value: 'hbase'
    },
    {
      name: 'hive_user',
      value: 'hive'
    }
  ],

  /**
   * security configs, which values should be modified after APPLY CONFIGURATIONS stage
   */
  secureConfigs: function () {
    var configs = [
      {
        name: 'zookeeper_principal_name',
        serviceName: 'ZOOKEEPER'
      },
      {
        name: 'knox_principal_name',
        serviceName: 'KNOX'
      },
      {
        name: 'storm_principal_name',
        serviceName: 'STORM'
      }
    ];
    if (App.get('isHadoop22Stack')) {
      configs.push({
        name: 'nimbus_principal_name',
        serviceName: 'STORM'
      });
    }
    return configs;
  }.property('App.isHadoop22Stack'),

  secureServices: function () {
    return this.get('content.services');
  }.property('content.services'),

  /**
   * prepare secure configs
   */
  prepareSecureConfigs: function () {
    var configs = this.get('content.serviceConfigProperties');
    this.set('configs', configs);
    this.loadStaticConfigs(); //Hack for properties which are declared in site_properties.js and not able to retrieve values declared in secure_properties.js
    this.loadUsersToConfigs();
    this.loadHostNames();
    this.loadPrimaryNames();
    var uiConfigs = this.loadUiSideSecureConfigs();
    this.set('configs', this.get('configs').concat(uiConfigs));
  },


  /**
   * push users to configs
   */
  loadUsersToConfigs: function () {
    if (!this.get('serviceUsers').length) {
      this.loadUsersFromServer();
    }
    App.router.get('mainAdminSecurityController.serviceUsers').forEach(function (_user) {
      this.get('configs').pushObject(_user);
    }, this);
  },

  /**
   * add component config that contain host name as value
   * @param serviceName
   * @param componentName
   * @param configName
   * @return {Boolean}
   */
  addHostConfig: function (serviceName, componentName, configName) {
    var service = App.Service.find(serviceName);
    var isServiceSecure = this.get('secureServices').someProperty('serviceName', serviceName);

    if (service.get('isLoaded') && isServiceSecure) {
      var hostComponent = service.get('hostComponents').findProperty('componentName', componentName);
      if (hostComponent) {
        var hostName = hostComponent.get('hostName');
        this.get('configs').push({
          id: 'puppet var',
          name: configName,
          value: hostName
        });
        return true;
      }
    }
    return false;
  },

  /**
   * add hosts' names to configs
   */
  loadHostNames: function () {
    var componentsConfig = this.get('componentsConfig');
    componentsConfig.forEach(function (host) {
      this.addHostConfig(host.serviceName, host.componentName, host.configName);
    }, this);
  },

  /**
   * load static configs
   */
  loadStaticConfigs: function () {
    this.get('configs').forEach(function (_property) {
      switch (_property.name) {
        case 'security_enabled':
          _property.value = 'true';
          break;
      }
    }, this);
  },

  /**
   * add principals to properties
   */
  loadPrimaryNames: function () {
    var principalProperties = this.getPrincipalNames();
    principalProperties.forEach(function (_principalProperty) {
      var name = _principalProperty.name.replace('principal', 'primary');
      var value = _principalProperty.value.split('/')[0];
      this.get('configs').push({name: name, value: value});
    }, this);
  },

  /**
   * gather and return properties with "principal_name"
   * @return {Array}
   */
  getPrincipalNames: function () {
    var principalNames = [];
    this.get('configs').forEach(function (_property) {
      if (/principal_name?$/.test(_property.name)) {
        principalNames.push(_property);
      }
    }, this);
    this.get('secureProperties').forEach(function (_secureProperty) {
      if (/principal_name?$/.test(_secureProperty.name)) {
        var principalName = principalNames.findProperty('name', _secureProperty.name);
        if (!principalName) {
          _secureProperty.value = _secureProperty.defaultValue;
          principalNames.push(_secureProperty);
        }
      }
    }, this);
    return principalNames;
  },

  /**
   * load users from server
   */
  loadUsersFromServer: function () {
    if (App.get('testMode')) {
      var serviceUsers = this.get('serviceUsers');
      this.get('testModeUsers').forEach(function (user) {
        user.id = 'puppet var';
        serviceUsers.push(user);
      }, this);
    } else {
      App.router.set('mainAdminSecurityController.serviceUsers', App.db.getSecureUserInfo());
    }
  },

  /**
   * load configs from UI side
   * @return {Array}
   */
  loadUiSideSecureConfigs: function () {
    var uiConfig = [];
    var configs = this.get('secureMapping').filterProperty('foreignKey', null).filter(function (_configProperty) {
      return (App.Service.find().mapProperty('serviceName').contains(_configProperty.serviceName));
    }, this);
    configs.forEach(function (_config) {
      var value = _config.value;
      if (_config.hasOwnProperty('dependedServiceName')) {
        value = this.checkServiceForConfigValue(value, _config.dependedServiceName);
      }
      value = this.getConfigValue(_config.templateName, value, _config.name);
      uiConfig.push({
        "id": "site property",
        "name": _config.name,
        "value": value,
        "filename": _config.filename
      });
    }, this);
    var dependentConfig = this.get('secureMapping').filterProperty('foreignKey');
    dependentConfig.forEach(function (_config) {
      if (App.Service.find().mapProperty('serviceName').contains(_config.serviceName)) {
        this.setConfigValue(_config);
        this.formatConfigName(uiConfig, _config);
        uiConfig.push({
          "id": "site property",
          "name": _config._name || _config.name,
          "value": _config.value,
          "filename": _config.filename
        });
      }
    }, this);
    return uiConfig;
  },

  /**
   * erase template rules from config value if service is not loaded
   * @param value
   * @param services
   * @return {*}
   */
  checkServiceForConfigValue: function (value, services) {
    services.forEach(function (_service) {
      if (!App.Service.find(_service.name).get('isLoaded')) {
        value = value.replace(_service.replace, '');
      }
    }, this);
    return value;
  },

  /**
   * Set all site property that are derived from other puppet-variable
   * @param templateName
   * @param expression
   * @param name
   * @return {String|null}
   */
  getConfigValue: function (templateName, expression, name) {
    var express = expression.match(/<(.*?)>/g);
    var value = expression;
    if (Em.isNone(express)) return expression;

    express.forEach(function (_express) {
      var index = parseInt(_express.match(/\[([\d]*)(?=\])/)[1]);
      var configs = this.get('configs').findProperty('name', templateName[index]);
      var configValue = templateName[index] == 'hive_metastore' ?
        configs.value.map(function (hostName) {
          return 'thrift://' + hostName + ':9083';
        }).join(',') : configs.value;

      if (!!value) {
        value = (configs) ? App.config.replaceConfigValues(name, _express, value, configValue) : null;
      }
    }, this);
    return value;
  },

  /**
   * format name of config values of configs which match foreignKey
   * @param uiConfig
   * @param config
   * @return {Boolean}
   */
  formatConfigName: function (uiConfig, config) {
    if (Em.isNone(config.value)) return false;

    var fkValue = config.name.match(/<(foreignKey.*?)>/g);
    if (fkValue) {
      fkValue.forEach(function (_fkValue) {
        var index = parseInt(_fkValue.match(/\[([\d]*)(?=\])/)[1]);
        var value;
        if (uiConfig.someProperty('name', config.foreignKey[index])) {
          value = uiConfig.findProperty('name', config.foreignKey[index]).value;
          config._name = config.name.replace(_fkValue, value);
        } else if (this.get('configs').someProperty('name', config.foreignKey[index])) {
          value = this.get('configs').findProperty('name', config.foreignKey[index]).value;
          config._name = config.name.replace(_fkValue, value);
        }
      }, this);
      return true;
    }
    return false;
  },

  /**
   * Set config value with values of configs which match template
   * @param config
   * @return {Boolean}
   */
  setConfigValue: function (config) {
    if (Em.isNone(config.value)) return false;

    //For properties in the configMapping file having foreignKey and templateName properties.
    var templateValue = config.value.match(/<(templateName.*?)>/g);
    if (templateValue) {
      templateValue.forEach(function (_value) {
        var index = parseInt(_value.match(/\[([\d]*)(?=\])/)[1]);
        var cfg = this.get('configs').findProperty('name', config.templateName[index]);

        if (cfg) {
          var cfgValue = config.templateName[index] == 'hive_metastore' ? cfg.value.join(',') : cfg.value;
          config.value = config.value.replace(_value, cfgValue);
        } else {
          config.value = null;
        }
      }, this);
      return true;
    }
    return false;
  },

  /**
   * set value of principal property
   * @param serviceName
   * @param principalName
   * @return {Boolean}
   */
  setPrincipalValue: function (serviceName, principalName) {
    var siteProperties = this.get('configs');

    var realmName = siteProperties.findProperty('name', 'kerberos_domain');

    if (this.get('secureServices').someProperty('serviceName', serviceName)) {
      var principalProperty = siteProperties.findProperty('name', principalName);
      principalProperty.value = principalProperty.value + '@' + realmName.value;
      return true;
    }
    return false;
  },

  /**
   * Generate stack descriptor configs.
   *
   * @returns {$.Deferred}
   */
  getDescriptorConfigs: function () {
    return this.loadDescriptorConfigs().pipe(this.createServicesStackDescriptorConfigs.bind(this));
  },

  /**
   *
   * @param {object[]} items - stack descriptor json response
   * @returns {App.ServiceConfigProperty[]}
   */
  createServicesStackDescriptorConfigs: function (items) {
    var self = this;
    var configs = [];
    var clusterConfigs = [];
    var kerberosDescriptor = items.artifact_data;
    this.set('kerberosDescriptor', kerberosDescriptor);
    // generate configs for root level properties object, currently realm, keytab_dir
    clusterConfigs = clusterConfigs.concat(this.expandKerberosStackDescriptorProps(kerberosDescriptor.properties, 'Cluster', 'stackConfigs'));
    // generate configs for root level identities object, currently spnego property
    clusterConfigs = clusterConfigs.concat(this.createConfigsByIdentities(kerberosDescriptor.identities, 'Cluster'));
    kerberosDescriptor.services.forEach(function (service) {
      var serviceName = service.name;
      // generate configs for service level identity objects
      configs = configs.concat(self.createResourceConfigs(service, serviceName));
      // generate configs for service component level identity  object
      service.components.forEach(function (component) {
        configs = configs.concat(self.createResourceConfigs(component, serviceName));
      });
    });
    // unite cluster, service and component configs
    configs = configs.concat(clusterConfigs);
    self.processConfigReferences(kerberosDescriptor, configs);
    return configs;
  },

  /**
   *
   * @param {Object} resource
   * @param {String} serviceName
   * @return {Array}
   */
  createResourceConfigs: function (resource, serviceName) {
    var identityConfigs = [];
    var resourceConfigs = [];
    if (resource.identities) {
      identityConfigs = this.createConfigsByIdentities(resource.identities, serviceName);
    }
    if (resource.configurations) {
      resource.configurations.forEach(function (_configuration) {
        for (var key in _configuration) {
          resourceConfigs = resourceConfigs.concat(this.expandKerberosStackDescriptorProps(_configuration[key], serviceName, key));
        }
      }, this);
    }
    return identityConfigs.concat(resourceConfigs);
  },

  /**
   * Create service properties based on component identity
   *
   * @param {object[]} identities
   * @param {string} serviceName
   * @returns {App.ServiceConfigProperty[]}
   */
  createConfigsByIdentities: function (identities, serviceName) {
    var self = this;
    var configs = [];
    identities.forEach(function (identity) {
      var defaultObject = {
        isOverridable: false,
        isVisible: true,
        isSecureConfig: true,
        serviceName: serviceName,
        name: identity.name,
        identityType: identity.principal && identity.principal.type
      };
      self.parseIdentityObject(identity).forEach(function (item) {
        configs.push(App.ServiceConfigProperty.create($.extend({}, defaultObject, item)));
      });
    });

    return configs;
  },

  /**
   * Bootstrap base object according to identity info. Generate objects will be converted to
   * App.ServiceConfigProperty model class instances.
   *
   * @param {object} identity
   * @returns {object[]}
   */
  parseIdentityObject: function (identity) {
    var result = [];
    var name = identity.name;
    var keys = Em.keys(identity).without('name');
    keys.forEach(function (item) {
      var configObject = {};
      var prop = identity[item];
      var itemValue = prop[{keytab: 'file', principal: 'value'}[item]];
      // skip inherited property without `configuration` and `keytab` or `file` values
      if (!prop.configuration && !itemValue) return;
      // inherited property with value should not observe value from reference
      if (name.startsWith('/') && !itemValue) {
        configObject.referenceProperty = name.substring(1) + ':' + item;
        configObject.isEditable = false;
      }
      configObject.defaultValue = configObject.value = itemValue;
      configObject.filename = prop.configuration ? prop.configuration.split('/')[0] : 'cluster-env';
      configObject.name = prop.configuration ? prop.configuration.split('/')[1] : name + '_' + item;
      configObject.displayName = configObject.filename == "cluster-env" ? App.format.normalizeName(configObject.name) : configObject.name;
      result.push(configObject);
    });
    return result;
  },

  /**
   * Wrap kerberos properties to App.ServiceConfigProperty model class instances.
   *
   * @param {object} kerberosProperties
   * @param {string} serviceName
   * @param {string} filename
   * @returns {App.ServiceConfigProperty[]}
   */
  expandKerberosStackDescriptorProps: function (kerberosProperties, serviceName, filename) {
    var configs = [];

    for (var propertyName in kerberosProperties) {
      var propertyObject = {
        name: propertyName,
        value: kerberosProperties[propertyName],
        defaultValue: kerberosProperties[propertyName],
        serviceName: serviceName,
        filename: filename,
        displayName: serviceName == "Cluster" ? App.format.normalizeName(propertyName) : propertyName,
        isOverridable: false,
        isEditable: propertyName != 'realm',
        isSecureConfig: true
      };
      configs.push(App.ServiceConfigProperty.create(propertyObject));
    }

    return configs;
  },


  /**
   * Take care about configs that should observe value from referenced configs.
   * Reference is set with `referenceProperty` key.
   *
   * @param {object[]} kerberosDescriptor
   * @param {App.ServiceConfigProperty[]} configs
   */
  processConfigReferences: function (kerberosDescriptor, configs) {
    var identities = kerberosDescriptor.identities;
    identities = identities.concat(kerberosDescriptor.services.map(function (service) {
      var _identities = service.identities || [];
      if (service.components && !!service.components.length) {
        identities = identities.concat(service.components.mapProperty('identities').reduce(function (p, c) {
          return p.concat(c);
        }, []));
        return identities;
      }
    }).reduce(function (p, c) {
      return p.concat(c);
    }, []));
    // clean up array
    identities = identities.compact().without(undefined);
    configs.forEach(function (item) {
      var reference = item.get('referenceProperty');
      if (!!reference) {
        var identity = identities.findProperty('name', reference.split(':')[0])[reference.split(':')[1]];
        if (identity && !!identity.configuration) {
          item.set('observesValueFrom', identity.configuration.split('/')[1]);
        } else {
          item.set('observesValueFrom', reference.replace(':', '_'));
        }
      }
    });
  },

  /**
   * update the kerberos descriptor to be put on cluster resource with user customizations
   * @param kerberosDescriptor {Object}
   * @param configs {Object}
   */
  updateKerberosDescriptor: function (kerberosDescriptor, configs) {
    configs.forEach(function (_config) {
      var isConfigUpdated;
      var isStackResouce = true;
      isConfigUpdated = this.updateResourceIdentityConfigs(kerberosDescriptor, _config, isStackResouce);
      if (!isConfigUpdated) {
        kerberosDescriptor.services.forEach(function (_service) {
          isConfigUpdated = this.updateResourceIdentityConfigs(_service, _config);
          if (!isConfigUpdated) {
            _service.components.forEach(function (_component) {
              isConfigUpdated = this.updateResourceIdentityConfigs(_component, _config);
            }, this);
          }
        }, this);
      }
    }, this);
  },

  /**
   * Updates the identity configs or configurations at a resource. A resource could be
   * 1) Stack
   * 2) Service
   * 3) Component
   * @param resource
   * @param config
   * @param isStackResource
   * @return boolean
   */
  updateResourceIdentityConfigs: function (resource, config, isStackResource) {
    var isConfigUpdated;
    var identities = resource.identities;
    var properties = !!isStackResource ? resource.properties : resource.configurations;
    isConfigUpdated = this.updateDescriptorConfigs(properties, config);
    if (!isConfigUpdated) {
      if (identities) {
        isConfigUpdated = this.updateDescriptorIdentityConfig(identities, config);
      }
    }
    return isConfigUpdated;
  },

  /**
   *
   * @param configurations
   * @param config
   * @return boolean
   */
  updateDescriptorConfigs: function (configurations, config) {
    var isConfigUpdated;
    if (!!configurations) {
      if (Array.isArray(configurations)) {
        configurations.forEach(function (_configuration) {
          for (var key in _configuration) {
            if (Object.keys(_configuration[key]).contains(config.name) && config.filename === key) {
              _configuration[key][config.name] = config.value;
              isConfigUpdated = true
            }
          }
        }, this);
      } else if (Object.keys(configurations).contains(config.name) && config.filename === 'stackConfigs') {
        configurations[config.name] = config.value;
        isConfigUpdated = true
      }
    }
    return isConfigUpdated;
  },


  /**
   *
   * @param identities
   * @param config
   * @return boolean
   */
  updateDescriptorIdentityConfig: function (identities, config) {
    var isConfigUpdated = false;
    identities.forEach(function (identity) {
      var keys = Em.keys(identity).without('name');
      keys.forEach(function (item) {
        var prop = identity[item];
        if (prop.configuration && prop.configuration.split('/')[0] === config.filename &&
          prop.configuration.split('/')[1] === config.name) {
          prop[{keytab: 'file', principal: 'value'}[item]] = config.value;
          isConfigUpdated = true;
        }
      });
    }, this);
    return isConfigUpdated;
  },

  /**
   * Make request for stack descriptor configs if cluster  is not secure
   * or cluster descriptor configs if cluster is secure
   * @returns {$.ajax}
   * @method loadStackDescriptorConfigs
   */
  loadDescriptorConfigs: function() {
    if (App.router.get('mainAdminKerberosController.securityEnabled')) {
      return this.loadClusterDescriptorConfigs();
    } else {
      return this.loadStackDescriptorConfigs();
    }
  },
  /**
   * Make request for stack descriptor configs.
   * @returns {$.ajax}
   * @method loadStackDescriptorConfigs
   */
  loadStackDescriptorConfigs: function () {
    return App.ajax.send({
      sender: this,
      name: 'admin.kerberize.stack_descriptor',
      data: {
        stackName: App.get('currentStackName'),
        stackVersionNumber: App.get('currentStackVersionNumber')
      }
    });
  },

  /**
   * Make request for cluster descriptor configs.
   * @returns {$.ajax}
   * @method loadClusterDescriptorConfigs
   */
  loadClusterDescriptorConfigs: function () {
    return App.ajax.send({
      sender: this,
      name: 'admin.kerberize.cluster_descriptor'
    });
  }
});

});

require.register("mixins/wizard/selectHost", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Mixin for host select view on step 5 install wizard
 * Implements base login for select, input types
 * Each view which use this mixin should redefine method <code>changeHandler</code> (with
 * observing <code>controller.hostNameCheckTrigger</code>)
 *
 * @type {Ember.Mixin}
 */
App.SelectHost = Em.Mixin.create({

  /**
   * Element of <code>controller.servicesMasters</code>
   * Binded from template
   * @type {object}
   */
  component: null,

  /**
   * List of avaiable host names
   * @type {string[]}
   */
  content: [],

  /**
   * Host component name
   * @type {string}
   */
  componentName: null,

  /**
   * Handler for selected value change
   * Triggers <code>changeHandler</code> execution
   * @method change
   */
  change: function () {
    if ('destroyed' === this.get('state')) return;
    this.set('controller.lastChangedComponent', this.get('component.component_name'));
    this.get('controller').toggleProperty('hostNameCheckTrigger');
  },

  /**
   * Add or remove <code>error</code> class from parent div-element
   * @param {bool} flag true - add class, false - remove
   * @method updateErrorStatus
   */
  updateErrorStatus: function(flag) {
    var parentBlock = this.$().parent('div');
    /* istanbul ignore next */
    if (flag) {
      parentBlock.removeClass('error');
    }
    else {
      parentBlock.addClass('error');
    }
  },

  /**
   * If component is multiple (defined in <code>controller.multipleComponents</code>),
   * should update <code>controller.componentToRebalance</code> and trigger rebalancing
   * @method tryTriggerRebalanceForMultipleComponents
   */
  tryTriggerRebalanceForMultipleComponents: function() {
    var componentIsMultiple = this.get('controller.multipleComponents').contains(this.get("component.component_name"));
    if(componentIsMultiple) {
      this.get('controller').set('componentToRebalance', this.get("component.component_name"));
      this.get('controller').incrementProperty('rebalanceComponentHostsCounter');
    }
  },

  /**
   * Handler for value changing
   * Should be redeclared in child views
   * Each redeclarion should contains code:
   * <code>
   *   if (!this.shouldChangeHandlerBeCalled()) return;
   * </code>
   * @method changeHandler
   */
  changeHandler: function() {}.observes('controller.hostNameCheckTrigger'),

  /**
   * If view is destroyed or not current component's host name was changed we should do nothing
   * This method should be called from <code>changeHandler</code>:
   * <code>
   *   changeHandler: function() {
   *     if (!this.shouldChangeHandlerBeCalled()) return;
   *     // your code
   *   }.observes(...)
   * </code>
   * @returns {boolean}
   * @method shouldChangeHandlerBeCalled
   */
  shouldChangeHandlerBeCalled: function() {
    return !(('destroyed' === this.get('state')) || (this.get('controller.lastChangedComponent') !== this.get("component.component_name")));
  },

  /**
   * If <code>component.isHostNameValid</code> was changed,
   * error status should be updated according to new value
   * @method isHostNameValidObs
   */
  isHostNameValidObs: function() {
    this.updateErrorStatus(this.get('component.isHostNameValid'));
  }.observes('component.isHostNameValid'),

  /**
   * Recalculate available hosts
   * This should be done only once per Ember loop
   * @method rebalanceComponentHosts
   */
  rebalanceComponentHosts: function () {
    Em.run.once(this, 'rebalanceComponentHostsOnce');
  }.observes('controller.rebalanceComponentHostsCounter'),

  /**
   * Recalculate available hosts
   * @method rebalanceComponentHostsOnce
   */
  rebalanceComponentHostsOnce: function() {
    if (this.get('component.component_name') === this.get('controller.componentToRebalance')) {
      this.initContent();
    }
  },

  /**
   * Get available hosts
   * multipleComponents component can be assigned to multiple hosts,
   * shared hosts among the same component should be filtered out
   * @return {string[]}
   * @method getAvailableHosts
   */
  getAvailableHosts: function () {
    var hosts = this.get('controller.hosts').slice(),
      componentName = this.get('component.component_name'),
      multipleComponents = this.get('controller.multipleComponents'),
      occupiedHosts = this.get('controller.selectedServicesMasters')
        .filterProperty('component_name', componentName)
        .mapProperty('selectedHost')
        .without(this.get('component.selectedHost'));

    if (multipleComponents.contains(componentName)) {
      return hosts.filter(function (host) {
        return !occupiedHosts.contains(host.get('host_name'));
      }, this);
    }
    return hosts;
  },

  /**
   * Extract hosts from controller,
   * filter out available to selection and
   * push them into form element content
   * @method initContent
   */
  initContent: function () {
    this.set("content", this.getAvailableHosts());
  }

});

});

require.register("mixins/wizard/wizardDeployProgressController", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Mixin for wizard controller for showing command progress on wizard pages
 * This should
 * @type {Ember.Mixin}
 */
App.wizardDeployProgressControllerMixin = Em.Mixin.create({

  /**
   * Ajax-requests count
   * @type {number}
   */
  ajaxQueueLength: 0,

  /**
   * Ajax-requests queue
   * @type {App.ajaxQueue}
   */
  ajaxRequestsQueue: null,

  /**
   * This flag when turned to true launches deploy progress bar
   */
  isDeployStarted: '',


  /**
   * We need to do a lot of ajax calls async in special order. To do this,
   * generate array of ajax objects and then send requests step by step. All
   * ajax objects are stored in <code>ajaxRequestsQueue</code>
   *
   * @param {Object} params object with ajax-request parameters like url, type, data etc
   * @method addRequestToAjaxQueue
   */
  addRequestToAjaxQueue: function (params) {
    if (App.get('testMode')) return;

    params = jQuery.extend({
      sender: this,
      error: 'ajaxQueueRequestErrorCallback'
    }, params);
    params.data['cluster'] = this.get('clusterName');

    this.get('ajaxRequestsQueue').addRequest(params);
  },

  /**
   * Navigate to next step after all requests are sent
   * @method ajaxQueueFinished
   */
  ajaxQueueFinished: function () {
    console.log('everything is loaded');
    App.router.send('next');
  },

  /**
   * Error callback for each queued ajax-request
   * @param {object} xhr
   * @param {string} status
   * @param {string} error
   * @method ajaxQueueRequestErrorCallback
   */
  ajaxQueueRequestErrorCallback: function (xhr, status, error) {
    var responseText = JSON.parse(xhr.responseText);
    var controller = App.router.get(App.clusterStatus.wizardControllerName);
    controller.registerErrPopup(Em.I18n.t('common.error'), responseText.message);
    this.set('hasErrorOccurred', true);
    // an error will break the ajax call chain and allow submission again
    this.set('isSubmitDisabled', false);
    this.set('isBackBtnDisabled', false);
    App.router.get(this.get('content.controllerName')).setStepsEnable();
  }
});



});

require.register("mixins/wizard/wizardDeployProgressView", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Mixin for wizard view for showing command progress on wizard pages
 * This should
 * @type {Ember.Mixin}
 */
App.wizardDeployProgressViewMixin = Em.Mixin.create({

  /**
   * Should ajax-queue progress bar be displayed
   * @method showLoadingIndicator
   */
  showLoadingIndicator: function () {
    if (!this.get('controller.isSubmitDisabled') || App.get('testMode')) {
      if (this.get('modalPopup')) {
        this.get('modalPopup').hide();
        this.set('modalPopup', null);
      }
      return;
    }
    // don't create popup if it already exists
    if (this.get('modalPopup')) {
      return;
    }
    this.set('modalPopup', App.ModalPopup.show({

      header: '',

      showFooter: false,

      showCloseButton: false,

      bodyClass: Em.View.extend({

        templateName: require('templates/wizard/step8/step8_log_popup'),

        controllerBinding: 'App.router.wizardStep8Controller',

        /**
         * Css-property for progress-bar
         * @type {string}
         */
        barWidth: '',

        /**
         * Popup-message
         * @type {string}
         */
        message: '',

        /**
         * Set progress bar width and popup message when ajax-queue requests are proccessed
         * @method ajaxQueueChangeObs
         */
        ajaxQueueChangeObs: function () {
          var length = this.get('controller.ajaxQueueLength');
          var left = this.get('controller.ajaxRequestsQueue.queue.length');
          this.set('barWidth', 'width: ' + ((length - left) / length * 100) + '%;');
          this.set('message', Em.I18n.t('installer.step8.deployPopup.message').format((length - left), length));
        }.observes('controller.ajaxQueueLength', 'controller.ajaxRequestsQueue.queue.length'),

        /**
         * Hide popup when ajax-queue is finished
         * @method autoHide
         */
        autoHide: function () {
          if (this.get('controller.servicesInstalled')) {
            this.get('parentView').hide();
          }
        }.observes('controller.servicesInstalled')
      })

    }));
  }.observes('controller.isDeployStarted')

});


});

require.register("mixins/wizard/wizardProgressPageController", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Mixin for wizard controller for showing command progress on wizard pages
 * This should
 * @type {Ember.Mixin}
 */
App.wizardProgressPageControllerMixin = Em.Mixin.create({
  controllerName: '',
  clusterDeployState: 'WIZARD_DEPLOY',
  status: 'IN_PROGRESS',
  tasks: [],
  commands: [],
  currentRequestIds: [], //todo: replace with using requestIds from tasks
  logs: [],
  currentTaskId: null,
  POLL_INTERVAL: 4000,
  isSubmitDisabled: true,
  isBackButtonDisabled: true,
  stages: [],
  /**
   * List of statuses that inform about the end of request progress.
   * @type {String[]}
   */
  completedStatuses: ['COMPLETED', 'FAILED', 'TIMEDOUT', 'ABORTED'],
  currentPageRequestId: null,
  isSingleRequestPage: false,
  isCommandLevelRetry: function () {
    return !this.get('isSingleRequestPage');
  }.property('isSingleRequestPage'),
  showRetry: false,
  /**
   * Show whether tasks data was loaded
   * @type {Boolean}
   */
  isLoading: false,

  k: Em.K,
  /**
   *  tasksMessagesPrefix should be overloaded by any controller including the mixin
   */
  tasksMessagesPrefix: '',

  loadStep: function () {
    this.clearStep();
    var self = this;
    if (!self.isSingleRequestPage) {
      this.initStep();
    } else {
      var requestIds = this.get('content.tasksRequestIds');
      var currentRequestId = requestIds && requestIds[0][0];
      if (!currentRequestId) {
        this.set('isLoaded', false);
        this.submitRequest();
      } else {
        self.set('currentPageRequestId', currentRequestId);
        self.doPollingForPageRequest();
      }
    }
  },

  initStep: function () {
    this.initializeTasks();
    if (!this.isSingleRequestPage) {
      this.loadTasks();
    }
    this.addObserver('tasks.@each.status', this, 'onTaskStatusChange');
    if (this.isSingleRequestPage) {
      var dfd = $.Deferred();
      var dfdObject = {
        deferred: dfd,
        isJqueryPromise: true
      };
      this.onTaskStatusChange(dfdObject);
      return dfd.promise();
    } else {
      this.onTaskStatusChange();
    }
  },

  clearStep: function () {
    this.removeObserver('tasks.@each.status', this, 'onTaskStatusChange');
    this.set('isSubmitDisabled', true);
    this.set('isBackButtonDisabled', true);
    this.set('tasks', []);
    this.set('currentRequestIds', []);
    this.set('isLoaded', false);
  },

  /**
   * Clear stages info for single page request.
   */
  clearStage: function() {
    this.setDBProperty('tasksRequestIds', null);
    this.setDBProperty('tasksStatuses', null);
    this.set('showRetry', false);
    this.set('content.tasksRequestIds', null);
    this.set('content.tasksStatuses', null);
    this.set('content.currentTaskId', null);
    this.get('stages').clear();
  },

  retry: function () {
    this.set('showRetry', false);
    this.get('tasks').setEach('status','PENDING');
    this.loadStep();
  },

  submitRequest: function () {
    return App.ajax.send({
      name: this.get('request.ajaxName'),
      data: this.get('request.ajaxData'),
      sender: this,
      error: 'onSingleRequestError',
      success: 'submitRequestSuccess',
      kdcCancelHandler: 'failTaskOnKdcCheck'
    });
  },

  submitRequestSuccess: function(data, result, request) {
    if (data) {
      this.set('currentPageRequestId', data.Requests.id);
      this.doPollingForPageRequest();
    } else {
      //Step has been successfully completed
      if (request.status === 200) {
        this.set('status', 'COMPLETED');
        this.set('isSubmitDisabled', false);
        this.set('isLoaded', true);
      }
    }
  },

  failTaskOnKdcCheck: function() {
    this.set('status', 'FAILED');
    this.set('isLoaded', true);
    this.set('showRetry', true);
  },

  doPollingForPageRequest: function () {
    App.ajax.send({
      name: 'admin.poll.kerberize.cluster.request',
      sender: this,
      data: {
        requestId: this.get('currentPageRequestId')
      },
      success: 'initializeStages'
    });
  },

  initializeStages: function (data) {
    var self = this;
    var stages = [];
    this.set('logs', []);
    data.stages.forEach(function (_stage) {
      stages.pushObject(Em.Object.create(_stage.Stage));
    }, this);
    if (!this.get('stages').length) {
      this.get('stages').pushObjects(stages);
      this.initStep().done(function(){
        self.updatePageWithPolledData(data);
      });
    } else {
      this.updatePageWithPolledData(data);
    }
  },

  updatePageWithPolledData: function(data) {
    // If all tasks completed no need to update each task status.
    // Preferable to skip polling of data for completed tasks after page refresh.
    if (this.get('status') === 'COMPLETED') return;
    var self = this;
    var tasks = [];
    var currentPageRequestId = this.get('currentPageRequestId');
    var currentTaskId = this.get('currentTaskId');
    var currentTask = this.get('tasks').findProperty('id', currentTaskId);
    var currentStage =  data.stages.findProperty('Stage.stage_id', currentTask.get('stageId'));
    var tasksInCurrentStage =  currentStage.tasks;
    this.set('logs',tasksInCurrentStage);

    this.setRequestIds(this.get('currentTaskId'), [this.get('currentPageRequestId')]);

    if (!tasksInCurrentStage.someProperty('Tasks.status', 'PENDING') && !tasksInCurrentStage.someProperty('Tasks.status', 'QUEUED') && !tasksInCurrentStage.someProperty('Tasks.status', 'IN_PROGRESS')) {
      this.set('currentRequestIds', []);
      if (tasksInCurrentStage.someProperty('Tasks.status', 'FAILED') || tasksInCurrentStage.someProperty('Tasks.status', 'TIMEDOUT') || tasksInCurrentStage.someProperty('Tasks.status', 'ABORTED')) {
        this.setTaskStatus(currentTaskId, 'FAILED');
      } else {
        this.setTaskStatus(currentTaskId, 'COMPLETED');
      }
    } else {
      var completedActions = tasksInCurrentStage.filterProperty('Tasks.status', 'COMPLETED').length
        + tasksInCurrentStage.filterProperty('Tasks.status', 'FAILED').length
        + tasksInCurrentStage.filterProperty('Tasks.status', 'ABORTED').length
        + tasksInCurrentStage.filterProperty('Tasks.status', 'TIMEDOUT').length;
      var queuedActions = tasksInCurrentStage.filterProperty('Tasks.status', 'QUEUED').length;
      var inProgressActions = tasksInCurrentStage.filterProperty('Tasks.status', 'IN_PROGRESS').length;
      var progress = Math.floor(((queuedActions * 0.09) + (inProgressActions * 0.35) + completedActions ) / tasksInCurrentStage.length * 100);
      this.get('tasks').findProperty('id', currentTaskId).set('progress', progress);
    }

    // start polling if current request not completed
    if (!(this.get('completedStatuses').contains(this.get('status')))) {
      window.setTimeout(function () {
        self.doPollingForPageRequest();
      }, self.POLL_INTERVAL);
    }
  },

  initializeTasks: function () {
    var self = this;
    var commands = this.isSingleRequestPage ? this.get('stages') : this.get('commands');
    var currentStep = App.router.get(this.get('content.controllerName') + '.currentStep');
    var tasksMessagesPrefix = this.get('tasksMessagesPrefix');
    // check that all stages have been completed for single request type
    var allStagesCompleted = commands.everyProperty('status', 'COMPLETED');
    for (var i = 0; i < commands.length; i++) {
      this.get('tasks').pushObject(Ember.Object.create({
        title: self.isSingleRequestPage ? commands[i].get('context') : Em.I18n.t(tasksMessagesPrefix + currentStep + '.task' + i + '.title'),
        // set COMPLETED status for task if all stages completed successfully
        status: allStagesCompleted ? 'COMPLETED' : 'PENDING',
        id: i,
        stageId: self.isSingleRequestPage ? commands[i].get('stage_id') : null,
        command: self.isSingleRequestPage ? 'k' : commands[i],
        showRetry: false,
        showRollback: false,
        name: self.isSingleRequestPage ? commands[i].get('context') : Em.I18n.t(tasksMessagesPrefix + currentStep + '.task' + i + '.title'),
        displayName: self.isSingleRequestPage ? commands[i].get('context') : Em.I18n.t(tasksMessagesPrefix + currentStep + '.task' + i + '.title'),
        progress: 0,
        isRunning: false,
        requestIds: self.isSingleRequestPage ? [this.get('stages')[0].request_id] : []
      }));
    }
    this.set('isLoaded', true);
  },

  loadTasks: function () {
    var self = this;
    var loadedStatuses = this.get('content.tasksStatuses');
    var loadedRequestIds = this.get('content.tasksRequestIds');
    if (loadedStatuses && loadedStatuses.length === this.get('tasks').length) {
      this.get('tasks').forEach(function (task, i) {
        self.setTaskStatus(task.get('id'), loadedStatuses[i]);
        self.setRequestIds(task.get('id'), loadedRequestIds[i]);
      });
      if (loadedStatuses.contains('IN_PROGRESS')) {
        var curTaskId = this.get('tasks')[loadedStatuses.indexOf('IN_PROGRESS')].get('id');
        this.set('currentRequestIds', this.get('content.requestIds'));
        this.set('currentTaskId', curTaskId);
        this.doPolling();
      } else if (loadedStatuses.contains('QUEUED')) {
        var curTaskId = this.get('tasks')[loadedStatuses.indexOf('QUEUED')].get('id');
        this.set('currentTaskId', curTaskId);
        this.runTask(curTaskId);
      }
    }
  },

  setTaskStatus: function (taskId, status) {
    this.get('tasks').findProperty('id', taskId).set('status', status);
  },

  setRequestIds: function (taskId, requestIds) {
    this.get('tasks').findProperty('id', taskId).set('requestIds', requestIds);
  },

  retryTask: function () {
    var task = this.get('tasks').findProperty('status', 'FAILED');
    task.set('showRetry', false);
    task.set('showRollback', false);
    task.set('status', 'PENDING');
  },

  onTaskStatusChange: function (dfdObject) {
    var statuses = this.get('tasks').mapProperty('status');
    var tasksRequestIds = this.get('tasks').mapProperty('requestIds');
    var requestIds = this.get('currentRequestIds');
    // save task info
    App.router.get(this.get('content.controllerName')).saveTasksStatuses(statuses);
    App.router.get(this.get('content.controllerName')).saveTasksRequestIds(tasksRequestIds);
    App.router.get(this.get('content.controllerName')).saveRequestIds(requestIds);
    // call saving of cluster status asynchronous
    // synchronous executing cause problems in Firefox
    var successCallbackData;
    if (dfdObject && dfdObject.isJqueryPromise) {
      successCallbackData =  {deferred: dfdObject.deferred};
    }
    App.clusterStatus.setClusterStatus({
      clusterName: App.router.getClusterName(),
      clusterState: this.get('clusterDeployState'),
      wizardControllerName: this.get('content.controllerName'),
      localdb: App.db.data
    }, {successCallback: this.statusChangeCallback, sender: this, successCallbackData: successCallbackData});
  },

  /**
   * Method that called after saving persist data to server.
   * Switch task according its status.
   */
  statusChangeCallback: function (data) {
    if (!this.get('tasks').someProperty('status', 'IN_PROGRESS') && !this.get('tasks').someProperty('status', 'QUEUED') && !this.get('tasks').someProperty('status', 'FAILED')) {
      var nextTask = this.get('tasks').findProperty('status', 'PENDING');
      if (nextTask) {
        this.set('status', 'IN_PROGRESS');
        var taskStatus = this.isSingleRequestPage ? 'IN_PROGRESS' : 'QUEUED';
        this.setTaskStatus(nextTask.get('id'), taskStatus);
        this.set('currentTaskId', nextTask.get('id'));
        this.runTask(nextTask.get('id'));
      } else {
        this.set('status', 'COMPLETED');
        this.set('isSubmitDisabled', false);
        this.set('isBackButtonDisabled', false);
      }
    } else if (this.get('tasks').someProperty('status', 'FAILED')) {
      this.set('status', 'FAILED');
      this.set('isBackButtonDisabled', false);
      if (this.get('isCommandLevelRetry')) {
        this.get('tasks').findProperty('status', 'FAILED').set('showRetry', true);
      } else {
        this.set('showRetry', true);
      }
      if (App.supports.autoRollbackHA) {
        this.get('tasks').findProperty('status', 'FAILED').set('showRollback', true);
      }
    }
    this.get('tasks').filterProperty('status', 'COMPLETED').setEach('showRetry', false);
    this.get('tasks').filterProperty('status', 'COMPLETED').setEach('showRollback', false);

    if (data && data.deferred) {
      data.deferred.resolve();
    }
  },

  /**
   * Run command of appropriate task
   */
  runTask: function (taskId) {
    this[this.get('tasks').findProperty('id', taskId).get('command')]();
  },

  onTaskError: function () {
    this.setTaskStatus(this.get('currentTaskId'), 'FAILED');
  },

  onSingleRequestError: function (jqXHR, ajaxOptions, error, opt) {
    App.ajax.defaultErrorHandler(jqXHR, opt.url, opt.method, jqXHR.status);
    this.set('status', 'FAILED');
    this.set('isLoaded', true);
    this.set('showRetry', true);
  },

  onTaskCompleted: function () {
    this.setTaskStatus(this.get('currentTaskId'), 'COMPLETED');
  },

  /**
   * check whether component installed on specified hosts
   * @param {string} componentName
   * @param {string[]} hostNames
   * @return {$.ajax}
   */
  checkInstalledComponents: function (componentName, hostNames) {
    return App.ajax.send({
      name: 'host_component.installed.on_hosts',
      sender: this,
      data: {
        componentName: componentName,
        hostNames: hostNames.join(',')
      }
    });
  },


  /**
   * Create component on single or multiple hosts.
   *
   * @method createComponent
   * @param {string} componentName - name of the component
   * @param {(string|string[])} hostName - host/hosts where components should be installed
   * @param {string} serviceName - name of the services
   */
  createComponent: function (componentName, hostName, serviceName) {
    var hostNames = (Array.isArray(hostName)) ? hostName : [hostName];
    var self = this;

    this.set('showRetry', false);

    this.checkInstalledComponents(componentName, hostNames).then(function (data) {
      var hostsWithComponents = data.items.mapProperty('HostRoles.host_name');
      var result = hostNames.map(function(item) {
        return {
          componentName: componentName,
          hostName: item,
          hasComponent: hostsWithComponents.contains(item)
        };
      });
      var hostsWithoutComponents = result.filterProperty('hasComponent', false).mapProperty('hostName');
      var taskNum = 1;
      var requestData = {
        "RequestInfo": {
          "query": hostsWithoutComponents.map(function(item) {
            return 'Hosts/host_name=' + item;
          }).join('|')
        },
        "Body": {
          "host_components": [
            {
              "HostRoles": {
                "component_name": componentName
              }
            }
          ]
        }
      };
      if (!!hostsWithoutComponents.length) {
        App.ajax.send({
          name: 'wizard.step8.register_host_to_component',
          sender: self,
          data: {
            data: JSON.stringify(requestData),
            hostName: result.mapProperty('hostName'),
            componentName: componentName,
            serviceName: serviceName,
            taskNum: taskNum,
            cluster: App.get('clusterName')
          },
          success: 'onCreateComponent',
          error: 'onCreateComponent'
        });
      } else {
        self.onCreateComponent(null, null, {
          hostName: result.mapProperty('hostName'),
          componentName: componentName,
          serviceName: serviceName,
          taskNum: taskNum
        }, self);
      }
    });
  },

  onCreateComponent: function () {
    var hostName = arguments[2].hostName;
    var componentName = arguments[2].componentName;
    var taskNum = arguments[2].taskNum;
    var serviceName = arguments[2].serviceName;
    this.updateComponent(componentName, hostName, serviceName, "Install", taskNum);
  },

  onCreateComponentError: function (error) {
    if (error.responseText.indexOf('org.apache.ambari.server.controller.spi.ResourceAlreadyExistsException') !== -1) {
      this.onCreateComponent();
    } else {
      this.onTaskError();
    }
  },

  /**
   * Update component status on selected hosts.
   *
   * @param {string} componentName
   * @param {(string|string[])} hostName
   * @param {string} serviceName
   * @param {string} context
   * @param {number} taskNum
   * @returns {$.ajax}
   */
  updateComponent: function (componentName, hostName, serviceName, context, taskNum) {
    if (!(hostName instanceof Array)) {
      hostName = [hostName];
    }
    var state = context.toLowerCase() == "start" ? "STARTED" : "INSTALLED";
    return App.ajax.send({
      name: 'common.host_components.update',
      sender: this,
      data: {
        HostRoles: {
          state: state
        },
        query: 'HostRoles/component_name=' + componentName + '&HostRoles/host_name.in(' + hostName.join(',') + ')&HostRoles/maintenance_state=OFF',
        context: context + " " + App.format.role(componentName),
        hostName: hostName,
        taskNum: taskNum || 1,
        componentName: componentName,
        serviceName: serviceName
      },
      success: 'startPolling',
      error: 'onTaskError'
    });
  },

  startPolling: function (data) {
    if (data) {
      this.get('currentRequestIds').push(data.Requests.id);
      var tasksCount = arguments[2].taskNum || 1;
      if (tasksCount === this.get('currentRequestIds').length) {
        this.setRequestIds(this.get('currentTaskId'), this.get('currentRequestIds'));
        this.doPolling();
      }
    } else {
      this.onTaskCompleted();
    }
  },

  doPolling: function () {
    this.setTaskStatus(this.get('currentTaskId'), 'IN_PROGRESS');
    var requestIds = this.get('currentRequestIds');
    this.set('logs', []);
    for (var i = 0; i < requestIds.length; i++) {
      App.ajax.send({
        name: 'background_operations.get_by_request',
        sender: this,
        data: {
          requestId: requestIds[i]
        },
        success: 'parseLogs',
        error: 'onTaskError'
      });
    }
  },

  parseLogs: function (logs) {
    this.get('logs').pushObject(logs.tasks);
    if (this.get('currentRequestIds').length === this.get('logs').length) {
      var tasks = [];
      this.get('logs').forEach(function (logs) {
        tasks.pushObjects(logs);
      }, this);
      var self = this;
      var currentTaskId = this.get('currentTaskId');
      if (!tasks.someProperty('Tasks.status', 'PENDING') && !tasks.someProperty('Tasks.status', 'QUEUED') && !tasks.someProperty('Tasks.status', 'IN_PROGRESS')) {
        this.set('currentRequestIds', []);
        if (tasks.someProperty('Tasks.status', 'FAILED') || tasks.someProperty('Tasks.status', 'TIMEDOUT') || tasks.someProperty('Tasks.status', 'ABORTED')) {
          this.setTaskStatus(currentTaskId, 'FAILED');
        } else {
          this.setTaskStatus(currentTaskId, 'COMPLETED');
        }
      } else {
        var actionsPerHost = tasks.length;
        var completedActions = tasks.filterProperty('Tasks.status', 'COMPLETED').length
          + tasks.filterProperty('Tasks.status', 'FAILED').length
          + tasks.filterProperty('Tasks.status', 'ABORTED').length
          + tasks.filterProperty('Tasks.status', 'TIMEDOUT').length;
        var queuedActions = tasks.filterProperty('Tasks.status', 'QUEUED').length;
        var inProgressActions = tasks.filterProperty('Tasks.status', 'IN_PROGRESS').length;
        var progress = Math.floor(((queuedActions * 0.09) + (inProgressActions * 0.35) + completedActions ) / actionsPerHost * 100);
        this.get('tasks').findProperty('id', currentTaskId).set('progress', progress);
        window.setTimeout(function () {
          self.doPolling();
        }, self.POLL_INTERVAL);
      }
    }
  },

  showHostProgressPopup: function (event) {
    var popupTitle = event.contexts[0].title;
    var requestIds = event.contexts[0].requestIds;
    var stageId = event.contexts[0].stageId;
    var hostProgressPopupController = App.router.get('highAvailabilityProgressPopupController');
    hostProgressPopupController.initPopup(popupTitle, requestIds, this, true, stageId);
  },

  done: function () {
    if (!this.get('isSubmitDisabled')) {
      this.removeObserver('tasks.@each.status', this, 'onTaskStatusChange');
      App.router.send('next');
    }
  },

  back: function () {
    if (!this.get('isBackButtonDisabled')) {
      this.removeObserver('tasks.@each.status', this, 'onTaskStatusChange');
      App.router.send('back');
    }
  }

});



});

require.register("mixins/wizard/wizardProgressPageView", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Mixin for wizard view for showing command progress on wizard pages
 * This should
 * @type {Ember.Mixin}
 */
App.wizardProgressPageViewMixin = Em.Mixin.create({

  /**
   * Following computed property needs to be overridden by the view implementing this mixin
   */

  currentStep: '',

  /**
   * Following computed property needs to be overridden by the view implementing this mixin
   */

  headerTitle: function () {

  }.property(),

  submitButtonText: Em.I18n.t('common.next'),

  noticeCompleted: Em.I18n.t('wizard.progressPage.notice.completed'),

  noticeFailed: function() {
    return (this.get('controller.isSingleRequestPage') ? Em.I18n.t('wizard.singleRequest.progressPage.notice.failed') : Em.I18n.t('wizard.progressPage.notice.failed'))
  }.property('controller.isSingleRequestPage'),

  /**
   * @noticeInProgress: Following computed property needs to be overridden to show the label text while the commands
   * on the page are progressing
   */
  noticeInProgress: function () {

  }.property(),

  /**
   * @showBackButton: Override this property to show back button on the wizard progress page
   */
  showBackButton: false,

  /**
   * Following computed property needs to be overridden by the view implementing this mixin
   */
  notice: '',

  noticeClass: 'alert alert-info',

  onStatusChange: function () {
    var status = this.get('controller.status');
    if (status === 'COMPLETED') {
      this.set('notice', this.get('noticeCompleted'));
      this.set('noticeClass', 'alert alert-success');
    } else if (status === 'FAILED') {
      this.set('notice', this.get('noticeFailed'));
      this.set('noticeClass', 'alert alert-error');
    } else {
      this.set('notice', this.get('noticeInProgress'));
      this.set('noticeClass', 'alert alert-info');
    }
  }.observes('controller.status'),

  taskView: Em.View.extend({
    icon: '',
    iconColor: '',
    linkClass: '',

    didInsertElement: function () {
      this.onStatus();
    },

    barWidth: function () {
      return 'width: ' + this.get('content.progress') + '%;';
    }.property('content.progress'),

    onStatus: function () {
      var linkClass = !!this.get('content.requestIds.length') ? 'active-link' : 'active-text';
      this.set('linkClass', linkClass);
      if (this.get('content.status') === 'IN_PROGRESS') {
        this.set('icon', 'icon-cog');
        this.set('iconColor', 'text-info');
      } else if (this.get('content.status') === 'FAILED') {
        this.set('icon', 'icon-exclamation-sign');
        this.set('iconColor', 'text-error');
      } else if (this.get('content.status') === 'COMPLETED') {
        this.set('icon', 'icon-ok');
        this.set('iconColor', 'text-success');
      } else {
        this.set('icon', 'icon-cog');
        this.set('iconColor', '');
        this.set('linkClass', 'not-active-link');
      }
    }.observes('content.status', 'content.hosts.length','content.requestIds'),

    showProgressBar: function () {
      return this.get('content.status') === "IN_PROGRESS";
    }.property('content.status')
  })

});



});

require.register("mixins/wizard/wizard_menu_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.WizardMenuMixin = Em.Mixin.create({

  isStepDisabled: function (index) {
    return this.get('controller.isStepDisabled').findProperty('step', index).get('value');
  },

  isStep0Disabled: function () {
    return this.isStepDisabled(0);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep1Disabled: function () {
    return this.isStepDisabled(1);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep2Disabled: function () {
    return this.isStepDisabled(2);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep3Disabled: function () {
    return this.isStepDisabled(3);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep4Disabled: function () {
    return this.isStepDisabled(4);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep5Disabled: function () {
    return this.isStepDisabled(5);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep6Disabled: function () {
    return this.isStepDisabled(6);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep7Disabled: function () {
    return this.isStepDisabled(7);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep8Disabled: function () {
    return this.isStepDisabled(8);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep9Disabled: function () {
    return this.isStepDisabled(9);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep10Disabled: function () {
    return this.isStepDisabled(10);
  }.property('controller.isStepDisabled.@each.value').cacheable()

});

});

require.register("models", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


// load all models here

require('models/form'); // should be the 1st
require('models/authentication');
require('models/cluster');
require('models/cluster_states');
require('models/hosts');
require('models/stack');
require('models/stack_version/version');
require('models/stack_version/repository_version');
require('models/stack_version/os');
require('models/stack_version/repository');
require('models/operating_system');
require('models/repository');
require('models/stack_service');
require('models/stack_service_component');
require('models/quick_links');
require('models/service');
require('models/service_config');
require('models/service_audit');
require('models/service/hdfs');
require('models/service/yarn');
require('models/service/mapreduce2');
require('models/service/hbase');
require('models/service/flume');
require('models/service/storm');
require('models/alert_definition');
require('models/alert_instance');
require('models/alert_notification');
require('models/alert_config');
require('models/alert_group');
require('models/user');
require('models/host');
require('models/rack');
require('models/background_operation');
require('models/client_component');
require('models/host_component');
require('models/target_cluster');
require('models/slave_component');
require('models/config_group');
require('models/service_config_version');
require('models/host_stack_version');
require('models/root_service');
require('models/upgrade_entity');
});

require.register("models/alert_config", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var validator = require('utils/validator');

App.AlertConfigProperty = Ember.Object.extend({

  /**
   * label to be shown for config property
   * @type {String}
   */
  label: '',

  /**
   * PORT|METRIC|AGGREGATE
   * @type {String}
   */
  type: '',

  /**
   * config property value
   * @type {*}
   */
  value: null,

  /**
   * property value cache to realise undo function
   * @type {*}
   */
  previousValue: null,

  /**
   * define either input is disabled or enabled
   * @type {Boolean}
   */
  isDisabled: false,

  /**
   * options that Select list will have
   * @type {Array}
   */
  options: [],

  /**
   * input displayType
   * one of 'textFields', 'textArea', 'select' or 'threshold'
   * @type {String}
   */
  displayType: '',

  /**
   * unit to be shown with value
   * @type {String}
   */
  unit: null,

  /**
   * space separated list of css class names to use
   * @type {String}
   */
  classNames: '',

  /**
   * define whether row with property should be shifted right
   * @type {Boolean}
   */
  isShifted: false,

  /**
   * name or names of properties related to config
   * may be either string for one property or array of strings for multiple properties
   * if this property is array, then <code>apiFormattedValue</code> should also be an array
   * example: <code>apiProperty[0]</code> relates to <code>apiFormattedValue[0]</code>
   * @type {String|Array}
   */
  apiProperty: '',

  /**
   * for some metrics properties may be set true or false
   * depending on what property is related to (JMX or Ganglia)
   */
  isJMXMetric: null,

  /**
   * define place to show label
   * if true - label is shown before input
   * if false - label is shown after input
   * @type {Boolean}
   */
  isPreLabeled: function () {
    var afterLabeledTypes = ['radioButton'];
    return !afterLabeledTypes.contains(this.get('displayType'));
  }.property('displayType'),

  /**
   * value converted to appropriate format for sending to server
   * should be computed property
   * should be defined in child class
   * @type {*}
   */
  apiFormattedValue: function () {
    return this.get('value');
  }.property('value'),

  /**
   * define if property was changed by user
   * @type {Boolean}
   */
  wasChanged: function () {
    return this.get('previousValue') !== null && this.get('value') !== this.get('previousValue');
  }.property('value', 'previousValue'),

  /**
   * view class according to <code>displayType</code>
   * @type {Em.View}
   */
  viewClass: function () {
    var displayType = this.get('displayType');
    switch (displayType) {
      case 'textField':
        return App.AlertConfigTextFieldView;
      case 'textArea':
        return App.AlertConfigTextAreaView;
      case 'select':
        return App.AlertConfigSelectView;
      case 'threshold':
        return App.AlertConfigThresholdView;
      case 'radioButton':
        return App.AlertConfigRadioButtonView;
      default:
        console.error('Unable to find viewClass for displayType ', displayType);
    }
  }.property('displayType'),

  /**
   * Define whether property is valid
   * Computed property
   * Should be defined in child class
   * @type {Boolean}
   */
  isValid: function () {
    return true;
  }.property()

});

App.AlertConfigProperties = {

  AlertName: App.AlertConfigProperty.extend({
    name: 'alert_name',
    label: 'Alert Name',
    displayType: 'textField',
    classNames: 'alert-text-input',
    apiProperty: 'name'
  }),

  AlertNameSelected: App.AlertConfigProperty.extend({
    name: 'alert_name',
    label: 'Alert Name',
    displayType: 'select',
    apiProperty: 'name'
  }),

  ServiceAlertType: App.AlertConfigProperty.extend({
    name: 'alert_type_service',
    label: 'Service Alert Definition',
    displayType: 'radioButton',
    group: 'alert_type'
  }),

  HostAlertType: App.AlertConfigProperty.extend({
    name: 'alert_type_host',
    label: 'Host Alert Definition',
    displayType: 'radioButton',
    group: 'alert_type'
  }),

  Service: App.AlertConfigProperty.extend({
    name: 'service',
    label: 'Service',
    displayType: 'select',
    apiProperty: 'service_name',
    apiFormattedValue: function () {
      return App.StackService.find().findProperty('displayName', this.get('value')).get('serviceName');
    }.property('value')
  }),

  Component: App.AlertConfigProperty.extend({
    name: 'component',
    label: 'Component',
    displayType: 'select',
    apiProperty: 'component_name',
    apiFormattedValue: function () {
      return App.StackServiceComponent.find().findProperty('displayName', this.get('value')).get('componentName');
    }.property('value')
  }),

  Scope: App.AlertConfigProperty.extend({
    name: 'scope',
    label: 'Scope',
    displayType: 'select',
    apiProperty: 'scope',
    apiFormattedValue: function () {
      return this.get('value').toUpperCase();
    }.property('value')
  }),

  Description: App.AlertConfigProperty.extend({
    name: 'description',
    label: 'Description',
    displayType: 'textArea',
    classNames: 'alert-config-text-area',
    // todo: check value after API will be provided
    apiProperty: 'description'
  }),

  Interval: App.AlertConfigProperty.extend({
    name: 'interval',
    label: 'Interval',
    displayType: 'textField',
    unit: 'Minute',
    classNames: 'alert-interval-input',
    apiProperty: 'interval',
    isValid: function () {
      var value = this.get('value');
      if (!value) return false;
      return String(value) === String(parseInt(value, 10)) && value >= 1;
    }.property('value')
  }),

  /**
   * Implements threshold
   * Main difference from other alertConfigProperties:
   * it has two editable parts - <code>value</code> and <code>text</code>
   * User may configure it to edit only one of them (use flags <code>showInputForValue</code> and <code>showInputForText</code>)
   * This flags also determines update value and text in the API-request or not (see <code>App.AlertConfigProperties.Thresholds</code> for more examples)
   *
   * @type {App.AlertConfigProperty.Threshold}
   */
  Threshold: App.AlertConfigProperty.extend({

    name: 'threshold',

    /**
     * Property text cache to realise undo function
     * @type {*}
     */
    previousText: null,

    label: '',

    /**
     * OK|WARNING|CRITICAL
     * @type {string}
     */
    badge: '',

    /**
     * threshold-value
     * @type {string}
     */
    value: '',

    /**
     * Type of value. This will be a fixed set of types (like %).
     */
    valueMetric: null,

    /**
     * Value actually displayed to the user. This value is transformed
     * based on the limited types of 'valueMetric's. Mappings from
     * 'value' to 'displayValue' is handled by observers.
     */
    displayValue: '',

    /**
     * threshold-text
     * @type {string}
     */
    text: '',

    displayType: 'threshold',

    classNames: 'alert-thresholds-input',

    apiProperty: [],

    init: function () {
      this.set('displayValue', this.getNewValue());
      this._super();
    },

    /**
     * @type {string[]}
     */
    apiFormattedValue: function () {
      var ret = [];
      if (this.get('showInputForValue')) {
        ret.push(this.get('value'));
      }
      if (this.get('showInputForText')) {
        ret.push(this.get('text'));
      }
      return ret;
    }.property('value', 'text', 'showInputForValue', 'showInputForText'),

    /**
     * Determines if <code>value</code> should be visible and editable (if not - won't update API-value)
     * @type {bool}
     */
    showInputForValue: true,

    /**
     * Determines if <code>text</code> should be visible and editable (if not - won't update API-text)
     * @type {bool}
     */
    showInputForText: true,

    /**
     * Custom css-class for different badges
     * type {string}
     */
    badgeCssClass: function () {
      return 'alert-state-' + this.get('badge');
    }.property('badge'),

    /**
     * Determines if <code>value</code> or <code>text</code> were changed
     * @type {bool}
     */
    wasChanged: function () {
      return (this.get('previousValue') !== null && this.get('value') !== this.get('previousValue')) ||
      (this.get('previousText') !== null && this.get('text') !== this.get('previousText'));
    }.property('value', 'text', 'previousValue', 'previousText'),

    /**
     * May be redefined in child-models, mixins etc
     * @method getValue
     * @returns {string}
     */
    getNewValue: function () {
      return this.get('value');
    },

    valueWasChanged: function () {
      var displayValue = this.get('displayValue');
      var newDisplayValue = this.getNewValue();
      if (newDisplayValue !== displayValue && !(isNaN(newDisplayValue) ||isNaN(displayValue))) {
        this.set('displayValue', newDisplayValue);
      }
    }.observes('value'),

    /**
     * May be redefined in child-models, mixins etc
     * @method getDisplayValue
     * @returns {string}
     */
    getNewDisplayValue: function () {
      return this.get('displayValue');
    },

    displayValueWasChanged: function () {
      var value = this.get('value');
      var newValue = this.getNewDisplayValue();
      if (newValue !== value && !(isNaN(newValue) ||isNaN(value))) {
        this.set('value', newValue);
      }
    }.observes('displayValue'),

    /**
     * Check if <code>displayValue</code> is valid float number
     * If this value isn't shown (see <code>showInputForValue</code>), result is always true
     * @return {boolean}
     */
    isValid: function () {
      if (!this.get('showInputForValue')) return true;
      var value = this.get('displayValue');
      if (Em.isNone(value)) return false;
      value = ('' + value).trim();
      return validator.isValidFloat(value);
    }.property('displayValue', 'showInputForValue')

  }),

  URI: App.AlertConfigProperty.extend({
    name: 'uri',
    label: 'URI',
    displayType: 'textField',
    classNames: 'alert-text-input',
    apiProperty: 'source.uri'
  }),

  URIExtended: App.AlertConfigProperty.extend({
    name: 'uri',
    label: 'URI',
    displayType: 'textArea',
    classNames: 'alert-config-text-area',
    apiProperty: 'source.uri',
    apiFormattedValue: function () {
      var result = {};
      try {
        result = JSON.parse(this.get('value'));
      } catch (e) {
        console.error('Wrong format of URI');
      }
      return result;
    }.property('value')
  }),

  DefaultPort: App.AlertConfigProperty.extend({
    name: 'default_port',
    label: 'Default Port',
    displayType: 'textField',
    classNames: 'alert-port-input',
    apiProperty: 'source.default_port'
  }),

  Path: App.AlertConfigProperty.extend({
    name: 'path',
    label: 'Path',
    displayType: 'textField',
    classNames: 'alert-text-input',
    apiProperty: 'source.path'
  }),

  Metrics: App.AlertConfigProperty.extend({
    name: 'metrics',
    label: 'JMX/Ganglia Metrics',
    displayType: 'textArea',
    classNames: 'alert-config-text-area',
    apiProperty: function () {
      return this.get('isJMXMetric') ? 'source.jmx.property_list' : 'source.ganglia.property_list'
    }.property('isJMXMetric'),
    apiFormattedValue: function () {
      return this.get('value').split(',\n');
    }.property('value')
  }),

  FormatString: App.AlertConfigProperty.extend({
    name: 'metrics_string',
    label: 'Format String',
    displayType: 'textArea',
    classNames: 'alert-config-text-area',
    apiProperty: function () {
      return this.get('isJMXMetric') ? 'source.jmx.value' : 'source.ganglia.value'
    }.property('isJMXMetric')
  })

};

App.AlertConfigProperties.Thresholds = {

  OkThreshold: App.AlertConfigProperties.Threshold.extend({

    badge: 'OK',

    name: 'ok_threshold',

    apiProperty: function () {
      var ret = [];
      if (this.get('showInputForValue')) {
        ret.push('source.reporting.ok.value');
      }
      if (this.get('showInputForText')) {
        ret.push('source.reporting.ok.text');
      }
      return ret;
    }.property('showInputForValue', 'showInputForText')

  }),

  WarningThreshold: App.AlertConfigProperties.Threshold.extend({

    badge: 'WARNING',

    name: 'warning_threshold',

    apiProperty: function () {
      var ret = [];
      if (this.get('showInputForValue')) {
        ret.push('source.reporting.warning.value');
      }
      if (this.get('showInputForText')) {
        ret.push('source.reporting.warning.text');
      }
      return ret;
    }.property('showInputForValue', 'showInputForText')

  }),

  CriticalThreshold: App.AlertConfigProperties.Threshold.extend({

    badge: 'CRITICAL',

    name: 'critical_threshold',

    apiProperty: function () {
      var ret = [];
      if (this.get('showInputForValue')) {
        ret.push('source.reporting.critical.value');
      }
      if (this.get('showInputForText')) {
        ret.push('source.reporting.critical.text');
      }
      return ret;
    }.property('showInputForValue', 'showInputForText')

  }),

  /**
   * Mixin for <code>App.AlertConfigProperties.Threshold</code>
   * Used to validate values in percentage range (0..1]
   * @type {Em.Mixin}
   */
  PercentageMixin: Em.Mixin.create({

    isValid: function () {
      var value = this.get('displayValue');
      if (!value) return false;
      value = ('' + value).trim();
      value = parseFloat(value);
      return this.get('showInputForValue') ? !isNaN(value) && value > 0 && value <= 100 : true;
    }.property('displayValue', 'showInputForValue'),

    /**
     * Return <code>value * 100</code>
     * @returns {string}
     */
    getNewValue: function () {
      var value = this.get('value');
      return (value && !isNaN(value)) ? (Number(value) * 100) + '' : value;
    },

    /**
     * Return <code>displayValue / 100</code>
     * @returns {string}
     */
    getNewDisplayValue: function () {
      var displayValue = this.get('displayValue');
      return (displayValue && !isNaN(displayValue)) ? (Number(displayValue) / 100) + '' : displayValue;
    }

  }),

  /**
   * Mixin for <code>App.AlertConfigProperties.Threshold</code>
   * Used to validate values that should be greater than 0
   * @type {Em.Mixin}
   */
  PositiveMixin: Em.Mixin.create({

    isValid: function () {
      if (!this.get('showInputForValue')) return true;
      var value = this.get('displayValue');
      if (!value) return false;
      value = ('' + value).trim();
      value = parseFloat(value);
      return !isNaN(value) && value > 0;
    }.property('displayValue', 'showInputForValue')

  })

};

});

require.register("models/alert_definition", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var dateUtils = require('utils/date');

App.AlertDefinition = DS.Model.extend({

  name: DS.attr('string'),
  label: DS.attr('string'),
  description: DS.attr('string'),
  service: DS.belongsTo('App.Service'),
  serviceName: DS.attr('string'),
  componentName: DS.attr('string'),
  enabled: DS.attr('boolean'),
  scope: DS.attr('string'),
  interval: DS.attr('number'),
  type: DS.attr('string'),
  groups: DS.hasMany('App.AlertGroup'),
  reporting: DS.hasMany('App.AlertReportDefinition'),
  lastTriggered: DS.attr('number'),

  //relates only to SCRIPT-type alert definition
  location: DS.attr('string'),
  //relates only to AGGREGATE-type alert definition
  alertName: DS.attr('string'),
  //relates only to WEB and METRIC types of alert definition
  uri: DS.belongsTo('App.AlertMetricsUriDefinition'),
  //relates only METRIC-type alert definition
  jmx: DS.belongsTo('App.AlertMetricsSourceDefinition'),
  ganglia: DS.belongsTo('App.AlertMetricsSourceDefinition'),
  //relates only PORT-type alert definition
  defaultPort: DS.attr('number'),
  portUri: DS.attr('string'),

  /**
   * Raw data from AlertDefinition/source
   * used to format request content for updating alert definition
   * @type {Object}
   */
  rawSourceData: {},

  /**
   * Counts of alert grouped by their status
   * Format:
   * <code>
   *   {
   *    "CRITICAL": {
   *      count: 1,
   *      maintenanceCount: 0
   *    },
   *    "OK": {
   *      count: 0,
   *      maintenanceCount: 1
   *    },
   *    "UNKNOWN": {
   *      count: 0,
   *      maintenanceCount: 0
   *    },
   *    "WARNING": {
   *      count: 1,
   *      maintenanceCount: 1
   *    }
   *   }
   * </code>
   * @type {object}
   */
  summary: {},

  /**
   * Formatted timestamp for latest alert triggering for current alertDefinition
   * @type {string}
   */
  lastTriggeredFormatted: function () {
    return dateUtils.dateFormat(this.get('lastTriggered'));
  }.property('lastTriggered'),

  /**
   * Formatted timestamp with <code>$.timeago</code>
   * @type {string}
   */
  lastTriggeredAgoFormatted: function () {
    var lastTriggered = this.get('lastTriggered');
    return lastTriggered ? $.timeago(new Date(lastTriggered)) : '';
  }.property('lastTriggered'),

  lastTriggeredVerboseDisplay: function () {
    var lastTriggered = this.get('lastTriggered');
    return Em.I18n.t('models.alert_definition.triggered.verbose').format(dateUtils.dateFormat(lastTriggered));
  }.property('lastTriggered'),

  /**
   * Formatted timestamp in format: for 4 days
   * @type {string}
   */
  lastTriggeredForFormatted: function () {
    var lastTriggered = this.get('lastTriggered');
    var previousSuffixAgo = $.timeago.settings.strings.suffixAgo;
    var previousPrefixAgo = $.timeago.settings.strings.prefixAgo;
    $.timeago.settings.strings.suffixAgo = null;
    $.timeago.settings.strings.prefixAgo = 'for';
    var triggeredFor = lastTriggered ? $.timeago(new Date(lastTriggered)) : '';
    $.timeago.settings.strings.suffixAgo = previousSuffixAgo;
    $.timeago.settings.strings.prefixAgo = previousPrefixAgo;
    return triggeredFor;
  }.property('lastTriggered'),

  /**
   * Formatted displayName for <code>componentName</code>
   * @type {String}
   */
  componentNameFormatted: function () {
    return App.format.role(this.get('componentName'));
  }.property('componentName'),

  /**
   * Status generates from child-alerts
   * Format: OK(1)  WARN(2)  CRIT(1)  UNKN(1)
   * If single host: show: OK/WARNING/CRITICAL/UNKNOWN
   * If some there are no alerts with some state, this state isn't shown
   * If no OK/WARN/CRIT/UNKN state, then show PENDING
   * Order is equal to example
   * @type {string}
   */
  status: function () {
    var order = this.get('order'),
        summary = this.get('summary'),
        hostCnt = 0,
        self = this;
    order.forEach(function (state) {
      hostCnt += summary[state] ? summary[state].count + summary[state].maintenanceCount : 0;
    });
    if (hostCnt > 1) {
      // multiple hosts
      return order.map(function (state) {
        var shortState = self.get('shortState')[state];
        var result = '';
        result += summary[state].count ? '<span class="alert-state-single-host label alert-state-' + state + '">' + shortState + ' (' + summary[state].count + ')</span>' : '';
        // add status with maintenance mode icon
        result += summary[state].maintenanceCount ?
        '<span class="alert-state-single-host label alert-state-PENDING"><span class="icon-medkit"></span> ' + shortState + ' (' + summary[state].maintenanceCount + ')</span>' : '';
        return result;
      }).without('').join(' ');
    } else if (hostCnt == 1) {
      // single host, single status
      return order.map(function (state) {
        var shortState = self.get('shortState')[state];
        var result = '';
        result += summary[state].count ? '<span class="alert-state-single-host label alert-state-' + state + '">' + shortState + '</span>' : '';
        // add status with maintenance mode icon
        result += summary[state].maintenanceCount ?
        '<span class="alert-state-single-host label alert-state-PENDING"><span class="icon-medkit"></span> ' + shortState + '</span>' : '';
        return result;
      }).without('').join(' ');
    } else if (hostCnt == 0) {
      // none
      return '<span class="alert-state-single-host label alert-state-PENDING">NONE</span>';
    }
    return '';
  }.property('summary'),

  latestText: function () {
    var order = this.get('order'), summary = this.get('summary'), text = '';
    order.forEach(function (state) {
      var cnt = summary[state] ? summary[state].count + summary[state].maintenanceCount : 0;
      if (cnt > 0) {
        text = summary[state].latestText;
      }
    });
    return text;
  }.property('summary'),

  isHostAlertDefinition: function () {
    var serviceID = (this.get('service')._id === "AMBARI"),
        component = (this.get('componentName') === "AMBARI_AGENT");
    return serviceID && component;
  }.property('service', 'componentName'),

  typeIconClass: function () {
    var typeIcons = this.get('typeIcons'),
        type = this.get('type');
    return typeIcons[type];
  }.property('type'),

  /**
   * if this definition is in state: CRITICAL / WARNING, if true, will show up in alerts fast access popup
   * instances with maintenance mode ON are ignored
   * @type {boolean}
   */
  isCriticalOrWarning: function () {
    return !!(this.get('summary.CRITICAL.count') || this.get('summary.WARNING.count'));
  }.property('summary'),

  /**
   * if this definition is in state: CRIT
   * @type {boolean}
   */
  isCritical: function () {
    var summary = this.get('summary');
    var state = 'CRITICAL';
    return !!summary[state] && !!(summary[state].count || summary[state].maintenanceCount);
  }.property('summary'),

  /**
   * if this definition is in state: WARNING
   * @type {boolean}
   */
  isWarning: function () {
    var summary = this.get('summary');
    var state = 'WARNING';
    return !!summary[state] && !!(summary[state].count || summary[state].maintenanceCount);
  }.property('summary'),

  /**
   * if this definition is in state: OK
   * @type {boolean}
   */
  isOK: function () {
    var summary = this.get('summary');
    var state = 'OK';
    return !!summary[state] && !!(summary[state].count || summary[state].maintenanceCount);
  }.property('summary'),

  /**
   * if this definition is in state: OK
   * @type {boolean}
   */
  isUnknown: function () {
    var summary = this.get('summary');
    var state = 'UNKNOWN';
    return !!summary[state] && !!(summary[state].count || summary[state].maintenanceCount);
  }.property('summary'),

  /**
   * For alerts we will have processes which are not typical
   * cluster services - like Ambari-Server. This method unifies
   * cluster services and other services into a common display-name.
   * @see App.AlertInstance#serviceDisplayName()
   */
  serviceDisplayName: function () {
    var serviceName = this.get('service.displayName');
    if (!serviceName) {
      serviceName = this.get('serviceName');
      if (serviceName) {
        serviceName = serviceName.toCapital();
      }
    }
    return serviceName;
  }.property('serviceName', 'service.displayName'),

  /**
   * List of css-classes for alert types
   * @type {object}
   */
  typeIcons: {
    'METRIC': 'icon-bolt',
    'SCRIPT': 'icon-file-text',
    'WEB': 'icon-globe',
    'PORT': 'icon-signin',
    'AGGREGATE': 'icon-plus'
  },

  /**
   * Sort on load definitions by this severity order
   */
  severityOrder: ['CRITICAL', 'WARNING', 'OK', 'UNKNOWN', 'PENDING'],
  order: ['OK', 'WARNING', 'CRITICAL', 'UNKNOWN'],

  shortState: {
    'CRITICAL': 'CRIT',
    'WARNING': 'WARN',
    'OK': 'OK',
    'UNKNOWN': 'UNKWN',
    'PENDING': 'NONE'
  }
});

App.AlertDefinition.reopenClass({

  /**
   * Return function to sort list of AlertDefinitions by their status
   * It sorts according to <code>severityOrder</code>
   * @param {boolean} order true - DESC, false - ASC
   * @returns {Function}
   * @method getSortDefinitionsByStatus
   */
  getSortDefinitionsByStatus: function (order) {
    return function (a, b) {
      var a_summary = a.get('summary'),
        b_summary = b.get('summary'),
        st_order = a.get('severityOrder'),
        ret = 0;
      for (var i = 0; i < st_order.length; i++) {
        var a_v = Em.isNone(a_summary[st_order[i]]) ? 0 : a_summary[st_order[i]].count + a_summary[st_order[i]].maintenanceCount,
          b_v = Em.isNone(b_summary[st_order[i]]) ? 0 : b_summary[st_order[i]].count + b_summary[st_order[i]].maintenanceCount;
        ret = b_v - a_v;
        if (ret !== 0) {
          break;
        }
      }
      return order ? ret : -ret;
    };
  }

});

App.AlertReportDefinition = DS.Model.extend({
  type: DS.attr('string'),
  text: DS.attr('string'),
  value: DS.attr('number')
});

App.AlertMetricsSourceDefinition = DS.Model.extend({
  propertyList: [],
  value: DS.attr('string')
});

App.AlertMetricsUriDefinition = DS.Model.extend({
  http: DS.attr('string'),
  https: DS.attr('string'),
  httpsProperty: DS.attr('string'),
  httpsPropertyValue: DS.attr('string')
});

App.AlertDefinition.FIXTURES = [];
App.AlertReportDefinition.FIXTURES = [];
App.AlertMetricsSourceDefinition.FIXTURES = [];
App.AlertMetricsUriDefinition.FIXTURES = [];

});

require.register("models/alert_group", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Represents an alert-group on the cluster.
 * A alert group is a collection of alert definitions
 *
 * Alert group hierarchy is at 2 levels. For
 * each service there is a 'Default' alert group
 * containing all definitions , this group is read-only
 *
 * User can create new alert group containing alert definitions from
 * any service.
 */
App.AlertGroup = DS.Model.extend({

  name: DS.attr('string'),

  description: DS.attr('string'),

  /**
   * Is this group default for some service
   * @type {boolean}
   */
  default: DS.attr('boolean'),

  /**
   * @type {App.AlertDefinition[]}
   */
  definitions: DS.hasMany('App.AlertDefinition'),

  /**
   * @type {App.AlertNotification[]}
   */
  targets: DS.hasMany('App.AlertNotification'),

  /**
   * @type {string}
   */
  displayName: function () {
    var name = this.get('name');
    if (name && name.length > App.config.CONFIG_GROUP_NAME_MAX_LENGTH) {
      var middle = Math.floor(App.config.CONFIG_GROUP_NAME_MAX_LENGTH / 2);
      name = name.substring(0, middle) + "..." + name.substring(name.length - middle);
    }
    return this.get('default') ? (name + ' Default') : name;
  }.property('name', 'default'),

  /**
   * @type {string}
   */
  displayNameDefinitions: function () {
    return this.get('displayName') + ' (' + this.get('definitions.length') + ')';
  }.property('displayName', 'definitions.length'),

  isAddDefinitionsDisabled: function () {
    return this.get('default');
  }.property('default')
});
App.AlertGroup.FIXTURES = [];



});

require.register("models/alert_instance", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var dateUtils = require('utils/date');

App.AlertInstance = DS.Model.extend({
  id: DS.attr('number'),
  label: DS.attr('string'),
  definitionName: DS.attr('string'),
  definitionId: DS.attr('number'),
  service: DS.belongsTo('App.Service'),
  serviceName: DS.attr('string'),
  componentName: DS.attr('string'),
  host: DS.belongsTo('App.Host'),
  hostName: DS.attr('string'),
  scope: DS.attr('string'),
  originalTimestamp: DS.attr('number'),
  latestTimestamp: DS.attr('number'),
  maintenanceState: DS.attr('string'),
  instance: DS.attr('string'),
  state: DS.attr('string'),
  text: DS.attr('string'),
  notification: DS.hasMany('App.AlertNotification'),

  /**
   * Status icon markup
   * @type {string}
   */
  status: function () {
    var isMaintenanceStateOn = this.get('maintenanceState') === 'ON';
    var state = this.get('state');
    var stateClass = isMaintenanceStateOn ? 'PENDING' : state;
    var shortState = this.get('shortState')[state];
    var maintenanceIcon = isMaintenanceStateOn ? '<span class="icon-medkit"></span> ' : '';
    return '<div class="label alert-state-single-host alert-state-' + stateClass + '">' + maintenanceIcon + shortState + '</div>';
  }.property('state'),

  /**
   * For alerts we will have processes which are not typical
   * cluster services - like Ambari-Server. This method unifies
   * cluster services and other services into a common display-name.
   * @see App.AlertDefinition#serviceDisplayName()
   */
  serviceDisplayName: function () {
    var serviceName = this.get('service.displayName');
    if (!serviceName) {
      serviceName = this.get('serviceName');
      if (serviceName) {
        serviceName = serviceName.toCapital();
      }
    }
    return serviceName;
  }.property('serviceName', 'service.displayName'),

  /**
   * Formatted timestamp for latest instance triggering
   * @type {string}
   */
  lastCheckedFormatted: function () {
    return dateUtils.dateFormat(this.get('latestTimestamp'));
  }.property('latestTimestamp'),

  /**
   * Formatted timestamp for latest instance triggering
   * @type {string}
   */
  lastTriggeredFormatted: function () {
    return dateUtils.dateFormat(this.get('originalTimestamp'));
  }.property('originalTimestamp'),

  /**
   * Formatted timestamp with <code>$.timeago</code>
   * @type {string}
   */
  lastTriggeredAgoFormatted: function () {
    var lastTriggered = this.get('originalTimestamp');
    return lastTriggered ? $.timeago(new Date(lastTriggered)) : '';
  }.property('originalTimestamp'),

  lastTriggeredVerboseDisplay: function () {
    var originalTimestamp = this.get('originalTimestamp');
    var latestTimestamp = this.get('latestTimestamp');
    return Em.I18n.t('models.alert_instance.tiggered.verbose').format(
        dateUtils.dateFormat(originalTimestamp),
        dateUtils.dateFormat(latestTimestamp));
  }.property('originalTimestamp', 'latestTimestamp'),

  /**
   * Formatted timestamp with <code>$.timeago</code>
   * @type {string}
   */
  lastTriggeredForFormatted: function () {
    var lastTriggered = this.get('originalTimestamp');
    var previousSuffixAgo = $.timeago.settings.strings.suffixAgo;
    var previousPrefixAgo = $.timeago.settings.strings.prefixAgo;
    $.timeago.settings.strings.suffixAgo = null;
    $.timeago.settings.strings.prefixAgo = 'for';
    var triggeredFor = lastTriggered ? $.timeago(new Date(lastTriggered)) : '';
    $.timeago.settings.strings.suffixAgo = previousSuffixAgo;
    $.timeago.settings.strings.prefixAgo = previousPrefixAgo;
    return triggeredFor;
  }.property('originalTimestamp'),

  /**
   * Formatted lastChecked and lastTriggered timestamp
   * @returns {string}
   */
  statusChangedAndLastCheckedFormatted: function () {
    var lastCheckedFormatted = this.get('lastCheckedFormatted');
    var lastTriggeredFormatted = this.get('lastTriggeredFormatted');
    return Em.I18n.t('models.alert_definition.triggered.checked').format(lastTriggeredFormatted, lastCheckedFormatted);
  }.property('lastCheckedFormatted', 'lastTriggeredFormatted'),

  /**
   * List of css-classes for alert instance status
   * @type {object}
   */
  typeIcons: {
    'DISABLED': 'icon-off'
  },

  /**
   * Define if definition serviceName is Ambari
   * Used in some logic in templates to distinguish definitions with Ambari serviceName
   * @returns {boolean}
   */
  isAmbariServiceName: function () {
    return this.get('serviceName') === 'AMBARI';
  }.property('serviceName'),

  shortState: {
    'CRITICAL': 'CRIT',
    'WARNING': 'WARN',
    'OK': 'OK',
    'UNKNOWN': 'UNKWN',
    'PENDING': 'NONE'
  }
});

App.AlertInstance.FIXTURES = [];

});

require.register("models/alert_notification", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.AlertNotification = DS.Model.extend({
  id: DS.attr('number'),
  name: DS.attr('string'),
  type: DS.attr('string'),
  description: DS.attr('string'),
  groups: DS.hasMany('App.AlertGroup'),
  global: DS.attr('boolean'),

  properties: {},
  alertStates: []
});

App.AlertNotification.FIXTURES = [];
});

require.register("models/authentication", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

App.Authentication = DS.Model.extend({
  method:DS.attr('boolean'), // use LDAP
  primaryServer:DS.attr('string'),
  secondaryServer:DS.attr('string'),
  useSsl:DS.attr('boolean'),
  bindMethod:DS.attr('boolean'), // use credentials
  bindUser:DS.attr('string'),
  password:DS.attr('string'),
  passwordRetype:DS.attr('string'),
  searchBaseDn:DS.attr('string'),
  usernameAttribute:DS.attr('string')
});

App.Authentication.FIXTURES = [
  {
    id:1,
    method:0,
    primary_server:"",
    secondary_server:"",
    use_ssl:false,
    bind_method:0,
    bind_user:"",
    password:"",
    password_retype:"",
    search_base_dn:"",
    username_attribute:""
  }
];

App.AuthenticationForm = App.Form.extend({
  testResult:false,
  fieldsOptions:[
    { name:"method", displayName:"", isRequired:false, displayType:"select",
      values:[
        {value:0, label:Em.I18n.t("admin.authentication.form.method.database")},
        {value:1, label:Em.I18n.t("admin.authentication.form.method.ldap")}
      ]
    },
    { name:"primaryServer", displayName:Em.I18n.t("admin.authentication.form.primaryServer"), /*validator:'ipaddress',*/
      isRequired:function () {
        return this.get('form.field.method.value');
      }.property('form.field.method.value')
    },
    { name:"secondaryServer", displayName:Em.I18n.t("admin.authentication.form.secondaryServer"), /*validator:'ipaddress',*/ isRequired:false},
    { name:"useSsl", displayName:Em.I18n.t("admin.authentication.form.useSsl"), displayType:"checkbox", isRequired:false },
    { name:"bindMethod", displayName:'', displayType:"select", isRequired:false,
      values:[
        {value:0, label:Em.I18n.t("admin.authentication.form.bind.anonymously")},
        {value:1, label:Em.I18n.t("admin.authentication.form.bind.useCrenedtials")}
      ]},
    { name:"bindUser", displayName:Em.I18n.t('admin.authentication.form.bindUserDN'), isRequired:function () {
      return this.get('form.field.bindMethod.value');
    }.property('form.field.bindMethod.value')},
    { name:"password", displayName:Em.I18n.t('common.password'), displayType:"password",
      isRequired:function () {
        return this.get('form.field.bindMethod.value');
      }.property('form.field.bindMethod.value') },
    { name:"passwordRetype", displayName:Em.I18n.t('form.passwordRetype'), displayType:"password",
      validator: "passwordRetype",
      isRequired:function () {
        return this.get('form.field.bindMethod.value');
      }.property('form.field.bindMethod.value')},
    { name:"searchBaseDn", displayName:Em.I18n.t('admin.authentication.form.searchBaseDN'),
      isRequired:function () {
        return this.get('form.field.method.value');
      }.property('form.field.method.value')
    },
    { name:"usernameAttribute", displayName:Em.I18n.t('admin.authentication.form.usernameAttribute'),
      isRequired:function () {
        return this.get('form.field.method.value');
      }.property('form.field.method.value')
    },

    { name:"userDN", displayName:Em.I18n.t('admin.authentication.form.userDN') },
    { name:"userPassword", displayName:Em.I18n.t('common.password'), displayType:'password'}
  ],
  fields:[],
  testConfiguration:function () {
    console.warn('Configuration test is randomized');
    this.set('testResult', parseInt(Math.random() * 2));
    return true;
  },
  testConfigurationMessage:function () {
    return this.get('testResult') ? Em.I18n.t('admin.authentication.form.test.success') : Em.I18n.t('admin.authentication.form.test.fail');
  }.property('testResult'),
  testConfigurationClass:function () {
    return this.get('testResult') ? "text-success" : "text-error";
  }.property('testConfigurationMessage')
})
;
});

require.register("models/background_operation", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.BackgroundOperation = DS.Model.extend({
  operationName:DS.attr('string'),
  events: DS.hasMany('App.BackgroundOperationEvent'),
  cluster:DS.belongsTo('App.Cluster'),
  host:DS.belongsTo('App.Host'),
  operationLog: DS.attr('string')
});

App.BackgroundOperation.FIXTURES = [
  {
    id:1,
    operation_name:'Decommissioning host1',
    operation_log:'Decommissioning log',
    events:[1,2],
    cluster_id:1,
    host_id:1
  },
  {
    id:2,
    operation_name:'Starting DataNode on host4',
    operation_log:'Starting DataNode log',
    events:[3],
    cluster_id:1,
    host_id:1
  }
];

App.BackgroundOperationEvent = DS.Model.extend({
  eventName:DS.attr('string'),
  operation:DS.belongsTo('App.BackgroundOperation'),
  eventDate: DS.attr('string')
});

App.BackgroundOperationEvent.FIXTURES = [
  {
    id:1,
    event_name:'Some intermediate operation',
    operation_id:1,
    event_date:'4 min ago'
  },
  {
    id:2,
    event_name:'Operation started',
    operation_id:1,
    event_date:'5 min ago'
  },
  {
    id:3,
    event_name:'Operation started',
    operation_id:2,
    event_date:'5 min ago'
  }
];


});

require.register("models/client_component", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

App.ClientComponent = DS.Model.extend({
  service: DS.belongsTo('App.Service'),
  componentName: DS.attr('string'),
  installedCount: DS.attr('number'),
  startedCount: DS.attr('number'),
  totalCount: DS.attr('number'),
  stackInfo: DS.belongsTo('App.StackServiceComponent'),

  displayName: function() {
    var displayName = App.format.role(this.get('componentName'));
    if (this.get('service.serviceName') === this.get('componentName')) {
      displayName += ' ' + Em.I18n.t('common.client');
    }
    return displayName;
  }.property('componentName'),

  displayNamePluralized: function() {
    return stringUtils.pluralize(this.get('installedCount'), this.get('displayName'));
  }.property('installedCount')
});

App.ClientComponent.FIXTURES = [];

});

require.register("models/cluster", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.Cluster = DS.Model.extend({
  clusterName: DS.attr('string'),
  stackName: DS.attr('string'),
  version: DS.attr('string'),
  hosts: DS.hasMany('App.Host'),
  racks: DS.hasMany('App.Rack'),
  maxHostsPerRack: DS.attr('number'),
  /**
   * Array containing desired configs. New array
   * should be set by instances of class.
   */
  desiredConfigs: null
});

App.Cluster.FIXTURES = [/*
  {
    id: 1,
    cluster_name: 'cluster1',
    stack_name: 'HDP',
    hosts: [1, 2, 3, 4],
    racks: [1, 2, 3, 4, 5, 6],
    max_hosts_per_rack: 10
  }*/
];
});

require.register("models/cluster_states", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');
require('mixins/common/userPref');

App.clusterStatus = Em.Object.create(App.UserPref, {

  /**
   * Cluster name
   * @type {string}
   */
  clusterName: '',

  /**
   * List valid cluster states
   * @type {string[]}
   */
  validStates: [
    'DEFAULT',
    'CLUSTER_NOT_CREATED_1',
    'CLUSTER_DEPLOY_PREP_2',
    'CLUSTER_INSTALLING_3',
    'SERVICE_STARTING_3',
    'CLUSTER_INSTALLED_4',
    'ADD_HOSTS_DEPLOY_PREP_2',
    'ADD_HOSTS_INSTALLING_3',
    'ADD_HOSTS_INSTALLED_4',
    'ADD_SERVICES_DEPLOY_PREP_2',
    'ADD_SERVICES_INSTALLING_3',
    'ADD_SERVICES_INSTALLED_4',
    'STOPPING_SERVICES',
    'STACK_UPGRADING',
    'STACK_UPGRADE_FAILED',
    'STACK_UPGRADED',
    'ADD_SECURITY_STEP_1',
    'ADD_SECURITY_STEP_2',
    'ADD_SECURITY_STEP_3',
    'ADD_SECURITY_STEP_4',
    'DISABLE_SECURITY',
    'HIGH_AVAILABILITY_DEPLOY',
    'ROLLBACK_HIGH_AVAILABILITY'],

  /**
   * Default cluster state
   * @type {string}
   */
  clusterState: 'CLUSTER_NOT_CREATED_1',

  /**
   * Current used wizard <code>controller.name</code>
   * @type {string|null}
   */
  wizardControllerName: null,

  /**
   * Local DB
   * @type {object|null}
   */
  localdb: null,

  /**
   * Persist key
   * @type {string}
   */
  key: 'CLUSTER_CURRENT_STATUS',

  /**
   * Is cluster installed
   * @type {bool}
   */
  isInstalled: function () {
    var notInstalledStates = ['CLUSTER_NOT_CREATED_1', 'CLUSTER_DEPLOY_PREP_2', 'CLUSTER_INSTALLING_3', 'SERVICE_STARTING_3'];
    return !notInstalledStates.contains(this.get('clusterState'));
  }.property('clusterState'),

  /**
   * General info about cluster
   * @type {{clusterName: string, clusterState: string, wizardControllerName: string, localdb: object}}
   */
  value: function () {
    return {
      clusterName: this.get('clusterName'),
      clusterState: this.get('clusterState'),
      wizardControllerName: this.get('wizardControllerName'),
      localdb: this.get('localdb')
    };
  }.property('clusterName', 'clusterState', 'localdb', 'wizardControllerName'),

  /**
   * get cluster data from server and update cluster status
   * @param {bool} overrideLocaldb
   * @return promise object for the get call
   * @method updateFromServer
   */
  updateFromServer: function (overrideLocaldb) {
    this.set('additionalData', {
      user: App.db.getUser(),
      login: App.db.getLoginName(),
      overrideLocaldb: !overrideLocaldb
    });
    return this.getUserPref(this.get('key'));
  },

  /**
   * Success callback for get-persist request
   * @param {object} response
   * @param {object} opt
   * @param {object} params
   * @method getUserPrefSuccessCallback
   */
  getUserPrefSuccessCallback: function (response, opt, params) {
    if (response) {
      if (response.clusterState) {
        this.set('clusterState', response.clusterState);
      }
      if (response.clusterName) {
        this.set('clusterName', response.clusterName);
      }
      if (response.wizardControllerName) {
        this.set('wizardControllerName', response.wizardControllerName);
      }
      if (response.localdb) {
        this.set('localdb', response.localdb);
        // restore HAWizard data if process was started
        var isHAWizardStarted = App.isAccessible('ADMIN') && !App.isEmptyObject(response.localdb.HighAvailabilityWizard);
        if (params.data.overrideLocaldb || isHAWizardStarted) {
          var localdbTables = (App.db.data.app && App.db.data.app.tables) ? App.db.data.app.tables : {};
          App.db.data = response.localdb;
          App.db.setLocalStorage();
          App.db.setUser(params.data.user);
          App.db.setLoginName(params.data.login);
          App.db.data.app.tables = localdbTables;
        }
      }
    }
    // this is to ensure that the local storage namespaces are initialized with all expected namespaces.
    // after upgrading ambari, loading local storage data from the "persist" data saved via an older version of
    // Ambari can result in missing namespaces that are defined in the new version of Ambari.
    App.db.mergeStorage();
  },

  /**
   * Error callback for get-persist request
   * @param {object} request
   * @param {object} ajaxOptions
   * @param {string} error
   * @method getUserPrefErrorCallback
   */
  getUserPrefErrorCallback: function (request, ajaxOptions, error) {
    if (request.status == 404) {
      // default status already set
      console.log('Persist API did NOT find the key CLUSTER_CURRENT_STATUS');
      return;
    }
    App.ModalPopup.show({
      header: Em.I18n.t('common.error'),
      secondary: false,
      bodyClass: Em.View.extend({
        template: Em.Handlebars.compile('<p>{{t common.update.error}}</p>')
      })
    });
  },

  /**
   * update cluster status and post it on server.
   * This function should always be called by admin user
   * @param {object} newValue
   * @param {object} opt - Can have additional params for ajax callBacks and sender
   *                 opt.successCallback
   *                 opt.successCallbackData
   *                 opt.errorCallback
   *                 opt.errorCallbackData
   *                 opt.alwaysCallback
   *                 opt.alwaysCallbackData
   *                 opt.sender
   * @method setClusterStatus
   * @return {*}
   */
  setClusterStatus: function (newValue, opt) {
    if (App.get('testMode')) return false;
    if (!App.isAccessible('ADMIN')) {
      Em.assert('Non-Admin user should not execute setClusterStatus function', true);
    }
    var user = App.db.getUser();
    var login = App.db.getLoginName();
    var val = {clusterName: this.get('clusterName')};
    if (newValue) {
      App.db.cleanTmp();
      //setter
      if (newValue.clusterName) {
        this.set('clusterName', newValue.clusterName);
        val.clusterName = newValue.clusterName;
      }

      if (newValue.clusterState) {
        this.set('clusterState', newValue.clusterState);
        val.clusterState = newValue.clusterState;
      }
      if (newValue.wizardControllerName) {
        this.set('wizardControllerName', newValue.wizardControllerName);
        val.wizardControllerName = newValue.wizardControllerName;
      }

      if (newValue.localdb) {
        if (newValue.localdb.app && newValue.localdb.app.user)
          delete newValue.localdb.app.user;
        if (newValue.localdb.app && newValue.localdb.app.loginName)
          delete newValue.localdb.app.loginName;
        if (newValue.localdb.app && newValue.localdb.app.tables)
          delete newValue.localdb.app.tables;
        this.set('localdb', newValue.localdb);
        val.localdb = newValue.localdb;
      } else {
        delete App.db.data.app.user;
        delete App.db.data.app.loginName;
        delete App.db.data.app.tables;
        val.localdb = App.db.data;
        App.db.setUser(user);
        App.db.setLoginName(login);
      }
      this.postUserPref(this.get('key'), val)
          .done(function () {
            !!opt && Em.typeOf(opt.successCallback) === 'function' && opt.successCallback.call(opt.sender || this, opt.successCallbackData);
          })
          .fail(function () {
            !!opt && Em.typeOf(opt.errorCallback) === 'function' && opt.errorCallback.call(opt.sender || this, opt.errorCallbackData);
          })
          .always(function () {
            !!opt && Em.typeOf(opt.alwaysCallback) === 'function' && opt.alwaysCallback.call(opt.sender || this, opt.alwaysCallbackData);
          });
      return newValue;
    }
  },

  /**
   * Error callback for post-persist request
   * @param {object} request
   * @param {object} ajaxOptions
   * @param {string} error
   * @method postUserPrefErrorCallback
   */
  postUserPrefErrorCallback: function (request, ajaxOptions, error) {
    console.log("ERROR");
    var msg = '', doc;
    try {
      msg = 'Error ' + (request.status) + ' ';
      doc = $.parseXML(request.responseText);
      msg += $(doc).find("body p").text();
    } catch (e) {
      msg += JSON.parse(request.responseText).message;
    }

    App.ModalPopup.show({
      header: Em.I18n.t('common.error'),
      secondary: false,
      response: msg,
      bodyClass: Em.View.extend({
        template: Em.Handlebars.compile('<p>{{t common.persist.error}} {{response}}</p>')
      })
    });
  }

});

});

require.register("models/config_group", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Represents a configuration-group on the cluster.
 * A configuration-group is a collection of hosts
 * on which a collection of configurations are applied.
 *
 * Configuration group hierarchy is at 2 levels. For
 * each service there is a 'Default' configuration group
 * containing all hosts not belonging to any group of that
 * service.
 *
 * A default configuration group has child configuration
 * groups which contain configuration overrides (deltas)
 * for a bunch of hosts. This allows different configurations
 * for different hosts in a heterogeneous cluster environment.
 */
App.ConfigGroup = Ember.Object.extend({
  id: null,
  name: null,
  description: null,
  isDefault: null,
  serviceName: null,

  /**
   * Parent configuration group for this group.
   * When {@link #isDefault} is true, this value is <code>null</code>
   * When {@link #isDefault} is false, this represents the configuration
   * deltas that are applied on the default.
   */
  parentConfigGroup: null,
  /**
   * all hosts that belong to cluster
   */
  clusterHostsBinding: 'App.router.manageConfigGroupsController.clusterHosts',

  /**
   * Children configuration groups for this group.
   * When {@link #isDefault} is false, this value is <code>null</code>
   * When {@link #isDefault} is true, this represents the various
   * configuration groups that override the default.
   */
  childConfigGroups: [],

  /**
   * Service for which this configuration-group
   * is applicable.
   */
  service: null,

  /**
   * Hosts on which this configuration-group
   * is to be applied. For a service, a host can
   * belong to only one non-default configuration-group.
   *
   * When {#isDefault} is false, this contains hosts
   * for which the overrides will apply.
   *
   * When {#isDefault} is true, this value is empty, as
   * it dynamically reflects hosts not belonging to other
   * non-default groups.
   *
   */
  hosts: [],

  /**
   * this flag is used for installed services' config groups
   * if user make changes to them - mark this flag to true
   */
  isForUpdate: false,

  /**
   * mark config groups for installed services
   */
  isForInstalledService: false,
  /**
   * Provides a display friendly name. This includes trimming
   * names to a certain length.
   */
  displayName: function () {
    var name = this.get('name');
    if (name && name.length>App.config.CONFIG_GROUP_NAME_MAX_LENGTH) {
      var middle = Math.floor(App.config.CONFIG_GROUP_NAME_MAX_LENGTH / 2);
      name = name.substring(0, middle) + "..." + name.substring(name.length-middle);
    }
    return name;
  }.property('name'),

  displayNameHosts: function () {
    return this.get('displayName') + ' (' + this.get('hosts.length') + ')';
  }.property('displayName', 'hosts.length'),

  apiResponse: null,

  /**
   * Provides hosts which are available for inclusion in
   * non-default configuration groups.
   */
  availableHosts: function () {
    if (this.get('isDefault')) return [];
    var unusedHostsMap = {};
    var availableHosts = [];
    var sharedHosts = this.get('clusterHosts');
    // parentConfigGroup.hosts(hosts from default group) - are available hosts, which don't belong to any group
    this.get('parentConfigGroup.hosts').forEach(function (hostName) {
      unusedHostsMap[hostName] = true;
    });
    sharedHosts.forEach(function (host) {
      if (unusedHostsMap[host.get('id')]) {
        availableHosts.pushObject(Ember.Object.create({
          selected: false,
          host: host,
          hostComponentNames: host.get('hostComponents').mapProperty('componentName')
        }));
      }
    });
    return availableHosts;
  }.property('isDefault', 'parentConfigGroup', 'childConfigGroups', 'parentConfigGroup.hosts.@each'),

  isAddHostsDisabled: function () {
    return (this.get('isDefault') || this.get('availableHosts.length') === 0);
  }.property('availableHosts.length'),

  /**
   * Collection of (site, tag) pairs representing properties.
   *
   * When {#isDefault} is true, this represents the
   * default cluster configurations for that service.
   *
   * When {#isDefault} is false, this represents the
   * configuration overrides on top of the cluster default for the
   * hosts identified by 'hosts'.
   */
  configSiteTags: [],

  properties: [],

  propertiesList: function () {
    var result = '';
    this.get('properties').forEach(function (item) {
      result += item.name + " : " + item.value + '<br/>';
    }, this);
    return result;
  }.property('properties.length'),

  hash: null
});

});

require.register("models/form", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');

// move this to models cause some errors
App.Form = Em.View.extend({
  /**
   * generating fields from fieldsOptions
   */
  classNames:["form-horizontal"],
  attributeBindings: ['autocomplete'],
  autocomplete: 'off',
  i18nprefix:'form.',
  fields:[],
  field:{},
  messages:[],
  object:false,
  result:0, // save result var (-1 - error; 0 - init; 1 - success)
  templateName:require('templates/common/form'),
  tagName:'form',

  init:function () {
    var thisForm = this;
    if (!this.fields.length) {
      this.fieldsOptions.forEach(
        function (options) {
          var field = App.FormField.create(options);
          field.set('form', thisForm);
          thisForm.fields.push(field);
          thisForm.set("field." + field.get('name'), field);
        }
      );
    }
    this._super();
  },

  /**
   * get field of form by name
   * @param name
   * @return {Object}
   */
  getField: function (name) {
    return this.get('fields').findProperty('name', name);
  },

  isValid:function () {
    var isValid = true;
    $.each(this.fields, function () {
      this.validate();
      if (!this.get('isValid')) {
        isValid = false;
        console.warn(this.get('name') + " IS INVALID : " + this.get('errorMessage'));
      }
    });

    return isValid;
  },

  updateValues:function () {
    var object = this.get('object');
    if (object instanceof Em.Object) {
      $.each(this.fields, function () {
        this.set('value', (this.get('displayType') == 'password') ? '' : object.get(this.get('name')));
      });
    } else {
      this.clearValues();
    }

  }.observes("object"),

  /**
   * reset values to default of every field in the form
   */
  clearValues: function () {
    this.get('fields').forEach(function (field) {
      var value = (field.get('defaultValue') === undefined) ? '' : field.get('defaultValue');
      field.set('value', value);
    }, this);
  },

  visibleFields:function () {
    return this.get('fields').filterProperty('isHiddenField', false);
  }.property('fields'),

  resultText:function () {
    var text = "";
    switch (this.get('result')) {
      case -1:
        text = this.t("form.saveError");
        break;
      case 1:
        text = this.t("form.saveSuccess");
        break;
    }

    return text;
  }.property('result')
});

App.FormField = Em.Object.extend({ // try to realize this as view
  name:'',
  displayName:'',
//  defaultValue:'', NOT REALIZED YET
  description:'',
  disabled:false,
  displayType:'string', // string, digits, number, directories, textarea, checkbox
  disableRequiredOnPresent:false,
  errorMessage:'',
  warnMessage:'',
  form:false,
  isRequired:true, // by default a config property is required
  unit:'',
  value:'',

  observeValue:function () {

    if (this.get('displayType') == 'hidden')
      console.warn(" FORM FIELD VALUE: ", this.get('value'));

  }.observes('value'),

  isValid:function () {
    return this.get('errorMessage') === '';
  }.property('errorMessage'),

  viewClass:function () {
    var options = {};
    var element = Em.TextField;
    switch (this.get('displayType')) {
      case 'checkbox':
        element = Em.Checkbox;
        options.checkedBinding = "value";
        break;
      case 'select':
        element = Em.Select;
        options.content = this.get('values');
        options.valueBinding = "value";
        options.optionValuePath = "content.value";
        options.optionLabelPath = "content.label";
        break;
      case 'password':
        options['type'] = 'password';
        break;
      case 'textarea':
        element = Em.TextArea;
        break;
      case 'hidden':
        options.type = "hidden";
        break;
    }

    return element.extend(options);
  }.property('displayType'),

  validate:function () {
    var value = this.get('value');
    var isError = false;
    this.set('errorMessage', '');

    if (this.get('isRequired') && (typeof value === 'string' && value.trim().length === 0)) {
      this.set('errorMessage', 'This is required');
      isError = true;
    }

    if (typeof value === 'string' && value.trim().length === 0) { // this is not to validate empty field.
      isError = true;
    }

    if (!isError) {
      this.set('errorMessage', '');
    }
  },

  isHiddenField:function () {
    return this.get('displayType') == 'hidden';
  }.property('type')
});

});

require.register("models/host", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');
var misc = require('utils/misc');

App.Host = DS.Model.extend({
  hostName: DS.attr('string'),
  publicHostName: DS.attr('string'),
  cluster: DS.belongsTo('App.Cluster'),
  hostComponents: DS.hasMany('App.HostComponent'),
  cpu: DS.attr('string'),
  cpuPhysical: DS.attr('string'),
  memory: DS.attr('string'),
  diskTotal: DS.attr('number'),
  diskFree: DS.attr('number'),
  osArch: DS.attr('string'),
  ip: DS.attr('string'),
  rack: DS.attr('string'),
  healthStatus: DS.attr('string'),
  lastHeartBeatTime: DS.attr('number'),
  osType: DS.attr("string"),
  diskInfo: DS.attr('object'),
  loadOne:DS.attr('number'),
  loadFive:DS.attr('number'),
  loadFifteen:DS.attr('number'),
  memTotal:DS.attr('number'),
  memFree:DS.attr('number'),
  cpuSystem:DS.attr('number'),
  cpuUser:DS.attr('number'),
  criticalWarningAlertsCount: DS.attr('number'),
  alertsSummary: DS.attr('object'),
  passiveState: DS.attr('string'),
  index: DS.attr('number'),
  stackVersions: DS.hasMany('App.HostStackVersion'),

  /**
   * Is host checked at the main Hosts page
   */
  selected:DS.attr('boolean'),
  /**
   * determine whether host is requested from server
   */
  isRequested: DS.attr('boolean'),

  currentVersion: function () {
    var current = this.get('stackVersions').findProperty('isCurrent');
    return current ? current.get('repoVersion') : null;
  }.property('stackVersions.@each.isCurrent'),

  /**
   * Overall CPU usage (system and user)
   * @returns {Number}
   */
  cpuUsage: function () {
    if (this.get('cpuSystem') && this.get('cpuUser')) {
      return this.get('cpuSystem') + this.get('cpuUser');
    }
    return 0;
  }.property('cpuSystem', 'cpuUser'),

  /**
   * Percent value of used memory
   * @returns {Number}
   */
  memoryUsage: function () {
    if (this.get('memFree') && this.get('memTotal')) {
      var memUsed = this.get('memTotal') - this.get('memFree');
      return (100 * memUsed) / this.get('memTotal');
    }
    return 0;
  }.property('memTotal', 'memFree'),

  componentsWithStaleConfigs: function () {
    return this.get('hostComponents').filterProperty('staleConfigs', true);
  }.property('hostComponents.@each.staleConfigs'),

  /**
   * Get count of host components with stale configs
   * @returns {Number}
   */
  componentsWithStaleConfigsCount: function() {
    return this.get('componentsWithStaleConfigs').length;
  }.property('componentsWithStaleConfigs.length'),

  /**
   * Get count of host components in passive state
   * @returns {Number}
   */
  componentsInPassiveStateCount: function() {
    return this.get('hostComponents').filter(function(component) {
      return component.get('passiveState') !== 'OFF';
    }).length;
  }.property('hostComponents.@each.passiveState'),

  /**
   * Count of mounted on host disks
   * @returns {Number}
   */
  disksMounted: function() {
    return this.get('diskInfo.length');
  }.property('diskInfo.length'),

  coresFormatted: function() {
    return this.get('cpu') + ' (' + this.get('cpuPhysical') + ')';
  }.property('cpu', 'cpuPhysical'),

  /**
   * API return diskTotal and diskFree. Need to save their different
   * @returns {Number}
   */
  diskUsed: function(){
    return this.get('diskTotal') - this.get('diskFree');
  }.property('diskFree', 'diskTotal'),

  /**
   * Format diskUsed value to float with 2 digits (also convert to GB)
   * @returns {String} Format: '*** GB'
   */
  diskUsedFormatted: function() {
    return Math.round(this.get('diskUsed') * Math.pow(10, 2)) / Math.pow(10, 2) + 'GB';
  }.property('diskUsed'),

  /**
   * Format diskTotal value to float with 2 digits (also convert to GB)
   * @returns {String} Format: '*** GB'
   */
  diskTotalFormatted: function() {
    return Math.round(this.get('diskTotal') * Math.pow(10, 2)) / Math.pow(10, 2) + 'GB';
  }.property('diskTotal'),

  /**
   * Percent value of used disk space
   * @returns {Number}
   */
  diskUsage: function() {
    return (this.get('diskUsed')) / this.get('diskTotal') * 100;
  }.property('diskUsed', 'diskTotal'),

  /**
   * Format diskUsage to float with 2 digits
   * @returns {String} Format: '**.** %'
   */
  diskUsageFormatted: function() {
    if (isNaN(this.get('diskUsage')) || this.get('diskUsage') < 0) {
      return Em.I18n.t('hosts.host.metrics.dataUnavailable');
    }
    var s = Math.round(this.get('diskUsage') * Math.pow(10, 2)) / Math.pow(10, 2);
    if (isNaN(s)) {
      s = 0;
    }
    return s + '%';
  }.property('diskUsage'),

  /**
   * Formatted string with data about disk usage
   * @returns {String}
   */
  diskInfoBar: function() {
    if (isNaN(this.get('diskUsage')) || this.get('diskUsage') < 0) {
      return this.get('diskUsageFormatted');
    }
    return this.get('diskUsedFormatted') + '/' + this.get('diskTotalFormatted') + ' (' + this.get('diskUsageFormatted')
      + ' ' + Em.I18n.t('services.service.summary.diskInfoBar.used') + ')';
  }.property('diskUsedFormatted', 'diskTotalFormatted'),

  /**
   * Formatted bytes to appropriate value
   * @returns {String}
   */
  memoryFormatted: function () {
    return misc.formatBandwidth(this.get('memory') * 1024);
  }.property('memory'),

  /**
   * Return true if the host <code>healthStatus</code> is UNKNOWN
   * @returns {bool}
   */
  isNotHeartBeating : function() {
    return (App.get('testMode')) ? false : (this.get('healthStatus') === "UNKNOWN");
  }.property('lastHeartBeatTime'),

  /**
   * Average load
   * @returns {Number}
   */
  loadAvg: function() {
    if (this.get('loadOne') != null) return this.get('loadOne').toFixed(2);
    if (this.get('loadFive') != null) return this.get('loadFive').toFixed(2);
    if (this.get('loadFifteen') != null) return this.get('loadFifteen').toFixed(2);
    return null;
  }.property('loadOne', 'loadFive', 'loadFifteen'),

  /**
   * Host health indicator
   * Based on  <code>healthStatus</code>
   * @returns {String}
   */
  healthClass: function(){
    if (this.get('passiveState')!= 'OFF') {
      return 'icon-medkit';
    }
    var statusMap = {
      'UNKNOWN': 'health-status-DEAD-YELLOW',
      'HEALTHY': 'health-status-LIVE',
      'UNHEALTHY': 'health-status-DEAD-RED',
      'ALERT': 'health-status-DEAD-ORANGE'
    };
    return statusMap[this.get('healthStatus')] || 'health-status-DEAD-YELLOW';
  }.property('healthStatus'),

  healthIconClass: function () {
    switch (this.get('healthClass')) {
      case 'health-status-LIVE':
        return App.healthIconClassGreen;
        break;
      case 'health-status-DEAD-RED':
        return App.healthIconClassRed;
        break;
      case 'health-status-DEAD-YELLOW':
        return App.healthIconClassYellow;
        break;
      case 'health-status-DEAD-ORANGE':
        return App.healthIconClassOrange;
        break;
      default:
        return "";
        break;
    }
  }.property('healthClass'),

  /**
   * Tooltip for host indicator
   * Contains affected host components names (based on <code>healthClass</code>)
   * @returns {String}
   */
  healthToolTip: function(){
    var hostComponents = this.get('hostComponents').filter(function(item){
      return item.get('workStatus') !== App.HostComponentStatus.started;
    });
    var output = '';
    if (this.get('passiveState') != 'OFF') {
      return Em.I18n.t('hosts.host.passive.mode');
    }
    switch (this.get('healthClass')){
      case 'health-status-DEAD-RED':
        hostComponents = hostComponents.filterProperty('isMaster', true);
        output = Em.I18n.t('hosts.host.healthStatus.mastersDown');
        hostComponents.forEach(function(hc, index){
          output += (index == (hostComponents.length-1)) ? hc.get('displayName') : (hc.get('displayName')+", ");
        }, this);
        break;
      case 'health-status-DEAD-YELLOW':
        output = Em.I18n.t('hosts.host.healthStatus.heartBeatNotReceived');
        break;
      case 'health-status-DEAD-ORANGE':
        hostComponents = hostComponents.filterProperty('isSlave', true);
        output = Em.I18n.t('hosts.host.healthStatus.slavesDown');
        hostComponents.forEach(function(hc, index){
          output += (index == (hostComponents.length-1)) ? hc.get('displayName') : (hc.get('displayName')+", ");
        }, this);
        break;
      case 'health-status-LIVE':
        output = Em.I18n.t('hosts.host.healthStatus.allUp');
        break;
    }
    return output;
  }.property('hostComponents.@each.workStatus','hostComponents.@each.passiveState')
});

App.Host.FIXTURES = [];

});

require.register("models/host_component", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HostComponent = DS.Model.extend({
  workStatus: DS.attr('string'),
  passiveState: DS.attr('string'),
  componentName: DS.attr('string'),
  haStatus: DS.attr('string'),
  displayNameAdvanced: DS.attr('string'),
  staleConfigs: DS.attr('boolean'),
  host: DS.belongsTo('App.Host'),
  hostName: DS.attr('string'),
  service: DS.belongsTo('App.Service'),
  adminState: DS.attr('string'),
  /**
   * Determine if component is client
   * @returns {bool}
   */
  isClient: function () {
    return App.get('components.clients').contains(this.get('componentName'));
  }.property('componentName'),
  /**
   * Determine if component is running now
   * Based on <code>workStatus</code>
   * @returns {bool}
   */
  isRunning: function () {
    return (this.get('workStatus') == 'STARTED' || this.get('workStatus') == 'STARTING');
  }.property('workStatus'),

  /**
   * Formatted <code>componentName</code>
   * @returns {String}
   */
  displayName: function () {
    return App.format.role(this.get('componentName'));
  }.property('componentName'),

  /**
   * Determine if component is master
   * @returns {bool}
   */
  isMaster: function () {
    return App.get('components.masters').contains(this.get('componentName'));
  }.property('componentName', 'App.components.masters'),

  /**
   * Determine if component is slave
   * @returns {bool}
   */
  isSlave: function () {
    return App.get('components.slaves').contains(this.get('componentName'));
  }.property('componentName'),
  /**
   * Only certain components can be deleted.
   * They include some from master components,
   * some from slave components, and rest from
   * client components.
   * @returns {bool}
   */
  isDeletable: function () {
    return App.get('components.deletable').contains(this.get('componentName'));
  }.property('componentName'),
  /**
   * A host-component is decommissioning when it is in HDFS service's list of
   * decomNodes.
   * @returns {bool}
   */
  isDecommissioning: function () {
    var decommissioning = false;
    var hostName = this.get('hostName');
    var componentName = this.get('componentName');
    var hdfsSvc = App.HDFSService.find().objectAt(0);
    if (componentName === 'DATANODE' && hdfsSvc) {
      var decomNodes = hdfsSvc.get('decommissionDataNodes');
      var decomNode = decomNodes != null ? decomNodes.findProperty("hostName", hostName) : null;
      decommissioning = decomNode != null;
    }
    return decommissioning;
  }.property('componentName', 'hostName', 'App.router.clusterController.isLoaded', 'App.router.updateController.isUpdated'),
  /**
   * User friendly host component status
   * @returns {String}
   */
  isActive: function () {
    return (this.get('passiveState') == 'OFF');
  }.property('passiveState'),

  passiveTooltip: function () {
    if (!this.get('isActive')) {
      return Em.I18n.t('hosts.component.passive.mode');
    }
  }.property('isActive'),

  statusClass: function () {
    return this.get('isActive') ? this.get('workStatus') : 'icon-medkit';
  }.property('workStatus', 'isActive'),

  statusIconClass: function () {
    switch (this.get('statusClass')) {
      case 'STARTED':
      case 'STARTING':
        return App.healthIconClassGreen;
        break;
      case 'INSTALLED':
      case 'STOPPING':
        return App.healthIconClassRed;
        break;
      case 'UNKNOWN':
        return App.healthIconClassYellow;
        break;
      default:
        return "";
        break;
    }
  }.property('statusClass'),

  componentTextStatus: function () {
    return App.HostComponentStatus.getTextStatus(this.get("workStatus"));
  }.property('workStatus', 'isDecommissioning')
});

App.HostComponent.FIXTURES = [];

App.HostComponentStatus = {
  started: "STARTED",
  starting: "STARTING",
  stopped: "INSTALLED",
  stopping: "STOPPING",
  install_failed: "INSTALL_FAILED",
  installing: "INSTALLING",
  upgrade_failed: "UPGRADE_FAILED",
  unknown: "UNKNOWN",
  disabled: "DISABLED",
  init: "INIT",

  /**
   * Get host component status in "machine" format
   * @param {String} value
   * @returns {String}
   */
  getKeyName: function (value) {
    switch (value) {
      case this.started:
        return 'started';
      case this.starting:
        return 'starting';
      case this.stopped:
        return 'installed';
      case this.stopping:
        return 'stopping';
      case this.install_failed:
        return 'install_failed';
      case this.installing:
        return 'installing';
      case this.upgrade_failed:
        return 'upgrade_failed';
      case this.disabled:
      case this.unknown:
        return 'unknown';
    }
    return 'unknown';
  },

  /**
   * Get user-friendly host component status
   * @param {String} value
   * @returns {String}
   */
  getTextStatus: function (value) {
    switch (value) {
      case this.installing:
        return 'Installing...';
      case this.install_failed:
        return 'Install Failed';
      case this.stopped:
        return 'Stopped';
      case this.started:
        return 'Started';
      case this.starting:
        return 'Starting...';
      case this.stopping:
        return 'Stopping...';
      case this.unknown:
        return 'Heartbeat Lost';
      case this.upgrade_failed:
        return 'Upgrade Failed';
      case this.disabled:
        return 'Disabled';
      case this.init:
        return 'Install Pending...';
    }
    return 'Unknown';
  },

  /**
   * Get list of possible <code>App.HostComponent</code> statuses
   * @returns {String[]}
   */
  getStatusesList: function () {
    var ret = [];
    for (var st in this) {
      if (this.hasOwnProperty(st) && Em.typeOf(this[st]) == 'string') {
        ret.push(this[st]);
      }
    }
    return ret;
  }
};


});

require.register("models/host_stack_version", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HostStackVersion = DS.Model.extend({
  stack: DS.attr('string'),
  version: DS.attr('string'),
  repo: DS.belongsTo('App.Repository'),
  repoVersion: DS.attr('string'),
  displayName: DS.attr('string'),
  isVisible: DS.attr('boolean', {defaultValue: true}),
  /**
   * possible property value defined at App.HostStackVersion.statusDefinition
   * @type {string}
   */
  status: DS.attr('string'),
  host: DS.belongsTo('App.Host'),
  hostName: DS.attr('string'),

  /**
   * @type {boolean}
   */
  isCurrent: function () {
    return this.get('status') === 'CURRENT'
  }.property('status'),

  /**
   * @type {boolean}
   */
  isInstalling: function () {
    return this.get('status') === 'INSTALLING';
  }.property('status'),

  /**
   * @type {boolean}
   */
  isOutOfSync: function () {
    return this.get('status') === 'OUT_OF_SYNC';
  }.property('status'),
  /**
   * @type {string}
   */
  displayStatus: function() {
    return App.HostStackVersion.formatStatus(this.get('status'));
  }.property('status'),

  /**
   * @type {boolean}
   */
  installEnabled: function () {
    return ['OUT_OF_SYNC', 'INSTALL_FAILED'].contains(this.get('status'));
  }.property('status'),
  installDisabled: Ember.computed.not('installEnabled')
});

App.HostStackVersion.FIXTURES = [];

/**
 * definition of possible statuses of Stack Version
 * @type {Array}
 */
App.HostStackVersion.statusDefinition = [
  "INSTALLED",
  "INSTALLING",
  "INSTALL_FAILED",
  "OUT_OF_SYNC",
  "CURRENT",
  "UPGRADING",
  "UPGRADE_FAILED"
];

/**
 * translate status to label
 * @param status
 * @return {string}
 */
App.HostStackVersion.formatStatus = function (status) {
  return App.HostStackVersion.statusDefinition.contains(status) ?
    Em.I18n.t('hosts.host.stackVersions.status.' + status.toLowerCase()) :
    status.toCapital();
};

});

require.register("models/hosts", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HostInfo = Ember.Object.extend({
  elementId: 'host',
  name: '',
  cpu: null,
  memory: null,
  message: 'Information',
  barColor: 'progress-info',
  isChecked: true,
  bootLog:null,
  bootStatus: 'PENDING',

  bootStatusForDisplay:function () {
    switch (this.get('bootStatus')) {
      case 'PENDING':
        return 'Preparing';
      case 'REGISTERED':
        return 'Success';
      case 'FAILED':
        return 'Failed';
      case 'RUNNING':
        return 'Installing';
      case 'DONE':
      case 'REGISTERING':
      default:
        return 'Registering';
    }
  }.property('bootStatus'),

  bootBarColor:function () {
    switch (this.get('bootStatus')) {
      case 'REGISTERED':
        return 'progress-success';
      case 'FAILED':
        return 'progress-danger';
      case 'PENDING':
      case 'RUNNING':
      case 'DONE':
      case 'REGISTERING':
      default:
        return 'progress-info';
    }
  }.property('bootStatus'),

  bootStatusColor:function () {
    switch (this.get('bootStatus')) {
      case 'REGISTERED':
        return 'text-success';
      case 'FAILED':
        return 'text-error';
      case 'PENDING':
      case 'RUNNING':
      case 'DONE':
      case 'REGISTERING':
      default:
        return 'text-info';
    }
  }.property('bootStatus'),

  isBootDone:function () {
    switch (this.get('bootStatus')) {
      case 'REGISTERED':
      case 'FAILED':
        return true;
      case 'PENDING':
      case 'RUNNING':
      case 'DONE':
      case 'REGISTERING':
      default:
        return false;
    }

  }.property('bootStatus')
});

});

require.register("models/operating_system", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.OperatingSystem = DS.Model.extend({
  id:  DS.attr('string'), // This is ${stackName}-${stackVersion}-${osType}.
  osType: DS.attr('string'),
  stackName: DS.attr('string'),
  stackVersion: DS.attr('string'),
  repositories: DS.hasMany('App.Repository'),
  stack: DS.belongsTo('App.Stack'),
  isSelected: DS.attr('boolean', {defaultValue: true})
});


App.OperatingSystem.FIXTURES = [];
});

require.register("models/quick_links", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var portRegex = '\\w*:(\\d+)';

App.QuickLinks = DS.Model.extend({
  label: DS.attr('string'),
  url: DS.attr('string'),
  service_id: DS.attr('string'),
  template: DS.attr('string'),
  http_config: DS.attr('string'),
  https_config: DS.attr('string'),
  site: DS.attr('string'),
  regex: DS.attr('string'),
  default_http_port: DS.attr('number'),
  default_https_port: DS.attr('number')
});


App.QuickLinks.FIXTURES = [
  {
    id:1,
    label:'NameNode UI',
    url:'%@://%@:%@',
    service_id: 'HDFS',
    template:'%@://%@:%@',
    http_config: 'dfs.namenode.http-address',
    https_config: 'dfs.namenode.https-address',
    site: 'hdfs-site',
    regex: portRegex,
    default_http_port: 50070,
    default_https_port: 50470
  },
  {
    id:2,
    label:'NameNode logs',
    url:'%@://%@:%@/logs',
    service_id: 'HDFS',
    template:'%@://%@:%@/logs',
    http_config: 'dfs.namenode.http-address',
    https_config: 'dfs.namenode.https-address',
    site: 'hdfs-site',
    regex: portRegex,
    default_http_port: 50070,
    default_https_port: 50470
  },
  {
    id:3,
    label:'NameNode JMX',
    url:'%@://%@:%@/jmx',
    service_id: 'HDFS',
    template:'%@://%@:%@/jmx',
    http_config: 'dfs.namenode.http-address',
    https_config: 'dfs.namenode.https-address',
    site: 'hdfs-site',
    regex: portRegex,
    default_http_port: 50070,
    default_https_port: 50470
  },
  {
    id:4,
    label:'Thread Stacks',
    url:'%@://%@:%@/stacks',
    service_id: 'HDFS',
    template:'%@://%@:%@/stacks',
    http_config: 'dfs.namenode.http-address',
    https_config: 'dfs.namenode.https-address',
    site: 'hdfs-site',
    regex: portRegex,
    default_http_port: 50070,
    default_https_port: 50470
  },
  {
    id:13,
    label:'HBase Master UI',
    url:'%@://%@:%@/master-status',
    service_id: 'HBASE',
    template:'%@://%@:%@/master-status',
    http_config: 'hbase.master.info.port',
    site: 'hbase-site',
    regex: '^(\\d+)$',
    default_http_port: 60010
  },
  {
    id:14,
    label:'HBase Logs',
    url:'%@://%@:60010/logs',
    service_id: 'HBASE',
    template:'%@://%@:%@/logs',
    http_config: 'hbase.master.info.port',
    site: 'hbase-site',
    regex: '^(\\d+)$',
    default_http_port: 60010
  },
  {
    id:15,
    label:'Zookeeper Info',
    url:'%@://%@:60010/zk.jsp',
    service_id: 'HBASE',
    template:'%@://%@:%@/zk.jsp',
    http_config: 'hbase.master.info.port',
    site: 'hbase-site',
    regex: '^(\\d+)$',
    default_http_port: 60010
  },
  {
    id:16,
    label:'HBase Master JMX',
    url:'%@://%@:60010/jmx',
    service_id: 'HBASE',
    template:'%@://%@:%@/jmx',
    http_config: 'hbase.master.info.port',
    site: 'hbase-site',
    regex: '^(\\d+)$',
    default_http_port: 60010
  },
  {
    id:17,
    label:'Debug Dump',
    url:'%@://%@:%@/dump',
    service_id: 'HBASE',
    template:'%@://%@:%@/dump',
    http_config: 'hbase.master.info.port',
    site: 'hbase-site',
    regex: '^(\\d+)$',
    default_http_port: 60010
  },
  {
    id:18,
    label:'Thread Stacks',
    url:'%@://%@:%@/stacks',
    service_id: 'HBASE',
    template:'%@://%@:%@/stacks',
    http_config: 'hbase.master.info.port',
    site: 'hbase-site',
    regex: '^(\\d+)$',
    default_http_port: 60010
  },
  {
    id:19,
    label:'Oozie Web UI',
    url:'%@://%@:%@/oozie',
    service_id: 'OOZIE',
    template:'%@://%@:%@/oozie',
    http_config: 'oozie.base.url',
    site: 'oozie-site',
    regex: portRegex,
    default_http_port: 11000
  },
  {
    id:20,
    label:'Ganglia Web UI',
    url:'%@://%@/ganglia',
    service_id: 'GANGLIA',
    template:'%@://%@/ganglia'

  },
  {
    id:23,
    label:'ResourceManager UI',
    url:'%@://%@:%@',
    service_id: 'YARN',
    template:'%@://%@:%@',
    http_config: 'yarn.resourcemanager.webapp.address',
    https_config: 'yarn.resourcemanager.webapp.https.address',
    site: 'yarn-site',
    regex: portRegex,
    default_http_port: 8088,
    default_https_port: 8090

  },
  {
    id:24,
    label:'ResourceManager logs',
    url:'%@://%@:%@/logs',
    service_id: 'YARN',
    template:'%@://%@:%@/logs',
    http_config: 'yarn.resourcemanager.webapp.address',
    https_config: 'yarn.resourcemanager.webapp.https.address',
    site: 'yarn-site',
    regex: portRegex,
    default_http_port: 8088,
    default_https_port: 8090
  },
  {
    id:25,
    label:'ResourceManager JMX',
    url:'%@://%@:%@/jmx',
    service_id: 'YARN',
    template:'%@://%@:%@/jmx',
    http_config: 'yarn.resourcemanager.webapp.address',
    https_config: 'yarn.resourcemanager.webapp.https.address',
    site: 'yarn-site',
    regex: portRegex,
    default_http_port: 8088,
    default_https_port: 8090
  },
  {
    id:26,
    label:'Thread Stacks',
    url:'%@://%@:%@/stacks',
    service_id: 'YARN',
    template:'%@://%@:%@/stacks',
    http_config: 'yarn.resourcemanager.webapp.address',
    https_config: 'yarn.resourcemanager.webapp.https.address',
    site: 'yarn-site',
    regex: portRegex,
    default_http_port: 8088,
    default_https_port: 8090
  },
  {
    id:27,
    label:'JobHistory UI',
    url:'%@://%@:%@',
    service_id: 'MAPREDUCE2',
    template:'%@://%@:%@',
    http_config: 'mapreduce.jobhistory.webapp.address',
    https_config: 'mapreduce.jobhistory.webapp.https.address',
    site: 'mapred-site',
    regex: portRegex,
    default_http_port: 19888
  },
  {
    id:28,
    label:'JobHistory logs',
    url:'%@://%@:%@/logs',
    service_id: 'MAPREDUCE2',
    template:'%@://%@:%@/logs',
    http_config: 'mapreduce.jobhistory.webapp.address',
    https_config: 'mapreduce.jobhistory.webapp.https.address',
    site: 'mapred-site',
    regex: portRegex,
    default_http_port: 19888
  },
  {
    id:29,
    label:'JobHistory JMX',
    url:'%@://%@:%@/jmx',
    service_id: 'MAPREDUCE2',
    template:'%@://%@:%@/jmx',
    http_config: 'mapreduce.jobhistory.webapp.address',
    https_config: 'mapreduce.jobhistory.webapp.https.address',
    site: 'mapred-site',
    regex: portRegex,
    default_http_port: 19888
  },
  {
    id:30,
    label:'Thread Stacks',
    url:'%@://%@:%@/stacks',
    service_id: 'MAPREDUCE2',
    template:'%@://%@:%@/stacks',
    http_config: 'mapreduce.jobhistory.webapp.address',
    https_config: 'mapreduce.jobhistory.webapp.https.address',
    site: 'mapred-site',
    regex: portRegex,
    default_http_port: 19888
  },
  {
    id:31,
    label:'Storm UI',
    url:'%@://%@:%@/',
    service_id: 'STORM',
    template:'%@://%@:%@/',
    http_config: 'ui.port',
    site: 'storm-site',
    regex: '^(\\d+)$',
    default_http_port: 8744
  },
  {
    id:32,
    label:'Falcon Web UI',
    url:'%@://%@:%@/',
    service_id: 'FALCON',
    template:'%@://%@:%@/',
    http_config: 'falcon_port',
    site: 'falcon-env',
    regex: '^(\\d+)$',
    default_http_port: 15000
  },
  {
    id: 33,
    label:'Ranger Admin UI',
    url:'%@://%@:6080/',
    service_id: 'RANGER',
    template:'%@://%@:6080/'
  },
  {
    id: 34,
    label:'Spark History Server UI',
    url:'%@://%@:%@/',
    service_id: 'SPARK',
    template:'%@://%@:%@/',
    http_config: 'spark.history.ui.port',
    site: 'spark-defaults',
    regex: '^(\\d+)$',
    default_http_port: 18080
  },
  {
    id: 35,
    label:'Eagle Web',
    url:'%@://%@:9099/eagle-service',
    service_id: 'EAGLE',
    template:'%@://%@:9099/eagle-service'
  },
  {
    id: 36,
    label:'Eagle Topology',
    url:'%@://%@:8744',
    service_id: 'EAGLE',
    template:'%@://%@:8744'
  }
];

});

require.register("models/rack", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.Rack = DS.Model.extend({
  name: DS.attr('string'),
  status: DS.attr('string'),
  criticalHostsCount: DS.attr('number'),
  deadHostsCount: DS.attr('number'),
  liveHostsCount: DS.attr('number'),
  hosts: []
});

App.Rack.FIXTURES = [
  //here example of data
  /*{
    id: 1,
    name: 'Rack-0',
    hosts: ['host01', 'host06', 'host05'],
    status: 'LIVE',
    live_hosts_count: 5,
    critical_hosts_count: 0,
    dead_hosts_count: 2
  }*/
];
});

require.register("models/repository", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.Repository = DS.Model.extend({
  id:  DS.attr('string'), // This is ${osType}-${repoId}.
  repoId: DS.attr('string'),
  osType: DS.attr('string'),
  baseUrl: DS.attr('string'),
  defaultBaseUrl: DS.attr('string'),
  latestBaseUrl: DS.attr('string'),
  repoName: DS.attr('string'),
  stackName: DS.attr('string'),
  stackVersion: DS.attr('string'),
  operatingSystem: DS.belongsTo('App.OperatingSystem'),

  validation: DS.attr('string', {defaultValue: ''}),
  errorContent: DS.attr('string', {defaultValue: ''}),
  errorTitle: DS.attr('string', {defaultValue: ''}),

  isSelected: function() {
    return this.get('operatingSystem.isSelected');
  }.property('id','operatingSystem.isSelected'),

  emptyError: function() {
    return !this.get('baseUrl');
  }.property('baseUrl'),

  invalidError: function() {
    return this.get('validation') == App.Repository.validation['INVALID'];
  }.property('validation'),

  undo: function() {
    return this.get('baseUrl') != this.get('latestBaseUrl');
  }.property('baseUrl','latestBaseUrl'),

  clearAll: function() {
    return this.get('baseUrl')
  }.property('baseUrl')
});

App.Repository.validation = {
  PENDING: '',
  INVALID: 'icon-exclamation-sign',
  OK: 'icon-ok',
  INPROGRESS: 'icon-repeat'
};


App.Repository.FIXTURES = [];
});

require.register("models/root_service", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * This model loads the rootService (Ambari)
 * The model maps to the  http://localhost:8080/api/v1/services/AMBARI?fields=components/RootServiceComponents
 * @type {*}
 */
App.RootService = DS.Model.extend({
  serviceName: DS.attr('string'),
  displayName: function() {
    return App.format.role(this.get('serviceName'));
  }.property('serviceName'),
  components: DS.hasMany('App.RootServiceComponents')
});

App.RootService.FIXTURES = [];

App.RootServiceComponents =   DS.Model.extend({
  componentName: DS.attr('string'),
  displayName: function() {
    return App.format.role(this.get('componentName'));
  }.property('componentName'),
  componentVersion: DS.attr('string'),
  serverClock: DS.attr('number'),
  serviceName: DS.attr('string'),
  properties: DS.attr('object')
});

App.RootServiceComponents.FIXTURES = [];


});

require.register("models/service", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');
require('utils/config');

App.Service = DS.Model.extend({
  serviceName: DS.attr('string'),
  displayName: function() {
    return App.format.role(this.get('serviceName'));
  }.property('serviceName'),
  passiveState: DS.attr('string'),
  workStatus: DS.attr('string'),
  rand: DS.attr('string'),
  toolTipContent: DS.attr('string'),
  quickLinks: DS.hasMany('App.QuickLinks'),  // mapped in app/mappers/service_metrics_mapper.js method - mapQuickLinks
  hostComponents: DS.hasMany('App.HostComponent'),
  serviceConfigsTemplate: App.config.get('preDefinedServiceConfigs'),
  /**
   * used by services("OOZIE", "ZOOKEEPER", "HIVE", "MAPREDUCE2", "TEZ", "SQOOP", "PIG","FALCON")
   * that have only client components
   */
  installedClients: DS.attr('number'),

  clientComponents: DS.hasMany('App.ClientComponent'),
  slaveComponents: DS.hasMany('App.SlaveComponent'),

  /**
   * @type {bool}
   */
  isInPassive: function() {
    return this.get('passiveState') === "ON";
  }.property('passiveState'),

  // Instead of making healthStatus a computed property that listens on hostComponents.@each.workStatus,
  // we are creating a separate observer _updateHealthStatus.  This is so that healthStatus is updated
  // only once after the run loop.  This is because Ember invokes the computed property every time
  // a property that it depends on changes.  For example, App.statusMapper's map function would invoke
  // the computed property too many times and freezes the UI without this hack.
  // See http://stackoverflow.com/questions/12467345/ember-js-collapsing-deferring-expensive-observers-or-computed-properties
  healthStatus: function(){
    switch(this.get('workStatus')){
      case 'STARTED':
        return 'green';
      case 'STARTING':
        return 'green-blinking';
      case 'INSTALLED':
        return 'red';
      case 'STOPPING':
        return 'red-blinking';
      case 'UNKNOWN':
      default:
        return 'yellow';
    }
  }.property('workStatus'),
  isStopped: function () {
    return this.get('workStatus') === 'INSTALLED';
  }.property('workStatus'),
  isStarted: function () {
    return this.get('workStatus') === 'STARTED';
  }.property('workStatus'),

  /**
   * Service Tagging by their type.
   * @type {String[]}
   **/
  serviceTypes: function() {
    var typeServiceMap = {
      GANGLIA: ['MONITORING'],
      HDFS: ['HA_MODE'],
      YARN: ['HA_MODE']
    };
    return typeServiceMap[this.get('serviceName')] || [];
  }.property('serviceName'),

  /**
   * For each host-component, if the desired_configs dont match the
   * actual_configs, then a restart is required.
   */
  isRestartRequired: function () {
    var rhc = App.HostComponent.find().filterProperty('service.serviceName', this.get('serviceName')).filterProperty('staleConfigs', true);
    var hc = {};
    rhc.forEach(function(_rhc) {
      var hostName = _rhc.get('hostName');
      if (!hc[hostName]) {
        hc[hostName] = [];
      }
      hc[hostName].push(_rhc.get('displayName'));
    });
    this.set('restartRequiredHostsAndComponents', hc);
    return (rhc.length>0);

  }.property('serviceName', 'hostComponents.@each.staleConfigs'),

  /**
   * Contains a map of which hosts and host_components
   * need a restart. This is populated when calculating
   * #isRestartRequired()
   * Example:
   * {
   *  'publicHostName1': ['TaskTracker'],
   *  'publicHostName2': ['JobTracker', 'TaskTracker']
   * }
   */
  restartRequiredHostsAndComponents: {},

  /**
   * Based on the information in #restartRequiredHostsAndComponents
   */
  restartRequiredMessage: function () {
    var restartHC = this.get('restartRequiredHostsAndComponents');
    var hostCount = 0;
    var hcCount = 0;
    var hostsMsg = "<ul>";
    for(var host in restartHC){
      hostCount++;
      hostsMsg += "<li>"+host+"</li><ul>";
      restartHC[host].forEach(function(c){
        hcCount++;
        hostsMsg += "<li>"+c+"</li>";
      });
      hostsMsg += "</ul>";
    }
    hostsMsg += "</ul>";
    return this.t('services.service.config.restartService.TooltipMessage').format(hcCount, hostCount, hostsMsg);
  }.property('restartRequiredHostsAndComponents'),

  hasCriticalAlerts: function () {
    var controller = App.router.get('mainAlertDefinitionsController');
    return controller.getCriticalAlertsCountForService(this.get('serviceName')) > 0;
  }.property('alertsCount'),

  alertsCount: function () {
    var controller = App.router.get('mainAlertDefinitionsController');
    return controller.getCriticalOrWarningAlertsCountForService(this.get('serviceName'));
  }.property('App.router.mainAlertDefinitionsController.content.@each.isCriticalOrWarning')
});

App.Service.Health = {
  live: "LIVE",
  dead: "DEAD-RED",
  starting: "STARTING",
  stopping: "STOPPING",
  unknown: "DEAD-YELLOW",

  getKeyName: function (value) {
    switch (value) {
      case this.live:
        return 'live';
      case this.dead:
        return 'dead';
      case this.starting:
        return 'starting';
      case this.stopping:
        return 'stopping';
      case this.unknown:
        return 'unknown';
    }
    return 'none';
  }
};

/**
 * association between service and extended model name
 * @type {Object}
 */
  App.Service.extendedModel = {
  'HDFS': 'HDFSService',
  'HBASE': 'HBaseService',
  'YARN': 'YARNService',
  'MAPREDUCE2': 'MapReduce2Service',
  'STORM': 'StormService',
  'FLUME': 'FlumeService'
};

App.Service.FIXTURES = [];

});

require.register("models/service/flume", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.FlumeService = App.Service.extend({
  version: DS.attr('string'),
  agents: DS.hasMany('App.FlumeAgent'),
  flumeHandlersTotal: DS.attr('number')
});

App.FlumeAgent = DS.Model.extend({
  /**
   * ID of a flume agent will be of the format
   * '<agent-name>-<host-name>'
   */
  id: DS.attr('string'),
  name: DS.attr('string'),
  /**
   * Status of agent. One of 'RUNNING', 'NOT_RUNNING', 'UNKNOWN'.
   */
  status: DS.attr('string'),
  host: DS.belongsTo('App.Host'),
  hostName: DS.attr('string'),

  channelsCount: DS.attr('number'),
  sourcesCount: DS.attr('number'),
  sinksCount: DS.attr('number'),

  healthClass : function() {
    switch (this.get('status')) {
    case 'RUNNING':
      return App.healthIconClassGreen;
      break;
    case 'NOT_RUNNING':
      return App.healthIconClassRed;
      break;
    case 'UNKNOWN':
    default:
      return App.healthIconClassYellow;
      break;
    }
  }.property('status'),

  displayStatus : function() {
    switch (this.get('status')) {
      case 'RUNNING':
        return "Running";
        break;
      case 'NOT_RUNNING':
        return "Stopped";
        break;
      case 'UNKNOWN':
      default:
        return "Unknown";
        break;
    }
  }.property('status')

});

App.FlumeAgent.FIXTURES = [];
App.FlumeService.FIXTURES = [];

});

require.register("models/service/hbase", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.HBaseService = App.Service.extend({
  version: DS.attr('string'),
  master: function () {
    return this.get('hostComponents').findProperty('componentName', 'HBASE_MASTER');
  }.property('hostComponents'),
  regionServersStarted: DS.attr('number'),
  regionServersInstalled: DS.attr('number'),
  regionServersTotal: DS.attr('number'),
  masterStartTime: DS.attr('number'),
  masterActiveTime: DS.attr('number'),
  averageLoad: DS.attr('number'),
  regionsInTransition: DS.attr('number'),
  revision: DS.attr('string'),
  heapMemoryUsed: DS.attr('number'),
  heapMemoryMax: DS.attr('number')
});

App.HBaseService.FIXTURES = [];

});

require.register("models/service/hdfs", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.HDFSService = App.Service.extend({
  version: DS.attr('string'),
  nameNode: function () {
    return this.get('hostComponents').findProperty('componentName', 'NAMENODE');
  }.property('hostComponents'),
  snameNode: function () {
    return this.get('hostComponents').findProperty('componentName', 'SECONDARY_NAMENODE');
  }.property('hostComponents'),
  activeNameNode: DS.belongsTo('App.HostComponent'),
  standbyNameNode: DS.belongsTo('App.HostComponent'),
  standbyNameNode2: DS.belongsTo('App.HostComponent'),
  dataNodesStarted: DS.attr('number'),
  dataNodesInstalled: DS.attr('number'),
  dataNodesTotal: DS.attr('number'),
  journalNodes: function () {
    return this.get('hostComponents').filterProperty('componentName', 'JOURNALNODE');
  }.property('hostComponents.@each'),
  nameNodeStartTime: DS.attr('number'),
  jvmMemoryHeapUsed: DS.attr('number'),
  jvmMemoryHeapMax: DS.attr('number'),
  decommissionDataNodes: DS.hasMany('App.HostComponent'),
  liveDataNodes: DS.hasMany('App.HostComponent'),
  deadDataNodes: DS.hasMany('App.HostComponent'),
  capacityUsed: DS.attr('number'),
  capacityTotal: DS.attr('number'),
  capacityRemaining: DS.attr('number'),
  dfsTotalBlocks: DS.attr('number'),
  dfsCorruptBlocks: DS.attr('number'),
  dfsMissingBlocks: DS.attr('number'),
  dfsUnderReplicatedBlocks: DS.attr('number'),
  dfsTotalFiles: DS.attr('number'),
  upgradeStatus: DS.attr('boolean'),
  safeModeStatus: DS.attr('string'),
  nameNodeRpc: DS.attr('number'),
  metricsNotAvailable: DS.attr('boolean')
});

App.HDFSService.FIXTURES = [];

});

require.register("models/service/mapreduce2", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.MapReduce2Service = App.Service.extend({
  version: DS.attr('string'),
  jobHistoryServer: function () {
    return this.get('hostComponents').findProperty('componentName', 'HISTORYSERVER');
  }.property('hostComponents'),
  mapReduce2Clients: DS.attr('number')
});

App.MapReduce2Service.FIXTURES = [];

});

require.register("models/service/storm", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.StormService = App.Service.extend({
  version: DS.attr('string'),
  superVisorsStarted: DS.attr('number'),
  superVisorsInstalled: DS.attr('number'),
  superVisorsTotal: DS.attr('number'),
  totalTasks: DS.attr('number'),
  totalSlots: DS.attr('number'),
  usedSlots: DS.attr('number'),
  freeSlots: DS.attr('number'),
  totalExecutors: DS.attr('number'),
  topologies: DS.attr('number'),
  nimbusUptime: DS.attr('string')
});

App.StormService.FIXTURES = [];

});

require.register("models/service/yarn", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
var objectUtils = require('utils/object_utils');

App.YARNService = App.Service.extend({
  version: DS.attr('string'),
  resourceManager: function() {
    return this.get('hostComponents').findProperty('componentName', 'RESOURCEMANAGER');
  }.property('hostComponents'),
  appTimelineServer: function() {
    return this.get('hostComponents').findProperty('componentName', 'APP_TIMELINE_SERVER');
  }.property('hostComponents'),
  nodeManagersStarted: DS.attr('number'),
  nodeManagersInstalled: DS.attr('number'),
  nodeManagersTotal: DS.attr('number'),
  nodeManagersCountActive: DS.attr('number'),
  nodeManagersCountLost: DS.attr('number'),
  nodeManagersCountUnhealthy: DS.attr('number'),
  nodeManagersCountRebooted: DS.attr('number'),
  nodeManagersCountDecommissioned: DS.attr('number'),
  containersAllocated: DS.attr('number'),
  containersPending: DS.attr('number'),
  containersReserved: DS.attr('number'),
  appsSubmitted: DS.attr('number'),
  appsRunning: DS.attr('number'),
  appsPending: DS.attr('number'),
  appsCompleted: DS.attr('number'),
  appsKilled: DS.attr('number'),
  appsFailed: DS.attr('number'),
  ahsWebPort: function() {
    var yarnConf = App.db.getConfigs().findProperty('type', 'yarn-site')
    if(yarnConf){
      return yarnConf.properties['yarn.timeline-service.webapp.address'].match(/:(\d+)/)[1];;
    }
    return "8188";
  }.property(),
  resourceManagerStartTime: DS.attr('number'),
  jvmMemoryHeapUsed: DS.attr('number'),
  jvmMemoryHeapMax: DS.attr('number'),
  allocatedMemory: DS.attr('number'),
  reservedMemory: DS.attr('number'),
  availableMemory: DS.attr('number'),
  queue: DS.attr('string'),
  queueFormatted: function() {
    var queue = JSON.parse(this.get('queue'));
    return objectUtils.recursiveTree(queue);
  }.property('queue'),
  queuesCount: function() {
    var queue = JSON.parse(this.get('queue'));
    return objectUtils.recursiveKeysCount(queue);
  }.property('queue'),
  allQueueNames: [],
  childQueueNames: [],

  maxMemory: function() {
    return this.get('allocatedMemory') + this.get('availableMemory');
  }.property('allocatedMemory','availableMemory'),

  /**
   * Provides a flat array of queue names.
   * Example: root, root/default
   */
  queueNames: function () {
    var queueString = this.get('queue');
    var allQueueNames = [];
    var childQueueNames = [];
    if (queueString != null) {
      var queues = JSON.parse(queueString);
      var addQueues = function (queuesObj, path){
        var names = [];
        for ( var subQueue in queuesObj) {
          if (queuesObj[subQueue] instanceof Object) {
            var qFN = path=='' ? subQueue : path+'/'+subQueue;
            names.push(qFN);
            var subNames = addQueues(queuesObj[subQueue], qFN);
            names = names.concat(subNames);
            if (!subNames || subNames.length < 1) {
              childQueueNames.push(qFN);
            }
          }
        }
        return names;
      };
      allQueueNames = addQueues(queues, '');
    }
    this.set('allQueueNames', allQueueNames);
    this.set('childQueueNames', childQueueNames);
  }.observes('queue')
  /**
   * ResourceManager's lost count is not accurate once RM is rebooted. Since
   * Ambari knows the total number of nodes and the counts of nodes in other
   * states, we calculate the lost count.
   *
  nodeManagersCountLost: function () {
    var totalCount = this.get('nodeManagersTotal');
    var activeCount = this.get('nodeManagersCountActive');
    var rebootedCount = this.get('nodeManagersCountRebooted');
    var unhealthyCount = this.get('nodeManagersCountUnhealthy');
    var decomCount = this.get('nodeManagersCountDecommissioned');
    var nonLostHostsCount = activeCount + rebootedCount + decomCount + unhealthyCount;
    return totalCount >= nonLostHostsCount ? totalCount - nonLostHostsCount : 0;
  }.property('nodeManagersTotal', 'nodeManagersCountActive', 'nodeManagersCountRebooted', 'nodeManagersCountUnhealthy', 'nodeManagersCountDecommissioned')*/
});

App.YARNService.FIXTURES = [];

});

require.register("models/service_audit", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ServiceAudit = DS.Model.extend({
  date: DS.attr('date'),
  service: DS.belongsTo('App.Service'),
  operationName: DS.attr('string'),
  user: DS.belongsTo('App.User')
});

App.ServiceAudit.FIXTURES = [
  {
    id: 1,
    date: 'September 12, 2012 17:00',
    operation_name: 'Reconfigure',
    user_id: 2,
    service_id: 1
  },
  {
    id: 2,
    date: 'September 13, 2012 17:00',
    operation_name: 'Start',
    user_id: 1,
    service_id: 1
  },
  {
    id: 3,
    date: 'September 14, 2012 17:00',
    operation_name: 'Install',
    user_id: 1,
    service_id: 1
  },
  {
    id: 4,
    date: 'September 12, 2012 17:00',
    operation_name: 'Reconfigure',
    user_id: 2,
    service_id: 2
  },
  {
    id: 5,
    date: 'September 13, 2012 17:00',
    operation_name: 'Start',
    user_id: 1,
    service_id: 2
  },
  {
    id: 6,
    date: 'September 14, 2012 17:00',
    operation_name: 'Install',
    user_id: 1,
    service_id: 2
  }
];
});

require.register("models/service_config", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var validator = require('utils/validator');

App.ServiceConfig = Ember.Object.extend({
  serviceName: '',
  configCategories: [],
  configs: null,
  restartRequired: false,
  restartRequiredMessage: '',
  restartRequiredHostsAndComponents: {},
  configGroups: [],
  initConfigsLength: 0, // configs length after initialization in order to watch changes
  errorCount: function () {
    var overrideErrors = 0;
    this.get('configs').filterProperty("overrides").forEach(function (e) {
      e.overrides.forEach(function (e) {
        if (e.error) {
          overrideErrors += 1;
        }
      })
    });
    var categoryNames = this.get('configCategories').mapProperty('name');
    var masterErrors = this.get('configs').filter(function (item) {
      return categoryNames.contains(item.get('category'));
    }).filterProperty('isValid', false).filterProperty('isVisible', true).get('length');
    var slaveErrors = 0;
    this.get('configCategories').forEach(function (_category) {
      slaveErrors += _category.get('slaveErrorCount');
    }, this);
    return masterErrors + slaveErrors + overrideErrors;
  }.property('configs.@each.isValid', 'configs.@each.isVisible', 'configCategories.@each.slaveErrorCount', 'configs.@each.overrideErrorTrigger'),

  isPropertiesChanged: function() {
    return this.get('configs').someProperty('isNotDefaultValue') ||
           this.get('configs').someProperty('isOverrideChanged') ||
           this.get('configs.length') !== this.get('initConfigsLength') ||
           (this.get('configs.length') === this.get('initConfigsLength') && this.get('configs').someProperty('defaultValue', null));
  }.property('configs.@each.isNotDefaultValue', 'configs.@each.isOverrideChanged', 'configs.length')
});

App.ServiceConfigCategory = Ember.Object.extend({
  name: null,
  /**
   *  We cant have spaces in the name as this is being used as HTML element id while rendering. Hence we introduced 'displayName' where we can have spaces like 'Secondary Name Node' etc.
   */
  displayName: null,
  slaveConfigs: null,
  /**
   * check whether to show custom view in category instead of default
   */
  isCustomView: false,
  customView: null,
  /**
   * Each category might have a site-name associated (hdfs-site, core-site, etc.)
   * and this will be used when determining which category a particular property
   * ends up in, based on its site.
   */
  siteFileName: null,
  /**
   * Can this category add new properties. Used for custom configurations.
   */
  canAddProperty: false,
  primaryName: function () {
    switch (this.get('name')) {
      case 'DataNode':
        return 'DATANODE';
        break;
      case 'TaskTracker':
        return 'TASKTRACKER';
        break;
      case 'RegionServer':
        return 'HBASE_REGIONSERVER';
    }
    return null;
  }.property('name'),


  isForMasterComponent: function () {
    var masterServices = [ 'NameNode', 'SNameNode', 'JobTracker', 'HBase Master', 'Oozie Master',
      'Hive Metastore', 'WebHCat Server', 'ZooKeeper Server', 'Ganglia' ];

    return (masterServices.contains(this.get('name')));
  }.property('name'),

  isForSlaveComponent: function () {
    var slaveComponents = ['DataNode', 'TaskTracker', 'RegionServer'];
    return (slaveComponents.contains(this.get('name')));
  }.property('name'),

  slaveErrorCount: function () {
    var length = 0;
    if (this.get('slaveConfigs.groups')) {
      this.get('slaveConfigs.groups').forEach(function (_group) {
        length += _group.get('errorCount');
      }, this);
    }
    return length;
  }.property('slaveConfigs.groups.@each.errorCount'),

  isAdvanced : function(){
    var name = this.get('name');
    return name.indexOf("Advanced") !== -1 ;
  }.property('name')
});


App.SlaveConfigs = Ember.Object.extend({
  componentName: null,
  displayName: null,
  hosts: null,
  groups: null
});

App.Group = Ember.Object.extend({
  name: null,
  hostNames: null,
  properties: null,
  errorCount: function () {
    if (this.get('properties')) {
      return this.get('properties').filterProperty('isValid', false).filterProperty('isVisible', true).get('length');
    }
  }.property('properties.@each.isValid', 'properties.@each.isVisible')
});


App.ServiceConfigProperty = Em.Object.extend({

  id: '', //either 'puppet var' or 'site property'
  name: '',
  displayName: '',
  value: '',
  retypedPassword: '',
  defaultValue: '',
  defaultDirectory: '',
  description: '',
  displayType: 'string', // string, digits, number, directories, custom
  unit: '',
  category: 'General',
  isRequired: true, // by default a config property is required
  isReconfigurable: true, // by default a config property is reconfigurable
  isEditable: true, // by default a config property is editable
  isNotEditable: Ember.computed.not('isEditable'),
  isFinal: false,
  hideFinalIcon: function () {
    return (!this.get('isFinal'))&& this.get('isNotEditable');
  }.property('isFinal', 'isNotEditable'),
  defaultIsFinal: false,
  supportsFinal: false,
  isVisible: true,
  isMock: false, // mock config created created only to displaying
  isRequiredByAgent: true, // Setting it to true implies property will be stored in configuration
  isSecureConfig: false,
  errorMessage: '',
  warnMessage: '',
  serviceConfig: null, // points to the parent App.ServiceConfig object
  filename: '',
  isOriginalSCP : true, // if true, then this is original SCP instance and its value is not overridden value.
  parentSCP: null, // This is the main SCP which is overridden by this. Set only when isOriginalSCP is false.
  selectedHostOptions : null, // contain array of hosts configured with overridden value
  overrides : null,
  overrideValues: [],
  group: null, // Contain group related to this property. Set only when isOriginalSCP is false.
  isUserProperty: null, // This property was added by user. Hence they get removal actions etc.
  isOverridable: true,
  compareConfigs: [],
  isComparison: false,
  hasCompareDiffs: false,
  showLabel: true,
  error: false,
  warn: false,
  overrideErrorTrigger: 0, //Trigger for overrridable property error
  isRestartRequired: false,
  restartRequiredMessage: 'Restart required',
  index: null, //sequence number in category
  editDone: false, //Text field: on focusOut: true, on focusIn: false
  serviceValidator: null,
  isNotSaved: false, // user property was added but not saved
  hasInitialValue: false, //if true then property value is defined and saved to server
  isHiddenByFilter: false, //if true then hide this property (filtered out)
  rowStyleClass: null, // CSS-Class to be applied on the row showing this config
  /**
   * Usage example see on <code>App.ServiceConfigRadioButtons.handleDBConnectionProperty()</code>
   *
   * @property {Ember.View} additionalView - custom view related to property
   **/
  additionalView: null,

  /**
   * On Overridable property error message, change overrideErrorTrigger value to recount number of errors service have
   */
  observeErrors: function () {
    this.set("overrideErrorTrigger", this.get("overrideErrorTrigger") + 1);
  }.observes("overrides.@each.errorMessage"),
  /**
   * No override capabilities for fields which are not edtiable
   * and fields which represent master hosts.
   */
  isPropertyOverridable: function () {
    var overrideable = this.get('isOverridable');
    var editable = this.get('isEditable');
    var overrides = this.get('overrides');
    var dt = this.get('displayType');
    return overrideable && (editable || !overrides || !overrides.length) && ("masterHost" != dt);
  }.property('isEditable', 'displayType', 'isOverridable', 'overrides.length'),
  isOverridden: function() {
    var overrides = this.get('overrides');
    return (overrides != null && overrides.get('length')>0) || !this.get('isOriginalSCP');
  }.property('overrides', 'overrides.length', 'isOriginalSCP'),
  isOverrideChanged: function () {
    if (Em.isNone(this.get('overrides')) && this.get('overrideValues.length') === 0) return false;
    return JSON.stringify(this.get('overrides').mapProperty('isFinal')) !== JSON.stringify(this.get('overrideIsFinalValues'))
      || JSON.stringify(this.get('overrides').mapProperty('value')) !== JSON.stringify(this.get('overrideValues'));
  }.property('isOverridden', 'overrides.@each.isNotDefaultValue'),
  isRemovable: function() {
    var isOriginalSCP = this.get('isOriginalSCP');
    var isUserProperty = this.get('isUserProperty');
    var isEditable = this.get('isEditable');
    var hasOverrides = this.get('overrides.length') > 0;
    // Removable when this is a user property, or it is not an original property and it is editable
    return isEditable && !hasOverrides && (isUserProperty || !isOriginalSCP);
  }.property('isUserProperty', 'isOriginalSCP', 'overrides.length'),
  init: function () {
    if(this.get("displayType")=="password"){
      this.set('retypedPassword', this.get('value'));
    }
    if ((this.get('id') === 'puppet var') && this.get('value') == '') {
      this.set('value', this.get('defaultValue'));
    }
    // TODO: remove mock data
  },

  /**
   * Indicates when value is not the default value.
   * Returns false when there is no default value.
   */
  isNotDefaultValue: function () {
    var value = this.get('value');
    var defaultValue = this.get('defaultValue');
    var supportsFinal = this.get('supportsFinal');
    var isFinal = this.get('isFinal');
    var defaultIsFinal = this.get('defaultIsFinal');
    var isEditable = this.get('isEditable');
    return isEditable && ((defaultValue != null && value !== defaultValue) || (supportsFinal && isFinal !== defaultIsFinal));
  }.property('value', 'defaultValue', 'isEditable', 'isFinal', 'defaultIsFinal'),

  /**
   * Don't show "Undo" for hosts on Installer Step7
   */
  cantBeUndone: function() {
    var types = ["masterHost", "slaveHosts", "masterHosts", "slaveHost","radio button"];
    var displayType = this.get('displayType');
    var result = false;
    types.forEach(function(type) {
      if (type === displayType) {
        result = true;
      }
    });
    return result;
  }.property('displayType'),

  initialValue: function (localDB, dependencies) {
    var masterComponentHostsInDB = localDB.masterComponentHosts;
    //console.log("value in initialvalue: " + JSON.stringify(masterComponentHostsInDB));
    var hostsInfo = localDB.hosts; // which we are setting in installerController in step3.
    var slaveComponentHostsInDB = localDB.slaveComponentHosts;
    var isOnlyFirstOneNeeded = true;
    var hostWithPort = "([\\w|\\.]*)(?=:)";
    var hostWithPrefix = ":\/\/" + hostWithPort;
    switch (this.get('name')) {
      case 'namenode_host':
        this.set('value', masterComponentHostsInDB.filterProperty('component', 'NAMENODE').mapProperty('hostName'));
        break;
      case 'dfs.http.address':
        var nnHost =  masterComponentHostsInDB.findProperty('component', 'NAMENODE').hostName;
        this.setDefaultValue(hostWithPort,nnHost);
        break;
      case 'dfs.namenode.http-address':
        var nnHost =  masterComponentHostsInDB.findProperty('component', 'NAMENODE').hostName;
        this.setDefaultValue(hostWithPort,nnHost);
        break;
      case 'dfs.https.address':
        var nnHost =  masterComponentHostsInDB.findProperty('component', 'NAMENODE').hostName;
        this.setDefaultValue(hostWithPort,nnHost);
        break;
      case 'dfs.namenode.https-address':
        var nnHost =  masterComponentHostsInDB.findProperty('component', 'NAMENODE').hostName;
        this.setDefaultValue(hostWithPort,nnHost);
        break;
      case 'fs.default.name':
        var nnHost = masterComponentHostsInDB.filterProperty('component', 'NAMENODE').mapProperty('hostName');
        this.setDefaultValue(hostWithPrefix,'://' + nnHost);
        break;
      case 'fs.defaultFS':
        var nnHost = masterComponentHostsInDB.filterProperty('component', 'NAMENODE').mapProperty('hostName');
        this.setDefaultValue(hostWithPrefix,'://' + nnHost);
        break;
      case 'hbase.rootdir':
        var nnHost = masterComponentHostsInDB.filterProperty('component', 'NAMENODE').mapProperty('hostName');
        this.setDefaultValue(hostWithPrefix,'://' + nnHost);
        break;
      case 'snamenode_host':
        // Secondary NameNode does not exist when NameNode HA is enabled
        var snn = masterComponentHostsInDB.findProperty('component', 'SECONDARY_NAMENODE');
        if (snn) {
          this.set('value', snn.hostName);
        }
        break;
      case 'dfs.secondary.http.address':
        var snnHost = masterComponentHostsInDB.findProperty('component', 'SECONDARY_NAMENODE');
        if (snnHost) {
          this.setDefaultValue(hostWithPort,snnHost.hostName);
        }
        break;
      case 'dfs.namenode.secondary.http-address':
        var snnHost = masterComponentHostsInDB.findProperty('component', 'SECONDARY_NAMENODE');
        if (snnHost) {
          this.setDefaultValue(hostWithPort,snnHost.hostName);
        }
        break;
      case 'datanode_hosts':
        this.set('value', slaveComponentHostsInDB.findProperty('componentName', 'DATANODE').hosts.mapProperty('hostName'));
        break;
      case 'hs_host':
        this.set('value', masterComponentHostsInDB.filterProperty('component', 'HISTORYSERVER').mapProperty('hostName'));
        break;
      case 'yarn.log.server.url':
        var hsHost = masterComponentHostsInDB.filterProperty('component', 'HISTORYSERVER').mapProperty('hostName');
        this.setDefaultValue(hostWithPrefix,'://' + hsHost);
        break;
      case 'mapreduce.jobhistory.webapp.address':
        var hsHost = masterComponentHostsInDB.filterProperty('component', 'HISTORYSERVER').mapProperty('hostName');
        this.setDefaultValue(hostWithPort,hsHost);
        break;
      case 'mapreduce.jobhistory.address':
        var hsHost = masterComponentHostsInDB.filterProperty('component', 'HISTORYSERVER').mapProperty('hostName');
        this.setDefaultValue(hostWithPort,hsHost);
        break;
      case 'rm_host':
        this.set('value', masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName);
        break;
      case 'ats_host':
        var atsHost =  masterComponentHostsInDB.findProperty('component', 'APP_TIMELINE_SERVER');
        if (atsHost)
          this.set('value', atsHost.hostName);
        else
          this.set('value', 'false');
        break;
      case 'yarn.resourcemanager.hostname':
        var rmHost = masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName;
        this.set('defaultValue',rmHost);
        this.set('value',this.get('defaultValue'));
        break;
      case 'yarn.resourcemanager.resource-tracker.address':
        var rmHost = masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName;
        this.setDefaultValue(hostWithPort,rmHost);
        break;
      case 'yarn.resourcemanager.webapp.https.address':
        var rmHost = masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName;
        this.setDefaultValue(hostWithPort,rmHost);
        break;
      case 'yarn.resourcemanager.webapp.address':
        var rmHost = masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName;
        this.setDefaultValue(hostWithPort,rmHost);
        break;
      case 'yarn.resourcemanager.scheduler.address':
        var rmHost = masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName;
        this.setDefaultValue(hostWithPort,rmHost);
        break;
      case 'yarn.resourcemanager.address':
        var rmHost = masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName;
        this.setDefaultValue(hostWithPort,rmHost);
        break;
      case 'yarn.resourcemanager.admin.address':
        var rmHost = masterComponentHostsInDB.findProperty('component', 'RESOURCEMANAGER').hostName;
        this.setDefaultValue(hostWithPort,rmHost);
        break;
      case 'yarn.timeline-service.webapp.address':
        var atsHost =  masterComponentHostsInDB.findProperty('component', 'APP_TIMELINE_SERVER');
        if (atsHost && atsHost.hostName) {
          this.setDefaultValue(hostWithPort,atsHost.hostName);
        }
        break;
      case 'yarn.timeline-service.webapp.https.address':
        var atsHost =  masterComponentHostsInDB.findProperty('component', 'APP_TIMELINE_SERVER');
        if (atsHost && atsHost.hostName) {
          this.setDefaultValue(hostWithPort,atsHost.hostName);
        }
        break;
      case 'yarn.timeline-service.address':
        var atsHost =  masterComponentHostsInDB.findProperty('component', 'APP_TIMELINE_SERVER');
        if (atsHost && atsHost.hostName) {
          this.setDefaultValue(hostWithPort,atsHost.hostName);
        }
        break;
      case 'nm_hosts':
        this.set('value', slaveComponentHostsInDB.findProperty('componentName', 'NODEMANAGER').hosts.mapProperty('hostName'));
        break;
      case 'jobtracker_host':
        this.set('value', masterComponentHostsInDB.findProperty('component', 'JOBTRACKER').hostName);
        break;
      case 'mapred.job.tracker':
        var jtHost = masterComponentHostsInDB.findProperty('component', 'JOBTRACKER').hostName;
        this.setDefaultValue(hostWithPort,jtHost);
        break;
      case 'mapred.job.tracker.http.address':
        var jtHost = masterComponentHostsInDB.findProperty('component', 'JOBTRACKER').hostName;
        this.setDefaultValue(hostWithPort,jtHost);
        break;
      case 'mapreduce.history.server.http.address':
        var jtHost = masterComponentHostsInDB.findProperty('component', 'HISTORYSERVER').hostName;
        this.setDefaultValue(hostWithPort,jtHost);
        break;
      case 'tasktracker_hosts':
        this.set('value', slaveComponentHostsInDB.findProperty('componentName', 'TASKTRACKER').hosts.mapProperty('hostName'));
        break;
      case 'hbasemaster_host':
        this.set('value', masterComponentHostsInDB.filterProperty('component', 'HBASE_MASTER').mapProperty('hostName'));
        break;
      case 'regionserver_hosts':
        this.set('value', slaveComponentHostsInDB.findProperty('componentName', 'HBASE_REGIONSERVER').hosts.mapProperty('hostName'));
        break;
      case 'hivemetastore_host':
        this.set('value', masterComponentHostsInDB.filterProperty('component', 'HIVE_METASTORE').mapProperty('hostName'));
        break;
      case 'hive_ambari_host':
        this.set('value', masterComponentHostsInDB.findProperty('component', 'HIVE_SERVER').hostName);
        break;
      case 'hive_master_hosts':
        var hostNames = masterComponentHostsInDB.filter(function (masterComponent) {
          return ['HIVE_METASTORE', 'HIVE_SERVER'].contains(masterComponent.component);
        });
        this.set('value', hostNames.mapProperty('hostName').uniq().join(','));
        break;
      case 'hive_database':
        var newMySQLDBOption = this.get('options').findProperty('displayName', 'New MySQL Database');
        if (newMySQLDBOption) {
          var isNewMySQLDBOptionHidden = !App.get('supports.alwaysEnableManagedMySQLForHive') && App.get('router.currentState.name') != 'configs' &&
            !App.get('isManagedMySQLForHiveEnabled');
          if (isNewMySQLDBOptionHidden && this.get('value') == 'New MySQL Database') {
            this.set('value', 'Existing MySQL Database');
          }
          Em.set(newMySQLDBOption, 'hidden', isNewMySQLDBOptionHidden);
        }
        break;
      case 'oozieserver_host':
        this.set('value', masterComponentHostsInDB.findProperty('component', 'OOZIE_SERVER').hostName);
        break;
      case 'oozie.base.url':
        var oozieHost = masterComponentHostsInDB.findProperty('component', 'OOZIE_SERVER').hostName;
        this.setDefaultValue(hostWithPrefix,'://' + oozieHost);
        break;
      case 'webhcatserver_host':
        this.set('value', masterComponentHostsInDB.findProperty('component', 'WEBHCAT_SERVER').hostName);
        break;
      case 'oozie_ambari_host':
        this.set('value', masterComponentHostsInDB.findProperty('component', 'OOZIE_SERVER').hostName);
        break;
      case 'hadoop_host':
        this.set('value', masterComponentHostsInDB.filterProperty('component', 'NAMENODE').mapProperty('hostName'));
        break;
      case 'hive_existing_mysql_host':
      case 'hive_existing_postgresql_host':
      case 'hive_existing_oracle_host':
      case 'hive_existing_mssql_server_host':
      case 'hive_existing_mssql_server_2_host':
        var hiveServerHost = masterComponentHostsInDB.findProperty('component', 'HIVE_SERVER').hostName;
        this.set('value', hiveServerHost).set('defaultValue', hiveServerHost);
        break;
      case 'hive.metastore.uris':
        var hiveMSUris = this.getHiveMetastoreUris(masterComponentHostsInDB, dependencies['hive.metastore.uris']);
        if (hiveMSUris) {
          this.setDefaultValue("(.*)", hiveMSUris);
        }
        break;
      case 'oozie_existing_mysql_host':
      case 'oozie_existing_postgresql_host':
      case 'oozie_existing_oracle_host':
      case 'oozie_existing_mssql_server_host':
      case 'oozie_existing_mssql_server_2_host':
        var oozieServerHost = masterComponentHostsInDB.findProperty('component', 'OOZIE_SERVER').hostName;
        this.set('value', oozieServerHost).set('defaultValue', oozieServerHost);
        break;
      case 'storm.zookeeper.servers':
      case 'zookeeperserver_hosts':
        this.set('value', masterComponentHostsInDB.filterProperty('component', 'ZOOKEEPER_SERVER').mapProperty('hostName'));
        break;
      case 'nimbus.host':
        this.set('value', masterComponentHostsInDB.findProperty('component', 'NIMBUS').hostName);
        break;
      case 'falconserver_host':
        this.set('value', masterComponentHostsInDB.findProperty('component', 'FALCON_SERVER').hostName);
        break;
      case 'drpcserver_host':
        var drpcHost = masterComponentHostsInDB.findProperty('component', 'DRPC_SERVER');
        if (drpcHost) {
          this.set('value', drpcHost.hostName);
        }
        break;
      case 'stormuiserver_host':
        this.set('value', masterComponentHostsInDB.findProperty('component', 'STORM_UI_SERVER').hostName);
        break;
      case 'storm_rest_api_host':
        var stormRresApiHost = masterComponentHostsInDB.findProperty('component', 'STORM_REST_API');
        if(stormRresApiHost) {
          this.set('value', stormRresApiHost.hostName);
        }
        break;
      case 'supervisor_hosts':
        this.set('value', slaveComponentHostsInDB.findProperty('componentName', 'SUPERVISOR').hosts.mapProperty('hostName'));
        break;
      case 'knox_gateway_host':
        this.set('value', masterComponentHostsInDB.findProperty('component', 'KNOX_GATEWAY').hostName);
        break;
      case 'kafka_broker_hosts':
        this.set('value', masterComponentHostsInDB.filterProperty('component', 'KAFKA_BROKER').mapProperty('hostName'));
        break;
      case 'kafka.ganglia.metrics.host':
        var gangliaHost =  masterComponentHostsInDB.findProperty('component', 'GANGLIA_SERVER');
        if (gangliaHost) {
          this.set('value', gangliaHost.hostName);
        }
        break;
      case 'hbase.zookeeper.quorum':
        if (this.get('filename') == 'hbase-site.xml') {
          var zkHosts = masterComponentHostsInDB.filterProperty('component', 'ZOOKEEPER_SERVER').mapProperty('hostName');
          this.setDefaultValue("(\\w*)", zkHosts);
        }
        break;
      case 'yarn.resourcemanager.zk-address':
        var value = masterComponentHostsInDB.findProperty('component', 'ZOOKEEPER_SERVER').hostName + ':' + dependencies.clientPort;
        this.setProperties({
          value: value,
          defaultValue: value
        });
        break;
      case 'zookeeper.connect':
      case 'hive.zookeeper.quorum':
      case 'templeton.zookeeper.hosts':
      case 'hadoop.registry.zk.quorum':
      case 'hive.cluster.delegation.token.store.zookeeper.connectString':
        var zkHosts = masterComponentHostsInDB.filterProperty('component', 'ZOOKEEPER_SERVER').mapProperty('hostName');
        var zkHostPort = zkHosts;
        var regex = "\\w*:(\\d+)";   //regex to fetch the port
        var portValue = this.get('defaultValue').match(new RegExp(regex));
        if (!portValue) return;
        if (portValue[1]) {
          for ( var i = 0; i < zkHosts.length; i++ ) {
            zkHostPort[i] = zkHosts[i] + ":" + portValue[1];
          }
        }
        this.setDefaultValue("(.*)", zkHostPort);
        break;
      case 'templeton.hive.properties':
        var hiveMSUris = this.getHiveMetastoreUris(masterComponentHostsInDB, dependencies['hive.metastore.uris']).replace(',', '\\,');
        if (/\/\/localhost:/g.test(this.get('value'))) {
          this.set('defaultValue', this.get('value') + ',hive.metastore.execute.setugi=true');
        }
        this.setDefaultValue("(hive\\.metastore\\.uris=)([^\\,]+)", "$1" + hiveMSUris);
        break;
      case 'dfs.name.dir':
      case 'dfs.namenode.name.dir':
      case 'dfs.data.dir':
      case 'dfs.datanode.data.dir':
      case 'yarn.nodemanager.local-dirs':
      case 'yarn.nodemanager.log-dirs':
      case 'mapred.local.dir':
      case 'log.dirs':  // for Kafka Broker
        this.unionAllMountPoints(!isOnlyFirstOneNeeded, localDB);
        break;
      case 'hbase.tmp.dir':
        if (this.get('filename') == 'hbase-site.xml') {
          this.unionAllMountPoints(isOnlyFirstOneNeeded, localDB);
        }
        break;
      case 'fs.checkpoint.dir':
      case 'dfs.namenode.checkpoint.dir':
      case 'yarn.timeline-service.leveldb-timeline-store.path':
      case 'dataDir':
      case 'oozie_data_dir':
      case 'storm.local.dir':
      case '*.falcon.graph.storage.directory':
      case '*.falcon.graph.serialize.path':
        this.unionAllMountPoints(isOnlyFirstOneNeeded, localDB);
        break;
      case '*.broker.url':
        var falconServerHost = masterComponentHostsInDB.findProperty('component', 'FALCON_SERVER').hostName;
        this.setDefaultValue('localhost', falconServerHost);
        break;
      case 'RANGER_HOST':
        var rangerAdminHost = masterComponentHostsInDB.findProperty('component', 'RANGER_ADMIN');
        if(rangerAdminHost) {
          this.set('value', rangerAdminHost.hostName);
        } else {
          this.set('isVisible', 'false');
          this.set('isRequired', 'false');
        }
        break;
    }
  },

  /**
   * Get hive.metastore.uris initial value
   * @param hosts
   * @param defaultValue
   * @returns {string}
   */
  getHiveMetastoreUris: function (hosts, defaultValue) {
    var hiveMSHosts = hosts.filterProperty('component', 'HIVE_METASTORE').mapProperty('hostName'),
      hiveMSUris = hiveMSHosts,
      regex = "\\w*:(\\d+)",
      portValue = defaultValue && defaultValue.match(new RegExp(regex));

    if (!portValue) return '';
    if (portValue[1]) {
      for (var i = 0; i < hiveMSHosts.length; i++) {
        hiveMSUris[i] = "thrift://" + hiveMSHosts[i] + ":" + portValue[1];
      }
    }
    return hiveMSUris.join(',');
  },

  /**
   * @param regex : String
   * @param replaceWith : String
   */
  setDefaultValue: function(regex,replaceWith) {
    var defaultValue = this.get('defaultValue');
    var re = new RegExp(regex);
    defaultValue = defaultValue.replace(re,replaceWith);
    this.set('defaultValue',defaultValue);
    this.set('value',this.get('defaultValue'));
  },

  unionAllMountPoints: function (isOnlyFirstOneNeeded, localDB) {
    var hostname = '';
    var mountPointsPerHost = [];
    var mountPointAsRoot;
    var masterComponentHostsInDB = localDB.masterComponentHosts;
    var slaveComponentHostsInDB = localDB.slaveComponentHosts;
    var hostsInfo = localDB.hosts; // which we are setting in installerController in step3.
    //all hosts should be in local storage without using App.Host model
    App.Host.find().forEach(function(item){
      if(!hostsInfo[item.get('id')]){
        hostsInfo[item.get('id')] = {
          name: item.get('id'),
          cpu: item.get('cpu'),
          memory: item.get('memory'),
          disk_info: item.get('diskInfo'),
          bootStatus: "REGISTERED",
          isInstalled: true
        };
      }
    });
    var temp = '';
    var setOfHostNames = [];
    var components = [];
    switch (this.get('name')) {
      case 'dfs.namenode.name.dir':
      case 'dfs.name.dir':
        components = masterComponentHostsInDB.filterProperty('component', 'NAMENODE');
        components.forEach(function (component) {
          setOfHostNames.push(component.hostName);
        }, this);
        break;
      case 'fs.checkpoint.dir':
      case 'dfs.namenode.checkpoint.dir':
        components = masterComponentHostsInDB.filterProperty('component', 'SECONDARY_NAMENODE');
        components.forEach(function (component) {
          setOfHostNames.push(component.hostName);
        }, this);
        break;
      case 'dfs.data.dir':
      case 'dfs.datanode.data.dir':
        temp = slaveComponentHostsInDB.findProperty('componentName', 'DATANODE');
        temp.hosts.forEach(function (host) {
          setOfHostNames.push(host.hostName);
        }, this);
        break;
      case 'mapred.local.dir':
        temp = slaveComponentHostsInDB.findProperty('componentName', 'TASKTRACKER') || slaveComponentHostsInDB.findProperty('componentName', 'NODEMANAGER');
        temp.hosts.forEach(function (host) {
          setOfHostNames.push(host.hostName);
        }, this);
        break;
      case 'yarn.nodemanager.log-dirs':
      case 'yarn.nodemanager.local-dirs':
        temp = slaveComponentHostsInDB.findProperty('componentName', 'NODEMANAGER');
        temp.hosts.forEach(function (host) {
          setOfHostNames.push(host.hostName);
        }, this);
        break;
      case 'yarn.timeline-service.leveldb-timeline-store.path':
        components = masterComponentHostsInDB.filterProperty('component', 'APP_TIMELINE_SERVER');
        components.forEach(function (component) {
          setOfHostNames.push(component.hostName);
        }, this);
        break;
      case 'dataDir':
        components = masterComponentHostsInDB.filterProperty('component', 'ZOOKEEPER_SERVER');
        components.forEach(function (component) {
          setOfHostNames.push(component.hostName);
        }, this);
        break;
      case 'oozie_data_dir':
        components = masterComponentHostsInDB.filterProperty('component', 'OOZIE_SERVER');
        components.forEach(function (component) {
          setOfHostNames.push(component.hostName);
        }, this);
        break;
      case 'hbase.tmp.dir':
        temp = slaveComponentHostsInDB.findProperty('componentName', 'HBASE_REGIONSERVER');
        temp.hosts.forEach(function (host) {
          setOfHostNames.push(host.hostName);
        }, this);
        break;
      case 'storm.local.dir':
        temp = slaveComponentHostsInDB.findProperty('componentName', 'SUPERVISOR');
        temp.hosts.forEach(function (host) {
          setOfHostNames.push(host.hostName);
        }, this);
        components = masterComponentHostsInDB.filterProperty('component', 'NIMBUS');
        components.forEach(function (component) {
          setOfHostNames.push(component.hostName);
        }, this);
        break;
      case '*.falcon.graph.storage.directory':
      case '*.falcon.graph.serialize.path':
        components = masterComponentHostsInDB.filterProperty('component', 'FALCON_SERVER');
        components.forEach(function (component) {
          setOfHostNames.push(component.hostName);
        }, this);
        break;
      case 'log.dirs':
        components = masterComponentHostsInDB.filterProperty('component', 'KAFKA_BROKER');
        components.forEach(function (component) {
          setOfHostNames.push(component.hostName);
        }, this);
        break;
    }

    // In Add Host Wizard, if we did not select this slave component for any host, then we don't process any further.
    if (setOfHostNames.length === 0) {
      return;
    }

    var allMountPoints = [];
    for (var i = 0; i < setOfHostNames.length; i++) {
      hostname = setOfHostNames[i];

      mountPointsPerHost = hostsInfo[hostname].disk_info;

      mountPointAsRoot = mountPointsPerHost.findProperty('mountpoint', '/');

      // If Server does not send any host details information then atleast one mountpoint should be presumed as root
      // This happens in a single container Linux Docker environment.
      if (!mountPointAsRoot) {
        mountPointAsRoot = {mountpoint: '/'};
      }

      mountPointsPerHost = mountPointsPerHost.filter(function (mPoint) {
        return !(['/', '/home', '/boot'].contains(mPoint.mountpoint)
        || ['devtmpfs', 'tmpfs', 'vboxsf'].contains(mPoint.type)
        || mPoint.available == 0
        || mPoint.type == 'CDFS');
      });

      mountPointsPerHost.forEach(function (mPoint) {
        if( !allMountPoints.findProperty("mountpoint", mPoint.mountpoint)) {
          allMountPoints.push(mPoint);
        }
      }, this);
    }
    if (allMountPoints.length == 0) {
      allMountPoints.push(mountPointAsRoot);
    }
    this.set('value', '');
    var winRegex = /^([a-z]):\\?$/;
    if (!isOnlyFirstOneNeeded) {
      allMountPoints.forEach(function (eachDrive) {
        var mPoint = this.get('value');
        if (!mPoint) {
          mPoint = "";
        }
        if (eachDrive.mountpoint === "/") {
          mPoint += this.get('defaultDirectory') + "\n";
        } else if(winRegex.test(eachDrive.mountpoint.toLowerCase())) {
          switch (this.get('name')) {
            case 'dfs.datanode.data.dir':
            case 'dfs.name.dir':
            case 'dfs.namenode.name.dir':
            case 'dfs.data.dir':
            case 'dfs.datanode.data.dir':
              var winDriveUrl = eachDrive.mountpoint.toLowerCase().replace(winRegex, "file:///$1:");
              mPoint += winDriveUrl + this.get('defaultDirectory') + "\n";
              break;
            default:
              var winDrive = eachDrive.mountpoint.toLowerCase().replace(winRegex, "$1:");
              var winDir = this.get('defaultDirectory').replace(/\//g, "\\");
              mPoint += winDrive + winDir + "\n";
          }
        } else {
          mPoint += eachDrive.mountpoint + this.get('defaultDirectory') + "\n";
        }
        this.set('value', mPoint);
        this.set('defaultValue', mPoint);
      }, this);
    } else {
      var mPoint = allMountPoints[0].mountpoint;
      if (mPoint === "/") {
        mPoint = this.get('defaultDirectory');
      } else if(winRegex.test(mPoint.toLowerCase())) {
        switch (this.get('name')) {
          case 'fs.checkpoint.dir':
          case 'dfs.namenode.checkpoint.dir':
            var winDriveUrl = mPoint.toLowerCase().replace(winRegex, "file:///$1:");
            mPoint = winDriveUrl + this.get('defaultDirectory') + "\n";
            break;
          case 'zk_data_dir':
            var winDrive = mPoint.toLowerCase().replace(winRegex, "$1:");
            var winDir = this.get('defaultDirectory').replace(/\//g, "\\\\");
            mPoint = winDrive + winDir + "\n";
            break;
          default:
            var winDrive = mPoint.toLowerCase().replace(winRegex, "$1:");
            var winDir = this.get('defaultDirectory').replace(/\//g, "\\");
            mPoint = winDrive + winDir + "\n";
        }
      } else {
        mPoint = mPoint + this.get('defaultDirectory');
      }
      this.set('value', mPoint);
      this.set('defaultValue', mPoint);
    }
  },

  isValid: function () {
    return this.get('errorMessage') === '';
  }.property('errorMessage'),

  viewClass: function () {
    switch (this.get('displayType')) {
      case 'checkbox':
        if (this.get('dependentConfigPattern')) {
          return App.ServiceConfigCheckboxWithDependencies;
        } else {
          return App.ServiceConfigCheckbox;
        }
      case 'password':
        return App.ServiceConfigPasswordField;
      case 'combobox':
        return App.ServiceConfigComboBox;
      case 'radio button':
        return App.ServiceConfigRadioButtons;
        break;
      case 'directories':
      case 'datanodedirs':
        return App.ServiceConfigTextArea;
        break;
      case 'content':
        return App.ServiceConfigTextAreaContent;
        break;
      case 'multiLine':
        return App.ServiceConfigTextArea;
        break;
      case 'custom':
        return App.ServiceConfigBigTextArea;
      case 'masterHost':
        return App.ServiceConfigMasterHostView;
      case 'label':
        return App.ServiceConfigLabelView;
      case 'masterHosts':
        return App.ServiceConfigMasterHostsView;
      case 'slaveHosts':
        return App.ServiceConfigSlaveHostsView;
      case 'supportTextConnection':
        return App.checkConnectionView;
      default:
        if (this.get('unit')) {
          return App.ServiceConfigTextFieldWithUnit;
        } else {
          return App.ServiceConfigTextField;
        }
    }
  }.property('displayType'),

  validate: function () {
    var value = this.get('value');
    var supportsFinal = this.get('supportsFinal');
    var isFinal = this.get('isFinal');
    var valueRange = this.get('valueRange');
    var values = [];//value split by "," to check UNIX users, groups list

    var isError = false;
    var isWarn = false;

    if (typeof value === 'string' && value.length === 0) {
      if (this.get('isRequired')) {
        this.set('errorMessage', 'This is required');
        isError = true;
      } else {
        return;
      }
    }

    if (!isError) {
      switch (this.get('displayType')) {
        case 'int':
          if (!validator.isValidInt(value)) {
            this.set('errorMessage', 'Must contain digits only');
            isError = true;
          } else {
            if(valueRange){
              if(value < valueRange[0] || value > valueRange[1]){
                this.set('errorMessage', 'Must match the range');
                isError = true;
              }
            }
          }
          break;
        case 'float':
          if (!validator.isValidFloat(value)) {
            this.set('errorMessage', 'Must be a valid number');
            isError = true;
          }
          break;
        case 'UNIXList':
          if(value != '*'){
            values = value.split(',');
            for(var i = 0, l = values.length; i < l; i++){
              if(!validator.isValidUNIXUser(values[i])){
                if(this.get('type') == 'USERS'){
                  this.set('errorMessage', 'Must be a valid list of user names');
                } else {
                  this.set('errorMessage', 'Must be a valid list of group names');
                }
                isError = true;
              }
            }
          }
          break;
        case 'checkbox':
          break;
        case 'datanodedirs':
          if (!validator.isValidDataNodeDir(value)) {
            this.set('errorMessage', 'dir format is wrong, can be "[{storage type}]/{dir name}"');
            isError = true;
          }
          else {
            if (!validator.isAllowedDir(value)) {
              this.set('errorMessage', 'Cannot start with "home(s)"');
              isError = true;
            }
          }
          break;
        case 'directories':
        case 'directory':
          if (!validator.isValidDir(value)) {
            this.set('errorMessage', 'Must be a slash or drive at the start');
            isError = true;
          }
          else {
            if (!validator.isAllowedDir(value)) {
              this.set('errorMessage', 'Can\'t start with "home(s)"');
              isError = true;
            }
          }
          break;
        case 'custom':
          break;
        case 'email':
          if (!validator.isValidEmail(value)) {
            this.set('errorMessage', 'Must be a valid email address');
            isError = true;
          }
          break;
        case 'host':
          var hiveOozieHostNames = ['hive_hostname','hive_existing_mysql_host','hive_existing_oracle_host','hive_ambari_host',
          'oozie_hostname','oozie_existing_mysql_host','oozie_existing_oracle_host','oozie_ambari_host'];
          if(hiveOozieHostNames.contains(this.get('name'))) {
            if (validator.hasSpaces(value)) {
              this.set('errorMessage', Em.I18n.t('host.spacesValidation'));
              isError = true;
            }
          } else {
            if (validator.isNotTrimmed(value)) {
              this.set('errorMessage', Em.I18n.t('host.trimspacesValidation'));
              isError = true;
            }
          }
          break;
        case 'advanced':
          if(this.get('name')=='javax.jdo.option.ConnectionURL' || this.get('name')=='oozie.service.JPAService.jdbc.url') {
            if (validator.isNotTrimmed(value)) {
              this.set('errorMessage', Em.I18n.t('host.trimspacesValidation'));
              isError = true;
            }
          }
          break;
        case 'password':
          // retypedPassword is set by the retypePasswordView child view of App.ServiceConfigPasswordField
          if (value !== this.get('retypedPassword')) {
            this.set('errorMessage', 'Passwords do not match');
            isError = true;
          }
      }
    }

    if (!isError) {
      // Check if this value is already in any of the overrides
      var self = this;
      var isOriginalSCP = this.get('isOriginalSCP');
      var parentSCP = this.get('parentSCP');
      if (!isOriginalSCP) {
        if (!isError && parentSCP != null) {
          if (value === parentSCP.get('value') && supportsFinal && isFinal === parentSCP.get('isFinal')) {
            this.set('errorMessage', 'Configuration overrides must have different value');
            isError = true;
          } else {
            var overrides = parentSCP.get('overrides');
            if (overrides) {
              overrides.forEach(function (override) {
                if (self != override && value === override.get('value')  && supportsFinal && isFinal === parentSCP.get('isFinal')) {
                  self.set('errorMessage', 'Multiple configuration overrides cannot have same value');
                  isError = true;
                }
              });
            }
          }
        }
      }
    }

    if (!isWarn || isError) { // Errors get priority
        this.set('warnMessage', '');
        this.set('warn', false);
    } else {
        this.set('warn', true);
    }

    if (!isError) {
        this.set('errorMessage', '');
        this.set('error', false);
      } else {
        this.set('error', true);
      }
  }.observes('value', 'isFinal', 'retypedPassword')

});

App.ConfigSiteTag = Ember.Object.extend({
  site: DS.attr('string'),
  tag: DS.attr('string'),
  /**
   * Object map of hostname->override-tag for overrides.
   * <b>Creators should set new object here.<b>
   */
  hostOverrides: null
});

});

require.register("models/service_config_version", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');
var dateUtil = require('utils/date');


App.ServiceConfigVersion = DS.Model.extend({
  serviceName: DS.attr('string'),
  displayName: function() {
    return App.format.role(this.get('serviceName'));
  }.property('serviceName'),
  groupName: DS.attr('string'),
  groupId: DS.attr('number'),
  version: DS.attr('number'),
  createTime: DS.attr('number'),
  author: DS.attr('string'),
  notes: DS.attr('string'),
  service: DS.belongsTo('App.Service'),
  hosts: DS.attr('array'),
  index: DS.attr('number'),
  isCurrent: DS.attr('boolean'),
  isDisplayed: DS.attr('boolean'),
  isDefault: function() {
    return this.get('groupName') === 'default';
  }.property('groupName'),
  currentTooltip: function () {
    return Em.I18n.t('dashboard.configHistory.table.current.tooltip').format(this.get('displayName'), this.get('configGroupName'));
  }.property('displayName', 'configGroupName'),
  configGroupName: function () {
    return this.get('isDefault') ? (this.get('displayName') + ' ' + Em.I18n.t('common.default')) : this.get('groupName');
  }.property('groupName','isDefault'),
  authorFormatted: function () {
    var author = this.get('author');
    if (author) {
      return author.length > 20 ? author.slice(0, 20) + '...' : author;
    }
  }.property('author'),
  fullNotes: function () {
    return (typeof this.get('notes') === 'string') ? this.get('notes') || Em.I18n.t('dashboard.configHistory.table.notes.no') : Em.I18n.t('dashboard.configHistory.table.notes.no');
  }.property('notes'),
  briefNotes: function () {
    return this.get('fullNotes').slice(0, 81);
  }.property('fullNotes'),
  moreNotesExists: function () {
    return (typeof this.get('notes') === 'string') ?  this.get('notes').length > 80 : false;
  }.property('notes'),
  versionText: function () {
    return Em.I18n.t('dashboard.configHistory.table.version.versionText').format(this.get('version'));
  }.property('version'),
  makeCurrentButtonText: function() {
    return Em.I18n.t('dashboard.configHistory.info-bar.revert.versionButton').format(this.get('versionText'));
  }.property('versionText'),
  createdDate: function () {
    return dateUtil.dateFormat(this.get('createTime'));
  }.property('createTime'),
  timeSinceCreated: function () {
    return $.timeago(this.get('createTime'));
  }.property('createTime'),
  /**
   * determine whether ServiceConfigVersion is requested from server
   */
  isRequested: DS.attr('boolean'),
  isRestartRequired: function () {
    if (this.get('service.isRestartRequired') && this.get('isCurrent')) {
      var hostNames = this.get('hosts');
      if (!hostNames.length) return false;
      for (var i = 0; i < hostNames.length; i++) {
        if (Object.keys(this.get('service.restartRequiredHostsAndComponents')).contains(hostNames[i])) {
          return true;
        }
      }
    }
    return false;
  }.property('service.isRestartRequired','isDefault', 'isCurrent', 'hosts', 'service.restartRequiredHostsAndComponents', 'router.mainServiceInfoConfigsController.configGroups'),
  disabledActionMessages: function () {
    return {
      view: (this.get('isDisplayed')) ? Em.I18n.t('dashboard.configHistory.info-bar.view.button.disabled') : '',
      compare: (this.get('isDisplayed')) ? Em.I18n.t('dashboard.configHistory.info-bar.compare.button.disabled') : '',
      revert: (this.get('isCurrent')) ? Em.I18n.t('dashboard.configHistory.info-bar.revert.button.disabled') : ''
    }
  }.property('isDisplayed', 'isCurrent'),
  disabledActionAttr: function () {
    return {
      view: (this.get('isDisplayed')) ? 'disabled' : false,
      compare: (this.get('isDisabled') || this.get('isDisplayed')) ? 'disabled' : false,
      revert: (this.get('isDisabled') || this.get('isCurrent')) ? 'disabled' : false
    }
  }.property('isDisplayed', 'isCurrent', 'isDisabled')
});

App.ServiceConfigVersion.FIXTURES = [];

});

require.register("models/slave_component", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

App.SlaveComponent = App.ClientComponent.extend({
  displayNamePluralized: function() {
    return stringUtils.pluralize(this.get('startedCount'), this.get('displayName'));
  }.property('startedCount')
});

App.SlaveComponent.FIXTURES = [];

});

require.register("models/stack", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.Stack = DS.Model.extend({
  id: DS.attr('string'), //  ${stackName}-${stackVersion}.
  stackName: DS.attr('string'),
  stackVersion: DS.attr('string'),
  active: DS.attr('boolean'),  // All of the instances should have this value to true. We should map only those stacks that has active flag set to true
  parentStackVersion: DS.attr('string'),
  minUpgradeVersion: DS.attr('string'),
  configTypes: DS.attr('object'),
  operatingSystems: DS.hasMany('App.OperatingSystem'),
  isSelected: DS.attr('boolean', {defaultValue: false}),

  /**
   * @return: {Array} returns supported repositories for all OperatingSystem's supported by a stack instance
   */
  repositories: function () {
    var operatingSystems = this.get('operatingSystems');
    var repositories = [];
    operatingSystems.forEach(function (os) {
      os.get('repositories').forEach(function (repository) {
        repositories.pushObject(repository);
      }, this);
    }, this);
    return repositories;
  }.property('id'),

  /**
   * @return: {Array} App.StackService instances for selected stack instance. For non-selected stack instance returns empty array
   */
  services: function () {
    var result = [];
    var isStackSelected = this.get('isSelected');
    var stackServices = App.StackService.find().get('length');
    if (isStackSelected && stackServices) {
      result = App.StackService.find();
    }
    return result;
  }.property('isSelected'),

  /**
   * Right now there ambari-web is not fetching this information from the server as it does not need as of present.
   * @TODO: This should return stack level configurations for selected stack instance i.e properties of cluster-env file
   */
  configurations: function() {
    return [];
  }.property('isSelected')

});


App.Stack.FIXTURES = [];

});

require.register("models/stack_service", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
require('utils/helper');
require('models/service_config');

/**
 * This model loads all services supported by the stack
 * The model maps to the  http://hostname:8080/api/v1/stacks/HDP/versions/${versionNumber}/services?fields=StackServices/*,serviceComponents/*
 * @type {*}
 */
App.StackService = DS.Model.extend({
  serviceName: DS.attr('string'),
  displayName: DS.attr('string'),
  comments: DS.attr('string'),
  configTypes: DS.attr('object'),
  serviceVersion: DS.attr('string'),
  serviceCheckSupported: DS.attr('boolean'),
  stackName: DS.attr('string'),
  stackVersion: DS.attr('string'),
  isSelected: DS.attr('boolean', {defaultValue: true}),
  isInstalled: DS.attr('boolean', {defaultValue: false}),
  isInstallable: DS.attr('boolean', {defaultValue: true}),
  stack: DS.belongsTo('App.Stack'),
  serviceComponents: DS.hasMany('App.StackServiceComponent'),
  configs: DS.attr('array'),
  requiredServices: DS.attr('array'),

  // Is the service a distributed filesystem
  isDFS: function () {
    var dfsServices = ['HDFS', 'GLUSTERFS'];
    return dfsServices.contains(this.get('serviceName'));
  }.property('serviceName'),

  // Primary DFS. used if there is more than one DFS in a stack.
  // Only one service in the stack should be tagged as primary DFS.
  isPrimaryDFS: function () {
    return this.get('serviceName') === 'HDFS';
  }.property('serviceName'),

  configTypesRendered: function () {
    var configTypes = this.get('configTypes');
    var renderedConfigTypes = $.extend(true, {}, configTypes);
    if (this.get('serviceName') == 'FALCON') {
      delete renderedConfigTypes['oozie-site'];
    }
    return renderedConfigTypes
  }.property('serviceName', 'configTypes'),

  displayNameOnSelectServicePage: function () {
    var displayName = this.get('displayName');
    console.info("displayName = " + displayName);
    var services = this.get('coSelectedServices').slice();
    var serviceDisplayNames = services.map(function (item) {
      return App.format.role(item);
    }, this);
    if (!!serviceDisplayNames.length) {
      serviceDisplayNames.unshift(displayName);
      displayName = serviceDisplayNames.join(" + ");
    }
    return displayName;
  }.property('coSelectedServices', 'serviceName'),

  isHiddenOnSelectServicePage: function () {
    var hiddenServices = ['MAPREDUCE2'];
    return hiddenServices.contains(this.get('serviceName')) || !this.get('isInstallable') || this.get('doNotShowAndInstall');
  }.property('serviceName', 'isInstallable'),

  doNotShowAndInstall: function () {
    var skipServices = [];
    if(!App.supports.installGanglia) {
      skipServices.push('GANGLIA');
    }
    if(App.router.get('clusterInstallCompleted') != true){
      skipServices.push('RANGER');
    }
    return skipServices.contains(this.get('serviceName'));
  }.property('serviceName'),

  // Is the service required for monitoring of other hadoop ecosystem services
  isMonitoringService: function () {
    var services = ['GANGLIA'];
    return services.contains(this.get('serviceName'));
  }.property('serviceName'),

  // Is the service required for reporting host metrics
  isHostMetricsService: function () {
      var services = ['GANGLIA'];
      return services.contains(this.get('serviceName'));
  }.property('serviceName'),

  // Is the service required for reporting hadoop service metrics
  isServiceMetricsService: function () {
      var services = ['GANGLIA'];
      return services.contains(this.get('serviceName'));
  }.property('serviceName'),

  coSelectedServices: function () {
    var coSelectedServices = App.StackService.coSelected[this.get('serviceName')];
    if (!!coSelectedServices) {
      return coSelectedServices;
    } else {
      return [];
    }
  }.property('serviceName'),

  hasClient: function () {
    var serviceComponents = this.get('serviceComponents');
    return serviceComponents.someProperty('isClient');
  }.property('serviceName'),

  hasMaster: function () {
    var serviceComponents = this.get('serviceComponents');
    return serviceComponents.someProperty('isMaster');
  }.property('serviceName'),

  hasSlave: function () {
    var serviceComponents = this.get('serviceComponents');
    return serviceComponents.someProperty('isSlave');
  }.property('serviceName'),

  hasNonMastersWithCustomAssignment: function () {
    var serviceComponents = this.get('serviceComponents');
    return serviceComponents.rejectProperty('isMaster').rejectProperty('cardinality', 'ALL').length > 0;
  }.property('serviceName'),

  isClientOnlyService: function () {
    var serviceComponents = this.get('serviceComponents');
    return serviceComponents.everyProperty('isClient');
  }.property('serviceName'),

  isNoConfigTypes: function () {
    var configTypes = this.get('configTypes');
    return !(configTypes && !!Object.keys(configTypes).length);
  }.property('configTypes'),

  customReviewHandler: function () {
    return App.StackService.reviewPageHandlers[this.get('serviceName')];
  }.property('serviceName'),

  /**
   * configCategories are fetched from  App.StackService.configCategories.
   * Also configCategories that does not match any serviceComponent of a service and not included in the permissible default pattern are omitted
   */
  configCategories: function () {
    var configCategories = [];
    var configTypes = this.get('configTypes');
    var serviceComponents = this.get('serviceComponents');
    if (configTypes && Object.keys(configTypes).length) {
      var pattern = ["General", "CapacityScheduler", "FaultTolerance", "Isolation", "Performance", "KDC", "Kadmin","^Advanced", "Env$", "^Custom", "Falcon - Oozie integration", "FalconStartupSite", "FalconRuntimeSite", "MetricCollector", "Settings$"];
      configCategories = App.StackService.configCategories.call(this).filter(function (_configCategory) {
        var serviceComponentName = _configCategory.get('name');
        var isServiceComponent = serviceComponents.someProperty('componentName', serviceComponentName);
        if (isServiceComponent) return  isServiceComponent;
        var result = false;
        pattern.forEach(function (_pattern) {
          var regex = new RegExp(_pattern);
          if (regex.test(serviceComponentName)) result = true;
        });
        return result;
      });
    }
    return configCategories;
  }.property('serviceName', 'configTypes', 'serviceComponents')
});

App.StackService.FIXTURES = [];

App.StackService.displayOrder = [
  'HDFS',
  'GLUSTERFS',
  'MAPREDUCE2',
  'YARN',
  'TEZ',
  'GANGLIA',
  'HIVE',
  'HBASE',
  'PIG',
  'SQOOP',
  'OOZIE',
  'ZOOKEEPER',
  'FALCON',
  'STORM',
  'FLUME'
];

//@TODO: Write unit test for no two keys in the object should have any intersecting elements in their values
App.StackService.coSelected = {
  'YARN': ['MAPREDUCE2']
};


App.StackService.reviewPageHandlers = {
  'HIVE': {
    'Database': 'loadHiveDbValue'
  },
  'OOZIE': {
    'Database': 'loadOozieDbValue'
  }
};

App.StackService.configCategories = function () {
  var serviceConfigCategories = [];
  switch (this.get('serviceName')) {
    case 'HDFS':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'NAMENODE', displayName: 'NameNode'}),
        App.ServiceConfigCategory.create({ name: 'SECONDARY_NAMENODE', displayName: 'Secondary NameNode'}),
        App.ServiceConfigCategory.create({ name: 'DATANODE', displayName: 'DataNode'}),
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General'})
      ]);
      break;
    case 'GLUSTERFS':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General'})
      ]);
      break;
    case 'YARN':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'RESOURCEMANAGER', displayName: 'Resource Manager'}),
        App.ServiceConfigCategory.create({ name: 'NODEMANAGER', displayName: 'Node Manager'}),
        App.ServiceConfigCategory.create({ name: 'APP_TIMELINE_SERVER', displayName: 'Application Timeline Server'}),
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General'}),
        App.ServiceConfigCategory.create({ name: 'FaultTolerance', displayName: 'Fault Tolerance'}),
        App.ServiceConfigCategory.create({ name: 'Isolation', displayName: 'Isolation'}),
        App.ServiceConfigCategory.create({ name: 'CapacityScheduler', displayName: 'Scheduler', siteFileName: 'capacity-scheduler.xml'})
      ]);
      break;
    case 'MAPREDUCE2':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'HISTORYSERVER', displayName: 'History Server'}),
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General'})
      ]);
      break;
    case 'HIVE':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'HIVE_METASTORE', displayName: 'Hive Metastore'}),
        App.ServiceConfigCategory.create({ name: 'WEBHCAT_SERVER', displayName: 'WebHCat Server'}),
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General'}),
        App.ServiceConfigCategory.create({ name: 'Performance', displayName: 'Performance'})
      ]);
      break;
    case 'HBASE':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'HBASE_MASTER', displayName: 'HBase Master'}),
        App.ServiceConfigCategory.create({ name: 'HBASE_REGIONSERVER', displayName: 'RegionServer'}),
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General'})
      ]);
      break;
    case 'ZOOKEEPER':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'ZOOKEEPER_SERVER', displayName: 'ZooKeeper Server'})
      ]);
      break;
    case 'OOZIE':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'OOZIE_SERVER', displayName: 'Oozie Server'})
      ]);
      break;
    case 'FALCON':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'FALCON_SERVER', displayName: 'Falcon Server'}),
        App.ServiceConfigCategory.create({ name: 'Falcon - Oozie integration', displayName: 'Falcon - Oozie integration'}),
        App.ServiceConfigCategory.create({ name: 'FalconStartupSite', displayName: 'Falcon startup.properties'}),
        App.ServiceConfigCategory.create({ name: 'FalconRuntimeSite', displayName: 'Falcon runtime.properties'}),
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General'})
      ]);
      break;
    case 'STORM':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'NIMBUS', displayName: 'Nimbus'}),
        App.ServiceConfigCategory.create({ name: 'SUPERVISOR', displayName: 'Supervisor'}),
        App.ServiceConfigCategory.create({ name: 'STORM_UI_SERVER', displayName: 'Storm UI Server'}),
        App.ServiceConfigCategory.create({ name: 'STORM_REST_API', displayName: 'Storm REST API Server'}),
        App.ServiceConfigCategory.create({ name: 'DRPC_SERVER', displayName: 'DRPC Server'}),
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General'})
      ]);
      break;
    case 'TEZ':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General'})
      ]);
      break;
    case 'FLUME':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'FLUME_HANDLER', displayName: 'flume.conf', siteFileName: 'flume-conf', canAddProperty: false})
      ]);
      break;
    case 'KNOX':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'KNOX_GATEWAY', displayName: 'Knox Gateway'})
      ]);
      break;
    case 'KAFKA':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'KAFKA_BROKER', displayName: 'Kafka Broker'})
      ]);
      break;
    case 'KERBEROS':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'KDC', displayName: 'KDC'}),
        App.ServiceConfigCategory.create({ name: 'Kadmin', displayName: 'Kadmin'}),
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General'})
      ]);
      break;
    case 'AMBARI_METRICS':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General'}),
        App.ServiceConfigCategory.create({ name: 'MetricCollector', displayName: 'Metric Collector'})
      ]);
      break;
    case 'RANGER':
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'AdminSettings', displayName: 'Admin Settings'}),
        App.ServiceConfigCategory.create({ name: 'DBSettings', displayName: 'DB Settings'}),
        App.ServiceConfigCategory.create({ name: 'RangerSettings', displayName: 'Ranger Settings'}),
        App.ServiceConfigCategory.create({ name: 'UnixAuthenticationSettings', displayName: 'Unix Authentication Settings'}),
        App.ServiceConfigCategory.create({ name: 'ADSettings', displayName: 'AD Settings'}),
        App.ServiceConfigCategory.create({ name: 'LDAPSettings', displayName: 'LDAP Settings'})
      ]);
      break;
    case 'PIG':
      break;
    case 'SQOOP':
      break;
    default:
      serviceConfigCategories.pushObjects([
        App.ServiceConfigCategory.create({ name: 'General', displayName: 'General'})
      ]);
  }
  serviceConfigCategories.pushObject(App.ServiceConfigCategory.create({ name: 'Advanced', displayName: 'Advanced'}));

  var configTypes = Object.keys(this.get('configTypes'));

  //Falcon has dependency on oozie-site but oozie-site advanced/custom section should not be shown on Falcon page
  if (this.get('serviceName') !== 'OOZIE') {
    configTypes = configTypes.without('oozie-site');
  }

  // Add Advanced section for every configType to all the services
  configTypes.forEach(function (type) {
    serviceConfigCategories.pushObject(App.ServiceConfigCategory.create({
      name: 'Advanced ' + type,
      displayName: Em.I18n.t('common.advanced') + " " + type,
      canAddProperty: false
    }));
  }, this);

  // Add custom section for every configType to all the services
  configTypes.forEach(function (type) {
    var configTypesWithNoCustomSection = ['capacity-scheduler','mapred-queue-acls','flume-conf', 'pig-properties','topology','users-ldif'];
    if (type.endsWith('-env') || type.endsWith('-log4j') || configTypesWithNoCustomSection.contains(type)) {
      return;
    }
    serviceConfigCategories.pushObject(App.ServiceConfigCategory.create({
      name: 'Custom ' + type,
      displayName: Em.I18n.t('common.custom') + " " + type,
      siteFileName: type + '.xml',
      canAddProperty: true
    }));
  }, this);
  return serviceConfigCategories;
};

});

require.register("models/stack_service_component", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var numberUtils = require('utils/number_utils');
/**
 * This model loads all serviceComponents supported by the stack
 * @type {*}
 */
App.StackServiceComponent = DS.Model.extend({
  componentName: DS.attr('string'),
  displayName: DS.attr('string'),
  cardinality: DS.attr('string'),
  customCommands: DS.attr('array'),
  dependencies: DS.attr('array'),
  serviceName: DS.attr('string'),
  componentCategory: DS.attr('string'),
  isMaster: DS.attr('boolean'),
  isClient: DS.attr('boolean'),
  stackName: DS.attr('string'),
  stackVersion: DS.attr('string'),
  stackService: DS.belongsTo('App.StackService'),
  serviceComponentId: DS.attr('number', {defaultValue: 1}), // this is used on Assign Master page for multiple masters

  /**
   * Minimum required count for installation.
   *
   * @property {Number} minToInstall
   **/
  minToInstall: function() {
    return numberUtils.getCardinalityValue(this.get('cardinality'), false);
  }.property('cardinality'),

  /**
   * Maximum required count for installation.
   *
   * @property {Number} maxToInstall
   **/
  maxToInstall: function() {
    return numberUtils.getCardinalityValue(this.get('cardinality'), true);
  }.property('cardinality'),

  /** @property {Boolean} isRequired - component required to install **/
  isRequired: function() {
    return this.get('minToInstall') > 0;
  }.property('cardinality'),

  /** @property {Boolean} isMultipleAllowed - component can be assigned for more than one host **/
  isMultipleAllowed: function() {
    return this.get('maxToInstall') > 1;
  }.property('cardinality'),

  /** @property {Boolean} isSlave **/
  isSlave: function() {
   return this.get('componentCategory') === 'SLAVE';
  }.property('componentCategory'),

  /** @property {Boolean} isRestartable - component supports restart action **/
  isRestartable: function() {
    return !this.get('isClient');
  }.property('isClient'),

  /** @property {Boolean} isReassignable - component supports reassign action **/
  isReassignable: function() {
    return ['NAMENODE', 'SECONDARY_NAMENODE', 'JOBTRACKER', 'RESOURCEMANAGER', 'APP_TIMELINE_SERVER', 'OOZIE_SERVER', 'WEBHCAT_SERVER', 'HIVE_SERVER', 'HIVE_METASTORE', 'MYSQL_SERVER'].contains(this.get('componentName'));
  }.property('componentName'),

  /** @property {Boolean} isRollinRestartAllowed - component supports rolling restart action **/
  isRollinRestartAllowed: function() {
    return this.get('isSlave');
  }.property('componentName'),

  /** @property {Boolean} isDecommissionAllowed - component supports decommission action **/
  isDecommissionAllowed: function() {
    return ["DATANODE", "TASKTRACKER", "NODEMANAGER", "HBASE_REGIONSERVER"].contains(this.get('componentName'));
  }.property('componentName'),

  /** @property {Boolean} isRefreshConfigsAllowed - component supports refresh configs action **/
  isRefreshConfigsAllowed: function() {
    return ["FLUME_HANDLER"].contains(this.get('componentName'));
  }.property('componentName'),

  /** @property {Boolean} isAddableToHost - component can be added on host details page **/
  isAddableToHost: function() {
    return ((this.get('isMasterAddableInstallerWizard') || (this.get('isSlave') && this.get('maxToInstall') > 2) || this.get('isClient')) && !this.get('isHAComponentOnly'));
  }.property('componentName'),

  /** @property {Boolean} isDeletable - component supports delete action **/
  isDeletable: function() {
    var ignored = ['HBASE_MASTER'];
    return this.get('isAddableToHost') && !ignored.contains(this.get('componentName'));
  }.property('componentName'),

  /** @property {Boolean} isShownOnInstallerAssignMasterPage - component visible on "Assign Masters" step of Install Wizard **/
  isShownOnInstallerAssignMasterPage: function() {
    var component = this.get('componentName');
    var mastersNotShown = ['MYSQL_SERVER', 'POSTGRESQL_SERVER'];
    return this.get('isMaster') && !mastersNotShown.contains(component);
  }.property('isMaster','componentName'),

  /** @property {Boolean} isShownOnInstallerSlaveClientPage - component visible on "Assign Slaves and Clients" step of Install Wizard**/
  isShownOnInstallerSlaveClientPage: function() {
    var component = this.get('componentName');
    var slavesNotShown = ['JOURNALNODE','ZKFC','APP_TIMELINE_SERVER'];
    return this.get('isSlave') && !this.get('isRequiredOnAllHosts') && !slavesNotShown.contains(component);
  }.property('isSlave','componentName', 'isRequiredOnAllHosts'),

  /** @property {Boolean} isShownOnAddServiceAssignMasterPage - component visible on "Assign Masters" step of Add Service Wizard **/
  isShownOnAddServiceAssignMasterPage: function() {
    var isVisible = this.get('isShownOnInstallerAssignMasterPage');
    if (App.get('isHaEnabled')) {
      isVisible =  isVisible && this.get('componentName') !== 'SECONDARY_NAMENODE';
    }
    return isVisible;
  }.property('isShownOnInstallerAssignMasterPage','App.isHaEnabled'),

  /** @property {Boolean} isMasterWithMultipleInstances **/
  isMasterWithMultipleInstances: function() {
    // @todo: safe removing JOURNALNODE from masters list
    return (this.get('isMaster') && this.get('isMultipleAllowed')) || this.get('componentName') == 'JOURNALNODE';
  }.property('componentName'),

  /**
   * Master component list that could be assigned for more than 1 host.
   * Some components like NameNode and ResourceManager have range cardinality value
   * like 1-2. We can assign only components with cardinality 1+/0+. Strict range value
   * show that this components will be assigned for 2 hosts only if HA mode activated.
   *
   * @property {Boolean} isMasterAddableInstallerWizard
   **/
  isMasterAddableInstallerWizard: function() {
    return this.get('isMaster') && this.get('isMultipleAllowed') && this.get('maxToInstall') > 2;
  }.property('componentName'),

  /** @property {Boolean} isHAComponentOnly - Components that can be installed only if HA enabled **/
  isHAComponentOnly: function() {
    var HAComponentNames = ['ZKFC','JOURNALNODE'];
    return HAComponentNames.contains(this.get('componentName'));
  }.property('componentName'),

  /** @property {Boolean} isRequiredOnAllHosts - Is It require to install the components on all hosts. used in step-6 wizard controller **/
  isRequiredOnAllHosts: function() {
    return this.get('minToInstall') == Infinity;
  }.property('stackService','isSlave'),

  /** @property {Number} defaultNoOfMasterHosts - default number of master hosts on Assign Master page: **/
  defaultNoOfMasterHosts: function() {
     if (this.get('isMasterAddableInstallerWizard')) {
       return this.get('componentName') == 'ZOOKEEPER_SERVER' ? 3 : this.get('minToInstall');
     }
  }.property('componentName'),

  /** @property {Boolean} coHostedComponents - components that are co-hosted with this component **/
  coHostedComponents: function() {
    var componentName = this.get('componentName');
    var key, coHostedComponents = [];
    for (key in App.StackServiceComponent.coHost) {
      if (App.StackServiceComponent.coHost[key] === componentName) {
        coHostedComponents.push(key)
      }
    }
    return coHostedComponents;
  }.property('componentName'),

  /** @property {Boolean} isOtherComponentCoHosted - Is any other component co-hosted with this component **/
  isOtherComponentCoHosted: function() {
    return !!this.get('coHostedComponents').length;
  }.property('coHostedComponents'),

  /** @property {Boolean} isCoHostedComponent - Is this component co-hosted with other component **/
  isCoHostedComponent: function() {
    var componentName = this.get('componentName');
    return !!App.StackServiceComponent.coHost[componentName];
  }.property('componentName'),

  /** @property {Boolean} showAddBtnInInstall - show add button for this component on Assign Masters  **/
  showAddBtnInInstall: function() {
    var doNotShowList = ['HIVE_METASTORE', 'HIVE_SERVER'];
    return !doNotShowList.contains(this.get('componentName'));
  }.property('componentName')

});

App.StackServiceComponent.FIXTURES = [];

App.StackServiceComponent.coHost = {
  'WEBHCAT_SERVER': 'HIVE_SERVER'
};

});

require.register("models/stack_version/os", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.OS = DS.Model.extend({
  id: DS.attr('string'),
  osType : DS.attr('string'),
  stackName : DS.attr('string'),
  stackVersion : DS.attr('string'),
  repositoryVersion: DS.belongsTo('App.RepositoryVersion'),
  repositories: DS.hasMany('App.Repository')
});

App.OS.FIXTURES = [];



});

require.register("models/stack_version/repository", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.Repository = DS.Model.extend({
  id: DS.attr('string'),
  baseUrl : DS.attr('string'),
  defaultBaseUrl : DS.attr('string'),
  latestBaseUrl : DS.attr('string'),
  mirrorsList : DS.attr('array'),
  osType : DS.attr('string'),
  repoId : DS.attr('string'),
  repoName : DS.attr('string'),
  stackName : DS.attr('string'),
  stackVersion : DS.attr('string'),
  operatingSystem: DS.belongsTo('App.OS')
});

App.Repository.FIXTURES = [];



});

require.register("models/stack_version/repository_version", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.RepositoryVersion = DS.Model.extend({
  displayName: DS.attr('string'),
  repositoryVersion: DS.attr('string'),
  upgradePack: DS.attr('string'),
  stackVersionType: DS.attr('string'),
  stackVersionNumber: DS.attr('string'),
  operatingSystems: DS.hasMany('App.OS'),
  stackVersion: DS.belongsTo('App.StackVersion'),
  stack: function () {
    return this.get('stackVersionType') + " " + this.get('stackVersionNumber');
  }.property('stackVersionType', 'stackVersionNumber'),

  /**
   * status used until corresponding stack version get created
   * @type {string}
   */
  defaultStatus: 'INIT',

  /**
   * @type {string}
   */
  status: function () {
    return this.get('stackVersion.state') || this.get('defaultStatus');
  }.property('stackVersion.state', 'defaultStatus'),

  /**
   * @type {Array}
   */
  notInstalledHosts: function () {
    return this.get('stackVersion.notInstalledHosts') || App.get('allHostNames');
  }.property('stackVersion.notInstalledHosts'),

  /**
   * @type {Array}
   */
  installedHosts: function () {
    return this.get('stackVersion.installedHosts') || [];
  }.property('stackVersion.installedHosts'),

  /**
   * @type {Array}
   */
  currentHosts: function () {
    return this.get('stackVersion.currentHosts') || [];
  }.property('stackVersion.currentHosts'),

  /**
   * @type {boolean}
   */
  noInstalledHosts: function () {
    return (this.get('stackVersion')) ? this.get('stackVersion.noInstalledHosts') : true;
  }.property('stackVersion.noInstalledHosts'),

  /**
   * @type {boolean}
   */
  noCurrentHosts: function () {
    return (this.get('stackVersion')) ? this.get('stackVersion.noCurrentHosts') : true;
  }.property('stackVersion.noCurrentHosts'),

  /**
   * @type {boolean}
   */
  noInitHosts: function () {
    return (this.get('stackVersion')) ? this.get('stackVersion.noInitHosts') : false;
  }.property('stackVersion.noInitHosts'),

  /**
   * @type {string}
   */
  noInitHostsTooltip: function () {
    return (this.get('noInitHosts')) ? Em.I18n.t('admin.stackVersions.version.emptyHostsTooltip') : Em.I18n.t('admin.stackVersions.version.hostsTooltip');
  }.property('noInitHosts'),

  /**
   * @type {string}
   */
  noCurrentHostsTooltip: function () {
    return (this.get('noCurrentHosts')) ? Em.I18n.t('admin.stackVersions.version.emptyHostsTooltip') : Em.I18n.t('admin.stackVersions.version.hostsTooltip');
  }.property('noCurrentHosts'),

  /**
   * @type {string}
   */
  noInstalledHostsTooltip: function () {
    return (this.get('noInstalledHosts')) ? Em.I18n.t('admin.stackVersions.version.emptyHostsTooltip') : Em.I18n.t('admin.stackVersions.version.hostsTooltip');
  }.property('noInstalledHosts'),

  /**
   * @type {boolean}
   */
  isVisible: true
});

App.RepositoryVersion.FIXTURES = [];


});

require.register("models/stack_version/version", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.StackVersion = DS.Model.extend({
  clusterName: DS.attr('string'),
  stack: DS.attr('string'),
  version: DS.attr('string'),
  name: function() {
    return this.get('stack') + " " + this.get('version');
  }.property('stack', 'version'),
  state: DS.attr('string'),
  repositoryVersion: DS.belongsTo('App.RepositoryVersion'),
  notInstalledHosts: DS.attr('array'),
  installingHosts: DS.attr('array'),
  installedHosts: DS.attr('array'),
  installFailedHosts: DS.attr('array'),
  outOfSyncHosts: DS.attr('array'),
  upgradingHosts: DS.attr('array'),
  upgradedHosts: DS.attr('array'),
  upgradeFailedHosts: DS.attr('array'),
  currentHosts: DS.attr('array'),

  noInstalledHosts:  function() {
    return this.get('installedHosts.length') == 0;
  }.property('installedHosts.length'),

  noCurrentHosts: function() {
    return this.get('currentHosts.length') == 0;
  }.property('currentHosts.length'),

  noInitHosts: function() {
    return this.get('notInstalledHosts.length') == 0;
  }.property('notInstalledHosts.length')
});

App.StackVersion.FIXTURES = [];


});

require.register("models/target_cluster", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.TargetCluster = DS.Model.extend({
  name: DS.attr('string'),
  execute: DS.attr('string'),
  workflow: DS.attr('string'),
  write: DS.attr('string'),
  readonly: DS.attr('string'),
  staging: DS.attr('string'),
  working: DS.attr('string'),
  temp: DS.attr('string')
});

App.TargetCluster.FIXTURES = [];
});

require.register("models/upgrade_entity", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
 * @type {Ember.Object}
 * @class
 */
App.upgradeEntity = Em.Object.extend({

  /**
   * type of entity "GROUP", "ITEM", "TASK"
   * @type {string}
   */
  type: null,

  /**
   * @type {boolean}
   */
  isExpanded: false,

  /**
   * @type {boolean}
   */
  isVisible: function () {
    return this.get('status') !== 'PENDING';
  }.property('status'),

  /**
   * status of tasks/items/groups which should be grayed out and disabled
   * @type {Array}
   */
  nonActiveStates: ['PENDING', 'ABORTED'],

  /**
   * @type {boolean}
   */
  isRunning: function () {
    return ['IN_PROGRESS'].contains(this.get('status'));
  }.property('status'),

  /**
   * @type {number}
   */
  progress: function () {
    return Math.floor(this.get('progress_percent'));
  }.property('progress_percent'),

  /**
   * indicate whether entity has active link
   * @type {boolean}
   */
  isActive: function () {
    return !this.get('nonActiveStates').contains(this.get('status'));
  }.property('status')
});
});

require.register("models/user", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var validator = require('utils/validator');

App.User = DS.Model.extend({
  userName:DS.attr('string'),
  id:function(){
    return this.get('userName');
  }.property('userName'),
  isLdap:DS.attr('boolean'),
  type: function(){
    if(this.get('isLdap')){
      return 'LDAP';
    }
    return 'Local';
  }.property('isLdap'),
  auditItems:DS.hasMany('App.ServiceAudit'),
  admin: DS.attr('boolean'),
  operator: DS.attr('boolean'),
  /**
   * List of permissions assigned to user
   *  Available permissions:
   *    AMBARI.ADMIN
   *    CLUSTER.READ
   *    CLUSTER.OPERATE
   *    VIEW.USE
   * @property {Array} permissions
   **/
  permissions: DS.attr('array')
});

App.EditUserForm = App.Form.extend({
  className:App.User,
  object:function () {
    return App.router.get('mainAdminUserEditController.content');
  }.property('App.router.mainAdminUserEditController.content'),

  fieldsOptions:[
    { name:"userName", displayName:"Username" },
    { name:"old_password", displayName:"Current Password", displayType:"password", isRequired: false },
    { name:"new_password", displayName:"New Password", displayType:"password",  isRequired: false },
    { name:"new_passwordRetype", displayName:"Retype New Password", displayType:"password", isRequired: false },
    { name:"admin", displayName:"Admin", displayType:"checkbox", isRequired:false },
    { name:"isLdap", displayName:"Type", isRequired:false, isHidden:true }
  ],
  fields:[],
  disableUsername:function () {
    this.getField("userName").set("disabled", "disabled");
  }.observes('object'),
  disableAdminCheckbox:function () {
    var object = this.get('object');
    if (object) {
      if (object.get('userName') == App.get('router').getLoginName()) {
        this.getField("admin").set("disabled", true);
      } else {
        this.getField("admin").set("disabled", false);
      }
    }
  }.observes('object'),

  isValid:function () {

    var isValid = this._super();
    var thisForm = this;

    var newPass = this.get('field.new_password');
    var oldPass = this.get('field.old_password');
    var passRetype = this.get('field.new_passwordRetype');

    if (!validator.empty(newPass.get('value'))) {
      if(validator.empty(oldPass.get('value'))){
        oldPass.set('errorMessage', this.t('admin.users.editError.requiredField'));
        isValid = false;
      }
      if (newPass.get('value') != passRetype.get('value')) {
        passRetype.set('errorMessage', this.t('admin.users.createError.passwordValidation'));
        isValid = false;
      }
    }

    return isValid;
  },

  save: function () {
    var object = this.get('object');
    var formValues = {};
    $.each(this.get('fields'), function () {
      formValues[this.get('name')] = this.get('value');
    });

    $.each(formValues, function (k, v) {
      object.set(k, v);
    });

    //App.store.commit();
    this.set('result', 1);

    return true;
  }
});
App.CreateUserForm = App.Form.extend({
  className:App.User,
  object:function () {
    return App.router.get('mainAdminUserCreateController.content');
  }.property('App.router.mainAdminUserCreateController.content'),

  fieldsOptions:[
    { name:"userName", displayName:"Username", toLowerCase: function(){var v = this.get('value'); this.set('value', v.toLowerCase())}.observes('value') },
    { name:"password", displayName:"Password", displayType:"password", isRequired: true },
    { name:"passwordRetype", displayName:"Retype Password", displayType:"password", validator:"passwordRetype", isRequired: true },
    { name:"admin", displayName:"Admin", displayType:"checkbox", isRequired:false, defaultValue: true}
  ],
  fields:[],

  isValid:function () {
    var isValid = this._super();

    var passField = this.get('field.password');
    var passRetype = this.get('field.passwordRetype');

    if (!validator.empty(passField.get('value'))) {
      if (passField.get('value') != passRetype.get('value')) {
        passRetype.set('errorMessage', this.t('admin.users.createError.passwordValidation'));
        isValid = false;
      }
    }

    if (isValid) {
      var users = App.User.find();
      var userNameField = this.getField('userName');
      var userName = userNameField.get('value');

      if (users.mapProperty('userName').contains(userName)) {
        userNameField.set('errorMessage', this.t('admin.users.createError.userNameExists'));
        return isValid = false;
      }
    }

    return isValid;
  },

  isWarn: function() {
    var isWarn = false;
    var userNameField = this.getField('userName');
    userNameField.set('warnMessage', '');
    var userName = userNameField.get('value');

    if (this.isValid() && !validator.isValidUserName(userName)) {
      userNameField.set('warnMessage', this.t('users.userName.validationFail'));
      isWarn = true;
    }
    return isWarn;
  },

  save: function () {

    var object = this.get('object');
    var formValues = {};
    $.each(this.get('fields'), function () {
      formValues[Ember.String.decamelize(this.get('name'))] = this.get('value');
    });

    if (this.get('className')) {
      App.store.load(this.get('className'), App.dateTime(), formValues);
    }
    else {
      console.log("Please define class name for your form " + this.constructor);
    }

    this.set('result', 1);

    return true;
  }
});
App.User.FIXTURES = [];


});

require.register("router", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.WizardRoute = Em.Route.extend({

  gotoStep0: Em.Router.transitionTo('step0'),

  gotoStep1: Em.Router.transitionTo('step1'),

  gotoStep2: Em.Router.transitionTo('step2'),

  gotoStep3: Em.Router.transitionTo('step3'),

  gotoStep4: Em.Router.transitionTo('step4'),

  gotoStep5: Em.Router.transitionTo('step5'),

  gotoStep6: Em.Router.transitionTo('step6'),

  gotoStep7: Em.Router.transitionTo('step7'),

  gotoStep8: Em.Router.transitionTo('step8'),

  gotoStep9: Em.Router.transitionTo('step9'),

  gotoStep10: Em.Router.transitionTo('step10'),

  isRoutable: function() {
    return (typeof this.get('route') === 'string' && App.router.get('loggedIn'));
  }.property('App.router.loggedIn')

});

App.Router = Em.Router.extend({

  enableLogging: true,
  isFwdNavigation: true,
  backBtnForHigherStep: false,

  /**
   * Is true, if cluster.provisioning_state is equal to 'INSTALLED'
   * @type {Boolean}
   */
  clusterInstallCompleted: false,
  /**
   * user prefered path to route
   */
  preferedPath: null,

  setNavigationFlow: function (step) {
    var matches = step.match(/\d+$/);
    var newStep;
    if (matches) {
      newStep = parseInt(matches[0]);
    }
    var previousStep = parseInt(this.getInstallerCurrentStep());
    this.set('isFwdNavigation', newStep >= previousStep);
  },


  clearAllSteps: function () {
    this.get('installerController').clear();
    this.get('addHostController').clear();
    this.get('addServiceController').clear();
    this.get('backgroundOperationsController').clear();
    for (var i = 1; i < 11; i++) {
      this.set('wizardStep' + i + 'Controller.hasSubmitted', false);
      this.set('wizardStep' + i + 'Controller.isDisabled', true);
    }
  },

  /**
   * Temporary fix for getting cluster name
   * @return {*}
   */

  getClusterName: function () {
    return App.router.get('clusterController').get('clusterName');
  },


  /**
   * Get current step of Installer wizard
   * @return {*}
   */
  getInstallerCurrentStep: function () {
    return this.getWizardCurrentStep('installer');
  },

  /**
   * Get current step for <code>wizardType</code> wizard
   * @param wizardType one of <code>installer</code>, <code>addHost</code>, <code>addServices</code>
   */
  getWizardCurrentStep: function (wizardType) {
    var loginName = this.getLoginName();
    var currentStep = App.db.getWizardCurrentStep(wizardType);
    console.log('getWizardCurrentStep: loginName=' + loginName + ", currentStep=" + currentStep);
    if (!currentStep) {
      currentStep = wizardType === 'installer' ? '0' : '1';
    }
    console.log('returning currentStep=' + currentStep);
    return currentStep;
  },

  loggedIn: App.db.getAuthenticated(),

  loginName: function() {
    return this.getLoginName();
  }.property('loggedIn'),

  getAuthenticated: function () {
    var dfd = $.Deferred();
    var self = this;
    var auth = App.db.getAuthenticated();
    var authResp = (auth && auth === true);
    if (authResp) {
      App.ajax.send({
        name: 'router.login.clusters',
        sender: this,
        success: 'onAuthenticationSuccess',
        error: 'onAuthenticationError'
      }).complete(function () {
          dfd.resolve(self.get('loggedIn'));
        });
    } else {
      this.set('loggedIn', false);
      dfd.resolve(false);
    }
    return dfd.promise();
  },

  onAuthenticationSuccess: function (data) {
    this.setAuthenticated(true);
    if (data.items.length) {
      this.setClusterInstalled(data);
    }
  },

  onAuthenticationError: function (data) {
    if (data.status === 403) {
      this.setAuthenticated(false);
    } else {
      console.log('error in getAuthenticated');
    }
  },

  setAuthenticated: function (authenticated) {
    console.log("TRACE: Entering router:setAuthenticated function");
    App.db.setAuthenticated(authenticated);
    this.set('loggedIn', authenticated);
  },

  getLoginName: function () {
    return App.db.getLoginName();
  },

  setLoginName: function (loginName) {
    App.db.setLoginName(loginName);
  },

  /**
   * Set user model to local storage
   * @param user
   */
  setUser: function (user) {
    App.db.setUser(user);
  },

  /**
   * Get user model from local storage
   * @return {*}
   */
  getUser: function () {
    return App.db.getUser();
  },

  setUserLoggedIn: function(userName) {
    this.setAuthenticated(true);
    this.setLoginName(userName);
    this.setUser(App.User.find().findProperty('id', userName));
  },

  /**
   * Set `clusterInstallCompleted` property based on cluster info response.
   *
   * @param {Object} clusterObject
   **/
  setClusterInstalled: function(clusterObject) {
    this.set('clusterInstallCompleted', clusterObject.items[0].Clusters.provisioning_state === 'INSTALLED')
  },

  login: function () {
    var controller = this.get('loginController');
    var loginName = controller.get('loginName').toLowerCase();
    controller.set('loginName', loginName);
    var hash = window.btoa(loginName + ":" + controller.get('password'));
    var usr = '';

    if (App.get('testMode')) {
      if (loginName === "admin" && controller.get('password') === 'admin') {
        usr = 'admin';
      } else if (loginName === 'user' && controller.get('password') === 'user') {
        usr = 'user';
      }
    }

    App.ajax.send({
      name: 'router.login',
      sender: this,
      data: {
        auth: "Basic " + hash,
        usr: usr,
        loginName: encodeURIComponent(loginName)
      },
      beforeSend: 'authBeforeSend',
      success: 'loginSuccessCallback',
      error: 'loginErrorCallback'
    });

  },

  authBeforeSend: function(opt, xhr, data) {
    xhr.setRequestHeader("Authorization", data.auth);
  },

  loginSuccessCallback: function(data, opt, params) {
    console.log('login success');
    App.usersMapper.map({"items": [data]});
    this.setUserLoggedIn(params.loginName);
    App.router.get('mainViewsController').loadAmbariViews();
    App.ajax.send({
      name: 'router.login.clusters',
      sender: this,
      data: {
        loginName: params.loginName,
        loginData: data
      },
      success: 'loginGetClustersSuccessCallback',
      error: 'loginGetClustersErrorCallback'
    });
  },

  loginErrorCallback: function(request, ajaxOptions, error, opt) {
    var controller = this.get('loginController');
    console.log("login error: " + error);
    this.setAuthenticated(false);
    if (request.status == 403) {
      var responseMessage = request.responseText;
      try{
        responseMessage = JSON.parse(request.responseText).message;
      }catch(e){}
      controller.postLogin(true, false, responseMessage);
    } else {
      controller.postLogin(false, false, null);
    }

  },

  loginGetClustersSuccessCallback: function (clustersData, opt, params) {
    var adminViewUrl = '/views/ADMIN_VIEW/2.0.0/INSTANCE/#/';
    //TODO: Replace hard coded value with query. Same in templates/application.hbs
    var loginController = this.get('loginController');
    var loginData = params.loginData;
    var privileges = loginData.privileges || [];
    var router = this;
    var permissionList = privileges.mapProperty('PrivilegeInfo.permission_name');
      var isAdmin = permissionList.contains('AMBARI.ADMIN');
      var transitionToApp = false;
      if (isAdmin) {
        App.set('isAdmin', true);
        if (clustersData.items.length) {
          router.setClusterInstalled(clustersData);
          transitionToApp = true;
        } else {
          window.location = adminViewUrl;
          return;
        }
      } else {
        if (clustersData.items.length) {
          router.setClusterInstalled(clustersData);
          //TODO: Iterate over clusters
          var clusterName = clustersData.items[0].Clusters.cluster_name;
          var clusterPermissions = privileges.filterProperty('PrivilegeInfo.cluster_name', clusterName).mapProperty('PrivilegeInfo.permission_name');
          if (clusterPermissions.contains('CLUSTER.OPERATE')) {
            App.set('isAdmin', true);
            App.set('isOperator', true);
            transitionToApp = true;
          } else if (clusterPermissions.contains('CLUSTER.READ')) {
            transitionToApp = true;
          }
        }
      }
      if (transitionToApp) {
        if (!Em.isNone(router.get('preferedPath'))) {
          window.location = router.get('preferedPath');
          router.set('preferedPath', null);
        } else {
          router.getSection(function (route) {
            router.transitionTo(route);
            loginController.postLogin(true, true);
          });
        }
      } else {
        router.transitionTo('main.views.index');
        loginController.postLogin(true,true);
      }
  },

  loginGetClustersErrorCallback: function (req) {
    console.log("Get clusters error: " + req.statusCode);
  },

  getSection: function (callback) {
    if (App.get('testMode')) {
      if (App.alwaysGoToInstaller) {
        callback('installer');
      } else {
        callback('main.dashboard.index');
      }
    } else {
      if (this.get('clusterInstallCompleted')) {
        App.clusterStatus.updateFromServer(false).complete(function () {
          var clusterStatusOnServer = App.clusterStatus.get('value');
          var route = 'main.dashboard.index';
          if (clusterStatusOnServer && clusterStatusOnServer.wizardControllerName === App.router.get('addHostController.name')) {
            // if wizardControllerName == "addHostController", then it means someone closed the browser or the browser was crashed when we were last in Add Hosts wizard
            route = 'main.hostAdd';
          } else if (clusterStatusOnServer && (clusterStatusOnServer.wizardControllerName === App.router.get('addSecurityController.name') || clusterStatusOnServer.wizardControllerName === App.router.get('mainAdminSecurityDisableController.name'))) {
            // if wizardControllerName == "addSecurityController", then it means someone closed the browser or the browser was crashed when we were last in Add Security wizard
            route = 'main.admin.adminSecurity';
          } else if (clusterStatusOnServer && (clusterStatusOnServer.wizardControllerName === App.router.get('kerberosWizardController.name'))) {
            // if wizardControllerName == "adminKerberosController", then it means someone closed the browser or the browser was crashed when we were last in Add Kerberos wizard
            route = 'main.admin.adminKerberos.adminAddKerberos';
          } else if (clusterStatusOnServer && clusterStatusOnServer.wizardControllerName === App.router.get('addServiceController.name')) {
            // if wizardControllerName == "addHostController", then it means someone closed the browser or the browser was crashed when we were last in Add Hosts wizard
            route = 'main.serviceAdd';
          } else if (clusterStatusOnServer && clusterStatusOnServer.wizardControllerName === App.router.get('reassignMasterController.name')) {
            // if wizardControllerName == "reassignMasterController", then it means someone closed the browser or the browser was crashed when we were last in Reassign Master wizard
            route = 'main.reassign';
          } else if (clusterStatusOnServer && clusterStatusOnServer.wizardControllerName === App.router.get('highAvailabilityWizardController.name')) {
            // if wizardControllerName == "highAvailabilityWizardController", then it means someone closed the browser or the browser was crashed when we were last in NameNode High Availability wizard
            route = 'main.services.enableHighAvailability';
          } else if (clusterStatusOnServer && clusterStatusOnServer.wizardControllerName === App.router.get('rMHighAvailabilityWizardController.name')) {
            // if wizardControllerName == "highAvailabilityWizardController", then it means someone closed the browser or the browser was crashed when we were last in NameNode High Availability wizard
            route = 'main.services.enableRMHighAvailability';
          } else if (clusterStatusOnServer && clusterStatusOnServer.wizardControllerName === App.router.get('rollbackHighAvailabilityWizardController.name')) {
            // if wizardControllerName == "highAvailabilityRollbackController", then it means someone closed the browser or the browser was crashed when we were last in NameNode High Availability Rollback wizard
            route = 'main.services.rollbackHighAvailability';
          } else if (clusterStatusOnServer && clusterStatusOnServer.wizardControllerName === App.router.get('mainAdminStackAndUpgradeController.name')) {
            // if wizardControllerName == "mainAdminStackAndUpgradeController", then it means someone closed the browser or the browser was crashed when we were last in Rolling Upgrade wizard
            route = 'main.admin.stackAndUpgrade';
          }
          callback(route);
        });
      } else {
        callback('installer');
      }
    }
  },

  logOff: function (context) {
    $('title').text(Em.I18n.t('app.name'));
    var hash = window.btoa(this.get('loginController.loginName') + ":" + this.get('loginController.password'));

    App.router.get('mainController').stopPolling();
    // App.db.cleanUp() must be called before router.clearAllSteps().
    // otherwise, this.set('installerController.currentStep, 0) would have no effect
    // since it's a computed property but we are not setting it as a dependent of App.db.
    App.db.cleanUp();
    App.set('isAdmin', false);
    App.set('isOperator', false);
    this.set('loggedIn', false);
    this.clearAllSteps();
    console.log("Log off: " + App.router.getClusterName());
    this.set('loginController.loginName', '');
    this.set('loginController.password', '');
    // When logOff is called by Sign Out button, context contains event object. As it is only case we should send logoff request, we are checking context below.
    if (!App.get('testMode') && context) {
      App.ajax.send({
        name: 'router.logoff',
        sender: this,
        data: {
          auth: "Basic " + hash
        },
        beforeSend: 'authBeforeSend',
        success: 'logOffSuccessCallback',
        error:'logOffErrorCallback'
      });
    }
    if (App.router.get('clusterController.isLoaded')) {
      window.location.reload();
    } else {
      this.transitionTo('login', context);
    }
  },

  logOffSuccessCallback: function (data) {
    console.log("invoked logout on the server successfully");
    var applicationController = App.router.get('applicationController');
    applicationController.set('isPollerRunning',false);
  },

  logOffErrorCallback: function (req) {
    console.log("failed to invoke logout on the server");
  },

  /**
   * initialize isAdmin if user is administrator
   */
  initAdmin: function(){
    if (App.db) {
      var user = App.db.getUser();
      if (user) {
        if (user.admin) {
          App.set('isAdmin', true);
          console.log('Administrator logged in');
        }
        if (user.operator) {
          App.set('isOperator', true);
        }
      }
    }
  },

  root: Em.Route.extend({
    index: Em.Route.extend({
      route: '/',
      redirectsTo: 'login'
    }),

    enter: function(router){
      router.initAdmin();
    },

    login: Em.Route.extend({
      route: '/login',

      /**
       *  If the user is already logged in, redirect to where the user was previously
       */
      enter: function (router, context) {
        router.getAuthenticated().done(function (loggedIn) {
          var location = router.location.location.hash;
          //key to parse URI for prefered path to route
          var key = '?targetURI=';

          if (loggedIn) {
            Ember.run.next(function () {
              console.log(router.getLoginName() + ' already authenticated.  Redirecting...');
              router.getSection(function (route) {
                router.transitionTo(route, context);
              });
            });
          } else {
            if (location.contains(key)) {
              router.set('preferedPath', location.slice(location.indexOf(key) + key.length));
            }
          }
        });
      },

      connectOutlets: function (router, context) {
        $('title').text(Em.I18n.t('app.name'));
        console.log('/login:connectOutlet');
        console.log('currentStep is: ' + router.getInstallerCurrentStep());
        console.log('authenticated is: ' + router.getAuthenticated());
        router.get('applicationController').connectOutlet('login');
      }
    }),

    installer: require('routes/installer'),

    main: require('routes/main'),

    adminView: Em.Route.extend({
      route: '/adminView',
      enter: function (router) {
        if (!router.get('loggedIn') || !App.isAccessible('upgrade_ADMIN') || App.isAccessible('upgrade_OPERATOR')) {
          Em.run.next(function () {
            router.transitionTo('login');
          });
        } else {
            window.location.replace('/views/ADMIN_VIEW/2.0.0/INSTANCE/#/');
        }
      }
    }),

    experimental: Em.Route.extend({
      route: '/experimental',
      enter: function (router, context) {
        if (App.isAccessible('upgrade_OPERATOR')) {
          Em.run.next(function () {
            if (router.get('clusterInstallCompleted')) {
              router.transitionTo("main.dashboard.widgets");
            } else {
              router.route("installer");
            }
          });
        } else if (!App.isAccessible('upgrade_ADMIN')) {
          Em.run.next(function () {
            router.transitionTo("main.views.index");
          });
        }
      },
      connectOutlets: function (router, context) {
        if (App.isAccessible('upgrade_ONLY_ADMIN')) {
          $('title').text(Em.I18n.t('app.name.subtitle.experimental'));
          console.log('/experimental:connectOutlet');
          router.get('applicationController').connectOutlet('experimental');
        }
      }
    }),

    logoff: function (router, context) {
      router.logOff(context);
    }

  })
});

});

require.register("routes/add_alert_definition_routes", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

module.exports = App.WizardRoute.extend({

  route: '/alerts/add',

  enter: function (router) {
    if (App.isAccessible('ADMIN')) {
      Em.run.next(function () {
        var addAlertDefinitionController = router.get('addAlertDefinitionController');
        App.router.get('updateController').set('isWorking', false);
        var popup = App.ModalPopup.show({
          classNames: ['full-width-modal'],
          header: Em.I18n.t('alerts.add.header'),
          bodyClass: App.AddAlertDefinitionView.extend({
            controllerBinding: 'App.router.addAlertDefinitionController'
          }),
          primary: Em.I18n.t('form.cancel'),
          showFooter: false,
          secondary: null,

          onPrimary: function () {
            this.hide();
            App.router.transitionTo('main.alerts.index');
          },
          onClose: function () {
            this.set('showCloseButton', false); // prevent user to click "Close" many times
            App.router.get('updateController').set('isWorking', true);
            router.transitionTo('main.alerts.index');
            this.hide();
          },
          didInsertElement: function () {
            this.fitHeight();
          }
        });
        addAlertDefinitionController.set('popup', popup);
        router.transitionTo('step' + addAlertDefinitionController.get('currentStep'));
      });
    } else {
      Em.run.next(function () {
        App.router.transitionTo('main.alerts');
      });
    }
  },

  step1: Em.Route.extend({

    route: '/step1',

    connectOutlets: function (router) {
      var controller = router.get('addAlertDefinitionController');
      controller.setCurrentStep('1');
      controller.set('content', controller.getDBProperty('content'));
      controller.connectOutlet('addAlertDefinitionStep1', controller.get('content'));
    },

    next: function (router) {
      var controller = router.get('addAlertDefinitionController');
      controller.setDBProperty('content', controller.get('content'));
      router.transitionTo('step2');
    }

  }),

  step2: Em.Route.extend({

    route: '/step2',

    connectOutlets: function (router) {
      var controller = router.get('addAlertDefinitionController');
      controller.setCurrentStep('2');
      controller.set('content', controller.getDBProperty('content'));
      controller.connectOutlet('addAlertDefinitionStep2', controller.get('content'));
    },

    back: Em.Router.transitionTo('step1'),

    next: function (router) {
      var controller = router.get('addAlertDefinitionController');
      controller.set('content.configs', App.router.get('mainAlertDefinitionConfigsController.configs'));
      var newDefinitionData = App.router.get('mainAlertDefinitionConfigsController').getPropertiesToUpdate(false);
      newDefinitionData['AlertDefinition/source'].type = controller.get('content.selectedType');
      newDefinitionData['AlertDefinition/label'] = newDefinitionData['AlertDefinition/name'];
      newDefinitionData['AlertDefinition/name'] = newDefinitionData['AlertDefinition/name'].toLowerCase().replace(/\s+/g, '_');
      controller.set('content.formattedToRequestConfigs', newDefinitionData);
      controller.setDBProperty('content', controller.get('content'));
      router.transitionTo('step3');
    }

  }),

  step3: Em.Route.extend({

    route: '/step3',

    connectOutlets: function (router) {
      var controller = router.get('addAlertDefinitionController');
      controller.setCurrentStep('3');
      controller.set('content', controller.getDBProperty('content'));
      controller.connectOutlet('addAlertDefinitionStep3', controller.get('content'));
    },

    back: Em.Router.transitionTo('step2'),

    done: function (router) {
      var controller = router.get('addAlertDefinitionController');
      controller.createNewAlertDefinition(Em.get(controller.getDBProperty('content'), 'formattedToRequestConfigs')).done(function () {
        controller.get('popup').hide();
        controller.setDBProperty('content', {});
        controller.finish();
        App.clusterStatus.setClusterStatus({
            clusterName: controller.get('content.cluster.name'),
            clusterState: 'DEFAULT',
            localdb: App.db.data
          },
          {
            alwaysCallback: function () {
              controller.get('popup').hide();
              router.transitionTo('main.alerts');
              location.reload();
            }
          });
      });
    }
  })

});

});

require.register("routes/add_host_routes", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

module.exports = App.WizardRoute.extend({
  route: '/host/add',

  leaveWizard: function (router,context) {
    var addHostController = router.get('addHostController');
    App.router.get('updateController').set('isWorking', true);
    addHostController.finish();
    App.clusterStatus.setClusterStatus({
      clusterName: App.router.get('content.cluster.name'),
      clusterState: 'DEFAULT',
      localdb: App.db.data
    }, {alwaysCallback: function() {context.hide();App.router.transitionTo('hosts.index');location.reload();}});
  },

  enter: function (router) {
    console.log('in /host/add:enter');

    var self = this;

    Ember.run.next(function () {
      var addHostController = router.get('addHostController');
      App.router.get('updateController').set('isWorking', false);
      App.ModalPopup.show({
        classNames: ['full-width-modal'],
        header:Em.I18n.t('hosts.add.header'),
        bodyClass:  App.AddHostView.extend({
          controllerBinding: 'App.router.addHostController'
        }),
        primary:Em.I18n.t('form.cancel'),
        secondary: null,
        showFooter: false,

        onPrimary:function () {
          this.hide();
          App.router.get('updateController').set('isWorking', true);
          router.transitionTo('hosts.index');
        },
        onClose: function() {
          var popupContext = this;
          if (addHostController.get('currentStep') == '6') {
            App.ModalPopup.show({
              header: Em.I18n.t('hosts.add.exit.header'),
              body: Em.I18n.t('hosts.add.exit.body'),
              onPrimary: function () {
                self.leaveWizard(router,popupContext);
              }
            });
          } else {
            self.leaveWizard(router,this);
          }
        },
        didInsertElement: function(){
          this.fitHeight();
        }
      });
      var currentClusterStatus = App.clusterStatus.get('value');
      if (currentClusterStatus) {
        switch (currentClusterStatus.clusterState) {
          case 'ADD_HOSTS_DEPLOY_PREP_2' :
            addHostController.setCurrentStep('4');
            break;
          case 'ADD_HOSTS_INSTALLING_3' :
          case 'SERVICE_STARTING_3' :
            addHostController.setCurrentStep('5');
            break;
          case 'ADD_HOSTS_INSTALLED_4' :
            addHostController.setCurrentStep('6');
            break;
          default:
            break;
        }
      }

      router.transitionTo('step' + addHostController.get('currentStep'));
    });

  },

  step1: Em.Route.extend({
    route: '/step1',
    connectOutlets: function (router) {
      console.log('in addHost.step1:connectOutlets');
      var controller = router.get('addHostController');
      controller.setCurrentStep('1');
      controller.set('hideBackButton', true);
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        var wizardStep2Controller = router.get('wizardStep2Controller');
        wizardStep2Controller.set('wizardController', controller);
        controller.connectOutlet('wizardStep2', controller.get('content'));
      });
    },

    next: function (router) {
      var controller = router.get('addHostController');
      controller.save('installOptions');
      //hosts was saved to content.hosts inside wizardStep2Controller
      controller.save('hosts');
      router.transitionTo('step2');
      controller.setDBProperty('bootStatus', false);
    },
    evaluateStep: function (router) {
      console.log('in addHost.step1:evaluateStep');
      var addHostController = router.get('addHostController');
      var wizardStep2Controller = router.get('wizardStep2Controller');

      wizardStep2Controller.set('hasSubmitted', true);

      if (!wizardStep2Controller.get('isSubmitDisabled')) {
        wizardStep2Controller.evaluateStep();
      }
    }
  }),

  step2: Em.Route.extend({
    route: '/step2',
    connectOutlets: function (router) {
      console.log('in addHost.step2:connectOutlets');
      var controller = router.get('addHostController');
      controller.setCurrentStep('2');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        var wizardStep3Controller = router.get('wizardStep3Controller');
        wizardStep3Controller.set('wizardController', controller);
        controller.connectOutlet('wizardStep3', controller.get('content'));
      });
    },
    back: function(router){
      router.transitionTo('step1');
    },
    exit: function (router) {
      router.get('wizardStep3Controller').set('stopBootstrap', true);
    },
    next: function (router, context) {
      var addHostController = router.get('addHostController');
      var wizardStep3Controller = router.get('wizardStep3Controller');
      var wizardStep6Controller = router.get('wizardStep6Controller');
      addHostController.saveConfirmedHosts(wizardStep3Controller);
      addHostController.saveClients();

      addHostController.setDBProperty('bootStatus', true);
      addHostController.setDBProperty('slaveComponentHosts', undefined);
      wizardStep6Controller.set('isClientsSet', false);
      router.transitionTo('step3');
    },
    /**
     * Wrapper for remove host action.
     * Since saving data stored in addHostController, we should call this from router
     * @param router
     * @param context Array of hosts to delete
     */
    removeHosts: function (router, context) {
      console.log('in addHost.step2.removeHosts:hosts to delete ', context);
      var controller = router.get('addHostController');
      controller.removeHosts(context);
    }
  }),

  step3: Em.Route.extend({
    route: '/step3',
    connectOutlets: function (router, context) {
      console.log('in addHost.step3:connectOutlets');
      var controller = router.get('addHostController');
      controller.setCurrentStep('3');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        var wizardStep6Controller = router.get('wizardStep6Controller');
        wizardStep6Controller.set('wizardController', controller);
        controller.connectOutlet('wizardStep6', controller.get('content'));
      });
    },
    back: Em.Router.transitionTo('step2'),
    next: function (router) {
      var addHostController = router.get('addHostController');
      var wizardStep6Controller = router.get('wizardStep6Controller');

      wizardStep6Controller.callValidation(function() {
        wizardStep6Controller.showValidationIssuesAcceptBox(function() {
          addHostController.saveSlaveComponentHosts(wizardStep6Controller);
          router.transitionTo('step4');
        });
      });
    }
  }),

  step4: Em.Route.extend({
    route: '/step4',
    connectOutlets: function (router, context) {
      console.log('in addHost.step4:connectOutlets');
      var controller = router.get('addHostController');
      var addHostStep4Controller = router.get('addHostStep4Controller');
      controller.setCurrentStep('4');
      controller.dataLoading().done(function () {
        addHostStep4Controller.loadConfigGroups();
        addHostStep4Controller.set('isConfigGroupLoaded', false);
        addHostStep4Controller.configGroupsLoading().done(function () {
          controller.loadAllPriorSteps();
          controller.loadServiceConfigGroups();
          addHostStep4Controller.set('wizardController', controller);
          controller.connectOutlet('addHostStep4', controller.get('content'));
        });
      });
    },
    back: function(router){
      router.transitionTo('step3');
    },
    next: function (router) {
      var addHostController = router.get('addHostController');
      addHostController.saveServiceConfigGroups();
      router.transitionTo('step5');
    }
  }),

  step5: Em.Route.extend({
    route: '/step5',
    connectOutlets: function (router, context) {
      console.log('in addHost.step5:connectOutlets');
      var controller = router.get('addHostController');
      controller.setCurrentStep('5');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.getServiceConfigGroups();
        var wizardStep8Controller = router.get('wizardStep8Controller');
        wizardStep8Controller.set('wizardController', controller);
        controller.connectOutlet('wizardStep8', controller.get('content'));
      });
    },
    back: function(router){
      if(!router.get('wizardStep8Controller.isBackBtnDisabled')) {
        router.transitionTo('step4');
      }
    },
    next: function (router) {
      var addHostController = router.get('addHostController');
      var wizardStep8Controller = router.get('wizardStep8Controller');
      addHostController.applyConfigGroup();
      addHostController.installServices(false, function () {
        addHostController.setInfoForStep9();
        // We need to do recovery based on whether we are in Add Host or Installer wizard
        addHostController.saveClusterState('ADD_HOSTS_INSTALLING_3');
        wizardStep8Controller.set('servicesInstalled', true);
        router.transitionTo('step6');
      });
    }
  }),

  step6: Em.Route.extend({
    route: '/step6',
    connectOutlets: function (router, context) {
      console.log('in addHost.step6:connectOutlets');
      var controller = router.get('addHostController');
      controller.setCurrentStep('6');
      controller.dataLoading().done(function () {
        var wizardStep9Controller = router.get('wizardStep9Controller');
        wizardStep9Controller.set('wizardController', controller);
        controller.loadAllPriorSteps();
        if (!App.get('testMode')) {              //if test mode is ON don't disable prior steps link.
          controller.setLowerStepsDisable(6);
        }
        controller.connectOutlet('wizardStep9', controller.get('content'));
      });
    },
    back: Em.Router.transitionTo('step5'),
    retry: function(router,context) {
      var addHostController = router.get('addHostController');
      var wizardStep9Controller = router.get('wizardStep9Controller');
      wizardStep9Controller.set('wizardController', addHostController);
      if (wizardStep9Controller.get('showRetry')) {
        if (wizardStep9Controller.get('content.cluster.status') === 'INSTALL FAILED') {
          var isRetry = true;
          addHostController.installServices(isRetry, function () {
            addHostController.setInfoForStep9();
            wizardStep9Controller.resetHostsForRetry();
            // We need to do recovery based on whether we are in Add Host or Installer wizard
            addHostController.saveClusterState('ADD_HOSTS_INSTALLING_3');
            wizardStep9Controller.navigateStep();
          });
        } else {
          wizardStep9Controller.navigateStep();
        }
      }
    },
    unroutePath: function() {
      return false;
    },
    next: function (router) {
      var addHostController = router.get('addHostController');
      var wizardStep9Controller = router.get('wizardStep9Controller');
      wizardStep9Controller.set('wizardController', addHostController);
      addHostController.saveInstalledHosts(wizardStep9Controller);

      // We need to do recovery based on whether we are in Add Host or Installer wizard
      addHostController.saveClusterState('ADD_HOSTS_INSTALLED_4');
      router.transitionTo('step7');
    }
  }),

  step7: Em.Route.extend({
    route: '/step7',
    connectOutlets: function (router, context) {
      console.log('in addHost.step7:connectOutlets');
      var controller = router.get('addHostController');
      controller.setCurrentStep('7');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        var wizardStep10Controller = router.get('wizardStep10Controller');
        wizardStep10Controller.set('wizardController', controller);
        if (!App.get('testMode')) {              //if test mode is ON don't disable prior steps link.
          controller.setLowerStepsDisable(7);
        }
        controller.connectOutlet('wizardStep10', controller.get('content'));
        router.get('updateController').set('isWorking', true);
      });
    },
    back: Em.Router.transitionTo('step6'),
    complete: function (router, context) {
      var addHostController = router.get('addHostController');
      var hostsUrl = '/hosts?fields=Hosts/host_name,Hosts/public_host_name,Hosts/cpu_count,Hosts/total_mem,' +
        'Hosts/host_status,Hosts/last_heartbeat_time,Hosts/os_arch,Hosts/os_type,Hosts/ip,host_components,' +
        'metrics/disk,metrics/load/load_one,metrics/cpu/cpu_system,metrics/cpu/cpu_user,metrics/memory/mem_total,metrics/memory/mem_free';
      router.get('clusterController').requestHosts(hostsUrl, function () {
        console.log('Request for hosts, with immutable parameters');
      });
      router.get('updateController').updateAll();
      $(context.currentTarget).parents("#modal").find(".close").trigger('click');
    }
  }),

  backToHostsList: function (router, event) {
    App.router.get('updateController').set('isWorking', true);
    router.transitionTo('hosts.index');
  }


});

});

require.register("routes/add_kerberos_routes", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');

module.exports = App.WizardRoute.extend({
  route: '/enable',
  enter: function (router) {
    router.get('mainController').dataLoading().done(function () {
      var kerberosWizardController = router.get('kerberosWizardController');
      App.router.get('updateController').set('isWorking', false);
      var popup = App.ModalPopup.show({
        classNames: ['full-width-modal'],
        header: Em.I18n.t('admin.kerberos.wizard.header'),
        bodyClass: App.KerberosWizardView.extend({
          controller: kerberosWizardController
        }),
        primary: Em.I18n.t('form.cancel'),
        showFooter: false,
        secondary: null,

        onClose: function () {
          var self = this;
          switch (kerberosWizardController.get('currentStep')) {
            case "2":
              var step2Controller = router.get('kerberosWizardStep2Controller');
              if (step2Controller.get('testConnectionInProgress')) {
                step2Controller.showConnectionInProgressPopup(function () {
                  kerberosWizardController.warnBeforeExitPopup(function () {
                    self.exitWizard();
                  }, false);
                });
              } else {
                kerberosWizardController.warnBeforeExitPopup(function () {
                  self.exitWizard();
                }, false);
              }
              break;
            case "6":
            case "7":
              kerberosWizardController.warnBeforeExitPopup(function () {
                self.exitWizard();
              }, true);
              break;
            default:
              kerberosWizardController.warnBeforeExitPopup(function () {
                self.exitWizard();
              }, false);
          }
        },
        didInsertElement: function () {
          this.fitHeight();
        },

        exitWizard: function () {
          var self = this;
          var kerberosProgressPageController = App.router.get('kerberosProgressPageController');
          var controller = App.router.get('kerberosWizardController');
          var exitPath = controller.getDBProperty('onClosePath') || 'adminKerberos.index';
          controller.clearTasksData();
          controller.finish();
          App.get('router.updateController').set('isWorking', true);
          if (App.get('testMode')) {
            App.get('router').transitionTo('adminKerberos.index');
            location.reload();
          }
          App.clusterStatus.setClusterStatus({
            clusterName: App.router.getClusterName(),
            clusterState: 'DEFAULT',
            localdb: App.db.data
          }, {
            alwaysCallback: function () {
              self.hide();
              App.get('router').transitionTo(exitPath);
              location.reload();
            }
          });
        }
      });
      kerberosWizardController.set('popup', popup);
      var currentClusterStatus = App.clusterStatus.get('value');
      if (currentClusterStatus) {
        if (App.get('testMode')) {
          kerberosWizardController.setCurrentStep(App.db.data.KerberosWizard.currentStep);
        } else {
          switch (currentClusterStatus.clusterState) {
            case 'KERBEROS_DEPLOY' :
              kerberosWizardController.setCurrentStep(currentClusterStatus.localdb.KerberosWizard.currentStep);
              break;
            default:
              var currStep = App.get('router.kerberosWizardController.currentStep');
              kerberosWizardController.setCurrentStep(currStep);
              break;
          }
        }

      }
      Em.run.next(function(){
        router.transitionTo('step' + kerberosWizardController.get('currentStep'));
      });
    });
  },

  step1: Em.Route.extend({
    route: '/step1',

    connectOutlets: function (router) {
      console.log('in addSecurity.step1:connectOutlets');
      var controller = router.get('kerberosWizardController');
      controller.dataLoading().done(function () {
        router.get('kerberosWizardController').setCurrentStep('1');
        controller.loadAllPriorSteps();
        controller.connectOutlet('kerberosWizardStep1', controller.get('content'));
      });
    },

    unroutePath: function () {
      return false;
    },

    next: function (router) {
      var kerberosWizardController = router.get('kerberosWizardController');
      var kerberosStep1controller = router.get('kerberosWizardStep1Controller');
      kerberosWizardController.saveKerberosOption(kerberosStep1controller);
      kerberosWizardController.setDBProperty('serviceConfigProperties', null);
      kerberosWizardController.setDBProperty('advancedServiceConfig', null);
      router.transitionTo('step2');
    }
  }),

  step2: Em.Route.extend({
    route: '/step2',

    connectOutlets: function (router) {
      console.log('in kerberosWizardController.step2:connectOutlets');
      var controller = router.get('kerberosWizardController');
      controller.dataLoading().done(function () {
        router.get('kerberosWizardController').setCurrentStep('2');
        controller.loadAllPriorSteps();
        var kerberosWizardStep2Controller = router.get('kerberosWizardStep2Controller');
        kerberosWizardStep2Controller.set('wizardController', controller);
        controller.connectOutlet('kerberosWizardStep2', controller.get('content'));
      });
    },
    unroutePath: function () {
      return false;
    },
    back: function(router) {
      var controller = router.get('kerberosWizardStep2Controller');
      if (!controller.get('isBackBtnDisabled')) {
        router.transitionTo('step1')
      }
    },

    next: function (router) {
      var kerberosWizardController = router.get('kerberosWizardController');
      var kerberosWizardStep2Controller = router.get('kerberosWizardStep2Controller');
      kerberosWizardController.saveServiceConfigProperties(kerberosWizardStep2Controller);
      kerberosWizardController.clearTasksData();
      router.transitionTo('step3');
    }
  }),

  step3: Em.Route.extend({
    route: '/step3',

    connectOutlets: function (router) {
      console.log('in kerberosWizardController.step3:connectOutlets');
      var controller = router.get('kerberosWizardController');
      controller.dataLoading().done(function () {
        router.get('kerberosWizardController').setCurrentStep('3');
        controller.loadAllPriorSteps();
        controller.connectOutlet('kerberosWizardStep3', controller.get('content'));
      });
    },
    unroutePath: function () {
      return false;
    },
    back: Em.Router.transitionTo('step2'),
    next: function (router) {
      var kerberosWizardController = router.get('kerberosWizardController');
      kerberosWizardController.setDBProperty('kerberosDescriptorConfigs', null);
      router.transitionTo('step4');
    }
  }),

  step4: Em.Route.extend({
    route: '/step4',

    connectOutlets: function (router) {
      console.log('in kerberosWizardController.step4:connectOutlets');
      var controller = router.get('kerberosWizardController');
      var step4Controller = router.get('kerberosWizardStep4Controller');
      controller.setCurrentStep(4);
      controller.dataLoading().done(function () {
        router.get('kerberosWizardController').setCurrentStep('4');
        controller.loadAllPriorSteps();
        controller.setLowerStepsDisable(4);
        step4Controller.set('wizardController', controller);
        controller.connectOutlet('kerberosWizardStep4', controller.get('content'));
      });
    },
    unroutePath: function () {
      return false;
    },
    back: Em.Router.transitionTo('step3'),
    next: function (router) {
      var kerberosWizardController = router.get('kerberosWizardController');
      var kerberosWizardStep4Controller = router.get('kerberosWizardStep4Controller');
      kerberosWizardController.saveServiceConfigProperties(kerberosWizardStep4Controller);
      kerberosWizardController.setDBProperty('tasksStatuses', null);
      kerberosWizardController.setDBProperty('tasksRequestIds', null);
      router.transitionTo('step5');
    }
  }),

  step5: Em.Route.extend({
    route: '/step5',

    connectOutlets: function (router) {
      console.log('in kerberosWizardController.step5:connectOutlets');
      var controller = router.get('kerberosWizardController');
      controller.dataLoading().done(function () {
        router.get('kerberosWizardController').setCurrentStep('5');
        controller.setLowerStepsDisable(5);
        controller.loadAllPriorSteps();
        controller.connectOutlet('kerberosWizardStep5', controller.get('content'));
      });
    },
    unroutePath: function () {
      return false;
    },
    back: Em.Router.transitionTo('step4'),
    next: function (router) {
      var kerberosWizardController = router.get('kerberosWizardController');
      kerberosWizardController.setDBProperty('tasksStatuses', null);
      kerberosWizardController.setDBProperty('tasksRequestIds', null);
      var step6Controller = router.get('kerberosWizardStep6Controller');
      var kerberosDescriptor = kerberosWizardController.get('kerberosDescriptorConfigs');
      step6Controller.postKerberosDescriptor(kerberosDescriptor).always(function (data, result, request) {
        if (result === 'error' && data.status === 409) {
          step6Controller.putKerberosDescriptor(kerberosDescriptor);
        } else {
          step6Controller.unkerberizeCluster();
        }
      });
    }
  }),

  step6: Em.Route.extend({
    route: '/step6',

    connectOutlets: function (router) {
      console.log('in kerberosWizardController.step6:connectOutlets');
      var controller = router.get('kerberosWizardController');
      var stepController = router.get('kerberosWizardStep6Controller');
      controller.dataLoading().done(function () {
        router.get('kerberosWizardController').setCurrentStep('6');
        controller.setLowerStepsDisable(6);
        controller.loadAllPriorSteps();
        stepController.setRequest();
        controller.connectOutlet('kerberosWizardStep6', controller.get('content'));
      });
    },
    retry: function (router) {
      router.get('kerberosWizardStep6Controller').setRequest(true);
    },
    unroutePath: function () {
      return false;
    },
    back: Em.Router.transitionTo('step4'),
    next: function (router) {
      router.transitionTo('step7');
    }
  }),

  step7: Em.Route.extend({
    route: '/step7',

    connectOutlets: function (router) {
      console.log('in kerberosWizardController.step7:connectOutlets');
      var controller = router.get('kerberosWizardController');
      controller.dataLoading().done(function () {
        router.get('kerberosWizardController').setCurrentStep('7');
        controller.setLowerStepsDisable(7);
        controller.loadAllPriorSteps();
        controller.connectOutlet('kerberosWizardStep7', controller.get('content'));
      });
    },
    unroutePath: function () {
      return false;
    },
    back: Em.Router.transitionTo('step6'),
    next: function (router) {
      var controller = router.get('kerberosWizardController');
      controller.finish();
      App.clusterStatus.setClusterStatus({
        clusterName: App.get('router').getClusterName(),
        clusterState: 'DEFAULT',
        localdb: App.db.data
      }, {
        alwaysCallback: function () {
          controller.get('popup').hide();
          App.get('router').transitionTo('adminKerberos.index');
          location.reload();
        }
      });

    }
  })
});

});

require.register("routes/add_security", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');

module.exports = App.WizardRoute.extend({
  route: '/addSecurity',
  enter: function (router) {
    console.log('in /security/add:enter');
    var controller = router.get('addSecurityController');
    controller.dataLoading().done(Ember.run.next(function () {
      //after refresh check if the wizard is open then restore it
      if (router.get('mainAdminSecurityController').getAddSecurityWizardStatus() === 'RUNNING') {
        var mainAdminSecurityController = router.get('mainAdminSecurityController');
        var addSecurityController = router.get('addSecurityController');
        var currentStep = router.get('addSecurityController').get('currentStep');
        App.router.get('updateController').set('isWorking', false);
        App.ModalPopup.show({
            classNames: ['full-width-modal'],
            header: Em.I18n.t('admin.addSecurity.header'),
            bodyClass: App.MainAdminSecurityAddMenuView.extend({
              controllerBinding: 'App.router.addSecurityController'
            }),
            primary: Em.I18n.t('form.cancel'),
            secondary: null,
            showFooter: false,

            onClose: function () {
              var self = this;
              if (router.get('addSecurityController.currentStep') == 4) {
                var controller = router.get('mainAdminSecurityAddStep4Controller');
                if (!controller.get('isSubmitDisabled')) {
                  router.get('mainAdminSecurityAddStep4Controller').clearStep();
                  self.proceedOnClose();
                  return;
                }
                var applyingConfigCommand = router.get('mainAdminSecurityAddStep4Controller.commands').findProperty('name', 'APPLY_CONFIGURATIONS');
                if (applyingConfigCommand) {
                  if (!applyingConfigCommand.get('isCompleted')) {
                    if (applyingConfigCommand.get('isStarted')) {
                      App.showAlertPopup(Em.I18n.t('admin.security.applying.config.header'), Em.I18n.t('admin.security.applying.config.body'));
                    } else {
                      App.showConfirmationPopup(function () {
                        self.proceedOnClose();
                      }, Em.I18n.t('admin.addSecurity.enable.onClose'));
                    }
                  } else {
                    App.showConfirmationPopup(function () {
                        self.proceedOnClose();
                      },
                      Em.I18n.t('admin.addSecurity.enable.after.stage2.onClose')
                    );
                  }
                  return;
                }
              }
              router.get('mainAdminSecurityAddStep4Controller').clearStep();
              App.db.setSecurityDeployCommands(undefined);
              self.proceedOnClose();
            },
            proceedOnClose: function () {
              var self = this;
              router.get('mainAdminSecurityAddStep4Controller').clearStep();
              router.get('addSecurityController.content.services').clear();
              router.set('addSecurityController.content.serviceConfigProperties', null);
              App.router.get('updateController').set('isWorking', true);
              mainAdminSecurityController.setAddSecurityWizardStatus(null);
              App.db.setSecurityDeployCommands(undefined);
              addSecurityController.finish();
              App.clusterStatus.setClusterStatus({
                clusterName: router.get('content.cluster.name'),
                clusterState: 'DEFAULT',
                localdb: App.db.data
              }, {alwaysCallback: function() {
                self.hide();
                router.transitionTo('adminSecurity.index');
                location.reload();   // this is needed because the ATS Component may be deleted in older HDP stacks.
              }});
            },
            didInsertElement: function () {
              this.fitHeight();
            }
          }
        );

        App.router.transitionTo('step' + currentStep);
      } else {
        router.transitionTo('adminSecurity.index');
      }
    }));
  },

  step1: Em.Route.extend({
    route: '/start',
    enter: function (router) {
      router.get('addSecurityController').setCurrentStep('1');
      if(!App.get('testMode')){
        App.clusterStatus.setClusterStatus({
          clusterName: this.get('clusterName'),
          clusterState: 'ADD_SECURITY_STEP_1',
          wizardControllerName: router.get('addSecurityController.name'),
          localdb: App.db.data
        });
      }
    },

    connectOutlets: function (router) {
      console.log('in addSecurity.step1:connectOutlets');
      var controller = router.get('addSecurityController');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('mainAdminSecurityAddStep1', controller.get('content'));
      })
    },

    unroutePath: function () {
      return false;
    },

    next: function (router) {
      var addSecurityController = router.get('addSecurityController');
      addSecurityController.get('content').set('serviceConfigProperties', null);
      App.db.setSecureConfigProperties(null);
      router.transitionTo('step2');
    }
  }),

  step2: Em.Route.extend({
    route: '/configure',

    enter: function (router) {
      router.get('addSecurityController').setCurrentStep('2');
      if(!App.get('testMode')){
        App.clusterStatus.setClusterStatus({
          clusterName: this.get('clusterName'),
          clusterState: 'ADD_SECURITY_STEP_2',
          wizardControllerName: router.get('addSecurityController.name'),
          localdb:  App.db.data
        });
      }
    },
    connectOutlets: function (router) {
      console.log('in addSecurity.step2:connectOutlets');
      var controller = router.get('addSecurityController');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('mainAdminSecurityAddStep2', controller.get('content'));
      })
    },
    unroutePath: function () {
      return false;
    },
    back: Em.Router.transitionTo('step1'),
    next: function (router) {
      var addSecurityController = router.get('addSecurityController');
      var addSecurityStep2Controller = router.get('mainAdminSecurityAddStep2Controller');
      addSecurityController.saveServiceConfigProperties(addSecurityStep2Controller);
      router.transitionTo('step3');
    }
  }),

  step3: Em.Route.extend({
    route: '/principal_keytab',

    enter: function (router) {
      router.get('addSecurityController').setCurrentStep('3');
      if(!App.get('testMode')){
        App.clusterStatus.setClusterStatus({
          clusterName: this.get('clusterName'),
          clusterState: 'ADD_SECURITY_STEP_3',
          wizardControllerName: router.get('addSecurityController.name'),
          localdb:  App.db.data
        });
      }
    },
    connectOutlets: function (router) {
      console.log('in addSecurity.step3:connectOutlets');
      var controller = router.get('addSecurityController');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('mainAdminSecurityAddStep3', controller.get('content'));
      })
    },
    unroutePath: function () {
      return false;
    },
    back: Em.Router.transitionTo('step2'),
    next: function (router) {
      App.db.setSecurityDeployCommands(undefined);
      router.transitionTo('step4');
    }
  }),

  step4: Em.Route.extend({
    route: '/apply',

    enter: function (router) {
      router.get('addSecurityController').setCurrentStep('4');
    },

    connectOutlets: function (router) {
      console.log('in addSecurity.step4:connectOutlets');
      var controller = router.get('addSecurityController');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.setLowerStepsDisable(4);
        controller.connectOutlet('mainAdminSecurityAddStep4', controller.get('content'));
      })
    },
    unroutePath: function () {
      return false;
    },
    back: function (router, context) {
      var controller = router.get('mainAdminSecurityAddStep4Controller');
      if (!controller.get('isBackBtnDisabled')) {
        router.transitionTo('step3');
      }
    },
    done: function (router, context) {
      var controller = router.get('mainAdminSecurityAddStep4Controller');
      if (!controller.get('isSubmitDisabled')) {
        $(context.currentTarget).parents("#modal").find(".close").trigger('click');
      }
    }
  })

});


});

require.register("routes/add_service_routes", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
// @todo: remove App.supports.kerberosAutomated after Kerberos Automation Wizard support

module.exports = App.WizardRoute.extend({
  route: '/service/add',

  enter: function (router) {
    console.log('in /service/add:enter');
    if (App.isAccessible('ADMIN')) {
      Em.run.next(function () {
        var addServiceController = router.get('addServiceController');
        App.router.get('updateController').set('isWorking', false);
        var popup = App.ModalPopup.show({
          classNames: ['full-width-modal'],
          header:Em.I18n.t('services.add.header'),
          bodyClass:  App.AddServiceView.extend({
            controllerBinding: 'App.router.addServiceController'
          }),
          primary:Em.I18n.t('form.cancel'),
          showFooter: false,
          secondary: null,

          onPrimary:function () {
            this.hide();
            App.router.transitionTo('main.services.index');
          },
          onClose: function() {
            this.set('showCloseButton', false); // prevent user to click "Close" many times
            App.router.get('updateController').set('isWorking', true);
            var self = this;
            App.router.get('updateController').updateServices(function(){
              App.router.get('updateController').updateServiceMetric();
            });
            var exitPath = addServiceController.getDBProperty('onClosePath') || 'main.services.index';
            addServiceController.finish();
            // We need to do recovery based on whether we are in Add Host or Installer wizard
            App.clusterStatus.setClusterStatus({
              clusterName: App.router.get('content.cluster.name'),
              clusterState: 'DEFAULT'
            }, {alwaysCallback: function() {self.hide();App.router.transitionTo(exitPath);location.reload();}});

          },
          didInsertElement: function(){
            this.fitHeight();
          }
        });
        addServiceController.set('popup',popup);
        var currentClusterStatus = App.clusterStatus.get('value');
        if (currentClusterStatus) {
          switch (currentClusterStatus.clusterState) {
            case 'ADD_SERVICES_DEPLOY_PREP_2' :
              addServiceController.setCurrentStep('5');
              break;
            case 'ADD_SERVICES_INSTALLING_3' :
            case 'SERVICE_STARTING_3' :
              addServiceController.setCurrentStep('6');
              break;
            case 'ADD_SERVICES_INSTALLED_4' :
              addServiceController.setCurrentStep('7');
              break;
            default:
              break;
          }
        }

        router.transitionTo('step' + addServiceController.get('currentStep'));
      });
    } else {
      Em.run.next(function () {
        App.router.transitionTo('main.services');
      });
    }

  },

  /*connectOutlets: function (router) {
    console.log('in /service/add:connectOutlets');
    router.get('mainController').connectOutlet('addService');
  },*/

  step1: Em.Route.extend({
    route: '/step1',
    connectOutlets: function (router) {
      console.log('in addService.step1:connectOutlets');
      var controller = router.get('addServiceController');
      controller.setCurrentStep('1');
      controller.set('hideBackButton', true);
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps().done(function () {
          var wizardStep4Controller = router.get('wizardStep4Controller');
          wizardStep4Controller.set('wizardController', controller);
          controller.connectOutlet('wizardStep4', controller.get('content.services').filterProperty('isInstallable', true));
        });
      });
    },
    next: function (router) {
      var addServiceController = router.get('addServiceController');
      var wizardStep4Controller = router.get('wizardStep4Controller');
      addServiceController.saveServices(wizardStep4Controller);
      addServiceController.saveClients(wizardStep4Controller);
      addServiceController.setDBProperty('masterComponentHosts', undefined);

      var wizardStep5Controller = router.get('wizardStep5Controller');
      wizardStep5Controller.clearRecommendations(); // Force reload recommendation between steps 1 and 2
      router.transitionTo('step2');
    }
  }),

  step2: Em.Route.extend({
    route: '/step2',
    connectOutlets: function (router) {
      console.log('in addService.step2:connectOutlets');
      var controller = router.get('addServiceController');
      var wizardStep2Controller = router.get('wizardStep5Controller');
      controller.setCurrentStep('2');
      controller.set('hideBackButton', false);
      wizardStep2Controller.set('isInitialLayout', true);
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps().done(function () {
          controller.connectOutlet('wizardStep5', controller.get('content'));
        });
      });

    },
    back: Em.Router.transitionTo('step1'),
    next: function (router) {
      var addServiceController = router.get('addServiceController');
      var wizardStep5Controller = router.get('wizardStep5Controller');
      var wizardStep6Controller = router.get('wizardStep6Controller');
      addServiceController.saveMasterComponentHosts(wizardStep5Controller);
      addServiceController.setDBProperty('slaveComponentHosts', undefined);
      wizardStep6Controller.set('isClientsSet', false);
      router.transitionTo('step3');
    }
  }),

  step3: Em.Route.extend({
    route: '/step3',
    connectOutlets: function (router) {
      console.log('in addService.step3:connectOutlets');
      var controller = router.get('addServiceController');
      controller.setCurrentStep('3');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps().done(function () {
          var wizardStep6Controller = router.get('wizardStep6Controller');
          wizardStep6Controller.set('wizardController', controller);
          controller.connectOutlet('wizardStep6', controller.get('content'));
        });
      });
    },
    back: function(router){
      var controller = router.get('addServiceController');
      if(!controller.get('content.skipMasterStep')){
        router.transitionTo('step2');
      } else {
        router.transitionTo('step1');
      }
    },
    next: function (router) {
      var addServiceController = router.get('addServiceController');
      var wizardStep6Controller = router.get('wizardStep6Controller');

      wizardStep6Controller.callValidation(function() {
        wizardStep6Controller.showValidationIssuesAcceptBox(function() {
          addServiceController.saveSlaveComponentHosts(wizardStep6Controller);
          addServiceController.get('content').set('serviceConfigProperties', null);
          addServiceController.setDBProperty('serviceConfigProperties', null);
          addServiceController.setDBProperty('groupsToDelete', []);
          router.transitionTo('step4');
        });
      });
    }
  }),

  step4: Em.Route.extend({
    route: '/step4',
    connectOutlets: function (router) {
      console.log('in addService.step4:connectOutlets');
      var controller = router.get('addServiceController');
      controller.setCurrentStep('4');
      controller.dataLoading().done(function () {
        var wizardStep7Controller = router.get('wizardStep7Controller');
        controller.loadAllPriorSteps().done(function () {
          controller.loadAdvancedConfigs(wizardStep7Controller);
          wizardStep7Controller.getConfigTags();
          wizardStep7Controller.set('wizardController', controller);
          controller.usersLoading().done(function () {
            controller.connectOutlet('wizardStep7', controller.get('content'));
          });
        });
      });
    },
    back: function(router){
      var controller = router.get('addServiceController');
      if(!controller.get('content.skipSlavesStep')){
        router.transitionTo('step3');
      } else if(!controller.get('content.skipMasterStep')) {
        router.transitionTo('step2');
      } else {
        router.transitionTo('step1');
      }
    },
    next: function (router) {
      var addServiceController = router.get('addServiceController');
      var wizardStep7Controller = router.get('wizardStep7Controller');
      addServiceController.saveServiceConfigProperties(wizardStep7Controller);
      addServiceController.saveServiceConfigGroups(wizardStep7Controller, true);
      if (App.supports.automatedKerberos) {
        if (router.get('mainAdminKerberosController.securityEnabled')) {
          router.transitionTo('step5');
          return;
        }
      }
      router.transitionTo('step6');
    }
  }),

  step5: Em.Route.extend({
    route: '/step5',
    connectOutlets: function (router) {
      var controller = router.get('addServiceController');
      controller.setCurrentStep('5');
      controller.dataLoading().done(function () {
        var kerberosStep4Controller = router.get('kerberosWizardStep4Controller');
        controller.loadAllPriorSteps().done(function () {
          kerberosStep4Controller.set('wizardController', controller);
          controller.connectOutlet('kerberosWizardStep4', controller.get('content'));
        });
      });
    },
    back: function(router){
      var controller = router.get('addServiceController');
      if(!controller.get('content.skipConfigStep')) {
        router.transitionTo('step4');
      }
      else {
        if(!controller.get('content.skipSlavesStep')) {
          router.transitionTo('step3');
        }
        else {
          if(!controller.get('content.skipMasterStep')) {
            router.transitionTo('step2');
          }
          else {
            router.transitionTo('step1');
          }
        }
      }
    },
    next: function (router) {
      if (App.supports.automatedKerberos && router.get('mainAdminKerberosController.securityEnabled')) {
        router.get('kerberosWizardStep2Controller').createKerberosAdminSession(router.get('kerberosWizardStep4Controller.stepConfigs')[0].get('configs'));
      }
      router.transitionTo('step6');
    }
  }),

  step6: Em.Route.extend({
    route: App.supports.automatedKerberos ? '/step6' : '/step5',
    connectOutlets: function (router, context) {
      console.log('in addService.step5:connectOutlets');
      var controller = router.get('addServiceController');
      controller.setCurrentStep(App.supports.automatedKerberos ? '6' : '5');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps().done(function () {
          var wizardStep8Controller = router.get('wizardStep8Controller');
          wizardStep8Controller.set('wizardController', controller);
          controller.connectOutlet('wizardStep8', controller.get('content'));
        });
      });
    },
    back: function(router){
      var controller = router.get('addServiceController');
      if (App.supports.automatedKerberos && router.get('mainAdminKerberosController.securityEnabled')) {
        router.transitionTo('step5');
        return;
      }
      if(!controller.get('content.skipConfigStep')) {
        router.transitionTo('step4');
      }
      else {
        if(!controller.get('content.skipSlavesStep')) {
          router.transitionTo('step3');
        }
        else {
          if(!controller.get('content.skipMasterStep')) {
            router.transitionTo('step2');
          }
          else {
            router.transitionTo('step1');
          }
        }
      }
    },
    next: function (router) {
      var addServiceController = router.get('addServiceController');
      addServiceController.installServices(function () {
        router.get('wizardStep8Controller').set('servicesInstalled', true);
        addServiceController.setInfoForStep9();
        addServiceController.saveClusterState('ADD_SERVICES_INSTALLING_3');
        App.router.transitionTo('step7');
      });
    }
  }),

  step7: Em.Route.extend({
    route: App.supports.automatedKerberos ? '/step7' : '/step6',
    connectOutlets: function (router, context) {
      console.log('in addService.step6:connectOutlets');
      var controller = router.get('addServiceController');
      controller.setCurrentStep(App.supports.automatedKerberos ? '7' : '6');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps().done(function () {
          var wizardStep9Controller = router.get('wizardStep9Controller');
          wizardStep9Controller.set('wizardController', controller);
          if (!App.get('testMode')) {              //if test mode is ON don't disable prior steps link.
            controller.setLowerStepsDisable(App.supports.automatedKerberos ? 7 : 6);
          }
          controller.connectOutlet('wizardStep9', controller.get('content'));
        });
      });
    },
    back: Em.Router.transitionTo('step6'),
    retry: function(router,context) {
      var addServiceController = router.get('addServiceController');
      var wizardStep9Controller = router.get('wizardStep9Controller');
      if (wizardStep9Controller.get('showRetry')) {
        if (wizardStep9Controller.get('content.cluster.status') === 'INSTALL FAILED') {
          addServiceController.installServices(function () {
            addServiceController.setInfoForStep9();
            wizardStep9Controller.resetHostsForRetry();
            // We need to do recovery based on whether we are in Add Host or Installer wizard
            addServiceController.saveClusterState('ADD_SERVICES_INSTALLING_3');
            wizardStep9Controller.navigateStep();
          });
        } else {
          wizardStep9Controller.navigateStep();
        }
      }
    },
    unroutePath: function() {
      return false;
    },
    next: function (router) {
      var addServiceController = router.get('addServiceController');
      var wizardStep9Controller = router.get('wizardStep9Controller');
      addServiceController.saveInstalledHosts(wizardStep9Controller);

      // We need to do recovery based on whether we are in Add Host or Installer wizard
      addServiceController.saveClusterState('ADD_SERVICES_INSTALLED_4');
      router.transitionTo('step8');
    }
  }),

  step8: Em.Route.extend({
    route: App.supports.automatedKerberos ? '/step8' : '/step7',
    connectOutlets: function (router, context) {
      console.log('in addService.step7:connectOutlets');
      var controller = router.get('addServiceController');
      controller.setCurrentStep(App.supports.automatedKerberos ? '8' : '7');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps().done(function () {
          controller.setLowerStepsDisable(App.supports.automatedKerberos ? 8 : 7);
          controller.connectOutlet('wizardStep10', controller.get('content'));
        });
      });
    },
    back: Em.Router.transitionTo('step7'),
    complete: function (router, context) {
      var addServiceController = router.get('addServiceController');
      addServiceController.get('popup').onClose();
    }
  }),

  backToServices: function (router) {
    App.router.get('updateController').set('isWorking', true);
    router.transitionTo('services');
  }

});

});

require.register("routes/high_availability_routes", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

module.exports = App.WizardRoute.extend({
  route: '/highAvailability/NameNode/enable',

  enter: function (router) {
    var highAvailabilityWizardController = router.get('highAvailabilityWizardController');
    highAvailabilityWizardController.dataLoading().done(function () {
        App.router.get('updateController').set('isWorking', false);
        var popup = App.ModalPopup.show({
          classNames: ['full-width-modal'],
          header: Em.I18n.t('admin.highAvailability.wizard.header'),
          bodyClass: App.HighAvailabilityWizardView.extend({
            controller: highAvailabilityWizardController
          }),
          primary: Em.I18n.t('form.cancel'),
          showFooter: false,
          secondary: null,
          hideCloseButton: function () {
            var currStep = App.router.get('highAvailabilityWizardController.currentStep');
            switch (currStep) {
              case "5" :
              case "7" :
              case "9" :
                if (App.supports.autoRollbackHA) {
                  this.set('showCloseButton', false);
                } else {
                  this.set('showCloseButton', true);
                }
                break;
              default :
                this.set('showCloseButton', true);
            }
          }.observes('App.router.highAvailabilityWizardController.currentStep'),

          onClose: function () {
            var self = this;
            var currStep = App.router.get('highAvailabilityWizardController.currentStep');
            var highAvailabilityProgressPageController = App.router.get('highAvailabilityProgressPageController');
            if (parseInt(currStep) > 4) {
              if (!App.supports.autoRollbackHA) {
                highAvailabilityProgressPageController.manualRollback();
              } else {
                this.hide();
                App.router.get('highAvailabilityWizardController').setCurrentStep('1');
                App.router.transitionTo('rollbackHighAvailability');
              }
            } else {
              var controller = App.router.get('highAvailabilityWizardController');
              controller.clearTasksData();
              controller.finish();
              App.router.get('updateController').set('isWorking', true);
              App.clusterStatus.setClusterStatus({
                clusterName: controller.get('content.cluster.name'),
                clusterState: 'DEFAULT',
                localdb: App.db.data
              }, {
                alwaysCallback: function () {
                  self.hide();
                  App.router.transitionTo('main.services.index');
                  location.reload();
                }
              });
            }
          },
          didInsertElement: function () {
            this.fitHeight();
          }
        });
        highAvailabilityWizardController.set('popup', popup);
        var currentClusterStatus = App.clusterStatus.get('value');
        if (currentClusterStatus) {
          switch (currentClusterStatus.clusterState) {
            case 'HIGH_AVAILABILITY_DEPLOY' :
              highAvailabilityWizardController.setCurrentStep(currentClusterStatus.localdb.HighAvailabilityWizard.currentStep);
              break;
            default:
              var currStep = App.router.get('highAvailabilityWizardController.currentStep');
              highAvailabilityWizardController.setCurrentStep(currStep);
              break;
          }
        }
      Em.run.next(function () {
        router.transitionTo('step' + highAvailabilityWizardController.get('currentStep'));
      });
    });
  },

  step1: Em.Route.extend({
    route: '/step1',
    connectOutlets: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.dataLoading().done(function () {
        controller.setCurrentStep('1');
        controller.usersLoading().done(function () {
          controller.saveHdfsUser();
          controller.loadAllPriorSteps().done(function () {
            controller.connectOutlet('highAvailabilityWizardStep1', controller.get('content'));
          });
        })
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.saveNameServiceId(router.get('highAvailabilityWizardStep1Controller.content.nameServiceId'));
      router.transitionTo('step2');
    }
  }),

  step2: Em.Route.extend({
    route: '/step2',
    connectOutlets: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.dataLoading().done(function () {
        controller.setCurrentStep('2');
        controller.loadAllPriorSteps().done(function () {
          controller.connectOutlet('highAvailabilityWizardStep2', controller.get('content'));
        });
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      var highAvailabilityWizardStep2Controller = router.get('highAvailabilityWizardStep2Controller');
      var addNN = highAvailabilityWizardStep2Controller.get('selectedServicesMasters').findProperty('isAddNameNode', true).get('selectedHost');
      var sNN = highAvailabilityWizardStep2Controller.get('selectedServicesMasters').findProperty('component_name','SECONDARY_NAMENODE').get('selectedHost');
      if(addNN){
        App.db.setRollBackHighAvailabilityWizardAddNNHost(addNN);
      }
      if(sNN){
        App.db.setRollBackHighAvailabilityWizardSNNHost(sNN);
      }

      controller.saveMasterComponentHosts(highAvailabilityWizardStep2Controller);
      controller.get('content').set('serviceConfigProperties', null);
      controller.setDBProperty('serviceConfigProperties', null);
      router.transitionTo('step3');
    },
    back: function (router) {
      router.transitionTo('step1');
    }
  }),

  step3: Em.Route.extend({
    route: '/step3',
    connectOutlets: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.dataLoading().done(function () {
        controller.setCurrentStep('3');
        controller.loadAllPriorSteps().done(function () {
          controller.connectOutlet('highAvailabilityWizardStep3',  controller.get('content'));
        });
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      var stepController = router.get('highAvailabilityWizardStep3Controller');
      controller.saveServiceConfigProperties(stepController);
      controller.saveConfigTag(stepController.get("hdfsSiteTag"));
      controller.saveConfigTag(stepController.get("coreSiteTag"));
      if (App.Service.find().someProperty('serviceName', 'HBASE')) {
        controller.saveConfigTag(stepController.get("hbaseSiteTag"));
      }
      router.transitionTo('step4');
    },
    back: Em.Router.transitionTo('step2')
  }),

  step4: Em.Route.extend({
    route: '/step4',
    connectOutlets: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.dataLoading().done(function () {
        controller.setCurrentStep('4');
        controller.setLowerStepsDisable(4);
        controller.loadAllPriorSteps().done(function () {
          controller.connectOutlet('highAvailabilityWizardStep4',  controller.get('content'));
        });
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.clearTasksData();
      router.transitionTo('step5');
    }
  }),

  step5: Em.Route.extend({
    route: '/step5',
    connectOutlets: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.dataLoading().done(function () {
        controller.setCurrentStep('5');
        controller.setLowerStepsDisable(5);
        controller.loadAllPriorSteps().done(function () {
          controller.connectOutlet('highAvailabilityWizardStep5',  controller.get('content'));
        });
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.clearTasksData();
      router.transitionTo('step6');
    }
  }),

  step6: Em.Route.extend({
    route: '/step6',
    connectOutlets: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.dataLoading().done(function () {
        controller.setCurrentStep('6');
        controller.setLowerStepsDisable(6);
        controller.loadAllPriorSteps().done(function () {
          controller.connectOutlet('highAvailabilityWizardStep6',  controller.get('content'));
        });
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      router.transitionTo('step7');
    }
  }),

  step7: Em.Route.extend({
    route: '/step7',
    connectOutlets: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.dataLoading().done(function () {
        controller.setCurrentStep('7');
        controller.setLowerStepsDisable(7);
        controller.loadAllPriorSteps().done(function () {
          controller.connectOutlet('highAvailabilityWizardStep7',  controller.get('content'));
        });
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.clearTasksData();
      router.transitionTo('step8');
    }
  }),

  step8: Em.Route.extend({
    route: '/step8',
    connectOutlets: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.dataLoading().done(function () {
        controller.setCurrentStep('8');
        controller.setLowerStepsDisable(8);
        controller.loadAllPriorSteps().done(function () {
          controller.connectOutlet('highAvailabilityWizardStep8',  controller.get('content'));
        });
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      App.showConfirmationPopup(function() {
        router.transitionTo('step9');
      }, Em.I18n.t('admin.highAvailability.wizard.step8.confirmPopup.body'));
    }
  }),

  step9: Em.Route.extend({
    route: '/step9',
    connectOutlets: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.dataLoading().done(function () {
        controller.setCurrentStep('9');
        controller.setLowerStepsDisable(9);
        controller.loadAllPriorSteps().done(function () {
          controller.connectOutlet('highAvailabilityWizardStep9',  controller.get('content'));
        });
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      var controller = router.get('highAvailabilityWizardController');
      controller.clearTasksData();
      controller.finish();
      App.clusterStatus.setClusterStatus({
        clusterName: controller.get('content.cluster.name'),
        clusterState: 'DEFAULT',
        localdb: App.db.data
      },{alwaysCallback: function() {controller.get('popup').hide();router.transitionTo('main.services.index');location.reload();}});
    }
  })

});

});

require.register("routes/installer", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

module.exports = Em.Route.extend(App.RouterRedirections, {
  route: '/installer',
  App: require('app'),

  enter: function (router) {
    console.log('in /installer:enter');
    var self = this;

    App.clusterStatus.set('wizardControllerName', App.router.get('installerController.name'));

    router.getAuthenticated().done(function (loggedIn) {
      if (loggedIn) {
        var applicationController = router.get('applicationController');
        applicationController.startKeepAlivePoller();
        // check server/web client versions match
        App.router.get('installerController').checkServerClientVersion().done(function () {

          var name = 'Cluster Install Wizard';
          $('title').text('Ambari - ' + name);

          App.router.get('mainViewsController').loadAmbariViews();
          if (App.isAccessible('ADMIN')) {
            router.get('mainController').stopPolling();
            console.log('In installer with successful authenticated');
            console.log('current step=' + router.get('installerController.currentStep'));
            Em.run.next(function () {
              App.clusterStatus.updateFromServer().complete(function () {
                var currentClusterStatus = App.clusterStatus.get('value');
                //@TODO: Clean up  following states. Navigation should be done solely via currentStep stored in the localDb and API persist endpoint.
                //       Actual currentStep value for the installer controller should always remain in sync with localdb and at persist store in the server.
                if (currentClusterStatus) {
                  if (self.get('installerStatuses').contains(currentClusterStatus.clusterState)) {
                    self.redirectToInstaller(router, currentClusterStatus, true);
                  }
                  else {
                    router.transitionTo('main.dashboard.index');
                  }
                }
              });
            });
          } else {
            Em.run.next(function () {
              App.router.transitionTo('main.views.index');
            });
          }
        });
      } else {
        console.log('In installer but its not authenticated');
        console.log('value of authenticated is: ' + router.getAuthenticated());
        Ember.run.next(function () {
          router.transitionTo('login');
        });
      }
    });
  },

  routePath: function (router, event) {
    console.log("INFO: value of router is: " + router);
    console.log("INFO: value of event is: " + event);
    router.setNavigationFlow(event);
    if (!router.isFwdNavigation) {
      this._super(router, event);
    } else {
      router.set('backBtnForHigherStep', true);

      var installerController = router.get('installerController');
      router.transitionTo('step' + installerController.get('currentStep'));
    }
  },

  connectOutlets: function (router, context) {
    console.log('in /installer:connectOutlets');
    router.get('applicationController').connectOutlet('installer');
  },

  step0: Em.Route.extend({
    route: '/step0',
    connectOutlets: function (router) {
      console.log('in installer.step0:connectOutlets');
      var controller = router.get('installerController');
      controller.setCurrentStep('0');
      controller.loadAllPriorSteps().done(function () {
        controller.connectOutlet('wizardStep0', controller.get('content'));
      });
    },

    next: function (router) {
      var installerController = router.get('installerController');
      installerController.save('cluster');
      App.db.setStacks(undefined);
      installerController.set('content.stacks',undefined);
      router.transitionTo('step1');
    }
  }),

  step1: Em.Route.extend({
    route: '/step1',
    connectOutlets: function (router) {
      console.log('in installer.step1:connectOutlets');
      var controller = router.get('installerController');
      controller.setCurrentStep('1');
      controller.loadAllPriorSteps().done(function () {
        controller.connectOutlet('wizardStep1', controller.get('content'));
      });
    },
    back: Em.Router.transitionTo('step0'),
    next: function (router) {
      var wizardStep1Controller = router.get('wizardStep1Controller');
      var installerController = router.get('installerController');
      installerController.checkRepoURL(wizardStep1Controller).done(function () {
        installerController.setDBProperty('service', undefined);
        installerController.setStacks();
        installerController.clearInstallOptions();
        router.transitionTo('step2');
      });
    }
  }),

  step2: Em.Route.extend({
    route: '/step2',
    connectOutlets: function (router, context) {
      router.setNavigationFlow('step2');

      var controller = router.get('installerController');
      controller.setCurrentStep('2');
      controller.loadAllPriorSteps().done(function () {
        controller.connectOutlet('wizardStep2', controller.get('content'));
      });
    },
    back: Em.Router.transitionTo('step1'),
    next: function (router) {
      var controller = router.get('installerController');
      controller.save('installOptions');
      //hosts was saved to content.hosts inside wizardStep2Controller
      controller.save('hosts');
      router.transitionTo('step3');
    }
  }),

  step3: Em.Route.extend({
    route: '/step3',
    connectOutlets: function (router) {
      console.log('in installer.step3:connectOutlets');
      var controller = router.get('installerController');
      controller.setCurrentStep('3');
      controller.loadAllPriorSteps().done(function () {
        var wizardStep3Controller = router.get('wizardStep3Controller');
        wizardStep3Controller.set('wizardController', controller);
        controller.connectOutlet('wizardStep3', controller.get('content'));
      });
    },
    back: function (router) {
      router.transitionTo('step2');
    },
    next: function (router, context) {
      var installerController = router.get('installerController');
      var wizardStep3Controller = router.get('wizardStep3Controller');
      installerController.saveConfirmedHosts(wizardStep3Controller);
      installerController.setDBProperty('bootStatus', true);
      installerController.setDBProperty('selectedServiceNames', undefined);
      installerController.setDBProperty('installedServiceNames', undefined);
      router.transitionTo('step4');
    },
    exit: function (router) {
      router.get('wizardStep3Controller').set('stopBootstrap', true);
    },
    /**
     * Wrapper for remove host action.
     * Since saving data stored in installerController, we should call this from router
     * @param router
     * @param context Array of hosts to delete
     */
    removeHosts: function (router, context) {
      console.log('in installer.step2.removeHosts:hosts to delete ', context);
      var controller = router.get('installerController');
      controller.removeHosts(context);
    }
  }),

  step4: Em.Route.extend({
    route: '/step4',
    connectOutlets: function (router, context) {
      router.setNavigationFlow('step4');
      var controller = router.get('installerController');
      controller.setCurrentStep('4');
      controller.loadAllPriorSteps().done(function () {
        var wizardStep4Controller = router.get('wizardStep4Controller');
        wizardStep4Controller.set('wizardController', controller);
        controller.connectOutlet('wizardStep4', App.StackService.find().filterProperty('isInstallable', true));
      });
    },
    back: Em.Router.transitionTo('step3'),

    next: function (router) {
      var controller = router.get('installerController');
      var wizardStep4Controller = router.get('wizardStep4Controller');
      controller.saveServices(wizardStep4Controller);
      controller.saveClients(wizardStep4Controller);

      router.get('wizardStep5Controller').clearRecommendations(); // Force reload recommendation between steps 4 and 5
      controller.setDBProperty('recommendations', undefined);
      controller.setDBProperty('masterComponentHosts', undefined);
      router.transitionTo('step5');
    }
  }),

  step5: Em.Route.extend({
    route: '/step5',
    connectOutlets: function (router, context) {
      router.setNavigationFlow('step5');

      var controller = router.get('installerController');
      var wizardStep5Controller = router.get('wizardStep5Controller');
      wizardStep5Controller.setProperties({
        servicesMasters: [],
        isInitialLayout: true
      });
      controller.setCurrentStep('5');
      controller.loadAllPriorSteps().done(function () {
        controller.connectOutlet('wizardStep5', controller.get('content'));
      });
    },
    back: Em.Router.transitionTo('step4'),
    next: function (router) {
      var controller = router.get('installerController');
      var wizardStep5Controller = router.get('wizardStep5Controller');
      var wizardStep6Controller = router.get('wizardStep6Controller');
      controller.saveMasterComponentHosts(wizardStep5Controller);
      controller.setDBProperty('slaveComponentHosts', undefined);
      controller.setDBProperty('recommendations', wizardStep5Controller.get('content.recommendations'));
      wizardStep6Controller.set('isClientsSet', false);
      router.transitionTo('step6');
    }
  }),

  step6: Em.Route.extend({
    route: '/step6',
    connectOutlets: function (router, context) {
      router.setNavigationFlow('step6');

      var controller = router.get('installerController');
      router.get('wizardStep6Controller').set('hosts', []);
      controller.setCurrentStep('6');
      controller.loadAllPriorSteps().done(function () {
        controller.connectOutlet('wizardStep6', controller.get('content'));
      });
    },
    back: Em.Router.transitionTo('step5'),

    next: function (router) {
      var controller = router.get('installerController');
      var wizardStep6Controller = router.get('wizardStep6Controller');
      var wizardStep7Controller = router.get('wizardStep7Controller');

      if (!wizardStep6Controller.get('submitDisabled')) {
        wizardStep6Controller.showValidationIssuesAcceptBox(function () {
          controller.saveSlaveComponentHosts(wizardStep6Controller);
          controller.get('content').set('serviceConfigProperties', null);
          controller.setDBProperty('serviceConfigProperties', null);
          controller.setDBProperty('advancedServiceConfig', null);
          controller.setDBProperty('serviceConfigGroups', null);
          controller.setDBProperty('recommendationsHostGroups', wizardStep6Controller.get('content.recommendationsHostGroups'));
          controller.setDBProperty('recommendationsConfigs', null);
          controller.loadAdvancedConfigs(wizardStep7Controller);
          wizardStep7Controller.set('isAdvancedConfigLoaded', false);
          router.transitionTo('step7');
        });
      }
    }
  }),

  step7: Em.Route.extend({
    route: '/step7',
    enter: function (router) {
      console.log('in /wizardStep7Controller:enter');
      var controller = router.get('installerController');
      controller.setCurrentStep('7');
    },
    connectOutlets: function (router, context) {
      var controller = router.get('installerController');

      controller.loadAllPriorSteps().done(function () {
        var wizardStep7Controller = router.get('wizardStep7Controller');
        wizardStep7Controller.set('wizardController', controller);
        controller.connectOutlet('wizardStep7', controller.get('content'));
      });
    },
    back: Em.Router.transitionTo('step6'),
    next: function (router) {
      var controller = router.get('installerController');
      var wizardStep7Controller = router.get('wizardStep7Controller');
      controller.saveServiceConfigProperties(wizardStep7Controller);
      controller.saveServiceConfigGroups(wizardStep7Controller);
      controller.setDBProperty('recommendationsConfigs', wizardStep7Controller.get('recommendationsConfigs'));
      router.transitionTo('step8');
    }
  }),

  step8: Em.Route.extend({
    route: '/step8',
    connectOutlets: function (router, context) {
      console.log('in installer.step8:connectOutlets');
      var controller = router.get('installerController');
      controller.setCurrentStep('8');
      controller.loadAllPriorSteps().done(function () {
        var wizardStep8Controller = router.get('wizardStep8Controller');
        wizardStep8Controller.set('wizardController', controller);
        controller.connectOutlet('wizardStep8', controller.get('content'));
      });
    },
    back: Em.Router.transitionTo('step7'),
    next: function (router) {
      var installerController = router.get('installerController');
      var wizardStep8Controller = router.get('wizardStep8Controller');
      // invoke API call to install selected services
      installerController.installServices(false, function () {
        installerController.setInfoForStep9();
        // We need to do recovery based on whether we are in Add Host or Installer wizard
        installerController.saveClusterState('CLUSTER_INSTALLING_3');
        wizardStep8Controller.set('servicesInstalled', true);
        router.transitionTo('step9');
      });
    }
  }),

  step9: Em.Route.extend({
    route: '/step9',
    connectOutlets: function (router, context) {
      console.log('in installer.step9:connectOutlets');
      var controller = router.get('installerController');
      controller.setCurrentStep('9');
      controller.loadAllPriorSteps().done(function () {
        if (!App.get('testMode')) {
          controller.setLowerStepsDisable(9);
        }
        var wizardStep9Controller = router.get('wizardStep9Controller');
        wizardStep9Controller.set('wizardController', controller);
        controller.connectOutlet('wizardStep9', controller.get('content'));
      });
    },
    back: Em.Router.transitionTo('step8'),
    retry: function (router) {
      var installerController = router.get('installerController');
      var wizardStep9Controller = router.get('wizardStep9Controller');
      if (wizardStep9Controller.get('showRetry')) {
        if (wizardStep9Controller.get('content.cluster.status') === 'INSTALL FAILED') {
          var isRetry = true;
          installerController.installServices(isRetry, function () {
            installerController.setInfoForStep9();
            wizardStep9Controller.resetHostsForRetry();
            // We need to do recovery based on whether we are in Add Host or Installer wizard
            installerController.saveClusterState('CLUSTER_INSTALLING_3');
            wizardStep9Controller.navigateStep();
          });
        } else {
          wizardStep9Controller.navigateStep();
        }
      }
    },
    unroutePath: function (router, context) {
      // exclusion for transition to Admin View
      if (context === '/adminView') {
        this._super(router, context);
      } else {
        return false;
      }
    },
    next: function (router) {
      var installerController = router.get('installerController');
      var wizardStep9Controller = router.get('wizardStep9Controller');
      installerController.saveInstalledHosts(wizardStep9Controller);

      installerController.saveClusterState('CLUSTER_INSTALLED_4');
      router.transitionTo('step10');
    }
  }),

  step10: Em.Route.extend({
    route: '/step10',
    connectOutlets: function (router, context) {
      console.log('in installer.step10:connectOutlets');
      var controller = router.get('installerController');
      controller.setCurrentStep('10');
      controller.loadAllPriorSteps().done(function () {
        if (!App.get('testMode')) {
          controller.setLowerStepsDisable(10);
        }
        controller.connectOutlet('wizardStep10', controller.get('content'));
      });
    },
    back: Em.Router.transitionTo('step9'),
    complete: function (router, context) {
      var controller = router.get('installerController');
      controller.finish();
      controller.setClusterProvisioningState('INSTALLED').complete(function () {
        // We need to do recovery based on whether we are in Add Host or Installer wizard
        controller.saveClusterState('DEFAULT');
        App.router.set('clusterController.isLoaded', false);
        router.set('clusterInstallCompleted', true);
        router.transitionTo('main.dashboard.index');
      });
    }
  }),

  gotoStep0: Em.Router.transitionTo('step0'),

  gotoStep1: Em.Router.transitionTo('step1'),

  gotoStep2: Em.Router.transitionTo('step2'),

  gotoStep3: Em.Router.transitionTo('step3'),

  gotoStep4: Em.Router.transitionTo('step4'),

  gotoStep5: Em.Router.transitionTo('step5'),

  gotoStep6: Em.Router.transitionTo('step6'),

  gotoStep7: Em.Router.transitionTo('step7'),

  gotoStep8: Em.Router.transitionTo('step8'),

  gotoStep9: Em.Router.transitionTo('step9'),

  gotoStep10: Em.Router.transitionTo('step10')

});

});

require.register("routes/main", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

module.exports = Em.Route.extend(App.RouterRedirections, {
  route: '/main',
  enter: function (router) {
    App.db.updateStorage();
    console.log('in /main:enter');
    var self = this;
    router.getAuthenticated().done(function (loggedIn) {
      if (loggedIn) {
        var applicationController = App.router.get('applicationController');
        applicationController.startKeepAlivePoller();
        App.router.get('mainController').checkServerClientVersion().done(function () {
          App.router.get('mainViewsController').loadAmbariViews();
          App.router.get('clusterController').loadClusterName(false).done(function () {
            if (App.get('testMode')) {
              router.get('mainController').initialize();
            } else {
              if (router.get('clusterInstallCompleted')) {
                App.router.get('clusterController').loadClientServerClockDistance().done(function () {
                  App.router.get('clusterController').checkDetailedRepoVersion().done(function () {
                    router.get('mainController').initialize();
                  });
                });
              }
              else {
                Em.run.next(function () {
                  App.clusterStatus.updateFromServer().complete(function () {
                    var currentClusterStatus = App.clusterStatus.get('value');
                    if (router.get('currentState.parentState.name') !== 'views'
                        && currentClusterStatus && self.get('installerStatuses').contains(currentClusterStatus.clusterState)) {
                      if (App.isAccessible('ADMIN')) {
                        self.redirectToInstaller(router, currentClusterStatus, false);
                      } else {
                        Em.run.next(function () {
                          App.router.transitionTo('main.views.index');
                        });
                      }
                    }
                  });
                });
                App.router.get('clusterController').set('isLoaded', true);
              }
            }
          });
        });
        // TODO: redirect to last known state
      } else {
        router.set('preferedPath', router.location.location.hash);
        Em.run.next(function () {
          router.transitionTo('login');
        });
      }
    });
  },
  /*
   routePath: function(router,event) {
   if (router.getAuthenticated()) {
   App.router.get('clusterController').loadClusterName(false);
   router.get('mainController').initialize();
   // TODO: redirect to last known state
   } else {
   Ember.run.next(function () {
   router.transitionTo('login');
   });
   }
   }, */

  index: Ember.Route.extend({
    route: '/',
    redirectsTo: 'dashboard.index'
  }),

  connectOutlets: function (router, context) {
    router.get('applicationController').connectOutlet('main');
  },

  test: Em.Route.extend({
    route: '/test',
    connectOutlets: function (router, context) {
      router.get('mainController').connectOutlet('mainTest');
    }
  }),

  dashboard: Em.Route.extend({
    route: '/dashboard',
    connectOutlets: function (router, context) {
      router.get('mainController').connectOutlet('mainDashboard');
    },
    index: Em.Route.extend({
      route: '/',
      enter: function (router) {
        Em.run.next(function () {
          router.transitionTo('main.dashboard.widgets');
        });
      }
    }),
    goToDashboardView: function (router, event) {
      router.transitionTo(event.context);
    },
    widgets: Em.Route.extend({
      route: '/metrics',
      connectOutlets: function (router, context) {
        router.set('mainDashboardController.selectedCategory', 'widgets');
        router.get('mainDashboardController').connectOutlet('mainDashboardWidgets');
      }
    }),
    charts: Em.Route.extend({
      route: '/charts',
      connectOutlets: function (router, context) {
        router.set('mainDashboardController.selectedCategory', 'charts');
        router.get('mainDashboardController').connectOutlet('mainCharts');
      },
      index: Ember.Route.extend({
        route: '/',
        enter: function (router) {
          Em.run.next(function () {
            router.transitionTo('heatmap');
          });
        }
      }),
      heatmap: Em.Route.extend({
        route: '/heatmap',
        connectOutlets: function (router, context) {
          router.get('mainChartsController').connectOutlet('mainChartsHeatmap');
        }
      }),
      horizon_chart: Em.Route.extend({
        route: '/horizon_chart',
        connectOutlets: function (router, context) {
          router.get('mainChartsController').connectOutlet('mainChartsHorizon');
        }
      }),
      showChart: function (router, event) {
        var parent = event.view._parentView;
        parent.deactivateChildViews();
        event.view.set('active', "active");
        router.transitionTo(event.context);
      }
    }),
    configHistory: Em.Route.extend({
      route: '/config_history',
      connectOutlets: function (router, context) {
        router.set('mainDashboardController.selectedCategory', 'configHistory');
        router.get('mainDashboardController').connectOutlet('mainConfigHistory');
      }
    }),
    goToServiceConfigs: function (router, event) {
      router.get('mainServiceItemController').set('routeToConfigs', true);
      router.get('mainServiceInfoConfigsController').set('preSelectedConfigVersion', event.context);
      router.transitionTo('main.services.service.configs', App.Service.find(event.context.get('serviceName')));
      router.get('mainServiceItemController').set('routeToConfigs', false);
    }
  }),

  views: require('routes/views'),

  hosts: Em.Route.extend({
    route: '/hosts',
    index: Ember.Route.extend({
      route: '/',
      connectOutlets: function (router, context) {
        router.get('mainController').connectOutlet('mainHost');
      }
    }),

    hostDetails: Em.Route.extend({
      route: '/:host_id',
      connectOutlets: function (router, host) {
        router.get('mainHostController').set('showFilterConditionsFirstLoad', true);
        router.get('mainController').connectOutlet('mainHostDetails', host);
      },

      index: Ember.Route.extend({
        route: '/',
        redirectsTo: 'summary'
      }),

      summary: Em.Route.extend({
        route: '/summary',
        connectOutlets: function (router, context) {
          router.get('mainController').dataLoading().done(function() {
            var controller = router.get('mainHostDetailsController');
            if ( App.Service.find().mapProperty('serviceName').contains('OOZIE')) {
              controller.loadConfigs('loadOozieConfigs');
              controller.isOozieConfigLoaded.always(function () {
                controller.connectOutlet('mainHostSummary');
              });
            }else {
              controller.connectOutlet('mainHostSummary');
            }
          });
        }
      }),

      configs: Em.Route.extend({
        route: '/configs',
        connectOutlets: function (router, context) {
          router.get('mainHostDetailsController').connectOutlet('mainHostConfigs');
        }
      }),

      alerts: Em.Route.extend({
        route: '/alerts',
        connectOutlets: function (router, context) {
          router.get('mainHostDetailsController').connectOutlet('mainHostAlerts');
        },
        exit: function (router) {
          router.set('mainAlertInstancesController.isUpdating', false);
        }
      }),

      metrics: Em.Route.extend({
        route: '/metrics',
        connectOutlets: function (router, context) {
          router.get('mainHostDetailsController').connectOutlet('mainHostMetrics');
        }
      }),

      stackVersions: Em.Route.extend({
        route: '/stackVersions',
        connectOutlets: function (router, context) {
          if (App.get('stackVersionsAvailable')) {
            router.get('mainHostDetailsController').connectOutlet('mainHostStackVersions');
          }
          else {
            router.transitionTo('summary');
          }
        }
      }),

      hostNavigate: function (router, event) {
        var parent = event.view._parentView;
        parent.deactivateChildViews();
        event.view.set('active', "active");
        router.transitionTo(event.context);
      }
    }),

    back: function (router, event) {
      var referer = router.get('mainHostDetailsController.referer');
      if (referer) {
        router.route(referer);
      }
      else {
        window.history.back();
      }
    },

    addHost: function (router) {
      router.transitionTo('hostAdd');
    }

  }),

  hostAdd: require('routes/add_host_routes'),

  alerts: Em.Route.extend({
    route: '/alerts',
    index: Em.Route.extend({
      route: '/',
      connectOutlets: function (router, context) {
        router.get('mainController').connectOutlet('mainAlertDefinitions');
      }
    }),

    alertDetails: Em.Route.extend({

      route: '/:alert_definition_id',

      connectOutlets: function (router, alertDefinition) {
        App.router.set('mainAlertDefinitionsController.showFilterConditionsFirstLoad', true);
        router.get('mainController').connectOutlet('mainAlertDefinitionDetails', alertDefinition);
      },

      exit: function (router) {
        router.set('mainAlertInstancesController.isUpdating', false);
      },

      unroutePath: function (router, context) {
        var controller = router.get('mainAlertDefinitionDetailsController');
        if (!controller.get('forceTransition') && controller.get('isEditing')) {
          controller.showSavePopup(context);
        } else {
          controller.set('forceTransition', false);
          this._super(router, context);
        }
      }
    }),

    back: function (router, event) {
      window.history.back();
    }
  }),

  alertAdd: require('routes/add_alert_definition_routes'),

  admin: Em.Route.extend({
    route: '/admin',
    enter: function (router, transition) {
      if (router.get('loggedIn') && !App.isAccessible('upgrade_ADMIN')) {
        Em.run.next(function () {
          router.transitionTo('main.dashboard.index');
        });
      }
    },

    routePath: function (router, event) {
      if (!App.isAccessible('upgrade_ADMIN')) {
        Em.run.next(function () {
          App.router.transitionTo('main.dashboard.index');
        });
      } else {
        this._super(router, event);
      }
    },
    connectOutlets: function (router, context) {
      router.get('mainController').connectOutlet('mainAdmin');
    },

    index: Em.Route.extend({
      /* enter: function(router, transition){
       var controller = router.get('mainAdminController');
       router.transitionTo('admin' + controller.get('category').capitalize());
       }, */
      route: '/',
      redirectsTo: 'stackAndUpgrade.index'
    }),

    adminAuthentication: Em.Route.extend({
      route: '/authentication',
      connectOutlets: function (router, context) {
        router.set('mainAdminController.category', "authentication");
        router.get('mainAdminController').connectOutlet('mainAdminAuthentication');
      }
    }),

    adminSecurity: Em.Route.extend({
      route: '/security',
      enter: function (router) {
        router.set('mainAdminController.category', "security");
        var controller = router.get('mainAdminSecurityController');
        if (!(controller.getAddSecurityWizardStatus() === 'RUNNING') && !(controller.getDisableSecurityStatus() === 'RUNNING')) {
          Em.run.next(function () {
            router.transitionTo('adminSecurity.index');
          });
        } else if (controller.getAddSecurityWizardStatus() === 'RUNNING') {
          Em.run.next(function () {
            router.transitionTo('adminAddSecurity');
          });
        } else if (controller.getDisableSecurityStatus() === 'RUNNING') {
          Em.run.next(function () {
            router.transitionTo('disableSecurity');
          });
        }
      },

      index: Em.Route.extend({
        route: '/',
        connectOutlets: function (router, context) {
          var controller = router.get('mainAdminController');
          controller.set('category', "security");
          controller.connectOutlet('mainAdminSecurity');
        }
      }),

      addSecurity: function (router, object) {
        router.get('mainAdminSecurityController').setAddSecurityWizardStatus('RUNNING');
        router.transitionTo('adminAddSecurity');
      },

      adminAddSecurity: require('routes/add_security')
    }),

    adminKerberos: Em.Route.extend({
      route: '/kerberos',
      index: Em.Route.extend({
        route: '/',
        connectOutlets: function (router, context) {
          router.set('mainAdminController.category', "kerberos");
          router.get('mainAdminController').connectOutlet('mainAdminKerberos');
        }
      }),
      adminAddKerberos: require('routes/add_kerberos_routes'),

      disableSecurity: Em.Route.extend({
        route: '/disableSecurity',
        enter: function (router) {
          App.router.get('updateController').set('isWorking', false);
          router.get('mainController').dataLoading().done(function() {
            App.ModalPopup.show({
              classNames: ['full-width-modal'],
              header: Em.I18n.t('admin.removeSecurity.header'),
              bodyClass: App.KerberosDisableView.extend({
                controllerBinding: 'App.router.kerberosDisableController'
              }),
              primary: Em.I18n.t('form.cancel'),
              secondary: null,
              showFooter: false,

              onClose: function () {
                var self = this;
                var controller = router.get('kerberosDisableController');
                if (!controller.get('isSubmitDisabled')) {
                  self.proceedOnClose();
                  return;
                }
                // warn user if disable kerberos command in progress
                var unkerberizeCommand = controller.get('tasks').findProperty('command', 'unkerberize');
                if (unkerberizeCommand && !unkerberizeCommand.get('isCompleted')) {
                  // user cannot exit wizard during removing kerberos
                  if (unkerberizeCommand.get('status') == 'IN_PROGRESS') {
                    App.showAlertPopup(Em.I18n.t('admin.kerberos.disable.unkerberize.header'), Em.I18n.t('admin.kerberos.disable.unkerberize.message'));
                  } else {
                    // otherwise show confirmation window
                    App.showConfirmationPopup(function () {
                      self.proceedOnClose();
                    }, Em.I18n.t('admin.addSecurity.disable.onClose'));
                  }
                } else {
                  self.proceedOnClose();
                }
              },
              proceedOnClose: function () {
                var self = this;
                var disableController = router.get('kerberosDisableController');
                disableController.clearStep();
                disableController.resetDbNamespace();
                App.db.setSecurityDeployCommands(undefined);
                App.router.get('updateController').set('isWorking', true);
                router.get('mainAdminKerberosController').setDisableSecurityStatus(undefined);
                router.get('addServiceController').finish();
                App.clusterStatus.setClusterStatus({
                  clusterName: router.get('content.cluster.name'),
                  clusterState: 'DEFAULT',
                  localdb: App.db.data
                }, {
                  alwaysCallback: function() {
                    self.hide();
                    router.transitionTo('adminKerberos.index');
                    location.reload();
                  }
                });
              },
              didInsertElement: function () {
                this.fitHeight();
              }
            });
          });
        },

        unroutePath: function () {
          return false;
        },
        next: function (router, context) {
          $("#modal").find(".close").trigger('click');
        },
        done: function (router, context) {
          var controller = router.get('kerberosDisableController');
          if (!controller.get('isSubmitDisabled')) {
            $(context.currentTarget).parents("#modal").find(".close").trigger('click');
          }
        }
      })
    }),

    stackAndUpgrade: Em.Route.extend({
      route: '/stack',
      connectOutlets: function (router) {
        router.set('mainAdminController.category', "stackAndUpgrade");
        router.get('mainAdminController').connectOutlet('mainAdminStackAndUpgrade');
      },

      index: Em.Route.extend({
        route: '/',
        redirectsTo: 'services'
      }),

      services: Em.Route.extend({
        route: '/services',
        connectOutlets: function (router, context) {
          router.get('mainAdminStackAndUpgradeController').connectOutlet('mainAdminStackServices');
        }
      }),

      versions: Em.Route.extend({
        route: '/versions',
        connectOutlets: function (router, context) {
          router.get('mainAdminStackAndUpgradeController').connectOutlet('MainAdminStackVersions');
        }
      }),

      stackNavigate: function (router, event) {
        var parent = event.view._parentView;
        parent.deactivateChildViews();
        event.view.set('active', "active");
        router.transitionTo(event.context);
      }
    }),
    stackUpgrade: require('routes/stack_upgrade_routes'),

    adminAdvanced: Em.Route.extend({
      route: '/advanced',
      connectOutlets: function (router) {
        router.set('mainAdminController.category', "advanced");
        router.get('mainAdminController').connectOutlet('mainAdminAdvanced');
      }
    }),
    adminServiceAccounts: Em.Route.extend({
      route: '/serviceAccounts',
      connectOutlets: function (router) {
        router.set('mainAdminController.category', "adminServiceAccounts");
        router.get('mainAdminController').connectOutlet('mainAdminServiceAccounts');
      }
    }),

    adminAudit: Em.Route.extend({
      route: '/audit',
      connectOutlets: function (router) {
        router.set('mainAdminController.category', "audit");
        router.get('mainAdminController').connectOutlet('mainAdminAudit');
      }
    }),
    upgradeStack: function (router, event) {
      if (!$(event.currentTarget).hasClass('inactive')) {
        router.transitionTo('stackUpgrade');
      }
    },


    adminNavigate: function (router, object) {
      router.transitionTo('admin' + object.context.capitalize());
    },

    //events
    goToAdmin: function (router, event) {
      router.transitionTo(event.context);
    }

  }),

  services: Em.Route.extend({
    route: '/services',
    index: Em.Route.extend({
      route: '/',
      enter: function (router) {
        Em.run.next(function () {
          var controller = router.get('mainController');
          controller.dataLoading().done(function () {
            if (router.currentState.parentState.name === 'services' && router.currentState.name === 'index') {
              var service = router.get('mainServiceItemController.content');
              if (!service || !service.get('isLoaded')) {
                service = App.Service.find().objectAt(0); // getting the first service to display
              }
              if (router.get('mainServiceItemController').get('routeToConfigs')) {
                router.transitionTo('service.configs', service);
              } else {
                router.transitionTo('service.summary', service);
              }
            }
          });
        });
      }
    }),
    connectOutlets: function (router, context) {
      router.get('mainController').connectOutlet('mainService');
    },
    service: Em.Route.extend({
      route: '/:service_id',
      connectOutlets: function (router, service) {
        router.get('mainServiceController').connectOutlet('mainServiceItem', service);
        if (service && router.get('mainServiceItemController').get('routeToConfigs')) {
          router.transitionTo('configs');
        } else {
          router.transitionTo('summary');
        }
      },
      index: Ember.Route.extend({
        route: '/'
      }),
      summary: Em.Route.extend({
        route: '/summary',
        connectOutlets: function (router, context) {
          var item = router.get('mainServiceItemController.content');
          //if service is not existed then route to default service
          if (item.get('isLoaded')) {
            router.get('mainServiceItemController').connectOutlet('mainServiceInfoSummary', item);
          } else {
            router.transitionTo('services.index');
          }
        }
      }),
      metrics: Em.Route.extend({
        route: '/metrics',
        connectOutlets: function (router, context) {
          var item = router.get('mainServiceItemController.content');
          router.get('mainServiceItemController').connectOutlet('mainServiceInfoMetrics', item);
        }
      }),
      configs: Em.Route.extend({
        route: '/configs',
        connectOutlets: function (router, context) {
          var item = router.get('mainServiceItemController.content');
          //if service is not existed then route to default service
          if (item.get('isLoaded')) {
            if (router.get('mainServiceItemController.isConfigurable')) {
              router.get('mainServiceItemController').connectOutlet('mainServiceInfoConfigs', item);
            }
            else {
              // if service doesn't have configs redirect to summary
              router.transitionTo('summary');
            }
          }
          else {
            item.set('routeToConfigs', true);
            router.transitionTo('services.index');
          }
        },
        unroutePath: function (router, context) {
          var controller = router.get('mainServiceInfoConfigsController');
          if (!controller.get('forceTransition') && controller.hasUnsavedChanges()) {
            controller.showSavePopup(context);
          } else {
            this._super(router, context);
          }
        }
      }),
      audit: Em.Route.extend({
        route: '/audit',
        connectOutlets: function (router, context) {
          var item = router.get('mainServiceItemController.content');
          router.get('mainServiceItemController').connectOutlet('mainServiceInfoAudit', item);
        }
      }),
      showInfo: function (router, event) {
        var mainServiceInfoConfigsController = App.router.get('mainServiceInfoConfigsController');
        if (event.context === 'summary' && mainServiceInfoConfigsController.hasUnsavedChanges()) {
          mainServiceInfoConfigsController.showSavePopup(router.get('location.lastSetURL').replace('configs', 'summary'));
          return false;
        }
        var parent = event.view.get('_parentView');
        parent.deactivateChildViews();
        event.view.set('active', "active");
        router.transitionTo(event.context);
      }
    }),
    showService: Em.Router.transitionTo('service'),
    addService: Em.Router.transitionTo('serviceAdd'),
    reassign: Em.Router.transitionTo('reassign'),

    enableHighAvailability: require('routes/high_availability_routes'),

    enableRMHighAvailability: require('routes/rm_high_availability_routes'),

    rollbackHighAvailability: require('routes/rollbackHA_routes')
  }),

  reassign: require('routes/reassign_master_routes'),

  serviceAdd: require('routes/add_service_routes'),

  selectService: Em.Route.transitionTo('services.service.summary'),
  selectHost: function (router, event) {
    router.get('mainHostDetailsController').set('isFromHosts', false);
    router.transitionTo('hosts.hostDetails.index', event.context);
  },
  filterHosts: function (router, component) {
    if (!component.context)
      return;
    router.get('mainHostController').filterByComponent(component.context);
    router.get('mainHostController').set('showFilterConditionsFirstLoad', true);
    router.transitionTo('hosts.index');
  },
  showDetails: function (router, event) {
    router.get('mainHostDetailsController').set('referer', router.location.lastSetURL);
    router.get('mainHostDetailsController').set('isFromHosts', true);
    router.transitionTo('hosts.hostDetails.summary', event.context);
  },
  gotoAlertDetails: function (router, event) {
    router.transitionTo('alerts.alertDetails', event.context);
  },

  /**
   * Open summary page of the selected service
   * @param {object} event
   * @method routeToService
   */
  routeToService: function (router, event) {
    var service = event.context;
    router.transitionTo('main.services.service.summary', service);
  }
});

});

require.register("routes/reassign_master_routes", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

module.exports = App.WizardRoute.extend({
  route: '/service/reassign',

  leaveWizard: function (router, context) {
    var reassignMasterController = router.get('reassignMasterController');
    App.router.get('updateController').set('isWorking', true);
    reassignMasterController.finish();
    App.clusterStatus.setClusterStatus({
      clusterName: App.router.get('content.cluster.name'),
      clusterState: 'DEFAULT',
      localdb: App.db.data
    }, {alwaysCallback: function () {
      context.hide();
      router.transitionTo('main.index');
      location.reload();
    }});
  },

  enter: function (router) {
    console.log('in /service/reassign:enter');
    var context = this;
    var reassignMasterController = router.get('reassignMasterController');

    reassignMasterController.dataLoading().done(function () {
      if (App.router.get('mainHostController.hostsCountMap.TOTAL') > 1) {
        Em.run.next(function () {
          App.router.get('updateController').set('isWorking', false);
          var popup = App.ModalPopup.show({
            classNames: ['full-width-modal'],
            header: Em.I18n.t('services.reassign.header'),
            bodyClass: App.ReassignMasterView.extend({
              controller: reassignMasterController
            }),
            primary: Em.I18n.t('form.cancel'),
            showFooter: false,
            secondary: null,

            onPrimary: function () {
              this.hide();
              App.router.get('updateController').set('isWorking', true);
              App.router.transitionTo('main.services.index');
            },
            onClose: function () {
              var currStep = reassignMasterController.get('currentStep');
              if (parseInt(currStep) > 3) {
                var self = this;

                var step4Controller = router.get('reassignMasterWizardStep4Controller');
                var testDBTaskId = step4Controller.get('tasks').filterProperty('command', 'testDBConnection').get('firstObject.id');

                if(currStep !== "7"
                   && testDBTaskId
                   && reassignMasterController.get('content.tasksStatuses').get(testDBTaskId) === "FAILED")
                {
                  App.showConfirmationPopup(function () {
                    App.router.transitionTo('step7');
                  }, Em.I18n.t('services.reassign.rollback.confirm'));
                } else {
                  App.showConfirmationPopup(function () {
                    router.get('reassignMasterWizardStep' + currStep + 'Controller').removeObserver('tasks.@each.status', this, 'onTaskStatusChange');
                    context.leaveWizard(router, self);
                  }, Em.I18n.t('services.reassign.closePopup').format(reassignMasterController.get('content.reassign.display_name')));
                }
              } else {
                context.leaveWizard(router, this);
              }
            },
            didInsertElement: function () {
              this.fitHeight();
            }
          });
          reassignMasterController.set('popup', popup);
          reassignMasterController.loadSecurityEnabled();
          reassignMasterController.loadComponentToReassign();
          var currStep = reassignMasterController.get('currentStep');
          var currentClusterStatus = App.clusterStatus.get('value');
          if (currentClusterStatus) {
            switch (currentClusterStatus.clusterState) {
              case 'REASSIGN_MASTER_INSTALLING' :
                if (currentClusterStatus.localdb.ReassignMaster.currentStep !== currStep) {
                  reassignMasterController.setCurrentStep(currentClusterStatus.localdb.ReassignMaster.currentStep);
                }
                break;
            }
          }
          router.transitionTo('step' + currStep);
        });
      } else {
        App.showAlertPopup(Em.I18n.t('common.error'), Em.I18n.t('services.reassign.error.fewHosts'), function () {
          router.transitionTo('main.services.index');
        })
      }
    });
  },

  step1: Em.Route.extend({
    route: '/step1',
    connectOutlets: function (router) {
      console.log('in reassignMaster.step1:connectOutlets');
      var controller = router.get('reassignMasterController');
      var step1Controller = router.get('reassignMasterWizardStep1Controller');
      controller.setCurrentStep('1');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('reassignMasterWizardStep1', controller.get('content'));
        step1Controller.loadConfigsTags();
      })
    },
    next: function (router) {
      App.db.setMasterComponentHosts(undefined);
      router.transitionTo('step2');
    },

    unroutePath: function () {
      return false;
    }
  }),

  step2: Em.Route.extend({
    route: '/step2',
    connectOutlets: function (router) {
      console.log('in reassignMaster.step2:connectOutlets');
      var controller = router.get('reassignMasterController');
      controller.setCurrentStep('2');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('reassignMasterWizardStep2', controller.get('content'));
      })

    },
    back: Em.Router.transitionTo('step1'),
    next: function (router) {
      var controller = router.get('reassignMasterController');
      var reassignMasterWizardStep2 = router.get('reassignMasterWizardStep2Controller');
      controller.saveMasterComponentHosts(reassignMasterWizardStep2);
      var reassignHosts = {};
      var componentName = reassignMasterWizardStep2.get('content.reassign.component_name');
      var masterAssignmentsHosts = reassignMasterWizardStep2.get('selectedServicesMasters').filterProperty('component_name', componentName).mapProperty('selectedHost');
      var currentMasterHosts = App.HostComponent.find().filterProperty('componentName', componentName).mapProperty('hostName');
      masterAssignmentsHosts.forEach(function (host) {
        if (!currentMasterHosts.contains(host)) {
          reassignHosts.target = host;
        }
      }, this);
      currentMasterHosts.forEach(function (host) {
        if (!masterAssignmentsHosts.contains(host)) {
          reassignHosts.source = host;
        }
      }, this);
      controller.saveReassignHosts(reassignHosts);
      router.transitionTo('step3');
    },

    unroutePath: function () {
      return false;
    }
  }),

  step3: Em.Route.extend({
    route: '/step3',
    connectOutlets: function (router) {
      console.log('in reassignMaster.step3:connectOutlets');
      var controller = router.get('reassignMasterController');
      controller.setCurrentStep('3');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('reassignMasterWizardStep3', controller.get('content'));
      })
    },
    back: Em.Router.transitionTo('step2'),
    next: function (router) {
      App.db.setReassignTasksStatuses(undefined);
      App.db.setReassignTasksRequestIds(undefined);
      App.clusterStatus.setClusterStatus({
        clusterName: router.get('reassignMasterController.content.cluster.name'),
        clusterState: 'REASSIGN_MASTER_INSTALLING',
        wizardControllerName: 'reassignMasterController',
        localdb: App.db.data
      });
      router.transitionTo('step4');
    },

    unroutePath: function () {
      return false;
    }
  }),

  step4: Em.Route.extend({
    route: '/step4',
    connectOutlets: function (router) {
      console.log('in reassignMaster.step4:connectOutlets');
      var controller = router.get('reassignMasterController');
      controller.setCurrentStep('4');
      controller.setLowerStepsDisable(4);
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('reassignMasterWizardStep4', controller.get('content'));
      })
    },
    next: function (router) {
      router.get('reassignMasterController').setCurrentStep('5');

      App.clusterStatus.setClusterStatus({
        clusterName: router.get('reassignMasterController.content.cluster.name'),
        clusterState: 'REASSIGN_MASTER_INSTALLING',
        wizardControllerName: 'reassignMasterController',
        localdb: App.db.data
      });

      router.transitionTo('step5');
    },

    complete: function (router) {
      var controller = router.get('reassignMasterController');
      var reassignMasterWizardStep4 = router.get('reassignMasterWizardStep4Controller');
      if (!reassignMasterWizardStep4.get('isSubmitDisabled')) {
        controller.finish();
        controller.get('popup').hide();
        App.clusterStatus.setClusterStatus({
          clusterName: router.get('reassignMasterController.content.cluster.name'),
          clusterState: 'DEFAULT',
          localdb: App.db.data
        }, {alwaysCallback: function () {
          controller.get('popup').hide();
          router.transitionTo('main.index');
          location.reload();
        }});
      }
    },

    unroutePath: function () {
      return false;
    }
  }),

  step5: Em.Route.extend({
    route: '/step5',
    connectOutlets: function (router) {
      console.log('in reassignMaster.step5:connectOutlets');
      var controller = router.get('reassignMasterController');
      controller.setCurrentStep('5');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.setLowerStepsDisable(5);
        if ((controller.get('content.reassign.component_name') === 'NAMENODE') || controller.get('content.reassign.component_name') === 'SECONDARY_NAMENODE') {
          controller.usersLoading().done(function () {
            controller.connectOutlet('reassignMasterWizardStep5', controller.get('content'));
          })
        } else {
          controller.connectOutlet('reassignMasterWizardStep5', controller.get('content'));
        }
      })
    },
    next: function (router) {
      App.showConfirmationPopup(function () {
        router.transitionTo('step6');
      }, Em.I18n.t('services.reassign.step5.confirmPopup.body'));
    },

    unroutePath: function () {
      return false;
    }
  }),

  step6: Em.Route.extend({
    route: '/step6',
    connectOutlets: function (router) {
      console.log('in reassignMaster.step6:connectOutlets');
      var controller = router.get('reassignMasterController');
      controller.setCurrentStep('6');
      controller.setLowerStepsDisable(6);
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('reassignMasterWizardStep6', controller.get('content'));
      })
    },

    next: function (router) {
      var controller = router.get('reassignMasterController');
      var reassignMasterWizardStep6 = router.get('reassignMasterWizardStep6Controller');
      if (!reassignMasterWizardStep6.get('isSubmitDisabled')) {
        controller.finish();
        controller.get('popup').hide();
        App.clusterStatus.setClusterStatus({
          clusterName: router.get('reassignMasterController.content.cluster.name'),
          clusterState: 'DEFAULT',
          localdb: App.db.data
        }, {alwaysCallback: function () {
          controller.get('popup').hide();
          router.transitionTo('main.index');
          location.reload();
        }});
      }
    },

    unroutePath: function () {
      return false;
    }
  }),

  step7: Em.Route.extend({
    route: '/step7',
    connectOutlets: function (router) {
      console.log('in reassignMaster.step7:connectOutlets');
      var controller = router.get('reassignMasterController');
      controller.setCurrentStep('7');
      controller.setLowerStepsDisable(7);
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('reassignMasterWizardStep7', controller.get('content'));
      });
    },

    next: function (router) {
      var controller = router.get('reassignMasterController');
      var reassignMasterWizardStep7 = router.get('reassignMasterWizardStep7Controller');
      if (!reassignMasterWizardStep7.get('isSubmitDisabled')) {
        controller.finish();
        controller.get('popup').hide();
        App.clusterStatus.setClusterStatus({
          clusterName: router.get('reassignMasterController.content.cluster.name'),
          clusterState: 'DEFAULT',
          localdb: App.db.data
        }, {alwaysCallback: function () {
          controller.get('popup').hide();
          router.transitionTo('main.index');
          location.reload();
        }});
      }
    },

    complete: function (router) {
      var controller = router.get('reassignMasterController');
      var reassignMasterWizardStep7 = router.get('reassignMasterWizardStep7Controller');
      if (!reassignMasterWizardStep7.get('isSubmitDisabled')) {
        controller.finish();
        controller.get('popup').hide();
        App.clusterStatus.setClusterStatus({
          clusterName: router.get('reassignMasterController.content.cluster.name'),
          clusterState: 'DEFAULT',
          localdb: App.db.data
        }, {alwaysCallback: function () {
          controller.get('popup').hide();
          router.transitionTo('main.index');
          location.reload();
        }});
      }
    },

    unroutePath: function () {
      return false;
    }
  }),

  gotoStep7: Em.Router.transitionTo('step7'),

  backToServices: function (router) {
    App.router.get('updateController').set('isWorking', true);
    router.transitionTo('services');
  }

});

});

require.register("routes/rm_high_availability_routes", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

module.exports = App.WizardRoute.extend({
  route: '/highAvailability/ResourceManager/enable',

  enter: function (router,transition) {
    var rMHighAvailabilityWizardController = router.get('rMHighAvailabilityWizardController');
    rMHighAvailabilityWizardController.dataLoading().done(function () {
      //Set YARN as current service
      App.router.set('mainServiceItemController.content', App.Service.find().findProperty('serviceName', 'YARN'));
    });
    Em.run.next(function () {
      App.router.get('updateController').set('isWorking', false);
      var popup = App.ModalPopup.show({
        classNames: ['full-width-modal'],
        header: Em.I18n.t('admin.rm_highAvailability.wizard.header'),
        bodyClass: App.RMHighAvailabilityWizardView.extend({
          controller: rMHighAvailabilityWizardController
        }),
        primary: Em.I18n.t('form.cancel'),
        showFooter: false,
        secondary: null,

        onClose: function () {
          var rMHighAvailabilityWizardController = router.get('rMHighAvailabilityWizardController'),
            currStep = rMHighAvailabilityWizardController.get('currentStep'),
            self = this;

          if (parseInt(currStep) === 4) {
            App.showConfirmationPopup(function () {
              router.get('updateController').set('isWorking', true);
              rMHighAvailabilityWizardController.finish();
              App.clusterStatus.setClusterStatus({
                clusterName: App.router.getClusterName(),
                clusterState: 'DEFAULT',
                localdb: App.db.data
              }, {alwaysCallback: function () {
                self.hide();
                router.transitionTo('main.services.index');
                location.reload();
              }});
            }, Em.I18n.t('admin.rm_highAvailability.closePopup'));
          } else {
            router.get('updateController').set('isWorking', true);
            rMHighAvailabilityWizardController.finish();
            App.clusterStatus.setClusterStatus({
              clusterName: App.router.getClusterName(),
              clusterState: 'DEFAULT',
              localdb: App.db.data
            }, {alwaysCallback: function () {
              self.hide();
              router.transitionTo('main.services.index');
              location.reload();
            }});
          }
        },
        didInsertElement: function () {
          this.fitHeight();
        }
      });
      rMHighAvailabilityWizardController.set('popup', popup);
      var currentClusterStatus = App.clusterStatus.get('value');
      if (currentClusterStatus) {
        switch (currentClusterStatus.clusterState) {
          case 'RM_HIGH_AVAILABILITY_DEPLOY' :
            rMHighAvailabilityWizardController.setCurrentStep(currentClusterStatus.localdb.RMHighAvailabilityWizard.currentStep);
            break;
          default:
            var currStep = App.router.get('rMHighAvailabilityWizardController.currentStep');
            rMHighAvailabilityWizardController.setCurrentStep(currStep);
            break;
        }
      }
      router.transitionTo('step' + rMHighAvailabilityWizardController.get('currentStep'));
    });
  },

  step1: Em.Route.extend({
    route: '/step1',
    connectOutlets: function (router) {
      var controller = router.get('rMHighAvailabilityWizardController');
      controller.setCurrentStep('1');
      controller.dataLoading().done(function () {
        controller.connectOutlet('rMHighAvailabilityWizardStep1', controller.get('content'));
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      router.get('rMHighAvailabilityWizardController').setDBProperty('rmHosts', undefined);
      router.transitionTo('step2');
    }
  }),

  step2: Em.Route.extend({
    route: '/step2',
    connectOutlets: function (router) {
      var controller = router.get('rMHighAvailabilityWizardController');
      controller.setCurrentStep('2');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('rMHighAvailabilityWizardStep2', controller.get('content'));
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      var wizardController = router.get('rMHighAvailabilityWizardController');
      var stepController = router.get('rMHighAvailabilityWizardStep2Controller');
      var currentRM = stepController.get('servicesMasters').findProperty('isAdditional', false);
      var additionalRM = stepController.get('servicesMasters').findProperty('isAdditional', true);
      var rmHost = {
        currentRM: currentRM.get('selectedHost'),
        additionalRM: additionalRM.get('selectedHost')
      };
      wizardController.saveRmHosts(rmHost);
      router.transitionTo('step3');
    },
    back: function (router) {
      router.transitionTo('step1');
    }
  }),

  step3: Em.Route.extend({
    route: '/step3',
    connectOutlets: function (router) {
      var controller = router.get('rMHighAvailabilityWizardController');
      controller.setCurrentStep('3');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('rMHighAvailabilityWizardStep3', controller.get('content'));
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      var wizardController = router.get('rMHighAvailabilityWizardController');
      var stepController = router.get('rMHighAvailabilityWizardStep3Controller');
      var configs = stepController.get('selectedService.configs');
      wizardController.saveConfigs(configs);
      router.transitionTo('step4');
    },
    back: Em.Router.transitionTo('step2')
  }),

  step4: Em.Route.extend({
    route: '/step4',
    connectOutlets: function (router) {
      var controller = router.get('rMHighAvailabilityWizardController');
      controller.setCurrentStep('4');
      controller.setLowerStepsDisable(4);
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('rMHighAvailabilityWizardStep4', controller.get('content'));
      })
    },
    unroutePath: function (router, path) {
      // allow user to leave route if wizard has finished
      if (router.get('rMHighAvailabilityWizardController').get('isFinished')) {
        this._super(router, path);
      } else {
        return false;
      }
    },
    next: function (router) {
      var controller = router.get('rMHighAvailabilityWizardController');
      controller.finish();
      App.clusterStatus.setClusterStatus({
        clusterName: controller.get('content.cluster.name'),
        clusterState: 'DEFAULT',
        localdb: App.db.data
      }, {alwaysCallback: function () {
        controller.get('popup').hide();
        router.transitionTo('main.services.index');
        location.reload();
      }});
    }
  })

});

});

require.register("routes/rollbackHA_routes", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

module.exports = App.WizardRoute.extend({
  route: '/highAvailability/NameNode/rollbackHA',

  enter: function (router) {
    Em.run.next(function () {
      var rollbackHighAvailabilityWizardController = router.get('rollbackHighAvailabilityWizardController');
      App.router.get('updateController').set('isWorking', false);
      var popup = App.ModalPopup.show({
        classNames: ['full-width-modal'],
        header: Em.I18n.t('admin.highAvailability.wizard.rollback.header.title'),
        bodyClass: App.RollbackHighAvailabilityWizardView.extend({
          controller: rollbackHighAvailabilityWizardController
        }),
        primary: Em.I18n.t('form.cancel'),
        showFooter: false,
        secondary: null,
        hideCloseButton: function () {
          var currStep = App.router.get('rollbackHighAvailabilityWizardController.currentStep');
          if(currStep == 3){
            this.set('showCloseButton', false);
          }else{
            this.set('showCloseButton', true);
          }
        }.observes('App.router.rollbackHighAvailabilityWizardController.currentStep'),

        onClose: function () {
          this.hide();
          App.router.get('rollbackHighAvailabilityWizardController').setCurrentStep('1');
          App.router.get('updateController').set('isWorking', true);
          App.router.transitionTo('main.services');
        },
        didInsertElement: function () {
          this.fitHeight();
        }
      });
      rollbackHighAvailabilityWizardController.set('popup', popup);
      var currentClusterStatus = App.clusterStatus.get('value');
      if (currentClusterStatus) {
        switch (currentClusterStatus.clusterState) {
          case 'ROLLBACK_HIGH_AVAILABILITY' :
            rollbackHighAvailabilityWizardController.setCurrentStep(currentClusterStatus.localdb.RollbackHighAvailabilityWizard.currentStep);
            break;
          default:
            var currStep = App.router.get('rollbackHighAvailabilityWizardController.currentStep');
            rollbackHighAvailabilityWizardController.setCurrentStep(currStep);
            break;
        }
      }
      router.transitionTo('step' + rollbackHighAvailabilityWizardController.get('currentStep'));
    });
  },

  step1: Em.Route.extend({
    route: '/step1',
    connectOutlets: function (router) {
      var controller = router.get('rollbackHighAvailabilityWizardController');
      controller.setCurrentStep('1');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('rollbackHighAvailabilityWizardStep1', controller.get('content'));
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      var controller = router.get('rollbackHighAvailabilityWizardController');
      controller.saveSelectedAddNN(controller.get('content.selectedAddNNHost'));
      controller.saveSelectedSNN(controller.get('content.selectedSNNHost'));
      router.transitionTo('step2');
    }
  }),

  step2: Em.Route.extend({
    route: '/step2',
    connectOutlets: function (router) {
      var controller = router.get('rollbackHighAvailabilityWizardController');
      controller.setCurrentStep('2');
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.usersLoading().done(function () {
          controller.connectOutlet('rollbackHighAvailabilityWizardStep2', controller.get('content'));
        });
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      router.transitionTo('step3');
    },
    back: function (router) {
      router.transitionTo('step1');
    }
  }),

  step3: Em.Route.extend({
    route: '/step3',
    connectOutlets: function (router) {
      var controller = router.get('rollbackHighAvailabilityWizardController');
      controller.setCurrentStep('3');
      controller.setLowerStepsDisable(3);
      App.clusterStatus.setClusterStatus({
        clusterName: router.get('content.cluster.name'),
        clusterState: 'ROLLBACK_HIGH_AVAILABILITY',
        wizardControllerName: 'rollbackHighAvailabilityWizardController',
        localdb: App.db.data
      });
      controller.dataLoading().done(function () {
        controller.loadAllPriorSteps();
        controller.connectOutlet('rollbackHighAvailabilityWizardStep3',  controller.get('content'));
      })
    },
    unroutePath: function () {
      return false;
    },
    next: function (router) {
      var controller = router.get('rollbackHighAvailabilityWizardController');
      controller.clearTasksData();
      controller.clearStorageData();
      controller.finish();
      controller.get('popup').hide();
      App.router.get('updateController').set('isWorking', true);
      App.clusterStatus.setClusterStatus({
        clusterName: router.get('content.cluster.name'),
        clusterState: 'DEFAULT',
        wizardControllerName: 'rollbackHighAvailabilityWizardController',
        localdb: App.db.data
      });
      router.transitionTo('main.index');
      location.reload();
    }
  })

});

});

require.register("routes/stack_upgrade_routes", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

module.exports = App.WizardRoute.extend({
  route: 'stack/upgrade',

  enter: function (router) {
    console.log('in /admin/stack/upgrade:enter');
    Ember.run.next(function () {
      //if upgrade id is absent then upgrade is completed
      if (Em.isNone(App.db.get('MainAdminStackAndUpgrade', 'upgradeId'))) {
        router.transitionTo('main.admin.stackAndUpgrade.versions');
        return null;
      }

      App.router.get('updateController').set('isWorking', false);

      return App.ModalPopup.show({
        classNames: ['full-width-modal'],
        header: function () {
          var controller = App.router.get('mainAdminStackAndUpgradeController');
          if (controller.get('isDowngrade')) {
            return Em.I18n.t('admin.stackUpgrade.dialog.downgrade.header').format(controller.get('upgradeVersion'));
          } else {
            return Em.I18n.t('admin.stackUpgrade.dialog.header').format(controller.get('upgradeVersion'));
          }
        }.property('App.router.mainAdminStackAndUpgradeController.upgradeVersion', 'App.router.mainAdminStackAndUpgradeController.isDowngrade'),
        bodyClass: App.upgradeWizardView,
        primary: Em.I18n.t('common.dismiss'),
        secondary: null,
        didInsertElement: function () {
          this.fitHeight();
          this.fitInnerHeight();
        },

        /**
         * fir height of scrollable block inside of modal body
         */
        fitInnerHeight: function () {
          var block = this.$().find('#modal > .modal-body');
          var scrollable = this.$().find('#modal .scrollable-block');

          scrollable.css('max-height', Number(block.css('max-height').slice(0, -2)) - block.height());
          block.css('max-height', 'none');
        },
        onPrimary: function () {
          this.closeWizard();
        },
        onClose: function () {
          this.closeWizard();
        },
        closeWizard: function () {
          App.router.get('updateController').set('isWorking', true);
          App.router.transitionTo('main.admin.stackAndUpgrade.versions');
          this.hide();
        }
      });
    });
  }
});

});

require.register("routes/views", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

module.exports = Em.Route.extend({
  route: '/views',
  enter: function (router) {
    router.get('mainViewsController').loadAmbariViews();
  },
  index: Em.Route.extend({
    route: '/',
    connectOutlets: function (router) {
      router.get('mainViewsController').dataLoading().done(function() {
        router.get('mainController').connectOutlet('mainViews');
      });
    }
  }),
  viewDetails: Em.Route.extend({
    route: '/:viewName/:version/:instanceName',
    connectOutlets: function (router, params) {
      // find and set content for `mainViewsDetails` and associated controller

      var href = ['/views', params.viewName, params.version, params.instanceName].join('/');
      var viewPath = this.parseViewPath(window.location.href.slice(window.location.href.indexOf('?')));
      if (viewPath) {
        href = ['/views', params.viewName, params.version, params.instanceName.slice(0, params.instanceName.lastIndexOf('?'))].join('/');
      }

      router.get('mainViewsController').dataLoading().done(function() {
        var content = App.router.get('mainViewsController.ambariViews').findProperty('href', href);
        if (content) content.set('viewPath', viewPath);
        router.get('mainController').connectOutlet('mainViewsDetails', content);
      });
    },
    /**
     * parse internal view path
     * "viewPath" - used as a key of additional path
     * Example:
     *   origin URL: viewName?viewPath=%2Fuser%2Fadmin%2Faddress&foo=bar&count=1
     * should be translated to
     *   view path: /user/admin/address?foo=bar&count=1
     *
     * @param {string} instanceName
     * @returns {string}
     */
    parseViewPath: function (instanceName) {
      var path = '';
      if (instanceName.contains('?')) {
        path = instanceName.slice(instanceName.indexOf('?'));
        if (path.contains('viewPath')) {
          path = decodeURIComponent(path.slice((path.lastIndexOf('?viewPath=') + 10))).replace('&', '?');
        }
      }
      return path;
    }
  })
});

});

require.register("templates", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */



// load templates here

require('templates/main/service/info/summary/base');
require('templates/common/progress');

});

require.register("templates/application", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            <a ");
  stack1 = {};
  stack2 = "topnav.logo.href";
  stack1['href'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + " class=\"logo\"><img src=\"/img/logo-white.png\" alt=\"Apache Ambari\"\n                                                                           title=\"Apache Ambari\"></a>\n            <a class=\"brand\" ");
  stack1 = {};
  stack2 = "topnav.logo.href";
  stack1['href'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + " title=\"Apache Ambari\">");
  stack1 = depth0;
  stack2 = "app.name";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n            <a class=\"brand cluster-name\" href=\"javascript:void(null);\" ");
  stack1 = {};
  stack2 = "clusterName";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <span ");
  stack1 = depth0;
  stack2 = "showPopup";
  stack3 = {};
  stack4 = "App.router.backgroundOperationsController";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " >");
  stack1 = depth0;
  stack2 = "clusterDisplayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " </span>\n              ");
  stack1 = depth0;
  stack2 = "App.router.backgroundOperationsController";
  stack3 = helpers['with'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "App.router.mainAlertDefinitionsController.unhealthyAlertInstancesCount";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(9, program9, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </a>\n          ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "allOperationsCount";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(5, program5, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                  <i class=\"icon-caret-left ops-count\"></i><span id=\"span-bg-operation-count\"\n                                                                 class=\"label operations-count\" ");
  stack1 = depth0;
  stack2 = "showPopup";
  stack3 = {};
  stack4 = "App.router.backgroundOperationsController";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                    ");
  stack1 = depth0;
  stack2 = "allOperationsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "allOperationsCount";
  stack3 = {};
  stack4 = "t:op";
  stack3['singular'] = stack4;
  stack4 = "t:ops";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n                ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                  <i class=\"icon-caret-left\"></i><span id=\"span-bg-operation-count\"\n                                                       class=\"label\" ");
  stack1 = depth0;
  stack2 = "showPopup";
  stack3 = {};
  stack4 = "App.router.backgroundOperationsController";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                    ");
  stack1 = depth0;
  stack2 = "allOperationsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "allOperationsCount";
  stack3 = {};
  stack4 = "t:op";
  stack3['singular'] = stack4;
  stack4 = "t:ops";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n                ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                <span ");
  stack1 = {};
  stack2 = ":label App.router.mainAlertDefinitionsController.isCriticalAlerts:alert-crit-count:alert-warn-count";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "showPopup";
  stack3 = {};
  stack4 = "App.router.mainAlertInstancesController";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                ");
  stack1 = depth0;
  stack2 = "App.router.mainAlertDefinitionsController.unhealthyAlertInstancesCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "App.router.mainAlertDefinitionsController.unhealthyAlertInstancesCount";
  stack3 = {};
  stack4 = "alert";
  stack3['singular'] = stack4;
  stack4 = "alerts";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n              ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                <span rel=\"tooltip\" ");
  stack1 = {};
  stack2 = "titlebar.alerts.noAlerts";
  stack1['title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + "\n                      class=\"label alerts-none-count\" ");
  stack1 = depth0;
  stack2 = "showPopup";
  stack3 = {};
  stack4 = "App.router.mainAlertInstancesController";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                ");
  stack1 = depth0;
  stack2 = "App.router.mainAlertDefinitionsController.unhealthyAlertInstancesCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "App.router.mainAlertDefinitionsController.unhealthyAlertInstancesCount";
  stack3 = {};
  stack4 = "alert";
  stack3['singular'] = stack4;
  stack4 = "alerts";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n              ");
  return buffer;}

function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <a class=\"logo\"><img src=\"/img/logo-white.png\" alt=\"Apache Ambari\" title=\"Apache Ambari\"></a>\n            <a class=\"brand\" title=\"Apache Ambari\">");
  stack1 = depth0;
  stack2 = "app.name";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n          ");
  return buffer;}

function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            <div class=\"top-nav-user btn-group\">\n              <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n                <i class=\"icon-user\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "App.router.loginName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "&nbsp;<span class=\"caret\"></span>\n              </button>\n              <ul class=\"dropdown-menu\">\n                <li><a href=\"\" id=\"about\" ");
  stack1 = depth0;
  stack2 = "showAboutPopup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "app.aboutAmbari";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                ");
  stack1 = depth0;
  stack2 = "App.router.clusterInstallCompleted";
  stack3 = helpers['if'];
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(18, program18, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "isClusterDataLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(21, program21, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                <li class=\"break\"></li>\n                <li><a href=\"\" id=\"sign-out\" ");
  stack1 = depth0;
  stack2 = "logoff";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "app.signout";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              </ul>\n            </div>\n          ");
  return buffer;}
function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "isClusterDataLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "upgrade_ONLY_ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(16, program16, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program16(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                      <li><a href=\"\"\n                             id=\"manage-ambari\" ");
  stack1 = depth0;
  stack2 = "goToAdminView";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "app.manageAmbari";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                      </li>\n                    ");
  return buffer;}

function program18(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "upgrade_ONLY_ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(19, program19, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program19(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                    <li><a href=\"\"\n                           id=\"manage-ambari\" ");
  stack1 = depth0;
  stack2 = "goToAdminView";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "app.manageAmbari";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                    </li>\n                  ");
  return buffer;}

function program21(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "upgrade_ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(22, program22, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program22(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                    <li><a href=\"\" ");
  stack1 = depth0;
  stack2 = "showSettingsPopup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "app.settings";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                  ");
  return buffer;}

function program24(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <div class=\"ru-badge span12\">\n          <div class=\"navbar navbar-static-top clearfix\">\n            <div class=\"span4 offset4\">\n              ");
  stack1 = depth0;
  stack2 = "isExistingClusterDataLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(25, program25, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </div>\n          </div>\n        </div>\n      ");
  return buffer;}
function program25(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <a class=\"brand cluster-name\" href=\"#\">\n                  ");
  stack1 = depth0;
  stack2 = "App.upgradeInProgress";
  stack3 = helpers['if'];
  tmp1 = self.program(26, program26, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "App.upgradeHolding";
  stack3 = helpers['if'];
  tmp1 = self.program(31, program31, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </a>\n              ");
  return buffer;}
function program26(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "App.router.mainAdminStackAndUpgradeController.isDowngrade";
  stack3 = helpers['if'];
  tmp1 = self.program(27, program27, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(29, program29, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program27(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                      <span\n                          class=\"label upgrade-in-progress\" ");
  stack1 = depth0;
  stack2 = "openUpgradeDialog";
  stack3 = {};
  stack4 = "App.router.mainAdminStackAndUpgradeController";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                             <i class=\"icon-cog\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.stackVersions.version.downgrade.running";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n                    ");
  return buffer;}

function program29(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                      <span\n                          class=\"label upgrade-in-progress\" ");
  stack1 = depth0;
  stack2 = "openUpgradeDialog";
  stack3 = {};
  stack4 = "App.router.mainAdminStackAndUpgradeController";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                             <i class=\"icon-cog\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.stackVersions.version.upgrade.running";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n                    ");
  return buffer;}

function program31(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "App.router.mainAdminStackAndUpgradeController.isDowngrade";
  stack3 = helpers['if'];
  tmp1 = self.program(32, program32, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(34, program34, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program32(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                      <span\n                          class=\"label upgrade-holding\" ");
  stack1 = depth0;
  stack2 = "openUpgradeDialog";
  stack3 = {};
  stack4 = "App.router.mainAdminStackAndUpgradeController";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                          <i class=\"icon-pause\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.stackVersions.version.downgrade.pause";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n                    ");
  return buffer;}

function program34(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                      <span\n                          class=\"label upgrade-holding\" ");
  stack1 = depth0;
  stack2 = "openUpgradeDialog";
  stack3 = {};
  stack4 = "App.router.mainAdminStackAndUpgradeController";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                          <i class=\"icon-pause\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.stackVersions.version.upgrade.pause";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n                    ");
  return buffer;}

  data.buffer.push("\n\n<div id=\"main\">\n  <div id=\"top-nav\">\n    <div class=\"navbar navbar-static-top\">\n      <div class=\"navbar-inner\">\n        <div class=\"container main-container\">\n          ");
  stack1 = depth0;
  stack2 = "isExistingClusterDataLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(11, program11, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n          ");
  stack1 = depth0;
  stack2 = "App.router.loggedIn";
  stack3 = helpers['if'];
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n          ");
  stack1 = depth0;
  stack2 = "App.MainMenuView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n\n        </div>\n      </div>\n    </div>\n  </div>\n  <div class=\"container main-container\">\n    <div id=\"content\">\n      ");
  stack1 = depth0;
  stack2 = "App.upgradeIsRunning";
  stack3 = helpers['if'];
  tmp1 = self.program(24, program24, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/about", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div class=\"about\">\n  <div class=\"logo\">\n    <img src=\"/img/logo.png\" alt=\"Apache Ambari\" title=\"Apache Ambari\" />\n  </div>\n  <div class=\"content\">\n    <div class=\"project\">Apache Ambari</div>\n      <br />\n      ");
  stack1 = depth0;
  stack2 = "app.aboutAmbari.version";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "view.ambariVersion";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      <br />\n      <br />\n      <a href=\"http://ambari.apache.org/\" target=\"_blank\">");
  stack1 = depth0;
  stack2 = "app.aboutAmbari.getInvolved";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n      <br />\n      <a href=\"http://www.apache.org/licenses/LICENSE-2.0\" target=\"_blank\">");
  stack1 = depth0;
  stack2 = "app.aboutAmbari.licensed";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n  </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/chart/linear_time", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div ");
  stack1 = {};
  stack2 = "view.leftArrowVisible:visibleArrow :arrow-left";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "switchTimeBack";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n  <div ");
  stack1 = {};
  stack2 = "view.rightArrowVisible:visibleArrow :arrow-right";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "forward";
  stack3 = depth0;
  stack4 = "switchTimeForward";
  stack5 = {};
  stack6 = "view.parentView";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n");
  return buffer;}

  data.buffer.push("\n\n<div ");
  stack1 = {};
  stack2 = "view.isReady:hide:show :screensaver :no-borders :chart-container";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n<div ");
  stack1 = {};
  stack2 = "view.isReady::hidden :time-label";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.parentView.currentTimeState.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n");
  stack1 = depth0;
  stack2 = "view.isTimePagingEnable";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n<div ");
  stack1 = {};
  stack2 = "view.containerId";
  stack1['id'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.isReady:show:hide view.containerClass :chart-container";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  <div ");
  stack1 = {};
  stack2 = "view.yAxisId";
  stack1['id'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.yAxisClass :chart-y-axis";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n  <div ");
  stack1 = {};
  stack2 = "view.xAxisId";
  stack1['id'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.xAxisClass :chart-x-axis";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n  <div ");
  stack1 = {};
  stack2 = "view.legendId";
  stack1['id'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.legendClass :chart-legend";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n  <div ");
  stack1 = {};
  stack2 = "view.chartId";
  stack1['id'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.chartClass :chart";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n  <div ");
  stack1 = {};
  stack2 = "view.titleId";
  stack1['id'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.titleClass :chart-title";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n</div>\n");
  stack1 = depth0;
  stack2 = "view.isTimePagingEnable";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/addPropertyWindow", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div ");
  stack1 = {};
  stack2 = "view.serviceConfigObj.bulkConfigError:error :each-row :control-group";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <label class=\"control-label\">\n        ");
  stack1 = depth0;
  stack2 = "services.service.config.addPropertyWindow.properties";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        <br>\n        <small>");
  stack1 = depth0;
  stack2 = "services.service.config.addPropertyWindow.propertiesHelper";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</small>\n      </label>\n\n      <div class=\"controls\">\n        ");
  stack1 = depth0;
  stack2 = "Ember.TextArea";
  stack3 = {};
  stack4 = "view.serviceConfigObj.bulkConfigValue";
  stack3['valueBinding'] = stack4;
  stack4 = "4";
  stack3['rows'] = stack4;
  stack4 = "input-block-level";
  stack3['classNames'] = stack4;
  stack4 = "services.service.config.addPropertyWindow.propertiesPlaceholder";
  stack3['placeholderTranslation'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "view.serviceConfigObj.bulkConfigErrorMessage";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</span>\n      </div>\n    </div>\n  ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div ");
  stack1 = {};
  stack2 = "view.serviceConfigObj.isKeyError:error :each-row :control-group";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "common.key";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n      <div class=\"controls\">\n        ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "view.serviceConfigObj.name";
  stack3['valueBinding'] = stack4;
  stack4 = "input-block-level";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.serviceConfigObj.isKeyError";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.serviceConfigObj.showFilterLink";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n    </div>\n    <div class=\"each-row control-group\">\n      <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "common.value";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n      <div class=\"controls\">\n        ");
  stack1 = depth0;
  stack2 = "Ember.TextArea";
  stack3 = {};
  stack4 = "view.serviceConfigObj.value";
  stack3['valueBinding'] = stack4;
  stack4 = "4";
  stack3['rows'] = stack4;
  stack4 = "input-block-level";
  stack3['classNames'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n  ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "view.serviceConfigObj.errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n        ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.notMisc";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          <a href=\"#\" class=\"btn-padding\" ");
  stack1 = depth0;
  stack2 = "filterByKey";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            ");
  stack1 = depth0;
  stack2 = "services.service.config.addPropertyWindow.filterKeyLink";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n          </a>\n            ");
  return buffer;}

  data.buffer.push("\n<form class=\"form-horizontal add-property-window\" autocomplete=\"off\">\n  <div class=\"each-row control-group\">\n    <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "common.type";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n    <div class=\"controls\">\n      <input class=\"span4\" type=\"text\" disabled ");
  stack1 = {};
  stack2 = "view.fileName";
  stack1['value'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "/>\n\n      <div class=\"btn-group add-mode pull-right\" ");
  stack1 = depth0;
  stack2 = "toggleBulkMode";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <a href=\"#\"\n          ");
  stack1 = {};
  stack2 = ":btn view.serviceConfigObj.isBulkMode::active";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = {};
  stack2 = "services.service.config.addPropertyWindow.singleMode";
  stack1['title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + "\n           data-toggle=\"tooltip\"><i class=\"icon-tag\"></i></a>\n        <a href=\"#\"\n          ");
  stack1 = {};
  stack2 = ":btn view.serviceConfigObj.isBulkMode:active";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = {};
  stack2 = "services.service.config.addPropertyWindow.bulkMode";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + "\n           data-toggle=\"tooltip\"><i class=\"icon-tags\"></i></a>\n      </div>\n    </div>\n  </div>\n  ");
  stack1 = depth0;
  stack2 = "view.serviceConfigObj.isBulkMode";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</form>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/compare_property", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div ");
  stack1 = {};
  stack2 = ":control-group :overrideField";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      ");
  stack1 = depth0;
  stack2 = "compareConfig.viewClass";
  stack3 = {};
  stack4 = "compareConfig";
  stack3['serviceConfigBinding'] = stack4;
  stack4 = "view.parentView.categoryConfigsAll";
  stack3['categoryConfigsAllBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      <span class=\"label label-info\">");
  stack1 = depth0;
  stack2 = "compareConfig.serviceVersion.versionText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n      ");
  stack1 = depth0;
  stack2 = "compareConfig.serviceVersion.isCurrent";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "compareConfig.isMock";
  stack3 = helpers.unless;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <span class=\"label label-success\">");
  stack1 = depth0;
  stack2 = "common.current";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n      ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "compareConfig.supportsFinal";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2;
  data.buffer.push("\n            <a disabled=\"disabled\" ");
  stack1 = {};
  stack2 = ":btn :btn-small :btn-final compareConfig.isFinal:active compareConfig.hideFinalIcon:hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                <i class=\"icon-lock\"></i>\n            </a>\n        ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.serviceConfigProperty.compareConfigs";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "compareConfig";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/config_history_flow", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.serviceVersionBox";
  stack3 = {};
  foundHelper = helpers.serviceVersion;
  stack4 = foundHelper || depth0.serviceVersion;
  stack3['serviceVersionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, stack8;
  data.buffer.push("\n      <div ");
  stack1 = {};
  stack2 = "view.showCompareVersionBar::hidden :version-info-bar";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <div class=\"row-fluid\">\n              <div class=\"span1 remove-compare-bar\" ");
  stack1 = depth0;
  stack2 = "removeCompareVersionBar";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " data-toggle=\"arrow-tooltip\" ");
  stack1 = {};
  stack2 = "services.service.config.configHistory.dismissIcon.tooltip";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n                  <i class=\"icon-remove-circle icon-large\"></i>\n              </div>\n              <div class=\"label-wrapper span8\"\n                   data-toggle=\"tooltip\" ");
  stack1 = {};
  stack2 = "view.compareServiceVersion.fullNotes";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  ");
  stack1 = depth0;
  stack2 = "services.service.config.configHistory.comparing";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                  <span class=\"label label-info\">");
  stack1 = depth0;
  stack2 = "view.displayedServiceVersion.versionText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n                  ...\n                  <span class=\"label label-info\">");
  stack1 = depth0;
  stack2 = "view.compareServiceVersion.versionText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n                ");
  stack1 = depth0;
  stack2 = "view.compareServiceVersion.isCurrent";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  <strong>");
  stack1 = depth0;
  stack2 = "view.compareServiceVersion.author";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong>&nbsp;");
  stack1 = depth0;
  stack2 = "dashboard.configHistory.info-bar.authoredOn";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                  &nbsp;<strong>");
  stack1 = depth0;
  stack2 = "view.compareServiceVersion.createdDate";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong>\n              </div>\n              <div class=\"pull-right operations-button\">\n                  <button class=\"btn btn-success\"  ");
  stack1 = depth0;
  stack2 = "view.actionTypes.REVERT";
  stack3 = depth0;
  stack4 = "view.serviceVersionsReferences.compare";
  stack5 = depth0;
  stack6 = "doAction";
  stack7 = {};
  stack8 = "view";
  stack7['target'] = stack8;
  stack8 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack7;
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack8.call(depth0, stack6, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.versionActionsDisabled";
  stack1['disabled'] = stack2;
  stack2 = "view.compareServiceVersion.isCurrent:hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.compareServiceVersion.makeCurrentButtonText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</button>\n              </div>\n          </div>\n      </div>\n    ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    <span class=\"label label-success\">");
  stack1 = depth0;
  stack2 = "common.current";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n                ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, stack8;
  data.buffer.push("\n                  <li ");
  stack1 = {};
  stack2 = ":pointer :dropdown-submenu serviceVersion.isDisplayed:not-allowed";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                    <div class=\"row-fluid version-in-dropdown \" ");
  stack1 = depth0;
  stack2 = "view.actionTypes.SWITCH";
  stack3 = depth0;
  stack4 = "serviceVersion";
  stack5 = depth0;
  stack6 = "doAction";
  stack7 = {};
  stack8 = "view";
  stack7['target'] = stack8;
  stack8 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack7;
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack8.call(depth0, stack6, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                      <div class=\"span2\">");
  stack1 = depth0;
  stack2 = "serviceVersion.versionText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                      <div class=\"span6\">");
  stack1 = depth0;
  stack2 = "serviceVersion.createdDate";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                      <div class=\"span3\">");
  stack1 = depth0;
  stack2 = "serviceVersion.author";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                      <div class=\"pull-right\"><i class=\"icon-caret-right\"></i></div>\n                    </div>\n                    <ul class=\"dropdown-menu version-info-operations\">\n                      <div class=\"content\">\n                        <span class=\"label label-info\">");
  stack1 = depth0;
  stack2 = "serviceVersion.versionText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n                        <div class=\"group\"><strong>");
  stack1 = depth0;
  stack2 = "serviceVersion.configGroupName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong></div>\n                        <div class=\"date\"><strong>");
  stack1 = depth0;
  stack2 = "serviceVersion.author";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong>&nbsp;");
  stack1 = depth0;
  stack2 = "dashboard.configHistory.info-bar.authoredOn";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "&nbsp;<strong>");
  stack1 = depth0;
  stack2 = "serviceVersion.createdDate";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong></div>\n                        <div class=\"notes\">");
  stack1 = depth0;
  stack2 = "serviceVersion.fullNotes";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</div>\n                      </div>\n                      <div class=\"version-operations-buttons\">\n                        <button ");
  stack1 = {};
  stack2 = "serviceVersion.disabledActionAttr.view";
  stack1['disabled'] = stack2;
  stack2 = ":btn serviceVersion.isDisplayed:not-allowed-cursor";
  stack1['class'] = stack2;
  stack2 = "serviceVersion.disabledActionMessages.view";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.actionTypes.SWITCH";
  stack3 = depth0;
  stack4 = "serviceVersion";
  stack5 = depth0;
  stack6 = "doAction";
  stack7 = {};
  stack8 = "view";
  stack7['target'] = stack8;
  stack8 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack7;
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack8.call(depth0, stack6, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-search\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "common.view";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n                        ");
  stack1 = depth0;
  stack2 = "MANAGER";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                      </div>\n                    </ul>\n                  </li>\n                ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, stack8;
  data.buffer.push("\n                        <button ");
  stack1 = {};
  stack2 = "serviceVersion.disabledActionAttr.compare";
  stack1['disabled'] = stack2;
  stack2 = ":btn serviceVersion.isDisplayed:not-allowed-cursor";
  stack1['class'] = stack2;
  stack2 = "serviceVersion.disabledActionMessages.compare";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.actionTypes.COMPARE";
  stack3 = depth0;
  stack4 = "serviceVersion";
  stack5 = depth0;
  stack6 = "doAction";
  stack7 = {};
  stack8 = "view";
  stack7['target'] = stack8;
  stack8 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack7;
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack8.call(depth0, stack6, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-copy\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "common.compare";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n                        <button ");
  stack1 = {};
  stack2 = "serviceVersion.disabledActionAttr.revert";
  stack1['disabled'] = stack2;
  stack2 = ":btn serviceVersion.isCurrent:not-allowed-cursor";
  stack1['class'] = stack2;
  stack2 = "serviceVersion.disabledActionMessages.revert";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.actionTypes.REVERT";
  stack3 = depth0;
  stack4 = "serviceVersion";
  stack5 = depth0;
  stack6 = "doAction";
  stack7 = {};
  stack8 = "view";
  stack7['target'] = stack8;
  stack8 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack7;
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack8.call(depth0, stack6, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "dashboard.configHistory.info-bar.revert.button";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n                        ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                    <li class=\"align-center pointer\" id=\"show_more\">\n                        <a ");
  stack1 = depth0;
  stack2 = "openFullList";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                          ");
  stack1 = depth0;
  stack2 = "dashboard.configHistory.info-bar.showMore";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "&nbsp;");
  stack1 = depth0;
  stack2 = "view.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                            &nbsp;<span class=\"lowercase ellipsis\">");
  stack1 = depth0;
  stack2 = "dashboard.configHistory.title";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n                        </a>\n                    </li>\n                ");
  return buffer;}

function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <span class=\"label label-info\">");
  stack1 = depth0;
  stack2 = "view.displayedServiceVersion.versionText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n              ");
  return buffer;}

function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  <span class=\"label label-success\">");
  stack1 = depth0;
  stack2 = "common.current";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n              ");
  return buffer;}

function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <strong>");
  stack1 = depth0;
  stack2 = "view.displayedServiceVersion.author";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong>&nbsp;");
  stack1 = depth0;
  stack2 = "dashboard.configHistory.info-bar.authoredOn";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "&nbsp;<strong>");
  stack1 = depth0;
  stack2 = "view.displayedServiceVersion.createdDate";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong>\n              ");
  return buffer;}

function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, stack8;
  data.buffer.push("\n            <div class=\"pull-right operations-button\">\n                <div ");
  stack1 = {};
  stack2 = "view.displayedServiceVersion.isCurrent::hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                    <button class=\"btn\" ");
  stack1 = depth0;
  stack2 = "doCancel";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.isDiscardDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.discard";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n                    <button class=\"btn btn-success\" ");
  stack1 = depth0;
  stack2 = "save";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.isSaveDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.save";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n                </div>\n                <button class=\"btn btn-success\"  ");
  stack1 = depth0;
  stack2 = "view.actionTypes.REVERT";
  stack3 = depth0;
  stack4 = "view.serviceVersionsReferences.displayed";
  stack5 = depth0;
  stack6 = "doAction";
  stack7 = {};
  stack8 = "view";
  stack7['target'] = stack8;
  stack8 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack7;
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack8.call(depth0, stack6, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.versionActionsDisabled";
  stack1['disabled'] = stack2;
  stack2 = "view.displayedServiceVersion.isCurrent:hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.displayedServiceVersion.makeCurrentButtonText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</button>\n            </div>\n          ");
  return buffer;}

  data.buffer.push("\n\n\n<div id=\"config_history_flow\" ");
  stack1 = {};
  stack2 = "view.showCompareVersionBar:two-stories-bar:one-story-bar";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  <div class=\"version-slider\">\n    <div ");
  stack1 = {};
  stack2 = ":icon-chevron-box :pull-left view.showLeftArrow::disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "shiftBack";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " data-toggle=\"arrow-tooltip\"\n      ");
  stack1 = {};
  stack2 = "view.leftArrowTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-chevron-left icon-3x\"></i></div>\n    <div ");
  stack1 = {};
  stack2 = ":icon-chevron-box :pull-left view.showRightArrow::disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "shiftForward";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " data-toggle=\"arrow-tooltip\"\n      ");
  stack1 = {};
  stack2 = "view.rightArrowTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-chevron-right icon-3x\"></i></div>\n    ");
  stack1 = depth0;
  stack2 = "view.visibleServiceVersion";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "serviceVersion";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n  </div>\n  <div class=\"version-info-bar-wrapper\">\n    ");
  stack1 = depth0;
  stack2 = "MANAGER";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      <div class=\"version-info-bar\">\n        <div class=\"row-fluid\">\n            <div class=\"btn-group pull-left\">\n              <button id=\"toggle-dropdown-button\" class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\" ");
  stack1 = depth0;
  stack2 = "hideFullList";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.versionActionsDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                <i class=\"icon-random\"></i>\n                <span class=\"caret\"></span>\n              </button>\n              <ul class=\"dropdown-menu\">\n                ");
  stack1 = depth0;
  stack2 = "view.dropDownList";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "serviceVersion";
  stack7 = helpers.each;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "view.showFullList";
  stack3 = helpers.unless;
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </ul>\n            </div>\n            <div class=\"label-wrapper span8\" data-toggle=\"tooltip\" ");
  stack1 = {};
  stack2 = "view.displayedServiceVersion.fullNotes";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "view.displayedServiceVersion.versionText";
  stack3 = helpers['if'];
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "view.displayedServiceVersion.isCurrent";
  stack3 = helpers['if'];
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "view.displayedServiceVersion.author";
  stack3 = helpers['if'];
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </div>\n          ");
  stack1 = depth0;
  stack2 = "MANAGER";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </div>\n      </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/config_recommendation_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "service.configs";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "property";
  stack7 = helpers.each;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "property.warn";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "property.error";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <tr>\n              <td>");
  stack1 = depth0;
  stack2 = "property.serviceName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n              <td>");
  stack1 = depth0;
  stack2 = "property.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n              <td>");
  stack1 = depth0;
  stack2 = "property.value";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n              <td>\n                <div class=\"property-message\">");
  stack1 = depth0;
  stack2 = "property.warnMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                <div class=\"property-description\">");
  stack1 = depth0;
  stack2 = "property.description";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n              </td>\n            </tr>\n          ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "property.isVisible";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              <tr>\n                <td>");
  stack1 = depth0;
  stack2 = "property.serviceName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n                <td>");
  stack1 = depth0;
  stack2 = "property.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n                <td>");
  stack1 = depth0;
  stack2 = "property.value";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n                <td>\n                  <div class=\"property-message\">");
  stack1 = depth0;
  stack2 = "property.errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                  <div class=\"property-description\">");
  stack1 = depth0;
  stack2 = "property.description";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                </td>\n              </tr>\n            ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <tr>\n          <td>");
  stack1 = depth0;
  stack2 = "message.serviceName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n          <td>");
  stack1 = depth0;
  stack2 = "message.propertyName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n          <td colspan=\"2\">");
  stack1 = depth0;
  stack2 = "message.warnMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " in ");
  stack1 = depth0;
  stack2 = "message.filename";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n        </tr>\n      ");
  return buffer;}

  data.buffer.push("\n\n<p>");
  stack1 = depth0;
  stack2 = "installer.step7.popup.validation.warning.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</p>\n<div id=\"config-validation-warnings\" class=\"limited-height-2\">\n  <table class=\"table no-borders\">\n    <thead>\n    <tr>\n      <th>");
  stack1 = depth0;
  stack2 = "common.service";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "common.property";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "common.value";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "common.description";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n    </tr>\n    </thead>\n    <tbody>\n      ");
  stack1 = depth0;
  stack2 = "stepConfigs";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "service";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "configValidationGlobalMessage";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "message";
  stack7 = helpers.each;
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </tbody>\n  </table>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/notifications_configs", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <div ");
  stack1 = {};
  stack2 = ":entry-row rowStyleClass isUserProperty:indent-1";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <span ");
  stack1 = {};
  stack2 = "errorMessage:error: :control-group :control-label-span";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <label class=\"control-label\">\n              ");
  stack1 = depth0;
  stack2 = "displayName";
  foundHelper = helpers.formatWordBreak;
  stack3 = foundHelper || depth0.formatWordBreak;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "formatWordBreak", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </label>\n          </span>\n\n          <div ");
  stack1 = {};
  stack2 = "showLabel:controls";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <div ");
  stack1 = {};
  stack2 = "errorMessage:error: warnMessage:warning: :control-group";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "viewClass";
  stack3 = {};
  stack4 = "this";
  stack3['serviceConfigBinding'] = stack4;
  stack4 = "view.categoryConfigsAll";
  stack3['categoryConfigsAllBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n\n              ");
  stack1 = depth0;
  stack2 = "isRemovable";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n              <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n              <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "warnMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n            </div>\n          </div>\n        </div>\n      ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "isComparison";
  stack3 = helpers.unless;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                    <a class=\"btn btn-small\" href=\"#\" data-toggle=\"tooltip\"\n                      ");
  stack1 = depth0;
  stack2 = "";
  stack3 = depth0;
  stack4 = "removeProperty";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                      ");
  stack1 = {};
  stack2 = "common.remove";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n                      <i class=\"icon-minus-sign\"></i>\n                    </a>\n                  ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.canEdit";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.configsAreDisabled";
  stack3 = helpers.unless;
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <div class=\"entry-row indent-1\">\n             <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "";
  stack3 = depth0;
  stack4 = "showAddPropertyWindow";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " >");
  stack1 = depth0;
  stack2 = "installer.step7.config.addProperty";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "...</a>\n            </div>\n          ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"accordion-heading\" ");
  stack1 = depth0;
  stack2 = "category";
  stack3 = depth0;
  stack4 = "onToggleBlock";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  <i ");
  stack1 = {};
  stack2 = ":pull-left :accordion-toggle view.category.isCollapsed:icon-caret-right:icon-caret-down";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n  <a class=\"accordion-toggle\">");
  stack1 = depth0;
  stack2 = "view.category.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n</div>\n\n<div class=\"accordion-body collapse in\" ");
  stack1 = {};
  stack2 = "view.isCategoryBodyVisible";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  <div class=\"accordion-inner service-config-section\">\n    <form class=\"form-horizontal\" autocomplete=\"off\">\n\n      <div class=\"entry-row\">\n        ");
  stack1 = depth0;
  stack2 = "Ember.RadioButton";
  stack3 = {};
  stack4 = "createNotification";
  stack3['name'] = stack4;
  stack4 = "view.createNotification";
  stack3['selectionBinding'] = stack4;
  stack4 = "no";
  stack3['value'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "installer.step7.misc.notification.configure.later";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n\n      <div class=\"entry-row\">\n        ");
  stack1 = depth0;
  stack2 = "Ember.RadioButton";
  stack3 = {};
  stack4 = "createNotification";
  stack3['name'] = stack4;
  stack4 = "view.createNotification";
  stack3['selectionBinding'] = stack4;
  stack4 = "yes";
  stack3['value'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "installer.step7.misc.notification.configure";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n\n      ");
  stack1 = depth0;
  stack2 = "view.categoryConfigs";
  stack3 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n      <div class=\"entry-row indent-1\">\n        ");
  stack1 = depth0;
  stack2 = "Ember.RadioButton";
  stack3 = {};
  stack4 = "tlsOrSsl";
  stack3['name'] = stack4;
  stack4 = "view.configsAreDisabled";
  stack3['disabledBinding'] = stack4;
  stack4 = "view.tlsOrSsl";
  stack3['selectionBinding'] = stack4;
  stack4 = "tls";
  stack3['value'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "installer.step7.misc.notification.use_tls";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n\n      <div class=\"entry-row indent-1\">\n        ");
  stack1 = depth0;
  stack2 = "Ember.RadioButton";
  stack3 = {};
  stack4 = "tlsOrSsl";
  stack3['name'] = stack4;
  stack4 = "view.configsAreDisabled";
  stack3['disabledBinding'] = stack4;
  stack4 = "view.tlsOrSsl";
  stack3['selectionBinding'] = stack4;
  stack4 = "ssl";
  stack3['value'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "installer.step7.misc.notification.use_ssl";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n      \n      ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </form>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/overriddenProperty", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  ");
  data.buffer.push("\n  <div ");
  stack1 = {};
  stack2 = "overriddenSCP.errorMessage:error: :control-group :overrideField";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    ");
  stack1 = depth0;
  stack2 = "overriddenSCP.viewClass";
  stack3 = {};
  stack4 = "overriddenSCP";
  stack3['serviceConfigBinding'] = stack4;
  stack4 = "view.categoryConfigs";
  stack3['categoryConfigsBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "overriddenSCP.supportsFinal";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "view.isDefaultGroupSelected";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(7, program7, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n      <a href=\"#\" data-toggle=\"tooltip\"\n        ");
  stack1 = {};
  stack2 = ":btn :btn-small :btn-final overriddenSCP.isFinal:active overriddenSCP.hideFinalIcon:hidden";
  stack1['class'] = stack2;
  stack2 = "overriddenSCP.isNotEditable";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "overriddenSCP";
  stack3 = depth0;
  stack4 = "toggleFinalFlag";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = {};
  stack2 = "services.service.config.final";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n        <i class=\"icon-lock\"></i>\n      </a>\n    ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "overriddenSCP.group";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n        <a href=\"#\" data-toggle=\"tooltip\" ");
  stack1 = {};
  stack2 = "overriddenSCP.group.switchGroupTextFull";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"action\" ");
  stack1 = depth0;
  stack2 = "overriddenSCP.group";
  stack3 = depth0;
  stack4 = "selectConfigGroup";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "overriddenSCP.group.switchGroupTextShort";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </a>\n      ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "overriddenSCP.isEditable";
  stack3 = helpers['if'];
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(13, program13, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "isNotDefaultValue";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "overriddenSCP.errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n      ");
  return buffer;}
function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n          <a class=\"btn btn-small\" href=\"#\" data-toggle=\"tooltip\"\n            ");
  stack1 = depth0;
  stack2 = "overriddenSCP";
  stack3 = depth0;
  stack4 = "doRestoreDefaultValue";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  stack1 = {};
  stack2 = "common.undo";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n            <i class=\"icon-undo\"></i>\n          </a>\n        ");
  return buffer;}

function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n          <a class=\"btn btn-small\" href=\"#\" data-toggle=\"tooltip\"\n            ");
  stack1 = depth0;
  stack2 = "overriddenSCP";
  stack3 = depth0;
  stack4 = "removeOverride";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  stack1 = {};
  stack2 = "common.remove";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n            <i class=\"icon-minus-sign\"></i>\n          </a>\n        ");
  return buffer;}

function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <a class=\"action\">");
  stack1 = depth0;
  stack2 = "overriddenSCP.group.switchGroupTextShort";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a> <i class=\"icon-spinner\"></i>\n      ");
  return buffer;}

  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.serviceConfigProperty.overrides";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "overriddenSCP";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/overrideWindow", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n	        <div class=\"text-warning\">\n	          ");
  stack1 = depth0;
  stack2 = "view.parentView.warningMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n	        </div>\n        ");
  return buffer;}

function program3(depth0,data) {


  data.buffer.push("\n                <i class='icon-ok-sign'></i>\n              ");}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n			                    <li>\n			                      <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "column";
  stack3 = depth0;
  stack4 = "selectFilterColumn";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n			                        ");
  stack1 = depth0;
  stack2 = "column.selected";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n			                        ");
  stack1 = depth0;
  stack2 = "column.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n			                      </a>\n			                    </li>\n			                  ");
  return buffer;}
function program6(depth0,data) {


  data.buffer.push("\n	                              <i class='icon-ok-sign'></i>\n	                            ");}

function program8(depth0,data) {


  data.buffer.push("\n                                <i class='icon-placeholder'></i>\n	                            ");}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                        <li>\n                          <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "component";
  stack3 = depth0;
  stack4 = "selectFilterComponent";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                            ");
  stack1 = depth0;
  stack2 = "component.selected";
  stack3 = helpers['if'];
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(13, program13, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n	                          ");
  stack1 = depth0;
  stack2 = "component.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n	                        </a>\n                        </li>\n                      ");
  return buffer;}
function program11(depth0,data) {


  data.buffer.push("\n                              <i class='icon-ok-sign'></i>\n                            ");}

function program13(depth0,data) {


  data.buffer.push("\n                              <i class='icon-placeholder'></i>\n	                          ");}

function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            <tr ");
  stack1 = {};
  stack2 = "entry.filtered::hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <td width=\"10%\">\n                ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "entry.selected";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </td>\n              <td width=\"45%\">\n                ");
  stack1 = depth0;
  stack2 = "entry.host.publicHostName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </td>\n              <td>\n                ");
  stack1 = depth0;
  stack2 = "entry.filterColumnValue";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </td>\n            </tr>\n          ");
  return buffer;}

function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div class=\"page-bar pull-right no-borders\">\n      <div class=\"items-on-page\">\n        <label>");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.rowsPerPageSelectView";
  stack3 = {};
  stack4 = "view.displayLength";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n      </div>\n      <div class=\"info\">");
  stack1 = depth0;
  stack2 = "view.paginationInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n      <div class=\"paging_two_button\">\n        ");
  stack1 = depth0;
  stack2 = "view.paginationFirst";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.paginationLeft";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.paginationRight";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.paginationLast";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n  ");
  return buffer;}

  data.buffer.push("\n\n");
  data.buffer.push("\n  <form class=\"form-horizontal mbm\" autocomplete=\"off\">\n      <div class=\"override-controls\">\n        <div class=\"alert alert-info\">");
  stack1 = depth0;
  stack2 = "view.parentView.dialogMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.warningMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        <table style=\"width: 100%;\">\n          <tr>\n            <td>\n              <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "toggleShowSelectedHosts";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.hostSelectMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n              ");
  stack1 = depth0;
  stack2 = "view.showOnlySelectedHosts";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </td>\n            <td width=\"30%\">\n              <div class=\"row\">\n                <div class=\"span2\" id=\"filter-dropdown-div\">\n                  <!-- Filter text box with drop-down -->\n		              <div class=\"input-append\">\n		                ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "view.filterText";
  stack3['valueBinding'] = stack4;
  stack4 = "appendedDropdownButton";
  stack3['id'] = stack4;
  stack4 = "view.filterTextPlaceholder";
  stack3['placeholderBinding'] = stack4;
  stack4 = "span2";
  stack3['class'] = stack4;
  stack4 = "view.isDisabled";
  stack3['disabledBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n		                <div class=\"btn-group\">\n		                  <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" ");
  stack1 = {};
  stack2 = "view.isDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n		                    <span class=\"caret\"></span>\n		                  </button>\n		                  <ul class=\"dropdown-menu\">\n			                  ");
  stack1 = depth0;
  stack2 = "view.filterColumns";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "column";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n		                  </ul>\n		                </div>\n		              </div>\n                </div>\n                <div class=\"span2\" id=\"component-dropdown-div\">\n                  <!-- Host-components drop-down -->\n		              <div class=\"btn-group\">\n		                <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\" ");
  stack1 = {};
  stack2 = "view.isDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n		                  ");
  stack1 = depth0;
  stack2 = "common.components";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n		                  <span class=\"caret\"></span>\n		                </button>\n		                <ul class=\"dropdown-menu\">\n		                  <!-- dropdown menu links -->\n		                  ");
  stack1 = depth0;
  stack2 = "view.filterComponents";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n		                </ul>\n                  </div>\n                </div>\n              </div>\n            </td>\n          </tr>\n        </table>\n        <table class=\"table table-striped hosts-table\">\n          <thead>\n            <tr class=\"success\">\n              <th width=\"10%\">\n                ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "view.allHostsSelected";
  stack3['checkedBinding'] = stack4;
  stack4 = "view.isDisabled";
  stack3['disabledBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </th>\n              <th width=\"45%\">");
  stack1 = depth0;
  stack2 = "common.host";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n              <th width=\"45%\">");
  stack1 = depth0;
  stack2 = "view.filterColumn.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n            </tr>\n          </thead>\n         </table>\n        <div class=\"hosts-table-container\">\n          <table class=\"table table-striped hosts-table\">\n          ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "entry";
  stack7 = helpers.each;
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n         </table>\n        </div>\n      </div>\n  </form>\n\n  ");
  stack1 = depth0;
  stack2 = "view.isPaginate";
  stack3 = helpers['if'];
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/propertyDependence", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <tr>\n      <td>");
  stack1 = depth0;
  stack2 = "property.serviceName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
  stack1 = depth0;
  stack2 = "property.propertyDisplayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
  stack1 = depth0;
  stack2 = "property.curValue";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
  stack1 = depth0;
  stack2 = "property.newValue";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n    </tr>\n  ");
  return buffer;}

  data.buffer.push("\n\n<table class=\"table table-bordered table-striped\">\n  <thead>\n  <tr>\n    <th>");
  stack1 = depth0;
  stack2 = "common.service";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n    <th>");
  stack1 = depth0;
  stack2 = "common.property";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n    <th>");
  stack1 = depth0;
  stack2 = "installer.step7.popup.currentValue";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n    <th>");
  stack1 = depth0;
  stack2 = "installer.step7.popup.adjustedValue";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n  </tr>\n  </thead>\n  <tbody>\n  ");
  stack1 = depth0;
  stack2 = "view.propertyChange";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "property";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </tbody>\n</table>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/propertyDependence_footer", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <button class=\"btn btn-warning\" ");
  stack1 = depth0;
  stack2 = "onIgnore";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.ignore";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n  ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"pull-right\">\n  <button class=\"btn\" ");
  stack1 = depth0;
  stack2 = "onUndo";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.cancel";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n  ");
  stack1 = depth0;
  stack2 = "view.canIgnore";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  <button class=\"btn btn-success\" ");
  stack1 = depth0;
  stack2 = "onApply";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.apply";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/queuePopup_body", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "isVisible";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <div class=\"rw-fluid control-group\">\n        <div ");
  stack1 = {};
  stack2 = "errorMessage:error :control-label-span :span4";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <label>");
  stack1 = depth0;
  stack2 = "displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n        </div>\n        <div ");
  stack1 = {};
  stack2 = "errorMessage:error :control-group :span8";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "viewClass";
  stack3 = {};
  stack4 = "this";
  stack3['serviceConfigBinding'] = stack4;
  stack4 = "view.categoryConfigs";
  stack3['categoryConfigsBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n        </div>\n      </div>\n    ");
  return buffer;}

  data.buffer.push("\n\n<form class=\"form-horizontal pre-scrollable\" autocomplete=\"off\">\n  ");
  stack1 = depth0;
  stack2 = "view.parentView.content.configs";
  stack3 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</form>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/queuePopup_footer", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div class=\"pull-left\">\n    <button class=\"btn btn-danger\" ");
  stack1 = depth0;
  stack2 = "onDelete";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      ");
  stack1 = depth0;
  stack2 = "view.parentView.delete";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </button>\n  </div>\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <button type=\"button\" class=\"btn\" ");
  stack1 = depth0;
  stack2 = "onPrimary";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      ");
  stack1 = depth0;
  stack2 = "view.parentView.primary";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </button>\n  ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <button type=\"button\" ");
  stack1 = {};
  stack2 = "view.parentView.isError";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-success\" ");
  stack1 = depth0;
  stack2 = "onSecondary";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      ");
  stack1 = depth0;
  stack2 = "view.parentView.secondary";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </button>\n  ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.parentView.delete";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n<p class=\"pull-right\">\n  ");
  stack1 = depth0;
  stack2 = "view.parentView.primary";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.parentView.secondary";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</p>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/saveConfigGroup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<div id=\"config-group-confirm-save\">\n  <div class=\"alert alert-success\">");
  stack1 = depth0;
  stack2 = "config.group.save.confirmation.configGroup";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " <em>");
  stack1 = depth0;
  stack2 = "groupName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</em> ");
  stack1 = depth0;
  stack2 = "config.group.save.confirmation.saved";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n  <div class=\"message\">");
  stack1 = depth0;
  stack2 = "config.group.save.confirmation.msg";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/save_configuration", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"row-fluid\">\n        <div class=\"span12 alert alert-warning\">");
  stack1 = depth0;
  stack2 = "services.service.config.exitPopup.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n");
  return buffer;}

  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.showSaveWarning";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n<div class=\"row-fluid\">\n    <div class=\"span1\">");
  stack1 = depth0;
  stack2 = "common.notes";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n    <div class=\"span10\">");
  stack1 = depth0;
  stack2 = "view.notesArea";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/selectCreateConfigGroup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    <div class=\"select-create-config-group-div\">\n      <span class=\"btn-group\">\n        <button class=\"btn\" ");
  stack1 = {};
  stack2 = "view.parentView.optionCreateConfigGroup";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.parentView.selectedConfigGroup.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</button>\n        <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" ");
  stack1 = {};
  stack2 = "view.parentView.optionCreateConfigGroup";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <span class=\"caret\"></span>\n        </button>\n        <ul class=\"dropdown-menu\">\n          <!-- available config group menu links -->\n          ");
  stack1 = depth0;
  stack2 = "view.parentView.availableConfigGroups";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "configGroup";
  stack7 = helpers.each;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </ul>\n      </span>\n      <div class=\"message\">");
  stack1 = depth0;
  stack2 = "config.group.selection.dialog.option.select.msg";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n             <li>\n               <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "configGroup";
  stack3 = depth0;
  stack4 = "doSelectConfigGroup";
  stack5 = {};
  stack6 = "view.parentView";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                 ");
  stack1 = depth0;
  stack2 = "configGroup.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n               </a>\n             </li>\n          ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"select-create-config-group-div\">\n      <div class=\"alert alert-block\">");
  stack1 = depth0;
  stack2 = "view.parentView.noGroups";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n  ");
  return buffer;}

  data.buffer.push("\n<div id=\"config-group-select-create-dialog\">\n  <div class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "view.parentView.subTitle";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <div class=\"text-warning\">\n    ");
  stack1 = depth0;
  stack2 = "view.parentView.warningMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <label class=\"radio\">\n    ");
  stack1 = depth0;
  stack2 = "view.selectConfigGroupRadioButton";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    <span ");
  stack1 = {};
  stack2 = "view.parentView.hasExistedGroups::message";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.parentView.selectExistingGroupLabel";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n  </label>\n  ");
  stack1 = depth0;
  stack2 = "view.parentView.availableConfigGroups.length";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  <label class=\"radio\">\n    ");
  stack1 = depth0;
  stack2 = "view.createConfigGroupRadioButton";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "view.parentView.createNewGroupLabel";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </label>\n  <div class=\"select-create-config-group-div\">\n    ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "view.parentView.newConfigGroupName";
  stack3['valueBinding'] = stack4;
  stack4 = "view.parentView.optionSelectConfigGroup";
  stack3['disabledBinding'] = stack4;
  stack4 = "255";
  stack3['maxlength'] = stack4;
  stack4 = "span5";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    <div class=\"message\">");
  stack1 = depth0;
  stack2 = "view.parentView.createNewGroupDescription";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/service_config", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "selectedService.restartRequired";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    <div class=\"alert\">\n      <div class=\"clearfix like_pointer collapsable\" ");
  stack1 = depth0;
  stack2 = "toggleRestartMessageView";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "view.isRestartMessageCollapsed";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(5, program5, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        <i class=\"icon-refresh\"></i>\n        <strong>");
  stack1 = depth0;
  stack2 = "dashboard.services.configs.popup.restartService.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</strong> ");
  stack1 = depth0;
  stack2 = "selectedService.restartRequiredMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n      <div class=\"service-body\">\n        <ul>\n          ");
  stack1 = depth0;
  stack2 = "selectedService.restartRequiredHostsAndComponents";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "host";
  stack7 = helpers.each;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </ul>\n      </div>\n    </div>\n  ");
  return buffer;}
function program3(depth0,data) {


  data.buffer.push("\n          <i class=\"icon-caret-down pull-left\"></i>\n        ");}

function program5(depth0,data) {


  data.buffer.push("\n          <i class=\"icon-caret-right pull-left\"></i>\n        ");}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n            <li>\n              <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "host.hostData";
  stack3 = depth0;
  stack4 = "showDetails";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "host.hostData.publicHostName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n              <ul>\n                ");
  stack1 = depth0;
  stack2 = "host.components";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </ul>\n            </li>\n          ");
  return buffer;}
function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  <li>");
  stack1 = depth0;
  stack2 = "component.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</li>\n                ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n  <div class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "common.group";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "&nbsp;\n	  <span class=\"btn-group\">\n		  <button ");
  stack1 = {};
  stack2 = "controller.isHostsConfigsPage";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn dropdown-toggle\"\n                                                                    data-toggle=\"dropdown\">");
  stack1 = depth0;
  stack2 = "selectedConfigGroup.displayNameHosts";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</button>\n		  <button ");
  stack1 = {};
  stack2 = "controller.isHostsConfigsPage";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n        <span class=\"caret\"></span>\n      </button>\n		  <ul class=\"dropdown-menu\">\n        <!-- available config group menu links -->\n        ");
  stack1 = depth0;
  stack2 = "configGroups";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "configGroup";
  stack7 = helpers.each;
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </ul>\n		</span>\n    ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <div class=\"pull-right\">\n      ");
  stack1 = depth0;
  stack2 = "App.FilterComboCleanableView";
  stack3 = {};
  stack4 = "view.filter";
  stack3['filterBinding'] = stack4;
  stack4 = "view.columns";
  stack3['columnsBinding'] = stack4;
  stack4 = "view.propertyFilterPopover";
  stack3['popoverDescriptionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n");
  return buffer;}
function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n          <li>\n            <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "configGroup";
  stack3 = depth0;
  stack4 = "selectConfigGroup";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "configGroup.displayNameHosts";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n        ");
  return buffer;}

function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "controller.isHostsConfigsPage";
  stack3 = helpers['if'];
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(16, program16, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        &nbsp;<a href=\"#\" ");
  stack1 = depth0;
  stack2 = "switchHostGroup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.change";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n      ");
  return buffer;}

function program16(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <a href=\"#\" class=\"link-left-pad\" ");
  stack1 = depth0;
  stack2 = "manageConfigurationGroup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "services.service.actions.manage_configuration_groups.short";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n      ");
  return buffer;}

function program18(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "App.ConfigHistoryFlowView";
  stack3 = {};
  stack4 = "selectedService";
  stack3['serviceBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;}

function program20(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "selectedService.configCategories";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "category";
  stack7 = helpers.each;
  tmp1 = self.program(21, program21, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program21(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"accordion\">\n      ");
  stack1 = depth0;
  stack2 = "category.isCustomView";
  stack3 = helpers['if'];
  tmp1 = self.program(22, program22, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(24, program24, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n  ");
  return buffer;}
function program22(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "category.customView";
  stack3 = {};
  stack4 = "category";
  stack3['categoryBinding'] = stack4;
  stack4 = "selectedService";
  stack3['serviceBinding'] = stack4;
  stack4 = "view.canEdit";
  stack3['canEditBinding'] = stack4;
  stack4 = "selectedService.configs";
  stack3['serviceConfigsBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  return buffer;}

function program24(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "App.ServiceConfigsByCategoryView";
  stack3 = {};
  stack4 = "category";
  stack3['categoryBinding'] = stack4;
  stack4 = "view.canEdit";
  stack3['canEditBinding'] = stack4;
  stack4 = "selectedService";
  stack3['serviceBinding'] = stack4;
  stack4 = "selectedService.configs";
  stack3['serviceConfigsBinding'] = stack4;
  stack4 = "view.supportsHostOverrides";
  stack3['supportsHostOverridesBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  return buffer;}

function program26(depth0,data) {


  data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.supportsHostOverrides";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.supportsHostOverrides";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.showConfigHistoryFeature";
  stack3 = helpers['if'];
  tmp1 = self.program(18, program18, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "versionLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(20, program20, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(26, program26, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/service_config_category", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <div ");
  stack1 = {};
  stack2 = ":entry-row isHiddenByFilter:hide isOverridden:overridden-property hasCompareDiffs:overridden-property";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "showLabel";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          <div ");
  stack1 = {};
  stack2 = "showLabel:controls";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            ");
  data.buffer.push("\n            <div ");
  stack1 = {};
  stack2 = "errorMessage:error: warnMessage:warning: :control-group";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "viewClass";
  stack3 = {};
  stack4 = "this";
  stack3['serviceConfigBinding'] = stack4;
  stack4 = "view.categoryConfigsAll";
  stack3['categoryConfigsAllBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "isComparison";
  stack3 = helpers['if'];
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "supportsFinal";
  stack3 = helpers['if'];
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "view.canEdit";
  stack3 = helpers['if'];
  tmp1 = self.program(16, program16, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n              <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "warnMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n            </div>\n            ");
  stack1 = depth0;
  stack2 = "isOverridden";
  stack3 = helpers['if'];
  tmp1 = self.program(32, program32, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "isComparison";
  stack3 = helpers['if'];
  tmp1 = self.program(34, program34, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n        </div>\n        ");
  stack1 = depth0;
  stack2 = "additionalView";
  stack3 = helpers['if'];
  tmp1 = self.program(36, program36, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <span ");
  stack1 = {};
  stack2 = "errorMessage:error: :control-group :control-label-span";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <label class=\"control-label\">\n                ");
  stack1 = depth0;
  stack2 = "displayName";
  foundHelper = helpers.formatWordBreak;
  stack3 = foundHelper || depth0.formatWordBreak;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "formatWordBreak", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                ");
  stack1 = depth0;
  stack2 = "isSecureConfig";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "view.supportsHostOverrides";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </label>\n            </span>\n          ");
  return buffer;}
function program3(depth0,data) {


  data.buffer.push("\n                  <a href=\"javascript:void(null);\"><i class=\"icon-lock\" rel=\"tooltip\" data-toggle=\"tooltip\"\n                                                      title=\"security knob\"></i></a>\n                ");}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "isRestartRequired";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2;
  data.buffer.push("\n                    <i class=\"icon-refresh restart-required-property\"\n                       rel=\"tooltip\" ");
  stack1 = {};
  stack2 = "restartRequiredMessage";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n                  ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "controller.selectedConfigGroup.isDefault";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(12, program12, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  return buffer;}
function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  <span class=\"label label-info\">");
  stack1 = depth0;
  stack2 = "dashboard.configHistory.table.version.prefix";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "controller.selectedVersion";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n                  ");
  stack1 = depth0;
  stack2 = "controller.isCurrentSelected";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    <span class=\"label label-success\">");
  stack1 = depth0;
  stack2 = "common.current";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n                  ");
  return buffer;}

function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  <span class=\"label label-info\">");
  stack1 = depth0;
  stack2 = "common.default";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "&nbsp;");
  stack1 = depth0;
  stack2 = "services.service.config.configHistory.configGroup";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n                ");
  return buffer;}

function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                <a href=\"#\" data-toggle=\"tooltip\"\n                  ");
  stack1 = {};
  stack2 = ":btn :btn-small :btn-final isFinal:active hideFinalIcon:hidden";
  stack1['class'] = stack2;
  stack2 = "isNotEditable";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  ");
  stack1 = depth0;
  stack2 = "";
  stack3 = depth0;
  stack4 = "toggleFinalFlag";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  ");
  stack1 = {};
  stack2 = "services.service.config.final";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n                  <i class=\"icon-lock\"></i>\n                </a>\n              ");
  return buffer;}

function program16(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "isPropertyOverridable";
  stack3 = helpers['if'];
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "cantBeUndone";
  stack3 = helpers.unless;
  tmp1 = self.program(22, program22, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "isRemovable";
  stack3 = helpers['if'];
  tmp1 = self.program(25, program25, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "isEditable";
  stack3 = helpers.unless;
  tmp1 = self.program(29, program29, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  return buffer;}
function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "view.supportsHostOverrides";
  stack3 = helpers['if'];
  tmp1 = self.program(18, program18, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program18(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(19, program19, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program19(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                      ");
  stack1 = depth0;
  stack2 = "isComparison";
  stack3 = helpers.unless;
  tmp1 = self.program(20, program20, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    ");
  return buffer;}
function program20(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                          <a class=\"btn btn-small\" href=\"#\" data-toggle=\"tooltip\"\n                            ");
  stack1 = depth0;
  stack2 = "";
  stack3 = depth0;
  stack4 = "createOverrideProperty";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                            ");
  stack1 = {};
  stack2 = "common.override";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n                              <i class=\"icon-plus-sign\"></i>\n                          </a>\n                      ");
  return buffer;}

function program22(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "isNotDefaultValue";
  stack3 = helpers['if'];
  tmp1 = self.program(23, program23, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program23(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                    <a class=\"btn btn-small\" href=\"#\" data-toggle=\"tooltip\"\n                      ");
  stack1 = depth0;
  stack2 = "";
  stack3 = depth0;
  stack4 = "doRestoreDefaultValue";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                      ");
  stack1 = {};
  stack2 = "common.undo";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n                      <i class=\"icon-undo\"></i>\n                    </a>\n                  ");
  return buffer;}

function program25(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(26, program26, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program26(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "isComparison";
  stack3 = helpers.unless;
  tmp1 = self.program(27, program27, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program27(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                        <a class=\"btn btn-small\" href=\"#\" data-toggle=\"tooltip\"\n                          ");
  stack1 = depth0;
  stack2 = "";
  stack3 = depth0;
  stack4 = "removeProperty";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                          ");
  stack1 = {};
  stack2 = "common.remove";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n                            <i class=\"icon-minus-sign\"></i>\n                        </a>\n                    ");
  return buffer;}

function program29(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "group";
  stack3 = helpers['if'];
  tmp1 = self.program(30, program30, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program30(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                    <a rel='SwitchGroupTooltip' ");
  stack1 = {};
  stack2 = "this.group.switchGroupTextShort";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"action\" ");
  stack1 = depth0;
  stack2 = "group";
  stack3 = depth0;
  stack4 = "selectConfigGroup";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                      ");
  stack1 = depth0;
  stack2 = "group.switchGroupTextShort";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                    </a>\n                  ");
  return buffer;}

function program32(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "App.ServiceConfigView.SCPOverriddenRowsView";
  stack3 = {};
  stack4 = "this";
  stack3['serviceConfigPropertyBinding'] = stack4;
  stack4 = "controller.selectedConfigGroup.isDefault";
  stack3['isDefaultGroupSelectedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program34(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "App.ServiceConfigView.SCPComparisonRowsView";
  stack3 = {};
  stack4 = "this";
  stack3['serviceConfigPropertyBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program36(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "additionalView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  return buffer;}

function program38(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.canEdit";
  stack3 = helpers['if'];
  tmp1 = self.program(39, program39, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program39(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.category.customCanAddProperty";
  stack3 = helpers['if'];
  tmp1 = self.program(40, program40, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program40(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <div>\n              <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "";
  stack3 = depth0;
  stack4 = "showAddPropertyWindow";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " >");
  stack1 = depth0;
  stack2 = "installer.step7.config.addProperty";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "...</a>\n            </div>\n          ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"accordion-heading\" ");
  stack1 = depth0;
  stack2 = "category";
  stack3 = depth0;
  stack4 = "onToggleBlock";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  <i ");
  stack1 = {};
  stack2 = ":pull-left :accordion-toggle view.category.isCollapsed:icon-caret-right:icon-caret-down";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n  <a class=\"accordion-toggle\">");
  stack1 = depth0;
  stack2 = "view.category.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n</div>\n\n<div class=\"accordion-body collapse in\" ");
  stack1 = {};
  stack2 = "view.isCategoryBodyVisible";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  <div class=\"accordion-inner service-config-section\">\n    <form class=\"form-horizontal\" autocomplete=\"off\">\n\n      ");
  stack1 = depth0;
  stack2 = "view.categoryConfigs";
  stack3 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n      ");
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(38, program38, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </form>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/service_config_wizard", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n  <div class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "common.group";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "&nbsp;\n	  <span class=\"btn-group\">\n		  <button ");
  stack1 = {};
  stack2 = "controller.isHostsConfigsPage";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn\">");
  stack1 = depth0;
  stack2 = "selectedConfigGroup.displayNameHosts";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</button>\n		  <button ");
  stack1 = {};
  stack2 = "controller.isHostsConfigsPage";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n        <span class=\"caret\"></span>\n      </button>\n		  <ul class=\"dropdown-menu\">\n        <!-- available config group menu links -->\n        ");
  stack1 = depth0;
  stack2 = "configGroups";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "configGroup";
  stack7 = helpers.each;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </ul>\n		</span>\n    ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <div class=\"pull-right\">\n      ");
  stack1 = depth0;
  stack2 = "App.FilterComboCleanableView";
  stack3 = {};
  stack4 = "view.filter";
  stack3['filterBinding'] = stack4;
  stack4 = "view.columns";
  stack3['columnsBinding'] = stack4;
  stack4 = "view.propertyFilterPopover";
  stack3['popoverDescriptionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n          <li>\n            <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "configGroup";
  stack3 = depth0;
  stack4 = "selectConfigGroup";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "configGroup.displayNameHosts";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n        ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "controller.isHostsConfigsPage";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(7, program7, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        &nbsp;<a href=\"#\" ");
  stack1 = depth0;
  stack2 = "switchHostGroup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.change";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n      ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <a href=\"#\"  class=\"link-left-pad\" ");
  stack1 = depth0;
  stack2 = "manageConfigurationGroup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "services.service.actions.manage_configuration_groups.short";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n      ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "catView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.supportsHostOverrides";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n<div class=\"accordion\">\n  ");
  stack1 = depth0;
  stack2 = "view.serviceConfigsByCategoryView.childViews";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "catView";
  stack7 = helpers.each;
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/service_version_box", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, stack8, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <div class=\"current-label\">\n          <span class=\"label label-success\">\n            ");
  stack1 = depth0;
  stack2 = "common.current";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            <i ");
  stack1 = {};
  stack2 = ":icon-refresh :restart-required-service serviceVersion.isRestartRequired::hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n          </span>\n        </div>\n      ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, stack8;
  data.buffer.push("\n          <button ");
  stack1 = {};
  stack2 = "serviceVersion.disabledActionAttr.compare";
  stack1['disabled'] = stack2;
  stack2 = ":btn serviceVersion.isDisplayed:not-allowed-cursor";
  stack1['class'] = stack2;
  stack2 = "serviceVersion.disabledActionMessages.compare";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.actionTypes.COMPARE";
  stack3 = depth0;
  stack4 = "serviceVersion";
  stack5 = depth0;
  stack6 = "doAction";
  stack7 = {};
  stack8 = "view.parentView";
  stack7['target'] = stack8;
  stack8 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack7;
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack8.call(depth0, stack6, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-copy\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "common.compare";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n          <button ");
  stack1 = {};
  stack2 = "serviceVersion.disabledActionAttr.revert";
  stack1['disabled'] = stack2;
  stack2 = ":btn serviceVersion.isCurrent:not-allowed-cursor";
  stack1['class'] = stack2;
  stack2 = "serviceVersion.disabledActionMessages.revert";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.actionTypes.REVERT";
  stack3 = depth0;
  stack4 = "serviceVersion";
  stack5 = depth0;
  stack6 = "doAction";
  stack7 = {};
  stack8 = "view.parentView";
  stack7['target'] = stack8;
  stack8 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack7;
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack8.call(depth0, stack6, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "dashboard.configHistory.info-bar.revert.button";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n        ");
  return buffer;}

  data.buffer.push("\n\n<div ");
  stack1 = {};
  stack2 = ":flow-element :pull-left serviceVersion.first:first";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  <div class=\"arrow-box pull-left\"></div>\n  <div class=\"version-box\">\n    <div ");
  stack1 = {};
  stack2 = ":version-info :box :pull-right serviceVersion.isDisplayed:displayed serviceVersion.isDisabled:grayedOut";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.actionTypes.SWITCH";
  stack3 = depth0;
  stack4 = "serviceVersion";
  stack5 = depth0;
  stack6 = "doAction";
  stack7 = {};
  stack8 = "view.parentView";
  stack7['target'] = stack8;
  stack8 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack7;
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack8.call(depth0, stack6, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <div class=\"top-label\">\n        <span class=\"label label-info\">");
  stack1 = depth0;
  stack2 = "serviceVersion.versionText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n        <span class=\"author pull-right\">");
  stack1 = depth0;
  stack2 = "serviceVersion.author";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n      </div>\n      <div class=\"content\">");
  stack1 = depth0;
  stack2 = "serviceVersion.timeSinceCreated";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n      ");
  stack1 = depth0;
  stack2 = "serviceVersion.isCurrent";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n\n    <div class=\"version-popover\">\n      <div class=\"content\">\n        <span class=\"label label-info\">");
  stack1 = depth0;
  stack2 = "serviceVersion.versionText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span> <span class=\"pull-right\"><strong>");
  stack1 = depth0;
  stack2 = "serviceVersion.configGroupName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong></span>\n        <div class=\"date\"><strong>");
  stack1 = depth0;
  stack2 = "serviceVersion.author";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong>&nbsp;");
  stack1 = depth0;
  stack2 = "dashboard.configHistory.info-bar.authoredOn";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "&nbsp;<strong>");
  stack1 = depth0;
  stack2 = "serviceVersion.createdDate";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong></div>\n        <div class=\"notes\">");
  stack1 = depth0;
  stack2 = "serviceVersion.fullNotes";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</div>\n      </div>\n      <div class=\"version-operations-buttons\">\n        <button ");
  stack1 = {};
  stack2 = "serviceVersion.disabledActionAttr.view";
  stack1['disabled'] = stack2;
  stack2 = ":btn serviceVersion.isDisplayed:not-allowed-cursor";
  stack1['class'] = stack2;
  stack2 = "serviceVersion.disabledActionMessages.view";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.actionTypes.SWITCH";
  stack3 = depth0;
  stack4 = "serviceVersion";
  stack5 = depth0;
  stack6 = "doAction";
  stack7 = {};
  stack8 = "view.parentView";
  stack7['target'] = stack8;
  stack8 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack7;
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack8.call(depth0, stack6, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-search\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "common.view";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n        ");
  stack1 = depth0;
  stack2 = "MANAGER";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/configs/services_config", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    <ul class=\"nav nav-tabs\">\n      ");
  stack1 = depth0;
  stack2 = "controller.stepConfigs";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "service";
  stack7 = helpers.each;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </ul>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "service.showConfig";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "App.ServiceConfigTab";
  stack3 = helpers.view;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n              <a href=\"#");
  stack1 = depth0;
  stack2 = "service.serviceName";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\" ");
  stack1 = {};
  stack2 = ":active service.selected:new";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                 data-toggle=\"tab\" ");
  stack1 = depth0;
  stack2 = "service";
  stack3 = depth0;
  stack4 = "selectService";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                ");
  stack1 = depth0;
  stack2 = "service.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "service.errorCount";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</a>\n          ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("<span\n                      class=\"badge badge-important\">");
  stack1 = depth0;
  stack2 = "service.errorCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>");
  return buffer;}

function program7(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n");}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <p class=\"loading align-center\"></p>\n    ");
  stack1 = depth0;
  stack2 = "isSubmitDisabled";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "submitButtonClicked";
  stack3 = helpers.unless;
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <div class=\"alert\">");
  stack1 = depth0;
  stack2 = "installer.step7.attentionNeeded";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n      ");
  return buffer;}

  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "controller.isRecommendedLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(7, program7, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "App.ServiceConfigContainerView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  stack1 = depth0;
  stack2 = "controller.isRecommendedLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/custom_date_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


  data.buffer.push("\n\n<div class=\"jobs-custom-dates\">\n  <div ");
  stack1 = {};
  stack2 = ":control-group view.isValid.isStartDateError:error";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <label>");
  stack1 = depth0;
  stack2 = "jobs.customDateFilter.startTime";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n    ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "controller.customDateFormFields.startDate";
  stack3['valueBinding'] = stack4;
  stack4 = "input-small datepicker";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "view.hourOptions";
  stack3['contentBinding'] = stack4;
  stack4 = "controller.customDateFormFields.hoursForStart";
  stack3['selectionBinding'] = stack4;
  stack4 = "input-mini";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "view.minuteOptions";
  stack3['contentBinding'] = stack4;
  stack4 = "controller.customDateFormFields.minutesForStart";
  stack3['selectionBinding'] = stack4;
  stack4 = "input-mini";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "view.middayPeriodOptions";
  stack3['contentBinding'] = stack4;
  stack4 = "controller.customDateFormFields.middayPeriodForStart";
  stack3['selectionBinding'] = stack4;
  stack4 = "input-mini";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "view.validationErrors.startDate";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n  </div>\n  <div>\n\n  </div>\n  <div ");
  stack1 = {};
  stack2 = ":control-group view.isValid.isEndDateError:error";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <label>");
  stack1 = depth0;
  stack2 = "jobs.customDateFilter.endTime";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n    ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "controller.customDateFormFields.endDate";
  stack3['valueBinding'] = stack4;
  stack4 = "input-small datepicker";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "view.hourOptions";
  stack3['contentBinding'] = stack4;
  stack4 = "controller.customDateFormFields.hoursForEnd";
  stack3['selectionBinding'] = stack4;
  stack4 = "input-mini";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "view.minuteOptions";
  stack3['contentBinding'] = stack4;
  stack4 = "controller.customDateFormFields.minutesForEnd";
  stack3['selectionBinding'] = stack4;
  stack4 = "input-mini";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "view.middayPeriodOptions";
  stack3['contentBinding'] = stack4;
  stack4 = "controller.customDateFormFields.middayPeriodForEnd";
  stack3['selectionBinding'] = stack4;
  stack4 = "input-mini";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "view.validationErrors.endDate";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/editable_list", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <li ");
  stack1 = {};
  stack2 = ":item item.global:global:deletable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <span>\n            <a href>");
  stack1 = depth0;
  stack2 = "item.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n            ");
  stack1 = depth0;
  stack2 = "item.global";
  stack3 = helpers.unless;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </span>\n        </li>\n      ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n              <button ");
  stack1 = depth0;
  stack2 = "item";
  stack3 = depth0;
  stack4 = "removeFromItems";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " type=\"button\" class=\"close\">\n                <span>&times;</span>\n              </button>\n            ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <li class=\"item add-item-input\">\n          ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "view.input";
  stack3['valueBinding'] = stack4;
  stack4 = "New";
  stack3['placeholder'] = stack4;
  stack4 = "input-area";
  stack3['classBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "view.typeahead";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </li>\n      ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n            <div class=\"typeahead-box\">\n              <ul>\n                ");
  stack1 = depth0;
  stack2 = "view.typeahead";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "item";
  stack7 = helpers.each;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </ul>\n            </div>\n          ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                  <li ");
  stack1 = depth0;
  stack2 = "item";
  stack3 = depth0;
  stack4 = "addItem";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"\">");
  stack1 = depth0;
  stack2 = "item.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</li>\n                ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.items";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(11, program11, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program9(depth0,data) {


  data.buffer.push("\n        ");}

function program11(depth0,data) {


  data.buffer.push("\n           <li class=\"item add-item\">Add</li>\n        ");}

function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div class=\"actions-panel\">\n      <button class=\"btn btn-default btn-xs\" ");
  stack1 = depth0;
  stack2 = "onSecondary";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <span class=\"icon-remove\"></span>\n      </button>\n      <button class=\"btn btn-primary btn-xs\" ");
  stack1 = depth0;
  stack2 = "onPrimary";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <span class=\"icon-ok\"></span>\n      </button>\n    </div>\n  ");
  return buffer;}

function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div class=\"pencil-box\" ");
  stack1 = depth0;
  stack2 = "enableEditMode";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <span class=\"icon-pencil\"></span>\n    </div>\n  ");
  return buffer;}

  data.buffer.push("\n\n<div ");
  stack1 = {};
  stack2 = ":editable-list-container :well view.editMode";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  <div class=\"items-box\" ");
  stack1 = depth0;
  stack2 = "enableEditMode";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <ul class=\"items-list\">\n      ");
  stack1 = depth0;
  stack2 = "view.items";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "item";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.editMode";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </ul>\n  </div>\n\n  ");
  stack1 = depth0;
  stack2 = "view.editMode";
  stack3 = helpers['if'];
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(15, program15, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/error_log_body", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <h5>structured_out: &nbsp;</h5>\n        <pre class=\"stdout\">");
  stack1 = depth0;
  stack2 = "view.openedTask.structuredOut";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</pre>\n      ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"task-detail-log-info\">\n  <div class=\"content-area\">\n    <div class=\"task-detail-log-clipboard-wrap\"></div>\n    <div class=\"task-detail-log-maintext\">\n      <h5>");
  stack1 = depth0;
  stack2 = "common.stderr";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": &nbsp; <span class=\"muted\">");
  stack1 = depth0;
  stack2 = "view.openedTask.errorLog";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " </span></h5>\n      <pre class=\"stderr\">");
  stack1 = depth0;
  stack2 = "view.openedTask.stderr";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</pre>\n      <h5>");
  stack1 = depth0;
  stack2 = "common.stdout";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": &nbsp; <span class=\"muted\"> ");
  stack1 = depth0;
  stack2 = "view.openedTask.outputLog";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " </span></h5>\n      <pre class=\"stdout\">");
  stack1 = depth0;
  stack2 = "view.openedTask.stdout";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</pre>\n      ");
  stack1 = depth0;
  stack2 = "view.openedTask.structuredOut";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/filter_combo_cleanable", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <li>\n                <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "column";
  stack3 = depth0;
  stack4 = "selectFilterColumn";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <i ");
  stack1 = {};
  stack2 = "column.selected:icon-ok-sign:icon-placeholder :text-success";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n                  ");
  stack1 = depth0;
  stack2 = "column.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                </a>\n            </li>\n        ");
  return buffer;}

  data.buffer.push("\n\n  ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "view.filter";
  stack3['valueBinding'] = stack4;
  stack4 = "view.placeHolder";
  stack3['placeholderBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  <div class=\"btn-group\">\n    <div id=\"cleanerContainer\" class=\"pull-left\">\n      <a href=\"#\" id=\"filterCleaner\" ");
  stack1 = depth0;
  stack2 = "clearFilter";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = {};
  stack2 = "view.filterNotEmpty::hidden :ui-icon :ui-icon-circle-close";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></a>\n    </div>\n      <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n          <span class=\"caret\"></span>\n      </button>\n      <ul class=\"dropdown-menu\">\n        ");
  stack1 = depth0;
  stack2 = "view.columns";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "column";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </ul>\n  </div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/filter_combobox", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <li>\n                <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "column";
  stack3 = depth0;
  stack4 = "selectFilterColumn";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <i ");
  stack1 = {};
  stack2 = "column.selected:icon-ok-sign:icon-placeholder :text-success";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n                  ");
  stack1 = depth0;
  stack2 = "column.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                </a>\n            </li>\n        ");
  return buffer;}

  data.buffer.push("\n\n  ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "view.filter";
  stack3['valueBinding'] = stack4;
  stack4 = "view.placeHolder";
  stack3['placeholderBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  <div class=\"btn-group\">\n      <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n          <span class=\"caret\"></span>\n      </button>\n      <ul class=\"dropdown-menu\">\n        ");
  stack1 = depth0;
  stack2 = "view.columns";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "column";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </ul>\n  </div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/form", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "App.FormFieldTemplate";
  stack3 = {};
  stack4 = "field";
  stack3['fieldBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.fields";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "field";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/form/check_db_connection", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div class=\"entry-row db-connection\">\n  <span class=\"control-label\"></span>\n  <div class=\"controls\">\n    <div class=\"control-group\">\n      <div class=\"span9\">\n        <span ");
  stack1 = {};
  stack2 = ":pull-left :btn :btn-primary view.isBtnDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "connectToDatabase";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.btnCaption";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n        <div class=\"pull-left connection-result mll\">\n          <a ");
  stack1 = {};
  stack2 = "view.isConnectionSuccess:mute:action";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "showLogsPopup";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.responseCaption";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n        </div>\n        <div ");
  stack1 = {};
  stack2 = ":spinner :mll :pull-left view.isConnecting::hide";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n        <i ");
  stack1 = {};
  stack2 = ":pull-right view.isConnectionSuccess:icon-ok-sign:icon-warning-sign view.isRequestResolved::hide";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n      </div>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/form/checkbox", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div class=\"controls\">\n  <label class=\"checkbox\" for=\"input");
  stack1 = depth0;
  stack2 = "view.field.name";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\">\n    ");
  stack1 = depth0;
  stack2 = "view.field.viewClass";
  stack3 = {};
  stack4 = "view.field.value";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.field.displayName";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </label>\n  <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "view.field.errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/form/field", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<label class=\"control-label\"\n       for=\"input");
  stack1 = depth0;
  stack2 = "view.field.name";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\">");
  stack1 = depth0;
  stack2 = "view.field.displayName";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n<div class=\"controls\">\n  ");
  stack1 = depth0;
  stack2 = "view.field.viewClass";
  stack3 = {};
  stack4 = "view.field.value";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "view.field.errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/grid/filter", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n<li>\n  <label class=\"checkbox\">\n    ");
  stack1 = depth0;
  stack2 = "Em.Checkbox";
  stack3 = {};
  stack4 = "filter.checked";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "filter.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </label>\n</li>\n");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.filters";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "filter";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/grid/header", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.filter";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  <a ");
  stack1 = depth0;
  stack2 = "applyFilter";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "apply";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "<i ");
  stack1 = depth0;
  stack2 = "toggleFilter";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"icon-filter\"></i>\n");
  stack1 = depth0;
  stack2 = "view.showFilter";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/grid/pager", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n    <li ");
  stack1 = {};
  stack2 = "page.activeClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ><a ");
  stack1 = depth0;
  stack2 = "page";
  stack3 = depth0;
  stack4 = "activatePage";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
  stack1 = depth0;
  stack2 = "page.number";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n  ");
  return buffer;}

  data.buffer.push("\n\n<ul>\n    <li ");
  stack1 = {};
  stack2 = "view.prevPageDisabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a ");
  stack1 = depth0;
  stack2 = "activatePrevPage";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
  stack1 = depth0;
  stack2 = "common.prev";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n  ");
  stack1 = depth0;
  stack2 = "view.pages";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "page";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <li ");
  stack1 = {};
  stack2 = "view.nextPageDisabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a ");
  stack1 = depth0;
  stack2 = "activateNextPage";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n</ul>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/host_progress_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n");
  data.buffer.push("\n\n  <div ");
  stack1 = {};
  stack2 = "view.parentView.isServiceListHidden:hidden :task-list-main-warp";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <div class=\"task-top-wrap\">\n      <div class=\"operation-name-top\">");
  stack1 = depth0;
  stack2 = "common.operations";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n      <div class=\"start-time-top\">");
  stack1 = depth0;
  stack2 = "common.startTime";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n      <div class=\"duration-top\">");
  stack1 = depth0;
  stack2 = "common.duration";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n      <div class=\"select-wrap\">\n        ");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":\n        ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "view.categories";
  stack3['contentBinding'] = stack4;
  stack4 = "content.value";
  stack3['optionValuePath'] = stack4;
  stack4 = "content.label";
  stack3['optionLabelPath'] = stack4;
  stack4 = "view.serviceCategory";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    <div id=\"service-info\">\n      ");
  stack1 = depth0;
  stack2 = "view.isServiceEmptyList";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.isShowMore";
  stack3 = helpers['if'];
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n  </div>\n\n\n  ");
  data.buffer.push("\n\n  <div ");
  stack1 = {};
  stack2 = "view.parentView.isHostListHidden:hidden :task-list-main-warp";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <div class=\"task-top-wrap\">\n      ");
  stack1 = depth0;
  stack2 = "controller.isBackgroundOperations";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      <span class=\"task-detail-log-rolename\">");
  stack1 = depth0;
  stack2 = "common.hosts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n\n      <div class=\"select-wrap\">\n        ");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":\n        ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "view.categories";
  stack3['contentBinding'] = stack4;
  stack4 = "content.value";
  stack3['optionValuePath'] = stack4;
  stack4 = "content.label";
  stack3['optionLabelPath'] = stack4;
  stack4 = "view.hostCategory";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    ");
  stack1 = depth0;
  stack2 = "view.isRequestSchedule";
  stack3 = helpers['if'];
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "view.isLevelLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(23, program23, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n\n\n  ");
  data.buffer.push("\n\n  <div ");
  stack1 = {};
  stack2 = "view.parentView.isTaskListHidden:hidden :task-list-main-warp";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <div class=\"task-top-wrap\">\n      <a class=\"task-detail-back-to-hosts\" href=\"javascript:void(null)\" ");
  stack1 = depth0;
  stack2 = "backToHostList";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ><i\n              class=\"icon-arrow-left\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "common.hosts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n\n      <div class=\"task-name-top task-detail-log-rolename\">");
  stack1 = depth0;
  stack2 = "common.tasks";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n      <div class=\"select-wrap tasks-list-select\">\n        ");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":\n        ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "view.categories";
  stack3['contentBinding'] = stack4;
  stack4 = "content.value";
  stack3['optionValuePath'] = stack4;
  stack4 = "content.label";
  stack3['optionLabelPath'] = stack4;
  stack4 = "view.taskCategory";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    <div id=\"host-log\">\n      ");
  stack1 = depth0;
  stack2 = "view.isTasksEmptyList";
  stack3 = helpers['if'];
  tmp1 = self.program(25, program25, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(27, program27, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n  </div>\n\n\n  <!-- TASK DETAILS --->\n\n  <div ");
  stack1 = {};
  stack2 = "view.parentView.isLogWrapHidden:hidden :task-detail-info";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <div class=\"task-top-wrap\">\n      <a class=\"task-detail-back\" href=\"javascript:void(null)\" ");
  stack1 = depth0;
  stack2 = "backToTaskList";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ><i\n              class=\"icon-arrow-left\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "common.tasks";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n\n      <div>\n        <i ");
  stack1 = {};
  stack2 = "view.openedTask.status :task-detail-status-ico view.openedTask.icon";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n\n        <div class=\"task-detail-ico-wrap\">\n          <a title=\"Click to Copy\" ");
  stack1 = depth0;
  stack2 = "taskInfo";
  stack3 = depth0;
  stack4 = "textTrigger";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-copy\"><i\n                  class=\"icon-copy\"></i> ");
  stack1 = depth0;
  stack2 = "common.copy";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n          <a title=\"Open in New Window\" ");
  stack1 = depth0;
  stack2 = "openTaskLogInDialog";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-open-dialog\"><i\n                  class=\"icon-external-link\"></i> ");
  stack1 = depth0;
  stack2 = "common.open";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n        </div>\n        <span class=\"task-detail-log-rolename\">");
  stack1 = depth0;
  stack2 = "view.openedTask.commandDetail";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n      </div>\n    </div>\n    ");
  stack1 = depth0;
  stack2 = "view.isLevelLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(30, program30, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(34, program34, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <div class=\"log-list-wrap\">");
  stack1 = depth0;
  stack2 = "hostPopup.noServicesToShow";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n      ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.services";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "servicesInfo";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n          <div ");
  stack1 = {};
  stack2 = "servicesInfo.isVisible::hidden :log-list-wrap";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <div ");
  stack1 = depth0;
  stack2 = "servicesInfo";
  stack3 = depth0;
  stack4 = "gotoHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-list-line-cursor\">\n              <div class=\"operation-name-icon-wrap\">\n                ");
  stack1 = depth0;
  stack2 = "statusIcon";
  stack3 = {};
  stack4 = "servicesInfo";
  stack3['servicesInfoBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                <a href=\"#\">\n                  ");
  stack1 = depth0;
  stack2 = "servicesInfo.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                </a>\n                ");
  stack1 = depth0;
  stack2 = "servicesInfo.isAbortable";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </div>\n              <div class=\"time-summary start-time-text\">");
  stack1 = depth0;
  stack2 = "servicesInfo.startTime";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n              <div class=\"time-summary duration-text\">");
  stack1 = depth0;
  stack2 = "servicesInfo.duration";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n              <div class=\"progress-bar span2\">\n                <div ");
  stack1 = {};
  stack2 = "servicesInfo.isInProgress:progress-striped servicesInfo.isInProgress:active servicesInfo.barColor :progress";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <div class=\"bar\" ");
  stack1 = {};
  stack2 = "servicesInfo.barWidth";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n                </div>\n              </div>\n              <div class=\"host-progress-num\">");
  stack1 = depth0;
  stack2 = "servicesInfo.progress";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "%</div>\n              <div class=\"show-details\"><i class=\"icon-caret-right\"></i></div>\n            </div>\n          </div>\n        ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "abortIcon";
  stack3 = {};
  stack4 = "servicesInfo";
  stack3['servicesInfoBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <div class=\"show-more\" ");
  stack1 = depth0;
  stack2 = "requestMoreOperations";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"#\">");
  stack1 = depth0;
  stack2 = "hostPopup.serviceInfo.showMore";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n        </div>\n      ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <a class=\"task-detail-back-to-hosts\" href=\"javascript:void(null)\" ");
  stack1 = depth0;
  stack2 = "backToServiceList";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " >\n          <i class=\"icon-arrow-left\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "common.operations";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </a>\n      ");
  return buffer;}

function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.sourceRequestScheduleRunning";
  stack3 = helpers['if'];
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.sourceRequestScheduleAborted";
  stack3 = helpers['if'];
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n        <div class=\"alert alert-info request-schedule-abort\">\n          ");
  stack1 = depth0;
  stack2 = "hostPopup.bgop.sourceRequestSchedule.running";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n          <button type=\"button\" class=\"btn btn-warning pull-right\"\n            ");
  stack1 = depth0;
  stack2 = "view.sourceRequestScheduleId";
  stack3 = depth0;
  stack4 = "doAbortRequestSchedule";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            ");
  stack1 = depth0;
  stack2 = "view.requestScheduleAbortLabel";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </button>\n        </div>\n      ");
  return buffer;}

function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <div class=\"alert alert-info request-schedule-abort\">\n          ");
  stack1 = depth0;
  stack2 = "hostPopup.bgop.sourceRequestSchedule.aborted";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      ");
  return buffer;}

function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <div id=\"host-info\" ");
  stack1 = {};
  stack2 = "view.isRequestSchedule:scheduled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "view.isHostEmptyList";
  stack3 = helpers['if'];
  tmp1 = self.program(18, program18, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(20, program20, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n    ");
  return buffer;}
function program18(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <div class=\"log-list-wrap\">");
  stack1 = depth0;
  stack2 = "hostPopup.noHostsToShow";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n        ");
  return buffer;}

function program20(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "hostInfo";
  stack7 = helpers.each;
  tmp1 = self.program(21, program21, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program21(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n            <div ");
  stack1 = {};
  stack2 = "hostInfo.isVisible::hidden :log-list-wrap";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <div ");
  stack1 = depth0;
  stack2 = "hostInfo";
  stack3 = depth0;
  stack4 = "gotoTasks";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-list-line-cursor\">\n                <div class=\"host-name-icon-wrap\">\n                  ");
  stack1 = depth0;
  stack2 = "statusIcon";
  stack3 = {};
  stack4 = "hostInfo";
  stack3['servicesInfoBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  <a href=\"#\">\n                    ");
  stack1 = depth0;
  stack2 = "hostInfo.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  </a>\n                </div>\n                <div class=\"progress-bar span2\">\n                  <div ");
  stack1 = {};
  stack2 = "hostInfo.isInProgress:progress-striped :active hostInfo.barColor :progress";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                    <div class=\"bar\" ");
  stack1 = {};
  stack2 = "hostInfo.barWidth";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n                  </div>\n                </div>\n                <div class=\"host-progress-num\">");
  stack1 = depth0;
  stack2 = "hostInfo.progress";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "%</div>\n                <div class=\"show-details\"><i class=\"icon-caret-right\"></i></div>\n              </div>\n            </div>\n          ");
  return buffer;}

function program23(depth0,data) {


  data.buffer.push("\n      <div class=\"spinner\"></div>\n    ");}

function program25(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <div class=\"log-list-wrap\">");
  stack1 = depth0;
  stack2 = "hostPopup.noTasksToShow";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n      ");
  return buffer;}

function program27(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.tasks";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "taskInfo";
  stack7 = helpers.each;
  tmp1 = self.program(28, program28, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program28(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n          <div ");
  stack1 = {};
  stack2 = "taskInfo.isVisible::hidden :log-list-wrap";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <div ");
  stack1 = depth0;
  stack2 = "taskInfo";
  stack3 = depth0;
  stack4 = "toggleTaskLog";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-list-line-cursor\">\n              <div class=\"operation-name-icon-wrap\">\n                ");
  stack1 = depth0;
  stack2 = "statusIcon";
  stack3 = {};
  stack4 = "taskInfo";
  stack3['servicesInfoBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                <a href=\"#\">\n                  ");
  stack1 = depth0;
  stack2 = "taskInfo.commandDetail";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              </div>\n              <div class=\"show-details\"><i class=\"icon-caret-right\"></i></div>\n            </div>\n          </div>\n        ");
  return buffer;}

function program30(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <div class=\"task-detail-log-info\">\n        <div class=\"content-area\">\n          <div class=\"task-detail-log-clipboard-wrap\"></div>\n          <div class=\"task-detail-log-maintext\">\n            ");
  stack1 = depth0;
  stack2 = "view.openedTask.isRebalanceHDFSTask";
  stack3 = helpers['if'];
  tmp1 = self.program(31, program31, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            <h5>");
  stack1 = depth0;
  stack2 = "common.stderr";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": &nbsp; <span class=\"muted\">");
  stack1 = depth0;
  stack2 = "view.openedTask.errorLog";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " </span></h5>\n            <pre class=\"stderr\">");
  stack1 = depth0;
  stack2 = "view.openedTask.stderr";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</pre>\n            <h5>");
  stack1 = depth0;
  stack2 = "common.stdout";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": &nbsp; <span class=\"muted\"> ");
  stack1 = depth0;
  stack2 = "view.openedTask.outputLog";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " </span></h5>\n            <pre class=\"stdout\">");
  stack1 = depth0;
  stack2 = "view.openedTask.stdout";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</pre>\n          </div>\n        </div>\n      </div>\n    ");
  return buffer;}
function program31(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              <h5>");
  stack1 = depth0;
  stack2 = "services.hdfs.rebalance.title";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h5>\n\n              <div class=\"progresspopup-rebalancehdfs\">\n                <div ");
  stack1 = {};
  stack2 = ":progress view.openedTask.isInProgress:progress-striped view.openedTask.barColor :active";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <div class=\"bar\" ");
  stack1 = {};
  stack2 = "view.openedTask.completionProgressStyle";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n                </div>\n              </div>\n              <div class=\"clearfix\">\n                <div class=\"pull-left\">\n                  ");
  stack1 = depth0;
  stack2 = "view.openedTask.dataMoved";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " moved /\n                  ");
  stack1 = depth0;
  stack2 = "view.openedTask.dataLeft";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " left /\n                  ");
  stack1 = depth0;
  stack2 = "view.openedTask.dataBeingMoved";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " being processed\n                </div>\n                ");
  stack1 = depth0;
  stack2 = "view.openedTask.isNotComplete";
  stack3 = helpers['if'];
  tmp1 = self.program(32, program32, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </div>\n              <hr>\n            ");
  return buffer;}
function program32(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  <button class=\"btn btn-danger pull-right\" ");
  stack1 = depth0;
  stack2 = "stopRebalanceHDFS";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.cancel";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n                ");
  return buffer;}

function program34(depth0,data) {


  data.buffer.push("\n      <div class=\"spinner\"></div>\n    ");}

function program36(depth0,data) {


  data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

function program38(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div class=\"page-bar pull-right no-borders mtm\">\n    <div class=\"items-on-page\">\n      <label>");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.rowsPerPageSelectView";
  stack3 = {};
  stack4 = "view.displayLength";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n    </div>\n    <div class=\"info\">");
  stack1 = depth0;
  stack2 = "view.paginationInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n    <div class=\"paging_two_button\">\n      ");
  stack1 = depth0;
  stack2 = "view.paginationFirst";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "view.paginationLeft";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "view.paginationRight";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "view.paginationLast";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n");
  return buffer;}

  data.buffer.push("\n\n<div class=\"host-component-popup-wrap\">\n\n");
  stack1 = depth0;
  stack2 = "view.parentView.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(36, program36, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  stack1 = depth0;
  stack2 = "view.isPaginate";
  stack3 = helpers['if'];
  tmp1 = self.program(38, program38, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/items_list_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <p>");
  stack1 = depth0;
  stack2 = "item";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.insertedItems";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "item";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/metric", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.itemView";
  stack3 = {};
  stack4 = "metric";
  stack3['metricBinding'] = stack4;
  stack4 = "view";
  stack3['widgetBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"accordion\" id=\"metricAccordion\">\n      ");
  stack1 = depth0;
  stack2 = "view.moreMetrics";
  stack3 = helpers.each;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n    ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <div class=\"accordion-group\">\n        <div class=\"accordion-heading\">\n          <a class=\"accordion-toggle\" data-toggle=\"collapse\" data-parent=\"#metricAccordion\"\n             href=\"#");
  stack1 = depth0;
  stack2 = "code";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "Collapse\">\n            <i class=\"icon-play\"></i>");
  stack1 = depth0;
  stack2 = "label";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </a>\n        </div>\n        <div id=\"");
  stack1 = depth0;
  stack2 = "code";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "Collapse\" class=\"accordion-body collapse\">\n          <div class=\"accordion-inner\">\n            ");
  stack1 = depth0;
  stack2 = "items.length";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n        </div>\n      </div>\n      ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n            <ul class=\"items\">\n              ");
  stack1 = depth0;
  stack2 = "items";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "metric";
  stack7 = helpers.each;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </ul>\n            ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "view.moreItemView";
  stack3 = {};
  stack4 = "metric";
  stack3['metricBinding'] = stack4;
  stack4 = "view";
  stack3['widgetBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "metric.notFound";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"span title\">\n  ");
  stack1 = depth0;
  stack2 = "common.metrics";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " <i class=\"icon-question-sign\"></i>\n</div>\n<ul class=\"display-inline-block nav nav-pills\">\n  ");
  stack1 = depth0;
  stack2 = "view.metrics";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "metric";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n  <li class=\"dropdown\">\n    <a ");
  stack1 = depth0;
  stack2 = "toggleMore";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n      ");
  stack1 = depth0;
  stack2 = "metric.more";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n      <b class=\"caret\"></b>\n    </a>\n\n    ");
  stack1 = depth0;
  stack2 = "view.showMore";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </li>\n</ul>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/modal_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <a class=\"close\" ");
  stack1 = depth0;
  stack2 = "onClose";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">x</a>\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "headerClass";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "header";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "bodyClass";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "encodeBody";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(12, program12, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "body";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  return buffer;}

function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "body";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}

function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "footerClass";
  stack3 = helpers['if'];
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(17, program17, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "footerClass";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  return buffer;}

function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <div class=\"modal-footer\">\n        ");
  stack1 = depth0;
  stack2 = "view.hasFooterCheckbox";
  stack3 = helpers['if'];
  tmp1 = self.program(18, program18, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.third";
  stack3 = helpers['if'];
  tmp1 = self.program(20, program20, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.secondary";
  stack3 = helpers['if'];
  tmp1 = self.program(22, program22, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.primary";
  stack3 = helpers['if'];
  tmp1 = self.program(24, program24, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n    ");
  return buffer;}
function program18(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          <label id=\"footer-checkbox\">");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "checkbox";
  stack3['classNames'] = stack4;
  stack4 = "view.isNotShowBgChecked";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " &nbsp;\n          ");
  stack1 = depth0;
  stack2 = "app.settings.notShowBgOperations";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n        ");
  return buffer;}

function program20(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          <button class=\"btn\" ");
  stack1 = {};
  stack2 = "view.disableThird";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "onThird";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.third";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</button>\n        ");
  return buffer;}

function program22(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          <button ");
  stack1 = {};
  stack2 = "view.disableSecondary";
  stack1['disabled'] = stack2;
  stack2 = ":btn view.secondaryClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "onSecondary";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.secondary";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</button>\n        ");
  return buffer;}

function program24(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          <button ");
  stack1 = {};
  stack2 = "view.disablePrimary";
  stack1['disabled'] = stack2;
  stack2 = ":btn view.primaryClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "onPrimary";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.primary";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</button>\n        ");
  return buffer;}

  data.buffer.push("\n\n\n<div class=\"modal-backdrop\"></div>\n<div class=\"modal\" id=\"modal\" tabindex=\"-1\" role=\"dialog\" aria-labelledby=\"modal-label\" aria-hidden=\"true\">\n  <div class=\"modal-header\">\n    ");
  stack1 = depth0;
  stack2 = "showCloseButton";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <h3 id=\"modal-label\">\n      ");
  stack1 = depth0;
  stack2 = "headerClass";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(5, program5, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </h3>\n  </div>\n  <div ");
  stack1 = {};
  stack2 = ":modal-body view.isHideBodyScroll:hidden-scroll";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    ");
  stack1 = depth0;
  stack2 = "bodyClass";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(9, program9, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  ");
  stack1 = depth0;
  stack2 = "showFooter";
  stack3 = helpers['if'];
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/modal_popups/alerts_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"alert-list-main-warp\">\n      <div class=\"top-wrap\">\n        <div class=\"service-top\">");
  stack1 = depth0;
  stack2 = "alerts.definition.details.serviceHost";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"name-top\">");
  stack1 = depth0;
  stack2 = "alerts.fastAccess.popup.body.name";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"last-triggered-top\">");
  stack1 = depth0;
  stack2 = "common.status";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n      </div>\n        <div id=\"alert-info\">\n          ");
  stack1 = depth0;
  stack2 = "view.isAlertEmptyList";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n    </div>\n\n    <div ");
  stack1 = {};
  stack2 = "view.filteringComplete:hidden :table-overlay";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <div class=\"spinner\"></div>\n    </div>\n\n    ");
  stack1 = depth0;
  stack2 = "view.isPaginate";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <div class=\"alert-list-wrap\">");
  stack1 = depth0;
  stack2 = "alerts.fastAccess.popup.body.noalerts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n          ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "alertInstance";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n              <div class=\"alert-list-wrap\">\n                <div class=\"alert-list-line-cursor\">\n                  <div class=\"service-text\"><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "alertInstance.service";
  stack3 = depth0;
  stack4 = "gotoService";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alertInstance.service.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                    ");
  stack1 = depth0;
  stack2 = "alertInstance.hostName";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  </div>\n                  <div class=\"name-text\">\n                    <div>\n                      <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "alertInstance";
  stack3 = depth0;
  stack4 = "gotoAlertDetails";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alertInstance.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                    </div>\n                    <div class=\"instance-text\">\n                      ");
  stack1 = depth0;
  stack2 = "alertInstance.text";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                    </div>\n                  </div>\n                  <div class=\"status-col\" rel=\"alert-status-tooltip\" ");
  stack1 = {};
  stack2 = "alertInstance.lastTriggered";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                    <span>\n                      <span class=\"status-icon\">");
  stack1 = depth0;
  stack2 = "alertInstance.status";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</span>\n                      <time class=\"timeago\" ");
  stack1 = {};
  stack2 = "alertInstance.lastTriggeredVerboseDisplay";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alertInstance.lastTriggeredForFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</time>\n                    </span>\n                  </div>\n              </div>\n            </div>\n          ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                      ");
  stack1 = depth0;
  stack2 = "alertInstance.service.displayName";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                      <a ");
  stack1 = depth0;
  stack2 = "alertInstance.host";
  stack3 = depth0;
  stack4 = "goToHostAlerts";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n                        ");
  stack1 = depth0;
  stack2 = "alertInstance.hostName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                      </a>\n                    ");
  return buffer;}
function program7(depth0,data) {


  data.buffer.push("\n                        &nbsp;/&nbsp;\n                      ");}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <div class=\"page-bar pull-right no-borders mtm\">\n        <div class=\"items-on-page\">\n          <label>");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.rowsPerPageSelectView";
  stack3 = {};
  stack4 = "view.displayLength";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n        </div>\n        <div class=\"info\">");
  stack1 = depth0;
  stack2 = "view.paginationInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"paging_two_button\">\n          ");
  stack1 = depth0;
  stack2 = "view.paginationFirst";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "view.paginationLeft";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "view.paginationRight";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "view.paginationLast";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    ");
  return buffer;}

function program11(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

  data.buffer.push("\n\n<div class=\"alerts-popup-wrap\">\n  ");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(11, program11, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/modal_popups/cluster_check_dialog", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <i class=\"icon-remove\"></i>&nbsp;<span>");
  stack1 = depth0;
  stack2 = "item.UpgradeChecks.check";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n        <pre>");
  stack1 = depth0;
  stack2 = "popup.clusterCheck.reason";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "item.UpgradeChecks.reason";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "<br/>");
  stack1 = depth0;
  stack2 = "popup.clusterCheck.failedOn";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "item.UpgradeChecks.failed_on";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</pre>\n    ");
  return buffer;}

  data.buffer.push("\n<div id=\"pre-upgrade-check\">\n  <h4>");
  stack1 = depth0;
  stack2 = "view.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</h4>\n  <div class=\"alert alert-warning\">\n    ");
  stack1 = depth0;
  stack2 = "view.alert";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <div class=\"limited-height-2\">\n    ");
  stack1 = depth0;
  stack2 = "view.checks";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "item";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/modal_popups/confirmation_feedback", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  <br />\n  <div class='alert'>\n    ");
  stack1 = depth0;
  stack2 = "view.parentView.additionalWarningMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <label class=\"checkbox\">");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "view.parentView.runMmOperation";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.parentView.turnOnMmMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n");
  return buffer;}

  data.buffer.push("\n<div>\n  ");
  stack1 = depth0;
  stack2 = "view.parentView.statusMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n");
  stack1 = depth0;
  stack2 = "view.parentView.additionalWarningMsg";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.parentView.putInMaintenance";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/modal_popups/invalid_KDC_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


  data.buffer.push("\n\n<div class=\"alert alert-warning\">\n  ");
  stack1 = depth0;
  stack2 = "view.warningMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n<form class=\"form-horizontal\">\n  <div class=\"control-group\">\n    <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "popup.invalid.KDC.admin.principal";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n    <div class=\"controls\">\n      ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "view.parentView.principal";
  stack3['valueBinding'] = stack4;
  stack4 = "form-control";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n  <div class=\"control-group\">\n    <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "popup.invalid.KDC.admin.password";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n    <div class=\"controls\">\n      ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "password";
  stack3['type'] = stack4;
  stack4 = "view.parentView.password";
  stack3['valueBinding'] = stack4;
  stack4 = "form-control";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n</form>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/modal_popups/prompt_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "view.parentView.errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n      ");
  return buffer;}

  data.buffer.push("\n<div class=\"prompt-popup\">\n  <form>\n    <div ");
  stack1 = {};
  stack2 = ":control-group view.parentView.isInvalid:error";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "view.text";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n      <div class=\"controls\">\n        ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "prompt-input";
  stack3['class'] = stack4;
  stack4 = "view.parentView.inputValue";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n      ");
  stack1 = depth0;
  stack2 = "view.parentView.isInvalid";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n  </form>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/progress", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "retry";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "common.retry";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "controller.tasks";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "task";
  stack7 = helpers.each;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.taskView";
  stack3 = {};
  stack4 = "task";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n        <div class=\"item\">\n          <div ");
  stack1 = {};
  stack2 = ":pull-left view.linkClass isKerberosWizard:span4 controller.isHA:span4 controller.isRMHA:span5 controller.isRollback:span3 controller.isReassign:span5";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <i ");
  stack1 = {};
  stack2 = "view.icon view.iconColor";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n            <a ");
  stack1 = depth0;
  stack2 = "task";
  stack3 = depth0;
  stack4 = "showHostProgressPopup";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " >");
  stack1 = depth0;
  stack2 = "task.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n          </div>\n          <div ");
  stack1 = {};
  stack2 = "view.showProgressBar::hide :row :span5 :pull-left";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <div ");
  stack1 = {};
  stack2 = ":progress-bar controller.isRollback::span8 controller.isRollback:span3";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <div class=\"progress-striped active progress-info progress\">\n                <div class=\"bar\" ");
  stack1 = {};
  stack2 = "view.barWidth";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n              </div>\n            </div>\n            <div ");
  stack1 = {};
  stack2 = ":span1 view.hidePercent:noDisplay";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "task.progress";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "&#37;</div>\n          </div>\n          <div>\n            ");
  stack1 = depth0;
  stack2 = "task.showRetry";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "task.showRollback";
  stack3 = helpers['if'];
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "task.showSkip";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n        </div>\n      ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              <a ");
  stack1 = depth0;
  stack2 = "retryTask";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-primary retry\"\n                                                          rel=\"tooltip\"\n                                                          data-trigger=\"click\" ");
  stack1 = {};
  stack2 = "view.showDBTooltip:testDBRetryTooltip";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                <i class=\"icon-repeat icon-white\"></i>\n                ");
  stack1 = depth0;
  stack2 = "common.retry";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              <a ");
  stack1 = depth0;
  stack2 = "rollback";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-primary retry\">\n                <i class=\"icon-repeat icon-white\"></i>\n                ");
  stack1 = depth0;
  stack2 = "common.rollBack";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              <a ");
  stack1 = depth0;
  stack2 = "skipTask";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-primary retry\">\n                <i class=\"icon-step-forward icon-white\"></i>\n                ");
  stack1 = depth0;
  stack2 = "common.skip";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            ");
  return buffer;}

function program12(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <button class=\"btn pull-left\" ");
  stack1 = {};
  stack2 = "controller.isBackButtonDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n    ");
  return buffer;}

  data.buffer.push("\n<div id=\"common-progress-page\">\n  <h2>");
  stack1 = depth0;
  stack2 = "view.headerTitle";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <div ");
  stack1 = {};
  stack2 = "view.noticeClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.notice";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "controller.showRetry";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  ");
  stack1 = depth0;
  stack2 = "controller.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(12, program12, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  <div class=\"btn-area\">\n    ");
  stack1 = depth0;
  stack2 = "view.showBackButton";
  stack3 = helpers['if'];
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <button class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "controller.isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "done";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.submitButtonText";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</button>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/rolling_restart_view", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <p>\n        ");
  stack1 = depth0;
  stack2 = "view.maintainanceMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </p>\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <p>\n        ");
  stack1 = depth0;
  stack2 = "view.suggestTurnOnMaintenanceMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </p>\n    ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <table>\n      <tr>\n        <td>");
  stack1 = depth0;
  stack2 = "common.restart";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "view.batchSize";
  stack3['valueBinding'] = stack4;
  stack4 = "span1";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
  stack1 = depth0;
  stack2 = "view.batchSizeMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <tr>\n        <td>");
  stack1 = depth0;
  stack2 = "rollingrestart.dialog.msg.timegap.prefix";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "view.interBatchWaitTimeSeconds";
  stack3['valueBinding'] = stack4;
  stack4 = "span1";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
  stack1 = depth0;
  stack2 = "rollingrestart.dialog.msg.timegap.suffix";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <tr>\n        <td>");
  stack1 = depth0;
  stack2 = "rollingrestart.dialog.msg.toleration.prefix";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "view.tolerateSize";
  stack3['valueBinding'] = stack4;
  stack4 = "span1";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
  stack1 = depth0;
  stack2 = "rollingrestart.dialog.msg.toleration.suffix";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n      </tr>\n      <tr>\n        <td>&nbsp;</td>\n        <td colspan=\"2\">\n          ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "view.staleConfigsOnly";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "view.staleConfigsOnlyMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n      <tr>\n        <td>&nbsp;</td>\n        <td colspan=\"2\">\n          ");
  stack1 = depth0;
  stack2 = "view.isServiceInMM";
  stack3 = helpers.unless;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </td>\n      </tr>\n\n    </table>\n  ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "view.turnOnMm";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  stack1 = depth0;
  stack2 = "view.turnOnMmMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  return buffer;}

function program8(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    <div class=\"alert alert-warn\">\n      <ul>\n      ");
  stack1 = depth0;
  stack2 = "view.errors";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "error";
  stack7 = helpers.each;
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </ul>\n    </div>\n  ");
  return buffer;}
function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <li>");
  stack1 = depth0;
  stack2 = "error";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</li>\n      ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"rolling-restart-view\">\n  <div class=\"alert alert-info\">\n    <p>\n      ");
  stack1 = depth0;
  stack2 = "view.restartMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </p>\n    ");
  stack1 = depth0;
  stack2 = "view.maintainanceMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "view.suggestTurnOnMaintenanceMsg";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  ");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.errors";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/selectable_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


  data.buffer.push("\n\n<div>\n  <div class=\"task-top-wrap\">\n    <div class=\"task-detail-ico-wrap\">\n      <a href=\"#\" title=\"Click to Copy\" ");
  stack1 = depth0;
  stack2 = "textTrigger";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-copy\"><i class=\"icon-copy\"></i> ");
  stack1 = depth0;
  stack2 = "common.copy";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    </div>\n  </div>\n  <div class=\"task-detail-log-info\">\n    <div class=\"content-area\" >\n      <div ");
  stack1 = {};
  stack2 = "view.textareaVisible::hidden :task-detail-log-clipboard-wrap";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <textarea class=\"task-detail-log-clipboard\"></textarea>\n      </div>\n      <div ");
  stack1 = {};
  stack2 = "view.textareaVisible:hidden :task-detail-log-maintext";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "content";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/settings", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


  data.buffer.push("\n\n<div class=\"admin-user-settings\">\n  <div>\n    <label>\n      ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "view.isNotShowBgChecked";
  stack3['checkedBinding'] = stack4;
  stack4 = "checkbox";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "app.settings.notshowBgOperationsPopup";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    </label>\n  </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/common/time_range", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "view.presetView";
  stack3 = {};
  stack4 = "preset";
  stack3['presetBinding'] = stack4;
  stack4 = "view";
  stack3['widgetBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"span title\">\n  ");
  stack1 = depth0;
  stack2 = "common.timeRange";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " <i class=\"icon-question-sign\"></i>\n</div>\n<ul class=\"display-inline-block nav nav-pills\">\n  ");
  stack1 = depth0;
  stack2 = "view.presets";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "preset";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "from";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.dateFromView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  stack1 = depth0;
  stack2 = "to";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.dateToView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n</ul>\n<div id=\"slider\">\n  <span class=\"period-marker\">");
  stack1 = depth0;
  stack2 = "view.rangeLabel";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n  <span class=\"now-marker\">");
  stack1 = depth0;
  stack2 = "view.nowLabel";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/experimental", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "App.supports";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(5, program5, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    <table class=\"table table-bordered table-striped\">\n      <thead>\n        <tr>\n          <td style=\"width: 20%; font-weight: bold;\">Experimental Functionality</td>\n          <td style=\"width: 20%; font-weight: bold;\">Enabled?</td>\n          <td style=\"width: 60%; font-weight: bold;\">Description</td>\n        </tr>\n      </thead>\n      <tbody>\n		    ");
  stack1 = depth0;
  stack2 = "view.supports";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "support";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </tbody>\n    </table>\n    <div class=\"control-group\" style=\"margin-bottom: 100px;\">\n	    <div class=\"controls pull-right\">\n	      <button class=\"btn\" ");
  stack1 = depth0;
  stack2 = "doCancel";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "form.cancel";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n	      <button class=\"btn btn-primary\" ");
  stack1 = depth0;
  stack2 = "doSave";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.save";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n	    </div>\n	  </div>\n  ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n		      <tr>\n		        <td>");
  stack1 = depth0;
  stack2 = "support.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n		        <td>");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "support.selected";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n		        <td></td>\n			    </tr>\n		    ");
  return buffer;}

function program5(depth0,data) {


  data.buffer.push("\n    <div class=\"alert alert-error\">\n      No application supports found. Please login to Ambari and revisit this page.\n    </div>\n  ");}

function program7(depth0,data) {


  data.buffer.push("\n  <div class=\"alert alert-error\">\n    No application found. Please login to Ambari and revisit this page.\n  </div>\n");}

  data.buffer.push("\n\n<div class=\"alert alert-error\">\n  <h2>Ambari's Experimental Functionality</h2>\n  <p>\n    <span class=\"label label-important\">Important</span> <strong>Changes made here are neither tested nor supported by Ambari</strong>\n  </p>\n  <p>\n    Ambari's experimental functionality is provided <strong>as-is</strong>.<br>\n    Any changes will effect application behavior, including data corruption. Users making changes do so at their own risk.<br>\n    Experimental functionality can be modified or removed at any time without notice.\n  </p>  \n</div>\n    \n");
  stack1 = depth0;
  stack2 = "App";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(7, program7, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/installer", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div id=\"version-mismatch-title\">");
  stack1 = depth0;
  stack2 = "app.versionMismatchAlert.title";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n  <div class=\"alert alert-danger\">\n    ");
  stack1 = depth0;
  stack2 = "versionConflictAlertBody";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div class=\"wizard\">\n    <div class=\"container\">\n      <div class=\"container-fluid\">\n        <div class=\"row-fluid\">\n          <div class=\"span3\">\n            <!--Sidebar content-->\n            <div class=\"well\">\n              <ul class=\"nav nav-pills nav-stacked\">\n                <li class=\"nav-header\">");
  stack1 = depth0;
  stack2 = "installer.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n                <li ");
  stack1 = {};
  stack2 = "isStep0:active view.isStep0Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep0";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step0.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep1:active view.isStep1Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep1";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep2:active view.isStep2Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep2";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep3:active view.isStep3Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep3";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep4:active view.isStep4Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep4";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step4.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep5:active view.isStep5Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep5";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step5.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep6:active view.isStep6Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep6";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step6.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep7:active view.isStep7Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep7";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step7.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep8:active view.isStep8Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep8";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step8.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep9:active view.isStep9Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep9";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step9.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep10:active view.isStep10Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\" ");
  stack1 = depth0;
  stack2 = "gotoStep10";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step10.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              </ul>\n            </div>\n          </div>\n          <div class=\"wizard-content well span9\">\n            ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </div>\n        </div>\n      </div>\n    </div>\n  </div>\n");
  return buffer;}

  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "isServerClientVersionMismatch";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/login", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n<div class=\"well login span4\">\n  <h2>");
  stack1 = depth0;
  stack2 = "login.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n  ");
  stack1 = depth0;
  stack2 = "errorMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  <label>");
  stack1 = depth0;
  stack2 = "login.username";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n  ");
  stack1 = depth0;
  stack2 = "view.loginTextField";
  stack3 = {};
  stack4 = "loginName";
  stack3['valueBinding'] = stack4;
  stack4 = "span4";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  <label>");
  stack1 = depth0;
  stack2 = "common.password";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n  ");
  stack1 = depth0;
  stack2 = "view.passTextField";
  stack3 = {};
  stack4 = "password";
  stack3['type'] = stack4;
  stack4 = "password";
  stack3['valueBinding'] = stack4;
  stack4 = "span4";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    <button class=\"btn btn-success login-btn\" ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "login.loginButton";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n</div>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  <div class=\"alert alert-error\">\n    ");
  stack1 = depth0;
  stack2 = "errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "isServerClientVersionMismatch";
  stack3 = helpers.unless;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div id=\"version-mismatch-title\">");
  stack1 = depth0;
  stack2 = "app.versionMismatchAlert.title";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n  <div class=\"alert alert-danger\">\n    ");
  stack1 = depth0;
  stack2 = "versionConflictAlertBody";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "isClusterDataLoaded";
  stack3 = helpers.unless;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "isClusterDataLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <h2>");
  stack1 = depth0;
  stack2 = "app.loadingPlaceholder";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " </h2>\n    <div class=\"progress progress-striped active\">\n      <div class=\"bar\" ");
  stack1 = {};
  stack2 = "controller.clusterDataLoadedPercent";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n    </div>\n  ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "isServerClientVersionMismatch";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.NavItemView";
  stack3 = {};
  stack4 = "category.name";
  stack3['itemBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n              <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "category.url";
  stack3 = depth0;
  stack4 = "goToAdmin";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " >");
  stack1 = depth0;
  stack2 = "category.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n          ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"row-fluid\">\n  <div id=\"main-admin-menu\" class=\"well span2\">\n      <ul class=\"nav nav-list\">\n        ");
  stack1 = depth0;
  stack2 = "view.categories";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "category";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </ul>\n  </div>\n  <div class=\"span10\">\n    <div class=\"row-fluid\">\n      ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/advanced", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<h5>");
  stack1 = depth0;
  stack2 = "common.advanced";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "<i class=\"icon-question-sign\"></i></h5>\n<div class=\"row\">\n  ");
  stack1 = depth0;
  stack2 = "admin.advanced.caution";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n<div class=\"row\">\n  <button ");
  stack1 = depth0;
  stack2 = "uninstall";
  stack3 = {};
  stack4 = "App.router.mainAdminAdvancedController";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn\"><i class=\"icon-trash\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.advanced.button.uninstallIncludingData";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n  <br/>\n  <button ");
  stack1 = depth0;
  stack2 = "view.params";
  stack3 = depth0;
  stack4 = "uninstall";
  stack5 = {};
  stack6 = "App.router.mainAdminAdvancedController";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn\"><i class=\"icon-trash\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.advanced.button.uninstallKeepData";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/advanced/uninstall", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "admin.confirmUninstall";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1));
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/audit", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "column";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "row";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.pager";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;}

  data.buffer.push("\n\n<table class=\"table table-striped\">\n  <thead>\n  <tr>\n    ");
  stack1 = depth0;
  stack2 = "view.columns";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "column";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </tr>\n  </thead>\n  <tbody>\n  ");
  stack1 = depth0;
  stack2 = "view.rows";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "row";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </tbody>\n</table>\n");
  stack1 = depth0;
  stack2 = "view.pager";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/authentication", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "App.FormFieldTemplate";
  stack3 = {};
  stack4 = "view.form.field.primaryServer";
  stack3['fieldBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  stack1 = depth0;
  stack2 = "App.FormFieldTemplate";
  stack3 = {};
  stack4 = "view.form.field.secondaryServer";
  stack3['fieldBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  stack1 = depth0;
  stack2 = "App.FormFieldTemplate";
  stack3 = {};
  stack4 = "view.form.field.useSsl";
  stack3['fieldBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  stack1 = depth0;
  stack2 = "App.FormFieldTemplate";
  stack3 = {};
  stack4 = "view.form.field.bindMethod";
  stack3['fieldBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  stack1 = depth0;
  stack2 = "view.useCredentials";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n    ");
  stack1 = depth0;
  stack2 = "App.FormFieldTemplate";
  stack3 = {};
  stack4 = "view.form.field.searchBaseDn";
  stack3['fieldBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "App.FormFieldTemplate";
  stack3 = {};
  stack4 = "view.form.field.usernameAttribute";
  stack3['fieldBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n\n    <h5>");
  stack1 = depth0;
  stack2 = "admin.authentication.form.configurationTest";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h5>\n    ");
  stack1 = depth0;
  stack2 = "App.FormFieldTemplate";
  stack3 = {};
  stack4 = "view.form.field.userDN";
  stack3['fieldBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "App.FormFieldTemplate";
  stack3 = {};
  stack4 = "view.form.field.userPassword";
  stack3['fieldBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n\n    <button ");
  stack1 = depth0;
  stack2 = "testConfiguration";
  stack3 = {};
  stack4 = "click";
  stack3['on'] = stack4;
  stack4 = "view.form";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn\">");
  stack1 = depth0;
  stack2 = "admin.authentication.form.testConfiguration";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n\n    ");
  stack1 = depth0;
  stack2 = "view.form.testConfigurationMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "App.FormFieldTemplate";
  stack3 = {};
  stack4 = "view.form.field.bindUser";
  stack3['fieldBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "App.FormFieldTemplate";
  stack3 = {};
  stack4 = "view.form.field.password";
  stack3['fieldBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "App.FormFieldTemplate";
  stack3 = {};
  stack4 = "view.form.field.passwordRetype";
  stack3['fieldBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <p ");
  stack1 = {};
  stack2 = "view.form.testConfigurationClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.form.testConfigurationMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n    ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "App.FormFieldTemplate";
  stack3 = {};
  stack4 = "view.form.field.method";
  stack3['fieldBinding'] = stack4;
  stack4 = "user-auth-method";
  stack3['id'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n\n");
  stack1 = depth0;
  stack2 = "view.ldapChecked";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n<div style=\"margin:40px 0\">\n  <button ");
  stack1 = depth0;
  stack2 = "updateValues";
  stack3 = {};
  stack4 = "view.form";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn\">");
  stack1 = depth0;
  stack2 = "form.cancel";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n  <button ");
  stack1 = depth0;
  stack2 = "view.form";
  stack3 = depth0;
  stack4 = "save";
  stack5 = {};
  stack6 = "App.router.mainAdminAuthenticationController";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-primary\">");
  stack1 = depth0;
  stack2 = "form.save";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/rollback", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<div class=\"wizard\">\n  ");
  stack1 = depth0;
  stack2 = "templates/common/progress";
  stack3 = helpers.template;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/rollbackHA/rollback_wizard", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div class=\"wizard\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li class=\"nav-header\">");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.rollback.header.title";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n              <li ");
  stack1 = {};
  stack2 = "isStep1:active view.isStep1Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep1";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.rollbackHighAvailability.wizard.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep2:active view.isStep2Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep2";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.rollbackHighAvailability.wizard.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep3:active view.isStep3Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep3";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.rollbackHighAvailability.wizard.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n            </ul>\n          </div>\n        </div>\n        <div class=\"wizard-content well span9\">\n          ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/rollbackHA/step1", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <h2>");
  stack1 = depth0;
  stack2 = "admin.rollbackHighAvailability.wizard.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n      <div class=\"alert alert-info\">\n          <p>\n              Select Additional NameNode host:\n            ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "view.addNNHosts";
  stack3['contentBinding'] = stack4;
  stack4 = "view.selectedAddNNHost";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </p>\n\n          <div class=\"code-snippet\">Tip: Suggested host <b>");
  stack1 = depth0;
  stack2 = "view.tipAddNNHost";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</b></div>\n      </div>\n      <div class=\"alert alert-info\">\n          <p>\n              Select Secondary NameNode host:\n            ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "view.sNNHosts";
  stack3['contentBinding'] = stack4;
  stack4 = "view.selectedSNNHost";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </p>\n\n          <div class=\"code-snippet\">Tip: Suggested host <b>");
  stack1 = depth0;
  stack2 = "view.tipSNNHost";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</b></div>\n      </div>\n      <div class=\"btn-area\">\n          <a ");
  stack1 = {};
  stack2 = ":btn controller.isNameServiceIdValid::disabled :btn-success :pull-right";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "done";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n      </div>\n  ");
  return buffer;}

function program3(depth0,data) {


  data.buffer.push("\n      <div class=\"spinner\"></div>\n  ");}

  data.buffer.push("\n<div>\n  ");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/rollbackHA/step2", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<div>\n  <h2>");
  stack1 = depth0;
  stack2 = "admin.rollbackHighAvailability.wizard.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n  <div class=\"alert alert-info\">\n      ");
  stack1 = depth0;
  stack2 = "view.step2BodyText";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n   <div class=\"btn-area\">\n      <a  ");
  stack1 = {};
  stack2 = ":btn :btn-success :pull-right";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n      <span class=\"pull-right btn-extra-info\">");
  stack1 = depth0;
  stack2 = "view.nnCheckPointText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n    </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/rollbackHA/step3", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<div>\n  <h2>");
  stack1 = depth0;
  stack2 = "admin.rollbackHighAvailability.wizard.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <div class=\"btn-area\">\n    <a ");
  stack1 = {};
  stack2 = ":btn controller.isNameServiceIdValid::disabled :btn-success :pull-right";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/step1", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<div id=\"ha-step1\">\n  <h2>");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <div class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step1.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    <div class=\"alert alert-warn\">\n      ");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step1.alert";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n\n  <div ");
  stack1 = {};
  stack2 = ":control-group view.showInputError:error :form-horizontal :nameserviceid-input";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step1.nameserviceid";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</label>\n\n    <div class=\"controls\">\n      ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "content.nameServiceId";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      <span  ");
  stack1 = {};
  stack2 = ":help-inline view.showInputError::hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step1.nameserviceid.error";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n      </span>\n    </div>\n  </div>\n  <div class=\"btn-area\">\n    <a ");
  stack1 = {};
  stack2 = ":btn controller.isNameServiceIdValid::disabled :btn-success :pull-right";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/step2", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <div class=\"control-group\">\n          <label class=\"control-label\">\n            ");
  stack1 = depth0;
  stack2 = "isCurNameNode";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "isAddNameNode";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "display_name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ":\n          </label>\n\n          <div class=\"controls\">\n            ");
  stack1 = depth0;
  stack2 = "view.shouldUseInputs";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "showAddControl";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "showRemoveControl";
  stack3 = helpers['if'];
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n        </div>\n      ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "common.current";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "common.additional";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "App.InputHostView";
  stack3 = {};
  stack4 = "this";
  stack3['componentBinding'] = stack4;
  stack4 = "isInstalled";
  stack3['disabledBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "App.SelectHostView";
  stack3 = {};
  stack4 = "this";
  stack3['componentBinding'] = stack4;
  stack4 = "isInstalled";
  stack3['disabledBinding'] = stack4;
  stack4 = "content.host_name";
  stack3['optionValuePath'] = stack4;
  stack4 = "content.host_info";
  stack3['optionLabelPath'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "App.AddControlView";
  stack3 = {};
  stack4 = "component_name";
  stack3['componentNameBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "App.RemoveControlView";
  stack3 = {};
  stack4 = "component_name";
  stack3['componentNameBinding'] = stack4;
  stack4 = "serviceComponentId";
  stack3['serviceComponentIddBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <div class=\"mapping-box round-corners well\">\n        <div class=\"hostString\"><span>");
  stack1 = depth0;
  stack2 = "hostInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span></div>\n        ");
  stack1 = depth0;
  stack2 = "masterServices";
  stack3 = helpers.each;
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n    ");
  return buffer;}
function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <span ");
  stack1 = {};
  stack2 = "isInstalled:assignedService:newService :round-corners";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "display_name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n        ");
  return buffer;}

function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <div class=\"remaining-hosts round-corners well\">\n        <span><strong>");
  stack1 = depth0;
  stack2 = "remainingHosts";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong> ");
  stack1 = depth0;
  stack2 = "installer.step5.attention";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span></div>\n    ");
  return buffer;}

  data.buffer.push("\n\n<h2>");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n<div class=\"alert alert-info\">\n  ");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step2.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n<div class=\"assign-masters\">\n  <div class=\"select-hosts span7\">\n    <form class=\"form-horizontal\" autocomplete=\"off\">\n      <!-- View for array controller -->\n      ");
  stack1 = depth0;
  stack2 = "servicesMasters";
  stack3 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </form>\n  </div>\n\n  <div class=\"host-assignments span5\">\n    ");
  stack1 = depth0;
  stack2 = "masterHostMapping";
  stack3 = helpers.each;
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n    ");
  stack1 = depth0;
  stack2 = "remainingHosts";
  stack3 = helpers['if'];
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  <div style=\"clear: both;\"></div>\n</div>\n<div class=\"btn-area\">\n  <a class=\"btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n  <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "submitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/step3", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              <li>");
  stack1 = depth0;
  stack2 = "item.hostName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</li>\n            ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.confirm.config.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "App.ServiceConfigView";
  stack3 = {};
  stack4 = "controller.isNotEditable";
  stack3['isNotEditableBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;}

function program5(depth0,data) {


  data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

  data.buffer.push("\n\n\n<h2>");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n<div class=\"alert alert-info\">\n  ");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.confirm.host.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n\n<div id=\"ha-step3-content\" class=\"well pre-scrollable\">\n  <div id=\"step8-info\">\n    <table id=\"ha-step3-review-table\">\n      <tr>\n        <td>");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.curNameNode";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
  stack1 = depth0;
  stack2 = "view.curNameNode";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td></td>\n      </tr>\n      <tr>\n        <td>");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.secNameNode";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
  stack1 = depth0;
  stack2 = "view.secondaryNameNode";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td><span class=\"to-be-disabled-red\"><i class=\"icon-minus\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.toBeDeleted";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span></td>\n      </tr>\n      <tr>\n        <td>");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.addNameNode";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>");
  stack1 = depth0;
  stack2 = "view.addNameNode";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td><span class=\"to-be-installed-green\"><i class=\"icon-plus\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.toBeInstalled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span></td>\n      </tr>\n      <tr>\n        <td>");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.journalNode";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td>\n          <ul>\n            ");
  stack1 = depth0;
  stack2 = "view.journalNodes";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "item";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </ul>\n        </td>\n        <td>\n          <ul>\n            <li><span class=\"to-be-installed-green\"><i\n                    class=\"icon-plus\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.toBeInstalled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span></li>\n            <li><span class=\"to-be-installed-green\"><i\n                    class=\"icon-plus\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.toBeInstalled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span></li>\n            <li><span class=\"to-be-installed-green\"><i\n                    class=\"icon-plus\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.toBeInstalled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span></li>\n          </ul>\n        </td>\n      </tr>\n    </table>\n  </div>\n</div>\n\n\n<div id=\"serviceConfig\">\n");
  stack1 = depth0;
  stack2 = "controller.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(5, program5, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n\n<div class=\"btn-area\">\n  <a class=\"btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n  <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "controller.isNextDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/step4", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"alert alert-error\">\n      ");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step4.error.nameNode";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  ");
  return buffer;}

  data.buffer.push("\n<div id=\"ha-step4\">\n  <h2>");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step4.bodyHeader";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n  ");
  stack1 = depth0;
  stack2 = "controller.isNameNodeStarted";
  stack3 = helpers.unless;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  <div class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "view.step4BodyText";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  <div class=\"btn-area\">\n    <a  ");
  stack1 = {};
  stack2 = "controller.isNextEnabled::disabled :btn :btn-success :pull-right";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "done";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n    <span class=\"pull-right btn-extra-info\">");
  stack1 = depth0;
  stack2 = "view.nnCheckPointText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/step5", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "templates/common/progress";
  stack3 = helpers.template;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/step6", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<div id=\"ha-step6\">\n  <h2>");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step6.bodyHeader";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n  <div class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "view.step6BodyText";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  <div class=\"btn-area\">\n    <a ");
  stack1 = {};
  stack2 = "controller.isNextEnabled::disabled :btn :btn-success :pull-right";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "done";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n    <span class=\"pull-right btn-extra-info\">");
  stack1 = depth0;
  stack2 = "view.jnCheckPointText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/step7", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "templates/common/progress";
  stack3 = helpers.template;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/step8", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<div id=\"ha-step8\">\n  <h2>");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step8.bodyHeader";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n  ");
  stack1 = depth0;
  stack2 = "view.step8BodyText";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  <div class=\"btn-area\">\n    <a class=\"btn btn-success pull-right\" ");
  stack1 = depth0;
  stack2 = "done";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n    <span class=\"pull-right btn-extra-info\">");
  data.buffer.push("</span>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/step9", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "templates/common/progress";
  stack3 = helpers.template;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/nameNode/wizard", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div class=\"wizard\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li class=\"nav-header\">");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n              <li ");
  stack1 = {};
  stack2 = "isStep1:active view.isStep1Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep1";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep2:active view.isStep2Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep2";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep3:active view.isStep3Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep3";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep4:active view.isStep4Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep4";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step4.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep5:active view.isStep5Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep5";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step5.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep6:active view.isStep6Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep6";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step6.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep7:active view.isStep7Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep7";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step7.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep8:active view.isStep8Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep8";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step8.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep9:active view.isStep9Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep9";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step9.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n            </ul>\n          </div>\n        </div>\n          <div class=\"wizard-content well span9\">\n            ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/progress", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.taskView";
  stack3 = {};
  stack4 = "task";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n    <div class=\"item\">\n      <div ");
  stack1 = {};
  stack2 = ":pull-left view.linkClass controller.isHA:span4 controller.isRMHA:span5 controller.isRollback:span3 controller.isReassign:span5";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <i ");
  stack1 = {};
  stack2 = "view.icon view.iconColor";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n        <a ");
  stack1 = depth0;
  stack2 = "task";
  stack3 = depth0;
  stack4 = "showHostProgressPopup";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " >");
  stack1 = depth0;
  stack2 = "task.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n      </div>\n      <div ");
  stack1 = {};
  stack2 = "view.showProgressBar::hide :row :span5 :pull-left";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <div ");
  stack1 = {};
  stack2 = ":progress-bar controller.isRollback::span8 controller.isRollback:span3";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <div class=\"progress-striped active progress-info progress\">\n            <div class=\"bar\" ");
  stack1 = {};
  stack2 = "view.barWidth";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n          </div>\n        </div>\n        <div ");
  stack1 = {};
  stack2 = ":span1 view.hidePercent:noDisplay";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "task.progress";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "&#37;</div>\n      </div>\n      <div>\n      ");
  stack1 = depth0;
  stack2 = "task.showRetry";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "task.showRollback";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "task.showSkip";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n    </div>\n  ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <a ");
  stack1 = depth0;
  stack2 = "retryTask";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-primary retry\"\n           rel=\"tooltip\" data-trigger=\"click\" ");
  stack1 = {};
  stack2 = "view.showDBTooltip:testDBRetryTooltip";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <i class=\"icon-repeat icon-white\"></i>\n          ");
  stack1 = depth0;
  stack2 = "common.retry";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </a>\n      ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <a ");
  stack1 = depth0;
  stack2 = "rollback";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-primary retry\">\n          <i class=\"icon-repeat icon-white\"></i>\n          ");
  stack1 = depth0;
  stack2 = "common.rollBack";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </a>\n      ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <a ");
  stack1 = depth0;
  stack2 = "skipTask";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-primary retry\">\n          <i class=\"icon-step-forward icon-white\"></i>\n          ");
  stack1 = depth0;
  stack2 = "common.skip";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </a>\n      ");
  return buffer;}

  data.buffer.push("\n<div id=\"ha-progress-page\">\n  <h2>");
  stack1 = depth0;
  stack2 = "view.headerTitle";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <div ");
  stack1 = {};
  stack2 = "view.noticeClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.notice";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</div>\n  ");
  stack1 = depth0;
  stack2 = "controller.tasks";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "task";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  <div class=\"btn-area\">\n    <button class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "controller.isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "done";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.submitButtonText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</button>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/resourceManager/step1", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<div>\n  <h2>");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <div class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.step1.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n\n  <div class=\"btn-area\">\n    <a class=\"btn btn-success pull-right\" ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/resourceManager/step2", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  <div class=\"assign-masters row-fluid\">\n    <div class=\"select-hosts span7\">\n      <div class=\"row-fluid\">\n        <div class=\"clearfix\"></div>\n        <div class=\"row-fluid\">\n          <div class=\"span12 control-group\">\n            <form class=\"form-horizontal\" autocomplete=\"off\">\n              <!-- View for array controller -->\n              ");
  stack1 = depth0;
  stack2 = "servicesMasters";
  stack3 = helpers.each;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </form>\n          </div>\n        </div>\n      </div>\n    </div>\n\n    <div class=\"host-assignments span5\">\n      ");
  stack1 = depth0;
  stack2 = "masterHostMapping";
  stack3 = helpers.each;
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n      ");
  stack1 = depth0;
  stack2 = "remainingHosts";
  stack3 = helpers['if'];
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n    <div class=\"clearfix\"></div>\n  </div>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <div class=\"row-fluid\">\n                  <div class=\"span5\">\n                    <div class=\"control-group\">\n                      <label class=\"pts pull-right\">\n                        ");
  stack1 = depth0;
  stack2 = "isAdditional";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(5, program5, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                        ");
  stack1 = depth0;
  stack2 = "display_name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ":\n                      </label>\n                    </div>\n                  </div>\n                  <div class=\"span7\">\n                    <div class=\"control-group\">\n                      ");
  stack1 = depth0;
  stack2 = "view.shouldUseInputs";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(9, program9, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    </div>\n                  </div>\n                </div>\n              ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                          ");
  stack1 = depth0;
  stack2 = "common.additional";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                        ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                          ");
  stack1 = depth0;
  stack2 = "common.current";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                        ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                        ");
  stack1 = depth0;
  stack2 = "App.InputHostView";
  stack3 = {};
  stack4 = "this";
  stack3['componentBinding'] = stack4;
  stack4 = "isInstalled";
  stack3['disabledBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                      ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                        ");
  stack1 = depth0;
  stack2 = "App.SelectHostView";
  stack3 = {};
  stack4 = "this";
  stack3['componentBinding'] = stack4;
  stack4 = "rm-host-select";
  stack3['class'] = stack4;
  stack4 = "isInstalled";
  stack3['disabledBinding'] = stack4;
  stack4 = "content.host_name";
  stack3['optionValuePath'] = stack4;
  stack4 = "content.host_info";
  stack3['optionLabelPath'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                      ");
  return buffer;}

function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <div class=\"mapping-box round-corners well\">\n          <div class=\"hostString\"><span>");
  stack1 = depth0;
  stack2 = "hostInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span></div>\n          ");
  stack1 = depth0;
  stack2 = "masterServicesToDisplay";
  stack3 = helpers.each;
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </div>\n      ");
  return buffer;}
function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <span ");
  stack1 = {};
  stack2 = "isInstalled:assignedService:newService :round-corners";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "display_name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n          ");
  return buffer;}

function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <div class=\"remaining-hosts round-corners well\">\n          <span><strong>");
  stack1 = depth0;
  stack2 = "remainingHosts";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong> ");
  stack1 = depth0;
  stack2 = "installer.step5.attention";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span></div>\n      ");
  return buffer;}

function program16(depth0,data) {


  data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

  data.buffer.push("\n<h2>");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n<div class=\"alert alert-info\">\n  ");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.step2.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n");
  stack1 = depth0;
  stack2 = "controller.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(16, program16, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n<div class=\"btn-area\">\n  <a class=\"btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n  <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "submitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/resourceManager/step3", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  <div id=\"serviceConfig\">\n\n    ");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.step3.confirm.config.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n    ");
  stack1 = depth0;
  stack2 = "App.ServiceConfigView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n");
  return buffer;}

function program3(depth0,data) {


  data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

  data.buffer.push("\n<h2>");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n<div class=\"alert alert-info\">\n  ");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.step3.confirm.host.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n\n<div id=\"ha-step3-content\" class=\"well pre-scrollable\">\n    <div id=\"step8-info\">\n        <table id=\"ha-step3-review-table\">\n            <tr>\n                <td>\n                    <b>");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.step3.currentRM";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</b>:\n                </td>\n                <td>");
  stack1 = depth0;
  stack2 = "controller.content.rmHosts.currentRM";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n                <td></td>\n            </tr>\n            <tr>\n                <td>\n                  <b>");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.step3.additionalRM";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</b>:\n                </td>\n                <td>");
  stack1 = depth0;
  stack2 = "controller.content.rmHosts.additionalRM";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n                <td><span class=\"to-be-installed-green\"><i class=\"icon-plus\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.highAvailability.wizard.step3.toBeInstalled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span></td>\n            </tr>\n        </table>\n    </div>\n</div>\n\n");
  stack1 = depth0;
  stack2 = "controller.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n<div class=\"btn-area\">\n    <a class=\"btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    <a class=\"btn btn-success pull-right\" ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "controller.isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/resourceManager/step4", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "templates/common/progress";
  stack3 = helpers.template;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/highAvailability/resourceManager/wizard", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program3(depth0,data) {


  data.buffer.push("\n                <div class=\"spinner\"></div>\n            ");}

  data.buffer.push("\n\n<div id=\"rm-ha-wizard\" class=\"wizard\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li class=\"nav-header\">");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n              <li ");
  stack1 = {};
  stack2 = "isStep1:active view.isStep1Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep1";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep2:active view.isStep2Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep2";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep3:active view.isStep3Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep3";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep4:active view.isStep4Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep4";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.rm_highAvailability.wizard.step4.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n            </ul>\n          </div>\n        </div>\n          <div class=\"wizard-content well span9\">\n            ");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/kerberos", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "securityEnabled";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div>\n      <p class=\"text-success\">");
  stack1 = depth0;
  stack2 = "admin.security.enabled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        <a class=\"btn btn-padding btn-warning admin-disable-security-btn\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "notifySecurityOffPopup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.kerberos.button.disable";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " </a>\n        <button class=\"btn btn-success\"");
  stack1 = depth0;
  stack2 = "regenerateKeytabs";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-repeat\"></i> ");
  stack1 = depth0;
  stack2 = "admin.kerberos.button.regenerateKeytabs";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n        <br/>\n      </p>\n    </div>\n    <div id=\"serviceConfig\">\n      ");
  stack1 = depth0;
  stack2 = "App.ServicesConfigView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div>\n      <p class=\"muted\">");
  stack1 = depth0;
  stack2 = "admin.security.disabled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        <a class=\"btn btn-padding btn-success admin-enable-security-btn\" ");
  stack1 = depth0;
  stack2 = "checkAndStartKerberosWizard";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.kerberos.button.enable";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " </a>\n        <br/>\n      </p>\n    </div>\n  ");
  return buffer;}

function program6(depth0,data) {


  data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "dataIsLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(6, program6, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/kerberos/disable", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div class=\"wizard\">\n  <div class=\"row-fluid\">\n    ");
  stack1 = depth0;
  stack2 = "templates/common/progress";
  stack3 = helpers.template;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/kerberos/regenerate_keytabs_popup_body", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<p class=\"alert alert-warning\">");
  stack1 = depth0;
  stack2 = "admin.kerberos.regenerate_keytabs.popup.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</p>\n<p>\n  <label>");
  stack1 = depth0;
  stack2 = "Em.Checkbox";
  stack3 = {};
  stack4 = "checkbox";
  stack3['classNames'] = stack4;
  stack4 = "view.parentView.regenerateKeytabsOnlyForMissing";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "admin.kerberos.regenerate_keytabs.checkbox.label";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n</p>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/kerberos/step1", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <label class=\"checkbox kerberos-step1-option-radiobutton\">\n        ");
  stack1 = depth0;
  stack2 = "Ember.RadioButton";
  stack3 = {};
  stack4 = "option.displayName";
  stack3['name'] = stack4;
  stack4 = "selectedItem";
  stack3['selectionBinding'] = stack4;
  stack4 = "option.value";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "  ");
  stack1 = depth0;
  stack2 = "option.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </label>\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <label class=\"checkbox kerberos-step1-prerequiste-checkboxes\"> ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "condition.checked";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "condition.displayText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " </label>\n    ");
  return buffer;}

  data.buffer.push("\n\n<div id=\"kerberos-wizard-step1\">\n  <h2>");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <p class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step1.info.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </p>\n\n  <div class=\"alert\">\n    ");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step1.alert.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n\n  <div id=\"kerberos-step1-options-body\">\n    ");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step1.body.text";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "options";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "option";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <br/>\n  </div>\n  <div class=\"alert alert-info\" id=\"kerberos-step1-prerequisites-body\">\n    <h5>");
  stack1 = depth0;
  stack2 = "selectedOption.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ":</h5>\n    <b>");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step1.prerequisites.label";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</b> <br/> <br/>\n    ");
  stack1 = depth0;
  stack2 = "selectedOption.preConditions";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "condition";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n\n\n  <div class=\"btn-area\">\n    <a id=\"submit-kerberos-step1\" class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/kerberos/step2", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "App.ServicesConfigView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  return buffer;}

function program3(depth0,data) {


  data.buffer.push("\n      <div class=\"spinner\"></div>\n    ");}

  data.buffer.push("\n\n<div id=\"kerberos-wizard-step2\">\n  <div id=\"serviceConfig\">\n    <h2>");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n    <div class=\"alert alert-info\">\n      ");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step2.info.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n    ");
  stack1 = depth0;
  stack2 = "isConfigsLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n    <div class=\"btn-area\">\n      <a id=\"back-kerberos-step2\" class=\"btn\" ");
  stack1 = {};
  stack2 = "isBackBtnDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n      <a id=\"submit-kerberos-step2\" class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/kerberos/step3", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<div id=\"kerberos-wizard-step3\">\n  ");
  stack1 = depth0;
  stack2 = "templates/common/progress";
  stack3 = helpers.template;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/kerberos/step4", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<div id=\"kerberos-wizard-step4\">\n  <div id=\"serviceConfig\">\n    <h2>");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step4.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n    <p class=\"alert alert-info\">\n      ");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step4.info.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    </p>\n\n    ");
  stack1 = depth0;
  stack2 = "App.ServicesConfigView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n\n    <div class=\"btn-area\">\n      <button id=\"back-kerberos-step4\" class=\"btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n      <button id=\"submit-kerberos-step4\" class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</button>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/kerberos/step5", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div id=\"kerberos-wizard-step5\">\n  ");
  stack1 = depth0;
  stack2 = "templates/common/progress";
  stack3 = helpers.template;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/kerberos/step6", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<div id=\"kerberos-wizard-step6\">\n  ");
  stack1 = depth0;
  stack2 = "templates/common/progress";
  stack3 = helpers.template;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/kerberos/step7", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<div id=\"kerberos-wizard-step7\">\n  ");
  stack1 = depth0;
  stack2 = "templates/common/progress";
  stack3 = helpers.template;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/kerberos/wizard", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div class=\"wizard\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li class=\"nav-header\"> <i class=\"icon-lock\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n              <li ");
  stack1 = {};
  stack2 = "isStep1:active view.isStep1Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep1";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep2:active view.isStep2Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep2";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep3:active view.isStep3Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep3";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep4:active view.isStep4Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep4";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step4.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep5:active view.isStep5Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep5";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step5.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep6:active view.isStep6Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep6";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step6.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep7:active view.isStep7Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep7";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step7.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n            </ul>\n          </div>\n        </div>\n        <div class=\"wizard-content well span9\">\n          ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/security", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "securityEnabled";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(5, program5, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div>\n      <p class=\"text-success\">");
  stack1 = depth0;
  stack2 = "admin.security.enabled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </p>\n    </div>\n    <div id=\"serviceConfig\">\n      ");
  stack1 = depth0;
  stack2 = "App.ServicesConfigView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          <a class=\"btn btn-padding btn-warning admin-disable-security-btn\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "notifySecurityOffPopup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.security.button.disable";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n          </a> <br/>\n        ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div>\n      <p class=\"muted\">");
  stack1 = depth0;
  stack2 = "admin.security.disabled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </p>\n    </div>\n  ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <a class=\"btn btn-padding btn-success admin-enable-security-btn\" ");
  stack1 = depth0;
  stack2 = "addSecurity";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.security.button.enable";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n          </a> <br/>\n        ");
  return buffer;}

function program8(depth0,data) {


  data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "dataIsLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/security/add/menu", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n\n<div class=\"wizard\" id=\"enable_security\">\n    <div class=\"container\">\n        <div class=\"container-fluid\">\n            <div class=\"row-fluid\">\n                <div class=\"span3\">\n                    <!--Sidebar content-->\n                    <div class=\"well\">\n                        <ul class=\"nav nav-pills nav-stacked\">\n                            <li class=\"nav-header\">");
  stack1 = depth0;
  stack2 = "admin.addSecurity.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n                            <li ");
  stack1 = {};
  stack2 = "isStep1:active view.isStep1Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep1";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.security.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                            <li ");
  stack1 = {};
  stack2 = "isStep2:active view.isStep2Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep2";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.security.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                            <li ");
  stack1 = {};
  stack2 = "isStep3:active view.isStep3Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep3";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.security.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                            <li ");
  stack1 = {};
  stack2 = "isStep4:active view.isStep4Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep4";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.security.step4.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                        </ul>\n                    </div>\n                </div>\n                <div class=\"wizard-content well span9\">\n                  ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                </div>\n            </div>\n        </div>\n    </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/security/add/step1", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <li>");
  stack1 = depth0;
  stack2 = "admin.security.step1.body.instruction5";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</li>\n      ");
  return buffer;}

  data.buffer.push("\n\n<h2>");
  stack1 = depth0;
  stack2 = "admin.security.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n<p class=\"alert\">\n  ");
  stack1 = depth0;
  stack2 = "admin.security.step1.body.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n</p>\n<div class=\"alert alert-info\">\n    <ol>\n        <li>");
  stack1 = depth0;
  stack2 = "admin.security.step1.body.instruction1";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n        <li>");
  stack1 = depth0;
  stack2 = "admin.security.step1.body.instruction2";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n        <li>");
  stack1 = depth0;
  stack2 = "admin.security.step1.body.instruction3";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n        <li>");
  stack1 = depth0;
  stack2 = "admin.security.step1.body.instruction4";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n      ");
  stack1 = depth0;
  stack2 = "view.shouldRemoveATS";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </ol>\n</div>\n\n\n<div class=\"btn-area\">\n    <a class=\"btn btn-success pull-right\" ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/security/add/step2", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<h2>");
  stack1 = depth0;
  stack2 = "admin.security.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n<div id=\"serviceConfig\">\n    <p class=\"alert alert-info\">\n      ");
  stack1 = depth0;
  stack2 = "admin.security.step2.body.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    </p>\n  ");
  stack1 = depth0;
  stack2 = "App.ServicesConfigView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n\n    <div class=\"btn-area\">\n        <a class=\"btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n\n        <button class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;\n        </button>\n    </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/security/add/step3", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div class=\"btn-area\">\n    <a class=\"btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    <div class=\"pull-right\">\n      <button class=\"btn btn-info\" ");
  stack1 = depth0;
  stack2 = "doDownloadCsv";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.security.step3.downloadCSV";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n      <button class=\"btn btn-success\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.apply";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</button>\n    </div>\n  </div>\n");
  return buffer;}

function program3(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n");}

  data.buffer.push("\n\n<h2>");
  stack1 = depth0;
  stack2 = "admin.security.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n<div class=\"alert alert-info\">");
  stack1 = depth0;
  stack2 = "admin.security.step3.notice";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n");
  stack1 = depth0;
  stack2 = "isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/security/add/step4", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <p ");
  stack1 = {};
  stack2 = "view.msgColor :alert";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.message";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n");
  return buffer;}

  data.buffer.push("\n\n<h2>");
  stack1 = depth0;
  stack2 = "admin.security.step4.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n");
  stack1 = depth0;
  stack2 = "view.message";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "App.MainServiceReconfigureView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n\n<div class=\"btn-area\">\n    <a class=\"btn\" ");
  stack1 = {};
  stack2 = "isBackBtnDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "done";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.done";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " </a>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/security/disable", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <p ");
  stack1 = {};
  stack2 = "view.msgColor :alert";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.message";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.message";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "App.MainServiceReconfigureView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n<div class=\"btn-area\">\n    <button class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "done";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.done";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " </button>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/security/notify_security_off_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div class=\"alert\">\n  ");
  stack1 = depth0;
  stack2 = "admin.security.disable.popup.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/serviceAccounts", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "controller.users";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(6, program6, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n      <div class=\"span6\">\n        <table class=\"table table-bordered table-striped\">\n          <thead>\n            <tr>\n              <th>");
  stack1 = depth0;
  stack2 = "common.name";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n              <th>");
  stack1 = depth0;
  stack2 = "common.value";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n            </tr>\n          </thead>\n          <tbody>\n            ");
  stack1 = depth0;
  stack2 = "controller.users";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "user";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </tbody>\n        </table>\n      </div>\n    ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "user.isVisible";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <tr>\n                  <td>");
  stack1 = depth0;
  stack2 = "user.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n                  <td>");
  stack1 = depth0;
  stack2 = "user.value";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n                </tr>\n              ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <p>");
  stack1 = depth0;
  stack2 = "admin.misc.nothingToShow";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</p>\n    ");
  return buffer;}

function program8(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

  data.buffer.push("\n<div class=\"admin-misc\">\n  <div class=\"header\">\n    <strong>");
  stack1 = depth0;
  stack2 = "admin.misc.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</strong>\n  </div>\n  ");
  stack1 = depth0;
  stack2 = "controller.dataIsLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/stack_and_upgrade", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div id=\"stack-upgrade-page\">\n  ");
  stack1 = depth0;
  stack2 = "App.MainAdminStackMenuView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/stack_upgrade/edit_repositories", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n  <div class=\"row-fluid os-block\">\n    <div class=\"span2\">\n      ");
  stack1 = depth0;
  stack2 = "os.osType";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  <div class=\"span10\">\n  ");
  stack1 = depth0;
  stack2 = "os.repositories";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "repository";
  stack7 = helpers.each;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n  </div>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div class=\"row-fluid\">\n      <div class=\"span3\">");
  stack1 = depth0;
  stack2 = "repository.repoName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n      <div ");
  stack1 = {};
  stack2 = "repository.hasError:error :control-group :span9";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "App.BaseUrlTextField";
  stack3 = {};
  stack4 = "repository";
  stack3['repositoryBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n        </div>\n      ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"alert alert-info\">\n  ");
  stack1 = depth0;
  stack2 = "admin.stackVersions.editRepositories.info";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n<div ");
  stack1 = {};
  stack2 = "view.parentView.hasErrors::hidden :alert :alert-warning";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  ");
  stack1 = depth0;
  stack2 = "admin.stackVersions.editRepositories.validation.warning";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n<div class=\"row-fluid\">\n  <div class=\"span2\"><strong>");
  stack1 = depth0;
  stack2 = "common.os";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</strong></div>\n  <div class=\"span10 row-fluid\">\n    <div class=\"span3\"><strong>");
  stack1 = depth0;
  stack2 = "common.name";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</strong></div>\n    <div class=\"span9\"><strong>");
  stack1 = depth0;
  stack2 = "admin.cluster.repositories.baseUrl";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</strong></div>\n  </div>\n</div>\n");
  stack1 = depth0;
  stack2 = "view.content.operatingSystems";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "os";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n<div>\n  <label>");
  stack1 = depth0;
  stack2 = "view.skipCheckBox";
  stack3 = {};
  stack4 = "view.parentView.skipValidation";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "installer.step1.advancedRepo.skipValidation.message";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    <i class=\"icon-question-sign\" rel=\"skip-validation-tooltip\"\n  data-toggle=\"tooltip\" ");
  stack1 = {};
  stack2 = "installer.step1.advancedRepo.skipValidation.tooltip";
  stack1['title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + "></i></label>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/stack_upgrade/services", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <tr>\n      <td class=\"service-display-name\">");
  stack1 = depth0;
  stack2 = "service.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n      <td class=\"service-stack-version\">");
  stack1 = depth0;
  stack2 = "service.serviceVersion";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n      <td class=\"stack-version-state\">\n        ");
  stack1 = depth0;
  stack2 = "service.isInstalled";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </td>\n      <td class=\"service-description\">");
  stack1 = depth0;
  stack2 = "service.comments";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</td>\n    </tr>\n    ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <span class=\"label label-success\">");
  stack1 = depth0;
  stack2 = "common.installed";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n        ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n          <a class=\"path-link\" ");
  stack1 = depth0;
  stack2 = "service.serviceName";
  stack3 = depth0;
  stack4 = "goToAddService";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            ");
  stack1 = depth0;
  stack2 = "services.service.add";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n        ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n\n  <div class=\"header\">\n    <strong>");
  stack1 = depth0;
  stack2 = "admin.cluster.repositories.repositories";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</strong>\n  </div>\n  <ul class=\"nav nav-tabs\">\n    <li class=\"active\">\n      <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "view.allRepositoriesGroups.stackVersion";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n    </li>\n  </ul>\n\n  <div class=\"admin-cluster\">\n    <div class=\"repositories-table\">\n      <div class=\"thead\">\n        <div class=\"th os-th\">");
  stack1 = depth0;
  stack2 = "common.os";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"th name-th\">");
  stack1 = depth0;
  stack2 = "common.name";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"th url-th\">");
  stack1 = depth0;
  stack2 = "installer.step1.advancedRepo.localRepo.column.baseUrl";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n      </div>\n      <div class=\"tbody\">\n        ");
  stack1 = depth0;
  stack2 = "view.allRepositoriesGroups";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "repoGroup";
  stack7 = helpers.each;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n    </div>\n  </div>\n\n");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n          <div class=\"trow\">\n            <div class=\"os-td\">\n              <label>\n                <span class=\"os\">");
  stack1 = depth0;
  stack2 = "repoGroup.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n              </label>\n            </div>\n            <div style=\"width:89%\">\n              ");
  stack1 = depth0;
  stack2 = "repoGroup.repositories";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "repository";
  stack7 = helpers.each;
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </div>\n          </div>\n        ");
  return buffer;}
function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <div class=\"sub-trow\">\n                  <div class=\"name-td\">");
  stack1 = depth0;
  stack2 = "repository.repoId";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                  <!--edit mode for current url-->\n                  ");
  stack1 = depth0;
  stack2 = "repository.onEdit";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(16, program16, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </div>\n              ");
  return buffer;}
function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                    <div ");
  stack1 = {};
  stack2 = ":url-td repository.empty-error:textfield-error repository.invalid-error:textfield-error";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                      ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "repository.baseUrl";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                    </div>\n                    <div class=\"clear-td\">\n                      ");
  stack1 = depth0;
  stack2 = "repository.clearAll";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    </div>\n                    <div class=\"edit-buttons-td\">\n                      <a class=\"btn\" ");
  stack1 = depth0;
  stack2 = "repository";
  stack3 = depth0;
  stack4 = "doCancel";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.cancel";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                    </div>\n                    <div class=\"edit-buttons-td\">\n                      ");
  stack1 = depth0;
  stack2 = "repository.empty-error";
  stack3 = helpers['if'];
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(14, program14, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    </div>\n                    <!--non-edit mode for current url-->\n                  ");
  return buffer;}
function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                        <a ");
  stack1 = depth0;
  stack2 = "repository";
  stack3 = depth0;
  stack4 = "clearGroupLocalRepository";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                          <i class=\"icon-remove-sign\"></i>\n                        </a>\n                      ");
  return buffer;}

function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                        <a class=\"btn btn-primary\" disabled=\"disabled\">");
  stack1 = depth0;
  stack2 = "common.save";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                      ");
  return buffer;}

function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                        <a class=\"btn btn-primary\" ");
  stack1 = depth0;
  stack2 = "repository";
  stack3 = depth0;
  stack4 = "saveRepoUrls";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.save";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                      ");
  return buffer;}

function program16(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                    <div class=\"url-text-td\">\n                      ");
  stack1 = depth0;
  stack2 = "repository.baseUrl";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                    </div>\n                    <div class=\"edit-td\">\n                      <a ");
  stack1 = depth0;
  stack2 = "repository";
  stack3 = depth0;
  stack4 = "onEditClick";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                        <i class=\"icon-edit\"></i> ");
  stack1 = depth0;
  stack2 = "common.edit";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                      </a>\n                    </div>\n                  ");
  return buffer;}

  data.buffer.push("\n\n\n<table class=\"table table-striped\">\n  <thead>\n  <tr>\n    <th>");
  stack1 = depth0;
  stack2 = "common.service";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n    <th>");
  stack1 = depth0;
  stack2 = "common.version";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n    <th>");
  stack1 = depth0;
  stack2 = "common.status";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n    <th>");
  stack1 = depth0;
  stack2 = "common.description";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n  </tr>\n  </thead>\n  <tbody>\n    ");
  stack1 = depth0;
  stack2 = "view.services";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "service";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </tbody>\n</table>\n\n");
  stack1 = depth0;
  stack2 = "App.stackVersionsAvailable";
  stack3 = helpers.unless;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/stack_upgrade/stack_upgrade_wizard", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n      <div>\n        ");
  stack1 = depth0;
  stack2 = "view.runningItem";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.failedItem";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.manualItem";
  stack3 = helpers['if'];
  tmp1 = self.program(23, program23, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.noActiveItem";
  stack3 = helpers['if'];
  tmp1 = self.program(34, program34, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n      ");
  stack1 = depth0;
  stack2 = "view.upgradeGroups";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "group";
  stack7 = helpers.each;
  tmp1 = self.program(36, program36, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <div class=\"box details-box\">\n            <div class=\"row-fluid\">\n              <div class=\"pull-left\">");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.dialog.inProgress";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "&nbsp;");
  stack1 = depth0;
  stack2 = "view.runningItem.text";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n              ");
  stack1 = depth0;
  stack2 = "view.isDetailsOpened";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </div>\n          </div>\n        ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                  <a href=\"#\" class=\"pull-right\" ");
  stack1 = depth0;
  stack2 = "toggleDetails";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.dialog.details.hide";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                ");
  stack1 = depth0;
  stack2 = "view.runningItem.isTasksLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(6, program6, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                  <div class=\"clear\">\n                    ");
  stack1 = depth0;
  stack2 = "App.upgradeTaskView";
  stack3 = {};
  stack4 = "view.taskDetails";
  stack3['contentBinding'] = stack4;
  stack4 = "view.outsideView";
  stack3['outsideViewBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  </div>\n                ");
  return buffer;}

function program6(depth0,data) {


  data.buffer.push("\n                  <div class=\"clear spinner\"></div>\n                ");}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                <a href=\"#\" class=\"pull-right\" ");
  stack1 = depth0;
  stack2 = "toggleDetails";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.dialog.details.open";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n              ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <div class=\"box details-box\">\n            <div class=\"row-fluid\">\n              <div class=\"pull-left\">");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.dialog.failed";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "&nbsp;");
  stack1 = depth0;
  stack2 = "view.failedItem.text";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n              ");
  stack1 = depth0;
  stack2 = "view.isDetailsOpened";
  stack3 = helpers['if'];
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(16, program16, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </div>\n            ");
  stack1 = depth0;
  stack2 = "view.isHoldingState";
  stack3 = helpers['if'];
  tmp1 = self.program(18, program18, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n        ");
  return buffer;}
function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                  <a href=\"#\" class=\"pull-right\" ");
  stack1 = depth0;
  stack2 = "toggleDetails";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.dialog.details.hide";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                ");
  stack1 = depth0;
  stack2 = "view.failedItem.isTasksLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(14, program14, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n               ");
  return buffer;}
function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                  <div class=\"clear\">\n                    ");
  stack1 = depth0;
  stack2 = "App.upgradeTaskView";
  stack3 = {};
  stack4 = "view.taskDetails";
  stack3['contentBinding'] = stack4;
  stack4 = "view.outsideView";
  stack3['outsideViewBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  </div>\n                ");
  return buffer;}

function program14(depth0,data) {


  data.buffer.push("\n                  <div class=\"clear spinner\"></div>\n                ");}

function program16(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                <a href=\"#\" class=\"pull-right\" ");
  stack1 = depth0;
  stack2 = "toggleDetails";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.dialog.details.open";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n              ");
  return buffer;}

function program18(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n              <div class=\"button-row\">\n                ");
  stack1 = depth0;
  stack2 = "view.isDowngradeAvailable";
  stack3 = helpers['if'];
  tmp1 = self.program(19, program19, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "view.failedItem.skippable";
  stack3 = helpers['if'];
  tmp1 = self.program(21, program21, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                <button class=\"btn\" ");
  stack1 = {};
  stack2 = "controller.requestInProgress";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.failedItem";
  stack3 = depth0;
  stack4 = "retry";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.retry";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n              </div>\n            ");
  return buffer;}
function program19(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                  <button class=\"btn btn-danger\" ");
  stack1 = {};
  stack2 = "controller.requestInProgress";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.failedItem";
  stack3 = depth0;
  stack4 = "confirmDowngrade";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.downgrade";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n                ");
  return buffer;}

function program21(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                  <button class=\"btn btn-warning\" ");
  stack1 = {};
  stack2 = "controller.requestInProgress";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.failedItem";
  stack3 = depth0;
  stack4 = "continue";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.dialog.continue";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n                ");
  return buffer;}

function program23(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.isFinalizeItem";
  stack3 = helpers['if'];
  tmp1 = self.program(24, program24, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(31, program31, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program24(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <div class=\"box details-box\">\n              <p><strong>");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.dialog.manual";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</strong></p>\n              ");
  stack1 = depth0;
  stack2 = "isDowngrade";
  stack3 = helpers['if'];
  tmp1 = self.program(25, program25, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(27, program27, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n              <label class=\"message\">\n                ");
  stack1 = depth0;
  stack2 = "Em.Checkbox";
  stack3 = {};
  stack4 = "view.isManualDone";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                ");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.dialog.manualDone";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </label>\n              <div class=\"button-row\">\n                ");
  stack1 = depth0;
  stack2 = "view.isDowngradeAvailable";
  stack3 = helpers['if'];
  tmp1 = self.program(29, program29, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                <button class=\"btn\" ");
  stack1 = depth0;
  stack2 = "closeWizard";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.finalize.later";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n                <button class=\"btn btn-success\" ");
  stack1 = {};
  stack2 = "view.isManualProceedDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.manualItem";
  stack3 = depth0;
  stack4 = "complete";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  ");
  stack1 = depth0;
  stack2 = "common.finalize";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                </button>\n              </div>\n            </div>\n          ");
  return buffer;}
function program25(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <p>");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.finalize.message.downgrade";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</p>\n              ");
  return buffer;}

function program27(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <p>");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.finalize.message.upgrade";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</p>\n              ");
  return buffer;}

function program29(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                  <button class=\"btn btn-danger\" ");
  stack1 = {};
  stack2 = "controller.requestInProgress";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.manualItem";
  stack3 = depth0;
  stack4 = "confirmDowngrade";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.downgrade";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n                ");
  return buffer;}

function program31(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <div class=\"box details-box\">\n              <p><strong>");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.dialog.manual";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</strong></p>\n              <p>");
  stack1 = depth0;
  stack2 = "view.manualItem.text";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n\n              <label class=\"message\">\n                ");
  stack1 = depth0;
  stack2 = "Em.Checkbox";
  stack3 = {};
  stack4 = "view.isManualDone";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                ");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.dialog.manualDone";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </label>\n              <div class=\"button-row\">\n                ");
  stack1 = depth0;
  stack2 = "view.isDowngradeAvailable";
  stack3 = helpers['if'];
  tmp1 = self.program(32, program32, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                <button class=\"btn\" ");
  stack1 = depth0;
  stack2 = "closeWizard";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.pauseUpgrade";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n                <button class=\"btn btn-success\" ");
  stack1 = {};
  stack2 = "view.isManualProceedDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.manualItem";
  stack3 = depth0;
  stack4 = "complete";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  ");
  stack1 = depth0;
  stack2 = "common.proceed";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                </button>\n              </div>\n            </div>\n          ");
  return buffer;}
function program32(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                  <button class=\"btn btn-danger\" ");
  stack1 = {};
  stack2 = "controller.requestInProgress";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.manualItem";
  stack3 = depth0;
  stack4 = "confirmDowngrade";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.downgrade";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n                ");
  return buffer;}

function program34(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <div class=\"box details-box\">\n            <div>\n              ");
  stack1 = depth0;
  stack2 = "admin.stackUpgrade.dialog.notActive";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </div>\n          </div>\n        ");
  return buffer;}

function program36(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "group.isVisible";
  stack3 = helpers['if'];
  tmp1 = self.program(37, program37, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program37(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "App.upgradeGroupView";
  stack3 = {};
  stack4 = "group";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  return buffer;}

function program39(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

  data.buffer.push("\n\n<div id=\"stack-upgrade-dialog\">\n  <div ");
  stack1 = {};
  stack2 = "view.isLoaded::hidden :row-fluid";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <div class=\"span3 task-list-main-warp\">");
  stack1 = depth0;
  stack2 = "controller.upgradeData.Upgrade.request_status";
  foundHelper = helpers.statusIcon;
  stack3 = foundHelper || depth0.statusIcon;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "statusIcon", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n      &nbsp;");
  stack1 = depth0;
  stack2 = "view.upgradeStatusLabel";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n    <div class=\"span8\">\n      ");
  stack1 = depth0;
  stack2 = "App.ProgressBarView";
  stack3 = {};
  stack4 = "view.overallProgress";
  stack3['progressBinding'] = stack4;
  stack4 = "controller.upgradeData.Upgrade.request_status";
  stack3['statusBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n    <div class=\"span1\">\n      ");
  stack1 = depth0;
  stack2 = "view.overallProgress";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "%\n    </div>\n  </div>\n\n  <div class=\"task-list scrollable-block task-list-main-warp\">\n    ");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  ");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers.unless;
  tmp1 = self.program(39, program39, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/stack_upgrade/upgrade_group", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, stack8, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div class=\"span3\">\n      ");
  stack1 = depth0;
  stack2 = "App.ProgressBarView";
  stack3 = {};
  stack4 = "view.content.progress";
  stack3['progressBinding'] = stack4;
  stack4 = "view.content.status";
  stack3['statusBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n    <div class=\"span1\">\n      <div>");
  stack1 = depth0;
  stack2 = "view.content.progress";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "%</div>\n    </div>\n  ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "item.isVisible";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, stack8;
  data.buffer.push("\n      <div class=\"row-fluid\">\n        <div ");
  stack1 = {};
  stack2 = "item.isActive::not-active-link :span8";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "item.status";
  foundHelper = helpers.statusIcon;
  stack3 = foundHelper || depth0.statusIcon;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "statusIcon", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n          <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.content.upgradeItems";
  stack3 = depth0;
  stack4 = "item";
  stack5 = depth0;
  stack6 = "toggleExpanded";
  stack7 = {};
  stack8 = "view";
  stack7['target'] = stack8;
  stack8 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack7;
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack8.call(depth0, stack6, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "item.context";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n        </div>\n        ");
  stack1 = depth0;
  stack2 = "item.isRunning";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n\n      ");
  stack1 = depth0;
  stack2 = "item.isExpanded";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n    ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          <div class=\"span3\">\n            ");
  stack1 = depth0;
  stack2 = "App.ProgressBarView";
  stack3 = {};
  stack4 = "item.progress";
  stack3['progressBinding'] = stack4;
  stack4 = "item.status";
  stack3['statusBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </div>\n          <div class=\"span1\">\n            <div>");
  stack1 = depth0;
  stack2 = "item.progress";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "%</div>\n          </div>\n        ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "item.isTasksLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(11, program11, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n        ");
  data.buffer.push("\n          <div class=\"task-list margin-bottom-5\">\n            ");
  stack1 = depth0;
  stack2 = "item.tasks";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "task";
  stack7 = helpers.each;
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n        ");
  return buffer;}
function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "App.upgradeTaskView";
  stack3 = {};
  stack4 = "task";
  stack3['contentBinding'] = stack4;
  stack4 = "item.tasks";
  stack3['tasksBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program11(depth0,data) {


  data.buffer.push("\n          <div class=\"spinner\"></div>\n        ");}

  data.buffer.push("\n\n\n<div class=\"row-fluid\">\n  <div ");
  stack1 = {};
  stack2 = "view.content.isActive::not-active-link :span8";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    ");
  stack1 = depth0;
  stack2 = "view.content.status";
  foundHelper = helpers.statusIcon;
  stack3 = foundHelper || depth0.statusIcon;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "statusIcon", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "controller.upgradeData.upgradeGroups";
  stack3 = depth0;
  stack4 = "view.content";
  stack5 = depth0;
  stack6 = "toggleExpanded";
  stack7 = {};
  stack8 = "view";
  stack7['target'] = stack8;
  stack8 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack7;
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack8.call(depth0, stack6, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.content.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n  </div>\n  ");
  stack1 = depth0;
  stack2 = "view.content.isRunning";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n\n<div ");
  stack1 = {};
  stack2 = "view.content.isExpanded::hidden :task-list";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n\n  ");
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.content.upgradeItems";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "item";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/stack_upgrade/upgrade_task", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, stack8;
  data.buffer.push("\n  <div ");
  stack1 = {};
  stack2 = "view.content.isActive::not-active-link";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.content.status";
  foundHelper = helpers.statusIcon;
  stack3 = foundHelper || depth0.statusIcon;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "statusIcon", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "<a\n    href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.tasks";
  stack3 = depth0;
  stack4 = "view.content";
  stack5 = depth0;
  stack6 = "toggleExpanded";
  stack7 = {};
  stack8 = "view.parentView";
  stack7['target'] = stack8;
  stack8 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack7;
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack8.call(depth0, stack6, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.content.command_detail";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n  </div>\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(6, program6, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n    <div class=\"task-details task-detail-info\">\n      <ul class=\"nav nav-tabs\">\n        <li class=\"active\"><a data-toggle=\"tab\" ");
  stack1 = {};
  stack2 = "view.logTabIdLink";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.stdout";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n        <li><a ");
  stack1 = {};
  stack2 = "view.errorTabIdLInk";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " data-toggle=\"tab\">");
  stack1 = depth0;
  stack2 = "common.stderr";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n      </ul>\n      <div class=\"tab-content\">\n        <div class=\"tab-pane active\" ");
  stack1 = {};
  stack2 = "view.logTabId";
  stack1['id'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <p>");
  stack1 = depth0;
  stack2 = "view.content.host_name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n          <div class=\"row-fluid\">\n            <p class=\"pull-left\">");
  stack1 = depth0;
  stack2 = "view.content.output_log";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n            <div class=\"manage-controls pull-right\">\n              <a title=\"Click to Copy\" ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "copyOutLog";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-copy\">\n                <i class=\"icon-copy\"></i> ");
  stack1 = depth0;
  stack2 = "common.copy";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n              <a title=\"Open in New Window\" ");
  stack1 = depth0;
  stack2 = "openOutLog";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-open-dialog\">\n                <i class=\"icon-external-link\"></i> ");
  stack1 = depth0;
  stack2 = "common.open";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </div>\n          </div>\n          <pre ");
  stack1 = {};
  stack2 = "view.outputLogOpened:hidden :stdout";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.content.stdout";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</pre>\n          ");
  stack1 = depth0;
  stack2 = "Ember.TextArea";
  stack3 = {};
  stack4 = "view.content.stdout";
  stack3['valueBinding'] = stack4;
  stack4 = "view.outputLogOpened::hidden";
  stack3['classBinding'] = stack4;
  stack4 = "readonly";
  stack3['readonly'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n        <div class=\"tab-pane\" ");
  stack1 = {};
  stack2 = "view.errorTabId";
  stack1['id'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <p>");
  stack1 = depth0;
  stack2 = "view.content.host_name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n          <div class=\"row-fluid\">\n            <p class=\"pull-left\">");
  stack1 = depth0;
  stack2 = "view.content.error_log";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n            <div class=\"manage-controls pull-right\">\n              <a title=\"Click to Copy\" ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "copyErrLog";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-copy\">\n                <i class=\"icon-copy\"></i> ");
  stack1 = depth0;
  stack2 = "common.copy";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n              <a title=\"Open in New Window\" ");
  stack1 = depth0;
  stack2 = "openErrorLog";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-open-dialog\">\n                <i class=\"icon-external-link\"></i> ");
  stack1 = depth0;
  stack2 = "common.open";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </div>\n          </div>\n          <pre ");
  stack1 = {};
  stack2 = "view.errorLogOpened:hidden :stderr";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.content.stderr";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</pre>\n          ");
  stack1 = depth0;
  stack2 = "Ember.TextArea";
  stack3 = {};
  stack4 = "view.content.stderr";
  stack3['valueBinding'] = stack4;
  stack4 = "view.errorLogOpened::hidden";
  stack3['classBinding'] = stack4;
  stack4 = "readonly";
  stack3['readonly'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  ");
  return buffer;}

function program6(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.outsideView";
  stack3 = helpers.unless;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.showContent";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/stack_upgrade/upgrade_version_box", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {


  data.buffer.push("\n  <i class=\"icon-exclamation-sign out-of-sync-badge\"></i>\n");}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <button class=\"btn btn-primary\"\n      ");
  stack1 = depth0;
  stack2 = "runAction";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = {};
  stack2 = "view.stateElement.isDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      ");
  stack1 = depth0;
  stack2 = "view.stateElement.text";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </button>\n  ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <span ");
  stack1 = {};
  stack2 = "view.stateElement.class";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.stateElement.text";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n  ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "view.stateElement.iconClass";
  stack3 = helpers['if'];
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "runAction";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.stateElement.text";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
  stack1 = depth0;
  stack2 = "view.stateElement.isInstalling";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program8(depth0,data) {

  var buffer = '', stack1, stack2;
  data.buffer.push("\n      <i ");
  stack1 = {};
  stack2 = "view.stateElement.iconClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n    ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "App.ProgressBarView";
  stack3 = {};
  stack4 = "view.installProgress";
  stack3['progressBinding'] = stack4;
  stack4 = "view.PROGRESS_STATUS";
  stack3['statusBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.isOutOfSync";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n<p class=\"version\">\n  <span>");
  stack1 = depth0;
  stack2 = "view.content.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n  <a class=\"pull-right not-active link-tooltip\" ");
  stack1 = depth0;
  stack2 = "editRepositories";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <i class=\"icon-edit\"></i>\n  </a>\n</p>\n<p class=\"repository-name\">(");
  stack1 = depth0;
  stack2 = "view.content.repositoryVersion";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ")</p>\n<div ");
  stack1 = {};
  stack2 = "view.stateElement.isInstalling:installing :align-center :state";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  ");
  stack1 = depth0;
  stack2 = "view.stateElement.isButton";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.stateElement.isLabel";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.stateElement.isLink";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n\n<div class=\"hosts-section\">\n  <div class=\"hosts-bar\">");
  stack1 = depth0;
  stack2 = "common.hosts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n  <div class=\"row-fluid host-link\">\n    <div class=\"span4 align-center\">\n      <div><a href=\"#\" class=\"hosts-tooltip not-active\" ");
  stack1 = {};
  stack2 = "view.content.noInitHostsTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.versionStateMap.not_installed";
  stack3 = depth0;
  stack4 = "showHosts";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "view.content.notInstalledHosts.length";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></div>\n      <div>");
  stack1 = depth0;
  stack2 = "admin.stackVersions.version.notInstalled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n    <div class=\"span4 align-center\">\n      <div><a href=\"#\" class=\"hosts-tooltip not-active\" ");
  stack1 = {};
  stack2 = "view.content.noInstalledHostsTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.versionStateMap.installed";
  stack3 = depth0;
  stack4 = "showHosts";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "view.content.installedHosts.length";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></div>\n      <div>");
  stack1 = depth0;
  stack2 = "common.installed";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n    <div class=\"span4 align-center\">\n      <div><a href=\"#\" class=\"hosts-tooltip not-active\" ");
  stack1 = {};
  stack2 = "view.content.noCurrentHostsTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.versionStateMap.current";
  stack3 = depth0;
  stack4 = "showHosts";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "view.content.currentHosts.length";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></div>\n      <div>");
  stack1 = depth0;
  stack2 = "common.current";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/admin/stack_upgrade/versions", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n          <li>\n            <a ");
  stack1 = depth0;
  stack2 = "category";
  stack3 = depth0;
  stack4 = "selectCategory";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n              ");
  stack1 = depth0;
  stack2 = "category.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n      ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <button class=\"btn btn-primary pull-right\" ");
  stack1 = depth0;
  stack2 = "goToVersions";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-external-link\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "admin.stackVersions.manageVersions";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n  ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "view.repoVersions";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "version";
  stack7 = helpers.each;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "version.isVisible";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "App.UpgradeVersionBoxView";
  stack3 = {};
  stack4 = "version";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  return buffer;}

function program9(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

  data.buffer.push("\n\n\n<div>\n  <div class=\"btn-group display-inline-block\">\n    <a class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n      <span class=\"filters-label\">");
  stack1 = depth0;
  stack2 = "common.filter";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": </span>\n            <span>\n              ");
  stack1 = depth0;
  stack2 = "view.selectedCategory.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              <span class=\"caret\"></span>\n            </span>\n    </a>\n    <ul class=\"dropdown-menu\">\n      ");
  stack1 = depth0;
  stack2 = "view.categories";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "category";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </ul>\n  </div>\n  ");
  stack1 = depth0;
  stack2 = "upgrade_ONLY_ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n<div class=\"row-fluid\">\n  ");
  stack1 = depth0;
  stack2 = "isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(9, program9, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div id=\"alert-actions-button\" class=\"pull-left\">\n    ");
  stack1 = depth0;
  stack2 = "App.MainAlertDefinitionActionsView";
  stack3 = {};
  stack4 = "App.router.mainAlertDefinitionActionsController";
  stack3['controllerBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.nameSort";
  stack3 = {};
  stack4 = "first name-sorting";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.statusSort";
  stack3 = {};
  stack4 = "status-sorting";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.serviceSort";
  stack3 = {};
  stack4 = "service-sorting";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.lastTriggeredSort";
  stack3 = {};
  stack4 = "last-triggred-sorting";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.enabledSort";
  stack3 = {};
  stack4 = "enabled-sorting";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "alertDefinition";
  stack7 = helpers.each;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n        <tr>\n          <td class=\"first\">\n            <span ");
  stack1 = {};
  stack2 = "alertDefinition.type";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = ":type-icon  alertDefinition.typeIconClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span>\n            <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "alertDefinition";
  stack3 = depth0;
  stack4 = "gotoAlertDetails";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alertDefinition.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n          </td>\n          <td>");
  stack1 = depth0;
  stack2 = "alertDefinition.status";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</td>\n          <td>");
  stack1 = depth0;
  stack2 = "alertDefinition.serviceDisplayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n          <td><time class=\"timeago\" ");
  stack1 = {};
  stack2 = "alertDefinition.lastTriggeredFormatted";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alertDefinition.lastTriggeredAgoFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</time></td>\n          <td class=\"last toggle-state-button\">\n            ");
  stack1 = depth0;
  stack2 = "alertDefinition.enabled";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(12, program12, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </td>\n        </tr>\n      ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "NON_ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "alertDefinition";
  stack3 = depth0;
  stack4 = "toggleState";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "alertDefinition.enabled:alert-definition-enable:alert-definition-disable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <span class=\"enable-disable-button\" ");
  stack1 = {};
  stack2 = "view.enabledTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                    ");
  stack1 = depth0;
  stack2 = "view.enabledDisplay";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  </span>\n                </a>\n              ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <span ");
  stack1 = {};
  stack2 = "alertDefinition.enabled:alert-definition-enable:alert-definition-disable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                      ");
  stack1 = depth0;
  stack2 = "alerts.table.state.enabled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                </span>\n              ");
  return buffer;}

function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "NON_ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "alertDefinition";
  stack3 = depth0;
  stack4 = "toggleState";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "alertDefinition.enabled:alert-definition-enable:alert-definition-disable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <span class=\"enable-disable-button\" ");
  stack1 = {};
  stack2 = "view.disabledTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                    ");
  stack1 = depth0;
  stack2 = "view.disabledDisplay";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  </span>\n                </a>\n              ");
  return buffer;}

function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <span ");
  stack1 = {};
  stack2 = "alertDefinition.enabled:alert-definition-enable:alert-definition-disable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                      ");
  stack1 = depth0;
  stack2 = "alerts.table.state.disabled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                </span>\n              ");
  return buffer;}

function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <tr>\n        <td class=\"first\"></td>\n        <td colspan=\"4\">\n          ");
  stack1 = depth0;
  stack2 = "alerts.table.noAlerts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n      ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n<div id=\"alert-groups-filter\" class=\"pull-left advanced-header-table groups-filter\">\n  ");
  stack1 = depth0;
  stack2 = "view.alertGroupFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n\n<div id=\"alerts\">\n  <div class=\"box-header row\">\n  </div>\n  <table class=\"table advanced-header-table table-bordered table-striped alerts-table\" id=\"alert-definitions-table\">\n    <thead>\n      ");
  stack1 = depth0;
  stack2 = "view.sortView";
  stack3 = {};
  stack4 = "label-row";
  stack3['classNames'] = stack4;
  stack4 = "view.filteredContent";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <tr class=\"filter-row\">\n      <th class=\"first\">");
  stack1 = depth0;
  stack2 = "view.nameFilterView";
  stack3 = {};
  stack4 = "name-filter";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "view.stateFilterView";
  stack3 = {};
  stack4 = "status-filter";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "view.serviceFilterView";
  stack3 = {};
  stack4 = "service-filter";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "view.triggeredFilterView";
  stack3 = {};
  stack4 = "last-triggered-filter";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "view.enabledFilterView";
  stack3 = {};
  stack4 = "enabled-filter";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n    </tr>\n    </thead>\n    <tbody>\n    ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(17, program17, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </tbody>\n  </table>\n\n  <div class=\"page-bar\">\n    <div class=\"filtered-info span4\">\n      <label>");
  stack1 = depth0;
  stack2 = "view.filteredContentInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " - <a ");
  stack1 = depth0;
  stack2 = "clearFilters";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          href=\"#\">");
  stack1 = depth0;
  stack2 = "tableView.filters.clearAllFilters";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></label>\n    </div>\n    <div class=\"selected-hosts-info span4\">\n    </div>\n    <div class=\"items-on-page\">\n      <label>");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.rowsPerPageSelectView";
  stack3 = {};
  stack4 = "view.displayLength";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n    </div>\n    <div class=\"info\">");
  stack1 = depth0;
  stack2 = "view.paginationInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n    <div class=\"paging_two_button\">\n      <a ");
  stack1 = {};
  stack2 = "view.paginationLeftClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "previousPage";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-arrow-left\"></i></a>\n      <a ");
  stack1 = {};
  stack2 = "view.paginationRightClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "nextPage";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-arrow-right\"></i></a>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/add_alert_definition/add_alert_definition", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div class=\"wizard\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li class=\"nav-header\">");
  stack1 = depth0;
  stack2 = "alerts.add.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n              <li ");
  stack1 = {};
  stack2 = "isStep1:active view.isStep1Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep1";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alerts.add.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep2:active view.isStep2Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep2";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alerts.add.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep3:active view.isStep3Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep3";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alerts.add.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n            </ul>\n          </div>\n        </div>\n        <div class=\"wizard-content well span9\">\n          ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/add_alert_definition/step1", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n  <a href=\"#\" ");
  stack1 = {};
  stack2 = ":btn type.isActive:active";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "type";
  stack3 = depth0;
  stack4 = "selectType";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <span ");
  stack1 = {};
  stack2 = "type.icon";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span> ");
  stack1 = depth0;
  stack2 = "type.value";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </a>\n");
  return buffer;}

  data.buffer.push("\n\n<h2>");
  stack1 = depth0;
  stack2 = "alerts.add.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n");
  stack1 = depth0;
  stack2 = "controller.alertDefinitionsTypes";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "type";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n<div class=\"btn-area\">\n  <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/add_alert_definition/step2", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<h2>");
  stack1 = depth0;
  stack2 = "alerts.add.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n");
  stack1 = depth0;
  stack2 = "App.AlertDefinitionConfigsView";
  stack3 = {};
  stack4 = true;
  stack3['canEdit'] = stack4;
  stack4 = true;
  stack3['isWizard'] = stack4;
  stack4 = "controller.content.selectedType";
  stack3['alertDefinitionTypeBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n\n<div class=\"btn-area\">\n  <a class=\"btn\" ");
  stack1 = {};
  stack2 = "isBackBtnDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n  <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/add_alert_definition/step3", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<h2>");
  stack1 = depth0;
  stack2 = "alerts.add.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n<div>\n  ");
  stack1 = depth0;
  stack2 = "alerts.add.step3.selectedType";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": <strong>");
  stack1 = depth0;
  stack2 = "controller.content.selectedType";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong>\n</div>\n<div class=\"control-group\">\n  <div class=\"controls\">\n    <pre class=\"alert-definition-review\">");
  stack1 = depth0;
  stack2 = "view.alertDefinitionToDisplay";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</pre>\n  </div>\n</div>\n\n<div class=\"btn-area\">\n  <a class=\"btn\" ");
  stack1 = {};
  stack2 = "isBackBtnDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n  <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "done";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.done";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " </a>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/add_custom_config_to_alert_notification_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"alert alert-danger\">\n      ");
  stack1 = depth0;
  stack2 = "view.errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  ");
  return buffer;}

  data.buffer.push("\n\n<form class=\"form-horizontal alert-configs\">\n  <div ");
  stack1 = {};
  stack2 = "view.isError:error :control-group";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "common.name";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " </label>\n\n    <div class=\"controls\">\n      ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "controller.newCustomProperty.name";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n\n  <div class=\"control-group\">\n    <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "common.value";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " </label>\n\n    <div class=\"controls\">\n      ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "controller.newCustomProperty.value";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n\n  ");
  stack1 = depth0;
  stack2 = "view.isError";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n</form>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/add_definition_to_group_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"text-warning\">\n      ");
  stack1 = depth0;
  stack2 = "view.parentView.warningMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  ");
  return buffer;}

function program3(depth0,data) {


  data.buffer.push("\n             <i class='icon-ok-sign'></i>\n          ");}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                    <li>\n                      <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "service";
  stack3 = depth0;
  stack4 = "selectFilterService";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                        ");
  stack1 = depth0;
  stack2 = "service.selected";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                        ");
  stack1 = depth0;
  stack2 = "service.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                      </a>\n                    </li>\n                  ");
  return buffer;}
function program6(depth0,data) {


  data.buffer.push("\n                            <i class='icon-ok-sign'></i>\n                        ");}

function program8(depth0,data) {


  data.buffer.push("\n                            <i class='icon-placeholder'></i>\n                        ");}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                    <li>\n                      <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "component";
  stack3 = depth0;
  stack4 = "selectFilterComponent";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                        ");
  stack1 = depth0;
  stack2 = "component.selected";
  stack3 = helpers['if'];
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(13, program13, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                        ");
  stack1 = depth0;
  stack2 = "component.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                      </a>\n                    </li>\n                  ");
  return buffer;}
function program11(depth0,data) {


  data.buffer.push("\n                            <i class='icon-ok-sign'></i>\n                        ");}

function program13(depth0,data) {


  data.buffer.push("\n                            <i class='icon-placeholder'></i>\n                        ");}

function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "entry";
  stack7 = helpers.each;
  tmp1 = self.program(16, program16, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program16(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            <tr ");
  stack1 = {};
  stack2 = "entry.filtered::hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <td width=\"10%\">\n                ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "entry.selected";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </td>\n              <td width=\"35%\">\n                ");
  stack1 = depth0;
  stack2 = "entry.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </td>\n              <td>\n                ");
  stack1 = depth0;
  stack2 = "entry.serviceNameDisplay";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </td>\n              <td>\n                ");
  stack1 = depth0;
  stack2 = "entry.componentNameDisplay";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </td>\n            </tr>\n          ");
  return buffer;}

function program18(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <tr>\n            <td width=\"30%\"></td>\n            <td>");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_groups_popup.addDefinition.noDefinitions";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n          </tr>\n        ");
  return buffer;}

function program20(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div class=\"page-bar pull-right no-borders\">\n    <div class=\"items-on-page\">\n        <label>");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.rowsPerPageSelectView";
  stack3 = {};
  stack4 = "view.displayLength";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n    </div>\n    <div class=\"info\">");
  stack1 = depth0;
  stack2 = "view.paginationInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n    <div class=\"paging_two_button\">\n      ");
  stack1 = depth0;
  stack2 = "view.paginationFirst";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "view.paginationLeft";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "view.paginationRight";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "view.paginationLast";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n");
  return buffer;}

  data.buffer.push("\n\n\n<form class=\"form-horizontal mbm\" autocomplete=\"off\">\n  <div class=\"override-controls\">\n    <div class=\"alert alert-info\">");
  stack1 = depth0;
  stack2 = "view.parentView.dialogMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n  ");
  stack1 = depth0;
  stack2 = "view.parentView.warningMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <table style=\"width: 100%;\">\n      <tr>\n        <td>\n            <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "toggleShowSelectedDefs";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.defSelectMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n          ");
  stack1 = depth0;
  stack2 = "view.showOnlySelectedDefs";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </td>\n        <td width=\"20%\">\n          <div class=\"row\">\n            <div class=\"span2\" id=\"filter-dropdown-div\">\n              <!-- services drop-down -->\n              <div class=\"btn-group\">\n                <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\" ");
  stack1 = {};
  stack2 = "view.isDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  ");
  stack1 = depth0;
  stack2 = "common.service";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                    <span class=\"caret\"></span>\n                </button>\n                <ul class=\"dropdown-menu\">\n                    <!-- dropdown menu links -->\n                  ");
  stack1 = depth0;
  stack2 = "view.filterServices";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "service";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </ul>\n              </div>\n            </div>\n            <div class=\"span2\" id=\"component-dropdown-div\">\n              <!-- definition-components drop-down -->\n              <div class=\"btn-group\">\n                <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\" ");
  stack1 = {};
  stack2 = "view.isDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  ");
  stack1 = depth0;
  stack2 = "common.component";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                    <span class=\"caret\"></span>\n                </button>\n                <ul class=\"dropdown-menu\">\n                    <!-- dropdown menu links -->\n                  ");
  stack1 = depth0;
  stack2 = "view.filterComponents";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </ul>\n              </div>\n            </div>\n          </div>\n        </td>\n      </tr>\n    </table>\n    <table class=\"table table-striped hosts-table\">\n      <thead>\n        <tr class=\"success\">\n          <th width=\"10%\">\n            ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "view.allDefsSelected";
  stack3['checkedBinding'] = stack4;
  stack4 = "view.isDisabled";
  stack3['disabledBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </th>\n          <th width=\"35%\">");
  stack1 = depth0;
  stack2 = "common.alertDefinition";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n          <th width=\"25%\">");
  stack1 = depth0;
  stack2 = "common.service";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n          <th width=\"30%\">");
  stack1 = depth0;
  stack2 = "common.component";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n        </tr>\n      </thead>\n    </table>\n    <div class=\"hosts-table-container\">\n      <table class=\"table table-striped hosts-table\">\n        ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = helpers['if'];
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(18, program18, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </table>\n    </div>\n</div>\n</form>\n\n");
  stack1 = depth0;
  stack2 = "view.isPaginate";
  stack3 = helpers['if'];
  tmp1 = self.program(20, program20, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/alert_definitions_actions", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n      <li>\n        <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "action";
  stack3 = depth0;
  stack4 = "actionHandler";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <span ");
  stack1 = {};
  stack2 = "action.icon";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span>\n          ");
  stack1 = depth0;
  stack2 = "action.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </a>\n      </li>\n      ");
  stack1 = depth0;
  stack2 = "action.showDivider";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program2(depth0,data) {


  data.buffer.push("\n        <li class=\"divider\"></li>\n      ");}

  data.buffer.push("\n\n<div class=\"dropdown\">\n  <a class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">");
  stack1 = depth0;
  stack2 = "common.actions";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " <span class=\"caret\"></span></a>\n  <ul class=\"dropdown-menu pull-left\">\n    ");
  stack1 = depth0;
  stack2 = "controller";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "action";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </ul>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/alert_groups/success_popup_body", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<p>");
  stack1 = depth0;
  stack2 = "alerts.groups.successPopup.body.created";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " - <strong>");
  stack1 = depth0;
  stack2 = "view.dataForSuccessPopup.created";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong></p>\n<p>");
  stack1 = depth0;
  stack2 = "alerts.groups.successPopup.body.updated";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " - <strong>");
  stack1 = depth0;
  stack2 = "view.dataForSuccessPopup.updated";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong></p>\n<p>");
  stack1 = depth0;
  stack2 = "alerts.groups.successPopup.body.deleted";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " - <strong>");
  stack1 = depth0;
  stack2 = "view.dataForSuccessPopup.deleted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong></p>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/configs", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div ");
  stack1 = {};
  stack2 = ":control-group property.name property.isValid::error";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      ");
  stack1 = depth0;
  stack2 = "property.isPreLabeled";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <label ");
  stack1 = {};
  stack2 = ":control-label property.isShifted:shifted";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "property.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div ");
  stack1 = {};
  stack2 = ":controls property.isShifted:shifted";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "property.viewClass";
  stack3 = {};
  stack4 = "property";
  stack3['propertyBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <div class=\"controls\">\n          <label class=\"radio\">\n            ");
  stack1 = depth0;
  stack2 = "property.viewClass";
  stack3 = {};
  stack4 = "property";
  stack3['propertyBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  stack1 = depth0;
  stack2 = "property.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </label>\n        </div>\n      ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"error-message controls\">\n      ");
  stack1 = depth0;
  stack2 = "view.errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  ");
  return buffer;}

  data.buffer.push("\n<form class=\"form-horizontal alert-configs\">\n  ");
  stack1 = depth0;
  stack2 = "controller.configs";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "property";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "controller.hasThresholdsError";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</form>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/configs/alert_config_text_field", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <span class=\"add-on\">");
  stack1 = depth0;
  stack2 = "view.property.unit";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n  ");
  return buffer;}

  data.buffer.push("\n<div ");
  stack1 = {};
  stack2 = "view.property.unit:input-append";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "view.property.value";
  stack3['valueBinding'] = stack4;
  stack4 = "view.property.isDisabled";
  stack3['disabledBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  stack1 = depth0;
  stack2 = "view.property.unit";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/configs/alert_config_threshold", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div ");
  stack1 = {};
  stack2 = ":span3 view.property.valueMetric:input-append";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "view.property.displayValue";
  stack3['valueBinding'] = stack4;
  stack4 = "view.property.isDisabled";
  stack3['disabledBinding'] = stack4;
  stack4 = "span7";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.property.valueMetric";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <span class=\"add-on\">");
  stack1 = depth0;
  stack2 = "view.property.valueMetric";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n        ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "view.property.text";
  stack3['valueBinding'] = stack4;
  stack4 = "view.property.isDisabled";
  stack3['disabledBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  return buffer;}

  data.buffer.push("\n\n<div>\n  <div class=\"span2 badge-container\"><span ");
  stack1 = {};
  stack2 = "view.property.badgeCssClass :alert-threshold-badge :alert-state-single-host :label";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.property.badge";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>&nbsp;</div>\n  ");
  stack1 = depth0;
  stack2 = "view.property.showInputForValue";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  <div ");
  stack1 = {};
  stack2 = ":alert-text-input view.property.showInputForValue:span7:span10";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    ");
  stack1 = depth0;
  stack2 = "view.property.showInputForText";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/create_alert_notification", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <div ");
  stack1 = {};
  stack2 = ":control-group controller.inputFields.email.errorMsg:error";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <label class=\"control-label\" for=\"inputEmail\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.email.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "controller.inputFields.email.value";
  stack3['valueBinding'] = stack4;
  stack4 = "inputEmail";
  stack3['id'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n\n        <div class=\"controls error-msg\">\n          ");
  stack1 = depth0;
  stack2 = "controller.inputFields.email.errorMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n\n      <div class=\"control-group\">\n        <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.SMTPServer.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "controller.inputFields.SMTPServer.value";
  stack3['valueBinding'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n\n      <div ");
  stack1 = {};
  stack2 = ":control-group controller.inputFields.SMTPPort.errorMsg:error";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.SMTPPort.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "controller.inputFields.SMTPPort.value";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n\n        <div class=\"controls error-msg\">\n          ");
  stack1 = depth0;
  stack2 = "controller.inputFields.SMTPPort.errorMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n\n      </div>\n\n      <div ");
  stack1 = {};
  stack2 = ":control-group controller.inputFields.emailFrom.errorMsg:error";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.emailFrom.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "controller.inputFields.emailFrom.value";
  stack3['valueBinding'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n\n        <div class=\"controls error-msg\">\n          ");
  stack1 = depth0;
  stack2 = "controller.inputFields.emailFrom.errorMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n\n      <div class=\"control-group\">\n        <label class=\"control-label\" for=\"inputUseAuthentication\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.SMTPUseAuthentication.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.Checkbox";
  stack3 = {};
  stack4 = "controller.inputFields.SMTPUseAuthentication.value";
  stack3['checkedBinding'] = stack4;
  stack4 = "inputUseAuthentication";
  stack3['id'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n\n      <div class=\"control-group\">\n        <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.SMTPUsername.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "controller.inputFields.SMTPUseAuthentication.invertedValue";
  stack3['disabledBinding'] = stack4;
  stack4 = "controller.inputFields.SMTPUsername.value";
  stack3['valueBinding'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n\n      <div class=\"control-group\">\n        <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.SMTPPassword.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "password";
  stack3['type'] = stack4;
  stack4 = "controller.inputFields.SMTPUseAuthentication.invertedValue";
  stack3['disabledBinding'] = stack4;
  stack4 = "controller.inputFields.SMTPPassword.value";
  stack3['valueBinding'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n\n      <div ");
  stack1 = {};
  stack2 = ":control-group controller.inputFields.retypeSMTPPassword.errorMsg:error";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.retypeSMTPPassword.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "password";
  stack3['type'] = stack4;
  stack4 = "controller.inputFields.SMTPUseAuthentication.invertedValue";
  stack3['disabledBinding'] = stack4;
  stack4 = "controller.inputFields.retypeSMTPPassword.value";
  stack3['valueBinding'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n\n        <div class=\"controls error-msg\">\n          ");
  stack1 = depth0;
  stack2 = "controller.inputFields.retypeSMTPPassword.errorMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n\n      <div class=\"control-group\">\n        <label class=\"control-label\" for=\"inputSMTPSTARTTLS\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.SMTPSTARTTLS.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.Checkbox";
  stack3 = {};
  stack4 = "controller.inputFields.SMTPUseAuthentication.invertedValue";
  stack3['disabledBinding'] = stack4;
  stack4 = "controller.inputFields.SMTPSTARTTLS.value";
  stack3['checkedBinding'] = stack4;
  stack4 = "inputSMTPSTARTTLS";
  stack3['id'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <div class=\"control-group\">\n        <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.version.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.Select";
  stack3 = {};
  stack4 = "controller.inputFields.version.value";
  stack3['selectionBinding'] = stack4;
  stack4 = "controller.SNMPVersions";
  stack3['contentBinding'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n\n      <div class=\"control-group\">\n        <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.OIDs.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "controller.inputFields.OIDs.value";
  stack3['valueBinding'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n\n      <div class=\"control-group\">\n        <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.community.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "controller.inputFields.community.value";
  stack3['valueBinding'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n\n      <div ");
  stack1 = {};
  stack2 = ":control-group controller.inputFields.port.errorMsg:error";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.port.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "controller.inputFields.port.value";
  stack3['valueBinding'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n\n        <div class=\"controls error-msg\">\n          ");
  stack1 = depth0;
  stack2 = "controller.inputFields.port.errorMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n      <div class=\"control-group\">\n        <label class=\"control-label\" for=\"inputGroups\">");
  stack1 = depth0;
  stack2 = "customProperty.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n        <div class=\"controls\">\n          ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "customProperty.value";
  stack3['valueBinding'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "customProperty";
  stack3 = depth0;
  stack4 = "removeCustomPropertyHandler";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"btn btn-small\"><span class=\"icon-minus-sign\"></span></a>\n        </div>\n      </div>\n    ");
  return buffer;}

  data.buffer.push("\n\n<div id=\"create-edit-alert-notification\">\n\n  <form class=\"form-horizontal\">\n\n\n    ");
  data.buffer.push("\n    <div ");
  stack1 = {};
  stack2 = ":control-group controller.inputFields.name.errorMsg:error";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <label class=\"control-label\" for=\"inputName\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.name.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n      <div class=\"controls\">\n        ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "controller.inputFields.name.value";
  stack3['valueBinding'] = stack4;
  stack4 = "inputName";
  stack3['id'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n\n      <div class=\"controls error-msg\">\n        ");
  stack1 = depth0;
  stack2 = "controller.inputFields.name.errorMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    ");
  data.buffer.push("\n\n    ");
  data.buffer.push("\n    <div class=\"control-group\">\n      <label class=\"control-label\" for=\"inputGroups\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.groups.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n      <div class=\"controls\">\n\n        <div>\n          ");
  stack1 = depth0;
  stack2 = "Ember.RadioButton";
  stack3 = {};
  stack4 = "allGroups";
  stack3['name'] = stack4;
  stack4 = "controller.inputFields.allGroups.value";
  stack3['selectionBinding'] = stack4;
  stack4 = "all";
  stack3['value'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "common.all";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n        <div>\n          ");
  stack1 = depth0;
  stack2 = "Ember.RadioButton";
  stack3 = {};
  stack4 = "allGroups";
  stack3['name'] = stack4;
  stack4 = "controller.inputFields.allGroups.value";
  stack3['selectionBinding'] = stack4;
  stack4 = "custom";
  stack3['value'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "common.custom";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n          ");
  stack1 = depth0;
  stack2 = "view.groupsSelectView";
  stack3 = {};
  stack4 = "true";
  stack3['multiple'] = stack4;
  stack4 = "inputGroups";
  stack3['id'] = stack4;
  stack4 = "input-large";
  stack3['class'] = stack4;
  stack4 = "controller.inputFields.allGroups.isAll";
  stack3['disabledBinding'] = stack4;
  stack4 = "controller.inputFields.groups.value";
  stack3['selectionBinding'] = stack4;
  stack4 = "controller.allAlertGroups";
  stack3['contentBinding'] = stack4;
  stack4 = "content.displayName";
  stack3['optionLabelPath'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "selectAllGroups";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "controller.inputFields.allGroups.isAll:disabled view.allGroupsSelected:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_notifications_popup.selectAll";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a> |\n          <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "clearAllGroups";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "controller.inputFields.allGroups.isAll:disabled view.noneGroupsSelected:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_notifications_popup.clearAll";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n      </div>\n    </div>\n    ");
  data.buffer.push("\n\n    ");
  data.buffer.push("\n    <div class=\"control-group\">\n      <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.severityFilter.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n      <div class=\"controls\">\n        ");
  stack1 = depth0;
  stack2 = "view.severitySelectView";
  stack3 = {};
  stack4 = "true";
  stack3['multiple'] = stack4;
  stack4 = "inputSeverity";
  stack3['id'] = stack4;
  stack4 = "input-large";
  stack3['class'] = stack4;
  stack4 = "controller.inputFields.severityFilter.value";
  stack3['selectionBinding'] = stack4;
  stack4 = "controller.severities";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "selectAllSeverity";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.allSeveritySelected:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_notifications_popup.selectAll";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a> |\n        <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "clearAllSeverity";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.noneSeveritySelected:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_notifications_popup.clearAll";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n      </div>\n    </div>\n    ");
  data.buffer.push("\n\n    ");
  data.buffer.push("\n    <div class=\"control-group\">\n      <label class=\"control-label\" for=\"inputDescription\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.description.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n      <div class=\"controls\">\n        ");
  stack1 = depth0;
  stack2 = "Em.TextArea";
  stack3 = {};
  stack4 = "controller.inputFields.description.value";
  stack3['valueBinding'] = stack4;
  stack4 = "inputDescription";
  stack3['id'] = stack4;
  stack4 = "4";
  stack3['rows'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    ");
  data.buffer.push("\n\n    ");
  data.buffer.push("\n    <div class=\"control-group\">\n      <label class=\"control-label\" for=\"inputMethod\">");
  stack1 = depth0;
  stack2 = "controller.inputFields.method.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n\n      <div class=\"controls\">\n        ");
  stack1 = depth0;
  stack2 = "Em.Select";
  stack3 = {};
  stack4 = "controller.methods";
  stack3['contentBinding'] = stack4;
  stack4 = "controller.inputFields.method.value";
  stack3['selectionBinding'] = stack4;
  stack4 = "inputMethod";
  stack3['id'] = stack4;
  stack4 = "input-xlarge";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    ");
  data.buffer.push("\n\n    ");
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "view.isEmailMethodSelected";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  data.buffer.push("\n\n    ");
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "controller.inputFields.customProperties";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "customProperty";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  data.buffer.push("\n\n    <a href=\"#\" class=\"add-custom-property\" ");
  stack1 = depth0;
  stack2 = "addCustomPropertyHandler";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step7.config.addProperty";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " ...</a>\n\n  </form>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/create_new_alert_group", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.parentView.warningMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  return buffer;}

function program3(depth0,data) {


  data.buffer.push("\n        &nbsp;\n    ");}

  data.buffer.push("\n\n<div class=\"new-config-group-div\" id=\"create-new-alert-group\">\n  <div class=\"text-warning\">\n    ");
  stack1 = depth0;
  stack2 = "view.parentView.warningMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  <table>\n    <tr>\n       <td class=\"title\">");
  stack1 = depth0;
  stack2 = "common.name";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": &nbsp;</td>\n       <td>");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "255";
  stack3['maxlength'] = stack4;
  stack4 = "alertGroupName";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n    </tr>\n  </table>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/definition_details", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n\n  <div id=\"alert-definition-details\">\n  <div class=\"row-fluid\">\n\n    ");
  data.buffer.push("\n    <div class=\"span9\">\n      ");
  data.buffer.push("\n      <div class=\"definition-name row-fluid\">\n        ");
  stack1 = depth0;
  stack2 = "controller.editing.label.isEditing";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "controller.editing.label.isEditing";
  stack3 = helpers.unless;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n      ");
  data.buffer.push("\n\n      <div><a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i\n          class=\"icon-arrow-left\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></div>\n\n      ");
  data.buffer.push("\n      <div class=\"box\">\n        <div class=\"box-header\">\n          <div class=\"pull-left\">\n            <h4>");
  stack1 = depth0;
  stack2 = "common.configuration";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h4>\n          </div>\n          <div class=\"pull-right span5 row-fluid\" style=\"padding:0 10px;\">\n            ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n        </div>\n        ");
  stack1 = depth0;
  stack2 = "App.AlertDefinitionConfigsView";
  stack3 = {};
  stack4 = "view.controller.content";
  stack3['contentBinding'] = stack4;
  stack4 = "view.controller.content.type";
  stack3['alertDefinitionTypeBinding'] = stack4;
  stack4 = false;
  stack3['canEdit'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "App.router.mainAlertDefinitionConfigsController.canEdit";
  stack3 = helpers['if'];
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n      ");
  data.buffer.push("\n    </div>\n    ");
  data.buffer.push("\n\n    ");
  data.buffer.push("\n    <div class=\"span3 right-column\">\n      <div class=\"status\">\n        ");
  stack1 = depth0;
  stack2 = "controller.content.status";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n      <div class=\"properties-list\">\n        <div class=\"row-fluid\">\n          <div class=\"span4 property-name\">");
  stack1 = depth0;
  stack2 = "alerts.table.state";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</div>\n          <div class=\"span8\">\n            ");
  stack1 = depth0;
  stack2 = "controller.content.enabled";
  stack3 = helpers['if'];
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(19, program19, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n        </div>\n        ");
  stack1 = depth0;
  stack2 = "controller.content.serviceDisplayName";
  stack3 = helpers['if'];
  tmp1 = self.program(24, program24, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "controller.content.componentNameFormatted";
  stack3 = helpers['if'];
  tmp1 = self.program(26, program26, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        <div class=\"row-fluid\">\n          <div class=\"span4 property-name\">");
  stack1 = depth0;
  stack2 = "common.type";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</div>\n          <div class=\"span8\">\n            <span ");
  stack1 = {};
  stack2 = ":type-icon  controller.content.typeIconClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span> ");
  stack1 = depth0;
  stack2 = "controller.content.type";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </div>\n        </div>\n        ");
  stack1 = depth0;
  stack2 = "controller.groupsList.length";
  stack3 = helpers['if'];
  tmp1 = self.program(28, program28, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        <div class=\"row-fluid\">\n          <div class=\"span4 property-name\">");
  stack1 = depth0;
  stack2 = "alerts.table.header.lastTrigger";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</div>\n          <div class=\"span8\">");
  stack1 = depth0;
  stack2 = "controller.content.lastTriggeredFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n        </div>\n      </div>\n    </div>\n  </div>\n    ");
  data.buffer.push("\n\n    ");
  data.buffer.push("\n  <div class=\"definition-details-block\">\n  <span>\n    <strong>");
  stack1 = depth0;
  stack2 = "alerts.definition.details.instances";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</strong>\n  </span>\n    <hr>\n    <div>\n\n      ");
  stack1 = depth0;
  stack2 = "App.router.mainAlertInstancesController.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(31, program31, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(40, program40, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n    </div>\n  </div>\n    ");
  data.buffer.push("\n\n  </div>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n          <div ");
  stack1 = {};
  stack2 = "controller.editing.label.isError:error :span5 :name-text-field :control-group";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "controller.editing.label.value";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </div>\n          <div class=\"edit-buttons\">\n            <button ");
  stack1 = depth0;
  stack2 = "controller.editing.label";
  stack3 = depth0;
  stack4 = "cancelEdit";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                class=\"btn cancel-button\">");
  stack1 = depth0;
  stack2 = "common.cancel";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n            <button ");
  stack1 = {};
  stack2 = "controller.editing.label.isError";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "controller.editing.label";
  stack3 = depth0;
  stack4 = "saveEdit";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                class=\"btn btn-primary save-button\">");
  stack1 = depth0;
  stack2 = "common.save";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </button>\n          </div>\n        ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <span>\n            ");
  stack1 = depth0;
  stack2 = "controller.content.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </span>\n        ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <a ");
  stack1 = depth0;
  stack2 = "controller.editing.label";
  stack3 = depth0;
  stack4 = "edit";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"edit-description-button\"><i\n                class=\"icon-pencil\"></i></a>\n          ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "App.router.mainAlertDefinitionConfigsController.canEdit";
  stack3 = helpers.unless;
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                <a ");
  stack1 = depth0;
  stack2 = "editConfigs";
  stack3 = {};
  stack4 = "App.router.mainAlertDefinitionConfigsController";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                    class=\"pull-right edit-link\">\n                  ");
  stack1 = depth0;
  stack2 = "common.edit";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              ");
  return buffer;}

function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          <div class=\"edit-buttons\">\n            <button ");
  stack1 = depth0;
  stack2 = "cancelEditConfigs";
  stack3 = {};
  stack4 = "App.router.mainAlertDefinitionConfigsController";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                class=\"btn\">");
  stack1 = depth0;
  stack2 = "common.cancel";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n            <button ");
  stack1 = {};
  stack2 = "App.router.mainAlertDefinitionConfigsController.hasErrors";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "saveConfigs";
  stack3 = {};
  stack4 = "App.router.mainAlertDefinitionConfigsController";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                class=\"btn btn-primary\">");
  stack1 = depth0;
  stack2 = "common.save";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </button>\n          </div>\n        ");
  return buffer;}

function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "NON_ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "controller.content";
  stack3 = depth0;
  stack4 = "toggleState";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "controller.content.enabled:alert-definition-enable:alert-definition-disable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <span\n                      class=\"enable-disable-button\" ");
  stack1 = {};
  stack2 = "alerts.table.state.enabled.tooltip";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n                    ");
  stack1 = depth0;
  stack2 = "view.enabledDisplay";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  </span>\n                </a>\n              ");
  return buffer;}

function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <span ");
  stack1 = {};
  stack2 = "controller.content.enabled:alert-definition-enable:alert-definition-disable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  ");
  stack1 = depth0;
  stack2 = "alerts.table.state.enabled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                </span>\n              ");
  return buffer;}

function program19(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(20, program20, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "NON_ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(22, program22, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program20(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "controller.content";
  stack3 = depth0;
  stack4 = "toggleState";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "controller.content.enabled:alert-definition-enable:alert-definition-disable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <span\n                      class=\"enable-disable-button\" ");
  stack1 = {};
  stack2 = "alerts.table.state.disabled.tooltip";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n                    ");
  stack1 = depth0;
  stack2 = "view.disabledDisplay";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  </span>\n                </a>\n              ");
  return buffer;}

function program22(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <span ");
  stack1 = {};
  stack2 = "controller.content.enabled:alert-definition-enable:alert-definition-disable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  ");
  stack1 = depth0;
  stack2 = "alerts.table.state.disabled";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                </span>\n              ");
  return buffer;}

function program24(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <div class=\"row-fluid\">\n            <div class=\"span4 property-name\">");
  stack1 = depth0;
  stack2 = "common.service";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</div>\n            <div class=\"span8\"><span>");
  stack1 = depth0;
  stack2 = "controller.content.serviceDisplayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span></div>\n          </div>\n        ");
  return buffer;}

function program26(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <div class=\"row-fluid\">\n            <div class=\"span4 property-name\">");
  stack1 = depth0;
  stack2 = "common.component";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</div>\n            <div class=\"span8\">");
  stack1 = depth0;
  stack2 = "controller.content.componentNameFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n          </div>\n        ");
  return buffer;}

function program28(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n          <div class=\"row-fluid\">\n            <div class=\"span4 property-name\">");
  stack1 = depth0;
  stack2 = "alerts.definition.details.groups";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</div>\n            <div class=\"span8\">\n              <ul>\n                ");
  stack1 = depth0;
  stack2 = "controller.groupsList";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "groupName";
  stack7 = helpers.each;
  tmp1 = self.program(29, program29, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </ul>\n            </div>\n          </div>\n        ");
  return buffer;}
function program29(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  <li>");
  stack1 = depth0;
  stack2 = "groupName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</li>\n                ");
  return buffer;}

function program31(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n\n        <table class=\"table advanced-header-table table-bordered table-striped alerts-table\" id=\"alert-instances-table\">\n          <thead>\n          <tr>\n            <th class=\"first\">");
  stack1 = depth0;
  stack2 = "alerts.definition.details.serviceHost";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n            <th>");
  stack1 = depth0;
  stack2 = "common.status";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n            <th>");
  stack1 = depth0;
  stack2 = "alerts.definition.details.24-hour";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n            <th>");
  stack1 = depth0;
  stack2 = "alerts.table.header.check.response";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n          </tr>\n          </thead>\n          <tbody>\n            ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = helpers['if'];
  tmp1 = self.program(32, program32, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(38, program38, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </tbody>\n        </table>\n\n        <div class=\"page-bar\">\n          <div class=\"items-on-page\">\n            <label>");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.rowsPerPageSelectView";
  stack3 = {};
  stack4 = "view.displayLength";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n          </div>\n          <div class=\"info\">");
  stack1 = depth0;
  stack2 = "view.paginationInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n          <div class=\"paging_two_button\">\n            <a ");
  stack1 = {};
  stack2 = "view.paginationLeftClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "previousPage";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i\n                class=\"icon-arrow-left\"></i></a>\n            <a ");
  stack1 = {};
  stack2 = "view.paginationRightClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "nextPage";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i\n                class=\"icon-arrow-right\"></i></a>\n          </div>\n        </div>\n\n      ");
  return buffer;}
function program32(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "instance";
  stack7 = helpers.each;
  tmp1 = self.program(33, program33, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program33(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "view.instanceTableRow";
  stack3 = helpers.view;
  tmp1 = self.program(34, program34, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  return buffer;}
function program34(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                <td class=\"first\">\n                  <a ");
  stack1 = depth0;
  stack2 = "instance.service";
  stack3 = depth0;
  stack4 = "goToService";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                      href=\"#\">");
  stack1 = depth0;
  stack2 = "instance.service.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                  ");
  stack1 = depth0;
  stack2 = "instance.hostName";
  stack3 = helpers['if'];
  tmp1 = self.program(35, program35, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </td>\n                <td>");
  stack1 = depth0;
  stack2 = "instance.status";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  <time class=\"timeago\"\n                        rel=\"tooltip\"  ");
  stack1 = {};
  stack2 = "instance.statusChangedAndLastCheckedFormatted";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "instance.lastTriggeredForFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</time>\n                </td>\n                <td>");
  stack1 = depth0;
  stack2 = "view.parentView.lastDayCount";
  stack3 = {};
  stack4 = "instance.hostName";
  stack3['hostNameBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n                <td><span class=\"alert-text\" rel=\"tooltip\" ");
  stack1 = {};
  stack2 = "instance.text";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                          class=\"alert-text\">");
  stack1 = depth0;
  stack2 = "instance.text";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span></td>\n                ");
  return buffer;}
function program35(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "instance.service.displayName";
  stack3 = helpers['if'];
  tmp1 = self.program(36, program36, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    <a ");
  stack1 = depth0;
  stack2 = "instance.host";
  stack3 = depth0;
  stack4 = "goToHostAlerts";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n                      ");
  stack1 = depth0;
  stack2 = "instance.hostName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                    </a>\n                  ");
  return buffer;}
function program36(depth0,data) {


  data.buffer.push("\n                      &nbsp;/&nbsp;\n                    ");}

function program38(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <tr>\n              <td class=\"first\" colspan=\"5\">\n                ");
  stack1 = depth0;
  stack2 = "alerts.definition.details.noAlerts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </td>\n            </tr>\n            ");
  return buffer;}

function program40(depth0,data) {


  data.buffer.push("\n        <div class=\"spinner\"></div>\n      ");}

function program42(depth0,data) {


  data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(42, program42, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/manage_alert_groups_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div class=\"row-fluid  manage-configuration-group-content\">\n    <div class=\"span12\">\n      <div class=\"row-fluid\">\n        <div class=\"span4\">\n          <span>&nbsp;</span>\n          ");
  stack1 = depth0;
  stack2 = "Em.Select";
  stack3 = {};
  stack4 = "alertGroups";
  stack3['contentBinding'] = stack4;
  stack4 = "content.label";
  stack3['optionLabelPath'] = stack4;
  stack4 = "view.selectedAlertGroup";
  stack3['selectionBinding'] = stack4;
  stack4 = "multiple";
  stack3['multiple'] = stack4;
  stack4 = "group-select select-group-box";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          <div class=\"btn-toolbar pull-right\">\n            <button rel=\"button-info\" class=\"btn add-group-button\" ");
  stack1 = {};
  stack2 = "view.addButtonTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "addAlertGroup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-plus\"></i></button>\n            <a rel=\"button-info\" class=\"btn remove-group-button\" ");
  stack1 = {};
  stack2 = "view.removeButtonTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = "controller.isRemoveButtonDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "confirmDelete";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-minus\"></i></a>\n            <div class=\"btn-group\">\n              <button class=\"btn dropdown-toggle actions-group-button\" data-toggle=\"dropdown\">\n                <i class=\"icon-cog\"></i>&nbsp;<span class=\"caret\"></span>\n              </button>\n              <ul class=\"dropdown-menu\">\n                <li ");
  stack1 = {};
  stack2 = "controller.isRenameButtonDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <a href=\"\" rel=\"button-info-dropdown\" ");
  stack1 = {};
  stack2 = "view.renameButtonTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "renameAlertGroup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_groups_popup.rename";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                </li>\n                <li ");
  stack1 = {};
  stack2 = "controller.isDuplicateButtonDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <a href=\"\" rel=\"button-info-dropdown\" ");
  stack1 = {};
  stack2 = "view.duplicateButtonTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "duplicateAlertGroup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_groups_popup.duplicate";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                </li>\n              </ul>\n            </div>\n          </div>\n        </div>\n          <div class=\"span8\">\n            <span>&nbsp;</span>\n            <div class=\"row-fluid\">\n              <div class=\"span12 pull-right\">\n                ");
  stack1 = depth0;
  stack2 = "Em.Select";
  stack3 = {};
  stack4 = "selectedAlertGroup.definitions";
  stack3['contentBinding'] = stack4;
  stack4 = "content.label";
  stack3['optionLabelPath'] = stack4;
  stack4 = "multiple";
  stack3['multiple'] = stack4;
  stack4 = "group-select select-definiton-box";
  stack3['class'] = stack4;
  stack4 = "selectedDefinitions";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </div>\n              <div class=\"button-group pull-right\">\n                  <a rel=\"button-info\" ");
  stack1 = {};
  stack2 = "view.addDefinitionTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = ":btn selectedAlertGroup.isAddDefinitionsDisabled:disabled :add-definition";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "addDefinitions";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-plus\"></i></a>\n                  <a rel=\"button-info\" ");
  stack1 = {};
  stack2 = "view.removeDefinitionTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = ":btn isDeleteDefinitionsDisabled:disabled :remove-definition";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "deleteDefinitions";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-minus\"></i></a>\n              </div>\n            </div>\n\n            <div class=\"clearfix\"></div>\n            <div class=\"row-fluid notification-editable-list\">\n              <div class=\"span4 title\"><span class=\"icon-envelope-alt\"></span>&nbsp;");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_groups_popup.notifications";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n            </div>\n            <div class=\"row-fluid notification-editable-list\">\n              <div class=\"span11\">\n                ");
  stack1 = depth0;
  stack2 = "App.EditableList";
  stack3 = {};
  stack4 = "selectedAlertGroup.notifications";
  stack3['itemsBinding'] = stack4;
  stack4 = "controller.alertNotifications";
  stack3['resourcesBinding'] = stack4;
  stack4 = "selectedAlertGroup.displayName";
  stack3['nameBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </div>\n            </div>\n\n          </div>\n          <div class=\"clearfix\"></div>\n          <div class=\"row-fluid\">\n            <div class=\"span12 text-error\" id=\"manage-config-group-error-div\">\n              ");
  stack1 = depth0;
  stack2 = "controller.errorMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </div>\n          </div>\n      </div>\n    </div>\n  </div>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "controller.errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  return buffer;}

function program4(depth0,data) {


  data.buffer.push("\n                 &nbsp;\n              ");}

function program6(depth0,data) {


  data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

  data.buffer.push("\n<div class=\"alert alert-info margin-bottom-5\">");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_groups_popup.notice";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n");
  stack1 = depth0;
  stack2 = "controller.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(6, program6, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/alerts/manage_alert_notifications_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div class=\"row-fluid manage-configuration-group-content\" id=\"manage-alert-notification-content\">\n    <div class=\"span12\">\n      <div class=\"row-fluid\">\n        <div class=\"span4 notification-list\">\n          <span>&nbsp;</span>\n          ");
  stack1 = depth0;
  stack2 = "Em.Select";
  stack3 = {};
  stack4 = "alertNotifications";
  stack3['contentBinding'] = stack4;
  stack4 = "content.name";
  stack3['optionLabelPath'] = stack4;
  stack4 = "multiple";
  stack3['multiple'] = stack4;
  stack4 = "group-select";
  stack3['class'] = stack4;
  stack4 = "view.selectedAlertNotification";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          <div class=\"btn-toolbar pull-right\">\n            <button rel=\"button-info\" class=\"btn add-notification-button\"\n              ");
  stack1 = {};
  stack2 = "alerts.actions.manage_alert_notifications_popup.addButton";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "addAlertNotification";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-plus\"></i></button>\n            <button rel=\"button-info\" class=\"btn remove-notification-button\"\n              ");
  stack1 = {};
  stack2 = "alerts.actions.manage_alert_notifications_popup.removeButton";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = {};
  stack2 = "view.isRemoveButtonDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "deleteAlertNotification";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-minus\"></i></button>\n            <div class=\"btn-group notification-actions-button\">\n              <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n                <i class=\"icon-cog\"></i>&nbsp;<span class=\"caret\"></span>\n              </button>\n              <ul class=\"dropdown-menu\">\n                <li ");
  stack1 = {};
  stack2 = "view.isEditButtonDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <a href=\"\" rel=\"button-info-dropdown\"\n                    ");
  stack1 = {};
  stack2 = "alerts.actions.manage_alert_notifications_popup.editButton";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + "\n                    ");
  stack1 = depth0;
  stack2 = "editAlertNotification";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.edit";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                </li>\n                <li ");
  stack1 = {};
  stack2 = "view.isDuplicateButtonDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <a href=\"\" rel=\"button-info-dropdown\"\n                    ");
  stack1 = {};
  stack2 = "alerts.actions.manage_alert_notifications_popup.duplicateButton";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + "\n                    ");
  stack1 = depth0;
  stack2 = "duplicateAlertNotification";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.duplicate";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                </li>\n              </ul>\n            </div>\n          </div>\n        </div>\n        <div class=\"span8 notification-info\">\n          <span>&nbsp;</span>\n\n          <div class=\"row-fluid\">\n            <div class=\"span12 pull-right\">\n              ");
  stack1 = depth0;
  stack2 = "alertNotifications.length";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(19, program19, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </div>\n          </div>\n        </div>\n        <div class=\"clearfix\"></div>\n        <div class=\"row-fluid\">\n          <div class=\"span12 text-error\" id=\"manage-alert-notifications-error\">\n            ");
  stack1 = depth0;
  stack2 = "controller.errorMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(21, program21, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(23, program23, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n        </div>\n      </div>\n    </div>\n  </div>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "selectedAlertNotification";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  <div class=\"row-fluid\">\n                    <div class=\"span3 input-label\">");
  stack1 = depth0;
  stack2 = "common.name";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n                    <div class=\"span9\">");
  stack1 = depth0;
  stack2 = "selectedAlertNotification.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                  </div>\n                  <div class=\"row-fluid\">\n                    <div class=\"span3 input-label\">");
  stack1 = depth0;
  stack2 = "common.groups";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n                    <div class=\"span9\">\n                      ");
  stack1 = depth0;
  stack2 = "selectedAlertNotification.global";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(6, program6, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    </div>\n                  </div>\n                  <div class=\"row-fluid\">\n                    <div class=\"span3 input-label\">");
  stack1 = depth0;
  stack2 = "common.severity";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n                    <div class=\"span9\">");
  stack1 = depth0;
  stack2 = "view.severities";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                  </div>\n                  <div class=\"row-fluid\">\n                    <div class=\"span3 input-label\">");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_notifications_popup.method";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n                    <div class=\"span9\">");
  stack1 = depth0;
  stack2 = "selectedAlertNotification.type";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                  </div>\n                  ");
  stack1 = depth0;
  stack2 = "view.showEmailDetails";
  stack3 = helpers['if'];
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "view.showSNMPDetails";
  stack3 = helpers['if'];
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  <div class=\"row-fluid\">\n                    <div class=\"span3 input-label\">");
  stack1 = depth0;
  stack2 = "common.description";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n                    ");
  stack1 = depth0;
  stack2 = "selectedAlertNotification.description";
  stack3 = helpers['if'];
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(17, program17, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  </div>\n                ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                        ");
  stack1 = depth0;
  stack2 = "common.all";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                      ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                        ");
  stack1 = depth0;
  stack2 = "view.selectedAlertNotificationGroups";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(9, program9, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                      ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                          ");
  stack1 = depth0;
  stack2 = "view.selectedAlertNotificationGroups";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                        ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                          ");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_notifications_popup.noGroup";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                        ");
  return buffer;}

function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    <div class=\"row-fluid\">\n                      <div class=\"span3 input-label\">");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_notifications_popup.email";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n                      <div class=\"span9\">");
  stack1 = depth0;
  stack2 = "view.email";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                    </div>\n                  ");
  return buffer;}

function program13(depth0,data) {


  data.buffer.push("\n                  ");}

function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                      <div class=\"span9 notification-description\">");
  stack1 = depth0;
  stack2 = "selectedAlertNotification.description";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                    ");
  return buffer;}

function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                      <div class=\"span9 notification-description\">");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_notifications_popup.noDescription";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n                    ");
  return buffer;}

function program19(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <div class=\"alert alert-info\">\n                  ");
  stack1 = depth0;
  stack2 = "alerts.actions.manage_alert_notifications_popup.noAlertNotification";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                </div>\n              ");
  return buffer;}

function program21(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "controller.errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program23(depth0,data) {


  data.buffer.push("\n              &nbsp;\n            ");}

function program25(depth0,data) {


  data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

  data.buffer.push("\n<div class=\"alert alert-info margin-bottom-5\">");
  stack1 = depth0;
  stack2 = "alerts.actions.manageNotifications.info";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n");
  stack1 = depth0;
  stack2 = "controller.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(25, program25, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/charts", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1));
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/charts/heatmap", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n               <li class=\"dropdown-submenu\">\n                <a tabindex=\"-1\" >");
  stack1 = depth0;
  stack2 = "category.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                <ul class=\"dropdown-menu\">\n                  ");
  stack1 = depth0;
  stack2 = "category.items";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "metric";
  stack7 = helpers.each;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </ul>\n               </li>\n            ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                    <li>\n                      <a tabindex=\"-1\" ");
  stack1 = depth0;
  stack2 = "metric";
  stack3 = depth0;
  stack4 = "showHeatMapMetric";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "metric.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                    </li>\n                  ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n					<table class=\"legend\">\n					  ");
  stack1 = depth0;
  stack2 = "controller.selectedMetric.slotDefinitions";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "slot";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n					</table>\n	        ");
  stack1 = depth0;
  stack2 = "common.maximum";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":\n	        <div id=\"inputMaximum\" class=\"control-group\">\n	          ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "text";
  stack3['type'] = stack4;
  stack4 = "8";
  stack3['maxlength'] = stack4;
  stack4 = "controller.inputMaximum";
  stack3['valueBinding'] = stack4;
  stack4 = "span6";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n	          ");
  stack1 = depth0;
  stack2 = "controller.selectedMetric.units";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n	        </div>\n        ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n					    <tr>\n					      <td>\n					        <div class=\"tile\" ");
  stack1 = {};
  stack2 = "slot.cssStyle";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n					      </td>\n					      <td>");
  stack1 = depth0;
  stack2 = "slot.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n					    </tr>\n					  ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n				    <div ");
  stack1 = {};
  stack2 = "controller.rackClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n				      ");
  stack1 = depth0;
  stack2 = "App.MainChartsHeatmapRackView";
  stack3 = {};
  stack4 = "rack";
  stack3['rackBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n				    </div>\n				  ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"heatmap\">\n\n  <div class=\"container-fluid\">\n	  <div class=\"row-fluid\">\n	    <div class=\"span2 legend-column\">\n		     <div class=\"btn-group\">\n				  <button class=\"btn heatmap-select-metric-btn\">");
  stack1 = depth0;
  stack2 = "charts.heatmap.selectMetric";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n				  <button class=\"btn dropdown-toggle heatmap-toggle-metrics-btn\" data-toggle=\"dropdown\">\n				    <span class=\"caret\"></span>\n				  </button>\n				  <ul class=\"dropdown-menu\">\n				    ");
  stack1 = depth0;
  stack2 = "controller.allMetrics";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "category";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n				  </ul>\n				</div>\n        ");
  stack1 = depth0;
  stack2 = "controller.selectedMetric";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n	    </div>\n	    <div class=\"span10 heatmap-content\">\n	      <h4 id=\"heatmap-metric-loading\">\n	        <span id=\"heatmap-metric-title\">");
  stack1 = depth0;
  stack2 = "controller.selectedMetric.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n	      </h4>\n	      <div class=\"row-fluid\">\n				  ");
  stack1 = depth0;
  stack2 = "controller.racks";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "rack";
  stack7 = helpers.each;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n			  </div>\n			  ");
  stack1 = depth0;
  stack2 = "App.MainChartsHeatmapHostDetailView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n	    </div>\n	  </div>\n  </div>\n\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/charts/heatmap/heatmap_host", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div ");
  stack1 = {};
  stack2 = "view.hostClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.hostTemperatureStyle";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  <a href=\"#\" class=\"heatmap-host\" ");
  stack1 = depth0;
  stack2 = "view.hostModelLink";
  stack3 = depth0;
  stack4 = "showDetails";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></a>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/charts/heatmap/heatmap_host_detail", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.details.metricName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.details.metricValue";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "<br/>\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "common.disk";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.details.diskUsage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "% <br/>\n");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "common.cpu";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.details.cpuUsage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "%<br/>\n");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "common.memory";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.details.memoryUsage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "%<br/>\n");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "common.components";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.details.hostComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "<br/>\n");
  return buffer;}

  data.buffer.push("\n\n<b>");
  stack1 = depth0;
  stack2 = "view.details.publicHostName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</b><br/>\n");
  stack1 = depth0;
  stack2 = "common.os";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.details.osType";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "<br/>\n");
  stack1 = depth0;
  stack2 = "common.ipAddress";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.details.ip";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "<br/>\n");
  stack1 = depth0;
  stack2 = "view.details.metricName";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.details.diskUsage";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.details.cpuUsage";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.details.memoryUsage";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.details.hostComponents";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/charts/heatmap/heatmap_rack", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <div ");
  stack1 = {};
  stack2 = "view.hostCssStyle";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "App.MainChartsHeatmapHostView";
  stack3 = {};
  stack4 = "this";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n    ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"rackHeader\">\n  <div class=\"statusName clearfix\">\n    <div ");
  stack1 = {};
  stack2 = "view.statusIndicator rack.status";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n    <div class=\"rackName\">");
  stack1 = depth0;
  stack2 = "rack.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n  </div>\n</div>\n<!--<div class=\"hostsSummary clearfix\">-->\n  <!--<div class=\"textBlock\">Total hosts: ");
  data.buffer.push("</div>-->\n  <!--");
  data.buffer.push("-->\n    <!--<div class=\"textBlock\">-->\n      <!--<div class=\"indicatorG\"></div>-->\n      <!--Hosts in live state: ");
  data.buffer.push("-->\n    <!--</div>-->\n    <!--<div class=\"textBlock\">-->\n      <!--<div class=\"indicatorY\"></div>-->\n      <!--Hosts in critical state: ");
  data.buffer.push("-->\n    <!--</div>-->\n    <!--<div class=\"textBlock\">-->\n      <!--<div class=\"indicatorR\"></div>-->\n      <!--Hosts in dead state: ");
  data.buffer.push("-->\n    <!--</div>-->\n  <!--");
  data.buffer.push("-->\n<!--</div>-->\n<div class=\"isActive hosts clearfix\">\n  <div ");
  stack1 = {};
  stack2 = "view.rack.isLoaded::hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    ");
  stack1 = depth0;
  stack2 = "view.hosts";
  stack3 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  <div ");
  stack1 = {};
  stack2 = "view.rack.isLoaded:hidden :spinner";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/charts/linear_time", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div id=\"");
  stack1 = depth0;
  stack2 = "view.id";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "-title\" class=\"chart-title\">");
  stack1 = depth0;
  stack2 = "view.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n  ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"screensaver chart-container\" ");
  stack1 = {};
  stack2 = "view.isReady:hide";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " >\n  <div id=\"");
  stack1 = depth0;
  stack2 = "view.id";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "-title\" class=\"chart-title\">");
  stack1 = depth0;
  stack2 = "view.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n</div>\n<div  id=\"");
  stack1 = depth0;
  stack2 = "view.id";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "-container\" class=\"chart-container hide\" ");
  stack1 = {};
  stack2 = "view.isReady:show";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "  rel=\"ZoomInTooltip\" title=\"Click to zoom\">\n  <div id=\"");
  stack1 = depth0;
  stack2 = "view.id";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "-yaxis\" class=\"chart-y-axis\" ");
  stack1 = depth0;
  stack2 = "showGraphInPopup";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n  <div id=\"");
  stack1 = depth0;
  stack2 = "view.id";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "-xaxis\" class=\"chart-x-axis\" ");
  stack1 = depth0;
  stack2 = "showGraphInPopup";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n  <div id=\"");
  stack1 = depth0;
  stack2 = "view.id";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "-legend\" class=\"chart-legend\"></div>\n  <div id=\"");
  stack1 = depth0;
  stack2 = "view.id";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "-chart\" class=\"chart\"  ");
  stack1 = depth0;
  stack2 = "showGraphInPopup";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n  <div id=\"");
  stack1 = depth0;
  stack2 = "view.id";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "-timeline\" class=\"timeline\" ");
  stack1 = depth0;
  stack2 = "showGraphInPopup";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n  ");
  stack1 = depth0;
  stack2 = "view.noTitleUnderGraph";
  stack3 = helpers.unless;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n</div>\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/dashboard", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.NavItemView";
  stack3 = {};
  stack4 = "category.name";
  stack3['itemBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n            <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "category.url";
  stack3 = depth0;
  stack4 = "goToDashboardView";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " >");
  stack1 = depth0;
  stack2 = "category.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n          ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"row-fluid\">\n  <div class=\"services-menu well span2 service-menu-width\" style=\"padding: 8px 0\">\n    ");
  stack1 = depth0;
  stack2 = "App.MainServiceMenuView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "App.AllServicesActionView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n    <div class=\"summary-width span10\" id=\"dashboard-widgets-container\">\n      <ul class=\"nav nav-tabs\">\n        ");
  stack1 = depth0;
  stack2 = "view.categories";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "category";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </ul>\n\n      <!--show widgets, heatmaps or configs in the content-->\n      ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/dashboard/config_history", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.parentView.versionSort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "view.parentView.configGroupSort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "view.parentView.modifiedSort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "view.parentView.authorSort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "view.parentView.notesSort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(16, program16, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "item";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.ConfigVersionView";
  stack3 = {};
  stack4 = "item";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n            <td class=\"first\">\n              <span class=\"label label-info\">");
  stack1 = depth0;
  stack2 = "item.versionText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n              <a ");
  stack1 = depth0;
  stack2 = "item";
  stack3 = depth0;
  stack4 = "goToServiceConfigs";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"service-name\">\n                ");
  stack1 = depth0;
  stack2 = "item.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n              <i ");
  stack1 = {};
  stack2 = ":icon-refresh :restart-required-service item.isRestartRequired::hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                rel=\"Tooltip\" ");
  stack1 = {};
  stack2 = "dashboard.configHistory.table.restart.tooltip";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + "></i>\n            </td>\n            <td>");
  stack1 = depth0;
  stack2 = "item.configGroupName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "item.isCurrent";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </td>\n            <td>");
  stack1 = depth0;
  stack2 = "item.createdDate";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n            <td><span rel=\"Tooltip\" ");
  stack1 = {};
  stack2 = "item.author";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "item.authorFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span></td>\n            <td class=\"notes\">\n              ");
  stack1 = depth0;
  stack2 = "item.moreNotesExists";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(14, program14, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </td>\n          ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <span class=\"label label-success\" rel=\"Tooltip\"\n                ");
  stack1 = {};
  stack2 = "item.currentTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.current";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                </span>\n              ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "view.showLessNotes";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(12, program12, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  return buffer;}
function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "item.briefNotes";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  <a ");
  stack1 = depth0;
  stack2 = "toggleShowLessStatus";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"show-more-button\">>> More</a>\n                ");
  return buffer;}

function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "item.fullNotes";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  <a ");
  stack1 = depth0;
  stack2 = "toggleShowLessStatus";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"show-more-button\"><< Less</a>\n                ");
  return buffer;}

function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "item.briefNotes";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  return buffer;}

function program16(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <tr>\n          <td class=\"first\" colspan=\"5\">\n            ");
  stack1 = depth0;
  stack2 = "dashboard.configHistory.table.empty";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n          </td>\n        </tr>\n      ");
  return buffer;}

function program18(depth0,data) {


  data.buffer.push("\n      <tr><td colspan=\"5\"><div class=\"spinner\"></div></td></tr>\n    ");}

  data.buffer.push("\n\n<div id=\"config_history\">\n  <table class=\"table advanced-header-table table-bordered table-striped\" >\n    <thead>\n    ");
  stack1 = depth0;
  stack2 = "view.sortView";
  stack3 = {};
  stack4 = "label-row config-history-sort-row";
  stack3['classNames'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n    <tr class=\"filter-row config-history-filter-row\">\n      <th class=\"first\">");
  stack1 = depth0;
  stack2 = "view.serviceFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "view.configGroupFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "view.modifiedFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "view.authorFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "view.notesFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n    </tr>\n    </thead>\n    <tbody class=\"services-menu\">\n    ");
  stack1 = depth0;
  stack2 = "view.filteringComplete";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(18, program18, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </tbody>\n</table>\n\n  <div class=\"page-bar\">\n    <div class=\"filtered-info span4\">\n      <label>");
  stack1 = depth0;
  stack2 = "view.filteredContentInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " - <a ");
  stack1 = depth0;
  stack2 = "clearFilters";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              href=\"#\">");
  stack1 = depth0;
  stack2 = "tableView.filters.clearAllFilters";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></label>\n    </div>\n    <div class=\"items-on-page\">\n      <label>");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.rowsPerPageSelectView";
  stack3 = {};
  stack4 = "view.displayLength";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n    </div>\n    <div class=\"info\">");
  stack1 = depth0;
  stack2 = "view.paginationInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n    <div class=\"paging_two_button\">\n      ");
  stack1 = depth0;
  stack2 = "view.paginationLeft";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "view.paginationRight";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/dashboard/edit_widget_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n         ");
  stack1 = depth0;
  stack2 = "view.configPropertyObj.isGreenOrangeRed";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n       ");
  return buffer;}
function program2(depth0,data) {


  data.buffer.push("\n             <div class=\"progress span9\">\n                 <div class=\"bar bar-success\" style=\"width: 33%;\"></div>\n                 <div class=\"bar bar-warning\" style=\"width: 33%;\"></div>\n                 <div class=\"bar bar-danger\" style=\"width: 34%;\"></div>\n             </div>\n         ");}

function program4(depth0,data) {


  data.buffer.push("\n             <div class=\"progress span9\">\n                 <div class=\"bar bar-danger\" style=\"width: 33%;\"></div>\n                 <div class=\"bar bar-warning\" style=\"width: 33%;\"></div>\n                 <div class=\"bar bar-success\" style=\"width: 34%;\"></div>\n             </div>\n         ");}

function program6(depth0,data) {


  data.buffer.push("\n         <div class=\"span9\" id=\"slider-range\"></div>\n       ");}

  data.buffer.push("\n<form class=\"form-horizontal\" autocomplete=\"off\">\n    <div class=\"each-row\">\n        <div class=\"alert alert-info\">\n          ");
  stack1 = depth0;
  stack2 = "view.configPropertyObj.hintInfo";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </div>\n    </div>\n\n    <div class=\"row-fluid\" id= \"min-height-limit\">\n       ");
  stack1 = depth0;
  stack2 = "view.configPropertyObj.isIE9";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(6, program6, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n\n    <div class=\"row-fluid\">\n        <div id=\"slider-value1\" class=\"value-on-slider span2\">0</div>\n        <div id=\"slider-value2\" ");
  stack1 = {};
  stack2 = "view.configPropertyObj.isThresh1Error:slider-error :value-on-slider :span4";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "view.configPropertyObj.thresh1";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "view.configPropertyObj.errorMessage1";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n        </div>\n        <div id=\"slider-value3\" ");
  stack1 = {};
  stack2 = "view.configPropertyObj.isThresh2Error:slider-error :value-on-slider :span4";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "view.configPropertyObj.thresh2";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "view.configPropertyObj.errorMessage2";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n        </div>\n        <div id=\"slider-value4\" class=\"value-on-slider span2\">");
  stack1 = depth0;
  stack2 = "view.configPropertyObj.maxValue";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n\n</form>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/dashboard/plus_button_filter", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    <li class=\"add-widgets-dropdown-list\">\n      <ul>\n        ");
  stack1 = depth0;
  stack2 = "view.hiddenWidgets";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "widget";
  stack7 = helpers.each;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </ul>\n    </li>\n    <li>\n      <button class=\"btn\" ");
  stack1 = depth0;
  stack2 = "closeFilter";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.cancel";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n      <button class=\"btn btn-primary\" ");
  stack1 = depth0;
  stack2 = "applyFilter";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.apply";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n    </li>\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          <li>\n            <label class=\"checkbox\">\n              ");
  stack1 = depth0;
  stack2 = "view.widgetCheckbox";
  stack3 = {};
  stack4 = "widget.checked";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "widget.displayName";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </label>\n          </li>\n        ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <li>\n      <ul>\n      <li>");
  stack1 = depth0;
  stack2 = "dashboard.widgets.nothing";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n      </ul>\n    </li>\n    <li>\n      <button class=\"btn\" ");
  stack1 = depth0;
  stack2 = "closeFilter";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.cancel";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n    </li>\n  ");
  return buffer;}

  data.buffer.push("\n\n<ul class=\"dropdown-menu\" >\n  ");
  stack1 = depth0;
  stack2 = "view.hiddenWidgets.length";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</ul>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/dashboard/widgets", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n\n  <div id=\"widgets-options-menu\" class=\"dropdown\">\n    <a class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n        <span>Metric Actions</span>\n        <span class=\"caret\"></span>\n    </a>\n    <ul class=\"dropdown-menu\">\n      <li class=\"dropdown-submenu add-widgets-text\">\n        <a href=\"javascript:void(null);\"><i class=\"icon-plus\"></i> &nbsp; ");
  stack1 = depth0;
  stack2 = "common.add";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n        ");
  stack1 = depth0;
  stack2 = "view.plusButtonFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </li>\n      <li class=\"dropdown-submenu\">\n        <a href=\"javascript:void(null);\"><i class=\"icon-cog\"></i> &nbsp; ");
  stack1 = depth0;
  stack2 = "common.edit";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n        <ul class=\"dropdown-menu\">\n          <li>\n            <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "resetAllWidgets";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <i class=\"icon-refresh\"></i> &nbsp; ");
  stack1 = depth0;
  stack2 = "dashboard.button.reset";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n          ");
  stack1 = depth0;
  stack2 = "App.router.clusterController.isGangliaInstalled";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </ul>\n      </li>\n    </ul>\n  </div>\n\n  <div class=\"dashboard-widgets-box\">\n    <div id=\"dashboard-widgets\"  class=\"widgets-container\">\n      <div class=\"thumbnails row-fluid\" id=\"sortable\">\n        ");
  stack1 = depth0;
  stack2 = "view.visibleWidgets.length";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n    </div>\n  </div>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <li>\n              <a target=\"_blank\" ");
  stack1 = {};
  stack2 = "view.gangliaUrl";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                <i class=\"icon-share\"></i> &nbsp; ");
  stack1 = depth0;
  stack2 = "dashboard.button.gangliaLink";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n          ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.visibleWidgets";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "widgetClass";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <div ");
  stack1 = {};
  stack2 = "widgetClass.class";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "widgetClass";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </div>\n          ");
  return buffer;}

  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.isDataLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/dashboard/widgets/cluster_metrics", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div class=\"cluster-metrics\">\n  <ul>\n    <li class=\"thumbnail row\">\n      <a ");
  stack1 = {};
  stack2 = ":corner-icon :span1 view.parentView.isMoving:hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\" ");
  stack1 = depth0;
  stack2 = "deleteWidget";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <i class=\"icon-remove-sign icon-large\"></i>\n      </a>\n      <div class=\"caption span10\">");
  stack1 = depth0;
  stack2 = "view.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n\n      <div class=\"widget-content\" >\n        ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </li>\n  </ul>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/dashboard/widgets/hbase_links", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n                <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.activeMaster.host";
  stack3 = depth0;
  stack4 = "showDetails";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "dashboard.services.hbase.masterServer";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n              ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "dashboard.services.hbase.noMasterServer";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <a ");
  stack1 = {};
  stack2 = "view.hbaseMasterWebUrl";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " target=\"_blank\">");
  stack1 = depth0;
  stack2 = "dashboard.services.hbase.masterWebUI";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n              ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "services.service.summary.notAvailable";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "App.QuickViewLinks";
  stack3 = {};
  stack4 = "view.model";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(10, program10, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <div class=\"btn-group\">\n              <a class=\"btn btn-mini dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n                ");
  stack1 = depth0;
  stack2 = "common.more";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                <span class=\"caret\"></span>\n              </a>\n                <ul class=\"dropdown-menu\">\n                  ");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(19, program19, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </ul>\n            </div>\n          ");
  return buffer;}
function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "view.quickLinksArray";
  stack3 = helpers['if'];
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(16, program16, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n                      <!--there are multiple masters eg, HBase multiple masters or HDFS HA enabled-->\n                      ");
  stack1 = depth0;
  stack2 = "view.quickLinksArray";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "quickLinks";
  stack7 = helpers.each;
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    ");
  return buffer;}
function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                        <li class=\"dropdown-submenu\">\n                          <a href=\"javascript:void(null)\">");
  stack1 = depth0;
  stack2 = "quickLinks.publicHostNameLabel";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " &nbsp;</a>\n                          <ul class=\"dropdown-menu\">\n                            ");
  stack1 = depth0;
  stack2 = "quickLinks";
  stack3 = helpers.each;
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                          </ul>\n                        </li>\n                      ");
  return buffer;}
function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                              <li><a ");
  stack1 = {};
  stack2 = "url";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.linkTarget";
  stack1['target'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                            ");
  return buffer;}

function program16(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                      ");
  stack1 = depth0;
  stack2 = "view.quickLinks";
  stack3 = helpers.each;
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    ");
  return buffer;}
function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                        <li><a ");
  stack1 = {};
  stack2 = "url";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.linkTarget";
  stack1['target'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                      ");
  return buffer;}

function program19(depth0,data) {


  data.buffer.push("\n                    <div class=\"spinner\"></div>\n                  ");}

  data.buffer.push("\n\n<div class=\"links\">\n  <ul>\n    <li class=\"thumbnail row\">\n      <a ");
  stack1 = {};
  stack2 = ":corner-icon :span1 view.parentView.isMoving:hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\" ");
  stack1 = depth0;
  stack2 = "deleteWidget";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-remove-sign icon-large\"></i></a>\n      <div class=\"caption span10\"> ");
  stack1 = depth0;
  stack2 = "view.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n\n      <div class=\"widget-content\" >\n        <table>\n          <!--hbase master server-->\n          <tr>\n            <td>\n              ");
  stack1 = depth0;
  stack2 = "view.activeMaster";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </td>\n          </tr>\n          <!--region servers-->\n          <tr>\n              <td><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.component";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.model.regionServersTotal";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "dashboard.services.hbase.regionServers";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></td>\n          </tr>\n          <!--hbase master Web UI-->\n          <tr>\n            <td>\n              ");
  stack1 = depth0;
  stack2 = "view.activeMaster";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(7, program7, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </td>\n          </tr>\n        </table>\n      </div>\n\n      <div class=\"link-button\">\n        ");
  stack1 = depth0;
  stack2 = "view.model.quickLinks.length";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n\n    </li>\n  </ul>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/dashboard/widgets/hdfs_links", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n        <table>\n          <!--Active NameNode-->\n          <tr>\n            ");
  stack1 = depth0;
  stack2 = "view.isActiveNNValid";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </tr>\n          <!--Standby NameNodes-->\n          <tr>\n            ");
  stack1 = depth0;
  stack2 = "view.isStandbyNNValid";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(11, program11, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </tr>\n          <!--Data Nodes-->\n          <tr>\n            <td>\n              <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.component";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.model.dataNodesTotal";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.datanodes";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n            </td>\n          </tr>\n        </table>\n      ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n              <td><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.model.activeNameNode.host";
  stack3 = depth0;
  stack4 = "showDetails";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "dashboard.widgets.HDFSLinks.activeNameNode";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></td>\n            ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              <td><a class=\"disabled-hdfs-link\">");
  stack1 = depth0;
  stack2 = "dashboard.widgets.HDFSLinks.activeNameNode";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></td>\n            ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "view.isTwoStandbyNN";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(9, program9, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n                <!--Two Standby NameNodes-->\n                <td><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.twoStandbyComponent";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "dashboard.widgets.HDFSLinks.standbyNameNodes";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></td>\n              ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n                <!--One Standby NameNode-->\n                <td><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.model.standbyNameNode.host";
  stack3 = depth0;
  stack4 = "showDetails";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "dashboard.widgets.HDFSLinks.standbyNameNode";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></td>\n              ");
  return buffer;}

function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              <td><a class=\"disabled-hdfs-link\">");
  stack1 = depth0;
  stack2 = "dashboard.widgets.HDFSLinks.standbyNameNode";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></td>\n            ");
  return buffer;}

function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n        <table>\n          <!--NameNode-->\n          <tr>\n            <td><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.model.nameNode.host";
  stack3 = depth0;
  stack4 = "showDetails";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.nanmenode";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></td>\n          </tr>\n          <!--SecondaryNameNode-->\n          <tr>\n            <td><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.model.snameNode.host";
  stack3 = depth0;
  stack4 = "showDetails";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.snanmenode";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></td>\n          </tr>\n          <!--Data Nodes-->\n          <tr>\n            <td>\n              <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.component";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.model.dataNodesTotal";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.datanodes";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n            </td>\n          </tr>\n        </table>\n      ");
  return buffer;}

function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "App.QuickViewLinks";
  stack3 = {};
  stack4 = "view.model";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(16, program16, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program16(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <div class=\"btn-group\">\n            <a class=\"btn btn-mini dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n              ");
  stack1 = depth0;
  stack2 = "common.more";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              <span class=\"caret\"></span>\n            </a>\n              <ul class=\"dropdown-menu\">\n                ");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(25, program25, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </ul>\n          </div>\n        ");
  return buffer;}
function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "view.quickLinksArray";
  stack3 = helpers['if'];
  tmp1 = self.program(18, program18, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(22, program22, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program18(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n                    <!--there are multiple masters eg, HBase multiple masters or HDFS HA enabled-->\n                    ");
  stack1 = depth0;
  stack2 = "view.quickLinksArray";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "quickLinks";
  stack7 = helpers.each;
  tmp1 = self.program(19, program19, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program19(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                      <li class=\"dropdown-submenu\">\n                        <a href=\"javascript:void(null)\">");
  stack1 = depth0;
  stack2 = "quickLinks.publicHostNameLabel";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " &nbsp;</a>\n                        <ul class=\"dropdown-menu\">\n                          ");
  stack1 = depth0;
  stack2 = "quickLinks";
  stack3 = helpers.each;
  tmp1 = self.program(20, program20, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                        </ul>\n                      </li>\n                    ");
  return buffer;}
function program20(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                            <li><a ");
  stack1 = {};
  stack2 = "url";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.linkTarget";
  stack1['target'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                          ");
  return buffer;}

function program22(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "view.quickLinks";
  stack3 = helpers.each;
  tmp1 = self.program(23, program23, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program23(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                      <li><a ");
  stack1 = {};
  stack2 = "url";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.linkTarget";
  stack1['target'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                    ");
  return buffer;}

function program25(depth0,data) {


  data.buffer.push("\n                  <div class=\"spinner\"></div>\n                ");}

  data.buffer.push("\n\n<div class=\"links\">\n  <ul>\n  <li class=\"thumbnail row\">\n    <a ");
  stack1 = {};
  stack2 = ":corner-icon :span1 view.parentView.isMoving:hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\" ");
  stack1 = depth0;
  stack2 = "deleteWidget";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <i class=\"icon-remove-sign icon-large\"></i>\n    </a>\n    <div class=\"caption span10\"> ");
  stack1 = depth0;
  stack2 = "view.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n\n    <div class=\"widget-content\" >\n      ");
  stack1 = depth0;
  stack2 = "view.isHAEnabled";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(13, program13, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n    </div>\n\n    <div class=\"link-button\">\n      ");
  stack1 = depth0;
  stack2 = "view.model.quickLinks.length";
  stack3 = helpers['if'];
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n  </li>\n    </ul>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/dashboard/widgets/pie_chart", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <tr><td>");
  stack1 = depth0;
  stack2 = "line";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td></tr>\n        ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <div ");
  stack1 = {};
  stack2 = ":widget-content view.hoverContentTopClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = {};
  stack4 = "view.model";
  stack3['modelBinding'] = stack4;
  stack4 = "view.thresh1";
  stack3['thresh1Binding'] = stack4;
  stack4 = "view.thresh2";
  stack3['thresh2Binding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <div class=\"widget-content-isNA\" >");
  stack1 = depth0;
  stack2 = "services.service.summary.notAvailable";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n    ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"has-hidden-info\">\n  <ul>\n    <li class=\"thumbnail row\">\n      <a ");
  stack1 = {};
  stack2 = ":corner-icon :span1 view.parentView.isMoving:hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\" ");
  stack1 = depth0;
  stack2 = "deleteWidget";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <i class=\"icon-remove-sign icon-large\"></i>\n      </a>\n      <div class=\"caption span10\">");
  stack1 = depth0;
  stack2 = "view.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n      <a class=\"corner-icon span1\" href=\"#\" ");
  stack1 = depth0;
  stack2 = "editWidget";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <i class=\"icon-edit\"></i>\n      </a>\n\n    <div ");
  stack1 = {};
  stack2 = ":hidden-info-general view.hiddenInfoClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <table align=\"center\">\n        ");
  stack1 = depth0;
  stack2 = "view.hiddenInfo";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "line";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </table>\n    </div>\n\n    ");
  stack1 = depth0;
  stack2 = "view.isPieExist";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(5, program5, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </li>\n  </ul>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/dashboard/widgets/simple_text", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <tr><td>");
  stack1 = depth0;
  stack2 = "line";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td></tr>\n          ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"has-hidden-info\">\n  <ul>\n    <li class=\"thumbnail row\" >\n      <a ");
  stack1 = {};
  stack2 = ":corner-icon :span1 view.parentView.isMoving:hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\" ");
  stack1 = depth0;
  stack2 = "deleteWidget";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <i class=\"icon-remove-sign icon-large\"></i>\n      </a>\n      <div class=\"caption span10\"> ");
  stack1 = depth0;
  stack2 = "view.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n      <a class=\"corner-icon span1\" href=\"#\" ");
  stack1 = depth0;
  stack2 = "editWidget";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <i class=\"icon-edit\"></i>\n      </a>\n\n      <div ");
  stack1 = {};
  stack2 = ":hidden-info-general view.hiddenInfoClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <table align=\"center\">\n          ");
  stack1 = depth0;
  stack2 = "view.hiddenInfo";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "line";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </table>\n      </div>\n\n      <div ");
  stack1 = {};
  stack2 = ":widget-content view.hoverContentTopClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n    </li>\n  </ul>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/dashboard/widgets/uptime", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <tr>\n              <td>\n                ");
  stack1 = depth0;
  stack2 = "line";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </td>\n            </tr>\n          ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <div class=\"widget-content-isNA uptime-content-isNA\" >");
  stack1 = depth0;
  stack2 = "services.service.summary.notAvailable";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n      ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <div class=\"widget-content uptime-content\" >\n          ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"has-hidden-info\">\n  <ul>\n    <li class=\"thumbnail row\" >\n      <a ");
  stack1 = {};
  stack2 = ":corner-icon :span1 view.parentView.isMoving:hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\" ");
  stack1 = depth0;
  stack2 = "deleteWidget";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <i class=\"icon-remove-sign icon-large\"></i>\n      </a>\n      <div class=\"caption span11\"> ");
  stack1 = depth0;
  stack2 = "view.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " </div>\n      <div ");
  stack1 = {};
  stack2 = ":hidden-info-general view.hiddenInfoClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <table align=\"center\">\n          ");
  stack1 = depth0;
  stack2 = "view.hiddenInfo";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "line";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </table>\n      </div>\n\n      ");
  stack1 = depth0;
  stack2 = "view.isNA";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(5, program5, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </li>\n  </ul>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/dashboard/widgets/yarn_links", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "App.QuickViewLinks";
  stack3 = {};
  stack4 = "view.model";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <div class=\"btn-group\">\n              <a class=\"btn btn-mini dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n                ");
  stack1 = depth0;
  stack2 = "common.more";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                <span class=\"caret\"></span>\n              </a>\n              <ul class=\"dropdown-menu\">\n                ");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(11, program11, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </ul>\n            </div>\n          ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "view.quickLinksArray";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n                    <!--there are multiple masters eg, HBase multiple masters or HDFS HA enabled-->\n                    ");
  stack1 = depth0;
  stack2 = "view.quickLinksArray";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "quickLinks";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                      <li class=\"dropdown-submenu\">\n                        <a href=\"javascript:void(null)\">");
  stack1 = depth0;
  stack2 = "quickLinks.publicHostNameLabel";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " &nbsp;</a>\n                        <ul class=\"dropdown-menu\">\n                          ");
  stack1 = depth0;
  stack2 = "quickLinks";
  stack3 = helpers.each;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                        </ul>\n                      </li>\n                    ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                            <li><a ");
  stack1 = {};
  stack2 = "url";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.linkTarget";
  stack1['target'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                          ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "view.quickLinks";
  stack3 = helpers.each;
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                      <li><a ");
  stack1 = {};
  stack2 = "url";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.linkTarget";
  stack1['target'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                    ");
  return buffer;}

function program11(depth0,data) {


  data.buffer.push("\n                  <div class=\"spinner\"></div>\n                ");}

  data.buffer.push("\n\n<div class=\"links\">\n  <ul>\n    <li class=\"thumbnail row\">\n      <a ");
  stack1 = {};
  stack2 = ":corner-icon :span1 view.parentView.isMoving:hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        href=\"#\" ");
  stack1 = depth0;
  stack2 = "deleteWidget";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <i class=\"icon-remove-sign icon-large\"></i>\n      </a>\n\n      <div class=\"caption span10\"> ");
  stack1 = depth0;
  stack2 = "view.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n\n      <div class=\"widget-content\">\n        <table>\n          <!-- Resource Manager-->\n          <tr>\n            <td>\n              <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.model.resourceManager.host";
  stack3 = depth0;
  stack4 = "showDetails";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "> ");
  stack1 = depth0;
  stack2 = "dashboard.services.yarn.resourceManager";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n            </td>\n          </tr>\n\n          <!-- Node Manages-->\n          <tr>\n            <td>\n              <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.component";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "> ");
  stack1 = depth0;
  stack2 = "view.model.nodeManagersTotal";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "dashboard.services.yarn.nodeManagers";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n            </td>\n          </tr>\n        </table>\n      </div>\n\n      <div class=\"link-button\">\n        ");
  stack1 = depth0;
  stack2 = "view.model.quickLinks.length";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n    </li>\n  </ul>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "App.HostTableMenuView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n        <div class=\"btn-group display-inline-block\">\n          <a class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n            <span class=\"filters-label\">");
  stack1 = depth0;
  stack2 = "common.filter";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": </span>\n            <span>\n              <span ");
  stack1 = {};
  stack2 = "view.class";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span>\n              ");
  stack1 = depth0;
  stack2 = "view.comboBoxLabel";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              <span class=\"caret\"></span>\n            </span>\n          </a>\n          <ul class=\"dropdown-menu\">\n            <li ");
  stack1 = {};
  stack2 = ":category-item view.filtersUsed::active";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <a ");
  stack1 = depth0;
  stack2 = "clearFilter";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
  stack1 = depth0;
  stack2 = "common.all";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " (");
  stack1 = depth0;
  stack2 = "view.parentView.totalCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ")</a>\n            </li>\n            ");
  stack1 = depth0;
  stack2 = "view.categories";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "category";
  stack7 = helpers.each;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n         </ul>\n       </div>\n      ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "category.isVisible";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n           ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                <li ");
  stack1 = {};
  stack2 = ":category-item category.itemClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <a ");
  stack1 = depth0;
  stack2 = "category";
  stack3 = depth0;
  stack4 = "selectCategory";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n                      <span ");
  stack1 = {};
  stack2 = "category.isHealthStatus:health-status category.healthClass category.class";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span>\n                    ");
  stack1 = depth0;
  stack2 = "category.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  </a>\n               </li>\n             ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <th class=\"first\"> </th>\n        <th> </th>\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.nameSort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        <th> </th>\n        <th> </th>\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.ipSort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.cpuSort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.memorySort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.diskUsageSort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.loadAvgSort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        <th ");
  stack1 = {};
  stack2 = "App.supports.stackUpgrade::hidden App.stackVersionsAvailable::hidden :sort-view-11";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "hosts.host.menu.stackVersions";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </th>\n        <th class=\"sort-view-6\">");
  stack1 = depth0;
  stack2 = "common.components";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n      ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "host";
  stack7 = helpers.each;
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.HostView";
  stack3 = {};
  stack4 = "host";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(11, program11, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n\n          <td class=\"first\">");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "host.selected";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n\n          <td class=\"health\">\n            <span rel=\"HealthTooltip\" ");
  stack1 = {};
  stack2 = "host.healthClass host.healthIconClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "host.healthToolTip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span>\n          </td>\n\n          <td class=\"name\">\n            <span class=\"trim_hostname\">\n              <a title=\"");
  stack1 = depth0;
  stack2 = "host.hostName";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\" href=\"#\" ");
  stack1 = depth0;
  stack2 = "host";
  stack3 = depth0;
  stack4 = "showDetails";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "host.hostName";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n            </span>\n            ");
  stack1 = depth0;
  stack2 = "host.criticalWarningAlertsCount";
  stack3 = helpers['if'];
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </td>\n          <td class=\"restart\">\n            ");
  stack1 = depth0;
  stack2 = "host.componentsWithStaleConfigsCount";
  stack3 = helpers['if'];
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </td>\n          <td class=\"passive-state\">\n              <span rel=\"ComponentsTooltip\" ");
  stack1 = {};
  stack2 = "view.componentsInPassiveStateMessage";
  stack1['data-original-title'] = stack2;
  stack2 = "host.componentsInPassiveStateCount:icon-medkit";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span>\n          </td>\n          <td class=\"host-ip\">");
  stack1 = depth0;
  stack2 = "host.ip";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n          <td class=\"cores-formatted\">");
  stack1 = depth0;
  stack2 = "host.coresFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n          <td class=\"memory-formatted\">");
  stack1 = depth0;
  stack2 = "host.memoryFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n\n          <td class=\"disk-info\">\n            <div class=\"progress progress-info\" ");
  stack1 = {};
  stack2 = "host.diskInfoBar";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " rel=\"UsageTooltip\">\n              <div class=\"bar\" ");
  stack1 = {};
  stack2 = "view.usageStyle";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n            </div>\n          </td>\n\n          <td class=\"load-avg\">");
  stack1 = depth0;
  stack2 = "host.loadAvg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n          <td ");
  stack1 = {};
  stack2 = "App.supports.stackUpgrade::hidden App.stackVersionsAvailable::hidden view.hasSingleVersion:not-active-link :host-table-versions";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"#\" class=\"expander black\" ");
  stack1 = depth0;
  stack2 = "toggleVersions";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "view.hasSingleVersion";
  stack3 = helpers.unless;
  tmp1 = self.program(19, program19, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "view.currentVersion";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </a>\n            <div ");
  stack1 = {};
  stack2 = "view.isVersionsCollapsed:noDisplay:display :stack-versions :collapsed-list";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "view.versionLabels";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </div>\n          </td>\n          <td class=\"host-table-components\">\n            <a href=\"#\" class=\"expander\" ");
  stack1 = depth0;
  stack2 = "toggleComponents";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "> <span ");
  stack1 = {};
  stack2 = "view.isComponentsCollapsed:icon-caret-right:icon-caret-down";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span>\n              ");
  stack1 = depth0;
  stack2 = "view.content.hostComponents.length";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.content.hostComponents.length";
  stack3 = {};
  stack4 = "t:common.component";
  stack3['singular'] = stack4;
  stack4 = "t:common.components";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n            <div ");
  stack1 = {};
  stack2 = "view.isComponentsCollapsed:noDisplay:display :host-components :collapsed-list";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "view.labels";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </div>\n          </td>\n        ");
  return buffer;}
function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "host.alertsSummary.CRITICAL";
  stack3 = helpers['if'];
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(15, program15, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                <span class=\"label alerts-crit-count\" ");
  stack1 = depth0;
  stack2 = "host";
  stack3 = depth0;
  stack4 = "goToHostAlerts";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "host.criticalWarningAlertsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n              ");
  return buffer;}

function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                <span class=\"label alerts-warn-count\" ");
  stack1 = depth0;
  stack2 = "host";
  stack3 = depth0;
  stack4 = "goToHostAlerts";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "host.criticalWarningAlertsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n              ");
  return buffer;}

function program17(depth0,data) {

  var buffer = '', stack1, stack2;
  data.buffer.push("\n              <span class=\"muted icon-refresh\" rel=\"ComponentsTooltip\" ");
  stack1 = {};
  stack2 = "view.restartRequiredComponentsMessage";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span>\n            ");
  return buffer;}

function program19(depth0,data) {

  var buffer = '', stack1, stack2;
  data.buffer.push("\n                <span ");
  stack1 = {};
  stack2 = "view.isVersionsCollapsed:icon-caret-right:icon-caret-down";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span>\n              ");
  return buffer;}

function program21(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <tr>\n        <td class=\"first\"> </td>\n        <td ");
  stack1 = {};
  stack2 = "view.colspan";
  stack1['colspan'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "hosts.table.noHosts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n    ");
  return buffer;}

function program23(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <div>\n          <a ");
  stack1 = depth0;
  stack2 = "filterSelected";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n            ");
  stack1 = depth0;
  stack2 = "view.selectedHosts.length";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  stack1 = depth0;
  stack2 = "view.selectedHostsCount";
  stack3 = {};
  stack4 = "t:hosts.filters.selectedHostInfo";
  stack3['singular'] = stack4;
  stack4 = "t:hosts.filters.selectedHostsInfo";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "\n          </a>\n        </div> - <a ");
  stack1 = depth0;
  stack2 = "clearSelection";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
  stack1 = depth0;
  stack2 = "hosts.filters.clearSelection";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n      ");
  return buffer;}

  data.buffer.push("\n\n<div id=\"hosts\">\n\n  <div class=\"box-header row\">\n    <div class=\"hosts-actions pull-left\">\n      ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n    <div class=\"health-status-bar pull-left\">\n      ");
  stack1 = depth0;
  stack2 = "view.statusFilter";
  stack3 = {};
  stack4 = "view.categories";
  stack3['categoriesBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n  </div>\n\n  <table class=\"table advanced-header-table table-bordered table-striped\" id=\"hosts-table\">\n    <thead>\n      ");
  stack1 = depth0;
  stack2 = "view.sortView";
  stack3 = {};
  stack4 = "label-row";
  stack3['classNames'] = stack4;
  stack4 = "view.filteredContent";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      <tr class=\"filter-row\">\n        <th class=\"first\"><div class=\"ember-view view-wrapper\">");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "view.selectAllHosts";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div></th>\n        <th> </th>\n        <th>");
  stack1 = depth0;
  stack2 = "view.nameFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n        <th> </th>\n        <th> </th>\n        <th>");
  stack1 = depth0;
  stack2 = "view.ipFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n        <th>");
  stack1 = depth0;
  stack2 = "view.cpuFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n        <th>");
  stack1 = depth0;
  stack2 = "view.ramFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n        <th> </th>\n        <th>");
  stack1 = depth0;
  stack2 = "view.loadAvgFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n        <th ");
  stack1 = {};
  stack2 = "App.supports.stackUpgrade::hidden App.stackVersionsAvailable::hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "view.versionsFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </th>\n        <th>");
  stack1 = depth0;
  stack2 = "view.componentsFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n      </tr>\n    </thead>\n    <tbody>\n    ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(21, program21, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </tbody>\n  </table>\n\n  <div ");
  stack1 = {};
  stack2 = "view.filteringComplete:hidden :table-overlay";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <div class=\"spinner\"></div>\n  </div>\n\n  <div class=\"page-bar\">\n    <div class=\"filtered-info span4\">\n      <label>");
  stack1 = depth0;
  stack2 = "view.filteredContentInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " - <a ");
  stack1 = depth0;
  stack2 = "clearFilters";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " id=\"hosts-page-clear-filters\" href=\"#\">");
  stack1 = depth0;
  stack2 = "tableView.filters.clearAllFilters";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></label>\n    </div>\n    <div class=\"selected-hosts-info span4\">\n      ");
  stack1 = depth0;
  stack2 = "view.showSelectedFilter";
  stack3 = helpers['if'];
  tmp1 = self.program(23, program23, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n    <div class=\"items-on-page\">\n      <label>");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.rowsPerPageSelectView";
  stack3 = {};
  stack4 = "view.displayLength";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n    </div>\n    <div class=\"info\">");
  stack1 = depth0;
  stack2 = "view.paginationInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n    <div class=\"paging_two_button\">\n      <a ");
  stack1 = {};
  stack2 = "view.paginationLeftClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "previousPage";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-arrow-left\"></i></a>\n      <a ");
  stack1 = {};
  stack2 = "view.paginationRightClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "nextPage";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-arrow-right\"></i></a>\n    </div>\n  </div>\n</div>\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/add", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program3(depth0,data) {


  data.buffer.push("\n                <div class=\"spinner\"></div>\n            ");}

  data.buffer.push("\n\n<div class=\"wizard\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n\n      <!--<a class=\"btn back\" ");
  stack1 = depth0;
  stack2 = "backToHostsList";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\\u2190 Back to Hosts</a>-->\n\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li class=\"nav-header\">");
  stack1 = depth0;
  stack2 = "hosts.add.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n              <li ");
  stack1 = {};
  stack2 = "isStep1:active view.isStep1Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep1";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep2:active view.isStep2Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep2";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep3:active view.isStep3Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep3";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step6.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep4:active view.isStep4Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep4";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "addHost.step4.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep5:active view.isStep5Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep5";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step8.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep6:active view.isStep6Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep6";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step9.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep7:active view.isStep7Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep7";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step10.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n            </ul>\n          </div>\n        </div>\n          <div class=\"wizard-content well span9\">\n            ");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/addHost/step4", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            <tr>\n              <td>");
  stack1 = depth0;
  stack2 = "service.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n              <td>\n                ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "service.configGroupsNames";
  stack3['contentBinding'] = stack4;
  stack4 = "service.selectedConfigGroup";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </td>\n            </tr>\n          ");
  return buffer;}

  data.buffer.push("\n<div>\n  <h2>");
  stack1 = depth0;
  stack2 = "addHost.step4.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n  <div class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "addHost.step4.title";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <div class=\"pre-scrollable\">\n      <table class=\"table\">\n        <thead>\n          <tr>\n            <th>");
  stack1 = depth0;
  stack2 = "common.service";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n            <th>");
  stack1 = depth0;
  stack2 = "common.conf.group";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n          </tr>\n        </thead>\n        <tbody>\n          ");
  stack1 = depth0;
  stack2 = "controller.content.configGroups";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "service";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </tbody>\n      </table>\n  </div>\n   <div class=\"btn-area\">\n      <a class=\"btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n      <a class=\"btn btn-success pull-right\" ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n    </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/bulk_operation_confirm_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "showAll";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "tableView.filters.showAll";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"alert alert-warning\">\n      ");
  stack1 = depth0;
  stack2 = "view.warningInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "<br />\n      <pre class=\"limited-height alert alert-warning\"><strong>");
  stack1 = depth0;
  stack2 = "hostNamesSkippedVisible";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong></pre>\n    </div>\n  ");
  return buffer;}

  data.buffer.push("\n\n<p>");
  stack1 = depth0;
  stack2 = "view.message";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</p>\n<div class=\"task-top-wrap\">\n  <div class=\"task-detail-ico-wrap\">\n    <a href=\"#\" title=\"Click to Copy\" ");
  stack1 = depth0;
  stack2 = "textTrigger";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-copy\"><i class=\"icon-copy\"></i> ");
  stack1 = depth0;
  stack2 = "common.copy";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n  </div>\n</div>\n<div class=\"task-detail-log-info\">\n  <div class=\"content-area\">\n    <div ");
  stack1 = {};
  stack2 = "view.textareaVisible::hidden :task-detail-log-clipboard-wrap";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <textarea class=\"task-detail-log-clipboard\"></textarea>\n    </div>\n    <div ");
  stack1 = {};
  stack2 = "view.textareaVisible:hidden :task-detail-log-maintext";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <pre class=\"limited-height\">");
  stack1 = depth0;
  stack2 = "visibleHosts";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</pre>\n    </div>\n    ");
  stack1 = depth0;
  stack2 = "expanded";
  stack3 = helpers.unless;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  ");
  stack1 = depth0;
  stack2 = "hostNamesSkipped";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/bulk_operation_menu", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <li><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "addHost";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-plus icon-white\"></i> ");
  stack1 = depth0;
  stack2 = "hosts.host.add";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n      <li class=\"divider\"></li>\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n              <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "view.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n              <ul class=\"dropdown-menu\">\n                ");
  stack1 = depth0;
  stack2 = "view.operationsInfo";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "operation";
  stack7 = helpers.each;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </ul>\n          ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "operation.label.length";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "view.operationView";
  stack3 = {};
  stack4 = "operation.operationData";
  stack3['contentBinding'] = stack4;
  stack4 = "s";
  stack3['selection'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                        <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "operation.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                    ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "view.slaveItemView";
  stack3 = {};
  stack4 = "component";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(9, program9, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n                <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "component.componentNameFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                <ul class=\"dropdown-menu\">\n                  ");
  stack1 = depth0;
  stack2 = "view.operationsInfo";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "operation";
  stack7 = helpers.each;
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </ul>\n            ");
  return buffer;}
function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "operation.decommission";
  stack3 = helpers['if'];
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(14, program14, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                      ");
  stack1 = depth0;
  stack2 = "view.advancedOperationView";
  stack3 = {};
  stack4 = "operation.operationData";
  stack3['contentBinding'] = stack4;
  stack4 = "s";
  stack3['selection'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(12, program12, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    ");
  return buffer;}
function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                          <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "operation.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                      ");
  return buffer;}

function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                      ");
  stack1 = depth0;
  stack2 = "view.commonOperationView";
  stack3 = {};
  stack4 = "operation.operationData";
  stack3['contentBinding'] = stack4;
  stack4 = "s";
  stack3['selection'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(15, program15, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    ");
  return buffer;}
function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                          <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "operation.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                      ");
  return buffer;}

function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n            <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "view.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n            <ul class=\"dropdown-menu\">\n              ");
  stack1 = depth0;
  stack2 = "view.operationsInfo";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "operation";
  stack7 = helpers.each;
  tmp1 = self.program(18, program18, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </ul>\n          ");
  return buffer;}
function program18(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "operation.label.length";
  stack3 = helpers['if'];
  tmp1 = self.program(19, program19, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  return buffer;}
function program19(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "view.operationView";
  stack3 = {};
  stack4 = "operation.operationData";
  stack3['contentBinding'] = stack4;
  stack4 = "f";
  stack3['selection'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(20, program20, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program20(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "operation.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                  ");
  return buffer;}

function program22(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "view.slaveItemView";
  stack3 = {};
  stack4 = "component";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(23, program23, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program23(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n                <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "component.componentNameFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                  <ul class=\"dropdown-menu\">\n                    ");
  stack1 = depth0;
  stack2 = "view.operationsInfo";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "operation";
  stack7 = helpers.each;
  tmp1 = self.program(24, program24, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  </ul>\n              ");
  return buffer;}
function program24(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                      ");
  stack1 = depth0;
  stack2 = "operation.decommission";
  stack3 = helpers['if'];
  tmp1 = self.program(25, program25, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(28, program28, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    ");
  return buffer;}
function program25(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                        ");
  stack1 = depth0;
  stack2 = "view.advancedOperationView";
  stack3 = {};
  stack4 = "operation.operationData";
  stack3['contentBinding'] = stack4;
  stack4 = "f";
  stack3['selection'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(26, program26, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                      ");
  return buffer;}
function program26(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                        <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "operation.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                        ");
  return buffer;}

function program28(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                        ");
  stack1 = depth0;
  stack2 = "view.commonOperationView";
  stack3 = {};
  stack4 = "operation.operationData";
  stack3['contentBinding'] = stack4;
  stack4 = "f";
  stack3['selection'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(29, program29, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                      ");
  return buffer;}
function program29(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                        <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "operation.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                        ");
  return buffer;}

function program31(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n              <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "view.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n              <ul class=\"dropdown-menu\">\n                ");
  stack1 = depth0;
  stack2 = "view.operationsInfo";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "operation";
  stack7 = helpers.each;
  tmp1 = self.program(32, program32, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </ul>\n          ");
  return buffer;}
function program32(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "operation.label.length";
  stack3 = helpers['if'];
  tmp1 = self.program(33, program33, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program33(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "view.operationView";
  stack3 = {};
  stack4 = "operation.operationData";
  stack3['contentBinding'] = stack4;
  stack4 = "a";
  stack3['selection'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(34, program34, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program34(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                        <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "operation.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                    ");
  return buffer;}

function program36(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "view.slaveItemView";
  stack3 = {};
  stack4 = "component";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(37, program37, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program37(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n                <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "component.componentNameFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                <ul class=\"dropdown-menu\">\n                  ");
  stack1 = depth0;
  stack2 = "view.operationsInfo";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "operation";
  stack7 = helpers.each;
  tmp1 = self.program(38, program38, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </ul>\n            ");
  return buffer;}
function program38(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "operation.decommission";
  stack3 = helpers['if'];
  tmp1 = self.program(39, program39, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(42, program42, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program39(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                      ");
  stack1 = depth0;
  stack2 = "view.advancedOperationView";
  stack3 = {};
  stack4 = "operation.operationData";
  stack3['contentBinding'] = stack4;
  stack4 = "a";
  stack3['selection'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(40, program40, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    ");
  return buffer;}
function program40(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                          <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "operation.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                      ");
  return buffer;}

function program42(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                      ");
  stack1 = depth0;
  stack2 = "view.commonOperationView";
  stack3 = {};
  stack4 = "operation.operationData";
  stack3['contentBinding'] = stack4;
  stack4 = "a";
  stack3['selection'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(43, program43, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    ");
  return buffer;}
function program43(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                          <a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "operation.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                      ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"dropdown bulk-menu\">\n  <a class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">");
  stack1 = depth0;
  stack2 = "common.actions";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " <span class=\"caret\"></span></a>\n  <ul class=\"dropdown-menu\">\n    ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <li class=\"dropdown-submenu\">\n      <a ");
  stack1 = {};
  stack2 = "view.parentView.showSelectedFilter::disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " tabindex=\"-1\" href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "view.menuItems.s.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        (");
  stack1 = depth0;
  stack2 = "view.parentView.selectedHosts.length";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ")</a>\n        <ul ");
  stack1 = {};
  stack2 = "view.parentView.showSelectedFilter::hidden :dropdown-menu";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "view.hostItemView";
  stack3 = helpers.view;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.components";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </ul>\n    </li>\n    <li class=\"dropdown-submenu\">\n      <a ");
  stack1 = {};
  stack2 = "view.parentView.hasFilteredItems::disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " tabindex=\"-1\" href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "view.menuItems.f.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        (");
  stack1 = depth0;
  stack2 = "view.parentView.filteredCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ")</a>\n        <ul ");
  stack1 = {};
  stack2 = "view.parentView.hasFilteredItems::hidden :dropdown-menu";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "view.hostItemView";
  stack3 = helpers.view;
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.components";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(22, program22, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </ul>\n    </li>\n    <li class=\"dropdown-submenu\">\n      <a tabindex=\"-1\" href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "view.menuItems.a.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        (");
  stack1 = depth0;
  stack2 = "view.parentView.totalCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ")</a>\n        <ul class=\"dropdown-menu\">\n          ");
  stack1 = depth0;
  stack2 = "view.hostItemView";
  stack3 = helpers.view;
  tmp1 = self.program(31, program31, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.components";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(36, program36, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </ul>\n    </li>\n  </ul>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/component_filter", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              <li>\n                <label class=\"checkbox\">\n                ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "component.checkedForHostFilter";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "component.displayName";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                </label>\n              </li>\n            ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              <li>\n                <label class=\"checkbox\">\n                ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "component.checkedForHostFilter";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "component.displayName";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                </label>\n              </li>\n            ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              <li>\n              <label class=\"checkbox\">\n              ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "component.checkedForHostFilter";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "component.displayName";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </label>\n              </li>\n            ");
  return buffer;}

  data.buffer.push("\n\n<button class=\"btn single-btn-group filter-btn\" id=\"host-components-filter-button\" ");
  stack1 = depth0;
  stack2 = "clickFilterButton";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.filter";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " <i class=\"icon-filter\"></i></button>\n    <ul class=\"dropdown-menu filter-components\">\n      <li id=\"title-bar\">");
  stack1 = depth0;
  stack2 = "hosts.filters.filterComponents";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        <a class=\"close\" ");
  stack1 = depth0;
  stack2 = "closeFilter";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">X</a>\n      </li>\n      <li id=\"list-area\">\n        <ul>\n          <li>\n            <label class=\"checkbox\">\n            ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "view.masterComponentsChecked";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "host.host.componentFilter.master";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":\n            </label>\n            <ul>\n            ");
  stack1 = depth0;
  stack2 = "view.masterComponents";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </ul>\n          </li>\n          <li>\n            <label class=\"checkbox\">\n            ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "view.slaveComponentsChecked";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "host.host.componentFilter.slave";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":\n            </label>\n            <ul>\n            ");
  stack1 = depth0;
  stack2 = "view.slaveComponents";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </ul>\n          </li>\n          <li>\n            <label class=\"checkbox\">\n            ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "view.clientComponentsChecked";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "host.host.componentFilter.client";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":\n            </label>\n            <ul>\n            ");
  stack1 = depth0;
  stack2 = "view.clientComponents";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </ul>\n          </li>\n        </ul>\n      </li>\n      <li id=\"button-bar\">\n        <button class=\"btn btn-inverse\" ");
  stack1 = depth0;
  stack2 = "applyFilter";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.apply";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n      </li>\n    </ul>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/configs", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"row-fluid\">\n        <div class=\"well span2 services-host-menu\">\n          ");
  stack1 = depth0;
  stack2 = "App.MainHostServiceMenuView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n        <div class=\"span10\">\n          ");
  stack1 = depth0;
  stack2 = "service_config_outlet";
  foundHelper = helpers.outlet;
  stack3 = foundHelper || depth0.outlet;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "outlet", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n    </div>\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"services-host-menu\">\n        <i class=\"message alert alert-info\">");
  stack1 = depth0;
  stack2 = "hosts.host.serviceNotAvailable";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</i>\n    </div>\n");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.isConfigAvailable";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/configs_service", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div class=\"clearfix\"></div>\n    ");
  stack1 = depth0;
  stack2 = "App.ServiceConfigView";
  stack3 = {};
  stack4 = "view.isConfigsEditable";
  stack3['canEditBinding'] = stack4;
  stack4 = "controller.filter";
  stack3['filterBinding'] = stack4;
  stack4 = "controller.filterColumns";
  stack3['columnsBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

function program3(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

  data.buffer.push("\n\n<div id=\"serviceConfig\">\n  ");
  stack1 = depth0;
  stack2 = "dataIsLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/decommission", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n\n<li rel='decommissionTooltip' ");
  stack1 = {};
  stack2 = "view.parentView.decommissionTooltipMessage";
  stack1['data-original-title'] = stack2;
  stack2 = "view.parentView.noActionAvailable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
  stack1 = {};
  stack2 = "view.parentView.isComponentDecommissionDisable:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      ");
  stack1 = depth0;
  stack2 = "view.text";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </a>\n</li>\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/details", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div id=\"host-details\">\n    <div class=\"status-info\">\n      <span rel=\"HealthTooltip\" ");
  stack1 = {};
  stack2 = "view.content.healthClass view.content.healthIconClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.content.healthToolTip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span><span\n            class='host-title'>");
  stack1 = depth0;
  stack2 = "view.content.hostName";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n      ");
  stack1 = depth0;
  stack2 = "view.isActive";
  stack3 = helpers.unless;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n    <div><a href=\"javascript:void(null)\" id=\"host-details-back-btn\" data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i\n            class=\"icon-arrow-left\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></div>\n    <div class=\"content\">\n      ");
  stack1 = depth0;
  stack2 = "App.MainHostMenuView";
  stack3 = {};
  stack4 = "view.content";
  stack3['hostBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <span class=\"host-maintenance-notice pull-right\"><span class=\"icon-medkit\"></span> ");
  stack1 = depth0;
  stack2 = "hosts.host.passive.mode";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n      ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n        <div class=\"service-button\">\n          <div class=\"btn-group display-inline-block\">\n            <a href=\"javascript:void(null)\" ");
  stack1 = {};
  stack2 = ":btn :dropdown-toggle";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " id=\"host-details-action-btn\" data-toggle=\"dropdown\">\n              ");
  stack1 = depth0;
  stack2 = "hosts.host.details.hostActions";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              <span class=\"caret\"></span>\n            </a>\n            <ul class=\"dropdown-menu pull-right\">\n              <!-- dropdown menu links -->\n              ");
  stack1 = depth0;
  stack2 = "view.maintenance";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "option";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "view.clients";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </ul>\n          </div>\n        </div>\n      ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                <li ");
  stack1 = {};
  stack2 = "controller.isStopDisabled:disabled option.liClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <a ");
  stack1 = depth0;
  stack2 = "option";
  stack3 = depth0;
  stack4 = "doAction";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = true;
  stack5['href'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i ");
  stack1 = {};
  stack2 = "option.cssClass option.liClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i> ");
  stack1 = depth0;
  stack2 = "option.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  </a>\n                </li>\n              ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n                <li class=\"dropdown-submenu submenu-left\">\n                  <a><i class=\"icon-download-alt\"></i> ");
  stack1 = depth0;
  stack2 = "services.service.actions.downloadClientConfigs";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n\n                  <div class=\"dropdown-menu-wrap\">\n                    <ul class=\"dropdown-menu\">\n                      ");
  stack1 = depth0;
  stack2 = "view.clientsWithConfigs";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "client";
  stack7 = helpers.each;
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    </ul>\n                  </div>\n                </li>\n              ");
  return buffer;}
function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                        <li>\n                          <a ");
  stack1 = depth0;
  stack2 = "client";
  stack3 = depth0;
  stack4 = "downloadClientConfigs";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = true;
  stack5['href'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "client.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                        </li>\n                      ");
  return buffer;}

function program10(depth0,data) {


  data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(10, program10, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/details/addComponentPopup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "addComponentMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "<br /><br />");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/details/deleteComponentPopup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div class=\"alert-error row-fluid\">\n        <div class='tinyspan tinyoffset'>");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "isChecked";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class='span10'>");
  stack1 = depth0;
  stack2 = "lastComponentError";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  <br />\n  <div class='alert'>");
  stack1 = depth0;
  stack2 = "deleteZkServerMsg";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  <br />\n  <div class='alert'>");
  stack1 = depth0;
  stack2 = "deleteHiveMetastoreMsg";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "deleteComponentMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "<br /><br />\n");
  stack1 = depth0;
  stack2 = "lastComponent";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "isZkServer";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "isHiveMetastore";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/details/doDeleteHostPopup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "hosts.delete.popup.unknownComponents";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  <br /><br />\n  <div class='row-fluid'>\n    <div class='tinyoffset span10'>\n      <i>");
  stack1 = depth0;
  stack2 = "unknownComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</i>\n    </div>\n  </div>\n  <br />\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div class=\"alert-error row-fluid\">\n        <div class='tinyspan tinyoffset'>");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "isChecked";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class='span10'>");
  stack1 = depth0;
  stack2 = "lastComponentError";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "hosts.delete.popup.body.msg.unknownComponents";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "unknownComponents";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n<p><i class=\"icon-warning-sign\"></i> ");
  stack1 = depth0;
  stack2 = "deletePopupBody";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</p>\n");
  stack1 = depth0;
  stack2 = "lastComponent";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n<br />\n<div class='alert'>");
  stack1 = depth0;
  stack2 = "common.important";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "unknownComponents";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "hosts.delete.popup.body.msg1";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n<div class='alert'>");
  stack1 = depth0;
  stack2 = "common.important";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push(" ");
  stack1 = depth0;
  stack2 = "hosts.delete.popup.body.msg3";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/details/host_component", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", blockHelperMissing=helpers.blockHelperMissing, helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n      <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "upgradeComponent";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " >\n        <i title=\"Component upgrade failed\" class=\"components-health icon-arrow-up\"></i>\n      </a>\n    ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2;
  data.buffer.push("\n    <span rel='componentHealthTooltip' ");
  stack1 = {};
  stack2 = "view.statusClass view.statusIconClass :components-health";
  stack1['class'] = stack2;
  stack2 = "view.componentTextStatus";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span>&nbsp;\n  ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "component.displayNameAdvanced";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "component.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

function program10(depth0,data) {


  data.buffer.push("\n    <span class=\"text-warning icon-refresh\"></span>\n  ");}

function program12(depth0,data) {


  data.buffer.push("\n    <span class=\"icon-medkit\"></span>\n  ");}

function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    <div class=\"btn-group pull-right\">\n      <a ");
  stack1 = {};
  stack2 = "view.disabled :btn :dropdown-toggle";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " data-toggle=\"dropdown\">\n        ");
  stack1 = depth0;
  stack2 = "view.componentTextStatus";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        <span class=\"caret pull-right\"></span>\n      </a>\n      <ul class=\"dropdown-menu\">\n        ");
  stack1 = depth0;
  stack2 = "view.isComponentDecommissionAvailable";
  stack3 = helpers['if'];
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.isComponentRecommissionAvailable";
  stack3 = helpers['if'];
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.isRestartableComponent";
  stack3 = helpers['if'];
  tmp1 = self.program(19, program19, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.isInstalling";
  stack3 = helpers.unless;
  tmp1 = self.program(21, program21, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.isInit";
  stack3 = helpers['if'];
  tmp1 = self.program(37, program37, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.isDeletableComponent";
  stack3 = helpers['if'];
  tmp1 = self.program(39, program39, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.isRefreshConfigsAllowed";
  stack3 = helpers['if'];
  tmp1 = self.program(41, program41, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n      ");
  stack1 = depth0;
  stack2 = "view.customCommands";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "command";
  stack7 = helpers.each;
  tmp1 = self.program(43, program43, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </ul>\n    </div>\n  ");
  return buffer;}
function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.decommissionView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  return buffer;}

function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.decommissionView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  return buffer;}

function program19(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n          <li ");
  stack1 = {};
  stack2 = "view.isRestartComponentDisabled:hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "restartComponent";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "common.restart";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n        ");
  return buffer;}

function program21(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.isStart";
  stack3 = helpers['if'];
  tmp1 = self.program(22, program22, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.isStart";
  stack3 = helpers.unless;
  tmp1 = self.program(24, program24, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.isUpgradeFailed";
  stack3 = helpers['if'];
  tmp1 = self.program(27, program27, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.isInstallFailed";
  stack3 = helpers['if'];
  tmp1 = self.program(29, program29, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.isReassignable";
  stack3 = helpers['if'];
  tmp1 = self.program(31, program31, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          <li ");
  stack1 = {};
  stack2 = "view.noActionAvailable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"javascript:void(null)\"\n               data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "toggleMaintenanceMode";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "view.isActive";
  stack3 = helpers['if'];
  tmp1 = self.program(33, program33, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(35, program35, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </a>\n          </li>\n        ");
  return buffer;}
function program22(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <li ");
  stack1 = {};
  stack2 = "view.isComponentDecommissioning:hidden view.noActionAvailable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "stopComponent";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                ");
  stack1 = depth0;
  stack2 = "common.stop";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n          ");
  return buffer;}

function program24(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "view.isInit";
  stack3 = helpers.unless;
  tmp1 = self.program(25, program25, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program25(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n              <li ");
  stack1 = {};
  stack2 = "view.isUpgradeFailed:hidden view.isInstallFailed:hidden view.isDecommissioning:hidden view.noActionAvailable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "startComponent";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  ");
  stack1 = depth0;
  stack2 = "common.start";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              </li>\n            ");
  return buffer;}

function program27(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <li ");
  stack1 = {};
  stack2 = "view.noActionAvailable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "upgradeComponent";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                ");
  stack1 = depth0;
  stack2 = "common.reUpgrade";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n          ");
  return buffer;}

function program29(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <li ");
  stack1 = {};
  stack2 = "view.noActionAvailable";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "installComponent";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                ");
  stack1 = depth0;
  stack2 = "common.reinstall";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n          ");
  return buffer;}

function program31(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <li ");
  stack1 = {};
  stack2 = "view.noActionAvailable view.isMoveComponentDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "moveComponent";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                ");
  stack1 = depth0;
  stack2 = "common.move";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n          ");
  return buffer;}

function program33(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "passiveState.turnOn";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  return buffer;}

function program35(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "passiveState.turnOff";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  return buffer;}

function program37(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n          <li ");
  stack1 = {};
  stack2 = "view.";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "installComponent";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "common.reinstall";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n        ");
  return buffer;}

function program39(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <li ");
  stack1 = {};
  stack2 = "view.isDeleteComponentDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "deleteComponent";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  ");
  stack1 = depth0;
  stack2 = "common.delete";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                </a>\n            </li>\n        ");
  return buffer;}

function program41(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <li>\n                <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "refreshComponentConfigs";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  ");
  stack1 = depth0;
  stack2 = "hosts.host.details.refreshConfigs";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                </a>\n            </li>\n        ");
  return buffer;}

function program43(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n        <li>\n          <a href=\"javascript:void(null)\" ");
  stack1 = depth0;
  stack2 = "command";
  stack3 = depth0;
  stack4 = "executeCustomCommand";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = true;
  stack5['href'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "command.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n        </li>\n      ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"span7\">\n  ");
  stack1 = depth0;
  stack2 = "view.isUpgradeFailed";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "component.displayNameAdvanced";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  &nbsp;/&nbsp;\n  <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "component.service";
  stack3 = depth0;
  stack4 = "routeToService";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "component.service.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n  &nbsp;\n  ");
  stack1 = depth0;
  stack2 = "component.staleConfigs";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "component.isActive";
  stack3 = helpers.unless;
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n<div class=\"span5 pull-right\">\n  ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/details/installComponentPopup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "installComponentMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "<br /><br />");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/details/raiseDeleteComponentErrorPopup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div class=\"warning-details\">");
  stack1 = depth0;
  stack2 = "componentsBodyEnd";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</div>\n");
  return buffer;}

  data.buffer.push("\n\n<div class=\"dialog-delete-component\">\n  <div class=\"warning\">\n  <i class=\"icon-warning-sign\"></i> <strong>");
  stack1 = depth0;
  stack2 = "componentsBody";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong>\n  </div>\n  <div class=\"row-fluid\">\n    <div class=\"tinyoffset span10 warning-list\">\n    ");
  stack1 = depth0;
  stack2 = "componentsStr";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n  </div>\n\n");
  stack1 = depth0;
  stack2 = "showBodyEnd";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/host_alerts", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n\n  <table class=\"table advanced-header-table table-bordered table-striped alerts-table\" id=\"host-alerts-table\">\n    <thead>\n      ");
  stack1 = depth0;
  stack2 = "view.sortView";
  stack3 = {};
  stack4 = "label-row";
  stack3['classNames'] = stack4;
  stack4 = "view.filteredContent";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <tr class=\"filter-row\">\n      <th class=\"first\">");
  stack1 = depth0;
  stack2 = "view.serviceFilterView";
  stack3 = {};
  stack4 = "service-filter";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "view.nameFilterView";
  stack3 = {};
  stack4 = "name-filter";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "view.stateFilterView";
  stack3 = {};
  stack4 = "status-filter";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "view.textView";
  stack3 = {};
  stack4 = "text-filter";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n    </tr>\n    </thead>\n    <tbody>\n    ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(10, program10, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </tbody>\n  </table>\n\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.serviceSort";
  stack3 = {};
  stack4 = "first service-sorting";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.nameSort";
  stack3 = {};
  stack4 = "name-sorting";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.statusSort";
  stack3 = {};
  stack4 = "status-sorting";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.textSort";
  stack3 = {};
  stack4 = "text-sorting";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "alertInstance";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n        <tr>\n          <td class=\"first\">\n            ");
  stack1 = depth0;
  stack2 = "alertInstance.isAmbariServiceName";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </td>\n          <td>\n            <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "alertInstance.definitionId";
  stack3 = depth0;
  stack4 = "routeToAlertDefinition";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alertInstance.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n          </td>\n          <td>");
  stack1 = depth0;
  stack2 = "alertInstance.status";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push(" <time class=\"timeago\" ");
  stack1 = {};
  stack2 = "alertInstance.statusChangedAndLastCheckedFormatted";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alertInstance.lastTriggeredForFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</time></td>\n          <td><span class=\"alert-text\" ");
  stack1 = {};
  stack2 = "alertInstance.text";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"alert-text\">");
  stack1 = depth0;
  stack2 = "alertInstance.text";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span></td>\n        </tr>\n      ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "alertInstance.serviceDisplayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n              <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "alertInstance.service";
  stack3 = depth0;
  stack4 = "routeToService";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alertInstance.serviceDisplayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n            ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <tr>\n        <td class=\"first\"></td>\n        <td colspan=\"3\">\n          ");
  stack1 = depth0;
  stack2 = "alerts.table.noAlerts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </td>\n      </tr>\n    ");
  return buffer;}

function program12(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

  data.buffer.push("\n\n<div id=\"alerts\">\n  <div class=\"box-header row\">\n  </div>\n\n  ");
  stack1 = depth0;
  stack2 = "App.router.mainAlertInstancesController.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(12, program12, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n  <div class=\"page-bar\">\n    <div class=\"filtered-info span4\">\n      <label>");
  stack1 = depth0;
  stack2 = "view.filteredContentInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " - <a ");
  stack1 = depth0;
  stack2 = "clearFilters";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          href=\"#\">");
  stack1 = depth0;
  stack2 = "tableView.filters.clearAllFilters";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></label>\n    </div>\n    <div class=\"selected-hosts-info span4\">\n    </div>\n    <div class=\"items-on-page\">\n      <label>");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.rowsPerPageSelectView";
  stack3 = {};
  stack4 = "view.displayLength";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n    </div>\n    <div class=\"info\">");
  stack1 = depth0;
  stack2 = "view.paginationInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n    <div class=\"paging_two_button\">\n      <a ");
  stack1 = {};
  stack2 = "view.paginationLeftClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "previousPage";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-arrow-left\"></i></a>\n      <a ");
  stack1 = {};
  stack2 = "view.paginationRightClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "nextPage";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-arrow-right\"></i></a>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/metrics", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div class=\"host-metrics\">\n    <table class=\"graphs\">\n        <tr>\n            <td>\n              ");
  stack1 = depth0;
  stack2 = "App.ChartHostMetricsCPU";
  stack3 = {};
  stack4 = "view.content";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </td>\n            <td>\n              ");
  stack1 = depth0;
  stack2 = "App.ChartHostMetricsDisk";
  stack3 = {};
  stack4 = "view.content";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </td>\n        </tr>\n        <tr>\n            <td>\n              ");
  stack1 = depth0;
  stack2 = "App.ChartHostMetricsLoad";
  stack3 = {};
  stack4 = "view.content";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </td>\n            <td>\n              ");
  stack1 = depth0;
  stack2 = "App.ChartHostMetricsMemory";
  stack3 = {};
  stack4 = "view.content";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </td>\n        </tr>\n        <tr>\n            <td>\n              ");
  stack1 = depth0;
  stack2 = "App.ChartHostMetricsNetwork";
  stack3 = {};
  stack4 = "view.content";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </td>\n            <td>\n              ");
  stack1 = depth0;
  stack2 = "App.ChartHostMetricsProcesses";
  stack3 = {};
  stack4 = "view.content";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </td>\n        </tr>\n    </table>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/stack_versions", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "view.parentView.stackSort";
  stack3 = {};
  stack4 = "first";
  stack3['classNames'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "view.parentView.repoVersionSort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "view.parentView.statusSort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    <th></th>\n  ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = helpers.each;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <tr>\n        <td class=\"first stack-name\">");
  stack1 = depth0;
  stack2 = "stack";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td class=\"repo-version\">");
  stack1 = depth0;
  stack2 = "displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n        <td class=\"host-stack-version-status\">\n          ");
  stack1 = depth0;
  stack2 = "isCurrent";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(7, program7, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </td>\n        <td class=\"install-repo-version align-center\">\n          ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </td>\n      </tr>\n    ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <span class=\"label label-success\">");
  stack1 = depth0;
  stack2 = "common.current";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n          ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "isInstalling";
  stack3 = helpers['if'];
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(10, program10, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n              <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "";
  stack3 = depth0;
  stack4 = "showInstallProgress";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-cog in_progress\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "displayStatus";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n            ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "displayStatus";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "isOutOfSync";
  stack3 = helpers['if'];
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "view.outOfSyncInfo";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  return buffer;}

function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <button class=\"btn\" ");
  stack1 = depth0;
  stack2 = "";
  stack3 = depth0;
  stack4 = "installVersionConfirmation";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "this.installDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-off\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "common.install";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n          ");
  return buffer;}

function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <tr>\n      <td colspan=\"4\" class=\"first\">\n        ");
  stack1 = depth0;
  stack2 = "hosts.host.stackVersions.table.noVersions";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n      </td>\n    </tr>\n  ");
  return buffer;}

  data.buffer.push("\n<table class=\"table advanced-header-table table-bordered table-striped\">\n  <thead>\n  ");
  stack1 = depth0;
  stack2 = "view.sortView";
  stack3 = {};
  stack4 = "label-row";
  stack3['classNames'] = stack4;
  stack4 = "view.filteredContent";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  <tr class=\"filter-row\">\n    <th class=\"first stack-name-filter\">");
  stack1 = depth0;
  stack2 = "view.stackFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n    <th class=\"repo-version-filter\">");
  stack1 = depth0;
  stack2 = "view.repoVersionFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n    <th class=\"status-filter\">");
  stack1 = depth0;
  stack2 = "view.statusFilterView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</th>\n    <th> </th>\n  </tr>\n  </thead>\n  <tbody>\n  ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(15, program15, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </tbody>\n</table>\n\n<div class=\"page-bar\">\n  <div class=\"filtered-info span4\">\n    <label>");
  stack1 = depth0;
  stack2 = "view.filteredContentInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " - <a ");
  stack1 = depth0;
  stack2 = "clearFilters";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      href=\"#\">");
  stack1 = depth0;
  stack2 = "tableView.filters.clearAllFilters";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></label>\n  </div>\n  <div class=\"items-on-page\">\n    <label>");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.rowsPerPageSelectView";
  stack3 = {};
  stack4 = "view.parentView.displayLength";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n  </div>\n  <div class=\"info\">");
  stack1 = depth0;
  stack2 = "view.paginationInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n  <div class=\"paging_two_button\">\n    ");
  stack1 = depth0;
  stack2 = "view.paginationLeft";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "view.paginationRight";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/summary", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n              <div class=\"btn-group pull-right\">\n                <button id=\"add_component\" data-toggle=\"dropdown\" ");
  stack1 = {};
  stack2 = "view.addComponentDisabled:disabled :btn :dropdown-toggle";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <span class=\"icon-plus\"></span>&nbsp;");
  stack1 = depth0;
  stack2 = "add";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                </button>\n                <ul class=\"dropdown-menu\">\n                  ");
  stack1 = depth0;
  stack2 = "view.addableComponents";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </ul>\n              </div>\n            ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                    <li>\n                      <a href=\"javascript:void(null)\" data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "component";
  stack3 = depth0;
  stack4 = "addComponentWithCheck";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                        ");
  stack1 = depth0;
  stack2 = "component.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                      </a>\n                    </li>\n                  ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n\n          ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n          ");
  stack1 = depth0;
  stack2 = "view.sortedComponents";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "view.content.componentsWithStaleConfigsCount";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              <div class=\"alert alert-warning clearfix\">\n                <i class=\"icon-refresh\"></i> ");
  stack1 = depth0;
  stack2 = "view.needToRestartMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  <button ");
  stack1 = {};
  stack2 = ":btn :restart-components :pull-right :btn-warning";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "restartAllStaleConfigComponents";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                    ");
  stack1 = depth0;
  stack2 = "hosts.host.details.needToRestart.button";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                  </button>\n              </div>\n            ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            <div class=\"row-fluid\">\n              ");
  stack1 = depth0;
  stack2 = "component.view";
  stack3 = {};
  stack4 = "component";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </div>\n          ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n                  <div class=\"span3 align-right\">");
  stack1 = depth0;
  stack2 = "common.clients";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "&nbsp;/</div>\n                <div class=\"span7\">\n                  ");
  stack1 = depth0;
  stack2 = "view.clients";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </div>\n              ");
  return buffer;}
function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "component.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "component.isInstallFailed";
  stack3 = helpers['if'];
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  stack1 = depth0;
  stack2 = "component.staleConfigs";
  stack3 = helpers['if'];
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  stack1 = depth0;
  stack2 = "component.service.isInPassive";
  stack3 = helpers['if'];
  tmp1 = self.program(16, program16, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  stack1 = depth0;
  stack2 = "component.isLast";
  stack3 = helpers.unless;
  tmp1 = self.program(18, program18, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  ");
  return buffer;}
function program12(depth0,data) {


  data.buffer.push("&nbsp;<span class=\"health-status-installed icon-warning-sign\"></span>");}

function program14(depth0,data) {


  data.buffer.push("&nbsp;<span class=\"text-warning icon-refresh\"></span>");}

function program16(depth0,data) {


  data.buffer.push("&nbsp;<span class=\"icon-medkit\"></span>");}

function program18(depth0,data) {


  data.buffer.push(",");}

function program20(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "view.clients.length";
  stack3 = helpers['if'];
  tmp1 = self.program(21, program21, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  return buffer;}
function program21(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n                  <div class=\"btn-group pull-right\">\n                    <button id=\"add_component\" data-toggle=\"dropdown\" ");
  stack1 = {};
  stack2 = ":btn :dropdown-toggle controller.content.isNotHeartBeating:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                      ");
  stack1 = depth0;
  stack2 = "common.installed";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                      <span class=\"caret pull-right\"></span>\n                    </button>\n                    <ul class=\"dropdown-menu\">\n                      <li>\n                        <a href=\"javascript:void(null)\" ");
  stack1 = {};
  stack2 = "view.areClientWithStaleConfigs::disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "view.clients";
  stack3 = depth0;
  stack4 = "refreshConfigs";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                          ");
  stack1 = depth0;
  stack2 = "hosts.host.details.refreshConfigs";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                        </a>\n                      </li>\n                      <li>\n                        <a href=\"javascript:void(null)\" ");
  stack1 = {};
  stack2 = "view.areClientsInstallFailed::disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "view.clients";
  stack3 = depth0;
  stack4 = "reinstallClients";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                          ");
  stack1 = depth0;
  stack2 = "host.host.details.installClients";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                        </a>\n                      </li>\n                      ");
  stack1 = depth0;
  stack2 = "view.clientsWithCustomCommands";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "option";
  stack7 = helpers.each;
  tmp1 = self.program(22, program22, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    </ul>\n                  </div>\n                ");
  return buffer;}
function program22(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n                        <li class=\"dropdown-submenu submenu-left\">\n                          <a href=\"javascript:void(null)\" >\n                            <i class=\"icon-play-circle\"></i>\n                            ");
  stack1 = depth0;
  stack2 = "option.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                          </a>\n                            <div class=\"dropdown-menu-wrap\">\n                              <ul class=\"dropdown-menu\">\n                                ");
  stack1 = depth0;
  stack2 = "option.commands";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "command";
  stack7 = helpers.each;
  tmp1 = self.program(23, program23, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                              </ul>\n                            </div>\n                        </li>\n                      ");
  return buffer;}
function program23(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                                  <li>\n                                    <a href=\"javascript:void(null)\" ");
  stack1 = depth0;
  stack2 = "command";
  stack3 = depth0;
  stack4 = "executeCustomCommand";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = true;
  stack5['href'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "command.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                                  </li>\n                                ");
  return buffer;}

function program25(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                          <dt class=\"summary-cur-ver-label\">");
  stack1 = depth0;
  stack2 = "hosts.host.summary.currentVersion";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</dt>\n                            <dd class=\"summary-cur-ver-value\">&nbsp;");
  stack1 = depth0;
  stack2 = "view.content.currentVersion";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</dd>\n                        ");
  return buffer;}

function program27(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <div class=\"span6\">\n        <div class=\"box\">\n          <div class=\"box-header\">\n            <h4>");
  stack1 = depth0;
  stack2 = "hosts.host.summary.hostMetrics";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h4>\n            ");
  stack1 = depth0;
  stack2 = "App.router.clusterController.isGangliaInstalled";
  stack3 = helpers['if'];
  tmp1 = self.program(28, program28, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n          <div>\n           ");
  stack1 = depth0;
  stack2 = "App.MainHostMetricsView";
  stack3 = {};
  stack4 = "view.content";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </div>\n        </div>\n      </div>\n    </div>\n  ");
  return buffer;}
function program28(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              <div class=\"btn-group\">\n                ");
  stack1 = depth0;
  stack2 = "App.router.clusterController.isGangliaUrlLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(29, program29, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(31, program31, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </div>\n            ");
  return buffer;}
function program29(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                  <a class=\"btn\" rel=\"tooltip\" title=\"Go to Ganglia\" ");
  stack1 = depth0;
  stack2 = "showGangliaCharts";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-link\"></i></a>\n                ");
  return buffer;}

function program31(depth0,data) {


  data.buffer.push("\n                  <div class=\"spinner\"></div>\n                ");}

  data.buffer.push("\n\n<div class=\"row-fluid host-block\">\n  <div class=\"span12\">\n    <div class=\"span6\">\n      ");
  data.buffer.push("\n      <div class=\"box\">\n        <div class=\"box-header\">\n          <div class=\"pull-left\">\n            <h4>");
  stack1 = depth0;
  stack2 = "common.components";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h4>\n          </div>\n          <div class=\"pull-right span5 row-fluid\" style=\"padding:0 10px;\">\n            ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n        </div>\n        <div class=\"host-components\">\n        ");
  stack1 = depth0;
  stack2 = "view.sortedComponents.length";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  data.buffer.push("\n          <div class=\"clients row-fluid\">\n            <div class=\"span7 row\">\n              ");
  stack1 = depth0;
  stack2 = "view.clients.length";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </div>\n            <div class=\"span5 row\">\n              ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(20, program20, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </div>\n          </div>\n        </div>\n      </div>\n      <div class=\"host-configuration\">\n            <div class=\"box\">\n                <div class=\"box-header\">\n                    <h4>");
  stack1 = depth0;
  stack2 = "hosts.host.summary.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h4>\n                </div>\n                <div class=\"host-summary-content\">\n                    <dl class=\"dl-horizontal host-details-summary-list\">\n\n                        <dt class=\"summary-hostname-label\">");
  stack1 = depth0;
  stack2 = "hosts.host.summary.hostname";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</dt>\n                          <dd class=\"summary-hostname-value\">&nbsp;");
  stack1 = depth0;
  stack2 = "view.content.hostName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</dd>\n\n                        <dt class=\"summary-ipaddress-label\">");
  stack1 = depth0;
  stack2 = "common.ipAddress";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</dt>\n                          <dd class=\"summary-ipaddress-value\">&nbsp;");
  stack1 = depth0;
  stack2 = "view.content.ip";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</dd>\n\n                        <dt class=\"summary-os-label\">");
  stack1 = depth0;
  stack2 = "common.os";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</dt>\n                          <dd class=\"summary-os-value\">&nbsp;");
  stack1 = depth0;
  stack2 = "view.content.osType";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "&nbsp;(");
  stack1 = depth0;
  stack2 = "view.content.osArch";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ")</dd>\n\n                        <dt class=\"summary-cores-label\">");
  stack1 = depth0;
  stack2 = "common.cores";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</dt>\n                          <dd class=\"summary-cores-value\">&nbsp;");
  stack1 = depth0;
  stack2 = "view.content.coresFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</dd>\n\n                        <dt class=\"summary-disk-label\">");
  stack1 = depth0;
  stack2 = "common.disk";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</dt>\n                          <dd class=\"summary-disk-value\">&nbsp;");
  stack1 = depth0;
  stack2 = "view.content.diskInfoBar";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</dd>\n\n                        <dt class=\"summary-memory-label\">");
  stack1 = depth0;
  stack2 = "common.memory";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</dt>\n                          <dd class=\"summary-memory-value\">&nbsp;");
  stack1 = depth0;
  stack2 = "view.content.memoryFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</dd>\n\n                        <dt class=\"summary-load-avg-label\">");
  stack1 = depth0;
  stack2 = "common.loadAvg";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</dt>\n                          <dd class=\"summary-load-avg-value\">&nbsp;");
  stack1 = depth0;
  stack2 = "view.content.loadAvg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</dd>\n\n                        <dt class=\"summary-agent-heart-label\">");
  stack1 = depth0;
  stack2 = "hosts.host.summary.agentHeartbeat";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</dt>\n                          <dd class=\"summary-agent-heart-value\">&nbsp;");
  stack1 = depth0;
  stack2 = "view.timeSinceHeartBeat";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</dd>\n\n                        ");
  stack1 = depth0;
  stack2 = "App.supports.stackUpgrade";
  stack3 = helpers['if'];
  tmp1 = self.program(25, program25, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    </dl>\n                </div>\n            </div>\n        </div>\n    </div>\n  ");
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.isNoHostMetricsService";
  stack3 = helpers.unless;
  tmp1 = self.program(27, program27, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/host/version_filter", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


  data.buffer.push("\n\n\n<button class=\"btn single-btn-group filter-btn\" id=\"host-version-filter-button\" ");
  stack1 = depth0;
  stack2 = "clickFilterButton";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  ");
  stack1 = depth0;
  stack2 = "common.filter";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  <i class=\"icon-filter\"></i>\n</button>\n<ul class=\"dropdown-menu filter-components\">\n  <li>");
  stack1 = depth0;
  stack2 = "view.versionSelectView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</li>\n  <li>");
  stack1 = depth0;
  stack2 = "view.statusSelectView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</li>\n  <li id=\"button-bar\">\n    <button class=\"btn btn-inverse\" ");
  stack1 = depth0;
  stack2 = "applyFilter";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.apply";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n  </li>\n</ul>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/menu", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n    <li ");
  stack1 = {};
  stack2 = "active";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ><a ");
  stack1 = depth0;
  stack2 = "routing";
  stack3 = depth0;
  stack4 = "navigate";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
  stack1 = depth0;
  stack2 = "label";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n    ");
  return buffer;}

  data.buffer.push("\n\n<ul class=\"nav nav-tabs\">\n    ");
  stack1 = depth0;
  stack2 = "view.items";
  stack3 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</ul>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/menu_item", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <span ");
  stack1 = {};
  stack2 = ":label view.hasCriticalAlerts:alerts-crit-count:alerts-warn-count";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </span>\n  ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "App.TopNavServiceMenuView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n  <ul class=\"top-nav-dropdown-menu\">\n    ");
  stack1 = depth0;
  stack2 = "view.dropdownCategories";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "category";
  stack7 = helpers.each;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </ul>\n");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.AdminDropdownItemView";
  stack3 = {};
  stack4 = "category.name";
  stack3['itemBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n        <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "category.url";
  stack3 = depth0;
  stack4 = "goToCategory";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "category.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n      ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  <ul class=\"top-nav-dropdown-menu\">\n    ");
  stack1 = depth0;
  stack2 = "view.content.views.length";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(13, program13, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </ul>\n");
  return buffer;}
function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.content.views";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "category";
  stack7 = helpers.each;
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n        <li><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "category";
  stack3 = depth0;
  stack4 = "setView";
  stack5 = {};
  stack6 = "App.router.mainViewsController";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "category.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n      ");
  return buffer;}

function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <li class=\"disabled\"><a href=\"javascript:void(null);\">");
  stack1 = depth0;
  stack2 = "menu.item.views.noViews";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n    ");
  return buffer;}

  data.buffer.push("\n\n\n<a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.content.routing";
  stack3 = depth0;
  stack4 = "goToSection";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  ");
  stack1 = depth0;
  stack2 = "view.content.label";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.hasAlertsLabel";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</a>\n\n<!--dropdown menu for the items had dropdowns-->\n");
  stack1 = depth0;
  stack2 = "view.isServicesItem";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.isAdminItem";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.isViewsItem";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<div class=\"row-fluid\">\n  <div class=\"services-menu well span2 service-menu-width\" style=\"padding: 8px 0\">\n    ");
  stack1 = depth0;
  stack2 = "App.MainServiceMenuView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "App.AllServicesActionView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <div class=\"span10 summary-width\">\n    ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/add", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                <li ");
  stack1 = {};
  stack2 = "isStep5:active view.isStep5Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep5";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "admin.kerberos.wizard.step4.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep6:active view.isStep6Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep5";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step8.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep7:active view.isStep7Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep6";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step9.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep8:active view.isStep8Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep7";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step10.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                <li ");
  stack1 = {};
  stack2 = "isStep5:active view.isStep5Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep5";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step8.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep6:active view.isStep6Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep6";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step9.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                <li ");
  stack1 = {};
  stack2 = "isStep7:active view.isStep7Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep7";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step10.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"wizard\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n\n      <!--<a class=\"btn back\" ");
  stack1 = depth0;
  stack2 = "backToServices";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">← Back to Services</a>-->\n\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li class=\"nav-header\">");
  stack1 = depth0;
  stack2 = "services.add.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n              <li ");
  stack1 = {};
  stack2 = "isStep1:active view.isStep1Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep1";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step4.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep2:active view.isStep2Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep2";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step5.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep3:active view.isStep3Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep3";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step6.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep4:active view.isStep4Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep4";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step7.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              ");
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "App.supports.automatedKerberos";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </ul>\n          </div>\n        </div>\n          <div class=\"wizard-content well span9\">\n            ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/add_host_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "selectHostMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "<br />\n  <div>\n    ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "hostsWithoutComponent";
  stack3['contentBinding'] = stack4;
  stack4 = "selectedHost";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n\n  ");
  stack1 = depth0;
  stack2 = "addComponentMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "<br /><br />\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "thereIsNoHostsMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "anyHostsWithoutComponent";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/all_services_actions", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div class=\"service-button\" style=\"margin-top: 10px;\">\n    <div class=\"btn-group display-inline-block span11 offset1\">\n      <a class=\"btn dropdown-toggle span10\" data-toggle=\"dropdown\" href=\"#\">");
  stack1 = depth0;
  stack2 = "common.actions";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        <span class=\"caret\"></span>\n      </a>\n      <ul class=\"pull-left dropdown-menu\">\n          <li ");
  stack1 = {};
  stack2 = "controller.isAllServicesInstalled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"#\"\n              ");
  stack1 = {};
  stack2 = "controller.isAllServicesInstalled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "gotoAddService";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <i class=\"icon-plus icon-white\"></i> ");
  stack1 = depth0;
  stack2 = "services.service.add";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n          </li>\n          <li class=\"divider\"></li>\n          <li ");
  stack1 = {};
  stack2 = "controller.isStartAllDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"#\" data-toggle=\"modal\"\n              ");
  stack1 = {};
  stack2 = "controller.isStartAllDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "startAllService";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <i ");
  stack1 = {};
  stack2 = ":icon-play controller.isStartAllDisabled:disabled:enabled ";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n              ");
  stack1 = depth0;
  stack2 = "services.service.startAll";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n          <li ");
  stack1 = {};
  stack2 = "controller.isStopAllDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"#\" data-toggle=\"modal\"\n              ");
  stack1 = {};
  stack2 = "controller.isStopAllDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "stopAllService";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <i ");
  stack1 = {};
  stack2 = ":icon-stop controller.isStopAllDisabled:disabled:enabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n              ");
  stack1 = depth0;
  stack2 = "services.service.stopAll";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n      </ul>\n    </div>\n  </div>\n");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/info/configs", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "controller.content.isRestartRequired";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <div class=\"clearfix\"></div>\n    ");
  stack1 = depth0;
  stack2 = "App.ServiceConfigView";
  stack3 = {};
  stack4 = "controller.filter";
  stack3['filterBinding'] = stack4;
  stack4 = "controller.filterColumns";
  stack3['columnsBinding'] = stack4;
  stack4 = "controller.canEdit";
  stack3['canEditBinding'] = stack4;
  stack4 = "controller.showConfigHistoryFeature";
  stack3['showConfigHistoryFeatureBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n        <div>\n          <div class=\"alert alert-warning clearfix\">\n            <i class=\"icon-refresh\"></i> ");
  stack1 = depth0;
  stack2 = "services.service.config.restartService.needToRestart";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "  <a\n                  href=\"#\" ");
  stack1 = depth0;
  stack2 = "controller.content.restartRequiredHostsAndComponents";
  stack3 = depth0;
  stack4 = "showComponentsShouldBeRestarted";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.componentsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.componentsCount";
  stack3 = {};
  stack4 = "t:common.component";
  stack3['singular'] = stack4;
  stack4 = "t:common.components";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "</a> ");
  stack1 = depth0;
  stack2 = "on";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "controller.content.restartRequiredHostsAndComponents";
  stack3 = depth0;
  stack4 = "showHostsShouldBeRestarted";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.hostsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.hostsCount";
  stack3 = {};
  stack4 = "t:common.host";
  stack3['singular'] = stack4;
  stack4 = "t:common.hosts";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n            <span class=\"restart-components pull-right\">&nbsp</span>\n\n            <div class=\"btn-group pull-right\">\n              <button type=\"button\" class=\"btn btn-default dropdown-toggle btn-warning\" data-toggle=\"dropdown\">\n                ");
  stack1 = depth0;
  stack2 = "hosts.host.details.needToRestart.button";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                <span class=\"caret\"></span>\n              </button>\n              <ul class=\"dropdown-menu\">\n                <li>\n                  <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "restartAllStaleConfigComponents";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "restart.service.all";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                </li>\n                ");
  stack1 = depth0;
  stack2 = "view.rollingRestartSlaveComponentName";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </ul>\n            </div>\n          </div>\n        </div>\n      ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                  <li>\n                    <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.rollingRestartSlaveComponentName";
  stack3 = depth0;
  stack4 = "rollingRestartStaleConfigSlaveComponents";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.rollingRestartActionName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                  </li>\n                ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "showConfigHistoryFeature";
  stack3 = helpers.unless;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          <p class=\"pull-right\">\n              <!--<input class=\"btn btn-primary\" type=\"button\" value=\"Save\" ");
  data.buffer.push(" />-->\n              <a class=\"btn\" ");
  stack1 = depth0;
  stack2 = "doCancel";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.cancel";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n              <a class=\"btn btn-primary\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                ");
  stack1 = depth0;
  stack2 = "restartServicePopup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.save";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n          </p>\n      ");
  return buffer;}

function program9(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

  data.buffer.push("\n\n<div id=\"serviceConfig\">\n  ");
  stack1 = depth0;
  stack2 = "dataIsLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(9, program9, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/info/configs_save_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n  <div class=\"pre-scrollable\" style=\"max-height: 250px;\">\n    <ul>\n    ");
  stack1 = depth0;
  stack2 = "view.getDisplayMessage";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "val";
  stack7 = helpers.each;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </ul>\n  </div>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <li>\n        ");
  stack1 = depth0;
  stack2 = "val";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </li>\n    ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.runningHosts";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.unknownHosts";
  stack3 = helpers['if'];
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n      <i class=\"icon-warning-sign\"></i>  ");
  stack1 = depth0;
  stack2 = "view.runningHostsMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      <table class=\"table-striped running-host-components-table\">\n          <tr><th>");
  stack1 = depth0;
  stack2 = "common.host";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th><th>");
  stack1 = depth0;
  stack2 = "common.components";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th></tr>\n        ");
  stack1 = depth0;
  stack2 = "view.runningHosts";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "host";
  stack7 = helpers.each;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </table>\n  ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <tr><td>");
  stack1 = depth0;
  stack2 = "host.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td><td>");
  stack1 = depth0;
  stack2 = "host.components";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td></tr>\n        ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n      <i class=\"icon-warning-sign\"></i>  ");
  stack1 = depth0;
  stack2 = "view.unknownHostsMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      <table class=\"table-striped running-host-components-table\">\n          <tr><th>");
  stack1 = depth0;
  stack2 = "common.host";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th><th>");
  stack1 = depth0;
  stack2 = "common.components";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th></tr>\n        ");
  stack1 = depth0;
  stack2 = "view.unknownHosts";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "host";
  stack7 = helpers.each;
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </table>\n  ");
  return buffer;}
function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <tr><td>");
  stack1 = depth0;
  stack2 = "host.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td><td>");
  stack1 = depth0;
  stack2 = "host.components";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td></tr>\n        ");
  return buffer;}

function program11(depth0,data) {


  data.buffer.push("\n  <div class=\"spinner\"></div>\n");}

  data.buffer.push("\n\n<div ");
  stack1 = {};
  stack2 = "view.messageClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.message";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n");
  stack1 = depth0;
  stack2 = "view.flag";
  stack3 = helpers.unless;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(11, program11, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/info/metric_graphs", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    <tr>\n      ");
  stack1 = depth0;
  stack2 = "graphs";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "graph";
  stack7 = helpers.each;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </tr>\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <td>\n          <div class=\"\">\n            ");
  stack1 = depth0;
  stack2 = "graph";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </div>\n        </td>\n      ");
  return buffer;}

  data.buffer.push("\n<table class=\"graphs\">\n  ");
  stack1 = depth0;
  stack2 = "view.serviceMetricGraphs";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "graphs";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</table>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/info/save_popup_footer", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;


  data.buffer.push("\n\n<div class=\"modal-footer\">\n  <button type=\"button\" class=\"btn btn\" ");
  stack1 = depth0;
  stack2 = "onCancel";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.cancel";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n  <button type=\"button\" class=\"btn btn\" ");
  stack1 = depth0;
  stack2 = "onDiscard";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.discard";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n  <button type=\"button\" class=\"btn btn-success\" ");
  stack1 = {};
  stack2 = "view.isSaveDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "onSave";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.save";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/info/service_alert_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n  <ul id='summary-alerts-popup' class=\"alerts\">\n    ");
  stack1 = depth0;
  stack2 = "view.alerts";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "alert";
  stack7 = helpers.each;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </ul>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n      <li class=\"alert-list-wrap\">\n        <div class=\"row-fluid\">\n          <div class=\"span7 name-text\">\n            <div>\n              <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "alert";
  stack3 = depth0;
  stack4 = "gotoAlertDetails";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "alert.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n            </div>\n            <div class=\"definition-latest-text\">\n              ");
  stack1 = depth0;
  stack2 = "alert.latestText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            </div>\n          </div>\n          <div class=\"span5 status-col\" rel=\"alert-status-tooltip\" ");
  stack1 = {};
  stack2 = "alert.lastTriggered";
  stack1['title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <span class=\"timeago\" ");
  stack1 = {};
  stack2 = "alert.lastTriggeredVerboseDisplay";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <span class=\"status-icon\">");
  stack1 = depth0;
  stack2 = "alert.status";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</span>\n              <time>");
  stack1 = depth0;
  stack2 = "alert.lastTriggeredForFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</time>\n            </span>\n          </div>\n        </div>\n      </li>\n    ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "services.service.summary.alerts.noAlerts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.alerts.length";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/info/summary", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n    <div id=\"summary-restart-bar\">\n      <div class=\"alert alert-warning clearfix\">\n        <i class=\"icon-refresh\"></i> ");
  stack1 = depth0;
  stack2 = "view.needToRestartMessage";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push(" ");
  stack1 = depth0;
  stack2 = "services.service.config.restartService.needToRestart";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "controller.content.restartRequiredHostsAndComponents";
  stack3 = depth0;
  stack4 = "showComponentsShouldBeRestarted";
  stack5 = {};
  stack6 = "App.router.mainServiceInfoConfigsController";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.componentsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.componentsCount";
  stack3 = {};
  stack4 = "t:common.component";
  stack3['singular'] = stack4;
  stack4 = "t:common.components";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "</a> ");
  stack1 = depth0;
  stack2 = "on";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "controller.content.restartRequiredHostsAndComponents";
  stack3 = depth0;
  stack4 = "showHostsShouldBeRestarted";
  stack5 = {};
  stack6 = "App.router.mainServiceInfoConfigsController";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.hostsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.hostsCount";
  stack3 = {};
  stack4 = "t:common.host";
  stack3['singular'] = stack4;
  stack4 = "t:common.hosts";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n        <span class=\"restart-components pull-right\">&nbsp</span>\n\n        <div class=\"btn-group pull-right\">\n          <button type=\"button\" class=\"btn btn-default dropdown-toggle btn-warning\" data-toggle=\"dropdown\">\n            ");
  stack1 = depth0;
  stack2 = "hosts.host.details.needToRestart.button";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            <span class=\"caret\"></span>\n          </button>\n          <ul class=\"dropdown-menu\">\n            <li><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "restartAllStaleConfigComponents";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "restart.service.all";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n            ");
  stack1 = depth0;
  stack2 = "view.rollingRestartSlaveComponentName";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </ul>\n        </div>\n      </div>\n    </div>\n  ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n              <li>\n                <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.rollingRestartSlaveComponentName";
  stack3 = depth0;
  stack4 = "rollingRestartStaleConfigSlaveComponents";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.rollingRestartActionName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n              </li>\n            ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <span ");
  stack1 = depth0;
  stack2 = "controller.content";
  stack3 = depth0;
  stack4 = "showServiceAlertsPopup";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = {};
  stack2 = ":label :pull-right view.hasCriticalAlerts:alerts-crit-count:alerts-warn-count";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = {};
  stack4 = "alert";
  stack3['singular'] = stack4;
  stack4 = "alerts";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n          ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <span ");
  stack1 = depth0;
  stack2 = "controller.content";
  stack3 = depth0;
  stack4 = "showServiceAlertsPopup";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              class=\"label pull-right no-alerts-label\">");
  stack1 = depth0;
  stack2 = "services.service.summary.alerts.noAlerts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n          ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "view.serviceSummaryView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  return buffer;}

function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "view.serviceMetricGraphs.length";
  stack3 = helpers['if'];
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n      <div class=\"span6 service-metrics-block\">\n        <div class=\"box\">\n          <div class=\"box-header\">\n            <h4>");
  stack1 = depth0;
  stack2 = "services.service.metrics";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h4>\n            <div class=\"btn-group pull-right\">\n              <button type=\"button\" class=\"btn btn-default dropdown-toggle\" data-toggle=\"dropdown\">\n                ");
  stack1 = depth0;
  stack2 = "view.currentTimeRange.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " &nbsp;<span class=\"caret\"></span>\n              </button>\n              <ul class=\"dropdown-menu\">\n                ");
  stack1 = depth0;
  stack2 = "view.timeRangeOptions";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "option";
  stack7 = helpers.each;
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </ul>\n            </div>\n          </div>\n          <div class=\"\">\n            <table class=\"graphs\">\n              ");
  stack1 = depth0;
  stack2 = "view.serviceMetricGraphs";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "graphs";
  stack7 = helpers.each;
  tmp1 = self.program(16, program16, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </table>\n          </div>\n        </div>\n      </div>\n    ");
  return buffer;}
function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                    <li><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "option";
  stack3 = depth0;
  stack4 = "setTimeRange";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "option.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                ");
  return buffer;}

function program16(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n                <tr>\n                  ");
  stack1 = depth0;
  stack2 = "graphs";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "graph";
  stack7 = helpers.each;
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </tr>\n              ");
  return buffer;}
function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    <td>\n                        <div class=\"\">\n                          ");
  stack1 = depth0;
  stack2 = "graph";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                        </div>\n                    </td>\n                  ");
  return buffer;}

function program19(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n  <div class=\"row-fluid\">\n    <div class=\"span12 metrics-collapsed-graphs\">\n      ");
  stack1 = depth0;
  stack2 = "view.collapsedSections";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "collapsedSection";
  stack7 = helpers.each;
  tmp1 = self.program(20, program20, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n  </div>\n");
  return buffer;}
function program20(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <div ");
  stack1 = {};
  stack2 = ":accordion collapsedSection.toggleIndex";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <div class=\"accordion-group box\">\n            <div class=\"accordion-heading box-header\">\n              <span class=\"pull-left accordion-toggle\"\n                    data-toggle=\"collapse\" ");
  stack1 = {};
  stack2 = "collapsedSection.toggleIndex";
  stack1['dataParent'] = stack2;
  stack2 = "collapsedSection.toggleIndex";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                <i class=\"pull-left icon-caret-toggle\"></i>\n                <span>");
  stack1 = depth0;
  stack2 = "collapsedSection.header";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n              </span>\n              <a class=\"btn pull-right\" ");
  stack1 = {};
  stack2 = "collapsedSection.url";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " target=\"_blank\">\n                <i class=\"icon-link\"></i>\n              </a>\n\n              <div class=\"clearfix\"></div>\n            </div>\n            <div class=\"accordion-body collapse in\" ");
  stack1 = {};
  stack2 = "collapsedSection.id";
  stack1['id'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <div class=\"accordion-inner\">\n                ");
  stack1 = depth0;
  stack2 = "collapsedSection.metricView";
  stack3 = helpers['if'];
  tmp1 = self.program(21, program21, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </div>\n            </div>\n          </div>\n        </div>\n      ");
  return buffer;}
function program21(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "collapsedSection.metricView";
  stack3 = {};
  stack4 = "collapsedSection.metricViewData";
  stack3['viewDataBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.service.isRestartRequired";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n<div class=\"row-fluid service-block\">\n  <div class=\"span6\">\n    <div class=\"box\">\n      <div class=\"box-header summary-box-header\">\n        <h4>");
  stack1 = depth0;
  stack2 = "controller.content.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "services.service.info.menu.summary";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h4>\n        ");
  stack1 = depth0;
  stack2 = "view.hasAlertDefinitions";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n      <div class=\"service-content\">\n        <table id=\"summary-info\" class=\"table no-borders table-condensed\">\n          <tbody>\n          ");
  stack1 = depth0;
  stack2 = "view.serviceSummaryView";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </tbody>\n        </table>\n      </div>\n    </div>\n  </div>\n\n  ");
  stack1 = depth0;
  stack2 = "view.isServiceMetricLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n</div>\n\n");
  stack1 = depth0;
  stack2 = "view.collapsedSections";
  stack3 = helpers['if'];
  tmp1 = self.program(19, program19, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n\n\n\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/info/summary/base", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "App.SummaryMasterComponentsView";
  stack3 = {};
  stack4 = "view.content.mastersObj";
  stack3['mastersCompBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  stack1 = depth0;
  stack2 = "App.SummarySlaveComponentsView";
  stack3 = {};
  stack4 = "view.content.slavesObj";
  stack3['slavesObjBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  stack1 = depth0;
  stack2 = "App.SummaryClientComponentsView";
  stack3 = {};
  stack4 = "view.content.clientObj";
  stack3['clientsObjBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/info/summary/client_components", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n  <tr>\n    <td class=\"summary-label\">\n      <a ");
  stack1 = depth0;
  stack2 = "clientComponent";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"javascript:void(null)\" >\n        ");
  stack1 = depth0;
  stack2 = "clientComponent.displayNamePluralized";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    </td>\n    <td>\n      <span class=\"green-live\">");
  stack1 = depth0;
  stack2 = "clientComponent.installedCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span> ");
  stack1 = depth0;
  stack2 = "clientComponent.displayNamePluralized";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "common.installed";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    </td>\n  </tr>\n");
  return buffer;}

  data.buffer.push("\n<tr class=\"hidden\"><td></td></tr>\n");
  stack1 = depth0;
  stack2 = "view.clientsObj";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "clientComponent";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/info/summary/master_components", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n  <tr ");
  stack1 = {};
  stack2 = "comp.isSubComponent:component-small";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <td>\n      <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "comp.host";
  stack3 = depth0;
  stack4 = "showDetails";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " title=\"");
  stack1 = depth0;
  stack2 = "comp.host.publicHostName";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\" rel=\"UsageTooltip\">\n        ");
  stack1 = depth0;
  stack2 = "comp.displayNameAdvanced";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </a>\n    </td>\n    <td>\n      <span rel='healthTooltip' ");
  stack1 = {};
  stack2 = "comp.statusClass comp.statusIconClass";
  stack1['class'] = stack2;
  stack2 = "comp.passiveTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span>\n      ");
  stack1 = depth0;
  stack2 = "comp.componentTextStatus";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </td>\n  </tr>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "comp.displayNameAdvanced";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "comp.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  return buffer;}

  data.buffer.push("\n<tr class=\"hidden\"><td></td></tr>\n");
  stack1 = depth0;
  stack2 = "view.mastersComp";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "comp";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/info/summary/slave_components", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n  <tr>\n    <td><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "slaveComponent";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "slaveComponent.displayNamePluralized";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n    </td>\n    <td>\n      <span>\n        ");
  stack1 = depth0;
  stack2 = "App.ComponentLiveTextView";
  stack3 = {};
  stack4 = "slaveComponent.startedCount";
  stack3['liveComponentsBinding'] = stack4;
  stack4 = "slaveComponent.totalCount";
  stack3['totalComponentsBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </span>\n      ");
  stack1 = depth0;
  stack2 = "slaveComponent.displayNamePluralized";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "dashboard.services.summary.slaves.live";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    </td>\n  </tr>\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.liveComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "/");
  stack1 = depth0;
  stack2 = "view.totalComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  return buffer;}

  data.buffer.push("\n<tr class=\"hidden\"><td></td></tr>\n\n");
  stack1 = depth0;
  stack2 = "view.slavesObj";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "slaveComponent";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/item", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, blockHelperMissing=helpers.blockHelperMissing;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "App.QuickViewLinks";
  stack3 = {};
  stack4 = "view.service";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <ul class=\"nav nav-pills move\">\n          <li class=\"dropdown\">\n            <a class=\"dropdown-toggle\" id=\"quick-links-dropdown-btn\" data-toggle=\"dropdown\" href=\"#\">");
  stack1 = depth0;
  stack2 = "common.quickLinks";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "<b class=\"caret\"></b></a>\n            <ul class=\"dropdown-menu\">\n              ");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(11, program11, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n           </ul>\n          </li>\n        </ul>\n      ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "view.quickLinksArray";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n                  <!--there are multiple masters eg, HBase multiple masters or HDFS HA enabled-->\n                  ");
  stack1 = depth0;
  stack2 = "view.quickLinksArray";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "quickLinks";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    <li class=\"dropdown-submenu\">\n                      <a href=\"javascript:void(null)\">");
  stack1 = depth0;
  stack2 = "quickLinks.publicHostNameLabel";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " &nbsp;</a>\n                      <ul class=\"dropdown-menu\">\n                        ");
  stack1 = depth0;
  stack2 = "quickLinks";
  stack3 = helpers.each;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                      </ul>\n                    </li>\n                  ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                          <li><a ");
  stack1 = {};
  stack2 = "url";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.linkTarget";
  stack1['target'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                        ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "view.quickLinks";
  stack3 = helpers.each;
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    <li><a ");
  stack1 = {};
  stack2 = "url";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "view.linkTarget";
  stack1['target'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n                  ");
  return buffer;}

function program11(depth0,data) {


  data.buffer.push("\n                <div class=\"spinner\"></div>\n              ");}

function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n<div class=\"service-button\">\n  ");
  stack1 = depth0;
  stack2 = "view.isMaintenanceActive";
  stack3 = helpers['if'];
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;}
function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    <div class=\"btn-group display-inline-block\">\n      <a class=\"btn dropdown-toggle\" id=\"service-actions-dropdown-btn\" data-toggle=\"dropdown\" href=\"#\">\n        ");
  stack1 = depth0;
  stack2 = "services.service.actions.serviceActions";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        <span class=\"caret\"></span>\n      </a>\n      <ul class=\"pull-right dropdown-menu\">\n        <!-- dropdown menu links -->\n\n        <!-- Start/Stop service actions -->\n        ");
  stack1 = depth0;
  stack2 = "controller.isClientsOnlyService";
  stack3 = helpers.unless;
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n        <!-- Other service actions -->\n        ");
  stack1 = depth0;
  stack2 = "view.maintenance";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "option";
  stack7 = helpers.each;
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </ul>\n    </div>\n  ");
  return buffer;}
function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          <li ");
  stack1 = {};
  stack2 = "controller.isStartDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"javascript:void(null)\" ");
  stack1 = {};
  stack2 = "controller.isStartDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "startService";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <i ");
  stack1 = {};
  stack2 = ":icon-play controller.isStartDisabled:disabled:enabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n              ");
  stack1 = depth0;
  stack2 = "services.service.start";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n          <li ");
  stack1 = {};
  stack2 = "controller.isStopDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <a href=\"javascript:void(null)\" ");
  stack1 = {};
  stack2 = "controller.isStopDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "stopService";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <i ");
  stack1 = {};
  stack2 = ":icon-stop controller.isStopDisabled:disabled:enabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n              ");
  stack1 = depth0;
  stack2 = "services.service.stop";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          </li>\n        ");
  return buffer;}

function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n        <li ");
  stack1 = {};
  stack2 = "option.disabled option.isHidden:hidden option.hasSubmenu:dropdown-submenu option.hasSubmenu:submenu-left";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <a ");
  stack1 = depth0;
  stack2 = "option";
  stack3 = depth0;
  stack4 = "doAction";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = true;
  stack5['href'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "option.tooltip";
  stack1['data-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " rel=\"HealthTooltip\">\n              <i ");
  stack1 = {};
  stack2 = "option.cssClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n              ");
  stack1 = depth0;
  stack2 = "option.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </a>\n          ");
  stack1 = depth0;
  stack2 = "option.hasSubmenu";
  stack3 = helpers['if'];
  tmp1 = self.program(18, program18, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </li>\n        ");
  return buffer;}
function program18(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n            <div class=\"dropdown-menu-wrap\">\n              <ul class=\"dropdown-menu\">\n                ");
  stack1 = depth0;
  stack2 = "option.submenuOptions";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "item";
  stack7 = helpers.each;
  tmp1 = self.program(19, program19, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </ul>\n            </div>\n          ");
  return buffer;}
function program19(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                  <li>\n                    <a ");
  stack1 = depth0;
  stack2 = "item";
  stack3 = depth0;
  stack4 = "doAction";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = true;
  stack5['href'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "item.context.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n                  </li>\n                ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "App.MainServiceInfoMenuView";
  stack3 = {};
  stack4 = "view.hasConfigTab";
  stack3['configTabBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  <div class=\"span3 quick-links-wrapper\">\n    ");
  stack1 = depth0;
  stack2 = "view.service.quickLinks.length";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n\n");
  stack1 = depth0;
  stack2 = "ADMIN";
  foundHelper = helpers.isAccessible;
  stack3 = foundHelper || depth0.isAccessible;
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  if(foundHelper && typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else { stack1 = blockHelperMissing.call(depth0, stack3, stack2, tmp1); }
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/manage_configuration_groups_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div class=\"row-fluid  manage-configuration-group-content\">\n        <div class=\"span12\">\n            <div class=\"row-fluid\">\n                <div class=\"span4\">\n                    <span>&nbsp;</span>\n                  ");
  stack1 = depth0;
  stack2 = "Em.Select";
  stack3 = {};
  stack4 = "configGroups";
  stack3['contentBinding'] = stack4;
  stack4 = "content.displayNameHosts";
  stack3['optionLabelPath'] = stack4;
  stack4 = "view.selectedConfigGroup";
  stack3['selectionBinding'] = stack4;
  stack4 = "multiple";
  stack3['multiple'] = stack4;
  stack4 = "group-select";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                    <div class=\"btn-toolbar pull-right\">\n                        <button rel=\"button-info\" class=\"btn\" ");
  stack1 = {};
  stack2 = "view.addButtonTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                          ");
  stack1 = depth0;
  stack2 = "addConfigGroup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-plus\"></i></button>\n                        <button rel=\"button-info\" class=\"btn\" ");
  stack1 = {};
  stack2 = "view.removeButtonTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = "view.isRemoveButtonDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                          ");
  stack1 = depth0;
  stack2 = "confirmDelete";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-minus\"></i></button>\n                        <div class=\"btn-group\">\n                            <button class=\"btn dropdown-toggle\" data-toggle=\"dropdown\">\n                                <i class=\"icon-cog\"></i>&nbsp;<span class=\"caret\"></span>\n                            </button>\n                            <ul class=\"dropdown-menu\">\n                                <li ");
  stack1 = {};
  stack2 = "view.isRenameButtonDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                                    <a href=\"\" rel=\"button-info-dropdown\" ");
  stack1 = {};
  stack2 = "view.renameButtonTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "renameConfigGroup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "services.service.config_groups_popup.rename";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                                </li>\n                                <li ");
  stack1 = {};
  stack2 = "view.isDuplicateButtonDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                                    <a href=\"\" rel=\"button-info-dropdown\" ");
  stack1 = {};
  stack2 = "view.duplicateButtonTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "duplicateConfigGroup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "services.service.config_groups_popup.duplicate";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                                </li>\n                            </ul>\n                        </div>\n                    </div>\n                </div>\n                <div class=\"span8\">\n                    <span>&nbsp;</span>\n                    <div class=\"row-fluid\">\n                        <div class=\"span12 pull-right\">\n                          ");
  stack1 = depth0;
  stack2 = "Em.Select";
  stack3 = {};
  stack4 = "selectedConfigGroup.hosts";
  stack3['contentBinding'] = stack4;
  stack4 = "multiple";
  stack3['multiple'] = stack4;
  stack4 = "group-select";
  stack3['class'] = stack4;
  stack4 = "selectedHosts";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                        </div>\n                        <div class=\"button-group pull-right\">\n                            <a rel=\"button-info\" ");
  stack1 = {};
  stack2 = "view.addHostTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = ":btn selectedConfigGroup.isAddHostsDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "addHosts";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-plus\"></i></a>\n                            <a rel=\"button-info\" ");
  stack1 = {};
  stack2 = "view.removeHostTooltip";
  stack1['data-original-title'] = stack2;
  stack2 = ":btn isDeleteHostsDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "deleteHosts";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ><i class=\"icon-minus\"></i></a>\n                        </div>\n                    </div>\n                    <div class=\"row-fluid\">\n                        <div class=\"span2\">");
  stack1 = depth0;
  stack2 = "common.overrides";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n                        <div class=\"span10\">\n                            <a href=\"\" class=\"properties-link\" ");
  stack1 = depth0;
  stack2 = "showProperties";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                              ");
  stack1 = {};
  stack2 = "selectedConfigGroup.propertiesList";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "selectedConfigGroup.properties.length";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "common.properties";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                        </div>\n                    </div>\n                    <div class=\"clearfix\"></div>\n                    <div class=\"row-fluid\">\n                        <div class=\"span2\"><span class=\"prrl\">");
  stack1 = depth0;
  stack2 = "common.description";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span></div>\n                        <div class=\"span10 group-description\">");
  stack1 = depth0;
  stack2 = "selectedConfigGroup.description";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                    </div>\n                </div>\n                <div class=\"clearfix\"></div>\n                <div class=\"row-fluid\">\n                    <div class=\"span12 text-error\" id=\"manage-config-group-error-div\">\n                      ");
  stack1 = depth0;
  stack2 = "errorMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(4, program4, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    </div>\n                </div>\n            </div>\n        </div>\n    </div>\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                        ");
  stack1 = depth0;
  stack2 = "errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                      ");
  return buffer;}

function program4(depth0,data) {


  data.buffer.push("\n                          &nbsp;\n                      ");}

function program6(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

  data.buffer.push("\n<div class=\"alert alert-info margin-bottom-5\">");
  stack1 = depth0;
  stack2 = "services.service.config_groups_popup.notice";
  stack3 = {};
  stack4 = "view.displayName";
  stack3['serviceNameBinding'] = stack4;
  foundHelper = helpers['t'];
  stack4 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n  ");
  stack1 = depth0;
  stack2 = "controller.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(6, program6, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/menu_item", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <span ");
  stack1 = {};
  stack2 = ":label :alerts-count view.hasCriticalAlerts:alerts-crit-count:alerts-warn-count";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </span>\n  ");
  return buffer;}

  data.buffer.push("\n\n<a ");
  stack1 = {};
  stack2 = "view.link";
  stack1['href'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "App.MainDashboardServiceHealthView";
  stack3 = {};
  stack4 = "service-health";
  stack3['class'] = stack4;
  stack4 = "view.content";
  stack3['serviceBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "&nbsp;\n  <span>");
  stack1 = depth0;
  stack2 = "view.content.displayName";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n  <i rel=\"tooltip\" ");
  stack1 = depth0;
  stack2 = "goToConfigs";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = {};
  stack2 = ":icon-refresh :restart-required-service view.content.isRestartRequired::hidden";
  stack1['class'] = stack2;
  stack2 = "view.restartRequiredMessage";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n</a>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/new_config_group", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.parentView.warningMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  return buffer;}

function program3(depth0,data) {


  data.buffer.push("\n     &nbsp;\n    ");}

  data.buffer.push("\n\n<div class=\"new-config-group-div\">\n  <div class=\"text-warning\">\n    ");
  stack1 = depth0;
  stack2 = "view.parentView.warningMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  <table>\n    <tr>\n      <td class=\"title\">");
  stack1 = depth0;
  stack2 = "services.service.config_groups_popup.group_name_lable";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":&nbsp;</td>\n      <td>");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "255";
  stack3['maxlength'] = stack4;
  stack4 = "configGroupName";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n    </tr>\n    <tr>\n      <td class=\"title\">");
  stack1 = depth0;
  stack2 = "services.service.config_groups_popup.group_desc_lable";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":&nbsp;</td>\n      <td class=\"row-fluid\">\n        ");
  stack1 = depth0;
  stack2 = "Ember.TextArea";
  stack3 = {};
  stack4 = "span12";
  stack3['classNames'] = stack4;
  stack4 = "configGroupDesc";
  stack3['valueBinding'] = stack4;
  stack4 = "4";
  stack3['rows'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </td>\n    </tr>\n  </table>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/reassign", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              <li ");
  stack1 = {};
  stack2 = "isStep5:active view.isStep5Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep3";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "services.reassign.step5.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep6:active view.isStep6Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep4";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "services.reassign.step6.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "outlet";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program5(depth0,data) {


  data.buffer.push("\n                <div class=\"spinner\"></div>\n            ");}

  data.buffer.push("\n\n<div class=\"wizard\">\n  <div class=\"container\">\n    <div class=\"container-fluid\">\n      <div class=\"row-fluid\">\n        <div class=\"span3\">\n          <!--Sidebar content-->\n          <div class=\"well\">\n            <ul class=\"nav nav-pills nav-stacked\">\n              <li class=\"nav-header\">");
  stack1 = depth0;
  stack2 = "services.reassign.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</li>\n              <li ");
  stack1 = {};
  stack2 = "isStep1:active view.isStep1Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep1";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "services.reassign.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep2:active view.isStep2Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep2";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "services.reassign.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep3:active view.isStep3Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep3";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "services.reassign.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              <li ");
  stack1 = {};
  stack2 = "isStep4:active view.isStep4Disabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><a href=\"javascript:void(null);\"  ");
  stack1 = depth0;
  stack2 = "gotoStep4";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "services.reassign.step4.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></li>\n              ");
  stack1 = depth0;
  stack2 = "controller.content.hasManualSteps";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </ul>\n          </div>\n        </div>\n          <div class=\"wizard-content well span9\">\n            ");
  stack1 = depth0;
  stack2 = "view.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(5, program5, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n      </div>\n    </div>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/reassign/step1", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "message";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}

  data.buffer.push("\n\n<h2>");
  stack1 = depth0;
  stack2 = "services.reassign.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n<div class=\"alert alert-info\">\n  ");
  stack1 = depth0;
  stack2 = "view.message";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "message";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n<div class=\"btn-area\">\n  <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/reassign/step3", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div class=\"alert alert-warning\">\n    ");
  stack1 = depth0;
  stack2 = "view.jdbcSetupMessage";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n");
  return buffer;}

  data.buffer.push("\n\n<h2>");
  stack1 = depth0;
  stack2 = "services.reassign.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n<div class=\"alert alert-info\">\n  ");
  stack1 = depth0;
  stack2 = "services.reassign.step3.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n</div>\n");
  stack1 = depth0;
  stack2 = "view.jdbcSetupMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n<div id=\"step8-content\" class=\"well pre-scrollable\">\n  <div id=\"printReview\">\n    <a class=\"btn btn-info pull-right\" ");
  stack1 = depth0;
  stack2 = "printReview";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.print";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a> <br/>\n  </div>\n  <div id=\"step8-info\">\n    <p><b>");
  stack1 = depth0;
  stack2 = "services.reassign.step3.component";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</b> ");
  stack1 = depth0;
  stack2 = "controller.content.reassign.display_name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n\n    <p><b>");
  stack1 = depth0;
  stack2 = "services.reassign.step3.sourceHost";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</b> ");
  stack1 = depth0;
  stack2 = "view.sourceHost";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n\n    <p><b>");
  stack1 = depth0;
  stack2 = "services.reassign.step3.targetHost";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</b> ");
  stack1 = depth0;
  stack2 = "view.targetHost";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n  </div>\n</div>\n<div class=\"btn-area\">\n  <a class=\"btn pull-left\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = {};
  stack4 = "true";
  stack3['href'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n  <a class=\"btn btn-success pull-right\"\n     id=\"spinner\" ");
  stack1 = {};
  stack2 = "controller.isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.deploy";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/reassign/step4", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "templates/common/progress";
  stack3 = helpers.template;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/reassign/step5", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<h2>");
  stack1 = depth0;
  stack2 = "services.reassign.step5.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n");
  stack1 = depth0;
  stack2 = "view.manualCommands";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  data.buffer.push("\n  ");
  data.buffer.push("\n");
  data.buffer.push("\n<div class=\"btn-area\">\n  <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n</div>\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/reassign/step6", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "templates/common/progress";
  stack3 = helpers.template;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/reassign/step7", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "templates/common/progress";
  stack3 = helpers.template;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/reconfigure", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "command.isVisible";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <tr>\n        <td>\n          ");
  stack1 = depth0;
  stack2 = "App.StageLabelView";
  stack3 = {};
  stack4 = "command";
  stack3['commandBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </td>\n        <td>\n          ");
  stack1 = depth0;
  stack2 = "command.isCompleted";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(11, program11, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </td>\n        <td>\n          ");
  stack1 = depth0;
  stack2 = "command.isError";
  stack3 = helpers['if'];
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </td>\n      </tr>\n    ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <p ");
  stack1 = {};
  stack2 = "view.isStarted::faintText";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "command.index";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ".&nbsp");
  stack1 = depth0;
  stack2 = "command.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n          ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "command.isSuccess";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(8, program8, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "App.StageSuccessView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "command.isError";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "App.StageFailureView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  return buffer;}

function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "command.isStarted";
  stack3 = helpers['if'];
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "command.isPolling";
  stack3 = helpers['if'];
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                <div class=\"progress-bar pull-left\">\n                  ");
  stack1 = depth0;
  stack2 = "App.StageInProgressView";
  stack3 = {};
  foundHelper = helpers.command;
  stack4 = foundHelper || depth0.command;
  stack3['commandBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                </div>\n                <div class=\"progress-percentage pull-left\">");
  stack1 = depth0;
  stack2 = "command.progress";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "%</div>\n              ");
  return buffer;}

function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            <a class=\"btn btn-primary\"\n               href=\"javascript:void(null)\" ");
  stack1 = depth0;
  stack2 = "retry";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i\n                    class=\"icon-repeat icon-white\"></i>\n              ");
  stack1 = depth0;
  stack2 = "common.retry";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </a>\n          ");
  return buffer;}

  data.buffer.push("\n\n<table id=\"security-stages\" class=\"table no-borders\">\n  <tbody>\n  ");
  stack1 = depth0;
  stack2 = "controller.commands";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "command";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </tbody>\n</table>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/services/flume", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "view.parentView.hostSort";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            <th>");
  stack1 = depth0;
  stack2 = "dashboard.services.flume.agent";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n            <th><i class=\"icon-signin\" ");
  stack1 = {};
  stack2 = "dashboard.services.flume.sources";
  stack1['title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + "></i></th>\n            <th><i class=\"icon-random\" ");
  stack1 = {};
  stack2 = "dashboard.services.flume.channels";
  stack1['title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + "></i></th>\n            <th><i class=\"icon-signout\" ");
  stack1 = {};
  stack2 = "dashboard.services.flume.sinks";
  stack1['title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + "></i></th>\n          ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "host";
  stack7 = helpers.each;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n              <tbody class=\"flume-agent-table-tbody\">\n                ");
  stack1 = depth0;
  stack2 = "view.agentView";
  stack3 = {};
  stack4 = "host";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  stack1 = depth0;
  stack2 = "host.otherAgents";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "agent";
  stack7 = helpers.each;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </tbody>\n            ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                  <td ");
  stack1 = {};
  stack2 = "host.rowspan";
  stack1['rowspan'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " id=\"flume-host-agent-row\" class=\"agent-host-name\">\n                      <a class=\"agent-host-link\" href=\"javascript:void(null)\" ");
  stack1 = depth0;
  stack2 = "host";
  stack3 = depth0;
  stack4 = "click";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                        ");
  stack1 = depth0;
  stack2 = "host.hostName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                      </a>\n                  </td>\n                  <td class=\"agent-status\">\n                    <div class=\"wrapp-flume-status\">\n                      <div class=\"pull-left\"><span ");
  stack1 = {};
  stack2 = "host.firtstAgent.healthClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span> ");
  stack1 = depth0;
  stack2 = "host.firtstAgent.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                      <div class=\"btn-wrapper\">\n                        <div class=\"btn-group display-inline-block flume-agents-actions\">\n                            <a ");
  stack1 = {};
  stack2 = ":btn :dropdown-toggle";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " data-toggle=\"dropdown\" href=\"javascript:void(null)\">");
  stack1 = depth0;
  stack2 = "host.firtstAgent.displayStatus";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                                <span class=\"caret\"></span>\n                            </a>\n                            <ul class=\"pull-left dropdown-menu\">\n                                <li ");
  stack1 = {};
  stack2 = "host.firtstAgent.isStartAgentDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                                    <a href=\"javascript:void(null)\"\n                                      ");
  stack1 = {};
  stack2 = "host.firtstAgent.isStartAgentDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                                      ");
  stack1 = depth0;
  stack2 = "host.firtstAgent";
  stack3 = depth0;
  stack4 = "startFlumeAgent";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                                      ");
  stack1 = depth0;
  stack2 = "services.service.summary.flume.startAgent";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                                </li>\n                                <li ");
  stack1 = {};
  stack2 = "host.firtstAgent.isStopAgentDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                                    <a href=\"javascript:void(null)\"\n                                      ");
  stack1 = {};
  stack2 = "host.firtstAgent.isStopAgentDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                                      ");
  stack1 = depth0;
  stack2 = "host.firtstAgent";
  stack3 = depth0;
  stack4 = "stopFlumeAgent";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                                      ");
  stack1 = depth0;
  stack2 = "services.service.summary.flume.stopAgent";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                                </li>\n                            </ul>\n                        </div>\n                        </div>\n                    </div>\n                  </td>\n                  <td class=\"flume-sources-count\">\n                    ");
  stack1 = depth0;
  stack2 = "host.firtstAgent.sourcesCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  </td>\n                  <td class=\"flume-channels-count\">\n                    ");
  stack1 = depth0;
  stack2 = "host.firtstAgent.channelsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  </td>\n                  <td class=\"flume-sinks-count\">\n                    ");
  stack1 = depth0;
  stack2 = "host.firtstAgent.sinksCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                  </td>\n                ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                  ");
  stack1 = depth0;
  stack2 = "view.agentView";
  stack3 = {};
  stack4 = "host";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(8, program8, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                ");
  return buffer;}
function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                    <td class=\"agent-status\">\n                        <div class=\"wrapp-flume-status\">\n                          <div class=\"pull-left\"><span ");
  stack1 = {};
  stack2 = "agent.healthClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></span> ");
  stack1 = depth0;
  stack2 = "agent.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                          <div class=\"btn-wrapper\">\n                            <div class=\"btn-group display-inline-block flume-agents-actions\">\n                                <a ");
  stack1 = {};
  stack2 = ":btn :dropdown-toggle";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " data-toggle=\"dropdown\" href=\"javascript:void(null)\">");
  stack1 = depth0;
  stack2 = "agent.displayStatus";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                                    <span class=\"caret\"></span>\n                                </a>\n                                <ul class=\"pull-left dropdown-menu\">\n                                    <li ");
  stack1 = {};
  stack2 = "agent.isStartAgentDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                                        <a href=\"javascript:void(null)\"\n                                          ");
  stack1 = {};
  stack2 = "agent.isStartAgentDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                                          ");
  stack1 = depth0;
  stack2 = "agent";
  stack3 = depth0;
  stack4 = "startFlumeAgent";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                                          ");
  stack1 = depth0;
  stack2 = "services.service.summary.flume.startAgent";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                                    </li>\n                                    <li ");
  stack1 = {};
  stack2 = "agent.isStopAgentDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                                        <a href=\"javascript:void(null)\"\n                                          ");
  stack1 = {};
  stack2 = "agent.isStopAgentDisabled:disabled";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                                          ");
  stack1 = depth0;
  stack2 = "agent";
  stack3 = depth0;
  stack4 = "stopFlumeAgent";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                                          ");
  stack1 = depth0;
  stack2 = "services.service.summary.flume.stopAgent";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                                    </li>\n                                </ul>\n                              </div>\n                            </div>\n                        </div>\n                    </td>\n                    <td class=\"flume-agent-sources-count\">\n                      ");
  stack1 = depth0;
  stack2 = "agent.sourcesCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                    </td>\n                    <td class=\"flume-agent-channels-count\">\n                      ");
  stack1 = depth0;
  stack2 = "agent.channelsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                    </td>\n                    <td class=\"flume-agent-sink-count\">\n                      ");
  stack1 = depth0;
  stack2 = "agent.sinksCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                    </td>\n                  ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <tr>\n              <td colspan=\"6\" id=\"flume-agents-empty-label\" class=\"empty-label\">\n                ");
  stack1 = depth0;
  stack2 = "services.service.summary.flume.noAgents";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </td>\n            </tr>\n          ");
  return buffer;}

  data.buffer.push("\n\n<!-- Flume Agents -->\n    <div id=\"flume-summary\">\n      <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.flumeHandlerComponent";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.summaryHeader";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n      <a href=\"#\" class=\"pull-right\" ");
  stack1 = depth0;
  stack2 = "gotoConfigs";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "dashboard.services.flume.summary.configure";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n      <div id=\"flume-agent-table-wrap\" class=\"scrollable-container\">\n        <table class=\"table table-hover table-bordered table-striped\" id=\"flume-agents-table\">\n          <thead>\n          ");
  stack1 = depth0;
  stack2 = "view.sortView";
  stack3 = {};
  stack4 = "view.filteredContent";
  stack3['contentBinding'] = stack4;
  stack4 = "label-row";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </thead>\n          ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(10, program10, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </table>\n      </div>\n    </div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/services/hbase", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.liveComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "/");
  stack1 = depth0;
  stack2 = "view.totalComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  return buffer;}

  data.buffer.push("\n\n<!-- HBase Master Server -->\n  ");
  stack1 = depth0;
  stack2 = "view.dashboardMasterComponentView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n<!-- RegionServers -->\n<tr>\n  <td><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.regionServerComponent";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "dashboard.services.hbase.regionServers";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n  </td>\n  <td>\n    <span>\n      ");
  stack1 = depth0;
  stack2 = "App.ComponentLiveTextView";
  stack3 = {};
  stack4 = "view.service.regionServersStarted";
  stack3['liveComponentsBinding'] = stack4;
  stack4 = "view.service.regionServersTotal";
  stack3['totalComponentsBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </span>\n      ");
  stack1 = depth0;
  stack2 = "services.service.summary.RegionServersLIVE";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n<!-- Regions in Transition -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.hbase.regions.transition";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.service.regionsInTransition";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- HBase Master Started Time -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.hbase.masterStarted";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.masterStartedTime";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- HBase Master Activated Time -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.hbase.masterActivated";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.masterStartedTime";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- Average Load -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.hbase.averageLoad";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.averageLoad";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- Master Server Heap -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.hbase.masterServerHeap";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.masterServerHeapSummary";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/services/hdfs", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.liveComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "/");
  stack1 = depth0;
  stack2 = "view.totalComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n  <tr>\n    <td><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.journalNodeComponent";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.journalnodes";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    </td>\n    <td>\n      <span>\n        ");
  stack1 = depth0;
  stack2 = "App.ComponentLiveTextView";
  stack3 = {};
  stack4 = "view.journalNodesLive";
  stack3['liveComponentsBinding'] = stack4;
  stack4 = "view.journalNodesTotal";
  stack3['totalComponentsBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </span>\n        ");
  stack1 = depth0;
  stack2 = "services.service.summary.JournalNodesLive";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    </td>\n  </tr>\n");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.liveComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "/");
  stack1 = depth0;
  stack2 = "view.totalComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.dashboardMasterComponentView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n<!-- Data Nodes -->\n<tr>\n  <td><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.dataNodeComponent";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.datanodes";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></td>\n  <td>\n    <span>\n      ");
  stack1 = depth0;
  stack2 = "App.ComponentLiveTextView";
  stack3 = {};
  stack4 = "view.service.dataNodesStarted";
  stack3['liveComponentsBinding'] = stack4;
  stack4 = "view.service.dataNodesTotal";
  stack3['totalComponentsBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </span>\n      ");
  stack1 = depth0;
  stack2 = "common.started";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n<!-- Data Node Counts -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.datanodecounts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>\n    <span ");
  stack1 = {};
  stack2 = "dashboard.services.hdfs.datanode.status.tooltip.live";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + " rel=\"tooltip\">");
  stack1 = depth0;
  stack2 = "view.service.liveDataNodes.length";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.nodes.live";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " </span> /\n    <span ");
  stack1 = {};
  stack2 = "dashboard.services.hdfs.datanode.status.tooltip.dead";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + " rel=\"tooltip\">");
  stack1 = depth0;
  stack2 = "view.service.deadDataNodes.length";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.nodes.dead";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " </span> /\n    <span ");
  stack1 = {};
  stack2 = "dashboard.services.hdfs.datanode.status.tooltip.decommission";
  stack1['data-original-title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + " rel=\"tooltip\">");
  stack1 = depth0;
  stack2 = "view.service.decommissionDataNodes.length";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.nodes.decom";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n  </td>\n</tr>\n<!-- JournalNodes -->\n");
  stack1 = depth0;
  stack2 = "view.showJournalNodes";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n<!-- NameNode Uptime -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.nodes.uptime";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.nodeUptime";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- NameNode Heap -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.nodes.heap";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.nodeHeap";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- HDFS Capacity (Disk Usage)-->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.capacity.dfsUsed";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.dfsUsedDisk";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.capacity.nonDfsUsed";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.nonDfsUsedDisk";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.capacity.remaining";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.remainingDisk";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- Blocks Total -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "services.service.summary.blocksTotal";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.dfsTotalBlocks";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- Block Errors -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "services.service.summary.blockErrors";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>\n    ");
  stack1 = depth0;
  stack2 = "view.blockErrorsMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n<!-- Total Files And Directories -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.hdfs.totalFilesAndDirs";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.dfsTotalFiles";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- Upgrade Status -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "services.service.summary.pendingUpgradeStatus";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>\n    <span ");
  stack1 = {};
  stack2 = "view.isUpgradeStatusWarning:upgrade-status-warning";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.upgradeStatus";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n  </td>\n</tr>\n<!-- Safe Mode Status -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "services.service.summary.safeModeStatus";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>\n    ");
  stack1 = depth0;
  stack2 = "view.safeModeStatus";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/services/hive", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n      <a href=\"#\" class=\"label label-important\" ");
  stack1 = depth0;
  stack2 = "view.service";
  stack3 = depth0;
  stack4 = "showAlertsPopup";
  stack5 = {};
  stack6 = "view.parentView";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n      <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "component.host";
  stack3 = depth0;
  stack4 = "showDetails";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "component.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>,\n    ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"clearfix\">\n  <div class=\"name span2\">\n    <i class=\"pull-left icon-empty\"></i>\n    ");
  stack1 = depth0;
  stack2 = "App.MainDashboardServiceHealthView";
  stack3 = {};
  stack4 = "view.service";
  stack3['serviceBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    <a ");
  stack1 = depth0;
  stack2 = "view.service";
  stack3 = depth0;
  stack4 = "selectService";
  stack5 = {};
  stack6 = true;
  stack5['href'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.service.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  <div class=\"summary span\">\n    ");
  stack1 = depth0;
  stack2 = "view.titleMasters";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n    <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.clients.component";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.clients.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n  </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/services/mapreduce2", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n      <a href=\"#\" class=\"label label-important\" ");
  stack1 = depth0;
  stack2 = "view.service";
  stack3 = depth0;
  stack4 = "showAlertsPopup";
  stack5 = {};
  stack6 = "view.parentView";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
  return buffer;}

  data.buffer.push("\n\n\n<div class=\"clearfix\">\n  <div class=\"name span2\">\n    <i class=\"pull-left icon-empty\"></i>\n    ");
  stack1 = depth0;
  stack2 = "App.MainDashboardServiceHealthView";
  stack3 = {};
  stack4 = "view.service";
  stack3['serviceBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    <a ");
  stack1 = depth0;
  stack2 = "view.service";
  stack3 = depth0;
  stack4 = "selectService";
  stack5 = {};
  stack6 = true;
  stack5['href'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.service.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  <div class=\"summary span\">\n    ");
  stack1 = depth0;
  stack2 = "view.titleInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/services/oozie", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n      <a href=\"#\" class=\"label label-important\" ");
  stack1 = depth0;
  stack2 = "view.service";
  stack3 = depth0;
  stack4 = "showAlertsPopup";
  stack5 = {};
  stack6 = "view.parentView";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n      <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "component.host";
  stack3 = depth0;
  stack4 = "showDetails";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "component.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>,\n    ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"clearfix\">\n  <div class=\"name span2\">\n    <i class=\"pull-left icon-empty\"></i>\n    ");
  stack1 = depth0;
  stack2 = "App.MainDashboardServiceHealthView";
  stack3 = {};
  stack4 = "view.service";
  stack3['serviceBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    <a ");
  stack1 = depth0;
  stack2 = "view.service";
  stack3 = depth0;
  stack4 = "selectService";
  stack5 = {};
  stack6 = true;
  stack5['href'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.service.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  <div class=\"summary span\">\n    ");
  stack1 = depth0;
  stack2 = "view.masters";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n    <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.clients.component";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.clients.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>,\n\n    <a href=\"");
  stack1 = depth0;
  stack2 = "view.webUi";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\" target=\"_blank\">");
  stack1 = depth0;
  stack2 = "services.oozie.webUi";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n\n  </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/services/ranger", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "item.isDisplayed";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <tr>\n      <td>");
  stack1 = depth0;
  stack2 = "item.pluginTitle";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n      <td>");
  stack1 = depth0;
  stack2 = "item.status";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n    </tr>\n  ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.dashboardMasterComponentView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n\n");
  stack1 = depth0;
  stack2 = "controller.rangerPlugins";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "item";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/services/storm", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.liveComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "/");
  stack1 = depth0;
  stack2 = "view.totalComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "App.SummaryMasterComponentsView";
  stack3 = {};
  stack4 = "view.parentView.mastersObj";
  stack3['mastersCompBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n<tr>\n  <td>\n    <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.filterComponent";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      ");
  stack1 = depth0;
  stack2 = "dashboard.services.storm.supervisors";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    </a>\n  </td>\n  <td>\n    <span>\n      ");
  stack1 = depth0;
  stack2 = "App.ComponentLiveTextView";
  stack3 = {};
  stack4 = "view.superVisorsLive";
  stack3['liveComponentsBinding'] = stack4;
  stack4 = "view.superVisorsTotal";
  stack3['totalComponentsBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </span> ");
  stack1 = depth0;
  stack2 = "services.service.summary.SupervisorsLIVE";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n<tr>\n  <td>\n    ");
  stack1 = depth0;
  stack2 = "services.service.summary.storm.freeslots";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n  <td>\n    ");
  stack1 = depth0;
  stack2 = "view.service.freeSlots";
  foundHelper = helpers.formatNull;
  stack3 = foundHelper || depth0.formatNull;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "formatNull", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " / ");
  stack1 = depth0;
  stack2 = "view.service.totalSlots";
  foundHelper = helpers.formatNull;
  stack3 = foundHelper || depth0.formatNull;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "formatNull", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " (");
  stack1 = depth0;
  stack2 = "view.freeSlotsPercentage";
  stack3 = {};
  stack4 = "0";
  stack3['empty'] = stack4;
  foundHelper = helpers.formatNull;
  stack4 = foundHelper || depth0.formatNull;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "formatNull", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "% ");
  stack1 = depth0;
  stack2 = "common.free";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ")\n  </td>\n</tr>\n<tr>\n  <td>\n    ");
  stack1 = depth0;
  stack2 = "services.service.summary.storm.tasks";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n  <td>\n    ");
  stack1 = depth0;
  stack2 = "view.service.totalTasks";
  foundHelper = helpers.formatNull;
  stack3 = foundHelper || depth0.formatNull;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "formatNull", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n<tr>\n  <td>\n    ");
  stack1 = depth0;
  stack2 = "services.service.summary.storm.executors";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n  <td>\n    ");
  stack1 = depth0;
  stack2 = "view.service.totalExecutors";
  foundHelper = helpers.formatNull;
  stack3 = foundHelper || depth0.formatNull;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "formatNull", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n<tr>\n  <td>\n    ");
  stack1 = depth0;
  stack2 = "services.service.summary.storm.topologies";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n  <td>\n    ");
  stack1 = depth0;
  stack2 = "view.service.topologies";
  foundHelper = helpers.formatNull;
  stack3 = foundHelper || depth0.formatNull;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "formatNull", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n<tr>\n  <td>\n    ");
  stack1 = depth0;
  stack2 = "services.service.summary.storm.nimbus.uptime";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n  <td>\n    ");
  stack1 = depth0;
  stack2 = "view.nimbusUptimeFormatted";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/services/yarn", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.liveComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "/");
  stack1 = depth0;
  stack2 = "view.totalComponents";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  return buffer;}

  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.dashboardMasterComponentView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n\n<!-- NodeManagers -->\n<tr>\n  <td><a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.nodeManagerComponent";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "dashboard.services.yarn.nodeManagers";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a></td>\n  <td>\n    <span>\n      ");
  stack1 = depth0;
  stack2 = "App.ComponentLiveTextView";
  stack3 = {};
  stack4 = "view.service.nodeManagersStarted";
  stack3['liveComponentsBinding'] = stack4;
  stack4 = "view.service.nodeManagersTotal";
  stack3['totalComponentsBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </span>");
  stack1 = depth0;
  stack2 = "common.started";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n<!-- NodeManagers status -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.yarn.nodeManagers.status";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.nodeManagersStatus";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- YARN Clients -->\n<tr>\n  <td class=\"summary-label\"><a ");
  stack1 = depth0;
  stack2 = "view.yarnClientComponent";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      href=\"javascript:void(null)\">");
  stack1 = depth0;
  stack2 = "view.service.installedClients";
  stack3 = {};
  stack4 = "t:dashboard.services.yarn.client";
  stack3['singular'] = stack4;
  stack4 = "t:dashboard.services.yarn.clients";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n  </td>\n  <td>\n    <span class=\"green-live\">");
  stack1 = depth0;
  stack2 = "view.service.installedClients";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " </span>\n    ");
  stack1 = depth0;
  stack2 = "view.service.installedClients";
  stack3 = {};
  stack4 = "t:dashboard.services.yarn.client";
  stack3['singular'] = stack4;
  stack4 = "t:dashboard.services.yarn.clients";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "common.installed";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </td>\n</tr>\n<!-- ResourceManager Uptime -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.yarn.resourceManager.uptime";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.nodeUptime";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- ResourceManager Heap -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.resourceManager.nodes.heap";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.nodeHeap";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- Containers -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.yarn.containers";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.containers";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- Applications -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.yarn.apps";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.apps";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- Memory -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.yarn.memory";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td>");
  stack1 = depth0;
  stack2 = "view.memory";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n</tr>\n<!-- Queues -->\n<tr>\n  <td>");
  stack1 = depth0;
  stack2 = "dashboard.services.yarn.queues";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</td>\n  <td><span rel=\"queue-tooltip\"\n            class=\"text-tooltip\" ");
  stack1 = {};
  stack2 = "view.service.queueFormatted";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.queues";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n  </td>\n</tr>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/service/services/zookeeper", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n      <a href=\"#\" class=\"label label-important\" ");
  stack1 = depth0;
  stack2 = "view.service";
  stack3 = depth0;
  stack4 = "showAlertsPopup";
  stack5 = {};
  stack6 = "view.parentView";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </a>\n    ");
  return buffer;}

  data.buffer.push("\n\n<div class=\"clearfix\">\n  <div class=\"name span2\">\n    <i class=\"pull-left icon-empty\"></i>\n    ");
  stack1 = depth0;
  stack2 = "App.MainDashboardServiceHealthView";
  stack3 = {};
  stack4 = "view.service";
  stack3['serviceBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    <a ");
  stack1 = depth0;
  stack2 = "view.service";
  stack3 = depth0;
  stack4 = "selectService";
  stack5 = {};
  stack6 = true;
  stack5['href'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.service.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
  stack1 = depth0;
  stack2 = "view.alertsCount";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  <div class=\"summary span\">\n    ");
  stack1 = depth0;
  stack2 = "view.titleInfo.pre";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.titleInfo.component";
  stack3 = depth0;
  stack4 = "filterHosts";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.titleInfo.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
  stack1 = depth0;
  stack2 = "services.zookeeper.postfix";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/main/views", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    <table class=\"table views-table\">\n      <tbody>\n        ");
  stack1 = depth0;
  stack2 = "view.views";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "instance";
  stack7 = helpers.each;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </tbody>\n    </table>\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "instance.visible";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <tr ");
  stack1 = depth0;
  stack2 = "instance";
  stack3 = depth0;
  stack4 = "setView";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <td class=\"icon\">\n                <a class=\"view-icon-image\" target=\"_blank\"><img ");
  stack1 = {};
  stack2 = "instance.iconPath";
  stack1['src'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></a>\n              </td>\n              <td class=\"label-version-description\">\n                <p class=\"view-label-text\">");
  stack1 = depth0;
  stack2 = "instance.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n                <p class=\"view-version-text\"> &nbsp;&nbsp;(");
  stack1 = depth0;
  stack2 = "instance.version";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ")</p> <br />\n                <p class=\"view-description-text\">");
  stack1 = depth0;
  stack2 = "instance.description";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n              </td>\n            </tr>\n          ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"muted no-views-text\">");
  stack1 = depth0;
  stack2 = "views.main.noViews";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n  ");
  return buffer;}

  data.buffer.push("\n\n<div id=\"views\">\n  <h4>");
  stack1 = depth0;
  stack2 = "views.main.yourViews";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h4>\n  ");
  stack1 = depth0;
  stack2 = "view.views.length";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(5, program5, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/utils/ajax", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  <br />\n  <br />\n  <pre><strong>");
  stack1 = depth0;
  stack2 = "utils.ajax.errorMessage";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": </strong><span class=\"text-error\">");
  stack1 = depth0;
  stack2 = "view.message";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span></pre>\n");
  return buffer;}

  data.buffer.push("\n\n<span class=\"text-error\">");
  stack1 = depth0;
  stack2 = "view.statusCode";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span> <span>");
  stack1 = depth0;
  stack2 = "view.api";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n");
  stack1 = depth0;
  stack2 = "view.showMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/utils/config_launch_switch_config_group_of_host", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "installer.controls.slaveComponentGroups";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "&#58;&nbsp;\n");
  stack1 = depth0;
  stack2 = "Em.Select";
  stack3 = {};
  stack4 = "view.parentView.configGroups";
  stack3['contentBinding'] = stack4;
  stack4 = "content.displayName";
  stack3['optionLabelPath'] = stack4;
  stack4 = "view.parentView.selectedConfigGroup";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/bootstrap_call_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div>\n        <div class=\"alert alert-error\">\n          ");
  stack1 = depth0;
  stack2 = "installer.step2.bootStrap.error";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      <pre>\n        ");
  stack1 = depth0;
  stack2 = "view.parentView.serverError";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </pre>\n    </div>\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div>\n      ");
  stack1 = depth0;
  stack2 = "installer.step2.bootStrap.inProgress";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n    <div class=\"spinner\"></div>\n");
  return buffer;}

  data.buffer.push("\n\n\n");
  stack1 = depth0;
  stack2 = "view.parentView.isError";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/controls_service_config_radio_buttons", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "option.hidden";
  stack3 = helpers.unless;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <label class=\"radio\">\n      ");
  stack1 = depth0;
  stack2 = "App.ServiceConfigRadioButton";
  stack3 = {};
  stack4 = "view.name";
  stack3['nameBinding'] = stack4;
  stack4 = "option.displayName";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "option.displayName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " &nbsp;\n    </label>\n  ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '';
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.options";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "option";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/controls_service_config_textfield_with_unit", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "App.ServiceConfigTextField";
  stack3 = {};
  stack4 = "view.serviceConfig";
  stack3['serviceConfigBinding'] = stack4;
  stack4 = "false";
  stack3['isPopoverEnabled'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n<span class=\"add-on\">");
  stack1 = depth0;
  stack2 = "view.serviceConfig.unit";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/controls_slave_component_groups_menu", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <span class=\"badge badge-important\">");
  stack1 = depth0;
  stack2 = "view.errorCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n  ");
  return buffer;}

  data.buffer.push("\n\n<a ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "showSlaveComponentGroup";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n  ");
  stack1 = depth0;
  stack2 = "view.content.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  stack1 = depth0;
  stack2 = "view.errorCount";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</a>\n<i ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "removeSlaveComponentGroup";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"icon-remove\"></i>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/master_hosts", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "installer.noHostsAssigned";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "value";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "showHosts";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "value.firstObject";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "and";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.otherLength";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.hasNoHosts";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.hasOneHost";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.hasMultipleHosts";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/master_hosts_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <li>");
  stack1 = depth0;
  stack2 = "host";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</li>\n  ");
  return buffer;}

  data.buffer.push("\n\n<ul>\n  ");
  stack1 = depth0;
  stack2 = "view.serviceConfig.value";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "host";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</ul>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/slave_component_hosts", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "none";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " -\n<a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.serviceConfig.category";
  stack3 = depth0;
  stack4 = "showEditSlaveComponentGroups";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  ");
  stack1 = depth0;
  stack2 = "installer.slaveComponentHosts.selectHosts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n</a>\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n<a href=\"#\" ");
  stack1 = depth0;
  stack2 = "view.serviceConfig.category";
  stack3 = depth0;
  stack4 = "showEditSlaveComponentGroups";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n  ");
  stack1 = depth0;
  stack2 = "view.hasMultipleHosts";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(6, program6, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</a>\n  ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.hosts.firstObject.hostName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "and";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.otherLength";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.hosts.firstObject.hostName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.hasNoHosts";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/slave_component_hosts_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <tr>\n      <td>\n        <label>");
  stack1 = depth0;
  stack2 = "host.hostName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n      </td>\n      <td>\n        ");
  stack1 = depth0;
  stack2 = "App.SlaveComponentDropDownGroupView";
  stack3 = {};
  stack4 = "host";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </td>\n    </tr>\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n        <select ");
  stack1 = depth0;
  stack2 = "changeGroup";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = "change";
  stack3['on'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "controller.getGroupsForDropDown";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "groupName";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </select>\n        ");
  return buffer;}
function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "view.optionTag";
  stack3 = {};
  stack4 = "groupName";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              <option value=\"");
  stack1 = depth0;
  stack2 = "groupName";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\" ");
  stack1 = {};
  stack2 = "view.selected";
  stack1['selected'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                ");
  stack1 = depth0;
  stack2 = "groupName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </option>\n            ");
  return buffer;}

  data.buffer.push("\n\n<div id=\"slave-hosts-popup\" class=\"alert alert-info\">");
  stack1 = depth0;
  stack2 = "header";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n<table class=\"table table-striped\">\n  <thead>\n    <tr>\n      <th>");
  stack1 = depth0;
  stack2 = "common.host";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "common.group";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n    </tr>\n  </thead>\n  <tbody>\n  ");
  stack1 = depth0;
  stack2 = "hosts";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "host";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </tbody>\n</table>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/slave_hosts", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "installer.noHostsAssigned";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.hasMultipleHosts";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(6, program6, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "showHosts";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "value.firstObject";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "and";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.otherLength";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n  ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "view.value.0";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.hasNoHosts";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step0", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<div id=\"get-started\">\n  <h2>");
  stack1 = depth0;
  stack2 = "installer.step0.body.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n  <p class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "installer.step0.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </p>\n  <div ");
  stack1 = {};
  stack2 = "view.onError:error :control-group";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <label class=\"control-label\" for=\"cluster-name\">");
  stack1 = depth0;
  stack2 = "installer.step0.clusterName";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n      <a href=\"javascript:void(null)\"\n         rel=\"popover\"\n        ");
  stack1 = {};
  stack2 = "installer.step0.clusterName.tooltip.title";
  stack1['title'] = stack2;
  stack2 = "installer.step0.clusterName.tooltip.content";
  stack1['data-content'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.learnMore";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    </label>\n\n    <div class=\"controls\">\n      ");
  stack1 = depth0;
  stack2 = "App.WizardStep0ViewClusterNameInput";
  stack3 = {};
  stack4 = "cluster-name";
  stack3['id'] = stack4;
  stack4 = "content.cluster.name";
  stack3['valueBinding'] = stack4;
  stack4 = "cluster name";
  stack3['placeholder'] = stack4;
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      <p class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "clusterNameError";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n    </div>\n  </div>\n\n  <div class=\"btn-area\">\n    <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "invalidClusterName";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>\n\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step1", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <label class=\"radio\">");
  stack1 = depth0;
  stack2 = "view.stackRadioButton";
  stack3 = {};
  stack4 = "stack";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "stack.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <span class=\"badge badge-important\">");
  stack1 = depth0;
  stack2 = "view.totalErrorCnt";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n          ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n                <div class=\"trow\">\n                  <div ");
  stack1 = {};
  stack2 = ":os-td operatingSystem.osType";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                    <label>\n                      ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "operatingSystem.isSelected";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                      <span ");
  stack1 = {};
  stack2 = ":os operatingSystem.isSelected::disabled-label";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "operatingSystem.osType";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n                    </label>\n                  </div>\n                  <div style=\"width:83%\">\n                    ");
  stack1 = depth0;
  stack2 = "operatingSystem.repositories";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "repository";
  stack7 = helpers.each;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                  </div>\n                </div>\n              ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                      <div class=\"sub-trow\">\n                        <div class=\"name-td\">");
  stack1 = depth0;
  stack2 = "repository.repoId";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n                        <div class=\"validation-td\">\n                          ");
  stack1 = depth0;
  stack2 = "repository.validation";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                        </div>\n                        <div ");
  stack1 = {};
  stack2 = ":url-td operatingSystem.osType repository.repoId operatingSystem.isSelected::disabled-textfield repository.emptyError:textfield-error repository.invalidError:textfield-error";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                          ");
  stack1 = depth0;
  stack2 = "Ember.TextField";
  stack3 = {};
  stack4 = "repository.baseUrl";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                        </div>\n                        <div class=\"clear-td\">\n                          ");
  stack1 = depth0;
  stack2 = "repository.clearAll";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                        </div>\n                        <div class=\"actions-td\">\n                          ");
  stack1 = depth0;
  stack2 = "repository.undo";
  stack3 = helpers['if'];
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                        </div>\n                      </div>\n                    ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                            ");
  stack1 = depth0;
  stack2 = "view.popoverView";
  stack3 = {};
  stack4 = "repository";
  stack3['repositoryBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                          ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                            <a ");
  stack1 = depth0;
  stack2 = "repository";
  stack3 = depth0;
  stack4 = "clearGroupLocalRepository";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = {};
  stack2 = "operatingSystem.isSelected::disabled-clear-button";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                              <i class=\"icon-remove-sign\"></i>\n                            </a>\n                          ");
  return buffer;}

function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n                            <a ");
  stack1 = depth0;
  stack2 = "repository";
  stack3 = depth0;
  stack4 = "undoGroupLocalRepository";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                              <i class=\"icon-undo\"></i>");
  stack1 = depth0;
  stack2 = "common.undo";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                            </a>\n                          ");
  return buffer;}

function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <div class=\"alert\">");
  stack1 = depth0;
  stack2 = "installer.step1.attentionNeeded";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n          ");
  return buffer;}

function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            <div class=\"alert\">\n              ");
  stack1 = depth0;
  stack2 = "installer.step1.invalidURLAttention";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              <a href=\"javascript:void(null)\" ");
  stack1 = depth0;
  stack2 = "retryRepoUrls";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step1.retryRepoUrls";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n            </div>\n          ");
  return buffer;}

function program17(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <div class=\"alert\">");
  stack1 = depth0;
  stack2 = "installer.step1.checkAtLeastOneAttention";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n          ");
  return buffer;}

  data.buffer.push("\n<div id=\"select-stack\">\n  <h2>");
  stack1 = depth0;
  stack2 = "installer.step1.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n  <p class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "installer.step1.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </p>\n  <p><b>");
  stack1 = depth0;
  stack2 = "common.stacks";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</b></p>\n  <form autocomplete=\"off\">\n    ");
  stack1 = depth0;
  stack2 = "view.stacks";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "stack";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </form>\n\n\n  <div class=\"accordion\" id=\"advancedRepoAccordion\">\n    <div class=\"accordion-group\">\n      <div class=\"accordion-heading\" ");
  stack1 = depth0;
  stack2 = "onToggleBlock";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n        <i ");
  stack1 = {};
  stack2 = ":pull-left :accordion-toggle view.isRLCollapsed:icon-caret-right:icon-caret-down";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n        <a class=\"accordion-toggle\">\n          ");
  stack1 = depth0;
  stack2 = "installer.step1.advancedRepo.title";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "view.isSubmitDisabled";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </a>\n\n      </div>\n      <div class=\"accordion-body collapse in\">\n        <div class=\"accordion-inner\">\n          <div class=\"alert alert-info\">");
  stack1 = depth0;
  stack2 = "installer.step1.advancedRepo.message";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n          <div class=\"alert alert-warning\">");
  stack1 = depth0;
  stack2 = "installer.step1.advancedRepo.importantMassage";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n\n          <div class=\"repositories-table\">\n            <div class=\"thead\">\n              <div class=\"first-th\">&nbsp;</div>\n              <div class=\"th os-th\">");
  stack1 = depth0;
  stack2 = "common.os";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n              <div class=\"th name-th\">");
  stack1 = depth0;
  stack2 = "common.name";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n              <div class=\"th url-th\">");
  stack1 = depth0;
  stack2 = "installer.step1.advancedRepo.localRepo.column.baseUrl";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n            </div>\n            <div class=\"tbody\">\n              ");
  stack1 = depth0;
  stack2 = "view.operatingSystems";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "operatingSystem";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </div>\n          </div>\n          <div id=\"skip-validation\">\n            <label>");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "skipValidationChecked";
  stack3['checkedBinding'] = stack4;
  stack4 = "checkbox";
  stack3['class'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "installer.step1.advancedRepo.skipValidation.message";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              <i class=\"icon-question-sign\" rel=\"skip-validation-tooltip\"\n                 data-toggle=\"tooltip\" ");
  stack1 = {};
  stack2 = "installer.step1.advancedRepo.skipValidation.tooltip";
  stack1['title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + "></i></label>\n          </div>\n          ");
  stack1 = depth0;
  stack2 = "view.emptyRepoExist";
  stack3 = helpers['if'];
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.invalidUrlExist";
  stack3 = helpers['if'];
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  stack1 = depth0;
  stack2 = "view.allRepoUnchecked";
  stack3 = helpers['if'];
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </div>\n      </div>\n    </div>\n  </div>\n\n  <a class=\"btn pull-left installer-back-btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n  <button class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "view.isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</button>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step10", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div id=\"step10-alert-message\" class=\"alert alert-danger\">\n      ");
  stack1 = depth0;
  stack2 = "view.serviceRestartText";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n  ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n        <li>\n          <span ");
  stack1 = {};
  stack2 = "item.color";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "item.displayStatement";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n          <ul>\n            ");
  stack1 = depth0;
  stack2 = "item.status";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "status";
  stack7 = helpers.each;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </ul>\n        </li>\n      ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n              <li>\n                <span ");
  stack1 = {};
  stack2 = "status.color";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "status.displayStatement";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n                <ul>\n                  ");
  stack1 = depth0;
  stack2 = "status.statements";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "statement";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </ul>\n              </li>\n            ");
  return buffer;}
function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    <li>\n                      <span ");
  stack1 = {};
  stack2 = "status.color";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "statement.displayStatement";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n                    </li>\n                  ");
  return buffer;}

  data.buffer.push("\n<div id=\"deploy-summary\">\n  <h2>");
  stack1 = depth0;
  stack2 = "installer.step10.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n  ");
  stack1 = depth0;
  stack2 = "view.serviceRestartText";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  <div class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "installer.step10.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <div id=\"step10-content\" class=\"well pre-scrollable\">\n    <ul>\n      ");
  stack1 = depth0;
  stack2 = "clusterInfo";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "item";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </ul>\n  </div>\n  <div class=\"btn-area\">\n    <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "complete";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.complete";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step1_addLocalRepository", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n					    <li>\n  					    <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "os";
  stack3 = depth0;
  stack4 = "selectOS";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  ");
  stack1 = depth0;
  stack2 = "os.osType";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              </li>\n				    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n	    <div class=\"alert\">\n			  <strong>");
  stack1 = depth0;
  stack2 = "common.warning";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</strong>  ");
  stack1 = depth0;
  stack2 = "view.errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n			</div>\n	  ");
  return buffer;}

function program5(depth0,data) {


  data.buffer.push("\n	    <div class=\"alert alert-info\">\n        Provide location of the local repository for the selected OS\n      </div>\n		");}

  data.buffer.push("\n  <form class=\"form-horizontal\" id=\"addLocalRepositoryPopup\" autocomplete=\"off\">\n    <div class=\"each-row\">\n      <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "installer.step1.advancedRepo.localRepo.label.os";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n      <div class=\"controls\">\n        <div class=\"btn-group\">\n				  <a class=\"btn dropdown-toggle\" data-toggle=\"dropdown\" href=\"#\">\n				    ");
  stack1 = depth0;
  stack2 = "view.selectedOS.osType";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n				    <span class=\"caret\"></span>\n				  </a>\n				  <ul class=\"dropdown-menu\">\n  				  ");
  stack1 = depth0;
  stack2 = "view.oses";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "os";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n				  </ul>\n				</div>\n      </div>\n    </div>\n    <div class=\"each-row\">\n      <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "installer.step1.advancedRepo.localRepo.label.baseUrl";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n      <div class=\"controls\">\n        ");
  stack1 = depth0;
  stack2 = "Em.TextField";
  stack3 = {};
  stack4 = "view.enteredUrl";
  stack3['valueBinding'] = stack4;
  stack4 = "span6";
  stack3['classNames'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    <div class=\"each-row\">\n      <label class=\"control-label\">");
  stack1 = depth0;
  stack2 = "installer.step1.advancedRepo.localRepo.label.stack";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n      <div class=\"controls\">\n        ");
  stack1 = depth0;
  stack2 = "view.stackName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n    <div style=\"height:100px;\"></div>\n    ");
  stack1 = depth0;
  stack2 = "view.errorMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(5, program5, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </form>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step2", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <p class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "hostsError";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n        ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <label class=\"radio\">\n          ");
  stack1 = depth0;
  stack2 = "view.providingSSHKeyRadioButton";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "installer.step2.useSsh.provide";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n          <a href=\"javascript:void(null)\"\n             rel=\"popover\"\n            ");
  stack1 = {};
  stack2 = "installer.step2.useSsh.tooltip.title";
  stack1['title'] = stack2;
  stack2 = "installer.step2.useSsh.tooltip.content";
  stack1['data-content'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n            ");
  stack1 = depth0;
  stack2 = "installer.step2.useSsh.tooltip.title";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n          ");
  stack1 = depth0;
  stack2 = "installer.step2.useSsh.provide_id_rsa";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </label>\n\n        <div class=\"ssh-key-input\">\n          ");
  stack1 = depth0;
  stack2 = "view.isFileApi";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          <div ");
  stack1 = {};
  stack2 = "sshKeyError:error :controls :control-group";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            ");
  stack1 = depth0;
  stack2 = "Ember.TextArea";
  stack3 = {};
  stack4 = "span6";
  stack3['class'] = stack4;
  stack4 = "3";
  stack3['rows'] = stack4;
  stack4 = "sshKey";
  stack3['id'] = stack4;
  stack4 = "ssh private key";
  stack3['placeholder'] = stack4;
  stack4 = "view.sshKeyState";
  stack3['disabledBinding'] = stack4;
  stack4 = "content.installOptions.sshKey";
  stack3['valueBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  stack1 = depth0;
  stack2 = "sshKeyError";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n          <div class=\"row-fluid\">\n            <label rel=\"tooltip\" ");
  stack1 = {};
  stack2 = "installer.step2.sshUser.toolTip";
  stack1['title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + " class=\"ssh-user pull-left span4\">\n              ");
  stack1 = depth0;
  stack2 = "installer.step2.sshUser";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </label>\n\n            <div ");
  stack1 = {};
  stack2 = "sshUserError:error :control-group";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              ");
  stack1 = depth0;
  stack2 = "view.textFieldView";
  stack3 = {};
  stack4 = "content.installOptions.sshUser";
  stack3['valueBinding'] = stack4;
  stack4 = "content.installOptions.useSsh";
  stack3['isEnabledBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "sshUserError";
  stack3 = helpers['if'];
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </div>\n          </div>\n          ");
  stack1 = depth0;
  stack2 = "App.supports.customizeAgentUserAccount";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n        </div>\n      ");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "App.SshKeyFileUploader";
  stack3 = {};
  stack4 = "view.sshKeyState";
  stack3['disabledBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "sshKeyError";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n            ");
  return buffer;}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "sshUserError";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n              ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            <div class=\"row-fluid\">\n              <label rel=\"tooltip\" ");
  stack1 = {};
  stack2 = "installer.step2.agentUser.toolTip";
  stack1['title'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + " class=\"ssh-user pull-left span4\">\n                ");
  stack1 = depth0;
  stack2 = "installer.step2.agentUser";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n              </label>\n\n              <div ");
  stack1 = {};
  stack2 = "agentUserError:error :control-group";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                ");
  stack1 = depth0;
  stack2 = "view.textFieldView";
  stack3 = {};
  stack4 = "content.installOptions.agentUser";
  stack3['valueBinding'] = stack4;
  stack4 = "content.installOptions.useSsh";
  stack3['isEnabledBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                ");
  stack1 = depth0;
  stack2 = "agentUserError";
  stack3 = helpers['if'];
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </div>\n            </div>\n          ");
  return buffer;}
function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  <span class=\"help-inline\">");
  stack1 = depth0;
  stack2 = "agentUserError";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n                ");
  return buffer;}

function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <a class=\"btn pull-left installer-back-btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
  return buffer;}

  data.buffer.push("\n\n<div id=\"installOptions\">\n  <h2>");
  stack1 = depth0;
  stack2 = "installer.step2.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <p class=\"alert alert-info\">");
  stack1 = depth0;
  stack2 = "installer.step2.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</p>\n\n  <div id=\"targetHosts\">\n    <h5>");
  stack1 = depth0;
  stack2 = "installer.step2.targetHosts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h5>\n\n    <div ");
  stack1 = {};
  stack2 = "hostsError:error :control-group :target-hosts-input";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <p>");
  stack1 = depth0;
  stack2 = "installer.step2.targetHosts.info";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ". ");
  stack1 = depth0;
  stack2 = "installer.step2.orUse";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        <a href=\"javascript:void(null)\"\n           rel=\"popover\"\n          ");
  stack1 = {};
  stack2 = "installer.step2.hostPattern.tooltip.title";
  stack1['title'] = stack2;
  stack2 = "installer.step2.hostPattern.tooltip.content";
  stack1['data-content'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "installer.step2.hostPattern.tooltip.title";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </a>\n      </p>\n\n      <div class=\"controls\">\n        ");
  stack1 = depth0;
  stack2 = "Ember.TextArea";
  stack3 = {};
  stack4 = "host-names";
  stack3['id'] = stack4;
  stack4 = "span6";
  stack3['class'] = stack4;
  stack4 = "content.installOptions.hostNames";
  stack3['valueBinding'] = stack4;
  stack4 = "5";
  stack3['rows'] = stack4;
  stack4 = "host names";
  stack3['placeholder'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "hostsError";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n    </div>\n  </div>\n\n  <div id=\"hostConnectivity\">\n    <div class=\"ambari-agents\">\n      <h5>");
  stack1 = depth0;
  stack2 = "installer.step2.sshKey";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h5>\n\n      ");
  stack1 = depth0;
  stack2 = "isSSHRegistrationEnabled";
  stack3 = helpers['if'];
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n      <label class=\"radio\">\n        ");
  stack1 = depth0;
  stack2 = "view.manualRegistrationRadioButton";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "installer.step2.manualInstall.perform";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        <a href=\"javascript:void(null)\"\n           rel=\"popover\"\n          ");
  stack1 = {};
  stack2 = "installer.step2.manualInstall.tooltip.title";
  stack1['title'] = stack2;
  stack2 = "installer.step2.manualInstall.tooltip.content";
  stack1['data-content'] = stack2;
  foundHelper = helpers.translateAttr;
  stack2 = foundHelper || depth0.translateAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  if(typeof stack2 === functionType) { stack1 = stack2.call(depth0, tmp1); }
  else if(stack2=== undef) { stack1 = helperMissing.call(depth0, "translateAttr", tmp1); }
  else { stack1 = stack2; }
  data.buffer.push(escapeExpression(stack1) + ">\n          ");
  stack1 = depth0;
  stack2 = "installer.step2.manualInstall.tooltip.title";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n        ");
  stack1 = depth0;
  stack2 = "installer.step2.manualInstall.perform_on_hosts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n      </label>\n\n    </div>\n  </div>\n\n  <div class=\"btn-area\">\n    ");
  stack1 = depth0;
  stack2 = "view.parentView.controller.hideBackButton";
  stack3 = helpers.unless;
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "evaluateStep";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      ");
  stack1 = depth0;
  stack2 = "installer.step2.registerAndConfirm";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step2ManualInstallPopup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<p>");
  stack1 = depth0;
  stack2 = "installer.step2.manualInstall.popup.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</p>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step2_installed_hosts_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n<p>");
  stack1 = depth0;
  stack2 = "installer.step2.evaluateStep.installedHosts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</p>\n<p>");
  stack1 = depth0;
  stack2 = "view.inputtedAgainHostNames";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n<p>");
  stack1 = depth0;
  stack2 = "installer.step2.evaluateStep.continueConfirm";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</p>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step3", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          <a class=\"btn btn-primary decommission\"\n             href=\"#\" ");
  stack1 = depth0;
  stack2 = "retrySelectedHosts";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-repeat icon-white\"></i>\n            ");
  stack1 = depth0;
  stack2 = "installer.step3.retryFailed";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n          </a>\n        ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n              <li ");
  stack1 = {};
  stack2 = ":filter-status category.itemClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                <a ");
  stack1 = depth0;
  stack2 = "category";
  stack3 = depth0;
  stack4 = "selectCategory";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n                  ");
  stack1 = depth0;
  stack2 = "category.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              </li>\n              ");
  stack1 = depth0;
  stack2 = "category.last";
  stack3 = helpers.unless;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program4(depth0,data) {


  data.buffer.push("\n                <li class=\"divider\">|</li>\n              ");}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "host";
  stack7 = helpers.each;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "App.WizardHostView";
  stack3 = {};
  stack4 = "controller.category";
  stack3['categoryBinding'] = stack4;
  stack4 = "host";
  stack3['hostInfoBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(8, program8, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n              <td class=\"step3-table-checkbox\">\n                ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "host.isChecked";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </td>\n              <td class=\"host\">\n                <span title=\"");
  stack1 = depth0;
  stack2 = "host.name";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\" class=\"trim_hostname\">");
  stack1 = depth0;
  stack2 = "host.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n              </td>\n              <td class=\"step3-table-progress\">\n                <div ");
  stack1 = {};
  stack2 = "host.bootBarColor host.isBootDone::progress-striped host.isBootDone::active :progress";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                  <div class=\"bar\" style=\"width:100%\">\n                  </div>\n                </div>\n              </td>\n              <td class=\"step3-table-status\">\n                <a href=\"javascript:void(null)\"\n                   data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "host";
  stack3 = depth0;
  stack4 = "hostLogPopup";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><span ");
  stack1 = {};
  stack2 = "host.bootStatusColor";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "host.bootStatusForDisplay";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span></a>\n              </td>\n              <td class=\"step3-table-action\">\n                <a class=\"btn btn-mini\" ");
  stack1 = depth0;
  stack2 = "remove";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-trash\"></i>\n                  ");
  stack1 = depth0;
  stack2 = "common.remove";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n                ");
  stack1 = depth0;
  stack2 = "view.isRetryable";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </td>\n              ");
  return buffer;}
function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("<a class=\"btn btn-mini\" ");
  stack1 = depth0;
  stack2 = "retry";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i\n                    class=\"icon-repeat\"></i>\n                  ");
  stack1 = depth0;
  stack2 = "common.retry";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>");
  return buffer;}

function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <tr>\n            <td class=\"step3-table-no-hosts\" colspan=\"5\">\n              ");
  stack1 = depth0;
  stack2 = "hosts.table.noHosts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n            </td>\n          </tr>\n          ");
  return buffer;}

function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n            <a ");
  stack1 = depth0;
  stack2 = "selectedHostsPopup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n              ");
  stack1 = depth0;
  stack2 = "view.selectedHostsCount";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "view.selectedHostsCount";
  stack3 = {};
  stack4 = "t:hosts.filters.selectedHostInfo";
  stack3['singular'] = stack4;
  stack4 = "t:hosts.filters.selectedHostsInfo";
  stack3['plural'] = stack4;
  foundHelper = helpers.pluralize;
  stack4 = foundHelper || depth0.pluralize;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack4 === functionType) { stack1 = stack4.call(depth0, stack2, tmp1); }
  else if(stack4=== undef) { stack1 = helperMissing.call(depth0, "pluralize", stack2, tmp1); }
  else { stack1 = stack4; }
  data.buffer.push(escapeExpression(stack1) + "\n            </a>\n            -\n            <a ");
  stack1 = depth0;
  stack2 = "unSelectAll";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
  stack1 = depth0;
  stack2 = "hosts.filters.clearSelection";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n          ");
  return buffer;}

function program15(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n    <div ");
  stack1 = {};
  stack2 = ":alert alert-warn";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "registeredHostsPopup";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.registeredHostsMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n    </div>\n  ");
  return buffer;}

function program17(depth0,data) {


  data.buffer.push("\n      <div class=\"spinner\"></div>\n    ");}

  data.buffer.push("\n\n<div id=\"confirm-hosts\">\n  <h2>");
  stack1 = depth0;
  stack2 = "installer.step3.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <p class=\"alert alert-info\">");
  stack1 = depth0;
  stack2 = "installer.step3.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</p>\n\n  <div class=\"box\">\n    <div class=\"box-header\">\n      <div class=\"button-section\">\n        <button class=\"btn btn-primary step3-remove-selected-btn\" ");
  stack1 = {};
  stack2 = "view.noHostsSelected";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "removeSelectedHosts";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "><i class=\"icon-trash icon-white\"></i>\n          ");
  stack1 = depth0;
  stack2 = "installer.step3.removeSelected";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n        </button>\n        ");
  stack1 = depth0;
  stack2 = "isRetryDisabled";
  stack3 = helpers.unless;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n        <div id=\"host-filter\" class=\"pull-right\">\n          <ul class=\"clearfix\">\n            <li class=\"first\">");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</li>\n            ");
  stack1 = depth0;
  stack2 = "view.categories";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "category";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </ul>\n        </div>\n      </div>\n    </div>\n\n    <div class=\"pre-scrollable\" style=\"max-height: 440px;\">\n      <table id=\"confirm-hosts-table\" class=\"table table-bordered table-striped\">\n        <thead>\n        <tr>\n          <th class=\"tinyspan\">\n            ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "view.pageChecked";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          </th>\n          <th class=\"span5\">");
  stack1 = depth0;
  stack2 = "common.host";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n          <!-- retrieved from local storage initially -->\n          <th class=\"span2\">");
  stack1 = depth0;
  stack2 = "common.progress";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n          <th class=\"span2\">");
  stack1 = depth0;
  stack2 = "common.status";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n          <!-- given by the parsing function that parses data from bootstrap call, dynamically assign the color -->\n          <th class=\"span2\">");
  stack1 = depth0;
  stack2 = "common.action";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n          <!-- trash icon -->\n          <!-- retry icon -->\n        </tr>\n        </thead>\n        <tbody>\n          ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(11, program11, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </tbody>\n      </table>\n    </div>\n    <div id=\"hosts\">\n      <div class=\"page-bar\">\n        <div class=\"selected-hosts-info pull-left\">\n          ");
  stack1 = depth0;
  stack2 = "view.selectedHostsCount";
  stack3 = helpers['if'];
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </div>\n        <div class=\"items-on-page\">\n          <label>");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.rowsPerPageSelectView";
  stack3 = {};
  stack4 = "view.displayLength";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n        </div>\n        <div class=\"info\">");
  stack1 = depth0;
  stack2 = "view.paginationInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"paging_two_button\">\n          ");
  stack1 = depth0;
  stack2 = "view.paginationFirst";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "view.paginationLeft";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "view.paginationRight";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "view.paginationLast";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  </div>\n  ");
  stack1 = depth0;
  stack2 = "hasMoreRegisteredHosts";
  stack3 = helpers['if'];
  tmp1 = self.program(15, program15, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  <div ");
  stack1 = {};
  stack2 = ":alert view.status isWarningsBoxVisible::hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    ");
  stack1 = depth0;
  stack2 = "view.message";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    <a href=\"#\" ");
  stack1 = depth0;
  stack2 = "warnings";
  stack3 = depth0;
  stack4 = "hostWarningsPopup";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.linkText";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
  stack1 = depth0;
  stack2 = "isWarningsLoaded";
  stack3 = helpers.unless;
  tmp1 = self.program(17, program17, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  <div class=\"btn-area\">\n    <button class=\"btn pull-left installer-back-btn\" ");
  stack1 = {};
  stack2 = "isBackDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</button>\n    <button class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</button>\n  </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step3/step3_host_log_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <div>\n        ");
  stack1 = depth0;
  stack2 = "view.textArea";
  stack3 = {};
  stack4 = "view.bootLog";
  stack3['contentBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <pre class=\"bootLog\">");
  stack1 = depth0;
  stack2 = "view.bootLog";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</pre>\n    ");
  return buffer;}

  data.buffer.push("\n\n\n<div id=\"host-log\">\n  <div class=\"content-area\">\n    <div class=\"textTrigger\">");
  stack1 = depth0;
  stack2 = "popup.highlight";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n    ");
  stack1 = depth0;
  stack2 = "view.isTextArea";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step3/step3_host_warning_popup_footer", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2;
  data.buffer.push("\n    <div class=\"progress-info active progress\">\n      <div class=\"bar\" ");
  stack1 = {};
  stack2 = "view.progressWidth";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n    </div>\n  ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <label ");
  stack1 = {};
  stack2 = "view.updateStatusClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.updateStatus";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n  ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <button type=\"button\"\n          class=\"btn btn-info\" ");
  stack1 = {};
  stack2 = "view.isUpdateInProgress";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "onSecondary";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    <i class=\"icon-repeat\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "view.parentView.secondary";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</button>\n");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <button type=\"button\" class=\"btn\" ");
  stack1 = depth0;
  stack2 = "onPrimary";
  stack3 = {};
  stack4 = "view.parentView";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.parentView.primary";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</button>\n");
  return buffer;}

  data.buffer.push("\n\n<div class=\"update-progress pull-left\">\n  ");
  stack1 = depth0;
  stack2 = "view.isUpdateInProgress";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  stack1 = depth0;
  stack2 = "view.parentView.secondary";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  stack1 = depth0;
  stack2 = "view.parentView.primary";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step3/step3_host_warnings_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      <div class=\"span3 offset2\">\n        <a href=\"javascript:void(null)\" title=\"Show Details\" ");
  stack1 = depth0;
  stack2 = "openWarningsInDialog";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n           class=\"task-detail-open-dialog\"><i\n            class=\"icon-external-link\"></i> ");
  stack1 = depth0;
  stack2 = "installer.step3.hostWarningsPopup.report";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n      </div>\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n      <div class=\"accordion-group block\">\n        <div class=\"accordion-heading\" ");
  stack1 = depth0;
  stack2 = "category";
  stack3 = depth0;
  stack4 = "onToggleBlock";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <i ");
  stack1 = {};
  stack2 = ":pull-left :accordion-toggle category.isCollapsed:icon-caret-right:icon-caret-down";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n          ");
  stack1 = depth0;
  stack2 = "category.warnings.length";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(6, program6, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          <a class=\"accordion-toggle\">\n            ");
  stack1 = depth0;
  stack2 = "category.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " (");
  stack1 = depth0;
  stack2 = "category.warnings.length";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ")\n          </a>\n        </div>\n        <div id=\"");
  stack1 = depth0;
  stack2 = "category.category";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\" class=\"accordion-body collapse in\" style=\"display: none\">\n          <div class=\"accordion-inner\">\n            ");
  stack1 = depth0;
  stack2 = "category.warnings.length";
  stack3 = helpers['if'];
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(18, program18, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n        </div>\n      </div>\n    ");
  return buffer;}
function program4(depth0,data) {


  data.buffer.push("\n            <i class=\"pull-right accordion-toggle icon-warning-sign\"></i>\n          ");}

function program6(depth0,data) {


  data.buffer.push("\n            <i class=\"pull-right accordion-toggle icon-ok\"></i>\n          ");}

function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n              <table>\n                <thead>\n                <tr>\n                  <th colspan=\"2\">");
  stack1 = depth0;
  stack2 = "category.message";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</th>\n                </tr>\n                <tr>\n                  <th colspan=\"2\"><b>");
  stack1 = depth0;
  stack2 = "category.type";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</b></th>\n                </tr>\n                </thead>\n                <tbody>\n                  ");
  stack1 = depth0;
  stack2 = "category.warnings";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "warning";
  stack7 = helpers.each;
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </tbody>\n              </table>\n            ");
  return buffer;}
function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n                  <tr>\n                    <td class=\"warning-name\" ");
  stack1 = {};
  stack2 = "warning.command";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " >");
  stack1 = depth0;
  stack2 = "warning.name";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</td>\n                    ");
  stack1 = depth0;
  stack2 = "warning.version";
  stack3 = helpers['if'];
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    ");
  stack1 = depth0;
  stack2 = "warning.target";
  stack3 = helpers['if'];
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    <td>");
  stack1 = depth0;
  stack2 = "category.action";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                      <a href=\"javascript:void(null);\" rel='HostsListTooltip' ");
  stack1 = {};
  stack2 = "warning.hostsList";
  stack1['data-original-title'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "warning.hosts";
  stack3 = depth0;
  stack4 = "showHostsPopup";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                        ");
  stack1 = depth0;
  stack2 = "warning.hosts.length";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                        ");
  stack1 = depth0;
  stack2 = "warning.onSingleHost";
  stack3 = helpers['if'];
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(16, program16, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                      </a>\n                    </td>\n                  </tr>\n                  ");
  return buffer;}
function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                      <td class=\"package-version\">");
  stack1 = depth0;
  stack2 = "warning.version";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n                    ");
  return buffer;}

function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                      <td class=\"package-version\">");
  stack1 = depth0;
  stack2 = "warning.target";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n                    ");
  return buffer;}

function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                          ");
  stack1 = depth0;
  stack2 = "installer.step3.hostWarningsPopup.host";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                        ");
  return buffer;}

function program16(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                          ");
  stack1 = depth0;
  stack2 = "installer.step3.hostWarningsPopup.hosts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n                        ");
  return buffer;}

function program18(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "installer.step3.hostWarningsPopup.emptyMessage";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "category.emptyName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  return buffer;}

  data.buffer.push("\n\n\n<div id=\"host-warnings\">\n  <div class=\"notice\">\n    <span>");
  stack1 = depth0;
  stack2 = "installer.step3.hostWarningsPopup.checks";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " <b>");
  stack1 = depth0;
  stack2 = "view.warningsNotice";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</b>.<br>");
  stack1 = depth0;
  stack2 = "installer.step3.hostWarningsPopup.notice";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n  </div>\n  <div class=\"row-fluid\">\n    <div class=\"span7\">\n      ");
  stack1 = depth0;
  stack2 = "common.hosts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "&nbsp;");
  stack1 = depth0;
  stack2 = "view.hostSelectView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n    </div>\n    ");
  stack1 = depth0;
  stack2 = "view.totalWarningsCount";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  <div class=\"accordion warnings-list\" id=\"accordion2\">\n    ");
  stack1 = depth0;
  stack2 = "view.content";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "category";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step3/step3_registered_hosts_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <li>");
  stack1 = depth0;
  stack2 = "host";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</li>\n  ");
  return buffer;}

  data.buffer.push("\n\n<p>");
  stack1 = depth0;
  stack2 = "view.message";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n<ul>\n  ");
  stack1 = depth0;
  stack2 = "view.registeredHosts";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "host";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</ul>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step4", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "isHiddenOnSelectServicePage";
  stack3 = helpers.unless;
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n        <tr ");
  stack1 = {};
  stack2 = "isSelected:success:";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <td><label class=\"checkbox\">");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "serviceName";
  stack3['classBinding'] = stack4;
  stack4 = "isInstalled";
  stack3['disabledBinding'] = stack4;
  stack4 = "isSelected";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "displayNameOnSelectServicePage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n          </td>\n          <td>");
  stack1 = depth0;
  stack2 = "serviceVersion";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</td>\n          <td>");
  stack1 = depth0;
  stack2 = "comments";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</td>\n        </tr>\n      ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <a class=\"btn pull-left installer-back-btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
  return buffer;}

  data.buffer.push("\n\n<div id=\"step4\">\n  <h2>");
  stack1 = depth0;
  stack2 = "installer.step4.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <div class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "installer.step4.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  <table class=\"table table-striped\">\n    <thead>\n    <tr>\n      <th class=\"span3\">\n          ");
  stack1 = depth0;
  stack2 = "Ember.Checkbox";
  stack3 = {};
  stack4 = "isInstalled";
  stack3['disabledBinding'] = stack4;
  stack4 = "isAllChecked";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "common.service";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n      </th>\n      <th>");
  stack1 = depth0;
  stack2 = "common.version";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th>");
  stack1 = depth0;
  stack2 = "common.description";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n    </tr>\n    </thead>\n    <tbody>\n    ");
  stack1 = depth0;
  stack2 = "controller";
  stack3 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </tbody>\n  </table>\n\n  <div class=\"btn-area\">\n    ");
  stack1 = depth0;
  stack2 = "view.parentView.controller.hideBackButton";
  stack3 = helpers.unless;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "> ");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step4/step4_ranger_requirements_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;


  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "installer.step4.rangerRequirements.popup.body.requirements";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n<label class=\"checkbox\">");
  stack1 = depth0;
  stack2 = "Em.Checkbox";
  stack3 = {};
  stack4 = "isChecked";
  stack3['checkedBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "installer.step4.rangerRequirements.popup.body.confirmation";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</label>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step5", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n      ");
  stack1 = depth0;
  stack2 = "view.coHostedComponentText";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"alert alert-error\">");
  stack1 = depth0;
  stack2 = "msg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n  ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <div class=\"alert alert-warning\">");
  stack1 = depth0;
  stack2 = "msg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n  ");
  return buffer;}

function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <div class=\"assign-masters row-fluid\">\n      <div class=\"select-hosts span7\">\n        <div class=\"row-fluid\">\n          ");
  stack1 = depth0;
  stack2 = "showCurrentHost";
  stack3 = helpers['if'];
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          <div class=\"clearfix\"></div>\n          <div class=\"row-fluid\">\n            <div class=\"span12 control-group\">\n              <form class=\"form-horizontal\" autocomplete=\"off\">\n                <!-- View for array controller -->\n                ");
  stack1 = depth0;
  stack2 = "servicesMasters";
  stack3 = helpers.each;
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              </form>\n            </div>\n          </div>\n        </div>\n      </div>\n\n      <div class=\"host-assignments span5\">\n        ");
  stack1 = depth0;
  stack2 = "masterHostMapping";
  stack3 = helpers.each;
  tmp1 = self.program(26, program26, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n        ");
  stack1 = depth0;
  stack2 = "remainingHosts";
  stack3 = helpers['if'];
  tmp1 = self.program(29, program29, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n      <div class=\"clearfix\"></div>\n    </div>\n  ");
  return buffer;}
function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <div class=\"span12 control-group mlc\">\n              <div class=\"row-fluid\">\n                <div class=\"span4\"><span class=\"pull-right control-label\">");
  stack1 = depth0;
  stack2 = "services.reassign.step2.currentHost";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span>\n                </div>\n                <div class=\"span8\"><span>");
  stack1 = depth0;
  stack2 = "currentHostId";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span></div>\n              </div>\n            </div>\n          ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  <div class=\"row-fluid\">\n                    <div class=\"span4\">\n                      <div class=\"control-group\">\n                        <label class=\"pts pull-right\">\n                          ");
  stack1 = depth0;
  stack2 = "display_name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ":\n                        </label>\n                      </div>\n                    </div>\n                    <div class=\"span8\">\n                      ");
  stack1 = depth0;
  stack2 = "isServiceCoHost";
  stack3 = helpers['if'];
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(13, program13, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                    </div>\n                  </div>\n                ");
  return buffer;}
function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                        <div class=\"hostName\">\n                          ");
  stack1 = depth0;
  stack2 = "selectedHost";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "<i class=\"icon-asterisks\">&#10037;</i>\n                        </div>\n                      ");
  return buffer;}

function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                        <div ");
  stack1 = {};
  stack2 = "errorMessage:error: warnMessage:warning: :control-group";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                          ");
  stack1 = depth0;
  stack2 = "view.shouldUseInputs";
  stack3 = helpers['if'];
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(16, program16, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                          ");
  stack1 = depth0;
  stack2 = "showAddControl";
  stack3 = helpers['if'];
  tmp1 = self.program(18, program18, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                          ");
  stack1 = depth0;
  stack2 = "showRemoveControl";
  stack3 = helpers['if'];
  tmp1 = self.program(20, program20, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n                          <span rel=\"popover\" title=\"Warning\" ");
  stack1 = {};
  stack2 = "warnMessage";
  stack1['data-content'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                            ");
  stack1 = depth0;
  stack2 = "warnMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(22, program22, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                          </span>\n                          <span rel=\"popover\" title=\"Error\" ");
  stack1 = {};
  stack2 = "errorMessage";
  stack1['data-content'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                            ");
  stack1 = depth0;
  stack2 = "errorMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(24, program24, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                          </span>\n                        </div>\n                      ");
  return buffer;}
function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                            ");
  stack1 = depth0;
  stack2 = "App.InputHostView";
  stack3 = {};
  stack4 = "this";
  stack3['componentBinding'] = stack4;
  stack4 = "isInstalled";
  stack3['disabledBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                          ");
  return buffer;}

function program16(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                            ");
  stack1 = depth0;
  stack2 = "App.SelectHostView";
  stack3 = {};
  stack4 = "this";
  stack3['componentBinding'] = stack4;
  stack4 = "isInstalled";
  stack3['disabledBinding'] = stack4;
  stack4 = "content.host_name";
  stack3['optionValuePath'] = stack4;
  stack4 = "content.host_info";
  stack3['optionLabelPath'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                          ");
  return buffer;}

function program18(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                            ");
  stack1 = depth0;
  stack2 = "App.AddControlView";
  stack3 = {};
  stack4 = "component_name";
  stack3['componentNameBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                          ");
  return buffer;}

function program20(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n                            ");
  stack1 = depth0;
  stack2 = "App.RemoveControlView";
  stack3 = {};
  stack4 = "component_name";
  stack3['componentNameBinding'] = stack4;
  stack4 = "serviceComponentId";
  stack3['serviceComponentIdBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                          ");
  return buffer;}

function program22(depth0,data) {


  data.buffer.push("\n                              <i class=\"icon-warning-sign\"></i>\n                            ");}

function program24(depth0,data) {


  data.buffer.push("\n                              <i class=\"icon-remove-sign\"></i>\n                            ");}

function program26(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <div class=\"mapping-box round-corners well\">\n            <div class=\"hostString\"><span>");
  stack1 = depth0;
  stack2 = "hostInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span></div>\n            ");
  stack1 = depth0;
  stack2 = "masterServicesToDisplay";
  stack3 = helpers.each;
  tmp1 = self.program(27, program27, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </div>\n        ");
  return buffer;}
function program27(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              <span ");
  stack1 = {};
  stack2 = "isInstalled:assignedService:newService :round-corners";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "display_name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n            ");
  return buffer;}

function program29(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <div class=\"remaining-hosts round-corners well\">\n            <span><strong>");
  stack1 = depth0;
  stack2 = "remainingHosts";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</strong> ");
  stack1 = depth0;
  stack2 = "installer.step5.attention";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</span></div>\n        ");
  return buffer;}

function program31(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

  data.buffer.push("\n<div id=\"assign-masters\">\n  <h2>");
  stack1 = depth0;
  stack2 = "title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</h2>\n  <div class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "installer.step5.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    ");
  stack1 = depth0;
  stack2 = "view.coHostedComponentText";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  </div>\n  ");
  stack1 = depth0;
  stack2 = "controller.generalErrorMessages";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "msg";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "controller.generalWarningMessages";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "msg";
  stack7 = helpers.each;
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "controller.isLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(31, program31, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n  <div class=\"btn-area\">\n    <a class=\"btn pull-left installer-back-btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = {};
  stack4 = "true";
  stack3['href'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "submitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step6", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n  <div class=\"alert alert-error\">\n    ");
  stack1 = depth0;
  stack2 = "installer.step6.validationSlavesAndClients.hasIssues";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n    <a href=\"javascript:void(null);\" ");
  stack1 = depth0;
  stack2 = "openSlavesAndClientsIssues";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "installer.step6.validationSlavesAndClients.click";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    ");
  stack1 = depth0;
  stack2 = "installer.step6.validationSlavesAndClients.forDetails";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n  ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n\n          <th>\n            <a href=\"#\" ");
  stack1 = {};
  stack2 = "header.allChecked:selected:deselected header.isDisabled:remove-link";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              ");
  stack1 = depth0;
  stack2 = "header";
  stack3 = depth0;
  stack4 = "selectAllNodes";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "all";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a> &nbsp;|&nbsp; <a\n                  href=\"#\" ");
  stack1 = {};
  stack2 = "header.noChecked:selected:deselected header.isDisabled:remove-link";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n            ");
  stack1 = depth0;
  stack2 = "header";
  stack3 = depth0;
  stack4 = "deselectAllNodes";
  stack5 = {};
  stack6 = "controller";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "none";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n          </th>\n\n        ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "host";
  stack7 = helpers.each;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n          <tr>\n            ");
  stack1 = depth0;
  stack2 = "App.WizardStep6HostView";
  stack3 = {};
  stack4 = "host";
  stack3['hostBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "host.checkboxes";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "checkbox";
  stack7 = helpers.each;
  tmp1 = self.program(10, program10, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </tr>\n          <tr ");
  stack1 = {};
  stack2 = "host.anyMessage::hidden";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <td ");
  stack1 = {};
  stack2 = "view.columnCount";
  stack1['colspan'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"no-borders\">\n              ");
  stack1 = depth0;
  stack2 = "host.errorMessages";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "errorMsg";
  stack7 = helpers.each;
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "host.warnMessages";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "warnMsg";
  stack7 = helpers.each;
  tmp1 = self.program(14, program14, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            </td>\n          </tr>\n        ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n              <span class=\"trim_hostname\">");
  stack1 = depth0;
  stack2 = "host.hostName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n              ");
  stack1 = depth0;
  stack2 = "host.hasMaster";
  stack3 = helpers['if'];
  tmp1 = self.program(8, program8, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program8(depth0,data) {


  data.buffer.push("\n                <i class=icon-asterisks>&#10037;</i>\n              ");}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n              <td ");
  stack1 = {};
  stack2 = "checkbox.hasErrorMessage:error checkbox.hasWarnMessage:warning";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                <label class=\"checkbox\">\n                  <input ");
  stack1 = {};
  stack2 = "checkbox.checked";
  stack1['checked'] = stack2;
  stack2 = "checkbox.isDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "checkbox";
  stack3 = depth0;
  stack4 = "checkboxClick";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                          type=\"checkbox\"/>");
  stack1 = depth0;
  stack2 = "checkbox.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                </label>\n              </td>\n            ");
  return buffer;}

function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                  <div class=\"alert alert-error\">");
  stack1 = depth0;
  stack2 = "errorMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n              ");
  return buffer;}

function program14(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                <div class=\"alert alert-warning\">");
  stack1 = depth0;
  stack2 = "warnMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n              ");
  return buffer;}

  data.buffer.push("\n\n<div id=\"step6\">\n  <h2>");
  stack1 = depth0;
  stack2 = "view.title";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <div class=\"alert alert-info\">");
  stack1 = depth0;
  stack2 = "view.label";
  stack3 = {};
  stack4 = "true";
  stack3['unescaped'] = stack4;
  stack4 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("</div>\n  ");
  stack1 = depth0;
  stack2 = "anyGeneralIssues";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n  <div class=\"pre-scrollable\">\n    <table class=\"table\" id=\"component_assign_table\">\n      <thead>\n      <tr>\n        <th>");
  stack1 = depth0;
  stack2 = "common.host";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n        ");
  stack1 = depth0;
  stack2 = "controller.headers";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "header";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </tr>\n      </thead>\n      <tbody>\n      ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </tbody>\n    </table>\n  </div>\n  <div id=\"hosts\">\n    <div class=\"page-bar\">\n      <div class=\"items-on-page\">\n        <label>");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.rowsPerPageSelectView";
  stack3 = {};
  stack4 = "view.displayLength";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n      </div>\n      <div class=\"info\">");
  stack1 = depth0;
  stack2 = "view.paginationInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n      <div class=\"paging_two_button\">\n        ");
  stack1 = depth0;
  stack2 = "view.paginationFirst";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.paginationLeft";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.paginationRight";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.paginationLast";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </div>\n    </div>\n  </div>\n  <div class=\"btn-area\">\n    <a class=\"btn installer-back-btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "submitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "next";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step6/step6_issues_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    <div class=\"alert alert-error\">\n      <ul>\n        ");
  stack1 = depth0;
  stack2 = "errorMessage";
  stack3 = helpers['if'];
  tmp1 = self.program(2, program2, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "controller.generalErrorMessages";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "msg";
  stack7 = helpers.each;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </ul>\n    </div>\n  ");
  return buffer;}
function program2(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <li>");
  stack1 = depth0;
  stack2 = "errorMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</li>\n        ");
  return buffer;}

function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <li>");
  stack1 = depth0;
  stack2 = "msg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</li>\n        ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    <div class=\"alert alert-warning\">\n      <ul>\n        ");
  stack1 = depth0;
  stack2 = "controller.generalWarningMessages";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "msg";
  stack7 = helpers.each;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </ul>\n    </div>\n  ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <li>");
  stack1 = depth0;
  stack2 = "msg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</li>\n        ");
  return buffer;}

  data.buffer.push("\n\n<p>");
  stack1 = depth0;
  stack2 = "installer.step6.validationSlavesAndClients.popup.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</p>\n<div class=\"limited-height-2\">\n  ");
  stack1 = depth0;
  stack2 = "anyGeneralErrors";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n  ");
  stack1 = depth0;
  stack2 = "anyGeneralWarnings";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step7", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "App.ServicesConfigView";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

function program3(depth0,data) {


  data.buffer.push("\n    <div class=\"spinner\"></div>\n  ");}

  data.buffer.push("\n\n<div id=\"serviceConfig\">\n  <h2>");
  stack1 = depth0;
  stack2 = "installer.step7.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1));
  stack1 = depth0;
  stack2 = "controller.finnalyLoaded";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <div class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "installer.step7.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n\n  ");
  stack1 = depth0;
  stack2 = "isConfigsLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n\n  <div class=\"btn-area\">\n    <a class=\"btn installer-back-btn\" ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    <a class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step8", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n        <p>\n          <b>");
  stack1 = depth0;
  stack2 = "item.display_name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</b> : ");
  stack1 = depth0;
  stack2 = "item.config_value";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </p>\n      ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n            <li>\n              <p><span class=\"text text-info\">");
  stack1 = depth0;
  stack2 = "item.os_type";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " (");
  stack1 = depth0;
  stack2 = "item.repo_id";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "): <br/></span>");
  stack1 = depth0;
  stack2 = "item.base_url";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n            </li>\n          ");
  return buffer;}

function program5(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <p><b>");
  stack1 = depth0;
  stack2 = "menu.item.services";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</b></p>\n          <ul>\n            ");
  stack1 = depth0;
  stack2 = "controller.services";
  stack3 = helpers.each;
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          </ul>\n        ");
  return buffer;}
function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n              <li>\n                <em><b>");
  stack1 = depth0;
  stack2 = "display_name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</b></em>\n                <ul>\n                  ");
  stack1 = depth0;
  stack2 = "service_components";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "component";
  stack7 = helpers.each;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n                </ul>\n              </li>\n            ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n                    <li>\n                      <span class=\"text text-info\">");
  stack1 = depth0;
  stack2 = "component.display_name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " : </span>");
  stack1 = depth0;
  stack2 = "component.component_value";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                    </li>\n                  ");
  return buffer;}

  data.buffer.push("\n<div id=\"deploy-review\">\n  <h2>");
  stack1 = depth0;
  stack2 = "installer.step8.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <div class=\"alert alert-info\">\n    ");
  stack1 = depth0;
  stack2 = "installer.step8.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n\n  <div id=\"step8-content\" class=\"well pre-scrollable\">\n    <div id=\"step8-info\">\n      ");
  stack1 = depth0;
  stack2 = "controller.clusterInfo";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "item";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n\n      <p><b>");
  stack1 = depth0;
  stack2 = "controller.clusterInfo.repoInfo.display_name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</b>:</p>\n\n      <div>\n        <ul>\n          ");
  stack1 = depth0;
  stack2 = "controller.clusterInfo.repoInfo";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "item";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </ul>\n      </div>\n\n      <div>\n        ");
  stack1 = depth0;
  stack2 = "controller.services.length";
  stack3 = helpers['if'];
  tmp1 = self.program(5, program5, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n    </div>\n  </div>\n  <div class=\"btn-area\">\n    <a class=\"btn pull-left installer-back-btn\" ");
  stack1 = {};
  stack2 = "controller.isBackBtnDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "back";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">&larr; ");
  stack1 = depth0;
  stack2 = "common.back";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n    <a class=\"btn btn-success pull-right\"\n       id=\"spinner\" ");
  stack1 = {};
  stack2 = "controller.isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.deploy";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</a>\n    <a class=\"btn btn-info pull-right\" ");
  stack1 = depth0;
  stack2 = "printReview";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.print";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n  </div>\n</div>\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step8/step8_log_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression;


  data.buffer.push("\n<p>");
  stack1 = depth0;
  stack2 = "view.message";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</p>\n<div class=\"progress log_popup\">\n    <div class=\"bar\" ");
  stack1 = {};
  stack2 = "view.barWidth";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n    </div>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step9", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <a class=\"btn btn-primary\" href=\"#\" ");
  stack1 = depth0;
  stack2 = "retry";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <i class=\"icon-repeat icon-white\"></i>\n            ");
  stack1 = depth0;
  stack2 = "common.retry";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n          </a>\n        ");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6;
  data.buffer.push("\n            <li ");
  stack1 = {};
  stack2 = ":filter-status category.itemClass";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n              <a ");
  stack1 = depth0;
  stack2 = "category";
  stack3 = depth0;
  stack4 = "selectCategory";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">\n                ");
  stack1 = depth0;
  stack2 = "category.label";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n              </a>\n            </li>\n            ");
  stack1 = depth0;
  stack2 = "category.last";
  stack3 = helpers.unless;
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program4(depth0,data) {


  data.buffer.push("\n              <li class=\"divider\">|</li>\n            ");}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n            ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "host";
  stack7 = helpers.each;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n          ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              ");
  stack1 = depth0;
  stack2 = "App.HostStatusView";
  stack3 = {};
  stack4 = "host";
  stack3['objBinding'] = stack4;
  stack4 = "controller";
  stack3['controllerBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = self.program(8, program8, data);
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n            ");
  return buffer;}
function program8(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n              <td class=\"host\">\n                <span title=\"");
  stack1 = depth0;
  stack2 = "host.name";
  stack3 = helpers.unbound;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\" class=\"trim_hostname\">");
  stack1 = depth0;
  stack2 = "host.name";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n              </td>\n              <td>\n                <div class=\"progress-bar pull-left\">\n                  <div ");
  stack1 = {};
  stack2 = "view.isHostCompleted::progress-striped view.isHostCompleted::active view.barColor :progress";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n                    <div class=\"bar\" ");
  stack1 = {};
  stack2 = "view.barWidth";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></div>\n                  </div>\n                </div>\n                <div class=\"progress-percentage pull-left\">");
  stack1 = depth0;
  stack2 = "host.progress";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "%</div>\n              </td>\n              <td>\n                <a ");
  stack1 = {};
  stack2 = "view.isFailed:text-error view.isSuccess:text-success view.isWarning:text-warning";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                    href=\"javascript:void(null)\"\n                    data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "hostLogPopup";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "host.message";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</a>\n              </td>\n              ");
  return buffer;}

function program10(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <tr>\n            <td colspan=\"3\"><p>");
  stack1 = depth0;
  stack2 = "installer.step3.hosts.noHosts";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</p></td>\n          </tr>\n          ");
  return buffer;}

function program12(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n      <p ");
  stack1 = {};
  stack2 = "view.resultMsgColor :alert";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "view.resultMsg";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        ");
  stack1 = depth0;
  stack2 = "view.isHostHeartbeatLost";
  stack3 = helpers['if'];
  tmp1 = self.program(13, program13, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </p>\n    ");
  return buffer;}
function program13(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4;
  data.buffer.push("\n          <a href=\"javascript:void(null)\"\n             data-toggle=\"modal\" ");
  stack1 = depth0;
  stack2 = "hostWithInstallFailed";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.showDetails";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n        ");
  return buffer;}

  data.buffer.push("\n\n<div id=\"deploy\">\n\n  <h2>");
  stack1 = depth0;
  stack2 = "installer.step9.header";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</h2>\n\n  <p class=\"alert alert-info\">");
  stack1 = depth0;
  stack2 = "installer.step9.body";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</p>\n\n  <div id=\"overallProgress\">\n    <div class=\"row-fluid\">\n      <div class=\"span10\">\n        <div ");
  stack1 = {};
  stack2 = "view.isStepCompleted::progress-striped view.isStepCompleted::active view.barColor :progress";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          <div class=\"bar\" ");
  stack1 = {};
  stack2 = "view.barWidth";
  stack1['style'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n          </div>\n        </div>\n      </div>\n      <div class=\"span2\">");
  stack1 = depth0;
  stack2 = "view.progressMessage";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n    </div>\n  </div>\n\n  <div class=\"box\">\n    <div class=\"box-header\">\n      <div class=\"pull-left\">\n        ");
  stack1 = depth0;
  stack2 = "controller.showRetry";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n      <!-- filter by host level -->\n      <div id=\"host-filter\" class=\"pull-right\">\n        <ul class=\"clearfix\">\n          <li class=\"first\">");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":</li>\n          ");
  stack1 = depth0;
  stack2 = "view.categories";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "category";
  stack7 = helpers.each;
  tmp1 = self.program(3, program3, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </ul>\n      </div>\n    </div>\n    <div class=\"pre-scrollable\">\n      <table id=\"deploy-status-by-host\" class=\"table table-bordered table-striped\">\n        <thead>\n        <tr>\n          <th class=\"host\">\n            ");
  stack1 = depth0;
  stack2 = "common.host";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n          </th>\n          <th class=\"status\">");
  stack1 = depth0;
  stack2 = "common.status";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n          <!--  given by the parsing function that parses data from bootstrap call -->\n          <th class=\"message\">");
  stack1 = depth0;
  stack2 = "common.message";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n          <!-- retrieved from local storage initially -->\n        </tr>\n        </thead>\n\n        <tbody>\n          ");
  stack1 = depth0;
  stack2 = "view.pageContent";
  stack3 = helpers['if'];
  tmp1 = self.program(6, program6, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(10, program10, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        </tbody>\n      </table>\n    </div>\n    <div id=\"hosts\">\n      <div class=\"page-bar\">\n        <div class=\"selected-hosts-info pull-left\">\n          ");
  stack1 = depth0;
  stack2 = "view.filteredHostsInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          -\n          <a ");
  stack1 = depth0;
  stack2 = "showAllHosts";
  stack3 = {};
  stack4 = "view";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " href=\"#\">");
  stack1 = depth0;
  stack2 = "tableView.filters.showAll";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n        </div>\n        <div class=\"items-on-page\">\n          <label>");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ": ");
  stack1 = depth0;
  stack2 = "view.rowsPerPageSelectView";
  stack3 = {};
  stack4 = "view.displayLength";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</label>\n        </div>\n        <div class=\"info\">");
  stack1 = depth0;
  stack2 = "view.paginationInfo";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"paging_two_button\">\n          ");
  stack1 = depth0;
  stack2 = "view.paginationFirst";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "view.paginationLeft";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "view.paginationRight";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n          ");
  stack1 = depth0;
  stack2 = "view.paginationLast";
  stack3 = helpers.view;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n    </div>\n  </div>\n\n  <div>\n    ");
  stack1 = depth0;
  stack2 = "view.resultMsg";
  stack3 = helpers['if'];
  tmp1 = self.program(12, program12, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    <div class=\"btn-area\">\n      <button\n          class=\"btn btn-success pull-right\" ");
  stack1 = {};
  stack2 = "isSubmitDisabled";
  stack1['disabled'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "submit";
  stack3 = {};
  stack4 = "controller";
  stack3['target'] = stack4;
  stack4 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">");
  stack1 = depth0;
  stack2 = "common.next";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + " &rarr;</button>\n    </div>\n  </div>\n\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step9/step9HostTasksLogPopup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, foundHelper, tmp1, self=this, functionType="function", helperMissing=helpers.helperMissing, undef=void 0, escapeExpression=this.escapeExpression;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  <div>\n    ");
  stack1 = depth0;
  stack2 = "installer.step9.host.heartbeat_lost_popup";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  </div>\n");
  return buffer;}

function program3(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n  ");
  stack1 = depth0;
  stack2 = "view.isNoTasksScheduled";
  stack3 = helpers['if'];
  tmp1 = self.program(4, program4, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(6, program6, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;}
function program4(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    ");
  stack1 = depth0;
  stack2 = "installer.step9.host.status.noTasks";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "\n  ");
  return buffer;}

function program6(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7;
  data.buffer.push("\n    <div ");
  stack1 = {};
  stack2 = "view.isLogWrapHidden::hidden :task-list-main-warp";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <div class=\"task-top-wrap\">\n        <div class=\"task-name-top task-detail-log-rolename\">");
  stack1 = depth0;
  stack2 = "common.tasks";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n        <div class=\"select-wrap\">\n          ");
  stack1 = depth0;
  stack2 = "common.show";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + ":\n          ");
  stack1 = depth0;
  stack2 = "Ember.Select";
  stack3 = {};
  stack4 = "view.categories";
  stack3['contentBinding'] = stack4;
  stack4 = "content.value";
  stack3['optionValuePath'] = stack4;
  stack4 = "content.label";
  stack3['optionLabelPath'] = stack4;
  stack4 = "view.category";
  stack3['selectionBinding'] = stack4;
  stack4 = helpers.view;
  tmp1 = {};
  tmp1.hash = stack3;
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack4.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n        </div>\n      </div>\n      <div id=\"host-log\">\n        ");
  stack1 = depth0;
  stack2 = "view.tasks";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "taskInfo";
  stack7 = helpers.each;
  tmp1 = self.program(7, program7, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n        ");
  stack1 = depth0;
  stack2 = "view.isEmptyList";
  stack3 = helpers['if'];
  tmp1 = self.program(9, program9, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n      </div>\n    </div>\n\n    <div ");
  stack1 = {};
  stack2 = "view.isLogWrapHidden:hidden :task-detail-info";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n      <div class=\"task-top-wrap\">\n        <a class=\"task-detail-back\" href=\"javascript:void(null)\" ");
  stack1 = depth0;
  stack2 = "backToTaskList";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ><i\n                class=\"icon-arrow-left\"></i>&nbsp;");
  stack1 = depth0;
  stack2 = "common.tasks";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n\n        <div>\n          <i ");
  stack1 = {};
  stack2 = "view.openedTask.status :task-detail-status-ico view.openedTask.icon";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n          <div class=\"task-detail-ico-wrap\">\n            <a title=\"Click to Copy\" ");
  stack1 = depth0;
  stack2 = "taskInfo";
  stack3 = depth0;
  stack4 = "textTrigger";
  stack5 = {};
  stack6 = "view";
  stack5['target'] = stack6;
  stack6 = helpers.action;
  tmp1 = {};
  tmp1.hash = stack5;
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack6.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-copy\"><i class=\"icon-copy\"></i> ");
  stack1 = depth0;
  stack2 = "common.copy";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n            <a title=\"Open in New Window\" ");
  stack1 = depth0;
  stack2 = "openTaskLogInDialog";
  stack3 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-detail-open-dialog\"><i class=\"icon-external-link\"></i> ");
  stack1 = depth0;
  stack2 = "common.open";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</a>\n          </div>\n          <span class=\"task-detail-log-rolename\">");
  stack1 = depth0;
  stack2 = "view.openedTask.role";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "view.openedTask.command";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</span>\n        </div>\n\n      </div>\n      ");
  stack1 = depth0;
  stack2 = "view.isTaskLoaded";
  stack3 = helpers['if'];
  tmp1 = self.program(11, program11, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(13, program13, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </div>\n  ");
  return buffer;}
function program7(depth0,data) {

  var buffer = '', stack1, stack2, stack3, stack4, stack5;
  data.buffer.push("\n          <div ");
  stack1 = {};
  stack2 = "taskInfo.isVisible::hidden :log-list-wrap";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + ">\n            <div ");
  stack1 = depth0;
  stack2 = "taskInfo";
  stack3 = depth0;
  stack4 = "toggleTaskLog";
  stack5 = helpers.action;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack5.call(depth0, stack4, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " class=\"task-list-line-cursor\">\n              <div class=\"operation-name-icon-wrap\">\n                <i ");
  stack1 = {};
  stack2 = "taskInfo.status taskInfo.icon";
  stack1['class'] = stack2;
  stack2 = helpers.bindAttr;
  tmp1 = {};
  tmp1.hash = stack1;
  tmp1.contexts = [];
  tmp1.data = data;
  stack1 = stack2.call(depth0, tmp1);
  data.buffer.push(escapeExpression(stack1) + "></i>\n                <a href=\"#\">\n                  ");
  stack1 = depth0;
  stack2 = "taskInfo.role";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " ");
  stack1 = depth0;
  stack2 = "taskInfo.command";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n                </a>\n              </div>\n              <div class=\"show-details\"><i class=\"icon-caret-right\"></i></div>\n            </div>\n          </div>\n        ");
  return buffer;}

function program9(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <div class=\"log-list-wrap\">");
  stack1 = depth0;
  stack2 = "installer.step9.hostLog.popup.noTasksToShow";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</div>\n        ");
  return buffer;}

function program11(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n          <div class=\"task-detail-log-info\">\n              <div class=\"content-area\">\n                  <div class=\"task-detail-log-clipboard-wrap\"></div>\n                  <div class=\"task-detail-log-maintext\">\n                      <h5>stderr: &nbsp; <span class=\"muted\">");
  stack1 = depth0;
  stack2 = "view.openedTask.errorLog";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " </span></h5>\n                      <pre class=\"stderr\">");
  stack1 = depth0;
  stack2 = "view.openedTask.stderr";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</pre>\n                      <h5>stdout: &nbsp; <span class=\"muted\"> ");
  stack1 = depth0;
  stack2 = "view.openedTask.outputLog";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + " </span></h5>\n                      <pre class=\"stdout\">");
  stack1 = depth0;
  stack2 = "view.openedTask.stdout";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "</pre>\n                  </div>\n              </div>\n          </div>\n      ");
  return buffer;}

function program13(depth0,data) {


  data.buffer.push("\n        <div class=\"spinner\"></div>\n      ");}

  data.buffer.push("\n\n");
  stack1 = depth0;
  stack2 = "view.isHeartbeatLost";
  stack3 = helpers['if'];
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.program(3, program3, data);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n");
  return buffer;
});
module.exports = module.id;
});

require.register("templates/wizard/step9/step9_install_host_popup", function(exports, require, module) {
Ember.TEMPLATES[module.id] = Ember.Handlebars.template(function anonymous(Handlebars,depth0,helpers,partials,data) {
helpers = helpers || Ember.Handlebars.helpers;
  var buffer = '', stack1, stack2, stack3, stack4, stack5, stack6, stack7, foundHelper, tmp1, self=this, escapeExpression=this.escapeExpression, functionType="function", helperMissing=helpers.helperMissing, undef=void 0;

function program1(depth0,data) {

  var buffer = '', stack1, stack2, stack3;
  data.buffer.push("\n    <tr>\n      <td>\n        ");
  stack1 = depth0;
  stack2 = "host.hostName";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </td>\n      <td>\n        ");
  stack1 = depth0;
  stack2 = "host.componentNames";
  stack3 = helpers._triageMustache;
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  stack1 = stack3.call(depth0, stack2, tmp1);
  data.buffer.push(escapeExpression(stack1) + "\n      </td>\n    </tr>\n  ");
  return buffer;}

  data.buffer.push("\n\n<div id=\"install-fail-popup\" class=\"pre-scrollable\">\n  <table class=\"table table-striped\">\n    <thead>\n    <tr>\n      <th class=\"span3\">");
  stack1 = depth0;
  stack2 = "common.host";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n      <th class=\"span3\">");
  stack1 = depth0;
  stack2 = "common.components";
  foundHelper = helpers['t'];
  stack3 = foundHelper || depth0['t'];
  tmp1 = {};
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack1);
  tmp1.data = data;
  if(typeof stack3 === functionType) { stack1 = stack3.call(depth0, stack2, tmp1); }
  else if(stack3=== undef) { stack1 = helperMissing.call(depth0, "t", stack2, tmp1); }
  else { stack1 = stack3; }
  data.buffer.push(escapeExpression(stack1) + "</th>\n    </tr>\n    </thead>\n    <tbody>\n  ");
  stack1 = depth0;
  stack2 = "view.failedHosts";
  stack3 = depth0;
  stack4 = "in";
  stack5 = depth0;
  stack6 = "host";
  stack7 = helpers.each;
  tmp1 = self.program(1, program1, data);
  tmp1.hash = {};
  tmp1.contexts = [];
  tmp1.contexts.push(stack5);
  tmp1.contexts.push(stack3);
  tmp1.contexts.push(stack1);
  tmp1.fn = tmp1;
  tmp1.inverse = self.noop;
  tmp1.data = data;
  stack1 = stack7.call(depth0, stack6, stack4, stack2, tmp1);
  if(stack1 || stack1 === 0) { data.buffer.push(stack1); }
  data.buffer.push("\n    </tbody>\n  </table>\n</div>");
  return buffer;
});
module.exports = module.id;
});

require.register("utils/action_sequence", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
 * Utility to define sequence of actions (sync or async either) and execute them in exact order
 * methods executed in order that described in sequence
 * each action obtain result of previous action <code>previousResponse</code>
 * For example:
 * loadMap: {
 *  '0': {
 *   type: 'async',
 *   callback: function (previousResponse) {
 *     //method1
 *   }
 *  },
 *  '1': {
 *    type: 'sync',
 *    callback: function (previousResponse) {
 *      //method2
 *    }
 *  }
 * }
 * However method1 is async, method2 will be executed only after method1 is completed
 * @type {*}
 */
App.actionSequence = Em.Object.extend({
  /**
   * List of actions which will be executed
   * format:
   * [
   *   {
   *     type: 'async',
   *     callback: function(previousResponse) {
   *       //async method should return Deferred object to continue sequence
   *     }
   *   },
   *   {
   *     type: 'sync',
   *     callback: function(previousResponse) {
   *       //code
   *     }
   *   }
   * ]
   * @type {Object[]}
   */
  sequence: [],
  /**
   * specific context for methods executed in actions
   */
  context: this,
  /**
   * Function called after sequence is completed
   * @type {callback}
   */
  finishedCallback: Em.K,
  /**
   * counter of actions
   */
  actionCounter: 0,
  /**
   * set sequence
   * @param sequence
   * @return {object}
   */
  setSequence: function (sequence) {
    if (Array.isArray(sequence)) {
      this.set('sequence', sequence);
    } else {
      console.log('ERROR: passed sequence has incorrect type');
    }
    return this;
  },
  /**
   * start executing sequence of actions
   */
  start: function () {
    this.set('actionCounter', this.get('sequence.length'));
    this.runNextAction(0, null);
  },
  /**
   * set <code>finishedCallback</code>
   * @param callback
   * @return {object}
   */
  onFinish: function (callback) {
    if (typeof(callback) === 'function') {
      this.set('finishedCallback', callback);
    }
    return this;
  },
  /**
   * run next action in sequence
   * @param index
   * @param previousResponse
   * @return {Boolean}
   */
  runNextAction: function (index, previousResponse) {
    var action = this.get('sequence')[index];
    var context = this.get('context');
    var self = this;

    index++;
    this.decrementProperty('actionCounter');

    if (this.get('actionCounter') >= 0 && !Em.isNone(action)) {
      if (action.type === 'sync') {
        this.runNextAction(index, action.callback.call(context, previousResponse));
        return true;
      } else if (action.type === 'async') {
        action.callback.call(context, previousResponse).done(function (response) {
          self.runNextAction(index, response);
        });
        return true;
      } else {
        console.log('WARNING: action has incorrect type, action skipped');
      }
    } else {
      //if no more actions left then finish sequence
      this.finishedCallback();
    }
    return false;
  }
});


});

require.register("utils/ajax/ajax", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');

/**
 * Config for each ajax-request
 *
 * Fields example:
 *  mock - testMode url
 *  real - real url (without API prefix)
 *  type - request type (also may be defined in the format method)
 *  format - function for processing ajax params after default formatRequest. May be called with one or two parameters (data, opt). Return ajax-params object
 *  testInProduction - can this request be executed on production tests (used only in tests)
 *
 * @type {Object}
 */
var urls = {

  'common.cluster.update' : {
    'type': 'PUT',
    'real': '/clusters/{clusterName}',
    'mock': '/data/wizard/deploy/poll_1.json',
    'format': function (data) {
      return {
        data: JSON.stringify(data.data)
      };
    }
  },

  'common.services.update' : {
    'real': '/clusters/{clusterName}/services?{urlParams}',
    'mock': '/data/wizard/deploy/poll_1.json',
    'format': function (data) {
      return {
        type: 'PUT',
        data: JSON.stringify({
          RequestInfo: {
            "context": data.context,
            "operation_level": {
              "level": "CLUSTER",
              "cluster_name" : data.clusterName
            }
          },
          Body: {
            ServiceInfo: data.ServiceInfo
          }
        })
      };
    }
  },

  'common.service.update' : {
    'real': '/clusters/{clusterName}/services/{serviceName}',
    'mock': '/data/wizard/deploy/poll_1.json',
    'format': function (data) {
      return {
        type: 'PUT',
        data: JSON.stringify({
          RequestInfo: {
            "context": data.context,
            "operation_level": {
              "level": "SERVICE",
              "cluster_name" : data.clusterName,
              "service_name" : data.serviceName
            }
          },
          Body: {
            ServiceInfo: data.ServiceInfo
          }
        })
      };
    }
  },

  'common.service_component.info' : {
    'real': '/clusters/{clusterName}/services/{serviceName}/components/{componentName}?{urlParams}',
    'mock': '/data/wizard/deploy/poll_1.json'
  },

  'common.host_component.update': {
    'real': '/clusters/{clusterName}/host_components',
    'mock': '',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify({
          RequestInfo: {
            context: data.context,
            query: data.query
          },
          Body: {
            "HostRoles": data.HostRoles
          }
        })
      }
    }
  },

  'common.host.host_components.update': {
    'real': '/clusters/{clusterName}/hosts/{hostName}/host_components?{urlParams}',
    'mock': '',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify({
          RequestInfo: {
            "context": data.context,
            "operation_level": {
              level: "HOST",
              cluster_name: data.clusterName,
              host_names: data.hostName
            },
            query: data.query
          },
          Body: {
            "HostRoles": data.HostRoles
          }
        })
      }
    }
  },

  'common.host.host_component.update': {
    'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}?{urlParams}',
    'mock': '/data/wizard/deploy/2_hosts/poll_9.json',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify({
          RequestInfo: {
            "context": data.context,
            "operation_level": {
              level: "HOST_COMPONENT",
              cluster_name: data.clusterName,
              host_name: data.hostName,
              service_name: data.serviceName || null
            }
          },
          Body: {
            "HostRoles": data.HostRoles
          }
        })
      }
    }
  },

  'common.service.configurations': {
    'real':'/clusters/{clusterName}',
    'mock':'',
    'format': function (data) {
      return {
        type: 'PUT',
        data: JSON.stringify({
          Clusters: {
            desired_config: data.desired_config
          }
        })
      }
    }
  },

  'common.across.services.configurations': {
    'type': 'PUT',
    'real':'/clusters/{clusterName}',
    'mock':'/data/services/ambari.json',
    'format': function(data) {
      return {
        dataType: 'text',
        data: data.data
      }
    }
  },

  'common.request.polling': {
    'real': '/clusters/{clusterName}/requests/{requestId}?fields=tasks/Tasks/request_id,tasks/Tasks/command,tasks/Tasks/command_detail,tasks/Tasks/start_time,tasks/Tasks/end_time,tasks/Tasks/exit_code,tasks/Tasks/host_name,tasks/Tasks/id,tasks/Tasks/role,tasks/Tasks/status,tasks/Tasks/structured_out,Requests/*&tasks/Tasks/stage_id={stageId}',
    'mock': '/data/background_operations/host_upgrade_tasks.json'
  },

  'service.ambari': {
    'real': '/services/AMBARI?fields=components/RootServiceComponents',
    'mock': '/data/services/ambari.json'
  },

  'service.flume.agent.command': {
    'real': '/clusters/{clusterName}/hosts/{host}/host_components/FLUME_HANDLER',
    'mock': '',
    'format': function (data) {
      return {
        type: 'PUT',
        data: JSON.stringify({
          "RequestInfo": {
            "context": data.context,
            "flume_handler": data.agentName,
            "operation_level": {
              level: "HOST_COMPONENT",
              cluster_name: data.clusterName,
              service_name: "FLUME",
              host_name: data.host
            }
          },
          "Body": {
            "HostRoles": {
              "state": data.state
            }
          }
        })
      }
    }
  },

  'common.host_components.update': {
    'real': '/clusters/{clusterName}/host_components?{urlParams}',
    'mock': '/data/wizard/deploy/poll_1.json',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify({
          RequestInfo: {
            "context": data.context,
            "operation_level": {
              level: data.level || "CLUSTER",
              cluster_name: data.clusterName
            },
            query: data.query
          },
          Body: {
            "HostRoles": data.HostRoles
          }
        })
      }
    }
  },

  'common.service.passive': {
    'real': '/clusters/{clusterName}/services/{serviceName}',
    'mock': '',
    'format': function (data) {
      return {
        type: 'PUT',
        data: JSON.stringify({
          RequestInfo: {
            "context": data.requestInfo
          },
          Body: {
            ServiceInfo: {
              maintenance_state: data.passive_state
            }
          }
        })
      };
    }
  },

  'common.host.host_component.passive': {
    'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}',
    'mock': '',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify({
          RequestInfo: {
            "context": data.context
          },
          Body: {
            HostRoles: {
              maintenance_state: data.passive_state
            }
          }
        })
      };
    }
  },

  'common.host.with_host_component': {
    'real': '/clusters/{clusterName}/hosts?host_components/HostRoles/component_name={componentName}&minimal_response=true',
    'mock': ''
  },

  'common.delete.host': {
    'real': '/clusters/{clusterName}/hosts/{hostName}',
    'type': 'DELETE'
  },
  'common.delete.host_component': {
    'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}',
    'type': 'DELETE'
  },
  'common.delete.user': {
    'real': '/users/{user}',
    'type': 'DELETE'
  },
  'common.delete.config_group': {
    'real': '/clusters/{clusterName}/config_groups/{id}',
    'type': 'DELETE'
  },
  'common.delete.cluster': {
    'real': '/clusters/{name}',
    'type': 'DELETE'
  },
  'common.delete.service': {
    'real': '/clusters/{clusterName}/services/{serviceName}',
    'mock': '/data/services/ambari.json',
    'type': 'DELETE'
  },
  'common.delete.request_schedule': {
    'real': '/clusters/{clusterName}/request_schedules/{request_schedule_id}',
    'type': 'DELETE'
  },
  'alerts.load_alert_groups': {
    'real': '/clusters/{clusterName}/alert_groups?fields=*',
    'mock': 'data/alerts/alertGroups.json'
  },
  'alerts.load_an_alert_group': {
    'real': '/clusters/{clusterName}/alert_groups/{group_id}',
    'mock': 'data/alerts/alertGroup.json'
  },
  'alert_groups.create': {
    'real': '/clusters/{clusterName}/alert_groups',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          "AlertGroup": {
            "name": data.name,
            "definitions": data.definitions,
            "targets": data.targets
          }
        })
      };
    }
  },
  'alert_groups.update': {
    'real': '/clusters/{clusterName}/alert_groups/{group_id}',
    'mock': '',
    'format': function (data) {
      return {
        type: 'PUT',
        data: JSON.stringify({
          "AlertGroup": {
            "name": data.name,
            "definitions": data.definitions,
            "targets": data.targets
          }
        })
      };
    }
  },
  'alert_groups.delete': {
    'real': '/clusters/{clusterName}/alert_groups/{group_id}',
    'mock': '',
    'format': function (data) {
      return {
        type: 'DELETE'
      };
    }
  },
  'alerts.load_all_alert_definitions': {
    'real': '/clusters/{clusterName}/alert_definitions?fields=*',
    'mock': 'data/alerts/alertDefinitions.json'
  },
  'alerts.notifications': {
    'real': '/alert_targets?fields=*',
    'mock': '/data/alerts/alertNotifications.json'
  },
  'alerts.instances': {
    'real': '/clusters/{clusterName}/alerts?fields=*',
    'mock': '/data/alerts/alert_instances.json'
  },
  'alerts.instances.unhealthy': {
    'real': '/clusters/{clusterName}/alerts?fields=*&Alert/state.in(CRITICAL,WARNING)&{paginationInfo}',
    'mock': '/data/alerts/alert_instances.json'
  },
  'alerts.instances.by_definition': {
    'real': '/clusters/{clusterName}/alerts?fields=*&Alert/definition_id={definitionId}',
    'mock': '/data/alerts/alert_instances.json'
  },
  'alerts.instances.by_host': {
    'real': '/clusters/{clusterName}/alerts?fields=*&Alert/host_name={hostName}',
    'mock': '/data/alerts/alert_instances.json'
  },
  'alerts.update_alert_definition': {
    'real': '/clusters/{clusterName}/alert_definitions/{id}',
    'mock': '',
    'format': function (data) {
      return {
        type: 'PUT',
        data: JSON.stringify(data.data)
      }
    }
  },
  'alerts.create_alert_definition': {
    'real': '/clusters/{clusterName}/alert_definitions/',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify(data.data)
      }
    }
  },
  'alerts.delete_alert_definition': {
    'real': '/clusters/{clusterName}/alert_definitions/{id}',
    'mock': '',
    'type': 'DELETE'
  },
  'alerts.create_alert_notification': {
    'real': '/alert_targets?{urlParams}',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify(data.data)
      }
    }
  },
  'alerts.update_alert_notification': {
    'real': '/alert_targets/{id}',
    'mock': '',
    'format': function (data) {
      return {
        type: 'PUT',
        data: JSON.stringify(data.data)
      }
    }
  },
  'alerts.delete_alert_notification': {
    'real': '/alert_targets/{id}',
    'mock': '',
    'type': 'DELETE'
  },
  'alerts.get_instances_history': {
    'real': '/clusters/{clusterName}/alert_history?(AlertHistory/definition_name={definitionName})&(AlertHistory/timestamp>={timestamp})',
    'mock': '/data/alerts/alert_instances_history.json'
  },
  'background_operations.get_most_recent': {
    'real': '/clusters/{clusterName}/requests?to=end&page_size={operationsCount}&fields=Requests',
    'mock': '/data/background_operations/list_on_start.json',
    'testInProduction': true
  },
  'background_operations.get_by_request': {
    'real': '/clusters/{clusterName}/requests/{requestId}?fields=*,tasks/Tasks/request_id,tasks/Tasks/command,tasks/Tasks/command_detail,tasks/Tasks/start_time,tasks/Tasks/end_time,tasks/Tasks/exit_code,tasks/Tasks/host_name,tasks/Tasks/id,tasks/Tasks/role,tasks/Tasks/status&minimal_response=true',
    'mock': '/data/background_operations/task_by_request{requestId}.json',
    'testInProduction': true
  },
  'background_operations.get_by_task': {
    'real': '/clusters/{clusterName}/requests/{requestId}/tasks/{taskId}',
    'mock': '/data/background_operations/list_on_start.json',
    'testInProduction': true
  },
  'background_operations.abort_request': {
    'real': '/clusters/{clusterName}/requests/{requestId}',
    'mock': '',
    'format': function () {
      return {
        type: 'PUT',
        data: JSON.stringify({
          "Requests": {
            "request_status": "ABORTED",
            "abort_reason": Em.I18n.t('hostPopup.bgop.abortRequest.reason')
          }
        })
      };
    }
  },
  'service.item.smoke': {
    'real': '/clusters/{clusterName}/requests',
    'mock': '/data/wizard/deploy/poll_1.json',
    'format': function (data) {
      var requestData = {
        "RequestInfo": {
          "context": data.displayName + " Service Check",
          "command": data.actionName
        },
        "Requests/resource_filters": [{"service_name": data.serviceName}]
      };
      if (data.operationLevel) {
        requestData.RequestInfo.operation_level = data.operationLevel;
      }
      return {
        'type': 'POST',
        data: JSON.stringify(requestData)
      };
    }
  },
  'service.item.rebalanceHdfsNodes': {
    'real': '/clusters/{clusterName}/requests',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          RequestInfo: {
            'context': Em.I18n.t('services.service.actions.run.rebalanceHdfsNodes.context'),
            'command': 'REBALANCEHDFS',
            'namenode': JSON.stringify({threshold: data.threshold})
          },
          "Requests/resource_filters": [{
            'service_name': 'HDFS',
            'component_name': 'NAMENODE',
            'hosts': data.hosts
          }]
        })
      }
    }
  },

  'cancel.background.operation': {
    'real': '/clusters/{clusterName}/requests/{requestId}',
    'mock': '',
    'format': function (data) {
      return {
        type: 'PUT',
        data: JSON.stringify({
          RequestInfo: {
            'context': 'Cancel operation',
            "parameters": {
              "cancel_policy": "SIGKILL"
            }
          },
          "Requests/request_status": 'ABORTED',
          "Requests/abort_reason": "Cancel background operation"
        })
      }
    }
  },


  'service.item.refreshQueueYarnRequest': {
    'real': '/clusters/{clusterName}/requests',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          RequestInfo: {
            'context': data.context,
            'command': data.command,
            'parameters/forceRefreshConfigTags': data.forceRefreshConfigTags
          },
          "Requests/resource_filters": [{
            "service_name": data.serviceName,
            "component_name": data.componentName,
            'hosts': data.hosts
          }]
        })
      }
    }
  },

  'service.item.startStopLdapKnox': {
    'real': '/clusters/{clusterName}/requests',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          RequestInfo: {
            'context': data.context,
            'command': data.command
          },
          "Requests/resource_filters": [{
            "service_name": data.serviceName,
            "component_name": data.componentName,
            'hosts': data.host
          }]
        })
      }
    }
  },

  'service.item.executeCustomCommand': {
    'real': '/clusters/{clusterName}/requests',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          RequestInfo: {
            'context': data.context,
            'command': data.command
          },
          "Requests/resource_filters": [{
            "service_name": data.serviceName,
            "component_name": data.componentName,
            'hosts': data.hosts
          }]
        })
      }
    }
  },

  'service.load_config_groups': {
    'real': '/clusters/{clusterName}/config_groups?ConfigGroup/tag={serviceName}&fields=*',
    'mock': '/data/configurations/config_group.json'
  },
  'reassign.load_configs': {
    'real': '/clusters/{clusterName}/configurations?{urlParams}',
    'mock': ''
  },

  'reassign.save_configs': {
    'real': '/clusters/{clusterName}',
    'mock': '',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify({
          Clusters: {
            desired_config: {
              "type": data.siteName,
              "tag": 'version' + (new Date).getTime(),
              "properties": data.properties,
              "service_config_version_note": data.service_config_version_note

            }
          }
        })
      }
    }
  },
  'config.cluster': {
    'real': '{stackVersionUrl}/configurations?fields=*',
    'mock': ''
  },
  'config.advanced': {
    'real': '{stackVersionUrl}/services/{serviceName}/configurations?fields=*',
    'mock': '/data/wizard/stack/hdp/version{stackVersion}/{serviceName}.json'
  },
  'config.advanced.partial': {
    'real': '{stackVersionUrl}/services/?StackServices/service_name.in({serviceList})&fields=configurations/*{queryFilter}',
    'mock': ''
  },
  'config.config_types': {
    'real': '{stackVersionUrl}/services/{serviceName}?fields=StackServices/config_types',
    'mock': ''
  },
  'config.tags': {
    'real': '/clusters/{clusterName}?fields=Clusters/desired_configs',
    'mock': '/data/clusters/cluster.json'
  },
  'config.tags_and_groups': {
    'real': '/clusters/{clusterName}?fields=Clusters/desired_configs,config_groups/*{urlParams}',
    'mock': '/data/clusters/tags_and_groups.json'
  },
  'config.ambari.database.info': {
    'real': '/services/AMBARI/components/AMBARI_SERVER?fields=hostComponents/RootServiceHostComponents/properties/server.jdbc.database_name,hostComponents/RootServiceHostComponents/properties/server.jdbc.url',
    'mock': ''
  },
  'config_groups.all_fields': {
    'real': '/clusters/{clusterName}/config_groups?fields=*',
    'mock': ''
  },
  'config_groups.get_config_group_by_id': {
    'real': '/clusters/{clusterName}/config_groups/{id}',
    'mock': ''
  },
  'config_groups.update_config_group': {
    'real': '/clusters/{clusterName}/config_groups/{id}',
    'mock': '',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify(
          [
            data.configGroup
          ]
        )
      }
    }
  },
  'config.on_site': {
    'real': '/clusters/{clusterName}/configurations?{params}',
    'mock': '/data/configurations/cluster_level_configs.json?{params}'
  },
  'config.host_overrides': {
    'real': '/clusters/{clusterName}/configurations?{params}',
    'mock': '/data/configurations/host_level_overrides_configs.json?{params}'
  },

  'config.cluster_env_site': {
    'real': '/clusters/{clusterName}/configurations?type=cluster-env',
    'mock': '/data/configuration/cluster_env_site.json'
  },

  'host.host_component.add_new_component': {
    'real': '/clusters/{clusterName}/hosts?Hosts/host_name={hostName}',
    'mock': '/data/wizard/deploy/poll_1.json',
    'format': function (data) {
      return {
        type: 'POST',
        data: data.data
      }
    }
  },

  'host.host_component.slave_desired_admin_state': {
    'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}/?fields=HostRoles/desired_admin_state',
    'mock': '/data/hosts/HDP2/decommission_state.json'
  },
  'host.host_component.decommission_status': {
    'real': '/clusters/{clusterName}/services/{serviceName}/components/{componentName}/?fields=ServiceComponentInfo,host_components/HostRoles/state',
    'mock': ''
  },
  'host_components.hbase_regionserver.active': {
    'real': '/clusters/{clusterName}/host_components?HostRoles/component_name=HBASE_REGIONSERVER&HostRoles/maintenance_state=OFF&HostRoles/desired_admin_state=INSERVICE&HostRoles/host_name.in({hostNames})',
    'mock': ''
  },
  'host.host_component.decommission_status_datanode': {
    'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}?fields=metrics/dfs/namenode',
    'mock': '/data/hosts/HDP2/decommission_state.json'
  },
  'host.region_servers.in_inservice': {
    'real': '/clusters/{clusterName}/host_components?HostRoles/component_name=HBASE_REGIONSERVER&HostRoles/desired_admin_state=INSERVICE&fields=HostRoles/host_name&minimal_response=true',
    'mock': ''
  },
  'host.host_component.decommission_slave': {
    'real': '/clusters/{clusterName}/requests',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          RequestInfo: {
            'context': data.context,
            'command': data.command,
            'parameters': {
              'slave_type': data.slaveType,
              'excluded_hosts': data.hostName
            },
            'operation_level': {
              level: "HOST_COMPONENT",
              cluster_name: data.clusterName,
              host_name: data.hostName,
              service_name: data.serviceName
            }
          },
          "Requests/resource_filters": [{"service_name": data.serviceName, "component_name": data.componentName}]
        })
      }
    }
  },
  'host.host_component.recommission_and_restart': {
    'real': '/clusters/{clusterName}/request_schedules',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify([{
          "RequestSchedule": {
            "batch": [{
              "requests": data.batches
            }, {
              "batch_settings": {
                "batch_separation_in_seconds": data.intervalTimeSeconds,
                "task_failure_tolerance": data.tolerateSize
              }
            }]
          }
        }])
      }
    }
  },

  'host.host_component.refresh_configs': {
    'real': '/clusters/{clusterName}/requests',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          "RequestInfo": {
            "command": "CONFIGURE",
            "context": data.context
          },
          "Requests/resource_filters": data.resource_filters
        })
      }
    }
  },

  'hosts.metrics': {
    'real': '/clusters/{clusterName}/hosts?fields={metricName}',
    'mock': '/data/cluster_metrics/cpu_1hr.json'
  },
  'hosts.metrics.host_component': {
    'real': '/clusters/{clusterName}/services/{serviceName}/components/{componentName}?fields=host_components/{metricName}',
    'mock': '/data/cluster_metrics/cpu_1hr.json'
  },
  'service.metrics.flume.channel_fill_percent': {
    'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_HANDLER?fields=host_components/metrics/flume/flume/CHANNEL/ChannelFillPercentage[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/flume/channelFillPct.json',
    'testInProduction': true
  },
  'service.metrics.flume.channel_size': {
    'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_HANDLER?fields=host_components/metrics/flume/flume/CHANNEL/ChannelSize[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/flume/channelSize.json',
    'testInProduction': true
  },
  'service.metrics.flume.sink_drain_success': {
    'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_HANDLER?fields=host_components/metrics/flume/flume/SINK/EventDrainSuccessCount[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/flume/sinkDrainSuccessCount.json',
    'testInProduction': true
  },
  'service.metrics.flume.sink_connection_failed': {
    'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_HANDLER?fields=host_components/metrics/flume/flume/SINK/ConnectionFailedCount[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/flume/sinkConnectionFailedCount.json',
    'testInProduction': true
  },
  'service.metrics.flume.source_accepted': {
    'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_HANDLER?fields=host_components/metrics/flume/flume/SOURCE/EventAcceptedCount[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/flume/sourceEventAccepted.json',
    'testInProduction': true
  },
  'service.metrics.flume.channel_size_for_all': {
    'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_HANDLER?fields=metrics/flume/flume/CHANNEL/ChannelSize/rate[{fromSeconds},{toSeconds},{stepSeconds}]'
  },
  'service.metrics.flume.gc': {
    'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_HANDLER?fields=host_components/metrics/jvm/gcTimeMillis[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/flume/jvmGcTime.json',
    'testInProduction': true
  },
  'service.metrics.flume.jvm_heap_used': {
    'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_HANDLER?fields=host_components/metrics/jvm/memHeapUsedM[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/flume/jvmMemHeapUsedM.json',
    'testInProduction': true
  },
  'service.metrics.flume.jvm_threads_runnable': {
    'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_HANDLER?fields=host_components/metrics/jvm/threadsRunnable[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/flume/jvmThreadsRunnable.json',
    'testInProduction': true
  },
  'service.metrics.flume.cpu_user': {
    'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_HANDLER?fields=host_components/metrics/cpu/cpu_user[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '',
    'testInProduction': true
  },
  'service.metrics.flume.incoming_event_put_successCount': {
    'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_HANDLER?fields=metrics/flume/flume/CHANNEL/EventPutSuccessCount/rate[{fromSeconds},{toSeconds},{stepSeconds}]'
  },
  'service.metrics.flume.outgoing_event_take_success_count': {
    'real': '/clusters/{clusterName}/services/FLUME/components/FLUME_HANDLER?fields=metrics/flume/flume/CHANNEL/EventTakeSuccessCount/rate[{fromSeconds},{toSeconds},{stepSeconds}]'
  },
  'service.metrics.hbase.cluster_requests': {
    'real': '/clusters/{clusterName}/services/HBASE/components/HBASE_MASTER?fields=metrics/hbase/master/cluster_requests[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hbase/cluster_requests.json',
    'testInProduction': true
  },
  'service.metrics.hbase.hlog_split_size': {
    'real': '/clusters/{clusterName}/services/HBASE/components/HBASE_MASTER?fields=metrics/hbase/master/splitSize_avg_time[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hbase/hlog_split_size.json',
    'testInProduction': true
  },
  'service.metrics.hbase.hlog_split_time': {
    'real': '/clusters/{clusterName}/services/HBASE/components/HBASE_MASTER?fields=metrics/hbase/master/splitTime_avg_time[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hbase/hlog_split_time.json',
    'testInProduction': true
  },
  'service.metrics.hbase.regionserver_queuesize': {
    'real': '/clusters/{clusterName}/services/HBASE/components/HBASE_REGIONSERVER?fields=metrics/hbase/regionserver/flushQueueSize[{fromSeconds},{toSeconds},{stepSeconds}],metrics/hbase/regionserver/compactionQueueSize[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hbase/regionserver_queuesize.json',
    'testInProduction': true
  },
  'service.metrics.hbase.regionserver_regions': {
    'real': '/clusters/{clusterName}/services/HBASE/components/HBASE_REGIONSERVER?fields=metrics/hbase/regionserver/regions[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hbase/regionserver_regions.json',
    'testInProduction': true
  },
  'service.metrics.hbase.regionserver_rw_requests': {
    'real': '/clusters/{clusterName}/services/HBASE/components/HBASE_REGIONSERVER?fields=metrics/hbase/regionserver/readRequestsCount[{fromSeconds},{toSeconds},{stepSeconds}],metrics/hbase/regionserver/writeRequestsCount[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hbase/regionserver_rw_requests.json',
    'testInProduction': true
  },
  'service.metrics.hdfs.block_status': {
    'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/dfs/FSNamesystem/PendingReplicationBlocks[{fromSeconds},{toSeconds},{stepSeconds}],metrics/dfs/FSNamesystem/UnderReplicatedBlocks[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hdfs/block_status.json',
    'testInProduction': true
  },
  'service.metrics.hdfs.file_operations': {
    'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/dfs/namenode/FileInfoOps[{fromSeconds},{toSeconds},{stepSeconds}],metrics/dfs/namenode/CreateFileOps[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hdfs/file_operations.json',
    'testInProduction': true
  },
  'service.metrics.hdfs.gc': {
    'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/jvm/gcTimeMillis[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hdfs/gc.json',
    'testInProduction': true
  },
  'service.metrics.hdfs.io': {
    'real': '/clusters/{clusterName}/services/HDFS/components/DATANODE?fields=metrics/dfs/datanode/bytes_written[{fromSeconds},{toSeconds},{stepSeconds}],metrics/dfs/datanode/bytes_read[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hdfs/io.json',
    'testInProduction': true
  },
  'service.metrics.hdfs.jvm_heap': {
    'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/jvm/memNonHeapUsedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memNonHeapCommittedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memHeapUsedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memHeapCommittedM[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hdfs/jvm_heap.json',
    'testInProduction': true
  },
  'service.metrics.hdfs.jvm_threads': {
    'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/jvm/threadsRunnable[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsBlocked[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsWaiting[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsTimedWaiting[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hdfs/jvm_threads.json',
    'testInProduction': true
  },
  'service.metrics.hdfs.rpc': {
    'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/rpc/RpcQueueTime_avg_time[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hdfs/rpc.json',
    'testInProduction': true
  },
  'service.metrics.hdfs.space_utilization': {
    'real': '/clusters/{clusterName}/hosts/{nameNodeName}/host_components/NAMENODE?fields=metrics/dfs/FSNamesystem/CapacityRemaining[{fromSeconds},{toSeconds},{stepSeconds}],metrics/dfs/FSNamesystem/CapacityUsed[{fromSeconds},{toSeconds},{stepSeconds}],metrics/dfs/FSNamesystem/CapacityTotal[{fromSeconds},{toSeconds},{stepSeconds}],metrics/dfs/FSNamesystem/CapacityNonDFSUsed[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/hdfs/space_utilization.json',
    'testInProduction': true
  },
  'service.metrics.yarn.gc': {
    'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/jvm/gcTimeMillis[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/yarn/gc.json',
    'testInProduction': true
  },
  'service.metrics.yarn.jobs_threads': {
    'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/jvm/threadsRunnable[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsBlocked[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsWaiting[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/threadsTimedWaiting[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/yarn/jvm_threads.json',
    'testInProduction': true
  },
  'service.metrics.yarn.rpc': {
    'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/rpc/RpcQueueTime_avg_time[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/yarn/rpc.json',
    'testInProduction': true
  },
  'service.metrics.yarn.jobs_heap': {
    'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/jvm/memNonHeapUsedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memNonHeapCommittedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memHeapUsedM[{fromSeconds},{toSeconds},{stepSeconds}],metrics/jvm/memHeapCommittedM[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/services/metrics/yarn/jvm_heap.json',
    'testInProduction': true
  },
  'service.metrics.yarn.queue.allocated': {
    'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/yarn/Queue/root/AvailableMB[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/PendingMB[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/AllocatedMB[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '',
    'testInProduction': true
  },
  'service.metrics.yarn.queue.allocated.container': {
    'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/yarn/Queue/root/AllocatedContainers[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/ReservedContainers[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/PendingContainers[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '',
    'testInProduction': true
  },
  'service.metrics.yarn.node.manager.statuses': {
    'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/yarn/ClusterMetrics/NumActiveNMs[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/ClusterMetrics/NumDecommissionedNMs[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/ClusterMetrics/NumLostNMs[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/ClusterMetrics/NumRebootedNMs[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/ClusterMetrics/NumUnhealthyNMs[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '',
    'testInProduction': true
  },
  'service.metrics.yarn.queue.memory.resource': {
    'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=',
    'mock': '',
    'format': function (data, opt) {
      var field1 = 'metrics/yarn/Queue/{queueName}/AllocatedMB[{fromSeconds},{toSeconds},{stepSeconds}]';
      var field2 = 'metrics/yarn/Queue/{queueName}/AvailableMB[{fromSeconds},{toSeconds},{stepSeconds}]';
      if (opt.url != null && data.queueNames != null && data.queueNames.length > 0) {
        data.queueNames.forEach(function (q) {
          data.queueName = q;
          opt.url += (formatUrl(field1, data) + ",");
          opt.url += (formatUrl(field2, data) + ",");
        });
      } else {
        opt.url += (formatUrl(field1, data) + ",");
        opt.url += (formatUrl(field2, data) + ",");
      }
    },
    'testInProduction': true
  },
  'service.metrics.yarn.queue.apps.states.current': {
    'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/yarn/Queue/root/AppsPending[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/AppsRunning[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '',
    'testInProduction': true
  },
  'service.metrics.yarn.queue.apps.states.finished': {
    'real': '/clusters/{clusterName}/hosts/{resourceManager}/host_components/RESOURCEMANAGER?fields=metrics/yarn/Queue/root/AppsKilled[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/AppsFailed[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/AppsSubmitted[{fromSeconds},{toSeconds},{stepSeconds}],metrics/yarn/Queue/root/AppsCompleted[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '',
    'testInProduction': true
  },
  'service.metrics.kafka.broker.topic': {
    'real': '/clusters/{clusterName}/services/KAFKA/components/KAFKA_BROKER?fields=metrics/kafka/server/BrokerTopicMetrics/AllTopicsBytesInPerSec/1MinuteRate[{fromSeconds},{toSeconds},{stepSeconds}],metrics/kafka/server/BrokerTopicMetrics/AllTopicsBytesOutPerSec/1MinuteRate[{fromSeconds},{toSeconds},{stepSeconds}],metrics/kafka/server/BrokerTopicMetrics/AllTopicsMessagesInPerSec/1MinuteRate[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': ''
  },
  'service.metrics.kafka.controller.KafkaController': {
    'real': '/clusters/{clusterName}/services/KAFKA/components/KAFKA_BROKER?fields=metrics/kafka/controller/KafkaController/ActiveControllerCount[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': ''
  },
  'service.metrics.kafka.controller.ControllerStats': {
    'real': '/clusters/{clusterName}/services/KAFKA/components/KAFKA_BROKER?fields=metrics/kafka/controller/ControllerStats/LeaderElectionRateAndTimeMs/1MinuteRate[{fromSeconds},{toSeconds},{stepSeconds}],metrics/kafka/controller/ControllerStats/UncleanLeaderElectionsPerSec/1MinuteRate[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': ''
  },
  'service.metrics.kafka.log.LogFlushStats': {
    'real': '/clusters/{clusterName}/services/KAFKA/components/KAFKA_BROKER?fields=metrics/kafka/log/LogFlushStats/LogFlushRateAndTimeMs/1MinuteRate[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': ''
  },
  'service.metrics.kafka.server.ReplicaManager': {
    'real': '/clusters/{clusterName}/services/KAFKA/components/KAFKA_BROKER?fields=metrics/kafka/server/ReplicaManager/PartitionCount[{fromSeconds},{toSeconds},{stepSeconds}],metrics/kafka/server/ReplicaManager/UnderReplicatedPartitions[{fromSeconds},{toSeconds},{stepSeconds}],metrics/kafka/server/BrokerTopicMetrics/ReplicaManager/LeaderCount[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': ''
  },
  'service.metrics.kafka.server.ReplicaFetcherManager': {
    'real': '/clusters/{clusterName}/services/KAFKA/components/KAFKA_BROKER?fields=metrics/kafka/server/ReplicaFetcherManager/Replica-MaxLag[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': ''
  },
  'service.metrics.storm.nimbus': {
    'real': '/clusters/{clusterName}/services/STORM/components/NIMBUS?fields={metricsTemplate}',
    'mock': ''
  },
  'dashboard.cluster_metrics.cpu': {
    'real': '/clusters/{clusterName}/?fields=metrics/cpu[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/cluster_metrics/cpu_1hr.json',
    'testInProduction': true
  },
  'dashboard.cluster_metrics.load': {
    'real': '/clusters/{clusterName}/?fields=metrics/load[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/cluster_metrics/load_1hr.json',
    'testInProduction': true
  },
  'dashboard.cluster_metrics.memory': {
    'real': '/clusters/{clusterName}/?fields=metrics/memory[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/cluster_metrics/memory_1hr.json',
    'testInProduction': true
  },
  'dashboard.cluster_metrics.network': {
    'real': '/clusters/{clusterName}/?fields=metrics/network[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/cluster_metrics/network_1hr.json',
    'testInProduction': true
  },
  'host.metrics.cpu': {
    'real': '/clusters/{clusterName}/hosts/{hostName}?fields=metrics/cpu/cpu_user[{fromSeconds},{toSeconds},{stepSeconds}],metrics/cpu/cpu_wio[{fromSeconds},{toSeconds},{stepSeconds}],metrics/cpu/cpu_nice[{fromSeconds},{toSeconds},{stepSeconds}],metrics/cpu/cpu_aidle[{fromSeconds},{toSeconds},{stepSeconds}],metrics/cpu/cpu_system[{fromSeconds},{toSeconds},{stepSeconds}],metrics/cpu/cpu_idle[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/hosts/metrics/cpu.json',
    'testInProduction': true
  },
  'host.metrics.disk': {
    'real': '/clusters/{clusterName}/hosts/{hostName}?fields=metrics/disk/disk_total[{fromSeconds},{toSeconds},{stepSeconds}],metrics/disk/disk_free[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/hosts/metrics/disk.json',
    'testInProduction': true
  },
  'host.metrics.load': {
    'real': '/clusters/{clusterName}/hosts/{hostName}?fields=metrics/load/load_fifteen[{fromSeconds},{toSeconds},{stepSeconds}],metrics/load/load_one[{fromSeconds},{toSeconds},{stepSeconds}],metrics/load/load_five[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/hosts/metrics/load.json',
    'testInProduction': true
  },
  'host.metrics.memory': {
    'real': '/clusters/{clusterName}/hosts/{hostName}?fields=metrics/memory/swap_free[{fromSeconds},{toSeconds},{stepSeconds}],metrics/memory/mem_shared[{fromSeconds},{toSeconds},{stepSeconds}],metrics/memory/mem_free[{fromSeconds},{toSeconds},{stepSeconds}],metrics/memory/mem_cached[{fromSeconds},{toSeconds},{stepSeconds}],metrics/memory/mem_buffers[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/hosts/metrics/memory.json',
    'testInProduction': true
  },
  'host.metrics.network': {
    'real': '/clusters/{clusterName}/hosts/{hostName}?fields=metrics/network/bytes_in[{fromSeconds},{toSeconds},{stepSeconds}],metrics/network/bytes_out[{fromSeconds},{toSeconds},{stepSeconds}],metrics/network/pkts_in[{fromSeconds},{toSeconds},{stepSeconds}],metrics/network/pkts_out[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/hosts/metrics/network.json',
    'testInProduction': true
  },
  'host.metrics.processes': {
    'real': '/clusters/{clusterName}/hosts/{hostName}?fields=metrics/process/proc_total[{fromSeconds},{toSeconds},{stepSeconds}],metrics/process/proc_run[{fromSeconds},{toSeconds},{stepSeconds}]',
    'mock': '/data/hosts/metrics/processes.json',
    'testInProduction': true
  },
  'admin.security_status': {
    'real': '/clusters/{clusterName}?fields=Clusters/security_type',
    'mock': '',
    'format': function () {
      return {
        timeout: 10000
      };
    }
  },
  'settings.get.user_pref': {
    'real': '/persist/{key}',
    'mock': '/data/user_settings/{key}.json'
  },
  'settings.post.user_pref': {
    'real': '/persist',
    'mock': '',
    'type': 'POST',
    'format': function (data) {
      return {
        data: JSON.stringify(data.keyValuePair)
      }
    }
  },
  'cluster.load_cluster_name': {
    'real': '/clusters',
    'mock': '/data/clusters/info.json'
  },
  'cluster.update_upgrade_version': {
    'real': '/stacks/{stackName}/versions?fields=services/StackServices,Versions',
    'mock': '/data/wizard/stack/stacks.json',
    'format': function (data) {
      return {
        data: data.data
      };
    }
  },
  'cluster.load_repositories': {
    'real': '/stacks/{stackName}/versions/{stackVersion}/operating_systems?fields=repositories/*',
    'mock': '/data/stacks/HDP-2.1/operating_systems.json',
    'format': function (data) {
      return {
        data: data.data
      };
    }
  },
  'cluster.load_detailed_repo_version': {
    'real': '/clusters/{clusterName}/stack_versions?ClusterStackVersions/state=CURRENT&fields=repository_versions/RepositoryVersions/repository_version&minimal_response=true',
    'mock': '/data/stack_versions/stack_version_all.json'
  },
  'cluster.save_provisioning_state': {
    'real': '/clusters/{clusterName}',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify({
          "Clusters": {
            "provisioning_state": data.state
          }
        })
      };
    }
  },
  'admin.high_availability.polling': {
    'real': '/clusters/{clusterName}/requests/{requestId}?fields=tasks/*,Requests/*',
    'mock': '/data/background_operations/host_upgrade_tasks.json'
  },
  'admin.high_availability.getNnCheckPointStatus': {
    'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/NAMENODE',
    'mock': ''
  },
  'admin.high_availability.getJnCheckPointStatus': {
    'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/JOURNALNODE?fields=metrics',
    'mock': ''
  },
  'admin.high_availability.getHostComponent': {
    'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/{componentName}',
    'mock': ''
  },
  'admin.high_availability.create_component': {
    'real': '/clusters/{clusterName}/hosts?Hosts/host_name={hostName}',
    'mock': '',
    'type': 'POST',
    'format': function (data) {
      return {
        data: JSON.stringify({
          "host_components": [
            {
              "HostRoles": {
                "component_name": data.componentName
              }
            }
          ]
        })
      }
    }
  },
  'admin.high_availability.create_journalnode': {
    'real': '/clusters/{clusterName}/services?ServiceInfo/service_name=HDFS',
    'mock': '',
    'type': 'POST',
    'format': function () {
      return {
        data: JSON.stringify({
          "components": [
            {
              "ServiceComponentInfo": {
                "component_name": "JOURNALNODE"
              }
            }
          ]
        })
      }
    }
  },
  'common.create_component': {
    'real': '/clusters/{clusterName}/services?ServiceInfo/service_name={serviceName}',
    'mock': '',
    'type': 'POST',
    'format': function (data) {
      return {
        data: JSON.stringify({
          "components": [
            {
              "ServiceComponentInfo": {
                "component_name": data.componentName
              }
            }
          ]
        })
      }
    }
  },
  'admin.high_availability.create_zkfc': {
    'real': '/clusters/{clusterName}/services?ServiceInfo/service_name=HDFS',
    'mock': '',
    'type': 'POST',
    'format': function () {
      return {
        data: JSON.stringify({
          "components": [
            {
              "ServiceComponentInfo": {
                "component_name": "ZKFC"
              }
            }
          ]
        })
      }
    }
  },
  'admin.high_availability.load_configs': {
    'real': '/clusters/{clusterName}/configurations?(type=core-site&tag={coreSiteTag})|(type=hdfs-site&tag={hdfsSiteTag})',
    'mock': ''
  },
  'admin.high_availability.save_configs': {
    'real': '/clusters/{clusterName}',
    'mock': '',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify({
          Clusters: {
            desired_config: {
              "type": data.siteName,
              "tag": 'version' + (new Date).getTime(),
              "properties": data.properties
            }
          }
        })
      }
    }
  },
  'admin.high_availability.load_hbase_configs': {
    'real': '/clusters/{clusterName}/configurations?type=hbase-site&tag={hbaseSiteTag}',
    'mock': ''
  },
  'admin.security.cluster_configs': {
    'real': '/clusters/{clusterName}',
    'mock': '',
    'format': function () {
      return {
        timeout: 10000
      };
    }
  },
  'admin.get.all_configurations': {
    'real': '/clusters/{clusterName}/configurations?{urlParams}',
    'mock': '',
    'format': function () {
      return {
        timeout: 10000
      };
    }
  },
  'admin.security.add.cluster_configs': {
    'real': '/clusters/{clusterName}' + '?fields=Clusters/desired_configs',
    'mock': '',
    'format': function () {
      return {
        timeout: 10000
      };
    }
  },

  'kerberos.session.state': {
    'real': '/clusters/{clusterName}/services/KERBEROS?fields=Services/attributes/kdc_validation_result,Services/attributes/kdc_validation_failure_details',
    'mock': ''
  },

  'admin.kerberize.cluster': {
    'type': 'PUT',
    'real': '/clusters/{clusterName}',
    'mock': '/data/wizard/kerberos/kerberize_cluster.json',
    'format' : function (data) {
      return {
        data: JSON.stringify(data.data)
      }
    }
  },
  'admin.unkerberize.cluster': {
    'type': 'PUT',
    'real': '/clusters/{clusterName}',
    'format': function (data) {
      return {
        data: JSON.stringify({
          Clusters: {
            security_type: "NONE"
          }
        })
      }
    }
  },
  'get.cluster.artifact': {
    'real': '/clusters/{clusterName}/artifacts/{artifactName}?fields=artifact_data',
    'mock': '/data/wizard/kerberos/stack_descriptors.json'
  },
  'admin.kerberize.stack_descriptor': {
    'real': '/stacks/{stackName}/versions/{stackVersionNumber}/artifacts/kerberos_descriptor?fields=artifact_data',
    'mock': '/data/wizard/kerberos/stack_descriptors.json'
  },
  'admin.kerberize.cluster_descriptor': {
    'real': '/clusters/{clusterName}/artifacts/kerberos_descriptor?fields=artifact_data',
    'mock': '/data/wizard/kerberos/stack_descriptors.json'
  },
  'admin.kerberos.cluster.artifact.create': {
    'type': 'POST',
    'real': '/clusters/{clusterName}/artifacts/{artifactName}',
    'format' : function (data) {
      return {
        data: JSON.stringify(data.data)
      }
    }
  },
  'admin.kerberos.cluster.artifact.update': {
    'type': 'PUT',
    'real': '/clusters/{clusterName}/artifacts/{artifactName}',
    'format' : function (data) {
      return {
        data: JSON.stringify(data.data)
      }
    }
  },
  'admin.poll.kerberize.cluster.request': {
    'real': '/clusters/{clusterName}/requests/{requestId}?fields=stages/Stage/context,stages/Stage/status,stages/Stage/progress_percent,stages/tasks/*,Requests/*',
    'mock': '/data/wizard/kerberos/kerberize_cluster.json'
  },
  'admin.stack_upgrade.run_upgrade': {
    'real': '/clusters/{clusterName}',
    'mock': '',
    'format': function (data) {
      return {
        type: 'PUT',
        data: data.data
      };
    }
  },
  'admin.user.create': {
    'real': '/users/{user}',
    'mock': '/data/users/users.json',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify(data.data)
      }
    }
  },

  'admin.user.edit': {
    'real': '/users/{user}',
    'mock': '/data/users/users.json',
    'format': function (data) {
      return {
        type: 'PUT',
        data: data.data
      }
    }
  },

  'admin.stack_upgrade.do_poll': {
    'real': '/clusters/{cluster}/requests/{requestId}?fields=tasks/*',
    'mock': '/data/wizard/{mock}'
  },
  'admin.upgrade.data': {
    'real': '/clusters/{clusterName}/upgrades/{id}?upgrade_groups/UpgradeGroup/status!=PENDING&fields=' +
    'Upgrade/progress_percent,Upgrade/request_context,Upgrade/request_status,Upgrade/direction,' +
    'upgrade_groups/UpgradeGroup,' +
    'upgrade_groups/upgrade_items/UpgradeItem/status,' +
    'upgrade_groups/upgrade_items/UpgradeItem/context,' +
    'upgrade_groups/upgrade_items/UpgradeItem/group_id,' +
    'upgrade_groups/upgrade_items/UpgradeItem/progress_percent,' +
    'upgrade_groups/upgrade_items/UpgradeItem/request_id,' +
    'upgrade_groups/upgrade_items/UpgradeItem/skippable,' +
    'upgrade_groups/upgrade_items/UpgradeItem/stage_id,' +
    'upgrade_groups/upgrade_items/UpgradeItem/status,' +
    'upgrade_groups/upgrade_items/UpgradeItem/text&' +
    'minimal_response=true',
    'mock': '/data/stack_versions/upgrade.json'
  },
  'admin.upgrade.state': {
    'real': '/clusters/{clusterName}/upgrades/{id}?fields=Upgrade',
    'mock': '/data/stack_versions/upgrade.json'
  },
  'admin.upgrade.upgrade_item': {
    'real': '/clusters/{clusterName}/upgrades/{upgradeId}/upgrade_groups/{groupId}/upgrade_items/{stageId}?fields=' +
    'UpgradeItem/group_id,' +
    'UpgradeItem/stage_id,' +
    'tasks/Tasks/*&' +
    'minimal_response=true',
    'mock': '/data/stack_versions/upgrade_item.json'
  },
  'admin.upgrade.start': {
    'real': '/clusters/{clusterName}/upgrades',
    'mock': '/data/stack_versions/start_upgrade.json',
    'type': 'POST',
    'format': function (data) {
      return {
        data: JSON.stringify({
          "Upgrade": {
            "repository_version": data.value
          }
        })
      }
    }
  },
  'admin.downgrade.start': {
    'real': '/clusters/{clusterName}/upgrades',
    'mock': '/data/stack_versions/start_upgrade.json',
    'type': 'POST',
    'format': function (data) {
      return {
        data: JSON.stringify({
          "Upgrade": {
            "repository_version": data.value,
            "force_downgrade": true
          }
        })
      }
    }
  },
  'admin.upgrade.abort': {
    'real': '/clusters/{clusterName}/upgrades/{upgradeId}',
    'mock': '',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify({
          "Upgrade": {
            "request_status": "ABORTED"
          }
        })
      }
    }
  },
  'admin.upgrade.upgradeItem.setState': {
    'real': '/clusters/{clusterName}/upgrades/{upgradeId}/upgrade_groups/{groupId}/upgrade_items/{itemId}',
    'mock': '',
    type: 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify({
          "UpgradeItem" : {
            "status" : data.status
          }
        })
      };
    }
  },
  'admin.stack_versions.all': {
    'real': '/clusters/{clusterName}/stack_versions?fields=ClusterStackVersions/*,repository_versions/RepositoryVersions/*&minimal_response=true',
    'mock': '/data/stack_versions/stack_version_all.json'
  },
  'admin.stack_version.install.repo_version': {
    'real': '/clusters/{clusterName}/stack_versions',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          ClusterStackVersions: data.ClusterStackVersions
        })
      }
    },
    'mock': ''
  },

  'admin.stack_versions.edit.repo': {
    'real': '/stacks/{stackName}/versions/{stackVersion}/repository_versions/{repoVersionId}',
    'mock': '',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify(data.repoVersion)
      }
    }
  },
  'admin.stack_versions.validate.repo': {
    'real': '/stacks/{stackName}/versions/{stackVersion}/operating_systems/{osType}/repositories/{repoId}?validate_only=true',
    'mock': '',
    'type': 'POST',
    'format': function (data) {
      return {
        data: JSON.stringify({
          "Repositories": {
            "base_url": data.baseUrl
          }
        })
      }
    }
  },

  'admin.rolling_upgrade.pre_upgrade_check': {
    'real': '/clusters/{clusterName}/rolling_upgrades_check?fields=*&UpgradeChecks/repository_version={value}',
    'mock': '/data/stack_versions/pre_upgrade_check.json'
  },

  'admin.kerberos_security.checks': {
    //TODO when api will be known
    'real': '',
    'mock': '/data/stack_versions/pre_upgrade_check.json'
  },

  'admin.kerberos_security.test_connection': {
    'real': '/kdc_check/{kdcHostname}',
    'mock': '',
    'format': function () {
      return {
        dataType: 'text'
      };
    }
  },

  'admin.kerberos_security.regenerate_keytabs': {
    'real': '/clusters/{clusterName}?regenerate_keytabs={type}',
    'mock': '',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify({
          "Clusters" : {
            "security_type" : "KERBEROS"
          }
        })
      }
    }
  },

  'wizard.advanced_repositories.valid_url': {
    'real': '/stacks/{stackName}/versions/{stackVersion}/operating_systems/{osType}/repositories/{repoId}',
    'mock': '',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify(data.data)
      }
    }
  },
  'wizard.service_components': {
    'real': '{stackUrl}/services?fields=StackServices/*,components/*,components/dependencies/Dependencies/scope',
    'mock': '/data/stacks/HDP-2.1/service_components.json',
    'format': function (data) {
      return {
        timeout: 10000
      };
    }
  },
  'wizard.step9.installer.get_host_status': {
    'real': '/clusters/{cluster}/hosts?fields=Hosts/host_state,host_components/HostRoles/state',
    'mock': '/data/wizard/deploy/5_hosts/get_host_state.json'
  },
  'wizard.step9.load_log': {
    'real': '/clusters/{cluster}/requests/{requestId}?fields=tasks/Tasks/command,tasks/Tasks/exit_code,tasks/Tasks/start_time,tasks/Tasks/end_time,tasks/Tasks/host_name,tasks/Tasks/id,tasks/Tasks/role,tasks/Tasks/status&minimal_response=true',
    'mock': '/data/wizard/deploy/5_hosts/poll_{numPolls}.json',
    'format': function () {
      return {
        dataType: 'text'
      };
    }
  },

  'wizard.step8.existing_cluster_names': {
    'real': '/clusters',
    'mock': ''
  },

  'wizard.step8.create_cluster': {
    'real': '/clusters/{cluster}',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        dataType: 'text',
        data: data.data
      }
    }
  },

  'wizard.step8.create_selected_services': {
    'type': 'POST',
    'real': '/clusters/{cluster}/services',
    'mock': '/data/stacks/HDP-2.1/recommendations.json',
    'format': function (data) {
      return {
        dataType: 'text',
        data: data.data
      }
    }
  },

  'wizard.step8.create_components': {
    'real': '/clusters/{cluster}/services?ServiceInfo/service_name={serviceName}',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        dataType: 'text',
        data: data.data
      }
    }
  },

  'wizard.step8.register_host_to_cluster': {
    'real': '/clusters/{cluster}/hosts',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        dataType: 'text',
        data: data.data
      }
    }
  },

  'wizard.step8.register_host_to_component': {
    'real': '/clusters/{cluster}/hosts',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        dataType: 'text',
        data: data.data
      }
    }
  },

  'wizard.step8.apply_configuration_groups': {
    'real': '/clusters/{cluster}/config_groups',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        dataType: 'text',
        data: data.data
      }
    }
  },

  'wizard.step8.set_local_repos': {
    'real': '{stackVersionURL}/operating_systems/{osType}/repositories/{repoId}',
    'mock': '',
    'format': function (data) {
      return {
        type: 'PUT',
        dataType: 'text',
        data: data.data
      }
    }
  },
  'wizard.step3.jdk_check': {
    'real': '/requests',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          "RequestInfo": {
            "context": "Check hosts",
            "action": "check_host",
            "parameters": {
              "threshold": "60",
              "java_home": data.java_home,
              "jdk_location": data.jdk_location,
              "check_execute_list": "java_home_check"
            }
          },
          "Requests/resource_filters": [{
            "hosts": data.host_names
          }]
        })
      }
    }
  },
  'wizard.step3.jdk_check.get_results': {
    'real': '/requests/{requestIndex}?fields=*,tasks/Tasks/host_name,tasks/Tasks/status,tasks/Tasks/structured_out',
    'mock': '/data/requests/host_check/jdk_check_results.json'
  },
  'wizard.step3.host_info': {
    'real': '/hosts?fields=Hosts/total_mem,Hosts/cpu_count,Hosts/disk_info,Hosts/last_agent_env,Hosts/host_name,Hosts/os_type,Hosts/os_arch,Hosts/ip',
    'mock': '/data/wizard/bootstrap/two_hosts_information.json',
    'format': function () {
      return {
        contentType: 'application/json'
      };
    }
  },


  'wizard.loadrecommendations': {
    'real': '{stackVersionUrl}/recommendations',
    'mock': '/data/stacks/HDP-2.1/recommendations.json',
    'type': 'POST',
    'format': function (data) {
      var q = {
        hosts: data.hosts,
        services: data.services,
        recommend: data.recommend
      };

      if (data.recommendations) {
        q.recommendations = data.recommendations;
      }

      return {
        data: JSON.stringify(q)
      }
    }
  },


  // TODO: merge with wizard.loadrecommendations query
  'wizard.step7.loadrecommendations.configs': {
    'real': '{stackVersionUrl}/recommendations',
    'mock': '/data/stacks/HDP-2.1/recommendations_configs.json',
    'type': 'POST',
    'format': function (data) {
      return {
        data: JSON.stringify({
          hosts: data.hosts,
          services: data.services,
          recommendations: data.recommendations,
          recommend: "configurations"
        })
      }
    }
  },

  'config.validations': {
    'real': '{stackVersionUrl}/validations',
    'mock': '/data/stacks/HDP-2.1/validations.json',
    'type': 'POST',
    'format': function (data) {
      return {
        data: JSON.stringify({
          hosts: data.hosts,
          services: data.services,
          validate: data.validate,
          recommendations: data.recommendations
        })
      }
    }
  },


  'preinstalled.checks': {
    'real': '/requests',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          "RequestInfo": data.RequestInfo,
          "Requests/resource_filters": [data.resource_filters]
        })
      }
    }
  },

  'preinstalled.checks.tasks': {
    'real': '/requests/{requestId}?fields=tasks/Tasks,Requests/inputs,Requests/request_status',
    'mock': '/data/requests/host_check/1.json'
  },

  'wizard.step3.rerun_checks': {
    'real': '/hosts?fields=Hosts/last_agent_env',
    'mock': '/data/wizard/bootstrap/two_hosts_information.json',
    'format': function () {
      return {
        contentType: 'application/json'
      };
    }
  },
  'wizard.step3.bootstrap': {
    'real': '/bootstrap/{bootRequestId}',
    'mock': '/data/wizard/bootstrap/poll_{numPolls}.json'
  },
  'wizard.step3.is_hosts_registered': {
    'real': '/hosts?fields=Hosts/host_status',
    'mock': '/data/wizard/bootstrap/single_host_registration.json'
  },
  'wizard.stacks': {
    'real': '/stacks',
    'mock': '/data/wizard/stack/stacks2.json'
  },
  'wizard.stacks_versions': {
    'real': '/stacks/{stackName}/versions?fields=Versions,operating_systems/repositories/Repositories',
    'mock': '/data/wizard/stack/{stackName}_versions.json'
  },
  'wizard.launch_bootstrap': {
    'real': '/bootstrap',
    'mock': '/data/wizard/bootstrap/bootstrap.json',
    'type': 'POST',
    'format': function (data) {
      return {
        contentType: 'application/json',
        data: data.bootStrapData,
        popup: data.popup
      }
    }
  },
  'router.login': {
    'real': '/users/{loginName}?fields=*,privileges/PrivilegeInfo/cluster_name,privileges/PrivilegeInfo/permission_name',
    'mock': '/data/users/user_{usr}.json',
    'format': function (data) {
      var statusCode = jQuery.extend({}, require('data/statusCodes'));
      statusCode['403'] = function () {
        console.log("Error code 403: Forbidden.");
      };
      return {
        statusCode: statusCode
      };
    }
  },
  'users.all': {
    real: '/users/?fields=*',
    mock: '/data/users/users.json'
  },
  'users.privileges': {
    real: '/privileges?fields=*',
    mock: '/data/users/privileges.json'
  },
  'router.user.privileges': {
    real: '/privileges?PrivilegeInfo/principal_name={userName}&fields=*',
    mock: '/data/users/privileges_{userName}.json'
  },
  'router.login.clusters': {
    'real': '/clusters?fields=Clusters/provisioning_state',
    'mock': '/data/clusters/info.json'
  },
  'router.logoff': {
    'real': '/logout',
    'mock': ''
  },
  'ambari.service.load_jdk_name': {
    'real': '/services/AMBARI/components/AMBARI_SERVER?fields=RootServiceComponents/properties/jdk.name,RootServiceComponents/properties/java.home,RootServiceComponents/properties/jdk_location',
    'mock': '/data/requests/host_check/jdk_name.json'
  },
  'ambari.service.load_server_version': {
    'real': '/services/AMBARI/components/AMBARI_SERVER?fields=RootServiceComponents/component_version,RootServiceComponents/properties/server.os_family&minimal_response=true',
    'mock': '/data/ambari_components/component_version.json'
  },
  'ambari.service': {
    'real': '/services/AMBARI/components/AMBARI_SERVER',
    'mock': '/data/services/ambari_server.json'
  },
  'ambari.service.load_server_clock': {
    'real': '/services/AMBARI/components/AMBARI_SERVER?fields=RootServiceComponents/server_clock',
    'mock': ''
  },

  'config_groups.create': {
    'real': '/clusters/{clusterName}/config_groups',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify([{
          "ConfigGroup": {
            "group_name": data.group_name,
            "tag": data.service_id,
            "description": data.description,
            "desired_configs": data.desired_configs,
            "hosts": data.hosts
          }
        }])
      }
    }
  },
  'config_groups.update': {
    'real': '/clusters/{clusterName}/config_groups/{id}',
    'mock': '',
    'format': function (data) {
      return {
        type: 'PUT',
        data: JSON.stringify(data.data)
      }
    }
  },
  'rolling_restart.post': {
    'real': '/clusters/{clusterName}/request_schedules',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify([{
          "RequestSchedule": {
            "batch": [{
              "requests": data.batches
            }, {
              "batch_settings": {
                "batch_separation_in_seconds": data.intervalTimeSeconds,
                "task_failure_tolerance": data.tolerateSize
              }
            }]
          }
        }])
      }
    }
  },
  'request_schedule.get': {
    'real': '/clusters/{clusterName}/request_schedules/{request_schedule_id}',
    'mock': ''
  },
  'restart.hostComponents': {
    'real': '/clusters/{clusterName}/requests',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          "RequestInfo": {
            "command": "RESTART",
            "context": data.context,
            "operation_level": data.operation_level
          },
          "Requests/resource_filters": data.resource_filters
        })
      }
    }
  },

  'bulk_request.decommission': {
    'real': '/clusters/{clusterName}/requests',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          'RequestInfo': {
            'context': data.context,
            'command': 'DECOMMISSION',
            'parameters': data.parameters,
            'operation_level': {
              'level': "CLUSTER",
              'cluster_name': data.clusterName
            }
          },
          "Requests/resource_filters": [{"service_name": data.serviceName, "component_name": data.componentName}]
        })
      }
    }
  },

  'bulk_request.hosts.passive_state': {
    'real': '/clusters/{clusterName}/hosts',
    'mock': '',
    'format': function (data) {
      return {
        type: 'PUT',
        data: JSON.stringify({
          RequestInfo: {
            context: data.requestInfo,
            query: 'Hosts/host_name.in(' + data.hostNames + ')'
          },
          Body: {
            Hosts: {
              maintenance_state: data.passive_state
            }
          }
        })
      }
    }
  },

  'bulk_request.hosts.all_components.passive_state': {
    'real': '/clusters/{clusterName}/host_components',
    'mock': '',
    'format': function (data) {
      return {
        type: 'PUT',
        data: JSON.stringify({
          RequestInfo: {
            context: data.requestInfo,
            query: data.query
          },
          Body: {
            HostRoles: {
              maintenance_state: data.passive_state
            }
          }
        })
      }
    }
  },
  'views.info': {
    'real': '/views',
    'mock': '/data/views/views.json'
  },
  /**
   * Get all instances of all views across versions
   */
  'views.instances': {
    'real': '/views?fields=versions/instances/ViewInstanceInfo,versions/ViewVersionInfo/label&versions/ViewVersionInfo/system=false',
    'mock': '/data/views/instances.json'
  },
  'host.host_component.flume.metrics': {
    'real': '/clusters/{clusterName}/hosts/{hostName}/host_components/FLUME_HANDLER?fields=metrics/flume/flume/{flumeComponent}/*',
    'mock': ''
  },
  'host.host_component.flume.metrics.timeseries': {
    'real': '',
    'mock': '',
    format: function (data) {
      return {
        url: data.url
      }
    }
  },
  'host.host_components.filtered': {
    'real': '/clusters/{clusterName}/hosts?{fields}',
    'mock': '',
    format: function (data) {
      return {
        headers: {
          'X-Http-Method-Override': 'GET'
        },
        type: 'POST',
        data: JSON.stringify({
          "RequestInfo": {"query": data.parameters}
        })
      };
    }
  },
  'host.status.counters': {
    'real': '/clusters/{clusterName}?fields=Clusters/health_report,Clusters/total_hosts,alerts_summary_hosts&minimal_response=true',
    'mock': '/data/hosts/HDP2/host_status_counters.json'
  },
  'host.stack_versions.install': {
    'real': '/clusters/{clusterName}/hosts/{hostName}/stack_versions',
    'mock': '',
    'type': 'POST',
    'format': function (data) {
      return {
        data: JSON.stringify({
          "HostStackVersions": {
            "stack": data.version.get('stack'),
            "version": data.version.get('version'),
            "repository_version": data.version.get('repoVersion')
          }
        })
      }
    }
  },
  'components.filter_by_status': {
    'real': '/clusters/{clusterName}/components?fields=host_components/HostRoles/host_name,ServiceComponentInfo/component_name,ServiceComponentInfo/started_count{urlParams}&minimal_response=true',
    'mock': ''
  },
  'hosts.all.install': {
    'real': '/hosts?minimal_response=true',
    'mock': ''
  },
  'hosts.all': {
    'real': '/clusters/{clusterName}/hosts?minimal_response=true',
    'mock': '/data/hosts/HDP2/hosts.json'
  },
  'hosts.with_public_host_names': {
    'real': '/clusters/{clusterName}/hosts?fields=Hosts/public_host_name&minimal_response=true',
    'mock': ''
  },
  'hosts.for_quick_links': {
    'real': '/clusters/{clusterName}/hosts?Hosts/host_name.in({masterHosts})&fields=Hosts/public_host_name,host_components/HostRoles/component_name{urlParams}&minimal_response=true',
    'mock': '/data/hosts/quick_links.json'
  },
  'hosts.confirmed.install': {
    'real': '/hosts?fields=Hosts/cpu_count,Hosts/disk_info,Hosts/total_mem&minimal_response=true',
    'mock': ''
  },
  'hosts.confirmed': {
    'real': '/clusters/{clusterName}/hosts?fields=Hosts/cpu_count,Hosts/disk_info,Hosts/total_mem,Hosts/os_type,Hosts/os_arch,Hosts/ip,host_components/HostRoles/state&minimal_response=true',
    'mock': '/data/hosts/HDP2/hosts.json'
  },
  'host_components.all': {
    'real': '/clusters/{clusterName}/host_components?fields=HostRoles/host_name&minimal_response=true',
    'mock': ''
  },
  'host_components.with_services_names': {
    'real': '/clusters/{clusterName}/host_components?fields=component/ServiceComponentInfo/service_name,HostRoles/host_name&minimal_response=true',
    'mock': ''
  },
  'components.get_installed': {
    'real': '/clusters/{clusterName}/components',
    'mock': ''
  },
  'hosts.heatmaps': {
    'real': '/clusters/{clusterName}/hosts?fields=Hosts/host_name,Hosts/public_host_name,Hosts/os_type,Hosts/ip,host_components,metrics/disk,metrics/cpu/cpu_system,metrics/cpu/cpu_user,metrics/memory/mem_total,metrics/memory/mem_free&minimal_response=true',
    'mock': '/data/hosts/HDP2/hosts.json'
  },
  'namenode.cpu_wio': {
    'real': '/clusters/{clusterName}/hosts/{nnHost}?fields=metrics/cpu',
    'mock': '/data/cluster_metrics/cpu.json'
  },

  'custom_action.create': {
    'real': '/requests',
    'mock': '',
    'format': function (data) {
      var requestInfo = {
        context: 'Check host',
        action: 'check_host',
        parameters: {}
      };
      $.extend(true, requestInfo, data.requestInfo);
      return {
        type: 'POST',
        data: JSON.stringify({
          'RequestInfo': requestInfo,
          'Requests/resource_filters': [{
            hosts: data.filteredHosts.join(',')
          }]
        })
      }
    }
  },
  'custom_action.request': {
    'real': '/requests/{requestId}/tasks/{taskId}',
    'mock': '/data/requests/1.json',
    'format': function (data) {
      return {
        requestId: data.requestId,
        taskId: data.taskId || ''
      }
    }
  },
  'hosts.high_availability.wizard': {
    'real': '/clusters/{clusterName}/hosts?fields=Hosts/cpu_count,Hosts/disk_info,Hosts/total_mem&minimal_response=true',
    'mock': ''
  },
  'hosts.security.wizard': {
    'real': '/clusters/{clusterName}/hosts?fields=host_components/HostRoles/service_name&minimal_response=true',
    'mock': ''
  },
  'host_component.installed.on_hosts': {
    'real': '/clusters/{clusterName}/host_components?HostRoles/component_name={componentName}&HostRoles/host_name.in({hostNames})&fields=HostRoles/host_name&minimal_response=true',
    'mock': ''
  },
  'hosts.by_component.one': {
    'real': '/clusters/{clusterName}/hosts?host_components/HostRoles/component_name.in({componentNames})&fields=host_components,Hosts/cpu_count,Hosts/disk_info,Hosts/total_mem,Hosts/ip,Hosts/os_type,Hosts/os_arch,Hosts/public_host_name&page_size=1&minimal_response=true',
    'mock': ''
  },
  'hosts.by_component.all': {
    'real': '/clusters/{clusterName}/hosts?host_components/HostRoles/component_name.in({componentNames})&fields=host_components,Hosts/cpu_count,Hosts/disk_info,Hosts/total_mem,Hosts/ip,Hosts/os_type,Hosts/os_arch,Hosts/public_host_name&minimal_response=true',
    'mock': ''
  },
  'hosts.config_groups': {
    'real': '/clusters/{clusterName}/hosts?fields=Hosts/cpu_count,Hosts/disk_info,Hosts/total_mem,Hosts/ip,Hosts/os_type,Hosts/os_arch,Hosts/public_host_name,host_components&minimal_response=true',
    'mock': ''
  },
  'hosts.host_components.pre_load': {
    real: '',
    mock: '/data/hosts/HDP2/hosts.json',
    format: function (data) {
      return {
        url: data.url
      }
    }
  },
  'hosts.bulk.operations': {
    real: '/clusters/{clusterName}/hosts?fields=Hosts/host_name,Hosts/maintenance_state,' +
    'host_components/HostRoles/state,host_components/HostRoles/maintenance_state,' +
    'Hosts/total_mem,stack_versions/HostStackVersions,stack_versions/repository_versions/RepositoryVersions/repository_version,' +
    'stack_versions/repository_versions/RepositoryVersions/id,' +
    'host_components/HostRoles/stale_configs&minimal_response=true',
    mock: '',
    format: function (data) {
      return {
        headers: {
          'X-Http-Method-Override': 'GET'
        },
        type: 'POST',
        data: JSON.stringify({
          "RequestInfo": {"query": data.parameters}
        })
      }
    }
  },
  'service.serviceConfigVersions.get': {
    real: '/clusters/{clusterName}/configurations/service_config_versions?service_name={serviceName}&fields=service_config_version,user,hosts,group_id,group_name,is_current,createtime,service_name,service_config_version_note&minimal_response=true',
    mock: '/data/configurations/service_versions.json'
  },
  'service.serviceConfigVersions.get.current': {
    real: '/clusters/{clusterName}?fields=Clusters/desired_service_config_versions&minimal_response=true',
    mock: ''
  },
  'service.serviceConfigVersions.get.total': {
    real: '/clusters/{clusterName}/configurations/service_config_versions?page_size=1&minimal_response=true',
    mock: '/data/configurations/service_versions_total.json'
  },
  'service.serviceConfigVersion.get': {
    real: '/clusters/{clusterName}/configurations/service_config_versions?service_name={serviceName}&service_config_version={serviceConfigVersion}',
    mock: '/data/configurations/service_version.json'
  },
  'service.serviceConfigVersions.get.multiple': {
    real: '/clusters/{clusterName}/configurations/service_config_versions?service_name={serviceName}&service_config_version.in({serviceConfigVersions})',
    mock: '/data/configurations/service_version.json',
    format: function (data) {
      return {
        serviceConfigVersions: data.serviceConfigVersions.join(',')
      }
    }
  },
  'service.serviceConfigVersion.revert': {
    'real': '/clusters/{clusterName}',
    'mock': '',
    'type': 'PUT',
    'format': function (data) {
      return {
        data: JSON.stringify(data.data)
      }
    }
  },

  'service.mysql.clean': {
    'real': '/clusters/{clusterName}/requests',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          "RequestInfo": {
            "command" : "CLEAN", "context" : "Clean MYSQL Server"
          },
          "Requests/resource_filters": [{
              "service_name" : "HIVE",
              "component_name" : "MYSQL_SERVER",
              "hosts": data.host
          }]
        })
      }
    }
  },

  'service.mysql.configure': {
    'real': '/clusters/{clusterName}/requests',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          "RequestInfo": {
            "command" : "CONFIGURE", "context" : "Configure MYSQL Server"
          },
          "Requests/resource_filters": [{
              "service_name" : "HIVE",
              "component_name" : "MYSQL_SERVER",
              "hosts": data.host
          }]
        })
      }
    }
  },

  'service.mysql.testHiveConnection': {
    'real': '/requests',
    'mock': '',
    'format': function (data) {
      return {
        type: 'POST',
        data: JSON.stringify({
          "RequestInfo": {
            "context": "Check host",
            "action": "check_host",
            "parameters":{
              "db_name": data.db_name,
              "user_name": data.db_user,
              "user_passwd": data.db_pass,
              "db_connection_url": data.db_connection_url,
              "jdk_location": data.jdk_location,
              "threshold":"60",
              "ambari_server_host": "c6403.ambari.apache.org",
              "check_execute_list":"db_connection_check",
              "java_home": data.java_home,
              "jdk_name": data.jdk_name
            }
          },
          "Requests/resource_filters": [{"hosts": data.hosts}]})
      }
    }
  }
};
/**
 * Replace data-placeholders to its values
 *
 * @param {String} url
 * @param {Object} data
 * @return {String}
 */
var formatUrl = function (url, data) {
  if (!url) return null;
  var keys = url.match(/\{\w+\}/g);
  keys = (keys === null) ? [] : keys;
  if (keys) {
    keys.forEach(function (key) {
      var raw_key = key.substr(1, key.length - 2);
      var replace;
      if (!data || !data[raw_key]) {
        replace = '';
      }
      else {
        replace = data[raw_key];
      }
      url = url.replace(new RegExp(key, 'g'), replace);
    });
  }
  return url;
};

/**
 * this = object from config
 * @return {Object}
 */
var formatRequest = function (data) {
  var opt = {
    type: this.type || 'GET',
    timeout: App.timeout,
    dataType: 'json',
    statusCode: require('data/statusCodes')
  };
  if (App.get('testMode')) {
    opt.url = formatUrl(this.mock ? this.mock : '', data);
    opt.type = 'GET';
  }
  else {
    var prefix = this.apiPrefix != null ? this.apiPrefix : App.apiPrefix;
    opt.url = prefix + formatUrl(this.real, data);
  }

  if (this.format) {
    jQuery.extend(opt, this.format(data, opt));
  }
  return opt;
};

/**
 * Wrapper for all ajax requests
 *
 * @type {Object}
 */
var ajax = Em.Object.extend({
  /**
   * Send ajax request
   *
   * @param {Object} config
   * @return {$.ajax} jquery ajax object
   *
   * config fields:
   *  name - url-key in the urls-object *required*
   *  sender - object that send request (need for proper callback initialization) *required*
   *  data - object with data for url-format
   *  beforeSend - method-name for ajax beforeSend response callback
   *  success - method-name for ajax success response callback
   *  error - method-name for ajax error response callback
   *  callback - callback from <code>App.updater.run</code> library
   */
  send: function (config) {

    if (!config.sender) {
      console.warn('Ajax sender should be defined!');
      return null;
    }

    // default parameters
    var params = {
      clusterName: (App.get('clusterName') || App.clusterStatus.get('clusterName'))
    };

    // extend default parameters with provided
    if (config.data) {
      jQuery.extend(params, config.data);
    }

    var opt = {};
    if (!urls[config.name]) {
      console.warn('Invalid name provided!');
      return null;
    }
    opt = formatRequest.call(urls[config.name], params);

    opt.context = this;

    // object sender should be provided for processing beforeSend, success and error responses
    opt.beforeSend = function (xhr) {
      if (config.beforeSend) {
        config.sender[config.beforeSend](opt, xhr, params);
      }
    };
    opt.success = function (data, textStatus, request) {
      console.log("TRACE: The url is: " + opt.url);
      if (config.success) {
        config.sender[config.success](data, opt, params, request);
      }
    };
    opt.error = function (request, ajaxOptions, error) {
      var KDCErrorMsg = this.getKDCErrorMgs(request);
      if (!Em.isNone(KDCErrorMsg)) {
        /**
         * run this handler before show KDC error popup
         */
        if (config.kdcFailHandler) {
          config.sender[config.kdcFailHandler](request, ajaxOptions, error, opt, params);
        }
        /**
         * run this handler when click cancle on KDC error popup
         */
        if (config.kdcCancelHandler) {
          opt.kdcCancelHandler = function() {
            config.sender[config.kdcCancelHandler]();
          };
        }
        this.defaultErrorKDCHandler(opt, KDCErrorMsg);
      } else if (config.error) {
        config.sender[config.error](request, ajaxOptions, error, opt, params);
      } else {
        this.defaultErrorHandler(request, opt.url, opt.type);
      }
    };
    opt.complete = function () {
      if (config.callback) {
        config.callback();
      }
    };
    if ($.mocho) {
      opt.url = 'http://' + $.hostName + opt.url;
    }
    return $.ajax(opt);
  },

  // A single instance of App.ModalPopup view
  modalPopup: null,
  /**
   * defaultErrorHandler function is referred from App.ajax.send function and App.HttpClient.defaultErrorHandler function
   * @jqXHR {jqXHR Object}
   * @url {string}
   * @method {String} Http method
   * @showStatus {number} HTTP response code which should be shown. Default is 500.
   */
  defaultErrorHandler: function (jqXHR, url, method, showStatus) {
    method = method || 'GET';
    var self = this;
    try {
      var json = $.parseJSON(jqXHR.responseText);
      var message = json.message;
    } catch (err) {
    }
    if (!showStatus) {
      showStatus = 500;
    }
    if (jqXHR.status === showStatus && !this.get('modalPopup')) {
      this.set('modalPopup', App.ModalPopup.show({
        header: Em.I18n.t('common.error'),
        secondary: false,
        onPrimary: function () {
          this.hide();
          self.set('modalPopup', null);
        },
        bodyClass: App.AjaxDefaultErrorPopupBodyView.extend({
          type: method,
          url: url,
          status: jqXHR.status,
          message: message
        })
      }));
    }
  },

  /**
   * defines if it's admin session expiration error
   * and if so returns short error message
   * @param jqXHR
   * @returns {string|null}
   */
  getKDCErrorMgs: function(jqXHR) {
    try {
      var message = $.parseJSON(jqXHR.responseText).message;
    } catch (err) {}
    if (jqXHR.status === 400 && message) {
      return App.format.kdcErrorMsg(message, true);
    }
  },

  /**
   * default handler for admin session expiration error
   * @param {object} opt
   * @param {string} msg
   * @returns {*}
   */
  defaultErrorKDCHandler: function(opt, msg) {
    return App.showInvalidKDCPopup(opt, msg);
  }

});

/**
 * Add few access-methods for test purposes
 */
if ($.mocho) {
  ajax.reopen({
    /**
     * Don't use it anywhere except tests!
     * @returns {Array}
     */
    fakeGetUrlNames: function () {
      return Em.keys(urls);
    },

    /**
     * Don't use it anywhere except tests!
     * @param name
     * @returns {*}
     */
    fakeGetUrl: function (name) {
      return urls[name];
    },

    /**
     * Don't use it anywhere except tests!
     * @param url
     * @param data
     * @returns {String}
     */
    fakeFormatUrl: function (url, data) {
      return formatUrl(url, data);
    },

    /**
     * Don't use it anywhere except tests!
     * @param urlObj
     * @param data
     * @returns {Object}
     */
    fakeFormatRequest: function (urlObj, data) {
      return formatRequest.call(urlObj, data);
    }
  });
}

App.ajax = ajax.create({});

});

require.register("utils/ajax/ajax_queue", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Simple util for executing ajax-requests queue
 *
 * @uses App.ajax
 * @type {Em.Object}
 *
 * Don't add requests directly to <code>queue</code>!
 * Bad example:
 * <code>
 *  var q = App.ajax.Queue.create();
 *  q.get('queue').pushObject({...}); // Don't do this!!!!
 * </code>
 *
 * Don't call <code>runNextRequest</code> directly!
 * Bad example:
 * <code>
 *  var q = App.ajax.Queue.create();
 *  q.runNextRequest(); // Don't do this!!!!
 * </code>
 *
 * Usage example 1:
 * <code>
 *  var q = App.ajax.Queue.create();
 *  q.addRequest({
 *    name: 'some_request',
 *    sender: this,
 *    success: 'someFunc'
 *  }).addRequest({
 *    name: 'some_request2',
 *    sender: this,
 *    success: 'someFunc2'
 *  }).start();
 * </code>
 *
 * Usage example 2:
 * <code>
 *  var q = App.ajax.Queue.create();
 *  q.addRequest({
 *    name: 'some_request',
 *    sender: this,
 *    success: 'someFunc'
 *  });
 *  q.addRequest({
 *    name: 'some_request2',
 *    sender: this,
 *    success: 'someFunc2'
 *  });
 *  q.start();
 * </code>
 *
 * Usage example 3:
 * <code>
 *  var q = App.ajax.Queue.create();
 *  q.addRequests(Em.A([{
 *    name: 'some_request',
 *    sender: this,
 *    success: 'someFunc'
 *  },
 *  {
 *    name: 'some_request2',
 *    sender: this,
 *    success: 'someFunc2'
 *  }]));
 *  q.start();
 * </code>
 */
App.ajaxQueue = Em.Object.extend({

  /**
   * List of requests
   * @type {Object[]}
   */
  queue: Em.A([]),

  /**
   * About query executing if some request failed
   * @type {bool}
   */
  abortOnError: true,

  /**
   * Function called after queue is complete
   * @type {callback}
   */
  finishedCallback: Em.K,

  /**
   * Add request to the <code>queue</code>
   * @param {Object} request object that uses in <code>App.ajax.send</code>
   * @return {App.ajaxQueue}
   * @method addRequest
   */
  addRequest: function(request) {
    Em.assert('Each ajax-request should has non-blank `name`', !Em.isBlank(Em.get(request, 'name')));
    Em.assert('Each ajax-request should has object `sender`', Em.typeOf(Em.get(request, 'sender')) !== 'object');
    this.get('queue').pushObject(request);
    return this;
  },

  /**
   * Add requests to the <code>queue</code>
   * @param {Object[]} requests list of objects that uses in <code>App.ajax.send</code>
   * @return {App.ajaxQueue}
   * @method addRequests
   */
  addRequests: function(requests) {
    requests.map(function(request) {
      this.addRequest(request);
    }, this);
    return this;
  },

  /**
   * Enter point to start requests executing
   * @method start
   */
  start: function() {
    this.runNextRequest();
  },

  /**
   * Execute first request from the <code>queue</code>
   * @method runNextRequest
   */
  runNextRequest: function() {
    var self = this;
    var queue = this.get('queue');
    if (queue.length === 0) {
      this.finishedCallback();
      return;
    }
    var r = App.ajax.send(queue.shift());
    this.propertyDidChange('queue');
    if (r) {
      r.complete(function(xhr) {
        if(xhr.status>=200 && xhr.status <= 299) {
          self.runNextRequest();
        }
        else {
          if (self.get('abortOnError')) {
            self.clear();
          }
          else {
            self.runNextRequest();
          }
        }
      });
    }
    else {
      if (this.get('abortOnError')) {
        this.clear();
      }
      else {
        this.runNextRequest();
      }
    }
  },

  /**
   * Remove all requests from <code>queue</code>
   * @method clear
   */
  clear: function() {
    this.get('queue').clear();
  }

});

});

require.register("utils/base64", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

(function () {

  var
    object = typeof window != 'undefined' ? window : exports,
    chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=',
    INVALID_CHARACTER_ERR = (function () {
      // fabricate a suitable error object
      try {
        document.createElement('$');
      }
      catch (error) {
        return error;
      }
    }());

  // encoder
  // [https://gist.github.com/999166] by [https://github.com/nignag]
  object.btoa || (
    object.btoa = function (input) {
      for (
        // initialize result and counter
        var block, charCode, idx = 0, map = chars, output = '';
        // if the next input index does not exist:
        // change the mapping table to "="
        // check if d has no fractional digits
        input.charAt(idx | 0) || (map = '=', idx % 1);
        // "8 - idx % 1 * 8" generates the sequence 2, 4, 6, 8
        output += map.charAt(63 & block >> 8 - idx % 1 * 8)
        ) {
        charCode = input.charCodeAt(idx += 3 / 4);
        if (charCode > 0xFF) throw INVALID_CHARACTER_ERR;
        block = block << 8 | charCode;
      }
      return output;
    });

  // decoder
  // [https://gist.github.com/1020396] by [https://github.com/atk]
  object.atob || (
    object.atob = function (input) {
      input = input.replace(/=+$/, '');
      if (input.length % 4 == 1) throw INVALID_CHARACTER_ERR;
      for (
        // initialize result and counters
        var bc = 0, bs, buffer, idx = 0, output = '';
        // get next character
        buffer = input.charAt(idx++);
        // character found in table? initialize bit storage and add its ascii value;
        ~buffer && (bs = bc % 4 ? bs * 64 + buffer : buffer,
          // and if not first of each 4 characters,
          // convert the first 8 bits to one ascii character
          bc++ % 4) ? output += String.fromCharCode(255 & bs >> (-2 * bc & 6)) : 0
        ) {
        // try to find character in table (0-63, not found => -1)
        buffer = chars.indexOf(buffer);
      }
      return output;
    });

}());
});

require.register("utils/batch_scheduled_requests", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * Default success callback for ajax-requests in this module
 * @type {Function}
 */
var defaultSuccessCallback = function(data, ajaxOptions, params) {
  App.router.get('applicationController').dataLoading().done(function(initValue) {
    params.query && params.query.set('status', 'SUCCESS');
    if (initValue) {
      App.router.get('backgroundOperationsController').showPopup();
    }
  });
};
/**
 * Default error callback for ajax-requests in this module
 * @param {Object} xhr
 * @param {String} textStatus
 * @param {String} error
 * @param {Object} opt
 * @type {Function}
 */
var defaultErrorCallback = function(xhr, textStatus, error, opt, params) {
  params.query && params.query.set('status', 'FAIL');
  App.ajax.defaultErrorHandler(xhr, opt.url, 'POST', xhr.status);
};

/**
 * Contains helpful utilities for handling batch and scheduled requests.
 */
module.exports = {

  /**
   * Some services have components which have a need for rolling restarts. This
   * method returns the name of the host-component which supports rolling
   * restarts for a service.
   * @param {String} serviceName
   */
  getRollingRestartComponentName: function(serviceName) {
    var rollingRestartComponents = {
      HDFS: 'DATANODE',
      YARN: 'NODEMANAGER',
      HBASE: 'HBASE_REGIONSERVER',
      STORM: 'SUPERVISOR'
    };
    return rollingRestartComponents[serviceName] ? rollingRestartComponents[serviceName] : null;
  },

  /**
   * Facade-function for restarting host components of specific service
   * @param {String} serviceName for which service hostComponents should be restarted
   * @param {bool} staleConfigsOnly restart only hostComponents with <code>staleConfig</code> true
   * @param {Object} query
   * @param {bool} runMmOperation
   */
  restartAllServiceHostComponents: function(serviceName, staleConfigsOnly, query, runMmOperation) {
    var self = this;
    var context = staleConfigsOnly ? Em.I18n.t('rollingrestart.context.allWithStaleConfigsForSelectedService').format(serviceName) : Em.I18n.t('rollingrestart.context.allForSelectedService').format(serviceName);

    if (runMmOperation) {
      this.turnOnOffPassiveRequest('ON', Em.I18n.t('passiveState.turnOnFor').format(serviceName), serviceName);
    }
    this.getComponentsFromServer({
      services: [serviceName],
      staleConfigs: staleConfigsOnly ? staleConfigsOnly : null,
      passiveState: 'OFF',
      displayParams: ['host_components/HostRoles/component_name']
    }, function (data) {
      var hostComponents = [];
      data.items.forEach(function (host) {
        host.host_components.forEach(function (hostComponent) {
          hostComponents.push(Em.Object.create({
            componentName: hostComponent.HostRoles.component_name,
            hostName: host.Hosts.host_name
          }))
        });
      });
      self.restartHostComponents(hostComponents, context, "SERVICE", query);
    });
  },

  /**
   * construct URL from parameters for request in <code>getComponentsFromServer()</code>
   * @param options
   * @return {{fields: string, params: string}}
   */
  constructComponentsCallUrl: function (options) {
    var multipleValueParams = {
        'services': 'host_components/HostRoles/service_name.in(<entity-names>)',
        'hosts': 'Hosts/host_name.in(<entity-names>)',
        'components': 'host_components/HostRoles/component_name.in(<entity-names>)'
      },
      singleValueParams = {
        staleConfigs: 'host_components/HostRoles/stale_configs=',
        passiveState: 'Hosts/maintenance_state=',
        workStatus: 'host_components/HostRoles/state='
      },
      displayParams = options.displayParams || [],
      urlParams = '',
      addAmpersand = false;

    for (var i in multipleValueParams) {
      var arrayParams = options[i];
      if (Array.isArray(arrayParams) && arrayParams.length > 0) {
        if (addAmpersand) {
          urlParams += '&';
          addAmpersand = false;
        }
        urlParams += multipleValueParams[i].replace('<entity-names>', arrayParams.join(','));
        addAmpersand = true;
      }
    }

    for (var j in singleValueParams) {
      var param = options[j];
      if (!Em.isNone(param)) {
        urlParams += (addAmpersand) ? '&' : '';
        urlParams += singleValueParams[j] + param.toString();
        addAmpersand = true;
      }
    }
    var params = urlParams,
      fields = '';
    displayParams.forEach(function (displayParam, index, array) {
      if (index === 0) {
        fields += (addAmpersand) ? '&' : '';
        fields += 'fields=';
      }
      fields += displayParam;
      fields += (array.length === (index + 1)) ? '' : ",";
    });
    fields += '&minimal_response=true';

    return {fields: fields.substring(1, fields.length), params: params};
  },

  /**
   * make GET call to server in order to obtain host-components
   * which correspond to filter params from <code>options</code>
   * @param options
   * @param callback
   */
  getComponentsFromServer: function (options, callback) {
    var request_parameters = this.constructComponentsCallUrl(options);
    return App.ajax.send({
      name: 'host.host_components.filtered',
      sender: this,
      data: {
        parameters: request_parameters.params,
        fields: request_parameters.fields,
        callback: callback
      },
      success: 'getComponentsFromServerSuccessCallback'
    });
  },

  /**
   * pass request outcome to <code>callback()<code>
   * @param data
   * @param opt
   * @param params
   */
  getComponentsFromServerSuccessCallback: function (data, opt, params) {
    params.callback(data);
  },

  /**
   * Restart list of host components
   * @param {Ember.Enumerable} hostComponentsList list of host components should be restarted
   * @param {String} context message to show in BG popup
   * @param {String} level - operation level, can be ("CLUSTER", "SERVICE", "HOST", "HOSTCOMPONENT")
   * @param {String} query
   */
  restartHostComponents: function (hostComponentsList, context, level, query) {
    context = context || Em.I18n.t('rollingrestart.context.default');
    /**
     * Format: {
     *  'DATANODE': ['host1', 'host2'],
     *  'NAMENODE': ['host1', 'host3']
     *  ...
     * }
     */
    var componentToHostsMap = {};
    var hosts = [];
    hostComponentsList.forEach(function(hc) {
      var hostName = hc.get('hostName');
      var componentName = hc.get('componentName');
      if (!componentToHostsMap[componentName]) {
        componentToHostsMap[componentName] = [];
      }
      componentToHostsMap[componentName].push(hostName);
      hosts.push(hostName);
    });
    var resource_filters = [];
    for (var componentName in componentToHostsMap) {
      if (componentToHostsMap.hasOwnProperty(componentName)) {
        resource_filters.push({
          service_name: App.StackServiceComponent.find(componentName).get('serviceName'),
          component_name: componentName,
          hosts: componentToHostsMap[componentName].join(",")
        });
      }
    }
    if (hostComponentsList.length > 0) {
      var serviceComponentName = hostComponentsList[0].get("componentName");
      var serviceName = App.StackServiceComponent.find(serviceComponentName).get('serviceName');
      var operation_level = this.getOperationLevelobject(level, serviceName,
        serviceComponentName);
    }


    if (resource_filters.length) {
      App.ajax.send({
        name: 'restart.hostComponents',
        sender: {
          successCallback: defaultSuccessCallback,
          errorCallback: defaultErrorCallback
        },
        data: {
          context: context,
          resource_filters: resource_filters,
          query: query,
          operation_level: operation_level
        },
        success: 'successCallback',
        error: 'errorCallback'
      });
    }
  },

  /**
   * @param {String} level - operation level name, can be ("CLUSTER", "SERVICE", "HOST", "HOSTCOMPONENT")
   * @param {String} serviceName
   * @param {String} componentName
   * @returns {Object} {{level: *, cluster_name: *}} - operation level object
   * @method getOperationLevelobject - create operation level object to be included into ajax query
   */
  getOperationLevelobject: function(level, serviceName, componentName) {
    var operationLevel = {
      "level": level,
      "cluster_name": App.get("clusterName")
    };
    if (level === "SERVICE") {
      operationLevel["service_name"] = serviceName;
    } else if(level !== "HOST") {
      operationLevel["service_name"] = serviceName;
      operationLevel["hostcomponent_name"] = componentName;
    }
    return operationLevel;
  },

  turnOnOffPassiveRequest: function(state, message, serviceName, callback) {
    return App.ajax.send({
      'name': 'common.service.passive',
      'sender': {
        'successCallback': callback || defaultSuccessCallback,
        'errorCallback': defaultErrorCallback
      },
      'data': {
        'requestInfo': message,
        'serviceName': serviceName,
        'passive_state': state
      },
      'success': 'successCallback'
    });
  },
  /**
   * Makes a REST call to the server requesting the rolling restart of the
   * provided host components.
   * @param {Array} restartHostComponents list of host components should be restarted
   * @param {Number} batchSize size of each batch
   * @param {Number} intervalTimeSeconds delay between two batches
   * @param {Number} tolerateSize task failure tolerance
   * @param {callback} successCallback
   * @param {callback} errorCallback
   */
  _doPostBatchRollingRestartRequest: function(restartHostComponents, batchSize, intervalTimeSeconds, tolerateSize, successCallback, errorCallback) {
    successCallback = successCallback || defaultSuccessCallback;
    errorCallback = errorCallback || defaultErrorCallback;
    if (!restartHostComponents.length) {
      console.log('No batch rolling restart if no restartHostComponents provided!');
      return;
    }
    App.ajax.send({
      name: 'rolling_restart.post',
      sender: {
        successCallback: successCallback,
        errorCallback: errorCallback
      },
      data: {
        intervalTimeSeconds: intervalTimeSeconds,
        tolerateSize: tolerateSize,
        batches: this.getBatchesForRollingRestartRequest(restartHostComponents, batchSize)
      },
      success: 'successCallback',
      error: 'errorCallback'
    });
  },

  /**
   * Create list of batches for rolling restart request
   * @param {Array} restartHostComponents list host components should be restarted
   * @param {Number} batchSize size of each batch
   * @returns {Array} list of batches
   */
  getBatchesForRollingRestartRequest: function(restartHostComponents, batchSize) {
    var hostIndex = 0,
        batches = [],
        batchCount = Math.ceil(restartHostComponents.length / batchSize),
        sampleHostComponent = restartHostComponents.objectAt(0),
        componentName = sampleHostComponent.get('componentName'),
        serviceName = sampleHostComponent.get('serviceName');

    for ( var count = 0; count < batchCount; count++) {
      var hostNames = [];
      for ( var hc = 0; hc < batchSize && hostIndex < restartHostComponents.length; hc++) {
        hostNames.push(restartHostComponents.objectAt(hostIndex++).get('hostName'));
      }
      if (hostNames.length > 0) {
        batches.push({
          "order_id" : count + 1,
          "type" : "POST",
          "uri" : App.apiPrefix + "/clusters/" + App.get('clusterName') + "/requests",
          "RequestBodyInfo" : {
            "RequestInfo" : {
              "context" : "_PARSE_.ROLLING-RESTART." + componentName + "." + (count + 1) + "." + batchCount,
              "command" : "RESTART"
            },
            "Requests/resource_filters": [{
              "service_name" : serviceName,
              "component_name" : componentName,
              "hosts" : hostNames.join(",")
            }]
          }
        });
      }
    }
    return batches;
  },

  /**
   * Launches dialog to handle rolling restarts of host components.
   *
   * Rolling restart is supported only for components listed in <code>getRollingRestartComponentName</code>
   * @see getRollingRestartComponentName
   * @param {String} hostComponentName
   *           Type of host-component to restart across cluster
   *          (ex: DATANODE)
   * @param {bool} staleConfigsOnly
   *           Pre-select host-components which have stale
   *          configurations
   */
  launchHostComponentRollingRestart: function(hostComponentName, serviceName, isMaintenanceModeOn, staleConfigsOnly, skipMaintenance) {
    if (App.get('components.rollinRestartAllowed').contains(hostComponentName)) {
      this.showRollingRestartPopup(hostComponentName, serviceName, isMaintenanceModeOn, staleConfigsOnly, null, skipMaintenance);
    }
    else {
      this.showWarningRollingRestartPopup(hostComponentName);
    }
  },

  /**
   * Show popup with rolling restart dialog
   * @param {String} hostComponentName name of the host components that should be restarted
   * @param {bool} staleConfigsOnly restart only components with <code>staleConfigs</code> = true
   * @param {App.hostComponent[]} hostComponents list of hostComponents that should be restarted (optional).
   * Using this parameter will reset hostComponentName
   */
  showRollingRestartPopup: function(hostComponentName, serviceName, isMaintenanceModeOn, staleConfigsOnly, hostComponents, skipMaintenance) {
    hostComponents = hostComponents || [];
    var componentDisplayName = App.format.role(hostComponentName);
    var self = this;
    if (!componentDisplayName) {
      componentDisplayName = hostComponentName;
    }
    var title = Em.I18n.t('rollingrestart.dialog.title').format(componentDisplayName);
    var viewExtend = {
      turnOnMmMsg: Em.I18n.t('passiveState.turnOnFor').format(serviceName),
      turnOnMm: false,
      staleConfigsOnly : staleConfigsOnly,
      hostComponentName : hostComponentName,
      skipMaintenance: skipMaintenance,
      serviceName: serviceName,
      isServiceInMM: isMaintenanceModeOn,
      didInsertElement: function () {
        var view = this;

        this.set('parentView.innerView', this);
        if (hostComponents.length) {
          view.initialize();
        } else {
          self.getComponentsFromServer({
            components: [hostComponentName],
            displayParams: ['host_components/HostRoles/stale_configs', 'Hosts/maintenance_state', 'host_components/HostRoles/maintenance_state'],
            staleConfigs: staleConfigsOnly ? staleConfigsOnly : null
          }, function (data) {
            var wrappedHostComponents = [];
            data.items.forEach(function (host) {
              host.host_components.forEach(function(hostComponent){
                wrappedHostComponents.push(Em.Object.create({
                  componentName: hostComponent.HostRoles.component_name,
                  serviceName: App.StackServiceComponent.find(hostComponent.HostRoles.component_name).get('serviceName'),
                  hostName: host.Hosts.host_name,
                  staleConfigs: hostComponent.HostRoles.stale_configs,
                  hostPassiveState: host.Hosts.maintenance_state,
                  passiveState: hostComponent.HostRoles.maintenance_state
                }));
              });
            });
            view.set('allHostComponents', wrappedHostComponents);
            view.initialize();
          });
        }
      }
    };
    if (hostComponents.length) {
      viewExtend.allHostComponents = hostComponents;
    }

    App.ModalPopup.show({
      header : title,
      hostComponentName : hostComponentName,
      serviceName: serviceName,
      isServiceInMM: isMaintenanceModeOn,
      staleConfigsOnly : staleConfigsOnly,
      skipMaintenance: skipMaintenance,
      innerView : null,
      bodyClass : App.RollingRestartView.extend(viewExtend),
      classNames : [ 'rolling-restart-popup' ],
      primary : Em.I18n.t('rollingrestart.dialog.primary'),
      onPrimary : function() {
        var dialog = this;
        var restartComponents = this.get('innerView.restartHostComponents');
        var batchSize = this.get('innerView.batchSize');
        var waitTime = this.get('innerView.interBatchWaitTimeSeconds');
        var tolerateSize = this.get('innerView.tolerateSize');
        if (this.get('innerView.turnOnMm')) {
          self.turnOnOffPassiveRequest('ON', Em.I18n.t('passiveState.turnOnFor').format(serviceName), serviceName.toUpperCase());
        }
        self._doPostBatchRollingRestartRequest(restartComponents, batchSize, waitTime, tolerateSize, function(data, ajaxOptions, params) {
          dialog.hide();
          defaultSuccessCallback(data, ajaxOptions, params);
        });
      },
      updateButtons : function() {
        var errors = this.get('innerView.errors');
        this.set('disablePrimary', (errors != null && errors.length > 0))
      }.observes('innerView.errors')
    });
  },

  /**
   * Show warning popup about not supported host components
   * @param {String} hostComponentName
   */
  showWarningRollingRestartPopup: function(hostComponentName) {
    var componentDisplayName = App.format.role(hostComponentName);
    if (!componentDisplayName) {
      componentDisplayName = hostComponentName;
    }
    var title = Em.I18n.t('rollingrestart.dialog.title').format(componentDisplayName);
    var msg = Em.I18n.t('rollingrestart.notsupported.hostComponent').format(componentDisplayName);
    console.log(msg);
    App.ModalPopup.show({
      header : title,
      secondary : false,
      msg : msg,
      bodyClass : Em.View.extend({
        template : Em.Handlebars.compile('<div class="alert alert-warning">{{msg}}</div>')
      })
    });
  },

  /**
   * Warn user that alerts will be updated in few minutes
   * @param {String} hostComponentName
   */
  infoPassiveState: function(passiveState) {
    var enabled = passiveState == 'OFF' ? 'enabled' : 'suppressed';
    App.ModalPopup.show({
      header: Em.I18n.t('common.information'),
      secondary: null,
      bodyClass: Ember.View.extend({
        template: Ember.Handlebars.compile('<p>{{view.message}}</p>'),
        message: function() {
          return Em.I18n.t('hostPopup.warning.alertsTimeOut').format(passiveState.toLowerCase(), enabled);
        }.property()
      })
    });
  },

  /**
   * Retrieves the latest information about a specific request schedule
   * identified by 'requestScheduleId'
   *
   * @param {Number} requestScheduleId ID of the request schedule to get
   * @param {Function} successCallback Called with request_schedule data from server. An
   *          empty object returned for invalid ID.
   * @param {Function} errorCallback Optional error callback. Default behavior is to
   *          popup default error dialog.
   */
  getRequestSchedule: function(requestScheduleId, successCallback, errorCallback) {
    if (requestScheduleId != null && !isNaN(requestScheduleId) && requestScheduleId > -1) {
      errorCallback = errorCallback ? errorCallback : defaultErrorCallback;
      App.ajax.send({
        name : 'request_schedule.get',
        sender : {
          successCallbackFunction : function(data) {
            successCallback(data);
          },
          errorCallbackFunction : function(xhr, textStatus, error, opt) {
            errorCallback(xhr, textStatus, error, opt);
          }
        },
        data : {
          request_schedule_id : requestScheduleId
        },
        success : 'successCallbackFunction',
        error : 'errorCallbackFunction'
      });
    } else {
      successCallback({});
    }
  },

  /**
   * Attempts to abort a specific request schedule identified by 'requestScheduleId'
   *
   * @param {Number} requestScheduleId ID of the request schedule to get
   * @param {Function} successCallback Called when request schedule successfully aborted
   * @param {Function} errorCallback Optional error callback. Default behavior is to
   *          popup default error dialog.
   */
  doAbortRequestSchedule: function(requestScheduleId, successCallback, errorCallback) {
    if (requestScheduleId != null && !isNaN(requestScheduleId) && requestScheduleId > -1) {
      errorCallback = errorCallback || defaultErrorCallback;
      App.ajax.send({
        name : 'common.delete.request_schedule',
        sender : {
          successCallbackFunction : function(data) {
            successCallback(data);
          },
          errorCallbackFunction : function(xhr, textStatus, error, opt) {
            errorCallback(xhr, textStatus, error, opt);
          }
        },
        data : {
          request_schedule_id : requestScheduleId
        },
        success : 'successCallbackFunction',
        error : 'errorCallbackFunction'
      });
    } else {
      successCallback({});
    }
  }
};

});

require.register("utils/blueprint", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

module.exports = {
  mergeBlueprints: function(masterBlueprint, slaveBlueprint) {
    var self = this;

    // Check edge cases
    if (!slaveBlueprint && !masterBlueprint) {
      throw 'slaveBlueprint or masterBlueprint should not be empty';
    } else if (slaveBlueprint && !masterBlueprint) {
      return slaveBlueprint;
    } else if (!slaveBlueprint && masterBlueprint) {
      return masterBlueprint;
    }

    // Work with main case (both blueprint are presented)
    var matches = self.matchGroups(masterBlueprint, slaveBlueprint);

    var res = {
      blueprint: { host_groups: [] },
      blueprint_cluster_binding: { host_groups: [] }
    };

    matches.forEach(function(match, i){
      var group_name = 'host-group-' + (i+1);

      var masterComponents = self.getComponentsFromBlueprintByGroupName(masterBlueprint, match.g1);
      var slaveComponents = self.getComponentsFromBlueprintByGroupName(slaveBlueprint, match.g2);

      res.blueprint.host_groups.push({
        name: group_name,
        components: masterComponents.concat(slaveComponents)
      });

      res.blueprint_cluster_binding.host_groups.push({
        name: group_name,
        hosts: self.getHostsFromBlueprintByGroupName(match.g1 ? masterBlueprint : slaveBlueprint, match.g1 ? match.g1 : match.g2)
      });
    });
    return res;
  },

  getHostsFromBlueprint: function(blueprint) {
    return blueprint.blueprint_cluster_binding.host_groups.mapProperty("hosts").reduce(function(prev, curr){ return prev.concat(curr); }, []).mapProperty("fqdn");
  },

  getHostsFromBlueprintByGroupName: function(blueprint, groupName) {
    if (groupName) {
      var group = blueprint.blueprint_cluster_binding.host_groups.find(function(g) {
        return g.name === groupName;
      });

      if (group) {
        return group.hosts;
      }
    }
    return [];
  },

  getComponentsFromBlueprintByGroupName: function(blueprint, groupName) {
    if (groupName) {
      var group = blueprint.blueprint.host_groups.find(function(g) {
        return g.name === groupName;
      });

      if (group) {
        return group.components;
      }
    }
    return [];
  },

  matchGroups: function(masterBlueprint, slaveBlueprint) {
    var self = this;
    var res = [];

    var groups1 = masterBlueprint.blueprint_cluster_binding.host_groups;
    var groups2 = slaveBlueprint.blueprint_cluster_binding.host_groups;

    var groups1_used = groups1.map(function() { return false; });
    var groups2_used = groups2.map(function() { return false; });

    self.matchGroupsWithLeft(groups1, groups2, groups1_used, groups2_used, res, false);
    self.matchGroupsWithLeft(groups2, groups1, groups2_used, groups1_used, res, true);

    return res;
  },

  matchGroupsWithLeft: function(groups1, groups2, groups1_used, groups2_used, res, inverse) {
    for (var i = 0; i < groups1.length; i++) {
      if (groups1_used[i]) {
        continue;
      }

      var group1 = groups1[i];
      groups1_used[i] = true;

      var group2 = groups2.find(function(g2, index) {
        if (group1.hosts.length != g2.hosts.length) {
          return false;
        }

        for (var gi = 0; gi < group1.hosts.length; gi++) {
          if (group1.hosts[gi].fqdn != g2.hosts[gi].fqdn) {
            return false;
          }
        }

        groups2_used[index] = true;
        return true;
      });

      var item = {};

      if (inverse) {
        item.g2 = group1.name;
        if (group2) {
          item.g1 = group2.name;
        }
      } else {
        item.g1 = group1.name;
        if (group2) {
          item.g2 = group2.name;
        }
      }
      res.push(item);
    }
  },

  /**
   * Remove from blueprint all components expect given components
   * @param blueprint
   * @param [string] components
   */
  filterByComponents: function(blueprint, components) {
    if (!blueprint) {
      return null;
    }

    var res = JSON.parse(JSON.stringify(blueprint));
    var emptyGroups = [];

    for (var i = 0; i < res.blueprint.host_groups.length; i++) {
      res.blueprint.host_groups[i].components = res.blueprint.host_groups[i].components.filter(function(c) {
        return components.contains(c.name);
      });

      if (res.blueprint.host_groups[i].components.length == 0) {
        emptyGroups.push(res.blueprint.host_groups[i].name);
      }
    }

    res.blueprint.host_groups = res.blueprint.host_groups.filter(function(g) {
      return !emptyGroups.contains(g.name);
    });

    res.blueprint_cluster_binding.host_groups = res.blueprint_cluster_binding.host_groups.filter(function(g) {
      return !emptyGroups.contains(g.name);
    });

    return res;
  },

  addComponentsToBlueprint: function(blueprint, components) {
    var res = JSON.parse(JSON.stringify(blueprint))

    res.blueprint.host_groups.forEach(function(group) {
      components.forEach(function(component) {
        group.components.push({ name: component });
      });
    });

    return res;
  },

  /**
   * @method buildConfisJSON - generet JSON according to blueprint format
   * @param {Em.Array} stepConfigs - array of Ember Objects
   * @param {Array} services
   * @returns {Object}
   * Example:
   * {
   *   "yarn-env": {
   *     "properties": {
   *       "content": "some value",
   *       "yarn_heapsize": "1024",
   *       "resourcemanager_heapsize": "1024",
   *     }
   *   },
   *   "yarn-log4j": {
   *     "properties": {
   *       "content": "some other value"
   *     }
   *   }
   * }
   */
  buildConfisJSON: function(services, stepConfigs) {
    var configurations = {};
    services.forEach(function(service) {
      var config = stepConfigs.findProperty('serviceName', service.get('serviceName'));
      if (config && service.get('configTypes')) {
        Object.keys(service.get('configTypes')).forEach(function(type) {
          if(!configurations[type]){
            configurations[type] = {
              properties: {}
            }
          }
        });
        config.get('configs').forEach(function(property){
          if (configurations[property.get('filename').replace('.xml','')]){
            configurations[property.get('filename').replace('.xml','')]['properties'][property.get('name')] = property.get('value');
          } else {
            console.warn(property.get('name') + " from " + property.get('filename') + " can't be validate");
          }
        });
      }
    });
    return configurations;
  },

  /**
   * @method generateHostGroups
   * @param {Array} hostNames - list of all hostNames
   * @param {Array} hostComponents - list of all hostComponents
   * @returns {{blueprint: {host_groups: Array}, blueprint_cluster_binding: {host_groups: Array}}}
   */
  generateHostGroups: function(hostNames, hostComponents) {
    var recommendations = {
      blueprint: {
        host_groups: []
      },
      blueprint_cluster_binding: {
        host_groups: []
      }
    };

    for (var i = 1; i <= hostNames.length; i++) {
      var host_group = {
        name: "host-group-" + i,
        components: []
      };
      var hcFiltered = hostComponents.filterProperty('hostName', hostNames[i-1]).mapProperty('componentName');
      for (var j = 0; j < hcFiltered.length; j++) {
        host_group.components.push({
          name: hcFiltered[j]
        });
      }
      recommendations.blueprint.host_groups.push(host_group);
      recommendations.blueprint_cluster_binding.host_groups.push({
        name: "host-group-" + i,
        hosts: [{
          fqdn: hostNames[i-1]
        }]
      });
    }
    return recommendations;
  }
};

});

require.register("utils/components", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

module.exports = {
  installHostComponent: function(hostName, component) {
    var
      self = this,
      componentName = component.get('componentName'),
      displayName = component.get('displayName');
    App.ajax.send({
      name: 'host.host_component.add_new_component',
      sender: self,
      data: {
        hostName: hostName,
        component: component,
        data: JSON.stringify({
          RequestInfo: {
            "context": Em.I18n.t('requestInfo.installHostComponent') + " " + displayName
          },
          Body: {
            host_components: [
              {
                HostRoles: {
                  component_name: componentName
                }
              }
            ]
          }
        })
      },
      success: 'addNewComponentSuccessCallback',
      error: 'ajaxErrorCallback'
    });
  },

  /**
   * Success callback for add host component request
   * @param {object} data
   * @param {object} opt
   * @param {object} params
   * @method addNewComponentSuccessCallback
   */
  addNewComponentSuccessCallback: function (data, opt, params) {
    console.log('Send request for ADDING NEW COMPONENT successfully');
    App.ajax.send({
      name: 'common.host.host_component.update',
      sender: App.router.get('mainHostDetailsController'),
      data: {
        hostName: params.hostName,
        componentName: params.component.get('componentName'),
        serviceName: params.component.get('serviceName'),
        component: params.component,
        "context": Em.I18n.t('requestInfo.installNewHostComponent') + " " + params.component.get('displayName'),
        HostRoles: {
          state: 'INSTALLED'
        },
        urlParams: "HostRoles/state=INIT"
      },
      success: 'installNewComponentSuccessCallback',
      error: 'ajaxErrorCallback'
    });
  },

  /**
   * Default error-callback for ajax-requests in current page
   * @param {object} request
   * @param {object} ajaxOptions
   * @param {string} error
   * @param {object} opt
   * @param {object} params
   * @method ajaxErrorCallback
   */
  ajaxErrorCallback: function (request, ajaxOptions, error, opt, params) {
    console.log('error on change component host status');
    App.ajax.defaultErrorHandler(request, opt.url, opt.method);
  },

  downloadClientConfigs: function (data) {
    var isForHost = typeof data.hostName !== 'undefined';
    var url = App.get('apiPrefix') + '/clusters/' + App.router.getClusterName() + '/' +
      (isForHost ? 'hosts/' + data.hostName + '/host_components/' : 'services/' + data.serviceName + '/components/') +
      data.componentName + '?format=client_config_tar';
    try {
      var self = this;
      $.fileDownload(url).fail(function (error) {
        var errorMessage = '';
        var isNoConfigs = false;
        if (error && $(error).text()) {
          var errorObj = JSON.parse($(error).text());
          if (errorObj && errorObj.message && errorObj.status) {
            isNoConfigs = errorObj.message.indexOf(Em.I18n.t('services.service.actions.downloadClientConfigs.fail.noConfigFile')) !== -1;
            errorMessage += isNoConfigs ? Em.I18n.t('services.service.actions.downloadClientConfigs.fail.noConfigFile') :
              Em.I18n.t('services.service.actions.downloadClientConfigs.fail.popup.body.errorMessage').format(data.displayName, errorObj.status, errorObj.message);
          } else {
            errorMessage += Em.I18n.t('services.service.actions.downloadClientConfigs.fail.popup.body.noErrorMessage').format(data.displayName);
          }
          errorMessage += isNoConfigs ? '' : Em.I18n.t('services.service.actions.downloadClientConfigs.fail.popup.body.question');
        } else {
          errorMessage += Em.I18n.t('services.service.actions.downloadClientConfigs.fail.popup.body.noErrorMessage').format(data.displayName) +
            Em.I18n.t('services.service.actions.downloadClientConfigs.fail.popup.body.question');
        }
        App.ModalPopup.show({
          header: Em.I18n.t('services.service.actions.downloadClientConfigs.fail.popup.header').format(data.displayName),
          bodyClass: Ember.View.extend({
            template: Em.Handlebars.compile(errorMessage)
          }),
          secondary: isNoConfigs ? false : Em.I18n.t('common.cancel'),
          onPrimary: function () {
            this.hide();
            if (!isNoConfigs) {
              self.downloadClientConfigs({
                context: Em.Object.create(data)
              })
            }
          }
        });
      });
    } catch (err) {
      var newWindow = window.open(url);
      newWindow.focus();
    }
  },
  /**
   * Check if all required components are installed on host.
   * Available options:
   *  scope: 'host' - dependency level `host`,`cluster` or `*`.
   *  hostName: 'example.com' - host name to search installed components
   *  installedComponents: ['A', 'B'] - names of installed components
   *
   * By default scope level is `*`
   * For host level dependency you should specify at least `hostName` or `installedComponents` attribute.
   *
   * @param {String} componentName
   * @param {Object} opt - options. Allowed options are `hostName`, `installedComponents`, `scope`.
   * @return {Array} - names of missed components
   */
  checkComponentDependencies: function(componentName, opt) {
    opt = opt || {};
    opt.scope = opt.scope || '*';
    var installedComponents;
    var dependencies = App.StackServiceComponent.find(componentName).get('dependencies');
    dependencies = opt.scope === '*' ? dependencies : dependencies.filterProperty('scope', opt.scope);
    if (dependencies.length == 0) return [];
    switch (opt.scope) {
      case 'host':
        Em.assert("You should pass at least `hostName` or `installedComponents` to options.", opt.hostName || opt.installedComponents);
        installedComponents = opt.installedComponents || App.HostComponent.find().filterProperty('hostName', opt.hostName).mapProperty('componentName').uniq();
        break;
      default:
        // @todo: use more appropriate value regarding installed components
        installedComponents = opt.installedComponents || App.HostComponent.find().mapProperty('componentName').uniq();
        break;
    }
    return dependencies.filter(function(dependency) {
      return !installedComponents.contains(dependency.componentName);
    }).mapProperty('componentName');
  }

};
});

require.register("utils/config", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

var configGroupsByTag = [];

App.config = Em.Object.create({

  CONFIG_GROUP_NAME_MAX_LENGTH: 18,

  /**
   * filename exceptions used to support substandard sitenames which don't have "xml" extension
   */
  filenameExceptions: [],

  preDefinedServiceConfigs: [],
  /**
   *
   * Returns file name version that stored on server.
   *
   * Example:
   *   App.config.getOriginalFileName('core-site') // returns core-site.xml
   *   App.config.getOriginalFileName('zoo.cfg') // returns zoo.cfg
   *
   * @param {String} fileName
   * @method getOriginalFileName
   **/
  getOriginalFileName: function (fileName) {
    if (/\.xml$/.test(fileName)) return fileName;
    return this.get('filenameExceptions').contains(fileName) ? fileName : fileName + '.xml';
  },

  /**
   * Check if Hive installation with new MySQL database created via Ambari is allowed
   * @param osFamily
   * @returns {boolean}
   */
  isManagedMySQLForHiveAllowed: function (osFamily) {
    var osList = ['redhat5', 'suse11'];
    return !osList.contains(osFamily);
  },

  /**
   *
   * Returns the configuration tagName from supplied filename
   *
   * Example:
   *   App.config.getConfigTagFromFileName('core-site.xml') // returns core-site
   *   App.config.getConfigTagFromFileName('zoo.cfg') // returns zoo.cfg
   *
   * @param {String} fileName
   * @method getConfigTagFromFileName
   **/
  getConfigTagFromFileName: function (fileName) {
    return fileName.endsWith('.xml') ? fileName.slice(0, -4) : fileName;
  },

  setPreDefinedServiceConfigs: function (isMiscTabToBeAdded) {
    var configs = this.get('preDefinedSiteProperties');
    var services = [];
    var stackServices = App.StackService.find().filterProperty('id');
    // Only include services that has configTypes related to them for service configuration page
    var servicesWithConfigTypes = stackServices.filter(function (service) {
      var configtypes = service.get('configTypes');
      return configtypes && !!Object.keys(configtypes).length;
    }, this);

    var allTabs;
    if (isMiscTabToBeAdded) {
      var nonServiceTab = require('data/service_configs');
      var miscService = nonServiceTab.findProperty('serviceName', 'MISC');
      var tagTypes = {};
      servicesWithConfigTypes.mapProperty('configTypes').forEach(function(configTypes) {
        for (var fileName in configTypes) {
          if (fileName.endsWith('-env') && !miscService.get('configTypes')[fileName]) {
            tagTypes[fileName] = configTypes[fileName];
          }
        }
      });
      miscService.set('configTypes', $.extend(miscService.get('configTypes'), tagTypes));
      allTabs = servicesWithConfigTypes.concat(nonServiceTab);
    } else {
      allTabs = servicesWithConfigTypes;
    }

    allTabs.forEach(function (service) {
      var serviceConfigs = configs.filterProperty('serviceName', service.get('serviceName'));
      service.set('configs', serviceConfigs);
      services.push(service);
    });
    this.set('preDefinedServiceConfigs', services);
  },

  configMapping: require('data/HDP2/config_mapping'),

  preDefinedSiteProperties: function () {
    var sitePropertiesForCurrentStack = this.preDefinedConfigFile('site_properties');
    if (sitePropertiesForCurrentStack) {
      return sitePropertiesForCurrentStack.configProperties;
    }

    if (App.get('isHadoop22Stack')) {
      return require('data/HDP2.2/site_properties').configProperties;
    }
    return require('data/HDP2/site_properties').configProperties;
  }.property('App.isHadoop22Stack', 'App.currentStackName'),

  preDefinedCustomConfigs: require('data/HDP2/custom_configs'),

  preDefinedConfigFile: function(file) {
    try {
      return require('data/{0}/{1}'.format(App.get('currentStackName'), file));
    } catch(err) {
      // the file doesn't exist, which might be expected.
    }
  },
  //categories which contain custom configs
  categoriesWithCustom: ['CapacityScheduler'],


  /**
   * Create array of service properties for Log4j files
   * @returns {Array}
   */
  createContentProperties: function (configs) {
    var services = App.StackService.find();
    var contentProperties = [];
    if (configs) {
      services.forEach(function (service) {
        if (service.get('configTypes')) {
          Object.keys(service.get('configTypes')).forEach(function (type) {
            var contentProperty = configs.filterProperty('filename', type + '.xml').someProperty('name', 'content');
            if (contentProperty && (type.endsWith('-log4j') || type.endsWith('-env'))) {
              var property = {
                "id": "site property",
                "name": "content",
                "displayName": type.endsWith('-env') ? type + ' template' : "content",
                "value": "",
                "defaultValue": "",
                "description": type + " properties",
                "displayType": "content",
                "isOverridable": true,
                "isRequired": false,
                "isVisible": true,
                "showLabel": type.endsWith('-env'),
                "serviceName": service.get('serviceName'),
                "filename": type + '.xml',
                "category": "Advanced " + type
              };
              contentProperties.pushObject(property);
            }
          }, this);
        }
      }, this);
    }
    return contentProperties;
  },

  //configs with these filenames go to appropriate category not in Advanced
  customFileNames: ['flume-conf.xml'],

  /**
   * Function should be used post-install as precondition check should not be done only after installer wizard
   * @param siteNames {String|Array}
   * @returns {Array}
   */
  getBySitename: function (siteNames) {
    var computedConfigs = this.get('configMapping').computed();
    var siteProperties = [];
    if (typeof siteNames === "string") {
      siteProperties = computedConfigs.filterProperty('filename', siteNames);
    } else if (siteNames instanceof Array) {
      siteNames.forEach(function (_siteName) {
        siteProperties = siteProperties.concat(computedConfigs.filterProperty('filename', _siteName));
      }, this);
    }
    return siteProperties;
  },

  /**
   * Cache of loaded configurations. This is useful in not loading
   * same configuration multiple times. It is populated in multiple
   * places.
   *
   * Example:
   * {
   *  'hdfs-site_version3': {...},
   * }
   */
  loadedConfigurationsCache: {},

  /**
   * identify category by filename of config
   * @param config
   * @return {object|null}
   */
  identifyCategory: function (config) {
    var category = null,
      serviceConfigMetaData = this.get('preDefinedServiceConfigs').findProperty('serviceName', config.serviceName),
      configCategories = (serviceConfigMetaData && serviceConfigMetaData.get('configCategories')) || [];

    if (config.filename.contains("env")) {
      if (config.category) {
        category = configCategories.findProperty("name", config.category);
      } else {
        configCategories.forEach(function (_category) {
          if (_category.name.contains(this.getConfigTagFromFileName(config.filename))) {
            category = _category;
          }
        }, this);
      }
    } else {
      configCategories.forEach(function (_category) {
        if (_category.siteFileNames && Array.isArray(_category.siteFileNames) && _category.siteFileNames.contains(config.filename)) {
          category = _category;
        }
      });
      category = Em.isNone(category) ? configCategories.findProperty('siteFileName', this.getOriginalFileName(config.filename)) : category;
    }
    return category;
  },
  /**
   * additional handling for special properties such as
   * checkbox and digital which values with 'm' at the end
   * @param config
   */
  handleSpecialProperties: function (config) {
    if (config.displayType === 'int' && /\d+m$/.test(config.value)) {
      config.value = config.value.slice(0, config.value.length - 1);
      config.defaultValue = config.value;
    }
  },
  /**
   * calculate config properties:
   * category, filename, isUserProperty, description
   * @param config
   * @param isAdvanced
   * @param advancedConfigs
   */
  calculateConfigProperties: function (config, isAdvanced, advancedConfigs) {
    if (!isAdvanced || this.get('customFileNames').contains(config.filename)) {
      var categoryMetaData = this.identifyCategory(config);
      if (categoryMetaData != null) {
        config.category = categoryMetaData.get('name');
        if (!isAdvanced) config.isUserProperty = true;
      }
      if (isAdvanced) {
        var advancedProperty = advancedConfigs.filterProperty('filename', config.filename).findProperty('name', config.name);
        if (advancedProperty) {
          config.description = advancedProperty.description;
        }
      }
    } else {
      var advancedProperty = null;
      var configType = this.getConfigTagFromFileName(config.filename);
      if (isAdvanced) {
        advancedProperty = advancedConfigs.filterProperty('filename', config.filename).findProperty('name', config.name);
      }
      config.category = config.category ? config.category : 'Advanced ' + configType;
      if (advancedProperty) {
        config.description = advancedProperty.description;
      }
    }
  },

  capacitySchedulerFilter: function () {
    var yarnRegex = /^yarn\.scheduler\.capacity\.root\.([a-z]([\_\-a-z0-9]{0,50}))\.(acl_administer_jobs|acl_submit_jobs|state|user-limit-factor|maximum-capacity|capacity)$/i;
    return function (_config) {
      return (yarnRegex.test(_config.name));
    }
  }.property(),
  /**
   * return:
   *   configs,
   *   mappingConfigs
   *
   * @param configCategories
   * @param advancedConfigs
   * @param tags
   * @param serviceName
   * @return {object}
   */
  mergePreDefinedWithLoaded: function (configCategories, advancedConfigs, tags, serviceName) {
    var configs = [];
    var contentProperties = this.createContentProperties(advancedConfigs);
    var preDefinedConfigs = this.get('preDefinedSiteProperties').concat(contentProperties);
    var mappingConfigs = [];
    var filenameExceptions = this.get('filenameExceptions');
    var selectedServiceNames = App.Service.find().mapProperty('serviceName');
    tags.forEach(function (_tag) {
      var isAdvanced = null;
      var filename = (filenameExceptions.contains(_tag.siteName)) ? _tag.siteName : _tag.siteName + ".xml";
      var siteConfig = configCategories.filter(function (serviceConfigProperties) {
        return _tag.tagName === serviceConfigProperties.tag && _tag.siteName === serviceConfigProperties.type;
      });
      siteConfig = siteConfig[0] || {};

      var attributes = siteConfig['properties_attributes'] || {};
      var finalAttributes = attributes.final || {};
      var properties = siteConfig.properties || {};
      for (var index in properties) {
        var configsPropertyDef = preDefinedConfigs.filterProperty('name', index).findProperty('filename', filename);
        if (!configsPropertyDef) {
          advancedConfigs.filterProperty('name', index).forEach(function (_advancedConfig) {
            var isServiceInstalled = selectedServiceNames.contains(_advancedConfig.serviceName);
            if (isServiceInstalled || _advancedConfig.serviceName == 'MISC') {
              configsPropertyDef = _advancedConfig;
            }
          }, this);
        }

        var serviceConfigObj = App.ServiceConfig.create({
          name: index,
          value: properties[index],
          defaultValue: properties[index],
          filename: filename,
          isUserProperty: false,
          isOverridable: true,
          isReconfigurable: true,
          isRequired: advancedConfigs.someProperty('name', index),
          isFinal: finalAttributes[index] === "true",
          defaultIsFinal: finalAttributes[index] === "true",
          showLabel: true,
          serviceName: serviceName,
          belongsToService: []
        });

        if (configsPropertyDef) {
          this.setServiceConfigUiAttributes(serviceConfigObj, configsPropertyDef);
          // check if defined UI config present in config list obtained from server.
          // in case when config is absent on server and defined UI config is required
          // by server, this config should be ignored
          var serverProperty = properties[serviceConfigObj.get('name')];
          if (Em.isNone(serverProperty) && serviceConfigObj.get('isRequiredByAgent')) {
            continue;
          }
        }

        this.tweakConfigVisibility(serviceConfigObj, properties);
        if (!this.getBySitename(serviceConfigObj.get('filename')).someProperty('name', index)) {
          if (configsPropertyDef) {
            if (configsPropertyDef.isRequiredByAgent === false) {
              continue;
            }
            this.handleSpecialProperties(serviceConfigObj);
          } else {
            serviceConfigObj.displayType = stringUtils.isSingleLine(serviceConfigObj.value) ? 'advanced' : 'multiLine';
          }

          isAdvanced = advancedConfigs.filterProperty('name', index).someProperty('filename', filename);
          serviceConfigObj.id = 'site property';
          serviceConfigObj.displayName = configsPropertyDef && configsPropertyDef.displayName ? configsPropertyDef.displayName : index;
          serviceConfigObj.options = configsPropertyDef ? configsPropertyDef.options : null;
          serviceConfigObj.radioName = configsPropertyDef ? configsPropertyDef.radioName : null;
          serviceConfigObj.serviceName = configsPropertyDef && configsPropertyDef.serviceName ? configsPropertyDef.serviceName : serviceName;
          serviceConfigObj.belongsToService = configsPropertyDef && configsPropertyDef.belongsToService ? configsPropertyDef.belongsToService : [];
          this.calculateConfigProperties(serviceConfigObj, isAdvanced, advancedConfigs);
          this.setValueByDisplayType(serviceConfigObj);
          configs.push(serviceConfigObj);
        } else {
          mappingConfigs.push(serviceConfigObj);
        }
      }
    }, this);
    return {
      configs: configs,
      mappingConfigs: mappingConfigs
    }
  },

  tweakConfigVisibility: function (config, allSiteConfigs) {
    var kdcType = allSiteConfigs['kdc_type'];
    if (kdcType === 'active-directory' && (config.name === 'container_dn' || config.name === 'ldap_url')) {
      config.isVisible = true;
    }
  },

  setValueByDisplayType: function(serviceConfigObj) {
    if (serviceConfigObj.get('displayType') == 'directories' && (serviceConfigObj.get('category') == 'DataNode' || serviceConfigObj.get('category') == 'NameNode')) {
      var dirs = serviceConfigObj.get('value').split(',').sort();
      serviceConfigObj.set('value', dirs.join(','));
      serviceConfigObj.set('defaultValue', dirs.join(','));
    }

    if (serviceConfigObj.get('displayType') == 'directory' && serviceConfigObj.get('category') == 'SNameNode') {
      var dirs = serviceConfigObj.get('value').split(',').sort();
      serviceConfigObj.set('value', dirs[0]);
      serviceConfigObj.set('defaultValue', dirs[0]);
    }

    if (serviceConfigObj.get('displayType') == 'masterHosts') {
      if (typeof(serviceConfigObj.get('value')) == 'string') {
        var value = serviceConfigObj.get('value').replace(/\[|]|'|&apos;/g, "").split(',');
        serviceConfigObj.set('value', value);
      }
    }
  },

  /**
   * @param serviceConfigObj : Object
   * @param configsPropertyDef : Object
   */
  setServiceConfigUiAttributes: function (serviceConfigObj, configsPropertyDef) {
    serviceConfigObj.displayType = configsPropertyDef.displayType;
    serviceConfigObj.isRequired = (configsPropertyDef.isRequired !== undefined) ? configsPropertyDef.isRequired : true;
    serviceConfigObj.isRequiredByAgent = (configsPropertyDef.isRequiredByAgent !== undefined) ? configsPropertyDef.isRequiredByAgent : true;
    serviceConfigObj.isReconfigurable = (configsPropertyDef.isReconfigurable !== undefined) ? configsPropertyDef.isReconfigurable : true;
    serviceConfigObj.isVisible = (configsPropertyDef.isVisible !== undefined) ? configsPropertyDef.isVisible : true;
    serviceConfigObj.unit = (configsPropertyDef.unit !== undefined) ? configsPropertyDef.unit : undefined;
    serviceConfigObj.description = (configsPropertyDef.description !== undefined) ? configsPropertyDef.description : undefined;
    serviceConfigObj.isOverridable = configsPropertyDef.isOverridable === undefined ? true : configsPropertyDef.isOverridable;
    serviceConfigObj.serviceName = configsPropertyDef ? configsPropertyDef.serviceName : null;
    serviceConfigObj.index = configsPropertyDef.index;
    serviceConfigObj.isSecureConfig = configsPropertyDef.isSecureConfig === undefined ? false : configsPropertyDef.isSecureConfig;
    serviceConfigObj.belongsToService = configsPropertyDef.belongsToService;
    serviceConfigObj.category = configsPropertyDef.category;
    serviceConfigObj.showLabel = configsPropertyDef.showLabel !== false;
    serviceConfigObj.dependentConfigPattern = configsPropertyDef.dependentConfigPattern
  },

  /**
   * synchronize order of config properties with order, that on UI side
   *
   * @method syncOrderWithPredefined
   * @param configSet {object}
   * @return {Object}
   */
  syncOrderWithPredefined: function (configSet) {
    var siteConfigs = configSet.configs,
      siteStart = [];

    var preDefinedSiteProperties = this.get('preDefinedSiteProperties').mapProperty('name');
    var contentProperties = this.createContentProperties(siteConfigs).mapProperty('name');
    var siteProperties = preDefinedSiteProperties.concat(contentProperties);
    siteProperties.forEach(function (name) {
      var _site = siteConfigs.filterProperty('name', name);
      if (_site.length == 1) {
        siteStart.push(_site[0]);
        siteConfigs = siteConfigs.without(_site[0]);
      } else if (_site.length > 1) {
        _site.forEach(function (site) {
          siteStart.push(site);
          siteConfigs = siteConfigs.without(site);
        }, this);
      }
    }, this);

    return {
      configs: siteStart.concat(siteConfigs.sortProperty('name')),
      mappingConfigs: configSet.mappingConfigs
    }
  },

  /**
   * merge stored configs with pre-defined
   * @param storedConfigs
   * @param advancedConfigs
   * @param selectedServiceNames
   * @return {Array}
   */
  mergePreDefinedWithStored: function (storedConfigs, advancedConfigs, selectedServiceNames) {
    var mergedConfigs = [];
    var contentProperties = advancedConfigs ? this.createContentProperties(advancedConfigs) : [];
    var preDefinedConfigs = this.get('preDefinedSiteProperties').concat(contentProperties);

    storedConfigs = (storedConfigs) ? storedConfigs : [];

    var preDefinedNames = preDefinedConfigs.mapProperty('name');
    var storedNames = storedConfigs.mapProperty('name');
    var names = preDefinedNames.concat(storedNames).uniq();
    names.forEach(function (name) {
      var storedCfgs = storedConfigs.filterProperty('name', name);
      var preDefinedCfgs = [];
      var preDefinedConfig = preDefinedConfigs.filterProperty('name', name);
      preDefinedConfig.forEach(function (_preDefinedConfig) {
        if (selectedServiceNames.contains(_preDefinedConfig.serviceName) || _preDefinedConfig.serviceName === 'MISC') {
          preDefinedCfgs.push(_preDefinedConfig);
        }
      }, this);

      var configData = {};
      var isAdvanced = advancedConfigs && advancedConfigs.someProperty('name', name);
      if (storedCfgs.length <= 1 && preDefinedCfgs.length <= 1) {
        var stored = storedCfgs[0];
        var preDefined = preDefinedCfgs[0];

        if (preDefined && stored) {
          configData = preDefined;
          configData.value = stored.value;
          configData.defaultValue = stored.defaultValue;
          configData.overrides = stored.overrides;
          configData.filename = stored.filename;
          configData.description = stored.description;
          configData.isVisible = stored.isVisible;
          configData.isFinal = stored.isFinal;
          configData.supportsFinal = stored.supportsFinal;
          configData.isRequired = (configData.isRequired !== undefined) ? configData.isRequired : true;
          configData.isRequiredByAgent = (configData.isRequiredByAgent !== undefined) ? configData.isRequiredByAgent : true;
          configData.showLabel = !!stored.showLabel;
        }
        else if (!preDefined && stored) {
          configData = this.addUserProperty(stored, isAdvanced, advancedConfigs);
        }
        else if (preDefined && !stored) {
          configData = preDefined;
          configData.isRequiredByAgent = (configData.isRequiredByAgent !== undefined) ? configData.isRequiredByAgent : true;
          if (isAdvanced) {
            var advanced = advancedConfigs.filterProperty('filename', configData.filename).findProperty('name', configData.name);
            this.setPropertyFromStack(configData, advanced);
          }
        }

        mergedConfigs.push(configData);
      } else {
        preDefinedCfgs.forEach(function (cfg) {
          configData = cfg;
          configData.isRequiredByAgent = (configData.isRequiredByAgent !== undefined) ? configData.isRequiredByAgent : true;
          var storedCfg = storedCfgs.findProperty('filename', cfg.filename);
          if (storedCfg) {
            configData.value = storedCfg.value;
            configData.defaultValue = storedCfg.defaultValue;
            configData.overrides = storedCfg.overrides;
            configData.filename = storedCfg.filename;
            configData.description = storedCfg.description;
            configData.isFinal = storedCfg.isFinal;
            configData.supportsFinal = storedCfg.supportsFinal;
            configData.showLabel = !!storedCfg.showLabel;
          } else if (isAdvanced) {
            advanced = advancedConfigs.filterProperty('filename', configData.filename).findProperty('name', configData.name);
            this.setPropertyFromStack(configData, advanced);
          }
          mergedConfigs.push(configData);
        }, this);
      }
    }, this);
    return mergedConfigs;
  },

  /**
   *
   * @param configData {Object} Configs that will be binded to the view on step-7 of installer wizard
   * @param advanced {Object} Config property loaded from Server side stack definition
   */
  setPropertyFromStack: function (configData, advanced) {

    // Password fields should be made blank by default in installer wizard
    // irrespective of whatever value is sent from stack definition.
    // This forces the user to fill the password field.
    if (configData.displayType == 'password') {
      configData.value = '';
    } else {
      configData.value = advanced ? advanced.value : configData.value;
    }
    configData.defaultValue = configData.value;
    configData.filename = advanced ? advanced.filename : configData.filename;
    configData.description = advanced ? advanced.description : configData.description;
    configData.isFinal = !!(advanced && (advanced.isFinal === "true"));
    configData.supportsFinal = !!(advanced && advanced.supportsFinal);
  },


  /**
   * look over advanced configs and add missing configs to serviceConfigs
   * filter fetched configs by service if passed
   * @param serviceConfigs
   * @param advancedConfigs
   * @param serviceName
   */
  addAdvancedConfigs: function (serviceConfigs, advancedConfigs, serviceName) {
    var miscConfigs = serviceConfigs.filterProperty('serviceName', 'MISC');
    var configsToVerifying = (serviceName) ? serviceConfigs.filterProperty('serviceName', serviceName).concat(miscConfigs) : serviceConfigs.slice();
    var definedService = this.get('preDefinedServiceConfigs').findProperty('serviceName', serviceName);
    if (definedService) {
      var definedConfigs = (serviceName) ? definedService.get('configs') : [];

      if (definedConfigs.length) {
        advancedConfigs = advancedConfigs.filter(function(property) {
          return !(definedConfigs.someProperty('name', property.name) && !serviceConfigs.someProperty('name', property.name));
        }, this);
      }
    }
    if (advancedConfigs) {
      advancedConfigs.forEach(function (_config) {
        var configType = this.getConfigTagFromFileName(_config.filename);
        var configCategory = _config.category || 'Advanced ' + configType;
        var categoryMetaData = null;
        if (_config) {
          if (!(this.get('configMapping').computed().someProperty('name', _config.name) ||
            configsToVerifying.filterProperty('name', _config.name).someProperty('filename', _config.filename))) {
            if (this.get('customFileNames').contains(_config.filename)) {
              categoryMetaData = this.identifyCategory(_config);
              if (categoryMetaData != null) {
                configCategory = categoryMetaData.get('name');
              }
            }
            _config.id = "site property";
            _config.category = configCategory;
            _config.displayName = _config.displayName || _config.name;
            _config.defaultValue = _config.value;
            // make all advanced configs optional and populated by default
            /*
             * if (/\${.*}/.test(_config.value) || (service.serviceName !==
             * 'OOZIE' && service.serviceName !== 'HBASE')) { _config.isRequired =
             * false; _config.value = ''; } else if
             * (/^\s+$/.test(_config.value)) { _config.isRequired = false; }
             */
            _config.isRequired = true;
            _config.displayType = _config.displayType ? _config.displayType : stringUtils.isSingleLine(_config.value) ? 'advanced' : 'multiLine';
            serviceConfigs.push(_config);
          }
        }
      }, this);
    }
  },
  /**
   * Render a custom conf-site box for entering properties that will be written in *-site.xml files of the services
   */
  addCustomConfigs: function (configs) {
    var preDefinedCustomConfigs = $.extend(true, [], this.get('preDefinedCustomConfigs'));
    var stored = configs.filter(function (_config) {
      return this.get('categoriesWithCustom').contains(_config.category);
    }, this);
  },

  miscConfigVisibleProperty: function (configs, serviceToShow) {
    configs.forEach(function (item) {
      if (item.get('isVisible') && item.belongsToService && item.belongsToService.length) {
        if (item.get('belongsToService').contains('Cluster') && item.get('displayType') == 'user') {
          item.set('isVisible', true);
          return;
        }
        item.set("isVisible", item.belongsToService.some(function (cur) {
          return serviceToShow.contains(cur)
        }));
      }
    });
    return configs;
  },

  /**
   * render configs, distribute them by service
   * and wrap each in ServiceConfigProperty object
   * @param configs
   * @param storedConfigs
   * @param allSelectedServiceNames
   * @param installedServiceNames
   * @param localDB
   * @return {Array}
   */
  renderConfigs: function (configs, storedConfigs, allSelectedServiceNames, installedServiceNames, localDB, recommended) {
    var renderedServiceConfigs = [];
    var services = [];

    this.get('preDefinedServiceConfigs').forEach(function (serviceConfig) {
      var serviceName = serviceConfig.get('serviceName');
      if (allSelectedServiceNames.contains(serviceName) || serviceName === 'MISC') {
        console.log('pushing ' + serviceName, serviceConfig);
        if (!installedServiceNames.contains(serviceName) || serviceName === 'MISC') {
          serviceConfig.set('showConfig', true);
        }
        services.push(serviceConfig);
      }
    });
    services.forEach(function (service) {
      var configsByService = [];
      var serviceConfigs = configs.filterProperty('serviceName', service.get('serviceName'));
      serviceConfigs.forEach(function (_config) {
        _config.isOverridable = (_config.isOverridable === undefined) ? true : _config.isOverridable;
        var serviceConfigProperty = App.ServiceConfigProperty.create(_config);
        this.updateHostOverrides(serviceConfigProperty, _config);
        if (!storedConfigs && !serviceConfigProperty.get('hasInitialValue')) {
          var hiveMetastoreUrisConfig = configs.filterProperty('filename', 'hive-site.xml').findProperty('name', 'hive.metastore.uris');
          var clientPortConfig = configs.filterProperty('filename', 'zoo.cfg.xml').findProperty('name', 'clientPort');
          var dependencies = {
            'hive.metastore.uris': hiveMetastoreUrisConfig && hiveMetastoreUrisConfig.defaultValue,
            'clientPort': clientPortConfig && clientPortConfig.defaultValue
          };
          serviceConfigProperty.initialValue(localDB, dependencies);
        }
        if (storedConfigs && storedConfigs.filterProperty('name', _config.name).length && !!_config.filename) {
          var storedConfig = storedConfigs.filterProperty('name', _config.name).findProperty('filename', _config.filename);
          if (storedConfig) {
            serviceConfigProperty.set('defaultValue', storedConfig.defaultValue);
            serviceConfigProperty.set('value', storedConfig.value);
          }
        }
        this.tweakDynamicDefaults(localDB, serviceConfigProperty, _config);
        serviceConfigProperty.validate();
        configsByService.pushObject(serviceConfigProperty);
      }, this);
      var serviceConfig = this.createServiceConfig(service.get('serviceName'));
      serviceConfig.set('showConfig', service.get('showConfig'));

      // Use calculated default values for some configs
      var recommendedDefaults = {};
      if (!storedConfigs && service.get('configTypes')) {
        Object.keys(service.get('configTypes')).forEach(function (type) {
          if (!recommended || !recommended[type]) {
            return;
          }
          var defaults = recommended[type].properties;
          for (var name in defaults) {
            var config = configsByService.findProperty('name', name);
            if (!config) {
              continue;
            }
            recommendedDefaults[name] = defaults[name];
            config.set('value', defaults[name]);
            config.set('defaultValue', defaults[name]);
          }
        });
      }
      serviceConfig.set('configs', configsByService);
      renderedServiceConfigs.push(serviceConfig);
    }, this);
    return renderedServiceConfigs;
  },
  /**
   Takes care of the "dynamic defaults" for the GLUSTERFS configs.  Sets
   some of the config defaults to previously user-entered data.
   **/
  tweakDynamicDefaults: function (localDB, serviceConfigProperty, config) {
    var firstHost;
    for (var host in localDB.hosts) {
      firstHost = host;
      break;
    }
    try {
      if (typeof(config.defaultValue) == "string" && config.defaultValue.indexOf("{firstHost}") >= 0) {
        serviceConfigProperty.set('value', serviceConfigProperty.value.replace(new RegExp("{firstHost}"), firstHost));
        serviceConfigProperty.set('defaultValue', serviceConfigProperty.defaultValue.replace(new RegExp("{firstHost}"), firstHost));
      }
    } catch (err) {
      // Nothing to worry about here, most likely trying indexOf on a non-string
    }
  },

  /**
   * Mark descriptor properties in configuration object.
   *
   * @param {Object[]} configs - config properties to change
   * @param {App.ServiceConfigProperty[]} descriptor - parsed kerberos descriptor
   */
  addKerberosDescriptorConfigs: function(configs, descriptor) {
    descriptor.forEach(function(item) {
      var property = configs.findProperty('name', item.get('name'));
      if (property) {
        Em.set(property, 'isSecureConfig', true);
        Em.set(property, 'displayName', Em.get(item, 'name'));
        Em.set(property, 'isUserProperty', false);
        Em.set(property, 'isOverridable', false);
        Em.set(property, 'category', 'Advanced ' + Em.get(item, 'filename'));
      }
    });
  },
  /**
   * create new child configs from overrides, attach them to parent config
   * override - value of config, related to particular host(s)
   * @param configProperty
   * @param storedConfigProperty
   */
  updateHostOverrides: function (configProperty, storedConfigProperty) {
    if (storedConfigProperty.overrides != null && storedConfigProperty.overrides.length > 0) {
      var overrides = [];
      storedConfigProperty.overrides.forEach(function (overrideEntry) {
        // create new override with new value
        var newSCP = App.ServiceConfigProperty.create(configProperty);
        newSCP.set('value', overrideEntry.value);
        newSCP.set('isOriginalSCP', false); // indicated this is overridden value,
        newSCP.set('parentSCP', configProperty);
        var hostsArray = Ember.A([]);
        overrideEntry.hosts.forEach(function (host) {
          hostsArray.push(host);
        });
        newSCP.set('selectedHostOptions', hostsArray);
        overrides.pushObject(newSCP);
      });
      configProperty.set('overrides', overrides);
    }
  },
  /**
   * create new ServiceConfig object by service name
   * @param serviceName
   */
  createServiceConfig: function (serviceName) {
    var preDefinedServiceConfig = App.config.get('preDefinedServiceConfigs').findProperty('serviceName', serviceName);
    var serviceConfig = App.ServiceConfig.create({
      serviceName: preDefinedServiceConfig.get('serviceName'),
      displayName: preDefinedServiceConfig.get('displayName'),
      configCategories: preDefinedServiceConfig.get('configCategories'),
      configs: [],
      configGroups: []
    });
    return serviceConfig;
  },
  /**
   * GETs all cluster level sites in one call.
   *
   * @return {object}
   */
  loadConfigsByTags: function (tags) {
    var urlParams = [];
    tags.forEach(function (_tag) {
      urlParams.push('(type=' + _tag.siteName + '&tag=' + _tag.tagName + ')');
    });
    var params = urlParams.join('|');
    return App.ajax.send({
      name: 'config.on_site',
      sender: this,
      data: {
        params: params
      }
    });
  },

  /**
   * Fetch cluster configs from server
   *
   * @param callback
   * @return {object|null}
   */
  loadClusterConfig: function (callback) {
    return App.ajax.send({
      name: 'config.cluster',
      sender: this,
      data: {
        stackVersionUrl: App.get('stackVersionURL'),
        callback: callback
      },
      success: 'loadClusterConfigSuccess',
      error: 'loadClusterConfigError'
    });
  },

  loadClusterConfigSuccess: function (data, opt, params) {
    console.log("TRACE: In success function for the loadClusterConfigSuccess; url is ", opt.url);
    var properties = [];
    if (data.items.length) {
      data.items.forEach(function (item) {
        item.StackLevelConfigurations.property_type = item.StackLevelConfigurations.property_type || [];
        item.StackLevelConfigurations.service_name = 'MISC';
        var property = this.createAdvancedPropertyObject(item.StackLevelConfigurations);
        if (property) properties.push(property);
      }, this);
    }
    params.callback(properties);
  },

  loadClusterConfigError: function (request, ajaxOptions, error, opt, params) {
    console.log('ERROR: Failed to load cluster-env configs');
    params.callback([]);
  },


  /**
   * Generate serviceProperties save it to localDB
   * called from stepController step6WizardController
   *
   * @method loadAdvancedConfig
   * @param serviceName
   * @param callback
   * @return {object|null}
   */
  loadAdvancedConfig: function (serviceName, callback) {
    return App.ajax.send({
      name: 'config.advanced',
      sender: this,
      data: {
        serviceName: serviceName,
        stackVersionUrl: App.get('stackVersionURL'),
        stackVersion: App.get('currentStackVersionNumber'),
        callback: callback
      },
      success: 'loadAdvancedConfigSuccess',
      error: 'loadAdvancedConfigError'
    });
  },

  /**
   * Load advanced configs by service names etc.
   * Use this method when you need to get configs for
   * particular services by single request
   *
   * @method loadAdvancedConfigPartial
   * @param {String[]} serviceNames
   * @param {Object} opt
   * @param {Function} callback
   * @returns {jqXHR}
   */
  loadAdvancedConfigPartial: function(serviceNames, opt, callback) {
    var data = {
      serviceList: serviceNames.join(','),
      stackVersionUrl: App.get('stackVersionURL'),
      stackVersion: App.get('currentStackVersionNumber'),
      queryFilter: ('&' + opt.queryFilter) || '',
      callback: callback
    };
    return App.ajax.send({
      name: 'config.advanced.partial',
      sender: this,
      data: data,
      success: 'loadAdvancedConfigPartialSuccess',
      error: 'loadAdvancedConfigError'
    });
  },

  loadAdvancedConfigSuccess: function (data, opt, params, request) {
    console.log("TRACE: In success function for the loadAdvancedConfig; url is ", opt.url);
    var properties = [];
    if (data.items.length) {
      data.items.forEach(function (item) {
        var property = this.createAdvancedPropertyObject(item.StackConfigurations);
        if (property) properties.push(property);
      }, this);
    }
    params.callback(properties, request);
  },

  loadAdvancedConfigError: function (request, ajaxOptions, error, opt, params) {
    console.log('ERROR: failed to load stack configs for', params.serviceName);
    params.callback([], request);
  },

  loadAdvancedConfigPartialSuccess: function(data, opt, params, request) {
    var properties = [];
    if(data.items.length && data.items.mapProperty('configurations').length) {
      var configurations = data.items.mapProperty('configurations').reduce(function(p,c) { return p.concat(c); });
      configurations.forEach(function(item) {
        var property = this.createAdvancedPropertyObject(item.StackConfigurations);
        if (property) properties.push(property);
      }, this);
    }
    params.callback(properties, request);
  },

  /**
   * Bootstrap configuration property object according to
   * format that we using in our application.
   *
   * @method createAdvancedPropertyObject
   * @param {Object} item
   * @returns {Object|Boolean}
   */
  createAdvancedPropertyObject: function(item) {
    var serviceName = item.service_name;
    var fileName = item.type;
    /**
     * Properties from mapred-queue-acls.xml are ignored
     * Properties from capacity-scheduler.xml are ignored unless HDP stack version is 2.x or
     * HDP stack version is 1.x
     */
    if (fileName == 'mapred-queue-acls.xml') return false;
    item.isVisible = fileName != 'cluster-env.xml';
    var property = {
      serviceName: serviceName,
      name: item.property_name,
      value: item.property_value,
      description: item.property_description,
      isVisible: item.isVisible,
      isFinal: item.final === "true",
      defaultIsFinal: item.final === "true",
      filename: item.filename || fileName
    };

    return $.extend(property, this.advancedConfigIdentityData(item));
  },

  /**
   * Add additional properties to advanced property config object.
   * Additional logic based on `property_type`.
   *
   * @method advancedConfigIdentityData
   * @param {Object} config
   * @return {Object}
   */
  advancedConfigIdentityData: function(config) {
    var propertyData = {};
    var proxyUserGroupServices = ['HIVE', 'OOZIE', 'FALCON'];
    var nameToDisplayNameMap = {
      'smokeuser': 'Smoke Test User',
      'user_group': 'Hadoop Group',
      'mapred_user': 'MapReduce User',
      'zk_user': 'ZooKeeper User',
      'ignore_groupsusers_create': 'Skip group modifications during install'
    };
    if (config.property_type.contains('USER') || config.property_type.contains('ADDITIONAL_USER_PROPERTY') || config.property_type.contains('GROUP')) {
      propertyData.id = "puppet var";
      propertyData.category = 'Users and Groups';
      propertyData.isVisible = !App.get('isHadoopWindowsStack');
      propertyData.serviceName = 'MISC';
      propertyData.isOverridable = false;
      propertyData.isReconfigurable = false;
      propertyData.displayName = nameToDisplayNameMap[config.property_name] || App.format.normalizeName(config.property_name);
      propertyData.displayType = config.property_name == 'ignore_groupsusers_create' ? 'checkbox' : 'user';
      if (config.service_name) {
        var propertyIndex = config.service_name == 'MISC' ? 30 : App.StackService.find().mapProperty('serviceName').indexOf(config.service_name);
        propertyData.belongsToService = [config.service_name];
        propertyData.index = propertyIndex;
      } else {
        propertyData.index = 30;
      }
      if (config.property_name == 'proxyuser_group') propertyData.belongsToService = proxyUserGroupServices;
    }

    if (config.property_type.contains('PASSWORD')) {
      propertyData.displayType = "password";
    }

    return propertyData;
  },

  /**
   * Get config types and config type attributes from stack service
   *
   * @param service
   * @return {object}
   */
  getConfigTypesInfoFromService: function (service) {
    var configTypes = service.get('configTypes');
    var configTypesInfo = {
      items: [],
      supportsFinal: []
    };
    if (configTypes) {
      for (var key in configTypes) {
        if (configTypes.hasOwnProperty(key)) {
          configTypesInfo.items.push(key);
          if (configTypes[key].supports && configTypes[key].supports.final === "true") {
            configTypesInfo.supportsFinal.push(key);
          }
        }
      }
    }
    return configTypesInfo;
  },

  /**
   * Get properties from server by type and tag with properties, that belong to group
   * push them to common {serviceConfigs} and call callback function
   */
  loadServiceConfigGroupOverrides: function (serviceConfigs, loadedGroupToOverrideSiteToTagMap, configGroups, callback, sender) {
    var configKeyToConfigMap = {};
    serviceConfigs.forEach(function (item) {
      if (!configKeyToConfigMap[item.filename]) {
        configKeyToConfigMap[item.filename] = {};
      }
      configKeyToConfigMap[item.filename][item.name] = item;
    });
    var typeTagToGroupMap = {};
    var urlParams = [];
    for (var group in loadedGroupToOverrideSiteToTagMap) {
      var overrideTypeTags = loadedGroupToOverrideSiteToTagMap[group];
      for (var type in overrideTypeTags) {
        var tag = overrideTypeTags[type];
        typeTagToGroupMap[type + "///" + tag] = configGroups.findProperty('name', group);
        urlParams.push('(type=' + type + '&tag=' + tag + ')');
      }
    }
    var params = urlParams.join('|');
    if (urlParams.length) {
      App.ajax.send({
        name: 'config.host_overrides',
        sender: this,
        data: {
          params: params,
          configKeyToConfigMap: configKeyToConfigMap,
          typeTagToGroupMap: typeTagToGroupMap,
          callback: callback,
          sender: sender,
          serviceConfigs: serviceConfigs
        },
        success: 'loadServiceConfigGroupOverridesSuccess'
      });
    } else {
      callback.call(sender, serviceConfigs);
    }
  },
  loadServiceConfigGroupOverridesSuccess: function (data, opt, params) {
    data.items.forEach(function (config) {
      App.config.loadedConfigurationsCache[config.type + "_" + config.tag] = config.properties;
      var group = params.typeTagToGroupMap[config.type + "///" + config.tag];
      var properties = config.properties;
      for (var prop in properties) {
        var fileName = this.getOriginalFileName(config.type);
        var serviceConfig = !!params.configKeyToConfigMap[fileName] ? params.configKeyToConfigMap[fileName][prop] : false;
        var hostOverrideValue = this.formatOverrideValue(serviceConfig, properties[prop]);
        var hostOverrideIsFinal = !!(config.properties_attributes && config.properties_attributes.final && config.properties_attributes.final[prop]);
        if (serviceConfig) {
          // Value of this property is different for this host.
          if (!Em.get(serviceConfig, 'overrides')) Em.set(serviceConfig, 'overrides', []);
          console.log("loadServiceConfigGroupOverridesSuccess(): [" + group + "] OVERRODE(" + serviceConfig.name + "): " + serviceConfig.value + " -> " + hostOverrideValue);
          serviceConfig.overrides.pushObject({value: hostOverrideValue, group: group, isFinal: hostOverrideIsFinal});
        } else {
          params.serviceConfigs.push(this.createCustomGroupConfig(prop, config, group));
        }
      }
    }, this);
    params.callback.call(params.sender, params.serviceConfigs);
  },
  /**
   * Create config with non default config group. Some custom config properties
   * can be created and assigned to non-default config group.
   *
   * @param {String} propertyName - name of the property
   * @param {Object} config - config info
   * @param {Em.Object} group - config group to set
   * @return {Object}
   **/
  createCustomGroupConfig: function (propertyName, config, group) {
    var propertyValue = config.properties[propertyName];
    var propertyObject = {
      name: propertyName,
      displayName: propertyName,
      defaultValue: propertyValue,
      value: propertyValue,
      displayType: stringUtils.isSingleLine(propertyValue) ? 'advanced' : 'multiLine',
      isSecureConfig: false,
      group: group,
      id: 'site property',
      serviceName: group.get('service.serviceName'),
      filename: this.getOriginalFileName(config.type),
      isUserProperty: true,
      isVisible: true,
      isOverridable: false
    };
    propertyObject.category = this.identifyCategory(propertyObject).name;
    group.set('switchGroupTextShort', Em.I18n.t('services.service.config_groups.switchGroupTextShort').format(group.get('name')));
    group.set('switchGroupTextFull', Em.I18n.t('services.service.config_groups.switchGroupTextFull').format(group.get('name')));
    return App.ServiceConfigProperty.create(propertyObject);
  },
  /**
   * format value of override of config
   * @param serviceConfig
   * @param hostOverrideValue
   */
  formatOverrideValue: function (serviceConfig, hostOverrideValue) {
    if (serviceConfig && serviceConfig.displayType === 'int') {
      if (/\d+m$/.test(hostOverrideValue)) {
        return hostOverrideValue.slice(0, hostOverrideValue.length - 1);
      }
    }
    return hostOverrideValue;
  },

  /**
   * Set all site property that are derived from other site-properties
   */
  setConfigValue: function (mappedConfigs, allConfigs, config) {
    var globalValue;
    if (config.value == null) {
      return;
    }
    var fkValue = config.value.match(/<(foreignKey.*?)>/g);
    var fkName = config.name.match(/<(foreignKey.*?)>/g);
    var templateValue = config.value.match(/<(templateName.*?)>/g);
    if (fkValue) {
      fkValue.forEach(function (_fkValue) {
        var index = parseInt(_fkValue.match(/\[([\d]*)(?=\])/)[1]);
        if (mappedConfigs.someProperty('name', config.foreignKey[index])) {
          globalValue = mappedConfigs.findProperty('name', config.foreignKey[index]).value;
          config.value = config.value.replace(_fkValue, globalValue);
        } else if (allConfigs.someProperty('name', config.foreignKey[index])) {
          if (allConfigs.findProperty('name', config.foreignKey[index]).value === '') {
            globalValue = allConfigs.findProperty('name', config.foreignKey[index]).defaultValue;
          } else {
            globalValue = allConfigs.findProperty('name', config.foreignKey[index]).value;
          }
          config.value = config.value.replace(_fkValue, globalValue);
        }
      }, this);
    }

    // config._name - formatted name from original config name
    if (fkName) {
      fkName.forEach(function (_fkName) {
        var index = parseInt(_fkName.match(/\[([\d]*)(?=\])/)[1]);
        if (mappedConfigs.someProperty('name', config.foreignKey[index])) {
          globalValue = mappedConfigs.findProperty('name', config.foreignKey[index]).value;
          config._name = config.name.replace(_fkName, globalValue);
        } else if (allConfigs.someProperty('name', config.foreignKey[index])) {
          if (allConfigs.findProperty('name', config.foreignKey[index]).value === '') {
            globalValue = allConfigs.findProperty('name', config.foreignKey[index]).defaultValue;
          } else {
            globalValue = allConfigs.findProperty('name', config.foreignKey[index]).value;
          }
          config._name = config.name.replace(_fkName, globalValue);
        }else{
          config.noMatchSoSkipThisConfig = true;
        }
      }, this);
    }

    //For properties in the configMapping file having foreignKey and templateName properties.
    if (templateValue) {
      templateValue.forEach(function (_value) {
        var index = parseInt(_value.match(/\[([\d]*)(?=\])/)[1]);
        if (allConfigs.someProperty('name', config.templateName[index])) {
          var globalValue = allConfigs.findProperty('name', config.templateName[index]).value;
          config.value = config.value.replace(_value, globalValue);
        } else {
          config.value = null;
        }
      }, this);
    }
  },

  /**
   * identify service name of config by its config's type
   * @param type
   * @return {string|null}
   */
  getServiceNameByConfigType: function (type) {
    var preDefinedServiceConfigs = this.get('preDefinedServiceConfigs');
    var service = preDefinedServiceConfigs.find(function (serviceConfig) {
      return !!serviceConfig.get('configTypes')[type];
    }, this);
    return service && service.get('serviceName');
  },

  /**
   * add user property
   * @param stored
   * @param isAdvanced
   * @param advancedConfigs
   * @return {Object}
   */
  addUserProperty: function (stored, isAdvanced, advancedConfigs) {
    var
      skipAttributeChanges = {
        displayType: ['ignore_groupsusers_create'],
        displayName: ['ignore_groupsusers_create', 'smokeuser', 'user_group', 'mapred_user', 'zk_user']
      },
      configData = {
        id: stored.id,
        name: stored.name,
        displayName: skipAttributeChanges.displayName.contains(stored.name) ?
          this.getOriginalConfigAttribute(stored, 'displayName', advancedConfigs) : App.format.normalizeName(stored.name),
        serviceName: stored.serviceName,
        value: stored.value,
        defaultValue: stored.defaultValue,
        displayType: skipAttributeChanges.displayType.contains(stored.name) ?
          this.getOriginalConfigAttribute(stored, 'displayType', advancedConfigs) :
          (stringUtils.isSingleLine(stored.value) ? 'advanced' : 'multiLine'),
        filename: stored.filename,
        isUserProperty: stored.isUserProperty === true,
        hasInitialValue: !!stored.hasInitialValue,
        isOverridable: true,
        overrides: stored.overrides,
        isRequired: false,
        isVisible: stored.isVisible,
        isFinal: stored.isFinal,
        defaultIsFinal: stored.defaultIsFinal,
        supportsFinal: stored.supportsFinal,
        showLabel: stored.showLabel !== false,
        category: stored.category
      };
    if (stored.category == 'Users and Groups') {
      configData.index = this.getOriginalConfigAttribute(stored, 'index', advancedConfigs);
    }

    App.get('config').calculateConfigProperties(configData, isAdvanced, advancedConfigs);
    return configData;
  },

  getOriginalConfigAttribute: function (stored, key, advancedConfigs) {
    return advancedConfigs.findProperty('name', stored.name) ?
      advancedConfigs.findProperty('name', stored.name)[key] : stored[key];
  },

  complexConfigsTemplate: [
    {
      "id": "site property",
      "name": "capacity-scheduler",
      "displayName": "Capacity Scheduler",
      "value": "",
      "defaultValue": "",
      "description": "Capacity Scheduler properties",
      "displayType": "custom",
      "isOverridable": true,
      "isRequired": true,
      "isVisible": true,
      "isReconfigurable": true,
      "supportsFinal": false,
      "serviceName": "YARN",
      "filename": "capacity-scheduler.xml",
      "category": "CapacityScheduler"
    }
  ],

  /**
   * transform set of configs from file
   * into one config with textarea content:
   * name=value
   * @param configs
   * @param filename
   * @return {*}
   */
  fileConfigsIntoTextarea: function (configs, filename) {
    var fileConfigs = configs.filterProperty('filename', filename);
    var value = '';
    var defaultValue = '';
    var template = this.get('complexConfigsTemplate').findProperty('filename', filename);
    var complexConfig = $.extend({}, template);
    if (complexConfig) {
      fileConfigs.forEach(function (_config) {
        value += _config.name + '=' + _config.value + '\n';
        defaultValue += _config.name + '=' + _config.defaultValue + '\n';
      }, this);
      var isFinal = fileConfigs.someProperty('isFinal', true);
      complexConfig.value = value;
      complexConfig.defaultValue = defaultValue;
      complexConfig.isFinal = isFinal;
      complexConfig.defaultIsFinal = isFinal;
      configs = configs.filter(function (_config) {
        return _config.filename !== filename;
      });
      configs.push(complexConfig);
    }
    return configs;
  },

  /**
   * transform one config with textarea content
   * into set of configs of file
   * @param configs
   * @param filename
   * @return {*}
   */
  textareaIntoFileConfigs: function (configs, filename) {
    var complexConfigName = this.get('complexConfigsTemplate').findProperty('filename', filename).name;
    var configsTextarea = configs.findProperty('name', complexConfigName);
    if (configsTextarea) {
      var properties = configsTextarea.get('value').split('\n');

      properties.forEach(function (_property) {
        var name, value;
        if (_property) {
          _property = _property.split('=');
          name = _property[0];
          value = (_property[1]) ? _property[1] : "";
          configs.push(Em.Object.create({
            id: configsTextarea.get('id'),
            name: name,
            value: value,
            defaultValue: value,
            serviceName: configsTextarea.get('serviceName'),
            filename: filename,
            isFinal: configsTextarea.get('isFinal'),
            isNotDefaultValue: configsTextarea.get('isNotDefaultValue'),
            group: null
          }));
        }
      });
      return configs.without(configsTextarea);
    }
    console.log('ERROR: textarea config - ' + complexConfigName + ' is missing');
    return configs;
  },

  /**
   * trim trailing spaces for all properties.
   * trim both trailing and leading spaces for host displayType and hive/oozie datebases url.
   * for directory or directories displayType format string for further using.
   * for password and values with spaces only do nothing.
   * @param {Object} property
   * @param {Boolean} isEmberObject
   * @returns {*}
   */
  trimProperty: function (property, isEmberObject) {
    var displayType = (isEmberObject) ? property.get('displayType') : property.displayType;
    var value = (isEmberObject) ? property.get('value') : property.value;
    var name = (isEmberObject) ? property.get('name') : property.name;
    var rez;
    switch (displayType) {
      case 'directories':
      case 'directory':
      case 'datanodedirs':
        rez = value.replace(/,/g,' ').trim().split(/\s+/g).join(',');
        break;
      case 'host':
        rez = value.trim();
        break;
      case 'password':
        break;
      case 'advanced':
        if (name == 'javax.jdo.option.ConnectionURL' || name == 'oozie.service.JPAService.jdbc.url') {
          rez = value.trim();
        }
      default:
        rez = (typeof value == 'string') ? value.replace(/(\s+$)/g, '') : value;
    }
    return ((rez == '') || (rez == undefined)) ? value : rez;
  },

  OnNnHAHideSnn: function (ServiceConfig) {
    var configCategories = ServiceConfig.get('configCategories');
    var snCategory = configCategories.findProperty('name', 'SNameNode');
    var isSnnPresent = !!App.HDFSService.find('HDFS').get('snameNode');
    if (snCategory && !isSnnPresent) {
      configCategories.removeObject(snCategory);
    }
  },

  /**
   * Launches a dialog where an existing config-group can be selected, or a new
   * one can be created. This is different than the config-group management
   * dialog where host membership can be managed.
   *
   * The callback will be passed the created/selected config-group in the form
   * of {id:2, name:'New hardware group'}. In the case of dialog being cancelled,
   * the callback is provided <code>null</code>
   *
   * @param {String} groupName
   *  is closed, cancelled or OK is pressed.
   */

  saveGroupConfirmationPopup: function (groupName) {
    App.ModalPopup.show({
      header: Em.I18n.t('config.group.save.confirmation.header'),
      secondary: Em.I18n.t('config.group.save.confirmation.manage.button'),
      groupName: groupName,
      bodyClass: Ember.View.extend({
        templateName: require('templates/common/configs/saveConfigGroup')
      }),
      onSecondary: function () {
        App.router.get('mainServiceInfoConfigsController').manageConfigurationGroups();
        this.hide();
      }
    });
  },

  //Persist config groups created in step7 wizard controller
  persistWizardStep7ConfigGroups: function () {
    var installerController = App.router.get('installerController');
    var step7Controller = App.router.get('wizardStep7Controller');
      installerController.saveServiceConfigGroups(step7Controller, step7Controller.get('content.controllerName') == 'addServiceController');
      App.clusterStatus.setClusterStatus({
        localdb: App.db.data
      });
  },
  /**
   * exclude configs that depends on services which are uninstalled
   * if config doesn't have serviceName or dependent service is installed then
   * config not excluded
   */
  excludeUnsupportedConfigs: function (configs, installedServices) {
    return configs.filter(function (config) {
      return !(config.serviceName && !installedServices.contains(config.serviceName));
    });
  },

  launchConfigGroupSelectionCreationDialog: function (serviceId, configGroups, configProperty, callback, isInstaller) {
    var self = this;
    var availableConfigGroups = configGroups.slice();
    // delete Config Groups, that already have selected property overridden
    var alreadyOverriddenGroups = [];
    if (configProperty.get('overrides')) {
      alreadyOverriddenGroups = configProperty.get('overrides').mapProperty('group.name');
    }
    var result = [];
    availableConfigGroups.forEach(function (group) {
      if (!group.get('isDefault') && (!alreadyOverriddenGroups.length || !alreadyOverriddenGroups.contains(group.name))) {
        result.push(group);
      }
    }, this);
    availableConfigGroups = result;
    var selectedConfigGroup = availableConfigGroups && availableConfigGroups.length > 0 ?
      availableConfigGroups[0] : null;
    var serviceName = App.format.role(serviceId);
    App.ModalPopup.show({
      classNames: [ 'sixty-percent-width-modal' ],
      header: Em.I18n.t('config.group.selection.dialog.title').format(serviceName),
      subTitle: Em.I18n.t('config.group.selection.dialog.subtitle').format(serviceName),
      selectExistingGroupLabel: Em.I18n.t('config.group.selection.dialog.option.select').format(serviceName),
      noGroups: Em.I18n.t('config.group.selection.dialog.no.groups').format(serviceName),
      createNewGroupLabel: Em.I18n.t('config.group.selection.dialog.option.create').format(serviceName),
      createNewGroupDescription: Em.I18n.t('config.group.selection.dialog.option.create.msg').format(serviceName),
      warningMessage: '&nbsp;',
      isWarning: false,
      optionSelectConfigGroup: true,
      optionCreateConfigGroup: function () {
        return !this.get('optionSelectConfigGroup');
      }.property('optionSelectConfigGroup'),
      hasExistedGroups: function () {
        return !!this.get('availableConfigGroups').length;
      }.property('availableConfigGroups'),
      availableConfigGroups: availableConfigGroups,
      selectedConfigGroup: selectedConfigGroup,
      newConfigGroupName: '',
      disablePrimary: function () {
        return !(this.get('optionSelectConfigGroup') || (this.get('newConfigGroupName').trim().length > 0 && !this.get('isWarning')));
      }.property('newConfigGroupName', 'optionSelectConfigGroup', 'warningMessage'),
      onPrimary: function () {
        if (this.get('optionSelectConfigGroup')) {
          var selectedConfigGroup = this.get('selectedConfigGroup');
          this.hide();
          callback(selectedConfigGroup);
        } else {
          var newConfigGroupName = this.get('newConfigGroupName').trim();
          var newConfigGroup = App.ConfigGroup.create({
            id: null,
            name: newConfigGroupName,
            description: Em.I18n.t('config.group.description.default').format(new Date().toDateString()),
            isDefault: false,
            parentConfigGroup: null,
            service: (isInstaller) ? Em.Object.create({id: serviceId}) : App.Service.find().findProperty('serviceName', serviceId),
            hosts: [],
            configSiteTags: [],
            properties: []
          });
          if (!isInstaller) {
            self.postNewConfigurationGroup(newConfigGroup);
          }
          if (newConfigGroup) {
            newConfigGroup.set('parentConfigGroup', configGroups.findProperty('isDefault'));
            configGroups.pushObject(newConfigGroup);
            if (isInstaller) {
              self.persistWizardStep7ConfigGroups();
            } else {
              self.saveGroupConfirmationPopup(newConfigGroupName);
            }
            this.hide();
            callback(newConfigGroup);
          }
        }
      },
      onSecondary: function () {
        this.hide();
        callback(null);
      },
      doSelectConfigGroup: function (event) {
        var configGroup = event.context;
        console.log(configGroup);
        this.set('selectedConfigGroup', configGroup);
      },
      validate: function () {
        var msg = '&nbsp;';
        var isWarning = false;
        var optionSelect = this.get('optionSelectConfigGroup');
        if (!optionSelect) {
          var nn = this.get('newConfigGroupName');
          if (nn && configGroups.mapProperty('name').contains(nn.trim())) {
            msg = Em.I18n.t("config.group.selection.dialog.err.name.exists");
            isWarning = true;
          }
        }
        this.set('warningMessage', msg);
        this.set('isWarning', isWarning);
      }.observes('newConfigGroupName', 'optionSelectConfigGroup'),
      bodyClass: Ember.View.extend({
        templateName: require('templates/common/configs/selectCreateConfigGroup'),
        controllerBinding: 'App.router.mainServiceInfoConfigsController',
        selectConfigGroupRadioButton: Ember.Checkbox.extend({
          tagName: 'input',
          attributeBindings: ['type', 'checked', 'disabled'],
          checked: function () {
            return this.get('parentView.parentView.optionSelectConfigGroup');
          }.property('parentView.parentView.optionSelectConfigGroup'),
          type: 'radio',
          disabled: false,
          click: function () {
            this.set('parentView.parentView.optionSelectConfigGroup', true);
          },
          didInsertElement: function () {
            if (!this.get('parentView.parentView.hasExistedGroups')) {
              this.set('disabled', true);
              this.set('parentView.parentView.optionSelectConfigGroup', false);
            }
          }
        }),
        createConfigGroupRadioButton: Ember.Checkbox.extend({
          tagName: 'input',
          attributeBindings: ['type', 'checked'],
          checked: function () {
            return !this.get('parentView.parentView.optionSelectConfigGroup');
          }.property('parentView.parentView.optionSelectConfigGroup'),
          type: 'radio',
          click: function () {
            this.set('parentView.parentView.optionSelectConfigGroup', false);
          }
        })
      })
    });
  },
  /**
   * launch dialog where can be assigned another group to host
   * @param selectedGroup
   * @param configGroups
   * @param hostName
   * @param callback
   */
  launchSwitchConfigGroupOfHostDialog: function (selectedGroup, configGroups, hostName, callback) {
    var self = this;
    App.ModalPopup.show({
      header: Em.I18n.t('config.group.host.switch.dialog.title'),
      configGroups: configGroups,
      selectedConfigGroup: selectedGroup,
      disablePrimary: function () {
        return !(this.get('selectedConfigGroup.name') !== selectedGroup.get('name'));
      }.property('selectedConfigGroup'),
      onPrimary: function () {
        var newGroup = this.get('selectedConfigGroup');
        if (selectedGroup.get('isDefault')) {
          selectedGroup.set('hosts.length', selectedGroup.get('hosts.length') - 1)
        } else {
          selectedGroup.get('hosts').removeObject(hostName);
        }
        if (!selectedGroup.get('isDefault')) {
          self.updateConfigurationGroup(selectedGroup, function () {
          }, Em.K);
        }

        if (newGroup.get('isDefault')) {
          newGroup.set('hosts.length', newGroup.get('hosts.length') + 1)
        } else {
          newGroup.get('hosts').pushObject(hostName);
        }
        callback(newGroup);
        if (!newGroup.get('isDefault')) {
          self.updateConfigurationGroup(newGroup, function () {
          }, Em.K);
        }
        this.hide();
      },
      bodyClass: Ember.View.extend({
        templateName: require('templates/utils/config_launch_switch_config_group_of_host')
      })
    });
  },
  /**
   * Create a new config-group for a service.
   *
   * @param newConfigGroupData   config group to post to server
   * @param callback    Callback function for Success or Error handling
   * @return  Returns the created config-group
   */
  postNewConfigurationGroup: function (newConfigGroupData, callback) {
    var dataHosts = [];
    newConfigGroupData.get('hosts').forEach(function (_host) {
      dataHosts.push({
        host_name: _host
      });
    }, this);
    var sendData = {
      name: 'config_groups.create',
      data: {
        'group_name': newConfigGroupData.get('name'),
        'service_id': newConfigGroupData.get('service.id'),
        'description': newConfigGroupData.get('description'),
        'hosts': dataHosts
      },
      success: 'successFunction',
      error: 'errorFunction',
      successFunction: function (response) {
        newConfigGroupData.set('id', response.resources[0].ConfigGroup.id);
        if (callback) {
          callback();
        }
      },
      errorFunction: function (xhr, text, errorThrown) {
        if (callback) {
          callback(xhr, text, errorThrown);
        }
        console.error('Error in creating new Config Group');
      }
    };
    sendData.sender = sendData;
    App.ajax.send(sendData);
    return newConfigGroupData;
  },

  /**
   * PUTs the new configuration-group on the server.
   * Changes possible here are the name, description and
   * host memberships of the configuration-group.
   *
   * @param {App.ConfigGroup} configGroup Configuration group to update
   * @param {Function} successCallback
   * @param {Function} errorCallback
   */
  updateConfigurationGroup: function (configGroup, successCallback, errorCallback) {
    var putConfigGroup = {
      ConfigGroup: {
        group_name: configGroup.get('name'),
        description: configGroup.get('description'),
        tag: configGroup.get('service.id'),
        hosts: [],
        desired_configs: []
      }
    };
    configGroup.get('hosts').forEach(function (h) {
      putConfigGroup.ConfigGroup.hosts.push({
        host_name: h
      });
    });
    configGroup.get('configSiteTags').forEach(function (cst) {
      putConfigGroup.ConfigGroup.desired_configs.push({
        type: cst.get('site'),
        tag: cst.get('tag')
      });
    });

    var sendData = {
      name: 'config_groups.update',
      data: {
        id: configGroup.get('id'),
        data: putConfigGroup
      },
      success: 'successFunction',
      error: 'errorFunction',
      successFunction: function () {
        if (successCallback) {
          successCallback();
        }
      },
      errorFunction: function (xhr, text, errorThrown) {
        if (errorCallback) {
          errorCallback(xhr, text, errorThrown);
        }
      }
    };
    sendData.sender = sendData;
    App.ajax.send(sendData);
  },

  clearConfigurationGroupHosts: function (configGroup, successCallback, errorCallback) {
    configGroup = jQuery.extend({}, configGroup);
    configGroup.set('hosts', []);
    this.updateConfigurationGroup(configGroup, successCallback, errorCallback);
  },

  deleteConfigGroup: function (configGroup, successCallback, errorCallback) {
    var sendData = {
      name: 'common.delete.config_group',
      sender: this,
      data: {
        id: configGroup.get('id')
      },
      success: 'successFunction',
      error: 'errorFunction',
      successFunction: function () {
        if (successCallback) {
          successCallback();
        }
      },
      errorFunction: function (xhr, text, errorThrown) {
        if (errorCallback) {
          errorCallback(xhr, text, errorThrown);
        }
      }
    };
    sendData.sender = sendData;
    App.ajax.send(sendData);
  },

  /**
   * Gets all the configuration-groups for the given service.
   *
   * @param serviceId
   *          (string) ID of the service. Ex: HDFS
   */
  getConfigGroupsForService: function (serviceId) {

  },

  /**
   * Gets all the configuration-groups for a host.
   *
   * @param hostName
   *          (string) host name used to register
   */
   getConfigGroupsForHost: function (hostName) {

  },

  /**
   * Generate minimal config property object used in *_properties.js files.
   * Example:
   * <code>
   *   var someProperties = App.config.generateConfigPropertiesByName([
   *    'property_1', 'property_2', 'property_3'], { category: 'General', filename: 'myFileName'});
   *   // someProperties contains Object[]
   *   [
   *    {
   *      name: 'property_1',
   *      displayName: 'property_1',
   *      isVisible: true,
   *      isReconfigurable: true,
   *      category: 'General',
   *      filename: 'myFileName'
   *    },
   *    .......
   *   ]
   * </code>
   * @param {Array} names
   * @param {Object} properties - additional properties which will merge with base object definition
   * @returns {*}
   */
  generateConfigPropertiesByName: function(names, properties) {
    return names.map(function (item) {
      var baseObj = {
        name: item,
        displayName: item,
        isVisible: true,
        isReconfigurable: true
      };
      if (properties) return $.extend(baseObj, properties);
      else return baseObj;
    });
  },

  /**
   * replace some values in config property
   * @param {string} name
   * @param {string} express
   * @param {string} value
   * @param {string} globValue
   * @return {string}
   * @private
   * @method replaceConfigValues
   */
  replaceConfigValues: function (name, express, value, globValue) {
    if (name == 'templeton.hive.properties') {
      globValue = globValue.replace(/,/g, '\\,');
    }
    return value.replace(express, globValue);
  }
});

});

require.register("utils/configs/modification_handlers/hbase", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
require('utils/configs/modification_handlers/modification_handler');

module.exports = App.ServiceConfigModificationHandler.create({
  serviceId : 'HBASE',

  getConfig : function(allConfigs, configName, configFilename, configServiceName) {
    return allConfigs.findProperty("serviceName", configServiceName).get("configs").find(function(config) {
      return configName == config.get('name') && (configFilename == null || configFilename == config.get('filename'));
    });
  },

  updateConfigClasses : function(configClasses, authEnabled, affectedProperties, addOldValue) {
    if (configClasses != null) {
      var xaAuthCoProcessorClass = "com.xasecure.authorization.hbase.XaSecureAuthorizationCoprocessor";
      var nonXAClass = 'org.apache.hadoop.hbase.security.access.AccessController';
      var currentClassesList = configClasses.get('value').trim().length > 0 ? configClasses.get('value').trim().split(',') : [];
      var newClassesList = null, xaClassIndex, nonXaClassIndex;

      if (authEnabled) {
        var nonXaClassIndex = currentClassesList.indexOf(nonXAClass);
        if (nonXaClassIndex > -1) {
          currentClassesList.splice(nonXaClassIndex, 1);
          newClassesList = currentClassesList;
        }
        var xaClassIndex = currentClassesList.indexOf(xaAuthCoProcessorClass);
        if (xaClassIndex < 0) {
          currentClassesList.push(xaAuthCoProcessorClass);
          newClassesList = currentClassesList;
        }
      } else {
        var xaClassIndex = currentClassesList.indexOf(xaAuthCoProcessorClass);
        if (xaClassIndex > -1) {
          currentClassesList.splice(xaClassIndex, 1);
          newClassesList = currentClassesList;
        }
        if (addOldValue) {
          var nonXaClassIndex = currentClassesList.indexOf(nonXAClass);
          if (nonXaClassIndex < 0) {
            currentClassesList.push(nonXAClass);
            newClassesList = currentClassesList;
          }
        }
      }

      if (newClassesList != null) {
        affectedProperties.push({
          serviceName : "HBASE",
          sourceServiceName : "HBASE",
          propertyName : configClasses.get('name'),
          propertyDisplayName : configClasses.get('name'),
          newValue : newClassesList.join(','),
          curValue : configClasses.get('value'),
          changedPropertyName : 'ranger-hbase-plugin-enabled',
          removed : false,
          filename : 'hbase-site.xml'
        });
      }
    }
  },

  getDependentConfigChanges : function(changedConfig, selectedServices, allConfigs, securityEnabled) {
    var affectedProperties = [];
    var newValue = changedConfig.get("value");
    var hbaseAuthEnabledPropertyName = "ranger-hbase-plugin-enabled";
    var affectedPropertyName = changedConfig.get("name");
    if (affectedPropertyName == hbaseAuthEnabledPropertyName) {
      var configAuthEnabled = this.getConfig(allConfigs, 'hbase.security.authorization', 'hbase-site.xml', 'HBASE');
      var configMasterClasses = this.getConfig(allConfigs, 'hbase.coprocessor.master.classes', 'hbase-site.xml', 'HBASE');
      var configRegionClasses = this.getConfig(allConfigs, 'hbase.coprocessor.region.classes', 'hbase-site.xml', 'HBASE');
      var configRpcProtection = this.getConfig(allConfigs, 'hbase.rpc.protection', 'hbase-site.xml', 'HBASE');

      var authEnabled = newValue == "Yes";
      var newAuthEnabledValue = authEnabled ? "true" : "false";
      var newRpcProtectionValue = authEnabled ? "privacy" : "authentication";

      // Add Hive-Ranger configs
      this.updateConfigClasses(configMasterClasses, authEnabled, affectedProperties, configAuthEnabled.get('value') == 'true');
      this.updateConfigClasses(configRegionClasses, authEnabled, affectedProperties, configAuthEnabled.get('value') == 'true');
      if (newRpcProtectionValue !== configRpcProtection.get('value')) {
        affectedProperties.push({
          serviceName : "HBASE",
          sourceServiceName : "HBASE",
          propertyName : 'hbase.rpc.protection',
          propertyDisplayName : 'hbase.rpc.protection',
          newValue : newRpcProtectionValue,
          curValue : configRpcProtection.get('value'),
          changedPropertyName : hbaseAuthEnabledPropertyName,
          removed : false,
          filename : 'hbase-site.xml'
        });
      }
      if (authEnabled && newAuthEnabledValue !== configAuthEnabled.get('value')) {
        affectedProperties.push({
          serviceName : "HBASE",
          sourceServiceName : "HBASE",
          propertyName : 'hbase.security.authorization',
          propertyDisplayName : 'hbase.security.authorization',
          newValue : newAuthEnabledValue,
          curValue : configAuthEnabled.get('value'),
          changedPropertyName : hbaseAuthEnabledPropertyName,
          removed : false,
          filename : 'hbase-site.xml'
        });
      }
    }
    return affectedProperties;
  }
});
});

require.register("utils/configs/modification_handlers/hdfs", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
require('utils/configs/modification_handlers/modification_handler');

module.exports = App.ServiceConfigModificationHandler.create({
  serviceId : 'HDFS',

  getConfig : function(allConfigs, configName, configFilename, configServiceName) {
    return allConfigs.findProperty("serviceName", configServiceName).get("configs").find(function(config) {
      return configName == config.get('name') && (configFilename == null || configFilename == config.get('filename'));
    });
  },

  getDependentConfigChanges : function(changedConfig, selectedServices, allConfigs, securityEnabled) {
    var affectedProperties = [];
    var newValue = changedConfig.get("value");
    var rangerPluginEnabledName = "ranger-hdfs-plugin-enabled";
    var affectedPropertyName = changedConfig.get("name");
    if (affectedPropertyName == rangerPluginEnabledName) {
      var configDfsPermissionsEnabled = this.getConfig(allConfigs, 'dfs.permissions.enabled', 'hdfs-site.xml', 'HDFS');

      var rangerPluginEnabled = newValue == "Yes";
      var newDfsPermissionsEnabled = rangerPluginEnabled ? "true" : "true";

      // Add Hive-Ranger configs
      if (configDfsPermissionsEnabled != null && newDfsPermissionsEnabled !== configDfsPermissionsEnabled.get('value')) {
        affectedProperties.push({
          serviceName : "HDFS",
          sourceServiceName : "HDFS",
          propertyName : 'dfs.permissions.enabled',
          propertyDisplayName : 'dfs.permissions.enabled',
          newValue : newDfsPermissionsEnabled,
          curValue : configDfsPermissionsEnabled.get('value'),
          changedPropertyName : rangerPluginEnabledName,
          removed : false,
          filename : 'hdfs-site.xml'
        });
      }
    }
    return affectedProperties;
  }
});
});

require.register("utils/configs/modification_handlers/hive", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
require('utils/configs/modification_handlers/modification_handler');

module.exports = App.ServiceConfigModificationHandler.create({
  serviceId : 'HIVE',

  getConfig : function(allConfigs, configName, configFilename, configServiceName) {
    return allConfigs.findProperty("serviceName", configServiceName).get("configs").find(function(config) {
      return configName == config.get('name') && (configFilename == null || configFilename == config.get('filename'));
    });
  },

  getDependentConfigChanges : function(changedConfig, selectedServices, allConfigs, securityEnabled) {
    var affectedProperties = [];
    var newValue = changedConfig.get("value");
    var rangerPluginEnabledName = "ranger-hive-plugin-enabled";
    var affectedPropertyName = changedConfig.get("name");
    if (affectedPropertyName == rangerPluginEnabledName) {
      var configAuthorizationEnabled = this.getConfig(allConfigs, 'hive.security.authorization.enabled', 'hive-site.xml', 'HIVE');
      var configAuthorizationManager = this.getConfig(allConfigs, 'hive.security.authorization.manager', 'hiveserver2-site.xml', 'HIVE');
      var configAuthenticatorManager = this.getConfig(allConfigs, 'hive.security.authenticator.manager', 'hiveserver2-site.xml', 'HIVE');
      var configRestrictedList = this.getConfig(allConfigs, 'hive.conf.restricted.list', 'hive-site.xml', 'HIVE');

      var rangerPluginEnabled = newValue == "Yes";
      var newConfigAuthorizationEnabledValue = rangerPluginEnabled ? "true" : "false";
      var newAuthorizationManagerValue = rangerPluginEnabled ? "com.xasecure.authorization.hive.authorizer.XaSecureHiveAuthorizerFactory"
          : "org.apache.hadoop.hive.ql.security.authorization.plugin.sqlstd.SQLStdHiveAuthorizerFactory";
      var newAuthenticatorManagerValue = rangerPluginEnabled ? "org.apache.hadoop.hive.ql.security.SessionStateUserAuthenticator"
          : "org.apache.hadoop.hive.ql.security.SessionStateUserAuthenticator";
      var enabledRestrictedMap = {
        "hive.security.authorization.enabled" : "hive.security.authorization.enabled",
        "hive.security.authorization.manager" : "hive.security.authorization.manager",
        "hive.security.authenticator.manager" : "hive.security.authenticator.manager"
      }
      var enabledRestrictedList = Object.keys(enabledRestrictedMap);
      var newRestrictedListValue = rangerPluginEnabled ? enabledRestrictedList : [];

      // Add Hive-Ranger configs
      if (configAuthorizationEnabled != null && newConfigAuthorizationEnabledValue !== configAuthorizationEnabled.get('value')) {
        affectedProperties.push({
          serviceName : "HIVE",
          sourceServiceName : "HIVE",
          propertyName : 'hive.security.authorization.enabled',
          propertyDisplayName : 'hive.security.authorization.enabled',
          newValue : newConfigAuthorizationEnabledValue,
          curValue : configAuthorizationEnabled.get('value'),
          changedPropertyName : rangerPluginEnabledName,
          removed : false,
          filename : 'hive-site.xml'
        });
      }
      if (configAuthorizationManager != null && newAuthorizationManagerValue !== configAuthorizationManager.get('value')) {
        affectedProperties.push({
          serviceName : "HIVE",
          sourceServiceName : "HIVE",
          propertyName : 'hive.security.authorization.manager',
          propertyDisplayName : 'hive.security.authorization.manager',
          newValue : newAuthorizationManagerValue,
          curValue : configAuthorizationManager.get('value'),
          changedPropertyName : rangerPluginEnabledName,
          removed : false,
          filename : 'hiveserver2-site.xml'
        });
      }
      if (configAuthenticatorManager != null && newAuthenticatorManagerValue !== configAuthenticatorManager.get('value')) {
        affectedProperties.push({
          serviceName : "HIVE",
          sourceServiceName : "HIVE",
          propertyName : 'hive.security.authenticator.manager',
          propertyDisplayName : 'hive.security.authenticator.manager',
          newValue : newAuthenticatorManagerValue,
          curValue : configAuthenticatorManager.get('value'),
          changedPropertyName : rangerPluginEnabledName,
          removed : false,
          filename : 'hiveserver2-site.xml'
        });
      }
      if (configRestrictedList != null) {
        var currentValueList = configRestrictedList.get('value').split(',');
        // 'newRestrictedListValue' elements should be found in existing list
        var newValueList = [];
        currentValueList.forEach(function(s) {
          if (enabledRestrictedMap[s] == s) {
            return;
          }
          newValueList.push(s);
        });
        if (newRestrictedListValue.length > 0)
          newValueList = newValueList.concat(newRestrictedListValue);

        if (newValueList.length != currentValueList.length) {
          // One of the value was not found - set all of them in.
          var newValueListString = newValueList.join(',');
          affectedProperties.push({
            serviceName : "HIVE",
            sourceServiceName : "HIVE",
            propertyName : 'hive.conf.restricted.list',
            propertyDisplayName : 'hive.conf.restricted.list',
            newValue : newValueListString,
            curValue : configRestrictedList.get('value'),
            changedPropertyName : rangerPluginEnabledName,
            removed : false,
            filename : 'hive-site.xml'
          });
        }
      }
    }
    return affectedProperties;
  }
});
});

require.register("utils/configs/modification_handlers/knox", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
require('utils/configs/modification_handlers/modification_handler');

module.exports = App.ServiceConfigModificationHandler.create({
  serviceId : 'KNOX',

  getConfig : function(allConfigs, configName, configFilename, configServiceName) {
    return allConfigs.findProperty("serviceName", configServiceName).get("configs").find(function(config) {
      return configName == config.get('name') && (configFilename == null || configFilename == config.get('filename'));
    });
  },

  getDependentConfigChanges : function(changedConfig, selectedServices, allConfigs, securityEnabled) {
    var affectedProperties = [];
    var newValue = changedConfig.get("value");
    var rangerPluginEnablePropertyName = "ranger-knox-plugin-enabled";
    var affectedPropertyName = changedConfig.get("name");
    if (affectedPropertyName == rangerPluginEnablePropertyName) {
      var topologyXmlContent = this.getConfig(allConfigs, 'content', 'topology.xml', 'KNOX');
      if (topologyXmlContent != null) {
        var topologyXmlContentString = topologyXmlContent.get('value');
        var newTopologyXmlContentString = null;
        var authEnabled = newValue == "Yes";
        var authXml = /<provider>[\s]*<role>[\s]*authorization[\s]*<\/role>[\s\S]*?<\/provider>/.exec(topologyXmlContentString);
        if (authXml != null && authXml.length > 0) {
          var nameArray = /<name>\s*(.*?)\s*<\/name>/.exec(authXml[0]);
          if (nameArray != null && nameArray.length > 1) {
            if (authEnabled && 'AclsAuthz' == nameArray[1]) {
              var newName = nameArray[0].replace('AclsAuthz', 'XASecurePDPKnox');
              var newAuthXml = authXml[0].replace(nameArray[0], newName);
              newTopologyXmlContentString = topologyXmlContentString.replace(authXml[0], newAuthXml);
            } else if (!authEnabled && 'XASecurePDPKnox' == nameArray[1]) {
              var newName = nameArray[0].replace('XASecurePDPKnox', 'AclsAuthz');
              var newAuthXml = authXml[0].replace(nameArray[0], newName);
              newTopologyXmlContentString = topologyXmlContentString.replace(authXml[0], newAuthXml);
            }
          }
        }
        if (newTopologyXmlContentString != null) {
          affectedProperties.push({
            serviceName : "KNOX",
            sourceServiceName : "KNOX",
            propertyName : 'content',
            propertyDisplayName : 'content',
            newValue : newTopologyXmlContentString,
            curValue : topologyXmlContent.get('value'),
            changedPropertyName : rangerPluginEnablePropertyName,
            removed : false,
            filename : 'topology.xml'
          });
        }
      }
    }
    return affectedProperties;
  }
});
});

require.register("utils/configs/modification_handlers/misc", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
require('utils/configs/modification_handlers/modification_handler');

module.exports = App.ServiceConfigModificationHandler.create({
  serviceId : 'MISC',
  getDependentConfigChanges : function(changedConfig, selectedServices, allConfigs, securityEnabled) {
    var affectedProperties = [];
    var newValue = changedConfig.get("value");
    var curConfigs = "";
    var affectedPropertyName = "dfs.permissions.superusergroup";
    if (changedConfig.get("name") == "hdfs_user") {
      curConfigs = allConfigs.findProperty("serviceName", "HDFS").get("configs");
      if (newValue != curConfigs.findProperty("name", affectedPropertyName).get("value")) {
        affectedProperties.push({
          serviceName : "HDFS",
          sourceServiceName : "MISC",
          propertyName : affectedPropertyName,
          propertyDisplayName : affectedPropertyName,
          newValue : newValue,
          curValue : curConfigs.findProperty("name", affectedPropertyName).get("value"),
          changedPropertyName : "hdfs_user",
          remove : false,
          filename : 'hdfs-site.xml'
        });
      }
      if ($.trim(newValue) != $.trim(curConfigs.findProperty("name", "dfs.cluster.administrators").get("value"))) {
        affectedProperties.push({
          serviceName : "HDFS",
          sourceServiceName : "MISC",
          propertyName : "dfs.cluster.administrators",
          propertyDisplayName : "dfs.cluster.administrators",
          newValue : " " + $.trim(newValue),
          curValue : curConfigs.findProperty("name", "dfs.cluster.administrators").get("value"),
          changedPropertyName : "hdfs_user",
          remove : false,
          filename : 'hdfs-site.xml'
        });
      }
    } else if (changedConfig.get("name") == "user_group") {
      if (!(selectedServices.indexOf("YARN") >= 0)) {
        return;
      }
      if (selectedServices.indexOf("MAPREDUCE2") >= 0) {
        curConfigs = allConfigs.findProperty("serviceName", "MAPREDUCE2").get("configs");
        if ($.trim(newValue) != $.trim(curConfigs.findProperty("name", "mapreduce.cluster.administrators").get("value"))) {
          affectedProperties.push({
            serviceName : "MAPREDUCE2",
            sourceServiceName : "MISC",
            propertyName : "mapreduce.cluster.administrators",
            propertyDisplayName : "mapreduce.cluster.administrators",
            newValue : " " + $.trim(newValue),
            curValue : curConfigs.findProperty("name", "mapreduce.cluster.administrators").get("value"),
            changedPropertyName : "user_group",
            filename : 'mapred-site.xml'
          });
        }
      }
      if (selectedServices.indexOf("YARN") >= 0) {
        curConfigs = allConfigs.findProperty("serviceName", "YARN").get("configs");
        if (newValue != curConfigs.findProperty("name", "yarn.nodemanager.linux-container-executor.group").get("value")) {
          affectedProperties.push({
            serviceName : "YARN",
            sourceServiceName : "MISC",
            propertyName : "yarn.nodemanager.linux-container-executor.group",
            propertyDisplayName : "yarn.nodemanager.linux-container-executor.group",
            newValue : newValue,
            curValue : curConfigs.findProperty("name", "yarn.nodemanager.linux-container-executor.group").get("value"),
            changedPropertyName : "user_group",
            filename : 'yarn-site.xml'
          })
        }
      }
    }
    return affectedProperties;
  }
});
});

require.register("utils/configs/modification_handlers/modification_handler", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * Config modification handler for a given service.
 *
 * This handler should take care of scenarios that arise from modification of a
 * config. Scenarios like updating dependency configs, or doing any special
 * processing.
 */
App.ServiceConfigModificationHandler = Em.Object.extend({

  /**
   * Service whose configs this handler will react to.
   */
  serviceId : DS.attr('string'),

  /**
   * Given a changed config, provide an array of changes to dependent configs
   * across all services. Dependent configs can be created, updated and deleted.
   *
   *  Example: Here, setting 'hdfs_user' updates 'dfs.permissions.superusergroup'.
   *  [
   *   {
   *     serviceName: "HDFS",
   *     sourceServiceName: "MISC",  // Cross service configs
   *     propertyName: "dfs.permissions.superusergroup",
   *     propertyDisplayName: "dfs.permissions.superusergroup",
   *     filename: 'hdfs-site',
   *     remove: false,
   *     curValue: "hdfs",
   *     newValue: "my_hdfs",
   *     changedPropertyName: "hdfs_user"
   *   }
   *  ]
   *
   * @param changedConfig
   * @param selectedServices
   * @param allConfigs
   * @param securityEnabled
   * @return Array of dependent config updates
   */
  getDependentConfigChanges : function(changedConfig, selectedServices, allConfigs, securityEnabled) {
    return [];
  }

});
});

require.register("utils/configs/modification_handlers/storm", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
require('utils/configs/modification_handlers/modification_handler');

module.exports = App.ServiceConfigModificationHandler.create({
  serviceId : 'STORM',

  getConfig : function(allConfigs, configName, configFilename, configServiceName) {
    return allConfigs.findProperty("serviceName", configServiceName).get("configs").find(function(config) {
      return configName == config.get('name') && (configFilename == null || configFilename == config.get('filename'));
    });
  },

  getDependentConfigChanges : function(changedConfig, selectedServices, allConfigs, securityEnabled) {
    var affectedProperties = [];
    var newValue = changedConfig.get("value");
    var rangerPluginEnablePropertyName = "ranger-storm-plugin-enabled";
    var affectedPropertyName = changedConfig.get("name");
    if (affectedPropertyName == rangerPluginEnablePropertyName) {
      var authEnabled = newValue == "Yes";
      var configNimbusAuthorizer = this.getConfig(allConfigs, 'nimbus.authorizer', 'storm-site.xml', 'STORM');
      if (configNimbusAuthorizer != null) {
        // Only when configuration is already present, do we act on it.
        // Unsecured clusters do not have this config, and hence we skip any
        // updates
        var newNimbusAuthorizer = authEnabled ? "com.xasecure.authorization.storm.authorizer.XaSecureStormAuthorizer"
            : "backtype.storm.security.auth.authorizer.SimpleACLAuthorizer";

        // Add Hive-Ranger configs
        if (newNimbusAuthorizer !== configNimbusAuthorizer.get('value')) {
          affectedProperties.push({
            serviceName : "STORM",
            sourceServiceName : "STORM",
            propertyName : 'nimbus.authorizer',
            propertyDisplayName : 'nimbus.authorizer',
            newValue : newNimbusAuthorizer,
            curValue : configNimbusAuthorizer.get('value'),
            changedPropertyName : rangerPluginEnablePropertyName,
            removed : false,
            filename : 'storm-site.xml'
          });
        }
      }
      if (authEnabled && affectedProperties.length < 1 && !securityEnabled) {
        App.ModalPopup.show({
          header : Em.I18n.t('services.storm.configs.range-plugin-enable.dialog.title'),
          primary : Em.I18n.t('ok'),
          secondary : false,
          showCloseButton : false,
          onPrimary : function() {
            this.hide();
          },
          body : Em.I18n.t('services.storm.configs.range-plugin-enable.dialog.message')
        });
      }
    }
    return affectedProperties;
  }
});
});

require.register("utils/data_manipulation", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

module.exports = {

  /**
   * Exclude objects from <code>collection</code> if its <code>key</code> exist in <code>valuesToReject</code>
   * Example:
   * <code>
   *   var collection = [{n: 'v1'}, {n: 'v2'}, {n: 'v3'}, {n: 'v4'}],
   *   key = 'n',
   *   valuesToReject = ['v2', 'v3'];
   *   var result = rejectPropertyValues(collection, key, valuesToReject); // [{n: 'v1'}, {n: 'v4'}]
   * </code>
   * @param {Object[]} collection array of objects
   * @param {String} key property name of each object to be checked
   * @param {String[]} valuesToReject list of <code>key</code> values. Objects with <code>key</code>,
   *                                    equal to one of them, will be excluded
   * @returns {Object[]} Rejected array
   */
  rejectPropertyValues: function(collection, key, valuesToReject) {
    return collection.filter(function (item) {
      var propertyValue = Em.get(item, key);
      return valuesToReject.contains(propertyValue);
    });
  },

  /**
   * Wrapper to <code>filterProperty</code>-method that allows using list of values to filter
   * Example:
   * <code>
   *   var collection = [{n: 'v1'}, {n: 'v2'}, {n: 'v3'}, {n: 'v4'}],
   *   key = 'n',
   *   valuesToFilter = ['v2', 'v3'];
   *   var result = filterPropertyValues(collection, key, valuesToFilter); // [{n: 'v2'}, {n: 'v3'}]
   * </code>
   * @param {Object[]} collection array of objects
   * @param {String} key property name of each object to be filtered
   * @param {String|String[]} valuesToFilter array of values (or one value-string) to filter
   * @returns {Object[]} Filtered array
   */
  filterPropertyValues: function(collection, key, valuesToFilter) {
    var type = Em.typeOf(valuesToFilter);
    if ('string' === type)
      return collection.filterProperty(key, valuesToFilter);

    var ret = [];
    if ('array' === type) {
      valuesToFilter.forEach(function (value) {
        ret = ret.concat(collection.filterProperty(key, value));
      });
    }
    return ret;
  },

  /**
   * Group <code>collection</code> items by <code>key</code> value
   * Example:
   * <code>
   *   var collection = [{n: 'v1'}, {n: 'v2'}, {n: 'v2'}, {n: 'v4'}],
   *   key = 'n';
   *   var result = groupPropertyValues(collection, key); // {v1: [{n: 'v1'}], v2: [{n: 'v2'}, {n: 'v2'}], v4: [{n: 'v4'}]}
   * </code>
   * @param {Object[]} collection array of objects
   * @param {String} key property name of each object to be grouped
   * @returns {*}
   */
  groupPropertyValues: function(collection, key) {
    var group = {};
    collection.forEach(function(item) {
      var value = Em.get(item, key);
      if (Em.isNone(group[value])) {
        group[value] = [item];
      }
      else {
        group[value].pushObject(item);
      }
    });
    return group;
  }

};

});

require.register("utils/date", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var validator = require('utils/validator');
var App = require('app');

module.exports = {

  /**
   * List of monthes short names
   * @type {string[]}
   */
  dateMonths: ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'],

  /**
   * List of days short names
   * @type {string[]}
   */
  dateDays: ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'],

  /**
   * Add leading zero
   *
   * @param {string} time
   * @returns {string}
   * @method dateFormatZeroFirst
   */
  dateFormatZeroFirst: function (time) {
    if (time < 10) return '0' + time;
    return ""  + time;
  },

  /**
   * Convert timestamp to date-string
   * default format - 'DAY_OF_THE_WEEK, MONTH DAY, YEAR HOURS:MINUTES'
   *
   * @param {number} timestamp
   * @param {bool} format
   * @return {*} date
   * @method dateFormat
   */
  dateFormat: function (timestamp, format) {
    if (!validator.isValidInt(timestamp)) {
      return timestamp;
    }
    format = format || 'ddd, MMM DD, YYYY HH:mm';

    return moment((new Date(timestamp))).format(format);
  },

  /**
   * Convert timestamp to date-string 'DAY_OF_THE_WEEK MONTH DAY YEAR'
   *
   * @param {string} timestamp
   * @return {string}
   * @method dateFormatShort
   */
  dateFormatShort: function (timestamp) {
    if (!validator.isValidInt(timestamp)) {
      return timestamp;
    }
    var format = 'ddd MMM DD YYYY';
    var date = moment((new Date(timestamp))).format(format);
    var today = moment((new Date())).format(format);
    if (date === today) {
      return 'Today ' + (new Date(timestamp)).toLocaleTimeString();
    }
    return date;
  },

  /**
   * Convert starTimestamp to 'DAY_OF_THE_WEEK, MONTH DAY, YEAR HOURS:MINUTES', except for the case: year equals 1969
   *
   * @param {string} startTimestamp
   * @return {string} startTimeSummary
   * @method startTime
   */
  startTime: function (startTimestamp) {
    if (!validator.isValidInt(startTimestamp)) {
      return '';
    }
    var startDate = new Date(startTimestamp);
    var months = this.dateMonths;
    var days = this.dateDays;
    // generate start time
    if (startDate.getFullYear() == 1969 || startTimestamp < 1) {
      return 'Not started';
    }
    var startTimeSummary = '';
    if (new Date(startTimestamp).setHours(0, 0, 0, 0) == new Date().setHours(0, 0, 0, 0)) { //today
      startTimeSummary = 'Today ' + this.dateFormatZeroFirst(startDate.getHours()) + ':' + this.dateFormatZeroFirst(startDate.getMinutes());
    } else {
      startTimeSummary = days[startDate.getDay()] + ' ' + months[startDate.getMonth()] + ' ' +
        this.dateFormatZeroFirst(startDate.getDate()) + ' ' + startDate.getFullYear() + ' '
        + this.dateFormatZeroFirst(startDate.getHours()) + ':' + this.dateFormatZeroFirst(startDate.getMinutes());
    }
    return startTimeSummary;
  },

  /**
   * Provides the duration between the given start and end timestamp. If start time
   * not valid, duration will be ''. If end time is not valid, duration will
   * be till now, showing 'Lasted for xxx secs'.
   *
   * @param {string} startTimestamp
   * @param {string} endTimestamp
   * @return {string} durationSummary
   * @method durationSummary
   */
  durationSummary: function (startTimestamp, endTimestamp) {
    // generate duration
    var durationSummary = '';
    var startDate = new Date(startTimestamp);
    var endDate = new Date(endTimestamp);
    var self = this;
    if (startDate.getFullYear() == 1969 || startTimestamp < 1) {
      // not started
      return Em.I18n.t('common.na');
    }
    if (endDate.getFullYear() != 1969 && endTimestamp > 0) {
      return '' + this.timingFormat(endTimestamp - startTimestamp, 1); //lasted for xx secs
    } else {
      // still running, duration till now
      var time = (App.dateTime() - startTimestamp) < 0 ? 0 : (App.dateTime() - startTimestamp);
      durationSummary = '' + this.timingFormat(time, 1);
    }
    return durationSummary;
  },

  /**
   * Convert time in mseconds to
   * 30 ms = 30 ms
   * 300 ms = 300 ms
   * 999 ms = 999 ms
   * 1000 ms = 1.00 secs
   * 3000 ms = 3.00 secs
   * 35000 ms = 35.00 secs
   * 350000 ms = 350.00 secs
   * 999999 ms = 999.99 secs
   * 1000000 ms = 16.66 mins
   * 3500000 secs = 58.33 mins
   *
   * @param {number} time
   * @param {bool} [zeroValid] for the case to show 0 when time is 0, not null
   * @return {string|null} formatted date
   * @method timingFormat
   */
  timingFormat: function (time, /* optional */ zeroValid) {
    var intTime = parseInt(time);
    if (zeroValid && intTime == 0) {
      return 0 + ' secs';
    }
    if (!intTime) {
      return null;
    }
    var timeStr = intTime.toString();
    var lengthOfNumber = timeStr.length;
    var oneMinMs = 60000;
    var oneHourMs = 3600000;
    var oneDayMs = 86400000;

    if (lengthOfNumber < 4) {
      return time + ' ms';
    } else if (lengthOfNumber < 7) {
      time = (time / 1000).toFixed(2);
      return time + ' secs';
    } else if (time < oneHourMs) {
      time = (time / oneMinMs).toFixed(2);
      return time + ' mins';
    } else if (time < oneDayMs) {
      time = (time / oneHourMs).toFixed(2);
      return time + ' hours';
    } else {
      time = (time / oneDayMs).toFixed(2);
      return time + ' days';
    }
  },

  /**
   * Provides the duration between the given start and end time. If start time
   * is not given, duration will be 0. If end time is not given, duration will
   * be till now.
   *
   * @param {Number} startTime Start time from epoch
   * @param {Number} endTime End time from epoch
   * @return {Number} duration
   * @method duration
   */
  duration: function (startTime, endTime) {
    var duration = 0;
    if (startTime && startTime > 0) {
      if (!endTime || endTime < 1) {
        endTime = App.dateTime();
      }
      duration = endTime - startTime;
    }
    return duration;
  }
};

});

require.register("utils/db", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');
App.db = {};
var InitialData =  {
  'app': {
    'loginName': '',
    'authenticated': false,
    'configs': [],
    'tables': {
      'filterConditions': {},
      'displayLength': {},
      'startIndex': {},
      'sortingConditions': {},
      'selectedItems': {}
    }
  },

  'Installer' : {},
  'AddHost' : {},
  'AddService' : {},
  'KerberosWizard': {},
  'StackUpgrade' : {},
  'ReassignMaster' : {},
  'AddSecurity': {},
  'AddAlertDefinition': {
    content: {}
  },
  'HighAvailabilityWizard': {},
  'RMHighAvailabilityWizard': {},
  'RollbackHighAvailabilityWizard': {},
  'MainAdminStackAndUpgrade': {},
  'KerberosDisable': {},
  'tmp': {}

};

if (typeof Storage !== 'undefined') {
  Storage.prototype.setObject = function (key, value) {
    this.setItem(key, JSON.stringify(value));
  };

  Storage.prototype.getObject = function (key) {
    var value = this.getItem(key);
    return value && JSON.parse(value);
  };
} else {
  // stub for unit testing purposes
  window.localStorage = {};
  localStorage.setItem = function (key, val) {
    this[key] = val;
  };
  localStorage.getItem = function (key) {
    return this[key];
  };
  window.localStorage.setObject = function (key, value) {
    this[key] = value;
  };
  window.localStorage.getObject = function (key, value) {
    return this[key];
  };
}

App.db.cleanUp = function () {
  console.log('TRACE: Entering db:cleanup function');
  App.db.data = InitialData;
  console.log("In cleanup./..");
  localStorage.setObject('ambari', App.db.data);
};

App.db.cleanTmp = function() {
  App.db.data.tmp = {};
  localStorage.setObject('ambari', App.db.data);
};

App.db.cleanAddHost = function() {
  App.db.data.AddHost = {};
  localStorage.setObject('ambari', App.db.data);
};

App.db.updateStorage = function() {
  App.db.data = localStorage.getObject('ambari');
  if (App.db.data && App.db.data.app && App.db.data.app.tables && App.db.data.app.configs) {
    return true;
  }
  console.warn("local storage is deprecated!");
  App.db.cleanUp();
  return false;
};

/*
  Initialize wizard namespaces if they are not initialized on login.
  This will be required during upgrade.
 */
App.db.mergeStorage = function() {
  if (localStorage.getObject('ambari') == null) {
    console.log('doing a cleanup');
    App.db.cleanUp();
  } else {
    localStorage.setObject('ambari', $.extend(true,{}, InitialData, App.db.data));
  }
};

// called whenever user logs in
if (localStorage.getObject('ambari') == null) {
  console.log('doing a cleanup');
  App.db.cleanUp();
}

App.db.get = function (namespace, key) {
  console.log('TRACE: Entering db:get' + key);
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data[namespace]) {
    App.db.data[namespace] = {};
  }
  return App.db.data[namespace][key];
};

App.db.set = function (namespace, key, value) {
  console.log('TRACE: Entering db:set' + key + ';value: ', value);
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data[namespace]) {
    App.db.data[namespace] = {};
  }
  App.db.data[namespace][key] = value;
  localStorage.setObject('ambari', App.db.data);
};
/*
 * setter methods
 */

App.db.setLoginName = function (name) {
  console.log('TRACE: Entering db:setLoginName function');
  App.db.data = localStorage.getObject('ambari');
  App.db.data.app.loginName = name;
  localStorage.setObject('ambari', App.db.data);
};

/**
 * Set user model to db
 * @param user
 */
App.db.setUser = function (user) {
  console.log('TRACE: Entering db:setUser function');
  App.db.data = localStorage.getObject('ambari');
  App.db.data.app.user = user;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setAuthenticated = function (authenticated) {
  console.log('TRACE: Entering db:setAuthenticated function');

  App.db.data = localStorage.getObject('ambari');
  console.log('present value of authentication is: ' + App.db.data.app.authenticated);
  console.log('desired value of authentication is: ' + authenticated);
  App.db.data.app.authenticated = authenticated;
  localStorage.setObject('ambari', App.db.data);
  App.db.data = localStorage.getObject('ambari');
  console.log('Now present value of authentication is: ' + App.db.data.app.authenticated);
};

App.db.setFilterConditions = function(name, filterConditions) {
  console.log('TRACE: Entering db:setFilterConditions function');
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data.app.tables.filterConditions) {
    App.db.data.app.tables.filterConditions = {};
  }
  App.db.data.app.tables.filterConditions[name] = filterConditions;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setDisplayLength = function(name, displayLength) {
  console.log('TRACE: Entering db:setDisplayLength function');
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data.app.tables.displayLength) {
    App.db.data.app.tables.displayLength = {};
  }
  App.db.data.app.tables.displayLength[name] = displayLength;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setStartIndex = function(name, startIndex) {
  console.log('TRACE: Entering db:setStartIndex function');
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data.app.tables.startIndex) {
    App.db.data.app.tables.startIndex = {};
  }
  App.db.data.app.tables.startIndex[name] = startIndex;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setSortingStatuses = function(name, sortingConditions) {
  console.log('TRACE: Entering db:setSortingConditions function');
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data.app.tables.sortingConditions) {
    App.db.data.app.tables.sortingConditions = {};
  }
  App.db.data.app.tables.sortingConditions[name] = sortingConditions;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setSelectedHosts = function(name, selectedHosts) {
  console.log('TRACE: Entering db:setSelectedHosts function');
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data.app.tables.selectedItems) {
    App.db.data.app.tables.selectedItems = {};
  }
  App.db.data.app.tables.selectedItems[name] = selectedHosts;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setAllHostNames = function (hostNames) {
  console.log('TRACE: Entering db:setAllHostNames function');
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.hostNames = hostNames;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setAllHostNamesPattern = function (hostNames) {
  console.log('TRACE: Entering db:setAllHostNamesPattern function');
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.hostNamesPattern = hostNames;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setHosts = function (hostInfo) {
  console.log('TRACE: Entering db:setHosts function');
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.hostInfo = hostInfo;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setInstallOptions = function (installOptions) {
  console.log('TRACE: Entering db:setInstallOptions function');
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.installOptions = installOptions;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setBootStatus = function (status) {
  console.log('TRACE: Entering db:setBootStatus function');
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.bootStatus = status;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setService = function (serviceInfo) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.serviceInfo = serviceInfo;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setSelectedServiceNames = function (serviceNames) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.selectedServiceNames = serviceNames;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setClientsForSelectedServices = function (clientInfo) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.clientInfo = clientInfo;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setMasterComponentHosts = function (masterComponentHosts) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.masterComponentHosts = masterComponentHosts;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setSlaveComponentHosts = function (slaveComponentHosts) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.slaveComponentHosts = slaveComponentHosts;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setServiceConfigs = function (serviceConfigs) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.serviceConfigs = serviceConfigs;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setAdvancedServiceConfig = function (serviceConfigs) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.advanceServiceConfigs = serviceConfigs;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setServiceConfigProperties = function (configProperties) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.configProperties = configProperties;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setCluster = function (status) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.clusterStatus = status;
  console.log('db.setClusterStatus called: ' + JSON.stringify(status));
  localStorage.setObject('ambari', App.db.data);
};

App.db.setUpgradeOptions = function (upgradeOptions) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.StackUpgrade.upgradeOptions = upgradeOptions;
  console.log('db.setUpgradeOptions called: ' + JSON.stringify(upgradeOptions));
  localStorage.setObject('ambari', App.db.data);
};

App.db.setMasterToReassign = function (masterComponent) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.ReassignMaster.masterComponent = masterComponent;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setReassignTasksStatuses = function (tasksStatuses) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.ReassignMaster.tasksStatuses = tasksStatuses;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setReassignTasksRequestIds = function (requestIds) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.ReassignMaster.tasksRequestIds = requestIds;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setStacks = function (stacks) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.Installer.stacksVersions = stacks;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setServices = function (services) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.app.services = services;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setConfigs = function (configs) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.app.configs = configs;
  localStorage.setObject('ambari', App.db.data);
};

/**
 * Set current step value for specified Wizard Type
 * @param wizardType
 * @param currentStep
 */
App.db.setWizardCurrentStep = function (wizardType, currentStep) {
  console.log('TRACE: Entering db:setWizardCurrentStep function');
  App.db.data = localStorage.getObject('ambari');
  App.db.data[wizardType.capitalize()].currentStep = currentStep;
  localStorage.setObject('ambari', App.db.data);
};

/**
 * Set localStorage with data from server
 */
App.db.setLocalStorage = function () {
  localStorage.setObject('ambari', App.db.data);
};

App.db.setSecurityWizardStatus = function (status) {
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data.AddSecurity) {
    App.db.data.AddSecurity = {};
  }
  App.db.data.AddSecurity.status = status;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setDisableSecurityStatus = function (status) {
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data.AddSecurity) {
    App.db.data.AddSecurity = {};
  }
  App.db.data.AddSecurity.disableSecurityStatus = status;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setSecurityDeployCommands = function (commands) {
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data.AddSecurity) {
    App.db.data.AddSecurity = {};
  }
  App.db.data.AddSecurity.securityDeployCommands = commands;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setSecureConfigProperties  = function (secureConfigs) {
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data.AddSecurity) {
    App.db.data.AddSecurity = {};
  }
  App.db.data.AddSecurity.secureConfigProperties = secureConfigs;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setSecureUserInfo  = function (userInfo) {
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data.AddSecurity) {
    App.db.data.AddSecurity = {};
  }
  App.db.data.AddSecurity.secureUserInfo = userInfo;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setIsNameNodeHa = function (haStatus) {
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data.AddSecurity) {
    App.db.data.AddSecurity = {};
  }
  App.db.data.AddSecurity.haStatus = haStatus;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setHighAvailabilityWizardConfigTag = function (tag) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.HighAvailabilityWizard[tag.name] = tag.value;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setHighAvailabilityWizardHdfsClientHosts = function (hostNames) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.HighAvailabilityWizard.hdfsClientHostNames = hostNames;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setHighAvailabilityWizardTasksStatuses = function (tasksStatuses) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.HighAvailabilityWizard.tasksStatuses = tasksStatuses;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setHighAvailabilityWizardTasksRequestIds = function (requestIds) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.HighAvailabilityWizard.tasksRequestIds = requestIds;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setHighAvailabilityWizardHdfsUser = function (hdfsUser) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.HighAvailabilityWizard.hdfsUser = hdfsUser;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setHighAvailabilityWizardRequestIds = function (requestIds) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.HighAvailabilityWizard.requestIds = requestIds;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setHighAvailabilityWizardNameServiceId = function (nameServiceId) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.HighAvailabilityWizard.nameServiceId = nameServiceId;
  localStorage.setObject('ambari', App.db.data);
};


App.db.setRollBackHighAvailabilityWizardAddNNHost = function (host) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.RollbackHighAvailabilityWizard.addNNHost = host;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setRollBackHighAvailabilityWizardSNNHost = function (host) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.RollbackHighAvailabilityWizard.sNNHost = host;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setRollBackHighAvailabilityWizardSelectedAddNN = function (host) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.RollbackHighAvailabilityWizard.selectedAddNN = host;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setRollBackHighAvailabilityWizardSelectedSNN = function (host) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.RollbackHighAvailabilityWizard.selectedSNNH = host;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setRollbackHighAvailabilityWizardTasksStatuses = function (tasksStatuses) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.RollbackHighAvailabilityWizard.tasksStatuses = tasksStatuses;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setRollbackHighAvailabilityWizardTasksRequestIds = function (requestIds) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.RollbackHighAvailabilityWizard.tasksRequestIds = requestIds;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setRollbackHighAvailabilityWizardRequestIds = function (requestIds) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.RollbackHighAvailabilityWizard.requestIds = requestIds;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setReassignMasterWizardRequestIds = function (requestIds) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.ReassignMaster.requestIds = requestIds;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setReassignMasterWizardComponentDir = function (componentDir) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.ReassignMaster.componentDir = componentDir;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setReassignMasterWizardReassignHosts = function (reassignHosts) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.ReassignMaster.reassignHosts = reassignHosts;
  localStorage.setObject('ambari', App.db.data);
};

App.db.setKerberosWizardConfigTag = function (tag) {
  App.db.data = localStorage.getObject('ambari');
  App.db.data.KerberosWizard[tag.name] = tag.value;
  localStorage.setObject('ambari', App.db.data);
};

/*
 *  getter methods
 */

/**
 * Get user model from db
 * @return {*}
 */
App.db.getUser = function () {
  console.log('TRACE: Entering db:getUser function');
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.app.user;
};

App.db.getLoginName = function () {
  console.log('Trace: Entering db:getLoginName function');
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.app.loginName;
};

App.db.getAuthenticated = function () {
  console.log('Trace: Entering db:getAuthenticated function');
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.app.authenticated;
};

App.db.getFilterConditions = function(name) {
  console.log('TRACE: Entering db:getFilterConditions function');
  App.db.data = localStorage.getObject('ambari');
  if (App.db.data.app.tables.filterConditions[name]) {
    return App.db.data.app.tables.filterConditions[name];
  }
  return null;
};

App.db.getDisplayLength = function(name) {
  console.log('TRACE: Entering db:getDisplayLength function');
  App.db.data = localStorage.getObject('ambari');
  if (App.db.data.app.tables.displayLength[name]) {
    return App.db.data.app.tables.displayLength[name];
  }
  return null;
};

App.db.getStartIndex = function(name) {
  console.log('TRACE: Entering db:getStartIndex function');
  App.db.data = localStorage.getObject('ambari');
  if (App.db.data.app.tables.startIndex[name]) {
    return App.db.data.app.tables.startIndex[name];
  }
  return null;
};

App.db.getSortingStatuses = function(name) {
  console.log('TRACE: Entering db:getSortingConditions function');
  App.db.data = localStorage.getObject('ambari');
  if (App.db.data.app.tables.sortingConditions[name]) {
    return App.db.data.app.tables.sortingConditions[name];
  }
  return null;
};

App.db.getSelectedHosts = function(name) {
  console.log('TRACE: Entering db:getSelectedHosts function');
  App.db.data = localStorage.getObject('ambari');
  if (App.db.data.app.tables.selectedItems[name]) {
    return App.db.data.app.tables.selectedItems[name];
  }
  return [];
};

/**
 * Return current step for specified Wizard Type
 * @param wizardType
 * @return {*}
 */
App.db.getWizardCurrentStep = function (wizardType) {
  console.log('Trace: Entering db:getWizardCurrentStep function for ', wizardType);
  if (App.db.data[wizardType.capitalize()]) {
    return App.db.data[wizardType.capitalize()].currentStep;
  }
  return 0;
};

App.db.getAllHostNames = function () {
  console.log('TRACE: Entering db:getHostNames function');
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.hostNames;
};

App.db.getAllHostNamesPattern = function () {
  console.log('TRACE: Entering db:getHostNamesPattern function');
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.hostNamesPattern;
};
App.db.getInstallOptions = function () {
  console.log('TRACE: Entering db:getInstallOptions function');
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.installOptions;
};

App.db.isCompleted = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.completed;
};

App.db.getHosts = function () {
  console.log('TRACE: Entering db:getHosts function');
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.hostInfo;
};

App.db.getBootStatus = function () {
  console.log('TRACE: Entering db:getBootStatus function');
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.bootStatus;
};

App.db.getService = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.serviceInfo;
};

App.db.getSelectedServiceNames = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.selectedServiceNames;
};

App.db.getClientsForSelectedServices = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.clientInfo;
};

App.db.getMasterComponentHosts = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.masterComponentHosts;
};

App.db.getSlaveComponentHosts = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.slaveComponentHosts;
};

App.db.getServiceConfigs = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.serviceConfigs;
};

App.db.getAdvancedServiceConfig = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.advanceServiceConfigs;
};

App.db.getServiceConfigProperties = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.configProperties;
};

App.db.getCluster = function () {
  console.log('TRACE: Entering db:getClusterStatus function');
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.clusterStatus;
};

App.db.getMasterToReassign = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.ReassignMaster.masterComponent;
};

App.db.getReassignTasksStatuses = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.ReassignMaster.tasksStatuses;
};

App.db.getReassignTasksRequestIds = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.ReassignMaster.tasksRequestIds;
};

App.db.getUpgradeOptions = function () {
  console.log('TRACE: Entering db:getUpgradeOptions function');
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.StackUpgrade.upgradeOptions;
};

App.db.getSecurityWizardStatus = function () {
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data.AddSecurity) {
    App.db.data.AddSecurity = {};
  }
  return App.db.data.AddSecurity.status;
};

App.db.getDisableSecurityStatus = function () {
  App.db.data = localStorage.getObject('ambari');
  if (!App.db.data.AddSecurity) {
    App.db.data.AddSecurity = {};
  }
  return App.db.data.AddSecurity.disableSecurityStatus;
};

App.db.getSecurityDeployCommands = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.AddSecurity.securityDeployCommands;
};

App.db.getSecureConfigProperties = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.AddSecurity.secureConfigProperties;
};

App.db.getSecureUserInfo  = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.AddSecurity.secureUserInfo;
};

App.db.getIsNameNodeHa = function (haStatus) {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.AddSecurity.haStatus;
};

App.db.getStacks = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.Installer.stacksVersions;
};

App.db.getServices = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.app.services;
};

App.db.getHighAvailabilityWizardHdfsUser = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.HighAvailabilityWizard.hdfsUser;
};

App.db.getHighAvailabilityWizardTasksStatuses = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.HighAvailabilityWizard.tasksStatuses;
};

App.db.getHighAvailabilityWizardTasksRequestIds = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.HighAvailabilityWizard.tasksRequestIds;
};

App.db.getHighAvailabilityWizardFailedTask = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.HighAvailabilityWizard.failedTask;
};

App.db.getHighAvailabilityWizardHdfsClientHosts = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.HighAvailabilityWizard.hdfsClientHostNames;
};

App.db.getHighAvailabilityWizardConfigTag = function (tag) {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.HighAvailabilityWizard[tag];
};

App.db.getHighAvailabilityWizardRequestIds = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.HighAvailabilityWizard.requestIds;
};

App.db.getHighAvailabilityWizardNameServiceId = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.HighAvailabilityWizard.nameServiceId;
};

App.db.getRollbackHighAvailabilityWizardTasksStatuses = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.RollbackHighAvailabilityWizard.tasksStatuses;
};

App.db.getRollbackHighAvailabilityWizardTasksRequestIds = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.ReassignMaster.tasksRequestIds;
};

App.db.getRollbackHighAvailabilityWizardRequestIds = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.RollbackHighAvailabilityWizard.requestIds;
};

App.db.getRollBackHighAvailabilityWizardAddNNHost = function (host) {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.RollbackHighAvailabilityWizard.addNNHost;
};

App.db.getRollBackHighAvailabilityWizardSNNHost = function (host) {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.RollbackHighAvailabilityWizard.sNNHost;
};

App.db.getReassignMasterWizardRequestIds = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.ReassignMaster.requestIds;
};

App.db.getReassignMasterWizardComponentDir = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.ReassignMaster.componentDir;
};

App.db.getConfigs = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.app.configs;
};

App.db.getReassignMasterWizardReassignHosts = function () {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.ReassignMaster.reassignHosts;
};

App.db.getKerberosWizardConfigTag = function (tag) {
  App.db.data = localStorage.getObject('ambari');
  return App.db.data.KerberosWizard[tag];
};

module.exports = App.db;

});

require.register("utils/ember_reopen", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 Merge the contents of two objects together into the first object.

 ```javascript
 Ember.merge({first: 'Tom'}, {last: 'Dale'}); // {first: 'Tom', last: 'Dale'}
 var a = {first: 'Yehuda'}, b = {last: 'Katz'};
 Ember.merge(a, b); // a == {first: 'Yehuda', last: 'Katz'}, b == {last: 'Katz'}
 ```

 @method merge
 @for Ember
 @param {Object} original The object to merge into
 @param {Object} updates The object to copy properties from
 @return {Object}
 */
Ember.merge = function(original, updates) {
  for (var prop in updates) {
    if (!updates.hasOwnProperty(prop)) { continue; }
    original[prop] = updates[prop];
  }
  return original;
};

/**
 Returns true if the passed value is null or undefined. This avoids errors
 from JSLint complaining about use of ==, which can be technically
 confusing.

 ```javascript
 Ember.isNone();              // true
 Ember.isNone(null);          // true
 Ember.isNone(undefined);     // true
 Ember.isNone('');            // false
 Ember.isNone([]);            // false
 Ember.isNone(function() {});  // false
 ```

 @method isNone
 @for Ember
 @param {Object} obj Value to test
 @return {Boolean}
 */
Ember.isNone = function(obj) {
  return obj === null || obj === undefined;
};

/**
 Verifies that a value is `null` or an empty string, empty array,
 or empty function.

 Constrains the rules on `Ember.isNone` by returning false for empty
 string and empty arrays.

 ```javascript
 Ember.isEmpty();                // true
 Ember.isEmpty(null);            // true
 Ember.isEmpty(undefined);       // true
 Ember.isEmpty('');              // true
 Ember.isEmpty([]);              // true
 Ember.isEmpty('Adam Hawkins');  // false
 Ember.isEmpty([0,1,2]);         // false
 ```

 @method isEmpty
 @for Ember
 @param {Object} obj Value to test
 @return {Boolean}
 */
Ember.isEmpty = function(obj) {
  return Ember.isNone(obj) || (obj.length === 0 && typeof obj !== 'function') || (typeof obj === 'object' && Ember.get(obj, 'length') === 0);
};

/**
 A value is blank if it is empty or a whitespace string.

 ```javascript
 Ember.isBlank();                // true
 Ember.isBlank(null);            // true
 Ember.isBlank(undefined);       // true
 Ember.isBlank('');              // true
 Ember.isBlank([]);              // true
 Ember.isBlank('\n\t');          // true
 Ember.isBlank('  ');            // true
 Ember.isBlank({});              // false
 Ember.isBlank('\n\t Hello');    // false
 Ember.isBlank('Hello world');   // false
 Ember.isBlank([1,2,3]);         // false
 ```

 @method isBlank
 @for Ember
 @param {Object} obj Value to test
 @return {Boolean}
 */
Ember.isBlank = function(obj) {
  return Ember.isEmpty(obj) || (typeof obj === 'string' && obj.match(/\S/) === null);
};

/**
 * Calculates sum of two numbers
 * Use this function as a callback on <code>invoke</code> etc
 *
 * @method sum
 * @param {Number} a
 * @param {Number} b
 * @returns {Number}
 */
Ember.sum = function (a, b) {
  return a + b;
};

/**
 *
 */
Ember.RadioButton = Ember.Checkbox.extend({
  tagName: "input",
  type: "radio",
  attributeBindings: [ "type", "name", "value", "checked", "style", "disabled" ],
  style: "vertical-align: middle; margin: 0px;",
  click: function () {
    this.set("selection", this.$().val())
  },
  checked: function () {
    return this.get("value") == this.get("selection");
  }.property('value', 'selection')
});

/**
 * Set value to obj by path
 * Create nested objects if needed
 * Example:
 * <code>
 *   var a = {b: {}};
 *   var path = 'b.c.d';
 *   var value = 1;
 *   Em.setFullPath(a, path, value); // a = {b: {c: {d: 1}}
 * </code>
 *
 * @param {object} obj
 * @param {string} path
 * @param {*} value
 */
Ember.setFullPath = function (obj, path, value) {
  var parts = path.split('.'),
    sub_path = '';
  parts.forEach(function(_path, _index) {
    Em.assert('path parts can\'t be empty', _path.length);
    sub_path += '.' + _path;
    if (_index === parts.length - 1) {
      Em.set(obj, sub_path, value);
      return;
    }
    if (Em.isNone(Em.get(obj, sub_path))) {
      Em.set(obj, sub_path, {});
    }
  });
};

Em.View.reopen({
  /**
   * overwritten set method of Ember.View to avoid uncaught errors
   * when trying to set property of destroyed view
   */
  set: function(attr, value){
    if(!this.get('isDestroyed') && !this.get('isDestroying')){
      this._super(attr, value);
    } else {
      console.debug('Calling set on destroyed view');
    }
  }
});

Ember.TextArea.reopen({
  attributeBindings: ['readonly']
});

});

require.register("utils/heatmap", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
module.exports = {
  mappers: Em.Mixin.create({
    metricMapperWithTransform: function (json, metricName, transformValueFunction) {
      var hostToValueMap = {};
      //var metricName = this.get('defaultMetric');
      if (json.host_components) {
        var props = metricName.split('.');
        //var transformValueFunction = this.get('transformValue');
        json.host_components.forEach(function (hc) {
          var value = hc;
          props.forEach(function (prop) {
            if (value != null && prop in value) {
              value = value[prop];
            } else {
              value = null;
            }
          });
          if (value != null) {
            if (transformValueFunction) {
              value = transformValueFunction(value);
            }
            var hostName = hc.HostRoles.host_name;
            hostToValueMap[hostName] = value;
          }
        });
      }
      return hostToValueMap;
    }
  })
};

});

require.register("utils/helper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var stringUtils = require('utils/string_utils');

/**
 * Remove spaces at beginning and ending of line.
 * @example
 *  var str = "  I'm a string  "
 *  str.trim() // return "I'm a string"
 * @method trim
 * @return {string}
 */
String.prototype.trim = function () {
  return this.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
};
/**
 * Determines whether string end within another string.
 *
 * @method endsWith
 * @param suffix {string}  substring for search
 * @return {boolean}
 */
String.prototype.endsWith = function(suffix) {
  return this.indexOf(suffix, this.length - suffix.length) !== -1;
};
/**
 * Determines whether string start within another string.
 *
 * @method startsWith
 * @param prefix {string} substring for search
 * @return {boolean}
 */
String.prototype.startsWith = function (prefix){
  return this.indexOf(prefix) == 0;
};
/**
 * Determines whether string founded within another string.
 *
 * @method contains
 * @param substring {string} substring for search
 * @return {boolean}
 */
String.prototype.contains = function(substring) {
  return this.indexOf(substring) != -1;
};
/**
 * Capitalize the first letter of string.
 * @method capitalize
 * @return {string}
 */
String.prototype.capitalize = function () {
  return this.charAt(0).toUpperCase() + this.slice(1);
};

/**
 * Capitalize the first letter of string.
 * And set to lowercase other part of string
 * @method toCapital
 * @return {string}
 */
String.prototype.toCapital = function () {
  return this.charAt(0).toUpperCase() + this.slice(1).toLowerCase();
};

/**
 * Finds the value in an object where this string is a key.
 * Optionally, the index of the key can be provided where the
 * value of the nth key in the hierarchy is returned.
 *
 * Example:
 *  var tofind = 'smart';
 *  var person = {'name': 'Bob Bob', 'smart': 'no', 'age': '28', 'personality': {'smart': 'yes', 'funny': 'yes', 'emotion': 'happy'} };
 *  tofind.findIn(person); // 'no'
 *  tofind.findIn(person, 0); // 'no'
 *  tofind.findIn(person, 1); // 'yes'
 *  tofind.findIn(person, 2); // null
 *
 *  @method findIn
 *  @param multi {object}
 *  @param index {number} Occurrence count of this key
 *  @return {*} Value of key at given index
 */
String.prototype.findIn = function(multi, index, _foundValues) {
  if (!index) {
    index = 0;
  }
  if (!_foundValues) {
    _foundValues = [];
  }
  multi = multi || '';
  var value = null;
  var str = this.valueOf();
  if (typeof multi == 'object') {
    for ( var key in multi) {
      if (value != null) {
        break;
      }
      if (key == str) {
        _foundValues.push(multi[key]);
      }
      if (_foundValues.length - 1 == index) {
        // Found the value
        return _foundValues[index];
      }
      if (typeof multi[key] == 'object') {
        value = value || this.findIn(multi[key], index, _foundValues);
      }
    }
  }
  return value;
};
/**
 * Replace {i} with argument. where i is number of argument to replace with.
 * @example
 *  var str = "{0} world{1}";
 *  str.format("Hello", "!") // return "Hello world!"
 *
 * @method format
 * @return {string}
 */
String.prototype.format = function () {
  var args = arguments;
  return this.replace(/{(\d+)}/g, function (match, number) {
    return typeof args[number] != 'undefined' ? args[number] : match;
  });
};
/**
 * Wrap words in string within template.
 *
 * @method highlight
 * @param {string[]} words - words to wrap
 * @param {string} [highlightTemplate="<b>{0}</b>"] - template for wrapping
 * @return {string}
 */
String.prototype.highlight = function (words, highlightTemplate) {
  var self = this;
  highlightTemplate = highlightTemplate ? highlightTemplate : "<b>{0}</b>";

  words.forEach(function (word) {
    var searchRegExp = new RegExp("\\b" + word + "\\b", "gi");
    self = self.replace(searchRegExp, function (found) {
      return highlightTemplate.format(found);
    });
  });

  return self;
};
/**
 * Convert time in milliseconds to object contained days, hours and minutes.
 * @typedef ConvertedTime
 *  @type {Object}
 *  @property {number} d - days
 *  @property {number} h - hours
 *  @property {string} m - minutes
 * @example
 *  var time = 1000000000;
 *  time.toDaysHoursMinutes() // {d: 11, h: 13, m: "46.67"}
 *
 * @method toDaysHoursMinutes
 * @return {object}
 */
Number.prototype.toDaysHoursMinutes = function () {
  var formatted = {},
    dateDiff = this,
    secK = 1000, //ms
    minK = 60 * secK, // sec
    hourK = 60 * minK, // sec
    dayK = 24 * hourK;

  dateDiff = parseInt(dateDiff);
  formatted.d = Math.floor(dateDiff / dayK);
  dateDiff -= formatted.d * dayK;
  formatted.h = Math.floor(dateDiff / hourK);
  dateDiff -= formatted.h * hourK;
  formatted.m = (dateDiff / minK).toFixed(2);

  return formatted;
};
/**
 Sort an array by the key specified in the argument.
 Handle only native js objects as element of array, not the Ember's object.

 Can be used as alternative to sortProperty method of Ember library
 in order to speed up executing on large data volumes

 @method sortBy
 @param {String} path name(s) to sort on
 @return {Array} The sorted array.
 */
Array.prototype.sortPropertyLight = function (path) {
  var realPath = (typeof path === "string") ? path.split('.') : [];
  this.sort(function (a, b) {
    var aProperty = a;
    var bProperty = b;
    realPath.forEach(function (key) {
      aProperty = aProperty[key];
      bProperty = bProperty[key];
    });
    if (aProperty > bProperty) return 1;
    if (aProperty < bProperty) return -1;
    return 0;
  });
  return this;
};
/** @namespace Em **/
Em.CoreObject.reopen({
  t:function (key, attrs) {
    return Em.I18n.t(key, attrs)
  }
});

Em.TextArea.reopen(Em.I18n.TranslateableAttributes);

/** @namespace Em.Handlebars **/
Em.Handlebars.registerHelper('log', function (variable) {
  console.log(variable);
});

Em.Handlebars.registerHelper('warn', function (variable) {
  console.warn(variable);
});

Em.Handlebars.registerHelper('highlight', function (property, words, fn) {
  var context = (fn.contexts && fn.contexts[0]) || this;
  property = Em.Handlebars.getPath(context, property, fn);

  words = words.split(";");

//  if (highlightTemplate == undefined) {
  var highlightTemplate = "<b>{0}</b>";
//  }

  words.forEach(function (word) {
    var searchRegExp = new RegExp("\\b" + word + "\\b", "gi");
    property = property.replace(searchRegExp, function (found) {
      return highlightTemplate.format(found);
    });
  });

  return new Em.Handlebars.SafeString(property);
});

Em.Handlebars.registerHelper('isAccessible', function (context, options) {
  if (App.isAccessible(context)) {
    return options.fn(this);
  }
});

/**
 * @namespace App
 */
App = require('app');

/**
 * Certain variables can have JSON in string
 * format, or in JSON format itself.
 *
 * @memberof App
 * @function parseJson
 * @param {string|object}
 * @return {object}
 */
App.parseJSON = function (value) {
  if (typeof value == "string") {
    return jQuery.parseJSON(value);
  }
  return value;
};
/**
 * Check for empty <code>Object</code>, built in Em.isEmpty()
 * doesn't support <code>Object</code> type
 *
 * @memberof App
 * @method isEmptyObject
 * @param obj {Object}
 * @return {Boolean}
 */
App.isEmptyObject = function(obj) {
  var empty = true;
  for (var prop in obj) { if (obj.hasOwnProperty(prop)) {empty = false; break;} }
  return empty;
}

/**
 * Convert object under_score keys to camelCase
 *
 * @param {Object} object
 * @return {Object}
 **/
App.keysUnderscoreToCamelCase = function(object) {
  var tmp = {};
  for (var key in object) {
    tmp[stringUtils.underScoreToCamelCase(key)] = object[key];
  }
  return tmp;
};

/**
 * Convert dotted keys to camelcase
 *
 * @param {Object} object
 * @return {Object}
 **/
App.keysDottedToCamelCase = function(object) {
  var tmp = {};
  for (var key in object) {
    tmp[key.split('.').reduce(function(p, c) { return p + c.capitalize()})] = object[key];
  }
  return tmp;
};
/**
 * Returns object with defined keys only.
 *
 * @memberof App
 * @method permit
 * @param {Object} obj - input object
 * @param {String|Array} keys - allowed keys
 * @return {Object}
 */
App.permit = function(obj, keys) {
  var result = {};
  if (typeof obj !== 'object' || App.isEmptyObject(obj)) return result;
  if (typeof keys == 'string') keys = Array(keys);
  keys.forEach(function(key) {
    if (obj.hasOwnProperty(key))
      result[key] = obj[key];
  });
  return result;
};
/**
 *
 * @namespace App
 * @namespace App.format
 */
App.format = {
  /**
   * @memberof App.format
   * @type {object}
   * @property components
   */
  components: {
    'API': 'API',
    'DECOMMISSION_DATANODE': 'Update Exclude File',
    'DRPC': 'DRPC',
    'FLUME_HANDLER': 'Flume',
    'GLUSTERFS': 'GLUSTERFS',
    'HBASE': 'HBase',
    'HBASE_REGIONSERVER': 'RegionServer',
    'HCAT': 'HCat Client',
    'HDFS': 'HDFS',
    'HISTORYSERVER': 'History Server',
    'HIVE_SERVER': 'HiveServer2',
    'JCE': 'JCE',
    'MAPREDUCE2': 'MapReduce2',
    'MYSQL': 'MySQL',
    'REST': 'REST',
    'SECONDARY_NAMENODE': 'SNameNode',
    'STORM_REST_API': 'Storm REST API Server',
    'WEBHCAT': 'WebHCat',
    'YARN': 'YARN',
    'UI': 'UI',
    'ZKFC': 'ZKFailoverController',
    'ZOOKEEPER': 'ZooKeeper',
    'ZOOKEEPER_QUORUM_SERVICE_CHECK': 'ZK Quorum Service Check'
  },

  /**
   * @memberof App.format
   * @property command
   * @type {object}
   */
  command: {
    'INSTALL': 'Install',
    'UNINSTALL': 'Uninstall',
    'START': 'Start',
    'STOP': 'Stop',
    'EXECUTE': 'Execute',
    'ABORT': 'Abort',
    'UPGRADE': 'Upgrade',
    'RESTART': 'Restart',
    'SERVICE_CHECK': 'Check',
    'Excluded:': 'Decommission:',
    'Included:': 'Recommission:'
  },

  /**
   * convert role to readable string
   *
   * @memberof App.format
   * @method role
   * @param {string} role
   * return {string}
   */
  role:function (role) {
    var result;
    var models = [App.StackService, App.StackServiceComponent];
    models.forEach(function(model){
      var instance =  model.find().findProperty('id',role);
      if (instance) {
        result = instance.get('displayName');
      }
    },this);
    if (!result)  {
      result =  this.normalizeName(role);
    }
    return result;
  },

  /**
   * Try to format non predefined names to readable format.
   *
   * @method normalizeName
   * @param name {String} - name to format
   * @return {String}
   */
  normalizeName: function(name) {
    if (!name || typeof name != 'string') return '';
    if (this.components[name]) return this.components[name];
    name = name.toLowerCase();
    var suffixNoSpaces = ['node','tracker','manager'];
    var suffixRegExp = new RegExp('(\\w+)(' + suffixNoSpaces.join('|') + ')', 'gi');
    if (/_/g.test(name)) {
      name = name.split('_').map(function(singleName) {
        return this.normalizeName(singleName.toUpperCase());
      }, this).join(' ');
    } else if(suffixRegExp.test(name)) {
      suffixRegExp.lastIndex = 0;
      var matches = suffixRegExp.exec(name);
      name = matches[1].capitalize() + matches[2].capitalize();
    }
    return name.capitalize();
  },

  /**
   * convert command_detail to readable string, show the string for all tasks name
   *
   * @memberof App.format
   * @method commandDetail
   * @param {string} command_detail
   * @param {string} request_inputs
   * @return {string}
   */
  commandDetail: function (command_detail, request_inputs) {
    var detailArr = command_detail.split(' ');
    var self = this;
    var result = '';
    detailArr.forEach( function(item) {
      // if the item has the pattern SERVICE/COMPONENT, drop the SERVICE part
      if (item.contains('/')) {
        item = item.split('/')[1];
      }
      if (item == 'DECOMMISSION,') {
        // ignore text 'DECOMMISSION,'( command came from 'excluded/included'), here get the component name from request_inputs
        item = (jQuery.parseJSON(request_inputs)) ? jQuery.parseJSON(request_inputs).slave_type : '';
      }
      if (self.components[item]) {
        result = result + ' ' + self.components[item];
      } else if (self.command[item]) {
        result = result + ' ' + self.command[item];
      } else {
        result = result + ' ' + self.role(item);
      }
    });

    if (result.indexOf('Decommission:') > -1 || result.indexOf('Recommission:') > -1) {
      // for Decommission command, make sure the hostname is in lower case
       result = result.split(':')[0] + ': ' + result.split(':')[1].toLowerCase();
    }
    //TODO check if UI use this
    if (result === ' Nagios Update Ignore Actionexecute') {
       result = Em.I18n.t('common.maintenance.task');
    }
    if (result.indexOf('Install Packages Actionexecute') != -1) {
      result = Em.I18n.t('common.installRepo.task');
    }
    if (result === ' Rebalancehdfs NameNode') {
       result = Em.I18n.t('services.service.actions.run.rebalanceHdfsNodes.title');
    }
    return result;
  },

  /**
   * Convert uppercase status name to lowercase.
   * <br>
   * <br>PENDING - Not queued yet for a host
   * <br>QUEUED - Queued for a host
   * <br>IN_PROGRESS - Host reported it is working
   * <br>COMPLETED - Host reported success
   * <br>FAILED - Failed
   * <br>TIMEDOUT - Host did not respond in time
   * <br>ABORTED - Operation was abandoned
   *
   * @memberof App.format
   * @method taskStatus
   * @param {string} _taskStatus
   * @return {string}
   *
   */
  taskStatus:function (_taskStatus) {
    return _taskStatus.toLowerCase();
  },

  /**
   * simplify kdc error msg
   * @param {string} message
   * @param {boolean} strict if this flag is true ignore not defined msgs return null
   *  else return input msg as is;
   * @returns {*}
   */
  kdcErrorMsg: function(message, strict) {
    /**
     * Error messages for KDC administrator credentials error
     * is used for checking if error message is caused by bad KDC credentials
     * @type {{missingKDC: string, invalidKDC: string}}
     */
    var specialMsg = {
      "missingKDC": "Missing KDC administrator credentials.",
      "invalidKDC": "Invalid KDC administrator credentials.",
      "missingRDCForRealm": "Failed to find a KDC for the specified realm - kadmin"
    };

    for (var m in specialMsg) {
      if (specialMsg.hasOwnProperty(m) && message.contains(specialMsg[m]))
        return specialMsg[m];
    }
    return strict ? null : message;
  }
};

/**
 * wrapper to bootstrap popover
 * fix issue when popover stuck on view routing
 *
 * @memberof App
 * @method popover
 * @param {DOMElement} self
 * @param {object} options
 */
App.popover = function (self, options) {
  self.popover(options);
  self.on("remove", function () {
    $(this).trigger('mouseleave');
  });
};

/**
 * wrapper to bootstrap tooltip
 * fix issue when tooltip stuck on view routing
 * @memberof App
 * @method tooltip
 * @param {DOMElement} self
 * @param {object} options
 */
App.tooltip = function (self, options) {
  self.tooltip(options);
  /* istanbul ignore next */
  self.on("remove", function () {
    $(this).trigger('mouseleave');
  });
};

/**
 * wrapper to Date().getTime()
 * fix issue when client clock and server clock not sync
 *
 * @memberof App
 * @method dateTime
 * @return {Number} timeStamp of current server clock
 */
App.dateTime = function() {
  return new Date().getTime() + App.clockDistance;
};

/**
 * Helper function for bound property helper registration
 * @memberof App
 * @method registerBoundHelper
 * @param name {String} name of helper
 * @param view {Em.View} view
 */
App.registerBoundHelper = function(name, view) {
  Em.Handlebars.registerHelper(name, function(property, options) {
    options.hash.contentBinding = property;
    return Em.Handlebars.helpers.view.call(this, view, options);
  });
};

/*
 * Return singular or plural word based on Em.I18n, view|controller context property key.
 *
 *  Example: {{pluralize hostsCount singular="t:host" plural="t:hosts"}}
 *           {{pluralize hostsCount singular="@view.hostName"}}
 */
App.registerBoundHelper('pluralize', Em.View.extend({
  tagName: 'span',
  template: Em.Handlebars.compile('{{view.wordOut}}'),

  wordOut: function() {
    var count, singular, plural;
    count = this.get('content');
    singular = this.get('singular');
    plural = this.get('plural');
    return this.getWord(count, singular, plural);
  }.property('content'),
  /**
   * Get computed word.
   *
   * @param {Number} count
   * @param {String} singular
   * @param {String} [plural]
   * @return {String}
   * @method getWord
   */
  getWord: function(count, singular, plural) {
    singular = this.parseValue(singular);
    // if plural not passed
    if (!plural) plural = singular + 's';
    else plural = this.parseValue(plural);
    if (singular && plural) {
      if (count == 1) {
        return singular;
      } else {
        return plural;
      }
    }
    return '';
  },
  /**
   * Detect and return value from its instance.
   *
   * @param {String} value
   * @return {*}
   * @method parseValue
   **/
  parseValue: function(value) {
    switch (value[0]) {
      case '@':
        value = this.getViewPropertyValue(value);
        break;
      case 't':
        value = this.tDetect(value);
        break;
      default:
        break;
    }
    return value;
  },
  /*
   * Detect for Em.I18n.t reference call
   * @params word {String}
   * return {String}
  */
  tDetect: function(word) {
    var splitted = word.split(':');
    if (splitted.length > 1 && splitted[0] == 't') {
      return Em.I18n.t(splitted[1]);
    } else {
      return splitted[0];
    }
  },
  /**
   * Get property value from view|controller by its key path.
   *
   * @param {String} value - key path
   * @return {*}
   * @method getViewPropertyValue
   **/
  getViewPropertyValue: function(value) {
    value = value.substr(1);
    var keyword = value.split('.')[0]; // return 'controller' or 'view'
    switch (keyword) {
      case 'controller':
        return Em.get(this, value);
      case 'view':
        return Em.get(this, value.replace(/^view/, 'parentView'))
      default:
        break;
    }
  }
  })
);
/**
 * Return defined string instead of empty if value is null/undefined
 * by default is `n/a`.
 *
 * @param empty {String} - value instead of empty string (not required)
 *  can be used with Em.I18n pass value started with't:'
 *
 * Examples:
 *
 * default value will be returned
 * {{formatNull service.someValue}}
 *
 * <code>empty<code> will be returned
 * {{formatNull service.someValue empty="I'm empty"}}
 *
 * Em.I18n translation will be returned
 * {{formatNull service.someValue empty="t:my.key.to.translate"
 */
App.registerBoundHelper('formatNull', Em.View.extend({
  tagName: 'span',
  template: Em.Handlebars.compile('{{view.result}}'),

  result: function() {
    var emptyValue = this.get('empty') ? this.get('empty') : Em.I18n.t('services.service.summary.notAvailable');
    emptyValue = emptyValue.startsWith('t:') ? Em.I18n.t(emptyValue.substr(2, emptyValue.length)) : emptyValue;
    return (this.get('content') || this.get('content') == 0) ? this.get('content') : emptyValue;
  }.property('content')
}));

/**
 * Return formatted string with inserted <code>wbr</code>-tag after each dot
 *
 * @param {String} content
 *
 * Examples:
 *
 * returns 'apple'
 * {{formatWordBreak 'apple'}}
 *
 * returns 'apple.<wbr />banana'
 * {{formatWordBreak 'apple.banana'}}
 *
 * returns 'apple.<wbr />banana.<wbr />uranium'
 * {{formatWordBreak 'apple.banana.uranium'}}
 */
App.registerBoundHelper('formatWordBreak', Em.View.extend({
  tagName: 'span',
  template: Em.Handlebars.compile('{{{view.result}}}'),

  /**
   * @type {string}
   */
  result: function() {
    return this.get('content') && this.get('content').replace(/\./g, '.<wbr />');
  }.property('content')
}));

/**
 * Return <i></i> with class that correspond to status
 *
 * @param {string} content - status
 *
 * Examples:
 *
 * {{statusIcon view.status}}
 * returns 'icon-cog'
 *
 */
App.registerBoundHelper('statusIcon', Em.View.extend({
  tagName: 'i',

  /**
   * relation map between status and icon class
   * @type {object}
   */
  statusIconMap: {
    'COMPLETED': 'icon-ok completed',
    'WARNING': 'icon-warning-sign',
    'FAILED': 'icon-exclamation-sign failed',
    'HOLDING_FAILED': 'icon-exclamation-sign failed',
    'PENDING': 'icon-cog pending',
    'QUEUED': 'icon-cog queued',
    'IN_PROGRESS': 'icon-cogs in_progress',
    'HOLDING': 'icon-pause',
    'ABORTED': 'icon-minus aborted',
    'TIMEDOUT': 'icon-time timedout',
    'HOLDING_TIMEDOUT': 'icon-time timedout'
  },

  classNameBindings: ['iconClass'],
  /**
   * @type {string}
   */
  iconClass: function () {
    return this.get('statusIconMap')[this.get('content')] || 'icon-question-sign';
  }.property('content')
}));

/**
 * Ambari overrides the default date transformer.
 * This is done because of the non-standard data
 * sent. For example Nagios sends date as "12345678".
 * The problem is that it is a String and is represented
 * only in seconds whereas Javascript's Date needs
 * milliseconds representation.
 */
DS.attr.transforms.date = {
  from: function (serialized) {
    var type = typeof serialized;
    if (type === Em.I18n.t('common.type.string')) {
      serialized = parseInt(serialized);
      type = typeof serialized;
    }
    if (type === Em.I18n.t('common.type.number')) {
      if (!serialized ){  //serialized timestamp = 0;
        return 0;
      }
      // The number could be seconds or milliseconds.
      // If seconds, then the length is 10
      // If milliseconds, the length is 13
      if (serialized.toString().length < 13) {
        serialized = serialized * 1000;
      }
      return new Date(serialized);
    } else if (serialized === null || serialized === undefined) {
      // if the value is not present in the data,
      // return undefined, not null.
      return serialized;
    } else {
      return null;
    }
  },
  to: function (deserialized) {
    if (deserialized instanceof Date) {
      return deserialized.getTime();
    } else if (deserialized === undefined) {
      return undefined;
    } else {
      return null;
    }
  }
};

DS.attr.transforms.object = {
  from: function(serialized) {
    return Ember.none(serialized) ? null : Object(serialized);
  },

  to: function(deserialized) {
    return Ember.none(deserialized) ? null : Object(deserialized);
  }
};

/**
 * Allows EmberData models to have array properties.
 *
 * Declare the property as <code>
 *  operations: DS.attr('array'),
 * </code> and
 * during load provide a JSON array for value.
 *
 * This transform simply assigns the same array in both directions.
 */
DS.attr.transforms.array = {
  from : function(serialized) {
    return serialized;
  },
  to : function(deserialized) {
    return deserialized;
  }
};

/**
 *  Utility method to delete all existing records of a DS.Model type from the model's associated map and
 *  store's persistence layer (recordCache)
 * @param type DS.Model Class
 */
App.resetDsStoreTypeMap = function(type) {
  var allRecords = App.get('store.recordCache');  //This fetches all records in the ember-data persistence layer
  var typeMaps = App.get('store.typeMaps');
  var guidForType = Em.guidFor(type);
  var typeMap = typeMaps[guidForType];
  if (typeMap) {
    var idToClientIdMap = typeMap.idToCid;
    for (var id in idToClientIdMap) {
      if (idToClientIdMap.hasOwnProperty(id) && idToClientIdMap[id]) {
        delete allRecords[idToClientIdMap[id]];  // deletes the cached copy of the record from the store
      }
    }
    typeMaps[guidForType] = {
      idToCid: {},
      clientIds: [],
      cidToHash: {},
      recordArrays: []
    };
  }
};

});

require.register("utils/host_progress_popup", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var batchUtils = require('utils/batch_scheduled_requests');
var date = require('utils/date');

/**
 * App.HostPopup is for the popup that shows up upon clicking already-performed or currently-in-progress operations
 */
App.HostPopup = Em.Object.create({

  name: 'hostPopup',

  servicesInfo: [],
  hosts: null,
  inputData: null,

  /**
   * @type {string}
   */
  serviceName: "",

  /**
   * @type {Number}
   */
  currentServiceId: null,
  previousServiceId: null,

  /**
   * @type {string}
   */
  popupHeaderName: "",

  operationInfo: null,
  /**
   * @type {App.Controller}
   */
  dataSourceController: null,

  /**
   * @type {bool}
   */
  isBackgroundOperations: false,

  /**
   * @type {string}
   */
  currentHostName: null,

  /**
   * @type {App.ModalPopup}
   */
  isPopup: null,

  detailedProperties: {
    stdout: 'stdout',
    stderr: 'stderr',
    outputLog: 'output_log',
    errorLog: 'error_log'
  },

  abortIcon: Em.View.extend({
    tagName: 'i',
    classNames: ['abort-icon', 'icon-remove-circle', 'pointer'],
    click: function () {
      this.get('controller').abortRequest(this.get('servicesInfo'));
      return false;
    },
    didInsertElement: function () {
      App.tooltip($(this.get('element')), {
        placement: "top",
        title: Em.I18n.t('hostPopup.bgop.abortRequest.title')
      });
    }
  }),

  statusIcon: Em.View.extend({
    tagName: 'i',
    classNames: ["service-status"],
    classNameBindings: ['servicesInfo.status', 'servicesInfo.icon', 'additionalClass'],
    attributeBindings: ['data-original-title'],
    'data-original-title': function() {
      return this.get('servicesInfo.status');
    }.property('servicesInfo.status'),
    didInsertElement: function () {
      App.tooltip($(this.get('element')));
    }
  }),

  /**
   * Determines if background operation can be aborted depending on its status
   * @param status
   * @returns {boolean}
   */
  isAbortableByStatus: function (status) {
    var statuses = this.get('statusesStyleMap');
    return !Em.keys(statuses).contains(status) || status == 'IN_PROGRESS';
  },

  /**
   * Send request to abort operation
   */
  abortRequest: function (serviceInfo) {
    var requestName = serviceInfo.get('name');
    var self = this;
    App.showConfirmationPopup(function () {
      serviceInfo.set('isAbortable', false);
      App.ajax.send({
        name: 'background_operations.abort_request',
        sender: self,
        data: {
          requestId: serviceInfo.get('id'),
          requestName: requestName,
          serviceInfo: serviceInfo
        },
        success: 'abortRequestSuccessCallback',
        error: 'abortRequestErrorCallback'
      });
    }, Em.I18n.t('hostPopup.bgop.abortRequest.confirmation.body').format(requestName));
    return false;
  },

  /**
   * Method called on successful sending request to abort operation
   */
  abortRequestSuccessCallback: function (response, request, data) {
    App.ModalPopup.show({
      header: Em.I18n.t('hostPopup.bgop.abortRequest.modal.header'),
      bodyClass: Em.View.extend({
        template: Em.Handlebars.compile(Em.I18n.t('hostPopup.bgop.abortRequest.modal.body').format(data.requestName))
      }),
      secondary: null
    });
  },

  /**
   * Method called on unsuccessful sending request to abort operation
   */
  abortRequestErrorCallback: function (xhr, textStatus, error, opt, data) {
    data.serviceInfo.set('isAbortable', this.isAbortableByStatus(data.serviceInfo.status));
    App.ajax.defaultErrorHandler(xhr, opt.url, 'PUT', xhr.status);
  },
  /**
   * Entering point of this component
   * @param {String} serviceName
   * @param {Object} controller
   * @param {Boolean} isBackgroundOperations
   * @param {Integer} requestId
   */
  initPopup: function (serviceName, controller, isBackgroundOperations, requestId) {
    if (!isBackgroundOperations) {
      this.clearHostPopup();
      this.set("popupHeaderName", serviceName);
    }

    this.set('currentServiceId', requestId);
    this.set("serviceName", serviceName);
    this.set("dataSourceController", controller);
    this.set("isBackgroundOperations", isBackgroundOperations);
    this.set("inputData", this.get("dataSourceController.services"));
    if(isBackgroundOperations){
      this.onServiceUpdate();
    } else {
      this.onHostUpdate();
    }
    return this.createPopup();
  },

  /**
   * clear info popup data
   */
  clearHostPopup: function () {
    this.set('servicesInfo', []);
    this.set('hosts', null);
    this.set('inputData', null);
    this.set('serviceName', "");
    this.set('currentServiceId', null);
    this.set('previousServiceId', null);
    this.set('popupHeaderName', "");
    this.set('dataSourceController', null);
    this.set('currentHostName', null);
    this.get('isPopup') ? this.get('isPopup').remove() : null;
  },

  /**
   * Depending on tasks status
   * @param {Array} tasks
   * @return {Array} [Status, Icon type, Progressbar color, is IN_PROGRESS]
   */
  getStatus: function(tasks){
    var isCompleted = true;
    var status;
    var tasksLength = tasks.length;
    var isFailed = false;
    var isAborted = false;
    var isTimedout = false;
    var isInProgress = false;
    for (var i = 0; i < tasksLength; i++) {
      if (tasks[i].Tasks.status !== 'COMPLETED') {
        isCompleted = false;
      }
      if(tasks[i].Tasks.status === 'FAILED'){
        isFailed = true;
      }
      if (tasks[i].Tasks.status === 'ABORTED') {
        isAborted = true;
      }
      if (tasks[i].Tasks.status === 'TIMEDOUT') {
        isTimedout = true;
      }
      if (tasks[i].Tasks.status === 'IN_PROGRESS') {
        isInProgress = true;
      }
    }
    if (isFailed) {
      status = ['FAILED', 'icon-exclamation-sign', 'progress-danger', false];
    } else if (isAborted) {
      status = ['ABORTED', 'icon-minus', 'progress-warning', false];
    } else if (isTimedout) {
      status = ['TIMEDOUT', 'icon-time', 'progress-warning', false];
    } else if (isInProgress) {
      status = ['IN_PROGRESS', 'icon-cogs', 'progress-info', true];
    }
    if(status){
      return status;
    } else if(isCompleted){
      return ['SUCCESS', 'icon-ok', 'progress-success', false];
    } else {
      return ['PENDING', 'icon-cog', 'progress-info', true];
    }
  },

  /**
   * Progress of host or service depending on tasks status
   * @param {Array} tasks
   * @return {Number} percent of completion
   */
  getProgress: function (tasks) {
    if (!tasks || tasks.length === 0) return 0;

    var completedActions = 0;
    var queuedActions = 0;
    var inProgressActions = 0;

    tasks.forEach(function (task) {
      if (['COMPLETED', 'FAILED', 'ABORTED', 'TIMEDOUT'].contains(task.Tasks.status)) {
        completedActions++;
      } else if (task.Tasks.status === 'QUEUED') {
        queuedActions++;
      } else if (task.Tasks.status === 'IN_PROGRESS') {
        inProgressActions++;
      }
    });
    return Math.ceil(((queuedActions * 0.09) + (inProgressActions * 0.35) + completedActions ) / tasks.length * 100);
  },

  /**
   * Count number of operations for select box options
   * @param {Object[]} obj
   * @param {Object[]} categories
   */
  setSelectCount: function (obj, categories) {
    if (!obj) return;
    var countAll = obj.length;
    var countPending = 0;
    var countInProgress = 0;
    var countFailed = 0;
    var countCompleted = 0;
    var countAborted = 0;
    var countTimedout = 0;
    obj.forEach(function(item){
      switch (item.status){
        case 'pending':
          countPending++;
          break;
        case 'queued':
          countPending++;
          break;
        case 'in_progress':
          countInProgress++;
          break;
        case 'failed':
          countFailed++;
          break;
        case 'success':
          countCompleted++;
          break;
        case 'completed':
          countCompleted++;
          break;
        case 'aborted':
          countAborted++;
          break;
        case 'timedout':
          countTimedout++;
          break;
      }
    });

    categories.findProperty("value", 'all').set("count", countAll);
    categories.findProperty("value", 'pending').set("count", countPending);
    categories.findProperty("value", 'in_progress').set("count", countInProgress);
    categories.findProperty("value", 'failed').set("count", countFailed);
    categories.findProperty("value", 'completed').set("count", countCompleted);
    categories.findProperty("value", 'aborted').set("count", countAborted);
    categories.findProperty("value", 'timedout').set("count", countTimedout);
  },

  /**
   * For Background operation popup calculate number of running Operations, and set popup header
   * @param {bool} isServiceListHidden
   */
  setBackgroundOperationHeader: function (isServiceListHidden) {
    if (this.get('isBackgroundOperations') && !isServiceListHidden) {
      var numRunning =  App.router.get('backgroundOperationsController.allOperationsCount');
      this.set("popupHeaderName", numRunning + Em.I18n.t('hostPopup.header.postFix').format(numRunning == 1 ? "" : "s"));
    }
  },

  // map to get css class with styles by service status
  statusesStyleMap: {
    'FAILED': ['FAILED', 'icon-exclamation-sign', 'progress-danger', false],
    'ABORTED': ['ABORTED', 'icon-minus', 'progress-warning', false],
    'TIMEDOUT': ['TIMEDOUT', 'icon-time', 'progress-warning', false],
    'IN_PROGRESS': ['IN_PROGRESS', 'icon-cogs', 'progress-info', true],
    'COMPLETED': ['SUCCESS', 'icon-ok', 'progress-success', false]
  },

  /**
   * Create services obj data structure for popup
   * Set data for services
   * @param {bool} isServiceListHidden
   */
  onServiceUpdate: function (isServiceListHidden) {
    if (this.get('isBackgroundOperations') && this.get("inputData")) {
      var statuses = this.get('statusesStyleMap');
      var servicesInfo = this.get("servicesInfo");
      var currentServices = [];
      this.get("inputData").forEach(function (service, index) {
        var updatedService;
        var id = service.id;
        currentServices.push(id);
        var existedService = servicesInfo.findProperty('id', id);
        updatedService = existedService;
        if (existedService) {
          updatedService = this.updateService(existedService, service);
        } else {
          updatedService = this.createService(service);
          servicesInfo.insertAt(index, updatedService);
        }
        updatedService.set('isAbortable',  App.isAccessible('MANAGER') &&  this.isAbortableByStatus(service.status));
      }, this);
      this.removeOldServices(servicesInfo, currentServices);
      this.setBackgroundOperationHeader(isServiceListHidden);
    }
  },

  /**
   * Create service object from transmitted data
   * @param service
   */
  createService: function (service) {
    var statuses = this.get('statusesStyleMap');
    var pendingStatus = ['PENDING', 'icon-cog', 'progress-info', true];
    var status = statuses[service.status] || pendingStatus;
    return Ember.Object.create({
      id: service.id,
      displayName: service.displayName,
      progress: service.progress,
      status: App.format.taskStatus(status[0]),
      isRunning: service.isRunning,
      name: service.name,
      isVisible: true,
      startTime: date.startTime(service.startTime),
      duration: date.durationSummary(service.startTime, service.endTime),
      icon: status[1],
      barColor: status[2],
      isInProgress: status[3],
      barWidth: "width:" + service.progress + "%;",
      sourceRequestScheduleId: service.get('sourceRequestScheduleId'),
      contextCommand: service.get('contextCommand')
    });
  },

  /**
   * Update properties of existed service with new data
   * @param service
   * @param newData
   * @returns {Ember.Object}
   */
  updateService: function (service, newData) {
    var statuses = this.get('statusesStyleMap');
    var pendingStatus = ['PENDING', 'icon-cog', 'progress-info', true];
    var status = statuses[newData.status] || pendingStatus;
    return service.setProperties({
      progress: newData.progress,
      status: App.format.taskStatus(status[0]),
      isRunning: newData.isRunning,
      startTime: date.startTime(newData.startTime),
      duration: date.durationSummary(newData.startTime, newData.endTime),
      icon: status[1],
      barColor: status[2],
      isInProgress: status[3],
      barWidth: "width:" + newData.progress + "%;",
      sourceRequestScheduleId: newData.get('sourceRequestScheduleId'),
      contextCommand: newData.get('contextCommand')
    });
  },

  /**
   * remove old requests
   * as API returns 10, or  20 , or 30 ...etc latest request, the requests that absent in response should be removed
   * @param services
   * @param currentServicesIds
   */
  removeOldServices: function (services, currentServicesIds) {
    services.forEach(function (service, index, services) {
      if (!currentServicesIds.contains(service.id)) {
        services.removeAt(index, 1);
      }
    });
  },

  /**
   * create task Ember object
   * @param {Object} _task
   * @return {Em.Object}
   */
  createTask: function (_task) {
    return Em.Object.create({
      id: _task.Tasks.id,
      hostName: _task.Tasks.host_name,
      command: ( _task.Tasks.command.toLowerCase() != 'service_check') ? _task.Tasks.command.toLowerCase() : '',
      commandDetail: App.format.commandDetail(_task.Tasks.command_detail, _task.Tasks.request_inputs),
      status: App.format.taskStatus(_task.Tasks.status),
      role: App.format.role(_task.Tasks.role),
      stderr: _task.Tasks.stderr,
      stdout: _task.Tasks.stdout,
      request_id: _task.Tasks.request_id,
      isVisible: true,
      startTime: date.startTime(_task.Tasks.start_time),
      duration: date.durationSummary(_task.Tasks.start_time, _task.Tasks.end_time),
      icon: function () {
        var statusIconMap = {
          'pending': 'icon-cog',
          'queued': 'icon-cog',
          'in_progress': 'icon-cogs',
          'completed': 'icon-ok',
          'failed': 'icon-exclamation-sign',
          'aborted': 'icon-minus',
          'timedout': 'icon-time'
        };
        return statusIconMap[this.get('status')] || 'icon-cog';
      }.property('status')
    });
  },

  /**
   * Create hosts and tasks data structure for popup
   * Set data for hosts and tasks
   */
  onHostUpdate: function () {
    var self = this;
    var inputData = this.get("inputData");
    if (inputData) {
      var hostsArr = [];
      var hostsData;
      var hostsMap = {};

      if(this.get('isBackgroundOperations') && this.get("currentServiceId")){
        //hosts popup for Background Operations
        hostsData = inputData.findProperty("id", this.get("currentServiceId"));
      } else if (this.get("serviceName")) {
        //hosts popup for Wizards
        hostsData = inputData.findProperty("name", this.get("serviceName"));
      }
      if (hostsData) {
        if (hostsData.hostsMap) {
          //hosts data come from Background Operations as object map
          hostsMap = hostsData.hostsMap;
        } else if (hostsData.hosts) {
          //hosts data come from Wizard as array
          hostsData.hosts.forEach(function (_host) {
            hostsMap[_host.name] = _host;
          });
        }
      }
      var existedHosts = self.get('hosts');

      if (existedHosts && (existedHosts.length > 0) && this.get('currentServiceId') === this.get('previousServiceId')) {
        existedHosts.forEach(function (host) {
          var newHostInfo = hostsMap[host.get('name')];
          //update only hosts with changed tasks or currently opened tasks of host
          if (newHostInfo && (!this.get('isBackgroundOperations') || newHostInfo.isModified || this.get('currentHostName') === host.get('name'))) {
            var hostStatus = self.getStatus(newHostInfo.logTasks);
            var hostProgress = self.getProgress(newHostInfo.logTasks);
            host.set('status', App.format.taskStatus(hostStatus[0]));
            host.set('icon', hostStatus[1]);
            host.set('barColor', hostStatus[2]);
            host.set('isInProgress', hostStatus[3]);
            host.set('progress', hostProgress);
            host.set('barWidth', "width:" + hostProgress + "%;");
            host.set('logTasks', newHostInfo.logTasks);
            var existTasks = host.get('tasks');
            if (existTasks) {
              newHostInfo.logTasks.forEach(function (_task) {
                var existTask = existTasks.findProperty('id', _task.Tasks.id);
                if (existTask) {
                  var status = _task.Tasks.status;
                  existTask.set('status', App.format.taskStatus(status));
                  Em.keys(this.get('detailedProperties')).forEach(function (key) {
                    var value = _task.Tasks[this.get('detailedProperties')[key]];
                    if (!Em.isNone(value)) {
                      existTask.set(key, value);
                    }
                  }, this);
                  existTask.set('startTime', date.startTime(_task.Tasks.start_time));
                  existTask.set('duration', date.durationSummary(_task.Tasks.start_time, _task.Tasks.end_time));
                  // Puts some command information to render it
                  var isRebalanceHDFSTask = (_task.Tasks.command === 'CUSTOM_COMMAND' && _task.Tasks.custom_command_name === 'REBALANCEHDFS');
                  existTask.set('isRebalanceHDFSTask', isRebalanceHDFSTask);
                  if(isRebalanceHDFSTask){
                    var structuredOut = _task.Tasks.structured_out;
                    if (!structuredOut || structuredOut === 'null') {
                      structuredOut = {};
                    }

                    var barColorMap = {
                      'FAILED': 'progress-danger',
                      'ABORTED': 'progress-warning',
                      'TIMEDOUT': 'progress-warning',
                      'IN_PROGRESS': 'progress-info',
                      'COMPLETED': 'progress-success'
                    };

                    existTask.set('dataMoved', structuredOut['dataMoved'] || '0');
                    existTask.set('dataLeft', structuredOut['dataLeft'] || '0');
                    existTask.set('dataBeingMoved', structuredOut['dataBeingMoved'] || '0');
                    existTask.set('barColor', barColorMap[status]);
                    existTask.set('isInProgress', status == 'IN_PROGRESS');
                    existTask.set('isNotComplete', ['QUEUED', 'IN_PROGRESS'].contains(status));
                    existTask.set('completionProgressStyle', 'width:' + (structuredOut['completePercent'] || 0) * 100 + '%;');

                    existTask.set('command', _task.Tasks.command);
                    existTask.set('custom_command_name', _task.Tasks.custom_command_name);
                  }
                } else {
                  existTasks.pushObject(this.createTask(_task));
                }
              }, this);
            }
          }
        }, this);
      } else {
        for (var hostName in hostsMap) {
          var _host = hostsMap[hostName];
          var tasks = _host.logTasks;
          var hostInfo = Ember.Object.create({
            name: hostName,
            publicName: _host.publicName,
            displayName: function () {
              return this.get('name').length < 43 ? this.get('name') : (this.get('name').substr(0, 40) + '...');
            }.property('name'),
            progress: 0,
            status: App.format.taskStatus("PENDING"),
            serviceName: _host.serviceName,
            isVisible: true,
            icon: "icon-cog",
            barColor: "progress-info",
            barWidth: "width:0%;"
          });

          if (tasks.length) {
            tasks = tasks.sortProperty('Tasks.id');
            var hostStatus = self.getStatus(tasks);
            var hostProgress = self.getProgress(tasks);
            hostInfo.set('status', App.format.taskStatus(hostStatus[0]));
            hostInfo.set('icon', hostStatus[1]);
            hostInfo.set('barColor', hostStatus[2]);
            hostInfo.set('isInProgress', hostStatus[3]);
            hostInfo.set('progress', hostProgress);
            hostInfo.set('barWidth', "width:" + hostProgress + "%;");
          }
          hostInfo.set('logTasks', tasks);
          hostsArr.push(hostInfo);
        }

        hostsArr = hostsArr.sortProperty('name');
        hostsArr.setEach("serviceName", this.get("serviceName"));
        self.set("hosts", hostsArr);
        self.set('previousServiceId', this.get('currentServiceId'));
      }
    }

    var operation = this.get('servicesInfo').findProperty('name', this.get('serviceName'));
    if (!operation || (operation && operation.get('progress') == 100)) {
      this.set('operationInfo', null);
    } else {
      this.set('operationInfo', operation);
    }
  },

  /**
   * Show popup
   * @return {App.ModalPopup} PopupObject For testing purposes
   */
  createPopup: function () {
    var self = this;
    var servicesInfo = this.get("servicesInfo");
    var isBackgroundOperations = this.get('isBackgroundOperations');
    var categoryObject = Em.Object.extend({
      value: '',
      count: 0,
      labelPath: '',
      label: function(){
        return Em.I18n.t(this.get('labelPath')).format(this.get('count'));
      }.property('count')
    });
    self.set('isPopup', App.ModalPopup.show({

      isLogWrapHidden: true,
      isTaskListHidden: true,
      isHostListHidden: true,
      isServiceListHidden: false,

      isHideBodyScroll: true,
      /**
       * no need to track is it loaded when popup contain only list of hosts
       * @type {bool}
       */
      isLoaded: !isBackgroundOperations,

      /**
       * is BG-popup opened
       * @type {bool}
       */
      isOpen: false,

      detailedProperties: self.get('detailedProperties'),

      didInsertElement: function(){
        this._super();
        this.set('isOpen', true);
      },

      /**
       * @type {Em.View}
       */
      headerClass: Em.View.extend({
        controller: this,
        template: Ember.Handlebars.compile('{{popupHeaderName}} ' +
            '{{#unless view.parentView.isHostListHidden}}{{#if controller.operationInfo.isAbortable}}' +
            '{{view controller.abortIcon servicesInfoBinding="controller.operationInfo"}}' +
            '{{/if}}{{/unless}}')
      }),

      /**
       * @type {String[]}
       */
      classNames: ['sixty-percent-width-modal', 'host-progress-popup'],

      /**
       * for the checkbox: do not show this dialog again
       * @type {bool}
       */
      hasFooterCheckbox: true,

      /**
       * Auto-display BG-popup
       * @type {bool}
       */
      isNotShowBgChecked : null,

      /**
       * Save user pref about auto-display BG-popup
       */
      updateNotShowBgChecked: function () {
        var curVal = !this.get('isNotShowBgChecked');
        var key = App.router.get('applicationController').persistKey();
        if (!App.get('testMode')) {
          App.router.get('applicationController').postUserPref(key, curVal);
        }
      }.observes('isNotShowBgChecked'),

      autoHeight: false,
      closeModelPopup: function () {
        this.set('isOpen', false);
        if(isBackgroundOperations){
          $(this.get('element')).detach();
          App.router.get('backgroundOperationsController').set('levelInfo.name', 'REQUESTS_LIST');
        } else {
          this.hide();
          self.set('isPopup', null);
        }
      },
      onPrimary: function () {
        this.closeModelPopup();
      },
      onClose: function () {
        this.closeModelPopup();
      },
      secondary: null,

      bodyClass: App.TableView.extend({
        templateName: require('templates/common/host_progress_popup'),
        showTextArea: false,
        isServiceEmptyList: true,
        isTasksEmptyList: true,
        controller: this,
        sourceRequestScheduleId: -1,
        sourceRequestScheduleRunning: false,
        sourceRequestScheduleAborted: false,
        sourceRequestScheduleCommand: null,
        hosts: self.get('hosts'),
        services: self.get('servicesInfo'),
        filterMap: {
          pending: ["pending", "queued"],
          in_progress: ["in_progress", "upgrading"],
          failed: ["failed"],
          completed: ["completed", "success"],
          aborted: ["aborted"],
          timedout: ["timedout"]
        },

        pagination: true,
        isPaginate: false,
        /**
         * Select box, display names and values
         */
        categories: [
          categoryObject.create({value: 'all', labelPath: 'hostPopup.status.category.all'}),
          categoryObject.create({value: 'pending', labelPath: 'hostPopup.status.category.pending'}),
          categoryObject.create({value: 'in_progress', labelPath: 'hostPopup.status.category.inProgress'}),
          categoryObject.create({value: 'failed', labelPath: 'hostPopup.status.category.failed'}),
          categoryObject.create({value: 'completed', labelPath: 'hostPopup.status.category.success'}),
          categoryObject.create({value: 'aborted', labelPath: 'hostPopup.status.category.aborted'}),
          categoryObject.create({value: 'timedout', labelPath: 'hostPopup.status.category.timedout'})
        ],

        /**
         * Selected option is bound to this values
         */
        serviceCategory: null,
        hostCategory: null,
        taskCategory: null,
        /**
         * flag to indicate whether level data has already been loaded
         * applied only to HOSTS_LIST and TASK_DETAILS levels, whereas async query used to obtain data
         */
        isLevelLoaded: true,
        isHostEmptyList: function() {
          return !this.get('pageContent.length');
        }.property('pageContent.length'),

        currentHost: function () {
          return this.get('hosts') && this.get('hosts').findProperty('name', this.get('controller.currentHostName'));
        }.property('controller.currentHostName'),

        tasks: function () {
          var currentHost = this.get('currentHost');
          if (currentHost) {
            return currentHost.get('tasks');
          }
          return [];
        }.property('currentHost.tasks', 'currentHost.tasks.@each.status'),

        willDestroyElement: function () {
          if (this.get('controller.dataSourceController.name') == 'highAvailabilityProgressPopupController') {
            this.set('controller.dataSourceController.isTaskPolling', false);
          }
        },

        /**
         * Preset values on init
         */
        setOnStart: function () {
          this.set('serviceCategory', this.get('categories').findProperty('value','all'));
          if (this.get("controller.isBackgroundOperations")) {
            this.get('controller').setSelectCount(this.get("services"), this.get('categories'));
            this.updateHostInfo();
          } else {
            this.set("parentView.isHostListHidden", false);
            this.set("parentView.isServiceListHidden", true);
          }
        },

        /**
         * force popup to show list of operations
         */
        resetState: function(){
          if(this.get('parentView.isOpen')){
            this.set('parentView.isLogWrapHidden', true);
            this.set('parentView.isTaskListHidden', true);
            this.set('parentView.isHostListHidden', true);
            this.set('parentView.isServiceListHidden', false);
            this.get("controller").setBackgroundOperationHeader(false);
            this.setOnStart();
          }
        }.observes('parentView.isOpen'),

        /**
         * When popup is opened, and data after polling has changed, update this data in component
         */
        updateHostInfo: function () {
          if(!this.get('parentView.isOpen')) return;
          this.set('parentView.isLoaded', false);
          this.get("controller").set("inputData", this.get("controller.dataSourceController.services"));
          this.get("controller").onServiceUpdate(this.get('parentView.isServiceListHidden'));
          this.get("controller").onHostUpdate();
          this.set('parentView.isLoaded', true);
          this.set("hosts", this.get("controller.hosts"));
          this.set("services", this.get("controller.servicesInfo"));
          this.set('isLevelLoaded', true);
        }.observes("controller.dataSourceController.serviceTimestamp"),

        /**
         * Depending on service filter, set which services should be shown
         */
        visibleServices: function () {
          if (this.get("services")) {
            this.set("isServiceEmptyList", true);
            if (this.get('serviceCategory.value')) {
              var filter = this.get('serviceCategory.value');
              var services = this.get('services');
              this.set("isServiceEmptyList", this.setVisibility(filter, services));
            }
          }
        }.observes('serviceCategory', 'services', 'services.@each.status'),

        /**
         * Depending on hosts filter, set which hosts should be shown
         */
        filter: function() {
          var _this = this,
              filter = this.get('hostCategory.value'),
              hosts = this.get('hosts') || [];
          if (!filter || !hosts.length) return;
          if (filter === 'all') {
            this.set('filteredContent', hosts);
          } else {
            this.set('filteredContent', hosts.filter(function(item) {
              return _this.get('filterMap')[filter].contains(item.status);
            }));
          }
        }.observes('hosts.length', 'hostCategory.value'),

        /**
         * Reset startIndex property back to 1 when filter type has been changed.
         */
        resetIndex: function() {
          if (this.get('hostCategory.value')) this.set('startIndex', 1)
        }.observes('hostCategory.value'),

        /**
         * Depending on tasks filter, set which tasks should be shown
         */
        visibleTasks: function () {
          this.set("isTasksEmptyList", true);
          if (this.get('taskCategory.value') && this.get('tasks')) {
            var filter = this.get('taskCategory.value');
            var tasks = this.get('tasks');
            this.set("isTasksEmptyList", this.setVisibility(filter, tasks));
          }
        }.observes('taskCategory', 'tasks', 'tasks.@each.status'),

        /**
         * Depending on selected filter type, set object visibility value
         * @param filter
         * @param obj
         * @return {bool} isEmptyList
         */
        setVisibility: function (filter, obj) {
          var isEmptyList = true;
          if (filter == "all") {
            obj.setEach("isVisible", true);
            isEmptyList = !(obj.length > 0);
          } else {
            obj.forEach(function(item){
              item.set('isVisible', this.get('filterMap')[filter].contains(item.status));
              isEmptyList = (isEmptyList) ? !item.get('isVisible') : false;
            }, this)
          }
          return isEmptyList;
        },

        /**
         * Depending on currently viewed tab, call setSelectCount function
         */
        updateSelectView: function () {
          var isPaginate;
          if (!this.get('parentView.isHostListHidden')) {
            //since lazy loading used for hosts, we need to get hosts info directly from controller, that always contains entire array of data
            this.get('controller').setSelectCount(this.get("controller.hosts"), this.get('categories'));
            isPaginate = true;
          } else if (!this.get('parentView.isTaskListHidden')) {
            this.get('controller').setSelectCount(this.get("tasks"), this.get('categories'));
          } else if (!this.get('parentView.isServiceListHidden')) {
            this.get('controller').setSelectCount(this.get("services"), this.get('categories'));
          }
          this.set('isPaginate', !!isPaginate);
        }.observes('tasks.@each.status', 'hosts.@each.status', 'parentView.isTaskListHidden', 'parentView.isHostListHidden', 'services.length', 'services.@each.status'),

        /**
         * control data uploading, depending on which display level is showed
         * @param levelName
         */
        switchLevel: function (levelName) {
          var dataSourceController = this.get('controller.dataSourceController');
          var securityControllers = [
            'mainAdminSecurityDisableController',
            'mainAdminSecurityAddStep4Controller'
          ];
          if (this.get("controller.isBackgroundOperations")) {
            var levelInfo = dataSourceController.get('levelInfo');
            levelInfo.set('taskId', this.get('openedTaskId'));
            levelInfo.set('requestId', this.get('controller.currentServiceId'));
            levelInfo.set('name', levelName);
            if (levelName === 'HOSTS_LIST') {
              this.set('isLevelLoaded', dataSourceController.requestMostRecent());
              this.set('hostCategory', this.get('categories').findProperty('value','all'));
            } else if (levelName === 'TASK_DETAILS') {
              dataSourceController.requestMostRecent();
              this.set('isLevelLoaded', false);
            } else if (levelName === 'REQUESTS_LIST') {
              this.set('serviceCategory', this.get('categories').findProperty('value','all'));
              this.get('controller.hosts').clear();
              dataSourceController.requestMostRecent();
            } else {
              this.set('taskCategory', this.get('categories').findProperty('value','all'));
            }
          } else if (securityControllers.contains(dataSourceController.get('name'))) {
            if (levelName === 'TASK_DETAILS') {
              this.set('isLevelLoaded', false);
              dataSourceController.startUpdatingTask(this.get('controller.currentServiceId'), this.get('openedTaskId'));
            } else {
              dataSourceController.stopUpdatingTask(this.get('controller.currentServiceId'));
            }
          } else if (dataSourceController.get('name') == 'highAvailabilityProgressPopupController') {
            if (levelName === 'TASK_DETAILS') {
              this.set('isLevelLoaded', false);
              dataSourceController.startTaskPolling(this.get('openedTask.request_id'), this.get('openedTask.id'));
              Em.keys(this.get('parentView.detailedProperties')).forEach(function (key) {
                dataSourceController.addObserver('taskInfo.' + this.get('parentView.detailedProperties')[key], this, 'updateTaskInfo');
              }, this);
            } else {
              dataSourceController.stopTaskPolling();
            }
          }
        },
        updateTaskInfo: function () {
          var dataSourceController = this.get('controller.dataSourceController');
          var openedTask = this.get('openedTask');
          if (openedTask && openedTask.get('id') == dataSourceController.get('taskInfo.id')) {
            this.set('isLevelLoaded', true);
            Em.keys(this.get('parentView.detailedProperties')).forEach(function (key) {
              openedTask.set(key, dataSourceController.get('taskInfo.' + key));
            }, this);
          }
        },
        /**
         * Onclick handler for button <-Tasks
         */
        backToTaskList: function () {
          this.destroyClipBoard();
          this.set("openedTaskId", 0);
          this.set("parentView.isLogWrapHidden", true);
          this.set("parentView.isTaskListHidden", false);
          this.switchLevel("TASKS_LIST");
        },

        /**
         * Onclick handler for button <-Hosts
         */
        backToHostList: function () {
          this.set("parentView.isHostListHidden", false);
          this.set("parentView.isTaskListHidden", true);
          this.get("controller").set("popupHeaderName", this.get("controller.serviceName"));
          this.get("controller").set("operationInfo", this.get('controller.servicesInfo').findProperty('name', this.get('controller.serviceName')));
          this.switchLevel("HOSTS_LIST");
        },

        /**
         * Onclick handler for button <-Services
         */
        backToServiceList: function () {
          this.get("controller").set("serviceName", "");
          this.set("parentView.isHostListHidden", true);
          this.set("parentView.isServiceListHidden", false);
          this.set("parentView.isTaskListHidden", true);
          this.set("parentView.isLogWrapHidden", true);
          this.set("hosts", null);
          this.get("controller").setBackgroundOperationHeader(false);
          this.switchLevel("REQUESTS_LIST");
        },

        /**
         * Onclick handler for Show more ..
         */
        requestMoreOperations: function () {
          var BGOController = App.router.get('backgroundOperationsController');
          var count = BGOController.get('operationsCount');
          BGOController.set('operationsCount', (count + 10));
          BGOController.requestMostRecent();
        },

        setShowMoreAvailable: function () {
          if (this.get('parentView.isOpen')) {
            this.set('isShowMore', App.router.get("backgroundOperationsController.isShowMoreAvailable"));
          }
        }.observes('parentView.isOpen', 'App.router.backgroundOperationsController.isShowMoreAvailable'),
        isShowMore: true,

        /**
         * Onclick handler for selected Service
         * @param {Object} event
         */
        gotoHosts: function (event) {
          this.get("controller").set("serviceName", event.context.get("name"));
          this.get("controller").set("currentServiceId", event.context.get("id"));
          this.get("controller").set("currentHostName", null);
          this.get("controller").onHostUpdate();
          this.switchLevel("HOSTS_LIST");
          var servicesInfo = this.get("controller.hosts");
          this.set("controller.popupHeaderName", event.context.get("name"));
          this.set("controller.operationInfo", event.context);

          //apply lazy loading on cluster with more than 100 nodes
          if (servicesInfo.length > 100) {
            this.set('hosts', servicesInfo.slice(0, 50));
          } else {
            this.set('hosts', servicesInfo);
          }
          this.set("parentView.isServiceListHidden", true);
          this.set("parentView.isHostListHidden", false);
          this.set("parentView.isTaskListHidden", true);
          $(".modal").scrollTop(0);
          $(".modal-body").scrollTop(0);
          if (servicesInfo.length > 100) {
            Em.run.next(this, function(){
              this.set('hosts', this.get('hosts').concat(servicesInfo.slice(50, servicesInfo.length)));
            });
          }
          // Determine if source request schedule is present
          this.set('sourceRequestScheduleId', event.context.get("sourceRequestScheduleId"));
          this.set('sourceRequestScheduleCommand', event.context.get('contextCommand'));
          this.refreshRequestScheduleInfo();
        },

        isRequestSchedule : function() {
          var id = this.get('sourceRequestScheduleId');
          return id != null && !isNaN(id) && id > -1;
        }.property('sourceRequestScheduleId'),

        refreshRequestScheduleInfo : function() {
          var self = this;
          var id = this.get('sourceRequestScheduleId');
          batchUtils.getRequestSchedule(id, function(data) {
            if (data != null && data.RequestSchedule != null &&
                data.RequestSchedule.status != null) {
              switch (data.RequestSchedule.status) {
              case 'DISABLED':
                self.set('sourceRequestScheduleRunning', false);
                self.set('sourceRequestScheduleAborted', true);
                break;
              case 'COMPLETED':
                self.set('sourceRequestScheduleRunning', false);
                self.set('sourceRequestScheduleAborted', false);
                break;
              case 'SCHEDULED':
                self.set('sourceRequestScheduleRunning', true);
                self.set('sourceRequestScheduleAborted', false);
                break;
              }
            } else {
              self.set('sourceRequestScheduleRunning', false);
              self.set('sourceRequestScheduleAborted', false);
            }
          }, function(xhr, textStatus, error, opt) {
            console.log("Error getting request schedule information: ", textStatus, error, opt);
            self.set('sourceRequestScheduleRunning', false);
            self.set('sourceRequestScheduleAborted', false);
          });
        }.observes('sourceRequestScheduleId'),

        /**
         * Attempts to abort the current request schedule
         */
        doAbortRequestSchedule: function(event){
          var self = this;
          var id = event.context;
          console.log("Aborting request schedule: ", id);
          batchUtils.doAbortRequestSchedule(id, function(){
            self.refreshRequestScheduleInfo();
          });
        },

        requestScheduleAbortLabel : function() {
          var label = Em.I18n.t("common.abort");
          var command = this.get('sourceRequestScheduleCommand');
          if (command != null && "ROLLING-RESTART" == command) {
            label = Em.I18n.t("hostPopup.bgop.abort.rollingRestart");
          }
          return label;
        }.property('sourceRequestScheduleCommand'),

        /**
         * Onclick handler for selected Host
         * @param {Object} event
         */
        gotoTasks: function (event) {
          var tasksInfo = [];
          event.context.logTasks.forEach(function (_task) {
            tasksInfo.pushObject(this.get("controller").createTask(_task));
          }, this);
          if (tasksInfo.length) {
            this.get("controller").set("popupHeaderName", event.context.publicName);
            this.get("controller").set("currentHostName", event.context.publicName);
          }
          this.switchLevel("TASKS_LIST");
          this.set('currentHost.tasks', tasksInfo);
          this.set("parentView.isHostListHidden", true);
          this.set("parentView.isTaskListHidden", false);
          $(".modal").scrollTop(0);
          $(".modal-body").scrollTop(0);
        },

        stopRebalanceHDFS: function () {
          var hostPopup = this;
          return App.showConfirmationPopup(function () {
          App.ajax.send({
            name : 'cancel.background.operation',
              sender : hostPopup,
            data : {
              requestId : hostPopup.get('controller.currentServiceId')
            }
          });
            hostPopup.backToServiceList();
          });
        },
        /**
         * Onclick handler for selected Task
         */
        openTaskLogInDialog: function () {
          if ($(".task-detail-log-clipboard").length > 0) {
            this.destroyClipBoard();
          }
          var newWindow = window.open();
          var newDocument = newWindow.document;
          newDocument.write($(".task-detail-log-info").html());
          newDocument.close();
        },

        openedTaskId: 0,

        /**
         * Return task detail info of opened task
         */
        openedTask: function () {
          if (!(this.get('openedTaskId') && this.get('tasks'))) {
            return Ember.Object.create();
          }
          return this.get('tasks').findProperty('id', this.get('openedTaskId'));
        }.property('tasks', 'tasks.@each.stderr', 'tasks.@each.stdout', 'openedTaskId'),

        /**
         * Onclick event for show task detail info
         * @param {Object} event
         */
        toggleTaskLog: function (event) {
          var taskInfo = event.context;
          this.set("parentView.isLogWrapHidden", false);
          if ($(".task-detail-log-clipboard").length > 0) {
            this.destroyClipBoard();
          }
          this.set("parentView.isHostListHidden", true);
          this.set("parentView.isTaskListHidden", true);
          this.set('openedTaskId', taskInfo.id);
          this.switchLevel("TASK_DETAILS");
          $(".modal").scrollTop(0);
          $(".modal-body").scrollTop(0);
        },

        /**
         * Onclick event for copy to clipboard button
         */
        textTrigger: function () {
          $(".task-detail-log-clipboard").length > 0 ? this.destroyClipBoard() : this.createClipBoard();
        },

        /**
         * Create Clip Board
         */
        createClipBoard: function () {
          var logElement = $(".task-detail-log-maintext");
          $(".task-detail-log-clipboard-wrap").html('<textarea class="task-detail-log-clipboard"></textarea>');
          $(".task-detail-log-clipboard")
            .html("stderr: \n" + $(".stderr").html() + "\n stdout:\n" + $(".stdout").html())
            .css("display", "block")
            .width(logElement.width())
            .height(logElement.height())
            .select();
          logElement.css("display", "none")
        },

        /**
         * Destroy Clip Board
         */
        destroyClipBoard: function () {
          $(".task-detail-log-clipboard").remove();
          $(".task-detail-log-maintext").css("display", "block");
        }

      })
    }));
    return self.get('isPopup');
  }

});

});

require.register("utils/hosts", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

require('views/common/table_view');

var App = require('app');
var lazyloading = require('utils/lazy_loading');

module.exports = {

  /**
   * Launches a dialog to select hosts from the provided available hosts.
   *
   * Once the user clicks OK or Cancel, the callback is called with the
   * array of hosts (App.Host[]) selected. If the dialog was cancelled
   * or closed, <code>null</code> is provided to the callback. Else
   * an array (maybe empty) will be provided to the callback.
   *
   * @param initialHosts  {App.Host[]} List of hosts to pick from
   * @param selectedHosts {App.Host[]} List of hosts already selected from the available hosts
   * @param selectAtleastOneHost  {boolean} If true atleast one host has to be selected
   * @param validComponents {App.HostComponent[]} List of host-component types to pick from.
   * @param callback  Callback function which is invoked when dialog
   * @param popupDescription {Object} Consist header and message for popup
   *   Example: {header: 'header', dialogMessage: 'message'}
   *  is closed, cancelled or OK is pressed.
   */
  launchHostsSelectionDialog : function(initialHosts, selectedHosts,
      selectAtleastOneHost, validComponents, callback, popupDescription) {
    // set default popup description
    var defaultPopupDescription = {
      header: Em.I18n.t('hosts.selectHostsDialog.title'),
      dialogMessage: Em.I18n.t('hosts.selectHostsDialog.message')
    };
    if (popupDescription !== null) {
      popupDescription = $.extend(true, defaultPopupDescription, popupDescription);
    }
    App.ModalPopup.show({
      classNames: [ 'sixty-percent-width-modal' ],
      header: popupDescription.header,
      dialogMessage: popupDescription.dialogMessage,
      warningMessage: null,
      availableHosts: [],
      onPrimary: function () {
        this.set('warningMessage', null);
        var arrayOfSelectedHosts = this.get('availableHosts').filterProperty('selected', true).mapProperty('host.id');
        if (selectAtleastOneHost && arrayOfSelectedHosts.length < 1) {
          this.set('warningMessage', Em.I18n.t('hosts.selectHostsDialog.message.warning'));
          return;
        }
        callback(arrayOfSelectedHosts);
        console.debug('(new-selectedHosts)=', arrayOfSelectedHosts);
        this.hide();
      },
      disablePrimary: function () {
        return !this.get('isLoaded');
      }.property('isLoaded'),
      onSecondary: function () {
        callback(null);
        this.hide();
      },
      bodyClass: App.TableView.extend({
        templateName: require('templates/common/configs/overrideWindow'),
        controllerBinding: 'App.router.mainServiceInfoConfigsController',
        isPaginate: true,
        filteredContent: function() {
          return this.get('parentView.availableHosts').filterProperty('filtered') || [];
        }.property('parentView.availableHosts.@each.filtered'),
        filterText: '',
        filterTextPlaceholder: Em.I18n.t('hosts.selectHostsDialog.filter.placeHolder'),
        filterColumn: null,
        filterColumns: Ember.A([
          Ember.Object.create({id: 'ip', name: 'IP Address', selected: true}),
          Ember.Object.create({id: 'cpu', name: 'CPU', selected: false}),
          Ember.Object.create({id: 'memory', name: 'RAM', selected: false}),
          Ember.Object.create({id: 'osArch', name: 'OS Architecture', selected: false}),
          Ember.Object.create({id: 'osType', name: 'OS Type', selected: false}),
          Ember.Object.create({id: 'diskTotal', name: 'Total Disks Capacity', selected: false}),
          Ember.Object.create({id: 'disksMounted', name: '# of Disk Mounts', selected: false})
        ]),
        showOnlySelectedHosts: false,
        filterComponents: validComponents,
        filterComponent: null,
        isDisabled: function () {
          return !this.get('parentView.isLoaded');
        }.property('parentView.isLoaded'),
        didInsertElement: function(){
          var defaultFilterColumn = this.get('filterColumns').findProperty('selected');
          this.set('filterColumn', defaultFilterColumn);
          initialHosts.setEach('filtered', true);
          this.set('parentView.availableHosts', initialHosts);
          this.set('parentView.isLoaded', true);
        },
        filterHosts: function () {
          var filterText = this.get('filterText');
          var showOnlySelectedHosts = this.get('showOnlySelectedHosts');
          var filterComponent = this.get('filterComponent');
          var filterColumn = this.get('filterColumn');

          this.get('parentView.availableHosts').forEach(function (host) {
            var skip = showOnlySelectedHosts && !host.get('selected');
            var value = host.get('host').get(filterColumn.id);
            var hostComponentNames = host.get('hostComponentNames');

            host.set('filterColumnValue', value);

            if (!skip && filterText) {
              if ((value == null || !value.toString().match(filterText)) && !host.get('host.publicHostName').match(filterText)) {
                skip = true;
              }
            }
            if (!skip && filterComponent) {
              if (hostComponentNames.length > 0) {
                skip = !hostComponentNames.contains(filterComponent.get('componentName'));
              }
            }
            host.set('filtered', !skip);
          }, this);

          this.set('startIndex', 1);
        }.observes('parentView.availableHosts', 'filterColumn', 'filterText', 'filterComponent', 'filterComponent.componentName', 'showOnlySelectedHosts'),
        hostSelectMessage: function () {
          var hosts = this.get('parentView.availableHosts');
          var selectedHosts = hosts.filterProperty('selected', true);
          return this.t('hosts.selectHostsDialog.selectedHostsLink').format(selectedHosts.get('length'), hosts.get('length'))
        }.property('parentView.availableHosts.@each.selected'),
        selectFilterColumn: function (event) {
          if (event != null && event.context != null && event.context.id != null) {
            var filterColumn = this.get('filterColumn');
            if (filterColumn != null) {
              filterColumn.set('selected', false);
            }
            event.context.set('selected', true);
            this.set('filterColumn', event.context);
          }
        },
        selectFilterComponent: function (event) {
          if (event != null && event.context != null && event.context.componentName != null) {
            var currentFilter = this.get('filterComponent');
            if (currentFilter != null) {
              currentFilter.set('selected', false);
            }
            if (currentFilter != null && currentFilter.componentName === event.context.componentName) {
              // selecting the same filter deselects it.
              this.set('filterComponent', null);
            } else {
              this.set('filterComponent', event.context);
              event.context.set('selected', true);
            }
          }
        },
        allHostsSelected: false,
        toggleSelectAllHosts: function (event) {
          this.get('parentView.availableHosts').filterProperty('filtered').setEach('selected', this.get('allHostsSelected'));
        }.observes('allHostsSelected'),
        toggleShowSelectedHosts: function () {
          var currentFilter = this.get('filterComponent');
          if (currentFilter != null) {
            currentFilter.set('selected', false);
          }
          this.set('filterComponent', null);
          this.set('filterText', null);
          this.set('showOnlySelectedHosts', !this.get('showOnlySelectedHosts'));
        }
      })
    });
  }
};

});

require.register("utils/http_client", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * App.HttpClient perform an ajax request
 */
App.HttpClient = Em.Object.create({

  /**
   *
   * @param jqXHR
   * @param textStatus
   * @param errorThrown
   * @param url api that invoked this callback function
   */
  defaultErrorHandler: function (jqXHR, textStatus, errorThrown, url) {
    try {
      var json = $.parseJSON(jqXHR.responseText);
    } catch (err) { }
    App.ajax.defaultErrorHandler(jqXHR, url);
    if (json) {
      Em.assert("HttpClient:", json);
    }
    else {
      if (!$.mocho) { // don't use this assert on tests
        Em.assert("HttpClient:", errorThrown);
      }
    }
  },

  /**
   * @param {string} url
   * @param {Object} ajaxOptions
   * @param {App.ServerDataMapper} mapper - json processor
   * @param {callback} errorHandler
   */
  request: function (url, ajaxOptions, mapper, errorHandler) {

    if (!errorHandler) {
      errorHandler = this.defaultErrorHandler;
    }

    var xhr = new XMLHttpRequest();
    var curTime = App.dateTime();

    xhr.open('GET', url + (url.indexOf('?') >= 0 ? '&_=' : '?_=') + curTime, true);
    xhr.send(null);

    this.onReady(xhr, "", ajaxOptions, mapper, errorHandler, url);
  },

  /**
   * Do POST-request equal to GET-request but with some params put to body
   * @param {string} url
   * @param {{params: string, success: callback, error: callback}} ajaxOptions
   * @param {App.QuickDataMapper} mapper
   * @param errorHandler
   * @method getAsPostRequest
   */
  getAsPostRequest: function (url, ajaxOptions, mapper, errorHandler) {

    if (!errorHandler) {
      errorHandler = this.defaultErrorHandler;
    }

    var xhr = new XMLHttpRequest(),
      curTime = App.dateTime(),
      params = JSON.stringify({
        "RequestInfo": {"query" : ajaxOptions.params }
      });

    xhr.open('POST', url + (url.indexOf('?') >= 0 ? '&_=' : '?_=') + curTime, true);
    xhr.setRequestHeader("X-Http-Method-Override", "GET");
    xhr.setRequestHeader("Content-type", "application/x-www-form-urlencoded");
    xhr.send(params);

    this.onReady(xhr, "", ajaxOptions, mapper, errorHandler, url);
  },

  /*
   This function checks if we get response from server
   Not using onreadystatechange cuz of possible closure
   */
  onReady: function (xhr, tm, tmp_val, mapper, errorHandler, url) {
    var self = this;
    clearTimeout(tm);
    var timeout = setTimeout(function () {
      if (xhr.readyState == 4) {
        if (xhr.status == 200) {
          try {
            App.store.commit();
          } catch (err) {
            console.warn('App.store.commit error:', err);
          }
          mapper.map($.parseJSON(xhr.responseText));
          tmp_val.complete.call(self);
          xhr.abort();
        } else {
          errorHandler(xhr, "error", xhr.statusText, url);
        }

        tmp_val = null;
        xhr = null;
        clearTimeout(timeout);
        timeout = null;
      }
      else {
        self.onReady(xhr, timeout, tmp_val, mapper, errorHandler, url);
      }
    }, 10);
  },

  /**
   * @param {string} url
   * @param {App.ServerDataMapper} mapper - json processor
   * @param {Object} data - ajax data property
   * @param {callback} errorHandler
   * @param {number} interval - frequency request
   */
  get: function (url, mapper, data, errorHandler, interval) {
    if (!errorHandler && data.error) {
      errorHandler = data.error;
    }
    var client = this,
      request = function () {
        if (data.doGetAsPost && !App.get('testMode')) {
          client.getAsPostRequest(url, data, mapper, errorHandler);
        }
        else {
          client.request(url, data, mapper, errorHandler);
        }
        url = null;
        data = null;
        mapper = null;
        errorHandler = null;
      };

    interval = "" + interval;
    if (interval.match(/\d+/)) {
      $.periodic({period: interval}, request);
    }
    else {
      request();
    }
  },

  /**
   * @param {string} url
   * @param {Object} data - ajax data property
   * @param {App.ServerDataMapper} mapper - json processor
   * @param {callback} errorHandler
   * @param {number} interval - frequecy request
   */
  post: function (url, data, mapper, errorHandler, interval) {
    this.get(url, data, mapper, errorHandler, interval);
  }
});

});

require.register("utils/lazy_loading", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

module.exports = {
  /**
   * Divide source array in chunks, then push each chunk into destination array
   * with delay time one by one. So then destination array gets more and more items
   * till all items are loaded.
   * @param options
   * options.initSize - number of items which will be pushed in array immediately
   * options.chunkSize - number of items which will be pushed after defined delay
   * options.delay - interval between each chunk push
   * options.destination - array where items will be pushed
   * options.source - source of items
   * options.context - the object that should know when data is completely loaded,
   * lazy loading will define "isLoaded" property for context object and update it
   * @return {Object} - instance of lazy loading run
   */
  run: function (options) {
    var initSize = options.initSize || 25,
      chunkSize = options.chunkSize || 50,
      delay = options.delay || 300,
      destination = options.destination,
      source = options.source,
      self = this,
      instance = {
        context: options.context,
        timeoutRef: null
      },
      chunks;
    if (Array.isArray(destination) && Array.isArray(source)) {
      destination.pushObjects(source.slice(0, initSize));
      if(source.length > initSize) {
        chunks = this.divideIntoChunks(source.slice(initSize, source.length), chunkSize);
        this.pushChunk(chunks, 0, delay, destination, instance);
      } else {
        instance.context.set('isLoaded', true);
      }
    } else {
      console.error('Lazy loading: source or destination has incorrect value');
    }
    return instance;
  },

  /**
   * push chunks into destination array in delay time
   * @param chunks
   * @param index
   * @param delay
   * @param destination
   * @param instance
   */
  pushChunk: function (chunks, index, delay, destination, instance) {
    var self = this;
    instance.timeoutRef = setTimeout(function () {
      destination.pushObjects(chunks[index]);
      if (chunks.length === (++index)) {
        instance.context.set('isLoaded', true);
      } else {
        self.pushChunk(chunks, index, delay, destination, instance);
      }
    }, delay);
  },

  /**
   * divide source array into chunks
   * @param source
   * @param chunkSize
   * @return {Array}
   */
  divideIntoChunks: function (source, chunkSize) {
    var chunk = [];
    var chunks = [];
    var counter = 0;
    source.forEach(function (item) {
      counter++;
      chunk.push(item);
      if (counter === chunkSize) {
        chunks.push(chunk);
        chunk = [];
        counter = 0;
      }
    });
    if (chunk.length > 0) {
      chunks.push(chunk);
    }
    return chunks;
  },
  /**
   * terminate lazy loading run
   * @param instance - lazy loading instance
   */
  terminate: function (instance) {
    clearTimeout(instance.timeoutRef);
  }
};

});

require.register("utils/misc", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
module.exports = {
  /**
   * Convert value from bytes to appropriate measure
   */
  formatBandwidth: function (value) {
    if (value) {
      if (value < 1024) {
        value = '<1KB';
      } else {
        if (value < 1048576) {
          value = (value / 1024).toFixed(1) + 'KB';
        } else  if (value >= 1048576 && value < 1073741824){
          value = (value / 1048576).toFixed(1) + 'MB';
        } else {
          value = (value / 1073741824).toFixed(2) + 'GB';
        }
      }
    }
    return value;
  },
  /**
   * Convert ip address to integer
   * @param ip
   * @return integer
   */
  ipToInt: function(ip){
    // *     example 1: ipToInt('192.0.34.166');
    // *     returns 1: 3221234342
    // *     example 2: ipToInt('255.255.255.256');
    // *     returns 2: false
    // Verify IP format.
    if (!/^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$/.test(ip)) {
      return false; // Invalid format.
    }
    // Reuse ip variable for component counter.
    var d = ip.split('.');
    return ((((((+d[0])*256)+(+d[1]))*256)+(+d[2]))*256)+(+d[3]);
  },

  sortByOrder: function (sortOrder, array) {
    var sorted = [];
    for (var i = 0; i < sortOrder.length; i++)
      for (var j = 0; j < array.length; j++) {
        if (sortOrder[i] == Em.get(array[j], 'id')) {
          sorted.push(array[j]);
        }
      }
    return sorted;
  },
};

});

require.register("utils/number_utils", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
module.exports = {

  /**
   * Convert byte size to other metrics.
   *
   * @param {Number} bytes to convert to string
   * @param {Number} precision Number to adjust precision of return value. Default is 0.
   * @param {String} parseType
   *           JS method name for parse string to number. Default is "parseInt".
   * @param {Number} multiplyBy bytes by this number if given. This is needed
   *          as <code>null * 1024 = 0</null>
   * @remarks The parseType argument can be "parseInt" or "parseFloat".
   * @return {String} Returns converted value with abbreviation.
   */
  bytesToSize: function (bytes, precision, parseType, multiplyBy) {
    if (bytes === null || bytes === undefined) {
      return 'n/a';
    } else {
      if (arguments[2] === undefined) {
        parseType = 'parseInt';
      }
      if (arguments[3] === undefined) {
        multiplyBy = 1;
      }
      var value = bytes * multiplyBy;
      var sizes = [ 'Bytes', 'KB', 'MB', 'GB', 'TB', 'PB' ];
      var posttxt = 0;
      while (value >= 1024) {
        posttxt++;
        value = value / 1024;
      }
      if (value === 0) {
        precision = 0;
      }
      var parsedValue = window[parseType](value);
      return parsedValue.toFixed(precision) + " " + sizes[posttxt];
    }
  },

  /**
   * Validates if the given string or number is an integer between the
   * values of min and max (inclusive). The minimum and maximum
   * checks are ignored if their valid is NaN.
   *
   * @method validateInteger
   * @param {string|number} str - input string
   * @param {string|number} [min]
   * @param {string|number} [max]
   */
  validateInteger : function(str, min, max) {
    if (str==null || str==undefined || (str + "").trim().length < 1) {
      return Em.I18n.t('number.validate.empty');
    } else {
      str = (str + "").trim();
      var number = parseInt(str);
      if (isNaN(number)) {
        return Em.I18n.t('number.validate.notValidNumber');
      } else {
        if (str.length != (number + "").length) {
          // parseInt("1abc") returns 1 as integer
          return Em.I18n.t('number.validate.notValidNumber');
        }
        if (!isNaN(min) && number < min) {
          return Em.I18n.t('number.validate.lessThanMinumum').format(min);
        }
        if (!isNaN(max) && number > max) {
          return Em.I18n.t('number.validate.moreThanMaximum').format(max);
        }
      }
    }
    return null;
  },
  /**
   * @param {String|Number} cardinality - value to parse
   * @param {Boolean} isMax - return maximum count
   * @return {Number}
   **/
  getCardinalityValue: function(cardinality, isMax) {
    if (cardinality) {
      var isOptional = cardinality.toString().split('-').length > 1;
      if (isOptional) {
        return parseInt(cardinality.split('-')[(isMax) ? 1 : 0]);
      } else {
        if (isMax) return /^\d+\+/.test(cardinality) || cardinality == 'ALL' ? Infinity : parseInt(cardinality);
        return cardinality == 'ALL' ? Infinity : parseInt(cardinality.toString().replace('+', ''))
      }
    } else {
      return 0;
    }
  }
};

});

require.register("utils/object_utils", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var stringUtils = require('utils/string_utils');

module.exports = {

  isChild: function(obj)
  {
    for (var k in obj) {
      if (obj.hasOwnProperty(k)) {
        if (obj[k] instanceof Object) {
          return false;
        }
      }
    }
    return true;
  },

  recursiveKeysCount: function(obj) {
    if (!(obj instanceof Object)) {
      return null;
    }
    var self = this;
    function r(obj) {
      var count = 0;
      for (var k in obj) {
        if(self.isChild(obj[k])){
          count++;
        } else {
          count += r(obj[k]);
        }
      }
      return count;
    }
    return r(obj);
  },

  recursiveTree: function(obj) {
    if (!(obj instanceof Object)) {
      return null;
    }
    var self = this;
    function r(obj,parent) {
      var leaf = '';
      for (var k in obj) {
        if(self.isChild(obj[k])){
          leaf += k + ' ('+parent+')' + '<br/>';
        } else {
          leaf += r(obj[k],parent +'/' + k);
        }
      }
      return leaf;
    }
    return r(obj,'');
  }
};

});

require.register("utils/polling", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
App.Poll = Em.Object.extend({
  name: '',
  stage: '',
  label: '',
  isVisible: true,
  isStarted: false,
  isPolling: true,
  clusterName: null,
  requestId: null,
  temp: false,
  progress: 0,
  url: null,
  testUrl: null,
  data: null,
  isError: false,
  isSuccess: false,
  POLL_INTERVAL: 4000,
  polledData: [],
  numPolls: 0,
  mockDataPrefix: '/data/wizard/deploy/5_hosts',
  currentTaskId: null,

  barWidth: function () {
    return 'width: ' + this.get('progress') + '%;';
  }.property('progress'),

  isCompleted: function () {
    return (this.get('isError') || this.get('isSuccess'));
  }.property('isError', 'isSuccess'),

  showLink: function () {
    return (this.get('isPolling') === true && this.get('isStarted') === true);
  }.property('isPolling', 'isStarted'),

  start: function () {
    if (Em.isNone(this.get('requestId'))) {
      this.setRequestId();
    } else {
      this.startPolling();
    }
  },

  setRequestId: function () {
    if (App.get('testMode')) {
      this.set('requestId', '1');
      this.doPolling();
      return;
    }
    var self = this;
    var url = this.get('url');
    var method = 'PUT';
    var data = this.get('data');

    $.ajax({
      type: method,
      url: url,
      data: data,
      dataType: 'text',
      timeout: App.timeout,
      success: function (data) {
        var jsonData = jQuery.parseJSON(data);
        console.log("TRACE: Polling -> value of the url is: " + url);
        console.log("TRACE: Polling-> value of the sent data is: " + self.get('data'));
        console.log("TRACE: Polling-> value of the received data is: " + jsonData);
        if (Em.isNone(jsonData)) {
          self.set('isSuccess', true);
          self.set('isError', false);
        } else {
          var requestId = jsonData.Requests.id;
          self.set('requestId', requestId);
          self.doPolling();
          console.log('requestId is: ' + requestId);
        }
      },

      error: function () {
        console.log("ERROR");
        self.set('isError', true);
        self.set('isSuccess', false);
      },

      statusCode: require('data/statusCodes')
    });
  },

  /**
   * set current task id and send request
   * @param taskId
   */
  updateTaskLog: function (taskId) {
    this.set('currentTaskId', taskId);
    this.pollTaskLog();
  },

  doPolling: function () {
    if (this.get('requestId')) {
      this.startPolling();
    }
  },

  /**
   * server call to obtain task logs
   */
  pollTaskLog: function () {
    if (this.get('currentTaskId')) {
      App.ajax.send({
        name: 'background_operations.get_by_task',
        sender: this,
        data: {
          requestId: this.get('requestId'),
          taskId: this.get('currentTaskId')
        },
        success: 'pollTaskLogSuccessCallback'
      })
    }
  },

  /**
   * update logs of current task
   * @param data
   */
  pollTaskLogSuccessCallback: function (data) {
    var currentTask = this.get('polledData').findProperty('Tasks.id', data.Tasks.id);
    currentTask.Tasks.stdout = data.Tasks.stdout;
    currentTask.Tasks.stderr = data.Tasks.stderr;
    Em.propertyDidChange(this, 'polledData');
  },

  /**
   * start polling operation data
   * @return {Boolean}
   */
  startPolling: function () {
    if (!this.get('requestId')) return false;

    this.pollTaskLog();
    App.ajax.send({
      name: 'background_operations.get_by_request',
      sender: this,
      data: {
        requestId: this.get('requestId')
      },
      success: 'startPollingSuccessCallback',
      error: 'startPollingErrorCallback'
    })
      .retry({times: App.maxRetries, timeout: App.timeout})
      .then(null, function () {
        App.showReloadPopup();
        console.log('Install services all retries failed');
      });
    return true;
  },

  startPollingSuccessCallback: function (data) {
    var self = this;
    var result = this.parseInfo(data);
    if (!result) {
      window.setTimeout(function () {
        self.startPolling();
      }, this.POLL_INTERVAL);
    }
  },

  startPollingErrorCallback: function (request, ajaxOptions, error) {
    console.log("TRACE: In error function for the GET data");
    console.log("TRACE: value of the url is: " + url);
    console.log("TRACE: error code status is: " + request.status);
    if (!this.get('isSuccess')) {
      this.set('isError', true);
    }
  },

  stopPolling: function () {
    //this.set('isSuccess', true);
  },

  replacePolledData: function (polledData) {
    var currentTaskId = this.get('currentTaskId');
    if (currentTaskId) {
      var task = this.get('polledData').findProperty('Tasks.id', currentTaskId);
      var currentTask = polledData.findProperty('Tasks.id', currentTaskId);
      if (task && currentTask) {
        currentTask.Tasks.stdout = task.Tasks.stdout;
        currentTask.Tasks.stderr = task.Tasks.stderr;
      }
    }
    this.set('polledData', polledData);
  },


  calculateProgressByTasks: function (tasksData) {
    var queuedTasks = tasksData.filterProperty('Tasks.status', 'QUEUED').length;
    var completedTasks = tasksData.filter(function (task) {
      return ['COMPLETED', 'FAILED', 'ABORTED', 'TIMEDOUT'].contains(task.Tasks.status);
    }).length;
    var inProgressTasks = tasksData.filterProperty('Tasks.status', 'IN_PROGRESS').length;
    return Math.ceil(((queuedTasks * 0.09) + (inProgressTasks * 0.35) + completedTasks ) / tasksData.length * 100)
  },

  isPollingFinished: function (polledData) {
    var runningTasks;
    runningTasks = polledData.filterProperty('Tasks.status', 'QUEUED').length;
    runningTasks += polledData.filterProperty('Tasks.status', 'IN_PROGRESS').length;
    runningTasks += polledData.filterProperty('Tasks.status', 'PENDING').length;
    if (runningTasks === 0) {
      if (polledData.everyProperty('Tasks.status', 'COMPLETED')) {
        this.set('isSuccess', true);
        this.set('isError', false);
      } else if (polledData.someProperty('Tasks.status', 'FAILED') || polledData.someProperty('Tasks.status', 'TIMEDOUT') || polledData.someProperty('Tasks.status', 'ABORTED')) {
        this.set('isSuccess', false);
        this.set('isError', true);
      }
      return true;
    } else {
      return false;
    }
  },


  parseInfo: function (polledData) {
    console.log('TRACE: Entering task info function');
    var self = this;
    var totalProgress = 0;
    var tasksData = polledData.tasks;
    console.log("The value of tasksData is: ", tasksData);
    if (!tasksData) {
      console.log("ERROR: NO tasks available to process");
    }
    var requestId = this.get('requestId');
    if (polledData.Requests && polledData.Requests.id && polledData.Requests.id != requestId) {
      // We don't want to use non-current requestId's tasks data to
      // determine the current install status.
      // Also, we don't want to keep polling if it is not the
      // current requestId.
      return false;
    }
    this.replacePolledData(tasksData);
    var totalProgress = this.calculateProgressByTasks(tasksData);
    this.set('progress', totalProgress.toString());
    console.log("INFO: right now the progress is: " + this.get('progress'));
    return this.isPollingFinished(tasksData);
  }

});


});

require.register("utils/string_utils", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

module.exports = {

  pad: function(str, len, pad, dir) {

    var STR_PAD_LEFT = 1;
    var STR_PAD_RIGHT = 2;
    var STR_PAD_BOTH = 3;

    if (typeof(len) == "undefined") { var len = 0; }
    if (typeof(pad) == "undefined") { var pad = ' '; }
    if (typeof(dir) == "undefined") { var dir = STR_PAD_RIGHT; }

    if (len + 1 >= str.length) {

      switch (dir){

        case STR_PAD_LEFT:
          str = Array(len + 1 - str.length).join(pad) + str;
          break;

        case STR_PAD_BOTH:
          var padlen = len - str.length;
          var right = Math.ceil((padlen) / 2);
          var left = padlen - right;
          str = Array(left+1).join(pad) + str + Array(right+1).join(pad);
          break;

        default:
          str = str + Array(len + 1 - str.length).join(pad);
          break;

      } // switch

    }
    return str;

  },
  underScoreToCamelCase: function(name){
    function replacer(str, p1, p2, offset, s) {
      return str[1].toUpperCase();
    }
    return name.replace(/_\w/g,replacer);
  },

  /**
   * Forces given string into upper camel-case representation. The first
   * character of each word will be capitalized with the rest in lower case.
   */
  getCamelCase : function(name) {
    if (name != null) {
      return name.toLowerCase().replace(/(\b\w)/g, function(f) {
        return f.toUpperCase();
      })
    }
    return name;
  },

  /**
   * Compare two versions by following rules:
   * first higher than second then return 1
   * first lower than second then return -1
   * first equal to second then return 0
   * @param first {string}
   * @param second {string}
   * @return {number}
   */
  compareVersions: function (first, second) {
    if (!(typeof first === 'string' && typeof second === 'string')) {
      return -1;
    }
    if (first === '' || second === '') {
      return -1;
    }
    var firstNumbers = first.split(/[\.-]/);
    var secondNumbers = second.split(/[\.-]/);
    var length = 0;
    var i = 0;
    var result = false;
    if (firstNumbers.length === secondNumbers.length) {
      length = firstNumbers.length;
    } else if (firstNumbers.length < secondNumbers.length) {
      length = secondNumbers.length;
    } else {
      length = firstNumbers.length;
    }

    while (i < length && !result) {
      firstNumbers[i] = (firstNumbers[i] === undefined) ? 0 : window.parseInt(firstNumbers[i]);
      secondNumbers[i] = (secondNumbers[i] === undefined) ? 0 : window.parseInt(secondNumbers[i]);
      if (firstNumbers[i] > secondNumbers[i]) {
        result = 1;
        break;
      } else if (firstNumbers[i] === secondNumbers[i]) {
        result = 0;
      } else if (firstNumbers[i] < secondNumbers[i]) {
        result = -1;
        break;
      }
      i++;
    }
    return result;
  },

  isSingleLine: function(string){
    return String(string).trim().indexOf("\n") == -1;
  },
  /**
   * transform array of objects into CSV format content
   * @param array
   * @return {Array}
   */
  arrayToCSV: function(array){
    var content = "";
    array.forEach(function(item){
      var row = [];
      for(var i in item){
        if(item.hasOwnProperty(i)){
          row.push(item[i]);
        }
      }
      content += row.join(',') + '\n';
    });
    return content;
  },

  /**
   * Extracts filename from linux/unix path
   * @param path
   * @return {string}: filename
   */
  getFileFromPath: function(path) {
    if (!path || typeof path !== 'string') {
      return '';
    }
    return path.replace(/^.*[\/]/, '');
  },

  getPath: function(path) {
    if (!path || typeof path !== 'string' || path[0] != '/') {
      return '';
    }
    var last_slash = path.lastIndexOf('/');
    return (last_slash!=0)?path.substr(0,last_slash):'/';
  },

  /**
   * @method getFormattedStringFromArray Get formatted string of elements to display on the UI
   * Example:
   * var arr = [ambari, bigdata, hadoop]
   * getFormattedStringFromArray(arr);  // ambari, bigdata and hadoop
   * @param array {Array}  Array of elements
   * @returns {String}
   */
  getFormattedStringFromArray: function (array) {
    var label = '';
    array.forEach(function (_arrElement) {
      if (array.length === 1) {
        label = _arrElement;
      }
      else {
        if (_arrElement !== array[array.length - 1]) {           // [clients.length - 1]
          label = label + ' ' + _arrElement;
          if (_arrElement !== array[array.length - 2]) {
            label = label + ',';
          }
        }
        else {
          label = label + ' ' + Em.I18n.t('and') + ' ' + _arrElement;
        }
      }
    }, this);
    return label.trim();
  },
  /**
   * Get plural|singular value of string by related count.
   *
   * @param {Number} count
   * @param {String} singular
   * @param {String} [plural]
   * @return {String}
   * @method pluralize
   */
  pluralize: function(count, singular, plural) {
    plural = plural || singular + 's';
    if (count > 1) {
      return plural;
    }
    return singular;
  }
};

});

require.register("utils/ui_effects", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

module.exports = {
  /**
   *
   * @param node {DOMElement} - DOM element which blinking
   * @param delay {number} - overall time of blinking
   * @param callback {function}
   * @param interval {number} - change frequence of blinking
   */
  pulsate: function (node, delay, callback, interval) {
    var self = this;
    /**
     * execute single blink
     * @param interval {number} - time of single blink
     * @param callback {function}
     * @param opacity {string|number}
     * @param iteration {number} - current iteration(default amount of iterations: 10)
     * @param isReverse {boolean} - flag, that mean opacity increase or decrease
     */
    var blink = function (interval, callback, opacity, iteration, isReverse) {
      var iterations = 10;
      opacity = opacity || 1;
      iteration = (iteration !== undefined) ? iteration : 10;
      if (iteration > 0) {
        iteration--;
        setTimeout(function () {
          isReverse = isReverse || (opacity <= 1 / (iterations / 2));
          opacity = (isReverse) ? opacity + (1 / (iterations / 2)) : opacity - (1 / (iterations / 2));
          node.css('opacity', opacity);
          blink(interval, callback, opacity, iteration, isReverse);
        }, interval / iterations);
      } else {
        node.css('opacity', 1);
        callback();
      }
    };
    interval = interval || 200;
    if (delay > 0) {
      delay -= interval;
      setTimeout(function () {
        blink(interval, function () {
          self.pulsate(node, delay, callback, interval);
        });
      }, interval);
    } else {
      callback();
    }
  }
};

});

require.register("utils/updater", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');

var states = {};

function update(obj, name, isWorking, interval){
  if(typeof isWorking == 'string' && !obj.get(isWorking)){
    return false;
  }

  var state = states[name];

  if(!state){
    var callback = function(){
      update(obj, name, isWorking, interval);
    };
    states[name] = state = {
      timeout: null,
      func: function(){
        if(typeof isWorking == 'string' && !obj.get(isWorking)){
          return false;
        }
        obj[name](callback);
        return true;
      },
      interval: interval,
      callback: callback
    };
  }

  clearTimeout(state.timeout);

  state.timeout = setTimeout(state.func, state.interval);
  return true;
}

function rerun(name){
  var state = states[name];
  if(state){
    clearTimeout(state.timeout);
    state.func();
  }
}

App.updater = {

  /**
   * Run function periodically with <code>App.contentUpdateInterval</code> delay.
   * Example 1(wrong way, will not be working):
   *    var obj = {
   *      method: function(callback){
   *        //do something
   *      }
   *    };
   *    App.updater.run(obj, 'method');
   *
   * Will be called only once, because <code>callback</code> will never execute. Below is right way:
   *
   * Example 2:
   *    var obj = {
   *      method: function(callback){
   *        //do something
   *        callback();
   *      }
   *    };
   *    App.updater.run(obj, 'method');
   *
   * Method will always be called.
   *
   * Example 3:
   *    var obj = {
   *      method: function(callback){
   *          //do something
   *          callback();
   *      },
   *      isWorking: true
   *    };
   *    App.updater.run(obj, 'method', 'isWorking');
   *
   * <code>obj.method</code> will be called automatically.
   * Warning: You should call <code>callback</code> parameter when function finished iteration.
   * Otherwise nothing happened next time.
   * If <code>isWorking</code> provided, library will check <code>obj.isWorking</code> before iteration and
   * stop working when it equals to false. Otherwise method will always be called.
   *
   *
   *
   * @param obj Object
   * @param name Method name
   * @param isWorking Property, which will be checked as a rule for working
   * @param interval Interval between calls
   * @return {*}
   */
  run: function(obj, name, isWorking, interval){
    interval = interval || App.contentUpdateInterval;
    return update(obj, name, isWorking, interval);
  },

  /**
   * Immediate run function, which is periodically running using <code>run</code> method
   * Example:
   *    App.updater.run(obj, 'clickIt');
   *    App.updater.immediateRun('clickIt');
   *
   * <code>clickIt</code> will be executed immediately and proceed executing periodically
   *
   * @param name Method name, which was used as a parameter in <code>run</code> method
   * @return {*}
   */
  immediateRun: function(name){
    return rerun(name);
  },
  /**
   * Update interval for state by its name.
   *
   * @param name - state name
   * @param value - new interval value
   **/
  updateInterval: function(name, value) {
    if (!states[name]) return;
    states[name].interval = value;
  }

};

});

require.register("utils/validator", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

module.exports = {

  isValidEmail: function(value) {
    var emailRegex = /^((([a-z]|\d|[!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+(\.([a-z]|\d|[!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+)*)|((\x22)((((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(([\x01-\x08\x0b\x0c\x0e-\x1f\x7f]|\x21|[\x23-\x5b]|[\x5d-\x7e]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(\\([\x01-\x09\x0b\x0c\x0d-\x7f]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]))))*(((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(\x22)))@((([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))\.)+(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))$/i;
    return emailRegex.test(value);
  },

  isValidInt: function(value) {
    var intRegex = /^-?\d+$/;
    return intRegex.test(value);
  },

  isValidUNIXUser: function(value){
    var regex = /^[a-z_][a-z0-9_-]{0,31}$/;
    return regex.test(value);
  },

  isValidFloat: function(value) {
    if (typeof value === 'string' && value.trim() === '') {
      return false;
    }
    var floatRegex = /^-?(?:\d+|\d{1,3}(?:,\d{3})+)?(?:\.\d+)?$/;
    return floatRegex.test(value);
  },
  /**
   * validate directory with slash or drive at the start
   * @param value
   * @return {Boolean}
   */
  isValidDir: function(value){
    var floatRegex = /^\/[0-9a-z]*/;
    var winRegex = /^[a-z]:\\[0-9a-zA-Z]*/;
    var winUrlRegex = /^file:\/\/\/[a-zA-Z]:\/[0-9a-zA-Z]*/;
    var dirs = value.replace(/,/g,' ').trim().split(new RegExp("\\s+", "g"));
    for(var i = 0; i < dirs.length; i++){
      if(!floatRegex.test(dirs[i]) && !winRegex.test(dirs[i]) && !winUrlRegex.test(dirs[i])){
        return false;
      }
    }
    return true;
  },

  /**
   * validate directory with slash at the start
   * @param value
   * @returns {boolean}
   */
  isValidDataNodeDir: function(value) {
    var dirRegex = /^(\[[0-9a-zA-Z]+\])?(\/[0-9a-z]*)/;
    var winRegex = /^(\[[0-9a-zA-Z]+\])?[a-zA-Z]:\\[0-9a-zA-Z]*/;
    var winUrlRegex = /^(\[[0-9a-zA-Z]+\])?file:\/\/\/[a-zA-Z]:\/[0-9a-zA-Z]*/;
    var dirs = value.replace(/,/g,' ').trim().split(new RegExp("\\s+", "g"));
    for(var i = 0; i < dirs.length; i++){
      if(!dirRegex.test(dirs[i]) && !winRegex.test(dirs[i]) && !winUrlRegex.test(dirs[i])){
        return false;
      }
    }
    return true;
  },

  /**
   * validate directory doesn't start "home" or "homes"
   * @param value
   * @returns {boolean}
   */
  isAllowedDir: function(value) {
    var dirs = value.replace(/,/g,' ').trim().split(new RegExp("\\s+", "g"));
    for(var i = 0; i < dirs.length; i++){
      if(dirs[i].startsWith('/home') || dirs[i].startsWith('/homes')) {
        return false;
      }
    }
    return true;
  },

  /**
   * validate ip address with port
   * @param value
   * @return {Boolean}
   */
  isIpAddress: function(value) {
    var ipRegex = /^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)($|\:[0-9]{1,5})$/;
    return ipRegex.test(value);
  },

  /**
   * validate hostname
   * @param value
   * @return {Boolean}
   */
  isHostname: function(value) {
    var regex = /(?=^.{3,254}$)(^([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])(\.([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9]))*(\.[a-zA-Z]{1,62})$)/;
    return regex.test(value);
  },

  hasSpaces: function(value) {
    var regex = /(\s+)/;
    return regex.test(value);
  },

  isNotTrimmed: function(value) {
    var regex = /(^\s+|\s+$)/;
    return regex.test(value);
  },
  /**
   * validate domain name with port
   * @param value
   * @return {Boolean}
   */
  isDomainName: function(value) {
    var domainRegex = /^([a-zA-Z0-9]([a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])?\.)+[a-zA-Z]{2,6}$/;
    return domainRegex.test(value);
  },

  /**
   * validate username
   * @param value
   * @return {Boolean}
   */
  isValidUserName: function(value) {
    var usernameRegex = /^[a-z]([-a-z0-9]{0,30})$/;
    return usernameRegex.test(value);
  },

  /**
   * validate key of configurations
   * @param value
   * @return {Boolean}
   */
  isValidConfigKey: function(value) {
    var configKeyRegex = /^[0-9a-z_\-\.\*]+$/i;
    return configKeyRegex.test(value);
  },

  /**
   * validate configuration group name
   * @param value
   * @return {Boolean}
   */
  isValidConfigGroupName: function(value) {
    var configKeyRegex = /^[\s0-9a-z_\-]+$/i;
    return configKeyRegex.test(value);
  },

  /**
   * validate alert group name
   * @param value
   * @return {Boolean}
   */
  isValidAlertGroupName: function(value) {
    var configKeyRegex = /^[\s0-9a-z_\-]+$/i;
    return configKeyRegex.test(value);
  },

  empty:function (e) {
    switch (e) {
      case "":
      case 0:
      case "0":
      case null:
      case false:
      case undefined:
      case typeof this == "undefined":
        return true;
      default :
        return false;
    }
  },
  /**
   * Validate string that will pass as parameter to .matches() url param.
   * Try to prevent invalid regexp.
   * For example: /api/v1/clusters/c1/hosts?Hosts/host_name.matches(.*localhost.)
   *
   * @param {String} value - string to validate
   * @return {Boolean}
   * @method isValidMatchesRegexp
   */
  isValidMatchesRegexp: function(value) {
    var checkPair = function(chars) {
      chars = chars.map(function(c) { return '\\' + c; });
      var charsReg = new RegExp(chars.join('|'), 'g');
      if (charsReg.test(value)) {
        var pairContentReg = new RegExp(chars.join('.*'), 'g');
        if (!pairContentReg.test(value)) return false;
        var pairCounts = chars.map(function(c) { return value.match(new RegExp(c, 'g')).length; });
        if (pairCounts[0] != pairCounts[1] ) return false;
      }
      return true;
    };
    if (/^[\?\|\*\!,]/.test(value)) return false;
    return /^((\.\*?)?([\w\[\]\?\-_,\|\*\!\{\}]*)?)+(\.\*?)?$/g.test(value) && (checkPair(['[',']'])) && (checkPair(['{','}']));
  },

  /**
  * Remove validation messages for components which are already installed
  */
  filterNotInstalledComponents: function(validationData) {
    var hostComponents = App.HostComponent.find();
    return validationData.resources[0].items.filter(function(item) {
      // true is there is no host with this component
      return hostComponents.filterProperty("componentName", item["component-name"]).filterProperty("hostName", item.host).length === 0;
    });
  }
};

});

require.register("views", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


// load all views here

require('views/application');
require('views/common/ajax_default_error_popup_body');
require('views/common/chart');
require('views/common/chart/pie');
require('views/common/chart/linear');
require('views/common/chart/linear_time');
require('views/common/modal_popup');
require('views/common/modal_popups/alert_popup');
require('views/common/modal_popups/confirmation_feedback_popup');
require('views/common/modal_popups/confirmation_popup');
require('views/common/modal_popups/prompt_popup');
require('views/common/modal_popups/reload_popup');
require('views/common/modal_popups/cluster_check_popup');
require('views/common/modal_popups/invalid_KDC_popup');
require('views/common/editable_list');
require('views/common/rolling_restart_view');
require('views/common/select_custom_date_view');
require('views/common/metric');
require('views/common/time_range');
require('views/common/form/field');
require('views/common/quick_view_link_view');
require('views/common/configs/services_config');
require('views/common/configs/service_config_container_view');
require('views/common/configs/service_configs_by_category_view');
require('views/common/configs/service_config_view');
require('views/common/configs/service_config_tab_view');
require('views/common/configs/overriddenProperty_view');
require('views/common/configs/compare_property_view');
require('views/common/configs/config_history_flow');
require('views/common/configs/custom_category_views/notification_configs_view');
require('views/common/filter_combobox');
require('views/common/filter_combo_cleanable');
require('views/common/table_view');
require('views/common/progress_bar_view');
require('views/common/controls_view');
require('views/login');
require('views/main');
require('views/main/menu');
require('views/main/alert_definitions_view');
require('views/main/alerts/definition_details_view');
require('views/main/alerts/alert_definitions_actions_view');
require('views/main/alerts/definition_configs_view');
require('views/main/alerts/manage_alert_groups/select_definitions_popup_body_view');
require('views/main/alerts/add_alert_definition/add_alert_definition_view');
require('views/main/alerts/add_alert_definition/step1_view');
require('views/main/alerts/add_alert_definition/step2_view');
require('views/main/alerts/add_alert_definition/step3_view');
require('views/main/alerts');
require('views/main/alerts/manage_alert_groups_view');
require('views/main/alerts/manage_alert_notifications_view');
require('views/main/charts');
require('views/main/views/details');
require('views/main/host');
require('views/main/host/hosts_table_menu_view');
require('views/main/host/details');
require('views/main/host/details/host_component_view');
require('views/main/host/details/host_component_views/datanode_view');
require('views/main/host/details/host_component_views/nodemanager_view');
require('views/main/host/details/host_component_views/regionserver_view');
require('views/main/host/details/host_component_views/tasktracker_view');
require('views/main/host/menu');
require('views/main/host/summary');
require('views/main/host/configs');
require('views/main/host/configs_service');
require('views/main/host/configs_service_menu');
require('views/main/host/metrics');
require('views/main/host/stack_versions_view');
require('views/main/host/add_view');
require('views/main/host/host_alerts_view');
require('views/main/host/metrics/cpu');
require('views/main/host/metrics/disk');
require('views/main/host/metrics/load');
require('views/main/host/metrics/memory');
require('views/main/host/metrics/network');
require('views/main/host/metrics/processes');
require('views/main/host/addHost/step4_view');
require('views/main/admin');
require('views/main/admin/highAvailability/nameNode/wizard_view');
require('views/main/admin/highAvailability/progress_view');
require('views/main/admin/highAvailability/nameNode/rollback_view');
require('views/main/admin/highAvailability/nameNode/step1_view');
require('views/main/admin/highAvailability/nameNode/step2_view');
require('views/main/admin/highAvailability/nameNode/step3_view');
require('views/main/admin/highAvailability/nameNode/step4_view');
require('views/main/admin/highAvailability/nameNode/step5_view');
require('views/main/admin/highAvailability/nameNode/step6_view');
require('views/main/admin/highAvailability/nameNode/step7_view');
require('views/main/admin/highAvailability/nameNode/step8_view');
require('views/main/admin/highAvailability/nameNode/step9_view');
require('views/main/admin/highAvailability/nameNode/rollbackHA/step1_view');
require('views/main/admin/highAvailability/nameNode/rollbackHA/step2_view');
require('views/main/admin/highAvailability/nameNode/rollbackHA/step3_view');
require('views/main/admin/highAvailability/nameNode/rollbackHA/rollback_wizard_view');
require('views/main/admin/highAvailability/resourceManager/wizard_view');
require('views/main/admin/highAvailability/resourceManager/step1_view');
require('views/main/admin/highAvailability/resourceManager/step2_view');
require('views/main/admin/highAvailability/resourceManager/step3_view');
require('views/main/admin/highAvailability/resourceManager/step4_view');
require('views/main/admin/serviceAccounts_view');
require('views/main/admin/stack_upgrade/upgrade_wizard_view');
require('views/main/admin/stack_upgrade/upgrade_version_box_view');
require('views/main/admin/stack_upgrade/upgrade_group_view');
require('views/main/admin/stack_upgrade/upgrade_task_view');
require('views/main/admin/stack_upgrade/services_view');
require('views/main/admin/stack_upgrade/versions_view');
require('views/main/admin/stack_upgrade/menu_view');
require('views/main/admin/stack_and_upgrade_view');
require('views/main/admin/advanced');
require('views/main/admin/advanced/password');
require('views/main/admin/audit');
require('views/main/admin/authentication');

require('views/main/admin/kerberos');
require('views/main/admin/kerberos/disable_view');
require('views/main/admin/kerberos/wizard_view');
require('views/main/admin/kerberos/progress_view');
require('views/main/admin/kerberos/step1_view');
require('views/main/admin/kerberos/step2_view');
require('views/main/admin/kerberos/step3_view');
require('views/main/admin/kerberos/step4_view');
require('views/main/admin/kerberos/step5_view');
require('views/main/admin/kerberos/step6_view');
require('views/main/admin/kerberos/step7_view');

require('views/main/admin/security');
require('views/main/admin/security/disable');
require('views/main/admin/security/add/menu');
require('views/main/admin/security/add/step1');
require('views/main/admin/security/add/step2');
require('views/main/admin/security/add/step3');
require('views/main/admin/security/add/step4');

require('views/main/dashboard');
require('views/main/dashboard/cluster_metrics/cpu');
require('views/main/dashboard/cluster_metrics/load');
require('views/main/dashboard/cluster_metrics/memory');
require('views/main/dashboard/cluster_metrics/network');

require('views/main/dashboard/widget');
require('views/main/dashboard/widgets');
require('views/main/dashboard/widgets/text_widget');
require('views/main/dashboard/widgets/uptime_text_widget');
require('views/main/dashboard/widgets/links_widget');
require('views/main/dashboard/widgets/pie_chart_widget');
require('views/main/dashboard/widgets/cluster_metrics_widget');
require('views/main/dashboard/widgets/namenode_heap');
require('views/main/dashboard/widgets/namenode_cpu');
require('views/main/dashboard/widgets/hdfs_capacity');
require('views/main/dashboard/widgets/datanode_live');
require('views/main/dashboard/widgets/namenode_rpc');
require('views/main/dashboard/widgets/metrics_memory');
require('views/main/dashboard/widgets/metrics_network');
require('views/main/dashboard/widgets/metrics_cpu');
require('views/main/dashboard/widgets/metrics_load');
require('views/main/dashboard/widgets/namenode_uptime');
require('views/main/dashboard/widgets/hdfs_links');
require('views/main/dashboard/widgets/yarn_links');
require('views/main/dashboard/widgets/hbase_links');
require('views/main/dashboard/widgets/hbase_master_heap');
require('views/main/dashboard/widgets/hbase_average_load');
require('views/main/dashboard/widgets/hbase_regions_in_transition');
require('views/main/dashboard/widgets/hbase_master_uptime');
require('views/main/dashboard/widgets/resource_manager_heap');
require('views/main/dashboard/widgets/resource_manager_uptime');
require('views/main/dashboard/widgets/node_managers_live');
require('views/main/dashboard/widgets/yarn_memory');
require('views/main/dashboard/widgets/supervisor_live');
require('views/main/dashboard/widgets/flume_agent_live');
require('views/main/dashboard/config_history_view');


require('views/main/service');
require('views/main/service/service');
require('views/main/service/services/hdfs');
require('views/main/service/services/yarn');
require('views/main/service/services/mapreduce2');
require('views/main/service/services/hbase');
require('views/main/service/services/hive');
require('views/main/service/services/zookeeper');
require('views/main/service/services/oozie');
require('views/main/service/services/flume');
require('views/main/service/services/storm');
require('views/main/service/services/ranger');
require('views/main/service/all_services_actions');
require('views/main/service/menu');
require('views/main/service/item');
require('views/main/service/reconfigure');
require('views/main/service/info/components_list_view');
require('views/main/service/info/menu');
require('views/main/service/info/summary');
require('views/main/service/info/configs');
require('views/main/service/info/metric_graphs_view');
require('views/main/service/info/metrics/hdfs/jvm_threads');
require('views/main/service/info/metrics/hdfs/jvm_heap');
require('views/main/service/info/metrics/hdfs/io');
require('views/main/service/info/metrics/hdfs/rpc');
require('views/main/service/info/metrics/hdfs/file_operations');
require('views/main/service/info/metrics/hdfs/gc');
require('views/main/service/info/metrics/hdfs/space_utilization');
require('views/main/service/info/metrics/hdfs/block_status');
require('views/main/service/info/metrics/yarn/gc');
require('views/main/service/info/metrics/yarn/jvm_threads');
require('views/main/service/info/metrics/yarn/jvm_heap');
require('views/main/service/info/metrics/yarn/rpc');
require('views/main/service/info/metrics/yarn/allocated');
require('views/main/service/info/metrics/yarn/allocated_container');
require('views/main/service/info/metrics/yarn/apps_current_states');
require('views/main/service/info/metrics/yarn/apps_finished_states');
require('views/main/service/info/metrics/yarn/nms');
require('views/main/service/info/metrics/yarn/qmr');
require('views/main/service/info/metrics/hbase/cluster_requests');
require('views/main/service/info/metrics/hbase/regionserver_rw_requests');
require('views/main/service/info/metrics/hbase/regionserver_regions');
require('views/main/service/info/metrics/hbase/regionserver_queuesize');
require('views/main/service/info/metrics/hbase/hlog_split_time');
require('views/main/service/info/metrics/hbase/hlog_split_size');
require('views/main/service/info/metrics/flume/channel_sum');
require('views/main/service/info/metrics/flume/channel_size_mma');
require('views/main/service/info/metrics/flume/flume_incoming_sum');
require('views/main/service/info/metrics/flume/flume_incoming_mma');
require('views/main/service/info/metrics/flume/flume_outgoing_sum');
require('views/main/service/info/metrics/flume/flume_outgoing_mma');
require('views/main/service/info/metrics/flume/channel_fill_pct');
require('views/main/service/info/metrics/flume/channel_size');
require('views/main/service/info/metrics/flume/sink_connection_failed');
require('views/main/service/info/metrics/flume/sink_drain_success');
require('views/main/service/info/metrics/flume/source_accepted');
require('views/main/service/info/metrics/flume/gc');
require('views/main/service/info/metrics/flume/jvm_heap');
require('views/main/service/info/metrics/flume/jvm_threads_runnable');
require('views/main/service/info/metrics/flume/cpu_user');
require('views/main/service/info/metrics/flume/flume_metric_graph');
require('views/main/service/info/metrics/flume/flume_metric_graphs');
require('views/main/service/info/metrics/storm/slots_number_metric');
require('views/main/service/info/metrics/storm/executors_metric');
require('views/main/service/info/metrics/storm/tasks_metric');
require('views/main/service/info/metrics/storm/topologies_metric');
require('views/main/service/info/metrics/kafka/broker_topic');
require('views/main/service/info/metrics/kafka/kafka_controller');
require('views/main/service/info/metrics/kafka/controller_status');
require('views/main/service/info/metrics/kafka/replica_manager');
require('views/main/service/info/metrics/kafka/replica_fetcher');
require('views/main/service/info/metrics/kafka/kafka_log_flush');

require('views/main/service/add_view');
require('views/main/service/reassign_view');
require('views/main/service/reassign/step1_view');
require('views/main/service/reassign/step2_view');
require('views/main/service/reassign/step3_view');
require('views/main/service/reassign/step4_view');
require('views/main/service/reassign/step5_view');
require('views/main/service/reassign/step6_view');
require('views/main/service/manage_config_groups_view');
require('views/main/charts/menu');
require('views/main/charts/heatmap');
require('views/main/charts/heatmap/heatmap_rack');
require('views/main/charts/heatmap/heatmap_host');
require('views/main/charts/heatmap/heatmap_host_detail');

require('views/main/views_view');

require('views/installer');
require('views/wizard/step0_view');
require('views/wizard/step1_view');
require('views/wizard/step2_view');
require('views/wizard/step3_view');
require('views/wizard/step3/hostLogPopupBody_view');
require('views/wizard/step3/hostWarningPopupBody_view');
require('views/wizard/step3/hostWarningPopupFooter_view');
require('views/wizard/step4_view');
require('views/wizard/step5_view');
require('views/wizard/step6_view');
require('views/wizard/step7_view');
require('views/main/service/reassign/step7_view');
require('views/wizard/step8_view');
require('views/wizard/step9_view');
require('views/wizard/step9/hostLogPopupBody_view');
require('views/wizard/step10_view');
require('views/loading');

require('views/experimental');

});

require.register("views/application", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.ApplicationView = Em.View.extend({
  templateName: require('templates/application'),

  didInsertElement: function () {
    // on 'Enter' pressed, trigger modal window primary button if primary button is enabled(green)
    // on 'Esc' pressed, close the modal
    $(document).keydown(function(event){
      if (event.which == 13 || event.keyCode == 13 ) {
        var primaryButton = $(document).find('#modal > .modal-footer > .btn-success').last();
        if ((!$("*:focus").is("textarea")) && primaryButton.length > 0 && primaryButton.attr('disabled') != 'disabled') {
          event.preventDefault();
          event.stopPropagation();
          primaryButton.click();
          return false;
        }
      }
      return true;
    });
    $(document).keyup(function(event){
      if (event.which == 27 || event.keyCode == 27) {
        var closeButton = $(document).find('#modal > .modal-header > .close').last();
        if (closeButton.length > 0) {
          event.preventDefault();
          event.stopPropagation();
          closeButton.click();
          return false;
        }
      }
      return true;
    });
  }
});
});

require.register("views/common/ajax_default_error_popup_body", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.AjaxDefaultErrorPopupBodyView = Em.View.extend({

  classNames: ['api-error'],
  templateName: require('templates/utils/ajax'),

  /**
   * HTTP request URL
   * @type {string}
   */
  url: '',

  /**
   * HTTP request type
   * @type {string}
   */
  type: '',

  /**
   * HTTP response status code
   * @type {number}
   */
  status: 0,

  /**
   * Received error message
   * @type {string}
   */
  message: '',

  /**
   * Status code string
   * @type {string}
   */
  statusCode: function () {
    return Em.I18n.t('utils.ajax.defaultErrorPopupBody.statusCode').format(this.get('status'));
  }.property('status'),

  /**
   * Indicates if error message should be displayed
   * @type {boolean}
   */
  showMessage: function () {
    return !!this.get('message');
  }.property('message'),

  /**
   * HTTP response error description
   * @type {string}
   */
  api: function () {
    return Em.I18n.t('utils.ajax.defaultErrorPopupBody.message').format(this.get('type'), this.get('url'));
  }.property('type', 'url')

});

});

require.register("views/common/chart", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ChartView = Em.View.extend({
  dateFormat:'dd/mm/yy',
  timeFormat:'h:m',
  w:900,
  p:30, // axis padding
  shift:30,
  ticksCount:10,
  pointsLimit:300,
  areaHeight:30, // px
  axis:false,
  x:false,
  y:false,

  init:function () {
    this._super();
    var renderer = this;

    this.x = d3.time.scale().domain([renderer.getMinDate({}), renderer.getMaxDate({})]).range([0, this.get('w')]);
    this.y = d3.scale.linear().domain([0, 50]).range([this.get('h'), 0]);
    this.axis = d3.svg.axis().orient("top").scale(this.x).ticks(this.get('ticksCount'));
  },

  h: function(){
    return this.get('p') + this.get('nodeAttributes').length * this.get('areaHeight'); //default: 160
  }.property('nodeAttributes', 'p'),

  activeH: function(){
    return this.get('nodeAttributes').length * this.get('areaHeight'); // default 160;
  }.property('h'),

  ruleHeight: function(){
    return this.get('nodeAttributes').length * this.get('areaHeight');
  }.property('nodeAttributes'),

  updateY: function(){
    this.y = d3.scale.linear().domain([0, 50]).range([this.get('h'), 0]);
  }.observes('h'),

  getMinDate:function (data) {
    if (data.length)
      return new Date(Date.parse(data[0]['date']));

    return new Date();
  },

  getMaxDate:function (data) {
    if (data.length)
      return new Date(Date.parse(data[data.length - 1]['date']));

    return new Date();
  },

  area:function () {
    var renderer = this;
    var area = d3.svg.area().x(function (d) {
      return renderer.x(renderer.getDate(d));
    });

    area.y1(function (d) {
      return renderer.get('h') - (renderer.get('h') - renderer.y(d[$(this).attr("getter")])) / renderer.get('koef');
    });

    area.y0(function (d) {
      return renderer.get('h');
    });

    return area;
  },

  line:function () {
    var renderer = this;
    var area = d3.svg.line().x(function (d) {
      return renderer.x(renderer.getDate(d));
    })
      .interpolate("basis");

    area.y(function (d) {
      return renderer.get('h');
    });

    return area;
  },

  /**
   * @todo: calculate this
   * coefficient of compression
   * @param shift
   * @return {Number}
   */

  koef:function () {
    // max value divide on area height;
    return 2 * (this.get('nodeAttributes').length + 1);
  }.property('h'),

  getDate:function (d) {
    return new Date(Date.parse(d.date));
  },

  dateTimeToDateObject:function (string) {
    var ren = this;
    return new Date($.datepicker.parseDateTime(ren.dateFormat, ren.timeFormat, string));
  },

  getDefaultShift:function () {
    return  -1 * this.get('areaHeight') * (this.get('nodeAttributes').length - 1);
  },

  percentScaleXDefaultTranslate:function () {
    return this.w + 3
  },

  clearPlot: function(){
    d3.select(this.get('chartContainerSelector')).selectAll("*").remove();
  },

  drawPlot:function () {
    this.clearPlot();

    var renderer = this;
    this.x.domain([renderer.getMinDate({}), renderer.getMaxDate({})]);

    var rule = $('<div></div>').addClass("rule").css('height', renderer.get('ruleHeight')).mouseenter(function () { $(this).hide(); });
    $(this.get('chartContainerSelector')).prepend(rule);

    var vis = d3.select(this.get('chartContainerSelector'))
      .append("svg:svg")
      .attr("width", renderer.get('w') + 5)
      .attr("height", renderer.get('h'))
      .attr("rendererId", this.get('elementId'))
      .on("mousemove", function () {

        var area = d3.select(this).select("path.line");
        var d = area.data()[0];
        var x = d3.mouse(this)[0];

        var renderer = Em.View.views[d3.select(this).attr('rendererId')];
        var container = $(this).parent();
        var scale = renderer.x;

        // first move rule
        var rule = $(container).children("div.rule");
        rule.css("left", (168 + x) + "px"); // 168 - left container margin
        rule.show();

        x = x + 5; // some correction
        var selectedDate = scale.invert(x);

        // search date between this coordinates
        var prevVal = false;
        var nextVal = d[0];

        $.each(d, function (i, point) {
          if (renderer.getDate(point).getTime() <= selectedDate.getTime()) {
            prevVal = nextVal;
            nextVal = point;
          }
        });

        var len1 = Math.abs(x - scale(renderer.getDate(prevVal)));
        var len2 = Math.abs(x - scale(renderer.getDate(nextVal)));

        var clearing = 5;
        var pointToShow = false;
        // if the first point if closer
        if ((len1 < len2) && (len1 <= clearing)) {
          pointToShow = prevVal;
        } else if (len2 <= clearing) { // the second point is closer
          pointToShow = nextVal;
        }

        $.each(renderer.get('nodeAttributes'), function (i, v) {
          var value = !pointToShow ? "" : pointToShow[v] + "%";
          $(rule).children("div." + v).html(value);
        });
      });

    vis.append("svg:g")
      .attr("class", "axis")
      .attr("transform", "translate(0," + this.get('p') + ")")
      .call(renderer.axis);

    $.each(this.get('nodeAttributes'), function (i, v) {
      var element = $('<div></div>').addClass(v).addClass("stateValue").html("");
      rule.append(element);
    });

    var shift = this.getDefaultShift();
    vis.append("svg:path")
      .attr("class", "horizontal-line")
      .data([
      {}
    ])
      .attr("transform", "translate(0," + (shift - this.get('areaHeight')) + ")")
      .attr("d", renderer.line())
      .style("stroke", "#000");

    $.each(this.get('nodeAttributes'), function (i, v) {
      vis.append("svg:path").data([
        {}
      ])
        .attr("class", "line")
        .attr("getter", v)
        .attr("transform", "translate(0, " + shift + ")")
        .attr("d", renderer.area())
        .style("fill", function () {
          return "#31a354";
        });

      vis.append("svg:path")
        .attr("class", "horizontal-line")
        .data([
        {}
      ])
        .attr("transform", "translate(0," + shift + ")")
        .attr("d", renderer.line())
        .style("stroke", "#000");

      shift += renderer.get('areaHeight');
    });
  },

  getData:function (containerId) {
    return (d3.select(containerId + " path.line").data())[0];
  },

  drawChart:function () {
    var containerSel = this.get('chartContainerSelector');
    var data = this.get('data');

    while (data.length > this.get('pointsLimit')) {
      data.shift();
    }

    var renderer = this;
    var minDate = this.getMinDate(data);
    var maxDate = this.getMaxDate(data);

    this.x.domain([minDate, maxDate]);

    var ticks = data.length > 10 ? 10 : data.length;
    this.axis.scale(renderer.x).ticks(ticks);

//        remove dots axis
    $(containerSel + " svg g.axis g").remove();
    d3.select(containerSel + " svg g.axis")
      .call(this.axis);

    $.each(this.get('nodeAttributes'), function (i, v) {
      d3.select(containerSel + " path.line[getter='" + v + "']")
        .data([data])
        .transition()
        .attr("d", renderer.area());
    });

    // lines between charts
    $(containerSel + " path.horizontal-line").each(
      function (i, path) {
        d3.select(path).data([
          [
            {date:minDate},
            {date:maxDate}
          ]
        ]).attr("d", renderer.line());
      }
    );
  }
});
});

require.register("views/common/chart/linear", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ChartLinearView = Em.View.extend({
  w:90,
  h:90,
  margin:2,

  classNames: ['linear'],

  init:function () {
    this._super();

    this.x = d3.scale.linear().domain([0, this.get('data').length]).range([0 + this.get('margin'), this.get('w') - this.get('margin')]);
    this.y = d3.scale.linear().domain([0, d3.max(this.get('data'))]).range([0 + this.get('margin'), this.get('h') - this.get('margin')]);
    this.line = d3.svg.line().x(function (d, i) { return this.x(i); }).y(function (d) {return -1 * this.y(d); })
  },

  didInsertElement:function () {
    this._super();
    this.appendSvg();
  },

  selector:function () {
    return '#' + this.get('elementId');
  }.property('elementId'),

  appendSvg:function () {
    var thisChart = this;

    this.set('svg', d3.select(this.get('selector'))
      .append("svg:svg")
      .attr("width", thisChart.get('w'))
      .attr("height", thisChart.get('h')));

    this.set('g', thisChart.get('svg').append("svg:g").attr("transform", "translate(0, " + thisChart.get('h') + ")"));
    this.get('g').append("svg:path").attr("d", thisChart.line(thisChart.get('data')));


    // axis
    this.get('g').append("svg:line")
      .attr("x1", thisChart.x(0))
      .attr("y1", -1 * thisChart.y(0))
      .attr("x2", thisChart.x(this.get('data').length))
      .attr("y2", -1 * thisChart.y(0));

    this.get('g').append("svg:line")
      .attr("x1", thisChart.x(0))
      .attr("y1", -1 * thisChart.y(0))
      .attr("x2", thisChart.x(0))
      .attr("y2", -1 * thisChart.y(d3.max(thisChart.get('data'))))
  }
});
});

require.register("views/common/chart/linear_time", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
var string_utils = require('utils/string_utils');

/**
 * @class
 *
 * This is a view which GETs data from a URL and shows it as a time based line
 * graph. Time is shown on the X axis with data series shown on Y axis. It
 * optionally also has the ability to auto refresh itself over a given time
 * interval.
 *
 * This is an abstract class which is meant to be extended.
 *
 * Extending classes should override the following:
 * <ul>
 * <li>url - from where the data can be retrieved
 * <li>title - Title to be displayed when showing the chart
 * <li>id - which uniquely identifies this chart in any page
 * <li>#transformToSeries(jsonData) - function to map server data into graph
 * series
 * </ul>
 *
 * Extending classes could optionally override the following:
 * <ul>
 * <li>#colorForSeries(series) - function to get custom colors per series
 * </ul>
 *
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartLinearTimeView = Ember.View.extend({
  templateName: require('templates/main/charts/linear_time'),

  /**
   * The URL from which data can be retrieved.
   *
   * This property must be provided for the graph to show properly.
   *
   * @type String
   * @default null
   */
  url: null,

  /**
   * A unique ID for this chart.
   *
   * @type String
   * @default null
   */
  id: null,

  /**
   * Title to be shown under the chart.
   *
   * @type String
   * @default null
   */
  title: null,

  /**
   * @private
   *
   * @type Rickshaw.Graph
   * @default null
   */
  _graph: null,

  /**
   * Array of classnames for each series (in widget)
   * @type Rickshaw.Graph
   */
  _popupGraph: null,

  /**
   * Array of classnames for each series
   * @type Array
   */
  _seriesProperties: null,

  /**
   * Array of classnames for each series (in widget)
   * @type Array
   */
  _seriesPropertiesWidget: null,

  /**
   * Renderer type
   * See <code>Rickshaw.Graph.Renderer</code> for more info
   * @type String
   */
  renderer: 'area',

  /**
   * Suffix used in DOM-elements selectors
   * @type String
   */
  popupSuffix: '-popup',

  /**
   * Is popup for current graph open
   * @type Boolean
   */
  isPopup: false,

  /**
   * Is graph ready
   * @type Boolean
   */
  isReady: false,

  /**
   * Is popup-graph ready
   * @type Boolean
   */
  isPopupReady: false,

  /**
   * Is data for graph available
   * @type Boolean
   */
  hasData: true,

  didInsertElement: function () {
    this.loadData();
    this.registerGraph();
  },

  registerGraph: function() {
    var graph = {
      name: this.get('title'),
      id: this.get('elementId'),
      popupId: this.get('id')
    };
    App.router.get('updateController.graphs').push(graph);
  },

  loadData: function() {
    App.ajax.send({
      name: this.get('ajaxIndex'),
      sender: this,
      data: this.getDataForAjaxRequest(),
      success: '_refreshGraph',
      error: 'loadDataErrorCallback'
    });
  },

  getDataForAjaxRequest: function() {
    var toSeconds = Math.round(App.dateTime() / 1000);
    var hostName = (this.get('content')) ? this.get('content.hostName') : "";

    var HDFSService = App.HDFSService.find().objectAt(0);
    var nameNodeName = "";
    var YARNService = App.YARNService.find().objectAt(0);
    var resourceManager = YARNService ? YARNService.get('resourceManager.hostName') : "";
    var timeUnit = this.get('timeUnitSeconds');
    if (HDFSService) {
      nameNodeName = (HDFSService.get('activeNameNode')) ? HDFSService.get('activeNameNode.hostName') : HDFSService.get('nameNode.hostName');
    }
    return {
      toSeconds: toSeconds,
      fromSeconds: toSeconds - timeUnit,
      stepSeconds: 15,
      hostName: hostName,
      nameNodeName: nameNodeName,
      resourceManager: resourceManager
    };
  },

  loadDataErrorCallback: function(xhr, textStatus, errorThrown) {
    this.set('isReady', true);
    if (xhr.readyState == 4 && xhr.status) {
      textStatus = xhr.status + " " + textStatus;
    }
    this._showMessage('warn', this.t('graphs.error.title'), this.t('graphs.error.message').format(textStatus, errorThrown));
    this.set('isPopup', false);
    this.set('hasData', false);
  },

  /**
   * Shows a yellow warning message in place of the chart.
   *
   * @param type  Can be any of 'warn', 'error', 'info', 'success'
   * @param title Bolded title for the message
   * @param message String representing the message
   * @type: Function
   */
  _showMessage: function(type, title, message) {
    var chartOverlay = '#' + this.id;
    var chartOverlayId = chartOverlay + '-chart';
    var chartOverlayY = chartOverlay + '-yaxis';
    var chartOverlayX = chartOverlay + '-xaxis';
    var chartOverlayLegend = chartOverlay + '-legend';
    var chartOverlayTimeline = chartOverlay + '-timeline';
    if (this.get('isPopup')) {
      chartOverlayId += this.get('popupSuffix');
      chartOverlayY += this.get('popupSuffix');
      chartOverlayX += this.get('popupSuffix');
      chartOverlayLegend += this.get('popupSuffix');
      chartOverlayTimeline += this.get('popupSuffix');
    }
    var typeClass;
    switch (type) {
      case 'error':
        typeClass = 'alert-error';
        break;
      case 'success':
        typeClass = 'alert-success';
        break;
      case 'info':
        typeClass = 'alert-info';
        break;
      default:
        typeClass = '';
        break;
    }
    $(chartOverlayId+', '+chartOverlayY+', '+chartOverlayX+', '+chartOverlayLegend+', '+chartOverlayTimeline).html('');
    $(chartOverlayId).append('<div class=\"alert '+typeClass+'\"><strong>'+title+'</strong> '+message+'</div>');
  },

  /**
   * Transforms the JSON data retrieved from the server into the series
   * format that Rickshaw.Graph understands.
   *
   * The series object is generally in the following format: [ { name :
   * "Series 1", data : [ { x : 0, y : 0 }, { x : 1, y : 1 } ] } ]
   *
   * Extending classes should override this method.
   *
   * @param seriesData
   *          Data retrieved from the server
   * @param displayName
   *          Graph title
   * @type: Function
   *
   */
  transformData: function (seriesData, displayName) {
    var seriesArray = [];
    if (seriesData != null) {
      // Is it a string?
      if ("string" == typeof seriesData) {
        seriesData = JSON.parse(seriesData);
      }
      // Is it a number?
      if ("number" == typeof seriesData) {
        // Same number applies to all time.
        var number = seriesData;
        seriesData = [];
        seriesData.push([number, App.dateTime()-(60*60)]);
        seriesData.push([number, App.dateTime()]);
      }
      // We have valid data
      var series = {};
      series.name = displayName;
      series.data = [];
      for ( var index = 0; index < seriesData.length; index++) {
        series.data.push({
          x: seriesData[index][1],
          y: seriesData[index][0]
        });
      }
      return series;
    }
    return null;
  },

  /**
   * Provides the formatter to use in displaying Y axis.
   *
   * Uses the App.ChartLinearTimeView.DefaultFormatter which shows 10K,
   * 300M etc.
   *
   * @type Function
   */
  yAxisFormatter: function(y) {
    return App.ChartLinearTimeView.DefaultFormatter(y);
  },

  /**
   * Provides the color (in any HTML color format) to use for a particular
   * series.
   * May be redefined in child views
   *
   * @param series
   *          Series for which color is being requested
   * @return color String. Returning null allows this chart to pick a color
   *         from palette.
   * @default null
   * @type Function
   */
  colorForSeries: function (series) {
    return null;
  },

  /**
  * Check whether seriesData is correct data for chart drawing
  * @param {Array} seriesData
  * @return {Boolean}
  */
  checkSeries : function(seriesData) {
    if(!seriesData || !seriesData.length) {
      return false;
    }
    var result = true;
    seriesData.forEach(function(item) {
      if(!item.data || !item.data.length || !item.data[0] || typeof item.data[0].x === 'undefined') {
        result = false;
      }
    });
    return result;
  },

  /**
   * @private
   *
   * Refreshes the graph with the latest JSON data.
   *
   * @type Function
   */
  _refreshGraph: function (jsonData) {
    if(this.get('isDestroyed')){
      return;
    }
    var seriesData = this.transformToSeries(jsonData);

      //if graph opened as modal popup
      var popup_path = $("#" + this.get('id') + "-container" + this.get('popupSuffix'));
      var graph_container = $("#" + this.get('id') + "-container");
      if(popup_path.length) {
        popup_path.children().each(function () {
          $(this).children().remove();
        });
        this.set('isPopup', true);
      }
      else {
        graph_container.children().each(function (index, value) {
          $(value).children().remove();
        });
      }
    if (this.checkSeries(seriesData)) {
      // Check container exists (may be not, if we go to another page and wait while graphs loading)
      if (graph_container.length) {
        this.draw(seriesData);
        this.set('hasData', true);
          //move yAxis value lower to make them fully visible
        $("#" + this.id + "-container").find('.y_axis text').attr('y',8);
      }
    }
    else {
      this.set('isReady', true);
      //if Axis X time interval is default(60 minutes)
      if(this.get('timeUnitSeconds') === 3600){
        this._showMessage('info', this.t('graphs.noData.title'), this.t('graphs.noData.message'));
        this.set('hasData', false);
      }
      else {
        this._showMessage('info', this.t('graphs.noData.title'), this.t('graphs.noDataAtTime.message'));
      }
      this.set('isPopup', false);
    }
  },

  /**
   * Returns a custom time unit, that depends on X axis interval length, for the graph's X axis.
   * This is needed as Rickshaw's default time X axis uses UTC time, which can be confusing
   * for users expecting locale specific time.
   *
   * If <code>null</code> is returned, Rickshaw's default time unit is used.
   *
   * @type Function
   * @return Rickshaw.Fixtures.Time
   */
  localeTimeUnit: function(timeUnitSeconds) {
    var timeUnit = new Rickshaw.Fixtures.Time();
    switch (timeUnitSeconds){
      case 604800:
        timeUnit = timeUnit.unit('day');
        break;
      case 2592000:
        timeUnit = timeUnit.unit('week');
        break;
      case 31104000:
        timeUnit = timeUnit.unit('month');
        break;
      default:
        timeUnit = {
          name: timeUnitSeconds / 240 + ' minute',
          seconds: timeUnitSeconds / 4,
          formatter: function (d) {
            // format locale specific time
            var minutes = d.getMinutes() > 9 ? "" + d.getMinutes() : "0" + d.getMinutes();
            var hours = d.getHours() > 9 ? "" + d.getHours() : "0" + d.getHours();
            return hours + ":" + minutes;
          }
        };
    }
    return timeUnit;
  },

  /**
   * temporary fix for incoming data for graph
   * to shift data time to correct time point
   * @param {Array} data
   */
  dataShiftFix: function(data) {
    var nowTime = Math.round(App.dateTime() / 1000);
    data.forEach(function(series){
      var l = series.data.length;
      var shiftDiff = nowTime - series.data[l - 1].x;
      if(shiftDiff > 3600){
        for(var i = 0;i < l;i++){
          series.data[i].x = series.data[i].x + shiftDiff;
        }
        series.data.unshift({
          x: nowTime - this.get('timeUnitSeconds'),
          y: 0
        });
      }
    }, this);
  },

  /**
   * calculate statistic data for popup legend and set proper colors for series
   * @param {Array} data
   */
  dataPreProcess: function(data) {
    var self = this;
    var palette = new Rickshaw.Color.Palette({ scheme: 'munin'});
    // Format series for display
    var series_min_length = 100000000;
    data.forEach(function (series, index) {
      var seriesColor = self.colorForSeries(series);
      if (seriesColor == null) {
        seriesColor = palette.color();
      }
      series.color = seriesColor;
      series.stroke = 'rgba(0,0,0,0.3)';
      if (this.get('isPopup')) {
        // calculate statistic data for popup legend
        var avg = 0;
        var min = Number.MAX_VALUE;
        var max = Number.MIN_VALUE;
        for (var i = 0; i < series.data.length; i++) {
          avg += series.data[i]['y'];
          if (series.data[i]['y'] < min) {
            min = series.data[i]['y'];
          }
          else {
            if (series.data[i]['y'] > max) {
              max = series.data[i]['y'];
            }
          }
        }
        series.name = string_utils.pad(series.name, 30, '&nbsp;', 2) +
                      string_utils.pad('min', 5, '&nbsp;', 3) +
                      string_utils.pad(this.get('yAxisFormatter')(min), 12, '&nbsp;', 3) +
                      string_utils.pad('avg', 5, '&nbsp;', 3) +
                      string_utils.pad(this.get('yAxisFormatter')(avg/series.data.length), 12, '&nbsp;', 3) +
                      string_utils.pad('max', 12, '&nbsp;', 3) +
                      string_utils.pad(this.get('yAxisFormatter')(max), 5, '&nbsp;', 3);
      }
      if (series.data.length < series_min_length) {
        series_min_length = series.data.length;
      }
    }.bind(this));

    // All series should have equal length
    data.forEach(function(series, index) {
      if (series.data.length > series_min_length) {
        series.data.length = series_min_length;
      }
    });
  },

  draw: function(seriesData) {
    var self = this;
    var isPopup = this.get('isPopup');
    var p = isPopup ? this.get('popupSuffix') : '';

    this.dataShiftFix(seriesData);
    this.dataPreProcess(seriesData);

    var chartId = "#" + this.get('id') + "-chart" + p;
    var chartOverlayId = "#" + this.get('id') + "-container" + p;
    var xaxisElementId = "#" + this.get('id') + "-xaxis" + p;
    var yaxisElementId = "#" + this.get('id') + "-yaxis" + p;
    var legendElementId = "#" + this.get('id') + "-legend" + p;

    var chartElement = document.querySelector(chartId);
    var overlayElement = document.querySelector(chartOverlayId);
    var xaxisElement = document.querySelector(xaxisElementId);
    var yaxisElement = document.querySelector(yaxisElementId);
    var legendElement = document.querySelector(legendElementId);

    var height = 150;
    var width = 400;
    var diff = 32;

    if(this.get('inWidget')) {
      height = 105; // for widgets view
      diff = 22;
    }
    if (isPopup) {
      height = 180;
      width = 670;
    }
    else {
      // If not in popup, the width could vary.
      // We determine width based on div's size.
      var thisElement = this.get('element');
      if (thisElement!=null) {
        var calculatedWidth = $(thisElement).width();
        if (calculatedWidth > diff) {
          width = calculatedWidth - diff;
        }
      }
    }

    var _graph = new Rickshaw.Graph({
      height: height,
      width: width,
      element: chartElement,
      series: seriesData,
      interpolation: 'step-after',
      stroke: true,
      renderer: this.get('renderer'),
      strokeWidth: (this.get('renderer') != 'area' ? 2 : 1)
    });

    if (this.get('renderer') === 'area') {
      _graph.renderer.unstack = false;
    }

    new Rickshaw.Graph.Axis.Time({
      graph: _graph,
      timeUnit: this.localeTimeUnit(this.get('timeUnitSeconds'))
    });

    new Rickshaw.Graph.Axis.Y({
      tickFormat: this.yAxisFormatter,
      element: yaxisElement,
      orientation: (isPopup ? 'left' : 'right'),
      graph: _graph
    });

    var legend = new Rickshaw.Graph.Legend({
      graph: _graph,
      element: legendElement
    });

    new Rickshaw.Graph.Behavior.Series.Toggle({
      graph: _graph,
      legend: legend
    });

    new Rickshaw.Graph.Behavior.Series.Order({
      graph: _graph,
      legend: legend
    });

    if (!isPopup) {
      overlayElement.addEventListener('mousemove', function () {
        $(xaxisElement).removeClass('hide');
        $(legendElement).removeClass('hide');
        $(chartElement).children("div").removeClass('hide');
      });
      overlayElement.addEventListener('mouseout', function () {
        $(legendElement).addClass('hide');
      });
      _graph.onUpdate(function () {
        $(legendElement).addClass('hide');
      });
    }

    //show the graph when it's loaded
    _graph.onUpdate(function() {
      self.set('isReady', true);
    });
    _graph.render();

    if (isPopup) {
      new Rickshaw.Graph.HoverDetail({
        graph: _graph,
        yFormatter:function (y) {
          return self.yAxisFormatter(y);
        },
        xFormatter:function (x) {
          return (new Date(x * 1000)).toLocaleTimeString();
        },
        formatter:function (series, x, y, formattedX, formattedY, d) {
          return formattedY + '<br />' + formattedX;
        }
      });
    }

    _graph = this.updateSeriesInGraph(_graph);
    if (isPopup) {
      //show the graph when it's loaded
      _graph.onUpdate(function() {
        self.set('isPopupReady', true);
      });
      _graph.update();

      var selector = '#'+this.get('id')+'-container'+this.get('popupSuffix');
      $(selector + ' li.line').click(function() {
        var series = [];
        $(selector + ' a.action').each(function(index, v) {
          series[index] = v.parentNode.classList;
        });
        self.set('_seriesProperties', series);
      });

      this.set('_popupGraph', _graph);
    }
    else {
      _graph.update();
      var selector = '#'+this.get('id')+'-container';
      $(selector + ' li.line').click(function() {
        var series = [];
        $(selector + ' a.action').each(function(index, v) {
          series[index] = v.parentNode.classList;
        });
        self.set('_seriesPropertiesWidget', series);
      });

      this.set('_graph', _graph);
    }
  },

  /**
   *
   * @param {Rickshaw.Graph} graph
   * @returns {Rickshaw.Graph}
   */
  updateSeriesInGraph: function(graph) {
    var id = this.get('id');
    var isPopup = this.get('isPopup');
    var popupSuffix = this.get('popupSuffix');
    var _series = isPopup ? this.get('_seriesProperties') : this.get('_seriesPropertiesWidget');
    graph.series.forEach(function(series, index) {
      if (_series !== null && _series[index] !== null && _series[index] !== undefined ) {
        if(_series[_series.length - index - 1].length > 1) {
          var s = '#' + id + '-container' + (isPopup ? popupSuffix : '') + ' a.action:eq(' + (_series.length - index - 1) + ')';
          $(s).parent('li').addClass('disabled');
          series.disable();
        }
      }
    });
    return graph;
  },

  showGraphInPopup: function() {
    if(!this.get('hasData')) {
      return;
    }

    this.set('isPopup', true);
    var self = this;

    App.ModalPopup.show({
      bodyClass: Em.View.extend({

        containerId: null,
        containerClass: null,
        yAxisId: null,
        yAxisClass: null,
        xAxisId: null,
        xAxisClass: null,
        legendId: null,
        legendClass: null,
        chartId: null,
        chartClass: null,
        titleId: null,
        titleClass: null,

        isReady: function() {
          return this.get('parentView.graph.isPopupReady');
        }.property('parentView.graph.isPopupReady'),

        didInsertElement: function() {
          $('#modal').addClass('modal-graph-line');
          var popupSuffix = this.get('parentView.graph.popupSuffix');
          var id = this.get('parentView.graph.id');
          var idTemplate = id + '-{element}' + popupSuffix;

          this.set('containerId', idTemplate.replace('{element}', 'container'));
          this.set('containerClass', 'chart-container' + popupSuffix);
          this.set('yAxisId', idTemplate.replace('{element}', 'yaxis'));
          this.set('yAxisClass', this.get('yAxisId').replace(popupSuffix, ''));
          this.set('xAxisId', idTemplate.replace('{element}', 'xaxis'));
          this.set('xAxisClass', this.get('xAxisId').replace(popupSuffix, ''));
          this.set('legendId', idTemplate.replace('{element}', 'legend'));
          this.set('legendClass', this.get('legendId').replace(popupSuffix, ''));
          this.set('chartId', idTemplate.replace('{element}', 'chart'));
          this.set('chartClass', this.get('chartId').replace(popupSuffix, ''));
          this.set('titleId', idTemplate.replace('{element}', 'title'));
          this.set('titleClass', this.get('titleId').replace(popupSuffix, ''));
        },

        templateName: require('templates/common/chart/linear_time'),
        /**
         * check is time paging feature is enable for graph
         */
        isTimePagingEnable: function() {
          return !self.get('isTimePagingDisable');
        }.property(),
        rightArrowVisible: function() {
          return (this.get('isReady') && (this.get('parentView.currentTimeIndex') != 0));
        }.property('isReady', 'parentView.currentTimeIndex'),
        leftArrowVisible: function() {
          return (this.get('isReady') && (this.get('parentView.currentTimeIndex') != 7));
        }.property('isReady', 'parentView.currentTimeIndex')
      }),
      header: this.get('title'),
      /**
       * App.ChartLinearTimeView
       */
      graph: self,
      secondary: null,
      onPrimary: function() {
        this.hide();
        self.set('isPopup', false);
      },
      onClose: function() {
        this.onPrimary();
      },
      /**
       * move graph back by time
       * @param event
       */
      switchTimeBack: function(event) {
        var index = this.get('currentTimeIndex');
        // 7 - number of last time state
        if(index < 7) {
          this.reloadGraphByTime(++index);
        }
      },
      /**
       * move graph forward by time
       * @param event
       */
      switchTimeForward: function(event) {
        var index = this.get('currentTimeIndex');
        if(index > 0) {
          this.reloadGraphByTime(--index);
        }
      },
      /**
       * reload graph depending on the time
       * @param index
       */
      reloadGraphByTime: function(index) {
        this.set('currentTimeIndex', index);
        self.set('currentTimeIndex', index);
        self.loadData();
      },
      currentTimeIndex: self.get('currentTimeIndex'),
      currentTimeState: function() {
        return self.get('timeStates').objectAt(this.get('currentTimeIndex'));
      }.property('currentTimeIndex')
    });
    Ember.run.next(function() {
      self.loadData();
      self.set('isPopupReady', false);
    });
  },
  timeStates: [
    {name: Em.I18n.t('graphs.timeRange.hour'), seconds: 3600},
    {name: Em.I18n.t('graphs.timeRange.twoHours'), seconds: 7200},
    {name: Em.I18n.t('graphs.timeRange.fourHours'), seconds: 14400},
    {name: Em.I18n.t('graphs.timeRange.twelveHours'), seconds: 43200},
    {name: Em.I18n.t('graphs.timeRange.day'), seconds: 86400},
    {name: Em.I18n.t('graphs.timeRange.week'), seconds: 604800},
    {name: Em.I18n.t('graphs.timeRange.month'), seconds: 2592000},
    {name: Em.I18n.t('graphs.timeRange.year'), seconds: 31104000}
  ],
  // should be set by time range control dropdown list when create current graph
  currentTimeIndex: 0,
  timeUnitSeconds: function() {
    return this.get('timeStates').objectAt(this.get('currentTimeIndex')).seconds;
  }.property('currentTimeIndex')
});

/**
 * A formatter which will turn a number into computer storage sizes of the
 * format '23 GB' etc.
 *
 * @type {Function}
 */
App.ChartLinearTimeView.BytesFormatter = function (y) {
  if (y == 0) return '0 B';
  var value = Rickshaw.Fixtures.Number.formatBase1024KMGTP(y);
  if (!y || y.length < 1) {
    value = '0 B';
  }
  else {
    if ("number" == typeof value) {
      value = String(value);
    }
    if ("string" == typeof value) {
      value = value.replace(/\.\d(\d+)/, function($0, $1){ // Remove only 1-digit after decimal part
        return $0.replace($1, '');
      });
      // Either it ends with digit or ends with character
      value = value.replace(/(\d$)/, '$1 '); // Ends with digit like '120'
      value = value.replace(/([a-zA-Z]$)/, ' $1'); // Ends with character like
      // '120M'
      value = value + 'B'; // Append B to make B, MB, GB etc.
    }
  }
  return value;
};

/**
 * A formatter which will turn a number into percentage display like '42%'
 *
 * @type {Function}
 */
App.ChartLinearTimeView.PercentageFormatter = function (percentage) {
  var value = percentage;
  if (!value || value.length < 1) {
    value = '0 %';
  } else {
    value = value.toFixed(3).replace(/0+$/, '').replace(/\.$/, '') + '%';
  }
  return value;
};

/**
 * A formatter which will turn elapsed time into display time like '50 ms',
 * '5s', '10 m', '3 hr' etc. Time is expected to be provided in milliseconds.
 *
 * @type {Function}
 */
App.ChartLinearTimeView.TimeElapsedFormatter = function (millis) {
  var value = millis;
  if (!value || value.length < 1) {
    value = '0 ms';
  } else if ("number" == typeof millis) {
    var seconds = millis > 1000 ? Math.round(millis / 1000) : 0;
    var minutes = seconds > 60 ? Math.round(seconds / 60) : 0;
    var hours = minutes > 60 ? Math.round(minutes / 60) : 0;
    var days = hours > 24 ? Math.round(hours / 24) : 0;
    if (days > 0) {
      value = days + ' d';
    } else if (hours > 0) {
      value = hours + ' hr';
    } else if (minutes > 0) {
      value = minutes + ' m';
    } else if (seconds > 0) {
      value = seconds + ' s';
    } else if (millis > 0) {
      value = millis.toFixed(3).replace(/0+$/, '').replace(/\.$/, '') + ' ms';
    } else {
      value = millis.toFixed(3).replace(/0+$/, '').replace(/\.$/, '') + ' ms';
    }
  }
  return value;
};

/**
 * The default formatter which uses Rickshaw.Fixtures.Number.formatKMBT
 * which shows 10K, 300M etc.
 *
 * @type {Function}
 */
App.ChartLinearTimeView.DefaultFormatter = function(y) {
  if(isNaN(y)){
    return 0;
  }
  var value = Rickshaw.Fixtures.Number.formatKMBT(y);
  if (value == '') return '0';
  value = String(value);
  var c = value[value.length - 1];
  if (!isNaN(parseInt(c))) {
    // c is digit
    value = parseFloat(value).toFixed(3).replace(/0+$/, '').replace(/\.$/, '');
  }
  else {
    // c in not digit
    value = parseFloat(value.substr(0, value.length - 1)).toFixed(3).replace(/0+$/, '').replace(/\.$/, '') + c;
  }
  return value;
};


/**
 * Creates and returns a formatter that can convert a 'value'
 * to 'value units/s'.
 *
 * @param unitsPrefix Prefix which will be used in 'unitsPrefix/s'
 * @param valueFormatter  Value itself will need further processing
 *        via provided formatter. Ex: '10M requests/s'. Generally
 *        should be App.ChartLinearTimeView.DefaultFormatter.
 * @return {Function}
 */
App.ChartLinearTimeView.CreateRateFormatter = function (unitsPrefix, valueFormatter) {
  var suffix = " "+unitsPrefix+"/s";
  return function (value) {
    value = valueFormatter(value) + suffix;
    return value;
  };
};
});

require.register("views/common/chart/pie", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ChartPieView = Em.View.extend({
  w:90,
  h:90,
  data:[300, 500],
  id:null,
  palette: new Rickshaw.Color.Palette({ scheme: 'munin'}),
  stroke: 'black',
  strokeWidth: 2,
  donut:d3.layout.pie().sort(null),
  existCenterText: false,
  centerTextColor: 'black',

  r:function () {
    return Math.min(this.get('w'), this.get('h')) / 2 - this.get('strokeWidth');
  }.property('w', 'h'),

  outerR:function () {
    return this.get('r'); // - 10;
  }.property('r'),

  innerR:function () {
    return 0; // this.get('r') - 20;
  }.property('r'),

  arc:function () {
    return d3.svg.arc().innerRadius(this.get('innerR')).outerRadius(this.get('outerR'));
  }.property(),

  didInsertElement:function () {
    this._super();
    this.appendSvg();
  },

  selector:function () {
    return '#' + this.get('elementId');
  }.property('elementId'),

  appendSvg:function () {

    var thisChart = this;
    var svg = d3.select(thisChart.get('selector')).append("svg:svg")
      .attr("id", thisChart.get('id'))
      .attr("width", thisChart.get('w'))
      .attr("height", thisChart.get('h'))
      .attr("stroke", thisChart.get('stroke'))
      .attr("stroke-width", thisChart.get('strokeWidth'));

    // set percentage data in center if there exist a center text
    if(thisChart.get('existCenterText')){
      this.set('svg', svg
        .append("svg:g")
        .attr("render-order", 1)
        .append("svg:text")
        .attr("stroke", thisChart.get('centerTextColor'))
        .attr("font-size", 17)
        .attr("transform", "translate(" + thisChart.get('w') / 2 + "," + ((thisChart.get('h') / 2) + 3) + ")")
        .attr("text-anchor", "middle")
        .text(function(d) {
                 return thisChart.get('data')[0] + '%';
              })
         );
    }

    this.set('svg', svg
      .append("svg:g")
      .attr("transform", "translate(" + thisChart.get('w') / 2 + "," + thisChart.get('h') / 2 + ")"));

    this.set('arcs', thisChart.get('svg').selectAll("path")
      .data(thisChart.donut(thisChart.get('data')))
      .enter().append("svg:path")
      .attr("fill", function (d, i) {
        return thisChart.palette.color(i);
      })
      .attr("d", thisChart.get('arc'))

    );

  }

});
});

require.register("views/common/configs/compare_property_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
// SCP means ServiceConfigProperty

var App = require('app');

App.ServiceConfigView.SCPComparisonRowsView = Ember.View.extend({
  templateName: require('templates/common/configs/compare_property'),
  serviceConfigProperty: null
});

});

require.register("views/common/configs/config_history_flow", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ConfigHistoryFlowView = Em.View.extend({
  templateName: require('templates/common/configs/config_history_flow'),

  /**
   * index of the first element(service version box) in viewport
   */
  startIndex: 0,
  showLeftArrow: false,
  showRightArrow: false,
  leftArrowTooltip: function () {
    return this.get('showLeftArrow') ? Em.I18n.t('services.service.config.configHistory.leftArrow.tooltip') : null;
  }.property('showLeftArrow'),
  rightArrowTooltip: function () {
    return this.get('showRightArrow') ? Em.I18n.t('services.service.config.configHistory.rightArrow.tooltip') : null;
  }.property('showRightArrow'),
  VERSIONS_IN_FLOW: 6,
  VERSIONS_IN_DROPDOWN: 6,
  /**
   * flag identify whether to show all versions or short list of them
   */
  showFullList: false,
  compareServiceVersion: null,

  /**
   * types of actions that can't be done to service config versions
   */
  actionTypes: {
    SWITCH: 'switchVersion',
    COMPARE: 'compare',
    REVERT: 'revert'
  },

  /**
   * In reason of absence of properties dynamic values support which passed to an action,
   * used property map to get latest values of properties for action
   */
  serviceVersionsReferences: {
    displayed: Em.Object.create({
      isReference: true,
      property: 'displayedServiceVersion'
    }),
    compare: Em.Object.create({
      isReference: true,
      property: 'compareServiceVersion'
    })
  },

  showCompareVersionBar: function() {
    return !Em.isNone(this.get('compareServiceVersion'));
  }.property('compareServiceVersion'),

  isSaveDisabled: function () {
    return (this.get('controller.isSubmitDisabled') || !this.get('controller.versionLoaded') || !this.get('controller.isPropertiesChanged')) ;
  }.property('controller.isSubmitDisabled', 'controller.versionLoaded', 'controller.isPropertiesChanged'),

  serviceName: function () {
    return this.get('controller.selectedService.serviceName');
  }.property('controller.selectedService.serviceName'),

  displayedServiceVersion: function () {
    return this.get('serviceVersions').findProperty('isDisplayed');
  }.property('serviceVersions.@each.isDisplayed'),
  /**
   * identify whether to show link that open whole content of notes
   */
  showMoreLink: function () {
    //100 is number of symbols that fit into label
    return (this.get('displayedServiceVersion.notes.length') > 100);
  }.property('displayedServiceVersion.notes.length'),
  /**
   * formatted notes ready to display
   */
  shortNotes: function () {
    //100 is number of symbols that fit into label
    if (this.get('showMoreLink')) {
      return this.get('displayedServiceVersion.notes').slice(0, 100) + '...';
    }
    return this.get('displayedServiceVersion.notes');
  }.property('displayedServiceVersion'),

  serviceVersions: function () {
    var allServiceVersions = App.ServiceConfigVersion.find().filterProperty('serviceName', this.get('serviceName'));
    var groupName = this.get('controller.selectedConfigGroup.isDefault') ? 'default'
        : this.get('controller.selectedConfigGroup.name');

    allServiceVersions.forEach(function (version) {
      version.set('isDisabled', !(version.get('groupName') === groupName));
    }, this);

    var serviceVersions = allServiceVersions.filter(function(s) {
      return s.get('groupName') == groupName || s.get('groupName') == 'default';
    });

    return serviceVersions.sort(function (a, b) {
      return Em.get(b, 'createTime') - Em.get(a, 'createTime');
    });
  }.property('serviceName', 'controller.selectedConfigGroup.name'),

  /**
   * service versions which in viewport and visible to user
   */
  visibleServiceVersion: function () {
    return this.get('serviceVersions').slice(this.get('startIndex'), (this.get('startIndex') + this.VERSIONS_IN_FLOW));
  }.property('startIndex', 'serviceVersions'),

  /**
   * enable actions to manipulate version only after it's loaded
   */
  versionActionsDisabled: function () {
    return !this.get('controller.versionLoaded') || this.get('dropDownList.length') === 0;
  }.property('controller.versionLoaded', 'dropDownList.length'),

  /**
   * enable discard to manipulate version only after it's loaded and any property is changed
   */
  isDiscardDisabled: function () {
    return !this.get('controller.versionLoaded') || !this.get('controller.isPropertiesChanged');
  }.property('controller.versionLoaded','controller.isPropertiesChanged'),
  /**
   * list of service versions
   * by default 6 is number of items in short list
   */
  dropDownList: function () {
    var serviceVersions = this.get('serviceVersions').slice(0);
    if (this.get('showFullList')) {
      return serviceVersions;
    }
    return serviceVersions.slice(0, this.VERSIONS_IN_DROPDOWN);
  }.property('serviceVersions', 'showFullList', 'displayedServiceVersion'),

  openFullList: function (event) {
    event.stopPropagation();
    this.set('showFullList', true);
  },
  hideFullList: function (event) {
    this.set('showFullList', !(this.get('serviceVersions.length') > this.VERSIONS_IN_DROPDOWN));
  },

  didInsertElement: function () {
    App.tooltip(this.$('[data-toggle=tooltip]'),{
      placement: 'bottom'
    });
    App.tooltip(this.$('[data-toggle=arrow-tooltip]'),{
      placement: 'top'
    });
  },

  serviceVersionBox: Em.View.extend({
    actionTypes: function() {
      return this.get('parentView.actionTypes');
    }.property('parentView.actionTypes'),
    templateName: require('templates/common/configs/service_version_box'),
    didInsertElement: function () {
      $('.version-box').hoverIntent(function() {
        var self = this;
        setTimeout(function() {
          if ($(self).is(':hover')) {
            $(self).find('.version-popover').fadeIn(200);
          }
        }, 700);
      }, function() {
        $(this).find('.version-popover').hide();
      });
      App.tooltip(this.$('[data-toggle=tooltip]'),{
        placement: 'bottom'
      });
      App.tooltip(this.$('[data-toggle=arrow-tooltip]'),{
        placement: 'top'
      });
    }
  }),

  willInsertElement: function () {
    var serviceVersions = this.get('serviceVersions');
    var startIndex = 0;
    var currentIndex = 0;
    var selectedVersion = this.get('controller.selectedVersion');

    serviceVersions.setEach('isDisplayed', false);

    serviceVersions.forEach(function (serviceVersion, index) {
      if (selectedVersion === serviceVersion.get('version')) {
        serviceVersion.set('isDisplayed', true);
        currentIndex = index;
      }
    }, this);

    // show current version as the last one
    if (currentIndex + 1 > this.VERSIONS_IN_FLOW) {
      startIndex = currentIndex + 1 - this.VERSIONS_IN_FLOW;
    }
    this.set('startIndex', startIndex);
    this.adjustFlowView();
    this.keepInfoBarAtTop();
  },

  onChangeConfigGroup: function () {
    var serviceVersions = this.get('serviceVersions');
    var selectedGroupName = this.get('controller.selectedConfigGroup.name');
    var startIndex = 0;
    var currentIndex = 0;

    serviceVersions.setEach('isDisplayed', false);
    //display the version belongs to current group
    if (this.get('controller.selectedConfigGroup.isDefault')) {
      // display current in default group
      serviceVersions.forEach(function (serviceVersion, index) {
        // find current in default group
        if (serviceVersion.get('isCurrent') && serviceVersion.get('groupName') == Em.I18n.t('dashboard.configHistory.table.configGroup.default')) {
          serviceVersion.set('isDisplayed', true);
          currentIndex = index + 1;
        }
      });
    } else {
      // display current in selected group
      serviceVersions.forEach(function (serviceVersion, index) {
        // find current in selected group
        if (serviceVersion.get('isCurrent') && serviceVersion.get('groupName') == selectedGroupName) {
          serviceVersion.set('isDisplayed', true);
          currentIndex = index + 1;
        }
      });
      // no current version for selected group, show default group current version
      if (currentIndex == 0) {
        serviceVersions.forEach(function (serviceVersion, index) {
          // find current in default group
          if (serviceVersion.get('isCurrent') && serviceVersion.get('groupName') == Em.I18n.t('dashboard.configHistory.table.configGroup.default')) {
            currentIndex = index + 1;
            serviceVersion.set('isDisplayed', true);
          }
        });
      }
    }
    // show current version as the last one
    if (currentIndex > this.VERSIONS_IN_FLOW) {
      startIndex = currentIndex - this.VERSIONS_IN_FLOW;
    }
    this.set('startIndex', startIndex);
    this.adjustFlowView();
    this.keepInfoBarAtTop();
  }.observes('controller.selectedConfigGroup.name'),

  /**
   * initialize event to keep info bar position at the top of the page
   */
  keepInfoBarAtTop: function () {
    var defaultTop, defaultLeft;
    var self = this;
    //reset defaultTop value in closure
    $(window).unbind('scroll');

    $(window).on('scroll', function (event) {
      var infoBar = $('#config_history_flow>.version-info-bar-wrapper');
      var versionSlider = $('#config_history_flow>.version-slider');
      var scrollTop = $(window).scrollTop();
      var scrollLeft = $(window).scrollLeft();
      if (infoBar.length === 0) {
        $(window).unbind('scroll');
        return;
      }
      //290 - default "top" property in px
      defaultTop = defaultTop || (infoBar.get(0).getBoundingClientRect() && infoBar.get(0).getBoundingClientRect().top) || 290;
      // keep the version info bar always aligned to version slider
      defaultLeft = (versionSlider.get(0).getBoundingClientRect() && versionSlider.get(0).getBoundingClientRect().left);
      self.setInfoBarPosition(infoBar, defaultTop, scrollTop, defaultLeft, scrollLeft);
    })
  },
  /**
   * calculate and reset top position of info bar
   * @param infoBar
   * @param defaultTop
   * @param scrollTop
   * @param defaultLeft
   * @param scrollLeft
   */
  setInfoBarPosition: function (infoBar, defaultTop, scrollTop, defaultLeft, scrollLeft) {
    if (scrollTop > defaultTop) {
      infoBar.css('top', '10px');
    } else if (scrollTop > 0) {
      infoBar.css('top', (defaultTop - scrollTop) + 'px');
    } else {
      infoBar.css('top', 'auto');
    }
    if (scrollLeft > 0) {
      infoBar.css('left', defaultLeft);
    } else {
      infoBar.css('left', 'auto');
    }
  },

  /**
   *  define the first element in viewport
   *  change visibility of arrows
   */
  adjustFlowView: function () {
    var startIndex = this.get('startIndex');
    this.get('serviceVersions').forEach(function (serviceVersion, index) {
      serviceVersion.set('first', (index === startIndex));
    });
    this.set('showLeftArrow', (startIndex !== 0));
    this.set('showRightArrow', (this.get('serviceVersions.length') > this.VERSIONS_IN_FLOW) && ((startIndex + this.VERSIONS_IN_FLOW) < this.get('serviceVersions.length')));
  },

  /**
   * check action constraints prior to invoke it
   * @param event
   */
  doAction: function (event) {
    var type = event.contexts[1],
        controller = this.get('controller'),
        self = this;

    if (type === 'switchVersion') {
      if (event.context.get("isDisplayed"))  return;
    } else {
      var isDisabled = event.context ? event.context.get('isDisabled') : false;
      if (isDisabled) return;
    }

    function callback() {
      self[type].call(self, event);
    }

    if (controller.hasUnsavedChanges()) {
      controller.showSavePopup(null, callback);
      return;
    }
    callback();
  },

  /**
   * switch configs view version to chosen
   */
  switchVersion: function (event) {
    var version = event.context.get('version');
    var versionIndex = 0;

    this.set('compareServiceVersion', null);
    this.get('serviceVersions').forEach(function (serviceVersion, index) {
      if (serviceVersion.get('version') === version) {
        serviceVersion.set('isDisplayed', true);
        versionIndex = index;
      } else {
        serviceVersion.set('isDisplayed', false);
      }
    });
    this.shiftFlowOnSwitch(versionIndex);
    this.get('controller').loadSelectedVersion(version);
  },

  /**
   * add config values of chosen version to view for comparison
   * add a second version-info-bar for the chosen version
   */
  compare: function (event) {
    this.set('controller.compareServiceVersion', event.context);
    this.set('compareServiceVersion', event.context);
    this.get('controller').onConfigGroupChange();
  },
  removeCompareVersionBar: function () {
    var displayedVersion = this.get('displayedServiceVersion.version');
    var versionIndex = 0;

    this.set('compareServiceVersion', null);
    this.get('serviceVersions').forEach(function (serviceVersion, index) {
      if (serviceVersion.get('version') === displayedVersion) {
        serviceVersion.set('isDisplayed', true);
        versionIndex = index;
      } else {
        serviceVersion.set('isDisplayed', false);
      }
    });
    this.shiftFlowOnSwitch(versionIndex);
    this.get('controller').loadSelectedVersion(displayedVersion);
  },
  clearCompareVersionBar: function () {
    this.set('compareServiceVersion', null);
  }.observes('controller.selectedConfigGroup'),
  /**
   * revert config values to chosen version and apply reverted configs to server
   */
  revert: function (event) {
    var self = this;
    var serviceConfigVersion = event.context || Em.Object.create({
      version: this.get('displayedServiceVersion.version'),
      serviceName: this.get('displayedServiceVersion.serviceName'),
      notes:''
    });
    if (serviceConfigVersion.get('isReference')) {
      serviceConfigVersion = this.get(serviceConfigVersion.get('property'));
    }
    var versionText = serviceConfigVersion.get('versionText');
    return App.ModalPopup.show({
      header: Em.I18n.t('dashboard.configHistory.info-bar.makeCurrent.popup.title'),
      serviceConfigNote: Em.I18n.t('services.service.config.configHistory.makeCurrent.message').format(versionText),
      bodyClass: Em.View.extend({
        templateName: require('templates/common/configs/save_configuration'),
        notesArea: Em.TextArea.extend({
          classNames: ['full-width'],
          value: Em.I18n.t('services.service.config.configHistory.makeCurrent.message').format(versionText),
          onChangeValue: function() {
            this.get('parentView.parentView').set('serviceConfigNote', this.get('value'));
          }.observes('value')
        })
      }),
      primary: Em.I18n.t('dashboard.configHistory.info-bar.revert.button'),
      secondary: Em.I18n.t('common.discard'),
      third: Em.I18n.t('common.cancel'),
      onPrimary: function () {
        serviceConfigVersion.set('serviceConfigNote', this.get('serviceConfigNote'));
        self.sendRevertCall(serviceConfigVersion);
        this.hide();
      }
    });
  },

  /**
   * send PUT call to revert config to selected version
   * @param serviceConfigVersion
   */
  sendRevertCall: function (serviceConfigVersion) {
    App.ajax.send({
      name: 'service.serviceConfigVersion.revert',
      sender: this,
      data: {
        data: {
          "Clusters": {
            "desired_service_config_versions": {
              "service_config_version": serviceConfigVersion.get('version'),
              "service_name": serviceConfigVersion.get('serviceName'),
              "service_config_version_note": serviceConfigVersion.get('serviceConfigNote')
            }
          }
        }
      },
      success: 'sendRevertCallSuccess'
    });
  },

  sendRevertCallSuccess: function (data, opt, params) {
    // revert to an old version would generate a new version with latest version number,
    // so, need to loadStep to update
     this.get('controller').loadStep();
  },

  /**
   * save configuration
   * @return {object}
   */
  save: function () {
    var self = this;
    return App.ModalPopup.show({
      header: Em.I18n.t('dashboard.configHistory.info-bar.save.popup.title'),
      serviceConfigNote: '',
      bodyClass: Em.View.extend({
        templateName: require('templates/common/configs/save_configuration'),
        notesArea: Em.TextArea.extend({
          classNames: ['full-width'],
          placeholder: Em.I18n.t('dashboard.configHistory.info-bar.save.popup.placeholder'),
          onChangeValue: function() {
            this.get('parentView.parentView').set('serviceConfigNote', this.get('value'));
          }.observes('value')
        })
      }),
      footerClass: Ember.View.extend({
        templateName: require('templates/main/service/info/save_popup_footer')
      }),
      primary: Em.I18n.t('common.save'),
      secondary: Em.I18n.t('common.cancel'),
      onSave: function () {
        self.get('controller').set('serviceConfigVersionNote', this.get('serviceConfigNote'));
        var newVersionToBeCreated = App.ServiceConfigVersion.find().filterProperty('serviceName', self.get('serviceName')).get('length') + 1;
        self.get('controller').set('preSelectedConfigVersion', Em.Object.create({
          version: newVersionToBeCreated,
          serviceName: self.get('displayedServiceVersion.serviceName'),
          groupName: self.get('controller.selectedConfigGroup.name')
        }));
        self.get('controller').restartServicePopup();
        this.hide();
      },
      onDiscard: function () {
        this.hide();
        self.set('controller.preSelectedConfigVersion', null);
        self.get('controller').loadStep();
      },
      onCancel: function () {
        this.hide();
      }
    });
  },
  /**
   * move back to the later service version
   */
  shiftBack: function () {
    if (!this.get('showLeftArrow')) return;
    this.decrementProperty('startIndex');
    this.adjustFlowView();
  },
  /**
   * move forward to the previous service version
   */
  shiftForward: function () {
    if (!this.get('showRightArrow')) return;
    this.incrementProperty('startIndex');
    this.adjustFlowView();
  },
  /**
   * shift flow view to position where selected version is visible
   * @param versionIndex
   */
  shiftFlowOnSwitch: function (versionIndex) {
    var serviceVersions = this.get('serviceVersions');

    if ((this.get('startIndex') + this.VERSIONS_IN_FLOW) < versionIndex || versionIndex < this.get('startIndex')) {
      versionIndex = (serviceVersions.length < (versionIndex + this.VERSIONS_IN_FLOW)) ? serviceVersions.length - this.VERSIONS_IN_FLOW : versionIndex;
      this.set('startIndex', versionIndex);
      this.adjustFlowView();
    }
  }
});

});

require.register("views/common/configs/custom_category_views/notification_configs_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Custom view for Notification section for MISC tab
 * Used only in the Install Wizard
 * Show configs for creating Init Notification (@see 'data/site_properties' for list of the configs used here)
 *
 * @type {App.NotificationsConfigsView}
 */
App.NotificationsConfigsView = App.ServiceConfigsByCategoryView.extend({

  templateName: require('templates/common/configs/notifications_configs'),

  /**
   * @type {string} "yes|no"
   */
  createNotification: 'no',

  categoryConfigsAll: function () {
    return this.get('serviceConfigs').filterProperty('filename', 'alert_notification');
  }.property('serviceConfigs.@each').cacheable(),

  /**
   * @type {string} "tls|ssl"
   */
  tlsOrSsl: 'tls',

  /**
   * Determines if notification configs should be disabled
   * @type {boolean}
   */
  configsAreDisabled: true,

  /**
   * Config with flag for user auth in the notification
   * @type {App.ServiceConfigProperty}
   */
  useAuthConfig: function () {
    return this.get('categoryConfigs').findProperty('name', 'smtp_use_auth');
  }.property(),

  /**
   * Empty categoryConfigsAll means that user isn't at Installer, so manage notification view shouldn't be processed
   * @method didInsertElement
   */
  didInsertElement: function () {
    this._super();
    if (!this.get('categoryConfigsAll.length')) return;
    this.set('createNotification', this.get('categoryConfigsAll').findProperty('name', 'create_notification').get('value'));
    this.set('tlsOrSsl', this.get('categoryConfigsAll').findProperty('name', 'mail.smtp.starttls.enable').get('value') ? 'tls' : 'ssl');
    var smtp_use_auth = this.get('categoryConfigsAll').findProperty('name', 'smtp_use_auth');
    var v = (smtp_use_auth.get('value') == 'true');
    smtp_use_auth.set('value', v);
    this.updateCategoryConfigs();
  },

  /**
   * If <code>tlsOrSsl</code> was changed two configs should be updated
   * TLS and SSL can't be enabled in the same time
   * @method onTlsOrSslChanged
   */
  onTlsOrSslChanged: function () {
    var tlsOrSsl = this.get('tlsOrSsl');
    this.get('categoryConfigsAll').findProperty('name', 'mail.smtp.starttls.enable').set('value', tlsOrSsl == 'tls');
    this.get('categoryConfigsAll').findProperty('name', 'mail.smtp.startssl.enable').set('value', tlsOrSsl == 'ssl');
  }.observes('tlsOrSsl'),

  /**
   * Update username and password fields state according to <code>useAuthConfig.value</code>
   * @method onUseAuthConfigChange
   */
  onUseAuthConfigChange: function () {
    var configsToUpdate = ['ambari.dispatch.credential.username', 'ambari.dispatch.credential.password'],
      useAuthConfigValue = this.get('useAuthConfig.value'),
      useAuthConfigIsEditable = this.get('useAuthConfig.isEditable'),
      self = this;
    this.get('categoryConfigs').forEach(function (config) {
      if (configsToUpdate.contains(config.get('name'))) {
        var flag = useAuthConfigIsEditable ? useAuthConfigValue : false;
        self.updateConfig(config, flag);
      }
    });
  }.observes('useAuthConfig.value'),

  /**
   * Update all notification fields state according to <code>createNotification</code>
   * If user select to create notification, all fields become editable and required (also run validation for each)
   * If user select don't create notification, all fields become disabled, not required and valid
   * @method updateCategoryConfigs
   */
  updateCategoryConfigs: function () {
    var createNotification = this.get('createNotification'),
      self = this;
    this.get('categoryConfigs').forEach(function (config) {
      var flag = (createNotification == 'yes');
      self.updateConfig(config, flag);
    });
    this.set('configsAreDisabled', this.get('createNotification') == 'no');
    this.onUseAuthConfigChange();
    this.get('categoryConfigsAll').findProperty('name', 'create_notification').set('value', createNotification);
  }.observes('createNotification'),

  /**
   * According to <code>flag</code>, enable/disable <code>config</code>
   * If <code>flag</code> is true - validate <code>config</code> value
   * If <code>flag</code> is false - mark <code>config</code> as valid
   * @param {App.ServiceConfigProperty} config
   * @param {boolean} flag
   * @method updateConfig
   */
  updateConfig: function (config, flag) {
    config.set('isRequired', flag);
    config.set('isEditable', flag);
    if (flag) {
      config.validate();
    }
    else {
      config.set('errorMessage', '');
      config.propertyDidChange('isValid');
    }
  }

});

});

require.register("views/common/configs/overriddenProperty_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
// SCP means ServiceConfigProperty

var App = require('app');

App.ServiceConfigView.SCPOverriddenRowsView = Ember.View.extend({
  templateName: require('templates/common/configs/overriddenProperty'),
  serviceConfigProperty: null, // is passed dynamically at runtime where ever
  // we are declaring this from configs.hbs ( we are initializing this from UI )
  categoryConfigs: null, // just declared as viewClass need it

  init: function () {
    this._super();
    if (this.get('controller.name') != 'mainServiceInfoConfigsController') {
      this.addObserver('isDefaultGroupSelected', this, 'setSwitchText');
    }
  },

  didInsertElement: function () {
    this.setSwitchText();
  },

  willDestroyElement: function () {
    if (this.get('controller.name') != 'mainServiceInfoConfigsController') {
      this.removeObserver('isDefaultGroupSelected', this, 'setSwitchText');
    }
  },

  setSwitchText: function () {
    Em.$('body>.tooltip').remove();
    if (this.get('isDefaultGroupSelected')) {
      var overrides = this.get('serviceConfigProperty.overrides');
      overrides.forEach(function(overriddenSCP) {
        overriddenSCP.get('group').set('switchGroupTextShort',
          Em.I18n.t('services.service.config_groups.switchGroupTextShort').format(overriddenSCP.get('group.displayName')));
        overriddenSCP.get('group').set('switchGroupTextFull',
          Em.I18n.t('services.service.config_groups.switchGroupTextFull').format(overriddenSCP.get('group.displayName')));
      });
      this.set('serviceConfigProperty.overrides', overrides);
    }
    App.tooltip(this.$('[data-toggle=tooltip]'),{
      placement: 'top'
    });
  },

  toggleFinalFlag: function (event) {
    var override = event.contexts[0];
    if (override.get('isNotEditable')) {
      return;
    }
    override.set('isFinal', !override.get('isFinal'));
  },

  removeOverride: function (event) {
    // arg 1 SCP means ServiceConfigProperty
    var scpToBeRemoved = event.contexts[0];
    var overrides = this.get('serviceConfigProperty.overrides');
    // remove override property from selectedService on installer 7-th step
    if (this.get('controller.name') == 'wizardStep7Controller') {
      var controller = this.get('controller');
      var group = controller.get('selectedService.configGroups').findProperty('name', controller.get('selectedConfigGroup.name'));
      group.get('properties').removeObject(scpToBeRemoved);
    }
    overrides = overrides.without(scpToBeRemoved);
    this.set('serviceConfigProperty.overrides', overrides);
    Em.$('body>.tooltip').remove(); //some tooltips get frozen when their owner's DOM element is removed
  }
});

});

require.register("views/common/configs/service_config_container_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

var lazyLoading = require('utils/lazy_loading');

App.ServiceConfigContainerView = Em.ContainerView.extend({

  view: null,

  lazyLoading: null,

  pushView: function () {
    if (this.get('controller.selectedService')) {
      var self = this;
      var controllerRoute = 'App.router.' + this.get('controller.name');
      if (!this.get('view')) {
        this.get('childViews').pushObject(App.ServiceConfigView.create({
          templateName: require('templates/common/configs/service_config_wizard'),
          controllerBinding: controllerRoute,
          isNotEditableBinding: controllerRoute + '.isNotEditable',
          filterBinding: controllerRoute + '.filter',
          columnsBinding: controllerRoute + '.filterColumns',
          selectedServiceBinding: controllerRoute + '.selectedService',
          serviceConfigsByCategoryView: Em.ContainerView.create(),
          willDestroyElement: function () {
            $('.loading').append(Em.I18n.t('app.loadingPlaceholder'));
          },
          didInsertElement: function () {
            $('.loading').empty();
            this._super();
          }
        }));
      }
      else {
        this.get('childViews').pushObject(this.get('view'));
      }
      var categoriesToPush = [];
      this.get('controller.selectedService.configCategories').forEach(function (item) {

        var categoryView = item.get('isCustomView') ? item.get('customView') : App.ServiceConfigsByCategoryView;
        if (categoryView !== null) {
          categoriesToPush.pushObject(categoryView.extend({
            category: item,
            controllerBinding: controllerRoute,
            canEditBinding: 'parentView.canEdit',
            serviceBinding: controllerRoute + '.selectedService',
            serviceConfigsBinding: controllerRoute + '.selectedService.configs',
            supportsHostOverridesBinding: 'parentView.supportsHostOverrides'
          }));
        }
      });
      this.set('lazyLoading', lazyLoading.run({
        destination: self.get('childViews.lastObject.serviceConfigsByCategoryView.childViews'),
        source: categoriesToPush,
        initSize: 3,
        chunkSize: 3,
        delay: 200,
        context: this
      }));
    }
  },

  selectedServiceObserver: function () {
    if (this.get('childViews.length')) {
      var view = this.get('childViews.firstObject');
      if (view.get('serviceConfigsByCategoryView.childViews.length')) {
        view.get('serviceConfigsByCategoryView.childViews').clear();
      }
      view.removeFromParent();
      this.set('view', view);
    }
    //terminate lazy loading when switch service
    if (this.get('lazyLoading')) lazyLoading.terminate(this.get('lazyLoading'));
    this.pushView();
  }.observes('controller.selectedService')

});

});

require.register("views/common/configs/service_config_tab_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ServiceConfigTab = Em.View.extend({

  tagName: 'li',

  selectService: function (event) {
    this.set('controller.selectedService', event.context);
  },

  didInsertElement: function () {
    var serviceName = this.get('controller.selectedService.serviceName');
    this.$('a[href="#' + serviceName + '"]').tab('show');
  }
});

});

require.register("views/common/configs/service_config_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ServiceConfigView = Em.View.extend({
  templateName: require('templates/common/configs/service_config'),
  isRestartMessageCollapsed: false,
  filter: '', //from template
  columns: [], //from template
  propertyFilterPopover: [Em.I18n.t('services.service.config.propertyFilterPopover.title'), Em.I18n.t('services.service.config.propertyFilterPopover.content')],
  canEdit: true, // View is editable or read-only?
  supportsHostOverrides: function () {
    switch (this.get('controller.name')) {
      case 'wizardStep7Controller':
        return this.get('controller.selectedService.serviceName') !== 'MISC';
      case 'mainServiceInfoConfigsController':
      case 'mainHostServiceConfigsController':
        return true;
      default:
        return false;
    }
  }.property('controller.name', 'controller.selectedService'),
  showConfigHistoryFeature: false,
  toggleRestartMessageView: function () {
    this.$('.service-body').toggle('blind', 200);
    this.set('isRestartMessageCollapsed', !this.get('isRestartMessageCollapsed'));
  },
  didInsertElement: function () {
    if (this.get('isNotEditable') === true) {
      this.set('canEdit', false);
    }
    if (this.$('.service-body')) {
      this.$('.service-body').hide();
    }
    App.tooltip($(".restart-required-property"), {html: true});
    App.tooltip($(".icon-lock"), {placement: 'right'});
    this.checkCanEdit();
  },

  /**
   * Check if we should show Custom Property category
   */
  checkCanEdit: function () {
    var controller = this.get('controller');
    if (!controller.get('selectedService.configCategories')) {
      return;
    }

    if (controller.get('selectedConfigGroup')) {
      controller.get('selectedService.configCategories').filterProperty('siteFileName').forEach(function (config) {
        config.set('customCanAddProperty', config.get('canAddProperty'));
      });
    }

  }.observes(
      'App.router.mainServiceInfoConfigsController.selectedConfigGroup.name',
      'App.router.wizardStep7Controller.selectedConfigGroup.name'
    )
});

});

require.register("views/common/configs/service_configs_by_category_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

var validator = require('utils/validator');
var stringUtils = require('utils/string_utils');
require('utils/configs/modification_handlers/modification_handler');

App.ServiceConfigsByCategoryView = Em.View.extend(App.UserPref, {

  templateName: require('templates/common/configs/service_config_category'),

  classNames: ['accordion-group', 'common-config-category'],

  classNameBindings: ['category.name', 'isShowBlock::hidden'],

  content: null,

  category: null,

  service: null,

  canEdit: true, // View is editable or read-only?

  serviceConfigs: null, // General, Advanced, NameNode, SNameNode, DataNode, etc.

  /**
   * @type {Array}
   */
  categoryConfigs: function () {
    var categoryConfigs = this.get('categoryConfigsAll');
    return this.orderContentAtLast(this.sortByIndex(categoryConfigs)).filterProperty('isVisible', true);
  }.property('categoryConfigsAll.@each.isVisible').cacheable(),

  /**
   * This method provides all the properties which apply
   * to this category, irrespective of visibility. This
   * is helpful in Oozie/Hive database configuration, where
   * MySQL etc. database options don't show up, because
   * they were not visible initially.
   */
  categoryConfigsAll: function () {
    return this.get('serviceConfigs').filterProperty('category', this.get('category.name'));
  }.property('serviceConfigs.@each').cacheable(),

  /**
   * If added/removed a serverConfigObject, this property got updated.
   * Without this property, all serviceConfigs Objects will show up even if some was collapsed before.
   * @type {boolean}
   */
  isCategoryBodyVisible: function () {
    return this.get('category.isCollapsed') ? "display: none;" : "display: block;"
  }.property('serviceConfigs.length'),

  childView: App.ServiceConfigsOverridesView,

  changeFlag: Ember.Object.create({
    val: 1
  }),

  /**
   * Should we show config group or not
   * @type {boolean}
   */
  isShowBlock: function () {
    return this.get('category.customCanAddProperty') || this.get('categoryConfigs').filterProperty('isHiddenByFilter', false).length > 0;
  }.property('category.customCanAddProperty', 'categoryConfigs.@each.isHiddenByFilter'),

  /**
   * Re-order the configs to list content displayType properties at last in the category
   * @param categoryConfigs
   */
  orderContentAtLast: function (categoryConfigs) {
    var contentProperties = categoryConfigs.filterProperty('displayType', 'content');
    var self = this;
    if (!contentProperties.length) {
      return categoryConfigs
    }
    else {
      var comparator;
      return categoryConfigs.sort(function (a, b) {
        var aContent = contentProperties.someProperty('name', a.get('name'));
        var bContent = contentProperties.someProperty('name', b.get('name'));
        if (aContent && bContent) {
          return 0;
        }
        else if (aContent) {
          return 1;
        }
        else {
          return -1;
        }
      });
    }
  },

  /**
   * Warn/prompt user to adjust Service props when changing user/groups in Misc
   * Is triggered when user ended editing text field
   */
  configChangeObserver: function (manuallyChangedProperty) {
    var changedProperty;
    if (manuallyChangedProperty.get("id")) {
      changedProperty = [manuallyChangedProperty];
    }
    else {
      changedProperty = this.get("serviceConfigs").filterProperty("editDone", true);
    }

    if (changedProperty.length > 0) {
      changedProperty = changedProperty.objectAt(0);
    }
    else {
      return;
    }
    this.affectedProperties = [];
    var stepConfigs = this.get("controller.stepConfigs");
    var serviceId = this.get('controller.selectedService.serviceName');
    var serviceConfigModificationHandler = null;
    try{
      serviceConfigModificationHandler = require('utils/configs/modification_handlers/'+serviceId.toLowerCase());
    }catch (e) {
      console.log("Unable to load modification handler for ", serviceId);
    }
    if (serviceConfigModificationHandler != null) {
      var securityEnabled = App.router.get('mainAdminSecurityController.securityEnabled');
      this.affectedProperties = serviceConfigModificationHandler.getDependentConfigChanges(changedProperty, this.get("controller.selectedServiceNames"), stepConfigs, securityEnabled);
    }
    changedProperty.set("editDone", false); // Turn off flag

    if (this.affectedProperties.length > 0 && !this.get("controller.miscModalVisible")) {
      this.newAffectedProperties = this.affectedProperties;
      var self = this;
      return App.ModalPopup.show({
        classNames: ['modal-690px-width'],
        showCloseButton: false,
        header: "Warning: you must also change these Service properties",
        onApply: function () {
          self.get("newAffectedProperties").forEach(function(item) {
            self.get("controller.stepConfigs").findProperty("serviceName", item.serviceName).get("configs").find(function(config) {
              return item.propertyName == config.get('name') && (item.filename == null || item.filename == config.get('filename'));
            }).set("value", item.newValue);
          });
          self.get("controller").set("miscModalVisible", false);
          this.hide();
        },
        onIgnore: function () {
          self.get("controller").set("miscModalVisible", false);
          this.hide();
        },
        onUndo: function () {
          var affected = self.get("newAffectedProperties").objectAt(0);
          self.get("controller.stepConfigs").findProperty("serviceName", affected.sourceServiceName).get("configs")
          .findProperty("name", affected.changedPropertyName).set("value", $.trim(affected.curValue));
          self.get("controller").set("miscModalVisible", false);
          this.hide();
        },
        footerClass: Ember.View.extend({
          classNames: ['modal-footer'],
          templateName: require('templates/common/configs/propertyDependence_footer'),
          canIgnore: serviceId == 'MISC'
        }),
        bodyClass: Ember.View.extend({
          templateName: require('templates/common/configs/propertyDependence'),
          controller: this,
          propertyChange: self.get("newAffectedProperties"),
          didInsertElement: function () {
            self.get("controller").set("miscModalVisible", true);
          }
        })
      });
    }
  }.observes('categoryConfigs.@each.editDone'),

  /**
   * When the view is in read-only mode, it marks
   * the properties as read-only.
   */
  updateReadOnlyFlags: function () {
    var configs = this.get('serviceConfigs');
    var canEdit = this.get('canEdit');
    if (!canEdit && configs) {
      configs.forEach(function (c) {
        c.set('isEditable', false);
        var overrides = c.get('overrides');
        if (overrides != null) {
          overrides.setEach('isEditable', false);
        }
      });
    }
  },

  /**
   * Filtered <code>categoryConfigs</code> array. Used to show filtered result
   */
  filteredCategoryConfigs: function () {
    $('.popover').remove();
    var filter = this.get('parentView.filter').toLowerCase();
    var selectedFilters = this.get('parentView.columns').filterProperty('selected');
    var filteredResult = this.get('categoryConfigs');
    var isInitialRendering = !arguments.length || arguments[1] != 'categoryConfigs';

    if (selectedFilters.length > 0 || filter.length > 0 || this.get('state') === 'inDOM') {
      filteredResult.forEach(function (config) {
        var passesFilters = true;

        selectedFilters.forEach(function (filter) {
          if (config.get(filter.attributeName) !== filter.attributeValue) {
            passesFilters = false;
          }
        });

        if (!passesFilters) {
          config.set('isHiddenByFilter', true);
          return false;
        }

        var searchString = config.get('defaultValue') + config.get('description') +
          config.get('displayName') + config.get('name') + config.get('value');

        if (config.get('overrides')) {
          config.get('overrides').forEach(function (overriddenConf) {
            searchString += overriddenConf.get('value') + overriddenConf.get('group.name');
          });
        }

        if (filter != null && typeof searchString === "string") {
          config.set('isHiddenByFilter', !(searchString.toLowerCase().indexOf(filter) > -1));
        } else {
          config.set('isHiddenByFilter', false);
        }
      });
    }
    filteredResult = this.sortByIndex(filteredResult);
    filteredResult = filteredResult.filterProperty('isHiddenByFilter', false);

    if (filter) {
      if (filteredResult.length && typeof this.get('category.collapsedByDefault') === 'undefined') {
        // Save state
        this.set('category.collapsedByDefault', this.get('category.isCollapsed'));
      }
      this.set('category.isCollapsed', !filteredResult.length);
    } else if (typeof this.get('category.collapsedByDefault') !== 'undefined') {
      // If user clear filter -- restore defaults
      this.set('category.isCollapsed', this.get('category.collapsedByDefault'));
      this.set('category.collapsedByDefault', undefined);
    } else if (isInitialRendering && !filteredResult.length) {
      this.set('category.isCollapsed', true);
    }

    var categoryBlock = $('.' + this.get('category.name').split(' ').join('.') + '>.accordion-body');
    this.get('category.isCollapsed') ? categoryBlock.hide() : categoryBlock.show();
  }.observes('categoryConfigs', 'parentView.filter', 'parentView.columns.@each.selected'),

  /**
   * sort configs in current category by index
   * @param configs
   * @return {Array}
   * @method sortByIndex
   */
  sortByIndex: function (configs) {
    var sortedConfigs = [];
    var unSorted = [];
    if (!configs.someProperty('index')) {
      return configs;
    }
    configs.forEach(function (config) {
      var index = config.get('index');
      if ((index !== null) && isFinite(index)) {
        sortedConfigs[index] ? sortedConfigs.splice(index, 0, config) : sortedConfigs[index] = config;
      } else {
        unSorted.push(config);
      }
    });
    // remove undefined elements from array
    sortedConfigs = sortedConfigs.filter(function (config) {
      return config !== undefined;
    });
    return sortedConfigs.concat(unSorted);
  },

  /**
   * Onclick handler for Config Group Header. Used to show/hide block
   * @method onToggleBlock
   */
  onToggleBlock: function () {
    this.$('.accordion-body').toggle('blind', 500);
    this.set('category.isCollapsed', !this.get('category.isCollapsed'));
  },

  didInsertElement: function () {
    var isCollapsed = this.get('category.isCollapsed') == undefined ? (this.get('category.name').indexOf('Advanced') != -1 || this.get('category.name').indexOf('CapacityScheduler') != -1 || this.get('category.name').indexOf('Custom') != -1) : this.get('category.isCollapsed');
    var self = this;
    this.set('category.isCollapsed', isCollapsed);
    if (isCollapsed) {
      this.$('.accordion-body').hide();
    }
    else {
      this.$('.accordion-body').show();
    }
    $('#serviceConfig').tooltip({
      selector: '[data-toggle=tooltip]',
      placement: 'top'
    });
    this.updateReadOnlyFlags();
    this.filteredCategoryConfigs();
    Em.run.next(function () {
      self.updateReadOnlyFlags();
    });
  },

  willDestroyElement: function () {
    if (this.get('parentView.controller.name') == 'mainServiceInfoConfigsController') {
      this.get('categoryConfigsAll').forEach(function (item) {
        item.set('isVisible', false);
      });
    }
  },

  isOneOfAdvancedSections: function () {
    var category = this.get('category');
    return category.indexOf("Advanced") != -1;
  },

  /**
   * @returns {string}
   */
  persistKey: function () {
    return 'admin-bulk-add-properties-' + App.router.get('loginName');
  },

  /**
   * Show popup for adding new config-properties
   * @method showAddPropertyWindow
   */
  showAddPropertyWindow: function () {
    var persistController = this;
    var modePersistKey = this.persistKey();
    var selectedConfigGroup = this.get('controller.selectedConfigGroup');

    persistController.getUserPref(modePersistKey).pipe(function (data) {
      return !!data;
    },function () {
      return false;
    }).always((function (isBulkMode) {

        var category = this.get('category');
        var siteFileName = category.get('siteFileName');

        var service = this.get('service');
        var serviceName = service.get('serviceName');

        var secureConfigs = this.get('controller.secureConfigs').filterProperty('filename', siteFileName);

        function isSecureConfig(configName) {
          return !!secureConfigs.findProperty('name', configName);
        }

        var configsOfFile = service.get('configs').filterProperty('filename', siteFileName);
        var siteFileProperties = App.config.get('configMapping').all().filterProperty('filename', siteFileName);

        function shouldSupportFinal(filename) {
          var stackService = App.StackService.find().findProperty('serviceName', serviceName);
          var supportsFinal = App.config.getConfigTypesInfoFromService(stackService).supportsFinal;
          var matchingConfigType = supportsFinal.find(function (configType) {
            return filename.startsWith(configType);
          });
          return !!matchingConfigType;
        }

        var supportsFinal = (serviceName == 'MISC') ? false: (siteFileName ? shouldSupportFinal(siteFileName) : false);

        function isDuplicatedConfigKey(name) {
          return siteFileProperties.findProperty('name', name) || configsOfFile.findProperty('name', name);
        }

        var serviceConfigs = this.get('serviceConfigs');

        function createProperty(propertyName, propertyValue) {
          serviceConfigs.pushObject(App.ServiceConfigProperty.create({
            name: propertyName,
            displayName: propertyName,
            value: propertyValue,
            displayType: stringUtils.isSingleLine(propertyValue) ? 'advanced' : 'multiLine',
            isSecureConfig: isSecureConfig(propertyName),
            category: category.get('name'),
            id: 'site property',
            serviceName: serviceName,
            defaultValue: null,
            supportsFinal: supportsFinal,
            filename: siteFileName || '',
            isUserProperty: true,
            isNotSaved: true,
            isRequired: false,
            group: selectedConfigGroup.get('isDefault') ? null : selectedConfigGroup,
            isOverridable: selectedConfigGroup.get('isDefault')
          }));
        }

        var serviceConfigObj = Ember.Object.create({
          isBulkMode: isBulkMode,
          bulkConfigValue: '',
          bulkConfigError: false,
          bulkConfigErrorMessage: '',

          name: '',
          value: '',
          isKeyError: false,
          showFilterLink: false,
          errorMessage: '',
          observeAddPropertyValue: function () {
            var name = this.get('name');
            if (name.trim() != '') {
              if (validator.isValidConfigKey(name)) {
                if (!isDuplicatedConfigKey(name)) {
                  this.set('showFilterLink', false);
                  this.set('isKeyError', false);
                  this.set('errorMessage', '');
                } else {
                  this.set('showFilterLink', true);
                  this.set('isKeyError', true);
                  this.set('errorMessage', Em.I18n.t('services.service.config.addPropertyWindow.error.derivedKey'));
                }
              } else {
                this.set('showFilterLink', false);
                this.set('isKeyError', true);
                this.set('errorMessage', Em.I18n.t('form.validator.configKey'));
              }
            } else {
              this.set('showFilterLink', false);
              this.set('isKeyError', true);
              this.set('errorMessage', Em.I18n.t('services.service.config.addPropertyWindow.error.required'));
            }
          }.observes('name')
        });

        function processConfig(config, callback) {
          var lines = config.split('\n');
          var errorMessages = [];
          var parsedConfig = {};
          var propertyCount = 0;

          function lineNumber(index) {
            return Em.I18n.t('services.service.config.addPropertyWindow.error.lineNumber').format(index + 1);
          }

          lines.forEach(function (line, index) {
            if (line.trim() === '') {
              return;
            }
            var delimiter = '=';
            var delimiterPosition = line.indexOf(delimiter);
            if (delimiterPosition === -1) {
              errorMessages.push(lineNumber(index) + Em.I18n.t('services.service.config.addPropertyWindow.error.format'));
              return;
            }
            var key = Em.Handlebars.Utils.escapeExpression(line.slice(0, delimiterPosition).trim());
            var value = line.slice(delimiterPosition + 1);
            if (validator.isValidConfigKey(key)) {
              if (!isDuplicatedConfigKey(key) && !(key in parsedConfig)) {
                parsedConfig[key] = value;
                propertyCount++;
              } else {
                errorMessages.push(lineNumber(index) + Em.I18n.t('services.service.config.addPropertyWindow.error.derivedKey.specific').format(key));
              }
            } else {
              errorMessages.push(lineNumber(index) + Em.I18n.t('form.validator.configKey.specific').format(key));
            }
          });

          if (errorMessages.length > 0) {
            callback(errorMessages.join('<br>'), parsedConfig);
          }
          else if (propertyCount === 0) {
            callback(Em.I18n.t('services.service.config.addPropertyWindow.propertiesPlaceholder', parsedConfig));
          }
          else {
            callback(null, parsedConfig);
          }
        }

        App.ModalPopup.show({
          classNames: ['sixty-percent-width-modal'],
          header: 'Add Property',
          primary: 'Add',
          secondary: 'Cancel',
          onPrimary: function () {
            if (serviceConfigObj.isBulkMode) {
              var popup = this;
              processConfig(serviceConfigObj.bulkConfigValue, function (error, parsedConfig) {
                if (error) {
                  serviceConfigObj.set('bulkConfigError', true);
                  serviceConfigObj.set('bulkConfigErrorMessage', error);
                } else {
                  for (var key in parsedConfig) {
                    if (parsedConfig.hasOwnProperty(key)) {
                      createProperty(key, parsedConfig[key]);
                    }
                  }
                  popup.hide();
                }
              });
            }
            else {
              serviceConfigObj.observeAddPropertyValue();
              /**
               * For the first entrance use this if (serviceConfigObj.name.trim() != '')
               */
              if (!serviceConfigObj.isKeyError) {
                createProperty(serviceConfigObj.get('name'), serviceConfigObj.get('value'));
                this.hide();
              }
            }
          },
          bodyClass: Em.View.extend({
            fileName: siteFileName,
            notMisc: function () {
              return serviceName !== 'MISC';
            }.property(),
            templateName: require('templates/common/configs/addPropertyWindow'),
            controllerBinding: 'App.router.mainServiceInfoConfigsController',
            serviceConfigObj: serviceConfigObj,
            didInsertElement: function () {
              App.tooltip(this.$("[data-toggle=tooltip]"), {
                placement: "top"
              });
            },
            toggleBulkMode: function () {
              this.toggleProperty('serviceConfigObj.isBulkMode');
              persistController.postUserPref(modePersistKey, this.get('serviceConfigObj.isBulkMode'));
            },
            filterByKey: function (event) {
              var controller = (App.router.get('currentState.name') == 'configs')
                ? App.router.get('mainServiceInfoConfigsController')
                : App.router.get('wizardStep7Controller');
              this.get('parentView').onClose();
              controller.set('filter', event.view.get('serviceConfigObj.name'));
            }
          })
        });

      }).bind(this));
  },

  /**
   * Toggle <code>isFinal</code> for selected config-property if <code>isNotEditable</code> is false
   * @param {object} event
   * @method toggleFinalFlag
   */
  toggleFinalFlag: function (event) {
    var serviceConfigProperty = event.contexts[0];
    if (serviceConfigProperty.get('isNotEditable')) {
      return;
    }
    serviceConfigProperty.toggleProperty('isFinal');
  },

  /**
   * Removes the top-level property from list of properties.
   * Should be only called on user properties.
   * @method removeProperty
   */
  removeProperty: function (event) {
    var serviceConfigProperty = event.contexts[0];
    this.get('serviceConfigs').removeObject(serviceConfigProperty);
    // push config's file name if this config was stored on server
    if (!serviceConfigProperty.get('isNotSaved')) {
      this.get('controller').get('modifiedFileNames').push(serviceConfigProperty.get('filename'));
    }
    Em.$('body>.tooltip').remove(); //some tooltips get frozen when their owner's DOM element is removed
  },

  /**
   * Restores given property's value to be its default value.
   * Does not update if there is no default value.
   * @method doRestoreDefaultValue
   */
  doRestoreDefaultValue: function (event) {
    var serviceConfigProperty = event.contexts[0];
    var value = serviceConfigProperty.get('value');
    var dValue = serviceConfigProperty.get('defaultValue');
    var supportsFinal = serviceConfigProperty.get('supportsFinal');
    var defaultIsFinal = serviceConfigProperty.get('defaultIsFinal');

    if (dValue != null) {
      if (serviceConfigProperty.get('displayType') === 'password') {
        serviceConfigProperty.set('retypedPassword', dValue);
      }
      serviceConfigProperty.set('value', dValue);
    }
    if (supportsFinal) {
      serviceConfigProperty.set('isFinal', defaultIsFinal);
    }
    this.configChangeObserver(serviceConfigProperty);
    Em.$('body>.tooltip').remove(); //some tooltips get frozen when their owner's DOM element is removed
  },

  /**
   *
   * @method createOverrideProperty
   */
  createOverrideProperty: function (event) {
    var serviceConfigProperty = event.contexts[0];
    var serviceConfigController = this.get('controller');
    var selectedConfigGroup = serviceConfigController.get('selectedConfigGroup');
    var isInstaller = (this.get('controller.name') === 'wizardStep7Controller');
    var configGroups = (isInstaller) ? serviceConfigController.get('selectedService.configGroups') : serviceConfigController.get('configGroups');

    //user property is added, and it has not been saved, not allow override
    if (serviceConfigProperty.get('isUserProperty') && serviceConfigProperty.get('isNotSaved') && !isInstaller) {
      App.ModalPopup.show({
        header: Em.I18n.t('services.service.config.configOverride.head'),
        body: Em.I18n.t('services.service.config.configOverride.body'),
        secondary: false
      });
      return;
    }
    if (selectedConfigGroup.get('isDefault')) {
      // Launch dialog to pick/create Config-group
      App.config.launchConfigGroupSelectionCreationDialog(this.get('service.serviceName'),
        configGroups, serviceConfigProperty, function (selectedGroupInPopup) {
          console.log("launchConfigGroupSelectionCreationDialog(): Selected/Created:", selectedGroupInPopup);
          if (selectedGroupInPopup) {
            serviceConfigController.set('overrideToAdd', serviceConfigProperty);
            serviceConfigController.set('selectedConfigGroup', selectedGroupInPopup);
          }
        }, isInstaller);
    }
    else {
      serviceConfigController.addOverrideProperty(serviceConfigProperty, selectedConfigGroup);
    }
  }

});

});

require.register("views/common/configs/services_config", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ServicesConfigView = Em.View.extend({

  templateName: require('templates/common/configs/services_config'),

  didInsertElement: function () {
    this.get('controller').loadStep();
  }

});

});

require.register("views/common/controls_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Abstract view for config fields.
 * Add popover support to control
 */
App.ServiceConfigPopoverSupport = Ember.Mixin.create({

  /**
   * Config object. It will instance of App.ServiceConfigProperty
   */
  serviceConfig: null,
  attributeBindings:['readOnly'],
  isPopoverEnabled: true,

  didInsertElement: function () {
    $('body').tooltip({
      selector: '[data-toggle=tooltip]',
      placement: 'top'
    });
    // if description for this serviceConfig not exist, then no need to show popover
    if (this.get('isPopoverEnabled') !== 'false' && this.get('serviceConfig.description')) {
      App.popover(this.$(), {
        title: Em.I18n.t('installer.controls.serviceConfigPopover.title').format(
          this.get('serviceConfig.displayName'),
          (this.get('serviceConfig.displayName') == this.get('serviceConfig.name')) ? '' : this.get('serviceConfig.name')
        ),
        content: this.get('serviceConfig.description'),
        placement: 'right',
        trigger: 'hover'
      });
    }
  },

  willDestroyElement: function() {
    this.$().popover('destroy');
  },

  readOnly: function () {
    return !this.get('serviceConfig.isEditable');
  }.property('serviceConfig.isEditable')
});

/**
 * mixin set class that serve as unique element identificator,
 * id not used in order to avoid collision with ember ids
 */
App.ServiceConfigCalculateId = Ember.Mixin.create({
  idClass: Ember.computed(function () {
    var label = Em.get(this, 'serviceConfig.name') ? Em.get(this, 'serviceConfig.name').toLowerCase().replace(/\./g, '-') : '',
        fileName = Em.get(this, 'serviceConfig.filename') ? Em.get(this, 'serviceConfig.filename').toLowerCase().replace(/\./g, '-') : '',
        group = Em.get(this, 'serviceConfig.group.name') || 'default';
        isOrigin = Em.get(this, 'serviceConfig.compareConfigs.length') > 0 ? '-origin' : '';
    return 'service-config-' + label + '-' + fileName + '-' + group + isOrigin;
  }),
  classNameBindings: 'idClass'
});

/**
 * Default input control
 * @type {*}
 */
App.ServiceConfigTextField = Ember.TextField.extend(App.ServiceConfigPopoverSupport, App.ServiceConfigCalculateId, {

  valueBinding: 'serviceConfig.value',
  classNameBindings: 'textFieldClassName',
  placeholderBinding: 'serviceConfig.defaultValue',

  keyPress: function (event) {
    if (event.keyCode == 13) {
      return false;
    }
  },
  //Set editDone true for last edited config text field parameter
  focusOut: function (event) {
    this.get('serviceConfig').set("editDone", true);
  },
  //Set editDone false for all current category config text field parameter
  focusIn: function (event) {
    if (!this.get('serviceConfig.isOverridden') && !this.get('serviceConfig.isComparison')) {
      this.get("parentView.categoryConfigsAll").setEach("editDone", false);
    }
  },

  textFieldClassName: function () {
    if (this.get('serviceConfig.unit')) {
      return ['input-small'];
    } else if (this.get('serviceConfig.displayType') === 'principal') {
      return ['span12'];
    } else {
      return ['span9'];
    }
  }.property('serviceConfig.displayType', 'serviceConfig.unit')

});

/**
 * Customized input control with Units type specified
 * @type {Em.View}
 */
App.ServiceConfigTextFieldWithUnit = Ember.View.extend(App.ServiceConfigPopoverSupport, {
  valueBinding: 'serviceConfig.value',
  classNames: ['input-append', 'with-unit'],
  placeholderBinding: 'serviceConfig.defaultValue',

  templateName: require('templates/wizard/controls_service_config_textfield_with_unit')
});

/**
 * Password control
 * @type {*}
 */
App.ServiceConfigPasswordField = Ember.TextField.extend({

  serviceConfig: null,
  type: 'password',
  attributeBindings:['readOnly'],
  valueBinding: 'serviceConfig.value',
  classNames: [ 'span4' ],
  placeholder: Em.I18n.t('form.item.placeholders.typePassword'),

  template: Ember.Handlebars.compile('{{view view.retypePasswordView}}'),

  keyPress: function (event) {
    if (event.keyCode == 13) {
      return false;
    }
  },

  retypePasswordView: Ember.TextField.extend({
    placeholder: Em.I18n.t('form.passwordRetype'),
    attributeBindings:['readOnly'],
    type: 'password',
    classNames: [ 'span4', 'retyped-password' ],
    keyPress: function (event) {
      if (event.keyCode == 13) {
        return false;
      }
    },
    valueBinding: 'parentView.serviceConfig.retypedPassword',
    readOnly: function () {
      return !this.get('parentView.serviceConfig.isEditable');
    }.property('parentView.serviceConfig.isEditable')
  }),

  readOnly: function () {
    return !this.get('serviceConfig.isEditable');
  }.property('serviceConfig.isEditable')

});

/**
 * Textarea control
 * @type {*}
 */
App.ServiceConfigTextArea = Ember.TextArea.extend(App.ServiceConfigPopoverSupport, App.ServiceConfigCalculateId, {

  valueBinding: 'serviceConfig.value',
  rows: 4,
  classNames: ['span9', 'directories']
});

/**
 * Textarea control for content type
 * @type {*}
 */
App.ServiceConfigTextAreaContent = Ember.TextArea.extend(App.ServiceConfigPopoverSupport, App.ServiceConfigCalculateId, {

  valueBinding: 'serviceConfig.value',
  rows: 20,
  classNames: ['span10']
});

/**
 * Textarea control with bigger height
 * @type {*}
 */
App.ServiceConfigBigTextArea = App.ServiceConfigTextArea.extend(App.ServiceConfigCalculateId, {
  rows: 10
});

/**
 * Checkbox control
 * @type {*}
 */
App.ServiceConfigCheckbox = Ember.Checkbox.extend(App.ServiceConfigPopoverSupport, App.ServiceConfigCalculateId, {

  allowedPairs: {
    'trueFalse': ["true", "false"],
    'YesNo': ["Yes", "No"],
    'YESNO': ["YES", "NO"],
    'yesNo': ["yes", "no"]
  },

  trueValue: true,
  falseValue: false,

  checked: false,

  /**
   * set appropriate config values pair
   * to define which value is positive (checked) property
   * and what value is negative (unchecked) proeprty
   */
  didInsertElement: function() {
    this._super();
    this.addObserver('serviceConfig.value', this, 'toggleChecker');
    Object.keys(this.get('allowedPairs')).forEach(function(key) {
      if (this.get('allowedPairs')[key].contains(this.get('serviceConfig.value'))) {
        this.set('trueValue', this.get('allowedPairs')[key][0]);
        this.set('falseValue', this.get('allowedPairs')[key][1]);
      }
    }, this);
    this.set('checked', this.get('serviceConfig.value') === this.get('trueValue'))
  },

  willDestroyElement: function() {
    this.removeObserver('serviceConfig.value', this, 'checkedBinding');
  },

  /***
   * defines if checkbox value appropriate to the config value
   * @returns {boolean}
   */
  isNotAppropriateValue: function() {
    return this.get('serviceConfig.value') !== this.get(this.get('checked') + 'Value');
  },

  /**
   * change service config value if click on checkbox
   */
  toggleValue: function() {
    if (this.isNotAppropriateValue()){
      this.set('serviceConfig.value', this.get(this.get('checked') + 'Value'));
      this.get('serviceConfig').set("editDone", true);
    }
  }.observes('checked'),

  /**
   * change checkbox value if click on undo
   */
  toggleChecker: function() {
    if (this.isNotAppropriateValue())
      this.set('checked', !this.get('checked'));
  },

  disabled: function () {
    return !this.get('serviceConfig.isEditable');
  }.property('serviceConfig.isEditable'),

  //Set editDone false for all current category config text field parameter
  focusIn: function (event) {
    if (!this.get('serviceConfig.isOverridden') && !this.get('serviceConfig.isComparison')) {
      this.get("parentView.categoryConfigsAll").setEach("editDone", false);
    }
  }
});

/**
 * Checkbox control which can hide or show dependent  properties
 * @type {*|void}
 */
App.ServiceConfigCheckboxWithDependencies = App.ServiceConfigCheckbox.extend({

  toggleDependentConfigs: function() {
    if (this.get('serviceConfig.dependentConfigPattern')) {
      if (this.get('serviceConfig.dependentConfigPattern') === "CATEGORY") {
        this.disableEnableCategoryConfigs();
      } else {
        this.showHideDependentConfigs();
      }
    }
  }.observes('checked'),

  disableEnableCategoryConfigs: function () {
    this.get('categoryConfigsAll').setEach('isEditable', this.get('checked'));
    this.set('serviceConfig.isEditable', true);
  },

  showHideDependentConfigs: function () {
    this.get('categoryConfigsAll').forEach(function (c) {
      if (c.get('name').match(this.get('serviceConfig.dependentConfigPattern')) && c.get('name') != this.get('serviceConfig.name'))
        c.set('isVisible', this.get('checked'))
    }, this);
  }
});

App.ServiceConfigRadioButtons = Ember.View.extend(App.ServiceConfigCalculateId, {
  templateName: require('templates/wizard/controls_service_config_radio_buttons'),

  didInsertElement: function () {
    // on page render, automatically populate JDBC URLs only for default database settings
    // so as to not lose the user's customizations on these fields
    if (['addServiceController', 'installerController'].contains(this.get('controller.wizardController.name'))) {
      if (/^New\s\w+\sDatabase$/.test(this.get('serviceConfig.value')) || this.dontUseHandleDbConnection.contains(this.get('serviceConfig.name'))) {
        this.onOptionsChange();
      } else {
        this.handleDBConnectionProperty();
      }
    }
  },

  /**
   * properties with these names don'use handleDBConnectionProperty mathod
   */
  dontUseHandleDbConnection: ['DB_FLAVOR', 'authentication_method'],

  configs: function () {
    if (this.get('controller.name') == 'mainServiceInfoConfigsController') return this.get('categoryConfigsAll');
    return this.get('categoryConfigsAll').filterProperty('isObserved', true);
  }.property('categoryConfigsAll'),

  serviceConfig: null,
  categoryConfigsAll: null,

  onOptionsChange: function () {
    // The following if condition will be satisfied only for installer wizard flow
    if (this.get('configs').length) {
      var connectionUrl = this.get('connectionUrl');
      if (connectionUrl) {
        var dbClass = this.get('dbClass');
        var hostName = this.get('hostName');
        var databaseName = this.get('databaseName');
        var hostNameDefault;
        var databaseNameDefault;
        var connectionUrlValue = connectionUrl.get('value');
        var connectionUrlDefaultValue = connectionUrl.get('defaultValue');
        var dbClassValue = dbClass.get('value');
        var serviceName = this.get('serviceConfig.serviceName');
        var isServiceInstalled = App.Service.find().someProperty('serviceName', serviceName);
        var postgresUrl = 'jdbc:postgresql://{0}:5432/{1}';
        var oracleUrl = 'jdbc:oracle:thin:@//{0}:1521/{1}';
        var mssqlUrl = 'jdbc:sqlserver://{0};databaseName={1}';
        var mssqlIntegratedAuthUrl = 'jdbc:sqlserver://{0};databaseName={1};integratedSecurity=true';
        var isNotExistingMySQLServer = this.get('serviceConfig.value') !== 'Existing MSSQL Server database with integrated authentication';
        var categoryConfigsAll = this.get('categoryConfigsAll');
        if (isServiceInstalled) {
          hostNameDefault = this.get('hostNameProperty.defaultValue');
          databaseNameDefault = this.get('databaseNameProperty.defaultValue');
        } else {
          hostNameDefault = hostName;
          databaseNameDefault = databaseName;
        }
        switch (serviceName) {
          case 'HIVE':
            var hiveDbType = this.get('parentView.serviceConfigs').findProperty('name', 'hive_database_type');
            var mysqlUrl = 'jdbc:mysql://{0}/{1}?createDatabaseIfNotExist=true';
            switch (this.get('serviceConfig.value')) {
              case 'New MySQL Database':
              case 'Existing MySQL Database':
                connectionUrlValue = mysqlUrl.format(hostName, databaseName);
                connectionUrlDefaultValue = mysqlUrl.format(hostNameDefault, databaseNameDefault);
                dbClassValue = 'com.mysql.jdbc.Driver';
                Em.set(hiveDbType, 'value', 'mysql');
                break;
              case Em.I18n.t('services.service.config.hive.oozie.postgresql'):
                connectionUrlValue = postgresUrl.format(hostName, databaseName);
                connectionUrlDefaultValue = postgresUrl.format(hostNameDefault, databaseNameDefault);
                dbClassValue = 'org.postgresql.Driver';
                Em.set(hiveDbType, 'value', 'postgres');
                break;
              case 'Existing Oracle Database':
                connectionUrlValue = oracleUrl.format(hostName, databaseName);
                connectionUrlDefaultValue = oracleUrl.format(hostNameDefault, databaseNameDefault);
                dbClassValue = 'oracle.jdbc.driver.OracleDriver';
                Em.set(hiveDbType, 'value', 'oracle');
                break;
              case 'Existing MSSQL Server database with SQL authentication':
                connectionUrlValue = mssqlUrl.format(hostName, databaseName);
                connectionUrlDefaultValue = mssqlUrl.format(hostNameDefault, databaseNameDefault);
                dbClassValue = 'com.microsoft.sqlserver.jdbc.SQLServerDriver';
                Em.set(hiveDbType, 'value', 'mssql');
                break;
              case 'Existing MSSQL Server database with integrated authentication':
                connectionUrlValue = mssqlIntegratedAuthUrl.format(hostName, databaseName);
                connectionUrlDefaultValue = mssqlIntegratedAuthUrl.format(hostNameDefault, databaseNameDefault);
                dbClassValue = 'com.microsoft.sqlserver.jdbc.SQLServerDriver';
                Em.set(hiveDbType, 'value', 'mssql');
                break;
            }
            categoryConfigsAll.findProperty('name', 'javax.jdo.option.ConnectionUserName').setProperties({
              isVisible: isNotExistingMySQLServer,
              isRequired: isNotExistingMySQLServer
            });
            categoryConfigsAll.findProperty('name', 'javax.jdo.option.ConnectionPassword').setProperties({
              isVisible: isNotExistingMySQLServer,
              isRequired: isNotExistingMySQLServer
            });
            break;
          case 'OOZIE':
            var derbyUrl = 'jdbc:derby:${oozie.data.dir}/${oozie.db.schema.name}-db;create=true';
            var mysqlUrl = 'jdbc:mysql://{0}/{1}';
            switch (this.get('serviceConfig.value')) {
              case 'New Derby Database':
                connectionUrlValue = derbyUrl;
                connectionUrlDefaultValue = derbyUrl;
                dbClassValue = 'org.apache.derby.jdbc.EmbeddedDriver';
                break;
              case 'Existing MySQL Database':
                connectionUrlValue = mysqlUrl.format(hostName, databaseName);
                connectionUrlDefaultValue = mysqlUrl.format(hostNameDefault, databaseNameDefault);
                dbClassValue = 'com.mysql.jdbc.Driver';
                break;
              case Em.I18n.t('services.service.config.hive.oozie.postgresql'):
                connectionUrlValue = postgresUrl.format(hostName, databaseName);
                connectionUrlDefaultValue = postgresUrl.format(hostNameDefault, databaseNameDefault);
                dbClassValue = 'org.postgresql.Driver';
                break;
              case 'Existing Oracle Database':
                connectionUrlValue = oracleUrl.format(hostName, databaseName);
                connectionUrlDefaultValue = oracleUrl.format(hostNameDefault, databaseNameDefault);
                dbClassValue = 'oracle.jdbc.driver.OracleDriver';
                break;
              case 'Existing MSSQL Server database with SQL authentication':
                connectionUrlValue = mssqlUrl.format(hostName, databaseName);
                connectionUrlDefaultValue = mssqlUrl.format(hostNameDefault, databaseNameDefault);
                dbClassValue = 'com.microsoft.sqlserver.jdbc.SQLServerDriver';
                break;
              case 'Existing MSSQL Server database with integrated authentication':
                connectionUrlValue = mssqlIntegratedAuthUrl.format(hostName, databaseName);
                connectionUrlDefaultValue = mssqlIntegratedAuthUrl.format(hostNameDefault, databaseNameDefault);
                dbClassValue = 'com.microsoft.sqlserver.jdbc.SQLServerDriver';
                break;
            }
            categoryConfigsAll.findProperty('name', 'oozie.service.JPAService.jdbc.username').setProperties({
              isVisible: isNotExistingMySQLServer,
              isRequired: isNotExistingMySQLServer
            });
            categoryConfigsAll.findProperty('name', 'oozie.service.JPAService.jdbc.password').setProperties({
              isVisible: isNotExistingMySQLServer,
              isRequired: isNotExistingMySQLServer
            });
            break;
        }
        connectionUrl.set('value', connectionUrlValue);
        connectionUrl.set('defaultValue', connectionUrlDefaultValue);
        dbClass.set('value', dbClassValue);
      }
    }
  }.observes('databaseName', 'hostName'),

  nameBinding: 'serviceConfig.radioName',

  databaseNameProperty: function () {
    switch (this.get('serviceConfig.serviceName')) {
      case 'HIVE':
        return this.get('categoryConfigsAll').findProperty('name', 'ambari.hive.db.schema.name');
      case 'OOZIE':
        return this.get('categoryConfigsAll').findProperty('name', 'oozie.db.schema.name');
      default:
        return null;
    }
  }.property('serviceConfig.serviceName'),

  databaseName: function () {
    return this.get('databaseNameProperty.value');
  }.property('databaseNameProperty.value'),

  hostNameProperty: function () {
    var value = this.get('serviceConfig.value');
    var returnValue;
    var hostname;
    if (this.get('serviceConfig.serviceName') === 'HIVE') {
      switch (value) {
        case 'New MySQL Database':
          hostname = this.get('categoryConfigsAll').findProperty('name', 'hive_ambari_host');
          break;
        case 'Existing MySQL Database':
          hostname = this.get('categoryConfigsAll').findProperty('name', 'hive_existing_mysql_host');
          break;
        case Em.I18n.t('services.service.config.hive.oozie.postgresql'):
          hostname = this.get('categoryConfigsAll').findProperty('name', 'hive_existing_postgresql_host');
          break;
        case 'Existing Oracle Database':
          hostname = this.get('categoryConfigsAll').findProperty('name', 'hive_existing_oracle_host');
          break;
        case 'Existing MSSQL Server database with SQL authentication':
          hostname = this.get('categoryConfigsAll').findProperty('name', 'hive_existing_mssql_server_host');
          break;
        case 'Existing MSSQL Server database with integrated authentication':
          hostname = this.get('categoryConfigsAll').findProperty('name', 'hive_existing_mssql_server_2_host');
          break;
      }
      if (hostname) {
        Em.set(hostname, 'isUserProperty', false);
        returnValue = hostname;
      } else {
        returnValue = this.get('categoryConfigsAll').findProperty('name', 'hive_hostname');
      }
    } else if (this.get('serviceConfig.serviceName') === 'OOZIE') {
      switch (value) {
        case 'New Derby Database':
          hostname = this.get('categoryConfigsAll').findProperty('name', 'oozie_ambari_host');
          break;
        case 'Existing MySQL Database':
          hostname = this.get('categoryConfigsAll').findProperty('name', 'oozie_existing_mysql_host');
          break;
        case Em.I18n.t('services.service.config.hive.oozie.postgresql'):
          hostname = this.get('categoryConfigsAll').findProperty('name', 'oozie_existing_postgresql_host');
          break;
        case 'Existing Oracle Database':
          hostname = this.get('categoryConfigsAll').findProperty('name', 'oozie_existing_oracle_host');
          break;
        case 'Existing MSSQL Server database with SQL authentication':
          hostname = this.get('categoryConfigsAll').findProperty('name', 'oozie_existing_mssql_server_host');
          break;
        case 'Existing MSSQL Server database with integrated authentication':
          hostname = this.get('categoryConfigsAll').findProperty('name', 'oozie_existing_mssql_server_2_host');
          break;
      }
      if (hostname) {
        Em.set(hostname, 'isUserProperty', false);
        returnValue = hostname;
      } else {
        returnValue = this.get('categoryConfigsAll').findProperty('name', 'oozie_hostname');
      }
    }
    return returnValue;
  }.property('serviceConfig.serviceName', 'serviceConfig.value'),

  hostName: function () {
    return this.get('hostNameProperty.value');
  }.property('hostNameProperty.value'),

  connectionUrl: function () {
    if (this.get('serviceConfig.serviceName') === 'HIVE') {
      return this.get('categoryConfigsAll').findProperty('name', 'javax.jdo.option.ConnectionURL');
    } else {
      return this.get('categoryConfigsAll').findProperty('name', 'oozie.service.JPAService.jdbc.url');
    }
  }.property('serviceConfig.serviceName'),

  dbClass: function () {
    if (this.get('serviceConfig.serviceName') === 'HIVE') {
      return this.get('categoryConfigsAll').findProperty('name', 'javax.jdo.option.ConnectionDriverName');
    } else {
      return this.get('categoryConfigsAll').findProperty('name', 'oozie.service.JPAService.jdbc.driver');
    }
  }.property('serviceConfig.serviceName'),

  /**
   * `Observer` that add <code>additionalView</code> to <code>App.ServiceConfigProperty</code>
   * that responsible for (if existing db selected)
   * 1. checking database connection
   * 2. showing jdbc driver setup warning msg.
   *
   * @method handleDBConnectionProperty
   **/
  handleDBConnectionProperty: function() {
    if (this.dontUseHandleDbConnection.contains(this.get('serviceConfig.name')))
      return;
    var handledProperties = ['oozie_database', 'hive_database'];
    var currentValue = this.get('serviceConfig.value');
    var databases = /MySQL|PostgreSQL|Oracle|Derby|MSSQL/gi;
    var currentDB = currentValue.match(databases)[0];
    var databasesTypes = /MySQL|PostgreS|Oracle|Derby|MSSQL/gi;
    var currentDBType = currentValue.match(databasesTypes)[0];
    var existingDatabase = /existing/gi.test(currentValue);
    // db connection check button show up if existed db selected
    var propertyAppendTo1 = this.get('categoryConfigsAll').findProperty('displayName', 'Database URL');
    if (currentDB && existingDatabase) {
      if (handledProperties.contains(this.get('serviceConfig.name'))) {
        if (propertyAppendTo1) propertyAppendTo1.set('additionalView', App.CheckDBConnectionView.extend({databaseName: currentDB}));
      }
    } else {
      propertyAppendTo1.set('additionalView', null);
    }
    // warning msg under database type radio buttons, to warn the user to setup jdbc driver if existed db selected
    var propertyHive = this.get('categoryConfigsAll').findProperty('displayName', 'Hive Database');
    var propertyOozie = this.get('categoryConfigsAll').findProperty('displayName', 'Oozie Database');
    var propertyAppendTo2 = propertyHive ? propertyHive : propertyOozie;
    if (currentDB && existingDatabase) {
      if (handledProperties.contains(this.get('serviceConfig.name'))) {
        if (propertyAppendTo2) {
          propertyAppendTo2.set('additionalView', Ember.View.extend({
            template: Ember.Handlebars.compile('<div class="alert">{{{view.message}}}</div>'),
            message: Em.I18n.t('services.service.config.database.msg.jdbcSetup').format(currentDBType.toLowerCase(), currentDBType.toLowerCase())
          }));
        }
      }
    } else {
      propertyAppendTo2.set('additionalView', null);
    }
  }.observes('serviceConfig.value'),

  optionsBinding: 'serviceConfig.options'
});

App.ServiceConfigRadioButton = Ember.Checkbox.extend({
  tagName: 'input',
  attributeBindings: ['type', 'name', 'value', 'checked', 'disabled'],
  checked: false,
  type: 'radio',
  name: null,
  value: null,

  didInsertElement: function () {
    console.debug('App.ServiceConfigRadioButton.didInsertElement');
    if (this.get('parentView.serviceConfig.value') === this.get('value')) {
      console.debug(this.get('name') + ":" + this.get('value') + ' is checked');
      this.set('checked', true);
    }
  },

  click: function () {
    this.set('checked', true);
    console.debug('App.ServiceConfigRadioButton.click');
    this.onChecked();
  },

  onChecked: function () {
    // Wrapping the call with Ember.run.next prevents a problem where setting isVisible on component
    // causes JS error due to re-rendering.  For example, this occurs when switching the Config Group
    // in Service Config page
    Em.run.next(this, function() {
      console.debug('App.ServiceConfigRadioButton.onChecked');
      this.set('parentView.serviceConfig.value', this.get('value'));
      var components = this.get('parentView.serviceConfig.options');
      if (components) {
        components.forEach(function (_component) {
          if (_component.foreignKeys) {
            _component.foreignKeys.forEach(function (_componentName) {
              if (this.get('parentView.parentView.serviceConfigs').someProperty('name', _componentName)) {
                var component = this.get('parentView.parentView.serviceConfigs').findProperty('name', _componentName);
                component.set('isVisible', _component.displayName === this.get('value'));
              }
            }, this);
          }
        }, this);
      }
    });
  }.observes('checked'),

  disabled: function () {
    return !this.get('parentView.serviceConfig.isEditable') ||
      !['addServiceController', 'installerController'].contains(this.get('controller.wizardController.name')) && /^New\s\w+\sDatabase$/.test(this.get('value'));
  }.property('parentView.serviceConfig.isEditable')
});

App.ServiceConfigComboBox = Ember.Select.extend(App.ServiceConfigPopoverSupport, App.ServiceConfigCalculateId, {
  contentBinding: 'serviceConfig.options',
  selectionBinding: 'serviceConfig.value',
  placeholderBinding: 'serviceConfig.defaultValue',
  classNames: [ 'span3' ]
});


/**
 * Base component for host config with popover support
 */
App.ServiceConfigHostPopoverSupport = Ember.Mixin.create({

  /**
   * Config object. It will instance of App.ServiceConfigProperty
   */
  serviceConfig: null,

  didInsertElement: function () {
    App.popover(this.$(), {
      title: this.get('serviceConfig.displayName'),
      content: this.get('serviceConfig.description'),
      placement: 'right',
      trigger: 'hover'
    });
  }
});

/**
 * Master host component.
 * Show hostname without ability to edit it
 * @type {*}
 */
App.ServiceConfigMasterHostView = Ember.View.extend(App.ServiceConfigHostPopoverSupport, App.ServiceConfigCalculateId, {

  classNames: ['master-host', 'span6'],
  valueBinding: 'serviceConfig.value',

  template: Ember.Handlebars.compile('{{value}}')

});

/**
 * text field property view that enables possibility
 * for check connectio
 * @type {*}
 */
App.checkConnectionView = App.ServiceConfigTextField.extend({
  didInsertElement: function() {
    this._super();
    var kdc = this.get('categoryConfigsAll').findProperty('name', 'kdc_type');
    var propertyAppendTo = this.get('categoryConfigsAll').findProperty('name', 'domains');
    if (propertyAppendTo) propertyAppendTo.set('additionalView', App.CheckDBConnectionView.extend({databaseName: kdc && kdc.get('value')}));
  }
});

/**
 * Show value as plain label in italics
 * @type {*}
 */
App.ServiceConfigLabelView = Ember.View.extend(App.ServiceConfigHostPopoverSupport, App.ServiceConfigCalculateId, {

  classNames: ['master-host', 'span6'],
  valueBinding: 'serviceConfig.value',

  template: Ember.Handlebars.compile('<i>{{view.value}}</i>')
});

/**
 * Base component to display Multiple hosts
 * @type {*}
 */
App.ServiceConfigMultipleHostsDisplay = Ember.Mixin.create(App.ServiceConfigHostPopoverSupport, App.ServiceConfigCalculateId, {

  hasNoHosts: function () {
    console.log('view', this.get('viewName')); //to know which View cause errors
    console.log('controller', this.get('controller').name); //should be slaveComponentGroupsController
    if (!this.get('value')) {
      return true;
    }
    return this.get('value').length === 0;
  }.property('value'),

  hasOneHost: function () {
    return this.get('value').length === 1;
  }.property('value'),

  hasMultipleHosts: function () {
    return this.get('value').length > 1;
  }.property('value'),

  otherLength: function () {
    var len = this.get('value').length;
    if (len > 2) {
      return Em.I18n.t('installer.controls.serviceConfigMultipleHosts.others').format(len - 1);
    } else {
      return Em.I18n.t('installer.controls.serviceConfigMultipleHosts.other');
    }
  }.property('value')

});


/**
 * Multiple master host component.
 * Show hostnames without ability to edit it
 * @type {*}
 */
App.ServiceConfigMasterHostsView = Ember.View.extend(App.ServiceConfigMultipleHostsDisplay, App.ServiceConfigCalculateId, {

  viewName: "serviceConfigMasterHostsView",
  valueBinding: 'serviceConfig.value',

  classNames: ['master-hosts', 'span6'],
  templateName: require('templates/wizard/master_hosts'),

  /**
   * Onclick handler for link
   */
  showHosts: function () {
    var serviceConfig = this.get('serviceConfig');
    App.ModalPopup.show({
      header: Em.I18n.t('installer.controls.serviceConfigMasterHosts.header').format(serviceConfig.category),
      bodyClass: Ember.View.extend({
        serviceConfig: serviceConfig,
        templateName: require('templates/wizard/master_hosts_popup')
      }),
      secondary: null
    });
  }

});

/**
 * Show tabs list for slave hosts
 * @type {*}
 */
App.SlaveComponentGroupsMenu = Em.CollectionView.extend(App.ServiceConfigCalculateId, {

  content: function () {
    return this.get('controller.componentGroups');
  }.property('controller.componentGroups'),

  tagName: 'ul',
  classNames: ["nav", "nav-tabs"],

  itemViewClass: Em.View.extend({
    classNameBindings: ["active"],

    active: function () {
      return this.get('content.active');
    }.property('content.active'),

    errorCount: function () {
      return this.get('content.properties').filterProperty('isValid', false).filterProperty('isVisible', true).get('length');
    }.property('content.properties.@each.isValid', 'content.properties.@each.isVisible'),

    templateName: require('templates/wizard/controls_slave_component_groups_menu')
  })

});

/**
 * <code>Add group</code> button
 * @type {*}
 */
App.AddSlaveComponentGroupButton = Ember.View.extend(App.ServiceConfigCalculateId, {

  tagName: 'span',
  slaveComponentName: null,

  didInsertElement: function () {
    App.popover(this.$(), {
      title: Em.I18n.t('installer.controls.addSlaveComponentGroupButton.title').format(this.get('slaveComponentName')),
      content: Em.I18n.t('installer.controls.addSlaveComponentGroupButton.content').format(this.get('slaveComponentName'), this.get('slaveComponentName'), this.get('slaveComponentName')),
      placement: 'right',
      trigger: 'hover'
    });
  }

});

/**
 * Multiple Slave Hosts component
 * @type {*}
 */
App.ServiceConfigSlaveHostsView = Ember.View.extend(App.ServiceConfigMultipleHostsDisplay, App.ServiceConfigCalculateId, {

  viewName: 'serviceConfigSlaveHostsView',

  classNames: ['slave-hosts', 'span6'],

  valueBinding: 'serviceConfig.value',

  templateName: require('templates/wizard/slave_hosts'),

  /**
   * Onclick handler for link
   */
  showHosts: function () {
    var serviceConfig = this.get('serviceConfig');
    App.ModalPopup.show({
      header: Em.I18n.t('installer.controls.serviceConfigMasterHosts.header').format(serviceConfig.category),
      bodyClass: Ember.View.extend({
        serviceConfig: serviceConfig,
        templateName: require('templates/wizard/master_hosts_popup')
      }),
      secondary: null
    });
  }

});

/**
 * properties for present active slave group
 * @type {*}
 */
App.SlaveGroupPropertiesView = Ember.View.extend(App.ServiceConfigCalculateId, {

  viewName: 'serviceConfigSlaveHostsView',

  group: function () {
    return this.get('controller.activeGroup');
  }.property('controller.activeGroup'),

  groupConfigs: function () {
    console.log("************************************************************************");
    console.log("The value of group is: " + this.get('group'));
    console.log("************************************************************************");
    return this.get('group.properties');
  }.property('group.properties.@each').cacheable(),

  errorCount: function () {
    return this.get('group.properties').filterProperty('isValid', false).filterProperty('isVisible', true).get('length');
  }.property('configs.@each.isValid', 'configs.@each.isVisible')
});

/**
 * DropDown component for <code>select hosts for groups</code> popup
 * @type {*}
 */
App.SlaveComponentDropDownGroupView = Ember.View.extend(App.ServiceConfigCalculateId, {

  viewName: "slaveComponentDropDownGroupView",

  /**
   * On change handler for <code>select hosts for groups</code> popup
   * @param event
   */
  changeGroup: function (event) {
    var host = this.get('content');
    var groupName = $('#' + this.get('elementId') + ' select').val();
    this.get('controller').changeHostGroup(host, groupName);
  },

  optionTag: Ember.View.extend({

    /**
     * Whether current value(OptionTag value) equals to host value(assigned to SlaveComponentDropDownGroupView.content)
     */
    selected: function () {
      return this.get('parentView.content.group') === this.get('content');
    }.property('content')
  })
});

/**
 * Show info about current group
 * @type {*}
 */
App.SlaveComponentChangeGroupNameView = Ember.View.extend(App.ServiceConfigCalculateId, {

  contentBinding: 'controller.activeGroup',
  classNames: ['control-group'],
  classNameBindings: 'error',
  error: false,
  setError: function () {
    this.set('error', false);
  }.observes('controller.activeGroup'),
  errorMessage: function () {
    return this.get('error') ? Em.I18n.t('installer.controls.slaveComponentChangeGroupName.error') : '';
  }.property('error'),

  /**
   * Onclick handler for saving updated group name
   * @param event
   */
  changeGroupName: function (event) {
    var inputVal = $('#' + this.get('elementId') + ' input[type="text"]').val();
    if (inputVal !== this.get('content.name')) {
      var result = this.get('controller').changeSlaveGroupName(this.get('content'), inputVal);
      this.set('error', result);
    }
  }
});
/**
 * View for testing connection to database.
 **/
App.CheckDBConnectionView = Ember.View.extend({
  templateName: require('templates/common/form/check_db_connection'),
  /** @property {string} btnCaption - text for button **/
  btnCaption: function() {
    return this.get('parentView.service.serviceName') === 'KERBEROS'
      ? Em.I18n.t('services.service.config.kdc.btn.idle')
      : Em.I18n.t('services.service.config.database.btn.idle')
  }.property('parentView.service.serviceName'),
  /** @property {string} responseCaption - text for status link **/
  responseCaption: null,
  /** @property {boolean} isConnecting - is request to server activated **/
  isConnecting: false,
  /** @property {boolean} isValidationPassed - check validation for required fields **/
  isValidationPassed: null,
  /** @property {string} databaseName- name of current database **/
  databaseName: null,
  /** @property {boolean} isRequestResolved - check for finished request to server **/
  isRequestResolved: false,
  /** @property {boolean} isConnectionSuccess - check for successful connection to database **/
  isConnectionSuccess: null,
  /** @property {string} responseFromServer - message from server response **/
  responseFromServer: null,
  /** @property {Object} ambariRequiredProperties - properties that need for custom action request **/
  ambariRequiredProperties: null,
  /** @property {Number} currentRequestId - current custom action request id **/
  currentRequestId: null,
  /** @property {Number} currentTaskId - current custom action task id **/
  currentTaskId: null,
  /** @property {jQuery.Deferred} request - current $.ajax request **/
  request: null,
  /** @property {Number} pollInterval - timeout interval for ajax polling **/
  pollInterval: 3000,
  /** @property {string} hostNameProperty - host name property based on service and database names **/
  hostNameProperty: function() {
    if (!/wizard/i.test(this.get('controller.name')) && this.get('parentView.service.serviceName') === 'HIVE') {
      return this.get('parentView.service.serviceName').toLowerCase() + '_hostname';
    } else if (this.get('parentView.service.serviceName') === 'KERBEROS') {
      return 'kdc_host';
    }
    return '{0}_existing_{1}_host'.format(this.get('parentView.service.serviceName').toLowerCase(), this.get('databaseName').toLowerCase());
  }.property('databaseName'),
  /** @property {boolean} isBtnDisabled - disable button on failed validation or active request **/
  isBtnDisabled: function() {
    return !this.get('isValidationPassed') || this.get('isConnecting');
  }.property('isValidationPassed', 'isConnecting'),
  /** @property {object} requiredProperties - properties that necessary for database connection **/
  requiredProperties: function() {
    var propertiesMap = {
      OOZIE: ['oozie.db.schema.name','oozie.service.JPAService.jdbc.username','oozie.service.JPAService.jdbc.password','oozie.service.JPAService.jdbc.driver','oozie.service.JPAService.jdbc.url'],
      HIVE: ['ambari.hive.db.schema.name','javax.jdo.option.ConnectionUserName','javax.jdo.option.ConnectionPassword','javax.jdo.option.ConnectionDriverName','javax.jdo.option.ConnectionURL'],
      KERBEROS: ['kdc_host']
    };
    return propertiesMap[this.get('parentView.service.serviceName')];
  }.property(),
  /** @property {Object} propertiesPattern - check pattern according to type of connection properties **/
  propertiesPattern: function() {
    var patterns = {
      db_connection_url: /jdbc\.url|connectionurl|kdc_host/ig
    };
    if (this.get('parentView.service.serviceName') != "KERBEROS") {
      patterns.user_name = /(username|dblogin)$/ig;
      patterns.user_passwd = /(dbpassword|password)$/ig;
    }
    return patterns;
  }.property('parentView.service.serviceName'),
  /** @property {String} masterHostName - host name location of Master Component related to Service **/
  masterHostName: function() {
    var serviceMasterMap = {
      'OOZIE': 'oozieserver_host',
      'HDFS': 'hadoop_host',
      'HIVE': 'hive_ambari_host',
      'KERBEROS': 'kdc_host'
    };
    return this.get('parentView.categoryConfigsAll').findProperty('name', serviceMasterMap[this.get('parentView.service.serviceName')]).get('value');
  }.property('parentView.service.serviceName', 'parentView.categoryConfigsAll.@each.value'),
  /** @property {Object} connectionProperties - service specific config values mapped for custom action request **/
  connectionProperties: function() {
    var propObj = {};
    for (var key in this.get('propertiesPattern')) {
      propObj[key] = this.getConnectionProperty(this.get('propertiesPattern')[key]);
    }
    return propObj;
  }.property('parentView.categoryConfigsAll.@each.value'),
  /**
   * Properties that stores in local storage used for handling
   * last success connection.
   *
   * @property {Object} preparedDBProperties
   **/
  preparedDBProperties: function() {
    var propObj = {};
    for (var key in this.get('propertiesPattern')) {
      var propName = this.getConnectionProperty(this.get('propertiesPattern')[key], true);
      propObj[propName] = this.get('parentView.categoryConfigsAll').findProperty('name', propName).get('value');
    }
    return propObj;
  }.property(),
  /** Check validation and load ambari properties **/
  didInsertElement: function() {
    var kdc = this.get('parentView.categoryConfigsAll').findProperty('name', 'kdc_type');
    if (kdc) {
      var name = kdc.get('value') == 'Existing MIT KDC' ? 'KDC' : 'AD';
      App.popover(this.$(), {
        title: Em.I18n.t('services.service.config.database.btn.idle'),
        content: Em.I18n.t('installer.controls.checkConnection.popover').format(name),
        placement: 'right',
        trigger: 'hover'
    });
    }
    this.handlePropertiesValidation();
    this.getAmbariProperties();
  },
  /** On view destroy **/
  willDestroyElement: function() {
    this.set('isConnecting', false);
    this._super();
  },
  /**
   * Observer that take care about enabling/disabling button based on required properties validation.
   *
   * @method handlePropertiesValidation
   **/
  handlePropertiesValidation: function() {
    this.restore();
    var isValid = true;
    var properties = [].concat(this.get('requiredProperties'));
    properties.push(this.get('hostNameProperty'));
    properties.forEach(function(propertyName) {
      var property = this.get('parentView.categoryConfigsAll').findProperty('name', propertyName);
      if(property && !property.get('isValid')) isValid = false;
    }, this);
    this.set('isValidationPassed', isValid);
  }.observes('parentView.categoryConfigsAll.@each.isValid', 'parentView.categoryConfigsAll.@each.value', 'databaseName'),

  getConnectionProperty: function(regexp, isGetName) {
    var _this = this;
    var propertyName = _this.get('requiredProperties').filter(function(item) {
      return regexp.test(item);
    })[0];
    return (isGetName) ? propertyName : _this.get('parentView.categoryConfigsAll').findProperty('name', propertyName).get('value');
  },
  /**
   * Set up ambari properties required for custom action request
   *
   * @method getAmbariProperties
   **/
  getAmbariProperties: function() {
    var clusterController = App.router.get('clusterController');
    var _this = this;
    if (!App.isEmptyObject(App.db.get('tmp', 'ambariProperties')) && !this.get('ambariProperties')) {
      this.set('ambariProperties', App.db.get('tmp', 'ambariProperties'));
      return;
    }
    if (App.isEmptyObject(clusterController.get('ambariProperties'))) {
      clusterController.loadAmbariProperties().done(function(data) {
        _this.formatAmbariProperties(data.RootServiceComponents.properties);
      });
    } else {
      this.formatAmbariProperties(clusterController.get('ambariProperties'));
    }
  },

  formatAmbariProperties: function(properties) {
    var defaults = {
      threshold: "60",
      ambari_server_host: location.hostname,
      check_execute_list : "db_connection_check"
    };
    var properties = App.permit(properties, ['jdk.name','jdk_location','java.home']);
    var renameKey = function(oldKey, newKey) {
      if (properties[oldKey]) {
        defaults[newKey] = properties[oldKey];
        delete properties[oldKey];
      }
    };
    renameKey('java.home', 'java_home');
    renameKey('jdk.name', 'jdk_name');
    $.extend(properties, defaults);
    App.db.set('tmp', 'ambariProperties', properties);
    this.set('ambariProperties', properties);
  },
  /**
   * `Action` method for starting connect to current database.
   *
   * @method connectToDatabase
   **/
  connectToDatabase: function() {
    if (this.get('isBtnDisabled')) return;
    this.set('isRequestResolved', false);
    App.db.set('tmp', this.get('parentView.service.serviceName') + '_connection', {});
    this.setConnectingStatus(true);
    if (App.get('testMode')) {
      this.startPolling();
    } else {
      this.runCheckConnection();
    }
  },

  /**
   * runs check connections methods depending on service
   * @return {void}
   * @method runCheckConnection
   */
  runCheckConnection: function() {
    if (this.get('parentView.service.serviceName') === 'KERBEROS') {
      this.runKDCCheck();
    } else {
      this.createCustomAction();
    }
  },

  /**
   * send ajax request to perforn kdc host check
   * @return {App.ajax}
   * @method runKDCCheck
   */
  runKDCCheck: function() {
    return App.ajax.send({
      name: 'admin.kerberos_security.test_connection',
      sender: this,
      data: {
        kdcHostname: this.get('masterHostName')
      },
      success: 'onRunKDCCheckSuccess',
      error: 'onCreateActionError'
    });
  },

  /**
   *
   * @param data
   */
  onRunKDCCheckSuccess: function(data) {
    var statusCode = {
      success: 'REACHABLE',
      failed: 'UNREACHABLE'
    };
    if (data == statusCode['success']) {
      this.setResponseStatus('success');
    } else {
      this.setResponseStatus('failed');
    }
    this.set('responseFromServer', data);
  },

  /**
   * Run custom action for database connection.
   *
   * @method createCustomAction
   **/
  createCustomAction: function() {
    var dbName = this.get('databaseName').toLowerCase() === 'postgresql' ? 'postgres' : this.get('databaseName').toLowerCase();
    var params = $.extend(true, {}, { db_name: dbName }, this.get('connectionProperties'), this.get('ambariProperties'));
    App.ajax.send({
      name: 'custom_action.create',
      sender: this,
      data: {
        requestInfo: {
          parameters: params
        },
        filteredHosts: [this.get('masterHostName')]
      },
      success: 'onCreateActionSuccess',
      error: 'onCreateActionError'
    });
  },
  /**
   * Run updater if task is created successfully.
   *
   * @method onConnectActionS
   **/
  onCreateActionSuccess: function(data) {
    this.set('currentRequestId', data.Requests.id);
    App.ajax.send({
      name: 'custom_action.request',
      sender: this,
      data: {
        requestId: this.get('currentRequestId')
      },
      success: 'setCurrentTaskId'
    });
  },

  setCurrentTaskId: function(data) {
    this.set('currentTaskId', data.items[0].Tasks.id);
    this.startPolling();
  },

  startPolling: function() {
    if (this.get('isConnecting'))
      this.getTaskInfo();
  },

  getTaskInfo: function() {
    var request = App.ajax.send({
      name: 'custom_action.request',
      sender: this,
      data: {
        requestId: this.get('currentRequestId'),
        taskId: this.get('currentTaskId')
      },
      success: 'getTaskInfoSuccess'
    });
    this.set('request', request);
  },

  getTaskInfoSuccess: function(data) {
    var task = data.Tasks;
    this.set('responseFromServer', {
      stderr: task.stderr,
      stdout: task.stdout
    });
    if (task.status === 'COMPLETED') {
      var structuredOut = task.structured_out.db_connection_check;
      if (structuredOut.exit_code != 0) {
        this.set('responseFromServer', {
          stderr: task.stderr,
          stdout: task.stdout,
          structuredOut: structuredOut.message
        });
        this.setResponseStatus('failed');
      } else {
        App.db.set('tmp', this.get('parentView.service.serviceName') + '_connection', this.get('preparedDBProperties'));
        this.setResponseStatus('success');
      }
    }
    if (task.status === 'FAILED') {
      this.setResponseStatus('failed');
    }
    if (/PENDING|QUEUED|IN_PROGRESS/.test(task.status)) {
      Em.run.later(this, function() {
        this.startPolling();
      }, this.get('pollInterval'));
    }
  },

  onCreateActionError: function(jqXhr, status, errorMessage) {
    this.setResponseStatus('failed');
    this.set('responseFromServer', errorMessage);
  },

  setResponseStatus: function(isSuccess) {
    var isSuccess = isSuccess == 'success';
    this.setConnectingStatus(false);
    this.set('responseCaption', isSuccess ? Em.I18n.t('services.service.config.database.connection.success') : Em.I18n.t('services.service.config.database.connection.failed'));
    this.set('isConnectionSuccess', isSuccess);
    this.set('isRequestResolved', true);
  },
  /**
   * Switch captions and statuses for active/non-active request.
   *
   * @method setConnectionStatus
   * @param {Boolean} [active]
   */
  setConnectingStatus: function(active) {
    if (active) {
      this.set('responseCaption', Em.I18n.t('services.service.config.database.connection.inProgress'));
    }
    this.set('controller.testConnectionInProgress', !!active);
    this.set('btnCaption', !!active ? Em.I18n.t('services.service.config.database.btn.connecting') : Em.I18n.t('services.service.config.database.btn.idle'));
    this.set('isConnecting', !!active);
  },
  /**
   * Set view to init status.
   *
   * @method restore
   **/
  restore: function() {
    if (this.get('request')) {
      this.get('request').abort();
      this.set('request', null);
    }
    this.set('responseCaption', null);
    this.set('responseFromServer', null);
    this.setConnectingStatus(false);
    this.set('isRequestResolved', false);
  },
  /**
   * `Action` method for showing response from server in popup.
   *
   * @method showLogsPopup
   **/
  showLogsPopup: function() {
    if (this.get('isConnectionSuccess')) return;
    var _this = this;
    var popup = App.showAlertPopup('Error: {0} connection'.format(this.get('databaseName')));
    if (typeof this.get('responseFromServer') == 'object') {
      popup.set('bodyClass', Em.View.extend({
        templateName: require('templates/common/error_log_body'),
        openedTask: _this.get('responseFromServer')
      }));
    } else {
      popup.set('body', this.get('responseFromServer'));
    }
    return popup;
  }
});

/**
 *
 * @type {*}
 */
App.BaseUrlTextField = Ember.TextField.extend({

  valueBinding: 'repository.baseUrl',

  keyUp: function (event) {
    if (Em.get(this, 'repository.hasError')) {
      Em.set(this, 'repository.hasError', false);
    }
  }

});


});

require.register("views/common/custom_date_popup", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

module.exports = Em.Object.create({

  // Fields values from Select Custom Dates form
  customDateFormFields: Ember.Object.create({
    startDate: null,
    hoursForStart: null,
    minutesForStart: null,
    middayPeriodForStart: null,
    endDate: null,
    hoursForEnd: null,
    minutesForEnd: null,
    middayPeriodForEnd: null
  }),

  errors: Ember.Object.create({
    isStartDateError: false,
    isEndDateError: false
  }),

  errorMessages: Ember.Object.create({
    startDate: '',
    endDate: ''
  }),

  showCustomDatePopup: function (context) {
    var self = this;

    return App.ModalPopup.show({
      header: Em.I18n.t('jobs.table.custom.date.header'),
      onPrimary: function () {
        self.validate();
        if(self.get('errors.isStartDateError') || self.get('errors.isEndDateError')) {
          return false;
        }

        var windowStart = self.createCustomStartDate();
        var windowEnd = self.createCustomEndDate();
        context.set('actualValues', {
          endTime: windowEnd.getTime(),
          startTime: windowStart.getTime()
        });
        this.hide();
      },
      onSecondary: function () {
        context.cancel();
        this.hide();
      },
      bodyClass: App.JobsCustomDatesSelectView.extend({
        controller: self,
        validationErrors: self.get('errorMessages'),
        isValid: self.get('errors')
      })
    });
  },

  createCustomStartDate : function () {
    var startDate = this.get('customDateFormFields.startDate');
    var hoursForStart = this.get('customDateFormFields.hoursForStart');
    var minutesForStart = this.get('customDateFormFields.minutesForStart');
    var middayPeriodForStart = this.get('customDateFormFields.middayPeriodForStart');
    if (startDate && hoursForStart && minutesForStart && middayPeriodForStart) {
      return new Date(startDate + ' ' + hoursForStart + ':' + minutesForStart + ' ' + middayPeriodForStart);
    }
    return null;
  },

  createCustomEndDate : function () {
    var endDate = this.get('customDateFormFields.endDate');
    var hoursForEnd = this.get('customDateFormFields.hoursForEnd');
    var minutesForEnd = this.get('customDateFormFields.minutesForEnd');
    var middayPeriodForEnd = this.get('customDateFormFields.middayPeriodForEnd');
    if (endDate && hoursForEnd && minutesForEnd && middayPeriodForEnd) {
      return new Date(endDate + ' ' + hoursForEnd + ':' + minutesForEnd + ' ' + middayPeriodForEnd);
    }
    return null;
  },

  clearErrors: function () {
    var errorMessages = this.get('errorMessages');
    Em.keys(errorMessages).forEach(function (key) {
      errorMessages.set(key, '');
    }, this);
    var errors = this.get('errors');
    Em.keys(errors).forEach(function (key) {
      errors.set(key, false);
    }, this);
  },

  // Validation for every field in customDateFormFields
  validate: function () {
    var formFields = this.get('customDateFormFields');
    var errors = this.get('errors');
    var errorMessages = this.get('errorMessages');
    this.clearErrors();
    // Check if feild is empty
    Em.keys(errorMessages).forEach(function (key) {
      if (!formFields.get(key)) {
        errors.set('is' + key.capitalize() + 'Error', true);
        errorMessages.set(key, Em.I18n.t('jobs.customDateFilter.error.required'));
      }
    }, this);
    // Check that endDate is after startDate
    var startDate = this.createCustomStartDate();
    var endDate = this.createCustomEndDate();
    if (startDate && endDate && (startDate > endDate)) {
      errors.set('isEndDateError', true);
      errorMessages.set('endDate', Em.I18n.t('jobs.customDateFilter.error.date.order'));
    }
  }
});

});

require.register("views/common/editable_list", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.EditableList = Ember.View.extend({

  templateName: require('templates/common/editable_list'),

  items: [], // items show on list
  resources: [], // original resources including all items
  itemsOriginal: [], //backup of items

  editMode: false,
  input: '',
  typeahead: [],
  selectedTypeahed: 0,

  init: function () {
    this._super();
    this.set('itemsOriginal', Em.copy(this.get('items')));
    this.set('input', '');
    this.set('editMode', false);
  },

  onPrimary: function (event) {
    this.set('editMode', false);
    this.set('input', '');
    if(event){
      event.stopPropagation();
    }
  },

  onSecondary: function () {
    // restore all items
    this.set('items', this.get('itemsOriginal'));
    this.set('input', '');
    this.set('editMode', false);
  },

  enableEditMode: function() {
    this.set('input', '');
    this.set('editMode', true);
  },

  removeFromItems: function(event) {
    var items = this.get('items');
    items.removeObject(event.context);
    this.set('input', '');
  },

  /**
   * available items to add, will show up typing ahead
   */
  availableItemsToAdd: function () {
    var allItems = Em.copy(this.get('resources'));
    var input = this.get('input');
    var toRemove = [];
    var existed = this.get('items');
    allItems.forEach(function(item) {
      if (item.name.indexOf(input) < 0 || existed.findProperty('name', item.name)) {
        toRemove.push(item);
      }
    });
    toRemove.forEach(function(item) {
      allItems.removeObject(item);
    });
    return allItems;
  }.property('items', 'input'),

  addItem: function(event) {
    var items = this.get('items');
    var toAdd;
    if (event.context) {
      toAdd = event.context;
    } else if (this.get('typeahead.length') > 0){
      toAdd = this.get('typeahead')[this.get('selectedTypeahead')];
    }
    items.pushObject(event.context);
    this.set('input', '');
  },

  updateTypeahead: function() {
    var newValue = this.get('input');
    var self = this;
    if(newValue){
      var newValue = newValue.split(',');
      if( newValue.length > 1){
        // If coma separated string, then just add all items to list
        newValue.forEach( function(item) {
          self.addItem(item);
        });
        self.clearInput();
      } else {
        // Load typeahed items based on current input
        var items = self.get('availableItemsToAdd');
        self.set('typeahead', items);
        self.set('selectedTypeahed', 0);
      }
    } else {
      self.set('typeahead', []);
      self.set('selectedTypeahed', 0);
    }

  }.observes('input')
});

});

require.register("views/common/filter_combo_cleanable", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');


/**
 * Filter component with custom dropdown and cross-circled cleaner. Build on Twitter Bootstrap styles
 * @type {*}
 */
App.FilterComboCleanableView = Ember.View.extend({
  templateName: require('templates/common/filter_combo_cleanable'),

  classNames: ['filter-combobox', 'input-append'],

  didInsertElement: function() {
    App.popover(this.$("input[type=text]"), {
      title: this.get('popoverDescription')[0],
      content: this.get('popoverDescription')[1],
      placement: 'bottom',
      trigger: 'hover'
    });
    this.clearFilter();
  },

  placeHolder: function(){
    return this.t('common.combobox.placeholder');
  }.property(),

  /**
   * Onclick handler for dropdown menu
   * @param event
   */
  selectFilterColumn: function(event){
    var column = event.context;
    column.set('selected', !column.get('selected'));
  },

  filterNotEmpty: function(target){
    return (this.get('filter').length > 0);
  }.property('filter'),

  /**
   * clear Filter textfield
   */
  clearFilter: function() {
    this.set('filter', '');
  }
});

});

require.register("views/common/filter_combobox", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');


/**
 * Filter component with custom dropdown. Build on Twitter Bootstrap styles
 * @type {*}
 */
App.FilterComboboxView = Ember.View.extend({
  templateName: require('templates/common/filter_combobox'),

  classNames: ['filter-combobox', 'input-append'],

  //filter: '', binding outside
  //columns: [], binding outside

  placeHolder: function(){
    return this.t('common.combobox.placeholder');
  }.property(),

  /**
   * Onclick handler for dropdown menu
   * @param event
   */
  selectFilterColumn: function(event){
    var column = event.context;
    column.set('selected', !column.get('selected'));
  }

});

});

require.register("views/common/filter_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * Wrapper View for all filter components. Layout template and common actions are located inside of it.
 * Logic specific for data component(input, select, or custom multi select, which fire any changes on interface) are
 * located in inner view - <code>filterView</code>.
 *
 * If we want to have input filter, put <code>textFieldView</code> to it.
 * All inner views implemented below this view.
 * @type {*}
 */

var App = require('app');

var wrapperView = Ember.View.extend({
  classNames: ['view-wrapper'],
  layout: Ember.Handlebars.compile('<a href="#" {{action "clearFilter" target="view"}} class="ui-icon ui-icon-circle-close"></a> {{yield}}'),
  template: Ember.Handlebars.compile(
    '{{#if view.fieldId}}<input type="hidden" id="{{unbound view.fieldId}}" value="" />{{/if}}' +
    '{{view view.filterView}}' +
    '{{#if view.showApply}}<button {{action "setValueOnApply" target="view"}} class="apply-btn btn"><span>{{t common.apply}}</span></button>{{/if}} '
  ),

  value: null,

  /**
   * Column index
   */
  column: null,

  /**
   * If this field is exists we dynamically create hidden input element and set value there.
   * Used for some cases, where this values will be used outside of component
   */
  fieldId: null,

  /**
   * This option is useful for Em.Select, if you want get events when same option selected more than one time a raw.
   * This property is Object[] (for example see App.MainConfigHistoryView.modifiedFilterView),
   * that have followed structure:
   *
   * <code>
   * [
   *  {
   *    values: ['DefinedValue1', 'DefinedValue2'], // this properties should be defined in `content` property
   *    displayAs: 'Choose me' // this value will be displayed
   *  }
   * ]
   * </code>
   * @type {Array}
   *
   **/
  //TODO delete if will not be used
  triggeredOnSameValue: null,

  clearFilter: function () {
    this.set('value', this.get('emptyValue'));
    if (this.get('setPropertyOnApply')) {
      this.setValueOnApply();
    }
    return false;
  },

  setValueOnApply: function() {
    if(this.get('value') == null){
      this.set('value', '')
    }
    this.set(this.get('setPropertyOnApply'), this.get('value'));
    return false;
  },

  /**
   * Use to determine whether filter is clear or not. Also when we want to set empty value
   */
  emptyValue: '',

  /**
   * empty value that actually applied to filtering on server,
   * used with Select filter where displayed and actual filter value are different
   */
  appliedEmptyValue: "",

  /**
   * Whether our <code>value</code> is empty or not
   * @return {Boolean}
   */
  isEmpty: function () {
    if (Em.isNone(this.get('value'))) {
      return true;
    }
    return this.get('value').toString() === this.get('emptyValue').toString();
  },

  setValue: function (value) {
    this.set('value', value);
  },

  /**
   * Show/Hide <code>Clear filter</code> button.
   * Also this method updates computed field related to <code>fieldId</code> if it exists.
   * Call <code>onChangeValue</code> callback when everything is done.
   */
  showClearFilter: function () {
    if (!this.get('parentNode')) {
      return;
    }
    // get the sort view element in the same column to current filter view to highlight them together
    var relatedSort = $(this.get('element')).parents('thead').find('.sort-view-' + this.get('column'));
    if (this.isEmpty()) {
      this.get('parentNode').removeClass('active-filter');
      this.get('parentNode').addClass('notActive');
      relatedSort.removeClass('active-sort');
    } else {
      this.get('parentNode').removeClass('notActive');
      this.get('parentNode').addClass('active-filter');
      relatedSort.addClass('active-sort');
    }

    if (this.get('fieldId')) {
      this.$('> input').eq(0).val(this.get('value'));
    }

    this.onChangeValue();
  }.observes('value'),

  /**
   * Callback for value changes
   */
  onChangeValue: Em.K,

  /**
   * Filter components is located here. Should be redefined
   */
  filterView: Em.View,

  /**
   * Update class of parentNode(hide clear filter button) on page load
   */
  didInsertElement: function () {
    var parent = this.$().parent();
    this.set('parentNode', parent);
    parent.addClass('notActive');
    //TODO delete if will not be used
    //this.checkSelectSpecialOptions();
  },

  /**
   * Check for Em.Select that should use dispatching event when option with same value selected more than one time.
   **/
    //TODO delete if will not be used
  checkSelectSpecialOptions: function () {
    // check predefined property
    if (!this.get('triggeredOnSameValue') || !this.get('triggeredOnSameValue').length) return;
    // add custom additional observer that will handle property changes
    this.addObserver('value', this, 'valueCustomObserver');
    // get the full class attribute to find our select
    var classInlineAttr = this.get('fieldType').split(',')
      .map(function (className) {
        return '.' + className.trim();
      }).join('');
    this.set('classInlineAttr', classInlineAttr);
    this.get('triggeredOnSameValue').forEach(function (triggeredValue) {
      triggeredValue.values.forEach(function (value, index) {
        // option with property `value`
        var $optionEl = $(this.get('element')).find(classInlineAttr)
          .find('option[value="' + value + '"]');
        // should be displayed with `displayAs` caption
        $optionEl.text(triggeredValue.displayAs);
        // the second one option should be hidden
        // as the result, on init stage we show only one option that could be selected
        if (index == 1) {
          $optionEl.css('display', 'none');
        }
      }, this);
    }, this);
  },
  /**
   *
   * Custom observer that used for special case of Em.Select related to dispatching event
   * when option with same value selected more than one time.
   *
   **/
  //TODO delete if will not be used
  valueCustomObserver: function () {
    var hiddenValue;
    this.get('triggeredOnSameValue').forEach(function (triggeredValue) {
      var values = triggeredValue.values;
      // find current selected value from `values` list
      var currentValueIndex = values.indexOf(this.get('value'));
      if (currentValueIndex < 0) return;
      // value assigned to hidden option
      hiddenValue = values[Number(currentValueIndex == 0)];
    }, this);
    if (hiddenValue) {
      // our select
      var $select = $(this.get('element')).find(this.get('classInlineAttr'));
      // now hide option with current value
      $select.find('option[value="{0}"]'.format(this.get('value'))).css('display', 'none');
      // and show option that was hidden
      $select.find('option[value="{0}"]'.format(hiddenValue)).css('display', 'block');
    }
  }
});

/**
 * Simple input control for wrapperView
 */
var textFieldView = Ember.TextField.extend({
  type: 'text',
  placeholder: Em.I18n.t('any'),
  valueBinding: "parentView.value"
});

/**
 * Simple multiselect control for wrapperView.
 * Used to render blue button and popup, which opens on button click.
 * All content related logic should be implemented manually outside of it
 */
var componentFieldView = Ember.View.extend({
  classNames: ['btn-group'],
  classNameBindings: ['isFilterOpen:open:'],

  /**
   * Whether popup is shown or not
   */
  isFilterOpen: false,

  /**
   * We have <code>value</code> property similar to inputs <code>value</code> property
   */
  valueBinding: 'parentView.value',

  /**
   * Clear filter to initial state
   */
  clearFilter: function () {
    this.set('value', '');
  },

  /**
   * Onclick handler for <code>cancel filter</code> button
   */
  closeFilter: function () {
    $(document).unbind('click');
    this.set('isFilterOpen', false);
  },

  /**
   * Onclick handler for <code>apply filter</code> button
   */
  applyFilter: function () {
    this.closeFilter();
  },

  /**
   * Onclick handler for <code>show component filter</code> button.
   * Also this function is used in some other places
   */
  clickFilterButton: function () {
    var self = this;
    this.toggleProperty('isFilterOpen');
    if (this.get('isFilterOpen')) {

      var dropDown = this.$('.filter-components');
      var firstClick = true;
      $(document).bind('click', function (e) {
        if (!firstClick && $(e.target).closest(dropDown).length == 0) {
          self.set('isFilterOpen', false);
          $(document).unbind('click');
        }
        firstClick = false;
      });
    }
  }
});

/**
 * Simple select control for wrapperView
 */
var selectFieldView = Ember.Select.extend({
  selectionBinding: 'parentView.selected',
  contentBinding: 'parentView.content',
  optionValuePath: "content.value",
  optionLabelPath: "content.label"
});

/**
 * Result object, which will be accessible outside
 * @type {Object}
 */
module.exports = {
  /**
   * You can access wrapperView outside
   */
  wrapperView: wrapperView,

  /**
   * And also controls views if need it
   */
  textFieldView: textFieldView,
  selectFieldView: selectFieldView,
  componentFieldView: componentFieldView,

  /**
   * Quick create input filters
   * @param config parameters of <code>wrapperView</code>
   */
  createTextView: function (config) {
    config.fieldType = config.fieldType || 'input-medium';
    config.filterView = textFieldView.extend({
      classNames: [ config.fieldType ]
    });

    return wrapperView.extend(config);
  },

  /**
   * Quick create multiSelect filters
   * @param config parameters of <code>wrapperView</code>
   */
  createComponentView: function (config) {
    config.clearFilter = function () {
      this.forEachChildView(function (item) {
        if (item.clearFilter) {
          item.clearFilter();
        }
      });
      return false;
    };

    return wrapperView.extend(config);
  },

  /**
   * Quick create select filters
   * @param config parameters of <code>wrapperView</code>
   */
  createSelectView: function (config) {
    config.fieldType = config.fieldType || 'input-medium';
    config.filterView = selectFieldView.extend({
      classNames: config.fieldType.split(','),
      attributeBindings: ['disabled', 'multiple'],
      disabled: false
    });

    config.valueBinding = 'selected.value';
    config.clearFilter = function () {
      this.set('selected', this.get('content').findProperty('value', this.get('emptyValue')));
    };
    config.setValue = function (value) {
      this.set('selected', this.get('content').findProperty('value', value));
    };
    return wrapperView.extend(config);
  },
  /**
   * returns the filter function, which depends on the type of property
   * @param type
   * @param isGlobal check is search global
   * @return {Function}
   */
  getFilterByType: function (type, isGlobal) {
    switch (type) {
      case 'ambari-bandwidth':
        return function (rowValue, rangeExp) {
          var compareChar = isNaN(rangeExp.charAt(0)) ? rangeExp.charAt(0) : false;
          var compareScale = rangeExp.charAt(rangeExp.length - 1);
          var compareValue = compareChar ? parseFloat(rangeExp.substr(1, rangeExp.length)) : parseFloat(rangeExp.substr(0, rangeExp.length));
          var match = false;
          if (rangeExp.length == 1 && compareChar !== false) {
            // User types only '=' or '>' or '<', so don't filter column values
            match = true;
            return match;
          }
          switch (compareScale) {
            case 'g':
              compareValue *= 1073741824;
              break;
            case 'm':
              compareValue *= 1048576;
              break;
            case 'k':
              compareValue *= 1024;
              break;
            default:
              //default value in GB
              compareValue *= 1073741824;
          }
          rowValue = (jQuery(rowValue).text()) ? jQuery(rowValue).text() : rowValue;

          var convertedRowValue;
          if (rowValue === '<1KB') {
            convertedRowValue = 1;
          } else {
            var rowValueScale = rowValue.substr(rowValue.length - 2, 2);
            switch (rowValueScale) {
              case 'KB':
                convertedRowValue = parseFloat(rowValue) * 1024;
                break;
              case 'MB':
                convertedRowValue = parseFloat(rowValue) * 1048576;
                break;
              case 'GB':
                convertedRowValue = parseFloat(rowValue) * 1073741824;
                break;
            }
          }

          switch (compareChar) {
            case '<':
              if (compareValue > convertedRowValue) match = true;
              break;
            case '>':
              if (compareValue < convertedRowValue) match = true;
              break;
            case false:
            case '=':
              if (compareValue == convertedRowValue) match = true;
              break;
          }
          return match;
        };
        break;
      case 'duration':
        return function (rowValue, rangeExp) {
          var compareChar = isNaN(rangeExp.charAt(0)) ? rangeExp.charAt(0) : false;
          var compareScale = rangeExp.charAt(rangeExp.length - 1);
          var compareValue = compareChar ? parseFloat(rangeExp.substr(1, rangeExp.length)) : parseFloat(rangeExp.substr(0, rangeExp.length));
          var match = false;
          if (rangeExp.length == 1 && compareChar !== false) {
            // User types only '=' or '>' or '<', so don't filter column values
            match = true;
            return match;
          }
          switch (compareScale) {
            case 's':
              compareValue *= 1000;
              break;
            case 'm':
              compareValue *= 60000;
              break;
            case 'h':
              compareValue *= 3600000;
              break;
            default:
              compareValue *= 1000;
          }
          rowValue = (jQuery(rowValue).text()) ? jQuery(rowValue).text() : rowValue;

          switch (compareChar) {
            case '<':
              if (compareValue > rowValue) match = true;
              break;
            case '>':
              if (compareValue < rowValue) match = true;
              break;
            case false:
            case '=':
              if (compareValue == rowValue) match = true;
              break;
          }
          return match;
        };
        break;
      case 'date':
        return function (rowValue, rangeExp) {
          var match = false;
          var timePassed = App.dateTime() - rowValue;
          switch (rangeExp) {
            case 'Past 1 hour':
              match = timePassed <= 3600000;
              break;
            case 'Past 1 Day':
              match = timePassed <= 86400000;
              break;
            case 'Past 2 Days':
              match = timePassed <= 172800000;
              break;
            case 'Past 7 Days':
              match = timePassed <= 604800000;
              break;
            case 'Past 14 Days':
              match = timePassed <= 1209600000;
              break;
            case 'Past 30 Days':
              match = timePassed <= 2592000000;
              break;
            case 'Any':
              match = true;
              break;
          }
          return match;
        };
        break;
      case 'number':
        return function (rowValue, rangeExp) {
          var compareChar = rangeExp.charAt(0);
          var compareValue;
          var match = false;
          if (rangeExp.length == 1) {
            if (isNaN(parseInt(compareChar))) {
              // User types only '=' or '>' or '<', so don't filter column values
              match = true;
              return match;
            }
            else {
              compareValue = parseFloat(parseFloat(rangeExp).toFixed(2));
            }
          }
          else {
            if (isNaN(parseInt(compareChar))) {
              compareValue = parseFloat(parseFloat(rangeExp.substr(1, rangeExp.length)).toFixed(2));
            }
            else {
              compareValue = parseFloat(parseFloat(rangeExp.substr(0, rangeExp.length)).toFixed(2));
            }
          }
          rowValue = parseFloat((jQuery(rowValue).text()) ? jQuery(rowValue).text() : rowValue);
          match = false;
          switch (compareChar) {
            case '<':
              if (compareValue > rowValue) match = true;
              break;
            case '>':
              if (compareValue < rowValue) match = true;
              break;
            case '=':
              if (compareValue == rowValue) match = true;
              break;
            default:
              if (rangeExp == rowValue) match = true;
          }
          return match;
        };
        break;
      case 'sub-resource':
        return function (origin, compareValue) {
          if (!Array.isArray(compareValue) || compareValue.length === 0) return true;

          return origin.some(function (item) {
            for (var i = 0, l = compareValue.length; i < l; i++) {
              if(item.get(compareValue[i].property) !== compareValue[i].value) return false
            }
            return true;
          });
        };
        break;
      case 'multiple':
        return function (origin, compareValue) {
          var options = compareValue.split(',');
          if (typeof (origin) === "string") {
            var rowValue = origin;
          } else {
            var rowValue = origin.mapProperty('componentName').join(" ");
          }
          var str = new RegExp(compareValue, "i");
          for (var i = 0; i < options.length; i++) {
            if (!isGlobal) {
              str = new RegExp('(\\W|^)' + options[i] + '(\\W|$)');
            }
            if (rowValue.search(str) !== -1) {
              return true;
            }
          }
          return false;
        };
        break;
      case 'boolean':
        return function (origin, compareValue) {
          return origin === compareValue;
        };
        break;
      case 'select':
        return function (origin, compareValue) {
          return origin == compareValue;
        };
        break;
      case 'os':
        return function (origin, compareValue) {
          return origin.getEach('osType').contains(compareValue)
        };
        break;
      case 'range':
        return function (origin, compareValue) {
          if (compareValue[1] && compareValue[0]) {
            return origin >= compareValue[0] && origin <= compareValue[1];
          } else if (compareValue[0]) {
            return origin >= compareValue[0];
          } else if (compareValue[1]) {
            return origin <= compareValue[1]
          }
          return true;
        };
        break;
      case 'alert_status':
        /**
         * origin - alertDefinition.summary
         * compareValue - "OK|WARNING|CRITICAL|UNKNOWN|PENDING"
         * PENDING means that OK is 0, WARNING is 0, CRITICAL is 0 and UNKNOWN is 0
         */
        return function (origin, compareValue) {
          if ('PENDING' === compareValue) {
            var isPending = true;
            Em.keys(origin).forEach(function(state) {
              if (origin[state] && (origin[state].count > 0 || origin[state].maintenanceCount > 0)) {
                isPending = false;
              }
            });
            return isPending;
          }
          return !!origin[compareValue] && (origin[compareValue].count > 0 || origin[compareValue].maintenanceCount > 0);
        };
        break;
      case 'alert_group':
        return function (origin, compareValue) {
          return origin.mapProperty('id').contains(compareValue);
        };
        break;
      case 'enable_disable':
        return function (origin, compareValue) {
          return origin == (compareValue == 'enabled');
        };
        break;
      case 'string':
      default:
        return function (origin, compareValue) {
          var regex = new RegExp(compareValue, "i");
          return regex.test(origin);
        }
    }
  }
};

});

require.register("views/common/form/field", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.FormFieldTemplate = Em.View.extend({
  templateDir: 'templates/common/form/',
  defaultTemplate: 'field',
  classNames: ["control-group"],
  init: function(){
    this._super();
    this.updateTemplate();
  },
  updateTemplate: function(){
    var field = this.get('field');
    if(field) {
      var templateName = '';
      switch(field.get('displayType')) {
        case 'checkbox':
          templateName = field.get('displayType');
          break;
        default:
          templateName = this.get('defaultTemplate');
          break;
      }

      this.set('templateName', require(this.get('templateDir') + templateName));
    }
  }.observes('field')
});
});

require.register("views/common/grid", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var validator = require('utils/validator');

App.GridFilterObject = Em.Object.extend({
  checked:false
});

App.GridFilter = Em.View.extend({
  tagName:"ul",
  classNames:['filter'],
  templateName:require('templates/common/grid/filter'),
  attributeBindings:['style'],
  getHeader:function () {
    return this.get('header')
  },
  filters:function () {
    return this.get('header._filters');
  }.property('header._filters')
});

App.GridHeader = Em.View.extend({
  templateName:require('templates/common/grid/header'),
  tagName:'th',
  filterable:true,
  showFilter:false,
  getGrid:function () {
    return this.get('grid');
  },
  _filters:[],
  doFilter:function () {
    console.log(this.get('grid'));
  },
  toggleFilter:function () {
    this.set('showFilter', 1 - this.get('showFilter'));
  },
  applyFilter:function () {
    console.warn('APPLYING FILTERS');

    var filters = this.get('_filters');
    var filterValues = [];
    $.each(filters, function(){
      if(this.get('checked')) {
        filterValues.push(this.get('value'));
      }
    });

    var grid = this.get('grid');
    grid.addFilters(this.get('name'), filterValues);
    this.set('showFilter', false);
  },
  init:function () {
    this._super();
    if (!this.get('_filters').length) {
      this.filterValues();
      var thisHeader = this;
      this.set('filter', App.GridFilter.extend({ header:thisHeader }));
    }
  },

  filterValues:function () {
    var gridFilters = this.get('grid._filters');
    if (gridFilters && gridFilters[this.get('name')]) {
      var filters = this.get('grid._filters')[this.get('name')];
      // there should be something like filter preparing
      var newFilters = [];
      $.each(filters, function (i, v) {
        newFilters.push(App.GridFilterObject.create({label:v, value:v}));
      });

      this.set('_filters', newFilters);
    }
  }.observes('grid._filters')
});

App.GridRow = Em.View.extend({
  tagName:'tr',
  init:function (options) {
    var object = this.get('object');
    var grid = this.get('grid');
    var fieldNames = grid.get('fieldNames');
    var template = '';

    if (fieldNames) {
      $.each(grid.get('fieldNames'), function (i, field) {
        template += "<td>" + object.get(field) + "</td>";
      });

      this.set('template', Em.Handlebars.compile(template));
    }
    return this._super();
  }
});

App.GridPage = Em.Object.extend({
  activeClass:function () {
    return this.get('active') ? "active" : "";
  }.property('active'),
  active:function () {
    return parseInt(this.get('number')) == parseInt(this.get('pager.grid.currentPage'));
  }.property('pager.grid.currentPage')
});

App.GridPager = Em.View.extend({

  pages:[],
  templateName:require('templates/common/grid/pager'),
  classNames:['pagination'],

  activatePrevPage:function () {
    var current = this.get('grid.currentPage');
    if (current > 1) this.set('grid.currentPage', current - 1);
  },
  activateNextPage:function () {
    var current = this.get('grid.currentPage');
    if (current < this.get('pages').length) this.set('grid.currentPage', current + 1);
  },

  prevPageDisabled:function () {
    return this.get('grid.currentPage') > 1 ? false : "disabled";
  }.property('grid.currentPage'),

  nextPageDisabled:function () {
    return this.get('grid.currentPage') < this.get('pages').length ? false : "disabled";
  }.property('grid.currentPage'),

  init:function () {
    this._super();
    this.clearPages();
    this.pushPages();
  },

  activatePage:function (event) {
    var page = event.context;
    this.get('grid').set('currentPage', parseInt(event.context.get('number')));
  },

  clearPages:function () {
    this.set('pages', []);
  },

  pushPages:function () {
    var thisPager = this;
    var pages = this.get('grid._pager.pages');
    $.each(pages, function () {
      var thisNumber = this;
      thisPager.get('pages').push(App.GridPage.create({
        number:thisNumber,
        pager:thisPager
      }));
    })
  }.observes('grid._pager')
});

App.Grid = Em.View.extend({
  _columns:{}, // not used
  _filters:{}, // prepared filters from data values
  _pager:{pages:[1, 2, 3, 4, 5]}, // observed by pager to config it

  _collection:{className:false, staticOptions:{}}, // collection config
  currentPage:1,
  fieldNames:[],
  appliedFilters:{},
  filteredArray:[],
  columns:[],
  collection:[],
  initComleted:false,
  rows:[],
  templateName:require('templates/main/admin/audit'),

  init:function () {
    console.warn("  Grid INIT  ");
    this._super();
    this.prepareColumns(); // should be the 1
    this.prepareCollection();
    this.preparePager();
  },

  preparePager:function () {
//    this.set('pager', App.GridPager.extend({ grid:this })); ask to hide
  },

  addFilters: function(field, values){
    var filters = this.get('appliedFilters');
    filters[field] = values;

    var collection = this.get('_collection.className');
    collection = collection.find();
    var arrayCollection = collection.filter(function(data) {
      var oneFilterFail = false;
      $.each(filters, function(fieldname, values){
        if(values.length && values.indexOf(data.get(fieldname)) == -1) {
          return oneFilterFail = true;
        }
      });
      return !oneFilterFail;
    });

    this.set('filteredArray', arrayCollection);
  },

  prepareCollection:function () {
    if (validator.empty(this.get('_collection.className'))) {
      throw "_collection.className field is not defined";
    }
    var collection = this.get('_collection.className');
    this.set('collection', collection.find(this.get('_collection.staticOptions')));
  },

  addColumn:function (options) {
    options.grid = this;
    if (validator.empty(options.name)) {
      throw "define column name";
    }

    if (this.get('_columns.' + options.name)) {
      throw "column with this '" + options.name + "' already exists";
    }

    var field = App.GridHeader.extend(options);
    this.columns.push(field);

    if (field.filterable || 1) { // .filterable - field not working :(
      this.fieldNames.push(options.name);
    }
  },

  clearColumns:function () {
    this.set('_columns', {});
    this.set('columns', []);
    this.set('fieldNames', []);
  },

  prepareColumns:function () {
    this.clearColumns();
  },

  prepareFilters:function () {
    var thisGrid = this;
    var collection = this.get('collection');
    var fieldNames = this.get('fieldNames');
    var options = {};

    if (collection && collection.content) {
      collection.forEach(function (object, i) {
        $.each(fieldNames, function (j, field) {
          if (!options[field]) {
            options[field] = [];
          }

          var filter = object.get(field);
          if (options[field].indexOf(filter) == -1) {
            options[field].push(filter);
          }
        });
      });

      thisGrid.set('_filters', options);
    }
  }.observes('collection.length'),


  clearRows:function () {
    this.set('rows', [])
  },

  prepareRows:function () {
    var collection = this.get('collection');
    var thisGrid = this;
    this.clearRows();
    console.warn("PREPARE ROWS LEN:", collection.get('length'));
    var i=1;

    if (collection && collection.content) {
      collection.forEach(function (object, i) {
        var row = App.GridRow.extend({grid:thisGrid, object:object});
        thisGrid.rows.push(row);
      });
    }
  }.observes('collection.length'),

  filteredRows:function () {
    var collection = this.get('filteredArray');
    var thisGrid = this;
    this.clearRows();

    collection.forEach(function (object) {
      var row = App.GridRow.extend({grid:thisGrid, object:object});
      thisGrid.rows.push(row);
    });
  }.observes('filteredArray')
});
});

require.register("views/common/metric", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * use: {{view App.MetricFilteringWidget controllerBinding="App.router.mainChartsController"}}
 * set controller.metric field with metric value
 * widget assign itself to controller like metricWidget (controller.get('metricWidget'))
 * @type {*}
 */
App.MetricFilteringWidget = Em.View.extend({
  classNames:['metric-filtering-widget'],
  /**
   * chosen metric value
   */
  chosenMetric:null,
  chosenMoreMetric:null,
  showMore:0, // toggle more metrics indicator
  /**
   * metrics
   */
  metrics:[
    Em.Object.create({ label:Em.I18n.t('metric.default'), value:null}),
    Em.Object.create({ label:Em.I18n.t('metric.cpu'), value:'cpu'}),
    Em.Object.create({ label:Em.I18n.t('metric.memory'), value:'memory'}),
    Em.Object.create({ label:Em.I18n.t('metric.network'), value:'network'}),
    Em.Object.create({ label:Em.I18n.t('metric.io'), value:'io'})
  ],


  moreMetrics:[
    Em.Object.create({ label:Em.I18n.t('metric.more.cpu'), code:'cpu', items:[
      Em.Object.create({value:"cpu_nice"}),
      Em.Object.create({value:"cpu_wio"}),
      Em.Object.create({value:"cpu_user"}),
      Em.Object.create({value:"cpu_idle"}),
      Em.Object.create({value:"cpu_system"}),
      Em.Object.create({value:"cpu_aidle"})
    ] }),

    Em.Object.create({ label:Em.I18n.t('metric.more.disk'), code:'disk',
      items:[
        Em.Object.create({value:'disk_free'}),
        Em.Object.create({value:'disk_total'}),
        Em.Object.create({value:'part_max_used'})
      ]
    }),

    Em.Object.create({ label:Em.I18n.t('metric.more.load'), code:'load',
      items:[
        Em.Object.create({value:'load_one'}),
        Em.Object.create({value:'load_five'}),
        Em.Object.create({value:'load_fifteen'})
      ]
    }),

    Em.Object.create({ label:Em.I18n.t('metric.more.memory'), code:'memory',
      items:[
        Em.Object.create({value:'swap_free'}),
        Em.Object.create({value:'cpu'})
      ]
    }),

    Em.Object.create({ label:Em.I18n.t('metric.more.network'), code:'network',
      items:[
        Em.Object.create({value:'bytes_out'}),
        Em.Object.create({value:'bytes_in'}),
        Em.Object.create({value:'pkts_in'}),
        Em.Object.create({value:'pkts_out'})
      ]
    }),

    Em.Object.create({ label:Em.I18n.t('metric.more.process'), code:'process',
      items:[
        Em.Object.create({value:'proc_run'}),
        Em.Object.create({value:'proc_total'})
      ]
    })

  ],

  /**
   * return array of chosen metrics
   */
  chosenMetrics:function () {
    return this.get('chosenMetric') ? [this.get('chosenMetric')] : this.get('defaultMetrics');
  }.property('chosenMetric'),

  /**
   * metric item view
   */
  itemView:Em.View.extend({
    tagName:'li',
    classNameBindings:['disabled'],
    disabled:function () {
      return this.get('isActive') ? "disabled" : false;
    }.property('isActive'),
    isActive:function () {
      return this.get('metric.value') == this.get('widget.chosenMetric');
    }.property('widget.chosenMetric'),
    label:function () {
      return this.get('metric.label');
    }.property('metric.label'),
    template:Em.Handlebars.compile('<a {{action activate view.metric.value target="view.widget" href="#" }}>{{unbound view.label}}</a>')
  }),

  moreItemView:function () {
    return this.get('itemView').extend({
      label:function () {
        return this.get('metric.value');
      }.property('metric.value')
    });
  }.property(),

  /**
   * return default selected metrics (currently - all)
   */
  defaultMetrics:function () {
    var values = [];
    $.each(this.get('metrics'), function () {
      if (this.value) {
        values.push(this.value);
      }
    });
    return values;
  }.property(),

  bindToController:function () {
    var thisW = this;
    var controller = this.get('controller');
    controller.set('metricWidget', thisW);
  },

  toggleMore:function () {
    this.set('showMore', 1 - this.get('showMore'));
  },

  /**
   * assign this widget to controller, prepare items by metricsConfig
   */
  init:function () {
    this._super();
    this.bindToController();
  },

  /**
   * write active metric to widget
   * @param event
   */
  activate:function (event) {
    this.set('chosenMetric', event.context);
  },

  templateName:require('templates/common/metric')
});
});

require.register("views/common/modal_popup", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ModalPopup = Ember.View.extend({

  viewName: 'modalPopup',
  templateName: require('templates/common/modal_popup'),
  header: '&nbsp;',
  body: '&nbsp;',
  encodeBody: true,
  // define bodyClass which extends Ember.View to use an arbitrary Handlebars template as the body
  primary: Em.I18n.t('ok'),
  secondary: Em.I18n.t('common.cancel'),
  third: null,
  autoHeight: true,
  marginBottom: 300,
  disablePrimary: false,
  disableSecondary: false,
  disableThird: false,
  primaryClass: 'btn-success',
  secondaryClass: '',
  onPrimary: function () {
    this.hide();
  },

  onSecondary: function () {
    this.hide();
  },

  onThird: function () {
    this.hide();
  },

  onClose: function () {
    this.hide();
  },

  hide: function () {
    this.destroy();
  },

  showFooter: true,

  /**
   * Hide or show 'X' button for closing popup
   */
  showCloseButton: true,

  didInsertElement: function () {
    if (this.autoHeight) {
      var block = this.$().find('#modal > .modal-body').first();
      if(block.offset()) {
        block.css('max-height', $(window).height() - block.offset().top  - this.marginBottom + $(window).scrollTop()); // fix popup height
      }
    }
    // If popup is opened from another popup it should be displayed above
    var existedPopups = $('.modal-backdrop');
    if (existedPopups) {
      var maxZindex = 1;
      existedPopups.each(function(index, popup) {
        if ($(popup).css('z-index') > maxZindex) {
          maxZindex = $(popup).css('z-index');
      }
      });
      this.$().find('.modal-backdrop').css('z-index', maxZindex * 2);
      this.$().find('.modal').css('z-index', maxZindex * 2 + 1);
    }

    var firstInputElement = this.$('#modal').find(':input').not(':disabled').first();
    this.focusElement(firstInputElement);
  },

  focusElement: function(elem) {
    elem.focus();
  },

  fitHeight: function () {
    var popup = this.$().find('#modal');
    var block = this.$().find('#modal > .modal-body');
    var wh = $(window).height();

    var top = wh * .05;
    popup.css({
      'top': top + 'px',
      'marginTop': 0
    });

    block.css('max-height', $(window).height() - top * 2 - (popup.height() - block.height()));
  }
});

App.ModalPopup.reopenClass({

  show: function (options) {
    var popup = this.create(options);
    popup.appendTo('#wrapper');
    return popup;
  }

});
});

require.register("views/common/modal_popups/alert_popup", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Show alert popup
 *
 * @param {String} header - header of the popup
 * @param {String} body - body of the popup
 * @param {Function} primary - function to call upon clicking the OK button
 * @return {*}
 */
App.showAlertPopup = function (header, body, primary) {
  return App.ModalPopup.show({
    primary: Em.I18n.t('ok'),
    secondary: null,
    header: header,
    body: body,
    onPrimary: function () {
      this.hide();
      if (primary) {
        primary();
      }
    }
  });
};

});

require.register("views/common/modal_popups/cluster_check_popup", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * popup to display requirements that are not met
 * for current action
 * @param data
 * @param header
 * @param title
 * @param alert
 * @returns {*|void}
 */
App.showClusterCheckPopup = function (data, header, title, alert) {
  return App.ModalPopup.show({
    primary: Em.I18n.t('common.dismiss'),
    secondary: false,
    header: header,
    bodyClass: Em.View.extend({
      title: title,
      alert: alert,
      templateName: require('templates/common/modal_popups/cluster_check_dialog'),
      checks: data.items.filterProperty('UpgradeChecks.status', "FAIL")
    })
  });
};

});

require.register("views/common/modal_popups/confirmation_feedback_popup", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Show confirmation popup
 * After sending command watch status of query,
 * and in case of failure provide ability to retry to launch an operation.
 *
 * @param {Function} primary - "OK" button click handler
 * @param {Object} bodyMessage - confirmMsg:{String},
 confirmButton:{String},
 additionalWarningMsg:{String},
 * @param {Function} secondary - "Cancel" button click handler
 * @return {*}
 */
App.showConfirmationFeedBackPopup = function (primary, bodyMessage, secondary) {
  if (!primary) {
    return false;
  }
  return App.ModalPopup.show({
    header: Em.I18n.t('popup.confirmation.commonHeader'),
    bodyClass: Em.View.extend({
      templateName: require('templates/common/modal_popups/confirmation_feedback')
    }),
    query: Em.Object.create({status: "INIT"}),
    primary: function () {
      return bodyMessage? bodyMessage.confirmButton : Em.I18n.t('ok');
    }.property('bodyMessage'),
    onPrimary: function () {
      this.set('query.status', "INIT");
      this.set('disablePrimary', true);
      this.set('disableSecondary', true);
      this.set('statusMessage', Em.I18n.t('popup.confirmationFeedBack.sending'));
      this.hide();
      primary(this.get('query'), this.get('runMmOperation'));
    },
    statusMessage: function () {
      return bodyMessage? bodyMessage.confirmMsg : Em.I18n.t('question.sure');
    }.property('bodyMessage'),
    additionalWarningMsg: function () {
      return bodyMessage? bodyMessage.additionalWarningMsg : null;
    }.property('bodyMessage'),
    putInMaintenance: function () {
      return bodyMessage ? bodyMessage.putInMaintenance : null;
    }.property('bodyMessage'),
    runMmOperation: false,
    turnOnMmMsg: function () {
      return bodyMessage ? bodyMessage.turnOnMmMsg : null;
    }.property('bodyMessage'),
    watchStatus: function() {
      if (this.get('query.status') === "SUCCESS") {
        this.hide();
      } else if(this.get('query.status') === "FAIL") {
        this.set('primaryClass', 'btn-primary');
        this.set('primary', Em.I18n.t('common.retry'));
        this.set('disablePrimary', false);
        this.set('disableSecondary', false);
        this.set('statusMessage', Em.I18n.t('popup.confirmationFeedBack.query.fail'));
      }
    }.observes('query.status'),
    onSecondary: function () {
      this.hide();
      if (secondary) {
        secondary();
      }
    }
  });
};
});

require.register("views/common/modal_popups/confirmation_popup", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Show confirmation popup
 *
 * @param {Function} primary - "OK" button click handler
 * @param {String} body - additional text constant. Will be placed in the popup-body
 * @param {Function} secondary
 * @param {String} header
 * @param {String} primaryText
 * @param {Boolean} isCritical
 * @return {*}
 */
App.showConfirmationPopup = function (primary, body, secondary, header, primaryText, isCritical) {
  if (!primary) {
    return false;
  }
  return App.ModalPopup.show({
    encodeBody: false,
    primary: primaryText || Em.I18n.t('ok'),
    header: header || Em.I18n.t('popup.confirmation.commonHeader'),
    body: body || Em.I18n.t('question.sure'),
    primaryClass: isCritical ? 'btn-danger' : 'btn-success',
    onPrimary: function () {
      this.hide();
      primary();
    },
    onSecondary: function () {
      this.hide();
      if (secondary) {
        secondary();
      }
    },
    onClose:  function () {
      this.hide();
      if (secondary) {
        secondary();
      }
    }
  });
};
});

require.register("views/common/modal_popups/invalid_KDC_popup", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * @param {Object} ajaxOpt - callbek funciton when clicking save
 * @param {Object} message - warning message
 * @return {*}
 */
App.showInvalidKDCPopup = function (ajaxOpt, message) {
  return App.ModalPopup.show({
    primary: Em.I18n.t('common.save'),
    header: Em.I18n.t('popup.invalid.KDC.header'),
    principal: "",
    password: "",
    bodyClass: Em.View.extend({
      warningMsg: message + Em.I18n.t('popup.invalid.KDC.msg'),
      templateName: require('templates/common/modal_popups/invalid_KDC_popup')
    }),
    onClose: function() {
      this.hide();
      if (ajaxOpt.kdcCancelHandler) {
        ajaxOpt.kdcCancelHandler();
      }
    },
    onSecondary: function() {
      this.hide();
      if (ajaxOpt.kdcCancelHandler) {
        ajaxOpt.kdcCancelHandler();
      }
    },
    onPrimary: function () {
      this.hide();
      App.get('router.clusterController').createKerberosAdminSession(this.get('principal'), this.get('password'), ajaxOpt);
    }
  });
};

});

require.register("views/common/modal_popups/prompt_popup", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * Show prompt popup
 *
 * @param {String} text - additional text constant. Will be placed on the top of the input field
 * @param {Function} primary - "OK" button click handler
 * @param {String} defaultValue - additional text constant. Will be default value for input field
 * @param {Function} secondary
 * @return {*}
 */
App.showPromptPopup = function (text, primary, defaultValue, secondary) {
  if (!primary) {
    return false;
  }
  return App.ModalPopup.show({
    header: Em.I18n.t('popup.prompt.commonHeader'),
    bodyClass: Em.View.extend({
      templateName: require('templates/common/modal_popups/prompt_popup'),
      text: text
    }),
    inputValue: defaultValue || '',
    isInvalid: false,
    errorMessage: '',
    onPrimary: function () {
      this.hide();
      primary(this.get('inputValue'));
    },
    onSecondary: function () {
      this.hide();
      if (secondary) {
        secondary();
      }
    }
  });
};

});

require.register("views/common/modal_popups/reload_popup", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.showReloadPopup = function () {
  return App.ModalPopup.show({
    primary: null,
    secondary: null,
    showFooter: false,
    header: this.t('app.reloadPopup.header'),
    body: "<div id='reload_popup' class='alert alert-info'><div class='spinner'><span>" + this.t('app.reloadPopup.text') + "</span></div></div><div><a href='#' onclick='location.reload();'>" + this.t('app.reloadPopup.link') + "</a></div>",
    encodeBody: false
  });
};
});

require.register("views/common/progress_bar_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ProgressBarView = Em.View.extend({
  template: Ember.Handlebars.compile('<div class="bar" {{bindAttr style="view.progressWidth"}}></div>'),
  classNameBindings: ['generalClass', 'barClass'],

  /**
   * @type {number}
   * @default null
   */
  progress: null,

  /**
   * @type {string}
   * @default null
   */
  status: null,

  /**
   * @type {string}
   */
  generalClass: 'progress',

  /**
   * string format: width:<number>%;
   * @type {string}
   */
  progressWidth: function () {
    return "width:" + this.get('progress') + "%;";
  }.property('progress'),

  /**
   * @type {string}
   */
  barClass: function () {
    switch (this.get('status')) {
      case 'FAILED':
        return 'progress-danger';
      case 'ABORTED':
      case 'TIMED_OUT':
        return 'progress-warning';
      case 'COMPLETED':
        return 'progress-success';
      case 'QUEUED':
      case 'PENDING':
      case 'IN_PROGRESS':
        return 'progress-info active progress-striped';
      default:
        return 'progress-info'
    }
  }.property('status')
});

});

require.register("views/common/quick_view_link_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

App.QuickViewLinks = Em.View.extend({

  isLoaded: false,

  loadTags: function () {
    App.ajax.send({
      name: 'config.tags',
      sender: this,
      success: 'loadTagsSuccess',
      error: 'loadTagsError'
    });
  },

  loadTagsSuccess: function (data) {
    this.get('actualTags').clear();
    var tags = [];
    var self = this;
    for (var prop in data.Clusters.desired_configs) {
      tags.push(Em.Object.create({
        siteName: prop,
        tagName: data.Clusters.desired_configs[prop]['tag']
      }));
    }
    this.get('actualTags').pushObjects(tags);
    this.setConfigProperties().done(function (data) {
      self.get('configProperties').pushObjects(data);
      self.getQuickLinksHosts();
    });
  },

  loadTagsError: function() {
    this.getQuickLinksHosts();
  },

  getQuickLinksHosts: function () {
    var masterHosts = App.HostComponent.find().filterProperty('isMaster').mapProperty('hostName').uniq();

    App.ajax.send({
      name: 'hosts.for_quick_links',
      sender: this,
      data: {
        clusterName: App.get('clusterName'),
        masterHosts: masterHosts.join(','),
        urlParams: ',host_components/metrics/hbase/master/IsActiveMaster'
      },
      success: 'setQuickLinksSuccessCallback'
    });
  },

  actualTags: [],

  configProperties: [],

  /**
   * list of files that contains properties for enabling/disabling ssl
   */
  requiredSiteNames: ['hadoop-env','yarn-env','hbase-env','oozie-env','mapred-env','storm-env', 'falcon-env', 'core-site', 'hdfs-site', 'hbase-site', 'oozie-site', 'yarn-site', 'mapred-site', 'storm-site', 'spark-defaults'],
  /**
   * Get public host name by its host name.
   *
   * @method getPublicHostName
   * @param {Object[]} hosts - list of hosts from response
   * @param {String} hostName
   * @return {String}
   **/
  getPublicHostName: function(hosts, hostName) {
    return Em.get(hosts.findProperty('Hosts.host_name', hostName), 'Hosts.public_host_name');
  },

  setConfigProperties: function () {
    this.get('configProperties').clear();
    var requiredSiteNames = this.get('requiredSiteNames');
    var tags = this.get('actualTags').filter(function (tag) {
      return requiredSiteNames.contains(tag.siteName);
    });
    return App.router.get('configurationController').getConfigsByTags(tags);
  },

  ambariProperties: function () {
    return App.router.get('clusterController.ambariProperties');
  },
  /**
   * Updated quick links. Here we put correct hostname to url
   */
  quickLinks: [],

  didInsertElement: function () {
    this.setQuickLinks();
  },

  findComponentHost: function (components, componentName) {
    var component = components.find(function (item) {
      return item.host_components.someProperty('HostRoles.component_name', componentName);
    });
    return component && component.Hosts.public_host_name;
  },

  setQuickLinks: function () {
    this.loadTags();
  }.observes('App.currentStackVersionNumber', 'App.singleNodeInstall'),

  setQuickLinksSuccessCallback: function (response) {
    var self = this;
    var quickLinks = [];
    var hosts = this.setHost(response, this.get('content.serviceName'));
    if (!hosts || !this.get('content.quickLinks')) {
      quickLinks = [{
          label: this.t('quick.links.error.label'),
          url: 'javascript:alert("' + this.t('contact.administrator') + '");return false;'
      }];
      this.set('quickLinks', quickLinks);
      this.set('isLoaded', true);
    } else if (hosts.length == 1) {

      quickLinks = this.get('content.quickLinks').map(function (item) {
        var protocol = self.setProtocol(item.get('service_id'), self.get('configProperties'), self.ambariProperties());
        if (item.get('template')) {
          var port = item.get('http_config') && self.setPort(item, protocol);
          item.set('url', item.get('template').fmt(protocol, hosts[0], port));
        }
        return item;
      });
      this.set('quickLinks', quickLinks);
      this.set('isLoaded', true);
    } else {
      // multiple hbase masters or HDFS HA enabled
      var quickLinksArray = [];
      hosts.forEach(function(host) {
        var quickLinks = [];
        self.get('content.quickLinks').forEach(function (item) {
          var newItem = {};
          var protocol = self.setProtocol(item.get('service_id'), self.get('configProperties'), self.ambariProperties());
          if (item.get('template')) {
            var port = item.get('http_config') && self.setPort(item, protocol);
            newItem.url = item.get('template').fmt(protocol, host.publicHostName, port);
            newItem.label = item.get('label');
          }
          quickLinks.push(newItem);
        });
        if (host.status) {
          quickLinks.set('publicHostNameLabel', Em.I18n.t('quick.links.publicHostName').format(host.publicHostName, host.status));
        } else {
          quickLinks.set('publicHostNameLabel', host.publicHostName);
        }
        quickLinksArray.push(quickLinks);
      }, this);
      this.set('quickLinksArray', quickLinksArray);
      this.set('isLoaded', true);
    }
  },

  /**
   * sets public host names for required masters of current service
   * @param {String} serviceName - selected serviceName
   * @param {JSON} response
   * @returns {Array} containing hostName(s)
   * @method setHost
   */
  setHost: function(response, serviceName) {
    if (App.get('singleNodeInstall')) {
      return [App.get('singleNodeAlias')];
    }
    var hosts = [];
    switch (serviceName) {
      case "HDFS":
        if (this.get('content.snameNode')) {
          // not HA
          hosts[0] = this.findComponentHost(response.items, 'NAMENODE');
        } else {
          // HA enabled, need both two namenodes hosts
          this.get('content.hostComponents').filterProperty('componentName', 'NAMENODE').forEach(function (component) {
            hosts.push({'publicHostName': response.items.findProperty('Hosts.host_name', component.get('hostName')).Hosts.public_host_name});
          });
          // assign each namenode status label
          if (this.get('content.activeNameNode')) {
            hosts.findProperty('publicHostName', this.getPublicHostName(response.items, this.get('content.activeNameNode.hostName'))).status = Em.I18n.t('quick.links.label.active');
          }
          if (this.get('content.standbyNameNode')) {
            hosts.findProperty('publicHostName', this.getPublicHostName(response.items, this.get('content.standbyNameNode.hostName'))).status = Em.I18n.t('quick.links.label.standby');
          }
          if (this.get('content.standbyNameNode2')) {
            hosts.findProperty('publicHostName', this.getPublicHostName(response.items, this.get('content.standbyNameNode2.hostName'))).status = Em.I18n.t('quick.links.label.standby');
          }
        }
        break;
      case "HBASE":
        var masterComponents = response.items.filter(function (item) {
          return item.host_components.someProperty('HostRoles.component_name', 'HBASE_MASTER');
        });
        var activeMaster, standbyMasters, otherMasters;
        activeMaster = masterComponents.filter(function (item) {
          return item.host_components.someProperty('metrics.hbase.master.IsActiveMaster', 'true');
        });
        standbyMasters = masterComponents.filter(function (item) {
          return item.host_components.someProperty('metrics.hbase.master.IsActiveMaster', 'false');
        });
        otherMasters = masterComponents.filter(function (item) {
          return !(item.host_components.someProperty('metrics.hbase.master.IsActiveMaster', 'true') || item.host_components.someProperty('metrics.hbase.master.IsActiveMaster', 'false'));
        });
        if (masterComponents.length > 1) {
          // need all hbase_masters hosts in quick links
          if (activeMaster) {
            activeMaster.forEach(function (item) {
              hosts.push({'publicHostName': item.Hosts.public_host_name, 'status': Em.I18n.t('quick.links.label.active')});
            });
          }
          if (standbyMasters) {
            standbyMasters.forEach(function (item) {
              hosts.push({'publicHostName': item.Hosts.public_host_name, 'status': Em.I18n.t('quick.links.label.standby')});
            });
          }
          if (otherMasters) {
            otherMasters.forEach(function (item) {
              hosts.push({'publicHostName': item.Hosts.public_host_name});
            });
          }
        } else {
          hosts[0] = masterComponents[0].Hosts.public_host_name;
        }
        break;
      case "YARN":
        if (App.get('isRMHaEnabled')) {
          this.get('content.hostComponents').filterProperty('componentName', 'RESOURCEMANAGER').forEach(function (component) {
            var newHost = {'publicHostName': response.items.findProperty('Hosts.host_name', component.get('hostName')).Hosts.public_host_name};
            var status = '';
            switch (component.get('haStatus')) {
              case 'ACTIVE':
                status = Em.I18n.t('quick.links.label.active');
                break;
              case 'STANDBY':
                status = Em.I18n.t('quick.links.label.standby');
                break;
            }
            if (status) {
              newHost.status = status;
            }
            hosts.push(newHost);
          }, this);
        } else {
          hosts[0] = this.findComponentHost(response.items, 'RESOURCEMANAGER');
        }
        break;
      case "STORM":
        hosts[0] = this.findComponentHost(response.items, "STORM_UI_SERVER");
        break;
      default:
        var service = App.StackService.find().findProperty('serviceName', serviceName);
        if (service && service.get('hasMaster')) {
          hosts[0] = this.findComponentHost(response.items, this.get('content.hostComponents') && this.get('content.hostComponents').findProperty('isMaster', true).get('componentName'));
        }
        break;
    }
    return hosts;
  },

  /**
   * services that supports security. this array is used to find out protocol.
   * becides GANGLIA, YARN, MAPREDUCE2. These properties use
   * their properties to know protocol
   */
  servicesSupportsHttps: ["HDFS", "HBASE"],

  /**
   * setProtocol - if cluster is secure for some services (GANGLIA, MAPREDUCE2, YARN and servicesSupportsHttps)
   * protocol becomes "https" otherwise "http" (by default)
   * @param {String} service_id - service name
   * @param {Object} configProperties
   * @param {Object} ambariProperties
   * @returns {string} "https" or "http" only!
   * @method setProtocol
   */
  setProtocol: function (service_id, configProperties, ambariProperties) {
    var hadoopSslEnabled = false;
    if (configProperties && configProperties.length > 0) {
      var hdfsSite = configProperties.findProperty('type', 'hdfs-site');
      hadoopSslEnabled = (hdfsSite && Em.get(hdfsSite, 'properties') && hdfsSite.properties['dfs.http.policy'] === 'HTTPS_ONLY');
    }
    switch (service_id) {
      case "GANGLIA":
        return (ambariProperties && ambariProperties['ganglia.https'] == "true") ? "https" : "http";
        break;
      case "YARN":
        var yarnProperties = configProperties && configProperties.findProperty('type', 'yarn-site');
        if (yarnProperties && yarnProperties.properties) {
          if (yarnProperties.properties['yarn.http.policy'] === 'HTTPS_ONLY') {
            return "https";
          } else if (yarnProperties.properties['yarn.http.policy'] === 'HTTP_ONLY') {
            return "http";
          }
        }
        return hadoopSslEnabled ? "https" : "http";
        break;
      case "MAPREDUCE2":
        var mapred2Properties = configProperties && configProperties.findProperty('type', 'mapred-site');
        if (mapred2Properties && mapred2Properties.properties) {
          if (mapred2Properties.properties['mapreduce.jobhistory.http.policy'] === 'HTTPS_ONLY') {
            return "https";
          } else if (mapred2Properties.properties['mapreduce.jobhistory.http.policy'] === 'HTTP_ONLY') {
            return "http";
          }
        }
        return hadoopSslEnabled ? "https" : "http";
        break;
      default:
        return this.get('servicesSupportsHttps').contains(service_id) && hadoopSslEnabled ? "https" : "http";
    }
  },

  /**
   * sets the port of quick link
   * @param item
   * @param protocol
   * @returns {*}
   * @method setPort
   */
  setPort: function (item, protocol) {
    var configProperties = this.get('configProperties');
    var config = item.get('http_config');
    var defaultPort = item.get('default_http_port');
    if (protocol === 'https' && item.get('https_config')) {
      config = item.get('https_config');
      if (item.get('default_https_port')) {
        defaultPort = item.get('default_https_port');
      }
    }
    var site = configProperties.findProperty('type', item.get('site'));
    var propertyValue = site && site.properties[config];
    if (!propertyValue) {
      return defaultPort;
    }

    var re = new RegExp(item.get('regex'));

    var portValue = propertyValue.match(re);
    return  portValue[1];
  },

  linkTarget: function () {
    switch (this.get('content.serviceName').toLowerCase()) {
      case "hdfs":
      case "yarn":
      case "mapreduce2":
      case "hbase":
      case "oozie":
      case "ganglia":
      case "storm":
      case "falcon":
        return "_blank";
        break;
      default:
        return "";
        break;
    }
  }.property('service')

});

});

require.register("views/common/rolling_restart_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
var numberUtils = require('utils/number_utils');

/**
 * View content of the rolling restart dialog.
 *
 * Callers provide the context in which this dialog is invoked.
 */
App.RollingRestartView = Em.View.extend({
  templateName : require('templates/common/rolling_restart_view'),

  /**
   * Component name for components that should be restarted
   * @type {String}
   */
  hostComponentName : null,

  /**
   * Service name for components that should be restarted
   * @type {String}
   */
  serviceName : null,

  /**
   * If service is in Maintenance Mode
   * @type {bool}
   */
  isServiceInMM: false,

  /**
   * If true service will be put in Maintenance mode before rolling restart
   * @type {bool}
   */
  turnOnMm: false,


  /**
   * Restart only components with <code>staleConfigs</code>
   * @type {bool}
   */
  staleConfigsOnly : false,

  /**
   * We should do rolling restart for components if we run
   * restart for service and service is in Maintenance mode
   * @type {bool}
   */
  skipMaintenance: false,

  /**
   * Count of host components in one batch
   * @type {Number}
   */
  batchSize : -1,

  /**
   * Delay between batches
   * @type {Number}
   */
  interBatchWaitTimeSeconds : -1,

  /**
   * @type {Number}
   */
  tolerateSize : -1,

  /**
   * Shows if all needed info is loaded and initialized
   * @type {Bool}
   */
  isLoaded: false,
  /**
   * List of error in batch-request properties
   * @type {Array}
   */
  errors : [],
  /**
   * Set initial values for batch-request properties
   */
  initialize : function() {
    if (this.get('batchSize') == -1 && this.get('interBatchWaitTimeSeconds') == -1 && this.get('tolerateSize') == -1) {
      var restartCount = this.get('restartHostComponents.length');
      var batchSize = 1;
      if (restartCount > 10) {
        batchSize = Math.ceil(restartCount / 10);
      }
      var tolerateCount = batchSize;
      this.set('batchSize', batchSize);
      this.set('tolerateSize', tolerateCount);
      this.set('interBatchWaitTimeSeconds', 120);
      this.set('isLoaded', true);
    }
  },

  /**
   * Validate batch-request properties
   * List of errors is saved to <code>errors</code>
   */
  validate : function() {
    var displayName = this.get('hostComponentDisplayName');
    var totalCount = this.get('restartHostComponents.length');
    var bs = this.get('batchSize');
    var ts = this.get('tolerateSize');
    var wait = this.get('interBatchWaitTimeSeconds');
    var errors = [];
    if (totalCount < 1) {
      errors.push(Em.I18n.t('rollingrestart.dialog.msg.noRestartHosts').format(displayName));
    } else {
      var bsError = numberUtils.validateInteger(bs, 1, totalCount);
      var tsError = numberUtils.validateInteger(ts, 0, totalCount);
      if (bsError != null) {
        errors.push(Em.I18n.t('rollingrestart.dialog.err.invalid.batchsize').format(bsError));
      }
      if (tsError != null) {
        errors.push(Em.I18n.t('rollingrestart.dialog.err.invalid.toleratesize').format(tsError));
      }
    }
    var waitError = numberUtils.validateInteger(wait, 0, NaN);
    if (waitError != null) {
      errors.push(Em.I18n.t('rollingrestart.dialog.err.invalid.waitTime').format(waitError));
    }
    this.set('errors', errors);
  }.observes('batchSize', 'interBatchWaitTimeSeconds', 'tolerateSize', 'restartHostComponents', 'hostComponentDisplayName'),

  /**
   * Formatted <code>hostComponentName</code>
   * @type {String}
   */
  hostComponentDisplayName : function() {
    return App.format.role(this.get('hostComponentName'));
  }.property('hostComponentName'),

  /**
   * List of all host components
   * @type {Array}
   */
  allHostComponents : [],

  /**
   * List of host components without components in out-of-service state
   * @type {Array}
   */
  nonMaintainanceHostComponents : function() {
    return this.get('allHostComponents').filterProperty('passiveState', 'OFF')
  }.property('allHostComponents', 'allHostComponents.@each.passiveState'),

  /**
   * List of host components with host in Maintenance mode
   * @type {Array}
   */
  componentsWithMaintenanceHost: function() {
    return this.get('allHostComponents').filterProperty('hostPassiveState', 'ON');
  }.property('allHostComponents', 'allHostComponents.@each.hostPassiveState'),
  /**
   * List of host components without components in out-of-service state
   * If <code>staleConfigsOnly</code> is true, components with <code>staleConfigs</code> = false are also filtered
   * @type {Array}
   */
  restartHostComponents: function () {
    var hostComponents = (this.get('skipMaintenance')) ? this.get('allHostComponents') : this.get('nonMaintainanceHostComponents');

    if (this.get('staleConfigsOnly')) {
      hostComponents = hostComponents.filterProperty('staleConfigs');
    }
    return hostComponents;
  }.property('nonMaintainanceHostComponents', 'staleConfigsOnly'),
  /**
   * @type {String}
   */
  suggestTurnOnMaintenanceMsg : function() {
    if (!this.get('isServiceInMM')) {
      return Em.I18n.t('rollingrestart.dialog.msg.serviceNotInMM').format(this.get('serviceName'));
    } else {
      return null;
    }
  }.property('isServiceInMM'),

  /**
   * @type {String}
   */
  restartMessage : function() {
    return Em.I18n.t('rollingrestart.dialog.msg.restart').format(this.get('hostComponentDisplayName'))
  }.property('hostComponentDisplayName'),

  /**
   * @type {String}
   */
  maintainanceMessage : function() {
    var count = this.get('componentsWithMaintenanceHost.length');
    if (count > 0) {
      var name = this.get('hostComponentDisplayName');
      if (count > 1) {
        return Em.I18n.t('rollingrestart.dialog.msg.maintainance.plural').format(count, name)
      }
      return Em.I18n.t('rollingrestart.dialog.msg.maintainance').format(count, name)
    }
    return null;
  }.property('componentsWithMaintenanceHost', 'hostComponentDisplayName'),

  /**
   * @type {String}
   */
  batchSizeMessage : function() {
    return Em.I18n.t('rollingrestart.dialog.msg.componentsAtATime').format(this.get('hostComponentDisplayName'));
  }.property('hostComponentDisplayName'),

  /**
   * @type {String}
   */
  staleConfigsOnlyMessage : function() {
    return Em.I18n.t('rollingrestart.dialog.msg.staleConfigsOnly').format(this.get('hostComponentDisplayName'));
  }.property('hostComponentDisplayName')
});

});

require.register("views/common/select_custom_date_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.JobsCustomDatesSelectView = Em.View.extend({

  name: 'jobsCustomDatesSelectView',

  templateName: require('templates/common/custom_date_popup'),

  middayPeriodOptions: [Em.I18n.t('jobs.table.custom.date.am'), Em.I18n.t('jobs.table.custom.date.pm')],

  hourOptions: ['01', '02', '03', '04', '05', '06', '07', '08', '09', '10', '11', '12'],

  minuteOptions: ['00', '05', '10', '15', '20', '25', '30', '35', '40', '45', '50', '55'],

  didInsertElement: function () {
    $('.datepicker').datepicker({
      format: 'mm/dd/yyyy'
    }).on('changeDate', function() {
      $(this).datepicker('hide');
    });
  }
});

});

require.register("views/common/sort_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var misc = require('utils/misc');
var stringUtils = require('utils/string_utils');
var App = require('app');

/**
 * Wrapper View for all sort components. Layout template and common actions are located inside of it.
 * Logic specific for sort fields
 * located in inner view - <code>fieldView</code>.
 *
 * @type {*}
 */
var wrapperView = Em.View.extend({
  tagName: 'tr',

  classNames: ['sort-wrapper'],

  willInsertElement: function () {
    if (this.get('parentView.tableFilteringComplete')) {
      this.get('parentView').set('filteringComplete', true);
    }
  },

  /**
   * Load sort statuses from local storage
   * Works only after finish filtering in the parent View
   */
  loadSortStatuses: function () {
    var statuses = App.db.getSortingStatuses(this.get('controller.name'));
    if (!this.get('parentView.filteringComplete')) return;
    if (statuses) {
      var childViews = this.get('childViews');
      var self = this;
      statuses.forEach(function (st) {
        if (st.status != 'sorting') {
          var sortOrder = false;
          if (st.status == 'sorting_desc') {
            sortOrder = true;
          }
          self.sort(childViews.findProperty('name', st.name), sortOrder);
          childViews.findProperty('name', st.name).set('status', (sortOrder) ? 'sorting_desc' : 'sorting_asc');
          self.get('controller').set('sortingColumn', childViews.findProperty('name', st.name));
        }
        else {
          childViews.findProperty('name', st.name).set('status', st.status);
        }
      });
    }
  }.observes('parentView.filteringComplete'),

  /**
   * Save sort statuses to local storage
   * Works only after finish filtering in the parent View
   */
  saveSortStatuses: function () {
    if (!this.get('parentView.filteringComplete')) return;

    var statuses = [];
    this.get('childViews').forEach(function (childView) {
      statuses.push({
        name: childView.get('name'),
        status: childView.get('status')
      });
    });
    App.db.setSortingStatuses(this.get('controller.name'), statuses);
  },

  /**
   * sort content by property
   * @param property {object}
   * @param order {Boolean} true - DESC, false - ASC
   * @param returnSorted {Boolean}
   */
  sort: function (property, order, returnSorted) {
    var content = this.get('content').toArray();
    var sortFunc = this.getSortFunc(property, order);
    var status = order ? 'sorting_desc' : 'sorting_asc';

    this.resetSort();
    this.get('childViews').findProperty('name', property.get('name')).set('status', status);
    this.saveSortStatuses(property, order);
    content.sort(sortFunc);

    if (!!returnSorted) {
      return content;
    } else {
      this.set('content', content);
    }
  },

  isSorting: false,

  onContentChange: function () {
    if (!this.get('isSorting') && this.get('content.length')) {
      this.get('childViews').forEach(function (view) {
        if (view.status !== 'sorting') {
          var status = view.get('status');
          this.set('isSorting', true);
          this.sort(view, status == 'sorting_desc');
          this.set('isSorting', false);
          view.set('status', status);
        }
      }, this);
    }
  }.observes('content.length'),

  /**
   * reset all sorts fields
   */
  resetSort: function () {
    this.get('childViews').setEach('status', 'sorting');
  },
  /**
   * determines sort function depending on the type of sort field
   * @param property
   * @param order
   * @return {*}
   */
  getSortFunc: function (property, order) {
    var func;
    switch (property.get('type')) {
      case 'ip':
        func = function (a, b) {
          var a = misc.ipToInt(a.get(property.get('name')));
          var b = misc.ipToInt(b.get(property.get('name')));
          if (order) {
            return a - b;
          } else {
            return b - a;
          }
        };
        break;
      case 'number':
        func = function (a, b) {
          var a_p = a.get(property.get('name'));
          var b_p = b.get(property.get('name'));
          a_p = Em.isNone(a_p) ? -Infinity : parseFloat(a_p);
          b_p = Em.isNone(b_p) ? -Infinity : parseFloat(b_p);
          if (order) {
            return a_p - b_p;
          } else {
            return b_p - a_p
          }
        };
        break;
      case 'version':
        func = function (a, b) {
          var res = stringUtils.compareVersions(a.get(property.get('name')), b.get(property.get('name')));
          if (order) {
            return -res;
          } else {
            return res;
          }
        };
        break;
      case 'alert_status':
        func = App.AlertDefinition.getSortDefinitionsByStatus(order);
        break;
      default:
        func = function (a, b) {
          var a_p = a.get(property.get('name'));
          var b_p = b.get(property.get('name'));
          a_p = Em.isNone(a_p) ? '' : '' + a_p;
          b_p = Em.isNone(b_p) ? '' : '' + b_p;
          return order ? b_p.localeCompare(a_p) : a_p.localeCompare(b_p);
        };
    }
    return func;
  }
});

/**
 * view that carry on sorting on server-side via <code>refresh()</code> in parentView
 * @type {*}
 */
var serverWrapperView = Em.View.extend({
  tagName: 'tr',

  classNames: ['sort-wrapper'],

  willInsertElement: function () {
    this.loadSortStatuses();
  },

  /**
   * Initialize and save sorting statuses: hostName sorting_asc
   */
  loadSortStatuses: function () {
    var statuses = [];
    var childViews = this.get('childViews');
    childViews.forEach(function (childView) {
      var sortStatus = (childView.get('name') == 'hostName' && childView.get('status') == 'sorting') ? 'sorting_asc' : childView.get('status');
      statuses.push({
        name: childView.get('name'),
        status: sortStatus
      });
      childView.set('status', sortStatus);
    });
    App.db.setSortingStatuses(this.get('controller.name'), statuses);
    this.get('controller').set('sortingColumn', childViews.findProperty('name', 'hostName'));
  },

  /**
   * Save sort statuses to local storage
   * Works only after finish filtering in the parent View
   */
  saveSortStatuses: function () {
    var statuses = [];
    this.get('childViews').forEach(function (childView) {
      statuses.push({
        name: childView.get('name'),
        status: childView.get('status')
      });
    });
    App.db.setSortingStatuses(this.get('controller.name'), statuses);
  },

  /**
   * sort content by property
   * @param property {object}
   * @param order {Boolean} true - DESC, false - ASC
   */
  sort: function (property, order) {
    var status = order ? 'sorting_desc' : 'sorting_asc';

    this.resetSort();
    this.get('childViews').findProperty('name', property.get('name')).set('status', status);
    this.saveSortStatuses();
    this.get('parentView').refresh();
  },

  /**
   * reset all sorts fields
   */
  resetSort: function () {
    this.get('childViews').setEach('status', 'sorting');
  }
});
/**
 * particular view that contain sort field properties:
 * name - name of property in content table
 * type(optional) - specific type to sort
 * displayName - label to display
 * @type {*}
 */
var fieldView = Em.View.extend({
  template: Em.Handlebars.compile('<span {{bindAttr class="view.status :column-name"}}>{{view.displayName}}</span>'),
  classNameBindings: ['viewNameClass'],
  tagName: 'th',
  name: null,
  displayName: null,
  status: 'sorting',
  viewNameClass: function () {
    return 'sort-view-' + this.get('column');
  }.property(),
  type: null,
  column: 0,
  /**
   * callback that run sorting and define order of sorting
   * @param event
   */
  click: function (event) {
    this.get('parentView').sort(this, (this.get('status') !== 'sorting_desc'));
    this.get('controller').set('sortingColumn', this);
  }
});

/**
 * Result object, which will be accessible outside
 * @type {Object}
 */
module.exports = {
  serverWrapperView: serverWrapperView,
  wrapperView: wrapperView,
  fieldView: fieldView
};
});

require.register("views/common/table_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var filters = require('views/common/filter_view');

App.TableView = Em.View.extend(App.UserPref, {

  init: function() {
    this.set('filterConditions', []);
    this._super();
  },

  /**
   * Defines to show pagination or show all records
   * @type {Boolean}
   */
  pagination: true,

  /**
   * Shows if all data is loaded and filtered
   * @type {Boolean}
   */
  filteringComplete: false,

  /**
   * intermediary for filteringComplete
   * @type {Boolean}
   */
  tableFilteringComplete: false,

  /**
   * The number of rows to show on every page
   * The value should be a number converted into string type in order to support select element API
   * Example: "10", "25"
   * @type {String}
   */
  displayLength: null,

  /**
   * default value of display length
   * The value should be a number converted into string type in order to support select element API
   * Example: "10", "25"
   */
  defaultDisplayLength: "10",

  /**
   * number of items in table after applying filters
   */
  filteredCount: function () {
    return this.get('filteredContent.length');
  }.property('filteredContent.length'),

  /**
   * total number of items in table before applying filters
   */
  totalCount: function () {
    return this.get('content.length');
  }.property('content.length'),

  /**
   * Do filtering, using saved in the local storage filter conditions
   */
  willInsertElement: function () {
    var self = this;
    var name = this.get('controller.name');
    if (!this.get('displayLength')) {
      if (App.db.getDisplayLength(name)) {
        Em.run.next(function () {
          self.set('displayLength', App.db.getDisplayLength(name));
          self.initFilters();
        });
      } else {
        if (!$.mocho) {
          this.getUserPref(this.displayLengthKey()).complete(function () {
            self.initFilters();
          });
        }
      }
    }
  },

  /**
   * initialize filters
   * restore values from local DB
   * or clear filters in case there is no filters to restore
   */
  initFilters: function () {
    var name = this.get('controller.name');
    var self = this;
    var filterConditions = App.db.getFilterConditions(name);
    if (filterConditions) {
      this.set('filterConditions', filterConditions);

      var childViews = this.get('childViews');

      filterConditions.forEach(function (condition, index, filteredConditions) {
        var view = !Em.isNone(condition.iColumn) && childViews.findProperty('column', condition.iColumn);
        if (view) {
          if (view.get('emptyValue')) {
            view.clearFilter();
            self.saveFilterConditions(condition.iColumn, view.get('appliedEmptyValue'), condition.type, false);
          } else {
            view.setValue(condition.value);
          }
          Em.run.next(function () {
            view.showClearFilter();
            // check if it is the last iteration
            if (filteredConditions.length - index - 1 === 0) {
              self.set('tableFilteringComplete', true);
            }
          });
        }
      });
    } else {
      this.set('tableFilteringComplete', true);
    }
  },

  /**
   * Persist-key of current table displayLength property
   * @param {String} loginName current user login name
   * @returns {String}
   */
  displayLengthKey: function (loginName) {
    if (App.get('testMode')) {
      return 'pagination_displayLength';
    }
    loginName = loginName ? loginName : App.router.get('loginName');
    return this.get('controller.name') + '-pagination-displayLength-' + loginName;
  },

  /**
   * Set received from server value to <code>displayLengthOnLoad</code>
   * @param {Number} response
   * @param {Object} request
   * @param {Object} data
   * @returns {*}
   */
  getUserPrefSuccessCallback: function (response, request, data) {
    console.log('Got DisplayLength value from server with key ' + data.key + '. Value is: ' + response);
    this.set('displayLength', response);
    return response;
  },

  /**
   * Set default value to <code>displayLength</code> (and send it on server) if value wasn't found on server
   * @returns {Number}
   */
  getUserPrefErrorCallback: function () {
    // this user is first time login
    console.log('Persist did NOT find the key');
    var displayLengthDefault = this.get('defaultDisplayLength');
    this.set('displayLength', displayLengthDefault);
    if (App.isAccessible('upgrade_ADMIN')) {
      this.saveDisplayLength();
    }
    this.filter();
    return displayLengthDefault;
  },

  /**
   * Return pagination information displayed on the page
   * @type {String}
   */
  paginationInfo: function () {
    return this.t('tableView.filters.paginationInfo').format(this.get('startIndex'), this.get('endIndex'), this.get('filteredCount'));
  }.property('filteredCount', 'endIndex'),

  paginationLeft: Ember.View.extend({
    tagName: 'a',
    template: Ember.Handlebars.compile('<i class="icon-arrow-left"></i>'),
    classNameBindings: ['class'],
    class: function () {
      if (this.get("parentView.startIndex") > 1) {
        return "paginate_previous";
      }
      return "paginate_disabled_previous";
    }.property("parentView.startIndex", 'parentView.filteredCount'),

    click: function () {
      if (this.get('class') === "paginate_previous") {
        this.get('parentView').previousPage();
      }
    }
  }),

  paginationRight: Ember.View.extend({
    tagName: 'a',
    template: Ember.Handlebars.compile('<i class="icon-arrow-right"></i>'),
    classNameBindings: ['class'],
    class: function () {
      if ((this.get("parentView.endIndex")) < this.get("parentView.filteredCount")) {
        return "paginate_next";
      }
      return "paginate_disabled_next";
    }.property("parentView.endIndex", 'parentView.filteredCount'),

    click: function () {
      if (this.get('class') === "paginate_next") {
        this.get('parentView').nextPage();
      }
    }
  }),

  paginationFirst: Ember.View.extend({
    tagName: 'a',
    template: Ember.Handlebars.compile('<i class="icon-step-backward"></i>'),
    classNameBindings: ['class'],
    class: function () {
      if ((this.get("parentView.endIndex")) > parseInt(this.get("parentView.displayLength"))) {
        return "paginate_previous";
      }
      return "paginate_disabled_previous";
    }.property("parentView.endIndex", 'parentView.filteredCount'),

    click: function () {
      if (this.get('class') === "paginate_previous") {
        this.get('parentView').firstPage();
      }
    }
  }),

  paginationLast: Ember.View.extend({
    tagName: 'a',
    template: Ember.Handlebars.compile('<i class="icon-step-forward"></i>'),
    classNameBindings: ['class'],
    class: function () {
      if (this.get("parentView.endIndex") !== this.get("parentView.filteredCount")) {
        return "paginate_next";
      }
      return "paginate_disabled_next";
    }.property("parentView.endIndex", 'parentView.filteredCount'),

    click: function () {
      if (this.get('class') === "paginate_next") {
        this.get('parentView').lastPage();
      }
    }
  }),

  /**
   * Select View with list of "rows-per-page" options
   * @type {Ember.View}
   */
  rowsPerPageSelectView: Em.Select.extend({
    content: ['10', '25', '50', '100'],
    change: function () {
      this.get('parentView').saveDisplayLength();
    }
  }),

  /**
   * Start index for displayed content on the page
   */
  startIndex: 1,

  /**
   * Calculate end index for displayed content on the page
   */
  endIndex: function () {
    if (this.get('pagination') && this.get('displayLength')) {
      return Math.min(this.get('filteredCount'), this.get('startIndex') + parseInt(this.get('displayLength')) - 1);
    } else {
      return this.get('filteredCount') || 0;
    }
  }.property('startIndex', 'displayLength', 'filteredCount'),

  /**
   * Onclick handler for previous page button on the page
   */
  previousPage: function () {
    var result = this.get('startIndex') - parseInt(this.get('displayLength'));
    this.set('startIndex', (result < 2) ? 1 : result);
  },

  /**
   * Onclick handler for next page button on the page
   */
  nextPage: function () {
    var result = this.get('startIndex') + parseInt(this.get('displayLength'));
    if (result - 1 < this.get('filteredCount')) {
      this.set('startIndex', result);
    }
  },
  /**
   * Onclick handler for first page button on the page
   */
  firstPage: function () {
    this.set('startIndex', 1);
  },
  /**
   * Onclick handler for last page button on the page
   */
  lastPage: function () {
    var pagesCount = this.get('filteredCount') / parseInt(this.get('displayLength'));
    var startIndex = (this.get('filteredCount') % parseInt(this.get('displayLength')) === 0) ?
      (pagesCount - 1) * parseInt(this.get('displayLength')) :
      Math.floor(pagesCount) * parseInt(this.get('displayLength'));
    this.set('startIndex', ++startIndex);
  },

  /**
   * Calculates default value for startIndex property after applying filter or changing displayLength
   */
  updatePaging: function (controller, property) {
    var displayLength = this.get('displayLength');
    var filteredContentLength = this.get('filteredCount');
    if (property == 'displayLength') {
      this.set('startIndex', Math.min(1, filteredContentLength));
    } else if (!filteredContentLength) {
      this.set('startIndex', 0);
    } else if (this.get('startIndex') > filteredContentLength) {
      this.set('startIndex', Math.floor((filteredContentLength - 1) / displayLength) * displayLength + 1);
    } else if (!this.get('startIndex')) {
      this.set('startIndex', 1);
    }
  }.observes('displayLength', 'filteredCount'),

  /**
   * Apply each filter to each row
   *
   * @param {Number} iColumn number of column by which filter
   * @param {Object} value
   * @param {String} type
   */
  updateFilter: function (iColumn, value, type) {
    this.saveFilterConditions(iColumn, value, type, false);
    this.filtersUsedCalc();
    this.filter();
  },

  /**
   * save filter conditions to local storage
   * @param iColumn {Number}
   * @param value {String|Array}
   * @param type {String}
   * @param skipFilter {Boolean}
   */
  saveFilterConditions: function(iColumn, value, type, skipFilter) {
    var filterCondition = this.get('filterConditions').findProperty('iColumn', iColumn);

    if (filterCondition) {
      filterCondition.value = value;
      filterCondition.skipFilter = skipFilter;
    } else {
      filterCondition = {
        skipFilter: skipFilter,
        iColumn: iColumn,
        value: value,
        type: type
      };
      this.get('filterConditions').push(filterCondition);
    }
    App.db.setFilterConditions(this.get('controller.name'), this.get('filterConditions'));
  },

  saveDisplayLength: function() {
    var self = this;
    Em.run.next(function() {
      App.db.setDisplayLength(self.get('controller.name'), self.get('displayLength'));
      if (!App.get('testMode')) {
        if (App.isAccessible('upgrade_ADMIN')) {
          self.postUserPref(self.displayLengthKey(), self.get('displayLength'));
        }
      }
    });
  },

  clearFilterCondition: function() {
    App.db.setFilterConditions(this.get('controller.name'), null);
  },

  /**
   * Contain filter conditions for each column
   * @type {Array}
   */
  filterConditions: [],

  /**
   * Contains content after implementing filters
   * @type {Array}
   */
  filteredContent: [],

  /**
   * Determine if <code>filteredContent</code> is empty or not
   * @type {Boolean}
   */
  hasFilteredItems: function() {
    return !!this.get('filteredCount');
  }.property('filteredCount'),

  /**
   * Contains content to show on the current page of data page view
   * @type {Array}
   */
  pageContent: function () {
    return this.get('filteredContent').slice(this.get('startIndex') - 1, this.get('endIndex'));
  }.property('filteredCount', 'startIndex', 'endIndex'),

  /**
   * Filter table by filterConditions
   */
  filter: function () {
    var content = this.get('content');
    var filterConditions = this.get('filterConditions').filterProperty('value');
    var result;
    var assoc = this.get('colPropAssoc');
    if (filterConditions.length) {
      result = content.filter(function (item) {
        var match = true;
        filterConditions.forEach(function (condition) {
          var filterFunc = filters.getFilterByType(condition.type, false);
          if (match) {
            match = filterFunc(item.get(assoc[condition.iColumn]), condition.value);
          }
        });
        return match;
      });
      this.set('filteredContent', result);
    } else {
      this.set('filteredContent', content.toArray());
    }
  }.observes('content.length'),

  /**
   * sort content by active sort column
   */
  sortContent: function() {
    var activeSort = App.db.getSortingStatuses(this.get('controller.name')).find(function (sort) {
      return (sort.status === 'sorting_asc' || sort.status === 'sorting_desc');
    });
    var sortIndexes = {
      'sorting_asc': 1,
      'sorting_desc': -1
    };

    this.get('content').sort(function (a, b) {
      if (a.get(activeSort.name) > b.get(activeSort.name)) return sortIndexes[activeSort.status];
      if (a.get(activeSort.name) < b.get(activeSort.name)) return -(sortIndexes[activeSort.status]);
      return 0;
    });
    this.filter();
  },

  /**
   * Does any filter is used on the page
   * @type {Boolean}
   */
  filtersUsed: false,

  /**
   * Determine if some filters are used on the page
   * Set <code>filtersUsed</code> value
   */
  filtersUsedCalc: function() {
    var filterConditions = this.get('filterConditions');
    if (!filterConditions.length) {
      this.set('filtersUsed', false);
      return;
    }
    var filtersUsed = false;
    filterConditions.forEach(function(filterCondition) {
      if (filterCondition.value.toString() !== '') {
        filtersUsed = true;
      }
    });
    this.set('filtersUsed', filtersUsed);
  },

  /**
   * Run <code>clearFilter</code> in the each child filterView
   */
  clearFilters: function() {
    this.set('filterConditions', []);
    this.get('_childViews').forEach(function(childView) {
      if (childView['clearFilter']) {
        childView.clearFilter();
      }
    });
  }

});

});

require.register("views/common/time_range", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * use: {{view App.TimeRangeWidget controllerBinding="App.router.mainChartsController"}}
 * set controller.preset field with preset value
 * widget assign itself to controller like presetWidget (controller.get('presetWidget'))
 * @type {*}
 */
App.TimeRangeWidget = Em.View.extend({
  classNames:['time-range-widget'],
  templateName:require('templates/common/time_range'),
  dateFrom: null,
  dateTo: null,

  /**
   * presets
   */
  presets:[
    Em.Object.create({ label:Em.I18n.t('timeRange.presets.1hour'), value:'1h', period: 3600000, step: 300000}),
    Em.Object.create({ label:Em.I18n.t('timeRange.presets.12hour'), value:'12h', period: 43200000, step: 3600000}),
    Em.Object.create({ label:Em.I18n.t('timeRange.presets.1day'), value:'1d', period: 86400000, step: 3600000}),
    Em.Object.create({ label:Em.I18n.t('timeRange.presets.1week'), value:'1wk', period: 604800000, step: 86400000}),
    Em.Object.create({ label:Em.I18n.t('timeRange.presets.1month'), value:'1mo', period: 2592000000, step: 86400000}),
    Em.Object.create({ label:Em.I18n.t('timeRange.presets.1year'), value:'1yr', period: 31536000000, step: 2592000000})
  ],
  /**
   * chosen preset value
   */
  chosenPreset: null,

  /**
   * return array of chosen presets
   */
  chosenPresets:function () {
    return this.get('chosenPreset') ? [this.get('chosenPreset')] : this.get('defaultPresets');
  }.property('chosenPreset'),

  /**
   * preset item view
   */
  presetView:Em.View.extend({
    tagName:'li',
    classNameBindings:['disabled'],
    disabled:function () {
      return this.get('isActive') ? "disabled" : false;
    }.property('isActive'),
    isActive:function () {
      return this.get('preset.value') == this.get('widget.chosenPreset.value');
    }.property('widget.chosenPreset'),
    template:Em.Handlebars.compile('<a {{action activate view.preset target="view.widget" href="true" }}>{{unbound view.preset.label}}</a>')
  }),

  /**
   * return default selected presets (currently - all)
   */
  defaultPresets:function () {
    var values = [];
    $.each(this.get('presets'), function () {
      if (this.value) {
        values.push(this.value);
      }
    });
    return values;
  }.property(),

  bindToController:function () {
    var thisW = this;
    var controller = this.get('controller');
    controller.set('presetWidget', thisW);
  },

  /**
   * assign this widget to controller, prepare items by presetsConfig
   */
  init:function () {
    this._super();
    this.bindToController();
  },

  /**
   * write active preset to widget
   * @param event
   */
  activate:function (event) {
    if (event.context == this.get('chosenPreset')) {
      this.set('chosenPreset', null);
    } else {
      this.set('chosenPreset', event.context);
    }
  },

  dateFromView: Ember.TextField.extend({
    elementId: 'timeRangeFrom',
    classNames: 'timeRangeFrom',
    attributeBindings:['readonly'],
    readonly: true,
    didInsertElement: function() {
      var self = this;
      this.$().datetimepicker({
        dateFormat: 'dd/mm/yy',
        timeFormat: 'hh:mm',
        maxDate: new Date(),
        onClose:function (dateText, inst) {
          var endDateTextBox = $('#timeRangeTo');
          if (endDateTextBox.val() != '') {
            var testStartDate = new Date(dateText);
            var testEndDate = new Date(endDateTextBox.val());
            if (testStartDate > testEndDate)
              endDateTextBox.val(dateText);
          } else {
            endDateTextBox.val(dateText);
          }
          self.set('dateFrom', dateText);
        },
        onSelect:function (selectedDateTime) {
          var start = $(this).datetimepicker('getDate');
          $('#timeRangeTo').datetimepicker('option', 'minDate', new Date(start.getTime()));
        }
      });
      self.set('dateFrom', this.get('value'));
    }
  }),

  dateToView: Ember.TextField.extend({
    elementId: 'timeRangeTo',
    classNames: 'timeRangeTo',
    attributeBindings:['readonly'],
    readonly: true,
    didInsertElement: function() {
      var self = this;
      this.$().datetimepicker({
        dateFormat: 'dd/mm/yy',
        timeFormat: 'hh:mm',
        maxDate: new Date(),
        onClose:function (dateText, inst) {
          var startDateTextBox = $('#timeRangeFrom');
          if (startDateTextBox.val() != '') {
            var testStartDate = new Date(startDateTextBox.val());
            var testEndDate = new Date(dateText);
            if (testStartDate > testEndDate)
              startDateTextBox.val(dateText);
          } else {
            startDateTextBox.val(dateText);
          }
          self.set('dateTo', dateText);
        },
        onSelect:function (selectedDateTime) {
          var end = $(this).datetimepicker('getDate');
          $('#timeRangeFrom').datetimepicker('option', 'maxDate', new Date(end.getTime()));
        }
      });
      self.set('dateTo', this.get('value'));
    }
  }),

  sliderOptions: Ember.Object.extend({
    end: null,
    period: null,
    start: function() {
      return this.get('end') - this.get('period');
    }.property('end', 'period')
  }),
  nowLabel: null,
  rangeLabel: null,
  buildSlider: function() {
    if (this.get('chosenPreset')) {
      var sliderOptions = this.sliderOptions.create({
        end: function() {
          var endDate = new Date();
          return endDate.getTime();
        }.property(),
        period: this.get('chosenPreset.period'),
        step: this.get('chosenPreset.step'),
        countTimeAgo: function(stepValue) {
          var msAgo = this.get('end') - stepValue;
          return msAgo.toDaysHoursMinutes();
        }
      });
      this.set('nowLabel', 'Now');
      this.set('rangeLabel', new Date(sliderOptions.get('start')));

      var self = this;
      $('#slider').slider({
        range: "max",
        min: sliderOptions.get('start'),
        max: sliderOptions.get('end'),
        value: sliderOptions.get('start'),
        step: sliderOptions.get('step'),
        stop: function(event, ui) {
          self.set('rangeLabel', new Date(ui.value));
//          self.set('rangeLabel', sliderOptions.countTimeAgo(ui.value).h);
        },
        slide: function(event, ui){
          self.set('rangeLabel', new Date(ui.value));
//          self.set('rangeLabel', sliderOptions.countTimeAgo(ui.value).h);
        }
      });
    } else {
      console.log(this.get('chosenPreset'));
      $("#slider").slider("destroy");
    }
  }.observes('chosenPreset')
});
});

require.register("views/experimental", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.ExperimentalView = Em.View.extend({
  templateName: require('templates/experimental'),
  supports: function () {
    var supports = [];
    Em.keys(App.get('supports')).forEach(function (sup) {
      supports.push(Ember.Object.create({
        name: sup,
        selected: App.get('supports')[sup]
      }));
    });
    return supports;
  }.property('App.supports'),

  doSave: function () {
    var supports = this.get('supports');
    supports.forEach(function(s){
      var propName = 'App.supports.' + s.get('name');
      var propValue = s.get('selected');
      console.log(">>>>>>>> " + propName + " = "+ propValue);
      Ember.set(propName, propValue);
    });
    App.router.transitionTo('root.index');
  },

  doCancel: function () {
    App.router.transitionTo('root.index');
  }
});
});

require.register("views/installer", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.InstallerView = Em.View.extend(App.WizardMenuMixin, {

  templateName: require('templates/installer')

});
});

require.register("views/loading", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.LoadingView = Em.View.extend({
    tagName: 'h2',
    template: Ember.Handlebars.compile('{{t app.loadingPlaceholder}}')
});
});

require.register("views/login", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.LoginView = Em.View.extend({

  templateName: require('templates/login'),
  loginTextField: Em.TextField.extend({
    didInsertElement: function(){
      this._super();
      this.$().focus();
    },
    classNames: ['login-user-name']
  }),
  passTextField : Em.TextField.extend({
    insertNewline: function(){
      this.get("controller").submit();
    },
    classNames: ['login-user-password']
  })
});
});

require.register("views/main", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.MainView = Em.View.extend({
  templateName:require('templates/main')
});
});

require.register("views/main/admin", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminView = Em.View.extend({
  templateName: require('templates/main/admin'),
  selectedBinding: 'controller.category',
  categories: function() {
    var items = [];
    items.push({
      name: 'stackAndUpgrade',
      url: 'stackAndUpgrade.index',
      label: Em.I18n.t('admin.stackUpgrade.title')
    });
    items.push({
      name: 'adminServiceAccounts',
      url: 'adminServiceAccounts',
      label: Em.I18n.t('common.serviceAccounts')
    });
    if (!App.get('isHadoopWindowsStack')) {
      if (App.get('supports.automatedKerberos')) {
        items.push({
          name: 'kerberos',
          url: 'adminKerberos.index',
          label: Em.I18n.t('common.kerberos')
        });
      } else {
        items.push({
          name: 'security',
          url: 'adminSecurity.index',
          label: Em.I18n.t('common.security')
        });
      }
    }
    return items;
  }.property(''),

  NavItemView: Ember.View.extend({
    tagName: 'li',
    classNameBindings: 'isActive:active'.w(),
    isActive: function () {
      return this.get('item') === this.get('parentView.selected');
    }.property('item', 'parentView.selected')
  }),

  willDestroyElement: function () {
    //reset selected category in Admin sub-menu after leaving admin section
    this.set('controller.category', null);
  }
});


});

require.register("views/main/admin/advanced", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminAdvancedView = Em.View.extend({
  templateName: require('templates/main/admin/advanced'),
  params: {keepData: 1}
});
});

require.register("views/main/admin/advanced/password", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminAdvancedPasswordView = App.Form.extend({
  fieldsOptions:[
    { name:"password", displayName:Em.I18n.t('common.password'), displayType:"password", disableRequiredOnExistent:true },
    { name:"passwordRetype", displayName:Em.I18n.t('form.passwordRetype'), displayType:"passwordRetype", disableRequiredOnExistent:true }
  ],
  fields: []
});
});

require.register("views/main/admin/audit", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

require('views/common/grid');

App.MainAdminAuditView = App.Grid.extend({
  _collection: {className: App.ServiceAudit},
  prepareColumns:function () {
    this._super();

    this.addColumn({
      name:"date",
      label:Em.I18n.t("admin.audit.grid.date")
    });
    this.addColumn({
      name:"service.label",
      label:Em.I18n.t("admin.audit.grid.service")
    });
    this.addColumn({
      name:"operationName",
      label:Em.I18n.t("admin.audit.grid.operationName")
    });
    this.addColumn({
      name:"user.userName",
      label:Em.I18n.t("admin.audit.grid.performedBy")
    });
  }
});
});

require.register("views/main/admin/authentication", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminAuthenticationView = Em.View.extend({
  templateName:require('templates/main/admin/authentication'),
  form:App.AuthenticationForm.create({}),
  pushAuthenticationToForm:function () {
    var auth = App.Authentication.find(1);
    App.router.set('mainAdminAuthenticationController.content', auth);
    this.form.set('object', auth);
  },
  didInsertElement: function (){
    this._super();
    this.pushAuthenticationToForm();
  },

  ldapChecked:function () {
    return this.get('form.field.method.value');
  }.property('form.field.method.value'),

  useCredentials:function () {
    return this.get('form.field.bindMethod.value');
  }.property('form.field.bindMethod.value')
});
});

require.register("views/main/admin/highAvailability/nameNode/rollbackHA/rollback_wizard_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.RollbackHighAvailabilityWizardView = Em.View.extend(App.WizardMenuMixin, {

  templateName: require('templates/main/admin/highAvailability/nameNode/rollbackHA/rollback_wizard'),

  didInsertElement: function() {
    var currentStep = this.get('controller.currentStep');
    if (currentStep > 4) {
      this.get('controller').setLowerStepsDisable(currentStep);
    }
  }

});

});

require.register("views/main/admin/highAvailability/nameNode/rollbackHA/step1_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.RollbackHighAvailabilityWizardStep1View = Em.View.extend({

  templateName: require('templates/main/admin/highAvailability/nameNode/rollbackHA/step1'),

  addNNHosts: null,
  sNNHosts: null,
  selectedSNNHost: null,
  selectedAddNNHost: null,

  isLoaded: false,

  loadHostsName: function () {
    App.ajax.send({
      name: 'hosts.all',
      sender: this,
      data: {},
      success: 'loadHostsNameSuccessCallback',
      error: 'loadHostsNameErrorCallback'
    });
  },

  loadHostsNameSuccessCallback: function (data) {
    var addNNHosts = App.HostComponent.find().filterProperty('componentName', 'NAMENODE');

    this.secondaryNNHosts = [];
    this.set('selectedSNNHost', this.get('controller.content.sNNHost'));
    this.set('selectedAddNNHost', this.get('controller.content.addNNHost'));

    if (addNNHosts.length == 2) {
      this.set('addNNHosts', addNNHosts.mapProperty('hostName'));
    }
    data.items.forEach(function (host) {
      this.secondaryNNHosts.push(host.Hosts.host_name);
    }, this);
    this.set('sNNHosts', this.secondaryNNHosts);
    this.set('isLoaded', true);
  },
  loadHostsNameErrorCallback: function(){
    this.set('isLoaded', true);
  },

  didInsertElement: function() {
    this.loadHostsName();
  },

  tipAddNNHost: function () {
    return this.get('controller.content.addNNHost');
  }.property('controller.content.addNNHost'),

  tipSNNHost: function () {
    return this.get('controller.content.sNNHost');
  }.property('controller.content.sNNHost'),

  done: function () {
    this.get('controller.content').set('selectedSNNHost', this.get('selectedSNNHost'));
    this.get('controller.content').set('selectedAddNNHost', this.get('selectedAddNNHost'));
    App.router.send('next');
  }

});

});

require.register("views/main/admin/highAvailability/nameNode/rollbackHA/step2_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.RollbackHighAvailabilityWizardStep2View = App.HighAvailabilityWizardStep4View.extend({

  templateName: require('templates/main/admin/highAvailability/nameNode/rollbackHA/step2'),

  step2BodyText: function () {
    var activeNN = App.HostComponent.find().findProperty('displayNameAdvanced','Active NameNode');
    if(!activeNN){
      activeNN = App.HostComponent.find().findProperty('componentName','NAMENODE');
    }
    activeNN = activeNN.get('hostName');
    this.get('controller.content').set('activeNNHost', activeNN);
    return Em.I18n.t('admin.highAvailability.rollback.step2.body').format(this.get('controller.content.hdfsUser'), activeNN);
  }.property()

});

});

require.register("views/main/admin/highAvailability/nameNode/rollbackHA/step3_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.RollbackHighAvailabilityWizardStep3View = Em.View.extend({

  templateName: require('templates/main/admin/highAvailability/nameNode/rollbackHA/step3'),

  didInsertElement: function() {

  }

});

});

require.register("views/main/admin/highAvailability/nameNode/rollback_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.HighAvailabilityRollbackView = App.HighAvailabilityProgressPageView.extend({
  templateName: require('templates/main/admin/highAvailability/nameNode/rollback'),

  notice: Em.I18n.t('admin.highAvailability.rollback.notice.inProgress'),
  submitButtonText: Em.I18n.t('common.done')


});

});

require.register("views/main/admin/highAvailability/nameNode/step1_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.HighAvailabilityWizardStep1View = Em.View.extend({

  templateName: require('templates/main/admin/highAvailability/nameNode/step1'),

  didInsertElement: function() {
    App.popover($("div.controls input[type=text]"), {'placement': 'right', 'trigger': 'hover', 'title': this.t('admin.highAvailability.wizard.step1.nameserviceid.tooltip.title'), 'content': this.t('admin.highAvailability.wizard.step1.nameserviceid.tooltip.content')});
  },

  showInputError: function () {
    return !this.get('controller.isNameServiceIdValid') && this.get('controller.content.nameServiceId').length != 0 ;
  }.property('controller.isNameServiceIdValid', 'controller.content.nameServiceId')

});

});

require.register("views/main/admin/highAvailability/nameNode/step2_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

require('views/wizard/step5_view');

App.HighAvailabilityWizardStep2View = App.WizardStep5View.extend({

  templateName: require('templates/main/admin/highAvailability/nameNode/step2')

});

});

require.register("views/main/admin/highAvailability/nameNode/step3_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.HighAvailabilityWizardStep3View = Em.View.extend({

  templateName: require('templates/main/admin/highAvailability/nameNode/step3'),
  didInsertElement: function () {
    this.get('controller').loadStep();
  },
  curNameNode: function () {
    return this.get('controller.content.masterComponentHosts').findProperty('isCurNameNode', true).hostName;
  }.property('controller.content.masterComponentHosts'),
  addNameNode: function () {
    return this.get('controller.content.masterComponentHosts').findProperty('isAddNameNode', true).hostName;
  }.property('controller.content.masterComponentHosts'),
  secondaryNameNode: function () {
    return this.get('controller.content.masterComponentHosts').findProperty('component', "SECONDARY_NAMENODE").hostName;
  }.property('controller.content.masterComponentHosts'),
  journalNodes: function () {
    return this.get('controller.content.masterComponentHosts').filterProperty('component', "JOURNALNODE");
  }.property('controller.content.masterComponentHosts')

});

});

require.register("views/main/admin/highAvailability/nameNode/step4_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.HighAvailabilityWizardStep4View = Em.View.extend({

  templateName: require('templates/main/admin/highAvailability/nameNode/step4'),

  didInsertElement: function() {
    this.get('controller').pullCheckPointStatus();
  },

  step4BodyText: function () {
    var nN = this.get('controller.content.masterComponentHosts').findProperty('isCurNameNode', true);
    return Em.I18n.t('admin.highAvailability.wizard.step4.body').format(this.get('controller.content.hdfsUser'), nN.hostName);
  }.property('controller.content.masterComponentHosts'),

  nnCheckPointText: function () {
    return (this.get('controller.isNextEnabled')) ?
      Em.I18n.t('admin.highAvailability.wizard.step4.ckCreated') :
      Em.I18n.t('admin.highAvailability.wizard.step4.ckNotCreated');
  }.property('controller.isNextEnabled')

});

});

require.register("views/main/admin/highAvailability/nameNode/step5_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.HighAvailabilityWizardStep5View = App.HighAvailabilityProgressPageView.extend({

  templateName: require('templates/main/admin/highAvailability/nameNode/step5')
});

});

require.register("views/main/admin/highAvailability/nameNode/step6_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.HighAvailabilityWizardStep6View = Em.View.extend({

  templateName: require('templates/main/admin/highAvailability/nameNode/step6'),

  didInsertElement: function() {
    this.get('controller').pullCheckPointStatus();
  },

  step6BodyText: function () {
    var nN = this.get('controller.content.masterComponentHosts').findProperty('isCurNameNode', true);
    return Em.I18n.t('admin.highAvailability.wizard.step6.body').format(this.get('controller.content.hdfsUser'), nN.hostName);
  }.property('controller.content.masterComponentHosts'),

  jnCheckPointText: function () {
    return this.get('controller.isNextEnabled') ?
      Em.I18n.t('admin.highAvailability.wizard.step6.jsInit') :
      Em.I18n.t('admin.highAvailability.wizard.step6.jsNoInit');
  }.property('controller.isNextEnabled')

});

});

require.register("views/main/admin/highAvailability/nameNode/step7_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.HighAvailabilityWizardStep7View = App.HighAvailabilityProgressPageView.extend({

  templateName: require('templates/main/admin/highAvailability/nameNode/step7')
});

});

require.register("views/main/admin/highAvailability/nameNode/step8_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.HighAvailabilityWizardStep8View = Em.View.extend({

  templateName: require('templates/main/admin/highAvailability/nameNode/step8'),

  step8BodyText: function () {
    var nN = this.get('controller.content.masterComponentHosts').findProperty('isCurNameNode', true);
    var addNN = this.get('controller.content.masterComponentHosts').findProperty('isAddNameNode', true);
    return Em.I18n.t('admin.highAvailability.wizard.step8.body').format(this.get('controller.content.hdfsUser'), nN.hostName, addNN.hostName);
  }.property('controller.content.masterComponentHosts')

});

});

require.register("views/main/admin/highAvailability/nameNode/step9_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.HighAvailabilityWizardStep9View = App.HighAvailabilityProgressPageView.extend({

  templateName: require('templates/main/admin/highAvailability/nameNode/step9'),

  submitButtonText: Em.I18n.t('common.done'),

  noticeCompleted: Em.I18n.t('admin.highAvailability.wizard.step9.notice.completed')
});

});

require.register("views/main/admin/highAvailability/nameNode/wizard_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.HighAvailabilityWizardView = Em.View.extend(App.WizardMenuMixin, {

  templateName: require('templates/main/admin/highAvailability/nameNode/wizard'),

  didInsertElement: function() {
    var currentStep = this.get('controller.currentStep');
    if (currentStep > 4) {
      this.get('controller').setLowerStepsDisable(currentStep);
    }
  }

});

});

require.register("views/main/admin/highAvailability/progress_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.HighAvailabilityProgressPageView = Em.View.extend(App.wizardProgressPageViewMixin, {

  didInsertElement: function () {
    this.get('controller').loadStep();
  },

  headerTitle: function () {
    var currentStep = App.router.get('highAvailabilityWizardController.currentStep');
    if (currentStep == 1) {
      return  Em.I18n.t('admin.highAvailability.wizard.rollback.header.title');
    } else {
      return  Em.I18n.t('admin.highAvailability.wizard.step' + currentStep + '.header.title');
    }
  }.property(),

  noticeInProgress: function () {
    var currentStep = App.router.get('highAvailabilityWizardController.currentStep');
    if (currentStep == 1) {
      return  Em.I18n.t('admin.highAvailability.rollback.notice.inProgress');
    } else {
      return  Em.I18n.t('admin.highAvailability.wizard.step' + currentStep + '.notice.inProgress');
    }
  }.property(),

  notice: Em.I18n.t('admin.highAvailability.wizard.progressPage.notice.inProgress'),

  noticeClass: 'alert alert-info',

  onStatusChange: function () {
    var status = this.get('controller.status');
    if (status === 'COMPLETED') {
      this.set('notice', this.get('noticeCompleted'));
      this.set('noticeClass', 'alert alert-success');
    } else if (status === 'FAILED') {
      this.set('notice', this.get('noticeFailed'));
      this.set('noticeClass', 'alert alert-error');
    } else {
      this.set('notice', this.get('noticeInProgress'));
      this.set('noticeClass', 'alert alert-info');
    }
  }.observes('controller.status'),

  taskView: Em.View.extend({
    icon: '',
    iconColor: '',
    linkClass: '',

    didInsertElement: function () {
      this.onStatus();
      $('body').tooltip({
        selector: '[rel=tooltip]'
      });
    },

    barWidth: function () {
      return 'width: ' + this.get('content.progress') + '%;';
    }.property('content.progress'),

    onStatus: function () {
      var linkClass = !!this.get('content.requestIds.length') ? 'active-link' : 'active-text';
      this.set('linkClass', linkClass);
      if (this.get('content.status') === 'IN_PROGRESS') {
        this.set('icon', 'icon-cog');
        this.set('iconColor', 'text-info');
      } else if (this.get('content.status') === 'FAILED') {
        this.set('icon', 'icon-exclamation-sign');
        this.set('iconColor', 'text-error');
      } else if (this.get('content.status') === 'COMPLETED') {
        this.set('icon', 'icon-ok');
        this.set('iconColor', 'text-success');
      } else {
        this.set('icon', 'icon-cog');
        this.set('iconColor', '');
        this.set('linkClass', 'not-active-link');
      }
    }.observes('content.status', 'content.hosts.length'),

    showProgressBar: function () {
      return this.get('content.status') === "IN_PROGRESS";
    }.property('content.status'),

    hidePercent: function() {
      return this.get('content.command') === 'testDBConnection';
    }.property('content.command'),

    showDBTooltip: function() {
      return this.get('content.command') !== 'testDBConnection';
    }.property('content.command')
  })
});

});

require.register("views/main/admin/highAvailability/resourceManager/step1_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.RMHighAvailabilityWizardStep1View = Em.View.extend({

  templateName: require('templates/main/admin/highAvailability/resourceManager/step1')

});

});

require.register("views/main/admin/highAvailability/resourceManager/step2_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.RMHighAvailabilityWizardStep2View = App.WizardStep5View.extend({

  templateName: require('templates/main/admin/highAvailability/resourceManager/step2')

});

});

require.register("views/main/admin/highAvailability/resourceManager/step3_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.RMHighAvailabilityWizardStep3View = Em.View.extend({

  templateName: require('templates/main/admin/highAvailability/resourceManager/step3'),

  didInsertElement: function () {
    this.get('controller').loadStep();
  }
});

});

require.register("views/main/admin/highAvailability/resourceManager/step4_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.RMHighAvailabilityWizardStep4View = App.HighAvailabilityProgressPageView.extend({

  templateName: require('templates/main/admin/highAvailability/resourceManager/step4'),

  headerTitle: Em.I18n.t('admin.rm_highAvailability.wizard.step4.header'),

  noticeInProgress: Em.I18n.t('admin.rm_highAvailability.wizard.step4.notice.inProgress'),

  noticeCompleted: Em.I18n.t('admin.rm_highAvailability.wizard.step4.notice.completed'),

  submitButtonText: Em.I18n.t('common.complete')

});

});

require.register("views/main/admin/highAvailability/resourceManager/wizard_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.RMHighAvailabilityWizardView = Em.View.extend(App.WizardMenuMixin, {

  didInsertElement: function() {
    var currentStep = this.get('controller.currentStep');
    if (currentStep > 3) {
      this.get('controller').setLowerStepsDisable(currentStep);
    }
  },

  templateName: require('templates/main/admin/highAvailability/resourceManager/wizard'),

  isLoaded: false,

  willInsertElement: function() {
    this.set('isLoaded', false);
    this.loadHosts();
  },

  /**
   * load hosts from server
   */
  loadHosts: function () {
    App.ajax.send({
      name: 'hosts.high_availability.wizard',
      data: {},
      sender: this,
      success: 'loadHostsSuccessCallback',
      error: 'loadHostsErrorCallback'
    });
  },

  loadHostsSuccessCallback: function (data, opt, params) {
    var hosts = {};

    data.items.forEach(function (item) {
      hosts[item.Hosts.host_name] = {
        name: item.Hosts.host_name,
        cpu: item.Hosts.cpu_count,
        memory: item.Hosts.total_mem,
        disk_info: item.Hosts.disk_info,
        bootStatus: "REGISTERED",
        isInstalled: true
      };
    });
    App.db.setHosts(hosts);
    this.set('controller.content.hosts', hosts);
    this.set('isLoaded', true);
  },

  loadHostsErrorCallback: function(){
    this.set('isLoaded', true);
  }
});

});

require.register("views/main/admin/kerberos", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminKerberosView = Em.View.extend({
  templateName: require('templates/main/admin/kerberos'),
  didInsertElement: function() {
    var controller = this.get('controller');
    controller.set('dataIsLoaded',false);
    controller.loadSecurityStatusFromServer();
  }
});
});

require.register("views/main/admin/kerberos/disable_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

require('views/main/admin/kerberos/progress_view');

App.KerberosDisableView = App.KerberosProgressPageView.extend({

  didInsertElement: function () {
    this._super();
    this.get('controller').loadStep();
  },

  headerTitle: '',

  noticeCompleted: Em.I18n.t('admin.kerberos.disable.notice.completed'),

  notice: Em.I18n.t('admin.kerberos.disable.inProgress'),

  templateName: require('templates/main/admin/kerberos/disable'),

  msgColor: 'alert-info',

  submitButtonText: Em.I18n.t('common.complete')

});

});

require.register("views/main/admin/kerberos/progress_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.KerberosProgressPageView = Em.View.extend(App.wizardProgressPageViewMixin, {

  didInsertElement: function () {
    this.get('controller').loadStep();
  },

  headerTitle: function () {
    var currentStep = App.router.get('kerberosWizardController.currentStep');
    return  Em.I18n.t('admin.kerberos.wizard.step' + currentStep + '.header');
  }.property(),

  noticeInProgress: function () {
    var currentStep = App.router.get('kerberosWizardController.currentStep');
    return  Em.I18n.t('admin.kerberos.wizard.step' + currentStep + '.notice.inProgress');
  }.property(),

  notice: Em.I18n.t('admin.kerberos.wizard.progressPage.notice.inProgress')
});

});

require.register("views/main/admin/kerberos/step1_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.KerberosWizardStep1View = Em.View.extend({

  templateName: require('templates/main/admin/kerberos/step1'),

  didInsertElement: function () {
    var controller = this.get('controller');
    controller.loadStep();
  },

  /**
   * whenever the KDC type is changed, all checkboxes for the precondition should be unchecked
   */
  resetCheckBoxes: function () {
    var options = this.get('controller.options');
    options.forEach(function (option) {
      option.preConditions.forEach(function (_condition) {
        _condition.set('checked', false);
      });
    }, this)
  }.observes('controller.selectedItem')

});



});

require.register("views/main/admin/kerberos/step2_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

require('views/wizard/step7_view');

App.KerberosWizardStep2View = App.WizardStep7View.extend({

  templateName: require('templates/main/admin/kerberos/step2'),

  didInsertElement: function() {
    var controller =  this.get('controller');
    controller.loadStep();
  }
});

});

require.register("views/main/admin/kerberos/step3_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.KerberosWizardStep3View = App.KerberosProgressPageView.extend({

  templateName: require('templates/main/admin/kerberos/step3'),

  noticeCompleted: Em.I18n.t('admin.kerberos.wizard.step3.notice.completed'),

  submitButtonText: Em.I18n.t('common.next') + '&rarr;',

  showBackButton: true

});
});

require.register("views/main/admin/kerberos/step4_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.KerberosWizardStep4View = Em.View.extend({

  templateName: require('templates/main/admin/kerberos/step4')

});
});

require.register("views/main/admin/kerberos/step5_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.KerberosWizardStep5View = App.KerberosProgressPageView.extend({

  templateName: require('templates/main/admin/kerberos/step5'),

  noticeCompleted: Em.I18n.t('admin.kerberos.wizard.step5.notice.completed'),

  submitButtonText: Em.I18n.t('common.next') + '&rarr;',

  showBackButton: true
});

});

require.register("views/main/admin/kerberos/step6_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.KerberosWizardStep6View = App.KerberosProgressPageView.extend({

  templateName: require('templates/main/admin/kerberos/step6'),

  noticeCompleted: Em.I18n.t('admin.kerberos.wizard.step6.notice.completed'),

  submitButtonText: Em.I18n.t('common.next') + '&rarr;',

  showBackButton: true

});
});

require.register("views/main/admin/kerberos/step7_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');

App.KerberosWizardStep7View = App.KerberosProgressPageView.extend({

  templateName: require('templates/main/admin/kerberos/step7'),

  noticeCompleted: Em.I18n.t('admin.kerberos.wizard.step7.notice.completed'),

  noticeFailed: Em.I18n.t('admin.kerberos.wizard.step7.notice.failed'),

  submitButtonText: Em.I18n.t('common.complete'),

  showBackButton: false

});
});

require.register("views/main/admin/kerberos/wizard_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.KerberosWizardView = Em.View.extend(App.WizardMenuMixin, {

  templateName: require('templates/main/admin/kerberos/wizard')
});



});

require.register("views/main/admin/security", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminSecurityView = Em.View.extend({
  templateName: require('templates/main/admin/security'),
  didInsertElement: function() {
    this.set('controller.dataIsLoaded',false);
    this.get('controller').setSecurityStatus();
  },

  configProperties: function () {
    var configProperties = [];
    var stepConfigs = this.get('controller.stepConfigs');
    if (stepConfigs) {
      this.get('controller.stepConfigs').mapProperty('configs').forEach(function (_stepProperties) {
        _stepProperties.forEach(function (_stepConfigProperty) {
          configProperties.pushObject(_stepConfigProperty);
        }, this);
      }, this);
    }
    return configProperties;
  }.property('controller.stepConfigs.@each.configs'),

  realmName: function () {
    return this.get('configProperties').findProperty('name', 'kerberos_domain');
  }.property('configProperties'),

  onRealmNameChange: function () {
    this.get('configProperties').forEach(function (_property) {
      if (/principal_name?$/.test(_property.get('name'))) {
        _property.set('unit', '@' + this.get('realmName.value'));
      }
    }, this);
  }.observes('realmName.value')

});
});

require.register("views/main/admin/security/add/menu", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminSecurityAddMenuView = Em.View.extend(App.WizardMenuMixin, {

  templateName: require('templates/main/admin/security/add/menu')

});



});

require.register("views/main/admin/security/add/step1", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminSecurityAddStep1View = Em.View.extend({

  templateName: require('templates/main/admin/security/add/step1'),
  shouldRemoveATS: false,
  didInsertElement: function(){
    var shouldRemoveATS = this.get('controller').shouldRemoveATS();
    this.set('shouldRemoveATS', shouldRemoveATS);
  }

});




});

require.register("views/main/admin/security/add/step2", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminSecurityAddStep2View = Em.View.extend({

  templateName: require('templates/main/admin/security/add/step2'),

  configProperties: function () {
    var configProperties = [];
    var stepConfigs = this.get('controller.stepConfigs');
    if (stepConfigs) {
      this.get('controller.stepConfigs').mapProperty('configs').forEach(function (_stepProperties) {
        _stepProperties.forEach(function (_stepConfigProperty) {
          configProperties.pushObject(_stepConfigProperty);
        }, this);
      }, this);
    }
    return configProperties;
  }.property('controller.stepConfigs.@each.configs'),

  realmName: function () {
    return this.get('configProperties').findProperty('name', 'kerberos_domain');
  }.property('configProperties'),

  onRealmNameChange: function () {
    this.get('configProperties').forEach(function (_property) {
      if (/principal_name?$/.test(_property.get('name')) || _property.get('name') == 'namenode_principal_name_falcon') {
        _property.set('unit', '@' + this.get('realmName.value'));
      }
    }, this);
  }.observes('realmName.value')

});

});

require.register("views/main/admin/security/add/step3", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminSecurityAddStep3View = Em.View.extend({
  templateName: require('templates/main/admin/security/add/step3'),
  didInsertElement: function() {
    this.set('controller.isLoaded', false);
    this.get('controller').loadHosts();
  }
});


});

require.register("views/main/admin/security/add/step4", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminSecurityAddStep4View = Em.View.extend({

  templateName: require('templates/main/admin/security/add/step4'),
  didInsertElement: function () {
    this.get('controller').loadStep();
  },
  msgColor: 'alert-info',
  message: Em.I18n.t('admin.security.step4.body.header'),
  onResult: function () {
    var stopServiceCommand = this.get('controller.commands').findProperty('name', 'STOP_SERVICES');
    var applyConfigCommand = this.get('controller.commands').findProperty('name', 'APPLY_CONFIGURATIONS');
      if (applyConfigCommand && applyConfigCommand.get('isSuccess') === true ) {
        this.set('message', Em.I18n.t('admin.security.step4.body.success.header'));
        this.set('msgColor','alert-success');
      } else if ((stopServiceCommand && stopServiceCommand.get('isError') === true) || (applyConfigCommand && applyConfigCommand.get('isError') === true)) {
        this.set('message', Em.I18n.t('admin.security.step4.body.failure.header'));
        this.set('msgColor','alert-error');
      } else {
        this.set('message', Em.I18n.t('admin.security.step4.body.header'));
        this.set('msgColor','alert-info');
      }
  }.observes('controller.commands.@each.isCompleted')

});

App.StageStatusView = Em.View.extend({
  tagName: 'tr',
  hasStarted: null,
  classNameBindings: ['faintText']
});

});

require.register("views/main/admin/security/disable", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminSecurityDisableView = Em.View.extend({

  templateName: require('templates/main/admin/security/disable'),

  didInsertElement: function () {
    this.get('controller').loadStep();
  },
  msgColor: 'alert-info',
  message: Em.I18n.t('admin.security.disable.body.header'),
  onResult: function () {
    var stopServiceCommand = this.get('controller.commands').findProperty('name', 'STOP_SERVICES');
    var applyConfigCommand = this.get('controller.commands').findProperty('name', 'APPLY_CONFIGURATIONS');
    if (applyConfigCommand && applyConfigCommand.get('isSuccess') === true ) {
      this.set('message', Em.I18n.t('admin.security.disable.body.success.header'));
      this.set('msgColor', 'alert-success');
    } else if ((stopServiceCommand && stopServiceCommand.get('isError') === true) || (applyConfigCommand && applyConfigCommand.get('isError') === true)) {
      this.set('message', Em.I18n.t('admin.security.disable.body.failure.header'));
      this.set('msgColor', 'alert-error');
    } else {
      this.set('message', Em.I18n.t('admin.security.disable.body.header'));
      this.set('msgColor', 'alert-info');
    }
  }.observes('controller.commands.@each.isCompleted')
});

});

require.register("views/main/admin/serviceAccounts_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminServiceAccountsView = Em.View.extend({
  templateName: require('templates/main/admin/serviceAccounts'),
  didInsertElement: function() {
    this.get('controller').loadUsers();
  }
});

});

require.register("views/main/admin/stack_and_upgrade_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

App.MainAdminStackAndUpgradeView = Em.View.extend({
  templateName: require('templates/main/admin/stack_and_upgrade')
});


});

require.register("views/main/admin/stack_upgrade/menu_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAdminStackMenuView = Em.CollectionView.extend({
  tagName: 'ul',
  classNames: ["nav", "nav-tabs"],
  defaultRoute: 'services',

  content: function () {
    return [
      Em.Object.create({
        name: 'services',
        label: Em.I18n.t('common.stack'),
        routing: 'services'
      }),
      Em.Object.create({
        name: 'versions',
        label: Em.I18n.t('common.versions'),
        routing: 'versions',
        hidden: !App.get('stackVersionsAvailable')
      })
    ]
  }.property('App.stackVersionsAvailable'),

  didInsertElement: function () {
    this.activateView();
  },

  activateView: function () {
    var defaultRoute = App.router.get('currentState.name') || this.get('defaultRoute');
    $.each(this._childViews, function () {
      this.set('active', (this.get('content.routing') == defaultRoute) ? "active" : "");
    });
  },

  deactivateChildViews: function () {
    $.each(this._childViews, function () {
      this.set('active', "");
    });
  },

  itemViewClass: Em.View.extend({
    classNameBindings: ["active"],
    active: "",
    template: Ember.Handlebars.compile('{{#unless view.content.hidden}}<a {{action stackNavigate view.content.routing }} href="#"> {{unbound view.content.label}}</a>{{/unless}}')
  })
});

});

require.register("views/main/admin/stack_upgrade/services_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.MainAdminStackServicesView = Em.View.extend({
  templateName: require('templates/main/admin/stack_upgrade/services'),

  /**
   * @type {Array}
   */
  services: function() {
    var services = App.supports.installGanglia ? App.StackService.find() : App.StackService.find().without(App.StackService.find('GANGLIA'));
    return services.map(function(s) {
      s.set('isInstalled', App.Service.find().someProperty('serviceName', s.get('serviceName')));
      return s;
    });
  }.property('App.router.clusterController.isLoaded'),

  didInsertElement: function () {
    if (!App.get('stackVersionsAvailable')) {
      this.get('controller').loadRepositories();
    }
  },

  /**
   * launch Add Service wizard
   * @param event
   */
  goToAddService: function (event) {
    if (event.context == "KERBEROS") {
      App.router.get('mainAdminKerberosController').checkAndStartKerberosWizard();
      App.router.get('kerberosWizardController').setDBProperty('onClosePath', 'main.admin.stackAndUpgrade.services');
    } else {
      App.router.get('addServiceController').set('serviceToInstall', event.context);
      App.router.get('addServiceController').setDBProperty('onClosePath', 'main.admin.stackAndUpgrade.services');
      App.get('router').transitionTo('main.serviceAdd');
    }
  },

  /**
   * List of all repo-groups
   * @type {Object[][]}
   */
  allRepositoriesGroups: function () {
    var repos = this.get('controller.allRepos');
    var reposGroup = [];
    var repositories = [];
    reposGroup.set('stackVersion', App.get('currentStackVersionNumber'));
    if (repos) {
      repos.forEach(function (group) {
        group.repositories.forEach (function(repo) {
          var cur_repo = Em.Object.create({
            'repoId': repo.repoId,
            'id': repo.repoId + '-' + repo.osType,
            'repoName' : repo.repoName,
            'stackName' : repo.stackName,
            'stackVersion' : repo.stackVersion,
            'baseUrl': repo.baseUrl,
            'originalBaseUrl': repo.baseUrl,
            'osType': repo.osType,
            'onEdit': false,
            'empty-error': !repo.baseUrl,
            'undo': false,
            'clearAll': repo.baseUrl
          });
          var cur_group = reposGroup.findProperty('name', group.name);
          if (!cur_group) {
            var cur_group = Ember.Object.create({
              name: group.name,
              repositories: []
            });
            reposGroup.push(cur_group);
          }
          cur_group.repositories.push(cur_repo);
          repositories.push(cur_repo);
        });
      });
    }
    this.set('allRepos', repositories);
    return reposGroup;
  }.property('controller.allRepos'),

  /**
   * Onclick handler for edit action of each repo, enter edit mode
   * @param {object} event
   */
  onEditClick:function (event) {
    var targetRepo = this.get('allRepos').findProperty('id', event.context.get('id'));
    if (targetRepo) {
      targetRepo.set('onEdit', true);
    }
  },

  /**
   * Onclick handler for undo action of each repo group
   * @method undoGroupLocalRepository
   * @param {object} event
   */
  undoGroupLocalRepository: function (event) {
    this.doActionForGroupLocalRepository(event, 'originalBaseUrl');
  },

  /**
   * Handler for clear icon click
   * @method clearGroupLocalRepository
   * @param {object} event
   */
  clearGroupLocalRepository: function (event) {
    this.doActionForGroupLocalRepository(event, '');
  },

  /**
   * Common handler for repo groups actions
   * @method doActionForGroupLocalRepository
   * @param {object} event
   * @param {string} newBaseUrlField
   */
  doActionForGroupLocalRepository: function (event, newBaseUrlField) {
    var targetRepo = this.get('allRepos').findProperty('id', event.context.get('id'));
    if (targetRepo) {
      targetRepo.set('baseUrl', Em.isEmpty(newBaseUrlField) ? '' : Em.get(targetRepo, newBaseUrlField));
    }
  },

  /**
   * Handler when editing any repo group BaseUrl
   * @method editGroupLocalRepository
   */
  editGroupLocalRepository: function (event) {
    var repos = this.get('allRepos');
    repos.forEach(function (targetRepo) {
      targetRepo.set('undo', targetRepo.get('baseUrl') != targetRepo.get('originalBaseUrl'));
      targetRepo.set('clearAll', targetRepo.get('baseUrl'));
      targetRepo.set('empty-error', !targetRepo.get('baseUrl'));

    });
  }.observes('allRepos.@each.baseUrl'),

  /**
   * onSuccess callback for save Repo URL.
   */
  doSaveRepoUrlsSuccessCallback: function (response, request, data) {
    var id = data.repoId + '-' + data.osType;
    console.log('Success in check Repo URL. data repoId+osType: ' + id);
    var targetRepo = this.get('allRepos').findProperty('id', id);
    if (!targetRepo) {
      return;
    } else {

      var modalCloseHandler = function() {
        this.hide();
        targetRepo.set('baseUrl', data.data.Repositories.base_url);
        targetRepo.set('originalBaseUrl', data.data.Repositories.base_url);
        targetRepo.set('onEdit', false);
      };

      App.ModalPopup.show({
        header: Em.I18n.t('admin.cluster.repositories.popup.header.success'),
        secondary: null,
        onPrimary: modalCloseHandler,
        onClose: modalCloseHandler,
        message: Em.I18n.t('admin.cluster.repositories.popup.body.success'),
        bodyClass: Em.View.extend({
          template: Em.Handlebars.compile('<div class="alert alert-success">{{{message}}}</div>')
        })
      })
    }
  },

  /**
   * onError callback for save Repo URL.
   */
  doSaveRepoUrlsErrorCallback: function (request, ajaxOptions, error, data) {
    console.log('Error in check Repo URL. The baseURL sent is:  ' + data.data);
    var self = this;
    var id = data.url.split('/')[10] + '-' + data.url.split('/')[8];
    var targetRepo = this.get('allRepos').findProperty('id', id);
    if (!targetRepo) {
      return;
    } else {
      App.ModalPopup.show({
        header: Em.I18n.t('admin.cluster.repositories.popup.header.fail'),
        primary: Em.I18n.t('common.saveAnyway'),
        secondary: Em.I18n.t('common.revert'),
        third: Em.I18n.t('common.cancel'),
        onPrimary: function () {
          // save anyway: Go ahead and save with Repo URL validation turned off and close Dialog when done.
          this.hide();
          self.doSaveRepoUrls(id, false);
        },
        onSecondary: function () {
          // Revert: Close dialog, revert URL value, go back to non-Edit mode
          this.hide();
          targetRepo.set('baseUrl', targetRepo.get('originalBaseUrl'));
          targetRepo.set('onEdit', false);
        },
        onThird: function () {
          // cancel: Close dialog but stay in Edit mode
          this.hide();
        },
        message: Em.I18n.t('admin.cluster.repositories.popup.body.fail'),
        bodyClass: Em.View.extend({
          template: Em.Handlebars.compile('<div class="alert alert-warning">{{{message}}}</div>')
        })
      })
    }
  },

  /**
   * Check validation and Save the customized local urls
   */
  doSaveRepoUrls: function (id, verifyBaseUrl) {
    var targetRepo = this.get('allRepos').findProperty('id', id);
    var verifyBaseUrl = verifyBaseUrl;
    App.ajax.send({
      name: 'wizard.advanced_repositories.valid_url',
      sender: this,
      data: {
        stackName: targetRepo.stackName,
        stackVersion: targetRepo.stackVersion,
        repoId: targetRepo.repoId,
        osType: targetRepo.osType,
        data: {
          'Repositories': {
            'base_url': targetRepo.baseUrl,
            "verify_base_url": verifyBaseUrl
          }
        }
      },
      success: 'doSaveRepoUrlsSuccessCallback',
      error: 'doSaveRepoUrlsErrorCallback'
    });
  },
  /**
   * Check validation and Save the customized local urls
   */
  saveRepoUrls: function (event) {
    this.doSaveRepoUrls(event.context.get('id'), true);
  },

  /**
   * on click handler 'Cancel' for current repo in edit mode
   */
  doCancel: function (event) {
    var targetRepo = this.get('allRepos').findProperty('id', event.context.get('id'));
    if (targetRepo) {
      targetRepo.set('baseUrl', targetRepo.get('originalBaseUrl'));
      targetRepo.set('onEdit', false);
    }
  }
});

});

require.register("views/main/admin/stack_upgrade/upgrade_group_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.upgradeGroupView = Em.View.extend({
  templateName: require('templates/main/admin/stack_upgrade/upgrade_group'),

  /**
   * Only one UpgradeGroup or UpgradeItem could be expanded at a time
   * @param {object} event
   */
  toggleExpanded: function (event) {
    var isExpanded = event.context.get('isExpanded');
    event.contexts[1].filterProperty('isExpanded').forEach(function (item) {
      this.collapseLowerLevels(item);
      item.set('isExpanded', false);
    }, this);
    this.collapseLowerLevels(event.context);
    event.context.set('isExpanded', !isExpanded);
    if (!isExpanded && event.context.get('type') === 'ITEM') {
      event.context.set('isTasksLoaded', false);
      this.doPolling(event.context);
    }
  },

  /**
   * collapse sub-entities of current
   * @param {App.upgradeEntity} entity
   */
  collapseLowerLevels: function (entity) {
    if (entity.get('isExpanded')) {
      if (entity.type === 'ITEM') {
        entity.get('tasks').setEach('isExpanded', false);
      } else if (entity.type === 'GROUP') {
        entity.get('upgradeItems').forEach(function (item) {
          this.collapseLowerLevels(item);
          item.set('isExpanded', false);
        }, this);
      }
    }
  },

  /**
   * poll for tasks when item is expanded
   */
  doPolling: function (item) {
    var self = this;

    if (item && item.get('isExpanded')) {
      this.get('controller').getUpgradeItem(item).complete(function () {
        self.set('timer', setTimeout(function () {
          self.doPolling(item);
        }, App.bgOperationsUpdateInterval));
      });
    } else {
      clearTimeout(this.get('timer'));
    }
  },

  willDestroyElement: function () {
    clearTimeout(this.get('timer'));
  }
});

});

require.register("views/main/admin/stack_upgrade/upgrade_task_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.upgradeTaskView = Em.View.extend({
  templateName: require('templates/main/admin/stack_upgrade/upgrade_task'),

  /**
   * view observed directly
   * @type {boolean}
   */
  outsideView: false,

  /**
   * @type {boolean}
   */
  showContent: function () {
    return this.get('outsideView') || this.get('content.isExpanded');
  }.property('content.isExpanded', 'outsideView'),

  /**
   * @type {boolean}
   */
  errorLogOpened: false,

  /**
   * @type {boolean}
   */
  outputLogOpened: false,

  /**
   * @type {App.upgradeEntity}
   * @default null
   */
  content: null,

  /**
   * @type {Array}
   */
  tasks: [],

  /**
   * @type {string}
   */
  logTabId: function () {
    return this.get('elementId') + '-log-tab'
  }.property(''),

  /**
   * @type {string}
   */
  errorTabId: function () {
    return this.get('elementId') + '-error-tab'
  }.property(''),

  /**
   * @type {string}
   */
  logTabIdLink: function () {
    return '#' + this.get('logTabId');
  }.property(''),

  /**
   * @type {string}
   */
  errorTabIdLInk: function () {
    return '#' + this.get('errorTabId');
  }.property(''),

  /**
   * open error log in textarea to give ability to cope content
   * @param {object} event
   */
  copyErrLog: function(event) {
    this.toggleProperty('errorLogOpened');
  },

  /**
   * open stdout log in textarea to give ability to cope content
   * @param {object} event
   */
  copyOutLog: function(event) {
    this.toggleProperty('outputLogOpened');
  },

  /**
   * open error log in new window
   */
  openErrorLog: function () {
    this.openLogWindow(this.get('content.stderr'));
  },

  /**
   * open stdout log in new window
   */
  openOutLog: function () {
    this.openLogWindow(this.get('content.stdout'));
  },

  /**
   * open logs in new window
   * @param {string} log
   */
  openLogWindow: function(log) {
    var newWindow = window.open();
    var newDocument = newWindow.document;
    newDocument.write(log);
    newDocument.close();
  }
});

});

require.register("views/main/admin/stack_upgrade/upgrade_version_box_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');
var stringUtils = require('utils/string_utils');

App.UpgradeVersionBoxView = Em.View.extend({
  templateName: require('templates/main/admin/stack_upgrade/upgrade_version_box'),
  classNames: ['span4', 'version-box'],
  classNameBindings: ['versionClass'],

  /**
   * @type {string}
   * @constant
   */
  PROGRESS_STATUS: 'IN_PROGRESS',

  /**
   * progress of version installation
   * @type {number}
   */
  installProgress: function() {
    var requestId = App.get('testMode') ? 1 : App.db.get('repoVersionInstall', 'id')[0];
    var installRequest = App.router.get('backgroundOperationsController.services').findProperty('id', requestId);
    return (installRequest) ? installRequest.get('progress') : 0;
  }.property('App.router.backgroundOperationsController.serviceTimestamp'),

  /**
   * version is upgrading
   * @type {boolean}
   */
  isUpgrading: function () {
    return (this.get('controller.upgradeVersion') === this.get('content.displayName') && App.get('upgradeState') !== 'INIT');
  }.property('App.upgradeState', 'content.displayName', 'controller.upgradeVersion'),

  /**
   * @type {string}
   */
  versionClass: function () {
    return this.get('content.status') === 'CURRENT' ? 'current-version-box' : '';
  }.property('content.status'),

  /**
   * @type {boolean}
   */
  isOutOfSync: function () {
    return this.get('content.status') === 'OUT_OF_SYNC';
  }.property('content.status'),

  /**
   * map containing version (id, label)
   * this is used as param for <code>showHosts<code> method
   * @type {Object}
   */
  versionStateMap: {
    'current': {
      'id': 'current',
      'property': 'currentHosts',
      'label': Em.I18n.t('admin.stackVersions.hosts.popup.header.current')
    },
    'installed': {
      'id': 'installed',
      'property': 'installedHosts',
      'label': Em.I18n.t('admin.stackVersions.hosts.popup.header.installed')
    },
    'not_installed': {
      'id': 'installing',
      'property': 'notInstalledHosts',
      'label': Em.I18n.t('admin.stackVersions.hosts.popup.header.not_installed')
    }
  },

  /**
   * @type {object}
   * @default null
   */
  content: null,

  /**
   * object that describes how content should be displayed
   * @type {Em.Object}
   * TODO remove <code>isUpgrading</code> condition when transition of version states in API fixed
   */
  stateElement: function () {
    var currentVersion = this.get('controller.currentVersion');
    var upgradeVersion = this.get('controller.upgradeVersion');
    var status = this.get('content.status');
    var element = Em.Object.create({
      status: status,
      isInstalling: function () {
        return this.get('status') === 'INSTALLING';
      }.property('status'),
      isDisabled: false
    });

    if (status === 'CURRENT') {
      element.set('isLabel', true);
      element.set('text', Em.I18n.t('common.current'));
      element.set('class', 'label label-success');
    } else if (['INIT', 'INSTALL_FAILED', 'OUT_OF_SYNC'].contains(status)) {
      element.set('isButton', true);
      element.set('text', Em.I18n.t('admin.stackVersions.version.installNow'));
      element.set('action', 'installRepoVersionConfirmation');
      element.set('isDisabled', !App.isAccessible('ADMIN') || this.get('controller.requestInProgress'));
    } else if (status === 'INSTALLING') {
      element.set('iconClass', 'icon-cog');
      element.set('isLink', true);
      element.set('text', Em.I18n.t('hosts.host.stackVersions.status.installing'));
      element.set('action', 'showProgressPopup');
    } else if (status === 'INSTALLED' && !this.get('isUpgrading')) {
      if (stringUtils.compareVersions(this.get('content.repositoryVersion'), currentVersion.repository_version) === 1) {
        element.set('isButton', true);
        element.set('text', Em.I18n.t('admin.stackVersions.version.performUpgrade'));
        element.set('action', 'confirmUpgrade');
        element.set('isDisabled', !App.isAccessible('ADMIN') || this.get('controller.requestInProgress'));
      } else {
        element.set('iconClass', 'icon-ok');
        element.set('isLink', true);
        element.set('text', Em.I18n.t('common.installed'));
        element.set('action', null);
      }
    } else if (['UPGRADING', 'UPGRADE_FAILED', 'UPGRADED'].contains(status) || this.get('isUpgrading')) {
      element.set('isLink', true);
      element.set('action', 'openUpgradeDialog');
      if (['HOLDING', 'HOLDING_FAILED', 'HOLDING_TIMEDOUT'].contains(App.get('upgradeState'))) {
        element.set('iconClass', 'icon-pause');
        if (this.get('controller.isDowngrade')) {
          element.set('text', Em.I18n.t('admin.stackVersions.version.downgrade.pause'));
        } else {
          element.set('text', Em.I18n.t('admin.stackVersions.version.upgrade.pause'));
        }
      } else {
        element.set('iconClass', 'icon-cog');
        if (this.get('controller.isDowngrade')) {
          element.set('text', Em.I18n.t('admin.stackVersions.version.downgrade.running'));
        } else {
          element.set('text', Em.I18n.t('admin.stackVersions.version.upgrade.running'));
        }
      }
    }
    return element;
  }.property('content.status', 'controller.isDowngrade', 'isUpgrading', 'controller.requestInProgress'),

  didInsertElement: function () {
    App.tooltip($('.link-tooltip'), {title: Em.I18n.t('admin.stackVersions.version.linkTooltip')});
    App.tooltip($('.hosts-tooltip'));
    App.tooltip($('.out-of-sync-badge'), {title: Em.I18n.t('hosts.host.stackVersions.status.out_of_sync')});
  },

  /**
   * run custom action of controller
   */
  runAction: function () {
    var action = this.get('stateElement.action');
    if (action) {
      this.get('controller')[action](this.get('content'));
    }
  },

  /**
   * show popup with repositories to edit
   * @return {App.ModalPopup}
   */
  editRepositories: function () {
    var self = this;
    var repoRecord = App.RepositoryVersion.find(this.get('content.id'));
    //make deep copy of repoRecord
    var repo = Em.Object.create({
      repoVersionId: repoRecord.get('id'),
      displayName: repoRecord.get('displayName'),
      repositoryVersion: repoRecord.get('displayName'),
      operatingSystems: repoRecord.get('operatingSystems').map(function (os) {
        return Em.Object.create({
          osType: os.get('osType'),
          isSelected: true,
          repositories: os.get('repositories').map(function (repository) {
            return Em.Object.create({
              repoName: repository.get('repoName'),
              repoId: repository.get('repoId'),
              baseUrl: repository.get('baseUrl'),
              hasError: false
            });
          })
        });
      })
    });

    return App.ModalPopup.show({
      classNames: ['repository-list', 'sixty-percent-width-modal'],
      skipValidation: false,
      autoHeight: false,
      hasErrors: false,
      bodyClass: Ember.View.extend({
        content: repo,
        skipCheckBox: Ember.Checkbox.extend({
          classNames: ["align-checkbox"],
          change: function() {
            this.get('parentView.content.operatingSystems').forEach(function(os) {
              if (Em.get(os, 'repositories.length') > 0) {
                os.get('repositories').forEach(function(repo) {
                  Em.set(repo, 'hasError', false);
                })
              }
            });
          }
        }),
        templateName: require('templates/main/admin/stack_upgrade/edit_repositories'),
        didInsertElement: function () {
          App.tooltip($("[rel=skip-validation-tooltip]"), {placement: 'right'});
        }
      }),
      header: Em.I18n.t('common.repositories'),
      primary: Em.I18n.t('common.save'),
      disablePrimary: !(App.get('isAdmin') && !App.get('isOperator')),
      onPrimary: function () {
        var self = this;
        App.get('router.mainAdminStackAndUpgradeController').saveRepoOS(repo, this.get('skipValidation')).done(function(data){
          if (data.length > 0) {
            self.set('hasErrors', true);
          } else {
            self.hide();
          }
        })
      }
    });
  },

  /**
   * shows popup with listed hosts wich has current state of hostStackVersion
   * @param {object} event
   * @returns {App.ModalPopup}
   * @method showHostsListPopup
   */
  showHosts: function (event) {
    var status = event.contexts[0];
    var displayName = this.get('content.displayName');
    var hosts = this.get('content').get(status['property']);
    var self = this;
    if (hosts.length) {
      return App.ModalPopup.show({
        bodyClass: Ember.View.extend({
          title: Em.I18n.t('admin.stackVersions.hosts.popup.title').format(displayName, status.label, hosts.length),
          hosts: hosts,
          template: Em.Handlebars.compile('<h4>{{view.title}}</h4><div class="limited-height-2">{{#each view.hosts}}<div>{{this}}</div>{{/each}}</div>')
        }),
        header: Em.I18n.t('admin.stackVersions.hosts.popup.header').format(status.label),
        primary: Em.I18n.t('admin.stackVersions.hosts.popup.primary'),
        secondary: Em.I18n.t('common.close'),
        onPrimary: function () {
          this.hide();
          self.filterHostsByStack(displayName, status.id);
        }
      });
    }
  },

  /**
   * goes to the hosts page with content filtered by repo_version_name and repo_version_state
   * @param displayName
   * @param state
   * @method filterHostsByStack
   */
  filterHostsByStack: function (displayName, state) {
    if (!displayName || !state) return;
    App.router.get('mainHostController').filterByStack(displayName, state);
    App.router.get('mainHostController').set('showFilterConditionsFirstLoad', true);
    App.router.transitionTo('hosts.index');
  }
});

});

require.register("views/main/admin/stack_upgrade/upgrade_wizard_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');
var stringUtils = require('utils/string_utils');

App.upgradeWizardView = Em.View.extend({
  controllerBinding: 'App.router.mainAdminStackAndUpgradeController',
  templateName: require('templates/main/admin/stack_upgrade/stack_upgrade_wizard'),

  /**
   * @type {Array}
   */
  failedStatuses: ['HOLDING_FAILED', 'HOLDING_TIMEDOUT', 'FAILED', 'TIMED_OUT'],

  /**
   * @type {Array}
   */
  activeStatuses: ['HOLDING_FAILED', 'HOLDING_TIMEDOUT', 'FAILED', 'TIMED_OUT', 'HOLDING', 'IN_PROGRESS'],

  /**
   * update timer
   * @type {number|null}
   * @default null
   */
  updateTimer: null,

  /**
   * update timer of Upgrade Item
   * @type {number|null}
   * @default null
   */
  upgradeItemTimer: null,

  /**
   * @type {boolean}
   */
  isLoaded: false,

  /**
   * @type {boolean}
   */
  isDetailsOpened: false,

  /**
   * @type {boolean}
   */
  outsideView: true,

  /**
   * Downgrade should be available only if target version higher than current, so we can't downgrade
   * when downgrade already started
   * @type {boolean}
   */
  isDowngradeAvailable: function () {
    return !this.get('controller.isDowngrade');
  }.property('controller.isDowngrade'),

  /**
   * progress value is rounded to floor
   * @type {number}
   */
  overallProgress: function () {
    return Math.floor(this.get('controller.upgradeData.Upgrade.progress_percent'));
  }.property('controller.upgradeData.Upgrade.progress_percent'),

  /**
   * upgrade groups, reversed and PENDING ones are hidden
   * @type {Array}
   */
  upgradeGroups: function () {
    return this.get('controller.upgradeData.upgradeGroups') || [];
  }.property('controller.upgradeData.upgradeGroups'),

  /**
   * currently active group
   * @type {object|undefined}
   */
  activeGroup: function () {
    return this.get('upgradeGroups').find(function (item) {
      return this.get('activeStatuses').contains(item.get('status'));
    }, this);
  }.property('upgradeGroups.@each.status'),

  /**
   * if upgrade group is in progress it should have currently running item
   * @type {object|undefined}
   */
  runningItem: function () {
    return this.get('activeGroup.upgradeItems') && this.get('activeGroup.upgradeItems').findProperty('status', 'IN_PROGRESS');
  }.property('activeGroup.upgradeItems.@each.status'),

  /**
   * if upgrade group is failed it should have failed item
   * @type {object|undefined}
   */
  failedItem: function () {
    return this.get('activeGroup.upgradeItems') && this.get('activeGroup.upgradeItems').find(function (item) {
      return this.get('failedStatuses').contains(item.get('status'));
    }, this);
  }.property('activeGroup.upgradeItems.@each.status'),

  /**
   * upgrade doesn't have any failed or manual or running item
   * @type {boolean}
   */
  noActiveItem: function () {
    return (Em.isNone(this.get('failedItem')) && Em.isNone(this.get('runningItem')) && Em.isNone(this.get('manualItem'))) &&
      !['INIT', 'COMPLETED'].contains(App.get('upgradeState'));
  }.property('failedItem', 'runningItem', 'manualItem', 'App.upgradeState'),

  /**
   * details of currently active task
   * @type {object|null}
   */
  taskDetails: function () {
    if (this.get('runningItem')) {
      return this.get('runningItem').get('tasks').findProperty('status', 'IN_PROGRESS');
    } else if (this.get('failedItem')) {
      return this.get('failedItem').get('tasks').find(function (task) {
        return this.get('failedStatuses').contains(task.get('status'));
      }, this);
    } else {
      return null;
    }
  }.property('failedItem.tasks.@each.status', 'runningItem.tasks.@each.status'),

  /**
   * indicate whether failed item can be skipped or retried in order to continue Upgrade
   * @type {boolean}
   */
  isHoldingState: function () {
    return Boolean(this.get('failedItem.status') && this.get('failedItem.status').contains('HOLDING'));
  }.property('failedItem.status'),

  /**
   * @type {boolean}
   */
  isManualDone: false,

  /**
   * @type {boolean}
   */
  isManualProceedDisabled: function () {
    return !this.get('isManualDone') || this.get('controller.requestInProgress');
  }.property('isManualDone'),

  /**
   * if upgrade group is manual it should have manual item
   * @type {object|undefined}
   */
  manualItem: function () {
    return this.get('activeGroup.upgradeItems') && this.get('activeGroup.upgradeItems').findProperty('status', 'HOLDING');
  }.property('activeGroup.upgradeItems.@each.status'),

  /**
   * indicate whether the step is Finalize
   * @type {boolean}
   */
  isFinalizeItem: function () {
    return this.get('manualItem.context') === 'Confirm Finalize';
  }.property('manualItem.context'),

  /**
   * label of Upgrade status
   * @type {string}
   */
  upgradeStatusLabel: function() {
    var labelKey = null;
    switch (this.get('controller.upgradeData.Upgrade.request_status')) {
      case 'QUEUED':
      case 'PENDING':
      case 'IN_PROGRESS':
        labelKey = 'admin.stackUpgrade.state.inProgress';
        break;
      case 'COMPLETED':
        labelKey = 'admin.stackUpgrade.state.completed';
        break;
      case 'ABORTED':
      case 'TIMEDOUT':
      case 'FAILED':
      case 'HOLDING_FAILED':
      case 'HOLDING_TIMEDOUT':
      case 'HOLDING':
        labelKey = 'admin.stackUpgrade.state.paused';
        break;
    }
    if (labelKey) {
      labelKey += (this.get('controller.isDowngrade')) ? '.downgrade' : "";
      return Em.I18n.t(labelKey);
    } else {
      return "";
    }
  }.property('controller.upgradeData.Upgrade.request_status', 'controller.isDowngrade'),

  /**
   * toggle details box
   */
  toggleDetails: function () {
    this.toggleProperty('isDetailsOpened');
  },

  /**
   * close details block if no active task present
   */
  closeDetails: function () {
    if (this.get('noActiveItem')) {
      this.set('isDetailsOpened', false);
    }
  }.observes('noActiveItem'),

  /**
   * start polling upgrade data
   */
  startPolling: function () {
    var self = this;
    if (App.get('clusterName')) {
      this.get('controller').loadUpgradeData().done(function () {
        self.set('isLoaded', true);
        self.doPolling();
      });
    }
  }.observes('App.clusterName'),

  /**
   * start polling upgrade data
   */
  willInsertElement: function () {
    this.startPolling();
  },

  /**
   * stop polling upgrade data
   */
  willDestroyElement: function () {
    clearTimeout(this.get('updateTimer'));
    clearTimeout(this.get('upgradeItemTimer'));
    this.set('isLoaded', false);
  },

  /**
   * load upgrade data with time interval
   */
  doPolling: function () {
    var self = this;
    this.set('updateTimer', setTimeout(function () {
      self.get('controller').loadUpgradeData().done(function() {
        self.doPolling();
      });
    }, App.bgOperationsUpdateInterval));
  },

  /**
   * poll for tasks when item is expanded
   */
  doUpgradeItemPolling: function () {
    var self = this;
    var item = this.get('runningItem') || this.get('failedItem');

    if (item && this.get('isDetailsOpened')) {
      this.get('controller').getUpgradeItem(item).complete(function () {
        self.set('upgradeItemTimer', setTimeout(function () {
          self.doUpgradeItemPolling();
        }, App.bgOperationsUpdateInterval));
      });
    } else {
      clearTimeout(this.get('upgradeItemTimer'));
    }
  }.observes('isDetailsOpened'),

  /**
   * set current upgrade item state to FAILED (for HOLDING_FAILED) or TIMED_OUT (for HOLDING_TIMED_OUT)
   * in order to ignore fail and continue Upgrade
   * @param {object} event
   */
  continue: function (event) {
    this.get('controller').setUpgradeItemStatus(event.context, event.context.get('status').slice(8));
    this.set('isDetailsOpened', false);
  },

  /**
   * set current upgrade item state to PENDING in order to retry Upgrade
   * @param {object} event
   */
  retry: function (event) {
    this.get('controller').setUpgradeItemStatus(event.context, 'PENDING');
    this.set('isDetailsOpened', false);
  },

  /**
   * set current upgrade item state to COMPLETED in order to proceed
   * @param {object} event
   */
  complete: function (event) {
    this.get('controller').setUpgradeItemStatus(event.context, 'COMPLETED');
    this.set('isManualDone', false);
  }
});

});

require.register("views/main/admin/stack_upgrade/versions_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');
var stringUtils = require('utils/string_utils');

App.MainAdminStackVersionsView = Em.View.extend({
  templateName: require('templates/main/admin/stack_upgrade/versions'),

  /**
   * update timer
   * @type {number|null}
   * @default null
   */
  updateTimer: null,

  /**
   * Not Installed = the version is not installed or out of sync
   * Upgrade Ready = the version is installed and ready for upgrade
   * Current = the version currently being used
   * Upgrade in Process = UPGRADING
   * Ready to Finalize = UPGRADED
   * Installed = All the versions that are installed BUT cannot be upgraded to (meaning: they are lower than the current)
   * @type {Array}
   */
  categories: [
    Em.Object.create({
      labelKey: 'admin.stackVersions.filter.all',
      value: '',
      isSelected: true
    }),
    Em.Object.create({
      labelKey: 'admin.stackVersions.filter.notInstalled',
      value: 'NOT_INSTALLED',
      isSelected: false
    }),
    Em.Object.create({
      labelKey: 'admin.stackVersions.filter.upgradeReady',
      value: 'UPGRADE_READY',
      isSelected: false
    }),
    Em.Object.create({
      labelKey: 'admin.stackVersions.filter.current',
      value: 'CURRENT',
      isSelected: false
    }),
    Em.Object.create({
      labelKey: 'admin.stackVersions.filter.installed',
      value: 'INSTALLED',
      isSelected: false
    }),
    Em.Object.create({
      labelKey: 'admin.stackVersions.filter.upgrading',
      value: 'UPGRADING',
      isSelected: false
    }),
    Em.Object.create({
      labelKey: 'admin.stackVersions.filter.upgraded',
      value: 'UPGRADED',
      isSelected: false
    })
  ],

  didInsertElement: function () {
    this.observesCategories();
  },

  /**
   * update categories labels
   */
  observesCategories: function () {
    this.get('categories').forEach(function (category) {
      category.set('label', Em.I18n.t(category.labelKey).format(this.filterBy(this.get('repoVersions'), category).length));
    }, this);
    this.filterVersions(this.get('selectedCategory'));
  }.observes('repoVersions.@each.stackVersion.state', 'controller.isLoaded'),

  /**
   * select category
   * @param event
   */
  selectCategory: function (event) {
    this.get('categories').filterProperty('isSelected').setEach('isSelected', false);
    event.context.set('isSelected', true);
    this.filterVersions(event.context);
  },

  /**
   * filter versions that match category
   * @param {Em.Object} category
   */
  filterVersions: function (category) {
    var filteredVersionIds = this.filterBy(this.get('repoVersions'), category).mapProperty('id');
    this.get('repoVersions').forEach(function (version) {
      version.set('isVisible', filteredVersionIds.contains(version.get('id')));
    });
  },

  /**
   * @type {object}
   */
  selectedCategory: function () {
    return this.get('categories').findProperty('isSelected');
  }.property('categories.@each.isSelected'),

  /**
   * @type {Em.Array}
   */
  repoVersions: App.RepositoryVersion.find(),

  /**
   * @type {Em.Array}
   */
  stackVersions: App.StackVersion.find(),

  /**
   * filter versions by category
   * @param versions
   * @param filter
   * @return {Array}
   */
  filterBy: function (versions, filter) {
    var currentVersion = this.get('controller.currentVersion');
    if (filter && filter.get('value')) {
      versions = versions.filter(function (version) {
        var status = version.get('status');
        if (status === 'INSTALLED' && ['UPGRADE_READY', 'INSTALLED'].contains(filter.get('value'))) {
          if (filter.get('value') === 'UPGRADE_READY') {
            return stringUtils.compareVersions(version.get('repositoryVersion'), Em.get(currentVersion, 'repository_version')) === 1;
          } else if (filter.get('value') === 'INSTALLED') {
            return stringUtils.compareVersions(version.get('repositoryVersion'), Em.get(currentVersion, 'repository_version')) < 1;
          }
        } else if (filter.get('value') === 'NOT_INSTALLED') {
          return ['INIT', 'INSTALL_FAILED', 'INSTALLING', 'OUT_OF_SYNC'].contains(status);
        } else {
          return status === filter.get('value');
        }
      }, this);
    }
    if (App.get('supports.displayOlderVersions')) {
      return versions.toArray();
    } else {
      return versions.filter(function(v) {
        return stringUtils.compareVersions(v.get('repositoryVersion'), Em.get(currentVersion, 'repository_version')) >= 0;
      }).toArray();
    }
  },

  /**
   * route to versions in Admin View
   * @return {App.ModalPopup}
   */
  goToVersions: function () {
    return App.showConfirmationPopup(function () {
      window.location.replace('/views/ADMIN_VIEW/2.0.0/INSTANCE/#/stackVersions');
    },
    Em.I18n.t('admin.stackVersions.manageVersions.popup.body'),
    null,
    Em.I18n.t('admin.stackVersions.manageVersions'));
  },

  /**
   * load ClusterStackVersions data
   */
  willInsertElement: function () {
    this.doPolling();
  },

  /**
   * stop polling upgrade state
   */
  willDestroyElement: function () {
    window.clearTimeout(this.get('updateTimer'));
  },

  /**
   * set timer for polling
   */
  doPolling: function () {
    var self = this;
    this.set('updateTimer', window.setTimeout(function () {
      self.poll.apply(self);
    }, App.bgOperationsUpdateInterval));
  },

  /**
   * poll data
   */
  poll: function () {
    var self = this;
    //skip call if Upgrade wizard opened
    if (App.router.get('updateController').get('isWorking')) {
      this.get('controller').load().done(function () {
        self.set('controller.isLoaded', true);
        self.doPolling();
      });
    }
  }

});

});

require.register("views/main/alert_definitions_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var filters = require('views/common/filter_view'),
  sort = require('views/common/sort_view'),
  date = require('utils/date');

App.MainAlertDefinitionsView = App.TableView.extend({

  templateName: require('templates/main/alerts'),

  content: function() {
    return this.get('controller.content') ? this.get('controller.content').toArray().sort(App.AlertDefinition.getSortDefinitionsByStatus(true)) : [];
  }.property('controller.content.@each'),

  willInsertElement: function () {
    if (!this.get('controller.showFilterConditionsFirstLoad')) {
      this.clearFilterCondition();
      this.clearStartIndex();
    }
    this.removeObserver('filteredCount', this, 'updatePaging');
    this.removeObserver('displayLength', this, 'updatePaging');
    var startIndex = App.db.getStartIndex(this.get('controller.name'));
    this._super();
    this.set('startIndex', startIndex ? startIndex : 1);
    this.addObserver('filteredCount', this, 'updatePaging');
    this.addObserver('displayLength', this, 'updatePaging');
  },

  /**
   * Method is same as in the parentView, but observes are set in the <code>willInsertElement</code>
   * and not in the declaration
   */
  updatePaging: function () {
    this._super();
  },

  didInsertElement: function () {
    var self = this;
    Em.run.next(function () {
      self.set('isInitialRendering', false);
      self.tooltipsUpdater();
      self.addObserver('pageContent.length', self, 'tooltipsUpdater');
    });
  },

  willDestroyElement: function () {
    this.removeObserver('pageContent.length', this, 'tooltipsUpdater');
  },

  /**
   * Save <code>startIndex</code> to the localStorage
   * @method saveStartIndex
   */
  saveStartIndex: function() {
    App.db.setStartIndex(this.get('controller.name'), this.get('startIndex'));
  }.observes('startIndex'),

  /**
   * Clear <code>startIndex</code> from the localStorage
   * @method clearStartIndex
   */
  clearStartIndex: function () {
    App.db.setStartIndex(this.get('controller.name'), null);
  },

  /**
   * @type {number}
   */
  totalCount: function () {
    return this.get('content.length');
  }.property('content.length'),

  colPropAssoc: ['', 'label', 'summary', 'serviceName', 'type', 'lastTriggered', 'enabled', 'groups'],

  /**
   * @type {string}
   */
  enabledTooltip: Em.I18n.t('alerts.table.state.enabled.tooltip'),

  /**
   * @type {string}
   */
  disabledTooltip: Em.I18n.t('alerts.table.state.disabled.tooltip'),

  /**
   * @type {string}
   */
  enabledDisplay: Em.I18n.t('alerts.table.state.enabled'),

  /**
   * @type {string}
   */
  disabledDisplay: Em.I18n.t('alerts.table.state.disabled'),

  sortView: sort.wrapperView,

  /**
   * Define whether initial view rendering has finished
   * @type {Boolean}
   */
  isInitialRendering: true,

  /**
   * Sorting header for <label>alertDefinition.label</label>
   * @type {Em.View}
   */
  nameSort: sort.fieldView.extend({
    column: 1,
    name: 'label',
    displayName: Em.I18n.t('alerts.table.header.definitionName')
  }),

  /**
   * Sorting header for <label>alertDefinition.status</label>
   * @type {Em.View}
   */
  statusSort: sort.fieldView.extend({
    column: 2,
    name: 'summary',
    displayName: Em.I18n.t('common.status'),
    type: 'alert_status',
    status: 'sorting_desc'
  }),

  /**
   * Sorting header for <label>alertDefinition.service.serviceName</label>
   * @type {Em.View}
   */
  serviceSort: sort.fieldView.extend({
    column: 3,
    name: 'service.serviceName',
    displayName: Em.I18n.t('common.service'),
    type: 'string'
  }),

  /**
   * Sorting header for <label>alertDefinition.type</label>
   * @type {Em.View}
   */
  typeSort: sort.fieldView.extend({
    column: 4,
    name: 'type',
    displayName: Em.I18n.t('common.type'),
    type: 'string'
  }),

  /**
   * Sorting header for <label>alertDefinition.lastTriggeredSort</label>
   * @type {Em.View}
   */
  lastTriggeredSort: sort.fieldView.extend({
    column: 5,
    name: 'lastTriggered',
    displayName: Em.I18n.t('alerts.table.header.lastTriggered'),
    type: 'number'
  }),

  /**
   * Sorting header for <label>alertDefinition.enabled</label>
   * @type {Em.View}
   */
  enabledSort: sort.fieldView.extend({
    template:Em.Handlebars.compile('<span {{bindAttr class="view.status :column-name"}}>{{t alerts.table.state}}</span>'),
    column: 6,
    name: 'enabled'
  }),

  /**
   * Filtering header for <label>alertDefinition.label</label>
   * @type {Em.View}
   */
  nameFilterView: filters.createTextView({
    column: 1,
    fieldType: 'filter-input-width',
    onChangeValue: function(){
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'string');
    }
  }),

  /**
   * Filtering header for <label>alertDefinition.status</label>
   * @type {Em.View}
   */
  stateFilterView: filters.createSelectView({
    column: 2,
    fieldType: 'filter-input-width',
    content: [
      {
        value: '',
        label: Em.I18n.t('common.all')
      },
      {
        value: 'OK',
        label: 'OK'
      },
      {
        value: 'WARNING',
        label: 'WARNING'
      },
      {
        value: 'CRITICAL',
        label: 'CRITICAL'
      },
      {
        value: 'UNKNOWN',
        label: 'UNKNOWN'
      },
      {
        value: 'PENDING',
        label: 'NONE'
      }
    ],
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'alert_status');
    }
  }),

  /**
   * Filtering header for <label>alertDefinition.service.serviceName</label>
   * @type {Em.View}
   */
  serviceFilterView: filters.createSelectView({
    column: 3,
    fieldType: 'filter-input-width',
    content: function () {
      return [
        {
          value: '',
          label: Em.I18n.t('common.all')
        }
      ].concat(App.Service.find().map(function (service) {
        return {
          value: service.get('serviceName'),
          label: service.get('displayName')
        }
      }).concat([
        {
          value: 'AMBARI',
          label: Em.I18n.t('app.name')
        }
      ]));
    }.property('App.router.clusterController.isLoaded'),
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'select');
    }
  }),

  /**
   * Filtering header for <label>alertDefinition.type</label>
   * @type {Em.View}
   */
  typeFilterView: filters.createSelectView({
    column: 4,
    fieldType: 'filter-input-width',
    content: [
      {
        value: '',
        label: Em.I18n.t('common.all')
      },
      {
        value: 'SCRIPT',
        label: 'SCRIPT'
      },
      {
        value: 'WEB',
        label: 'WEB'
      },
      {
        value: 'PORT',
        label: 'PORT'
      },
      {
        value: 'METRIC',
        label: 'METRIC'
      },
      {
        value: 'AGGREGATE',
        label: 'AGGREGATE'
      }
    ],
    onChangeValue: function(){
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'select');
    }
  }),

  /**
   * Filtering header for <label>alertDefinition.lastTriggered</label>
   * @type {Em.View}
   */
  triggeredFilterView: filters.createSelectView({
    column: 5,
    appliedEmptyValue: ["", ""],
    fieldType: 'filter-input-width,modified-filter',
    content: [
      {
        value: 'Any',
        label: Em.I18n.t('any')
      },
      {
        value: 'Past 1 hour',
        label: 'Past 1 hour'
      },
      {
        value: 'Past 1 Day',
        label: 'Past 1 Day'
      },
      {
        value: 'Past 2 Days',
        label: 'Past 2 Days'
      },
      {
        value: 'Past 7 Days',
        label: 'Past 7 Days'
      },
      {
        value: 'Past 14 Days',
        label: 'Past 14 Days'
      },
      {
        value: 'Past 30 Days',
        label: 'Past 30 Days'
      }
    ],
    emptyValue: 'Any',
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'date');
    }
  }),

  /**
   * Filtering header for <label>alertDefinition.enabled</label>
   * @type {Em.View}
   */
  enabledFilterView:  filters.createSelectView({
    column: 6,
    fieldType: 'filter-input-width',
    content: [
      {
        value: '',
        label: Em.I18n.t('common.all')
      },
      {
        value: 'enabled',
        label: Em.I18n.t('alerts.table.state.enabled')
      },
      {
        value: 'disabled',
        label: Em.I18n.t('alerts.table.state.disabled')
      }
    ],
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'enable_disable');
    }
  }),

  /**
   * Filtering header for <label>alertDefinition</label> groups
   * @type {Em.View}
   */
  alertGroupFilterView: filters.createSelectView({
    column: 7,
    fieldType: 'filter-input-width',
    template: Ember.Handlebars.compile(
      '<div class="btn-group display-inline-block">' +
        '<a class="btn dropdown-toggle" data-toggle="dropdown" href="#">' +
          '<span class="filters-label">Groups:&nbsp;&nbsp;</span>' +
          '<span>{{view.selected.label}}&nbsp;<span class="caret"></span></span>' +
        '</a>' +
          '<ul class="dropdown-menu">' +
            '{{#each category in view.content}}' +
              '<li {{bindAttr class=":category-item category.selected:active"}}>' +
                '<a {{action selectCategory category target="view"}} href="#">' +
                   '<span {{bindAttr class="category.class"}}></span>{{category.label}}</a>' +
              '</li>'+
            '{{/each}}' +
          '</ul>'+
      '</div>'
    ),
    content: [],

    didInsertElement: function() {
      this._super();
      this.updateContent();
      this.set('value', '');
    },

    emptyValue: '',

    /**
     * Update list of <code>App.AlertGroup</code> used in the filter
     * @method updateContent
     */
    updateContent: function() {
      var defaultGroups = [];
      var customGroups = [];
      App.AlertGroup.find().forEach(function (group) {
        var item = Em.Object.create({
          value: group.get('id'),
          label: group.get('displayNameDefinitions')
        });
        if (group.get('default')) {
          defaultGroups.push(item);
        } else {
          customGroups.push(item);
        }
      });
      defaultGroups = defaultGroups.sortProperty('label');
      customGroups = customGroups.sortProperty('label');

      this.set('content', [
        Em.Object.create({
          value: '',
          label: Em.I18n.t('common.all') + ' (' + this.get('parentView.controller.content.length') + ')'
        })
      ].concat(defaultGroups, customGroups));
      this.onValueChange();
    }.observes('App.router.clusterController.isLoaded', 'App.router.manageAlertGroupsController.changeTrigger'),

    selectCategory: function (event) {
      var category = event.context;
      this.set('value', category.value);
      this.get('parentView').updateFilter(this.get('column'), category.value, 'alert_group');
    },

    onValueChange: function () {
      var value = this.get('value');
      if (value != undefined) {
        this.get('content').setEach('selected', false);
        this.set('selected', this.get('content').findProperty('value', value));
        var selectEntry = this.get('content').findProperty('value', value);
        if (selectEntry) {
          selectEntry.set('selected', true);
        }
        this.get('parentView').updateFilter(this.get('column'), value, 'alert_group');
      } else {
        this.set('value', '');
        this.get('parentView').updateFilter(this.get('column'), '', 'alert_group');
      }
    }.observes('value')
  }),

  /**
   * Filtered number of all content number information displayed on the page footer bar
   * @returns {String}
   */
  filteredContentInfo: function () {
    return this.t('alerts.filters.filteredAlertsInfo').format(this.get('filteredCount'), this.get('totalCount'));
  }.property('filteredCount', 'totalCount'),

  /**
   * Determines how display "back"-link - as link or text
   * @type {string}
   */
  paginationLeftClass: function () {
    if (this.get("startIndex") > 1) {
      return "paginate_previous";
    }
    return "paginate_disabled_previous";
  }.property("startIndex", 'filteredCount'),

  /**
   * Determines how display "next"-link - as link or text
   * @type {string}
   */
  paginationRightClass: function () {
    if ((this.get("endIndex")) < this.get("filteredCount")) {
      return "paginate_next";
    }
    return "paginate_disabled_next";
  }.property("endIndex", 'filteredCount'),

  /**
   * Show previous-page if user not in the first page
   * @method previousPage
   */
  previousPage: function () {
    if (this.get('paginationLeftClass') === 'paginate_previous') {
      this._super();
    }
  },

  /**
   * Show next-page if user not in the last page
   * @method nextPage
   */
  nextPage: function () {
    if (this.get('paginationRightClass') === 'paginate_next') {
      this._super();
    }
  },

  /**
   * Update tooltips when <code>pageContent</code> is changed
   * @method tooltipsUpdater
   */
  tooltipsUpdater: function () {
    Em.run.next(this, function () {
      App.tooltip($(".enable-disable-button, .timeago"));
    });
  },

  updateFilter: function (iColumn, value, type) {
    if (!this.get('isInitialRendering')) {
      this._super(iColumn, value, type);
    }
  }

});
});

require.register("views/main/alerts", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var filters = require('views/common/filter_view');
var sort = require('views/common/sort_view');
var date = require('utils/date');

App.MainAlertsView = App.TableView.extend(App.TableServerViewMixin, {
  templateName:require('templates/main/alerts'),

  tableName: 'Alerts'
});

});

require.register("views/main/alerts/add_alert_definition/add_alert_definition_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.AddAlertDefinitionView = Em.View.extend(App.WizardMenuMixin, {

  templateName: require('templates/main/alerts/add_alert_definition/add_alert_definition')

});

});

require.register("views/main/alerts/add_alert_definition/step1_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.AddAlertDefinitionStep1View = Em.View.extend({

  templateName: require('templates/main/alerts/add_alert_definition/step1'),

  didInsertElement: function() {
    this.get('controller').loadStep();
  }

});

});

require.register("views/main/alerts/add_alert_definition/step2_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.AddAlertDefinitionStep2View = Em.View.extend({

  templateName: require('templates/main/alerts/add_alert_definition/step2')

});

});

require.register("views/main/alerts/add_alert_definition/step3_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.AddAlertDefinitionStep3View = Em.View.extend({

  /**
   * @type {string}
   */
  alertDefinitionToDisplay: '',

  willInsertElement: function () {
    this.set('alertDefinitionToDisplay', JSON.stringify(this.get('controller.content.formattedToRequestConfigs'),  null, 4));
  },

  templateName: require('templates/main/alerts/add_alert_definition/step3')

});

});

require.register("views/main/alerts/alert_definitions_actions_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

App.MainAlertDefinitionActionsView = Em.View.extend({

  templateName: require('templates/main/alerts/alert_definitions_actions')

});
});

require.register("views/main/alerts/definition_configs_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.AlertDefinitionConfigsView = Em.View.extend({

  controllerBinding: 'App.router.mainAlertDefinitionConfigsController',

  templateName: require('templates/main/alerts/configs'),

  /**
   * Define whether configs are editable
   * is set in template
   * @type {Boolean}
   */
  canEdit: true,

  /**
   * List of classes applied to all inputs
   * @type {String}
   */
  basicClass: 'span9',

  init: function () {
    this.set('controller.canEdit', this.get('canEdit'));
    this.set('controller.isWizard', this.get('isWizard'));
    this.set('controller.alertDefinitionType', this.get('alertDefinitionType'));
    this.set('controller.content', this.get('content'));
    this.get('controller').renderConfigs();
    this._super();
  },

  errorMessage: Em.I18n.t('alerts.definition.details.configs.thresholdsErrorMsg')

});

App.AlertConfigTextFieldView = Em.View.extend({
  templateName: require('templates/main/alerts/configs/alert_config_text_field'),
  classNameBindings: ['property.classNames', 'parentView.basicClass']
});

App.AlertConfigTextAreaView = Em.TextArea.extend({
  valueBinding: 'property.value',
  disabledBinding: 'property.isDisabled',
  classNameBindings: ['property.classNames', 'parentView.basicClass']
});

App.AlertConfigSelectView = Em.Select.extend({
  attributeBindings: ['disabled'],
  selectionBinding: 'property.value',
  disabledBinding: 'property.isDisabled',
  contentBinding: 'property.options',
  classNameBindings: ['property.classNames', 'parentView.basicClass']
});

App.AlertConfigThresholdView = Em.View.extend({
  templateName: require('templates/main/alerts/configs/alert_config_threshold'),
  classNameBindings: ['property.classNames', 'parentView.basicClass']
});

App.AlertConfigRadioButtonView = Em.Checkbox.extend({
  attributeBindings: ['type', 'name', 'value', 'checked', 'disabled'],
  type: 'radio',
  nameBinding: 'property.group',
  checkedBinding: 'property.value',

  change: function () {
    this.set('property.value', true);
    this.get('parentView.controller.configs').filterProperty('group', this.get('name')).without(this.get('property')).setEach('value', false);
    this.get('parentView.controller').changeType(this.get('property.name'));
  },

  classNameBindings: ['property.classNames']
});

});

require.register("views/main/alerts/definition_details_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAlertDefinitionDetailsView = App.TableView.extend({

  templateName: require('templates/main/alerts/definition_details'),

  /**
   * Determines if <code>controller.content</code> is loaded
   * @type {bool}
   */
  isLoaded: false,

  /**
   * @type {string}
   */
  enabledDisplay: Em.I18n.t('alerts.table.state.enabled'),

  /**
   * @type {string}
   */
  disabledDisplay: Em.I18n.t('alerts.table.state.disabled'),

  content: function () {
    return this.get('controller.alerts');
  }.property('controller.alerts.@each'),

  willInsertElement: function () {
    this._super();
    this.get('controller').clearStep();
    var self = this,
      updater = App.router.get('updateController');
    if (self.get('controller.content.isLoaded')) {
      self.set('isLoaded', true);
      self.get('controller').loadAlertInstances();
    }
    else {
      updater.updateAlertGroups(function () {
        updater.updateAlertDefinitions(function () {
          updater.updateAlertDefinitionSummary(function () {
            self.set('isLoaded', true);
            // App.AlertDefinition doesn't represents real models
            // Real model (see AlertDefinition types) should be used
            self.set('controller.content', App.AlertDefinition.find().findProperty('id', parseInt(self.get('controller.content.id'))));
            self.get('controller').loadAlertInstances();
          });
        });
      });
    }
  },

  didInsertElement: function () {
    this.filter();
    this.tooltipsUpdater();
  },

  /**
   * Update tooltips when <code>pageContent</code> is changed
   * @method tooltipsUpdater
   */
  tooltipsUpdater: function () {
    Em.run.next(function () {
      App.tooltip($(".enable-disable-button"));
    });
  }.observes('controller.content.enabled'),

  /**
   * View calculates and represents count of alerts on appropriate host during last day
   */
  lastDayCount: Em.View.extend({
    hostName: '', // binding from template
    template: Ember.Handlebars.compile('<span>{{view.count}}</span>'),
    count: function () {
      var lastDayAlertsCount = this.get('parentView.controller.lastDayAlertsCount');
      return lastDayAlertsCount ? lastDayAlertsCount[this.get('hostName')] || 0 : Em.I18n.t('app.loadingPlaceholder');
    }.property('parentView.controller.lastDayAlertsCount', 'hostName')
  }),

  /**
   * View represents each row of instances table
   */
  instanceTableRow: Em.View.extend({
    tagName: 'tr',
    didInsertElement: function () {
      App.tooltip($("[rel=tooltip]"));
    }
  }),

  paginationLeftClass: function () {
    if (this.get("startIndex") > 1) {
      return "paginate_previous";
    }
    return "paginate_disabled_previous";
  }.property("startIndex", 'filteredCount'),

  /**
   * Determines how display "next"-link - as link or text
   * @type {string}
   */
  paginationRightClass: function () {
    if ((this.get("endIndex")) < this.get("filteredCount")) {
      return "paginate_next";
    }
    return "paginate_disabled_next";
  }.property("endIndex", 'filteredCount'),

  /**
   * Show previous-page if user not in the first page
   * @method previousPage
   */
  previousPage: function () {
    if (this.get('paginationLeftClass') === 'paginate_previous') {
      this._super();
    }
  },

  /**
   * Show next-page if user not in the last page
   * @method nextPage
   */
  nextPage: function () {
    if (this.get('paginationRightClass') === 'paginate_next') {
      this._super();
    }
  }

});

});

require.register("views/main/alerts/manage_alert_groups/select_definitions_popup_body_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.SelectDefinitionsPopupBodyView = App.TableView.extend({

  templateName: require('templates/main/alerts/add_definition_to_group_popup'),

  controllerBinding: 'App.router.manageAlertGroupsController',

  isPaginate: true,

  filteredContent: function () {
    return this.get('parentView.availableDefs').filterProperty('filtered') || [];
  }.property('parentView.availableDefs.@each.filtered'),

  showOnlySelectedDefs: false,

  filterComponent: null,

  filterService: null,

  isDisabled: function () {
    return !this.get('parentView.isLoaded');
  }.property('parentView.isLoaded'),

  didInsertElement: function () {
    var initialDefs = this.get('initialDefs');
    initialDefs.setEach('filtered', true);
    this.set('parentView.availableDefs', initialDefs);
    this.set('parentView.isLoaded', true);
  },

  filterDefs: function () {
    var showOnlySelectedDefs = this.get('showOnlySelectedDefs');
    var filterComponent = this.get('filterComponent');
    var filterService = this.get('filterService');
    this.get('parentView.availableDefs').forEach(function (defObj) {
      var componentOnObj = true;
      var serviceOnObj = true;
      if (filterComponent) {
        componentOnObj = (defObj.componentName == filterComponent.get('componentName'));
      }
      if (defObj.serviceName && filterService) {
        serviceOnObj = (defObj.serviceName == filterService.get('serviceName'));
      }
      defObj.set('filtered', showOnlySelectedDefs ? (componentOnObj && serviceOnObj && defObj.get('selected')) : (componentOnObj && serviceOnObj));
    }, this);
    this.set('startIndex', 1);
  }.observes('parentView.availableDefs', 'filterService', 'filterService.serviceName', 'filterComponent', 'filterComponent.componentName', 'showOnlySelectedDefs'),

  defSelectMessage: function () {
    var defs = this.get('parentView.availableDefs');
    var selectedDefs = defs.filterProperty('selected', true);
    return this.t('alerts.actions.manage_alert_groups_popup.selectDefsDialog.selectedDefsLink').format(selectedDefs.get('length'), defs.get('length'));
  }.property('parentView.availableDefs.@each.selected'),

  selectFilterComponent: function (event) {
    if (event != null && event.context != null && event.context.componentName != null) {
      var currentFilter = this.get('filterComponent');
      if (currentFilter != null) {
        currentFilter.set('selected', false);
      }
      if (currentFilter != null && currentFilter.componentName === event.context.componentName) {
        // selecting the same filter deselects it.
        this.set('filterComponent', null);
      } else {
        this.set('filterComponent', event.context);
        event.context.set('selected', true);
      }
    }
  },

  selectFilterService: function (event) {
    if (event != null && event.context != null && event.context.serviceName != null) {
      var currentFilter = this.get('filterService');
      if (currentFilter != null) {
        currentFilter.set('selected', false);
      }
      if (currentFilter != null && currentFilter.serviceName === event.context.serviceName) {
        // selecting the same filter deselects it.
        this.set('filterService', null);
      } else {
        this.set('filterService', event.context);
        event.context.set('selected', true);
      }
    }
  },

  /**
   * Determines if all alert definitions are selected
   * @type {boolean}
   */
  allDefsSelected: false,

  /**
   * Inverse selection for alert definitions
   * @method toggleSelectAllDefs
   */
  toggleSelectAllDefs: function () {
    this.get('parentView.availableDefs').filterProperty('filtered').setEach('selected', this.get('allDefsSelected'));
  }.observes('allDefsSelected'),

  toggleShowSelectedDefs: function () {
    var filter1 = this.get('filterComponent');
    if (filter1 != null) {
      filter1.set('selected', false);
    }
    var filter2 = this.get('filterService');
    if (filter2 != null) {
      filter2.set('selected', false);
    }
    this.set('filterComponent', null);
    this.set('filterService', null);
    this.set('showOnlySelectedDefs', !this.get('showOnlySelectedDefs'));
  }
});

});

require.register("views/main/alerts/manage_alert_groups_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainAlertsManageAlertGroupView = Em.View.extend({

  templateName: require('templates/main/alerts/manage_alert_groups_popup'),

  /**
   * @type {object}
   */
  selectedAlertGroup: null,

  /**
   * Observer to prevent user select more than 1 alert group
   * @method onGroupSelect
   */
  onGroupSelect: function () {
    var selectedAlertGroup = this.get('selectedAlertGroup');
    // to unable user select more than one alert group at a time
    if (selectedAlertGroup && selectedAlertGroup.length) {
      this.set('controller.selectedAlertGroup', selectedAlertGroup[selectedAlertGroup.length - 1]);
    }
    if (selectedAlertGroup && selectedAlertGroup.length > 1) {
      this.set('selectedAlertGroup', selectedAlertGroup[selectedAlertGroup.length - 1]);
    }
    this.set('controller.selectedDefinitions', []);
  }.observes('selectedAlertGroup'),

  /**
   * Observer to provide two-way binding for group selection
   */
  setGroupInController: function () {
    if (!this.get('controller.isLoaded')) return;
    this.set('selectedAlertGroup', [this.get('controller.selectedAlertGroup')]);
  }.observes('controller.selectedAlertGroup'),

  /**
   * Select first alert group when all groups are loaded
   * @method onLoad
   */
  onLoad: function () {
    if (this.get('controller.isLoaded')) {
      this.setTooltips();
    }
  }.observes('controller.isLoaded'),

  /**
   * Load alert groups, definitions and notification to have all new data
   * Useful if user delete some definition and immediately open "Manage Alert Groups" popup
   * @method willInsertElement
   */
  willInsertElement: function() {
    this.get('controller').loadAlertNotifications();
  },

  /**
   * Add tooltips and try to select first alert group
   * @method didInsertElement
   */
  didInsertElement: function () {
    this.onLoad();
  },

  /**
   * Setup tooltips for view elements
   */
  setTooltips: function () {
    Em.run.next(function(){
      App.tooltip($("[rel='button-info']"));
      App.tooltip($("[rel='button-info-dropdown']"), {placement: 'left'});
    });
  },

  /**
   * Tooltip for "Add group"-button
   * @type {string}
   */
  addButtonTooltip: Em.I18n.t('alerts.actions.manage_alert_groups_popup.addButton'),

  /**
   * Tooltip for "Remove group"-button
   * @type {string}
   */
  removeButtonTooltip: function () {
    return this.get('controller.isRemoveButtonDisabled') ? Em.I18n.t('alerts.actions.manage_alert_groups_popup.removeButtonDisabled') :
        Em.I18n.t('alerts.actions.manage_alert_groups_popup.removeButton');
  }.property('controller.isRemoveButtonDisabled'),

  /**
   * Tooltip for "Rename"-button
   * @type {string}
   */
  renameButtonTooltip: Em.I18n.t('alerts.actions.manage_alert_groups_popup.renameButton'),

  /**
   * Tooltip for "Duplicate group"-button
   * @type {string}
   */
  duplicateButtonTooltip: Em.I18n.t('alerts.actions.manage_alert_groups_popup.duplicateButton'),

  /**
   * Tooltip for "Add definition to group"-button
   * @type {string}
   */
  addDefinitionTooltip: function () {
    if (this.get('controller.selectedAlertGroup.default')) {
      return Em.I18n.t('alerts.actions.manage_alert_groups_popup.addDefinitionToDefault');
    }
    else
      if (this.get('controller.selectedAlertGroup.isAddDefinitionsDisabled')) {
        return Em.I18n.t('alerts.actions.manage_alert_groups_popup.addDefinitionDisabled');
      }
      else {
        return  Em.I18n.t('alerts.actions.manage_alert_groups_popup.addDefinition');
      }
  }.property('controller.selectedAlertGroup.default', 'controller.selectedAlertGroup.isAddDefinitionsDisabled'),

  /**
   * Tooltip for "Remove definition from group"-button
   * @type {string}
   */
  removeDefinitionTooltip: function () {
    if (this.get('controller.selectedAlertGroup.default')) {
      return Em.I18n.t('alerts.actions.manage_alert_groups_popup.removeDefinitionDisabled');
    }
    else
      if (this.get('controller.isDeleteDefinitionsDisabled')) {
        return 'Nothing to delete';
      } else {
        return Em.I18n.t('alerts.actions.manage_alert_groups_popup.removeDefinition');
      }
  }.property('controller.selectedAlertGroup.default', 'controller.isDeleteDefinitionsDisabled')

});

});

require.register("views/main/alerts/manage_alert_notifications_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ManageAlertNotificationsView = Em.View.extend({

  templateName: require('templates/main/alerts/manage_alert_notifications_popup'),

  /**
   * @type {App.AlertNotification}
   */
  selectedAlertNotification: null,

  selectedAlertNotificationGroups: function () {
    return this.get('controller.selectedAlertNotification.groups').toArray().mapProperty('displayName').join(', ');
  }.property('controller.selectedAlertNotification', 'controller.selectedAlertNotification.groups.@each', 'controller.isLoaded'),

  isEditButtonDisabled: true,

  isRemoveButtonDisabled: true,

  isDuplicateButtonDisabled: true,

  /**
   * Show EMAIL information if selected alert notification has type EMAIL
   * @type {boolean}
   */
  showEmailDetails: function () {
    return this.get('controller.selectedAlertNotification.type') === 'EMAIL';
  }.property('controller.selectedAlertNotification.type'),

  /**
   * Show SNMP information if selected alert notification has type SNMP
   * @type {boolean}
   */
  showSNMPDetails: function () {
    return this.get('controller.selectedAlertNotification.type') === 'SNMP';
  }.property('controller.selectedAlertNotification.type'),

  email: function () {
    return this.get('controller.selectedAlertNotification.properties')['ambari.dispatch.recipients'];
  }.property('controller.selectedAlertNotification'),

  severities: function () {
    return this.get('controller.selectedAlertNotification.alertStates').join(', ');
  }.property('controller.selectedAlertNotification'),

  /**
   * Enable/disable "edit"/"remove"/"duplicate" buttons basing on <code>controller.selectedAlertNotification</code>
   * @method buttonObserver
   */
  buttonObserver: function () {
    var selectedAlertNotification = this.get('controller.selectedAlertNotification');
    this.set('isEditButtonDisabled', !selectedAlertNotification);
    this.set('isRemoveButtonDisabled', !selectedAlertNotification);
    this.set('isDuplicateButtonDisabled', !selectedAlertNotification);
  }.observes('controller.selectedAlertNotification'),

  /**
   * Prevent user select more than 1 alert notification
   * @method onAlertNotificationSelect
   */
  onAlertNotificationSelect: function () {
    var selectedAlertNotification = this.get('selectedAlertNotification');
    if (selectedAlertNotification && selectedAlertNotification.length) {
      this.set('controller.selectedAlertNotification', selectedAlertNotification[selectedAlertNotification.length - 1]);
    }
    if (selectedAlertNotification && selectedAlertNotification.length > 1) {
      this.set('selectedAlertNotification', selectedAlertNotification[selectedAlertNotification.length - 1]);
    }
  }.observes('selectedAlertNotification'),

  /**
   * Set first alert notification as selected (if they are already loaded)
   * Add some tooltips on manage buttons
   * @method onLoad
   */
  onLoad: function () {
    if (this.get('controller.isLoaded')) {
      var notifications = this.get('controller.alertNotifications');
      if (notifications && notifications.length) {
        this.set('selectedAlertNotification', this.get('controller.selectedAlertNotification') || notifications[0]);
        this.buttonObserver();
      }  else {
        this.set('selectedAlertNotification', null);
      }
      Em.run.later(this, function () {
        App.tooltip(this.$("[rel='button-info']"));
        App.tooltip(this.$("[rel='button-info-dropdown']"), {placement: 'left'});
      }, 50) ;
    }
  }.observes('controller.isLoaded'),

  willInsertElement: function () {
    this.get('controller').loadAlertNotifications();
  },

  didInsertElement: function () {
    this.onLoad();
  }

});

});

require.register("views/main/charts", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainChartsView = Em.View.extend({
    templateName: require('templates/main/charts')
});
});

require.register("views/main/charts/heatmap", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.MainChartsHeatmapView = Em.View.extend({
  templateName: require('templates/main/charts/heatmap'),
  spinner: null,
  didInsertElement: function () {
    this._super();
    // set default metric
    this.set('controller.selectedMetric', this.get('controller.allMetrics')[0].get('items')[0]);
    $("#heatmapDetailsBlock").hide();
  },
  /**
   * show spinner while loading selected metric
   */
  showLoading: function () {
    if (this.get('controller.selectedMetric.loading')) {
      var e = document.getElementById("heatmap-metric-loading");
      if (e) {
        $(e).children('div.spinner').remove();
        var spinOpts = {
          lines: 9,
          length: 4,
          width: 2,
          radius: 3,
          top: '0',
          left: '0'
        };
        this.set('spinner', new Spinner(spinOpts).spin(e));
      }
    } else {
      var spinner = this.get('spinner');
      if (spinner) {
        spinner.stop();
      }
      this.set('spinner', null);
    }
  }.observes('controller.selectedMetric.loading')
});
});

require.register("views/main/charts/heatmap/heatmap_host", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var date = require('utils/date');

var App = require('app');

App.MainChartsHeatmapHostView = Em.View.extend({
  templateName: require('templates/main/charts/heatmap/heatmap_host'),
  /** @private */
  hostClass: 'hostBlock',

  /**
   * link to host record in App.Host model
   */
  hostModelLink: function () {
    return App.Host.find(this.get('content.hostName'));
  }.property('content.hostName'),

  /**
   * show Host details block and move it near the cursor
   *
   * @param {Object} event
   * @this App.MainChartsHeatmapHostView
   */
  mouseEnter: function (event) {
    var host = this.get('content');
    var view = App.MainChartsHeatmapHostDetailView.create();
    var self = this;
    var nonClientComponents = App.get('components.slaves').concat(App.get('components.masters'));

    $.each(view.get('details'), function (i) {
      var val = host[i];

      switch (i) {
        case 'diskUsage':
          val = self.getUsage(host, 'diskTotal', 'diskFree');
          break;
        case 'cpuUsage':
          val = 0;
          if (Number.isFinite(host.cpuSystem) && Number.isFinite(host.cpuUser)) {
            val = host.cpuSystem + host.cpuUser;
          }
          val = val.toFixed(1);
          break;
        case 'memoryUsage':
          val = self.getUsage(host, 'memTotal', 'memFree');
          break;
        case 'hostComponents':
          val = [];
          host[i].forEach(function (componentName) {
            if (nonClientComponents.contains(componentName)) {
              val.push(App.format.role(componentName));
            }
          }, this);
          val = val.join(', ')
      }

      view.set('details.' + i, val);
    });
    this.setMetric(view, host);
    this.openDetailsBlock(event);
  },

  /**
   * show tooltip with host's details
   */
  openDetailsBlock: function (event) {
    var detailsBlock = $("#heatmapDetailsBlock");

    detailsBlock.css('top', event.pageY + 10);
    detailsBlock.css('left', event.pageX + 10);
    detailsBlock.show();
  },

  /**
   * set name and value of selected metric
   * @param view
   * @param host
   */
  setMetric: function (view, host) {
    var selectedMetric = this.get('controller.selectedMetric');

    if (selectedMetric) {
      var metricName = selectedMetric.get('name');
      var h2vMap = selectedMetric.get('hostToValueMap');
      if (h2vMap && metricName) {
        var value = h2vMap[host.hostName];
        if (Em.isNone(value)) {
          value = this.t('charts.heatmap.unknown');
        } else {
          if (metricName == 'Garbage Collection Time') {
            value = date.timingFormat(parseInt(value));
          } else {
            if (isNaN(value)) {
              value = this.t('charts.heatmap.unknown');
            } else {
              value = value + selectedMetric.get('units');
            }
          }
        }
        view.set('details.metricName', metricName);
        view.set('details.metricValue', value);
      }
    }
  },
  /**
   * get relative usage of metric in percents
   * @param item
   * @param totalProperty
   * @param freeProperty
   * @return {String}
   */
  getUsage: function (item, totalProperty, freeProperty) {
    var result = 0;
    var total = item[totalProperty];

    if (Number.isFinite(total) && Number.isFinite(item[freeProperty]) && total > 0) {
      result = ((total - item[freeProperty]) / total) * 100;
    }
    return result.toFixed(1);
  },

  /**
   * hide Host details block
   *
   * @param {Object} e
   * @this App.MainChartsHeatmapHostView
   */
  mouseLeave: function (e) {
    $("#heatmapDetailsBlock").hide();
  },

  hostTemperatureStyle: function () {
    var controller = this.get('controller');
    var h2sMap = controller.get('hostToSlotMap');
    if (h2sMap) {
      var hostname = this.get('content.hostName');
      if (hostname) {
        var slot = h2sMap[hostname];
        if (slot > -1) {
          var slotDefs = controller.get('selectedMetric.slotDefinitions');
          if (slotDefs && slotDefs.length > slot) {
            return slotDefs[slot].get('cssStyle');
          }
        }
      }
    }
    return '';
  }.property('controller.hostToSlotMap')
});
});

require.register("views/main/charts/heatmap/heatmap_host_detail", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainChartsHeatmapHostDetailView = Em.View.extend({
  templateName: require('templates/main/charts/heatmap/heatmap_host_detail'),
  /** @private */ classNames:['heatmap_host_details'],
  /** @private */ elementId:'heatmapDetailsBlock',
  /** @private */ details:{
    hostName:'test node',
    publicHostName:'test node',
    osType: 'OS',
    ip: '192.168.0.0',
    metricName: 'metric-name',
    metricValue: 'metric-value',
    diskUsage: '10',
    cpuUsage: '10',
    memoryUsage: '10',
    hostComponents: 'host components'
  }
});
});

require.register("views/main/charts/heatmap/heatmap_rack", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var lazyloading = require('utils/lazy_loading');

App.MainChartsHeatmapRackView = Em.View.extend({
  templateName: require('templates/main/charts/heatmap/heatmap_rack'),
  classNames: ['rack'],
  classNameBindings: ['visualSchema'],

  /** rack status block class */
  statusIndicator: 'statusIndicator',
  /** loaded hosts of rack */
  hosts: [],

  willDestroyElement: function () {
    this.get('hosts').clear();
  },

  /**
   * get hosts from server
   */
  getHosts: function () {
    App.ajax.send({
      name: 'hosts.heatmaps',
      sender: this,
      data: {},
      success: 'getHostsSuccessCallback',
      error: 'getHostsErrorCallback'
    });
  },

  getHostsSuccessCallback: function (data, opt, params) {
    this.pushHostsToRack(data);
    this.displayHosts();
  },

  /**
   * display hosts of rack
   */
  displayHosts: function () {
    var rackHosts = this.get('rack.hosts');

    if (this.get('hosts.length') === 0) {
      if (rackHosts.length > 100) {
        lazyloading.run({
          initSize: 100,
          chunkSize: 200,
          delay: 100,
          destination: this.get('hosts'),
          source: rackHosts,
          context: this.get('rack')
        });
      } else {
        this.set('hosts', rackHosts);
        this.set('rack.isLoaded', true);
      }
    }
  },

  getHostsErrorCallback: function (request, ajaxOptions, error, opt, params) {
    this.set('rack.isLoaded', true);
  },
  /**
   * push hosts to rack
   * @param data
   */
  pushHostsToRack: function (data) {
    var newHostsData = [];
    var rackHosts = this.get('rack.hosts');

    data.items.forEach(function (item) {
      newHostsData.push({
        hostName: item.Hosts.host_name,
        publicHostName: item.Hosts.public_host_name,
        osType: item.Hosts.os_type,
        ip: item.Hosts.ip,
        diskTotal: item.metrics ? item.metrics.disk.disk_total : 0,
        diskFree: item.metrics ? item.metrics.disk.disk_free : 0,
        cpuSystem: item.metrics ? item.metrics.cpu.cpu_system : 0,
        cpuUser: item.metrics ? item.metrics.cpu.cpu_user : 0,
        memTotal: item.metrics ? item.metrics.memory.mem_total : 0,
        memFree: item.metrics ? item.metrics.memory.mem_free : 0,
        hostComponents: item.host_components.mapProperty('HostRoles.component_name')
      })
    });

    if (rackHosts.length > 0) {
      this.updateLoadedHosts(rackHosts, newHostsData);
    } else {
      this.set('rack.hosts', newHostsData);
    }
  },

  updateLoadedHosts: function (rackHosts, newHostsData) {
    var rackHostsMap = {};
    var isNewHosts = false;

    //create map
    rackHosts.forEach(function (host) {
      rackHostsMap[host.hostName] = host;
    });

    newHostsData.forEach(function (item) {
      var currentHostInfo = rackHostsMap[item.hostName];

      if (currentHostInfo) {
        ['diskTotal', 'diskFree', 'cpuSystem', 'cpuUser', 'memTotal', 'memFree', 'hostComponents'].forEach(function (property) {
          currentHostInfo[property] = item[property];
        });
        delete rackHostsMap[item.hostName];
      } else {
        isNewHosts = true;
      }
    }, this);

    //if hosts were deleted or added then reload hosts view
    if (!App.isEmptyObject(rackHostsMap) || isNewHosts) {
      this.redrawHostsView(newHostsData)
    }
  },

  /**
   * reload hosts rack
   * @param newHostsData
   */
  redrawHostsView: function (newHostsData) {
    this.set('rack.isLoaded', false);
    this.get('hosts').clear();
    this.set('rack.hosts', newHostsData);
  },

  /**
   * call metrics update after hosts of rack are loaded
   */
  updateMetrics: function(){
    if (this.get('rack.isLoaded')) {
      this.get('controller').loadMetrics();
    }
  }.observes('rack.isLoaded'),

  didInsertElement: function () {
    this.set('rack.isLoaded', false);
    if (this.get('rack.hosts.length') > 0) {
      this.displayHosts();
    }
    this.getHosts();
  },
  /**
   * Provides the CSS style for an individual host.
   * This can be used as the 'style' attribute of element.
   */
  hostCssStyle: function () {
    var rack = this.get('rack');
    var widthPercent = 100;
    var hostCount = rack.get('hosts.length');
    if (hostCount && hostCount < 11) {
      widthPercent = (100 / hostCount) - 0.5;
    } else {
      widthPercent = 10; // max out at 10%
    }
    return "width:" + widthPercent + "%;float:left;";
  }.property('rack.isLoaded')
});

});

require.register("views/main/charts/menu", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainChartsMenuView = Em.CollectionView.extend({
  tagName: 'ul',
  classNames: ["nav", "nav-tabs"],
  content:[
    /*{ label:'Heatmap', routing:'heatmap', active:"active"},
    { label:'Horizon Chart', routing:'horizon_chart'}*/
  ],

  init: function(){ this._super(); this.activateView(); },

  activateView:function () {
    $.each(this._childViews, function () {
      this.set('active', (this.get('content.routing') == 'heatmap' ? "active" : ""));
    });
  },

  deactivateChildViews: function() {
    $.each(this._childViews, function(){
      this.set('active', "");
    });
  },

  itemViewClass: Em.View.extend({
    classNameBindings: ["active"],
    active: "",
    template: Ember.Handlebars.compile('<a {{action showChart view.content.routing }} href="#"> {{unbound view.content.label}}</a>')
  })
});
});

require.register("views/main/dashboard", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainDashboardView = Em.View.extend({

  name: 'mainDashboardView',
  templateName: require('templates/main/dashboard'),

  selectedBinding: 'controller.selectedCategory',
  categories: [
    {
      name: 'widgets',
      url: 'dashboard.index',
      label: Em.I18n.t('dashboard.widgets.title')
    },
    {
      name: 'charts',
      url: 'dashboard.charts.index',
      label: Em.I18n.t('dashboard.heatmaps.title')
    },
    {
      name: 'configHistory',
      url: 'dashboard.configHistory',
      label: Em.I18n.t('dashboard.configHistory.title')
    }
  ],

  NavItemView: Ember.View.extend({
    tagName: 'li',
    classNameBindings: 'isActive:active'.w(),
    isActive: function () {
      return this.get('item') === this.get('parentView.selected');
    }.property('item', 'parentView.selected'),
    elementId: Ember.computed(function(){
      var label = Em.get(this, 'templateData.keywords.category.label');
      return label ? 'dashboard-view-tab-' + label.toLowerCase().replace(/\s/g, '-') : "";
    })
  })
});

});

require.register("views/main/dashboard/cluster_metrics/cpu", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartClusterMetricsCPU = App.ChartLinearTimeView.extend({
  id: "cluster-metrics-cpu",

  ajaxIndex: 'dashboard.cluster_metrics.cpu',

  title: Em.I18n.t('dashboard.clusterMetrics.cpu'),
  yAxisFormatter: App.ChartLinearTimeView.PercentageFormatter,
  isTimePagingDisable: true,
  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var idle = null;

    if (jsonData && jsonData.metrics && jsonData.metrics.cpu) {
      for (var name in jsonData.metrics.cpu) {
        var seriesData = jsonData.metrics.cpu[name];
        if (seriesData) {
          var s = this.transformData(seriesData, name);
          if (name === 'Idle') {
            //CPU idle metric should be the last in series array
            idle = s;
            continue;
          }
          seriesArray.push(s);
        }
      }
      if (idle) {
        seriesArray.push(idle);
      }
    }
    return seriesArray;
  },

  colorForSeries: function (series) {
    if (Em.I18n.t('dashboard.clusterMetrics.cpu.displayNames.idle') == series.name){
      return '#CFECEC';
    }
    return null;
  }
});
});

require.register("views/main/dashboard/cluster_metrics/load", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster load
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartClusterMetricsLoad = App.ChartLinearTimeView.extend({
  id: "cluster-metrics-load",

  ajaxIndex: 'dashboard.cluster_metrics.load',
  isTimePagingDisable: true,
  renderer: 'line',
  title: Em.I18n.t('dashboard.clusterMetrics.load'),

  transformToSeries: function(jsonData){
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.load) {
      for ( var name in jsonData.metrics.load) {
        var displayName = name;
        var seriesData = jsonData.metrics.load[name];
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/dashboard/cluster_metrics/memory", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster memory metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartClusterMetricsMemory = App.ChartLinearTimeView.extend({
  id: "cluster-metrics-memory",

  ajaxIndex: 'dashboard.cluster_metrics.memory',

  isTimePagingDisable: true,
  title: Em.I18n.t('dashboard.clusterMetrics.memory'),
  yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
  renderer: 'line',
  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.memory) {
      var isAmbariMetricsAvailable = App.StackService.find().someProperty('serviceName', 'AMBARI_METRICS');
      var isAmbariMetricsInstalled = App.Service.find().someProperty('serviceName', 'AMBARI_METRICS');
      var isGangliaInstalled = App.Service.find().someProperty('serviceName', 'GANGLIA');
      var shouldConvertToBytes = isAmbariMetricsInstalled || isAmbariMetricsAvailable && !isGangliaInstalled;
      var KB = Math.pow(2, 10);
      for ( var name in jsonData.metrics.memory) {
        var displayName = name;
        var seriesData = jsonData.metrics.memory[name];
        if (seriesData) {
          var s = this.transformData(seriesData, displayName);
          if (shouldConvertToBytes) {
            for (var i = 0; i < s.data.length; i++) {
              s.data[i].y *= KB;
            }
          }
          seriesArray.push(s);
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/dashboard/cluster_metrics/network", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster network metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartClusterMetricsNetwork = App.ChartLinearTimeView.extend({
  id: "cluster-metrics-network",

  ajaxIndex: 'dashboard.cluster_metrics.network',

  isTimePagingDisable: true,
  title: Em.I18n.t('dashboard.clusterMetrics.network'),
  yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
  renderer: 'line',

  transformToSeries : function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.network) {
      for ( var name in jsonData.metrics.network) {
        var displayName = name;
        var seriesData = jsonData.metrics.network[name];
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/dashboard/config_history_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var filters = require('views/common/filter_view');
var sort = require('views/common/sort_view');

App.MainConfigHistoryView = App.TableView.extend(App.TableServerViewMixin, {
  templateName: require('templates/main/dashboard/config_history'),

  controllerBinding: 'App.router.mainConfigHistoryController',

  /**
   * @type {boolean}
   * @default false
   */
  filteringComplete: false,
  isInitialRendering: true,

  /**
   * return filtered number of all content number information displayed on the page footer bar
   * @returns {String}
   */
  filteredContentInfo: function () {
    return this.t('tableView.filters.filteredConfigVersionInfo').format(this.get('filteredCount'), this.get('totalCount'));
  }.property('filteredCount', 'totalCount'),

  didInsertElement: function () {
    this.addObserver('startIndex', this, 'updatePagination');
    this.addObserver('displayLength', this, 'updatePagination');
    this.set('isInitialRendering', true);
    this.refresh();
    this.set('controller.isPolling', true);
    this.get('controller').doPolling();
  },

  /**
   * stop polling after leaving config history page
   */
  willDestroyElement: function () {
    this.set('controller.isPolling', false);
    clearTimeout(this.get('controller.timeoutRef'));
  },

  updateFilter: function (iColumn, value, type) {
    if (!this.get('isInitialRendering')) {
      this._super(iColumn, value, type);
    }
  },

  sortView: sort.serverWrapperView,
  versionSort: sort.fieldView.extend({
    column: 1,
    name: 'serviceVersion',
    displayName: Em.I18n.t('dashboard.configHistory.table.version.title'),
    classNames: ['first']
  }),
  configGroupSort: sort.fieldView.extend({
    column: 2,
    name: 'configGroup',
    displayName: Em.I18n.t('dashboard.configHistory.table.configGroup.title')
  }),
  modifiedSort: sort.fieldView.extend({
    column: 3,
    name: 'createTime',
    status: 'sorting_desc',
    displayName: Em.I18n.t('dashboard.configHistory.table.created.title')
  }),
  authorSort: sort.fieldView.extend({
    column: 4,
    name: 'author',
    displayName: Em.I18n.t('common.author')
  }),
  notesSort: sort.fieldView.extend({
    column: 5,
    name: 'notes',
    displayName: Em.I18n.t('common.notes')
  }),

  serviceFilterView: filters.createSelectView({
    column: 1,
    fieldType: 'filter-input-width',
    content: function () {
      return [
        {
          value: '',
          label: Em.I18n.t('common.all')
        }
      ].concat(App.Service.find().map(function (service) {
        return {
          value: service.get('serviceName'),
          label: service.get('displayName')
        }
      }));
    }.property('App.router.clusterController.isLoaded'),
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'select');
    }
  }),

  configGroupFilterView: filters.createSelectView({
    column: 2,
    fieldType: 'filter-input-width',
    content: function () {
      var groupName = App.ServiceConfigVersion.find().mapProperty('groupName').uniq();
      if (groupName.indexOf(null) > -1) {
        groupName.splice(groupName.indexOf(null), 1);
      }
      return [
        {
          value: '',
          label: Em.I18n.t('common.all')
        }
      ].concat(groupName.map(function (item) {
        return {
          value: item,
          label: item
        }
      }));
    }.property('parentView.isInitialRendering'),
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'select');
    }
  }),

  modifiedFilterView: filters.createSelectView({
    column: 3,
    appliedEmptyValue: ["", ""],
    fieldType: 'filter-input-width,modified-filter',
    emptyValue: 'Any',
    contentBinding: "controller.modifiedFilter.content",
    onChangeValue: function () {
      this.set("controller.modifiedFilter.optionValue", this.get('selected'));
      this.get('parentView').updateFilter(this.get('column'), [this.get('controller.modifiedFilter.actualValues.startTime'), this.get('controller.modifiedFilter.actualValues.endTime')], 'range');
    }
  }),

  authorFilterView: filters.createTextView({
    column: 4,
    fieldType: 'filter-input-width',
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'string');
    }
  }),

  notesFilterView: filters.createTextView({
    column: 5,
    fieldType: 'filter-input-width',
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'string');
    }
  }),

  ConfigVersionView: Em.View.extend({
    tagName: 'tr',
    showLessNotes: true,
    toggleShowLessStatus: function () {
      this.toggleProperty('showLessNotes');
    },
    didInsertElement: function () {
      App.tooltip(this.$("[rel='Tooltip']"));
    }
  }),

  /**
   * refresh table content
   */
  refresh: function () {
    var self = this;
    this.set('filteringComplete', false);
    this.get('controller').load().done(function () {
      self.refreshDone.apply(self);
    });
  },

  /**
   * callback executed after refresh call done
   * @method refreshDone
   */
  refreshDone: function () {
    this.set('isInitialRendering', false);
    this.set('filteringComplete', true);
    this.propertyDidChange('pageContent');
    this.set('controller.resetStartIndex', false);
  },

  /**
   * associations between host property and column index
   * @type {Array}
   */
  colPropAssoc: function () {
    return this.get('controller.colPropAssoc');
  }.property('controller.colPropAssoc')
});

});

require.register("views/main/dashboard/widget", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.DashboardWidgetView = Em.View.extend({

  /**
   * @type {string}
   * @default null
   */
  title: null,
  templateName: null, // each has specific template

  /**
   * Setup model for widget by `model_type`. Usually `model_type` is a lowercase service name,
   * for example `hdfs`, `yarn`, etc. You need to set `model_type` in extended object View, for example
   * look App.DataNodeUpView.
   * @type {object} - model that set up in App.MainDashboardView.setWidgetsDataModel()
   */
  model : function () {
    if (!this.get('model_type')) return {};
    return this.get('parentView').get(this.get('model_type') + '_model');
  }.property(),

  /**
   * id 1-10 used to identify
   * @type {number}
   * @default null
   */
  id: null,

  /**
   * html id bind to view-class: widget-(1)
   * used by re-sort
   * @type {string}
   */
  viewID: function () {
    return 'widget-' + this.get('id');
  }.property('id'),
  attributeBindings: ['viewID'],

  /**
   * @type {boolean}
   */
  isPieChart: false,

  /**
   * @type {boolean}
   */
  isText: false,

  /**
   * @type {boolean}
   */
  isProgressBar: false,

  /**
   * @type {boolean}
   */
  isLinks: false,

  /**
   * widget content pieChart/ text/ progress bar/links/ metrics. etc
   * @type {Array}
   * @default null
   */
  content: null,

  /**
   * more info details
   * @type {Array}
   */
  hiddenInfo: [],

  /**
   * @type {string}
   */
  hiddenInfoClass: "hidden-info-two-line",

  /**
   * @type {number}
   * @default null
   */
  thresh1: null,

  /**
   * @type {number}
   * @default null
   */
  thresh2: null,

  /**
   * @type {Em.Object}
   * @class
   */
  widgetConfig: Ember.Object.extend({
    thresh1: '',
    thresh2: '',
    hintInfo: function () {
      return Em.I18n.t('dashboard.widgets.hintInfo.common').format(this.get('maxValue'));
    }.property('maxValue'),
    isThresh1Error: false,
    isThresh2Error: false,
    errorMessage1: "",
    errorMessage2: "",
    maxValue: 0,
    observeThresh1Value: function () {
      var thresh1 = this.get('thresh1');
      var thresh2 = this.get('thresh2');
      var maxValue = this.get('maxValue');

      if (thresh1.trim() != "") {
        if (isNaN(thresh1) || thresh1 > maxValue || thresh1 < 0) {
          this.set('isThresh1Error', true);
          this.set('errorMessage1', Em.I18n.t('dashboard.widgets.error.invalid').format(maxValue));
        } else if (this.get('isThresh2Error') === false && parseFloat(thresh2) <= parseFloat(thresh1)) {
          this.set('isThresh1Error', true);
          this.set('errorMessage1', Em.I18n.t('dashboard.widgets.error.smaller'));
        } else {
          this.set('isThresh1Error', false);
          this.set('errorMessage1', '');
        }
      } else {
        this.set('isThresh1Error', true);
        this.set('errorMessage1', Em.I18n.t('admin.users.editError.requiredField'));
      }
      this.updateSlider();
    }.observes('thresh1', 'maxValue'),
    observeThresh2Value: function () {
      var thresh2 = this.get('thresh2');
      var maxValue = this.get('maxValue');

      if (thresh2.trim() != "") {
        if (isNaN(thresh2) || thresh2 > maxValue || thresh2 < 0) {
          this.set('isThresh2Error', true);
          this.set('errorMessage2', Em.I18n.t('dashboard.widgets.error.invalid').format(maxValue));
        } else {
          this.set('isThresh2Error', false);
          this.set('errorMessage2', '');
        }
      } else {
        this.set('isThresh2Error', true);
        this.set('errorMessage2', Em.I18n.t('admin.users.editError.requiredField'));
      }
      this.updateSlider();
    }.observes('thresh2', 'maxValue'),
    updateSlider: function () {
      var thresh1 = this.get('thresh1');
      var thresh2 = this.get('thresh2');
      // update the slider handles and color
      if (this.get('isThresh1Error') === false && this.get('isThresh2Error') === false) {
        $("#slider-range").slider('values', 0, parseFloat(thresh1));
        $("#slider-range").slider('values', 1, parseFloat(thresh2));
      }
    }
  }),

  didInsertElement: function () {
    App.tooltip(this.$("[rel='ZoomInTooltip']"), {placement : 'left'});
  },

  /**
   * delete widget
   * @param {object} event
   */
  deleteWidget: function (event) {
    var parent = this.get('parentView');
    var self = this;

    if (App.get('testMode')) {
      //update view on dashboard
      var objClass = parent.widgetsMapper(this.get('id'));
      parent.get('visibleWidgets').removeObject(objClass);
      parent.get('hiddenWidgets').pushObject(Em.Object.create({displayName: this.get('title'), id: this.get('id'), checked: false}));
    } else {
      //reconstruct new persist value then post in persist
      parent.getUserPref(parent.get('persistKey')).complete(function () {
        self.deleteWidgetComplete.apply(self);
      });
    }
  },

  /**
   * delete widget complete callback
   */
  deleteWidgetComplete: function () {
    var parent = this.get('parentView');
    var oldValue = parent.get('currentPrefObject');
    var deletedId = this.get('id');
    var newValue = Em.Object.create({
      dashboardVersion: oldValue.dashboardVersion,
      visible: oldValue.visible.slice(0).without(deletedId),
      hidden: oldValue.hidden,
      threshold: oldValue.threshold
    });
    newValue.hidden.push([deletedId, this.get('title')]);
    parent.postUserPref(parent.get('persistKey'), newValue);
    parent.translateToReal(newValue);
  },

  /**
   * edit widget
   * @param {object} event
   */
  editWidget: function (event) {
    var configObj = this.get('widgetConfig').create({
      thresh1: this.get('thresh1') + '',
      thresh2: this.get('thresh2') + '',
      maxValue: parseFloat(this.get('maxValue'))
    });
    this.showEditDialog(configObj)
  },

  /**
   *  show edit dialog
   * @param {Em.Object} configObj
   * @returns {App.ModalPopup}
   */
  showEditDialog: function (configObj) {
    var self = this;
    var maxValue = this.get('maxValue');

    return App.ModalPopup.show({
      header: Em.I18n.t('dashboard.widgets.popupHeader'),
      classNames: [ 'sixty-percent-width-modal-edit-widget' ],
      bodyClass: Ember.View.extend({
        templateName: require('templates/main/dashboard/edit_widget_popup'),
        configPropertyObj: configObj
      }),
      primary: Em.I18n.t('common.apply'),
      onPrimary: function () {
        configObj.observeThresh1Value();
        configObj.observeThresh2Value();
        if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
          self.set('thresh1', parseFloat(configObj.get('thresh1')));
          self.set('thresh2', parseFloat(configObj.get('thresh2')));

          if (!App.get('testMode')) {
            // save to persist
            var parent = self.get('parentView');
            parent.getUserPref(parent.get('persistKey')).complete(function () {
              var oldValue = parent.get('currentPrefObject');
              oldValue.threshold[parseInt(self.get('id'))] = [configObj.get('thresh1'), configObj.get('thresh2')];
              parent.postUserPref(parent.get('persistKey'), oldValue);
            });
          }

          this.hide();
        }
      },

      didInsertElement: function () {
        var browserVersion = self.getInternetExplorerVersion();
        var handlers = [configObj.get('thresh1'), configObj.get('thresh2')];
        var colors = [App.healthStatusGreen, App.healthStatusOrange, App.healthStatusRed]; //color green, orange ,red

        if (browserVersion == -1 || browserVersion > 9) {
          configObj.set('isIE9', false);
          configObj.set('isGreenOrangeRed', true);
          $("#slider-range").slider({
            range: true,
            min: 0,
            max: maxValue,
            values: handlers,
            create: function (event, ui) {
              updateColors(handlers);
            },
            slide: function (event, ui) {
              updateColors(ui.values);
              configObj.set('thresh1', ui.values[0] + '');
              configObj.set('thresh2', ui.values[1] + '');
            },
            change: function (event, ui) {
              updateColors(ui.values);
            }
          });

          function updateColors(handlers) {
            var colorstops = colors[0] + ", "; // start with the first color
            for (var i = 0; i < handlers.length; i++) {
              colorstops += colors[i] + " " + handlers[i] * 100 / maxValue + "%,";
              colorstops += colors[i + 1] + " " + handlers[i] * 100 / maxValue + "%,";
            }
            colorstops += colors[colors.length - 1];
            var sliderElement = $('#slider-range');
            var css1 = '-webkit-linear-gradient(left,' + colorstops + ')'; // chrome & safari
            sliderElement.css('background-image', css1);
            var css2 = '-ms-linear-gradient(left,' + colorstops + ')'; // IE 10+
            sliderElement.css('background-image', css2);
            //$('#slider-range').css('filter', 'progid:DXImageTransform.Microsoft.gradient( startColorStr= ' + colors[0] + ', endColorStr= ' + colors[2] +',  GradientType=1 )' ); // IE 10-
            var css3 = '-moz-linear-gradient(left,' + colorstops + ')'; // Firefox
            sliderElement.css('background-image', css3);

            sliderElement.find('.ui-widget-header').css({'background-color': '#FF8E00', 'background-image': 'none'}); // change the  original ranger color
          }
        } else {
          configObj.set('isIE9', true);
          configObj.set('isGreenOrangeRed', true);
        }
      }
    });
  },

  /**
   * @returns {number}
   */
  getInternetExplorerVersion: function () {
    var rv = -1; //return -1 for other browsers
    if (navigator.appName == 'Microsoft Internet Explorer') {
      var ua = navigator.userAgent;
      var re = new RegExp("MSIE ([0-9]{1,}[\.0-9]{0,})");
      if (re.exec(ua) != null)
        rv = parseFloat(RegExp.$1); // IE version 1-10
    }
    var isFirefox = typeof InstallTrigger !== 'undefined';   // Firefox 1.0+
    if (isFirefox) {
      return -2;
    } else {
      return rv;
    }
  },

  /**
   * for widgets has hidden info(hover info),
   * calculate the hover content top number
   * based on how long the hiddenInfo is
   * @returns {string}
   */
  hoverContentTopClass: function () {
    var lineNum = this.get('hiddenInfo.length');
    if (lineNum == 2) {
      return "content-hidden-two-line";
    } else if (lineNum == 3) {
      return "content-hidden-three-line";
    } else if (lineNum == 4) {
      return "content-hidden-four-line";
    } else if (lineNum == 5) {
      return "content-hidden-five-line";
    } else if (lineNum == 6) {
      return "content-hidden-six-line";
    }
    return '';
  }.property('hiddenInfo.length')

});


App.DashboardWidgetView.reopenClass({
  class: 'span2p4'
});

});

require.register("views/main/dashboard/widgets", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var filters = require('views/common/filter_view');

App.MainDashboardWidgetsView = Em.View.extend(App.UserPref, App.LocalStorage, {

  name: 'mainDashboardWidgetsView',

  templateName: require('templates/main/dashboard/widgets'),

  didInsertElement: function () {
    this.setWidgetsDataModel();
    this.setInitPrefObject();
    this.setOnLoadVisibleWidgets();
    this.set('isDataLoaded', true);
    Em.run.next(this, 'makeSortable');
  },

  /**
   * List of services
   * @type {Ember.Enumerable}
   */
  content: [],

  /**
   * @type {boolean}
   */
  isDataLoaded: false,

  /**
   * Define if some widget is currently moving
   * @type {boolean}
   */
  isMoving: false,

  /**
   * Make widgets' list sortable on New Dashboard style
   */
  makeSortable: function () {
    var self = this;
    $("#sortable").sortable({
      items: "> div",
      //placeholder: "sortable-placeholder",
      cursor: "move",
      tolerance: "pointer",
      scroll: false,
      update: function (event, ui) {
        if (!App.get('testMode')) {
          // update persist then translate to real
          var widgetsArray = $('div[viewid]'); // get all in DOM
          self.getUserPref(self.get('persistKey')).complete(function () {
            var oldValue = self.get('currentPrefObject') || self.getDBProperty(self.get('persistKey'));
            var newValue = Em.Object.create({
              dashboardVersion: oldValue.dashboardVersion,
              visible: [],
              hidden: oldValue.hidden,
              threshold: oldValue.threshold
            });
            var size = oldValue.visible.length;
            for (var j = 0; j <= size - 1; j++) {
              var viewID = widgetsArray.get(j).getAttribute('viewid');
              var id = viewID.split("-").get(1);
              newValue.visible.push(id);
            }
            self.postUserPref(self.get('persistKey'), newValue);
            self.setDBProperty(self.get('persistKey'), newValue);
            //self.translateToReal(newValue);
          });
        }
      },
      activate: function (event, ui) {
        self.set('isMoving', true);
      },
      deactivate: function (event, ui) {
        self.set('isMoving', false);
      }
    }).disableSelection();
  },

  /**
   * Set Service model values
   */
  setWidgetsDataModel: function () {
    if (App.get('services.hostMetrics').length > 0) {
      this.set('host_metrics_model', App.get('services.hostMetrics'));
    }
    App.Service.find().forEach(function (item) {
      var extendedModel = App.Service.extendedModel[item.get('serviceName')];
      var key = item.get('serviceName').toLowerCase() + '_model';
      if (extendedModel && App[extendedModel].find(item.get('id'))) {
        this.set(key, App[extendedModel].find(item.get('id')));
      } else {
        this.set(key, item);
      }
    }, this);
  },

  /**
   * Load widget statuses to <code>initPrefObject</code>
   */
  setInitPrefObject: function () {
    //in case of some service not installed
    var visibleFull = [
      '2', '4', '11', //hdfs
      '6', '7', '8', '9', //host metrics
      '1', '5', '3',  '10', //hdfs
      '13', '12', '14', '16', //hbase
      '17', '18', '19', '20', '23', // all yarn
      '21', // storm
      '22' // flume
    ]; // all in order
    var hiddenFull = [
      ['15', 'Region In Transition']
    ];

    // Display widgets for host metrics if the stack definition has a host metrics service to display it.
    if (this.get('host_metrics_model') == null) {
      var hostMetrics = ['6', '7', '8', '9'];
      hostMetrics.forEach(function (item) {
        visibleFull = visibleFull.without(item);
      }, this);
    }

    if (this.get('hdfs_model') == null) {
      var hdfs = ['1', '2', '3', '4', '5', '10', '11'];
      hdfs.forEach(function (item) {
        visibleFull = visibleFull.without(item);
      }, this);
    }
    if (this.get('hbase_model') == null) {
      var hbase = ['12', '13', '14', '16'];
      hbase.forEach(function (item) {
        visibleFull = visibleFull.without(item);
      }, this);
      hiddenFull = [];
    }
    if (this.get('yarn_model') == null) {
      var yarn = ['17', '18', '19', '20', '23'];
      yarn.forEach(function (item) {
        visibleFull = visibleFull.without(item);
      }, this);
    }
    if (this.get('storm_model') == null) {
      var storm = ['21'];
      storm.forEach(function (item) {
        visibleFull = visibleFull.without(item);
      }, this);
    }
    if (this.get('flume_model') == null) {
      var flume = ['22'];
      flume.forEach(function (item) {
        visibleFull = visibleFull.without(item);
      }, this);
    }
    var obj = this.get('initPrefObject');
    obj.set('visible', visibleFull);
    obj.set('hidden', hiddenFull);
  },

  host_metrics_model: null,

  hdfs_model: null,

  mapreduce2_model: null,

  yarn_model: null,

  hbase_model: null,

  storm_model: null,

  flume_model: null,

  /**
   * List of visible widgets
   * @type {Ember.Enumerable}
   */
  visibleWidgets: [],

  /**
   * List of hidden widgets
   * @type {Ember.Enumerable}
   */
  hiddenWidgets: [], // widget child view will push object in this array if deleted

  /**
   * Submenu view for New Dashboard style
   * @type {Ember.View}
   * @class
   */
  plusButtonFilterView: Ember.View.extend({
    templateName: require('templates/main/dashboard/plus_button_filter'),
    hiddenWidgetsBinding: 'parentView.hiddenWidgets',
    visibleWidgetsBinding: 'parentView.visibleWidgets',
    valueBinding: '',
    widgetCheckbox: Em.Checkbox.extend({
      didInsertElement: function () {
        $('.checkbox').click(function (event) {
          event.stopPropagation();
        });
      }
    }),
    closeFilter: Em.K,
    applyFilter: function () {
      var self = this;
      var parent = this.get('parentView');
      var hiddenWidgets = this.get('hiddenWidgets');
      var checkedWidgets = hiddenWidgets.filterProperty('checked', true);

      if (App.get('testMode')) {
        var visibleWidgets = this.get('visibleWidgets');
        checkedWidgets.forEach(function (item) {
          var newObj = parent.widgetsMapper(item.id);
          visibleWidgets.pushObject(newObj);
          hiddenWidgets.removeObject(item);
        }, this);
      } else {
        //save in persist
        parent.getUserPref(parent.get('persistKey')).complete(function(){
          self.applyFilterComplete.apply(self);
        });
      }
    },
    applyFilterComplete: function () {
      var parent = this.get('parentView'),
        hiddenWidgets = this.get('hiddenWidgets'),
        oldValue = parent.get('currentPrefObject'),
        newValue = Em.Object.create({
          dashboardVersion: oldValue.dashboardVersion,
          visible: oldValue.visible,
          hidden: [],
          threshold: oldValue.threshold
        });
      hiddenWidgets.filterProperty('checked').forEach(function (item) {
        newValue.visible.push(item.id);
        hiddenWidgets.removeObject(item);
      }, this);
      hiddenWidgets.forEach(function (item) {
        newValue.hidden.push([item.id, item.displayName]);
      }, this);
      parent.postUserPref(parent.get('persistKey'), newValue);
      parent.translateToReal(newValue);
    }
  }),

  /**
   * Translate from Json value got from persist to real widgets view
   */
  translateToReal: function (value) {
    var version = value.dashboardVersion;
    var visible = value.visible;
    var hidden = value.hidden;
    var threshold = value.threshold;

    if (version == 'new') {
      var visibleWidgets = [];
      var hiddenWidgets = [];
      // re-construct visibleWidgets and hiddenWidgets
      for (var i = 0; i < visible.length; i++) {
        var id = visible[i];
        var widgetClass = this.widgetsMapper(id);
        //override with new threshold
        if (threshold[id].length > 0) {
          widgetClass.reopen({
            thresh1: threshold[id][0],
            thresh2: threshold[id][1]
          });
        }
        visibleWidgets.pushObject(widgetClass);
      }
      for (var j = 0; j < hidden.length; j++) {
        var title = hidden[j][1];
        hiddenWidgets.pushObject(Em.Object.create({displayName: title, id: hidden[j][0], checked: false}));
      }
      this.set('visibleWidgets', visibleWidgets);
      this.set('hiddenWidgets', hiddenWidgets);
    }
  },

  /**
   * Set visibility-status for widgets
   */
  setOnLoadVisibleWidgets: function () {
    var self = this;
    if (App.get('testMode')) {
      this.translateToReal(this.get('initPrefObject'));
    } else {
      // called when first load/refresh/jump back page
      this.getUserPref(this.get('persistKey')).complete(function () {
        self.setOnLoadVisibleWidgetsComplete.apply(self);
      });
    }
  },

  /**
   * complete load of visible widgets
   */
  setOnLoadVisibleWidgetsComplete: function () {
    var currentPrefObject = this.get('currentPrefObject') || this.getDBProperty(this.get('persistKey'));
    if (currentPrefObject) { // fit for no dashboard version
      if (!currentPrefObject.dashboardVersion) {
        currentPrefObject.dashboardVersion = 'new';
        this.postUserPref(this.get('persistKey'), currentPrefObject);
        this.setDBProperty(this.get('persistKey'), currentPrefObject);
      }
      this.set('currentPrefObject', this.checkServicesChange(currentPrefObject));
      this.translateToReal(this.get('currentPrefObject'));
    }
    else {
      // post persist then translate init object
      this.postUserPref(this.get('persistKey'), this.get('initPrefObject'));
      this.setDBProperty(this.get('persistKey'), this.get('initPrefObject'));
      this.translateToReal(this.get('initPrefObject'));
    }
  },

  /**
   * Remove widget from visible and hidden lists
   * @param {Object} value
   * @param {Object} widget
   * @returns {*}
   */
  removeWidget: function (value, widget) {
    value.visible = value.visible.without(widget);
    for (var j = 0; j < value.hidden.length; j++) {
      if (value.hidden[j][0] == widget) {
        value.hidden.splice(j, 1);
      }
    }
    return value;
  },

  /**
   * Check if widget is in visible or hidden list
   * @param {Object} value
   * @param {Object} widget
   * @returns {bool}
   */
  containsWidget: function (value, widget) {
    var flag = value.visible.contains(widget);
    for (var j = 0; j < value.hidden.length; j++) {
      if (!flag && value.hidden[j][0] == widget) {
        flag = true;
        break;
      }
    }
    return flag;
  },

  /**
   * check if stack has upgraded from HDP 1.0 to 2.0 OR add/delete services.
   * Update the value on server if true.
   * @param {Object} currentPrefObject
   * @return {Object}
   */
  checkServicesChange: function (currentPrefObject) {
    var toDelete = $.extend(true, {}, currentPrefObject);
    var toAdd = [];
    var serviceWidgetsMap = {
      hdfs_model: ['1', '2', '3', '4', '5', '10', '11'],
      host_metrics_model: ['6', '7', '8', '9'],
      hbase_model: ['12', '13', '14', '15', '16'],
      yarn_model: ['17', '18', '19', '20', '23'],
      storm_model: ['21'],
      flume_model: ['22']
    };

    // check each service, find out the newly added service and already deleted service
    Em.keys(serviceWidgetsMap).forEach(function (modelName) {
      if (!Em.isNone(this.get(modelName))) {
        var ids = serviceWidgetsMap[modelName];
        var flag = this.containsWidget(toDelete, ids[0]);
        if (flag) {
          ids.forEach(function (item) {
            toDelete = this.removeWidget(toDelete, item);
          }, this);
        } else {
          toAdd = toAdd.concat(ids);
        }
      }
    }, this);

    var value = currentPrefObject;
    if (toDelete.visible.length || toDelete.hidden.length) {
      toDelete.visible.forEach(function (item) {
        value = this.removeWidget(value, item);
      }, this);
      toDelete.hidden.forEach(function (item) {
        value = this.removeWidget(value, item[0]);
      }, this);
    }
    if (toAdd.length) {
      value.visible = value.visible.concat(toAdd);
      var allThreshold = this.get('initPrefObject').threshold;
      // add new threshold OR override with default value
      toAdd.forEach(function (item) {
        value.threshold[item] = allThreshold[item];
      }, this);
    }
    return value;
  },

  /**
   * Get view for widget by widget's id
   * @param {string} id
   * @returns {Ember.View}
   */
  widgetsMapper: function (id) {
    return Em.get({
      '1': App.NameNodeHeapPieChartView,
      '2': App.NameNodeCapacityPieChartView,
      '3': App.NameNodeCpuPieChartView,
      '4': App.DataNodeUpView,
      '5': App.NameNodeRpcView,
      '6': App.ChartClusterMetricsMemoryWidgetView,
      '7': App.ChartClusterMetricsNetworkWidgetView,
      '8': App.ChartClusterMetricsCPUWidgetView,
      '9': App.ChartClusterMetricsLoadWidgetView,
      '10': App.NameNodeUptimeView,
      '11': App.HDFSLinksView,
      '12': App.HBaseLinksView,
      '13': App.HBaseMasterHeapPieChartView,
      '14': App.HBaseAverageLoadView,
      '15': App.HBaseRegionsInTransitionView,
      '16': App.HBaseMasterUptimeView,
      '17': App.ResourceManagerHeapPieChartView,
      '18': App.ResourceManagerUptimeView,
      '19': App.NodeManagersLiveView,
      '20': App.YARNMemoryPieChartView,
      '21': App.SuperVisorUpView,
      '22': App.FlumeAgentUpView,
      '23': App.YARNLinksView
    }, id);
  },

  /**
   * @type {Object|null}
   */
  currentPrefObject: null,

  /**
   * @type {Ember.Object}
   */
  initPrefObject: Em.Object.create({
    dashboardVersion: 'new',
    visible: [],
    hidden: [],
    threshold: {1: [80, 90], 2: [85, 95], 3: [90, 95], 4: [80, 90], 5: [1000, 3000], 6: [], 7: [], 8: [], 9: [], 10: [], 11: [], 12: [], 13: [70, 90], 14: [150, 250], 15: [3, 10], 16: [],
      17: [70, 90], 18: [], 19: [50, 75], 20: [50, 75], 21: [85, 95], 22: [85, 95], 23: []} // id:[thresh1, thresh2]
  }),

  /**
   * Key-name to store data in Local Storage and Persist
   * @type {string}
   */
  persistKey: function () {
    return 'user-pref-' + App.router.get('loginName') + '-dashboard';
  }.property(),

  getUserPrefSuccessCallback: function (response, request, data) {
    if (response) {
      console.log('Got persist value from server with key ' + data.key + '. Value is: ' + response);
      var initPrefObject = this.get('initPrefObject');
      initPrefObject.get('threshold');
      for(var k in response.threshold) {
        if (response.threshold.hasOwnProperty(k)) {
          if (response.threshold[k].length === 0 && initPrefObject.get('threshold')[k] && initPrefObject.get('threshold')[k].length) {
            response.threshold[k] = initPrefObject.get('threshold')[k];
          }
        }
      }
      this.set('currentPrefObject', response);
    }
  },

  getUserPrefErrorCallback: function (request) {
    // this user is first time login
    if (request.status == 404) {
      console.log('Persist did NOT find the key');
    }
  },

  /**
   * Reset widgets visibility-status
   */
  resetAllWidgets: function () {
    var self = this;
    App.showConfirmationPopup(function () {
      if (!App.get('testMode')) {
        self.postUserPref(self.get('persistKey'), self.get('initPrefObject'));
        self.setDBProperty(self.get('persistKey'), self.get('initPrefObject'));
      }
      self.translateToReal(self.get('initPrefObject'));
    });
  },

  /**
   * @type {string}
   */
  gangliaUrl: function () {
    return App.router.get('clusterController.gangliaUrl') + "/?r=hour&cs=&ce=&m=&s=by+name&c=HDPSlaves&tab=m&vn=";
  }.property('App.router.clusterController.gangliaUrl'),

  showAlertsPopup: Em.K

});


});

require.register("views/main/dashboard/widgets/cluster_metrics_widget", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ClusterMetricsDashboardWidgetView = App.DashboardWidgetView.extend({

  templateName: require('templates/main/dashboard/widgets/cluster_metrics')

});

});

require.register("views/main/dashboard/widgets/datanode_live", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.DataNodeUpView = App.TextDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.DataNodeUp'),
  id: '4',

  isPieChart: false,
  isText: true,
  isProgressBar: false,
  model_type: 'hdfs',

  hiddenInfo: function () {
    var result = [];
    result.pushObject(this.get('dataNodesLive') + ' ' + Em.I18n.t('dashboard.services.hdfs.nodes.live'));
    result.pushObject(this.get('dataNodesDead') + ' ' + Em.I18n.t('dashboard.services.hdfs.nodes.dead'));
    result.pushObject(this.get('dataNodesDecom')+ ' ' + Em.I18n.t('dashboard.services.hdfs.nodes.decom'));
    return result;
  }.property('dataNodesLive', 'dataNodesDead', 'dataNodesDecom'),
  hiddenInfoClass: "hidden-info-three-line",

  thresh1: 40,
  thresh2: 70,
  maxValue: 100,

  dataNodesLive: function () {
    return this.get('model.metricsNotAvailable') ? Em.I18n.t('services.service.summary.notAvailable') : this.get('model.liveDataNodes.length');
  }.property('model.liveDataNodes.length'),
  dataNodesDead: function () {
    return this.get('model.metricsNotAvailable') ? Em.I18n.t('services.service.summary.notAvailable') : this.get('model.deadDataNodes.length');
  }.property('model.deadDataNodes.length'),
  dataNodesDecom: function () {
    return this.get('model.metricsNotAvailable') ? Em.I18n.t('services.service.summary.notAvailable') : this.get('model.decommissionDataNodes.length');
  }.property('model.decommissionDataNodes.length'),

  data: function () {
    if (this.get('model.metricsNotAvailable')) {
      return null;
    } else {
      return ((this.get('dataNodesLive') / this.get('model.dataNodesTotal')).toFixed(2)) * 100;
    }
  }.property('model.dataNodesTotal', 'dataNodesLive'),

  content: function () {
    return this.get('model.metricsNotAvailable') ? Em.I18n.t('services.service.summary.notAvailable') : this.get('dataNodesLive') + "/" + this.get('model.dataNodesTotal');
  }.property('model.dataNodesTotal', 'dataNodesLive'),

  editWidget: function (event) {
    var parent = this;
    var max_tmp =  parseFloat(parent.get('maxValue'));
    var configObj = Ember.Object.create({
      thresh1: parent.get('thresh1') + '',
      thresh2: parent.get('thresh2') + '',
      hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint1').format(max_tmp),
      isThresh1Error: false,
      isThresh2Error: false,
      errorMessage1: "",
      errorMessage2: "",
      maxValue: max_tmp,
      observeNewThresholdValue: function () {
        var thresh1 = this.get('thresh1');
        var thresh2 = this.get('thresh2');
        if (thresh1.trim() != "") {
          if (isNaN(thresh1) || thresh1 > max_tmp || thresh1 < 0){
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Invalid! Enter a number between 0 - ' + max_tmp);
          } else if ( this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)) {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
          } else {
            this.set('isThresh1Error', false);
            this.set('errorMessage1', '');
          }
        } else {
          this.set('isThresh1Error', true);
          this.set('errorMessage1', 'This is required');
        }

        if (thresh2.trim() != "") {
          if (isNaN(thresh2) || thresh2 > max_tmp || thresh2 < 0) {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'Invalid! Enter a number between 0 - ' + max_tmp);
          } else {
            this.set('isThresh2Error', false);
            this.set('errorMessage2', '');
          }
        } else {
          this.set('isThresh2Error', true);
          this.set('errorMessage2', 'This is required');
        }

        // update the slider handles and color
        if (this.get('isThresh1Error') === false && this.get('isThresh2Error') === false) {
          $("#slider-range").slider('values', 0 , parseFloat(thresh1));
          $("#slider-range").slider('values', 1 , parseFloat(thresh2));
        }
      }.observes('thresh1', 'thresh2')

    });

    var browserVerion = this.getInternetExplorerVersion();
    App.ModalPopup.show({
      header: Em.I18n.t('dashboard.widgets.popupHeader'),
      classNames: [ 'sixty-percent-width-modal-edit-widget'],
      bodyClass: Ember.View.extend({
        templateName: require('templates/main/dashboard/edit_widget_popup'),
        configPropertyObj: configObj
      }),
      primary: Em.I18n.t('common.apply'),
      onPrimary: function () {
        configObj.observeNewThresholdValue();
        if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
          parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
          parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
          if (!App.get('testMode')) {
            var big_parent = parent.get('parentView');
            big_parent.getUserPref(big_parent.get('persistKey'));
            var oldValue = big_parent.get('currentPrefObject');
            oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
            big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
          }
          this.hide();
        }
      },

      didInsertElement: function () {
        var handlers = [configObj.get('thresh1'), configObj.get('thresh2')];
        var colors = [App.healthStatusRed, App.healthStatusOrange, App.healthStatusGreen]; //color red, orange, green

        if (browserVerion == -1 || browserVerion > 9) {
          configObj.set('isIE9', false);
          configObj.set('isGreenOrangeRed', false);
          $("#slider-range").slider({
            range: true,
            min: 0,
            max: max_tmp,
            values: handlers,
            create: function (event, ui) {
              parent.updateColors(handlers, colors);
            },
            slide: function (event, ui) {
              parent.updateColors(ui.values, colors);
              configObj.set('thresh1', ui.values[0] + '');
              configObj.set('thresh2', ui.values[1] + '');
            },
            change: function (event, ui) {
              parent.updateColors(ui.values, colors);
            }
          });
        } else {
          configObj.set('isIE9', true);
          configObj.set('isGreenOrangeRed', false);
        }
      }
    });
  }
});

});

require.register("views/main/dashboard/widgets/flume_agent_live", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.FlumeAgentUpView = App.TextDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.FlumeAgentUp'),
  id: '22',

  isPieChart: false,
  isText: true,
  isProgressBar: false,
  model_type: 'flume',

  hiddenInfo: function () {
    var result = [];
    result.pushObject(this.get('flumeAgentsLive').length + ' ' + Em.I18n.t('dashboard.services.hdfs.nodes.live'));
    result.pushObject(this.get('flumeAgentsDead').length + ' ' + Em.I18n.t('dashboard.services.hdfs.nodes.dead'));
    return result;
  }.property('flumeAgentsLive', 'flumeAgentsDead'),
  hiddenInfoClass: "hidden-info-two-line",

  thresh1: 40,
  thresh2: 70,
  maxValue: 100,

  flumeAgentComponents: function() {
    return this.get('model.hostComponents').filterProperty('componentName', 'FLUME_HANDLER');
  }.property('model.hostComponents.length'),

  flumeAgentsLive: function () {
    return this.get('flumeAgentComponents').filterProperty("workStatus", "STARTED");
  }.property('flumeAgentComponents.@each.workStatus'),

  flumeAgentsDead: function () {
    return this.get('flumeAgentComponents').filterProperty("workStatus", "INSTALLED");
  }.property('flumeAgentComponents.@each.workStatus'),


  data: function () {
    if ( !this.get('flumeAgentComponents.length')) {
      return -1;
    } else {
      return ((this.get('flumeAgentsLive').length / this.get('model.hostComponents').filterProperty('componentName', 'FLUME_HANDLER').length).toFixed(2)) * 100;
    }
  }.property('model.hostComponents.length', 'flumeAgentsLive'),

  content: function () {
    return this.get('flumeAgentsLive').length + "/" + this.get('flumeAgentComponents').length;
  }.property('flumeAgentComponents.length', 'flumeAgentsLive'),

  editWidget: function (event) {
    var parent = this;
    var max_tmp =  parseFloat(parent.get('maxValue'));
    var configObj = Ember.Object.create({
      thresh1: parent.get('thresh1') + '',
      thresh2: parent.get('thresh2') + '',
      hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint1').format(max_tmp),
      isThresh1Error: false,
      isThresh2Error: false,
      errorMessage1: "",
      errorMessage2: "",
      maxValue: max_tmp,
      observeNewThresholdValue: function () {
        var thresh1 = this.get('thresh1');
        var thresh2 = this.get('thresh2');
        if (thresh1.trim() != "") {
          if (isNaN(thresh1) || thresh1 > max_tmp || thresh1 < 0){
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Invalid! Enter a number between 0 - ' + max_tmp);
          } else if ( this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)) {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
          } else {
            this.set('isThresh1Error', false);
            this.set('errorMessage1', '');
          }
        } else {
          this.set('isThresh1Error', true);
          this.set('errorMessage1', 'This is required');
        }

        if (thresh2.trim() != "") {
          if (isNaN(thresh2) || thresh2 > max_tmp || thresh2 < 0) {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'Invalid! Enter a number between 0 - ' + max_tmp);
          } else {
            this.set('isThresh2Error', false);
            this.set('errorMessage2', '');
          }
        } else {
          this.set('isThresh2Error', true);
          this.set('errorMessage2', 'This is required');
        }

        // update the slider handles and color
        if (this.get('isThresh1Error') === false && this.get('isThresh2Error') === false) {
          $("#slider-range").slider('values', 0 , parseFloat(thresh1));
          $("#slider-range").slider('values', 1 , parseFloat(thresh2));
        }
      }.observes('thresh1', 'thresh2')

    });

    var browserVerion = this.getInternetExplorerVersion();
    App.ModalPopup.show({
      header: Em.I18n.t('dashboard.widgets.popupHeader'),
      classNames: [ 'sixty-percent-width-modal-edit-widget'],
      bodyClass: Ember.View.extend({
        templateName: require('templates/main/dashboard/edit_widget_popup'),
        configPropertyObj: configObj
      }),
      primary: Em.I18n.t('common.apply'),
      onPrimary: function () {
        configObj.observeNewThresholdValue();
        if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
          parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
          parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
          if (!App.get('testMode')) {
            var big_parent = parent.get('parentView');
            big_parent.getUserPref(big_parent.get('persistKey'));
            var oldValue = big_parent.get('currentPrefObject');
            oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
            big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
          }
          this.hide();
        }
      },

      didInsertElement: function () {
        var handlers = [configObj.get('thresh1'), configObj.get('thresh2')];
        var colors = [App.healthStatusRed, App.healthStatusOrange, App.healthStatusGreen]; //color red, orange, green

        if (browserVerion == -1 || browserVerion > 9) {
          configObj.set('isIE9', false);
          configObj.set('isGreenOrangeRed', false);
          $("#slider-range").slider({
            range: true,
            min: 0,
            max: max_tmp,
            values: handlers,
            create: function (event, ui) {
              parent.updateColors(handlers, colors);
            },
            slide: function (event, ui) {
              parent.updateColors(ui.values, colors);
              configObj.set('thresh1', ui.values[0] + '');
              configObj.set('thresh2', ui.values[1] + '');
            },
            change: function (event, ui) {
              parent.updateColors(ui.values, colors);
            }
          });
        } else {
          configObj.set('isIE9', true);
          configObj.set('isGreenOrangeRed', false);
        }
      }
    });
  }
});

});

require.register("views/main/dashboard/widgets/hbase_average_load", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HBaseAverageLoadView = App.TextDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.HBaseAverageLoad'),
  id: '14',

  model_type: 'hbase',
  hiddenInfo: function () {
    var avgLoad = this.get('model.averageLoad');
    if (isNaN(avgLoad)) {
      avgLoad = Em.I18n.t('services.service.summary.notAvailable');
    }
    var result = [];
    result.pushObject(Em.I18n.t('dashboard.services.hbase.averageLoadPerServer').format(avgLoad));
    return result;
  }.property("model.averageLoad"),

  isGreen: function () {
    return this.get('data') <= this.get('thresh1');
  }.property('data','thresh1','thresh2'),
  isRed: function () {
    return this.get('data') > this.get('thresh2');
  }.property('data','thresh1','thresh2'),
  isNA: function (){
    return this.get('data') === null || isNaN(this.get('data'));
  }.property('data'),

  thresh1: 0.5,
  thresh2: 2,
  maxValue: 'infinity',

  data: function () {
    return this.get('model.averageLoad');
  }.property("model.averageLoad"),

  content: function (){
    if(this.get('data') || this.get('data') == 0){
      return this.get('data') + "";
    }else{
      return Em.I18n.t('services.service.summary.notAvailable');
    }
  }.property('model.averageLoad'),

  editWidget: function (event) {
    var parent = this;
    var configObj = Ember.Object.create({
      thresh1: parent.get('thresh1') + '',
      thresh2: parent.get('thresh2') + '',
      hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint2'),
      isThresh1Error: false,
      isThresh2Error: false,
      errorMessage1: "",
      errorMessage2: "",
      maxValue: 'infinity',
      observeNewThresholdValue: function () {
        var thresh1 = this.get('thresh1');
        var thresh2 = this.get('thresh2');
        if (thresh1.trim() != "") {
          if (isNaN(thresh1) || thresh1 < 0) {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Invalid! Enter a number larger than 0');
          } else if ( this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)){
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
          } else {
            this.set('isThresh1Error', false);
            this.set('errorMessage1', '');
          }
        } else {
          this.set('isThresh1Error', true);
          this.set('errorMessage1', 'This is required');
        }

        if (thresh2.trim() != "") {
          if (isNaN(thresh2) || thresh2 < 0) {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'Invalid! Enter a number larger than 0');
          } else {
            this.set('isThresh2Error', false);
            this.set('errorMessage2', '');
          }
        } else {
          this.set('isThresh2Error', true);
          this.set('errorMessage2', 'This is required');
        }

      }.observes('thresh1', 'thresh2')

    });

    var browserVerion = this.getInternetExplorerVersion();
    App.ModalPopup.show( {
      header: Em.I18n.t('dashboard.widgets.popupHeader'),
      classNames: [ 'sixty-percent-width-modal-edit-widget'],
      bodyClass: Ember.View.extend({
        templateName: require('templates/main/dashboard/edit_widget_popup'),
        configPropertyObj: configObj
      }),
      primary: Em.I18n.t('common.apply'),
      onPrimary: function () {
        configObj.observeNewThresholdValue();
        if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
          parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
          parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
          if (!App.get('testMode')) {
            //save to persist
            var big_parent = parent.get('parentView');
            big_parent.getUserPref(big_parent.get('persistKey'));
            var oldValue = big_parent.get('currentPrefObject');
            oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
            big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
          }

          this.hide();
        }
      },

      didInsertElement: function () {
        var colors = [App.healthStatusGreen, App.healthStatusOrange, App.healthStatusRed]; //color green, orange ,red
        var handlers = [33, 66]; //fixed value

        if (browserVerion == -1 || browserVerion > 9) {
          configObj.set('isIE9', false);
          configObj.set('isGreenOrangeRed', true);
          $("#slider-range").slider({
            range:true,
            disabled:true, //handlers cannot move
            min: 0,
            max: 100,
            values: handlers,
            create: function (event, ui) {
              parent.updateColors(handlers, colors);
            }
          });
        } else {
          configObj.set('isIE9', true);
          configObj.set('isGreenOrangeRed', true);
        }
      }
    });
  }

});

});

require.register("views/main/dashboard/widgets/hbase_links", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HBaseLinksView = App.LinkDashboardWidgetView.extend({

  templateName: require('templates/main/dashboard/widgets/hbase_links'),
  title: Em.I18n.t('dashboard.widgets.HBaseLinks'),
  id: '12',

  model_type: 'hbase',

  port: '60010',

  componentName: 'HBASE_REGIONSERVER',

  /**
   * All master components
   */
  masters: function () {
    return this.get('model.hostComponents').filterProperty('isMaster', true);
  }.property('model.hostComponents.@each'),
  /**
   * Passive master components
   */
  passiveMasters: function () {
    return this.get('masters').filterProperty('haStatus', 'false');
  }.property('masters'),
  /**
   * One(!) active master component
   */
  activeMaster: function () {
    return this.get('masters').findProperty('haStatus', 'true');
  }.property('masters'),

  activeMasterTitle: function(){
    return this.t('service.hbase.activeMaster');
  }.property('activeMaster'),

  hbaseMasterWebUrl: function () {
    if (this.get('activeMaster.host') && this.get('activeMaster.host').get('publicHostName')) {
      return "http://" + this.get('activeMaster.host').get('publicHostName') + ':' + this.get('port');
    }
    return '';
  }.property('activeMaster'),

  calcWebUrl: function() {
    return '';
  }

});
});

require.register("views/main/dashboard/widgets/hbase_master_heap", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var numberUtils = require('utils/number_utils');

App.HBaseMasterHeapPieChartView = App.PieChartDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.HBaseMasterHeap'),
  id: '13',

  model_type: 'hbase',
  modelFieldMax: 'heapMemoryMax',
  modelFieldUsed: 'heapMemoryUsed',
  widgetHtmlId: 'widget-hbase-heap',

  didInsertElement: function() {
    this._super();
    this.calc();
  },

  getUsed: function() {
    return (this.get('model').get(this.get('modelFieldUsed')) / (1024 * 1024)) || 0;
  },

  getMax: function() {
    return (this.get('model').get(this.get('modelFieldMax')) / (1024 * 1024)) || 0;
  }
});
});

require.register("views/main/dashboard/widgets/hbase_master_uptime", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HBaseMasterUptimeView = App.UptimeTextDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.HBaseMasterUptime'),
  id: '16',

  model_type: 'hbase',

  component: 'Hbase Master',
  modelField: 'masterStartTime',

  didInsertElement: function() {
    this._super();
    this.calc();
  }

});
});

require.register("views/main/dashboard/widgets/hbase_regions_in_transition", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HBaseRegionsInTransitionView = App.TextDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.HBaseRegionsInTransition'),
  id: '15',

  model_type: 'hbase',
  hiddenInfo: function () {
    var result = [];
    result.pushObject(this.get("model.regionsInTransition") + " regions");
    result.pushObject("in transition");
    return result;
  }.property("model.regionsInTransition"),

  classNameBindings: ['isRed', 'isOrange', 'isGreen', 'isNA'],
  isGreen: function () {
    return this.get('data') <= this.get('thresh1');
  }.property('data','thresh1'),
  isOrange: function () {
    return (this.get('data') <= this.get('thresh2') && this.get('data') > this.get('thresh1') );
  }.property('data','thresh1','thresh2'),
  isRed: function () {
    return this.get('data') > this.get('thresh2');
  }.property('data','thresh2'),
  isNA: function () {
    return this.get('data') === null;
  }.property('data'),

  thresh1: 0.5,
  thresh2: 2,
  maxValue: 'infinity',

  data: function () {
    return this.get('model.regionsInTransition');
  }.property("model.regionsInTransition"),

  content: function (){
    return this.get('data') + "";
  }.property('model.regionsInTransition'),

  editWidget: function (event) {
    var parent = this;
    var configObj = Ember.Object.create({
      thresh1: parent.get('thresh1') + '',
      thresh2: parent.get('thresh2') + '',
      hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint2'),
      isThresh1Error: false,
      isThresh2Error: false,
      errorMessage1: "",
      errorMessage2: "",
      maxValue: 'infinity',
      observeNewThresholdValue: function () {
        var thresh1 = this.get('thresh1');
        var thresh2 = this.get('thresh2');
        if (thresh1.trim() != "") {
          if (isNaN(thresh1) || thresh1 < 0) {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Invalid! Enter a number larger than 0');
          } else if ( this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)){
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
          } else {
            this.set('isThresh1Error', false);
            this.set('errorMessage1', '');
          }
        } else {
          this.set('isThresh1Error', true);
          this.set('errorMessage1', 'This is required');
        }

        if (thresh2.trim() != "") {
          if (isNaN(thresh2) || thresh2 < 0) {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'Invalid! Enter a number larger than 0');
          } else {
            this.set('isThresh2Error', false);
            this.set('errorMessage2', '');
          }
        } else {
          this.set('isThresh2Error', true);
          this.set('errorMessage2', 'This is required');
        }
      }.observes('thresh1', 'thresh2')

    });

    var browserVerion = this.getInternetExplorerVersion();
    App.ModalPopup.show({
      header: Em.I18n.t('dashboard.widgets.popupHeader'),
      classNames: [ 'sixty-percent-width-modal-edit-widget'],
      bodyClass: Ember.View.extend({
        templateName: require('templates/main/dashboard/edit_widget_popup'),
        configPropertyObj: configObj
      }),
      primary: Em.I18n.t('common.apply'),
      onPrimary: function () {
        configObj.observeNewThresholdValue();
        if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
          parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
          parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
          if (!App.get('testMode')) {
            //save to persist
            var big_parent = parent.get('parentView');
            big_parent.getUserPref(big_parent.get('persistKey'));
            var oldValue = big_parent.get('currentPrefObject');
            oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
            big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
          }

          this.hide();
        }
      },

      didInsertElement: function () {
        var colors = [App.healthStatusGreen, App.healthStatusOrange, App.healthStatusRed]; //color green, orange ,red
        var handlers = [33, 66]; //fixed value

        if (browserVerion == -1 || browserVerion > 9) {
          configObj.set('isIE9', false);
          configObj.set('isGreenOrangeRed', true);
          $("#slider-range").slider({
            range:true,
            disabled:true, //handlers cannot move
            min: 0,
            max: 100,
            values: handlers,
            create: function (event, ui) {
              parent.updateColors(handlers, colors);
            }
          });
        } else {
          configObj.set('isIE9', true);
          configObj.set('isGreenOrangeRed', true);
        }
      }
    });
  }

});

});

require.register("views/main/dashboard/widgets/hdfs_capacity", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var numberUtils = require('utils/number_utils');

App.NameNodeCapacityPieChartView = App.PieChartDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.HDFSDiskUsage'),
  id: '2',

  model_type: 'hdfs',
  modelFieldMax: 'capacityTotal',
  /**
   * HDFS model has 'remaining' value, but not 'used'
   */
  modelFieldUsed: 'capacityRemaining',
  widgetHtmlId: 'widget-nn-capacity',
  hiddenInfoClass: "hidden-info-six-line",

  didInsertElement: function() {
    this._super();
    this.calc();
  },

  calcHiddenInfo: function () {
    var total = this.get('model').get(this.get('modelFieldMax')) + 0;
    var remaining = this.get('model.capacityRemaining');
    var dfsUsed = this.get('model.capacityUsed');
    var nonDfsUsed = total !== null && remaining !== null && dfsUsed !== null ? total - remaining - dfsUsed : null;
    var dfsPercent = total > 0 ? ((dfsUsed * 100) / total).toFixed(2) : 0;
    var nonDfsPercent = total > 0 ? ((nonDfsUsed * 100) / total).toFixed(2) : 0;
    var remainingPercent = total > 0 ? ((remaining * 100) / total).toFixed(2) : 0;
    if (dfsPercent == "NaN" || dfsPercent < 0) {
      dfsPercent = Em.I18n.t('services.service.summary.notAvailable') + " ";
    }
    if (nonDfsPercent == "NaN" || nonDfsPercent < 0) {
      nonDfsPercent = Em.I18n.t('services.service.summary.notAvailable') + " ";
    }
    if (remainingPercent == "NaN" || remainingPercent < 0) {
      remainingPercent = Em.I18n.t('services.service.summary.notAvailable') + " ";
    }
    var result = [];
    result.pushObject(Em.I18n.t('dashboard.widgets.HDFSDiskUsage.DFSused'));
    result.pushObject(Em.I18n.t('dashboard.widgets.HDFSDiskUsage.info').format(numberUtils.bytesToSize(dfsUsed, 1, 'parseFloat'), dfsPercent));
    result.pushObject(Em.I18n.t('dashboard.widgets.HDFSDiskUsage.nonDFSused'));
    result.pushObject(Em.I18n.t('dashboard.widgets.HDFSDiskUsage.info').format(numberUtils.bytesToSize(nonDfsUsed, 1, 'parseFloat'), nonDfsPercent));
    result.pushObject(Em.I18n.t('dashboard.widgets.HDFSDiskUsage.remaining'));
    result.pushObject(Em.I18n.t('dashboard.widgets.HDFSDiskUsage.info').format(numberUtils.bytesToSize(remaining, 1, 'parseFloat'), remainingPercent));
    return result;
  },

  calcDataForPieChart: function() {
    var total = this.get('model').get(this.get('modelFieldMax')) * 1024 * 1024;
    var used = total - this.get('model').get(this.get('modelFieldUsed')) * 1024 * 1024;
    var percent = total > 0 ? ((used)*100 / total).toFixed() : 0;
    var percent_precise = total > 0 ? ((used)*100 / total).toFixed(1) : 0;
    return [percent, percent_precise];
  }

});

});

require.register("views/main/dashboard/widgets/hdfs_links", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HDFSLinksView = App.LinkDashboardWidgetView.extend({

  templateName: require('templates/main/dashboard/widgets/hdfs_links'),
  title: Em.I18n.t('dashboard.widgets.HDFSLinks'),
  id: '11',

  model_type: 'hdfs',

  port: '50070',

  componentName: 'DATANODE',

  modelField: 'nameNode',

  didInsertElement: function() {
    this._super();
    this.calc();
  },

  isHAEnabled: function() {
    return !this.get('model.snameNode');
  }.property('model.snameNode'),
  isActiveNNValid: function () {
    return this.get('model.activeNameNode') != null;
  }.property('model.activeNameNode'),
  isStandbyNNValid: function () {
    return this.get('model.standbyNameNode') != null;
  }.property('model.standbyNameNode'),
  isTwoStandbyNN: function () {
    return (this.get('model.standbyNameNode') != null && this.get('model.standbyNameNode2') != null);
  }.property('model.standbyNameNode', 'model.standbyNameNode2'),
  twoStandbyComponent: function () {
    return App.HostComponent.find().findProperty('componentName', 'NAMENODE');
  }.property()
});
});

require.register("views/main/dashboard/widgets/links_widget", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var date = require('utils/date');

App.LinkDashboardWidgetView = App.DashboardWidgetView.extend({

  componentName: null,

  component: function() {
    return App.HostComponent.find().findProperty('componentName', this.get('componentName'));
  }.property(),

  port: null,

  modelField: null,

  webUrl: null,

  didInsertElement: function() {
    this._super();
    this.addObserver('model.' + this.get('modelField'), this, this.calc);
  },
  calc: function() {
    this.set('webUrl', this.calcWebUrl());
  },
  calcWebUrl: function() {
    if (this.get('model')) {
      return "http://" + this.get('model').get(this.get('modelField')).get('publicHostName') + ':' + this.get('port');
    }
    return '';
  }
});

});

require.register("views/main/dashboard/widgets/metrics_cpu", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ChartClusterMetricsCPUWidgetView = App.ClusterMetricsDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.clusterMetrics.cpu'),
  id: '8',

  content: App.ChartClusterMetricsCPU.extend({
    noTitleUnderGraph: true,
    inWidget: true
  })
});
});

require.register("views/main/dashboard/widgets/metrics_load", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ChartClusterMetricsLoadWidgetView = App.ClusterMetricsDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.clusterMetrics.load'),
  id: '9',

  content: App.ChartClusterMetricsLoad.extend({
    noTitleUnderGraph: true,
    inWidget: true
  })
});
});

require.register("views/main/dashboard/widgets/metrics_memory", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ChartClusterMetricsMemoryWidgetView = App.ClusterMetricsDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.clusterMetrics.memory'),
  id: '6',

  content: App.ChartClusterMetricsMemory.extend({
    noTitleUnderGraph: true,
    inWidget: true
  })
});
});

require.register("views/main/dashboard/widgets/metrics_network", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.ChartClusterMetricsNetworkWidgetView = App.ClusterMetricsDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.clusterMetrics.network'),
  id: '7',

  content: App.ChartClusterMetricsNetwork.extend({
    noTitleUnderGraph: true,
    inWidget: true
  })
});
});

require.register("views/main/dashboard/widgets/namenode_cpu", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.NameNodeCpuPieChartView = App.PieChartDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.NameNodeCpu'),
  id: '3',

  model_type: 'hdfs',
  widgetHtmlId: 'widget-nn-cpu',
  cpuWio: null,
  nnHostName:"",
  intervalId: null,

  willDestroyElement: function() {
    clearInterval(this.get("intervalId"));
  },

  didInsertElement: function() {
    this._super();
    var self = this,
    intervalId;
    if (App.get('isHaEnabled')) {
      this.set('nnHostName', this.get('model').get('activeNameNode.hostName'));
    }else{
     this.set('nnHostName', this.get('model').get('nameNode.hostName'));
    }
    if (this.get('nnHostName')) {
      this.getValue();
      intervalId = setInterval(function () {
        self.getValue()
      }, App.componentsUpdateInterval);
      this.set('intervalId', intervalId);
    }
  },

  getValue: function () {
    App.ajax.send({
      name: 'namenode.cpu_wio',
      sender: this,
      data: {
        nnHost: this.get('nnHostName')
      },
      success: 'updateValueSuccess',
      error: 'updateValueError'
    });
  },

  updateValueError: function () {
    this.calc();
  },

  updateValueSuccess: function (response) {
    this.set('cpuWio', Em.get(response, 'metrics.cpu.cpu_wio'));
    this.calc();
  },

  calcHiddenInfo: function() {
    var value = this.get('cpuWio');
    var obj1;
    if( value == null) {
      obj1 = Em.I18n.t('services.service.summary.notAvailable');
    }
    else {
      value = value >= 100 ? 100: value;
      obj1 = (value + 0).toFixed(2) + '%';
    }
    var result = [];
    result.pushObject(obj1);
    result.pushObject('CPU wait I/O');
    return result;
  },

  calcIsPieExists: function() {
    return (!Em.isNone(this.get('cpuWio')));
  },

  calcDataForPieChart: function () {
    var value = this.get('cpuWio');
    value = value >= 100 ? 100 : value;
    var percent = (value + 0).toFixed(1);
    var percent_precise = (value + 0).toFixed(2);
    return [ percent, percent_precise];
  }
});
});

require.register("views/main/dashboard/widgets/namenode_heap", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var numberUtils = require('utils/number_utils');

App.NameNodeHeapPieChartView = App.PieChartDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.NameNodeHeap'),
  id: '1',

  model_type: 'hdfs',
  modelFieldMax: 'jvmMemoryHeapMax',
  modelFieldUsed: 'jvmMemoryHeapUsed',
  widgetHtmlId: 'widget-nn-heap',

  getUsed: function() {
    return (this.get('model').get(this.get('modelFieldUsed')) / (1024 * 1024)) || 0;
  },

  getMax: function() {
    return (this.get('model').get(this.get('modelFieldMax')) / (1024 * 1024)) || 0;
  },

  didInsertElement: function() {
    this._super();
    this.calc();
  }
});
});

require.register("views/main/dashboard/widgets/namenode_rpc", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.NameNodeRpcView = App.TextDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.NameNodeRpc'),
  id: '5',

  model_type: 'hdfs',
  hiddenInfo: function () {
    var result = [];
    result.pushObject(this.get('content') + ' average RPC');
    result.pushObject('queue wait time');
    return result;
  }.property('model.nameNodeRpc'),

  thresh1: 0.5,
  thresh2: 2,
  maxValue: 'infinity',

  isGreen: function () {
    return (this.get('data') <= this.get('thresh1'));
  }.property('data','thresh1'),

  isRed: function () {
    return (this.get('data') > this.get('thresh2'));
  }.property('data','thresh2'),

  data: function () {
    if (this.get('model.nameNodeRpc')) {
      return (this.get('model.nameNodeRpc')).toFixed(2);
    } else {
      if (this.get('model.nameNodeRpc') == 0) {
        return 0;
      } else {
        return null;
      }
    }
  }.property('model.nameNodeRpc'),

  content: function () {
    if (this.get('data') || this.get('data') == 0) {
      return this.get('data') + " ms";
    }
    else {
      return Em.I18n.t('services.service.summary.notAvailable');
    }
  }.property('model.nameNodeRpc'),

  editWidget: function (event) {
    var parent = this;
    var configObj = Ember.Object.create({
      thresh1: parent.get('thresh1') + '',
      thresh2: parent.get('thresh2') + '',
      hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint3'),
      isThresh1Error: false,
      isThresh2Error: false,
      errorMessage1: "",
      errorMessage2: "",
      maxValue: 'infinity',
      observeNewThresholdValue: function () {
        var thresh1 = this.get('thresh1');
        var thresh2 = this.get('thresh2');
        if (thresh1.trim() != "") {
          if (isNaN(thresh1) || thresh1 < 0) {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Invalid! Enter a number larger than 0');
          } else if ( this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)){
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
          } else {
            this.set('isThresh1Error', false);
            this.set('errorMessage1', '');
          }
        } else {
          this.set('isThresh1Error', true);
          this.set('errorMessage1', 'This is required');
        }

        if (thresh2.trim() != "") {
          if (isNaN(thresh2) || thresh2 < 0) {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'Invalid! Enter a number larger than 0');
          } else {
            this.set('isThresh2Error', false);
            this.set('errorMessage2', '');
          }
        } else {
          this.set('isThresh2Error', true);
          this.set('errorMessage2', 'This is required');
        }
      }.observes('thresh1', 'thresh2')

    });

    var browserVerion = this.getInternetExplorerVersion();
    App.ModalPopup.show({
      header: Em.I18n.t('dashboard.widgets.popupHeader'),
      classNames: ['sixty-percent-width-modal-edit-widget'],
      bodyClass: Ember.View.extend({
        templateName: require('templates/main/dashboard/edit_widget_popup'),
        configPropertyObj: configObj
      }),
      primary: Em.I18n.t('common.apply'),
      onPrimary: function () {
        configObj.observeNewThresholdValue();
        if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
          parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
          parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
          if (!App.get('testMode')) {
            //save to persist
            var big_parent = parent.get('parentView');
            big_parent.getUserPref(big_parent.get('persistKey'));
            var oldValue = big_parent.get('currentPrefObject');
            oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
            big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
          }

          this.hide();
        }
      },

      didInsertElement: function () {
        var colors = [App.healthStatusGreen, App.healthStatusOrange, App.healthStatusRed]; //color green, orange ,red
        var handlers = [33, 66]; //fixed value

        if (browserVerion == -1 || browserVerion > 9) {
          configObj.set('isIE9', false);
          configObj.set('isGreenOrangeRed', true);
          $("#slider-range").slider({
            range:true,
            disabled:true, //handlers cannot move
            min: 0,
            max: 100,
            values: handlers,
            create: function (event, ui) {
              parent.updateColors(handlers, colors);
            }
          });
        } else {
          configObj.set('isIE9', true);
          configObj.set('isGreenOrangeRed', true);
        }
      }
    });
  }

});

});

require.register("views/main/dashboard/widgets/namenode_uptime", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.NameNodeUptimeView = App.UptimeTextDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.NameNodeUptime'),
  id: '10',

  model_type: 'hdfs',

  component: 'NameNode',
  modelField: 'nameNodeStartTime',

  didInsertElement: function() {
    this._super();
    this.calc();
  }

});
});

require.register("views/main/dashboard/widgets/node_managers_live", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.NodeManagersLiveView = App.TextDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.NodeManagersLive'),
  id: '19',

  model_type: 'yarn',

  hiddenInfo: function () {
    var nmActive = this.get('model.nodeManagersCountActive') == null ? Em.I18n.t('services.service.summary.notAvailable') : this.get('model.nodeManagersCountActive');
    var nmLost = this.get('model.nodeManagersCountLost') == null ? Em.I18n.t('services.service.summary.notAvailable') : this.get('model.nodeManagersCountLost');
    var nmUnhealthy = this.get('model.nodeManagersCountUnhealthy') == null ? Em.I18n.t('services.service.summary.notAvailable') : this.get('model.nodeManagersCountUnhealthy');
    var nmRebooted = this.get('model.nodeManagersCountRebooted') == null ? Em.I18n.t('services.service.summary.notAvailable'): this.get('model.nodeManagersCountRebooted');
    var nmDecom = this.get('model.nodeManagersCountDecommissioned') == null ? Em.I18n.t('services.service.summary.notAvailable') : this.get('model.nodeManagersCountDecommissioned');
    var result = [];
      result.pushObject(nmActive + " active");
      result.pushObject(nmLost + " lost");
      result.pushObject(nmUnhealthy + " unhealthy");
      result.pushObject(nmRebooted + " rebooted");
      result.pushObject(nmDecom + " decommissioned");
    return result;
  }.property('model.nodeManagersCountActive', 'model.nodeManagersCountLost',
    'model.nodeManagersCountUnhealthy', 'model.nodeManagersCountRebooted', 'model.nodeManagersCountDecommissioned'),
  hiddenInfoClass: "hidden-info-five-line",

  thresh1: 40,
  thresh2: 70,
  maxValue: 100,

  nodeManagersLive: function () {
    return this.get('model.nodeManagersCountActive');
  }.property('model.nodeManagersCountActive'),

  data: function () {
    var nodeManagers = this.get('model.nodeManagersTotal');
    var nodeManagersLive = this.get('nodeManagersLive');
    if (nodeManagersLive == null) {
      return null;
    } else {
      return (nodeManagersLive / nodeManagers).toFixed(2) * 100;
    }
  }.property('model.nodeManagersTotal', 'nodeManagersLive'),

  content: function () {
    return this.get('nodeManagersLive') == null ? Em.I18n.t('services.service.summary.notAvailable') : this.get('nodeManagersLive') + '/' + this.get('model.nodeManagersTotal');
  }.property('model.nodeManagersTotal', 'nodeManagersLive'),

  editWidget: function (event) {
    var parent = this;
    var max_tmp =  parseFloat(parent.get('maxValue'));
    var configObj = Ember.Object.create({
      thresh1: parent.get('thresh1') + '',
      thresh2: parent.get('thresh2') + '',
      hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint1').format(max_tmp),
      isThresh1Error: false,
      isThresh2Error: false,
      errorMessage1: "",
      errorMessage2: "",
      maxValue: max_tmp,
      observeNewThresholdValue: function () {
        var thresh1 = this.get('thresh1');
        var thresh2 = this.get('thresh2');
        if (thresh1.trim() != "") {
          if (isNaN(thresh1) || thresh1 > max_tmp || thresh1 < 0) {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Invalid! Enter a number between 0 - ' + max_tmp);
          } else if (this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)){
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
          } else {
            this.set('isThresh1Error', false);
            this.set('errorMessage1', '');
          }
        } else {
          this.set('isThresh1Error', true);
          this.set('errorMessage1', 'This is required');
        }

        if (thresh2.trim() != "") {
          if (isNaN(thresh2) || thresh2 > max_tmp || thresh2 < 0) {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'Invalid! Enter a number between 0 - ' + max_tmp);
          } else {
            this.set('isThresh2Error', false);
            this.set('errorMessage2', '');
          }
        } else {
          this.set('isThresh2Error', true);
          this.set('errorMessage2', 'This is required');
        }

        // update the slider handles and color
        if (this.get('isThresh1Error') === false && this.get('isThresh2Error') === false) {
          $("#slider-range").slider('values', 0 , parseFloat(thresh1));
          $("#slider-range").slider('values', 1 , parseFloat(thresh2));
        }
      }.observes('thresh1', 'thresh2')
    });

    var browserVerion = this.getInternetExplorerVersion();
    App.ModalPopup.show({
      header: 'Customize Widget',
      classNames: [ 'sixty-percent-width-modal-edit-widget'],
      bodyClass: Ember.View.extend({
        templateName: require('templates/main/dashboard/edit_widget_popup'),
        configPropertyObj: configObj
      }),
      primary: Em.I18n.t('common.apply'),
      onPrimary: function() {
        configObj.observeNewThresholdValue();
        if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
          parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
          parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
          if (!App.get('testMode')) {
            //save to persit
            var big_parent = parent.get('parentView');
            big_parent.getUserPref(big_parent.get('persistKey'));
            var oldValue = big_parent.get('currentPrefObject');
            oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
            big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
          }
          this.hide();
        }
      },

      didInsertElement: function () {
        var self = this;
        var handlers = [configObj.get('thresh1'), configObj.get('thresh2')];
        var colors = [App.healthStatusRed, App.healthStatusOrange, App.healthStatusGreen]; //color red, orange, green

        if (browserVerion == -1 || browserVerion > 9) {
          configObj.set('isIE9', false);
          configObj.set('isGreenOrangeRed', false);
          $("#slider-range").slider({
            range: true,
            min: 0,
            max: max_tmp,
            values: handlers,
            create: function (event, ui) {
              parent.updateColors(handlers,colors);
            },
            slide: function (event, ui) {
              parent.updateColors(ui.values,colors);
              configObj.set('thresh1', ui.values[0] + '');
              configObj.set('thresh2', ui.values[1] + '');
            },
            change: function (event, ui) {
              parent.updateColors(ui.values,colors);
            }
          });
        } else {
          configObj.set('isIE9', true);
          configObj.set('isGreenOrangeRed', false);
        }
      }
    });
  }

});

});

require.register("views/main/dashboard/widgets/pie_chart_widget", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var numberUtils = require('utils/number_utils');

App.PieChartDashboardWidgetView = App.DashboardWidgetView.extend({

  templateName: require('templates/main/dashboard/widgets/pie_chart'),

  maxValue: 100,

  modelFieldMax: null,
  modelFieldUsed: null,

  hiddenInfo: null,

  isPieExist: null,

  dataForPieChart: null,

  widgetHtmlId: null,

  getUsed: function() {
    return this.get('model').get(this.get('modelFieldUsed')) || 0;
  },

  getMax: function() {
    return this.get('model').get(this.get('modelFieldMax')) || 0;
  },

  calcHiddenInfo: function() {
    var used = this.getUsed();
    var max = this.getMax();
    var percent = max > 0 ? 100 * used / max : 0;
    var result = [];
    result.pushObject(percent.toFixed(1) + '% used');
    result.pushObject(numberUtils.bytesToSize(used, 1, 'parseFloat', 1024 * 1024) + ' of ' + numberUtils.bytesToSize(max, 1, 'parseFloat', 1024 * 1024));
    return result;
  },

  calcIsPieExists: function() {
    return (this.get('model').get(this.get('modelFieldMax')) > 0);
  },

  calcDataForPieChart: function() {
    var used = this.getUsed();
    var total = this.getMax();
    var percent = total > 0 ? ((used)*100 / total).toFixed() : 0;
    var percent_precise = total > 0 ? ((used)*100 / total).toFixed(1) : 0;
    return [percent, percent_precise];
  },

  calc: function() {
    this.set('hiddenInfo', this.calcHiddenInfo());
    var isPieExists = this.calcIsPieExists();
    this.set('isPieExist', isPieExists);
    if (isPieExists) {
      this.set('dataForPieChart', this.calcDataForPieChart());
    }
  },

  didInsertElement: function() {
    this._super();
    this.addObserver('model.' + this.get('modelFieldMax'), this, this.calc);
    this.addObserver('model.' + this.get('modelFieldUsed'), this, this.calc);
  },

  content: App.ChartPieView.extend({
    model: null,  //data bind here
    id: function() {
      return this.get('parentView.widgetHtmlId');
    }.property('parentView.widgetHtmlId'), // html id
    stroke: '#D6DDDF', //light grey
    thresh1: null, //bind from parent
    thresh2: null,
    innerR: 25,

    existCenterText: true,
    centerTextColor: function () {
      return this.get('contentColor');
    }.property('contentColor'),

    palette: new Rickshaw.Color.Palette({
      scheme: [ '#FFFFFF', '#D6DDDF'].reverse()
    }),

    data: function() {
      var ori_data = this.get('parentView.dataForPieChart');
      return [ ori_data[0], 100 - ori_data[0]];
    }.property(),

    setData: function() {
      this.set('data', this.get('parentView.dataForPieChart'));
    }.observes('parentView.dataForPieChart'),

    contentColor: function () {
      var used = parseFloat(this.get('parentView.dataForPieChart')[1]);
      var thresh1 = parseFloat(this.get('thresh1'));
      var thresh2 = parseFloat(this.get('thresh2'));
      var color_green = App.healthStatusGreen;
      var color_red = App.healthStatusRed;
      var color_orange = App.healthStatusOrange;
      if (used <= thresh1) {
        this.set('palette', new Rickshaw.Color.Palette({
          scheme: [ '#FFFFFF', color_green  ].reverse()
        }));
        return color_green;
      } else if (used <= thresh2) {
        this.set('palette', new Rickshaw.Color.Palette({
          scheme: [ '#FFFFFF', color_orange  ].reverse()
        }));
        return color_orange;
      } else {
        this.set('palette', new Rickshaw.Color.Palette({
          scheme: [ '#FFFFFF', color_red  ].reverse()
        }));
        return color_red;
      }
    }.property('data', 'thresh1', 'thresh2'),

    // refresh text and color when data in model changed
    refreshSvg: function () {
      // remove old svg
      var old_svg =  $("#" + this.get('id'));
      if(old_svg){
        old_svg.remove();
      }

      // draw new svg
      this.appendSvg();
    }.observes('data', 'thresh1', 'thresh2')
  })
});

});

require.register("views/main/dashboard/widgets/resource_manager_heap", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var numberUtils = require('utils/number_utils');

App.ResourceManagerHeapPieChartView = App.PieChartDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.ResourceManagerHeap'),
  id: '17',

  model_type: 'yarn',
  modelFieldMax: 'jvmMemoryHeapMax',
  modelFieldUsed: 'jvmMemoryHeapUsed',
  widgetHtmlId: 'widget-rm-heap',

  getUsed: function() {
    return (this.get('model').get(this.get('modelFieldUsed')) / (1024 * 1024)) || 0;
  },

  getMax: function() {
    return (this.get('model').get(this.get('modelFieldMax')) / (1024 * 1024)) || 0;
  },

  didInsertElement: function() {
    this._super();
    this.calc();
  }

});
});

require.register("views/main/dashboard/widgets/resource_manager_uptime", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var date = require('utils/date');

App.ResourceManagerUptimeView = App.UptimeTextDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.ResourceManagerUptime'),
  id: '18',

  model_type: 'yarn',

  component: 'ResourceManager',
  modelField: 'resourceManagerStartTime',

  didInsertElement: function() {
    this._super();
    this.calc();
  }

});

});

require.register("views/main/dashboard/widgets/supervisor_live", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.SuperVisorUpView = App.TextDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.SuperVisorUp'),
  id: '21',

  isPieChart: false,
  isText: true,
  isProgressBar: false,
  model_type: 'storm',

  hiddenInfo: function () {
    var result = [];
    result.pushObject(this.get('superVisorsLive') + ' ' + Em.I18n.t('dashboard.services.hdfs.nodes.live'));
    result.pushObject(this.get('superVisorsDead') + ' ' + Em.I18n.t('dashboard.services.hdfs.nodes.dead'));
    return result;
  }.property('superVisorsLive', 'superVisorsDead'),
  hiddenInfoClass: "hidden-info-two-line",

  thresh1: 40,
  thresh2: 70,
  maxValue: 100,

  superVisorsLive: function () {
    return this.get('model.superVisorsStarted');
  }.property('model.superVisorsStarted'),

  superVisorsDead: function () {
    return this.get('model.superVisorsInstalled');
  }.property('model.superVisorsInstalled'),

  superVisorsTotal: function() {
    return this.get('model.superVisorsTotal');
  }.property('model.superVisorsTotal'),

  data: function () {
    if ( !this.get('superVisorsTotal')) {
      return -1;
    } else {
      return ((this.get('superVisorsLive') / this.get('superVisorsTotal')).toFixed(2)) * 100;
    }
  }.property('superVisorsTotal', 'superVisorsLive'),

  content: function () {
    return this.get('superVisorsLive') + "/" + this.get('superVisorsTotal');
  }.property('superVisorsLive', 'superVisorsTotal'),

  editWidget: function (event) {
    var parent = this;
    var max_tmp =  parseFloat(parent.get('maxValue'));
    var configObj = Ember.Object.create({
      thresh1: parent.get('thresh1') + '',
      thresh2: parent.get('thresh2') + '',
      hintInfo: Em.I18n.t('dashboard.widgets.hintInfo.hint1').format(max_tmp),
      isThresh1Error: false,
      isThresh2Error: false,
      errorMessage1: "",
      errorMessage2: "",
      maxValue: max_tmp,
      observeNewThresholdValue: function () {
        var thresh1 = this.get('thresh1');
        var thresh2 = this.get('thresh2');
        if (thresh1.trim() != "") {
          if (isNaN(thresh1) || thresh1 > max_tmp || thresh1 < 0){
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Invalid! Enter a number between 0 - ' + max_tmp);
          } else if ( this.get('isThresh2Error') === false && parseFloat(thresh2)<= parseFloat(thresh1)) {
            this.set('isThresh1Error', true);
            this.set('errorMessage1', 'Threshold 1 should be smaller than threshold 2 !');
          } else {
            this.set('isThresh1Error', false);
            this.set('errorMessage1', '');
          }
        } else {
          this.set('isThresh1Error', true);
          this.set('errorMessage1', 'This is required');
        }

        if (thresh2.trim() != "") {
          if (isNaN(thresh2) || thresh2 > max_tmp || thresh2 < 0) {
            this.set('isThresh2Error', true);
            this.set('errorMessage2', 'Invalid! Enter a number between 0 - ' + max_tmp);
          } else {
            this.set('isThresh2Error', false);
            this.set('errorMessage2', '');
          }
        } else {
          this.set('isThresh2Error', true);
          this.set('errorMessage2', 'This is required');
        }

        // update the slider handles and color
        if (this.get('isThresh1Error') === false && this.get('isThresh2Error') === false) {
          $("#slider-range").slider('values', 0 , parseFloat(thresh1));
          $("#slider-range").slider('values', 1 , parseFloat(thresh2));
        }
      }.observes('thresh1', 'thresh2')

    });

    var browserVerion = this.getInternetExplorerVersion();
    App.ModalPopup.show({
      header: Em.I18n.t('dashboard.widgets.popupHeader'),
      classNames: [ 'sixty-percent-width-modal-edit-widget'],
      bodyClass: Ember.View.extend({
        templateName: require('templates/main/dashboard/edit_widget_popup'),
        configPropertyObj: configObj
      }),
      primary: Em.I18n.t('common.apply'),
      onPrimary: function () {
        configObj.observeNewThresholdValue();
        if (!configObj.isThresh1Error && !configObj.isThresh2Error) {
          parent.set('thresh1', parseFloat(configObj.get('thresh1')) );
          parent.set('thresh2', parseFloat(configObj.get('thresh2')) );
          if (!App.get('testMode')) {
            var big_parent = parent.get('parentView');
            big_parent.getUserPref(big_parent.get('persistKey'));
            var oldValue = big_parent.get('currentPrefObject');
            oldValue.threshold[parseInt(parent.id)] = [configObj.get('thresh1'), configObj.get('thresh2')];
            big_parent.postUserPref(big_parent.get('persistKey'),oldValue);
          }
          this.hide();
        }
      },

      didInsertElement: function () {
        var handlers = [configObj.get('thresh1'), configObj.get('thresh2')];
        var colors = [App.healthStatusRed, App.healthStatusOrange, App.healthStatusGreen]; //color red, orange, green

        if (browserVerion == -1 || browserVerion > 9) {
          configObj.set('isIE9', false);
          configObj.set('isGreenOrangeRed', false);
          $("#slider-range").slider({
            range: true,
            min: 0,
            max: max_tmp,
            values: handlers,
            create: function (event, ui) {
              parent.updateColors(handlers, colors);
            },
            slide: function (event, ui) {
              parent.updateColors(ui.values, colors);
              configObj.set('thresh1', ui.values[0] + '');
              configObj.set('thresh2', ui.values[1] + '');
            },
            change: function (event, ui) {
              parent.updateColors(ui.values, colors);
            }
          });
        } else {
          configObj.set('isIE9', true);
          configObj.set('isGreenOrangeRed', false);
        }
      }
    });
  }
});

});

require.register("views/main/dashboard/widgets/text_widget", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var date = require('utils/date');

App.TextDashboardWidgetView = App.DashboardWidgetView.extend({

  templateName: require('templates/main/dashboard/widgets/simple_text'),

  classNameBindings: ['isRed', 'isOrange', 'isGreen', 'isNA'],

  isRed: function () {
    return (this.get('data') <= this.get('thresh1'));
  }.property('data','thresh1'),

  isOrange: function () {
    return (this.get('data') <= this.get('thresh2') && this.get('data') > this.get('thresh1'));
  }.property('data','thresh1','thresh2'),

  isGreen: function () {
    return (this.get('data') > this.get('thresh2'));
  }.property('data','thresh2'),

  isNA: function () {
    return this.get('data') === null;
  }.property('data'),

  hiddenInfo: [],
  thresh1: null,
  thresh2: null,
  maxValue: null,
  updateColors: function(handlers, colors) {
    var colorstops = colors[0] + ", "; // start with the first color
    for (var i = 0; i < handlers.length; i++) {
      colorstops += colors[i] + " " + handlers[i] + "%,";
      colorstops += colors[i+1] + " " + handlers[i] + "%,";
    }
    colorstops += colors[colors.length - 1];
    var cssForChromeAndSafari = '-webkit-linear-gradient(left,' + colorstops + ')'; // chrome & safari
    var slider = $('#slider-range');
    slider.css('background-image', cssForChromeAndSafari);
    var cssForIE = '-ms-linear-gradient(left,' + colorstops + ')'; // IE 10+
    slider.css('background-image', cssForIE);
    //$('#slider-range').css('filter', 'progid:DXImageTransform.Microsoft.gradient( startColorStr= ' + colors[0] + ', endColorStr= ' + colors[2] +',  GradientType=1 )' ); // IE 10-
    var cssForFireFox = '-moz-linear-gradient(left,' + colorstops + ')'; // Firefox
    slider.css('background-image', cssForFireFox);

    slider.find('.ui-widget-header').css({'background-color': '#FF8E00', 'background-image': 'none'}); // change the  original ranger color
  }
});

});

require.register("views/main/dashboard/widgets/uptime_text_widget", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var date = require('utils/date');

App.UptimeTextDashboardWidgetView = App.TextDashboardWidgetView.extend({

  templateName: require('templates/main/dashboard/widgets/uptime'),

  timeUnit: null,

  hiddenInfoClass: "hidden-info-three-line",

  thresh1: 5,
  thresh2: 10,
  maxValue: 'infinity',

  component: null,
  /**
   * Model's field that used to calculate data and content
   * Should be defined in every child
   */
  modelField: null,

  data: null,
  content: null,

  isGreen: function () {
    return this.get('data') != null;
  }.property('data'),

  isOrange: function () {
    return false;
  }.property('data'),

  isRed: function () {
    return false;
  }.property('data'),

  timeConverter: function (timestamp) {
    var m = moment((new Date(timestamp)));
    return [m.format('ddd MMM DD YYYY'), m.format('HH:mm:ss')];
  },

  /**
   * All children should have such code
   * <code>
   * didInsertElement: function() {
   *   this._super();
   *   this.calc();
   * }
   * </code>
   */
  didInsertElement: function () {
    this._super();
    this.addObserver('model.' + this.get('modelField'), this, this.calc);
  },

  calc: function () {
    this.set('data', this.calcData());
    this.set('content', this.calcContent());
  },

  uptimeProcessing: function (uptime) {
    var uptimeString = this.timeConverter(uptime);
    var diff = App.dateTime() - uptime;
    if (diff < 0) {
      diff = 0;
    }
    var formatted = date.timingFormat(diff); //17.67 days
    var timeUnit = null;
    if (formatted != null) {
      switch (formatted.split(" ")[1]) {
        case 'secs':
          timeUnit = 's';
          break;
        case 'hours':
          timeUnit = 'hr';
          break;
        case 'days':
          timeUnit = 'd';
          break;
        case 'mins':
          timeUnit = 'min';
          break;
        default:
          timeUnit = formatted.split(" ")[1];
      }
      this.set('timeUnit', timeUnit);
      this.set('hiddenInfo', []);
      this.get('hiddenInfo').pushObject(formatted);
      this.get('hiddenInfo').pushObject(uptimeString[0]);
      this.get('hiddenInfo').pushObject(uptimeString[1]);
    }
    return formatted;
  },

  calcData: function () {
    var field = this.get('modelField');
    var uptime = this.get('model').get(field);
    if (uptime && uptime > 0) {
      var formatted = this.uptimeProcessing(uptime);
      if (formatted != null) {
        return parseFloat(formatted.split(" ")[0]);
      }
    }
    this.set('hiddenInfo', [this.get('component'), 'Not Running']);
    return null;
  },

  calcContent: function () {
    var data = this.get('data');
    if (data) {
      return data.toFixed(1) + ' ' + this.get('timeUnit');
    }
    else {
      return Em.I18n.t('services.service.summary.notAvailable');
    }
  }
});

});

require.register("views/main/dashboard/widgets/yarn_links", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.YARNLinksView = App.LinkDashboardWidgetView.extend({
  templateName: require('templates/main/dashboard/widgets/yarn_links'),
  title: Em.I18n.t('dashboard.widgets.YARNLinks'),
  id: '23',

  model_type: 'yarn',

  componentName : 'NODEMANAGER'
});

});

require.register("views/main/dashboard/widgets/yarn_memory", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var numberUtils = require('utils/number_utils');

App.YARNMemoryPieChartView = App.PieChartDashboardWidgetView.extend({

  title: Em.I18n.t('dashboard.widgets.YARNMemory'),
  id: '20',

  widgetHtmlId: 'widget-yarn-memory',
  model_type: 'yarn',
  modelFieldUsed: 'allocatedMemory',
  modelFieldMax: 'maxMemory',

  didInsertElement: function() {
    this._super();
    this.calc();
  }
});
});

require.register("views/main/host", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var filters = require('views/common/filter_view');
var sort = require('views/common/sort_view');
var date = require('utils/date');

App.MainHostView = App.TableView.extend(App.TableServerViewMixin, {
  templateName:require('templates/main/host'),

  tableName: 'Hosts',
  updaterBinding: 'App.router.updateController',
  filterConditions: [],

  /**
   * Select/deselect all visible hosts flag
   * @property {bool}
   */
  selectAllHosts: false,

  /**
   * Contains all selected hosts on cluster
   */
  selectedHosts: [],

  /**
   * Request error data
   */
  requestError: null,

  colspan: function () {
    return App.get('supports.stackUpgrade') ? 11 : 10;
  }.property("App.supports.stackUpgrade"),

  /**
   * List of hosts in cluster
   * @type {Array}
   */
  contentBinding: 'controller.content',

  onRequestErrorHandler: function() {
    this.set('requestError', null);
    this.get('controller').get('dataSource').setEach('isRequested', false);
    this.set('filteringComplete', true);
    this.propertyDidChange('pageContent');
  }.observes('requestError'),

  /**
   * flag to toggle displaying selected hosts counter
   */
  showSelectedFilter: function () {
    return this.get('selectedHosts.length') > 0;
  }.property('selectedHosts'),

  /**
   * return filtered number of all content number information displayed on the page footer bar
   * @returns {String}
   */
  filteredContentInfo: function () {
    return this.t('hosts.filters.filteredHostsInfo').format(this.get('filteredCount'), this.get('totalCount'));
  }.property('filteredCount', 'totalCount'),

  /**
   * request latest data filtered by new parameters
   * called when trigger property(<code>refreshTriggers</code>) is changed
   */
  refresh: function () {
    var self = this;
    this.set('filteringComplete', false);
    var updaterMethodName = this.get('updater.tableUpdaterMap')[this.get('tableName')];
    this.get('updater')[updaterMethodName](function () {
      self.set('filteringComplete', true);
      self.propertyDidChange('pageContent');
    }, function() {
      self.set('requestError', arguments);
    });
    return true;
  },
  /**
   * reset filters value by column to which filter belongs
   * @param columns {Array}
   */
  resetFilterByColumns: function (columns) {
    var filterConditions = this.get('filterConditions');
    columns.forEach(function (iColumn) {
      var filterCondition = filterConditions.findProperty('iColumn', iColumn);

      if (filterCondition) {
        filterCondition.value = '';
        this.saveFilterConditions(filterCondition.iColumn, filterCondition.value, filterCondition.type, filterCondition.skipFilter);
      }
    }, this);
  },

  /**
   * Return pagination information displayed on the page
   * @type {String}
   */
  paginationInfo: function () {
    return this.t('tableView.filters.paginationInfo').format(this.get('startIndex'), this.get('endIndex'), this.get('filteredCount'));
  }.property('startIndex', 'endIndex', 'filteredCount'),

  paginationLeftClass: function () {
    if (this.get("startIndex") > 1 && this.get('filteringComplete')) {
      return "paginate_previous";
    }
    return "paginate_disabled_previous";
  }.property("startIndex", 'filteringComplete'),

  previousPage: function () {
    if (this.get('paginationLeftClass') === 'paginate_previous') {
      this._super();
    }
  },

  paginationRightClass: function () {
    if ((this.get("endIndex")) < this.get("filteredCount") && this.get('filteringComplete')) {
      return "paginate_next";
    }
    return "paginate_disabled_next";
  }.property("endIndex", 'filteredCount', 'filteringComplete'),

  nextPage: function () {
    if (this.get('paginationRightClass') === 'paginate_next') {
      this._super();
    }
  },

  /**
   * Select View with list of "rows-per-page" options
   * @type {Ember.View}
   */
  rowsPerPageSelectView: Em.Select.extend({
    content: ['10', '25', '50', '100'],
    attributeBindings: ['disabled'],
    disabled: true,

    disableView: function () {
      Em.run.next(this, function(){
        this.set('disabled', !this.get('parentView.filteringComplete'));
      });
    }.observes('parentView.filteringComplete'),

    change: function () {
      this.get('parentView').saveDisplayLength();
      var self = this;
      if (this.get('parentView.startIndex') === 1 || this.get('parentView.startIndex') === 0) {
        Ember.run.next(function () {
          self.get('parentView').updatePagination();
        });
      } else {
        Ember.run.next(function () {
          self.set('parentView.startIndex', 1);
        });
      }
    }
  }),

  saveStartIndex: function () {
    this.set('controller.startIndex', this.get('startIndex'));
  }.observes('startIndex'),

  clearFiltersObs: function() {
    var self = this;
    Em.run.next(function() {
      if (self.get('controller.clearFilters')) {
        self.clearFilters();
      }
    });
  },
  /**
   * Restore filter properties in view
   */
  willInsertElement: function () {
    if (!this.get('controller.showFilterConditionsFirstLoad')) {
      this.clearFilterCondition();
    }
    this._super();
    this.set('startIndex', this.get('controller.startIndex'));
    this.addObserver('pageContent.@each.selected', this, this.selectedHostsObserver);
  },

  /**
   * stub for filter function in TableView
   */
  filter: function () {
    //Since filtering moved to server side, function is empty
  },

  didInsertElement: function() {
    this.addObserver('controller.clearFilters', this, this.clearFiltersObs);
    this.clearFiltersObs();
    this.addObserver('selectAllHosts', this, this.toggleAllHosts);
    this.addObserver('filteringComplete', this, this.overlayObserver);
    this.addObserver('startIndex', this, 'updateHostsPagination');
    this.addObserver('displayLength', this, 'updatePagination');
    this.addObserver('filteredCount', this, this.updatePaging);
    this.overlayObserver();
  },

  updateHostsPagination: function () {
    this.clearExpandedSections();
    this.updatePagination();
  },

  willDestroyElement: function () {
    this.clearExpandedSections();
  },

  clearExpandedSections: function () {
    this.get('controller.expandedComponentsSections').clear();
    this.get('controller.expandedVersionsSections').clear();
  },

  onInitialLoad: function () {
    if (this.get('tableFilteringComplete')) {
      this.refresh();
    }
  }.observes('tableFilteringComplete'),

  /**
   * Set <code>selected</code> property for each App.Host
   */
  toggleAllHosts: function() {
    this.get('pageContent').setEach('selected', this.get('selectAllHosts'));
  },

  /**
   * Trigger updating <code>selectedHostsCount</code> only 1 time
   */
  selectedHostsObserver: function() {
    Ember.run.once(this, 'updateCheckedFlags');
  },

  /**
   * Update <code>selectAllHosts</code> value
   */
  updateCheckedFlags: function() {
    this.removeObserver('selectAllHosts', this, this.toggleAllHosts);
    if (this.get('pageContent').length) {
      this.set('selectAllHosts', this.get('pageContent').everyProperty('selected', true));
    }
    else {
      this.set('selectAllHosts', false);
    }
    this.combineSelectedFilter();
    //10 is an index of selected column
    var controllerName = this.get('controller.name');
    App.db.setSelectedHosts(controllerName, this.get('selectedHosts'));

    this.addObserver('selectAllHosts', this, this.toggleAllHosts);
  },
  /**
   * combine selected hosts on page with selected hosts which are filtered out but added to cluster
   */
  combineSelectedFilter: function () {
    var controllerName = this.get('controller.name');
    var previouslySelectedHosts = App.db.getSelectedHosts(controllerName);
    var selectedHosts = [];
    var hostsOnPage = this.get('pageContent').mapProperty('hostName');
    selectedHosts = this.get('pageContent').filterProperty('selected').mapProperty('hostName');

    previouslySelectedHosts.forEach(function (hostName) {
      if (!hostsOnPage.contains(hostName)) {
        selectedHosts.push(hostName);
      }
    }, this);
    this.set('selectedHosts', selectedHosts);
  },

  /**
   * filter selected hosts
   */
  filterSelected: function() {
    //10 is an index of selected column
    this.updateFilter(10, this.get('selectedHosts'), 'multiple');
  },

  /**
   * Show spinner when filter/sorting request is in processing
   * @method overlayObserver
   */
  overlayObserver: function() {
    var $tbody = this.$('tbody'),
      $overlay = this.$('.table-overlay'),
      $spinner = $($overlay).find('.spinner');
    if (!this.get('filteringComplete')) {
      if (!$tbody) return;
      var tbodyPos =  $tbody.position();
      if (!tbodyPos) return;
      $spinner.css('display', 'block');
      $overlay.css({
        top: tbodyPos.top + 1,
        left: tbodyPos.left + 1,
        width: $tbody.width() - 1,
        height: $tbody.height() - 1
      });
    }
  },

  /**
   * Clear selectedFilter
   * Set <code>selected</code> to false for each host
   */
  clearSelection: function() {
    this.get('pageContent').setEach('selected', false);
    this.set('selectAllHosts', false);
    App.db.setSelectedHosts(this.get('controller.name'), []);
    this.get('selectedHosts').clear();
    this.filterSelected();
  },

  /**
   * Returs all hostNames if amount is less than {minShown} or
   * first elements of array (number of elements - {minShown}) converted to string
   * @param {Array} hostNames - array of all listed hostNames
   * @param {String} divider - string to separate hostNames
   * @param {Number} minShown - min amount of hostName to be shown
   * @returns {String} hostNames
   * @method showHostNames
   */
  showHostNames: function(hostNames, divider, minShown) {
    if (hostNames.length > minShown) {
      return hostNames.slice(0, minShown).join(divider) + divider + Em.I18n.t("installer.step8.other").format(hostNames.length - minShown);
    } else {
      return hostNames.join(divider);
    }
  },

  /**
   * Confirmation Popup for bulk Operations
   */
  bulkOperationConfirm: function(operationData, selection) {
    var hostsNames = [],
      queryParams = [];
    switch(selection) {
      case 's':
        hostsNames = this.get('selectedHosts');
        if(hostsNames.length > 0){
          queryParams.push({
            key: 'Hosts/host_name',
            value: hostsNames,
            type: 'MULTIPLE'
          });
        }
        break;
      case 'f':
        queryParams = this.get('controller').getQueryParameters(true).filter(function (obj) {
          return !(obj.key == 'page_size' || obj.key == 'from');
        });
        break;
    }

    var loadingPopup = App.ModalPopup.show({
      header: Em.I18n.t('jobs.loadingTasks'),
      primary: false,
      secondary: false,
      bodyClass: Ember.View.extend({
        template: Ember.Handlebars.compile('<div class="spinner"></div>')
      })
    });
    var parameters = App.router.get('updateController').computeParameters(queryParams);
    if (!parameters.length) parameters = '&';
    App.ajax.send({
      name: 'hosts.bulk.operations',
      sender: this,
      data: {
        parameters: parameters.substring(0, parameters.length - 1),
        operationData: operationData,
        loadingPopup: loadingPopup
      },
      success: 'getHostsForBulkOperationSuccessCallback'
    });
  },

  convertHostsObjects: function(hosts) {
    var newHostArr = [];
    hosts.forEach(function (host) {
      newHostArr.push({
        index:host.index,
        id:host.id,
        clusterId: host.cluster_id,
        passiveState: host.passive_state,
        isRequested: host.is_requested,
        hostName: host.host_name,
        hostComponents: host.host_components
      })
    });
    return newHostArr;
  },

  getHostsForBulkOperationSuccessCallback: function(json, opt, param) {
    var self = this,
    operationData = param.operationData,
    hosts = this.convertHostsObjects(App.hostsMapper.map(json, true));
    // no hosts - no actions
    if (!hosts.length) {
      console.log('No bulk operation if no hosts selected.');
      return;
    }
    var hostNames = hosts.mapProperty('hostName');
    var hostsToSkip = [];
    if (operationData.action == "DECOMMISSION") {
      var hostComponentStatusMap = {}; // "DATANODE_c6401.ambari.apache.org" => "STARTED"
      var hostComponentIdMap = {}; // "DATANODE_c6401.ambari.apache.org" => "DATANODE"
      if (json.items) {
        json.items.forEach(function(host) {
          if (host.host_components) {
            host.host_components.forEach(function(component) {
              hostComponentStatusMap[component.id] = component.HostRoles.state;
              hostComponentIdMap[component.id] = component.HostRoles.component_name;
            });
          }
        });
      }
      hostsToSkip = hosts.filter(function(host) {
        var invalidStateComponents = host.hostComponents.filter(function(component) {
          return hostComponentIdMap[component] == operationData.realComponentName && hostComponentStatusMap[component] == 'INSTALLED';
        });
        return invalidStateComponents.length > 0;
      });
    }
    var hostNamesSkipped = hostsToSkip.mapProperty('hostName');
    var message;
    if (operationData.componentNameFormatted) {
      message = Em.I18n.t('hosts.bulkOperation.confirmation.hostComponents').format(operationData.message, operationData.componentNameFormatted, hostNames.length);
    }
    else {
      message = Em.I18n.t('hosts.bulkOperation.confirmation.hosts').format(operationData.message, hostNames.length);
    }
    param.loadingPopup.hide();
    App.ModalPopup.show({
      header: Em.I18n.t('hosts.bulkOperation.confirmation.header'),
      hostNames: hostNames.join("\n"),
      visibleHosts: self.showHostNames(hostNames, "\n", 3),
      hostNamesSkippedVisible: self.showHostNames(hostNamesSkipped, "\n", 3),
      hostNamesSkipped: function() {
        if (hostNamesSkipped.length) {
          return hostNamesSkipped.join("\n");
        }
        return false;
      }.property(),
      expanded: false,
      didInsertElement: function() {
        this.set('expanded', hostNames.length <= 3);
      },
      onPrimary: function() {
        self.get('controller').bulkOperation(operationData, hosts);
        this._super();
      },
      bodyClass: Em.View.extend({
        templateName: require('templates/main/host/bulk_operation_confirm_popup'),
        message: message,
        warningInfo: Em.I18n.t('hosts.bulkOperation.warningInfo.body'),
        textareaVisible: false,
        textTrigger: function() {
          this.set('textareaVisible', !this.get('textareaVisible'));
        },
        showAll: function() {
          this.set('parentView.visibleHosts', this.get('parentView.hostNames'));
          this.set('parentView.hostNamesSkippedVisible', this.get('parentView.hostNamesSkipped'));
          this.set('parentView.expanded', true);
        },
        putHostNamesToTextarea: function() {
          var hostNames = this.get('parentView.hostNames');
          if (this.get('textareaVisible')) {
            var wrapper = $(".task-detail-log-maintext");
            $('.task-detail-log-clipboard').html(hostNames).width(wrapper.width()).height(250);
            Em.run.next(function() {
              $('.task-detail-log-clipboard').select();
            });
          }
        }.observes('textareaVisible')
      })
    });
  },

  sortView: sort.serverWrapperView,
  nameSort: sort.fieldView.extend({
    column: 1,
    name:'hostName',
    displayName: Em.I18n.t('common.name')
  }),
  ipSort: sort.fieldView.extend({
    column: 2,
    name:'ip',
    displayName: Em.I18n.t('common.ipAddress'),
    type: 'ip'
  }),
  cpuSort: sort.fieldView.extend({
    column: 3,
    name:'cpu',
    displayName: Em.I18n.t('common.cores'),
    type: 'number'
  }),
  memorySort: sort.fieldView.extend({
    column: 4,
    name:'memoryFormatted',
    displayName: Em.I18n.t('common.ram'),
    type: 'number'
  }),
  diskUsageSort: sort.fieldView.extend({
    name:'diskUsage',
    displayName: Em.I18n.t('common.diskUsage')
  }),
  loadAvgSort: sort.fieldView.extend({
    column: 5,
    name:'loadAvg',
    displayName: Em.I18n.t('common.loadAvg'),
    type: 'number'
  }),

  HostView:Em.View.extend({
    content:null,
    tagName: 'tr',
    didInsertElement: function(){
      var hostName = this.get('content.hostName');
      App.tooltip(this.$("[rel='HealthTooltip'], [rel='UsageTooltip'], [rel='ComponentsTooltip']"));
      this.set('isComponentsCollapsed', !this.get('controller.expandedComponentsSections').contains(hostName));
      this.set('isVersionsCollapsed', !this.get('controller.expandedVersionsSections').contains(hostName));
    },

    toggleList: function (flagName, arrayName) {
      var arrayPropertyName = 'controller.' + arrayName;
      var hostNameArray = this.get(arrayPropertyName);
      var hostName = this.get('content.hostName');
      this.toggleProperty(flagName);
      if (this.get(flagName)) {
        this.set(arrayPropertyName, hostNameArray.without(hostName));
      } else {
        hostNameArray.push(hostName);
      }
    },

    toggleComponents: function () {
      this.toggleList('isComponentsCollapsed', 'expandedComponentsSections');
    },

    toggleVersions: function () {
      this.toggleList('isVersionsCollapsed', 'expandedVersionsSections');
    },

    /**
     * Tooltip message for "Restart Required" icon
     * @returns {String}
     */
    restartRequiredComponentsMessage: function() {
      var restartRequiredComponents = this.get('content.hostComponents').filterProperty('staleConfigs', true);
      var count = restartRequiredComponents.length;
      if (count <= 5) {
        var word = (count == 1) ? Em.I18n.t('common.component') : Em.I18n.t('common.components');
        return Em.I18n.t('hosts.table.restartComponents.withNames').format(restartRequiredComponents.getEach('displayName').join(', ')) + ' ' + word.toLowerCase();
      }
      return Em.I18n.t('hosts.table.restartComponents.withoutNames').format(count);
    }.property('content.hostComponents.@each.staleConfigs'),

    /**
     * Tooltip message for "Maintenance" icon
     * @returns {String}
     */
    componentsInPassiveStateMessage: function() {
      var componentsInPassiveState = this.get('content.hostComponents').filter(function(component) {
        return component.get('passiveState') !== 'OFF';
      });
      var count = componentsInPassiveState.length;
      if (count <= 5) {
        return Em.I18n.t('hosts.table.componentsInPassiveState.withNames').format(componentsInPassiveState.getEach('displayName').join(', '));
      }
      return Em.I18n.t('hosts.table.componentsInPassiveState.withoutNames').format(count);
    }.property('content.hostComponents.@each.passiveState'),

    /**
     * String with list of host components <code>displayName</code>
     * @returns {String}
     */
    labels: function() {
      return this.get('content.hostComponents').getEach('displayName').join("<br />");
    }.property('content.hostComponents.length'),

    /**
     * true if host has only one repoversion
     * in this case expander in version column is hidden
     * @returns {Boolean}
     */
    hasSingleVersion: function() {
      return this.get('content.stackVersions').filterProperty('isVisible', true).length < 2;
    }.property('content.stackVersions.length'),

    /**
     * this version is always shown others hidden unless expander is open
     * host may have no stack versions
     * @returns {String}
     */
    currentVersion: function() {
      var currentRepoVersion = this.get('content.stackVersions').findProperty('isCurrent') || this.get('content.stackVersions').objectAt(0);
      return currentRepoVersion ? currentRepoVersion.get('displayName') + " (" + currentRepoVersion.get('displayStatus') + ")" : "";
    }.property('content.stackVersions'),

    /**
     * String with list of host components <code>displayName</code>
     * @returns {String}
     */
    versionLabels: function () {
      return this.get('content.stackVersions').filter(function(sv) {
        return sv.get('isVisible') === true && sv.get('isCurrent') === false;
      }).map(function (version) {
        return version.get('displayName');
      }).join("<br />");
    }.property('content.stackVersions.length'),

    /**
     * CSS value for disk usage bar
     * @returns {String}
     */
    usageStyle:function () {
      return "width:" + this.get('content.diskUsage') + "%";
    }.property('content.diskUsage')

  }),

  /**
   * Update <code>hostsCount</code> in every category
   */
  updateHostsCount: function() {
    var hostsCountMap = this.get('controller.hostsCountMap');

    this.get('categories').forEach(function(category) {
      var hostsCount = (category.get('healthStatus').trim() === "") ? hostsCountMap['TOTAL'] : hostsCountMap[category.get('healthStatus')];

      if (!Em.isNone(hostsCount)) {
        category.set('hostsCount', hostsCount);
        category.set('hasHosts', (hostsCount > 0));
      }
    }, this);
  }.observes('controller.hostsCountMap'),

  /**
   * Category view for all hosts
   * @type {Object}
   */
  //@TODO maybe should be separated to two types (basing on <code>isHealthStatus</code>)
  categoryObject: Em.Object.extend({

    /**
     * Text used with <code>hostsCount</code> in category label
     * @type {String}
     */
    value: null,
    /**
     * Is category based on host health status
     * @type {bool}
     */
    isHealthStatus: true,
    /**
     * host health status (used if <code>isHealthStatus</code> is true)
     * @type {String}
     */
    healthStatusValue: '',
    /**
     * Should category be displayed on the top of the hosts table
     * @type {bool}
     */
    isVisible: true,
    /**
     * Is category selected now
     * @type {bool}
     */
    isActive: false,
    /**
     * String with path that category should observe
     * @type {String}
     */
    observes: null,
    /**
     * CSS-class for span in the category-link (used if <code>isHealthStatus</code> is false)
     * @type {String}
     */
    class: null,
    /**
     * Associated column number
     * @type {Number}
     */
    column: null,
    /**
     * Type of filter value (string, number, boolean)
     * @type {String}
     */
    type: null,
    /**
     * @type {String|Number|bool}
     */
    filterValue: null,
    /**
     * <code>App.Host</code> property that should be used to calculate <code>hostsCount</code> (used if <code>isHealthStatus</code> is false)
     * @type {String}
     */
    hostProperty: null,

    /**
     * Number of host in current category
     * @type {Number}
     */
    hostsCount: 0,

    /**
     * Determine if category has hosts
     * @type {bool}
     */
    hasHosts: false,

    /**
     * Add "active" class for category span-wrapper if current category is selected
     * @type {String}
     */
    itemClass: function() {
      return this.get('isActive') ? 'active' : '';
    }.property('isActive'),

    /**
     * Text shown on the right of category icon
     * @type {String}
     */
    label: function () {
      return "%@ (%@)".fmt(this.get('value'), this.get('hostsCount'));
    }.property('hostsCount')
  }),

  /**
   * List of categories used to filter hosts
   * @type {Array}
   */
  categories: function () {
    var self = this;
    var category_mocks = require('data/host/categories');

    return category_mocks.map(function(category_mock) {
      return self.categoryObject.create(category_mock);
    });
  }.property(),

  /**
   * Category for <code>selected</code> property of each App.Host
   */
  selectedCategory: function() {
    return this.get('categories').findProperty('selected', true);
  }.property('categories.@each.selected'),

  statusFilter: Em.View.extend({
    column: 0,
    categories: [],
    value: null,
    class: "",
    comboBoxLabel: function(){
      var selected = this.get('categories').findProperty('isActive');
      if (!this.get('value') || !selected) {
        return "%@ (%@)".fmt(Em.I18n.t('common.all'), this.get('parentView.totalCount'));
      } else {
        return "%@ (%@)".fmt(selected.get('value'), selected.get('hostsCount'))
      }
    }.property('value', 'parentView.totalCount'),
    /**
     * switch active category label
     */
    onCategoryChange: function () {
      this.get('categories').setEach('isActive', false);
      var selected = this.get('categories').findProperty('healthStatus', this.get('value'));
      selected.set('isActive', true);
      this.set('class', selected.get('class') + ' ' + selected.get('healthClass'));
    }.observes('value'),

    showClearFilter: function () {
      var mockEvent = {
        context: this.get('categories').findProperty('healthStatus', this.get('value'))
      };
      this.selectCategory(mockEvent);
    },
    /**
     * Trigger on Category click
     * @param {Object} event
     */
    selectCategory: function(event){
      var category = event.context;

      this.set('value', category.get('healthStatus'));
      this.get('parentView').resetFilterByColumns([0, 7, 8, 9]);
      if (category.get('isHealthStatus')) {
        this.get('parentView').updateFilter(0, category.get('healthStatus'), 'string');
      } else {
        this.get('parentView').updateFilter(category.get('column'), category.get('filterValue'), category.get('type'));
      }
    },

    /**
     * set value
     * @param {string} value
     */
    setValue: function (value) {
      this.set('value', value);
    },

    clearFilter: function() {
      this.get('categories').setEach('isActive', false);
      this.set('value', '');
      this.set('class', '');
      this.showClearFilter();
    }
  }),

  /**
   * Filter view for name column
   * Based on <code>filters</code> library
   */
  nameFilterView: filters.createTextView({
    column: 1,
    fieldType: 'filter-input-width',
    onChangeValue: function(){
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'string');
    }
  }),

  /**
   * Filter view for ip column
   * Based on <code>filters</code> library
   */
  ipFilterView: filters.createTextView({
    column: 2,
    fieldType: 'filter-input-width',
    onChangeValue: function(){
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'string');
    }
  }),

  /**
   * Filter view for Cpu column
   * Based on <code>filters</code> library
   */
  cpuFilterView: filters.createTextView({
    fieldType: 'filter-input-width',
    fieldId: 'cpu_filter',
    column: 3,
    onChangeValue: function(){
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'number');
    }
  }),

  /**
   * Filter view for Ram column
   * Based on <code>filters</code> library
   */
  ramFilterView: filters.createTextView({
    fieldType: 'filter-input-width',
    fieldId: 'ram_filter',
    column: 4,
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'ambari-bandwidth');
    }
  }),

  /**
   * Filter view for LoadAverage column
   * Based on <code>filters</code> library
   */
  loadAvgFilterView: filters.createTextView({
    fieldType: 'filter-input-width',
    fieldId: 'load_avg_filter',
    column: 5,
    onChangeValue: function(){
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'number');
    }
  }),

  /**
   * Filter view for HostComponents column
   * Based on <code>filters</code> library
   */
  componentsFilterView: filters.createComponentView({

    column: 6,

    /**
     * Inner FilterView. Used just to render component. Value bind to <code>mainview.value</code> property
     * Base methods was implemented in <code>filters.componentFieldView</code>
     */
    filterView: filters.componentFieldView.extend({
      templateName: require('templates/main/host/component_filter'),


      /**
       * Components which will be shown in component filter
       * @returns {Array}
       */
      componentsForFilter: function () {
        var installedComponents = App.StackServiceComponent.find().toArray();
        installedComponents.setEach('checkedForHostFilter', false);
        return installedComponents;
      }.property('App.router.clusterController.isLoaded'),

      /**
       * Master components
       * @returns {Array}
       */
      masterComponents: function () {
        return this.get('componentsForFilter').filterProperty('isMaster', true);
      }.property('componentsForFilter'),

      /**
       * Slave components
       * @returns {Array}
       */
      slaveComponents: function () {
        return this.get('componentsForFilter').filterProperty('isSlave', true);
      }.property('componentsForFilter'),

      /**
       * Client components
       * @returns {Array}
       */
      clientComponents: function () {
        return this.get('componentsForFilter').filterProperty('isClient', true);
      }.property('componentsForFilter'),

      /**
       * Checkbox for quick selecting/deselecting of master components
       */
      masterComponentsChecked:false,
      toggleMasterComponents:function () {
        this.get('masterComponents').setEach('checkedForHostFilter', this.get('masterComponentsChecked'));
      }.observes('masterComponentsChecked'),

      /**
       * Checkbox for quick selecting/deselecting of slave components
       */
      slaveComponentsChecked:false,
      toggleSlaveComponents:function () {
        this.get('slaveComponents').setEach('checkedForHostFilter', this.get('slaveComponentsChecked'));
      }.observes('slaveComponentsChecked'),

      /**
       * Checkbox for quick selecting/deselecting of client components
       */
      clientComponentsChecked: false,
      toggleClientComponents: function() {
        this.get('clientComponents').setEach('checkedForHostFilter', this.get('clientComponentsChecked'));
      }.observes('clientComponentsChecked'),

      /**
       * Clear filter.
       * Called by parent view, when user clicks on <code>x</code> button(clear button)
       */
      clearFilter:function() {
        this.set('masterComponentsChecked', false);
        this.set('slaveComponentsChecked', false);
        this.set('clientComponentsChecked', false);

        this.get('masterComponents').setEach('checkedForHostFilter', false);
        this.get('slaveComponents').setEach('checkedForHostFilter', false);
        this.get('clientComponents').setEach('checkedForHostFilter', false);

        this._super();
      },

      /**
       * Onclick handler for <code>Apply filter</code> button
       */
      applyFilter:function() {
        this._super();
        var self = this;
        var chosenComponents = [];

        this.get('masterComponents').filterProperty('checkedForHostFilter', true).forEach(function(item){
          chosenComponents.push(item.get('id'));
        });
        this.get('slaveComponents').filterProperty('checkedForHostFilter', true).forEach(function(item){
          chosenComponents.push(item.get('id'));
        });
        this.get('clientComponents').filterProperty('checkedForHostFilter', true).forEach(function(item){
          chosenComponents.push(item.get('id'));
        });
        Em.run.next(function() {
          self.set('value', chosenComponents);
        });
      },

      /**
       * Verify that checked checkboxes are equal to value stored in hidden field (components ids list)
       */
      checkComponents: function() {
        var components = this.get('value');
        var self = this;
        if (components) {
          components.forEach(function(componentId) {
            if(!self.tryCheckComponent(self, 'masterComponents', componentId)) {
              if(!self.tryCheckComponent(self, 'slaveComponents', componentId)) {
                self.tryCheckComponent(self, 'clientComponents', componentId);
              }
            }
          });
        }
      }.observes('value'),

      tryCheckComponent: function(self, category, componentId) {
        var c = self.get(category).findProperty('id', componentId);
        if (c) {
          if (!c.get('checkedForHostFilter')) {
            c.set('checkedForHostFilter', true);
            return true;
          }
        }
        return false;
      }

    }),
    onChangeValue: function(){
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'multiple');
    }
  }),

  versionsFilterView: filters.wrapperView.extend({
    column: 11,
    filterView: filters.componentFieldView.extend({
      templateName: require('templates/main/host/version_filter'),
      selectedVersion: null,
      selectedStatus: null,
      value: [],

      versionSelectView: filters.createSelectView({
        classNames: ['notActive'],
        fieldType: 'filter-input-width',
        filterPropertyName: 'repository_versions/RepositoryVersions/display_name',
        content: function () {
          return  [
            {
              value: '',
              label: Em.I18n.t('common.all')
            }
          ].concat(this.get('controller.allHostStackVersions').filterProperty('isVisible', true).mapProperty('displayName').uniq().map(function (version) {
            return {
              value: version,
              label: version
            }
          }));
        }.property('App.router.clusterController.isLoaded', 'controller.allHostStackVersions.length'),
        onChangeValue: function () {
          this.set('parentView.selectedVersion', this.get('value'));
        }
      }),
      statusSelectView: filters.createSelectView({
        classNames: ['notActive'],
        fieldType: 'filter-input-width',
        filterPropertyName: 'HostStackVersions/state',
        content: function () {
          return [
            {
              value: '',
              label: Em.I18n.t('common.all')
            }
          ].concat(App.HostStackVersion.statusDefinition.map(function (status) {
            return {
              value: status,
              label: App.HostStackVersion.formatStatus(status)
            }
          }));
        }.property('App.router.clusterController.isLoaded'),
        onChangeValue: function () {
          this.set('parentView.selectedStatus', this.get('value'));
        }
      }),
      /**
       * Onclick handler for <code>Apply filter</code> button
       */
      applyFilter: function () {
        this._super();
        var self = this;
        var filterProperties = [];
        if (this.get('selectedVersion')) {
          filterProperties.push({
            property: 'repository_versions/RepositoryVersions/display_name',
            value: this.get('selectedVersion')
          });
        }
        if (this.get('selectedStatus')) {
          filterProperties.push({
            property: 'HostStackVersions/state',
            value: this.get('selectedStatus')
          });
        }
        self.set('value', filterProperties);
      },
      /**
       * Clear filter to initial state
       */
      clearFilter: function () {
        this.set('value', []);
        this.get('childViews').forEach(function (view) {
          if (typeof view.clearFilter === "function") view.clearFilter();
        });
      }
    }),
    setValue: function (value) {
      var versionSelectView = this.get('childViews')[0];

      //restore selected options in Select views
      versionSelectView.get('childViews').forEach(function (view) {
        var filter = value.findProperty('property', view.get('filterPropertyName'));
        if (filter && view.get('content').findProperty('value', filter.value)) {
          view.set('selected', view.get('content').findProperty('value', filter.value));
        }
      }, this);
      this._super(value);
    },
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'sub-resource');
    },
    clearFilter: function () {
      this._super();
      this.get('childViews').forEach(function (view) {
        if (typeof view.clearFilter === "function") view.clearFilter();
      });
    }
  }),

  /**
   * associations between host property and column index
   * @type {Array}
   */
  colPropAssoc: function () {
    return this.get('controller.colPropAssoc');
  }.property('controller.colPropAssoc')
});

});

require.register("views/main/host/addHost/step4_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.AddHostStep4View = Em.View.extend({
  templateName: require('templates/main/host/addHost/step4')
});

});

require.register("views/main/host/add_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.AddHostView = Em.View.extend(App.WizardMenuMixin, {

  templateName: require('templates/main/host/add'),

  isLoaded: false,

  willInsertElement: function () {
    if (this.get('controller').getDBProperty('hosts')) {
      this.set('isLoaded', true);
    } else {
      this.loadHosts();
    }
  },

  loadHosts: function () {
    App.ajax.send({
      name: 'hosts.confirmed',
      sender: this,
      data: {},
      success: 'loadHostsSuccessCallback',
      error: 'loadHostsErrorCallback'
    });
  },

  loadHostsSuccessCallback: function (response) {
    var installedHosts = {};

    response.items.forEach(function (item) {
      installedHosts[item.Hosts.host_name] = {
        name: item.Hosts.host_name,
        cpu: item.Hosts.cpu_count,
        memory: item.Hosts.total_mem,
        disk_info: item.Hosts.disk_info,
        bootStatus: "REGISTERED",
        isInstalled: true,
        hostComponents: item.host_components
      };
    });
    this.get('controller').setDBProperty('hosts', installedHosts);
    this.set('controller.content.hosts', installedHosts);
    this.set('isLoaded', true);
  },

  loadHostsErrorCallback: function(){
    this.set('isLoaded', true);
  }

});

});

require.register("views/main/host/configs", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var date = require('utils/date');

App.MainHostConfigsView = Em.View.extend({
  templateName: require('templates/main/host/configs'),

  content: function(){
    return App.router.get('mainHostDetailsController.content');
  }.property('App.router.mainHostDetailsController.content'),

  isConfigAvailable: function () {
    return !!this.get('content.hostComponents.length');
  }.property('content.hostComponents.length')

});

});

require.register("views/main/host/configs_service", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.MainHostServiceConfigsView = Em.View.extend({
  templateName: require('templates/main/host/configs_service'),
  didInsertElement: function () {
    var controller = this.get('controller');
    controller.loadStep();
  },
  isConfigsEditable: false,
  content: function () {
    return App.router.get('mainHostDetailsController.content');
  }.property('App.router.mainHostDetailsController.content')
});

});

require.register("views/main/host/configs_service_menu", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var misc = require('utils/misc');

App.MainHostServiceMenuView = Em.CollectionView.extend({
  content: function () {
    var host = this.get('host');
    var hostComponents = host.get('hostComponents');
    var services = Em.A([]);
    if (hostComponents) {
      hostComponents.forEach(function (hc) {
        var service = hc.get('service');
        if (service) {
          var serviceName = service.get('serviceName');
          if (!App.get('services.noConfigTypes').contains(serviceName)) {
            if (!services.findProperty('serviceName', serviceName)) {
              services.push(service);
            }
          }
        } else {
          console.warn("service not found for " + hc.get('componentName'));
        }
      });
    }
    var stackServices = App.StackService.find().mapProperty('serviceName');
    return misc.sortByOrder(stackServices, services);
  }.property('host'),

  host: function () {
    return App.router.get('mainHostDetailsController.content');
  }.property('App.router.mainHostDetailsController.content'),

  selectedService: null,

  showHostService: function (event) {
    var service = event.contexts[0];
    if (!Em.isNone(service)) {
      this.set('selectedService', service);
      var context = service;
      context.host = this.get('host');
      this.get('controller').connectOutlet('service_config_outlet', 'mainHostServiceConfigs', context);
    }
  },

  didInsertElement: function () {
    var event = {
      contexts: [this.get('content').objectAt(0)]
    };
    this.showHostService(event);
  },

  activeServiceId: null,

  tagName: 'ul',
  classNames: ["nav", "nav-list", "nav-services"],

  itemViewClass: Em.View.extend({
    classNameBindings: ["active", "clients"],
    active: function () {
      return this.get('content.serviceName') == this.get('parentView.selectedService.serviceName') ? 'active' : '';
    }.property('parentView.selectedService.serviceName'),
    template: Ember.Handlebars.compile('<a href="#" {{action showHostService view.content target="view.parentView"}} >{{view.content.displayName}}</a>')
  })
});
});

require.register("views/main/host/details", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var date = require('utils/date');

App.MainHostDetailsView = Em.View.extend({
  templateName: require('templates/main/host/details'),

  /**
   * flag identify whether current host exist and loaded to model
   */
  isLoaded: false,

  content: function(){
    return App.router.get('mainHostDetailsController.content');
  }.property('App.router.mainHostDetailsController.content'),

  clients: function () {
    return this.get('content.hostComponents').filterProperty('isClient');
  }.property('content.hostComponents.@each'),

  clientsWithConfigs: function() {
    return this.get('content.hostComponents').filterProperty('isClient').filter(function(client) {
      return !App.get('services.noConfigTypes').contains(client.get('service.serviceName'));
    });
  }.property('content.hostComponents.@each'),

  isActive: function() {
    return this.get('controller.content.passiveState') === "OFF";
  }.property('controller.content.passiveState'),

  maintenance: function(){
    var onOff = this.get('isActive') ? "On" : "Off";
    return [
      {action: 'startAllComponents', liClass: (this.get('controller.content.isNotHeartBeating')?'disabled':'enabled'), cssClass: 'icon-play', 'label': this.t('hosts.host.details.startAllComponents')},
      {action: 'stopAllComponents', liClass: (this.get('controller.content.isNotHeartBeating')?'disabled':'enabled'), cssClass: 'icon-stop', 'label': this.t('hosts.host.details.stopAllComponents')},
      {action: 'restartAllComponents', liClass: (this.get('controller.content.isNotHeartBeating')?'disabled':'enabled'), cssClass: 'icon-repeat', 'label': this.t('hosts.host.details.restartAllComponents')},
      {action: 'onOffPassiveModeForHost', liClass:'', cssClass: 'icon-medkit', active:this.get('isActive'), 'label': this.t('passiveState.turn' + onOff)},
      {action: 'deleteHost', liClass:'', cssClass: 'icon-remove', 'label': this.t('hosts.host.details.deleteHost')}];
  }.property('controller.content','isActive', 'controller.content.isNotHeartBeating'),
  didInsertElement: function() {
    var self = this;

    this.set('isLoaded', false);
    App.router.get('updateController').updateHost(function () {
      self.set('isLoaded', true);
      if (!self.get('content.isLoaded')) {
        //if host is not existed then route to list of hosts
        App.router.transitionTo('main.hosts.index');
      }
    });
    App.tooltip($("[rel='HealthTooltip']"));
  }
});

});

require.register("views/main/host/details/host_component_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var uiEffects = require('utils/ui_effects');

App.HostComponentView = Em.View.extend({

  templateName: require('templates/main/host/details/host_component'),

  /**
   * @type {App.HostComponent}
   */
  content: null,

  /**
   * @type {App.HostComponent}
   */
  hostComponent: function () {
    var hostComponent = null;
    var serviceComponent = this.get('content');
    var host = App.router.get('mainHostDetailsController.content');
    if (host) {
      hostComponent = host.get('hostComponents').findProperty('componentName', serviceComponent.get('componentName'));
    }
    return hostComponent;
  }.property('content', 'App.router.mainHostDetailsController.content'),
  /**
   * @type {String}
   */
  workStatus: function () {
    var workStatus = this.get('content.workStatus');
    var hostComponent = this.get('hostComponent');
    if (hostComponent) {
      workStatus = hostComponent.get('workStatus');
    }
    return workStatus;
  }.property('content.workStatus', 'hostComponent.workStatus'),

  /**
   * Return host component text status
   * @type {String}
   */
  componentTextStatus: function () {
    var componentTextStatus = this.get('content.componentTextStatus');
    var hostComponent = this.get('hostComponent');
    if (hostComponent) {
      componentTextStatus = hostComponent.get('componentTextStatus');
    }
    return componentTextStatus;
  }.property('content.passiveState','workStatus'),

  /**
   * CSS-class for host component status
   * @type {String}
   */
  statusClass: function () {
    //Class when install failed
    if (this.get('workStatus') === App.HostComponentStatus.install_failed) {
      return 'health-status-color-red icon-cog';
    }

    //Class when installing
    if (this.get('workStatus') === App.HostComponentStatus.installing) {
      return 'health-status-color-blue icon-cog';
    }

    //For all other cases
    return 'health-status-' + App.HostComponentStatus.getKeyName(this.get('workStatus'));

  }.property('workStatus'),

  /**
   * CSS-icon-class for host component status
   * @type {String}
   */
  statusIconClass: function () {
    return Em.getWithDefault({
      'health-status-started': App.healthIconClassGreen,
      'health-status-starting': App.healthIconClassGreen,
      'health-status-installed': App.healthIconClassRed,
      'health-status-stopping': App.healthIconClassRed,
      'health-status-unknown': App.healthIconClassYellow,
      'health-status-DEAD-ORANGE': App.healthIconClassOrange
    }, this.get('statusClass'), '');
  }.property('statusClass'),

  /**
   * CSS-class for disabling drop-down menu with list of host component actions
   * Disabled if host's <code>healthClass</code> is health-status-DEAD-YELLOW (lost heartbeat)
   * @type {String}
   */
  disabled: function () {
    return (this.get('parentView.content.healthClass') === "health-status-DEAD-YELLOW") ? 'disabled' : '';
  }.property('parentView.content.healthClass'),

  /**
   * For Upgrade failed state
   * @type {bool}
   */
  isUpgradeFailed: function () {
    return App.HostComponentStatus.getKeyName(this.get('workStatus')) === "upgrade_failed";
  }.property("workStatus"),

  /**
   * For Install failed state
   * @type {bool}
   */
  isInstallFailed: function () {
    return App.HostComponentStatus.getKeyName(this.get('workStatus')) === "install_failed";
  }.property("workStatus"),

  /**
   * For Started and Starting states
   * @type {bool}
   */
  isStart: function () {
    return [App.HostComponentStatus.started, App.HostComponentStatus.starting].contains(this.get('workStatus'));
  }.property('workStatus'),

  /**
   * For Installed state
   * @type {bool}
   */
  isStop: function () {
    return (this.get('workStatus') == App.HostComponentStatus.stopped);
  }.property('workStatus'),

  /**
   * For Installing state
   * @type {bool}
   */
  isInstalling: function () {
    return (this.get('workStatus') == App.HostComponentStatus.installing);
  }.property('workStatus'),

  /**
   * For Init state
   * @type {bool}
   */
  isInit: function() {
    return this.get('workStatus') == App.HostComponentStatus.init;
  }.property('workStatus'),

  /**
   * No action available while component is starting/stopping/unknown
   * @type {String}
   */
  noActionAvailable: function () {
    var workStatus = this.get('workStatus');
    return [App.HostComponentStatus.starting, App.HostComponentStatus.stopping,
      App.HostComponentStatus.unknown, App.HostComponentStatus.disabled].contains(workStatus) ? "hidden" : '';
  }.property('workStatus'),

  /**
   * For Stopping or Starting states
   * @type {bool}
   */
  isInProgress: function () {
    return (this.get('workStatus') === App.HostComponentStatus.stopping ||
      this.get('workStatus') === App.HostComponentStatus.starting);
  }.property('workStatus'),

  /**
   * For OFF <code>passiveState</code> of host component
   * @type {bool}
   */
  isActive: function () {
    return (this.get('content.passiveState') == "OFF");
  }.property('content.passiveState'),

  /**
   * Shows whether we need to show Delete button
   * @type {bool}
   */
  isDeletableComponent: function () {
    return App.get('components.deletable').contains(this.get('content.componentName'));
  }.property('content'),

  /**
   * Host component with some <code>workStatus</code> can't be moved (so, disable such action in the dropdown list)
   * @type {bool}
   */
  isMoveComponentDisabled: function () {
    return App.allHostNames.length === App.HostComponent.find().filterProperty('componentName', this.get('content.componentName')).mapProperty('hostName').length;
  }.property('content'),

  /**
   * Host component with some <code>workStatus</code> can't be deleted (so, disable such action in the dropdown list)
   * @type {bool}
   */
  isDeleteComponentDisabled: function () {
    var stackComponentCount = App.StackServiceComponent.find(this.get('hostComponent.componentName')).get('minToInstall');
    var installedCount = this.componentCounter();
    return (installedCount <= stackComponentCount)
      || ![App.HostComponentStatus.stopped, App.HostComponentStatus.unknown, App.HostComponentStatus.install_failed, App.HostComponentStatus.upgrade_failed, App.HostComponentStatus.init].contains(this.get('workStatus'));
  }.property('workStatus'),

  /**
   * gets number of current component that are applied to the cluster;
   * @returns {Number}
   */
  componentCounter: function() {
    return App.StackServiceComponent.find(this.get('hostComponent.componentName')).get('isMaster')
      ? App.HostComponent.find().filterProperty('componentName', this.get('content.componentName')).length
      : App.SlaveComponent.find().findProperty('componentName', this.get('content.componentName')).get('totalCount');
  },

  /**
   * Check if component may be reassinged to another host
   * @type {bool}
   */
  isReassignable: function () {
    return App.get('components.reassignable').contains(this.get('content.componentName')) && App.router.get('mainHostController.hostsCountMap')['TOTAL'] > 1;
  }.property('content.componentName'),

  /**
   * Check if component is restartable
   * @type {bool}
   */
  isRestartableComponent: function() {
    return App.get('components.restartable').contains(this.get('content.componentName'));
  }.property('content'),

  /**
   * Host component with some <code>workStatus</code> can't be restarted (so, disable such action in the dropdown list)
   * @type {bool}
   */
  isRestartComponentDisabled: function() {
    return ![App.HostComponentStatus.started].contains(this.get('workStatus'));
  }.property('workStatus'),

  /**
   * Check if component configs can be refreshed
   * @type {bool}
   */
  isRefreshConfigsAllowed: function() {
    return App.get('components.refreshConfigsAllowed').contains(this.get('content.componentName'));
  }.property('content'),

  didInsertElement: function () {
    App.tooltip($('[rel=componentHealthTooltip]'));
    App.tooltip($('[rel=passiveTooltip]'));
    if (this.get('isInProgress')) {
      this.doBlinking();
    }
  },

  /**
   * Do blinking for 1 minute
   */
  doBlinking: function () {
    var workStatus = this.get('workStatus');
    var self = this;
    var pulsate = [ App.HostComponentStatus.starting, App.HostComponentStatus.stopping, App.HostComponentStatus.installing].contains(workStatus);
    if (pulsate && !self.get('isBlinking')) {
      self.set('isBlinking', true);
      uiEffects.pulsate(self.$('.components-health'), 1000, function () {
        self.set('isBlinking', false);
        self.doBlinking();
      });
    }
  },

  /**
   * Start blinking when host component is starting/stopping
   */
  startBlinking: function () {
    this.$('.components-health').stop(true, true);
    this.$('.components-health').css({opacity: 1.0});
    this.doBlinking();
  }.observes('workStatus'),

  /**
   * Get custom commands for slave components
   */
  customCommands: function() {
    var hostComponent = this.get('content');
    var component = App.StackServiceComponent.find(hostComponent.get('componentName'));
    var customCommands = [];
    var commands;

    if (component.get('isSlave')) {
      commands = component.get('customCommands');
      commands.forEach(function(command) {
        customCommands.push({
          label: Em.I18n.t('services.service.actions.run.executeCustomCommand.menu').format(command),
          service: component.get('serviceName'),
          hosts: hostComponent.get('hostName'),
          component: component.get('componentName'),
          command: command
        });
      });
    }

    return customCommands;
  }.property('content')

});

});

require.register("views/main/host/details/host_component_views/datanode_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.DataNodeComponentView = App.HostComponentView.extend(App.Decommissionable, {

  componentForCheckDecommission: 'NAMENODE',

  /**
   * Get component decommission status from server
   * @returns {$.ajax}
   */
  getDNDecommissionStatus: function () {
    // always get datanode decommission statue from active namenode (if NN HA enabled)
    var hdfs = App.HDFSService.find().objectAt(0);
    var activeNNHostName = (!hdfs.get('snameNode') && hdfs.get('activeNameNode')) ? hdfs.get('activeNameNode.hostName') : hdfs.get('nameNode.hostName');
    return App.ajax.send({
      name: 'host.host_component.decommission_status_datanode',
      sender: this,
      data: {
        hostName: activeNNHostName,
        componentName: this.get('componentForCheckDecommission')
      },
      success: 'getDNDecommissionStatusSuccessCallback',
      error: 'getDNDecommissionStatusErrorCallback'
    });
  },

  /**
   * Set received value or null to <code>decommissionedStatusObject</code>
   * @param {Object} response
   * @returns {Object|null}
   */
  getDNDecommissionStatusSuccessCallback: function (response) {
    var statusObject = Em.get(response, 'metrics.dfs.namenode');
    if (!Em.isNone(statusObject)) {
      this.computeStatus(statusObject);
      return statusObject;
    }
    return null;
  },

  /**
   * Set null to <code>decommissionedStatusObject</code> if server returns error
   * @returns {null}
   */
  getDNDecommissionStatusErrorCallback: function () {
    this.set('decommissionedStatusObject', null);
    return null;
  },

  /**
   * load Recommission/Decommission status from adminState of each live node
   */
  loadComponentDecommissionStatus: function () {
    return this.getDNDecommissionStatus();
  },

  setDesiredAdminState: function (desired_admin_state) {
    this.setStatusAs(desired_admin_state);
  },

  /**
   * compute and set decommission state by namenode metrics
   * @param curObj
   */
  computeStatus: function (curObj) {
    var hostName = this.get('content.hostName');

    if (curObj) {
      var liveNodesJson = App.parseJSON(curObj.LiveNodes);
      // HDP-2 stack
      if (liveNodesJson && liveNodesJson[hostName]) {
        switch (liveNodesJson[hostName].adminState) {
          case "In Service":
            this.setStatusAs('INSERVICE');
            break;
          case "Decommission In Progress":
            this.setStatusAs('DECOMMISSIONING');
            break;
          case "Decommissioned":
            this.setStatusAs('DECOMMISSIONED');
            break;
        }
      } else {
        // if namenode is down, get desired_admin_state to decide if the user had issued a decommission
        this.getDesiredAdminState();
      }
    }
  }
});

});

require.register("views/main/host/details/host_component_views/nodemanager_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.NodeManagerComponentView = App.HostComponentView.extend(App.Decommissionable, {

  componentForCheckDecommission: 'RESOURCEMANAGER',

  setDesiredAdminState: function (desiredAdminState) {
    var self = this;
    switch (desiredAdminState) {
      case "INSERVICE":
        // can be decommissioned if already started
        this.setStatusAs(desiredAdminState);
        break;
      case "DECOMMISSIONED":
        this.getDecommissionStatus();
        break;
    }
  },

  /**
   * set decommission status according to component status and resource manager metrics
   */
  setDecommissionStatus: function (curObj) {
    var rmComponent = this.get('content.service.hostComponents').findProperty('componentName', this.get('componentForCheckDecommission')),
        hostName = this.get('content.hostName');

    rmComponent.set('workStatus', curObj.component_state);
    if (curObj.rm_metrics) {
      // Update RESOURCEMANAGER status
      var nodeManagersArray = App.parseJSON(curObj.rm_metrics.cluster.nodeManagers);
      if (nodeManagersArray.findProperty('HostName', hostName)) {
        // decommisioning ..
        this.setStatusAs('DECOMMISSIONING');
      } else {
        // decommissioned ..
        this.setStatusAs('DECOMMISSIONED');
      }
    } else {
      // in this case ResourceManager not started. Set status to Decommissioned
      this.setStatusAs('DECOMMISSIONED');
    }
  }
});

});

require.register("views/main/host/details/host_component_views/regionserver_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.RegionServerComponentView = App.HostComponentView.extend(App.Decommissionable, {

  componentForCheckDecommission: 'HBASE_MASTER',

  setDesiredAdminState: function (desiredAdminState) {
    switch (desiredAdminState) {
      case "INSERVICE":
        this.setStatusAs(desiredAdminState);
        break;
      case "DECOMMISSIONED":
        this.setStatusAs("RS_DECOMMISSIONED");
        break;
    }
  }
});

});

require.register("views/main/host/details/host_component_views/tasktracker_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.TaskTrackerComponentView = App.HostComponentView.extend(App.Decommissionable, {

  componentForCheckDecommission: 'JOBTRACKER',

  setDesiredAdminState: function (desiredAdminState) {
    switch (desiredAdminState) {
      case "INSERVICE":
        // can be decommissioned if already started
        this.setStatusAs('INSERVICE');
        break;
      case "DECOMMISSIONED":
        this.getDecommissionStatus();
        break;
    }
  },

  setDecommissionStatus: function (curObj) {
    var hostName = this.get('content.hostName');

    if (curObj) {
      var aliveNodesArray = App.parseJSON(curObj.AliveNodes);
      if (!Em.isNone(aliveNodesArray)) {
        if (aliveNodesArray.findProperty('hostname', hostName)) {
          //decommissioning ..
          this.setStatusAs("DECOMMISSIONING");
        } else {
          //decommissioned
          this.setStatusAs("DECOMMISSIONED");
        }
      }
    }
  }
});

});

require.register("views/main/host/host_alerts_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var filters = require('views/common/filter_view'),
    sort = require('views/common/sort_view');

App.MainHostAlertsView = App.TableView.extend({
  templateName: require('templates/main/host/host_alerts'),

  content: function () {
    return this.get('controller.content');
  }.property('controller.content.@each'),

  willInsertElement: function () {
    var hostName = this.get('parentView.controller.content.hostName');
    App.router.get('mainAlertInstancesController').loadAlertInstancesByHost(hostName);
    App.router.set('mainAlertInstancesController.isUpdating', true);
    this._super();
  },

  didInsertElement: function () {
    this.tooltipsUpdater();
  },

  /**
   * @type {number}
   */
  totalCount: function () {
    return this.get('content.length');
  }.property('content.length'),

  colPropAssoc: ['', 'serviceName', 'label', 'latestTimestamp', 'state', 'text'],

  sortView: sort.wrapperView,

  /**
   * Sorting header for <label>alertDefinition.label</label>
   * @type {Em.View}
   */
  nameSort: sort.fieldView.extend({
    column: 2,
    name: 'label',
    displayName: Em.I18n.t('alerts.definition.name')
  }),

  /**
   * Sorting header for <label>alertDefinition.status</label>
   * @type {Em.View}
   */
  statusSort: sort.fieldView.extend({
    column: 4,
    name: 'state',
    displayName: Em.I18n.t('common.status'),
    type: 'select'
  }),

  /**
   * Sorting header for <label>alertDefinition.service.serviceName</label>
   * @type {Em.View}
   */
  serviceSort: sort.fieldView.extend({
    column: 1,
    name: 'serviceName',
    displayName: Em.I18n.t('common.service'),
    type: 'string'
  }),

  /**
   * Sorting header for <label>alertDefinition.label</label>
   * @type {Em.View}
   */
  textSort: sort.fieldView.extend({
    column: 5,
    name: 'text',
    displayName: Em.I18n.t('alerts.table.header.check.response')
  }),


  /**
   * Filtering header for <label>alertDefinition.label</label>
   * @type {Em.View}
   */
  nameFilterView: filters.createTextView({
    column: 2,
    fieldType: 'filter-input-width',
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'string');
    }
  }),

  /**
   * Filtering header for <label>alertDefinition.status</label>
   * @type {Em.View}
   */
  stateFilterView: filters.createSelectView({
    column: 4,
    fieldType: 'filter-input-width',
    content: [
      {
        value: '',
        label: Em.I18n.t('common.all')
      },
      {
        value: 'OK',
        label: 'OK'
      },
      {
        value: 'WARNING',
        label: 'WARNING'
      },
      {
        value: 'CRITICAL',
        label: 'CRITICAL'
      },
      {
        value: 'UNKNOWN',
        label: 'UNKNOWN'
      }
    ],
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'select');
    }
  }),

  /**
   * Filtering header for <label>alertDefinition.service.serviceName</label>
   * @type {Em.View}
   */
  serviceFilterView: filters.createSelectView({
    column: 1,
    fieldType: 'filter-input-width',
    content: function () {
      return [
        {
          value: '',
          label: Em.I18n.t('common.all')
        }
      ].concat(App.Service.find().map(function (service) {
            return {
              value: service.get('serviceName'),
              label: service.get('displayName')
            }
          })).concat({
            value: 'AMBARI',
            label: Em.I18n.t('app.name')
          });
    }.property('App.router.clusterController.isLoaded'),
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'select');
    }
  }),

  /**
   * Filtering header for <label>alertDefinition.service.serviceName</label>
   * @type {Em.View}
   */
  textView: filters.createTextView({
    column: 5,
    fieldType: 'filter-input-width',
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'string');
    }
  }),

  /**
   * Filtered number of all content number information displayed on the page footer bar
   * @returns {String}
   */
  filteredContentInfo: function () {
    return this.t('alerts.filters.filteredAlertsInfo').format(this.get('filteredCount'), this.get('totalCount'));
  }.property('filteredCount', 'totalCount'),

  /**
   * Determines how display "back"-link - as link or text
   * @type {string}
   */
  paginationLeftClass: function () {
    if (this.get("startIndex") > 1) {
      return "paginate_previous";
    }
    return "paginate_disabled_previous";
  }.property("startIndex", 'filteredCount'),

  /**
   * Determines how display "next"-link - as link or text
   * @type {string}
   */
  paginationRightClass: function () {
    if ((this.get("endIndex")) < this.get("filteredCount")) {
      return "paginate_next";
    }
    return "paginate_disabled_next";
  }.property("endIndex", 'filteredCount'),

  /**
   * Show previous-page if user not in the first page
   * @method previousPage
   */
  previousPage: function () {
    if (this.get('paginationLeftClass') === 'paginate_previous') {
      this._super();
    }
  },

  /**
   * Show next-page if user not in the last page
   * @method nextPage
   */
  nextPage: function () {
    if (this.get('paginationRightClass') === 'paginate_next') {
      this._super();
    }
  },

  /**
   * Update tooltips when <code>pageContent</code> is changed
   * @method tooltipsUpdater
   */
  tooltipsUpdater: function () {
    Em.run.next(this, function () {
      App.tooltip($(".enable-disable-button, .timeago, .alert-text"));
    });
  }.observes('pageContent.@each'),

  /**
   * Run <code>clearFilter</code> in the each child filterView
   */
  clearFilters: function() {
    this.set('filterConditions', []);
    this.get('childViews').forEach(function(childView) {
      if (childView['clearFilter']) {
        childView.clearFilter();
      }
    });
  }

});

});

require.register("views/main/host/hosts_table_menu_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.HostTableMenuView = Em.View.extend({

  templateName: require('templates/main/host/bulk_operation_menu'),

  menuItems: function () {
    return {
      s: {label: Em.I18n.t('hosts.table.menu.l1.selectedHosts')},
      f: {label: Em.I18n.t('hosts.table.menu.l1.filteredHosts')},
      a: {label: Em.I18n.t('hosts.table.menu.l1.allHosts')}
    };
  }.property('App.router.clusterController.isLoaded'),

  components: function(){
    var menuItems = [
    Em.Object.create({
      serviceName: 'HDFS',
      componentName: 'DATANODE',
      masterComponentName: 'NAMENODE',
      componentNameFormatted: Em.I18n.t('dashboard.services.hdfs.datanodes')
    }),
    Em.Object.create({
      serviceName: 'YARN',
      componentName: 'NODEMANAGER',
      masterComponentName: 'RESOURCEMANAGER',
      componentNameFormatted: Em.I18n.t('dashboard.services.yarn.nodeManagers')
    }),
    Em.Object.create({
      serviceName: 'HBASE',
      componentName: 'HBASE_REGIONSERVER',
      masterComponentName: 'HBASE_MASTER',
      componentNameFormatted: Em.I18n.t('dashboard.services.hbase.regionServers')
    }),
    Em.Object.create({
      serviceName: 'STORM',
      componentName: 'SUPERVISOR',
      masterComponentName: 'SUPERVISOR',
      componentNameFormatted: Em.I18n.t('dashboard.services.storm.supervisors')
    })];

    return menuItems.filter(function(item){
      return App.Service.find().findProperty('serviceName',item.serviceName);
    });
  }.property(),


  /**
   * slaveItemView build second-level menu
   * for slave component
   */

  slaveItemView: Em.View.extend({

    tagName: 'li',

    classNames: ['dropdown-submenu'],

    /**
     * Get third-level menu items ingo for slave components
     * operationData format:
     * <code>
     *  {
     *    action: 'STARTED|INSTALLED|RESTART|DECOMMISSION|DECOMMISSION_OFF', // action for selected host components
     *    message: 'some text', // just text to BG popup
     *    componentName: 'DATANODE|NODEMANAGER...', //component name that should be processed
     *    realComponentName: 'DATANODE|NODEMANAGER...', // used only for decommission(_off) actions
     *    serviceName: 'HDFS|YARN|HBASE...', // service name of the processed component
     *    componentNameFormatted: 'DataNodes|NodeManagers...' // "user-friendly" string with component name (used in BG popup)
     *  }
     *  </code>
     *
     * @returns {Array}
     */

    operationsInfo: function () {
      var content = this.get('content');
      var menuItems = Em.A([
        Em.Object.create({
          label: Em.I18n.t('common.start'),
          operationData: Em.Object.create({
            action: App.HostComponentStatus.started,
            message: Em.I18n.t('common.start'),
            componentName: content.componentName,
            serviceName: content.serviceName,
            componentNameFormatted: content.componentNameFormatted
          })
        }),
        Em.Object.create({
          label: Em.I18n.t('common.stop'),
          operationData: Em.Object.create({
            action: App.HostComponentStatus.stopped,
            message: Em.I18n.t('common.stop'),
            componentName: content.componentName,
            serviceName: content.serviceName,
            componentNameFormatted: content.componentNameFormatted
          })
        }),
        Em.Object.create({
          label: Em.I18n.t('common.restart'),
          operationData: Em.Object.create({
            action: 'RESTART',
            message: Em.I18n.t('common.restart'),
            componentName: content.componentName,
            serviceName: content.serviceName,
            componentNameFormatted: content.componentNameFormatted
          })
        })
      ]);
      if (App.get('components.decommissionAllowed').contains(content.componentName)) {
        menuItems.pushObject(Em.Object.create({
          label: Em.I18n.t('common.decommission'),
          decommission: true,
          operationData: Em.Object.create({
            action: 'DECOMMISSION',
            message: Em.I18n.t('common.decommission'),
            componentName: content.masterComponentName,
            realComponentName: content.componentName,
            serviceName: content.serviceName,
            componentNameFormatted: content.componentNameFormatted
          })
        }));
        menuItems.pushObject(Em.Object.create({
          label: Em.I18n.t('common.recommission'),
          decommission: true,
          operationData: Em.Object.create({
            action: 'DECOMMISSION_OFF',
            message: Em.I18n.t('common.recommission'),
            componentName: content.masterComponentName,
            realComponentName: content.componentName,
            serviceName: content.serviceName,
            componentNameFormatted: content.componentNameFormatted
          })
        }));
      }
      return menuItems;
    }.property("content"),

    /**
     * commonOperationView is used for third-level menu items
     * for simple operations ('START','STOP','RESTART')
     */
    commonOperationView: Em.View.extend({
      tagName: 'li',

      /**
       * click function use
       * App.MainHostView as a thirl level parent
       * and runs it's function
       */
      click: function () {
        this.get('parentView.parentView.parentView').bulkOperationConfirm(this.get('content'), this.get('selection'));
      }
    }),

    /**
     * advancedOperationView is used for third level menu item
     * for advanced operations ('RECOMMISSION','DECOMMISSION')
     */
    advancedOperationView: Em.View.extend({
      tagName: 'li',
      rel: 'menuTooltip',
      classNameBindings: ['disabledElement'],
      attributeBindings: ['tooltipMsg:data-original-title'],

      service: function () {
        return App.router.get('mainServiceController.content').findProperty('serviceName', this.get('content.serviceName'))
      }.property('App.router.mainServiceController.content.@each', 'content'),

      tooltipMsg: function () {
        return (this.get('disabledElement') == 'disabled') ?
           Em.I18n.t('hosts.decommission.tooltip.warning').format(this.get('content.message'),  App.format.role(this.get('content.componentName'))) : '';
      }.property('disabledElement','content.componentName'),

      disabledElement: function () {
        return (this.get('service.workStatus') != 'STARTED') ? 'disabled' : '';
      }.property('service.workStatus'),

      /**
       * click function use
       * App.MainHostView as a thirl level parent
       * and runs it's function
       */
      click: function () {
        if (this.get('disabledElement') == 'disabled') {
          return;
        }
        this.get('parentView.parentView.parentView').bulkOperationConfirm(this.get('content'), this.get('selection'));
      },

      didInsertElement: function () {
        App.tooltip($(this.get('element')));
      }
    })
  }),

  /**
   * hostItemView build second-level menu
   * for host
   */

  hostItemView: Em.View.extend({

    tagName: 'li',

    classNames: ['dropdown-submenu'],

    label: Em.I18n.t('common.hosts'),

    /** Get third-level menu items for Hosts
     * operationData format:
     * <code>
     *  {
     *    action: 'STARTED|INSTALLED|RESTART..', // action for selected hosts (will be applied for each host component in selected hosts)
     *    actionToCheck: 'INSTALLED|STARTED..' // state to filter host components should be processed
     *    message: 'some text', // just text to BG popup
     *  }
     *  </code>
     * @returns {Array}
     */
    operationsInfo: function () {
      return Em.A([
        Em.Object.create({
          label: Em.I18n.t('hosts.host.details.startAllComponents'),
          operationData: Em.Object.create({
            action: 'STARTED',
            actionToCheck: 'INSTALLED',
            message: Em.I18n.t('hosts.host.details.startAllComponents')
          })
        }),
        Em.Object.create({
          label: Em.I18n.t('hosts.host.details.stopAllComponents'),
          operationData: Em.Object.create({
            action: 'INSTALLED',
            actionToCheck: 'STARTED',
            message: Em.I18n.t('hosts.host.details.stopAllComponents')
          })
        }),
        Em.Object.create({
          label: Em.I18n.t('hosts.table.menu.l2.restartAllComponents'),
          operationData: Em.Object.create({
            action: 'RESTART',
            message: Em.I18n.t('hosts.table.menu.l2.restartAllComponents')
          })
        }),
        Em.Object.create({
          label: Em.I18n.t('passiveState.turnOn'),
          operationData: Em.Object.create({
            state: 'ON',
            action: 'PASSIVE_STATE',
            message: Em.I18n.t('passiveState.turnOnFor').format('hosts')
          })
        }),
        Em.Object.create({
          label: Em.I18n.t('passiveState.turnOff'),
          operationData: Em.Object.create({
            state: 'OFF',
            action: 'PASSIVE_STATE',
            message: Em.I18n.t('passiveState.turnOffFor').format('hosts')
          })
        })
      ]);
    }.property(),

    /**
     * commonOperationView is used for third-level menu items
     * for all operations for host
     */
    operationView: Em.View.extend({
      tagName: 'li',

      /**
       * click function use
       * App.MainHostView as a thirl level parent
       * and runs it's function
       */
      click: function () {
        this.get('parentView.parentView.parentView').bulkOperationConfirm(this.get('content'), this.get('selection'));
      }
    })
  })
});

});

require.register("views/main/host/menu", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainHostMenuView = Em.CollectionView.extend({
  tagName: 'ul',
  classNames: ["nav", "nav-tabs"],
  host: null,

  content: function () {
    return [
      Em.Object.create({
        name: 'summary',
        label: Em.I18n.t('common.summary'),
        routing: 'summary',
        id: 'host-details-summary-tab'
      }),
      Em.Object.create({
        name: 'configs',
        label: Em.I18n.t('common.configs'),
        routing: 'configs',
        id: 'host-details-summary-configs'
      }),
      Em.Object.create({
        name: 'alerts',
        label: Em.I18n.t('hosts.host.alerts.label'),
        routing: 'alerts',
        badgeText: '0',
        badgeClasses: 'label',
        id: 'host-details-summary-alerts'
      }),
      Em.Object.create({
        name: 'versions',
        label: Em.I18n.t('hosts.host.menu.stackVersions'),
        routing: 'stackVersions',
        hidden: function () {
          return !App.get('supports.stackUpgrade') || !App.get('stackVersionsAvailable')
        }.property('App.supports.stackUpgrade'),
        id: 'host-details-summary-version'
      })
    ];
  }.property('App.stackVersionsAvailable'),

  /**
   * Update Alerts menu option counter text and class
   */
  updateAlertCounter: function () {
    var criticalWarningCount = this.get('host.criticalWarningAlertsCount');
    var criticalCount = this.get('host.alertsSummary.CRITICAL');
    var warningCount = this.get('host.alertsSummary.WARNING');
    var badgeText = "" + criticalWarningCount;
    var badgeClasses = "label";
    if (criticalCount > 0) {
      badgeClasses += " alerts-crit-count";
    } else if (warningCount > 0) {
      badgeClasses += " alerts-warn-count";
    }
    var alertOption = this.get('content').findProperty('name', 'alerts');
    alertOption.set('badgeText', badgeText);
    alertOption.set('badgeClasses', badgeClasses);
  }.observes('host.alertsSummary.CRITICAL', 'host.alertsSummary.WARNING', 'host.criticalWarningAlertsCount'),

  init: function () {
    this._super();
    this.updateAlertCounter();
    this.activateView();
  },

  activateView: function () {
    var defaultRoute = App.router.get('currentState.name') || "summary";
    $.each(this._childViews, function () {
      this.set('active', (this.get('content.routing') == defaultRoute ? "active" : ""));
    });
  },

  deactivateChildViews: function () {
    $.each(this._childViews, function () {
      this.set('active', "");
    });
  },

  itemViewClass: Em.View.extend({
    classNameBindings: ["active"],
    active: "",
    template: Ember.Handlebars.compile('{{#unless view.content.hidden}}<a {{action hostNavigate view.content.routing }} {{bindAttr id="view.content.id"}} href="#"> {{unbound view.content.label}} ' +
    '{{#if view.content.badgeText}} ' +
    '<span {{bindAttr class="view.content.badgeClasses"}}> ' +
    '{{view.content.badgeText}}' +
    '</span>  {{/if}}</a>{{/unless}}')
  })
});

});

require.register("views/main/host/metrics", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainHostMetricsView = Em.View.extend({
  templateName: require('templates/main/host/metrics'),
  content:function(){
    return App.router.get('mainHostDetailsController.content');
  }.property('App.router.mainHostDetailsController.content')
});
});

require.register("views/main/host/metrics/cpu", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing Host CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartHostMetricsCPU = App.ChartLinearTimeView.extend({
  id: "host-metrics-cpu",
  title: Em.I18n.t('hosts.host.metrics.cpu'),
  yAxisFormatter: App.ChartLinearTimeView.PercentageFormatter,

  ajaxIndex: 'host.metrics.cpu',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.cpu) {
      var cpu_idle;
      for ( var name in jsonData.metrics.cpu) {
        var displayName;
        var seriesData = jsonData.metrics.cpu[name];
        switch (name) {
          case "cpu_wio":
            displayName = Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_wio');
            break;
          case "cpu_idle":
            displayName = Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_idle');
            break;
          case "cpu_nice":
            displayName = Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_nice');
            break;
          case "cpu_aidle":
            displayName = Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_aidle');
            break;
          case "cpu_system":
            displayName = Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_system');
            break;
          case "cpu_user":
            displayName = Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_user');
            break;
          default:
            break;
        }
        if (seriesData) {
          var s = this.transformData(seriesData, displayName);
          if (Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_idle') == s.name) {
            cpu_idle = s;
          }
          else {
            seriesArray.push(s);
          }
        }
      }
      seriesArray.push(cpu_idle);
    }
    return seriesArray;
  },

  colorForSeries: function (series) {
    if (Em.I18n.t('hosts.host.metrics.cpu.displayNames.cpu_idle') == series.name) {
      return '#CFECEC';
    }
    return null;
  }
});
});

require.register("views/main/host/metrics/disk", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing host disk usage
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartHostMetricsDisk = App.ChartLinearTimeView.extend({
  id: "host-metrics-disk",
  title: Em.I18n.t('hosts.host.metrics.disk'),
  yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
  renderer: 'line',

  ajaxIndex: 'host.metrics.disk',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var GB = Math.pow(2, 30);
    if (jsonData && jsonData.metrics && jsonData.metrics.disk) {
      if(jsonData.metrics.part_max_used){
        jsonData.metrics.disk.part_max_used = jsonData.metrics.part_max_used;
      }
      for ( var name in jsonData.metrics.disk) {
        var displayName;
        var seriesData = jsonData.metrics.disk[name];
        switch (name) {
          case "disk_total":
            displayName = Em.I18n.t('hosts.host.metrics.disk.displayNames.disk_total');
            break;
          case "disk_free":
            displayName = Em.I18n.t('hosts.host.metrics.disk.displayNames.disk_free');
            break;
          default:
            break;
        }
        if (seriesData) {
          var s = this.transformData(seriesData, displayName);
          for (var i = 0; i < s.data.length; i++) {
            s.data[i].y *= GB;
          }
          seriesArray.push(s);
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/host/metrics/load", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing host load
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartHostMetricsLoad = App.ChartLinearTimeView.extend({
  id: "host-metrics-load",
  title: Em.I18n.t('hosts.host.metrics.load'),
  renderer: 'line',

  ajaxIndex: 'host.metrics.load',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.load) {
      for ( var name in jsonData.metrics.load) {
        var displayName;
        var seriesData = jsonData.metrics.load[name];
        switch (name) {
          case "load_fifteen":
            displayName = Em.I18n.t('hosts.host.metrics.load.displayNames.load_fifteen');
            break;
          case "load_one":
            displayName = Em.I18n.t('hosts.host.metrics.load.displayNames.load_one');
            break;
          case "load_five":
            displayName = Em.I18n.t('hosts.host.metrics.load.displayNames.load_five');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/host/metrics/memory", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing host memory metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartHostMetricsMemory = App.ChartLinearTimeView.extend({
  id: "host-metrics-memory",
  title: Em.I18n.t('hosts.host.metrics.memory'),
  yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
  renderer: 'line',

  ajaxIndex: 'host.metrics.memory',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var KB = Math.pow(2, 10);
    if (jsonData && jsonData.metrics && jsonData.metrics.memory) {
      for ( var name in jsonData.metrics.memory) {
        var displayName;
        var seriesData = jsonData.metrics.memory[name];
        switch (name) {
          case "mem_shared":
            displayName = Em.I18n.t('hosts.host.metrics.memory.displayNames.mem_shared');
            break;
          case "swap_free":
            displayName = Em.I18n.t('hosts.host.metrics.memory.displayNames.swap_free');
            break;
          case "mem_buffers":
            displayName = Em.I18n.t('hosts.host.metrics.memory.displayNames.mem_buffers');
            break;
          case "mem_free":
            displayName = Em.I18n.t('hosts.host.metrics.memory.displayNames.mem_free');
            break;
          case "mem_cached":
            displayName = Em.I18n.t('hosts.host.metrics.memory.displayNames.mem_cached');
            break;
          default:
            break;
        }
        if (seriesData) {
          var s = this.transformData(seriesData, displayName);
          for (var i = 0; i < s.data.length; i++) {
            s.data[i].y *= KB;
          }
          seriesArray.push(s);
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/host/metrics/network", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing host network metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartHostMetricsNetwork = App.ChartLinearTimeView.extend({
  id: "host-metrics-network",
  title: Em.I18n.t('hosts.host.metrics.network'),
  yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
  renderer: 'line',

  ajaxIndex: 'host.metrics.network',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.network) {
      for ( var name in jsonData.metrics.network) {
        var displayName;
        var seriesData = jsonData.metrics.network[name];
        switch (name) {
          case "pkts_out":
            displayName = Em.I18n.t('hosts.host.metrics.network.displayNames.pkts_out');
            break;
          case "bytes_in":
            displayName = Em.I18n.t('hosts.host.metrics.network.displayNames.bytes_in');
            break;
          case "bytes_out":
            displayName = Em.I18n.t('hosts.host.metrics.network.displayNames.bytes_out');
            break;
          case "pkts_in":
            displayName = Em.I18n.t('hosts.host.metrics.network.displayNames.pkts_in');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/host/metrics/processes", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing host process counts
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartHostMetricsProcesses = App.ChartLinearTimeView.extend({
  id: "host-metrics-processes",
  title: Em.I18n.t('hosts.host.metrics.processes'),
  renderer: 'line',

  ajaxIndex: 'host.metrics.processes',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.process) {
      for ( var name in jsonData.metrics.process) {
        var displayName;
        var seriesData = jsonData.metrics.process[name];
        switch (name) {
          case "proc_total":
            displayName = Em.I18n.t('hosts.host.metrics.processes.displayNames.proc_total');
            break;
          case "proc_run":
            displayName = Em.I18n.t('hosts.host.metrics.processes.displayNames.proc_run');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/host/stack_versions_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var filters = require('views/common/filter_view');
var sort = require('views/common/sort_view');

App.MainHostStackVersionsView = App.TableView.extend({
  templateName: require('templates/main/host/stack_versions'),
  classNames: ['host-tab-content'],

  /**
   * @type {Ember.Object}
   */
  host: function () {
    return App.router.get('mainHostDetailsController.content');
  }.property('App.router.mainHostDetailsController.content'),

  /**
   * @type {Ember.Array}
   */
  content: function () {
    return this.get('host.stackVersions').filterProperty('isVisible', true);
  }.property('host.stackVersions.length'),

  /**
   * return filtered number of all content number information displayed on the page footer bar
   * @returns {string}
   */
  filteredContentInfo: function () {
    return this.t('hosts.host.stackVersions.table.filteredInfo').format(this.get('filteredCount'), this.get('totalCount'));
  }.property('filteredCount', 'totalCount'),

  /**
   * @type {Ember.View}
   */
  sortView: sort.wrapperView,

  /**
   * @type {Ember.View}
   */
  stackSort: sort.fieldView.extend({
    column: 1,
    name: 'stack',
    displayName: Em.I18n.t('common.stack'),
    type: 'version'
  }),

  /**
   * @type {Ember.View}
   */
  repoVersionSort: sort.fieldView.extend({
    column: 2,
    name: 'displayName',
    displayName: Em.I18n.t('common.name'),
    type: 'version'
  }),

  /**
   * @type {Ember.View}
   */
  statusSort: sort.fieldView.extend({
    column: 3,
    name: 'status',
    displayName: Em.I18n.t('common.status')
  }),

  /**
   * Filter view for stackName column
   * Based on <code>filters</code> library
   * @type {Ember.View}
   */
  stackFilterView: filters.createSelectView({
    column: 1,
    fieldType: 'filter-input-width',
    content: function () {
      return [
        {
          value: '',
          label: Em.I18n.t('common.all')
        }
      ].concat(this.get('parentView.content').mapProperty('stack').uniq().map(function (item) {
        return {
          value: item,
          label: item
        }
      }));
    }.property('App.router.clusterController.isLoaded'),
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'select');
    }
  }),

  /**
   * Filter view for version column
   * Based on <code>filters</code> library
   * @type {Ember.View}
   */
  repoVersionFilterView: filters.createSelectView({
    column: 2,
    fieldType: 'filter-input-width',
    content: function () {
      return [
        {
          value: '',
          label: Em.I18n.t('common.all')
        }
      ].concat(this.get('parentView.content').mapProperty('displayName').uniq().map(function (version) {
        return {
          value: version,
          label: version
        }
      }));
    }.property('App.router.clusterController.isLoaded'),
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'select');
    }
  }),

  /**
   * Filter view for status column
   * Based on <code>filters</code> library
   * @type {Ember.View}
   */
  statusFilterView: filters.createSelectView({
    column: 3,
    fieldType: 'filter-input-width',
    content: function () {
      return [
        {
          value: '',
          label: Em.I18n.t('common.all')
        }
      ].concat(App.HostStackVersion.statusDefinition.map(function (status) {
        return {
          value: status,
          label: App.HostStackVersion.formatStatus(status)
        }
      }));
    }.property(),
    onChangeValue: function () {
      this.get('parentView').updateFilter(this.get('column'), this.get('value'), 'select');
    }
  }),

  /**
   * show progress of installation of version on host
   */
  showInstallProgress: function (event) {
    App.router.get('mainAdminStackAndUpgradeController').showProgressPopup(event.context);
  },

  outOfSyncInfo: Em.View.extend({
    tagName: 'i',
    classNames: ['icon-question-sign'],
    didInsertElement: function() {
      App.tooltip($(this.get('element')), {
        placement: "top",
        title: Em.I18n.t('hosts.host.stackVersions.outOfSync.tooltip')
      });
    }
  }),

  /**
   * @type {Array}
   */
  colPropAssoc: function () {
    var associations = [];
    associations[1] = 'stack';
    associations[2] = 'displayName';
    associations[3] = 'status';
    return associations;
  }.property()

});

});

require.register("views/main/host/summary", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainHostSummaryView = Em.View.extend({

  templateName: require('templates/main/host/summary'),

  /**
   * List of custom view for some host components
   * @type {Em.Object}
   * Format:
   *  <code>
   *    {
   *      COMPONENT_NAME1: VIEW1,
   *      COMPONENT_NAME2: VIEW2
   *      ....
   *    }
   *  </code>
   */
  hostComponentViewMap: Em.Object.create({
    'DATANODE': App.DataNodeComponentView,
    'NODEMANAGER': App.NodeManagerComponentView,
    'HBASE_REGIONSERVER': App.RegionServerComponentView,
    'TASKTRACKER': App.TaskTrackerComponentView
  }),

  /**
   * @type {bool}
   */
  isStopCommand: true,

  /**
   * @type {App.Host}
   */
  content: function () {
    return App.router.get('mainHostDetailsController.content');
  }.property('App.router.mainHostDetailsController.content'),

  showGangliaCharts: function () {
    var name = this.get('content.hostName');
    var gangliaMobileUrl = App.router.get('clusterController.gangliaUrl') + "/mobile_helper.php?show_host_metrics=1&h=" + name + "&c=HDPSlaves&r=hour&cs=&ce=";
    window.open(gangliaMobileUrl);
  },

  /**
   * Host metrics panel not displayed when Metrics service (ex:Ganglia) is not in stack definition.
   */
  isNoHostMetricsService: function() {
    return !App.get('services.hostMetrics').length;
  }.property('App.services.hostMetrics'),

  /**
   * Message for "restart" block
   * @type {String}
   */
  needToRestartMessage: function() {
    var componentsCount, word;
    componentsCount = this.get('content.componentsWithStaleConfigsCount');
    if (componentsCount > 1) {
      word = Em.I18n.t('common.components').toLowerCase();
    } else {
      word = Em.I18n.t('common.component').toLowerCase();
    }
    return Em.I18n.t('hosts.host.details.needToRestart').format(this.get('content.componentsWithStaleConfigsCount'), word);
  }.property('content.componentsWithStaleConfigsCount'),

  /**
   * Reset <code>sortedComponents</code>
   * Used when some component was deleted from host
   */
  redrawComponents: function() {
    if (App.router.get('mainHostDetailsController.redrawComponents')) {
      this.set('sortedComponents', []);
      this.sortedComponentsFormatter();
      App.router.set('mainHostDetailsController.redrawComponents', false);
    }
  }.observes('App.router.mainHostDetailsController.redrawComponents'),

  willInsertElement: function() {
    this.set('sortedComponents', []);
    this.sortedComponentsFormatter();
    this.addObserver('content.hostComponents.length', this, 'sortedComponentsFormatter');
  },

  didInsertElement: function () {
    this.addToolTip();
  },

  /**
   * Create tooltip for "Add" button if nothing to add to the current host
   */
  addToolTip: function() {
    if (this.get('addComponentDisabled')) {
      App.tooltip($('#add_component'), {title: Em.I18n.t('services.nothingToAdd')});
    }
  }.observes('addComponentDisabled'),

  /**
   * List of installed services
   * @type {String[]}
   */
  installedServices: function () {
    return App.Service.find().mapProperty('serviceName');
  }.property('App.router.clusterController.dataLoadList.serviceMetrics'),

  /**
   * List of installed masters and slaves
   * Masters first, then slaves
   * @type {App.HostComponent[]}
   */
  sortedComponents: [],

  /**
   * Update <code>sortedComponents</code>
   * Master components first, then slaves
   */
  sortedComponentsFormatter: function() {
    var updatebleProperties = Em.A(['workStatus', 'passiveState', 'staleConfigs', 'haStatus']);
    var self = this;
    var hostComponentViewMap = this.get('hostComponentViewMap');
    // Remove deleted components
    this.get('sortedComponents').forEach(function(sortedComponent, index) {
      if (!self.get('content.hostComponents').findProperty('id', sortedComponent.get('id'))) {
        self.get('sortedComponents').removeAt(index, 1);
      }
    });

    this.get('content.hostComponents').forEach(function (component) {
      if (component.get('isMaster') || component.get('isSlave')) {
        var obj = this.get('sortedComponents').findProperty('id', component.get('id'));
        if (obj) {
          // Update existing component
          updatebleProperties.forEach(function(property) {
            obj.set(property, component.get(property));
          });
        }
        else {
          // Add new component
          component.set('view', hostComponentViewMap[component.get('componentName')] ? hostComponentViewMap[component.get('componentName')] : App.HostComponentView);
          if (component.get('isMaster')) {
            // Masters should be before slaves
            var lastMasterIndex = 0, atLeastOneMasterExists = false;
            this.get('sortedComponents').forEach(function(sortedComponent, index) {
              if (sortedComponent.get('isMaster')) {
                lastMasterIndex = index;
                atLeastOneMasterExists = true;
              }
            });
            this.get('sortedComponents').insertAt(atLeastOneMasterExists ? lastMasterIndex + 1 : 0, component);
          }
          else {
            // it is slave 100%
            this.get('sortedComponents').pushObject(component);
          }
        }
      }
    }, this);
  },

  /**
   * List of installed clients
   * @type {App.HostComponent[]}
   */
  clients: function () {
    var clients = [];
    this.get('content.hostComponents').forEach(function (component) {
      if (!component.get('componentName')) {
        //temporary fix because of different data in hostComponents and serviceComponents
        return;
      }
      if (!component.get('isSlave') && !component.get('isMaster')) {
        if (clients.length) {
          clients[clients.length - 1].set('isLast', false);
        }
        component.set('isLast', true);
        if (['INSTALL_FAILED', 'INIT'].contains(component.get('workStatus'))) {
          component.set('isInstallFailed', true);
        }
        clients.push(component);
      }
    }, this);
    return clients;
  }.property('content.hostComponents.length'),
  /**
   * Check if some clients not installed or started
   *
   * @type {bool}
   **/
  areClientsInstallFailed: function() {
    return this.get('clients').someProperty('isInstallFailed', true);
  }.property('clients.@each.workStatus'),

  /**
   * Check if some clients have stale configs
   * @type {bool}
   */
  areClientWithStaleConfigs: function() {
    return !!this.get('clients').filter(function(component) {
      return component.get('staleConfigs');
    }).length;
  }.property('clients.@each.staleConfigs'),

  /**
   * Template for addable component
   * @type {Em.Object}
   */
  addableComponentObject: Em.Object.extend({
    componentName: '',
    subComponentNames: null,
    displayName: function () {
      if (this.get('componentName') === 'CLIENTS') {
        return this.t('common.clients');
      }
      return App.format.role(this.get('componentName'));
    }.property('componentName')
  }),

  /**
   * If host lost heartbeat, components can't be added on it
   * @type {bool}
   */
  isAddComponent: function () {
    return this.get('content.healthClass') !== 'health-status-DEAD-YELLOW';
  }.property('content.healthClass'),

  /**
   * Disable "Add" button if components can't be added to the current host
   * @type {bool}
   */
  addComponentDisabled: function() {
    return (!this.get('isAddComponent')) || (this.get('addableComponents.length') == 0);
  }.property('isAddComponent', 'addableComponents.length'),

  /**
   * List of client's that may be installed to the current host
   * @type {String[]}
   */
  installableClientComponents: function() {
    var clientComponents = App.StackServiceComponent.find().filterProperty('isClient');
    var installedServices = this.get('installedServices');
    var installedClients = this.get('clients').mapProperty('componentName');
    var installableClients = clientComponents.filter(function(componentName) {
      // service for current client is installed but client isn't installed on current host
      return installedServices.contains(clientComponents.get('serviceName')) && !installedClients.contains(clientComponents.get('componentName'));
    });
    return installableClients.mapProperty('componentName');
  }.property('content.hostComponents.length', 'installedServices.length'),

  /**
   * List of components that may be added to the current host
   * @type {Em.Object[]}
   */
  addableComponents: function () {
    var components = [];
    var self = this;
    var installableClients = this.get('installableClientComponents');
    var installedComponents = this.get('content.hostComponents').mapProperty('componentName');
    var addableToHostComponents = App.StackServiceComponent.find().filterProperty('isAddableToHost');
    var installedServices = this.get('installedServices');

    addableToHostComponents.forEach(function(addableComponent) {
      if(installedServices.contains(addableComponent.get('serviceName')) && !installedComponents.contains(addableComponent.get('componentName'))) {
        if ((addableComponent.get('componentName') === 'OOZIE_SERVER') && !App.router.get('mainHostDetailsController.isOozieServerAddable')) {
          return;
        }
        components.pushObject(self.addableComponentObject.create({'componentName': addableComponent.get('componentName'), 'serviceName': addableComponent.get('serviceName')}));
      }
    });
    if (installableClients.length > 0) {
      components.pushObject(this.addableComponentObject.create({ 'componentName': 'CLIENTS', subComponentNames: installableClients }));
    }

    return components;
  }.property('content.hostComponents.length', 'installableClientComponents', 'App.components.addableToHost.@each'),

  /**
   * Formatted with <code>$.timeago</code> value of host's last heartbeat
   * @type {String}
   */
  timeSinceHeartBeat: function () {
    var d = this.get('content.lastHeartBeatTime');
    if (d) {
      return $.timeago(d);
    }
    return "";
  }.property('content.lastHeartBeatTime'),

  /**
   * Get clients with custom commands
   */
  clientsWithCustomCommands: function() {
    var clients = this.get('clients').rejectProperty('componentName', 'KERBEROS_CLIENT');
    var options = [];
    var clientWithCommands;
    clients.forEach(function(client) {
      var componentName = client.get('componentName');
      var customCommands = App.StackServiceComponent.find(componentName).get('customCommands');

      if (customCommands.length) {
        clientWithCommands = {
          label: client.get('displayName'),
          commands: []
        };
        customCommands.forEach(function(command) {
          clientWithCommands.commands.push({
            label: Em.I18n.t('services.service.actions.run.executeCustomCommand.menu').format(command),
            service: client.get('service.serviceName'),
            hosts: client.get('hostName'),
            component: componentName,
            command: command
          });
        });

        options.push(clientWithCommands);
      }
    });

    return options;
  }.property('controller')
});

});

require.register("views/main/menu", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

/**
 * this menu extended by other with modifying content and itemViewClass.template
 * @type {*}
 */
App.MainMenuView = Em.CollectionView.extend({
  tagName: 'ul',
  classNames: ['nav', 'top-nav-menu'],

  views: function () {
    return App.router.get('mainViewsController.ambariViews');
  }.property('App.router.mainViewsController.ambariViews'),

  content: function () {
    var result = [];
    if (App.router.get('loggedIn')) {

      if (App.router.get('clusterController.isLoaded') && App.get('router.clusterInstallCompleted')) {

        result.push(
          { label: Em.I18n.t('menu.item.dashboard'), routing: 'dashboard', active: 'active'},
          { label: Em.I18n.t('menu.item.services'), routing: 'services'},
          { label: Em.I18n.t('menu.item.hosts'), routing: 'hosts', hasAlertsLabel: true},
          { label: Em.I18n.t('menu.item.alerts'), routing: 'alerts'}
        );
        if (App.isAccessible('upgrade_ADMIN')) {
          result.push({ label: Em.I18n.t('menu.item.admin'), routing: 'admin'});
        }
      }
      result.push({ label: Em.I18n.t('menu.item.views'), routing: 'views.index', isView: true, views: this.get('views').filterProperty('visible')});
    }
    return result;
  }.property('App.router.loggedIn', 'views.length',
    'App.router.clusterController.isLoaded', 'App.router.clusterInstallCompleted'),

  itemViewClass: Em.View.extend({

    classNameBindings: ['active', ':top-nav-dropdown'],

    active: function () {
      if (App.get('clusterName') && App.router.get('clusterController.isLoaded')) {
        var last_url = App.router.location.lastSetURL || location.href.replace(/^[^#]*#/, '');
        if (last_url.substr(1, 4) !== 'main' || !this._childViews) {
          return;
        }
        var reg = /^\/main\/([a-z]+)/g;
        var sub_url = reg.exec(last_url);
        var chunk = (null != sub_url) ? sub_url[1] : 'dashboard';
        return this.get('content.routing').indexOf(chunk) === 0 ? "active" : "";
      }
      return "";
    }.property('App.router.location.lastSetURL', 'App.router.clusterController.isLoaded'),

    alertsCount: function () {
      return App.router.get('mainHostController.hostsCountMap.health-status-WITH-ALERTS');
    }.property('App.router.mainHostController.hostsCountMap'),

    hasCriticalAlerts: function () {
      return App.router.get('mainHostController.hostsCountMap.health-status-CRITICAL') > 0;
    }.property('content.hasAlertsLabel', 'alertsCount'),

    hasAlertsLabel: function () {
      return this.get('content.hasAlertsLabel') && this.get('alertsCount') > 0;
    }.property('content.hasAlertsLabel', 'alertsCount'),

    templateName: require('templates/main/menu_item'),

    dropdownMenu: function () {
      var item = this.get('content').routing;
      var itemsWithDropdown = ['services', 'admin', 'views'];
      return itemsWithDropdown.contains(item);
    }.property(''),
    isAdminItem: function () {
      return this.get('content').routing == 'admin';
    }.property(''),
    isServicesItem: function () {
      return this.get('content').routing == 'services';
    }.property(''),
    isViewsItem: function () {
      return this.get('content').routing.contains('views');
    }.property(''),
    goToSection: function (event) {
      if (event.context === 'hosts') {
        App.router.set('mainHostController.showFilterConditionsFirstLoad', false);
      } else if (event.context === 'views') {
        App.router.route('views');
        return;
      } else if (event.context === 'alerts') {
        App.router.set('mainAlertDefinitionsController.showFilterConditionsFirstLoad', false);
      }
      App.router.route('main/' + event.context);
    },

    selectedAdminItemBinding: 'App.router.mainAdminController.category',

    dropdownCategories: function () {
      var itemName = this.get('content').routing;
      var categories = [];
      // create dropdown categories for each menu item
      if (itemName == 'admin') {
        categories = [];
        categories.push({
          name: 'stackAndUpgrade',
          url: 'stack',
          label: Em.I18n.t('admin.stackUpgrade.title')
        });
        categories.push({
          name: 'adminServiceAccounts',
          url: 'serviceAccounts',
          label: Em.I18n.t('common.serviceAccounts')
        });
        if (!App.get('isHadoopWindowsStack')) {
          if (App.get('supports.automatedKerberos')) {
            categories.push({
              name: 'kerberos',
              url: 'kerberos/',
              label: Em.I18n.t('common.kerberos')
            });
          } else {
            categories.push({
              name: 'security',
              url: 'security/',
              label: Em.I18n.t('common.security')
            });
          }
        }
      }
      return categories;
    }.property(''),

    AdminDropdownItemView: Ember.View.extend({
      tagName: 'li',
      classNameBindings: 'isActive:active'.w(),
      isActive: function () {
        return this.get('item') === this.get('parentView.selectedAdminItem');
      }.property('item', 'parentView.selectedAdminItem'),

      goToCategory: function (event) {
        var itemName = this.get('parentView').get('content').routing;
        // route to correct category of current menu item
        // skip routing to already selected category
        if (itemName === 'admin' && !this.get('isActive')) {
          App.router.route('main/admin/' + event.context);
        }
      }
    })
  })
});

});

require.register("views/main/service", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainServiceView = Em.View.extend({
  templateName:require('templates/main/service'),
  didInsertElement: function() {
    this.addToolTip();
  },
  addToolTip: function() {
    /* istanbul ignore next */
    if (this.get('controller.isAllServicesInstalled')) {
      App.tooltip($('.add-service-button a'), {title: Em.I18n.t('services.nothingToAdd')});
    }
  }.observes('controller.isAllServicesInstalled')
});
});

require.register("views/main/service/add_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.AddServiceView = Em.View.extend(App.WizardMenuMixin, {

  templateName: require('templates/main/service/add')

});

});

require.register("views/main/service/all_services_actions", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.AllServicesActionView = Em.View.extend({
  templateName: require('templates/main/service/all_services_actions')
});

});

require.register("views/main/service/info/components_list_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.SummaryMasterComponentsView = Em.View.extend({
  templateName: require('templates/main/service/info/summary/master_components'),
  mastersComp: null,
  didInsertElement: function() {
    App.tooltip($('[rel=healthTooltip]'));
  }
});

App.SummaryClientComponentsView = Em.View.extend({
  templateName: require('templates/main/service/info/summary/client_components'),
  clientsObj: []
});

App.SummarySlaveComponentsView = Em.View.extend({
  templateName: require('templates/main/service/info/summary/slave_components'),
  slavesObj: []
});

});

require.register("views/main/service/info/configs", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var batchUtils = require('utils/batch_scheduled_requests');

App.MainServiceInfoConfigsView = Em.View.extend({
  templateName: require('templates/main/service/info/configs'),
  didInsertElement: function () {
    var controller = this.get('controller');
    controller.loadStep();
  },

  componentsCount: null,
  hostsCount: null,
  isStopCommand:true,

  updateComponentInformation: function() {
    var hc = this.get('controller.content.restartRequiredHostsAndComponents');
    var hostsCount = 0;
    var componentsCount = 0;
    for (var host in hc) {
      hostsCount++;
      componentsCount += hc[host].length;
    }
    this.set('componentsCount', componentsCount);
    this.set('hostsCount', hostsCount);
  }.observes('controller.content.restartRequiredHostsAndComponents'),

  rollingRestartSlaveComponentName : function() {
    return batchUtils.getRollingRestartComponentName(this.get('controller.content.serviceName'));
  }.property('controller.content.serviceName'),

  rollingRestartActionName : function() {
    var label = null;
    var componentName = this.get('rollingRestartSlaveComponentName');
    if (componentName) {
      label = Em.I18n.t('rollingrestart.dialog.title').format(App.format.role(componentName));
    }
    return label;
  }.property('rollingRestartSlaveComponentName')
});

});

require.register("views/main/service/info/menu", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainServiceInfoMenuView = Em.CollectionView.extend({
  tagName: 'ul',
  classNames: ["nav", "nav-tabs"],
  content:function(){
    var menuItems = [
      { label: Em.I18n.t('services.service.info.menu.summary'), id: 'summary-service-tab',routing:'summary', active:"active"}
      //{ label:'Audit', routing:'audit'}
    ];
    if(this.get('configTab')) menuItems.push({ label: Em.I18n.t('services.service.info.menu.configs'), id: 'configs-service-tab', routing:'configs'});
    return menuItems;
  }.property(),

  init: function(){ this._super(); this.activateView(); },

  activateView:function () {
    $.each(this._childViews, function () {
      this.set('active', (document.URL.endsWith(this.get('content.routing')) ? "active" : ""));
    });
  }.observes('App.router.location.lastSetURL'),

  deactivateChildViews: function() {
    $.each(this._childViews, function(){
      this.set('active', "");
    });
  },

  itemViewClass: Em.View.extend({
    classNameBindings: ["active"],
    active: "",
    template: Ember.Handlebars.compile('<a {{action showInfo view.content.routing}} {{bindAttr id="view.content.id"}} href="#"> {{unbound view.content.label}}</a>')
  })
});
});

require.register("views/main/service/info/metric_graphs_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
require('views/main/service/service');

App.MainServiceInfoSummaryMetricGraphsView = Em.View.extend({
  templateName: require('templates/main/service/info/metric_graphs'),

  /**
   * Data used by this view which is set from the template.
   */
  viewData: null,

  /**
   * Contains graphs for this particular service
   */
  serviceMetricGraphs: null
});

});

require.register("views/main/service/info/metrics/flume/channel_fill_pct", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_ChannelFillPercent = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-channel-fill-percent",
  title: Em.I18n.t('services.service.info.metrics.flume.channelFillPercent'),
  yAxisFormatter: App.ChartLinearTimeView.PercentageFormatter,

  ajaxIndex: 'service.metrics.flume.channel_fill_percent',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;
    if (jsonData && jsonData.host_components) {
      jsonData.host_components.forEach(function (hc) {
        var hostName = hc.HostRoles.host_name;
        var host = App.Host.find(hostName);
        if (host && host.get('publicHostName')) {
          hostName = host.get('publicHostName');
        }
        if (hc.metrics && hc.metrics.flume && hc.metrics.flume.flume && hc.metrics.flume.flume.CHANNEL) {
          for ( var cname in hc.metrics.flume.flume.CHANNEL) {
            var seriesName = Em.I18n.t('services.service.info.metrics.flume.channelName').format(cname + " (" + hostName + ")");
            var seriesData = hc.metrics.flume.flume.CHANNEL[cname]['ChannelFillPercentage'];
            if (seriesData) {
              seriesArray.push(self.transformData(seriesData, seriesName));
            }
          }
        }
      });
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/flume/channel_size", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_ChannelSize = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-channel-size",
  title: Em.I18n.t('services.service.info.metrics.flume.channelSize'),

  ajaxIndex: 'service.metrics.flume.channel_size',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;
    if (jsonData && jsonData.host_components) {
      jsonData.host_components.forEach(function (hc) {
        var hostName = hc.HostRoles.host_name;
        var host = App.Host.find(hostName);
        if (host && host.get('publicHostName')) {
          hostName = host.get('publicHostName');
        }
        if (hc.metrics && hc.metrics.flume && hc.metrics.flume.flume && hc.metrics.flume.flume.CHANNEL) {
          for ( var cname in hc.metrics.flume.flume.CHANNEL) {
            var seriesName = Em.I18n.t('services.service.info.metrics.flume.channelName').format(cname + " (" + hostName + ")");
            var seriesData = hc.metrics.flume.flume.CHANNEL[cname]['ChannelSize'];
            if (seriesData) {
              seriesArray.push(self.transformData(seriesData, seriesName));
            }
          }
        }
      });
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/flume/channel_size_mma", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_ChannelSizeMMA = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-channel-size-mma",
  title: Em.I18n.t('services.service.info.metrics.flume.channelSizeMMA'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.flume.channel_size_for_all',
  yAxisFormatter: App.ChartLinearTimeView.CreateRateFormatter('',
    App.ChartLinearTimeView.DefaultFormatter),

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;

    if (Em.get(jsonData, "metrics.flume.flume.CHANNEL.ChannelSize.rate")) {
      for ( var cname in jsonData.metrics.flume.flume.CHANNEL.ChannelSize.rate) {
        if(cname != "sum"){
          var seriesName = Em.I18n.t('services.service.info.metrics.flume.channelType').format(cname);
          var seriesData = jsonData.metrics.flume.flume.CHANNEL.ChannelSize.rate[cname];
          if (seriesData) {
            seriesArray.push(self.transformData(seriesData, seriesName));
          }
        }
      }
    }
    return seriesArray;
  },

  colorForSeries: function (series) {
    if (Em.I18n.t('services.service.info.metrics.flume.channelType').format("avg") == series.name){
      return '#0066b3';
    }else if (Em.I18n.t('services.service.info.metrics.flume.channelType').format("min") == series.name){
      return '#00CC00';
    }else if (Em.I18n.t('services.service.info.metrics.flume.channelType').format("max") == series.name){
      return '#FF8000';
    }
    return null;
  }
});

});

require.register("views/main/service/info/metrics/flume/channel_sum", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 *
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_ChannelSizeSum = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-channel-size-sum",
  title: Em.I18n.t('services.service.info.metrics.flume.channelSizeSum'),
  yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,

  ajaxIndex: 'service.metrics.flume.channel_size_for_all',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;
    if(Em.get(jsonData, "metrics.flume.flume.CHANNEL.ChannelSize.rate.sum")){
      var seriesName = Em.I18n.t('services.service.info.metrics.flume.channelSizeSum');
      var seriesData = jsonData.metrics.flume.flume.CHANNEL.ChannelSize.rate.sum;
      if (seriesData) {
        seriesArray.push(self.transformData(seriesData, seriesName));
      }
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/flume/cpu_user", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_CPUUser = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-cpu-user",
  title: Em.I18n.t('services.service.info.metrics.flume.cpu.user'),
  yAxisFormatter: App.ChartLinearTimeView.PercentageFormatter,
  renderer: 'line',

  ajaxIndex: 'service.metrics.flume.cpu_user',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;
    if (jsonData && jsonData.host_components) {
      jsonData.host_components.forEach(function (hc) {
        var hostName = hc.HostRoles.host_name;
        var host = App.Host.find(hostName);
        if (host && host.get('publicHostName')) {
          hostName = host.get('publicHostName');
        }
        if (hc.metrics && hc.metrics.cpu && hc.metrics.cpu.cpu_user) {
          var seriesName = Em.I18n.t('services.service.info.metrics.flume.hostName').format(hostName);
          var seriesData = hc.metrics.cpu.cpu_user;
          if (seriesData) {
            seriesArray.push(self.transformData(seriesData, seriesName));
          }
        }
      });
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/flume/flume_incoming_mma", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_IncommingMMA = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-incoming_mma",
  renderer: 'line',
  title: Em.I18n.t('services.service.info.metrics.flume.incoming.mma'),

  ajaxIndex: 'service.metrics.flume.incoming_event_put_successCount',
  yAxisFormatter: App.ChartLinearTimeView.CreateRateFormatter('',
      App.ChartLinearTimeView.DefaultFormatter),

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;

    if (Em.get(jsonData, "metrics.flume.flume.CHANNEL.EventPutSuccessCount.rate")) {
      for ( var cname in jsonData.metrics.flume.flume.CHANNEL.EventPutSuccessCount.rate) {
        if(cname != "sum"){
          var seriesName = Em.I18n.t('services.service.info.metrics.flume.incoming_mma').format(cname);
          var seriesData = jsonData.metrics.flume.flume.CHANNEL.EventPutSuccessCount.rate[cname];
          if (seriesData) {
            seriesArray.push(self.transformData(seriesData, seriesName));
          }
        }
      }
    }
    return seriesArray;
  },

  colorForSeries: function (series) {
    var name = Em.I18n.t('services.service.info.metrics.flume.incoming_mma');
    if (name.format("avg") == series.name){
      return '#0066b3';
    }else if (name.format("min") == series.name){
      return '#00CC00';
    }else if (name.format("max") == series.name){
      return '#FF8000';
    }
    return null;
  }
});

});

require.register("views/main/service/info/metrics/flume/flume_incoming_sum", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_IncommingSum = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-incoming_sum",
  title: Em.I18n.t('services.service.info.metrics.flume.incoming.sum'),

  ajaxIndex: 'service.metrics.flume.incoming_event_put_successCount',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;
    if(Em.get(jsonData, "metrics.flume.flume.CHANNEL.EventPutSuccessCount.rate.sum")){
      var seriesName = Em.I18n.t('services.service.info.metrics.flume.incoming.sum');
      var seriesData = jsonData.metrics.flume.flume.CHANNEL.EventPutSuccessCount.rate.sum;
      if (seriesData) {
        seriesArray.push(self.transformData(seriesData, seriesName));
      }
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/flume/flume_metric_graph", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing individual graphs of Flume.
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceFlumeMetricGraph = App.ChartLinearTimeView.extend({
  /**
   * One of 'SOURCE', 'SINK' or 'CHANNEL'.
   */
  metricType: null,
  metricName: null,
  hostName: null,
  metricItems: null,

  id: function(){
    return "service-metrics-flume-metric-graph-" + this.get('metricType') + '-' + this.get('metricName');
  }.property('metricType', 'metricName'),

  title: function(){
    return this.get('metricName');
  }.property('metricName'),

  ajaxIndex: 'host.host_component.flume.metrics.timeseries',

  getDataForAjaxRequest: function() {
    var data = this._super();

    var urlFields = '';
    this.get('metricItems').forEach(function (metricItem, index) {
      urlFields += index === 0 ? '' : ',';
      urlFields += 'metrics/flume/flume/' + this.get('metricType') + '/' + metricItem + '/' + this.get('metricName') +
          '[' + data.fromSeconds + ',' + data.toSeconds + ',' + data.stepSeconds + ']'
    }, this);

    data.url = App.get('apiPrefix') + '/clusters/' + App.get('clusterName') + '/hosts/' + this.get('hostName') + '/host_components/FLUME_HANDLER?fields=' + urlFields;
    return data;
  },

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;
    if (jsonData && jsonData.metrics && jsonData.metrics.flume && jsonData.metrics.flume.flume &&
        jsonData.metrics.flume.flume[this.get('metricType')]) {
      var metricTypeData = jsonData.metrics.flume.flume[this.get('metricType')];
      for ( var componentName in metricTypeData) {
        var seriesName = componentName;
        var seriesData = metricTypeData[componentName][this.get('metricName')];
        if (seriesData) {
          seriesArray.push(self.transformData(seriesData, seriesName));
        }
      }
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/flume/flume_metric_graphs", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
require('views/main/service/service');

App.MainServiceInfoFlumeGraphsView = App.MainServiceInfoSummaryMetricGraphsView.extend({

  serviceMetricGraphs: [],

  loadMetrics: function () {
    var graphRows = [];
    var viewData = this.get('viewData');
    if (viewData != null) {
      var metricType = viewData.metricType;
      var hostName = viewData.agent.get('hostName');
      App.ajax.send({
        'name': 'host.host_component.flume.metrics',
        'sender': this,
        'success': 'onLoadMetricsSuccess',
        'data': {
          hostName: hostName,
          flumeComponent: metricType
        }
      });
    }
    return graphRows;
  }.observes('viewData', 'metricType'),

  onLoadMetricsSuccess: function (data) {
    var graphRows = [];
    var viewData = this.get('viewData');
    var metricType = viewData.metricType;
    var hostName = viewData.agent.get('hostName');
    var metricNames = {};
    var metricItems = [];
    if (data != null && data.metrics != null && data.metrics.flume != null && data.metrics.flume.flume != null && data.metrics.flume.flume[metricType] != null) {
      for (var name in data.metrics.flume.flume[metricType]) {
        for (var metricName in data.metrics.flume.flume[metricType][name]) {
          metricNames[metricName] = name;
        }
        metricItems.push(name);
      }
    }
    // Now that we have collected all metric names, we create
    // views for each of them and store them 4 in a row.
    graphRows.push([]);
    var graphs = graphRows[0];
    for (var metricName in metricNames) {
      if (graphs.length > 1) {
        graphRows.push([]);
        graphs = graphRows[graphRows.length - 1];
      }
      graphs.push(App.ChartServiceFlumeMetricGraph.extend({
        metricType: metricType,
        metricName: metricName,
        hostName: hostName,
        metricItems: metricItems
      }));
    }
    this.set('serviceMetricGraphs', graphRows);
  },

  didInsertElement: function () {
    this.loadMetrics();
  }

});

});

require.register("views/main/service/info/metrics/flume/flume_outgoing_mma", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_OutgoingMMA = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-outgoing_mma",
  renderer: 'line',
  title: Em.I18n.t('services.service.info.metrics.flume.outgoing.mma'),

  ajaxIndex: 'service.metrics.flume.outgoing_event_take_success_count',
  yAxisFormatter: App.ChartLinearTimeView.CreateRateFormatter('',
      App.ChartLinearTimeView.DefaultFormatter),

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;

    if (Em.get(jsonData, "metrics.flume.flume.CHANNEL.EventTakeSuccessCount.rate")) {
      for ( var cname in jsonData.metrics.flume.flume.CHANNEL.EventTakeSuccessCount.rate) {
        if(cname != "sum"){
          var seriesName = Em.I18n.t('services.service.info.metrics.flume.outgoing_mma').format(cname);
          var seriesData = jsonData.metrics.flume.flume.CHANNEL.EventTakeSuccessCount.rate[cname];
          if (seriesData) {
            seriesArray.push(self.transformData(seriesData, seriesName));
          }
        }
      }
    }
    return seriesArray;
  },

  colorForSeries: function (series) {
    var name = Em.I18n.t('services.service.info.metrics.flume.outgoing_mma');
    if (name.format("avg") == series.name){
      return '#0066b3';
    }else if (name.format("min") == series.name){
      return '#00CC00';
    }else if (name.format("max") == series.name){
      return '#FF8000';
    }
    return null;
  }
});

});

require.register("views/main/service/info/metrics/flume/flume_outgoing_sum", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_OutgoingSum = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-outgoing_sum",
  title: Em.I18n.t('services.service.info.metrics.flume.outgoing.sum'),

  ajaxIndex: 'service.metrics.flume.outgoing_event_take_success_count',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;
    if(Em.get(jsonData, "metrics.flume.flume.CHANNEL.EventTakeSuccessCount.rate.sum")){
      var seriesName = Em.I18n.t('services.service.info.metrics.flume.outgoing.sum');
      var seriesData = jsonData.metrics.flume.flume.CHANNEL.EventTakeSuccessCount.rate.sum;
      if (seriesData) {
        seriesArray.push(self.transformData(seriesData, seriesName));
      }
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/flume/gc", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_GarbageCollection = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-jvm-gc",
  title: Em.I18n.t('services.service.info.metrics.flume.gc'),
  yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

  ajaxIndex: 'service.metrics.flume.gc',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;
    if (jsonData && jsonData.host_components) {
      jsonData.host_components.forEach(function (hc) {
        var hostName = hc.HostRoles.host_name;
        var host = App.Host.find(hostName);
        if (host && host.get('publicHostName')) {
          hostName = host.get('publicHostName');
        }
        if (hc.metrics && hc.metrics.jvm && hc.metrics.jvm.gcTimeMillis) {
          var seriesName = Em.I18n.t('services.service.info.metrics.flume.hostName').format(hostName);
          var seriesData = hc.metrics.jvm.gcTimeMillis;
          if (seriesData) {
            seriesArray.push(self.transformData(seriesData, seriesName));
          }
        }
      });
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/flume/jvm_heap", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_JVMHeapUsed = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-jvm-heap-used",
  title: Em.I18n.t('services.service.info.metrics.flume.jvmHeapUsed'),
  yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
  renderer: 'line',

  ajaxIndex: 'service.metrics.flume.jvm_heap_used',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;
    var MB = Math.pow(2, 20);
    if (jsonData && jsonData.host_components) {
      jsonData.host_components.forEach(function (hc) {
        var hostName = hc.HostRoles.host_name;
        var host = App.Host.find(hostName);
        if (host && host.get('publicHostName')) {
          hostName = host.get('publicHostName');
        }
        if (hc.metrics && hc.metrics.jvm && hc.metrics.jvm.memHeapUsedM) {
          var seriesName = Em.I18n.t('services.service.info.metrics.flume.hostName').format(hostName);
          var seriesData = hc.metrics.jvm.memHeapUsedM;
          if (seriesData) {
            var s = self.transformData(seriesData, seriesName);
            for (var i = 0; i < s.data.length; i++) {
              s.data[i].y *= MB;
            }
            seriesArray.push(s);
          }
        }
      });
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/flume/jvm_threads_runnable", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_JVMThreadsRunnable = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-jvm-threads-runnable",
  title: Em.I18n.t('services.service.info.metrics.flume.jvmThreadsRunnable'),
  renderer: 'line',

  ajaxIndex: 'service.metrics.flume.jvm_threads_runnable',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;
    var MB = Math.pow(2, 20);
    if (jsonData && jsonData.host_components) {
      jsonData.host_components.forEach(function (hc) {
        var hostName = hc.HostRoles.host_name;
        var host = App.Host.find(hostName);
        if (host && host.get('publicHostName')) {
          hostName = host.get('publicHostName');
        }
        if (hc.metrics && hc.metrics.jvm && hc.metrics.jvm.threadsRunnable) {
          var seriesName = Em.I18n.t('services.service.info.metrics.flume.hostName').format(hostName);
          var seriesData = hc.metrics.jvm.threadsRunnable;
          if (seriesData) {
            var s = self.transformData(seriesData, seriesName);
            seriesArray.push(s);
          }
        }
      });
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/flume/sink_connection_failed", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_SinkConnectionFailedCount = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-sink-connection-failed",
  title: Em.I18n.t('services.service.info.metrics.flume.sinkConnectionFailed'),

  ajaxIndex: 'service.metrics.flume.sink_connection_failed',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;
    if (jsonData && jsonData.host_components) {
      jsonData.host_components.forEach(function (hc) {
        var hostName = hc.HostRoles.host_name;
        var host = App.Host.find(hostName);
        if (host && host.get('publicHostName')) {
          hostName = host.get('publicHostName');
        }
        if (hc.metrics && hc.metrics.flume && hc.metrics.flume.flume && hc.metrics.flume.flume.SINK) {
          for ( var cname in hc.metrics.flume.flume.SINK) {
            var seriesName = Em.I18n.t('services.service.info.metrics.flume.sinkName').format(cname + " (" + hostName + ")");
            var seriesData = hc.metrics.flume.flume.SINK[cname]['ConnectionFailedCount'];
            if (seriesData) {
              seriesArray.push(self.transformData(seriesData, seriesName));
            }
          }
        }
      });
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/flume/sink_drain_success", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_SinkDrainSuccessCount = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-sink-drain-success",
  title: Em.I18n.t('services.service.info.metrics.flume.sinkDrainSuccess'),

  ajaxIndex: 'service.metrics.flume.sink_drain_success',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;
    if (jsonData && jsonData.host_components) {
      jsonData.host_components.forEach(function (hc) {
        var hostName = hc.HostRoles.host_name;
        var host = App.Host.find(hostName);
        if (host && host.get('publicHostName')) {
          hostName = host.get('publicHostName');
        }
        if (hc.metrics && hc.metrics.flume && hc.metrics.flume.flume && hc.metrics.flume.flume.SINK) {
          for ( var cname in hc.metrics.flume.flume.SINK) {
            var seriesName = Em.I18n.t('services.service.info.metrics.flume.sinkName').format(cname + " (" + hostName + ")");
            var seriesData = hc.metrics.flume.flume.SINK[cname]['EventDrainSuccessCount'];
            if (seriesData) {
              seriesArray.push(self.transformData(seriesData, seriesName));
            }
          }
        }
      });
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/flume/source_accepted", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsFlume_SourceAcceptedCount = App.ChartLinearTimeView.extend({
  id: "service-metrics-flume-source-accepted",
  title: Em.I18n.t('services.service.info.metrics.flume.sourceAccepted'),

  ajaxIndex: 'service.metrics.flume.source_accepted',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var self = this;
    if (jsonData && jsonData.host_components) {
      jsonData.host_components.forEach(function (hc) {
        var hostName = hc.HostRoles.host_name;
        var host = App.Host.find(hostName);
        if (host && host.get('publicHostName')) {
          hostName = host.get('publicHostName');
        }
        if (hc.metrics && hc.metrics.flume && hc.metrics.flume.flume && hc.metrics.flume.flume.SOURCE) {
          for ( var cname in hc.metrics.flume.flume.SOURCE) {
            var seriesName = Em.I18n.t('services.service.info.metrics.flume.sourceName').format(cname + " (" + hostName + ")");
            var seriesData = hc.metrics.flume.flume.SOURCE[cname]['EventAcceptedCount'];
            if (seriesData) {
              seriesArray.push(self.transformData(seriesData, seriesName));
            }
          }
        }
      });
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/hbase/cluster_requests", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHBASE_ClusterRequests = App.ChartLinearTimeView.extend({
  id: "service-metrics-hbase-cluster-requests",
  title: Em.I18n.t('services.service.info.metrics.hbase.clusterRequests'),

  ajaxIndex: 'service.metrics.hbase.cluster_requests',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.hbase && jsonData.metrics.hbase.master) {
      for ( var name in jsonData.metrics.hbase.master) {
        var displayName;
        var seriesData = jsonData.metrics.hbase.master[name];
        switch (name) {
          case "cluster_requests":
            displayName = Em.I18n.t('services.service.info.metrics.hbase.clusterRequests.displayNames.requestCount');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/hbase/hlog_split_size", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHBASE_HlogSplitSize = App.ChartLinearTimeView.extend({
  id: "service-metrics-hbase-hlog-split-size",
  title: Em.I18n.t('services.service.info.metrics.hbase.hlogSplitSize'),
  yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,

  ajaxIndex: 'service.metrics.hbase.hlog_split_size',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.hbase && jsonData.metrics.hbase.master) {
      for ( var name in jsonData.metrics.hbase.master) {
        var displayName;
        var seriesData = jsonData.metrics.hbase.master[name];
        switch (name) {
          case "splitSize_avg_time":
            displayName = Em.I18n.t('services.service.info.metrics.hbase.hlogSplitSize.displayNames.splitSize');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/hbase/hlog_split_time", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHBASE_HlogSplitTime = App.ChartLinearTimeView.extend({
  id: "service-metrics-hbase-hlog-split-time",
  title: Em.I18n.t('services.service.info.metrics.hbase.hlogSplitTime'),
  yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

  ajaxIndex: 'service.metrics.hbase.hlog_split_time',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.hbase && jsonData.metrics.hbase.master) {
      for ( var name in jsonData.metrics.hbase.master) {
        var displayName;
        var seriesData = jsonData.metrics.hbase.master[name];
        switch (name) {
          case "splitTime_avg_time":
            displayName = Em.I18n.t('services.service.info.metrics.hbase.hlogSplitTime.displayNames.splitTime');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/hbase/regionserver_queuesize", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHBASE_RegionServerQueueSize = App.ChartLinearTimeView.extend({
  id: "service-metrics-hbase-regionserver-queuesize",
  title: Em.I18n.t('services.service.info.metrics.hbase.regionServerQueueSize'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.hbase.regionserver_queuesize',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.hbase && jsonData.metrics.hbase.regionserver) {
      for ( var name in jsonData.metrics.hbase.regionserver) {
        var displayName;
        var seriesData = jsonData.metrics.hbase.regionserver[name];
        switch (name) {
          case "compactionQueueSize":
            displayName = Em.I18n.t('services.service.info.metrics.hbase.regionServerQueueSize.displayNames.compactionQueueSize');
            break;
          case "flushQueueSize":
            displayName = Em.I18n.t('services.service.info.metrics.hbase.regionServerQueueSize.displayNames.flushQueueSize');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/hbase/regionserver_regions", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHBASE_RegionServerRegions = App.ChartLinearTimeView.extend({
  id: "service-metrics-hbase-regionserver-regions",
  title: Em.I18n.t('services.service.info.metrics.hbase.regionServerRegions'),

  ajaxIndex: 'service.metrics.hbase.regionserver_regions',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.hbase && jsonData.metrics.hbase.regionserver) {
      for ( var name in jsonData.metrics.hbase.regionserver) {
        var displayName;
        var seriesData = jsonData.metrics.hbase.regionserver[name];
        switch (name) {
          case "regions":
            displayName = Em.I18n.t('services.service.info.metrics.hbase.regionServerRegions.displayNames.regions');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/hbase/regionserver_rw_requests", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing HBase Cluster Requests
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHBASE_RegionServerReadWriteRequests = App.ChartLinearTimeView.extend({
  id: "service-metrics-hbase-regionserver-rw-requests",
  renderer: 'line',
  ajaxIndex: 'service.metrics.hbase.regionserver_rw_requests',
  yAxisFormatter: App.ChartLinearTimeView.CreateRateFormatter('req',
      App.ChartLinearTimeView.DefaultFormatter),
  title: Em.I18n.t('services.service.info.metrics.hbase.regionServerRequests.2'),

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.hbase && jsonData.metrics.hbase.regionserver) {
      for ( var name in jsonData.metrics.hbase.regionserver) {
        var displayName;
        var seriesData = jsonData.metrics.hbase.regionserver[name];
        switch (name) {
          case "writeRequestsCount":
            displayName = Em.I18n.t('services.service.info.metrics.hbase.regionServerRequests.displayNames.writeRequests');
            break;
          case "readRequestsCount":
            displayName = Em.I18n.t('services.service.info.metrics.hbase.regionServerRequests.displayNames.readRequests');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/hdfs/block_status", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHDFS_BlockStatus = App.ChartLinearTimeView.extend({
  id: "service-metrics-hdfs-block-status",
  title: Em.I18n.t('services.service.info.metrics.hdfs.blockStatus'),
  renderer: 'line',

  ajaxIndex: 'service.metrics.hdfs.block_status',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.dfs && jsonData.metrics.dfs.FSNamesystem) {
      for ( var name in jsonData.metrics.dfs.FSNamesystem) {
        var displayName;
        var seriesData = jsonData.metrics.dfs.FSNamesystem[name];
        switch (name) {
          case "PendingReplicationBlocks":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.blockStatus.displayNames.pendingReplicationBlocks');
            break;
          case "UnderReplicatedBlocks":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.blockStatus.displayNames.underReplicatedBlocks');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/hdfs/file_operations", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHDFS_FileOperations = App.ChartLinearTimeView.extend({
  id: "service-metrics-hdfs-file-operations",
  title: Em.I18n.t('services.service.info.metrics.hdfs.fileOperations'),
  renderer: 'line',

  ajaxIndex: 'service.metrics.hdfs.file_operations',
  yAxisFormatter: App.ChartLinearTimeView.CreateRateFormatter('ops',
      App.ChartLinearTimeView.DefaultFormatter),

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.dfs && jsonData.metrics.dfs.namenode) {
      for ( var name in jsonData.metrics.dfs.namenode) {
        var displayName;
        var seriesData = jsonData.metrics.dfs.namenode[name];
        switch (name) {
          case "FileInfoOps":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.fileOperations.displayNames.fileInformationOperations');
            break;
          case "DeleteFileOps":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.fileOperations.displayNames.deleteFileOperations');
            break;
          case "CreateFileOps":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.fileOperations.displayNames.createFileOperations');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/hdfs/gc", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHDFS_GC = App.ChartLinearTimeView.extend({
  id: "service-metrics-hdfs-gc",
  title: Em.I18n.t('services.service.info.metrics.hdfs.gc'),
  yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

  ajaxIndex: 'service.metrics.hdfs.gc',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
      for ( var name in jsonData.metrics.jvm) {
        var displayName;
        var seriesData = jsonData.metrics.jvm[name];
        switch (name) {
          case "gcTimeMillis":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.gc.displayNames.gcTimeMillis');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/hdfs/io", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHDFS_IO = App.ChartLinearTimeView.extend({
  id: "service-metrics-hdfs-io",
  title: Em.I18n.t('services.service.info.metrics.hdfs.io'),
  yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
  renderer: 'line',

  ajaxIndex: 'service.metrics.hdfs.io',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.dfs && jsonData.metrics.dfs.datanode) {
      for ( var name in jsonData.metrics.dfs.datanode) {
        var displayName;
        var seriesData = jsonData.metrics.dfs.datanode[name];
        switch (name) {
          case "bytes_written":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.io.displayNames.bytesWritten');
            break;
          case "bytes_read":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.io.displayNames.bytesRead');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/hdfs/jvm_heap", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHDFS_JVMHeap = App.ChartLinearTimeView.extend({
  id: "service-metrics-hdfs-jvm-heap",
  title: Em.I18n.t('services.service.info.metrics.hdfs.jvmHeap'),
  yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
  renderer: 'line',

  ajaxIndex: 'service.metrics.hdfs.jvm_heap',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var MB = Math.pow(2, 20);
    if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
      for ( var name in jsonData.metrics.jvm) {
        var displayName;
        var seriesData = jsonData.metrics.jvm[name];
        switch (name) {
          case "memHeapCommittedM":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmHeap.displayNames.memHeapCommittedM');
            break;
          case "memNonHeapUsedM":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmHeap.displayNames.memNonHeapUsedM');
            break;
          case "memHeapUsedM":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmHeap.displayNames.memHeapUsedM');
            break;
          case "memNonHeapCommittedM":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmHeap.displayNames.memNonHeapCommittedM');
            break;
          default:
            break;
        }
        if (seriesData) {
          var s = this.transformData(seriesData, displayName);
          for (var i = 0; i < s.data.length; i++) {
            s.data[i].y *= MB;
          }
          seriesArray.push(s);
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/hdfs/jvm_threads", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHDFS_JVMThreads = App.ChartLinearTimeView.extend({
  id: "service-metrics-hdfs-jvm-threads",
  title: Em.I18n.t('services.service.info.metrics.hdfs.jvmThreads'),
  renderer: 'line',

  ajaxIndex: 'service.metrics.hdfs.jvm_threads',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
      for ( var name in jsonData.metrics.jvm) {
        var displayName;
        var seriesData = jsonData.metrics.jvm[name];
        switch (name) {
          case "threadsBlocked":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsBlocked');
            break;
          case "threadsWaiting":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsWaiting');
            break;
          case "threadsTimedWaiting":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsTimedWaiting');
            break;
          case "threadsRunnable":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.jvmThreads.displayNames.threadsRunnable');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/hdfs/rpc", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHDFS_RPC = App.ChartLinearTimeView.extend({
  id: "service-metrics-hdfs-rpc",
  title: Em.I18n.t('services.service.info.metrics.hdfs.rpc'),
  yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

  ajaxIndex: 'service.metrics.hdfs.rpc',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.rpc) {
      for ( var name in jsonData.metrics.rpc) {
        var displayName;
        var seriesData = jsonData.metrics.rpc[name];
        switch (name) {
          case "RpcQueueTime_avg_time":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.rpc.displayNames.rpcQueueTimeAvgTime');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/hdfs/space_utilization", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsHDFS_SpaceUtilization = App.ChartLinearTimeView.extend({
  id: "service-metrics-hdfs-space-utilization",
  title: Em.I18n.t('services.service.info.metrics.hdfs.spaceUtilization'),
  yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
  renderer: 'line',
  ajaxIndex: 'service.metrics.hdfs.space_utilization',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.dfs && jsonData.metrics.dfs.FSNamesystem) {
      for ( var name in jsonData.metrics.dfs.FSNamesystem) {
        var displayName;
        var seriesData = jsonData.metrics.dfs.FSNamesystem[name];
        switch (name) {
          case "CapacityRemaining":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityRemaining');
            break;
          case "CapacityUsed":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityUsed');
            break;
          case "CapacityTotal":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityTotal');
            break;
          case "CapacityNonDFSUsed":
            displayName = Em.I18n.t('services.service.info.metrics.hdfs.spaceUtilization.displayNames.capacityNonDFSUsed');
            break;
          default:
            break;
        }
        if (seriesData) {
          var s = this.transformData(seriesData, displayName);
          seriesArray.push(s);
        }
      }
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/kafka/broker_topic", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing Kafka_BrokerTopicMetrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsKafka_BrokerTopicMetrics = App.ChartLinearTimeView.extend({
  id: "service-metrics-kafka-broker-topic-metrics",
  title: Em.I18n.t('services.service.info.metrics.kafka.server.brokerTopic.title'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.kafka.broker.topic',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (Em.get(jsonData, 'metrics.kafka.server.BrokerTopicMetrics')) {
      for (var name in Em.get(jsonData, 'metrics.kafka.server.BrokerTopicMetrics')) {
        var displayName = null;
        var seriesData = Em.get(jsonData, 'metrics.kafka.server.BrokerTopicMetrics.' + name + '.1MinuteRate');
        switch (name) {
          case "AllTopicsBytesOutPerSec":
            displayName = Em.I18n.t('services.service.info.metrics.kafka.server.brokerTopic.displayNames.AllTopicsBytesOutPerSec');
            break;
          case "AllTopicsBytesInPerSec":
            displayName = Em.I18n.t('services.service.info.metrics.kafka.server.brokerTopic.displayNames.AllTopicsBytesInPerSec');
            break;
          case "AllTopicsMessagesInPerSec":
            displayName = Em.I18n.t('services.service.info.metrics.kafka.server.brokerTopic.displayNames.AllTopicsMessagesInPerSec');
            break;
          default:
            break;
        }
        if (seriesData != null && displayName) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/kafka/controller_status", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing Kafka_ControllerStatus
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsKafka_ControllerStatus = App.ChartLinearTimeView.extend({
  id: "service-metrics-kafka-controler-status-metrics",
  title: Em.I18n.t('services.service.info.metrics.kafka.controller.ControllerStats.title'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.kafka.controller.ControllerStats',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (Em.get(jsonData, 'metrics.kafka.controller.ControllerStats')) {
      for (var name in Em.get(jsonData, 'metrics.kafka.controller.ControllerStats')) {
        var displayName = null;
        var seriesData = Em.get(jsonData, 'metrics.kafka.controller.ControllerStats.' + name + '.1MinuteRate');
        switch (name) {
          case "LeaderElectionRateAndTimeMs":
            displayName = Em.I18n.t('services.service.info.metrics.kafka.controller.ControllerStats.displayNames.LeaderElectionRateAndTimeMs');
            break;
          case "UncleanLeaderElectionsPerSec":
            displayName = Em.I18n.t('services.service.info.metrics.kafka.controller.ControllerStats.displayNames.UncleanLeaderElectionsPerSec');
            break;
          default:
            break;
        }
        if (seriesData != null && displayName) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/kafka/kafka_controller", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing Kafka_Controller
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsKafka_Controller = App.ChartLinearTimeView.extend({
  id: "service-metrics-kafka-controller-metrics",
  title: Em.I18n.t('services.service.info.metrics.kafka.controller.KafkaController.title'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.kafka.controller.KafkaController',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var seriesData = Em.get(jsonData, 'metrics.kafka.controller.KafkaController.ActiveControllerCount');
    if (seriesData != null) {
      var displayName = Em.I18n.t('services.service.info.metrics.kafka.controller.KafkaController.displayNames.ActiveControllerCount');
      seriesArray.push(this.transformData(seriesData, displayName));
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/kafka/kafka_log_flush", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing Kafka_LogFlush
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsKafka_LogFlush = App.ChartLinearTimeView.extend({
  id: "service-metrics-kafka-log-metrics",
  title: Em.I18n.t('services.service.info.metrics.kafka.log.LogFlushStats.title'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.kafka.log.LogFlushStats',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var seriesData = Em.get(jsonData, 'metrics.kafka.log.LogFlushStats.LogFlushRateAndTimeMs.1MinuteRate');
    if (seriesData != null) {
      var displayName = Em.I18n.t('services.service.info.metrics.kafka.log.LogFlushStats.displayNames.LogFlushRateAndTimeMs');
      seriesArray.push(this.transformData(seriesData, displayName));
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/kafka/replica_fetcher", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing Kafka_ReplicaFetcher
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsKafka_ReplicaFetcher = App.ChartLinearTimeView.extend({
  id: "service-metrics-kafka-replica-fetcher-metrics",
  title: Em.I18n.t('services.service.info.metrics.kafka.server.ReplicaFetcherManager.title'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.kafka.server.ReplicaFetcherManager',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var seriesData = Em.get(jsonData, 'metrics.kafka.server.ReplicaFetcherManager.Replica-MaxLag');
    if (seriesData != null) {
      var displayName = Em.I18n.t('services.service.info.metrics.kafka.server.ReplicaFetcherManager.displayNames.Replica-MaxLag');
      seriesArray.push(this.transformData(seriesData, displayName));
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/kafka/replica_manager", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing Kafka_BrokerTopicMetrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsKafka_ReplicaManager = App.ChartLinearTimeView.extend({
  id: "service-metrics-kafka-replica-manager-metrics",
  title: Em.I18n.t('services.service.info.metrics.kafka.server.ReplicaManager.title'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.kafka.server.ReplicaManager',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (Em.get(jsonData, 'metrics.kafka.server.ReplicaManager')) {
      for (var name in Em.get(jsonData, 'metrics.kafka.server.ReplicaManager')) {
        var displayName = null;
        var seriesData = Em.get(jsonData, 'metrics.kafka.server.ReplicaManager.' + name);
        switch (name) {
          case "LeaderCount":
            displayName = Em.I18n.t('services.service.info.metrics.kafka.server.ReplicaManager.displayNames.LeaderCount');
            break;
          case "UnderReplicatedPartitions":
            displayName = Em.I18n.t('services.service.info.metrics.kafka.server.ReplicaManager.displayNames.UnderReplicatedPartitions');
            break;
          case "PartitionCount":
            displayName = Em.I18n.t('services.service.info.metrics.kafka.server.ReplicaManager.displayNames.PartitionCount');
            break;
          default:
            break;
        }
        if (seriesData != null && displayName) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/storm/executors_metric", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
require('mixins/common/chart/storm_linear_time');

App.ChartServiceMetricsSTORM_Executors = App.ChartLinearTimeView.extend(App.StormLinearTimeChartMixin, {
  id: "service-metrics-storm-executors",
  title: Em.I18n.t('services.storm.executors.metrics.title'),
  renderer: 'line',
  yAxisFormatter: App.ChartLinearTimeView.DefaultFormatter,

  stormChartDefinition: [
    {
      name: Em.I18n.t('services.storm.executors.metrics.total'),
      field: 'totalexecutors'
    }
  ]

});

});

require.register("views/main/service/info/metrics/storm/slots_number_metric", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
require('mixins/common/chart/storm_linear_time');

App.ChartServiceMetricsSTORM_SlotsNumber = App.ChartLinearTimeView.extend(App.StormLinearTimeChartMixin, {
  id: "service-metrics-storm-supervisor-allocated",
  title: Em.I18n.t('services.storm.slots.metrics.title'),
  renderer: 'line',
  yAxisFormatter: App.ChartLinearTimeView.DefaultFormatter,

  stormChartDefinition: [
    {
      name: Em.I18n.t('services.storm.slots.metrics.total'),
      field: 'totalslots'
    },
    {
      name: Em.I18n.t('services.storm.slots.metrics.free'),
      field: 'freeslots'
    },
    {
      name: Em.I18n.t('services.storm.slots.metrics.used'),
      field: 'usedslots'
    }
  ]

});

});

require.register("views/main/service/info/metrics/storm/tasks_metric", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
require('mixins/common/chart/storm_linear_time');

App.ChartServiceMetricsSTORM_Tasks = App.ChartLinearTimeView.extend(App.StormLinearTimeChartMixin, {
  id: "service-metrics-storm-tasks",
  title: Em.I18n.t('services.storm.tasks.metrics.title'),
  renderer: 'line',
  yAxisFormatter: App.ChartLinearTimeView.DefaultFormatter,

  stormChartDefinition: [
    {
      name: Em.I18n.t('services.storm.tasks.metrics.total'),
      field: 'totaltasks'
    }
  ]

});

});

require.register("views/main/service/info/metrics/storm/topologies_metric", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
require('mixins/common/chart/storm_linear_time');

App.ChartServiceMetricsSTORM_Topologies = App.ChartLinearTimeView.extend(App.StormLinearTimeChartMixin, {
  id: "service-metrics-storm-topologies",
  title: Em.I18n.t('services.storm.topology.metrics.title'),
  renderer: 'line',
  yAxisFormatter: App.ChartLinearTimeView.DefaultFormatter,

  stormChartDefinition: [
    {
      name: Em.I18n.t('services.storm.topology.metrics.total'),
      field: 'topologies'
    }
  ]

});

});

require.register("views/main/service/info/metrics/yarn/allocated", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsYARN_AllocatedMemory = App.ChartLinearTimeView.extend({
  id: "service-metrics-yarn-queue-allocated",
  title: Em.I18n.t('services.service.info.metrics.yarn.allocated.memory'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.yarn.queue.allocated',
  yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var MB = Math.pow(2, 20);
    if (jsonData && jsonData.metrics && jsonData.metrics.yarn.Queue && jsonData.metrics.yarn.Queue.root) {
      for (var name in jsonData.metrics.yarn.Queue.root) {
        var displayName = null;
        var seriesData = jsonData.metrics.yarn.Queue.root[name];
        switch (name) {
          case "AvailableMB":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.allocated.memory.displayNames.available');
            break;
          case "PendingMB":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.allocated.memory.displayNames.pending');
            break;
          case "AllocatedMB":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.allocated.memory.displayNames.allocated');
            break;
          default:
            break;
        }
        if (seriesData && displayName) {
          var s = this.transformData(seriesData, displayName);
          for (var i = 0; i < s.data.length; i++) {
            s.data[i].y *= MB;
          }
          seriesArray.push(s);
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/yarn/allocated_container", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsYARN_AllocatedContainer = App.ChartLinearTimeView.extend({
  id: "service-metrics-yarn-queue-allocated-container",
  title: Em.I18n.t('services.service.info.metrics.yarn.allocated.container'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.yarn.queue.allocated.container',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.yarn.Queue && jsonData.metrics.yarn.Queue.root) {
      for (var name in jsonData.metrics.yarn.Queue.root) {
        var displayName = null;
        var seriesData = jsonData.metrics.yarn.Queue.root[name];
        switch (name) {
          case "ReservedContainers":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.allocated.container.displayNames.reserved');
            break;
          case "PendingContainers":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.allocated.container.displayNames.pending');
            break;
          case "AllocatedContainers":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.allocated.container.displayNames.allocated');
            break;
          default:
            break;
        }
        if (seriesData != null && displayName) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/yarn/apps_current_states", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsYARN_ApplicationCurrentStates = App.ChartLinearTimeView.extend({
  id: "service-metrics-yarn-apps-current-states",
  title: Em.I18n.t('services.service.info.metrics.yarn.apps.states.current.title'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.yarn.queue.apps.states.current',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.yarn.Queue && jsonData.metrics.yarn.Queue.root) {
      for (var name in jsonData.metrics.yarn.Queue.root) {
        var displayName = null;
        var seriesData = jsonData.metrics.yarn.Queue.root[name];
        switch (name) {
          case "AppsPending":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.apps.states.pending');
            break;
          case "AppsRunning":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.apps.states.running');
            break;
          default:
            break;
        }
        if (seriesData != null && displayName) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/yarn/apps_finished_states", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsYARN_ApplicationFinishedStates = App.ChartLinearTimeView.extend({
  id: "service-metrics-yarn-apps-finished-states",
  title: Em.I18n.t('services.service.info.metrics.yarn.apps.states.finished.title'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.yarn.queue.apps.states.finished',
  yAxisFormatter: App.ChartLinearTimeView.CreateRateFormatter('apps',
      App.ChartLinearTimeView.DefaultFormatter),

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.yarn.Queue && jsonData.metrics.yarn.Queue.root) {
      for (var name in jsonData.metrics.yarn.Queue.root) {
        var displayName = null;
        var seriesData = jsonData.metrics.yarn.Queue.root[name];
        switch (name) {
          case "AppsCompleted":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.apps.states.completed');
            break;
//          case "AppsFailed":
//            displayName = Em.I18n.t('services.service.info.metrics.yarn.apps.states.failed');
//            break;
          case "AppsKilled":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.apps.states.killed');
            break;
          case "AppsSubmitted":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.apps.states.submitted');
            break;
          default:
            break;
        }
        if (seriesData != null && displayName) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/yarn/gc", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsYARN_GC = App.ChartLinearTimeView.extend({
  id: "service-metrics-yarn-gc",
  title: Em.I18n.t('services.service.info.metrics.yarn.gc'),
  yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

  ajaxIndex: 'service.metrics.yarn.gc',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
      for ( var name in jsonData.metrics.jvm) {
        var displayName;
        var seriesData = jsonData.metrics.jvm[name];
        switch (name) {
          case "gcTimeMillis":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.gc.displayNames.gcTimeMillis');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/yarn/jvm_heap", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsYARN_JVMHeap = App.ChartLinearTimeView.extend({
  id: "service-metrics-yarn-jvm-heap",
  title: Em.I18n.t('services.service.info.metrics.yarn.jvmHeap'),
  yAxisFormatter: App.ChartLinearTimeView.BytesFormatter,
  renderer: 'line',

  ajaxIndex: 'service.metrics.yarn.jobs_heap',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    var MB = Math.pow(2, 20);
    if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
      for ( var name in jsonData.metrics.jvm) {
        var displayName;
        var seriesData = jsonData.metrics.jvm[name];
        switch (name) {
          case "memHeapCommittedM":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.jvmHeap.displayNames.memHeapCommittedM');
            break;
          case "memNonHeapUsedM":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.jvmHeap.displayNames.memNonHeapUsedM');
            break;
          case "memHeapUsedM":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.jvmHeap.displayNames.memHeapUsedM');
            break;
          case "memNonHeapCommittedM":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.jvmHeap.displayNames.memNonHeapCommittedM');
            break;
          default:
            break;
        }
        if (seriesData) {
          var s = this.transformData(seriesData, displayName);
          for (var i = 0; i < s.data.length; i++) {
            s.data[i].y *= MB;
          }
          seriesArray.push(s);
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/yarn/jvm_threads", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsYARN_JVMThreads = App.ChartLinearTimeView.extend({
  id: "service-metrics-yarn-jvm-threads",
  title: Em.I18n.t('services.service.info.metrics.yarn.jvmThreads'),
  renderer: 'line',

  ajaxIndex: 'service.metrics.yarn.jobs_threads',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.jvm) {
      for ( var name in jsonData.metrics.jvm) {
        var displayName;
        var seriesData = jsonData.metrics.jvm[name];
        switch (name) {
          case "threadsBlocked":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.jvmThreads.displayNames.threadsBlocked');
            break;
          case "threadsWaiting":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.jvmThreads.displayNames.threadsWaiting');
            break;
          case "threadsTimedWaiting":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.jvmThreads.displayNames.threadsTimedWaiting');
            break;
          case "threadsRunnable":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.jvmThreads.displayNames.threadsRunnable');
            break;
          default:
            break;
        }

        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/yarn/nms", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsYARN_NMS = App.ChartLinearTimeView.extend({
  id: "service-metrics-yarn-node-manager-statuses",
  title: Em.I18n.t('services.service.info.metrics.yarn.nodemanager.statuses'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.yarn.node.manager.statuses',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.yarn && jsonData.metrics.yarn.ClusterMetrics) {
      for (var name in jsonData.metrics.yarn.ClusterMetrics) {
        var displayName;
        var seriesData = jsonData.metrics.yarn.ClusterMetrics[name];
        switch (name) {
          case "NumActiveNMs":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.nodemanager.statuses.displayNames.active');
            break;
          case "NumDecommissionedNMs":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.nodemanager.statuses.displayNames.decommissioned');
            break;
          case "NumLostNMs":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.nodemanager.statuses.displayNames.lost');
            break;
          case "NumRebootedNMs":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.nodemanager.statuses.displayNames.rebooted');
            break;
          case "NumUnhealthyNMs":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.nodemanager.statuses.displayNames.unhealthy');
            break;
          default:
            break;
        }
        if (seriesData != null) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/metrics/yarn/qmr", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
var objUtils = require('utils/object_utils');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsYARN_QMR = App.ChartLinearTimeView.extend({
  id: "service-metrics-yarn-queue-memory-resource",
  title: Em.I18n.t('services.service.info.metrics.yarn.queueMemoryResource'),
  renderer: 'line',
  ajaxIndex: 'service.metrics.yarn.queue.memory.resource',
  yAxisFormatter: App.ChartLinearTimeView.PercentageFormatter,

  getDataForAjaxRequest: function () {
    var data = this._super();
    var svc = App.YARNService.find().objectAt(0);
    var queueNames = [];
    if (svc != null) {
      queueNames = svc.get('childQueueNames');
    }
    data.queueNames = queueNames;
    return data;
  },

  transformToSeries: function (jsonData) {
    var self = this;
    var seriesArray = [];
    var MB = Math.pow(2, 20);
    var svc = App.YARNService.find().objectAt(0);
    var queueNames = [];
    if (svc != null) {
      queueNames = svc.get('childQueueNames');
    }
    if (jsonData && jsonData.metrics && jsonData.metrics.yarn.Queue) {
      queueNames.forEach(function (qName) {
        var qPath = qName.replace(/\//g, '.');
        var displayName;
        var allocatedData = Em.get(jsonData.metrics.yarn.Queue, qPath + '.AllocatedMB');
        var availableData = Em.get(jsonData.metrics.yarn.Queue, qPath + '.AvailableMB');
        displayName = Em.I18n.t('services.service.info.metrics.yarn.queueMemoryResource.displayName').format(qName);
        var seriesData = null;
        if (allocatedData != null && availableData != null) {
          if (typeof allocatedData == "number" && typeof availableData == "number") {
            seriesData = (availableData == 0 && allocatedData == 0) ? 0 : (allocatedData * 100) / (allocatedData + availableData);
          } else if (allocatedData.length > 0 && availableData.length > 0) {
            seriesData = [];
            for ( var c = 0; c < Math.min(availableData.length, allocatedData.length); c++) {
              var allocDivAvail = (availableData[c][0] == 0 && allocatedData[c][0] == 0) ? 0 : (allocatedData[c][0] * 100) / (availableData[c][0] + allocatedData[c][0]);
              seriesData.push([allocDivAvail, allocatedData[c][1] ]);
            }
          } else {
            console.log("Skipping data series for Queue " + qName);
          }
        }
        if (seriesData != null) {
          seriesArray.push(self.transformData(seriesData, displayName));
        }
      });
    }
    return seriesArray;
  }
});

});

require.register("views/main/service/info/metrics/yarn/rpc", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

/**
 * @class
 *
 * This is a view for showing cluster CPU metrics
 *
 * @extends App.ChartLinearTimeView
 * @extends Ember.Object
 * @extends Ember.View
 */
App.ChartServiceMetricsYARN_RPC = App.ChartLinearTimeView.extend({
  id: "service-metrics-yarn-rpc",
  title: Em.I18n.t('services.service.info.metrics.yarn.rpc'),
  yAxisFormatter: App.ChartLinearTimeView.TimeElapsedFormatter,

  ajaxIndex: 'service.metrics.yarn.rpc',

  transformToSeries: function (jsonData) {
    var seriesArray = [];
    if (jsonData && jsonData.metrics && jsonData.metrics.rpc) {
      for ( var name in jsonData.metrics.rpc) {
        var displayName;
        var seriesData = jsonData.metrics.rpc[name];
        switch (name) {
          case "RpcQueueTime_avg_time":
            displayName = Em.I18n.t('services.service.info.metrics.yarn.rpc.displayNames.RpcQueueTimeAvgTime');
            break;
          default:
            break;
        }
        if (seriesData) {
          seriesArray.push(this.transformData(seriesData, displayName));
        }
      }
    }
    return seriesArray;
  }
});
});

require.register("views/main/service/info/summary", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
var batchUtils = require('utils/batch_scheduled_requests');
require('views/main/service/service');
require('data/service_graph_config');

App.MainServiceInfoSummaryView = Em.View.extend(App.UserPref, {
  templateName: require('templates/main/service/info/summary'),
  attributes:null,

  /**
   *  @property {String} templatePathPrefix - base path for custom templates
   *    if you want to add custom template, add <service_name>.hbs file to
   *    templates/main/service/info/summary folder.
   */
  templatePathPrefix: 'templates/main/service/info/summary/',
  /** @property {Ember.View} serviceSummaryView - view to embed, computed in
   *  <code>loadServiceSummary()</code>
   */
  serviceSummaryView: null,
  /**
   * @property {Object} serviceCustomViewsMap - custom views to embed
   *
   */
  serviceCustomViewsMap: function() {
    return {
      HBASE: App.MainDashboardServiceHbaseView,
      HDFS: App.MainDashboardServiceHdfsView,
      STORM: App.MainDashboardServiceStormView,
      YARN: App.MainDashboardServiceYARNView,
      RANGER: App.MainDashboardServiceRangerView,
      FLUME: Em.View.extend({
        template: Em.Handlebars.compile('' +
          '<tr>' +
            '<td>' +
              '{{view App.MainDashboardServiceFlumeView serviceBinding="view.service"}}' +
            '</td>' +
          '</tr>')
      })
    }
  }.property('serviceName'),
  /** @property collapsedMetrics {object[]} - metrics list for collapsed section
   *    structure of element from list:
   *      @property {string} header - title for section
   *      @property {string} id - id of section for toggling, like: metric1
   *      @property {string} toggleIndex - passed to `data-parent` attribute, like: #metric1
   *      @property {Em.View} metricView - metric view class
   */
  collapsedSections: null,

  servicesHaveClients: function() {
    return App.get('services.hasClient');
  }.property('App.services.hasClient'),

  hasManyServers: function () {
    return this.get('servers').length > 1;
  }.property('servers'),

  clientsHostText: function () {
    if (this.get('controller.content.installedClients').length == 0) {
      return '';
    } else if (this.get("hasManyClients")) {
      return Em.I18n.t('services.service.summary.viewHosts');
    } else {
      return Em.I18n.t('services.service.summary.viewHost');
    }
  }.property("hasManyClients"),

  hasManyClients: function () {
    return this.get('controller.content.installedClients').length > 1;
  }.property('service.installedClients'),

  servers: function () {
    var result = [];
    var service = this.get('controller.content');
    if (service.get("id") == "ZOOKEEPER" || service.get("id") == "FLUME") {
      var servers = service.get('hostComponents').filterProperty('isMaster');
      if (servers.length > 0) {
        result = [
          {
            'host': servers[0].get('displayName'),
            'isComma': false,
            'isAnd': false
          }
        ];
      }
      if (servers.length > 1) {
        result[0].isComma = true;
        result.push({
          'host': servers[1].get('displayName'),
          'isComma': false,
          'isAnd': false
        });
      }
      if (servers.length > 2) {
        result[1].isAnd = true;
        result.push({
          'host': Em.I18n.t('services.service.info.summary.serversHostCount').format(servers.length - 2),
          'isComma': false,
          'isAnd': false
        });
      }
    }
    return result;
  }.property('controller.content'),

  historyServerUI: function(){
    var service=this.get('controller.content');
    return (App.singleNodeInstall ? "http://" + App.singleNodeAlias + ":19888" : "http://" + service.get("hostComponents").findProperty('isMaster', true).get("host").get("publicHostName")+":19888");
  }.property('controller.content'),
  /**
   * Property related to ZOOKEEPER service, is unused for other services
   * @return {Object}
   */
  serversHost: function() {
    var service = this.get('controller.content');
    if (service.get("id") == "ZOOKEEPER" || service.get("id") == "FLUME") {
      var servers = service.get('hostComponents').filterProperty('isMaster');
      if (servers.length > 0) {
        return servers[0];
      }
    }
    return {};
  }.property('controller.content'),

  mastersObj: function() {
    return this.get('service.hostComponents').filterProperty('isMaster', true);
  }.property('service'),

  /**
   * Contain array with list of client components models <code>App.ClientComponent</code>.
   * @type {Array}
   */
  clientObj: function () {
    var clientComponents = this.get('controller.content.clientComponents').toArray();
    return clientComponents.get('length') ? clientComponents : [];
  }.property('service.clientComponents.@each.totalCount'),

  /**
   * Contain array with list of slave components models <code>App.SlaveComponent</code>.
   * @type {Array}
   */
  slavesObj: function() {
    var slaveComponents = this.get('controller.content.slaveComponents').toArray();
    return slaveComponents.get('length') ? slaveComponents : [];
  }.property('service.slaveComponents.@each.totalCount', 'service.slaveComponents.@each.startedCount'),

  data:{
    hive:{
      "database":"PostgreSQL",
      "databaseName":"hive",
      "user":"hive"
    }
  },

  /**
   * Wrapper for displayName. used to render correct display name for mysql_server
   */
  componentNameView: Ember.View.extend({
    template: Ember.Handlebars.compile('{{view.displayName}}'),
    comp : null,
    displayName: function(){
      if(this.get('comp.componentName') == 'MYSQL_SERVER'){
        return this.t('services.hive.databaseComponent');
      }
      return this.get('comp.displayName');
    }.property('comp')
  }),

  service:function () {
    var svc = this.get('controller.content');
    var svcName = svc.get('serviceName');
    if (svcName) {
      switch (svcName.toLowerCase()) {
        case 'hdfs':
          svc = App.HDFSService.find().objectAt(0);
          break;
        case 'yarn':
          svc = App.YARNService.find().objectAt(0);
          break;
        case 'hbase':
          svc = App.HBaseService.find().objectAt(0);
          break;
        case 'flume':
          svc = App.FlumeService.find().objectAt(0);
          break;
        case 'storm':
          svc = App.StormService.find().objectAt(0);
          break;
        default:
          break;
      }
    }
    return svc;
  }.property('controller.content.serviceName').volatile(),

  isHide:true,
  moreStatsView:Em.View.extend({
    tagName:"a",
    template:Ember.Handlebars.compile('{{t services.service.summary.moreStats}}'),
    attributeBindings:[ 'href' ],
    classNames:[ 'more-stats' ],
    click:function (event) {
      this._parentView._parentView.set('isHide', false);
      this.remove();
    },
    href:'javascript:void(null)'
  }),

  serviceName:function () {
    return this.get('service.serviceName');
  }.property('service'),

  oldServiceName:'',

  /*
   * 'Restart Required bar' start
   */
  componentsCount: null,
  hostsCount: null,

  alertsCount: function () {
    return this.get('controller.content.alertsCount');
  }.property('controller.content.alertsCount'),

  hasCriticalAlerts: function () {
    return this.get('controller.content.hasCriticalAlerts');
  }.property('controller.content.alertsCount'),

  /**
   * Define if service has alert definitions defined
   * @type {Boolean}
   */
  hasAlertDefinitions: function () {
    return App.AlertDefinition.find().someProperty('serviceName', this.get('controller.content.serviceName'));
  }.property('controller.content.serviceName'),

  updateComponentInformation: function() {
    var hc = this.get('controller.content.restartRequiredHostsAndComponents');
    var hostsCount = 0;
    var componentsCount = 0;
    for (var host in hc) {
      hostsCount++;
      componentsCount += hc[host].length;
    }
    this.set('componentsCount', componentsCount);
    this.set('hostsCount', hostsCount);
  }.observes('controller.content.restartRequiredHostsAndComponents'),

  rollingRestartSlaveComponentName : function() {
    return batchUtils.getRollingRestartComponentName(this.get('serviceName'));
  }.property('serviceName'),
  rollingRestartActionName : function() {
    var label = null;
    var componentName = this.get('rollingRestartSlaveComponentName');
    if (componentName) {
      label = Em.I18n.t('rollingrestart.dialog.title').format(App.format.role(componentName));
    }
    return label;
  }.property('rollingRestartSlaveComponentName'),

  restartAllStaleConfigComponents: function () {
    var self = this;
    var serviceDisplayName = this.get('service.displayName');
    var bodyMessage = Em.Object.create({
      confirmMsg: Em.I18n.t('services.service.restartAll.confirmMsg').format(serviceDisplayName),
      confirmButton: Em.I18n.t('services.service.restartAll.confirmButton'),
      additionalWarningMsg: this.get('service.passiveState') === 'OFF' ? Em.I18n.t('services.service.restartAll.warningMsg.turnOnMM').format(serviceDisplayName) : null
    });
    return App.showConfirmationFeedBackPopup(function (query) {
      var selectedService = self.get('service.id');
      batchUtils.restartAllServiceHostComponents(selectedService, true, query);
    }, bodyMessage);
  },
  rollingRestartStaleConfigSlaveComponents: function (componentName) {
    batchUtils.launchHostComponentRollingRestart(componentName.context, this.get('service.displayName'), this.get('service.passiveState') === "ON", true);
  },
  /*
   * 'Restart Required bar' ended
   */

   /*
   * Find the graph class associated with the graph name, and split
   * the array into sections of 2 for displaying on the page
   * (will only display rows with 2 items)
   */
  constructGraphObjects: function(graphNames) {
    var result = [], graphObjects = [], chunkSize = 2;
    var self = this;

    if (!graphNames) {
      self.set('serviceMetricGraphs', []);
      self.set('isServiceMetricLoaded', true);
      return;
    }
    // load time range for current service from server
    self.getUserPref(self.get('persistKey')).complete(function () {
      var index = self.get('currentTimeRangeIndex');
      graphNames.forEach(function(graphName) {
        graphObjects.push(App["ChartServiceMetrics" + graphName].extend({
          currentTimeIndex : index
        }));
      });

      while(graphObjects.length) {
        result.push(graphObjects.splice(0, chunkSize));
      }
      self.set('serviceMetricGraphs', result);
      self.set('isServiceMetricLoaded', true);
    });
  },

  /**
   * Contains graphs for this particular service
   */
  serviceMetricGraphs: [],
  isServiceMetricLoaded: false,

  /**
   * Key-name to store time range in Persist
   * @type {string}
   */
  persistKey: function () {
    return 'time-range-service-' + this.get('service.serviceName');
  }.property(),

  getUserPrefSuccessCallback: function (response, request, data) {
    if (response) {
      console.log('Got persist value from server with key ' + data.key + '. Value is: ' + response);
      this.set('currentTimeRangeIndex', response);
    }
  },

  getUserPrefErrorCallback: function (request) {
    if (request.status == 404) {
      console.log('Persist did NOT find the key');
      this.postUserPref(this.get('persistKey'), 0);
      this.set('currentTimeRangeIndex', 0);
    }
  },

  /**
   * time range options for service metrics, a dropdown will list all options
   */
  timeRangeOptions: [
    {index: 0, name: Em.I18n.t('graphs.timeRange.hour'), seconds: 3600},
    {index: 1, name: Em.I18n.t('graphs.timeRange.twoHours'), seconds: 7200},
    {index: 2, name: Em.I18n.t('graphs.timeRange.fourHours'), seconds: 14400},
    {index: 3, name: Em.I18n.t('graphs.timeRange.twelveHours'), seconds: 43200},
    {index: 4, name: Em.I18n.t('graphs.timeRange.day'), seconds: 86400},
    {index: 5, name: Em.I18n.t('graphs.timeRange.week'), seconds: 604800},
    {index: 6, name: Em.I18n.t('graphs.timeRange.month'), seconds: 2592000},
    {index: 7, name: Em.I18n.t('graphs.timeRange.year'), seconds: 31104000}
  ],

  currentTimeRangeIndex: 0,
  currentTimeRange: function() {
    return this.get('timeRangeOptions').objectAt(this.get('currentTimeRangeIndex'));
  }.property('currentTimeRangeIndex'),

  /**
   * onclick handler for a time range option
   */
  setTimeRange: function (event) {
    var self = this;
    if (event && event.context) {
      self.postUserPref(self.get('persistKey'), event.context.index);
      self.set('currentTimeRangeIndex', event.context.index);
      var svcName = self.get('service.serviceName');
      if (svcName) {
        var result = [], graphObjects = [], chunkSize = 2;
        App.service_graph_config[svcName.toLowerCase()].forEach(function(graphName) {
          graphObjects.push(App["ChartServiceMetrics" + graphName].extend({
            currentTimeIndex : event.context.index
          }));
        });
        while(graphObjects.length) {
          result.push(graphObjects.splice(0, chunkSize));
        }
        self.set('serviceMetricGraphs', result);
        self.set('isServiceMetricLoaded', true);
      }
    }
  },

  loadServiceSummary: function () {
    var serviceName = this.get('serviceName');
    var serviceSummaryView = null;

    if (!serviceName) {
      return;
    }

    if (this.get('oldServiceName')) {
      // do not delete it!
      return;
    }

    var customServiceView = this.get('serviceCustomViewsMap')[serviceName];
    if (customServiceView) {
      serviceSummaryView  = customServiceView.extend({
        service: this.get('service')
      });
    } else  {
      serviceSummaryView = Em.View.extend({
        templateName: this.get('templatePathPrefix') + 'base',
        content: this
      });
    }
    this.set('serviceSummaryView', serviceSummaryView);
    this.set('oldServiceName', serviceName);
  }.observes('serviceName'),


  /**
   * Service metrics panel not displayed when metrics service (ex:Ganglia) is not in stack definition.
   */
  isNoServiceMetricsService: function() {
    return !App.get('services.serviceMetrics').length;
  }.property('App.services.serviceMetrics'),

  gangliaUrl:function () {
    var gangliaUrl = App.router.get('clusterController.gangliaUrl');
    if (!gangliaUrl) return null;
    var svcName = this.get('service.serviceName');
    if (svcName) {
      switch (svcName.toLowerCase()) {
        case 'hdfs':
          gangliaUrl += "/?r=hour&cs=&ce=&m=&s=by+name&c=HDPSlaves&tab=m&vn=";
          break;
        case 'hbase':
          gangliaUrl += "?r=hour&cs=&ce=&m=&s=by+name&c=HDPHBaseMaster&tab=m&vn=";
          break;
        default:
          break;
      }
    }
    return gangliaUrl;
  }.property('App.router.clusterController.gangliaUrl', 'service.serviceName'),

  willInsertElement: function () {
    App.router.get('updateController').updateServiceMetric(Em.K);
  },

  didInsertElement: function () {
    var svcName = this.get('service.serviceName');
    var isMetricsSupported = svcName != 'STORM' || App.get('isStormMetricsSupported');
    if (svcName && isMetricsSupported) {
      this.constructGraphObjects(App.service_graph_config[svcName.toLowerCase()]);
    }
    // adjust the summary table height
    var summaryTable = document.getElementById('summary-info');
    if (summaryTable) {
      var rows = $(summaryTable).find('tr');
      if (rows != null && rows.length > 0) {
        var minimumHeightSum = 20;
        var summaryActualHeight = summaryTable.clientHeight;
        // for summary window
        if (summaryActualHeight <= minimumHeightSum) {
          $(summaryTable).attr('style', "height:" + minimumHeightSum + "px;");
        }
      }
    }
  }
});

});

require.register("views/main/service/item", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var batchUtils = require('utils/batch_scheduled_requests');

App.MainServiceItemView = Em.View.extend({
  templateName: require('templates/main/service/item'),

  serviceName: function() {
    return this.get('controller.content.serviceName');
  }.property('controller.content.serviceName'),

  displayName: function() {
    return this.get('controller.content.displayName');
  }.property('controller.content.displayName'),

  isPassive: function() {
    return this.get('controller.content.passiveState') === 'ON';
  }.property('controller.content.passiveState'),

  /**
   * Some custom commands need custom logic to be executed
   */
  mastersExcludedCommands: {
    'NAMENODE': ['DECOMMISSION', 'REBALANCEHDFS'],
    'RESOURCEMANAGER': ['DECOMMISSION', 'REFRESHQUEUES'],
    'HBASE_MASTER': ['DECOMMISSION'],
    'KNOX_GATEWAY': ['STARTDEMOLDAP','STOPDEMOLDAP']
  },

  actionMap: function() {
    return {
      RESTART_ALL: {
        action: 'restartAllHostComponents',
        context: this.get('serviceName'),
        label: Em.I18n.t('restart.service.all'),
        cssClass: 'icon-repeat',
        disabled: false
      },
      RUN_SMOKE_TEST: {
        action: 'runSmokeTest',
        label: Em.I18n.t('services.service.actions.run.smoke'),
        cssClass: 'icon-thumbs-up-alt'
      },
      REFRESH_CONFIGS: {
        action: 'refreshConfigs',
        label: Em.I18n.t('hosts.host.details.refreshConfigs'),
        cssClass: 'icon-refresh',
        disabled: !this.get('controller.content.isRestartRequired')
      },
      REFRESH_YARN_QUEUE: {
        action: 'refreshYarnQueues',
        customCommand: 'REFRESHQUEUES',
        label: Em.I18n.t('services.service.actions.run.yarnRefreshQueues.menu'),
        cssClass: 'icon-refresh',
        disabled: false
      },
      ROLLING_RESTART: {
        action: 'rollingRestart',
        context: this.get('rollingRestartComponent'),
        label: Em.I18n.t('rollingrestart.dialog.title'),
        cssClass: 'icon-time',
        disabled: false
      },
      TOGGLE_PASSIVE: {
        action: 'turnOnOffPassive',
        context: this.get('isPassive') ? Em.I18n.t('passiveState.turnOffFor').format(this.get('displayName')) : Em.I18n.t('passiveState.turnOnFor').format(this.get('displayName')),
        label: this.get('isPassive') ? Em.I18n.t('passiveState.turnOff') : Em.I18n.t('passiveState.turnOn'),
        cssClass: 'icon-medkit',
        disabled: false
      },
      TOGGLE_NN_HA: {
        action: App.get('isHaEnabled') ? 'disableHighAvailability' : 'enableHighAvailability',
        label: App.get('isHaEnabled') ? Em.I18n.t('admin.highAvailability.button.disable') : Em.I18n.t('admin.highAvailability.button.enable'),
        cssClass: App.get('isHaEnabled') ? 'icon-arrow-down' : 'icon-arrow-up',
        isHidden: (App.get('isHaEnabled') || (/^1.3/.test(App.get('currentStackVersionNumber')))),
        disabled: App.get('isSingleNode')
      },
      TOGGLE_RM_HA: {
        action: 'enableRMHighAvailability',
        label: Em.I18n.t('admin.rm_highAvailability.button.enable'),
        cssClass: 'icon-arrow-up',
        isHidden: App.get('isRMHaEnabled'),
        disabled: App.get('isSingleNode')
      },
      MOVE_COMPONENT: {
        action: 'reassignMaster',
        context: '',
        label: Em.I18n.t('services.service.actions.reassign.master'),
        cssClass: 'icon-share-alt'
      },
      STARTDEMOLDAP: {
        action: 'startLdapKnox',
        customCommand: 'STARTDEMOLDAP',
        label: Em.I18n.t('services.service.actions.run.startLdapKnox.context'),
        cssClass: 'icon-play-sign',
        disabled: false
      },
      STOPDEMOLDAP: {
        action: 'stopLdapKnox',
        customCommand: 'STOPDEMOLDAP',
        label: Em.I18n.t('services.service.actions.run.stopLdapKnox.context'),
        cssClass: 'icon-stop',
        disabled: false
      },
      REBALANCE_HDFS: {
        action: 'rebalanceHdfsNodes',
        customCommand: 'REBALANCEHDFS',
        context: Em.I18n.t('services.service.actions.run.rebalanceHdfsNodes.context'),
        label: Em.I18n.t('services.service.actions.run.rebalanceHdfsNodes'),
        cssClass: 'icon-refresh',
        disabled: false
      },
      DOWNLOAD_CLIENT_CONFIGS: {
        action: this.get('controller.isSeveralClients') ? '' : 'downloadClientConfigs',
        label: Em.I18n.t('services.service.actions.downloadClientConfigs'),
        cssClass: 'icon-download-alt',
        isHidden: !!this.get('controller.content.clientComponents') ? this.get('controller.content.clientComponents').rejectProperty('totalCount', 0).length == 0 : false,
        disabled: false,
        hasSubmenu: this.get('controller.isSeveralClients'),
        submenuOptions: this.get('controller.clientComponents')
      },
      MASTER_CUSTOM_COMMAND: {
        action: 'executeCustomCommand',
        cssClass: 'icon-play-circle',
        isHidden: false,
        disabled: false
      }
    }
  },

   addActionMap: function() {
     return [
      {
        cssClass: 'icon-plus',
        'label': '{0} {1}'.format(Em.I18n.t('add'), Em.I18n.t('dashboard.services.hbase.masterServer')),
        service: 'HBASE',
        component: 'HBASE_MASTER'
      },
      {
       cssClass: 'icon-plus',
       'label': '{0} {1}'.format(Em.I18n.t('add'), Em.I18n.t('dashboard.services.hive.metastore')),
       service: 'HIVE',
       component: 'HIVE_METASTORE',
       isHidden: !App.get('isHadoop22Stack')
      },
      {
       cssClass: 'icon-plus',
       'label': '{0} {1}'.format(Em.I18n.t('add'), Em.I18n.t('dashboard.services.hive.server2')),
       service: 'HIVE',
       component: 'HIVE_SERVER',
       isHidden: !App.get('isHadoop22Stack')
      },
      {
        cssClass: 'icon-plus',
        'label': '{0} {1}'.format(Em.I18n.t('add'), Em.I18n.t('dashboard.services.zookeeper.server')),
        service: 'ZOOKEEPER',
        component: 'ZOOKEEPER_SERVER'
      },
      {
        cssClass: 'icon-plus',
        'label': '{0} {1}'.format(Em.I18n.t('add'), Em.I18n.t('dashboard.services.flume.agentLabel')),
        service: 'FLUME',
        component: 'FLUME_HANDLER'
      }
    ]
  },
  /**
   * Create option for MOVE_COMPONENT or ROLLING_RESTART task.
   *
   * @param {Object} option - one of the options that return by <code>actionMap()</code>
   * @param {Object} fields - option fields to add/rewrite
   * @return {Object}
   */
  createOption: function(option, fields) {
    return $.extend(true, {}, option, fields);
  },

  maintenance: [],

  observeMaintenance: function() {
    Em.run.once(this, 'observeMaintenanceOnce');
  },

  observeMaintenanceOnce: function() {
    var self = this;
    var options = [];
    var service = this.get('controller.content');
    var allMasters = service.get('hostComponents').filterProperty('isMaster').mapProperty('componentName').uniq();
    var allSlaves = service.get('slaveComponents').rejectProperty('totalCount', 0).mapProperty('componentName');
    var actionMap = this.actionMap();
    var serviceCheckSupported = App.get('services.supportsServiceCheck').contains(service.get('serviceName'));
    var hasConfigTab = this.get('hasConfigTab');
    var excludedCommands = this.get('mastersExcludedCommands');

    if (this.get('controller.isClientsOnlyService')) {
      if (serviceCheckSupported) {
        options.push(actionMap.RUN_SMOKE_TEST);
      }
      if (hasConfigTab) {
	      options.push(actionMap.REFRESH_CONFIGS);
	    }
    } else {
      if (this.get('serviceName') === 'FLUME') {
        options.push(actionMap.REFRESH_CONFIGS);
      }
      if (this.get('serviceName') === 'YARN') {
        options.push(actionMap.REFRESH_YARN_QUEUE);
      }
      options.push(actionMap.RESTART_ALL);
      allSlaves.filter(function (slave) {
        return App.get('components.rollinRestartAllowed').contains(slave);
      }).forEach(function(slave) {
        options.push(self.createOption(actionMap.ROLLING_RESTART, {
          context: slave,
          label: actionMap.ROLLING_RESTART.label.format(App.format.role(slave))
        }));
      });
      allMasters.filter(function(master) {
        return App.get('components.reassignable').contains(master);
      }).forEach(function(master) {
        options.push(self.createOption(actionMap.MOVE_COMPONENT, {
          context: master,
          label: actionMap.MOVE_COMPONENT.label.format(App.format.role(master)),
          disabled: App.allHostNames.length === App.HostComponent.find().filterProperty('componentName', master).mapProperty('hostName').length
        }));
      });
      if (service.get('serviceTypes').contains('HA_MODE')) {
        switch (service.get('serviceName')) {
          case 'HDFS':
            options.push(actionMap.TOGGLE_NN_HA);
            break;
          case 'YARN':
            options.push(actionMap.TOGGLE_RM_HA);
            break;
        }
      }
      if (serviceCheckSupported) {
        options.push(actionMap.RUN_SMOKE_TEST);
      }
      options.push(actionMap.TOGGLE_PASSIVE);
      var serviceName = service.get('serviceName');
      var nnComponent = App.StackServiceComponent.find().findProperty('componentName','NAMENODE');
      var knoxGatewayComponent = App.StackServiceComponent.find().findProperty('componentName','KNOX_GATEWAY');
      if (serviceName === 'HDFS' && nnComponent) {
        var namenodeCustomCommands = nnComponent.get('customCommands');
        if (namenodeCustomCommands && namenodeCustomCommands.contains('REBALANCEHDFS'))
        options.push(actionMap.REBALANCE_HDFS);
      }

      if (serviceName === 'KNOX' && knoxGatewayComponent) {
        var knoxGatewayCustomCommands = knoxGatewayComponent.get('customCommands');
        knoxGatewayCustomCommands.forEach(function(command) {
          if (actionMap[command]) {
            options.push(actionMap[command]);
          }
        });
      }
      self.addActionMap().filterProperty('service', serviceName).forEach(function(item) {
        item.action = 'add' + item.component;
        item.disabled = self.get('controller.isAddDisabled-' + item.component);
        item.tooltip = self.get('controller.addDisabledTooltip' + item.component);
        options.push(item);
      });

      allMasters.forEach(function(master) {
        var component = App.StackServiceComponent.find(master);
        var commands = component.get('customCommands');

        if (!commands.length) {
          return false;
        }

        commands.forEach(function(command) {
          if (excludedCommands[master] && excludedCommands[master].contains(command)){
            return false;
          }

          options.push(self.createOption(actionMap.MASTER_CUSTOM_COMMAND, {
            label: Em.I18n.t('services.service.actions.run.executeCustomCommand.menu').format(command),
            context: {
              label: Em.I18n.t('services.service.actions.run.executeCustomCommand.menu').format(command),
              service: component.get('serviceName'),
              component: component.get('componentName'),
              command: command
            }
          }));
        });
      });
    }

    if (hasConfigTab) {
      options.push(actionMap.DOWNLOAD_CLIENT_CONFIGS);
    }

    if (!this.get('maintenance').length) {
      this.set('maintenance', options);
    } else {
      this.get('maintenance').forEach(function(option, index) {
        if ( JSON.stringify(option) != JSON.stringify(options[index])  ) {
          self.get('maintenance').removeAt(index).insertAt(index, options[index]);
        }
      });
      options.forEach(function(opt, index) {
        if ( JSON.stringify(opt) != JSON.stringify(self.get('maintenance')[index])  ) {
          self.get('maintenance').pushObject(opt);
        }
      });
    }
  },

  isMaintenanceActive: function() {
    return this.get('state') !== 'inDOM' || this.get('maintenance').length !== 0;
  }.property('maintenance'),

  hasConfigTab: function() {
    return !App.get('services.noConfigTypes').contains(this.get('controller.content.serviceName'));
  }.property('controller.content.serviceName','App.services.noConfigTypes'),

  didInsertElement: function () {
    this.get('controller').setStartStopState();
  },

  willInsertElement: function () {
    this.addObserver('controller.isStopDisabled', this, 'observeMaintenance');
    this.addObserver('controller.isClientsOnlyService', this, 'observeMaintenance');
    this.addObserver('controller.content.isRestartRequired', this, 'observeMaintenance');
  },

  willDestroyElement: function() {
    this.removeObserver('controller.isStopDisabled', this, 'observeMaintenance');
    this.removeObserver('controller.isClientsOnlyService', this, 'observeMaintenance');
    this.removeObserver('controller.content.isRestartRequired', this, 'observeMaintenance');
  },
  service:function () {
    var svc = this.get('controller.content');
    var svcName = svc.get('serviceName');
    if (svcName) {
      switch (svcName.toLowerCase()) {
        case 'hdfs':
          svc = App.HDFSService.find().objectAt(0);
          break;
        case 'yarn':
          svc = App.YARNService.find().objectAt(0);
          break;
        case 'hbase':
          svc = App.HBaseService.find().objectAt(0);
          break;
        case 'flume':
          svc = App.FlumeService.find().objectAt(0);
          break;
        default:
          break;
      }
    }
    return svc;
  }.property('controller.content.serviceName').volatile()
});

});

require.register("views/main/service/manage_config_groups_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainServiceManageConfigGroupView = Em.View.extend({

  templateName: require('templates/main/service/manage_configuration_groups_popup'),

  selectedConfigGroup: null,

  isRemoveButtonDisabled: true,

  isRenameButtonDisabled: true,

  isDuplicateButtonDisabled: true,

  //Disable actions remove and rename for Default config group
  buttonObserver: function () {
    var selectedConfigGroup = this.get('controller.selectedConfigGroup');
    if(selectedConfigGroup.isDefault){
      this.set('isRemoveButtonDisabled', true);
      this.set('isRenameButtonDisabled', true);
      this.set('isDuplicateButtonDisabled', false);
    }else{
      this.set('isRemoveButtonDisabled', false);
      this.set('isRenameButtonDisabled', false);
      this.set('isDuplicateButtonDisabled', false);
    }
  }.observes('controller.selectedConfigGroup'),

  onGroupSelect: function () {
    var selectedConfigGroup = this.get('selectedConfigGroup');
    // to unable user select more than one config group at a time
    if (selectedConfigGroup.length) {
      this.set('controller.selectedConfigGroup', selectedConfigGroup[selectedConfigGroup.length - 1]);
    }
    if (selectedConfigGroup.length > 1) {
      this.set('selectedConfigGroup', selectedConfigGroup[selectedConfigGroup.length - 1]);
    }
    this.set('controller.selectedHosts', []);
  }.observes('selectedConfigGroup'),

  onLoad: function () {
    if (this.get('controller.isLoaded')) {
      this.set('selectedConfigGroup', this.get('controller.configGroups')[0])
    }
  }.observes('controller.isLoaded', 'controller.configGroups'),

  willInsertElement: function() {
    this.get('controller').loadHosts();
  },

  didInsertElement: function () {
    this.selectDefaultGroup();
    App.tooltip($('.properties-link'));
    App.tooltip($("[rel='button-info']"));
    App.tooltip($("[rel='button-info-dropdown']"), {placement: 'left'});
  },

  selectDefaultGroup: function () {
    if (this.get('controller.isLoaded')) {
      this.set('selectedConfigGroup', [this.get('controller.configGroups').findProperty('isDefault')]);
    }
  }.observes('controller.isLoaded'),

  addButtonTooltip: function () {
    return  Em.I18n.t('services.service.config_groups_popup.addButton');
  }.property(),
  removeButtonTooltip: function () {
    return  Em.I18n.t('services.service.config_groups_popup.removeButton');
  }.property(),
  renameButtonTooltip: function () {
    return  Em.I18n.t('services.service.config_groups_popup.renameButton');
  }.property(),
  duplicateButtonTooltip: function () {
    return  Em.I18n.t('services.service.config_groups_popup.duplicateButton');
  }.property(),
  addHostTooltip: function () {
    if (!this.get('controller.selectedConfigGroup.isDefault') && this.get('controller.selectedConfigGroup.isAddHostsDisabled')) {
      return Em.I18n.t('services.service.config_groups_popup.addHostDisabled');
    } else {
      return  Em.I18n.t('services.service.config_groups_popup.addHost');
    }
  }.property('controller.selectedConfigGroup.isDefault', 'controller.selectedConfigGroup.isAddHostsDisabled'),
  removeHostTooltip: function () {
    return  Em.I18n.t('services.service.config_groups_popup.removeHost');
  }.property()

});

});

require.register("views/main/service/menu", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var misc = require('utils/misc');

App.MainServiceMenuView = Em.CollectionView.extend({
  disabledServices: [],

  content:function () {
    var items = App.router.get('mainServiceController.content').filter(function(item){
      return !this.get('disabledServices').contains(item.get('id'));
    }, this);
    var stackServices = App.StackService.find().mapProperty('serviceName');
    return misc.sortByOrder(stackServices, items);
  }.property('App.router.mainServiceController.content', 'App.router.mainServiceController.content.length'),

  didInsertElement:function () {
    App.router.location.addObserver('lastSetURL', this, 'renderOnRoute');
    this.renderOnRoute();
    App.tooltip($(".restart-required-service"), {html:true, placement:"right"});
  },

  activeServiceId:null,

  /**
   *    Syncs navigation menu with requested URL
   */
  renderOnRoute:function () {
    var last_url = App.router.location.lastSetURL || location.href.replace(/^[^#]*#/, '');
    if (last_url.substr(1, 4) !== 'main' || !this._childViews) {
      return;
    }
    var reg = /^\/main\/services\/(\S+)\//g;
    var sub_url = reg.exec(last_url);
    var service_id = (null != sub_url) ? sub_url[1] : 1;
    this.set('activeServiceId', service_id);

  },

  tagName:'ul',
  classNames:[ "nav", "nav-list", "nav-services"],

  itemViewClass:Em.View.extend({

    classNameBindings:["active", "clients"],
    templateName:require('templates/main/service/menu_item'),
    restartRequiredMessage: null,

    shouldBeRestarted: function() {
      return this.get('content.hostComponents').someProperty('staleConfigs', true);
    }.property('content.hostComponents.@each.staleConfigs'),

    active:function () {
      return this.get('content.id') == this.get('parentView.activeServiceId') ? 'active' : '';
    }.property('parentView.activeServiceId'),

    alertsCount: function () {
      return this.get('content.alertsCount');
    }.property('content.alertsCount'),

    hasCriticalAlerts: function () {
      return this.get('content.hasCriticalAlerts');
    }.property('content.hasCriticalAlerts'),

    isConfigurable: function () {
      return !App.get('services.noConfigTypes').contains(this.get('content.serviceName'));
    }.property('App.services.noConfigTypes','content.serviceName'),

    link: function() {
      var stateName = (['summary','configs'].contains(App.router.get('currentState.name')))
        ? this.get('isConfigurable') && this.get('parentView.activeServiceId') != this.get('content.id') ?  App.router.get('currentState.name') : 'summary'
        : 'summary';
      return "#/main/services/" + this.get('content.id') + "/" + stateName;
    }.property('App.router.currentState.name', 'parentView.activeServiceId', 'isConfigurable'),

    goToConfigs: function () {
      App.router.set('mainServiceItemController.routeToConfigs', true);
      App.router.transitionTo('services.service.configs', this.get('content'));
      App.router.set('mainServiceItemController.routeToConfigs', false);
    },

    refreshRestartRequiredMessage: function() {
      var restarted, componentsCount, hostsCount, message, tHosts, tComponents;
      restarted = this.get('content.restartRequiredHostsAndComponents');
      componentsCount = 0;
      hostsCount = 0;
      message = "";
      for (var host in restarted) {
        hostsCount++;
        componentsCount += restarted[host].length;
      }
      if (hostsCount > 1) {
        tHosts = Em.I18n.t('common.hosts');
      } else {
        tHosts = Em.I18n.t('common.host');
      }
      if (componentsCount > 1) {
        tComponents = Em.I18n.t('common.components');
      } else {
        tComponents = Em.I18n.t('common.component');
      }
      message += componentsCount + ' ' + tComponents + ' ' + Em.I18n.t('on') + ' ' +
        hostsCount + ' ' + tHosts + ' ' + Em.I18n.t('services.service.config.restartService.needToRestartEnd');
      this.set('restartRequiredMessage', message);
    }.observes('content.restartRequiredHostsAndComponents')
  })

});

App.TopNavServiceMenuView = Em.CollectionView.extend({
  disabledServices: [],

  content:function () {
    var items = App.router.get('mainServiceController.content').filter(function(item){
      return !this.get('disabledServices').contains(item.get('id'));
    }, this);
    var stackServices = App.StackService.find().mapProperty('serviceName');
    return misc.sortByOrder(stackServices, items);
  }.property('App.router.mainServiceController.content', 'App.router.mainServiceController.content.length'),

  didInsertElement:function () {
    App.router.location.addObserver('lastSetURL', this, 'renderOnRoute');
    this.renderOnRoute();
    App.tooltip($(".restart-required-service"), {html:true, placement:"right"});
  },

  activeServiceId:null,
  /**
   *    Syncs navigation menu with requested URL
   */
  renderOnRoute:function () {
    var last_url = App.router.location.lastSetURL || location.href.replace(/^[^#]*#/, '');
    if (last_url.substr(1, 4) !== 'main' || !this._childViews) {
      return;
    }
    var reg = /^\/main\/services\/(\S+)\//g;
    var sub_url = reg.exec(last_url);
    var service_id = (null != sub_url) ? sub_url[1] : 1;
    this.set('activeServiceId', service_id);
  },

  tagName:'ul',
  classNames:[ "top-nav-dropdown-menu"],

  itemViewClass:Em.View.extend({

    classNameBindings:["active", "clients"],
    templateName:require('templates/main/service/menu_item'),
    restartRequiredMessage: null,

    shouldBeRestarted: function() {
      return this.get('content.hostComponents').someProperty('staleConfigs', true);
    }.property('content.hostComponents.@each.staleConfigs'),

    active:function () {
      return this.get('content.id') == this.get('parentView.activeServiceId') ? 'active' : '';
    }.property('parentView.activeServiceId'),

    alertsCount: function () {
      return this.get('content.alertsCount');
    }.property('content.alertsCount'),

    hasCriticalAlerts: function () {
      return this.get('content.hasCriticalAlerts');
    }.property('content.hasCriticalAlerts'),

    isConfigurable: function () {
      return !App.get('services.noConfigTypes').contains(this.get('content.serviceName'));
    }.property('App.services.noConfigTypes','content.serviceName'),

    link: function() {
      var stateName = (['summary','configs'].contains(App.router.get('currentState.name')))
        ? this.get('isConfigurable') ?  App.router.get('currentState.name') : 'summary'
        : 'summary';
      return "#/main/services/" + this.get('content.id') + "/" + stateName;
    }.property('App.router.currentState.name', 'parentView.activeServiceId','isConfigurable'),

    goToConfigs: function () {
      App.router.set('mainServiceItemController.routeToConfigs', true);
      App.router.transitionTo('services.service.configs', this.get('content'));
      App.router.set('mainServiceItemController.routeToConfigs', false);
    },

    refreshRestartRequiredMessage: function() {
      var restarted, componentsCount, hostsCount, message, tHosts, tComponents;
      restarted = this.get('content.restartRequiredHostsAndComponents');
      componentsCount = 0;
      hostsCount = 0;
      message = "";
      for (var host in restarted) {
        hostsCount++;
        componentsCount += restarted[host].length;
      }
      if (hostsCount > 1) {
        tHosts = Em.I18n.t('common.hosts');
      } else {
        tHosts = Em.I18n.t('common.host');
      }
      if (componentsCount > 1) {
        tComponents = Em.I18n.t('common.components');
      } else {
        tComponents = Em.I18n.t('common.component');
      }
      message += componentsCount + ' ' + tComponents + ' ' + Em.I18n.t('on') + ' ' +
        hostsCount + ' ' + tHosts + ' ' + Em.I18n.t('services.service.config.restartService.needToRestartEnd');
      this.set('restartRequiredMessage', message);
    }.observes('content.restartRequiredHostsAndComponents')
  })

});

});

require.register("views/main/service/reassign/step1_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.ReassignMasterWizardStep1View = Em.View.extend({

  message: function () {
    var messages = [
      Em.I18n.t('services.reassign.step1.message1').format(this.get('controller.content.reassign.display_name')),
      Em.I18n.t('services.reassign.step1.message3').format(this.get('controller.content.reassign.display_name'))
    ];
    if (this.get('controller.content.hasManualSteps')) {
      messages.splice(1,0, Em.I18n.t('services.reassign.step1.message2').format(this.get('controller.content.reassign.display_name')));
    }
    return messages;
  }.property('controller.content.reassign.display_name','controller.content.hasManualSteps'),

  templateName: require('templates/main/service/reassign/step1')

});

});

require.register("views/main/service/reassign/step2_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.ReassignMasterWizardStep2View = App.WizardStep5View.extend({
  body: function () {
    if (this.get('controller.content.reassign.component_name') === 'NAMENODE' && App.get('isHaEnabled')) {
      return Em.I18n.t('services.reassign.step2.body.namenodeHA').format(this.get('controller.content.reassign.display_name'));
    }
    return Em.I18n.t('services.reassign.step2.body').format(this.get('controller.content.reassign.display_name'));
  }.property('controller.content.reassign.component_name', 'controller.content.reassign.display_name')
});

});

require.register("views/main/service/reassign/step3_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.ReassignMasterWizardStep3View = Em.View.extend({

  templateName: require('templates/main/service/reassign/step3'),

  sourceHost: function(){
    return this.get('controller.content.reassignHosts.source');
  }.property('controller.content.reassignHosts.source'),

  targetHost: function(){
    return this.get('controller.content.reassignHosts.target');
  }.property('controller.content.reassignHosts.target'),

  printReview: function () {
    $("#step8-info").jqprint();
  },

  jdbcSetupMessage: function() {
    if(['HIVE_SERVER', 'HIVE_METASTORE', 'OOZIE_SERVER'].contains(this.get('controller.content.reassign.component_name'))) {
      if(this.get('controller.content.reassign.component_name') === 'OOZIE_SERVER' && this.get('controller.content.databaseType') === 'derby') {
        return false;
      }

      return Em.I18n.t('services.service.config.database.msg.jdbcSetup').format(this.get('controller.content.databaseType'), this.get('controller.content.databaseType'));
    }

    return false;
  }.property('controller.content.reassign.display_name, controller.content.databaseType')
});

});

require.register("views/main/service/reassign/step4_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.ReassignMasterWizardStep4View = App.HighAvailabilityProgressPageView.extend({

  headerTitle: Em.I18n.t('services.reassign.step4.header'),

  noticeInProgress: function () {
    return Em.I18n.t('services.reassign.step4.status.info').format(App.format.role(this.get('controller.content.reassign.component_name')))
  }.property('controller.content.reassign.component_name'),

  noticeFailed: function () {
    return Em.I18n.t('services.reassign.step4.status.failed').format(App.format.role(this.get('controller.content.reassign.component_name')),this.get('controller.content.reassignHosts.source'),this.get('controller.content.reassignHosts.target'));
  }.property('controller.content.reassign.component_name','controller.content.reassignHosts.source','controller.content.reassignHosts.target'),

  noticeCompleted: function () {
    if (this.get('controller.content.hasManualSteps')) {
      return Em.I18n.t('services.reassign.step4.status.success.withManualSteps').format(App.format.role(this.get('controller.content.reassign.component_name')));
    } else {
      return Em.I18n.t('services.reassign.step4.status.success').format(App.format.role(this.get('controller.content.reassign.component_name')),this.get('controller.content.reassignHosts.source'),this.get('controller.content.reassignHosts.target'));
    }
  }.property('controller.content.reassign.component_name','controller.content.reassignHosts.source','controller.content.reassignHosts.target'),

  submitButtonText: function () {
    if (this.get('controller.content.hasManualSteps')) {
      return Em.I18n.t('common.next') + ' &rarr;';
    } else {
      return Em.I18n.t('common.complete');
    }
  }.property('controller.content.hasManualSteps'),

  templateName: require('templates/main/service/reassign/step4')
});

});

require.register("views/main/service/reassign/step5_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.ReassignMasterWizardStep5View = Em.View.extend({

  templateName: require('templates/main/service/reassign/step5'),

  manualCommands: function () {
    if (!this.get('controller.content.componentsWithManualCommands').contains(this.get('controller.content.reassign.component_name'))) {
      return '';
    }
    var componentDir = this.get('controller.content.componentDir') || '';
    var componentDirCmd = componentDir.replace(/,/g, ' ');
    var sourceHost = this.get('controller.content.reassignHosts.source');
    var targetHost = this.get('controller.content.reassignHosts.target');
    var ha = '';
    if (this.get('controller.content.reassign.component_name') === 'NAMENODE' && App.get('isHaEnabled')) {
      ha = '_ha';
      var nnStartedHost = this.get('controller.content.masterComponentHosts').filterProperty('component', 'NAMENODE').mapProperty('hostName').without(sourceHost);
    }

    var user = this.get('controller.content.hdfsUser');

    if(this.get('controller.content.reassign.component_name') === 'APP_TIMELINE_SERVER') {
      user = this.get('controller.content.serviceProperties.yarn-env.yarn_user');
    }

    return Em.I18n.t('services.reassign.step5.body.' + this.get('controller.content.reassign.component_name').toLowerCase() + ha).format(componentDir, sourceHost, targetHost, user, nnStartedHost,this.get('controller.content.group'), componentDirCmd);
  }.property('controller.content.reassign.component_name', 'controller.content.componentDir', 'controller.content.masterComponentHosts', 'controller.content.reassign.host_id', 'controller.content.hdfsUser'),

  /**
   * security notice to generate keytab manually is not used any more
   */
  securityNotice: function () {
    var secureConfigs = this.get('controller.content.secureConfigs');
    var proceedMsg = Em.I18n.t('services.reassign.step5.body.proceedMsg');
    var hasSecureConfigs = !this.get('controller.content.componentsWithoutSecurityConfigs').contains(this.get('controller.content.reassign.component_name'));
    if(!hasSecureConfigs) {
      secureConfigs = [];
    }
    if (!this.get('controller.content.securityEnabled') || !secureConfigs.length) {
      return proceedMsg;
    }
    var formattedText = '<ul>';
    secureConfigs.forEach(function (config) {
      formattedText += '<li>' + Em.I18n.t('services.reassign.step5.body.securityConfigsList').format(config.keytab,
          config.principal.replace('_HOST', this.get('controller.content.reassignHosts.target')), this.get('controller.content.reassignHosts.target')) + '</li>';
    }, this);
    formattedText += '</ul>';
    return Em.I18n.t('services.reassign.step5.body.securityNotice').format(formattedText) + proceedMsg;
  }.property('controller.content.securityEnabled', 'controller.content.secureConfigs', 'controller.content.reassignHosts.target')
});

});

require.register("views/main/service/reassign/step6_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.ReassignMasterWizardStep6View = App.HighAvailabilityProgressPageView.extend({

  headerTitle: Em.I18n.t('services.reassign.step6.header'),

  noticeInProgress: function () {
    return Em.I18n.t('services.reassign.step6.status.info').format(App.format.role(this.get('controller.content.reassign.component_name')))
  }.property('controller.content.reassign.component_name'),

  noticeFailed: function () {
    return Em.I18n.t('services.reassign.step6.status.failed').format(App.format.role(this.get('controller.content.reassign.component_name')),this.get('controller.content.reassignHosts.source'),this.get('controller.content.reassignHosts.target'))
  }.property('controller.content.reassign.component_name','controller.content.reassignHosts.source','controller.content.reassignHosts.target'),

  noticeCompleted: function () {
    return Em.I18n.t('services.reassign.step6.status.success').format(App.format.role(this.get('controller.content.reassign.component_name')),this.get('controller.content.reassignHosts.source'),this.get('controller.content.reassignHosts.target'))
  }.property('controller.content.reassign.component_name','controller.content.reassignHosts.source','controller.content.reassignHosts.target'),

  submitButtonText: Em.I18n.t('common.complete'),

  templateName: require('templates/main/service/reassign/step6')
});

});

require.register("views/main/service/reassign/step7_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.ReassignMasterWizardStep7View = App.HighAvailabilityProgressPageView.extend({

  headerTitle: Em.I18n.t('services.reassign.step7.header'),

  noticeInProgress: function () {
    return Em.I18n.t('services.reassign.step7.info');
  }.property(),

  noticeFailed: function () {
    return Em.I18n.t('services.reassign.step7.failed');
  }.property(),

  noticeCompleted: function () {
    return Em.I18n.t('services.reassign.step7.success');
  }.property(),

  submitButtonText: function () {
    return Em.I18n.t('common.complete') + ' &rarr;';
  }.property(),

  templateName: require('templates/main/service/reassign/step7')
});

});

require.register("views/main/service/reassign_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.ReassignMasterView = Em.View.extend(App.WizardMenuMixin, {

  templateName: require('templates/main/service/reassign'),

  isStep1Disabled: function () {
    return this.isStepDisabled(1);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep2Disabled: function () {
    return this.isStepDisabled(2);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep3Disabled: function () {
    return this.isStepDisabled(3);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep4Disabled: function () {
    return this.isStepDisabled(4);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep5Disabled: function () {
    return this.isStepDisabled(5);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep6Disabled: function () {
    return this.isStepDisabled(6);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStep7Disabled: function () {
    return this.isStepDisabled(7);
  }.property('controller.isStepDisabled.@each.value').cacheable(),

  isStepDisabled: function (index) {
    return this.get('controller.isStepDisabled').findProperty('step', index).get('value');
  },

  isLoaded: false,

  willInsertElement: function() {
    this.set('isLoaded', false);
    this.loadHosts();
  },

  /**
   * load hosts from server
   */
  loadHosts: function () {
    App.ajax.send({
      name: 'hosts.high_availability.wizard',
      data: {},
      sender: this,
      success: 'loadHostsSuccessCallback',
      error: 'loadHostsErrorCallback'
    });
  },

  loadHostsSuccessCallback: function (data, opt, params) {
    var hosts = {};

    data.items.forEach(function (item) {
      hosts[item.Hosts.host_name] = {
        name: item.Hosts.host_name,
        cpu: item.Hosts.cpu_count,
        memory: item.Hosts.total_mem,
        disk_info: item.Hosts.disk_info,
        bootStatus: "REGISTERED",
        isInstalled: true
      };
    });
    App.db.setHosts(hosts);
    this.set('controller.content.hosts', hosts);
    this.set('isLoaded', true);
  },

  loadHostsErrorCallback: function(){
    this.set('isLoaded', true);
  }

});

});

require.register("views/main/service/reconfigure", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var App = require('app');

App.MainServiceReconfigureView = Em.View.extend({

  templateName: require('templates/main/service/reconfigure')

});

App.StageLabelView = Em.View.extend({
  tagName: 'a',
  classNameBindings: ['removeLink'],
  attributeBindings: ['href'],
  href: '#',
  removeLink: null,
  didInsertElement: function() {
   this.onLink();
  },
  onLink: function() {
   if (this.get('showLink') === true) {
     this.set('removeLink',null);
   } else {
     this.set('removeLink','remove-link');
   }
  }.observes('showLink'),
  command: null,
  click: function () {
    if (this.get('command') && this.get('showLink')) {
      this.showHostPopup(this.get('command'));
    }
  },

  showHostPopup: function (command) {
    var controller = this.get("controller");
    App.router.get('applicationController').dataLoading().done(function (initValue) {
      var popupView = App.HostPopup.initPopup(command.get('label'), controller, false, command.get('requestId'));
      popupView.set('isNotShowBgChecked', !initValue);
    })
  },

  isStarted: function () {
    return  (this.get('command') && this.get('command.isStarted'));
  }.property('command.isStarted'),

  showLink: function () {
    return (this.get('command') && this.get('command.showLink'));
  }.property('command.showLink')
});

App.StageSuccessView = Em.View.extend({
  layout: Ember.Handlebars.compile('<i class="icon-ok icon-large"></i> {{t common.done}}')
});

App.StageFailureView = Em.View.extend({
  layout: Ember.Handlebars.compile('<i class="icon-remove icon-large"></i> {{t common.failed}}')
});

App.StageInProgressView = Em.View.extend({
  command: null,
  classNames: ['progress-striped', 'active', 'progress'],
  template: Ember.Handlebars.compile('<div class="bar" {{bindAttr style="command.barWidth"}}></div>'),

  isStageCompleted: function () {
    return this.get('obj.progress') == 100 || this.get('controller.isStepCompleted');
  }.property('controller.isStepCompleted', 'obj.progress')

});
});

require.register("views/main/service/service", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var uiEffects = require('utils/ui_effects');

App.MainDashboardServiceHealthView = Em.View.extend({
  classNameBindings: ["healthStatus", "healthStatusClass"],
  //template: Em.Handlebars.compile(""),
  blink: false,
  tagName: 'span',
  attributeBindings:['rel', 'title','data-original-title'],
  rel: 'HealthTooltip',
  'data-original-title': '',

  updateToolTip: function () {
    this.set('data-original-title', this.get('service.toolTipContent'));
  }.observes('service.toolTipContent'),

  startBlink: function () {
    this.set('blink', true);
  },

  doBlink: function () {
    var self = this;
    if (this.get('blink') && (this.get("state") == "inDOM")) {
      uiEffects.pulsate(self.$(), 1000, function(){
        self.doBlink();
      });
    }
  }.observes('blink'),

  stopBlink: function () {
    this.set('blink', false);
  },

  healthStatus: function () {
    var isClientOnlyService = App.get('services.clientOnly').contains(this.get('service.serviceName'));
    if (isClientOnlyService) {
      return 'icon-laptop';
    }
    if (this.get('service.passiveState') != 'OFF') {
      return 'icon-medkit';
    }
    var status = this.get('service.healthStatus');
    switch (status) {
      case 'green':
        status = App.Service.Health.live;
        this.stopBlink();
        break;
      case 'green-blinking':
        status = App.Service.Health.live;
        this.startBlink();
        break;
      case 'red-blinking':
        status = App.Service.Health.dead;
        this.startBlink();
        break;
      case 'yellow':
        status = App.Service.Health.unknown;
        break;
      default:
        status = App.Service.Health.dead;
        this.stopBlink();
        break;
    }

    return 'health-status-' + status;
  }.property('service.healthStatus','service.passiveState','service.serviceName'),

  healthStatusClass: function () {
    if (this.get('service.passiveState') != 'OFF' || App.get('services.clientOnly').contains(this.get('service.serviceName')))
      return '';
    switch (this.get('service.healthStatus')) {
      case 'green':
      case 'green-blinking':
        return App.healthIconClassGreen;
        break;
      case 'red':
      case 'red-blinking':
        return App.healthIconClassRed;
        break;
      case 'yellow':
        return App.healthIconClassYellow;
      default:
        return '';
    }
  }.property('service.healthStatus','service.passiveState','service.serviceName'),

  didInsertElement: function () {
    this.updateToolTip();
    App.tooltip($("[rel='HealthTooltip']"));
    this.doBlink(); // check for blink availability
  }
});

App.ComponentLiveTextView =  Em.View.extend({
  classNameBindings: ['color:service-summary-component-red-dead:service-summary-component-green-live'],
  liveComponents: null,
  totalComponents: null,
  color: function() {
    return this.get("liveComponents") === 0;
  }.property("liveComponents")
});

App.MainDashboardServiceView = Em.View.extend({
  classNames: ['service', 'clearfix'],

  data: function () {
    return this.get('controller.data.' + this.get('serviceName'));
  }.property('controller.data'),

  dashboardMasterComponentView : Em.View.extend({
    didInsertElement: function() {
      App.tooltip($('[rel=healthTooltip]'));
    },
    templateName: require('templates/main/service/info/summary/master_components'),
    mastersComp: function () {
      return this.get('parentView.service.hostComponents').filterProperty('isMaster', true);
    }.property("service")
  }),

  formatUnavailable: function(value){
    return (value || value == 0) ? value : this.t('services.service.summary.notAvailable');
  },

  alertsCount: function () {
    return this.get('service.alertsCount');
  }.property('service.alertsCount'),

  hasCriticalAlerts: function () {
    return this.get('service.hasCriticalAlerts');
  }.property('service.hasCriticalAlerts'),

  isCollapsed: false,

  toggleInfoView: function () {
    this.$('.service-body').toggle('blind', 200);
    this.set('isCollapsed', !this.isCollapsed);
  },

  masters: function(){
    return this.get('service.hostComponents').filterProperty('isMaster', true);
  }.property('service'),

  clients: function(){
    var clients = this.get('service.hostComponents').filterProperty('isClient', true);
    var len = clients.length;
    var template = 'dashboard.services.{0}.client'.format(this.get('serviceName').toLowerCase());
    if(len > 1){
      template += 's';
    }

    return {
      title: this.t(template).format(len),
      component: clients.objectAt(0)
    };
  }.property('service')

});

});

require.register("views/main/service/services/flume", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
var date = require('utils/date');
var sort = require('views/common/sort_view');

App.MainDashboardServiceFlumeView = App.TableView.extend({
  templateName: require('templates/main/service/services/flume'),

  pagination: false,

  selectedHost: null,

  flumeAgentsCount:0,

  content: function () {
    var flumeAgents = this.get('service.agents'),
    hostNames = flumeAgents.mapProperty('hostName').uniq(),
    content = [],
    self = this;
    hostNames.forEach(function(hostName) {
      var agents = flumeAgents.filterProperty('hostName', hostName);
      content.push(
        Em.Object.create({
          hostName: hostName,
          agents: agents,
          rowspan: agents.length,
          firtstAgent: agents[0],
          otherAgents: agents.without(agents[0])
        })
      );
    });
    return content;
  }.property('App.FlumeAgent.@each.id', 'flumeAgentsCount'),



  summaryHeader: function () {
    var agents = App.FlumeService.find().objectAt(0).get('agents'),
    agentCount = agents.get('length'),
    hostCount = this.get('service.flumeHandlersTotal');
    return this.t("dashboard.services.flume.summary.title").format(hostCount, (hostCount > 1 ? "s" : ""), agentCount,  (agentCount > 1 ? "s" : ""));
  }.property('service.agents', 'service.hostComponents.length'),

  flumeHandlerComponent: function () {
    return Em.Object.create({
      componentName: 'FLUME_HANDLER'
    });
    //return App.HostComponent.find().findProperty('componentName', 'FLUME_HANDLER');
  }.property(),

  agentView: Em.View.extend({
    content: null,
    tagName: 'tr',

    click: function (e) {
      var numberOfAgents = this.get('content.agents').length;
      if($(e.target).attr('class') == "agent-host-name" || $(e.target).attr('class') == "agent-host-link"){
        var currentTargetRow = $(e.currentTarget);
        if ($(e.target).attr('class') == "agent-host-link") {
          currentTargetRow = currentTargetRow.parent(".agent-host-name").parent();
        }
        currentTargetRow.parents("table:first").find('tr').removeClass('highlight');
        currentTargetRow.addClass('highlight').nextAll("tr").slice(0,numberOfAgents - 1).addClass('highlight');
        this.get('parentView').showAgentInfo(this.get('content'));
      }
    }
  }),

  sortView: sort.wrapperView,

  hostSort: sort.fieldView.extend({
    column: '1',
    name: 'hostName',
    displayName: Em.I18n.t('common.host')
  }),

  didInsertElement: function () {
    this.filter();
  },

  /**
   * Change classes for dropdown DOM elements after status change of selected agent
   */
  setActionsDropdownClasses: function () {
    this.get('content').forEach(function(hosts){
      hosts.agents.forEach(function (agent) {
        agent.set('isStartAgentDisabled', agent.get('status') !== 'NOT_RUNNING');
        agent.set('isStopAgentDisabled', agent.get('status') !== 'RUNNING');
      });
    });
  }.observes('service.agents.@each.status'),

  updateFlumeAgentsCount: function () {
    this.set('flumeAgentsCount', this.get('service.agents.length'));
  }.observes('service.agents.length'),

  /**
   * Action handler from flume tepmlate.
   * Highlight selected row and show metrics graphs of selected agent.
   *
   * @method showAgentInfo
   * @param {object} agent
   */
  showAgentInfo: function (host) {
    this.set('selectedHost', host);
    this.setAgentMetrics(host);
  },
  /**
   * Show Flume agent metric.
   *
   * @method setFlumeAgentMetric
   * @param {object} agent - DS.model of agent
   */
  setAgentMetrics: function(host) {
    var getMetricTitle = function(metricTypeKey, hostName) {
      var metricType = Em.I18n.t('services.service.info.metrics.flume.' + metricTypeKey).format(Em.I18n.t('common.metrics'));
      return  metricType + ' - ' + hostName;
    };
    var gangliaUrlTpl = App.router.get('clusterController.gangliaUrl') + '/?r=hour&cs=&ce=&m=load_one&s=by+name&c=HDPFlumeServer&h={0}&host_regex=&max_graphs=0&tab=m&vn=&sh=1&z=small&hc=4';
    var agentHostMock = host.get('hostName');
    var mockMetricData = [
      {
        header: 'channelName',
        metricView: App.MainServiceInfoFlumeGraphsView.extend(),
        metricViewData: {
          agent: host,
          metricType: 'CHANNEL'
        }
      },
      {
        header: 'sinkName',
        metricView: App.MainServiceInfoFlumeGraphsView.extend(),
        metricViewData: {
          agent: host,
          metricType: 'SINK'
        }
      },
      {
        header: 'sourceName',
        metricView: App.MainServiceInfoFlumeGraphsView.extend(),
        metricViewData: {
          agent: host,
          metricType: 'SOURCE'
        }
      }
    ];
    mockMetricData.forEach(function(mockData, index) {
      mockData.header = getMetricTitle(mockData.header, agentHostMock);
      mockData.url = gangliaUrlTpl.format(agentHostMock);
      mockData.id = 'metric' + index;
      mockData.toggleIndex = '#' + mockData.id;
    });
    this.set('parentView.parentView.collapsedSections', mockMetricData);
  }
});

});

require.register("views/main/service/services/hbase", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
var date = require('utils/date');
var numberUtils = require('utils/number_utils');

App.MainDashboardServiceHbaseView = App.MainDashboardServiceView.extend({
  templateName: require('templates/main/service/services/hbase'),
  serviceName: 'hbase',
  /**
   * All master components
   */
  masters: function () {
    return this.get('service.hostComponents').filterProperty('isMaster', true);
  }.property('service.hostComponents.length'),
  /**
   * Passive master components
   */
  passiveMasters: function () {
    return this.get('masters').filterProperty('haStatus', 'false');
  }.property('masters'),



  regionServesText: function () {
    if (this.get('service.regionServersTotal') == 0) {
      return '';
    } else if (this.get('service.regionServersTotal') > 1) {
      return Em.I18n.t('services.service.summary.viewHosts');
    } else {
      return Em.I18n.t('services.service.summary.viewHost');
    }
  }.property("service"),

  /**
   * One(!) active master component
   */
  activeMaster: function () {
    return this.get('masters').findProperty('haStatus', 'true');
  }.property('masters'),

  activeMasterTitle: function(){
    return this.t('service.hbase.activeMaster');
  }.property('activeMaster'),

  masterServerHeapSummary: function () {
    var heapUsed = this.get('service').get('heapMemoryUsed');
    var heapMax = this.get('service').get('heapMemoryMax');
    var percent = heapMax > 0 ? 100 * heapUsed / heapMax : 0;
    var heapString = numberUtils.bytesToSize(heapUsed, 1, "parseFloat");
    var heapMaxString = numberUtils.bytesToSize(heapMax, 1, "parseFloat");
    return this.t('dashboard.services.hbase.masterServerHeap.summary').format(heapString, heapMaxString, percent.toFixed(1));
  }.property('service.heapMemoryUsed', 'service.heapMemoryMax'),

  summaryHeader: function () {
    var avgLoad = this.get('service.averageLoad');
    if (isNaN(avgLoad)) {
      avgLoad = this.t("services.service.summary.unknown");
    }
    return this.t("dashboard.services.hbase.summary").format(this.get('service.regionServersTotal'), avgLoad);
  }.property('service.regionServersTotal', 'service.averageLoad'),

  hbaseMasterWebUrl: function () {
    if (this.get('activeMaster.host') && this.get('activeMaster.host').get('publicHostName')) {
      return "http://" + (App.singleNodeInstall ? App.singleNodeAlias : this.get('activeMaster.host').get('publicHostName')) + ":60010";
    }
  }.property('activeMaster'),

  averageLoad: function () {
    var avgLoad = this.get('service.averageLoad');
    if (isNaN(avgLoad)) {
      avgLoad = this.t('services.service.summary.notAvailable');
    }
    return this.t('dashboard.services.hbase.averageLoadPerServer').format(avgLoad);
  }.property("service.averageLoad"),

  masterStartedTime: function () {
    var uptime = this.get('service').get('masterStartTime');
    if (uptime && uptime > 0) {
      var diff = App.dateTime() - uptime;
      if (diff < 0) {
        diff = 0;
      }
      var formatted = date.timingFormat(diff);
      return this.t('dashboard.services.uptime').format(formatted);
    }
    return this.t('services.service.summary.notRunning');
  }.property("service.masterStartTime"),

  masterActivatedTime: function () {
    var uptime = this.get('service').get('masterActiveTime');
    if (uptime && uptime > 0) {
      var diff = App.dateTime() - uptime;
      if (diff < 0) {
        diff = 0;
      }
      var formatted = date.timingFormat(diff);
      return this.t('dashboard.services.uptime').format(formatted);
    }
    return this.t('services.service.summary.notRunning');
  }.property("service.masterActiveTime"),

  regionServerComponent: function () {
    return Em.Object.create({
      componentName: 'HBASE_REGIONSERVER'
    });
    //return this.get('service.regionServers').objectAt(0);
  }.property()

});
});

require.register("views/main/service/services/hdfs", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
var date = require('utils/date');
var numberUtils = require('utils/number_utils');

App.MainDashboardServiceHdfsView = App.MainDashboardServiceView.extend({
  templateName: require('templates/main/service/services/hdfs'),
  serviceName: 'HDFS',
  Chart: App.ChartPieView.extend({
    service: null,
    color: '#0066B3',
    stroke: '#0066B3',
    palette: new Rickshaw.Color.Palette({
      scheme: [ 'rgba(0,102,179,0)', 'rgba(0,102,179,1)'].reverse()
    }),
    data: function () {
      var total = this.get('service.capacityTotal') + 0;
      var remaining = (this.get('service.capacityRemaining') + 0);
      var used = total - remaining;
      return [ used, remaining ];
    }.property('service.capacityUsed', 'service.capacityTotal')
  }),

  dashboardMasterComponentView: Em.View.extend({
    didInsertElement: function() {
      App.tooltip($('[rel=healthTooltip]'));
    },
    templateName: require('templates/main/service/info/summary/master_components'),
    mastersComp: function() {
      var masterComponents = [];
      var zkfcs = this.get('parentView.service.hostComponents').filterProperty('componentName', 'ZKFC');

      this.get('parentView.service.hostComponents').forEach(function (comp) {
        if (comp.get('isMaster') && comp.get('componentName') !== 'JOURNALNODE') {
          masterComponents.push(comp);
          var zkfc = zkfcs.findProperty('hostName', comp.get('hostName'));
          if (zkfc) {
            zkfc.set('isSubComponent', true);
            masterComponents.push(zkfc);
          }
        }
      });
      return masterComponents;
    }.property('parentView.service.hostComponents.length')
  }),

  didInsertElement: function() {
    App.tooltip($("[rel='tooltip']"));
  },
  dataNodesLive: function () {
    return this.get('service.dataNodesStarted');
  }.property('service.dataNodesStarted'),
  dataNodesDead: function () {
    return this.get('service.dataNodesInstalled');
  }.property('service.dataNodesInstalled'),

  showJournalNodes: function () {
    return this.get('service.journalNodes.length') > 0;
  }.property('service.journalNodes.length'),

  journalNodesLive: function () {
    return this.get('service.journalNodes').filterProperty("workStatus", "STARTED").get("length");
  }.property("service.journalNodes.@each.workStatus"),

  journalNodesTotal: function () {
    return this.get('service.journalNodes').get("length");
  }.property("service.journalNodes.length"),

  dfsTotalBlocks: function(){
    return this.formatUnavailable(this.get('service.dfsTotalBlocks'));
  }.property('service.dfsTotalBlocks'),
  dfsTotalFiles: function(){
    return this.formatUnavailable(this.get('service.dfsTotalFiles'));
  }.property('service.dfsTotalFiles'),
  dfsCorruptBlocks: function(){
    return this.formatUnavailable(this.get('service.dfsCorruptBlocks'));
  }.property('service.dfsCorruptBlocks'),
  dfsMissingBlocks: function(){
    return this.formatUnavailable(this.get('service.dfsMissingBlocks'));
  }.property('service.dfsMissingBlocks'),
  dfsUnderReplicatedBlocks: function(){
    return this.formatUnavailable(this.get('service.dfsUnderReplicatedBlocks'));
  }.property('service.dfsUnderReplicatedBlocks'),

  blockErrorsMessage: function() {
    return Em.I18n.t('dashboard.services.hdfs.blockErrors').format(this.get('dfsCorruptBlocks'), this.get('dfsMissingBlocks'), this.get('dfsUnderReplicatedBlocks'));
  }.property('dfsCorruptBlocks','dfsMissingBlocks','dfsUnderReplicatedBlocks'),

  nodeUptime: function () {
    var uptime = this.get('service').get('nameNodeStartTime');
    if (uptime && uptime > 0){
      var diff = App.dateTime() - uptime;
      if (diff < 0) {
        diff = 0;
      }
      var formatted = date.timingFormat(diff);
      return this.t('dashboard.services.uptime').format(formatted);
    }
    return this.t('services.service.summary.notRunning');
  }.property("service.nameNodeStartTime"),

  nodeWebUrl: function () {
    return "http://" + (App.singleNodeInstall ? App.singleNodeAlias :  this.get('service').get('nameNode').get('publicHostName')) + ":50070";
  }.property('service.nameNode'),

  nodeHeap: function () {
    var memUsed = this.get('service').get('jvmMemoryHeapUsed');
    var memMax = this.get('service').get('jvmMemoryHeapMax');
    var percent = memMax > 0 ? ((100 * memUsed) / memMax) : 0;
    return this.t('dashboard.services.hdfs.nodes.heapUsed').format(
      numberUtils.bytesToSize(memUsed, 1, 'parseFloat'),
      numberUtils.bytesToSize(memMax, 1, 'parseFloat'),
      percent.toFixed(1));
  }.property('service.jvmMemoryHeapUsed', 'service.jvmMemoryHeapMax'),

  summaryHeader: function () {
    var text = this.t("dashboard.services.hdfs.summary");
    var service = this.get('service');
    var liveCount = service.get('dataNodesStarted');
    var totalCount = service.get('dataNodesTotal');
    var total = service.get('capacityTotal') + 0;
    var remaining = service.get('capacityRemaining') + 0;
    var used = total - remaining;
    var percent = total > 0 ? ((used * 100) / total).toFixed(1) : 0;
    if (percent == "NaN" || percent < 0) {
      percent = Em.I18n.t('services.service.summary.notAvailable') + " ";
    }
    return text.format(liveCount, totalCount, percent);
  }.property('service.dataNodesStarted', 'service.dataNodesTotal', 'service.capacityUsed', 'service.capacityTotal'),

  dfsUsedDisk: function () {
    var text = this.t("dashboard.services.hdfs.capacityUsed");
    var total = this.get('service.capacityTotal');
    var dfsUsed = this.get('service.capacityUsed');
    var percent = total > 0 ? ((dfsUsed * 100) / total).toFixed(2) : 0;
    if (percent == "NaN" || percent < 0) {
      percent = Em.I18n.t('services.service.summary.notAvailable') + " ";
    }
    return text.format(numberUtils.bytesToSize(dfsUsed, 1, 'parseFloat'), numberUtils.bytesToSize(total, 1, 'parseFloat'), percent);
  }.property('service.capacityUsed', 'service.capacityTotal'),
  nonDfsUsedDisk: function () {
    var text = this.t("dashboard.services.hdfs.capacityUsed");
    var total = this.get('service.capacityTotal');
    var remaining = this.get('service.capacityRemaining');
    var dfsUsed = this.get('service.capacityUsed');
    var nonDfsUsed = total !== null && remaining !== null && dfsUsed !== null ? total - remaining - dfsUsed : null;
    var percent = total > 0 ? ((nonDfsUsed * 100) / total).toFixed(2) : 0;
    if (percent == "NaN" || percent < 0) {
      percent = Em.I18n.t('services.service.summary.notAvailable') + " ";
    }
    return text.format(numberUtils.bytesToSize(nonDfsUsed, 1, 'parseFloat'), numberUtils.bytesToSize(total, 1, 'parseFloat'), percent);
  }.property('service.capacityUsed', 'service.capacityRemaining', 'service.capacityTotal'),
  remainingDisk: function () {
    var text = this.t("dashboard.services.hdfs.capacityUsed");
    var total = this.get('service.capacityTotal');
    var remaining = this.get('service.capacityRemaining');
    var percent = total > 0 ? ((remaining * 100) / total).toFixed(2) : 0;
    if (percent == "NaN" || percent < 0) {
      percent = Em.I18n.t('services.service.summary.notAvailable') + " ";
    }
    return text.format(numberUtils.bytesToSize(remaining, 1, 'parseFloat'), numberUtils.bytesToSize(total, 1, 'parseFloat'), percent);
  }.property('service.capacityRemaining', 'service.capacityTotal'),

  dataNodeComponent: function () {
    return Em.Object.create({
      componentName: 'DATANODE'
    });
    //return this.get('service.dataNodes').objectAt(0);
  }.property(),

  journalNodeComponent: function () {
    return this.get('service.journalNodes').objectAt(0);
  }.property(),

  safeModeStatus: function () {
    var safeMode = this.get('service.safeModeStatus');
    if (safeMode == null) {
      return Em.I18n.t("services.service.summary.notAvailable");
    } else if (safeMode.length == 0) {
      return Em.I18n.t("services.service.summary.safeModeStatus.notInSafeMode");
    } else {
      return Em.I18n.t("services.service.summary.safeModeStatus.inSafeMode");
    }
  }.property('service.safeModeStatus'),
  upgradeStatus: function () {
    var upgradeStatus = this.get('service.upgradeStatus');
    var healthStatus = this.get('service.healthStatus');
    if (upgradeStatus) {
      return Em.I18n.t('services.service.summary.pendingUpgradeStatus.notPending');
    } else if (healthStatus == 'green') {
      return Em.I18n.t('services.service.summary.pendingUpgradeStatus.notFinalized');
    } else {
      return Em.I18n.t("services.service.summary.notAvailable");
    }
  }.property('service.upgradeStatus', 'service.healthStatus'),
  isUpgradeStatusWarning: function () {
    var upgradeStatus = this.get('service.upgradeStatus');
    var healthStatus = this.get('service.healthStatus');
    return !upgradeStatus && healthStatus == 'green';
  }.property('service.upgradeStatus', 'service.healthStatus')

});

});

require.register("views/main/service/services/hive", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainDashboardServiceHiveView = App.MainDashboardServiceView.extend({
  templateName: require('templates/main/service/services/hive'),
  serviceName: 'hive',

  titleMasters: function(){
    var masters = this.get('masters');
    return [masters.findProperty('componentName', 'HIVE_SERVER'), masters.findProperty('componentName', 'HIVE_METASTORE')];
  }.property('service')
});
});

require.register("views/main/service/services/mapreduce2", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.MainDashboardServiceMapreduce2View = App.MainDashboardServiceView.extend({
    templateName: require('templates/main/service/services/mapreduce2'),
    serviceName: 'MAPREDUCE2',

    titleInfo: function(){
      var hs = this.get('service.hostComponents').findProperty('componentName', 'HISTORYSERVER');
      if (!hs) {
        return this.t('services.mapreduce2.history.unknown');
      } else if (hs.get('workStatus') === App.HostComponentStatus.started) {
        return this.t('services.mapreduce2.history.running');
      } else {
        return this.t('services.mapreduce2.history.stopped');
      }
    }.property('service')
  });
});

require.register("views/main/service/services/oozie", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainDashboardServiceOozieView = App.MainDashboardServiceView.extend({
  serviceName: 'oozie',
  templateName: require('templates/main/service/services/oozie'),

  webUi: function () {
    var hostName = App.singleNodeInstall ? App.singleNodeAlias : this.get('service.hostComponents').findProperty('componentName', 'OOZIE_SERVER').get('host.publicHostName');
    return "http://{0}:11000/oozie".format(hostName);
  }.property('service')
});
});

require.register("views/main/service/services/ranger", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.MainDashboardServiceRangerView = App.MainDashboardServiceView.extend({

  templateName: require('templates/main/service/services/ranger'),

  serviceName: 'RANGER',

  didInsertElement: function () {
    this.set('controller.isRangerUpdateWorking', true);
    App.updater.run(this.get('controller'), 'updateRangerPluginsStatus', 'isRangerUpdateWorking', App.bgOperationsUpdateInterval);
    App.updater.immediateRun('updateRangerPluginsStatus');
  },

  willDestroyElement: function () {
    this.set('controller.isRangerUpdateWorking', false);
  }

});

});

require.register("views/main/service/services/storm", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var date = require('utils/date');

App.MainDashboardServiceStormView = App.MainDashboardServiceView.extend({
  templateName: require('templates/main/service/services/storm'),
  serviceName: 'STORM',

  /**
   * this parameter is used to fiter hosts by component name
   * used in mainHostController.filterByComponent() method
   */
  filterComponent: function() {
    return Em.Object.create({componentName: 'SUPERVISOR'});
  }.property(),

  freeSlotsPercentage: function() {
    return Math.round(this.get('service.freeSlots')/this.get('service.totalSlots')*100);
  }.property('service.freeSlots', 'service.totalSlots'),

  superVisorsLive: function () {
    return this.get('service.superVisorsStarted');
  }.property('service.superVisorsStarted'),

  superVisorsTotal: function() {
    return this.get('service.superVisorsTotal');
  }.property('service.superVisorsTotal'),

  nimbusUptimeFormatted: function() {
    return this.get('service.nimbusUptime') || Em.I18n.t('services.service.summary.notRunning');
  }.property('service.nimbusUptime')
});

});

require.register("views/main/service/services/yarn", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');
var date = require('utils/date');
var numberUtils = require('utils/number_utils');

App.MainDashboardServiceYARNView = App.MainDashboardServiceView.extend({
  templateName: require('templates/main/service/services/yarn'),
  serviceName: 'YARN',

  nodeHeap: function () {
    var memUsed = this.get('service').get('jvmMemoryHeapUsed');
    var memMax = this.get('service').get('jvmMemoryHeapMax');
    var percent = memMax > 0 ? ((100 * memUsed) / memMax) : 0;
    return this.t('dashboard.services.hdfs.nodes.heapUsed').format(
      numberUtils.bytesToSize(memUsed, 1, 'parseFloat'),
      numberUtils.bytesToSize(memMax, 1, 'parseFloat'),
      percent.toFixed(1));
  }.property('service.jvmMemoryHeapUsed', 'service.jvmMemoryHeapMax'),

  summaryHeader: function () {
    var text = this.t("dashboard.services.yarn.summary");
    var totalCount = this.get('service.nodeManagersTotal');
    var liveCount = this.get('service.nodeManagersStarted');
    return text.format(liveCount, totalCount);
  }.property('service.nodeManagersStarted', 'service.nodeManagersTotal'),

  nodeManagerComponent: function () {
    return Em.Object.create({
      componentName: 'NODEMANAGER'
    });
    //return this.get('service.nodeManagerNodes').objectAt(0);
  }.property(),

  yarnClientComponent: function () {
    return Em.Object.create({
      componentName: 'YARN_CLIENT'
    });
    //return this.get('service.hostComponents').findProperty('componentName', 'YARN_CLIENT');
  }.property(),

  hasManyYarnClients: function () {
    return (this.get('service.installedClients') > 1);
  }.property('service.installedClients'),

  nodeUptime: function () {
    var uptime = this.get('service').get('resourceManagerStartTime');
    if (uptime && uptime > 0){
      var diff = App.dateTime() - uptime;
      if (diff < 0) {
        diff = 0;
      }
      var formatted = date.timingFormat(diff);
      return this.t('dashboard.services.uptime').format(formatted);
    }
    return this.t('services.service.summary.notRunning');
  }.property("service.resourceManagerStartTime"),

  nodeManagerText: function () {
    if (this.get("service.nodeManagersTotal") === 0) {
      return '';
    } else if (this.get("service.nodeManagersTotal") > 1) {
      return Em.I18n.t('services.service.summary.viewHosts');
    } else {
      return Em.I18n.t('services.service.summary.viewHost');
    }
  }.property("service.nodeManagersTotal"),

  nodeManagersStatus: function () {
    var nmActive = this.get('service.nodeManagersCountActive');
    var nmLost = this.get('service.nodeManagersCountLost');
    var nmUnhealthy = this.get('service.nodeManagersCountUnhealthy');
    var nmRebooted = this.get('service.nodeManagersCountRebooted');
    var nmDecom = this.get('service.nodeManagersCountDecommissioned');
    return this.t('dashboard.services.yarn.nodeManagers.status.msg').format(
      this.formatUnavailable(nmActive),
      this.formatUnavailable(nmLost),
      this.formatUnavailable(nmUnhealthy),
      this.formatUnavailable(nmRebooted),
      this.formatUnavailable(nmDecom)
    );
  }.property('service.nodeManagersCountActive', 'service.nodeManagersCountLost',
      'service.nodeManagersCountUnhealthy', 'service.nodeManagersCountRebooted', 'service.nodeManagersCountDecommissioned'),

  containers: function () {
    var allocated = this.get('service.containersAllocated');
    var pending = this.get('service.containersPending');
    var reserved = this.get('service.containersReserved');
    return this.t('dashboard.services.yarn.containers.msg').format(
      this.formatUnavailable(allocated),
      this.formatUnavailable(pending),
      this.formatUnavailable(reserved)
    );
  }.property('service.containersAllocated', 'service.containersPending', 'service.containersReserved'),

  apps: function () {
    var appsSubmitted = this.get('service.appsSubmitted');
    var appsRunning = this.get('service.appsRunning');
    var appsPending = this.get('service.appsPending');
    var appsCompleted = this.get('service.appsCompleted');
    var appsKilled = this.get('service.appsKilled');
    var appsFailed = this.get('service.appsFailed');
    return this.t('dashboard.services.yarn.apps.msg').format(
      this.formatUnavailable(appsSubmitted),
      this.formatUnavailable(appsRunning),
      this.formatUnavailable(appsPending),
      this.formatUnavailable(appsCompleted),
      this.formatUnavailable(appsKilled),
      this.formatUnavailable(appsFailed));
  }.property('service.appsSubmitted', 'service.appsRunning', 'service.appsPending', 'service.appsCompleted', 'service.appsKilled', 'service.appsFailed'),

  memory: function () {
    return Em.I18n.t('dashboard.services.yarn.memory.msg').format(
        numberUtils.bytesToSize(this.get('service.allocatedMemory'), 1, 'parseFloat', 1024 * 1024),
        numberUtils.bytesToSize(this.get('service.reservedMemory'), 1, 'parseFloat', 1024 * 1024),
        numberUtils.bytesToSize(this.get('service.availableMemory'), 1, 'parseFloat', 1024 * 1024));
  }.property('service.allocatedMemory', 'service.reservedMemory', 'service.availableMemory'),

  queues: function() {
    return Em.I18n.t('dashboard.services.yarn.queues.msg').format(this.formatUnavailable(this.get('service.queuesCount')));
  }.property('service.queuesCount'),

  didInsertElement: function(){
    App.tooltip($("[rel='queue-tooltip']"), {html: true, placement: "right"});
  }

});

});

require.register("views/main/service/services/zookeeper", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainDashboardServiceZookeperView = App.MainDashboardServiceView.extend({
  templateName: require('templates/main/service/services/zookeeper'),
  serviceName: 'zookeeper',

  titleInfo: function(){
    var components = this.get('service.hostComponents').filterProperty('componentName', 'ZOOKEEPER_SERVER');
    var running = 0;
    components.forEach(function(item){
      if(item.get('workStatus') === App.HostComponentStatus.started){
        running += 1;
      }
    });

    return {
      pre: this.t('services.zookeeper.prefix').format(running),
      title: this.t('services.zookeeper.title').format(components.length),
      component: components.objectAt(0)
    };
  }.property('service')
});
});

require.register("views/main/views/details", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainViewsDetailsView = Em.View.extend({

  name: "mainViewsDetailsView",

  tagName: "iframe",
  classNames: ["views_sizes"],
  attributeBindings: ['src','seamless'],
  seamless: "seamless",

  interval: null,

  /**
   * Drop autoHeight timer
   */
  willDestroyElement: function() {
    var interval = this.get('interval');
    if (interval) {
      clearInterval(interval);
    }
  },

  /**
   * Updates iframe height every 5s
   */
  didInsertElement : function() {
    var interval, self = this;
    interval = setInterval(function() {
      self.resizeFunction();
    }, 5000);
    self.set('interval', interval);
    this.resizeFunction();
  },

  resizeFunction : function() {
    var body = $(document.body);
    var footer = $("footer", body);
    var header = $("#top-nav", body);
    var iframe = $("iframe", body);

    var bodyHeight = body.outerHeight();
    var footerHeight = footer != null ? footer.outerHeight() : 0;
    var headerHeight = header != null ? header.outerHeight() : 0;

    var defaultHeight = bodyHeight - footerHeight - headerHeight;
    console.debug("IFrame default Height = " + defaultHeight + " ("
        + bodyHeight + " - " + headerHeight + " - " + footerHeight + ")");

    if (iframe != null && iframe.length > 0) {
      var childrenHeight = 0;
      var iframeElement = iframe[0];
      var pageScrollTop = $(window).scrollTop();
      // set iframe height to 'auto' to get actual scrollHeight
      iframeElement.style.height = 'auto';
      if (iframeElement.contentWindow != null
          && iframeElement.contentWindow.document != null
          && iframeElement.contentWindow.document.body != null) {
        var iFrameContentBody = iframeElement.contentWindow.document.body;
        childrenHeight = iFrameContentBody.scrollHeight;
        console.debug("IFrame content Height = " + childrenHeight);
      }
      var iFrameHeight = Math.max(childrenHeight, defaultHeight);
      console.debug("IFrame final height = ", iFrameHeight);
      iframe.css('height', iFrameHeight);
      $(window).scrollTop(pageScrollTop);
    }
  },

  src: function() {
    // can't use window.location.origin because IE doesn't support it
    return window.location.protocol + '//' + window.location.host + this.get('controller.content.href') + this.get('controller.content.viewPath');
  }.property('controller.content')

});

});

require.register("views/main/views_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.MainViewsView = Em.View.extend({

  name: "mainViewsView",
  templateName: require('templates/main/views'),
  classNames: [""],

  views: function() {
    return App.router.get('mainViewsController.ambariViews');
  }.property('App.router.mainViewsController.ambariViews.@each'),

  selectedViewBinding: 'controller.selectedView'

});
});

require.register("views/wizard/step0_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.WizardStep0View = Em.View.extend({

  tagName: "form",

  attributeBindings: ['autocomplete'],

  /**
   * Disable autocomplete for form's fields
   */
  autocomplete: 'off',

  templateName: require('templates/wizard/step0'),

  didInsertElement: function () {
    App.popover($("[rel=popover]"), {'placement': 'right', 'trigger': 'hover'});
    this.get('controller').loadStep();
  },

  /**
   * Is some error with provided cluster name
   * @type {bool}
   */
  onError: function () {
    return !Em.isEmpty(this.get('controller.clusterNameError'));
  }.property('controller.clusterNameError')

});

/**
 * Field for cluster name
 * @type {Ember.TextField}
 */
App.WizardStep0ViewClusterNameInput = Em.TextField.extend({

  /**
   * Submit form if "Enter" pressed
   * @method keyPress
   * @param {object} event
   * @returns {bool}
   */
  keyPress: function(event) {
    if (event.keyCode == 13) {
      this.get('parentView.controller').submit();
      return false;
    }
    return true;
  }
});

});

require.register("views/wizard/step10_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.WizardStep10View = Em.View.extend({

  templateName: require('templates/wizard/step10'),

  didInsertElement: function () {
    this.get('controller').loadStep();
  },

  serviceRestartText: function () {
    var controller = this.get('controller');
    if (controller.get('isAddServiceWizard'))
      return Em.I18n.t('common.important') + Em.I18n.t('installer.step10.staleServicesRestartRequired');
    else
      return '';
  }.property('controller.isAddServiceWizard')
});
});

require.register("views/wizard/step1_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

var App = require('app');

App.WizardStep1View = Em.View.extend({

  templateName: require('templates/wizard/step1'),

  /**
   * Is Repositories Accordion collapsed
   * @type {bool}
   */
  isRLCollapsed: true,

  didInsertElement: function () {
    if (this.get('isRLCollapsed')) {
      this.$('.accordion-body').hide();
    }
    $("[rel=skip-validation-tooltip]").tooltip({ placement: 'right'});
  },

  /**
   * List of available stacks
   * @type {Em.Object[]}
   */
  stacks: function () {
    return this.get('controller.content.stacks').toArray().map(function (stack) {
      return Em.Object.create({
        name: stack.get('id').replace('-', ' '),
        isSelected: stack.get('isSelected')
      });
    });
  }.property('controller.selectedStack'),

  operatingSystems: function () {
    var selectedStack = this.get('controller.selectedStack');
    var result = [];
    if (!!selectedStack)
      result = selectedStack.get('operatingSystems');
    return result;
  }.property('controller.selectedStack'),

  /**
   * List of all repositories under selected stack operatingSystems
   * API and ember data model structure:
   * stack = [{OS-1},{OS-2}]
   * OS-1 = [{repository-1},{repository-2}]
   * OS-2 = [{repository-3},{repository-4}]
   * @return: [{repository-1},{repository-2},{repository-3},{repository-4}]
   */
  allRepositories: function () {
    var selectedStack = this.get('controller.selectedStack');
    var result = [];
    if (!!selectedStack)
      result = selectedStack.get('repositories');
    return result;
  }.property('controller.selectedStack'),

  /**
   * Verify if some repo has empty base-url
   * @type {bool}
   */
  emptyRepoExist: function () {
    return this.get('allRepositories').someProperty('emptyError', true);
  }.property('allRepositories.@each.emptyError'),

  /**
   * Disable submit button flag
   * @type {bool}
   */
  isSubmitDisabled: function () {
    return this.get('emptyRepoExist') || this.get('isNoOsChecked') || this.get('invalidUrlExist');
  }.property('emptyRepoExist', 'isNoOsChecked', 'invalidUrlExist'),

  /**
   * Verify if some invalid repo-urls exist
   * @type {bool}
   */
  invalidUrlExist: function () {
    return this.get('allRepositories').someProperty('validation', App.Repository.validation['INVALID']);
  }.property('allRepositories.@each.validation'),

  /**
   * If all repo links are unchecked
   * @type {bool}
   */
  isNoOsChecked: function () {
    return this.get('operatingSystems').everyProperty('isSelected', false);
  }.property('operatingSystems.@each.isSelected'),

  /**
   * Overall errors count
   * @type {number}
   */
  totalErrorCnt: function () {
    var emptyCnt = this.get('allRepositories').filterProperty('emptyError').length;
    var invalidCnt = this.get('allRepositories').filterProperty('validation', App.Repository.validation['INVALID']).length;
    if (this.get('isNoOsChecked')) {
      return 1;
    } else if (emptyCnt || invalidCnt) {
      return emptyCnt + invalidCnt;
    } else {
      return 0;
    }
  }.property('allRepositories.@each.emptyError', 'isNoOsChecked', 'allRepositories.@each.validation'),

  /**
   * Checkbox for each stack
   * @type {Ember.Checkbox}
   */
  stackRadioButton: Em.Checkbox.extend({
    tagName: 'input',
    attributeBindings: [ 'type', 'checked' ],
    checked: function () {
      return this.get('content.isSelected');
    }.property('content.isSelected'),
    type: 'radio',

    click: function () {
      this.get('controller.content.stacks').setEach('isSelected', false);
      this.get('controller.content.stacks').findProperty('id', this.get('content.name').replace(' ', '-')).set('isSelected', true);
    }
  }),

  /**
   * Popover for repo-url error indicator
   * @type {Em.View}
   */
  popoverView: Em.View.extend({
    tagName: 'i',
    classNameBindings: ['repository.validation'],
    attributeBindings: ['repository.errorTitle:title', 'repository.errorContent:data-content'],
    didInsertElement: function () {
      App.popover($(this.get('element')), {'trigger': 'hover'});
    }
  }),

  /**
   * Onclick handler for Config Group Header. Used to show/hide block
   * @method onToggleBlock
   */
  onToggleBlock: function () {
    this.$('.accordion-body').toggle('blind', 500);
    this.set('isRLCollapsed', !this.get('isRLCollapsed'));
  },

  /**
   * Onclick handler for recheck repos urls. Used in Advanced Repository Options.
   */
  retryRepoUrls: function () {
    App.router.get('installerController').checkRepoURL(this.get('controller'));
  },


  /**
   * Onclick handler for checkbox of each repo group
   * @method updateByCheckbox
   */
  updateByCheckbox: function () {
    //upload to content
    var operatingSystems = this.get('operatingSystems');
    if (operatingSystems) {
      operatingSystems.forEach(function (os) {
        if (!os.get('isSelected')) {
          os.get('repositories').forEach(function (repository) {
            repository.set('baseUrl', repository.get('latestBaseUrl'));
            repository.set('validation', App.Repository.validation['PENDING']);
          });
        } else {
          os.get('repositories').forEach(function (repository) {
            if (this.get('controller.skipValidationChecked')) {
              repository.set('validation', App.Repository.validation['PENDING']);
            }
          }, this);
        }
      }, this);
    }
  }.observes('operatingSystems.@each.isSelected', 'controller.skipValidationChecked'),

  /**
   * Onclick handler for undo action of each repo group
   * @method undoGroupLocalRepository
   * @param {object} event
   */
  undoGroupLocalRepository: function (event) {
    event.context.set('baseUrl', event.context.get('latestBaseUrl'));
    event.context.set('validation', App.Repository.validation['PENDING']);
  },

  /**
   * Handler for clear icon click
   * @method clearGroupLocalRepository
   * @param {object} event
   */
  clearGroupLocalRepository: function (event) {
    if (!event.context.get('isSelected')) {
      return;
    }
    event.context.set('baseUrl', '');
    event.context.set('validation', App.Repository.validation['PENDING']);
  },

  /**
   * Handler when editing any repo BaseUrl
   * @method editLocalRepository
   */
  editLocalRepository: function () {
    //upload to content
    var repositories = this.get('allRepositories');
    repositories.forEach(function (repository) {
      if (repository.get('lastBaseUrl') != repository.get('baseUrl')) {
        repository.set('lastBaseUrl', repository.get('baseUrl'));
        repository.set('validation', App.Repository.validation['PENDING']);
      }
    }, this);
  }.observes('allRepositories.@each.baseUrl')

});

});

require.register("views/wizard/step2_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * 'License'); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an 'AS IS' BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.SshKeyFileUploader = Em.View.extend({
  //TODO: rewrite it using tagName and attribute binding
  //TODO: rewrite it as independent component and place it somewhere in utils
  // alternative is to move it to App.WizardStep2View
  template: Em.Handlebars.compile('<input type="file" {{bindAttr disabled="view.disabled"}} />'),

  classNames: ['ssh-key-input-indentation'],

  change: function (e) {
    var self = this;
    if (e.target.files && e.target.files.length == 1) {
      var file = e.target.files[0];
      var reader = new FileReader();

      reader.onload = (function () {
        return function (e) {
          $('#sshKey').html(e.target.result);
          self.get("controller").setSshKey(e.target.result);
        };
      })(file);
      reader.readAsText(file);
    }
  }

});

App.WizardStep2View = Em.View.extend({

  templateName: require('templates/wizard/step2'),

  didInsertElement: function () {
    App.popover($("[rel=popover]"), {'placement': 'right', 'trigger': 'hover'});
    App.tooltip($("[rel=tooltip]"), {'placement': 'top', 'trigger': 'hover'});
    //todo: move them to conroller
    this.set('controller.hostsError', null);
    this.set('controller.sshKeyError', null);
  },

  /**
   * Is manualInstall selected
   * @type {bool}
   */
  sshKeyState: function () {
    return this.get("controller.content.installOptions.manualInstall");
  }.property("controller.content.installOptions.manualInstall"),

  /**
   * Is File API available
   * @type {bool}
   * TODO: incupsulate it inside of App.SshKeyFileUploader
   */
  isFileApi: function () {
    /* istanbul ignore next */
    return window.File && window.FileReader && window.FileList;
  }.property(),

  /**
   * Checkbox for activate SSH fields
   * @type {Ember.Checkbox}
   * TODO: replace next 2 properties with new one used in both places
   */
  providingSSHKeyRadioButton: Em.Checkbox.extend({

    tagName: 'input',

    attributeBindings: ['type', 'checked'],

    classNames: ['radio-btn-provide-ssh-key'],

    checked: function () {
      return this.get('controller.content.installOptions.useSsh');
    }.property('controller.content.installOptions.useSsh'),

    type: 'radio',

    click: function () {
      this.set('controller.content.installOptions.useSsh', true);
      this.set('controller.content.installOptions.manualInstall', false);
    }
  }),

  /**
   * Checkbox for manual registration
   * @type {Ember.Checkbox}
   */
  manualRegistrationRadioButton: Em.Checkbox.extend({
    tagName: 'input',

    attributeBindings: ['type', 'checked'],

    classNames: ['radio-btn-manual-reg'],

    type: 'radio',

    checked: function () {
      return this.get('controller.content.installOptions.manualInstall');
    }.property('controller.content.installOptions.manualInstall'),

    click: function () {
      this.set('controller.content.installOptions.manualInstall', true);
      this.set('controller.content.installOptions.useSsh', false);
    }
  }),

  /**
   * Textarea with ssh-key
   * @type {Ember.TextField}
   */
  textFieldView: Em.TextField.extend({

    /**
     * Is textfield disabled
     * @type {bool}
     */
    disabled: function () {
      return !this.get('isEnabled');
    }.property('isEnabled')
  })

});



});

require.register("views/wizard/step3/hostLogPopupBody_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.WizardStep3HostLogPopupBody = Em.View.extend({

  templateName: require('templates/wizard/step3/step3_host_log_popup'),

  /**
   * Host's boot log
   * @type {string}
   */
  bootLog: function() {
    return this.get('parentView.host.bootLog');
  }.property('parentView.host.bootLog'),

  /**
   * Is textarea view active
   * @type {bool}
   */
  isTextArea: false,

  /**
   * Textbox with host's boot log
   * @type {Ember.TextArea}
   */
  textArea: Em.TextArea.extend({

    didInsertElement: function () {
      /* istanbul ignore next: simple DOM manipulations */
      $(this.get('element'))
        .width($(this.get('parentView').get('element')).width() - 10)
        .height($(this.get('parentView').get('element')).height())
        .select()
        .css('resize', 'none');
    },

    /**
     * Edit disabled
     * @type {bool}
     */
    readOnly: true,

    /**
     * <code>parentView.bootLog</code>
     * @type {string}
     */
    value: function () {
      return this.get('content');
    }.property('content')

  }),

  didInsertElement: function () {
    var self = this;
    var button = $(this.get('element')).find('.textTrigger');
    button.click(function () {
      $(this).text(self.get('isTextArea') ? Em.I18n.t('installer.step3.hostLogPopup.highlight') : Em.I18n.t('installer.step3.hostLogPopup.copy'));
      self.set('isTextArea', !self.get('isTextArea'));
    });
    /* istanbul ignore next: difficult to test */
    $(this.get('element')).find('.content-area')
      .mouseenter(
      function () {
        $(this).css('border', '1px solid #dcdcdc');
        button.css('visibility', 'visible');
      })
      .mouseleave(
      function () {
        $(this).css('border', 'none');
        button.css('visibility', 'hidden');
      });
  }

});

});

require.register("views/wizard/step3/hostWarningPopupBody_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var lazyloading = require('utils/lazy_loading');
var numberUtils = require('utils/number_utils');

App.WizardStep3HostWarningPopupBody = Em.View.extend({

  templateName: require('templates/wizard/step3/step3_host_warnings_popup'),

  classNames: ['host-check'],

  bodyControllerBinding: 'App.router.wizardStep3Controller',

  /**
   * Listbox with host filters
   * @type {Ember.Select}
   */
  hostSelectView: Em.Select.extend({

    selectionBinding: "parentView.category",

    /**
     * Content has default value "All Hosts" to bind selection to category
     * @type {string[]}
     */
    content: ['All Hosts'],

    /**
     * List of filtered hostnames
     * @type {string[]}
     */
    hosts: function () {
      var warningsByHost = this.get('parentView.warningsByHost');
      if (Em.isNone(warningsByHost)) return [];
      return warningsByHost.mapProperty('name');
    }.property('parentView.warningsByHost'),

    /**
     * Is data loaded
     * @type {bool}
     */
    isLoaded: false,

    didInsertElement: function () {
      this.initContent();
    },

    /**
     * Check browser and set content for listbox
     * @method initContent
     */
    initContent: function () {
      this.set('isLoaded', false);
      //The lazy loading for select elements supported only by Firefox and Chrome
      var isBrowserSupported = $.browser.mozilla || ($.browser.safari && navigator.userAgent.indexOf('Chrome') !== -1);
      var isLazyLoading = isBrowserSupported && this.get('hosts').length > 100;
      this.set('isLazyLoading', isLazyLoading);
      if (isLazyLoading) {
        //select need at least 30 hosts to have scrollbar
        this.set('content', this.get('hosts').slice(0, 30));
      } else {
        this.set('content', this.get('hosts'));
        this.set('isLoaded', true);
      }
    }.observes('parentView.warningsByHost'),

    /**
     * On click start lazy loading
     * @method click
     */
    click: function () {
      if (!this.get('isLoaded') && this.get('isLazyLoading')) {
        //filter out hosts, which already pushed in select
        var source = this.get('hosts').filter(function (_host) {
          return !this.get('content').contains(_host);
        }, this).slice();
        lazyloading.run({
          destination: this.get('content'),
          source: source,
          context: this,
          initSize: 30,
          chunkSize: 200,
          delay: 50
        });
      }
    }
  }),

  /**
   * List of warnings grouped by host
   * Same to <code>bodyController.warningsByHost</code>
   * @type {Ember.Enumerable}
   */
  warningsByHost: function () {
    return this.get('bodyController.warningsByHost');
  }.property('bodyController.warningsByHost'),

  /**
   * List of all warnings
   * Same to <code>bodyController.warnings</code>
   * @type {Ember.Enumerable}
   */
  warnings: function () {
    return this.get('bodyController.warnings');
  }.property('bodyController.warnings'),

  /**
   * Selected category
   * @type {string}
   */
  category: 'All Hosts',

  /**
   * List of warnings for selected <code>category</code>
   * @type {Ember.Enumerable}
   */
  categoryWarnings: function () {
    var warningsByHost = this.get('warningsByHost');
    if (Em.isNone(warningsByHost)) return [];
    var category = warningsByHost.findProperty('name', this.get('category'));
    return Em.isNone(category) ? []: category.warnings;
  }.property('warningsByHost', 'category'),

  /**
   * List of warnings grouped by <code>category</code>
   * @type {Ember.Object[]}
   */
  content: function () {
    var hostCheckWarnings = this.get('bodyController.hostCheckWarnings');
    var repoCategoryWarnings = this.get('bodyController.repoCategoryWarnings');
    var diskCategoryWarnings = this.get('bodyController.diskCategoryWarnings');
    var jdkCategoryWarnings = this.get('bodyController.jdkCategoryWarnings') || [];
    var thpCategoryWarnings = this.get('bodyController.thpCategoryWarnings');
    var categoryWarnings = this.get('categoryWarnings');
    var warningsArray = [
      Em.Object.create({
        warnings: thpCategoryWarnings,
        title: Em.I18n.t('installer.step3.hostWarningsPopup.thp'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.thp.message'),
        type: Em.I18n.t('common.issues'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.thp'),
        action: Em.I18n.t('installer.step3.hostWarningsPopup.action.enabled'),
        category: 'thp'
      }),
      Em.Object.create({
        warnings: jdkCategoryWarnings,
        title: Em.I18n.t('installer.step3.hostWarningsPopup.jdk'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.jdk.message'),
        type: Em.I18n.t('common.issues'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.jdk'),
        action: Em.I18n.t('installer.step3.hostWarningsPopup.action.exists'),
        category: 'jdk'
      }),
      Em.Object.create({
        warnings: diskCategoryWarnings,
        title: Em.I18n.t('installer.step3.hostWarningsPopup.disk'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.disk.message'),
        type: Em.I18n.t('common.issues'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.disk'),
        action: Em.I18n.t('installer.step3.hostWarningsPopup.action.exists'),
        category: 'disk'
      }),
      Em.Object.create({
        warnings: repoCategoryWarnings,
        title: Em.I18n.t('installer.step3.hostWarningsPopup.repositories'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.repositories.message'),
        type: Em.I18n.t('common.issues'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.repositories'),
        action: Em.I18n.t('installer.step3.hostWarningsPopup.action.invalid'),
        category: 'repositories'
      }),
      Em.Object.create({
        warnings: categoryWarnings.filterProperty('category', 'firewall'),
        title: Em.I18n.t('installer.step3.hostWarningsPopup.firewall'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.firewall.message'),
        type: Em.I18n.t('common.issues'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.firewall'),
        action: Em.I18n.t('installer.step3.hostWarningsPopup.action.running'),
        category: 'firewall'
      }),
      Em.Object.create({
        warnings: categoryWarnings.filterProperty('category', 'processes'),
        title: Em.I18n.t('installer.step3.hostWarningsPopup.process'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.processes.message'),
        type: Em.I18n.t('common.process'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.processes'),
        action: Em.I18n.t('installer.step3.hostWarningsPopup.action.running'),
        category: 'process'
      }),
      Em.Object.create({
        warnings: categoryWarnings.filterProperty('category', 'packages'),
        title: Em.I18n.t('installer.step3.hostWarningsPopup.package'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.packages.message'),
        type: Em.I18n.t('common.package'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.packages'),
        action: Em.I18n.t('installer.step3.hostWarningsPopup.action.installed'),
        category: 'package'
      }),
      Em.Object.create({
        warnings: categoryWarnings.filterProperty('category', 'fileFolders'),
        title: Em.I18n.t('installer.step3.hostWarningsPopup.fileAndFolder'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.fileFolders.message'),
        type: Em.I18n.t('common.path'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.filesAndFolders'),
        action: Em.I18n.t('installer.step3.hostWarningsPopup.action.exists'),
        category: 'fileFolders'
      }),
      Em.Object.create({
        warnings: categoryWarnings.filterProperty('category', 'services'),
        title: Em.I18n.t('installer.step3.hostWarningsPopup.service'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.services.message'),
        type: Em.I18n.t('common.service'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.services'),
        action: Em.I18n.t('installer.step3.hostWarningsPopup.action.notRunning'),
        category: 'service'
      }),
      Em.Object.create({
        warnings: categoryWarnings.filterProperty('category', 'users'),
        title: Em.I18n.t('installer.step3.hostWarningsPopup.user'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.users.message'),
        type: Em.I18n.t('common.user'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.users'),
        action: Em.I18n.t('installer.step3.hostWarningsPopup.action.exists'),
        category: 'user'
      }),
      Em.Object.create({
        warnings: categoryWarnings.filterProperty('category', 'misc'),
        title: Em.I18n.t('installer.step3.hostWarningsPopup.misc'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.misc.message'),
        type: Em.I18n.t('installer.step3.hostWarningsPopup.misc.umask'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.empty.misc'),
        action: Em.I18n.t('installer.step3.hostWarningsPopup.action.exists'),
        category: 'misc'
      }),
      Em.Object.create({
        warnings: categoryWarnings.filterProperty('category', 'alternatives'),
        title: Em.I18n.t('installer.step3.hostWarningsPopup.alternatives'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.alternatives.message'),
        type: Em.I18n.t('installer.step3.hostWarningsPopup.alternatives.umask'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.alternatives.empty'),
        action: Em.I18n.t('installer.step3.hostWarningsPopup.action.exists'),
        category: 'alternatives'
      }),
      Em.Object.create({
        warnings: categoryWarnings.filterProperty('category', 'reverseLookup'),
        title: Em.I18n.t('installer.step3.hostWarningsPopup.reverseLookup'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.reverseLookup.message'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.reverseLookup.empty'),
        category: 'reverseLookup'
      }),
      Em.Object.create({
        warnings: hostCheckWarnings,
        title: Em.I18n.t('installer.step3.hostWarningsPopup.resolution.validation.name'),
        message: Em.I18n.t('installer.step3.hostWarningsPopup.resolution.validation.message'),
        type: Em.I18n.t('common.issues'),
        emptyName: Em.I18n.t('installer.step3.hostWarningsPopup.resolution.validation.empty'),
        action: Em.I18n.t('installer.step3.hostWarningsPopup.action.failed'),
        category: 'hostNameResolution'
      })
    ];
    warningsArray.forEach(function (warningType) {
      warningType.set('isCollapsed', true);
      warningType.get('warnings').forEach(function (warn) {
        var hosts = Em.get(warn, 'hosts');
        if (hosts) {
          var hostsList = hosts.length < 11 ? hosts.join('<br>') :
            hosts.slice(0, 10).join('<br>') + '<br> ' +
              Em.I18n.t('installer.step3.hostWarningsPopup.moreHosts').format(warn.hosts.length - 10);
          Em.set(warn, 'hostsList', hostsList);
        }
      });
    });
    return warningsArray;
  }.property('category', 'warningsByHost', 'bodyController.jdkCategoryWarnings', 'bodyController.repoCategoryWarnings', 'bodyController.diskCategoryWarnings', 'bodyController.hostCheckWarnings',  'bodyController.thpCategoryWarnings'),

  /**
   * Message with info about warnings
   * @return {string}
   */
  warningsNotice: function () {
    var issuesNumber = this.get('warnings.length') + this.get('bodyController.repoCategoryWarnings.length') + this.get('bodyController.diskCategoryWarnings.length')
      + this.get('bodyController.jdkCategoryWarnings.length') + this.get('bodyController.hostCheckWarnings.length') + this.get('bodyController.thpCategoryWarnings.length');
    this.set('totalWarningsCount', issuesNumber);
    var issues = issuesNumber + ' ' + (issuesNumber.length === 1 ? Em.I18n.t('installer.step3.hostWarningsPopup.issue') : Em.I18n.t('installer.step3.hostWarningsPopup.issues'));
    var hostsCnt = this.warningHostsNamesCount();
    var hosts = hostsCnt + ' ' + (hostsCnt === 1 ? Em.I18n.t('installer.step3.hostWarningsPopup.host') : Em.I18n.t('installer.step3.hostWarningsPopup.hosts'));
    return Em.I18n.t('installer.step3.hostWarningsPopup.summary').format(issues, hosts);
  }.property('warnings', 'warningsByHost', 'bodyController.jdkCategoryWarnings', 'bodyController.repoCategoryWarnings', 'bodyController.diskCategoryWarnings', 'bodyController.hostCheckWarnings',  'bodyController.diskCategoryWarnings'),

  /**
   * Detailed content to show it in new window
   * @return {string}
   */
  contentInDetails: function () {
    var content = this.get('content');
    var newContent = '';
    newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.header') + new Date;
    newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.hosts');
    newContent += this.get('hostNamesWithWarnings').join(' ');
    if (content.findProperty('category', 'thp').warnings.length) {
      newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.thp');
      newContent += content.findProperty('category', 'thp').warnings[0].hosts.join(' ');
    }
    if (content.findProperty('category', 'jdk').warnings.length) {
      newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.jdk');
      newContent += content.findProperty('category', 'jdk').warnings[0].hosts.join('<br>');
    }
    if (content.findProperty('category', 'disk').warnings.length) {
      newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.disk');
      newContent += content.findProperty('category', 'disk').warnings[0].hosts.join('<br>');
    }
    if (content.findProperty('category', 'repositories').warnings.length) {
      newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.repositories');
      newContent += content.findProperty('category', 'repositories').warnings[0].hosts.join('<br>');
    }
    if (content.findProperty('category', 'hostNameResolution').warnings.length) {
      newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.hostNameResolution');
      newContent += content.findProperty('category', 'hostNameResolution').warnings[0].hosts.join('<br>');
    }
    if (content.findProperty('category', 'firewall').warnings.length) {
      newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.firewall');
      newContent += content.findProperty('category', 'firewall').warnings.mapProperty('name').join('<br>');
    }
    if (content.findProperty('category', 'fileFolders').warnings.length) {
      newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.fileFolders');
      newContent += content.findProperty('category', 'fileFolders').warnings.mapProperty('name').join(' ');
    }
    if (content.findProperty('category', 'reverseLookup').warnings.length) {
      newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.reverseLookup');
      newContent += content.findProperty('category', 'reverseLookup').warnings[0].hosts.join(' ');
    }
    if (content.findProperty('category', 'process').warnings.length) {
      newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.process');
      content.findProperty('category', 'process').warnings.forEach(function (process, i) {
        process.hosts.forEach(function (host, j) {
          if (!!i || !!j) {
            newContent += ',';
          }
          newContent += '(' + host + ',' + process.user + ',' + process.pid + ')';
        });
      });
    }
    if (content.findProperty('category', 'package').warnings.length) {
      newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.package');
      newContent += content.findProperty('category', 'package').warnings.mapProperty('name').join(' ');
    }
    if (content.findProperty('category', 'service').warnings.length) {
      newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.service');
      newContent += content.findProperty('category', 'service').warnings.mapProperty('name').join(' ');
    }
    if (content.findProperty('category', 'user').warnings.length) {
      newContent += Em.I18n.t('installer.step3.hostWarningsPopup.report.user');
      newContent += content.findProperty('category', 'user').warnings.mapProperty('name').join(' ');
    }
    newContent += '</p>';
    return newContent;
  }.property('content', 'warningsByHost', 'bodyController.jdkCategoryWarnings', 'bodyController.repoCategoryWarnings', 'bodyController.diskCategoryWarnings', 'bodyController.hostCheckWarnings'),

  didInsertElement: function () {
    Em.run.next(this, function () {
      App.tooltip(this.$("[rel='HostsListTooltip']"), {html: true, placement: "right"});
      App.tooltip(this.$('#process .warning-name'), {html: true, placement: "top"});
    });
  }.observes('content'),

  /**
   * Show popup with selected hostnames
   * @param {object} hosts
   * @method showHostsPopup
   * @return {App.ModalPopup}
   */
  showHostsPopup: function (hosts) {
    $('.tooltip').hide();
    return App.ModalPopup.show({
      header: Em.I18n.t('installer.step3.hostWarningsPopup.allHosts'),
      bodyClass: Em.View.extend({
        hosts: hosts.context,
        template: Em.Handlebars.compile('<ul>{{#each host in view.hosts}}<li>{{host}}</li>{{/each}}</ul>')
      }),
      secondary: null
    });
  },

  /**
   * Open/Close selected category
   * @param {object} category
   * @method onToggleBlock
   */
  onToggleBlock: function (category) {
    this.$('#' + category.context.category).toggle('blind', 500);
    category.context.set("isCollapsed", !category.context.get("isCollapsed"));
  },

  /**
   * Generate number of hosts which had warnings, avoid duplicated host names in different warnings.
   * @method warningHostsNamesCount
   * @return {number}
   */
  warningHostsNamesCount: function () {
    var hostNameMap = Em.Object.create();
    if (Em.isNone(this.get('bodyController.warningsByHost'))) return 0;
    var warningsByHost = this.get('bodyController.warningsByHost').slice();
    warningsByHost.shift();
    warningsByHost.forEach(function (_host) {
      if (_host.warnings.length) {
        hostNameMap[_host.name] = true;
      }
    });
    var repoCategoryWarnings = this.get('bodyController.repoCategoryWarnings');
    var diskCategoryWarnings = this.get('bodyController.diskCategoryWarnings');
    var jdkCategoryWarnings = this.get('bodyController.jdkCategoryWarnings');
    var thpCategoryWarnings = this.get('bodyController.thpCategoryWarnings');
    var hostResolutionWarnings = this.get('bodyController.hostCheckWarnings');

    if (repoCategoryWarnings.length) {
      repoCategoryWarnings[0].hostsNames.forEach(function (_hostName) {
        if (!hostNameMap[_hostName]) {
          hostNameMap[_hostName] = true;
        }
      });
    }
    if (diskCategoryWarnings.length) {
      diskCategoryWarnings[0].hostsNames.forEach(function (_hostName) {
        if (!hostNameMap[_hostName]) {
          hostNameMap[_hostName] = true;
        }
      });
    }
    if (jdkCategoryWarnings && jdkCategoryWarnings.length) {
      jdkCategoryWarnings[0].hostsNames.forEach(function (_hostName) {
        if (!hostNameMap[_hostName]) {
          hostNameMap[_hostName] = true;
        }
      });
    }
    if (hostResolutionWarnings && hostResolutionWarnings.length) {
      hostResolutionWarnings[0].hostsNames.forEach(function (_hostName) {
        if (!hostNameMap[_hostName]) {
          hostNameMap[_hostName] = true;
        }
      });
    }
    if (thpCategoryWarnings.length) {
      thpCategoryWarnings[0].hostsNames.forEach(function (_hostName) {
        if (!hostNameMap[_hostName]) {
          hostNameMap[_hostName] = true;
        }
      });
    }
    this.set('hostNamesWithWarnings', Em.keys(hostNameMap));
    return Em.keys(hostNameMap).length;
  },

  /**
   * Open new browser tab with detailed content
   * @method openWarningsInDialog
   */
  openWarningsInDialog: function () {
    var newWindow = window.open('');
    var newDocument = newWindow.document;
    newDocument.write(this.get('contentInDetails'));
    newWindow.focus();
  }

});
});

require.register("views/wizard/step3/hostWarningPopupFooter_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');

App.WizardStep3HostWarningPopupFooter = Em.View.extend({

  templateName: require('templates/wizard/step3/step3_host_warning_popup_footer'),

  classNames: ['modal-footer', 'host-checks-update'],

  footerControllerBinding: 'App.router.wizardStep3Controller',

  progressWidth: function () {
    return 'width:' + this.get('footerController.checksUpdateProgress') + '%';
  }.property('footerController.checksUpdateProgress'),

  isUpdateInProgress: function () {
    return (this.get('footerController.checksUpdateProgress') > 0) &&
      (this.get('footerController.checksUpdateProgress') < 100);
  }.property('footerController.checksUpdateProgress'),

  updateStatusClass: function () {
    var status = this.get('footerController.checksUpdateStatus');
    if (status === 'SUCCESS') {
      return 'text-success';
    }
    else {
      if (status === 'FAILED') {
        return 'text-error';
      }
      else {
        return null;
      }
    }
  }.property('footerController.checksUpdateStatus'),

  updateStatus: function () {
    var status = this.get('footerController.checksUpdateStatus');
    if (status === 'SUCCESS') {
      return Em.I18n.t('installer.step3.warnings.updateChecks.success');
    } else if (status === 'FAILED') {
      return Em.I18n.t('installer.step3.warnings.updateChecks.failed');
    } else {
      return null;
    }
  }.property('footerController.checksUpdateStatus')

});

});

require.register("views/wizard/step3_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.WizardStep3View = App.TableView.extend({

  templateName: require('templates/wizard/step3'),

  /**
   * List of hosts
   * Same to <code>controller.hosts</code>
   * @type {Ember.Enumerable}
   */
  content:function () {
    return this.get('controller.hosts');
  }.property('controller.hosts.length'),

  /**
   * Message with info about registration result
   * @type {string}
   */
  message:'',

  /**
   * Text to link for hosts' warnings popup
   * @type {string}
   */
  linkText: '',

  /**
   * Registration status
   * @type {string}
   */
  status: '',

  /**
   * Active category
   * @type {string}
   */
  selectedCategory: function() {
    return this.get('categories').findProperty('isActive');
  }.property('categories.@each.isActive'),

  /**
   * Message about other registered hosts (not included in current registration)
   * @type {string}
   */
  registeredHostsMessage: '',

  /**
   * Number of visible hosts on page
   * @type {string}
   */
  displayLength: "25",

  /**
   * All checkboxes on page are checked
   * @type {bool}
   */
  pageChecked: false,

  /**
   * bootStatus category object
   * @type {Ember.Object}
   */
  categoryObject: Em.Object.extend({
    hostsCount: 0,
    label: function () {
      return "%@ (%@)".fmt(this.get('value'), this.get('hostsCount'));
    }.property('value', 'hostsCount'),
    isActive: false,
    itemClass: function () {
      return this.get('isActive') ? 'active' : '';
    }.property('isActive')
  }),

  /**
   * List of bootStatus categories
   * @type {categoryObject[]}
   */
  categories: function () {
    return [
      this.categoryObject.create({value: Em.I18n.t('common.all'), hostsBootStatus: 'ALL', isActive: true}),
      this.categoryObject.create({value: Em.I18n.t('installer.step3.hosts.status.installing'), hostsBootStatus: 'RUNNING'}),
      this.categoryObject.create({value: Em.I18n.t('installer.step3.hosts.status.registering'), hostsBootStatus: 'REGISTERING'}),
      this.categoryObject.create({value: Em.I18n.t('common.success'), hostsBootStatus: 'REGISTERED' }),
      this.categoryObject.create({value: Em.I18n.t('common.fail'), hostsBootStatus: 'FAILED', last: true })
    ];
  }.property(),

  didInsertElement: function () {
    this.get('controller').loadStep();
  },

  /**
   * Select checkboxes of hosts on page
   * @method onPageChecked
   */
  onPageChecked: function () {
    if (this.get('selectionInProgress')) return;
    this.get('pageContent').setEach('isChecked', this.get('pageChecked'));
  }.observes('pageChecked'),

  /**
   * Select checkboxes of all hosts
   * @method selectAll
   */
  selectAll: function () {
    this.get('content').setEach('isChecked', true);
  },

  /**
   * Reset checkbox of all hosts
   * @method unSelectAll
   */
  unSelectAll: function() {
    this.get('content').setEach('isChecked', false);
  },

  /**
   * Call <code>watchSelection</code> only once
   * @method watchSelectionOnce
   */
  watchSelectionOnce: function () {
    Em.run.once(this, 'watchSelection');
  }.observes('content.@each.isChecked', 'pageContent'),

  /**
   * Watch selection and calculate such flags as:
   * <ul>
   *  <li>noHostsSelected</li>
   *  <li>selectedHostsCount</li>
   *  <li>pageChecked</li>
   * </ul>
   * @method watchSelection
   */
  watchSelection: function() {
    this.set('selectionInProgress', true);
    this.set('pageChecked', !!this.get('pageContent.length') && this.get('pageContent').everyProperty('isChecked', true));
    this.set('selectionInProgress', false);
    var noHostsSelected = true;
    var selectedHostsCount = 0;
    this.get('content').forEach(function(host){
      selectedHostsCount += host.get('isChecked') ? 1 : 0;
      noHostsSelected = (noHostsSelected) ? !host.get('isChecked') : noHostsSelected;
    });
    this.set('noHostsSelected', noHostsSelected);
    this.set('selectedHostsCount', selectedHostsCount);
  },

  /**
   * Update <code>registeredHostsMessage</code> according to <code>controller.registeredHots.length</code>
   * @method setRegisteredHosts
   */
  setRegisteredHosts: function(){
    this.set('registeredHostsMessage',Em.I18n.t('installer.step3.warning.registeredHosts').format(this.get('controller.registeredHosts').length));
  }.observes('controller.registeredHosts'),

  /**
   * Call filters and counters one time
   * @method hostBootStatusObserver
   */
  hostBootStatusObserver: function(){
    Em.run.once(this, 'countCategoryHosts');
    Em.run.once(this, 'filter');
    Em.run.once(this, 'monitorStatuses');
  }.observes('content.@each.bootStatus'),

  /**
   * Calculate host count grouped by <code>bootStatus</code>
   * @method countCategoryHosts
   */
  countCategoryHosts: function () {
    var counters = {
      "RUNNING": 0,
      "REGISTERING": 0,
      "REGISTERED": 0,
      "FAILED": 0
    };
    this.get('content').forEach(function (host) {
      if (!Em.isNone(counters[host.get('bootStatus')])) {
        counters[host.get('bootStatus')]++;
      }
    }, this);
    counters["ALL"] = this.get('content.length');
    this.get('categories').forEach(function(category) {
      category.set('hostsCount', counters[category.get('hostsBootStatus')]);
    }, this);
  },


  /**
   * Filter hosts by category
   * @method filter
   */
  filter: function () {
    var self = this;
    Em.run.next(function () {
      self.doFilter();
    });
  }.observes('selectedCategory'),

  /**
   * Real filter-method
   * Called from <code>filter</code> in Em.run.next-wrapper
   * @method doFilter
   */
  doFilter: function() {
    var result = [];
    var selectedCategory = this.get('selectedCategory');
    if (!selectedCategory || selectedCategory.get('hostsBootStatus') === 'ALL') {
      result = this.get('content');
    } else {
      result = this.get('content').filterProperty('bootStatus', this.get('selectedCategory.hostsBootStatus'));
    }
    this.set('filteredContent', result);
  },

  /**
   * Trigger on Category click
   * @param {Object} event
   * @method selectCategory
   */
  selectCategory: function (event) {
    var categoryStatus = event.context.get('hostsBootStatus');
    this.get('categories').forEach(function (category) {
      category.set('isActive', (category.get('hostsBootStatus') === categoryStatus));
    });
    this.watchSelection();
  },

  /**
   * Select "All" hosts category
   * run registration of failed hosts again
   * @method retrySelectedHosts
   */
  retrySelectedHosts: function () {
    var eventObject = {context: Em.Object.create({hostsBootStatus: 'ALL'})};
    this.selectCategory(eventObject);
    this.get('controller').retrySelectedHosts();
  },

  /**
   * Update <code>status</code>, <code>linkText</code>, <code>message</code> according to hosts statuses
   * @method monitorStatuses
   */
  monitorStatuses: function() {
    var hosts = this.get('controller.bootHosts');
    var failedHosts = hosts.filterProperty('bootStatus', 'FAILED').length;

    if (hosts.length === 0) {
      this.set('status', 'alert-warn');
      this.set('linkText', '');
      this.set('message', Em.I18n.t('installer.step3.warnings.missingHosts'));
    }
    else {
      if (!this.get('controller.isWarningsLoaded')) {
        this.set('status', 'alert-info');
        this.set('linkText', '');
        this.set('message', Em.I18n.t('installer.step3.warning.loading'));
      }
      else {
        if (this.get('controller.isHostHaveWarnings') || this.get('controller.repoCategoryWarnings.length') || this.get('controller.diskCategoryWarnings.length') || this.get('controller.jdkCategoryWarnings.length')
          || this.get('controller.hostCheckWarnings.length') || this.get('controller.thpCategoryWarnings.length')) {
          this.set('status', 'alert-warn');
          this.set('linkText', Em.I18n.t('installer.step3.warnings.linkText'));
          this.set('message', Em.I18n.t('installer.step3.warnings.fails').format(hosts.length - failedHosts));
        }
        else {
          this.set('status', 'alert-success');
          this.set('linkText', Em.I18n.t('installer.step3.noWarnings.linkText'));
          if (failedHosts == 0) {
            // all are ok
            this.set('message', Em.I18n.t('installer.step3.warnings.noWarnings').format(hosts.length));
          }
          else {
            if (failedHosts == hosts.length) {
              // all failed
              this.set('status', 'alert-warn');
              this.set('linkText', '');
              this.set('message', Em.I18n.t('installer.step3.warnings.allFailed').format(failedHosts));
            }
            else {
              // some failed
              this.set('message', Em.I18n.t('installer.step3.warnings.someWarnings').format((hosts.length - failedHosts), failedHosts));
            }
          }
        }
      }
    }

  }.observes('controller.isWarningsLoaded','controller.hostCheckWarnings', 'controller.isHostHaveWarnings', 'controller.repoCategoryWarnings', 'controller.diskCategoryWarnings', 'controller.jdkCategoryWarnings', 'controller.thpCategoryWarnings')
});

//todo: move it inside WizardStep3View
App.WizardHostView = Em.View.extend({

  tagName: 'tr',

  classNameBindings: ['hostInfo.bootStatus'],

  /**
   * Host from parent view
   * @type {Object}
   */
  hostInfo: null,

  /**
   * @type {bool}
   */
  isRetryable: function() {
    // return ['FAILED'].contains(this.get('hostInfo.bootStatus'));
    return false;
  }.property('hostInfo.bootStatus'),

  /**
   * Remove selected host
   * @method remove
   */
  remove: function () {
    this.get('controller').removeHost(this.get('hostInfo'));
  },

  /**
   * Retry register selected host
   * @method retry
   */
  retry: function() {
    this.get('controller').retryHost(this.get('hostInfo'));
  }

});



});

require.register("views/wizard/step4_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.WizardStep4View = Em.View.extend({

  templateName: require('templates/wizard/step4')

});
});

require.register("views/wizard/step5_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var stringUtils = require('utils/string_utils');

App.WizardStep5View = Em.View.extend({

  templateName: require('templates/wizard/step5'),

  /**
   * If install more than 25 hosts, should use App.InputHostView for hosts selection
   * Otherwise - App.SelectHostView
   * @type {bool}
   */
  shouldUseInputs: function() {
    return this.get('controller.hosts.length') > 25;
  }.property('controller.hosts.length'),

  didInsertElement: function () {
    this.get('controller').loadStep();
    this.setCoHostedComponentText();
  },

  coHostedComponentText: '',

  setCoHostedComponentText: function () {
    var coHostedComponents = App.StackServiceComponent.find().filterProperty('isOtherComponentCoHosted').filterProperty('stackService.isSelected');
    var coHostedComponentsText = '';
    if (!this.get('controller').get('isReassignWizard')) {
      coHostedComponents.forEach(function (serviceComponent, index) {
        var coHostedComponentsDisplayNames = serviceComponent.get('coHostedComponents').map(function (item) {
          return App.StackServiceComponent.find().findProperty('componentName', item).get('displayName');
        });
        var componentTextArr = [serviceComponent.get('displayName')].concat(coHostedComponentsDisplayNames);
        coHostedComponents[index] = stringUtils.getFormattedStringFromArray(componentTextArr);
        coHostedComponentsText += '<br/>' + Em.I18n.t('installer.step5.body.coHostedComponents').format(coHostedComponents[index]);
      }, this);
    }

    this.set('coHostedComponentText', coHostedComponentsText);
  }
});

App.InputHostView = Em.TextField.extend(App.SelectHost, {

  attributeBindings: ['disabled'],
  /**
   * Saved typeahead component
   * @type {$}
   */
  typeahead: null,

  /**
   * When <code>value</code> (host_info) is changed this method is triggered
   * If new hostname is valid, this host is assigned to master component
   * @method changeHandler
   */
  changeHandler: function() {
    if (!this.shouldChangeHandlerBeCalled()) return;
    var host = this.get('controller.hosts').findProperty('host_name', this.get('value'));
    if (Em.isNone(host)) {
      this.get('controller').updateIsHostNameValidFlag(this.get("component.component_name"), this.get("component.serviceComponentId"), false);
      return;
    }
    this.get('controller').assignHostToMaster(this.get("component.component_name"), host.get('host_name'), this.get("component.serviceComponentId"));
    this.tryTriggerRebalanceForMultipleComponents();
  }.observes('controller.hostNameCheckTrigger'),

  didInsertElement: function () {
    this.initContent();
    var value = this.get('content').findProperty('host_name', this.get('component.selectedHost')).get('host_name');
    this.set("value", value);
    var content = this.get('content').mapProperty('host_info'),
      self = this,
      updater = function (item) {
        return self.get('content').findProperty('host_info', item).get('host_name');
      },
      typeahead = this.$().typeahead({items: 10, source: content, updater: updater, minLength: 0});
    typeahead.on('blur', function() {
      self.change();
    }).on('keyup', function(e) {
        self.set('value', $(e.currentTarget).val());
        self.change();
      });
    this.set('typeahead', typeahead);
  },

  /**
   * Extract hosts from controller,
   * filter out available to selection and
   * push them into Em.Select content
   * @method initContent
   */
  initContent: function () {
    this._super();
    this.updateTypeaheadData(this.get('content').mapProperty('host_info'));
  },

  /**
   * Update <code>source</code> property of <code>typeahead</code> with a new list of hosts
   * @param {string[]} hosts
   * @method updateTypeaheadData
   */
  updateTypeaheadData: function(hosts) {
    if (this.get('typeahead')) {
      this.get('typeahead').data('typeahead').source = hosts;
    }
  }

});

App.SelectHostView = Em.Select.extend(App.SelectHost, {

  attributeBindings: ['disabled'],

  didInsertElement: function () {
    this.initContent();
    this.set("value", this.get("component.selectedHost"));
    App.popover($("[rel=popover]"), {'placement': 'right', 'trigger': 'hover'});
  },

  /**
   * Handler for selected value change
   * @method change
   */
  changeHandler: function () {
    if (!this.shouldChangeHandlerBeCalled()) return;
    this.get('controller').assignHostToMaster(this.get("component.component_name"), this.get("value"), this.get("component.serviceComponentId"));
    this.tryTriggerRebalanceForMultipleComponents();
  }.observes('controller.hostNameCheckTrigger'),

  /**
   * On click handler
   * @method click
   */
  click: function () {
    this.initContent();
  }

});

App.AddControlView = Em.View.extend({

  /**
   * Current component name
   * @type {string}
   */
  componentName: null,

  tagName: "span",

  classNames: ["badge", "badge-important"],

  template: Em.Handlebars.compile('+'),

  /**
   * Onclick handler
   * Add selected component
   * @method click
   */
  click: function () {
    this.get('controller').addComponent(this.get('componentName'));
  }
});

App.RemoveControlView = Em.View.extend({

  /**
   * Index for multiple component
   * @type {number}
   */
  serviceComponentId: null,

  /**
   * Current component name
   * @type {string}
   */
  componentName: null,

  tagName: "span",

  classNames: ["badge", "badge-important"],

  template: Em.Handlebars.compile('-'),

  /**
   * Onclick handler
   * Remove current component
   * @method click
   */
  click: function () {
    this.get('controller').removeComponent(this.get('componentName'), this.get("serviceComponentId"));
  }
});

});

require.register("views/wizard/step6_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.WizardStep6View = App.TableView.extend({

  templateName: require('templates/wizard/step6'),

  title: '',

  /**
   * Numbe rof visible rows
   * @type {string}
   */
  displayLength: "25",

  /**
   * List of hosts
   * @type {object[]}
   */
  content: function () {
    return this.get('controller.hosts');
  }.property('controller.hosts'),

  /**
   * Synonym to <code>content</code> in this <code>App.TableView</code>
   * @type {object[]}
   */
  filteredContent: function () {
    return this.get('content');
  }.property('content'),

  /**
   * Set <code>label</code>, <code>title</code> and do <code>loadStep</code>
   * @method didInsertElement
   */
  didInsertElement: function () {
    var controller = this.get('controller');
    this.set('title', Em.I18n.t('installer.step6.header'));
    this.setLabel();

    App.tooltip($('body'), {selector: '[rel=tooltip]'});
    controller.loadStep();
  },

  /**
   * Set <code>label</code> value
   * @method setLabel
   */
  setLabel: function () {
    var clients = this.get('controller.content.clients');
    var label = !!clients.length ? Em.I18n.t('installer.step6.body') +  Em.I18n.t('installer.step6.body.clientText') : Em.I18n.t('installer.step6.body');

    clients.forEach(function (_client) {
      if (clients.length === 1) {
        label = label + ' ' + _client.display_name;
      } else {
        if (_client !== clients[clients.length - 1]) {           // [clients.length - 1]
          label = label + ' ' + _client.display_name;
          if (_client !== clients[clients.length - 2]) {
            label = label + ',';
          }
        }
        else {
          label = label + ' ' + Em.I18n.t('and') + ' ' + _client.display_name + '.';
        }
      }
    }, this);
    this.set('label', label);
  },

  checkboxClick: function(e) {
    var checkbox = e.context;
    checkbox.toggleProperty('checked');
    this.get('controller').checkCallback(checkbox.component);
    this.get('controller').callValidation();
  },

  columnCount: function() {
    var hosts = this.get('controller.hosts');
    if  (hosts && hosts.length > 0) {
      var checkboxes = hosts[0].get('checkboxes');
      return checkboxes.length + 1;
    }
    return 1;
  }.property('controller.hosts.@each.checkboxes')
});

App.WizardStep6HostView = Em.View.extend({

  /**
   * Binded <code>host</code> object
   * @type {object}
   */
  host: null,

  tagName: 'td',

  /**
   * Create hover-labels for hostName with list of installed master-components
   * @method didInsertElement
   */
  didInsertElement: function () {
    var components = this.get('controller').getMasterComponentsForHost(this.get('host.hostName'));
    if (components && components.length > 0) {
      components = components.map(function (_component) {
        return App.format.role(_component);
      });
      components = components.join(" /\n");
      App.popover(this.$(), {
        title: Em.I18n.t('installer.step6.wizardStep6Host.title').format(this.get('host.hostName')),
        content: components,
        placement: 'right',
        trigger: 'hover'
      });
    }
  }
});

});

require.register("views/wizard/step7_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.WizardStep7View = Em.View.extend({
  templateName: require('templates/wizard/step7')
});

});

require.register("views/wizard/step8_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


var App = require('app');

App.WizardStep8View = Em.View.extend(App.wizardDeployProgressViewMixin, {

  templateName: require('templates/wizard/step8'),

  didInsertElement: function () {
    this.get('controller').loadStep();
  },

  /**
   * Print review-report
   * @method printReview
   */
  printReview: function () {
    var o = $("#step8-info");
    o.jqprint();
  },

  /**
   * Reference to modalPopup to make sure only one instance is created
   * @type {App.ModalPopup|null}
   */
  modalPopup: null
});


});

require.register("views/wizard/step9/hostLogPopupBody_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var date = require('utils/date');

App.WizardStep9HostLogPopupBodyView = Em.View.extend({

  templateName: require('templates/wizard/step9/step9HostTasksLogPopup'),

  /**
   * Does host lost heartbeat
   * @type {bool}
   */
  isHeartbeatLost: function() {
    return (this.get('parentView.host.status') === 'heartbeat_lost');
  }.property('parentView.host.status'),

  /**
   * Does host doesn't have scheduled tasks for install
   * @type {bool}
   */
  isNoTasksScheduled: function() {
    return this.get('parentView.host.isNoTasksForInstall');
  }.property('parentView.host.isNoTasksForInstall'),

  /**
   * Is log-box hidden
   * @type {bool}
   */
  isLogWrapHidden: true,

  /**
   * Is log-textarea visible
   * @type {bool}
   */
  showTextArea: false,

  /**
   * No tasks shown
   * @type {bool}
   */
  isEmptyList: true,

  /**
   * Is task logs loaded
   * @type {Boolean}
   */
  isTaskLoaded: true,

  /**
   * Checks if no visible tasks are in popup
   * @method visibleTasks
   */
  visibleTasks: function () {
    this.set("isEmptyList", true);
    if (this.get('category.value')) {
      var filter = this.get('category.value');
      var tasks = this.get('tasks');
      tasks.setEach("isVisible", false);

      if (filter == "all") {
        tasks.setEach("isVisible", true);
      }
      else if (filter == "pending") {
        tasks.filterProperty("status", "pending").setEach("isVisible", true);
        tasks.filterProperty("status", "queued").setEach("isVisible", true);
      }
      else if (filter == "in_progress") {
        tasks.filterProperty("status", "in_progress").setEach("isVisible", true);
      }
      else if (filter == "failed") {
        tasks.filterProperty("status", "failed").setEach("isVisible", true);
      }
      else if (filter == "completed") {
        tasks.filterProperty("status", "completed").setEach("isVisible", true);
      }
      else if (filter == "aborted") {
        tasks.filterProperty("status", "aborted").setEach("isVisible", true);
      }
      else if (filter == "timedout") {
        tasks.filterProperty("status", "timedout").setEach("isVisible", true);
      }

      if (tasks.filterProperty("isVisible", true).length > 0) {
        this.set("isEmptyList", false);
      }
    }
  }.observes('category', 'tasks'),

  /**
   * List categories (implements possible values for task status)
   * @type {Em.Object[]}
   */
  categories: [
    Em.Object.create({value: 'all', label: Em.I18n.t('installer.step9.hostLog.popup.categories.all') }),
    Em.Object.create({value: 'pending', label: Em.I18n.t('installer.step9.hostLog.popup.categories.pending')}),
    Em.Object.create({value: 'in_progress', label: Em.I18n.t('installer.step9.hostLog.popup.categories.in_progress')}),
    Em.Object.create({value: 'failed', label: Em.I18n.t('installer.step9.hostLog.popup.categories.failed') }),
    Em.Object.create({value: 'completed', label: Em.I18n.t('installer.step9.hostLog.popup.categories.completed') }),
    Em.Object.create({value: 'aborted', label: Em.I18n.t('installer.step9.hostLog.popup.categories.aborted') }),
    Em.Object.create({value: 'timedout', label: Em.I18n.t('installer.step9.hostLog.popup.categories.timedout') })
  ],

  /**
   * Current category
   * @type {Em.Object}
   */
  category: null,

  /**
   * List of tasks
   * @type {Em.Object[]}
   */
  tasks: function () {
    var tasksArr = [];
    var host = this.get('parentView.host');
    var tasks = this.getStartedTasks(host);
    tasks = tasks.sortProperty('Tasks.id');
    if (tasks.length) {
      tasks.forEach(function (_task) {
        var taskInfo = Em.Object.create({});
        taskInfo.set('id', _task.Tasks.id);
        taskInfo.set('requestId', _task.Tasks.request_id);
        taskInfo.set('command', _task.Tasks.command.toLowerCase() === 'service_check' ? '' : _task.Tasks.command.toLowerCase());
        taskInfo.set('status', App.format.taskStatus(_task.Tasks.status));
        taskInfo.set('role', App.format.role(_task.Tasks.role));
        taskInfo.set('stderr', _task.Tasks.stderr);
        taskInfo.set('stdout', _task.Tasks.stdout);
        taskInfo.set('outputLog', _task.Tasks.output_log);
        taskInfo.set('errorLog', _task.Tasks.error_log);
        taskInfo.set('startTime',  date.startTime(_task.Tasks.start_time));
        taskInfo.set('duration', date.durationSummary(_task.Tasks.start_time, _task.Tasks.end_time));
        taskInfo.set('isVisible', true);
        taskInfo.set('icon', '');
        taskInfo.set('hostName', _task.Tasks.host_name);
        if (taskInfo.get('status') == 'pending' || taskInfo.get('status') == 'queued') {
          taskInfo.set('icon', 'icon-cog');
        } else if (taskInfo.get('status') == 'in_progress') {
          taskInfo.set('icon', 'icon-cogs');
        } else if (taskInfo.get('status') == 'completed') {
          taskInfo.set('icon', 'icon-ok');
        } else if (taskInfo.get('status') == 'failed') {
          taskInfo.set('icon', 'icon-exclamation-sign');
        } else if (taskInfo.get('status') == 'aborted') {
          taskInfo.set('icon', 'icon-minus');
        } else if (taskInfo.get('status') == 'timedout') {
          taskInfo.set('icon', 'icon-time');
        }
        tasksArr.push(taskInfo);
      }, this);
    }
    this.set('isTaskLoaded', true);
    return tasksArr;
  }.property('parentView.c.logTasksChangesCounter'),

  /**
   * Navigate to task list from task view
   * @method backToTaskList
   */
  backToTaskList: function () {
    this.destroyClipBoard();
    this.set("isLogWrapHidden", true);
  },

  /**
   * Get list of host's started tasks
   * @param {object} host
   * @returns {object[]}
   * @method getStartedTasks
   */
  getStartedTasks: function (host) {
    return host.logTasks.filter(function (task) {
      return task.Tasks.status;
    });
  },

  /**
   * Open new window with task's log
   * @method openTaskLogInDialog
   */
  openTaskLogInDialog: function () {
    var newwindow = window.open();
    var newdocument = newwindow.document;
    newdocument.write($(".task-detail-log-info").html());
    newdocument.close();
  },

  /**
   * Currently open task
   * @type {Em.Object}
   */
  openedTask: function () {
    return this.get('tasks').findProperty('id', this.get('parentView.c.currentOpenTaskId'))
  }.property('parentView.c.currentOpenTaskId', 'tasks.@each'),

  /**
   * Click-handler for toggle task's log view (textarea to box and back)
   * @param {object} event
   * @method toggleTaskLog
   */
  toggleTaskLog: function (event) {
    if (this.get('isLogWrapHidden')) {
      var taskInfo = event.context;
      this.set("isLogWrapHidden", false);
      this.set('parentView.c.currentOpenTaskId', taskInfo.id);
      this.set('parentView.c.currentOpenTaskRequestId', taskInfo.requestId);
      this.set('isTaskLoaded', false);
      this.get('parentView.c').loadCurrentTaskLog();
      $(".modal").scrollTop(0);
      $(".modal-body").scrollTop(0);
    }
    else {
      this.set("isLogWrapHidden", true);
      this.set('parentView.c.currentOpenTaskId', 0);
      this.set('parentView.c.currentOpenTaskRequestId', 0);
    }
  },

  /**
   * Create (if doesn't exist) or destroy (if exists) clipboard textarea
   * @method textTrigger
   */
  textTrigger: function () {
    if ($(".task-detail-log-clipboard").length > 0) {
      this.destroyClipBoard();
    }
    else {
      this.createClipBoard();
    }
  },

  /**
   * Create clipboard with task's log
   * @method createClipBoard
   */
  createClipBoard: function () {
    var log = $(".task-detail-log-maintext");
    $(".task-detail-log-clipboard-wrap").html('<textarea class="task-detail-log-clipboard"></textarea>');
    $(".task-detail-log-clipboard")
      .html("stderr: \n" + $(".stderr").html() + "\n stdout:\n" + $(".stdout").html())
      .css("display", "block")
      .width(log.width())
      .height(log.height())
      .select();
    log.css("display", "none")
  },

  /**
   * Destroy clipboard with task's log
   * @method destroyClipBoard
   */
  destroyClipBoard: function () {
    $(".task-detail-log-clipboard").remove();
    $(".task-detail-log-maintext").css("display", "block");
  }
});
});

require.register("views/wizard/step9_view", function(exports, require, module) {
/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var App = require('app');
var date = require('utils/date');

App.WizardStep9View = App.TableView.extend({

  templateName: require('templates/wizard/step9'),

  /**
   * Overall progress-bar color
   * @type {string}
   */
  barColor: '',

  /**
   * Overall result message
   * @type {string}
   */
  resultMsg: '',

  /**
   * Overall message color
   * @type {string}
   */
  resultMsgColor: '',

  /**
   * When progress is 100, step is completed
   * @type {bool}
   */
  isStepCompleted: function() {
   return (this.get('controller.progress') === '100');
  }.property('controller.progress'),

  /**
   * Number of visible hosts
   * @type {string}
   */
  displayLength: "25",

  /**
   * Same to <code>controller.hosts</code>
   * @type {object[]}
   */
  content: function () {
    return this.get('controller.hosts');
  }.property('controller.hosts'),

  /**
   * Active category
   * @type {Ember.Object}
   */
  selectedCategory: function() {
    return this.get('categories').findProperty('isActive');
  }.property('categories.@each.isActive'),

  /**
   * Ember Object category. This object also contains
   * <code>
   *   hostStatus: {String} A valid status of a host.
   *               Used to filter hosts in that status.
   *   hostsCount: {Int} Dynamic count of hosts displayed in the category label
   *   label : {String} status and hosts in that status displayed which consists as a category on the page
   *   isActive: {boolean} Gets set when the category is selected/clicked by the user
   *   itemClass: {computed property} Binds the category link to active class when user clicks on the link
   * </code>
   */
  categoryObject: Em.Object.extend({
    hostsCount: 0,
    label: function () {
      return "%@ (%@)".fmt(this.get('value'), this.get('hostsCount'));
    }.property('value', 'hostsCount'),
    isActive: false,
    itemClass: function () {
      return this.get('isActive') ? 'active' : '';
    }.property('isActive')
  }),

  /**
   * Domputed property creates the category objects on the load of the page and sets 'All' as the active category
   * @Returns {Em.Object[]} All created categories which are binded and iterated in the template
   */
  categories: function () {
    return [
      this.categoryObject.create({value: Em.I18n.t('common.all'), hostStatus: 'all', isActive: true}),
      this.categoryObject.create({value: Em.I18n.t('installer.step9.hosts.status.label.inProgress'), hostStatus: 'inProgress'}),
      this.categoryObject.create({value: Em.I18n.t('installer.step9.hosts.status.label.warning'), hostStatus: 'warning'}),
      this.categoryObject.create({value: Em.I18n.t('common.success'), hostStatus: 'success'}),
      this.categoryObject.create({value: Em.I18n.t('common.fail'), hostStatus: 'failed', last: true })
    ];
  }.property(),

  /**
   * True if <code>controller.hostsWithHeartbeatLost</code> contains some values
   * @type {bool}
   */
  isHostHeartbeatLost: function () {
    return (this.get('controller.hostsWithHeartbeatLost').length > 0);
  }.property('controller.hostsWithHeartbeatLost.@each'),

  /**
   * Css-string to overall progress-bar width-property
   * @type {string}
   */
  barWidth: function () {
    var controller = this.get('controller');
    return 'width: ' + controller.get('progress') + '%;';
  }.property('controller.progress'),

  /**
   * Filter hosts info shown up on bottom of the box. Set by filter function, when 'seletedCategory' changed
   * @type {string}
   */
  filteredHostsInfo: '',

  /**
   * Message for overall progress
   * @type {string}
   */
  progressMessage: function () {
    return Em.I18n.t('installer.step9.overallProgress').format(this.get('controller.progress'));
  }.property('controller.progress'),

  /**
   * Run <code>countCategoryHosts</code>, <code>filter</code> only once
   * @method hostStatusObserver
   */
  hostStatusObserver: function(){
    Em.run.once(this, 'countCategoryHosts');
    Em.run.once(this, 'filter');
  }.observes('content.@each.status'),

  /**
   * Count each category hosts to update label
   * @method countCategoryHosts
   */
  countCategoryHosts: function () {
    var counters = {
      "info": 0,
      "pending": 0,
      "in_progress": 0,
      "heartbeat_lost": 0,
      "warning": 0,
      "success": 0,
      "failed": 0
    };

    this.get('content').forEach(function (host) {
      if (counters[host.get('status')] !== undefined) {
        counters[host.get('status')]++;
      }
    }, this);
    counters["all"] = this.get('content.length');
    counters["inProgress"] = counters["info"] + counters["pending"] + counters["in_progress"];
    counters["failed"] += counters["heartbeat_lost"];
    this.get('categories').forEach(function(category) {
      category.set('hostsCount', counters[category.get('hostStatus')]);
    }, this);
  },

  /**
   * Filter hosts by category
   * @method filter
   */
  filter: function () {
    var self = this;
    Em.run.next(function () {
      self.doFilter();
    });
  }.observes('selectedCategory'),

  /**
   * Real filter-method
   * Called from <code>filter</code> in Em.run.next-wrapper
   * @method doFilter
   */
  doFilter: function() {
    var result = [];
    var content = this.get('content');
    var selectedCategory = this.get('selectedCategory');
    if (!selectedCategory || selectedCategory.get('hostStatus') === 'all') {
      result = content;
    } else if (selectedCategory.get('hostStatus') == 'inProgress') {
      result = content.filter(function (_host) {
        return (_host.get('status') == 'info' || _host.get('status') == 'pending' || _host.get('status') == 'in_progress');
      });
    } else if (selectedCategory.get('hostStatus') == 'failed') {
      result = content.filter(function (_host) {
        return (_host.get('status') == 'failed' || _host.get('status') == 'heartbeat_lost');
      });
    } else {
      result = content.filterProperty('status', selectedCategory.get('hostStatus'));
    }
    this.set('filteredContent', result);
    this.set('filteredHostsInfo', Em.I18n.t('installer.step9.hosts.filteredHostsInfo').format(result.get('length'), content.get('length')));
  },

  /**
   * On click handler for 'show all' link
   * @method showAllHosts
   */
  showAllHosts: function () {
    this.get('categories').forEach(function (category) {
      category.set('isActive', (category.get('hostStatus') === 'all'));
    });
  },

  /**
   * Trigger on Category click
   * @param {Object} event
   * @method selectCategory
   */
  selectCategory: function (event) {
    var categoryStatus = event.context.get('hostStatus');
    this.get('categories').forEach(function (category) {
      category.set('isActive', (category.get('hostStatus') === categoryStatus));
    });
  },

  didInsertElement: function () {
    this.onStatus();
    this.get('controller').navigateStep();
  },

  /**
   * Set <code>resultMsg</code>, <code>resultMsg</code>, <code>resultMsgColor</code> according to
   * <code>controller.status</code>, <code>controller.startCallFailed</code>, <code>isHostHeartbeatLost</code>
   * @method onStatus
   */
  onStatus: function () {
    if (this.get('controller.status') === 'info') {
      this.set('resultMsg', '');
      this.set('barColor', 'progress-info');
    } else if (this.get('controller.status') === 'warning') {
      this.set('barColor', 'progress-warning');
      this.set('resultMsg', Em.I18n.t('installer.step9.status.warning'));
      this.set('resultMsgColor', 'alert-warning');
    } else if (this.get('controller.status') === 'failed') {
      this.set('barColor', 'progress-danger');
      this.set('resultMsgColor', 'alert-error');
      console.log('TRACE: Inside error view step9');
      if (this.get('isHostHeartbeatLost')) {
        // When present requests succeeds but some host components are in UNKNOWN or INSTALL_FAILED state and
        // hosts are in HEARTBEAT_LOST state
        this.set('resultMsg', Em.I18n.t('installer.step9.status.hosts.heartbeat_lost').format(this.get('controller.hostsWithHeartbeatLost').length));
      } else if (this.get('controller.startCallFailed')) {
        this.set('resultMsg', Em.I18n.t('installer.step9.status.start.services.failed'));
      } else {
        this.set('resultMsg', Em.I18n.t('installer.step9.status.failed'));
      }
    } else if (this.get('controller.status') === 'success') {
      console.log('TRACE: Inside success view step9');
      this.set('barColor', 'progress-success');
      this.set('resultMsg', Em.I18n.t('installer.step9.status.success'));
      this.set('resultMsgColor', 'alert-success');
    }
  }.observes('controller.status', 'controller.startCallFailed','isHostHeartbeatLost'),

  /**
   * Show popup with info about failed hosts
   * @return {App.ModalPopup}
   * @method hostWithInstallFailed
   */
  hostWithInstallFailed: function () {
    var controller = this.get('controller');
    return App.ModalPopup.show({
      header: Em.I18n.t('installer.step9.host.heartbeat_lost.header'),
      classNames: ['sixty-percent-width-modal'],
      autoHeight: false,
      secondary: null,

      bodyClass: Em.View.extend({
        templateName: require('templates/wizard/step9/step9_install_host_popup'),
        c: controller,
        failedHosts: function () {
          return controller.get('hostsWithHeartbeatLost');
        }.property()
      })
    });
  }

});

App.HostStatusView = Em.View.extend({

  tagName: 'tr',

  /**
   * Current host
   * @type {Em.Object}
   */
  obj: null,

  /**
   * wizardStep9Controller
   * @type {App.WizardStep9Controller}
   */
  controller: null,

  /**
   * Progress-bar color for current host
   * @type {string}
   */
  barColor: '',

  /**
   * Css-string to progress-bar width-property
   * @type {string}
   */
  barWidth: function () {
    return 'width: ' + this.get('obj.progress') + '%;';
  }.property('obj.progress'),

  /**
   * Is current host failed
   * @type {bool}
   */
  isFailed: function () {
    return (this.get('isHostCompleted') && (this.get('obj.status') === 'failed' || this.get('obj.status') === 'heartbeat_lost'));
  }.property('obj.status', 'isHostCompleted'),

  /**
   * Is current host successfully installed
   * @type {bool}
   */
  isSuccess: function () {
    return (this.get('isHostCompleted') && this.get('obj.status') === 'success');
  }.property('obj.status', 'isHostCompleted'),

  /**
   * Current host has warnings
   * @type {bool}
   */
  isWarning: function () {
    return(this.get('isHostCompleted') && this.get('obj.status') === 'warning');
  }.property('obj.status', 'isHostCompleted'),

  /**
   * Current host completed all its tasks
   * @type {bool}
   */
  isHostCompleted: function () {
    return this.get('obj.progress') == 100;
  }.property('obj.progress'),

  didInsertElement: function () {
    this.onStatus();
  },

  /**
   * Set <code>barColor</code>, <code>obj.progress</code>, <code>obj.message</code> according to
   * <code>obj.status</code>, <code>obj.progress</code>, <code>controller.progress</code>
   * @method onStatus
   */
  onStatus: function () {
    if (this.get('obj.status') === 'info') {
      this.set('barColor', 'progress-info');
    } else if (this.get('obj.status') === 'warning') {
      this.set('barColor', 'progress-warning');
      if (this.get('obj.progress') === '100') {
        this.set('obj.message', Em.I18n.t('installer.step9.host.status.warning'));
      }
    } else if (this.get('obj.status') === 'failed') {
      this.set('barColor', 'progress-danger');
      if (this.get('obj.progress') === '100') {
        this.set('obj.message', Em.I18n.t('installer.step9.host.status.failed'));
      }
    } else if (this.get('obj.status') === 'heartbeat_lost') {
      this.set('barColor', 'progress-danger');
      if (this.get('obj.progress') === '100') {
        this.set('obj.message', Em.I18n.t('installer.step9.host.heartbeat_lost'));
      }
    } else if (this.get('obj.status') === 'success' && this.get('isHostCompleted') && parseInt(this.get('controller.progress')) > 34) {
        this.set('barColor', 'progress-success');
        this.set('obj.message', Em.I18n.t('installer.step9.host.status.success'));
    }
  }.observes('obj.status', 'obj.progress', 'controller.progress'),

  /**
   * Show popup with host logs
   * @return {App.ModalPopup}
   * @method hostLogPopup
   */
  hostLogPopup: function () {
    var controller = this.get('controller');
    var host = this.get('obj');

    return App.ModalPopup.show({

      header: host.get('name'),

      classNames: ['sixty-percent-width-modal'],

      autoHeight: false,

      secondary: null,

      /**
       * Current host
       * @type {Em.Object}
       */
      host: host,

      /**
       * wizardStep9Controller
       * @type {App.WizardStep9Controller}
       */
      c: controller,

      onClose: function () {
        this.set('c.currentOpenTaskId', 0);
        this.hide();
      },

      bodyClass: App.WizardStep9HostLogPopupBodyView
    });
  }

});
});

